program TestWriteAMDdata
  !  #[ Documentation
  !     a test program for the WriteAMDdata module
  !
  !     The program reads raw aux met data in ascii format
  !     and writes to EE-format.
  !
  !     written by: David Tan (ECMWF)
  !     created:    02-Feb-2006
  !
  !     Modifications:
  !     02-Feb-2006  D. Tan       Original version
  !     03-Feb-2006  D. Tan       Off_nadir datasets filled via convertor
  !     30-Mar-2006  D. Tan       Fill nadir datasets with copy of off-nadir.
  !     31-Mar-2006  D. Tan       Remove incorrect looping
  !     03-May-2006  J. de Kloe   Removed returns (these are not allowed 
  !                               in a main program)
  !     12-May-2006  D. Tan       Null use of ParKind1 module
  !     17-Aug-2006  P. Poli      Removed obsolete variables
  !     24-Apr-2007  J. de Kloe   phase out the SummaryVars code
  !     06-Jun-2007  P. Poli      changed num_BRC, added knmi_file,
  !                               num_knmi_profiles_to_skip_in, infile
  !                               added environment variables KNMINUMBRC,
  !                               INFILENOINC, KNMIFILENOINC
  !     28-Jun-2007  J. de Kloe   moved all code related to handling KNMI
  !                               ascii files to a new tool named
  !                               ConvertKnmiAscToAMD.F90
  !     08-Aug-2007  D. Tan       New interface to GetAuxiliaryData which
  !                               returns multiple profiles
  !     09-Aug-2007  D. Tan       Process Profile_Type flags
  !     10-Aug-2007  D. Tan       Pass Num_Model_Layers to GetAuxiliaryData
  !     19-Oct-2007  P. Poli      split AMD_DataType in two parts:
  !                               AMD_HDR_DataType and AMD_BRC_DataType
  !     26-Oct-2007  J. de Kloe   added a forgotten initialisation and a
  !                               trim() in a print command
  !     16-Jan-2008  J. de Kloe   phase out integer kind i_
  !     17-Jan-2008  J. de Kloe   removed a last few 64bit incompatibilities
  !
  !  #]
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE Numerics, only: r8_
  ! define the date and time types
  USE DateTimeMod, only: date_type, time_type, set_time, set_date
  ! define some parameters for verbosity control
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, VerbosityReport
  ! define a module to handle errors and program aborts
  USE ErrorHandler, only: no_error, program_abort, error_allocate, &
                          error_programming

  ! define a module to read ascii met data into a profile_type
  USE AuxiliaryModule, only: profile_type, GetAuxiliaryData, &
       TestPrintAuxiliaryData,        DeallocAuxiliaryData, &
       InitAuxiliaryModule, RawProfile_to_AMD_DataStructure

  USE EE_CFI_Datatypes, only: DateTimeType,SetDateTime, IntAus, IntAl

  ! define a module to hold the AMD_DataStructure
  USE AMD_DataStructure, only: AMD_HDR_DataType, AMD_BRC_DataType, &
                               Init_AMD_HDR_DataStructure, &
                               Init_AMD_BRC_DataStructure, &
                               Delete_AMD_HDR_DataStructure, &
                               Delete_AMD_BRC_DataStructure, &
                               Print_AMD_HDR_DataStructure, &
                               Print_AMD_BRC_DataStructure, &
                               Handle_AMDHeaders
  USE AMD_Geoloc_ADS,   only: Init_AMD_Geoloc_ADS
  USE AMD_Met_MDS,      only: Init_AMD_Met_MDS
  USE WriteAMDdata,     only: InitWriteAMDfileModule, WriteAMDfile

  !  #]
  !  #[ local variables and parameters

  implicit none

  ! Variables related to reading of ascii met file

  integer :: num_ecmwf_profiles
! TODO: make num_ecmwf_profiles a command line option.

!  integer        :: num_profiles_to_process  ! nr of profiles to process from in_file
! TODO: delete, or make num_profiles_to_process a command line option

  integer :: Num_model_layers
! TODO: make num_model_layers a command line option, and pass to GetAuxiliaryData.

  integer            :: nprofiles  ! nr of profiles read
  TYPE(profile_type), dimension(:), pointer :: profiles
  character(len=256) :: in_file ! filename to read aux data from

  character(len=256) :: text_string, conversion_message

  integer            :: Num_DSR1
  integer            :: Num_DSR2
  integer            :: Num_DSR

  integer :: current_dsr_DS1, current_dsr_DS2, &
             num_expected_L1B_obs, &
             num_profiles_L1B_Obs, num_missing_L1B_Obs, &
             num_computed_locations
  integer :: dsr_for_filling

  ! Variables related to output EE file

  integer            :: error_flag, AllocStatus

  ! P. Poli: modified this structure so that it can store several BRCs
  TYPE(AMD_BRC_DataType), dimension(:), allocatable :: AMD
  TYPE(AMD_HDR_DataType)                            :: AMD_Headers

  integer            :: i
  logical            :: fill_off_nadir_flag
  !  #]
  !  #[ program code

  print *,"starting TestWriteAMDdata"

  !*** initialisations ***
  error_flag = no_error
  nullify(profiles)

  call InitWriteAMDfileModule(error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"InitWriteAMDfileModule()")

  call InitAuxiliaryModule(error_flag)
  if (error_flag .ne. no_error) &
       call program_abort(error_flag,"InitAuxiliaryModule")

  ! set defaults to be used when no command line option is used to change it
  Num_model_layers   = 60
  Num_ecmwf_profiles = 38 ! 19
  in_file = "./infile.dat"
  num_expected_L1B_obs = 200 ! TODO: remove placeholder
  num_profiles_L1B_Obs = 0

  ! handle commandline arguments
  ! TODO: commandline arguments

  print *,"-------------------------------------"
  print *,'TestWriteAMDdata: assuming ',Num_model_layers,' model layers'
  print *,'TestWriteAMDdata: assuming ',Num_ecmwf_profiles,' ecmwf profiles'
  print *,'TestWriteAMDdata: using geolocation and met data from '//trim(in_file)
  print *,"-------------------------------------"

  ! Initialize counters
  current_dsr_DS1 = 0
  current_dsr_DS2 = 0
  Num_DSR1 = 0
  Num_DSR2 = 0
  num_computed_locations = 0

     call GetAuxiliaryData( in_file,             & ! input
                            num_model_layers,    & ! input
                            num_ecmwf_profiles,  & ! input
                            nprofiles,profiles,  & ! output
                            error_flag)            ! output
     
     if (error_flag .ne. no_error) &
          call program_abort(error_flag,"get_auxiliary_data")

  count_dsr_loop: do i=1,nprofiles
     !  #[
     if ( profiles(i)%OffNadir ) then
        Num_DSR1 = Num_DSR1 + 1
     else
        Num_DSR2 = Num_DSR2 + 1
     end if

     if ( profiles(i)%Computed_L1B_Location ) then
        num_computed_locations = num_computed_locations + 1
     else
        num_profiles_L1B_Obs = num_profiles_L1B_Obs + 1
     end if
  !  #]
  end do count_dsr_loop

  Num_DSR = max( Num_DSR1, Num_DSR2 )

  call Init_AMD_HDR_DataStructure(AMD_Headers, error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"Init_AMD_HDR_DataStructure")

  allocate(AMD(Num_DSR),stat=AllocStatus)
  IF (AllocStatus .ne. 0) &
       call program_abort(error_allocate,"TestWriteAMDdata")

  init_dsr_loop: do i=1,Num_DSR
     call Init_AMD_BRC_DataStructure(AMD(i), Num_Model_Layers, error_flag)
     IF (error_flag .ne. no_error) &
          call program_abort(error_flag,"Init_AMD_BRC_DataStructure")

     print *,'TestWriteAMDdata: initialized AMD(',i,')'

     if ( i .le. Num_DSR1 ) then
        call Init_AMD_Geoloc_ADS(AMD(i)%Geoloc_ADS_off_nadir, &
                                 error_flag)
        if (error_flag .ne. no_error) &
           call Program_Abort(error_flag,"Init_AMD_Geoloc_ADS(offnadir)")

        call Init_AMD_Met_MDS(AMD(i)%Met_MDS_off_nadir, &
                              Num_Model_Layers, error_flag)
        if (error_flag .ne. no_error) &
           call Program_Abort(error_flag,"Init_AMD_Met_MDS(offnadir)")
     end if

     if ( i .le. Num_DSR2 ) then
        call Init_AMD_Geoloc_ADS(AMD(i)%Geoloc_ADS_nadir, &
                                 error_flag)
        if (error_flag .ne. no_error) &
           call Program_Abort(error_flag,"Init_AMD_Geoloc_ADS(nadir)")

        call Init_AMD_Met_MDS(AMD(i)%Met_MDS_nadir, &
                              Num_Model_Layers, error_flag)
        if (error_flag .ne. no_error) &
           call Program_Abort(error_flag,"Init_AMD_Met_MDS(nadir)")
     end if

  end do init_dsr_loop

  profile_loop: do i=1,nprofiles
     !  #[

     !print *,'TestWriteAMDdata: Processing BRC #',i,' =============='

     ! print the result before conversion
     print *,"Pre-conversion result, profile number ", i
     call TestPrintAuxiliaryData(profiles(i))

     ! Check the OffNadir flag to know whether DS1 or DS2 should be filled
     fill_off_nadir_flag=profiles(i)%OffNadir
     if(fill_off_nadir_flag)then
       current_dsr_DS1 = current_dsr_DS1 + 1
       dsr_for_filling = current_dsr_DS1
     else
       current_dsr_DS2 = current_dsr_DS2 + 1
       dsr_for_filling = current_dsr_DS2
     endif

     write(text_string,'(I5)',err=997)i
     conversion_message = 'TestWriteAMDdata: converting profile '//&
                          trim(text_string)//' to AMD('
     write(text_string,'(I5)',err=996)dsr_for_filling
     conversion_message=trim(conversion_message)//trim(text_string)//')%'
     if(fill_off_nadir_flag)then
        conversion_message=trim(conversion_message)//'...DS1 (off_nadir)'
     else
        conversion_message=trim(conversion_message)//'...DS2 (nadir)'
     endif
     print *,trim(conversion_message)

     ! Transfer from raw profiles to AMD_DataStructure
     call RawProfile_to_AMD_DataStructure(fill_off_nadir_flag,profiles(i),&
                                          AMD(dsr_for_filling),error_flag)

     ! Print after conversion has been applied
     print *,"Post-conversion result:"
     call Print_AMD_BRC_DataStructure(AMD(dsr_for_filling), error_flag)
     IF (error_flag .ne. no_error) &
          & call program_abort(error_flag,"Print_AMD_BRC_DataStructure")

  !  #]
  end do profile_loop ! i

  ! Handle Headers
     num_missing_L1B_Obs = num_expected_L1B_obs - num_profiles_L1B_Obs ! current_dsr_DS2 ! TODO: remove placeholder

     print *,'TestWriteAMDdata: initializing AMD_Headers'
     call InitWriteAMDfileModule(error_flag)
     IF (error_flag .ne. no_error) &
          call program_abort(error_flag,"InitWriteAMDfile")

     call Handle_AMDHeaders(AMD_Headers, AMD(1), &
          error_flag, &
          int(profiles(1)%nlevels,IntAus), &
          int(num_DSR1,IntAl), &
          int(num_DSR2,IntAl), &
          int(num_profiles_L1B_Obs,IntAl), &
          int(num_missing_L1B_Obs,IntAl),  &
          int(num_computed_locations,IntAl), &
          switch_init=.true.)
     IF (error_flag .ne. no_error) &
          & call program_abort(error_flag,"Handle_AMDHeaders")
     print *,'TestWriteAMDdata: initialized AMD(1) Headers'

  ! Write AMD DataStructure to EE file

  print *,'TestWriteAMDdata: finished processing BRCs'
  print *,'=========================================='

  print *,'TestWriteAMDdata: calling WriteAMDfile'
  call WriteAMDfile(Num_DSR, AMD_Headers, AMD, error_flag)
  IF (error_flag .ne. no_error) call program_abort(error_flag,"WriteAMDfile")

  ! clean the used memory
  DO i=1,size(AMD)
     call Delete_AMD_BRC_DataStructure(AMD(i))
  END DO
  deallocate(AMD)
  call Delete_AMD_HDR_DataStructure(AMD_Headers)

  call DeallocAuxiliaryData(profiles)
  

  goto 111

996 print *,"ERROR: could not convert dsr_for_filling ",dsr_for_filling," to a string ..."
  call program_abort(error_programming,"TestWriteAMDdata")

997 print *,"ERROR: could not convert profile_number i=",i," to a string ..."
  call program_abort(error_programming,"TestWriteAMDdata")

  ! error handler for commandline handling problems
!TODO: 999 CALL Print_Usage
  !call program_abort(error_cmdline_options,"TestWriteAMDdata")

111 continue
  print *,"done"
  ! the end
  !  #]
end program TestWriteAMDdata
