module ReadAMDdata
  !  #[ Description:
  !
  !     a module to read the 2 files describing the auxiliary meteorological
  !     data product for use with Aeolus Level-2 processing (L2B and L2A)
  !
  !     created:     07-Mar-2006 by D. Tan, ECMWF
  !     based on:    ReadL2Bdata  by J. de Kloe, KNMI
  !
  !     Modifications:
  !     07-Mar-2006  D. Tan       Original version (incomplete)
  !     09-Mar-2006  D. Tan       Adapt to changes in the XML_module
  !     14-Mar-2006  D. Tan       Read KVT DSDs and requested Geoloc/Met DSR
  !     29-Mar-2006  D. Tan       Restore Compare in Read_MPH_from_KVT call.
  !     30-Mar-2006  D. Tan       Correct use of MDS2_DSD for nadir.
  !     07-Apr-2006  D. Tan       Remove debug prints
  !     21-Apr-2006  D. Tan       Explicit workaround=.false.
  !     25-Apr-2006  D. Tan       Upgrade for IODD V1.0 (Apr 2006)
  !     12-May-2006  D. Tan       Null use of ParKind1 module
  !     24-Jul-2006  D. Tan       Fill AMD%SPH from xml-tree
  !     17-Aug-2006  P. Poli      Fixed memory leak with AMD%SPH in reading,
  !                               removed obsolete variables
  !     10-Nov-2006  J. de Kloe   moved Load_All_AMD_Profiles from l2bp_module
  !                               to this module and moved Read_AMD_SPH_KVT 
  !                               routine to l2b_amd_sph.F90
  !     17-Apr-2007  D. Tan       Remove use_ECF switches.
  !     07-Jun-2007  J. de Kloe   adapted to changed xml interface
  !     18-Jun-2007  J. de Kloe   add a nullify statement
  !     19-Oct-2007  P. Poli      Split AMD_DataType in two parts:
  !                               AMD_HDR_DataType and AMD_BRC_DataType
  !                               Modified ReadAMDfile to allow using AMD 
  !                               header information from another datastructure
  !     12-Nov-2007 J. de Kloe    added check for xmlns attribute in xml file
  !     07-Dec-2007 J. de Kloe    adapted to changed Print_FH call
  !     16-Jan-2008 J. de Kloe    phase out integer kind i_
  !     17-Jan-2008 J. de Kloe    removed a last few 64bit incompatibilities
  !     21-Jan-2008 J. de Kloe    added (temporary) workaround switch in the
  !                               call to Read_MPH_from_KVT
  !     21-Feb-2008 J. de Kloe    removed the workaround again
  !  #]
  !  #[ Modules used:

  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE FixedHeader, only: FH_type, Print_FH, get_fh_from_xml_tree, &
                         file_type_AMD
  USE MainProductHeader, only: MPH_type, Print_MPH, &
                               Get_MPH_from_xml_tree, &
                               Read_MPH_from_KVT
  USE AMD_DataStructure, only: AMD_HDR_DataType, &
                               AMD_BRC_DataType, &
                               Init_AMD_HDR_DataStructure, &
                               Init_AMD_BRC_DataStructure
  USE AMD_Geoloc_ADS,    only: Init_AMD_Geoloc_ADS, Read_AMD_Geoloc_ADS, &
                               Print_AMD_Geoloc_ADS
  USE AMD_Met_MDS,       only: Init_AMD_Met_MDS, Read_AMD_Met_MDS, &
                               Print_AMD_Met_MDS
  USE L2B_AMD_SPH, only:       Print_L2B_AMD_SPH, &
                               Read_AMD_SPH_KVT, &
                               L2B_AMD_SPH_type, &
                               Get_L2B_AMD_SPH_from_xml_tree, &
                               AMD_DS_Name_GeoADS1, AMD_DS_Type_GeoADS1, &
                               AMD_DS_Name_GeoADS2, AMD_DS_Type_GeoADS2, &
                               AMD_DS_Name_MDS1,    AMD_DS_Type_MDS1, &
                               AMD_DS_Name_MDS2,    AMD_DS_Type_MDS2
  USE DataSetDescriptor, only: Init_DSD, Print_DSD, DSD_type, &
                               Read_DSD_from_KVT
  USE KVT_module, only: Init_KVT_module
  USE XML_module, only: xml_node_type, Init_XML, xml_load_file, &
                        xml_tree_print, delete_xml_tree, ee_h_name, &
                        xml_tree_get_namespace, uri_ns_header
  USE DirectBinaryIO, only: Init_DirectBinaryIO_module, &
                            BigEndian
  USE EE_CFI_Datatypes, only: IntAus, GetIntAus, IntAl, GetIntAl
  USE ErrorHandler, only: no_error, error_programming, &
                          error_opening_file, error_reading_file,&
                          error_writing_file, error_allocate
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE LunManager, only: get_lun, free_lun
  USE AeolusConstants, only: iodd_version_AuxMet
  USE Logging, only: logmsg,log_debug,log_info,log_progress,log_warn,log_error

  !  #]  
  !  #[ Parameter/Variable/Type declarations

  implicit none

  ! make all module contents private, unless explicitely stated otherwise
  private

  integer, parameter :: max_str_len = 512

  ! the default verbosity setting is Error (as silent as possible)
  integer, save :: LocalVerbosity = VerbosityError

  ! remember the initialisation state of the module
  logical, save :: module_is_initialised = .false.

  !  #]
  !  #[ Module interface
  public :: ReadAMDfile, InitReadAMDfileModule, Load_All_AMD_Profiles
  !  #]
contains
  !---------------------------------
  subroutine InitReadAMDfileModule(error_flag, ChoosenVerbosity)
    !  #[

    integer,           intent(out) :: error_flag
    integer, optional, intent(in)  :: ChoosenVerbosity

    error_flag = no_error

    IF (present(ChoosenVerbosity)) LocalVerbosity = ChoosenVerbosity

    IF (LocalVerbosity .ge. VerbosityReport) &
         call print_verbosity_setting(LocalVerbosity,ModuleName="ReadAMDdata")

    call Init_KVT_module()

    call Init_DirectBinaryIO_module(error_flag,FileByteOrder=BigEndian)
    if (error_flag .ne. no_error) return

    call Init_XML(error_flag) !(ChoosenVerbosity=VerbosityReport)
    if (error_flag .ne. no_error) return

    module_is_initialised = .true.

  end subroutine InitReadAMDfileModule
    !  #]
  subroutine ReadAMDfile(AMD_Headers, filename, error_flag, AMD, Which_DSR, &
       ReadAll            , ReadHeaders,      &
       ReadGeoloc_offnadir, ReadGeoloc_nadir, &
       ReadMet_offnadir   , ReadMet_nadir )
    !  #[
    type(AMD_HDR_DataType),intent(inout) :: AMD_Headers
    character(len=*),      intent(in)    :: filename ! without extension
    integer,               intent(out)   :: error_flag
    TYPE(AMD_BRC_DataType), optional, intent(inout) :: AMD
    integer    , optional, intent(in)    :: Which_DSR
    logical,     optional, intent(in)    :: ReadAll ! headers and datasets
    logical,     optional, intent(in)    :: ReadHeaders
    logical,     optional, intent(in)    :: ReadGeoloc_offnadir
    logical,     optional, intent(in)    :: ReadGeoloc_nadir
    logical,     optional, intent(in)    :: ReadMet_offnadir
    logical,     optional, intent(in)    :: ReadMet_nadir

    ! local variables
    character(len=max_str_len) :: filename_hdr
    character(len=max_str_len) :: filename_dbl
    logical                    :: do_ReadHeaders

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"Error in ReadAMDfile():"
       print *,"the ReadAMDdata module must be initialised before use ..."
       error_flag = error_programming
       return
    END IF

    filename_hdr = trim(filename)//".HDR"
    filename_dbl = trim(filename)//".DBL"

    do_ReadHeaders = .false.
    IF (present(ReadHeaders)) do_ReadHeaders = ReadHeaders
    IF (present(ReadAll)) THEN
       IF (ReadAll) do_ReadHeaders = .true.
    END IF

    IF (do_ReadHeaders) THEN
       ! needs to be called just once per set of HDR/DBL files
       call ReadAMDfile_HDR(AMD_Headers, filename_hdr, error_flag)
       IF (error_flag .ne. no_error) return
! to add: check if the filename is consistent with the file contents
    END IF

    ! needs to be called for every BRC present in a set of HDR/DBL files
    call ReadAMDfile_DBL(AMD_Headers, filename_dbl, error_flag, AMD, Which_DSR, &
         ReadAll, ReadHeaders, &
         ReadGeoloc_offnadir, ReadGeoloc_nadir, &
         ReadMet_offnadir   , ReadMet_nadir )
    IF (error_flag .ne. no_error) return

    return

  end subroutine ReadAMDfile
    !  #]
  !---------------------------------
  ! private routines
  subroutine ReadAMDfile_HDR(AMD_Headers, filename_hdr, error_flag)
    !  #[

    type(AMD_HDR_DataType), intent(inout)  :: AMD_Headers
    character(len=*),       intent(in)     :: filename_hdr
    integer,                intent(out)    :: error_flag

    ! local variables
    TYPE(xml_node_type), pointer :: xml_tree
    character(len=256)           :: path
    character(len=256)           :: xml_namespace_AMD
    character(len=256)           :: xml_namespace_uri

    ! init
    error_flag = no_error

! todo: add check to see if: AMD_Headers%FH%File_Name .eq. filename

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in ReadAMDfile:"
       print *,"the ReadAMDdata module should be initialised before use !"
       error_flag = error_programming
       return
    END IF

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"reading data from AMD header file: ",trim(filename_hdr)

    nullify(xml_tree)
    call xml_load_file(filename_hdr, xml_tree, error_flag)
    IF (error_flag .ne. no_error) return

    ! define which xml namespace should be present
    !example: xmlns="http://www.esa.int/schemas/ae/ALD_U_N_2B_01.01"
    !uri_ns_header    ! defined in xml_module
    !file_type_L2B_N  ! defined in FixedHeader
    !iodd_version_L2B ! defined in AeolusConstants
    xml_namespace_AMD = trim(uri_ns_header)//trim(file_type_AMD)//&
                        "_"//trim(iodd_version_AuxMet)

    ! check whether this namespace is really present in the input file
    call xml_tree_get_namespace(xml_tree,xml_namespace_uri,error_flag)
    if (error_flag .ne. no_error) then
       ! allow this attribute to be missing for now, to allow use
       ! of outdated files.
       ! TODO: remove this exception
       error_flag = no_error
       call logmsg(log_warn,"WARNING in ReadAMDfile_HDR: "//&
                   "could not find xmlns attribute in the root tag "//&
                   "of file: "//trim(filename_hdr))
    else
       if (trim(xml_namespace_AMD) .ne. trim(xml_namespace_uri)) &
            call logmsg(log_warn,"WARNING in ReadAMDfile_HDR: "//&
                        "expected an xmlns attribute with content: "//&
                        trim(xml_namespace_AMD)//", but found an "//&
                        "xmlns attribute with content: "//&
                        trim(xml_namespace_uri)//". This input file may "//&
                        "not be compatible with the current software!")
    end if


    call logmsg(log_debug,"ReadAMDfile_HDR: the xml file was "//&
                          "successfully read")

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"==================="
       call xml_tree_print(xml_tree, error_flag)
       IF (error_flag .ne. no_error) return
       print *,"==================="
    END IF
 
    !------------------------------
    ! copying the FH XML elements
    !------------------------------
    print *,"copying items to the FH datastructure"
    path = "/"//ee_h_name
    call Get_FH_from_xml_tree(xml_tree, path, AMD_Headers%FH, error_flag)
    IF (error_flag .ne. no_error) return
    AMD_Headers%FH_is_available = .true.

    !------------------------------
    ! copying the MPH XML elements
    !------------------------------
    print *,"copying items to the MPH datastructure"
    call Get_MPH_from_xml_tree(xml_tree, path, AMD_Headers%MPH, error_flag)
    IF (error_flag .ne. no_error) return
    AMD_Headers%MPH_is_available = .true.

    !------------------------------
    ! copying the SPH XML elements (with DSD's)
    !------------------------------
    print *,"copying items to the AMD_Headers%SPH datastructure"
    call Get_L2B_AMD_SPH_from_xml_tree(xml_tree, path, AMD_Headers%SPH, error_flag)
    IF (error_flag .ne. no_error) return
    AMD_Headers%SPH_is_available = .true.

    !------------------------------
    ! done                         
    !------------------------------

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"AMD_Headers%FH%..."
       call Print_FH(AMD_Headers%FH,error_flag)
       IF (error_flag .ne. no_error) return
       print *,"AMD_Headers%MPH%..."
       call Print_MPH(AMD_Headers%MPH,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"deallocating xml items"
    call delete_xml_tree(xml_tree, error_flag)
    IF (error_flag .ne. no_error) return

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"done with reading AMD header file"

    return

  end subroutine ReadAMDfile_HDR

    !  #]
  subroutine ReadAMDfile_DBL(AMD_Headers, filename_dbl, error_flag, AMD, Which_DSR, &
       ReadAll, ReadHeaders, ReadGeoloc_offnadir, ReadGeoloc_nadir, &
         ReadMet_offnadir   , ReadMet_nadir )
    !  #[ read all KVT and binary values for one BRC

    TYPE(AMD_HDR_DataType),intent(inout) :: AMD_Headers
    character(len=*),      intent(in)    :: filename_dbl
    integer,               intent(out)   :: error_flag
    TYPE(AMD_BRC_DataType), optional, intent(inout) :: AMD
    integer,     optional, intent(in)    :: Which_DSR
    logical,     optional, intent(in)    :: ReadAll ! headers and datasets
    logical,     optional, intent(in)    :: ReadHeaders
    logical,     optional, intent(in)    :: ReadGeoloc_offnadir
    logical,     optional, intent(in)    :: ReadGeoloc_nadir
    logical,     optional, intent(in)    :: ReadMet_offnadir
    logical,     optional, intent(in)    :: ReadMet_nadir

    ! local variables
!    character(len=256) :: exp_ds_name ! never used
!    character(len=256) :: exp_ds_type ! never used
    integer            :: fileunit, AllocStatus !, N_Meas_Max, N_range_gates
    integer            :: DataSetRecord_to_read
!    integer            :: filepos ! never used
!    integer            :: DS_distance ! never used
!    logical            :: order_as_expected ! never used
!    logical            :: distance_check_ok ! never used
!    logical            :: size_check_ok ! never used
    ! switches to copy the optional input switches to
    logical            :: do_ReadHeaders
    logical            :: do_ReadGeoloc_offnadir
    logical            :: do_ReadGeoloc_nadir
    logical            :: do_ReadMet_offnadir
    logical            :: do_ReadMet_nadir

! remark: expected_size_Geoloc_DSR_offnadir is longer than 
! 31 characters, so not allowed in standard Fortran90

!    integer :: exp_size_Geoloc_DSR_offnadir   ! never used
!    integer :: expected_size_Geoloc_DSR_nadir ! never used
!    integer :: expected_size_Met_DSR_offnadir ! never used
!    integer :: expected_size_Met_DSR_nadir    ! never used

!    logical, optional, intent(in)  :: Compare
    logical :: Compare

    integer(IntAus)  :: N_Model_Layers

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in ReadAMDfile_DBL():"
       print *,"the ReadAMDdata module should be initialised before use !"
       error_flag = error_programming
       return
    END IF

    DataSetRecord_to_read = 1 ! by default just read the 1st dataset
    IF (present(Which_DSR)) THEN
       DataSetRecord_to_read = Which_DSR
    END IF

    do_ReadHeaders         = .false.
    do_ReadGeoloc_offnadir = .false.
    do_ReadGeoloc_nadir    = .false.
    do_ReadMet_offnadir    = .false.
    do_ReadMet_nadir       = .false.

    IF (present(ReadAll)) THEN
       do_ReadHeaders         = .true.
       do_ReadGeoloc_offnadir = .true.
       do_ReadGeoloc_nadir    = .true.
       do_ReadMet_offnadir    = .true.
       do_ReadMet_nadir       = .true.
    END IF

    IF (present(ReadHeaders         )) do_ReadHeaders          = ReadHeaders
    IF (present(ReadGeoloc_offnadir )) do_ReadGeoloc_offnadir  = ReadGeoloc_offnadir
    IF (present(ReadGeoloc_nadir    )) do_ReadGeoloc_nadir     = ReadGeoloc_nadir
    IF (present(ReadMet_offnadir    )) do_ReadMet_offnadir     = ReadMet_offnadir
    IF (present(ReadMet_nadir       )) do_ReadMet_nadir        = ReadMet_nadir

    ! if nothing was selected to be read, just read everything
    IF ( .not. do_ReadHeaders         .and. &
         .not. do_ReadGeoloc_offnadir .and. &
         .not. do_ReadGeoloc_nadir    .and. &
         .not. do_ReadMet_offnadir    .and. &
         .not. do_ReadMet_nadir       ) THEN
       do_ReadHeaders = .true.
       do_ReadGeoloc_offnadir = .true.
       do_ReadGeoloc_nadir    = .true.
       do_ReadMet_offnadir    = .true.
       do_ReadMet_nadir       = .true.
    END IF

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"reading data from AMD datafile: ",trim(filename_dbl)

    !------------------------------
    ! open the binaryfile
    !------------------------------
    fileunit = get_lun()
    IF (fileunit .lt. 0) THEN
       error_flag = error_opening_file
       print *,"ERROR in ReadAMDfile_DBL():"
       print *,"no more free fileunits seem available ..."
       error_flag = error_reading_file
       return
    ENDIF
    
    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"opening file:"
       print *,trim(filename_dbl)
    END IF

    open(unit=fileunit,file=filename_dbl,status="old",&
         action="read",access="direct",recl=1,form="UNFORMATTED",ERR=999)
    
    IF (do_ReadHeaders) THEN
       !  #[

       !------------------------------
       ! allocate memory for the MPH and read it
       !------------------------------
       IF (.not. AMD_Headers%MPH_is_available) THEN

          ! the MPH is allocated and filled when reading the XML file
          ! so this part of the code is usually skipped.
          
          IF (.not. associated(AMD_Headers%MPH)) THEN
             print *,"Error in ReadAMDfile_DBL:"
             print *,"AMD_Headers%MPH seems not associated !"
             print *,"this should have been done in Init_AMD_HDR_DataStructure"
             error_flag = error_programming
             return
          END IF

          call Read_MPH_from_KVT(AMD_Headers%MPH, fileunit, error_flag)
          IF (error_flag .ne. no_error) return

          AMD_Headers%MPH_is_available = .true.
       ELSE
          ! Poli: modified call. 
          ! Now replace the values in memory by what is read.
          ! read the KVT values anyway, and compare them to what was
          ! read from the XML file

          call Read_MPH_from_KVT(AMD_Headers%MPH, fileunit, error_flag,&
                                  Compare=.true.)
!                                  CompareOnly=.true.)
          IF (error_flag .ne. no_error) return
       END IF

       ! an extra check
       ! this check is only possible if the FH is already loaded 
       IF (AMD_Headers%FH_is_available) THEN
          IF (associated(AMD_Headers%FH)) THEN
             IF (trim(AMD_Headers%MPH%PRODUCT) .ne. trim(AMD_Headers%FH%File_Name)) THEN
                print *,"ERROR in ReadAMDfile_DBL:"
                print *,"FH File name seems not to match with MPH productname!"
                print *,"AMD_Headers%FH%File_Name = ",trim(AMD_Headers%FH%File_Name)
                print *,"AMD_Headers%MPH%PRODUCT  = ",trim(AMD_Headers%MPH%PRODUCT)
                error_flag = error_reading_file
                return
             END IF
          END IF
       END IF

       !------------------------------
       ! allocate memory for the SPH and read it
       !------------------------------
       IF (.not. AMD_Headers%SPH_is_available) THEN

          IF (.not. associated(AMD_Headers%SPH)) THEN
             print *,"Error in ReadAMDfile_DBL:"
             print *,"AMD_Headers%SPH seems not associated !"
             print *,"this should have been done in Init_AMD_HDR_DataStructure"
             error_flag = error_programming
                return
             END IF

          call Read_AMD_SPH_KVT(AMD_Headers%SPH, fileunit, error_flag)
          IF (error_flag .ne. no_error) return
       
          AMD_Headers%SPH_is_available = .true.
       else
          ! reread the KVT items and compare then
          ! (not yet implemented, TODO: ) with the values read from xml
          ! this MUST be done here, otherwise the filepointer in the binary
          ! file stays at the start of the SPH block and the reading of 
          ! the DSD's below will fail...
          call Read_AMD_SPH_KVT(AMD_Headers%SPH, fileunit, error_flag)
          IF (error_flag .ne. no_error) return
       END IF

       !------------------------------
       ! read and check the 4 DSD's 
       !------------------------------
       
       !  #[

       !------------------------------
       ! allocate memory for the DSD1 and read it
       !------------------------------

! Adapt the RBC_SPH module
! If successful then TODO: Move to AMD_SPH module (currently L2B_AMD_SPH module)
    IF (.not. associated(AMD_Headers%SPH%GeoADS1_DSD)) THEN
       ! allow this, to enable the possibility of reading the dbl file only,
       ! without reading the HDR files.
       allocate(AMD_Headers%SPH%GeoADS1_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine ReadAMDfile_DBL/Read_AMD_SPH_from_KVT()"
          error_flag = error_allocate
          goto 998
       END IF
       call Init_DSD(AMD_Headers%SPH%GeoADS1_DSD,error_flag)
       IF (error_flag .ne. no_error) goto 998

       ! Read the AMD_Headers%GeoADS1_DSD without comparing
       call Read_DSD_from_KVT(AMD_Headers%SPH%GeoADS1_DSD,fileunit, &
                              AMD_DS_Name_GeoADS1, &
                              AMD_DS_Type_GeoADS1, error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the AMD_Headers%GeoADS1_DSD and compare if needed
       Compare = .false. ! TODO: Bring in line with RBC_SPH module
       call Read_DSD_from_KVT(AMD_Headers%SPH%GeoADS1_DSD,fileunit,&
                              AMD_DS_Name_GeoADS1, &
                              AMD_DS_Type_GeoADS1, error_flag,Compare)
       IF (error_flag .ne. no_error) goto 998
    END IF

       
       !------------------------------
       ! allocate memory for the DSD2 and read it
       !------------------------------
! Adapt the RBC_SPH module
! If successful then TODO: Move to AMD_SPH module (currently L2B_AMD_SPH module)
    IF (.not. associated(AMD_Headers%SPH%GeoADS2_DSD)) THEN
       ! allow this, to enable the possibility of reading the dbl file only,
       ! without reading the HDR files.
       allocate(AMD_Headers%SPH%GeoADS2_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine ReadAMDfile_DBL/Read_AMD_SPH_from_KVT()"
          error_flag = error_allocate
          goto 998
       END IF
       call Init_DSD(AMD_Headers%SPH%GeoADS2_DSD,error_flag)
       IF (error_flag .ne. no_error) goto 998

       ! Read the AMD_Headers%GeoADS2_DSD without comparing
       call Read_DSD_from_KVT(AMD_Headers%SPH%GeoADS2_DSD,fileunit,&
                              AMD_DS_Name_GeoADS2, &
                              AMD_DS_Type_GeoADS2, error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the AMD_Headers%GeoADS2_DSD and compare if needed
       Compare = .false. ! TODO: Bring in line with RBC_SPH module
       call Read_DSD_from_KVT(AMD_Headers%SPH%GeoADS2_DSD,fileunit,&
                              AMD_DS_Name_GeoADS2, &
                              AMD_DS_Type_GeoADS2, error_flag,Compare)
       IF (error_flag .ne. no_error) goto 998
    END IF

       
       !------------------------------
       ! allocate memory for the DSD3 and read it
       !------------------------------
! Adapt the RBC_SPH module
! If successful then TODO: Move to AMD_SPH module (currently L2B_AMD_SPH module)
    IF (.not. associated(AMD_Headers%SPH%MDS1_DSD)) THEN
       ! allow this, to enable the possibility of reading the dbl file only,
       ! without reading the HDR files.
       allocate(AMD_Headers%SPH%MDS1_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine ReadAMDfile_DBL/Read_AMD_SPH_from_KVT()"
          error_flag = error_allocate
          goto 998
       END IF
       call Init_DSD(AMD_Headers%SPH%MDS1_DSD,error_flag)
       IF (error_flag .ne. no_error) goto 998

       ! Read the AMD_Headers%MDS1_DSD without comparing
       call Read_DSD_from_KVT(AMD_Headers%SPH%MDS1_DSD,fileunit,&
                              AMD_DS_Name_MDS1,&
                              AMD_DS_Type_MDS1,error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the AMD_Headers%MDS1_DSD and compare if needed
       Compare = .false. ! TODO: Bring in line with RBC_SPH module
       call Read_DSD_from_KVT(AMD_Headers%SPH%MDS1_DSD,fileunit,&
                              AMD_DS_Name_MDS1,&
                              AMD_DS_Type_MDS1,error_flag,Compare)
       IF (error_flag .ne. no_error) goto 998
    END IF

       
       !------------------------------
       ! allocate memory for the DSD4 and read it
       !------------------------------
! Adapt the RBC_SPH module
! If successful then TODO: Move to AMD_SPH module (currently L2B_AMD_SPH module)
    IF (.not. associated(AMD_Headers%SPH%MDS2_DSD)) THEN
       ! allow this, to enable the possibility of reading the dbl file only,
       ! without reading the HDR files.
       allocate(AMD_Headers%SPH%MDS2_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine ReadAMDfile_DBL/Read_AMD_SPH_from_KVT()"
          error_flag = error_allocate
          goto 998
       END IF
       call Init_DSD(AMD_Headers%SPH%MDS2_DSD,error_flag)
       IF (error_flag .ne. no_error) goto 998

       ! Read the AMD_Headers%MDS2_DSD without comparing
       call Read_DSD_from_KVT(AMD_Headers%SPH%MDS2_DSD,fileunit,&
                              AMD_DS_Name_MDS2,&
                              AMD_DS_Type_MDS2,error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the AMD_Headers%MDS2_DSD and compare if needed
       Compare = .false. ! TODO: Bring in line with RBC_SPH module
       call Read_DSD_from_KVT(AMD_Headers%SPH%MDS2_DSD,fileunit,&
                              AMD_DS_Name_MDS2,&
                              AMD_DS_Type_MDS2,error_flag,Compare)
       IF (error_flag .ne. no_error) goto 998
    END IF

       !  #]

       !  #]
    END IF ! do_ReadHeaders

    !------------------------------
    ! Determine the number of model layers
    !------------------------------

    call GetIntAus(AMD_Headers%SPH%Num_of_Model_Layers, N_Model_Layers, error_flag)
    IF (error_flag .ne. no_error) return

    !------------------------------
    ! read the binary data sets 
    !------------------------------

    ! check that AMD data part is present
    IF (.not.(present(AMD)) .and. &
      & (do_ReadGeoloc_offnadir.or.do_ReadGeoloc_nadir.or. &
      &  do_ReadMet_offnadir.or.do_ReadMet_nadir)) THEN
        print *,"ERROR in ReadAMDfile_DBL()"
        print *,"AMD data argument is missing"
        error_flag = error_programming
        return
    ENDIF

    !print *,"test: do_ReadGeoloc_offnadir = ",do_ReadGeoloc_offnadir
    IF (do_ReadGeoloc_offnadir) THEN

       !  #[

!
! D.Tan, TODO or REMOVE: define AMD%headers_are_available
!
        print *,"Warning in ReadAMDfile_DBL(): no check on whether headers are available"
!
!       IF (.not. AMD%headers_are_available) THEN
!          print *,"ERROR in ReadAMDfile_DBL()==>Read_AMD_Geoloc_ADS_off_nadir:"
!          print *,"reading a binary dataset is only possible after reading"
!          print *,"the AMD headers, but these seem not yet available ..."
!          error_flag = error_programming
!          return
!       END IF

       IF (associated(AMD%Geoloc_ADS_off_nadir)) THEN
          ! REMARK: this structure is allocated at this point, but its sub-
          ! structures should still be nullified, and are only allocated in the
          ! init routine below. After calling Init_AMD_DataStructure this is
          ! the case, since Init_AMD_Geoloc_ADS() is called with N_Meas=0
          ! and N_range_gates=0. However, for the 2nd and next BRC's an
          ! explicit call to Empty_AMD_BinaryDatasets() is needed to
          ! restore this condition, before entering this reading code.
          !
          ! REMARK: this structure uses an extra rangegate to store the
          ! top of the upper rangebin, and the regular rangegates to store
          ! the bottom of each rangegate
          call Init_AMD_Geoloc_ADS(AMD%Geoloc_ADS_off_nadir, &
          ! N_Model_Layers, &
                                   error_flag)

          IF (error_flag .ne. no_error) return
       ELSE
          print *,"ERROR in ReadAMDfile_DBL()==>Read_AMD_Geoloc_ADS_off_nadir:"
          print *,"the AMD%Geoloc_ADS_off_nadir structure seems not allocated !!!"
          error_flag = error_programming
          return
       END IF

       IF ( (DataSetRecord_to_read .ge. 1) .and. &
            (DataSetRecord_to_read .le. &
                     AMD_Headers%SPH%GeoADS1_DSD%nr_of_dataset_records) ) THEN
          call Read_AMD_Geoloc_ADS(AMD_Headers%SPH%GeoADS1_DSD, fileunit, DataSetRecord_to_read, &
                                   AMD%Geoloc_ADS_off_nadir,error_flag)
          IF (error_flag .ne. no_error) return
       ELSE
          print *,"ERROR in ReadAMDfile_DBL()==>Read_AMD_Geoloc_ADS_off_nadir:"
          print *,"DataSetRecord_to_read = ",DataSetRecord_to_read
          print *,"This is outside the available range of datasets !!!"
          print *,"AMD_Headers%SPH%GeoADS1_DSD%nr_of_dataset_records = ",&
               AMD_Headers%SPH%GeoADS1_DSD%nr_of_dataset_records
          error_flag = error_reading_file
          return
       END IF

       print *,"done reading AMD%Geoloc_ADS_off_nadir"

       !  #]
    END IF ! do_ReadGeoloc_offnadir

    !print *,"test: do_ReadGeoloc_nadir = ",do_ReadGeoloc_nadir
    IF (do_ReadGeoloc_nadir) THEN

       !  #[

!
! D.Tan, TODO or REMOVE: define AMD_Headers%headers_are_available
!
        print *,"Warning in ReadAMDfile_DBL(): no check on whether headers are available"
!
!       IF (.not. AMD_Headers%headers_are_available) THEN
!          print *,"ERROR in ReadAMDfile_DBL()==>Read_AMD_Geoloc_ADS_nadir:"
!          print *,"reading a binary dataset is only possible after reading"
!          print *,"the AMD headers, but these seem not yet available ..."
!          error_flag = error_programming
!          return
!       END IF

       IF (associated(AMD%Geoloc_ADS_nadir)) THEN
          ! REMARK: this structure is allocated at this point, but its sub-
          ! structures should still be nullified, and are only allocated in the
          ! init routine below. After calling Init_AMD_DataStructure this is
          ! the case, since Init_AMD_Geoloc_ADS() is called with N_Meas=0
          ! and N_range_gates=0. However, for the 2nd and next BRC's an
          ! explicit call to Empty_AMD_BinaryDatasets() is needed to
          ! restore this condition, before entering this reading code.
          !
          ! REMARK: this structure uses an extra rangegate to store the
          ! top of the upper rangebin, and the regular rangegates to store
          ! the bottom of each rangegate
          call Init_AMD_Geoloc_ADS(AMD%Geoloc_ADS_nadir, &
          ! N_Model_Layers, &
                                   error_flag)

          IF (error_flag .ne. no_error) return
       ELSE
          print *,"ERROR in ReadAMDfile_DBL()==>Read_AMD_Geoloc_ADS_nadir:"
          print *,"the AMD%Geoloc_ADS_nadir structure seems not allocated !!!"
          error_flag = error_programming
          return
       END IF

       IF ( (DataSetRecord_to_read .ge. 1) .and. &
            (DataSetRecord_to_read .le. &
                     AMD_Headers%SPH%GeoADS2_DSD%nr_of_dataset_records) ) THEN
          call Read_AMD_Geoloc_ADS(AMD_Headers%SPH%GeoADS2_DSD, fileunit, DataSetRecord_to_read, &
                                   AMD%Geoloc_ADS_nadir,error_flag)
          IF (error_flag .ne. no_error) return
       ELSE
          print *,"ERROR in ReadAMDfile_DBL()==>Read_AMD_Geoloc_ADS_nadir:"
          print *,"DataSetRecord_to_read = ",DataSetRecord_to_read
          print *,"This is outside the available range of datasets !!!"
          print *,"AMD_Headers%SPH%GeoADS2_DSD%nr_of_dataset_records = ",&
               AMD_Headers%SPH%GeoADS2_DSD%nr_of_dataset_records
          error_flag = error_reading_file
          return
       END IF

       print *,"done reading AMD%Geoloc_ADS_nadir"

       !  #]
    END IF ! do_ReadGeoloc_nadir

    !print *,"test: do_ReadMet_offnadir = ",do_ReadMet_offnadir
    IF (do_ReadMet_offnadir) THEN
       
       !  #[

!
! D.Tan, TODO or REMOVE: define AMD%headers_are_available
!
        print *,"Warning in ReadAMDfile_DBL(): no check on whether headers are available"
!
!       IF (.not. AMD%headers_are_available) THEN
!          print *,"ERROR in ReadAMDfile_DBL()==>Read_AMD_Met_MDS_off_nadir:"
!          print *,"reading a binary dataset is only possible after reading"
!          print *,"the AMD headers, but these seem not yet available ..."
!          error_flag = error_programming
!          return
!       END IF

       IF (associated(AMD%Met_MDS_off_nadir)) THEN
          ! REMARK: this structure is allocated at this point, but its sub-
          ! structures should still be nullified, and are only allocated in the
          ! init routine below. After calling Init_AMD_DataStructure this is
          ! the case, since Init_AMD_Met_MDS() is called with N_Meas=0
          ! and N_range_gates=0. However, for the 2nd and next BRC's an
          ! explicit call to Empty_AMD_BinaryDatasets() is needed to
          ! restore this condition, before entering this reading code.
          !
          ! REMARK: this structure uses an extra rangegate to store the
          ! top of the upper rangebin, and the regular rangegates to store
          ! the bottom of each rangegate
          call Init_AMD_Met_MDS(AMD%Met_MDS_off_nadir, int(N_Model_Layers), &
                                   error_flag)

          IF (error_flag .ne. no_error) return
       ELSE
          print *,"ERROR in ReadAMDfile_DBL()==>Read_AMD_Met_MDS_off_nadir:"
          print *,"the AMD%Met_MDS_off_nadir structure seems not allocated !!!"
          error_flag = error_programming
          return
       END IF

       IF ( (DataSetRecord_to_read .ge. 1) .and. &
            (DataSetRecord_to_read .le. &
                     AMD_Headers%SPH%MDS1_DSD%nr_of_dataset_records) ) THEN
          call Read_AMD_Met_MDS(AMD_Headers%SPH%MDS1_DSD, fileunit, DataSetRecord_to_read, &
                                   AMD%Met_MDS_off_nadir,error_flag)
          IF (error_flag .ne. no_error) return
       ELSE
          print *,"ERROR in ReadAMDfile_DBL()==>Read_AMD_Met_MDS_off_nadir:"
          print *,"DataSetRecord_to_read = ",DataSetRecord_to_read
          print *,"This is outside the available range of datasets !!!"
          print *,"AMD_Headers%SPH%MDS1_DSD%nr_of_dataset_records = ",&
               AMD_Headers%SPH%MDS1_DSD%nr_of_dataset_records
          error_flag = error_reading_file
          return
       END IF

       print *,"done reading AMD%Met_MDS_off_nadir"

       !  #]
    END IF ! do_ReadMet_offnadir

    !print *,"test: do_ReadMet_nadir = ",do_ReadMet_nadir
    IF (do_ReadMet_nadir) THEN

       !  #[

!
! D.Tan, TODO or REMOVE: define AMD%headers_are_available
!
        print *,"Warning in ReadAMDfile_DBL(): no check on whether headers are available"
!
!       IF (.not. AMD%headers_are_available) THEN
!          print *,"ERROR in ReadAMDfile_DBL()==>Read_AMD_Met_MDS_nadir:"
!          print *,"reading a binary dataset is only possible after reading"
!          print *,"the AMD headers, but these seem not yet available ..."
!          error_flag = error_programming
!          return
!       END IF

       IF (associated(AMD%Met_MDS_nadir)) THEN
          ! REMARK: this structure is allocated at this point, but its sub-
          ! structures should still be nullified, and are only allocated in the
          ! init routine below. After calling Init_AMD_DataStructure this is
          ! the case, since Init_AMD_Met_MDS() is called with N_Meas=0
          ! and N_range_gates=0. However, for the 2nd and next BRC's an
          ! explicit call to Empty_AMD_BinaryDatasets() is needed to
          ! restore this condition, before entering this reading code.
          !
          ! REMARK: this structure uses an extra rangegate to store the
          ! top of the upper rangebin, and the regular rangegates to store
          ! the bottom of each rangegate
          call Init_AMD_Met_MDS(AMD%Met_MDS_nadir, int(N_Model_Layers), &
                                   error_flag)

          IF (error_flag .ne. no_error) return
       ELSE
          print *,"ERROR in ReadAMDfile_DBL()==>Read_AMD_Met_MDS_nadir:"
          print *,"the AMD%Met_MDS_nadir structure seems not allocated !!!"
          error_flag = error_programming
          return
       END IF

       IF ( (DataSetRecord_to_read .ge. 1) .and. &
            (DataSetRecord_to_read .le. &
                     AMD_Headers%SPH%MDS2_DSD%nr_of_dataset_records) ) THEN
          call Read_AMD_Met_MDS(AMD_Headers%SPH%MDS2_DSD, fileunit, DataSetRecord_to_read, &
                                   AMD%Met_MDS_nadir,error_flag)
          IF (error_flag .ne. no_error) return
       ELSE
          print *,"ERROR in ReadAMDfile_DBL()==>Read_AMD_Met_MDS_nadir:"
          print *,"DataSetRecord_to_read = ",DataSetRecord_to_read
          print *,"This is outside the available range of datasets !!!"
          print *,"AMD_Headers%SPH%MDS2_DSD%nr_of_dataset_records = ",&
               AMD_Headers%SPH%MDS2_DSD%nr_of_dataset_records
          error_flag = error_reading_file
          return
       END IF

       print *,"done reading AMD%Met_MDS_nadir"

       !  #]
    END IF ! do_ReadMet_nadir

    close(unit=fileunit)
    call free_lun(fileunit)

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"done with reading AMD datafile"

    !------------------------------
    ! print all datasets as test
    !------------------------------

!print *,"----------------------------------------------------------"
!print *,"--------------- extra test prints ------------------------"
!print *,"----------------------------------------------------------"
!print *,"AMD_Headers%SPH%GeoADS1_DSD%..."
!call Print_DSD(AMD_Headers%SPH%GeoADS1_DSD)
!print *,"----------------------------------------------------------"
!print *,"AMD%Geoloc_ADS_off_nadir%... "
!call Print_AMD_Geoloc_ADS(AMD%Geoloc_ADS_off_nadir,error_flag,PrintAll=.true.)
!print *,"----------------------------------------------------------"
!print *,"AMD%Geoloc_ADS_nadir%... "
!call Print_AMD_Geoloc_ADS(AMD%Geoloc_ADS_nadir,error_flag,PrintAll=.true.)
!print *,"----------------------------------------------------------"
!print *,"AMD%Met_MDS_off_nadir%... "
!!call Print_AMD_Met_MDS(AMD%Met_MDS_off_nadir,error_flag,PrintAll=.true.)
!call Print_AMD_Met_MDS(AMD%Met_MDS_off_nadir,error_flag)
!print *,"----------------------------------------------------------"
!print *,"AMD%Met_MDS_nadir%... "
!!call Print_AMD_Met_MDS(AMD%Met_MDS_nadir,error_flag,PrintAll=.true.)
!call Print_AMD_Met_MDS(AMD%Met_MDS_nadir,error_flag)
!print *,"----------------------------------------------------------"

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"----------------------------------------------------------"
       print *,"AMD_Headers%MPH%..."
       call Print_MPH(AMD_Headers%MPH,error_flag,prefix="AMD_Headers%MPH")
       print *,"----------------------------------------------------------"
       print *,"AMD_Headers%SPH%..."
       call Print_L2B_AMD_SPH(AMD_Headers%SPH,error_flag)
       print *,"----------------------------------------------------------"
    END IF

    return

998 print *,"ABORTING subroutine ReadAMDfile_DBL/Read_AMD_SPH_from_KVT"
    return

999 print *,"ERROR could not open inputfile: ",trim(filename_dbl)
    error_flag = error_opening_file
    return

  end subroutine ReadAMDfile_DBL

    !  #]
  subroutine Load_All_AMD_Profiles(NWPfilename,AMD_Headers,AMD,error_flag)
    !  #[ read all KVT and binary values for all BRCs
    character(len=*),       intent(in)    :: NWPfilename
    TYPE(AMD_HDR_DataType), intent(inout) :: AMD_Headers ! result
    TYPE(AMD_BRC_DataType), dimension(:), pointer     :: AMD ! result
    integer,                intent(out)   :: error_flag

    ! local variables
    integer        :: AllocStatus, iDSR
    integer(IntAl) :: num_AMD_DSR
    integer        :: Num_model_layers
    
    print *, "L2BP_primary: reading aux met data from EE(.hdr, .dbl): "
    print *, trim(NWPfilename)

       !  #[ load a proper EE_CFI formatted AMD file
       
       ! loading the AMD headers
       
       call Init_AMD_HDR_DataStructure(AMD_Headers, error_flag)
       IF (error_flag .ne. no_error) return
       
       ! read the header only from AUX_MET file in EE-format
       ! to extract the nr of available profiles
       call ReadAMDfile(AMD_Headers, NWPfilename, error_flag, &
                        ReadHeaders=.true.)
       IF (error_flag .ne. no_error) return

!       ! get the nr of off-nadir profiles in this file
       ! call GetIntAl(AMD(1)%SPH%Num_Records_in_DS1,num_AMD_DSR,error_flag)
       ! IF (error_flag .ne. no_error) return

       ! WORKAROUND because this field seems not yet filled
       ! in the current AMD datafiles.....
       call GetIntAl(AMD_Headers%SPH%GeoADS1_DSD%num_dsr,num_AMD_DSR,error_flag)
       IF (error_flag .ne. no_error) return
       !num_AMD_DSR=1
       ! NEXT remark: if this number is set to 2 or higher the program
       ! crashes with a segmentation fault. This should never happen!!!
       ! The ADM reading software should return with a proper error
       ! if the input parameters are wrong....
       !num_AMD_DSR=2 ! 1
       ! DT: Should be fixed now that call to Init_AMD_DataStructure passes
       !     AMD(iDSR) and not AMD(1).

       allocate(AMD(num_AMD_DSR),stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation problem for AMD in Load_All_AMD_Profiles()"
          error_flag = error_allocate
          return
       END IF

       DO iDSR=1,num_AMD_DSR
          ! init this AMD DSR
          Num_Model_Layers = 0 ! For initialization purposes
          call Init_AMD_BRC_DataStructure(AMD(iDSR), Num_Model_Layers, error_flag)
          IF (error_flag .ne. no_error) return

          ! read this AMD DSR
          call ReadAMDfile(AMD_Headers, NWPfilename, error_flag, &
               AMD = AMD(iDSR), &
               Which_DSR = iDSR,  &
               ReadHeaders=.true.,         &
               ReadGeoloc_offnadir=.true., &
               ReadGeoloc_nadir=.true.,    & 
               ReadMet_offnadir=.true.   , &
               ReadMet_nadir=.true.     )
          IF (error_flag .ne. no_error) return
       END DO
       !  #]

  end subroutine Load_All_AMD_Profiles
    !  #]
  !---------------------------------
end module ReadAMDdata
