module WriteAMDdata
  !  #[ Description:
  !
  !     a module to write the 2 files describing the AMD product
  !     of the Aeolus instrument
  !
  ! created:     07-Feb-2006 by P. Poli, Meteo-France
  !
  ! History:
  !   07-Feb-2006 P. Poli    Original version
  !   21-Feb-2006 J. de Kloe adapted to changes in Write_MPH_to_KVT
  !   02-Mar-2006 J. de Kloe adapted to changes in XML_module
  !   30-Mar-2006 D. Tan     Restore writing of AMD()%Met_MDS_nadir
  !   12-May-2006 D. Tan     Null use of ParKind1 module
  !   31-Jul-2006 J. de Kloe renamed set_file_pointer to SetFilePointer
  !   17-Aug-2006 P. Poli    removed obsolete variables
  !   17-Nov-2006 J. de Kloe removed initialisation of unused filenames
  !                          and small interface change in Fill_DSD_items
  !   24-Apr-2007 J. de Kloe phase out the SummaryVars code
  !   06-Jun-2007 J. de Kloe adapted to changes in xml_module interface
  !   20-Jul-2007 J. de Kloe added call to xml_tree_set_schema()
  !   18-Oct-2007 J. de Kloe added a nullify statement
  !   19-Oct-2007 P. Poli    Split AMD_DataType in two parts:
  !                          AMD_HDR_DataType and AMD_BRC_DataType
  !                          Modified WriteAMDfile to allow using AMD header
  !                          information from another datastructure
  !   01-Nov-2007 J. de Kloe added call to xml_tree_set_namespace 
  !                          to replace xml_tree_set_schema
  !   14-Nov-2007 J. de Kloe implemented filling of Input DSD's with dummy data
  !   16-Jan-2008 J. de Kloe phase out integer kind i_
  !   17-Jan-2008 J. de Kloe removed a last few 64bit incompatibilities
  !   21-Jan-2008 J. de kloe added workaround in call to Write_MPH_to_KVT
  !                          and to GetByteSize_MPH_KVT
  !                          to not change the fileformat for now.
  !   21-Feb-2008 J. de Kloe removed the workaround again, because it was
  !                          caused by a misunderstanding on L1B side
  !
  !  #]
  !  #[ Modules used:

  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE AMD_DataStructure, only: AMD_HDR_DataType, AMD_BRC_DataType
  USE FixedHeader, only: Copy_FH_to_xml_tree, file_type_AMD
  USE MainProductHeader, only: Copy_MPH_to_xml_tree, &
                               Write_MPH_to_KVT, Fill_MPH_Sizes, &
                               GetByteSize_MPH_KVT
  USE L2B_AMD_SPH, only: Copy_L2B_AMD_SPH_to_xml_tree, &
                         Write_L2B_AMD_SPH_to_KVT, &
                         GetByteSize_AMD_SPH_KVT, &
                         AMD_DS_Name_GeoADS1, AMD_DS_Type_GeoADS1, &
                         AMD_DS_Name_GeoADS2, AMD_DS_Type_GeoADS2, &
                         AMD_DS_Name_MDS1,    AMD_DS_Type_MDS1, &
                         AMD_DS_Name_MDS2,    AMD_DS_Type_MDS2, &
                         AMD_DS_Type_Input
  USE AMD_Geoloc_ADS, only: Write_AMD_Geoloc_ADS, &
                            GetByteSize_AMD_Geoloc_ADS
  USE AMD_Met_MDS, only: Write_AMD_Met_MDS, &
                         GetByteSize_AMD_Met_MDS
  USE DataSetDescriptor, only: Init_DSD, Fill_DSD_items, DSD_type, &
                               GetByteSize_DSD_KVT, maxlen_ds_name, &
                               maxlen_dsd_filename=>maxlen_filename
  USE KVT_module, only: max_str_len, Init_KVT_module
  USE DirectBinaryIO, only: Init_DirectBinaryIO_module , BigEndian, &
                            SetFilePointer
  USE XML_module, only: xml_node_type, &
       Init_XML, xml_tree_print, xml_tree_write, create_xml_tree, &
       xml_tree_create_header, delete_xml_tree, ee_h_name, &
       xml_tree_set_namespace, uri_ns_header

  USE EE_CFI_Datatypes, only: IntAd, IntAul, IntAl, maxlen_IntAd, &
                              SetIntAd
  ! support modules
  USE Logging, only: logmsg,log_debug,log_info,log_progress,log_warn,log_error
  USE Numerics, only: i8_
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting

  USE LunManager, only: get_lun, free_lun
  USE ErrorHandler, only: no_error, error_xml, error_programming, &
                          error_opening_file, & ! error_reading_file,&
                          error_writing_file, error_allocate
  USE c_support, only: get_filesize
  USE AeolusConstants, only: iodd_version_AuxMet
  !  #]  
  !  #[ Parameter/Variable/Type declarations

  implicit none

  ! make all module contents private, unless explicitely stated otherwise
  private

  ! the default verbosity setting is Error (as silent as possible)
  integer, save :: LocalVerbosity = VerbosityError

  ! remember the initialisation state of the module
  logical, save :: module_is_initialised = .false.

  !  #]
  !  #[ Module interface
  public :: WriteAMDfile, InitWriteAMDfileModule !, InitWriteAMD
  !  #]
contains
  !---------------------------------
  subroutine InitWriteAMDfileModule(error_flag, ChoosenVerbosity)
    !  #[

    integer,           intent(out) :: error_flag
    integer, optional, intent(in)  :: ChoosenVerbosity

    error_flag = no_error

    IF (present(ChoosenVerbosity)) LocalVerbosity = ChoosenVerbosity

    IF (LocalVerbosity .ge. VerbosityReport) &
         call print_verbosity_setting(LocalVerbosity,ModuleName="WriteAMDfile")

    call Init_KVT_module()

    call Init_DirectBinaryIO_module(error_flag,FileByteOrder=BigEndian)
    if (error_flag .ne. no_error) return

    call Init_XML(error_flag) ! (ChoosenVerbosity=VerbosityReport)
    if (error_flag .ne. no_error) return

    module_is_initialised = .true.

  end subroutine InitWriteAMDfileModule
    !  #]
  !---------------------------------
  subroutine WriteAMDfile(Num_BRC, AMD_Headers, AMD, error_flag)
    !  #[
    integer,                intent(in)    :: Num_BRC
    type(AMD_HDR_DataType), intent(inout) :: AMD_Headers
    type(AMD_BRC_DataType), dimension(1:Num_BRC), intent(inout) :: AMD
    integer,                intent(out)   :: error_flag

    ! local variables
    character(len=256) :: filename_hdr
    character(len=256) :: filename_dbl
    character(len=12)  :: filesize_string
    character(len=256) :: path
    character(len=256) :: xml_namespace_AMD

    ! general xml items
    TYPE(xml_node_type), pointer :: xml_tree

    ! init
    error_flag = no_error
    filename_hdr = trim(AMD_Headers%FH%File_Name)//".HDR"
    filename_dbl = trim(AMD_Headers%FH%File_Name)//".DBL"

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in WriteAMDfile:"
       print *,"the WriteAMDfile module should be initialised before use !"
       error_flag = error_programming
       return
    END IF

    call logmsg(log_info,"WriteAMDfile: Writing AMD output file...")

    ! Write the DBL file including MPH, SPH, and data and figure out sizes
    ! NOTE that this WriteAMDtoBinaryFile fills the sizes in the MPH, SPH
    !      and DSD structures, and therefore it must be called first
    !       before trying to construct the XML header file !!!
    call WriteAMDtoBinaryFile(Num_BRC,AMD_Headers,AMD,filename_dbl,error_flag)
    IF (error_flag .ne. no_error) return

    ! allocate a datastructure to hold the xml data
    ! and get a pointer to it, and fill it with the xml root tag pair:
    ! <Earth_Explorer_File><\Earth_Explorer_File>

    ! allocate and init the datastructure
    nullify(xml_tree)
    call create_xml_tree(xml_tree, error_flag)
    IF (error_flag .ne. no_error) return

    ! create a blank header (fixed+variable) and define the root
    call xml_tree_create_header(xml_tree, ee_h_name, error_flag)
    IF (error_flag .ne. no_error) return

    !example: xmlns="http://www.esa.int/schemas/ae/AUX_MET_12_01.01"
    !uri_ns_header    ! defined in xml_module
    !file_type_AMD    ! defined in fixedheader
    !iodd_version_AMD ! defined in AeolusConstants
    xml_namespace_AMD = trim(uri_ns_header)//trim(file_type_AMD)//&
                        "_"//trim(iodd_version_AuxMet)

    ! define which xml namespace should be used
    call xml_tree_set_namespace(xml_tree,xml_namespace_AMD,error_flag)
    IF (error_flag .ne. no_error) return

    ! fill all FH items in the xml tree
    path = "/"//ee_h_name ! = /Earth_Explorer_File
    call Copy_FH_to_xml_tree(AMD_Headers%FH, path, xml_tree, error_flag)
    IF (error_flag .ne. no_error) return

    ! => write the MPH
    call Copy_MPH_to_xml_tree(AMD_Headers%MPH, path, xml_tree, error_flag)
    IF (error_flag .ne. no_error) return

    ! => write the SPH
    call Copy_L2B_AMD_SPH_to_xml_tree(AMD_Headers%SPH, path, xml_tree, &
                                      error_flag)
    IF (error_flag .ne. no_error) return

    ! Write the HDR file including FH, MPH, and SPH
    call xml_tree_write(xml_tree, filename_hdr, error_flag)
    IF (error_flag .ne. no_error) return

    call delete_xml_tree(xml_tree,error_flag)
    IF (error_flag .ne. no_error) return

    call logmsg(log_progress,"Successfully written AMD file: "//&
                trim(AMD_Headers%FH%File_Name))
    write(filesize_string,"(i12)",err=999) &
         get_filesize(trim(AMD_Headers%FH%File_Name)//".DBL")
    call logmsg(log_progress,"Size of binary AMD file was: "//&
                             trim(filesize_string)//" bytes")

    ! TODO: add a GetByteSize_AMD_DataStructure routine
    !       and check if this number is as expected

    return

999 call logmsg(log_error,"ERROR in WriteAMDfile: could not convert size "//&
                          "of binary product file to a string")
    error_flag = error_writing_file
    return

  end subroutine WriteAMDfile
    !  #]
  !---------------------------------
  subroutine WriteAMDtoBinaryFile(Num_BRC, AMD_Headers, AMD, &
             filename_dbl,error_flag)
! NWPfilename, Paramfilename)
    !  #[
    integer,                intent(in)    :: Num_BRC
    type(AMD_HDR_DataType), intent(inout) :: AMD_Headers
    type(AMD_BRC_DataType), dimension(1:Num_BRC), intent(inout) :: AMD
    integer,                intent(out)   :: error_flag
    character(len=256),     intent(in)    :: filename_dbl
!    character(len=*), optional, intent(in)                  :: NWPfilename, &
!                                                               Paramfilename

    ! local variables
    integer            :: fileunit, AllocStatus
    integer            :: iBRC
    ! remark: variables used in the inquire() call must be default integers
    ! so don't force them to be 4 or 8 bytes 
    integer            :: filepos_before, filepos_after, endpos, begpos
    integer            :: loc_size, size_of_ds_calc, Num_Input_DSDs, i
    integer(IntAl)     :: Dsr_Size, Num_Dsd, Num_Attached_DS, Num_Dsr
    integer(IntAd)     :: Ds_Offset, Tot_Product_Size
    character(len=6)   :: str_cntr
!    character(len=256) :: NWPfilename_local,Paramfilename_local
    character(maxlen_IntAd) :: str_value_i1,str_value_i2
    character(maxlen_dsd_filename) :: Ds_filename
    character(maxlen_ds_name)      :: Ds_name
    type(DSD_type), pointer :: pointer_to_dsd

    call logmsg(log_info,"WriteAMDtoBinaryFile: writing AMD data "//&
                         "to binary file: "//trim(filename_dbl))

    error_flag = no_error

! WARNING, remark for myself: SPH%Input_DSD must have been allocated
!          at this point, otherwise GetByteSize_L2B_AMD_SPH_KVT
!          will return a wrong result

    !  #[ fill GeoADS1_DSD
    IF (.not. associated(AMD_Headers%SPH%GeoADS1_DSD)) THEN
       allocate(AMD_Headers%SPH%GeoADS1_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"WriteAMDtoBinaryFile: ERROR: Allocation failure:"
          error_flag = error_allocate
          goto 998
       END IF
    END IF
    call Init_DSD(AMD_Headers%SPH%GeoADS1_DSD,error_flag)
    IF (error_flag .ne. no_error) goto 998

    Dsr_size  = GetByteSize_AMD_Geoloc_ADS()
    Ds_offset = GetByteSize_MPH_KVT() + &
                GetByteSize_AMD_SPH_KVT(AMD_Headers%SPH)

    ! Geoloc_ADS_off_nadir is an attached annotation dataset, 
    ! so no need to set the filename.
    call Fill_DSD_items (AMD_Headers%SPH%GeoADS1_DSD, &
             AMD_DS_Name_GeoADS1, AMD_DS_Type_GeoADS1, &
             Ds_Offset, int(Num_BRC,IntAl), Dsr_Size, error_flag)
    IF (error_flag .ne. no_error) goto 999

        ! remember this, for use in the next DSD
    Ds_offset = Ds_offset + Dsr_size*Num_BRC
    !  #]
    !  #[ fill GeoADS2_DSD
    IF (.not. associated(AMD_Headers%SPH%GeoADS2_DSD)) THEN
       allocate(AMD_Headers%SPH%GeoADS2_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"WriteAMDtoBinaryFile: ERROR: Allocation failure:"
          error_flag = error_allocate
          goto 998
       END IF
    END IF
    call Init_DSD(AMD_Headers%SPH%GeoADS2_DSD,error_flag)
    IF (error_flag .ne. no_error) goto 998

    Dsr_size = GetByteSize_AMD_Geoloc_ADS()
    ! Ds_offset has already been set above

    ! Geoloc_ADS_nadir is an attached annotation dataset, 
    ! so no need to set the filename.
    call Fill_DSD_items (AMD_Headers%SPH%GeoADS2_DSD, &
             AMD_DS_Name_GeoADS2, AMD_DS_Type_GeoADS2, &
             Ds_Offset, int(Num_BRC,IntAl), Dsr_Size, error_flag)
    IF (error_flag .ne. no_error) goto 999

    ! remember this, for use in the next DSD
    Ds_offset = Ds_offset + Dsr_size*Num_BRC
    !  #]
    !  #[ fill MDS1_DSD
    IF (.not. associated(AMD_Headers%SPH%MDS1_DSD)) THEN
       allocate(AMD_Headers%SPH%MDS1_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"WriteAMDtoBinaryFile: ERROR: Allocation failure:"
          error_flag = error_allocate
          goto 998
       END IF
    END IF
    call Init_DSD(AMD_Headers%SPH%MDS1_DSD,error_flag)
    IF (error_flag .ne. no_error) goto 998

    Dsr_size = GetByteSize_AMD_Met_MDS(AMD(1)%Met_MDS_off_nadir)
    ! Ds_offset has already been set above

    ! Geoloc_ADS_nadir is an attached annotation dataset, 
    ! so no need to set the filename.
    call Fill_DSD_items (AMD_Headers%SPH%MDS1_DSD, &
             AMD_DS_Name_MDS1, AMD_DS_Type_MDS1, &
             Ds_Offset, int(Num_BRC,IntAl), Dsr_Size, error_flag)
    IF (error_flag .ne. no_error) goto 999

    ! remember this, for use in the next DSD
    Ds_offset = Ds_offset + Dsr_size*Num_BRC
    !  #]
    !  #[ fill MDS2_DSD
    IF (.not. associated(AMD_Headers%SPH%MDS2_DSD)) THEN
       allocate(AMD_Headers%SPH%MDS2_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"WriteAMDtoBinaryFile: ERROR: Allocation failure:"
          error_flag = error_allocate
          goto 998
       END IF
    END IF
    call Init_DSD(AMD_Headers%SPH%MDS2_DSD,error_flag)
    IF (error_flag .ne. no_error) goto 998

    Dsr_size = GetByteSize_AMD_Met_MDS(AMD(1)%Met_MDS_nadir)
    ! Ds_offset has already been set above

    ! Geoloc_ADS_nadir is an attached annotation dataset, 
    ! so no need to set the filename.
    call Fill_DSD_items (AMD_Headers%SPH%MDS2_DSD, &
             AMD_DS_Name_MDS2, AMD_DS_Type_MDS2, &
             Ds_Offset, int(Num_BRC,IntAl), Dsr_Size, error_flag)
    IF (error_flag .ne. no_error) goto 999

    ! remember this, for use in the next DSD
    Ds_offset = Ds_offset + Dsr_size*Num_BRC
    !  #]

    !  #[ fill Input_DSD's
    IF (associated(AMD_Headers%SPH%Input_DSD)) THEN
       ! NOTE: Num_Input_DSDs is hardcoded to be 10 in l2b_amd_sph.F90
       !       (line 202). Don't forget to change this when this DSD
       !       is to be filled with real data.
       Num_Input_DSDs = size(AMD_Headers%SPH%Input_DSD)
       do i=1,Num_Input_DSDs
          pointer_to_dsd => AMD_Headers%SPH%Input_DSD(i)
          call Init_DSD(pointer_to_dsd,error_flag)
          IF (error_flag .ne. no_error) goto 998

          write(Ds_Name,"(a,i2.2)",err=996) "Input DS ",i
          Num_Dsr     = 0_IntAl
          Dsr_Size    = 0_IntAl
          Ds_Filename = "[Undefined]"
          call Fill_DSD_items (pointer_to_dsd, DS_Name, AMD_DS_Type_Input, &
                               0_IntAd, Num_Dsr, Dsr_Size, error_flag, &
                               Ds_Filename)
          IF (error_flag .ne. no_error) goto 999
       end do
    END IF
    !  #]

    !  #[ fill L2B MPH sizes
    Tot_Product_Size = Ds_offset

    Num_Attached_DS = 4
    Num_Dsd         = Num_Attached_DS 
    if (associated(AMD_Headers%SPH%Input_DSD)) then
       Num_Dsd = Num_Dsd + int(size(AMD_Headers%SPH%Input_DSD),IntAl)
    end if

    call Fill_MPH_Sizes(AMD_Headers%MPH,&
              Tot_Size      = Tot_Product_Size,&
              Sph_Size      = GetByteSize_AMD_SPH_KVT(AMD_Headers%SPH),&
              Num_Dsd       = Num_Dsd,&
              Dsd_Size      = GetByteSize_DSD_KVT(),&
              Num_Data_Sets = Num_Attached_DS,&
              error_flag    = error_flag)
    IF (error_flag .ne. no_error) goto 999
    !  #]

    !  #[ 1. open the binary file
    fileunit = get_lun()
    IF (fileunit .lt. 0) THEN
       ! no fileunits available, so report an error
       error_flag = error_opening_file
       return
    END IF

    open(unit=fileunit,file=filename_dbl,status="replace",&
         action="write",access="direct",recl=1,form="UNFORMATTED",ERR=999)
    !  #]

    !  #[ 2. write AMD MPH
    call logmsg(log_info,'WriteAMDtoBinaryFile: writing MPH')
    call Write_MPH_to_KVT(AMD_Headers%MPH,fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 998 ! abort writing the file
    !  #]
    !  #[ 3. write AMD SPH
    call logmsg(log_info,'WriteAMDtoBinaryFile: writing SPH')
    call Write_L2B_AMD_SPH_to_KVT(AMD_Headers%SPH,fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 998 ! abort writing the file
    !  #]

    !  #[ 4. write AMD GeoADS1

    ! 4.2 get position before writing GeoADS1
    inquire(unit=fileunit,nextrec=filepos_before)
    ! 4.3 writing GeoADS1, get size of one DS and check sizes are
    ! constant
    do iBRC=1,Num_BRC
       write(str_cntr,"(i6)",err=997) iBRC
       call logmsg(log_info,'WriteAMDtoBinaryFile: '//&
                            'writing Geoloc_ADS_off_nadir (GeoADS1) for '//&
                            'BRC #'//trim(str_cntr))

       inquire(unit=fileunit,nextrec=begpos)
       call Write_AMD_Geoloc_ADS(fileunit,AMD(iBRC)%Geoloc_ADS_off_nadir,&
                                 error_flag)
       IF (error_flag .ne. no_error) goto 998 ! abort writing the file
       inquire(unit=fileunit,nextrec=endpos)
       endpos=endpos-1
       loc_size=endpos-begpos+1
       ! verify that all the DSRs have the same size
       IF (loc_size.ne.AMD_Headers%SPH%GeoADS1_DSD%size_of_one_dsr) THEN
          call logmsg(log_warn,'WriteAMDtoBinaryFile: WARNING: '//&
                      'Geoloc_ADS_off_nadir (GeoADS1) DSRs have '//&
                      'variable sizes')
       ENDIF
    enddo
    ! 4.4 get position and size after writing one DS
    inquire(unit=fileunit,nextrec=filepos_after)
    size_of_ds_calc = AMD_Headers%SPH%GeoADS1_DSD%size_of_one_dsr * &
                    & AMD_Headers%SPH%GeoADS1_DSD%nr_of_dataset_records
    IF (size_of_ds_calc .ne. (filepos_after-filepos_before)) THEN
       call SetIntAd(int(filepos_after-filepos_before,IntAd),&
                     str_value_i1,error_flag)
       call SetIntAd(int(size_of_ds_calc,IntAd),str_value_i2,error_flag)
       call logmsg(log_warn,"WriteAMDtoBinaryFile WARNING: "//&
            "Total size of Geoloc_ADS_off_nadir (GeoADS1) "//&
            "is not consistent with DSD content !!! "//&
            "actual DS size: "//str_value_i1//&
            "GetByteSize result in DSD: "//str_value_i2)
    ENDIF
    !  #]
    !  #[ 5. write AMD GeoADS2

    ! 5.2 get position before writing GeoADS2
    inquire(unit=fileunit,nextrec=filepos_before)

    ! 5.3 writing GeoADS2, get size of one DS and check sizes are
    ! constant
    do iBRC=1,Num_BRC
       write(str_cntr,"(i6)",err=997) iBRC
       call logmsg(log_info,'WriteAMDtoBinaryFile: '//&
                            'writing Geoloc_ADS_nadir (GeoADS2) for '//&
                            'BRC #'//trim(str_cntr))

       inquire(unit=fileunit,nextrec=begpos)
       call Write_AMD_Geoloc_ADS(fileunit,AMD(iBRC)%Geoloc_ADS_nadir,&
                                 error_flag)
       IF (error_flag .ne. no_error) goto 998 ! abort writing the file
       inquire(unit=fileunit,nextrec=endpos)
       endpos   = endpos-1
       loc_size = endpos-begpos+1
       ! verify that all the DSRs have the same size
       IF (loc_size.ne.AMD_Headers%SPH%GeoADS2_DSD%size_of_one_dsr) THEN
          call logmsg(log_warn,'WriteAMDtoBinaryFile: WARNING: '//&
                      'Geoloc_ADS_nadir (GeoADS2) DSRs have '//&
                      'variable sizes')
       ENDIF
    enddo
    ! 5.4 get position and size after writing one DS
    inquire(unit=fileunit,nextrec=filepos_after)
    size_of_ds_calc = AMD_Headers%SPH%GeoADS2_DSD%size_of_one_dsr * &
                    & AMD_Headers%SPH%GeoADS2_DSD%nr_of_dataset_records
    IF (size_of_ds_calc .ne. (filepos_after-filepos_before)) THEN
       call SetIntAd(int(filepos_after-filepos_before,IntAd),&
                     str_value_i1,error_flag)
       call SetIntAd(int(size_of_ds_calc,IntAd),str_value_i2,error_flag)
       call logmsg(log_warn,"WriteAMDtoBinaryFile WARNING: "//&
            "Total size of Geoloc_ADS_nadir (GeoADS2) "//&
            "is not consistent with DSD content !!! "//&
            "actual DS size: "//str_value_i1//&
            "GetByteSize result in DSD: "//str_value_i2)
    ENDIF
    !  #]
    !  #[ 6. write AMD MDS1

    ! 6.2 get position before writing MDS1
    inquire(unit=fileunit,nextrec=filepos_before)

    ! 6.3 writing MDS1, get size of one DS and check sizes are
    ! constant
    do iBRC=1,Num_BRC
       write(str_cntr,"(i6)",err=997) iBRC
       call logmsg(log_info,'WriteAMDtoBinaryFile: '//&
                            'writing Met_MDS_off_nadir (MDS1) '//&
                            'for BRC #'//trim(str_cntr))

       inquire(unit=fileunit,nextrec=begpos)
       call Write_AMD_Met_MDS(fileunit,AMD(iBRC)%Met_MDS_off_nadir,error_flag)
       IF (error_flag .ne. no_error) goto 998 ! abort writing the file
       inquire(unit=fileunit,nextrec=endpos)
       endpos   = endpos-1
       loc_size = endpos-begpos+1
       ! verify that all the DSRs have the same size
       IF (loc_size.ne.AMD_Headers%SPH%MDS1_DSD%size_of_one_dsr) THEN
          call logmsg(log_warn,'WriteAMDtoBinaryFile: WARNING: '//&
                      'Met_MDS_off_nadir (MDS1) DSRs have '//&
                      'variable sizes')
       ENDIF
    enddo
    ! 6.4 get position and size after writing one DS
    inquire(unit=fileunit,nextrec=filepos_after)

    size_of_ds_calc = AMD_Headers%SPH%MDS1_DSD%size_of_one_dsr * &
                    & AMD_Headers%SPH%MDS1_DSD%nr_of_dataset_records
    IF (size_of_ds_calc .ne. (filepos_after-filepos_before)) THEN
       call SetIntAd(int(filepos_after-filepos_before,IntAd),&
                     str_value_i1,error_flag)
       call SetIntAd(int(size_of_ds_calc,IntAd),str_value_i2,error_flag)
       call logmsg(log_warn,"WriteAMDtoBinaryFile WARNING: "//&
            "Total size of Met_MDS_off_nadir (MDS1) "//&
            "is not consistent with DSD content !!! "//&
            "actual DS size: "//str_value_i1//&
            "GetByteSize result in DSD: "//str_value_i2)
    ENDIF
    !  #]
    !  #[ 7. write AMD MDS2

    ! 7.2 get position before writing MDS2
    inquire(unit=fileunit,nextrec=filepos_before)

    ! 7.3 writing MDS2, get size of one DS and check sizes are
    ! constant
    do iBRC=1,Num_BRC
       write(str_cntr,"(i6)",err=997) iBRC
       call logmsg(log_info,'WriteAMDtoBinaryFile: '//&
                            'writing Met_MDS_nadir (MDS2) '//&
                            'for BRC #'//trim(str_cntr))
       inquire(unit=fileunit,nextrec=begpos)
       call Write_AMD_Met_MDS(fileunit,AMD(iBRC)%Met_MDS_nadir,error_flag)
       IF (error_flag .ne. no_error) goto 998 ! abort writing the file
       inquire(unit=fileunit,nextrec=endpos)
       endpos   = endpos-1
       loc_size = endpos-begpos+1
       ! verify that all the DSRs have the same size
       IF (loc_size.ne.AMD_Headers%SPH%MDS2_DSD%size_of_one_dsr) THEN
          call logmsg(log_warn,'WriteAMDtoBinaryFile: WARNING: '//&
                      'Met_MDS_nadir (MDS2) DSRs have '//&
                      'variable sizes')
       ENDIF
    enddo
    ! 7.4 get position and size after writing one DS
    inquire(unit=fileunit,nextrec=filepos_after)

    size_of_ds_calc = AMD_Headers%SPH%MDS2_DSD%size_of_one_dsr * &
                    & AMD_Headers%SPH%MDS2_DSD%nr_of_dataset_records
    IF (size_of_ds_calc .ne. (filepos_after-filepos_before)) THEN
       call SetIntAd(int(filepos_after-filepos_before,IntAd),&
                     str_value_i1,error_flag)
       call SetIntAd(int(size_of_ds_calc,IntAd),str_value_i2,error_flag)
       call logmsg(log_warn,"WriteAMDtoBinaryFile WARNING: "//&
            "Total size of Met_MDS_nadir (MDS2) "//&
            "is not consistent with DSD content !!! "//&
            "actual DS size: "//str_value_i1//&
            "GetByteSize result in DSD: "//str_value_i2)
    ENDIF
    !  #]

    !------------------------------
    ! done
    !------------------------------
    close(unit=fileunit)

    return

996 call logmsg(log_error,"Error in WriteAMDtoBinaryFile: "//&
                "Could not convert counter i to a 2 digit string")
    error_flag = error_writing_file
    return

997 call logmsg(log_error,"Error in WriteAMDtoBinaryFile: "//&
                "Could not convert BRC counter to a 6 digit string")
    error_flag = error_writing_file
    return

998 print *,"ABORTING subroutine WriteAMDtoBinaryFile"
    close(unit=fileunit)
    call free_lun(fileunit)    
    return

999 print *,"ERROR could not open outputfile: ",trim(filename_dbl)
    error_flag = error_opening_file
    return

  end subroutine WriteAMDtoBinaryFile
    !  #]  
  !---------------------------------
end module WriteAMDdata

