program application_client_example
  !  #[ Description
  !     An example of the calling sequence for the L2B processor subroutines
  !
  !     This example can serve for implementation inside an NWP assimilation
  !     system, for example.
  !     
  !     Created 18-12-2007 by: P. Poli, METEO-FRANCE
  !
  !     Note: Try not to be scared by the length of this program. 
  !     Half of it is made of instructions (catch_error) whose only 
  !     purposes are to stop the program if an error occurs
  !
  !     Modifications:
  !        18-Dec-2007 P. Poli     Initial version
  !        16-Jan-2008 J. de Kloe  phase out integer kind i_
  !  #]
  !  #[ modules used

  ! ParKind1 contains the integer/float variable type definitions
  USE ParKind1
  
  ! L2BP_module contains high-level points-of-entry/exit for the processor
  USE L2BP_module,            only:L2BP_primary, &
                                   aeolus_l2bp_setup, &
                                   aeolus_l2bp_import_L1B_Headers, &
                                   aeolus_l2bp_import_one_L1B_BRC, &
                                   aeolus_l2bp_import_AMD_Headers, &
                                   aeolus_l2bp_import_one_AMD_BRC, &
                                   aeolus_l2bp_prepare_primary, &
                                   aeolus_l2bp_export_one_L1B_BRC, &
                                   aeolus_l2bp_fill_L2B_Headers, &
                                   aeolus_l2bp_export_L1B_Headers, &
                                   aeolus_l2bp_export_L2B_Headers, &
                                   aeolus_l2bp_export_one_L2B_BRC, &
                                   aeolus_l2bp_export_AMD_Headers, &
                                   aeolus_l2bp_export_one_AMD_BRC, &
                                   aeolus_l2bp_unsetup, &
                                   aeolus_l2bp_load_data_to_vdas, &
                                   aeolus_l2bp_clear_vdas

  USE L2B_Proc_Settings,      only:L2BP_Settings_type
  USE ReadAMDdata,            only:Load_All_AMD_Profiles
  USE JobOrder_DataStructure, only:JobOrderData_type
  USE ReadL1Bdata,            only:ReadL1Bfile
  USE L1B_DataStructure,      only:L1B_HDR_DataType, L1B_BRC_DataType, &
                                   InitLatLonDateTimeRanges, &
                                   DetermineLatLonDateTimeRanges,&
                                   GetLatLonRanges, GetDateTimeRange
  USE L2BC_DataStructure,     only:L2BC_HDR_DataType, L2BC_BRC_DataType, &
                                   Init_L2BC_BRC_DataStructure
  USE WriteL2BCdata,          only:WriteL2BCfile, WriteProductList
  USE AMD_DataStructure,      only:AMD_HDR_DataType, &
                                   AMD_BRC_DataType
  USE RBC_DataStructure,      only:RBC_DataType
  USE Working_DataStructure,  only:WorkingDataType
  USE EE_CFI_DataTypes,       only:DateTimeType, IntAuc, IntAus, IntAl, &
                                   CheckWithinTimeRange
  ! support modules
  USE ErrorHandler,           only:catch_error, &
                                   no_error, error_allocate, &
                                   error_deallocate
  !  #]
  !  #[ Variables

  implicit none 

  ! local variables
  integer                  :: iBRC, Num_BRC, Num_AMD_BRC, AMD_iBRC
  integer                  :: AMD_file, L2B_iBRC
  integer                  :: error_flag, AllocStatus, DeAllocStatus
  logical                  :: ProcessBRC,IsWithinTimeRange
  integer(IntAus)          :: Num_L2B_BRCs
  integer(IntAl)           :: lat_start, lat_stop, lon_start, lon_stop
  integer(IntAuc)          :: M_Meas0, N_Meas_to_write
  integer(IntAuc)          :: M_Mie0, N_Obs_Mie_to_write
  integer(IntAuc)          :: M_Rayleigh0, N_Obs_Rayleigh_to_write
  character(len=256)       :: filename_JobOrder
  type(DateTimeType)       :: DateTimeStart, DateTimeStop
  logical                  :: do_standalone_processing

  ! a structure to hold the JobOrder settings
  TYPE(JobOrderData_type)  :: JobOrderData
  ! a structure to hold all processor settings
  TYPE(L2BP_Settings_type) :: L2BP_Settings
  ! a struct to hold the necessary input data for the processor
  TYPE(L1B_BRC_DataType)   :: L1B
  ! a struct to hold the necessary input data for the processor
  TYPE(L1B_HDR_DataType)   :: L1B_Headers
  ! a struct to hold all results from the processor
  TYPE(L2BC_BRC_DataType), dimension(:), pointer :: L2B
  ! a struct to hold all header results from the processor
  TYPE(L2BC_HDR_DataType)  :: L2B_Headers
  ! a struct to hold all NWP data
  TYPE(AMD_BRC_DataType), dimension(:), pointer :: AMD
  TYPE(AMD_HDR_DataType)   :: AMD_Headers
  ! a struct to hold all RBC data
  TYPE(RBC_DataType)       :: RBC
  ! a struct to hold all WorkingData
  TYPE(WorkingDataType)    :: WorkingData

  !  #]
  !  #[ Main Program code

  ! basic init
  error_flag = no_error
  do_standalone_processing=.FALSE.

  ! 1.1 L2BP initialization, get L2BP settings and filenames
  call aeolus_l2bp_setup (filename_JobOrder, JobOrderData, L2BP_Settings, RBC, error_flag, &
          standalone_version=do_standalone_processing)
  call catch_error(error_flag,"aeolus_l2bp_setup")

  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! 1.2 Read all the L1B data and all the AMD data ; store them inside a virtual
  ! data assimilation system (module virtual_das)
  call aeolus_l2bp_load_data_to_vdas(L2BP_Settings, error_flag)
  call catch_error(error_flag,"aeolus_l2bp_load_input_data_to_virtual_das")

  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! 2.0 loop on all the AMD files
  AMD_file_loop: DO AMD_file = 1, L2BP_Settings%NumNWPfiles
  !  #[

          ! 2.1 prepare processing for AMD file number AMD_file: get L1B_Headers, 
          ! allocate L2B array, and read AMD data
          call InitLatLonDateTimeRanges ()
          call aeolus_l2bp_import_L1B_Headers (L1B_Headers, &
                  Num_BRC, error_flag)
          call catch_error(error_flag,"aeolus_l2bp_import_L1B_Headers")
          allocate (L2B(Num_BRC),stat=AllocStatus)
          IF (AllocStatus.ne.0) error_flag=error_allocate
          call catch_error(error_flag,"allocate L2B(Num_AMD_BRC)")
          DO iBRC = 1, Num_BRC
                  call Init_L2BC_BRC_DataStructure (L2B(iBRC), &
                                       0, 0_IntAuc, 0_IntAuc, 0, error_flag)
                  call catch_error(error_flag,"Init_L2BC_BRC_DataStructure")
          ENDDO
          call aeolus_l2bp_import_AMD_Headers (AMD_Headers, &
                  Num_AMD_BRC, error_flag, &
                  iAMD_file=AMD_file)
          call catch_error(error_flag,"aeolus_l2bp_import_AMD_Headers")
          allocate (AMD(Num_AMD_BRC),stat=AllocStatus)
          IF (AllocStatus.ne.0) error_flag=error_allocate
          call catch_error(error_flag,"allocate AMD(Num_AMD_BRC)")
          DO AMD_iBRC = 1, Num_AMD_BRC
                  call aeolus_l2bp_import_one_AMD_BRC (AMD_Headers,  &
                          AMD(AMD_iBRC),  &
                          AMD_iBRC, error_flag,  &
                          iAMD_file=AMD_file)
                  call catch_error(error_flag,"aeolus_l2bp_import_one_AMD_BRC")
          ENDDO

          !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
          ! 2.2 start loop processing for all BRCs corresponding to AMD file number AMD_file
          L2B_iBRC = 0
          !  #[
          brc_loop: DO iBRC = 1, Num_BRC

                  ! 2.2.1 get the instrument data for that particular BRC, and verify time matching
                  call aeolus_l2bp_import_one_L1B_BRC (L1B_Headers, L1B, iBRC, &
                          error_flag) !, L1B_filename=L2BP_Settings%L1Bfile%name)
                  call catch_error(error_flag,"aeolus_l2bp_import_one_L1B_BRC")
                  ProcessBRC = .TRUE.
                  call CheckWithinTimeRange &
                          (L1B%Geoloc_ADS%start_of_obs_datetime, &
                          L2BP_Settings%L1Bfile%start, L2BP_Settings%L1Bfile%stop, &
                          IsWithinTimeRange, error_flag)
                  call catch_error(error_flag,"CheckWithinTimeRange L1B vs L2B")
                  IF (.not.(IsWithinTimeRange)) ProcessBRC=.FALSE.
                  call CheckWithinTimeRange &
                          (L1B%Geoloc_ADS%start_of_obs_datetime, &
                          L2BP_Settings%NWPfile(AMD_file)%start, &
                          L2BP_Settings%NWPfile(AMD_file)%stop, &
                          IsWithinTimeRange, error_flag)
                  call catch_error(error_flag,"CheckWithinTimeRange L1B vs AMD")
                  IF (.not.(IsWithinTimeRange)) ProcessBRC=.FALSE.

                  ! 2.2.2 do actual processing
                  IF (ProcessBRC) THEN

                          ! 2.2.2.1 start L2B processing, including finding a matching profile for the BRC 
                          ! in the meteorological data
                          L2B_iBRC = L2B_iBRC + 1
                          call aeolus_l2bp_prepare_primary (L1B, L2B(L2B_iBRC), &
                                  AMD, iBRC, &
                                  L2BP_Settings%L2B_AuxPar, &
                                  WorkingData, error_flag)
                          call catch_error(error_flag,"aeolus_l2bp_prepare_primary")
                          call L2BP_primary (L1B, L2B(L2B_iBRC), WorkingData, &
                                  RBC, L2BP_Settings%L2B_Auxpar, error_flag)
                          call catch_error(error_flag,"L2BP_primary")

                          ! 2.2.2.2 update information regarding the date/time ranges
                          call DetermineLatLonDateTimeRanges (L1B, error_flag)
                          call catch_error(error_flag,"DetermineLatLonDateTimeRanges")
                  ENDIF

                  ! 2.2.3 clear instrument data from memory
                  call aeolus_l2bp_export_one_L1B_BRC (L1B, error_flag)
                  call catch_error(error_flag,"aeolus_l2bp_export_one_L1B_BRC")
          !  #]
          ENDDO brc_loop
          ! done looping on the brcs
          !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

          ! 2.3 gather statistics on the processed BRCs and fill L2B_Headers
          Num_L2B_BRCs = int(L2B_iBRC, IntAus)
          call GetLatLonRanges (lat_start, lat_stop, lon_start, lon_stop, error_flag)
          call catch_error(error_flag,"GetLatLonRanges")
          call GetDateTimeRange (DateTimeStart, DateTimeStop, error_flag)
          call catch_error(error_flag,"GetDateTimeRanges")
          call aeolus_l2bp_fill_L2B_Headers (L1B_Headers, L2B, L2B_Headers, &
                  L2BP_Settings, &
                  Num_L2B_BRCs, DateTimeStart, DateTimeStop, &
                  lat_start, lat_stop, lon_start, lon_stop, &
                  M_Meas0, M_Mie0, M_Rayleigh0, error_flag)                
          call catch_error(error_flag,"aeolus_l2bp_fill_L2B_Headers")
          N_Meas_to_write = M_Meas0
          N_Obs_Mie_to_write = M_Mie0
          N_Obs_Rayleigh_to_write = M_Rayleigh0

          ! 2.4 write the L2B data to file
          call WriteL2BCfile (L2B_iBRC, L2B_Headers, L2B, N_Meas_to_write, &
                  N_Obs_Mie_to_write, N_Obs_Rayleigh_to_write, &
                  L2BCresultdir=L2BP_Settings%L2Bresultdir, &
                  L1Bfilename=L2BP_Settings%L1Bfile%name, &
                  NWPfilename=L2BP_Settings%NWPfile(AMD_file)%name, &
                  RBCfilename=L2BP_Settings%RBCfilename, &
                  CLMfilename="[undefined]", CALfilename="[undefined]", &
                  ParamFilename=L2BP_Settings%L2BAuxPar_filename, &
                  error_flag=error_flag)
          call catch_error(error_flag,"WriteL2BCfile")

          ! 2.5 clear L1B_Headers, L2B_Headers, and all AMD data from memory
          call aeolus_l2bp_export_L1B_Headers(L1B_Headers, error_flag)
          call catch_error(error_flag,"aeolus_l2bp_export_L1B_Headers")
          call aeolus_l2bp_export_L2B_Headers(L2B_Headers, error_flag)
          call catch_error(error_flag,"aeolus_l2bp_export_L2B_Headers")
          IF (associated(L2B)) THEN
                  DO iBRC = 1, size(L2B)
                          call aeolus_l2bp_export_one_L2B_BRC (L2B(iBRC), error_flag)
                          call catch_error(error_flag,"aeolus_l2bp_export_one_L2B_BRC")
                  ENDDO
          ENDIF
          deallocate (L2B,stat=DeAllocStatus)
          IF (DeAllocStatus.ne.0) error_flag=error_deallocate
          call catch_error(error_flag,"deallocate L2B")
          call aeolus_l2bp_export_AMD_Headers (AMD_Headers, error_flag)
          call catch_error(error_flag,"aeolus_l2bp_export_AMD_Headers")
          IF (associated(AMD)) THEN
                  DO AMD_iBRC = 1, size(AMD)        
                          call aeolus_l2bp_export_one_AMD_BRC (AMD(AMD_iBRC), &
                                  error_flag)
                          call catch_error(error_flag,"aeolus_l2bp_export_one_AMD_BRC")
                  ENDDO
          ENDIF
          deallocate (AMD,stat=DeAllocStatus)
          IF (DeAllocStatus.ne.0) error_flag=error_deallocate
          call catch_error(error_flag,"deallocate AMD")
  !  #]
  ENDDO AMD_file_loop
  ! done looping on the AMD files
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! 3.0 clean-up remaining items from memory
  call aeolus_l2bp_unsetup (JobOrderData, L2BP_Settings, RBC)
  
  ! 3.1 clear virtual data assimilation system (module virtual_das)
  call aeolus_l2bp_clear_vdas (error_flag)
  call catch_error(error_flag,"aeolus_l2bp_clear_vdas")

  ! if you manage to get there, you (probably) did not encounter 
  ! any visible error... good for you!

  !  #]
end program application_client_example

