program TestWriteAuxClimData
  !  #[ Documentation
  !     a test program for the WriteAuxClimData module
  !     written by: J. de Kloe
  !     created:    26-Jul-2006
  !
  !     Modifications:
  !     01-Aug-2006 J. de Kloe  added routine to fill the struct with data
  !     30-Jul-2007 J. de Kloe  adapt to changed interface of Fill_FH_Items
  !     14-Nov-2007 J. de Kloe  subtracted 1 millisecond from timestep
  !     07-Dec-2007 J. de Kloe  adapted to FH/MPH interface changes
  !     16-Jan-2008 J. de Kloe  phase out integer kind i_
  !     25-Feb-2008 D. Tan      Proc_Center from FixedHeader.System
  !     25-Feb-2008 J. de Kloe  merged in comments for same fix
  !  #]
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE WriteAuxClimData, only: InitWriteAuxClimFileModule,&
                              WriteAuxClimFile,Fill_AuxClim_Headers
  USE AuxClim_DataStructure, only: AuxClim_DataType, &
                                   Init_AuxClim_DataStructure, &
                                   Delete_AuxClim_DataStructure, &
                                   Init_AuxClim_DS, ClimData_type, &
                                   Construct_Test_Climatology, &
                                   Print_AuxClim_DataStructure
  USE AuxClim_SPH, only: maxlen_AuxClim_Ref_Name

  USE FixedHeader,  only: systemstr_LOCAL_MET_INST, &
                          maxlen_file_class, maxlen_system
  USE MainProductHeader, only: maxlen_Proc_Stage, maxlen_Proc_Center
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE ErrorHandler, only: no_error, program_abort, error_allocate, &
                          error_programming
  USE Numerics, only: r8_
  USE EE_CFI_Datatypes, only: &
       DateTimeType, Construct_KVT_DateTime, PrintDateTime, SetDateTime, &
       IntAs, IntAl, FAdoxy, maxlen_DateTime_KVT, IncrementDateTime
  !  #]
  !  #[ local variables and parameters
  implicit none

  ! used to fill the FH
  character(len=maxlen_file_class) :: file_class
  TYPE(DateTimeType)               :: validity_start_datetime
  TYPE(DateTimeType)               :: validity_stop_datetime
  real(FAdoxy)                     :: validity_period 
  integer                          :: file_version
  character(len=maxlen_system)     :: system
  ! used to fill the MPH
  character(len=maxlen_Proc_Stage)  :: proc_stage
  character(len=maxlen_Proc_Center) :: proc_center
  logical            :: product_err
  ! used to fill the SPH
  character(len=maxlen_AuxClim_Ref_Name) :: AuxClim_Ref_Name
  ! others
  integer                :: error_flag, AllocStatus
  TYPE(AuxClim_DataType) :: AuxClim
  character(len=maxlen_DateTime_KVT) :: DateTimeString
  !  #]
  !  #[ program code
  print *,"starting TestWriteAuxClimData"

  !*** initialisations ***
  error_flag = no_error

  call InitWriteAuxClimFileModule(error_flag, ChoosenVerbosity=VerbosityReport)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"InitWriteAuxClimFileModule()")

  call Init_AuxClim_DataStructure(AuxClim)

  allocate(AuxClim%AuxClim_DS,STAT=AllocStatus)
  IF (AllocStatus .ne. 0) &
       call program_abort(error_allocate,"TestWriteAuxClimData")

  ! allocate and initialise the whole structure
  ! (and specify the dimensions of the look-up table)
  call Init_AuxClim_DS(AuxClim%AuxClim_DS,&
       Num_DateTime_Ranges  =  4_IntAs,& ! small test file used for
       Num_Latitude_Ranges  = 18_IntAs,& ! Gert-Jan's first climatology
       Num_Longitude_Ranges = 36_IntAs,& ! proposal
       Num_Altitude_Ranges  =  4_IntAs,&
!       Num_DateTime_Ranges  = 4_IntAs,& ! very small test file
!       Num_Latitude_Ranges  = 6_IntAs,&
!       Num_Longitude_Ranges = 2_IntAs,&
!       Num_Altitude_Ranges  = 5_IntAs,&
!       Num_DateTime_Ranges  = 4_IntAs,&  ! medium sized test file
!       Num_Latitude_Ranges  = 90_IntAs,&
!       Num_Longitude_Ranges = 180_IntAs,&
!       Num_Altitude_Ranges  = 10_IntAs,&
!       Num_DateTime_Ranges  = 12_IntAs,& ! huge test file
!       Num_Latitude_Ranges  = 180_IntAs,&
!       Num_Longitude_Ranges = 360_IntAs,&
!       Num_Altitude_Ranges  = 30_IntAs,&
       error_flag=error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"Init_AuxClim_DS")

!  ! fill the datastruct with dummy data
!  call Fill_RBC_DS_with_Dummy_data(RBC%RBC_DS,error_flag)
!
  ! set some test values for FH input
  file_class = "TEST"
  file_version = 1
  system = systemstr_LOCAL_MET_INST
  call SetDateTime(validity_start_datetime,2001,03,12,12,34,56.0_FAdoxy)
  validity_period = 1._R8_*(10*365+2)*24*60*60 - 0.001 ! 10 years in seconds
  ! +2 days is added because there are 2 leap years in this period !
  
  ! NOTE: subtract 1 millisecond to make sure the ceil() returns
  ! with the proper time here, otherwise the ceil() used to get
  ! FH%Validity_Period%Validity_Stop might return with 1 more
  ! due to machine inaccuracy ! (And this makes the result not portable!)

  validity_stop_datetime = validity_start_datetime
  call IncrementDateTime(validity_stop_datetime,validity_period)

  ! set some test values for MPH input
  ! proc_stage might be: N (near-real-time), R (reprocessing) or T (test)
  proc_stage  = "T"     ! indicating this is a only testfile
  proc_center = system(1:maxlen_Proc_Center)
                        ! the location where this file was produced
                        ! Copy (possibly truncated) of FixedHeader system
                        ! (new requirement in spr 28)
  product_err = .false. ! no errors in this test product

  ! some test prints
  call PrintDateTime(validity_start_datetime)

  ! this 2 step approach is needed to prevent the xlf90 compiler
  ! from complaining about recursive IO
  DateTimeString = Construct_KVT_DateTime(validity_start_datetime)
  print *,"validity_start_datetime:KVT formatted = "//trim(DateTimeString)
       

  ! set some test values for SPH input
  AuxClim_Ref_Name = "test of AuxClim ref name"

  call Fill_AuxClim_Headers(AuxClim, &
       file_class, &                                      ! FH inputs
       validity_start_datetime, validity_stop_datetime, & ! FH inputs
       file_version, system, &                            ! FH inputs
       proc_stage, proc_center, product_err, &            ! MPH inputs
       AuxClim_Ref_Name, &                                ! SPH inputs
       error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"Fill_RBC_Headers()")

  print *,"filling the datastruct with data"
  call Construct_Test_Climatology(AuxClim,error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"Construct_Test_Climatology()")

  print *,'TestWriteAuxClimData: calling WriteAuxClimFile'
  ! for the 4x90x180x10 file this takes 38 seconds
  call WriteAuxClimfile(AuxClim, error_flag)
  ! for the 4x90x180x10 file this takes 26 seconds
  !call WriteAuxClimfile(AuxClim, error_flag,UseBuffering=.true.)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"WriteAuxClimFile")

  ! test print
  call Print_AuxClim_DataStructure(AuxClim,prefix="AuxClim",PrintAll=.true.,&
                                   error_flag=error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"Print_AuxClim_DataStructure")

  call Delete_AuxClim_DataStructure(AuxClim)

  print *,"done writing AuxClim test datafile"
  !  #]
end program TestWriteAuxClimData
