module WriteAuxClimData
  !  #[ Description:
  !
  !     a module to write the 2 files describing the Auxiliary
  !     climatology look-up table of the Aeolus instrument
  !
  ! created:     26-Jul-2006 by J. de Kloe
  ! 
  ! History:
  !   17-Nov-2006 J. de Kloe  small interface change for Fill_DSD_items
  !   07-Jun-2007 J. de Kloe  adapted to changed xml interface
  !   20-Jul-2007 J. de Kloe  added call to xml_tree_set_schema()
  !   30-Jul-2007 J. de Kloe  adapt to changed interface of Fill_FH_Items
  !   18-Oct-2007 J. de Kloe  adapt to changed interface of Fill_MPH_Items
  !   05-Nov-2007 J. de Kloe  added setting of the xml namespace
  !   07-Dec-2007 J. de Kloe  adapted to FH/MPH interface changes
  !   16-Jan-2008 J. de Kloe  phase out integer kind i_
  !   21-Jan-2008 J. de kloe  added workaround in call to GetByteSize_MPH_KVT
  !                           to not change the fileformat for now.
  !   21-Feb-2008 J. de Kloe  removed the workaround again, since it was
  !                           caused by a misunderstanding on L1B side
  !   22-Feb-2008 D. Tan      Change Ref_Doc value
  !   25-Feb-2008 J. de Kloe  corrected filling of mph%ref_doc field
  !
  !  #]
  !  #[ Modules used:
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE AuxClim_DataStructure, only: AuxClim_DataType, &
                                   Init_AuxClim_DataStructure,&
                                   Delete_AuxClim_DataStructure, &
                                   GetByteSize_AuxClimDataStr_KVT, &
                                   GetByteSize_AuxClim_DS, &
                                   Write_AuxClim_DataStructure
  USE FixedHeader, only: Init_FH, Copy_FH_to_xml_tree, &
                         Fill_FH_Items, file_type_AuxClim, &
                         maxlen_file_type, maxlen_file_description, &
                         maxlen_notes, maxlen_creator, maxlen_creator_version
  USE MainProductHeader, only: Init_MPH, Copy_MPH_to_xml_tree, &
                               GetByteSize_MPH_KVT, Fill_MPH_items, &
                               maxlen_Product, maxlen_Ref_Doc, &
                               maxlen_Software_Ver
  USE AuxClim_SPH, only: Copy_AuxClim_SPH_to_xml_tree, &
                         Write_AuxClim_SPH_to_KVT, &
                         maxlen_SPH_Descriptor, maxlen_AuxClim_Ref_Name, &
                         Fill_AuxClim_SPH_items, Init_AuxClim_SPH, &
                         GetByteSize_AuxClim_SPH_KVT
  USE DataSetDescriptor, only: Init_DSD, Print_DSD, DSD_type, &
       Fill_DSD_items,  maxlen_DS_Name, maxlen_DS_Type

  USE XML_module, only: xml_node_type, &
       Init_XML, xml_tree_print, xml_tree_write, create_xml_tree, &
       xml_tree_create_header, delete_xml_tree, ee_h_name, &
       xml_tree_set_namespace, uri_ns_header

  USE KVT_module, only: max_str_len, Init_KVT_module
  USE DirectBinaryIO, only: Init_DirectBinaryIO_module , &
                            OpenDirectBinaryFile, CloseDirectBinaryFile, &
                            BigEndian, write_zero_bytes_bin_file, &
                            DirectBinaryIO_fileunit_type
  USE ErrorHandler, only: no_error, error_xml, error_programming, &
                          error_allocate, &
                          error_writing_file 
!                          error_opening_file, & ! error_reading_file,&
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE LunManager, only: get_lun, free_lun
  USE Compiler_Features, only: newline
  USE EE_CFI_Datatypes, only: FAdoxy, IntAd, IntAl
  ! IntAul, IntAs, IntAus, &
  ! FAdoxy_to_IntAul, FAdoxy_to_IntAus

  USE EE_CFI_Datatypes, only: DateTimeType!,Construct_KVT_DateTime, &
!                              SetDateTime, PrintDateTime
!                              InitDateTime, IncrementDateTime, &
!                              TestDateTime, &
!                              DateTime2String, GetCurrentDateTime
  USE AeolusConstants, only: software_version_L2BP => software_version, &
                             iodd_version_AuxClim, Ref_Doc_AuxCLM
  !  #]  
  !  #[ Parameter/Variable/Type declarations

  implicit none

  ! make all module contents public, unless explicitely stated otherwise
  public

  ! the default verbosity setting is Error (as silent as possible)
  integer, private, save :: LocalVerbosity = VerbosityError

  ! remember the initialisation state of the module
  logical, private, save :: module_is_initialised = .false.

  !  #]
  !  #[ Module interface
  !private :: 
  !  #]
contains
  !---------------------------------
  subroutine InitWriteAuxClimFileModule(error_flag, ChoosenVerbosity)
    !  #[

    integer,           intent(out) :: error_flag
    integer, optional, intent(in)  :: ChoosenVerbosity

    error_flag = no_error

    IF (present(ChoosenVerbosity)) LocalVerbosity = ChoosenVerbosity

    IF (LocalVerbosity .ge. VerbosityReport) &
         call print_verbosity_setting(LocalVerbosity,&
                                      ModuleName="WriteAuxClimData")

    call Init_KVT_module()

    call Init_DirectBinaryIO_module(error_flag,FileByteOrder=BigEndian)
    if (error_flag .ne. no_error) return

    call Init_XML(error_flag) ! (ChoosenVerbosity=VerbosityReport)
    if (error_flag .ne. no_error) return

    module_is_initialised = .true.

  end subroutine InitWriteAuxClimFileModule
    !  #]
  subroutine WriteAuxClimFile(AuxClim, error_flag, UseBuffering)
    !  #[
    type(AuxClim_DataType), intent(inout) :: AuxClim
    integer,                intent(out)   :: error_flag
    logical,      optional, intent(in)    :: UseBuffering

    ! local variables
    character(len=256) :: filename_hdr
    character(len=256) :: filename_dbl
    character(len=256) :: path
    character(len=256) :: xml_namespace_CLM

    ! general xml items
    character(len=*), parameter :: xml_root = "Earth_Explorer_File"
    TYPE(xml_node_type), pointer :: xml_tree

    ! init
    error_flag = no_error

    ! check if we have a FH
    IF (.not. associated(AuxClim%FH)) THEN
       print *,"ERROR in WriteAuxClimFile():"
       print *,"the AuxClim%FH pointer seems not associated !!!!"
       print *,"Unable to write this datafile ...."
       error_flag = error_programming
       return
    END IF

    ! check if we have a MPH
    IF (.not. associated(AuxClim%MPH)) THEN
       print *,"ERROR in WriteAuxClimFile():"
       print *,"the AuxClim%MPH pointer seems not associated !!!!"
       error_flag = error_programming
       return
    END IF

    ! check if we have a SPH
    IF (.not. associated(AuxClim%SPH)) THEN
       print *,"ERROR in WriteAuxClimFile():"
       print *,"the AuxClim%SPH pointer seems not associated !!!!"
       error_flag = error_programming
       return
    END IF

    ! construct the filenames to be used by adding the extensions
    filename_hdr = trim(AuxClim%FH%File_Name)//".HDR"
    filename_dbl = trim(AuxClim%FH%File_Name)//".DBL"

    !IF (LocalVerbosity .ge. VerbosityReport) THEN &
    print *,"filename_hdr = ",trim(filename_hdr)
    print *,"filename_dbl = ",trim(filename_dbl)
    !END IF

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in subroutine WriteAuxClimFile:"
       print *,"the WriteAuxClimData module should be initialised before use !"
       error_flag = error_programming
       return
    END IF

    !IF (LocalVerbosity .ge. VerbosityReport) &
    print *,"WriteAuxClimFile: Writing AuxClim output file..."

    ! allocate a datastructure to hold the xml data
    ! get a pointer to it, and fill it with the xml root tag pair:
    ! <Earth_Explorer_File><\Earth_Explorer_File>

    ! allocate and init the datastructure
    nullify(xml_tree)
    call create_xml_tree(xml_tree, error_flag)
    IF (error_flag .ne. no_error) return

    ! create a blank header (fixed+variable) and define the root
    call xml_tree_create_header(xml_tree, ee_h_name, error_flag)
    IF (error_flag .ne. no_error) return

    !example: xmlns="http://www.esa.int/schemas/ae/AUX_CLM_L2_01.01"
    !uri_ns_header       ! defined in xml_module
    !file_type_AuxClim   ! defined in FixedHeader
    !iodd_version_AuxCLM ! defined in AeolusConstants
    xml_namespace_CLM = trim(uri_ns_header)//trim(file_type_AuxClim)//&
                        "_"//trim(iodd_version_AuxClim)

    ! define which  xml namespace should be used
    call xml_tree_set_namespace(xml_tree,xml_namespace_CLM,error_flag)
    IF (error_flag .ne. no_error) return

    ! fill all FH items in the xml tree
    path = "/"//ee_h_name
    call Copy_FH_to_xml_tree(AuxClim%FH, path, xml_tree, error_flag)
    IF (error_flag .ne. no_error) return

    ! fill all MPH items in the xml tree
    call Copy_MPH_to_xml_tree(AuxClim%MPH, path, xml_tree, error_flag)
    IF (error_flag .ne. no_error) return

    ! fill all SPH items in the xml tree
    call Copy_AuxClim_SPH_to_xml_tree(AuxClim%SPH, path, xml_tree, error_flag)
    IF (error_flag .ne. no_error) return

    print *,'AuxClim%SPH%AuxCLim_DSD%start_of_ds = ', &
             AuxClim%SPH%AuxClim_DSD%start_of_ds

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"WriteAuxClimFile: Writing XML header file ..."
       print *,trim(filename_hdr)
    END IF

    ! Write the HDR file including FH, MPH, and SPH (with DSD)
    call xml_tree_write(xml_tree, filename_hdr, error_flag)
    IF (error_flag .ne. no_error) return

    !IF (LocalVerbosity .ge. VerbosityReport) THEN
       !print *,"==================="
       !print *,"the generated xml tree was successfully written to file"
       !print *,"==================="
       !call xml_tree_print(xml_tree, error_flag)
       !IF (error_flag .ne. no_error) return
       !print *,"==================="
    !END IF

    call delete_xml_tree(xml_tree,error_flag)
    IF (error_flag .ne. no_error) return

    ! Write the DBL file including MPH, SPH, and data and figure out size
    call WriteAuxClimtoBinaryFile(AuxClim,filename_dbl,error_flag, &
                                  UseBuffering)

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"WriteAuxClimfile: done writing AuxClim output file"

  end subroutine WriteAuxClimFile
    !  #]
  !---------------------------------
  subroutine WriteAuxClimtoBinaryFile(AuxClim,filename_dbl,error_flag, &
                                      UseBuffering)
    !  #[
    type(AuxCLim_DataType), intent(inout) :: AuxCLim
    integer,                intent(out)   :: error_flag
    character(len=256),     intent(in)    :: filename_dbl
    logical,      optional, intent(in)    :: UseBuffering

    ! local variables
    TYPE(DirectBinaryIO_fileunit_type) :: fileunit

    error_flag = no_error

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"WriteAuxClimtoBinaryFile: writing AuxClim data to binary file:"
       print *,trim(filename_dbl)
    END IF

    call OpenDirectBinaryFile(filename_dbl,"write",fileunit,error_flag, &
                              UseBuffering)
    IF (error_flag .ne. no_error) goto 999

    call Write_AuxClim_DataStructure(fileunit,AuxClim,error_flag)
    IF (error_flag .ne. no_error) goto 999

    !------------------------------
    ! done
    !------------------------------
    call CloseDirectBinaryFile(fileunit, error_flag)
    IF (error_flag .ne. no_error) return

    return

999 print *,"ABORTING subroutine WriteAuxCLimtoBinaryFile"
    call CloseDirectBinaryFile(fileunit, error_flag)
    return
    
  end subroutine WriteAuxClimtoBinaryFile
    !  #]  
  !---------------------------------
  subroutine Fill_AuxClim_Headers(AuxClim, & ! the data struct to be filled
       file_class, &                                      ! FH input
       validity_start_datetime, validity_stop_datetime, & ! FH inputs
       file_version, system, &                            ! FH inputs
       proc_stage, proc_center, product_err, &            ! MPH inputs
       AuxClim_Ref_Name, &                                ! SPH input
       error_flag)                                        ! result
    !  #[
    ! the datastructure to be filled
    type(AuxClim_DataType), intent(inout):: AuxClim
    ! inputs needed to fill the FH
    character(len=*),   intent(in)   :: file_class
    TYPE(DateTimeType), intent(in)   :: validity_start_datetime
    TYPE(DateTimeType), intent(in)   :: validity_stop_datetime
    integer,            intent(in)   :: file_version
    character(len=*),   intent(in)   :: system
    ! inputs needed to fill the MPH
    character(len=1),   intent(in)   :: proc_stage
    character(len=*),   intent(in)   :: proc_center    
    logical,            intent(in)   :: product_err
    ! inputs needed to fill the SPH
    character(len=maxlen_AuxClim_Ref_Name) :: AuxClim_Ref_Name
    ! success or fail ?
    integer,            intent(out)  :: error_flag
    
    ! local variables
    integer :: AllocStatus
    
    ! used to fill the FH (filled with constants below)
    character(len=maxlen_file_type)        :: file_type
    character(len=maxlen_file_description) :: file_description
    character(len=maxlen_notes)            :: notes
    character(len=maxlen_creator)          :: creator
    character(len=maxlen_creator_version)  :: creator_version
    
    ! used to fill the MPH
    character(len=maxlen_Product)      :: product
    character(len=maxlen_Ref_Doc)      :: ref_doc
    character(len=maxlen_Software_Ver) :: software_version
    integer(IntAd) :: tot_size
    integer(IntAl) :: sph_size, num_DSD, num_data_sets
    
    ! used to fill the SPH
    character(len=maxlen_SPH_Descriptor) :: SPH_Descriptor

    ! used to fill the DSD
    character(len=maxlen_DS_Name)       :: DS_Name
    character(len=maxlen_DS_Type)       :: DS_Type
    integer(IntAd)                      :: Ds_Offset
    integer(IntAl)                      :: Num_Dsr
    integer(IntAl)                      :: Dsr_Size
    
    error_flag = no_error
    
    !*** set the Fixed Header ***
    allocate(AuxClim%FH,STAT=AllocStatus)
    IF (AllocStatus.ne. 0) THEN
       print *,"Allocation problem in Fill_AuxClim_Headers"
       error_flag = error_allocate
       return
    END IF
    
    call Init_FH(AuxClim%FH, error_flag)
    IF (error_flag .ne. no_error) return
    
    ! these settings are constant, independent of the location
    ! where the code is executed, so fill them here
    file_description = "AuxClim look-up table file"
    notes            = "Room for some"//newline//"additional remarks"
    file_type        = file_type_AuxClim
    creator          = "AuxClimGen"
    creator_version  = software_version_L2BP
    call Fill_FH_items(AuxClim%FH, &
         file_description, notes, file_class, file_type, &
         validity_start_datetime, validity_stop_datetime, &
         file_version, system, &
         creator,creator_version,error_flag)
    IF (error_flag .ne. no_error) return
    
    print *,'Fill_AuxClim_Headers: Filled FH'
    
    !*** set the Main Product Header ***
    allocate(AuxClim%MPH,STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation problem in Fill_AuxClim_Headers"
       error_flag = error_allocate
       return
    END IF
    
    call Init_MPH(AuxClim%MPH, error_flag)
    IF (error_flag .ne. no_error) return

    product          = AuxClim%FH%File_Name
    !proc_stage       = defined in the calling program
    ref_doc          = Ref_Doc_AuxCLM
    !proc_center      = defined in the calling program
    software_version = software_version_L2BP
    !product_err      = defined in the calling program
    
    ! these are constants for the AuxClim filetype
    ! see L2B IODD V1.0, sec. 5.2
    tot_size      = GetByteSize_AuxClimDataStr_KVT(AuxClim) ! = variable !
    print *,"tot_size = ",tot_size
    sph_size      = GetByteSize_AuxClim_SPH_KVT() ! = 486
    print *,"sph_size = ",sph_size
    num_DSD       = 1_IntAl
    num_data_sets = 1_IntAl

    call Fill_MPH_items(AuxClim%MPH, product, proc_stage, ref_doc, &
                        proc_center, software_version, product_err, tot_size, &
                        sph_size, num_dsd, num_data_sets, &
                        validity_start_datetime, validity_stop_datetime, &
                        error_flag)
    IF (error_flag .ne. no_error) return
    
    print *,'Fill_AuxClim_Headers: Filled MPH'
    
    !*** set the Specific Product Header ***
    allocate(AuxClim%SPH,STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation problem in Fill_AuxClim_Headers"
       error_flag = error_allocate
       return
    END IF
    
    call Init_AuxClim_SPH(AuxClim%SPH, error_flag)
    IF (error_flag .ne. no_error) return
    
    SPH_Descriptor = "AuxClim look-up table file"
    
    call Fill_AuxClim_SPH_items(AuxClim%SPH, error_flag, &
                                SPH_Descriptor,AuxClim_Ref_Name)
    IF (error_flag .ne. no_error) return
    
    print *,'Fill_AuxClim_Headers: Filled SPH'
    
    !*** set the AuxClim Data Set Descriptor ***
    allocate(AuxClim%SPH%AuxClim_DSD,STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation problem in Fill_AuxClim_Headers"
       error_flag = error_allocate
       return
    END IF
    
    call Init_DSD(AuxClim%SPH%AuxClim_DSD, error_flag)
    IF (error_flag .ne. no_error) return
    
    DS_Name    = "AuxClim_ADS"
    ! WARNING: in the L2B datafile it is "R" but here it is "A" !
    DS_Type    = "A" 
    Ds_Offset  = GetByteSize_MPH_KVT() + &
                 GetByteSize_AuxClim_SPH_KVT() ! = 1733
    print *,"Ds_Offset = ",Ds_Offset
    Num_Dsr    = 1_IntAl
    Dsr_Size   = GetByteSize_AuxClim_DS(AuxClim%AuxClim_DS) ! = variable !
    print *,"Dsr_Size = ",Dsr_Size

    call Fill_DSD_items(AuxClim%SPH%AuxClim_DSD, &
         DS_Name, DS_Type, DS_Offset, Num_Dsr, Dsr_Size, error_flag)
    
    print *,'Fill_AuxClim_Headers: Filled DSD'
    
  end subroutine Fill_AuxClim_Headers
  !  #]
  !---------------------------------
end module WriteAuxClimData
