module L1B_BufrUtil
  !---------------------------------
  !  #[ Description:
  !
  ! - a module to form bufr sections and message
  !   describing the ADM Aeolus instrument data products
  !
  ! - here: L1B observation granularity
  !         = BUFR subtype 251
  !
  ! created:     16-11-2005 by M.Rohn (ECMWF)
  ! last change: 07-03-2006
  !
  ! Modifications:
  !   17-Jan-2006 M.Rohn            single template (OBS+MEAS)
  !   06-Mar-2006 M.Rohn            consolidated L1B input definition
  !   01-Sep-2006 D. Tan I. Mallas  remove  parameter
  !          L1B%PCD_ADS%Measurement_PCD(imeas)%mie_emitted_freq_std_dev
  !              set the related bufr value to missing
  !                                 set new parameter 
  !          L1B%PCD_ADS%Measurement_PCD(imeas)%reference_pulse_FWHM 
  !                                 for future use
  !   30-Nov-2006 D. Tan I. Mallas  MDA L1bP V1.04: Padding for Frequency 
  !                                 Step Response, increase NSPECTROMETER
  !   30-Nov-2006 D. Tan I. Mallas  Remove 2000 offset for ksec1/year
  !   15-Dec-2006 D. Tan            Array dimensions for Bufr edition 310
  !   19-Jan-2007 J. de Kloe        added extra tests to prevent problems
  !                                 with the log() function used in the dB
  !                                 calculation
  !   22-Jan-2007 J. de loe         added some annotations on the meaning of 
  !                                 the descriptors in the set*dd routines
  !   08-May-2007 J. de Kloe        added use of the descriptions in the
  !                                 fill_one_element routine
  !   03-Sep-2007 P. Poli           modified arguments of L1B_bufr_encode,
  !                                 fill_L1B_bufr_header to separate between
  !                                 header information and BRC-level
  !                                 change L1B_DataType to L1B_BRC_DataType
  !                                 and L1B_HDR_DataType
  !   22-Feb-2008 J. de Kloe        added optional FakeDateTime in the
  !                                 L1B_bufr_encode routine
  !
  !  #] ---------------------------------
  !  #[ Modules Used
  USE L1B_DataStructure, only: L1B_BRC_DataType, L1B_HDR_DataType

  USE DateTimeMod,       only: date_type, time_type, string2date 
  USE EE_CFI_Datatypes,  only: DateTimeType, Get_Date_from_DateTime, &
                               Get_Time_from_DateTime, &
                               IntAs, IntAuc
  USE L1B_Cal_ADS,       only: Meas_Calib_Valid_type, Ref_Pulse_Calib_Valid_type
  USE Logging, only: logmsg, log_debug,log_info,log_progress,log_warn,log_error
  USE LunManager, only: fileunit_stderr
  USE numerics, only: is_NaN, is_PosInf, is_NegInf, missing_real, missing_int,&
                      real2txt_extra_checks
  !                   real2dB this one is only available for 4 byte reals
  !                   TODO: create an 8 byte version in the numerics module
  !  #]
  !  #[ 0.0 Parameter/Variable/Type declarations

  implicit none


  ! 0.1 make all module contents public
  public

  ! 0.2 define precisions 
  integer, parameter :: i1_  = Selected_Int_Kind( 2)       ! = integer*1
  integer, parameter :: r8_  = Selected_Real_Kind(15,307)  ! = real*8

  ! 0.3 constants
  logical, parameter :: bufr_test = .false.

  integer, parameter :: BCOMPRESSION_FLAG            = 64 ! 64=compression / 0=no compression

  integer, parameter :: NMAX_OBS_SUBSET              = 10 ! pack max 10 subsets in message 

  ! for DEVELOPMENT
  integer, parameter :: NOBSERVATION_FOR_DEVELOPMENT = 10 ! currently L1B product contains only 10 obs

  integer, parameter :: NPROFILE                     =  24
  integer, parameter :: NSPECTROMETER                =  20
  integer, parameter :: NTRIPOD_CORR                 =  16
  integer, parameter :: NFREQUENCY_STEP              =  60

  integer, parameter :: NELEMENTS_L1B_HEAD           =    17
  integer, parameter :: NELEMENTS_L1B_OBS            =  1579
  integer, parameter :: NELEMENTS_L1B_MEAS           =  1348

  real(r8_), parameter :: DEGREE_SCALE               =  1000000.0
  real(r8_), parameter :: DEGREE_REF_LON             =      180.0

  ! 0.31 template types
  integer, parameter :: BEDITION              =   4
  integer, parameter :: BTABLE_MASTER         =   0
  integer, parameter :: BTABLE_MASTER_VERSION =  15
  integer, parameter :: BTABLE_LOCAL_VERSION  =   1

  integer, parameter :: BOBSTYPE         =    3 ! sounding
  integer, parameter :: RDBTYPE          =    2 ! sounding
  integer, parameter :: BSUBTYPE_L1B     =  251 ! l1b
  integer, parameter :: BSUBTYPE_L2B     =  253 ! l2b

  ! 0.32 code tables
  integer, parameter :: BCODE_SATID      =  600 ! ADM
  integer, parameter :: BCODE_INSTRUMENT =  130 ! ALADIN
  integer, parameter :: BCODE_CENTRE     =   98 ! ECMWF
  integer, parameter :: BCODE_SUBCENTRE  =    0 ! L2B processing facility = ? 

  ! significance data section
  integer, parameter :: BCODE_DATA_HEA =    0 ! header
  integer, parameter :: BCODE_DATA_GEO =    1 ! geolocation
  integer, parameter :: BCODE_DATA_PCD =    2 ! product confidence
  integer, parameter :: BCODE_DATA_GWD =    3 ! ground wind detection
  integer, parameter :: BCODE_DATA_MEA =    4 ! measurement
  integer, parameter :: BCODE_DATA_US  =    5 ! useful signal
  integer, parameter :: BCODE_DATA_WV  =    6 ! wind velocity
  integer, parameter :: BCODE_DATA_CC  =    7 ! calibration and characterization
  integer, parameter :: BCODE_DATA_AMD =    8 ! auxiliary meteorological data

  ! significance position
  integer, parameter :: BCODE_POS_START_BIN    =    0 ! start point height bin
  integer, parameter :: BCODE_POS_END_BIN      =    1 ! end point height bin
  integer, parameter :: BCODE_POS_BOTTOM_BIN   =    2 ! bottom height bin
  integer, parameter :: BCODE_POS_BOTTOM_TOP   =    3 ! top height bin
  integer, parameter :: BCODE_POS_CENTROID_BIN =    4 ! centroid height bin
  integer, parameter :: BCODE_POS_LOS_GEOID    =    5 ! intersection LOS Geoid

  ! significance time
  integer, parameter :: BCODE_TIME_VALID_START  =   0 ! start validity
  integer, parameter :: BCODE_TIME_VALID_END    =   1 ! end validity
  integer, parameter :: BCODE_TIME_LEAP_SECOND  =   2 ! occurence leap second
  integer, parameter :: BCODE_TIME_STATE_VECTOR =   3 ! state vector
  integer, parameter :: BCODE_TIME_SENSE_START  =  19 ! sensing start
  integer, parameter :: BCODE_TIME_SENSE_STOP   =  20 ! sensing stop

  integer, parameter :: BCODE_TIME_SCAN_START   =  28 ! sensing start
  integer, parameter :: BCODE_TIME_SCAN_STOP    =  29 ! sensing stop

  integer, parameter :: BCODE_TIME_DIFFERENCE   =   2 ! UTC-SAT_CLOCK

  ! receiver channel
  integer, parameter :: BCODE_CHANNEL_MIE       =   0 ! mie
  integer, parameter :: BCODE_CHANNEL_RAY       =   1 ! rayleigh
  integer, parameter :: BCODE_CHANNEL_RAY_A     =   2 ! rayleigh A
  integer, parameter :: BCODE_CHANNEL_RAY_B     =   3 ! rayleigh B

  ! significance statistics
  integer, parameter :: BCODE_STATISTIC_STDEV   =  10 ! stdev


  ! 0.33 bufr scaling
  ! these are identical to the missing_indicators used elsewhere
  ! in the L2BP software
  real(r8_), parameter :: RVIND = 1.7E38 !_r8_
  integer, parameter   :: NVIND = 2147483647

  integer, parameter :: jsec0  =       3, &
                        jsec1  =      40, &
                        jsec2  =    4096, &
                        jsec3  =       4, &
                        jsec4  =       2, &
                        jkey   =      52, &
                        ktdlen =     300, & ! max no unexpanded descriptors
                        krdlen =       1, & ! max no delayed replication factor
                        kelem  =  160000, & ! max no expanded descriptors
                        kvals  = 4096000, & ! max no values in message
                        jbufl  = 1600000    ! max length bufr message  

  ! 0.4 variables
  integer::             ksec0(jsec0), &
                        ksec1(jsec1), &
                        ksec2(jsec2), &
                        ksec3(jsec3), &
                        ksec4(jsec4), &
                        key(jkey), &      ! packed RDB key
                        ktdlst(ktdlen), & ! unxepanded descriptor list
                        kdata(krdlen), &  ! delayed replication values : not used
                        kbufr(jbufl)

  character(len=80)::   cvals(kvals)

  character(len=64)::   cnames(kelem)
  character(len=24)::   cunits(kelem)

  real(r8_)::           values(kvals)

  integer::             ktdlst_size, ktdexp_size

  real(r8_)::           key_lat_min, key_lat_max, &
                        key_lon_min, key_lon_max

  integer      :: key_year, &
                  key_month, & 
                  key_day, & 
                  key_hour, & 
                  key_minute, &
                  key_second

  !  #]
  !---------------------------------
contains
  !---------------------------------
  ! 1.0 L1B_bufr_sec0
  subroutine set_L1B_bufr_sec0()
    !  #[
   !
   integer:: i

   ksec0( 1)= 0
   ksec0( 2)= 0
   ksec0( 3)= BEDITION

   if(bufr_test) then
      write(*, '(a)'   ) ' ------ set_L1B_bufr_sec0 --------'
      write(*, '(a)'   ) ' ------ ksec0 '
      write(*, '(i15)' ) ( ksec0(i), i=1,jsec0 )
   endif

   return
  end subroutine set_L1B_bufr_sec0
  !  #]
  !---------------------------------
  ! 2.0 L1B_bufr_sec1
  subroutine set_L1B_bufr_sec1( &
                year,month,day,hour,minute )
    !  #[
   ! arguments
   integer, intent(in)    :: year,month,day,hour,minute

   ! local variables
   integer:: i

   ! 2.1 init
   do i  = 1, jsec1
      ksec1(i)= 0
   enddo

   ! 2.2 set
   ksec1( 1)=  22            ! length sec1 bytes
   ksec1( 2)= BEDITION       ! bufr edition
   ksec1( 3)= BCODE_CENTRE   ! originating centre
   ksec1( 4)=   1            ! update sequence
   ksec1( 5)= 128            ! (PRESENCE SECT 2) (0/128 = no/yes)
   ksec1( 6)= BOBSTYPE       ! message type 
   ksec1( 7)= BSUBTYPE_L1B   ! subtype
   ksec1( 8)= BTABLE_LOCAL_VERSION ! version of local table
   ksec1( 9)= year           ! Without offset year - 2000
   ksec1(10)= month
   ksec1(11)= day
   ksec1(12)= hour
   ksec1(13)= minute
   ksec1(14)= BTABLE_MASTER         ! master table
   ksec1(15)= BTABLE_MASTER_VERSION ! version of master table
   ksec1(16)= BCODE_SUBCENTRE ! originating subcentre: new entry for L2B processing facility?
   ksec1(17)=   0
   ksec1(18)=   0

   if(bufr_test) then
      write(*, '(a)'   ) ' ------ set_L1B_bufr_sec1 --------'
      write(*, '(a)'   ) ' ------ ksec1 '
      write(*, '(i15)' ) ( ksec1(i), i=1,jsec1 )
   endif

   return
  end subroutine set_L1B_bufr_sec1
  !  #]
  !---------------------------------
  ! 3.0 L1B_bufr_sec2
  subroutine set_L1B_bufr_sec2( nsubset )
    !  #[
    ! arguments
    integer,   intent(in)  :: nsubset

    ! local variables
    integer    i, error_flag

    ! modify unpacked RDB keys (ECMWF only)
    !---------------------------------
    do i=1,jkey
        key( i )= 0
    end do
    do i=1,jsec2
        ksec2( i )= 0
    end do

    key( 1)= 52                ! length sec2
    key( 2)= RDBTYPE           ! RDB obtype
    key( 3)= BSUBTYPE_L1B      ! RDB subtype

    key( 4)= ksec1( 9)         ! Without offset ksec1( 9) + 2000
    key( 5)= ksec1(10)
    key( 6)= ksec1(11)
    key( 7)= ksec1(12)
    key( 8)= ksec1(13)
    key( 9)= 0          ! second

    key(10)= nint( key_lon_min* 100000.) + 18000000      ! Lon 1
    key(11)= nint( key_lat_min* 100000.) +  9000000      ! Lat 1
    key(12)= nint( key_lon_max* 100000.) + 18000000      ! Lon 2
    key(13)= nint( key_lat_max* 100000.) +  9000000      ! Lat 2

    key(14)= nsubset                ! no subsets

    key(15)= BCODE_SATID          ! satellite ID
    ! CCITTIA5 idents
        do  i=1,9
          key(15+i)=ichar('1')
        enddo

!    key(25)= nbufl        ! length of BUFR message in bytes, not known before packing

    ! RDB, MDB insertion time stamp tbd
    key(26)= 0 ! day
    key(27)= 0 ! hour
    key(28)= 0 ! minute
    key(29)= 0 ! second
    key(30)= 0 ! day
    key(31)= 0 ! hour
    key(32)= 0 ! minute
    key(33)= 0 ! second

    key(46)= 0                 ! minimum % confidence

    !---------------------------------
    ! pack RDB

    ksec2( 1)= 52               ! length sec2 

    call bupkey(key,ksec1,ksec2,error_flag)

    if(bufr_test) then
        write(*, '(a)'    ) ' ------ set_L1B_bufr_sec2 --------'
        write(*, '(a)'    ) ' ------ ksec2 '
        write(*, '(5i15)' ) ( ksec2(i), i=1,jsec2 )
    endif

    return
  end subroutine set_L1B_bufr_sec2
  !  #]
  !---------------------------------
  ! 4.0 L1B_bufr_sec3
  subroutine set_L1B_bufr_sec3( nsubset )
    !  #[
    ! arguments
   integer, intent(in)    :: nsubset

   integer                :: i

   ksec3( 1)= 0
   ksec3( 2)= 0
   ksec3( 3)= nsubset            ! no of data subsets
   ksec3( 4)= BCOMPRESSION_FLAG  ! 64 FOR COMPRESSION/ 0 MANY SUBSETS

   if(bufr_test) then
      write(*, '(a)'   ) ' ------ set_L1B_bufr_sec3 --------'
      write(*, '(a)'   ) ' ------ ksec3 '
      write(*, '(i15)' ) ( ksec3(i), i=1,jsec3 )
   endif

   return
  end subroutine set_L1B_bufr_sec3
  !  #]
  !---------------------------------
  ! 5.0 set_LxB_bufr_header_dd
  subroutine set_LxB_bufr_header_dd( no_dd_start, no_dd_stop )
    !  #[
    ! arguments
    integer, intent(in )    :: no_dd_start
    integer, intent(out)    :: no_dd_stop
   
    ! local variables
    integer                :: i

    i = no_dd_start-1 

    ! 5.1 set descriptors
    !   origin satellite, centre
    i=i+1;   ktdlst( i )=001007 ! satellite identifier
    i=i+1;   ktdlst( i )=002019 ! satellite instruments
    i=i+1;   ktdlst( i )=001033 ! id of originating/generating center
    i=i+1;   ktdlst( i )=001034 ! id of originating/generating sub-center

    ! start obs time
    i=i+1;   ktdlst( i )=008021 ! time significance
    i=i+1;   ktdlst( i )=301011 ! will be unpacked as:
                       ! 004001 ! year
                       ! 004002 ! month
                       ! 004003 ! day
    i=i+1;   ktdlst( i )=301012 ! will be unpacked as:
                       ! 004004 ! hour
                       ! 004005 ! minute
    i=i+1;   ktdlst( i )=004007 ! second

    ! obs position DEM
    i=i+1;   ktdlst( i )=017005 ! ADM target pos. significance
    i=i+1;   ktdlst( i )=331004 ! will be unpacked as:
                       ! 207001 ! increase scale, ref. value and data width
                       ! 005001 ! latitude, high acc.
                       ! 006001 ! longitude, high acc.
                       ! 207000 ! undo the scale change 
    i=i+1;   ktdlst( i )=017010 ! ADM height bin altitude

    ! L1B header info
    i=i+1;   ktdlst( i )=017002 ! ADM data section type which follows

    ! file version
    i=i+1;   ktdlst( i )=017100 ! product name
!    i=i+1;   ktdlst( i )=017101 ! ???


    ! 5.2 final size 
    no_dd_stop = i

    if(bufr_test) then
      write(*, '(a)'   )   ' ------ set_LxB_bufr_header_dd --------'
      write(*, '(a,2I15)' ) ' ------ no_dd_start,no_dd_stop ', no_dd_start, no_dd_stop
      write(*, '(a)'   )   ' ------ ktdlst '
      write(*, '(i15)' ) ( ktdlst(i), i=no_dd_start,no_dd_stop )
    endif

    return
  end subroutine set_LxB_bufr_header_dd
  !  #]
  !---------------------------------
  ! 6.0 set_L1B_bufr_data_obs_dd
  subroutine set_L1B_bufr_data_obs_dd( no_dd_start, no_dd_stop )
    !  #[
    ! arguments
    integer, intent( in)    :: no_dd_start
    integer, intent(out)    :: no_dd_stop
   
    ! local variables
    integer                :: i

    i = no_dd_start-1

    ! 6.1 set descriptors

    ! ----------------------------------
    ! GEO
    ! ----------------------------------
    i=i+1;   ktdlst( i )=017002 ! ADM data section
    
    i=i+1;   ktdlst( i )=017006 ! Geoid separation
    ! mie
    i=i+1;   ktdlst( i )=017001 ! ADM reciever channel
    i=i+1;   ktdlst( i )=102000+(NPROFILE+1) ! replication factor
    i=i+1;   ktdlst( i )=017003 ! height bin index
    i=i+1;   ktdlst( i )=331001 ! will be unpacked as:
                       ! 331004 ! will be unpacked as:
                       !     207001 increase scale, ref. value and data width
                       !     005001 latitude, high accuracy
                       !     006001 longitude, high accuracy
                       !     207000 undio scale change
                       ! 017010 ! ADM height bin altitude
                       ! 005021 ! bearing or azimuth
                       ! 007021 ! elevation
                       ! 008065 ! Sun-glint indicator
    ! rayleigh
    i=i+1;   ktdlst( i )=017001
    i=i+1;   ktdlst( i )=102000+(NPROFILE+1)
    i=i+1;   ktdlst( i )=017003
    i=i+1;   ktdlst( i )=331001

    ! -----------------------------------
    ! PCD
    ! -----------------------------------
    i=i+1;   ktdlst( i )=017002

    i=i+1;   ktdlst( i )=102000+(NPROFILE+1)
    i=i+1;   ktdlst( i )=017003
    i=i+1;   ktdlst( i )=017020
    ! Mie 
    i=i+1;   ktdlst( i )=017001
    i=i+1;   ktdlst( i )=103000+(NPROFILE+1)
    i=i+1;   ktdlst( i )=017003
    i=i+1;   ktdlst( i )=008023
    i=i+1;   ktdlst( i )=017062
    ! Rayleigh 
    i=i+1;   ktdlst( i )=017001
    i=i+1;   ktdlst( i )=103000+(NPROFILE+1)
    i=i+1;   ktdlst( i )=017003
    i=i+1;   ktdlst( i )=008023
    i=i+1;   ktdlst( i )=017062

 
    ! -----------------------------------
    ! GWD
    ! -----------------------------------
    i=i+1;   ktdlst( i )=017002
    ! mie
    i=i+1;   ktdlst( i )=017001
    i=i+1;   ktdlst( i )=017030
    i=i+1;   ktdlst( i )=017031
    ! rayleigh
    i=i+1;   ktdlst( i )=017001
    i=i+1;   ktdlst( i )=017030
    i=i+1;   ktdlst( i )=017031

    ! -----------------------------------
    ! MEAS
    ! -----------------------------------
    i=i+1;   ktdlst( i )=017002
    ! mie
    i=i+1;   ktdlst( i )=017001
    i=i+1;   ktdlst( i )=017041
    i=i+1;   ktdlst( i )=102000+(NPROFILE)
    i=i+1;   ktdlst( i )=017003
    i=i+1;   ktdlst( i )=017040
    ! rayleigh
    i=i+1;   ktdlst( i )=017001
    i=i+1;   ktdlst( i )=017041
    i=i+1;   ktdlst( i )=102000+(NPROFILE)
    i=i+1;   ktdlst( i )=017003
    i=i+1;   ktdlst( i )=017040

    ! -----------------------------------
    ! US 
    ! only in measurement scale
    ! -----------------------------------

    ! -----------------------------------
    ! WV
    ! -----------------------------------
    i=i+1;   ktdlst( i )=017002
    ! mie
    i=i+1;   ktdlst( i )=017001
    i=i+1;   ktdlst( i )=102000+(NPROFILE)
    i=i+1;   ktdlst( i )=017003
    i=i+1;   ktdlst( i )=331002
    ! rayleigh
    i=i+1;   ktdlst( i )=017001
    i=i+1;   ktdlst( i )=102000+(NPROFILE)
    i=i+1;   ktdlst( i )=017003
    i=i+1;   ktdlst( i )=331002

    ! -----------------------------------
    !  CC calibration and characterization
    ! -----------------------------------
    i=i+1;   ktdlst( i )=017002

    ! --- L1B characterisation 
    i=i+1;   ktdlst( i )=017200

    ! mie
    i=i+1;   ktdlst( i )=017001

    i=i+1;   ktdlst( i )=017201
    i=i+1;   ktdlst( i )=017201
    i=i+1;   ktdlst( i )=017201

    ! rayleigh
    i=i+1;   ktdlst( i )=017001

    i=i+1;   ktdlst( i )=017201
    i=i+1;   ktdlst( i )=017201
    i=i+1;   ktdlst( i )=017201
    i=i+1;   ktdlst( i )=017201

    ! tripod corr.
    i=i+1;   ktdlst( i )=101000+(NTRIPOD_CORR)
    i=i+1;   ktdlst( i )=017202

    ! Mie response calibration 
    i=i+1;   ktdlst( i )=017001

    i=i+1;   ktdlst( i )=017203
    ! -----------------------
    i=i+1;   ktdlst( i )=106000+(NFREQUENCY_STEP)
    i=i+1;   ktdlst( i )=017205
    i=i+1;   ktdlst( i )=017204
    i=i+1;   ktdlst( i )=017206
    i=i+1;   ktdlst( i )=017207
    i=i+1;   ktdlst( i )=017208
    i=i+1;   ktdlst( i )=017209
    ! -----------------------
    i=i+1;   ktdlst( i )=017218
    i=i+1;   ktdlst( i )=017219
    i=i+1;   ktdlst( i )=017220
    i=i+1;   ktdlst( i )=017221
    i=i+1;   ktdlst( i )=017222
    i=i+1;   ktdlst( i )=017223
    i=i+1;   ktdlst( i )=017224
    i=i+1;   ktdlst( i )=017225

    ! Rayleigh response calibration 
    i=i+1;   ktdlst( i )=017001

    i=i+1;   ktdlst( i )=017203
    ! -----------------------
    i=i+1;   ktdlst( i )=106000+(NFREQUENCY_STEP)
    i=i+1;   ktdlst( i )=017205
    i=i+1;   ktdlst( i )=017204
    i=i+1;   ktdlst( i )=017206
    i=i+1;   ktdlst( i )=017207
    i=i+1;   ktdlst( i )=017208
    i=i+1;   ktdlst( i )=017209
    ! -----------------------
    i=i+1;   ktdlst( i )=017218
    i=i+1;   ktdlst( i )=017219
    i=i+1;   ktdlst( i )=017220
    i=i+1;   ktdlst( i )=017221
    i=i+1;   ktdlst( i )=017222
    i=i+1;   ktdlst( i )=017223
    i=i+1;   ktdlst( i )=017224
    i=i+1;   ktdlst( i )=017225


    !----------------------------------

    ! 6.2 final size 
    no_dd_stop = i

    if(bufr_test) then
      write(*, '(a)'   )   ' ------ set_L1B_bufr_data_obs_dd --------'
      write(*, '(a,2I15)' ) ' ------ no_dd_start,no_dd_stop ', no_dd_start, no_dd_stop
      write(*, '(a)'   )   ' ------ ktdlst '
      write(*, '(i15)' ) ( ktdlst(i), i=no_dd_start,no_dd_stop )
    endif

    return
  end subroutine set_L1B_bufr_data_obs_dd
  !  #]
  !---------------------------------
  ! 7.0 set_L1B_bufr_data_meas_dd
  subroutine set_L1B_bufr_data_meas_dd( no_dd_start, no_dd_stop )
    !  #[

    ! arguments
    integer, intent( in)    :: no_dd_start
    integer, intent(out)    :: no_dd_stop
   
    ! local variables
    integer                :: i

    i = no_dd_start-1

    ! 7.1 set descriptors

    ! delayed replication
    i=i+1;   ktdlst( i )=163000 ! ???
    i=i+1;   ktdlst( i )=031001 ! delayed descriptor replication factor


    ! -----------------------------------
    ! measurement granularity
    ! -----------------------------------
    i=i+1;   KTDLST( i )=017004 ! imeas ! ADM measurement index

    ! -----------------------------------
    ! GEO 
    ! -----------------------------------
    i=i+1;   KTDLST( i )=017002 ! BCODE_DATA_GEO ! adm data section type which follows

    i=i+1;   KTDLST( i )=017005 ! BCODE_POS_LOS_GEOID ! adm instr. target pos. significance
    i=i+1;   KTDLST( i )=017062 ! L1B%Geoloc_ADS%Meas_Geoloc(imeas)%AOCS_LOS_Velocity
    i=i+1;   KTDLST( i )=331004 ! combination of: 207001, 005001, 006001, 207000 (lat,lon high accuracy)
    i=i+1;   KTDLST( i )=017010 !  L1B%Geoloc_ADS%Meas_Geoloc(imeas)%alt_of_DEM_intersection (height bin alt)
    ! mie
    i=i+1;   KTDLST( i )=017001 ! adm receiver channel (Mie or Rayleigh)
    i=i+1;   KTDLST( i )=103000+(NPROFILE+1) ! ???? ! BCODE_CHANNEL_MIE
    i=i+1;   KTDLST( i )=017003 ! iprof-1 !adm height bin index
    i=i+1;   KTDLST( i )=331004 ! combination of: 207001, 005001, 006001, 207000 (lat,lon high accuracy)
    i=i+1;   KTDLST( i )=017010 ! altitude
    ! rayleigh
    i=i+1;   KTDLST( i )=017001 ! adm reciever channel (Mie or Rayleigh)
    i=i+1;   KTDLST( i )=103000+(NPROFILE+1) ! ???? ! BCODE_CHANNEL_MIE
    i=i+1;   KTDLST( i )=017003 ! iprof-1 !adm height bin index
    i=i+1;   KTDLST( i )=331004 ! combination of: 207001, 005001, 006001, 207000 (lat,lon high accuracy)
    i=i+1;   KTDLST( i )=017010 ! altitude

    ! -----------------------------------
    ! PCD 
    ! -----------------------------------
    i=i+1;   KTDLST( i )=017002 ! BCODE_DATA_PCD ! adm data section type which follows

    ! frequency sequence
    i=i+1;   KTDLST( i )=331005 ! combination of twice (017001, 017205, 008023, 017205) = 
                                ! (adm reciever channel, sat. channel freq. offset, first order stats, sat. freq. offset)
    ! scattering ratio
    i=i+1;   KTDLST( i )=102000+(NPROFILE+1)
    i=i+1;   KTDLST( i )=017003 ! adm height bin index
    i=i+1;   KTDLST( i )=017020 ! backscatter [dB]

    ! signal to noise ratio
    ! mie
    i=i+1;   KTDLST( i )=017001
    i=i+1;   KTDLST( i )=102000+(NPROFILE+1)
    i=i+1;   KTDLST( i )=017003
    i=i+1;   KTDLST( i )=017021
    ! rayleigh A
    i=i+1;   KTDLST( i )=017001
    i=i+1;   KTDLST( i )=102000+(NPROFILE+1)
    i=i+1;   KTDLST( i )=017003
    i=i+1;   KTDLST( i )=017021
    ! rayleigh B
    i=i+1;   KTDLST( i )=017001
    i=i+1;   KTDLST( i )=102000+(NPROFILE+1)
    i=i+1;   KTDLST( i )=017003
    i=i+1;   KTDLST( i )=017021

    ! -----------------------------------
    ! MEAS
    ! -----------------------------------
    i=i+1;   KTDLST( i )=017002

    ! spectrometer sequence contains dimension = NSPECTROMETER:
    i=i+1;   KTDLST( i )=331006

    ! measurement data: 25 x NSPECTROMETER
    i=i+1;   KTDLST( i )=017001

    i=i+1;   KTDLST( i )=103000+(NPROFILE+1)
    i=i+1;   KTDLST( i )=017003

    i=i+1;   KTDLST( i )=101000+(NSPECTROMETER)
    i=i+1;   KTDLST( i )=017227

    ! -----------------------------------
    ! US 
    ! -----------------------------------
    i=i+1;   KTDLST( i )=017002

    ! mie
    i=i+1;   KTDLST( i )=017001
    i=i+1;   KTDLST( i )=103000+(NPROFILE+1)
    i=i+1;   KTDLST( i )=017003
    i=i+1;   KTDLST( i )=017050
    i=i+1;   KTDLST( i )=017051
    ! rayleigh
    i=i+1;   KTDLST( i )=017001
    i=i+1;   KTDLST( i )=106000+(NPROFILE+1)
    i=i+1;   KTDLST( i )=017003
    i=i+1;   KTDLST( i )=017050
    ! - B
    i=i+1;   KTDLST( i )=017001
    i=i+1;   KTDLST( i )=017051
    ! - A
    i=i+1;   KTDLST( i )=017001
    i=i+1;   KTDLST( i )=017051

    ! -----------------------------------
    ! WV 
    ! -----------------------------------
    i=i+1;   KTDLST( i )=017002

    ! mie
    i=i+1;   KTDLST( i )=017001
    i=i+1;   KTDLST( i )=102000+(NPROFILE)
    i=i+1;   KTDLST( i )=017003
    i=i+1;   KTDLST( i )=331002
    ! rayleigh
    i=i+1;   KTDLST( i )=017001
    i=i+1;   KTDLST( i )=102000+(NPROFILE)
    i=i+1;   KTDLST( i )=017003
    i=i+1;   KTDLST( i )=331002


    !----------------------------------
    ! 7.2 final size 
    no_dd_stop = i

    if(bufr_test) then
      write(*, '(a)'   )   ' ------ set_L1B_bufr_data_meas_dd --------'
      write(*, '(a,2I15)' ) ' ------ no_dd_start,no_dd_stop ', no_dd_start, no_dd_stop
      write(*, '(a)'   )   ' ------ ktdlst '
      write(*, '(i15)' ) ( ktdlst(i), i=no_dd_start,no_dd_stop )
    endif

    return
  end subroutine set_L1B_bufr_data_meas_dd

  !  #]
  !---------------------------------
  ! 10.0 L1B_bufr_encode
  subroutine L1B_bufr_encode(L1B_Headers, L1B, isubset, &
                             workaround, FakeDateTime)
    !  #[
    ! arguments
    TYPE(L1B_HDR_DataType), intent(in)  :: L1B_Headers
    TYPE(L1B_BRC_DataType), intent(in)  :: L1B
    integer,                intent(in)  :: isubset
    logical,            optional, intent(in) :: workaround
    TYPE(DateTimeType), optional, intent(in) :: FakeDateTime

    ! 0.0 local variables

    ! 0.1 indices, sizes
    integer    :: no_element_offset, &
                  imeasurement, &
                  nmeasurement
!                  no_element_start, & ! not used
!                  no_element_stop, &  ! not used

    ! 0.2 date variables
    integer    :: year, &
                  month, & 
                  day, & 
                  hour, & 
                  minute!, &
!                  second ! not used

    ! 0.3 position variables
    real(r8_)  :: lat, &
                  lon


    !---------------------------------
    ! 1.0 build L1B observation array (single subset)

    ! 1.1 header elements
    call fill_L1B_bufr_header( L1B_Headers, L1B, isubset, &
             year,month,day,hour,minute, & 
             lat, lon, FakeDateTime) 

    ! 1.2 observation scale
    call fill_L1B_bufr_data_obs( L1B, isubset, workaround=workaround)

    ! 1.3 measurement scale
    no_element_offset = (isubset-1) * ktdexp_size & 
                      + NELEMENTS_L1B_HEAD &
                      + NELEMENTS_L1B_OBS

    ! 1.3.1 set delayed replication factor

    write(*,*) "TODO : fetch actual number measurements for each observation : currently missing in testdata  "
!    nmeasurement =  L1B%nr_of_measurements
    nmeasurement = kdata(1)

    values(no_element_offset+1) = nmeasurement

 
    ! 1.3.2 fetch measurements 
    if (kdata(1) >= nmeasurement ) then

      do imeasurement = 1, nmeasurement

        call fill_L1B_bufr_data_meas( L1B, imeasurement, isubset,&
                                      workaround=workaround)

      end do

    else
      write(*,*) "Error: L1B_bufr_encode : nmeasurement_max < nmeasurement "
      ! exit to leave the code is not Fortran90 !!!!
      ! The Fortran90 exit commandd only allows leaving do loops.
      !call exit(2)
      stop 2
    endif

    ! end collect measurements

    ! 2.0 keep key for sec2

    ! 2.1 construct position box
    if (key_lat_min > lat) key_lat_min = lat
    if (key_lat_max < lat) key_lat_max = lat
    if (key_lon_min > lat) key_lon_min = lat
    if (key_lon_max < lat) key_lon_max = lat

    ! 2.2 keep obs date
    key_year    = year
    key_month   = month
    key_day     = day 
    key_hour    = hour  
    key_minute  = minute

    ! why is key_second never updated? 
    ! even if second is available and defined in the EE file??? JK.

    ! end collect observation
    !---------------------------------

    return
  end subroutine L1B_bufr_encode
  !  #]
  !---------------------------------
  ! 11.0 L1B_bufr_encode_init
  subroutine L1B_bufr_encode_init( nmeasurement )
    !  #[

    integer,            intent( in) :: nmeasurement

    ! 0.0 local variables
    integer    :: no_dd_start, &
                  no_dd_stop!, &
!                  error_flag ! not used


    ! 1.0 initialise
    kdata(1)    = nmeasurement
    ktdlst(:)   = 0
    ktdlst_size = 0
    ktdexp_size = 0

    ! values arrays
    values(:)   = RVIND
    cvals(:)    = ''
    kbufr(:)    = NVIND

    ! key for sec2 
    key_lat_min =   90
    key_lat_max = - 90
    key_lon_min =  180
    key_lon_max = -180

    key_year    = 0
    key_month   = 0
    key_day     = 0 
    key_hour    = 0  
    key_minute  = 0
    key_second  = 0


    ! 2.1 section 0
    call set_L1B_bufr_sec0()

    ! 2.2 declare template
    no_dd_start = 1
    call set_LxB_bufr_header_dd( no_dd_start, no_dd_stop )

    no_dd_start = no_dd_stop+1
    call set_L1B_bufr_data_obs_dd( no_dd_start, no_dd_stop )

    no_dd_start = no_dd_stop+1
    call set_L1B_bufr_data_meas_dd( no_dd_start, no_dd_stop )

    ! 2.3 set dimensions
    ktdlst_size = no_dd_stop 

    ktdexp_size   = NELEMENTS_L1B_HEAD &
                  + NELEMENTS_L1B_OBS &
                  + 1 &
                  + kdata(1) * NELEMENTS_L1B_MEAS
    

    return
  end subroutine L1B_bufr_encode_init

  !  #]
  !---------------------------------
  ! 12.0 L1B_bufr_encode_finish
  subroutine L1B_bufr_encode_finish( nsubset, kbufr, kbufr_size )
    !  #[

    integer,            intent( in) :: nsubset
    integer,            intent(out) :: kbufr(jbufl)
    integer,            intent(out) :: kbufr_size


    ! 0.0 local variables
    integer     :: i, &
                   iprint, &
                   nelements_l1b, &
                   error_flag

    integer    :: ktdexp(kelem)


    ! 1.0 build sections
    call buprq(1,0,0)

    ! 1.1 section 1
    call set_L1B_bufr_sec1( key_year,key_month,key_day,key_hour,key_minute ) 
 
    ! 1.2 section 2
    call set_L1B_bufr_sec2( nsubset )

    ! 1.3 section 3
    call set_L1B_bufr_sec3( nsubset )


    ! 2.0 check template
! (bufr_test == .true.) is not allowed in Fortran90
! just use the following:
    if (bufr_test) then
      iprint = 1
      write(*, '(a)'     ) ' L1B EE2BUFR: PRINT template '
    endif
    call buxdes(iprint,ksec1,ktdlst_size,ktdlst,krdlen,kdata,kelem, &
             ktdexp_size,ktdexp,cnames,cunits,error_flag)
    if(error_flag.gt.0) then
      print*,' ERROR buxdes  ------- ', error_flag
      ! exit to leave the code is not Fortran90 !!!!
      ! The Fortran90 exit commandd only allows leaving do loops.
      stop 2
      !call exit(2)
    endif

    nelements_l1b = NELEMENTS_L1B_HEAD &
                  + NELEMENTS_L1B_OBS &
                  + 1 &
                  + kdata(1) * NELEMENTS_L1B_MEAS

    if(ktdexp_size /= nelements_l1b) then
      write(*, '(a)'     ) ' L1B EE2BUFR: ERROR unexpected template length '
      write(*, '(a,2i15)' ) ' ktdexp_size /= NELEMENTS_L1B ', ktdexp_size, nelements_l1b
      ! exit to leave the code is not Fortran90 !!!!
      ! The Fortran90 exit commandd only allows leaving do loops.
      stop 2
      !call exit(2)
    endif


    ! 3.0 encode message
    call bufren( &
             ksec0, ksec1, ksec2, ksec3, ksec4, &
             ktdlst_size, ktdlst, krdlen, kdata, &
             ktdexp_size, kvals, values, cvals, kbufr_size, kbufr, & 
             error_flag )
    if(error_flag.gt.0) then
      print*,' ERROR bufren  ------- ', error_flag
      ! exit to leave the code is not Fortran90 !!!!
      ! The Fortran90 exit commandd only allows leaving do loops.
      stop 2
      !call exit(2)
    endif

    if(bufr_test) then
      write(*, '(a)'     ) ' -------------- L1B_bufr_encode_finish -------'
      write(*, '(a,i15)' ) ' kbufr_size   ', kbufr_size
      write(*, '(a)'     ) ' ------ kbufr '
      write(*, '(i15)' ) (kbufr(i),i=1,kbufr_size)
    endif


    return
  end subroutine L1B_bufr_encode_finish

  !  #]
  !---------------------------------
  ! 20.0 fill_L1B_bufr_header
  subroutine fill_L1B_bufr_header( L1B_Headers, L1B, isubset, &
                year,month,day,hour,minute, &
                lat, lon, FakeDateTime)
    !  #[
    ! arguments
    TYPE(L1B_HDR_DataType),  intent(in)  :: L1B_Headers
    TYPE(L1B_BRC_DataType),  intent(in)  :: L1B
    integer,             intent(in)  :: isubset
    integer,             intent(out) :: year,month,day,hour,minute
    real(r8_),           intent(out) :: lat, lon
    TYPE(DateTimeType), optional, intent(in) :: FakeDateTime

    ! local variables
    Type(date_type)        :: date
    Type(time_type)        :: time

    integer                :: no_start, no_stop
    integer                :: i

    real(r8_)              :: tmp_lon

    ! initialise
    no_start = (isubset-1) * ktdexp_size + 1

    ! map elements
    i = no_start-1

    i=i+1;  values( i) = BCODE_SATID
    i=i+1;  values( i) = BCODE_INSTRUMENT
    i=i+1;  values( i) = BCODE_CENTRE
    i=i+1;  values( i) = BCODE_SUBCENTRE

    i=i+1;  values( i) = BCODE_TIME_SENSE_START

    if (present(FakeDateTime)) then
       date = Get_Date_from_DateTime( FakeDateTime )
       time = Get_Time_from_DateTime( FakeDateTime )
    else
       date = Get_Date_from_DateTime( L1B%Geoloc_ADS%start_of_obs_datetime )
       time = Get_Time_from_DateTime( L1B%Geoloc_ADS%start_of_obs_datetime )
    endif

    i=i+1;  values( i) = date%year
    i=i+1;  values( i) = date%month
    i=i+1;  values( i) = date%day
    i=i+1;  values( i) = time%hour
    i=i+1;  values( i) = time%minute
    i=i+1;  values( i) = time%second

    i=i+1;  values( i) = BCODE_POS_LOS_GEOID
    i=i+1;  values( i) = L1B%Geoloc_ADS%Obs_Geoloc%lat_of_DEM_intersection / DEGREE_SCALE

    tmp_lon =  L1B%Geoloc_ADS%Obs_Geoloc%lon_of_DEM_intersection / DEGREE_SCALE
    if ( tmp_lon > DEGREE_REF_LON ) tmp_lon = tmp_lon - 2*DEGREE_REF_LON
    i=i+1;  values( i) = tmp_lon

    i=i+1;  values( i) = L1B%Geoloc_ADS%Obs_Geoloc%alt_of_DEM_intersection

    i=i+1;  values( i) = BCODE_DATA_HEA

    ! product name
    i=i+1;  values( i) = isubset * 1000 + 64;  cvals( isubset ) = trim( L1B_Headers%MPH%product( 1:62) )

    ! return time / position
    no_stop = i

    i = (no_start-1)
    year    = int(values( i + 6 ))
    month   = int(values( i + 7 ))
    day     = int(values( i + 8 ))
    hour    = int(values( i + 9 ))
    minute  = int(values( i +10 ))

    lat     = values( i +13 )
    lon     = values( i +14 )

    ! test 
    if(bufr_test) then
        write(*, '(a)'    )     ' -------------- fill_L1B_bufr_header  -------'
        write(*, '(a,2i15)' )   ' values no_start / no_stop  ', no_start, no_stop
        write(*, '(a,5i15)' )   ' time      ', year,month,day,hour,minute
        write(*, '(a,2f15.6)' ) ' position  ', lat, lon
    endif

    return
  end subroutine fill_L1B_bufr_header
  !  #]
  !---------------------------------
  ! 21.0 fill_L1B_bufr_data_obs
  subroutine fill_L1B_bufr_data_obs( L1B, isubset, workaround)
    !  #[

    ! arguments
    TYPE(L1B_BRC_DataType),  intent(in)  :: L1B
    integer,             intent(in)  :: isubset
    logical, optional,   intent(in)  :: workaround


    ! local variables
!    Type(date_type)        :: date ! never used
!    Type(time_type)        :: time ! never used

    integer                :: no_start, no_stop
    integer                :: i, iprof, nprof, itrip, ntrip, ifreq, nfreq
!    character(len=32)      :: string ! never used

    integer                :: iwarn

    real(r8_)              :: tmp_lon

    ! initialise
    no_start = (isubset-1) * ktdexp_size + NELEMENTS_L1B_HEAD + 1

    ! map elements
    i = no_start-1

    ! GEO
    i=i+1;  values( i) = BCODE_DATA_GEO

    i=i+1;  values( i) = L1B%Geoloc_ADS%Obs_Geoloc%Geoid_Separation_of_Height_Bin

    i=i+1;  values( i) = BCODE_CHANNEL_MIE

    nprof = size( L1B%Geoloc_ADS%Obs_Geoloc%Mie_Rangebin_Geoloc ) 
    if ( nprof /= NPROFILE+1 ) then
      write(*,*) "Error: fill_L1B_bufr_data_obs : Mie unexpected No profiles ", nprof
      ! exit to leave the code is not Fortran90 !!!!
      ! The Fortran90 exit commandd only allows leaving do loops.
      stop 2
      !call exit(2)
    endif

    ! set top upper bin index=0
    do iprof=1,nprof
      i=i+1;  values( i) = iprof-1
      i=i+1;  values( i) = L1B%Geoloc_ADS%Obs_Geoloc%Mie_Rangebin_Geoloc(iprof)%lat_of_height_bin / DEGREE_SCALE 

      tmp_lon =  L1B%Geoloc_ADS%Obs_Geoloc%Mie_Rangebin_Geoloc(iprof)%lon_of_height_bin / DEGREE_SCALE
      if ( tmp_lon > DEGREE_REF_LON ) tmp_lon = tmp_lon - 2*DEGREE_REF_LON
      i=i+1;  values( i) = tmp_lon

      i=i+1;  values( i) = L1B%Geoloc_ADS%Obs_Geoloc%Mie_Rangebin_Geoloc(iprof)%alt_of_height_bin
      i=i+1;  values( i) = L1B%Geoloc_ADS%Obs_Geoloc%Mie_Rangebin_Geoloc(iprof)%topoc_azim_of_height_bin
      i=i+1;  values( i) = L1B%Geoloc_ADS%Obs_Geoloc%Mie_Rangebin_Geoloc(iprof)%topoc_elev_of_height_bin
      i=i+1;  values( i) = code_sun_visibility( L1B%Geoloc_ADS%Obs_Geoloc%Mie_Rangebin_Geoloc(iprof)%Target_to_Sun_Visib_Flag )
    end do

    i=i+1;  values( i) = BCODE_CHANNEL_RAY

    nprof = size( L1B%Geoloc_ADS%Obs_Geoloc%Rayleigh_Rangebin_Geoloc ) 
    if ( nprof /= NPROFILE+1 ) then
      write(*,*) "Error: fill_L1B_bufr_data_obs : Rayleigh unexpected No profiles ", nprof
      ! exit to leave the code is not Fortran90 !!!!
      ! The Fortran90 exit commandd only allows leaving do loops.
      stop 2
      !call exit(2)
    endif

    ! set top upper bin index=0
    do iprof=1,nprof
      i=i+1;  values( i) = iprof-1
      i=i+1;  values( i) = L1B%Geoloc_ADS%Obs_Geoloc%Rayleigh_Rangebin_Geoloc(iprof)%lat_of_height_bin / DEGREE_SCALE 

      tmp_lon =  L1B%Geoloc_ADS%Obs_Geoloc%Rayleigh_Rangebin_Geoloc(iprof)%lon_of_height_bin / DEGREE_SCALE
      if ( tmp_lon > DEGREE_REF_LON ) tmp_lon = tmp_lon - 2*DEGREE_REF_LON
      i=i+1;  values( i) = tmp_lon

      i=i+1;  values( i) = L1B%Geoloc_ADS%Obs_Geoloc%Rayleigh_Rangebin_Geoloc(iprof)%alt_of_height_bin
      i=i+1;  values( i) = L1B%Geoloc_ADS%Obs_Geoloc%Rayleigh_Rangebin_Geoloc(iprof)%topoc_azim_of_height_bin
      i=i+1;  values( i) = L1B%Geoloc_ADS%Obs_Geoloc%Rayleigh_Rangebin_Geoloc(iprof)%topoc_elev_of_height_bin
      i=i+1;  values( i) = code_sun_visibility( L1B%Geoloc_ADS%Obs_Geoloc%Rayleigh_Rangebin_Geoloc(iprof)%Target_to_Sun_Visib_Flag )
    end do

    ! PCD
    i=i+1;  values( i) = BCODE_DATA_PCD

    nprof = size( L1B%PCD_ADS%Observation_PCD%Obs_Alt_Bin_PCD ) 
    if ( nprof /= NPROFILE+1 ) then
      write(*,*) "Error: fill_L1B_bufr_data_obs : PCD unexpected No profiles ", nprof
      ! exit to leave the code is not Fortran90 !!!!
      ! The Fortran90 exit commandd only allows leaving do loops.
      stop 2
      !call exit(2)
    endif

    ! set top upper bin index=0
    iwarn = 0
    do iprof=1,nprof
      i=i+1;  values( i) = iprof-1

      if ( L1B%PCD_ADS%Observation_PCD%Obs_Alt_Bin_PCD(iprof)%scattering_ratio == 0.0 ) then
        iwarn=iwarn+1
        i=i+1
      else
!        i=i+1;  values( i) = 10.0 * log10( L1B%PCD_ADS%Observation_PCD%Obs_Alt_Bin_PCD(iprof)%scattering_ratio )
         ! replaced by
         call fill_one_element(i,L1B%PCD_ADS%Observation_PCD%&
                               Obs_Alt_Bin_PCD(iprof)%scattering_ratio,&
                               "scattering_ratio",dB_encode=.true.,&
                               workaround=workaround)
      endif

    end do
    if ( iwarn > 0 ) then
        write(*,*) "Warning: fill_L1B_bufr_data_obs : PCD ZERO scattering_ratio for ", iwarn, " profiles" 
    endif

    ! Mie hlos stdev
    i=i+1;  values( i) = BCODE_CHANNEL_MIE
    do iprof=1,nprof
      i=i+1;  values( i) = iprof-1

      i=i+1;  values( i) = BCODE_STATISTIC_STDEV
!      i=i+1;  values( i) = L1B%PCD_ADS%Observation_PCD%Obs_Alt_Bin_PCD(iprof)%Mie_wind_velocity_std_dev
      ! replaced by:
      call fill_one_element(i,L1B%PCD_ADS%Observation_PCD%&
                            Obs_Alt_Bin_PCD(iprof)%Mie_wind_velocity_std_dev,&
                            "Mie_wind_velocity_std_dev",workaround=workaround)
    end do

    ! Rayleigh hlos stdev
    i=i+1;  values( i) = BCODE_CHANNEL_RAY
    do iprof=1,nprof
      i=i+1;  values( i) = iprof-1

      i=i+1;  values( i) = BCODE_STATISTIC_STDEV
!      i=i+1;  values( i) = L1B%PCD_ADS%Observation_PCD%Obs_Alt_Bin_PCD(iprof)%Rayleigh_wind_velocity_std_dev
      ! replaced by:
      call fill_one_element(i,L1B%PCD_ADS%Observation_PCD%&
                     Obs_Alt_Bin_PCD(iprof)%Rayleigh_wind_velocity_std_dev,&
                     "Rayleigh_wind_velocity_std_dev",workaround=workaround)
    end do


    ! GWD
    i=i+1;  values( i) = BCODE_DATA_GWD

    i=i+1;  values( i) = BCODE_CHANNEL_MIE
    i=i+1;  values( i) = L1B%GWD_ADS%Mie_ground_corr_velocity
    i=i+1;  values( i) = L1B%GWD_ADS%Mie_grnd_corr_wght_factor

    i=i+1;  values( i) = BCODE_CHANNEL_RAY
    i=i+1;  values( i) = L1B%GWD_ADS%Rayleigh_ground_corr_velocity
    i=i+1;  values( i) = L1B%GWD_ADS%Rayleigh_grnd_corr_wght_factor

    ! MEAS
    i=i+1;  values( i) = BCODE_DATA_MEA

    ! - Mie
    i=i+1;  values( i) = BCODE_CHANNEL_MIE
    i=i+1;  values( i) = L1B%Meas_ADS%Mie_time_delays%background_integration_time

    nprof = size( L1B%Meas_ADS%Mie_time_delays%bin_layer_integration_time ) 
    if ( nprof /= NPROFILE ) then
      write(*,*) "Error: fill_L1B_bufr_data_obs : Mie Meas unexpected No profiles ", nprof
      ! exit to leave the code is not Fortran90 !!!!
      ! The Fortran90 exit commandd only allows leaving do loops.
      stop 2
!      call exit(2)
    endif

    do iprof=1,nprof
      i=i+1;  values( i) = iprof
      i=i+1;  values( i) = L1B%Meas_ADS%Mie_time_delays%bin_layer_integration_time(iprof)
    end do

    ! - Rayleigh
    i=i+1;  values( i) = BCODE_CHANNEL_RAY
    i=i+1;  values( i) = L1B%Meas_ADS%Rayleigh_time_delays%background_integration_time

    nprof = size( L1B%Meas_ADS%Rayleigh_time_delays%bin_layer_integration_time ) 
    if ( nprof /= NPROFILE ) then
      write(*,*) "Error: fill_L1B_bufr_data_obs : Rayleigh Meas unexpected No profiles ", nprof
      ! exit to leave the code is not Fortran90 !!!!
      ! The Fortran90 exit commandd only allows leaving do loops.
      stop 2
      !call exit(2)
    endif

    do iprof=1,nprof
      i=i+1;  values( i) = iprof
      i=i+1;  values( i) = L1B%Meas_ADS%Rayleigh_time_delays%bin_layer_integration_time(iprof)
    end do

    ! WV
    i=i+1;  values( i) = BCODE_DATA_WV

    ! - Mie
    i=i+1;  values( i) = BCODE_CHANNEL_MIE

    nprof = size( L1B%WV_MDS%Obs_Wind_Profile%Mie_alt_bin_wind_info ) 
    if ( nprof /= NPROFILE ) then
      write(*,*) "Error: fill_L1B_bufr_data_obs : Mie WV unexpected No profiles ", nprof
      ! exit to leave the code is not Fortran90 !!!!
      ! The Fortran90 exit commandd only allows leaving do loops.
      stop 2
      !call exit(2)
    endif

    do iprof=1,nprof
      i=i+1;  values( i) = iprof

      i=i+1;  values( i) = code_wind_bin_quality_flag( L1B%WV_MDS%Obs_Wind_Profile%Mie_alt_bin_wind_info(iprof)%bin_quality_flag )
      i=i+1;  values( i) = L1B%WV_MDS%Obs_Wind_Profile%Mie_alt_bin_wind_info(iprof)%wind_velocity
    end do

    ! - Rayleigh
    i=i+1;  values( i) = BCODE_CHANNEL_RAY

    nprof = size( L1B%WV_MDS%Obs_Wind_Profile%Rayleigh_alt_bin_wind_info ) 
    if ( nprof /= NPROFILE ) then
      write(*,*) "Error: fill_L1B_bufr_data_obs : Rayleigh WV unexpected No profiles ", nprof
      ! exit to leave the code is not Fortran90 !!!!
      ! The Fortran90 exit commandd only allows leaving do loops.
      stop 2
!      call exit(2)
    endif

    do iprof=1,nprof
      i=i+1;  values( i) = iprof

      i=i+1;  values( i) = code_wind_bin_quality_flag( L1B%WV_MDS%Obs_Wind_Profile%&
                                                           Rayleigh_alt_bin_wind_info(iprof)%bin_quality_flag )
      i=i+1;  values( i) = L1B%WV_MDS%Obs_Wind_Profile%Rayleigh_alt_bin_wind_info(iprof)%wind_velocity
    end do


    ! CC - calibration and characterization
    i=i+1;  values( i) = BCODE_DATA_CC

    i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Sat_Char_Data%Laser_Wavelength

    ! - Mie
    i=i+1;  values( i) = BCODE_CHANNEL_MIE

    i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Sat_Char_Data%Err_Quant%Mie_Err_Quant_K1
    i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Sat_Char_Data%Err_Quant%Mie_Err_Quant_K2
    i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Sat_Char_Data%Err_Quant%Mie_Err_Quant_K3

    ! - Rayleigh
    i=i+1;  values( i) = BCODE_CHANNEL_RAY

    i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Sat_Char_Data%Err_Quant%Rayl_Err_Quant_Ka2
    i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Sat_Char_Data%Err_Quant%Rayl_Err_Quant_Ka3
    i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Sat_Char_Data%Err_Quant%Rayl_Err_Quant_Kb2
    i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Sat_Char_Data%Err_Quant%Rayl_Err_Quant_Kb3

    ! Tripod
    ntrip = size( L1B%Cal_ADS%L1B_Char_Data%Sat_Char_Data%List_Tripod_Obscur_Corr ) 

    if ( ntrip /= NTRIPOD_CORR ) then
      write(*,*) "Error: fill_L1B_bufr_data_obs : Obs unexpected Dimension Tripod correction ", ntrip
      ! exit to leave the code is not Fortran90 !!!!
      ! The Fortran90 exit commandd only allows leaving do loops.
      stop 2
!      call exit(2)
    endif

    do itrip=1,ntrip
      i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Sat_Char_Data%List_Tripod_Obscur_Corr(itrip)%Tripod_Obscur_Corr
    enddo

    ! Response Calibration

    ! - Mie
    i=i+1;  values( i) = BCODE_CHANNEL_MIE

    i=i+1;  if( L1B%Cal_ADS%L1B_Char_Data%Mie_Resp_Calib_Data%Calib_Valid ) then
              values( i) = 0
            else
              values( i) = 1
            endif

    ! -- frequency step
    nfreq = size( L1B%Cal_ADS%L1B_Char_Data%Mie_Resp_Calib_Data%List_Mie_Freq_Step_Res ) 

    if ( nfreq > NFREQUENCY_STEP ) then
      write(*,*) "Error: fill_L1B_bufr_data_obs : Obs Mie unexpected Dimension Frequency step ", nfreq
      ! exit to leave the code is not Fortran90 !!!!
      ! The Fortran90 exit commandd only allows leaving do loops.
      stop 2
!      call exit(2)
    endif

    do ifreq=1,nfreq
      i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Mie_Resp_Calib_Data%List_Mie_Freq_Step_Res(ifreq)%Mie_Freq_Step_Res%Freq_Offset
      i=i+1;  if( L1B%Cal_ADS%L1B_Char_Data%Mie_Resp_Calib_Data%List_Mie_Freq_Step_Res(ifreq)%Mie_Freq_Step_Res%Freq_Valid ) then
                values( i)=0
              else
                values( i)=1
              endif
      i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Mie_Resp_Calib_Data%&
                               List_Mie_Freq_Step_Res(ifreq)%Mie_Freq_Step_Res%Meas_Resp
      i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Mie_Resp_Calib_Data%&
                               List_Mie_Freq_Step_Res(ifreq)%Mie_Freq_Step_Res%Meas_Err_Mie_Resp
      i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Mie_Resp_Calib_Data%&
                               List_Mie_Freq_Step_Res(ifreq)%Mie_Freq_Step_Res%Ref_Pulse_Resp
      i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Mie_Resp_Calib_Data%&
                               List_Mie_Freq_Step_Res(ifreq)%Mie_Freq_Step_Res%Ref_Pulse_Err_Mie_Resp

    enddo

    ! -- Mie frequency step padding
    do ifreq=1+nfreq,NFREQUENCY_STEP
      i=i+1;  values( i) = RVIND
      i=i+1;  values( i) = NVIND
      i=i+1;  values( i) = RVIND
      i=i+1;  values( i) = RVIND
      i=i+1;  values( i) = RVIND
      i=i+1;  values( i) = RVIND
    enddo

    ! -- response calib
    i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Mie_Resp_Calib_Data%Mie_Meas_Resp_Calib%Meas_Mean_Sensitivity
    i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Mie_Resp_Calib_Data%Mie_Meas_Resp_Calib%Meas_Zero_Freq
    i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Mie_Resp_Calib_Data%Mie_Meas_Resp_Calib%Meas_Err_Mie_Resp_Stdev

    i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Mie_Resp_Calib_Data%Mie_Ref_Pulse_Resp_Calib%Ref_Pulse_Mean_Sensitivity
    i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Mie_Resp_Calib_Data%Mie_Ref_Pulse_Resp_Calib%Ref_Pulse_Zero_Freq
    i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Mie_Resp_Calib_Data%Mie_Ref_Pulse_Resp_Calib%Ref_Pulse_Err_Mie_Resp_Stdev

    i=i+1;  values( i) = code_meas_calibration_flag( L1B%Cal_ADS%L1B_Char_Data%Mie_Resp_Calib_Data%Meas_Calib_Valid )
    i=i+1;  values( i) = code_refp_calibration_flag( L1B%Cal_ADS%L1B_Char_Data%Mie_Resp_Calib_Data%Ref_Pulse_Calib_Valid )

    ! - Rayleigh
    i=i+1;  values( i) = BCODE_CHANNEL_RAY

    i=i+1;  if( L1B%Cal_ADS%L1B_Char_Data%Rayl_Resp_Calib_Data%Calib_Valid ) then
              values( i) = 0
            else
              values( i) = 1
            endif

    ! -- frequency step
    nfreq = size( L1B%Cal_ADS%L1B_Char_Data%Rayl_Resp_Calib_Data%List_Rayl_Freq_Step_Res ) 

    if ( nfreq > NFREQUENCY_STEP ) then
      write(*,*) "Error: fill_L1B_bufr_data_obs : Obs Rayleigh unexpected Dimension Frequency step ", nfreq
      ! exit to leave the code is not Fortran90 !!!!
      ! The Fortran90 exit commandd only allows leaving do loops.
      stop 2
!      call exit(2)
    endif

    do ifreq=1,nfreq
      i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Rayl_Resp_Calib_Data%&
                               List_Rayl_Freq_Step_Res(ifreq)%Rayl_Freq_Step_Res%Freq_Offset
      i=i+1;  if( L1B%Cal_ADS%L1B_Char_Data%Rayl_Resp_Calib_Data%List_Rayl_Freq_Step_Res(ifreq)%Rayl_Freq_Step_Res%Freq_Valid ) then
                values( i)=0
              else
                values( i)=1
              endif
      i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Rayl_Resp_Calib_Data%&
                               List_Rayl_Freq_Step_Res(ifreq)%Rayl_Freq_Step_Res%Meas_Resp
      i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Rayl_Resp_Calib_Data%&
                               List_Rayl_Freq_Step_Res(ifreq)%Rayl_Freq_Step_Res%Meas_Err_Rayl_Resp
      i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Rayl_Resp_Calib_Data%&
                               List_Rayl_Freq_Step_Res(ifreq)%Rayl_Freq_Step_Res%Ref_Pulse_Resp
      i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Rayl_Resp_Calib_Data%&
                               List_Rayl_Freq_Step_Res(ifreq)%Rayl_Freq_Step_Res%Ref_Pulse_Err_Rayl_Resp

    enddo

    ! -- Rayleigh frequency step padding
    do ifreq=1+nfreq,NFREQUENCY_STEP
      i=i+1;  values( i) = RVIND
      i=i+1;  values( i) = NVIND
      i=i+1;  values( i) = RVIND
      i=i+1;  values( i) = RVIND
      i=i+1;  values( i) = RVIND
      i=i+1;  values( i) = RVIND
    enddo

    ! -- response calib
    i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Rayl_Resp_Calib_Data%Rayl_Meas_Resp_Calib%Meas_Mean_Sensitivity
    i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Rayl_Resp_Calib_Data%Rayl_Meas_Resp_Calib%Meas_Zero_Freq
    i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Rayl_Resp_Calib_Data%Rayl_Meas_Resp_Calib%Meas_Err_Rayl_Resp_Stdev

    i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Rayl_Resp_Calib_Data%Rayl_Ref_Pulse_Resp_Calib%Ref_Pulse_Mean_Sensitivity
    i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Rayl_Resp_Calib_Data%Rayl_Ref_Pulse_Resp_Calib%Ref_Pulse_Zero_Freq
    i=i+1;  values( i) = L1B%Cal_ADS%L1B_Char_Data%Rayl_Resp_Calib_Data%Rayl_Ref_Pulse_Resp_Calib%Ref_Pulse_Err_Rayl_Resp_Stdev

    i=i+1;  values( i) = code_meas_calibration_flag( L1B%Cal_ADS%L1B_Char_Data%Rayl_Resp_Calib_Data%Meas_Calib_Valid )
    i=i+1;  values( i) = code_refp_calibration_flag( L1B%Cal_ADS%L1B_Char_Data%Rayl_Resp_Calib_Data%Ref_Pulse_Calib_Valid )




    ! return position
    no_stop = i

    ! test 
    if(bufr_test) then
        write(*, '(a)'    )    ' -------------- fill_L1B_bufr_data_obs  -------'
        write(*, '(a,2i15)' ) ' values no_start / no_stop  ', no_start, no_stop
    endif

    return
  end subroutine fill_L1B_bufr_data_obs

  !  #]
  !---------------------------------
  ! 22.0 fill_L1B_bufr_data_meas
  subroutine fill_L1B_bufr_data_meas( L1B, imeas, isubset, workaround)
    !  #[
    ! arguments
    TYPE(L1B_BRC_DataType),  intent(in)  :: L1B
    integer,             intent(in)  :: imeas
    integer,             intent(in)  :: isubset
    logical, optional,   intent(in)  :: workaround

    ! local variables
!    Type(date_type)        :: date ! never used
!    Type(time_type)        :: time ! never used

    integer                :: no_start, no_stop
    integer                :: i, iprof, nprof, ispec, nspec!, k ! not used

    integer                :: iwarn

    real(r8_)              :: tmp_lon

    ! initialise
    no_start = (isubset-1) * ktdexp_size &
             + NELEMENTS_L1B_HEAD &
             + NELEMENTS_L1B_OBS &
             + 1 &
             + (imeas-1) * NELEMENTS_L1B_MEAS &
             + 1

    ! map elements
    i = no_start-1

    ! measurement index
    i=i+1;  values( i) = imeas

    ! GEO
    i=i+1;  values( i) = BCODE_DATA_GEO

    i=i+1;  values( i) = BCODE_POS_LOS_GEOID

!    i=i+1;  values( i) = L1B%Geoloc_ADS%Meas_Geoloc(imeas)%AOCS_LOS_Velocity
    ! replaced by:
    call fill_one_element(i,L1B%Geoloc_ADS%Meas_Geoloc(imeas)%&
                       AOCS_LOS_Velocity,"AOCS_LOS_Velocity",&
                       workaround=workaround)

    i=i+1;  values( i) = L1B%Geoloc_ADS%Meas_Geoloc(imeas)%lat_of_DEM_intersection / DEGREE_SCALE

    tmp_lon = L1B%Geoloc_ADS%Meas_Geoloc(imeas)%lon_of_DEM_intersection / DEGREE_SCALE
    if ( tmp_lon > DEGREE_REF_LON ) tmp_lon = tmp_lon - 2*DEGREE_REF_LON
    i=i+1;  values( i) = tmp_lon

    i=i+1;  values( i) = L1B%Geoloc_ADS%Meas_Geoloc(imeas)%alt_of_DEM_intersection

    i=i+1;  values( i) = BCODE_CHANNEL_MIE

    nprof = size( L1B%Geoloc_ADS%Meas_Geoloc(imeas)%Mie_Rangebin_Geoloc ) 
    if ( nprof /= NPROFILE+1 ) then
      write(*,*) "Error: fill_L1B_bufr_data_meas : Mie unexpected No profiles ", nprof
      ! exit to leave the code is not Fortran90 !!!!
      ! The Fortran90 exit commandd only allows leaving do loops.
      stop 2
!      call exit(2)
    endif

    ! set top upper bin index=0
    do iprof=1,nprof
      i=i+1;  values( i) = iprof-1
      i=i+1;  values( i) = L1B%Geoloc_ADS%Meas_Geoloc(imeas)%Mie_Rangebin_Geoloc(iprof)%lat_of_height_bin / DEGREE_SCALE 

      tmp_lon = L1B%Geoloc_ADS%Meas_Geoloc(imeas)%Mie_Rangebin_Geoloc(iprof)%lon_of_height_bin / DEGREE_SCALE
      if ( tmp_lon > DEGREE_REF_LON ) tmp_lon = tmp_lon - 2*DEGREE_REF_LON
      i=i+1;  values( i) = tmp_lon

      i=i+1;  values( i) = L1B%Geoloc_ADS%Meas_Geoloc(imeas)%Mie_Rangebin_Geoloc(iprof)%alt_of_height_bin
    end do

    i=i+1;  values( i) = BCODE_CHANNEL_RAY

    nprof = size( L1B%Geoloc_ADS%Meas_Geoloc(imeas)%Rayleigh_Rangebin_Geoloc ) 
    if ( nprof /= NPROFILE+1 ) then
      write(*,*) "Error: fill_L1B_bufr_data_meas : Rayleigh unexpected No profiles ", nprof
      ! exit to leave the code is not Fortran90 !!!!
      ! The Fortran90 exit commandd only allows leaving do loops.
      stop 2
!      call exit(2)
    endif

    ! set top upper bin index=0
    do iprof=1,nprof
      i=i+1;  values( i) = iprof-1
      i=i+1;  values( i) = L1B%Geoloc_ADS%Meas_Geoloc(imeas)%Rayleigh_Rangebin_Geoloc(iprof)%lat_of_height_bin / DEGREE_SCALE 

      tmp_lon = L1B%Geoloc_ADS%Meas_Geoloc(imeas)%Rayleigh_Rangebin_Geoloc(iprof)%lon_of_height_bin / DEGREE_SCALE
      if ( tmp_lon > DEGREE_REF_LON ) tmp_lon = tmp_lon - 2*DEGREE_REF_LON
      i=i+1;  values( i) = tmp_lon

      i=i+1;  values( i) = L1B%Geoloc_ADS%Meas_Geoloc(imeas)%Rayleigh_Rangebin_Geoloc(iprof)%alt_of_height_bin
    end do

    ! PCD
    i=i+1;  values( i) = BCODE_DATA_PCD

    i=i+1;  values( i) = BCODE_CHANNEL_MIE
    i=i+1;  values( i) = L1B%PCD_ADS%Measurement_PCD(imeas)%mie_mean_emitted_freq
    i=i+1;  values( i) = BCODE_STATISTIC_STDEV
       
    !parameter no longer present in 1B input
    ! i=i+1;  values( i) = L1B%PCD_ADS%Measurement_PCD(imeas)%mie_emitted_freq_std_dev

!????? this one seems not present in the descriptor list definition ????
    i=i+1;  values( i) = RVIND

    ! TODO November 2006
    ! i=i+1;  values( i) = L1B%PCD_ADS%Measurement_PCD(imeas)%reference_pulse_FWHM

    i=i+1;  values( i) = BCODE_CHANNEL_RAY
    i=i+1;  values( i) = L1B%PCD_ADS%Measurement_PCD(imeas)%rayleigh_mean_emitted_freq
    i=i+1;  values( i) = BCODE_STATISTIC_STDEV
    i=i+1;  values( i) = L1B%PCD_ADS%Measurement_PCD(imeas)%rayleigh_emitted_freq_std_dev

    nprof = size( L1B%PCD_ADS%Measurement_PCD(imeas)%Meas_Alt_Bin_PCD ) 
    if ( nprof /= NPROFILE+1 ) then
      write(*,*) "Error: fill_L1B_bufr_data_meas : PCD unexpected No profiles ", nprof
      ! exit to leave the code is not Fortran90 !!!!
      ! The Fortran90 exit commandd only allows leaving do loops.
      stop 2
!      call exit(2)
    endif

    ! set top upper bin index=0
    iwarn = 0
    do iprof=1,nprof
      i=i+1;  values( i) = iprof-1
      if ( L1B%PCD_ADS%Measurement_PCD(imeas)%Meas_Alt_Bin_PCD(iprof)%scattering_ratio == 0.0 ) then
        iwarn=iwarn+1
        i=i+1
      else

!        i=i+1;  values( i) = 10.0 * log10( L1B%PCD_ADS%Measurement_PCD(imeas)%Meas_Alt_Bin_PCD(iprof)%scattering_ratio )
!replaced by:
         call fill_one_element(i,L1B%PCD_ADS%Measurement_PCD(imeas)%&
                                     Meas_Alt_Bin_PCD(iprof)%scattering_ratio,&
                               "scattering_ratio",dB_encode=.true.,&
                               workaround=workaround)

      endif
    end do
    if ( iwarn > 0 ) then
        write(*,*) "Warning: fill_L1B_bufr_data_meas : PCD ZERO scattering_ratio for ", iwarn, " profiles" 
    endif

    i=i+1;  values( i) = BCODE_CHANNEL_MIE

    iwarn = 0
    do iprof=1,nprof
      i=i+1;  values( i) = iprof-1
!      if ( L1B%PCD_ADS%Measurement_PCD(imeas)%Meas_Alt_Bin_PCD(iprof)%Mie_snr == 0.0 ) then
!        iwarn=iwarn+1
!        i=i+1
!      else
!        i=i+1;  values( i) = 10.0 * log10( L1B%PCD_ADS%Measurement_PCD(imeas)%Meas_Alt_Bin_PCD(iprof)%Mie_snr )
         ! replaced by
         call fill_one_element(i,L1B%PCD_ADS%Measurement_PCD(imeas)%&
                               Meas_Alt_Bin_PCD(iprof)%Mie_snr, &
                               "Mie_snr",dB_encode=.true.,&
                               workaround=workaround)
!      endif
    end do
    if ( iwarn > 0 ) then
        write(*,*) "Warning: fill_L1B_bufr_data_obs : PCD ZERO Mie_snr for ", iwarn, " profiles" 
    endif

    i=i+1;  values( i) = BCODE_CHANNEL_RAY_A

    iwarn = 0
    do iprof=1,nprof
      i=i+1;  values( i) = iprof-1
!      if ( L1B%PCD_ADS%Measurement_PCD(imeas)%Meas_Alt_Bin_PCD(iprof)%Rayleigh_snr_channel_A == 0.0 ) then
!        iwarn=iwarn+1
!        i=i+1
!      else
!        i=i+1;  values( i) = 10.0 * log10( L1B%PCD_ADS%Measurement_PCD(imeas)%Meas_Alt_Bin_PCD(iprof)%Rayleigh_snr_channel_A )
         ! replaced by
         call fill_one_element(i,L1B%PCD_ADS%Measurement_PCD(imeas)%&
                               Meas_Alt_Bin_PCD(iprof)%Rayleigh_snr_channel_A,&
                               "Rayleigh_snr_channel_A",dB_encode=.true.,&
                               workaround=workaround)
!      endif
    end do
    if ( iwarn > 0 ) then
        write(*,*) "Warning: fill_L1B_bufr_data_obs : PCD ZERO Rayleigh_snr_channel_A for ", iwarn, " profiles" 
    endif

    i=i+1;  values( i) = BCODE_CHANNEL_RAY_B

    iwarn = 0
    do iprof=1,nprof
      i=i+1;  values( i) = iprof-1
!      if ( L1B%PCD_ADS%Measurement_PCD(imeas)%Meas_Alt_Bin_PCD(iprof)%Rayleigh_snr_channel_B == 0.0 ) then
!        iwarn=iwarn+1
!        i=i+1
!      else
!        i=i+1;  values( i) = 10.0 * log10( L1B%PCD_ADS%Measurement_PCD(imeas)%Meas_Alt_Bin_PCD(iprof)%Rayleigh_snr_channel_B )
         ! replaced by
         call fill_one_element(i,L1B%PCD_ADS%Measurement_PCD(imeas)%&
                               Meas_Alt_Bin_PCD(iprof)%Rayleigh_snr_channel_B,&
                               "Rayleigh_snr_channel_B",dB_encode=.true.,&
                               workaround=workaround)
!      endif
    end do
    if ( iwarn > 0 ) then
        write(*,*) "Warning: fill_L1B_bufr_data_obs : PCD ZERO Rayleigh_snr_channel_B for ", iwarn, " profiles" 
    endif

    ! MEAS
    i=i+1;  values( i) = BCODE_DATA_MEA

    ! - Mie
    i=i+1;  values( i) = BCODE_CHANNEL_MIE

    nspec = size( L1B%Meas_ADS%Mie_reference_pulses(imeas)%spectrometer_count ) 

    if ( nspec /= NSPECTROMETER ) then
      write(*,*) "Error: fill_L1B_bufr_data_meas : Mie Meas unexpected Dimension spectrometer ", nspec
      ! exit to leave the code is not Fortran90 !!!!
      ! The Fortran90 exit commandd only allows leaving do loops.
      stop 2
!      call exit(2)
    endif

    do ispec=1,nspec
      i=i+1;  values( i) = L1B%Meas_ADS%Mie_reference_pulses(imeas)%spectrometer_count(ispec)
    end do

    ! - Rayleigh A
    i=i+1;  values( i) = BCODE_CHANNEL_RAY_A
    i=i+1;  values( i) = L1B%Meas_ADS%Rayleigh_reference_pulses_A(imeas)%spectrometer_count

    ! - Rayleigh B
    i=i+1;  values( i) = BCODE_CHANNEL_RAY_B
    i=i+1;  values( i) = L1B%Meas_ADS%Rayleigh_reference_pulses_B(imeas)%spectrometer_count

    ! - Mie
    i=i+1;  values( i) = BCODE_CHANNEL_MIE

    nprof = size( L1B%Meas_ADS%Mie_measurement_data(imeas)%Mie_spectra_alt_bin ) 
    if ( nprof /= NPROFILE+1 ) then
      write(*,*) "Error: fill_L1B_bufr_data_meas : Mie_spectra_alt_bin unexpected No profiles ", nprof
      ! exit to leave the code is not Fortran90 !!!!
      ! The Fortran90 exit commandd only allows leaving do loops.
      STOP 2
!      call exit(2)
    endif

    nspec = size( L1B%Meas_ADS%Mie_measurement_data(imeas)%Mie_spectra_alt_bin(1)%spectrometer_count ) 

    if ( nspec /= NSPECTROMETER ) then
      write(*,*) "Error: fill_L1B_bufr_data_meas : Mie_spectra_alt_bin unexpected Dimension spectrometer ", nspec
      ! exit to leave the code is not Fortran90 !!!!
      ! The Fortran90 exit commandd only allows leaving do loops.
      STOP 2
!      call exit(2)
    endif

    do iprof=1,nprof
      i=i+1;  values( i) = iprof-1

      do ispec=1,nspec
        i=i+1;  values( i) = L1B%Meas_ADS%Mie_measurement_data(imeas)%Mie_spectra_alt_bin(iprof)%spectrometer_count(ispec)
      end do

    end do

    ! US
    i=i+1;  values( i) = BCODE_DATA_US

    ! - Mie
    i=i+1;  values( i) = BCODE_CHANNEL_MIE

    nprof = size( L1B%US_MDS%Meas_US(imeas)%Mie ) 
    if ( nprof /= NPROFILE+1 ) then
      write(*,*) "Error: fill_L1B_bufr_data_obs : Mie US unexpected No profiles ", nprof
      ! exit to leave the code is not Fortran90 !!!!
      ! The Fortran90 exit commandd only allows leaving do loops.
      stop 2
!      call exit(2)
    endif

    ! set top upper bin index=0
    do iprof=1,nprof
      i=i+1;  values( i) = iprof-1

      i=i+1;  values( i) = code_data_quality_flag( L1B%US_MDS%Meas_US(imeas)%Mie(iprof)%Data_Quality_Flag )
      i=i+1;  values( i) = L1B%US_MDS%Meas_US(imeas)%Mie(iprof)%Useful_Signal
    end do

    ! - Rayleigh
    i=i+1;  values( i) = BCODE_CHANNEL_RAY

    nprof = size( L1B%US_MDS%Meas_US(imeas)%Rayleigh ) 
    if ( nprof /= NPROFILE+1 ) then
      write(*,*) "Error: fill_L1B_bufr_data_obs : Rayleigh US unexpected No profiles ", nprof
      ! exit to leave the code is not Fortran90 !!!!
      ! The Fortran90 exit commandd only allows leaving do loops.
      stop 2
!      call exit(2)
    endif

    ! set top upper bin index=0
    do iprof=1,nprof
      i=i+1;  values( i) = iprof-1

      i=i+1;  values( i) = code_data_quality_flag( L1B%US_MDS%Meas_US(imeas)%Rayleigh(iprof)%Data_Quality_Flag )

      i=i+1;  values( i) = BCODE_CHANNEL_RAY_A
      i=i+1;  values( i) = L1B%US_MDS%Meas_US(imeas)%Rayleigh(iprof)%Useful_Signal_Channel_A
      i=i+1;  values( i) = BCODE_CHANNEL_RAY_B
      i=i+1;  values( i) = L1B%US_MDS%Meas_US(imeas)%Rayleigh(iprof)%Useful_Signal_Channel_B
    end do


    ! WV
    i=i+1;  values( i) = BCODE_DATA_WV

    ! - Mie
    i=i+1;  values( i) = BCODE_CHANNEL_MIE

    nprof = size( L1B%WV_MDS%Meas_Wind_Profile(imeas)%Mie_alt_bin_wind_info )
    if ( nprof /= NPROFILE ) then
      write(*,*) "Error: fill_L1B_bufr_data_obs : Mie WV unexpected No profiles ", nprof
      ! exit to leave the code is not Fortran90 !!!!
      ! The Fortran90 exit commandd only allows leaving do loops.
      stop 2
!      call exit(2)
    endif

    do iprof=1,nprof
      i=i+1;  values( i) = iprof

      i=i+1;  values( i) = code_wind_bin_quality_flag( L1B%WV_MDS%Obs_Wind_Profile%Mie_alt_bin_wind_info(iprof)%bin_quality_flag )

      i=i+1;  values( i) = L1B%WV_MDS%Meas_Wind_Profile(imeas)%Mie_alt_bin_wind_info(iprof)%wind_velocity
    end do

    ! - Rayleigh
    i=i+1;  values( i) = BCODE_CHANNEL_RAY

    nprof = size( L1B%WV_MDS%Meas_Wind_Profile(imeas)%Rayleigh_alt_bin_wind_info )
    if ( nprof /= NPROFILE ) then
      write(*,*) "Error: fill_L1B_bufr_data_obs : Rayleigh WV unexpected No profiles ", nprof
      ! exit to leave the code is not Fortran90 !!!!
      ! The Fortran90 exit commandd only allows leaving do loops.
      stop 2
!      call exit(2)
    endif

    do iprof=1,nprof
      i=i+1;  values( i) = iprof

      i=i+1;  values( i) = code_wind_bin_quality_flag( L1B%WV_MDS%Obs_Wind_Profile%Mie_alt_bin_wind_info(iprof)%bin_quality_flag )

      i=i+1;  values( i) = L1B%WV_MDS%Meas_Wind_Profile(imeas)%Rayleigh_alt_bin_wind_info(iprof)%wind_velocity
    end do


    ! return position
    no_stop = i

    ! test 
    if(bufr_test) then
        write(*, '(a)'    )    ' -------------- fill_L1B_bufr_data_meas  -------'
        write(*, '(a,2i15)' ) ' values no_start / no_stop  ', no_start, no_stop
    endif

    return
  end subroutine fill_L1B_bufr_data_meas
  !  #]
  !---------------------------------
  subroutine fill_one_element(index, value, description, dB_encode, &
                              int_encode, workaround)
    !  #[
    ! this routine does some extra checks when filling the values()
    ! array of the bufr message, and enables to trap some more problems
    ! and report them in a more clear way (compared to the bare BUFR library)

    integer,           intent(inout) :: index
    real(r8_),         intent(in)    :: value
    character(len=*),  intent(in)    :: description
    logical, optional, intent(in)    :: dB_encode
    logical, optional, intent(in)    :: int_encode
    logical, optional, intent(in)    :: workaround

    ! the index is the full index, i.e. 
    !    index = (isubset-1) * ktdexp_size + sub_index
    ! with sub_index the index within the current subset

    ! note that the index points to the last filled item, so we
    ! increment it by one, before filling the value()

    ! local variables
    integer  :: error_flag
    ! (later we might want to return this as a result of this routine)

    real(r8_) :: temp_value
    logical   :: use_dB_encode
    logical   :: use_int_encode
    logical   :: use_workaround
    character(len=50) :: str_value

    ! init
    error_flag = 0
    use_dB_encode = .false.
    if (present(dB_encode)) use_dB_encode = dB_encode
    use_int_encode = .false.
    if (present(int_encode)) use_int_encode = int_encode
    use_workaround = .false.
    if (present(workaround)) use_workaround = workaround

    ! increment the index and check its sanity
    index = index+1
    if (index .gt. size(values)) then
       error_flag = 1
       print *,"ERROR in fill_one_element(): values array seems too small"
       print *,"current index value is: ",index
       print *,"while max. allowed value is: size(values) = ",size(values)
       print *,"current element is: ",trim(description)
    end if

    temp_value = value

    ! do some sanity checks on value
    if (is_NaN(value)) then
       if (use_workaround) then
          call logmsg(log_warn,"WARNING in fill_one_element(): value seems "//&
               "to be NaN! Using missing indicator instead ! "//&
               "current element is: "//trim(description))
       else
          error_flag = 1
          call logmsg(log_error,"ERROR in fill_one_element(): value seems "//&
               "to be NaN! Using missing indicator instead ! "//&
               "current element is: "//trim(description))
       end if
       temp_value = RVIND
    end if

    if (is_PosInf(value) .or. is_NegInf(value)) then
       if (use_workaround) then
          call logmsg(log_warn,"WARNING in fill_one_element(): value seems "//&
               "to be Infinite! Using missing indicator instead ! "//&
               "current element is: "//trim(description))
       else
          error_flag = 1
          call logmsg(log_error,"ERROR in fill_one_element(): value seems "//&
               "to be Infinite! Using missing indicator instead ! "//&
               "current element is: "//trim(description))
       end if
       temp_value = RVIND
    end if

    if (use_dB_encode) then
       ! convert this value to dB
       !temp_value = real2dB(temp_value)
       ! an 8 byte version is not yet available in numerics, so for now
       ! use the following lines:
       
       if (missing_real(temp_value)) then
          ! do nothing. a dB value is not defined in this case
          ! so the number should remain the missing_indicator
          ! note that this is a valid result, so no need to report any errors
       else
          if (temp_value .le. 0._r8_) then
             str_value = real2txt_extra_checks(temp_value)
             if (use_workaround) then
                call logmsg(log_warn,"WARNING in fill_one_element(): "//&
                     "converting negative or zero numbers to dB values "//&
                     "is not possible ! Value seems to be: "//&
                     trim(str_value)//" possibly this value is already "//&
                     "in dB? using missing indicator instead ! "//&
                     "current element is: "//trim(description),&
                     realacc_precision=12)
             else
                error_flag = 1
                call logmsg(log_error,"ERROR in fill_one_element(): "//&
                     "converting negative or zero numbers to dB values "//&
                     "is not possible ! Value seems to be: "//&
                     trim(str_value)//" possibly this value is already "//&
                     "in dB? using missing indicator instead ! "//&
                     "current element is: "//trim(description),&
                     realacc_precision=12)
             end if
             temp_value = RVIND
          else
             ! now finally convert this value to dB
             temp_value = 10._r8_ * log10(temp_value)
          end if
       end if
    end if

    if (use_int_encode) then
       ! this exception is needed because the integer missing indicator NVIND
       ! is different from the real missing indicator RVIND !
       if (missing_int(nint(temp_value))) then
          temp_value = RVIND
       end if
    end if

    values(index) = temp_value

    if (error_flag .ne. 0) then
       call logmsg(log_error,"a problem was detected when filling: "//&
            trim(description))
       ! todo: use the arrays returned by the buxdes() routine to provide 
       ! some additional information on the problematic item
    end if

  end subroutine fill_one_element
    !  #]
  !---------------------------------
  ! A.1 code_sun_visible()
  function code_sun_visibility(ee_code) result(bufr_code)
    !  #[
  !
  ! retrieve code value from
  ! target_to_sun_visibility_flag
  !
  !   sun eclipsed by earth : -1 ==> 0
  !   sun center visible    : +1 ==> 1
  !
    integer(IntAs), intent( in) :: ee_code
    integer                     :: bufr_code ! result
    ! remark: intent statements are not allowed for function results. JK.

    logical :: error_occurred

    bufr_code = -1
    error_occurred = .false.

    select case (ee_code)
      case ( -1 )
         bufr_code =  0
      case ( +1 )
         bufr_code =  1
      case default
         error_occurred = .true.
    end select

    if (error_occurred) then
      write(*,*) "Warning: code_sun_visibility UNEXPECTED", ee_code
      bufr_code = 3
    end if

  end function code_sun_visibility
  !  #]
  !---------------------------------
  ! A.2 code_wind_bin_quality_flag()
  function code_wind_bin_quality_flag(ee_code) result(bufr_code)
    !  #[
  !
  ! retrieve code value from
  ! bin_quality_flag
  !
  ! table 017060 matches EE definition:
  ! check range
  !
  !   bit no. ==> description
  !       0       data overall invalid
  !       1       mie/rayleigh peak not found
  !       2       data saturation found
  !       3       data spike found
  !       4       reference pulse invalid
  !       5       source packet invalid
  !       6       laser frequency not locked
  !       7       spacecraft attitude not on target
  !   all 9       MISSING
  !
    integer(IntAuc), intent( in) :: ee_code
    integer                      :: bufr_code ! result
    ! remark: intent statements are not allowed for function results. JK.

    logical :: error_occurred

    bufr_code = -1
    error_occurred = .false.

    if  ( ee_code >= 0 .and. ee_code < 255 ) then
      bufr_code = ee_code
    else
      error_occurred = .true.
    end if

    if (error_occurred) then
      write(*,*) "Warning: code_wind_bin_quality_flag UNEXPECTED ", ee_code
      bufr_code = 511
    end if

  end function code_wind_bin_quality_flag
  !  #]
  !---------------------------------
  ! A.3 code_data_quality_flag()
  function code_data_quality_flag(ee_code) result(bufr_code)
    !  #[
  ! retrieve code value from
  ! bin_quality_flag
  !
  ! table 017050 matches EE definition:
  ! check range
  !
  !   bit no. ==> description
  !       0       data overall valid
  !       1       bin level validity
  !       2       data saturation found
  !       3       data spike found
  !       4       measurement validity
  !       5       source packet invalid
  !       6       laser frequency not locked
  !       7       spacecraft attitude not on target
  !   all 9       MISSING
  !
  !
    integer(IntAuc), intent( in) :: ee_code
    integer                      :: bufr_code ! result
    ! remark: intent statements are not allowed for function results. JK.

    logical :: error_occurred

    bufr_code = -1
    error_occurred = .false.

    if  ( ee_code >= 0 .and. ee_code <= 255 ) then
      bufr_code = ee_code
    else
      error_occurred = .true.
    end if

    if (error_occurred) then
      write(*,*) "Warning: code_data_quality_flag UNEXPECTED ", ee_code
      bufr_code = 511
    end if

  end function code_data_quality_flag
  !  #]
  !---------------------------------
  ! A.4 code_meas_calibration_flag()
  function code_meas_calibration_flag(ee_code) result(bufr_code)
    !  #[
  !
  ! retrieve code value from
  ! Measuremnt_calibration_Flag
  !
  ! table 017224 matches EE definition:
  !
  !   bit no. ==> description
  !       0       mean sensitivity within threshold range
  !       1       error response standard deviation within threshold range
  !       2       zero frequency response within threshold range
  !       3       measurement data is monotonic
  !   all 5       MISSING
  !
    TYPE(Meas_Calib_Valid_type), intent( in) :: ee_code
    integer                                  :: bufr_code ! result
    ! remark: intent statements are not allowed for function results. JK.

    bufr_code = 0

    if  ( ee_code%Mean_Sensitivity_Valid ) bufr_code = bufr_code+1
    if  ( ee_code%Error_Resp_Stdev_Valid ) bufr_code = bufr_code+2
    if  ( ee_code%Zero_Freq_Resp_Valid   ) bufr_code = bufr_code+4
    if  ( ee_code%Data_Monotonic         ) bufr_code = bufr_code+8

  end function code_meas_calibration_flag
  !  #]
  !---------------------------------
  ! A.5 code_refp_calibration_flag()
  function code_refp_calibration_flag(ee_code) result(bufr_code)
    !  #[
  !
  ! retrieve code value from
  ! Reference_Pulse_calibration_Flag
  !
  ! table 017225 matches EE definition:
  !
  !   bit no. ==> description
  !       0       mean sensitivity within threshold range
  !       1       error response standard deviation within threshold range
  !       2       zero frequency response within threshold range
  !       3       measurement data is monotonic
  !   all 5       MISSING
  !
  !
    TYPE(Ref_Pulse_Calib_Valid_type), intent( in) :: ee_code
    integer                                       :: bufr_code ! result
    ! remark: intent statements are not allowed for function results. JK.

    bufr_code = 0

    if  ( ee_code%Mean_Sensitivity_Valid ) bufr_code = bufr_code+1
    if  ( ee_code%Error_Resp_Stdev_Valid ) bufr_code = bufr_code+2
    if  ( ee_code%Zero_Freq_Resp_Valid   ) bufr_code = bufr_code+4
    if  ( ee_code%Data_Monotonic         ) bufr_code = bufr_code+8

  end function code_refp_calibration_flag
  !  #]
! ----------------------------

end module L1B_BufrUtil
