program L1B_ee2bufr
  !---------------------------------
  !  #[ Description:
  ! - program to convert L1B
  !   from EarthExplorer format 
  !   into BUFR presentation 
  !
  ! - based on
  !   program for the ReadL1Bdata module
  !   written by: J. de Kloe (KNMI)
  !
  ! created    : 24-11-2005 M.Rohn
  !
  ! Modifications:
  !   17-Jan-2006 M.Rohn     adopt single template (OBS+MEAS)
  !                          remove write XML-header
  !   18-Dec-2006 D.Tan      Activate reading of Cal_Char_GADS (via ReadAll)
  !   19-Jan-2007 J. de Kloe Fixed memory leak, and added proper program
  !                          termination when an error occurs
  !   03-Sep-2007 P. Poli    Introduced L1B_Headers
  !   04-Sep-2007 P. Poli    change L1B_DataType to L1B_HDR_DataType and
  !                          L1B_BRC_DataType
  !   16-Jan-2008 J. de Kloe phase out integer kind i_, added folding marks
  !   21-Feb-2008 J. de Kloe added commandline handling to allow inserting
  !                          a fake datetime
  !
  !  #]
  !---------------------------------
  !  #[ Modules used
  USE L1B_DataStructure, only: L1B_HDR_DataType, &
                               L1B_BRC_DataType, &
                               Init_L1B_HDR_DataStructure, &
                               Init_L1B_BRC_DataStructure, &
                               Delete_L1B_HDR_DataStructure, &
                               Delete_L1B_BRC_DataStructure
  USE EE_CFI_DataTypes, only: IntAl, GetIntAl, DateTimeType, &
                              Convert_XML_to_Num_DateTime, &
                              InitDateTime, DateTimeMissing

  USE ReadL1Bdata, only: ReadL1Bfile, InitReadL1BfileModule

  USE ErrorHandler, only: no_error, program_abort, error_cmdline_options, &
                          error_reading_file
  USE Compiler_Features, only: iargc_aeolus,getarg_aeolus
  
  USE L1B_BufrUtil, only: L1B_bufr_encode, &
                          L1B_bufr_encode_init, &
                          L1B_bufr_encode_finish, &
                          kbufr, &
                          NMAX_OBS_SUBSET, &
                          NOBSERVATION_FOR_DEVELOPMENT

  ! Modules used for testing
  USE L1B_Geoloc_ADS, only: Print_L1B_Geoloc_ADS
  USE L1B_PCD_ADS, only: Print_L1B_PCD_ADS
  USE L1B_GWD_ADS, only: Print_L1B_GWD_ADS
  USE L1B_Meas_ADS, only: Print_L1B_Meas_ADS
  USE L1B_US_MDS, only: Print_L1B_US_MDS
  USE L1B_WV_MDS, only: Print_L1B_WV_MDS
  USE L1B_Cal_ADS, only: Print_L1B_Cal_ADS
  USE Logging, only: Init_Logging
  !  #]
  !  #[ variables
  implicit none

  integer :: error_flag, nr_of_arguments, current_arg_nr

  character(len=256) :: filename        ! without extension
  character(len=256) :: arg_name, arg_value

  TYPE(L1B_BRC_DataType) :: L1B
  TYPE(L1B_HDR_DataType) :: L1B_Headers
  TYPE(DateTimeType)     :: FakeDateTime

  ! BUFR related
  character(len=256) :: filename_bufr   ! with    extension

  integer            :: iunit_bufr, &
                        iobservation, &
                        nmeasurement_max, &
                        nsubset, &
                        kbufr_size
  integer(IntAl) :: nobservation

  !  #]
  !---------------------------------
  !  #[ 1.0 init
  !
  ! suppress warnings (for testing purposes only!)
!  call Init_Logging(Stdout_Log_Level_txt="ERROR",&
!                    Stderr_Log_Level_txt="NOOP",&
!                    error_flag=error_flag)
  ! normal logging
  call Init_Logging(Stdout_Log_Level_txt="INFO",&
                    Stderr_Log_Level_txt="NOOP",&
                    error_flag=error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"Init_Logging")

  ! init to missing
  call InitDateTime(FakeDateTime)

  ! 1.1 L1B read access
  call InitReadL1BfileModule(error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"InitReadL1BfileModule")

  ! 1.2 fetch arguments
  ! read the filename without extension as 1st commandline parameter:
  nr_of_arguments = iargc_aeolus()
  IF ( (nr_of_arguments .ne. 1) .and. &
       (nr_of_arguments .ne. 3) .and. &
       (nr_of_arguments .ne. 5)       ) THEN
     print *,"ERROR: wrong number of commandline arguments."
     print *,"nr_of_arguments = ",nr_of_arguments
     call PrintUsage
     call program_abort(error_cmdline_options,"L1B_ee2bufr")
  END IF

  call getarg_aeolus(1,filename)

  ! 1.3 filenames
  filename_bufr = trim( filename )//".BUFR"

  ! 1.4 open BUFR file
  call pbopen(iunit_bufr,filename_bufr,'w',error_flag)
  if(error_flag.ne.0) then
    print*,'Open error on ', filename_bufr
     call program_abort(error_flag,"pbopen")
  end if
  !  #]
  !---------------------------------
  !  #[ 2.0 fill BUFR
  !
  !     build multi subset message
  !

  ! init message dimensions
  call Init_L1B_HDR_DataStructure(L1B_Headers,error_flag )
  IF (error_flag .ne. no_error) &
     call program_abort(error_flag,"Init_L1B_HDR_DataStructure")

  ! Call ReadL1Bfile and only specify to ReadHeaders, no binary DSR.
  ! The optional flags for the binary datasets are false by default.
  ! So no need to update this call when more binary datasets are added.
  call ReadL1Bfile(L1B_Headers, filename, error_flag, ReadHeaders = .true.)
  IF (error_flag .ne. no_error) &
     call program_abort(error_flag,"ReadL1Bfile")  

  !read( L1B%SPH%TOTAL_NUM_OF_OBSERVATIONS, *) nobservation
  ! replaced by:
  call GetIntAl(L1B_Headers%SPH%TOTAL_NUM_OF_OBSERVATIONS,nobservation,error_flag)
  IF (error_flag .ne. no_error) &
     call program_abort(error_flag,"GetIntAl")  

  ! thanks to the checking above, we know at this point that
  ! the nr of commandline arguments is eaither 1, 3 or 5
  current_arg_nr = 1
  if (nr_of_arguments .gt. 1) then
     argloop: do while (current_arg_nr .lt. nr_of_arguments)
        current_arg_nr = current_arg_nr + 1
        call getarg_aeolus(current_arg_nr,arg_name)
        current_arg_nr = current_arg_nr + 1
        call getarg_aeolus(current_arg_nr,arg_value)
        select case (arg_name)
        case("-NumBRC")
           ! read NumBRC from the commandline and store in nobservation
           print *,"Reducing number of read BRC's from ",nobservation
           read(arg_value,*,err=999) nobservation
           print *,"to ",nobservation
        case("-FakeDateTime")
           ! read a datetime from the commandline and store in FakeDateTime
           print *,"Using fake Datetime: "//trim(arg_value)
           FakeDateTime = Convert_XML_to_Num_DateTime("UTC="//trim(arg_value))
        case default
           print *,"ERROR: invalid commandline argument: "
           print *,"ERROR: trim(arg_name)  = ["//trim(arg_name)//"]"
           print *,"ERROR: trim(arg_value) = ["//trim(arg_value)//"]"
           call PrintUsage
           call program_abort(error_cmdline_options,"L1B_ee2bufr")
        end select
     end do argloop
  end if

  nmeasurement_max = L1B_Headers%SPH%N_Meas_Max

  ! 2.1 build obs message
  nsubset = 0
  call L1B_bufr_encode_init( nmeasurement_max )

  do iobservation = 1,nobservation

    ! init before each read
    call Init_L1B_BRC_DataStructure(L1B,error_flag )
    IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"Init_L1B_BRC_DataStructure")  

    ! Call ReadL1Bfile and now specify to ReadAll.
    ! The optional flags for the binary datasets are autmoatically updated.
    ! So no need to update this call when more binary datasets are added.
    call ReadL1Bfile(L1B_Headers, filename, error_flag,&
                     L1B = L1B, &
                     Which_DSR = iobservation, ReadAll = .true.)

    IF (error_flag .ne. no_error) call program_abort(error_flag,"ReadL1Bfile")

    ! 2.2 encode L1B observations
    nsubset = nsubset+1

    if (DateTimeMissing(FakeDateTime)) then
       call L1B_bufr_encode( L1B_Headers, L1B, nsubset, &
                             workaround=.true.)
    else
       call L1B_bufr_encode( L1B_Headers, L1B, nsubset, &
                             workaround=.true., FakeDateTime=FakeDateTime)
    end if

    ! 2.3 encode and write message L1B observations
    if ( nsubset == NMAX_OBS_SUBSET .OR. iobservation == nobservation ) then

      ! encode message
      call L1B_bufr_encode_finish( nsubset, kbufr, kbufr_size )

      ! write message
      call pbwrite(iunit_bufr,kbufr,kbufr_size*4,error_flag)
      if(error_flag.lt.0) then
        print*,'ERROR pbwrite: Write error on ', filename_bufr
        call program_abort(error_flag,"pbwrite")  
      end if

      ! init multi subset message
      nsubset = 0
      call L1B_bufr_encode_init( nmeasurement_max )

    end if

    ! 3.1 free L1B structure (BRC level)
    call Delete_L1B_BRC_DataStructure( L1B )

  enddo
  !  #]
  !---------------------------------
  !  #[ 3.0 clean

  ! 3.1 free L1B structure (header only)
  call Delete_L1B_HDR_DataStructure( L1B_Headers )

  ! 3.2 close BUFR
  call pbclose(iunit_bufr,error_flag)
  if(error_flag.ne.0) then
    print*,'Close error on ', filename_bufr
    call program_abort(error_flag,"pbclose")  
  end if 
  !  #]
  !---------------------------------
  !  #[ error handler
  goto 111 ! goto the end of the program without using stop
           ! (that messes up our testsuite for certain compilers)

999 print *,"ERROR: could not convert NumBRC given on the command line"
    print *,"into an integer ...."
    print *,"argument string value is: "//trim(arg_value)
    call PrintUsage
    call program_abort(error_cmdline_options,"L1B_ee2bufr")

111 continue
  !  #]    
contains
  !-------------------
  subroutine PrintUsage
    print *,"Usage: L1B_ee2bufr <filename[without extension]>"
    print *,"                   [optional: -NumBRC <count>]"
    print *,"                   [optional: -FakeDateTime <DateTime>]"
    print *,"with <DateTime> having the format:"
    print *,"                   <2004-10-02T00:00:00.000000>"
    print *,""
  end subroutine PrintUsage
  !-------------------
end program L1B_ee2bufr
