module Classification
  !  #[ Description:

  !     A module to
  !     -classify the by ADM determined backscatter profiles
  !      to determine whether they contain cloud, or are cloud free
  !     (later this could be expanded to determine aerosol or dust 
  !      layers as well)
  !
  ! assumption: at the moment it is assumed that the mie and rayleight
  !             channels both have the same set of range-gates
  !   
  ! created:       08-04-2005, by: J. de Kloe, KNMI
  ! Modifications: 
  !   22-04-2005 J. de Kloe
  !   09-05-2005 David Tan, Interim calculation of T/P_ref
  !   03-06-2005 J. de Kloe, Mie/Rayleigh Rangebin definitions
  !                          may differ now, and L2B struct split into L1B, L2B
  !   16-06-2005 David Tan, Account for geoid-WGS84 offset
  !   16-06-2005 David Tan, Interpolation infrastructure
  !   03-08-2005 J. de Kloe, some print format changes
  !   10-08-2005 David Tan, Interface to Rayleigh-Brillouin/ILIAD
  !   12-08-2005 David Tan, Rayleigh-Brillouin I/O in ConstructObservations
  !   06-09-2005 J. de Kloe, reverted the print format changes
  !   07-oct-2005 D. Tan, Compute weighted Mie and Rayleigh responses
  !   18-oct              Issue warnings for placeholders and modules 
  !                       not implemented Weighted V_LOS.
  !   19-oct              Flag excessive rayleigh response 
  !                       (more exception handling needed)
  !                       Temporary placeholder sin(35 degrees)
  !   20-oct              Test RRmes_weighted_in_range.
  !   26-Oct-2005 J. de Kloe, ReadMeasData adapted to variable nr of 
  !                           Measurements in the L1B datastructure
  !   02-Nov-2005 J. de Kloe, corrected div by zero problem
  !   16-Nov-2005 J. de Kloe, architecture change: splitted Classify and
  !                       ConstructObservations in a number of smaller
  !                       routines to adapt to the different datastructures 
  !                       for Mie and Rayleigh, and to make the thing
  !                       more flexible.
  !   17-Nov-2005 J. de Kloe, removed code from ReadMeasData() and placed
  !                       in the new Match_Mie_And_Rayleigh_Bins routine
  !   21-Nov-2005 J. de Kloe, added calculation for extinction
  !   01-Dec-2005 David Tan, Fill further Rayleigh_MDS parameters
  !   06-Dec-2005 David Tan, Factor 100 for hPa to Pa.
  !   21-Dec-2005 David Tan, Fill some Mie_MDS parameters
  !                          Implement spectral nonlinearity correction
  !   23-Dec-2005 David Tan, Add corr0dR from RayleighBrillouin.
  !                          Generalize avg: sum_with_weights, normalize optional
  !   02-Jan-2006 J. de Kloe, updated Mie-Rayleigh match-up code, since the
  !                       rangebin definitions seem different between 
  !                       measurements within the same BRC in our L1B testfile
  !   22-Feb-2006 P. Poli   upgrade to DLR IODD L1B-IODD.231105:
  !                         removed function sum_with_weights_spectra which
  !                         was not used, and replaced 
  !                     L1B%Geoloc_ADS%Meas_Geoloc(m)%Line_of_Sight_Velocity by
  !                     L1B%Geoloc_ADS%Meas_Geoloc(m)%AOCS_LOS_Velocity
  !   23-Feb-2006 David Tan, Add placeholder Rayleigh hlos error estimates
  !   12-May-2006  D. Tan       Null use of ParKind1 module
  !   15-May-2006 David Tan, Revise centre-of-gravity calculations
  !   16-May-2006 David Tan, Fill some Rayleigh Geoloc parameters
  !   18-May-2006 David Tan, ILIAD error: prototype Rayleigh response error
  !   19-May-2006 David Tan, Replace placeholder scattering ratio with L1B Obs scale
  !   22-May-2006 David Tan, Interface to MieResponse (Core) algorithm
  !   23-May-2006 David Tan, Pass L2B_AuxPar settings to ComputeMieResponse
  !   24-May-2006 David Tan, Upgrade datatypes for Compute_MieResponse arguments.
  !   25-May-2006 David Tan,    Replace placeholder inputs to Rayleigh error quantifier
  !   25-May-2006 David Tan, Get tobs from L1B, prototype accd input.
  !   30-May-2006 David Tan, Prototype Mie_SNR tests (L1B Obs scale, alt_bin and int_ref)
  !   01-Jun-2006 David Tan, Correct selection of internal reference pulses
  !   01-Jun-2006 David Tan, Scattering ratio from weighted L1B Meas scale
  !   02-Jun-2006 David Tan, Remove placeholders for Mie Response Slope terms
  !   05-Jun-2006 David Tan, Transfer Mie c-of-g to output datastructure
  !   07-Jun-2006 David Tan, Mie spectral nonlinearity correction (optional)
  !   07-Jun-2006 Tan/Poli,  Mie hlos_error estimate
  !   27-Jun-2006 David Tan, Fill all Ray/Mie Geolocation params (elevation with placeholder)
  !   03-Jul-2006 David Tan, Text formatting for message MieSNR < Threshold
  !   10-Jul-2006 Poli,      Upgrade Mie hlos_error estimate: weights, tobs, offset
  !   13-Jul-2006 David Tan, Adjust indexing of tobs in Mie hlos_error
  !   25-Jul-2006 David Tan, No normalization for accd_meas and accd_intref
  !   26-Jul-2006 David Tan, Equal weights all 1000, not sum of 1000
  !   09-Aug-2006 J. de Kloe added generic Classify routine
  !   17-Aug-2006  P. Poli   Removed obsolete variables and precision loss problems
  !   25-Aug-2006 David Tan, Additional REALACC around prints.
  !   30-Aug-2006 David Tan, HLOS retrieval routines moved to 
  !                          HLOS_retrieval/hlos_retrieval.F90
  !   14-Sep-2006 J. de Kloe added a stub for ext. based classification
  !   15-Sep-2006 Paul Poli, Moved Construct_Rayleigh_Obs and Construct_Mie_Obs
  !                          to separate module Construct_Obs
  !   15-Sep-2006 Paul Poli, Rejection of screened-out data in Classify_*
  !   06-Oct-2006 J. de Kloe added use of the WorkingDataType
  !   23-Oct-2006 D. Tan     Use of WorkingDataType for old L2B-working
  !   06-Jun-2007 P. Poli    Add L2B%PCD_ADS as input of subroutine Classify.
  !                          Exclude measurement range-bins for which Bin_QC 
  !                          is not OK according to L1B_Input_Screening
  !   28-Aug-2007 D. Tan     Remove use of old L1B-working
  !   03-Sep-2007 P. Poli    remove obsolete use of L1B_datatype
  !   02-Oct-2007 J. de Kloe implemented classification on extinction threshold
  !   05-Oct-2007 J. de Kloe taken out the Bin_QC test and moved it to the
  !                          new Meas_Selection_Weighting module
  !   16-Jan-2008 J. de Kloe phase out integer kind i_
  !
  !
  !  #]
  !  #[ Modules used:

  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE L2B_AuxPar_DataStructure, only: Classification_Params_type, &
                                      Class_No_Clouds, Class_Backscat_Ratio, &
                                      Class_Ext_Threshold, &
                                      Get_BackscRatio_Thresh_Profile, &
                                      Get_Extinction_Thresh_Profile
  USE L2B_Parameters, only: no_classification_possible, &
                            classified_as_cloud, classified_as_no_cloud, &
                            nr_of_classes, GetClassTypeText
  USE Working_DataStructure, only: WorkingDataType, &
                           WD_MieRangeGateType      => MieRangeGateType,&
                           WD_RayleighRangeGateType => RayleighRangeGateType
  USE EE_CFI_Datatypes,  only: IntAuc,IntAc,IntAs

  !support modules
  USE Numerics, only: r8_, missing_real
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE ErrorHandler, only: no_error, error_programming
  USE Logging, only: logmsg,log_debug,log_info,log_progress,log_warn,log_error
  ! rename this type, to prevent confusion with the other profile_type 
  ! defined in auxiliary.F90. Dito for the interpolate() function.
  USE ProfileInterpolate, only: ProfileInterpolate_type => profile_type, &
                                PrInterpolate => interpolate, &
                                print_profile, delete_profile

  !  #]
  !  #[ Parameter/Variable/Type declarations

  implicit none

  ! the default verbosity setting is Error (as silent as possible)
  integer, private, save :: LocalVerbosity = VerbosityError

  ! Threshold profiles
  TYPE(ProfileInterpolate_type), save :: BackscRatio_Thresh_Profile
  TYPE(ProfileInterpolate_type), save :: Extinction_Thresh_Profile

  ! a variable to remember if the module was initialised or not.
  logical, save :: classification_initialised = .false.
  !  #]
  !  #[ Module interface

  !  #]
contains
  !---------------------------------------------
  ! public subroutines
  !---------------------------------------------
  subroutine InitClassification(Classification_Params,error_flag, &
                                ChoosenVerbosity)
    !  #[
    TYPE(Classification_Params_type), intent(in)    :: Classification_Params
    integer,                          intent(out)   :: error_flag ! output
    integer, optional, intent(in) :: ChoosenVerbosity

    error_flag = no_error
    
    IF (present(ChoosenVerbosity)) LocalVerbosity = ChoosenVerbosity

    if (LocalVerbosity >= VerbosityWarn) &
         call print_verbosity_setting(LocalVerbosity,&
                                      ModuleName="Classification")

    ! convert the threshold profiles to a type that can be used
    ! by the ProfileInterpolate module
    call Get_BackscRatio_Thresh_Profile(Classification_Params,&
         BackscRatio_Thresh_Profile,error_flag)
    IF (error_flag .ne. no_error) return
    call Get_Extinction_Thresh_Profile(Classification_Params,&
         Extinction_Thresh_Profile, error_flag)
    IF (error_flag .ne. no_error) return

    classification_initialised = .true. 

  end subroutine InitClassification

    !  #]
  subroutine CloseClassification()
    !  #[
    ! a routine intended to free the allocated memory

    classification_initialised = .false.

    ! free the momory used by the threshold profiles
    call delete_profile(BackscRatio_Thresh_Profile)
    call delete_profile(Extinction_Thresh_Profile)

  end subroutine CloseClassification
    !  #]
  subroutine Classify(WorkingData, Classification_Params, error_flag)
    !  #[
    TYPE(WorkingDataType),            intent(inout) :: WorkingData ! input/output
    TYPE(Classification_Params_type), intent(in)    :: Classification_Params
    integer,                          intent(out)   :: error_flag ! output

    ! init
    error_flag = no_error

    IF (.not. classification_initialised) THEN
       call logmsg(log_error,"Error in Classify: the classification module"//&
            "most be initialised before use by calling InitClassification!!!")
       error_flag = error_programming
       return
    END IF

    ! do the classification in cloud/no_cloud points for the Rayleighdata
    select case (Classification_Params%Classification_Type_Rayleigh)
    case(Class_No_Clouds)
       call Classify_Rayl_Meas_No_Clouds(WorkingData)
    case(Class_Backscat_Ratio)
       call Classify_Rayl_Meas_Backsc_Ratio(WorkingData, &
                          BackscRatio_Thresh_Profile, error_flag)
       IF (error_flag .ne. no_error) return
    case(Class_Ext_Threshold)
       call Classify_Rayl_Ext_Threshold(WorkingData, &
                                     Extinction_Thresh_Profile, error_flag)
       IF (error_flag .ne. no_error) return
    case default
       error_flag = error_programming
       call logmsg(log_error,"Classify: Classification_Type_Rayleigh"//&
                             " has an unknown value")
       return
    end select

    ! do the classification in cloud/no_cloud points for the Mie data
    select case (Classification_Params%Classification_Type_Mie)
    case(Class_No_Clouds)
       call Classify_Mie_Meas_No_Clouds(WorkingData)
    case(Class_Backscat_Ratio)
       call Classify_Mie_Meas_Backsc_Ratio(WorkingData, &
                                     BackscRatio_Thresh_Profile, error_flag)
       IF (error_flag .ne. no_error) return
    case(Class_Ext_Threshold)
       call Classify_Mie_Ext_Threshold(WorkingData, &
                                     Extinction_Thresh_Profile, error_flag)
       IF (error_flag .ne. no_error) return
    case default
       error_flag = error_programming
       call logmsg(log_error,"Classify: Classification_Type_Mie"//&
                             " has an unknown value")
       return
    end select

  end subroutine Classify
    !  #]
  subroutine Classify_Mie_Meas_No_Clouds(WorkingData)
    !  #[

    ! this routine implements a dummy classification of Rayleigh 
    ! and Mie rangebins, and just assumes no clouds everywhere

    TYPE(WorkingDataType), intent(inout) :: WorkingData ! input/output

    ! local variables
    integer :: m,r

    IF (LocalVerbosity >= VerbosityReport) &
       print *,"starting no-cloud classification of Mie measurements"

    measloop: DO m=1, WorkingData%nr_of_measurements
       Rangegateloop: DO r=1, WorkingData%MieMeas(m)%nr_of_range_gates
          WorkingData%MieMeas(m)%RangeGate(r)%class = classified_as_no_cloud
       END DO Rangegateloop
    END DO measloop

  end subroutine Classify_Mie_Meas_No_Clouds

    !  #]
  subroutine Classify_Rayl_Meas_No_Clouds(WorkingData)
    !  #[
    ! this routine implements a dummy classification of Rayleigh 
    ! and Mie rangebins, and just assumes no clouds everywhere

    TYPE(WorkingDataType), intent(inout) :: WorkingData ! input/output

    ! local variables
    integer :: m,r

    IF (LocalVerbosity >= VerbosityReport) &
       print *,"starting no-cloud classification of Rayleigh measurements"

    measloop: DO m=1, WorkingData%nr_of_measurements
       Rangegateloop: DO r=1, WorkingData%RayleighMeas(m)%nr_of_range_gates
          WorkingData%RayleighMeas(m)%RangeGate(r)%class = classified_as_no_cloud
       END DO Rangegateloop
    END DO measloop

  end subroutine Classify_Rayl_Meas_No_Clouds
    !  #]
  subroutine Classify_Mie_Meas_Backsc_Ratio(WorkingData, &
                                            BackscRatio_Thresh_Profile,&
                                            error_flag)
    !  #[
    ! this routine implements classification of Mie rangebins
    ! using a threshold on the backscatter ratio

    TYPE(WorkingDataType),         intent(inout) :: WorkingData ! input/output
    ! intent needs to be inout because the ProfileInterpolate module
    ! stores some items (like last used index) inside this struct
    TYPE(ProfileInterpolate_type), intent(inout) :: BackscRatio_Thresh_Profile
    integer,                       intent(out)   :: error_flag ! output

    ! local variables
    real(r8_) :: BackscatterRatio, altitude, Threshold_BackscatterRatio
    integer   :: m,r
    TYPE(WD_MieRangeGateType), pointer  :: WD_MieRangeGate 

    ! init
    error_flag = no_error

    IF (LocalVerbosity >= VerbosityReport) THEN
       print *,"starting classification of Mie measurements using"
       print *,"a threshold on the backscatter ratio"
       call print_profile(BackscRatio_Thresh_Profile)
       print "(a8,a16,a)","altitude","BackscatterRatio","Class"
    END IF

    measloop: DO m=1, WorkingData%nr_of_measurements
       Rangegateloop: DO r=1, WorkingData%MieMeas(m)%nr_of_range_gates
          !  #[ Classify the Mie data using BackscatterRatio

          ! point to this RangeGate for shorter notation
          WD_MieRangeGate => WorkingData%MieMeas(m)%RangeGate(r)
          BackscatterRatio = WD_MieRangeGate%backscatter_ratio
          IF ( missing_real(BackscatterRatio) .or. &
               .not.(WD_MieRangeGate%rho_ok)        ) THEN
             WD_MieRangeGate%class = no_classification_possible
          ELSE
             IF ( missing_real(WD_MieRangeGate%alt_start) .or. &
                  missing_real(WD_MieRangeGate%alt_end  )       ) THEN
                WD_MieRangeGate%class = no_classification_possible
             ELSE
                altitude = 0.5*(WD_MieRangeGate%alt_start + &
                                WD_MieRangeGate%alt_end     )
                Threshold_BackscatterRatio = &
                     PrInterpolate(BackscRatio_Thresh_Profile,altitude)
                ! then if the backscatter_ratio is above the threshold
                ! mark this rangebin as cloud
                IF (BackscatterRatio .gt. Threshold_BackscatterRatio) THEN
                   WD_MieRangeGate%class = classified_as_cloud
                ELSE
                   WD_MieRangeGate%class = classified_as_no_cloud
                END IF
             END IF

             IF (LocalVerbosity >= VerbosityReport) THEN
                print "(i5,3x,f7.3,9x,a)",&
                     nint(altitude),BackscatterRatio,&
                     trim(GetClassTypeText(WD_MieRangeGate%class))
             END IF

          END IF ! Bin_QC and missing check

       !  #]
       END DO Rangegateloop
    END DO measloop

  end subroutine Classify_Mie_Meas_Backsc_Ratio
    !  #]
  subroutine Classify_Rayl_Meas_Backsc_Ratio(WorkingData, &
                                             BackscRatio_Thresh_Profile,&
                                             error_flag)
    !  #[
    ! this routine implements classification of Rayleigh rangebins
    ! using a threshold on the backscatter ratio

    TYPE(WorkingDataType),         intent(inout)    :: WorkingData ! input/output
    ! intent needs to be inout because the ProfileInterpolate module
    ! stores some items (like last used index) inside this struct
    TYPE(ProfileInterpolate_type), intent(inout) :: BackscRatio_Thresh_Profile
    integer,                       intent(out)   :: error_flag ! output

    ! local variables
    real(r8_) :: BackscatterRatio, altitude, Threshold_BackscatterRatio
    integer   :: m,r
    TYPE(WD_RayleighRangeGateType),  pointer :: WD_RayleighRangeGate 

    ! init
    error_flag = no_error

    IF (LocalVerbosity >= VerbosityReport) THEN
       print *,"starting classification of Rayleigh measurements using"
       print *,"printing BackscRatio_Thresh_Profile:"
       call print_profile(BackscRatio_Thresh_Profile)
    END IF

    measloop: DO m=1, WorkingData%nr_of_measurements
       Rangegateloop: DO r=1, WorkingData%RayleighMeas(m)%nr_of_range_gates
          !  #[ Classify the Rayleigh data using BackscatterRatio

          ! point to this RangeGate for shorter notation
          WD_RayleighRangeGate => WorkingData%RayleighMeas(m)%RangeGate(r)
          BackscatterRatio = WD_RayleighRangeGate%backscatter_ratio
          IF (missing_real(BackscatterRatio).or. &
               .not.(WD_RayleighRangeGate%rho_ok)) THEN
!          IF (missing_real(BackscatterRatio)) THEN
             WD_RayleighRangeGate%class = no_classification_possible
          ELSE
             IF ( missing_real(WD_RayleighRangeGate%alt_start) .or. &
                  missing_real(WD_RayleighRangeGate%alt_end  )      ) THEN
                WD_RayleighRangeGate%class = no_classification_possible
                !print *,"testjos: MISSING ALTITUDE !!!"
             ELSE
                altitude = 0.5*(WD_RayleighRangeGate%alt_start + &
                                WD_RayleighRangeGate%alt_end     )
                Threshold_BackscatterRatio = &
                     PrInterpolate(BackscRatio_Thresh_Profile,altitude)
                ! then if the backscatter_ratio is above the threshold
                ! mark this rangebin as cloud
                IF (BackscatterRatio .ge. Threshold_BackscatterRatio) THEN
                   WD_RayleighRangeGate%class = classified_as_cloud
                ELSE
                   WD_RayleighRangeGate%class = classified_as_no_cloud
                END IF
             END IF
          END IF

       !  #]
       END DO Rangegateloop
    END DO measloop

  end subroutine Classify_Rayl_Meas_Backsc_Ratio
    !  #]
  subroutine Classify_Mie_Ext_Threshold(WorkingData, &
                                        Extinction_Thresh_Profile,&
                                        error_flag)
    !  #[

    ! this routine implements a classification by applying a
    ! threshold on the extinction values for the Mie rangebins

    TYPE(WorkingDataType),         intent(inout) :: WorkingData ! input/output
    ! intent needs to be inout because the ProfileInterpolate module
    ! stores some items (like last used index) inside this struct
    TYPE(ProfileInterpolate_type), intent(inout) :: Extinction_Thresh_Profile
    integer,                       intent(out)   :: error_flag ! output

    ! local variables
    real(r8_) :: Extinction, altitude, Threshold_Extinction
    integer   :: m,r
    TYPE(WD_MieRangeGateType), pointer  :: WD_MieRangeGate 

    ! init
    error_flag = no_error

    IF (LocalVerbosity >= VerbosityReport) THEN
       print *,"starting extinction threshold classification on"
       print *,"Mie measurements"
       call print_profile(Extinction_Thresh_Profile)
       print "(a8,a16,a)","altitude","Extinction","Class"
    END IF

    measloop: DO m=1, WorkingData%nr_of_measurements
       Rangegateloop: DO r=1, WorkingData%MieMeas(m)%nr_of_range_gates
          !  #[ Classify the Mie data using Extinction

          ! point to this RangeGate for shorter notation
          WD_MieRangeGate => WorkingData%MieMeas(m)%RangeGate(r)
          Extinction = WD_MieRangeGate%extinction

          IF ( missing_real(Extinction) ) THEN
             WD_MieRangeGate%class = no_classification_possible
          ELSE
             IF ( missing_real(WD_MieRangeGate%alt_start) .or. &
                  missing_real(WD_MieRangeGate%alt_end  )       ) THEN
                WD_MieRangeGate%class = no_classification_possible
             ELSE
                altitude = 0.5*(WD_MieRangeGate%alt_start + &
                                WD_MieRangeGate%alt_end     )
                Threshold_Extinction = &
                     PrInterpolate(Extinction_Thresh_Profile,altitude)
                ! then if the extinction is above the threshold
                ! mark this rangebin as cloud
                IF (Extinction .gt. Threshold_Extinction) THEN
                   WD_MieRangeGate%class = classified_as_cloud
                ELSE
                   WD_MieRangeGate%class = classified_as_no_cloud
                END IF
             END IF

             IF (LocalVerbosity >= VerbosityReport) THEN
                print "(i5,3x,f7.3,9x,a)",&
                     nint(altitude),Extinction,&
                     trim(GetClassTypeText(WD_MieRangeGate%class))
             END IF
          END IF ! Bin_QC and missing check
          !  #]
       END DO Rangegateloop
    END DO measloop

  end subroutine Classify_Mie_Ext_Threshold
    !  #]
  subroutine Classify_Rayl_Ext_Threshold(WorkingData, &
                                         Extinction_Thresh_Profile,&
                                         error_flag)
    !  #[

    ! this routine implements a classification by applying a
    ! threshold on the extinction values for the Rayleigh rangebins

    TYPE(WorkingDataType),         intent(inout) :: WorkingData ! input/output
    ! intent needs to be inout because the ProfileInterpolate module
    ! stores some items (like last used index) inside this struct
    TYPE(ProfileInterpolate_type), intent(inout) :: Extinction_Thresh_Profile
    integer,                       intent(out)   :: error_flag ! output

    ! local variables
    real(r8_) :: Extinction, altitude, Threshold_Extinction
    integer   :: m,r
    TYPE(WD_RayleighRangeGateType), pointer  :: WD_RayleighRangeGate 

    ! init
    error_flag = no_error

    IF (LocalVerbosity >= VerbosityReport) THEN
       print *,"starting extinction threshold classification on"
       print *,"Rayleigh measurements"
       call print_profile(Extinction_Thresh_Profile)
       print "(a8,a16,a)","altitude","Extinction","Class"
    END IF

    measloop: DO m=1, WorkingData%nr_of_measurements
       Rangegateloop: DO r=1, WorkingData%RayleighMeas(m)%nr_of_range_gates
          !  #[

          ! point to this RangeGate for shorter notation
          WD_RayleighRangeGate => WorkingData%RayleighMeas(m)%RangeGate(r)
          Extinction = WD_RayleighRangeGate%extinction

          IF ( missing_real(Extinction) ) THEN
             WD_RayleighRangeGate%class = no_classification_possible
          ELSE
             IF ( missing_real(WD_RayleighRangeGate%alt_start) .or. &
                  missing_real(WD_RayleighRangeGate%alt_end  )       ) THEN
                WD_RayleighRangeGate%class = no_classification_possible
             ELSE
                altitude = 0.5*(WD_RayleighRangeGate%alt_start + &
                                WD_RayleighRangeGate%alt_end     )
                Threshold_Extinction = &
                     PrInterpolate(Extinction_Thresh_Profile,altitude)
                ! then if the extinction is above the threshold
                ! mark this rangebin as cloud
                IF (Extinction .gt. Threshold_Extinction) THEN
                   WD_RayleighRangeGate%class = classified_as_cloud
                ELSE
                   WD_RayleighRangeGate%class = classified_as_no_cloud
                END IF
             END IF

             IF (LocalVerbosity >= VerbosityReport) THEN
                print "(i5,3x,f7.3,9x,a)",&
                     nint(altitude),Extinction,&
                     trim(GetClassTypeText(WD_RayleighRangeGate%class))
             END IF
          END IF ! Bin_QC and missing check
          !  #]
       END DO Rangegateloop
    END DO measloop

  end subroutine Classify_Rayl_Ext_Threshold
    !  #]
  !------------------------------------------

end module Classification
