module Construct_Obs
  !  #[ Description:

  !     A module to
  !     -construct the observations based on the results 
  !      of the classification algorithm
  !
  ! created:       15-Sep-2006, by:
  !                 P. Poli and J. de Kloe (METEO FRANCE and KNMI)
  ! Modifications: 
  !   15-Sep-2006 Paul Poli, Code moved from Classification
  !
  !   06-Oct-2006 J. de Kloe added use of the WorkingDataType
  !   23-Oct-2006 D. Tan     Use of WorkingDataType for old L2B-working
  !   28-Aug-2007 D. Tan     Remove use of old L1B-working
  !   03-Sep-2007 P. Poli    change L2BC_DataType to L2BC_BRC_DataType
  !   05-Oct-2007 J. de Kloe taken out the filling of the weights and moved 
  !                          it to the new Meas_Selection_Weighting module
  !   16-Jan-2008 J. de Kloe phase out integer kind i_
  !  #]
  !  #[ Modules used:

  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE L2BC_DataStructure, only: L2BC_BRC_DataType, &
                               L2B_MieRangeGateType, &
                               L2B_RayleighRangeGateType
  USE L2B_Rayleigh_MDS, only: get_Rayl_nr_of_cloud_points, &
                              get_Rayl_nr_of_clear_points
  USE L2B_Mie_MDS, only: get_Mie_nr_of_cloud_points, &
                         get_Mie_nr_of_clear_points
  USE Meas_Map, only: Meas_Map_type
  USE L2B_Parameters, only: nr_of_classes, GetClassTypeText
  USE Working_DataStructure, only: WorkingDataType, &
                           WD_MieRangeGateType      => MieRangeGateType,&
                           WD_RayleighRangeGateType => RayleighRangeGateType
  USE EE_CFI_Datatypes,  only: IntAuc,IntAc,IntAs

  !support modules
  USE Numerics, only: r8_, missing_real
  USE ErrorHandler, only: no_error, error_programming
  USE Logging, only: logmsg,log_debug,log_info,log_progress,log_warn,log_error

  !  #]
  !  #[ Parameter/Variable/Type declarations
  implicit none
  !  #]
  !  #[ Module interface

  ! all routines/functions are public unless stated otherwise below
  private Constr_Rayl_Meas_String, Constr_Rayl_Obs_String, &
          Constr_Mie_Meas_String, Constr_Mie_Obs_String

  !  #]
contains
  !---------------------------------------------
  ! public subroutines
  !---------------------------------------------
  subroutine Construct_Rayleigh_Obs(WorkingData, L2B, error_flag)
    !  #[
    TYPE(WorkingDataType),   intent(inout) :: WorkingData ! input/output
    TYPE(L2BC_BRC_DataType), intent(inout) :: L2B         ! input/output
    integer,                 intent(out)   :: error_flag  ! output

    ! local variables
    integer :: r,m,class,o
    integer :: nr_cases_of_this_class
!    TYPE(L2B_ObsRangeGateType), pointer :: L2B_ObsRangeGate  ! never used
    TYPE(Meas_Map_type),        pointer :: Map

    ! init
    error_flag = no_error
    ! point to the map for shorter notation
    Map => L2B%Rayleigh_MDS%Map_of_L1B_Measurements_Used

    ! algorithm:
    ! -loop over the possible classes (cloud/no_cloud)
    ! -count if there are any valid measurements in this class (on any level!)
    ! -if there are some, add an observation

    ! remember: 
    ! not_yet_classified     = 0
    ! classified_as_cloud    = 1
    ! classified_as_no_cloud = 2
    loopclasses: DO class = 1,nr_of_classes

       nr_cases_of_this_class = 0
       measloop1: DO m=1, L2B%nr_of_measurements
          rangegateloop1: DO r=1, WorkingData%RayleighMeas(m)%nr_of_range_gates
             IF (WorkingData%RayleighMeas(m)%RangeGate(r)%class .eq. class) THEN
                ! due to possible loss of accuracy, dont use this yet
                !IF (Map%Row(m)%Bin(r)%Weight .gt. 0) &
                ! but use this in stead
                IF (WorkingData%RayleighMeas(m)%RangeGate(r)%weight .gt. 0._r8_) &
                     nr_cases_of_this_class = nr_cases_of_this_class + 1
             END IF
          END DO rangegateloop1
       END DO measloop1

       addobs: IF (nr_cases_of_this_class .gt. 0) THEN
          ! OK, measurements of this class are present
          ! on some level and for some measurement, somewhere in this BRC
          ! so create a new observation for it.
          L2B%Rayleigh_MDS%N_Obs_Rayleigh_Actual = &
               L2B%Rayleigh_MDS%N_Obs_Rayleigh_Actual + 1_IntAs
          ! copy this Observation nr. for sake of shorter notation
          o = L2B%Rayleigh_MDS%N_Obs_Rayleigh_Actual

          ! sanity check on the datastructure 
          IF ( o .gt. size(L2B%Rayleigh_MDS%Rayleigh_Profile)) THEN
             print *,"ERROR in Construct_Rayleigh_Obs():"
             print *,"nr of observations to be generated is larger that the"
             print *,"space available in the L2B%Rayleigh_MDS datastructure !"
             error_flag = error_programming
             return
          END IF

          ! remember which class this observation has (cloud/nocloud)
          ! this is needed since only the observation types that
          ! contain data are defined, so the order is not know a priori
          L2B%Rayleigh_MDS%Rayleigh_Profile(o)%Obs_Type = int(class,IntAc)

          rangegateloop2: DO r=1, WorkingData%RayleighMeas(1)%nr_of_range_gates
             !  #[ define the Observation for this level

             ! count at each level how many cases we have of this class
             nr_cases_of_this_class = 0
             measloop2: DO m=1, L2B%nr_of_measurements
                IF (WorkingData%RayleighMeas(m)%RangeGate(r)%class .eq. class) THEN
                   nr_cases_of_this_class = nr_cases_of_this_class + 1
                   ! remember which measurement contributed to this obs.
                   Map%Row(m)%Bin(r)%which_L2B_Obs = int(o,IntAuc)
                END IF
             END DO measloop2

             ! remember the count for this class in the work structure
             L2B%Obs(o)%RangeGate(r)%nr_of_measurements = &
                  nr_cases_of_this_class

             ! copy some basic properties of the range-bins to the L2B%obs(:)
             ! struct note that this is a working structure for intermediate 
             ! results and an extra copy the Geolocation_ADS is needed
             
!             ! point to this ObsRangeGate for shorter notation
!             L2B_ObsRangeGate => L2B%Obs(o)%RangeGate(r) ! not used
             
             ! copy the altitudes of the range gates from the Measurement
! is this still used?
             L2B%Obs(o)%RangeGate(r)%alt_start = &
                  WorkingData%RayleighMeas(1)%RangeGate(r)%alt_start
             L2B%Obs(o)%RangeGate(r)%alt_end   = &
                  WorkingData%RayleighMeas(1)%RangeGate(r)%alt_end

             !  #]
          END DO rangegateloop2

          ! testjos
          print *,"Succesfully added a Rayleigh observation for ",&
               "class: ",class," = ",trim(GetClassTypeText(class))
          print *,"L2B%Rayleigh_MDS%N_Obs_Actual = ",&
               L2B%Rayleigh_MDS%N_Obs_Rayleigh_Actual
          print *,"L2B%Rayleigh_MDS%N_Obs_Max    = ",&
               size(L2B%Rayleigh_MDS%Rayleigh_Profile)
!               L2B%Rayleigh_MDS%N_Obs_Max
       ELSE
          print *,"Not added a Rayleigh observation for: ",&
               "class: ",class," = ",trim(GetClassTypeText(class))
       END IF addobs

    END DO loopclasses

  end subroutine Construct_Rayleigh_Obs
    !  #]
  subroutine Construct_Mie_Obs(WorkingData, L2B, error_flag)
    !  #[
    TYPE(WorkingDataType),   intent(inout) :: WorkingData ! input
    TYPE(L2BC_BRC_DataType), intent(inout) :: L2B         ! input/output
    integer,                 intent(out)   :: error_flag  ! output

    ! local variables
    integer :: r,m,class,o
    integer :: nr_cases_of_this_class, offset
!    TYPE(L2B_ObsRangeGateType),  pointer :: L2B_ObsRangeGate 
    TYPE(Meas_Map_type), pointer :: Map

    ! init
    error_flag = no_error
    ! point to the map for shorter notation
    Map => L2B%Mie_MDS%Map_of_L1B_Measurements_Used

    offset = L2B%Rayleigh_MDS%N_Obs_Rayleigh_Actual

    ! algorithm:
    ! -loop over the possible classes (cloud/no_cloud)
    ! -count if there are any valid measurements in this class (on any level!)
    ! -if there are some, add an observation

    ! remember: 
    ! not_yet_classified     = 0
    ! classified_as_cloud    = 1
    ! classified_as_no_cloud = 2
    loopclasses: DO class = 1,nr_of_classes

       nr_cases_of_this_class = 0
       measloop1: DO m=1, L2B%nr_of_measurements
          rangegateloop1: DO r=1, WorkingData%MieMeas(m)%nr_of_range_gates
             IF (WorkingData%MieMeas(m)%RangeGate(r)%class .eq. class) THEN
                ! due to possible loss of accuracy, dont use this yet
                !IF (Map%Row(m)%Bin(r)%Weight .gt. 0) &
                ! but use this in stead
                IF (WorkingData%MieMeas(m)%RangeGate(r)%weight .gt. 0._r8_) &
                     nr_cases_of_this_class = nr_cases_of_this_class + 1
             END IF
          END DO rangegateloop1
       END DO measloop1

       addobs: IF (nr_cases_of_this_class .gt. 0) THEN
          ! OK, measurements of this class are present
          ! on some level and for some measurement, somewhere in this BRC
          ! so create a new observation for it.
          L2B%Mie_MDS%N_Obs_Mie_Actual = L2B%Mie_MDS%N_Obs_Mie_Actual + 1_IntAs
          ! copy this Observation nr. for sake of shorter notation
          o = L2B%Mie_MDS%N_Obs_Mie_Actual

          ! sanity check on the datastructure 
          IF ( o .gt. size(L2B%Mie_MDS%Mie_Profile)) THEN
             print *,"ERROR in Construct_Mie_Obs():"
             print *,"nr of observations to be generated is larger that the"
             print *,"space available in the L2B%Mie_MDS datastructure !"
             error_flag = error_programming
             return
          END IF

          ! remember which class this observation has (cloud/nocloud)
          ! this is needed since only the observation types that
          ! contain data are defined, so the order is not know a priori
          L2B%Mie_MDS%Mie_Profile(o)%Obs_Type = int(class,IntAc)

          rangegateloop2: DO r=1, WorkingData%MieMeas(1)%nr_of_range_gates
             !  #[ define the Observation for this level

             ! count at each level how many cases we have of this class
             nr_cases_of_this_class = 0
             measloop2: DO m=1, L2B%nr_of_measurements
                IF (WorkingData%MieMeas(m)%RangeGate(r)%class .eq. class) THEN
                   nr_cases_of_this_class = nr_cases_of_this_class + 1
                   ! remember which measurement contributed to this obs.
                   Map%Row(m)%Bin(r)%which_L2B_Obs = int(o,IntAuc)
                END IF
             END DO measloop2

             ! remember the count for this class in the work structure
             L2B%Obs(o+offset)%RangeGate(r)%nr_of_measurements = &
                  nr_cases_of_this_class

             ! copy some basic properties of the range-bins to the L2B%obs(:)
             ! struct note that this is a working structure for intermediate 
             ! results and an extra copy the Geolocation_ADS is needed
             
             ! copy the altitudes of the range gates from the Measurement
             L2B%Obs(o+offset)%RangeGate(r)%alt_start = &
                  WorkingData%MieMeas(1)%RangeGate(r)%alt_start
             L2B%Obs(o+offset)%RangeGate(r)%alt_end   = &
                  WorkingData%MieMeas(1)%RangeGate(r)%alt_end

             !  #]
          END DO rangegateloop2

          ! testjos
          print *,"Succesfully added a Mie observation for ",&
               "class: ",class," = ",trim(GetClassTypeText(class))
          print *,"L2B%Mie_MDS%N_Obs_Actual = ", &
               L2B%Mie_MDS%N_Obs_Mie_Actual
          print *,"L2B%Mie_MDS%N_Obs_Max    = ", &
               size(L2B%Mie_MDS%Mie_Profile)
!               L2B%Mie_MDS%N_Obs_Max
       ELSE
          print *,"Not added a Mie observation for: ",&
               "class: ",class," = ",trim(GetClassTypeText(class))
       END IF addobs

    END DO loopclasses

  end subroutine Construct_Mie_Obs
    !  #]
  !---------------------------------------------
  ! printing subroutines
  !---------------------------------------------
  ! do a simple print, only counts for cloud/no_cloud for Mie/Rayl/Obs
  subroutine Print_Constructed_Obs(L2B)
    !  #[
    TYPE(L2BC_BRC_DataType), intent(in) :: L2B ! input

    ! local variables
    integer :: r
    integer :: rayleigh_nr_of_cloud_points
    integer :: rayleigh_nr_of_clear_points
    integer :: rayleigh_nr_of_missing_points
    integer :: mie_nr_of_cloud_points
    integer :: mie_nr_of_clear_points
    integer :: mie_nr_of_missing_points

    rangegateloop4: DO r=1, L2B%Obs(1)%nr_of_range_gates
       
       rayleigh_nr_of_cloud_points = &
            get_Rayl_nr_of_cloud_points(L2B%Rayleigh_MDS,r)
       rayleigh_nr_of_clear_points = &
            get_Rayl_nr_of_clear_points(L2B%Rayleigh_MDS,r)
       
       rayleigh_nr_of_missing_points = L2B%nr_of_measurements - &
                                       rayleigh_nr_of_cloud_points - &
                                       rayleigh_nr_of_clear_points 
       
       mie_nr_of_cloud_points = get_Mie_nr_of_cloud_points(L2B%Mie_MDS,r)
       mie_nr_of_clear_points = get_Mie_nr_of_clear_points(L2B%Mie_MDS,r)
       mie_nr_of_missing_points = L2B%nr_of_measurements - &
                                  mie_nr_of_cloud_points - &
                                  mie_nr_of_clear_points 
    
       ! print first Mie, then Rayleigh to be able to compare
       ! to previous versions
       ! a number of items are printed twice, to remain 
       ! compatible with older versions ....
       print "(3(i2,1x),':',1x,3(i2,1x),':',1x,4(i2,1x))",&
            mie_nr_of_missing_points, &
            mie_nr_of_cloud_points, mie_nr_of_clear_points, &
            rayleigh_nr_of_missing_points, &
            rayleigh_nr_of_cloud_points, rayleigh_nr_of_clear_points, &
            mie_nr_of_cloud_points, mie_nr_of_clear_points, &
            rayleigh_nr_of_cloud_points, rayleigh_nr_of_clear_points
    END DO rangegateloop4
    
  end subroutine Print_Constructed_Obs
    !  #]
  !---------------------------------------------
  ! private subroutines
  !---------------------------------------------
  ! 4 helper functions for the Print_Constr_Obs_Matched routine below
  function Constr_Rayl_Meas_String(L2B,r) result(txt)
    !  #[
    TYPE(L2BC_BRC_DataType), intent(in) :: L2B ! input
    integer,                 intent(in) :: r   ! input
    character(len=9)                    :: txt ! result
 
    ! local variables
    integer :: rayleigh_nr_of_missing_points
    integer :: rayleigh_nr_of_cloud_points
    integer :: rayleigh_nr_of_clear_points 

    rayleigh_nr_of_cloud_points = &
         get_Rayl_nr_of_cloud_points(L2B%Rayleigh_MDS,r)
    rayleigh_nr_of_clear_points = &
         get_Rayl_nr_of_clear_points(L2B%Rayleigh_MDS,r)
    rayleigh_nr_of_missing_points = &
         L2B%nr_of_measurements - &
         rayleigh_nr_of_cloud_points - &
         rayleigh_nr_of_clear_points 
    
    write(txt,"(3(i2,1x))") rayleigh_nr_of_missing_points, &
                            rayleigh_nr_of_cloud_points, &
                            rayleigh_nr_of_clear_points 

  end function Constr_Rayl_Meas_String
  !  #]
  function Constr_Mie_Meas_String(L2B,r) result(txt)
    !  #[
    TYPE(L2BC_BRC_DataType), intent(in) :: L2B ! input
    integer,                 intent(in) :: r   ! input
    character(len=9)                    :: txt ! result
 
    ! local variables
    integer :: mie_nr_of_missing_points
    integer :: mie_nr_of_cloud_points
    integer :: mie_nr_of_clear_points 

    mie_nr_of_cloud_points = get_Mie_nr_of_cloud_points(L2B%Mie_MDS,r)
    mie_nr_of_clear_points = get_Mie_nr_of_clear_points(L2B%Mie_MDS,r)
    mie_nr_of_missing_points = L2B%nr_of_measurements - &
                               mie_nr_of_cloud_points - &
                               mie_nr_of_clear_points 
    
    write(txt,"(3(i2,1x))") mie_nr_of_missing_points, &
                            mie_nr_of_cloud_points, &
                            mie_nr_of_clear_points 

  end function Constr_Mie_Meas_String
  !  #]
  function Constr_Rayl_Obs_String(L2B,r) result(txt)
    !  #[ 
    TYPE(L2BC_BRC_DataType), intent(in) :: L2B ! input
    integer,                 intent(in) :: r   ! input
    character(len=6)                    :: txt ! result

    ! local variables
    integer :: rayleigh_nr_of_cloud_points
    integer :: rayleigh_nr_of_clear_points 

    rayleigh_nr_of_cloud_points = &
         get_Rayl_nr_of_cloud_points(L2B%Rayleigh_MDS,r)
    rayleigh_nr_of_clear_points = &
         get_Rayl_nr_of_clear_points(L2B%Rayleigh_MDS,r)

    write(txt,"(2(i2,1x))") rayleigh_nr_of_cloud_points,&
                            rayleigh_nr_of_clear_points

  end function Constr_Rayl_Obs_String
  !  #]
  function Constr_Mie_Obs_String(L2B,r) result(txt)
    !  #[ 
    TYPE(L2BC_BRC_DataType), intent(in) :: L2B ! input
    integer,                 intent(in) :: r   ! input
    character(len=6)                    :: txt ! result

    ! local variables
    integer :: mie_nr_of_cloud_points
    integer :: mie_nr_of_clear_points 

    mie_nr_of_cloud_points = get_Mie_nr_of_cloud_points(L2B%Mie_MDS,r)
    mie_nr_of_clear_points = get_Mie_nr_of_clear_points(L2B%Mie_MDS,r)

    write(txt,"(2(i2,1x))") mie_nr_of_cloud_points, mie_nr_of_clear_points

  end function Constr_Mie_Obs_String
  !  #]
  !---------------------------------------------
  ! again a public subroutine
  !---------------------------------------------
  ! dito, but now matching rangebins properly to each other
  subroutine Print_Constr_Obs_Matched(WorkingData,L2B)
    !  #[
    TYPE(WorkingDataType), intent(in) :: WorkingData ! input
    TYPE(L2BC_BRC_DataType),   intent(in) :: L2B         ! input

    ! local variables
    character(len=9) :: txt_Mie_Meas
    character(len=9) :: txt_Rayl_Meas
    character(len=6) :: txt_Mie_Obs
    character(len=6) :: txt_Rayl_Obs
    integer          :: mrb1, mrb2, r_m, r, last_mie_rangebin_printed

    ! remember which Mie rangebin was last printed along with a corresponding
    ! Rayleigh rangebin, in order to be able to print excess Mie rangebins
    ! below the lowest Rayleigh rangebin.
    last_mie_rangebin_printed = 0

    write(*,"(2a)") "Mie        Rayleigh   Mie     Rayleigh"
    write(*,"(2a)") "Meas       Meas       Obs     Obs  "
    write(*,"(2a)") "mi cl nc : mi cl nc : cl nc : cl nc"

    ! loop along all Rayleigh rangebins first, since they normally
    ! starts highest up in the atmosphere
    do r=1,WorkingData%RayleighMeas(1)%nr_of_range_gates
       ! copy some Rayleigh rangebin fields for shorter notation
       mrb1 = WorkingData%RayleighMeas(1)%RangeGate(r)%first_mie_rangebin
       mrb2 = WorkingData%RayleighMeas(1)%RangeGate(r)%last_mie_rangebin

       txt_Rayl_Meas = Constr_Rayl_Meas_String(L2B,r)
       txt_Rayl_Obs  = Constr_Rayl_Obs_String(L2B,r)

       if ((mrb1 .eq. -1) .or. (mrb2 .eq. -1)) then
          ! we have only Rayleigh data at this level
          print "(9('-'),': ',a9,': ',6('-'),': ',a6)",&
               txt_Rayl_Meas,txt_Rayl_Obs
       elseif (mrb1 .eq. mrb2) then
          txt_Mie_Meas = Constr_Mie_Meas_String(L2B,mrb1)
          txt_Mie_Obs  = Constr_Mie_Obs_String(L2B,mrb1)
          print "(a9,': ',a9,': ',a6,': ',a6)",&
               txt_Mie_Meas,txt_Rayl_Meas,txt_Mie_Obs,txt_Rayl_Obs
          last_mie_rangebin_printed = mrb1
       else      
          txt_Mie_Meas = Constr_Mie_Meas_String(L2B,mrb1)
          txt_Mie_Obs  = Constr_Mie_Obs_String(L2B,mrb1)
          print "(a9,': ',a9,': ',a6,': ',a6)",&
               txt_Mie_Meas,txt_Rayl_Meas,txt_Mie_Obs,txt_Rayl_Obs
          DO r_m=mrb1+1,mrb2
             txt_Mie_Meas = Constr_Mie_Meas_String(L2B,r_m)
             txt_Mie_Obs  = Constr_Mie_Obs_String(L2B,r_m)
             print "(a9,': ',9x,': ',a6,': ',6x)",&
                  txt_Mie_Meas,txt_Mie_Obs
          END DO
          last_mie_rangebin_printed = mrb2
       end if
    end do

    do r=last_mie_rangebin_printed+1,WorkingData%MieMeas(1)%nr_of_range_gates
       txt_Mie_Meas = Constr_Mie_Meas_String(L2B,r)
       txt_Mie_Obs  = Constr_Mie_Obs_String(L2B,r)
       print "(a9,': ',9('-'),': ',a6,': ',6('-'))",&
            txt_Mie_Meas,txt_Mie_Obs
    end do

  end subroutine Print_Constr_Obs_Matched
    !  #]
  !------------------------------------------
end module Construct_Obs
