module AMD_DataStructure
  !  #[ Description
  !     a module to define the datastructure 
  !     for auxiliary meteorological data
  !     as used by the L2A and L2B processors,
  !     and the routines needed for initialisation, allocation
  !     and deallocation of all data structs inside it.
  !     
  !     created:     31-Jan-2006, by: David Tan, ECMWF
  !     based on:    L2B_DataStructure by Jos de Kloe.
  !
  !     Modifications
  !      D. Tan     31-Jan-2006 Original version
  !      D. Tan     02-Feb-2006 Activate both off_nadir datasets.
  !      P. Poli    07-Feb-2006 Implemented writing and headers handling
  !      D. Tan     14-Feb-2006 Change default validity period
  !      J. de Kloe 20-Feb-2006 some small corrections to prevent memory leaks
  !      J. de Kloe 20-Feb-2006 adapted to changes in the XML_module
  !      J. de Kloe 22-Feb-2006 Change precision, r_ to r8_.
  !      J. de Kloe 06-Mar-2006 removed use of program_abort
  !      D. Tan     14-Mar-2006 Activate nadir datasets.
  !      D. Tan     12-May-2006 Null use of ParKind1 module
  !      P. Poli    26-Jul-2006 Added reset of AMD%*H_is_available in Delete
  !      P. Poli    17-Aug-2006 Removed obsolete variables
  !      P. Poli    18-Sep-2006 Call routine Fill_L2B_AMD_SPH_items
  !      J. de Kloe 23-Apr-2007 phase out fillwithzeros code
  !      J. de Kloe 24-Apr-2007 phase out SummaryVars
  !      J. de Kloe 30-Jul-2007 adapt to changed interface of Fill_FH_Items
  !      P. Poli    03-Sep-2007 changed L1B to L1B_Headers in Handle_AMDHeaders
  !      P. Poli    04-Sep-2007 changed L1B_DataType to L1B_HDR_DataType
  !      J. de Kloe 18-Oct-2007 adapt to changed interface of Fill_MPH_Items
  !      P. Poli    19-Oct-2007 - Split AMD_DataType in two parts:
  !                             AMD_HDR_DataType and AMD_BRC_DataType
  !                             - Suppressed nr_of_model_levels inside
  !                             AMD datastructure
  !                             - Split routine Init_AMD_DataStructure
  !                             into Init_AMD_HDR_DataStructure and
  !                             Init_AMD_BRC_DataStructure
  !                             - Split routine Delete_AMD_DataStructure
  !                             into Delete_AMD_HDR_DataStructure and
  !                             Delete_AMD_BRC_DataStructure
  !                             - Split routine Print_AMD_DataStructure
  !                             into Print_AMD_HDR_DataStructure and
  !                             Print_AMD_BRC_DataStructure
  !                             - Replaced GetByteSize_AMD_DataStructArray
  !                             by GetByteSize_AMD_BRC_DataStructArray
  !                             - Split GetByteSize_AMD_DataStructure
  !                             into GetByteSize_AMD_HDR_DataStructure and
  !                             GetByteSize_AMD_BRC_DataStructure
  !                             - Allocation of FH,MPH,SPH now done in 
  !                             Init_AMD_HDR_DataStructure, to ensure symmetry
  !                             with Delete_AMD_HDR_DataStructure.
  !      J. de Kloe 26-Oct-2007 fixed too long subroutine names
  !      J. de Kloe 14-Nov-2007 changed the timestep value slighthly
  !      J. de Kloe 06-Dec-2007 changed interface for Fill_FH and
  !                             Fill_MPH routines
  !      16-Jan-2008 J. de Kloe phase out integer kind i_
  !      17-Jan-2008 J. de Kloe removed a last few 64bit incompatibilities
  !      21-Jan-2008 J. de kloe added workaround in call to GetByteSize_MPH_KVT
  !                             to not change the fileformat for now.
  !      21-Feb-2008 J. de Kloe removed the workaround again, since it was
  !                             caused by a misunderstanding on L1B side
  !      25-Feb-2008 D. Tan     New Ref_Doc and Proc_Center
  !      25-Feb-2008 J. de Kloe merged in a comment
  !  #]
  !  #[ modules used

  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE FixedHeader, only: FH_type, Init_FH, &
                         Fill_FH_items, systemstr_LOCAL_MET_INST, &
                         file_type_AMD
  USE MainProductHeader, only: MPH_type, Init_MPH, Print_MPH, &
                               Fill_MPH_items, &
                               Fill_L2B_MPH_From_L1B_MPH, &
                               GetByteSize_MPH_KVT, maxlen_Ref_Doc, &
                               maxlen_Product, maxlen_Proc_Stage, &
                               maxlen_Proc_Center, maxlen_Software_Ver
  USE L2B_AMD_SPH, only: L2B_AMD_SPH_type, Init_L2B_AMD_SPH, &
                         Fill_L2B_AMD_SPH_items, &
                         Delete_L2B_AMD_SPH, Print_L2B_AMD_SPH, &
                         GetByteSize_AMD_SPH_KVT
  USE AMD_Met_MDS, only: AMD_Met_MDS_type, Init_AMD_Met_MDS, &
                         Delete_AMD_Met_MDS, Print_AMD_Met_MDS, &
                         GetByteSize_AMD_Met_MDS
  USE AMD_Geoloc_ADS, only: AMD_Geoloc_ADS_type, Init_AMD_Geoloc_ADS, &
                            Print_AMD_Geoloc_ADS, &
                            GetByteSize_AMD_Geoloc_ADS
  USE L1B_DataStructure, only: L1B_HDR_DataType

  ! support modules
  USE EE_CFI_Datatypes, only: DateTimeType,&
                              InitDateTime, SetDateTime, PrintDateTime, &
                              IntAus, IntAd, IntAl, GetIntAl, IntAs, &
                              IncrementDateTime
                              ! Construct_KVT_DateTime, maxlen_DateTime_KVT
  USE numerics, only: r8_, missing_indicator_real
  USE ErrorHandler, only: no_error, error_programming, error_allocate
  USE Compiler_Features, only: newline
  USE AeolusConstants, only: software_version_L2BP => software_version,&
                             software_ProcName, Ref_Doc_AuxMet

  !  #]
  !  #[ Parameter and data type definitions
  implicit none

! end of working arrays definition

  TYPE AMD_HDR_DataType
     !  #[
     !  the AMD headers
     logical :: FH_is_available 
     logical :: MPH_is_available 
     logical :: SPH_is_available 
     TYPE(FH_type),      pointer :: FH
     TYPE(MPH_type),     pointer :: MPH
     TYPE(L2B_AMD_SPH_type), pointer :: SPH
  END TYPE AMD_HDR_DataType
     !  #]

  TYPE AMD_BRC_DataType
     !  #[
     ! the 4 attached datasets
     TYPE(AMD_Geoloc_ADS_type),   pointer :: Geoloc_ADS_off_nadir
     TYPE(AMD_Geoloc_ADS_type),   pointer :: Geoloc_ADS_nadir
     TYPE(AMD_Met_MDS_type),      pointer :: Met_MDS_off_nadir
     TYPE(AMD_Met_MDS_type),      pointer :: Met_MDS_nadir
  END TYPE AMD_BRC_DataType
     !  #]

  !  #]
  !  #[ Module interface
  !     everything is public in this module
  !  #]
contains
  !---------------------------------------------
  subroutine Init_AMD_HDR_DataStructure(AMD_Headers, &
                                    error_flag)
    !  #[

    TYPE(AMD_HDR_DataType), intent(out) :: AMD_Headers ! output

    integer, intent(out) :: error_flag     ! output

    ! local variable
    integer :: AllocStatus

    ! init
    error_flag = no_error

    !------------------------------------------------
    ! init the header dataset structures
    !------------------------------------------------
    AMD_Headers%FH_is_available = .false.
    nullify(AMD_Headers%FH)
    AMD_Headers%MPH_is_available = .false.
    nullify(AMD_Headers%MPH)
    AMD_Headers%SPH_is_available = .false.
    nullify(AMD_Headers%SPH)

    !------------------------------
    ! allocate memory to hold the FH, MPH and SPH elements
    !------------------------------
    allocate(AMD_Headers%FH,AMD_Headers%MPH,AMD_Headers%SPH,STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation failure in subroutine Init_AMD_HDR_DataStructure"
       error_flag = error_allocate
       return
    END IF
    call Init_FH(AMD_Headers%FH,error_flag)
    IF (error_flag .ne. no_error) return
    call Init_MPH(AMD_Headers%MPH,error_flag)
    IF (error_flag .ne. no_error) return
    call Init_L2B_AMD_SPH(AMD_Headers%SPH,error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine Init_AMD_HDR_DataStructure
    !  #]
  subroutine Init_AMD_BRC_DataStructure(AMD, N_model_levels, &
                                    error_flag)
    !  #[

    TYPE(AMD_BRC_DataType), intent(out) :: AMD ! output

    integer, intent(in)  :: N_model_levels ! input
    integer, intent(out) :: error_flag     ! output

    ! local variable
    integer :: AllocStatus

    ! init
    error_flag = no_error

    !------------------------------------------------
    ! allocate and init the binary dataset structures
    !------------------------------------------------
    IF (N_model_levels .ge. 0) THEN
       allocate(AMD%Geoloc_ADS_off_nadir,stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_AMD_DataStructure"
          error_flag = error_allocate
          return
       END IF
       ! by setting N_model_levels=0, we indicate that no 
       ! memory is yet to be allocated. Allocation is only done after the
       ! SPH is read, because only then we know the nr of model levels
       ! in each profile
       call Init_AMD_Geoloc_ADS(AMD%Geoloc_ADS_off_nadir, &
       ! 0_IntAus, &
                                error_flag)
       IF (error_flag .ne. no_error) return
    ELSE
       nullify(AMD%Geoloc_ADS_off_nadir)
    END IF

    IF (N_model_levels .ge. 0) THEN
       allocate(AMD%Geoloc_ADS_nadir,stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_AMD_DataStructure"
          error_flag = error_allocate
          return
       END IF
       ! by setting N_model_levels=0, we indicate that no 
       ! memory is yet to be allocated. Allocation is only done after the
       ! SPH is read, because only then we know the nr of model levels
       ! in each profile
       call Init_AMD_Geoloc_ADS(AMD%Geoloc_ADS_nadir, &
       ! 0_IntAus, &
                                error_flag)
       IF (error_flag .ne. no_error) return
    ELSE
       nullify(AMD%Geoloc_ADS_nadir)
    END IF

    IF (N_model_levels .ge. 0) THEN
       allocate(AMD%Met_MDS_off_nadir,stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_AMD_DataStructure"
          error_flag = error_allocate
          return
       END IF
       ! by setting N_model_levels=0, we indicate that no 
       ! memory is yet to be allocated. Allocation is only done after the
       ! SPH is read, because only then we know the nr of model levels
       ! in each profile
       call Init_AMD_Met_MDS(AMD%Met_MDS_off_nadir, 0, error_flag)
       IF (error_flag .ne. no_error) return
    ELSE
       nullify(AMD%Met_MDS_off_nadir)
    END IF

    IF (N_model_levels .ge. 0) THEN
       allocate(AMD%Met_MDS_nadir,stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_AMD_DataStructure"
          error_flag = error_allocate
          return
       END IF
       ! by setting N_model_levels=0, we indicate that no 
       ! memory is yet to be allocated. Allocation is only done after the
       ! SPH is read, because only then we know the nr of model levels
       ! in each profile
       call Init_AMD_Met_MDS(AMD%Met_MDS_nadir, 0, error_flag)
       IF (error_flag .ne. no_error) return
    ELSE
       nullify(AMD%Met_MDS_nadir)
    END IF

  end subroutine Init_AMD_BRC_DataStructure
    !  #]
  subroutine Delete_AMD_HDR_DataStructure(AMD)
    !  #[

    TYPE(AMD_HDR_DataType), intent(inout) :: AMD ! input/output
    ! (contents will be deallocated after this routine)

    ! clean up the header datastructures
    IF (associated(AMD%SPH)) call Delete_L2B_AMD_SPH(AMD%SPH)

    IF (associated(AMD%FH)) deallocate(AMD%FH)
    IF (associated(AMD%MPH)) deallocate(AMD%MPH)
    IF (associated(AMD%SPH)) deallocate(AMD%SPH)

    AMD%FH_is_available = .false.
    AMD%MPH_is_available = .false.
    AMD%SPH_is_available = .false.

  end subroutine Delete_AMD_HDR_DataStructure
    !  #]
  subroutine Delete_AMD_BRC_DataStructure(AMD)
    !  #[

    TYPE(AMD_BRC_DataType), intent(inout) :: AMD ! input/output
    ! (contents will be deallocated after this routine)

    ! clean up the datasets
! TODO: PP: why are the four lines below commented out?
! JK: the AMD_Geoloc_ADS_type contains no allocated arrays or pointers
!     so there is no memory to be freed in this case.
!     Therefore these routines are obsolete.
!
!    IF (associated(AMD%Geoloc_ADS_off_nadir)) &
!         call Delete_AMD_Geoloc_ADS(AMD%Geoloc_ADS_off_nadir)
!    IF (associated(AMD%Geoloc_ADS_nadir)) &
!         call Delete_AMD_Geoloc_ADS(AMD%Geoloc_ADS_nadir)
    IF (associated(AMD%Met_MDS_off_nadir)) &
         call Delete_AMD_Met_MDS(AMD%Met_MDS_off_nadir)
    IF (associated(AMD%Met_MDS_nadir)) &
         call Delete_AMD_Met_MDS(AMD%Met_MDS_nadir)

    IF (associated(AMD%Geoloc_ADS_off_nadir))   deallocate(AMD%Geoloc_ADS_off_nadir)
    IF (associated(AMD%Geoloc_ADS_nadir    ))   deallocate(AMD%Geoloc_ADS_nadir)

    IF (associated(AMD%Met_MDS_off_nadir))      deallocate(AMD%Met_MDS_off_nadir)
    IF (associated(AMD%Met_MDS_nadir    ))      deallocate(AMD%Met_MDS_nadir)

  end subroutine Delete_AMD_BRC_DataStructure
    !  #]
  subroutine Print_AMD_HDR_DataStructure(AMD, error_flag)
    !  #[ ascii output to screen
    TYPE(AMD_HDR_DataType), intent(in)  :: AMD        ! input
    integer,                intent(out) :: error_flag ! output

    ! DT: 31 Jan 2006
    !     This routine never called?
    !     No equivalent L1B routine?
    !     Is the equivalent L2B routine ever called?

    error_flag = no_error
    
    print *,"=================================="
    print *,"contents of AMD HDR datastructure:"
    print *,"=================================="

    print *,"...%FH_is_available    = ",AMD%FH_is_available 
    print *,"...%MPH_is_available   = ",AMD%MPH_is_available 
    print *,"...%SPH_is_available   = ",AMD%SPH_is_available 

  end subroutine Print_AMD_HDR_DataStructure
    !  #]
  subroutine Print_AMD_BRC_DataStructure(AMD, error_flag)
    !  #[ ascii output to screen
    TYPE(AMD_BRC_DataType), intent(in)  :: AMD        ! input
    integer,                intent(out) :: error_flag ! output

    ! DT: 31 Jan 2006
    !     This routine never called?
    !     No equivalent L1B routine?
    !     Is the equivalent L2B routine ever called?

    error_flag = no_error
    
    print *,"=================================="
    print *,"contents of AMD BRC datastructure:"
    print *,"=================================="

    ! the 4 attached datasets

    print *,"...%Geoloc_ADS_off_nadir%..."
    call Print_AMD_Geoloc_ADS(AMD%Geoloc_ADS_off_nadir, error_flag)
    if (error_flag .ne. no_error) return

    print *,"...%Geoloc_ADS_nadir%..."
    call Print_AMD_Geoloc_ADS(AMD%Geoloc_ADS_nadir, error_flag)
    if (error_flag .ne. no_error) return

    print *,"...%Met_MDS_off_nadir%..."
    call Print_AMD_Met_MDS(AMD%Met_MDS_off_nadir, error_flag)
    if (error_flag .ne. no_error) return

    print *,"...%Met_MDS_nadir%..."
    call Print_AMD_Met_MDS(AMD%Met_MDS_nadir, error_flag)
    if (error_flag .ne. no_error) return

  end subroutine Print_AMD_BRC_DataStructure
    !  #]
          !1234567890123456789012345678901
  function GetByteSize_AMD_HDR_DataStruct(AMD) result(c_bytesize)
    !  #[
    TYPE(AMD_HDR_DataType), intent(in)  :: AMD        ! input
    integer(IntAl)                  :: c_bytesize ! result
    
    c_bytesize = 0
    if (associated(AMD%MPH)) then
       c_bytesize = c_bytesize + GetByteSize_MPH_KVT()
    endif
    if (associated(AMD%SPH)) then
       c_bytesize = c_bytesize + GetByteSize_AMD_SPH_KVT(AMD%SPH)
    end if

    ! typical size:
    ! fortran size = ... bytes
    ! c       size = ... bytes

  end function GetByteSize_AMD_HDR_DataStruct
    !  #]
  function GetByteSize_AMD_BRC_DataStruct(AMD) result(c_bytesize)
    !  #[
    TYPE(AMD_BRC_DataType), intent(in)  :: AMD        ! input
    integer(IntAl)                      :: c_bytesize ! result
    
    c_bytesize = 0
    if (associated(AMD%Geoloc_ADS_off_nadir)) then
       c_bytesize = c_bytesize + GetByteSize_AMD_Geoloc_ADS()
    end if
    if (associated(AMD%Geoloc_ADS_nadir)) then
       c_bytesize = c_bytesize + GetByteSize_AMD_Geoloc_ADS()
    end if
    if (associated(AMD%Met_MDS_off_nadir)) then
       c_bytesize = c_bytesize + &
            GetByteSize_AMD_Met_MDS(AMD%Met_MDS_off_nadir)
    end if
    if (associated(AMD%Met_MDS_nadir)) then
       c_bytesize = c_bytesize + &
            GetByteSize_AMD_Met_MDS(AMD%Met_MDS_nadir)
    end if

    ! typical size:
    ! fortran size = ... bytes
    ! c       size = ... bytes

  end function GetByteSize_AMD_BRC_DataStruct
    !  #]
  function GetByteSize_AMD_BRC_DataStrArr(Num_BRC,AMD) result(c_bytesize)
    !  #[
    integer,                             intent(in) :: Num_BRC ! input
    type(AMD_BRC_DataType), dimension(1:Num_BRC), intent(in) :: AMD     ! input
    integer(IntAl) :: c_bytesize ! result

    ! local variable
    integer :: iBRC
    
    ! the AMD datastructures do not all have the same, size. 
    ! To allow for further differences, and to keep it simple,
    ! use a loop to get the overal size
    c_bytesize = 0
    DO iBRC=1,Num_BRC
       c_bytesize = c_bytesize + GetByteSize_AMD_BRC_DataStruct(AMD(iBRC))
    END DO

    ! typical size:
    ! fortran size = ... bytes
    ! c       size = ... bytes

  end function GetByteSize_AMD_BRC_DataStrArr
    !  #]
  !---------------------------------------------
  subroutine Handle_AMDHeaders (AMD_Header, AMD, &
!             Num_BRC, &
             error_flag, &
!             POfilename, &
             Num_of_Model_Layers, &
             Num_Records_in_DS1, Num_Records_in_DS2, &
             Num_Avail_L1B_Obs, Num_Missing_L1B_Obs, &
             Num_Computed_Locations, &
             L1B_Headers, switch_init, switch_update)
  !  #[

  !  07-02-2005  P. Poli    Original code. This routine fills in the FH, MPH,
  !                         and SPH as the L2B processor makes it way through
  !                         the succesive BRCs.
  ! Note:
  ! (*) AMD_Header is the AMD datastructure which contains the Headers you want
  ! to modify.
  ! (*) AMD is the datastructure from which you want to extract the information
  ! to update the headers of AMD_Header.
  type(AMD_HDR_DataType),         intent(inout):: AMD_Header
  type(AMD_BRC_DataType),         intent(inout):: AMD
!  integer,               intent(in)   :: Num_BRC
  integer,                intent(out)  :: error_flag
!  character(len=*),  optional,intent(in)   :: POfilename
  integer(IntAus),            intent(in)   :: Num_of_Model_Layers
  integer(IntAl),             intent(in)   :: Num_Records_in_DS1, Num_Records_in_DS2, &
                                              Num_Avail_L1B_Obs, Num_Missing_L1B_Obs, &
                                              Num_Computed_Locations
  type(L1B_HDR_DataType),optional,intent(in):: L1B_Headers
  logical,           optional,intent(in)   :: switch_init
  logical,           optional,intent(in)   :: switch_update

  ! local variables
  TYPE(DateTimeType) :: datetime_0    ! struct(YYYY,MM,DD)
  TYPE(DateTimeType) :: DateTimeStart ! struct(YYYY,MM,DD)
  TYPE(DateTimeType) :: DateTimeStop  ! struct(YYYY,MM,DD)
  real(r8_)          :: timestep      ! in seconds
  logical            :: do_init, do_update

  ! remark: this one is set during do_init, and used during do_update
  !         so the save keyword is needed !!
!  integer, save  :: N_meas

  ! used to fill the FH
  character(len=256) :: file_description, notes, file_class, file_type
  character(len=256) :: creator, creator_version
  integer            :: file_version=1
  character(len=*), parameter :: system=systemstr_LOCAL_MET_INST

  ! used to fill the MPH
  character(len=maxlen_Product)      :: product
  character(len=maxlen_Proc_Stage)   :: proc_stage
  character(len=maxlen_Ref_Doc)      :: ref_doc
  character(len=maxlen_Proc_Center)  :: proc_center
  character(len=maxlen_Software_Ver) :: software_version
  logical            :: product_err
  integer(IntAd)     :: tot_size
  integer(IntAl)     :: sph_size
  integer(IntAl)     :: num_DSD
  integer(IntAl)     :: num_data_sets

  ! used to fill the SPH
  character(len=256) :: text
  integer(IntAl)     :: Num_Records_in_DS1_old, Num_Records_in_DS2_old, &
                        Num_Avail_L1B_Obs_old, Num_Missing_L1B_Obs_old, &
                        Num_Computed_Locations_old, &
                        Num_Records_in_DS1_update, Num_Records_in_DS2_update, &
                        Num_Avail_L1B_Obs_update, Num_Missing_L1B_Obs_update, &
                        Num_Computed_Locations_update
    
  ! init
  error_flag = no_error

  do_init=.false.
  IF (present(switch_init)) do_init=switch_init

  do_update=.false.
  IF (present(switch_update)) do_update=switch_update

  IF (do_init .and. do_update) THEN
     print *,'Handle_AMDHeaders: PROBLEM -- both switch_init and switch_update'
     print *,'Handle_AMDHeaders: are true.'
     error_flag = error_programming
     return
  ENDIF

  ! date used as example
  call SetDateTime(datetime_0,2004,12,01,00,00,00.0_r8_)

  !*** set the Fixed Header ***
  IF (do_init) THEN
     IF (.not.(associated(AMD_Header%FH))) THEN
        print *,"AMD_Header%FH seems not associated !!!"
        print *,"programming error ...."
        error_flag = error_programming
        return
     END IF

! if no L1B datastructure is available, set dummy date/date/timestep
!        ! TO CHANGE LATER.
     print *,'Handle_AMDHeaders: no L1B datastructure in memory ->'
     print *,'Handle_AMDHeaders: setting dummy date, time, and timestep in FH'
     ! Validity period from prototype ascii met data
     timestep = 1799.999_r8_ ! seconds
     ! NOTE: use a value just below 1800 to make sure the ceil() returns
     ! with 1800. If you give 1800.0 here, the ceil() used to get
     ! FH%Validity_Period%Validity_Stop might return with 1801
     ! due to machine inaccuracy ! (And this makes the result not portable!)
     DateTimeStart = datetime_0
     DateTimeStop  = DateTimeStart
     call IncrementDateTime(DateTimeStop,timestep)

! some test prints
!call PrintDateTime(datetime)
!DateTimeString = Construct_KVT_DateTime(datetime)
!print *,"datetime:KVT formatted="//trim(DateTimeString)


     file_description = "Aux Met Data Product"
     notes            = "Room for some"//newline//"additional remarks"
     file_class       = "TEST"
     file_type        = file_type_AMD
     creator          = software_ProcName
     creator_version  = software_version_L2BP
     call Fill_FH_items(AMD_Header%FH, &
          file_description, notes, file_class, file_type, &
          DateTimeStart, DateTimeStop, file_version, &
          system, creator, creator_version, error_flag)
     IF (error_flag .ne. no_error) return
     print *,'Handle_AMDHeaders: Filled FH'
  ENDIF ! do_init on FH

  !*** set the Main Product Header ***
  IF (do_init) THEN
     IF (.not.(associated(AMD_Header%MPH))) THEN
        print *,"AMD_Header%MPH seems not associated !!!"
        print *,"programming error ...."
        error_flag = error_programming
        return
     END IF
     IF (present(L1B_Headers)) THEN
        ! TODO!
        print *,"Handle_AMDHeaders: DID YOU MAKE SURE L1B HEADERS ARE DEFINED??"
        call Fill_L2B_MPH_From_L1B_MPH(L1B_Headers%MPH,AMD_Header%MPH,error_flag)
        IF (error_flag .ne. no_error) return
     ELSE
        print *,'Handle_AMDHeaders: no L1B datastructure in memory ->'
        print *,'Handle_AMDHeaders: information missing in AMD MPH.'
     ENDIF

     product       = AMD_Header%FH%File_Name
     proc_stage = "T"  ! to be moved into ?? L2B_Proc_Settings.F90 ??
     ref_doc    = Ref_Doc_AuxMet
     ! proc_stage might be: N (near-real-time), R (reprocessing) or T (test)
     proc_center = system(1:maxlen_Proc_Center)
                        ! the location where this file was produced
                        ! Copy (possibly truncated) of FixedHeader system
                        ! new requirement in spr 28
     software_version = software_version_L2BP
     product_err   = .false. ! assume no errors are reported
     tot_size      = 0_IntAd ! to be filled later
     sph_size      = 1778_IntAl ! GetByteSize_AMD_SPH_KVT() ! to be implemented
     num_DSD       = 4_IntAl ! a constant for the AMD file type
     num_data_sets = 4_IntAl ! a constant for the AMD file type

     ! Note that DateTimeStart, DateTimeStop have been assigned dummy test
     ! values in the do_init (FH) block above, so they should contain 
     ! some values

     ! proc_stage might be: N (near-real-time), R (reprocessing) or T (test)
     call Fill_MPH_items(AMD_Header%MPH, product, proc_stage, ref_doc, &
                         proc_center, software_version, product_err, tot_size,&
                         sph_size, num_DSD, num_data_sets, &
                         DateTimeStart, DateTimeStop,&
                         error_flag)
     IF (error_flag .ne. no_error) return
     print *,'Handle_AMDHeaders: Filled MPH'
  ENDIF ! do_init on MPH

  !*** set the Specific Product Header ***
  IF (do_init) THEN
     IF (.not.(associated(AMD_Header%SPH))) THEN
        print *,"AMD_Header%SPH seems not associated !!!"
        print *,"programming error ...."
        error_flag = error_programming
        return
     END IF

     call Fill_L2B_AMD_SPH_items(AMD_Header%SPH, &
             error_flag, &
             SPH_Descriptor="AEOLUS_AMD_SPECIFIC_HEADER", &
             Ref_NWP_Suite="CY31R1_hlosdir_reint", &
             Fcst_Initial_Time=DateTimeStart, &
             Model_Timestep=982_IntAl, Model_Grid_Type="GF", &
             Model_Resol_Par1=358_IntAs, Model_Resol_Par2=24_IntAs, &
             Num_of_Model_Layers=Num_of_Model_Layers, &
             Num_Records_in_DS1=Num_Records_in_DS1, &
             Num_Records_in_DS2=Num_Records_in_DS2, &
             Num_Avail_L1B_Obs=Num_Avail_L1B_Obs, &
             Num_Missing_L1B_Obs=Num_Missing_L1B_Obs, &
             Num_Computed_Locations=Num_Computed_Locations)
     IF (error_flag .ne. no_error) return
     print *,"Handle_AMDHeaders: Filled SPH"
  ENDIF ! do_init on SPH

  IF (do_update) THEN
        text(:)=' '
        text=AMD_Header%SPH%Num_Records_in_DS1
        call GetIntAl(text,Num_Records_in_DS1_old,error_flag)
        Num_Records_in_DS1_update=Num_Records_in_DS1_old + &
                                & Num_Records_in_DS1
                                
        text(:)=' '
        text=AMD_Header%SPH%Num_Records_in_DS2
        call GetIntAl(text,Num_Records_in_DS2_old,error_flag)
        Num_Records_in_DS2_update=Num_Records_in_DS2_old + &
                                & Num_Records_in_DS2
                                
        text(:)=' '
        text=AMD_Header%SPH%Num_Avail_L1B_Obs
        call GetIntAl(text,Num_Avail_L1B_Obs_old,error_flag)
        Num_Avail_L1B_Obs_update=Num_Avail_L1B_Obs_old + &
                               & Num_Avail_L1B_Obs
                               
        text(:)=' '
        text=AMD_Header%SPH%Num_Missing_L1B_Obs
        call GetIntAl(text,Num_Missing_L1B_Obs_old,error_flag)
        Num_Missing_L1B_Obs_update=Num_Missing_L1B_Obs_old + &
                               & Num_Missing_L1B_Obs

        text(:)=' '
        text=AMD_Header%SPH%Num_Computed_Locations
        call GetIntAl(text,Num_Computed_Locations_old,error_flag)
        Num_Computed_Locations_update=Num_Computed_Locations_old + &
                               & Num_Computed_Locations
        
        call Fill_L2B_AMD_SPH_items(AMD_Header%SPH, &
             error_flag, &
             Num_Records_in_DS1=Num_Records_in_DS1_update, &
             Num_Records_in_DS2=Num_Records_in_DS2_update, &
             Num_Avail_L1B_Obs=Num_Avail_L1B_Obs_update, &
             Num_Missing_L1B_Obs=Num_Missing_L1B_Obs_update, &
             Num_Computed_Locations=Num_Computed_Locations_update)
     IF (error_flag .ne. no_error) return
     print *,"Handle_AMDHeaders: Updated SPH"
  ENDIF ! do_update on SPH

  end subroutine Handle_AMDHeaders
  !  #]
  !---------------------------------------------

end module AMD_DataStructure
