module AMD_Geoloc_ADS
  !  #[ Documentation
  !
  !     a module to handle the AMD Geolocation dataset
  !     as defined in AE-IF_ECMWF-L2BP-001 (L2B IODD),
  !     sec.5.1.3, table 62.
  !
  !     Written by David Tan, ECMWF
  !     based on L2B_Rayleigh_MDS by Jos de Kloe, KNMI
  !
  !     created:     02-Feb-2006
  !
  !     Modifications:
  !      02-Feb-2006  D. Tan     Original version
  !      09-Feb-2006  P. Poli    Added writing and reading routines
  !      14-Mar-2006  D. Tan     Enable request of Which_dsr.
  !      14-Mar-2006  D. Tan     Correct typos in two Print statements.
  !      24-Mar-2006  D. Tan     Remove Num_of_Model_Layers.
  !      09-May-2006  D. Tan     Remove offset bug compensation
  !      12-May-2006  D. Tan     Null use of ParKind1 module
  !      31-Jul-2006  J. de Kloe renamed set_file_pointer to SetFilePointer
  !      17-Aug-2006  P. Poli    removed obsolete variables
  !      23-Apr-2007  J. de Kloe phase out fillwithzero code
  !      24-Apr-2007  J. de Kloe phase out the SummaryVars code
  !      09-Jan-2008  J. de kloe added DateTimeMissing checks in print routine
  !      16-Jan-2008  J. de Kloe phase out integer kind i_
  !
  !  #]
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE DataSetDescriptor, only: DSD_type
  USE EE_CFI_Datatypes, only:            c_bytelen_Bool,   &
       IntAus, maxlen_IntAus, SetIntAus, c_bytelen_IntAus, & 
       write_IntAus_to_bin_file, read_IntAus_from_bin_file,  &
       IntAl,  maxlen_IntAl,  SetIntAl, minval_IntAl, c_bytelen_IntAl,  &
       write_IntAl_to_bin_file,  read_IntAl_from_bin_file,  &
       FAdoxy,                           c_bytelen_FAdoxy, &
       DateTimeType, InitDateTime, maxlen_DateTime_XML, &
       c_bytelen_DateTime, Construct_XML_DateTime, Construct_KVT_DateTime, &
       write_DateTime_to_bin_file, read_DateTime_from_bin_file, &
       maxlen_DateTime_KVT,DateTimeMissing
  USE DirectBinaryIO, only: SetFilePointer

  USE Numerics, only: i8_, &
                      missing_indicator_integer, &
                      missing_indicator_real
  USE ErrorHandler, only: no_error, error_programming, error_allocate, error_reading_file
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting

  !  #]
  !  #[ type definitions

  implicit none

  TYPE AMD_Geoloc_ADS_type 
     !  #[
     ! see sec. 6.1.2, table 60, of TN2.1, vers. 2.0, 30-Jan-2006
     !
     ! With some re-ordering pending discussion with Paul Poli.
     ! PP: Re-ordering validated after discussion with DT. 09-02-2006
     type(DateTimeType)  :: DateTime    ! UTC

     integer(IntAl)  :: Latitude         ! [10^-6 degN]
     integer(IntAl)  :: Longitude        ! [10^-6 degE]
     integer(IntAl)  :: geoid_height     ! [cm]

  END TYPE AMD_Geoloc_ADS_type
     !  #]

  ! the default verbosity setting is Err (as silent as possible)
  integer, private, save :: LocalVerbosity = VerbosityError

  !  #]
contains
  !---------------------------------------------------------------------
   subroutine Init_AMD_Geoloc_ADS_module(ChoosenVerbosity)
    !  #[
    integer, intent(in), optional :: ChoosenVerbosity

    IF (present(ChoosenVerbosity)) LocalVerbosity = ChoosenVerbosity

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"AMD_Geoloc_ADS module has been initialised"
       call print_verbosity_setting(LocalVerbosity,ModuleName="AMD_Geoloc_ADS")
    END IF

  end subroutine Init_AMD_Geoloc_ADS_module
    !  #]
  !---------------------------------------------------------------------
 subroutine Init_AMD_Geoloc_ADS(AMD_Geoloc_ADS, &
 ! N_Model_Layers, &
                                error_flag)
    !  #[
    TYPE(AMD_Geoloc_ADS_type), pointer         :: AMD_Geoloc_ADS    ! input/output
!    integer(IntAus),        intent(in)  :: N_Model_Layers ! input
    integer,                intent(out) :: error_flag     ! output

    ! no local variable

    error_flag = no_error

    IF (.not. associated(AMD_Geoloc_ADS)) THEN
       print *,"ERROR in Init_AMD_Geoloc_ADS: "
       print *,"pointer AMD_Geoloc_ADS seems not associated !!!"
       error_flag = error_programming
       return
    END IF
    
    call InitDateTime(AMD_Geoloc_ADS%DateTime)

    AMD_Geoloc_ADS%latitude     = missing_indicator_integer
    AMD_Geoloc_ADS%longitude    = missing_indicator_integer

    AMD_Geoloc_ADS%geoid_height = missing_indicator_integer

  end subroutine Init_AMD_Geoloc_ADS
    !  #]
!  subroutine Delete_AMD_Geoloc_ADS(AMD_Geoloc_ADS)
    !  #[
!    TYPE(AMD_Geoloc_ADS_type), pointer :: AMD_Geoloc_ADS
!
!    IF (.not. associated(AMD_Geoloc_ADS)) return

! DT: should this routine have more code, e.g.
!    deallocate(AMD_Geoloc_ADS)
! JK: no. If there is nothing allocated in this datastructure
!     then the delete routine has nothing to do, and need not even be defined

!  end subroutine Delete_AMD_Geoloc_ADS
    !  #]
  function GetByteSize_AMD_Geoloc_ADS() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result
    
    c_bytesize = &
         1*c_bytelen_DateTime + &
         3*c_bytelen_IntAl

    ! typical size:
    ! fortran size = 16 + 3*4 = 28 bytes
    ! c       size = 12 + 3*4 = 24 bytes

  end function GetByteSize_AMD_Geoloc_ADS
    !  #]
  !---------------------------------------------------------------------
  subroutine Print_AMD_Geoloc_ADS(Geoloc_ADS,error_flag) !,PrintAll)
    !  #[

    TYPE(AMD_Geoloc_ADS_type), pointer     :: Geoloc_ADS
    integer,                   intent(out) :: error_flag       ! output
!    logical, optional,           intent(in)  :: PrintAll

    ! local variables
!    integer       :: p, s, step
!    logical       :: do_PrintAll

!    character(len=13) :: txt

    character(len=maxlen_IntAl) :: str_IntAl
    character(len=maxlen_DateTime_KVT) :: DateTimeString

    error_flag = no_error

!    do_PrintAll = .false.
!    IF (present(PrintAll)) do_PrintAll = PrintAll

    IF (associated(Geoloc_ADS)) THEN
       print *,"Geoloc_ADS%..."
    ELSE
       print *,"Geoloc_ADS [seems not associated]"
       return
    END IF

    ! print the date and time
    if (DateTimeMissing(Geoloc_ADS%DateTime)) then
       DateTimeString = "[missing]"
    else
       DateTimeString = Construct_KVT_DateTime(Geoloc_ADS%DateTime)
    end if
    print *,"...%DateTime: "//trim(DateTimeString)

    ! latitude handling
    IF (Geoloc_ADS%Latitude .eq. minval_IntAl) THEN
       print *,"...%Latitude            = [missing]"
    ELSE
       call SetIntAl(Geoloc_ADS%Latitude,str_IntAl,error_flag)
       if (error_flag .ne. no_error) return
       print *,"...%Latitude            = ",str_IntAl," [10^-6 degN]"
    END IF

    ! longitude handling
    IF (Geoloc_ADS%Longitude .eq. minval_IntAl) THEN
       print *,"...%Longitude           = [missing]"
    ELSE
       call SetIntAl(Geoloc_ADS%Longitude,str_IntAl,error_flag)
       if (error_flag .ne. no_error) return
       print *,"...%Longitude           = ",str_IntAl," [10^-6 degE]"
    END IF

    ! geiod height handling
    IF (Geoloc_ADS%geoid_height .eq. minval_IntAl) THEN
       print *,"...%geoid_height        = [missing]"
    ELSE
       call SetIntAl(Geoloc_ADS%geoid_height,str_IntAl,error_flag)
       if (error_flag .ne. no_error) return
       print *,"...%geoid_height        = ",str_IntAl," [cm]"
    END IF

  end subroutine Print_AMD_Geoloc_ADS
    !  #]
  subroutine Write_AMD_Geoloc_ADS(fileunit,Geoloc_ADS,error_flag)
    !  #[
    !  P. Poli   09-02-2006   Original code

    integer,                   intent(in)  :: fileunit    ! input
    TYPE(AMD_Geoloc_ADS_type), pointer     :: Geoloc_ADS  ! input
    integer,                   intent(out) :: error_flag  ! output

    error_flag = no_error

    IF (.not. associated(Geoloc_ADS)) THEN
       print *,"WARNING in Write_AMD_Geoloc_ADS():"
       print *,"The pointer to this Geoloc_ADS seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! AMD_DateTime
    call write_DateTime_to_bin_file(fileunit, &
         Geoloc_ADS%DateTime, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! AMD_Latitude
    call write_IntAl_to_bin_file(fileunit, &
         Geoloc_ADS%Latitude,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! AMD_Longitude
    call write_IntAl_to_bin_file(fileunit, &
         Geoloc_ADS%Longitude,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! AMD_zg
    call write_IntAl_to_bin_file(fileunit, &
         Geoloc_ADS%geoid_height,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
  
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"Write_AMD_Geoloc_ADS: done with Write_AMD_Geoloc_ADS"

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Write_AMD_Geoloc_ADS"
    return

  end subroutine Write_AMD_Geoloc_ADS
    !  #]
  subroutine Read_AMD_Geoloc_ADS(Geoloc_DSD, fileunit, Which_DSR, Geoloc_ADS, &
                                 error_flag)
    !  #[

    !  P. Poli  09-02-2006   Original code

    TYPE(DSD_type),            pointer     :: Geoloc_DSD     ! input
    integer,                   intent(in)  :: fileunit       ! input
    integer,                   intent(in)  :: which_DSR      ! input
    TYPE(AMD_Geoloc_ADS_type), pointer     :: Geoloc_ADS
    integer,                   intent(out) :: error_flag

    ! local variables
    integer(i8_)      :: pos, endpos
    integer           :: currentpos
    character(len=maxlen_DateTime_KVT) :: DateTimeString

    !------------------------------
    ! reading the Geoloc ADS elements
    !------------------------------

    error_flag = no_error
    IF (.not. associated(Geoloc_ADS)) THEN
       print *,"ERROR in Read_AMD_Geoloc_ADS():"
       print *,"The pointer to this Geoloc_ADS seems not associated !!"
       error_flag = error_programming
       return
    END IF

    pos = Geoloc_DSD%start_of_ds + (which_DSR-1)*Geoloc_DSD%size_of_one_dsr
    call SetFilePointer(fileunit,pos,error_flag)
    IF (error_flag .ne. no_error) return
    
    print *,"trying to read the binary Geoloc_ADS nr. ",which_DSR
    
    ! AMD_DateTime
    call read_DateTime_from_bin_file(fileunit,&
         Geoloc_ADS%DateTime,error_flag)
    IF (error_flag .ne. no_error) goto 998
    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"Read_AMD_Geoloc_ADS: AMD_DateTime is: "
       if (DateTimeMissing(Geoloc_ADS%DateTime)) then
          DateTimeString = "[missing]"          
       else
          DateTimeString = Construct_KVT_DateTime(Geoloc_ADS%DateTime)
       end if
       print *,DateTimeString
    END IF
    
    ! AMD_Latitude
    call read_IntAl_from_bin_file(fileunit, &
         Geoloc_ADS%Latitude,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! AMD_Longitude
    call read_IntAl_from_bin_file(fileunit, &
         Geoloc_ADS%Longitude,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! AMD_zg
    call read_IntAl_from_bin_file(fileunit, &
         Geoloc_ADS%geoid_height,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! calculate expected end-position after reading this block of data
    endpos = Geoloc_DSD%start_of_ds + which_DSR*Geoloc_DSD%size_of_one_dsr
    ! request the current file position
    inquire(unit=fileunit,nextrec=currentpos)
    
    IF (currentpos .ne. endpos) THEN
       print *,"ERROR in Read_AMD_Geoloc_ADS():"
       print *,"there seems to be a problem while reading this binary dataset"
       print *,"expected file position after this DSR is: ",endpos
       print *,"but the actual position seems to be:      ",currentpos
       print *,"Maybe the fileformat definition has changed ???"
       error_flag = error_reading_file
       return
    END IF
    
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"Read_AMD_Geoloc_ADS: done"
    
    !------------------------------
    ! done
    !------------------------------
    return
    
998 print *,"ABORTING subroutine Read_AMD_Geoloc_ADS"
    return

  end subroutine Read_AMD_Geoloc_ADS

    !  #]
  !---------------------------------------------------------------------

end module AMD_Geoloc_ADS
