module AMD_Met_MDS
  !  #[ Documentation
  !
  !     a module to handle the AMD Meteorological dataset
  !     as defined in AE-IF-ECMWF-L2BP-001 (L2B IODD)
  !     sec.5.1.3.2, tables 63-64
  !
  !     Written by David Tan, ECMWF
  !     based on L2B_Rayleigh_MDS by Jos de Kloe, KNMI
  !
  !     created:     31-Jan-2006
  !
  !     Modifications:
  !      31-Jan-2006  D. Tan     Original version
  !      02-Feb-2006  D. Tan     Upgrade to TN2.1, V2.0 (30 Jan 2006)
  !      09-Feb-2006  P. Poli    Added writing and reading routines
  !      14-Mar-2006  D. Tan     Enable request of Which_dsr.
  !      24-Mar-2006  D. Tan     Remove Num_of_Model_Layers
  !      09-May-2006  D. Tan     Remove offset bug compensation
  !      12-May-2006  D. Tan     Null use of ParKind1 module
  !      31-Jul-2006  J. de Kloe renamed set_file_pointer to SetFilePointer
  !      23-Apr-2007  J. de Kloe phase out fillwithzeros code
  !      24-Apr-2007  J. de Kloe Annotated GetByteSize routines
  !      24-Apr-2007  J. de Kloe phase out the SummaryVars code
  !      16-Jan-2008  J. de Kloe phase out integer kind i_
  !      17-Jan-2008  J. de Kloe removed a last few 64bit incompatibilities
  !      22-Feb-2008  J. de Kloe allow met_ads to be empty now
  !
  !  #]
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE DataSetDescriptor, only: DSD_type
  USE EE_CFI_Datatypes, only:            c_bytelen_Bool,   &
       IntAuc, maxlen_IntAuc, SetIntAuc, c_bytelen_IntAuc, &
       write_IntAuc_to_bin_file,  read_IntAuc_from_bin_file,  &
       IntAc,                            c_bytelen_IntAc,  &
       write_IntAc_to_bin_file,  read_IntAc_from_bin_file,  &
       IntAs,  maxlen_IntAs,  SetIntAs,  c_bytelen_IntAs,  &
       write_IntAs_to_bin_file,  read_IntAs_from_bin_file,  &
       IntAus, maxlen_IntAus, SetIntAus, c_bytelen_IntAus, & 
       write_IntAus_to_bin_file, read_IntAus_from_bin_file,  &
       IntAul, maxlen_IntAul, SetIntAul, c_bytelen_IntAul, &
       write_IntAul_to_bin_file,  read_IntAul_from_bin_file,  &
       IntAl,  maxlen_IntAl,  SetIntAl,  c_bytelen_IntAl,  &
       write_IntAl_to_bin_file,  read_IntAl_from_bin_file,  &
       FAdoxy,                           c_bytelen_FAdoxy, &
       write_FAdoxy_to_bin_file, read_FAdoxy_from_bin_file,  &
       DateTimeType, InitDateTime, maxlen_DateTime_XML, &
       c_bytelen_DateTime, Construct_XML_DateTime, Construct_KVT_DateTime, &
       write_DateTime_to_bin_file, read_DateTime_from_bin_file, &
       maxlen_DateTime_KVT, &
       maxval_IntAuc, maxval_IntAs, maxval_IntAus, &
       maxval_IntAl,  maxval_IntAul, &
       write_SpareBytes_to_bin_file, &
       read_SpareBytes_from_bin_file

  USE DirectBinaryIO, only: SetFilePointer

  USE Numerics, only: i8_, real2txt, &
                      missing_indicator_integer, &
                      missing_indicator_real
  USE ErrorHandler, only: no_error, error_programming, error_allocate,error_reading_file
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting

  !  #]
  !  #[ type definitions

  implicit none

  TYPE AMD_Met_Layer_type 
     !  #[
     ! see sec. 6.1.5, table 31, of TN2.1, vers. 1.2, 06-09-2005
     !
     integer(IntAc)  :: Validity_Flag
     integer(IntAul) :: pbase     ! pressure at base of layer in Pa
     integer(IntAul) :: ptop      ! pressure at top  of layer in Pa
     integer(IntAul) :: pnom      ! nominal pressure of layer in Pa
     integer(IntAl)  :: zbase     ! height at base of layer in cm
     integer(IntAl)  :: ztop      ! height at top  of layer in cm
     integer(IntAl)  :: znom      ! nominal height of layer in cm
     integer(IntAus) :: T         ! temperature at pnom in 0.01K
     integer(IntAus) :: err_T     ! Std error for T in 0.01K
     integer(IntAs)  :: u         ! east-west wind in cm/s
     integer(IntAs)  :: v         ! north-south wind in cm/s
     !
     ! spare_1 (4 bytes)
     !
     Integer(IntAuc) :: rh        ! Relative humidity at pnom
     real(FAdoxy)    :: err_rh    ! Std error for rh

     real(FAdoxy)    :: q         ! Specific humidity at pnom
     Integer(IntAuc) :: cc        ! Cloud cover
     real(FAdoxy)    :: clwc      ! Cloud liquid water content
     real(FAdoxy)    :: ciwc      ! Cloud ice    water content

     !                  1234567890123456789012345678901
  END TYPE AMD_Met_Layer_type
     !  #]

  TYPE AMD_Met_MDS_type 
     !  #[
     ! see sec. 6.1.5, table 30, of TN2.1, vers. 1.2, 06-09-2005
!     type(DateTimeType)  :: Start_of_Obs_DateTime    ! UTC
     !
     ! spare_1 (2 bytes)
     !
     integer(IntAs)  :: us     ! east-west   "surface" wind in cm/s
     integer(IntAs)  :: vs     ! north-south "surface" wind in cm/s
     integer(IntAul) :: ps     ! surface pressure in Pa
     real(FAdoxy)    :: err_ps ! error for surface pressure in Pa
     integer(IntAl)  :: zs     ! surface geopotential height in cm
     !
     ! spare_2 (2 bytes)
     !
     ! an array for Num_of_Model_Layers substructures for this
     ! meteorological profile
     TYPE(AMD_Met_Layer_type), dimension(:), &
          pointer :: Met_Layer
     !              1234567890123456789012345678901
     !
     ! spare_3 (2 bytes)
     !
  END TYPE AMD_Met_MDS_type
     !  #]

  ! the default verbosity setting is Err (as silent as possible)
  integer, private, save :: LocalVerbosity = VerbosityError

  ! this is a temporary solution.
  ! David, Paul if you agree with defining the missing indicators like
  ! this we should add these definitions to the EE_CFI_Datatypes module.
  integer(IntAuc), parameter :: missing_indicator_IntAuc = maxval_IntAuc
  integer(IntAs),  parameter :: missing_indicator_IntAs  = maxval_IntAs
  integer(IntAus), parameter :: missing_indicator_IntAus = maxval_IntAus
  integer(IntAul), parameter :: missing_indicator_IntAul = maxval_IntAul
  integer(IntAl),  parameter :: missing_indicator_IntAl  = maxval_IntAl
  real(FAdoxy),    parameter :: missing_indicator_FAdoxy = &
                                missing_indicator_real

  !  #]
contains
  !---------------------------------------------------------------------
  subroutine Init_AMD_Met_MDS_module(ChoosenVerbosity)
    !  #[
    integer, intent(in), optional :: ChoosenVerbosity

    IF (present(ChoosenVerbosity)) LocalVerbosity = ChoosenVerbosity

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"AMD_Met_MDS module has been initialised"
       call print_verbosity_setting(LocalVerbosity,ModuleName="AMD_Met_MDS")
    END IF

  end subroutine Init_AMD_Met_MDS_module
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_AMD_Met_Layer(Met_Layer, error_flag)
    !  #[
    TYPE(AMD_Met_Layer_type), intent(out) :: Met_Layer ! output
    integer,    intent(out) :: error_flag    ! output

    error_flag = no_error

    Met_Layer%Validity_Flag = 0_IntAc ! .false.

    Met_Layer%pbase  = missing_indicator_IntAul
    Met_Layer%ptop   = missing_indicator_IntAul
    Met_Layer%pnom   = missing_indicator_IntAul

    Met_Layer%zbase  = missing_indicator_IntAl
    Met_Layer%ztop   = missing_indicator_IntAl
    Met_Layer%znom   = missing_indicator_IntAl

    Met_Layer%T      = missing_indicator_IntAus
    Met_Layer%err_T  = missing_indicator_IntAus
    Met_Layer%u      = missing_indicator_IntAs
    Met_Layer%v      = missing_indicator_IntAs
    Met_Layer%rh     = missing_indicator_IntAuc
    Met_Layer%err_rh = missing_indicator_FAdoxy

    Met_Layer%q      = missing_indicator_FAdoxy
    Met_Layer%cc     = missing_indicator_IntAuc
    Met_Layer%clwc   = missing_indicator_FAdoxy
    Met_Layer%ciwc   = missing_indicator_FAdoxy

  end subroutine Init_AMD_Met_Layer
    !  #]
  subroutine Print_Met_Layer(Met_Layer, error_flag)
    !  #[
    TYPE(AMD_Met_Layer_type), intent(in)  :: Met_Layer
    integer,                  intent(out) :: error_flag    ! output

    ! local variables

    character(len=13) :: txt

    character(len=maxlen_IntAs)  :: str_IntAs
    character(len=maxlen_IntAus) :: str_IntAus
    character(len=maxlen_IntAl)  :: str_IntAl
    character(len=maxlen_IntAul) :: str_IntAul
    character(len=maxlen_IntAuc) :: str_IntAuc

    error_flag = no_error

    if (Met_Layer%Validity_Flag .eq. 1_IntAc) then
       print *,"    ...%Validity_Flag  = Valid"
    else
       print *,"    ...%Validity_Flag  = Not Valid"
    endif

    call SetIntAul(Met_Layer%pbase,&
                  str_IntAul,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%pbase          = ",str_IntAul

    call SetIntAul(Met_Layer%ptop,&
                  str_IntAul,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%ptop           = ",str_IntAul

    call SetIntAul(Met_Layer%pnom,&
                  str_IntAul,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%pnom           = ",str_IntAul

    call SetIntAl(Met_Layer%zbase,&
                  str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%zbase          = ",str_IntAl

    call SetIntAl(Met_Layer%ztop,&
                  str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%ztop           = ",str_IntAl

    call SetIntAl(Met_Layer%znom,&
                  str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%znom           = ",str_IntAl

    call SetIntAus(Met_Layer%T,&
                  str_IntAus,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%T              = ",str_IntAus

    call SetIntAus(Met_Layer%err_T,&
                  str_IntAus,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%err_T          = ",str_IntAus

    call SetIntAs(Met_Layer%u,&
                  str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%u              = ",str_IntAs

    call SetIntAs(Met_Layer%v,&
                  str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%v              = ",str_IntAs

    call SetIntAuc(Met_Layer%rh,&
                   str_IntAuc,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%rh             = ",str_IntAuc

    print *,"REALACC(6)"
    txt = real2txt(Met_Layer%err_rh)
    print *,"    ...%err_rh         = ",txt

    txt = real2txt(Met_Layer%q)
    print *,"    ...%q              = ",txt

    call SetIntAuc(Met_Layer%cc,&
                  str_IntAuc,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%cc             = ",str_IntAuc

    txt = real2txt(Met_Layer%clwc)
    print *,"    ...%clwc           = ",txt

    txt = real2txt(Met_Layer%ciwc)
    print *,"    ...%ciwc           = ",txt
    print *,"ENDREALACC"

  end subroutine Print_Met_Layer
    !  #]
  function GetByteSize_Met_Layer() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result
    
    c_bytesize = &
         1*c_bytelen_IntAc     + &
         3*c_bytelen_IntAul    + &
         3*c_bytelen_IntAl     + &
         2*c_bytelen_IntAus    + &
         2*c_bytelen_IntAs     + &
         4                     + & ! 4 spare bytes
         1*c_bytelen_IntAuc    + &
         2*c_bytelen_FAdoxy    + &
         1*c_bytelen_IntAuc    + &
         2*c_bytelen_FAdoxy

    ! typical size:
    ! fortran size = 1+3*8+3*4+2*4+2*2+4+2+2*8+2+2*8 = 99 bytes
    ! c       size = 1+3*4+3*4+2*2+2*2+4+1+2*8+1+2*8 = 71 bytes

  end function GetByteSize_Met_Layer
    !  #]
  subroutine Write_Met_Layer(fileunit,Met_Layer,error_flag)
    !  #[
    !  P. Poli   09-02-2006   Original code

    integer,                 intent(in)   :: fileunit    ! input
    TYPE(AMD_Met_Layer_type),intent(inout):: Met_Layer   ! input
    integer,                 intent(out)  :: error_flag  ! output
    
    ! local variable
    integer :: nr_of_zero_bytes

    error_flag = no_error

    ! AMD_Validity_Flag
    call write_IntAc_to_bin_file(fileunit, &
            Met_Layer%Validity_Flag, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_pbase
    call write_IntAul_to_bin_file(fileunit, &
            Met_Layer%pbase, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_ptop
    call write_IntAul_to_bin_file(fileunit, &
            Met_Layer%ptop, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_pnom
    call write_IntAul_to_bin_file(fileunit, &
            Met_Layer%pnom, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_zbase
    call write_IntAl_to_bin_file(fileunit, &
            Met_Layer%zbase, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_ztop
    call write_IntAl_to_bin_file(fileunit, &
            Met_Layer%ztop, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_znom
    call write_IntAl_to_bin_file(fileunit, &
            Met_Layer%znom, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_T
    call write_IntAus_to_bin_file(fileunit, &
            Met_Layer%T, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_err_T
    call write_IntAus_to_bin_file(fileunit, &
            Met_Layer%err_T, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_u
    call write_IntAs_to_bin_file(fileunit, &
            Met_Layer%u, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_v
    call write_IntAs_to_bin_file(fileunit, &
            Met_Layer%v, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 4 spare bytes
    nr_of_zero_bytes = 4
    call write_SpareBytes_to_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! AMD_rh
    call write_IntAuc_to_bin_file(fileunit, &
            Met_Layer%rh, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_err_rh
    call write_FAdoxy_to_bin_file(fileunit, &
            Met_Layer%err_rh, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_q
    call write_FAdoxy_to_bin_file(fileunit, &
            Met_Layer%q, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_cc
    call write_IntAuc_to_bin_file(fileunit, &
            Met_Layer%cc, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_clwc
    call write_FAdoxy_to_bin_file(fileunit, &
            Met_Layer%clwc, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_ciwc
    call write_FAdoxy_to_bin_file(fileunit, &
            Met_Layer%ciwc, error_flag)
    IF (error_flag .ne. no_error) goto 998

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"Write_Met_Layer: done with Write_Met_Layer"

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Write_Met_Layer"
    return

  end subroutine Write_Met_Layer
    !  #]
  subroutine Read_Met_Layer(fileunit,Met_Layer,error_flag)
    !  #[
    !  P. Poli   09-02-2006   Original code

    integer,                 intent(in)    :: fileunit    ! input
    TYPE(AMD_Met_Layer_type),intent(inout) :: Met_Layer   ! input
    integer,                 intent(out)   :: error_flag  ! output
    
    ! local variable
    integer :: nr_of_zero_bytes

    error_flag = no_error

    ! AMD_Validity_Flag
    call read_IntAc_from_bin_file(fileunit, &
            Met_Layer%Validity_Flag, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_pbase
    call read_IntAul_from_bin_file(fileunit, &
            Met_Layer%pbase, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_ptop
    call read_IntAul_from_bin_file(fileunit, &
            Met_Layer%ptop, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_pnom
    call read_IntAul_from_bin_file(fileunit, &
            Met_Layer%pnom, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_zbase
    call read_IntAl_from_bin_file(fileunit, &
            Met_Layer%zbase, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_ztop
    call read_IntAl_from_bin_file(fileunit, &
            Met_Layer%ztop, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_znom
    call read_IntAl_from_bin_file(fileunit, &
            Met_Layer%znom, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_T
    call read_IntAus_from_bin_file(fileunit, &
            Met_Layer%T, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_err_T
    call read_IntAus_from_bin_file(fileunit, &
            Met_Layer%err_T, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_u
    call read_IntAs_from_bin_file(fileunit, &
            Met_Layer%u, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_v
    call read_IntAs_from_bin_file(fileunit, &
            Met_Layer%v, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 4 spare bytes
    nr_of_zero_bytes = 4
    call read_SpareBytes_from_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! AMD_rh
    call read_IntAuc_from_bin_file(fileunit, &
            Met_Layer%rh, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_err_rh
    call read_FAdoxy_from_bin_file(fileunit, &
            Met_Layer%err_rh, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_q
    call read_FAdoxy_from_bin_file(fileunit, &
            Met_Layer%q, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_cc
    call read_IntAuc_from_bin_file(fileunit, &
            Met_Layer%cc, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_clwc
    call read_FAdoxy_from_bin_file(fileunit, &
            Met_Layer%clwc, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_ciwc
    call read_FAdoxy_from_bin_file(fileunit, &
            Met_Layer%ciwc, error_flag)
    IF (error_flag .ne. no_error) goto 998

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"Read_Met_Layer: done with Read_Met_Layer"

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Read_Met_Layer"
    return

  end subroutine Read_Met_Layer
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_AMD_Met_MDS(AMD_Met_MDS, N_Model_Layers, error_flag)
    !  #[
    TYPE(AMD_Met_MDS_type), pointer     :: AMD_Met_MDS    ! input/output
    integer,                intent(in)  :: N_Model_Layers ! input
    integer,                intent(out) :: error_flag     ! output

    ! local variable
    integer :: p, AllocStatus

    error_flag = no_error

    IF (.not. associated(AMD_Met_MDS)) THEN
       print *,"ERROR in Init_AMD_Met_MDS: "
       print *,"pointer AMD_Met_MDS seems not associated !!!"
       error_flag = error_programming
       return
    END IF
    
!    call InitDateTime(AMD_Met_MDS%Start_of_Obs_DateTime)
    
    AMD_Met_MDS%us     = missing_indicator_IntAs
    AMD_Met_MDS%vs     = missing_indicator_IntAs
    AMD_Met_MDS%ps     = missing_indicator_IntAul
    AMD_Met_MDS%err_ps = missing_indicator_FAdoxy
    AMD_Met_MDS%zs     = missing_indicator_IntAl

    IF (N_Model_Layers .eq. 0) THEN
       ! init it to be empty
       nullify(AMD_Met_MDS%Met_Layer)
    ELSE
       allocate(AMD_Met_MDS%Met_Layer(N_Model_Layers),&
                STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_AMD_Met_MDS"
          error_flag = error_allocate
          return
       END IF

       DO p=1,N_Model_Layers
          call Init_AMD_Met_Layer(AMD_Met_MDS%Met_Layer(p),&
                                     error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    END IF

  end subroutine Init_AMD_Met_MDS
    !  #]
  subroutine Delete_AMD_Met_MDS(AMD_Met_MDS)
    !  #[
    TYPE(AMD_Met_MDS_type), pointer :: AMD_Met_MDS

    IF (.not. associated(AMD_Met_MDS)) return
    
    IF (associated(AMD_Met_MDS%Met_Layer)) THEN
       deallocate(AMD_Met_MDS%Met_Layer)
    END IF

  end subroutine Delete_AMD_Met_MDS
    !  #]
  function GetByteSize_AMD_Met_MDS(Met_MDS) result(c_bytesize)
    !  #[
    TYPE(AMD_Met_MDS_type), intent(in) :: Met_MDS
    integer(IntAl) :: c_bytesize ! result
    
    c_bytesize = &
         int(&
!             1*c_bytelen_DateTime + &
             2                     + & ! spare bytes
             2*c_bytelen_IntAs     + &
             1*c_bytelen_IntAul    + &
             1*c_bytelen_FAdoxy    + &
             1*c_bytelen_IntAl     + &
             2                     + & ! spare bytes
             size(Met_MDS%Met_Layer)*GetByteSize_Met_Layer() + &
             2                      ,IntAl)   ! spare bytes

    ! typical size:
    ! fortran size = 2+2*2+8+8+4+2+60*99+2 = 5970 bytes
    ! c       size = 2+2*2+4+8+4+2+60*71+2 = 4286 bytes

  end function GetByteSize_AMD_Met_MDS
    !  #]
  !---------------------------------------------------------------------
  subroutine Print_AMD_Met_MDS(Met_MDS,error_flag,PrintAll)
    !  #[

    TYPE(AMD_Met_MDS_type), pointer     :: Met_MDS
    integer,                intent(out) :: error_flag       ! output
    logical, optional,      intent(in)  :: PrintAll

    ! local variables
    integer           :: p, s, step
    logical           :: do_PrintAll

    character(len=13) :: txt

    character(len=maxlen_IntAs)  :: str_IntAs
    character(len=maxlen_IntAl)  :: str_IntAl
    character(len=maxlen_IntAul) :: str_IntAul
    !character(len=maxlen_DateTime_KVT) :: DateTimeString

    error_flag = no_error

    do_PrintAll = .false.
    IF (present(PrintAll)) do_PrintAll = PrintAll

    IF (associated(Met_MDS)) THEN
       print *,"Met_MDS%..."
    ELSE
       print *,"Met_MDS [seems not associated]"
       return
    END IF

!    ! print the date and time
!    DateTimeString = Construct_KVT_DateTime(Met_MDS%Start_of_Obs_DateTime)
!    print *,"...%start_of_obs_datetime: "//trim(DateTimeString)

    call SetIntAs(Met_MDS%us,&
                  str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    print *,"...%us                = ",str_IntAs

    call SetIntAs(Met_MDS%vs,&
                  str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    print *,"...%vs                = ",str_IntAs

    call SetIntAul(Met_MDS%ps,&
                  str_IntAul,error_flag)
    if (error_flag .ne. no_error) return
    print *,"...%ps                = ",str_IntAul

    txt = real2txt(Met_MDS%err_ps)
    print *,"...%err_ps            = ",txt

    call SetIntAl(Met_MDS%zs,&
                  str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    print *,"...%zs                = ",str_IntAl

    IF (associated(Met_MDS%Met_Layer)) THEN
       s = size(Met_MDS%Met_Layer)

       IF (do_PrintAll .or. (s .le. 1) ) THEN
          step=1
          print "(a,i2)"," ==> Num_of_Model_Layers = ",s
       ELSE
          step = s-1
          print "(a,i2,a)",&
               " ==> Num_of_Model_Layers = ",s," [printing first and last one]"
       END IF

       DO p=1,s,step
          print "(' ...%Met_Layer(',i2.2,')%...')",p
          call Print_Met_Layer(&
                   Met_MDS%Met_Layer(p), error_flag)
          if (error_flag .ne. no_error) return
       END DO
    ELSE
       print *,"...%Met_Layer(:) = [NOT ASSOCIATED]"
    END IF

  end subroutine Print_AMD_Met_MDS
    !  #]
  subroutine Write_AMD_Met_MDS(fileunit,Met_MDS,error_flag)
    !  #[
    !  P. Poli   09-02-2006   Original code

    integer,                intent(in)  :: fileunit    ! input
    TYPE(AMD_Met_MDS_type), pointer     :: Met_MDS
    integer,                intent(out) :: error_flag  ! output
    
    ! local variable
    integer :: s, i, nr_of_zero_bytes

    error_flag = no_error

    IF (.not. associated(Met_MDS)) THEN
       print *,"ERROR in Write_AMD_Met_MDS():"
       print *,"The pointer to this Met_MDS seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! 2 spare bytes
    nr_of_zero_bytes = 2
    call write_SpareBytes_to_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! AMD_us
    call write_IntAs_to_bin_file(fileunit, Met_MDS%us, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! AMD_vs
    call write_IntAs_to_bin_file(fileunit, Met_MDS%vs, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! AMD_ps
    call write_IntAul_to_bin_file(fileunit, Met_MDS%ps, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! AMD_err_ps
    call write_FAdoxy_to_bin_file(fileunit, Met_MDS%err_ps, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! AMD_zs
    call write_IntAl_to_bin_file(fileunit, Met_MDS%zs, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! 2 spare bytes
    nr_of_zero_bytes = 2
    call write_SpareBytes_to_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! note that one of the two datasets, nadir/offnadir, may be empty
    ! and will not be associated in that case. This is a valid condition.
    IF (associated(Met_MDS%Met_Layer)) THEN
       s = size(Met_MDS%Met_Layer)
    
       DO i=1,s
          call Write_Met_Layer(fileunit, Met_MDS%Met_Layer(i), error_flag)
          IF (error_flag .ne. no_error) goto 998
       ENDDO
    END IF

    ! 2 spare bytes
    nr_of_zero_bytes = 2
    call write_SpareBytes_to_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"Write_AMD_Met_MDS: done with Write_AMD_Met_MDS"

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Write_AMD_Met_MDS"
    return

  end subroutine Write_AMD_Met_MDS
    !  #]
  subroutine Read_AMD_Met_MDS(Met_DSD,fileunit,which_DSR,Met_MDS, &
                              error_flag)
    !  #[
    !  P. Poli   09-02-2006   Original code

    TYPE(DSD_type),         pointer     :: Met_DSD    ! input
    integer,                intent(in)  :: fileunit    ! input
    integer,                intent(in)  :: which_DSR      ! input
    TYPE(AMD_Met_MDS_type), pointer     :: Met_MDS
    integer,                intent(out) :: error_flag  ! output
    
    ! local variable
    integer      :: s, i, nr_of_zero_bytes
    integer(i8_) :: pos, endpos
    integer      :: currentpos

    error_flag = no_error

    IF (.not. associated(Met_MDS)) THEN
       print *,"ERROR in Read_AMD_Met_MDS():"
       print *,"The pointer to this Met_MDS seems not associated !!"
       error_flag = error_programming
       return
    END IF

    pos = Met_DSD%start_of_ds + (which_DSR-1)*Met_DSD%size_of_one_dsr
    call SetFilePointer(fileunit,pos,error_flag)
    IF (error_flag .ne. no_error) return
    
    print *,"trying to read the binary Met_MDS nr. ",which_DSR
    
    ! 2 spare bytes
    nr_of_zero_bytes = 2
    call read_SpareBytes_from_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! AMD_us
    call read_IntAs_from_bin_file(fileunit, Met_MDS%us, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! AMD_vs
    call read_IntAs_from_bin_file(fileunit, Met_MDS%vs, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! AMD_ps
    call read_IntAul_from_bin_file(fileunit, Met_MDS%ps, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! AMD_err_ps
    call read_FAdoxy_from_bin_file(fileunit, Met_MDS%err_ps, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! AMD_zs
    call read_IntAl_from_bin_file(fileunit, Met_MDS%zs, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! 2 spare bytes
    nr_of_zero_bytes = 2
    call read_SpareBytes_from_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! note that one of the two datasets, nadir/offnadir, may be empty
    ! and will not be associated in that case. This is a valid condition.
    IF (associated(Met_MDS%Met_Layer)) THEN
       s = size(Met_MDS%Met_Layer)
    
       DO i=1,s
          call Read_Met_Layer(fileunit, Met_MDS%Met_Layer(i), error_flag)
          IF (error_flag .ne. no_error) goto 998
       ENDDO
    END IF

    ! 2 spare bytes
    nr_of_zero_bytes = 2
    call read_SpareBytes_from_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! calculate expected end-position after reading this block of data
    endpos = Met_DSD%start_of_ds + which_DSR*Met_DSD%size_of_one_dsr
    ! request the current file position
    inquire(unit=fileunit,nextrec=currentpos)
    
    IF (currentpos .ne. endpos) THEN
       print *,"ERROR in Read_AMD_Met_MDS():"
       print *,"there seems to be a problem while reading this binary dataset"
       print *,"expected file position after this DSR is: ",endpos
       print *,"but the actual position seems to be:      ",currentpos
       print *,"Maybe the fileformat definition has changed ???"
       error_flag = error_reading_file
       return
    END IF
    
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"Read_AMD_Met_MDS: done with Read_AMD_Met_MDS"
    
    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Read_AMD_Met_MDS"
    return

  end subroutine Read_AMD_Met_MDS
    !  #]
   !---------------------------------------------------------------------
end module AMD_Met_MDS
