module AuxClim_DataStructure
  !  #[ Description
  !     a module to define the datastructure needed to stiore the
  !     auxiliary climatology dataset, needed to hold values for
  !     extinction-to-backscatter ratio.
  !
  !     created:    21-Jul-2006, by: J. de Kloe
  !
  !     Modifications:
  !     31-Jul-2006  J. de Kloe  writing fully implemented now
  !     01-Aug-2006  J. de Kloe  added AsciiWrite/Print routines
  !                              and Construct_Test_Climatology()
  !     04-Apr-2007  J. de Kloe  clean up of USE FixedHeader
  !     24-Apr-2007  J. de Kloe  phase out the SummaryVars code
  !     14-Nov-2007  J. de Kloe  replace outdated maxlen_DateTime parameter
  !                              by maxlen_DateTime_KVT
  !     16-Jan-2008  J. de Kloe  phase out integer kind i_
  !     21-Jan-2008  J. de kloe  added workaround in call to Write_MPH_to_KVT
  !                              and to GetByteSize_MPH_KVT
  !                              to not change the fileformat for now.
  !     21-Feb-2008  J. de Kloe  removed this workaround again since it was
  !                              caused by a misunderstanding on L1B side
  !
  !  #]
  !  #[ modules used 
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE FixedHeader, only: FH_type, Init_FH, Fill_FH_items, &
                         AsciiWrite_FH
  USE MainProductHeader, only: MPH_type, Init_MPH, Print_MPH, &
                               Fill_MPH_items, GetByteSize_MPH_KVT, &
                               Write_MPH_to_KVT, AsciiWrite_MPH
  USE AuxClim_SPH, only: AuxClim_SPH_type, &
                         Delete_AuxClim_SPH, & 
                         GetByteSize_AuxClim_SPH_KVT, &
                         Write_AuxClim_SPH_to_KVT,AsciiWrite_AuxClim_SPH
!, Print_AuxClim_SPH, &
!       Fill_AuxClim_SPH_items, 

  ! support modules
  USE DirectBinaryIO, only: DirectBinaryIO_fileunit_type, &
                            GetFilePointer
  USE EE_CFI_Datatypes, only: &
       DateTimeType, InitDateTime, SetDateTime, maxlen_DateTime_KVT, &
       write_DateTime_to_bin_file, c_bytelen_DateTime, &
       Construct_KVT_DateTime, &
       IntAs, c_bytelen_IntAs, write_IntAs_to_bin_file, maxlen_IntAs, &
       IntAl, c_bytelen_IntAl, write_IntAl_to_bin_file, maxlen_IntAl, &
       IntAd, &
       missing_indicator_IntAl, &
       SetIntAs, SetIntAl
!, SetDateTime, PrintDateTime, &
!       FAdoxy, c_bytelen_fadoxy, &
!       IntAus, maxval_IntAus, maxlen_IntAus, SetIntAus, c_bytelen_IntAus, &
!       IntAl,  maxval_IntAl,  maxlen_IntAl,  SetIntAl,  c_bytelen_IntAl,  &
!       IntAul, maxval_IntAul, maxlen_IntAul, SetIntAul, c_bytelen_IntAul, &
!       IntAd,  maxval_IntAd,  maxlen_IntAd,  SetIntAd,  c_bytelen_IntAd,  &
!       write_IntAus_to_bin_file, read_IntAus_from_bin_file, &
!       write_IntAul_to_bin_file, read_IntAul_from_bin_file, &
!       write_IntAd_to_bin_file,  read_IntAd_from_bin_file, &
!       write_FAdoxy_to_bin_file, read_FAdoxy_from_bin_file, &
!       BufferedFile_type,&
!       read_IntAus_from_buffbin_file, &
!       read_IntAul_from_buffbin_file, &
!       read_IntAd_from_buffbin_file,  &
!       read_FAdoxy_from_buffbin_file
  USE LunManager, only: fileunit_stdout
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE numerics, only: i8_, r8_
  USE ErrorHandler, only: no_error, error_programming, error_allocate, &
                          error_writing_file
!  USE Compiler_Features, only: newline
  !  #]
  !  #[ Parameter and data type definitions

  implicit none

  ! see TN2.1, sec ...
  TYPE ClimData_type
     integer(IntAl)     :: StartAltitude
     integer(IntAl)     :: EndAltitude
     integer(IntAl)     :: S
     integer(IntAl)     :: S_stdev
  END TYPE ClimData_type

  ! see TN2.1, sec ...
  TYPE ClimAlt_type
     integer(IntAl)     :: StartLongitude
     integer(IntAl)     :: EndLongitude
     integer(IntAs)     :: Num_Altitude_Ranges
     TYPE(ClimData_type), dimension(:), pointer :: ClimData
  END TYPE ClimAlt_type

  ! see TN2.1, sec ...
  TYPE ClimLonAlt_type
     integer(IntAl)     :: StartLatitude
     integer(IntAl)     :: EndLatitude
     integer(IntAs)     :: Num_Longitude_Ranges
     TYPE(ClimAlt_type), dimension(:), pointer :: ClimAlt
  END TYPE ClimLonAlt_type

  ! see TN2.1, sec ...
  TYPE ClimLatLonAlt_type
     TYPE(DateTimeType) :: StartDateTime
     TYPE(DateTimeType) :: EndDateTime
     integer(IntAs)     :: Num_Latitude_Ranges
     TYPE(ClimLonAlt_type), dimension(:), pointer :: ClimLonAlt
  END TYPE ClimLatLonAlt_type

  ! see TN2.1, sec ...
  TYPE AuxClim_DS_type
     integer(IntAs) :: Num_DateTime_Ranges
     TYPE(ClimLatLonAlt_type), dimension(:), pointer :: ClimLatLonAlt
  END TYPE AuxClim_DS_type

  TYPE AuxClim_DataType
     ! default headers needed for all EE_CFI type files
     TYPE(FH_type),          pointer :: FH
     TYPE(MPH_type),         pointer :: MPH
     TYPE(AuxClim_SPH_type), pointer :: SPH
     !
     ! the attached datasets
     TYPE(AuxClim_DS_type), pointer :: AuxClim_DS
  END TYPE AuxClim_DataType

  ! the default verbosity setting is Err (as silent as possible)
  integer, private, save :: LocalVerbosity = VerbosityError

  !  #]   
  !  #[ Module interface
  !     everything is public in this module,
  !     unless explicitely given the private keyword below
!  private :: PrintSizeAndGetStep
  !  #]   
contains
  !---------------------------------------------
  subroutine Init_AuxClim_DataStruct_Module(ChoosenVerbosity)
    !  #[
    integer, intent(in), optional :: ChoosenVerbosity

    IF (present(ChoosenVerbosity)) LocalVerbosity = ChoosenVerbosity

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"AuxClim_DataStructure module has been initialised"
       call print_verbosity_setting(LocalVerbosity,&
                                    ModuleName="AuxClim_DataStructure")
    END IF

  end subroutine Init_AuxClim_DataStruct_Module
    !  #]
  !---------------------------------------------
  subroutine Init_ClimData(ClimData)
    !  #[
    TYPE(ClimData_type), intent(out) :: ClimData ! result

    ClimData%StartAltitude = missing_indicator_IntAl
    ClimData%EndAltitude   = missing_indicator_IntAl
    ClimData%S             = missing_indicator_IntAl
    ClimData%S_stdev       = missing_indicator_IntAl

  end subroutine Init_ClimData
    !  #]
  function GetByteSize_ClimData() result(c_bytesize)
    !  #[
    integer(IntAl)                  :: c_bytesize ! result

    c_bytesize = 4*c_bytelen_IntAl

  end function GetByteSize_ClimData
    !  #]
  subroutine Write_ClimData(fileunit,ClimData,error_flag)
    !  #[
    TYPE(DirectBinaryIO_fileunit_type), intent(in)  :: fileunit
    TYPE(ClimData_type),                intent(in)  :: ClimData
    integer,                            intent(out) :: error_flag

    error_flag = no_error

    call write_IntAl_to_bin_file(fileunit,ClimData%StartAltitude,&
                                 error_flag)
    IF (error_flag .ne. no_error) return

    call write_IntAl_to_bin_file(fileunit,ClimData%EndAltitude,&
                                 error_flag)
    IF (error_flag .ne. no_error) return

    call write_IntAl_to_bin_file(fileunit,ClimData%S,&
                                 error_flag)
    IF (error_flag .ne. no_error) return

    call write_IntAl_to_bin_file(fileunit,ClimData%S_stdev,&
                                 error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine Write_ClimData
    !  #]
  subroutine Print_ClimData(ClimData,prefix,error_flag)
    !  #[
    TYPE(ClimData_type), intent(in)  :: ClimData   ! input
    character(len=*),    intent(in)  :: prefix     ! input
    integer,             intent(out) :: error_flag ! output

    call AsciiWrite_ClimData(ClimData,fileunit_stdout,prefix,error_flag)

  end subroutine Print_ClimData
    !  #]
  subroutine AsciiWrite_ClimData(ClimData,fileunit,prefix,error_flag)
    !  #[
    TYPE(ClimData_type), intent(in)  :: ClimData   ! input
    integer,             intent(in)  :: fileunit   ! input
    character(len=*),    intent(in)  :: prefix     ! input
    integer,             intent(out) :: error_flag ! output

    ! local variables
    character(len=maxlen_IntAl)    :: str_IntAl

    error_flag = no_error

    call SetIntAl(ClimData%StartAltitude,str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%StartAltitude = "//trim(str_IntAl)

    call SetIntAl(ClimData%EndAltitude,str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%EndAltitude   = "//trim(str_IntAl)

    call SetIntAl(ClimData%S,str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%S             = "//trim(str_IntAl)

    call SetIntAl(ClimData%S_stdev,str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%S_stdev       = "//trim(str_IntAl)

    return

999 print *,"ERROR in AsciiWrite_ClimData:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_ClimData
    !  #]
  !---------------------------------------------
  subroutine Init_ClimAlt(ClimAlt,Num_Altitude_Ranges,error_flag)
    !  #[
    TYPE(ClimAlt_type), intent(out) :: ClimAlt ! result
    integer(IntAs),     intent(in)  :: Num_Altitude_Ranges
    integer,            intent(out) :: error_flag

    ! local variable
    integer :: i, AllocStatus

    ClimAlt%StartLongitude      = missing_indicator_IntAl
    ClimAlt%EndLongitude        = missing_indicator_IntAl
    ClimAlt%Num_Altitude_Ranges = Num_Altitude_Ranges

   IF (Num_Altitude_Ranges .eq. 0) THEN
       ! init the struct to hold no data
       nullify(ClimAlt%ClimData)
    ELSE
       allocate(ClimAlt%ClimData(Num_Altitude_Ranges),STAT=AllocStatus)
 
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_ClimAlt()"
          error_flag = error_allocate
          return
       END IF
       DO i=1,Num_Altitude_Ranges
          call Init_ClimData(ClimAlt%ClimData(i))
       END DO
    END IF

  end subroutine Init_ClimAlt
    !  #]
  subroutine Delete_ClimAlt(ClimAlt)
    !  #[
    TYPE(ClimAlt_type), intent(inout) :: ClimAlt ! deleted

    IF (associated(ClimAlt%ClimData)) &
         deallocate(ClimAlt%ClimData)

  end subroutine Delete_ClimAlt
    !  #]
  function GetByteSize_ClimAlt(ClimAlt) result(c_bytesize)
    !  #[
    TYPE(ClimAlt_type), intent(in) :: ClimAlt
    integer(IntAl)                 :: c_bytesize ! result

    c_bytesize = 2*c_bytelen_IntAl + c_bytelen_IntAs
    IF (associated(ClimAlt%ClimData)) THEN
       c_bytesize = c_bytesize + &
            int(size(ClimAlt%ClimData) * GetByteSize_ClimData(),IntAl)
    ELSE
       print *,"WARNING in GetByteSize_ClimAlt():"
       print *,"ClimAlt%ClimData [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

  end function GetByteSize_ClimAlt
    !  #]
  subroutine Write_ClimAlt(fileunit,ClimAlt,error_flag)
    !  #[
    TYPE(DirectBinaryIO_fileunit_type), intent(in)  :: fileunit
    TYPE(ClimAlt_type),                 intent(in)  :: ClimAlt
    integer,                            intent(out) :: error_flag

    ! local variable
    integer :: i
    
    error_flag = no_error

    call write_IntAl_to_bin_file(fileunit,ClimAlt%StartLongitude,&
                                 error_flag)
    IF (error_flag .ne. no_error) return

    call write_IntAl_to_bin_file(fileunit,ClimAlt%EndLongitude,&
                                 error_flag)
    IF (error_flag .ne. no_error) return

    call write_IntAs_to_bin_file(fileunit,ClimAlt%Num_Altitude_Ranges, &
                                 error_flag)
    IF (error_flag .ne. no_error) return

    DO i=1,ClimAlt%Num_Altitude_Ranges
       call Write_ClimData(fileunit,ClimAlt%ClimData(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine Write_ClimAlt
    !  #]
  subroutine Print_ClimAlt(ClimAlt,prefix,PrintAll,error_flag)
    !  #[
    TYPE(ClimAlt_type), intent(in)  :: ClimAlt    ! input
    character(len=*),   intent(in)  :: prefix     ! input
    logical,            intent(in)  :: PrintAll   ! input
    integer,            intent(out) :: error_flag ! output

    call AsciiWrite_ClimAlt(ClimAlt,fileunit_stdout,prefix,&
                            PrintAll,error_flag)

  end subroutine Print_ClimAlt
    !  #]
  subroutine AsciiWrite_ClimAlt(ClimAlt,fileunit,prefix,&
                                PrintAll,error_flag)
    !  #[
    TYPE(ClimAlt_type), intent(in)  :: ClimAlt    ! input
    integer,            intent(in)  :: fileunit   ! input
    character(len=*),   intent(in)  :: prefix     ! input
    logical,            intent(in)  :: PrintAll   ! input
    integer,            intent(out) :: error_flag ! output

    ! local variables
    integer                        :: i,s,step
    character(len=maxlen_IntAs)    :: str_IntAs
    character(len=maxlen_IntAl)    :: str_IntAl
    character(len=2)               :: str_cntr

    error_flag = no_error

    call SetIntAl(ClimAlt%StartLongitude,str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%StartLongitude      = "//trim(str_IntAl)

    call SetIntAl(ClimAlt%EndLongitude,str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%EndLongitude        = "//trim(str_IntAl)

    call SetIntAs(ClimAlt%Num_Altitude_Ranges,str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Num_Altitude_Ranges = "//trim(str_IntAs)

    IF (associated(ClimAlt%ClimData)) THEN
       s=ClimAlt%Num_Altitude_Ranges
       IF (PrintAll .or. (s .le. 1)) THEN
          step=1
       ELSE
          step = s - 1
       END IF
       DO i=1,s,step
          write(str_cntr,"(i2.2)",err=998) i
          call AsciiWrite_ClimData(&
               ClimAlt%ClimData(i), fileunit, &
               trim(prefix)//"%ClimData("//trim(str_cntr)//")",error_flag)
          IF (error_flag .ne. no_error) goto 999
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%ClimData(:) [seems not associated]"
    END IF

    return

998 print *,"conversion error in AsciiWrite_ClimAlt:"
    print *,"could not convert the counter ",i," to a string .."
    error_flag = error_writing_file
    return

999 print *,"ERROR in AsciiWrite_ClimAlt:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_ClimAlt
    !  #]
  !---------------------------------------------
  subroutine Init_ClimLonAlt(ClimLonAlt,&
       Num_Longitude_Ranges,Num_Altitude_Ranges,error_flag)
    !  #[
    TYPE(ClimLonAlt_type), intent(out) :: ClimLonAlt ! result
    integer(IntAs),        intent(in)  :: Num_Longitude_Ranges
    integer(IntAs),        intent(in)  :: Num_Altitude_Ranges
    integer,               intent(out) :: error_flag

    ! local variable
    integer :: i, AllocStatus

    ClimLonAlt%StartLatitude        = missing_indicator_IntAl
    ClimLonAlt%EndLatitude          = missing_indicator_IntAl
    ClimLonAlt%Num_Longitude_Ranges = Num_Longitude_Ranges

   IF (Num_Longitude_Ranges .eq. 0) THEN
       ! init the struct to hold no data
       nullify(ClimLonAlt%ClimAlt)
    ELSE
       allocate(ClimLonAlt%ClimAlt(Num_Longitude_Ranges),STAT=AllocStatus)
 
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_ClimLonAlt()"
          error_flag = error_allocate
          return
       END IF
       DO i=1,Num_Longitude_Ranges
          call Init_ClimAlt(ClimLonAlt%ClimAlt(i),Num_Altitude_Ranges,&
                            error_flag)
       END DO
    END IF

  end subroutine Init_ClimLonAlt
    !  #]
  subroutine Delete_ClimLonAlt(ClimLonAlt)
    !  #[
    TYPE(ClimLonAlt_type), intent(inout) :: ClimLonAlt ! result

    ! local variable
    integer :: i

    IF (associated(ClimLonAlt%ClimAlt)) THEN
       DO i=1,size(ClimLonAlt%ClimAlt)
          call Delete_ClimAlt(ClimLonAlt%ClimAlt(i))
       END DO
       deallocate(ClimLonAlt%ClimAlt)
    END IF

  end subroutine Delete_ClimLonAlt
    !  #]
  function GetByteSize_ClimLonAlt(ClimLonAlt) result(c_bytesize)
    !  #[
    TYPE(ClimLonAlt_type), intent(in) :: ClimLonAlt
    integer(IntAl)                    :: c_bytesize ! result

    ! local variable
    integer :: i

    c_bytesize = 2*c_bytelen_IntAl + c_bytelen_IntAs
    IF (associated(ClimLonAlt%ClimAlt)) THEN
       ! remember that this datastructure is intended to accomodate also
       ! irregular grids, therefore we have to inspect the size of each array.
       ! Assuming that the size will just be
       !   size(ClimLonAlt%ClimAlt) * &
       !   GetByteSize_ClimAlt(ClimLonAlt%ClimAlt(1))
       ! may prove to be wrong in the future !
       DO i=1,size(ClimLonAlt%ClimAlt)
          c_bytesize = c_bytesize + GetByteSize_ClimAlt(ClimLonAlt%ClimAlt(i))
       END DO
    ELSE
       print *,"WARNING in GetByteSize_ClimLonAlt():"
       print *,"ClimLonAlt%ClimAlt [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

  end function GetByteSize_ClimLonAlt
    !  #]
  subroutine Write_ClimLonAlt(fileunit,ClimLonAlt,error_flag)
    !  #[
    TYPE(DirectBinaryIO_fileunit_type), intent(in)  :: fileunit
    TYPE(ClimLonAlt_type),              intent(in)  :: ClimLonAlt
    integer,                            intent(out) :: error_flag

    ! local variable
    integer :: i
    
    error_flag = no_error

    call write_IntAl_to_bin_file(fileunit,ClimLonAlt%StartLatitude,&
                                 error_flag)
    IF (error_flag .ne. no_error) return

    call write_IntAl_to_bin_file(fileunit,ClimLonAlt%EndLatitude,&
                                 error_flag)
    IF (error_flag .ne. no_error) return

    call write_IntAs_to_bin_file(fileunit,ClimLonAlt%Num_Longitude_Ranges, &
                                 error_flag)
    IF (error_flag .ne. no_error) return

    DO i=1,ClimLonAlt%Num_Longitude_Ranges
       call Write_ClimAlt(fileunit,ClimLonAlt%ClimAlt(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine Write_ClimLonAlt
    !  #]
  subroutine Print_ClimLonAlt(ClimLonAlt,prefix,&
                              PrintAll,error_flag)
    !  #[
    TYPE(ClimLonAlt_type), intent(in)  :: ClimLonAlt    ! input
    character(len=*),      intent(in)  :: prefix        ! input
    logical,               intent(in)  :: PrintAll      ! input
    integer,               intent(out) :: error_flag    ! output

    call AsciiWrite_ClimLonAlt(ClimLonAlt,fileunit_stdout,prefix,&
                               PrintAll,error_flag)

  end subroutine Print_ClimLonAlt
    !  #]
  subroutine AsciiWrite_ClimLonAlt(ClimLonAlt,fileunit,prefix,&
                                   PrintAll,error_flag)
    !  #[
    TYPE(ClimLonAlt_type), intent(in)  :: ClimLonAlt    ! input
    integer,               intent(in)  :: fileunit      ! input
    character(len=*),      intent(in)  :: prefix        ! input
    logical,               intent(in)  :: PrintAll      ! input
    integer,               intent(out) :: error_flag    ! output

    ! local variables
    integer :: i,s,step
    character(len=maxlen_IntAs)    :: str_IntAs
    character(len=maxlen_IntAl)    :: str_IntAl
    character(len=2)               :: str_cntr

    error_flag = no_error

    call SetIntAl(ClimLonAlt%StartLatitude,str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%StartLatitude        = "//trim(str_IntAl)

    call SetIntAl(ClimLonAlt%EndLatitude,str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%EndLatitude          = "//trim(str_IntAl)

    call SetIntAs(ClimLonAlt%Num_Longitude_Ranges,str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Num_Longitude_Ranges = "//trim(str_IntAs)

    IF (associated(ClimLonAlt%ClimAlt)) THEN
       s=ClimLonAlt%Num_Longitude_Ranges
       IF (PrintAll .or. (s .le. 1)) THEN
          step=1
       ELSE
          step = s - 1
       END IF
       DO i=1,s,step
          write(str_cntr,"(i2.2)",err=998) i
          call AsciiWrite_ClimAlt(&
               ClimLonAlt%ClimAlt(i), fileunit, &
               trim(prefix)//"%ClimAlt("//trim(str_cntr)//")",&
               PrintAll,error_flag)
          IF (error_flag .ne. no_error) goto 999
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%ClimAlt(:) [seems not associated]"
    END IF

    return

998 print *,"conversion error in AsciiWrite_ClimLonAlt:"
    print *,"could not convert the counter ",i," to a string .."
    error_flag = error_writing_file
    return

999 print *,"ERROR in AsciiWrite_ClimLonAlt:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_ClimLonAlt
    !  #]
  !---------------------------------------------
  subroutine Init_ClimLatLonAlt(ClimLatLonAlt,&
       Num_Latitude_Ranges,Num_Longitude_Ranges,Num_Altitude_Ranges,&
       error_flag)
    !  #[
    TYPE(ClimLatLonAlt_type), intent(out) :: ClimLatLonAlt ! result
    integer(IntAs),           intent(in)  :: Num_Latitude_Ranges
    integer(IntAs),           intent(in)  :: Num_Longitude_Ranges
    integer(IntAs),           intent(in)  :: Num_Altitude_Ranges
    integer,                  intent(out) :: error_flag

    ! local variable
    integer :: i, AllocStatus

    call InitDateTime(ClimLatLonAlt%StartDateTime)
    call InitDateTime(ClimLatLonAlt%EndDateTime)
    ClimLatLonAlt%Num_Latitude_Ranges = Num_Latitude_Ranges

    IF (Num_Latitude_Ranges .eq. 0) THEN
       ! init the struct to hold no data
       nullify(ClimLatLonAlt%ClimLonAlt)
    ELSE
       allocate(ClimLatLonAlt%ClimLonAlt(Num_Latitude_Ranges),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_ClimLatLonAlt()"
          error_flag = error_allocate
          return
       END IF
       DO i=1,Num_Latitude_Ranges
          call Init_ClimLonAlt(ClimLatLonAlt%ClimLonAlt(i),&
               Num_Longitude_Ranges, Num_Altitude_Ranges, error_flag)
       END DO
    END IF

  end subroutine Init_ClimLatLonAlt
    !  #]
  subroutine Delete_ClimLatLonAlt(ClimLatLonAlt)
    !  #[
    TYPE(ClimLatLonAlt_type), intent(inout) :: ClimLatLonAlt ! result

    ! local variable
    integer :: i

    IF (associated(ClimLatLonAlt%ClimLonAlt)) THEN
       DO i=1,size(ClimLatLonAlt%ClimLonAlt)
          call Delete_ClimLonAlt(ClimLatLonAlt%ClimLonAlt(i))
       END DO
       deallocate(ClimLatLonAlt%ClimLonAlt)
    END IF

  end subroutine Delete_ClimLatLonAlt
    !  #]
  function GetByteSize_ClimLatLonAlt(ClimLatLonAlt) result(c_bytesize)
    !  #[
    TYPE(ClimLatLonAlt_type), intent(in) :: ClimLatLonAlt
    integer(IntAl)                       :: c_bytesize ! result

    ! local variable
    integer :: i

    c_bytesize = 2*c_bytelen_DateTime + c_bytelen_IntAs
    IF (associated(ClimLatLonAlt%ClimLonAlt)) THEN
       ! remember that this datastructure is intended to accomodate also
       ! irregular grids, therefore we have to inspect the size of each array.
       ! Assuming that the size will just be
       !    size(ClimLatLonAlt%ClimLonAlt) * &
       !    GetByteSize_ClimLonAlt(ClimLatLonAlt%ClimLonAlt(1))
       ! may prove to be wrong in the future !
       DO i=1,size(ClimLatLonAlt%ClimLonAlt)
          c_bytesize = c_bytesize + &
                       GetByteSize_ClimLonAlt(ClimLatLonAlt%ClimLonAlt(i))
       END DO
    ELSE
       print *,"WARNING in GetByteSize_ClimLatLonAlt():"
       print *,"ClimLatLonAlt%ClimLonAlt [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

  end function GetByteSize_ClimLatLonAlt
    !  #]
  subroutine Write_ClimLatLonAlt(fileunit,ClimLatLonAlt,error_flag)
    !  #[
    TYPE(DirectBinaryIO_fileunit_type), intent(in)  :: fileunit
    TYPE(ClimLatLonAlt_type),           intent(in)  :: ClimLatLonAlt
    integer,                            intent(out) :: error_flag

    ! local variable
    integer :: i
    
    error_flag = no_error

    call write_DateTime_to_bin_file(fileunit,ClimLatLonAlt%StartDateTime,&
                                    error_flag)
    IF (error_flag .ne. no_error) return

    call write_DateTime_to_bin_file(fileunit,ClimLatLonAlt%EndDateTime,&
                                    error_flag)
    IF (error_flag .ne. no_error) return

    call write_IntAs_to_bin_file(fileunit,ClimLatLonAlt%Num_Latitude_Ranges, &
                                 error_flag)
    IF (error_flag .ne. no_error) return

    DO i=1,ClimLatLonAlt%Num_Latitude_Ranges
       call Write_ClimLonAlt(fileunit,ClimLatLonAlt%ClimLonAlt(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine Write_ClimLatLonAlt
    !  #]
  subroutine Print_ClimLatLonAlt(ClimLatLonAlt,prefix,&
                                 PrintAll,error_flag)
    !  #[
    TYPE(ClimLatLonAlt_type), intent(in)  :: ClimLatLonAlt    ! input
    character(len=*),         intent(in)  :: prefix           ! input
    logical,                  intent(in)  :: PrintAll         ! input
    integer,                  intent(out) :: error_flag       ! output

    call AsciiWrite_ClimLatLonAlt(ClimLatLonAlt,fileunit_stdout,prefix,&
                                      PrintAll,error_flag)

  end subroutine Print_ClimLatLonAlt
    !  #]
  subroutine AsciiWrite_ClimLatLonAlt(ClimLatLonAlt,fileunit,prefix,&
                                      PrintAll,error_flag)
    !  #[
    TYPE(ClimLatLonAlt_type), intent(in)  :: ClimLatLonAlt    ! input
    integer,                  intent(in)  :: fileunit         ! input
    character(len=*),         intent(in)  :: prefix           ! input
    logical,                  intent(in)  :: PrintAll         ! input
    integer,                  intent(out) :: error_flag       ! output

    ! local variables
    integer                            :: i,s,step
    character(len=maxlen_DateTime_KVT) :: str_DateTime
    character(len=maxlen_IntAs)        :: str_IntAs
    character(len=2)                   :: str_cntr

    error_flag = no_error

    str_DateTime = Construct_KVT_DateTime(ClimLatLonAlt%StartDateTime)
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%StartDateTime       = "//trim(str_DateTime)

    str_DateTime = Construct_KVT_DateTime(ClimLatLonAlt%EndDateTime)
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%EndDateTime         = "//trim(str_DateTime)

    call SetIntAs(ClimLatLonAlt%Num_Latitude_Ranges,str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Num_Latitude_Ranges = "//trim(str_IntAs)

    IF (associated(ClimLatLonAlt%ClimLonAlt)) THEN
       s=ClimLatLonAlt%Num_Latitude_Ranges
       IF (PrintAll .or. (s .le. 1)) THEN
          step=1
       ELSE
          step = s - 1
       END IF
       DO i=1,s,step
          write(str_cntr,"(i2.2)",err=998) i
          call AsciiWrite_ClimLonAlt(&
               ClimLatLonAlt%ClimLonAlt(i), fileunit, &
               trim(prefix)//"%ClimLonAlt("//trim(str_cntr)//")",&
               PrintAll,error_flag)
          IF (error_flag .ne. no_error) goto 999
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%ClimLonAlt(:) [seems not associated]"
    END IF

    return

998 print *,"conversion error in AsciiWrite_ClimLatLonAlt:"
    print *,"could not convert the counter ",i," to a string .."
    error_flag = error_writing_file
    return

999 print *,"ERROR in AsciiWrite_ClimLatLonAlt:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_ClimLatLonAlt
    !  #]
  !---------------------------------------------
  subroutine Init_AuxClim_DS(AuxClim_DS,Num_DateTime_Ranges,&
       Num_Latitude_Ranges,Num_Longitude_Ranges,Num_Altitude_Ranges,&
       error_flag)
    !  #[
    TYPE(AuxClim_DS_type), intent(out) :: AuxClim_DS ! result
    integer(IntAs),        intent(in)  :: Num_DateTime_Ranges
    integer(IntAs),        intent(in)  :: Num_Latitude_Ranges
    integer(IntAs),        intent(in)  :: Num_Longitude_Ranges
    integer(IntAs),        intent(in)  :: Num_Altitude_Ranges
    integer,               intent(out) :: error_flag

    ! local variable
    integer :: i, AllocStatus

    AuxClim_DS%Num_DateTime_Ranges = Num_DateTime_Ranges

    IF (Num_DateTime_Ranges .eq. 0) THEN
       ! init the struct to hold no data
       nullify(AuxClim_DS%ClimLatLonAlt)
    ELSE
       allocate(AuxClim_DS%ClimLatLonAlt(Num_DateTime_Ranges),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_AuxClim_DS()"
          error_flag = error_allocate
          return
       END IF
       DO i=1,Num_DateTime_Ranges
          call Init_ClimLatLonAlt(AuxClim_DS%ClimLatLonAlt(i),&
               Num_Latitude_Ranges, Num_Longitude_Ranges, &
               Num_Altitude_Ranges, error_flag)
       END DO
    END IF

  end subroutine Init_AuxClim_DS
    !  #]
  subroutine Delete_AuxClim_DS(AuxClim_DS)
    !  #[
    TYPE(AuxClim_DS_type), intent(inout) :: AuxClim_DS ! result

    ! local variable
    integer :: i

    IF (associated(AuxClim_DS%ClimLatLonAlt)) THEN
       DO i=1,size(AuxClim_DS%ClimLatLonAlt)
          call Delete_ClimLatLonAlt(AuxClim_DS%ClimLatLonAlt(i))
       END DO
       deallocate(AuxClim_DS%ClimLatLonAlt)
    END IF

  end subroutine Delete_AuxClim_DS
    !  #]
  function GetByteSize_AuxClim_DS(AuxClim_DS) result(c_bytesize)
    !  #[
    TYPE(AuxClim_DS_type), intent(in) :: AuxClim_DS
    integer(IntAl)                    :: c_bytesize ! result

    ! local variable
    integer :: i

    c_bytesize = c_bytelen_IntAs
    IF (associated(AuxClim_DS%ClimLatLonAlt)) THEN
       ! remember that this datastructure is intended to accomodate also
       ! irregular grids, therefore we have to inspect the size of each array.
       ! Assuming that the size will just be
       !    size(AuxClim_DS%ClimLatLonAlt) * &
       !    GetByteSize_ClimLatLonAlt(AuxClim_DS%ClimLatLonAlt(1))
       ! may prove to be wrong in the future !
       DO i=1,size(AuxClim_DS%ClimLatLonAlt)
          c_bytesize = c_bytesize + &
                       GetByteSize_ClimLatLonAlt(AuxClim_DS%ClimLatLonAlt(i))
       END DO
    ELSE
       print *,"WARNING in GetByteSize_AuxClim_DS():"
       print *,"AuxClim_DS%ClimLatLonAlt [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

  end function GetByteSize_AuxClim_DS
    !  #]
  subroutine Write_AuxClim_DS(fileunit,AuxClim_DS,error_flag)
    !  #[
    TYPE(DirectBinaryIO_fileunit_type), intent(in)  :: fileunit
    TYPE(AuxClim_DS_type),              intent(in)  :: AuxClim_DS
    integer,                            intent(out) :: error_flag

    ! local variable
    integer :: i

    error_flag = no_error

    call write_IntAs_to_bin_file(fileunit,AuxClim_DS%Num_DateTime_Ranges, &
                                 error_flag)
    IF (error_flag .ne. no_error) return

    DO i=1,AuxClim_DS%Num_DateTime_Ranges
       print *,'Write_AuxClim_DS: writing ClimLatLonAlt(',i,')'
       call Write_ClimLatLonAlt(fileunit,AuxClim_DS%ClimLatLonAlt(i),&
                                error_flag)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine Write_AuxClim_DS
    !  #]
  subroutine Print_AuxClim_DS(AuxClim_DS,prefix,&
                              PrintAll,error_flag)
    !  #[
    TYPE(AuxClim_DS_type),  pointer     :: AuxClim_DS       ! input
    character(len=*),       intent(in)  :: prefix           ! input
    logical,                intent(in)  :: PrintAll         ! input
    integer,                intent(out) :: error_flag       ! output

    call AsciiWrite_AuxClim_DS(AuxClim_DS,fileunit_stdout,prefix,&
                               PrintAll,error_flag)

  end subroutine Print_AuxClim_DS
    !  #]
  subroutine AsciiWrite_AuxClim_DS(AuxClim_DS,fileunit,prefix,&
                                   PrintAll,error_flag)
    !  #[
    TYPE(AuxClim_DS_type),  pointer     :: AuxClim_DS       ! input
    integer,                intent(in)  :: fileunit         ! input
    character(len=*),       intent(in)  :: prefix           ! input
    logical,                intent(in)  :: PrintAll         ! input
    integer,                intent(out) :: error_flag       ! output

    ! local variables
    integer                        :: i,s,step
    character(len=maxlen_IntAs)    :: str_IntAs
    character(len=2)               :: str_cntr

    error_flag = no_error

    IF (.not. associated(AuxClim_DS)) THEN
       write(fileunit,"(a)",err=999) trim(prefix)//&
            " [seems not associated]"
       return
    END IF
    
    call SetIntAs(AuxClim_DS%Num_DateTime_Ranges,str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Num_DateTime_Ranges = "//trim(str_IntAs)

    IF (associated(AuxClim_DS%ClimLatLonAlt)) THEN
       s=AuxClim_DS%Num_DateTime_Ranges
       IF (PrintAll .or. (s .le. 1)) THEN
          step=1
       ELSE
          step = s - 1
       END IF
       DO i=1,s,step
          write(str_cntr,"(i2.2)",err=998) i
          call AsciiWrite_ClimLatLonAlt(&
               AuxClim_DS%ClimLatLonAlt(i), fileunit, &
               trim(prefix)//"%ClimLatLonAlt("//trim(str_cntr)//")",&
               PrintAll,error_flag)
          IF (error_flag .ne. no_error) goto 999
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%ClimLatLonAlt(:) [seems not associated]"
    END IF

    return

998 print *,"conversion error in AsciiWrite_AuxClim_DS:"
    print *,"could not convert the counter ",i," to a string .."
    error_flag = error_writing_file
    return

999 print *,"ERROR in AsciiWrite_AuxClim_DS:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_AuxClim_DS
    !  #]
  !---------------------------------------------
  subroutine Init_AuxClim_DataStructure(AuxClimData)
    !  #[
    TYPE(AuxClim_DataType), intent(out) :: AuxClimData ! result

    nullify(AuxClimData%FH)
    nullify(AuxClimData%MPH)
    nullify(AuxClimData%SPH)
    nullify(AuxClimData%AuxClim_DS)

  end subroutine Init_AuxClim_DataStructure
    !  #]
  subroutine Delete_AuxClim_DataStructure(AuxClimData)
    !  #[
    TYPE(AuxClim_DataType), intent(inout) :: AuxClimData ! result

    ! clean up the header datastructures
    IF (associated(AuxClimData%FH)) call Delete_AuxClim_SPH(AuxClimData%SPH)

    IF (associated(AuxClimData%FH))  deallocate(AuxClimData%FH)
    IF (associated(AuxClimData%MPH)) deallocate(AuxClimData%MPH)
    IF (associated(AuxClimData%SPH)) deallocate(AuxClimData%SPH)

    ! clean up the dataset
    IF (associated(AuxClimData%AuxClim_DS)) THEN
       call Delete_AuxClim_DS(AuxClimData%AuxClim_DS)
       deallocate(AuxClimData%AuxClim_DS)
    END IF

  end subroutine Delete_AuxClim_DataStructure
    !  #]
  function GetByteSize_AuxClimDataStr_KVT(AuxClimData) result(c_bytesize)
    !  #[
    TYPE(AuxClim_DataType), intent(in) :: AuxClimData ! input
    integer(IntAl)                     :: c_bytesize  ! result

    c_bytesize = 0
    c_bytesize = c_bytesize + GetByteSize_MPH_KVT()
    c_bytesize = c_bytesize + GetByteSize_AuxClim_SPH_KVT()

    IF (associated(AuxClimData%AuxClim_DS)) THEN
       c_bytesize = c_bytesize + GetByteSize_AuxClim_DS(AuxClimData%AuxClim_DS)
    ELSE
       print *,"WARNING in GetByteSize_AuxClim_DataStr_KVT():"
       print *,"AuxClimData%AuxClim_DS [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

  end function GetByteSize_AuxClimDataStr_KVT
    !  #]
  subroutine Write_AuxClim_DataStructure(fileunit,AuxClimData,error_flag)
    !  #[
    TYPE(DirectBinaryIO_fileunit_type), intent(in)  :: fileunit
    TYPE(AuxClim_DataType),             intent(in)  :: AuxClimData
    integer,                            intent(out) :: error_flag

    ! local variables
    integer(i8_) :: pos_start, pos_end
    integer(i8_) :: nbytes_written, nbytes_expected

    error_flag = no_error

    ! write AuxClim MPH
    IF (associated(AuxClimData%MPH)) THEN
       call Write_MPH_to_KVT(AuxClimData%MPH,fileunit,error_flag)
       IF (error_flag .ne. no_error) return
    ELSE
       print *,"ERROR in Write_AuxClim_DataStructure:"
       print *,"the pointer AuxClim%MPH seems not associated !!!"
       error_flag = error_writing_file
       return
    END IF

    ! 3. write AuxClim SPH
    IF (associated(AuxClimData%SPH)) THEN
       call Write_AuxClim_SPH_to_KVT(AuxClimData%SPH,fileunit,error_flag)
       IF (error_flag .ne. no_error) return
    ELSE
       print *,"ERROR in Write_AuxClim_DataStructure:"
       print *,"the pointer AuxClim%SPH seems not associated !!!"
       error_flag = error_writing_file
       return
    END IF

    ! write AuxClim DS
    IF (associated(AuxClimData%AuxClim_DS)) THEN

       call GetFilePointer(fileunit,pos_start,error_flag)
       IF (error_flag .ne. no_error) return
       !inquire(unit=fileunit,nextrec=pos_start)
       print *,'WriteAuxClimtoBinaryFile: writing AuxClim_DS'

       call Write_AuxClim_DS(fileunit,AuxClimData%AuxClim_DS,error_flag)
       IF (error_flag .ne. no_error) return

       call GetFilePointer(fileunit,pos_end,error_flag)
       IF (error_flag .ne. no_error) return
       !inquire(unit=fileunit,nextrec=pos_end)

       nbytes_written  = pos_end-pos_start
       nbytes_expected = GetByteSize_AuxClim_DS(AuxClimData%AuxClim_DS)

       IF (nbytes_written .ne. nbytes_expected) THEN
          print *,"ERROR in Write_AuxClim_DataStructure:"
          print *,"pos_start       = ",pos_start
          print *,"pos_end         = ",pos_end
          print *,"nbytes_written  = ",nbytes_written
          print *,"nbytes_expected = ",nbytes_expected
          print *,"something went wrong when writing the AuxClim_DS to the"
          print *,"binary file, probably a programming error ..."
          error_flag = error_programming
          return
       END IF
    ELSE
       print *,"ERROR in Write_AuxClim_DataStructure:"
       print *,"the pointer AuxClim%AuxClim_DS seems not associated !!!"
       error_flag = error_writing_file
       return
    END IF

  end subroutine Write_AuxClim_DataStructure
    !  #]
  subroutine Print_AuxClim_DataStructure(AuxClimData,prefix,PrintAll,&
                                         error_flag)
    !  #[
    TYPE(AuxClim_DataType), intent(in)  :: AuxClimData    
    character(len=*),       intent(in)  :: prefix           ! input
    logical,                intent(in)  :: PrintAll         ! input
    integer,                intent(out) :: error_flag       ! output

    call AsciiWrite_AuxClim_DataStruct(AuxClimData,fileunit_stdout,prefix,&
                                       PrintAll,error_flag)

  end subroutine Print_AuxClim_DataStructure
    !  #]
  subroutine AsciiWrite_AuxClim_DataStruct(AuxClimData,fileunit,prefix,&
                                           PrintAll,error_flag)
    !  #[
    TYPE(AuxClim_DataType), intent(in)  :: AuxClimData    
    integer,                intent(in)  :: fileunit         ! input
    character(len=*),       intent(in)  :: prefix           ! input
    logical,                intent(in)  :: PrintAll         ! input
    integer,                intent(out) :: error_flag       ! output

    error_flag = no_error

    write(fileunit,"(a)") "=================================="
    write(fileunit,"(a)") "contents of the "//trim(prefix)//" datastructure:"
    write(fileunit,"(a)") "=================================="

    call AsciiWrite_FH(AuxClimData%FH, fileunit,prefix=trim(prefix)//"%FH",&
                       error_flag=error_flag)
    IF(error_flag .ne. no_error) return
    call AsciiWrite_MPH(AuxClimData%MPH,fileunit,prefix=trim(prefix)//"%MPH",&
                        error_flag=error_flag)
    IF(error_flag .ne. no_error) return
    call AsciiWrite_AuxClim_SPH(AuxClimData%SPH,fileunit,&
                        prefix=trim(prefix)//"%SPH",error_flag=error_flag)
    IF(error_flag .ne. no_error) return

    call AsciiWrite_AuxClim_DS(AuxClimData%AuxClim_DS,fileunit,&
                               prefix=trim(prefix)//"%AuxClim_DS",&
                               PrintAll=PrintAll,error_flag=error_flag)
    IF(error_flag .ne. no_error) return

  end subroutine AsciiWrite_AuxClim_DataStruct
    !  #]
  !---------------------------------------------
  subroutine Construct_Test_Climatology(AuxClim, error_flag)
    !  #[
    ! fill the AuxClim datastructure with the simple climatology
    ! proposed by Gert-Jan
    TYPE(AuxClim_DataType), intent(inout) :: AuxClim
    integer,                intent(out) :: error_flag

    ! local variables
    TYPE(ClimData_type), pointer :: ClimData
    integer(IntAs)     :: i_dt, i_lat, i_lon, i_alt, timestep_months
    real(r8_)          :: latstep, lonstep
    real(r8_)          :: lat1, lon1, lat2, lon2

    error_flag = no_error

    IF (.not. associated(AuxClim%AuxClim_DS)) THEN
       print *,"ERROR in Construct_Test_Climatology:"
       print *,"pointer AuxClim%AuxClim_DS seems not associated"
       error_flag = error_programming
       return
    END IF

    timestep_months = 12_IntAs/AuxClim%AuxClim_DS%Num_DateTime_Ranges
    DO i_dt = 1,AuxClim%AuxClim_DS%Num_DateTime_Ranges
       ! remark: the year should be ingnored when using these dates
       call SetDateTime(AuxClim%AuxClim_DS%ClimLatLonAlt(i_dt)%StartDateTime,&
                        2000,01+timestep_months*(i_dt-1),01,0,0,0.0_r8_)
       IF (i_dt .lt. AuxClim%AuxClim_DS%Num_DateTime_Ranges) THEN
          call SetDateTime(AuxClim%AuxClim_DS%ClimLatLonAlt(i_dt)%EndDateTime,&
                           2000,01+timestep_months* i_dt,   01,0,0,0.0_r8_)
       ELSE
          call SetDateTime(AuxClim%AuxClim_DS%ClimLatLonAlt(i_dt)%EndDateTime,&
                           2001,1,1,0,0,0.0_r8_)
       END IF
       print *,"i_dt, timestep_months = ",i_dt,timestep_months
       latstep = 180./AuxClim%AuxClim_DS%ClimLatLonAlt(i_dt)%&
                              Num_Latitude_Ranges
       DO i_lat=1,AuxClim%AuxClim_DS%ClimLatLonAlt(i_dt)%Num_Latitude_Ranges
          lat1 = -90._r8_ + latstep*(i_lat-1)
          lat2 = -90._r8_ + latstep* i_lat
          !print *,"i_lat,lat1,lat2 = ",i_lat,lat1,lat2
          !print *,"int(lat1*1.e6,IntAl) = ",int(lat1*1.e6,IntAl)
          AuxClim%AuxClim_DS%ClimLatLonAlt(i_dt)%&
               ClimLonAlt(i_lat)%StartLatitude = int(lat1*1.e6,IntAl)
          AuxClim%AuxClim_DS%ClimLatLonAlt(i_dt)%&
               ClimLonAlt(i_lat)%EndLatitude   = int(lat2*1.e6,IntAl)
          lonstep = 360./AuxClim%AuxClim_DS%ClimLatLonAlt(i_dt)%&
               ClimLonAlt(i_lat)%Num_Longitude_Ranges
          DO i_lon=1,AuxClim%AuxClim_DS%ClimLatLonAlt(i_dt)%&
               ClimLonAlt(i_lat)%Num_Longitude_Ranges
             lon1 = -180._r8_ + lonstep*(i_lon-1)
             lon2 = -180._r8_ + lonstep* i_lon
             !print *,"i_lon,lon1,lon2 = ",i_lon,lon1,lon2
             AuxClim%AuxClim_DS%ClimLatLonAlt(i_dt)%ClimLonAlt(i_lat)%&
                  ClimAlt(i_lon)%StartLongitude = int(lon1*1.e6,IntAl)
             AuxClim%AuxClim_DS%ClimLatLonAlt(i_dt)%ClimLonAlt(i_lat)%&
                  ClimAlt(i_lon)%EndLongitude   = int(lon2*1.e6,IntAl)
             DO i_alt=1,AuxClim%AuxClim_DS%ClimLatLonAlt(i_dt)%&
                  ClimLonAlt(i_lat)%ClimAlt(i_lon)%Num_Altitude_Ranges
                ! See Gert-Jan's email of 14-Jul-2006
                !               alt_range     S  stdev
                ! - PBL       :  0 -  3 km : (40, 15)
                ! - low_TROP  :  3 -  6 km : (40, 15)
                ! - UTLS      :  6 - 10 km : (40, 10)
                ! - STRAT     : 10 - 30 km : (30, 10)
                ClimData => AuxClim%AuxClim_DS%ClimLatLonAlt(i_dt)%&
                     ClimLonAlt(i_lat)%ClimAlt(i_lon)%ClimData(i_alt)
                select case(i_alt)
                case(1) ! PBL
                   ClimData%StartAltitude = int(0.e3,IntAl)
                   ClimData%EndAltitude   = int(3.e3,IntAl)
                   ClimData%S             = int(40.*1.e3,IntAl)
                   ClimData%S_stdev       = int(15.*1.e3,IntAl)
                case(2) ! low_TROP
                   ClimData%StartAltitude = int(3.e3,IntAl)
                   ClimData%EndAltitude   = int(6.e3,IntAl)
                   ClimData%S             = int(40.*1.e3,IntAl)
                   ClimData%S_stdev       = int(15.*1.e3,IntAl)
                case(3) ! PBL
                   ClimData%StartAltitude = int(6.e3,IntAl)
                   ClimData%EndAltitude   = int(10.e3,IntAl)
                   ClimData%S             = int(40.*1.e3,IntAl)
                   ClimData%S_stdev       = int(10.*1.e3,IntAl)
                case(4) ! PBL
                   ClimData%StartAltitude = int(10.e3,IntAl)
                   ClimData%EndAltitude   = int(30.e3,IntAl)
                   ClimData%S             = int(30.*1.e3,IntAl)
                   ClimData%S_stdev       = int(10.*1.e3,IntAl)
                case default
                   print *,"Error: only 4 altitude layers are defined"
                   print *,"by gert-Jan in his first climatology ..."
                   error_flag = error_programming
                   return
                end select
             END DO
          END DO
       END DO
    END DO

  end subroutine Construct_Test_Climatology
    !  #]
  !---------------------------------------------
end module AuxClim_DataStructure
