module AuxClim_SPH
  !  #[ Documentation
  !
  ! this module defines the Specific Product Header for the ee_cfi fileformat
  ! for the L2B processor AuxClim look-up table files
  ! The SPH format may differ for different filedefinitions, therefore
  ! we use a private definition for this filetype only
  !
  ! for a description, see 
  ! TN2.1, sec.6.3.1, table 64
  !
  ! Written by J. de Kloe
  !
  ! History:
  !      21-Jul-2006  J. de Kloe   Initial version
  !      31-Jul-2006  J. de Kloe   added GetFilePointer() function
  !      09-Mar-2007  D. Tan       Use ee_sph_name.
  !      24-Apr-2007  J. de Kloe   phase out the SummaryVars code
  !      07-Jun-2007  J. de Kloe   adapted to changed xml interface
  !      16-Jan-2008  J. de Kloe   phase out integer kind i_
  !
  !  #] 
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE XML_module, only: xml_node_type, &
                    max_str_len, xml_node_get_string, xml_set_fixed_header, &
                    check_length_xml_item, check_string_length, &
                    xml_tree_set_string_node_value, xml_tree_add_child, &
                    ee_vh_name, ee_sph_name, xml_add_SPH_item, &
                    xml_tree_add_and_set_attribute
  USE EE_CFI_Datatypes, only: IntAs,  maxlen_IntAs,  SetIntAs,  &
                              IntAl,  maxlen_IntAl,  SetIntAl
  USE DataSetDescriptor, only: DSD_type, Init_DSD, Print_DSD, &
                               Add_DSD_to_xml_SPH, Write_DSD_to_KVT, &
                               Get_DSD_from_xml_tree, GetByteSize_DSD_KVT, &
                               Read_DSD_from_KVT, &
                               maxlen_DS_Name, maxlen_DS_Type
  USE KVT_module, only: write_kvt_string, write_spare_kvt_string, &
                        read_kvt_string, read_spare_kvt_string, &
                        Compare_XML_to_KVT_item_Repl
  USE DirectBinaryIO, only: DirectBinaryIO_fileunit_type, &
                            GetFilePointer
  ! support modules
  USE ErrorHandler, only: no_error, error_allocate, error_programming, &
                          error_reading_file, error_writing_file
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE Numerics, only: i8_
  USE Compiler_Features, only: newline
  USE LunManager, only: fileunit_stdout
  !  #]
  !  #[ Type definitions and parameters
  implicit none

  public

  character(len=maxlen_DS_Name) ,parameter :: AUXCLIM_DS_Name_AuxClim_ADS=&
                                     "AuxClim_ADS"
  character(len=maxlen_DS_Type) ,parameter :: AUXCLIM_DS_Type_AuxClim_ADS="A"

  integer, parameter :: maxlen_SPH_Descriptor   = 28
  integer, parameter :: maxlen_AuxClim_Ref_Name = 50

  TYPE AuxClim_SPH_type
     !  #[
     character(len=maxlen_SPH_Descriptor)   :: SPH_Descriptor
     ! Spare 1
     character(len=maxlen_AuxClim_Ref_Name) :: AuxClim_Ref_Name
     ! Spare 2
     TYPE(DSD_type), pointer :: AuxClim_DSD
     !
  END TYPE AuxClim_SPH_type
    !  #]

  ! the default verbosity setting is Err (as silent as possible)
  ! (this switch is not yet set by an init routine, and can be set 
  !  to VerbosityReport manually for debugging purposes)
  integer, save :: LocalVerbosity = VerbosityError
!  integer, save :: LocalVerbosity = VerbosityReport
  
  !  #]
  !  #[ Module interface
  ! all is public for the moment, unless specified explicitely otherwise
  !  #]
contains
  !---------------------------------
  subroutine Init_AuxClim_SPH(AuxClim_SPH,error_flag)
    !  #[
    TYPE(AuxClim_SPH_type), pointer :: AuxClim_SPH
    integer, intent(out) :: error_flag
    
    error_flag = no_error
    IF (.not. associated(AuxClim_SPH)) THEN
       print *,"ERROR in Init_AuxClim_SPH():"
       print *,"The pointer to this AuxClim_SPH seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! see the remark in MainProductHeader.F90 for the reason why I use
    ! the form x(:)=' ' to initialise the strings

    AuxClim_SPH%SPH_Descriptor(:) = ' '
    ! Spare 1
    AuxClim_SPH%AuxClim_Ref_Name(:)  = ' '
    ! Spare 2
    nullify(AuxClim_SPH%AuxClim_DSD)

  end subroutine Init_AuxClim_SPH
    !  #]
  subroutine Print_AuxClim_SPH(AuxClim_SPH,error_flag,prefix)
    !  #[
    TYPE(AuxClim_SPH_type), pointer     :: AuxClim_SPH
    integer,                intent(out) :: error_flag
    character(len=*),       intent(in)  :: prefix

    call AsciiWrite_AuxClim_SPH(AuxClim_SPH,fileunit_stdout,prefix,error_flag)

  end subroutine Print_AuxClim_SPH
    !  #]
  subroutine AsciiWrite_AuxClim_SPH(AuxClim_SPH,fileunit,prefix,error_flag)
    !  #[
    TYPE(AuxClim_SPH_type), pointer     :: AuxClim_SPH ! input
    integer,                intent(in)  :: fileunit    ! input
    character(len=*),       intent(in)  :: prefix      ! input
    integer,                intent(out) :: error_flag  ! output

    ! local variable
    character(len=256) :: txt

    error_flag = no_error

    IF (.not. associated(AuxClim_SPH)) THEN
       write(fileunit,"(a)",err=999) trim(prefix)//" [seems not associated]"
       return
    END IF

    txt = AuxClim_SPH%SPH_Descriptor
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Sph_Descriptor   = "//trim(txt)

    ! Spare 1

    txt = AuxClim_SPH%AuxClim_Ref_Name
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%AuxClim_Ref_Name = "//trim(txt)

    ! Spare 2

    call Print_DSD(AuxClim_SPH%AuxClim_DSD,&
                   trim(prefix)//"%AuxClim_DSD",error_flag)
    if (error_flag .ne. no_error) return

    return

999 print *,"ERROR in AsciiWrite_AuXClim_SPH:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_AuxClim_SPH
    !  #]
  subroutine Delete_AuxClim_SPH(AuxClim_SPH)
    !  #[
    TYPE(AuxClim_SPH_type), pointer :: AuxClim_SPH

    IF (associated(AuxClim_SPH%AuxClim_DSD)) &
         deallocate(AuxClim_SPH%AuxClim_DSD)

  end subroutine Delete_AuxClim_SPH
    !  #]
  subroutine Fill_AuxClim_SPH_items(AuxClim_SPH, error_flag, &
                                    SPH_Descriptor, AuxClim_Ref_Name)
    !  #[
    ! Remark: the DSD structure is not yet filled inside this routine
    ! To get the needed sizes the GetByteSize functions are used,
    ! this is done in the WriteAuxClimData module.
    ! If we would do it here, the SPH module would become dependant
    ! on the AuxClim_Datastructure module, which in turn is dependent on this
    ! SPH module, and this cyclic dependency cannot be properly compiled.

    TYPE(AuxClim_SPH_type), pointer     :: AuxClim_SPH     ! filled on output
    integer,                intent(out) :: error_flag     ! result
    character(len=*),       intent(in)  :: SPH_Descriptor
    character(len=*),       intent(in)  :: AuxClim_Ref_Name

    error_flag = no_error

    IF (.not. associated(AuxClim_SPH)) THEN
       print *,"ERROR in Fill_AuxClim_SPH_items():"
       print *,"The pointer to this AuxClim_SPH seems not associated !!"
       error_flag = error_programming
       return
    END IF

    call  check_string_length(SPH_Descriptor,"Sph_Descriptor",&
         maxlen_SPH_Descriptor,"Fill_AuxClim_SPH_items") 
    AuxClim_SPH%SPH_Descriptor = SPH_Descriptor

    call  check_string_length(AuxClim_Ref_Name,"AuxClim_Ref_Name",&
         maxlen_AuxClim_Ref_Name,"Fill_AuxClim_SPH_items") 
    AuxClim_SPH%AuxClim_Ref_Name = AuxClim_Ref_Name

    ! the DSD is not filled in this subroutine
    
  end subroutine Fill_AuxClim_SPH_items
    !  #]
  !---------------------------------
  subroutine Copy_AuxClim_SPH_to_xml_tree(AuxClim_SPH, path_in, &
                                          xml_tree, error_flag)
    !  #[
    TYPE(AuxClim_SPH_type), pointer    :: AuxClim_SPH ! input
    character(len=*),       intent(in) :: path_in     ! input
    TYPE(xml_node_type),    pointer    :: xml_tree   ! points to updated struct
    integer,                intent(out):: error_flag ! output

    ! local variables
    character(len=max_str_len) :: path
    character(len=max_str_len) :: path_dsd

    error_flag = no_error

    ! add the <Specific_Product_Header></Specific_Product_Header> tags
    path = trim(path_in)//"/"//ee_vh_name
    call xml_tree_add_child(xml_tree,path,ee_sph_name,error_flag)
    IF (error_flag .ne. no_error) return

    ! fill all items inside the AuxClim_SPH
    call xml_add_SPH_item(xml_tree, path_in, "Sph_Descriptor",&
                          AuxClim_SPH%SPH_Descriptor,error_flag)
    IF (error_flag .ne. no_error) return

    ! Spare_1
    call xml_add_SPH_item(xml_tree, path_in, "Spare_1",newline,error_flag)
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "AuxClim_Ref_Name",&
                          AuxClim_SPH%AuxClim_Ref_Name,error_flag)
    IF (error_flag .ne. no_error) return

    ! Spare_2
    call xml_add_SPH_item(xml_tree, path_in, "Spare_2",newline,error_flag)
    IF (error_flag .ne. no_error) return

    ! add the <List_of_DSDs></List_of_DSDs> tags
    path = trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name
    call xml_tree_add_child(xml_tree,path,"List_of_Dsds",error_flag)
    IF (error_flag .ne. no_error) return

    path_dsd = trim(path)//"/List_of_Dsds"
    call xml_tree_add_and_set_attribute(xml_tree,path_dsd,"count","1",&
                                        error_flag)

    call Add_DSD_to_xml_SPH(xml_tree, path_in, &
                            AuxClim_SPH%AuxClim_DSD, 1, error_flag)
    IF (error_flag .ne. no_error) return

    return

  end subroutine Copy_AuxClim_SPH_to_xml_tree
    !  #]
  subroutine Get_AuxClim_SPH_from_xml_tree(xml_tree, path_in, &
                                           AuxClim_SPH, error_flag)
    !  #[
    TYPE(xml_node_type),     pointer     :: xml_tree    ! input
    character(len=*),        intent(in)  :: path_in     ! input
    TYPE(AuxClim_SPH_type),  pointer     :: AuxClim_SPH ! result
    integer,                 intent(out) :: error_flag  ! output

    ! local variables
    character(len=max_str_len) :: path
    character(len=max_str_len) :: text
    integer                    :: AllocStatus

    ! init
    error_flag = no_error

    IF (.not. associated(AuxClim_SPH)) THEN
       print *,"ERROR in Get_AuxClim_SPH_from_xml_tree():"
       print *,"the pointer AuxClim_SPH seems not associated !!!"
       error_flag = error_programming
       return
    END IF

    IF (.not. associated(xml_tree)) THEN
       print *,"ERROR in Get_AuxClim_SPH_from_xml_tree():"
       print *,"the pointer xml_tree seems not associated !!!"
       error_flag = error_programming
       return
    END IF

    !-------------------------------------
    ! copying the AuxClim_SPH XML elements
    !-------------------------------------

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Sph_Descriptor"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    AuxClim_SPH%SPH_Descriptor = &
         check_length_xml_item(text,path,maxlen_SPH_Descriptor,&
                               "Get_AuxClim_SPH_from_xml_tree")

    ! Spare_1: no need to read this one

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/AuxClim_Ref_Name"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    AuxClim_SPH%AuxClim_Ref_Name = &
         check_length_xml_item(text,path,maxlen_AuxClim_Ref_Name,&
                               "Get_AuxClim_SPH_from_xml_tree")

    ! Spare_2: no need to read this one

    allocate(AuxClim_SPH%AuxClim_DSD,STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       error_flag = error_allocate
       return
    END IF

    call Init_DSD(AuxClim_SPH%AuxClim_DSD,error_flag)
    IF (error_flag .ne. no_error) return

    ! get the AuxClim_DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree, path_in, &
                               AuxClim_SPH%AuxClim_DSD,1,&
                               AUXCLIM_DS_Name_AuxClim_ADS, &
                               AUXCLIM_DS_Type_AuxClim_ADS, error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine Get_AuxClim_SPH_from_xml_tree
    !  #]
  !---------------------------------
  subroutine Write_AuxClim_SPH_to_KVT(AuxClim_SPH, fileunit, error_flag)
    !  #[
    ! Writes the AuxClim Specific Product Header (KVT values).

    TYPE(AuxClim_SPH_type), pointer     :: AuxClim_SPH
    TYPE(DirectBinaryIO_fileunit_type), intent(in)  :: fileunit
    integer,                intent(out) :: error_flag

    ! local variables
    character(len=64) :: varname
    character(len=64) :: value
    integer           :: max_value_length
    integer(i8_)      :: filepos_end,filepos_start
    integer(i8_)      :: nbytes_written, nbytes_expected

    !-------------------------------------
    ! Writing the AuxClim SPH KVT elements
    !-------------------------------------

    error_flag = no_error
    IF (.not. associated(AuxClim_SPH)) THEN
       print *,"ERROR in Write_AuxClim_SPH_to_KVT():"
       print *,"The pointer to this AuxClim_SPH seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! as double check, we compare the number of bytes written
    ! to what was expected from the GetByteSize_AuxClim_SPH_KVT() function
    ! so store the filepointer before we write anything
    call GetFilePointer(fileunit,filepos_start,error_flag)
    IF (error_flag .ne. no_error) goto 998
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"in Write_AuxClim_SPH_to_KVT: filepos_start = ",filepos_start

    ! SPH_Descriptor 
    varname          = "Sph_Descriptor"
    value            = AuxClim_SPH%SPH_Descriptor
    max_value_length = maxlen_SPH_Descriptor
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag,&
                          use_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998
   
    ! Spare_1: this is a spare value without content
    max_value_length = 40
    call write_spare_kvt_string(fileunit,max_value_length,error_flag)
    IF (error_flag .ne. no_error) return

    ! AuxClim_Ref_Name
    varname          = "AuxClim_Ref_Name"
    value            = AuxClim_SPH%AuxClim_Ref_Name
    max_value_length = maxlen_AuxClim_Ref_Name
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag,&
                          use_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998

    ! Spare_2: this is a spare value without content
    max_value_length = 40
    call write_spare_kvt_string(fileunit,max_value_length,error_flag)
    IF (error_flag .ne. no_error) return

    ! Write RBC_DSD
    call Write_DSD_to_KVT(AuxClim_SPH%AuxClim_DSD,fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! as double check, compare the number of bytes written
    ! to what was expected from the GetByteSize_AuxClim_SPH_KVT() function
    call GetFilePointer(fileunit,filepos_end,error_flag)
    IF (error_flag .ne. no_error) goto 998
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"in Write_AuxClim_SPH_to_KVT: filepos_end = ",filepos_end

    nbytes_written  = filepos_end-filepos_start
    nbytes_expected = GetByteSize_AuxClim_SPH_KVT()

    IF (nbytes_written .ne. nbytes_expected) THEN
       print *,"ERROR in Write_AuxClim_SPH_to_KVT():"
       print *,"nbytes_written  = ",nbytes_written
       print *,"nbytes_expected = ",nbytes_expected
       print *,"something went wrong when writing the AuxClim_SPH to the KVT"
       print *,"part of the binary file, probably a programming error ..."
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Write_AuxClim_SPH_to_KVT"
    return

  end subroutine Write_AuxClim_SPH_to_KVT
    !  #]
  subroutine Read_AuxClim_SPH_from_KVT(AuxClim_SPH,fileunit,error_flag,Compare)
    !  #[
    ! Reads the AuxClim Specific Product Header (KVT values).

    ! if Compare=.true. then
    ! assume the values are already read from the XML file
    ! so read and compare them, and then replace the ones in memory by
    ! those read, because those read are in KVT format, which is the
    ! convention of choice.

    TYPE(AuxClim_SPH_type), pointer     :: AuxClim_SPH
    integer,                intent(in)  :: fileunit
    integer,                intent(out) :: error_flag
    logical, optional,      intent(in)  :: Compare

    ! local variables
    character(len=max_str_len) :: exp_name
    character(len=max_str_len) :: value
    integer                    :: exp_val_length
    integer                    :: AllocStatus
    logical                    :: do_Compare

    !-------------------------------------
    ! Reading the AuxClim SPH KVT elements
    !-------------------------------------

    error_flag = no_error
    IF (.not. associated(AuxClim_SPH)) THEN
       print *,"ERROR in Read_AuxClim_SPH_from_KVT():"
       print *,"The pointer to this AuxClim_SPH seems not associated !!"
       error_flag = error_programming
       return
    END IF

    do_Compare = .false.
    IF (present(Compare)) do_Compare = Compare

    IF (do_Compare) THEN
       print *,"Read_AuxClim_SPH_from_KVT:"//&
               "reading and comparing AuxClim SPH-KVT items"
       print *,"with the XML items in memory"
    ELSE
       print *,"Read_AuxClim_SPH_from_KVT: reading AuxClim SPH-KVT items "
    END IF

    ! read the SPH_Descriptor 
    exp_val_length = len(AuxClim_SPH%SPH_Descriptor)
    exp_name       = "Sph_Descriptor"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expect_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(AuxClim_SPH%SPH_Descriptor, value, &
                                    "AuxClim_SPH%Sph_Descriptor",error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       AuxClim_SPH%SPH_Descriptor = value
    END IF

    ! Spare_1: this is a spare value without content
    exp_val_length = 40
    call read_spare_kvt_string(fileunit,exp_val_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! read the AuxClim_Ref_Name
    exp_val_length = len(AuxClim_SPH%AuxClim_Ref_Name)
    exp_name       = "AuxClim_Ref_Name"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expect_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(AuxClim_SPH%AuxClim_Ref_Name, value, &
                                    "AuxClim_SPH%AuxClim_Ref_Name",error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       AuxClim_SPH%AuxClim_Ref_Name = value
    END IF

    ! Spare_2: this is a spare value without content
    exp_val_length = 40
    call read_spare_kvt_string(fileunit,exp_val_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    IF (.not. associated(AuxClim_SPH%AuxClim_DSD)) THEN
       ! allow this, to enable the possibility of reading the dbl file only,
       ! without reading the HDR files.
       allocate(AuxClim_SPH%AuxClim_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine "//&
                  "Read_AuxClim_SPH_from_KVT()"
          error_flag = error_allocate
          goto 998
       END IF
       call Init_DSD(AuxClim_SPH%AuxClim_DSD,error_flag)
       IF (error_flag .ne. no_error) goto 998

       ! Read the AuxClim_DSD without comparing
       call Read_DSD_from_KVT(AuxClim_SPH%AuxClim_DSD,fileunit,&
                              "AuxClim_ADS","A",error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the RBC_DSD and compare if needed
       call Read_DSD_from_KVT(AuxClim_SPH%AuxClim_DSD,fileunit,&
                              "AuxClim_ADS","A",error_flag,Compare)
       IF (error_flag .ne. no_error) goto 998
    END IF

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Read_AuxClim_SPH_from_KVT"
    return

  end subroutine Read_AuxClim_SPH_from_KVT
    !  #]
  !---------------------------------
  function GetByteSize_AuxClim_SPH_KVT() result(KVT_bytesize)
    !  #[
    integer(IntAl) :: KVT_bytesize
    
    ! KVT sizes have size(key)+size(=)+size(value)+size(terminator)
    KVT_bytesize = &
         !     len(key)          len(=)  len(value)     len(unit)   len(terminator)
         !                                           or len(quotes)
         int(len("Sph_Descriptor")   + 1 + maxlen_SPH_Descriptor    + 2   + 1 + &
             40+1+& ! spare_1
             len("AuxClim_Ref_Name") + 1 + maxlen_AuxClim_Ref_Name  + 2   + 1 + &
             40+1+& ! spare_2
             GetByteSize_DSD_KVT(),IntAl)
    
    !print *,"tests lengths of items in AuxClim_SPH KVT part"
    !print *,len("Sph_Descriptor")   + 1 + maxlen_SPH_Descriptor   +2+1 
    !print *,40+1
    !print *,len("AuxClim_Ref_Name") + 1 + maxlen_AuxClim_Ref_Name +2+1
    !print *,40+1
    !print *,GetByteSize_DSD_KVT()
    
  end function GetByteSize_AuxClim_SPH_KVT
  !  #]
  !---------------------------------
end module AUXCLIM_SPH
