! TODO: according to the Envisat product specification (which defines 
!       missing values for not used items in the MPH/SPH/DSD)
!       the FILENAME field in the DSD MAYBE SET to "NOT USED"
!       so lets just do this as a clean-up action. JK, 2-mar-2007

module DataSetDescriptor
  !  #[ Documentation

  !
  ! this module defines the DataSetDescriptors for the ee_cfi fileformat
  !
  ! the next type definition is generic for aeolus files
  ! and is used for both the L1B and L2B file headers
  !
  ! for a description, see section 3.3.3 of the L1B_iodd_2.1
  ! or TN2.1, version 1.2, (6-9-2005), sec.4.6, p.17, table 11
  !
  ! Written by J. de Kloe (KNMI)
  !
  ! Modifications:
  !    01-Dec-2005  P. Poli    Added routines Fill_L2B_DSD_items,
  !                            Write_DSD_to_KVT
  !    21-Feb-2006  J. de Kloe Get_DSD_from_xml_tree, Write_DSD_to_KVT_Jos 
  !    28-Feb-2006  J. de Kloe added prefix parameter to print routine
  !    01-Mar-2006  J. de Kloe added routines Read_DSD_from_KVT
  !    02-Mar-2006  J. de Kloe adapted to changes in XML_module, and added
  !                            Compare switch to the Read_from_KVT routine
  !    13-Mar-2006  J. de Kloe Added the new AsciiWrite routine
  !    24-Mar-2006  D. Tan     added name and type for RBC FP_MDS
  !    30-Mar-2006  J. de Kloe Bug fix for Ds_offset.
  !    30-Mar-2006  J. de Kloe Adapted Init_DSD to use -1 as missing indicator
  !                            (can be replaced later by NMDI if needed)
  !    18-Apr-2006  J. de Kloe Added the missing_indicator_bytes* parameters
  !    12-May-2006  D. Tan     Null use of ParKind1 module
  !    17-Jul-2006  D. Tan     added name and type for L1B_Product and 
  !                            CLM_Product, revise names of L2B DataSets
  !    31-Jul-2006  J. de Kloe made Write_DSD_to_KVT_Jos_buff generic
  !    17-Aug-2006  P. Poli    removed obsolete variables
  !    27-Oct-2006  D. Tan     Name and type for L2C-related datasets
  !    15-Nov-2006  J. de Kloe moved some parameters to l2bc_sph
  !    15-Mar-2007  J. de Kloe remove usage of SummaryVars
  !    24-Apr-2007  J. de Kloe phase out the SummaryVars code
  !    07-Jun-2007  J. de Kloe adapted to changed xml interface
  !    13-Jun-2007  J. de Kloe adapted Fill_DSD_items to remove a possible
  !                            path that might be present in the Filename field
  !    08-Jan-2008  J. de Kloe added some explicit truncations for string 
  !                            copying, to suppress warnings for gfortran
  !    16-Jan-2008  J. de Kloe phase out integer kind i_ 
  !    18-Jan-2008  J. de Kloe implement workaround to adapt to the changed 
  !                            DSD tagname in L1B v1.08 products
  !
  !  #] 
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE XML_module, only: xml_node_type, max_str_len, &
                        ee_vh_name, ee_sph_name, &
                        xml_tree_set_string_node_value, xml_tree_add_child, &
                        xml_tree_read_string_attribute, check_length_xml_item,&
                        xml_node_get_string
  USE EE_CFI_Datatypes, only: maxlen_IntAl, maxlen_IntAd, &
                              SetIntAl, SetIntAd, GetIntAd, GetIntAl, &
                              IntAl, IntAd
  USE KVT_module, only: write_kvt_string, write_spare_kvt_string, &
                        read_kvt_string, read_spare_kvt_string, &
                        Compare_XML_to_KVT_item_Repl
  USE Buffering, only: BufferedFile_type
  USE DirectBinaryIO, only: DirectBinaryIO_fileunit_type, GetFilePointer, &
                            Init_DirectBinaryIO_fileunit

  ! support modules
  USE Logging, only: logmsg,log_debug,log_info,log_progress,log_warn,log_error
  USE ErrorHandler, only: no_error, error_programming, &
                          error_reading_file !, error_allocate
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE Numerics, only: i8_
  USE LunManager, only: fileunit_stdout
  USE Compiler_Features, only: newline
  USE c_support, only: get_filesize
  USE StringTools, only: Split_path_and_filename

  !  #]
  !  #[ Type definitions and parameters
  implicit none

  integer, parameter :: maxlen_DS_Name    = 28
  integer, parameter :: maxlen_DS_Type    =  1
  integer, parameter :: maxlen_Filename   = 62
  integer, parameter :: maxlen_Byte_Order =  4

  ! this missing indicator is to be used inside this DSD module only
  ! I use -1 as missing value because this value is impossible
  ! for dataset/filesizes and offsets.
  ! Note that this mising indicator CAN be written to the outputfiles
  ! since the datatypes used are signed integers ......
  integer, parameter :: missing_indicator_bytes = -1
  character(len=*), parameter :: missing_indicator_bytes_IntAd = &
                                                       "-00000000000000000001"
  character(len=*), parameter :: missing_indicator_bytes_IntAl = "-0000000001"

  character(len=maxlen_Byte_Order) :: L2B_DS_Byte_Order="3210"

  TYPE DSD_type ! see section 3.3.3 of the L1B_iodd_2.1 (table 3.8)
     !  #[ data set descriptor
     character(len=maxlen_DS_Name)    :: DS_Name
     character(len=maxlen_DS_Type)    :: DS_Type
     character(len=maxlen_Filename)   :: Filename
     character(len=maxlen_IntAd)      :: DS_Offset
     character(len=maxlen_IntAl)      :: DS_Size
     character(len=maxlen_IntAl)      :: NUM_DSR
     character(len=maxlen_IntAl)      :: DSR_Size
     character(len=maxlen_Byte_Order) :: Byte_Order
     !character(len=2) :: Spare_1
     !
     ! the next items are numerical values derived from the stringvalues above
     integer(IntAd) :: start_of_ds            ! = DS_Offset + 1
     integer(IntAd) :: end_of_ds              ! = DS_Offset+DS_SIZE
     integer(IntAl) :: nr_of_dataset_records  ! = NUM_DSR
     integer(IntAl) :: size_of_one_dsr        ! = DSR_SIZE
     integer(IntAl) :: size_of_ds             ! = DS_SIZE
  END TYPE DSD_type
     !  #]

  ! the default verbosity setting is Err (as silent as possible)
  ! (this switch is not yet set by an init routine, and can be set 
  !  to VerbosityReport manually for debugging purposes)
!  integer, save :: LocalVerbosity = VerbosityError
  integer, save :: LocalVerbosity = VerbosityReport

  !  #]
  !  #[ interfaces for overloading functions
  ! TODO: phase out use of Print_DSD_old
!  INTERFACE Print_DSD
!     module procedure Print_DSD_old, Print_DSD_new
!  END INTERFACE
  ! TODO: Then phase out use of 
  !       Write_DSD_to_KVT_Jos_unbuff and Write_DSD_to_KVT_Jos_buff. 
  !       After that rename Write_DSD_to_KVT_Jos_gen to Write_DSD_to_KVT
  INTERFACE Write_DSD_to_KVT
     module procedure Write_DSD_to_KVT_unbuff, & ! old direct type
                      Write_DSD_to_KVT_buff,   & ! old buffered type
                      Write_DSD_to_KVT_gen       ! new generic type
  END INTERFACE
  !  #]
  !  #[ Module interface
  ! Type definitions
  public :: DSD_type  
  ! subroutines
  public :: Init_DSD, Print_DSD, Fill_DSD_items, &
            Write_DSD_to_KVT, Add_DSD_to_xml_SPH, Get_DSD_from_xml_tree, &
            GetByteSize_DSD_KVT
  !  #]
contains
  !-------------------------
  subroutine Init_DSD(DSD, error_flag)
    !  #[
    TYPE(DSD_type), pointer :: DSD
    integer, intent(out) :: error_flag
    
    error_flag = no_error
    IF (.not. associated(DSD)) THEN
       print *,"ERROR in Init_DSD():"
       print *,"The pointer to this DSD seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! see the remark in MainProductHeader.F90 for the reason why I use
    ! the form x(:)=' ' to initialise the strings

    DSD%DS_NAME(:)    = ' '
    DSD%DS_TYPE(:)    = ' '
    DSD%FILENAME(:)   = ' '
    DSD%DS_OFFSET(:)  = ' '
    DSD%DS_SIZE(:)    = ' '
    DSD%NUM_DSR(:)    = ' '
    DSD%DSR_SIZE(:)   = ' '
    DSD%BYTE_ORDER(:) = ' '
    !DSD%Spare_1

    DSD%start_of_ds            = missing_indicator_bytes
    DSD%end_of_ds              = missing_indicator_bytes
    DSD%nr_of_dataset_records  = missing_indicator_bytes
    DSD%size_of_one_dsr        = missing_indicator_bytes
    DSD%size_of_ds             = missing_indicator_bytes

  end subroutine Init_DSD
    !  #]
  function GetByteSize_DSD_KVT() result(KVT_bytesize)
    !  #[
    integer(IntAl) :: KVT_bytesize
    
    ! KVT sizes have size(key)+size(=)+size(value)+size(terminator)
    KVT_bytesize = &
         ! len(key)      len(=)  len(value)         len(unit)   len(terminator)
         !                                       or len(quotes)
         len("DS_Name")    + 1 + maxlen_DS_Name    + 2              + 1 + &
         len("DS_Type")    + 1 + maxlen_DS_Type                     + 1 + &
         len("Filename")   + 1 + maxlen_Filename   + 2              + 1 + &
         len("DS_Offset")  + 1 + maxlen_IntAd      + len("<bytes>") + 1 + &
         len("DS_Size")    + 1 + maxlen_IntAl                       + 1 + &
         len("NUM_DSR")    + 1 + maxlen_IntAl      + len("<bytes>") + 1 + &
         len("DSR_Size")   + 1 + maxlen_IntAl      + len("<bytes>") + 1 + &
         len("Byte_Order") + 1 + maxlen_Byte_Order + 2              + 1 + &
         32+1 ! spare_1
                   
  end function GetByteSize_DSD_KVT
    !  #]
  subroutine Print_DSD(DSD,prefix,error_flag)
    !  #[
    TYPE(DSD_type),   pointer     :: DSD
    character(len=*), intent(in)  :: prefix
    integer,          intent(out) :: error_flag

    ! init
    error_flag = no_error

    call AsciiWrite_DSD(DSD,fileunit_stdout,prefix,error_flag)

  end subroutine Print_DSD
    !  #]
  subroutine AsciiWrite_DSD(DSD,fileunit,prefix,error_flag)
    !  #[
    TYPE(DSD_type),   pointer     :: DSD
    integer,          intent(in)  :: fileunit
    character(len=*), intent(in)  :: prefix
    integer,          intent(out) :: error_flag

    ! local variable
    character(len=256) :: txt
    character(len=maxlen_IntAd)      :: txt_IntAd
    character(len=maxlen_IntAl)      :: txt_IntAl

    error_flag = no_error

    IF (.not. associated(DSD)) THEN
       write(fileunit,"(a)") trim(prefix)//" [seems not associated]"
       return
    END IF

    txt = DSD%DS_NAME
    IF (len_trim(txt).eq.0) txt = "[empty]"

    write(fileunit,"(a)") trim(prefix)//"%DS_NAME                = "//trim(txt)

    txt = DSD%DS_TYPE
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") trim(prefix)//"%DS_TYPE                = "//trim(txt)

    txt = DSD%FILENAME
    IF (len_trim(txt).eq.0) txt = "<THIS DSD HAS NO FILE>" 
    write(fileunit,"(a)") trim(prefix)//"%FILENAME               = "//trim(txt)

    txt = DSD%DS_OFFSET
    IF (len_trim(txt).eq.0) THEN ! this should never happen
       print *,"WARNING in AsciiWrite_DSD():"
       print *,"Empty string for %DS_OFFSET is not allowed for KVT output!"
       print *,"Allowing it for now, since we used this practice as missing"
       print *,"value for the last year ..."
       !error_flag = error_writing_file
       txt = "[empty]"
    END IF
    IF (trim(txt) .eq. missing_indicator_bytes_IntAd) txt = "[missing]"
    write(fileunit,"(a)") trim(prefix)//"%DS_OFFSET              = "//trim(txt)

    txt = DSD%DS_SIZE
    IF (len_trim(txt).eq.0) THEN ! this should never happen
       print *,"WARNING in AsciiWrite_DSD():"
       print *,"Empty string for %DS_SIZE is not allowed for KVT output!"
       print *,"Allowing it for now, since we used this practice as missing"
       print *,"value for the last year ..."
       !error_flag = error_writing_file
       txt = "[empty]"
    END IF
    IF (trim(txt) .eq. missing_indicator_bytes_IntAl) txt = "[missing]"
    write(fileunit,"(a)") trim(prefix)//"%DS_SIZE                = "//trim(txt)

    txt = DSD%NUM_DSR
    IF (len_trim(txt).eq.0) THEN ! this should never happen
       print *,"WARNING in AsciiWrite_DSD():"
       print *,"Empty string for %NUM_DSR is not allowed for KVT output!"
       print *,"Allowing it for now, since we used this practice as missing"
       print *,"value for the last year ..."
       !error_flag = error_writing_file
       txt = "[empty]"
    END IF
    IF (trim(txt) .eq. missing_indicator_bytes_IntAl) txt = "[missing]"
    write(fileunit,"(a)") trim(prefix)//"%NUM_DSR                = "//trim(txt)

    txt = DSD%DSR_SIZE
    IF (len_trim(txt).eq.0) THEN ! this should never happen
       print *,"WARNING in AsciiWrite_DSD():"
       print *,"Empty string for %DSR_SIZE is not allowed for KVT output!"
       print *,"Allowing it for now, since we used this practice as missing"
       print *,"value for the last year ..."
       !error_flag = error_writing_file
       txt = "[empty]"
    END IF
    IF (trim(txt) .eq. missing_indicator_bytes_IntAl) txt = "[missing]"
    write(fileunit,"(a)") trim(prefix)//"%DSR_SIZE               = "//trim(txt)

    txt = DSD%BYTE_ORDER
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") trim(prefix)//"%BYTE_ORDER             = "//trim(txt)

    ! Spare_1

    IF (DSD%start_of_ds .eq. missing_indicator_bytes) THEN
       txt_IntAd = "[missing]"
    ELSE
       call SetIntAd(DSD%start_of_ds,txt_IntAd,error_flag)
       IF (error_flag .ne. no_error) return
    END IF
    write(fileunit,"(a)") &
         trim(prefix)//"%start_of_ds            = "//trim(txt_IntAd)

    IF (DSD%end_of_ds .eq. missing_indicator_bytes) THEN
       txt_IntAd = "[missing]"
    ELSE
       call SetIntAd(DSD%end_of_ds,txt_IntAd,error_flag)
       IF (error_flag .ne. no_error) return
    END IF
    write(fileunit,"(a)") &
         trim(prefix)//"%end_of_ds              = "//trim(txt_IntAd)

    IF (DSD%nr_of_dataset_records .eq. missing_indicator_bytes) THEN
       txt_IntAl = "[missing]"
    ELSE
       call SetIntAl(DSD%nr_of_dataset_records,txt_IntAl,error_flag)
       IF (error_flag .ne. no_error) return
    END IF
    write(fileunit,"(a)") &
         trim(prefix)//"%nr_of_dataset_records  = "//trim(txt_IntAl)

    IF (DSD%size_of_one_dsr .eq. missing_indicator_bytes) THEN
       txt_IntAl = "[missing]"
    ELSE
       call SetIntAl(DSD%size_of_one_dsr,txt_IntAl,error_flag)
       IF (error_flag .ne. no_error) return
    END IF
    write(fileunit,"(a)") &
         trim(prefix)//"%size_of_one_dsr        = "//trim(txt_IntAl)

    IF (DSD%size_of_ds .eq. missing_indicator_bytes) THEN
       txt_IntAl = "[missing]"
    ELSE
       call SetIntAl(DSD%size_of_ds,txt_IntAl,error_flag)
       IF (error_flag .ne. no_error) return
    END IF
    write(fileunit,"(a)") &
         trim(prefix)//"%size_of_ds             = "//trim(txt_IntAl)

  end subroutine AsciiWrite_DSD
    !  #]
  !-------------------------
  subroutine Add_DSD_to_xml_SPH(xml_tree, path_in, DSD, which_DSD, error_flag)
    !  #[
    TYPE(xml_node_type), pointer     :: xml_tree   ! updated
    character(len=*),    intent(in)  :: path_in    ! input
    TYPE(DSD_type),      pointer     :: DSD        ! input
    integer,             intent(in)  :: which_DSD  ! input
    integer,             intent(out) :: error_flag ! output

    ! local variables
    character(len=max_str_len) :: path
    character(len=max_str_len) :: path_dsd
    character(len=max_str_len) :: item
    character(len=max_str_len) :: item_name
    character(len=2)           :: count

    ! initialise
    error_flag = no_error

    write(count,"(i2)") which_DSD
    path_dsd = trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/List_of_Dsds/Dsd["//trim(adjustl(count))//"]"
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"Writing xml data to: ",trim(path_dsd)

    ! add the <DSD></DSD> tags
    path = trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//"/List_of_Dsds"
    call xml_tree_add_child(xml_tree,path,"Dsd",error_flag)
    IF (error_flag .ne. no_error) return

    ! initialise, to use empty strings, in case the DSD pointer
    ! was not yet associated (i.e. writing an undefined DSD)
    item(:) = ' '

    item_name = "Ds_Name"
    if (associated(DSD)) item = DSD%Ds_Name
    ! add this item
    call xml_tree_add_child(xml_tree,path_dsd,item_name,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(path_dsd)//"/"//trim(item_name)
    call xml_tree_set_string_node_value(xml_tree,path,trim(adjustl(item)),&
                                        error_flag)
    IF (error_flag .ne. no_error) return

    item_name = "Ds_Type"
    if (associated(DSD)) item = DSD%Ds_Type
    ! add this item
    call xml_tree_add_child(xml_tree,path_dsd,item_name,error_flag)
    IF (error_flag .ne. no_error) return
    path = trim(path_dsd)//"/"//trim(item_name)
    call xml_tree_set_string_node_value(xml_tree,path,trim(adjustl(item)),&
                                        error_flag)
    IF (error_flag .ne. no_error) return

    item_name = "Filename"
    if (associated(DSD)) item = DSD%Filename
    ! add this item
    call xml_tree_add_child(xml_tree,path_dsd,item_name,error_flag)
    IF (error_flag .ne. no_error) return
    path = trim(path_dsd)//"/"//trim(item_name)
    call xml_tree_set_string_node_value(xml_tree,path,trim(adjustl(item)),&
                                        error_flag)
    IF (error_flag .ne. no_error) return

    item_name = "Ds_Offset"
    if (associated(DSD)) item = DSD%Ds_Offset
    ! add this item
    call xml_tree_add_child(xml_tree,path_dsd,item_name,error_flag)
    IF (error_flag .ne. no_error) return
    path = trim(path_dsd)//"/"//trim(item_name)
    call xml_tree_set_string_node_value(xml_tree,path,trim(adjustl(item)),&
                                        error_flag)
    IF (error_flag .ne. no_error) return

    item_name = "Ds_Size"
    if (associated(DSD)) item = DSD%Ds_Size
    ! add this item
    call xml_tree_add_child(xml_tree,path_dsd,item_name,error_flag)
    IF (error_flag .ne. no_error) return
    path = trim(path_dsd)//"/"//trim(item_name)
    call xml_tree_set_string_node_value(xml_tree,path,trim(adjustl(item)),&
                                        error_flag)
    IF (error_flag .ne. no_error) return

    item_name = "Num_Dsr"
    if (associated(DSD)) item = DSD%Num_Dsr
    ! add this item
    call xml_tree_add_child(xml_tree,path_dsd,item_name,error_flag)
    IF (error_flag .ne. no_error) return
    path = trim(path_dsd)//"/"//trim(item_name)
    call xml_tree_set_string_node_value(xml_tree,path,trim(adjustl(item)),&
                                        error_flag)
    IF (error_flag .ne. no_error) return

    item_name = "Dsr_Size"
    if (associated(DSD)) item = DSD%Dsr_Size
    ! add this item
    call xml_tree_add_child(xml_tree,path_dsd,item_name,error_flag)
    IF (error_flag .ne. no_error) return
    path = trim(path_dsd)//"/"//trim(item_name)
    call xml_tree_set_string_node_value(xml_tree,path,trim(adjustl(item)),&
                                        error_flag)
    IF (error_flag .ne. no_error) return

    item_name = "Byte_Order"
    if (associated(DSD)) item = DSD%Byte_Order
    ! add this item
    call xml_tree_add_child(xml_tree,path_dsd,item_name,error_flag)
    IF (error_flag .ne. no_error) return
    path = trim(path_dsd)//"/"//trim(item_name)
    call xml_tree_set_string_node_value(xml_tree,path,trim(adjustl(item)),&
                                        error_flag)
    IF (error_flag .ne. no_error) return

    ! Spare 1
    item_name = "Spare_1"
    item=newline
    ! add this item
    call xml_tree_add_child(xml_tree,path_dsd,item_name,error_flag)
    IF (error_flag .ne. no_error) return
    path = trim(path_dsd)//"/"//trim(item_name)
    call xml_tree_set_string_node_value(xml_tree,path,trim(adjustl(item)),&
                                        error_flag)
    IF (error_flag .ne. no_error) return
    
  end subroutine Add_DSD_to_xml_SPH
    !  #]
  subroutine Get_DSD_from_xml_tree(xml_tree, path_in, &
                                   DSD, which_DSD, exp_ds_name, exp_ds_type,&
                                   error_flag,workaround)
    !  #[
    TYPE(xml_node_type), pointer     :: xml_tree    ! input
    character(len=*),    intent(in)  :: path_in     ! input
    TYPE(DSD_type),      pointer     :: DSD         ! result
    integer,             intent(in)  :: which_DSD   ! input
    character(len=*),    intent(in)  :: exp_ds_name ! input
    character(len=*),    intent(in)  :: exp_ds_type ! input
    integer,             intent(out) :: error_flag  ! output
    logical, optional,   intent(in)  :: workaround  ! optional input

    ! local variables                                                          
    integer                     :: ios
    character(len=max_str_len)  :: path
    character(len=max_str_len)  :: dsd_path
    character(len=max_str_len)  :: text
    character(len=max_str_len)  :: attr_name
    character(len=max_str_len)  :: attr_value
    character(len=3)            :: which_DSD_string
    integer                     :: DSD_count
    integer(IntAd)              :: value_IntAd
    integer(IntAl)              :: value_IntAl
    integer                     :: pos
    logical                     :: use_workaround

    ! init
    error_flag = no_error

    use_workaround = .false.
    if (present(workaround)) use_workaround = workaround

    IF (.not. associated(DSD)) THEN
       print *,"ERROR in Get_DSD_from_xml_tree():"
       print *,"the pointer DSD seems not associated !!!"
       error_flag = error_programming
       return
    END IF

    IF (.not. associated(xml_tree)) THEN
       print *,"ERROR in Get_DSD_from_xml_tree():"
       print *,"the pointer xml_tree seems not associated !!!"
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! request the DSD count attribute and check if the
    ! requested DSD actually exists
    !------------------------------

    ! this change is only present in the L1B product v1.08 for now,
    ! but it will probably be necessary to adapt all our files.
    ! To enable a gradual shift, use this workaround switch.
    path = trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name
    if (use_workaround) then
       path = trim(path)//"/List_of_Data_Set_Descriptors"
    else
       path = trim(path)//"/List_of_Dsds"
    end if

    attr_name = "count"
    call xml_tree_read_string_attribute(xml_tree,path,attr_name,attr_value,&
         error_flag) 
    IF (error_flag .ne. no_error) return
    read(attr_value,"(i2)",iostat=ios) DSD_count
    IF (ios .ne. 0) goto 999

    IF (which_DSD .gt. DSD_count) THEN
       print *,"ERROR in Get_DSD_from_xml_tree():"
       print *,"DSD nr ",which_DSD," was requested"
       print *,"but this file only seems to contain ",DSD_count," DSD's !!!"
       error_flag = error_reading_file
       return
    END IF

    !------------------------------
    ! copying the DSD XML elements
    !------------------------------

    write(which_DSD_string,"(i2.2)") which_DSD

    if (use_workaround) then
       dsd_path = trim(path)//"/Data_Set_Descriptor["//&
                  trim(which_DSD_string)//"]"
    else
       dsd_path = trim(path)//"/Dsd["//trim(which_DSD_string)//"]"
    end if

    path = trim(dsd_path)//"/Ds_Name"
    call xml_node_get_string(xml_tree,path,text,error_flag) 
    IF (error_flag .ne. no_error) return
    DSD%DS_Name = & 
         check_length_xml_item(text,path,maxlen_DS_Name,& 
                               "Get_DSD_from_xml_tree")

    ! check if the name of this dataset equals the expected name
    IF (trim(DSD%DS_Name) .ne. trim(exp_ds_name)) THEN
       print *,"ERROR in Get_DSD_from_xml_tree():"
       print *,"expected a dataset named:  ",trim(exp_ds_name)
       print *,"but found a dataset named: ",trim(DSD%DS_Name)
       error_flag = error_reading_file
       return
    END IF

    path = trim(dsd_path)//"/Ds_Type"
    call xml_node_get_string(xml_tree,path,text,error_flag) 
    IF (error_flag .ne. no_error) return
    DSD%DS_Type = & 
         check_length_xml_item(text,path,maxlen_DS_Type,& 
                               "Get_DSD_from_xml_tree")

    ! check the result
    pos = verify(trim(DSD%DS_Type),"MAGR")
    IF (pos .ne. 0) THEN
       print *,"ERROR in Get_DSD_from_xml_tree():"
       print *,"item: DS_TYPE is only allowed to have the values M,A,G or R"
       print *,"but is seems to contain: ",trim(DSD%DS_Type)
       error_flag = error_reading_file
       return
    END IF

    ! check if the type of this dataset equals the expected type
    IF (trim(DSD%DS_Type) .ne. trim(exp_ds_type)) THEN
       print *,"ERROR in Get_DSD_from_xml_tree():"
       print *,"expected a dataset of type:  ",trim(exp_ds_type)
       print *,"but found a dataset of type: ",trim(DSD%DS_Type)
       error_flag = error_reading_file
       return
    END IF

    path = trim(dsd_path)//"/Filename"
    call xml_node_get_string(xml_tree,path,text,error_flag) 
    IF (error_flag .ne. no_error) return
    DSD%Filename = & 
         check_length_xml_item(text,path,maxlen_Filename,& 
                               "Get_DSD_from_xml_tree")

    path = trim(dsd_path)//"/Ds_Offset"
    call xml_node_get_string(xml_tree,path,text,error_flag) 
    IF (error_flag .ne. no_error) return
    DSD%DS_Offset = & 
         check_length_xml_item(text,path,maxlen_IntAd,& 
                               "Get_DSD_from_xml_tree")

    path = trim(dsd_path)//"/Ds_Size"
    call xml_node_get_string(xml_tree,path,text,error_flag) 
    IF (error_flag .ne. no_error) return
    DSD%DS_Size = & 
         check_length_xml_item(text,path,maxlen_IntAl,& 
                               "Get_DSD_from_xml_tree")

    path = trim(dsd_path)//"/Num_Dsr"
    call xml_node_get_string(xml_tree,path,text,error_flag) 
    IF (error_flag .ne. no_error) return
    DSD%NUM_DSR = & 
         check_length_xml_item(text,path,maxlen_IntAl,& 
                               "Get_DSD_from_xml_tree")

    path = trim(dsd_path)//"/Dsr_Size"
    call xml_node_get_string(xml_tree,path,text,error_flag) 
    IF (error_flag .ne. no_error) return
    DSD%DSR_Size = & 
         check_length_xml_item(text,path,maxlen_IntAl,& 
                               "Get_DSD_from_xml_tree")

    path = trim(dsd_path)//"/Byte_Order"
    call xml_node_get_string(xml_tree,path,text,error_flag) 
    IF (error_flag .ne. no_error) return
    DSD%Byte_Order = & 
         check_length_xml_item(text,path,maxlen_Byte_Order,& 
                               "Get_DSD_from_xml_tree")

    call GetIntAd(DSD%DS_Offset,value_IntAd,error_flag)
    DSD%start_of_ds     = value_IntAd+1
    call GetIntAl(DSD%NUM_DSR,value_IntAl,error_flag)
    DSD%nr_of_dataset_records  = value_IntAl
    call GetIntAl(DSD%DSR_SIZE,value_IntAl,error_flag)
    DSD%size_of_one_dsr = value_IntAl
    call GetIntAl(DSD%DS_SIZE,value_IntAl,error_flag)
    DSD%size_of_ds      = value_IntAl

    DSD%end_of_ds       = DSD%start_of_ds+DSD%size_of_ds-1

    return

999 print *,"ERROR in Get_DSD_from_xml_tree():"
    print *,"error converting DSD_count from string to real..."
    print *,"attr_value holding DSD_count = ",trim(attr_value)
    error_flag = error_reading_file
    return

  end subroutine Get_DSD_from_xml_tree
    !  #]
  !-------------------------
  subroutine Fill_DSD_items (DSD, &
             DS_Name, DS_Type, Ds_Offset, Num_Dsr, Dsr_Size, &
             error_flag, Filename)
    !  #[
    !  01-12-2005  P. Poli    Original code.
    !  23-02-2006  J. de Kloe adapted to be able to fill all fields
    !  17-11-2006  J. de Kloe made Filename optional and added a sanity check
    TYPE(DSD_type),       pointer              :: DSD
    character(len=maxlen_DS_Name), intent(in)  :: DS_Name
    character(len=maxlen_DS_Type), intent(in)  :: DS_Type
    integer(IntAd),                intent(in)  :: Ds_Offset
    integer(IntAl),                intent(in)  :: Num_Dsr
    integer(IntAl),                intent(in)  :: Dsr_Size    
    integer,                       intent(out) :: error_flag
    character(len=*),optional,     intent(in)  :: Filename

    ! local variables
    character(len=256) :: path
    character(len=256) :: filename_without_path
    character(len=10)  :: txt_maxlen, txt_len

    ! init
    error_flag = no_error

    IF (.not. associated(DSD)) THEN
       call logmsg(log_error,"ERROR in Fill_DSD_items(): "//&
            "The pointer to this DSD seems not associated !!")
       error_flag = error_programming
       return
    END IF

    ! sanity check
    IF (DS_Type .eq. 'R') THEN
       IF (.not. present(Filename)) THEN
          ! for referenced datasets a filename is required
          call logmsg(log_error,"Error in Fill_DSD_items(): "//&
               "DSD named "//trim(DS_Name)//" is of type R so a filename "//&
               "should be provided when filling this DSD, but this seems "//&
               "to be forgotten ...")
          error_flag = error_programming
          return
       END IF
    ELSE
       IF (present(Filename)) THEN
          ! for attached datasets (ADS/MDS) a filename is forbidden
          call logmsg(log_error,"Error in Fill_DSD_items(): "//&
               "DSD named "//trim(DS_Name)//" is of type "//trim(DS_Type)//&
               " so a filename is not allowed when filling this DSD, but "//&
               "it seems a filename is provided anyway ...")
          error_flag = error_programming
          return
       END IF
    END IF

    DSD%DS_Name  = DS_Name
    DSD%DS_Type  = DS_Type
    DSD%Filename(:) = ' '
    IF (present(Filename)) THEN
       ! a path might be prepended, this is allowed in the JobOrder
       ! fields used to define the Filenames, so remove the path
       ! before storing the value

       ! split path from filename
       call Split_path_and_filename(Filename,path,filename_without_path)

       ! check length of Filename
       if (len_trim(filename_without_path) .gt. maxlen_Filename) then
          write(txt_maxlen,"(i10)",err=998) maxlen_Filename
          write(txt_len,   "(i10)",err=999) len_trim(filename_without_path)
          call logmsg(log_warn,"WARNING in Fill_DSD_items(): "//&
               "filename to be stored in this DSD is too long: "//&
               trim(filename_without_path)//&
               ". This name has length: "//trim(txt_len)//&
               " but the max. length is "//trim(txt_maxlen)//&
               " The name will be truncated in the output file.")
       end if
       ! store the result
       DSD%Filename = filename_without_path(1:maxlen_Filename)
    END IF


    DSD%start_of_ds = Ds_Offset+1
    call SetIntAd(Ds_Offset,DSD%DS_Offset,error_flag)
    IF (error_flag .ne. no_error) THEN
       call logmsg(log_error,"ERROR in Fill_DSD_items: "//&
            "problem with conversion of Ds_Offset.")
       return
    ENDIF

    IF ( (Dsr_Size .eq. missing_indicator_bytes) .and. &
         (Num_Dsr  .eq. missing_indicator_bytes)) THEN
       DSD%size_of_ds = missing_indicator_bytes ! indicates missing condition
    ELSE
       DSD%size_of_ds = Dsr_Size * Num_Dsr
    END IF
    call SetIntAl(DSD%size_of_ds,DSD%DS_Size,error_flag)
    IF (error_flag .ne. no_error) THEN
       call logmsg(log_error,"ERROR in Fill_DSD_items: "//&
            "problem with conversion of size_of_ds.")
       return
    ENDIF

    DSD%end_of_ds = Ds_Offset + DSD%size_of_ds

    DSD%nr_of_dataset_records = Num_Dsr
    call SetIntAl(Num_Dsr,DSD%NUM_DSR,error_flag)
    IF (error_flag .ne. no_error) THEN
        call logmsg(log_error,"ERROR in Fill_DSD_items: "//&
             "problem with conversion of Num_Dsr.")
       return
    ENDIF

    DSD%size_of_one_dsr = Dsr_Size
    call SetIntAl(Dsr_Size,DSD%DSR_Size,error_flag)
    IF (error_flag .ne. no_error) THEN
        call logmsg(log_error,"ERROR in Fill_DSD_items: "//&
             "problem with conversion of Dsr_Size.")
       return
    ENDIF

    ! this is a constant for all ee_cfi file types
    ! so no inputs are needed to fill this field
    DSD%Byte_Order=L2B_DS_Byte_Order

    return

998 call logmsg(log_error,"ERROR in Fill_DSD_items(): could not convert "//&
         "maxlen_Filename to a string. Probably this is a programming error")
    error_flag = error_programming
    return

999 call logmsg(log_error,"ERROR in Fill_DSD_items(): could not convert "//&
         "lentrim(filename_without_path) to a string. "//&
         "Probably this is a programming error")
    error_flag = error_programming
    return

  end subroutine Fill_DSD_items
  !  #]
  !-------------------------
  subroutine Write_DSD_to_KVT_gen(DSD, fileunit, error_flag)
    !  #[

    !  01-12-2005  P. Poli    Original code.
    !  Note: when this routine is called for writing a non-associated DSD,
    !        it writes blanks instead and saves the position in the file.
    !  09-02-2006  P. Poli    Added switch_AMD to save DSD position for AMD
    !                         files
    !  21-02-2006  J. de Kloe, simplified version, without dependency on
    !                          L2B_SummaryVars or AMD specifics
    !  reason: Stored_Dsd_Size can be kep local easily
    !          DSD's should have identical structure for ALL files
    !          so having an exception for AMD files is a bad idea.
    !
    TYPE(DSD_type), pointer     :: DSD
    TYPE(DirectBinaryIO_fileunit_type), intent(in) :: fileunit
    integer,        intent(out) :: error_flag

    ! local variables
    character(len=64) :: varname
    character(len=64) :: value
    character(len=64) :: phys_unit
    integer           :: max_value_length
    integer(i8_)      :: begpos, endpos
    integer(IntAl)    :: Dsd_Size, expected_Dsd_Size

    ! only used to check whether sizes remain identicall between 
    ! different calls to this routine
    integer(IntAl), save :: Stored_Dsd_Size = 0_IntAl

    !---------------------------------
    ! Writing the DSD elements
    !---------------------------------

    error_flag = no_error

    ! get first position
    call GetFilePointer(fileunit,begpos,error_flag)
    IF (error_flag .ne. no_error) goto 998

    IF (.not. associated(DSD)) THEN
       print *,"Write_DSD_to_KVT: The pointer to this DSD seems not associated"
       print *,"Write_DSD_to_KVT: Writing empty tags instead"
       ! Spare_1: this is a spare value without content
       max_value_length = 254
       call write_spare_kvt_string(fileunit,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! writing DS_Name
       varname="DS_NAME"
       value=DSD%DS_Name
       max_value_length=maxlen_DS_Name
       call write_kvt_string(fileunit,varname,value,max_value_length, &
                             error_flag, use_quotes=.true.)
       IF (error_flag .ne. no_error) goto 998

       ! writing DS_Type
       varname="DS_TYPE"
       value=DSD%DS_Type
       max_value_length=maxlen_DS_Type
       call write_kvt_string(fileunit,varname,value,max_value_length, &
                             error_flag)
       IF (error_flag .ne. no_error) goto 998

       ! writing DS_Filename
       varname="FILENAME"
       value=DSD%Filename
       max_value_length=maxlen_Filename
       call write_kvt_string(fileunit,varname,value,max_value_length, &
                             error_flag, use_quotes=.true.)
       IF (error_flag .ne. no_error) goto 998

       ! writing DS_Offset
       varname="DS_OFFSET"
       value=DSD%DS_Offset
       max_value_length=maxlen_IntAd  ! =21
       phys_unit="bytes"
       call write_kvt_string(fileunit,varname,value,max_value_length, &
                             error_flag, phys_unit=phys_unit)
       IF (error_flag .ne. no_error) goto 998

       ! writing DS_Size
       varname="DS_SIZE"
       value=DSD%DS_Size
       max_value_length=maxlen_IntAl  ! =11
       phys_unit="bytes"
       call write_kvt_string(fileunit,varname,value,max_value_length, &
                             error_flag, phys_unit=phys_unit)
       IF (error_flag .ne. no_error) goto 998
       
       ! writing NUM_Dsret
       varname="NUM_DSR"
       value=DSD%NUM_DSR
       max_value_length=maxlen_IntAl  ! =11
       call write_kvt_string(fileunit,varname,value,max_value_length,&
                             error_flag)
       IF (error_flag .ne. no_error) goto 998
   
       ! writing DSR_Size
       varname="DSR_SIZE"
       value=DSD%DSR_Size
       max_value_length=maxlen_IntAl  ! =11
       phys_unit="bytes"
       call write_kvt_string(fileunit,varname,value,max_value_length, &
                             error_flag, phys_unit=phys_unit)
       IF (error_flag .ne. no_error) goto 998
      
       ! writing Byte_Order
       varname="BYTE_ORDER"
       value=DSD%Byte_Order
       max_value_length=maxlen_Byte_Order
       call write_kvt_string(fileunit,varname,value,max_value_length, &
                             error_flag, use_quotes=.true.)
       IF (error_flag .ne. no_error) goto 998
    ENDIF

    ! Spare_1: this is a spare value without content
    max_value_length = 32
    call write_spare_kvt_string(fileunit,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! get last position and calculate size.
    call GetFilePointer(fileunit,endpos,error_flag)
    IF (error_flag .ne. no_error) goto 998
    endpos=endpos-1
    Dsd_Size=int(endpos-begpos+1,IntAl)

    IF (Stored_Dsd_Size .eq. 0_IntAl) THEN
       Stored_Dsd_Size = Dsd_Size

       expected_Dsd_Size = GetByteSize_DSD_KVT()
       IF (Dsd_Size .ne. expected_Dsd_Size) THEN
          print *,"ERROR in Write_DSD_to_KVT_gen():"
          print *,"nbytes_written  = ",Dsd_Size
          print *,"nbytes_expected = ",expected_Dsd_Size
          print *,"something went wrong when writing a DSD to the KVT"
          print *,"part of the binary file, probably a programming error ..."
          error_flag = error_programming
          return
       END IF

    ELSE
       ! compare with size calculated previously
       IF (Stored_Dsd_Size .ne. Dsd_Size) THEN
          print *,'Write_DSD_to_KVT: '//&
               'WARNING: The size of L2B DSDs seems to change'
       ENDIF
    ENDIF

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Write_DSD_to_KVT"
    return

  end subroutine Write_DSD_to_KVT_gen
    !  #]
  subroutine Write_DSD_to_KVT_unbuff(DSD, fileunit, error_flag)
    !  #[
    TYPE(DSD_type), pointer     :: DSD
    integer,        intent(in)  :: fileunit
    integer,        intent(out) :: error_flag

    ! local variable
    TYPE(DirectBinaryIO_fileunit_type) :: local_fileunit

    error_flag = no_error

    call Init_DirectBinaryIO_fileunit(local_fileunit)
    local_fileunit%use_buffering = .false.
    local_fileunit%fileunit      = fileunit

    call Write_DSD_to_KVT_gen(DSD, local_fileunit, error_flag)

  end subroutine Write_DSD_to_KVT_unbuff
    !  #]
  subroutine Write_DSD_to_KVT_buff(DSD, fileunit, error_flag)
    !  #[
    TYPE(DSD_type),          pointer     :: DSD
    TYPE(BufferedFile_type), pointer     :: fileunit
    integer,                 intent(out) :: error_flag

    ! local variable
    TYPE(DirectBinaryIO_fileunit_type) :: local_fileunit

    error_flag = no_error

    call Init_DirectBinaryIO_fileunit(local_fileunit)
    local_fileunit%use_buffering     =  .true.
    local_fileunit%buffered_fileunit => fileunit

    call Write_DSD_to_KVT_gen(DSD, local_fileunit, error_flag)

  end subroutine Write_DSD_to_KVT_buff
    !  #]
  subroutine Read_DSD_from_KVT(DSD, fileunit, exp_ds_name, exp_ds_type, &
                               error_flag, Compare)
    !  #[
    ! Reads a DataSetDescriptor (KVT values).
    ! and compares with the already loaded values from the xml header

    TYPE(DSD_type),    pointer     :: DSD
    integer,           intent(in)  :: fileunit
    character(len=*),  intent(in)  :: exp_ds_name
    character(len=*),  intent(in)  :: exp_ds_type
    integer,           intent(out) :: error_flag
    logical, optional, intent(in)  :: Compare

    ! local variables
    character(len=max_str_len) :: exp_name
    character(len=max_str_len) :: value
    character(len=max_str_len) :: exp_phys_unit
    integer                    :: exp_val_length
    logical                    :: do_Compare
    integer(IntAd)             :: Ds_Offset
!    integer                    :: filesize
    integer                    :: pos

    !-----------------------------
    ! Reading the DSD KVT elements
    !-----------------------------

    error_flag = no_error
    IF (.not. associated(DSD)) THEN
       print *,"ERROR in Read_DSD_from_KVT():"
       print *,"The pointer to this DSD seems not associated !!"
       error_flag = error_programming
       return
    END IF

    do_Compare = .false.
    IF (present(Compare)) do_Compare = Compare

    IF (do_Compare) THEN
       print *,"Read_DSD_from_KVT: reading and comparing DSD-KVT items "
       print *,"for ",trim(exp_ds_name)," DSD with the XML items in memory"
    ELSE
       print *,"Read_DSD_from_KVT: reading DSD-KVT items "       
       print *,"for ",trim(exp_ds_name)," DSD"
    END IF

    !------------------------------
    ! reading the DSD KVT elements
    !------------------------------

    ! read the DS_NAME
    exp_val_length = len(DSD%DS_Name)
    exp_name       = "DS_NAME"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expect_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(DSD%DS_Name, value, &
                                         "DSD%DS_Name",error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       DSD%DS_Name = value(1:maxlen_DS_Name)
    END IF

    ! check if the name of this dataset equals the expected name
    IF (trim(value) .ne. trim(exp_ds_name)) THEN
       print *,"ERROR in Read_DSD_from_KVT():"
       print *,"expected a dataset named:  ",trim(exp_ds_name)
       print *,"but found a dataset named: ",trim(value)
       error_flag = error_reading_file
       return
    END IF

    ! read the DS_TYPE
    exp_val_length = len(DSD%DS_Type)
    exp_name       = "DS_TYPE"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(DSD%DS_Type, value, &
                                         "DSD%DS_Type",error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       DSD%DS_Type = value(1:maxlen_DS_Type)
    END IF

    ! check the result
    pos = verify(trim(value),"MAGR")
    IF (pos .ne. 0) THEN
       print *,"ERROR in Read_DSD_from_KVT():"
       print *,"item: DS_TYPE is only allowed to have the values M,A,G or R"
       print *,"but is seems to contain: ",trim(value)
       error_flag = error_reading_file
       return
    END IF

    ! check if the type of this dataset equals the expected type
    IF (trim(value) .ne. trim(exp_ds_type)) THEN
       print *,"ERROR in Read_DSD_from_KVT():"
       print *,"expected a dataset of type:  ",trim(exp_ds_type)
       print *,"but found a dataset of type: ",trim(value)
       error_flag = error_reading_file
       return
    END IF

    ! read the FILENAME
    exp_val_length = len(DSD%Filename)
    exp_name       = "FILENAME"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expect_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(DSD%Filename, value, &
                                         "DSD%Filename",error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       DSD%Filename = value(1:maxlen_Filename)
    END IF

    IF (DSD%DS_TYPE .eq. "R") THEN
       IF (len_trim(value) .eq. 0) THEN
          IF (LocalVerbosity .ge. VerbosityWarn) THEN
             print *,"WARNING in Read_DSD_from_KVT():"
             print *,"the DSD%FILENAME field should have a value for"
             print *,"DS_TYPE=R, but the filename seems to be missing !!!"
             print *,"DSD%DS_NAME = ",trim(DSD%DS_NAME)
             print *,"continuing ..."
          END IF
! disabled this test, since reference file don't need to be present
! when reading/using a file. They just describe what was used when
! this file was produced. JK, 02-Aug-2006
!       ELSE
!          ! check if the external reference file is present !
!          ! use the c-stat function to do this, which returns -1
!          ! in case the file can not be found
!          filesize = get_filesize(value)
!          IF (filesize .lt. 0) THEN
!             IF (LocalVerbosity .ge. VerbosityWarn) THEN
!                print *,"WARNING in Read_DSD_from_KVT():"
!                print *,"the file given by DSD%FILENAME could not be found !!"
!                print *,"DSD%DS_NAME  = ",trim(DSD%DS_NAME)
!                print *,"DSD%FILENAME = ",trim(value)
!                print *,"continuing ..."
!             END IF
!          END IF
       END IF
    ELSE
       IF (len_trim(value) .ne. 0) THEN
          print *,"ERROR in Read_DSD_from_KVT():"
          print *,"the DSD%FILENAME field can only have a value for"
          print *,"DS_TYPE=R, but this DSD with type: ",trim(DSD%DS_TYPE)
          print *,"has FILENAME=[",trim(value),"]"
          error_flag = error_reading_file
          return
       END IF
    END IF

    ! read the DS_Offset
    exp_val_length = len(DSD%DS_Offset)
    exp_name       = "DS_OFFSET"
    exp_phys_unit  = "bytes"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,&
         expected_phys_unit=exp_phys_unit, is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(DSD%DS_Offset, value, &
            "DSD%DS_Offset",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       DSD%DS_Offset = value(1:maxlen_IntAd)
    END IF

    ! read the DS_Size
    exp_val_length = len(DSD%DS_Size)
    exp_name       = "DS_SIZE"
    exp_phys_unit  = "bytes"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,&
         expected_phys_unit=exp_phys_unit, is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(DSD%DS_Size, value, &
            "DSD%DS_Size",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       DSD%DS_Size = value(1:maxlen_IntAl)
    END IF

    ! read the NUM_DSR
    exp_val_length = len(DSD%NUM_DSR)
    exp_name       = "NUM_DSR"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(DSD%NUM_DSR, value, &
            "DSD%NUM_DSR",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       DSD%NUM_DSR = value(1:maxlen_IntAl)
    END IF

    ! read the DSR_Size
    exp_val_length = len(DSD%DSR_Size)
    exp_name       = "DSR_SIZE"
    exp_phys_unit  = "bytes"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,&
         expected_phys_unit=exp_phys_unit, is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(DSD%DSR_Size, value, &
            "DSD%DSR_Size",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       DSD%DSR_Size = value(1:maxlen_IntAl)
    END IF

    ! read the BYTE_ORDER
    exp_val_length = len(DSD%BYTE_ORDER)
    exp_name       = "BYTE_ORDER"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expect_quotes=.true.)

    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(DSD%BYTE_ORDER, value, &
            "DSD%BYTE_ORDER",error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       DSD%BYTE_ORDER = value(1:maxlen_Byte_Order)
    END IF

    ! ee_cfi formatted files MUST be big-endian
    ! so check if the byte order is specified accordingly
    IF (value .ne. "3210") THEN
       print *,"ERROR in Read_DSD_from_KVT():"
       print *,"expected byteorder is: [3210] (bigendian)"
       print *,"but this DSD seems to use: [",trim(value),"]"
       goto 998
    END IF

    ! Spare_1: this is a spare value without content
    exp_val_length = 32
    call read_spare_kvt_string(fileunit,exp_val_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    !------------------------------
    ! fill the numerical values by converting some strings
    !------------------------------

    call GetIntAd(DSD%DS_Offset,Ds_Offset,error_flag)
    IF (error_flag .ne. no_error) goto 998
    call GetIntAl(DSD%DS_Size,DSD%size_of_ds,error_flag)
    IF (error_flag .ne. no_error) goto 998
    call GetIntAl(DSD%NUM_DSR,DSD%nr_of_dataset_records,error_flag)
    IF (error_flag .ne. no_error) goto 998
    call GetIntAl(DSD%DSR_Size,DSD%size_of_one_dsr,error_flag)
    IF (error_flag .ne. no_error) goto 998
    DSD%start_of_ds = Ds_Offset+1
    DSD%end_of_ds   = Ds_Offset + DSD%size_of_ds

    IF (DSD%size_of_one_dsr .ne. missing_indicator_bytes) THEN
       IF (DSD%size_of_ds .ne. DSD%nr_of_dataset_records*DSD%size_of_one_dsr) THEN
          print *,"ERROR in Read_DSD_from_KVT():"
          print *,"ERROR in the definition of this dataset"
          print *,"DSD%size_of_ds             = ",DSD%size_of_ds
          print *,"DSD%nr_of_dataset_records  = ",DSD%nr_of_dataset_records
          print *,"DSD%size_of_one_dsr        = ",DSD%size_of_one_dsr
          print *,"DSD%size_of_ds should be identical to: "
          print *,"DSD%nr_of_dataset_records*DSD%size_of_one_dsr = ",&
                   DSD%nr_of_dataset_records*DSD%size_of_one_dsr
          print *,"but it is not...."
          error_flag = error_reading_file
          return
       END IF
    ELSE
       IF (LocalVerbosity .ge. VerbosityWarn) THEN
          print *,"WARNING: No check defined yet for the case of dsr_size=-1"
          print *,"Assuming all is OK...."
       END IF
    END IF

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Read_DSD_from_KVT"
    return

  end subroutine Read_DSD_from_KVT
    !  #]
  !-------------------------
end module DataSetDescriptor
