module EE_CFI_Datatypes
  !  #[ Documentation
  !
  ! a module to contain type definitions for the datatypes
  ! used in the ee_cfi file definitions
  !
  !  Written by:  J. de Kloe, KNMI
  !  Created:     29-Sep-2005
  !
  !  Modifications:
  !     28-Nov-2005 P. Poli    Fixed bug in Convert_KVT_to_XML_DateTime
  !     29-Nov-2005 P. Poli    Added routine Convert_DateTime_to_XMLKVT
  !     01-Dec-2005 P. Poli    Changed Remove_Plus_Sign. Routine now also
  !                            removes unnecessary zeros.
  !     06-Jan-2006 J. de Kloe DateTime module updated to use r8_ for seconds
  !     06-Jan-2006 J. de Kloe added helper routines for the new DateTimeType
  !     23-Jan-2006 J. de Kloe added wrappers around the to/from_bin_file 
  !                            routines to define which EE_CFI datatype maps
  !                            to which routine in DirectBinaryIO
  !     25-Jan-2006 J. de Kloe replaced the interface statements by renaming
  !                            inside the use statement: "use ..=>.."
  !     18-Feb-2006 J. de Kloe added SetBoolXML/KVT routines
  !     20-Feb-2006 J. de Kloe added Real2Int*() functions
  !     22-Feb-2006 J. de Kloe added Real8_to_Int*() functions
  !     27-Feb-2006 D. Tan     fix for Real2IntAul
  !     06-Mar-2006 D. Tan     Overload 5 Real2IntA* functions
  !     13-Mar-2006 J. de Kloe added GetIntAuc and GetIntAus routines
  !     15-Mar-2006 J. de Kloe added the SetFAdoFull_or_Missing routine
  !     11-Apr-2006 J. de Kloe added the SetIntAs_or_missing,
  !                            SetIntAus_or_missing and GetFAdoxy
  !     12-May-2006 D. Tan     Null use of ParKind1 module
  !     13-Sep-2006 J. de Kloe renamed the real8_to_Int* functions to
  !                            FAdoxy_to_Int*, and deleted the realr_to_Int*
  !                            functions and related interface blocks
  !     31-Oct-2006 J. de Kloe resolved some gfortran compatibility ussues
  !     01-Nov-2006 J. de Kloe added get_averaged_datetime/get_time_difference
  !     06-Nov-2006 P. Poli    moved int2txt to the present module (to handle
  !                            unsigned integers, unknown in module numerics)
  !     15-Feb-2007 J. de Kloe added function Convert_JobOrd_to_Num_DateTime
  !     09-Mar-2007 J. de Kloe added write_SpareBytes_to_bin_file and
  !                            read_SpareBytes_from_bin_file aliasses
  !     27-Jul-2007 J. de Kloe added function DateTimeMissing()
  !                            added subroutine CheckWithinTimeRange()
  !     31-Jul-2007 J. de Kloe added missing handlers for several DateTime
  !                            subroutines and functions
  !     13-Nov-2007 J. de Kloe changed xml datetime format to have
  !                            microsecond precision
  !     14-Nov-2007 J. de Kloe introduced missing string values
  !                            missing_DateTime_XML and missing_DateTime_KVT
  !     08-Jan-2008 J. de Kloe added subroutine GetIntAc
  !     09-Jan-2008 J. de Kloe changed the XML/KVT missing string
  !                            values for DateTime
  !     16-Jan-2008 J. de Kloe phase out integer kind i_ 
  !     21-Feb-2008 J. de Kloe made workarounds in the DateTime xml handling
  !                            explicit with workaround switches
  !     26-Feb-2008 P. Poli    Added SetIntAc_or_missing
  !                            Removed int1_2txt from generic int2txt
  !
  !  #]
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  ! specify more explicitely which fortran data type corresponds to
  ! which ee_cfi data type, by renaming the reading/writing routines,
  USE DirectBinaryIO, only: &
       DirectBinaryIO_fileunit_type, &
       write_SpareBytes_to_bin_file  => write_zero_bytes_bin_file, &
       read_SpareBytes_from_bin_file => read_zero_bytes_bin_file,  &
       write_IntAc_to_bin_file   => write_int1_to_bin_file,      &
       write_IntAs_to_bin_file   => write_int2_to_bin_file,      &
       write_IntAl_to_bin_file   => write_int4_to_bin_file,      &
       write_IntAd_to_bin_file   => write_int8_to_bin_file,      &
       read_IntAc_from_bin_file  => read_int1_from_bin_file,     &
       read_IntAs_from_bin_file  => read_int2_from_bin_file,     &
       read_IntAl_from_bin_file  => read_int4_from_bin_file,     &
       read_IntAd_from_bin_file  => read_int8_from_bin_file,     &
       write_IntAuc_to_bin_file  => write_uns_int1_to_bin_file,  &
       write_IntAus_to_bin_file  => write_uns_int2_to_bin_file,  &
       write_IntAul_to_bin_file  => write_uns_int4_to_bin_file,  &
       read_IntAuc_from_bin_file => read_uns_int1_from_bin_file, &
       read_IntAus_from_bin_file => read_uns_int2_from_bin_file, &
       read_IntAul_from_bin_file => read_uns_int4_from_bin_file, &
       write_FAdoxy_to_bin_file  => write_real8_to_bin_file,     &
       read_FAdoxy_from_bin_file => read_real8_from_bin_file,    &
       BufferedFile_type, &
       read_IntAus_from_buffbin_file => read_uns_int2_from_buffbin_file, &
       read_IntAul_from_buffbin_file => read_uns_int4_from_buffbin_file, &
       read_IntAd_from_buffbin_file  => read_int8_from_buffbin_file,     &
       read_FAdoxy_from_buffbin_file => read_real8_from_buffbin_file
  USE numerics, only: i8_, i4_, i2_, i1_, r8_, r4_, missing_real,&
                      missing_indicator_integer_i2_, &
                      missing_indicator_integer_i4_, &
                      missing_indicator_integer_i8_, &
                      missing_indicator_real_r8, &
                      is_NaN_real8, is_PosInf_real8, is_NegInf_real8

  USE ErrorHandler, only: no_error, error_reading_file, error_writing_file, &
                          error_programming
  USE DateTimeMod, only: date_type, init_date, print_date, set_date, &
                         time_type, init_time, print_time, set_time, &
                         convert_int_to_date, convert_date_to_int,&
                         conv_date_to_daycount, conv_daycount_to_date,&
                         convert_real_to_time, convert_time_to_real, &
                         get_current_date_time, time2string, date2string,&
                         time2string_sep, date2string_sep, test_time, &
                         test_date, increment_datetime, &
                         set_date_to_missing, set_time_to_missing
  USE Logging, only: logmsg,log_debug,log_info,log_progress,log_warn,log_error
  !  #]
  !  #[ Type definitions and parameters
  implicit none

  !  #[ type definitions

  ! type String  corresponds to fortran type character(len=*)
  ! type Enum    corresponds to fortran type character(len=1) or i1_
  ! (so only 256 values are possible, but these may have human readable
  !  names, so the string representation may be much longer than 1 char)
  ! type Boolean corresponds to fortran type logical
  ! type IntAuc corresponds to fortran type i2_ (c: 1 byte unsigned char int)
  integer, parameter :: IntAuc = i2_
  ! type IntAc  corresponds to fortran type i1_ (c: 1 byte signed char int)
  integer, parameter :: IntAc  = i1_
  ! type IntAus corresponds to fortran type i4_ (c: 2 byte unsigned short int)
  integer, parameter :: IntAus = i4_
  ! type IntAs  corresponds to fortran type i2_ (c: 2 byte signed short int)
  integer, parameter :: IntAs  = i2_
  ! type IntAul corresponds to fortran type i8_ (c: 4 byte unsigned int)
  integer, parameter :: IntAul = i8_
  ! type IntAl  corresponds to fortran type i4_ (c: 4 byte signed int)
  integer, parameter :: IntAl  = i4_
  ! type IntAd  corresponds to fortran type i8_ (c: 8 byte signed int)
  integer, parameter :: IntAd  = i8_
  ! type FAdoxy corresponds to fortran type r8_ (c: 8 byte real)
  ! (represented as string with x digits before and y digits after 
  !  the decimal point, with the sign at the beginning. 
  !  So 123.4567 is op type FOdo34 )
  integer, parameter :: FAdoxy = r8_

  ! type DateTime corresponds to a fortran type character(len=*)
  !     with length 23 for xml files and 27 for KVT format
  !     (who invented this kind of nonsense?)
  !     The c-datatype is a struct of 3 long integers, 1 signed, 2 unsigned
  ! the numerical fortran representation will be:
  TYPE DateTimeType
     integer(IntAl)  :: daycount ! since 1-1-2000 0:0 h (may be negative)
     integer(IntAul) :: seconds  ! elapsed since beginning of the day
     integer(IntAul) :: microseconds ! elapsed since start of last second
  END TYPE DateTimeType
  ! bytesize in fortran will be: 4+8+8=20
  ! bytesize in c       will be: 4+4+4=12
  !  #]

  !  #[ allowed lengths for Boolean

  integer(IntAl), parameter :: c_bytelen_Bool = 1
  integer, parameter :: maxlen_BoolXML = 5 ! allowed values are: True/False
  integer, parameter :: maxlen_BoolKVT = 1 ! allowed values are: 1/0

  !  #]
  !  #[ allowed integer string lengths, ranges and formats

  ! lengths needed to convert to strings       ! example/valid range
  integer, parameter :: maxlen_IntAuc = 4  !           0 ...   +255
  integer, parameter :: maxlen_IntAc  = 4  !        -128 ...   +127
  integer, parameter :: maxlen_IntAus = 6  !           0 ... +65535
  integer, parameter :: maxlen_IntAs  = 6  !      -32768 ... +32767
  integer, parameter :: maxlen_IntAul = 11 !           0 ... +4294967295
  integer, parameter :: maxlen_IntAl  = 11 ! -2147483648 ... +2147483647
  integer, parameter :: maxlen_IntAd  = 21 ! -9223372036854775808 ...
  ! should this not be 20 ???????????????      !        +9223372036854775807 

  ! IntAuc used for storing flags
  integer, parameter :: maxlen_IntAucFlags = 8  ! FFFFFFFF ... TTTTTTTT

  ! allowed integer ranges
  integer(IntAuc), parameter :: minval_IntAuc =      0
  integer(IntAuc), parameter :: maxval_IntAuc =    255
  integer(IntAc),  parameter :: minval_IntAc  =   -128
  integer(IntAc),  parameter :: maxval_IntAc  =    127
  integer(IntAus), parameter :: minval_IntAus =      0
  integer(IntAus), parameter :: maxval_IntAus =  65535
  integer(IntAs),  parameter :: minval_IntAs  = -32768
  integer(IntAs),  parameter :: maxval_IntAs  =  32767
  integer(IntAul), parameter :: minval_IntAul =           0
  integer(IntAul), parameter :: maxval_IntAul =  4294967295_i8_
  ! WARNING: gfortran interprets the constant -2147483648
  ! as the integer +2147483648 and then applies the unary_minus()
  ! function on it to get the value. Therefore the next line fails
  ! with an out of range warning !!!
  !integer(IntAl),  parameter :: minval_IntAl  = -2147483648
  ! so replace it with this simple workaround to make it portable:
  integer(IntAl),  parameter :: minval_IntAl  = -2147483647 - 1
  integer(IntAl),  parameter :: maxval_IntAl  =  2147483647
  ! warning this number is too large for the ifort 8.1 compiler
  ! so I had to increase the minvalue by 1. This means the full range
  ! of the 8byte c integer cannot be used in fortran. but for the moment
  ! this does not seem to be a problem to me...
  !  integer(i8_), parameter :: minval_IntAd  = -9223372036854775808_i8_
  integer(IntAd),  parameter :: minval_IntAd  = -9223372036854775807_i8_
  integer(IntAd),  parameter :: maxval_IntAd  =  9223372036854775807_i8_

  integer(IntAc),  parameter :: missing_indicator_IntAc  = maxval_IntAc
  integer(IntAuc), parameter :: missing_indicator_IntAuc = maxval_IntAuc
  integer(IntAs),  parameter :: missing_indicator_IntAs  = maxval_IntAs
  integer(IntAus), parameter :: missing_indicator_IntAus = maxval_IntAus
  integer(IntAl),  parameter :: missing_indicator_IntAl  = maxval_IntAl
  integer(IntAul), parameter :: missing_indicator_IntAul = maxval_IntAul
  integer(IntAd),  parameter :: missing_indicator_IntAd  = maxval_IntAd

  ! integer formats
  character(len=*), parameter :: format_IntAuc = "(i4)"
  character(len=*), parameter :: format_IntAc  = "(i4)"
  character(len=*), parameter :: format_IntAus = "(i6)"
  character(len=*), parameter :: format_IntAs  = "(i6)"
  character(len=*), parameter :: format_IntAul = "(i11)"
  character(len=*), parameter :: format_IntAl  = "(i11)"
  character(len=*), parameter :: format_IntAd  = "(i21)"

  ! to compose the DataSetDescriptors we need the size of these datatypes
  ! when written in c-format to the binary file, so specify them here:
  integer(IntAl), parameter :: c_bytelen_IntAuc = 1
  integer(IntAl), parameter :: c_bytelen_IntAc  = 1
  integer(IntAl), parameter :: c_bytelen_IntAus = 2
  integer(IntAl), parameter :: c_bytelen_IntAs  = 2
  integer(IntAl), parameter :: c_bytelen_IntAul = 4
  integer(IntAl), parameter :: c_bytelen_IntAl  = 4
  integer(IntAl), parameter :: c_bytelen_IntAd  = 8

  !  #]
  !  #[ allowed length for DateTime

  ! lengths for string representations of each type
  integer, parameter :: maxlen_DateTime_XML      = 30
  integer, parameter :: maxlen_DateTime_XML_old  = 23
  integer, parameter :: maxlen_DateTime_KVT      = 27
  integer, parameter :: maxlen_DateTime_JobOrder = 21
  ! note that this joborder datetime length is redefined in
  ! joborder_datastructure.F90. Using that one here would mean
  ! a dependency for ee_cfi_datatypes on that module, which is not
  ! very desired right now.
  ! format "YYYYMMDD_HHMMSSSSSSSS"

  ! use these missing strings to fill missing fields in our files, rather 
  ! than "[missing]" to prevent complaints from ESA that we are using
  ! "invalid" strings that they cannot parse with their validation tools
  character(len=*), parameter :: missing_DateTime_XML = ""
!      "UTC=1970-01-01T00:00:00.000000"
  character(len=*), parameter :: missing_DateTime_KVT = &
       "                           "
!      "01-JAN-1970 00:00:00.000000"

  integer(IntAl), parameter :: c_bytelen_DateTime = &
       c_bytelen_IntAl + 2*c_bytelen_IntAul ! = 3*4=12

  !  #]
  !  #[ allowed real string lengths, ranges and formats

  ! lengths needed to convert to strings        ! example
  integer, parameter :: maxlen_FAdo06  = 8  !     +.123456
  integer, parameter :: maxlen_FAdo14  = 7  !    +3.4567
  integer, parameter :: maxlen_FAdo22  = 6  !   +23.45
  integer, parameter :: maxlen_FAdo31  = 6  !  +123.4
  integer, parameter :: maxlen_FAdo32  = 7  !  +123.45
  integer, parameter :: maxlen_FAdo36  = 11 !  +123.456789
  integer, parameter :: maxlen_FAdo56  = 13 ! +12345.123456
  integer, parameter :: maxlen_FAdo41  = 7  ! +1234.5
  integer, parameter :: maxlen_FAdo46  = 12 ! +1234.567123
  integer, parameter :: maxlen_FAdo73  = 12 !     +.123456
  ! the MDA L1B_iodd document (version 2, 6-5-2005)
  ! does not specify the values for xy, just the total field
  ! width of 15, so I'll need this next one too
  integer, parameter :: maxlen_FAdoxy         = 15 !
  integer, parameter :: maxlen_FAdoxy_full    = 25 !

  ! allowed real ranges
  real(r8_), parameter :: minval_FAdo06  =   -0.999999_r8_
  real(r8_), parameter :: maxval_FAdo06  =    0.999999_r8_
  !
  real(r8_), parameter :: minval_FAdo22  =  -99.99_r8_
  real(r8_), parameter :: maxval_FAdo22  =   99.99_r8_
  !
  real(r8_), parameter :: minval_FAdo32  = -999.99_r8_
  real(r8_), parameter :: maxval_FAdo32  =  999.99_r8_
  real(r8_), parameter :: minval_FAdo36  = -999.999999_r8_
  real(r8_), parameter :: maxval_FAdo36  =  999.999999_r8_
  !
  real(r8_), parameter :: minval_FAdo46  = -9999.999999_r8_
  real(r8_), parameter :: maxval_FAdo46  =  9999.999999_r8_
  !       WARNING: for some compiler implementations this maxval
  !                might be represented as: 10000.000000 !!!!!!!!!
  !                due to rounding/representation errors
  !
  real(r8_), parameter :: minval_FAdo73  = -9999999.999_r8_
  real(r8_), parameter :: maxval_FAdo73  =  9999999.999_r8_

  real(FAdoxy), parameter :: missing_indicator_FAdoxy = missing_indicator_real_r8

  ! real formats
  character(len=*), parameter :: format_FAdo06 = "(f8.6)"
  character(len=*), parameter :: format_FAdo22 = "(f6.2)"
  character(len=*), parameter :: format_FAdo32 = "(f7.2)"
  character(len=*), parameter :: format_FAdo36 = "(f11.6)"
  character(len=*), parameter :: format_FAdo46 = "(f12.6)"
  character(len=*), parameter :: format_FAdo73 = "(f12.3)"

  ! to compose the DataSetDescriptors we need the size of these datatypes
  ! when written in c-format to the binary file, so specify them here:
  integer(IntAl), parameter :: c_bytelen_FAdoxy = 8

  !  #]

  !  #]
  !  #[ interfaces for overloading functions
  ! overload the write_DateTime_to_bin_file routine to allow
  ! different types of fileunits
  INTERFACE write_DateTime_to_bin_file
     module procedure write_DateTime_to_buffbin_file,  & ! old buffered type
                      write_DateTime_to_unbuffbinfile, & ! old direct type
                      write_DateTime_to_genbin_file      ! new generic type
  END INTERFACE
  ! overload the function int2txt
  INTERFACE int2txt
     module procedure int2_2txt, int4_2txt, int8_2txt
  END INTERFACE
  !  #]
  !  #[ Module interface
  ! all is public for the moment
  !  #]
contains
  !---------------------------------------
  subroutine GetIntAuc(str_value,value,error_flag)
    !  #[

    character(len=maxlen_IntAuc), intent(in)  :: str_value
    integer(IntAuc),              intent(out) :: value
    integer,                      intent(out) :: error_flag

    ! local variable
    integer :: ios
    character(len=256) :: error_text

    error_flag = no_error

    read(str_value,*,iostat=ios) value
    IF (ios .ne. 0) THEN
       call logmsg(log_error,"ERROR in GetIntAuc: "//&
            "Conversion from string to type IntAuc failed"//&
            "for string: ["//trim(str_value)//"]",&
             SignalFatalError = .true.,FatalError=error_reading_file)
       value = missing_indicator_IntAuc
       return
    END IF

    ! check: see if value falls within allowed range
    IF ( (value .lt. minval_IntAuc) .or. (value .gt. maxval_IntAuc)) THEN
! This is an experiment for transforming the error prints that we
! have in this file into proper calls to the logmsg() function.
! Note that the FatalError needs to be checked later in the program
! since this GetIntAuc function will always return with a value,
! although it maybe missing in case of trouble.
! JK. 15-feb-2007
       write(error_text,"(a,i4,a,i4,a,i4)") &
            "value out of range in: GetIntAuc(), value = ",value,&
            "allowed range: ",minval_IntAuc," ... ",maxval_IntAuc
       call logmsg(log_error,"ERROR in GetIntAuc: "//trim(error_text),&
            SignalFatalError = .true.,FatalError=error_reading_file)
       value = missing_indicator_IntAuc
       return
    END IF

    return

  end subroutine GetIntAuc

    !  #]
  subroutine GetIntAc(str_value,value,error_flag)
    !  #[
    character(len=maxlen_IntAc), intent(in)  :: str_value
    integer(IntAc),              intent(out) :: value
    integer,                     intent(out) :: error_flag

    ! local variable
    integer :: ios
    character(len=256) :: error_text

    error_flag = no_error

    read(str_value,*,iostat=ios) value
    IF (ios .ne. 0) THEN
       call logmsg(log_error,"ERROR in GetIntAc: "//&
            "Conversion from string to type IntAc failed"//&
            "for string: ["//trim(str_value)//"]",&
             SignalFatalError = .true.,FatalError=error_reading_file)
       value = missing_indicator_IntAc
       return
    END IF

    ! check: see if value falls within allowed range
    IF ( (value .lt. minval_IntAc) .or. (value .gt. maxval_IntAc)) THEN
! This is an experiment for transforming the error prints that we
! have in this file into proper calls to the logmsg() function.
! Note that the FatalError needs to be checked later in the program
! since this GetIntAc function will always return with a value,
! although it maybe missing in case of trouble.
! JK. 15-feb-2007
       write(error_text,"(a,i4,a,i4,a,i4)") &
            "value out of range in: GetIntAc(), value = ",value,&
            "allowed range: ",minval_IntAc," ... ",maxval_IntAc
       call logmsg(log_error,"ERROR in GetIntAc: "//trim(error_text),&
            SignalFatalError = .true.,FatalError=error_reading_file)
       value = missing_indicator_IntAc
       return
    END IF

    return

  end subroutine GetIntAc
    !  #]
  subroutine GetIntAus(str_value,value,error_flag)
    !  #[
    character(len=maxlen_IntAus), intent(in)  :: str_value
    integer(IntAus),              intent(out) :: value
    integer,                      intent(out) :: error_flag

    ! local variable
    integer :: ios

    error_flag = no_error

    read(str_value,*,iostat=ios) value
    IF (ios .ne. 0) THEN
       print *,"ERROR in GetIntAus:"
       print *,"Conversion from string to type IntAus failed"
       print *,"for string: [",str_value,"]"
       print *,"error: istat = ",ios
       error_flag = error_reading_file
       return
    END IF

    ! check: see if value falls within allowed range
    IF ( (value .lt. minval_IntAus) .or. (value .gt. maxval_IntAus)) THEN
       print *,"value out of range in: GetIntAus(), value = ",value
       print *,"allowed range: ",minval_IntAus," ... ",maxval_IntAus
       error_flag = error_reading_file
       return
    END IF

    return

  end subroutine GetIntAus
    !  #]
  subroutine GetIntAs(str_value,value,error_flag)
    !  #[
    character(len=maxlen_IntAs), intent(in)  :: str_value
    integer(IntAs),              intent(out) :: value
    integer,                     intent(out) :: error_flag

    ! local variable
    integer :: ios

    error_flag = no_error

    read(str_value,*,iostat=ios) value
    IF (ios .ne. 0) goto 999

    ! check: see if value falls within allowed range
    IF ( (value .lt. minval_IntAs) .or. (value .gt. maxval_IntAs)) THEN
       print *,"value out of range in: GetIntAs(), value = ",value
       print *,"allowed range: ",minval_IntAs," ... ",maxval_IntAs
       error_flag = error_reading_file
       return
    END IF

    return

999 print *,"ERROR in GetIntAs:"
    print *,"Conversion from string to type IntAs failed"
    print *,"for string: [",str_value,"]"
    error_flag = error_reading_file
    return
    
  end subroutine GetIntAs
    !  #]
  subroutine GetIntAul(str_value,value,error_flag)
    !  #[
    character(len=maxlen_IntAul), intent(in)  :: str_value
    integer(IntAul),              intent(out) :: value
    integer,                      intent(out) :: error_flag

    ! local variable
    integer :: ios

    error_flag = no_error

    read(str_value,*,iostat=ios) value
    IF (ios .ne. 0) THEN
       print *,"ERROR in GetIntAul:"
       print *,"Conversion from string to type IntAul failed"
       print *,"for string: [",str_value,"]"
       print *,"error: istat = ",ios
       error_flag = error_reading_file
       return
    END IF

    ! check: see if value falls within allowed range
    IF ( (value .lt. minval_IntAul) .or. (value .gt. maxval_IntAul)) THEN
       print *,"value out of range in: GetIntAul(), value = ",value
       print *,"allowed range: ",minval_IntAul," ... ",maxval_IntAul
       error_flag = error_reading_file
       return
    END IF

    return

  end subroutine GetIntAul
    !  #]
  subroutine GetIntAl(str_value,value,error_flag)
    !  #[
    character(len=maxlen_IntAl), intent(in)  :: str_value
    integer(IntAl),              intent(out) :: value
    integer,                     intent(out) :: error_flag

    ! local variable
    integer :: ios

    error_flag = no_error

    read(str_value,*,iostat=ios) value
    IF (ios .ne. 0) goto 999

    ! check: see if value falls within allowed range
    IF ( (value .lt. minval_IntAl) .or. (value .gt. maxval_IntAl)) THEN
       print *,"value out of range in: GetIntAl(), value = ",value
       print *,"allowed range: ",minval_IntAl," ... ",maxval_IntAl
       error_flag = error_reading_file
       return
    END IF

    return

999 print *,"ERROR in GetIntAl:"
    print *,"Conversion from string to type IntAl failed"
    print *,"for string: [",str_value,"]"
    error_flag = error_reading_file
    return
    
  end subroutine GetIntAl
    !  #]
  subroutine GetIntAd(str_value,value,error_flag)
    !  #[
    character(len=maxlen_IntAd), intent(in)  :: str_value
    integer(IntAd),              intent(out) :: value
    integer,                     intent(out) :: error_flag

    ! local variable
    integer :: ios

    error_flag = no_error

    read(str_value,*,iostat=ios) value
    IF (ios .ne. 0) goto 999

    ! check: see if value falls within allowed range
    IF ( (value .lt. minval_IntAd) .or. (value .gt. maxval_IntAd)) THEN
       print *,"value out of range in: GetIntAd(), value = ",value
       print *,"allowed range: ",minval_IntAd," ... ",maxval_IntAd
       error_flag = error_reading_file
       return
    END IF

    return

999 print *,"ERROR in GetIntAd:"
    print *,"Conversion from string to type IntAd failed"
    print *,"for string: [",str_value,"]"
    error_flag = error_reading_file
    return
    
  end subroutine GetIntAd
    !  #]
  ! REMARK: the following routines can be used to construct strings
  !         from numerical values. These can be used directly for
  !         the KVT formatted files. However, for XML the positive
  !         numbers the prepended '+' character needs to be removed.
  !         (who invents these things ???)
  !         The subroutine Remove_Plus_Sign() can be used for this.
  ! (stripping '+' signs is much easier than adding one later, therefore
  !  I choose to store the KVT formatted strings, and not the xml ones)
  ! REMARK2: because there is no convention for handling missing numbers
  !          for the EE_CFI datatypes, I will use the minimum value allowed
  !          for the given datatype to indicate missing (for the moment
  !          this is only applied to the FAdoxy types)
  ! P.Poli 24/11/2005 : changed Add_Sign_to_KVT_String so that the output
  !             string is, for example,  "-003"
  !             (instead of "  -3" previously). For consistency with L1B files.
  ! REMARK 3: the *or_missing routines are intended for use in print
  !           and AsciiWrite routines. They might return the "[missing]" string
  !           which is not a valid result for KVT/XML use!
  subroutine SetBoolXML(value,str_value)
    !  #[
    logical,                       intent(in)  :: value
    character(len=maxlen_BoolXML), intent(out) :: str_value

    str_value="False"
    IF (value) str_value="True"

  end subroutine SetBoolXML
    !  #]
  subroutine SetBoolKVT(value,str_value)
    !  #[
    logical,                       intent(in)  :: value
    character(len=maxlen_BoolKVT), intent(out) :: str_value

    str_value="0"
    IF (value) str_value="1"

  end subroutine SetBoolKVT
    !  #]
  subroutine SetIntAuc(value,str_value,error_flag)
    !  #[
    integer(IntAuc),              intent(in)  :: value
    character(len=maxlen_IntAuc), intent(out) :: str_value
    integer,                      intent(out) :: error_flag

    error_flag = no_error

    ! check: see if value falls within allowed range
    IF ( (value .lt. minval_IntAuc) .or. (value .gt. maxval_IntAuc)) THEN
       print *,"value out of range in: SetIntAuc(), value = ",value
       print *,"allowed range: ",minval_IntAuc," ... ",maxval_IntAuc
       error_flag = error_writing_file
       return
    END IF

    write(str_value, format_IntAuc, err=999) abs(value)

    call Add_Sign_to_KVT_String(str_value,value.ge.0,error_flag)
    IF (error_flag .ne. no_error) return

    return

999 print *,"ERROR in SetIntAuc:"
    print *,"Conversion from type IntAuc to string failed for value: ",value
    error_flag = error_writing_file
    return

  end subroutine SetIntAuc
    !  #]
  subroutine SetIntAc(value,str_value,error_flag)
    !  #[
    integer(IntAc),              intent(in)  :: value
    character(len=maxlen_IntAc), intent(out) :: str_value
    integer,                     intent(out) :: error_flag

    error_flag = no_error

    ! check: see if value falls within allowed range
    IF ( (value .lt. minval_IntAc) .or. (value .gt. maxval_IntAc)) THEN
       print *,"value out of range in: SetIntAc(), value = ",value
       print *,"allowed range: ",minval_IntAc," ... ",maxval_IntAc
       error_flag = error_writing_file
       return
    END IF

    ! WARNING: this code breaks down if value=-128 because 
    !          a 1 byte integer cannot hold the value +128 !!!!!!
    ! Therefore many compilers let the abs() functions return
    ! with -128 if you do this .....
    ! We need to discuss the ranges and missing numbers 
    ! and agree on them with ESA.
    !write(str_value, format_IntAc, err=999) abs(value)

    IF (value .eq. -128) THEN
       str_value = "-128"
    ELSE
       write(str_value, format_IntAc, err=999) abs(value)
       call Add_Sign_to_KVT_String(str_value,value.ge.0,error_flag)
       IF (error_flag .ne. no_error) return
    END IF
    return

999 print *,"ERROR in SetIntAc:"
    print *,"Conversion from type IntAc to string failed for value: ",value
    error_flag = error_writing_file
    return

  end subroutine SetIntAc
    !  #]
  subroutine SetIntAus(value,str_value,error_flag)
    !  #[

    integer(IntAus),              intent(in)  :: value
    character(len=maxlen_IntAus), intent(out) :: str_value
    integer,                      intent(out) :: error_flag

    error_flag = no_error

    ! check: see if value falls within allowed range
    IF ( (value .lt. minval_IntAus) .or. (value .gt. maxval_IntAus)) THEN
       print *,"value out of range in: SetIntAus(), value = ",value
       print *,"allowed range: ",minval_IntAus," ... ",maxval_IntAus
       error_flag = error_writing_file
       return
    END IF

    write(str_value, format_IntAus, err=999) abs(value)

    call Add_Sign_to_KVT_String(str_value,value.ge.0,error_flag)
    IF (error_flag .ne. no_error) return

    return

999 print *,"ERROR in SetIntAus:"
    print *,"Conversion from type IntAus to string failed for value: ",value
    error_flag = error_writing_file
    return

  end subroutine SetIntAus

    !  #]
  subroutine SetIntAs(value,str_value,error_flag)
    !  #[
    integer(IntAs),              intent(in)  :: value
    character(len=maxlen_IntAs), intent(out) :: str_value
    integer,                     intent(out) :: error_flag

    error_flag = no_error

    ! check: see if value falls within allowed range
    IF ( (value .lt. minval_IntAs) .or. (value .gt. maxval_IntAs)) THEN
       print *,"value out of range in: SetIntAs(), value = ",value
       print *,"allowed range: ",minval_IntAs," ... ",maxval_IntAs
       error_flag = error_writing_file
       return
    END IF

    ! WARNING: this code breaks down if value=-32768 because 
    !          a 2 byte integer cannot hold the value +32768 !!!!!!
    ! Therefore many compilers let the abs() functions return
    ! with -32768 if you do this .....
    ! We need to discuss the ranges and missing numbers 
    ! and agree on them with ESA.
    !write(str_value, format_IntAs, err=999) abs(value)

    IF (value .eq. -32768) THEN
       str_value = "-32768"
    ELSE
       write(str_value, format_IntAs, err=999) abs(value)
       call Add_Sign_to_KVT_String(str_value,value.ge.0,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    return

999 print *,"ERROR in SetIntAs:"
    print *,"Conversion from type IntAs to string failed for value: ",value
    error_flag = error_writing_file
    return

  end subroutine SetIntAs
    !  #]
  subroutine SetIntAul(value,str_value,error_flag)
    !  #[
    integer(IntAul),              intent(in)  :: value
    character(len=maxlen_IntAul), intent(out) :: str_value
    integer,                      intent(out) :: error_flag

    error_flag = no_error

    ! check: see if value falls within allowed range
    IF ( (value .lt. minval_IntAul) .or. (value .gt. maxval_IntAul)) THEN
       print *,"value out of range in: SetIntAul(), value = ",value
       print *,"allowed range: ",minval_IntAul," ... ",maxval_IntAul
       error_flag = error_writing_file
       return
    END IF

    write(str_value, format_IntAul, err=999) abs(value)

    call Add_Sign_to_KVT_String(str_value,value.ge.0,error_flag)
    IF (error_flag .ne. no_error) return

    return

999 print *,"ERROR in SetIntAul:"
    print *,"Conversion from type IntAul to string failed for value: ",value
    error_flag = error_writing_file
    return

  end subroutine SetIntAul
    !  #]
  subroutine SetIntAl(value,str_value,error_flag)
    !  #[
    integer(IntAl),              intent(in)  :: value
    character(len=maxlen_IntAl), intent(out) :: str_value
    integer,                     intent(out) :: error_flag

    error_flag = no_error

    ! check: see if value falls within allowed range
    IF ( (value .lt. minval_IntAl) .or. (value .gt. maxval_IntAl)) THEN
       print *,"value out of range in: SetIntAl(), value = ",value
       print *,"allowed range: ",minval_IntAl," ... ",maxval_IntAl
       error_flag = error_writing_file
       return
    END IF

    ! WARNING: this code breaks down if value=-2147483648 because 
    !          a 4 byte integer cannot hold the value +2147483648 !!!!!!
    ! Therefore many compilers let the abs() functions return
    ! with -2147483648 if you do this .....
    ! We need to discuss the ranges and missing numbers 
    ! and agree on them with ESA.
    !write(str_value, format_IntAl, err=999) abs(value)

    ! warning the constant -2147483648 is forbidden in gfortran
    ! so use minval_IntAl instead (which contains the same number)
    !IF (value .eq. -2147483648) THEN
    IF (value .eq. minval_IntAl) THEN
       str_value = "-2147483648"
    ELSE
       write(str_value, format_IntAl, err=999) abs(value)
       call Add_Sign_to_KVT_String(str_value,value.ge.0,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    return

999 print *,"ERROR in SetIntAl:"
    print *,"Conversion from type IntAl to string failed for value: ",value
    error_flag = error_writing_file
    return

  end subroutine SetIntAl
    !  #]
  subroutine SetIntAd(value,str_value,error_flag)
    !  #[
    integer(IntAd),              intent(in)  :: value
    character(len=maxlen_IntAd), intent(out) :: str_value
    integer,                     intent(out) :: error_flag

    error_flag = no_error

    ! check: see if value falls within allowed range
    IF ( (value .lt. minval_IntAd) .or. (value .gt. maxval_IntAd)) THEN
       print *,"value out of range in: SetIntAd(), value = ",value
       print *,"allowed range: ",minval_IntAd," ... ",maxval_IntAd
       error_flag = error_writing_file
       return
    END IF

    write(str_value, format_IntAd, err=999) abs(value)

    call Add_Sign_to_KVT_String(str_value,value.ge.0,error_flag)
    IF (error_flag .ne. no_error) return

    return

999 print *,"ERROR in SetIntAd:"
    print *,"Conversion from type IntAd to string failed for value: ",value
    error_flag = error_writing_file
    return

  end subroutine SetIntAd
    !  #]

  subroutine SetIntAc_or_missing(value,str_value,error_flag)
    !  #[
    integer(IntAc),              intent(in)  :: value
    character(len=maxlen_IntAc), intent(out) :: str_value
    integer,                     intent(out) :: error_flag

    error_flag = no_error

    str_value(:)=' '
    IF (value .eq. missing_indicator_IntAc) THEN
       str_value = "[mi]"
    ELSE
       call SetIntAc(value,str_value,error_flag)
       if (error_flag .ne. no_error) return
    END IF

  end subroutine SetIntAc_or_missing
    !  #]
  subroutine SetIntAus_or_missing(value,str_value,error_flag)
    !  #[
    integer(IntAus),              intent(in)  :: value
    character(len=maxlen_IntAus), intent(out) :: str_value
    integer,                      intent(out) :: error_flag

    error_flag = no_error

    ! warning: using a generic missing_integer() routine is not possible
    ! here, since for the fortran compier IntAus and IntAl are identical
    ! since they are both mapped onto the i4_ 4 byte integer type.
    IF (value .eq. missing_indicator_IntAus) THEN
       str_value = "[miss]"
    ELSE
       call SetIntAus(value,str_value,error_flag)
       if (error_flag .ne. no_error) return
    END IF

  end subroutine SetIntAus_or_missing
    !  #]
  subroutine SetIntAs_or_missing(value,str_value,error_flag)
    !  #[
    integer(IntAs),              intent(in)  :: value
    character(len=maxlen_IntAs), intent(out) :: str_value
    integer,                     intent(out) :: error_flag

    error_flag = no_error

    ! warning: using a generic missing_integer() routine is not possible
    ! here, since for the fortran compier IntAs and IntAuc are identical
    ! since they are both mapped onto the i2_ 2 byte integer type.
    IF (value .eq. missing_indicator_IntAs) THEN
       str_value = "[miss]"
    ELSE
       call SetIntAs(value,str_value,error_flag)
       if (error_flag .ne. no_error) return
    END IF

  end subroutine SetIntAs_or_missing
    !  #]
  subroutine SetIntAul_or_missing(value,str_value,error_flag)
    !  #[
    integer(IntAul),              intent(in)  :: value
    character(len=maxlen_IntAul), intent(out) :: str_value
    integer,                      intent(out) :: error_flag

    error_flag = no_error

    ! warning: using a generic missing_integer() routine is not possible
    ! here, since for the fortran compier IntAul and IntAd are identical
    ! since they are both mapped onto the i8_ 8 byte integer type.
    IF (value .eq. missing_indicator_IntAul) THEN
       str_value = "[miss]"
    ELSE
       call SetIntAul(value,str_value,error_flag)
       if (error_flag .ne. no_error) return
    END IF

  end subroutine SetIntAul_or_missing
    !  #]
  subroutine SetIntAl_or_missing(value,str_value,error_flag)
    !  #[
    integer(IntAl),              intent(in)  :: value
    character(len=maxlen_IntAl), intent(out) :: str_value
    integer,                     intent(out) :: error_flag

    error_flag = no_error

    ! warning: using a generic missing_integer() routine is not possible
    ! here, since for the fortran compier IntAus and IntAl are identical
    ! since they are both mapped onto the i4_ 4 byte integer type.
    IF (value .eq. missing_indicator_IntAl) THEN
       str_value = "[missing]"
    ELSE
       call SetIntAl(value,str_value,error_flag)
       if (error_flag .ne. no_error) return
    END IF

  end subroutine SetIntAl_or_missing
    !  #]

  subroutine SetIntAucFlags(value,str_value,error_flag)
    !  #[

    integer(IntAuc),                   intent(in)  :: value
    character(len=maxlen_IntAucFlags), intent(out) :: str_value
    integer,                           intent(out) :: error_flag

    ! local variable
    integer :: bit

    error_flag = no_error

    ! check: see if value falls within allowed range
    IF ( (value .lt. minval_IntAuc) .or. (value .gt. maxval_IntAuc)) THEN
       print *,"value out of range in: SetIntAucFlags(), value = ",value
       print *,"allowed range: ",minval_IntAuc," ... ",maxval_IntAuc
       error_flag = error_writing_file
       return
    END IF

    str_value(:) = ' '
    DO bit=1,8
       IF (btest(value,bit-1)) THEN
          str_value(bit:bit) = 'T'
       ELSE
          str_value(bit:bit) = 'F'
       END IF
    END DO

    return

!999 print *,"ERROR in SetIntAucFlags:"
!    print *,"Conversion from type IntAuc to flags-string failed for value: ",&
!         value
!    error_flag = error_writing_file
!    return

  end subroutine SetIntAucFlags

    !  #]

  subroutine SetFAdoxy(value,x,y,str_value,error_flag)
    !  #[
    real(FAdoxy),     intent(in)  :: value
    integer,          intent(in)  :: x ! ndigits before .
    integer,          intent(in)  :: y ! ndigits after  .
    character(len=*), intent(out) :: str_value
    integer,          intent(out) :: error_flag

    ! local variables
    real(FAdoxy) :: tmp_value, min_value, max_value, eps
    character(len=2) :: xy
    character(len=7) :: format_FAdoxy

    ! initialise
    error_flag = no_error

    eps = 10.**(-1*y)
    max_value = 10.**x - eps
    min_value = -1.*max_value 

    write(xy,"(i1,i1)") x,y
    write(format_FAdoxy,"('(f',i2.2,'.',i1,')')") x+y+2,y

    ! check to see if length for str_value is correct
    if (len(str_value) .ne. x+y+2) then
       print *,"Error in: SetFAdoxy():"
       print *,"length of str_value does not match setting for x,y !!!"
       print *,"expected length = x+y+2 = ",x+y+2," (xy=",xy,")"
       print *,"but: len(str_value) = ",len(str_value)
       print *,"please use the correct maxlen_FAdoxy parameter to"
       print *,"define a string of the correct length."
       error_flag = error_programming
       return
    end if

    ! workaround for missing numbers
    tmp_value = value
    IF (missing_real(value)) tmp_value = max_value

    !print *,"test inside SetFAdoxy:"
    !print *,"xy = ",xy," value = ",tmp_value
    !print *,"allowed range: ",min_value," ... ",max_value
    !print *,"format_FAdoxy = [",format_FAdoxy,"]"
    !stop 1

    ! check: see if value falls within allowed range
    !        taking care of representation/rounding errors
    IF ( (tmp_value .lt. min_value-1.1*eps) .or. &
         (tmp_value .gt. max_value+1.1*eps)      ) THEN
       print *,"value out of range in: SetFAdoxy():"
       print *,"xy = ",xy," value = ",tmp_value
       print *,"allowed range: ",min_value," ... ",max_value
       print *,"allowed range+/-eps: ",&
            min_value-1.1*eps," ... ",max_value+1.1*eps
       error_flag = error_writing_file
       return
    END IF
    IF (tmp_value .lt. min_value) tmp_value = min_value
    IF (tmp_value .gt. max_value) tmp_value = max_value

    write(str_value, format_FAdoxy, err=999) abs(tmp_value)

    call Add_Sign_to_KVT_String(str_value,tmp_value.ge.0.,error_flag,&
                                no_digits_before_point=(x.eq.0))

    IF (error_flag .ne. no_error) return

    return

999 print *,"ERROR in SetFAdoxy:"
    print *,"Conversion from type FAdoxy to string failed for value: ",value
    error_flag = error_writing_file
    return

  end subroutine SetFAdoxy
    !  #]
  subroutine SetFAdo06(value,str_value,error_flag)
    !  #[
    real(FAdoxy),                 intent(in)  :: value
    character(len=maxlen_FAdo06), intent(out) :: str_value
    integer,                      intent(out) :: error_flag

    ! local variable
    real(FAdoxy) :: tmp_value

    error_flag = no_error

    ! workaround for missing numbers
    tmp_value = value
    IF (missing_real(value)) tmp_value = minval_FAdo06

    ! check: see if value falls within allowed range
    IF ( (tmp_value .lt. minval_FAdo06) .or. &
         (tmp_value .gt. maxval_FAdo06)      ) THEN
       print *,"value out of range in: SetFAdo06(), value = ",tmp_value
       print *,"allowed range: ",minval_FAdo06," ... ",maxval_FAdo06
       error_flag = error_writing_file
       return
    END IF

    write(str_value, format_FAdo06, err=999) abs(tmp_value)

    call Add_Sign_to_KVT_String(str_value,tmp_value.ge.0.,error_flag,&
                                no_digits_before_point=.true.)
    IF (error_flag .ne. no_error) return

    return

999 print *,"ERROR in SetFAdo06:"
    print *,"Conversion from type FAdo06 to string failed for value: ",value
    error_flag = error_writing_file
    return

  end subroutine SetFAdo06
    !  #]
  subroutine SetFAdo22(value,str_value,error_flag)
    !  #[
    real(FAdoxy),                 intent(in)  :: value
    character(len=maxlen_FAdo22), intent(out) :: str_value
    integer,                      intent(out) :: error_flag

    ! local variable
    real(FAdoxy) :: tmp_value

    error_flag = no_error

    ! workaround for missing numbers
    tmp_value = value
    IF (missing_real(value)) tmp_value = minval_FAdo22

    ! check: see if value falls within allowed range
    IF ( (tmp_value .lt. minval_FAdo22) .or. &
         (tmp_value .gt. maxval_FAdo22)       ) THEN
       print *,"value out of range in: SetFAdo22(), value = ",tmp_value
       print *,"allowed range: ",minval_FAdo22," ... ",maxval_FAdo22
       error_flag = error_writing_file
       return
    END IF

    write(str_value, format_FAdo22, err=999) abs(tmp_value)

    call Add_Sign_to_KVT_String(str_value,tmp_value.ge.0.,error_flag)
    IF (error_flag .ne. no_error) return

    return

999 print *,"ERROR in SetFAdo22:"
    print *,"Conversion from type FAdo22 to string failed for value: ",value
    error_flag = error_writing_file
    return

  end subroutine SetFAdo22
    !  #]
  subroutine SetFAdo32(value,str_value,error_flag)
    !  #[
    real(FAdoxy),                 intent(in)  :: value
    character(len=maxlen_FAdo32), intent(out) :: str_value
    integer,                      intent(out) :: error_flag

    ! local variable
    real(FAdoxy) :: tmp_value

    error_flag = no_error

    ! workaround for missing numbers
    tmp_value = value
    IF (missing_real(value)) tmp_value = minval_FAdo32

    ! check: see if value falls within allowed range
    IF ( (tmp_value .lt. minval_FAdo32) .or. &
         (tmp_value .gt. maxval_FAdo32)) THEN
       print *,"value out of range in: SetFAdo32(), value = ",tmp_value
       print *,"allowed range: ",minval_FAdo32," ... ",maxval_FAdo32
       error_flag = error_writing_file
       return
    END IF

    write(str_value, format_FAdo32, err=999) abs(tmp_value)

    call Add_Sign_to_KVT_String(str_value,tmp_value.ge.0.,error_flag)
    IF (error_flag .ne. no_error) return

    return

999 print *,"ERROR in SetFAdo32:"
    print *,"Conversion from type FAdo32 to string failed for value: ",value
    error_flag = error_writing_file
    return

  end subroutine SetFAdo32
    !  #]
  subroutine SetFAdo36(value,str_value,error_flag)
    !  #[
    real(FAdoxy),                 intent(in)  :: value
    character(len=maxlen_FAdo36), intent(out) :: str_value
    integer,                      intent(out) :: error_flag

    ! local variable
    real(FAdoxy) :: tmp_value

    error_flag = no_error

    ! workaround for missing numbers
    tmp_value = value
    IF (missing_real(value)) tmp_value = minval_FAdo36

    ! check: see if value falls within allowed range
    IF ( (tmp_value .lt. minval_FAdo36) .or. &
         (tmp_value .gt. maxval_FAdo36)      ) THEN
       print *,"value out of range in: SetFAdo36(), value = ",tmp_value
       print *,"allowed range: ",minval_FAdo36," ... ",maxval_FAdo36
       error_flag = error_writing_file
       return
    END IF

    write(str_value, format_FAdo36, err=999) abs(tmp_value)

    call Add_Sign_to_KVT_String(str_value,tmp_value.ge.0.,error_flag)
    IF (error_flag .ne. no_error) return

    return

999 print *,"ERROR in SetFAdo36:"
    print *,"Conversion from type FAdo36 to string failed for value: ",value
    error_flag = error_writing_file
    return

  end subroutine SetFAdo36
    !  #]
  subroutine SetFAdo46(value,str_value,error_flag)
    !  #[
    real(FAdoxy),                 intent(in)  :: value
    character(len=maxlen_FAdo46), intent(out) :: str_value
    integer,                      intent(out) :: error_flag

    ! local variable
    real(FAdoxy) :: tmp_value

    error_flag = no_error

    ! workaround for missing numbers
    tmp_value = value
    IF (missing_real(value)) tmp_value = minval_FAdo46

    ! check: see if value falls within allowed range
    IF ( (tmp_value .lt. minval_FAdo46) .or. &
         (tmp_value .gt. maxval_FAdo46)       ) THEN
       print *,"value out of range in: SetFAdo46(), value = ",tmp_value
       print *,"allowed range: ",minval_FAdo46," ... ",maxval_FAdo46
       error_flag = error_writing_file
       return
    END IF

    write(str_value, format_FAdo46, err=999) abs(tmp_value)

    call Add_Sign_to_KVT_String(str_value,tmp_value.ge.0.,error_flag)
    IF (error_flag .ne. no_error) return

    return

999 print *,"ERROR in SetFAdo46:"
    print *,"Conversion from type FAdo46 to string failed for value: ",value
    error_flag = error_writing_file
    return

  end subroutine SetFAdo46
    !  #]
  subroutine SetFAdo73(value,str_value,error_flag)
    !  #[
    real(FAdoxy),                 intent(in)  :: value
    character(len=maxlen_FAdo73), intent(out) :: str_value
    integer,                      intent(out) :: error_flag

    ! local variable
    real(FAdoxy) :: tmp_value

    error_flag = no_error

    ! workaround for missing numbers
    tmp_value = value
    IF (missing_real(value)) tmp_value = minval_FAdo73

    ! check: see if value falls within allowed range
    IF ( (tmp_value .lt. minval_FAdo73) .or. &
         (tmp_value .gt. maxval_FAdo73)      ) THEN
       print *,"value out of range in: SetFAdo73(), value = ",tmp_value
       print *,"allowed range: ",minval_FAdo73," ... ",maxval_FAdo73
       error_flag = error_writing_file
       return
    END IF

    write(str_value, format_FAdo73, err=999) abs(tmp_value)

    call Add_Sign_to_KVT_String(str_value,tmp_value.ge.0.,error_flag)
    IF (error_flag .ne. no_error) return

    return

999 print *,"ERROR in SetFAdo73:"
    print *,"Conversion from type FAdo73 to string failed for value: ",value
    error_flag = error_writing_file
    return

  end subroutine SetFAdo73
    !  #]

  ! remark: we could also give x and y to the next routine and
  !         check for the correct location of the dot, but thats not yet
  !         implemented now
  subroutine GetFAdoxy(str_value,value,error_flag)
    !  #[
    character(len=*), intent(in)   :: str_value
    real(FAdoxy),     intent(out)  :: value
    integer,          intent(out)  :: error_flag    

    ! local variable
    integer :: ios

    read(str_value,*,iostat=ios) value
    IF (ios .ne. 0) goto 999

    return

999 print *,"ERROR in GetFAdoxy:"
    print *,"Conversion from string to FAdoxy failed for string: ["//&
         trim(str_value)//"]"
    error_flag = error_reading_file
    return

  end subroutine GetFAdoxy
    !  #]
  ! remark: I know there is a real8_to_text routine in Numerics
  !         but I want to print the full accuracy of the reals, and
  !         real8_to_text does not use the default"*" format, so might
  !         loose some digits, so I defined the routine below for that
  ! besides, I think these routines handling the ee_cfi types like FAdoxy
  ! belong in this module. JK.
  subroutine SetFAdoFull_or_Missing(value,str_value,error_flag)
    !  #[
    real(FAdoxy),     intent(in)  :: value
    character(len=*), intent(out) :: str_value
    integer,          intent(out) :: error_flag

    ! local variable
    character(len=50) :: temp_string
    integer           :: needed_length

    error_flag = no_error

    ! give the string "[missing]" for missing numbers
    IF (missing_real(value)) THEN
       str_value = "[missing]"
    ELSE IF (is_NaN_real8(value)) THEN
       str_value = "[not-a-number]"
    ELSE IF (is_PosInf_real8(value)) THEN
       str_value = "[+Infinite]"
    ELSE IF (is_NegInf_real8(value)) THEN
       str_value = "[-Infinite]"
    ELSE
       write(temp_string, *, err=999) value
       needed_length = len_trim(adjustl(temp_string))
       IF (needed_length .gt. len(str_value) ) THEN
          print *,"ERROR in SetFAdoFull_or_Missing:"
          print *,"Conversion from type FAdoxy to string failed for value: ",&
               value
          print *,"available string str_value has length ",len(str_value)
          print *,"needed_length = ",needed_length
          print *,"temp_string = [",trim(temp_string),"]"
          print *,"the provided string is too short for your purpose!"
          error_flag = error_writing_file
          return
       END IF
       str_value =  trim(adjustl(temp_string))
    END IF

    return

999 print *,"ERROR in SetFAdoFull_or_Missing:"
    print *,"Conversion from type FAdoxy to string failed for value: ",value
    print *,"(Are you sure this fits within the 50 characters of the"
    print *," available temporaty string ?)"
    error_flag = error_writing_file
    return
    
  end subroutine SetFAdoFull_or_Missing
  !  #]

  subroutine Add_Sign_to_KVT_String(str_value,flag_plus_or_minus,&
                                    error_flag,no_digits_before_point)
    !  #[
    ! J. de Kloe    Original code
    ! P. Poli       Modified to add sign at the beginning of the string
    !               and zeroes in between the sign and the beginning of
    !               the number (example: -003)
    ! J. de Kloe 01-03-2006 added no_digits_before_point switch
    !
    character(len=*),  intent(inout) :: str_value
    logical,           intent(in)    :: flag_plus_or_minus
    integer,           intent(out)   :: error_flag
    logical, optional, intent(in)    :: no_digits_before_point

    ! local variable
    logical :: use_no_digits_before_point
    logical :: valid_string

    ! Remark1:
    ! if flag_plus_or_minus == .T. : add a '+' sign, otherwise, add a '-' sign
    ! Remark2:
    ! for FAdoxy types with x=0 there should not be printed any digits before
    ! the decimal point. To achieve this some exceptions are needed below.

    error_flag = no_error

    use_no_digits_before_point = .false.
    IF (present(no_digits_before_point)) THEN
       use_no_digits_before_point = no_digits_before_point
    END IF

    valid_string = .true.
    IF (use_no_digits_before_point) THEN
       IF (str_value(1:1) .ne. '0') valid_string = .false.
    ELSE
       IF (str_value(1:1) .ne. ' ') valid_string = .false.
    END IF
    IF (.not. valid_string) THEN
       print *,"ERROR in Add_Sign_to_KVT_String():"
       print *,"it seems that the string in input does not"
       print *,"have room for the sign. The string is (between pipes)"
       print *,"|"//str_value//"|"
       print *,"Something must be wrong with the file writing code !!"
       error_flag = error_programming
       return
    ENDIF
    IF (flag_plus_or_minus) THEN
       str_value(1:1)='+'
    ELSE
       str_value(1:1)='-'
    ENDIF

    call Fillblanks(str_value)

  end subroutine Add_Sign_to_KVT_String
    !  #]
  function Remove_Plus_Sign(str_value_in) result(str_value_out)
    !  #[
    !  Modifications
    !    01-12-2005  P. Poli    Removing also zeros between sign and number
    character(len=*), intent(in)     :: str_value_in
    character(len=len(str_value_in)) :: str_value_out

    ! local variables
    integer :: pos, ls
    logical :: exit_loop

    ! copy input to output first
    str_value_out = str_value_in

!    ! see if the '+' sign is present, and if so remove it
!    pos = scan(trim(str_value_in),"+")
!    IF (pos .ne. 0) THEN
!       ! a "+" character seems present
!       str_value_out(pos:pos) = " "
!       str_value_out = adjustl(str_value_out)
!    END IF

    IF (len_trim(str_value_out) .ne. 0) THEN
    pos = 1
    exit_loop = .false.
    DO WHILE ( (str_value_out(pos:pos) .ne. "1" .or. &
              & str_value_out(pos:pos) .ne. "2" .or. &
              & str_value_out(pos:pos) .ne. "3" .or. &
              & str_value_out(pos:pos) .ne. "4" .or. &
              & str_value_out(pos:pos) .ne. "5" .or. &
              & str_value_out(pos:pos) .ne. "6" .or. &
              & str_value_out(pos:pos) .ne. "7" .or. &
              & str_value_out(pos:pos) .ne. "8" .or. &
              & str_value_out(pos:pos) .ne. "9") .and. &
              & .not. (exit_loop))
       ! is it a plus sign ?
       IF (str_value_out(pos:pos) .eq. "+") THEN
          ! -- it is a plus sign. Remove it.
          str_value_out(pos:pos) = " "
          str_value_out = adjustl(str_value_out)
       ELSE
          ! -- it is not a plus sign.
          ! is it a minus sign ?
          IF (str_value_out(pos:pos) .eq. "-") THEN
             ! -- it is a minus sign. Move forward.
             pos = pos + 1
          ELSE
             ! -- it is not a minus sign.
             ! is it a zero ?
             IF (str_value_out(pos:pos) .eq. "0") THEN
                ! -- it is a zero
                ls = len_trim(str_value_out)
                ! is there a character after ?
                IF (ls .gt. pos) THEN
                   ! -- there is a character after
                   ! is that character a dot ?
                   IF (str_value_out(pos+1:pos+1) .eq. ".") THEN
                      ! -- it is a dot. done.
                      exit_loop = .true.
                   ELSE
                      ! -- there is no dot after the zero. Remove the zero.
                      str_value_out(pos:pos) = " "
                      str_value_out(pos:ls-1) = &
                      & str_value_out(pos+1:ls)
                      str_value_out(ls:ls) = " "
                   ENDIF
                ELSE
                   ! -- there is no character after the zero. done.
                   exit_loop = .true.
                ENDIF
             ELSE
                ! it is neither a plus, nor a minus, nor a zero. done.
                exit_loop = .true.
             ENDIF
          ENDIF
       ENDIF
    ENDDO
    ENDIF

  end function Remove_Plus_Sign
    !  #]
  !---------------------------------------
  ! conversion from real(r_) to Int* types, taking missing into account
  !---------------------------------------

! to be removed (since we should not use the r4_ type anywhere

!  function Real4_to_IntAuc(rval) result(ival)
    !  #[
!    real(r4_)       :: rval ! input
!    integer(IntAuc) :: ival ! result
!
!    IF (missing_real(rval)) THEN
!       ival = missing_indicator_IntAuc
!    ELSE
!       !convert using floor
!       ival = int(floor(rval),IntAuc)
!    END IF
!
!    ! check: see if value falls within allowed range
!    IF ( (ival .lt. minval_IntAuc) .or. (ival .gt. maxval_IntAuc)) THEN
!       print *,"value out of range in: Real4_to_IntAuc(), value = ",ival
!       print *,"allowed range: ",minval_IntAuc," ... ",maxval_IntAuc
!       print *,"returning with missing indicator ..."
!       ival = missing_indicator_IntAuc
!
!       ! if needed, we could convert this function to a subroutine
!       ! with the extra error_flag as result...
!       !error_flag = error_writing_file
!       !return
!    END IF
!
!  END function Real4_to_IntAuc
    !  #]
!  function Real4_to_IntAs(rval) result(ival)
    !  #[
!    real(r4_)      :: rval ! input
!    integer(IntAs) :: ival ! result
!
!    IF (missing_real(rval)) THEN
!       ival = missing_indicator_IntAs
!    ELSE
!       !convert using floor
!       ival = int(floor(rval),IntAs)
!    END IF
!
!    ! check: see if value falls within allowed range
!    IF ( (ival .lt. minval_IntAs) .or. (ival .gt. maxval_IntAs)) THEN
!       print *,"value out of range in: Real4_to_IntAs(), value = ",ival
!       print *,"allowed range: ",minval_IntAs," ... ",maxval_IntAs
!       print *,"returning with missing indicator ..."
!       ival = missing_indicator_IntAs
!
!       ! if needed, we could convert this function to a subroutine
!       ! with the extra error_flag as result...
!       !error_flag = error_writing_file
!       !return
!    END IF
!
!  END function Real4_to_IntAs
    !  #]
!  function Real4_to_IntAus(rval) result(ival)
    !  #[
!    real(r4_)       :: rval ! input
!    integer(IntAus) :: ival ! result
!
!    IF (missing_real(rval)) THEN
!       ival = missing_indicator_IntAus
!    ELSE
!       !convert using floor
!       ival = int(floor(rval),IntAus)
!    END IF
!
!    ! check: see if value falls within allowed range
!    IF ( (ival .lt. minval_IntAus) .or. (ival .gt. maxval_IntAus)) THEN
!       print *,"value out of range in: Real4_to_IntAus(), value = ",ival
!       print *,"allowed range: ",minval_IntAus," ... ",maxval_IntAus
!       print *,"returning with missing indicator ..."
!       ival = missing_indicator_IntAus
!
!       ! if needed, we could convert this function to a subroutine
!       ! with the extra error_flag as result...
!       !error_flag = error_writing_file
!       !return
!    END IF
!
!  END function Real4_to_IntAus
    !  #]
!  function Real4_to_IntAl(rval) result(ival)
    !  #[
!    real(r4_)       :: rval ! input
!    integer(IntAl) :: ival ! result
!
!    IF (missing_real(rval)) THEN
!       ival = missing_indicator_IntAl
!    ELSE
!       !convert using floor
!       ival = int(floor(rval),IntAl)
!    END IF
!
!    ! check: see if value falls within allowed range
!    IF ( (ival .lt. minval_IntAl) .or. (ival .gt. maxval_IntAl)) THEN
!       print *,"value out of range in: Real4_to_IntAl(), value = ",ival
!       print *,"allowed range: ",minval_IntAl," ... ",maxval_IntAl
!       print *,"returning with missing indicator ..."
!       ival = missing_indicator_IntAl
!
!       ! if needed, we could convert this function to a subroutine
!       ! with the extra error_flag as result...
!       !error_flag = error_writing_file
!       !return
!    END IF
!
!  END function Real4_to_IntAl
    !  #]
!  function Real4_to_IntAul(rval) result(ival)
    !  #[
!    real(r4_)       :: rval ! input
!    integer(IntAul) :: ival ! result
!
!    IF (missing_real(rval)) THEN
!       ival = missing_indicator_IntAul
!    ELSE
!       !convert using floor
!       !ival = floor(rval,IntAul)
!       ! DT: Specification of IntAul must be omitted on morgana.
!       ! Is it ok for other platforms to omit the spec always?
!       ! JK: yes, I looked for it in my documentation and the IntAul 
!       !     specification for floor() is Fortran95 not Fortran90.
!       !     It can be done as follows in Fortran90:
!       ival = int(floor(rval),IntAul)
!       !ival = floor(rval)
!       ival = int(floor(rval),IntAul) 
!    END IF
!
!    ! check: see if value falls within allowed range
!    IF ( (ival .lt. minval_IntAul) .or. (ival .gt. maxval_IntAul)) THEN
!       print *,"value out of range in: Real4_to_IntAul(), value = ",ival
!       print *,"allowed range: ",minval_IntAul," ... ",maxval_IntAul
!       print *,"returning with missing indicator ..."
!       ival = missing_indicator_IntAul
!
!       ! if needed, we could convert this function to a subroutine
!       ! with the extra error_flag as result...
!       !error_flag = error_writing_file
!       !return
!    END IF
!
!  END function Real4_to_IntAul
    !  #]

  function FAdoxy_to_IntAuc(rval,scale) result(ival)
    !  #[
    real(r8_),           intent(in) :: rval  ! input
    integer(IntAuc)                 :: ival  ! result
    real(r8_), optional, intent(in) :: scale ! optionalinput

    IF (missing_real(rval)) THEN
       ival = missing_indicator_IntAuc
    ELSE
       !convert using floor
       if (present(scale)) then
          ival = int(floor(rval*scale),IntAuc)
       else
          ival = int(floor(rval),IntAuc)
       end if
    END IF

    ! check: see if value falls within allowed range
    IF ( (ival .lt. minval_IntAuc) .or. (ival .gt. maxval_IntAuc)) THEN
       print *,"value out of range in: Real8_to_IntAuc(), value = ",ival
       print *,"allowed range: ",minval_IntAuc," ... ",maxval_IntAuc
       print *,"returning with missing indicator ..."
       ival = missing_indicator_IntAuc

       ! if needed, we could convert this function to a subroutine
       ! with the extra error_flag as result...
       !error_flag = error_writing_file
       !return
    END IF

  END function FAdoxy_to_IntAuc
    !  #]
  function FAdoxy_to_IntAs(rval,scale) result(ival)
    !  #[
    real(r8_),           intent(in)   :: rval  ! input
    integer(IntAs)                    :: ival  ! result
    real(r8_), optional, intent(in)   :: scale ! optionalinput

    IF (missing_real(rval)) THEN
       ival = missing_indicator_IntAs
    ELSE
       !convert using floor
       if (present(scale)) then
          ival = int(floor(rval*scale),IntAs)
       else
          ival = int(floor(rval),IntAs)
       end if
    END IF

    ! check: see if value falls within allowed range
    IF ( (ival .lt. minval_IntAs) .or. (ival .gt. maxval_IntAs)) THEN
       print *,"value out of range in: Real8_to_IntAs(), value = ",ival
       print *,"allowed range: ",minval_IntAs," ... ",maxval_IntAs
       print *,"returning with missing indicator ..."
       ival = missing_indicator_IntAs

       ! if needed, we could convert this function to a subroutine
       ! with the extra error_flag as result...
       !error_flag = error_writing_file
       !return
    END IF

  END function FAdoxy_to_IntAs
    !  #]
  function FAdoxy_to_IntAus(rval,scale) result(ival)
    !  #[
    real(r8_),           intent(in) :: rval  ! input
    integer(IntAus)                 :: ival  ! result
    real(r8_), optional, intent(in) :: scale ! optionalinput

    IF (missing_real(rval)) THEN
       ival = missing_indicator_IntAus
    ELSE
       !convert using floor
       if (present(scale)) then
          ival = int(floor(rval*scale),IntAus)
       else
          ival = int(floor(rval),IntAus)
       end if
    END IF

    ! check: see if value falls within allowed range
    IF ( (ival .lt. minval_IntAus) .or. (ival .gt. maxval_IntAus)) THEN
       print *,"value out of range in: Real8_to_IntAus(), value = ",ival
       print *,"allowed range: ",minval_IntAus," ... ",maxval_IntAus
       print *,"returning with missing indicator ..."
       ival = missing_indicator_IntAus

       ! if needed, we could convert this function to a subroutine
       ! with the extra error_flag as result...
       !error_flag = error_writing_file
       !return
    END IF

  END function FAdoxy_to_IntAus
    !  #]
  function FAdoxy_to_IntAl(rval,scale) result(ival)
    !  #[
    real(r8_),           intent(in) :: rval  ! input
    integer(IntAl)                  :: ival  ! result
    real(r8_), optional, intent(in) :: scale ! optionalinput

    IF (missing_real(rval)) THEN
       ival = missing_indicator_IntAl
    ELSE
       !convert using floor
       if (present(scale)) then
          ival = int(floor(rval*scale),IntAl)
       else
          ival = int(floor(rval),IntAl)
       END if
    END IF

    ! check: see if value falls within allowed range
    IF ( (ival .lt. minval_IntAl) .or. (ival .gt. maxval_IntAl)) THEN
       print *,"value out of range in: Real8_to_IntAl(), value = ",ival
       print *,"allowed range: ",minval_IntAl," ... ",maxval_IntAl
       print *,"returning with missing indicator ..."
       ival = missing_indicator_IntAl

       ! if needed, we could convert this function to a subroutine
       ! with the extra error_flag as result...
       !error_flag = error_writing_file
       !return
    END IF

  END function FAdoxy_to_IntAl
    !  #]
  function FAdoxy_to_IntAul(rval,scale) result(ival)
    !  #[
    real(r8_),           intent(in) :: rval  ! input
    integer(IntAul)                 :: ival  ! result
    real(r8_), optional, intent(in) :: scale ! optionalinput

    IF (missing_real(rval)) THEN
       ival = missing_indicator_IntAul
    ELSE
       !convert using floor
       if (present(scale)) then
          ival = int(floor(rval*scale),IntAul)
       else
          ival = int(floor(rval),IntAul)
       end if
    END IF

    ! check: see if value falls within allowed range
    IF ( (ival .lt. minval_IntAul) .or. (ival .gt. maxval_IntAul)) THEN
       print *,"value out of range in: Real8_to_IntAul(), value = ",ival
       print *,"allowed range: ",minval_IntAul," ... ",maxval_IntAul
       print *,"returning with missing indicator ..."
       ival = missing_indicator_IntAul

       ! if needed, we could convert this function to a subroutine
       ! with the extra error_flag as result...
       !error_flag = error_writing_file
       !return
    END IF

  END function FAdoxy_to_IntAul
    !  #]
  !---------------------------------------
  ! routines and functions related to the derived DateTime type
  !---------------------------------------
  ! WARNING: KVT_DateTime and the  DateTimeType
  !          have higher precision in the seconds
  !          (upto 1 micro-second) so conversion KVT->XML
  !          or DateTime->XML is irreversable !
  ! WARNING2: handling of missing dates is not yet implemented
  subroutine InitDateTime(DateTime)
    !  #[
    TYPE(DateTimeType), intent(out) :: DateTime

    DateTime%daycount     = minval_IntAl ! use this as missing for the moment
    DateTime%seconds      = 0
    DateTime%microseconds = 0

  end subroutine InitDateTime
    !  #]
  subroutine SetDateTime(DateTime,y,mon,d,h,min,s)
    !  #[
    TYPE(DateTimeType), intent(out) :: DateTime
    integer,   intent(in) :: y,mon,d,h,min
    real(r8_), intent(in) :: s

    ! local variables
    type(date_type) :: date
    type(time_type) :: time

    date = set_date(y,mon,d)
    time = set_time(h,min,s)
    call InitDateTime(DateTime)
    call Set_Date_in_DateTime(DateTime,date)
    call Set_Time_in_DateTime(DateTime,time)
    
  end subroutine SetDateTime
    !  #]
  function DateTimeMissing(DateTime) result(is_missing)
    !  #[
    TYPE(DateTimeType), intent(in) :: DateTime   ! input
    logical                        :: is_missing ! result

    is_missing = .false.
    if ( (DateTime%daycount     .eq. minval_IntAl) .and. &
         (DateTime%seconds      .eq.            0) .and. &
         (DateTime%microseconds .eq.            0)       ) then
       is_missing = .true.
    end if

  end function DateTimeMissing
    !  #]
  subroutine TestDateTime(DateTime,error_flag)
    !  #[ test the validity if the provided date and time 
    TYPE(DateTimeType), intent(in)  :: DateTime
    integer,            intent(out) :: error_flag

    ! local variables
    type(date_type) :: date
    type(time_type) :: time

    error_flag = no_error

    ! missing is an allowed value
    if (DateTimeMissing(DateTime)) return

    date = Get_Date_from_DateTime(DateTime)
    time = Get_Time_from_DateTime(DateTime)
    
    call test_date(date,error_flag)
    IF (error_flag .ne. no_error) return
    call test_time(time,error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine TestDateTime
    !  #]
  subroutine PrintDateTime(DateTime)
    !  #[
    TYPE(DateTimeType), intent(in) :: DateTime

    if (DateTimeMissing(DateTime)) then
       print *,"DateTime = [missing]"
    else
       print *,"daycount:",    DateTime%daycount,&
               "seconds:",     DateTime%seconds,&
               "microseconds:",DateTime%microseconds
    end if

  end subroutine PrintDateTime
    !  #]
  subroutine IncrementDateTime(DateTime,timestep)
    !  #[
    TYPE(DateTimeType), intent(inout) :: DateTime
    real(r8_),          intent(in)    :: timestep

    ! local variables
    type(date_type) :: date
    type(time_type) :: time

    ! nothing usefull to do when the input is missing
    ! todo: maybe issue an error in this case
    if (DateTimeMissing(DateTime)) return

    date = Get_Date_from_DateTime(DateTime)
    time = Get_Time_from_DateTime(DateTime)
    
    call increment_datetime(date,time,timestep)

    call Set_Date_in_DateTime(DateTime,date)
    call Set_Time_in_DateTime(DateTime,time)

  end subroutine IncrementDateTime
    !  #]
  function get_averaged_datetime(datetime1,datetime2) result(avg_datetime)
    !  #[
    TYPE(DateTimeType), intent(in) :: datetime1
    TYPE(DateTimeType), intent(in) :: datetime2
    TYPE(DateTimeType)             :: avg_datetime ! result

    ! local variable
    real(r8_) :: timestep

    if ( DateTimeMissing(datetime1) .or. &
         DateTimeMissing(datetime2)      ) then
       ! set to missing and return
       call InitDateTime(avg_dateTime)
       return
    end if

    ! calc. howmany seconds should be added to datetime1 to get datetime2
    timestep = get_time_difference(datetime1,datetime2)

    ! add half the timestep to datetime1 to get the average
    avg_dateTime=dateTime1
    call IncrementDateTime(avg_dateTime,0.5_r8_*timestep)

  end function get_averaged_datetime
    !  #]
  function get_time_difference(datetime1,datetime2) result(time_step)
    !  #[
    TYPE(DateTimeType), intent(in) :: datetime1
    TYPE(DateTimeType), intent(in) :: datetime2
    real(r8_)                      :: time_step ! result

    if ( DateTimeMissing(datetime1) .or. &
         DateTimeMissing(datetime2)      ) then
       ! set to missing and return
       time_step = missing_indicator_real_r8
    end if

    ! calc. how many seconds should be added to datetime1 to get datetime2
    time_step = &
         24._r8_*60*60*(datetime2%daycount     - datetime1%daycount) + &
                       (datetime2%seconds      - datetime1%seconds ) + &
             1.e-6_r8_*(datetime2%microseconds - datetime1%microseconds)

  end function get_time_difference
    !  #]
  function DateTime_1_after_or_eq_2(datetime1,datetime2) &
                                    result(one_after_or_eq_two)
    !  #[
    TYPE(DateTimeType), intent(in) :: datetime1
    TYPE(DateTimeType), intent(in) :: datetime2
    logical                        :: one_after_or_eq_two ! result

    if ( DateTimeMissing(datetime1) .or. &
         DateTimeMissing(datetime2)      ) then
       ! always return false when one of the inputs is missing
       one_after_or_eq_two = .false.
       return
    end if

    ! check the days first
    if (datetime1%daycount .gt. datetime2%daycount) then
       one_after_or_eq_two = .true.
    else if (datetime1%daycount .lt. datetime2%daycount) then
       one_after_or_eq_two = .false.
    else
       ! if the daycounts are identical, check the seconds
       if (datetime1%seconds .gt. datetime2%seconds) then
          one_after_or_eq_two = .true.
       else if (datetime1%seconds .lt. datetime2%seconds) then
          one_after_or_eq_two = .false.
       else
          ! if the day and second counts are equal, check the microseconds
          if (datetime1%microseconds .gt. datetime2%microseconds) then
             one_after_or_eq_two = .true.
          else if (datetime1%microseconds .lt. datetime2%microseconds) then
             one_after_or_eq_two = .false.
          else
             ! datetime1 is identical to datetime2
             one_after_or_eq_two = .true.
          end if
       end if
    end if

  end function DateTime_1_after_or_eq_2
    !  #]
  subroutine CheckWithinTimeRange(DateTime, StartDateTime, StopDateTime, &
                                  IsWithinTimeRange, error_flag)
    !  #[
    TYPE(DateTimeType), intent(in)  :: DateTime          ! input
    TYPE(DateTimeType), intent(in)  :: StartDateTime     ! input
    TYPE(DateTimeType), intent(in)  :: StopDateTime      ! input
    logical,            intent(out) :: IsWithinTimeRange ! result/output
    integer,            intent(out) :: error_flag        ! output
    
    ! init
    error_flag        = no_error
    IsWithinTimeRange = .false.

    ! missing is always out of range
    if (DateTimeMissing(DateTime)) return

    ! if start or stop are missing issue an error
    if ( DateTimeMissing(StartDateTime) .or. & 
         DateTimeMissing(StopDateTime)       ) then
       error_flag = error_programming
       return
    end if

    IF ( DateTime_1_after_or_eq_2(DateTime,StartDateTime) .and. &
         DateTime_1_after_or_eq_2(StopDateTime,DateTime)        ) THEN
       IsWithinTimeRange = .true.
    END IF

  end subroutine CheckWithinTimeRange
    !  #]
  function Get_Date_from_DateTime(DateTime) result(date)
    !  #[
    TYPE(DateTimeType), intent(in) :: DateTime
    type(date_type) :: date ! result

    ! local variables
    integer, save :: daycount_20000101
    logical, save :: first_call = .true.
    ! the DateTimeMod defines the a daycount since 19000101,
    ! the EE_CFI daycount is defined since 20000101, so correct for this
    integer   :: intdate ! has format YYYY*100*100+MM*100+DD

    if (DateTimeMissing(DateTime)) then
       call set_date_to_missing(date)
       return
    end if

    IF (first_call) THEN
       daycount_20000101 = conv_date_to_daycount(20000101)
       first_call = .false.
    END IF

    intdate = conv_daycount_to_date(daycount_20000101+DateTime%daycount)
    date = convert_int_to_date(intdate)

    !print *,"daycount_20000101 = ",daycount_20000101
    !print *,"DateTime%daycount = ",DateTime%daycount
    !print *,"intdate = ",intdate
    !call print_date(date)

  end function Get_Date_from_DateTime
    !  #]
  function Get_Time_from_DateTime(DateTime) result(time)
    !  #[ 
    TYPE(DateTimeType), intent(in) :: DateTime
    type(time_type) :: time ! result

    ! local variable
    real(r8_) :: rtime   ! has format HH*60*60+MM*60+SS.SSSSSS

    if (DateTimeMissing(DateTime)) then
       call set_time_to_missing(time)
       return
    end if

    rtime = 1._r8_*DateTime%seconds + 1.e-6_r8_*DateTime%microseconds
    time = convert_real_to_time(rtime)

    !print *,"rtime = ",rtime
    !call print_time(time)

  end function Get_Time_from_DateTime
    !  #]
  function DateTime2String(DateTime,sep_date,sep_time,&
                           round_floor,round_ceil) result(datetimestring)
    !  #[ 
    TYPE(DateTimeType), intent(in) :: DateTime       ! inout
    character(len=19)              :: datetimestring ! result
    character(len=1), optional, intent(in) :: sep_date ! optional separator
    character(len=1), optional, intent(in) :: sep_time ! optional separator
    logical,          optional, intent(in) :: round_floor ! howto round micro-
    logical,          optional, intent(in) :: round_ceil  ! seconds (up/down)

    ! local variables
    type(date_type)   :: date
    type(time_type)   :: time
    character(len=10) :: datestring
    character(len=8)  :: timestring
    logical           :: return_missing

    return_missing = .false.
    if (present(round_floor) .and. present(round_ceil)) then
       return_missing=.true. ! programming error !
    end if

    if (DateTimeMissing(DateTime)) then
       return_missing=.true. ! datetime seems missing
    end if

    if (return_missing) then
       if (present(sep_date)) then
          datestring = "00"//sep_date//"00"//sep_date//"0000"
       else
          datestring = "00000000"
       end if
       if (present(sep_time)) then
          timestring = "00"//sep_time//"00"//sep_time//"00"
       else
          timestring = "000000"
       end if
       datetimestring = trim(datestring)//"T"//trim(timestring)
       return
    end if

    date = Get_Date_from_DateTime(DateTime)
    time = Get_Time_from_DateTime(DateTime)

    ! round downwards to whole seconds (needed for FH%Validity_Period fields)
    if (present(round_floor)) then
       if (round_floor) then
          time%second = floor(time%second)
       end if
    end if

    ! round upwards to whole seconds (needed for FH%Validity_Period fields)
    if (present(round_ceil)) then
       if (round_ceil) then
          call increment_datetime(date,time,1._r8_)
          time%second = floor(time%second)
       end if
    end if

    IF (present(sep_date)) THEN
       datestring = date2string_sep(date,sep_date)
       ! 10 chars, format: yyyy-mm-dd
       ! in which the '-' char is defined by sep_date
    ELSE
       datestring = date2string(date) ! 8 chars, format: yyyymmdd
    END IF

    IF (present(sep_time)) THEN
       timestring = time2string_sep(time,sep_time)
       !  8 chars, format: hh:mm:ss
       ! in which the ':' char is defined by sep_time
    ELSE
       timestring = time2string(time) ! 6 chars, format: hhmmss
    END IF

    datetimestring = trim(datestring)//'T'//trim(timestring)

  end function DateTime2String
    !  #]
  function GetCurrentDateTime() result(DateTimeCurrent)
    !  #[
    TYPE(DateTimeType) :: DateTimeCurrent

    ! local variables
    type(date_type) :: date_current
    type(time_type) :: time_current

    ! get the current date/time when the program is running
    call get_current_date_time(date_current,time_current)

    call Set_Date_in_DateTime(DateTimeCurrent,date_current)
    call Set_Time_in_DateTime(DateTimeCurrent,time_current)

  end function GetCurrentDateTime
    !  #]
  subroutine Set_Date_in_DateTime(DateTime,date)
    !  #[
    TYPE(DateTimeType), intent(inout) :: DateTime
    type(date_type),    intent(in)    :: date

    ! local variables
    integer, save :: daycount_20000101
    logical, save :: first_call = .true.
    ! the DateTimeMod defines the a daycount since 19000101,
    ! the EE_CFI daycount is defined since 20000101, so correct for this
    integer   :: intdate ! has format YYYY*100*100+MM*100+DD
    integer   :: daycount

    IF (first_call) THEN
       daycount_20000101 = conv_date_to_daycount(20000101)
       first_call = .false.
    END IF

    intdate = convert_date_to_int(date)
    daycount = conv_date_to_daycount(intdate)
    DateTime%daycount = int(daycount - daycount_20000101,IntAl)
    
  end subroutine Set_Date_in_DateTime
    !  #]
  subroutine Set_Time_in_DateTime(DateTime,time)
    !  #[
    TYPE(DateTimeType), intent(inout) :: DateTime
    type(time_type),    intent(in)    :: time

    ! local variable
    real(r8_) :: rtime   ! has format HH*60*60+MM*60+SS.SSSSSS

    rtime = convert_time_to_real(time)
    DateTime%seconds      = floor(rtime)
    DateTime%microseconds = nint(1.e6*(rtime - DateTime%seconds))

  end subroutine Set_Time_in_DateTime
    !  #]
  subroutine write_DateTime_to_unbuffbinfile(fileunit,DateTime,error_flag)
    !  #[
    integer,            intent(in)  :: fileunit
    TYPE(DateTimeType), intent(in)  :: DateTime
    integer,            intent(out) :: error_flag

    error_flag = no_error

    call write_IntAl_to_bin_file(fileunit,  DateTime%daycount,     error_flag)
    IF (error_flag .ne. no_error) return
    call write_IntAul_to_bin_file(fileunit, DateTime%seconds,      error_flag)
    IF (error_flag .ne. no_error) return
    call write_IntAul_to_bin_file(fileunit, DateTime%microseconds, error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine write_DateTime_to_unbuffbinfile
    !  #]
  subroutine write_DateTime_to_buffbin_file(BufferedFile,DateTime,error_flag)
    !  #[
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    TYPE(DateTimeType),      intent(in)  :: DateTime
    integer,                 intent(out) :: error_flag

    error_flag = no_error

    call write_IntAl_to_bin_file(BufferedFile, DateTime%daycount,&
                                 error_flag)
    IF (error_flag .ne. no_error) return
    call write_IntAul_to_bin_file(BufferedFile, DateTime%seconds,&
                                  error_flag)
    IF (error_flag .ne. no_error) return
    call write_IntAul_to_bin_file(BufferedFile, DateTime%microseconds,&
                                  error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine write_DateTime_to_buffbin_file
    !  #]
  subroutine write_DateTime_to_genbin_file(fileunit,DateTime,error_flag)
    !  #[
    TYPE(DirectBinaryIO_fileunit_type), intent(in)  :: fileunit
    TYPE(DateTimeType), intent(in)  :: DateTime
    integer,            intent(out) :: error_flag

    error_flag = no_error

    call write_IntAl_to_bin_file(fileunit,  DateTime%daycount,     error_flag)
    IF (error_flag .ne. no_error) return
    call write_IntAul_to_bin_file(fileunit, DateTime%seconds,      error_flag)
    IF (error_flag .ne. no_error) return
    call write_IntAul_to_bin_file(fileunit, DateTime%microseconds, error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine write_DateTime_to_genbin_file
    !  #]
  subroutine read_DateTime_from_bin_file(fileunit,DateTime,error_flag)
    !  #[
    integer,            intent(in)  :: fileunit
    TYPE(DateTimeType), intent(out) :: DateTime
    integer,            intent(out) :: error_flag

    error_flag = no_error

    call read_IntAl_from_bin_file(fileunit,  DateTime%daycount,     error_flag)
    IF (error_flag .ne. no_error) return
    call read_IntAul_from_bin_file(fileunit, DateTime%seconds,      error_flag)
    IF (error_flag .ne. no_error) return
    call read_IntAul_from_bin_file(fileunit, DateTime%microseconds, error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine read_DateTime_from_bin_file
    !  #]
  function Construct_XML_DateTime(DateTime) result(DateTime_XML)
    !  #[
    TYPE(DateTimeType),    intent(in)  :: DateTime
    character(len=maxlen_DateTime_XML) :: DateTime_XML ! result

    ! local variables
    TYPE(date_type)  :: date ! as defined in DateTimeMod
    TYPE(time_type)  :: time ! as defined in DateTimeMod
    character(len=4) :: Year
    character(len=2) :: Month_XML
    character(len=2) :: Day
    character(len=2) :: Hour
    character(len=2) :: Minute
    character(len=9) :: Seconds_XML

    ! example XML DateTime: UTC=2007-10-02T00:00:01.123456
    !                       123456789012345678901234567890
    !
    ! Warning: the current XML implementation uses microsecond presision
    ! but earlier versions of the different processors used second presision!
    !
    ! NOTE that the FH fields that have only seconds precision are
    ! filled using the DateTime2String function

    ! init
    DateTime_XML(:) = ' '
    call init_date(date)
    call init_time(time)

    IF (DateTime%daycount .eq. minval_IntAl) THEN
       ! handler for missing dates
       DateTime_XML = missing_DateTime_XML
       return
    END IF

    date = Get_Date_from_DateTime(DateTime)

    write(Day,      "(i2.2)") date%day
    write(Month_XML,"(i2.2)") date%month
    write(Year,     "(i4.4)") date%year

    time = Get_Time_from_DateTime(DateTime)

    write(Hour,       "(i2.2)") time%hour
    write(Minute,     "(i2.2)") time%minute
    write(Seconds_XML,"(f9.6)") time%second

    ! check if the Seconds_XML has a leading space, and if so, replace it
    ! with a zero
    if (Seconds_XML(1:1) .eq. ' ') Seconds_XML(1:1) = '0'

    write(DateTime_XML,"('UTC=',a4,'-',a2,'-',a2,'T',a2,':',a2,':',a9)") &
          Year, Month_XML, Day, Hour, Minute, Seconds_XML

  end function Construct_XML_DateTime
    !  #]
  function Construct_KVT_DateTime(DateTime) result(DateTime_KVT)
    !  #[
    TYPE(DateTimeType),    intent(in)  :: DateTime
    character(len=maxlen_DateTime_KVT) :: DateTime_KVT ! result

    ! local variables
    TYPE(date_type)  :: date ! as defined in DateTimeMod
    TYPE(time_type)  :: time ! as defined in DateTimeMod
    character(len=4) :: Year
    character(len=3) :: Month_KVT
    character(len=2) :: Day
    character(len=2) :: Hour
    character(len=2) :: Minute
    character(len=9) :: Seconds_KVT
    ! remark: the KVT datetime format stores microseconds as well

    ! example KVT DateTime: 02-OCT-2007 00:00:01.999816
    !                       123456789012345678901234567

    ! init
    DateTime_KVT(:) = ' '
    call init_date(date)
    call init_time(time)

    IF (DateTime%daycount .eq. minval_IntAl) THEN
       ! handler for missing dates
       DateTime_KVT = missing_DateTime_KVT
       return
    END IF

    date = Get_Date_from_DateTime(DateTime)

    write(Day,      "(i2.2)") date%day
    write(Year,     "(i4.4)") date%year

    select case(date%month)
    case( 1); Month_KVT = "JAN"
    case( 2); Month_KVT = "FEB"
    case( 3); Month_KVT = "MAR"
    case( 4); Month_KVT = "APR"
    case( 5); Month_KVT = "MAY"
    case( 6); Month_KVT = "JUN"
    case( 7); Month_KVT = "JUL"
    case( 8); Month_KVT = "AUG"
    case( 9); Month_KVT = "SEP"
    case(10); Month_KVT = "OCT"
    case(11); Month_KVT = "NOV"
    case(12); Month_KVT = "DEC"
    case default
       ! this should never happen !!!
       print *,"WARNING: illegal month number found inside:"
       print *,"==>Construct_KVT_DateTime "
       print *,"input for this subroutine was: "
       call PrintDateTime(DateTime)
       print *,"setting string to missing"
       DateTime_KVT = "[missing]"
       return       
    end select

    time = Get_Time_from_DateTime(DateTime)

    write(Hour,       "(i2.2)") time%hour
    write(Minute,     "(i2.2)") time%minute
    write(Seconds_KVT,"(f9.6)") time%second

    ! check if the Seconds_KVT has a leading space, and if so, replace it
    ! with a zero
    if (Seconds_KVT(1:1) .eq. ' ') Seconds_KVT(1:1) = '0'

    write(DateTime_KVT,"(a2,'-',a3,'-',a4,1x,a2,':',a2,':',a9)") &
         Day, Month_KVT, Year, Hour, Minute, Seconds_KVT

  end function Construct_KVT_DateTime
    !  #]
  function Convert_KVT_to_Num_DateTime(DateTime_KVT) result(DateTime)
    !  #[

    character(len=maxlen_DateTime_KVT), intent(in) :: DateTime_KVT
    TYPE(DateTimeType):: DateTime ! result

    ! local variables
    integer          :: ios
    TYPE(date_type)  :: date ! as defined in DateTimeMod
    TYPE(time_type)  :: time ! as defined in DateTimeMod
    logical          :: CorrectFormat
    character(len=4) :: Year
    character(len=3) :: Month_KVT
    character(len=2) :: Day
    character(len=2) :: Hour
    character(len=2) :: Minute
    character(len=9) :: Seconds_KVT
    ! remark: the KVT datetime format stores microseconds as well

    ! example KVT DateTime: 02-OCT-2007 00:00:01.999816
    !                       123456789012345678901234567

    ! init
    call InitDateTime(DateTime)
    call init_date(date)
    call init_time(time)

    ! return when datetime is missing
    IF (trim(DateTime_KVT) .eq. trim(missing_DateTime_KVT)) return

    ! allow empty DateTime fields (which need no conversion!)
    ! this will happen a lot during development of the L2BP
    !IF (len_trim(DateTime_KVT) .eq. 0) return

    Day         = DateTime_KVT(1:2)
    Month_KVT   = DateTime_KVT(4:6)
    Year        = DateTime_KVT(8:11)
    Hour        = DateTime_KVT(13:14)
    Minute      = DateTime_KVT(16:17)
    Seconds_KVT = DateTime_KVT(19:27)

    CorrectFormat = .true.
    ! check '-', ' ', '.' and ':'
    IF (DateTime_KVT(3:3)   .ne. '-'   ) CorrectFormat = .false.
    IF (DateTime_KVT(7:7)   .ne. '-'   ) CorrectFormat = .false.
    IF (DateTime_KVT(12:12) .ne. ' '   ) CorrectFormat = .false.
    IF (DateTime_KVT(15:15) .ne. ':'   ) CorrectFormat = .false.
    IF (DateTime_KVT(18:18) .ne. ':'   ) CorrectFormat = .false.
    IF (DateTime_KVT(21:21) .ne. '.'   ) CorrectFormat = .false.
    
    ! check if the numbers are really numbers
    IF (verify(Year,       "0123456789") .ne. 0) CorrectFormat = .false.
    IF (verify(Day,        "0123456789") .ne. 0) CorrectFormat = .false.
    IF (verify(Hour,       "0123456789") .ne. 0) CorrectFormat = .false.
    IF (verify(Minute,     "0123456789") .ne. 0) CorrectFormat = .false.
    IF (verify(Seconds_KVT,"0123456789.") .ne. 0) CorrectFormat = .false.

    ! convert the strings to numbers
    read(Day, "(i2)",iostat=ios) date%day
    IF (ios .ne. 0) goto 999
    read(Year,"(i4)",iostat=ios)date%year
    IF (ios .ne. 0) goto 999

    select case(Month_KVT)
    case("JAN"); date%month = 1
    case("FEB"); date%month = 2
    case("MAR"); date%month = 3
    case("APR"); date%month = 4
    case("MAY"); date%month = 5
    case("JUN"); date%month = 6
    case("JUL"); date%month = 7
    case("AUG"); date%month = 8
    case("SEP"); date%month = 9
    case("OCT"); date%month = 10
    case("NOV"); date%month = 11
    case("DEC"); date%month = 12
    case default
       CorrectFormat = .false.
    end select

    read(Hour,       "(i2)",  iostat=ios) time%hour
    IF (ios .ne. 0) goto 999
    read(Minute,     "(i2)",  iostat=ios) time%minute
    IF (ios .ne. 0) goto 999
    read(Seconds_KVT,"(f9.6)",iostat=ios) time%second
    IF (ios .ne. 0) goto 999

    goto 111

    ! catch errors in the read() conversion statement
999 continue
    CorrectFormat = .false.

111 continue

    IF (.not. CorrectFormat) THEN
       print *,"ERROR in Convert_KVT_to_Num_DateTime()"
       print *,"Could not interpret KVT DateTime string, since it does"
       print *,"not conform to the definition: <DD-MMM-YYYY HH:MM:SS.SSSSSS>"
       print *,"DateTime_KVT = ",trim(DateTime_KVT)
       ! leave the DateTime value missing
       return
    END IF

    call Set_Date_in_DateTime(DateTime,date)
    call Set_Time_in_DateTime(DateTime,time)

  end function Convert_KVT_to_Num_DateTime
    !  #]
  function Convert_XML_to_Num_DateTime(DateTime_XML,WorkaroundMicroSeconds) &
       result(DateTime)
    !  #[
    character(len=maxlen_DateTime_XML), intent(in) :: DateTime_XML
    TYPE(DateTimeType)                             :: DateTime ! result
    logical, optional,                  intent(in) :: WorkaroundMicroSeconds

    ! local variables
    integer          :: ios
    TYPE(date_type)  :: date ! as defined in DateTimeMod
    TYPE(time_type)  :: time ! as defined in DateTimeMod
    logical          :: CorrectFormat
    character(len=4) :: Year
    character(len=2) :: Month_XML
    character(len=2) :: Day
    character(len=2) :: Hour
    character(len=2) :: Minute
    character(len=9) :: Seconds_XML
    logical          :: old_format

    ! remark: the XML datetime format also uses microseconds now !
    ! to allow using the old style without microseconds,
    ! use the WorkaroundMicroSeconds flag

    ! example XML DateTime: UTC=2007-10-02T00:00:01.123456
    !                       123456789012345678901234567890

    ! example old format:   UTC=1970-01-01T00:00:00
    !                       12345678901234567890123

    ! init
    call InitDateTime(DateTime)
    call init_date(date)
    call init_time(time)

    old_format=.false.
    if (present(WorkaroundMicroSeconds)) &
         old_format = WorkaroundMicroSeconds

    ! return when datetime is missing
    ! note that the xml missing indicator currently is just an empty
    ! string, so having the old or the new format does not make
    ! any difference here.
    IF (trim(DateTime_XML) .eq. trim(missing_DateTime_XML)) return

    ! allow empty DateTime fields (which need no conversion!)
    ! this will happen a lot during development of the L2BP
    !IF (len_trim(DateTime_XML) .eq. 0) return

    Year        = DateTime_XML(5:8)
    Month_XML   = DateTime_XML(10:11)
    Day         = DateTime_XML(13:14)
    Hour        = DateTime_XML(16:17)
    Minute      = DateTime_XML(19:20)
    if (old_format) then
       Seconds_XML = DateTime_XML(22:23)//".000000"
    else
       Seconds_XML = DateTime_XML(22:30)
    end if

    CorrectFormat = .true.

    ! check the string length
    if (old_format) then
       ! workaround.
       ! many old input files are still in use (mainly L1B)
       ! which use the convention of having NO microseconds
       ! in the xml fields. For now, silently allow this, and fill
       ! the microseconds with zeros
       if (len_trim(DateTime_XML) .ne. maxlen_DateTime_XML_old) then
          CorrectFormat = .false.
          goto 111
       end if
    else
       if (len_trim(DateTime_XML) .ne. maxlen_DateTime_XML) then
          CorrectFormat = .false.
          goto 111
       end if
    end if

    ! check "UTC=", '-' and 'T' and ':'
    IF (DateTime_XML(1:4)   .ne. "UTC=") CorrectFormat = .false.
    IF (DateTime_XML(9:9)   .ne. '-'   ) CorrectFormat = .false.
    IF (DateTime_XML(12:12) .ne. '-'   ) CorrectFormat = .false.
    IF (DateTime_XML(15:15) .ne. 'T'   ) CorrectFormat = .false.
    IF (DateTime_XML(18:18) .ne. ':'   ) CorrectFormat = .false.
    IF (DateTime_XML(21:21) .ne. ':'   ) CorrectFormat = .false.
    if (.not. old_format) then
       IF (DateTime_XML(24:24) .ne. '.'   ) CorrectFormat = .false.
    end if

    ! check if the numbers are really numbers
    IF (verify(Year,       "0123456789") .ne. 0) CorrectFormat = .false.
    IF (verify(Month_XML,  "0123456789") .ne. 0) CorrectFormat = .false.
    IF (verify(Day,        "0123456789") .ne. 0) CorrectFormat = .false.
    IF (verify(Hour,       "0123456789") .ne. 0) CorrectFormat = .false.
    IF (verify(Minute,     "0123456789") .ne. 0) CorrectFormat = .false.
    IF (verify(Seconds_XML,"0123456789.") .ne. 0) CorrectFormat = .false.

    read(Day,        "(i2)",iostat=ios) date%day
    IF (ios .ne. 0) goto 999
    read(Month_XML,  "(i2)",iostat=ios) date%month
    IF (ios .ne. 0) goto 999
    read(Year,       "(i4)",iostat=ios) date%year
    IF (ios .ne. 0) goto 999
    read(Hour,       "(i2)",iostat=ios) time%hour
    IF (ios .ne. 0) goto 999
    read(Minute,     "(i2)",iostat=ios) time%minute
    IF (ios .ne. 0) goto 999
    read(Seconds_XML,"(f9.6)",iostat=ios) time%second
    IF (ios .ne. 0) goto 999

    goto 111

    ! catch errors in the read() conversion statement
999 continue
    CorrectFormat = .false.

111 continue


    IF (.not. CorrectFormat) THEN
       print *,"ERROR in Convert_XML_to_Num_DateTime()"
       print *,"Could not interpret XML DateTime string, since it does"
       if (old_format) then
          print *,"not conform to the definition: <UTC=YYYY-MM-DDTHH:MM:SS>"
       else
          print *,"not conform to the definition: "//&
               "<UTC=YYYY-MM-DDTHH:MM:SS.UUUUUU>"
       end if
       print *,"DateTime_XML = ",trim(DateTime_XML)
       ! leave the DateTime value missing
       return
    END IF

    call Set_Date_in_DateTime(DateTime,date)
    call Set_Time_in_DateTime(DateTime,time)

  end function Convert_XML_to_Num_DateTime
    !  #]
  function Convert_JobOrd_to_Num_DateTime(DateTime_JobOrder) result(DateTime)
    !  #[
    character(len=maxlen_DateTime_JobOrder), intent(in) :: DateTime_JobOrder
    TYPE(DateTimeType):: DateTime ! result

    ! local variables
    integer          :: ios
    TYPE(date_type)  :: date ! as defined in DateTimeMod
    TYPE(time_type)  :: time ! as defined in DateTimeMod
    logical          :: CorrectFormat
    integer          :: error_flag
    character(len=4) :: Year
    character(len=2) :: Month
    character(len=2) :: Day
    character(len=2) :: Hour
    character(len=2) :: Minute
    character(len=8) :: Seconds
    integer(i8_)     :: microseconds
    ! remark: the KVT datetime format stores microseconds as well

    ! example JobOrder DateTime: <Start>20060907_221155059000</Start>
    !                                   123456789012345678901 = 21 chars
    ! format                           "YYYYMMDD_HHMMSSSSSSSS"

    ! init
    call InitDateTime(DateTime)
    call init_date(date)
    call init_time(time)

    Year    = DateTime_JobOrder(1:4)
    Month   = DateTime_JobOrder(5:6)
    Day     = DateTime_JobOrder(7:8)
    Hour    = DateTime_JobOrder(10:11)
    Minute  = DateTime_JobOrder(12:13)
    Seconds = DateTime_JobOrder(14:21) ! actually these are microseconds

    CorrectFormat = .true.
    ! check '-', ' ', '.' and ':'
    IF (DateTime_JobOrder(9:9)   .ne. '_'   ) CorrectFormat = .false.
    
    ! check if the numbers are really numbers
    IF (verify(Year,   "0123456789") .ne. 0) CorrectFormat = .false.
    IF (verify(Month,  "0123456789") .ne. 0) CorrectFormat = .false.
    IF (verify(Day,    "0123456789") .ne. 0) CorrectFormat = .false.
    IF (verify(Hour,   "0123456789") .ne. 0) CorrectFormat = .false.
    IF (verify(Minute, "0123456789") .ne. 0) CorrectFormat = .false.
    IF (verify(Seconds,"0123456789") .ne. 0) CorrectFormat = .false.

    ! convert the strings to numbers
    read(Year, "(i4)",iostat=ios)date%year
    IF (ios .ne. 0) goto 999
    read(Month,"(i2)",iostat=ios)date%month
    IF (ios .ne. 0) goto 999
    read(Day,  "(i2)",iostat=ios) date%day
    IF (ios .ne. 0) goto 999

    read(Hour,   "(i2)",  iostat=ios) time%hour
    IF (ios .ne. 0) goto 999
    read(Minute, "(i2)",  iostat=ios) time%minute
    IF (ios .ne. 0) goto 999
    read(Seconds,"(i8)",iostat=ios)   microseconds
    IF (ios .ne. 0) goto 999

    goto 111

    ! catch errors in the read() conversion statement
999 continue
    CorrectFormat = .false.
    error_flag = error_reading_file

111 continue

    time%second = 1.e-6_r8_ * microseconds

    call test_date(date,error_flag)
    IF (error_flag .ne. no_error) CorrectFormat = .false.

    call test_time(time,error_flag)
    IF (error_flag .ne. no_error) CorrectFormat = .false.

    IF (.not. CorrectFormat) THEN
       call logmsg(log_error,"ERROR in Convert_JobOrd_to_Num_DateTime(): "//&
            "Could not interpret JobOrder DateTime string, since it does "//&
            "not conform to the definition: YYYYMMDD_HHMMSSSSSSSS "//&
            "DateTime_JobOrder = "//trim(DateTime_JobOrder),&
            SignalFatalError = .true.,FatalError=error_flag)
       ! leave the DateTime value missing
       return
    END IF

    call Set_Date_in_DateTime(DateTime,date)
    call Set_Time_in_DateTime(DateTime,time)

  end function Convert_JobOrd_to_Num_DateTime
    !  #]
  function Convert_KVT_to_XML_DateTime(DateTime_KVT) result(DateTime_XML)
    !  #[
    character(len=maxlen_DateTime_KVT), intent(in) :: DateTime_KVT ! len = 27
    character(len=maxlen_DateTime_XML) :: DateTime_XML ! result    ! len = 30 

    ! local variable
    TYPE(DateTimeType)  :: DateTime

    DateTime     = Convert_KVT_to_Num_DateTime(DateTime_KVT)
    DateTime_XML = Construct_XML_DateTime(DateTime)

  end function Convert_KVT_to_XML_DateTime
    !  #]
  function Convert_XML_to_KVT_DateTime(DateTime_XML,WorkaroundMicroSeconds) &
       result(DateTime_KVT)
    !  #[
    character(len=maxlen_DateTime_XML), &
                            intent(in) :: DateTime_XML ! len = 30 (or 23)
    character(len=maxlen_DateTime_KVT) :: DateTime_KVT ! result    ! len = 27
    logical, optional,      intent(in) :: WorkaroundMicroSeconds

    ! local variable
    TYPE(DateTimeType)  :: DateTime

    DateTime     = Convert_XML_to_Num_DateTime(DateTime_XML, &
                                               WorkaroundMicroSeconds)
    DateTime_KVT = Construct_KVT_DateTime(DateTime)

  end function Convert_XML_to_KVT_DateTime
    !  #]
  subroutine Fillblanks(chain)
    !  #[
    ! This subroutine fills blanks with zeroes
    ! P. Poli 24/11/2005.
    character (len=*) :: chain
    
    integer :: ipos
    
    do ipos=1,len_trim(chain)
       if (chain(ipos:ipos) .eq. ' ') chain(ipos:ipos)='0'
    enddo
  end subroutine Fillblanks
    !  #]
  !---------------------------------------
  function int1_2txt(r,int_type) result(txt)
    !  #[
    !     Poli   21-02-2006
    ! convert an integer to text, taking into account that
    ! the input integer might have the missing value

    integer(i1_),      intent(in)  :: r
    character(len=*),  optional, intent(in) :: int_type
    character(len=13)              :: txt ! result

    ! local variable
    character(len=10) :: formatstring
    logical           :: missing_value
    
    formatstring = "(i13)"

    missing_value=.false.
    missing_value=r.eq.missing_indicator_IntAc
    if (present(int_type)) then
       print *,'ERROR: int1_2txt received an optional argument ',int_type
    endif
    
    if (missing_value) then
       txt = "[missing]"
    else
       write(txt,formatstring) r
    endif

  end function int1_2txt
    !  #]
  function int2_2txt(r,int_type) result(txt)
    !  #[
    !     Poli   21-02-2006
    ! convert an integer to text, taking into account that
    ! the input integer might have the missing value

    integer(i2_),      intent(in)  :: r
    character(len=*),  optional, intent(in) :: int_type
    character(len=13)              :: txt ! result

    ! local variable
    character(len=10) :: formatstring
    logical           :: missing_value
    
    formatstring = "(i13)"

    missing_value=.false.
    missing_value=r.eq.missing_indicator_IntAs
    if (present(int_type)) then
       if (trim(int_type) .eq. 'IntAuc') &
          missing_value=r.eq.missing_indicator_IntAuc
    endif

    if (missing_value) then
       txt = "[missing]"
    else
       write(txt,formatstring) r
    endif

  end function int2_2txt
    !  #]
  function int4_2txt(r,int_type) result(txt)
    !  #[
    !     Poli   21-02-2006
    ! convert an integer to text, taking into account that
    ! the input integer might have the missing value

    integer(i4_),      intent(in)  :: r
    character(len=*),  optional, intent(in) :: int_type
    character(len=13)              :: txt ! result

    ! local variable
    character(len=10) :: formatstring
    logical           :: missing_value
    
    formatstring = "(i13)"

    missing_value=.false.
    missing_value=r.eq.missing_indicator_IntAl
    if (present(int_type)) then
       if (trim(int_type) .eq. 'IntAus') &
          missing_value=r.eq.missing_indicator_IntAus
    endif

    if (missing_value) then
       txt = "[missing]"
    else
       write(txt,formatstring) r
    endif

  end function int4_2txt
    !  #]
  function int8_2txt(r,int_type) result(txt)
    !  #[
    !     Poli   21-02-2006
    ! convert an integer to text, taking into account that
    ! the input integer might have the missing value

    integer(i8_),      intent(in)  :: r
    character(len=*),  optional, intent(in) :: int_type
    character(len=13)              :: txt ! result

    ! local variable
    character(len=10) :: formatstring
    logical           :: missing_value
    
    formatstring = "(i13)"

    missing_value=.false.
    missing_value=r.eq.missing_indicator_IntAd
    if (present(int_type)) then
       if (trim(int_type) .eq. 'IntAul') &
          missing_value=r.eq.missing_indicator_IntAul
    endif

    if (missing_value) then
       txt = "[missing]"
    else
       write(txt,formatstring) r
    endif

  end function int8_2txt
    !  #]
  !---------------------------------------
end module EE_CFI_Datatypes
