module FixedHeader
  !  #[ Documentation

  !
  ! this module defines the Fixed Header for the ee_cfi fileformat
  !
  ! the next type definitions are generic for aeolus files
  ! and are used for both the L1B and L2B file headers
  !
  ! for a description, see section 3.3.1 of the L1B_iodd_2.1
  ! or TN2.1, version 1.2, (6-9-2005), sec.4.3, p.13, tables 6,7,8
  !
  ! Written by J. de Kloe (KNMI)
  !
  ! Modifications:
  !   15-Sep-2005 D. Tan     Upgrade to DLR TDS v1.2 (Aug 2005)
  !   28-Oct-2005 J. de Kloe moved check_length_xml_item to xml_module
  !   02-Dec-2005 P. Poli    Aux filename uses 'AE' rather than mission_name.
  !   06-Jan-2006 J. de Kloe updated DateTime module: timestep is now r8_ type
  !   26-Jan-2006 P. Poli    Upgrade to L1B IODD 06-09-2005, in line 
  !                          with TN2.1 v1.3
  !   14-Feb-2006 D. Tan     New file_type_AMD = AUX_MET_EC.
  !   28-Feb-2006 J. de Kloe added prefix parameter to print routine
  !   02-Mar-2006 J. de Kloe adapted to changes in XML_module
  !   13-Mar-2006 J. de Kloe added the AsciiWrite_FH routine
  !   26-Apr-2006 D. Tan     New file_type_AMD = AUX_MET_12.
  !   12-May-2006 D. Tan     Null use of ParKind1 module
  !   25-Oct-2006 D. Tan     Add file_type_2C_[N,R,T].
  !   03-Nov-2006 P. Poli    upgrade to L1BP v1.04
  !   05-Mar-2007 J. de Kloe updated check on value of file_version
  !   04-Apr-2007 J. de Kloe disallow: file_type_L2B_R/file_type_L2B_T/
  !                          file_type_L2C_R/file_type_L2C_T
  !                          changed file_type_RBC from AUX_RBC_AX to 
  !                          AUX_RBC_L2; changed file_type_AuxClim from
  !                          AUX_CLM_AX to AUX_CLM_L2, and 
  !                          introduced file_type_AuxCal
  !   17-Apr-2007 D. Tan     Remove all refs to AUX_ECF_AX and AUX_MET_EC.
  !   06-Jun-2007 J. de Kloe added path_in parameter to the subroutines
  !                          Get_FH_from_xml_tree() and Copy_FH_to_xml_tree()
  !   30-Jul-2007 J. de Kloe changed the interface to Fill_FH_Items
  !   05-Nov-2007 J. de Kloe renamed file_type_RBC to file_type_AuxRBC
  !   13-Nov-2007 J. de Kloe added file_type_L1B_N and file_type_AuxPar
  !   13-Nov-2007 J. de Kloe adapted to round to floor/ceil when filling
  !                          FH%Validity_Period and FH%Source%Creation_Date
  !   06-Dec-2007 J. de Kloe adapted to new allowed values for systemstr* 
  !                          see SPR 28 for details, and changed interface
  !                          for Fill_FH_items
  !   09-Jan-2008 J. de Kloe added check for missing_DateTime_KVT
  !                          in AsciiWrite_FH routine
  !   16-Jan-2008 J. de Kloe phase out integer kind i_ 
  !
  !  #] 
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE XML_module, only: max_str_len, xml_node_type, xml_node_get_string, &
                        xml_set_fixed_header, check_length_xml_item, &
                        ee_fh_name

  ! support modules
  USE ErrorHandler, only: no_error, error_allocate, error_programming, &
                          error_reading_file, error_writing_file
  USE EE_CFI_Datatypes, only: DateTimeType, &
                              InitDateTime, IncrementDateTime, &
                              TestDateTime, missing_DateTime_KVT, &
                              DateTime2String, GetCurrentDateTime
  USE Logging, only: logmsg,log_debug,log_info,log_progress,log_warn,log_error
  USE Numerics, only: r8_
  USE LunManager, only: fileunit_stdout
  USE AeolusConstants, only: software_version
  !  #]
  !  #[ Type definitions and parameters

  implicit none

  public

  ! allowed filetypes

! TODO: add some python script to check these filetypes with the ones in
!       ../ThinLayer/TaskTable.AE_L1B_L2B_WIND.xml
!       to ensure consistency of the software

  ! for Aux. Met. Data files
  character(len=10), parameter :: file_type_AMD     = "AUX_MET_12"

  ! for Nominal L1B input files
  character(len=10), parameter :: file_type_L1B_N   = "ALD_U_N_1B"

  ! for Nominal L2B output files
  character(len=10), parameter :: file_type_L2B_N   = "ALD_U_N_2B"
  ! for Reprocessed L2B output files
! still under consideration (but makes the ThinLayer more complicated ...)
!  character(len=10), parameter :: file_type_L2B_R   = "ALD_U_R_2B"
  ! for Test L2B output files
! not needed anymore (Test files use cccc=TEST, i.e. AE_TEST...)
!  character(len=10), parameter :: file_type_L2B_T   = "ALD_U_T_2B"

  ! for Nominal L2C output files
  character(len=10), parameter :: file_type_L2C_N   = "ALD_U_N_2C"
  ! for Reprocessed L2C output files
! not allowed anymore (makes the ThinLayer too complicated ...)
!  character(len=10), parameter :: file_type_L2C_R   = "ALD_U_R_2C"
  ! for Test L2C output files
! not allowed anymore (makes the ThinLayer too complicated ...)
!  character(len=10), parameter :: file_type_L2C_T   = "ALD_U_T_2C"

  ! for the auxiliary files
  character(len=10), parameter :: file_type_AuxPar  = "AUX_PAR_2B"
  character(len=10), parameter :: file_type_AuxRBC  = "AUX_RBC_L2"
  character(len=10), parameter :: file_type_AuxClim = "AUX_CLM_L2"
  character(len=10), parameter :: file_type_AuxCal  = "AUX_CAL_L2"

  ! max length for systemstr is 12 characters
  ! changed to values mentioned in SPR 28
  character(len=6), parameter :: systemstr_ECMWF_IFS      = "L2PF"
  character(len=6), parameter :: systemstr_ESRIN_REPR     = "LTA"
  character(len=6), parameter :: systemstr_LOCAL_MET_INST = "LOCAL"
  character(len=6), parameter :: systemstr_ACMF           = "ACMF"

  ! L2PF  if file was generated at ECMWF
  ! LTA   if file was generated after reprocessing at ESRIN
  ! ACMF  if file was generated at ACMF
  ! LOCAL if file was generated at any other Met. Institute, for example KNMI

  ! NOTE!!!
  ! The following maxlen parameters are NOT restrictive! Even if they
  ! are mentioned in the IODD, we may alter them, since FH is only written 
  ! to xml files, there is no limit to their lengths, the value in the IODD
  ! is just an indication of what the length might be. 
  ! However, since we need to define a length inside our
  ! fortran code, we need some numbers anyway.
  !
  ! EXCEPTIONS are the fields used for constructing the filename
  ! like file_class, file_type and the datetime fields
  
  ! file_name_length is expected to be either 55 or 56 characters
  ! (see 2 possible definitions in function Create_EE_CFI_Filename() )
  ! max. allowed length is 56 characters
  ! Poli: Upgrade to L1B IODD 06-09-2005, in line with TN2.1 v1.3
  integer, parameter :: expected_len_file_name  = 61
  ! set to 59 for now, to allow the L1B testfiles from MDA
  ! Poli: Upgrade to L1B IODD 06-09-2005, in line with TN2.1 v1.3
  integer, parameter :: maxlen_file_name        = 62 ! 59 ! 56
  integer, parameter :: maxlen_file_description = 32
  integer, parameter :: maxlen_notes            = 32
  integer, parameter :: maxlen_mission          = 6
  integer, parameter :: maxlen_file_class       = 4
  !                            = 'OPER' or 'TEST', 'REPR' is TBC
  !      (this is according to the example in fig.3, on p.14,
  !       and corresponds to the L1B testfile that we recieved)
  integer, parameter :: maxlen_file_type        = 10
  !                            = 'ALD_U_X_2B' or 'AUX_MET_12'
  !
  ! REMARK: in TN2.1, the descriptions of file_class and file_type
  !         in table 6 seem to be mixed up
  !
  integer, parameter :: maxlen_file_version     = 4 
  !                           (version is stored as integer)
  integer, parameter :: default_file_version    = 1 
  ! note that only the database system behind the ThinLayer knows what 
  ! version have been created of which file, and what version number
  ! the file should have. The processor does not have this information.
  ! Therefore this version is always set to 1, and the ThinLayer is expected
  ! to update the filename, and the File_Version element in the FixedHeader
  ! if needed.
  !
  ! (see TN2.1, sectie 4.3, table 6.)
  integer, parameter :: maxlen_validity_start   = 23
  integer, parameter :: maxlen_validity_stop    = 23
    ! (see TN2.1, sectie 4.3, table 7.)
! WARNING: according to TN2.1, section 4.3, p.13, table 6
! max. length of the system string should be 12 characters
! However, the first L1B testfile that we recieved had a
! systemstring of 19 characters: "AE_RAW_L1B_WIND_CAL"
! so this parameter has been adapted to allow reading this file
!  integer, parameter :: maxlen_system           = 12
  ! Poli: Upgrade to L1B IODD 06-09-2005, in line with TN2.1 v1.3
  integer, parameter :: maxlen_system           = 19
! changed from 12 to 40 after discussing with David and Christoph
! J. de Kloe, 06-Dec-2007.
  integer, parameter :: maxlen_creator          = 40
  integer, parameter :: maxlen_creator_version  = 12 
  integer, parameter :: maxlen_creation_date    = 23

  TYPE FH_Validity_Period_type
     !  #[
     character(len=maxlen_validity_start) :: Validity_Start
     character(len=maxlen_validity_stop)  :: Validity_Stop
  END TYPE FH_Validity_Period_type
     !  #]

  TYPE FH_Source_type
     !  #[
     character(len=maxlen_system)          :: System
     character(len=maxlen_creator)         :: Creator
     character(len=maxlen_creator_version) :: Creator_Version
     character(len=maxlen_creation_date)   :: Creation_Date
  END TYPE FH_Source_type
     !  #]

  TYPE FH_type
     !  #[ 
     character(len=maxlen_file_name)        :: File_name
     character(len=maxlen_file_description) :: File_Description
     character(len=maxlen_notes)            :: Notes
     character(len=maxlen_mission)          :: Mission
     character(len=maxlen_file_class)       :: File_Class
     character(len=maxlen_file_type)        :: File_Type
     TYPE(FH_Validity_Period_type)          :: Validity_Period
     ! this version has type integer here, and not string, to match
     ! the datatypes used in the ee_cfi library
     integer                                :: File_Version
     TYPE(FH_Source_type)                   :: Source
  END TYPE FH_type
    !  #]

  ! Fixed Header xml definition:
  ! => see TN2.1, version 1.2, (6-9-2005), sec.4.3, p.13, tables 6,7,8
  ! MODIFIED : Poli, TN2.1 v1.3
  ! ----> File_Name        : tag(11)+string(60)+endtag(13)
  ! ----> File_Description : tag(17)+string(32)+endtag(20)
  ! ----> Notes            : tag(7) +multi-line-string(32)+endtag(9)
  ! ----> Mission          : tag(9) +string(6) +endtag(11)
  ! ----> File_Class       : tag(12)+string(10)+endtag(14)
  ! ----> File_Type        : tag(11)+string(4) +endtag(13)
  ! ----> Validity_Period  : structure 18+152+19 bytes (see table 6)
  !       ----> Validity_Start : tag(26)+DateTimeType(23)+endtag(28)
  !       ----> Validity_Stop  : tag(25)+DateTimeType(23)+endtag(27)
  ! ----> File_Version     : tag(14)+string(4)+endtag(16)
  ! ----> Source                : structure 9+173+10 (see table 7)
  !       ----> System          : tag(8) +string(6) +endtag(10)
  !       ----> Creator         : tag(9) +string(4) +endtag(11)
  !       ----> Creator_Version : tag(17)+string(12)+endtag(19)
  !       ----> Creation_Date   : tag(26)+DateTimeType(23)+endtag(28)

  !  #]
  !  #[ interfaces for overloading functions
  INTERFACE Print_FH
     module procedure Print_FH_new
     !module procedure Print_FH_old, Print_FH_new
  END INTERFACE
  !  #]
  !  #[ Module interface

  ! all is public except for:
  private :: Create_EE_CFI_Filename

  !  #]
contains
  !---------------------------------
  subroutine Init_FH(FH,error_flag)
    !  #[
    TYPE(FH_type), pointer     :: FH
    integer,       intent(out) :: error_flag
    
    error_flag = no_error
    IF (.not. associated(FH)) THEN
       print *,"ERROR in Init_FH():"
       print *,"The pointer to this FH seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! see the remark in MainProductHeader.F90 for the reason why I use
    ! the form x(:)=' ' to initialise the strings

    FH%File_Name(:)        = ' '
    FH%File_Description(:) = ' '
    FH%Notes(:)            = ' '
    FH%Mission(:)          = ' '
    FH%File_Class(:)       = ' '
    FH%File_Type(:)        = ' '
    FH%Validity_Period%Validity_Start(:) = ' '
    FH%Validity_Period%Validity_Stop(:)  = ' '
    FH%File_Version                      = 1
    FH%Source%System(:)                  = ' '
    FH%Source%Creator(:)                 = ' '
    FH%Source%Creator_Version(:)         = ' '
    FH%Source%Creation_Date(:)           = ' '

  end subroutine Init_FH
    !  #]
  subroutine Print_FH_old(FH,prefix)
    !  #[
    TYPE(FH_type),              pointer    :: FH
    character(len=*), optional, intent(in) :: prefix

    ! local variable
    character(len=256) :: tmp_prefix
    integer            :: error_flag

    tmp_prefix = "..."
    IF (present(prefix)) tmp_prefix = prefix

    call AsciiWrite_FH(FH,fileunit_stdout,tmp_prefix,error_flag)

  end subroutine Print_FH_old
    !  #]
  subroutine Print_FH_new(FH,error_flag,prefix)
    !  #[
    TYPE(FH_type),              pointer     :: FH
    integer,                    intent(out) :: error_flag
    character(len=*), optional, intent(in)  :: prefix

    ! local variable
    character(len=256) :: tmp_prefix

    error_flag = no_error

    tmp_prefix = "..."
    IF (present(prefix)) tmp_prefix = prefix

    call AsciiWrite_FH(FH,fileunit_stdout,tmp_prefix,error_flag)

  end subroutine Print_FH_new
    !  #]
  subroutine AsciiWrite_FH(FH,fileunit,prefix,error_flag)
    !  #[

    TYPE(FH_type),    pointer     :: FH
    integer,          intent(in)  :: fileunit
    character(len=*), intent(in)  :: prefix
    integer,          intent(out) :: error_flag

    ! local variable
    character(len=256) :: txt

    error_flag = no_error

    IF (.not. associated(FH)) THEN
       write(fileunit,"(a)",err=999) trim(prefix)//" [seems not associated]"
       return
    END IF

    txt = FH%File_Name
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%File_Name                      = "//trim(txt)

    txt = FH%File_Description
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%File_Description               = "//trim(txt)

    txt = FH%Notes
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Notes                          = "//trim(txt)

    txt = FH%Mission
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Mission                        = "//trim(txt)

    txt = FH%File_Class
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%File_Class                     = "//trim(txt)

    txt = FH%File_Type
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%File_Type                      = "//trim(txt)

    txt = FH%Validity_Period%Validity_Start
    IF (trim(txt) .eq. trim(missing_DateTime_KVT)) THEN
       txt = "[missing]"
    ELSE 
       IF (len_trim(txt).eq.0) txt = "[empty]"
    END IF
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Validity_Period%Validity_Start = "//trim(txt)

    txt = FH%Validity_Period%Validity_Stop
    IF (trim(txt) .eq. trim(missing_DateTime_KVT)) THEN
       txt = "[missing]"
    ELSE 
       IF (len_trim(txt).eq.0) txt = "[empty]"
    END IF
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Validity_Period%Validity_Stop  = "//trim(txt)

    IF (FH%File_Version .ge. 0) THEN 
       write(fileunit,"(2a,i4.4)",err=999) &
            trim(prefix),"%File_Version                   = ",FH%File_Version
    ELSE
       write(fileunit,"(2a,i4.3)",err=999) &
            trim(prefix),"%File_Version                   = ",FH%File_Version
    END IF

    txt = FH%Source%System
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Source%System                  = "//trim(txt)

    txt = FH%Source%Creator
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Source%Creator                 = "//trim(txt)

    txt = FH%Source%Creator_Version
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Source%Creator_Version         = "//trim(txt)

    txt = FH%Source%Creation_Date
    IF (trim(txt) .eq. trim(missing_DateTime_KVT)) THEN
       txt = "[missing]"
    ELSE 
       IF (len_trim(txt).eq.0) txt = "[empty]"
    END IF
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Source%Creation_Date           = "//trim(txt)

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_FH "//&
                          "error while writing to outputfile")
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_FH

    !  #]
  subroutine Get_FH_from_xml_tree(xml_tree, path_in, FH, error_flag)
    !  #[
    TYPE(xml_node_type), pointer     :: xml_tree   ! input
    character(len=*),    intent(in)  :: path_in    ! input
    TYPE(FH_type),       pointer     :: FH         ! output
    integer,             intent(out) :: error_flag ! output

    ! local variables
    integer                     :: ios
    character(len=max_str_len)  :: File_Version_String
    character(len=max_str_len)  :: path
    character(len=max_str_len)  :: text
    character(len=max_str_len)  :: path_FH

    ! init
    error_flag = no_error
    path_FH = trim(path_in)//"/"//ee_fh_name

    IF (.not. associated(FH)) THEN
       print *,"ERROR in Get_FH_from_xml_tree():"
       print *,"the pointer FH seems not associated !!!"
       error_flag = error_programming
       return
    END IF

    IF (.not. associated(xml_tree)) THEN
       print *,"ERROR in Get_FH_from_xml_tree():"
       print *,"the pointer xml_tree seems not associated !!!"
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! copying the FH XML elements
    !------------------------------

    path  =trim(path_FH)//"/File_Name"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    FH%File_Name = check_length_xml_item(text,path,maxlen_file_name,&
         "Get_FH_from_xml_tree")

    path = trim(path_FH)//"/File_Description"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    FH%File_Description = &
         check_length_xml_item(text,path,maxlen_file_description,&
         "Get_FH_from_xml_tree")

    path = trim(path_FH)//"/Notes"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    FH%Notes = check_length_xml_item(text,path,maxlen_notes,&
         "Get_FH_from_xml_tree")

    path = trim(path_FH)//"/Mission"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    FH%Mission = check_length_xml_item(text,path,maxlen_mission,&
         "Get_FH_from_xml_tree")

    path = trim(path_FH)//"/File_Class"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    FH%File_Class = check_length_xml_item(text,path,maxlen_file_class,&
         "Get_FH_from_xml_tree")

    path = trim(path_FH)//"/File_Type"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    FH%File_Type = check_length_xml_item(text,path,maxlen_file_type,&
         "Get_FH_from_xml_tree")

    path = trim(path_FH)//"/Validity_Period/Validity_Start"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    FH%Validity_Period%Validity_Start = &
         check_length_xml_item(text,path,maxlen_validity_start,&
         "Get_FH_from_xml_tree")

    path = trim(path_FH)//"/Validity_Period/Validity_Stop"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    FH%Validity_Period%Validity_Stop = &
         check_length_xml_item(text,path,maxlen_validity_stop,&
         "Get_FH_from_xml_tree")

    path = trim(path_FH)//"/File_Version"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    File_Version_String = check_length_xml_item(text,path,maxlen_file_version,&
         "Get_FH_from_xml_tree")

    ! convert the string to an integer, before storing it.
    ! (error handling of this conversion is done below, see label 999)
    read(File_Version_String,"(i4)",iostat=ios) FH%File_Version
    IF (ios .ne. 0) goto 999

    path = trim(path_FH)//"/Source/System"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    FH%Source%System = check_length_xml_item(text,path,maxlen_system,&
         "Get_FH_from_xml_tree")

    path = trim(path_FH)//"/Source/Creator"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    FH%Source%Creator = check_length_xml_item(text,path,maxlen_creator,&
         "Get_FH_from_xml_tree")

    path = trim(path_FH)//"/Source/Creator_Version"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    FH%Source%Creator_Version = &
         check_length_xml_item(text,path,maxlen_creator_version,&
         "Get_FH_from_xml_tree")

    path = trim(path_FH)//"/Source/Creation_Date"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    FH%Source%Creation_Date = &
         check_length_xml_item(text,path,maxlen_creation_date,&
         "Get_FH_from_xml_tree")

    return

    ! error handler
999 print *,"Error converting File_Version_String into an integer"
    print *,"unable to fill: FH%File_Version"
    print *,"File_Version_String: ",trim(File_Version_String)
    error_flag = error_reading_file
    return

  end subroutine Get_FH_from_xml_tree
    !  #]
  subroutine Fill_FH_items(FH, &
                           file_description, notes, &
                           file_class, file_type, &
                           ValidityPeriodStart, ValidityPeriodStop,&
                           file_version, system, &
                           creator, creator_version, &
                           error_flag)
    !  #[
    TYPE(FH_type),       pointer    :: FH                  ! filled on output
    ! file_name : is generated from the other fields
    character(len=*),    intent(in) :: file_description    ! text
    character(len=*),    intent(in) :: notes               ! text
    ! mission : is a fixed string "Aeolus"
    character(len=*),    intent(in) :: file_class          ! text
    character(len=*),    intent(in) :: file_type           ! text
    TYPE(DateTimeType),  intent(in) :: ValidityPeriodStart ! struct
    TYPE(DateTimeType),  intent(in) :: ValidityPeriodStop  ! struct
    integer,             intent(in) :: file_version        ! counter
    character(len=*),    intent(in) :: system              ! text
    character(len=*),    intent(in) :: creator             ! text
    character(len=*),    intent(in) :: creator_version     ! text
    ! creation_date : generated using the system clock
    integer,             intent(out):: error_flag          ! result

    ! local variables

    ! workaround for portland compiler bug
    character(len=max_str_len) :: tmp_txt

    ! local variables
    character(len=max_str_len) :: File_Version_String
    TYPE(DateTimeType)         :: datetime_current
    character(len=max_str_len) :: tmp_filename

    error_flag = no_error
    IF (.not. associated(FH)) THEN
       print *,"ERROR in Fill_FH_items():"
       print *,"The pointer to this FH seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! test the validity if the provided datetimes
    call TestDateTime(ValidityPeriodStart,error_flag)
    IF (error_flag .ne. no_error) return
    call TestDateTime(ValidityPeriodStop,error_flag)
    IF (error_flag .ne. no_error) return
    
    IF (len_trim(file_description) .gt. maxlen_file_description) THEN
       call logmsg(log_warn,"WARNING in Fill_FH_items: "//&
                   "the file_description string is longer than allowed. "//&
                   "file_description ["//trim(file_description)//"] will "//&
                   "be truncated to ["//&
                   trim(file_description(1:maxlen_file_description))//"]")
    END IF
    FH%File_Description = file_description

    IF (len_trim(notes) .gt. maxlen_notes) THEN
       call logmsg(log_warn,"WARNING in Fill_FH_items: "//&
                   "the notes string is longer than allowed. "//&
                   "notes ["//trim(notes)//"] will "//&
                   "be truncated to ["//&
                   trim(notes(1:maxlen_notes))//"]")
    END IF
    FH%Notes            = notes

    ! this is always the same for our software (so no checks needed)
    ! ("Aeolus" seems also possible)
    ! Poli: fixed bug -- previous version was using Aeolus
    FH%mission          = "Aeolus"

    IF (len_trim(file_class) .gt. maxlen_file_class) THEN
       call logmsg(log_warn,"WARNING in Fill_FH_items: "//&
                   "the file_class string is longer than allowed. "//&
                   "file_class ["//trim(file_class)//"] will "//&
                   "be truncated to ["//&
                   trim(file_class(1:maxlen_file_class))//"]")
    END IF

    ! disabled this test 6-Dec-2007, JK
    ! since according to SPR 28 the ACMF and LTA wish to have their own choice
    ! in filling this field, which should just be copied from the JobOrder file

    !IF ( (trim(file_class) .ne. "OPER") .and. &
    !     (trim(file_class) .ne. "TEST") .and. & 
    !     (trim(file_class) .ne. "REPR")       ) THEN
    !   print *,"WARNING in Fill_FH_items():"
    !   print *,"file_class is allowed to have the strings"
    !   print *,"OPER and REPR, but this seems yet another class:"
    !   print *,"file_class = [",trim(file_class),"]"
    !   print *,"are you sure this is correct ?"
    !END IF
    FH%file_class       = file_class

    IF (len_trim(file_type) .gt. maxlen_file_type) THEN
       call logmsg(log_warn,"WARNING in Fill_FH_items: "//&
                   "the file_type string is longer than allowed. "//&
                   "file_type ["//trim(file_type)//"] will "//&
                   "be truncated to ["//&
                   trim(file_type(1:maxlen_file_type))//"]")
    END IF
    FH%File_Type        = file_type

    ! The portland compiler seems not able to take a function result
    ! directly and write it to an internal file (this is a BUG in this
    ! compiler!) so an intermediate step is needed ....

    tmp_txt = DateTime2String(ValidityPeriodStart,sep_date='-',sep_time=':',&
                              round_floor=.true.)
    write(FH%Validity_Period%Validity_Start,"('UTC=',a19)") tmp_txt

    tmp_txt = DateTime2String(ValidityPeriodStop,sep_date='-',sep_time=':',&
                              round_ceil=.true.)
    write(FH%Validity_Period%Validity_Stop,"('UTC=',a19)") tmp_txt
    
    IF (file_version .ne. 1) THEN
       write(File_Version_String,"(i4.4)",err=999) file_version
       call logmsg(log_warn,"WARNING in Fill_FH_items(): The user should "//&
            "always use file_version=1. "//&
            "Only the ThinLayer database system is expected to rename "//&
            "files to have higher version numbers! Are you sure the "//&
            "current value of "//File_Version_String//" is correct?")
    END IF
    FH%File_Version    = file_version

    IF (len_trim(system) .gt. maxlen_system) THEN
       call logmsg(log_warn,"WARNING in Fill_FH_items: "//&
                   "the system string is longer than allowed. "//&
                   "system ["//trim(system)//"] will "//&
                   "be truncated to ["//&
                   trim(system(1:maxlen_system))//"]")
    END IF
    FH%Source%System = system

    IF (len_trim(creator) .gt. maxlen_creator) THEN
       call logmsg(log_warn,"WARNING in Fill_FH_items: "//&
                   "the creator string is longer than allowed. "//&
                   "creator ["//trim(creator)//"] will "//&
                   "be truncated to ["//&
                   trim(creator(1:maxlen_creator))//"]")
    END IF
    FH%Source%Creator = creator

    IF (len_trim(creator_version) .gt. maxlen_creator_version) THEN
       call logmsg(log_warn,"WARNING in Fill_FH_items: "//&
                   "the creator_version string is longer than allowed. "//&
                   "creator_version ["//trim(creator_version)//"] will "//&
                   "be truncated to ["//&
                   trim(creator_version(1:maxlen_creator_version))//"]")
    END IF
    FH%Source%Creator_Version  = creator_version

    ! get the current date/time when the program is running
    datetime_current = GetCurrentDateTime()

    tmp_txt = DateTime2String(datetime_current,sep_date='-',sep_time=':',&
                              round_floor=.true.)
    write(FH%Source%creation_date,"('UTC=',a19)") tmp_txt

    ! the filename is generated using several other FH fields
    ! therefore this field is filled last
    tmp_filename = &
         Create_EE_CFI_Filename(FH%mission, FH%file_class, FH%file_type, &
                                FH%file_version, ValidityPeriodStart, &
                                ValidityPeriodStop)

    ! extra safety check on the length of the generated filename
    IF (len_trim(tmp_filename) .gt. maxlen_file_name) THEN
       call logmsg(log_warn,"WARNING in Fill_FH_items: "//&
                   "the length of the generated filename is larger than "//&
                   "what is allowed !!! "//&
                   "tmp_filename = ["//trim(tmp_filename)//"] "//&
                   "will be truncated to ["//&
                   trim(tmp_filename(1:maxlen_file_name))//"]")
    END IF
    FH%File_Name = tmp_filename(1:maxlen_file_name)

    return

    ! error handler
999 call logmsg(log_error,"Error in Fill_FH_items(): "//&
         "Error converting integer file_version number into a string "//&
         "Unable to display: FH%File_Version")
    error_flag = error_writing_file
    return

  end subroutine Fill_FH_items

    !  #]
  subroutine Copy_FH_to_xml_tree(FH, path, xml_tree, error_flag)
    !  #[

    TYPE(FH_type),       pointer    :: FH           ! input
    character(len=*),    intent(in) :: path         ! input
    TYPE(xml_node_type), pointer    :: xml_tree     ! points to updated struct
    integer,             intent(out):: error_flag   ! result

    ! local variable
    character(len=max_str_len)  :: path_FH

    ! init
    error_flag = no_error
    path_FH = trim(path)//"/Fixed_Header"

    IF (.not. associated(FH)) THEN
       print *,"ERROR in  Copy_FH_to_xml_tree():"
       print *,"the input FH pointer seems not associated !!!!"
       error_flag = error_programming
       return
    END IF

    ! fill all items inside the FixedHeader
    call xml_set_fixed_header(xml_tree, path_FH, &
         FH%File_Name, &
         FH%File_Description, &
         FH%Notes, &
         FH%Mission, &
         FH%File_Class, &
         FH%File_Type, &
         FH%Validity_Period%validity_start, &
         FH%Validity_Period%validity_stop, &
         FH%File_Version, &
         FH%Source%System, &
         FH%Source%Creator, &
         FH%Source%Creator_Version, &
         FH%Source%Creation_Date, &
         error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine Copy_FH_to_xml_tree

    !  #]
  !---------------------------------
  ! private routine for this module
  function Create_EE_CFI_Filename(mission,file_class,file_type,file_version,&
                    DateTimeStart, DateTimeStop) result(aux_filename)
    !  #[

    character(len=*)   :: mission       ! input
    character(len=*)   :: file_class    ! input
    character(len=*)   :: file_type     ! input
    integer            :: file_version  ! input
    TYPE(DateTimeType) :: DateTimeStart ! input struct(YYYY,MM,DD)
    TYPE(DateTimeType) :: DateTimeStop  ! input struct(YYYY,MM,DD)
    character(len=max_str_len) :: aux_filename ! result

    ! local variables
    character(len=max_str_len) :: file_version_str

    ! used filename definitions:
    ! AMD filename: see L2B/C IODD and ExtICD
    ! AE_OPER_AUX_MET_12_yyyymmddThhmmss_yyyymmddThhmmss_vvvv.EEE
    !
    ! 1234567890123456789012345678901234567890123456789012345
    !          1         2         3         4         5
    !
    ! L2B filename: see TN2.2 (version 1.1, 06-04-2005), p.25, sec.6.1
    ! AE_cccc_ALD_U_X_2B_yyyymmddThhmmss_yyyymmddThhmmss_vvvv.EEE
    !
    ! where cccc='OPER' or 'TEST', and
    ! where X='N' for near-real-time (or nominal) processed data
    !   and X='R' for reprocessed data (TBC)
    !   and X='N' for test data
    ! and yyyymmddThhmmss = start/stop datetime os the sensing sequence
    ! and vvvv=file version number (0001=initial version)
    ! and EEE=header=".HDR" or ".DBL"
    !
    ! However, an alternative definition seems to exist for the
    ! L2B filename: see TN2.1 (version 1.1, 14-02-2005), p.15, sec.5.1
    ! AE_OPER_ALD_C_X_2B_yyyymmddThhmmss_uuuuuuuuu_oooooo_vvvv.EEE
    ! where uuuuuuuuu = the duration of the sensing period in [ms]
    !   and oooooo = the absolute orbit number
    ! [this is what is used for the L1B filename]
    !
    ! I will use the first definition for now.

    file_version_str(:) = ' '
    write(file_version_str,"(i4.4)") file_version
    aux_filename = 'AE'//'_'//&      ! 'AE'
         trim(file_class)//'_'//&   ! 'OPER', 'TEST' or 'REPR'
         trim(file_type)//'_'//&    ! 'ALD_U_X_2B' or 'AUX_MET_12'
         trim(DateTime2String(DateTimeStart,round_floor=.true.))//'_'//&
         trim(DateTime2String(DateTimeStop, round_ceil =.true.))//'_'//&
         trim(file_version_str)

  end function Create_EE_CFI_Filename

    !  #]
  !---------------------------------
end module FixedHeader
