module JobOrder_DataStructure
  !  #[ Description:

  !
  !     a module to define the JobOrder datastructure needed to be able to
  !     load a JobOrder file
  !             
  ! code extracted from the file ../ThinLayer/readjoborderdata.F90
  ! on: 15-Aug-2006 by J. de Kloe, KNMI
  ! 
  ! Modifications:
  ! 31-Aug-2006 J. de Kloe  removed the Breakpoint_Enable element of 
  !                         Proc_Instructions_type since this tag is not
  !                         present in JobOrder files generated by the
  !                         ThinLayer emulator.
  ! 07-Sep-2006 J. de Kloe  implemented workaround for STEM file definitions
  !                         in the JobOrder file.
  ! 08-Sep-2006 J. de Kloe  renamed software_ExecutableName to
  !                         software_TaskName
  ! 13-Oct-2006 D. Tan      Change Workaround from Stem to Physical.DBL
  ! 03-Apr-2007 J. de Kloe  updated the list of allowed order types
  ! 27-Jul-2007 J. de Kloe  activated error reporting when the problem_found
  !                         switch is set in subroutine Check_Proc_Instructions
  ! 26-Oct-2007 J. de Kloe  remove filename extensions, except when the file
  !                         ends with .EEF this removes the need for the 
  !                         workaround flag that we had here.
  ! 13-Nov-2007 J. de Kloe  remove some no longer needed xml-module items
  ! 06-Dec-2007 J. de Kloe  remove an obsolete test on Order_Type and
  !                         reduce size of Order_Type to 4 chars
  ! 16-Jan-2008 J. de Kloe  phase out integer kind i_ 
  !
  !  #]
  !  #[ Modules used:
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE XML_module, only: xml_node_type, max_str_len, &
                        xml_node_get_string, check_length_xml_item, &
                        xml_tree_read_string_attribute

  USE StringTools, only: to_lowercase, Split_filename_and_extension
  USE AeolusConstants, only: software_TaskName, software_version
  USE Logging, only: logmsg,log_debug,log_info,log_progress,log_warn,log_error
  USE Numerics, only: r8_
  USE ErrorHandler, only: no_error, error_xml, error_programming, &
                          error_opening_file, error_reading_file,&
                          error_allocate, error_writing_file
  USE LunManager, only: fileunit_stdout
  USE DateTimeMod, only: date_type, init_date, test_date, &
                         time_type, init_time, test_time
  !  #]
  !  #[ Parameter/Variable/Type declarations

  implicit none

  ! make all module contents public, unless explicitely statet otherwise
  public

  ! see: PCS-IPF-ICD, Generic Interface Guidelines
  ! document: ESA-ID-ACS-GS-0001, issue 1.1, dd.20/02/2004

  !  #[ declarations needed for Global_Configuration
  ! REMARK: this is the <Ipf_Conf> block in the JobOrder file.

  integer, parameter :: maxlen_proc_name    = 15
  ! no requirements are set on this field. This field defines a symbolic
  ! name for the chains of processes (i.e. L2BP+RepGen in our case)
  integer, parameter :: maxlen_version      =  5
  ! format "XX.XX" with X an integer between 0-9
  integer, parameter :: maxlen_order_type   = 4
  ! Should be either "TEST", or another 4 character string 
  ! See SPR 28 on this requirement
  integer, parameter :: maxlen_log_level    =  8
  ! should be one of: DEBUG, INFO, PROGRESS, WARNING, ERROR
  ! REMARK: "NOOP" is not specified in the interface guidelines,
  ! but seems to be used by the L1BP anyway ???
  integer, parameter :: maxlen_Boolean      =  5
  integer, parameter :: maxlen_acq_station  = 20
  integer, parameter :: maxlen_proc_station = 20
  integer, parameter :: maxlen_jobdatetime  = 21
  ! format "YYYYMMDD_HHMMSSSSSSSS"
  ! (remark: this differs from the XML and the KVT datetime definition ...)

  TYPE Global_Configuration_type
     character(len=maxlen_proc_name)    :: Processor_Name
     character(len=maxlen_version)      :: Version
     character(len=maxlen_order_type)   :: Order_Type
     character(len=maxlen_log_level)    :: Stdout_Log_Level
     character(len=maxlen_log_level)    :: Stderr_Log_Level
     character(len=maxlen_Boolean)      :: Test
     character(len=maxlen_Boolean)      :: Breakpoint_Enable
     character(len=maxlen_acq_station)  :: Acquisition_Station
     character(len=maxlen_proc_station) :: Processing_Station
     ! not used element (for now)
     ! Config_Files: since we don't intent to use multiple config spaces
     character(len=maxlen_jobdatetime)  :: Sensing_Time_Start
     character(len=maxlen_jobdatetime)  :: Sensing_Time_Stop
  END TYPE Global_Configuration_type
  !  #]
  !  #[ declarations needed for Proc_Instructions
  ! REMARK: this is the <List_of_Ipf_Procs> block in the JobOrder file.

  integer, parameter :: maxlen_file_name = 256
  ! the field in FixedHeader.F90 used 62, but we have to add at least 4
  ! for possible extensions, and many more if a leading path is given
  ! so set it to 256 for the time being

! WARNING: it is not yet clear to me if it is possible to have the
!          special begin/end of mission codes. These have been defined
!          for the validity start/stop fields in the FH, but might be
!          usefull here as well.
!          See sec. 4.1.5.2 on p.17/18 and sec. 7.1, table 7.1.0-2 on p.38
!          of: PE=TN-ESA-GS-0001, issue 1.4, dd.13-Jun-2003
! These codes would be: "0000-00-00T00:00:00" 
!                  and: "9999-99-99T99:99:99"
! The Check_JobOrder_DateTime() routine cannot yet handle these special
! code and will generate an error if they occur.

  TYPE Time_Interval_Type
     character(len=maxlen_jobdatetime)  :: Start
     character(len=maxlen_jobdatetime)  :: Stop
     character(len=maxlen_file_name)    :: File_Name
  END TYPE Time_Interval_Type

  integer, parameter :: maxlen_file_type = 10 ! copied from FixedHeader.F90
  ! possible values for the L2BP are: (see also in FixedHeader.F90)
  ! ALD_U_N_1B or ALD_*_*_1B for the L1B input file
  ! ...
  ! possible value for the ReportGenerator
  ! (its not yet sure whether we have to program it ourself or not)
  ! LOG for the logfile containing all processor log messages
  ! Note that the example JobOrder file uses types
  ! ProductReport, ProductList for the ReportGenerator task !
  ! which need maxlen_file_type=15 !!

  integer, parameter :: maxlen_file_name_type = 9
  ! possible values are:
  ! Physical, Logical, Stem, Prefix, Regexp
  ! and for outputs also Directory is possible
  ! maxlen_file_name is defined above, just above TYPE Time_Interval_Type

  TYPE Input_type
     character(len=maxlen_file_type)      :: File_Type 
     character(len=maxlen_file_name_type) :: File_Name_Type
     character(len=maxlen_file_name), dimension(:), pointer :: File_Name
     TYPE(Time_Interval_type),        dimension(:), pointer :: Time_Interval
  END TYPE Input_type

  TYPE Output_type
     character(len=maxlen_file_type)      :: File_Type 
     character(len=maxlen_file_name_type) :: File_Name_Type
     character(len=maxlen_file_name)      :: File_Name
  END TYPE Output_type

  integer, parameter :: maxlen_task_name    = 15
  ! no requirements are set on this field. I assume this field defines
  ! the executable name for the process to run (i.e. L2BP in our case)
  ! maxlen_version is defined above, just above TYPE Global_Configuration_type
  ! maxlen_Boolean is defined above, just above TYPE Global_Configuration_type

  TYPE Proc_Instructions_type
     character(len=maxlen_task_name)    :: Task_Name
     character(len=maxlen_version)      :: Task_Version
! this element is not present in the JobOrder files generated by
! the ThinLayer emulator....
!     character(len=maxlen_Boolean)      :: Breakpoint_Enable
!
     ! not used element (for now)
     ! Breakpoint/List_of_Brk_Files
     TYPE(Input_type),  dimension(:), pointer :: Input
     TYPE(Output_type), dimension(:), pointer :: Output
  END TYPE Proc_Instructions_type
  !  #]
  !  #[ declarations needed for Special_Settings
  ! REMARK: this is the <Processor_Conf> block in th JobOrder file.

  TYPE Special_Settings_type
     ! NOTE: this file holds (additional?) processor settings. Remind
     ! however that in JobOrder jargon the processor is a sequence of
     ! Tasks/executables, so this settings file applies to ALL OF THEM.
     ! Therefore this is not the same as the L2B_Aux_Par file that we use
     ! to define input settings for the L2BP !!!!!!! 
     ! (which is just specified as one of the input files on the input 
     ! files list of this task)
     ! and probably we won't use this file at all for the L2BP.
     character(len=maxlen_file_name) :: File_Name
  END TYPE Special_Settings_type
  !  #]

  TYPE JobOrderData_type
     TYPE(Global_Configuration_type) :: Global_Configuration
     ! only the instructions for the L2BP will be loaded from the
     ! Proc_Instructions section, so no array of Proc_Instructions is needed
     TYPE(Proc_Instructions_type)    :: Proc_Instructions
     TYPE(Special_Settings_type)     :: Special_Settings
  END TYPE JobOrderData_type

  !  #]
  !  #[ Module interface
  ! all is public for now
  !  #]
contains
  !-------------------------------------------------
  ! Global_Configuration handling
  !-------------------------------------------------
  subroutine Init_Global_Configuration(Global_Configuration)
    !  #[

    TYPE(Global_Configuration_type), intent(out) :: Global_Configuration

    Global_Configuration%Processor_Name(:)      = ' '
    Global_Configuration%Version(:)             = ' '
    Global_Configuration%Order_Type(:)          = ' '
    Global_Configuration%Stdout_Log_Level(:)    = ' '
    Global_Configuration%Stderr_Log_Level(:)    = ' '
    Global_Configuration%Test(:)                = ' '
    Global_Configuration%Breakpoint_Enable(:)   = ' '
    Global_Configuration%Acquisition_Station(:) = ' '
    Global_Configuration%Processing_Station(:)  = ' '
    Global_Configuration%Sensing_Time_Start(:)  = ' '
    Global_Configuration%Sensing_Time_Stop(:)   = ' '
    
  end subroutine Init_Global_Configuration

    !  #]
  subroutine AsciiWrite_Global_Configuration(Global_Configuration,&
                                             fileunit,prefix,error_flag)
    !  #[
    TYPE(Global_Configuration_type), &
                      intent(in)  :: Global_Configuration
    integer,          intent(in)  :: fileunit
    character(len=*), intent(in)  :: prefix
    integer,          intent(out) :: error_flag

    ! local variable
    character(len=256) :: txt

    error_flag = no_error

    txt = Global_Configuration%Processor_Name
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Processor_Name      = "//trim(txt)

    txt = Global_Configuration%Version
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Version             = "//trim(txt)

    txt = Global_Configuration%Order_Type
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Order_Type          = "//trim(txt)

    txt = Global_Configuration%Stdout_Log_Level
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Stdout_Log_Level    = "//trim(txt)

    txt = Global_Configuration%Stderr_Log_Level
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Stderr_Log_Level    = "//trim(txt)

    txt = Global_Configuration%Test
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Test                = "//trim(txt)

    txt = Global_Configuration%Breakpoint_Enable
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Breakpoint_Enable   = "//trim(txt)

    txt = Global_Configuration%Acquisition_Station
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Acquisition_Station = "//trim(txt)

    txt = Global_Configuration%Processing_Station
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Processing_Station  = "//trim(txt)

    ! not used element (for now)
    ! Config_Files: since we don't intent to use multiple config spaces

    txt = Global_Configuration%Sensing_Time_Start
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Sensing_Time_Start  = "//trim(txt)

    txt = Global_Configuration%Sensing_Time_Stop
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Sensing_Time_Stop   = "//trim(txt)

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_Global_Configuration:"//&
         " error while writing to outputfile")
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Global_Configuration
    !  #]
  subroutine Get_Global_Config_from_xml_tree(xml_tree, xml_path, &
                                       Global_Configuration, error_flag)
    !  #[
    TYPE(xml_node_type), pointer     :: xml_tree ! input
    character(len=*),    intent(in)  :: xml_path ! input
    TYPE(Global_Configuration_type), &
                         intent(out) :: Global_Configuration
    integer,             intent(out) :: error_flag

    ! local variables
    character(len=max_str_len)  :: text
    character(len=max_str_len)  :: path

    ! init
    error_flag = no_error

    ! copying xml items
    path = trim(xml_path)//"/Processor_Name"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,&
                      maxlen_proc_name,"Get_Global_Config_from_xml_tree")
    Global_Configuration%Processor_Name = text

    path = trim(xml_path)//"/Version"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,&
                      maxlen_version,"Get_Global_Config_from_xml_tree")
    Global_Configuration%Version = text

    path = trim(xml_path)//"/Order_Type"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,&
                      maxlen_order_type,"Get_Global_Config_from_xml_tree")
    Global_Configuration%Order_Type = text

    path = trim(xml_path)//"/Stdout_Log_Level"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,&
                      maxlen_log_level,"Get_Global_Config_from_xml_tree")
    Global_Configuration%Stdout_Log_Level = text

    path = trim(xml_path)//"/Stderr_Log_Level"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,&
                      maxlen_log_level,"Get_Global_Config_from_xml_tree")
    Global_Configuration%Stderr_Log_Level = text

    path = trim(xml_path)//"/Test"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,&
                      maxlen_Boolean,"Get_Global_Config_from_xml_tree")
    Global_Configuration%Test = text

    path = trim(xml_path)//"/Breakpoint_Enable"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,&
                      maxlen_Boolean,"Get_Global_Config_from_xml_tree")
    Global_Configuration%Breakpoint_Enable = text

    path = trim(xml_path)//"/Acquisition_Station"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,&
                   maxlen_acq_station,"Get_Global_Config_from_xml_tree")
    Global_Configuration%Acquisition_Station = text

    path = trim(xml_path)//"/Processing_Station"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,&
                   maxlen_proc_station,"Get_Global_Config_from_xml_tree")
    Global_Configuration%Processing_Station = text

    path = trim(xml_path)//"/Sensing_Time/Start"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,&
                      maxlen_jobdatetime,"Get_Global_Config_from_xml_tree")
    Global_Configuration%Sensing_Time_Start = text

    path = trim(xml_path)//"/Sensing_Time/Stop"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,&
                      maxlen_jobdatetime,"Get_Global_Config_from_xml_tree")
    Global_Configuration%Sensing_Time_Stop = text

  end subroutine Get_Global_Config_from_xml_tree
    !  #]
  subroutine Check_Global_Configuration(Global_Configuration, error_flag)
    !  #[

    ! => do a sanity check on the input values 
    TYPE(Global_Configuration_type), &
                      intent(in)  :: Global_Configuration
    integer,          intent(out) :: error_flag

    ! local variables
    logical :: problem_found

    problem_found = .false.
    error_flag = no_error

! the way I interpret this now is that this is a symbolic name for the
! chain of processes (in our case the L2BP+ReportGenerator) so no
! check is possible on this name for now

    ! check the Processor_Name
!    IF (trim(Global_Configuration%Processor_Name) .ne. software_ProcName) THEN
!       call logmsg(log_error,"ERROR in Check_Global_Configuration():"//&
!            " this JobOrder requests use of a processor named: ["//&
!            trim(Global_Configuration%Processor_Name)//&
!            "] but the current processor is named: ["//software_ProcName//"]")
!       problem_found = .true.
!    END IF

! the way I interpret this now is that this is a version for the
! chain of processes (in our case the L2BP+ReportGenerator) so no
! check is possible on this version for now

    ! check the  Version
!    IF (trim(Global_Configuration%Version) .ne. software_version) THEN
!       call logmsg(log_error,"ERROR in Check_Global_Configuration():"//&
!            " this JobOrder requests use of processor version: ["//&
!            trim(Global_Configuration%Version)//&
!            "] but the current processor has version: ["//&
!            software_Version//"]")
!       problem_found = .true.
!    END IF

    ! check for the correct format of the version string: XX.XX
    call Check_JobOrder_Version(Global_Configuration%Version,&
                                "Global_Configuration%Version",&
                                problem_found)

! Disabled this check, 6-Dec-2007, JK
! Because SPR 28 states that the ACMF and LTA wish to be able to implement
! their own private order types. Therefor we should just copy the JobOrder
! Order_Type field, and check that it has 4 characters, and nothing more.

    ! check the Order_Type
    ! should be one of: NRT, SYSTEMATIC, REPROCESSING, SUBS
    ! JK: 03-Apr-2007, updated to the list currently defined in our
    !     example  WorkstationConfigurationFile.
    ! new list is: NRT, OFFL, RPRO, Reprocess, Systematic, Backlog_Processing
!    select case(trim(Global_Configuration%Order_Type))
!    case("NRT","OFFL","RPRO","Reprocess","Systematic","Backlog_Processing")
!       ! ok, this is a known Order_Type
!    case default
!       call logmsg(log_error,"ERROR in Check_Global_Configuration():"//&
!            " this JobOrder requests an Order_Type: ["//&
!            trim(Global_Configuration%Order_Type)//"] but the allowed"//&
!            " types are: "//&
!            "[NRT|OFFL|RPRO|Reprocess|SYSTEMATIC|Backlog_Processing]")
!!            "[NRT|SYSTEMATIC|REPROCESSING|SUBS]")
!       problem_found = .true.
!    end select

    ! check the  Stdout_Log_Level
    ! should be one of: DEBUG, INFO, PROGRESS, WARNING, ERROR
    ! REMARK: "NOOP" is not specified in the interface guidelines,
    ! but seems to be used anyway ???
    select case(trim(Global_Configuration%Stdout_Log_Level))
    case("DEBUG","INFO","PROGRESS","WARNING","ERROR")
       ! ok, this is a known Log_Level type
    case("NOOP")
       call logmsg(log_warn,"WARNING in Check_Global_Configuration():"//&
            " this JobOrder requests a Stdout_Log_Level of: ["//&
            trim(Global_Configuration%Stdout_Log_Level)//"] which is not"//&
            " one of the allowed types, according to the Generic Interface"//&
            " Guidelines. Allowing this for now, since L1BP also uses this")
    case default
       call logmsg(log_error,"ERROR in Check_Global_Configuration():"//&
            " this JobOrder requests a Stdout_Log_Level of: ["//&
            trim(Global_Configuration%Stdout_Log_Level)//"] but the allowed"//&
            " types are: [DEBUG|INFO|PROGRESS|WARNING|ERROR] "//&
            "(and maybe NOOP???)")
       problem_found = .true.
    end select

    ! check the  Stderr_Log_Level
    ! should be one of: DEBUG, INFO, PROGRESS, WARNING, ERROR
    ! REMARK: "NOOP" is not specified in the interface guidelines,
    ! but seems to be used anyway ???
    select case(trim(Global_Configuration%Stderr_Log_Level))
    case("DEBUG","INFO","PROGRESS","WARNING","ERROR")
       ! ok, this is a known Log_Level type
    case("NOOP")
       call logmsg(log_warn,"WARNING in Check_Global_Configuration():"//&
            " this JobOrder requests a Stderr_Log_Level of: ["//&
            trim(Global_Configuration%Stderr_Log_Level)//"] which is not"//&
            " one of the allowed types, according to the Generic Interface"//&
            " Guidelines. Allowing this for now, since L1BP also uses this")
    case default
       call logmsg(log_error,"ERROR in Check_Global_Configuration():"//&
            " this JobOrder requests a Stderr_Log_Level of: ["//&
            trim(Global_Configuration%Stderr_Log_Level)//" but the allowed"//&
            "] types are: [DEBUG|INFO|PROGRESS|WARNING|ERROR] "//&
            "(and maybe NOOP???)")
       problem_found = .true.
    end select

    ! check the Test
    ! should be either true or false
    select case(to_lowercase(trim(Global_Configuration%Test)))
    case("false","true")
       ! ok, this is a known boolean type
    case default
       call logmsg(log_error,"ERROR in Check_Global_Configuration():"//&
            " this JobOrder has a value for the boolean Test field of: ["//&
            trim(Global_Configuration%Test)//" but the allowed"//&
            "] types are: [true|false]")
       problem_found = .true.
    end select

    ! check the Breakpoint_Enable item
    ! should be either true or false
    select case(to_lowercase(trim(Global_Configuration%Breakpoint_Enable)))
    case("false")
       ! ok, this is a known boolean type
    case("true")
       ! this is a valid boolean but this feature is not implemented
       call logmsg(log_error,"ERROR in Check_Global_Configuration():"//&
            " this JobOrder has a value for the boolean Breakpoint_Enable"//&
            " field of: ["//trim(Global_Configuration%Breakpoint_Enable)//&
            "] but the use of breakpoints"//&
            " is not implemented in the L2B-Processor")
    case default
       call logmsg(log_error,"ERROR in Check_Global_Configuration():"//&
            " this JobOrder has a value for the boolean Breakpoint_Enable"//&
            " field of: ["//trim(Global_Configuration%Breakpoint_Enable)//&
            "] but the allowed types are: [true|false]")
       problem_found = .true.
    end select

    ! no check implemented on the Acquisition_Station
    ! no check implemented on the Processing_Station

    ! not used element (for now)
    ! Config_Files: since we don't intent to use multiple config spaces

    ! check the Sensing_Time_Start
    call Check_JobOrder_DateTime(Global_Configuration%Sensing_Time_Start,&
                                 "Global_Configuration%Sensing_Time_Start",&
                                 problem_found)

    ! check the Sensing_Time_Stop
    call Check_JobOrder_DateTime(Global_Configuration%Sensing_Time_Stop,&
                                 "Global_Configuration%Sensing_Time_Stop",&
                                 problem_found)

!    IF (problem_found) error_flag = error_reading_file

  end subroutine Check_Global_Configuration
    !  #]
  !-------------------------------------------------
  ! Proc_Instructions handling
  !-------------------------------------------------
  subroutine Init_Time_Interval(Time_Interval)
    !  #[
    TYPE(Time_Interval_type), intent(out) :: Time_Interval

    Time_Interval%Start(:)     = ' '
    Time_Interval%Stop(:)      = ' '
    Time_Interval%File_Name(:) = ' '

  end subroutine Init_Time_Interval
    !  #]
  subroutine AsciiWrite_Time_Interval(Time_Interval,fileunit,prefix,error_flag)
    !  #[
    TYPE(Time_Interval_type), intent(in)  :: Time_Interval
    integer,                  intent(in)  :: fileunit
    character(len=*),         intent(in)  :: prefix
    integer,                  intent(out) :: error_flag

    ! local variable
    character(len=256) :: txt

    error_flag = no_error

    txt = Time_Interval%Start
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) trim(prefix)//"%Start     = "//trim(txt)

    txt = Time_Interval%Stop
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) trim(prefix)//"%Stop      = "//trim(txt)

    txt = Time_Interval%File_Name
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) trim(prefix)//"%File_Name = "//trim(txt)

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_Time_Interval:"//&
         " error while writing to outputfile")
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Time_Interval
    !  #]
  subroutine Get_Time_Interv_from_xml_tree(xml_tree, xml_path, &
                                           Time_Interval, error_flag)
    !  #[
    TYPE(xml_node_type),      pointer     :: xml_tree ! input
    character(len=*),         intent(in)  :: xml_path ! input
    TYPE(Time_Interval_type), intent(out) :: Time_Interval
    integer,                  intent(out) :: error_flag

    ! local variables
    character(len=max_str_len)  :: text
    character(len=max_str_len)  :: path
    character(len=max_str_len)  :: filename
    character(len=max_str_len)  :: extension

    ! init
    error_flag = no_error

    ! copying xml items
    path = trim(xml_path)//"/Start"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,&
                      maxlen_jobdatetime,"Get_Time_Interv_from_xml_tree")
    Time_Interval%Start = text

    path = trim(xml_path)//"/Stop"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,&
                      maxlen_jobdatetime,"Get_Time_Interv_from_xml_tree")
    Time_Interval%Stop = text

    path = trim(xml_path)//"/File_Name"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,&
                      maxlen_file_name,"Get_Time_Interv_from_xml_tree")

    filename(:)=' '
    extension(:)=' '
    call Split_filename_and_extension(text,filename,extension)
    if (trim(extension) .eq. "EEF") then
       Time_Interval%File_Name = text
    else
       Time_Interval%File_Name = filename
    end if

  end subroutine Get_Time_Interv_from_xml_tree
    !  #]
  subroutine Check_Time_Interval(Time_Interval,problem_found)
    !  #[
    TYPE(Time_Interval_type), intent(in)    :: Time_Interval
    logical,                  intent(inout) :: problem_found

    call Check_JobOrder_DateTime(Time_Interval%Start,&
                                 "Time_Interval%Start",problem_found)
    call Check_JobOrder_DateTime(Time_Interval%Stop,&
                                 "Time_Interval%Stop",problem_found)

  end subroutine Check_Time_Interval
    !  #]
  !-------------------------------------------------
  subroutine Init_Input(Input)
    !  #[
    TYPE(Input_type), intent(out) :: Input

    Input%File_Type(:)      = ' '
    Input%File_Name_Type(:) = ' '
    nullify(Input%File_Name)
    nullify(Input%Time_Interval)

  end subroutine Init_Input
    !  #]
  subroutine Delete_Input(Input)
    !  #[
    TYPE(Input_type), intent(inout) :: Input

    IF (associated(Input%File_Name)) THEN
       deallocate(Input%File_Name)
    END IF

    IF (associated(Input%Time_Interval)) THEN
       deallocate(Input%Time_Interval)
    END IF

  end subroutine Delete_Input
    !  #]
  subroutine AsciiWrite_Input(Input,fileunit,prefix,error_flag)
    !  #[
    TYPE(Input_type), intent(in)  :: Input
    integer,          intent(in)  :: fileunit
    character(len=*), intent(in)  :: prefix
    integer,          intent(out) :: error_flag

    ! local variable
    character(len=256) :: txt
    character(len=2)   :: str_cntr
    integer            :: i,s

    error_flag = no_error

    txt = Input%File_Type
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) trim(prefix)//"%File_Type      = "//trim(txt)

    txt = Input%File_Name_Type
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) trim(prefix)//"%File_Name_Type = "//trim(txt)

    IF (associated(Input%File_Name)) THEN
       s = size(Input%File_Name)
       DO i=1,s
          write(str_cntr,"(i2.2)",err=998) i
          txt = Input%File_Name(i)
          IF (len_trim(txt).eq.0) txt = "[empty]"
          write(fileunit,"(a)",err=999) trim(prefix)//"%File_Name("//&
               trim(str_cntr)//")  = "//trim(txt)
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%File_Name(:) [seems not associated]"
    END IF

    IF (associated(Input%Time_Interval)) THEN
       s = size(Input%Time_Interval)
       DO i=1,s
          write(str_cntr,"(i2.2)",err=998) i
          call AsciiWrite_Time_Interval(Input%Time_Interval(i),&
               fileunit,trim(prefix)//"%Time_Interval("//trim(str_cntr)//")",&
               error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%Time_Interval(:) [seems not associated]"
    END IF

    return

998 call logmsg(log_error,"ERROR conversion error in AsciiWrite_Input:"//&
         " could not convert the counter i to a string ...")
    error_flag = error_writing_file
    return

999 call logmsg(log_error,"ERROR in AsciiWrite_Input:"//&
         " error while writing to outputfile")
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Input
    !  #]
  subroutine Get_Input_from_xml_tree(xml_tree, xml_path, Input, error_flag, &
                                     workaround)
    !  #[
    TYPE(xml_node_type), pointer     :: xml_tree ! input
    character(len=*),    intent(in)  :: xml_path ! input
    TYPE(Input_type),    intent(out) :: Input
    integer,             intent(out) :: error_flag
    logical, optional,   intent(in)  :: workaround

    ! local variables
    character(len=max_str_len) :: text
    character(len=max_str_len) :: path
    character(len=max_str_len) :: attr_name, attr_value
    character(len=max_str_len) :: filename
    character(len=max_str_len) :: extension
    character(len=3)           :: count_string
    integer                    :: file_count, ios, i, AllocStatus
    integer                    :: time_int_count

    ! init
    error_flag = no_error

    ! copying xml items
    path = trim(xml_path)//"/File_Type"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,&
                      maxlen_file_type,"Get_Input_from_xml_tree")
    Input%File_Type = text

    path = trim(xml_path)//"/File_Name_Type"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,&
                      maxlen_file_name_type,"Get_Input_from_xml_tree")
    Input%File_Name_Type = text

    ! determine the number of input files of this type
    path = trim(xml_path)//"/List_of_File_Names"
    attr_name = "count"
    call xml_tree_read_string_attribute(xml_tree,path,attr_name,&
                                        attr_value,error_flag)
    IF (error_flag .ne. no_error) return
    read(attr_value,"(i3)",iostat=ios) file_count
    IF (ios .ne. 0) goto 999

    ! allocate space to store the list of files
    !print *,"nr of input files: ",file_count
    allocate(Input%File_Name(file_count),STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       call logmsg(log_error,"ERROR allocation problem in "//&
            "Get_Input_from_xml_tree()")
       error_flag = error_allocate
       return
    END IF

    ! Initialise and read the input files
    DO i=1,file_count
       Input%File_Name(i) = "[undefined]"

       write(count_string,"(i3)",err=998) i
       path = trim(xml_path)//"/List_of_File_Names/File_Name["//&
            trim(adjustl(count_string))//"]"

       call xml_node_get_string(xml_tree,path,text,error_flag)
       IF (error_flag .ne. no_error) return
       text = check_length_xml_item(text,path,&
            maxlen_file_name,"Get_Input_from_xml_tree")

       ! this split function replaces the workaround below
       filename(:)=' '
       extension(:)=' '
       call Split_filename_and_extension(text,filename,extension)
       if (trim(extension) .eq. "EEF") then
          Input%File_Name(i) = text
       else
          Input%File_Name(i) = filename
       end if

! this is no longer needed, so the workaround switch may be phased out here.

!       IF (present(workaround)) THEN
!          IF (workaround .and. (Input%File_Name_Type .eq. "Physical")) THEN
!             IF (len_trim(text) .gt. 4) THEN
!                IF (text(len_trim(text)-3:len_trim(text)) .eq. ".DBL") THEN
!                   ! implement workaround for automatic removal of a file
!                   ! extension (like .DBL). This is needed because the
!                   ! ThinLayer emulator has a bug: it adds an extension 
!                   ! even if only the stem of the file should be given...
!                   print *,"Applying workaround for bug in JobOrder generation"
!                   print *,"for inputfile type Physical: "//&
!                           "removing extension from"
!                   print *,"file: ",trim(text)
!                   Input%File_Name(i) = text(1:len_trim(text)-4)
!                END IF 
!                ! no else needed, if no dot is present we have a filename stem,
!                ! so all is OK. (this happens in our own test-suite)
!             ELSE
!                call logmsg(log_error,"Error in: Get_Input_from_xml_tree()"//&
!                     " file name seems too short to strip a 4 char extension"//&
!                     " from it: "//trim(text))
!                error_flag = error_reading_file
!                return
!             END IF
!          END IF
!       END IF
    END DO

    ! determine the number of time intervals
    path = trim(xml_path)//"/List_of_Time_Intervals"
    attr_name = "count"
    call xml_tree_read_string_attribute(xml_tree,path,attr_name,&
                                        attr_value,error_flag)
    IF (error_flag .ne. no_error) return
    read(attr_value,"(i3)",iostat=ios) time_int_count
    IF (ios .ne. 0) goto 999

    ! allocate space to store the list of time intervals
    !print *,"nr of time intervals: ",time_int_count
    allocate(Input%Time_Interval(time_int_count),STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       call logmsg(log_error,"ERROR allocation problem in "//&
            "Get_Input_from_xml_tree()")
       error_flag = error_allocate
       return
    END IF

    ! Initialise and read the time intervals
    DO i=1,time_int_count
       call Init_Time_Interval(Input%Time_Interval(i))

       write(count_string,"(i3)",err=998) i
       path = trim(xml_path)//"/List_of_Time_Intervals/Time_Interval["//&
            trim(adjustl(count_string))//"]"

       call Get_Time_Interv_from_xml_tree(xml_tree, path, &
                    Input%Time_Interval(i), error_flag)

       ! TODO: ==>Time_Interval%File_Name
       ! I could check whether this name is a valid name in the
       ! list of File_names above, but selecting a time interval
       ! is not yet implemented in the L2BP so its not yet clear
       ! how to do that.
    END DO

    return

998 call logmsg(log_error,&
         "ERROR conversion error in Get_Input_from_xml_tree:"//&
         " could not convert the counter i to a string ...")
    error_flag = error_writing_file
    return

999 call logmsg(log_error,"ERROR in Get_Input_from_xml_tree():"//&
         " could not convert count attribute of the "//trim(xml_path)//&
         " element into an integer number")
    error_flag = error_reading_file
    return

  end subroutine Get_Input_from_xml_tree
    !  #]
  subroutine Check_Input(Input,problem_found)
    !  #[
    TYPE(Input_type), intent(in)    :: Input
    logical,          intent(inout) :: problem_found

    ! local variable
    logical :: type_problem_found
    integer :: i,s

    !Input%File_Type
    ! todo: add a check on valid known filetypes for the L2BP
    ! this is done below, in combination of the File_Name_Type
    ! because each file can only be specified with a specific type

    type_problem_found = .false.
    select case(trim(Input%File_Name_Type))
    case("Physical")
       ! ok, all *.EEF files are of this type
    case("Logical")
       ! this type is probably not used for the L2BP
       type_problem_found = .true.
    case("Stem")
       ! ok, this type is used for the L1B input file
       ! and several aux input files
    case("Prefix")
       ! this type is probably not used for the L2BP
       type_problem_found = .true.
    case("Regexp")
       ! do we use this?
       type_problem_found = .true.
    case default
       ! unknown type
       call logmsg(log_error,"ERROR in Check_Input():"//&
            " unknown file name type "//trim(Input%File_Name_Type))
       problem_found = .true.
       return
    end select

    IF (type_problem_found) THEN 
       ! issue only a warning for now, not a real problem
       call logmsg(log_warn,"WARNING in Check_Input():"//&
            " input type "//trim(Input%File_Type)//&
            " should not occur in combination with file name type "//&
            trim(Input%File_Name_Type))
    END IF

    ! Input%File_Name
    ! I could check here if the file exists, but I trust the ThinLayer
    ! takes care of that. Also, the processor will issue a proper 
    ! warning once it tries to open the file, so no check is implemented.
    ! Note also that if executables are chained, some files will not
    ! have been created at the time the first executable reads the
    ! joborder file. Therefore checking whether all needed input
    ! files are present is done directly after reading the joborder data
    ! in the main program.

    s=size(Input%Time_Interval)
    DO i=1,s
       call Check_Time_Interval(Input%Time_Interval(i),problem_found)
    END DO

  end subroutine Check_Input
    !  #]
  !-------------------------------------------------
  subroutine Init_Output(Output)
    !  #[
    TYPE(Output_type), intent(out) :: Output

    Output%File_Type(:)      = ' '
    Output%File_Name_Type(:) = ' '
    Output%File_Name(:)      = ' '

  end subroutine Init_Output
    !  #]
  subroutine AsciiWrite_Output(Output,fileunit,prefix,error_flag)
    !  #[
    TYPE(Output_type), intent(in)  :: Output
    integer,           intent(in)  :: fileunit
    character(len=*),  intent(in)  :: prefix
    integer,           intent(out) :: error_flag

    ! local variable
    character(len=256) :: txt

    error_flag = no_error

    txt = Output%File_Type
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) trim(prefix)//"%File_Type      = "//trim(txt)

    txt = Output%File_Name_Type
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) trim(prefix)//"%File_Name_Type = "//trim(txt)

    txt = Output%File_Name
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) trim(prefix)//"%File_Name      = "//trim(txt)

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_Output:"//&
         " error while writing to outputfile")
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Output
    !  #]
  subroutine Get_Output_from_xml_tree(xml_tree, xml_path, Output, error_flag,&
                                      workaround)
    !  #[

    TYPE(xml_node_type), pointer     :: xml_tree ! input
    character(len=*),    intent(in)  :: xml_path ! input
    TYPE(Output_type),   intent(out) :: Output
    integer,             intent(out) :: error_flag
    logical, optional,   intent(in)  :: workaround

    ! local variables
    character(len=max_str_len)  :: text
    character(len=max_str_len)  :: path
    character(len=max_str_len)  :: filename
    character(len=max_str_len)  :: extension

    ! init
    error_flag = no_error

    ! copying xml items
    path = trim(xml_path)//"/File_Type"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,&
                      maxlen_file_type,"Get_Output_from_xml_tree")
    Output%File_Type = text

    path = trim(xml_path)//"/File_Name_Type"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,&
                      maxlen_file_name_type,"Get_Output_from_xml_tree")
    Output%File_Name_Type = text

    path = trim(xml_path)//"/File_Name"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,&
                      maxlen_file_name,"Get_Output_from_xml_tree")
    ! this split function replaces the workaround below
    filename(:)=' '
    extension(:)=' '
    call Split_filename_and_extension(text,filename,extension)
    if (trim(extension) .eq. "EEF") then
       Output%File_Name = text
    else
       Output%File_Name = filename
    end if

! this is no longer needed, so the workaround switch may be phased out here.

!    IF (present(workaround)) THEN
!       IF (workaround .and. (Output%File_Name_Type .eq. "Physical")) THEN
!          IF (len_trim(text) .gt. 4) THEN
!             IF (text(len_trim(text)-3:len_trim(text)) .eq. ".DBL") THEN
!                print *,"Applying workaround for bug in JobOrder generation"
!                print *,"for outputfile type Physical: "//&
!                        "removing extension from: "
!                print *,"file: ",trim(text)
!                Output%File_Name = text(1:len_trim(text)-4)
!             END IF
!                ! no else needed, if no dot is present we have a filename stem,
!                ! so all is OK. (this happens in our own test-suite)
!          ELSE
!             call logmsg(log_error,"Error in: Get_Output_from_xml_tree()"//&
!                  " file name seems too short to strip a 4 char extension"//&
!                  " from it: "//trim(text))
!             error_flag = error_reading_file
!             return
!          END IF
!       END IF
!    END IF

  end subroutine Get_Output_from_xml_tree

    !  #]
  subroutine Check_Output(Output,problem_found)
    !  #[
    TYPE(Output_type), intent(in)    :: Output
    logical,           intent(inout) :: problem_found

    ! local variable
    logical     :: type_problem_found

    !Output%File_Type
    ! todo: add a check on valid known filetypes for the L2BP
    ! this is done below, in combination of the File_Name_Type
    ! because each file can only be specified with a specific type

    type_problem_found = .false.
    select case(trim(Output%File_Name_Type))
    case("Physical")
       ! this type is probably not used for the L2BP output
       type_problem_found = .true.
    case("Logical")
       ! this type is probably not used for the L2BP output
       type_problem_found = .true.
    case("Stem")
       ! this type is probably not used for the L2BP output
       type_problem_found = .true.
    case("Prefix")
       ! this type is probably not used for the L2BP output
       type_problem_found = .true.
    case("Regexp")
       ! do we use this?
       type_problem_found = .true.
    case("Directory")
       ! this is the expected setting, since the filenames for the
       ! output product are generated, so only the dir needs to be specified
    case default
       ! unknown type
       call logmsg(log_error,"ERROR in Check_Output():"//&
            " unknown file name type "//trim(Output%File_Name_Type))
       problem_found = .true.
       return
    end select

    IF (type_problem_found) THEN 
       ! issue only a warning for now, not a real problem
       call logmsg(log_warn,"WARNING in Check_output():"//&
            " Output type "//trim(Output%File_Type)//&
            " should not occur in combination with file name type "//&
            trim(Output%File_Name_Type))
    END IF

    ! Output%File_Name
    ! no check implemented on this item.
    ! we could check:
    ! -whether this directory exists
    ! -whether it is writable

  end subroutine Check_Output
    !  #]
  !-------------------------------------------------
  subroutine Init_Proc_Instructions(Proc_Instructions)
    !  #[
    TYPE(Proc_Instructions_type), intent(out) :: Proc_Instructions
    
    Proc_Instructions%Task_Name(:)         = ' '
    Proc_Instructions%Task_Version(:)      = ' '
! this element is not present in the JobOrder files generated by
! the ThinLayer emulator....
!    Proc_Instructions%Breakpoint_Enable(:) = ' '
    nullify(Proc_Instructions%Input)
    nullify(Proc_Instructions%Output)

  end subroutine Init_Proc_Instructions
    !  #]
  subroutine Delete_Proc_Instructions(Proc_Instructions)
    !  #[
    TYPE(Proc_Instructions_type), intent(inout) :: Proc_Instructions

    ! local variables
    integer :: s, i
    
    IF (associated(Proc_Instructions%Input)) THEN
       s=size(Proc_Instructions%Input)
       DO i=1,s
          call Delete_Input(Proc_Instructions%Input(i))
       END DO
       deallocate(Proc_Instructions%Input)
    END IF

    IF (associated(Proc_Instructions%Output)) THEN
       deallocate(Proc_Instructions%Output)
    END IF

  end subroutine Delete_Proc_Instructions
    !  #]
  subroutine AsciiWrite_Proc_Instructions(Proc_Instructions,&
                                          fileunit,prefix,error_flag)
    !  #[
    TYPE(Proc_Instructions_type), &
                      intent(in)  :: Proc_Instructions
    integer,          intent(in)  :: fileunit
    character(len=*), intent(in)  :: prefix
    integer,          intent(out) :: error_flag

    ! local variable
    character(len=256) :: txt
    character(len=2)   :: str_cntr
    integer            :: i,s

    error_flag = no_error

    txt = Proc_Instructions%Task_Name
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Task_Name         = "//trim(txt)

    txt = Proc_Instructions%Task_Version
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Task_Version      = "//trim(txt)

! this element is not present in the JobOrder files generated by
! the ThinLayer emulator....
!    txt = Proc_Instructions%Breakpoint_Enable
!    IF (len_trim(txt).eq.0) txt = "[empty]"
!    write(fileunit,"(a)",err=999) &
!         trim(prefix)//"%Breakpoint_Enable = "//trim(txt)

     ! not used element (for now)
     ! List_of_Brk_Files

    IF (associated(Proc_Instructions%Input)) THEN
       s = size(Proc_Instructions%Input)
       DO i=1,s
          write(str_cntr,"(i2.2)",err=998) i
          call AsciiWrite_Input(Proc_Instructions%Input(i),&
               fileunit,trim(prefix)//"%Input("//trim(str_cntr)//")",&
               error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%Input(:) [seems not associated]"
    END IF

    IF (associated(Proc_Instructions%Output)) THEN
       s = size(Proc_Instructions%Output)
       DO i=1,s
          write(str_cntr,"(i2.2)",err=998) i
          call AsciiWrite_Output(Proc_Instructions%Output(i),&
               fileunit,trim(prefix)//"%Output("//trim(str_cntr)//")",&
               error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%Outputs(:) [seems not associated]"
    END IF

    return

998 call logmsg(log_error,"ERROR conversion error in "//&
         "AsciiWrite_Proc_Instructions: could not convert the"//&
         " counter i to a string ...")
    error_flag = error_writing_file
    return
    
999 call logmsg(log_error,"ERROR in AsciiWrite_Proc_Instructions:"//&
         " error while writing to outputfile")
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Proc_Instructions
    !  #]
  subroutine Get_Proc_Instruc_from_xml_tree(xml_tree, xml_path, &
                                            Proc_Instructions, error_flag, &
                                            workaround)
    !  #[
    TYPE(xml_node_type), pointer     :: xml_tree ! input
    character(len=*),    intent(in)  :: xml_path ! input
    TYPE(Proc_Instructions_type), &
                         intent(out) :: Proc_Instructions
    integer,             intent(out) :: error_flag
    logical, optional,   intent(in)  :: workaround

    ! local variables
    character(len=max_str_len)  :: text
    character(len=max_str_len)  :: path
    character(len=max_str_len)  :: attr_name, attr_value
    character(len=3)            :: count_string
    integer                     :: inp_count, outp_count
    integer                     :: ios, i, AllocStatus

    ! init
    error_flag = no_error

    ! copying xml items
    path = trim(xml_path)//"/Task_Name"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,&
                      maxlen_task_name,"Get_Proc_Instruc_from_xml_tree")
    Proc_Instructions%Task_Name = text

    path = trim(xml_path)//"/Task_Version"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,&
                      maxlen_version,"Get_Proc_Instruc_from_xml_tree")
    Proc_Instructions%Task_Version = text

! this element is not present in the JobOrder files generated by
! the ThinLayer emulator....
!    path = trim(xml_path)//"/Breakpoint/Enable"
!    call xml_node_get_string(xml_tree,path,text,error_flag)
!    IF (error_flag .ne. no_error) return
!    text = check_length_xml_item(text,path,&
!                      maxlen_boolean,"Get_Proc_Instruc_from_xml_tree")
!    Proc_Instructions%Breakpoint_Enable = text

    ! determine the number of input blocks
    path = trim(xml_path)//"/List_of_Inputs"
    attr_name = "count"
    call xml_tree_read_string_attribute(xml_tree,path,attr_name,&
                                        attr_value,error_flag)
    IF (error_flag .ne. no_error) return
    read(attr_value,"(i3)",iostat=ios) inp_count
    IF (ios .ne. 0) goto 999

    ! allocate space to store the inputs
    !print *,"nr of input blocks: ",inp_count
    allocate(Proc_Instructions%Input(inp_count),STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       call logmsg(log_error,"ERROR allocation problem in "//&
            "Get_Proc_Instruc_from_xml_tree()")
       error_flag = error_allocate
       return
    END IF

    ! Initialise and read the input blocks
    DO i=1,inp_count
       call Init_Input(Proc_Instructions%Input(i))
       IF (error_flag .ne. no_error) return

       write(count_string,"(i3)",err=998) i

       path = trim(xml_path)//"/List_of_Inputs/Input["//&
              trim(adjustl(count_string))//"]"
       call Get_Input_from_xml_tree(xml_tree, path, &
                         Proc_Instructions%Input(i), error_flag, workaround)
       IF (error_flag .ne. no_error) return
    END DO

    ! determine the number of Output blocks
    path = trim(xml_path)//"/List_of_Outputs"
    attr_name = "count"
    call xml_tree_read_string_attribute(xml_tree,path,attr_name,&
                                        attr_value,error_flag)
    IF (error_flag .ne. no_error) return
    read(attr_value,"(i3)",iostat=ios) outp_count
    IF (ios .ne. 0) goto 999

    ! allocate space to store the outputs
    !print *,"nr of output blocks: ",outp_count
    allocate(Proc_Instructions%Output(outp_count),STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       call logmsg(log_error,"ERROR allocation problem in "//&
            "Get_Proc_Instruc_from_xml_tree()")
       error_flag = error_allocate
       return
    END IF

    ! Initialise and read the output blocks
    DO i=1,outp_count
       call Init_Output(Proc_Instructions%Output(i))
       IF (error_flag .ne. no_error) return

       write(count_string,"(i3)",err=998) i

       path = trim(xml_path)//"/List_of_Outputs/Output["//&
              trim(adjustl(count_string))//"]"
       call Get_Output_from_xml_tree(xml_tree, path, &
                         Proc_Instructions%Output(i), error_flag, workaround)
       IF (error_flag .ne. no_error) return
    END DO

    return

998 call logmsg(log_error,"ERROR conversion error in "//&
         "Get_Proc_Instruc_from_xml_tree: could not convert the"//&
         " counter i to a string ...")
    error_flag = error_writing_file
    return

999 call logmsg(log_error,"ERROR in Get_Proc_Instruc_from_xml_tree():"//&
         " could not convert count attribute of the "//trim(xml_path)//&
         " element into an integer number")
    error_flag = error_reading_file
    return

  end subroutine Get_Proc_Instruc_from_xml_tree
    !  #]
  subroutine Check_Proc_Instructions(Proc_Instructions, error_flag)
    !  #[

    ! => do a sanity check on the input values 
    TYPE(Proc_Instructions_type), &
                      intent(in)  :: Proc_Instructions
    integer,          intent(out) :: error_flag

    ! local variables
    logical :: problem_found
    integer :: i,s

    problem_found = .false.
    error_flag = no_error

    ! check the Task_Name
    ! I assume this must be the name of the executable
    ! of the L2BP main program, but I'm not completely sure about this
    IF (trim(Proc_Instructions%Task_Name) .ne. software_TaskName) THEN
       call logmsg(log_error,"ERROR in Check_Proc_Instructions():"//&
            " this JobOrder requests use of a processor executable Task"//&
            " named: ["//trim(Proc_Instructions%Task_Name)//&
            "] but the current processor executable Task name is defined "//&
            "as: ["//software_TaskName//"]")
       problem_found = .true.
    END IF

    ! check the Task_Version
    IF (trim(Proc_Instructions%Task_Version) .ne. software_version) THEN
       call logmsg(log_error,"ERROR in Check_Proc_Instructions():"//&
            " this JobOrder requests use of processor version: ["//&
            trim(Proc_Instructions%Task_Version)//&
            "] but the current processor has version: ["//&
            software_Version//"]")
       problem_found = .true.
    END IF

    ! check for the correct format of the version string: XX.XX
    call Check_JobOrder_Version(Proc_Instructions%Task_Version,&
                                "Proc_Instructions%Task_Version",&
                                problem_found)

! this element is not present in the JobOrder files generated by
! the ThinLayer emulator....
!    ! check the Breakpoint_Enable item
!    ! should be either true or false
!    select case(to_lowercase(trim(Proc_Instructions%Breakpoint_Enable)))
!    case("false")
!       ! ok, this is a known boolean type
!    case("true")
!       ! this is a valid boolean but this feature is not implemented
!       call logmsg(log_error,"ERROR in Check_Proc_Instructions():"//&
!            " this JobOrder has a value for the boolean Breakpoint_Enable"//&
!            " field of: ["//trim(Proc_Instructions%Breakpoint_Enable)//&
!            "] but the use of breakpoints"//&
!            " is not implemented in the L2B-Processor")
!    case default
!       call logmsg(log_error,"ERROR in Check_Proc_Instructions():"//&
!            " this JobOrder has a value for the boolean Breakpoint_Enable"//&
!            " field of: ["//trim(Proc_Instructions%Breakpoint_Enable)//&
!            "] but the allowed types are: [true|false]")
!       problem_found = .true.
!    end select

    IF (associated(Proc_Instructions%Input)) THEN
       s=size(Proc_Instructions%Input)
       DO i=1,s
          call Check_Input(Proc_Instructions%Input(i),problem_found)
       END DO
    END IF

    IF (associated(Proc_Instructions%Output)) THEN
       s=size(Proc_Instructions%Output)
       DO i=1,s
          call Check_Output(Proc_Instructions%Output(i),problem_found)
       END DO
    END IF

    IF (problem_found) error_flag = error_reading_file

  end subroutine Check_Proc_Instructions
    !  #]
  !-------------------------------------------------
  ! Global_Configuration handling
  !-------------------------------------------------
  subroutine Init_Special_Settings(Special_Settings)
    !  #[
    TYPE(Special_Settings_type), intent(out) :: Special_Settings

    Special_Settings%File_Name(:) = ' '
    
  end subroutine Init_Special_Settings
    !  #]
  subroutine AsciiWrite_Special_Settings(Special_Settings,&
                                         fileunit,prefix,error_flag)
    !  #[
    TYPE(Special_Settings_type), &
                      intent(in)  :: Special_Settings
    integer,          intent(in)  :: fileunit
    character(len=*), intent(in)  :: prefix
    integer,          intent(out) :: error_flag

    ! local variable
    character(len=256) :: txt

    error_flag = no_error

    txt = Special_Settings%File_Name
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) trim(prefix)//"%File_Name = "//trim(txt)

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_Special_Settings:"//&
         " error while writing to outputfile")
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Special_Settings
    !  #]
  subroutine Get_Spec_Settings_from_xml_tree(xml_tree, xml_path, &
                                       Special_Settings, error_flag)
    !  #[
    TYPE(xml_node_type), pointer     :: xml_tree ! input
    character(len=*),    intent(in)  :: xml_path ! input
    TYPE(Special_Settings_type), &
                         intent(out) :: Special_Settings
    integer,             intent(out) :: error_flag

    ! local variables
    character(len=max_str_len)  :: text
    character(len=max_str_len)  :: path

    ! init
    error_flag = no_error

    ! copying xml items
    path = trim(xml_path)//"/File_Name"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,&
                      maxlen_file_name,"Get_Spec_Settings_from_xml_tree")
    Special_Settings%File_Name = text

  end subroutine Get_Spec_Settings_from_xml_tree
    !  #]    
  !subroutine Check_Special_Settings(Special_Settings, error_flag)
    !  #[ not  implemented
  !end subroutine Check_Special_Settings
    !  #]
  !-------------------------------------------------
  ! JobOrderData handling
  !-------------------------------------------------
  subroutine Init_JobOrderData(JobOrderData)
    !  #[
    TYPE(JobOrderData_type), intent(out) :: JobOrderData

    call Init_Global_Configuration(JobOrderData%Global_Configuration)
    call Init_Proc_Instructions(JobOrderData%Proc_Instructions)
    call Init_Special_Settings(JobOrderData%Special_Settings)

  end subroutine Init_JobOrderData
    !  #]
  subroutine Delete_JobOrderData(JobOrderData)
    !  #[
    TYPE(JobOrderData_type), intent(inout) :: JobOrderData

    ! JobOrderData%Global_Configuration
    ! contains no allocated items, so needs no deallocation

    call Delete_Proc_Instructions(JobOrderData%Proc_Instructions)

    ! JobOrderData%Special_Settings
    ! contains no allocated items, so needs no deallocation

  end subroutine Delete_JobOrderData
    !  #]
  subroutine Print_JobOrderData(JobOrderData,Task_Number,prefix,error_flag)
    !  #[
    TYPE(JobOrderData_type), intent(in)  :: JobOrderData
    integer,                 intent(in)  :: Task_Number
    character(len=*),        intent(in)  :: prefix
    integer,                 intent(out) :: error_flag

    call AsciiWrite_JobOrderData(JobOrderData,fileunit_stdout,&
                                 Task_Number, prefix,error_flag)

  end subroutine Print_JobOrderData
    !  #]
  subroutine AsciiWrite_JobOrderData(JobOrderData,fileunit,&
                                     Task_Number, prefix,error_flag)
    !  #[
    TYPE(JobOrderData_type), intent(in)  :: JobOrderData
    integer,                 intent(in)  :: fileunit
    integer,                 intent(in)  :: Task_Number
    character(len=*),        intent(in)  :: prefix
    integer,                 intent(out) :: error_flag

    ! local variables
    character(len=3)             :: count_string

    error_flag = no_error

    call AsciiWrite_Global_Configuration(JobOrderData%Global_Configuration,&
         fileunit,trim(prefix)//"%Global_Configuration",error_flag)
    IF (error_flag .ne. no_error) return

    write(count_string,"(i3)",err=997) Task_Number

    call AsciiWrite_Proc_Instructions(JobOrderData%Proc_Instructions,&
         fileunit,trim(prefix)//"%Proc_Instructions("//&
         trim(adjustl(count_string))//")",error_flag)
    IF (error_flag .ne. no_error) return

    call AsciiWrite_Special_Settings(JobOrderData%Special_Settings,&
         fileunit,trim(prefix)//"%Special_Settings",error_flag)

    return

997 call logmsg(log_error,"ERROR conversion error in"//&
         " AsciiWrite_JobOrderData:"//&
         " could not convert the counter Task_Number to a string ...")
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_JobOrderData
    !  #]
  subroutine Get_JobOrderData_from_xml(xml_tree, &
                                       JobOrderData, Task_Number, error_flag,&
                                       workaround)
    !  #[
    TYPE(xml_node_type),     pointer     :: xml_tree ! input
    TYPE(JobOrderData_type), intent(out) :: JobOrderData
    integer,                 intent(out) :: Task_Number
    integer,                 intent(out) :: error_flag
    logical, optional,       intent(in)  :: workaround

    ! local variables
    character(len=max_str_len) :: xml_path, path, text
    character(len=max_str_len) :: attr_name, attr_value
    character(len=3)           :: count_string
    integer                    :: proc_count, ios, i

    ! init
    error_flag = no_error
    
    ! copying xml items

    ! copy and check the global_configuration settings
    xml_path = "/Ipf_Job_Order/Ipf_Conf"
    call Get_Global_Config_from_xml_tree(xml_tree, xml_path, &
                         JobOrderData%Global_Configuration, error_flag)
    IF (error_flag .ne. no_error) return

    call Check_Global_Configuration(JobOrderData%Global_Configuration, &
                                    error_flag)
    IF (error_flag .ne. no_error) return

    ! determine which block to read from this list:
    xml_path = "/Ipf_Job_Order/List_of_Ipf_Procs"
    attr_name = "count"
    call xml_tree_read_string_attribute(xml_tree,xml_path,attr_name,&
                                        attr_value,error_flag)
    IF (error_flag .ne. no_error) return
    read(attr_value,"(i3)",iostat=ios) proc_count
    IF (ios .ne. 0) goto 999

    Task_Number = -1
    print *,"test: this JobOrder file calls proc_count = ", proc_count,&
         " programs (or Tasks)"
    DO i=1,proc_count
       write(count_string,"(i3)",err=998) i
       path = trim(xml_path)//"/Ipf_Proc["//trim(adjustl(count_string))//&
            "]/Task_Name"
       call xml_node_get_string(xml_tree,path,text,error_flag)
       IF (error_flag .ne. no_error) return
       text = check_length_xml_item(text,path,&
                              maxlen_task_name,"Get_JobOrderData_from_xml")

       !print *,"path = ",trim(path)
       print *,"Task_Name = ["//trim(text)//"]"
       IF (trim(text) .eq. software_TaskName) THEN
          IF (Task_Number .eq. -1) THEN
             Task_Number = i
          ELSE
             call logmsg(log_error,"ERROR in Get_JobOrderData_from_xml():"//&
                  " multiple items in the "//trim(xml_path)//&
                  " list have a Task_Name matching the current processor"//&
                  " Task name, which is: ["//trim(software_TaskName)//"]."//&
                  " Don't know how to choose between them...")
             error_flag = error_reading_file
             return
          END IF
       END IF
    END DO

    IF (Task_Number .eq. -1) THEN
       call logmsg(log_error,"ERROR in Get_JobOrderData_from_xml():"//&
            " could not find an item in the "//trim(xml_path)//&
            " list that has a Task_Name matching the current processor"//&
            " Task name, which is: ["//trim(software_TaskName)//"]")
       error_flag = error_reading_file
       return
    ELSE
       print *,"loading settings for task nr. ",Task_Number,&
            " named: ",software_TaskName
    END IF

    write(count_string,"(i3)",err=997) Task_Number
    xml_path = "/Ipf_Job_Order/List_of_Ipf_Procs/Ipf_Proc["//&
                 trim(adjustl(count_string))//"]"

    call Get_Proc_Instruc_from_xml_tree(xml_tree, xml_path, &
                         JobOrderData%Proc_Instructions, error_flag,&
                         workaround)
    IF (error_flag .ne. no_error) return

    call Check_Proc_Instructions(JobOrderData%Proc_Instructions, &
                                 error_flag)
    IF (error_flag .ne. no_error) return

    xml_path = "/Ipf_Job_Order/Processor_Conf"
    call Get_Spec_Settings_from_xml_tree(xml_tree, xml_path, &
                          JobOrderData%Special_Settings, error_flag)
    IF (error_flag .ne. no_error) return

    ! not yet implemented
    !call Check_Special_Settings(Special_Settings, error_flag)
    
    return

997 call logmsg(log_error,"ERROR conversion error in "//&
         "Get_JobOrderData_from_xml: "//&
         "could not convert the counter Task_Number to a string ...")
    error_flag = error_writing_file
    return

998 call logmsg(log_error,"ERROR conversion error in "//&
         "Get_JobOrderData_from_xml: "//&
         "could not convert the counter i to a string ...")
    error_flag = error_writing_file
    return

999 call logmsg(log_error,"ERROR in Get_JobOrderData_from_xmle():"//&
         " could not convert count attribute of the "//trim(xml_path)//&
         " element into an integer number")
    error_flag = error_reading_file
    return

  end subroutine Get_JobOrderData_from_xml
    !  #]
  !-------------------------------------------------
  ! some general helper routines
  !-------------------------------------------------
  subroutine Check_JobOrder_DateTime(DateTimeString,ItemName,problem_found)
    !  #[
    character(len=maxlen_jobdatetime), intent(in)    :: DateTimeString
    character(len=*),                  intent(in)    :: ItemName
    logical,                           intent(inout) :: problem_found

    ! local variables
    logical          :: CorrectFormat
    character(len=4) :: Year
    character(len=2) :: Month, Day, Hour, Minute, Second
    character(len=6) :: MicroSecond

    TYPE(date_type)  :: date ! as defined in DateTimeMod
    TYPE(time_type)  :: time ! as defined in DateTimeMod
    integer          :: temp_seconds, temp_micro_seconds, error_flag, ios

    Year        = DateTimeString(1:4)
    Month       = DateTimeString(5:6)
    Day         = DateTimeString(7:8)
    Hour        = DateTimeString(10:11)
    Minute      = DateTimeString(12:13)
    Second      = DateTimeString(14:15)
    MicroSecond = DateTimeString(16:21)

    ! check for the correct format of this JobOrder DateTime: YYYYMMDD_HHMMSS
    ! note that this format differs from the usual xml and kvt formats !
    CorrectFormat = .true.
    IF (DateTimeString(9:9) .ne. '_') CorrectFormat = .false.

    ! check if the numbers are really numbers
    IF (verify(Year,       "0123456789") .ne. 0) CorrectFormat = .false.
    IF (verify(Month,      "0123456789") .ne. 0) CorrectFormat = .false.
    IF (verify(Day,        "0123456789") .ne. 0) CorrectFormat = .false.
    IF (verify(Hour,       "0123456789") .ne. 0) CorrectFormat = .false.
    IF (verify(Minute,     "0123456789") .ne. 0) CorrectFormat = .false.
    IF (verify(Second,     "0123456789") .ne. 0) CorrectFormat = .false.
    IF (verify(MicroSecond,"0123456789") .ne. 0) CorrectFormat = .false.

    ! give a proper error if something is wrong
    ! (might indicate a damaged file, or a programming/configuration error)
    IF (.not. CorrectFormat) THEN
       call logmsg(log_error,"ERROR in Check_JobOrder_DateTime():"//&
            " this JobOrder requests use of "//trim(ItemName)//": ["//&
            trim(DateTimeString)//"] which does not comply to the required"//&
            " format of [YYYYYYMMDD_HHMMSSSSSSSS]")
       problem_found = .true.
    END IF

    call init_date(date)
    call init_time(time)

    ! interpret the numbers to see if they are sensible
    read(Day,   "(i2)",iostat=ios) date%day
    IF (ios .ne. 0) goto 999
    read(Month, "(i2)",iostat=ios) date%month
    IF (ios .ne. 0) goto 999
    read(Year,  "(i4)",iostat=ios) date%year
    IF (ios .ne. 0) goto 999
    read(Hour,  "(i2)",iostat=ios) time%hour
    IF (ios .ne. 0) goto 999
    read(Minute,"(i2)",iostat=ios) time%minute
    IF (ios .ne. 0) goto 999
    read(Second,"(i2)",iostat=ios) temp_seconds
    IF (ios .ne. 0) goto 999
    read(MicroSecond,"(i6)",iostat=ios) temp_micro_seconds
    IF (ios .ne. 0) goto 999
    time%second = 1._r8_*temp_seconds + 1.e-6_r8_*temp_micro_seconds

    error_flag = no_error
    call test_date(date,error_flag)
    IF (error_flag .ne. no_error) goto 999

    call test_time(time,error_flag)
    IF (error_flag .ne. no_error) goto 999

    return

    ! catch errors in the read() conversion statement
999 continue
    call logmsg(log_error,"ERROR in Check_JobOrder_DateTime():"//&
         " this JobOrder requests use of "//trim(ItemName)//": ["//&
         trim(DateTimeString)//"] but this string forms no valid date/time")
    problem_found = .true.
    return

  end subroutine check_JobOrder_DateTime
    !  #]
  subroutine Check_JobOrder_Version(VersionString,ItemName,problem_found)
    !  #[
    character(len=maxlen_version), intent(in)    :: VersionString
    character(len=*),              intent(in)    :: ItemName
    logical,                       intent(inout) :: problem_found

    ! local variables
    logical :: CorrectFormat

    ! check for the correct format of the version string: XX.XX
    ! with X a digit between 0-9
    CorrectFormat = .true.
    IF (VersionString(3:3) .ne. '.') CorrectFormat = .false.
    ! check if the numbers are really numbers
    IF (verify(VersionString(1:2),"0123456789") .ne. 0) CorrectFormat = .false.
    IF (verify(VersionString(4:5),"0123456789") .ne. 0) CorrectFormat = .false.

    IF (.not. CorrectFormat) THEN
       call logmsg(log_error,"ERROR in Check_JobOrder_Version():"//&
            " this JobOrder requests use of "//trim(ItemName)//": ["//&
            trim(VersionString)//&
            "] which does not comply to the required format of [XX.XX]")
       problem_found = .true.
    END IF

  end subroutine Check_JobOrder_Version
    !  #]
  !-------------------------------------------------
end module JobOrder_DataStructure
