module L1B_Cal_ADS
  !  #[ Documentation
  !
  !     a module to handle the Calibration Annotated dataset
  !
  !     See DLR IODD AE.TN.DLR.IODD-L1B.231105
  !
  !     Written by P. Poli, Meteo France
  !     created:     17-02-2006
  !
  !     Modifications:
  !      28-Mar-2006 P. Poli    Upgrade to DLR IODD AE.TN.DLR.APM-L1B.150206
  !      12-May-2006 D. Tan     Null use of ParKind1 module
  !      03-Nov-2006 P. Poli    upgrade to L1BP v1.04 ;
  !                             moved and cleaned-up reading part
  !      16-Jan-2008 J. de Kloe phase out integer kind i_ 
  !
  !  #]
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE Numerics, only: i8_, real2txt, missing_int
  USE DataSetDescriptor, only: DSD_type
  USE EE_CFI_Datatypes, only: DateTimeType,Construct_KVT_DateTime, InitDateTime,  &
                              read_DateTime_from_bin_file, &
                              read_IntAl_from_bin_file, missing_indicator_IntAl, IntAl, &
                              read_FAdoxy_from_bin_file, missing_indicator_FAdoxy, FAdoxy, &
                              int2txt, maxlen_DateTime_KVT
  USE DirectBinaryIO, only: SetFilePointer, AdvanceFilePointer, read_logical1_from_bin_file
  USE ErrorHandler, only: no_error, error_programming, error_allocate, error_reading_file
  !  #]
  !  #[ type definitions

  implicit none

  TYPE Ref_Pulse_Calib_Valid_type
     logical :: Mean_Sensitivity_Valid
     logical :: Error_Resp_Stdev_Valid
     logical :: Zero_Freq_Resp_Valid
     logical :: Data_Monotonic
  END TYPE Ref_Pulse_Calib_Valid_type

  TYPE Meas_Calib_Valid_type
     logical :: Mean_Sensitivity_Valid
     logical :: Error_Resp_Stdev_Valid
     logical :: Zero_Freq_Resp_Valid
     logical :: Data_Monotonic
  END TYPE Meas_Calib_Valid_type

  TYPE Rayl_Ref_Pulse_Resp_Calib_type
     real(FAdoxy)    :: Ref_Pulse_Mean_Sensitivity
     real(FAdoxy)    :: Ref_Pulse_Zero_Freq
     real(FAdoxy)    :: Ref_Pulse_Err_Rayl_Resp_Stdev
  END TYPE Rayl_Ref_Pulse_Resp_Calib_type

  TYPE Mie_Ref_Pulse_Resp_Calib_type
     real(FAdoxy)    :: Ref_Pulse_Mean_Sensitivity
     real(FAdoxy)    :: Ref_Pulse_Zero_Freq
     real(FAdoxy)    :: Ref_Pulse_Err_Mie_Resp_Stdev
  END TYPE Mie_Ref_Pulse_Resp_Calib_type

  TYPE Rayl_Meas_Resp_Calib_type
     real(FAdoxy)    :: Meas_Mean_Sensitivity
     real(FAdoxy)    :: Meas_Zero_Freq
     real(FAdoxy)    :: Meas_Err_Rayl_Resp_Stdev
  END TYPE Rayl_Meas_Resp_Calib_type

  TYPE Mie_Meas_Resp_Calib_type
     real(FAdoxy)    :: Meas_Mean_Sensitivity
     real(FAdoxy)    :: Meas_Zero_Freq
     real(FAdoxy)    :: Meas_Err_Mie_Resp_Stdev
  END TYPE Mie_Meas_Resp_Calib_type

  TYPE Freq_Step_Data_Stat_type
     integer(IntAl) :: Num_Valid_Meas
     integer(IntAl) :: Num_Meas_Usable
     integer(IntAl) :: Num_Ref_Pulse_Usable
     integer(IntAl) :: Num_Meas_Freq_Unlock
     integer(IntAl) :: Num_Ref_Pulse_Freq_Unlock
     integer(IntAl) :: Num_Sat_Not_On_Target_Meas
     integer(IntAl) :: Num_Corrupt_Meas
     integer(IntAl) :: Num_Corrupt_Ref_Pulse
  END TYPE Freq_Step_Data_Stat_type

  TYPE Rayl_Freq_Step_Res_type
     real(FAdoxy)    :: Freq_Offset
     logical      :: Freq_Valid
     logical      :: Meas_Resp_Valid
     logical      :: Ref_Pulse_Resp_Valid
     real(FAdoxy)    :: Meas_Resp
     real(FAdoxy)    :: Meas_Err_Rayl_Resp
     real(FAdoxy)    :: Ref_Pulse_Resp
     real(FAdoxy)    :: Ref_Pulse_Err_Rayl_Resp
     type(Freq_Step_Data_Stat_type) :: Freq_Step_Data_Stat
  END TYPE Rayl_Freq_Step_Res_type

  TYPE Mie_Freq_Step_Res_type
     real(FAdoxy)    :: Freq_Offset
     logical      :: Freq_Valid
     logical      :: Meas_Resp_Valid
     logical      :: Ref_Pulse_Resp_Valid
     real(FAdoxy)    :: Meas_Resp
     real(FAdoxy)    :: Meas_Err_Mie_Resp
     real(FAdoxy)    :: Ref_Pulse_Resp
     real(FAdoxy)    :: Ref_Pulse_Err_Mie_Resp
     type(Freq_Step_Data_Stat_type) :: Freq_Step_Data_Stat
  END TYPE Mie_Freq_Step_Res_type

  TYPE List_Rayl_Freq_Step_Res_type
     type(Rayl_Freq_Step_Res_type) :: Rayl_Freq_Step_Res
  END TYPE List_Rayl_Freq_Step_Res_type

  TYPE List_Mie_Freq_Step_Res_type
     type(Mie_Freq_Step_Res_type) :: Mie_Freq_Step_Res
  END TYPE List_Mie_Freq_Step_Res_type

  integer, parameter :: numMRCFreqSteps = 40

  TYPE Rayl_Resp_Calib_Data_type
     logical      :: Calib_Valid
     !in L1B v1.04 the following array decreased in size for 60 to 40 elements
     type(List_Rayl_Freq_Step_Res_type), dimension(1:numMRCFreqSteps) :: List_Rayl_Freq_Step_Res
     type(Rayl_Meas_Resp_Calib_type) :: Rayl_Meas_Resp_Calib
     type(Rayl_Ref_Pulse_Resp_Calib_type) :: Rayl_Ref_Pulse_Resp_Calib
     type(Meas_Calib_Valid_type) :: Meas_Calib_Valid
     type(Ref_Pulse_Calib_Valid_type) :: Ref_Pulse_Calib_Valid
  END TYPE Rayl_Resp_Calib_Data_type

  TYPE Mie_Resp_Calib_Data_type
     logical      :: Calib_Valid
     !in L1B v1.04 the following array decreased in size for 60 to 40 elements
     type(List_Mie_Freq_Step_Res_type), dimension(1:numMRCFreqSteps) :: List_Mie_Freq_Step_Res
     type(Mie_Meas_Resp_Calib_type) :: Mie_Meas_Resp_Calib
     type(Mie_Ref_Pulse_Resp_Calib_type) :: Mie_Ref_Pulse_Resp_Calib
     type(Meas_Calib_Valid_type) :: Meas_Calib_Valid
     type(Ref_Pulse_Calib_Valid_type) :: Ref_Pulse_Calib_Valid
  END TYPE Mie_Resp_Calib_Data_type

  TYPE List_Tripod_Obscur_Corr_type
     real(FAdoxy) :: Tripod_Obscur_Corr
  END TYPE List_Tripod_Obscur_Corr_type

  TYPE Err_Quant_type
     real(FAdoxy) :: Mie_Err_Quant_K1
     real(FAdoxy) :: Mie_Err_Quant_K2
     real(FAdoxy) :: Mie_Err_Quant_K3
     real(FAdoxy) :: Rayl_Err_Quant_Ka2
     real(FAdoxy) :: Rayl_Err_Quant_Ka3
     real(FAdoxy) :: Rayl_Err_Quant_Kb2
     real(FAdoxy) :: Rayl_Err_Quant_Kb3
  END TYPE Err_Quant_type

  TYPE Sat_Char_Data_type
     real(FAdoxy) :: Laser_Wavelength
     type(Err_Quant_type) :: Err_Quant
     type(List_Tripod_Obscur_Corr_type), dimension(1:16) :: List_Tripod_Obscur_Corr
     integer(IntAl) :: NF_Order
     real(FAdoxy), dimension(1:1024) :: Mie_harmonic_bias_coeff_A
     real(FAdoxy), dimension(1:1024) :: Mie_harmonic_bias_coeff_B
     real(FAdoxy), dimension(1:1024) :: Rayl_harmonic_bias_coeff_A
     real(FAdoxy), dimension(1:1024) :: Rayl_harmonic_bias_coeff_B
  END TYPE Sat_Char_Data_type

  TYPE L1B_Char_Data_type
     type(Sat_Char_Data_type) :: Sat_Char_Data
     type(Mie_Resp_Calib_Data_type) :: Mie_Resp_Calib_Data
     type(Rayl_Resp_Calib_Data_type) :: Rayl_Resp_Calib_Data
  END TYPE L1B_Char_Data_type

  TYPE L1B_Cal_ADS_type
     type(DateTimeType) :: MRC_First_Start_of_Obs_DateTime
     type(DateTimeType) :: MRC_Last_Start_of_Obs_DateTime
     type(DateTimeType) :: RRC_First_Start_of_Obs_DateTime
     type(DateTimeType) :: RRC_Last_Start_of_Obs_DateTime
     type(L1B_Char_Data_type) :: L1B_Char_Data
  END TYPE L1B_Cal_ADS_type

  !  #]
contains
  !---------------------------------------------------------------------
  subroutine Init_Ref_Pulse_Calib_Valid(Ref_Pulse_Calib_Valid)
    !  #[
    TYPE(Ref_Pulse_Calib_Valid_type), intent(out) :: Ref_Pulse_Calib_Valid

     Ref_Pulse_Calib_Valid%Mean_Sensitivity_Valid = .FALSE.
     Ref_Pulse_Calib_Valid%Error_Resp_Stdev_Valid = .FALSE.
     Ref_Pulse_Calib_Valid%Zero_Freq_Resp_Valid = .FALSE.
     Ref_Pulse_Calib_Valid%Data_Monotonic = .FALSE.
  end subroutine Init_Ref_Pulse_Calib_Valid
    !  #]
  subroutine Print_Ref_Pulse_Calib_Valid(Ref_Pulse_Calib_Valid)
    !  #[
    TYPE(Ref_Pulse_Calib_Valid_type), intent(in) :: Ref_Pulse_Calib_Valid

    print *,"            ...%Mean_Sensitivity_Valid = ",Ref_Pulse_Calib_Valid%Mean_Sensitivity_Valid
    print *,"            ...%Error_Resp_Stdev_Valid = ",Ref_Pulse_Calib_Valid%Error_Resp_Stdev_Valid
    print *,"            ...%Zero_Freq_Resp_Valid   = ",Ref_Pulse_Calib_Valid%Zero_Freq_Resp_Valid
    print *,"            ...%Data_Monotonic         = ",Ref_Pulse_Calib_Valid%Data_Monotonic
  end subroutine Print_Ref_Pulse_Calib_Valid
    !  #]
  subroutine Read_Ref_Pulse_Calib_Valid(Ref_Pulse_Calib_Valid, fileunit, error_flag)
    !  #[
    TYPE(Ref_Pulse_Calib_Valid_type), intent(out) :: Ref_Pulse_Calib_Valid
    integer,                          intent(in)  :: fileunit
    integer,                          intent(out) :: error_flag

    logical :: value_logical

    error_flag = no_error

    ! Mean_Sensitivity_Valid
    call read_logical1_from_bin_file(fileunit, value_logical, error_flag)
    IF (error_flag .ne. no_error) return
    Ref_Pulse_Calib_Valid%Mean_Sensitivity_Valid = value_logical

    ! Error_Response_Std_Dev_Valid
    call read_logical1_from_bin_file(fileunit, value_logical, error_flag)
    IF (error_flag .ne. no_error) return
    Ref_Pulse_Calib_Valid%Error_Resp_Stdev_Valid = value_logical

    ! Zero_Freq_Response_Valid
    call read_logical1_from_bin_file(fileunit, value_logical, error_flag)
    IF (error_flag .ne. no_error) return
    Ref_Pulse_Calib_Valid%Zero_Freq_Resp_Valid = value_logical

    ! Data_Monotonic
    call read_logical1_from_bin_file(fileunit, value_logical, error_flag)
    IF (error_flag .ne. no_error) return
    Ref_Pulse_Calib_Valid%Data_Monotonic = value_logical

  end subroutine Read_Ref_Pulse_Calib_Valid
    !  #]
  !---------------------------------
  subroutine Init_Meas_Calib_Valid(Meas_Calib_Valid)
    !  #[
    TYPE(Meas_Calib_Valid_type), intent(out) :: Meas_Calib_Valid

     Meas_Calib_Valid%Mean_Sensitivity_Valid = .FALSE.
     Meas_Calib_Valid%Error_Resp_Stdev_Valid = .FALSE.
     Meas_Calib_Valid%Zero_Freq_Resp_Valid = .FALSE.
     Meas_Calib_Valid%Data_Monotonic = .FALSE.
  end subroutine Init_Meas_Calib_Valid
    !  #]
  subroutine Print_Meas_Calib_Valid(Meas_Calib_Valid)
    !  #[
    TYPE(Meas_Calib_Valid_type), intent(in) :: Meas_Calib_Valid

    print *,"            ...%Mean_Sensitivity_Valid = ",Meas_Calib_Valid%Mean_Sensitivity_Valid
    print *,"            ...%Error_Resp_Stdev_Valid = ",Meas_Calib_Valid%Error_Resp_Stdev_Valid
    print *,"            ...%Zero_Freq_Resp_Valid   = ",Meas_Calib_Valid%Zero_Freq_Resp_Valid
    print *,"            ...%Data_Monotonic         = ",Meas_Calib_Valid%Data_Monotonic
  end subroutine Print_Meas_Calib_Valid
    !  #]
  subroutine Read_Meas_Calib_Valid(Meas_Calib_Valid, fileunit, error_flag)
    !  #[
    TYPE(Meas_Calib_Valid_type), intent(out) :: Meas_Calib_Valid
    integer,                     intent(in)  :: fileunit
    integer,                     intent(out) :: error_flag

    logical :: value_logical

    error_flag = no_error

    ! Mean_Sensitivity_Valid
    call read_logical1_from_bin_file(fileunit, value_logical, error_flag)
    IF (error_flag .ne. no_error) return
    Meas_Calib_Valid%Mean_Sensitivity_Valid = value_logical

    ! Error_Response_Std_Dev_Valid
    call read_logical1_from_bin_file(fileunit, value_logical, error_flag)
    IF (error_flag .ne. no_error) return
    Meas_Calib_Valid%Error_Resp_Stdev_Valid = value_logical

    ! Zero_Freq_Response_Valid
    call read_logical1_from_bin_file(fileunit, value_logical, error_flag)
    IF (error_flag .ne. no_error) return
    Meas_Calib_Valid%Zero_Freq_Resp_Valid = value_logical

    ! Data_Monotonic
    call read_logical1_from_bin_file(fileunit, value_logical, error_flag)
    IF (error_flag .ne. no_error) return
    Meas_Calib_Valid%Data_Monotonic = value_logical

  end subroutine Read_Meas_Calib_Valid
    !  #]
  !---------------------------------
  subroutine Init_Rayl_Ref_Pulse_Resp_Calib(Rayl_Ref_Pulse_Resp_Calib)
    !  #[
    TYPE(Rayl_Ref_Pulse_Resp_Calib_type), intent(out) :: Rayl_Ref_Pulse_Resp_Calib

    Rayl_Ref_Pulse_Resp_Calib%Ref_Pulse_Mean_Sensitivity    = missing_indicator_FAdoxy
    Rayl_Ref_Pulse_Resp_Calib%Ref_Pulse_Zero_Freq           = missing_indicator_FAdoxy
    Rayl_Ref_Pulse_Resp_Calib%Ref_Pulse_Err_Rayl_Resp_Stdev = missing_indicator_FAdoxy
  end subroutine Init_Rayl_Ref_Pulse_Resp_Calib
    !  #]
  subroutine Print_Rayl_Ref_Pulse_Resp_Calib(Rayl_Ref_Pulse_Resp_Calib)
    !  #[
    TYPE(Rayl_Ref_Pulse_Resp_Calib_type), intent(in) :: Rayl_Ref_Pulse_Resp_Calib
    character(len=13) :: txt

    txt = real2txt(Rayl_Ref_Pulse_Resp_Calib%Ref_Pulse_Mean_Sensitivity)
    print *,"            ...%Ref_Pulse_Mean_Sensitivity    = ",txt
    txt = real2txt(Rayl_Ref_Pulse_Resp_Calib%Ref_Pulse_Zero_Freq)
    print *,"            ...%Ref_Pulse_Zero_Freq           = ",txt
    txt = real2txt(Rayl_Ref_Pulse_Resp_Calib%Ref_Pulse_Err_Rayl_Resp_Stdev)
    print *,"            ...%Ref_Pulse_Err_Rayl_Resp_Stdev = ",txt
  end subroutine Print_Rayl_Ref_Pulse_Resp_Calib
    !  #]
  subroutine Init_Mie_Ref_Pulse_Resp_Calib(Mie_Ref_Pulse_Resp_Calib)
    !  #[
    TYPE(Mie_Ref_Pulse_Resp_Calib_type), intent(out) :: Mie_Ref_Pulse_Resp_Calib

    Mie_Ref_Pulse_Resp_Calib%Ref_Pulse_Mean_Sensitivity   = missing_indicator_FAdoxy
    Mie_Ref_Pulse_Resp_Calib%Ref_Pulse_Zero_Freq          = missing_indicator_FAdoxy
    Mie_Ref_Pulse_Resp_Calib%Ref_Pulse_Err_Mie_Resp_Stdev = missing_indicator_FAdoxy
  end subroutine Init_Mie_Ref_Pulse_Resp_Calib
    !  #]
  subroutine Print_Mie_Ref_Pulse_Resp_Calib(Mie_Ref_Pulse_Resp_Calib)
    !  #[
    TYPE(Mie_Ref_Pulse_Resp_Calib_type), intent(in) :: Mie_Ref_Pulse_Resp_Calib
    character(len=13) :: txt

    txt = real2txt(Mie_Ref_Pulse_Resp_Calib%Ref_Pulse_Mean_Sensitivity)
    print *,"            ...%Ref_Pulse_Mean_Sensitivity    = ",txt
    txt = real2txt(Mie_Ref_Pulse_Resp_Calib%Ref_Pulse_Zero_Freq)
    print *,"            ...%Ref_Pulse_Zero_Freq           = ",txt
    txt = real2txt(Mie_Ref_Pulse_Resp_Calib%Ref_Pulse_Err_Mie_Resp_Stdev)
    print *,"            ...%Ref_Pulse_Err_Mie_Resp_Stdev = ",txt
  end subroutine Print_Mie_Ref_Pulse_Resp_Calib
    !  #]
  subroutine Init_Rayl_Meas_Resp_Calib(Rayl_Meas_Resp_Calib)
    !  #[
    TYPE(Rayl_Meas_Resp_Calib_type), intent(out) :: Rayl_Meas_Resp_Calib

    Rayl_Meas_Resp_Calib%Meas_Mean_Sensitivity    = missing_indicator_FAdoxy
    Rayl_Meas_Resp_Calib%Meas_Zero_Freq           = missing_indicator_FAdoxy
    Rayl_Meas_Resp_Calib%Meas_Err_Rayl_Resp_Stdev = missing_indicator_FAdoxy
  end subroutine Init_Rayl_Meas_Resp_Calib
    !  #]
  subroutine Print_Rayl_Meas_Resp_Calib(Rayl_Meas_Resp_Calib)
    !  #[
    TYPE(Rayl_Meas_Resp_Calib_type), intent(in) :: Rayl_Meas_Resp_Calib
    character(len=13) :: txt

    txt = real2txt(Rayl_Meas_Resp_Calib%Meas_Mean_Sensitivity)
    print *,"            ...%Meas_Mean_Sensitivity    = ",txt
    txt = real2txt(Rayl_Meas_Resp_Calib%Meas_Zero_Freq)
    print *,"            ...%Meas_Zero_Freq           = ",txt
    txt = real2txt(Rayl_Meas_Resp_Calib%Meas_Err_Rayl_Resp_Stdev)
    print *,"            ...%Meas_Err_Rayl_Resp_Stdev = ",txt
  end subroutine Print_Rayl_Meas_Resp_Calib
    !  #]
  subroutine Init_Mie_Meas_Resp_Calib(Mie_Meas_Resp_Calib)
    !  #[
    TYPE(Mie_Meas_Resp_Calib_type), intent(out) :: Mie_Meas_Resp_Calib

    Mie_Meas_Resp_Calib%Meas_Mean_Sensitivity   = missing_indicator_FAdoxy
    Mie_Meas_Resp_Calib%Meas_Zero_Freq          = missing_indicator_FAdoxy
    Mie_Meas_Resp_Calib%Meas_Err_Mie_Resp_Stdev = missing_indicator_FAdoxy
  end subroutine Init_Mie_Meas_Resp_Calib
    !  #]
  subroutine Print_Mie_Meas_Resp_Calib(Mie_Meas_Resp_Calib)
    !  #[
    TYPE(Mie_Meas_Resp_Calib_type), intent(in) :: Mie_Meas_Resp_Calib
    character(len=13) :: txt

    txt = real2txt(Mie_Meas_Resp_Calib%Meas_Mean_Sensitivity)
    print *,"            ...%Meas_Mean_Sensitivity    = ",txt
    txt = real2txt(Mie_Meas_Resp_Calib%Meas_Zero_Freq)
    print *,"            ...%Meas_Zero_Freq           = ",txt
    txt = real2txt(Mie_Meas_Resp_Calib%Meas_Err_Mie_Resp_Stdev)
    print *,"            ...%Meas_Err_Mie_Resp_Stdev = ",txt
  end subroutine Print_Mie_Meas_Resp_Calib
    !  #]
  subroutine Init_Freq_Step_Data_Stat(Freq_Step_Data_Stat)
    !  #[
    TYPE(Freq_Step_Data_Stat_type), intent(out) :: Freq_Step_Data_Stat

    Freq_Step_Data_Stat%Num_Valid_Meas             = missing_indicator_IntAl
    Freq_Step_Data_Stat%Num_Meas_Usable            = missing_indicator_IntAl
    Freq_Step_Data_Stat%Num_Ref_Pulse_Usable       = missing_indicator_IntAl
    Freq_Step_Data_Stat%Num_Meas_Freq_Unlock       = missing_indicator_IntAl
    Freq_Step_Data_Stat%Num_Ref_Pulse_Freq_Unlock  = missing_indicator_IntAl
    Freq_Step_Data_Stat%Num_Sat_Not_On_Target_Meas = missing_indicator_IntAl
    Freq_Step_Data_Stat%Num_Corrupt_Meas           = missing_indicator_IntAl
    Freq_Step_Data_Stat%Num_Corrupt_Ref_Pulse      = missing_indicator_IntAl
  end subroutine Init_Freq_Step_Data_Stat
    !  #]
  subroutine Print_Freq_Step_Data_Stat(Freq_Step_Data_Stat)
    !  #[
    TYPE(Freq_Step_Data_Stat_type), intent(in) :: Freq_Step_Data_Stat

    character(len=13) :: txt

    txt=int2txt(Freq_Step_Data_Stat%Num_Valid_Meas)
    print *,"                    ...%Num_Valid_Meas             = ", txt
    txt=int2txt(Freq_Step_Data_Stat%Num_Meas_Usable)
    print *,"                    ...%Num_Meas_Usable            = ", txt
    txt=int2txt(Freq_Step_Data_Stat%Num_Ref_Pulse_Usable)
    print *,"                    ...%Num_Ref_Pulse_Usable       = ", txt
    txt=int2txt(Freq_Step_Data_Stat%Num_Meas_Freq_Unlock)
    print *,"                    ...%Num_Meas_Freq_Unlock       = ", txt
    txt=int2txt(Freq_Step_Data_Stat%Num_Ref_Pulse_Freq_Unlock)
    print *,"                    ...%Num_Ref_Pulse_Freq_Unlock  = ", txt
    txt=int2txt(Freq_Step_Data_Stat%Num_Sat_Not_On_Target_Meas)
    print *,"                    ...%Num_Sat_Not_On_Target_Meas = ", txt
    txt=int2txt(Freq_Step_Data_Stat%Num_Corrupt_Meas)
    print *,"                    ...%Num_Corrupt_Meas           = ", txt
    txt=int2txt(Freq_Step_Data_Stat%Num_Corrupt_Ref_Pulse)
    print *,"                    ...%Num_Corrupt_Ref_Pulse      = ", txt
  end subroutine Print_Freq_Step_Data_Stat
    !  #]  
  subroutine Read_Freq_Step_Data_Stat(Freq_Step_Data_Stat,fileunit,error_flag)
    !  #[
    TYPE(Freq_Step_Data_Stat_type), intent(out) :: Freq_Step_Data_Stat
    integer,                        intent(in)  :: fileunit
    integer,                        intent(out) :: error_flag

    integer(IntAl) :: value_IntAl
    
    error_flag = no_error

    ! Num_Valid_Measurements
    call read_IntAl_from_bin_file(fileunit, value_IntAl, error_flag)
    IF (error_flag .ne. no_error) return
    Freq_Step_Data_Stat%Num_Valid_Meas = value_IntAl

    ! Num_Measurements_Usable
    call read_IntAl_from_bin_file(fileunit, value_IntAl, error_flag)
    IF (error_flag .ne. no_error) return
    Freq_Step_Data_Stat%Num_Meas_Usable = value_IntAl

    ! Num_Reference_Pulses_Usable
    call read_IntAl_from_bin_file(fileunit, value_IntAl, error_flag)
    IF (error_flag .ne. no_error) return
    Freq_Step_Data_Stat%Num_Ref_Pulse_Usable = value_IntAl

    ! Num_Measurement_Laser_Freq_Unlocked
    call read_IntAl_from_bin_file(fileunit, value_IntAl, error_flag)
    IF (error_flag .ne. no_error) return
    Freq_Step_Data_Stat%Num_Meas_Freq_Unlock = value_IntAl

    ! Num_Reference_Pulse_Laser_Freq_Unlocked
    call read_IntAl_from_bin_file(fileunit, value_IntAl, error_flag)
    IF (error_flag .ne. no_error) return
    Freq_Step_Data_Stat%Num_Ref_Pulse_Freq_Unlock = value_IntAl

    ! Num_Sat_Not_on_Target_Measurements
    call read_IntAl_from_bin_file(fileunit, value_IntAl, error_flag)
    IF (error_flag .ne. no_error) return
    Freq_Step_Data_Stat%Num_Sat_Not_On_Target_Meas = value_IntAl

    ! Num_Corrupt_Measurements
    call read_IntAl_from_bin_file(fileunit, value_IntAl, error_flag)
    IF (error_flag .ne. no_error) return
    Freq_Step_Data_Stat%Num_Corrupt_Meas = value_IntAl

    ! Num_Corrupt_Reference_Pulses
    call read_IntAl_from_bin_file(fileunit, value_IntAl, error_flag)
    IF (error_flag .ne. no_error) return
    Freq_Step_Data_Stat%Num_Corrupt_Ref_Pulse = value_IntAl

  end subroutine Read_Freq_Step_Data_Stat
    !  #]
  !---------------------------------
  subroutine Init_Rayl_Freq_Step_Res(Rayl_Freq_Step_Res)
    !  #[
    TYPE(Rayl_Freq_Step_Res_type), intent(out) :: Rayl_Freq_Step_Res

    Rayl_Freq_Step_Res%Freq_Offset             = missing_indicator_FAdoxy
    Rayl_Freq_Step_Res%Freq_Valid              = .FALSE.
    Rayl_Freq_Step_Res%Meas_Resp_Valid         = .FALSE.
    Rayl_Freq_Step_Res%Ref_Pulse_Resp_Valid    = .FALSE.
    Rayl_Freq_Step_Res%Meas_Resp               = missing_indicator_FAdoxy
    Rayl_Freq_Step_Res%Meas_Err_Rayl_Resp      = missing_indicator_FAdoxy
    Rayl_Freq_Step_Res%Ref_Pulse_Resp          = missing_indicator_FAdoxy
    Rayl_Freq_Step_Res%Ref_Pulse_Err_Rayl_Resp = missing_indicator_FAdoxy
    call Init_Freq_Step_Data_Stat(Rayl_Freq_Step_Res%Freq_Step_Data_Stat)
  end subroutine Init_Rayl_Freq_Step_Res
    !  #]
  subroutine Print_Rayl_Freq_Step_Res(Rayl_Freq_Step_Res)
    !  #[
    TYPE(Rayl_Freq_Step_Res_type), intent(in) :: Rayl_Freq_Step_Res
    character(len=13) :: txt

    txt=real2txt(Rayl_Freq_Step_Res%Freq_Offset)
    print *,"                ...%Freq_Offset             = ",txt
    print *,"                ...%Freq_Valid              = ",Rayl_Freq_Step_Res%Freq_Valid
    print *,"                ...%Meas_Resp_Valid         = ",Rayl_Freq_Step_Res%Meas_Resp_Valid
    print *,"                ...%Ref_Pulse_Resp_Valid    = ",Rayl_Freq_Step_Res%Ref_Pulse_Resp_Valid
    txt=real2txt(Rayl_Freq_Step_Res%Meas_Resp)
    print *,"                ...%Meas_Resp               = ",txt
    txt=real2txt(Rayl_Freq_Step_Res%Meas_Err_Rayl_Resp)
    print *,"                ...%Meas_Err_Rayl_Resp      = ",txt
    txt=real2txt(Rayl_Freq_Step_Res%Ref_Pulse_Resp)
    print *,"                ...%Ref_Pulse_Resp          = ",txt
    txt=real2txt(Rayl_Freq_Step_Res%Ref_Pulse_Err_Rayl_Resp)
    print *,"                ...%Ref_Pulse_Err_Rayl_Resp = ",txt
    print *,"                ...%Freq_Step_Data_Stat :"
    call Print_Freq_Step_Data_Stat(Rayl_Freq_Step_Res%Freq_Step_Data_Stat)
  end subroutine Print_Rayl_Freq_Step_Res
    !  #]
  subroutine Init_Mie_Freq_Step_Res(Mie_Freq_Step_Res)
    !  #[
    TYPE(Mie_Freq_Step_Res_type), intent(out) :: Mie_Freq_Step_Res

    Mie_Freq_Step_Res%Freq_Offset             = missing_indicator_FAdoxy
    Mie_Freq_Step_Res%Freq_Valid              = .FALSE.
    Mie_Freq_Step_Res%Meas_Resp_Valid         = .FALSE.
    Mie_Freq_Step_Res%Ref_Pulse_Resp_Valid    = .FALSE.
    Mie_Freq_Step_Res%Meas_Resp               = missing_indicator_FAdoxy
    Mie_Freq_Step_Res%Meas_Err_Mie_Resp       = missing_indicator_FAdoxy
    Mie_Freq_Step_Res%Ref_Pulse_Resp          = missing_indicator_FAdoxy
    Mie_Freq_Step_Res%Ref_Pulse_Err_Mie_Resp  = missing_indicator_FAdoxy
    call Init_Freq_Step_Data_Stat(Mie_Freq_Step_Res%Freq_Step_Data_Stat)
  end subroutine Init_Mie_Freq_Step_Res
    !  #]
  subroutine Print_Mie_Freq_Step_Res(Mie_Freq_Step_Res)
    !  #[
    TYPE(Mie_Freq_Step_Res_type), intent(in) :: Mie_Freq_Step_Res
    character(len=13) :: txt

    txt=real2txt(Mie_Freq_Step_Res%Freq_Offset)
    print *,"                ...%Freq_Offset             = ",txt
    print *,"                ...%Freq_Valid              = ",Mie_Freq_Step_Res%Freq_Valid
    print *,"                ...%Meas_Resp_Valid         = ",Mie_Freq_Step_Res%Meas_Resp_Valid
    print *,"                ...%Ref_Pulse_Resp_Valid    = ",Mie_Freq_Step_Res%Ref_Pulse_Resp_Valid
    txt=real2txt(Mie_Freq_Step_Res%Meas_Resp)
    print *,"                ...%Meas_Resp               = ",txt
    txt=real2txt(Mie_Freq_Step_Res%Meas_Err_Mie_Resp)
    print *,"                ...%Meas_Err_Mie_Resp       = ",txt
    txt=real2txt(Mie_Freq_Step_Res%Ref_Pulse_Resp)
    print *,"                ...%Ref_Pulse_Resp          = ",txt
    txt=real2txt(Mie_Freq_Step_Res%Ref_Pulse_Err_Mie_Resp)
    print *,"                ...%Ref_Pulse_Err_Mie_Resp  = ",txt
    print *,"                ...%Freq_Step_Data_Stat :"
    call Print_Freq_Step_Data_Stat(Mie_Freq_Step_Res%Freq_Step_Data_Stat)
  end subroutine Print_Mie_Freq_Step_Res
    !  #]
  subroutine Init_List_Rayl_Freq_Step_Res(List_Rayl_Freq_Step_Res)
    !  #[
    TYPE(List_Rayl_Freq_Step_Res_type), intent(out) :: List_Rayl_Freq_Step_Res
    
    call Init_Rayl_Freq_Step_Res(List_Rayl_Freq_Step_Res%Rayl_Freq_Step_Res)
  end subroutine Init_List_Rayl_Freq_Step_Res
    !  #]
  subroutine Print_List_Rayl_Freq_Step_Res(List_Rayl_Freq_Step_Res)
    !  #[
    TYPE(List_Rayl_Freq_Step_Res_type), intent(in) :: List_Rayl_Freq_Step_Res

    print *,"            ...%Rayl_Freq_Step_Res: "
    call Print_Rayl_Freq_Step_Res(List_Rayl_Freq_Step_Res%Rayl_Freq_Step_Res)
  end subroutine Print_List_Rayl_Freq_Step_Res
    !  #]
  subroutine Init_List_Mie_Freq_Step_Res(List_Mie_Freq_Step_Res)
    !  #[
    TYPE(List_Mie_Freq_Step_Res_type), intent(out) :: List_Mie_Freq_Step_Res
    
    call Init_Mie_Freq_Step_Res(List_Mie_Freq_Step_Res%Mie_Freq_Step_Res)
  end subroutine Init_List_Mie_Freq_Step_Res
    !  #]
  subroutine Print_List_Mie_Freq_Step_Res(List_Mie_Freq_Step_Res)
    !  #[
    TYPE(List_Mie_Freq_Step_Res_type), intent(in) :: List_Mie_Freq_Step_Res

    print *,"            ...%Mie_Freq_Step_Res: "
    call Print_Mie_Freq_Step_Res(List_Mie_Freq_Step_Res%Mie_Freq_Step_Res)
  end subroutine Print_List_Mie_Freq_Step_Res
    !  #]
  subroutine Init_Rayl_Resp_Calib_Data(Rayl_Resp_Calib_Data)
    !  #[
    TYPE(Rayl_Resp_Calib_Data_type), intent(out) :: Rayl_Resp_Calib_Data

    integer :: i,s

    Rayl_Resp_Calib_Data%Calib_Valid = .FALSE.
    s = size(Rayl_Resp_Calib_Data%List_Rayl_Freq_Step_Res)
    DO i=1,s
      call Init_List_Rayl_Freq_Step_Res(Rayl_Resp_Calib_Data%List_Rayl_Freq_Step_Res(i))
    ENDDO
    call Init_Rayl_Meas_Resp_Calib(Rayl_Resp_Calib_Data%Rayl_Meas_Resp_Calib)
    call Init_Rayl_Ref_Pulse_Resp_Calib(Rayl_Resp_Calib_Data%Rayl_Ref_Pulse_Resp_Calib)
    call Init_Meas_Calib_Valid(Rayl_Resp_Calib_Data%Meas_Calib_Valid)
    call Init_Ref_Pulse_Calib_Valid(Rayl_Resp_Calib_Data%Ref_Pulse_Calib_Valid)
  end subroutine Init_Rayl_Resp_Calib_Data
    !  #]
  subroutine Print_Rayl_Resp_Calib_Data(Rayl_Resp_Calib_Data)
    !  #[
    TYPE(Rayl_Resp_Calib_Data_type), intent(in) :: Rayl_Resp_Calib_Data

    integer :: i,s

    print *,"        ...%Calib_Valid = ",Rayl_Resp_Calib_Data%Calib_Valid
    s = size(Rayl_Resp_Calib_Data%List_Rayl_Freq_Step_Res)
    DO i=1,s
      print *,"        ...%List_Rayl_Freq_Step_Res(",i,") :"
      call Print_List_Rayl_Freq_Step_Res(Rayl_Resp_Calib_Data%List_Rayl_Freq_Step_Res(i))
    ENDDO 
    print *,"        ...%Rayl_Meas_Resp_Calib : "
    call Print_Rayl_Meas_Resp_Calib(Rayl_Resp_Calib_Data%Rayl_Meas_Resp_Calib)
    print *,"        ...%Rayl_Ref_Pulse_Resp_Calib : "
    call Print_Rayl_Ref_Pulse_Resp_Calib(Rayl_Resp_Calib_Data%Rayl_Ref_Pulse_Resp_Calib)
    print *,"        ...%Meas_Calib_Valid : "
    call Print_Meas_Calib_Valid(Rayl_Resp_Calib_Data%Meas_Calib_Valid)
    print *,"        ...%Ref_Pulse_Calib_Valid : "
    call Print_Ref_Pulse_Calib_Valid(Rayl_Resp_Calib_Data%Ref_Pulse_Calib_Valid)
  end subroutine Print_Rayl_Resp_Calib_Data
    !  #]
  subroutine Read_Rayl_Resp_Calib_Data(Rayl_Resp_Calib_Data,fileunit,error_flag)
    !  #[
    TYPE(Rayl_Resp_Calib_Data_type), intent(out) :: Rayl_Resp_Calib_Data ! result
    integer,                         intent(in)  :: fileunit      ! input
    integer,                         intent(out) :: error_flag    ! result

    integer      :: i,s
    real(FAdoxy) :: value_FAdoxy
    logical      :: value_logical

    error_flag = no_error

    ! Calibration_Valid
    call read_logical1_from_bin_file(fileunit, value_logical, error_flag)
    IF (error_flag .ne. no_error) return
    Rayl_Resp_Calib_Data%Calib_Valid = value_logical

    ! List_of_Rayl_Frequency_Step_Results x 16
    s = size(Rayl_Resp_Calib_Data%List_Rayl_Freq_Step_Res)
    do i = 1, s
       ! Rayl_Frequency_Step_Result
           ! Frequency_Offset
           call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
           IF (error_flag .ne. no_error) return
           Rayl_Resp_Calib_Data%List_Rayl_Freq_Step_Res(i)%Rayl_Freq_Step_Res%Freq_Offset = &
           & value_FAdoxy
           
           ! Frequency_Valid
           call read_logical1_from_bin_file(fileunit, value_logical, error_flag)
           IF (error_flag .ne. no_error) return
           Rayl_Resp_Calib_Data%List_Rayl_Freq_Step_Res(i)%Rayl_Freq_Step_Res%Freq_Valid = &
           & value_logical

           ! Measurement_Response_Valid
           call read_logical1_from_bin_file(fileunit, value_logical, error_flag)
           IF (error_flag .ne. no_error) return
           Rayl_Resp_Calib_Data%List_Rayl_Freq_Step_Res(i)%Rayl_Freq_Step_Res%Meas_Resp_Valid = &
           & value_logical

           ! Reference_Pulse_Response_Valid
           call read_logical1_from_bin_file(fileunit, value_logical, error_flag)
           IF (error_flag .ne. no_error) return
           Rayl_Resp_Calib_Data%List_Rayl_Freq_Step_Res(i)%Rayl_Freq_Step_Res%Ref_Pulse_Resp_Valid = &
           & value_logical

           ! Measurement_Response
           call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
           IF (error_flag .ne. no_error) return
           Rayl_Resp_Calib_Data%List_Rayl_Freq_Step_Res(i)%Rayl_Freq_Step_Res%Meas_Resp = &
           & value_FAdoxy

           ! Measurement_Error_Rayl_Response
           call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
           IF (error_flag .ne. no_error) return
           Rayl_Resp_Calib_Data%List_Rayl_Freq_Step_Res(i)%Rayl_Freq_Step_Res%Meas_Err_Rayl_Resp = &
           & value_FAdoxy

           ! Reference_Pulse_Response
           call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
           IF (error_flag .ne. no_error) return
           Rayl_Resp_Calib_Data%List_Rayl_Freq_Step_Res(i)%Rayl_Freq_Step_Res%Ref_Pulse_Resp = &
           & value_FAdoxy

           ! Reference_Pulse_Error_Rayl_Response
           call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
           IF (error_flag .ne. no_error) return
           Rayl_Resp_Calib_Data%List_Rayl_Freq_Step_Res(i)%Rayl_Freq_Step_Res%Ref_Pulse_Err_Rayl_Resp = &
           & value_FAdoxy

           ! Frequency_Step_Data_Statistics
           call Read_Freq_Step_Data_Stat( & 
           Rayl_Resp_Calib_Data%List_Rayl_Freq_Step_Res(i)%Rayl_Freq_Step_Res%Freq_Step_Data_Stat, &
           fileunit, error_flag)
           IF (error_flag .ne. no_error) return

       ! end of Rayl_Frequency_Step_Result
    enddo

    ! Rayl_Measurement_Response_Calibration
       ! Measurement_Mean_Sensitivity
       call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
       IF (error_flag .ne. no_error) return
       Rayl_Resp_Calib_Data%Rayl_Meas_Resp_Calib%Meas_Mean_Sensitivity = &
       & value_FAdoxy

       ! Measurement_Zero_Frequency
       call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
       IF (error_flag .ne. no_error) return
       Rayl_Resp_Calib_Data%Rayl_Meas_Resp_Calib%Meas_Zero_Freq = &
       & value_FAdoxy

       ! Measurement_Error_Rayl_Response_Std_Dev
       call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
       IF (error_flag .ne. no_error) return
       Rayl_Resp_Calib_Data%Rayl_Meas_Resp_Calib%Meas_Err_Rayl_Resp_Stdev = &
       & value_FAdoxy

    ! Rayl_Reference_Pulse_Response_Calibration
       ! Reference_Pulse_Mean_Sensitivity
       call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
       IF (error_flag .ne. no_error) return
       Rayl_Resp_Calib_Data%Rayl_Ref_Pulse_Resp_Calib%Ref_Pulse_Mean_Sensitivity = &
       & value_FAdoxy

       ! Reference_Pulse_Zero_Frequency
       call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
       IF (error_flag .ne. no_error) return
       Rayl_Resp_Calib_Data%Rayl_Ref_Pulse_Resp_Calib%Ref_Pulse_Zero_Freq = &
       & value_FAdoxy

       ! Reference_Pulse_Error_Rayl_Response_Std_Dev
       call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
       IF (error_flag .ne. no_error) return
       Rayl_Resp_Calib_Data%Rayl_Ref_Pulse_Resp_Calib%Ref_Pulse_Err_Rayl_Resp_Stdev = &
       & value_FAdoxy

    ! Measurement_Calibration_Validity
    call Read_Meas_Calib_Valid(Rayl_Resp_Calib_Data%Meas_Calib_Valid, &
         fileunit, error_flag)
    IF (error_flag .ne. no_error) return

    ! Reference_Pulse_Calibration_Validity
    call Read_Ref_Pulse_Calib_Valid(Rayl_Resp_Calib_Data%Ref_Pulse_Calib_Valid, &
         fileunit, error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine Read_Rayl_Resp_Calib_Data
    !  #]
  !---------------------------------
  subroutine Init_Mie_Resp_Calib_Data(Mie_Resp_Calib_Data)
    !  #[
    TYPE(Mie_Resp_Calib_Data_type), intent(out) :: Mie_Resp_Calib_Data

    integer :: i,s

    Mie_Resp_Calib_Data%Calib_Valid = .FALSE.
    s = size(Mie_Resp_Calib_Data%List_Mie_Freq_Step_Res)
    DO i=1,s
      call Init_List_Mie_Freq_Step_Res(Mie_Resp_Calib_Data%List_Mie_Freq_Step_Res(i))
    ENDDO
    call Init_Mie_Meas_Resp_Calib(Mie_Resp_Calib_Data%Mie_Meas_Resp_Calib)
    call Init_Mie_Ref_Pulse_Resp_Calib(Mie_Resp_Calib_Data%Mie_Ref_Pulse_Resp_Calib)
    call Init_Meas_Calib_Valid(Mie_Resp_Calib_Data%Meas_Calib_Valid)
    call Init_Ref_Pulse_Calib_Valid(Mie_Resp_Calib_Data%Ref_Pulse_Calib_Valid)
  end subroutine Init_Mie_Resp_Calib_Data
    !  #]
  subroutine Print_Mie_Resp_Calib_Data(Mie_Resp_Calib_Data)
    !  #[
    TYPE(Mie_Resp_Calib_Data_type), intent(in) :: Mie_Resp_Calib_Data

    integer :: i,s

    print *,"        ...%Calib_Valid = ",Mie_Resp_Calib_Data%Calib_Valid
    s = size(Mie_Resp_Calib_Data%List_Mie_Freq_Step_Res)
    DO i=1,s
      print *,"        ...%List_Mie_Freq_Step_Res(",i,") :"
      call Print_List_Mie_Freq_Step_Res(Mie_Resp_Calib_Data%List_Mie_Freq_Step_Res(i))
    ENDDO 
    print *,"        ...%Mie_Meas_Resp_Calib : "
    call Print_Mie_Meas_Resp_Calib(Mie_Resp_Calib_Data%Mie_Meas_Resp_Calib)
    print *,"        ...%Mie_Ref_Pulse_Resp_Calib : "
    call Print_Mie_Ref_Pulse_Resp_Calib(Mie_Resp_Calib_Data%Mie_Ref_Pulse_Resp_Calib)
    print *,"        ...%Meas_Calib_Valid : "
    call Print_Meas_Calib_Valid(Mie_Resp_Calib_Data%Meas_Calib_Valid)
    print *,"        ...%Ref_Pulse_Calib_Valid : "
    call Print_Ref_Pulse_Calib_Valid(Mie_Resp_Calib_Data%Ref_Pulse_Calib_Valid)
  end subroutine Print_Mie_Resp_Calib_Data
    !  #]
  subroutine Read_Mie_Resp_Calib_Data(Mie_Resp_Calib_Data,fileunit,error_flag)
    !  #[
    TYPE(Mie_Resp_Calib_Data_type), intent(out) :: Mie_Resp_Calib_Data ! result
    integer,                        intent(in)  :: fileunit      ! input
    integer,                        intent(out) :: error_flag    ! result

    integer      :: i,s
    real(FAdoxy) :: value_FAdoxy
    logical      :: value_logical

    error_flag = no_error

    ! Calibration_Valid
    call read_logical1_from_bin_file(fileunit, value_logical, error_flag)
    IF (error_flag .ne. no_error) return
    Mie_Resp_Calib_Data%Calib_Valid = value_logical

    ! List_of_Mie_Frequency_Step_Results x 16
    s = size(Mie_Resp_Calib_Data%List_Mie_Freq_Step_Res)
    do i = 1, s
       ! Mie_Frequency_Step_Result
           ! Frequency_Offset
           call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
           IF (error_flag .ne. no_error) return
           Mie_Resp_Calib_Data%List_Mie_Freq_Step_Res(i)%Mie_Freq_Step_Res%Freq_Offset = &
           & value_FAdoxy
           
           ! Frequency_Valid
           call read_logical1_from_bin_file(fileunit, value_logical, error_flag)
           IF (error_flag .ne. no_error) return
           Mie_Resp_Calib_Data%List_Mie_Freq_Step_Res(i)%Mie_Freq_Step_Res%Freq_Valid = &
           & value_logical

           ! Measurement_Response_Valid
           call read_logical1_from_bin_file(fileunit, value_logical, error_flag)
           IF (error_flag .ne. no_error) return
           Mie_Resp_Calib_Data%List_Mie_Freq_Step_Res(i)%Mie_Freq_Step_Res%Meas_Resp_Valid=&
           & value_logical

           ! Reference_Pulse_Response_Valid
           call read_logical1_from_bin_file(fileunit, value_logical, error_flag)
           IF (error_flag .ne. no_error) return
           Mie_Resp_Calib_Data%List_Mie_Freq_Step_Res(i)%Mie_Freq_Step_Res%Ref_Pulse_Resp_Valid=&
           & value_logical

           ! Measurement_Response
           call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
           IF (error_flag .ne. no_error) return
           Mie_Resp_Calib_Data%List_Mie_Freq_Step_Res(i)%Mie_Freq_Step_Res%Meas_Resp = &
           & value_FAdoxy

           ! Measurement_Error_Mie_Response
            call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
           IF (error_flag .ne. no_error) return
           Mie_Resp_Calib_Data%List_Mie_Freq_Step_Res(i)%Mie_Freq_Step_Res%Meas_Err_Mie_Resp=&
           & value_FAdoxy

           ! Reference_Pulse_Response
           call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
           IF (error_flag .ne. no_error) return
           Mie_Resp_Calib_Data%List_Mie_Freq_Step_Res(i)%Mie_Freq_Step_Res%Ref_Pulse_Resp =&
           & value_FAdoxy

           ! Reference_Pulse_Error_Mie_Response
           call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
           IF (error_flag .ne. no_error) return
           Mie_Resp_Calib_Data%List_Mie_Freq_Step_Res(i)%Mie_Freq_Step_Res%Ref_Pulse_Err_Mie_Resp =&
           & value_FAdoxy

           ! Frequency_Step_Data_Statistics
           call Read_Freq_Step_Data_Stat( & 
           Mie_Resp_Calib_Data%List_Mie_Freq_Step_Res(i)%Mie_Freq_Step_Res%Freq_Step_Data_Stat, &
           fileunit, error_flag)
           IF (error_flag .ne. no_error) return

       ! end of Mie_Frequency_Step_Result
    enddo

    ! Mie_Measurement_Response_Calibration
       ! Measurement_Mean_Sensitivity
       call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
       IF (error_flag .ne. no_error) return
       Mie_Resp_Calib_Data%Mie_Meas_Resp_Calib%Meas_Mean_Sensitivity = &
       & value_FAdoxy

       ! Measurement_Zero_Frequency
       call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
       IF (error_flag .ne. no_error) return
       Mie_Resp_Calib_Data%Mie_Meas_Resp_Calib%Meas_Zero_Freq = &
       & value_FAdoxy

       ! Measurement_Error_Mie_Response_Std_Dev
       call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
       IF (error_flag .ne. no_error) return
       Mie_Resp_Calib_Data%Mie_Meas_Resp_Calib%Meas_Err_Mie_Resp_Stdev = &
       & value_FAdoxy

    ! Mie_Reference_Pulse_Response_Calibration
       ! Reference_Pulse_Mean_Sensitivity
       call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
       IF (error_flag .ne. no_error) return
       Mie_Resp_Calib_Data%Mie_Ref_Pulse_Resp_Calib%Ref_Pulse_Mean_Sensitivity = &
       & value_FAdoxy

       ! Reference_Pulse_Zero_Frequency
       call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
       IF (error_flag .ne. no_error) return
       Mie_Resp_Calib_Data%Mie_Ref_Pulse_Resp_Calib%Ref_Pulse_Zero_Freq = &
       & value_FAdoxy

       ! Reference_Pulse_Error_Mie_Response_Std_Dev
       call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
       IF (error_flag .ne. no_error) return
       Mie_Resp_Calib_Data%Mie_Ref_Pulse_Resp_Calib%Ref_Pulse_Err_Mie_Resp_Stdev = &
       & value_FAdoxy

    ! Measurement_Calibration_Validity
    call Read_Meas_Calib_Valid(Mie_Resp_Calib_Data%Meas_Calib_Valid, &
         fileunit, error_flag)
    IF (error_flag .ne. no_error) return

    ! Reference_Pulse_Calibration_Validity
    call Read_Ref_Pulse_Calib_Valid(Mie_Resp_Calib_Data%Ref_Pulse_Calib_Valid, &
         fileunit, error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine Read_Mie_Resp_Calib_Data
    !  #]
  !---------------------------------
  subroutine Init_List_Tripod_Obscur_Corr(List_Tripod_Obscur_Corr)
    !  #[
    TYPE(List_Tripod_Obscur_Corr_type), intent(out) :: List_Tripod_Obscur_Corr

    List_Tripod_Obscur_Corr%Tripod_Obscur_Corr = missing_indicator_FAdoxy
  end subroutine Init_List_Tripod_Obscur_Corr
    !  #]
  subroutine Print_List_Tripod_Obscur_Corr(List_Tripod_Obscur_Corr)
    !  #[
    TYPE(List_Tripod_Obscur_Corr_type), intent(in) :: List_Tripod_Obscur_Corr
    character(len=13) :: txt

    txt=real2txt(List_Tripod_Obscur_Corr%Tripod_Obscur_Corr)
    print *,"            ...%Tripod_Obscur_Cor = ",txt
  end subroutine Print_List_Tripod_Obscur_Corr
    !  #]
  subroutine Init_Err_Quant(Err_Quant)
    !  #[
    TYPE(Err_Quant_type), intent(out) :: Err_Quant

    Err_Quant%Mie_Err_Quant_K1 = missing_indicator_FAdoxy
    Err_Quant%Mie_Err_Quant_K2 = missing_indicator_FAdoxy
    Err_Quant%Mie_Err_Quant_K3 = missing_indicator_FAdoxy
    Err_Quant%Rayl_Err_Quant_Ka2 = missing_indicator_FAdoxy
    Err_Quant%Rayl_Err_Quant_Ka3 = missing_indicator_FAdoxy
    Err_Quant%Rayl_Err_Quant_Kb2 = missing_indicator_FAdoxy
    Err_Quant%Rayl_Err_Quant_Kb3 = missing_indicator_FAdoxy
  end subroutine Init_Err_Quant
    !  #]
  subroutine Print_Err_Quant(Err_Quant)
    !  #[
    TYPE(Err_Quant_type), intent(in) :: Err_Quant
    character(len=13) :: txt

    txt = real2txt(Err_Quant%Mie_Err_Quant_K1)
    print *,"            ...%Mie_Err_Quant_K1   = ",txt
    txt = real2txt(Err_Quant%Mie_Err_Quant_K2)
    print *,"            ...%Mie_Err_Quant_K2   = ",txt
    txt = real2txt(Err_Quant%Mie_Err_Quant_K3)
    print *,"            ...%Mie_Err_Quant_K3   = ",txt
    txt = real2txt(Err_Quant%Rayl_Err_Quant_Ka2)
    print *,"            ...%Rayl_Err_Quant_Ka2 = ",txt
    txt = real2txt(Err_Quant%Rayl_Err_Quant_Ka3)
    print *,"            ...%Rayl_Err_Quant_Ka3 = ",txt
    txt = real2txt(Err_Quant%Rayl_Err_Quant_Kb2)
    print *,"            ...%Rayl_Err_Quant_Kb2 = ",txt
    txt = real2txt(Err_Quant%Rayl_Err_Quant_Kb3)
    print *,"            ...%Rayl_Err_Quant_Kb3 = ",txt
  end subroutine Print_Err_Quant
    !  #]
  subroutine Init_Sat_Char_Data(Sat_Char_Data)
    !  #[
    TYPE(Sat_Char_Data_type), intent(out) :: Sat_Char_Data

    integer :: i,s

    Sat_Char_Data%Laser_Wavelength = missing_indicator_FAdoxy
    call Init_Err_Quant(Sat_Char_Data%Err_Quant)
    s = size(Sat_Char_Data%List_Tripod_Obscur_Corr)
    do i=1,s
      call Init_List_Tripod_Obscur_Corr(Sat_Char_Data%List_Tripod_Obscur_Corr(i))
    enddo
    Sat_Char_Data%NF_Order = missing_indicator_IntAl
    Sat_Char_Data%Mie_harmonic_bias_coeff_A(:)  = missing_indicator_FAdoxy
    Sat_Char_Data%Mie_harmonic_bias_coeff_B(:)  = missing_indicator_FAdoxy
    Sat_Char_Data%Rayl_harmonic_bias_coeff_A(:) = missing_indicator_FAdoxy
    Sat_Char_Data%Rayl_harmonic_bias_coeff_B(:) = missing_indicator_FAdoxy
  end subroutine Init_Sat_Char_Data
    !  #]
  subroutine Print_Sat_Char_Data(Sat_Char_Data)
    !  #[
    TYPE(Sat_Char_Data_type), intent(in) :: Sat_Char_Data

    integer           :: i,s
    character(len=13) :: txt
   
    txt=real2txt(Sat_Char_Data%Laser_Wavelength)
    print *,"        ...%Laser_Wavelength = ",txt
    print *,"        ...%Err_Quant : "
    call Print_Err_Quant(Sat_Char_Data%Err_Quant)
    s = size(Sat_Char_Data%List_Tripod_Obscur_Corr)
    do i=1,s
      print *,"        ...%List_Tripod_Obscur_Corr(",i,") :"
      call Print_List_Tripod_Obscur_Corr(Sat_Char_Data%List_Tripod_Obscur_Corr(i))
    enddo
    txt=int2txt(Sat_Char_Data%NF_Order)
    print *,"        ...%NF_Order = ",txt
    if (missing_int(Sat_Char_Data%NF_Order)) then
       s=1024
    else
       s=min(Sat_Char_Data%NF_Order+1,1024)
    endif
    do i=1,s
      txt=real2txt(Sat_Char_Data%Mie_harmonic_bias_coeff_A(i))
      print *,"        ...%Mie_harmonic_bias_coeff_A(",i,") = ", txt
    enddo
    do i=1,s
      txt=real2txt(Sat_Char_Data%Mie_harmonic_bias_coeff_B(i))
      print *,"        ...%Mie_harmonic_bias_coeff_B(",i,") = ", txt
    enddo
    do i=1,s
      txt=real2txt(Sat_Char_Data%Rayl_harmonic_bias_coeff_A(i))
      print *,"        ...%Rayl_harmonic_bias_coeff_A(",i,") = ", txt
    enddo
    do i=1,s
      txt=real2txt(Sat_Char_Data%Rayl_harmonic_bias_coeff_B(i))
      print *,"        ...%Rayl_harmonic_bias_coeff_B(",i,") = ", txt
    enddo
  end subroutine Print_Sat_Char_Data
    !  #]
  subroutine Read_Sat_Char_Data(Sat_Char_Data,fileunit,NF_Order_Max,error_flag)
    !  #[
    TYPE(Sat_Char_Data_type), intent(out) :: Sat_Char_Data ! result
    integer,              intent(in)  :: fileunit      ! input
    integer,              intent(in)  :: NF_Order_Max  ! input (found in SPH)
    integer,              intent(out) :: error_flag    ! result

    integer        :: i,s
    integer(IntAl) :: value_IntAl
    real(FAdoxy)   :: value_FAdoxy

    error_flag = no_error

    ! Laser_Wavelength
    call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
    IF (error_flag .ne. no_error) return
    Sat_Char_Data%Laser_Wavelength = value_FAdoxy

    ! Error_Quantifiers
    call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
    IF (error_flag .ne. no_error) return
    Sat_Char_Data%Err_Quant%Mie_Err_Quant_K1 = value_FAdoxy
    call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
    IF (error_flag .ne. no_error) return
    Sat_Char_Data%Err_Quant%Mie_Err_Quant_K2 = value_FAdoxy
    call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
    IF (error_flag .ne. no_error) return
    Sat_Char_Data%Err_Quant%Mie_Err_Quant_K3 = value_FAdoxy
    call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
    IF (error_flag .ne. no_error) return
    Sat_Char_Data%Err_Quant%Rayl_Err_Quant_Ka2 = value_FAdoxy
    call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
    IF (error_flag .ne. no_error) return
    Sat_Char_Data%Err_Quant%Rayl_Err_Quant_Ka3 = value_FAdoxy
    call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
    IF (error_flag .ne. no_error) return
    Sat_Char_Data%Err_Quant%Rayl_Err_Quant_Kb2 = value_FAdoxy
    call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
    IF (error_flag .ne. no_error) return
    Sat_Char_Data%Err_Quant%Rayl_Err_Quant_Kb3 = value_FAdoxy

    ! List_of_Tripod_Obscuration_Corrections 
    s = size(Sat_Char_Data%List_Tripod_Obscur_Corr)
    do i = 1, s
      call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
      IF (error_flag .ne. no_error) return
      Sat_Char_Data%List_Tripod_Obscur_Corr(i)%Tripod_Obscur_Corr = value_FAdoxy
    enddo

    call read_IntAl_from_bin_file(fileunit, value_IntAl, error_flag)
    IF (error_flag .ne. no_error) return
    Sat_Char_Data%NF_Order = value_IntAl
    IF (Sat_Char_Data%NF_Order .ne. NF_Order_Max) THEN
       print *,"ERROR in Read_Sat_Char_Data: NF_Order found in Sat_Char_Data differs from SPH"
       print *,"Sat_Char_Data%NF_Order = ",Sat_Char_Data%NF_Order
       print *,"SPH%NF_Order_Max       = ",NF_Order_Max
       error_flag = error_reading_file
       return
    END IF
    ! this is required because the structure found in the file cannot hold more
    ! than 1024 elements (cf IODD ADM-IC-52-1666 dated oct 30 2006, table 6.28)
    s=min(NF_Order_Max+1,1024)
    do i = 1, s
      call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
      IF (error_flag .ne. no_error) return
      Sat_Char_Data%Mie_harmonic_bias_coeff_A(i) = value_FAdoxy
    enddo
    do i = 1, s
      call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
      IF (error_flag .ne. no_error) return
      Sat_Char_Data%Mie_harmonic_bias_coeff_B(i) = value_FAdoxy
    enddo
    do i = 1, s
      call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
      IF (error_flag .ne. no_error) return
      Sat_Char_Data%Rayl_harmonic_bias_coeff_A(i) = value_FAdoxy
    enddo
    do i = 1, s
      call read_FAdoxy_from_bin_file(fileunit, value_FAdoxy, error_flag)
      IF (error_flag .ne. no_error) return
      Sat_Char_Data%Rayl_harmonic_bias_coeff_B(i) = value_FAdoxy
    enddo
  end subroutine Read_Sat_Char_Data
    !  #]
  !---------------------------------
  subroutine Init_L1B_Char_Data(L1B_Char_Data)
    !  #[
    TYPE(L1B_Char_Data_type), intent(out) :: L1B_Char_Data

    call Init_Sat_Char_Data(L1B_Char_Data%Sat_Char_Data)
    call Init_Mie_Resp_Calib_Data(L1B_Char_Data%Mie_Resp_Calib_Data)
    call Init_Rayl_Resp_Calib_Data(L1B_Char_Data%Rayl_Resp_Calib_Data)
  end subroutine Init_L1B_Char_Data
    !  #]
  subroutine Print_L1B_Char_Data(L1B_Char_Data)
    !  #[
    TYPE(L1B_Char_Data_type), intent(in) :: L1B_Char_Data

    print *,"    ...%Sat_Char_Data :"
    call Print_Sat_Char_Data(L1B_Char_Data%Sat_Char_Data)
    print *,"    ...%Mie_Resp_Calib_Data :"
    call Print_Mie_Resp_Calib_Data(L1B_Char_Data%Mie_Resp_Calib_Data)
    print *,"    ...%Rayl_Resp_Calib_Data :"
    call Print_Rayl_Resp_Calib_Data(L1B_Char_Data%Rayl_Resp_Calib_Data)
  end subroutine Print_L1B_Char_Data
    !  #]
  subroutine Read_L1B_Char_Data(L1B_Char_Data,fileunit,NF_Order_Max,error_flag)
    !  #[
    TYPE(L1B_Char_Data_type), intent(out) :: L1B_Char_Data ! result
    integer,                  intent(in)  :: fileunit      ! input
    integer,                  intent(in)  :: NF_Order_Max  ! input (found in SPH)
    integer,                  intent(out) :: error_flag    ! result    

    error_flag = no_error
    call Read_Sat_Char_Data(L1B_Char_Data%Sat_Char_Data, &
         fileunit,NF_Order_Max,error_flag)
    IF (error_flag .ne. no_error) return
    call Read_Mie_Resp_Calib_Data(L1B_Char_Data%Mie_Resp_Calib_Data, &
         fileunit,error_flag)
    IF (error_flag .ne. no_error) return
    call Read_Rayl_Resp_Calib_Data(L1B_Char_Data%Rayl_Resp_Calib_Data, &
         fileunit,error_flag)
    IF (error_flag .ne. no_error) return
  end subroutine Read_L1B_Char_Data
    !  #]
  !---------------------------------
  subroutine Init_L1B_Cal_ADS(Cal_ADS, error_flag)
    !  #[
    TYPE(L1B_Cal_ADS_type), pointer     :: Cal_ADS
    integer,                intent(out) :: error_flag

    error_flag = no_error
    IF (.not. associated(Cal_ADS)) THEN
       print *,"ERROR in Init_L1B_Cal_ADS: "
       print *,"pointer Cal_ADS seems not associated !!!"
       error_flag = error_programming
       return
    END IF
    call InitDateTime(Cal_ADS%MRC_First_Start_of_Obs_DateTime)
    call InitDateTime(Cal_ADS%MRC_Last_Start_of_Obs_DateTime)
    call InitDateTime(Cal_ADS%RRC_First_Start_of_Obs_DateTime)
    call InitDateTime(Cal_ADS%RRC_Last_Start_of_Obs_DateTime)
    call Init_L1B_Char_Data(Cal_ADS%L1B_Char_Data)
  end subroutine Init_L1B_Cal_ADS
    !  #]
  subroutine Print_L1B_Cal_ADS(Cal_ADS, error_flag)
    !  #[
    TYPE(L1B_Cal_ADS_type), pointer :: Cal_ADS
    integer,     intent(out)        :: error_flag

    ! local variable
    character(len=maxlen_DateTime_KVT) :: DateTimeString

    error_flag = no_error
    IF (associated(Cal_ADS)) THEN
       print *,"Cal_ADS%..."
    ELSE
       print *,"Cal_ADS [seems not associated]"
       return
    END IF

    DateTimeString = Construct_KVT_DateTime(Cal_ADS%MRC_First_Start_of_Obs_DateTime)
    print *,"...%MRC_First_Start_of_Obs_DateTime = "//trim(DateTimeString)
    DateTimeString = Construct_KVT_DateTime(Cal_ADS%MRC_Last_Start_of_Obs_DateTime)
    print *,"...%MRC_Last_Start_of_Obs_DateTime  = "//trim(DateTimeString)
    DateTimeString = Construct_KVT_DateTime(Cal_ADS%RRC_First_Start_of_Obs_DateTime)
    print *,"...%RRC_First_Start_of_Obs_DateTime = "//trim(DateTimeString)
    DateTimeString = Construct_KVT_DateTime(Cal_ADS%RRC_Last_Start_of_Obs_DateTime)
    print *,"...%RRC_Last_Start_of_Obs_DateTime  = "//trim(DateTimeString)
    print *,"...%L1B_Char_Data :"
    call Print_L1B_Char_Data(Cal_ADS%L1B_Char_Data)

  end subroutine Print_L1B_Cal_ADS
    !  #]
  subroutine Delete_L1B_Cal_ADS(Cal_ADS)
    !  #[
    TYPE(L1B_Cal_ADS_type), pointer :: Cal_ADS

    IF (.not. associated(Cal_ADS)) return
  end subroutine Delete_L1B_Cal_ADS
    !  #]
  subroutine Read_L1B_Cal_ADS(Cal_DSD,fileunit, & !which_DS,&
                             NF_Order_Max, &
                             Cal_ADS,error_flag)
    !  #[

    TYPE(DSD_type),         pointer     :: Cal_DSD      ! input
    integer,                intent(in)  :: fileunit     ! input
!    integer,                intent(in)  :: which_DS     ! input
    integer,                intent(in)  :: NF_Order_Max ! input (found in SPH)
    TYPE(L1B_Cal_ADS_type), pointer     :: Cal_ADS      ! result
    integer,                intent(out) :: error_flag   ! result    

    ! local variables
    integer(i8_)      :: pos, endpos
    integer           :: currentpos
!    integer           :: m ! not used

    error_flag = no_error

    IF (.not. associated(Cal_DSD)) THEN
       print *,"ERROR in Read_L1B_Cal_ADS()"
       print *,"the pointer Cal_DSD seems not associated"
       error_flag = error_programming
       return
    END IF

    IF (.not. associated(Cal_ADS)) THEN
       print *,"ERROR in Read_L1B_Cal_ADS:"
       print *,"the pointer Cal_ADS seems not associated"
       error_flag = error_programming
       return
    END IF

    pos = Cal_DSD%start_of_ds ! + (which_DS-1)*Cal_DSD%size_of_one_dsr
    call SetFilePointer(fileunit,pos,error_flag)
    IF (error_flag .ne. no_error) return

    print *,"trying to read the binary Calibration_Char_GADS " !,"nr. ",which_DS

    ! read the 4 12 byte datetime fields
    call read_DateTime_from_bin_file(fileunit,&
         Cal_ADS%MRC_First_Start_of_Obs_DateTime,error_flag)
    IF (error_flag .ne. no_error) return
    call read_DateTime_from_bin_file(fileunit,&
         Cal_ADS%MRC_Last_Start_of_Obs_DateTime,error_flag)
    IF (error_flag .ne. no_error) return
    call read_DateTime_from_bin_file(fileunit,&
         Cal_ADS%RRC_First_Start_of_Obs_DateTime,error_flag)
    IF (error_flag .ne. no_error) return
    call read_DateTime_from_bin_file(fileunit,&
         Cal_ADS%RRC_Last_Start_of_Obs_DateTime,error_flag)
    IF (error_flag .ne. no_error) return

    call Read_L1B_Char_Data(Cal_ADS%L1B_Char_Data, fileunit, NF_Order_Max, error_flag)
    IF (error_flag .ne. no_error) return

    ! calculate expected end-position after reading this block of data
    !endpos = Cal_DSD%start_of_ds + which_DS*Cal_DSD%size_of_one_dsr
    endpos = Cal_DSD%start_of_ds + Cal_DSD%size_of_one_dsr
    ! request the current file position
    inquire(unit=fileunit,nextrec=currentpos)
    
    IF (currentpos .ne. endpos) THEN
       print *,"ERROR in Read_L1B_Cal_ADS():"
       print *,"there seems to be a problem while reading this binary dataset"
       print *,"expected file position after this DSR is: ",endpos
       print *,"but the actual position seems to be:      ",currentpos
       print *,"Maybe the fileformat definition has changed ???"
       error_flag = error_reading_file
       return
    END IF
  end subroutine Read_L1B_Cal_ADS
    !  #]
  !---------------------------------------------------------------------
end module L1B_Cal_ADS
