module L1B_DataStructure
  !  #[ Description
  !     a module to define the datastructure used by the L2B processor,
  !     for reading the L1B data.
  !     NOTE: after reading all data in this struct is to be regarded
  !           as read-only!!!
  !     and the routines needed for initialisation, allocation
  !     and deallocation of all data structs inside it.
  !     
  !     split from the L2B_DataStructur 30-05-2005, by: J. de Kloe, KNMI
  !     last change: 29-09-2005
  !     Modifications:
  !        01-Sep-2005  D. Tan         Add Measurement ADS + Geoid_Model
  !        29-Sep-2005  J. de Kloe     MPH/SPH improvements
  !        11-Jan-2006  J. de Kloe     added initialisation of datetime
  !        12-May-2006  D. Tan         Null use of ParKind1 module
  !        20-Jul-2006  J. de Kloe     updated get_L1B_latlondatetime_range
  !        19-Sep-2006  P. Poli        add flag rho_ok in L1B_MieRangeGateType
  !        05-Oct-2006  J. de Kloe     disabled L1B_MieRangeGateType and
  !                                    L1B_RayleighMeasurementType and
  !                                    activated the working datastrcucture
  !                                    to replace it.
  !        18-Oct-2006  J. de Kloe     moved the flag rho_ok to the working
  !                                    data structure.
  !        20-Feb-2006  J. de Kloe     added LatLonRanges handling
  !        30-Jul-2007  J. de Kloe     added DateTimeRange handling
  !                                    removed get_L1B_latlondatetime_range
  !        28-Aug-2007  D. Tan         Remove working elements
  !        03-Sep-2007  P. Poli        split Init_L1B_DataStructure and
  !                                    Delete_L1B_DataStructure in two
  !                                    parts: HDR and BRC 
  !                                    Allocation of FH,MPH,SPH now done inside 
  !                                    Init_L1B_HDR_DataStructure, to ensure symmetry 
  !                                    with Delete_L1B_HDR_DataStructure. 
  !                                    Changed routine DetermineLatLonDateTimeRanges.
  !        04-Sep-2007  P. Poli        split L1B_DataType in two parts:
  !                                    L1B_HDR_DataType and L1B_BRC_DataType
  !        19-Oct-2007  P. Poli        removed FH,MPH,SPH from L1B_BRC_DataType
  !        03-Jan-2008  J. de Kloe     added nullify statements to make the code
  !                                    more robust
  !        16-Jan-2008  J. de Kloe     phase out integer kind i_ 
  !
  !  #]
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE FixedHeader, only: FH_type, Init_FH, Print_FH
  USE MainProductHeader, only: MPH_type, Init_MPH, Print_MPH
  USE DataSetDescriptor, only: DSD_type, Init_DSD
  USE L1B_SPH, only: L1B_SPH_type, Init_L1B_SPH, Print_L1B_SPH,&
                     Delete_L1B_SPH
  USE L1B_Geoloc_ADS, only: L1B_Geoloc_ADS_type, Init_L1B_Geoloc_ADS, &
                            Delete_L1B_Geoloc_ADS
  USE L1B_PCD_ADS, only: L1B_PCD_ADS_type, Init_L1B_PCD_ADS, Delete_L1B_PCD_ADS
  USE L1B_GWD_ADS, only: L1B_GWD_ADS_type, Init_L1B_GWD_ADS, Delete_L1B_GWD_ADS
  USE L1B_Meas_ADS,only: L1B_Meas_ADS_type, Init_L1B_Meas_ADS, &
                         Delete_L1B_Meas_ADS
  USE L1B_US_MDS,  only: L1B_US_MDS_type, Init_L1B_US_MDS, &
                         Delete_L1B_US_MDS
  USE L1B_WV_MDS,  only: L1B_WV_MDS_type, Init_L1B_WV_MDS, &
                         Delete_L1B_WV_MDS
  USE L1B_Cal_ADS, only: L1B_Cal_ADS_type, Init_L1B_Cal_ADS, &
                         Delete_L1B_Cal_ADS
  USE EE_CFI_Datatypes, only: DateTimeType,&
                         InitDateTime, SetDateTime, DateTimeMissing, &
                         DateTime_1_after_or_eq_2, &
                         Convert_KVT_to_Num_DateTime, &
                         IntAl, missing_indicator_intal
  ! support modules
  USE numerics, only: r8_, i8_, missing_indicator_real, &
                                missing_indicator_integer
  USE ErrorHandler, only: no_error, error_allocate, error_programming
  !  #]
  !  #[ Parameter and data type definitions
  implicit none

  ! make all module contents public
  public

     ! XML file
     !==>FH_items (XML) (fixed header) (see sec. 3.3.1)
     !==>VH (variable header) 
     !   +->MPH_items (XML) (main product header) (see sec. 3.3.2)
     !   +->SPH (specific product header)
     !      +->SPHD_items (XML) (SPH descriptor)
     !      +->DSD1_items (XML) (data set descriptor) (see sec. 3.3.3)
     !      +->DSD2_items (XML) (data set descriptor)
     !      +->etc...
     !==>DB (data block) (maybe xml for small block or binary for large)
     !   +->DS1 (data set) (3 types: measurement dataset: MDS,
     !      |               annotation dataset: ADS,or global ADS: GADS)
     !      +->DSR1 (XML) (data set record)
     !      +->DSR2 (XML) (data set record)
     !      +->etc...
     !   +->DS2 (XML) (data set)
     !   +->etc...
     !
     ! KVT/bin file
     !==>DBH (datablock header) (KVT/ASCII)
     !   +->MPH_items (KVT) (main product header) (see sec. 3.3.2)
     !   +->SPH (specific product header)
     !      +->SPHD_items (KVT) (SPH descriptor)
     !      +->DSD1_items (KVT) (data set descriptor) (see sec. 3.3.3)
     !      +->DSD2_items (KVT) (data set descriptor)
     !      +->etc... (12 DSD items in total)
     !==>DB (data block) (maybe xml for small block or binary for large)
     !   +->DS1 (data set) (3 types: measurement dataset: MDS,
     !      |               annotation dataset: ADS,or global ADS: GADS)
     !      +->DSR1 (BIN) (data set record)
     !      +->DSR2 (BIN)(data set record)
     !      +->etc...
     !   +->DS2 (data set)
     !   +->etc...
     !==>

  TYPE L1B_HDR_DataType
     !  #[
     ! the L1B headers
     logical :: headers_are_available
     logical :: FH_is_available 
     logical :: MPH_is_available
     logical :: SPH_is_available 
     TYPE(FH_type),      pointer :: FH
     TYPE(MPH_type),     pointer :: MPH
     TYPE(L1B_SPH_type), pointer :: SPH
  END TYPE L1B_HDR_DataType
     !  #]

  TYPE L1B_BRC_DataType
     !  #[
     ! the 7 attached datasets 
     TYPE(L1B_Geoloc_ADS_type), pointer :: Geoloc_ADS
     TYPE(L1B_PCD_ADS_type),    pointer :: PCD_ADS
     TYPE(L1B_GWD_ADS_type),    pointer :: GWD_ADS
     TYPE(L1B_Meas_ADS_type),   pointer :: Meas_ADS
     TYPE(L1B_Cal_ADS_type),    pointer :: Cal_ADS
     TYPE(L1B_US_MDS_type),     pointer :: US_MDS
     TYPE(L1B_WV_MDS_type),     pointer :: WV_MDS
     ! the 8 external datasets (do we need all of these ?)
     !TYPE(), pointer :: Aeolus_L1A_Product
     !TYPE(), pointer :: Aeolus_MRC
     !TYPE(), pointer :: Aeolus_RRC
     !TYPE(), pointer :: Aeolus_ZWC
     !TYPE(), pointer :: L1B_Proc_Params
     !TYPE(), pointer :: Sat_Characterisation
     !TYPE(), pointer :: Orbit_Scenario_File
     !TYPE(), pointer :: Geoid_Model_File
  END TYPE L1B_BRC_DataType
     !  #]

  !  #]
  !  #[ Variables
  integer(IntAl),     save :: lat_start, lat_stop, lon_start, lon_stop
  type(DateTimeType), save :: DateTimeStart, DateTimeStop
  logical,            save :: latlon_ranges_initialised = .false.
  
  !  #]
  !  #[ Module interface
  !     everything is public in this module
  !  #]
contains
  !---------------------------------------------
  ! a routine which can be used for determining
  ! lat-lon and DateTime ranges when looping over L1B BRC's
  ! (needed for filling the L2B FH and SPH)
  subroutine InitLatLonDateTimeRanges()
    !  #[
    latlon_ranges_initialised = .true.
    lat_start = missing_indicator_IntAl
    lat_stop  = missing_indicator_IntAl
    lon_start = missing_indicator_IntAl
    lon_stop  = missing_indicator_IntAl

    call InitDateTime(DateTimeStart)
    call InitDateTime(DateTimeStop)

  end subroutine InitLatLonDateTimeRanges
    !  #]
  subroutine DetermineLatLonDateTimeRanges(L1B, error_flag)
    !  #[
    type(L1B_BRC_DataType),intent(in)  :: L1B
    integer,               intent(out) :: error_flag 

    ! local variable
    integer            :: N
    type(DateTimeType) :: DateTime

    error_flag = no_error
    IF (.not. latlon_ranges_initialised) THEN
       print *,"ERROR in DetermineLatLonDateTimeRanges():"
       print *,"latlon/datetime ranges should be initialised by calling"
       print *,"InitLatLonDateTimeRanges, before"
       print *,"DetermineLatLonDateTimeRanges can be used."
       error_flag = error_programming
       return
    END IF

! Poli: comment out:
!    N = L1B%SPH%N_Meas_Max
! and replace by:
    N = size(L1B%Geoloc_ADS%Meas_AOCS)

    ! remember the lat-lon value of meas 1 of the first L1B BRC
    IF (lat_start .eq. missing_indicator_IntAl) &
         lat_start = L1B%Geoloc_ADS%Meas_Geoloc(1)%lat_of_DEM_intersection
    IF (lon_start .eq. missing_indicator_IntAl) &
         lon_start = L1B%Geoloc_ADS%Meas_Geoloc(1)%lon_of_DEM_intersection

    ! overwrite the lat-lon value of meas N of each BRC processed
    lat_stop = L1B%Geoloc_ADS%Meas_Geoloc(N)%lat_of_DEM_intersection
    lon_stop = L1B%Geoloc_ADS%Meas_Geoloc(N)%lon_of_DEM_intersection

    ! start time of this BRC
    DateTime = L1B%Geoloc_ADS%Meas_AOCS(1)%obs_centroid_datetime

    ! remark: actually the above element is meas_centroid_datetime
    ! as defined in the L1B IODD sec.5.3.1, table 5-5, and not
    ! obs_centroid_datetime, but since we reuse the AOCS_type defined 
    ! for observation also for each measurement
    ! the name has changed in our software.
    ! TODO: Should we fix this?

    if (DateTimeMissing(DateTimeStart)) then
       DateTimeStart=DateTime
    else
       if (DateTime_1_after_or_eq_2(DateTimeStart,DateTime)) &
            DateTimeStart = DateTime
    end if

    ! end time of this BRC
    DateTime = L1B%Geoloc_ADS%Meas_AOCS(N)%obs_centroid_datetime

    if (DateTimeMissing(DateTimeStop)) then
       DateTimeStop = DateTime
    else
       if (DateTime_1_after_or_eq_2(DateTime,DateTimeStop)) &
            DateTimeStop = DateTime
    end if
         
  end subroutine DetermineLatLonDateTimeRanges
    !  #]
  subroutine GetLatLonRanges(lat_start_out,lat_stop_out,&
                             lon_start_out,lon_stop_out,error_flag)
    !  #[
    integer(IntAl), intent(out) :: lat_start_out,lat_stop_out
    integer(IntAl), intent(out) :: lon_start_out,lon_stop_out
    integer,        intent(out) :: error_flag 

    error_flag = no_error
    IF (.not. latlon_ranges_initialised) THEN
       print *,"ERROR in GetLatLonRanges():"
       print *,"latlon ranges should be initialised by calling"
       print *,"InitLatLonDateTimeRanges, before GetLatLonRanges can be used."
       error_flag = error_programming
       return
    END IF
    lat_start_out = lat_start
    lat_stop_out  = lat_stop 
    lon_start_out = lon_start
    lon_stop_out  = lon_stop 

  end subroutine GetLatLonRanges
    !  #]
  subroutine GetDateTimeRange(DT_Start,DT_Stop,error_flag)
    !  #[
    type(DateTimeType), intent(out) :: DT_Start, DT_Stop
    integer,            intent(out) :: error_flag 

    error_flag = no_error
    IF (.not. latlon_ranges_initialised) THEN
       print *,"ERROR in GetDateTimeRange():"
       print *,"datetime range should be initialised by calling"
       print *,"InitLatLonDateTimeRanges, before GetDateTimeRange can be used."
       error_flag = error_programming
       return
    END IF
    DT_Start = DateTimeStart
    DT_Stop  = DateTimeStop

  end subroutine GetDateTimeRange
    !  #]
  !---------------------------------------------
  subroutine Init_L1B_HDR_DataStructure(L1B_Headers, error_flag)
    !  #[

    TYPE(L1B_HDR_DataType), intent(out) :: L1B_Headers ! output

    integer, intent(out) :: error_flag    ! output

    ! local variables
    integer :: AllocStatus

    print *,"initialising L1B HDR datastructure ..."
    
    ! init
    error_flag = no_error

    ! allocate the needed memory and init the module datastructure
    L1B_Headers%headers_are_available = .false.

    !------------------------------
    ! allocate memory to hold the FH elements
    !------------------------------
    L1B_Headers%FH_is_available       = .false.
    nullify(L1B_Headers%FH)
    allocate(L1B_Headers%FH,STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation failure in subroutine Init_L1B_HDR_DataStructure"
       error_flag = error_allocate
       return
    END IF
    call Init_FH(L1B_Headers%FH,error_flag)
    IF (error_flag .ne. no_error) return

    !------------------------------
    ! allocate memory to hold the MPH elements
    !------------------------------
    L1B_Headers%MPH_is_available      = .false.
    nullify(L1B_Headers%MPH)
    allocate(L1B_Headers%MPH,STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation failure in subroutine Init_L1B_HDR_DataStructure"
       error_flag = error_allocate
       return
    END IF
    call Init_MPH(L1B_Headers%MPH,error_flag)
    IF (error_flag .ne. no_error) return

    !------------------------------
    ! allocate memory to hold the SPH elements
    !------------------------------
    L1B_Headers%SPH_is_available      = .false.
    nullify(L1B_Headers%SPH)
    allocate(L1B_Headers%SPH,STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation failure in subroutine Init_L1B_HDR_DataStructure"
       error_flag = error_allocate
       return
    END IF
    call Init_L1B_SPH(L1B_Headers%SPH,error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine Init_L1B_HDR_DataStructure
    !  #]
  !---------------------------------------------
  subroutine Init_L1B_BRC_DataStructure(L1B, error_flag, &
       use_binary_datasets, use_Geoloc_dataset, use_PCD_dataset, &
       use_GWD_dataset, use_Meas_dataset, use_Cal_dataset, &
       use_US_dataset, use_WV_dataset)
    !  #[

    TYPE(L1B_BRC_DataType), intent(out) :: L1B ! output

    integer,           intent(out) :: error_flag    ! output
    logical, optional, intent(in)  :: use_binary_datasets
    logical, optional, intent(in)  :: use_Geoloc_dataset
    logical, optional, intent(in)  :: use_PCD_dataset
    logical, optional, intent(in)  :: use_GWD_dataset
    logical, optional, intent(in)  :: use_Meas_dataset
    logical, optional, intent(in)  :: use_Cal_dataset
    logical, optional, intent(in)  :: use_US_dataset
    logical, optional, intent(in)  :: use_WV_dataset

    ! local variables
    integer :: AllocStatus
    logical :: do_use_binary_datasets
    logical :: do_use_Geoloc_dataset
    logical :: do_use_PCD_dataset
    logical :: do_use_GWD_dataset
    logical :: do_use_Meas_dataset
    logical :: do_use_Cal_dataset
    logical :: do_use_US_dataset
    logical :: do_use_WV_dataset

    print *,"initialising L1B BRC datastructure ..."
    
    do_use_binary_datasets = .true.
    IF (present(use_binary_datasets)) &
         do_use_binary_datasets = use_binary_datasets

    IF (do_use_binary_datasets) THEN
       print *,"(allocating and initialising binary datasets as well)"
       do_use_Geoloc_dataset = .true.
       do_use_PCD_dataset    = .true.
       do_use_GWD_dataset    = .true.
       do_use_Meas_dataset   = .true.
       do_use_Cal_dataset    = .true.
       do_use_US_dataset     = .true.
       do_use_WV_dataset     = .true.
    ELSE
       print *,"(not using binary datasets)"
       do_use_Geoloc_dataset = .false.
       do_use_PCD_dataset    = .false.
       do_use_GWD_dataset    = .false.
       do_use_Meas_dataset   = .false.
       do_use_Cal_dataset    = .false.
       do_use_US_dataset     = .false.
       do_use_WV_dataset     = .false.
       IF (present(use_Geoloc_dataset)) &
            do_use_Geoloc_dataset = use_Geoloc_dataset
       IF (present(use_PCD_dataset)) &
            do_use_PCD_dataset = use_PCD_dataset
       IF (present(use_GWD_dataset)) &
            do_use_GWD_dataset = use_GWD_dataset
       IF (present(use_Meas_dataset)) &
            do_use_Meas_dataset = use_Meas_dataset
       IF (present(use_Cal_dataset)) &
            do_use_Cal_dataset = use_Cal_dataset
       IF (present(use_US_dataset)) &
            do_use_US_dataset = use_US_dataset
       IF (present(use_WV_dataset)) &
            do_use_WV_dataset = use_WV_dataset
       
       IF (do_use_Geoloc_dataset) &
            print *,"(except for the Geolocation Dataset, which is used)"
       IF (do_use_PCD_dataset) &
            print *,"(except for the Product Confidence Dataset, ",&
                    "which is used)"
       IF (do_use_GWD_dataset) &
            print *,"(except for the Ground Wind Detection Dataset, ",&
                    "which is used)"
       IF (do_use_Cal_dataset) &
            print *,"(except for the Calibration and Characterization Dataset, which is used)"
       IF (do_use_US_dataset) &
            print *,"(except for the Useful Signal Dataset, which is used)"
       IF (do_use_Meas_dataset) &
            print *,"(except for the Measurement Dataset, which is used)"
       IF (do_use_WV_dataset) &
            print *,"(except for the Wind Velocity Dataset, which is used)"
    END IF

    ! init
    error_flag = no_error

    !------------------------------------------------
    ! allocate and init the binary dataset structures
    !------------------------------------------------
    IF (do_use_Geoloc_dataset) THEN
       allocate(L1B%Geoloc_ADS,stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L1B_BRC_DataStructure"
          error_flag = error_allocate
          return
       END IF
       ! by setting N_meas=0, N_range_gates=0 we indicate that no 
       ! memory is yet to be allocated. Allocation is only done after the
       ! SPH is read, because only then we know the nr of measurements
       ! in each BRC
       call Init_L1B_Geoloc_ADS(L1B%Geoloc_ADS, 0, 0, error_flag)
       IF (error_flag .ne. no_error) return
    ELSE
       nullify(L1B%Geoloc_ADS)
    END IF

    IF (do_use_PCD_dataset) THEN
       allocate(L1B%PCD_ADS,stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L1B_BRC_DataStructure"
          error_flag = error_allocate
          return
       END IF
       ! by setting N_meas=0, N_range_gates=0 we indicate that no 
       ! memory is yet to be allocated. Allocation is only done after the
       ! SPH is read, because only then we know the nr of measurements
       ! in each BRC
       call Init_L1B_PCD_ADS(L1B%PCD_ADS, 0, 0, error_flag)
!       call Init_L1B_PCD_ADS(L1B%PCD_ADS, N_meas, N_range_gates, error_flag)
       IF (error_flag .ne. no_error) return
    ELSE
       nullify(L1B%PCD_ADS)
    END IF

    IF (do_use_GWD_dataset) THEN
       allocate(L1B%GWD_ADS,stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L1B_BRC_DataStructure"
          error_flag = error_allocate
          return
       END IF
       ! by setting N_meas=0 we indicate that no 
       ! memory is yet to be allocated. Allocation is only done after the
       ! SPH is read, because only then we know the nr of measurements
       ! in each BRC
       call Init_L1B_GWD_ADS(L1B%GWD_ADS, 0, error_flag)
!       call Init_L1B_GWD_ADS(L1B%GWD_ADS,N_meas, error_flag)
       IF (error_flag .ne. no_error) return
    ELSE
       nullify(L1B%GWD_ADS)
    END IF

    IF (do_use_Meas_dataset) THEN
       allocate(L1B%Meas_ADS,stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L1B_BRC_DataStructure"
          error_flag = error_allocate
          return
       END IF
       ! by setting N_meas=0, N_range_gates=0 we indicate that no 
       ! memory is yet to be allocated. Allocation is only done after the
       ! SPH is read, because only then we know the nr of measurements
       ! in each BRC
       call Init_L1B_Meas_ADS(L1B%Meas_ADS, 0,0 ,error_flag)
!       call Init_L1B_Meas_ADS(L1B%Meas_ADS,N_meas,N_range_gates,error_flag)
       IF (error_flag .ne. no_error) return
    ELSE
       nullify(L1B%Meas_ADS)
    END IF

    IF (do_use_Cal_dataset) THEN
       allocate(L1B%Cal_ADS,stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L1B_BRC_DataStructure"
          error_flag = error_allocate
          return
       END IF
       call Init_L1B_Cal_ADS(L1B%Cal_ADS, error_flag)
       IF (error_flag .ne. no_error) return
    ELSE
       nullify(L1B%Cal_ADS)
    END IF

    IF (do_use_US_dataset) THEN
       allocate(L1B%US_MDS,stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L1B_BRC_DataStructure"
          error_flag = error_allocate
          return
       END IF
       ! by setting N_meas=0 we indicate that no 
       ! memory is yet to be allocated. Allocation is only done after the
       ! SPH is read, because only then we know the nr of measurements
       ! in each BRC
       call Init_L1B_US_MDS(L1B%US_MDS, 0, 0,error_flag)
!       call Init_L1B_US_MDS(L1B%US_MDS,N_meas,N_range_gates,error_flag)
       IF (error_flag .ne. no_error) return
    ELSE
       nullify(L1B%US_MDS)
    END IF

    IF (do_use_WV_dataset) THEN
       allocate(L1B%WV_MDS,stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L1B_BRC_DataStructure"
          error_flag = error_allocate
          return
       END IF
       ! by setting N_meas=0 we indicate that no 
       ! memory is yet to be allocated. Allocation is only done after the
       ! SPH is read, because only then we know the nr of measurements
       ! in each BRC
       call Init_L1B_WV_MDS(L1B%WV_MDS, 0, 0,error_flag)
!       call Init_L1B_WV_MDS(L1B%WV_MDS,N_meas,N_range_gates,error_flag)
       IF (error_flag .ne. no_error) return       
    ELSE
       nullify(L1B%WV_MDS)
    END IF

  end subroutine Init_L1B_BRC_DataStructure

    !  #]
  subroutine Delete_L1B_HDR_DataStructure(L1B_Headers)
    !  #[

    TYPE(L1B_HDR_DataType), intent(inout) :: L1B_Headers ! input/output
    ! (contents will be deallocated after this routine)

    IF (associated(L1B_Headers%SPH)) call Delete_L1B_SPH(L1B_Headers%SPH)

    !------------------------------------------------------
    ! then deallocate all associated pointers
    !------------------------------------------------------
    IF (associated(L1B_Headers%FH)) THEN
       deallocate(L1B_Headers%FH)
       nullify(L1B_Headers%FH)
    END IF

    IF (associated(L1B_Headers%MPH)) THEN
       deallocate(L1B_Headers%MPH)
       nullify(L1B_Headers%MPH)
    END IF

    IF (associated(L1B_Headers%SPH)) THEN
       deallocate(L1B_Headers%SPH)
       nullify(L1B_Headers%SPH)
    END IF

  end subroutine Delete_L1B_HDR_DataStructure
    !  #]
  subroutine Delete_L1B_BRC_DataStructure(L1B)
    !  #[

    TYPE(L1B_BRC_DataType), intent(inout) :: L1B ! input/output
    ! (the binary datasets will be deallocated after this routine)

    !------------------------------------------------------
    ! call the delete functions for the structures that contain
    ! pointers, to make sure all substructures are deleted
    !------------------------------------------------------
    call Empty_L1B_BinaryDatasets(L1B)

    !------------------------------------------------------
    ! then deallocate all associated pointers
    !------------------------------------------------------
    IF (associated(L1B%Geoloc_ADS)) deallocate(L1B%Geoloc_ADS)
    IF (associated(L1B%PCD_ADS))    deallocate(L1B%PCD_ADS)
    IF (associated(L1B%GWD_ADS))    deallocate(L1B%GWD_ADS)
    IF (associated(L1B%Meas_ADS))   deallocate(L1B%Meas_ADS)
    IF (associated(L1B%US_MDS))     deallocate(L1B%US_MDS)
    IF (associated(L1B%WV_MDS))     deallocate(L1B%WV_MDS)
    IF (associated(L1B%Cal_ADS))    deallocate(L1B%Cal_ADS)

    nullify(L1B%Geoloc_ADS)
    nullify(L1B%PCD_ADS)
    nullify(L1B%GWD_ADS)
    nullify(L1B%Meas_ADS)
    nullify(L1B%US_MDS)
    nullify(L1B%WV_MDS)
    nullify(L1B%Cal_ADS)

  end subroutine Delete_L1B_BRC_DataStructure
    !  #]
  subroutine Empty_L1B_BinaryDatasets(L1B)
    !  #[

    TYPE(L1B_BRC_DataType), intent(inout) :: L1B ! input/output
    ! (the binary datasets will be emptied but not deallocated
    !  after this routine)

    !------------------------------------------------------
    ! call the delete functions for the structures that contain
    ! pointers, to make sure all substructures are deleted
    !------------------------------------------------------
    IF (associated(L1B%Geoloc_ADS)) call Delete_L1B_Geoloc_ADS(L1B%Geoloc_ADS)
    IF (associated(L1B%PCD_ADS))    call Delete_L1B_PCD_ADS(L1B%PCD_ADS)
    IF (associated(L1B%GWD_ADS))    call Delete_L1B_GWD_ADS(L1B%GWD_ADS)
    IF (associated(L1B%Meas_ADS))   call Delete_L1B_Meas_ADS(L1B%Meas_ADS)
    IF (associated(L1B%US_MDS))     call Delete_L1B_US_MDS(L1B%US_MDS)
    IF (associated(L1B%WV_MDS))     call Delete_L1B_WV_MDS(L1B%WV_MDS)
    IF (associated(L1B%Cal_ADS))    call Delete_L1B_Cal_ADS(L1B%Cal_ADS)

  end subroutine Empty_L1B_BinaryDatasets

    !  #]
  !---------------------------------------------
end module L1B_DataStructure
