module L1B_Geoloc_ADS
  !  #[ Documentation
  !     a module to handle the Geolocation annotation dataset
  !
  !     See also the L1BP files: 
  !          L1bP_Draft_1/aeolus/common/PFR/src/
  !             PFR_LevelOneBProductGeolocationDSR.cc
  !             PFR_LevelOneBProductGeolocationDSR.hh
  !     for the code that defines this datablock in the L1BP
  !
  !     Written by J. de Kloe, KNMI 
  !     created:     13-07-2005
  !
  !     Modifications:
  !      02-Sep-2005 D. Tan     Upgrade to DLR TDS v1.2
  !      10-Jan-2006 J. de Kloe new DateTime type
  !      21-Feb-2006 P. Poli    Upgrade to DLR IODD L1B-IODD.231105
  !      28-Mar-2006 P. Poli    Upgrade to DLR IODD AE.TN.DLR.APM-L1B.150206
  !      12-May-2006 D. Tan     Null use of ParKind1 module
  !      03-Nov-2006 P. Poli    use proper missing real/integer indicators;
  !                             moved and cleaned-up reading part
  !      03-Jan-2008 J. de Kloe added nullify statements to make the code
  !                             more robust
  !      16-Jan-2008 J. de Kloe phase out integer kind i_ 
  !
  !  #]
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE Numerics, only: i8_, real2txt
  USE DataSetDescriptor, only: DSD_type
  USE EE_CFI_Datatypes, only: DateTimeType,Construct_KVT_DateTime, InitDateTime,  &
                              read_DateTime_from_bin_file, &
                              read_IntAl_from_bin_file, IntAl, missing_indicator_IntAl, &
                              read_IntAs_from_bin_file, IntAs, missing_indicator_IntAs, &
                              read_FAdoxy_from_bin_file, FAdoxy, missing_indicator_FAdoxy, &
                              int2txt, maxlen_DateTime_KVT
  USE DirectBinaryIO, only: SetFilePointer, AdvanceFilePointer
  USE ErrorHandler, only: no_error, error_programming, error_allocate, error_reading_file
  !  #]
  !  #[ type definitions
  implicit none

  TYPE AOCS_type ! see sec. 5.3.1, table 5-4, of L1B_iodd 
     type(DateTimeType) :: obs_centroid_datetime ! UTC
     real(FAdoxy)       :: X_Position        ! m
     real(FAdoxy)       :: Y_Position        ! m
     real(FAdoxy)       :: Z_Position        ! m
     real(FAdoxy)       :: X_Velocity        ! m/s
     real(FAdoxy)       :: Y_Velocity        ! m/s
     real(FAdoxy)       :: Z_Velocity        ! m/s
     real(FAdoxy)       :: Roll_Angle        ! deg
     real(FAdoxy)       :: Pitch_Angle       ! deg
     real(FAdoxy)       :: Yaw_Angle         ! deg
     ! spare 8 bytes
  END TYPE AOCS_type

! Poli: took the following type as a baseline for deriving
!  Rangebin_Obs_Geoloc_type and Rangebin_Meas_Geoloc_type,
!  both needed in DLR IODD IODD-L1B.2031105
!  TYPE Rangebin_Geoloc_type ! see sec. 5.3.1,tables 5-10/5-11/5-13, of L1B_iodd
!     !
!     ! Add items from AE.TN.DLR.IODD-L1B.050705, Tables 2-10/2-11/2-12/2-13.
!     !
!     integer(i4_) :: lon_of_height_bin ! 10^-6 deg E
!     integer(i4_) :: lat_of_height_bin ! 10^-6 deg N
!     real(r8_)    :: alt_of_height_bin ! m
!     real(r8_)    :: topoc_azim_of_height_bin ! deg
!     real(r8_)    :: topoc_elev_of_height_bin ! deg
!     real(r8_)    :: tgt_ht_los_intersec_height_bin ! in metres?
!     real(r8_)    :: satellite_range_of_height_bin ! m
!  END TYPE Rangebin_Geoloc_type

  TYPE Rangebin_Obs_Geoloc_type
     integer(IntAl) :: lon_of_height_bin ! 10^-6 deg E
     integer(IntAl) :: lat_of_height_bin ! 10^-6 deg N
     real(FAdoxy)   :: alt_of_height_bin ! m
     real(FAdoxy)   :: topoc_azim_of_height_bin ! deg
     real(FAdoxy)   :: topoc_elev_of_height_bin ! deg
     integer(IntAs) :: Target_to_Sun_Visib_Flag
     real(FAdoxy)   :: satellite_range_of_height_bin ! m
  END TYPE Rangebin_Obs_Geoloc_type

  TYPE Rangebin_Meas_Geoloc_type
     integer(IntAl) :: lon_of_height_bin ! 10^-6 deg E
     integer(IntAl) :: lat_of_height_bin ! 10^-6 deg N
     real(FAdoxy)   :: alt_of_height_bin ! m
  END TYPE Rangebin_Meas_Geoloc_type

! Poli: took the following type as a baseline for deriving
!  Obs_Geolocation_type and Meas_Geolocation_type,
!  both needed in DLR IODD IODD-L1B.2031105
!  TYPE Geolocation_type ! see sec. 5.3.1, table 5-7 and 5-9, of L1B_iodd
!     ! values valid for the lower edge of the rangebin,
!     ! except for the 1st element, which gives the values for the
!     ! upper edge of the the top most rangebin
!     !
!     ! Add geoid (AE.TN.DLR.IODD-L1B.050705, Table 2-7).
!     !
!     TYPE(Rangebin_Geoloc_type),dimension(:),pointer:: Mie_Rangebin_Geoloc
!     TYPE(Rangebin_Geoloc_type),dimension(:),pointer:: Rayleigh_Rangebin_Geoloc
!     ! then 5 items (32 bytes) for the DEM geolocation : see table 5-14
!     integer(i4_) :: lat_of_DEM_intersection ! 10^-6 deg N
!     integer(i4_) :: lon_of_DEM_intersection ! 10^-6 deg E
!     real(r8_)    :: alt_of_DEM_intersection ! m
!     real(r8_)    :: topoc_azim_of_DEM_intersection ! deg
!     real(r8_)    :: topoc_elev_of_DEM_intersection ! deg
!     ! then the final 8 byte element from table 5-7 and 5-9
!     real(r8_)    :: Line_of_Sight_Velocity  ! m/s
!     real(r8_)    :: Geoid_Separation_of_Height_Bin  ! m
!     ! spare: 8 bytes
!     !               1234567890123456789012345678901
!  END TYPE Geolocation_type

  TYPE Obs_Geolocation_type
     TYPE(Rangebin_Obs_Geoloc_type),dimension(:),pointer:: Mie_Rangebin_Geoloc
     TYPE(Rangebin_Obs_Geoloc_type),dimension(:),pointer:: Rayleigh_Rangebin_Geoloc
     integer(IntAl) :: lat_of_DEM_intersection ! 10^-6 deg N
     integer(IntAl) :: lon_of_DEM_intersection ! 10^-6 deg E
     real(FAdoxy)   :: alt_of_DEM_intersection ! m
     real(FAdoxy)   :: Line_of_Sight_Velocity  ! m/s
     real(FAdoxy)   :: Geoid_Separation_of_Height_Bin  ! m
     ! spare: 8 bytes 
  END TYPE Obs_Geolocation_type

  TYPE Meas_Geolocation_type
     TYPE(Rangebin_Meas_Geoloc_type),dimension(:),pointer:: Mie_Rangebin_Geoloc
     TYPE(Rangebin_Meas_Geoloc_type),dimension(:),pointer:: Rayleigh_Rangebin_Geoloc
     integer(IntAl) :: lat_of_DEM_intersection ! 10^-6 deg N
     integer(IntAl) :: lon_of_DEM_intersection ! 10^-6 deg E
     real(FAdoxy)   :: alt_of_DEM_intersection ! m
     real(FAdoxy)   :: AOCS_LOS_Velocity  ! m/s
     ! spare: 8 bytes 
  END TYPE Meas_Geolocation_type

  TYPE L1B_Geoloc_ADS_type ! see sec. 5.3.1, table 5-3, of L1B_iodd 
     type(DateTimeType)                     :: start_of_obs_datetime ! UTC
     TYPE(AOCS_type)                        :: Obs_AOCS
     TYPE(AOCS_type), dimension(:), pointer :: Meas_AOCS
     TYPE(Obs_Geolocation_type)                         :: Obs_Geoloc
     TYPE(Meas_Geolocation_type), dimension(:), pointer :: Meas_Geoloc
  END TYPE L1B_Geoloc_ADS_type
  !  #]
contains
  !---------------------------------------------------------------------
  subroutine Init_AOCS(AOCS)
    !  #[
    TYPE(AOCS_type), intent(out) :: AOCS

    call InitDateTime(AOCS%obs_centroid_datetime)
    AOCS%X_Position  = missing_indicator_FAdoxy
    AOCS%Y_Position  = missing_indicator_FAdoxy
    AOCS%Z_Position  = missing_indicator_FAdoxy
    AOCS%X_Velocity  = missing_indicator_FAdoxy
    AOCS%Y_Velocity  = missing_indicator_FAdoxy
    AOCS%Z_Velocity  = missing_indicator_FAdoxy
    AOCS%Roll_Angle  = missing_indicator_FAdoxy
    AOCS%Pitch_Angle = missing_indicator_FAdoxy
    AOCS%Yaw_Angle   = missing_indicator_FAdoxy

  end subroutine Init_AOCS
    !  #]
  subroutine Print_AOCS(AOCS)
    !  #[
    TYPE(AOCS_type), intent(in) :: AOCS

    ! local variables
    character(len=13) :: txt

    ! print the date and time
    print *,"...%obs_centroid_datetime: ",&
         Construct_KVT_DateTime(AOCS%obs_centroid_datetime)

    txt = real2txt(AOCS%X_position)
    print *,"...%X_Position  = ",txt
    txt = real2txt(AOCS%Y_position)
    print *,"...%Y_Position  = ",txt
    txt = real2txt(AOCS%Z_position)
    print *,"...%Z_Position  = ",txt
    txt = real2txt(AOCS%X_Velocity)
    print *,"...%X_Velocity  = ",txt
    txt = real2txt(AOCS%Y_Velocity)
    print *,"...%Y_Velocity  = ",txt
    txt = real2txt(AOCS%Z_Velocity)
    print *,"...%Z_Velocity  = ",txt
    txt = real2txt(AOCS%Roll_Angle)
    print *,"...%Roll_Angle  = ",txt
    txt = real2txt(AOCS%Pitch_Angle)
    print *,"...%Pitch_Angle = ",txt
    txt = real2txt(AOCS%Yaw_Angle)
    print *,"...%Yaw_Angle   = ",txt

  end subroutine Print_AOCS
    !  #]
  subroutine Read_AOCS(fileunit,AOCS,error_flag)
    !  #[
    integer,         intent(in)  :: fileunit     ! input
    TYPE(AOCS_type), intent(out) :: AOCS         ! result
    integer,         intent(out) :: error_flag   ! result

    ! local variables
    real(FAdoxy)         :: value_FAdoxy

    error_flag = no_error

    ! read the 12 byte datetime field
    call read_DateTime_from_bin_file(fileunit,&
         AOCS%obs_centroid_datetime,error_flag)

    ! read X_Position
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    AOCS%X_Position = value_FAdoxy

    ! read Y_Position
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    AOCS%Y_Position = value_FAdoxy

    ! read Z_Position
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    AOCS%Z_Position = value_FAdoxy

    ! read X_Velocity
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    AOCS%X_Velocity = value_FAdoxy

    ! read Y_Velocity
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    AOCS%Y_Velocity = value_FAdoxy

    ! read Z_Velocity
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    AOCS%Z_Velocity = value_FAdoxy

    ! read Roll_Angle
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    AOCS%Roll_Angle = value_FAdoxy

    ! read Pitch_Angle
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    AOCS%Pitch_Angle = value_FAdoxy

    ! read Yaw_Angle
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    AOCS%Yaw_Angle = value_FAdoxy

    ! skip the 8 byte spare
    ! to be replaced by a call to: read_zero_bytes_bin_file
    call AdvanceFilePointer(fileunit,8_i8_,error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine Read_AOCS
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_Rangebin_Obs_Geoloc(Rangebin_Obs_Geoloc)
    !  #[
    TYPE(Rangebin_Obs_Geoloc_type), intent(out) :: Rangebin_Obs_Geoloc

    Rangebin_Obs_Geoloc%lon_of_height_bin = missing_indicator_IntAl
    Rangebin_Obs_Geoloc%lat_of_height_bin = missing_indicator_IntAl
    Rangebin_Obs_Geoloc%alt_of_height_bin = missing_indicator_FAdoxy
    Rangebin_Obs_Geoloc%topoc_azim_of_height_bin = missing_indicator_FAdoxy
    Rangebin_Obs_Geoloc%topoc_elev_of_height_bin = missing_indicator_FAdoxy
!    Rangebin_Geoloc%tgt_ht_los_intersec_height_bin = missing_indicator_FAdoxy
    Rangebin_Obs_Geoloc%Target_to_Sun_Visib_Flag = missing_indicator_IntAs
    Rangebin_Obs_Geoloc%satellite_range_of_height_bin = missing_indicator_FAdoxy

  end subroutine Init_Rangebin_Obs_Geoloc
    !  #]
  subroutine Print_Rangebin_Obs_Geoloc(Rangebin_Obs_Geoloc)
    !  #[
    TYPE(Rangebin_Obs_Geoloc_type), intent(in) :: Rangebin_Obs_Geoloc

    ! local variables
    character(len=13) :: txt

    txt=int2txt(Rangebin_Obs_Geoloc%lon_of_height_bin)
    print *,"...%lon_of_height_bin              = ",txt," *10^-6 deg"

    txt=int2txt(Rangebin_Obs_Geoloc%lat_of_height_bin)
    print *,"...%lat_of_height_bin              = ",txt," *10^-6 deg"

    txt = real2txt(Rangebin_Obs_Geoloc%alt_of_height_bin)
    print *,"...%alt_of_height_bin              = ",txt

    txt = real2txt(Rangebin_Obs_Geoloc%topoc_azim_of_height_bin)
    print *,"...%topoc_azim_of_height_bin       = ",txt

    txt = real2txt(Rangebin_Obs_Geoloc%topoc_elev_of_height_bin)
    print *,"...%topoc_elev_of_height_bin       = ",txt

!    txt = real2txt(Rangebin_Geoloc%tgt_ht_los_intersec_height_bin)
!    print *,"...%tgt_ht_los_intersec_height_bin = ",txt

    txt = int2txt(Rangebin_Obs_Geoloc%Target_to_Sun_Visib_Flag)
    print *,"...%Target_to_Sun_Visib_Flag       = ",txt

    txt = real2txt(Rangebin_Obs_Geoloc%satellite_range_of_height_bin)
    print *,"...%satellite_range_of_height_bin  = ",txt

  end subroutine Print_Rangebin_Obs_Geoloc
    !  #]
  subroutine Read_Rangebin_Obs_Geoloc(fileunit,Rangebin_Geoloc,error_flag)
    !  #[
    integer,                        intent(in)  :: fileunit        ! input
    TYPE(Rangebin_Obs_Geoloc_type), intent(out) :: Rangebin_Geoloc ! result
    integer,                        intent(out) :: error_flag      ! result
    
    ! local variables
    integer(IntAl) :: value_IntAl
    integer(IntAs) :: value_IntAs
    real(FAdoxy)    :: value_FAdoxy

    error_flag = no_error

    ! read lon_of_height_bin
    call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    Rangebin_Geoloc%lon_of_height_bin = value_IntAl

    ! read lat_of_height_bin
    call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    Rangebin_Geoloc%lat_of_height_bin = value_IntAl

    ! read alt_of_height_bin
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Rangebin_Geoloc%alt_of_height_bin = value_FAdoxy

    ! read topoc_azim_of_height_bin
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Rangebin_Geoloc%topoc_azim_of_height_bin = value_FAdoxy

    ! read topoc_elev_of_height_bin
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Rangebin_Geoloc%topoc_elev_of_height_bin = value_FAdoxy

! Poli: removed in DLR IODD IODD-L1B.231105
!    ! read tgt_ht_los_intersec_of_height_bin
!    call read_real8_from_bin_file(fileunit,value_FAdoxy,error_flag)
!    IF (error_flag .ne. no_error) return
!    Rangebin_Geoloc%tgt_ht_los_intersec_height_bin = value_FAdoxy
! Poli: new in DLR IODD IODD-L1B.231105
    ! read Target_to_Sun_Visibility_Flag
    call read_IntAs_from_bin_file(fileunit, value_IntAs, error_flag)
    IF (error_flag .ne. no_error) return
    Rangebin_Geoloc%Target_to_Sun_Visib_Flag = value_IntAs

    ! read satellite_range_of_height_bin
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Rangebin_Geoloc%satellite_range_of_height_bin = value_FAdoxy

  end subroutine Read_Rangebin_Obs_Geoloc
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_Rangebin_Meas_Geoloc(Rangebin_Meas_Geoloc)
    !  #[
    TYPE(Rangebin_Meas_Geoloc_type), intent(out) :: Rangebin_Meas_Geoloc

    Rangebin_Meas_Geoloc%lon_of_height_bin = missing_indicator_IntAl
    Rangebin_Meas_Geoloc%lat_of_height_bin = missing_indicator_IntAl
    Rangebin_Meas_Geoloc%alt_of_height_bin = missing_indicator_FAdoxy

  end subroutine Init_Rangebin_Meas_Geoloc
    !  #]
  subroutine Print_Rangebin_Meas_Geoloc(Rangebin_Meas_Geoloc)
    !  #[
    TYPE(Rangebin_Meas_Geoloc_type), intent(in) :: Rangebin_Meas_Geoloc

    ! local variables
    character(len=13) :: txt

    txt=int2txt(Rangebin_Meas_Geoloc%lon_of_height_bin)
    print *,"...%lon_of_height_bin              = ",txt," *10^-6 deg"

    txt=int2txt(Rangebin_Meas_Geoloc%lat_of_height_bin)
    print *,"...%lat_of_height_bin              = ",txt," *10^-6 deg"

    txt = real2txt(Rangebin_Meas_Geoloc%alt_of_height_bin)
    print *,"...%alt_of_height_bin              = ",txt

  end subroutine Print_Rangebin_Meas_Geoloc
    !  #]
  subroutine Read_Rangebin_Meas_Geoloc(fileunit,Rangebin_Geoloc,error_flag)
    !  #[
    integer,                         intent(in)  :: fileunit        ! input
    TYPE(Rangebin_Meas_Geoloc_type), intent(out) :: Rangebin_Geoloc ! result
    integer,                         intent(out) :: error_flag      ! result
    
    ! local variables
    integer(IntAl) :: value_IntAl
    real(FAdoxy)    :: value_FAdoxy

    error_flag = no_error

    ! read lon_of_height_bin
    call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    Rangebin_Geoloc%lon_of_height_bin = value_IntAl

    ! read lat_of_height_bin
    call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    Rangebin_Geoloc%lat_of_height_bin = value_IntAl

    ! read alt_of_height_bin
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Rangebin_Geoloc%alt_of_height_bin = value_FAdoxy

  end subroutine Read_Rangebin_Meas_Geoloc
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_Obs_Geolocation(Obs_Geoloc, N_range_gates, error_flag)
    !  #[
    TYPE(Obs_Geolocation_type), intent(out) :: Obs_Geoloc
    integer,                    intent(in)  :: N_range_gates
    integer,                    intent(out) :: error_flag

    ! local variables
    integer :: AllocStatus, r

    error_flag = no_error

    IF (N_range_gates .eq. 0) THEN
       ! initialise it to be empty
       nullify(Obs_Geoloc%Mie_Rangebin_Geoloc)
       nullify(Obs_Geoloc%Rayleigh_Rangebin_Geoloc)
    ELSE
       allocate(Obs_Geoloc%Mie_Rangebin_Geoloc(N_range_gates),stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_Obs_Geolocation"
          error_flag = error_allocate
          return
       END IF
       DO r=1,N_range_gates
          call Init_Rangebin_Obs_Geoloc(Obs_Geoloc%Mie_Rangebin_Geoloc(r))
       END DO

       allocate(Obs_Geoloc%Rayleigh_Rangebin_Geoloc(N_range_gates),stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_Obs_Geolocation"
          error_flag = error_allocate
          return
       END IF
       DO r=1,N_range_gates
          call Init_Rangebin_Obs_Geoloc(Obs_Geoloc%Rayleigh_Rangebin_Geoloc(r))
       END DO
    END IF

    Obs_Geoloc%lat_of_DEM_intersection        = missing_indicator_IntAl
    Obs_Geoloc%lon_of_DEM_intersection        = missing_indicator_IntAl
    Obs_Geoloc%alt_of_DEM_intersection        = missing_indicator_FAdoxy
!    Obs_Geoloc%topoc_azim_of_DEM_intersection = missing_indicator_FAdoxy
!    Obs_Geoloc%topoc_elev_of_DEM_intersection = missing_indicator_FAdoxy
    Obs_Geoloc%Line_of_Sight_Velocity         = missing_indicator_FAdoxy
    Obs_Geoloc%Geoid_Separation_of_Height_Bin = missing_indicator_FAdoxy

  end subroutine Init_Obs_Geolocation
    !  #]
  subroutine Delete_Obs_Geolocation(Obs_Geoloc)
    !  #[
    TYPE(Obs_Geolocation_type), intent(inout) :: Obs_Geoloc

    IF (associated(Obs_Geoloc%Mie_Rangebin_Geoloc)) THEN
       deallocate(Obs_Geoloc%Mie_Rangebin_Geoloc)
       nullify(Obs_Geoloc%Mie_Rangebin_Geoloc)
    END IF

    IF (associated(Obs_Geoloc%Rayleigh_Rangebin_Geoloc)) THEN
       deallocate(Obs_Geoloc%Rayleigh_Rangebin_Geoloc)
       nullify(Obs_Geoloc%Rayleigh_Rangebin_Geoloc)
    END IF

  end subroutine Delete_Obs_Geolocation
    !  #]
  subroutine Print_Obs_Geolocation(Obs_Geoloc)
    !  #[
    TYPE(Obs_Geolocation_type), intent(in) :: Obs_Geoloc

    ! local variable
    integer           :: s, r
    character(len=13) :: txt

    ! print: Obs_Geoloc%Mie_Rangebin_Geoloc(:)
    IF (associated(Obs_Geoloc%Mie_Rangebin_Geoloc)) THEN
       s = size(Obs_Geoloc%Mie_Rangebin_Geoloc)
       print "(a,i2,a)","==> size(Obs_Geoloc%Mie_Rangebin_Geoloc)      = ",s,&
            " [printing first and last one]"
       DO r=1,s,s-1
          print "(' Obs_Geoloc%Mie_Rangebin_Geoloc(',i2.2,')%...')",r
          call Print_Rangebin_Obs_Geoloc(Obs_Geoloc%Mie_Rangebin_Geoloc(r))
       END DO
    ELSE
       print *,"...%Mie_Rangebin_Geoloc(:) [not associated]"
    END IF

    ! print: Obs_Geoloc%Rayleigh_Rangebin_Geoloc(:)
    IF (associated(Obs_Geoloc%Rayleigh_Rangebin_Geoloc)) THEN
       s = size(Obs_Geoloc%Rayleigh_Rangebin_Geoloc)
       print "(a,i2,a)","==> size(Obs_Geoloc%Rayleigh_Rangebin_Geoloc) = ",s,&
            " [printing first and last one]"
       DO r=1,s,s-1
          print "(' Obs_Geoloc%Rayleigh_Rangebin_Geoloc(',i2.2,')%...')",r
          call Print_Rangebin_Obs_Geoloc(Obs_Geoloc%Rayleigh_Rangebin_Geoloc(r))
       END DO
    ELSE
       print *,"...%Rayleigh_Rangebin_Geoloc(:) [not associated]"
    END IF

    txt = int2txt(Obs_Geoloc%lat_of_DEM_intersection)
    print *,"...%lat_of_DEM_intersection        = ",txt," *10^-6 deg"
    txt = int2txt(Obs_Geoloc%lon_of_DEM_intersection)
    print *,"...%lon_of_DEM_intersection        = ",txt," *10^-6 deg"
    txt = real2txt(Obs_Geoloc%alt_of_DEM_intersection)
    print *,"...%alt_of_DEM_intersection        = ",txt
!    txt = real2txt(Obs_Geoloc%topoc_azim_of_DEM_intersection)
!    print *,"...%topoc_azim_of_DEM_intersection = ",txt
!    txt = real2txt(Obs_Geoloc%topoc_elev_of_DEM_intersection)
!    print *,"...%topoc_elev_of_DEM_intersection = ",txt
    txt = real2txt(Obs_Geoloc%Line_of_Sight_Velocity)
    print *,"...%Line_of_Sight_Velocity         = ",txt
    txt = real2txt(Obs_Geoloc%Geoid_Separation_of_Height_Bin)
    print *,"...%Geoid_Separation_of_Height_Bin = ",txt

  end subroutine Print_Obs_Geolocation
    !  #]
  subroutine Read_Obs_Geolocation(fileunit, Obs_Geoloc, N_range_gates, error_flag)
    !  #[
    integer,                    intent(in)    :: fileunit      ! input
    TYPE(Obs_Geolocation_type), intent(inout) :: Obs_Geoloc    ! result
    integer,                    intent(in)    :: N_range_gates ! input
    !                           (used for checking validity of the dataset !)
    integer,                    intent(out)   :: error_flag    ! result
    
    ! local variables
    integer        :: r
    integer(IntAl) :: value_IntAl
    real(FAdoxy)   :: value_FAdoxy

    error_flag = no_error

    IF (.not. associated(Obs_Geoloc%Mie_Rangebin_Geoloc)) THEN
       print *,"ERROR in Read_Obs_Geolocation()"
       print *,"the datastructure Obs_Geoloc%Mie_Rangebin_Geoloc(:) seems to be"
       print *,"not associated ...."
       error_flag = error_programming
       return
    END IF

    ! check if the struct was initialised with the correct nr of rangegates
    IF (N_range_gates .ne. size(Obs_Geoloc%Mie_Rangebin_Geoloc)) THEN
       print *,"ERROR in Read_Obs_Geolocation()"
       print *,"the datastructure Obs_Geoloc%Mie_Rangebin_Geoloc(:) has room for",&
            size(Obs_Geoloc%Mie_Rangebin_Geoloc)
       print *,"rangegates, but the nr of rangegates in the current"
       print *,"datafile seems to be ",N_range_gates
       error_flag = error_programming
       return
    END IF

    ! read the Rangebin_geoloc structure for the 24+1 Mie rangegates
    DO r=1,N_range_gates ! 24+1
       call Read_Rangebin_Obs_Geoloc(fileunit,Obs_Geoloc%Mie_Rangebin_Geoloc(r),&
                                 error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! read the Rangebin_geoloc structure for the 24+1 Rayleigh rangegates
    DO r=1,N_range_gates ! 24+1
       call Read_Rangebin_Obs_Geoloc(fileunit,Obs_Geoloc%Rayleigh_Rangebin_Geoloc(r),&
                                 error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! read lat_of_DEM_intersection
    call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    Obs_Geoloc%lat_of_DEM_intersection = value_IntAl

    ! read lon_of_DEM_intersection
    call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    Obs_Geoloc%lon_of_DEM_intersection = value_IntAl

    ! read alt_of_DEM_intersection
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Obs_Geoloc%alt_of_DEM_intersection = value_FAdoxy
    
!    ! read topoc_azim_of_DEM_intersection
!    call read_real8_from_bin_file(fileunit,value_FAdoxy,error_flag)
!    IF (error_flag .ne. no_error) return
!    Geoloc%topoc_azim_of_DEM_intersection = value_FAdoxy
!
!    ! read topoc_elev_of_DEM_intersection
!    call read_real8_from_bin_file(fileunit,value_FAdoxy,error_flag)
!    IF (error_flag .ne. no_error) return
!    Geoloc%topoc_elev_of_DEM_intersection = value_FAdoxy

    ! read Line_of_Sight_Velocity
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Obs_Geoloc%Line_of_Sight_Velocity = value_FAdoxy

    ! read Geoid_Separation_of_Height_Bin
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Obs_Geoloc%Geoid_Separation_of_Height_Bin = value_FAdoxy

    ! skip the 8 byte spare
    ! to be replaced by a call to: read_zero_bytes_bin_file
    call AdvanceFilePointer(fileunit,8_i8_,error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine Read_Obs_Geolocation
    !  #]
  subroutine Init_Meas_Geolocation(Meas_Geoloc, N_range_gates, error_flag)
    !  #[
    TYPE(Meas_Geolocation_type), intent(out) :: Meas_Geoloc
    integer,                     intent(in)  :: N_range_gates
    integer,                     intent(out) :: error_flag

    ! local variables
    integer :: AllocStatus, r

    error_flag = no_error

    IF (N_range_gates .eq. 0) THEN
       ! initialise it to be empty
       nullify(Meas_Geoloc%Mie_Rangebin_Geoloc)
       nullify(Meas_Geoloc%Rayleigh_Rangebin_Geoloc)
    ELSE
       allocate(Meas_Geoloc%Mie_Rangebin_Geoloc(N_range_gates),stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_Meas_Geolocation"
          error_flag = error_allocate
          return
       END IF
       DO r=1,N_range_gates
          call Init_Rangebin_Meas_Geoloc(Meas_Geoloc%Mie_Rangebin_Geoloc(r))
       END DO

       allocate(Meas_Geoloc%Rayleigh_Rangebin_Geoloc(N_range_gates),stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_Meas_Geolocation"
          error_flag = error_allocate
          return
       END IF
       DO r=1,N_range_gates
          call Init_Rangebin_Meas_Geoloc(Meas_Geoloc%Rayleigh_Rangebin_Geoloc(r))
       END DO
    END IF

    Meas_Geoloc%lat_of_DEM_intersection        = missing_indicator_IntAl
    Meas_Geoloc%lon_of_DEM_intersection        = missing_indicator_IntAl
    Meas_Geoloc%alt_of_DEM_intersection        = missing_indicator_FAdoxy
    Meas_Geoloc%AOCS_LOS_Velocity              = missing_indicator_FAdoxy

  end subroutine Init_Meas_Geolocation
    !  #]
  subroutine Delete_Meas_Geolocation(Meas_Geoloc)
    !  #[
    TYPE(Meas_Geolocation_type), intent(inout) :: Meas_Geoloc

    IF (associated(Meas_Geoloc%Mie_Rangebin_Geoloc)) THEN
       deallocate(Meas_Geoloc%Mie_Rangebin_Geoloc)
       nullify(Meas_Geoloc%Mie_Rangebin_Geoloc)
    END IF

    IF (associated(Meas_Geoloc%Rayleigh_Rangebin_Geoloc)) THEN
       deallocate(Meas_Geoloc%Rayleigh_Rangebin_Geoloc)
       nullify(Meas_Geoloc%Rayleigh_Rangebin_Geoloc)
    END IF

  end subroutine Delete_Meas_Geolocation
    !  #]
  subroutine Print_Meas_Geolocation(Meas_Geoloc)
    !  #[
    TYPE(Meas_Geolocation_type), intent(in) :: Meas_Geoloc

    ! local variable
    integer           :: s, r
    character(len=13) :: txt

    ! print: Meas_Geoloc%Mie_Rangebin_Geoloc(:)
    IF (associated(Meas_Geoloc%Mie_Rangebin_Geoloc)) THEN
       s = size(Meas_Geoloc%Mie_Rangebin_Geoloc)
       print "(a,i2,a)","==> size(Meas_Geoloc%Mie_Rangebin_Geoloc)      = ",s,&
            " [printing first and last one]"
       DO r=1,s,s-1
          print "(' Meas_Geoloc%Mie_Rangebin_Geoloc(',i2.2,')%...')",r
          call Print_Rangebin_Meas_Geoloc(Meas_Geoloc%Mie_Rangebin_Geoloc(r))
       END DO
    ELSE
       print *,"...%Mie_Rangebin_Geoloc(:) [not associated]"
    END IF

    ! print: Meas_Geoloc%Rayleigh_Rangebin_Geoloc(:)
    IF (associated(Meas_Geoloc%Rayleigh_Rangebin_Geoloc)) THEN
       s = size(Meas_Geoloc%Rayleigh_Rangebin_Geoloc)
       print "(a,i2,a)","==> size(Meas_Geoloc%Rayleigh_Rangebin_Geoloc) = ",s,&
            " [printing first and last one]"
       DO r=1,s,s-1
          print "(' Meas_Geoloc%Rayleigh_Rangebin_Geoloc(',i2.2,')%...')",r
          call Print_Rangebin_Meas_Geoloc(Meas_Geoloc%Rayleigh_Rangebin_Geoloc(r))
       END DO
    ELSE
       print *,"...%Rayleigh_Rangebin_Geoloc(:) [not associated]"
    END IF

    txt = int2txt(Meas_Geoloc%lat_of_DEM_intersection)
    print *,"...%lat_of_DEM_intersection        = ",txt," *10^-6 deg"
    txt = int2txt(Meas_Geoloc%lon_of_DEM_intersection)
    print *,"...%lon_of_DEM_intersection        = ",txt," *10^-6 deg"
    txt = real2txt(Meas_Geoloc%alt_of_DEM_intersection)
    print *,"...%alt_of_DEM_intersection        = ",txt
    txt = real2txt(Meas_Geoloc%AOCS_LOS_Velocity)
    print *,"...%AOCS_LOS_Velocity              = ",txt

  end subroutine Print_Meas_Geolocation
    !  #]
  subroutine Read_Meas_Geolocation(fileunit, Meas_Geoloc, N_range_gates, error_flag)
    !  #[
    integer,                     intent(in)    :: fileunit       ! input
    TYPE(Meas_Geolocation_type), intent(inout) :: Meas_Geoloc    ! result
    integer,                      intent(in)    :: N_range_gates ! input
    !                            (used for checking validity of the dataset !)
    integer,                     intent(out)   :: error_flag     ! result
    
    ! local variables
    integer        :: r
    integer(IntAl) :: value_IntAl
    real(FAdoxy)   :: value_FAdoxy

    error_flag = no_error

    IF (.not. associated(Meas_Geoloc%Mie_Rangebin_Geoloc)) THEN
       print *,"ERROR in Read_Meas_Geolocation()"
       print *,"the datastructure Meas_Geoloc%Mie_Rangebin_Geoloc(:) seems to be"
       print *,"not associated ...."
       error_flag = error_programming
       return
    END IF

    ! check if the struct was initialised with the correct nr of rangegates
    IF (N_range_gates .ne. size(Meas_Geoloc%Mie_Rangebin_Geoloc)) THEN
       print *,"ERROR in Read_Meas_Geolocation()"
       print *,"the datastructure Meas_Geoloc%Mie_Rangebin_Geoloc(:) has room for",&
            size(Meas_Geoloc%Mie_Rangebin_Geoloc)
       print *,"rangegates, but the nr of rangegates in the current"
       print *,"datafile seems to be ",N_range_gates
       error_flag = error_programming
       return
    END IF

    ! read the Rangebin_geoloc structure for the 24+1 Mie rangegates
    DO r=1,N_range_gates ! 24+1
       call Read_Rangebin_Meas_Geoloc(fileunit, &
                                 Meas_Geoloc%Mie_Rangebin_Geoloc(r),&
                                 error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! read the Rangebin_geoloc structure for the 24+1 Rayleigh rangegates
    DO r=1,N_range_gates ! 24+1
       call Read_Rangebin_Meas_Geoloc(fileunit, &
                                 Meas_Geoloc%Rayleigh_Rangebin_Geoloc(r),&
                                 error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! read lat_of_DEM_intersection
    call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    Meas_Geoloc%lat_of_DEM_intersection = value_IntAl

    ! read lon_of_DEM_intersection
    call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    Meas_Geoloc%lon_of_DEM_intersection = value_IntAl

    ! read alt_of_DEM_intersection
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Meas_Geoloc%alt_of_DEM_intersection = value_FAdoxy
    
    ! read AOCS_LOS_Velocity
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Meas_Geoloc%AOCS_LOS_Velocity = value_FAdoxy

    ! skip the 8 byte spare
    ! to be replaced by a call to: read_zero_bytes_bin_file
    call AdvanceFilePointer(fileunit,8_i8_,error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine Read_Meas_Geolocation
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_L1B_Geoloc_ADS(Geoloc_ADS, N_meas, N_range_gates, error_flag)
    !  #[
    TYPE(L1B_Geoloc_ADS_type), pointer     :: Geoloc_ADS
    integer,                   intent(in)  :: N_meas        ! input
    integer,                   intent(in)  :: N_range_gates ! input
    integer,                   intent(out) :: error_flag    ! output

    ! local variable
    integer :: m, AllocStatus

    error_flag = no_error

    IF (.not. associated(Geoloc_ADS)) THEN
       print *,"ERROR in Init_L1B_Geoloc_ADS: "
       print *,"pointer Geoloc_ADS seems not associated !!!"
       error_flag = error_programming
       return
    END IF

    call InitDateTime(Geoloc_ADS%start_of_obs_datetime)
    call Init_AOCS(Geoloc_ADS%Obs_AOCS)

    IF (N_meas .eq. 0) THEN
       ! initialise it to be empty
       nullify(Geoloc_ADS%Meas_AOCS)
    ELSE
       allocate(Geoloc_ADS%Meas_AOCS(N_meas),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L1B_Geoloc_ADS"
          error_flag = error_allocate
          return
       END IF
    
       DO m=1,N_meas
          call Init_AOCS(Geoloc_ADS%Meas_AOCS(m))
       END DO
    END IF

    call Init_Obs_Geolocation(Geoloc_ADS%Obs_Geoloc, N_range_gates, error_flag)
    IF (error_flag .ne. no_error) return

    IF (N_meas .eq. 0) THEN
       ! initialise it to be empty
       nullify(Geoloc_ADS%Meas_Geoloc)
    ELSE
       allocate(Geoloc_ADS%Meas_Geoloc(N_meas),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L1B_Geoloc_ADS"
          error_flag = error_allocate
          return
       END IF
       
       DO m=1,N_meas
          call Init_Meas_Geolocation(Geoloc_ADS%Meas_Geoloc(m), N_range_gates, &
                                error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    END IF

  end subroutine Init_L1B_Geoloc_ADS
    !  #]
  subroutine Delete_L1B_Geoloc_ADS(Geoloc_ADS)
    !  #[
    TYPE(L1B_Geoloc_ADS_type), pointer :: Geoloc_ADS

    ! local variable
    integer :: m

    IF (.not. associated(Geoloc_ADS)) return

    IF (associated(Geoloc_ADS%Meas_AOCS)) THEN
       deallocate(Geoloc_ADS%Meas_AOCS)
       nullify(Geoloc_ADS%Meas_AOCS)
    END IF

    call Delete_Obs_Geolocation(Geoloc_ADS%Obs_Geoloc)

    IF (associated(Geoloc_ADS%Meas_Geoloc)) THEN
       DO m=1,size(Geoloc_ADS%Meas_Geoloc) 
          call Delete_Meas_Geolocation(Geoloc_ADS%Meas_Geoloc(m))
       END DO
       deallocate(Geoloc_ADS%Meas_Geoloc)
       nullify(Geoloc_ADS%Meas_Geoloc)
    END IF

  end subroutine Delete_L1B_Geoloc_ADS
    !  #]
  subroutine Print_L1B_Geoloc_ADS(Geoloc_ADS)
    !  #[
    TYPE(L1B_Geoloc_ADS_type), pointer :: Geoloc_ADS

    ! local variables
    integer :: m, s
    character(len=maxlen_DateTime_KVT) :: DateTimeString

    IF (associated(Geoloc_ADS)) THEN
       print *,"Geoloc_ADS%..."
    ELSE
       print *,"Geoloc_ADS [seems not associated]"
       return
    END IF

    ! print the date and time
    DateTimeString = Construct_KVT_DateTime(Geoloc_ADS%start_of_obs_datetime)
    print *,"...%start_of_obs_datetime: "//trim(DateTimeString)

    print *,"Geoloc_ADS%Obs_AOCS%..."
    call Print_AOCS(Geoloc_ADS%Obs_AOCS)

    IF (associated(Geoloc_ADS%Meas_AOCS)) THEN
       s = size(Geoloc_ADS%Meas_AOCS)
       print "(a,i2,a)","==> size(Geoloc_ADS%Meas_AOCS)   = ",s,&
            " [printing first and last one]"
       DO m=1,s,s-1
          print "(' Geoloc_ADS%Meas_AOCS(',i2.2,')%...')",m
          call Print_AOCS(Geoloc_ADS%Meas_AOCS(m))
       END DO
    ELSE
       print *,"Geoloc_ADS%Meas_AOCS(:) = [NOT ASSOCIATED]"
    END IF

    print *,"Geoloc_ADS%Obs_Geoloc%..."
    call Print_Obs_Geolocation(Geoloc_ADS%Obs_Geoloc)

    IF (associated(Geoloc_ADS%Meas_Geoloc)) THEN
       s = size(Geoloc_ADS%Meas_Geoloc)
       print "(a,i2,a)","==> size(Geoloc_ADS%Meas_Geoloc) = ",s,&
            " [printing first and last one]"
       DO m=1,s,s-1
          print "(' Geoloc_ADS%Meas_Geoloc(',i2.2,')%...')",m
          call Print_Meas_Geolocation(Geoloc_ADS%Meas_Geoloc(m))
       END DO

    ELSE
       print *,"Geoloc_ADS%Meas_Geoloc(:) = [NOT ASSOCIATED]"
    END IF

  end subroutine Print_L1B_Geoloc_ADS
    !  #]
  !---------------------------------------------------------------------
  subroutine Read_L1B_Geoloc_ADS(Geoloc_DSD,fileunit,which_DS,N_Meas_Max,&
                                 N_range_gates, Geoloc_ADS,error_flag)
    !  #[

    TYPE(DSD_type),            pointer     :: Geoloc_DSD    ! input
    integer,                   intent(in)  :: fileunit      ! input
    integer,                   intent(in)  :: which_DS      ! input
    integer,                   intent(in)  :: N_Meas_Max    ! input
    !                       (used for checking validity of the dataset !)
    integer,                   intent(in)  :: N_range_gates ! input
    !                       (used for checking validity of the dataset !)
    TYPE(L1B_Geoloc_ADS_type), pointer     :: Geoloc_ADS    ! result
    integer,                   intent(out) :: error_flag    ! result

    ! local variables
    integer           :: m
    integer(i8_)      :: pos, endpos
    integer           :: currentpos

    IF (.not. associated(Geoloc_DSD)) THEN
       print *,"ERROR in Read_L1B_Geoloc_ADS:"
       print *,"the pointer Geoloc_DSD seems not associated"
       error_flag = error_programming
       return
    END IF

    IF (.not. associated(Geoloc_ADS)) THEN
       print *,"ERROR in Read_L1B_Geoloc_ADS:"
       print *,"the pointer Geoloc_ADS seems not associated"
       error_flag = error_programming
       return
    END IF

    pos = Geoloc_DSD%start_of_ds + (which_DS-1)*Geoloc_DSD%size_of_one_dsr
    call SetFilePointer(fileunit,pos,error_flag)
    IF (error_flag .ne. no_error) return

    print *,"trying to read the binary Geoloc_ADS nr. ",which_DS

    ! read the 12 byte datetime field
    call read_DateTime_from_bin_file(fileunit,&
         Geoloc_ADS%start_of_obs_datetime,error_flag)

    ! read the AOCS structure for the observation
    call Read_AOCS(fileunit,Geoloc_ADS%Obs_AOCS,error_flag)

    IF (.not. associated(Geoloc_ADS%Meas_AOCS)) THEN
       print *,"ERROR in Read_L1B_Geoloc_ADS()"
       print *,"the datastructure L1B%Geoloc_ADS%Meas_AOCS(:) seems to be"
       print *,"not associated ...."
       error_flag = error_programming
       return
    END IF

    ! check if the struct was initialised with the correct nr of measurements
    IF (N_meas_Max .ne. size(Geoloc_ADS%Meas_AOCS)) THEN
       print *,"ERROR in Read_L1B_Geoloc_ADS()"
       print *,"the datastructure L1B%Geoloc_ADS%Meas_AOCS(:) has room for ",&
            size(Geoloc_ADS%Meas_AOCS)
       print *,"measurements, but the nr of measurements in the current"
       print *,"datafile seems to be ",N_meas_Max
       error_flag = error_programming
       return
    END IF

    ! read the AOCS structure for the 67 measurements
    DO m=1,N_Meas_Max ! 67
       call Read_AOCS(fileunit,Geoloc_ADS%Meas_AOCS(m),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! read the Obs_Geoloc: 848/1306 bytes
    call Read_Obs_Geolocation(fileunit, Geoloc_ADS%Obs_Geoloc, &
                          N_range_gates, error_flag)
    IF (error_flag .ne. no_error) return

     IF (.not. associated(Geoloc_ADS%Meas_Geoloc)) THEN
       print *,"ERROR in Read_L1B_Geoloc_ADS()"
       print *,"the datastructure L1B%Geoloc_ADS%Meas_Geoloc(:) seems to be"
       print *,"not associated ...."
       error_flag = error_programming
       return
    END IF

    ! check if the struct was initialised with the correct nr of measurements
    IF (N_meas_Max .ne. size(Geoloc_ADS%Meas_Geoloc)) THEN
       print *,"ERROR in Read_L1B_Geoloc_ADS()"
       print *,"the datastructure L1B%Geoloc_ADS%Meas_Geoloc(:) has room for",&
            size(Geoloc_ADS%Meas_Geoloc)
       print *,"measurements, but the nr of measurements in the current"
       print *,"datafile seems to be ",N_meas_Max
       error_flag = error_programming
       return
    END IF

    ! read the Meas_Geoloc structure for the 67 measurements
    DO m=1,N_Meas_Max ! 67
       call Read_Meas_Geolocation(fileunit,Geoloc_ADS%Meas_Geoloc(m), &
                             N_range_gates, error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! calculate expected end-position after reading this block of data
    endpos = Geoloc_DSD%start_of_ds + which_DS*Geoloc_DSD%size_of_one_dsr
    ! request the current file position
    inquire(unit=fileunit,nextrec=currentpos)
    
    IF (currentpos .ne. endpos) THEN
       print *,"ERROR in Read_L1B_Geoloc_ADS():"
       print *,"there seems to be a problem while reading this binary dataset"
       print *,"expected file position after this DSR is: ",endpos
       print *,"but the actual position seems to be:      ",currentpos
       print *,"Maybe the fileformat definition has changed ???"
       error_flag = error_reading_file
       return
    END IF

  end subroutine Read_L1B_Geoloc_ADS
    !  #]
  !---------------------------------
end module L1B_Geoloc_ADS
