module L1B_GWD_ADS
  !  #[ Documentation
  !     a module to handle the ground wind detection/correction 
  !     annotation dataset
  !
  !     See also the L1BP files: 
  !          L1bP_Draft_1/aeolus/common/PFR/src/
  !             PFR_LevelOneBProductGroundWindDetectionDSR.cc
  !             PFR_LevelOneBProductGroundWindDetectionDSR.hh
  !     for the code that defines this datablock in the L1BP
  !
  !     Written by J. de Kloe, KNMI 
  !     created:     07-07-2005
  ! 
  !     Modifications:
  !      10-Jan-2006 J. de Kloe new DateTime type
  !      28-Mar-2006 P. Poli    Upgrade to DLR IODD AE.TN.DLR.APM-L1B.150206
  !      12-May-2006 D. Tan     Null use of ParKind1 module
  !      03-Nov-2006 P. Poli    upgrade to L1BP v1.04 ;
  !                             moved and cleaned-up reading part
  !      03-Jan-2008 J. de Kloe added nullify statements to make the code
  !                             more robust
  !      16-Jan-2008 J. de Kloe phase out integer kind i_ 
  !      21-Jan-2008 J. de Kloe following IODD v1.08 change:
  !                             Mie_ch_tot_zero_wind_corr  from IntAl to FAdoxy
  !                             Rayl_ch_tot_zero_wind_corr from IntAl to FAdoxy
  !                             and implemented a workaround to allow reading 
  !                             the old format for now.
  !  #]
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE Numerics, only: i8_, real2txt
  USE DataSetDescriptor, only: DSD_type
  USE EE_CFI_Datatypes, only: DateTimeType,Construct_KVT_DateTime, InitDateTime,  &
                              read_DateTime_from_bin_file, &
                              read_IntAl_from_bin_file, IntAl, missing_indicator_IntAl, &
                              read_IntAuc_from_bin_file, IntAuc, missing_indicator_IntAuc, &
                              read_FAdoxy_from_bin_file, FAdoxy, missing_indicator_FAdoxy, &
                              int2txt, maxlen_DateTime_KVT
  USE DirectBinaryIO, only: SetFilePointer, AdvanceFilePointer, read_logical1_from_bin_file
  USE ErrorHandler, only: no_error, error_programming, error_allocate, error_reading_file
  !  #]
  !  #[ type definitions
  implicit none

  TYPE Validation_Criteria_type ! see sec. 6.3.3, table 6-13, of L1B_iodd
     integer(IntAuc) :: min_n_Mie_grnd_echo_meas_bins
     integer(IntAuc) :: min_n_Rayl_grnd_echo_meas_bins
     integer(IntAuc) :: nr_of_Mie_ground_bins
     integer(IntAuc) :: nr_of_Rayleigh_ground_bins
     !              1234567890123456789012345678901
  END TYPE Validation_Criteria_type

  TYPE Meas_GWD_type
     ! Mie part
     logical         :: Mie_grnd_wind_detected
     integer(IntAuc) :: Mie_grnd_bin_number
     real(FAdoxy)    :: Mie_grnd_bin_thick_ab_DEM
     ! Rayleigh part
     logical         :: Rayleigh_grnd_wind_detected
     integer(IntAuc) :: Rayleigh_grnd_bin_number
     real(FAdoxy)    :: Rayleigh_grnd_bin_thick_ab_DEM
     !              1234567890123456789012345678901
  END TYPE Meas_GWD_type

  TYPE L1B_GWD_ADS_type ! see sec. 6.3.3, table 6-12, of L1B_iodd
     type(DateTimeType) :: start_of_obs_datetime           ! UTC
     real(FAdoxy)       :: Mie_ground_corr_velocity        ! [m/s]
     real(FAdoxy)       :: Rayleigh_ground_corr_velocity   ! [m/s]
     logical            :: updated_Mie_ground_corr_vel     ! flag
     logical            :: updated_Rayleigh_grnd_corr_vel  ! flag
     real(FAdoxy)       :: Mie_avg_grnd_wind_bin_alt       ! [m]
     real(FAdoxy)       :: Rayleigh_avg_grnd_wind_bin_alt  ! [m]
     real(FAdoxy)       :: Mie_avg_grnd_wind_bin_thick     ! [m]
     real(FAdoxy)       :: Rayleigh_avg_grnd_wnd_bin_thick ! [m]
     TYPE(Validation_Criteria_type) :: Validation_Criteria 
     TYPE(Meas_GWD_type), dimension(:), pointer :: Meas_GWD
     real(FAdoxy)       :: Mie_grnd_corr_wght_factor       ! [no unit]
     real(FAdoxy)       :: Rayleigh_grnd_corr_wght_factor  ! [no unit]
     ! should these next 2 not have [m/s] as unit?
     real(FAdoxy)       :: Mie_ch_tot_zero_wind_corr       ! [no unit]
     real(FAdoxy)       :: Rayl_ch_tot_zero_wind_corr      ! [no unit
     real(FAdoxy)       :: HBE_Mie_grnd_corr_vel           ! [m/s]
     real(FAdoxy)       :: HBE_Rayl_grnd_corr_vel          ! [m/s]
     ! spare_1 (16 bytes)
     !                     1234567890123456789012345678901
  END TYPE L1B_GWD_ADS_type
   !  #]
contains
  !---------------------------------------------------------------------
  subroutine Init_Meas_GWD(Meas_GWD)
    !  #[
    TYPE(Meas_GWD_type), intent(out) :: Meas_GWD
    ! Mie part
    Meas_GWD%Mie_grnd_wind_detected    = .false.
    Meas_GWD%Mie_grnd_bin_number       = missing_indicator_IntAuc
    Meas_GWD%Mie_grnd_bin_thick_ab_DEM = missing_indicator_FAdoxy
    ! Rayleigh part
    Meas_GWD%Rayleigh_grnd_wind_detected    = .false.
    Meas_GWD%Rayleigh_grnd_bin_number       = missing_indicator_IntAuc
    Meas_GWD%Rayleigh_grnd_bin_thick_ab_DEM = missing_indicator_FAdoxy

  end subroutine Init_Meas_GWD
    !  #]
  subroutine Print_Meas_GWD(Meas_GWD)
    !  #[
    TYPE(Meas_GWD_type), intent(in) :: Meas_GWD
    
    ! local variables
    character(len=13) :: txt

    ! Mie part
    print *,"...%Mie_grnd_wind_detected          = ",&
         Meas_GWD%Mie_grnd_wind_detected
    txt = int2txt(Meas_GWD%Mie_grnd_bin_number,int_type="IntAuc")
    print *,"...%Mie_grnd_bin_number             = ",txt
    txt = real2txt(Meas_GWD%Mie_grnd_bin_thick_ab_DEM)
    print *,"...%Mie_grnd_bin_thick_ab_DEM       = ",txt

    ! Rayleigh part
    print *,"...%Rayleigh_grnd_wind_detected     = ",&
         Meas_GWD%Rayleigh_grnd_wind_detected
    txt = int2txt(Meas_GWD%Rayleigh_grnd_bin_number,int_type="IntAuc")
    print *,"...%Rayleigh_grnd_bin_number        = ",txt
    txt = real2txt(Meas_GWD%Rayleigh_grnd_bin_thick_ab_DEM)
    print *,"...%Rayleigh_grnd_bin_thick_ab_DEM  = ",txt

  end subroutine Print_Meas_GWD
    !  #]
  subroutine Read_Meas_GWD(fileunit,Meas_GWD,error_flag)
    !  #[

    integer,             intent(in)  :: fileunit   ! input
    TYPE(Meas_GWD_type), intent(out) :: Meas_GWD   ! result
    integer,             intent(out) :: error_flag ! result

    ! local parameters
    logical      :: value_l
    integer(IntAuc) :: value_IntAuc
    real(FAdoxy)    :: value_FAdoxy

    error_flag = no_error

    ! read Mie_grnd_wind_detected
    call read_logical1_from_bin_file(fileunit,value_l,error_flag)
    IF (error_flag .ne. no_error) THEN
       print *,"element: Meas_GWD%Mie_grnd_wind_detected"
       return
    END IF
!    IF (error_flag .ne. no_error) return
    Meas_GWD%Mie_grnd_wind_detected = value_l

    ! read Mie_grnd_bin_number 
    call read_IntAuc_from_bin_file(fileunit,value_IntAuc,error_flag)
    IF (error_flag .ne. no_error) return
    Meas_GWD%Mie_grnd_bin_number = value_IntAuc

    ! read Mie_grnd_bin_thick_ab_DEM
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Meas_GWD%Mie_grnd_bin_thick_ab_DEM = value_FAdoxy

    ! read Rayleigh_grnd_wind_detected
    call read_logical1_from_bin_file(fileunit,value_l,error_flag)
    IF (error_flag .ne. no_error) THEN
       print *,"element: Meas_GWD%Rayleigh_grnd_wind_detected"
! the N=30 testfile seems to contain this error, so to be able to proceed
! testing the reading code, I'll disable this test for now ...
error_flag=no_error
!       return
    END IF
!    IF (error_flag .ne. no_error) return
    Meas_GWD%Rayleigh_grnd_wind_detected = value_l

    ! read Rayleigh_grnd_bin_number 
    call read_IntAuc_from_bin_file(fileunit,value_IntAuc,error_flag)
    IF (error_flag .ne. no_error) return
    Meas_GWD%Rayleigh_grnd_bin_number = value_IntAuc

    ! read Rayleigh_grnd_bin_thick_ab_DEM
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Meas_GWD%Rayleigh_grnd_bin_thick_ab_DEM = value_FAdoxy
    
  end subroutine Read_Meas_GWD

    !  #]
  !---------------------------------------------------------------------
  subroutine Init_L1B_GWD_ADS(GWD_ADS, N_meas, error_flag)
    !  #[
    TYPE(L1B_GWD_ADS_type), pointer     :: GWD_ADS    ! input/output
    integer,                intent(in)  :: N_meas     ! input
    integer,                intent(out) :: error_flag ! output

    ! local variable
    integer :: m, AllocStatus

    error_flag = no_error

    IF (.not. associated(GWD_ADS)) THEN
       print *,"ERROR in Init_L1B_GWD_ADS: "
       print *,"pointer GWD_ADS seems not associated !!!"
       error_flag = error_programming
       return
    END IF

    call InitDateTime(GWD_ADS%start_of_obs_datetime)

    GWD_ADS%Mie_ground_corr_velocity        = missing_indicator_FAdoxy
    GWD_ADS%Rayleigh_ground_corr_velocity   = missing_indicator_FAdoxy
    GWD_ADS%updated_Mie_ground_corr_vel     = .false.
    GWD_ADS%updated_Rayleigh_grnd_corr_vel  = .false.
    GWD_ADS%Mie_avg_grnd_wind_bin_alt       = missing_indicator_FAdoxy
    GWD_ADS%Rayleigh_avg_grnd_wind_bin_alt  = missing_indicator_FAdoxy
    GWD_ADS%Mie_avg_grnd_wind_bin_thick     = missing_indicator_FAdoxy
    GWD_ADS%Rayleigh_avg_grnd_wnd_bin_thick = missing_indicator_FAdoxy
    ! validation criteria
    GWD_ADS%Validation_Criteria%min_n_Mie_grnd_echo_meas_bins  = &
         missing_indicator_IntAuc
    GWD_ADS%Validation_Criteria%min_n_Rayl_grnd_echo_meas_bins = &
         missing_indicator_IntAuc
    GWD_ADS%Validation_Criteria%nr_of_Mie_ground_bins          = &
         missing_indicator_IntAuc
    GWD_ADS%Validation_Criteria%nr_of_Rayleigh_ground_bins     = &
         missing_indicator_IntAuc

    IF (N_meas .eq. 0) THEN
       ! init it to be empty
       nullify(GWD_ADS%Meas_GWD)
    ELSE
       allocate(GWD_ADS%Meas_GWD(1:N_meas), STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L1B_GWD_ADS"
          error_flag = error_allocate
          return
       END IF
       DO m=1,N_meas
          call Init_Meas_GWD(GWD_ADS%Meas_GWD(m))
       END DO
    END IF

    GWD_ADS%Mie_grnd_corr_wght_factor       = missing_indicator_FAdoxy
    GWD_ADS%Rayleigh_grnd_corr_wght_factor  = missing_indicator_FAdoxy
    GWD_ADS%Mie_ch_tot_zero_wind_corr       = missing_indicator_FAdoxy
    GWD_ADS%Rayl_ch_tot_zero_wind_corr      = missing_indicator_FAdoxy
    GWD_ADS%HBE_Mie_grnd_corr_vel           = missing_indicator_FAdoxy
    GWD_ADS%HBE_Rayl_grnd_corr_vel          = missing_indicator_FAdoxy

  end subroutine Init_L1B_GWD_ADS
    !  #]
  subroutine Delete_L1B_GWD_ADS(GWD_ADS)
    !  #[
    TYPE(L1B_GWD_ADS_type), pointer :: GWD_ADS

    IF (.not. associated(GWD_ADS)) return

    IF (associated(GWD_ADS%Meas_GWD)) THEN
       deallocate(GWD_ADS%Meas_GWD)
       nullify(GWD_ADS%Meas_GWD)
    END IF

  end subroutine Delete_L1B_GWD_ADS
    !  #]
  subroutine Print_L1B_GWD_ADS(GWD_ADS)
    !  #[
    TYPE(L1B_GWD_ADS_type), pointer :: GWD_ADS

    ! local variables
    integer           :: m, s
    character(len=13) :: txt
    character(len=maxlen_DateTime_KVT) :: DateTimeString

    IF (.not. associated(GWD_ADS)) THEN
       print *,"GWD_ADS [seems not associated]"
       return
    END IF

    print *,"GWD_ADS%..."

    ! print the date and time
    
    DateTimeString = Construct_KVT_DateTime(GWD_ADS%start_of_obs_datetime)
    print *,"...%start_of_obs_datetime: "//trim(DateTimeString)

    txt = real2txt(GWD_ADS%Mie_ground_corr_velocity)
    print *,"...%Mie_ground_corr_velocity        = ",txt
    txt = real2txt(GWD_ADS%Rayleigh_ground_corr_velocity)
    print *,"...%Rayleigh_ground_corr_velocity   = ",txt

    print *,"...%updated_Mie_ground_corr_vel     = ",&
         GWD_ADS%updated_Mie_ground_corr_vel
    print *,"...%updated_Rayleigh_grnd_corr_vel  = ",&
         GWD_ADS%updated_Rayleigh_grnd_corr_vel

    txt = real2txt(GWD_ADS%Mie_avg_grnd_wind_bin_alt)
    print *,"...%Mie_avg_grnd_wind_bin_alt       = ",txt
    txt = real2txt(GWD_ADS%Rayleigh_avg_grnd_wind_bin_alt)
    print *,"...%Rayleigh_avg_grnd_wind_bin_alt  = ",txt
    txt = real2txt(GWD_ADS%Mie_avg_grnd_wind_bin_thick)
    print *,"...%Mie_avg_grnd_wind_bin_thick     = ",txt
    txt = real2txt(GWD_ADS%Rayleigh_avg_grnd_wnd_bin_thick)
    print *,"...%Rayleigh_avg_grnd_wnd_bin_thick = ",txt

    print *,"GWD_ADS%Validation_Criteria%..."

    txt = int2txt(GWD_ADS%Validation_Criteria%min_n_Mie_grnd_echo_meas_bins,int_type="IntAuc")
    print *,"...%min_n_Mie_grnd_echo_meas_bins   = ",txt
    txt = int2txt(GWD_ADS%Validation_Criteria%min_n_Rayl_grnd_echo_meas_bins,int_type="IntAuc")
    print *,"...%min_n_Rayl_grnd_echo_meas_bins  = ",txt
    txt = int2txt(GWD_ADS%Validation_Criteria%nr_of_Mie_ground_bins,int_type="IntAuc")
    print *,"...%nr_of_Mie_ground_bins           = ",txt
    txt = int2txt(GWD_ADS%Validation_Criteria%nr_of_Rayleigh_ground_bins,int_type="IntAuc")
    print *,"...%nr_of_Rayleigh_ground_bins      = ",txt

    IF (associated(GWD_ADS%Meas_GWD)) THEN
       s = size(GWD_ADS%Meas_GWD)
       print "(a,i2,a)","==> size(GWD_ADS%Meas_GWD) = ",s,&
            " [printing first and last one]"
       DO m=1,s,s-1
          print "(' GWD_ADS%Meas_GWD(',i2.2,')%...')",m
          call Print_Meas_GWD(GWD_ADS%Meas_GWD(m))
       END DO
    ELSE
       print *,"GWD_ADS%Meas_GWD(:) = [NOT ASSOCIATED]"
    END IF

    print *,"GWD_ADS%... (continued)"

    txt = real2txt(GWD_ADS%Mie_grnd_corr_wght_factor)
    print *,"...%Mie_grnd_corr_wght_factor       = ",txt
    txt = real2txt(GWD_ADS%Rayleigh_grnd_corr_wght_factor)
    print *,"...%Rayleigh_grnd_corr_wght_factor  = ",txt
    txt = real2txt(GWD_ADS%Mie_ch_tot_zero_wind_corr)
    print *,"...%Mie_ch_tot_zero_wind_corr       = ",txt
    txt = real2txt(GWD_ADS%Rayl_ch_tot_zero_wind_corr)
    print *,"...%Rayl_ch_tot_zero_wind_corr      = ",txt
    txt = real2txt(GWD_ADS%HBE_Mie_grnd_corr_vel)
    print *,"...%HBE_Mie_grnd_corr_vel           = ",txt
    txt = real2txt(GWD_ADS%HBE_Rayl_grnd_corr_vel)
    print *,"...%HBE_Rayl_grnd_corr_vel          = ",txt

  end subroutine Print_L1B_GWD_ADS
    !  #]
  subroutine Read_L1B_GWD_ADS(GWD_DSD,fileunit,which_DS,N_Meas_Max,&
                              GWD_ADS,error_flag,workaround_pre108)
    !  #[
    TYPE(DSD_type),         pointer     :: GWD_DSD      ! input
    integer,                intent(in)  :: fileunit     ! input
    integer,                intent(in)  :: which_DS     ! input
    integer,                intent(in)  :: N_Meas_Max   ! input 
    !                       (used for checking validity of the dataset !)
    TYPE(L1B_GWD_ADS_type), pointer     :: GWD_ADS      ! result
    integer,                intent(out) :: error_flag   ! result
    logical,                intent(in)  :: workaround_pre108 ! input
   
    ! local variables
    integer(i8_)      :: pos, endpos
    integer           :: currentpos
    real(FAdoxy)      :: value_FAdoxy
    logical           :: value_l
    integer           :: m
    integer(IntAl)    :: value_IntAl

    ! See the files: L1bP_Draft_1/aeolus/common/PFR/src/
    !        PFR_LevelOneBProductGroundWindDetectionDSR.cc/.hh       
    ! for the code that defines this datablock in the L1BP

    IF (.not. associated(GWD_DSD)) THEN
       print *,"ERROR in Read_L1B_GWD_ADS:"
       print *,"the pointer GWD_DSD seems not associated"
       error_flag = error_programming
       return
    END IF

    IF (.not. associated(GWD_ADS)) THEN
       print *,"ERROR in Read_L1B_GWD_ADS:"
       print *,"the pointer GWD_ADS seems not associated"
       error_flag = error_programming
       return
    END IF

    pos = GWD_DSD%start_of_ds + (which_DS-1)*GWD_DSD%size_of_one_dsr
    call SetFilePointer(fileunit,pos,error_flag)
    IF (error_flag .ne. no_error) return

    print *,"trying to read the binary GWD_ADS nr. ",which_DS

    ! read the 12 byte datetime field
    call read_DateTime_from_bin_file(fileunit,&
         GWD_ADS%start_of_obs_datetime,error_flag)

    ! read Mie_ground_corr_velocity
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    GWD_ADS%Mie_ground_corr_velocity = value_FAdoxy

    ! read Rayleigh_ground_corr_velocity
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    GWD_ADS%Rayleigh_ground_corr_velocity = value_FAdoxy

    ! read updated_Mie_ground_corr_vel
    call read_logical1_from_bin_file(fileunit,value_l,error_flag)
!    IF (error_flag .ne. no_error) return
    IF (error_flag .ne. no_error) THEN
       print *,"element: GWD_ADS%updated_Mie_ground_corr_vel"
       return
    END IF
    GWD_ADS%updated_Mie_ground_corr_vel = value_l

    ! read updated_Mie_ground_corr_vel
    call read_logical1_from_bin_file(fileunit,value_l,error_flag)
    IF (error_flag .ne. no_error) THEN
       print *,"element: GWD_ADS%updated_Rayleigh_ground_corr_vel"
       return
    END IF
!    IF (error_flag .ne. no_error) return
    GWD_ADS%updated_Rayleigh_grnd_corr_vel = value_l

    ! read Mie_avg_grnd_wind_bin_alt
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    GWD_ADS%Mie_avg_grnd_wind_bin_alt = value_FAdoxy

    ! read Rayleigh_avg_grnd_wind_bin_alt
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    GWD_ADS%Rayleigh_avg_grnd_wind_bin_alt = value_FAdoxy

    ! read Mie_avg_grnd_wind_bin_thick
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    GWD_ADS%Mie_avg_grnd_wind_bin_thick = value_FAdoxy

    ! read Rayleigh_avg_grnd_wnd_bin_thick
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    GWD_ADS%Rayleigh_avg_grnd_wnd_bin_thick = value_FAdoxy

    ! read the validation criteria (12 bytes)
    call Read_Validation_Criteria(fileunit, &
         GWD_ADS%Validation_Criteria,error_flag)
    IF (error_flag .ne. no_error) return

    IF (.not. associated(GWD_ADS%Meas_GWD)) THEN
       print *,"ERROR in Read_L1B_GWD_ADS()"
       print *,"the datastructure L1B%GWD_ADS%Meas_GWD(:) seems to be"
       print *,"not associated ...."
       error_flag = error_programming
       return
    END IF

    ! check if the struct was initialised with the correct nr of measurements
    IF (N_meas_Max .ne. size(GWD_ADS%Meas_GWD)) THEN
       print *,"ERROR in Read_L1B_GWD_ADS()"
       print *,"the datastructure L1B%GWD_ADS%Meas_GWD(:) has room for ",&
            size(GWD_ADS%Meas_GWD)
       print *,"measurements, but the nr of measurements in the current"
       print *,"datafile seems to be ",N_meas_Max
       error_flag = error_programming
       return
    END IF

    ! 67 structs of 20 bytes each, total: 1340 bytes
    DO m=1,N_Meas_Max ! 67
       call Read_Meas_GWD(fileunit,GWD_ADS%Meas_GWD(m),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! read Mie_grnd_corr_wght_factor
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    GWD_ADS%Mie_grnd_corr_wght_factor = value_FAdoxy

    ! read Rayleigh_grnd_corr_wght_factor
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    GWD_ADS%Rayleigh_grnd_corr_wght_factor = value_FAdoxy

    ! read Mie_ch_tot_zero_wind_corr
    if (workaround_pre108) then
       call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
       IF (error_flag .ne. no_error) return
       GWD_ADS%Mie_ch_tot_zero_wind_corr = real(value_IntAl,FAdoxy)
    else
       call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
       IF (error_flag .ne. no_error) return
       GWD_ADS%Mie_ch_tot_zero_wind_corr = value_FAdoxy
    end if

    ! read Rayl_ch_tot_zero_wind_corr
    if (workaround_pre108) then
       call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
       IF (error_flag .ne. no_error) return
       GWD_ADS%Rayl_ch_tot_zero_wind_corr = real(value_IntAl,FAdoxy)
    else
       call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
       IF (error_flag .ne. no_error) return
       GWD_ADS%Rayl_ch_tot_zero_wind_corr = value_FAdoxy
    end if

    ! read HBE_Mie_grnd_corr_vel
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    GWD_ADS%HBE_Mie_grnd_corr_vel = value_FAdoxy

    ! read HBE_Rayl_grnd_corr_vel
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    GWD_ADS%HBE_Rayl_grnd_corr_vel = value_FAdoxy

    ! skip the 16 byte spare
    ! to be replaced by a call to: read_zero_bytes_bin_file
    call AdvanceFilePointer(fileunit,16_i8_,error_flag)
    IF (error_flag .ne. no_error) return

    ! calculate expected end-position after reading this block of data
    endpos = GWD_DSD%start_of_ds + which_DS*GWD_DSD%size_of_one_dsr
    ! request the current file position
    inquire(unit=fileunit,nextrec=currentpos)
    
    IF (currentpos .ne. endpos) THEN
       print *,"ERROR in Read_L1B_GWD_ADS():"
       print *,"there seems to be a problem while reading this binary dataset"
       print *,"expected file position after this DSR is: ",endpos
       print *,"but the actual position seems to be:      ",currentpos
       print *,"Maybe the fileformat definition has changed ???"
       error_flag = error_reading_file
       return
    END IF

  end subroutine Read_L1B_GWD_ADS
    !  #]
  subroutine Read_Validation_Criteria(fileunit,Validation_Criteria,error_flag)
    !  #[

    integer,                        intent(in)  :: fileunit     ! input
    TYPE(Validation_Criteria_type), intent(out) :: Validation_Criteria ! result
    integer,                        intent(out) :: error_flag   ! result

    ! local variables
    integer(IntAuc)   :: value_IntAuc

    error_flag = no_error

    ! read min_n_Mie_grnd_echo_meas_bins
    call read_IntAuc_from_bin_file(fileunit,value_IntAuc,error_flag)
    IF (error_flag .ne. no_error) return
    Validation_Criteria%min_n_Mie_grnd_echo_meas_bins = value_IntAuc

    ! read min_n_Rayl_grnd_echo_meas_bins
    call read_IntAuc_from_bin_file(fileunit,value_IntAuc,error_flag)
    IF (error_flag .ne. no_error) return
    Validation_Criteria%min_n_Rayl_grnd_echo_meas_bins = value_IntAuc

    ! read nr_of_Mie_ground_bins
    call read_IntAuc_from_bin_file(fileunit,value_IntAuc,error_flag)
    IF (error_flag .ne. no_error) return
    Validation_Criteria%nr_of_Mie_ground_bins = value_IntAuc

    ! read nr_of_Rayleigh_ground_bins
    call read_IntAuc_from_bin_file(fileunit,value_IntAuc,error_flag)
    IF (error_flag .ne. no_error) return
    Validation_Criteria%nr_of_Rayleigh_ground_bins = value_IntAuc

    ! skip the 8 byte spare
    ! to be replaced by a call to: read_zero_bytes_bin_file
    call AdvanceFilePointer(fileunit,8_i8_,error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine Read_Validation_Criteria
    !  #]
  !---------------------------------------------------------------------
end module L1B_GWD_ADS
