module L1B_Meas_ADS
  !  #[ Documentation

  !     a module to handle the Measurement (Meas) annotation dataset
  !
  !     See also the L1BP files: 
  !          L1bP_Draft_1/aeolus/common/PFR/src/
  !             PFR_LevelOneBProductUsefulSignalDSR.cc
  !             PFR_LevelOneBProductUsefulSignalDSR.hh
  !     for the code that defines this datablock in the L1BP
  !
  !     Written by D. Tan, ECMWF
  !       based on L1B_PCD_MDS and L1B_US_MDS written by J. de Kloe, KNMI 
  !     created:     01-09-2005
  !
  !     Modifications:
  !      15-Sep-2005 D. Tan,    Fixed length for reference pulse allocation.
  !      25-Okt-2005 J. de Kloe, corrected end label for L1B_Meas_ADS_type,
  !                              removed some memory leaks, 
  !                              repaired error handling
  !      10-Jan-2006 J. de Kloe new DateTime type
  !      21-Feb-2006 P. Poli    Upgraded to DLR IODD L1B-IODD.231105
  !      28-Mar-2006 P. Poli    Upgrade to DLR IODD AE.TN.DLR.APM-L1B.150206
  !      12-May-2006 D. Tan     Null use of ParKind1 module
  !      03-Nov-2006 P. Poli    upgrade to L1BP v1.04 ;
  !                             moved and cleaned-up reading part
  !      30-Nov-2006 D. Tan     Type change IntAs to IntAus for Mie 
  !                             spectrometer counts
  !      03-Jan-2008 J. de Kloe added nullify statements to make the code
  !                             more robust
  !      16-Jan-2008 J. de Kloe phase out integer kind i_ 
  !
  !  #]
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE Numerics, only: i8_, real2txt
  USE DataSetDescriptor, only: DSD_type
  USE EE_CFI_Datatypes, only: DateTimeType,Construct_KVT_DateTime, InitDateTime,  &
                              read_DateTime_from_bin_file, &
                              read_IntAus_from_bin_file, missing_indicator_IntAus, IntAus, &
                              read_IntAul_from_bin_file, missing_indicator_IntAul, IntAul, &
                              read_IntAl_from_bin_file, missing_indicator_IntAl, IntAl, &
                              read_FAdoxy_from_bin_file, missing_indicator_FAdoxy, FAdoxy, &
                              int2txt, maxlen_DateTime_KVT
  USE DirectBinaryIO, only: SetFilePointer, AdvanceFilePointer, read_logical1_from_bin_file
  USE ErrorHandler, only: no_error, error_programming, error_allocate, error_reading_file
  !  #]
  !  #[ type definitions

  implicit none

  integer, parameter          :: max_number_of_reference_pulses = 1003

! not used ?
!  TYPE Rangebin_Meas_type
!     integer(i1_) :: Data_Quality_Flag ! 8 bits flag 
!     real(r8_)    :: Useful_Signal
!  END TYPE Rangebin_Meas_type

! not used ?
!  TYPE Meas_type
!     TYPE(Rangebin_Meas_type), dimension(:), pointer :: Mie
!     TYPE(Rangebin_Meas_type), dimension(:), pointer :: Rayleigh
!  END TYPE Meas_type

  TYPE Mie_spectrometer_count_type
  !in L1BP v1.04, the following array increased in size from 18 to 20 elements:
  !and MDA confirm that the type is IntAus, not IntAs
     integer(IntAus), dimension(20) :: spectrometer_count
  END TYPE Mie_spectrometer_count_type

  TYPE Rayl_spectrometer_count_type
     real(FAdoxy) :: spectrometer_count
  END TYPE Rayl_spectrometer_count_type

  TYPE mie_measurement_spectrum_type
     type(Mie_spectrometer_count_type), dimension(:), pointer :: Mie_spectra_alt_bin
  END TYPE mie_measurement_spectrum_type

  TYPE time_delay_type
     integer(IntAl), dimension(:), pointer :: bin_layer_integration_time
     integer(IntAl)                        :: background_integration_time
  END TYPE time_delay_type

  TYPE validity_indicator_type
     logical :: measurement_data_present
     logical :: Mie_measurement_sp_valid
     logical :: Rayleigh_measurement_sp_valid
     logical :: measurement_laser_freq_locked
     logical :: spacecraft_attitude_on_target
  END TYPE validity_indicator_type

  TYPE L1B_Meas_ADS_type ! see sec. 2.3.4, table 2-32, of L1B_iodd (DLR, 05 Jul 2005)
     type(DateTimeType)  :: start_of_obs_datetime    ! UTC
     integer(IntAul)     :: Num_of_Reference_Pulses ! Unsigned int4 stored as i8_.
     ! 4-byte spare
     type(Mie_spectrometer_count_type),  dimension(:), pointer :: Mie_reference_pulses
     type(Rayl_spectrometer_count_type), dimension(:), pointer :: Rayleigh_reference_pulses_A
     type(Rayl_spectrometer_count_type), dimension(:), pointer :: Rayleigh_reference_pulses_B
     type(mie_measurement_spectrum_type),dimension(:), pointer :: Mie_measurement_data
     type(time_delay_type)         :: Mie_time_delays
     type(time_delay_type)         :: Rayleigh_time_delays
     type(validity_indicator_type),      dimension(:), pointer :: validity_indicators

     ! not used ? also not present in the IODD version 1.0, 21-10-2005
     !TYPE(Meas_type)                        :: Obs_Meas
     !TYPE(Meas_type), dimension(:), pointer :: Meas_Meas

  END TYPE L1B_Meas_ADS_type
  !  #]
contains
  !---------------------------------------------------------------------
! not used ?
!  subroutine Print_Rangebin_Meas(Rangebin_Meas)
    !  #[
!    TYPE(Rangebin_Meas_type), intent(in) :: Rangebin_Meas
!    character(len=13) :: txt
!    print *,"...%Data_Quality_Flag = ",Rangebin_Meas%Data_Quality_Flag
!    txt = real2txt(Rangebin_Meas%Useful_Signal)
!    print *,"...%Useful_Signal     = ",txt
!!    IF (Rangebin_Meas%Data_Quality_Flag .ne. 0) &
!!         call Print_Data_Quality_Flags(Rangebin_Meas%Data_Quality_Flag)
!
!  end subroutine Print_Rangebin_Meas
    !  #]
!  subroutine Print_Data_Quality_Flags(Flags)
    !  #[
!  end subroutine Print_Data_Quality_Flags
    !  #]
  !---------------------------------------------------------------------
! not used ?
!  subroutine Print_Meas(Meas)
    !  #[
!    TYPE(Meas_type), intent(in) :: Meas
!
!    ! local variables
!    integer :: r, s
!
!    IF (associated(Meas%Mie)) THEN
!       print *,"Meas%Mie%..."
!       s = size(Meas%Mie)
!       print "(a,i2,a)","==> size(Meas%Mie) = ",s,&
!            " [printing first and last one]"
!       DO r=1,s,s-1
!          print "(' Meas%Mie(',i2.2,')%...')",r
!!          call Print_Rangebin_Meas(Meas%Mie(r))
!       END DO
!    ELSE
!       print *,"Meas%Mie [seems not associated]"
!    END IF
!
!    IF (associated(Meas%Rayleigh)) THEN
!       print *,"Meas%Rayleigh%..."
!       s = size(Meas%Rayleigh)
!       print "(a,i2,a)","==> size(Meas%Rayleigh) = ",s,&
!            " [printing first and last one]"
!       DO r=1,s,s-1
!          print "(' Meas%Rayleigh(',i2.2,')%...')",r
!!          call Print_Rangebin_Meas(Meas%Rayleigh(r))
!       END DO
!    ELSE
!       print *,"Meas%Rayleigh [seems not associated]"
!    END IF
!
!  end subroutine Print_Meas
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_L1B_Meas_ADS(Meas_ADS, N_meas, N_range_gates, error_flag)
    !  #[
    TYPE(L1B_Meas_ADS_type), pointer     :: Meas_ADS
    integer,                 intent(in)  :: N_meas        ! input
    integer,                 intent(in)  :: N_range_gates ! input
!    integer,                 intent(in)  :: N_reference_pulses ! input
    integer,                 intent(out) :: error_flag    ! output
    
    ! local variables
    integer :: AllocStatus, i

    error_flag = no_error
    IF (.not. associated(Meas_ADS) ) THEN
       print *,"ERROR in Init_L1B_Meas_ADS:"
       print *,"the pointer Meas_ADS seems not associated !!!"
       error_flag = error_programming
       return
    ENDIF

    call InitDateTime(Meas_ADS%start_of_obs_datetime)

    Meas_ADS%Num_of_Reference_Pulses = missing_indicator_IntAul

    ! ideally this test should check the (number_of_reference_pulses .eq. 0) 
    ! but since a max value is used for now, use N_meas=0 to signal
    ! that no allocation is to be done just yet
    IF (N_meas .eq. 0) THEN
       nullify(Meas_ADS%Mie_reference_pulses)
    ELSE
       ! Allocate the Mie reference pulses
!       allocate(Meas_ADS%Mie_reference_pulses(1:max_number_of_reference_pulses),&
       allocate(Meas_ADS%Mie_reference_pulses(1:N_meas),&
                stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L1B_Meas_ADS"
          error_flag = error_allocate
          return
       END IF
       
       ! Initialize the Mie reference pulses
       do i=1,N_meas
          call Init_Mie_reference_pulses(Meas_ADS%Mie_reference_pulses(i))
       enddo
    END IF

    ! ideally this test should check the (number_of_reference_pulses .eq. 0) 
    ! but since a max value is used for now, use N_meas=0 to signal
    ! that no allocation is to be done just yet
    IF (N_meas .eq. 0) THEN
       nullify(Meas_ADS%Rayleigh_reference_pulses_A)
    ELSE
       ! Allocate the Rayleigh reference pulses
!       allocate(Meas_ADS%Rayleigh_reference_pulses_A(1:max_number_of_reference_pulses),&
       allocate(Meas_ADS%Rayleigh_reference_pulses_A(1:N_meas),&
                stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L1B_Meas_ADS"
          error_flag = error_allocate
          return
       END IF
       
       ! Initialize the Rayleigh reference pulses
       do i=1,N_meas
          call Init_Rayl_reference_pulses(Meas_ADS%Rayleigh_reference_pulses_A(i))
       enddo
    END IF

    ! ideally this test should check the (number_of_reference_pulses .eq. 0) 
    ! but since a max value is used for now, use N_meas=0 to signal
    ! that no allocation is to be done just yet
    IF (N_meas .eq. 0) THEN
       nullify(Meas_ADS%Rayleigh_reference_pulses_B)
    ELSE
       ! Allocate the Rayleigh reference pulses
!       allocate(Meas_ADS%Rayleigh_reference_pulses_B(1:max_number_of_reference_pulses),&
       allocate(Meas_ADS%Rayleigh_reference_pulses_B(1:N_meas),&
                stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L1B_Meas_ADS"
          error_flag = error_allocate
          return
       END IF
       
       ! Initialize the Rayleigh reference pulses
       do i=1,N_meas
          call Init_Rayl_reference_pulses(Meas_ADS%Rayleigh_reference_pulses_B(i))
       enddo
    END IF

    IF (N_meas .eq. 0) THEN
       nullify(Meas_ADS%Mie_measurement_data)
    ELSE
       ! Allocate and initialize the Mie measurement data
       allocate(Meas_ADS%Mie_measurement_data(1:N_meas),stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L1B_Meas_ADS"
          error_flag = error_allocate
          return
       END IF
       DO i=1,N_meas
          call Init_Mie_measurement_data(Meas_ADS%Mie_measurement_data(i), &
                                         N_range_gates, error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    ENDIF

    ! Initialize the time delays - Mie channel
    call Init_Measurement_time_delay(Meas_ADS%Mie_time_delays, N_range_gates, &
                              error_flag)
    IF (error_flag .ne. no_error) return

    ! Initialize the time delays - Rayleigh channel
    call Init_Measurement_time_delay(Meas_ADS%Rayleigh_time_delays, &
                                     N_range_gates, error_flag)
    IF (error_flag .ne. no_error) return

    IF (N_meas .eq. 0) THEN
        nullify(Meas_ADS%validity_indicators)
     ELSE
        ! Allocate and initialize the validity indicators
        allocate(Meas_ADS%validity_indicators(1:N_meas),stat=AllocStatus)
        IF (AllocStatus .ne. 0) THEN
           print *,"Allocation failure in subroutine Init_Meas_ADS"
           error_flag = error_allocate
           return
        END IF
        DO i=1,N_meas
           call Init_validity_indicators(Meas_ADS%validity_indicators(i))
        END DO
     END IF

     ! should these not be initialised?
     !Obs_Meas
     !Meas_Meas

  end subroutine Init_L1B_Meas_ADS
    !  #]
  subroutine Delete_L1B_Meas_ADS(Meas_ADS)
    !  #[
    TYPE(L1B_Meas_ADS_type), pointer     :: Meas_ADS

    ! local variable
    integer :: m

    IF (.not. associated(Meas_ADS)) return  ! nothing to do

    IF (associated(Meas_ADS%Mie_reference_pulses)) THEN
       deallocate(Meas_ADS%Mie_reference_pulses)
       nullify(Meas_ADS%Mie_reference_pulses)
    END IF

    ! Rayleigh reference pulses

    IF (associated(Meas_ADS%Rayleigh_reference_pulses_A)) THEN
       deallocate(Meas_ADS%Rayleigh_reference_pulses_A)
       nullify(Meas_ADS%Rayleigh_reference_pulses_A)
    END IF
    IF (associated(Meas_ADS%Rayleigh_reference_pulses_B)) THEN
       deallocate(Meas_ADS%Rayleigh_reference_pulses_B)
       nullify(Meas_ADS%Rayleigh_reference_pulses_B)
    END IF


    ! Mie measurement data
    IF (associated(Meas_ADS%Mie_measurement_data)) THEN
       DO m=1,size(Meas_ADS%Mie_measurement_data) 
          call Delete_Mie_measurement_data(Meas_ADS%Mie_measurement_data(m))
       END DO
       deallocate(Meas_ADS%Mie_measurement_data)
       nullify(Meas_ADS%Mie_measurement_data)
    END IF

    call Delete_Measurement_time_delay(Meas_ADS%Mie_time_delays)
    call Delete_Measurement_time_delay(Meas_ADS%Rayleigh_time_delays)

    IF (associated(Meas_ADS%validity_indicators)) THEN
       deallocate(Meas_ADS%validity_indicators)
       nullify(Meas_ADS%validity_indicators)
    END IF

  end subroutine Delete_L1B_Meas_ADS
    !  #]
  subroutine Print_L1B_Meas_ADS(Meas_ADS)
    !  #[

    TYPE(L1B_Meas_ADS_type), pointer     :: Meas_ADS

     ! local variables
    integer           :: m, s
    character(len=13) :: txt
    character(len=maxlen_DateTime_KVT) :: DateTimeString

    IF (.not. associated(Meas_ADS)) THEN
       print *,"Meas_ADS [seems not associated]"
       return
    END IF

    print *,"Meas_ADS%..."
    
    ! print the date and time
    DateTimeString = Construct_KVT_DateTime(Meas_ADS%start_of_obs_datetime)
    print *,"...%start_of_obs_datetime: "//trim(DateTimeString)

    txt = int2txt(Meas_ADS%Num_of_Reference_Pulses,int_type="IntAul")
    print *,"...%Num_of_Reference_Pulses = ", txt

    ! Print the Mie reference pulses
    IF (associated(Meas_ADS%Mie_reference_pulses)) THEN
       s = size(Meas_ADS%Mie_reference_pulses)
       print "(a,i4,a)","==> size(Meas_ADS%Mie_reference_pulses) = ",s,&
            " [printing first and last one]"
       DO m=1,s,s-1
          print "(' Meas_ADS%Mie_reference_pulses(',i4.4,')%...')",m
          call Print_Mie_reference_pulses(Meas_ADS%Mie_reference_pulses(m))
       END DO
    ELSE
       print *,"Meas_ADS%Mie_reference_pulses(:) = [NOT ASSOCIATED]"
    END IF

    ! Print the Rayleigh reference pulses
    IF (associated(Meas_ADS%Rayleigh_reference_pulses_A)) THEN
       s = size(Meas_ADS%Rayleigh_reference_pulses_A)
       print "(a,i4,a)","==> size(Meas_ADS%Rayleigh_reference_pulses_A) = ",s,&
            " [printing first and last one]"
       DO m=1,s,s-1
          print "(' Meas_ADS%Rayleigh_reference_pulses_A(',i4.4,')%...')",m
          call Print_Rayl_reference_pulses(Meas_ADS%Rayleigh_reference_pulses_A(m))
       END DO
    ELSE
       print *,"Meas_ADS%Rayleigh_reference_pulses_A(:) = [NOT ASSOCIATED]"
    END IF

    IF (associated(Meas_ADS%Rayleigh_reference_pulses_B)) THEN
       s = size(Meas_ADS%Rayleigh_reference_pulses_B)
       print "(a,i4,a)","==> size(Meas_ADS%Rayleigh_reference_pulses_B) = ",s,&
            " [printing first and last one]"
       DO m=1,s,s-1
          print "(' Meas_ADS%Rayleigh_reference_pulses_B(',i4.4,')%...')",m
          call Print_Rayl_reference_pulses(Meas_ADS%Rayleigh_reference_pulses_B(m))
       END DO
    ELSE
       print *,"Meas_ADS%Rayleigh_reference_pulses_B(:) = [NOT ASSOCIATED]"
    END IF

    ! Print the Mie measurement data
    IF (associated(Meas_ADS%Mie_measurement_data)) THEN
       s = size(Meas_ADS%Mie_measurement_data)
       print "(a,i2,a)","==> size(Meas_ADS%Mie_measurement_data) = ",s,&
            " [printing first and last one]"
       DO m=1,s,s-1
          print "(' Meas_ADS%Mie_measurement_data(',i2.2,')%...')",m
          call Print_Mie_measurement_data(Meas_ADS%Mie_measurement_data(m))
       END DO
    ELSE
       print *,"Meas_ADS%Mie_measurement_data(:) = [NOT ASSOCIATED]"
    END IF

    ! Print the time delays - Mie channel
    print *,"Meas_ADS%Mie_time_delays%..."
    call Print_Measurement_time_delay(Meas_ADS%Mie_time_delays)

    ! Print the time delays - Rayleigh channel
    print *,"Meas_ADS%Rayleigh_time_delays%..."
    call Print_Measurement_time_delay(Meas_ADS%Rayleigh_time_delays)

    ! Print the validity indicators
    IF (associated(Meas_ADS%validity_indicators)) THEN
       s = size(Meas_ADS%validity_indicators)
       print "(a,i2,a)","==> size(Meas_ADS%validity_indicators) = ",s,&
            " [printing first and last one]"
       DO m=1,s,s-1
          print "(' Meas_ADS%validity_indicators(',i2.2,')%...')",m
          call Print_validity_indicators(Meas_ADS%validity_indicators(m))
       END DO
    ELSE
       print *,"Meas_ADS%validity_indicators(:) = [NOT ASSOCIATED]"
    END IF

  end subroutine Print_L1B_Meas_ADS

    !  #]
  subroutine Read_L1B_Meas_ADS(Measurement_DSD,fileunit,which_DS,N_Meas_Max,&
                             N_range_gates,Meas_ADS,error_flag)
    !  #[

    TYPE(DSD_type),         pointer     :: Measurement_DSD       ! input
    integer,                intent(in)  :: fileunit     ! input
    integer,                intent(in)  :: which_DS     ! input
    integer,                intent(in)  :: N_Meas_Max   ! input 
    !                       (used for checking validity of the dataset !)
    integer,                intent(in)  :: N_range_gates ! input
    !                       (used for checking validity of the dataset !)
    TYPE(L1B_Meas_ADS_type),pointer     :: Meas_ADS     ! result
    integer,                intent(out) :: error_flag   ! result    

    ! local variables
    integer(i8_)      :: pos, endpos
    integer           :: currentpos
    integer(IntAul)   :: value_IntAul
    integer           :: m

    error_flag = no_error

    ! See the files: L1bP_Draft_1/aeolus/common/PFR/src/
    !        PFR_LevelOneBProductGroundWindDetectionDSR.cc/.hh       
    ! for the code that defines this datablock in the L1BP

    IF (.not. associated(Measurement_DSD)) THEN
       print *,"ERROR in Read_L1B_Meas_ADS()"
       print *,"the pointer Measurement_DSD seems not associated"
       error_flag = error_programming
       return
    END IF

    IF (.not. associated(Meas_ADS)) THEN
       print *,"ERROR in Read_L1B_Meas_ADS:"
       print *,"the pointer Meas_ADS seems not associated"
       error_flag = error_programming
       return
    END IF

    pos = Measurement_DSD%start_of_ds + (which_DS-1)*Measurement_DSD%size_of_one_dsr
    call SetFilePointer(fileunit,pos,error_flag)
    IF (error_flag .ne. no_error) return

    print *,"trying to read the binary Meas_ADS nr. ",which_DS

    ! read the 12 byte datetime field
    call read_DateTime_from_bin_file(fileunit,&
         Meas_ADS%start_of_obs_datetime,error_flag)

    ! read Num_of_Reference_Pulses
    call read_IntAul_from_bin_file(fileunit,value_IntAul,error_flag)
    IF (error_flag .ne. no_error) return
    Meas_ADS%Num_of_Reference_Pulses = value_IntAul

    ! skip the 4 byte spare
    ! to be replaced by a call to: read_zero_bytes_bin_file
    call AdvanceFilePointer(fileunit,4_i8_,error_flag)
    IF (error_flag .ne. no_error) return

    ! read List_of_Mie_Reference_Pulses

    IF (.not. associated(Meas_ADS%Mie_reference_pulses)) THEN
       print *,"ERROR in Read_L1B_Meas_ADS()"
       print *,"the datastructure L1B%Meas_ADS%Mie_reference_pulses(:) seems to be"
       print *,"not associated ...."
       error_flag = error_programming
       return
    END IF

    ! check if the struct was initialised with the correct nr of reference pulses (Mie)
    IF (N_meas_Max .ne. size(Meas_ADS%Mie_reference_pulses)) THEN
       print *,"ERROR in Read_L1B_Meas_ADS()"
       print *,"the datastructure L1B%Meas_ADS%Mie_reference_pulses(:) ",&
            "has room for ",size(Meas_ADS%Mie_reference_pulses)
       print *,"co-added referece pulses, but the nr of measurements in the current"
       print *,"datafile seems to be ",N_meas_Max
       error_flag = error_programming
       return
    END IF

    ! read the Measurement_PCD structure for the N_meas reference pulses (Mie)
    DO m=1,size(Meas_ADS%Mie_reference_pulses) !N_meas ! 67
       call Read_Mie_reference_pulses(fileunit,Meas_ADS%Mie_reference_pulses(m),&
                                            error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! read List_of_Rayleigh_Reference_Pulses

    IF (.not. associated(Meas_ADS%Rayleigh_reference_pulses_A)) THEN
       print *,"ERROR in Read_L1B_Meas_ADS()"
       print *,"the datastructure L1B%Meas_ADS%Rayleigh_reference_pulses_A(:) seems to be"
       print *,"not associated ...."
       error_flag = error_programming
       return
    END IF

    IF (.not. associated(Meas_ADS%Rayleigh_reference_pulses_B)) THEN
       print *,"ERROR in Read_L1B_Meas_ADS()"
       print *,"the datastructure L1B%Meas_ADS%Rayleigh_reference_pulses_B(:) seems to be"
       print *,"not associated ...."
       error_flag = error_programming
       return
    END IF

    ! check if the struct was initialised with the correct nr of reference pulses (Rayleigh)
    IF (N_meas_Max .ne. size(Meas_ADS%Rayleigh_reference_pulses_A)) THEN
       print *,"ERROR in Read_L1B_Meas_ADS()"
       print *,"the datastructure L1B%Meas_ADS%Rayleigh_reference_pulses_A(:) ",&
            "has room for ",size(Meas_ADS%Rayleigh_reference_pulses_A)
       print *,"co-added referece pulses, but the nr of measurements in the current"
       print *,"datafile seems to be ",N_meas_Max
       error_flag = error_programming
       return
    END IF
    IF (N_meas_Max .ne. size(Meas_ADS%Rayleigh_reference_pulses_B)) THEN
       print *,"ERROR in Read_L1B_Meas_ADS()"
       print *,"the datastructure L1B%Meas_ADS%Rayleigh_reference_pulses_B(:) ",&
            "has room for ",size(Meas_ADS%Rayleigh_reference_pulses_B)
       print *,"co-added referece pulses, but the nr of measurements in the current"
       print *,"datafile seems to be ",N_meas_Max
       error_flag = error_programming
       return
    END IF

    ! read the Measurement_PCD structure for the N_meas reference pulses (Rayleigh)
    DO m=1,size(Meas_ADS%Rayleigh_reference_pulses_A) !N_meas_Max ! 67
       call Read_Rayl_reference_pulses(fileunit,Meas_ADS%Rayleigh_reference_pulses_A(m),&
                                            error_flag)
       IF (error_flag .ne. no_error) return
    END DO
    DO m=1,size(Meas_ADS%Rayleigh_reference_pulses_B) !N_meas_Max ! 67
       call Read_Rayl_reference_pulses(fileunit,Meas_ADS%Rayleigh_reference_pulses_B(m),&
                                            error_flag)
       IF (error_flag .ne. no_error) return
    END DO


    ! read Mie_Measurement_Data

    IF (.not. associated(Meas_ADS%Mie_measurement_data)) THEN
       print *,"ERROR in Read_L1B_Meas_ADS()"
       print *,"the datastructure L1B%Meas_ADS%Mie_measurement_data(:) seems to be"
       print *,"not associated ...."
       error_flag = error_programming
       return
    END IF

    ! check if the struct was initialised with the correct nr of measurements
    IF (N_meas_Max .ne. size(Meas_ADS%Mie_measurement_data)) THEN
       print *,"ERROR in Read_L1B_Meas_ADS()"
       print *,"the datastructure L1B%Meas_ADS%Mie_measurement_data(:) ",&
            "has room for ",size(Meas_ADS%Mie_measurement_data)
       print *,"measurements, but the nr of measurements in the current"
       print *,"datafile seems to be ",N_meas_Max
       error_flag = error_programming
       return
    END IF

    ! read the validity_indicator structure for the 67 measurements
    DO m=1,size(Meas_ADS%Mie_measurement_data) !N_meas_Max ! 67
      call Read_Mie_measurement_data(fileunit,Meas_ADS%Mie_measurement_data(m),N_range_gates+1,&
                                         error_flag)
      IF (error_flag .ne. no_error) return
    END DO

    ! read Mie_Time_Delays
    call Read_measurement_time_delay(fileunit,Meas_ADS%Mie_time_delays,&
                                         error_flag)
    IF (error_flag .ne. no_error) return

    ! read Rayleigh_Time_Delays
    call Read_measurement_time_delay(fileunit,Meas_ADS%Rayleigh_time_delays,&
                                         error_flag)
    IF (error_flag .ne. no_error) return

    ! read List_of_Measurement_Validity_Indicators

    IF (.not. associated(Meas_ADS%validity_indicators)) THEN
       print *,"ERROR in Read_L1B_Meas_ADS()"
       print *,"the datastructure L1B%Meas_ADS%validity_indicators(:) seems to be"
       print *,"not associated ...."
       error_flag = error_programming
       return
    END IF

    ! check if the struct was initialised with the correct nr of measurements
    IF (N_meas_Max .ne. size(Meas_ADS%validity_indicators)) THEN
       print *,"ERROR in Read_L1B_Meas_ADS()"
       print *,"the datastructure L1B%Meas_ADS%validity_indicators(:) ",&
            "has room for ",size(Meas_ADS%validity_indicators)
       print *,"measurements, but the nr of measurements in the current"
       print *,"datafile seems to be ",N_meas_Max
       error_flag = error_programming
       return
    END IF

    ! read the validity_indicator structure for the 67 measurements
    DO m=1,N_meas_Max ! 67
      call Read_validity_indicators(fileunit,Meas_ADS%validity_indicators(m),&
                                         error_flag)
      IF (error_flag .ne. no_error) return
    END DO

    ! calculate expected end-position after reading this block of data
    endpos = Measurement_DSD%start_of_ds + which_DS*Measurement_DSD%size_of_one_dsr
    ! request the current file position
    inquire(unit=fileunit,nextrec=currentpos)
    
    IF (currentpos .ne. endpos) THEN
       print *,"ERROR in Read_L1B_Meas_ADS():"
       print *,"there seems to be a problem while reading this binary dataset"
       print *,"expected file position after this DSR is: ",endpos
       print *,"but the actual position seems to be:      ",currentpos
       print *,"Maybe the fileformat definition has changed ???"
       error_flag = error_reading_file
       return
    END IF

  end subroutine Read_L1B_Meas_ADS
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_Mie_reference_pulses(Mie_reference_pulses)
    !  #[
    TYPE(Mie_spectrometer_count_type),intent(out) :: Mie_reference_pulses ! output

    call Init_alt_bin_Mie_spectrometer(Mie_reference_pulses)

  end subroutine Init_Mie_reference_pulses
    !  #]
  subroutine Print_Mie_reference_pulses(Mie_reference_pulses)
    !  #[
    TYPE(Mie_spectrometer_count_type),intent(in) :: Mie_reference_pulses

    ! local variable
    character      :: indent_string

    indent_string = ' '
    call Print_alt_bin_Mie_spectrometer( Mie_reference_pulses, indent_string )

  end subroutine Print_Mie_reference_pulses
    !  #]
  subroutine Read_Mie_reference_pulses(fileunit,Mie_spectrometer_readout,&
                                                  error_flag)
    !  #[
    integer,          intent(in)  :: fileunit             ! input
    TYPE(Mie_spectrometer_count_type) :: Mie_spectrometer_readout ! output
    integer,          intent(out) :: error_flag           ! result    

    ! local variables
    integer         :: i, s
    integer(IntAus) :: value_IntAus

    s = size(Mie_spectrometer_readout%spectrometer_count)
    DO i = 1, s
      call read_IntAus_from_bin_file(fileunit,value_IntAus, error_flag)
      IF (error_flag .ne. no_error) return
      Mie_spectrometer_readout%spectrometer_count(i)=value_IntAus
    END DO

  end subroutine Read_Mie_reference_pulses
    !  #]
  subroutine Init_Rayl_reference_pulses(Rayl_reference_pulses)
    !  #[
    TYPE(Rayl_spectrometer_count_type),intent(out) :: Rayl_reference_pulses ! output

    Rayl_reference_pulses%spectrometer_count = missing_indicator_FAdoxy 
    !call Init_alt_bin_Rayl_spectrometer_count(Rayl_reference_pulses(i))

  end subroutine Init_Rayl_reference_pulses
    !  #]
  subroutine Print_Rayl_reference_pulses(Rayl_reference_pulses)
    !  #[
    TYPE(Rayl_spectrometer_count_type),intent(in) :: Rayl_reference_pulses

    ! local variable
    character      :: indent_string
    character(len=13) :: txt

    indent_string = ' '
    txt = real2txt(Rayl_reference_pulses%spectrometer_count)
    print *,indent_string,"...%Rayl_reference_pulses%spectrometer_count = ", txt
!    call Print_Rayl_spectrometer_counts( Rayl_reference_pulses, indent_string )

  end subroutine Print_Rayl_reference_pulses
    !  #]
  subroutine Read_Rayl_reference_pulses(fileunit,Rayl_spectrometer_readout,&
                                                  error_flag)
    !  #[
    integer,          intent(in)  :: fileunit             ! input
    TYPE(Rayl_spectrometer_count_type) :: Rayl_spectrometer_readout ! output
    integer,          intent(out) :: error_flag           ! result    

    ! local variables
    real(FAdoxy)   :: value_FAdoxy

    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy, error_flag)
    IF (error_flag .ne. no_error) return
    Rayl_spectrometer_readout%spectrometer_count=value_FAdoxy

  end subroutine Read_Rayl_reference_pulses
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_Mie_measurement_data(Mie_measurement_spectrum, N_range_gates, &
                                      error_flag)
    !  #[
    TYPE(mie_measurement_spectrum_type)     :: Mie_measurement_spectrum ! output
    integer,                    intent(in)  :: N_range_gates            ! input
    integer,                    intent(out) :: error_flag               ! output

    ! local variables
    integer :: AllocStatus, r
    integer :: N_rows

    N_rows = N_range_gates + 1 ! include one background radiation sample

    allocate(Mie_measurement_spectrum%Mie_spectra_alt_bin(N_rows),stat=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation failure in subroutine Init_Mie_measurement_data"
       error_flag = error_allocate
       return
    END IF

    DO r=1,N_rows
       call Init_alt_bin_Mie_spectrometer(Mie_measurement_spectrum%Mie_spectra_alt_bin(r))
    END DO

  end subroutine Init_Mie_measurement_data
    !  #]
  subroutine Delete_Mie_measurement_data(Mie_measurement_spectrum)
    !  #[
    TYPE(mie_measurement_spectrum_type) :: Mie_measurement_spectrum

    IF (associated(Mie_measurement_spectrum%Mie_spectra_alt_bin)) THEN
       deallocate(Mie_measurement_spectrum%Mie_spectra_alt_bin)
       nullify(Mie_measurement_spectrum%Mie_spectra_alt_bin)
    END IF

  end subroutine Delete_Mie_measurement_data
    !  #]
  subroutine Print_Mie_measurement_data(Mie_measurement_spectrum)
    !  #[
    TYPE(mie_measurement_spectrum_type)     :: Mie_measurement_spectrum

    ! local variables
    integer      :: r, s
    character(3) :: indent_string

    indent_string='   '
    IF (associated(Mie_measurement_spectrum%Mie_spectra_alt_bin)) THEN
       s = size(Mie_measurement_spectrum%Mie_spectra_alt_bin)
       print "(a,i2,a)","==> size(...%Mie_spectra_alt_bin) = ",s,&
            " [printing first and last one]"
       DO r=1,s,s-1
          print "(' ...%Mie_spectra_alt_bin(',i2.2,')%...')",r
          call Print_alt_bin_Mie_spectrometer(Mie_measurement_spectrum%Mie_spectra_alt_bin(r),&
                                         indent_string)
       END DO
    ELSE
       print *,"...%Mie_spectra_alt_bin [not associated]"
    END IF
  end subroutine Print_Mie_measurement_data
    !  #]
  subroutine Read_Mie_measurement_data(fileunit,Mie_meas_spec_data,N_range_gates,&
                                     error_flag)
    !  #[
    integer, intent(in)                 :: fileunit     ! input
    TYPE(mie_measurement_spectrum_type) :: Mie_meas_spec_data  ! output
    integer, intent(in)                 :: N_range_gates ! input
    !                       (used for checking validity of the dataset !)
    integer, intent(out)                :: error_flag   ! result    

    ! local variables
    integer  :: r

    ! check if the struct was initialised with the correct nr of rangegates
    IF (N_range_gates .ne. size(Mie_meas_spec_data%Mie_spectra_alt_bin)) THEN
       print *,"ERROR in Read_Mie_measurement_data()"
       print *,"the datastructure Mie_meas_spec_data(:) has room for",&
            size(Mie_meas_spec_data%Mie_spectra_alt_bin)
       print *,"rangegates, but the nr of rangegates in the current"
       print *,"datafile seems to be ",N_range_gates
       error_flag = error_programming
       return
    END IF

    ! read the Rangebin_geoloc structure for the 24+1 Mie rangegates
    DO r=1,N_range_gates ! 24+1
       call Read_Mie_reference_pulses(fileunit,Mie_meas_spec_data%Mie_spectra_alt_bin(r),&
                                                  error_flag)
       IF (error_flag .ne. no_error) return
    END DO
  end subroutine Read_Mie_measurement_data
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_Measurement_time_delay(time_delays, N_range_gates, &
                              error_flag)
    !  #[
    TYPE(time_delay_type)             :: time_delays     ! output
    integer,              intent(in)  :: N_range_gates   ! input
    integer,              intent(out) :: error_flag      ! output

    ! local variables
    integer :: AllocStatus, r

    error_flag = no_error

    IF (N_range_gates .eq. 0) THEN
       nullify(time_delays%bin_layer_integration_time)
    ELSE
       allocate(time_delays%bin_layer_integration_time(N_range_gates),stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_Measurement_time_delay"
          error_flag = error_allocate
          return
       END IF
       DO r=1,N_range_gates
          call Init_Alt_Bin_time_delays(time_delays%bin_layer_integration_time(r))
       END DO
    END IF

    time_delays%background_integration_time = missing_indicator_IntAl

  end subroutine Init_Measurement_time_delay
    !  #]
  subroutine Print_Measurement_time_delay(time_delays)
    !  #[
    TYPE(time_delay_type)       :: time_delays

    ! local variables
    integer           :: m, s
    character(len=13) :: txt

       s = size(time_delays%bin_layer_integration_time)
       print "(a,i2,a)","==> size(...%bin_layer_integration_time) = ",s,&
            " [printing first and last one]"
       DO m=1,s,s-1
          txt = int2txt(time_delays%bin_layer_integration_time(m))
          print "(' ...%bin_layer_integration_time(',i2.2,') =    ',a)",m,txt
       END DO

       txt = int2txt(time_delays%background_integration_time)
       print *,"...%background_integration_time    =    ",txt

  end subroutine Print_Measurement_time_delay
    !  #]
  subroutine Delete_Measurement_time_delay(time_delays)
    !  #[
    TYPE(time_delay_type)       :: time_delays                 ! output

    IF (associated(time_delays%bin_layer_integration_time)) THEN
       deallocate(time_delays%bin_layer_integration_time)
       nullify(time_delays%bin_layer_integration_time)
    END IF

  end subroutine Delete_Measurement_time_delay
    !  #]
  subroutine Read_measurement_time_delay(fileunit, time_delays, error_flag)
    !  #[
    integer,  intent(in)  :: fileunit     ! input
    TYPE(time_delay_type) :: time_delays  ! output
    integer,  intent(out) :: error_flag   ! result    

    ! local variables
    integer(IntAl) :: value_IntAl
    integer        :: i, s

    ! read bin_layer_integration_time
    s = size(time_delays%bin_layer_integration_time)
    do i = 1, s
      call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
      IF (error_flag .ne. no_error) return
      time_delays%bin_layer_integration_time(i)=value_IntAl
    enddo

    ! read background_integration_time
    call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    time_delays%background_integration_time = value_IntAl

  end subroutine Read_measurement_time_delay
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_validity_indicators( validity_indicators )
    !  #[
    TYPE(validity_indicator_type)       :: validity_indicators ! output

    validity_indicators%measurement_data_present      = .false.
    validity_indicators%Mie_measurement_sp_valid      = .false.
    validity_indicators%Rayleigh_measurement_sp_valid = .false.
    validity_indicators%measurement_laser_freq_locked = .false.
    validity_indicators%spacecraft_attitude_on_target = .false.

  end subroutine Init_validity_indicators
    !  #]
  subroutine Print_validity_indicators( validity_indicators )
    !  #[
    TYPE(validity_indicator_type)       :: validity_indicators

    print *,"...%measurement_data_present      = ", validity_indicators%measurement_data_present
    print *,"...%Mie_measurement_sp_valid      = ", validity_indicators%Mie_measurement_sp_valid
    print *,"...%Rayleigh_measurement_sp_valid = ", validity_indicators%Rayleigh_measurement_sp_valid
    print *,"...%measurement_laser_freq_locked = ", validity_indicators%measurement_laser_freq_locked
    print *,"...%spacecraft_attitude_on_target = ", validity_indicators%spacecraft_attitude_on_target

  end subroutine Print_validity_indicators
    !  #]
  subroutine Read_validity_indicators(fileunit,validity_indicators,&
                                         error_flag)
    !  #[
    integer,          intent(in)  :: fileunit            ! input
    TYPE(validity_indicator_type) :: validity_indicators ! output
    integer,          intent(out) :: error_flag          ! result

    ! local variables
    logical      :: value_l

    call read_logical1_from_bin_file(fileunit,value_l,error_flag)
    IF (error_flag .ne. no_error) return
    validity_indicators%measurement_data_present      = value_l

    call read_logical1_from_bin_file(fileunit,value_l,error_flag)
    IF (error_flag .ne. no_error) return
    validity_indicators%Mie_measurement_sp_valid      = value_l

    call read_logical1_from_bin_file(fileunit,value_l,error_flag)
    IF (error_flag .ne. no_error) return
    validity_indicators%Rayleigh_measurement_sp_valid = value_l

    call read_logical1_from_bin_file(fileunit,value_l,error_flag)
    IF (error_flag .ne. no_error) return
    validity_indicators%measurement_laser_freq_locked = value_l

    call read_logical1_from_bin_file(fileunit,value_l,error_flag)
    IF (error_flag .ne. no_error) return
    validity_indicators%spacecraft_attitude_on_target = value_l

  end subroutine Read_validity_indicators
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_alt_bin_Mie_spectrometer(alt_bin_Mie_spectrometer_count)
    !  #[
    TYPE(Mie_spectrometer_count_type), intent(out) :: alt_bin_Mie_spectrometer_count

    alt_bin_Mie_spectrometer_count%spectrometer_count(:) = missing_indicator_IntAus

  end subroutine Init_alt_bin_Mie_spectrometer
    !  #]
  subroutine Print_alt_bin_Mie_spectrometer( alt_bin_Mie_spectrometer_counts, indent_string )
    !  #[
    TYPE(Mie_spectrometer_count_type), intent(in)   :: alt_bin_Mie_spectrometer_counts

    character(*)      :: indent_string
    character(len=13) :: txt
    integer           :: i

    do i=1,size(alt_bin_Mie_spectrometer_counts%spectrometer_count), &
         size(alt_bin_Mie_spectrometer_counts%spectrometer_count)-1
       txt = int2txt(alt_bin_Mie_spectrometer_counts%spectrometer_count(i),&
                     int_type="IntAus")
       print '(a,a,i2.2,a,a)',indent_string,"...%spectrometer_count(",i,") = ",txt
    enddo

  end subroutine Print_alt_bin_Mie_spectrometer
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_Alt_Bin_time_delays(alt_bin_layer_integration_time)
    !  #[
    integer(IntAl), intent(out) :: alt_bin_layer_integration_time       ! output

    alt_bin_layer_integration_time = missing_indicator_IntAl

  end subroutine Init_Alt_Bin_time_delays
    !  #]
  !---------------------------------------------------------------------
end module L1B_Meas_ADS
