module L1B_PCD_ADS ! L1B Product Confidence Data ADS
  !  #[ Documentation
  !     a module to handle the product confidence data annotation dataset
  !
  !     See also the L1BP files: 
  !          L1bP_Draft_1/aeolus/common/PFR/src/
  !             PFR_LevelOneBProductConfidenceDSR.cc
  !             PFR_LevelOneBProductConfidenceDSR.hh
  !     for the code that defines this datablock in the L1BP
  !
  ! WARNING: the datastructure defined in this c++ code (and the first L1B
  !          testfiles that we recieved) differs from the definition
  !          in the L1B_iodd document !!!!!!
  !          (the error_quantifier has been moved from table 6-4, to table 6-6)
  !
  !     Written by J. de Kloe, KNMI
  !     created:     07-07-2005
  !
  !     Modifications:
  !      02-Sep-2005 D. Tan     Upgrade to DLR TDS v1.2
  !                             Ref AE.TN.DLR.IODD-L1B.050705 (draft).
  !      28-Sep-2005 P. Poli    fixed incorrect intent in Observation_PCD
  !      25-Oct-2005 J. de Kloe corrected some more wrong intent statements
  !                             and introduced a 1 byte missing int number
  !      10-Jan-2006 J. de Kloe new DateTime type
  !      21-Feb-2006 P. Poli    Upgrade to DLR IODD L1B-IODD.231105
  !      22-Feb-2006 P. Poli    Upgrade to DLR IODD AE.TN.DLR.APM-L1B.121205
  !      28-Mar-2006 P. Poli    Upgrade to DLR IODD AE.TN.DLR.APM-L1B.150206
  !      12-May-2006 D. Tan     Null use of ParKind1 module
  !      03-Nov-2006 P. Poli    upgrade to L1BP v1.04 ;
  !                             moved and cleaned-up reading part
  !      06-Nov-2006 P. Poli    removed Mie_Emitted_Std_Dev listed in table 6.7
  !                             but not contained in L1B files
  !      20-Nov-2006 D. Tan     Re-introduce Mie_Emitted_Std_Dev, now in 
  !                             post-FAT TDS
  !      21-Nov-2006 D. Tan     Read_*_Alt_Bin_PCD workaround for all 
  !                             range gates
  !      20-Mar-2007 J. de Kloe changed workarounds to issue warnings in stead
  !                             of errors, as was requested by ESA, as long
  !                             as we don't have proper L1B files that do no
  !                             longer need any workarounds ....
  !      08-May-2007 J. de Kloe added code to recognise the -1 used by MDA
  !                             as missing indicator for scattering_ratio
  !                             and Mie_snr/Rayleigh_snr
  !      09-May-2007 J. de Kloe add some debugging prints to demonstrate that
  !                             the problem described in SPR 63, point 6, 
  !                             (non-boolean values for boolean elements) 
  !                             still exist.
  !      03-Jan-2008 J. de Kloe added nullify statements to make the code
  !                             more robust
  !      16-Jan-2008 J. de Kloe phase out integer kind i_ 
  !
  !  #]
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE Numerics, only: i8_, real2txt
  USE DataSetDescriptor, only: DSD_type
  USE EE_CFI_Datatypes, only: DateTimeType,Construct_KVT_DateTime, InitDateTime, &
                              read_DateTime_from_bin_file, &
                              read_IntAs_from_bin_file, missing_indicator_IntAs, IntAs, &
                              read_IntAuc_from_bin_file, missing_indicator_IntAuc, IntAuc, &
                              read_IntAl_from_bin_file, missing_indicator_IntAl, IntAl, &
                              read_FAdoxy_from_bin_file, missing_indicator_FAdoxy, FAdoxy, &
                              int2txt, maxlen_DateTime_KVT
  USE DirectBinaryIO, only: SetFilePointer, AdvanceFilePointer, read_logical1_from_bin_file
  USE ErrorHandler, only: no_error, error_programming, error_allocate, error_reading_file
  !  #]
  !  #[ type definitions
  implicit none

  integer,      parameter :: N_quality_quantifiers = 20

  real(FAdoxy), parameter :: minus_one = -1._FAdoxy
  real(FAdoxy), parameter :: delta     =  1.e-10_FAdoxy

  TYPE Mie_Core_Char_type
     real(FAdoxy) :: Peak_Position
     real(FAdoxy) :: FWHM
     real(FAdoxy) :: Amplitude
     real(FAdoxy) :: Offset
     logical      :: Error_flag
     real(FAdoxy) :: Residual_Error
     integer(IntAuc):: Num_Iterations
  END TYPE Mie_Core_Char_type

  TYPE ErrorQuantifierType ! see AE.TN.DLR.IODD-l1B.050705, Table 2.16/2.19
                           ! Define now but note no data in DLR TDS v1.2.
     real(FAdoxy) :: error_quant_Mie
     real(FAdoxy) :: error_quant_Rayleigh
     real(FAdoxy) :: error_quant_Rayleigh_channel_A
     real(FAdoxy) :: error_quant_Rayleigh_channel_B
  END TYPE ErrorQuantifierType

  TYPE Obs_Alt_Bin_PCD_type ! see sec. 6.3.2, tables 6-6+6-7, of L1B_iodd
                            ! 4 error_quant's still not moved in DLR TDS v1.2?
     TYPE(ErrorQuantifierType) :: error_quantifier
!     real(r8_)   :: error_quant_Mie
!     real(r8_)   :: error_quant_Rayleigh
!     real(r8_)   :: error_quant_Rayleigh_channel_A
!     real(r8_)   :: error_quant_Rayleigh_channel_B
     real(FAdoxy) :: Mie_wind_velocity_std_dev
     real(FAdoxy) :: Rayleigh_wind_velocity_std_dev
     real(FAdoxy) :: Mie_useful_signal_std_dev
     real(FAdoxy) :: Rayleigh_useful_signal_stdev_A
     real(FAdoxy) :: Rayleigh_useful_signal_stdev_B
     TYPE(Mie_Core_Char_type) :: Mie_Core_Char
     real(FAdoxy) :: scattering_ratio
     real(FAdoxy) :: Mie_snr
     real(FAdoxy) :: Rayleigh_snr_channel_A
     real(FAdoxy) :: Rayleigh_snr_channel_B
  END TYPE Obs_Alt_Bin_PCD_type

  TYPE Meas_Alt_Bin_PCD_type
     logical     :: Mie_meas_invalid
     logical     :: Rayleigh_meas_invalid
     logical     :: Invalid_Mie_peak
     TYPE(Mie_Core_Char_type) :: Mie_Core_Char
     real(FAdoxy):: scattering_ratio
     real(FAdoxy):: Mie_snr
     real(FAdoxy):: Rayleigh_snr_channel_A
     real(FAdoxy):: Rayleigh_snr_channel_B
  END TYPE Meas_Alt_Bin_PCD_type

!  TYPE QualityCharacteristicsType ! see AE.TN.DLR.IODD-l1B.050705, Table 2.16/2.21/2.24
!     integer(i1_):: processing_modules_quality_flag
!     real(r8_), dimension(N_quality_quantifiers) :: quality_quantifiers
!  END TYPE QualityCharacteristicsType

  TYPE Observation_PCD_type ! see sec. 6.3.2, table 6-4, of L1B_iodd
     integer(IntAl) :: nr_meas_laser_freq_unlocked
     integer(IntAl) :: nr_ref_laser_freq_unlocked
     integer(IntAl) :: nr_sat_not_on_target_meas
     integer(IntAl) :: nr_corrupt_mie_meas
     integer(IntAl) :: nr_corrupt_rayleigh_meas
     integer(IntAl) :: nr_corrupt_mie_ref_pulses
     integer(IntAl) :: nr_corrupt_rayleigh_ref_pulses
!     TYPE(ErrorQuantifierType) :: error_quantifier
     ! This above item is defined here in the L1B_iodd but seems to be
     ! moved to inside the Obs_Alt_Bin_PCD_type
     real(FAdoxy)   :: avg_laser_freq_offset
     real(FAdoxy)   :: avg_UV_energy
     real(FAdoxy)   :: laser_freq_offset_std_dev
     real(FAdoxy)   :: UV_energy_std_dev
     integer(IntAl) :: nr_invalid_mie_peak
     TYPE(Obs_Alt_Bin_PCD_type), dimension(:), pointer :: Obs_Alt_Bin_PCD
     integer(IntAl) :: nr_mie_invalid_meas
     integer(IntAl) :: nr_mie_invalid_ref_pulses
     integer(IntAl) :: nr_rayleigh_invalid_meas
     integer(IntAl) :: nr_rayleigh_invalid_ref_pulses
     real(FAdoxy)   :: mie_mean_emitted_freq
     real(FAdoxy)   :: mie_emitted_freq_std_dev
     real(FAdoxy)   :: rayleigh_mean_emitted_freq
     real(FAdoxy)   :: rayleigh_emitted_freq_std_dev
     ! added in L1B v1.04:
     logical        :: multimode_ratio_quality_flag
!     TYPE(QualityCharacteristicsType) :: quality_characteristics
     ! 4-byte spare
  END TYPE Observation_PCD_type

  TYPE Measurement_PCD_type ! see sec. 6.3.2, table 6-9, of L1B_iodd
     integer(IntAl) :: num_mie_invalid_ref_pulses
     integer(IntAl) :: num_rayleigh_invalid_ref_pulses
     real(FAdoxy)   :: avg_laser_freq_offset
     real(FAdoxy)   :: avg_UV_energy
     real(FAdoxy)   :: laser_freq_offset_std_dev
     real(FAdoxy)   :: UV_energy_std_dev
     !              1234567890123456789012345678901
     TYPE(Meas_Alt_Bin_PCD_type), dimension(:), pointer :: Meas_Alt_Bin_PCD
     real(FAdoxy)   :: vel_of_att_uncertainty_error
     real(FAdoxy)   :: mie_mean_emitted_freq
     ! removed in L1B v1.3:
     !real(FAdoxy)   :: mie_emitted_freq_std_dev
     ! added in L1BP IODD, present in L1B v1.04 (post-FAT, 20061114):
     real(FAdoxy)   :: mie_emitted_std_dev
     ! added in L1B v1.3:
     real(FAdoxy)   :: reference_pulse_FWHM
     real(FAdoxy)   :: rayleigh_mean_emitted_freq
     real(FAdoxy)   :: rayleigh_emitted_freq_std_dev
!     TYPE(QualityCharacteristicsType) :: quality_characteristics
     ! added in L1B v1.04:
     logical        :: uv_energy_quality_flag
     ! 8-byte spare
  END TYPE Measurement_PCD_type

  TYPE L1B_PCD_ADS_type ! see sec. 6.3.2, table 6-3, of L1B_iodd
     type(DateTimeType)    :: start_of_obs_datetime
     integer(IntAs)        :: nmeas
     integer(IntAs)        :: npulses
     ! 8-byte spare
     TYPE(Observation_PCD_type)                        :: Observation_PCD
     TYPE(Measurement_PCD_type), dimension(:), pointer :: Measurement_PCD
     ! 8-byte spare
  END TYPE L1B_PCD_ADS_type

  !  #]
contains
  !---------------------------------------------------
  subroutine Init_ErrorQuantifier(error_quantifier)
    !  #[
    TYPE(ErrorQuantifierType), intent(out) :: error_quantifier

    error_quantifier%error_quant_Mie                = missing_indicator_FAdoxy
    error_quantifier%error_quant_Rayleigh           = missing_indicator_FAdoxy
    error_quantifier%error_quant_Rayleigh_channel_A = missing_indicator_FAdoxy
    error_quantifier%error_quant_Rayleigh_channel_B = missing_indicator_FAdoxy

  end subroutine Init_ErrorQuantifier
    !  #]
  subroutine Print_ErrorQuantifier(error_quantifier)
    !  #[
    TYPE(ErrorQuantifierType), intent(in) :: error_quantifier

    ! local variables
    character(len=13) :: txt

    txt = real2txt(error_quantifier%error_quant_Mie)
    print *,"...Error_Quantifier%error_quant_Mie                = ",txt
    txt = real2txt(error_quantifier%error_quant_Rayleigh)
    print *,"...Error_Quantifier%error_quant_Rayleigh           = ",txt
    txt = real2txt(error_quantifier%error_quant_Rayleigh_channel_A)
    print *,"...Error_Quantifier%error_quant_Rayleigh_channel_A = ",txt
    txt = real2txt(error_quantifier%error_quant_Rayleigh_channel_B)
    print *,"...Error_Quantifier%error_quant_Rayleigh_channel_B = ",txt

  end subroutine Print_ErrorQuantifier
    !  #]  
  subroutine Read_ErrorQuantifier(fileunit, Error_Quantifier, error_flag)
    !  #[
    integer,                   intent(in)  :: fileunit
    TYPE(ErrorQuantifierType), intent(out) :: Error_Quantifier
    integer,                   intent(out) :: error_flag

    real(FAdoxy)       :: value_FAdoxy
!    integer(IntAuc) :: value_IntAuc ! never used

    error_flag = no_error

    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Error_Quantifier%error_quant_Mie                = value_FAdoxy

    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Error_Quantifier%error_quant_Rayleigh           = value_FAdoxy

    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Error_Quantifier%error_quant_Rayleigh_channel_A = value_FAdoxy

    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Error_Quantifier%error_quant_Rayleigh_channel_B = value_FAdoxy
   
  end subroutine Read_ErrorQuantifier
    !  #]
  !---------------------------------------------------
!  subroutine Init_QualityCharacteristics(quality_characteristics)
!    !  #[
!    TYPE(QualityCharacteristicsType), intent(out) :: quality_characteristics
!
!    quality_characteristics%processing_modules_quality_flag = missing_indicator_integer_i1_
!    quality_characteristics%quality_quantifiers(:)          = missing_indicator_FAdoxy
!
!  end subroutine Init_QualityCharacteristics
!    !  #]
!  subroutine Print_QualityCharacteristics(quality_characteristics)
!    !  #[
!    TYPE(QualityCharacteristicsType), intent(in) :: quality_characteristics
!
!    ! local variables
!    character(len=13) :: txt
!
!    print *,"...%quality_characteristics%processing_modules_quality_flag = ",&
!         quality_characteristics%processing_modules_quality_flag
!
!    txt = real2txt(quality_characteristics%quality_quantifiers(1))
!    print *,"...%quality_characteristics%quality_quantifiers( 1)         = ",txt
!
!    print *,"...%quality_characteristics%quality_quantifiers last index  = ",&
!         N_quality_quantifiers
!    txt = real2txt(quality_characteristics%quality_quantifiers(N_quality_quantifiers))
!    print *,"...%quality_characteristics%quality_quantifiers(last index) = ",txt
!
!!    txt = real2txt(error_quantifier%error_quant_Rayleigh)
!!    print *,"...Error_Quantifier%error_quant_Rayleigh           = ",txt
!!    txt = real2txt(error_quantifier%error_quant_Rayleigh_channel_A)
!!    print *,"...Error_Quantifier%error_quant_Rayleigh_channel_A = ",txt
!!    txt = real2txt(error_quantifier%error_quant_Rayleigh_channel_B)
!!    print *,"...Error_Quantifier%error_quant_Rayleigh_channel_B = ",txt
!
!  end subroutine Print_QualityCharacteristics
!    !  #]  
  !---------------------------------------------------
  subroutine Init_Mie_Core_Char(Mie_Core_Char)
    !  #[
    TYPE(Mie_Core_Char_type), intent(out) :: Mie_Core_Char

    Mie_Core_Char%Peak_Position = missing_indicator_FAdoxy
    Mie_Core_Char%FWHM          = missing_indicator_FAdoxy
    Mie_Core_Char%Amplitude     = missing_indicator_FAdoxy
    Mie_Core_Char%Offset        = missing_indicator_FAdoxy
    Mie_Core_Char%Error_Flag    = .FALSE.
    Mie_Core_Char%Residual_Error= missing_indicator_FAdoxy
    Mie_Core_Char%Num_Iterations= missing_indicator_IntAuc
  end subroutine Init_Mie_Core_Char
    !  #]
  subroutine Print_Mie_Core_Char(Mie_Core_Char)
    !  #[
    TYPE(Mie_Core_Char_type), intent(in) :: Mie_Core_Char

    character(len=13) :: txt

    txt = real2txt(Mie_Core_Char%Peak_Position)
    print *,"...%Peak_Position = ",txt
    txt = real2txt(Mie_Core_Char%FWHM)
    print *,"...%FWHM          = ",txt
    txt = real2txt(Mie_Core_Char%Amplitude)
    print *,"...%Amplitude     = ",txt
    txt = real2txt(Mie_Core_Char%Offset)
    print *,"...%Offset        = ",txt
    print *,"...%Error_Flag    = ",Mie_Core_Char%Error_Flag
    txt = real2txt(Mie_Core_Char%Residual_Error)
    print *,"...%Residual_Error= ",txt
    txt = int2txt(Mie_Core_Char%Num_Iterations,int_type="IntAuc")
    print *,"...%Num_Iterations= ",txt
  end subroutine Print_Mie_Core_Char
    !  #]
  subroutine Read_Mie_Core_Char(fileunit, Mie_Core_Char, error_flag, workaround_in)
    !  #[
    integer,                  intent(in)  :: fileunit
    TYPE(Mie_Core_Char_type), intent(out) :: Mie_Core_Char
    integer,                  intent(out) :: error_flag
    logical,        optional, intent(in)  :: workaround_in

    real(FAdoxy) :: value_FAdoxy
    integer(IntAuc) :: value_IntAuc
    logical   :: value_logical, workaround

    error_flag = no_error
    workaround=.FALSE.
    IF (present(workaround_in)) workaround=workaround_in

    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Mie_Core_Char%Peak_Position = value_FAdoxy

    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Mie_Core_Char%FWHM          = value_FAdoxy

    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Mie_Core_Char%Amplitude     = value_FAdoxy

    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Mie_Core_Char%Offset        = value_FAdoxy

    call read_logical1_from_bin_file(fileunit,value_logical,error_flag,&
                            workaround,description="Mie_Core_Char%Error_Flag")
    IF (error_flag .ne. no_error) return
    Mie_Core_Char%Error_Flag    = value_logical

    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Mie_Core_Char%Residual_Error= value_FAdoxy

    call read_IntAuc_from_bin_file(fileunit,value_IntAuc,error_flag)
    IF (error_flag .ne. no_error) return
    Mie_Core_Char%Num_Iterations= value_IntAuc

  end subroutine Read_Mie_Core_Char
    !  #]
  !---------------------------------------------------
  subroutine Init_Obs_Alt_Bin_PCD(Obs_Alt_Bin_PCD)
    !  #[
    TYPE(Obs_Alt_Bin_PCD_type), intent(out) :: Obs_Alt_Bin_PCD

    call Init_ErrorQuantifier(Obs_Alt_Bin_PCD%error_quantifier)
!    Obs_Alt_Bin_PCD%error_quant_Mie                = missing_indicator_FAdoxy
!    Obs_Alt_Bin_PCD%error_quant_Rayleigh           = missing_indicator_FAdoxy
!    Obs_Alt_Bin_PCD%error_quant_Rayleigh_channel_A = missing_indicator_FAdoxy
!    Obs_Alt_Bin_PCD%error_quant_Rayleigh_channel_B = missing_indicator_FAdoxy
    Obs_Alt_Bin_PCD%Mie_wind_velocity_std_dev      = missing_indicator_FAdoxy
    Obs_Alt_Bin_PCD%Rayleigh_wind_velocity_std_dev = missing_indicator_FAdoxy
    Obs_Alt_Bin_PCD%Mie_useful_signal_std_dev      = missing_indicator_FAdoxy
    Obs_Alt_Bin_PCD%Rayleigh_useful_signal_stdev_A = missing_indicator_FAdoxy
    Obs_Alt_Bin_PCD%Rayleigh_useful_signal_stdev_B = missing_indicator_FAdoxy
    call Init_Mie_Core_Char(Obs_Alt_Bin_PCD%Mie_Core_Char)
    Obs_Alt_Bin_PCD%scattering_ratio               = missing_indicator_FAdoxy
    Obs_Alt_Bin_PCD%Mie_snr                        = missing_indicator_FAdoxy
    Obs_Alt_Bin_PCD%Rayleigh_snr_channel_A         = missing_indicator_FAdoxy
    Obs_Alt_Bin_PCD%Rayleigh_snr_channel_B         = missing_indicator_FAdoxy

  end subroutine Init_Obs_Alt_Bin_PCD
    !  #]
  subroutine Print_Obs_Alt_Bin_PCD(Obs_Alt_Bin_PCD)
    !  #[
    TYPE(Obs_Alt_Bin_PCD_type), intent(in) :: Obs_Alt_Bin_PCD

    ! local variables
    character(len=13) :: txt

    print *,"...%error_quantifier                : "
    call Print_ErrorQuantifier(Obs_Alt_Bin_PCD%error_quantifier)
!    txt = real2txt(Obs_Alt_Bin_PCD%error_quant_Mie)
!    print *,"...%error_quant_Mie                = ",txt
!    txt = real2txt(Obs_Alt_Bin_PCD%error_quant_Rayleigh)
!    print *,"...%error_quant_Rayleigh           = ",txt
!    txt = real2txt(Obs_Alt_Bin_PCD%error_quant_Rayleigh_channel_A)
!    print *,"...%error_quant_Rayleigh_channel_A = ",txt
!    txt = real2txt(Obs_Alt_Bin_PCD%error_quant_Rayleigh_channel_B)
!    print *,"...%error_quant_Rayleigh_channel_B = ",txt
    txt = real2txt(Obs_Alt_Bin_PCD%Mie_wind_velocity_std_dev)
    print *,"...%Mie_wind_velocity_std_dev      = ",txt
    txt = real2txt(Obs_Alt_Bin_PCD%Rayleigh_wind_velocity_std_dev)
    print *,"...%Rayleigh_wind_velocity_std_dev = ",txt
    txt = real2txt(Obs_Alt_Bin_PCD%Mie_useful_signal_std_dev)
    print *,"...%Mie_useful_signal_std_dev      = ",txt
    txt = real2txt(Obs_Alt_Bin_PCD%Rayleigh_useful_signal_stdev_A)
    print *,"...%Rayleigh_useful_signal_stdev_A = ",txt
    txt = real2txt(Obs_Alt_Bin_PCD%Rayleigh_useful_signal_stdev_B)
    print *,"...%Rayleigh_useful_signal_stdev_B = ",txt
    print *,"...%Mie_Core_Char                  : "
    call Print_Mie_Core_Char(Obs_Alt_Bin_PCD%Mie_Core_Char)
    txt = real2txt(Obs_Alt_Bin_PCD%scattering_ratio)
    print *,"...%scattering_ratio               = ",txt
    txt = real2txt(Obs_Alt_Bin_PCD%Mie_snr)
    print *,"...%Mie_snr                        = ",txt
    txt = real2txt(Obs_Alt_Bin_PCD%Rayleigh_snr_channel_A)
    print *,"...%Rayleigh_snr_channel_A         = ",txt
    txt = real2txt(Obs_Alt_Bin_PCD%Rayleigh_snr_channel_B)
    print *,"...%Rayleigh_snr_channel_B         = ",txt

  end subroutine Print_Obs_Alt_Bin_PCD
    !  #]  
  subroutine Read_Obs_Alt_Bin_PCD(fileunit, Obs_Alt_Bin_PCD, error_flag, &
                                  workaround_in)
    !  #[

    integer,                    intent(in)  :: fileunit        ! input
    TYPE(Obs_Alt_Bin_PCD_type), intent(out) :: Obs_Alt_Bin_PCD ! result
    integer,                    intent(out) :: error_flag      ! result
    logical,          optional, intent(in)  :: workaround_in

    ! local variables
    real(FAdoxy)   :: value_FAdoxy
    logical        :: workaround

    error_flag = no_error
    workaround=.FALSE.
    IF (present(workaround_in)) workaround=workaround_in

    ! read Error_Quantifier
    call Read_ErrorQuantifier(fileunit, &
         Obs_Alt_Bin_PCD%Error_Quantifier, error_flag)
    IF (error_flag .ne. no_error) return

!!   For DLR TDS v1.2, the 4 error quants are not yet moved to read_Obs_PCD?
!    call read_real8_from_bin_file(fileunit,value_FAdoxy,error_flag)
!    IF (error_flag .ne. no_error) return
!    Obs_Alt_Bin_PCD%error_quant_Mie                = value_FAdoxy
!
!    call read_real8_from_bin_file(fileunit,value_FAdoxy,error_flag)
!    IF (error_flag .ne. no_error) return
!    Obs_Alt_Bin_PCD%error_quant_Rayleigh           = value_FAdoxy
!
!    call read_real8_from_bin_file(fileunit,value_FAdoxy,error_flag)
!    IF (error_flag .ne. no_error) return
!    Obs_Alt_Bin_PCD%error_quant_Rayleigh_channel_A = value_FAdoxy
!
!    call read_real8_from_bin_file(fileunit,value_FAdoxy,error_flag)
!    IF (error_flag .ne. no_error) return
!    Obs_Alt_Bin_PCD%error_quant_Rayleigh_channel_B = value_FAdoxy

    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Obs_Alt_Bin_PCD%Mie_wind_velocity_std_dev      = value_FAdoxy

    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Obs_Alt_Bin_PCD%Rayleigh_wind_velocity_std_dev = value_FAdoxy

    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Obs_Alt_Bin_PCD%Mie_useful_signal_std_dev      = value_FAdoxy

    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Obs_Alt_Bin_PCD%Rayleigh_useful_signal_stdev_A = value_FAdoxy

    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Obs_Alt_Bin_PCD%Rayleigh_useful_signal_stdev_B = value_FAdoxy

    !print *,"doing Read_Mie_Core_Char "

    ! read Mie_Core_Characteristics
    call Read_Mie_Core_Char(fileunit, &
         Obs_Alt_Bin_PCD%Mie_Core_Char, error_flag,workaround_in=workaround)
    IF (error_flag .ne. no_error) return

    !print *,"done with Read_Mie_Core_Char "

    ! read Scattering_Ratio
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Obs_Alt_Bin_PCD%scattering_ratio               = value_FAdoxy

    ! exception: define -1 to be missing for this item
    if ( abs(value_FAdoxy-minus_one) .lt. delta) &
         Obs_Alt_Bin_PCD%scattering_ratio = missing_indicator_FAdoxy

    ! read Mie_Signal_to_Noise_Ratio
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Obs_Alt_Bin_PCD%Mie_snr                        = value_FAdoxy

    ! exception: define -1 to be missing for this item
    if ( abs(value_FAdoxy-minus_one) .lt. delta) &
         Obs_Alt_Bin_PCD%Mie_snr = missing_indicator_FAdoxy

    ! read Rayleigh_Signal_to_Noise_Ratio_Channel_A
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Obs_Alt_Bin_PCD%Rayleigh_snr_channel_A         = value_FAdoxy

    ! exception: define -1 to be missing for this item
    if ( abs(value_FAdoxy-minus_one) .lt. delta) &
         Obs_Alt_Bin_PCD%Rayleigh_snr_channel_A = missing_indicator_FAdoxy

    ! read Rayleigh_Signal_to_Noise_Ratio_Channel_B
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Obs_Alt_Bin_PCD%Rayleigh_snr_channel_B         = value_FAdoxy

    ! exception: define -1 to be missing for this item
    if ( abs(value_FAdoxy-minus_one) .lt. delta) &
         Obs_Alt_Bin_PCD%Rayleigh_snr_channel_B = missing_indicator_FAdoxy

  end subroutine Read_Obs_Alt_Bin_PCD

    !  #]
  !---------------------------------------------------
  subroutine Init_Meas_Alt_Bin_PCD(Meas_Alt_Bin_PCD)
    !  #[
    TYPE(Meas_Alt_Bin_PCD_type), intent(out) :: Meas_Alt_Bin_PCD

    Meas_Alt_Bin_PCD%Mie_meas_invalid      = .false.
    Meas_Alt_Bin_PCD%Rayleigh_meas_invalid = .false.
    Meas_Alt_Bin_PCD%Invalid_Mie_peak      = .false.
    call Init_Mie_Core_Char(Meas_Alt_Bin_PCD%Mie_Core_Char)
    Meas_Alt_Bin_PCD%scattering_ratio      = missing_indicator_FAdoxy
    Meas_Alt_Bin_PCD%Mie_snr               = missing_indicator_FAdoxy
    Meas_Alt_Bin_PCD%Rayleigh_snr_channel_A= missing_indicator_FAdoxy
    Meas_Alt_Bin_PCD%Rayleigh_snr_channel_B= missing_indicator_FAdoxy

  end subroutine Init_Meas_Alt_Bin_PCD
    !  #]
  subroutine Print_Meas_Alt_Bin_PCD(Meas_Alt_Bin_PCD)
    !  #[
    TYPE(Meas_Alt_Bin_PCD_type), intent(in) :: Meas_Alt_Bin_PCD

    ! local variables
    character(len=13) :: txt

    print *,"...%Mie_meas_invalid      = ",&
         Meas_Alt_Bin_PCD%Mie_meas_invalid
    print *,"...%Rayleigh_meas_invalid = ",&
         Meas_Alt_Bin_PCD%Rayleigh_meas_invalid
    print *,"...%Invalid_Mie_peak      = ",&
         Meas_Alt_Bin_PCD%Invalid_Mie_peak

    print *,"...%Mie_Core_Char         : "
    call Print_Mie_Core_Char(Meas_Alt_Bin_PCD%Mie_Core_Char)

    txt = real2txt(Meas_Alt_Bin_PCD%scattering_ratio)
    print *,"...%scattering_ratio               = ",txt
    txt = real2txt(Meas_Alt_Bin_PCD%Mie_snr)
    print *,"...%Mie_snr                        = ",txt
    txt = real2txt(Meas_Alt_Bin_PCD%Rayleigh_snr_channel_A)
    print *,"...%Rayleigh_snr_channel_A         = ",txt
    txt = real2txt(Meas_Alt_Bin_PCD%Rayleigh_snr_channel_B)
    print *,"...%Rayleigh_snr_channel_B         = ",txt

  end subroutine Print_Meas_Alt_Bin_PCD
    !  #]
  subroutine Read_Meas_Alt_Bin_PCD(fileunit,Meas_Alt_Bin_PCD, error_flag, &
                                   workaround_in)
    !  #[
    integer,                     intent(in)  :: fileunit         ! input
    TYPE(Meas_Alt_Bin_PCD_type), intent(out) :: Meas_Alt_Bin_PCD ! result
    integer,                     intent(out) :: error_flag       ! result
    logical,           optional, intent(in)  :: workaround_in

    ! local variables
    logical   :: value_l, workaround
    real(FAdoxy) :: value_FAdoxy

    error_flag = no_error
    workaround=.FALSE.
    IF (present(workaround_in)) workaround=workaround_in

    ! read Mie_Measurement_Invalid
    call read_logical1_from_bin_file(fileunit,value_l,error_flag,&
                workaround,description="Meas_Alt_Bin_PCD%Mie_meas_invalid")
    IF (error_flag .ne. no_error) return
    Meas_Alt_Bin_PCD%Mie_meas_invalid = value_l

    ! read Rayleigh_Measurement_Invalid
    call read_logical1_from_bin_file(fileunit,value_l,error_flag,&
              workaround,description="Meas_Alt_Bin_PCD%Rayleigh_meas_invalid")
    IF (error_flag .ne. no_error) return
    Meas_Alt_Bin_PCD%Rayleigh_meas_invalid = value_l

    ! read Invalid_Mie_peak
    call read_logical1_from_bin_file(fileunit,value_l,error_flag,&
              workaround,description="Meas_Alt_Bin_PCD%Invalid_Mie_peak")
    IF (error_flag .ne. no_error) return
    Meas_Alt_Bin_PCD%Invalid_Mie_peak = value_l

    ! read Mie_Core_Characteristics
    call Read_Mie_Core_Char(fileunit, &
         Meas_Alt_Bin_PCD%Mie_Core_Char, error_flag,workaround_in=workaround)
    IF (error_flag .ne. no_error) return

    ! read Scattering_Ratio
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Meas_Alt_Bin_PCD%scattering_ratio               = value_FAdoxy

    ! exception: define -1 to be missing for this item
    if ( abs(value_FAdoxy-minus_one) .lt. delta) &
         Meas_Alt_Bin_PCD%scattering_ratio = missing_indicator_FAdoxy

    ! read Mie_Signal_to_Noise_Ratio
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Meas_Alt_Bin_PCD%Mie_snr                        = value_FAdoxy

    ! exception: define -1 to be missing for this item
    if ( abs(value_FAdoxy-minus_one) .lt. delta) &
         Meas_Alt_Bin_PCD%Mie_snr = missing_indicator_FAdoxy

    ! read Rayleigh_Signal_to_Noise_Ratio_Channel_A
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Meas_Alt_Bin_PCD%Rayleigh_snr_channel_A         = value_FAdoxy

    ! exception: define -1 to be missing for this item
    if ( abs(value_FAdoxy-minus_one) .lt. delta) &
         Meas_Alt_Bin_PCD%Rayleigh_snr_channel_A = missing_indicator_FAdoxy

    ! read Rayleigh_Signal_to_Noise_Ratio_Channel_B
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Meas_Alt_Bin_PCD%Rayleigh_snr_channel_B         = value_FAdoxy

    ! exception: define -1 to be missing for this item
    if ( abs(value_FAdoxy-minus_one) .lt. delta) &
         Meas_Alt_Bin_PCD%Rayleigh_snr_channel_B = missing_indicator_FAdoxy

  end subroutine Read_Meas_Alt_Bin_PCD
    !  #]
  !---------------------------------------------------
  subroutine Init_Observation_PCD(Observation_PCD, N_range_gates, error_flag)
    !  #[

    TYPE(Observation_PCD_type), intent(out) :: Observation_PCD ! output
    integer,                    intent(in)  :: N_range_gates   ! input
    integer,                    intent(out) :: error_flag      ! output

    ! local variables
    integer :: AllocStatus, r

    error_flag = no_error

    Observation_PCD%nr_meas_laser_freq_unlocked    = missing_indicator_IntAl
    Observation_PCD%nr_ref_laser_freq_unlocked     = missing_indicator_IntAl
    Observation_PCD%nr_sat_not_on_target_meas      = missing_indicator_IntAl
    Observation_PCD%nr_corrupt_mie_meas            = missing_indicator_IntAl
    Observation_PCD%nr_corrupt_rayleigh_meas       = missing_indicator_IntAl
    Observation_PCD%nr_corrupt_mie_ref_pulses      = missing_indicator_IntAl
    Observation_PCD%nr_corrupt_rayleigh_ref_pulses = missing_indicator_IntAl
!    ! the Observation_PCD%Error_Quantifier is now present in this type
!    call Init_ErrorQuantifier(Observation_PCD%error_quantifier)
    Observation_PCD%avg_laser_freq_offset     = missing_indicator_FAdoxy
    Observation_PCD%avg_UV_energy             = missing_indicator_FAdoxy
    Observation_PCD%laser_freq_offset_std_dev = missing_indicator_FAdoxy
    Observation_PCD%UV_energy_std_dev         = missing_indicator_FAdoxy
    Observation_PCD%nr_invalid_mie_peak       = missing_indicator_IntAl

    IF (N_range_gates .eq. 0) THEN
       ! init it to be empty
       nullify(Observation_PCD%Obs_Alt_Bin_PCD)
    ELSE
       allocate(Observation_PCD%Obs_Alt_Bin_PCD(N_range_gates+1),&
            stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_Observation_PCD"
          error_flag = error_allocate
          return
       END IF

       DO r=1,N_range_gates+1
          call Init_Obs_Alt_Bin_PCD(Observation_PCD%Obs_Alt_Bin_PCD(r))
       END DO
    END IF

    Observation_PCD%nr_mie_invalid_meas            = missing_indicator_IntAl
    Observation_PCD%nr_mie_invalid_ref_pulses      = missing_indicator_IntAl
    Observation_PCD%nr_rayleigh_invalid_meas       = missing_indicator_IntAl
    Observation_PCD%nr_rayleigh_invalid_ref_pulses = missing_indicator_IntAl

    Observation_PCD%mie_mean_emitted_freq          = missing_indicator_FAdoxy
    Observation_PCD%mie_emitted_freq_std_dev       = missing_indicator_FAdoxy
    Observation_PCD%rayleigh_mean_emitted_freq     = missing_indicator_FAdoxy
    Observation_PCD%rayleigh_emitted_freq_std_dev  = missing_indicator_FAdoxy
    Observation_PCD%multimode_ratio_quality_flag   = .FALSE.

!    call Init_QualityCharacteristics(Observation_PCD%quality_characteristics)

  end subroutine Init_Observation_PCD
    !  #]
  subroutine Delete_Observation_PCD(Observation_PCD)
    !  #[
    TYPE(Observation_PCD_type), intent(inout) :: Observation_PCD

    IF (associated(Observation_PCD%Obs_Alt_Bin_PCD)) THEN
       deallocate(Observation_PCD%Obs_Alt_Bin_PCD)
       nullify(Observation_PCD%Obs_Alt_Bin_PCD)
    END IF

  end subroutine Delete_Observation_PCD
    !  #]
  subroutine Print_Observation_PCD(Observation_PCD)
    !  #[
    TYPE(Observation_PCD_type), intent(in) :: Observation_PCD

    ! local variable
    integer :: s, r
    character(len=13) :: txt

    print *,"Observation_PCD%..."
    txt = int2txt(Observation_PCD%nr_meas_laser_freq_unlocked)
    print *,"...%nr_meas_laser_freq_unlocked    = ",txt
    txt = int2txt(Observation_PCD%nr_ref_laser_freq_unlocked)
    print *,"...%nr_ref_laser_freq_unlocked     = ",txt
    txt = int2txt(Observation_PCD%nr_sat_not_on_target_meas)
    print *,"...%nr_sat_not_on_target_meas      = ",txt
    txt = int2txt(Observation_PCD%nr_corrupt_mie_meas)
    print *,"...%nr_corrupt_mie_meas            = ",txt
    txt = int2txt(Observation_PCD%nr_corrupt_rayleigh_meas)
    print *,"...%nr_corrupt_rayleigh_meas       = ",txt
    txt = int2txt(Observation_PCD%nr_corrupt_mie_ref_pulses)
    print *,"...%nr_corrupt_mie_ref_pulses      = ",txt
    txt = int2txt(Observation_PCD%nr_corrupt_rayleigh_ref_pulses)
    print *,"...%nr_corrupt_rayleigh_ref_pulses = ",txt

!    call Print_ErrorQuantifier(Observation_PCD%error_quantifier)

    txt = real2txt(Observation_PCD%avg_laser_freq_offset)
    print *,"...%avg_laser_freq_offset          = ",txt
    txt = real2txt(Observation_PCD%avg_UV_energy)
    print *,"...%avg_UV_energy                  = ",txt
    txt = real2txt(Observation_PCD%laser_freq_offset_std_dev)
    print *,"...%laser_freq_offset_std_dev      = ",txt
    txt = real2txt(Observation_PCD%UV_energy_std_dev)
    print *,"...%UV_energy_std_dev              = ",txt
    txt = int2txt(Observation_PCD%nr_invalid_mie_peak)
    print *,"...%nr_invalid_mie_peak            = ",txt

    IF (associated(Observation_PCD%Obs_Alt_Bin_PCD)) THEN
       s = size(Observation_PCD%Obs_Alt_Bin_PCD)
       print "(a,i2,a)","==> size(Observation_PCD%Obs_Alt_Bin_PCD) = ",s,&
            " [printing first and last one]"
       DO r=1,s,s-1
          print "(' Observation_PCD%Obs_Alt_Bin_PCD(',i2.2,')%...')",r
          call Print_Obs_Alt_Bin_PCD(Observation_PCD%Obs_Alt_Bin_PCD(r))
       END DO
    ELSE
       print *,"...%Obs_Alt_Bin_PCD [not associated]"
    END IF

    print *,"Observation_PCD%...(continued)"
    txt = int2txt(Observation_PCD%nr_mie_invalid_meas)
    print *,"...%nr_mie_invalid_meas            = ",txt
    txt = int2txt(Observation_PCD%nr_mie_invalid_ref_pulses)
    print *,"...%nr_mie_invalid_ref_pulses      = ",txt
    txt = int2txt(Observation_PCD%nr_rayleigh_invalid_meas)
    print *,"...%nr_rayleigh_invalid_meas       = ",txt
    txt = int2txt(Observation_PCD%nr_rayleigh_invalid_ref_pulses)
    print *,"...%nr_rayleigh_invalid_ref_pulses = ",txt

    txt = real2txt(Observation_PCD%mie_mean_emitted_freq)
    print *,"...%mie_mean_emitted_freq          = ",txt
    txt = real2txt(Observation_PCD%mie_emitted_freq_std_dev)
    print *,"...%mie_emitted_freq_std_dev       = ",txt
    txt = real2txt(Observation_PCD%rayleigh_mean_emitted_freq)
    print *,"...%rayleigh_mean_emitted_freq     = ",txt
    txt = real2txt(Observation_PCD%rayleigh_emitted_freq_std_dev)
    print *,"...%rayleigh_emitted_freq_std_dev  = ",txt

    print *,"...%multimode_ratio_quality_flag   = ", &
    Observation_PCD%multimode_ratio_quality_flag

!    call Print_QualityCharacteristics(Observation_PCD%quality_characteristics)

  end subroutine Print_Observation_PCD
    !  #]
  subroutine Read_Observation_PCD(fileunit, Observation_PCD, &
                              N_range_gates, error_flag)
    !  #[
    integer,                    intent(in)    :: fileunit        ! input
    TYPE(Observation_PCD_type), intent(inout) :: Observation_PCD ! result
    integer,                    intent(in)    :: N_range_gates   ! input
    integer,                    intent(out)   :: error_flag      ! result

    ! local variable
    integer(IntAl) :: value_IntAl
    real(FAdoxy)   :: value_FAdoxy
    integer        :: r
    logical        :: value_logical

    error_flag = no_error

    ! read nr_meas_laser_freq_unlocked
    call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    Observation_PCD%nr_meas_laser_freq_unlocked = value_IntAl

    ! read nr_ref_laser_freq_unlocked
    call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    Observation_PCD%nr_ref_laser_freq_unlocked = value_IntAl

    ! read nr_sat_not_on_target_meas
    call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    Observation_PCD%nr_sat_not_on_target_meas = value_IntAl

    ! read nr_corrupt_mie_meas
    call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    Observation_PCD%nr_corrupt_mie_meas = value_IntAl

    ! read nr_corrupt_rayleigh_meas
    call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    Observation_PCD%nr_corrupt_rayleigh_meas = value_IntAl

    ! read nr_corrupt_mie_ref_pulses
    call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    Observation_PCD%nr_corrupt_mie_ref_pulses = value_IntAl

    ! read nr_corrupt_rayleigh_ref_pulses
    call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    Observation_PCD%nr_corrupt_rayleigh_ref_pulses = value_IntAl

    ! this type was present here in the L1B_iodd document of 6-5-2005
    ! but seems to have been moved in the testfile ...
    ! TYPE(ErrorQuantifierType) :: Error_Quantifier (32 bytes)
    ! Still not correctly placed in DLR TDS v1.2?
!    call read_real8_from_bin_file(fileunit,value_FAdoxy,error_flag)
!    IF (error_flag .ne. no_error) return
!    Observation_PCD%error_quantifier%error_quant_Mie                = value_FAdoxy
!
!    call read_real8_from_bin_file(fileunit,value_FAdoxy,error_flag)
!    IF (error_flag .ne. no_error) return
!    Observation_PCD%error_quantifier%error_quant_Rayleigh           = value_FAdoxy
!
!    call read_real8_from_bin_file(fileunit,value_FAdoxy,error_flag)
!    IF (error_flag .ne. no_error) return
!    Observation_PCD%error_quantifier%error_quant_Rayleigh_channel_A = value_FAdoxy
!
!    call read_real8_from_bin_file(fileunit,value_FAdoxy,error_flag)
!    IF (error_flag .ne. no_error) return
!    Observation_PCD%error_quantifier%error_quant_Rayleigh_channel_B = value_FAdoxy

    ! read avg_laser_freq_offset
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Observation_PCD%avg_laser_freq_offset = value_FAdoxy

    ! read avg_UV_energy
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Observation_PCD%avg_UV_energy = value_FAdoxy

    ! read laser_freq_offset_std_dev
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Observation_PCD%laser_freq_offset_std_dev = value_FAdoxy

    ! read UV_energy_std_dev
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Observation_PCD%UV_energy_std_dev = value_FAdoxy

    ! read nr_invalid_mie_peak
    call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    Observation_PCD%nr_invalid_mie_peak = value_IntAl

    IF (.not. associated(Observation_PCD%Obs_Alt_Bin_PCD)) THEN
       print *,"ERROR in Read_Observation_PCD()"
       print *,"the datastructure: Observation_PCD%Obs_Alt_Bin_PCD(:)"
       print *,"seems to be not associated ...."
       error_flag = error_programming
       return
    END IF

    ! check if the struct was initialised with the correct nr of rangegates
    IF (N_range_gates .ne. size(Observation_PCD%Obs_Alt_Bin_PCD)) THEN
       print *,"ERROR in Read_Observation_PCD()"
       print *,"the datastructure Observation_PCD%Obs_Alt_Bin_PCD(:) ",&
            "has room for ",size(Observation_PCD%Obs_Alt_Bin_PCD)
       print *,"rangegates, but the nr of rangegates in the current"
       print *,"datafile seems to be ",N_range_gates
       error_flag = error_programming
       return
    END IF

    ! read the Obs_Alt_Bin_PCD structure for the 24 rangegates
    DO r=1,N_range_gates ! 25
       !print *,"doing Read_Obs_Alt_Bin_PCD for r=",r
       call Read_Obs_Alt_Bin_PCD(fileunit,Observation_PCD%Obs_Alt_Bin_PCD(r),&
                                 error_flag,workaround_in=.TRUE.)
       IF (error_flag .ne. no_error) return
    END DO

    ! read nr_mie_invalid_meas
    call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    Observation_PCD%nr_mie_invalid_meas = value_IntAl

    ! read nr_mie_invalid_ref_pulses
    call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    Observation_PCD%nr_mie_invalid_ref_pulses = value_IntAl

    ! read nr_rayleigh_invalid_meas
    call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    Observation_PCD%nr_rayleigh_invalid_meas = value_IntAl

    ! read nr_rayleigh_invalid_ref_pulses
    call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    Observation_PCD%nr_rayleigh_invalid_ref_pulses = value_IntAl

    ! read mie_mean_emitted_freq
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Observation_PCD%mie_mean_emitted_freq = value_FAdoxy

    ! read mie_emitted_freq_std_dev
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Observation_PCD%mie_emitted_freq_std_dev = value_FAdoxy

    ! read rayleigh_mean_emitted_freq
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Observation_PCD%rayleigh_mean_emitted_freq = value_FAdoxy

    ! read rayleigh_emitted_freq_std_dev
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Observation_PCD%rayleigh_emitted_freq_std_dev = value_FAdoxy

    ! read multimode_ratio_quality_flag
    call read_logical1_from_bin_file(fileunit,value_logical,error_flag)
    IF (error_flag .ne. no_error) return
    Observation_PCD%multimode_ratio_quality_flag = value_logical

!    ! read Quality_Characteristics
!    call read_quality_characteristics(fileunit,Observation_PCD%quality_characteristics,error_flag)
!    IF (error_flag .ne. no_error) return

    ! skip the 4 byte spare_3
    ! to be replaced by a call to: read_zero_bytes_bin_file
    call AdvanceFilePointer(fileunit,4_i8_,error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine Read_Observation_PCD
    !  #]
  !---------------------------------------------------
  subroutine Init_Measurement_PCD(Measurement_PCD, N_range_gates, error_flag)
    !  #[
    TYPE(Measurement_PCD_type), intent(inout) :: Measurement_PCD ! output
    integer,                    intent(in)    :: N_range_gates   ! input
    integer,                    intent(out)   :: error_flag      ! output

    ! local variables
    integer :: AllocStatus, r

    error_flag = no_error

    Measurement_PCD%num_mie_invalid_ref_pulses      = missing_indicator_IntAl
    Measurement_PCD%num_rayleigh_invalid_ref_pulses = missing_indicator_IntAl
    Measurement_PCD%avg_laser_freq_offset           = missing_indicator_FAdoxy
    Measurement_PCD%avg_UV_energy                   = missing_indicator_FAdoxy
    Measurement_PCD%laser_freq_offset_std_dev       = missing_indicator_FAdoxy
    Measurement_PCD%UV_energy_std_dev               = missing_indicator_FAdoxy

    IF (N_range_gates .eq. 0) THEN
       ! init it to be empty
       nullify(Measurement_PCD%Meas_Alt_Bin_PCD)
    ELSE
       allocate(Measurement_PCD%Meas_Alt_Bin_PCD(N_range_gates+1),&
            stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_Measurement_PCD:"
          error_flag = error_allocate
          return
       ENDIF
       
       DO r=1,N_range_gates+1
          call Init_Meas_Alt_Bin_PCD(Measurement_PCD%Meas_Alt_Bin_PCD(r))
       END DO
    END IF

    Measurement_PCD%vel_of_att_uncertainty_error   = missing_indicator_FAdoxy

    Measurement_PCD%mie_mean_emitted_freq          = missing_indicator_FAdoxy
    Measurement_PCD%mie_emitted_std_dev            = missing_indicator_FAdoxy
    Measurement_PCD%reference_pulse_FWHM           = missing_indicator_FAdoxy
    Measurement_PCD%rayleigh_mean_emitted_freq     = missing_indicator_FAdoxy
    Measurement_PCD%rayleigh_emitted_freq_std_dev  = missing_indicator_FAdoxy
    Measurement_PCD%uv_energy_quality_flag         = .FALSE.

!    call Init_QualityCharacteristics(Measurement_PCD%quality_characteristics)

  end subroutine Init_Measurement_PCD
    !  #]
  subroutine Delete_Measurement_PCD(Measurement_PCD)
    !  #[
    TYPE(Measurement_PCD_type), intent(inout) :: Measurement_PCD ! output

    IF (associated(Measurement_PCD%Meas_Alt_Bin_PCD)) THEN
       deallocate(Measurement_PCD%Meas_Alt_Bin_PCD)
       nullify(Measurement_PCD%Meas_Alt_Bin_PCD)
    END IF

  end subroutine Delete_Measurement_PCD
    !  #]
  subroutine Print_Measurement_PCD(Measurement_PCD)
    !  #[
    TYPE(Measurement_PCD_type), intent(in) :: Measurement_PCD
    
    ! local variables
    integer :: s, r
    character(len=13) :: txt

    print *,"Measurement_PCD%..."
    txt = int2txt(Measurement_PCD%num_mie_invalid_ref_pulses)
    print *,"...%num_mie_invalid_ref_pulses             = ",txt
    txt = int2txt(Measurement_PCD%num_rayleigh_invalid_ref_pulses)
    print *,"...%num_rayleigh_invalid_ref_pulses        = ",txt
    txt = real2txt(Measurement_PCD%avg_laser_freq_offset)
    print *,"...%avg_laser_freq_offset                  = ",txt
    txt = real2txt(Measurement_PCD%avg_UV_energy)
    print *,"...%avg_UV_energy                          = ",txt
    txt = real2txt(Measurement_PCD%laser_freq_offset_std_dev)
    print *,"...%laser_freq_offset_std_dev              = ",txt
    txt = real2txt(Measurement_PCD%UV_energy_std_dev)
    print *,"...%UV_energy_std_dev                      = ",txt

    IF (associated(Measurement_PCD%Meas_Alt_Bin_PCD)) THEN
       s = size(Measurement_PCD%Meas_Alt_Bin_PCD)
       print "(a,i2,a)","==> size(Measurement_PCD%Meas_Alt_Bin_PCD) = ",s,&
            " [printing first and last one]"
       DO r=1,s,s-1
          print "(' Measurement_PCD%Meas_Alt_Bin_PCD(',i2.2,')%...')",r
          call Print_Meas_Alt_Bin_PCD(Measurement_PCD%Meas_Alt_Bin_PCD(r))
       END DO
    ELSE
       print *,"...%Meas_Alt_Bin_PCD [not associated]"
    END IF

    print *,"Measurement_PCD%...(continued)"
    txt = real2txt(Measurement_PCD%vel_of_att_uncertainty_error)
    print *,"...%vel_of_att_uncertainty_error   = ",txt

    txt = real2txt(Measurement_PCD%mie_mean_emitted_freq)
    print *,"...%mie_mean_emitted_freq          = ",txt
    txt = real2txt(Measurement_PCD%mie_emitted_std_dev)
    print *,"...%mie_emitted_std_dev            = ",txt
    txt = real2txt(Measurement_PCD%reference_pulse_FWHM)
    print *,"...%reference_pulse_FWHM           = ",txt
    txt = real2txt(Measurement_PCD%rayleigh_mean_emitted_freq)
    print *,"...%rayleigh_mean_emitted_freq     = ",txt
    txt = real2txt(Measurement_PCD%rayleigh_emitted_freq_std_dev)
    print *,"...%rayleigh_emitted_freq_std_dev  = ",txt
    print *,"...%uv_energy_quality_flag         = ", &
      Measurement_PCD%uv_energy_quality_flag

!    call Print_QualityCharacteristics(Measurement_PCD%quality_characteristics)

  end subroutine Print_Measurement_PCD
    !  #]
  subroutine Read_Measurement_PCD(fileunit, Measurement_PCD, &
                               N_range_gates, error_flag)
    !  #[
    integer,                    intent(in)    :: fileunit        ! input
    TYPE(Measurement_PCD_type), intent(inout) :: Measurement_PCD ! result
    integer,                    intent(in)    :: N_range_gates   ! input
    integer,                    intent(out)   :: error_flag      ! result

    ! local variables
    integer(IntAl) :: value_IntAl
    real(FAdoxy)   :: value_FAdoxy
    integer        :: r
    logical        :: value_logical

    error_flag = no_error

    ! skip the 96 bytes
    ! to be replaced by a call to: read_zero_bytes_bin_file
!    call AdvanceFilePointer(fileunit,96_i8_,error_flag)
!    IF (error_flag .ne. no_error) return

    call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    Measurement_PCD%num_mie_invalid_ref_pulses = value_IntAl

    call read_IntAl_from_bin_file(fileunit,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    Measurement_PCD%num_rayleigh_invalid_ref_pulses = value_IntAl
    
    ! read avg_laser_freq_offset
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Measurement_PCD%avg_laser_freq_offset = value_FAdoxy

    ! read avg_UV_energy
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Measurement_PCD%avg_UV_energy = value_FAdoxy

    ! read laser_freq_offset_std_dev
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Measurement_PCD%laser_freq_offset_std_dev = value_FAdoxy

    ! read UV_energy_std_dev
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Measurement_PCD%UV_energy_std_dev = value_FAdoxy

    IF (.not. associated(Measurement_PCD%Meas_Alt_Bin_PCD)) THEN
       print *,"ERROR in Read_Measurement_PCD()"
       print *,"the datastructure: Measurement_PCD%Meas_Alt_Bin_PCD(:)"
       print *,"seems to be not associated ...."
       error_flag = error_programming
       return
    END IF

    ! check if the struct was initialised with the correct nr of rangegates
    IF (N_range_gates .ne. size(Measurement_PCD%Meas_Alt_Bin_PCD)) THEN
       print *,"ERROR in Read_Measurement_PCD()"
       print *,"the datastructure Measurement_PCD%Meas_Alt_Bin_PCD(:) ",&
            "has room for ",size(Measurement_PCD%Meas_Alt_Bin_PCD)
       print *,"rangegates, but the nr of rangegates in the current"
       print *,"datafile seems to be ",N_range_gates
       error_flag = error_programming
       return
    END IF

    ! read the Meas_Alt_Bin_PCD structure for the 24 rangegates
    DO r=1,N_range_gates ! 25
       call Read_Meas_Alt_Bin_PCD(fileunit,&
            Measurement_PCD%Meas_Alt_Bin_PCD(r), error_flag, &
            workaround_in=.TRUE.)
       IF (error_flag .ne. no_error) return
    END DO
    
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Measurement_PCD%vel_of_att_uncertainty_error = value_FAdoxy

    ! read mie_mean_emitted_freq
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Measurement_PCD%mie_mean_emitted_freq = value_FAdoxy

! PP: L1BPv1.04 does not contain the following item although it is
! listed in the IODD dated oct 30, 2006
! DT: Contained in post-FAT TDS, 20061114.
    ! read mie_emitted_std_dev
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Measurement_PCD%mie_emitted_std_dev = value_FAdoxy

    ! read reference_pulse_FWHM
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Measurement_PCD%reference_pulse_FWHM = value_FAdoxy

    ! read rayleigh_mean_emitted_freq
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Measurement_PCD%rayleigh_mean_emitted_freq = value_FAdoxy

    ! read rayleigh_emitted_freq_std_dev
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Measurement_PCD%rayleigh_emitted_freq_std_dev = value_FAdoxy

!    ! read Quality_Characteristics
!    call read_quality_characteristics(fileunit,Measurement_PCD%quality_characteristics,error_flag)
!    IF (error_flag .ne. no_error) return

    ! read uv_energy_quality_flag
    call read_logical1_from_bin_file(fileunit,value_logical,error_flag)
    IF (error_flag .ne. no_error) return
    Measurement_PCD%uv_energy_quality_flag = value_logical

    ! skip the 8 byte spare_4
    ! to be replaced by a call to: read_zero_bytes_bin_file
    call AdvanceFilePointer(fileunit,8_i8_,error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine Read_Measurement_PCD
    !  #]
  !---------------------------------------------------
  subroutine Init_L1B_PCD_ADS(PCD_ADS, N_meas, N_range_gates, error_flag)
    !  #[
    TYPE(L1B_PCD_ADS_type), pointer     :: PCD_ADS         ! output
    integer,                intent(in)  :: N_meas          ! input
    integer,                intent(in)  :: N_range_gates   ! input
    integer,                intent(out) :: error_flag      ! output

    ! local variables
    integer :: AllocStatus, i

    error_flag = no_error
    IF (.not. associated(PCD_ADS) ) THEN
       print *,"ERROR in Init_PCD_ADS:"
       print *,"the pointer PCD_ADS seems not associated !!!"
       error_flag = error_programming
       return
    ENDIF

    call InitDateTime(PCD_ADS%start_of_obs_datetime)

    PCD_ADS%nmeas   = missing_indicator_IntAs
    PCD_ADS%npulses = missing_indicator_IntAs

    call Init_Observation_PCD(PCD_ADS%Observation_PCD, N_range_gates, &
                              error_flag)
    IF (error_flag .ne. no_error) return

    IF (N_meas .eq. 0) THEN
       ! init it to be empty
       nullify(PCD_ADS%Measurement_PCD)
    ELSE
       allocate(PCD_ADS%Measurement_PCD(1:N_meas),stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_PCD_ADS"
          error_flag = error_allocate
          return
       END IF
       DO i=1,N_meas
          call Init_Measurement_PCD(PCD_ADS%Measurement_PCD(i),N_range_gates, &
                                    error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    END IF

  end subroutine Init_L1B_PCD_ADS
    !  #]
  subroutine Delete_L1B_PCD_ADS(PCD_ADS)
    !  #[

    TYPE(L1B_PCD_ADS_type), pointer :: PCD_ADS ! input/output

    ! local variable
    integer :: m

    IF (.not. associated(PCD_ADS) ) return  ! nothing to do

    call Delete_Observation_PCD(PCD_ADS%Observation_PCD)

    IF (associated(PCD_ADS%Measurement_PCD)) THEN
       DO m=1,size(PCD_ADS%Measurement_PCD)
          call Delete_Measurement_PCD(PCD_ADS%Measurement_PCD(m))
       END DO
       deallocate(PCD_ADS%Measurement_PCD)
       nullify(PCD_ADS%Measurement_PCD)
    END IF

  end subroutine Delete_L1B_PCD_ADS
    !  #]
  subroutine Print_L1B_PCD_ADS(PCD_ADS)
    !  #[
    TYPE(L1B_PCD_ADS_type), pointer :: PCD_ADS ! input

     ! local variables
    integer           :: m, s
    character(len=13) :: txt
    character(len=maxlen_DateTime_KVT) :: DateTimeString

    IF (.not. associated(PCD_ADS)) THEN
       print *,"PCD_ADS [seems not associated]"
       return
    END IF

    print *,"PCD_ADS%..."
    
    ! print the date and time
    DateTimeString = Construct_KVT_DateTime(PCD_ADS%start_of_obs_datetime)
    print *,"...%start_of_obs_datetime: "//trim(DateTimeString)

    txt = int2txt(PCD_ADS%nmeas)
    print *,"...%nmeas   = ",txt
    txt = int2txt(PCD_ADS%npulses)
    print *,"...%npulses = ",txt

    call Print_Observation_PCD(PCD_ADS%Observation_PCD)
    IF (associated(PCD_ADS%Measurement_PCD)) THEN
       s = size(PCD_ADS%Measurement_PCD)
       print "(a,i2,a)","==> size(PCD_ADS%Measurement_PCD) = ",s,&
            " [printing first and last one]"
       DO m=1,s,s-1
          print "(' PCD_ADS%Measurement_PCD(',i2.2,')%...')",m
          call Print_Measurement_PCD(PCD_ADS%Measurement_PCD(m))
       END DO
    ELSE
       print *,"PCD_ADS%Measurement_PCD(:) = [NOT ASSOCIATED]"
    END IF

  end subroutine Print_L1B_PCD_ADS
    !  #]
  subroutine Read_L1B_PCD_ADS(PCD_DSD,fileunit,which_DS,N_Meas_Max,&
                              N_range_gates, PCD_ADS,error_flag)
    !  #[
    TYPE(DSD_type),         pointer     :: PCD_DSD    ! input
    integer,                intent(in)  :: fileunit   ! input
    integer,                intent(in)  :: which_DS   ! input
    integer,                intent(in)  :: N_Meas_Max    ! input
    !                       (used for checking validity of the dataset !)
    integer,                intent(in)  :: N_range_gates ! input
    !                       (used for checking validity of the dataset !)
    TYPE(L1B_PCD_ADS_type), pointer     :: PCD_ADS    ! result
    integer,                intent(out) :: error_flag ! result

    ! local variables
    integer(i8_)      :: pos, endpos
    integer           :: currentpos
    integer(IntAs)    :: value_IntAs

    ! local variables
    integer :: m

    IF (.not. associated(PCD_DSD)) THEN
       print *,"ERROR in Read_L1B_PCD_ADS:"
       print *,"the pointer PCD_DSD seems not associated"
       error_flag = error_programming
       return
    END IF

    IF (.not. associated(PCD_ADS)) THEN
       print *,"ERROR in Read_L1B_PCD_ADS:"
       print *,"the pointer PCD_ADS seems not associated"
       error_flag = error_programming
       return
    END IF

    pos = PCD_DSD%start_of_ds + (which_DS-1)*PCD_DSD%size_of_one_dsr
    call SetFilePointer(fileunit,pos,error_flag)
    IF (error_flag .ne. no_error) return

    print *,"trying to read the binary PCD_ADS nr. ",which_DS

    ! read the 12 byte datetime field
    call read_DateTime_from_bin_file(fileunit,&
         PCD_ADS%start_of_obs_datetime,error_flag)

    ! read 2 byte integer
    call read_IntAs_from_bin_file(fileunit,value_IntAs,error_flag)
    IF (error_flag .ne. no_error) return
    PCD_ADS%nmeas = value_IntAs

    ! read 2 byte integer
    call read_IntAs_from_bin_file(fileunit,value_IntAs,error_flag)
    IF (error_flag .ne. no_error) return
    PCD_ADS%npulses = value_IntAs

    ! skip the 8 byte spare_1
    ! to be replaced by a call to: read_zero_bytes_bin_file
    call AdvanceFilePointer(fileunit,8_i8_,error_flag)
    IF (error_flag .ne. no_error) return

    call Read_Observation_PCD(fileunit, PCD_ADS%Observation_PCD,&
                          N_range_gates+1, error_flag)
    IF (error_flag .ne. no_error) return

    IF (.not. associated(PCD_ADS%Measurement_PCD)) THEN
       print *,"ERROR in Read_L1B_PCD_ADS()"
       print *,"the datastructure L1B%PCD_ADS%Measurement_PCD(:) seems to be"
       print *,"not associated ...."
       error_flag = error_programming
       return
    END IF

    ! check if the struct was initialised with the correct nr of measurements
    IF (N_meas_Max .ne. size(PCD_ADS%Measurement_PCD)) THEN
       print *,"ERROR in Read_L1B_PCD_ADS()"
       print *,"the datastructure L1B%PCD_ADS%Measurement_PCD(:) ",&
            "has room for ",size(PCD_ADS%Measurement_PCD)
       print *,"measurements, but the nr of measurements in the current"
       print *,"datafile seems to be ",N_meas_Max
       error_flag = error_programming
       return
    END IF

    ! read the Measurement_PCD structure for the 67 measurements
    DO m=1,N_Meas_Max ! 67
       call  Read_Measurement_PCD(fileunit,PCD_ADS%Measurement_PCD(m),&
                               N_range_gates+1, error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! skip the 8 byte spare_2
    ! to be replaced by a call to: read_zero_bytes_bin_file
    call AdvanceFilePointer(fileunit,8_i8_,error_flag)
    IF (error_flag .ne. no_error) return

    ! calculate expected end-position after reading this block of data
    endpos = PCD_DSD%start_of_ds + which_DS*PCD_DSD%size_of_one_dsr
    ! request the current file position
    inquire(unit=fileunit,nextrec=currentpos)
    
    IF (currentpos .ne. endpos) THEN
       print *,"ERROR in Read_L1B_PCD_ADS():"
       print *,"there seems to be a problem while reading this binary dataset"
       print *,"expected file position after this DSR is: ",endpos
       print *,"but the actual position seems to be:      ",currentpos
       print *,"Maybe the fileformat definition has changed ???"
       error_flag = error_reading_file
       return
    END IF

  end subroutine Read_L1B_PCD_ADS
    !  #]
  !---------------------------------------------------
end module L1B_PCD_ADS
