module L1B_SPH
  !  #[ Documentation

  !
  ! this module defines the Specific Product Header for the ee_cfi fileformat
  ! for the L1B processor wind measurement output data files 
  ! The SPH format may differ for different filedefinitions, therefore
  ! we use a private definition for this filetype only
  !
  ! for a description, see 
  ! L1B_iodd, version 2, (6-5-2005), sec.6.2, p.6-2, table 6-1
  !
  ! Written by J. de Kloe (KNMI)
  !
  !     Modifications:
  !        29-Sep-2005  J. de Kloe   Original version
  !        19-Oct-2005  D. Tan       Add Measurement_DSD + Geoid_Model_DSD
  !        02-Mar-2006  J. de Kloe   adapted to changes in XML_module
  !        13-Mar-2006  J. de Kloe   added error_flag/prefix to print_DSD 
  !        12-May-2006  D. Tan       Null use of ParKind1 module
  !        03-Nov-2006  P. Poli      upgrade to L1BP v1.04 ;
  !                                  moved and cleaned-up reading part
  !        07-May-2007  J. de Kloe   added dsd handling in Read_L1B_SPH_KVT
  !                                  and adapted to enable reading L1BP v1.06 
  !                                  files
  !        08-Jun-2007  J. de Kloe   adapted to changed xml interface, and
  !                                  implemented Get_L1B_SPH_from_xml_tree()
  !        15-Jun-2007  J. de Kloe   changed to check for integer L1B version 
  !                                  numbers when doing a few workarounds
  !        03-Sep-2007  P. Poli      added tests to prevent allocation of 
  !                                  DSDs if they have already been allocated
  !                                  inside routine Get_L1B_SPH_from_xml_tree
  !        03-Jan-2008  J. de Kloe   added nullify statements to make the code
  !                                  more robust
  !        16-Jan-2008  J. de Kloe   phase out integer kind i_ 
  !        18-Jan-2008  J. de Kloe   implemented workaround for changed tag
  !                                  name for the DSD's in v1.08
  !        21-Feb-2008  J. de Kloe   adapted to L1B v1.09
  !  #] 
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE XML_module, only: xml_node_type, xml_node_get_string, &
                        ee_h_name, ee_vh_name, ee_sph_name, &
                        check_length_xml_item
!                        check_string_length

  USE EE_CFI_Datatypes, only: maxlen_IntAl, maxlen_FAdoxy
  USE DataSetDescriptor, only: DSD_type, Init_DSD, Print_DSD, &
                               maxlen_DS_Name, maxlen_DS_Type, &
                               Read_DSD_from_KVT, Get_DSD_from_xml_tree
  ! support modules
  USE KVT_module, only: max_str_len, Init_KVT_module, &
                        read_kvt_string, read_spare_kvt_string, &
                        Compare_XML_to_KVT_item_Repl
  USE ErrorHandler, only: no_error, error_allocate, error_programming, &
                          error_reading_file
  USE Numerics, only: i8_, i4_

  !  #]
  !  #[ Type definitions and parameters
  implicit none

  ! Names of reference datasets in L1B products
  character(len=maxlen_DS_Name) ,parameter :: L1B_DS_Name_Geolocation_ADS=&
                                              "Geolocation_ADS"
  character(len=maxlen_DS_Name) ,parameter :: L1B_DS_Name_Prod_Conf_Data_ADS=&
                                              "Product_Confidence_Data_ADS"
  character(len=maxlen_DS_Name) ,parameter :: L1B_DS_Name_Ground_Wind_Det_ADS=&
                                              "Ground_Wind_Detection_ADS"
  character(len=maxlen_DS_Name) ,parameter :: L1B_DS_Name_Measurement_ADS=&
                                              "Measurement_ADS"
  character(len=maxlen_DS_Name) ,parameter :: L1B_DS_Name_Calibr_Char_GADS=&
                                              "Calibration_Char_GADS"
  character(len=maxlen_DS_Name) ,parameter :: L1B_DS_Name_Useful_Signal_DSD=&
                                              "Useful_Signal_MDS"
  character(len=maxlen_DS_Name) ,parameter :: L1B_DS_Name_Wind_Velocity_MDS=&
                                              "Wind_Velocity_MDS"
  character(len=maxlen_DS_Name) ,parameter :: L1B_DS_Name_Aeolus_L1AP=&
                                              "Aeolus_Level_1A_Product"
  character(len=maxlen_DS_Name) ,parameter :: L1B_DS_Name_Aeolus_MRC=&
                                              "Aeolus_MRC"
  character(len=maxlen_DS_Name) ,parameter :: L1B_DS_Name_Aeolus_RRC=&
                                              "Aeolus_RRC"
  character(len=maxlen_DS_Name) ,parameter :: L1B_DS_Name_Aeolus_ZWC=&
                                              "Aeolus_ZWC"
  character(len=maxlen_DS_Name) ,parameter :: L1B_DS_Name_L1B_Proc_Pars=&
                                              "Level_1B_Proc_Params"
  character(len=maxlen_DS_Name) ,parameter :: L1B_DS_Name_Sat_Char=&
                                              "Satellite_Characterisation"
  character(len=maxlen_DS_Name) ,parameter :: L1B_DS_Name_Orbit_Scenario_File=&
                                              "Orbit_Scenario_File"
  character(len=maxlen_DS_Name) ,parameter :: L1B_DS_Name_DEM=&
                                              "Digital_Elevation_Model"
  character(len=maxlen_DS_Name) ,parameter :: L1B_DS_Name_Geoid_Model=&
                                              "Geoid_Model"

  ! Types (Annotation/Measurement) of attached datasets in L1B products
  character(len=maxlen_DS_Type) ,parameter :: L1B_DS_Type_Geolocation_ADS     = "A"
  character(len=maxlen_DS_Type) ,parameter :: L1B_DS_Type_Prod_Conf_Data_ADS  = "A"
  character(len=maxlen_DS_Type) ,parameter :: L1B_DS_Type_Ground_Wind_Det_ADS = "A"
  character(len=maxlen_DS_Type) ,parameter :: L1B_DS_Type_Measurement_ADS     = "A"
  character(len=maxlen_DS_Type) ,parameter :: L1B_DS_Type_Calibr_Char_GADS    = "G"
  character(len=maxlen_DS_Type) ,parameter :: L1B_DS_Type_Useful_Signal_DSD   = "M"
  character(len=maxlen_DS_Type) ,parameter :: L1B_DS_Type_Wind_Velocity_MDS   = "M"
  character(len=maxlen_DS_Type) ,parameter :: L1B_DS_Type_Aeolus_L1AP         = "R"
  character(len=maxlen_DS_Type) ,parameter :: L1B_DS_Type_Aeolus_MRC          = "R"
  character(len=maxlen_DS_Type) ,parameter :: L1B_DS_Type_Aeolus_RRC          = "R"
  character(len=maxlen_DS_Type) ,parameter :: L1B_DS_Type_Aeolus_ZWC          = "R"
  character(len=maxlen_DS_Type) ,parameter :: L1B_DS_Type_L1B_Proc_Pars       = "R"
  character(len=maxlen_DS_Type) ,parameter :: L1B_DS_Type_Sat_Char            = "R"
  character(len=maxlen_DS_Type) ,parameter :: L1B_DS_Type_Orbit_Scenario_File = "R"
  character(len=maxlen_DS_Type) ,parameter :: L1B_DS_Type_DEM                 = "R"
  character(len=maxlen_DS_Type) ,parameter :: L1B_DS_Type_Geoid_Model         = "R"

  integer, parameter :: maxlen_SPH_Descriptor = 28

  TYPE L1B_SPH_type ! see section 6.2 of the L1B_iodd_2.1
     !  #[ the L1B specific product header
     character(len=maxlen_SPH_Descriptor) :: SPH_Descriptor
     character(len=maxlen_IntAl)          :: Intersect_Start_Lat
     character(len=maxlen_IntAl)          :: Intersect_Start_Long
     character(len=maxlen_IntAl)          :: Intersect_Stop_Lat
     character(len=maxlen_IntAl)          :: Intersect_Stop_Long
     character(len=maxlen_FAdoxy)         :: Sat_Track
     ! Spare 1
     !                                       1234567890123456789012345678901
     character(len=maxlen_FAdoxy)         :: Base_Laser_Frequency
     character(len=maxlen_IntAl)          :: N_Max
     character(len=maxlen_IntAl)          :: N_Max_Actual
     character(len=maxlen_IntAl)          :: Total_Num_of_Observations
     character(len=maxlen_IntAl)          :: Total_Num_of_Measurements
     character(len=maxlen_IntAl)          :: Total_Num_of_Reference_Pulses
     ! Spare 2
     character(len=maxlen_IntAl)          :: Num_of_Mie_Obs_Used
     character(len=maxlen_IntAl)          :: Num_of_Rayleigh_Obs_Used
     character(len=maxlen_IntAl)          :: Num_of_Mie_Meas_Used
     character(len=maxlen_IntAl)          :: Num_of_Rayleigh_Meas_Used
     character(len=maxlen_IntAl)          :: Num_of_Mie_Ref_Pulses_Used
     character(len=maxlen_IntAl)          :: Num_of_Rayleigh_Ref_Pulses_Used
     ! Spare 3
     !                                       1234567890123456789012345678901
     character(len=maxlen_IntAl)          :: Num_of_Mie_Zero_Wind_Detected
     character(len=maxlen_IntAl)          :: Num_of_Rayleigh_Zero_Wind_Det
     character(len=maxlen_IntAl)          :: Num_of_Mie_Ground_Echo_Bins
     character(len=maxlen_IntAl)          :: Num_of_Rayleigh_Grnd_Echo_Bins
     ! Spare 4
     character(len=maxlen_IntAl)          :: Tnum_of_Meas_Laser_Fr_Unlocked
     character(len=maxlen_IntAl)          :: Tnum_of_Ref_Pls_Lsr_Fr_Unlocked
     character(len=maxlen_IntAl)          :: Tnum_of_Sat_Not_on_Target_Meas
     character(len=maxlen_IntAl)          :: Tnum_of_Corrupt_Mie_Meas
     character(len=maxlen_IntAl)          :: Tnum_of_Corrupt_Rayleigh_Meas
     character(len=maxlen_IntAl)          :: Tnum_of_Corrupt_Mie_Ref_Pulses
     character(len=maxlen_IntAl)          :: Tnum_of_Corr_Rayleigh_Ref_Pulss
     ! added in L1B v1.04:
     character(len=maxlen_IntAl)          :: NF_Order
     ! Spare 5
     !                                       1234567890123456789012345678901
     ! the list of 16 DSD's
     logical :: DSDs_are_available ! set .true. when reading was successfull
     TYPE(DSD_type), pointer :: Geolocation_DSD
     TYPE(DSD_type), pointer :: Product_Conf_Data_DSD
     TYPE(DSD_type), pointer :: Ground_Wind_Detection_DSD
     TYPE(DSD_type), pointer :: Measurement_DSD
     TYPE(DSD_type), pointer :: Cal_DSD
     TYPE(DSD_type), pointer :: Useful_Signal_DSD
     TYPE(DSD_type), pointer :: Wind_Velocity_DSD
     TYPE(DSD_type), pointer :: Aeolus_L1A_Product_DSD
     TYPE(DSD_type), pointer :: Aeolus_MRC_DSD
     TYPE(DSD_type), pointer :: Aeolus_RRC_DSD
     TYPE(DSD_type), pointer :: Aeolus_ZWC_DSD
     TYPE(DSD_type), pointer :: L1B_Proc_Params_DSD
     TYPE(DSD_type), pointer :: Sat_Characterisation_DSD
     TYPE(DSD_type), pointer :: Orbit_Scenario_File_DSD
     TYPE(DSD_type), pointer :: Digital_Elevation_Model_DSD ! added in v1.06
     TYPE(DSD_type), pointer :: Geoid_Model_DSD
     !
     ! the next items are numerical values derived from the stringvalues above
     ! (since only numbers we actually use need to be converted)
     integer :: N_Meas_Max    ! corresponds to N_Max
     integer :: NF_Order_Max
  END TYPE L1B_SPH_type
    !  #]

  !  #]
  !  #[ Module interface

  ! Type definitions
  public :: L1B_SPH_type  
  ! subroutines
  public :: Init_L1B_SPH, Print_L1B_SPH, Delete_L1B_SPH, Read_L1B_SPH_KVT, &
            Get_L1B_SPH_from_xml_tree
  !  #]
contains
  !---------------------------------
  subroutine Init_L1B_SPH(L1B_SPH,error_flag)
    !  #[
    TYPE(L1B_SPH_type), pointer :: L1B_SPH
    integer, intent(out) :: error_flag
    
    error_flag = no_error
    IF (.not. associated(L1B_SPH)) THEN
       print *,"ERROR in Init_L1B_SPH():"
       print *,"The pointer to this L1B_SPH seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! see the remark in MainProductHeader.F90 for the reason why I use
    ! the form x(:)=' ' to initialise the strings

    L1B_SPH%SPH_Descriptor(:)       = ' '
    L1B_SPH%Intersect_Start_Lat(:)  = ' '
    L1B_SPH%Intersect_Start_Long(:) = ' '
    L1B_SPH%Intersect_Stop_Lat(:)   = ' '
    L1B_SPH%Intersect_Stop_Long(:)  = ' '
    L1B_SPH%Sat_Track(:)            = ' '
    ! Spare 1
    L1B_SPH%Base_Laser_Frequency(:)          = ' '
    L1B_SPH%N_Max(:)                         = ' '
    L1B_SPH%N_Max_Actual(:)                  = ' '
    L1B_SPH%Total_Num_of_Observations(:)     = ' '
    L1B_SPH%Total_Num_of_Measurements(:)     = ' '
    L1B_SPH%Total_Num_of_Reference_Pulses(:) = ' '
    ! Spare 2
    L1B_SPH%Num_of_Mie_Obs_Used(:)             = ' '
    L1B_SPH%Num_of_Rayleigh_Obs_Used(:)        = ' '
    L1B_SPH%Num_of_Mie_Meas_Used(:)            = ' '
    L1B_SPH%Num_of_Rayleigh_Meas_Used(:)       = ' '
    L1B_SPH%Num_of_Mie_Ref_Pulses_Used(:)      = ' '
    L1B_SPH%Num_of_Rayleigh_Ref_Pulses_Used(:) = ' '
    ! Spare 3
    L1B_SPH%Num_of_Mie_Zero_Wind_Detected(:)   = ' '
    L1B_SPH%Num_of_Rayleigh_Zero_Wind_Det(:)   = ' '
    L1B_SPH%Num_of_Mie_Ground_Echo_Bins(:)     = ' '
    L1B_SPH%Num_of_Rayleigh_Grnd_Echo_Bins(:)  = ' '
    ! Spare 4
    L1B_SPH%Tnum_of_Meas_Laser_Fr_Unlocked(:)  = ' '
    L1B_SPH%Tnum_of_Ref_Pls_Lsr_Fr_Unlocked(:) = ' '
    L1B_SPH%Tnum_of_Sat_Not_on_Target_Meas(:)  = ' '
    L1B_SPH%Tnum_of_Corrupt_Mie_Meas(:)        = ' '
    L1B_SPH%Tnum_of_Corrupt_Rayleigh_Meas(:)   = ' '
    L1B_SPH%Tnum_of_Corrupt_Mie_Ref_Pulses(:)  = ' '
    L1B_SPH%Tnum_of_Corr_Rayleigh_Ref_Pulss(:) = ' '
    L1B_SPH%NF_Order(:)                        = ' '
    ! Spare 5
    ! the list of 14 DSD's
    L1B_SPH%DSDs_are_available = .false.
    nullify(L1B_SPH%Geolocation_DSD)
    nullify(L1B_SPH%Product_Conf_Data_DSD)
    nullify(L1B_SPH%Ground_Wind_Detection_DSD)
    nullify(L1B_SPH%Measurement_DSD)
    nullify(L1B_SPH%Cal_DSD)
    nullify(L1B_SPH%Useful_Signal_DSD)
    nullify(L1B_SPH%Wind_Velocity_DSD)
    nullify(L1B_SPH%Aeolus_L1A_Product_DSD)
    nullify(L1B_SPH%Aeolus_MRC_DSD)
    nullify(L1B_SPH%Aeolus_RRC_DSD)
    nullify(L1B_SPH%Aeolus_ZWC_DSD)
    nullify(L1B_SPH%L1B_Proc_Params_DSD)
    nullify(L1B_SPH%Sat_Characterisation_DSD)
    nullify(L1B_SPH%Orbit_Scenario_File_DSD)
    nullify(L1B_SPH%Digital_Elevation_Model_DSD)
    nullify(L1B_SPH%Geoid_Model_DSD)

    ! the derived numerical values
    L1B_SPH%N_Meas_Max = -1
    L1B_SPH%NF_Order_Max = -1
  end subroutine Init_L1B_SPH
    !  #]
  subroutine Print_L1B_SPH(L1B_SPH,error_flag)
    !  #[
    TYPE(L1B_SPH_type), pointer     :: L1B_SPH
    integer,            intent(out) :: error_flag

    ! local variable
    character(len=256) :: txt

    error_flag = no_error

    IF (.not. associated(L1B_SPH)) THEN
       print *,"[pointer to L1B_SPH is not associated so nothing to print]"
       return
    END IF

    txt = L1B_SPH%SPH_Descriptor
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%SPH_Descriptor                  = ",trim(txt)

    txt = L1B_SPH%Intersect_Start_Lat
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Intersect_Start_Lat             = ",trim(txt)

    txt = L1B_SPH%Intersect_Start_Long
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Intersect_Start_Long            = ",trim(txt)

    txt = L1B_SPH%Intersect_Stop_Lat
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Intersect_Stop_Lat              = ",trim(txt)

    txt = L1B_SPH%Intersect_Stop_Long
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Intersect_Stop_Long             = ",trim(txt)

    txt = L1B_SPH%Sat_Track
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Sat_Track                       = ",trim(txt)

    ! Spare 1

    txt = L1B_SPH%Base_Laser_Frequency
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Base_Laser_Frequency            = ",trim(txt)

    txt = L1B_SPH%N_Max
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%N_Max                           = ",trim(txt)

    txt = L1B_SPH%N_Max_Actual
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%N_Max_Actual                    = ",trim(txt)

    txt = L1B_SPH%Total_Num_of_Observations
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Total_Num_of_Observations       = ",trim(txt)

    txt = L1B_SPH%Total_Num_of_Measurements
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Total_Num_of_Measurements       = ",trim(txt)

    txt = L1B_SPH%Total_Num_of_Reference_Pulses
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Total_Num_of_Reference_Pulses   = ",trim(txt)

    ! Spare 2

    txt = L1B_SPH%Num_of_Mie_Obs_Used
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_of_Mie_Obs_Used             = ",trim(txt)

    txt = L1B_SPH%Num_of_Rayleigh_Obs_Used
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_of_Rayleigh_Obs_Used        = ",trim(txt)

    txt = L1B_SPH%Num_of_Mie_Meas_Used
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_of_Mie_Meas_Used            = ",trim(txt)

    txt = L1B_SPH%Num_of_Rayleigh_Meas_Used
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_of_Rayleigh_Meas_Used       = ",trim(txt)

    txt = L1B_SPH%Num_of_Mie_Ref_Pulses_Used
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_of_Mie_Ref_Pulses_Used      = ",trim(txt)

    txt = L1B_SPH%Num_of_Rayleigh_Ref_Pulses_Used
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_of_Rayleigh_Ref_Pulses_Used = ",trim(txt)

    ! Spare 3

    txt = L1B_SPH%Num_of_Mie_Zero_Wind_Detected
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_of_Mie_Zero_Wind_Detected   = ",trim(txt)

    txt = L1B_SPH%Num_of_Rayleigh_Zero_Wind_Det
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_of_Rayleigh_Zero_Wind_Det   = ",trim(txt)

    txt = L1B_SPH%Num_of_Mie_Ground_Echo_Bins
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_of_Mie_Ground_Echo_Bins     = ",trim(txt)

    txt = L1B_SPH%Num_of_Rayleigh_Grnd_Echo_Bins
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_of_Rayleigh_Grnd_Echo_Bins  = ",trim(txt)

    ! Spare 4

    txt = L1B_SPH%Tnum_of_Meas_Laser_Fr_Unlocked
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Tnum_of_Meas_Laser_Fr_Unlocked  = ",trim(txt)

    txt = L1B_SPH%Tnum_of_Ref_Pls_Lsr_Fr_Unlocked
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Tnum_of_Ref_Pls_Lsr_Fr_Unlocked  = ",trim(txt)

    txt = L1B_SPH%Tnum_of_Sat_Not_on_Target_Meas
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Tnum_of_Sat_Not_on_Target_Meas   = ",trim(txt)

    txt = L1B_SPH%Tnum_of_Corrupt_Mie_Meas
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Tnum_of_Corrupt_Mie_Meas         = ",trim(txt)

    txt = L1B_SPH%Tnum_of_Corrupt_Rayleigh_Meas
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Tnum_of_Corrupt_Rayleigh_Meas    = ",trim(txt)

    txt = L1B_SPH%Tnum_of_Corrupt_Mie_Ref_Pulses
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Tnum_of_Corrupt_Mie_Ref_Pulses   = ",trim(txt)

    txt = L1B_SPH%Tnum_of_Corr_Rayleigh_Ref_Pulss
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Tnum_of_Corr_Rayleigh_Ref_Pulss  = ",trim(txt)

    txt = L1B_SPH%NF_Order
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%NF_Order                         = ",trim(txt)

    ! Spare_5

    ! list of DSD's
    print *,"...%DSDs_are_available = ",L1B_SPH%DSDs_are_available
    print *,"L1B_SPH%Geolocation_DSD%..."
    call Print_DSD(L1B_SPH%Geolocation_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    print *,"L1B_SPH%Product_Conf_Data_DSD%..."
    call Print_DSD(L1B_SPH%Product_Conf_Data_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    print *,"L1B_SPH%Ground_Wind_Detection_DSD%..."
    call Print_DSD(L1B_SPH%Ground_Wind_Detection_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    print *,"L1B_SPH%Measurement_DSD%..."
    call Print_DSD(L1B_SPH%Measurement_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    print *,"L1B_SPH%Cal_DSD%..."
    call Print_DSD(L1B_SPH%Cal_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    print *,"L1B_SPH%Useful_Signal_DSD%..."
    call Print_DSD(L1B_SPH%Useful_Signal_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    print *,"L1B_SPH%Wind_Velocity_DSD%..."
    call Print_DSD(L1B_SPH%Wind_Velocity_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    print *,"L1B_SPH%Aeolus_L1A_Product_DSD%..."
    call Print_DSD(L1B_SPH%Aeolus_L1A_Product_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    print *,"L1B_SPH%Aeolus_MRC_DSD%..."
    call Print_DSD(L1B_SPH%Aeolus_MRC_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    print *,"L1B_SPH%Aeolus_RRC_DSD%..."
    call Print_DSD(L1B_SPH%Aeolus_RRC_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    print *,"L1B_SPH%Aeolus_ZWC_DSD%..."
    call Print_DSD(L1B_SPH%Aeolus_ZWC_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    print *,"L1B_SPH%L1B_Proc_Params_DSD%..."
    call Print_DSD(L1B_SPH%L1B_Proc_Params_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    print *,"L1B_SPH%Sat_Characterisation_DSD%..."
    call Print_DSD(L1B_SPH%Sat_Characterisation_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    print *,"L1B_SPH%Orbit_Scenario_File_DSD%..."
    call Print_DSD(L1B_SPH%Orbit_Scenario_File_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    if (associated(L1B_SPH%Digital_Elevation_Model_DSD)) then
       print *,"L1B_SPH%Digital_Elevation_Model_DSD%..."
       call Print_DSD(L1B_SPH%Digital_Elevation_Model_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    end if
    print *,"L1B_SPH%Geoid_Model_DSD%..."
    call Print_DSD(L1B_SPH%Geoid_Model_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    
  end subroutine Print_L1B_SPH
    !  #]
  subroutine Delete_L1B_SPH(L1B_SPH)
    !  #[
    TYPE(L1B_SPH_type), pointer :: L1B_SPH

    IF (associated(L1B_SPH%Geolocation_DSD)) THEN
       deallocate(L1B_SPH%Geolocation_DSD)
       nullify(L1B_SPH%Geolocation_DSD)
    END IF

    IF (associated(L1B_SPH%Product_Conf_Data_DSD)) THEN
       deallocate(L1B_SPH%Product_Conf_Data_DSD)
       nullify(L1B_SPH%Product_Conf_Data_DSD)
    END IF

    IF (associated(L1B_SPH%Ground_Wind_Detection_DSD)) THEN
       deallocate(L1B_SPH%Ground_Wind_Detection_DSD)
       nullify(L1B_SPH%Ground_Wind_Detection_DSD)
    END IF

    IF (associated(L1B_SPH%Measurement_DSD)) THEN
       deallocate(L1B_SPH%Measurement_DSD)
       nullify(L1B_SPH%Measurement_DSD)
    END IF

    IF (associated(L1B_SPH%Cal_DSD)) THEN
       deallocate(L1B_SPH%Cal_DSD)
       nullify(L1B_SPH%Cal_DSD)
    END IF

    IF (associated(L1B_SPH%Useful_Signal_DSD)) THEN
       deallocate(L1B_SPH%Useful_Signal_DSD)
       nullify(L1B_SPH%Useful_Signal_DSD)
    END IF

    IF (associated(L1B_SPH%Wind_Velocity_DSD)) THEN
       deallocate(L1B_SPH%Wind_Velocity_DSD)
       nullify(L1B_SPH%Wind_Velocity_DSD)
    END IF

    IF (associated(L1B_SPH%Aeolus_L1A_Product_DSD)) THEN
       deallocate(L1B_SPH%Aeolus_L1A_Product_DSD)
       nullify(L1B_SPH%Aeolus_L1A_Product_DSD)
    END IF

    IF (associated(L1B_SPH%Aeolus_MRC_DSD)) THEN
       deallocate(L1B_SPH%Aeolus_MRC_DSD)
       nullify(L1B_SPH%Aeolus_MRC_DSD)
    END IF

    IF (associated(L1B_SPH%Aeolus_RRC_DSD)) THEN
       deallocate(L1B_SPH%Aeolus_RRC_DSD)
       nullify(L1B_SPH%Aeolus_RRC_DSD)
    END IF

    IF (associated(L1B_SPH%Aeolus_ZWC_DSD)) THEN
       deallocate(L1B_SPH%Aeolus_ZWC_DSD)
       nullify(L1B_SPH%Aeolus_ZWC_DSD)
    END IF

    IF (associated(L1B_SPH%L1B_Proc_Params_DSD)) THEN
       deallocate(L1B_SPH%L1B_Proc_Params_DSD)
       nullify(L1B_SPH%L1B_Proc_Params_DSD)
    END IF

    IF (associated(L1B_SPH%Sat_Characterisation_DSD)) THEN
       deallocate(L1B_SPH%Sat_Characterisation_DSD)
       nullify(L1B_SPH%Sat_Characterisation_DSD)
    END IF

    IF (associated(L1B_SPH%Orbit_Scenario_File_DSD)) THEN
       deallocate(L1B_SPH%Orbit_Scenario_File_DSD)
       nullify(L1B_SPH%Orbit_Scenario_File_DSD)
    END IF

    IF (associated(L1B_SPH%Digital_Elevation_Model_DSD)) THEN
       deallocate(L1B_SPH%Digital_Elevation_Model_DSD)
       nullify(L1B_SPH%Digital_Elevation_Model_DSD)
    END IF

    IF (associated(L1B_SPH%Geoid_Model_DSD)) THEN
       deallocate(L1B_SPH%Geoid_Model_DSD)
       nullify(L1B_SPH%Geoid_Model_DSD)
    END IF

  end subroutine Delete_L1B_SPH
    !  #]
  subroutine Read_L1B_SPH_KVT(SPH, fileunit, &
                              Software_Version_Major_Number, &
                              Software_Version_Minor_Number, &
                              error_flag, Compare)
    !  #[ read the specific product header (SVT values)

    ! compare with the file:
    ! L1bP_Draft_1/aeolus/common/PFR/src/PFR_LevelOneBRead.cc
    ! lines 340-620    

    TYPE(L1B_SPH_type), pointer     :: SPH
    integer,            intent(in)  :: fileunit
    integer,            intent(in)  :: Software_Version_Major_Number
    integer,            intent(in)  :: Software_Version_Minor_Number
    integer,            intent(out) :: error_flag
    logical, optional,  intent(in)  :: Compare

    ! local variables
    character(len=max_str_len) :: exp_name
    character(len=max_str_len) :: exp_ds_type
    logical                    :: order_as_expected
    logical                    :: distance_check_ok
    integer(i8_)               :: DS_distance
    integer                    :: filepos
    character(len=max_str_len) :: value
    integer                    :: exp_val_length, N_Meas_Max, NF_Order_Max, ios
    integer                    :: which_DSD, AllocStatus
    logical                    :: do_Compare

    !------------------------------
    ! reading the SPH KVT elements
    !------------------------------

    ! init
    error_flag = no_error

    do_Compare = .false.
    IF (present(Compare)) do_Compare = Compare

    IF (.not. associated(SPH)) THEN
       print *,"ERROR in Read_L1B_SPH_KVT():"
       print *,"The pointer to this SPH seems not associated !!"
       error_flag = error_programming
       return
    END IF

    !  #[ read the specific product header descriptor
    exp_val_length = len(SPH%SPH_DESCRIPTOR)
    exp_name       = "SPH_DESCRIPTOR"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expect_quotes=.true.)
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%SPH_Descriptor, value, &
            "L1B%SPH%Sph_Descriptor",error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%SPH_DESCRIPTOR = value
    END IF
    !  #]
    !  #[ read the lat. of the intersection of WGS84 DEM and los for meas 1 brc 1
    exp_val_length = len(SPH%Intersect_Start_Lat)
    exp_name       = "INTERSECT_START_LAT"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expected_phys_unit="10-6DegN")
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Intersect_Start_Lat, value, &
            "L1B%SPH%Intersect_Start_Lat",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Intersect_Start_Lat = value
    END IF
    !  #]
    !  #[ read the lon. of the intersection of WGS84 DEM and los for meas 1 brc 1
    exp_val_length = len(SPH%Intersect_Start_Long)
    exp_name       = "INTERSECT_START_LONG"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expected_phys_unit="10-6DegE")
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Intersect_Start_Long, value, &
            "L1B%SPH%Intersect_Start_Long",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Intersect_Start_Long = value
    END IF
    !  #]
    !  #[ read the lat. of the intersection of WGS84 DEM and los for
    !     the last meas of the last brc
    exp_val_length = len(SPH%Intersect_Stop_Lat)
    exp_name       = "INTERSECT_STOP_LAT"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expected_phys_unit="10-6DegN")
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Intersect_Stop_Lat, value, &
            "L1B%SPH%Intersect_Stop_Lat",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Intersect_Stop_Lat = value
    END IF
    !  #]
    !  #[ read the lon. of the intersection of WGS84 DEM and los for 
    !     the last meas of the last brc
    exp_val_length = len(SPH%Intersect_Stop_Long)
    exp_name       = "INTERSECT_STOP_LONG"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expected_phys_unit="10-6DegE")
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Intersect_Stop_Long, value, &
            "L1B%SPH%Intersect_Stop_Long",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Intersect_Stop_Long = value
    END IF
    !  #]
    !  #[ read the sub-sat. track heading direction at sensing start time
    !     as specified in the MPH block
    exp_val_length = len(SPH%Sat_Track)
    exp_name       = "SAT_TRACK"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expected_phys_unit="deg")
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Sat_Track, value, &
            "L1B%SPH%Sat_Track",error_flag,CompareAsReal=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Sat_Track = value
    END IF
    !  #]

    !  #[ Spare_1: this is a spare value without content
    exp_val_length = 50
    call read_spare_kvt_string(fileunit,exp_val_length,error_flag)
    IF (error_flag .ne. no_error) return
    !  #]

    !  #[ read the Base Laser Transmit frequency
    exp_val_length = len(SPH%Base_Laser_Frequency)
    exp_name       = "BASE_LASER_FREQUENCY"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expected_phys_unit="GigaHertz")
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Base_Laser_Frequency, value, &
            "L1B%SPH%Base_Laser_Frequency",error_flag,CompareAsReal=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Base_Laser_Frequency = value
    END IF
    !  #]
    !  #[ read the measurement array size (max nr of measurements present in L1A)
    exp_val_length = len(SPH%N_Max)
    exp_name       = "N_MAX"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%N_Max, value, &
            "L1B%SPH%N_Max",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%N_Max = value
    END IF
    !  #]
    !  #[ read the measurement array size 
    !     (max nr of measurements that can be used in the L1B output)
    exp_val_length = len(SPH%N_Max_Actual)
    exp_name       = "N_MAX_ACTUAL"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%N_Max_Actual, value, &
            "L1B%SPH%N_Max_Actual",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%N_Max_Actual = value
    END IF
    !  #]
    !  #[ read the nr of observations
    exp_val_length = len(SPH%Total_Num_of_Observations)
    exp_name       = "TOTAL_NUM_OF_OBSERVATIONS"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Total_Num_of_Observations, value, &
            "L1B%SPH%Total_Num_of_Observations",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Total_Num_of_Observations = value
    END IF
    !  #]
    !  #[ read the nr of measurements
    exp_val_length = len(SPH%Total_Num_of_Measurements)
    exp_name       = "TOTAL_NUM_OF_MEASUREMENTS"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Total_Num_of_Measurements, value, &
            "L1B%SPH%Total_Num_of_Measurements",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Total_Num_of_Measurements = value
    END IF
    !  #]
    !  #[ read the total nr of reference pulses
    exp_val_length = len(SPH%Total_Num_of_Reference_Pulses)
    exp_name       = "TOTAL_NUM_OF_REFERENCE_PULSES"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Total_Num_of_Reference_Pulses, value, &
            "L1B%SPH%Total_Num_of_Reference_Pulses",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Total_Num_of_Reference_Pulses = value
    END IF
    !  #]

    !  #[ Spare_2: this is a spare value without content
    exp_val_length = 50
    call read_spare_kvt_string(fileunit,exp_val_length,error_flag)
    IF (error_flag .ne. no_error) return
    !  #]

    !  #[ read the nr of obs. used for Mie wind velocity estimates 
    !     (24, 1 for each altitude bin)
    exp_val_length = len(SPH%Num_of_Mie_Obs_Used)
    exp_name       = "NUM_MIE_OBSERVATIONS_USED"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Num_of_Mie_Obs_Used, value, &
            "L1B%SPH%Num_of_Mie_Obs_Used",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Num_of_Mie_Obs_Used = value
    END IF
    !  #]
    !  #[ read the nr of obs. used for Rayleigh wind velocity estimates 
    !     (24, 1 for each altitude bin)
    exp_val_length = len(SPH%Num_of_Rayleigh_Obs_Used)
    exp_name       = "NUM_RAYLEIGH_OBSERVATIONS_USED"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Num_of_Rayleigh_Obs_Used, value, &
            "L1B%SPH%Num_of_Rayleigh_Obs_Used",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Num_of_Rayleigh_Obs_Used = value
    END IF
    !  #]
    !  #[ read the nr of meas. used for Mie wind velocity estimates 
    !     (24, 1 for each altitude bin)
    exp_val_length = len(SPH%Num_of_Mie_Meas_Used)
    exp_name       = "NUM_MIE_MEASUREMENTS_USED"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Num_of_Mie_Meas_Used, value, &
            "L1B%SPH%Num_of_Mie_Meas_Used",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Num_of_Mie_Meas_Used = value
    END IF
    !  #]
    !  #[ read the nr of meas. used for Rayleigh wind velocity estimates 
    !     (24, 1 for each altitude bin)
    exp_val_length = len(SPH%Num_of_Rayleigh_Meas_Used)
    exp_name       = "NUM_RAYLEIGH_MEASUREMENTS_USED"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Num_of_Rayleigh_Meas_Used, value, &
            "L1B%SPH%Num_of_Rayleigh_Meas_Used",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Num_of_Rayleigh_Meas_Used = value
    END IF
    !  #]
    !  #[ read the nr of Mie reference pulses used
    exp_val_length = len(SPH%Num_of_Mie_Ref_Pulses_Used)
    exp_name       = "NUM_MIE_REFERENCE_PULSES_USED"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Num_of_Mie_Ref_Pulses_Used, value, &
            "L1B%SPH%Num_of_Mie_Ref_Pulses_Used",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Num_of_Mie_Ref_Pulses_Used = value
    END IF
    !  #]
    !  #[ read the nr of Rayleigh reference pulses used
    exp_val_length = len(SPH%Num_of_Rayleigh_Ref_Pulses_Used)
    exp_name       = "NUM_RAYLEIGH_REFERENCE_PULSES_USED"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Num_of_Rayleigh_Ref_Pulses_Used, value, &
            "L1B%SPH%Num_of_Rayleigh_Ref_Pulses_Used",error_flag,&
            CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Num_of_Rayleigh_Ref_Pulses_Used = value
    END IF
    !  #]

    !  #[ Spare_3: this is a spare value without content
    exp_val_length = 100
    call read_spare_kvt_string(fileunit,exp_val_length,error_flag)
    IF (error_flag .ne. no_error) return
    !  #]

    !  #[ read the nr of obs. in which Mie zero wind is detected
    exp_val_length = len(SPH%Num_of_Mie_Zero_Wind_Detected)
    exp_name       = "NUM_MIE_ZERO_WIND_DETECTED"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Num_of_Mie_Zero_Wind_Detected, value, &
            "L1B%SPH%Num_of_Mie_Zero_Wind_Detected",error_flag,&
            CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Num_of_Mie_Zero_Wind_Detected = value
    END IF
    !  #]
    !  #[ read the nr of obs. in which Rayleigh zero wind is detected
    exp_val_length = len(SPH%Num_of_Rayleigh_Zero_Wind_Det)
    exp_name       = "NUM_RAYLEIGH_ZERO_WIND_DETECTED"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return
    
    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Num_of_Rayleigh_Zero_Wind_Det, value, &
            "L1B%SPH%Num_of_Rayleigh_Zero_Wind_Det",error_flag,&
            CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Num_of_Rayleigh_Zero_Wind_Det = value
    END IF
    !  #]
    !  #[ read the nr of Mie ground echo bins
    exp_val_length = len(SPH%Num_of_Mie_Ground_Echo_Bins)
    exp_name       = "NUM_MIE_GROUND_ECHO_BINS"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Num_of_Mie_Ground_Echo_Bins, value, &
            "L1B%SPH%Num_of_Mie_Ground_Echo_Bins",error_flag,&
            CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Num_of_Mie_Ground_Echo_Bins = value
    END IF
    !  #]
    !  #[ read the nr of Rayleigh ground echo bins
    exp_val_length = len(SPH%Num_of_Rayleigh_Grnd_Echo_Bins)
    exp_name       = "NUM_RAYLEIGH_GROUND_ECHO_BINS"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Num_of_Rayleigh_Grnd_Echo_Bins, value, &
            "L1B%SPH%Num_of_Rayleigh_Grnd_Echo_Bins",error_flag,&
            CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Num_of_Rayleigh_Grnd_Echo_Bins = value
    END IF
    !  #]

    !  #[ Spare_4: this is a spare value without content
    exp_val_length = 100
    call read_spare_kvt_string(fileunit,exp_val_length,error_flag)
    IF (error_flag .ne. no_error) return
    !  #]

    !  #[ read the total nr of meas. with laser transm. freq. unlocked
    exp_val_length = len(SPH%Tnum_of_Meas_Laser_Fr_Unlocked)
    exp_name       = "TOTAL_NUM_OF_MEASUREMENT_LASER_FREQ_UNLOCKED"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Tnum_of_Meas_Laser_Fr_Unlocked, value, &
            "L1B%SPH%Tnum_of_Meas_Laser_Fr_Unlocked",error_flag,&
            CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Tnum_of_Meas_Laser_Fr_Unlocked = value
    END IF
    !  #]
    !  #[ read the total nr of ref. pulses with laser transm. freq. unlocked
    exp_val_length = len(SPH%Tnum_of_Ref_Pls_Lsr_Fr_Unlocked)
    exp_name       = "TOTAL_NUM_OF_REFERENCE_PULSE_LASER_FREQ_UNLOCKED"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Tnum_of_Ref_Pls_Lsr_Fr_Unlocked, value, &
            "L1B%SPH%Tnum_of_Ref_Pls_Lsr_Fr_Unlocked",error_flag,&
            CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Tnum_of_Ref_Pls_Lsr_Fr_Unlocked = value
    END IF
    !  #]
    !  #[ read the total nr of meas.with sat. not on target status
    exp_val_length = len(SPH%Tnum_of_Sat_Not_On_Target_Meas)
    exp_name       = "TOTAL_NUM_OF_SAT_NOT_ON_TARGET_MEASUREMENTS"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Tnum_of_Sat_Not_On_Target_Meas, value, &
            "L1B%SPH%Tnum_of_Sat_Not_On_Target_Meas",error_flag,&
            CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Tnum_of_Sat_Not_On_Target_Meas = value
    END IF
    !  #]
    !  #[ read the total nr of corrupt Mie measurements
    exp_val_length = len(SPH%Tnum_of_Corrupt_Mie_Meas)
    exp_name       = "TOTAL_NUM_OF_CORRUPT_MIE_MEASUREMENTS"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Tnum_of_Corrupt_Mie_Meas, value, &
            "L1B%SPH%Tnum_of_Corrupt_Mie_Meas",error_flag,&
            CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Tnum_of_Corrupt_Mie_Meas = value
    END IF
    !  #]
    !  #[ read the total nr of corrupt Rayleigh measurements
    exp_val_length = len(SPH%Tnum_of_Corrupt_Rayleigh_Meas)
    exp_name       = "TOTAL_NUM_OF_CORRUPT_RAYLEIGH_MEASUREMENTS"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Tnum_of_Corrupt_Rayleigh_Meas, value, &
            "L1B%SPH%Tnum_of_Corrupt_Rayleigh_Meas",error_flag,&
            CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Tnum_of_Corrupt_Rayleigh_Meas = value
    END IF
    !  #]
    !  #[ read the total nr of corrupt Mie reference pulses
    exp_val_length = len(SPH%Tnum_of_Corrupt_Mie_Ref_Pulses)
    exp_name       = "TOTAL_NUM_OF_CORRUPT_MIE_REFERENCE_PULSES"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Tnum_of_Corrupt_Mie_Ref_Pulses, value, &
            "L1B%SPH%Tnum_of_Corrupt_Mie_Ref_Pulses",error_flag,&
            CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Tnum_of_Corrupt_Mie_Ref_Pulses = value
    END IF
    !  #]
    !  #[ read the total nr of corrupt Rayleigh reference pulses
    exp_val_length = len(SPH%Tnum_of_Corr_Rayleigh_Ref_Pulss)
    exp_name       = "TOTAL_NUM_OF_CORRUPT_RAYLEIGH_REFERENCE_PULSES"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Tnum_of_Corr_Rayleigh_Ref_Pulss, value, &
            "L1B%SPH%Tnum_of_Corr_Rayleigh_Ref_Pulss",error_flag,&
            CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Tnum_of_Corr_Rayleigh_Ref_Pulss = value
    END IF
    !  #]
    !  #[ read the number of Fourier terms used in the harmonic bias estimation
    exp_val_length = len(SPH%Nf_Order)
    exp_name       = "NF_ORDER"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(SPH%Nf_Order, value, &
            "L1B%SPH%Nf_Order",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! store the result
       SPH%Nf_Order = value
    END IF
    !  #]

    !  #[ Spare_5: this is a spare value without content
    exp_val_length = 100
    call read_spare_kvt_string(fileunit,exp_val_length,error_flag)
    IF (error_flag .ne. no_error) return
    !  #]

    !------------------------------
    ! read and check the DSD's 
    !------------------------------
    
    which_DSD = 0
    !  #[ DSD 1 Geolocation_DSD
    which_DSD = which_DSD + 1
    IF (.not. associated(SPH%Geolocation_DSD)) THEN
       allocate(SPH%Geolocation_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L1B_SPH_KVT:"
          error_flag = error_allocate
          return
       END IF
       call Init_DSD(SPH%Geolocation_DSD,error_flag)
       IF (error_flag .ne. no_error) return

       ! Read the Geolocation_DSD without comparing
       call Read_DSD_from_KVT(SPH%Geolocation_DSD, fileunit, &
            L1B_DS_Name_Geolocation_ADS, L1B_DS_Type_Geolocation_ADS, &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the Geolocation_DSD and compare if needed
       call Read_DSD_from_KVT(SPH%Geolocation_DSD,fileunit,&
            L1B_DS_Name_Geolocation_ADS, L1B_DS_Type_Geolocation_ADS, &
            error_flag,Compare=.true.)
       IF (error_flag .ne. no_error) goto 998
    END IF
    !  #]
    !  #[ DSD 2 Product_Conf_Data_DSD
    which_DSD = which_DSD + 1
    IF (.not. associated(SPH%Product_Conf_Data_DSD)) THEN
       allocate(SPH%Product_Conf_Data_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L1B_SPH_KVT:"
          error_flag = error_allocate
          return
       END IF
       call Init_DSD(SPH%Product_Conf_Data_DSD,error_flag)
       IF (error_flag .ne. no_error) return

       ! Read the Product_Conf_Data_DSD without comparing
       call Read_DSD_from_KVT(SPH%Product_Conf_Data_DSD, fileunit, &
            L1B_DS_Name_Prod_Conf_Data_ADS, L1B_DS_Type_Prod_Conf_Data_ADS, &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the Product_Conf_Data_DSD and compare if needed
       call Read_DSD_from_KVT(SPH%Product_Conf_Data_DSD,fileunit,&
            L1B_DS_Name_Prod_Conf_Data_ADS, L1B_DS_Type_Prod_Conf_Data_ADS, &
            error_flag,Compare=.true.)
       IF (error_flag .ne. no_error) goto 998
    END IF
    !  #]
    !  #[ DSD 3 Ground_Wind_Detection_DSD
    which_DSD = which_DSD + 1
    IF (.not. associated(SPH%Ground_Wind_Detection_DSD)) THEN
       allocate(SPH%Ground_Wind_Detection_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L1B_SPH_KVT:"
          error_flag = error_allocate
          return
       END IF
       call Init_DSD(SPH%Ground_Wind_Detection_DSD,error_flag)
       IF (error_flag .ne. no_error) return

       ! Read the Ground_Wind_Detection_DSD without comparing
       call Read_DSD_from_KVT(SPH%Ground_Wind_Detection_DSD, fileunit, &
            L1B_DS_Name_Ground_Wind_Det_ADS, L1B_DS_Type_Ground_Wind_Det_ADS, &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the Ground_Wind_Detection_DSD and compare if needed
       call Read_DSD_from_KVT(SPH%Ground_Wind_Detection_DSD,fileunit,&
            L1B_DS_Name_Ground_Wind_Det_ADS, L1B_DS_Type_Ground_Wind_Det_ADS, &
            error_flag,Compare=.true.)
       IF (error_flag .ne. no_error) goto 998
    END IF
    !  #]
    !  #[ DSD 4 Measurement_DSD
    which_DSD = which_DSD + 1
    IF (.not. associated(SPH%Measurement_DSD)) THEN
       allocate(SPH%Measurement_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L1B_SPH_KVT:"
          error_flag = error_allocate
          return
       END IF
       call Init_DSD(SPH%Measurement_DSD,error_flag)
       IF (error_flag .ne. no_error) return

       ! Read the Measurement_DSD without comparing
       call Read_DSD_from_KVT(SPH%Measurement_DSD, fileunit, &
            L1B_DS_Name_Measurement_ADS, L1B_DS_Type_Measurement_ADS, &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the Measurement_DSD and compare if needed
       call Read_DSD_from_KVT(SPH%Measurement_DSD,fileunit,&
            L1B_DS_Name_Measurement_ADS, L1B_DS_Type_Measurement_ADS, &
            error_flag,Compare=.true.)
       IF (error_flag .ne. no_error) goto 998
    END IF
    !  #]
    !  #[ DSD 5 Calibration_Char_GADS
    which_DSD = which_DSD + 1
    IF (.not. associated(SPH%Cal_DSD)) THEN
       allocate(SPH%Cal_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L1B_SPH_KVT:"
          error_flag = error_allocate
          return
       END IF
       call Init_DSD(SPH%Cal_DSD,error_flag)
       IF (error_flag .ne. no_error) return

       ! Read the Cal_DSD without comparing
       call Read_DSD_from_KVT(SPH%Cal_DSD, fileunit, &
            L1B_DS_Name_Calibr_Char_GADS, L1B_DS_Type_Calibr_Char_GADS, &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the Cal_DSD and compare if needed
       call Read_DSD_from_KVT(SPH%Cal_DSD,fileunit,&
            L1B_DS_Name_Calibr_Char_GADS, L1B_DS_Type_Calibr_Char_GADS, &
            error_flag,Compare=.true.)
       IF (error_flag .ne. no_error) goto 998
    END IF
    !  #]
    !  #[ DSD 6 Useful_Signal_DSD

    which_DSD = which_DSD + 1
    IF (Software_Version_Major_Number .eq. 1) THEN
       select case(Software_Version_Minor_Number)
       case (4,5) ! workaround
          ! Poli: this is inconsistent with L1B IODD v1.04 (M)
          exp_ds_type = "A" 
       case default
          exp_ds_type = L1B_DS_Type_Useful_Signal_DSD
       end select
    ELSE
       exp_ds_type = L1B_DS_Type_Useful_Signal_DSD
    END IF
    IF (.not. associated(SPH%Useful_Signal_DSD)) THEN
       allocate(SPH%Useful_Signal_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L1B_SPH_KVT:"
          error_flag = error_allocate
          return
       END IF
       call Init_DSD(SPH%Useful_Signal_DSD,error_flag)
       IF (error_flag .ne. no_error) return

       ! Read the Useful_Signal_DSD without comparing
       call Read_DSD_from_KVT(SPH%Useful_Signal_DSD, fileunit, &
            L1B_DS_Name_Useful_Signal_DSD, exp_ds_type, &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the Useful_Signal_DSD and compare if needed
       call Read_DSD_from_KVT(SPH%Useful_Signal_DSD,fileunit,&
            L1B_DS_Name_Useful_Signal_DSD, exp_ds_type, &
            error_flag,Compare=.true.)
       IF (error_flag .ne. no_error) goto 998
    END IF

    !  #]
    !  #[ DSD 7 Wind_Velocity_DSD

    which_DSD = which_DSD + 1
    IF (Software_Version_Major_Number .eq. 1) THEN
       select case(Software_Version_Minor_Number)
       case (4,5) ! workaround
          ! Poli: this is inconsistent with L1B IODD v1.04 (M)
          exp_ds_type = "A" 
       case default
          exp_ds_type = L1B_DS_Type_Wind_Velocity_MDS
       end select
    ELSE
       exp_ds_type = L1B_DS_Type_Wind_Velocity_MDS
    END IF
    IF (.not. associated(SPH%Wind_Velocity_DSD)) THEN
       allocate(SPH%Wind_Velocity_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L1B_SPH_KVT:"
          error_flag = error_allocate
          return
       END IF
       call Init_DSD(SPH%Wind_Velocity_DSD,error_flag)
       IF (error_flag .ne. no_error) return

       ! Read the Wind_Velocity_DSD without comparing
       call Read_DSD_from_KVT(SPH%Wind_Velocity_DSD, fileunit, &
            L1B_DS_Name_Wind_Velocity_MDS, exp_ds_type, &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the Wind_Velocity_DSD and compare if needed
       call Read_DSD_from_KVT(SPH%Wind_Velocity_DSD,fileunit,&
            L1B_DS_Name_Wind_Velocity_MDS, exp_ds_type, &
            error_flag,Compare=.true.)
       IF (error_flag .ne. no_error) goto 998
    END IF

    !  #]
    !  #[ DSD 8 Aeolus_L1A_Product_DSD
    which_DSD = which_DSD + 1
    IF (.not. associated(SPH%Aeolus_L1A_Product_DSD)) THEN
       allocate(SPH%Aeolus_L1A_Product_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L1B_SPH_KVT:"
          error_flag = error_allocate
          return
       END IF
       call Init_DSD(SPH%Aeolus_L1A_Product_DSD,error_flag)
       IF (error_flag .ne. no_error) return

       ! Read the Aeolus_L1A_Product_DSD without comparing
       call Read_DSD_from_KVT(SPH%Aeolus_L1A_Product_DSD, fileunit, &
            L1B_DS_Name_Aeolus_L1AP, L1B_DS_Type_Aeolus_L1AP, &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the Aeolus_L1A_Product_DSD and compare if needed
       call Read_DSD_from_KVT(SPH%Aeolus_L1A_Product_DSD,fileunit,&
            L1B_DS_Name_Aeolus_L1AP, L1B_DS_Type_Aeolus_L1AP, &
            error_flag,Compare=.true.)
       IF (error_flag .ne. no_error) goto 998
    END IF
    !  #]       
    !  #[ DSD 9 Aeolus_MRC_DSD
    which_DSD = which_DSD + 1
    IF (.not. associated(SPH%Aeolus_MRC_DSD)) THEN
       allocate(SPH%Aeolus_MRC_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L1B_SPH_KVT:"
          error_flag = error_allocate
          return
       END IF
       call Init_DSD(SPH%Aeolus_MRC_DSD,error_flag)
       IF (error_flag .ne. no_error) return

       ! Read the Aeolus_MRC_DSD without comparing
       call Read_DSD_from_KVT(SPH%Aeolus_MRC_DSD, fileunit, &
            L1B_DS_Name_Aeolus_MRC, L1B_DS_Type_Aeolus_MRC, &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the Aeolus_MRC_DSD and compare if needed
       call Read_DSD_from_KVT(SPH%Aeolus_MRC_DSD,fileunit,&
            L1B_DS_Name_Aeolus_MRC, L1B_DS_Type_Aeolus_MRC, &
            error_flag,Compare=.true.)
       IF (error_flag .ne. no_error) goto 998
    END IF
    !  #]       
    !  #[ DSD 10 Aeolus_RRC_DSD
    which_DSD = which_DSD + 1
    IF (.not. associated(SPH%Aeolus_RRC_DSD)) THEN
       allocate(SPH%Aeolus_RRC_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L1B_SPH_KVT:"
          error_flag = error_allocate
          return
       END IF
       call Init_DSD(SPH%Aeolus_RRC_DSD,error_flag)
       IF (error_flag .ne. no_error) return

       ! Read the Aeolus_RRC_DSD without comparing
       call Read_DSD_from_KVT(SPH%Aeolus_RRC_DSD, fileunit, &
            L1B_DS_Name_Aeolus_RRC, L1B_DS_Type_Aeolus_RRC, &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the Aeolus_RRC_DSD and compare if needed
       call Read_DSD_from_KVT(SPH%Aeolus_RRC_DSD,fileunit,&
            L1B_DS_Name_Aeolus_RRC, L1B_DS_Type_Aeolus_RRC, &
            error_flag,Compare=.true.)
       IF (error_flag .ne. no_error) goto 998
    END IF
    !  #]       
    !  #[ DSD 11 Aeolus_ZWC_DSD
    which_DSD = which_DSD + 1
    IF (.not. associated(SPH%Aeolus_ZWC_DSD)) THEN
       allocate(SPH%Aeolus_ZWC_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L1B_SPH_KVT:"
          error_flag = error_allocate
          return
       END IF
       call Init_DSD(SPH%Aeolus_ZWC_DSD,error_flag)
       IF (error_flag .ne. no_error) return

       ! Read the Aeolus_ZWC_DSD without comparing
       call Read_DSD_from_KVT(SPH%Aeolus_ZWC_DSD, fileunit, &
            L1B_DS_Name_Aeolus_ZWC, L1B_DS_Type_Aeolus_ZWC, &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the Aeolus_ZWC_DSD and compare if needed
       call Read_DSD_from_KVT(SPH%Aeolus_ZWC_DSD,fileunit,&
            L1B_DS_Name_Aeolus_ZWC, L1B_DS_Type_Aeolus_ZWC, &
            error_flag,Compare=.true.)
       IF (error_flag .ne. no_error) goto 998
    END IF
    !  #]       
    !  #[ DSD 12 Level_1B_Proc_Params_DSD
    which_DSD = which_DSD + 1
    IF (.not. associated(SPH%L1B_Proc_Params_DSD)) THEN
       allocate(SPH%L1B_Proc_Params_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L1B_SPH_KVT:"
          error_flag = error_allocate
          return
       END IF
       call Init_DSD(SPH%L1B_Proc_Params_DSD,error_flag)
       IF (error_flag .ne. no_error) return

       ! Read the Aeolus_L1B_Proc_Params_DSD without comparing
       call Read_DSD_from_KVT(SPH%L1B_Proc_Params_DSD, fileunit, &
            L1B_DS_Name_L1B_Proc_Pars, L1B_DS_Type_L1B_Proc_Pars, &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the Aeolus_L1B_Proc_Params_DSD and compare if needed
       call Read_DSD_from_KVT(SPH%L1B_Proc_Params_DSD,fileunit,&
            L1B_DS_Name_L1B_Proc_Pars, L1B_DS_Type_L1B_Proc_Pars, &
            error_flag,Compare=.true.)
       IF (error_flag .ne. no_error) goto 998
    END IF
    !  #]       
    !  #[ DSD 13 Sat_Characterisation_DSD
    which_DSD = which_DSD + 1
    IF (.not. associated(SPH%Sat_Characterisation_DSD)) THEN
       allocate(SPH%Sat_Characterisation_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L1B_SPH_KVT:"
          error_flag = error_allocate
          return
       END IF
       call Init_DSD(SPH%Sat_Characterisation_DSD,error_flag)
       IF (error_flag .ne. no_error) return

       ! Read the Sat_Characterisation_DSD without comparing
       call Read_DSD_from_KVT(SPH%Sat_Characterisation_DSD, fileunit, &
            L1B_DS_Name_Sat_Char, L1B_DS_Type_Sat_Char, &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the Sat_Characterisation_DSD and compare if needed
       call Read_DSD_from_KVT(SPH%Sat_Characterisation_DSD,fileunit,&
            L1B_DS_Name_Sat_Char, L1B_DS_Type_Sat_Char, &
            error_flag,Compare=.true.)
       IF (error_flag .ne. no_error) goto 998
    END IF
    !  #]       
    !  #[ DSD 14 Orbit_Scenario_File_DSD
    which_DSD = which_DSD + 1
    IF (.not. associated(SPH%Orbit_Scenario_File_DSD)) THEN
       allocate(SPH%Orbit_Scenario_File_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L1B_SPH_KVT:"
          error_flag = error_allocate
          return
       END IF
       call Init_DSD(SPH%Orbit_Scenario_File_DSD,error_flag)
       IF (error_flag .ne. no_error) return

       ! Read the Orbit_Scenario_File_DSD without comparing
       call Read_DSD_from_KVT(SPH%Orbit_Scenario_File_DSD, fileunit, &
            L1B_DS_Name_Orbit_Scenario_File, L1B_DS_Type_Orbit_Scenario_File, &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the Orbit_Scenario_File_DSD and compare if needed
       call Read_DSD_from_KVT(SPH%Orbit_Scenario_File_DSD,fileunit,&
            L1B_DS_Name_Orbit_Scenario_File, L1B_DS_Type_Orbit_Scenario_File, &
            error_flag,Compare=.true.)
       IF (error_flag .ne. no_error) goto 998
    END IF
    !  #]       

    ! skip this DSD for software versions: "ADM_L1bP/1.04","ADM_L1bP/1.05"
    IF ( ( (Software_Version_Major_Number .eq. 1) .and. &
           (Software_Version_Minor_Number .ge. 6)       ) .or. &
         (Software_Version_Major_Number .gt. 1)                ) THEN
       ! this DSD has been added in version 1.06
       !  #[ DSD 15 Digital_Elevation_Model
       which_DSD = which_DSD + 1
       IF (.not. associated(SPH%Digital_Elevation_Model_DSD)) THEN
          allocate(SPH%Digital_Elevation_Model_DSD,STAT=AllocStatus)
          IF (AllocStatus .ne. 0) THEN
             print *,"Allocation failure in subroutine Read_L1B_SPH_KVT:"
             error_flag = error_allocate
             return
          END IF
          call Init_DSD(SPH%Digital_Elevation_Model_DSD,error_flag)
          IF (error_flag .ne. no_error) return
          
          ! Read the Digital_Elevation_Model_DSD without comparing
          call Read_DSD_from_KVT(SPH%Digital_Elevation_Model_DSD, fileunit, &
               L1B_DS_Name_DEM, L1B_DS_Type_DEM, &
               error_flag)
          IF (error_flag .ne. no_error) goto 998
       ELSE
          ! Read the Digital_Elevation_Model_DSD and compare if needed
          call Read_DSD_from_KVT(SPH%Digital_Elevation_Model_DSD,fileunit,&
               L1B_DS_Name_DEM, L1B_DS_Type_DEM, &
               error_flag,Compare=.true.)
          IF (error_flag .ne. no_error) goto 998
       END IF
       !  #]       
    END IF

    !  #[ DSD 15 (or 16) Geoid_Model
    which_DSD = which_DSD + 1
    IF (.not. associated(SPH%Geoid_Model_DSD)) THEN
       allocate(SPH%Geoid_Model_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L1B_SPH_KVT:"
          error_flag = error_allocate
          return
       END IF
       call Init_DSD(SPH%Geoid_Model_DSD,error_flag)
       IF (error_flag .ne. no_error) return
       
       ! Read the Geoid_Model_DSD without comparing
       call Read_DSD_from_KVT(SPH%Geoid_Model_DSD, fileunit, &
            L1B_DS_Name_Geoid_Model, L1B_DS_Type_Geoid_Model, &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the Sat_Characterisation_DSD and compare if needed
       call Read_DSD_from_KVT(SPH%Geoid_Model_DSD,fileunit,&
            L1B_DS_Name_Geoid_Model, L1B_DS_Type_Geoid_Model, &
            error_flag,Compare=.true.)
       IF (error_flag .ne. no_error) goto 998
    END IF
    !  #]       
    
    !  #[ check if the DS's are in the expected order
    order_as_expected = .true.
    IF ( SPH%Geolocation_DSD%start_of_ds .gt. &
         SPH%Product_Conf_Data_DSD%start_of_ds) order_as_expected = .false.
    IF ( SPH%Product_Conf_Data_DSD%start_of_ds .gt. &
         SPH%Ground_Wind_Detection_DSD%start_of_ds) order_as_expected = .false.
    IF ( SPH%Ground_Wind_Detection_DSD%start_of_ds .gt. &
         SPH%Measurement_DSD%start_of_ds) order_as_expected = .false.
    IF ( SPH%Measurement_DSD%start_of_ds .gt. &
         SPH%Useful_Signal_DSD%start_of_ds) order_as_expected = .false.
    IF ( SPH%Useful_Signal_DSD%start_of_ds .gt. &
         SPH%Wind_Velocity_DSD%start_of_ds) order_as_expected = .false.    
       
    !     Geolocation_DSD
    !     Product_Conf_Data_DSD
    !     Ground_Wind_Detection_DSD
    !     Measurement_DSD
    !     Cal_DSD
    !     Useful_Signal_DSD
    !     Wind_Velocity_DSD
    
    ! the next 9 datasets are in external files, so their filepointer start
    ! values cannot be checked
    
    print *,"-------------------------------------------------"
    IF (order_as_expected) THEN
       print *,"DSD order is OK"
    ELSE
       print *,"DSD order is not as expected"
    END IF
    !  #]

    !  #[ check if the DS's are following each other without gaps
    distance_check_ok = .true.
       
    ! request a filepointer pointing to the first byte AFTER
    ! the last byte of the KVT header
    inquire(unit=fileunit,nextrec=filepos)
    DS_distance = filepos - SPH%Geolocation_DSD%start_of_ds
    IF ( DS_distance .ne. 0) THEN
       print *,"WARNING: there seem to be unexpected extra bytes"
       print *,"between the KVT header and the Geolocation_ADS"
       print *,"DS_distance = ",DS_distance
       distance_check_ok = .false.
    END IF
    
    DS_distance = SPH%Product_Conf_Data_DSD%start_of_ds - &
                  SPH%Geolocation_DSD%end_of_ds - 1
    IF ( DS_distance .ne. 0) THEN
       print *,"WARNING: there seem to be unexpected extra bytes"
       print *,"between the Geolocation_ADS and the Product_Conf_ADS"
       print *,"DS_distance = ",DS_distance
       distance_check_ok = .false.
    END IF
    
    DS_distance = SPH%Ground_Wind_Detection_DSD%start_of_ds - &
                  SPH%Product_Conf_Data_DSD%end_of_ds - 1
    IF ( DS_distance .ne. 0) THEN
       print *,"WARNING: there seem to be unexpected extra bytes"
       print *,"between the Product_Conf_ADS and the Ground_Wind_Detection_ADS"
       print *,"DS_distance = ",DS_distance
       distance_check_ok = .false.
    END IF
       
    DS_distance = SPH%Measurement_DSD%start_of_ds - &
                  SPH%Ground_Wind_Detection_DSD%end_of_ds - 1

    IF ( DS_distance .ne. 0) THEN
       print *,"WARNING: there seem to be unexpected extra bytes"
       print *,"between the Ground_Wind_Detection_ADS and the Measurement_ADS"
       print *,"DS_distance = ",DS_distance
       distance_check_ok = .false.
    END IF
       
    DS_distance = SPH%Cal_DSD%start_of_ds - &
                  SPH%Measurement_DSD%end_of_ds - 1
    IF ( DS_distance .ne. 0) THEN
       print *,"WARNING: there seem to be unexpected extra bytes"
       print *,"between the Measurement_ADS and the Cal_ADS"
       print *,"DS_distance = ",DS_distance
       distance_check_ok = .false.
    END IF

    DS_distance = SPH%Useful_Signal_DSD%start_of_ds - &
                  SPH%Cal_DSD%end_of_ds - 1
    IF ( DS_distance .ne. 0) THEN
       print *,"WARNING: there seem to be unexpected extra bytes"
       print *,"between the Cal_ADS and the Useful_Signal_MDS"
       print *,"DS_distance = ",DS_distance
       distance_check_ok = .false.
    END IF
       
    DS_distance = SPH%Wind_Velocity_DSD%start_of_ds - &
                  SPH%Useful_Signal_DSD%end_of_ds - 1
    IF ( DS_distance .ne. 0) THEN
       print *,"WARNING: there seem to be unexpected extra bytes"
       print *,"between the Useful_Signal_MDS and the Wind_Velocity_MDS"
       print *,"DS_distance = ",DS_distance
       distance_check_ok = .false.
    END IF
       
    !     Geolocation_DSD
    !     Product_Conf_Data_DSD
    !     Ground_Wind_Detection_DSD
    !     Measurement_DSD
    !     Cal_DSD
    !     Useful_Signal_DSD
    !     Wind_Velocity_DSD
    
    ! the next 9 datasets are in external files, so their distance
    ! values have no meaning
       
    IF (distance_check_ok) THEN
       print *,"DS distance check OK"
    ELSE
       print *,"DS distance check failed"
    END IF
    !  #]

    !     if something is wrong (for debugging)
    IF ( .not. order_as_expected .or. &
         .not. distance_check_ok      ) THEN
       !  #[ print binary dataset start/end filepointers

       print *,"-------------------------------------------------"
       ! check filepos
       inquire(unit=fileunit,nextrec=filepos)
       print *,"filepos = ",filepos
       
       print *,"SPH%Geolocation_DSD%start_of_ds           = ",&
            SPH%Geolocation_DSD%start_of_ds
       print *,"SPH%Geolocation_DSD%end_of_ds             = ",&
            SPH%Geolocation_DSD%end_of_ds
       
       print *,"SPH%Product_Conf_Data_DSD%start_of_ds     = ",&
            SPH%Product_Conf_Data_DSD%start_of_ds
       print *,"SPH%Product_Conf_Data_DSD%end_of_ds       = ",&
            SPH%Product_Conf_Data_DSD%end_of_ds
       
       print *,"SPH%Ground_Wind_Detection_DSD%start_of_ds = ",&
            SPH%Ground_Wind_Detection_DSD%start_of_ds
       print *,"SPH%Ground_Wind_Detection_DSD%end_of_ds   = ",&
            SPH%Ground_Wind_Detection_DSD%end_of_ds
       
       print *,"SPH%Measurement_DSD%start_of_ds           = ",&
            SPH%Measurement_DSD%start_of_ds
       print *,"SPH%Measurement_DSD%end_of_ds             = ",&
            SPH%Measurement_DSD%end_of_ds
       
       print *,"SPH%Cal_Char_DSD%start_of_ds              = ",&
            SPH%Cal_DSD%start_of_ds
       print *,"SPH%Cal_Char_DSD%end_of_ds                = ",&
            SPH%Cal_DSD%end_of_ds
       
       print *,"SPH%Useful_Signal_DSD%start_of_ds         = ",&
            SPH%Useful_Signal_DSD%start_of_ds
       print *,"SPH%Useful_Signal_DSD%end_of_ds           = ",&
            SPH%Useful_Signal_DSD%end_of_ds
       
       print *,"SPH%Wind_Velocity_DSD%start_of_ds         = ",&
            SPH%Wind_Velocity_DSD%start_of_ds
       print *,"SPH%Wind_Velocity_DSD%end_of_ds           = ",&
            SPH%Wind_Velocity_DSD%end_of_ds
       
       print *,"-------------------------------------------------"
       print *,"Please check whether the datafile definition"
       print *,"has been updated ......"
       print *,"-------------------------------------------------"
       !       error_flag = error_reading_file
       !       return
    END IF
    
    !  #]
    
    SPH%DSDs_are_available = .true.

    !------------------------------
    ! convert some selected strings to their numerical value
    !------------------------------

    read(SPH%N_MAX, "(i11)",iostat=ios) N_Meas_Max
    IF (ios .ne. 0) goto 999
    SPH%N_Meas_Max = N_Meas_Max

    read(SPH%NF_ORDER, "(i11)",iostat=ios) NF_Order_Max
    IF (ios .ne. 0) goto 999
    SPH%NF_Order_Max = NF_Order_Max

    !------------------------------
    ! done
    !------------------------------

    return

998 print *,"ABORTING subroutine Read_L1B_SPH_KVT"
    return

999 print *,"ERROR in Read_L1B_SPH_KVT:"
    print *,"could not read integer from string: ",trim(SPH%N_MAX)
    error_flag = error_reading_file
    return

  end subroutine Read_L1B_SPH_KVT
    !  #]
  !---------------------------------------------
! to add to this next routine:
! -a check to see if the count attribute in <List_of_Dsds count="14">
!  has the correct value
  subroutine Get_L1B_SPH_from_xml_tree(xml_tree, path_in, L1B_SPH, &
       Software_Version_Major_Number, Software_Version_Minor_Number, &
       error_flag)
    !  #[ read the specific product header (XML values)

    TYPE(xml_node_type), pointer     :: xml_tree         ! input
    character(len=*),    intent(in)  :: path_in          ! input
    TYPE(L1B_SPH_type),  pointer     :: L1B_SPH          ! result
    integer,             intent(in)  :: Software_Version_Major_Number ! input
    integer,             intent(in)  :: Software_Version_Minor_Number ! input
    integer,             intent(out) :: error_flag       ! output

    ! local variables
    character(len=max_str_len) :: path
    character(len=max_str_len) :: text
    character(len=1)           :: exp_ds_type
    integer                    :: AllocStatus
    integer                    :: which_DSD
    logical                    :: use_dsd_workaround

    ! init
    error_flag = no_error

    IF (.not. associated(L1B_SPH)) THEN
       print *,"ERROR in Get_L1B_SPH_from_xml_tree():"
       print *,"the pointer L1B_SPH seems not associated !!!"
       error_flag = error_programming
       return
    END IF

    IF (.not. associated(xml_tree)) THEN
       print *,"ERROR in Get_L1B_SPH_from_xml_tree():"
       print *,"the pointer xml_tree seems not associated !!!"
       error_flag = error_programming
       return
    END IF

    !----------------------------------!
    ! copying the L1B_SPH XML elements !
    !----------------------------------!

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Sph_Descriptor"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%SPH_Descriptor = &
         check_length_xml_item(text,path,maxlen_SPH_Descriptor,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Intersect_Start_Lat"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Intersect_Start_Lat = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Intersect_Start_Long"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Intersect_Start_Long = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Intersect_Stop_Lat"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Intersect_Stop_Lat = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Intersect_Stop_Long"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Intersect_Stop_Long = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Sat_Track"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Sat_Track = &
         check_length_xml_item(text,path,maxlen_FAdoxy,&
                               "Get_L1B_SPH_from_xml_tree")

    ! Spare_1: no need to read this one

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Base_Laser_Frequency"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Base_Laser_Frequency = &
         check_length_xml_item(text,path,maxlen_FAdoxy,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/N_Max"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%N_Max = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/N_Max_Actual"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%N_Max_Actual = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Total_Num_of_Observations"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Total_Num_of_Observations = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Total_Num_of_Measurements"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Total_Num_of_Measurements = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Total_Num_of_Reference_Pulses"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Total_Num_of_Reference_Pulses = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    ! Spare_2: no need to read this one

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Mie_Observations_Used"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Num_of_Mie_Obs_Used = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Rayleigh_Observations_Used"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Num_of_Rayleigh_Obs_Used = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Mie_Measurements_Used"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Num_of_Mie_Meas_Used = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Rayleigh_Measurements_Used"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Num_of_Rayleigh_Meas_Used = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Mie_Reference_Pulses_Used"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Num_of_Mie_Ref_Pulses_Used = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Rayleigh_Reference_Pulses_Used"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Num_of_Rayleigh_Ref_Pulses_Used = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    ! Spare_3: no need to read this one

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Mie_Zero_Wind_Detected"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Num_of_Mie_Zero_Wind_Detected = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Rayleigh_Zero_Wind_Detected"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Num_of_Rayleigh_Zero_Wind_Det = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Mie_Ground_Echo_Bins"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Num_of_Mie_Ground_Echo_Bins = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Rayleigh_Ground_Echo_Bins"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Num_of_Rayleigh_Grnd_Echo_Bins = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    ! Spare_4: no need to read this one

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Total_Num_of_Measurement_Laser_Freq_Unlocked"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Tnum_of_Meas_Laser_Fr_Unlocked = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Total_Num_of_Reference_Pulse_Laser_Freq_Unlocked"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Tnum_of_Ref_Pls_Lsr_Fr_Unlocked = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Total_Num_of_Sat_Not_on_Target_Measurements"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Tnum_of_Sat_Not_on_Target_Meas = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Total_Num_of_Corrupt_Mie_Measurements"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Tnum_of_Corrupt_Mie_Meas = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Total_Num_of_Corrupt_Rayleigh_Measurements"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Tnum_of_Corrupt_Rayleigh_Meas = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Total_Num_of_Corrupt_Mie_Reference_Pulses"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Tnum_of_Corrupt_Mie_Ref_Pulses = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Total_Num_of_Corrupt_Rayleigh_Reference_Pulses"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%Tnum_of_Corr_Rayleigh_Ref_Pulss = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/NF_Order"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L1B_SPH%NF_Order = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L1B_SPH_from_xml_tree")

    ! Spare_5: no need to read this one

    use_dsd_workaround = .false.
    IF ( (Software_Version_Major_Number .eq. 1) .and. &
         (Software_Version_Minor_Number .eq. 8)       ) THEN

       ! this is a workaround for the xml format change in the DSD's 
       ! that has been added in version 1.08. This was a mistake
       ! and has been reverted in L1B v1.09
       use_dsd_workaround = .true.
    END IF

    which_DSD = 0
    !  #[ DSD 1 Geolocation_DSD
    which_DSD = which_DSD + 1
    IF (.not.associated(L1B_SPH%Geolocation_DSD)) THEN
       allocate(L1B_SPH%Geolocation_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L1B_SPH%Geolocation_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L1B_SPH%Geolocation_DSD,which_DSD,&
                               L1B_DS_Name_Geolocation_ADS, &
                               L1B_DS_Type_Geolocation_ADS, &
                               error_flag, use_dsd_workaround)
    IF (error_flag .ne. no_error) return
    !  #]
    !  #[ DSD 2 Product_Conf_Data_DSD
    which_DSD = which_DSD + 1
    IF (.not.associated(L1B_SPH%Product_Conf_Data_DSD)) THEN
       allocate(L1B_SPH%Product_Conf_Data_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L1B_SPH%Product_Conf_Data_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L1B_SPH%Product_Conf_Data_DSD,which_DSD,&
                               L1B_DS_Name_Prod_Conf_Data_ADS, &
                               L1B_DS_Type_Prod_Conf_Data_ADS, &
                               error_flag, use_dsd_workaround)
    IF (error_flag .ne. no_error) return
    !  #]
    !  #[ DSD 3 Ground_Wind_Detection_DSD
    which_DSD = which_DSD + 1
    IF (.not.associated(L1B_SPH%Ground_Wind_Detection_DSD)) THEN
       allocate(L1B_SPH%Ground_Wind_Detection_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L1B_SPH%Ground_Wind_Detection_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L1B_SPH%Ground_Wind_Detection_DSD,which_DSD,&
                               L1B_DS_Name_Ground_Wind_Det_ADS, &
                               L1B_DS_Type_Ground_Wind_Det_ADS, &
                               error_flag, use_dsd_workaround)
    IF (error_flag .ne. no_error) return
    !  #]
    !  #[ DSD 4 Measurement_DSD
    which_DSD = which_DSD + 1
    IF (.not.associated(L1B_SPH%Measurement_DSD)) THEN
       allocate(L1B_SPH%Measurement_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L1B_SPH%Measurement_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L1B_SPH%Measurement_DSD,which_DSD,&
                               L1B_DS_Name_Measurement_ADS, &
                               L1B_DS_Type_Measurement_ADS, &
                               error_flag, use_dsd_workaround)
    IF (error_flag .ne. no_error) return
    !  #]
    !  #[ DSD 5 Cal_DSD
    which_DSD = which_DSD + 1
    IF (.not.associated(L1B_SPH%Cal_DSD)) THEN
       allocate(L1B_SPH%Cal_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L1B_SPH%Cal_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L1B_SPH%Cal_DSD,which_DSD,&
                               L1B_DS_Name_Calibr_Char_GADS, &
                               L1B_DS_Type_Calibr_Char_GADS, &
                               error_flag, use_dsd_workaround)
    IF (error_flag .ne. no_error) return
    !  #]
    !  #[ DSD 6 Useful_Signal_DSD
    which_DSD = which_DSD + 1
    IF (.not.associated(L1B_SPH%Useful_Signal_DSD)) THEN
       allocate(L1B_SPH%Useful_Signal_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L1B_SPH%Useful_Signal_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    IF (Software_Version_Major_Number .eq. 1) THEN
       select case(Software_Version_Minor_Number)
       case (4,5) ! workaround
          ! Poli: this is inconsistent with L1B IODD v1.04 (M)
          exp_ds_type = "A" 
       case default
          exp_ds_type = L1B_DS_Type_Useful_Signal_DSD
       end select
    ELSE
       exp_ds_type = L1B_DS_Type_Useful_Signal_DSD
    END IF
    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L1B_SPH%Useful_Signal_DSD,which_DSD,&
                               L1B_DS_Name_Useful_Signal_DSD, &
                               exp_ds_type, &
                               error_flag, use_dsd_workaround)
    IF (error_flag .ne. no_error) return
    !  #]
    !  #[ DSD 7 Wind_Velocity_DSD
    which_DSD = which_DSD + 1
    IF (.not.associated(L1B_SPH%Wind_Velocity_DSD)) THEN
       allocate(L1B_SPH%Wind_Velocity_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L1B_SPH%Wind_Velocity_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    IF (Software_Version_Major_Number .eq. 1) THEN
       select case(Software_Version_Minor_Number)
       case (4,5) ! workaround
          ! Poli: this is inconsistent with L1B IODD v1.04 (M)
          exp_ds_type = "A" 
       case default
          exp_ds_type = L1B_DS_Type_Wind_Velocity_MDS
       end select
    ELSE
       exp_ds_type = L1B_DS_Type_Wind_Velocity_MDS
    END IF
    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L1B_SPH%Wind_Velocity_DSD,which_DSD,&
                               L1B_DS_Name_Wind_Velocity_MDS, &
                               exp_ds_type, &
                               error_flag, use_dsd_workaround)
    IF (error_flag .ne. no_error) return
    !  #]
    !  #[ DSD 8 Aeolus_L1A_Product_DSD
    which_DSD = which_DSD + 1
    IF (.not.associated(L1B_SPH%Aeolus_L1A_Product_DSD)) THEN
       allocate(L1B_SPH%Aeolus_L1A_Product_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L1B_SPH%Aeolus_L1A_Product_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L1B_SPH%Aeolus_L1A_Product_DSD,which_DSD,&
                               L1B_DS_Name_Aeolus_L1AP, &
                               L1B_DS_Type_Aeolus_L1AP, &
                               error_flag, use_dsd_workaround)
    IF (error_flag .ne. no_error) return
    !  #]
    !  #[ DSD 9 Aeolus_MRC_DSD
    which_DSD = which_DSD + 1
    IF (.not.associated(L1B_SPH%Aeolus_MRC_DSD)) THEN
       allocate(L1B_SPH%Aeolus_MRC_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L1B_SPH%Aeolus_MRC_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L1B_SPH%Aeolus_MRC_DSD,which_DSD,&
                               L1B_DS_Name_Aeolus_MRC, &
                               L1B_DS_Type_Aeolus_MRC, &
                               error_flag, use_dsd_workaround)
    IF (error_flag .ne. no_error) return
    !  #]
    !  #[ DSD 10 Aeolus_RRC_DSD
    which_DSD = which_DSD + 1
    IF (.not.associated(L1B_SPH%Aeolus_RRC_DSD)) THEN
       allocate(L1B_SPH%Aeolus_RRC_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L1B_SPH%Aeolus_RRC_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L1B_SPH%Aeolus_RRC_DSD,which_DSD,&
                               L1B_DS_Name_Aeolus_RRC, &
                               L1B_DS_Type_Aeolus_RRC, &
                               error_flag, use_dsd_workaround)
    IF (error_flag .ne. no_error) return
    !  #]
    !  #[ DSD 11 Aeolus_ZWC_DSD
    which_DSD = which_DSD + 1
    IF (.not.associated(L1B_SPH%Aeolus_ZWC_DSD)) THEN
       allocate(L1B_SPH%Aeolus_ZWC_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L1B_SPH%Aeolus_ZWC_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L1B_SPH%Aeolus_ZWC_DSD,which_DSD,&
                               L1B_DS_Name_Aeolus_ZWC, &
                               L1B_DS_Type_Aeolus_ZWC, &
                               error_flag, use_dsd_workaround)
    IF (error_flag .ne. no_error) return
    !  #]
    !  #[ DSD 12 L1B_Proc_Params_DSD
    which_DSD = which_DSD + 1
    IF (.not.associated(L1B_SPH%L1B_Proc_Params_DSD)) THEN
       allocate(L1B_SPH%L1B_Proc_Params_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L1B_SPH%L1B_Proc_Params_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L1B_SPH%L1B_Proc_Params_DSD,which_DSD,&
                               L1B_DS_Name_L1B_Proc_Pars, &
                               L1B_DS_Type_L1B_Proc_Pars, &
                               error_flag, use_dsd_workaround)
    IF (error_flag .ne. no_error) return
    !  #]
    !  #[ DSD 13 Sat_Characterisation_DSD
    which_DSD = which_DSD + 1
    IF (.not.associated(L1B_SPH%Sat_Characterisation_DSD)) THEN
       allocate(L1B_SPH%Sat_Characterisation_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L1B_SPH%Sat_Characterisation_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L1B_SPH%Sat_Characterisation_DSD,which_DSD,&
                               L1B_DS_Name_Sat_Char, &
                               L1B_DS_Type_Sat_Char, &
                               error_flag, use_dsd_workaround)
    IF (error_flag .ne. no_error) return
    !  #]
    !  #[ DSD 14 Orbit_Scenario_File_DSD
    which_DSD = which_DSD + 1
    IF (.not.associated(L1B_SPH%Orbit_Scenario_File_DSD)) THEN
       allocate(L1B_SPH%Orbit_Scenario_File_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L1B_SPH%Orbit_Scenario_File_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L1B_SPH%Orbit_Scenario_File_DSD,which_DSD,&
                               L1B_DS_Name_Orbit_Scenario_File, &
                               L1B_DS_Type_Orbit_Scenario_File, &
                               error_flag, use_dsd_workaround)
    IF (error_flag .ne. no_error) return
    !  #]

    ! skip this DSD for software versions: "ADM_L1bP/1.04","ADM_L1bP/1.05"
    IF ( ( (Software_Version_Major_Number .eq. 1) .and. &
           (Software_Version_Minor_Number .ge. 6)       ) .or. &
         (Software_Version_Major_Number .gt. 1)                ) THEN
       ! this DSD has been added in version 1.06
       !  #[ DSD 15 Digital_Elevation_Model_DSD
       which_DSD = which_DSD + 1
       IF (.not.associated(L1B_SPH%Digital_Elevation_Model_DSD)) THEN
       allocate(L1B_SPH%Digital_Elevation_Model_DSD,STAT=AllocStatus)
          IF (AllocStatus .ne. 0) THEN
             error_flag = error_allocate
             return
          END IF
       
          call Init_DSD(L1B_SPH%Digital_Elevation_Model_DSD,error_flag)
          IF (error_flag .ne. no_error) return
       END IF
       
       ! get the DSD from the xml_tree structure
       call Get_DSD_from_xml_tree(xml_tree,path_in,&
                                  L1B_SPH%Digital_Elevation_Model_DSD,which_DSD,&
                                  L1B_DS_Name_DEM, &
                                  L1B_DS_Type_DEM, &
                                  error_flag, use_dsd_workaround)
       IF (error_flag .ne. no_error) return
       !  #]
    END IF

    !  #[ DSD 16 Geoid_Model_DSD
    which_DSD = which_DSD + 1
    IF (.not.associated(L1B_SPH%Geoid_Model_DSD)) THEN
       allocate(L1B_SPH%Geoid_Model_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L1B_SPH%Geoid_Model_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L1B_SPH%Geoid_Model_DSD,which_DSD,&
                               L1B_DS_Name_Geoid_Model, &
                               L1B_DS_Type_Geoid_Model, &
                               error_flag, use_dsd_workaround)
    IF (error_flag .ne. no_error) return
    !  #]

    return

  end subroutine Get_L1B_SPH_from_xml_tree

    !  #]
  !---------------------------------
end module L1B_SPH
