module L1B_US_MDS
  !  #[ Documentation
  !
  !     a module to handle the Useful Signal (US) Measurement dataset
  !
  !     See also the L1BP files: 
  !          L1bP_Draft_1/aeolus/common/PFR/src/
  !             PFR_LevelOneBProductUsefulSignalDSR.cc
  !             PFR_LevelOneBProductUsefulSignalDSR.hh
  !     for the code that defines this datablock in the L1BP
  !
  !     Written by J. de Kloe, KNMI 
  !     created:     25-07-2005
  !
  !     Modifications:
  !      05-Sep-2005 D. Tan     Upgrade to DLR TDS v1.2
  !                             Ref AE.TN.DLR.IODD-L1B.050705 (draft).
  !      10-Jan-2006 J. de Kloe new DateTime type
  !      22-Feb-2006 P. Poli    Upgrade to DLR IODD L1B-IODD.231105
  !      28-Mar-2006 P. Poli    Upgrade to DLR IODD AE.TN.DLR.APM-L1B.150206
  !      12-May-2006 D. Tan     Null use of ParKind1 module
  !      03-Nov-2006 P. Poli    moved and cleaned-up reading part
  !      03-Jan-2008 J. de Kloe added nullify statements to make the code
  !                             more robust
  !      16-Jan-2008 J. de Kloe phase out integer kind i_ 
  !
  !  #]
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE Numerics, only: i8_, r8_, real2txt
  USE DataSetDescriptor, only: DSD_type
  USE EE_CFI_Datatypes, only: DateTimeType,Construct_KVT_DateTime, InitDateTime, &
                              read_DateTime_from_bin_file, maxlen_DateTime_KVT, &
                              read_IntAuc_from_bin_file, missing_indicator_IntAuc, IntAuc, &
                              read_FAdoxy_from_bin_file, missing_indicator_FAdoxy, FAdoxy
  USE DirectBinaryIO, only: SetFilePointer, AdvanceFilePointer, read_logical1_from_bin_file
  USE ErrorHandler, only: no_error, error_programming, error_allocate, error_reading_file
  !  #]
  !  #[ type definitions

  implicit none

  TYPE Rangebin_US_Mie_type
     integer(IntAuc) :: Data_Quality_Flag ! 8 bits flag 
     real(FAdoxy)    :: Useful_Signal
  END TYPE Rangebin_US_Mie_type

  TYPE Rangebin_US_Rayleigh_type
     integer(IntAuc) :: Data_Quality_Flag ! 8 bits flag 
     real(FAdoxy)    :: Useful_Signal_Channel_A
     real(FAdoxy)    :: Useful_Signal_Channel_B
  END TYPE Rangebin_US_Rayleigh_type

  TYPE US_type
     TYPE(Rangebin_US_Mie_type),      dimension(:), pointer :: Mie
     TYPE(Rangebin_US_Rayleigh_type), dimension(:), pointer :: Rayleigh
  END TYPE US_type

  TYPE L1B_US_MDS_type ! see sec. 6.3.4, table 6-18, of L1B_iodd 
     type(DateTimeType)   :: start_of_obs_datetime ! UTC
     TYPE(US_type)                        :: Obs_US
     TYPE(US_type), dimension(:), pointer :: Meas_US
     !              1234567890123456789012345678901
  END TYPE L1B_US_MDS_type

  !  #]
contains
  !---------------------------------------------------------------------
  subroutine Init_Rangebin_US_Mie(Rangebin_US_Mie)
    !  #[

    TYPE(Rangebin_US_Mie_type), intent(out) :: Rangebin_US_Mie

    Rangebin_US_Mie%Data_Quality_Flag = 0
    Rangebin_US_Mie%Useful_Signal = missing_indicator_FAdoxy

  end subroutine Init_Rangebin_US_Mie

    !  #]
  subroutine Print_Rangebin_US_Mie(Rangebin_US_Mie)
    !  #[
    TYPE(Rangebin_US_Mie_type), intent(in) :: Rangebin_US_Mie
    character(len=13) :: txt
    print *,"...%Data_Quality_Flag = ",Rangebin_US_Mie%Data_Quality_Flag
    txt = real2txt(Rangebin_US_Mie%Useful_Signal)
    print *,"...%Useful_Signal     = ",txt
    IF (Rangebin_US_Mie%Data_Quality_Flag .ne. 0) &
         call Print_Data_Quality_Flags(Rangebin_US_Mie%Data_Quality_Flag)

  end subroutine Print_Rangebin_US_Mie
    !  #]
  subroutine Read_Rangebin_US_Mie(Rangebin_US_Mie,fileunit,error_flag)
    !  #[
    TYPE(Rangebin_US_Mie_type), intent(out) :: Rangebin_US_Mie ! result
    integer,                    intent(in)  :: fileunit    ! input
    integer,                    intent(out) :: error_flag  ! result

    ! local variables
    integer(IntAuc) :: value_IntAuc
    real(FAdoxy)    :: value_FAdoxy

    error_flag = no_error

    ! read the byte holding the 8 data quality flags
    call read_IntAuc_from_bin_file(fileunit,value_IntAuc,error_flag)
    IF (error_flag .ne. no_error) return
    Rangebin_US_Mie%Data_Quality_Flag = value_IntAuc

    ! bit 1: overall validity. Data invalid 1, otherwise 0
    ! bit 2: 0 [Bin-level validity spare. Default 0]
    ! bit 3: Data saturation found 1, otherwise 0
    ! bit 4: Data spike found 1, otherwise 0
    ! bit 5: 0 [Measurement-level validity spare. Default 0]
    ! bit 6: Source packet invalid 1, otherwise 0
    ! bit 7: Laser Frequency not locked 1, otherwise 0
    ! bit 8: Spacecraft attitude not on target 1, otherwise 0

    ! check to make sure that bits 2 and 5 are zero
    IF (btest(value_IntAuc, 2 - 1)) THEN
       print *,"Error in Read_Rangebin_US_Mie:"
       print *,"bit 2 of Rangebin_US_Mie%Data_Quality_Flag was expected to be 0"
       print *,"but this flags seems to be set to 1 !!!!"
       error_flag = error_reading_file
       return
    END IF
    IF (btest(value_IntAuc, 5 - 1)) THEN
       print *,"Error in Read_Rangebin_US_Mie:"
       print *,"bit 5 of Rangebin_US_Mie%Data_Quality_Flag was expected to be 0"
       print *,"but this flags seems to be set to 1 !!!!"
       error_flag = error_reading_file
       return
    END IF

    ! test prints
    !IF (value_i1 .ne. 0) THEN
    !   print *,"btest(value_i1, 1 - 1) = ",btest(value_i1, 1 - 1)
    !   print *,"btest(value_i1, 2 - 1) = ",btest(value_i1, 2 - 1)
    !   print *,"btest(value_i1, 3 - 1) = ",btest(value_i1, 3 - 1)
    !   print *,"btest(value_i1, 4 - 1) = ",btest(value_i1, 4 - 1)
    !   print *,"btest(value_i1, 5 - 1) = ",btest(value_i1, 5 - 1)
    !   print *,"btest(value_i1, 6 - 1) = ",btest(value_i1, 6 - 1)
    !   print *,"btest(value_i1, 7 - 1) = ",btest(value_i1, 7 - 1)
    !   print *,"btest(value_i1, 8 - 1) = ",btest(value_i1, 8 - 1)
    !   print *,"value_i1 = ",value_i1
    !   stop
    !END IF

    ! read the usefull signal strength
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Rangebin_US_Mie%Useful_Signal = value_FAdoxy
 
  end subroutine Read_Rangebin_US_Mie
    !  #]
  !---------------------------------
  subroutine Init_Rangebin_US_Rayleigh(Rangebin_US_Rayleigh)
    !  #[

    TYPE(Rangebin_US_Rayleigh_type), intent(out) :: Rangebin_US_Rayleigh

    Rangebin_US_Rayleigh%Data_Quality_Flag = 0
    Rangebin_US_Rayleigh%Useful_Signal_Channel_A = missing_indicator_FAdoxy
    Rangebin_US_Rayleigh%Useful_Signal_Channel_B = missing_indicator_FAdoxy

  end subroutine Init_Rangebin_US_Rayleigh

    !  #]
  subroutine Print_Rangebin_US_Rayleigh(Rangebin_US_Rayleigh)
    !  #[
    TYPE(Rangebin_US_Rayleigh_type), intent(in) :: Rangebin_US_Rayleigh
    character(len=13) :: txt
    print *,"...%Data_Quality_Flag = ",Rangebin_US_Rayleigh%Data_Quality_Flag
    txt = real2txt(Rangebin_US_Rayleigh%Useful_Signal_Channel_A)
    print *,"...%Useful_Signal_Channel_A = ",txt
    txt = real2txt(Rangebin_US_Rayleigh%Useful_Signal_CHannel_B)
    print *,"...%Useful_Signal_Channel_B = ",txt
    IF (Rangebin_US_Rayleigh%Data_Quality_Flag .ne. 0) &
         call Print_Data_Quality_Flags(Rangebin_US_Rayleigh%Data_Quality_Flag)

  end subroutine Print_Rangebin_US_Rayleigh
    !  #]
  subroutine Read_Rangebin_US_Rayleigh(Rangebin_US_Rayleigh,fileunit,error_flag)
    !  #[
    TYPE(Rangebin_US_Rayleigh_type), intent(out) :: Rangebin_US_Rayleigh ! result
    integer,                intent(in)  :: fileunit    ! input
    integer,                intent(out) :: error_flag  ! result

    ! local variables
    integer(IntAuc) :: value_IntAuc
    real(FAdoxy)    :: value_FAdoxy

    error_flag = no_error

    ! read the byte holding the 8 data quality flags
    call read_IntAuc_from_bin_file(fileunit,value_IntAuc,error_flag)
    IF (error_flag .ne. no_error) return
    Rangebin_US_Rayleigh%Data_Quality_Flag = value_IntAuc

    ! bit 1: overall validity. Data invalid 1, otherwise 0
    ! bit 2: 0 [Bin-level validity spare. Default 0]
    ! bit 3: Data saturation found 1, otherwise 0
    ! bit 4: Data spike found 1, otherwise 0
    ! bit 5: 0 [Measurement-level validity spare. Default 0]
    ! bit 6: Source packet invalid 1, otherwise 0
    ! bit 7: Laser Frequency not locked 1, otherwise 0
    ! bit 8: Spacecraft attitude not on target 1, otherwise 0

    ! check to make sure that bits 2 and 5 are zero
    IF (btest(value_IntAuc, 2 - 1)) THEN
       print *,"Error in Read_Rangebin_US_Rayleigh:"
       print *,"bit 2 of Rangebin_US_Rayleigh%Data_Quality_Flag was expected to be 0"
       print *,"but this flags seems to be set to 1 !!!!"
       error_flag = error_reading_file
       return
    END IF
    IF (btest(value_IntAuc, 5 - 1)) THEN
       print *,"Error in Read_Rangebin_US_Rayleigh:"
       print *,"bit 5 of Rangebin_US_Rayleigh%Data_Quality_Flag was expected to be 0"
       print *,"but this flags seems to be set to 1 !!!!"
       error_flag = error_reading_file
       return
    END IF

    ! test prints
    !IF (value_i1 .ne. 0) THEN
    !   print *,"btest(value_i1, 1 - 1) = ",btest(value_i1, 1 - 1)
    !   print *,"btest(value_i1, 2 - 1) = ",btest(value_i1, 2 - 1)
    !   print *,"btest(value_i1, 3 - 1) = ",btest(value_i1, 3 - 1)
    !   print *,"btest(value_i1, 4 - 1) = ",btest(value_i1, 4 - 1)
    !   print *,"btest(value_i1, 5 - 1) = ",btest(value_i1, 5 - 1)
    !   print *,"btest(value_i1, 6 - 1) = ",btest(value_i1, 6 - 1)
    !   print *,"btest(value_i1, 7 - 1) = ",btest(value_i1, 7 - 1)
    !   print *,"btest(value_i1, 8 - 1) = ",btest(value_i1, 8 - 1)
    !   print *,"value_i1 = ",value_i1
    !   stop
    !END IF

    ! read the usefull signal strength, Channel A
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Rangebin_US_Rayleigh%Useful_Signal_Channel_A = value_FAdoxy
 
    ! read the usefull signal strength, Channel B
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    Rangebin_US_Rayleigh%Useful_Signal_Channel_B = value_FAdoxy
 
  end subroutine Read_Rangebin_US_Rayleigh
    !  #]
  subroutine Print_Data_Quality_Flags(Flags)
    !  #[
    integer(IntAuc), intent(in) :: Flags

    ! bit 1: overall validity. Data invalid 1, otherwise 0
    ! bit 2: 0 [Bin-level validity spare. Default 0]
    ! bit 3: Data saturation found 1, otherwise 0
    ! bit 4: Data spike found 1, otherwise 0
    ! bit 5: 0 [Measurement-level validity spare. Default 0]
    ! bit 6: Source packet invalid 1, otherwise 0
    ! bit 7: Laser Frequency not locked 1, otherwise 0
    ! bit 8: Spacecraft attitude not on target 1, otherwise 0

    print *,"bit 1 = ",btest(Flags, 1 - 1)
    print *,"bit 2 = ",btest(Flags, 2 - 1)
    print *,"bit 3 = ",btest(Flags, 3 - 1)
    print *,"bit 4 = ",btest(Flags, 4 - 1)
    print *,"bit 5 = ",btest(Flags, 5 - 1)
    print *,"bit 6 = ",btest(Flags, 6 - 1)
    print *,"bit 7 = ",btest(Flags, 7 - 1)
    print *,"bit 8 = ",btest(Flags, 8 - 1)
    print *,"Flags = ",Flags

    IF (btest(Flags, 1 - 1)) &
         print *,"flag 1: overall validity. Data invalid"
    IF (btest(Flags, 2 - 1)) &
         print *,"flag 2: spare [this flag should never be set]"
    IF (btest(Flags, 3 - 1)) &
         print *,"flag 3: Data saturation found"
    IF (btest(Flags, 4 - 1)) &
         print *,"flag 4: Data spike found"
    IF (btest(Flags, 5 - 1)) &
         print *,"flag 5: spare [this flag should never be set]"
    IF (btest(Flags, 6 - 1)) &
         print *,"flag 6: Source packet invalid"
    IF (btest(Flags, 7 - 1)) &
         print *,"flag 7: Laser Frequency not locked"
    IF (btest(Flags, 8 - 1)) &
         print *,"flag 8: Spacecraft attitude not on target"
    
  end subroutine Print_Data_Quality_Flags
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_US(US, N_range_gates, error_flag)
    !  #[
    TYPE(US_type), intent(out) :: US            ! output
    integer,       intent(in)  :: N_range_gates ! input
    integer,       intent(out) :: error_flag    ! output

        ! local variable
    integer :: r, AllocStatus

    error_flag = no_error

    IF (N_range_gates .eq. 0) THEN
       ! init them to be empty
       nullify(US%Mie)
       nullify(US%Rayleigh)
    ELSE
       allocate(US%Mie(N_range_gates+1),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_US"
          error_flag = error_allocate
          return
       END IF
       
       allocate(US%Rayleigh(N_range_gates+1),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_US"
          error_flag = error_allocate
          return
       END IF
       
       DO r=1,N_range_gates+1
          call Init_Rangebin_US_Mie(US%Mie(r))
          call Init_Rangebin_US_Rayleigh(US%Rayleigh(r))
       END DO
    END IF

  end subroutine Init_US
    !  #]
  subroutine Delete_US(US)
    !  #[
    TYPE(US_type), intent(inout) :: US

    IF (associated(US%Mie)) THEN
       deallocate(US%Mie)
       nullify(US%Mie)
    END IF

    IF (associated(US%Rayleigh)) THEN
       deallocate(US%Rayleigh)
       nullify(US%Rayleigh)
    END IF

  end subroutine Delete_US
    !  #]
  subroutine Print_US(US)
    !  #[
    TYPE(US_type), intent(in) :: US

    ! local variables
    integer :: r, s

    IF (associated(US%Mie)) THEN
       print *,"US%Mie%..."
       s = size(US%Mie)
       print "(a,i2,a)","==> size(US%Mie) = ",s,&
            " [printing first and last one]"
       DO r=1,s,s-1
          print "(' US%Mie(',i2.2,')%...')",r
          call Print_Rangebin_US_Mie(US%Mie(r))
       END DO
    ELSE
       print *,"US%Mie [seems not associated]"
    END IF

    IF (associated(US%Rayleigh)) THEN
       print *,"US%Rayleigh%..."
       s = size(US%Rayleigh)
       print "(a,i2,a)","==> size(US%Rayleigh) = ",s,&
            " [printing first and last one]"
       DO r=1,s,s-1
          print "(' US%Rayleigh(',i2.2,')%...')",r
          call Print_Rangebin_US_Rayleigh(US%Rayleigh(r))
       END DO
    ELSE
       print *,"US%Rayleigh [seems not associated]"
    END IF

  end subroutine Print_US
    !  #]
  subroutine Read_US(US, fileunit, N_range_gates, error_flag)
    !  #[
    TYPE(US_type), intent(inout) :: US          ! result
    integer,       intent(in)  :: fileunit      ! input
    integer,       intent(in)  :: N_range_gates ! input
    integer,       intent(out) :: error_flag    ! result

    ! local variables
    integer :: r

    error_flag = no_error
    IF (.not. associated(US%Mie)) THEN
       print *,"ERROR in Read_US()"
       print *,"the datastructure US%Mie(:) seems to be not associated ...."
       error_flag = error_programming
       return
    END IF

    IF (.not. associated(US%Rayleigh)) THEN
       print *,"ERROR in Read_US()"
       print *,"the datastructure US%Rayleigh(:) seems to be not associated..."
       error_flag = error_programming
    END IF

    ! check if the struct was initialised with the correct nr of rangegates
    IF (N_range_gates .ne. size(US%Mie)) THEN
       print *,"ERROR in Read_US()"
       print *,"the datastructure US%Mie(:) has room for",size(US%Mie)
       print *,"rangegates, but the nr of rangegates in the current"
       print *,"datafile seems to be ",N_range_gates
       error_flag = error_programming
       return
    END IF

    ! read the Rangebin_US_Mie structure for the 25 rangegates
    DO r=1,N_range_gates ! 25
       call Read_Rangebin_US_Mie(US%Mie(r),fileunit,error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! check if the struct was initialised with the correct nr of rangegates
    IF (N_range_gates .ne. size(US%Rayleigh)) THEN
       print *,"ERROR in Read_US()"
       print *,"the datastructure US%Rayleigh(:) has room for",&
            size(US%Rayleigh)
       print *,"rangegates, but the nr of rangegates in the current"
       print *,"datafile seems to be ",N_range_gates
       error_flag = error_programming
       return
    END IF

    ! read the Rangebin_US_Rayleigh structure for the 25 rangegates
    DO r=1,N_range_gates ! 25
       call Read_Rangebin_US_Rayleigh(US%Rayleigh(r),fileunit,error_flag)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine Read_US
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_L1B_US_MDS(US_MDS, N_meas, N_range_gates, error_flag)
    !  #[
    TYPE(L1B_US_MDS_type), pointer     :: US_MDS
    integer,               intent(in)  :: N_meas        ! input
    integer,               intent(in)  :: N_range_gates ! input
    integer,               intent(out) :: error_flag    ! output

    ! local variable
    integer :: m, AllocStatus

    error_flag = no_error

    IF (.not. associated(US_MDS)) THEN
       print *,"ERROR in Init_L1B_US_MDS: "
       print *,"pointer US_MDS seems not associated !!!"
       error_flag = error_programming
       return
    END IF
    
    call InitDateTime(US_MDS%start_of_obs_datetime)
    call Init_US(US_MDS%Obs_US, N_range_gates, error_flag)
    IF (error_flag .ne. no_error) return

    IF (N_meas .eq. 0) THEN
       ! init it to be empty
       nullify(US_MDS%Meas_US)
    ELSE
       allocate(US_MDS%Meas_US(N_meas),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L1B_US_MDS"
          error_flag = error_allocate
          return
       END IF
       
       DO m=1,N_meas
          call Init_US(US_MDS%Meas_US(m), N_range_gates, error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    END IF

  end subroutine Init_L1B_US_MDS
    !  #]
  subroutine Delete_L1B_US_MDS(US_MDS)
    !  #[
    TYPE(L1B_US_MDS_type), pointer     :: US_MDS

    ! local variable
    integer :: m

    IF (.not. associated(US_MDS)) return
    
    call Delete_US(US_MDS%Obs_US)

    IF (associated(US_MDS%Meas_US)) THEN
       DO m=1,size(US_MDS%Meas_US) 
          call Delete_US(US_MDS%Meas_US(m))
       END DO
       deallocate(US_MDS%Meas_US)
       nullify(US_MDS%Meas_US)
    END IF

  end subroutine Delete_L1B_US_MDS
    !  #]
  subroutine Print_L1B_US_MDS(US_MDS)
    !  #[
    TYPE(L1B_US_MDS_type), pointer     :: US_MDS

    ! local variables
    integer       :: m, s
    character(len=maxlen_DateTime_KVT) :: DateTimeString

    IF (associated(US_MDS)) THEN
       print *,"US_MDS%..."
    ELSE
       print *,"US_MDS [seems not associated]"
       return
    END IF

    ! print the date and time
    DateTimeString = Construct_KVT_DateTime(US_MDS%start_of_obs_datetime)
    print *,"...%start_of_obs_datetime: "//trim(DateTimeString)

    print *,"...%Obs_US%..."
    call Print_US(US_MDS%Obs_US)

    IF (associated(US_MDS%Meas_US)) THEN
       s = size(US_MDS%Meas_US)
       print "(a,i2,a)","==> size(US_MDS%Meas_US) = ",s,&
            " [printing first and last one]"
       DO m=1,s,s-1
          print "(' US_MDS%Meas_US(',i2.2,')%...')",m
          call Print_US(US_MDS%Meas_US(m))
       END DO
    ELSE
       print *,"...%Meas_US(:) = [NOT ASSOCIATED]"
    END IF

  end subroutine Print_L1B_US_MDS
    !  #]
  subroutine Read_L1B_US_MDS(US_DSD,fileunit,which_DS,N_Meas_Max,&
                             N_range_gates,US_MDS,error_flag)
    !  #[

    TYPE(DSD_type),         pointer     :: US_DSD       ! input
    integer,                intent(in)  :: fileunit     ! input
    integer,                intent(in)  :: which_DS     ! input
    integer,                intent(in)  :: N_Meas_Max   ! input 
    !                       (used for checking validity of the dataset !)
    integer,                intent(in)  :: N_range_gates ! input
    !                       (used for checking validity of the dataset !)
    TYPE(L1B_US_MDS_type),  pointer     :: US_MDS       ! result
    integer,                intent(out) :: error_flag   ! result    

    ! local variables
    integer(i8_)      :: pos, endpos
    integer           :: currentpos
    integer           :: m

    error_flag = no_error

    ! See the files: L1bP_Draft_1/aeolus/common/PFR/src/
    !        PFR_LevelOneBProductGroundWindDetectionDSR.cc/.hh       
    ! for the code that defines this datablock in the L1BP

    IF (.not. associated(US_DSD)) THEN
       print *,"ERROR in Read_L1B_US_MDS()"
       print *,"the pointer US_DSD seems not associated"
       error_flag = error_programming
       return
    END IF

    IF (.not. associated(US_MDS)) THEN
       print *,"ERROR in Read_L1B_US_MDS:"
       print *,"the pointer US_MDS seems not associated"
       error_flag = error_programming
       return
    END IF

    pos = US_DSD%start_of_ds + (which_DS-1)*US_DSD%size_of_one_dsr
    call SetFilePointer(fileunit,pos,error_flag)
    IF (error_flag .ne. no_error) return

    print *,"trying to read the binary US_MDS nr. ",which_DS

    ! read the 12 byte datetime field
    call read_DateTime_from_bin_file(fileunit,&
         US_MDS%start_of_obs_datetime,error_flag)

    call Read_US(US_MDS%Obs_US, fileunit, N_range_gates+1, error_flag)
    IF (error_flag .ne. no_error) return

    IF (.not. associated(US_MDS%Meas_US)) THEN
       print *,"ERROR in Read_L1B_US_MDS()"
       print *,"the datastructure US_MDS%Meas_US(:) seems to be"
       print *,"not associated ...."
       error_flag = error_programming
       return
    END IF

    ! check if the struct was initialised with the correct nr of measurements
    IF (N_meas_Max .ne. size(US_MDS%Meas_US)) THEN
       print *,"ERROR in Read_L1B_US_MDS()"
       print *,"the datastructure L1B%US_MDS%Meas_US(:) has room for ",&
            size(US_MDS%Meas_US)
       print *,"measurements, but the nr of measurements in the current"
       print *,"datafile seems to be ",N_meas_Max
       error_flag = error_programming
       return
    END IF

    DO m=1,size(US_MDS%Meas_US)
       call Read_US(US_MDS%Meas_US(m), fileunit, N_range_gates+1, error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! calculate expected end-position after reading this block of data
    endpos = US_DSD%start_of_ds + which_DS*US_DSD%size_of_one_dsr
    ! request the current file position
    inquire(unit=fileunit,nextrec=currentpos)
    
    IF (currentpos .ne. endpos) THEN
       print *,"ERROR in Read_L1B_US_MDS():"
       print *,"there seems to be a problem while reading this binary dataset"
       print *,"expected file position after this DSR is: ",endpos
       print *,"but the actual position seems to be:      ",currentpos
       print *,"Maybe the fileformat definition has changed ???"
       error_flag = error_reading_file
       return
    END IF

  end subroutine Read_L1B_US_MDS
    !  #]
  !---------------------------------------------------------------------
end module L1B_US_MDS
