module L1B_WV_MDS
  !  #[ Documentation
  !
  !     a module to handle the L1B Wind Velocity (WV) Measurement dataset
  !
  !     See also the L1BP files: 
  !          L1bP_Draft_1/aeolus/common/PFR/src/
  !             PFR_LevelOneBProduct???WindVelocity???DSR.cc/.hh
  !
  !     for the code that defines this datablock in the L1BP
  !
  !     Written by J. de Kloe, KNMI 
  !     created:     04-08-2005
  !
  !     Modifications:
  !      05-Sep-2005 D. Tan     Upgrade to DLR TDS v1.2
  !                             Ref AE.TN.DLR.IODD-L1B.050705 (draft).
  !      10-Jan-2006 J. de Kloe new DateTime type
  !      28-Mar-2006 P. Poli    Upgrade to DLR IODD AE.TN.DLR.APM-L1B.150206
  !      12-May-2006 D. Tan     Null use of ParKind1 module
  !      03-Nov-2006 P. Poli    moved and cleaned-up reading part
  !      03-Jan-2008 J. de Kloe added nullify statements to make the code
  !                             more robust
  !      16-Jan-2008 J. de Kloe phase out integer kind i_ 
  !
  !  #]
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE Numerics, only: i1_, i8_, real2txt
  USE DataSetDescriptor, only: DSD_type
  USE EE_CFI_Datatypes, only: DateTimeType,Construct_KVT_DateTime,InitDateTime,  &
                              read_DateTime_from_bin_file, maxlen_DateTime_KVT, &
                              maxlen_IntAc, &
                              read_IntAuc_from_bin_file, SetIntAc_or_missing, IntAuc, &
                              read_FAdoxy_from_bin_file, FAdoxy, missing_indicator_FAdoxy, &
                              int2txt
  USE DirectBinaryIO, only: SetFilePointer, AdvanceFilePointer, read_logical1_from_bin_file, &
                            read_int1_from_bin_file
  USE ErrorHandler, only: no_error, error_programming, error_allocate, error_reading_file
  !  #]
  !  #[ type definitions

  implicit none

  TYPE Alt_bin_wind_info_type ! see sec. 6.3.5, table 6-32+6-33, of L1B_iodd 
     integer(IntAuc) :: bin_quality_flag
!     integer(IntAuc) :: bin_wind_reliability_flag
     real(FAdoxy)    :: wind_velocity
  END TYPE Alt_bin_wind_info_type

  TYPE Wind_Profile_type ! see sec. 6.3.5, table 6-27+6-29, of L1B_iodd 
     integer(IntAuc) :: Mie_ref_pulse_quality_flag
     integer(IntAuc) :: Rayleigh_ref_pulse_quality_flag
     TYPE(Alt_bin_wind_info_type), dimension(:), pointer :: Mie_alt_bin_wind_info
     TYPE(Alt_bin_wind_info_type), dimension(:), pointer :: Rayleigh_alt_bin_wind_info
     !                  1234567890123456789012345678901
  END TYPE Wind_Profile_type

  TYPE L1B_WV_MDS_type ! see sec. 6.3.5, table 6-26, of L1B_iodd 
     type(DateTimeType)  :: start_of_obs_datetime    ! UTC
     integer(i1_)        :: Line_of_Sight_wind_flag
     TYPE(Wind_Profile_type)                        :: Obs_Wind_Profile
     TYPE(WInd_Profile_type), dimension(:), pointer :: Meas_Wind_Profile
     !                      1234567890123456789012345678901
  END TYPE L1B_WV_MDS_type

  !  #]
contains
  !---------------------------------------------------------------------
  subroutine Init_alt_bin_wind_info(Alt_bin_wind_info)
    !  #[
    TYPE(Alt_bin_wind_info_type), intent(out) :: Alt_bin_wind_info

    Alt_bin_wind_info%bin_quality_flag = 0
    Alt_bin_wind_info%wind_velocity = missing_indicator_FAdoxy

  end subroutine Init_alt_bin_wind_info
    !  #]
  subroutine Print_alt_bin_wind_info(Alt_bin_wind_info)
    !  #[
    TYPE(Alt_bin_wind_info_type), intent(in) :: Alt_bin_wind_info
    character(len=13) :: txt

    print *,"...%bin_quality_flag = ",Alt_bin_wind_info%bin_quality_flag
    IF (Alt_bin_wind_info%bin_quality_flag .ne. 0) &
         call Print_alt_bin_Quality_Flags(Alt_bin_wind_info%bin_quality_flag)

!    print *,"...%bin_wind_reliability_flag = ",Alt_bin_wind_info%bin_wind_reliability_flag
!    IF (Alt_bin_wind_info%bin_wind_reliability_flag .ne. 0) &
!         call Print_alt_bin_Wind_Reliab_Flags(Alt_bin_wind_info%bin_wind_reliability_flag)

    txt = real2txt(Alt_bin_wind_info%wind_velocity)
    print *,"...%wind_velocity    = ",txt

  end subroutine Print_alt_bin_wind_info
    !  #]
  subroutine Read_Alt_Bin_Wind_Info(alt_bin_wind_info, fileunit, &
                                    error_flag, ReadingRayleigh)
    !  #[
    TYPE(Alt_bin_wind_info_type), intent(out) :: alt_bin_wind_info ! result
    integer, intent(in)  :: fileunit        ! input
    integer, intent(out) :: error_flag      ! result
    logical, intent(in)  :: ReadingRayleigh ! input

    ! local variables
    integer(IntAuc) :: value_IntAuc
    real(FAdoxy)    :: value_FAdoxy

    error_flag = no_error

    ! read the byte holding the 8 data quality flags
    call read_IntAuc_from_bin_file(fileunit,value_IntAuc,error_flag)
    IF (error_flag .ne. no_error) return
    alt_bin_wind_info%bin_quality_flag = value_IntAuc

    ! bit 1: overall validity. Data invalid 1, otherwise 0
    ! bit 2: for Mie: peak not found 1, otherwise 0; for Rayleigh always 0
    ! bit 3: Data saturation found 1, otherwise 0
    ! bit 4: Data spike found 1, otherwise 0
    ! bit 5: Reference pulse invalid 1, otherwise 0
    ! bit 6: Source packet invalid 1, otherwise 0
    ! bit 7: Laser Frequency not locked 1, otherwise 0
    ! bit 8: Spacecraft attitude not on target 1, otherwise 0

    ! check to make sure that bit 2 is zero (for Rayleigh)
    IF (ReadingRayleigh) THEN
       IF (btest(value_IntAuc, 2 - 1)) THEN
          print *,"Error in Read_Alt_Bin_Wind_Info:"
          print *,"bit 2 of alt_bin_wind_info%bin_quality_flag was"
          print *,"expected to be 0, but this flags seems to be set to 1 !!!"

          ! test prints
          print *,"8bit_flag = ",value_IntAuc
          IF (value_IntAuc .ne. 0) THEN
             print *,"btest(value_IntAuc, 1 - 1) = ",btest(value_IntAuc, 1 - 1)
             print *,"btest(value_IntAuc, 2 - 1) = ",btest(value_IntAuc, 2 - 1)
             print *,"btest(value_IntAuc, 3 - 1) = ",btest(value_IntAuc, 3 - 1)
             print *,"btest(value_IntAuc, 4 - 1) = ",btest(value_IntAuc, 4 - 1)
             print *,"btest(value_IntAuc, 5 - 1) = ",btest(value_IntAuc, 5 - 1)
             print *,"btest(value_IntAuc, 6 - 1) = ",btest(value_IntAuc, 6 - 1)
             print *,"btest(value_IntAuc, 7 - 1) = ",btest(value_IntAuc, 7 - 1)
             print *,"btest(value_IntAuc, 8 - 1) = ",btest(value_IntAuc, 8 - 1)
             print *,"value_IntAuc = ",value_IntAuc
          END IF

          error_flag = error_reading_file
          return
       END IF
    END IF

!    ! read the byte holding the 8 data wind reliability flags
!    call read_int1_from_bin_file(fileunit,value_i1,error_flag)
!    IF (error_flag .ne. no_error) return
!    alt_bin_wind_info%bin_wind_reliability_flag = value_i1

   ! Check the bits (when suitably defined).

    ! read the wind velocity
    call read_FAdoxy_from_bin_file(fileunit,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return
    alt_bin_wind_info%wind_velocity = value_FAdoxy

  end subroutine Read_Alt_Bin_Wind_Info
    !  #]
  !---------------------------------
  subroutine Print_alt_bin_Quality_Flags(Flags)
    !  #[
    integer(IntAuc), intent(in) :: Flags

    ! bit 1: overall validity. Data invalid 1, otherwise 0
    ! bit 2: for Mie: peak not found 1, otherwise 0; for Rayleigh always 0
    ! bit 3: Data saturation found 1, otherwise 0
    ! bit 4: Data spike found 1, otherwise 0
    ! bit 5: Reference pulse invalid 1, otherwise 0
    ! bit 6: Source packet invalid 1, otherwise 0
    ! bit 7: Laser Frequency not locked 1, otherwise 0
    ! bit 8: Spacecraft attitude not on target 1, otherwise 0

    IF (btest(Flags, 1 - 1)) &
         print *,"flag 1: overall validity. Data invalid"
    IF (btest(Flags, 2 - 1)) &
         print *,"flag 2: Mie peak not found"
    IF (btest(Flags, 3 - 1)) &
         print *,"flag 3: data saturation found"
    IF (btest(Flags, 4 - 1)) &
         print *,"flag 4: data spike found"
    IF (btest(Flags, 5 - 1)) &
         print *,"flag 5: ref. pulse invalid"
    IF (btest(Flags, 6 - 1)) &
         print *,"flag 6: source packet invalid"
    IF (btest(Flags, 7 - 1)) &
         print *,"flag 7: laser frequency not locked"
    IF (btest(Flags, 8 - 1)) &
         print *,"flag 8: spacecraft attitude not on target"

  end subroutine Print_alt_bin_Quality_Flags
    !  #]
  subroutine Print_alt_bin_Wind_Reliab_Flags(Flags)
    !  #[
    integer(i1_), intent(in) :: Flags

    ! bit 1: TBD
    ! bit 2: TBD
    ! bit 3: TBD
    ! bit 4: TBD
    ! bit 5: TBD
    ! bit 6: TBD
    ! bit 7: TBD
    ! bit 8: TBD

    IF (btest(Flags, 1 - 1)) &
         print *,"flag 1: TBD"
    IF (btest(Flags, 2 - 1)) &
         print *,"flag 2: TBD"
    IF (btest(Flags, 3 - 1)) &
         print *,"flag 3: TBD"
    IF (btest(Flags, 4 - 1)) &
         print *,"flag 4: TBD"
    IF (btest(Flags, 5 - 1)) &
         print *,"flag 5: TBD"
    IF (btest(Flags, 6 - 1)) &
         print *,"flag 6: TBD"
    IF (btest(Flags, 7 - 1)) &
         print *,"flag 7: TBD"
    IF (btest(Flags, 8 - 1)) &
         print *,"flag 8: TBD"

  end subroutine Print_alt_bin_Wind_Reliab_Flags
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_Wind_Profile(Wind_Profile, N_range_gates, error_flag)
    !  #[
    TYPE(Wind_Profile_type), intent(out) :: Wind_Profile ! output
    integer,   intent(in)  :: N_range_gates ! input
    integer,   intent(out) :: error_flag    ! output

        ! local variable
    integer :: r, AllocStatus

    error_flag = no_error

    IF (N_range_gates .eq. 0) THEN
       ! init them to be empty
       nullify(Wind_Profile%Mie_alt_bin_wind_info)
       nullify(Wind_Profile%Rayleigh_alt_bin_wind_info)
    ELSE
       allocate(Wind_Profile%Mie_alt_bin_wind_info(N_range_gates),&
            STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_Wind_Profile"
          error_flag = error_allocate
          return
       END IF
       
       allocate(Wind_Profile%Rayleigh_alt_bin_wind_info(N_range_gates),&
            STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_Wind_Profile"
          error_flag = error_allocate
          return
       END IF

       DO r=1,N_range_gates
          call Init_Alt_Bin_Wind_Info(Wind_Profile%Mie_alt_bin_wind_info(r))
          call Init_Alt_Bin_Wind_Info(Wind_Profile%Rayleigh_alt_bin_wind_info(r))
       END DO
    END IF

  end subroutine Init_Wind_Profile
    !  #]
  subroutine Delete_Wind_Profile(Wind_Profile)
    !  #[
    TYPE(Wind_Profile_type), intent(inout) :: Wind_Profile

    IF ( associated(Wind_Profile%Mie_alt_bin_wind_info)) THEN
       deallocate(Wind_Profile%Mie_alt_bin_wind_info)
       nullify(Wind_Profile%Mie_alt_bin_wind_info)
    END IF

    IF ( associated(Wind_Profile%Rayleigh_alt_bin_wind_info)) THEN
       deallocate(Wind_Profile%Rayleigh_alt_bin_wind_info)
       nullify(Wind_Profile%Rayleigh_alt_bin_wind_info)
    END IF

  end subroutine Delete_Wind_Profile
    !  #]
  subroutine Print_Wind_Profile(Wind_Profile)
    !  #[
    TYPE(Wind_Profile_type), intent(in) :: Wind_Profile

    ! local variables
    integer :: r, s

    print *,"...%Mie_ref_pulse_quality_flag      = ",&
         Wind_Profile%Mie_ref_pulse_quality_flag
    IF (Wind_Profile%Mie_ref_pulse_quality_flag .ne. 0) &
         call Print_Ref_Pulse_Quality_Flags(&
                        Wind_Profile%Mie_ref_pulse_quality_flag)

    print *,"...%Rayleigh_ref_pulse_quality_flag = ",&
         Wind_Profile%Rayleigh_ref_pulse_quality_flag
    IF (Wind_Profile%Rayleigh_ref_pulse_quality_flag .ne. 0) &
         call Print_Ref_Pulse_Quality_Flags(&
                        Wind_Profile%Rayleigh_ref_pulse_quality_flag)

    IF (associated(Wind_Profile%Mie_alt_bin_wind_info)) THEN
       print *,"Wind_Profile%Mie_alt_bin_wind_info"
       s = size(Wind_Profile%Mie_alt_bin_wind_info)
       print "(a,i2,a)","==> size(Wind_Profile%Mie_alt_bin_wind_info) = ",s,&
            " [printing first and last one]"
       DO r=1,s,s-1
          print "(' Wind_Profile%Mie_alt_bin_wind_info(',i2.2,')%...')",r
          call Print_Alt_bin_wind_info(Wind_Profile%Mie_alt_bin_wind_info(r))
       END DO
    ELSE
       print *,"Wind_Profile%Mie_alt_bin_wind_info [seems not associated]"
    END IF

    IF (associated(Wind_Profile%Rayleigh_alt_bin_wind_info)) THEN
       print *,"Wind_Profile%Rayleigh_alt_bin_wind_info"
       s = size(Wind_Profile%Rayleigh_alt_bin_wind_info)
       print "(a,i2,a)","==> size(Wind_Profile%Rayleigh_alt_bin_wind_info) = ",s,&
            " [printing first and last one]"
       DO r=1,s,s-1
          print "(' Wind_Profile%Rayleigh_alt_bin_wind_info(',i2.2,')%...')",r
          call Print_Alt_bin_wind_info(Wind_Profile%Rayleigh_alt_bin_wind_info(r))
       END DO
    ELSE
       print *,"Wind_Profile%Rayleigh_alt_bin_wind_info [seems not associated]"
    END IF

  end subroutine Print_Wind_Profile
    !  #]
  subroutine Read_Wind_Profile(Wind_Profile,fileunit,N_range_gates,error_flag)
    !  #[
    TYPE(Wind_Profile_type), intent(inout) :: Wind_Profile ! result
    integer,   intent(in)  :: fileunit      ! input
    integer,   intent(in)  :: N_range_gates ! input
    integer,   intent(out) :: error_flag    ! result

    ! local variables
    logical         :: invalid_flag_found
    integer         :: r
    integer(IntAuc) :: value_IntAuc

    error_flag = no_error

    IF (.not. associated(Wind_Profile%Mie_alt_bin_wind_info)) THEN
       print *,"ERROR in Read_Wind_Profile()"
       print *,"the datastructure Wind_Profile%Mie_alt_bin_wind_info(:)"
       print *,"seems to be not associated ...."
       error_flag = error_programming
       return
    END IF
    IF (.not. associated(Wind_Profile%Rayleigh_alt_bin_wind_info)) THEN
       print *,"ERROR in Read_Wind_Profile()"
       print *,"the datastructure Wind_Profile%Mie_alt_bin_wind_info(:)"
       print *,"seems to be not associated..."
       error_flag = error_programming
    END IF

    ! bit 1: overall validity. Data invalid 1, otherwise 0
    ! bit 2: for Mie: peak not found 1, otherwise 0; for Rayleigh always 0
    ! bit 3: all reference pulses corresponding to the current measurement
    !        eliminated (source packet invalid, data corruption or
    !        laser frequency unlucked) 1, otherwise 0
    ! bit 4: always 0
    ! bit 5: always 0
    ! bit 6: always 0
    ! bit 7: always 0
    ! bit 8: always 0

    call read_IntAuc_from_bin_file(fileunit,value_IntAuc,error_flag)
    IF (error_flag .ne. no_error) return
    Wind_Profile%Mie_ref_pulse_quality_flag = value_IntAuc

    IF (value_IntAuc .ne. 0) THEN
       ! check to make sure that bits 4 upto 8 are zero
       invalid_flag_found = .false.
!       IF (btest(value_IntAuc, 4 - 1)) invalid_flag_found = .true.
!       IF (btest(value_IntAuc, 5 - 1)) invalid_flag_found = .true.
!       IF (btest(value_IntAuc, 6 - 1)) invalid_flag_found = .true.
!       IF (btest(value_IntAuc, 7 - 1)) invalid_flag_found = .true.
!       IF (btest(value_IntAuc, 8 - 1)) invalid_flag_found = .true.
       
       IF (invalid_flag_found) THEN
          print *,"Error in read_Wind_Profile:"
          print *,"invalid bit value found in:"
          print *,"Wind_Profile%Mie_ref_pulse_quality_flag"
          
          ! test prints
          print *,"8bit_flag = ",value_IntAuc
          IF (value_IntAuc .ne. 0) THEN
             print *,"btest(value_IntAuc, 1 - 1) = ",btest(value_IntAuc, 1 - 1)
             print *,"btest(value_IntAuc, 2 - 1) = ",btest(value_IntAuc, 2 - 1)
             print *,"btest(value_IntAuc, 3 - 1) = ",btest(value_IntAuc, 3 - 1)
             print *,"btest(value_IntAuc, 4 - 1) = ",btest(value_IntAuc, 4 - 1)
             print *,"btest(value_IntAuc, 5 - 1) = ",btest(value_IntAuc, 5 - 1)
             print *,"btest(value_IntAuc, 6 - 1) = ",btest(value_IntAuc, 6 - 1)
             print *,"btest(value_IntAuc, 7 - 1) = ",btest(value_IntAuc, 7 - 1)
             print *,"btest(value_IntAuc, 8 - 1) = ",btest(value_IntAuc, 8 - 1)
             print *,"value_IntAuc = ",value_IntAuc
          END IF
          
          error_flag = error_reading_file
          return
       END IF
    END IF

    call read_IntAuc_from_bin_file(fileunit,value_IntAuc,error_flag)
    IF (error_flag .ne. no_error) return
    Wind_Profile%Rayleigh_ref_pulse_quality_flag = value_IntAuc

    IF (value_IntAuc .ne. 0) THEN
       ! check to make sure that bits 4 upto 8 are zero
       invalid_flag_found = .false.
!       IF (btest(value_IntAuc, 4 - 1)) invalid_flag_found = .true.
!       IF (btest(value_IntAuc, 5 - 1)) invalid_flag_found = .true.
!       IF (btest(value_IntAuc, 6 - 1)) invalid_flag_found = .true.
!       IF (btest(value_IntAuc, 7 - 1)) invalid_flag_found = .true.
!       IF (btest(value_IntAuc, 8 - 1)) invalid_flag_found = .true.
       
       IF (invalid_flag_found) THEN
          print *,"Error in read_Wind_Profile:"
          print *,"invalid bit value found in:"
          print *,"Wind_Profile%Rayleigh_ref_pulse_quality_flag"
          
          ! test prints
          print *,"8bit_flag = ",value_IntAuc
          IF (value_IntAuc .ne. 0) THEN
             print *,"btest(value_IntAuc, 1 - 1) = ",btest(value_IntAuc, 1 - 1)
             print *,"btest(value_IntAuc, 2 - 1) = ",btest(value_IntAuc, 2 - 1)
             print *,"btest(value_IntAuc, 3 - 1) = ",btest(value_IntAuc, 3 - 1)
             print *,"btest(value_IntAuc, 4 - 1) = ",btest(value_IntAuc, 4 - 1)
             print *,"btest(value_IntAuc, 5 - 1) = ",btest(value_IntAuc, 5 - 1)
             print *,"btest(value_IntAuc, 6 - 1) = ",btest(value_IntAuc, 6 - 1)
             print *,"btest(value_IntAuc, 7 - 1) = ",btest(value_IntAuc, 7 - 1)
             print *,"btest(value_IntAuc, 8 - 1) = ",btest(value_IntAuc, 8 - 1)
             print *,"value_IntAuc = ",value_IntAuc
          END IF
          
          error_flag = error_reading_file
          return
       END IF
    END IF

    ! check if the struct was initialised with the correct nr of rangegates
    IF (N_range_gates .ne. size(Wind_Profile%Mie_alt_bin_wind_info)) THEN
       print *,"ERROR in Read_Wind_Profile()"
       print *,"the datastructure Wind_Profile%Mie_alt_bin_wind_info(:)"
       print *,"has room for",size(Wind_Profile%Mie_alt_bin_wind_info)
       print *,"rangegates, but the nr of rangegates in the current"
       print *,"datafile seems to be ",N_range_gates
       error_flag = error_programming
       return
    END IF

    ! read the alt_bin_wind_info structure for the 24 Mie rangegates
    DO r=1,N_range_gates ! 24
       call read_Alt_Bin_Wind_Info(Wind_Profile%Mie_alt_bin_wind_info(r),&
                                   fileunit,error_flag,ReadingRayleigh=.false.)
       IF (error_flag .ne. no_error) return
    END DO

    ! check if the struct was initialised with the correct nr of rangegates
    IF (N_range_gates .ne. size(Wind_Profile%Rayleigh_alt_bin_wind_info)) THEN
       print *,"ERROR in Read_Wind_Profile()"
       print *,"the datastructure Wind_Profile%Rayleigh_alt_bin_wind_info(:)"
       print *,"has room for",size(Wind_Profile%Rayleigh_alt_bin_wind_info)
       print *,"rangegates, but the nr of rangegates in the current"
       print *,"datafile seems to be ",N_range_gates
       error_flag = error_programming
       return
    END IF

    ! read the alt_bin_wind_info structure for the 24 Rayleigh rangegates
    DO r=1,N_range_gates ! 24
       call read_Alt_Bin_Wind_Info(Wind_Profile%Rayleigh_alt_bin_wind_info(r),&
                                   fileunit,error_flag,ReadingRayleigh=.true.)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine Read_Wind_Profile
    !  #]
  !---------------------------------
  subroutine Print_Ref_Pulse_Quality_Flags(Flags)
    !  #[
    integer(IntAuc), intent(in) :: Flags

    ! bit 1: overall validity. Data invalid 1, otherwise 0
    ! bit 2: for Mie: peak not found 1, otherwise 0; for Rayleigh always 0
    ! bit 3: all reference pulses corresponding to the current measurement
    !        eliminated (source packet invalid, data corruption or
    !        laser frequency unlucked) 1, otherwise 0
    ! bit 4: always 0
    ! bit 5: always 0
    ! bit 6: always 0
    ! bit 7: always 0
    ! bit 8: always 0

    IF (btest(Flags, 1 - 1)) &
         print *,"flag 1: overall validity. Data invalid"
    IF (btest(Flags, 2 - 1)) &
         print *,"flag 2: Mie peak not found"
    IF (btest(Flags, 3 - 1)) &
         print *,"flag 3: all ref. pulses are eliminated"

  end subroutine Print_Ref_Pulse_Quality_Flags
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_L1B_WV_MDS(WV_MDS, N_meas, N_range_gates, error_flag)
    !  #[
    TYPE(L1B_WV_MDS_type), pointer     :: WV_MDS
    integer,               intent(in)  :: N_meas        ! input
    integer,               intent(in)  :: N_range_gates ! input
    integer,               intent(out) :: error_flag    ! output

    ! local variable
    integer :: m, AllocStatus

    error_flag = no_error

    IF (.not. associated(WV_MDS)) THEN
       print *,"ERROR in Init_L1B_WV_MDS: "
       print *,"pointer WV_MDS seems not associated !!!"
       error_flag = error_programming
       return
    END IF
    
    call InitDateTime(WV_MDS%start_of_obs_datetime)
    call Init_Wind_Profile(WV_MDS%Obs_Wind_Profile, N_range_gates, error_flag)
    IF (error_flag .ne. no_error) return

    IF (N_meas .eq. 0) THEN
       ! init it to be empty
       nullify(WV_MDS%Meas_Wind_Profile)
    ELSE
       allocate(WV_MDS%Meas_Wind_Profile(N_meas),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L1B_WV_MDS"
          error_flag = error_allocate
          return
       END IF
       
       DO m=1,N_meas
          call Init_Wind_Profile(WV_MDS%Meas_Wind_Profile(m), &
               N_range_gates, error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    END IF

  end subroutine Init_L1B_WV_MDS
    !  #]
  subroutine Delete_L1B_WV_MDS(WV_MDS)
    !  #[
    TYPE(L1B_WV_MDS_type), pointer     :: WV_MDS

    ! local variable
    integer :: m

    IF (.not. associated(WV_MDS)) return
    
    call Delete_Wind_Profile(WV_MDS%Obs_Wind_Profile)

    IF (associated(WV_MDS%Meas_Wind_Profile)) THEN
       DO m=1,size(WV_MDS%Meas_Wind_Profile) 
          call Delete_Wind_Profile(WV_MDS%Meas_Wind_Profile(m))
       END DO
       deallocate(WV_MDS%Meas_Wind_Profile)
       nullify(WV_MDS%Meas_Wind_Profile)
    END IF

  end subroutine Delete_L1B_WV_MDS
    !  #]
  subroutine Print_L1B_WV_MDS(WV_MDS)
    !  #[
    TYPE(L1B_WV_MDS_type), pointer :: WV_MDS

    ! local variables
    integer           :: m, s, error_flag_local
    character(len=maxlen_IntAc) :: txt
    character(len=maxlen_DateTime_KVT) :: DateTimeString

    IF (associated(WV_MDS)) THEN
       print *,"WV_MDS%..."
    ELSE
       print *,"WV_MDS [seems not associated]"
       return
    END IF

    ! print the date and time
    DateTimeString = Construct_KVT_DateTime(WV_MDS%start_of_obs_datetime)
    print *,"...%start_of_obs_datetime: "//trim(DateTimeString)

    call SetIntAc_or_missing(WV_MDS%Line_of_Sight_wind_flag,txt,error_flag_local)
    print *,"...%Line_of_Sight_wind_flag = ", txt

    print *,"...%Obs_Wind_Profile%..."
    call Print_Wind_Profile(WV_MDS%Obs_Wind_Profile)

    IF (associated(WV_MDS%Meas_Wind_Profile)) THEN
       s = size(WV_MDS%Meas_Wind_Profile)
       print "(a,i2,a)","==> size(WV_MDS%Meas_Wind_Profile) = ",s,&
            " [printing first and last one]"
       DO m=1,s,s-1
          print "(' WV_MDS%Meas_Wind_Profile(',i2.2,')%...')",m
          call Print_Wind_Profile(WV_MDS%Meas_Wind_Profile(m))
       END DO
    ELSE
       print *,"...%Meas_Wind_Profile(:) = [NOT ASSOCIATED]"
    END IF

  end subroutine Print_L1B_WV_MDS
    !  #]
  subroutine Read_L1B_WV_MDS(WV_DSD,fileunit,which_DS,N_Meas_Max,&
                             N_range_gates,WV_MDS,error_flag)
    !  #[

    TYPE(DSD_type),        pointer     :: WV_DSD       ! input
    integer,               intent(in)  :: fileunit     ! input
    integer,               intent(in)  :: which_DS     ! input
    integer,               intent(in)  :: N_Meas_Max   ! input 
    !                      (used for checking validity of the dataset !)
    integer,               intent(in)  :: N_range_gates ! input
    !                      (used for checking validity of the dataset !)
    TYPE(L1B_WV_MDS_type), pointer     :: WV_MDS       ! result
    integer,               intent(out) :: error_flag   ! result    

    ! local variables
    integer(i8_)      :: pos, endpos
    integer           :: currentpos
    integer(i1_)      :: value_i1
    integer           :: m

    error_flag = no_error

    ! See the files: L1bP_Draft_1/aeolus/common/PFR/src/
    !        PFR_LevelOneBProduct???WindVelocityDSR.cc/.hh       
    ! for the code that defines this datablock in the L1BP

    IF (.not. associated(WV_DSD)) THEN
       print *,"ERROR in Read_L1B_WV_MDS()"
       print *,"the pointer WV_DSD seems not associated"
       error_flag = error_programming
       return
    END IF

    IF (.not. associated(WV_MDS)) THEN
       print *,"ERROR in Read_L1B_WV_MDS:"
       print *,"the pointer wv_MDS seems not associated"
       error_flag = error_programming
       return
    END IF

    pos = WV_DSD%start_of_ds + (which_DS-1)*WV_DSD%size_of_one_dsr
    call SetFilePointer(fileunit,pos,error_flag)
    IF (error_flag .ne. no_error) return

    print *,"trying to read the binary WV_MDS nr. ",which_DS

    ! read the 12 byte datetime field
    call read_DateTime_from_bin_file(fileunit,&
         WV_MDS%start_of_obs_datetime,error_flag)

    call read_int1_from_bin_file(fileunit,value_i1,error_flag)
    IF (error_flag .ne. no_error) return
    WV_MDS%Line_of_Sight_wind_flag = value_i1
    
    call Read_Wind_Profile(WV_MDS%Obs_Wind_Profile, fileunit, &
                           N_range_gates, error_flag)
    IF (error_flag .ne. no_error) return

    IF (.not. associated(WV_MDS%Meas_Wind_Profile)) THEN
       print *,"ERROR in Read_L1B_WV_MDS()"
       print *,"the datastructure WV_MDS%Meas_Wind_Profile(:) seems to be"
       print *,"not associated ...."
       error_flag = error_programming
       return
    END IF

    ! check if the struct was initialised with the correct nr of measurements
    IF (N_meas_Max .ne. size(WV_MDS%Meas_Wind_Profile)) THEN
       print *,"ERROR in Read_L1B_WV_MDS()"
       print *,"the datastructure L1B%WV_MDS%Meas_Wind_Profile(:) "
       print *,"has room for ",size(WV_MDS%Meas_Wind_Profile)
       print *,"measurements, but the nr of measurements in the current"
       print *,"datafile seems to be ",N_meas_Max
       error_flag = error_programming
       return
    END IF

    DO m=1,size(WV_MDS%Meas_Wind_Profile)
       call Read_Wind_Profile(WV_MDS%Meas_Wind_Profile(m), fileunit, &
                              N_range_gates, error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! calculate expected end-position after reading this block of data
    endpos = WV_DSD%start_of_ds + which_DS*WV_DSD%size_of_one_dsr
    ! request the current file position
    inquire(unit=fileunit,nextrec=currentpos)
    
    IF (currentpos .ne. endpos) THEN
       print *,"ERROR in Read_L1B_WV_MDS():"
       print *,"there seems to be a problem while reading this binary dataset"
       print *,"expected file position after this DSR is: ",endpos
       print *,"but the actual position seems to be:     ",currentpos
       print *,"Maybe the fileformat definition has changed ???"
       error_flag = error_reading_file
       return
    END IF

  end subroutine Read_L1B_WV_MDS
    !  #]
  !---------------------------------------------------------------------
end module L1B_WV_MDS
