module L2B_AMD_SPH
  !  #[ Documentation

  !
  ! this module defines the Specific Product Header for the ee_cfi fileformat
  ! for the Auxiliary Meteoroligical data files
  ! The SPH format may differ for different filedefinitions, therefore
  ! we use a private definition for thsi filetype only
  !
  ! for a description, see 
  !    AE-IF-ECMWF-L2BP-001 (the L2B IODD), section 5.1.2.
  !
  ! Written by J. de Kloe (KNMI)
  !
  ! Modifications:
  !  Poli       09-Feb-2006 Update in line with TN 2.1 v2.0
  !                         Added routines Delete_L2B_AMD_SPH and
  !                         Write_L2B_AMD_SPH_to_KVT
  !  J. de Kloe 02-Mar-2006 adapted to changes in the XML_module
  !  J. de Kloe 13-Mar-2006 added prefix/error_flag to the routine print_DSD
  !  D. Tan     25-Apr-2006 Add Num_Input_Files and Input DSDs 
  !                         (IODD draft v1.0C)
  !  D. Tan     12-May-2006 Null use of ParKind1 module
  !  P. Poli    17-Aug-2006 Removed obsolete variables
  !  P. Poli    18-Sep-2006 Complete routine Fill_L2B_AMD_SPH_items
  !  J. de Kloe 10-Nov-2006 moved Read_AMD_SPH_KVT routine from readamddata.F90
  !                         into this file
  !  J. de Kloe 02-Mar-2007 replaced subroutine add_L2B_AMD_SPH_item() by the
  !                         more generic xml_add_SPH_item defined in 
  !                         xml_module.F90
  !  D. Tan     09-Mar-2007 Use ee_sph_name.
  !  J. de Kloe 24-Apr-2007 added function GetByteSize_AMD_SPH_KVT
  !  J. de Kloe 24-Apr-2007 phase out the SummaryVars code
  !  J. de Kloe 07-Jun-2007 adapted to changed xml interface
  !  J. de Kloe 14-Nov-2007 fixed xml format for Fcst_Initial_Time,
  !                         added AMD_DS_Type_Input parameter, and 
  !                         some small clean-ups.
  !  J. de Kloe 16-Jan-2008 phase out integer kind i_ 
  !
  !  #] 
  !  #[ Modules used

  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE XML_module, only: max_str_len, xml_node_type, xml_node_get_string, &
                        xml_set_fixed_header, check_length_xml_item, &
                        check_string_length, xml_tree_set_string_node_value, &
                        xml_add_SPH_item, xml_tree_add_child, &
                        ee_sph_name, ee_vh_name, &
                        xml_tree_add_and_set_attribute

  USE EE_CFI_Datatypes, only: maxlen_DateTime_KVT, &
                              maxlen_IntAs, maxlen_IntAus, &
                              maxlen_IntAl, &
                              IntAus, IntAs, IntAul, IntAl, &
                              SetIntAus, SetIntAs, SetIntAul, SetIntAl, &
                              GetIntAus, &
                              Remove_Plus_Sign, DateTimeType, &
                              Construct_KVT_DateTime, &
                              Convert_XML_to_KVT_DateTime, &
                              Convert_KVT_to_XML_DateTime, &
                              maxlen_DateTime_XML

  USE DataSetDescriptor, only: DSD_type, Init_DSD, Print_DSD, &
                               Add_DSD_to_xml_SPH, Write_DSD_to_KVT, &
                               GetByteSize_DSD_KVT, &
                               maxlen_DS_Name, maxlen_DS_Type

  USE KVT_module, only: write_kvt_string, write_spare_kvt_string, &
                        Init_KVT_module, &
                        read_kvt_string, read_spare_kvt_string
 
  ! support modules
  USE ErrorHandler, only: no_error, error_allocate, error_programming, &
                          error_reading_file
  USE Compiler_Features, only: newline
  USE Logging, only: logmsg,log_debug,log_info,log_progress,log_warn,log_error

  !  #]
  !  #[ Type definitions and parameters
  implicit none

  ! Names of attached datasets in AMD products
  character(len=maxlen_DS_Name) ,parameter :: AMD_DS_Name_GeoADS1=&
                                     "Geolocation_ADS1 off-nadir  "
  character(len=maxlen_DS_Name) ,parameter :: AMD_DS_Name_GeoADS2=&
                                     "Geolocation_ADS2 nadir      "
  character(len=maxlen_DS_Name) ,parameter :: AMD_DS_Name_MDS1=&
                                     "Meteorological DS1 off-nadir"
  character(len=maxlen_DS_Name) ,parameter :: AMD_DS_Name_MDS2=&
                                     "Meteorological DS2 nadir    "

  ! Types (Annotation/Measurement) of attached datasets in AMD products
  character(len=maxlen_DS_Type) ,parameter :: AMD_DS_Type_GeoADS1=&
                                     "A"
  character(len=maxlen_DS_Type) ,parameter :: AMD_DS_Type_GeoADS2=&
                                     "A"
  character(len=maxlen_DS_Type) ,parameter :: AMD_DS_Type_MDS1=&
                                     "M"
  character(len=maxlen_DS_Type) ,parameter :: AMD_DS_Type_MDS2=&
                                     "M"
  character(len=maxlen_DS_Type) ,parameter :: AMD_DS_Type_Input=&
                                     "R"

  integer, parameter :: maxlen_SPH_Descriptor   = 28
  integer, parameter :: maxlen_Ref_NWP_Suite    = 20
  integer, parameter :: maxlen_Model_Grid_Type  = 2
!  integer, parameter :: maxlen_DSD_Met_ADS      = 65 ! never used

  TYPE L2B_AMD_SPH_type
     !  #[ 
!! currently in use, but should probably be renamed
!     character(len=maxlen_Model_Name)           :: Model_Name
!               ! to be renamed to: Ref_NWP_Suite ?
!! currently in use, should these be added to TN2.1?
!     character(len=maxlen_Model_Version_Number) :: Model_Version_Number
!! to be replaced by a character string
!     integer                :: Num_BRC

! this one is not in the IODD, so removed it. JK, 24-Apr-2007
! it has been used by the auxiliary module in previous versions
!     character(len=max_str_len) :: DSD_Met_ADS

! names defined in TN2.1
     character(len=maxlen_SPH_Descriptor)   :: SPH_Descriptor
     ! Spare_1
     character(len=maxlen_Ref_NWP_Suite)    :: Ref_NWP_Suite
     character(len=maxlen_DateTime_KVT)     :: Fcst_Initial_Time
     character(len=maxlen_IntAl)            :: Model_Timestep
     character(len=maxlen_Model_Grid_Type)  :: Model_Grid_Type
     character(len=maxlen_IntAs)            :: Model_Resol_Par1
     character(len=maxlen_IntAs)            :: Model_Resol_Par2
     character(len=maxlen_IntAus)           :: Num_of_Model_Layers
     character(len=maxlen_IntAl)            :: Num_Records_in_DS1
     character(len=maxlen_IntAl)            :: Num_Records_in_DS2
     character(len=maxlen_IntAl)            :: Num_Avail_L1B_Obs
     character(len=maxlen_IntAl)            :: Num_Missing_L1B_Obs
     character(len=maxlen_IntAl)            :: Num_Computed_Locations
     ! Spare_2
     character(len=maxlen_IntAus)           :: Num_Input_Files
     character(len=maxlen_IntAus)           :: Num_Files_Predict_Orbit
     ! Spare_3
     logical :: DSDs_are_available ! set .true. when reading was successfull
     TYPE(DSD_type), pointer :: GeoADS1_DSD
     TYPE(DSD_type), pointer :: GeoADS2_DSD
     TYPE(DSD_type), pointer :: MDS1_DSD
     TYPE(DSD_type), pointer :: MDS2_DSD
     TYPE(DSD_type), dimension(:), pointer :: Input_DSD
  END TYPE L2B_AMD_SPH_type
    !  #]

  !  #]
  !  #[ Module interface
! all is public for the moment
  ! Type definitions
  public :: L2B_AMD_SPH_type  
  ! subroutines
  public :: Init_L2B_AMD_SPH, Print_L2B_AMD_SPH, Delete_L2B_AMD_SPH, &
            Fill_L2B_AMD_SPH_items, Copy_L2B_AMD_SPH_to_xml_tree, &
            Write_L2B_AMD_SPH_to_KVT
  !  #]
contains
  !---------------------------------
! WARNING: this init routine sets num_input_files=10
! and uses this to allocate L2B_AMD_SPH%Input_DSD(:)
! TODO: Dont forget to remove this line when actual data is entered into these DSD's!
  subroutine Init_L2B_AMD_SPH(L2B_AMD_SPH,error_flag)
    !  #[

    TYPE(L2B_AMD_SPH_type), pointer :: L2B_AMD_SPH
    integer, intent(out) :: error_flag
    
    ! local variable
    integer(IntAus)         :: Num_Input_DSDs
    integer                 :: p, AllocStatus
    TYPE(DSD_Type), pointer :: dsd_ptr

    error_flag = no_error
    IF (.not. associated(L2B_AMD_SPH)) THEN
       print *,"ERROR in Init_L2B_AMD_SPH():"
       print *,"The pointer to this L2B_AMD_SPH seems not associated !!"
       error_flag = error_programming
       return
    END IF

! never used
!    L2B_AMD_SPH%DSD_Met_ADS(:)          = ' '

    ! see the remark in MainProductHeader.F90 for the reason why I use
    ! the form x(:)=' ' to initialise the strings

    L2B_AMD_SPH%SPH_Descriptor(:)      = ' '
     ! Spare_1
    L2B_AMD_SPH%Ref_NWP_Suite(:)       = ' '
    L2B_AMD_SPH%Fcst_Initial_Time(:)   = ' '
    L2B_AMD_SPH%Model_Timestep(:)      = ' '
    L2B_AMD_SPH%Model_Grid_Type(:)     = ' '
    L2B_AMD_SPH%Model_Resol_Par1(:)    = ' '
    L2B_AMD_SPH%Model_Resol_Par2(:)    = ' '
    L2B_AMD_SPH%Num_of_Model_Layers(:) = ' '
    L2B_AMD_SPH%Num_Records_in_DS1(:)  = ' '
    L2B_AMD_SPH%Num_Records_in_DS2(:)  = ' '
    L2B_AMD_SPH%Num_Avail_L1B_Obs(:)   = ' '
    L2B_AMD_SPH%Num_Missing_L1B_Obs(:) = ' '
    L2B_AMD_SPH%Num_Computed_Locations(:) = ' '
     ! Spare_2
    L2B_AMD_SPH%Num_Input_Files(:)         = '+00010'
    L2B_AMD_SPH%Num_Files_Predict_Orbit(:) = '+00004'
    ! Spare_3
    L2B_AMD_SPH%DSDs_are_available = .false.
    nullify(L2B_AMD_SPH%GeoADS1_DSD)
    nullify(L2B_AMD_SPH%GeoADS2_DSD)
    nullify(L2B_AMD_SPH%MDS1_DSD)
    nullify(L2B_AMD_SPH%MDS2_DSD)

    call GetIntAus( L2B_AMD_SPH%Num_Input_Files, Num_Input_DSDs, error_flag )
    IF (error_flag .ne. no_error) return

    IF (Num_Input_DSDs .eq. 0) THEN
       ! init it to be empty
       nullify(L2B_AMD_SPH%Input_DSD)
    ELSE
       allocate(L2B_AMD_SPH%Input_DSD(Num_Input_DSDs),&
                STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L2B_AMD_SPH"
          error_flag = error_allocate
          return
       END IF

       DO p=1,Num_Input_DSDs
          dsd_ptr => L2B_AMD_SPH%Input_DSD(p)
          call init_dsd(dsd_ptr, error_flag)
          IF (error_flag .ne. no_error) return
       END DO
   END IF

  end subroutine Init_L2B_AMD_SPH
    !  #]
  subroutine Print_L2B_AMD_SPH(L2B_AMD_SPH,error_flag)
    !  #[
    TYPE(L2B_AMD_SPH_type), pointer     :: L2B_AMD_SPH
    integer,                intent(out) :: error_flag

    ! local variable
    character(len=256) :: txt
    character(len=256) :: prefix_string
    integer            :: p, s
    TYPE(DSD_Type), pointer :: dsd_ptr

    error_flag = no_error

    IF (.not. associated(L2B_AMD_SPH)) THEN
       print *,"[pointer to L2B_AMD_SPH is not associated so nothing to print]"
       return
    END IF

    txt = L2B_AMD_SPH%SPH_Descriptor(:)
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%SPH_Descriptor                = ",trim(txt)

    ! Spare_1

    txt = L2B_AMD_SPH%Ref_NWP_Suite(:)
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Ref_NWP_Suite                 = ",trim(txt)

    txt = L2B_AMD_SPH%Fcst_Initial_Time(:)
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Fcst_Initial_Time             = ",trim(txt)

    txt = L2B_AMD_SPH%Model_Timestep(:)
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Model_Timestep                = ",trim(txt)

    txt = L2B_AMD_SPH%Model_Grid_Type(:)
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Model_Grid_Type               = ",trim(txt)

    txt = L2B_AMD_SPH%Model_Resol_Par1(:)
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Model_Resol_Par1              = ",trim(txt)

    txt = L2B_AMD_SPH%Model_Resol_Par2(:)
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Model_Resol_Par2              = ",trim(txt)

    txt = L2B_AMD_SPH%Num_of_Model_Layers(:)
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_of_Model_Layers           = ",trim(txt)

    txt = L2B_AMD_SPH%Num_Records_in_DS1(:)
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Records_in_DS1            = ",trim(txt)

    txt = L2B_AMD_SPH%Num_Records_in_DS2(:)
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Records_in_DS2            = ",trim(txt)

    txt = L2B_AMD_SPH%Num_Avail_L1B_Obs(:)
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Avail_L1B_Obs             = ",trim(txt)

    txt = L2B_AMD_SPH%Num_Missing_L1B_Obs(:)
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Missing_L1B_Obs           = ",trim(txt)

    txt = L2B_AMD_SPH%Num_Computed_Locations(:)
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Computed_Locations        = ",trim(txt)

     ! Spare_2

    txt = L2B_AMD_SPH%Num_Input_Files(:)
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Input_Files               = ",trim(txt)

    txt = L2B_AMD_SPH%Num_Files_Predict_Orbit(:)
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Files_Predict_Orbit       = ",trim(txt)

    ! Spare_3

    ! the list of 4 DSD's
    print *,"...%DSDs_are_available            = ",&
         L2B_AMD_SPH%DSDs_are_available
    print *,"L2B_AMD_SPH%GeoADS1_DSD%..."
    call Print_DSD(L2B_AMD_SPH%GeoADS1_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    print *,"L2B_AMD_SPH%GeoADS2_DSD%..."
    call Print_DSD(L2B_AMD_SPH%GeoADS2_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    print *,"L2B_AMD_SPH%MDS1_DSD%..."
    call Print_DSD(L2B_AMD_SPH%MDS1_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    print *,"L2B_AMD_SPH%MDS2_DSD%..."
    call Print_DSD(L2B_AMD_SPH%MDS2_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return

    IF (associated(L2B_AMD_SPH%Input_DSD)) THEN
       s = size(L2B_AMD_SPH%Input_DSD)
       DO p=1,s

          write(prefix_string, &
               "('L2B_AMD_SPH%Input_DSD(',i2.2,')')", err=999) p

          dsd_ptr => L2B_AMD_SPH%Input_DSD(p)
          call print_dsd(dsd_ptr, prefix=prefix_string, &
                         error_flag=error_flag)
          IF (error_flag .ne. no_error) return

       END DO
    ELSE
       print *,"L2B_AMD_SPH%Input_DSD(:) = [NOT ASSOCIATED]"
    END IF

    return

999 print *,"ERROR in Print_L2B_AMD_SPH:"
    print *,"Conversion from type Integer to string failed for value: ",p
    error_flag = error_programming
    return

  end subroutine Print_L2B_AMD_SPH
    !  #]
  subroutine Delete_L2B_AMD_SPH(L2B_AMD_SPH)
    !  #[
    TYPE(L2B_AMD_SPH_type), pointer :: L2B_AMD_SPH

    IF (associated(L2B_AMD_SPH%GeoADS1_DSD)) &
         deallocate(L2B_AMD_SPH%GeoADS1_DSD)
    IF (associated(L2B_AMD_SPH%GeoADS2_DSD)) &
         deallocate(L2B_AMD_SPH%GeoADS2_DSD)
    IF (associated(L2B_AMD_SPH%MDS1_DSD)) &
         deallocate(L2B_AMD_SPH%MDS1_DSD)
    IF (associated(L2B_AMD_SPH%MDS2_DSD)) &
         deallocate(L2B_AMD_SPH%MDS2_DSD)
    IF (associated(L2B_AMD_SPH%Input_DSD)) &
         deallocate(L2B_AMD_SPH%Input_DSD)

  end subroutine Delete_L2B_AMD_SPH
    !  #]
  subroutine Get_L2B_AMD_SPH_from_xml_tree(xml_tree, path_in, &
                                           L2B_AMD_SPH, error_flag)
    !  #[

    ! History:
    !   JdK    Original code
    !   PPoli  Update to match TN 2.1 v2.0
    TYPE(xml_node_type),    pointer     :: xml_tree    ! input
    character(len=*),       intent(in)  :: path_in     ! input
    TYPE(L2B_AMD_SPH_type), pointer     :: L2B_AMD_SPH ! result
    integer,                intent(out) :: error_flag  ! output

    ! local variables
    character(len=max_str_len)       :: path
    character(len=max_str_len)       :: text
!    character(len=maxlen_Num_Layers) :: Num_Layers_String
!    character(len=maxlen_Num_BRC)    :: Num_BRC_String
!    character(len=maxlen_Num_DSD)    :: Num_DSD_String
!    character(len=5)                 :: format_str

    ! init
    error_flag = no_error

    IF (.not. associated(L2B_AMD_SPH)) THEN
       print *,"ERROR in Get_L2B_AMD_SPH_from_xml_tree():"
       print *,"the pointer L2B_AMD_SPH seems not associated !!!"
       error_flag = error_programming
       return
    END IF

    IF (.not. associated(xml_tree)) THEN
       print *,"ERROR in Get_L2B_AMD_SPH_from_xml_tree():"
       print *,"the pointer xml_tree seems not associated !!!"
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! copying the L2B_AMD_SPH XML elements
    !------------------------------

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Sph_Descriptor"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2B_AMD_SPH%SPH_Descriptor = &
         check_length_xml_item(text,path,maxlen_SPH_Descriptor,&
                               "Get_L2B_AMD_SPH_from_xml_tree")

    ! Spare_1

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Ref_NWP_Suite"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2B_AMD_SPH%Ref_NWP_Suite= &
         check_length_xml_item(text,path,maxlen_Ref_NWP_Suite,&
                               "Get_L2B_AMD_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Fcst_Initial_Time"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,maxlen_DateTime_XML,&
                                 "Get_L2B_AMD_SPH_from_xml_tree")
    L2B_AMD_SPH%Fcst_Initial_Time= Convert_XML_to_KVT_DateTime(text)        

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Model_Timestep"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2B_AMD_SPH%Model_Timestep= &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L2B_AMD_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Model_Grid_Type"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2B_AMD_SPH%Model_Grid_Type= &
         check_length_xml_item(text,path,maxlen_Model_Grid_Type,&
                               "Get_L2B_AMD_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Model_Resol_Par1"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2B_AMD_SPH%Model_Resol_Par1= &
         check_length_xml_item(text,path,maxlen_IntAs,&
                               "Get_L2B_AMD_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Model_Resol_Par2"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2B_AMD_SPH%Model_Resol_Par2= &
         check_length_xml_item(text,path,maxlen_IntAs,&
                               "Get_L2B_AMD_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_of_Model_Layers"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2B_AMD_SPH%Num_of_Model_Layers= &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2B_AMD_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Records_in_DS1"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2B_AMD_SPH%Num_Records_in_DS1= &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L2B_AMD_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Records_in_DS2"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2B_AMD_SPH%Num_Records_in_DS2= &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L2B_AMD_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Avail_L1B_Obs"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2B_AMD_SPH%Num_Avail_L1B_Obs= &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L2B_AMD_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Missing_L1B_Obs"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2B_AMD_SPH%Num_Missing_L1B_Obs= &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L2B_AMD_SPH_from_xml_tree")

    ! Spare_2

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Input_Files"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2B_AMD_SPH%Num_Input_Files= &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2B_AMD_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Files_Predict_Orbit"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2B_AMD_SPH%Num_Files_Predict_Orbit= &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2B_AMD_SPH_from_xml_tree")

  end subroutine Get_L2B_AMD_SPH_from_xml_tree

    !  #]
! Remark: the DSD structures are not yet filled inside the next routine
! this is only possible after writing the binary data blocks, since
! only then the size of the Ds/Dsr is known
  subroutine Fill_L2B_AMD_SPH_items(L2B_AMD_SPH, error_flag, &
                                    SPH_Descriptor, Ref_NWP_Suite, &
                                    Fcst_Initial_Time, &
                                    Model_Timestep, Model_Grid_Type, &
                                    Model_Resol_Par1, Model_Resol_Par2, &
                                    Num_of_Model_Layers, &
                                    Num_Records_in_DS1, Num_Records_in_DS2, &
                                    Num_Avail_L1B_Obs, Num_Missing_L1B_Obs, &
                                    Num_Computed_Locations)
    !  #[

    TYPE(L2B_AMD_SPH_type),       pointer     :: L2B_AMD_SPH     ! filled on output
    integer,                      intent(out) :: error_flag      ! result
    character(len=*), intent(in), optional :: SPH_Descriptor
    character(len=*), intent(in), optional :: Ref_NWP_Suite
    type(DateTimeType),intent(in),optional :: Fcst_Initial_Time
    integer(IntAl),   intent(in), optional :: Model_Timestep
    character(len=*), intent(in), optional :: Model_Grid_Type
    integer(IntAs),   intent(in), optional :: Model_Resol_Par1
    integer(IntAs),   intent(in), optional :: Model_Resol_Par2
    integer(IntAus),  intent(in), optional :: Num_of_Model_Layers
    integer(IntAl),   intent(in), optional :: Num_Records_in_DS1
    integer(IntAl),   intent(in), optional :: Num_Records_in_DS2
    integer(IntAl),   intent(in), optional :: Num_Avail_L1B_Obs
    integer(IntAl),   intent(in), optional :: Num_Missing_L1B_Obs
    integer(IntAl),   intent(in), optional :: Num_Computed_Locations
    
    ! local variables

    error_flag = no_error
    IF (.not. associated(L2B_AMD_SPH)) THEN
       print *,"ERROR in Copy_L2B_AMD_SPH_to_xml_tree():"
       print *,"The pointer to this L2B_AMD_SPH seems not associated !!"
       error_flag = error_programming
       return
    END IF

    IF (present(SPH_Descriptor)) THEN
       call  check_string_length(SPH_Descriptor,"Sph_Descriptor",&
            maxlen_SPH_Descriptor,"Fill_L2B_AMD_SPH_items") 
       L2B_AMD_SPH%SPH_Descriptor = SPH_Descriptor
    ENDIF
    
    IF (present(Ref_NWP_Suite)) THEN
       call  check_string_length(Ref_NWP_Suite,"Ref_NWP_Suite",&
            maxlen_Ref_NWP_Suite,"Fill_L2B_AMD_SPH_items") 
       L2B_AMD_SPH%Ref_NWP_Suite = Ref_NWP_Suite
    ENDIF
    
    IF (present(Fcst_Initial_Time)) THEN
       L2B_AMD_SPH%Fcst_Initial_Time = Construct_KVT_DateTime(Fcst_Initial_Time)
     ENDIF
    
    IF (present(Model_Timestep)) THEN
       call SetIntAl(Model_Timestep,L2B_AMD_SPH%Model_Timestep,error_flag)
       IF (error_flag .ne. no_error) return
    ENDIF
    
    IF (present(Model_Grid_Type)) THEN
       call  check_string_length(Model_Grid_Type,"Model_Grid_Type",&
            maxlen_Model_Grid_Type,"Fill_L2B_AMD_SPH_items")
       L2B_AMD_SPH%Model_Grid_Type = Model_Grid_Type
    ENDIF
     
    IF (present(Model_Resol_Par1)) THEN
       call SetIntAs(Model_Resol_Par1,L2B_AMD_SPH%Model_Resol_Par1,error_flag)
       IF (error_flag .ne. no_error) return
    ENDIF
     
    IF (present(Model_Resol_Par2)) THEN
       call SetIntAs(Model_Resol_Par2,L2B_AMD_SPH%Model_Resol_Par2,error_flag)
       IF (error_flag .ne. no_error) return
    ENDIF
     
    IF (present(Num_of_Model_Layers)) THEN
       IF (Num_of_Model_Layers .le. 0 .or. Num_of_Model_Layers .gt. 9999) THEN
          print *,"WARNING in Fill_L2B_AMD_SPH_items():"
          print *,"unlikely value for Num_of_Model_Layers, are you sure this is correct?"
          print *,"Num_of_Model_Layers = ",Num_of_Model_Layers
       ENDIF
       call SetIntAus(Num_of_Model_Layers,L2B_AMD_SPH%Num_of_Model_Layers,error_flag)
       IF (error_flag .ne. no_error) return
    ENDIF
     
    IF (present(Num_Records_in_DS1)) THEN
       call SetIntAl(Num_Records_in_DS1,L2B_AMD_SPH%Num_Records_in_DS1,error_flag)
       IF (error_flag .ne. no_error) return
    ENDIF
     
    IF (present(Num_Records_in_DS2)) THEN
       call SetIntAl(Num_Records_in_DS2,L2B_AMD_SPH%Num_Records_in_DS2,error_flag)
       IF (error_flag .ne. no_error) return
    ENDIF

    IF (present(Num_Avail_L1B_Obs)) THEN
       call SetIntAl(Num_Avail_L1B_Obs,L2B_AMD_SPH%Num_Avail_L1B_Obs,error_flag)
       IF (error_flag .ne. no_error) return
    ENDIF
     
    IF (present(Num_Missing_L1B_Obs)) THEN
       call SetIntAl(Num_Missing_L1B_Obs,L2B_AMD_SPH%Num_Missing_L1B_Obs,error_flag)
       IF (error_flag .ne. no_error) return
    ENDIF
     
    IF (present(Num_Computed_Locations)) THEN
       call SetIntAl(Num_Computed_Locations,L2B_AMD_SPH%Num_Computed_Locations,error_flag)
       IF (error_flag .ne. no_error) return
    ENDIF
     
  end subroutine Fill_L2B_AMD_SPH_items

    !  #]
  subroutine Copy_L2B_AMD_SPH_to_xml_tree(L2B_AMD_SPH, path_in, &
                                          xml_tree, error_flag)
    !  #[

    TYPE(L2B_AMD_SPH_type), pointer    :: L2B_AMD_SPH ! input
    character(len=*),       intent(in) :: path_in     ! input
    TYPE(xml_node_type),    pointer    :: xml_tree   ! points to updated struct
    integer,                intent(out):: error_flag ! output

    ! local variables
!    character(len=maxlen_Num_Layers) :: Num_Layers_str
!    character(len=maxlen_Num_BRC)    :: Num_BRC_str
!    character(len=maxlen_Num_DSD)    :: Num_DSD_str

    character(len=max_str_len) :: path
    character(len=max_str_len) :: path_dsd
    character(len=max_str_len) :: text
    integer                    :: p, dsd_number_attached, dsd_number
    TYPE(DSD_Type), pointer    :: dsd_ptr

    error_flag = no_error

    ! add the <Specific_Product_Header></Specific_Product_Header> tags
    path = trim(path_in)//"/"//ee_vh_name
    call xml_tree_add_child(xml_tree,path,ee_sph_name,error_flag)
    IF (error_flag .ne. no_error) return

    ! fill all items inside the L2B_AMD_SPH
    call xml_add_SPH_item(xml_tree, path_in, "Sph_Descriptor", &
                          L2B_AMD_SPH%SPH_Descriptor,&
                          error_flag)
    IF (error_flag .ne. no_error) return

    ! Spare_1
    call xml_add_SPH_item(xml_tree, path_in, "Spare_1",newline,error_flag)
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Ref_NWP_Suite", &
                          L2B_AMD_SPH%Ref_NWP_Suite,&
                          error_flag)
    IF (error_flag .ne. no_error) return

    text = Convert_KVT_to_XML_DateTime(L2B_AMD_SPH%Fcst_Initial_Time)
    call xml_add_SPH_item(xml_tree, path_in, "Fcst_Initial_Time", &
                          text,error_flag)
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Model_Timestep", &
                          L2B_AMD_SPH%Model_Timestep,&
                          error_flag)
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Model_Grid_Type", &
                          L2B_AMD_SPH%Model_Grid_Type,&
                          error_flag)
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Model_Resol_Par1", &
                          L2B_AMD_SPH%Model_Resol_Par1,&
                          error_flag)
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Model_Resol_Par2", &
                          L2B_AMD_SPH%Model_Resol_Par2,&
                          error_flag)
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_of_Model_Layers", &
                          Remove_Plus_Sign(L2B_AMD_SPH%Num_of_Model_Layers),&
                          error_flag)
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Records_in_DS1", &
                          L2B_AMD_SPH%Num_Records_in_DS1,&
                          error_flag)
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Records_in_DS2", &
                          L2B_AMD_SPH%Num_Records_in_DS2,&
                          error_flag)
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Avail_L1B_Obs", &
                          L2B_AMD_SPH%Num_Avail_L1B_Obs,&
                          error_flag)
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Missing_L1B_Obs", &
                          L2B_AMD_SPH%Num_Missing_L1B_Obs,&
                          error_flag)
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Computed_Locations", &
                          L2B_AMD_SPH%Num_Computed_Locations,&
                          error_flag)
    IF (error_flag .ne. no_error) return

    ! Spare_2
    call xml_add_SPH_item(xml_tree, path_in, "Spare_2",newline,error_flag)
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Input_Files", &
                          Remove_Plus_Sign(L2B_AMD_SPH%Num_Input_Files),&
                          error_flag)
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Files_Predict_Orbit", &
                          Remove_Plus_Sign(L2B_AMD_SPH%Num_Files_Predict_Orbit),&
                          error_flag)
    IF (error_flag .ne. no_error) return

    ! Spare_3
    call xml_add_SPH_item(xml_tree, path_in, "Spare_3",newline,error_flag)
    IF (error_flag .ne. no_error) return

    ! add the <List_of_DSDs></List_of_DSDs> tags
    path = trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name
    call xml_tree_add_child(xml_tree,path,"List_of_Dsds",error_flag)
    IF (error_flag .ne. no_error) return

    path_dsd = trim(path)//"/List_of_Dsds"
    ! TODO: logic to determine Number of DSDs which should
    !       be sum of num of attached datasets plus num_input_files
    call xml_tree_add_and_set_attribute(xml_tree,path_dsd,"count","14",&
                                        error_flag)

    call Add_DSD_to_xml_SPH(xml_tree, path_in, &
                            L2B_AMD_SPH%GeoADS1_DSD, 1, error_flag)
    IF (error_flag .ne. no_error) return

    call Add_DSD_to_xml_SPH(xml_tree, path_in, &
                            L2B_AMD_SPH%GeoADS2_DSD, 2, error_flag)
    IF (error_flag .ne. no_error) return

    call Add_DSD_to_xml_SPH(xml_tree, path_in, &
                            L2B_AMD_SPH%MDS1_DSD, 3, error_flag)
    IF (error_flag .ne. no_error) return

    call Add_DSD_to_xml_SPH(xml_tree, path_in, &
                            L2B_AMD_SPH%MDS2_DSD, 4, error_flag)
    IF (error_flag .ne. no_error) return

    dsd_number_attached = 4 
    ! Assuming GeoADS1, GeoADS2, MetMDS1, MetMDS2 all present

    IF (associated(L2B_AMD_SPH%Input_DSD)) THEN
       DO p=1,size(L2B_AMD_SPH%Input_DSD)
          dsd_number = dsd_number_attached + p
          dsd_ptr => L2B_AMD_SPH%Input_DSD(p)
          call Add_DSD_to_xml_SPH(xml_tree, path_in, &
                                  dsd_ptr, dsd_number, error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    END IF

  end subroutine Copy_L2B_AMD_SPH_to_xml_tree
    !  #]
  !---------------------------------
  subroutine Write_L2B_AMD_SPH_to_KVT(L2B_AMD_SPH, fileunit, error_flag)
    !  #[ Writes the AMD Specific Product Header (KVT values) to file
    !     09-02-2006   P. Poli  Original code. From TN2.1 v2.0

    TYPE(L2B_AMD_SPH_type), pointer     :: L2B_AMD_SPH
    integer,                intent(in)  :: fileunit
    integer,                intent(out) :: error_flag

    ! local variables
    character(len=64) :: varname
    character(len=64) :: value
    character(len=64) :: phys_unit
    integer           :: max_value_length,begpos,endpos
    integer(IntAl)    :: size_SPH_calc
    integer           :: p
    TYPE(DSD_Type), pointer :: dsd_ptr
    !---------------------------------
    ! Writing the AMD SPH KVT elements
    !---------------------------------

    error_flag = no_error
    IF (.not. associated(L2B_AMD_SPH)) THEN
       print *,"ERROR in Write_L2B_AMD_SPH_to_KVT():"
       print *,"The pointer to this SPH seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! save start position
    inquire(unit=fileunit,nextrec=begpos)
       
    !SPH_Descriptor
    varname="Sph_Descriptor"
    value=L2B_AMD_SPH%SPH_Descriptor
    max_value_length=maxlen_SPH_Descriptor
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag,&
                          use_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998

    ! Spare_1: this is a spare value without content
    max_value_length = 40
    call write_spare_kvt_string(fileunit,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    !Ref_NWP_Suite
    varname="Ref_NWP_Suite"
    value=L2B_AMD_SPH%Ref_NWP_Suite
    max_value_length=maxlen_Ref_NWP_Suite
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag,&
                          use_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998

    !Fcst_Initial_Time
    varname="Fcst_Initial_Time"
    value=L2B_AMD_SPH%Fcst_Initial_Time
    max_value_length=maxlen_DateTime_KVT
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag,&
                          use_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998

    !Model_Timestep
    varname="Model_Timestep"
    value=L2B_AMD_SPH%Model_Timestep
    max_value_length=maxlen_IntAl
    phys_unit="s"
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag,&
                          phys_unit=phys_unit)
    IF (error_flag .ne. no_error) goto 998

    !Model_Grid_Type
    varname="Model_Grid_Type"
    value=L2B_AMD_SPH%Model_Grid_Type
    max_value_length=maxlen_Model_Grid_Type
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    !Model_Resol_Par1
    varname="Model_Resol_Par1"
    value=L2B_AMD_SPH%Model_Resol_Par1
    max_value_length=maxlen_IntAs
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    !Model_Resol_Par2
    varname="Model_Resol_Par2"
    value=L2B_AMD_SPH%Model_Resol_Par2
    max_value_length=maxlen_IntAs
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    !Num_of_Model_Layers
    varname="Num_of_Model_Layers"
    value=L2B_AMD_SPH%Num_of_Model_Layers
    max_value_length=maxlen_IntAus
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    !Num_Records_in_DS1
    varname="Num_Records_in_DS1"
    value=L2B_AMD_SPH%Num_Records_in_DS1
    max_value_length=maxlen_IntAl
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    !Num_Records_in_DS2
    varname="Num_Records_in_DS2"
    value=L2B_AMD_SPH%Num_Records_in_DS2
    max_value_length=maxlen_IntAl
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    !Num_Avail_L1B_Obs
    varname="Num_Avail_L1B_Obs"
    value=L2B_AMD_SPH%Num_Avail_L1B_Obs
    max_value_length=maxlen_IntAl
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    !Num_Missing_L1B_Obs
    varname="Num_Missing_L1B_Obs"
    value=L2B_AMD_SPH%Num_Missing_L1B_Obs
    max_value_length=maxlen_IntAl
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    !Num_Computed_Locations
    varname="Num_Computed_Locations"
    value=L2B_AMD_SPH%Num_Computed_Locations
    max_value_length=maxlen_IntAl
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Spare_2: this is a spare value without content
    max_value_length = 40
    call write_spare_kvt_string(fileunit,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    !Num_Input_Files
    varname="Num_Input_Files"
    value=L2B_AMD_SPH%Num_Input_Files
    max_value_length=maxlen_IntAus
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    !Num_Files_Predict_Orbit
    varname="Num_Files_Predict_Orbit"
    value=L2B_AMD_SPH%Num_Files_Predict_Orbit
    max_value_length=maxlen_IntAus
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! Spare_3: this is a spare value without content
    max_value_length = 40
    call write_spare_kvt_string(fileunit,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Write GeoADS1_DSD
    call Write_DSD_to_KVT(L2B_AMD_SPH%GeoADS1_DSD,fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Write GeoADS2_DSD
    call Write_DSD_to_KVT(L2B_AMD_SPH%GeoADS2_DSD,fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Write MDS1_DSD
    call Write_DSD_to_KVT(L2B_AMD_SPH%MDS1_DSD,fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Write MDS2_DSD
    call Write_DSD_to_KVT(L2B_AMD_SPH%MDS2_DSD,fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Write Input_DSD(:)
    IF (associated(L2B_AMD_SPH%Input_DSD)) THEN
       DO p=1,size(L2B_AMD_SPH%Input_DSD)
          dsd_ptr => L2B_AMD_SPH%Input_DSD(p)
          call Write_DSD_to_KVT(dsd_ptr,fileunit,error_flag)
          IF (error_flag .ne. no_error) goto 998
       END DO
    END IF

    ! save end position
    inquire(unit=fileunit,nextrec=endpos)

    ! calculate AMD SPH Size
    size_SPH_calc = int(endpos-begpos,IntAl)

    ! compare size with GetBytesize result
    IF (size_SPH_calc .ne. GetByteSize_AMD_SPH_KVT(L2B_AMD_SPH)) THEN
       call logmsg(log_error,"ERROR in Write_L2B_AMD_SPH_to_KVT: "//&
            "SPH size calculated by GetByteSize_AMD_SPH_KVT does not "//&
            "match actual nr of bytes written to file !!!")
       error_flag = error_programming
       return
    ENDIF

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Write_L2B_AMD_SPH_to_KVT"
    return

  end subroutine Write_L2B_AMD_SPH_to_KVT
    !  #]
  subroutine Read_AMD_SPH_KVT(SPH, fileunit, error_flag)
    !  #[ read the specific product header (KVT values) from file

    !
    ! compare with the file:
    ! L1bP_Draft_1/aeolus/common/PFR/src/PFR_LevelOneBRead.cc
    ! lines 340-620    

    TYPE(L2B_AMD_SPH_type), pointer     :: SPH
    integer,                intent(in)  :: fileunit
    integer,                intent(out) :: error_flag

    ! local variables
    character(len=max_str_len) :: exp_name
    character(len=max_str_len) :: value
    character(len=max_str_len) :: exp_phys_unit
    integer                    :: exp_val_length !, N_Meas_Max

    !------------------------------
    ! reading the SPH KVT elements
    !------------------------------

    error_flag = no_error
    IF (.not. associated(SPH)) THEN
       print *,"ERROR in Read_AMD_SPH_KVT():"
       print *,"The pointer to this SPH seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! read the specific product header descriptor
    exp_val_length = len(SPH%Sph_Descriptor)
    exp_name       = "Sph_Descriptor"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expect_quotes=.true.)
    IF (error_flag .ne. no_error) return

!    call Fill_or_Compare_MPH_item(SPH%Sph_Descriptor, value, &
!         "SPH%Sph_Descriptor",do_CompareOnly, do_Compare)

    ! store the result
    SPH%Sph_Descriptor = value

    ! Spare_1: this is a spare value without content
    exp_val_length = 40
    call read_spare_kvt_string(fileunit,exp_val_length,error_flag)
    IF (error_flag .ne. no_error) return

    ! read the reference of the NWP suite that produced the met data
    exp_val_length = len(SPH%Ref_NWP_Suite)
    exp_name       = "Ref_NWP_Suite"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expect_quotes=.true.)
    IF (error_flag .ne. no_error) return

!    call Fill_or_Compare_MPH_item(SPH%Ref_NWP_Suite, value, &
!         "SPH%Ref_NWP_Suite",do_CompareOnly, do_Compare)

    ! store the result
    SPH%Ref_NWP_Suite = value

    ! read the forecast initial date/time
    exp_val_length = len(SPH%Fcst_Initial_Time)
    exp_name       = "Fcst_Initial_Time"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expect_quotes=.true.)
    IF (error_flag .ne. no_error) return

!    call Fill_or_Compare_MPH_item(SPH%Fcst_Initial_Time, value, &
!         "SPH%Fcst_Initial_Time",do_CompareOnly, do_Compare,&
!         CompareAsDate=.true.)

    ! store the result
    SPH%Fcst_Initial_Time = value

    ! read the model timestep
    exp_val_length = len(SPH%Model_Timestep)
    exp_name       = "Model_Timestep"
    exp_phys_unit  = "s"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.,&
         expected_phys_unit=exp_phys_unit)
    IF (error_flag .ne. no_error) return

!    call Fill_or_Compare_MPH_item(SPH%Model_Timestep, value, &
!         "SPH%Model_Timestep",do_CompareOnly, do_Compare,is_numerical=.true.)

    ! store the result
    SPH%Model_Timestep = value

    ! read the model grid type
    exp_val_length = len(SPH%Model_Grid_Type)
    exp_name       = "Model_Grid_Type"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length) !,expect_quotes=.true.)
    IF (error_flag .ne. no_error) return

!    call Fill_or_Compare_MPH_item(SPH%Model_Grid_Type, value, &
!         "SPH%Model_Grid_Type",do_CompareOnly, do_Compare,is_numerical=.true.)

    ! store the result
    SPH%Model_Grid_Type = value

    ! read the model resolution parameter#1
    exp_val_length = len(SPH%Model_Resol_Par1)
    exp_name       = "Model_Resol_Par1"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

!    call Fill_or_Compare_MPH_item(SPH%Model_Resol_Par1, value, &
!         "SPH%Model_Resol_Par1",do_CompareOnly, do_Compare,is_numerical=.true.)

    ! store the result
    SPH%Model_Resol_Par1 = value

    ! read the model resolution parameter#2
    exp_val_length = len(SPH%Model_Resol_Par2)
    exp_name       = "Model_Resol_Par2"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

!    call Fill_or_Compare_MPH_item(SPH%Model_Resol_Par2, value, &
!         "SPH%Model_Resol_Par2",do_CompareOnly, do_Compare,is_numerical=.true.)

    ! store the result
    SPH%Model_Resol_Par2 = value

    ! read the number of model layers
    exp_val_length = len(SPH%Num_of_Model_Layers)
    exp_name       = "Num_of_Model_Layers"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

!    call Fill_or_Compare_MPH_item(SPH%Num_of_Model_Layers, value, &
!         "SPH%Num_of_Model_Layers",do_CompareOnly, do_Compare,is_numerical=.true.)

    ! store the result
    SPH%Num_of_Model_Layers = value

    ! read the number of records in DS#1
    exp_val_length = len(SPH%Num_Records_in_DS1)
    exp_name       = "Num_Records_in_DS1"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

!    call Fill_or_Compare_MPH_item(SPH%Num_Records_in_DS1, value, &
!         "SPH%Num_Recods_in_DS1",do_CompareOnly, do_Compare,is_numerical=.true.)

    ! store the result
    SPH%Num_Records_in_DS1 = value

    ! read the number of records in DS#2
    exp_val_length = len(SPH%Num_Records_in_DS2)
    exp_name       = "Num_Records_in_DS2"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

!    call Fill_or_Compare_MPH_item(SPH%Num_Records_in_DS2, value, &
!         "SPH%Num_Recods_in_DS2",do_CompareOnly, do_Compare,is_numerical=.true.)

    ! store the result
    SPH%Num_Records_in_DS2 = value

    ! read the number of available L1B observations
    exp_val_length = len(SPH%Num_Avail_L1B_Obs)
    exp_name       = "Num_Avail_L1B_Obs"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

!    call Fill_or_Compare_MPH_item(SPH%Num_Avail_L1B_Obs, value, &
!         "SPH%Num_Avail_L1B_Obs",do_CompareOnly, do_Compare,is_numerical=.true.)

    ! store the result
    SPH%Num_Avail_L1B_Obs = value

    ! read the number of missing L1B observations
    exp_val_length = len(SPH%Num_Missing_L1B_Obs)
    exp_name       = "Num_Missing_L1B_Obs"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

!    call Fill_or_Compare_MPH_item(SPH%Num_Missing_L1B_Obs, value, &
!         "SPH%Num_Missing_L1B_Obs",do_CompareOnly, do_Compare,is_numerical=.true.)

    ! store the result
    SPH%Num_Missing_L1B_Obs = value

    ! read the number of computed locations (from orbit predictions)
    exp_val_length = len(SPH%Num_Computed_Locations)
    exp_name       = "Num_Computed_Locations"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

!    call Fill_or_Compare_MPH_item(SPH%Num_Computed_Locations, value, &
!         "SPH%Num_Computed_Locations",do_CompareOnly, do_Compare,is_numerical=.true.)

    ! store the result
    SPH%Num_Computed_Locations = value

    ! Spare_2: this is a spare value without content
    exp_val_length = 40
    call read_spare_kvt_string(fileunit,exp_val_length,error_flag)
    IF (error_flag .ne. no_error) return

    ! read the number of input files (L1B or Predicted Orbit)
    exp_val_length = len(SPH%Num_Input_Files)
    exp_name       = "Num_Input_Files"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

!    call Fill_or_Compare_MPH_item(SPH%Num_Input_Files, value, &
!         "SPH%Num_Input_Files",do_CompareOnly, do_Compare,is_numerical=.true.)

    ! store the result
    SPH%Num_Input_Files = value

    ! read the number of predicted orbit files
    exp_val_length = len(SPH%Num_Files_Predict_Orbit)
    exp_name       = "Num_Files_Predict_Orbit"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

!    call Fill_or_Compare_MPH_item(SPH%Num_Files_Predict_Orbit, value, &
!         "SPH%Num_Files_Predict_Orbit",do_CompareOnly, do_Compare,is_numerical=.true.)

    ! store the result
    SPH%Num_Files_Predict_Orbit = value

    ! Spare_3: this is a spare value without content
    exp_val_length = 40
    call read_spare_kvt_string(fileunit,exp_val_length,error_flag)
    IF (error_flag .ne. no_error) return

!   DSDs are read elsewhere

    !------------------------------
    ! convert some selected strings to their numerical value
    !------------------------------

!    read(SPH%N_MAX, "(i11)") N_Meas_Max
!    SPH%N_Meas_Max = N_Meas_Max

    !------------------------------
    ! done
    !------------------------------

  end subroutine Read_AMD_SPH_KVT

    !  #]
  !---------------------------------
  function GetByteSize_AMD_SPH_KVT(SPH) result(KVT_bytesize)
    !  #[
    TYPE(L2B_AMD_SPH_type), pointer :: SPH          ! input
    integer(IntAl)                  :: KVT_bytesize ! result

    ! KVT sizes have size(key)+size(=)+size(value)+size(terminator)
    KVT_bytesize = &
         !     len(key)                len(=)  len(value)      len(unit)   len(terminator)
         !                                                  or len(quotes)
         int(len("Sph_Descriptor")         + 1 + maxlen_SPH_Descriptor + 2 + 1 + & ! 18+28=46
             40+1+& ! spare_1                                                      ! 41
             len("Ref_NWP_Suite")          + 1 + maxlen_Ref_NWP_Suite + 2  + 1 + & ! 17+20=37
             len("Fcst_Initial_Time")      + 1 + maxlen_DateTime_KVT + 2   + 1 + & ! 21+27=48
             len("Model_Timestep")         + 1 + maxlen_IntAl + len("<s>") + 1 + & ! 19+11=30
             len("Model_Grid_Type")        + 1 + maxlen_Model_Grid_Type    + 1 + & ! 17+2=19 !no quotes
             len("Model_Resol_Par1")       + 1 + maxlen_IntAs              + 1 + & ! 18+6=24
             len("Model_Resol_Par2")       + 1 + maxlen_IntAs              + 1 + & ! 18+6=24
             len("Num_of_Model_Layers")    + 1 + maxlen_IntAus             + 1 + & ! 21+6=27
             len("Num_Records_in_DS1")     + 1 + maxlen_IntAl              + 1 + & ! 20+11=31
             len("Num_Records_in_DS2")     + 1 + maxlen_IntAl              + 1 + & ! 20+11=31
             len("Num_Avail_L1B_Obs")      + 1 + maxlen_IntAl              + 1 + & ! 19+11=30
             len("Num_Missing_L1B_Obs")    + 1 + maxlen_IntAl              + 1 + & ! 21+11=32
             len("Num_Computed_Locations") + 1 + maxlen_IntAl              + 1 + & ! 24+11=35
             40+1+& ! spare_2                                                      ! 41
             len("Num_Input_Files")        + 1 + maxlen_IntAus             + 1 + & ! 17+6=23
             len("Num_Files_Predict_Orbit")+ 1 + maxlen_IntAus             + 1 + & ! 25+6=31
             40+1+& ! spare_3                                                      ! 41
             4*GetByteSize_DSD_KVT(),IntAl) ! the 4 attached AMD DSD's             ! 4*288=1152

    if (associated(SPH%Input_DSD)) then
       KVT_bytesize = KVT_bytesize + & ! the Num_Input_Files reference AMD DSD's
            int(size(SPH%Input_DSD) * GetByteSize_DSD_KVT(),IntAl)                 ! 10*288=2880
    end if

  end function GetByteSize_AMD_SPH_KVT
    !  #]  !---------------------------------
  !---------------------------------
end module L2B_AMD_SPH
