module L2B_Geoloc_ADS
  !  #[ Documentation
  !
  !     a module to handle the L2B Geolocation and AOCS ADS dataset
  !     as defined in TN2.1, version 1.2 (06-09-2005), 
  !     sec. 5.2, p. 18-19, tables 12-14.
  !
  !     Written by J. de Kloe, KNMI 
  !     created:   20-01-2006
  ! 
  !     Modifications:
  !      20-Jan-2006  J. de Kloe   Original version
  !      12-May-2006  D. Tan       Null use of ParKind1 module
  !      14-Nov-2006  J. de Kloe   reintroduced the old AsciiWrite routines
  !                                originally written for the L2BtoTXT
  !                                branch of 15-Mar-2006
  !      12-Mar-2007  J. de Kloe   updated GetByteSize routine
  !      15-Mar-2007  J. de Kloe   removed use of SummaryVars
  !      23-Apr-2007  J. de Kloe   removed fillwithzeros code
  !      09-Jan-2007  J. de Kloe   changed all print routines in calls
  !                                to the corresponding AsciiWrite routine
  !      16-Jan-2008  J. de Kloe   phase out integer kind i_ 
  !
  !  #]
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE EE_CFI_Datatypes, only: IntAuc, &
       IntAs,  maxlen_IntAs,  SetIntAs, minval_IntAs, c_bytelen_IntAs,  &
       IntAl,  maxlen_IntAl,  SetIntAl, minval_IntAl, c_bytelen_IntAl,  &
       FAdoxy, maxlen_FAdo32, SetFAdo32,              c_bytelen_FAdoxy, &
       write_IntAs_to_bin_file,  read_IntAs_from_bin_file,  &
       write_IntAl_to_bin_file,  read_IntAl_from_bin_file,  &
       write_FAdoxy_to_bin_file, read_FAdoxy_from_bin_file, &
       DateTimeType,Construct_KVT_DateTime, InitDateTime,  &
       write_DateTime_to_bin_file, read_DateTime_from_bin_file, &
       c_bytelen_DateTime, SetFAdoFull_or_Missing, maxlen_DateTime_KVT, &
       write_SpareBytes_to_bin_file, read_SpareBytes_from_bin_file, &
       DateTimeMissing

  USE LunManager, only: fileunit_stdout
  USE Numerics, only: missing_indicator_real, missing_real
  USE ErrorHandler, only: no_error, error_programming, error_allocate, &
                          error_writing_file
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  !  #]
  !  #[ type definitions

  implicit none

  TYPE Geolocation_Bin_type
     !  #[
     ! see sec. 5.2, table 14 of TN2.1, vers. 1.2, 06-09-2005
     integer(IntAl) :: Latitude_Start        ! [10^-6 degN]
     integer(IntAl) :: Latitude_Stop         ! [10^-6 degN]
     integer(IntAl) :: Latitude_COG          ! [10^-6 degN]
     integer(IntAl) :: Longitude_Start       ! [10^-6 degE]
     integer(IntAl) :: Longitude_Stop        ! [10^-6 degE]
     integer(IntAl) :: Longitude_COG         ! [10^-6 degE]
     integer(IntAl) :: Altitude_Bottom       ! [m]
     integer(IntAl) :: Altitude_Top          ! [m]
     integer(IntAl) :: Altitude_COG          ! [m]
     real(FAdoxy)   :: LOS_Azimuth           ! [deg]
     real(FAdoxy)   :: LOS_Elevation         ! [deg]
     real(FAdoxy)   :: LOS_Sat_Velocity      ! [m/s]
     !                 1234567890123456789012345678901
  END TYPE Geolocation_Bin_type
     !  #]

  TYPE Geolocation_type
     !  #[
     ! Start_DateTime not added since it seems not usefull, and adds nothing
     ! new compared to the Start_of_Obs_DateTime in the L2B_Geoloc_ADS_type
     ! below.
     !
     ! see sec. 5.2, table 13 of TN2.1, vers. 1.2, 06-09-2005
     ! this profile is an array of size Num_Rangegates
     TYPE(Geolocation_Bin_type), dimension(:), pointer :: Height_Bin
     !
     ! for these numbers we will take the location of measurement
     ! that has maximum altitude, within the set of used measurements
     integer(IntAl) :: Lat_of_DEM_Intersection ! [10^-6 degE]
     integer(IntAl) :: Lon_of_DEM_Intersection ! [10^-6 degE]
     integer(IntAl) :: Alt_of_DEM_Intersection ! [m]
  END TYPE Geolocation_type
     !  #]

  TYPE L2B_Geoloc_ADS_type 
     !  #[
     ! see sec. 5.2, table 12 of TN2.1, vers. 1.2, 06-09-2005
     type(DateTimeType)  :: Start_of_Obs_DateTime    ! UTC
!     integer(IntAs)      :: N_Obs_Mie_Max      ! max nr of Obs. in this product
!     integer(IntAs)      :: N_Obs_Rayleigh_Max ! dito
     integer(IntAs)      :: N_Obs_Mie_Actual   ! actual nr of Obs. for this BRC
     integer(IntAs)      :: N_Obs_Rayleigh_Actual ! dito
! Poli: following telecon of 25-01-2006, it is decided that N_Obs_*_Actual
! refers to the number of Mie and Rayleigh profiles with meaningful values.
! Each BRC contains M_Mie and M_Rayleigh profiles (these two numbers are fixed
! for a given file), but in some BRCs there might be less than M_Mie profiles;
! in such case N_Obs_Mie_Actual is less than M_Mie. Same story for Rayleigh.
!   - when initialising, N_Obs_Mie_Actual is set to zero. The structure
!     is dimensioned using M_Mie.
!   - after each BRC is processed, a call to Handle_L2BHeaders will set
!     N_Obs_Mie_Actual to the proper value
!   - when writing, M_Mie profiles are written. The number N_Obs_Mie_Actual
!     is saved so that the user will know when reading the BRC that
!     only the first N_Obs_Mie_Actual profiles are meaningful.
!
!   - when reading, the structure should be dimensioned first to M_Mie.
!  Same story for Rayleigh.
!
!     ! Remark: the nr of generated observations may vary from one BRC to 
!     ! the next but it was agreed that for one product file the maximum
!     ! N_Obs_max should be used for dimensioning the datastructures 
!     ! that are written to the output file:
!     !         N_Obs_max = max(N_Obs_Actual(first_BRC:last_BRC))
!     ! (see email of David Tan, 20051012, 19.44 utc)
!     ! So: -when initialising N_Obs_Max is used, N_Obs_Actual is set to 0
!     !     -when filling the datastructure N_Obs_Actual is updated
!     !      everytime an observation is generated
!     !     -when printing only N_Obs_Actual is used
!     !     -when writing to file N_Obs_Actual is used, 
!     !      and also all not used Observations are written to file
!     !
!     ! an array for the geolocation for all Mie observation profiles
!     ! this profile is an array of size N_Obs_Mie_Max
     TYPE(Geolocation_type), dimension(:), pointer :: Mie_Geolocation
     !
     ! an array for the geolocation for all Rayleigh observation profiles
     ! this profile is an array of size N_Obs_Rayleigh_Max
     TYPE(Geolocation_type), dimension(:), pointer :: Rayleigh_Geolocation
     !
     ! the difference between the geoid and WGS84 surface in [m]
     integer(IntAl) :: WGS84_to_Geoid_Altitude
     !
  END TYPE L2B_Geoloc_ADS_type
     !  #]

  ! the default verbosity setting is Err (as silent as possible)
  integer, private, save :: LocalVerbosity = VerbosityError

  !  #]
contains
  !---------------------------------------------------------------------
  subroutine Init_L2B_Geoloc_ADS_module(ChoosenVerbosity)
    !  #[
    integer, intent(in), optional :: ChoosenVerbosity

    IF (present(ChoosenVerbosity)) LocalVerbosity = ChoosenVerbosity

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"L2B_Geoloc_ADS module has been initialised"
       call print_verbosity_setting(LocalVerbosity,ModuleName="L2B_Geoloc_ADS")
    END IF

  end subroutine Init_L2B_Geoloc_ADS_module
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_Geoloc_Height_Bin(Geoloc_Height_Bin)
    !  #[
    TYPE(Geolocation_Bin_type), intent(out) :: Geoloc_Height_Bin

    ! I am using minval_IntAl as missing for the moment
    Geoloc_Height_Bin%Latitude_Start   = minval_IntAl
              ! to be replaced by: missing_indicator_IntAl=maxval_IntAl
    Geoloc_Height_Bin%Latitude_Stop    = minval_IntAl
              ! to be replaced by: missing_indicator_IntAl=maxval_IntAl
    Geoloc_Height_Bin%Latitude_COG     = minval_IntAl
              ! to be replaced by: missing_indicator_IntAl=maxval_IntAl

    Geoloc_Height_Bin%Longitude_Start  = minval_IntAl
              ! to be replaced by: missing_indicator_IntAl=maxval_IntAl
    Geoloc_Height_Bin%Longitude_Stop   = minval_IntAl
              ! to be replaced by: missing_indicator_IntAl=maxval_IntAl
    Geoloc_Height_Bin%Longitude_COG    = minval_IntAl
              ! to be replaced by: missing_indicator_IntAl=maxval_IntAl

    Geoloc_Height_Bin%Altitude_Bottom  = minval_IntAl
              ! to be replaced by: missing_indicator_IntAl=maxval_IntAl
    Geoloc_Height_Bin%Altitude_Top     = minval_IntAl
              ! to be replaced by: missing_indicator_IntAl=maxval_IntAl
    Geoloc_Height_Bin%Altitude_COG     = minval_IntAl
              ! to be replaced by: missing_indicator_IntAl=maxval_IntAl

    Geoloc_Height_Bin%LOS_Azimuth      = missing_indicator_real
              ! to be replaced by: missing_indicator_FAdoxy
    Geoloc_Height_Bin%LOS_Elevation    = missing_indicator_real
              ! to be replaced by: missing_indicator_FAdoxy
    Geoloc_Height_Bin%LOS_Sat_Velocity = missing_indicator_real
              ! to be replaced by: missing_indicator_FAdoxy

  end subroutine Init_Geoloc_Height_Bin
    !  #]
  subroutine Print_Geoloc_Height_Bin(Geoloc_Height_Bin, prefix, error_flag)
    !  #[
    TYPE(Geolocation_Bin_type), intent(in)  :: Geoloc_Height_Bin
    character(len=*),           intent(in)  :: prefix            ! input
    integer,                    intent(out) :: error_flag

    call AsciiWrite_Geoloc_Height_Bin(Geoloc_Height_Bin, fileunit_stdout, &
                                      prefix, error_flag)

  end subroutine Print_Geoloc_Height_Bin
    !  #]
  subroutine AsciiWrite_Geoloc_Height_Bin(Geoloc_Height_Bin, fileunit, &
                                          prefix, error_flag)
    !  #[
    TYPE(Geolocation_Bin_type), intent(in)  :: Geoloc_Height_Bin ! input
    integer,                    intent(in)  :: fileunit          ! input
    character(len=*),           intent(in)  :: prefix            ! input
    integer,                    intent(out) :: error_flag        ! output

    ! local variables
    character(len=maxlen_IntAl)  :: str_IntAl
    character(len=25)            :: str_FAdoxy

    error_flag = no_error

    ! latitude handling
    !TODO: replace this if then by a call to SetIntAl_or_missing()
    IF (Geoloc_Height_Bin%Latitude_Start .eq. minval_IntAl) THEN
       str_IntAl = "[missing]"
    ELSE
       call SetIntAl(Geoloc_Height_Bin%Latitude_Start,str_IntAl,error_flag)
       if (error_flag .ne. no_error) return
    END IF
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Latitude_Start   = "//trim(str_IntAl)//" [10^-6 degN]"

    !TODO: replace this if then by a call to SetIntAl_or_missing()
    IF (Geoloc_Height_Bin%Latitude_Stop .eq. minval_IntAl) THEN
       str_IntAl = "[missing]"
    ELSE
       call SetIntAl(Geoloc_Height_Bin%Latitude_Stop,str_IntAl,error_flag)
       if (error_flag .ne. no_error) return
    END IF
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Latitude_Stop    = "//trim(str_IntAl)//" [10^-6 degN]"

    !TODO: replace this if then by a call to SetIntAl_or_missing()
    IF (Geoloc_Height_Bin%Latitude_COG .eq. minval_IntAl) THEN
       str_IntAl = "[missing]"
    ELSE
       call SetIntAl(Geoloc_Height_Bin%Latitude_COG,str_IntAl,error_flag)
       if (error_flag .ne. no_error) return
    END IF
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Latitude_COG     = "//trim(str_IntAl)//" [10^-6 degN]"

    ! longitude handling
    !TODO: replace this if then by a call to SetIntAl_or_missing()
    IF (Geoloc_Height_Bin%Longitude_Start .eq. minval_IntAl) THEN
       str_IntAl = "[missing]"
    ELSE
       call SetIntAl(Geoloc_Height_Bin%Longitude_Start,str_IntAl,error_flag)
       if (error_flag .ne. no_error) return
    END IF
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Longitude_Start  = "//trim(str_IntAl)//" [10^-6 degE]"

    !TODO: replace this if then by a call to SetIntAl_or_missing()
    IF (Geoloc_Height_Bin%Longitude_Stop .eq. minval_IntAl) THEN
       str_IntAl = "[missing]"
    ELSE
       call SetIntAl(Geoloc_Height_Bin%Longitude_Stop,str_IntAl,error_flag)
       if (error_flag .ne. no_error) return
    END IF
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Longitude_Stop   = "//trim(str_IntAl)//" [10^-6 degE]"

    !TODO: replace this if then by a call to SetIntAl_or_missing()
    IF (Geoloc_Height_Bin%Longitude_COG .eq. minval_IntAl) THEN
       str_IntAl = "[missing]"
    ELSE
       call SetIntAl(Geoloc_Height_Bin%Longitude_COG,str_IntAl,error_flag)
       if (error_flag .ne. no_error) return
    END IF
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Longitude_COG    = "//trim(str_IntAl)//" [10^-6 degE]"

    ! altitude handling
    !TODO: replace this if then by a call to SetIntAl_or_missing()
    IF (Geoloc_Height_Bin%Altitude_Bottom .eq. minval_IntAl) THEN
       str_IntAl = "[missing]"
    ELSE
       call SetIntAl(Geoloc_Height_Bin%Altitude_Bottom,str_IntAl,error_flag)
       if (error_flag .ne. no_error) return
    END IF
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Altitude_Bottom  = "//trim(str_IntAl)//" [m]"

    !TODO: replace this if then by a call to SetIntAl_or_missing()
    IF (Geoloc_Height_Bin%Altitude_Top .eq. minval_IntAl) THEN
       str_IntAl = "[missing]"
    ELSE
       call SetIntAl(Geoloc_Height_Bin%Altitude_Top,str_IntAl,error_flag)
       if (error_flag .ne. no_error) return
    END IF
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Altitude_Top     = "//trim(str_IntAl)//" [m]"

    !TODO: replace this if then by a call to SetIntAl_or_missing()
    IF (Geoloc_Height_Bin%Altitude_COG .eq. minval_IntAl) THEN
       str_IntAl = "[missing]"
    ELSE
       call SetIntAl(Geoloc_Height_Bin%Altitude_COG,str_IntAl,error_flag)
       if (error_flag .ne. no_error) return
    END IF
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Altitude_COG     = "//trim(str_IntAl)//" [m]"

    ! LOS handling
    write(fileunit,"(a)",err=999) "REALACC(12)"

    ! SetFAdoFull_or_Missing already takes care of filling in the string
    ! "[missing]" for missing numbers, so no extra code needed for that
    call SetFAdoFull_or_Missing(Geoloc_Height_Bin%LOS_Azimuth,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%LOS_Azimuth      = "//trim(str_FAdoxy)//" [deg]"

    call SetFAdoFull_or_Missing(Geoloc_Height_Bin%LOS_Elevation,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%LOS_Elevation    = "//trim(str_FAdoxy)//" [deg]"

    call SetFAdoFull_or_Missing(Geoloc_Height_Bin%LOS_Sat_Velocity,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%LOS_Sat_Velocity = "//trim(str_FAdoxy)//" [m/s]"

    write(fileunit,"(a)") "ENDREALACC"

    return

999 print *,"ERROR in AsciiWrite_Geoloc_Height_Bin"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Geoloc_Height_Bin
    !  #]
  function GetByteSize_Geoloc_Height_Bin() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result

    c_bytesize = 9*c_bytelen_IntAl + &
                 3*c_bytelen_FAdoxy    ! 12 items in total

    ! typical sizes
    ! memory size = 9*4+3*8 = 60 bytes
    ! file   size = 9*4+3*8 = 60 bytes

  end function GetByteSize_Geoloc_Height_Bin
    !  #]                                           
  subroutine Write_Geoloc_Height_Bin(fileunit, Geoloc_Height_Bin, error_flag)
    !  #[
    integer,                    intent(in)  :: fileunit
    TYPE(Geolocation_Bin_type), intent(in)  :: Geoloc_Height_Bin
    integer,                    intent(out) :: error_flag

    error_flag = no_error

    ! handle latitudes
    call write_IntAl_to_bin_file(fileunit, &
         Geoloc_Height_Bin%Latitude_Start, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAl_to_bin_file(fileunit, &
         Geoloc_Height_Bin%Latitude_Stop, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAl_to_bin_file(fileunit, &
         Geoloc_Height_Bin%Latitude_COG, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! handle longitudes
    call write_IntAl_to_bin_file(fileunit, &
         Geoloc_Height_Bin%Longitude_Start, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAl_to_bin_file(fileunit, &
         Geoloc_Height_Bin%Longitude_Stop, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAl_to_bin_file(fileunit, &
         Geoloc_Height_Bin%Longitude_COG, error_flag)
    
    ! handle altitudes
    IF (error_flag .ne. no_error) goto 998
    call write_IntAl_to_bin_file(fileunit, &
         Geoloc_Height_Bin%Altitude_Bottom, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAl_to_bin_file(fileunit, &
         Geoloc_Height_Bin%Altitude_Top, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAl_to_bin_file(fileunit, &
         Geoloc_Height_Bin%Altitude_COG, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! handle LOS
    call write_FAdoxy_to_bin_file(fileunit, &
         Geoloc_Height_Bin%LOS_Azimuth, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_FAdoxy_to_bin_file(fileunit, &
         Geoloc_Height_Bin%LOS_Elevation, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_FAdoxy_to_bin_file(fileunit, &
         Geoloc_Height_Bin%LOS_Sat_Velocity, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
!    IF (LocalVerbosity .ge. VerbosityReport) &
!         print *,"done with Write_Geoloc_Height_Bin"

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Write_Geoloc_Height_Bin"
    return

  end subroutine Write_Geoloc_Height_Bin
    !  #]
  subroutine Read_Geoloc_Height_Bin(fileunit, Geoloc_Height_Bin, error_flag)
    !  #[
    integer,                    intent(in)  :: fileunit
    TYPE(Geolocation_Bin_type), intent(out) :: Geoloc_Height_Bin
    integer,                    intent(out) :: error_flag
    
    error_flag = no_error

    ! handle latitudes
    call read_IntAl_from_bin_file(fileunit, &
         Geoloc_Height_Bin%Latitude_Start, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_IntAl_from_bin_file(fileunit, &
         Geoloc_Height_Bin%Latitude_Stop, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_IntAl_from_bin_file(fileunit, &
         Geoloc_Height_Bin%Latitude_COG, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! handle longitudes
    call read_IntAl_from_bin_file(fileunit, &
         Geoloc_Height_Bin%Longitude_Start, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_IntAl_from_bin_file(fileunit, &
         Geoloc_Height_Bin%Longitude_Stop, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_IntAl_from_bin_file(fileunit, &
         Geoloc_Height_Bin%Longitude_COG, error_flag)
    
    ! handle altitudes
    IF (error_flag .ne. no_error) goto 998
    call read_IntAl_from_bin_file(fileunit, &
         Geoloc_Height_Bin%Altitude_Bottom, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_IntAl_from_bin_file(fileunit, &
         Geoloc_Height_Bin%Altitude_Top, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_IntAl_from_bin_file(fileunit, &
         Geoloc_Height_Bin%Altitude_COG, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! handle LOS
    call read_FAdoxy_from_bin_file(fileunit, &
         Geoloc_Height_Bin%LOS_Azimuth, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_FAdoxy_from_bin_file(fileunit, &
         Geoloc_Height_Bin%LOS_Elevation, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_FAdoxy_from_bin_file(fileunit, &
         Geoloc_Height_Bin%LOS_Sat_Velocity, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
!    IF (LocalVerbosity .ge. VerbosityReport) &
!         print *,"done with Read_Geoloc_Height_Bin"

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Read_Geoloc_Height_Bin"
    return

  end subroutine Read_Geoloc_Height_Bin
    !  #]
  !---------------------------------------------------------------------
  !          1234567890123456789012345678901
  subroutine Init_Geolocation(Geolocation, N_Range_Gates0, error_flag)
    !  #[
    TYPE(Geolocation_type), intent(out) :: Geolocation   ! output
    integer(IntAs),         intent(in)  :: N_Range_Gates0 ! input
    integer,                intent(out) :: error_flag    ! output

    ! local variable
    integer :: r, AllocStatus

    error_flag = no_error
    Geolocation%Lat_of_DEM_Intersection = minval_IntAl
              ! to be replaced by: missing_indicator_IntAl=maxval_IntAl
    Geolocation%Lon_of_DEM_Intersection = minval_IntAl
              ! to be replaced by: missing_indicator_IntAl=maxval_IntAl
    Geolocation%Alt_of_DEM_Intersection = minval_IntAl
              ! to be replaced by: missing_indicator_IntAl=maxval_IntAl

    IF (N_Range_Gates0 .eq. 0) THEN
       ! init them to be empty
       nullify(Geolocation%Height_Bin)
    ELSE
       allocate(Geolocation%Height_Bin(N_Range_Gates0),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_Geolocation"
          error_flag = error_allocate
          return
       END IF
       
       DO r=1,N_Range_Gates0
          call Init_Geoloc_Height_Bin(Geolocation%Height_Bin(r))
       END DO
    END IF

     end subroutine Init_Geolocation
    !  #]
  subroutine Delete_Geolocation(Geolocation)
    !  #[
    TYPE(Geolocation_type), intent(inout) :: Geolocation

    IF (associated(Geolocation%Height_Bin)) deallocate(Geolocation%Height_Bin)

  end subroutine Delete_Geolocation
    !  #]
  subroutine Print_Geolocation(Geolocation, prefix, PrintAll, error_flag)
    !  #[
    TYPE(Geolocation_type), intent(in)  :: Geolocation ! input
    character(len=*),       intent(in)  :: prefix      ! input
    logical,                intent(in)  :: PrintAll    ! input
    integer,                intent(out) :: error_flag  ! output

    call AsciiWrite_Geolocation(Geolocation, fileunit_stdout, prefix, &
                                PrintAll, error_flag)

  end subroutine Print_Geolocation
    !  #]
  subroutine AsciiWrite_Geolocation(Geolocation, fileunit, prefix, &
                                    PrintAll, error_flag)
    !  #[
    TYPE(Geolocation_type), intent(in)  :: Geolocation ! input
    integer,                intent(in)  :: fileunit    ! input
    character(len=*),       intent(in)  :: prefix      ! input
    logical,                intent(in)  :: PrintAll    ! input
    integer,                intent(out) :: error_flag  ! output

    ! local variables
    integer                     :: r, s, step
    character(len=maxlen_IntAl) :: str_IntAl
    character(len=2)            :: str_cntr

    error_flag = no_error

    IF (associated(Geolocation%Height_Bin)) THEN
       ! get the sizes of the array
       s = size(Geolocation%Height_Bin)
       IF (PrintAll .or. (s .le. 1) ) THEN
          step=1
       ELSE
          step = s-1
          !print "(a,i2,a)","    ==> size(Geolocation%Height_Bin) = ",s,&
          !     " [printing first and last one]"
       END IF
       DO r=1,s,step
          write(str_cntr,"(i2.2)",err=998) r
          call AsciiWrite_Geoloc_Height_Bin(Geolocation%Height_Bin(r),&
               fileunit, trim(prefix)//"%Height_Bin("//trim(str_cntr)//")",&
               error_flag)
          if (error_flag .ne. no_error) return
       END DO
    ELSE
       write(fileunit,"(a)",err=999) &
            trim(prefix)//"%Height_Bin(:) [seems not associated]"
    END IF

    !TODO: replace this if then by a call to SetIntAl_or_missing()
    IF (Geolocation%Lat_of_DEM_Intersection .eq. minval_IntAl) THEN
       str_IntAl = "[missing]"
    ELSE
       call SetIntAl(Geolocation%Lat_of_DEM_Intersection,str_IntAl,error_flag)
       if (error_flag .ne. no_error) return
    END IF

    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Lat_of_DEM_Intersection  = "//&
         trim(str_IntAl)//" [10^-6 degN]"

    !TODO: replace this if then by a call to SetIntAl_or_missing()
    IF (Geolocation%Lon_of_DEM_Intersection .eq. minval_IntAl) THEN
       str_IntAl = "[missing]"
    ELSE
       call SetIntAl(Geolocation%Lon_of_DEM_Intersection,str_IntAl,error_flag)
       if (error_flag .ne. no_error) return
    END IF

    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Lon_of_DEM_Intersection  = "//&
         trim(str_IntAl)//" [10^-6 degE]"

    !TODO: replace this if then by a call to SetIntAl_or_missing()
    IF (Geolocation%Alt_of_DEM_Intersection .eq. minval_IntAl) THEN
       str_IntAl = "[missing]"
    ELSE
       call SetIntAl(Geolocation%Alt_of_DEM_Intersection,str_IntAl,error_flag)
       if (error_flag .ne. no_error) return
    END IF

    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Alt_of_DEM_Intersection  = "//trim(str_IntAl)//" [m]"

    return

998 print *,"conversion error in AsciiWrite_Geolocation:"
    print *,"could not convert the counter ",r," to a string .."
    error_flag = error_writing_file
    return

999 print *,"ERROR in AsciiWrite_Geolocation"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Geolocation
    !  #]
  function GetByteSize_Geolocation(Geolocation) result(c_bytesize)
    !  #[
    integer(IntAl)                     :: c_bytesize ! result
    TYPE(Geolocation_Type), intent(in) :: Geolocation

    c_bytesize = 3*c_bytelen_IntAl
    IF (associated(Geolocation%Height_Bin)) THEN
       c_bytesize = c_bytesize + &
            int(size(Geolocation%Height_Bin)*&
                GetByteSize_Geoloc_Height_Bin(),IntAl)
    END IF
    
    ! typical sizes
    ! memory size = 2  +24*2 = 50 bytes
    ! file   size = 2+1+24*2 = 51 bytes
    
  end function GetByteSize_Geolocation
    !  #]
  subroutine Write_Geolocation(fileunit, Geolocation, error_flag)
    !  #[
    integer,                intent(in)  :: fileunit
    TYPE(Geolocation_type), intent(in)  :: Geolocation
    integer,                intent(out) :: error_flag

    ! local variables
    integer :: r, s

    error_flag = no_error
    IF (.not.(associated(Geolocation%Height_Bin))) THEN
       print *,"ERROR in Write_Geolocation():"
       print *,"The pointer to the Geolocation%Height_Bin array"
       print *,"seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! get the size of the array
    s = size(Geolocation%Height_Bin)
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"Write_Geolocation: size(Geolocation%Height_Bin) =",s
    DO r=1,s
       call Write_Geoloc_Height_Bin(fileunit, &
            Geolocation%Height_Bin(r), error_flag)
       if (error_flag .ne. no_error) goto 998
    END DO
    call write_IntAl_to_bin_file(fileunit, &
         Geolocation%Lat_of_DEM_Intersection, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAl_to_bin_file(fileunit, &
         Geolocation%Lon_of_DEM_Intersection, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAl_to_bin_file(fileunit, &
         Geolocation%Alt_of_DEM_Intersection, error_flag)
    IF (error_flag .ne. no_error) goto 998

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"done with Write_Geolocation"
    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Write_Geolocation"
    return

  end subroutine Write_Geolocation
    !  #]
  subroutine Read_Geolocation(fileunit, Geolocation, error_flag)
    !  #[
    integer,                intent(in)   :: fileunit
    TYPE(Geolocation_type), intent(inout):: Geolocation
    integer,                intent(out)  :: error_flag

    ! local variables
    integer :: r, s

    error_flag = no_error
    IF (.not.(associated(Geolocation%Height_Bin))) THEN
       print *,"ERROR in Read_Geolocation():"
       print *,"The pointer to the Geolocation%Height_Bin array"
       print *,"seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! get the size of the array
    s = size(Geolocation%Height_Bin)
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"Read_Geolocation: size(Geolocation%Height_Bin) =",s
    DO r=1,s
       call Read_Geoloc_Height_Bin(fileunit, &
            Geolocation%Height_Bin(r), error_flag)
       if (error_flag .ne. no_error) goto 998
    END DO

    call read_IntAl_from_bin_file(fileunit, &
         Geolocation%Lat_of_DEM_Intersection, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_IntAl_from_bin_file(fileunit, &
         Geolocation%Lon_of_DEM_Intersection, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_IntAl_from_bin_file(fileunit, &
         Geolocation%Alt_of_DEM_Intersection, error_flag)
    IF (error_flag .ne. no_error) goto 998

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"done with Read_Geolocation"
    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Read_Geolocation"
    return

  end subroutine Read_Geolocation
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_L2B_Geoloc_ADS(Geoloc_ADS, & 
                                 N_Obs_Mie_Max0, N_Obs_Rayleigh_Max0, &
                                 N_Range_Gates0, error_flag)
    !  #[
    TYPE(L2B_Geoloc_ADS_type),  pointer      :: Geoloc_ADS      ! input/output
    integer(IntAs),             intent(in)  :: N_Obs_Mie_Max0      ! input
    integer(IntAs),             intent(in)  :: N_Obs_Rayleigh_Max0 ! input
    integer(IntAs),             intent(in)  :: N_Range_Gates0      ! input
    integer,                    intent(out) :: error_flag       ! output

    ! local variable
    integer :: p, AllocStatus

    error_flag = no_error

    IF (.not. associated(Geoloc_ADS)) THEN
       print *,"ERROR in Init_L2B_Geoloc_ADS: "
       print *,"pointer Geoloc_ADS seems not associated !!!"
       error_flag = error_programming
       return
    END IF
    
    call InitDateTime(Geoloc_ADS%Start_of_Obs_DateTime)

    ! these are filled during processing, so init them to zero
    Geoloc_ADS%N_Obs_Mie_Actual      = 0
    Geoloc_ADS%N_Obs_Rayleigh_Actual = 0

    IF (N_Obs_Mie_Max0 .eq. 0) THEN
       ! init it to be empty
       nullify(Geoloc_ADS%Mie_Geolocation)
    ELSE
       allocate(Geoloc_ADS%Mie_Geolocation(N_Obs_Mie_Max0), &
                STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L2B_Geoloc_ADS"
          error_flag = error_allocate
          return
       END IF
       DO p=1,N_Obs_Mie_Max0
          call Init_Geolocation(Geoloc_ADS%Mie_Geolocation(p),&
                                N_Range_Gates0, error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    END IF

    IF (N_Obs_Rayleigh_Max0 .eq. 0) THEN
       ! init it to be empty
       nullify(Geoloc_ADS%Rayleigh_Geolocation)
    ELSE
       allocate(Geoloc_ADS%Rayleigh_Geolocation(N_Obs_Rayleigh_Max0),&
                STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L2B_Geoloc_ADS"
          error_flag = error_allocate
          return
       END IF
       DO p=1,N_Obs_Rayleigh_Max0
          call Init_Geolocation(Geoloc_ADS%Rayleigh_Geolocation(p),&
                                N_Range_Gates0, error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    END IF

    Geoloc_ADS%WGS84_to_Geoid_Altitude = minval_IntAl
              ! to be replaced by: missing_indicator_IntAl=maxval_IntAl

  end subroutine Init_L2B_Geoloc_ADS
    !  #]
  subroutine Delete_L2B_Geoloc_ADS(Geoloc_ADS)
    !  #[
    TYPE(L2B_Geoloc_ADS_type), pointer :: Geoloc_ADS

    ! local variable
    integer :: p

    IF (.not. associated(Geoloc_ADS)) return
    
    IF (associated(Geoloc_ADS%Mie_Geolocation)) THEN
       DO p=1,size(Geoloc_ADS%Mie_Geolocation)
          call Delete_Geolocation(Geoloc_ADS%Mie_Geolocation(p))
       END DO
       deallocate(Geoloc_ADS%Mie_Geolocation)
    END IF

    IF (associated(Geoloc_ADS%Rayleigh_Geolocation)) THEN
       DO p=1,size(Geoloc_ADS%Rayleigh_Geolocation)
          call Delete_Geolocation(Geoloc_ADS%Rayleigh_Geolocation(p))
       END DO
       deallocate(Geoloc_ADS%Rayleigh_Geolocation)
    END IF

  end subroutine Delete_L2B_Geoloc_ADS
    !  #]
  subroutine Print_L2B_Geoloc_ADS(Geoloc_ADS,prefix,PrintAll,error_flag)
    !  #[
    TYPE(L2B_Geoloc_ADS_type), pointer     :: Geoloc_ADS
    character(len=*),          intent(in)  :: prefix
    logical,                   intent(in)  :: PrintAll
    integer,                   intent(out) :: error_flag       ! output

    call AsciiWrite_L2B_Geoloc_ADS(Geoloc_ADS, fileunit_stdout, prefix,&
                                   PrintAll, error_flag)

  end subroutine Print_L2B_Geoloc_ADS
    !  #]
  subroutine AsciiWrite_L2B_Geoloc_ADS(Geoloc_ADS, fileunit, prefix,&
                                       PrintAll, error_flag)
    !  #[
    TYPE(L2B_Geoloc_ADS_type), pointer     :: Geoloc_ADS
    integer,                   intent(in)  :: fileunit
    character(len=*),          intent(in)  :: prefix
    logical,                   intent(in)  :: PrintAll
    integer,                   intent(out) :: error_flag       ! output

    ! local variables
    integer                            :: p, s, step
    character(len=maxlen_DateTime_KVT) :: str_DateTime
    character(len=maxlen_IntAs)        :: str_IntAs
    character(len=maxlen_IntAl)        :: str_IntAl
    character(len=2)                   :: str_cntr

    error_flag = no_error

    IF (.not. associated(Geoloc_ADS)) THEN
       write(fileunit,"(a)",err=999) trim(prefix)//" [seems not associated]"
       return
    END IF

    ! WARNING: my version of the portland pgf90 compiler has a bug here.
    ! if you give the Construct_KVT_DateTime() result directly to the 
    ! write statement the executable program crashes with a segmentation
    ! fault or sometime a deallocation/not allocated error.....

    if (DateTimeMissing(Geoloc_ADS%Start_of_Obs_DateTime)) then
       str_DateTime = "[missing]"
    else
       str_DateTime = Construct_KVT_DateTime(Geoloc_ADS%Start_of_Obs_DateTime)
    end if
    ! print the date and time
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Start_of_Obs_DateTime = "//&
         trim(str_DateTime)
! so don't use the next line with pgf90 !
!         trim(Construct_KVT_DateTime(Geoloc_ADS%Start_of_Obs_DateTime))

    !TODO: replace this if then by a call to SetIntAS_or_missing()
    IF (Geoloc_ADS%N_Obs_Mie_Actual .eq. minval_IntAs) THEN
       str_IntAs = "[miss]"
    ELSE
       call SetIntAs(Geoloc_ADS%N_Obs_Mie_Actual,str_IntAs,error_flag)
       if (error_flag .ne. no_error) return
    END IF
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%N_Obs_Mie_Actual      = "//trim(str_IntAs)

    !TODO: replace this if then by a call to SetIntAS_or_missing()
    IF (Geoloc_ADS%N_Obs_Rayleigh_Actual .eq. minval_IntAs) THEN
       str_IntAs = "[miss]"
    ELSE
       call SetIntAs(Geoloc_ADS%N_Obs_Rayleigh_Actual,str_IntAs,error_flag)
       if (error_flag .ne. no_error) return
    END IF
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%N_Obs_Rayleigh_Actual = "//trim(str_IntAs)

    IF (associated(Geoloc_ADS%Mie_Geolocation)) THEN
       s = size(Geoloc_ADS%Mie_Geolocation)

       IF (PrintAll .or. (s .le. 1) ) THEN
          step=1
       ELSE
          step = s-1
          !write(fileunit,"(a,i2,a)",err=999) &
          !     " ==> N_Obs_Mie_Actual = ",s," [printing first and last one]"
       END IF

       DO p=1,s,step
          write(str_cntr,"(i2.2)",err=998) p
          call AsciiWrite_Geolocation(Geoloc_ADS%Mie_Geolocation(p), &
               fileunit, trim(prefix)//"%Mie_Geolocation("//&
               trim(str_cntr)//")",PrintAll, error_flag)
       END DO
    ELSE
       write(fileunit,"(a)",err=999) &
            trim(prefix)//"%Mie_Geolocation(:) [NOT ASSOCIATED]"
    END IF

    IF (associated(Geoloc_ADS%Rayleigh_Geolocation)) THEN
       s = size(Geoloc_ADS%Rayleigh_Geolocation)

       IF (PrintAll .or. (s .le. 1) ) THEN
          step=1
       ELSE
          step = s-1
          write(fileunit,"(a,i2,a)",err=999) &
               " ==> N_Obs_Rayleigh_Actual = ",s,&
               " [printing first and last one]"
       END IF

       DO p=1,s,step
          write(str_cntr,"(i2.2)",err=998) p
          call AsciiWrite_Geolocation(Geoloc_ADS%Rayleigh_Geolocation(p),&
               fileunit, trim(prefix)//"%Rayleigh_Geolocation("//&
               trim(str_cntr)//")", PrintAll, error_flag)
       END DO
    ELSE
       write(fileunit,"(a)",err=999) &
            trim(prefix)//"%Rayleigh_Geolocation(:) [NOT ASSOCIATED]"
    END IF

    !TODO: replace this if then by a call to SetIntAl_or_missing()
    IF (Geoloc_ADS%WGS84_to_Geoid_Altitude .eq. minval_IntAl) THEN
       str_IntAl = "[missing]"
    ELSE
       call SetIntAl(Geoloc_ADS%WGS84_to_Geoid_Altitude,str_IntAl,error_flag)
       if (error_flag .ne. no_error) return
    END IF
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%WGS84_to_Geoid_Altitude  = "//trim(str_IntAl)

    return

998 print *,"conversion error in AsciiWrite_L2B_Geoloc_ADS:"
    print *,"could not convert the counter ",p," to a string .."
    error_flag = error_writing_file
    return

999 print *,"ERROR in AsciiWrite_L2B_Geoloc_ADS:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2B_Geoloc_ADS
    !  #]
  function GetByteSize_L2B_Geoloc_ADS(Geoloc_ADS, N_Obs_Mie_to_write, &
                                      N_Obs_Rayleigh_to_write) result(c_bytesize)
    !  #[
    integer(IntAl)                        :: c_bytesize ! result
    TYPE(L2B_Geoloc_ADS_type), pointer    :: Geoloc_ADS
    integer(IntAuc),           intent(in) :: N_Obs_Mie_to_write 
    integer(IntAuc),           intent(in) :: N_Obs_Rayleigh_to_write
 
    ! local variable
    integer :: s

    c_bytesize = 0
    IF (.not. associated(Geoloc_ADS)) return

    c_bytesize =   c_bytelen_DateTime + &
                 2*c_bytelen_IntAs    + &
                   c_bytelen_IntAl
    IF (associated(Geoloc_ADS%Mie_Geolocation)) THEN
       s = size(Geoloc_ADS%Mie_Geolocation)
       IF (N_Obs_Mie_to_write .gt.s) THEN
          print *,"WARNING in GetByteSize_L2B_Geoloc_ADS()"
          print *,"invalid value for N_Obs_Mie_to_write: ",N_Obs_Mie_to_write
          print *,"max. possible value is: size(Geoloc_ADS%Mie_Geolocation) = ",s
          print *,"Get_ByteSize result will be useless...."
       END IF
       c_bytesize = c_bytesize + N_Obs_Mie_to_write*&
            GetByteSize_Geolocation(Geoloc_ADS%Mie_Geolocation(1))
    END IF
    IF (associated(Geoloc_ADS%Rayleigh_Geolocation)) THEN
       s = size(Geoloc_ADS%Rayleigh_Geolocation)
       IF (N_Obs_Rayleigh_to_write .gt. s) THEN
          print *,"WARNING in GetByteSize_L2B_Geoloc_ADS()"
          print *,"invalid value for N_Obs_Rayleigh_to_write: ",&
               N_Obs_Rayleigh_to_write
          print *,"max. possible value is: size(Geoloc_ADS%Mie_Geolocation) = ",s
          print *,"Get_ByteSize result will be useless...."
       END IF
       c_bytesize = c_bytesize + N_Obs_Rayleigh_to_write*&
            GetByteSize_Geolocation(Geoloc_ADS%Rayleigh_Geolocation(1))
    END IF
    
    ! typical sizes
    ! memory size = 2  +24*2 = 50 bytes
    ! file   size = 2+1+24*2 = 51 bytes
    
  end function GetByteSize_L2B_Geoloc_ADS
    !  #]  
  subroutine Write_L2B_Geoloc_ADS(fileunit,Geoloc_ADS, &
                                  N_Obs_Mie_to_write, &
                                  N_Obs_Rayleigh_to_write, error_flag)
    !  #[

    integer,                   intent(in)  :: fileunit                ! input
    TYPE(L2B_Geoloc_ADS_type), pointer     :: Geoloc_ADS              ! input
    integer(IntAuc),           intent(in)  :: N_Obs_Mie_to_write      ! input
    integer(IntAuc),           intent(in)  :: N_Obs_Rayleigh_to_write ! input
    integer,                   intent(out) :: error_flag   ! output

    ! local variables
    integer :: p, s

    !------------------------------
    ! writing the Geoloc ADS elements
    !------------------------------

    error_flag = no_error
    IF (.not. associated(Geoloc_ADS)) THEN
       print *,"ERROR in Write_L2B_Geoloc_ADS():"
       print *,"The pointer to this Geoloc_ADS seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! Start_of_Obs_DateTime
    call write_DateTime_to_bin_file(fileunit, &
         Geoloc_ADS%Start_of_Obs_DateTime, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_IntAs_to_bin_file(fileunit, &
         Geoloc_ADS%N_Obs_Mie_Actual,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_IntAs_to_bin_file(fileunit, &
         Geoloc_ADS%N_Obs_Rayleigh_Actual,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! remind that N_Obs_Mie_Max sets the maximum nr of profiles, and this
    ! should never be zero. Therefore the next pointer should always be
    ! associated after initialisation, otherwise we have a programming
    ! or usage error (i.e. the user should not be allowed to set 
    ! N_Obs_Mie_Max to zero).
    IF (associated(Geoloc_ADS%Mie_Geolocation)) THEN
       ! remind that N_Obs_Mie_to_write is allowed to be zero, in case the
       ! program could not construct any usefull Mie profiles.
       ! In that case s=0, and the loop below will never be executed
       s = N_Obs_Mie_to_write
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"Write_L2B_Geoloc_ADS: Geoloc_ADS%N_Obs_Mie_Actual =",&
            Geoloc_ADS%N_Obs_Mie_Actual
       DO p=1,s
          call Write_Geolocation(fileunit,&
               Geoloc_ADS%Mie_Geolocation(p), error_flag)
          IF (error_flag .ne. no_error) goto 998
       ENDDO
    ELSE
       print *,"ERROR in Write_L2B_Geoloc_ADS: "
       print *,"Geoloc_ADS%Mie_Geolocation seems not associated ..."
       error_flag = error_programming
       return
    END IF
    
    IF (associated(Geoloc_ADS%Rayleigh_Geolocation)) THEN
       s = N_Obs_Rayleigh_to_write
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"Write_L2B_Geoloc_ADS: Geoloc_ADS%N_Obs_Rayleigh_Actual =",&
            Geoloc_ADS%N_Obs_Rayleigh_Actual
       DO p=1,s
          call Write_Geolocation(fileunit,&
               Geoloc_ADS%Rayleigh_Geolocation(p), error_flag)
          IF (error_flag .ne. no_error) goto 998
       ENDDO
    ELSE
       print *,"ERROR in Write_L2B_Geoloc_ADS: "
       print *,"Geoloc_ADS%Rayleigh_Geolocation seems not associated ..."
       error_flag = error_programming
       return
    END IF
    
    call write_IntAl_to_bin_file(fileunit, &
         Geoloc_ADS%WGS84_to_Geoid_Altitude,error_flag)
    IF (error_flag .ne. no_error) goto 998

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"Write_L2B_Geoloc_ADS: done with Write_L2B_Geoloc_ADS"
    
    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Write_L2B_Geoloc_ADS"
    return

  end subroutine Write_L2B_Geoloc_ADS
    !  #]
  subroutine Read_L2B_Geoloc_ADS(fileunit,Geoloc_ADS,error_flag)
    !  #[
    integer,                   intent(in)  :: fileunit
    TYPE(L2B_Geoloc_ADS_type), pointer     :: Geoloc_ADS
    integer,                   intent(out) :: error_flag

    ! local variables
    integer                            :: p, s
    character(len=maxlen_DateTime_KVT) :: DateTimeString

    !------------------------------
    ! reading the Geoloc ADS elements
    !------------------------------

    error_flag = no_error
    IF (.not. associated(Geoloc_ADS)) THEN
       print *,"ERROR in Read_L2B_Geoloc_ADS():"
       print *,"The pointer to this Geoloc_ADS seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! Start_of_Obs_DateTime
    call read_DateTime_from_bin_file(fileunit,&
         Geoloc_ADS%Start_of_Obs_DateTime,error_flag)
    
    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"Read_L2B_Geoloc_ADS: Start_of_Obs_DateTime is: "
       DateTimeString = Construct_KVT_DateTime(Geoloc_ADS%Start_of_Obs_DateTime)
       print *,DateTimeString
    END IF
    
    call read_IntAs_from_bin_file(fileunit, &
         Geoloc_ADS%N_Obs_Mie_Actual,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call read_IntAs_from_bin_file(fileunit, &
         Geoloc_ADS%N_Obs_Rayleigh_Actual,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! for N_Obs=0 this pointer may be un-associated
    ! so we have to check for that before requesting its size
    IF (associated(Geoloc_ADS%Mie_Geolocation)) THEN
       s = size(Geoloc_ADS%Mie_Geolocation)
       
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"Read_L2B_Mie_MDS: Geoloc_ADS%N_Obs_Actual =",&
            Geoloc_ADS%N_Obs_Rayleigh_Actual
       
       DO p=1,s
          call Read_Geolocation(fileunit,Geoloc_ADS%Mie_Geolocation(p),&
               error_flag)
          IF (error_flag .ne. no_error) goto 998
       ENDDO
    END IF
    
    ! for N_Obs=0 this pointer may be un-associated
    ! so we have to check for that before requesting its size
    IF (associated(Geoloc_ADS%Rayleigh_Geolocation)) THEN
       s = size(Geoloc_ADS%Rayleigh_Geolocation)
       
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"Read_L2B_Rayleigh_MDS: "//&
            "Geoloc_ADS%N_Obs_Rayleigh_Actual =",&
            Geoloc_ADS%N_Obs_Rayleigh_Actual
       DO p=1,s
          call Read_Geolocation(fileunit,&
               Geoloc_ADS%Rayleigh_Geolocation(p), error_flag)
          IF (error_flag .ne. no_error) goto 998
       ENDDO
    END IF
    
    call read_IntAl_from_bin_file(fileunit, &
         Geoloc_ADS%WGS84_to_Geoid_Altitude,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"Read_L2B_Geoloc_ADS: done"

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Read_L2B_Geoloc_ADS"
    return

  end subroutine Read_L2B_Geoloc_ADS
    !  #]
  !---------------------------------------------------------------------
    subroutine Update_L2B_Geoloc_N_Obs_Actual(Geoloc_ADS, &
         N_Obs_Mie_Actual0, N_Obs_Rayleigh_Actual0, error_flag)
    !  #[
    !  30-01-2006 P. Poli   Original code
    TYPE(L2B_Geoloc_ADS_Type), pointer    :: Geoloc_ADS
    integer(IntAs),         intent(in)    :: N_Obs_Mie_Actual0
    integer(IntAs),         intent(in)    :: N_Obs_Rayleigh_Actual0
    integer,                intent(out)   :: error_flag

    error_flag = no_error
    IF (.not.(associated(Geoloc_ADS))) THEN
       print *,"ERROR in Update_L2B_Geoloc_N_Obs_Actual():"
       print *,"The pointer to this Geoloc_ADS seems not associated !!"
       error_flag = error_programming
       return
    END IF

    Geoloc_ADS%N_Obs_Mie_Actual = N_Obs_Mie_Actual0
    Geoloc_ADS%N_Obs_Rayleigh_Actual = N_Obs_Rayleigh_Actual0

    end subroutine Update_L2B_Geoloc_N_Obs_Actual
    !  #]
end module L2B_Geoloc_ADS
