module L2B_Mie_MDS
  !  #[ Documentation
  !
  !     a module to handle the L2B Mie Measurement dataset
  !     as defined in TN2.1, version 1.1 (14-2-2005), 
  !     sec...., p. ..., tables ...
  !
  !     Written by J. de Kloe, KNMI 
  !                p. Poli,    Meteo-France
  !     created:   10-11-2005
  ! 
  !     Modifications:
  !      15-Nov-2005 J. de Kloe Original version (Copy of the Rayleigh_MDS,
  !              find and replace to change all Rayleigh strings into Mie.)
  !      24-Nov-2005 P. Poli    Added Write_L2B_Mie_MDS
  !      25-Nov-2005 P. Poli    Added Write_L2B_Mie_Profile
  !      28-Nov-2005 P. Poli    Added Write L2B_Mie_Height_Bin
  !      28-Nov-2005 P. Poli    Mie MDS as in TN2.1 v1.2 06-09-2005
  !      30-Nov-2005 P. Poli    Added Update_L2B_Mie_SummaryVars
  !      19-Dec-2005 J. de Kloe small correction in USE statements
  !      06-Jan-2006 J. de Kloe adapted to seconds of type r8_
  !      10-Jan-2006 P. Poli    Added reading routines (NOT tested yet)
  !      11-Jan-2006 J. de Kloe introduced use of the new DateTime type
  !      26-Jan-2006 P. Poli    Removed N_Obs_Max from Mie_MDS
  !      12-May-2006 D. Tan     Null use of ParKind1 module
  !      23-Feb-2007 J. de Kloe reintroduced the old AsciiWrite routines
  !                             originally written for the L2BtoTXT
  !                             branch of 15-Mar-2006
  !      02-Mar-2007 J. de Kloe adapted to IODD change, and some cleaning up
  !      12-Mar-2007 J. de Kloe updated GetByteSize routine
  !      15-Mar-2007 J. de Kloe remove use of SummaryVars
  !      23-Apr-2007 J. de Kloe phase out fillwithzero code
  !      10-Oct-2007 J. de Kloe use AsciiWrite within Print routines
  !      10-Oct-2007 J. de Kloe replaced all print statements with logmsg calls
  !      09-Jan-2008 J. de Kloe added missing checks for DateTime in 
  !                             AsciiWrite_L2B_Mie_MDS
  !      16-Jan-2008 J. de Kloe phase out integer kind i_ 
  !
  !  #]
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE EE_CFI_Datatypes, only: &
       IntAuc, maxlen_IntAuc, SetIntAuc, c_bytelen_IntAuc, &
       IntAc,  &
       IntAs,  maxlen_IntAs,  SetIntAs,  c_bytelen_IntAs,  &
       IntAus, maxlen_IntAus, SetIntAus, c_bytelen_IntAus, &
       IntAul, maxlen_IntAul, SetIntAul, c_bytelen_IntAul, &
       IntAl, &
       c_bytelen_Bool,   &
       DateTimeType,Construct_KVT_DateTime, DateTimeMissing, &
       InitDateTime, maxlen_DateTime_KVT, c_bytelen_DateTime, &
       write_DateTime_to_bin_file, read_DateTime_from_bin_file, &
       write_IntAuc_to_bin_file, read_IntAuc_from_bin_file, &
       write_IntAs_to_bin_file,  read_IntAs_from_bin_file,  &
       write_IntAus_to_bin_file, read_IntAus_from_bin_file, &
       write_IntAul_to_bin_file, read_IntAul_from_bin_file, &
       write_IntAl_to_bin_file,  read_IntAl_from_bin_file,  &
       write_SpareBytes_to_bin_file,   read_SpareBytes_from_bin_file

  USE L2B_parameters, only: Obs_Type_undefined, Obs_Type_cloudy_returns, &
                            Obs_Type_clear_returns, GetObsTypeText
  USE Meas_Map, only: Meas_Map_type, Init_Meas_Map, &
                      AsciiWrite_Meas_Map, Print_Meas_Map, Delete_Meas_Map, &
                      Count_Nr_Points_On_Level, &
                      Write_Meas_Map, Read_Meas_Map, &
                      Get_ByteSize_Meas_Map
  USE LunManager, only: fileunit_stdout
  USE ErrorHandler, only: no_error, error_programming, error_allocate, &
                          error_writing_file
  USE DirectBinaryIO, only: write_logical1_to_bin_file, &
                            read_logical1_from_bin_file
  USE Logging, only: Init_Logging, logmsg, &
                     log_debug,log_info,log_progress,log_warn,log_error
  !  #]
  !  #[ type definitions

  implicit none

  TYPE Mie_Height_Bin_type
     !  #[
     ! see sec. 5.4, table 19 of TN2.1, vers. 1.2, 06-09-2005
     logical         :: Validity_Flag        ! .true. = valid = 1
     integer(IntAs)  :: Mie_Wind_Velocity    ! [cm/s]
     integer(IntAus) :: Mie_Error_Quantifier ! [cm/s]
     integer(IntAul) :: Integration_Length   ! [m]
  END TYPE Mie_Height_Bin_type
     !  #]

  TYPE Mie_Profile_type
     !  #[
     ! added by JdK 7-11-2005. We need this to remember if this was
     ! a cloud or no-cloud profile (or any other type we define)
     integer(IntAuc) ::Obs_Type
     !
     ! Spare 1 (36 bytes reserved for Obs_Type_String)
     !
     ! see sec. 5.4, table 18 of TN2.1, vers. 1.2, 06-09-2005
     ! this profile is an array of size Num_Rangegates
     TYPE(Mie_Height_Bin_type), dimension(:), pointer :: Height_Bin
  END TYPE Mie_Profile_type
     !  #]

  TYPE L2B_Mie_MDS_type 
     !  #[
     ! see sec. 5.4, table 15 of TN2.1, vers. 1.2, 06-09-2005
     type(DateTimeType)  :: Start_of_Obs_DateTime    ! UTC
     integer(IntAs)      :: N_Meas       ! nr of Measurements in this product
     integer(IntAs)      :: N_Obs_Mie_Actual ! actual nr of Obs. for this BRC
     ! Poli: following telecon of 25-01-2006, it is decided that N_Obs_*_Actual
     ! refers to the number of Mie and Rayleigh profiles with meaningful values.
     ! Each BRC contains M_Mie and M_Rayleigh profiles (these two numbers are fixed
     ! for a given file, and defined in the SPH), but in some BRCs there might be 
     ! less than M_Mie profiles;
     ! in such case N_Obs_Mie_Actual is less than M_Mie. Same story for Rayleigh.
     !   - when initialising, N_Obs_Mie_Actual is set to zero. The structure
     !     is dimensioned using M_Mie.
     !   - after a BRC is processed, a call to Construct_Mie/Rayleigh_Obs will set
     !     N_Obs_Mie_Actual to the proper value
     !   - before writing M_Mie may be adjusted to hold the maximum of the
     !     N_Obs_*_Actual values of all BRC's to prevent unnecessary writing
     !     of empty datasets.
     !     Remark: the nr of generated observations may vary from one BRC to 
     !     the next but it was agreed that for one product file the maximum
     !     N_Obs_max should be used for dimensioning the datastructures 
     !     that are written to the output file:
     !         N_Obs_max = max(N_Obs_Actual(first_BRC:last_BRC))
     !     (see email of David Tan, 20051012, 19.44 utc)
     !   - when writing, M_Mie profiles are written. The number N_Obs_Mie_Actual
     !     is saved so that the user will know when reading the BRC that
     !     only the first N_Obs_Mie_Actual profiles are meaningful.
     !   - when reading, the structure should be dimensioned first to M_Mie.
     !   
     ! So: -when initialising N_Obs_Max is used, N_Obs_Actual is set to 0
     !     -when filling the datastructure N_Obs_Actual is updated
     !      everytime an observation is generated
     !     -when printing only N_Obs_Actual is used
     !     -when writing to file N_Obs_Mie_to_write is used, 
     !      and also all not used Observations are written to file
     !      (these are filled with missing values)
     !
     integer(IntAs)      :: P ! total nr of laser pulses in one Meas.
     ! a datastructure to store which measurements were used for
     ! creating which observation (the size will be N_Meas*N_Range_Bins)
     ! and what weight was used
     TYPE(Meas_Map_type), pointer :: Map_of_L1B_Measurements_Used
     !
     ! this next array is not implemented. In stead we added the Weight element
     ! to the L1B_Meas_Map_Bin_type type which saves a lot of double coding
     !
     ! an array of size N of rows containing the map weights
     ! TYPE(L1B_Meas_Weight_row_type), dimension(:), &
     !       pointer :: Map_of_L1B_Measurements_Used
     !
     ! an array for N_Obs_Max profiles for this BRC
     TYPE(Mie_Profile_type), dimension(:), pointer :: Mie_Profile
     !
  END TYPE L2B_Mie_MDS_type
     !  #]

  !  #]
contains
  !---------------------------------------------------------------------
  subroutine Init_Mie_Height_Bin(Mie_Height_Bin)
    !  #[
    TYPE(Mie_Height_Bin_type), intent(out) :: Mie_Height_Bin

    Mie_Height_Bin%Validity_Flag        =.false.
    Mie_Height_Bin%Mie_Wind_Velocity    = 0
    Mie_Height_Bin%Mie_Error_Quantifier = 0
    Mie_Height_Bin%Integration_Length   = 0

  end subroutine Init_Mie_Height_Bin
    !  #]
  subroutine Print_Mie_Height_Bin(Mie_Height_Bin, prefix, error_flag)
    !  #[
    TYPE(Mie_Height_Bin_type), intent(in)  :: Mie_Height_Bin ! input
    character(len=*),          intent(in)  :: prefix         ! input
    integer,                   intent(out) :: error_flag     ! output

    call AsciiWrite_Mie_Height_Bin(Mie_Height_Bin, fileunit_stdout, prefix,&
                                   error_flag)

  end subroutine Print_Mie_Height_Bin
    !  #]
  subroutine AsciiWrite_Mie_Height_Bin(Mie_Height_Bin, fileunit, prefix,&
                                       error_flag)
    !  #[
    TYPE(Mie_Height_Bin_type), intent(in)  :: Mie_Height_Bin ! input
    integer,                   intent(in)  :: fileunit       ! input
    character(len=*),          intent(in)  :: prefix         ! input
    integer,                   intent(out) :: error_flag     ! output

    ! local variables
    character(len=maxlen_IntAs)  :: str_IntAs
    character(len=maxlen_IntAus) :: str_IntAus
    character(len=maxlen_IntAul) :: str_IntAul
    character(len=10)            :: str_Valid

    error_flag = no_error

    if (Mie_Height_Bin%Validity_Flag) then
       str_Valid = "Valid"
    else
       str_Valid = "Not Valid"
    endif
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Validity_Flag              = "//trim(str_Valid)

    call SetIntAs(Mie_Height_Bin%Mie_Wind_Velocity,&
                  str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Mie_Wind_Velocity          = "//trim(str_IntAs)//&
         " [cm/s]"

    call SetIntAus(Mie_Height_Bin%Mie_Error_Quantifier,&
                  str_IntAus,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Mie_Error_Quantifier       = "//trim(str_IntAus)//&
         " [cm/s]"

    call SetIntAul(Mie_Height_Bin%Integration_Length,&
                  str_IntAul,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Integration_Length         = "//trim(str_IntAul)//&
         " [m]"

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_Mie_Height_Bin: "//&
                          "error while writing to outputfile")
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Mie_Height_Bin
    !  #]
  subroutine Write_L2B_Mie_Height_Bin(fileunit, &
             Mie_Height_Bin, error_flag)
    !  #[
    ! 28-11-2005 P. Poli  Original code. From TN2.1 v1.2 06-09-2005
    integer,                   intent(in)  :: fileunit
    TYPE(Mie_Height_Bin_type), intent(in)  :: Mie_Height_Bin
    integer,                   intent(out) :: error_flag

    error_flag = no_error

    call write_logical1_to_bin_file(fileunit, &
         Mie_Height_Bin%Validity_Flag, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAs_to_bin_file(fileunit, &
         Mie_Height_Bin%Mie_Wind_Velocity, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAus_to_bin_file(fileunit, &
         Mie_Height_Bin%Mie_Error_Quantifier, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAul_to_bin_file(fileunit, &
         Mie_Height_Bin%Integration_Length, error_flag)
    IF (error_flag .ne. no_error) goto 998

    !------------------------------
    ! done
    !------------------------------
    return

998 call logmsg(log_error,"ABORTING subroutine Write_L2B_Mie_Height_Bin")
    return

    end subroutine Write_L2B_Mie_Height_Bin
    !  #]
  subroutine Read_L2B_Mie_Height_Bin(fileunit, &
             Mie_Height_Bin, error_flag)
    !  #[
    ! 28-11-2005 P. Poli  Original code. From TN2.1 v1.2 06-09-2005
    integer,                   intent(in)    :: fileunit
    TYPE(Mie_Height_Bin_type), intent(inout) :: Mie_Height_Bin
    integer,                   intent(out)   :: error_flag

    error_flag = no_error

    call read_logical1_from_bin_file(fileunit, &
         Mie_Height_Bin%Validity_Flag, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_IntAs_from_bin_file(fileunit, &
         Mie_Height_Bin%Mie_Wind_Velocity, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_IntAus_from_bin_file(fileunit, &
         Mie_Height_Bin%Mie_Error_Quantifier, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_IntAul_from_bin_file(fileunit, &
         Mie_Height_Bin%Integration_Length, error_flag)
    IF (error_flag .ne. no_error) goto 998

    !------------------------------
    ! done
    !------------------------------
    return

998 call logmsg(log_error,"ABORTING subroutine Read_L2B_Mie_Height_Bin")
    return

  end subroutine Read_L2B_Mie_Height_Bin
    !  #]
  function Get_ByteSize_L2B_Mie_Height_Bin() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result

    c_bytesize = c_bytelen_Bool   + &
                 c_bytelen_IntAs  + &
                 c_bytelen_IntAus + &
                 c_bytelen_IntAul     ! 4 items in total

    ! typical sizes
    ! memory size = 4+2+4+8 = 18 bytes
    ! file   size = 1+2+2+4 =  9 bytes

  end function Get_ByteSize_L2B_Mie_Height_Bin
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_Mie_Profile(Mie_Profile, N_Range_Gates0, error_flag)
    !  #[
    TYPE(Mie_Profile_type), intent(out) :: Mie_Profile    ! output
    integer(IntAs),         intent(in)  :: N_Range_Gates0 ! input
    integer,                intent(out) :: error_flag     ! output

    ! local variable
    integer :: r, AllocStatus

    error_flag = no_error

    Mie_Profile%Obs_Type = Obs_Type_undefined

    IF (N_Range_Gates0 .eq. 0) THEN
       ! init them to be empty
       nullify(Mie_Profile%Height_Bin)
    ELSE
       allocate(Mie_Profile%Height_Bin(N_Range_Gates0),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          call logmsg(log_error,"Allocation failure in subroutine "//&
                                "Init_Mie_Profile")
          error_flag = error_allocate
          return
       END IF
       
       DO r=1,N_Range_Gates0
          call Init_Mie_Height_Bin(Mie_Profile%Height_Bin(r))
       END DO
    END IF

  end subroutine Init_Mie_Profile
    !  #]
  subroutine Delete_Mie_Profile(Mie_Profile)
    !  #[
    TYPE(Mie_Profile_type), intent(inout) :: Mie_Profile

    IF (associated(Mie_Profile%Height_Bin)) &
         deallocate(Mie_Profile%Height_Bin)

  end subroutine Delete_Mie_Profile
    !  #]
  subroutine Print_Mie_Profile(Mie_Profile, prefix, PrintAll, error_flag)
    !  #[
    TYPE(Mie_Profile_type), intent(in)  :: Mie_Profile ! input
    character(len=*),       intent(in)  :: prefix      ! input
    logical,                intent(in)  :: PrintAll    ! input
    integer,                intent(out) :: error_flag  ! output

    call AsciiWrite_Mie_Profile(Mie_Profile, fileunit_stdout, &
                                prefix, PrintAll, error_flag)

  end subroutine Print_Mie_Profile
    !  #]
  subroutine AsciiWrite_Mie_Profile(Mie_Profile, fileunit, &
                                    prefix, PrintAll, error_flag)
    !  #[
    TYPE(Mie_Profile_type), intent(in)  :: Mie_Profile ! input
    integer,                intent(in)  :: fileunit    ! input
    character(len=*),       intent(in)  :: prefix      ! input
    logical,                intent(in)  :: PrintAll    ! input
    integer,                intent(out) :: error_flag  ! output

    ! local variables
    integer            :: r, s, step
    character(len=256) :: ObsTypeText
    character(len=2)   :: str_cntr

    error_flag = no_error

    ObsTypeText = GetObsTypeText(Mie_Profile%Obs_Type)
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Obs_Type = "//trim(ObsTypeText)

    IF (associated(Mie_Profile%Height_Bin)) THEN
       ! get the sizes of the array
       s = size(Mie_Profile%Height_Bin)
       IF (PrintAll .or. (s .le. 1) ) THEN
          step = 1
       ELSE
          step = s-1
       END IF
       DO r=1,s,step
          write(str_cntr,"(i2.2)",err=998) r
          call AsciiWrite_Mie_Height_Bin(Mie_Profile%Height_Bin(r), fileunit,&
               trim(prefix)//"%Height_Bin("//trim(str_cntr)//")", error_flag)
          if (error_flag .ne. no_error) return
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%Height_Bin(:) [seems not associated]"
    END IF

    return

998 call logmsg(log_error,"conversion error in AsciiWrite_Mie_Profile: "//&
                "could not convert the counter r to a string ..")
    error_flag = error_writing_file
    return

999 call logmsg(log_error,"ERROR in AsciiWrite_Mie_Profile: "//&
                "error while writing to outputfile")
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Mie_Profile
    !  #]
  subroutine Write_L2B_Mie_Profile(fileunit, &
             Mie_Profile, error_flag)
    !  #[
    ! 25-11-2005 P. Poli  Original code. From TN2.1 v1.2 06-09-2005
    integer,                intent(in)  :: fileunit
    TYPE(Mie_Profile_type), intent(in)  :: Mie_Profile
    integer,                intent(out) :: error_flag

    ! local variables
    integer          :: r, s, nr_of_zero_bytes
    character(len=6) :: txt

    error_flag = no_error
    IF (.not.(associated(Mie_Profile%Height_Bin))) THEN
       call logmsg(log_error,"ERROR in Write_L2B_Mie_Profile(): "//&
            "The pointer to this Mie_Profile seems not associated !!")
       error_flag = error_programming
       return
    END IF

    ! Obs_Type
    call write_IntAuc_to_bin_file(fileunit, Mie_Profile%Obs_Type,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 36 spare bytes
    nr_of_zero_bytes = 36
    call write_SpareBytes_to_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! get the size of the array
    s = size(Mie_Profile%Height_Bin)
    write(txt,"(i6)",err=999) s
    call logmsg(log_info,"Write_L2B_Mie_Profile: "//&
                "size(Mie_Profile%Height_Bin) ="//trim(adjustl(txt)))
    DO r=1,s
       call Write_L2B_Mie_Height_Bin(fileunit, &
                                     Mie_Profile%Height_Bin(r), error_flag)
       if (error_flag .ne. no_error) goto 998
    END DO

    !------------------------------
    ! done
    !------------------------------
    return

998 call logmsg(log_error,"ABORTING subroutine Write_L2B_Mie_Profile")
    return

999 call logmsg(log_error,"ERROR in Write_L2B_Mie_Profile: "//&
                          "could not convert integer s to a string ...")
    return

  end subroutine Write_L2B_Mie_Profile
    !  #]
  subroutine Read_L2B_Mie_Profile(fileunit, &
             Mie_Profile, error_flag)
    !  #[
    ! 10-01-2006 P. Poli  Original code. From TN2.1 v1.2 06-09-2005
    integer,                intent(in)  :: fileunit
    TYPE(Mie_Profile_type), intent(inout):: Mie_Profile
    integer,                intent(out) :: error_flag

    ! local variables
    integer          :: r, s, nr_of_zero_bytes
    character(len=6) :: txt

    error_flag = no_error
    IF (.not.(associated(Mie_Profile%Height_Bin))) THEN
       call logmsg(log_error,"ERROR in Read_L2B_Mie_MDS(): "//&
                "The pointer to this Mie_Profile seems not associated !!")
       error_flag = error_programming
       return
    END IF

    ! Obs_Type
    call read_IntAuc_from_bin_file(fileunit, Mie_Profile%Obs_Type,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 36 spare bytes
    nr_of_zero_bytes = 36
    call read_SpareBytes_from_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! get the size of the array
    s = size(Mie_Profile%Height_Bin)
    write(txt,"(i6)",err=999) s
    call logmsg(log_info,"Read_L2B_Mie_Profile: "//&
                "size(Mie_Profile%Height_Bin) ="//trim(adjustl(txt)))
    DO r=1,s
       call Read_L2B_Mie_Height_Bin(fileunit, &
            Mie_Profile%Height_Bin(r), error_flag)
       if (error_flag .ne. no_error) goto 998
    END DO

    !------------------------------
    ! done
    !------------------------------
    return

998 call logmsg(log_error,"ABORTING subroutine Read_L2B_Mie_Profile")
    return

999 call logmsg(log_error,"ERROR in Read_L2B_Mie_Profile: "//&
                          "could not convert integer s to a string ...")
    return

  end subroutine Read_L2B_Mie_Profile
    !  #]
  function Get_ByteSize_L2B_Mie_Profile(Mie_Profile) result(c_bytesize)
    !  #[
    integer(IntAl)                    :: c_bytesize ! result
    TYPE(Mie_Profile_type), intent(in):: Mie_Profile

    c_bytesize = 1*c_bytelen_IntAuc + &
                 36_IntAl ! 36 spare bytes
    IF (associated(Mie_Profile%Height_Bin)) THEN
       c_bytesize = c_bytesize + &
            int(size(Mie_Profile%Height_Bin)*&
                Get_ByteSize_L2B_Mie_Height_Bin(),IntAl)
    END IF

    ! typical sizes
    ! memory size = 2+24*18 = 434 bytes
    ! file   size = 1+24*9  = 217 bytes
  end function Get_ByteSize_L2B_Mie_Profile
    !  #]
! todo: update typical sizes
  !---------------------------------------------------------------------
  subroutine Init_L2B_Mie_MDS(Mie_MDS, &
                      N_Meas0, N_Obs_Max_Mie0, N_Range_Gates0, error_flag)
    !  #[
    !  Modifications:
    !     P. Poli    02-12-2005    Saving N_Obs_Max in L2B_SummaryVars%M_Mie
    !     P. Poli    26-01-2006    Removed N_Obs_Max from Mie_MDS
    TYPE(L2B_Mie_MDS_type), pointer     :: Mie_MDS  ! input/output
    integer(IntAs),         intent(in)  :: N_Meas0        ! input
    integer(IntAs),         intent(in)  :: N_Obs_Max_Mie0     ! input
    integer(IntAs),         intent(in)  :: N_Range_Gates0 ! input
    integer,                intent(out) :: error_flag    ! output

    ! local variable
    integer :: p, AllocStatus

    error_flag = no_error

    IF (.not. associated(Mie_MDS)) THEN
       call logmsg(log_error,"ERROR in Init_L2B_Mie_MDS: "//&
                   "pointer Mie_MDS seems not associated !!!")
       error_flag = error_programming
       return
    END IF
    
    call InitDateTime(Mie_MDS%Start_of_Obs_DateTime)

    ! the first one is used for allocating/building the datastructure
    ! so keep a copy for convenience
    Mie_MDS%N_Meas       = N_Meas0
    ! these 2 are filled during processing, so init them to zero
    Mie_MDS%N_Obs_Mie_Actual = 0
    Mie_MDS%P                = 0

    allocate(Mie_MDS%Map_of_L1B_Measurements_Used,STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       call logmsg(log_error,"Allocation failure in subroutine "//&
                   "Init_L2B_Mie_MDS")
       error_flag = error_allocate
       return
    END IF
    call Init_Meas_Map(Mie_MDS%Map_of_L1B_Measurements_Used,&
                       N_Meas0, N_Range_Gates0, error_flag)
    IF (error_flag .ne. no_error) return

    IF (N_Obs_Max_Mie0 .eq. 0) THEN
       ! init it to be empty
       nullify(Mie_MDS%Mie_Profile)
    ELSE
       allocate(Mie_MDS%Mie_Profile(N_Obs_Max_Mie0),&
                STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          call logmsg(log_error,"Allocation failure in subroutine "//&
                      "Init_L2B_Mie_MDS")
          error_flag = error_allocate
          return
       END IF
       
       DO p=1,N_Obs_Max_Mie0
          call Init_Mie_Profile(Mie_MDS%Mie_Profile(p),&
                                     N_Range_Gates0, error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    END IF

  end subroutine Init_L2B_Mie_MDS
    !  #]
  subroutine Delete_L2B_Mie_MDS(Mie_MDS)
    !  #[
    TYPE(L2B_Mie_MDS_type), pointer :: Mie_MDS

    ! local variable
    integer :: p

    IF (.not. associated(Mie_MDS)) return
    
    IF (associated(Mie_MDS%Map_of_L1B_Measurements_Used)) THEN
       call Delete_Meas_Map(Mie_MDS%Map_of_L1B_Measurements_Used)
       deallocate(Mie_MDS%Map_of_L1B_Measurements_Used)
    END IF

    IF (associated(Mie_MDS%Mie_Profile)) THEN
       DO p=1,size(Mie_MDS%Mie_Profile)
          call Delete_Mie_Profile(Mie_MDS%Mie_Profile(p))
       END DO
       deallocate(Mie_MDS%Mie_Profile)
    END IF

  end subroutine Delete_L2B_Mie_MDS
    !  #]
  !---------------------------------------------------------------------
  function get_Mie_nr_of_cloud_points(Mie_MDS,r) result(n)
    !  #[
    TYPE(L2B_Mie_MDS_type), pointer :: Mie_MDS ! input
    integer, intent(in) :: r ! input
    integer             :: n ! result

    ! local variable
    integer :: o, Obs_Type

    n = -1
    IF (.not. associated(Mie_MDS)) return

    ! the counts are not stored as such, so use a little loop 
    ! to construct the numbers
    n = 0
    do o=1,Mie_MDS%N_Obs_Mie_Actual
       Obs_Type = Mie_MDS%Mie_Profile(o)%Obs_Type
       if (Obs_Type .eq. Obs_Type_cloudy_returns) then
          n = n + Count_Nr_Points_On_Level(&
                      Mie_MDS%Map_of_L1B_Measurements_Used,r,o)
       end if
    end do

  end function get_Mie_nr_of_cloud_points
    !  #]
  function get_Mie_nr_of_clear_points(Mie_MDS,r) result(n)
    !  #[
    TYPE(L2B_Mie_MDS_type), pointer :: Mie_MDS ! input
    integer, intent(in) :: r ! input
    integer             :: n ! result

    ! local variable
    integer :: o, Obs_Type

    n = -1
    IF (.not. associated(Mie_MDS)) return

    ! the counts are not stored as such, so use a little loop 
    ! to construct the numbers
    n = 0
    do o=1,Mie_MDS%N_Obs_Mie_Actual
       Obs_Type = Mie_MDS%Mie_Profile(o)%Obs_Type
       if (Obs_Type .eq. Obs_Type_clear_returns) then
          n = n + Count_Nr_Points_On_Level(&
                      Mie_MDS%Map_of_L1B_Measurements_Used,r,o)
       end if
    end do

  end function get_Mie_nr_of_clear_points
    !  #]
  !---------------------------------------------------------------------
  subroutine Print_L2B_Mie_MDS(Mie_MDS,prefix,PrintAll,error_flag)
    !  #[
    TYPE(L2B_Mie_MDS_type), pointer     :: Mie_MDS    ! input
    character(len=*),       intent(in)  :: prefix     ! input
    logical,                intent(in)  :: PrintAll   ! input
    integer,                intent(out) :: error_flag ! output

    call AsciiWrite_L2B_Mie_MDS(Mie_MDS,fileunit_stdout,prefix,PrintAll,&
                                error_flag)

  end subroutine Print_L2B_Mie_MDS
    !  #]
  subroutine AsciiWrite_L2B_Mie_MDS(Mie_MDS,fileunit,prefix,PrintAll,&
                                    error_flag)
    !  #[
    TYPE(L2B_Mie_MDS_type), pointer     :: Mie_MDS    ! input
    integer,                intent(in)  :: fileunit   ! input
    character(len=*),       intent(in)  :: prefix     ! input
    logical,                intent(in)  :: PrintAll   ! input
    integer,                intent(out) :: error_flag ! output

    ! local variables
    integer                            :: p, s, step
    character(len=maxlen_DateTime_KVT) :: str_DateTime
    character(len=maxlen_IntAs)        :: str_IntAs
    character(len=2)                   :: str_cntr

    error_flag = no_error

    IF (.not. associated(Mie_MDS)) THEN
       call logmsg(log_error,"Error in AsciiWrite_L2B_Mie_MDS(): "//&
                   "Mie_MDS [seems not associated]")
       error_flag = error_programming
       return
    END IF

    if (DateTimeMissing(Mie_MDS%start_of_obs_datetime)) then
       str_DateTime = "[missing]"
    else
       str_DateTime = Construct_KVT_DateTime(Mie_MDS%start_of_obs_datetime)
    end if
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%start_of_obs_datetime = "//trim(str_DateTime)

    call SetIntAs(Mie_MDS%N_Meas,str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%N_Meas                = "//trim(str_IntAs)

    call SetIntAs(Mie_MDS%N_Obs_Mie_Actual,str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%N_Obs_Mie_Actual      = "//trim(str_IntAs)

    call SetIntAs(Mie_MDS%P,str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%P                     = "//trim(str_IntAs)

    IF (associated(Mie_MDS%Map_of_L1B_Measurements_Used)) THEN
       call AsciiWrite_Meas_Map(Mie_MDS%Map_of_L1B_Measurements_Used,&
            fileunit,trim(prefix)//"%Map_of_L1B_Measurements_Used",&
            PrintAsMap=.false.,PrintAll=PrintAll,error_flag=error_flag)
          IF (error_flag .ne. no_error) goto 999          
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%Map_of_L1B_Measurements_Used = [seems not associated]"
    END IF

    IF (associated(Mie_MDS%Mie_Profile)) THEN
       s = size(Mie_MDS%Mie_Profile)
       IF (PrintAll .or. (s .le. 1) ) THEN
          step = 1
       ELSE
          step = s-1
       END IF
       DO p=1,s,step
          write(str_cntr,"(i2.2)",err=998) p
          call AsciiWrite_Mie_Profile(Mie_MDS%Mie_Profile(p), fileunit, &
               trim(prefix)//"%Mie_Profile("//trim(str_cntr)//")",&
               PrintAll,error_flag)
          IF (error_flag .ne. no_error) goto 999          
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%Mie_Profile(:) [seems not associated]"
    END IF

    return

998 call logmsg(log_error,"conversion error in AsciiWrite_L2B_Mie_MDS: "//&
                "could not convert the counter p to a string ..")
    error_flag = error_writing_file
    return

999 call logmsg(log_error,"ERROR in AsciiWrite_L2B_Mie_MDS: "//&
                "error while writing to outputfile")
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2B_Mie_MDS
    !  #]
  subroutine Print_L2B_Mie_hlos_profile(Mie_MDS,error_flag,Obs_Nr)
    !  #[
    TYPE(L2B_Mie_MDS_type), pointer     :: Mie_MDS
    integer,                intent(out) :: error_flag ! output
    integer,                intent(in)  :: Obs_Nr

    call AsciiWrite_L2B_Mie_hlos_profile(Mie_MDS,error_flag,&
                                         Obs_Nr, fileunit_stdout)

  end subroutine Print_L2B_Mie_hlos_profile
    !  #]
  subroutine AsciiWrite_L2B_Mie_hlos_profile(Mie_MDS,error_flag,&
                                             Obs_Nr, Fileunit)
    !  #[
    TYPE(L2B_Mie_MDS_type), pointer     :: Mie_MDS
    integer,                intent(out) :: error_flag ! output
    integer,                intent(in)  :: Obs_Nr
    integer,                intent(in)  :: Fileunit

    ! local variables
    integer                      :: Bin, Nr_Bins
    character(len=maxlen_IntAs)  :: str_IntAs
    character(len=1)             :: Validity_Flag

    error_flag = no_error

    IF (.not. associated(Mie_MDS)) THEN
       call logmsg(log_error,"ERROR: Mie_MDS [seems not associated]")
       error_flag = error_programming
       return
    END IF

    IF (.not. associated(Mie_MDS%Mie_Profile)) THEN
       call logmsg(log_error,"ERROR: Mie_MDS%Mie_Profile "//&
                   "[seems not associated]")
       error_flag = error_programming
       return
    END IF
   
    write(Fileunit,"(a,i2)") " ==> N_Obs_Actual = ",Mie_MDS%N_Obs_Mie_Actual
    IF (Obs_Nr .gt. Mie_MDS%N_Obs_Mie_Actual) THEN
       write(Fileunit,*) "Sorry, Obs_Nr ",Obs_Nr," was not (yet) defined"
       write(Fileunit,*) "Nothing to print."
       return
    END IF

    write(Fileunit,*) "Mie hlos-profile for Obs. ",Obs_Nr
    Nr_Bins = size(Mie_MDS%Mie_Profile(Obs_Nr)%Height_Bin)
    write(Fileunit,*) "nr height bins = ",Nr_Bins
    
    DO Bin=1,Nr_Bins
       call SetIntAs(Mie_MDS%Mie_Profile(Obs_Nr)%&
            Height_Bin(Bin)%Mie_Wind_Velocity,&
            str_IntAS, error_flag)
       if (error_flag .ne. no_error) return
       IF (Mie_MDS%Mie_Profile(Obs_Nr)%Height_Bin(Bin)%Validity_Flag) THEN
          Validity_Flag="1"
       ELSE
          Validity_Flag="0"
       ENDIF
       write(Fileunit,*) "hlos = ",str_IntAs,"  ",Validity_Flag
    END DO

  end subroutine AsciiWrite_L2B_Mie_hlos_profile
    !  #]
  !---------------------------------------------------------------------
  subroutine Write_L2B_Mie_MDS(fileunit, Mie_MDS,& 
                               N_Meas_to_write, N_Obs_Mie_to_write,error_flag)
    !  #[

    ! Writes the Mie MDS in binary format
    ! 24-11-2005 P. Poli Original code. From L2B TN2.1 v1.2 06-09-2005
    ! 10-01-2005 P. Poli Modified writing of DateTime. Now calls ymd2julian
    ! 23-feb-2007 J. de Kloe removed L2B_SummaryVars

    integer,                intent(in)  :: fileunit           ! input
    TYPE(L2B_Mie_MDS_type), pointer     :: Mie_MDS            ! input
    integer(IntAuc),        intent(in)  :: N_Meas_to_write    ! input
    integer(IntAuc),        intent(in)  :: N_Obs_Mie_to_write ! input
    integer,                intent(out) :: error_flag         ! output

    ! local variables
    integer          :: p, s
    character(len=6) :: txt1, txt2

    !------------------------------
    ! writing the Mie MDS elements
    !------------------------------

    error_flag = no_error
    IF (.not. associated(Mie_MDS)) THEN
       call logmsg(log_error,"ERROR in Write_L2B_Mie_MDS(): "//&
                   "The pointer to this Mie_MDS seems not associated !!")
       error_flag = error_programming
       return
    END IF

    ! Start_of_Obs_DateTime
    call write_DateTime_to_bin_file(fileunit, &
         Mie_MDS%Start_of_Obs_DateTime,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call write_IntAs_to_bin_file(fileunit,Mie_MDS%N_Meas,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call write_IntAs_to_bin_file(fileunit,Mie_MDS%N_Obs_Mie_Actual,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call write_IntAs_to_bin_file(fileunit,Mie_MDS%P,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call Write_Meas_Map(fileunit, &
         Mie_MDS%Map_of_L1B_Measurements_Used, N_Meas_to_write, &
         error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Changed this code, J. de Kloe, 30-mar-2006:
    ! Unused profiles should be initialised now with missing values,
    ! so there is no problem to just write them to file.
    ! This simplifies both the reading and writing routines a lot

    s = size(Mie_MDS%Mie_Profile)
    write(txt1,"(i6)",err=995) N_Obs_Mie_to_write
    write(txt2,"(i6)",err=996) s
    IF (N_Obs_Mie_to_write .gt.s) THEN
       call logmsg(log_error,"ERROR in Write_L2B_Mie_MDS(): "//&
            "invalid value for N_Obs_Mie_to_write: "//trim(adjustl(txt1))//&
            " max. possible value is: size(Mie_MDS%Mie_Profile) = "//&
            trim(adjustl(txt2)))
       error_flag = error_programming
       return
    END IF

    write(txt1,"(i6)",err=997) Mie_MDS%N_Obs_Mie_Actual
    call logmsg(log_info,"Write_L2B_Mie_MDS: Mie_MDS%N_Obs_Actual = "//&
                trim(adjustl(txt1)))
    DO p=1,N_Obs_Mie_to_write
       call Write_L2B_Mie_Profile(fileunit,Mie_MDS%Mie_Profile(p),error_flag)
       IF (error_flag .ne. no_error) goto 998
    ENDDO

    !------------------------------
    ! done
    !------------------------------
    return

995 call logmsg(log_error,"ERROR in Write_L2B_Mie_MDS: Could not convert "//&
         "integer N_Obs_Mie_to_write to a string.")
    return

996 call logmsg(log_error,"ERROR in Write_L2B_Mie_MDS: Could not convert "//&
         "integer s to a string")
    return

997 call logmsg(log_error,"ERROR in Write_L2B_Mie_MDS: Could not convert "//&
         "integer Mie_MDS%N_Obs_Mie_Actual to a string")
    return

998 call logmsg(log_error,"ABORTING subroutine Write_L2B_Mie_MDS")
    return

  end subroutine Write_L2B_Mie_MDS
    !  #]
  subroutine Read_L2B_Mie_MDS(fileunit,Mie_MDS,error_flag)
    !  #[
    ! 10-01-2005 P. Poli Original code.
    integer,                intent(in)  :: fileunit
    TYPE(L2B_Mie_MDS_type), pointer     :: Mie_MDS
    integer,                intent(out) :: error_flag
    character(len=maxlen_DateTime_KVT) :: DateTimeString

    ! local variables
    integer          :: p, s
    character(len=6) :: txt

    !------------------------------
    ! writing the Mie MDS elements
    !------------------------------

    error_flag = no_error
    IF (.not. associated(Mie_MDS)) THEN
       call logmsg(log_error,"ERROR in Read_L2B_Mie_MDS(): "//&
                   "The pointer to this Mie_MDS seems not associated !!")
       error_flag = error_programming
       return
    END IF

    ! Start_of_Obs_DateTime
    call read_DateTime_from_bin_file(fileunit,&
                      Mie_MDS%Start_of_Obs_DateTime,error_flag)

    if (DateTimeMissing(Mie_MDS%Start_of_Obs_DateTime)) then
       DateTimeString = "[missing]"
    else
       DateTimeString = Construct_KVT_DateTime(Mie_MDS%Start_of_Obs_DateTime)
    end if
    call logmsg(log_info,"Read_L2B_Mie_MDS: Start_of_Obs_DateTime is: "//&
                DateTimeString)

    call read_IntAs_from_bin_file(fileunit,Mie_MDS%N_Meas,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call read_IntAs_from_bin_file(fileunit,Mie_MDS%N_Obs_Mie_Actual,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call read_IntAs_from_bin_file(fileunit,Mie_MDS%P,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call Read_Meas_Map(fileunit, &
         Mie_MDS%Map_of_L1B_Measurements_Used, &
         error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! for N_Obs=0 this pointer may be un-associated
    ! so we have to check for that before requesting its size
    IF (associated(Mie_MDS%Mie_Profile)) THEN
       s = size(Mie_MDS%Mie_Profile)
       write(txt,"(i6)",err=999) Mie_MDS%N_Obs_Mie_Actual
       call logmsg(log_info,"Read_L2B_Mie_MDS: Mie_MDS%N_Obs_Actual ="//&
                   trim(adjustl(txt)))
       DO p=1,s
          call Read_L2B_Mie_Profile(fileunit,Mie_MDS%Mie_Profile(p),error_flag)
          IF (error_flag .ne. no_error) goto 998
       ENDDO
    END IF

    !------------------------------
    ! done
    !------------------------------
    return

998 call logmsg(log_error,"ABORTING subroutine Read_L2B_Mie_MDS")
    return

999 call logmsg(log_error,"conversion error in Read_L2B_Mie_MDS: "//&
                "could not convert the integer Mie_MDS%N_Obs_Mie_Actual "//&
                "to a string ..")
    error_flag = error_writing_file
    return

  end subroutine Read_L2B_Mie_MDS
    !  #]
  function Get_ByteSize_L2B_Mie_MDS(Mie_MDS, N_Meas_to_write, &
                                    N_Obs_Mie_to_write) result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result
    TYPE(L2B_Mie_MDS_type), pointer :: Mie_MDS
    integer(IntAuc), intent(in) :: N_Meas_to_write
    integer(IntAuc), intent(in) :: N_Obs_Mie_to_write

    ! local variable
    integer          :: s
    character(len=6) :: txt1, txt2

    c_bytesize = 0
    IF (.not. associated(Mie_MDS)) return
       
    c_bytesize = c_bytesize + c_bytelen_DateTime
    c_bytesize = c_bytesize + c_bytelen_IntAs ! N_Meas_Actual
    c_bytesize = c_bytesize + c_bytelen_IntAs ! N_Obs_Mie_Actual
    c_bytesize = c_bytesize + c_bytelen_IntAs ! nr laser pulses

    IF (associated(Mie_MDS%Map_of_L1B_Measurements_Used)) THEN
       c_bytesize = c_bytesize + &
            Get_ByteSize_Meas_Map(Mie_MDS%Map_of_L1B_Measurements_Used,&
                                  N_Meas_to_write)
    END IF

    IF (associated(Mie_MDS%Mie_Profile)) THEN
       s = size(Mie_MDS%Mie_Profile)
       IF (N_Obs_Mie_to_write .gt. s) THEN
          write(txt1,"(i6)",err=998) N_Obs_Mie_to_write
          write(txt2,"(i6)",err=999) s
          call logmsg(log_warn,"WARNING in Get_ByteSize_L2B_Mie_MDS(): "//&
               "invalid value for N_Obs_Mie_to_write: "//trim(adjustl(txt1))//&
               "max. possible value is: size(Mie_MDS%Mie_Profile) = "//&
               trim(adjustl(txt2))//" Get_ByteSize result will be useless....")
       END IF

       c_bytesize = c_bytesize + N_Obs_Mie_to_write * &
            Get_ByteSize_L2B_Mie_Profile(Mie_MDS%Mie_Profile(1))
    END IF

    ! typical sizes
    ! memory size = 20 + 6 + 9648 + 5*434 + = 11844 bytes
    ! file   size = 12 + 6 + 4824 + 5*217 + =  5927 bytes

    return

998 call logmsg(log_warn,"conversion problem in Get_ByteSize_L2B_Mie_MDS: "//&
                "could not convert the integer N_Obs_Mie_to_write "//&
                "to a string ... Get_ByteSize result will be useless....")
    return

999 call logmsg(log_warn,"conversion problem in Get_ByteSize_L2B_Mie_MDS: "//&
                "could not convert the integer s "//&
                "to a string ... Get_ByteSize result will be useless....")
    return

  end function Get_ByteSize_L2B_Mie_MDS
    !  #]
  !---------------------------------------------------------------------
end module L2B_Mie_MDS
