module L2B_Proc_Settings
  !  #[ Documentation
  !
  ! this module defines the program settings as used by the 
  ! L2B processor.
  ! (and eventually will define the
  !  L2B_Proc_Settings_Parameters_ADS global annotation dataset (GADS)
  !  to store these settings in.
  !
  ! Written by J. de Kloe (KNMI)
  ! Created:     09-Nov-2005
  !
  ! Modifications:
  !    28-Nov-2005 P.Poli     Modified Get_Commandline_Options
  !    23-Dec-2005 D.Tan      Error Quantifier Methods for Mie and Rayleigh
  !    05-Jan-2006 J. de Kloe added switches to control backscatter ratio calc.
  !    08-Feb-2006 J. de Kloe expanded the classify adjacent option to 2 types
  !    11-Apr-2006 P.Poli     added command line argument -RBCfile
  !    19-Apr-2006 J. de Kloe removed the -L2BasciiFile option since we will
  !                           probably not use that output type anymore
  !                           (and no tests are done on this file)
  !    24-Apr-2006 J. de Kloe moved the Read_L2B_AuxPar_file and the
  !                           Read_JobOrderFile routines to their own dirs
  !                           and rearranged a bit.
  !    09-May-2006 D.Tan      Amend default settings when JobOrder is specified
  !    12-May-2006 D. Tan     Null use of ParKind1 module
  !    23-Jun-2006 J. de Kloe added switches to control particle and molecular
  !                           extinction calculation.
  !    09-Aug-2006 J. de Kloe removed a number of commandline switches and
  !                           activated the corresponding JobOrder or AuxPar
  !                           items
  !    10-Nov-2006 J. de Kloe moved Copy_JobOrderData_Items in here from 
  !                           l2bp_module. this was needed to enable use of
  !                           this routine from other directories than main
  !    15-Feb-2007 J. de Kloe implemented use of L1B timerange as specified
  !                           in the JobOrder file
  !    01-Mar-2007 J. de Kloe implemented checks on presence of all needed
  !                           files and filename definition
  !    01-Mar-2007 J. de Kloe added additional command line option to enable
  !                           use of the optional logfile
  !    17-Apr-2007 D.Tan      Phase out AUX_???_AX.
  !    26-Jul-2007 J. de Kloe adapted to handle multiple AMD input files
  !    27-Jul-2007 J. de Kloe adapted to handle timeranges for the multiple 
  !                           AMD input files, and made the timerange handling
  !                           code generic, so it can be used for the L1B
  !                           input file as well.
  !    03-Oct-2007 J. de Kloe added an undocumented commandline option
  !                           "-ExpExtAlgo" which sets the switch 
  !                           UseExperimentalExtAlgorithm, which allows 
  !                           Gert-Jan to use his experimental extinction 
  !                           code more easily when he further develops it.
  !                           And some logic to stop the program when a user
  !                           tries to use Class_Ext_Threshold without this 
  !                           commandline switch.
  !    06-Dec-2007 J. de Kloe added some code to copy several additional
  !                           JobOrder items to the settings struct, needed
  !                           to fill some FH/MPH fields (see SPR 28)
  !    17-Dec-2007 J. de Kloe added L2C_flag to Init_L2B_Proc_Settings
  !    16-Jan-2008 J. de Kloe phase out integer kind i_ 
  !
  !  #] 
  !  #[ Modules used

  USE L2B_AuxPar_DataStructure, only: L2B_AuxPar_Type, Init_L2B_AuxPar, &
       AsciiWrite_L2B_AuxPar, Delete_L2B_AuxPar, &
       Class_No_Clouds, Class_Copy_E2S_Input, &
       Class_Ext_Threshold, Class_Cal_Mie_Threshold, Class_Rayleigh_Slope, &
       Class_Mie_Slope, Class_Backscat_Ratio, Class_Adjacent_Mie_Backscatter, &
       Class_Adjacent_Rayl_Backscatter, Class_Tau, &
       Scat_Ratio_from_L1B, Scat_Ratio_from_MieRayl, Scat_Ratio_from_RaylOnly,&
       Scat_Ratio_Dont_Use
  USE JobOrder_DataStructure, only: JobOrderData_type, Input_type
  USE FixedHeader, only: maxlen_system, maxlen_creator, &
                         maxlen_creator_version, maxlen_notes, &
                         maxlen_file_class, systemstr_ECMWF_IFS, &
                         systemstr_ESRIN_REPR, systemstr_LOCAL_MET_INST
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE EE_CFI_Datatypes, only: FAdoxy, SetFAdoFull_or_Missing, DateTimeType,&
                              Convert_JobOrd_to_Num_DateTime, &
                              Construct_KVT_DateTime, InitDateTime, &
                              maxlen_DateTime_KVT,DateTimeMissing
  USE ErrorHandler, only: no_error, error_cmdline_options, &
                          request_program_to_stop, error_writing_file, &
                          error_reading_file, error_allocate, error_programming
  USE Compiler_Features, only: iargc_aeolus,getarg_aeolus
  USE Logging, only: logmsg,log_debug,log_info,log_progress,log_warn,&
                     log_error,CheckIfFatalErrorOccurred,GetFatalError
  USE LunManager, only: fileunit_stdout
  USE c_support, only: get_filesize
  USE AeolusConstants, only: software_TaskName, software_version
  USE StringTools, only: to_lowercase
  !  #]
  !  #[ Variable and Parameter definitions

  implicit none
  !-------------------------------------------------------------
  ! list of choices/options for the L2BP
  !-------------------------------------------------------------

  ! TODO/WARNING: all these parameters will be moved to the
  !               L2B_AuxPar_DataStructure module in a later clean-up 

  ! posible sources for extinction by particles
  integer, parameter :: PartExt_Method_Undefined   = 0
  integer, parameter :: PartExt_Method_Old         = 1
  integer, parameter :: PartExt_Method_Iterative   = 2

  ! posible sources for extinction by molecules (and also backscatter
  ! which only has the added factor 8*pi/3)
  integer, parameter :: MolExt_Method_Undefined    = 0
  integer, parameter :: MolExt_Method_Parametrised = 1
  integer, parameter :: MolExt_Method_Proper       = 2

  !-------------------------------------------------------------
  ! a struct containing switches that control the L2BP processing.
  ! This is needed because the JobOrder structure is generic and holds 
  ! many items not needed for the L2BP task. Therefore items that we
  ! actually used are copied into this structure.
  ! Since we define the L2B_AuxPar structure ourselves, we just 
  ! add it here.
  !-------------------------------------------------------------

  integer, parameter :: maxlen_file_name = 256

  ! definitions copied from: ThinLayer/readjoborderdata.F90
  integer, parameter :: maxlen_log_level    =  8

  ! looks that this type looks like the Time_Interval_Type
  ! defined in joborder_datastructure.F90
  ! but it is not the same, since below we use the DateTimeType
  ! which is usual in the L2BP software, while in Time_Interval_Type
  ! the datetime is store as a string in a format again different from
  ! all xml/kvt formats that we already had...
  TYPE FileProperties
     TYPE(DateTimeType)              :: start
     TYPE(DateTimeType)              :: stop
     character(len=maxlen_file_name) :: Name
  END TYPE FileProperties

  TYPE L2BP_Settings_type
     ! selected items copied from the JobOrder file
     ! log level to stdout
     character(len=maxlen_log_level) :: Stdout_Log_Level
     ! log level to stderr
     character(len=maxlen_log_level) :: Stderr_Log_Level
     ! output filename for the optional logfile
     character(len=maxlen_file_name) :: LogFilename
     ! input  filename for L1B data (without extension)
     !character(len=256)              :: L1Bfilename
     !TYPE(DateTimeType)              :: L1Bfile_start
     !TYPE(DateTimeType)              :: L1Bfile_stop
     ! replaced by
     type(FileProperties)             :: L1Bfile
     ! selected timerange for the L1B input file
     ! Note that the JobOrder system allows for multiple L1B input
     ! files, and for multiple allowed timeranges for each L1b file,
     ! but for now we assume we will only have one L1B file, 
     ! and this file will have only one timewindow definition for it.
     ! output directory for L2B data
     !
     ! note that the L2Bfilename is generated, so there is no need to
     ! pass it to the processor.
     character(len=maxlen_file_name) :: L2Bresultdir
     ! input  filename for RBC data (without extension)
     character(len=maxlen_file_name) :: RBCfilename
     ! todo: implement use of the joborder specified timerange
     ! input  filename for L2BP auxiliary parameter file
     character(len=maxlen_file_name) :: L2BAuxPar_filename
     ! todo: implement use of the joborder specified timerange
     ! input  filename for auxiliary calibration file
     character(len=maxlen_file_name) :: AuxClm_filename
     ! todo: implement use of the joborder specified timerange
     ! input  filename for auxiliary calibration file
     character(len=maxlen_file_name) :: AuxCal_filename
     ! todo: implement use of the joborder specified timerange
     ! input/output filename for NWP data
     integer                         :: NumNWPfiles
     type(FileProperties), dimension(:), pointer   :: NWPfile
     ! todo: implement use of the joborder specified timerange
     character(len=maxlen_system)          :: System          ! maxlen=19
     character(len=maxlen_creator)         :: Creator         ! maxlen=40
     character(len=maxlen_creator_version) :: Creator_Version ! maxlen=12
     character(len=maxlen_notes)           :: Notes           ! maxlen=32
     character(len=maxlen_file_class)      :: File_Class      ! maxlen=4
     ! ...
     ! more to follow
     !
     ! the full AuxPar datastructure
     TYPE(L2B_AuxPar_Type) :: L2B_AuxPar
  END TYPE L2BP_Settings_type

  !-------------------------------------------------------------
  ! switches that control the processing used by the L2BP
  ! all these should be replaced by items in the L2B_AuxPar_Type
  ! or added as copied JobOrder items in the L2BP_Settings_type
  ! defined above. (except for items that are for internal use only
  ! like the use_dummylatlon and use_lite_test_data switches)
  !-------------------------------------------------------------

  ! this one is not yet added to the L2B_AuxPar definition !

  ! REMARK: these variables will be set by calling the
  !         Set_Default_L2B_Proc_Settings subroutine

  ! a variable to save the method to be used to get the extinction by molecules
  integer, save :: PartExt_Method

  ! this one is not yet added to the L2B_AuxPar definition !

  ! a variable to save the method to be used to get the extinction by particles
  integer, save :: MolExt_Method

  ! a switch to enable the experimental extinction algorithm
  ! in development at KNMI 
  logical, save :: UseExperimentalExtAlgorithm

  ! the next 2 items are not available oin the L2B_AuxPar file definitions
  ! but are kept for our own convenience for testing purposes

  ! use dummy lat-lon values, in stead of the values specified in the L1B file
  ! todo: phase out this test setting
  logical, save :: use_dummy_lat_lon

  ! use Lite data for testing the processor
  ! todo: phase out this test setting
  logical, save :: use_lite_test_data

  ! used by preprocessing only

  ! a small text file holding geolocation needed to extract NWP data
  ! todo: phase out this test setting, once we create a dedicated
  ! example test program to do this preprocessing (and move it out of the 
  ! main L2B_processor program).
  character(len=256), save :: NWPmetadataFilename

  !-------------------------------------------------------------
  ! derived switches
  !-------------------------------------------------------------
  ! actions to be performed by the main program
  logical, save :: do_preprocessing_step
  logical, save :: do_primary_processing

  ! this one can be derived from the choosen classification type
  ! but for the moment, for testing purposes, just calculate it always
  logical, save :: Extinction_Needed

  ! this is in fact always needed for RayleighBrillouin processing,
  ! and optionally also for classification. However, in case we
  ! ever need to switch it off, this switch is kept for now
  logical, save :: BackscatterRatio_Needed
  !-------------------------------------------------------------

  !  #]
contains
  !-----------------------
  subroutine InitFileProperties(File)
    !  #[
    type(FileProperties), intent(out) :: File

    file%name         = "[undefined]"
    call InitDateTime(file%start)
    call InitDateTime(file%stop)

  end subroutine InitFileProperties
    !  #]  
  subroutine AsciiWriteFileProperties(File,fileunit,prefix,error_flag)
    !  #[
    type(FileProperties), intent(in)  :: File             ! input
    integer,              intent(in)  :: fileunit         ! input
    character(len=*),     intent(in)  :: prefix           ! input
    integer,              intent(out) :: error_flag       ! output

    ! local variable
    character(len=maxlen_DateTime_KVT) :: tmp_txt_datetime

    ! init
    error_flag = no_error

    write(fileunit,"(a)",err=999) trim(prefix)//"%name  = "//&
         trim(File%name)
    ! WARNING: this 2 step copy is needed for pgf90 !
    ! if you glue Construct_KVT_DateTime() directly to the string
    ! in the write statement, the program crashes with a mysterious
    ! "Not Allocated" error, which looks like a compiler bug to me.....
    tmp_txt_datetime = Construct_KVT_DateTime(File%start)
    write(fileunit,"(a)",err=999) trim(prefix)//"%start = "//&
         trim(tmp_txt_datetime)
    tmp_txt_datetime = Construct_KVT_DateTime(File%stop)
    write(fileunit,"(a)",err=999) trim(prefix)//"%stop  = "//&
         trim(tmp_txt_datetime)

    return

999 call logmsg(log_error,"ERROR in AsciiWriteFileProperties: "//&
                          "error while writing to outputfile")
    error_flag = error_writing_file
    return

  end subroutine AsciiWriteFileProperties
    !  #]
  subroutine Init_L2B_Proc_Settings(L2BP_Settings,error_flag,L2C_flag)
    !  #[
    ! a wrapper for now, around Set_Default_L2B_Proc_Settings
    ! todo: remove Set_Default_L2B_Proc_Settings
    !       and use Init_L2B_Proc_Settings whenever needed
    TYPE(L2BP_Settings_type), intent(out) :: L2BP_Settings
    integer,                  intent(out) :: error_flag
    logical, optional,        intent(in)  :: L2C_flag

    ! these settings can be changed by setting the different
    ! commandline options
    ! or later by using the thin-layer xml files

    logical :: L2C_flag_local

    error_flag = no_error
    L2C_flag_local = .false.
    if (present(L2C_flag)) L2C_flag_local = L2C_flag

    L2BP_Settings%Stdout_Log_Level(:) = 'DEBUG'
    L2BP_Settings%Stderr_Log_Level(:) = 'NOOP'
    L2BP_Settings%LogFilename         = "[undefined]"
    call InitFileProperties(L2BP_Settings%L1Bfile)
    L2BP_Settings%L2Bresultdir        = "[undefined]"
    L2BP_Settings%RBCfilename         = "[undefined]"
    L2BP_Settings%L2BAuxPar_filename  = "[undefined]"
    L2BP_Settings%AuxClm_filename     = "[undefined]"
    L2BP_Settings%AuxCal_filename     = "[undefined]"
    L2BP_Settings%NumNWPfiles         = 0
    nullify(L2BP_Settings%NWPfile)
    L2BP_Settings%System              = "[undefined]"
    L2BP_Settings%Creator             = "[undefined]"
    L2BP_Settings%Creator_Version     = "[undefined]"
    L2BP_Settings%File_Class          = "TEST"

    ! hardcoded the processor name and version, as suggested in SPR 28
    if (L2C_flag_local) then
       L2BP_Settings%Notes            = "L2C_processor, version: "//&
                                        software_version
    else
       L2BP_Settings%Notes            = "L2B_processor, version: "//&
                                        software_version
    end if

    call Init_L2B_AuxPar(L2BP_Settings%L2B_AuxPar,error_flag)
    IF (error_flag .ne. no_error) return

    ! old settings, not yet included in the L2BP_Settings_type
    ! (TODO: to be phased out)

    do_preprocessing_step = .false.
    do_primary_processing = .false.

    ! use dummy lat-lon values, in stead of the values specified in L1B file
    use_dummy_lat_lon  = .false.
    ! use Lite data for testing the processor
    use_lite_test_data = .false.

    NWPmetadataFilename = "[undefined]"

    ! use the old extinction calculation to not break any existing tests
    ! in main and Classification
    PartExt_Method = PartExt_Method_Old
!    PartExt_Method = PartExt_Method_Iterative

    ! a switch to enable the experimental extinction algorithm
    ! in development at KNMI. By default disable users unknown to 
    ! this switch to use this code (for now)
    UseExperimentalExtAlgorithm = .false.

    ! use a proper calculation from p and T when derivinf molecular 
    ! extinction (and backscatter)
    MolExt_Method  = MolExt_Method_Proper

  end subroutine Init_L2B_Proc_Settings
    !  #]
  subroutine Delete_L2B_Proc_Settings(L2BP_Settings)
    !  #[

    TYPE(L2BP_Settings_type), intent(inout) :: L2BP_Settings

    ! deallocate all allocatable/pointer elements in this datastructure
    if (associated(L2BP_Settings%NWPfile)) &
         deallocate(L2BP_Settings%NWPfile)

    call Delete_L2B_AuxPar(L2BP_Settings%L2B_AuxPar)
  
  end subroutine Delete_L2B_Proc_Settings
    !  #]
  subroutine Get_JobOrderName_from_CmdLine(filename_JobOrder,error_flag)
    !  #[ get only the 1st argument (always the JobOrder filename)

    !
    character(len=256), intent(out) :: filename_JobOrder
    integer,            intent(out) :: error_flag

    ! local variables
    integer :: nr_of_arguments
    
    error_flag = no_error
    nr_of_arguments = iargc_aeolus()

    IF (nr_of_arguments .eq. 0) THEN
       CALL Print_Usage_Commandline_Options
       error_flag = request_program_to_stop
       return
    ENDIF

    ! get the 1st argument.
    ! From now on, this must always be the JobOrder filename
    call getarg_aeolus(1,filename_JobOrder)
    ! Now that JobOrder is specified, amend some previous default settings

    call logmsg(log_info,"Get_JobOrderName_from_CmdLine: "//&
                          "using JobOrder file: "//trim(filename_JobOrder))

    do_primary_processing = .true.

  end subroutine Get_JobOrderName_from_CmdLine

    !  #]
  subroutine Get_Additional_Cmdline_Options(L2BP_Settings,error_flag)
    !  #[ get additional commandline options

    ! Modifications:
    !  28-11-2005 P. Poli  If no arguments, display usage and stop.
    !
    TYPE(L2BP_Settings_type), intent(inout) :: L2BP_Settings
    integer,                  intent(out)   :: error_flag

    ! local variables
    integer            :: nr_of_arguments, current_argument_nr
    character(len=256) :: current_argument
    !character(len=256) :: scatteringratio_type
    character(len=256) :: molext_type
    character(len=256) :: partext_type
!    character(len=256) :: classification_type
!    character(len=256) :: error_quantifier_type
    
    error_flag = no_error
    nr_of_arguments = iargc_aeolus()

    ! return if no more arguments are present
    IF (nr_of_arguments .le. 1) return

    current_argument_nr = 1
    
    argloop: DO WHILE (current_argument_nr .lt. nr_of_arguments)
       !  #[ process the given commandline arguments

       !     but ignore the first one (filename_JobOrder)
       !     which was already read by Get_JobOrderName_from_CmdLine()
       current_argument_nr = current_argument_nr+1
       call getarg_aeolus(current_argument_nr,current_argument)

       SELECT CASE (trim(current_argument))
       CASE ("-1","--preprocess") 
          !print *,"running L2BP preprocess step:"
          do_preprocessing_step = .true.
          do_primary_processing = .false.
       CASE ("-2","--primary") 
          !print *,"running L2BP primary code:"
          do_preprocessing_step = .false.
          do_primary_processing = .true.
       CASE("-LogFile")
          !  #[
          current_argument_nr = current_argument_nr+1
          IF (current_argument_nr .gt. nr_of_arguments) THEN
             print *,"filename is missing after -LogFile option !!!"
             CALL Print_Usage_Commandline_Options;
             error_flag = error_cmdline_options
             return
          END IF
          call getarg_aeolus(current_argument_nr,L2BP_Settings%LogFilename)
          print *,'using LogFile: '//trim(L2BP_Settings%LogFilename)
          !  #]
       CASE("-NWPmetaDataFile")
          !  #[
          current_argument_nr = current_argument_nr+1
          IF (current_argument_nr .gt. nr_of_arguments) THEN
             print *,"filename is missing after -NWPmetaDataFile option !!!"
             CALL Print_Usage_Commandline_Options;
             error_flag = error_cmdline_options
             return
          END IF
          call getarg_aeolus(current_argument_nr,NWPMetaDataFilename)
          print *,'using NWP metadata input/outputfile: '
          print *,trim(NWPmetaDataFilename)
          !  #]
       CASE("-MolExt")
          !  #[
          current_argument_nr = current_argument_nr+1
          IF (current_argument_nr .gt. nr_of_arguments) THEN
             print *,"type is missing after -MolExt option !!!"
             CALL Print_Usage_Commandline_Options;
             error_flag = error_cmdline_options
             return
          END IF
          call getarg_aeolus(current_argument_nr,molext_type)
          select case (molext_type)
          case("Parametrised")
             MolExt_Method = MolExt_Method_Parametrised
          case("Proper")
             MolExt_Method = MolExt_Method_Proper
          case default
             print *,"unrecognised MolExt method: ",&
                  trim(molext_type)
             CALL Print_Usage_Commandline_Options;
             error_flag = error_cmdline_options
             return
          end select
          print *,'using MolExt method: ',trim(molext_type)
          !  #]
       CASE("-PartExt")
          !  #[
          current_argument_nr = current_argument_nr+1
          IF (current_argument_nr .gt. nr_of_arguments) THEN
             print *,"type is missing after -PartExt option !!!"
             CALL Print_Usage_Commandline_Options;
             error_flag = error_cmdline_options
             return
          END IF
          call getarg_aeolus(current_argument_nr,partext_type)
          select case (partext_type)
          case("Old")
             PartExt_Method = PartExt_Method_Old
          case("Iterative")
             PartExt_Method = PartExt_Method_Iterative
          case default
             print *,"unrecognised PartExt method: ",&
                  trim(partext_type)
             CALL Print_Usage_Commandline_Options;
             error_flag = error_cmdline_options
             return
          end select
          print *,'using PartExt method: ',trim(partext_type)
          !  #]
       CASE("-ExpExtAlgo")
          !  #[
          UseExperimentalExtAlgorithm = .true.
          print *,'using experimental KNMI Extinction algorithm'
          !  #]
       CASE ("-uselite")
          use_lite_test_data = .true.
          print *,"using Lite test data"
       CASE ("-dontuselite")
          use_lite_test_data = .false.
          print *,"NOT using Lite test data"
       CASE ("-usedummylatlon")
          use_dummy_lat_lon  = .true.
          print *,"using dummy lat-lon values"
       CASE ("-dontusedummylatlon")
          use_dummy_lat_lon  = .false.
          print *,"NOT using dummy lat-lon values"
       CASE ("-h","--help") 
          CALL Print_Usage_Commandline_Options;
          error_flag = request_program_to_stop
          return
       CASE DEFAULT
          !  #[
          CALL Print_Usage_Commandline_Options;
          print *,"============================================"
          print *,"unrecognised commandline option: ",trim(current_argument)
          IF (current_argument(1:1) .ne. '-') THEN
             print *,"This argument does not start with: -"
             print *,"Maybe you forgot the - character to indicate that this"
             print *,"is an option, or maybe you supplied the wrong number"
             print *,"of items to an option."
             print *,"Remind also that in case of using the JobOrder file"
             print *,"no additional options are allowed!"
          END IF
          print *,"============================================"
          error_flag = error_cmdline_options
          return
          !  #]
       END SELECT

       !  #]
    END DO argloop

  end subroutine Get_Additional_Cmdline_Options
    !  #]
  subroutine Set_Derived_L2B_Proc_Settings(L2BP_Settings,error_flag)
    !  #[
    TYPE(L2BP_Settings_type), intent(in)  :: L2BP_Settings
    integer,                  intent(out) :: error_flag

    error_flag = no_error

    IF (.not. associated(L2BP_Settings%L2B_AuxPar%Data)) THEN
       call logmsg(log_error,&
            "Error in Set_Derived_L2B_Proc_Settings: "//&
            "the pointer L2BP_Settings%L2B_AuxPar%Data "//&
            "seems not associated.")
       error_flag = error_programming
       return
    END IF

    IF (.not. do_preprocessing_step .and. .not. do_primary_processing) THEN
       ! if no separate processing step was selected, just do all steps
       do_preprocessing_step = .true.
       do_primary_processing = .true.
    END IF

    Extinction_Needed       = .false.
    select case (L2BP_Settings%L2B_AuxPar%Data%WVM_Params%&
                    Classification_Params%Classification_Type_Mie)
    case(Class_Ext_Threshold,Class_Cal_Mie_Threshold)
       Extinction_Needed       = .true.
    case(Class_Backscat_Ratio)
       ! explicitely state that we need the backscatter ratio
       ! even though it already is set to true
       BackscatterRatio_Needed = .true.
    end select

    select case (L2BP_Settings%L2B_AuxPar%Data%WVM_Params%&
                    Classification_Params%Classification_Type_Rayleigh)
    case(Class_Ext_Threshold,Class_Cal_Mie_Threshold)
       Extinction_Needed       = .true.
    case(Class_Backscat_Ratio)
       ! explicitely state that we need the backscatter ratio
       ! even though it already is set to true
       BackscatterRatio_Needed = .true.
    end select

    ! a bit of extra logic to:
    ! -switch on the extinction calculation when Gert-Jan uses the "-ExpExtAlgo" flag
    ! -disable the use of this algorithm for ordinary users switching on Class_Ext_Threshold

    IF (L2BP_Settings%L2B_AuxPar%Data%WVM_Params%&
            Classification_Params%Classification_Type_Rayleigh .eq. Class_Ext_Threshold) THEN
       IF (UseExperimentalExtAlgorithm) THEN
          Extinction_Needed = .true.
       ELSE
          error_flag = error_programming
          call logmsg(log_error,"Set_Derived_L2B_Proc_Settings: Classification_Type_Rayleigh="//&
                                "Class_Ext_Threshold is not yet fully implemented and validated"//&
                                " and cannot yet be used")
          return
       END IF
    END IF

    IF (L2BP_Settings%L2B_AuxPar%Data%WVM_Params%&
            Classification_Params%Classification_Type_Mie .eq. Class_Ext_Threshold) THEN
       IF (UseExperimentalExtAlgorithm) THEN
          Extinction_Needed = .true.
       ELSE
          error_flag = error_programming
          call logmsg(log_error,"Set_Derived_L2B_Proc_Settings: Classification_Type_Mie="//&
                                "Class_Ext_Threshold is not yet fully implemented and validated"//&
                                " and cannot yet be used")
          return
       END IF
    END IF

  end subroutine Set_Derived_L2B_Proc_Settings
    !  #]
  subroutine Print_L2B_Proc_Settings(L2BP_Settings,prefix,error_flag)
    !  #[
    TYPE(L2BP_Settings_type), intent(in)  :: L2BP_Settings    ! input
    character(len=*),         intent(in)  :: prefix           ! input
    integer,                  intent(out) :: error_flag       ! output

    call AsciiWrite_L2B_Proc_Settings(L2BP_Settings,fileunit_stdout,prefix,&
                                      error_flag)

  end subroutine Print_L2B_Proc_Settings
    !  #]
  subroutine AsciiWrite_L2B_Proc_Settings(L2BP_Settings,fileunit,prefix,&
                                          error_flag)
    !  #[

    TYPE(L2BP_Settings_type), intent(in)  :: L2BP_Settings    ! input
    integer,                  intent(in)  :: fileunit         ! input
    character(len=*),         intent(in)  :: prefix           ! input
    integer,                  intent(out) :: error_flag       ! output

    ! local variable
    !character(len=25)           :: str_FAdoxy
    character(len=256)                 :: tmp_txt
    integer                            :: i

    ! init
    error_flag = no_error

    write(fileunit,"(a)",err=999) trim(prefix)//"%Stdout_Log_Level   = "//&
         trim(L2BP_Settings%Stdout_Log_Level)
    write(fileunit,"(a)",err=999) trim(prefix)//"%Stderr_Log_Level   = "//&
         trim(L2BP_Settings%Stderr_Log_Level)
    write(fileunit,"(a)",err=999) trim(prefix)//"%LogFilename        = "//&
         trim(L2BP_Settings%LogFilename)

    call AsciiWriteFileProperties(L2BP_Settings%L1Bfile,fileunit,&
         prefix=trim(prefix)//"%L1Bfile",error_flag=error_flag)

    write(fileunit,"(a)",err=999) trim(prefix)//"%L2Bresultdir       = "//&
         trim(L2BP_Settings%L2Bresultdir)
    write(fileunit,"(a)",err=999) trim(prefix)//"%RBCfilename        = "//&
         trim(L2BP_Settings%RBCfilename)
    write(fileunit,"(a)",err=999) trim(prefix)//"%L2BAuxPar_filename = "//&
         trim(L2BP_Settings%L2BAuxPar_filename)
    write(fileunit,"(a)",err=999) trim(prefix)//"%AuxClm_filename    = "//&
         trim(L2BP_Settings%AuxClm_filename)
    write(fileunit,"(a)",err=999) trim(prefix)//"%AuxCal_filename    = "//&
         trim(L2BP_Settings%AuxCal_filename)
    write(tmp_txt,"(i6)",err=998) L2BP_Settings%NumNWPfiles
    write(fileunit,"(a)",err=999) trim(prefix)//"%NumNWPfiles        = "//&
         trim(adjustl(tmp_txt))
    if (associated(L2BP_Settings%NWPfile)) then
       do i=1,L2BP_Settings%NumNWPfiles
          write(tmp_txt,"(i6)",err=998) i
          call AsciiWriteFileProperties(L2BP_Settings%NWPfile(i),fileunit,&
               prefix=trim(prefix)//"%NWPfile("//trim(adjustl(tmp_txt))//")",&
               error_flag=error_flag)
       end do
    else
       write(fileunit,"(a)",err=999) trim(prefix)//"%NWPfilename(:)"//&
            " [seems not associated]"
    end if
    write(fileunit,"(a)",err=999) trim(prefix)//"%System             = "//&
         trim(L2BP_Settings%System)
    write(fileunit,"(a)",err=999) trim(prefix)//"%Creator            = "//&
         trim(L2BP_Settings%Creator)
    write(fileunit,"(a)",err=999) trim(prefix)//"%Creator_Version    = "//&
         trim(L2BP_Settings%Creator_Version)
    write(fileunit,"(a)",err=999) trim(prefix)//"%Notes              = "//&
         trim(L2BP_Settings%Notes)
    write(fileunit,"(a)",err=999) trim(prefix)//"%File_Class         = "//&
         trim(L2BP_Settings%File_Class)

    call AsciiWrite_L2B_AuxPar(L2BP_Settings%L2B_AuxPar,fileunit,&
                               trim(prefix)//"%L2B_AuxPar",&
                               error_flag)
    IF(error_flag .ne. no_error) return
    
    print *,"old style items:"

    ! print the MolExt_method for determining extinction by molecules
    select case (MolExt_Method)
       !  #[
    case(MolExt_Method_Undefined)
       write(fileunit,"(a)",err=999) &
            "MolExt_Method                = MolExt_Method_Undefined"
    case(MolExt_Method_Parametrised)
       write(fileunit,"(a)",err=999) &
            "MolExt_Method                = MolExt_Method_Parametrised"
    case(MolExt_Method_Proper)
       write(fileunit,"(a)",err=999) &
            "MolExt_Method                = MolExt_Method_Proper"
    case default
       write(fileunit,"(a)",err=999) &
            "MolExt_Method                = ??unknown type??"
       write(fileunit,"(a)",err=999) &
            "please update the printing routine in L2B_Proc_Settings..."
    end select
       !  #]

    ! print the PartExt_method for determining extinction by particles
    select case (PartExt_Method)
       !  #[
    case(PartExt_Method_Undefined)
       write(fileunit,"(a)",err=999) &
            "PartExt_Method               = PartExt_Method_Undefined"
    case(PartExt_Method_Old)
       write(fileunit,"(a)",err=999) &
            "PartExt_Method               = PartExt_Method_Old"
    case(PartExt_Method_Iterative)
       write(fileunit,"(a)",err=999) &
            "PartExt_Method               = PartExt_Method_Iterative"
    case default
       write(fileunit,"(a)",err=999) &
            "PartExt_Method                = ??unknown type??"
       write(fileunit,"(a)",err=999) &
            "please update the printing routine in L2B_Proc_Settings..."
    end select
       !  #]

    IF (use_dummy_lat_lon) THEN
       write(fileunit,"(a)",err=999) "use_dummy_lat_lon         = .true."
    ELSE
       write(fileunit,"(a)",err=999) "use_dummy_lat_lon         = .false."
    END IF

    IF (use_lite_test_data) THEN
       write(fileunit,"(a)",err=999) "use_lite_test_data        = .true."
    ELSE
       write(fileunit,"(a)",err=999) "use_lite_test_data        = .false."
    END IF

    ! derived settings
    IF (do_preprocessing_step) THEN
       write(fileunit,"(a)",err=999) "do_preprocessing_step     = .true."
    ELSE
       write(fileunit,"(a)",err=999) "do_preprocessing_step     = .false."
    END IF

    IF (do_primary_processing) THEN
       write(fileunit,"(a)",err=999) "do_primary_processing     = .true."
    ELSE
       write(fileunit,"(a)",err=999) "do_primary_processing     = .false."
    END IF

    IF (Extinction_Needed) THEN
       write(fileunit,"(a)",err=999) "Extinction_Needed         = .true."
    ELSE
       write(fileunit,"(a)",err=999) "Extinction_Needed         = .false."
    END IF

    IF (BackscatterRatio_Needed) THEN
       write(fileunit,"(a)",err=999) "BackscatterRatio_Needed   = .true."
    ELSE
       write(fileunit,"(a)",err=999) "BackscatterRatio_Needed   = .false."
    END IF

    return

998 call logmsg(log_error,"ERROR in AsciiWrite_L2B_Proc_Settings "//&
                          "error while converting value NumNWPfiles "//&
                          "to a string")
    error_flag = error_writing_file
    return

999 call logmsg(log_error,"ERROR in AsciiWrite_L2B_Proc_Settings "//&
                          "error while writing to outputfile")
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2B_Proc_Settings
    !  #]
  subroutine Print_Usage_Commandline_Options
    !  #[
    print *,""
    print *,"Possible commandline options for the L2B_processor main program:"
    print *,""
    print *,"L2B_processor <JobOrderFilename.xml>"
    print *,"   "
    print *,"   or"
    print *,"   "
    print *,"L2B_processor"
    print *,"   [<JobOrderFilename.xml>]"
    print *,"   [-h/--help      ]"
    print *,"   [-1/--preprocess]"
    print *,"   [-2/--primary   ]"
    print *,"   [-LogFile <filename>]"
    print *,"   [-NWPmetaDataFile <filename>]"
    print *,"   [-MolExt         [Parametrised|Proper]]"
    print *,"   [-PartExt        [Old|Iterative]]"
! for now undocumented (deliberately)
!    print *,"   [-ExpExtAlgo]"
    print *,"   [-uselite/-dontuselite]"
    print *,"   [-usedummylatlon/-dontusedummylatlon]"
    print *,""
  end subroutine Print_Usage_Commandline_Options
  !  #]
  subroutine SetFileTimeInterval(File,Input,error_flag)
    !  #[
    ! the file properties struct, with filled filename, but blank start/stop
    type(FileProperties), intent(inout) :: File       ! updated
    ! the input block from the Joborder file
    type(Input_type),     intent(in)    :: Input      ! input
    integer,              intent(out)   :: error_flag ! output

    ! local variable
    integer :: i,nfiles

    ! init
    error_flag = no_error

    ! remember that in cases where multiple input files of the same
    ! type are allowed, you have to match the Time_Interval with the
    ! File_Name to ensure they refer to the same file.

    nfiles = size(Input%File_Name)
    do i=1,nfiles ! loop over time intervals
       ! double check:
       ! see if this timeinterval is associated with this
       ! aux.met.file. If each file has just one
       ! time interval, then you would expect them to be in the
       ! same order, but there is no requirement on this, so
       ! it might just happen that they are reversed ...
       if ( trim(Input%Time_Interval(i)%File_Name) .eq. &
            trim(File%name) ) then
          if ( DateTimeMissing(File%start) .and. &
               DateTimeMissing(File%stop)        ) then
             File%start = Convert_JobOrd_to_Num_DateTime(&
                                        Input%Time_Interval(i)%Start )
             File%stop  = Convert_JobOrd_to_Num_DateTime(&
                                        Input%Time_Interval(i)%Stop )
             IF (CheckIfFatalErrorOccurred()) THEN
                ! the logmsg() has already been called from within the
                ! Convert_JobOrd_to_Num_DateTime() above if this
                ! point is reached
                error_flag = GetFatalError()
                return
             END IF
          else
             ! Note that multiple time intervals on the same input file are
             ! allowed by the JobOrder format, but we have not implemented 
             ! this use, so warn the user if he tries to do this....
             call logmsg(log_error,"ERROR in SetFileTimeInterval: "//&
                  "only 1 timeinterval per input file has been "//&
                  "implemented, but the JobOrder input section "//&
                  "contains more than 1 timeinterval for file: "//&
                  trim(File%name) )
             error_flag = error_reading_file
             return
          end if
       end if
 
    end do ! loop over time intervals in JobOrder file

    ! extra check, to see if this file has a timerange defined
    if ( DateTimeMissing(File%start) .or. DateTimeMissing(File%stop)   ) then
       call logmsg(log_error,"ERROR in Copy_JobOrderData_Items:"//&
            " no valid timerange was specified for file: "//trim(File%name))
       error_flag = error_reading_file
       return
    end if

  end subroutine SetFileTimeInterval
    !  #]
  subroutine Copy_JobOrderData_Items(JobOrderData,L2BP_Settings,error_flag)
    !  #[

    ! copy items from the JobOrderData to the variables in the
    ! L2B_Proc_Settings module, so the processor will use them
    TYPE(JobOrderData_type),  intent(in)    :: JobOrderData
    TYPE(L2BP_Settings_type), intent(inout) :: L2BP_Settings
    integer,                  intent(out)   :: error_flag

    ! local variables
    integer :: i,j,s,nfiles
    logical :: L1Bfilename_set,     L2BAuxPar_filename_set 
    logical :: NWPfilename_set,     RBCfilename_set
    logical :: AuxClm_filename_set, AuxCal_filename_set
    logical :: L2Bresultdir_set
    logical :: file_missing
    character(len=256) :: missing_files
    character(len=256) :: TempCreatorString
    integer :: AllocStatus

    ! init
    error_flag = no_error

    ! these global settings define what log level is used in the logging module
    L2BP_Settings%Stdout_Log_Level = &
         JobOrderData%Global_Configuration%Stdout_Log_Level
    L2BP_Settings%Stderr_Log_Level = &
         JobOrderData%Global_Configuration%Stderr_Log_Level

    ! these switches are used to make sure each file/dirname is only set once
    ! (a user might try for example to input multiple L1B files using 
    ! the JobOrder mechanism, but our processor only handles one, 
    ! so issue a proper warning if this happens)

    ! input files
    L1Bfilename_set        = .false.
    L2BAuxPar_filename_set = .false.
    AuxClm_filename_set    = .false.
    AuxCal_filename_set    = .false.
    NWPfilename_set        = .false.
    RBCfilename_set        = .false.

    ! output dir
    L2Bresultdir_set       = .false.
    
    ! a switch used to signal trouble
    file_missing           = .false.

    ! loop over the inputs and fill the needed filename variables
    ! also check if the files are present, and terminate the program
    ! only after the loop, to make sure all missing files get reported.
    s = size(JobOrderData%Proc_Instructions%Input)
    DO i=1,s
       !  #[
       select case (JobOrderData%Proc_Instructions%Input(i)%File_Type)
          ! meaning of these codes (see L1B_IODD, sec. 3.1)
          ! ALD_<u/c flag>_<proc_stage>_<prod_id>
          ! ALD: denotes data from the Aladin instrument
          ! u/c flag: U (unconsolidated)
          !           C (consolidated)
          ! proc_stage: N (near-real-time, or nominal), 
          !             R (reprocessing) (or B for backlog processing?)
          !             T (test)
          ! prod_id: 0, 1A, 1B (L1B input file)
       case("ALD_U_N_1B","ALD_U_R_1B","ALD_U_T_1B",&
            "ALD_C_N_1B","ALD_C_R_1B","ALD_C_T_1B")
          !  #[ handle the L1B input file
            nfiles = size(JobOrderData%Proc_Instructions%Input(i)%File_Name)
            IF (nfiles .gt. 1) THEN
               call logmsg(log_error,"ERROR in Copy_JobOrderData_Items:"//&
                    " only 1 L1B file can be processed at a time, but the "//&
                    " JobOrder input section contains more than 1 file!")
               error_flag = error_reading_file
               return
            END IF
            IF (.not. L1Bfilename_set) THEN
               L1Bfilename_set = .true.
               call InitFileProperties(L2BP_Settings%L1Bfile)
               L2BP_Settings%L1Bfile%name = &
                    JobOrderData%Proc_Instructions%Input(i)%File_Name(1)
               ! check if the file is present
               call check_ee_file_presence(L2BP_Settings%L1Bfile%name,&
                                           file_missing)
               call SetFileTimeInterval(L2BP_Settings%L1Bfile,&
                              JobOrderData%Proc_Instructions%Input(i),&
                              error_flag)
               if (error_flag .ne. no_error) return
            ELSE
               call logmsg(log_error,"ERROR in Copy_JobOrderData_Items:"//&
                    " only 1 L1B file can be processed at a time, but the "//&
                    " JobOrder contains more than 1 input section for"//&
                    " this filetype!")
               error_flag = error_reading_file
               return
            END IF
            !  #]
       case("AUX_PAR_2B")
          !  #[ handle the L2B Aux. Par. input file
            nfiles = size(JobOrderData%Proc_Instructions%Input(i)%File_Name)
            IF (nfiles .gt. 1) THEN
               call logmsg(log_error,"ERROR in Copy_JobOrderData_Items:"//&
                    " only 1 AUX_PAR_2B file can be used, but the "//&
                    " JobOrder input sections contains more than 1 file!")
               error_flag = error_reading_file
               return
            END IF
            IF (.not. L2BAuxPar_filename_set) THEN
               L2BAuxPar_filename_set = .true.
               ! this switch is temporary and will be removed in the end!
               L2BP_Settings%L2BAuxPar_filename = &
                    JobOrderData%Proc_Instructions%Input(i)%File_Name(1)

               ! check if the file is present
               call check_ee_file_presence(L2BP_Settings%L2BAuxPar_filename,&
                                           file_missing)
            ELSE
               call logmsg(log_error,"ERROR in Copy_JobOrderData_Items:"//&
                    " only 1 AUX_PAR_2B file can be used at a time, but "//&
                    " the JobOrder contains more than 1 input section for"//&
                    " this filetype!")
               error_flag = error_reading_file
               return
            END IF
            !  #]
       case("AUX_MET_12")  ! new
          !  #[ handle the NWP input file
          nfiles = size(JobOrderData%Proc_Instructions%Input(i)%File_Name)
          IF (.not. NWPfilename_set) THEN
             NWPfilename_set = .true.
             L2BP_Settings%NumNWPfiles = nfiles
             allocate(L2BP_Settings%NWPfile(nfiles),STAT=AllocStatus)
             if (AllocStatus .ne. 0) then
                call logmsg(log_error,"Allocation failure in subroutine "//&
                                      "Copy_JobOrderData_Items()")
                error_flag = error_allocate
                return
             end if
             do j=1,nfiles
                call InitFileProperties(L2BP_Settings%NWPfile(j))
                L2BP_Settings%NWPfile(j)%name = &
                     JobOrderData%Proc_Instructions%Input(i)%File_Name(j)
                ! check if the file is present
                call check_ee_file_presence(L2BP_Settings%NWPfile(j)%name,&
                                            file_missing)
                call SetFileTimeInterval(L2BP_Settings%NWPfile(j),&
                              JobOrderData%Proc_Instructions%Input(i),&
                              error_flag)
                if (error_flag .ne. no_error) return
             end do
          ELSE
             call logmsg(log_warn,"WARNING in Copy_JobOrderData_Items:"//&
                  "just one input block for file type AUX_MET_12 was "//&
                  "expected! ignoring the extra NWP input block ...")
          END IF
          !  #]
       case("AUX_RBC_L2")  ! new
          !  #[ handle the RBC input file
            nfiles = size(JobOrderData%Proc_Instructions%Input(i)%File_Name)
            IF (nfiles .gt. 1) THEN
               call logmsg(log_error,"ERROR in Copy_JobOrderData_Items:"//&
                    " only 1 RBC file can be used at a time, but the "//&
                    " JobOrder input sections contains more than 1 file!")
               error_flag = error_reading_file
               return
            END IF
            IF (.not. RBCfilename_set) THEN
               RBCfilename_set = .true.
               L2BP_Settings%RBCfilename = JobOrderData%Proc_Instructions%&
                                              Input(i)%File_Name(1)
               ! check if the file is present
               call check_ee_file_presence(L2BP_Settings%RBCfilename,&
                                           file_missing)
            ELSE
               call logmsg(log_error,"ERROR in Copy_JobOrderData_Items:"//&
                    " only 1 RBC file can be used, but the "//&
                    " JobOrder contains more than 1 input section for"//&
                    " this filetype!")
               error_flag = error_reading_file
               return
            END IF
            !  #]
       case("AUX_CLM_L2")  ! new
          !  #[ handle the AuxClm input file
            nfiles = size(JobOrderData%Proc_Instructions%Input(i)%File_Name)
            IF (nfiles .gt. 1) THEN
               call logmsg(log_error,"ERROR in Copy_JobOrderData_Items:"//&
                    " only 1 AuxClm file can be used at a time, but the "//&
                    " JobOrder input sections contains more than 1 file!")
               error_flag = error_reading_file
               return
            END IF
            IF (.not. AuxClm_filename_set) THEN
               AuxClm_filename_set = .true.
               L2BP_Settings%AuxClm_filename = JobOrderData%Proc_Instructions%&
                                                    Input(i)%File_Name(1)
               ! check if the file is present
               call check_ee_file_presence(L2BP_Settings%AuxClm_filename,&
                                           file_missing)
            ELSE
               call logmsg(log_error,"ERROR in Copy_JobOrderData_Items:"//&
                    " only 1 AuxClm file can be used, but the "//&
                    " JobOrder contains more than 1 input section for"//&
                    " this filetype!")
               error_flag = error_reading_file
               return
            END IF
            !  #]
       case("AUX_CAL_L2")  ! new
          !  #[ handle the AuxCal input file
            nfiles = size(JobOrderData%Proc_Instructions%Input(i)%File_Name)
            IF (nfiles .gt. 1) THEN
               call logmsg(log_error,"ERROR in Copy_JobOrderData_Items:"//&
                    " only 1 AuxCal file can be used at a time, but the "//&
                    " JobOrder input sections contains more than 1 file!")
               error_flag = error_reading_file
               return
            END IF
            IF (.not. AuxCal_filename_set) THEN
               AuxCal_filename_set = .true.
               L2BP_Settings%AuxCal_filename = JobOrderData%Proc_Instructions%&
                                                    Input(i)%File_Name(1)
               ! check if the file is present
! skipping this test for now, since we have no example file yet
! todo: activate this test once we have this file
!               call check_ee_file_presence(L2BP_Settings%AuxCal_filename,&
!                                           file_missing)
            ELSE
               call logmsg(log_error,"ERROR in Copy_JobOrderData_Items:"//&
                    " only 1 AuxCal file can be used, but the "//&
                    " JobOrder contains more than 1 input section for"//&
                    " this filetype!")
               error_flag = error_reading_file
               return
            END IF
            !  #]
       case default
          call logmsg(log_warn,"WARNING in  Copy_JobOrderData_Items: "//&
               "filetype: "//&
               trim(JobOrderData%Proc_Instructions%Input(i)%File_Type)//&
               " is not recognised as valid input"//&
               " file for the L2BP. Ignoring this file.")
       end select
       !  #]
    END DO

    ! TODO: we could add a check here to see, in case multiple files of
    !       the same filetype occur (may happen for AMD files)
    !       whether the time intervals are overlapping or not.
    !       -first we should decide then if this is allowed or not?
    !       -and whether they should follow each other without gaps?

    ! report an error if any of the defined files is not present or empty
    IF (file_missing) THEN
       call logmsg(log_error,"ERROR in Copy_JobOrderData_Items:"//&
            " one of the input files specified in the joborder is"//&
            " not present, or has size zero !")
       error_flag = error_reading_file
       return
    END IF

    ! report an error if not all required filenames have been defined
    IF (.not. ( L1Bfilename_set     .and. L2BAuxPar_filename_set .and. &
                NWPfilename_set     .and. RBCfilename_set        .and. &
                AuxClm_filename_set  ) )  THEN
!                AuxClm_filename_set .and. AuxCal_filename_set  ) )  THEN
       missing_files = ""
       IF (.not. L1Bfilename_set) &
            missing_files = trim(missing_files)//" L1B "
       IF (.not. L2BAuxPar_filename_set) &
            missing_files = trim(missing_files)//" L2B_AuxPar "
       IF (.not. NWPfilename_set) &
            missing_files = trim(missing_files)//" AMD"
       IF (.not. RBCfilename_set) &
            missing_files = trim(missing_files)//" RBC"
       IF (.not. AuxClm_filename_set) &
            missing_files = trim(missing_files)//" AuxClm"
       IF (.not. AuxCal_filename_set) &
            missing_files = trim(missing_files)//" AuxCal"
       call logmsg(log_error,"ERROR in Copy_JobOrderData_Items: "//&
            "required filetype(s): "//trim(missing_files)//&
            " is/are not defined in the current joborder file!")
       error_flag = error_reading_file
       return
    END IF

    ! loop over the outputs and fill the needed filename variables
    s = size(JobOrderData%Proc_Instructions%Output)
    DO i=1,s
       !  #[
       !print *,"type = ",JobOrderData%Proc_Instructions%Output(i)%File_Type
       select case (JobOrderData%Proc_Instructions%Output(i)%File_Type)
          ! meaning of these codes (see L1B_IODD, sec. 3.1)
          ! ALD_<u/c flag>_<proc_stage>_<prod_id>
          ! ALD: denotes data from the Aladin instrument
          ! u/c flag: U (unconsolidated)
          !           C (consolidated)
          ! proc_stage: N (near-real-time, or nominal), 
          !             R (reprocessing) (or B for backlog processing?)
          !             T (test)
          ! prod_id: 2B (L2B input file)
       case("ALD_U_N_2B","ALD_U_R_2B","ALD_U_T_2B",&
            "ALD_C_N_2B","ALD_C_R_2B","ALD_C_T_2B")
          !  #[ handle the L2B output directory
          IF (.not. L2Bresultdir_set) THEN
             L2Bresultdir_set = .true.
             L2BP_Settings%L2Bresultdir = &
                    JobOrderData%Proc_Instructions%Output(i)%File_Name
          ELSE
             call logmsg(log_error,"ERROR in Copy_JobOrderData_Items:"//&
                  " only 1 L2Bresultdir can be used at a time, but "//&
                  " the JobOrder contains more than 1 output section for"//&
                  " this filetype!")
             error_flag = error_reading_file
             return
          END IF
            !  #]
       case default
       end select
       !  #]
    END DO

    ! fill some other fields that we have to use from the JobOrder file
    ! See SPR 28 for details

    ! the System name to be used in the FH (as requested in SPR 28)
    L2BP_Settings%System = &
         JobOrderData%Global_Configuration%Processing_Station
    ! sanity check 1:
    ! the length of Processing_Station in the JobOrder may be up to 20 chars
    ! but we can only store 19 in the system field, so check for this
    if (len_trim(JobOrderData%Global_Configuration%Processing_Station) .gt. &
         maxlen_system) then
       call logmsg(log_warn,"WARNING in Copy_JoborderData_Items: "//&
                   "length of system string should not exceed "//&
                   "19 characters, but the Processing_Station field from "//&
                   "the JobOrder file seems longer. The string will be "//&
                   "truncated!")
    end if

    ! sanity check 2:
    ! according to the SPR the following strings are allowed for the L2B/C:
    ! L2PF (ECMWF), LTA (Long-term archive), LOCAL (for testing)
    select case (trim(L2BP_Settings%System))
    case(systemstr_ECMWF_IFS)      ! = "L2PF"  ! ok
    case(systemstr_ESRIN_REPR)     ! = "LTA"   ! ok
    case(systemstr_LOCAL_MET_INST) ! = "LOCAL" ! ok
    ! remind that although systemstr_ACMF is defined in fixedheader.F90
    ! that one is currently only used for producing AuxRBC files
    ! so it is not a valid setting for the L2BP itself.
    case default
       call logmsg(log_warn,"WARNING in Copy_JobOrderData_Items: "//&
            "Processing_Station should be one of "//&
            trim(systemstr_ECMWF_IFS)//"/"//&
            trim(systemstr_ESRIN_REPR)//"/"//&
            trim(systemstr_LOCAL_MET_INST)//", "//&
            "but it seems to be: "//trim(L2BP_Settings%System))
    end select

    TempCreatorString = &
         trim(JobOrderData%Global_Configuration%Processor_Name)//"/"//&
         trim(software_TaskName)
    ! sanity check 1:
    if (len_trim(TempCreatorString) .gt. maxlen_creator) then
       call logmsg(log_warn,"WARNING in Copy_JoborderData_Items: "//&
                   "length of creator string should not exceed "//&
                   "40 characters, but the string used to fill it "//&
                   "seems longer, so will be truncated! "//&
                   "String used for filling creator field was: "//&
                   trim(TempCreatorString))
    end if

    L2BP_Settings%Creator = TempCreatorString

    ! this one should always fit
    ! format "XX.XX" with X an integer between 0-9
    ! this format was already checked when reading the JobOrder file
    ! so no need to do it again here.
    ! Note that this version does not have to be identical to the
    ! software version. The processor in JobOrder terms is the chain
    ! of executables (in our case L2BP+RepGen) and may have a different
    ! version number than the executables them selves.
    L2BP_Settings%Creator_Version = & ! maxlen=12
         trim(JobOrderData%Global_Configuration%Version) ! maxlen=5
    ! sanity checks

    ! the new rule for setting the fileclass is as follows (see SPR 28)
    ! if the Test flag is set to true, File_Class="TEST",
    ! otherwise copy the content of the Order_Type field.
    select case (to_lowercase(trim(JobOrderData%Global_Configuration%Test)))
    case("false")
       L2BP_Settings%File_Class = &
            trim(JobOrderData%Global_Configuration%Order_Type)
    case("true")
       L2BP_Settings%File_Class = "TEST"
    case default
       ! This case should never occur, since the correct filling of this
       ! field is already tested when the JobOrder file is read
       call logmsg(log_warn,"WARNING in Copy_JobOrderData_Items: "//&
                   "the field JobOrderData%Global_Configuration%Test "//&
                   "should be either TRUE or FALSE, but it seems to be: "//&
                   trim(JobOrderData%Global_Configuration%Test))
    end select

  end subroutine Copy_JobOrderData_Items
  !  #]
  subroutine check_ee_file_presence(filename_in,file_missing)
    !  #[ check whether an EE file is present or not
    character(len=*), intent(in)    :: filename_in  ! input
    logical,          intent(inout) :: file_missing ! input/output

    ! local variables
    character(len=256) :: filename_hdr
    character(len=256) :: filename_dbl
    character(len=256) :: filename_eef
    character(len=4)   :: extension
    integer            :: filename_len

    filename_hdr(:) = ' '
    filename_dbl(:) = ' '
    filename_eef(:) = ' '
    filename_len = len_trim(filename_in)
    extension = filename_in(filename_len-3:filename_len)

    select case (extension)
    case(".EEF")
       filename_eef = filename_in
       call check_file_presence(filename_eef,file_missing)
    case(".DBL")
       filename_hdr(1:filename_len-4) = filename_in(1:filename_len-4)
       filename_hdr(filename_len-3:filename_len) = ".HDR"
       filename_dbl = filename_in
       call check_file_presence(filename_hdr,file_missing)
       call check_file_presence(filename_dbl,file_missing)
    case default
       ! the file has no (known) extension, so add the default .DBL/.HDR
       filename_hdr = trim(filename_in)//".HDR"
       filename_dbl = trim(filename_in)//".DBL"
       call check_file_presence(filename_hdr,file_missing)
       call check_file_presence(filename_dbl,file_missing)
    end select

  END subroutine check_ee_file_presence
    !  #]
  subroutine check_file_presence(filename,file_missing)
    !  #[ check whether an physical file is present or not
    character(len=*), intent(in)    :: filename
    logical,          intent(inout) :: file_missing

    ! local variables
    integer :: filesize

    filesize = get_filesize(filename)
    IF (filesize .lt. 0) THEN
       call logmsg(log_error,"Input file: "//trim(filename)//" not found !")
       file_missing = .true.
    END IF
    IF (filesize .eq. 0) THEN
       call logmsg(log_error,"Input file: "//trim(filename)//" seems empty !")
       file_missing = .true.
    END IF
  END subroutine check_file_presence
    !  #]
  !-----------------------
end module L2B_Proc_Settings
