! todo: remove all error prints, replace by logmsg()

module L2B_Rayleigh_MDS
  !  #[ Documentation
  !
  !     a module to handle the L2B Rayleigh Measurement dataset
  !     as defined in TN2.1, version 1.1 (14-2-2005), 
  !     sec.5.6, p. 23, tables 19-23
  !
  !     Written by J. de Kloe, KNMI 
  !     created:     06-10-2005
  !
  !     Modifications:
  !     30-Nov-2005 P. Poli Rayleigh MDS as in TN2.1 v1.2 06-09-2005
  !     30-Nov-2005 P. Poli Added Update_L2B_Rayleigh_SummaryVars
  !     01-Dec-2005 P. Poli Added Write_L2B_Rayleigh_MDS, 
  !                         Write_L2B_Rayleigh_Profile,
  !                         Write L2B_Rayleigh_Height_Bin
  !     19-Dec-2005 J. de Kloe small correction in USE statements
  !     06-Jan-2006 J. de Kloe adapted to seconds of type r8_
  !     06-Jan-2006 J. de Kloe added the Get_ByteSize routines
  !     06-Jan-2006 J. de Kloe added the new DateTime type
  !     10-Jan-2006 P. Poli    Added reading routines (NOT tested yet)
  !     26-Jan-2006 P. Poli    Removed N_Obs_Max from Rayleigh_MDS
  !     12-May-2006 D. Tan     Null use of ParKind1 module
  !     23-Feb-2007 J. de Kloe reintroduced the old AsciiWrite routines
  !                            originally written for the L2BtoTXT
  !                            branch of 15-Mar-2006. Removed all 
  !                            fillwithzero code, added N_*_to_write
  !                            variables to the write routines that need it
  !     28-Feb-2007 D. Tan     Re-introduce one print in Write_L2B_Rayleigh_MDS
  !     02-Mar-2007 J. de Kloe adapted to IODD change, and some cleaning up
  !     12-Mar-2007 J. de Kloe updated GetByteSize routine
  !     15-Mar-2007 J. de Kloe removed use of SummaryVars
  !     23-Apr-2007 J. de Kloe phase out fillwithzero code
  !     09-Jan-2008 J. de Kloe added DateTimeMissing tests
  !     16-Jan-2008 J. de Kloe phase out integer kind i_  
  !
  !  #]
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE EE_CFI_Datatypes, only:            c_bytelen_Bool,   &
       IntAuc, maxlen_IntAuc, SetIntAuc, c_bytelen_IntAuc, &
       IntAc,                            c_bytelen_IntAc,  &
       IntAs,  maxlen_IntAs,  SetIntAs,  c_bytelen_IntAs,  &
       IntAus, maxlen_IntAus, SetIntAus, c_bytelen_IntAus, & 
       IntAul, maxlen_IntAul, SetIntAul, c_bytelen_IntAul, &
       IntAl,                            c_bytelen_IntAl,  &
       DateTimeType, InitDateTime, maxlen_DateTime_XML, maxlen_DateTime_KVT, &
       c_bytelen_DateTime, Construct_XML_DateTime, Construct_KVT_DateTime, &
       write_DateTime_to_bin_file, read_DateTime_from_bin_file, &
       write_IntAuc_to_bin_file, read_IntAuc_from_bin_file, &
       write_IntAs_to_bin_file,  read_IntAs_from_bin_file,  &
       write_IntAus_to_bin_file, read_IntAus_from_bin_file, &
       write_IntAul_to_bin_file, read_IntAul_from_bin_file, &
       write_SpareBytes_to_bin_file, read_SpareBytes_from_bin_file, &
       DateTimeMissing

  USE L2B_parameters, only: Obs_Type_undefined, Obs_Type_cloudy_returns, &
                            Obs_Type_clear_returns, GetObsTypeText
  USE Meas_Map, only: Meas_Map_type, Init_Meas_Map, Print_Meas_Map, &
                      AsciiWrite_Meas_Map, Delete_Meas_Map, &
                      Count_Nr_Points_On_Level, &
                      Write_Meas_Map, Read_Meas_Map, &
                      Get_ByteSize_Meas_Map
  USE LunManager, only: fileunit_stdout
  USE ErrorHandler, only: no_error, error_programming, error_allocate, &
                          error_writing_file
  USE DirectBinaryIO, only: write_logical1_to_bin_file, &
                            read_logical1_from_bin_file
  !  #]
  !  #[ type definitions

  implicit none

  TYPE Rayleigh_Height_Bin_type
     !  #[
     ! see sec. 5.5, table 22, p. 22, of TN2.1, vers. 1.2, 06-09-2005
     logical         :: Validity_Flag                   ! .true. = valid = 1
     integer(IntAs)  :: Rayleigh_Wind_Velocity          ! [cm/s]
     integer(IntAs)  :: Rayleigh_Wind_to_Pressure       ! [10^-6 m/s/Pa]
     integer(IntAs)  :: Rayleigh_Wind_to_Temperature    ! [cm/s/K]
     integer(IntAs)  :: Rayleigh_Wind_to_Backscat_Ratio ! [cm/s]
     integer(IntAul) :: Reference_Pressure              ! [Pa]
     integer(IntAus) :: Reference_Temperature           ! [10^-2 K]
     integer(IntAul) :: Reference_Backscatter_Ratio     ! [10^-6]
     integer(IntAus) :: Rayleigh_Error_Quantifier       ! [cm/s]
     integer(IntAul) :: Integration_Length              ! [m]
!                       1234567890123456789012345678901
  END TYPE Rayleigh_Height_Bin_type
     !  #]

  TYPE Rayleigh_Profile_type
     !  #[
     ! added by JdK 7-11-2005. We need this to remember if this was
     ! a cloud or no-cloud profile (or any other type we define)
     integer(IntAuc) ::Obs_Type
     !
     ! Spare 1 (36 bytes reserved for Obs_Type_String)
     !
     ! see sec. 5.5, table 21, p. 22, of TN2.1, vers. 1.2, 06-09-2005
     ! this profile is an array of size Num_Rangegates
     TYPE(Rayleigh_Height_Bin_type), dimension(:), pointer :: Height_Bin
  END TYPE Rayleigh_Profile_type
     !  #]

  TYPE L2B_Rayleigh_MDS_type 
     !  #[
     ! see sec. 5.5, table 20, p. 21/22, of TN2.1, vers. 1.2, 06-09-2005
     type(DateTimeType)  :: Start_of_Obs_DateTime    ! UTC
     integer(IntAs)      :: N_Meas       ! nr of Measurements in this product
     integer(IntAs)      :: N_Obs_Rayleigh_Actual ! actual nr Obs. for this BRC
! Poli: following telecon of 25-01-2006, it is decided that N_Obs_*_Actual
! refers to the number of Mie and Rayleigh profiles with meaningful values.
! Each BRC contains M_Mie and M_Rayleigh profiles (these two numbers are fixed
! for a given file), but in some BRCs there might be less than M_Mie profiles;
! in such case N_Obs_Mie_Actual is less than M_Mie. Same story for Rayleigh.
!   - when initialising, N_Obs_Rayleigh_Actual is set to zero. The structure
!     is dimensioned using M_Rayleigh.
!   - after each BRC is processed, a call to Handle_L2BHeaders will set
!     N_Obs_Rayleigh_Actual to the proper value
!   - when writing, M_Rayleigh profiles are written. The number N_Obs_Rayleigh_Actual
!     is saved so that the user will know when reading the BRC that
!     only the first N_Obs_Rayleigh_Actual profiles are meaningful.
!
!   - when reading, the structure should be dimensioned first to M_Rayleigh.
!     ! Remark: the nr of generated observations may vary from one BRC to 
!     ! the next but it was agreed that for one product file the maximum
!     ! N_Obs_max should be used for dimensioning the datastructures 
!     ! that are written to the output file:
!     !         N_Obs_max = max(N_Obs_Actual(first_BRC:last_BRC))
!     ! (see email of David Tan, 20051012, 19.44 utc)
!     ! So: -when initialising N_Obs_Max is used, N_Obs_Actual is set to 0
!     !     -when filling the datastructure N_Obs_Actual is updated
!     !      everytime an observation is generated
!     !     -when printing only N_Obs_Actual is used
!     !     -when writing to file N_Obs_Actual is used, 
!     !      and also all not used Observations are written to file
     integer(IntAs)      :: P ! total nr of laser pulses in one Meas.
     ! a datastructure to store which measurements were used for
     ! creating which observation (the size will be N_Meas*N_Range_Bins)
     ! and what weight was used
     TYPE(Meas_Map_type), pointer :: Map_of_L1B_Measurements_Used
     !
! this next array is not implemented. In stead I added the Weight element
! to the L1B_Meas_Map_Bin_type type which save a lot of double coding
!
! an array of size N of rows containing the map weights
! TYPE(L1B_Meas_Weight_row_type), dimension(:), &
!       pointer :: Map_of_L1B_Measurements_Used
     !
     ! an array for N_Obs_Max profiles for this BRC
     TYPE(Rayleigh_Profile_type), dimension(:), &
          pointer :: Rayleigh_Profile
     !              1234567890123456789012345678901
  END TYPE L2B_Rayleigh_MDS_type
     !  #]

  !  #]
contains
  !---------------------------------------------------------------------
  subroutine Init_Rayleigh_Height_Bin(Rayleigh_Height_Bin)
    !  #[
    TYPE(Rayleigh_Height_Bin_type), intent(out) :: Rayleigh_Height_Bin

    Rayleigh_Height_Bin%Validity_Flag                   =.false.
    Rayleigh_Height_Bin%Rayleigh_Wind_Velocity          = 0
    Rayleigh_Height_Bin%Rayleigh_Wind_to_Pressure       = 0
    Rayleigh_Height_Bin%Rayleigh_Wind_to_Temperature    = 0
    Rayleigh_Height_Bin%Rayleigh_Wind_to_Backscat_Ratio = 0
    Rayleigh_Height_Bin%Reference_Pressure              = 0
    Rayleigh_Height_Bin%Reference_Temperature           = 0
    Rayleigh_Height_Bin%Reference_Backscatter_Ratio     = 0
    Rayleigh_Height_Bin%Rayleigh_Error_Quantifier       = 0
    Rayleigh_Height_Bin%Integration_Length              = 0

  end subroutine Init_Rayleigh_Height_Bin
    !  #]
  subroutine Print_Rayleigh_Height_Bin(Rayleigh_Height_Bin, error_flag)
    !  #[
    TYPE(Rayleigh_Height_Bin_type), intent(in)  :: Rayleigh_Height_Bin
    integer,                        intent(out) :: error_flag    ! output

    ! local variables
    character(len=maxlen_IntAs)  :: str_IntAs
    character(len=maxlen_IntAus) :: str_IntAus
    character(len=maxlen_IntAul) :: str_IntAul

    error_flag = no_error

    if (Rayleigh_Height_Bin%Validity_Flag) then
       print *,"    ...%Validity_Flag                   = Valid"
    else
       print *,"    ...%Validity_Flag                   = Not Valid"
    endif

    call SetIntAs(Rayleigh_Height_Bin%Rayleigh_Wind_Velocity,&
                  str_IntAS,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%Rayleigh_Wind_Velocity          = ",str_IntAs

    call SetIntAs(Rayleigh_Height_Bin%Rayleigh_Wind_to_Pressure,&
                  str_IntAS,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%Rayleigh_Wind_to_Pressure       = ",str_IntAs

    call SetIntAs(Rayleigh_Height_Bin%Rayleigh_Wind_to_Temperature,&
                  str_IntAS,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%Rayleigh_Wind_to_Temperature    = ",str_IntAs

    call SetIntAs(Rayleigh_Height_Bin%Rayleigh_Wind_to_Backscat_Ratio,&
                  str_IntAS,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%Rayleigh_Wind_to_Backscat_Ratio = ",str_IntAs

    call SetIntAul(Rayleigh_Height_Bin%Reference_Pressure,&
                  str_IntAul,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%Reference_Pressure              = ",str_IntAul

    call SetIntAus(Rayleigh_Height_Bin%Reference_Temperature,&
                  str_IntAus,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%Reference_Temperature           = ",str_IntAus

    call SetIntAul(Rayleigh_Height_Bin%Reference_Backscatter_Ratio,&
                  str_IntAul,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%Reference_Backscatter_Ratio     = ",str_IntAul

    call SetIntAus(Rayleigh_Height_Bin%Rayleigh_Error_Quantifier,&
                  str_IntAus,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%Rayleigh_Error_Quantifier       = ",str_IntAus

    call SetIntAul(Rayleigh_Height_Bin%Integration_Length,&
                  str_IntAul,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%Integration_Length              = ",str_IntAul

  end subroutine Print_Rayleigh_Height_Bin
    !  #]
  subroutine AsciiWrite_Rayleigh_Height_Bin(Rayleigh_Height_Bin, fileunit, &
                                            prefix, error_flag)
    !  #[
    TYPE(Rayleigh_Height_Bin_type), intent(in)  :: Rayleigh_Height_Bin
    integer,                        intent(in)  :: fileunit       ! input
    character(len=*),               intent(in)  :: prefix         ! input
    integer,                        intent(out) :: error_flag     ! output

    ! local variables
    character(len=maxlen_IntAs)  :: str_IntAs
    character(len=maxlen_IntAus) :: str_IntAus
    character(len=maxlen_IntAul) :: str_IntAul
    character(len=10)            :: str_Valid

    error_flag = no_error

    if (Rayleigh_Height_Bin%Validity_Flag) then
       str_Valid = "Valid"
    else
       str_Valid = "Not Valid"
    endif
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Validity_Flag                   = "//trim(str_Valid)

    call SetIntAs(Rayleigh_Height_Bin%Rayleigh_Wind_Velocity,&
                  str_IntAS,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Rayleigh_Wind_Velocity          = "//&
         trim(str_IntAs)//" [cm/s]"

    call SetIntAs(Rayleigh_Height_Bin%Rayleigh_Wind_to_Pressure,&
                  str_IntAS,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Rayleigh_Wind_to_Pressure       = "//&
         trim(str_IntAs)//" [10-6m/s/Pa]"

    call SetIntAs(Rayleigh_Height_Bin%Rayleigh_Wind_to_Temperature,&
                  str_IntAS,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Rayleigh_Wind_to_Temperature    = "//&
         trim(str_IntAs)//" [cm/s/K]"

    call SetIntAs(Rayleigh_Height_Bin%Rayleigh_Wind_to_Backscat_Ratio,&
                  str_IntAS,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Rayleigh_Wind_to_Backscat_Ratio = "//&
         trim(str_IntAs)//" [cm/s]"

    call SetIntAul(Rayleigh_Height_Bin%Reference_Pressure,&
                   str_IntAul,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Reference_Pressure              = "//&
         trim(str_IntAul)//" [Pa]"

    call SetIntAus(Rayleigh_Height_Bin%Reference_Temperature,&
                   str_IntAus,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Reference_Temperature           = "//&
         trim(str_IntAus)//" [10-2K]"

    call SetIntAul(Rayleigh_Height_Bin%Reference_Backscatter_Ratio,&
                   str_IntAul,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Reference_Backscatter_Ratio     = "//&
         trim(str_IntAul)//" [10-6]"

    call SetIntAus(Rayleigh_Height_Bin%Rayleigh_Error_Quantifier,&
                   str_IntAus,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Rayleigh_Error_Quantifier       = "//&
         trim(str_IntAus)//" [cm/s]"

    call SetIntAul(Rayleigh_Height_Bin%Integration_Length,&
                   str_IntAul,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Integration_Length              = "//&
         trim(str_IntAul)//" [m]"

    ! 1 spare byte

    return

999 print *,"ERROR in AsciiWrite_Rayleigh_Height_Bin:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Rayleigh_Height_Bin
    !  #]
  function Get_ByteSize_Rayl_Height_Bin() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result
    
    c_bytesize = 1*c_bytelen_Bool   + &
                 4*c_bytelen_IntAs  + &
                 3*c_bytelen_IntAul + &
                 2*c_bytelen_IntAus     ! 10 items in total

    ! fortran size = 44 bytes
    ! c       size = 25 bytes

  end function Get_ByteSize_Rayl_Height_Bin
    !  #]
  subroutine Write_L2B_Rayleigh_Height_Bin(fileunit, &
             Rayleigh_Height_Bin, error_flag)
    !  #[
    ! 28-11-2005 P. Poli  Original code. From TN2.1 v1.2 06-09-2005
    integer,                        intent(in)  :: fileunit
    TYPE(Rayleigh_Height_Bin_type), intent(in)  :: Rayleigh_Height_Bin
    integer,                        intent(out) :: error_flag

    error_flag = no_error

    call write_logical1_to_bin_file(fileunit, &
         Rayleigh_Height_Bin%Validity_Flag, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAs_to_bin_file(fileunit, &
         Rayleigh_Height_Bin%Rayleigh_Wind_Velocity, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAs_to_bin_file(fileunit, &
         Rayleigh_Height_Bin%Rayleigh_Wind_to_Pressure, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAs_to_bin_file(fileunit, &
         Rayleigh_Height_Bin%Rayleigh_Wind_to_Temperature, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAs_to_bin_file(fileunit, &
         Rayleigh_Height_Bin%Rayleigh_Wind_to_Backscat_Ratio, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAul_to_bin_file(fileunit, &
         Rayleigh_Height_Bin%Reference_Pressure, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAus_to_bin_file(fileunit, &
         Rayleigh_Height_Bin%Reference_Temperature, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAul_to_bin_file(fileunit, &
         Rayleigh_Height_Bin%Reference_Backscatter_Ratio, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAus_to_bin_file(fileunit, &
         Rayleigh_Height_Bin%Rayleigh_Error_Quantifier, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAul_to_bin_file(fileunit, &
         Rayleigh_Height_Bin%Integration_Length, error_flag)
    IF (error_flag .ne. no_error) goto 998

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Write_L2B_Rayleigh_Height_Bin"
    return

  end subroutine Write_L2B_Rayleigh_Height_Bin
    !  #]
  subroutine Read_L2B_Rayleigh_Height_Bin(fileunit, &
             Rayleigh_Height_Bin, error_flag)
    !  #[
    ! 10-01-2006 P. Poli  Original code. From TN2.1 v1.2 06-09-2005
    integer,                        intent(in)  :: fileunit
    TYPE(Rayleigh_Height_Bin_type), intent(inout):: Rayleigh_Height_Bin
    integer,                        intent(out) :: error_flag

    error_flag = no_error

    call read_logical1_from_bin_file(fileunit, &
         Rayleigh_Height_Bin%Validity_Flag, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_IntAs_from_bin_file(fileunit, &
         Rayleigh_Height_Bin%Rayleigh_Wind_Velocity, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_IntAs_from_bin_file(fileunit, &
         Rayleigh_Height_Bin%Rayleigh_Wind_to_Pressure, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_IntAs_from_bin_file(fileunit, &
         Rayleigh_Height_Bin%Rayleigh_Wind_to_Temperature, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_IntAs_from_bin_file(fileunit, &
         Rayleigh_Height_Bin%Rayleigh_Wind_to_Backscat_Ratio, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_IntAul_from_bin_file(fileunit, &
         Rayleigh_Height_Bin%Reference_Pressure, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_IntAus_from_bin_file(fileunit, &
         Rayleigh_Height_Bin%Reference_Temperature, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_IntAul_from_bin_file(fileunit, &
         Rayleigh_Height_Bin%Reference_Backscatter_Ratio, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_IntAus_from_bin_file(fileunit, &
         Rayleigh_Height_Bin%Rayleigh_Error_Quantifier, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_IntAul_from_bin_file(fileunit, &
         Rayleigh_Height_Bin%Integration_Length, error_flag)
    IF (error_flag .ne. no_error) goto 998

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Read_L2B_Rayleigh_Height_Bin"
    return

    end subroutine Read_L2B_Rayleigh_Height_Bin
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_Rayleigh_Profile(Rayleigh_Profile, N_Range_Gates0, error_flag)
    !  #[
    TYPE(Rayleigh_Profile_type), intent(out) :: Rayleigh_Profile ! output
    integer(IntAs), intent(in)  :: N_Range_Gates0 ! input
    integer,        intent(out) :: error_flag    ! output

    ! local variable
    integer :: r, AllocStatus

    error_flag = no_error

    Rayleigh_Profile%Obs_Type = Obs_Type_undefined

    IF (N_Range_Gates0 .eq. 0) THEN
       ! init them to be empty
       nullify(Rayleigh_Profile%Height_Bin)
    ELSE
       allocate(Rayleigh_Profile%Height_Bin(N_Range_Gates0),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_Rayleigh_Profile"
          error_flag = error_allocate
          return
       END IF
       
       DO r=1,N_Range_Gates0
          call Init_Rayleigh_Height_Bin(Rayleigh_Profile%Height_Bin(r))
       END DO
    END IF

  end subroutine Init_Rayleigh_Profile
    !  #]
  subroutine Delete_Rayleigh_Profile(Rayleigh_Profile)
    !  #[
    TYPE(Rayleigh_Profile_type), intent(inout) :: Rayleigh_Profile

    IF (associated(Rayleigh_Profile%Height_Bin)) &
         deallocate(Rayleigh_Profile%Height_Bin)

  end subroutine Delete_Rayleigh_Profile
    !  #]
  subroutine Print_Rayleigh_Profile(Rayleigh_Profile, error_flag)
    !  #[
    TYPE(Rayleigh_Profile_type), intent(in)  :: Rayleigh_Profile ! input
    integer,                     intent(out) :: error_flag       ! output

    ! local variables
    integer :: r, s

    error_flag = no_error

    print *,"Rayleigh_Profile%Obs_Type = ",&
         trim(GetObsTypeText(Rayleigh_Profile%Obs_Type))

    IF (associated(Rayleigh_Profile%Height_Bin)) THEN
       ! get the sizes of the array
       s = size(Rayleigh_Profile%Height_Bin)
       print "(a,i2,a)","    ==> size(Rayleigh_Profile%Height_Bin) = ",s,&
                        " [printing first and last one]"
       DO r=1,s,s-1
          print "('    Rayleigh_Profile%Height_Bin(',i2.2,')%...')",r
          call Print_Rayleigh_Height_Bin(&
               Rayleigh_Profile%Height_Bin(r),error_flag)
          if (error_flag .ne. no_error) return
       END DO
    ELSE
       print *,"    Rayleigh_Profile%Height_Bin(:) [seems not associated]"
    END IF

  end subroutine Print_Rayleigh_Profile
    !  #]
  subroutine AsciiWrite_Rayleigh_Profile(Rayleigh_Profile, fileunit, prefix, &
                                         PrintAll, error_flag)
    !  #[
    TYPE(Rayleigh_Profile_type), intent(in)  :: Rayleigh_Profile ! input
    integer,                     intent(in)  :: fileunit         ! input
    character(len=*),            intent(in)  :: prefix           ! input
    logical,                     intent(in)  :: PrintAll         ! input
    integer,                     intent(out) :: error_flag       ! output

    ! local variables
    integer            :: r, s, step
    character(len=256) :: ObsTypeText
    character(len=2)   :: str_cntr

    error_flag = no_error

    ObsTypeText = GetObsTypeText(Rayleigh_Profile%Obs_Type)
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Obs_Type = "//trim(ObsTypeText)

    IF (associated(Rayleigh_Profile%Height_Bin)) THEN
       ! get the sizes of the array
       s = size(Rayleigh_Profile%Height_Bin)
       IF (PrintAll .or. (s .le. 1) ) THEN
          step = 1
       ELSE
          step = s-1
          !print "(a,i2,a)","    ==> size(Rayleigh_Profile%Height_Bin) = ",s,&
          !                 " [printing first and last one]"
       END IF
       DO r=1,s,step
          write(str_cntr,"(i2.2)",err=998) r
          call AsciiWrite_Rayleigh_Height_Bin(Rayleigh_Profile%Height_Bin(r),&
               fileunit,&
               trim(prefix)//"%Height_Bin("//trim(str_cntr)//")", error_flag)
          if (error_flag .ne. no_error) return
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%Height_Bin(:) [seems not associated]"
    END IF

    return

998 print *,"conversion error in AsciiWrite_Rayleigh_Profile:"
    print *,"could not convert the counter ",r," to a string .."
    error_flag = error_writing_file
    return

999 print *,"ERROR in AsciiWrite_Rayleigh_Profile:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Rayleigh_Profile
    !  #]
  function Get_ByteSize_Rayleigh_Profile(Rayleigh_Profile) result(c_bytesize)
    !  #[
    TYPE(Rayleigh_Profile_type), intent(in)  :: Rayleigh_Profile ! input
    integer(IntAl) :: c_bytesize ! result
    
    c_bytesize = 1*c_bytelen_IntAuc    + &
                 36_IntAl ! 36 spare bytes
    IF (associated(Rayleigh_Profile%Height_Bin)) THEN
       c_bytesize = c_bytesize + &
            int(size(Rayleigh_Profile%Height_Bin)*&
                Get_ByteSize_Rayl_Height_Bin(),IntAl)
    END IF

    ! fortran size = 44*24+1  = 1057 bytes
    ! c       size = 25*24+1 =  601 bytes

  end function Get_ByteSize_Rayleigh_Profile
    !  #]
! todo: update typical sizes
  subroutine Write_L2B_Rayleigh_Profile(fileunit, &
             Rayleigh_Profile, error_flag)
    !  #[
    ! 25-11-2005 P. Poli  Original code. From TN2.1 v1.2 06-09-2005
    integer,                     intent(in)  :: fileunit
    TYPE(Rayleigh_Profile_type), intent(in)  :: Rayleigh_Profile
    integer,                     intent(out) :: error_flag

    ! local variables
    integer :: r, s, nr_of_zero_bytes

    error_flag = no_error
    IF (.not.(associated(Rayleigh_Profile%Height_Bin))) THEN
       print *,"ERROR in Write_L2B_Rayleigh_Profile():"
       print *,"The pointer to this Rayleigh_Profile seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! Obs_Type
    call write_IntAuc_to_bin_file(fileunit, Rayleigh_Profile%Obs_Type,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 36 spare bytes
    nr_of_zero_bytes = 36
    call write_SpareBytes_to_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! get the size of the array
    s = size(Rayleigh_Profile%Height_Bin)
    print *,"Write_L2B_Rayleigh_Profile: size(Rayleigh_Profile%Height_Bin) =",s
    DO r=1,s
       call Write_L2B_Rayleigh_Height_Bin(fileunit, &
            Rayleigh_Profile%Height_Bin(r), error_flag)
       if (error_flag .ne. no_error) goto 998
    END DO

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Write_L2B_Rayleigh_Profile"
    return

  end subroutine Write_L2B_Rayleigh_Profile
    !  #]
  subroutine Read_L2B_Rayleigh_Profile(fileunit, &
             Rayleigh_Profile, error_flag)
    !  #[
    ! 10-01-2006 P. Poli  Original code. From TN2.1 v1.2 06-09-2005
    integer,                     intent(in)   :: fileunit
    TYPE(Rayleigh_Profile_type), intent(inout):: Rayleigh_Profile
    integer,                     intent(out)  :: error_flag

    ! local variables
    integer :: r, s, nr_of_zero_bytes

    error_flag = no_error
    IF (.not.(associated(Rayleigh_Profile%Height_Bin))) THEN
       print *,"ERROR in Read_L2B_Rayleigh_MDS():"
       print *,"The pointer to this Rayleigh_Profile seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! Obs_Type
    call read_IntAuc_from_bin_file(fileunit, Rayleigh_Profile%Obs_Type,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 36 spare bytes
    nr_of_zero_bytes = 36
    call read_SpareBytes_from_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! get the size of the array
    s = size(Rayleigh_Profile%Height_Bin)
    print *,"Read_L2B_Rayleigh_Profile: size(Rayleigh_Profile%Height_Bin) =",s
    DO r=1,s
       call Read_L2B_Rayleigh_Height_Bin(fileunit, &
            Rayleigh_Profile%Height_Bin(r), error_flag)
       if (error_flag .ne. no_error) goto 998
    END DO

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Read_L2B_Rayleigh_Profile"
    return

  end subroutine Read_L2B_Rayleigh_Profile
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_L2B_Rayleigh_MDS(Rayleigh_MDS, &
                      N_Meas0, N_Obs_Max_Rayleigh0, N_Range_Gates0, error_flag)
    !  #[
    !  Modifications:
    !     P. Poli    02-12-2005  Saving N_Obs_Max in L2B_SummaryVars%M_Rayleigh
    !     P. Poli    26-01-2006  Removed N_Obs_Max from Rayleigh_MDS
    TYPE(L2B_Rayleigh_MDS_type), pointer     :: Rayleigh_MDS  ! input/output
    integer(IntAs),              intent(in)  :: N_Meas0        ! input
    integer(IntAs),              intent(in)  :: N_Obs_Max_Rayleigh0     ! input
    integer(IntAs),              intent(in)  :: N_Range_Gates0 ! input
    integer,                     intent(out) :: error_flag    ! output

    ! local variable
    integer :: p, AllocStatus

    error_flag = no_error

    IF (.not. associated(Rayleigh_MDS)) THEN
       print *,"ERROR in Init_L2B_Rayleigh_MDS: "
       print *,"pointer Rayleigh_MDS seems not associated !!!"
       error_flag = error_programming
       return
    END IF
    
    call InitDateTime(Rayleigh_MDS%Start_of_Obs_DateTime)


    ! the first one is used for allocating/building the datastructure
    ! so I'll keep a copy for convenience
    Rayleigh_MDS%N_Meas       = N_Meas0
    ! these 2 are filled during processing, so init them to zero
    Rayleigh_MDS%N_Obs_Rayleigh_Actual = 0
    Rayleigh_MDS%P            = 0

    allocate(Rayleigh_MDS%Map_of_L1B_Measurements_Used,STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation failure in subroutine Init_L2B_Rayleigh_MDS"
       error_flag = error_allocate
       return
    END IF
    call Init_Meas_Map(Rayleigh_MDS%Map_of_L1B_Measurements_Used,&
                       N_Meas0, N_Range_Gates0, error_flag)
    IF (error_flag .ne. no_error) return

    IF (N_Obs_Max_Rayleigh0 .eq. 0) THEN
       ! init it to be empty
       nullify(Rayleigh_MDS%Rayleigh_Profile)
    ELSE
       allocate(Rayleigh_MDS%Rayleigh_Profile(N_Obs_Max_Rayleigh0),&
                STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L2B_Rayleigh_MDS"
          error_flag = error_allocate
          return
       END IF
       
       DO p=1,N_Obs_Max_Rayleigh0
          call Init_Rayleigh_Profile(Rayleigh_MDS%Rayleigh_Profile(p),&
                                     N_Range_Gates0, error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    END IF

  end subroutine Init_L2B_Rayleigh_MDS
    !  #]
  subroutine Delete_L2B_Rayleigh_MDS(Rayleigh_MDS)
    !  #[
    TYPE(L2B_Rayleigh_MDS_type), pointer :: Rayleigh_MDS

    ! local variable
    integer :: p

    IF (.not. associated(Rayleigh_MDS)) return
    
    IF (associated(Rayleigh_MDS%Map_of_L1B_Measurements_Used)) THEN
       call Delete_Meas_Map(Rayleigh_MDS%Map_of_L1B_Measurements_Used)
       deallocate(Rayleigh_MDS%Map_of_L1B_Measurements_Used)
    END IF

    IF (associated(Rayleigh_MDS%Rayleigh_Profile)) THEN
       DO p=1,size(Rayleigh_MDS%Rayleigh_Profile)
          call Delete_Rayleigh_Profile(Rayleigh_MDS%Rayleigh_Profile(p))
       END DO
       deallocate(Rayleigh_MDS%Rayleigh_Profile)
    END IF

  end subroutine Delete_L2B_Rayleigh_MDS
    !  #]
  function Get_ByteSize_L2B_Rayleigh_MDS(Rayleigh_MDS,N_Meas_to_write,&
                               N_Obs_Rayleigh_to_write) result(c_bytesize)
    !  #[
    TYPE(L2B_Rayleigh_MDS_type), intent(in) :: Rayleigh_MDS
    integer(IntAuc),             intent(in) :: N_Meas_to_write
    integer(IntAuc),             intent(in) :: N_Obs_Rayleigh_to_write
    integer(IntAl)                          :: c_bytesize ! result

    ! local variable
    integer :: s
    
    c_bytesize = c_bytelen_DateTime + &
                 3*c_bytelen_IntAs    

    IF (associated(Rayleigh_MDS%Map_of_L1B_Measurements_Used)) THEN
       c_bytesize = c_bytesize + &
            Get_ByteSize_Meas_Map(Rayleigh_MDS%Map_of_L1B_Measurements_Used,&
                                  N_Meas_to_write)
    END IF

    IF (associated(Rayleigh_MDS%Rayleigh_Profile)) THEN
       s = size(Rayleigh_MDS%Rayleigh_Profile)
       IF (N_Obs_Rayleigh_to_write .gt. s) THEN
          print *,"WARNING in GetByteSize_L2B_Rayleigh_MDS()"
          print *,"invalid value for N_Obs_Rayleigh_to_write: ",&
               N_Obs_Rayleigh_to_write
          print *,"max. possible value is: "
          print *,"size(Rayleigh_MDS%Rayleigh_Profile) = ",s
          print *,"GetByteSize result will be useless...."
       END IF
       c_bytesize = c_bytesize + N_Obs_Rayleigh_to_write*&
            Get_ByteSize_Rayleigh_Profile(Rayleigh_MDS%Rayleigh_Profile(1))
    END IF

    ! typical size:
    ! fortran size = 16 + 3*2 + 9648 + 5*1057 = 14955 bytes
    ! c       size = 12 + 3*2 + 4824 + 5* 601 =  7847 bytes

  end function Get_ByteSize_L2B_Rayleigh_MDS
    !  #]
  !---------------------------------------------------------------------
  function get_Rayl_nr_of_cloud_points(Rayleigh_MDS,r) result(n)
    !  #[
    TYPE(L2B_Rayleigh_MDS_type), pointer :: Rayleigh_MDS ! input
    integer, intent(in) :: r ! input
    integer             :: n ! result

    ! local variable
    integer :: o, Obs_Type

    n = -1
    IF (.not. associated(Rayleigh_MDS)) return

    ! the counts are not stored as such, so use a little loop 
    ! to construct the numbers
    n = 0
    do o=1,Rayleigh_MDS%N_Obs_Rayleigh_Actual
       Obs_Type = Rayleigh_MDS%Rayleigh_Profile(o)%Obs_Type
       if (Obs_Type .eq. Obs_Type_cloudy_returns) then
          n = n + Count_Nr_Points_On_Level(&
                      Rayleigh_MDS%Map_of_L1B_Measurements_Used,r,o)
       end if
    end do

  end function get_Rayl_nr_of_cloud_points
    !  #]
  function get_Rayl_nr_of_clear_points(Rayleigh_MDS,r) result(n)
    !  #[
    TYPE(L2B_Rayleigh_MDS_type), pointer :: Rayleigh_MDS ! input
    integer, intent(in) :: r ! input
    integer             :: n ! result

    ! local variable
    integer :: o, Obs_Type

    n = -1
    IF (.not. associated(Rayleigh_MDS)) return

    ! the counts are not stored as such, so use a little loop 
    ! to construct the numbers
    n = 0
    do o=1,Rayleigh_MDS%N_Obs_Rayleigh_Actual
       Obs_Type = Rayleigh_MDS%Rayleigh_Profile(o)%Obs_Type
       if (Obs_Type .eq. Obs_Type_clear_returns) then
          n = n + Count_Nr_Points_On_Level(&
                      Rayleigh_MDS%Map_of_L1B_Measurements_Used,r,o)
       end if
    end do

  end function get_Rayl_nr_of_clear_points
    !  #]
  !---------------------------------------------------------------------
  subroutine Print_L2B_Rayleigh_MDS(Rayleigh_MDS,error_flag,PrintAll)
    !  #[

    TYPE(L2B_Rayleigh_MDS_type), pointer     :: Rayleigh_MDS
    integer,                     intent(out) :: error_flag       ! output
    logical, optional,           intent(in)  :: PrintAll

    ! local variables
    integer           :: p, s, step
    logical           :: do_PrintAll
    character(len=maxlen_DateTime_KVT) :: DateTimeString

    error_flag = no_error

    do_PrintAll = .false.
    IF (present(PrintAll)) do_PrintAll = PrintAll

    IF (associated(Rayleigh_MDS)) THEN
       print *,"Rayleigh_MDS%..."
    ELSE
       print *,"Rayleigh_MDS [seems not associated]"
       return
    END IF

    ! print the date and time
    if (DateTimeMissing(Rayleigh_MDS%Start_of_Obs_DateTime)) then
       DateTimeString = "[missing]"
    else
       DateTimeString = &
            Construct_KVT_DateTime(Rayleigh_MDS%Start_of_Obs_DateTime)
    end if
    print *,"...%start_of_obs_datetime: "//trim(DateTimeString)
    print *,"...%N_Meas       = ",Rayleigh_MDS%N_Meas
    ! Poli: keeping the following line otherwise all expected output have to be
    ! changed....
!    print *,"...%N_Obs_Max    = ",Rayleigh_MDS%N_Obs_Max
    print *,"...%N_Obs_Max    = ",size(Rayleigh_MDS%Rayleigh_Profile)
    print *,"...%N_Obs_Actual = ",Rayleigh_MDS%N_Obs_Rayleigh_Actual
    print *,"...%P            = ",Rayleigh_MDS%P

    IF (associated(Rayleigh_MDS%Map_of_L1B_Measurements_Used)) THEN
       call Print_Meas_Map(Rayleigh_MDS%Map_of_L1B_Measurements_Used)
! printing of the weights is not yet implemented
    ELSE
       print *,"...%Map_of_L1B_Measurements_Used [seems not associated]"
    END IF

    IF (associated(Rayleigh_MDS%Rayleigh_Profile)) THEN
       s = Rayleigh_MDS%N_Obs_Rayleigh_Actual

       IF (do_PrintAll .or. (s .le. 1) ) THEN
          step=1
          print "(a,i2)"," ==> N_Obs_Actual = ",s
       ELSE
          step = s-1
          print "(a,i2,a)",&
               " ==> N_Obs_Actual = ",s," [printing first and last one]"
       END IF

       DO p=1,s,step
          print "(' ...%Rayleigh_Profile(',i2.2,')%...')",p
          call Print_Rayleigh_Profile(&
                   Rayleigh_MDS%Rayleigh_Profile(p), error_flag)
       END DO
    ELSE
       print *,"...%Rayleigh_Profile(:) = [NOT ASSOCIATED]"
    END IF

  end subroutine Print_L2B_Rayleigh_MDS
    !  #]
  subroutine AsciiWrite_L2B_Rayleigh_MDS(Rayleigh_MDS,fileunit,&
                                         prefix,PrintAll,error_flag)
    !  #[
    TYPE(L2B_Rayleigh_MDS_type), pointer     :: Rayleigh_MDS     ! input
    integer,                     intent(in)  :: fileunit         ! input
    character(len=*),            intent(in)  :: prefix           ! input
    logical,                     intent(in)  :: PrintAll         ! input
    integer,                     intent(out) :: error_flag       ! output

    ! local variables
    integer                            :: p, s, step
    character(len=maxlen_DateTime_KVT) :: str_DateTime
    character(len=maxlen_IntAs)        :: str_IntAs
    character(len=2)                   :: str_cntr

    error_flag = no_error

    IF (.not. associated(Rayleigh_MDS)) THEN
       print *,"Error in AsciiWrite_L2B_Rayleigh_MDS():"
       print *,"Rayleigh_MDS [seems not associated]"
       error_flag = error_programming
       return
    END IF

    if (DateTimeMissing(Rayleigh_MDS%start_of_obs_datetime)) then
       str_DateTime = "[missing]"
    else
       str_DateTime = &
            Construct_KVT_DateTime(Rayleigh_MDS%start_of_obs_datetime)
    end if
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%start_of_obs_datetime = "//trim(str_DateTime)

    call SetIntAs(Rayleigh_MDS%N_Meas,str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%N_Meas                = "//trim(str_IntAs)

    call SetIntAs(Rayleigh_MDS%N_Obs_Rayleigh_Actual,str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%N_Obs_Rayleigh_Actual = "//trim(str_IntAs)

    call SetIntAs(Rayleigh_MDS%P,str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%P                     = "//trim(str_IntAs)

    IF (associated(Rayleigh_MDS%Map_of_L1B_Measurements_Used)) THEN
       call AsciiWrite_Meas_Map(Rayleigh_MDS%Map_of_L1B_Measurements_Used,&
            fileunit,trim(prefix)//"%Map_of_L1B_Measurements_Used",&
            PrintAsMap=.false.,PrintAll=PrintAll,error_flag=error_flag)
       IF (error_flag .ne. no_error) goto 999
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%Map_of_L1B_Measurements_Used = [seems not associated]"
    END IF



    IF (associated(Rayleigh_MDS%Rayleigh_Profile)) THEN
       s = Rayleigh_MDS%N_Obs_Rayleigh_Actual
       IF (PrintAll .or. (s .le. 1) ) THEN
          step=1
       ELSE
          step = s-1
          !print "(a,i2,a)",&
          !     " ==> N_Obs_Actual = ",s," [printing first and last one]"
       END IF
       DO p=1,s,step
          write(str_cntr,"(i2.2)",err=998) p
          call AsciiWrite_Rayleigh_Profile(&
               Rayleigh_MDS%Rayleigh_Profile(p), fileunit, &
               trim(prefix)//"%Rayleigh_Profile("//trim(str_cntr)//")",&
               PrintAll,error_flag)
          IF (error_flag .ne. no_error) goto 999          
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%Rayleigh_Profile(:) [seems not associated]"
    END IF

    return

998 print *,"conversion error in AsciiWrite_L2B_Rayleigh_MDS:"
    print *,"could not convert the counter ",p," to a string .."
    error_flag = error_writing_file
    return

999 print *,"ERROR in AsciiWrite_L2B_Rayleigh_MDS:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2B_Rayleigh_MDS
    !  #]
  !---------------------------------------------------------------------
  subroutine Write_L2B_Rayleigh_MDS(fileunit, Rayleigh_MDS,&
       N_Meas_to_write, N_Obs_Rayleigh_to_write, error_flag)
    !  #[
    ! Writes the Rayleigh MDS in binary format
    ! 01-12-2005 P. Poli Original code. From L2B TN2.1 v1.2 06-09-2005

    integer,                     intent(in)  :: fileunit
    TYPE(L2B_Rayleigh_MDS_type), pointer     :: Rayleigh_MDS
    integer(IntAuc),             intent(in)  :: N_Meas_to_write
    integer(IntAuc),             intent(in)  :: N_Obs_Rayleigh_to_write
    integer,                     intent(out) :: error_flag

    ! local variables
    integer :: p, s

    !-----------------------------------
    ! writing the Rayleigh MDS elements
    !-----------------------------------

    error_flag = no_error
    IF (.not. associated(Rayleigh_MDS)) THEN
       print *,"ERROR in Write_L2B_Rayleigh_MDS():"
       print *,"The pointer to this Rayleigh_MDS seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! Start_of_Obs_Date
    call write_DateTime_to_bin_file(fileunit, &
         Rayleigh_MDS%Start_of_Obs_DateTime,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call write_IntAs_to_bin_file(fileunit,Rayleigh_MDS%N_Meas,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call write_IntAs_to_bin_file(fileunit,Rayleigh_MDS%N_Obs_Rayleigh_Actual,&
                                error_flag)
    IF (error_flag .ne. no_error) goto 998

    call write_IntAs_to_bin_file(fileunit,Rayleigh_MDS%P,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call Write_Meas_Map(fileunit, Rayleigh_MDS%Map_of_L1B_Measurements_Used,&
                        N_Meas_to_write, error_flag)
    IF (error_flag .ne. no_error) goto 998

    s = size(Rayleigh_MDS%Rayleigh_Profile)
    IF (N_Obs_Rayleigh_to_write .gt. s) THEN
       print *,"ERROR in Write_L2B_Rayleigh_MDS():"
       print *,"invalid value for N_Obs_Rayleigh_to_write: ",&
            N_Obs_Rayleigh_to_write
       print *,"max. possible value is: "
       print *,"size(Rayleigh_MDS%Rayleigh_Profile) = ",s
       error_flag = error_programming
       return
    END IF

    print *,"Write_L2B_Rayleigh_MDS: Rayleigh_MDS%N_Obs_Actual =",&
                Rayleigh_MDS%N_Obs_Rayleigh_Actual
    DO p=1,N_Obs_Rayleigh_to_write
       call Write_L2B_Rayleigh_Profile(fileunit, &
                  Rayleigh_MDS%Rayleigh_Profile(p), error_flag)
       IF (error_flag .ne. no_error) goto 998
    ENDDO

    !print *,"Write_L2B_Rayleigh_MDS: done with Write_L2B_Rayleigh_MDS"

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Write_L2B_Rayleigh_MDS"
    return

  end subroutine Write_L2B_Rayleigh_MDS
    !  #]
  subroutine Read_L2B_Rayleigh_MDS(fileunit, Rayleigh_MDS,error_flag)
    !  #[

    ! 10-01-2006 P. Poli Original code. From L2B TN2.1 v1.2 06-09-2005
    integer,                     intent(in)  :: fileunit
    TYPE(L2B_Rayleigh_MDS_type), pointer     :: Rayleigh_MDS
    integer,                     intent(out) :: error_flag

    ! local variables
    integer  :: p, s
    character(len=maxlen_DateTime_KVT) :: DateTimeString

    !-----------------------------------
    ! reading the Rayleigh MDS elements
    !-----------------------------------

    error_flag = no_error
    IF (.not. associated(Rayleigh_MDS)) THEN
       print *,"ERROR in Read_L2B_Rayleigh_MDS():"
       print *,"The pointer to this Rayleigh_MDS seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! Start_of_Obs_DateTime
    call read_DateTime_from_bin_file(fileunit,&
                      Rayleigh_MDS%Start_of_Obs_DateTime,error_flag)

    print *,"Read_L2B_Rayleigh_MDS: Start_of_Obs_DateTime is: "
    if (DateTimeMissing(Rayleigh_MDS%Start_of_Obs_DateTime)) then
       DateTimeString = "[missing]"
    else
       DateTimeString = &
            Construct_KVT_DateTime(Rayleigh_MDS%Start_of_Obs_DateTime)
    end if
    print *,DateTimeString

    call read_IntAs_from_bin_file(fileunit,Rayleigh_MDS%N_Meas,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call read_IntAs_from_bin_file(fileunit,Rayleigh_MDS%N_Obs_Rayleigh_Actual,&
                                 error_flag)
    IF (error_flag .ne. no_error) goto 998

    call read_IntAs_from_bin_file(fileunit,Rayleigh_MDS%P,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call Read_Meas_Map(fileunit, Rayleigh_MDS%Map_of_L1B_Measurements_Used, &
                       error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! for N_Obs=0 this pointer may be un-associated
    ! so we have to check for that before requesting its size
    IF (associated(Rayleigh_MDS%Rayleigh_Profile)) THEN
       s = size(Rayleigh_MDS%Rayleigh_Profile)
       DO p=1,s
          call Read_L2B_Rayleigh_Profile(fileunit, &
               Rayleigh_MDS%Rayleigh_Profile(p), &
               error_flag)
          IF (error_flag .ne. no_error) goto 998
       ENDDO
    END IF
    !print *,"Read_L2B_Rayleigh_MDS: done with Read_L2B_Rayleigh_MDS"


    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Read_L2B_Rayleigh_MDS"
    return

  end subroutine Read_L2B_Rayleigh_MDS

    !  #]
  !---------------------------------------------------------------------
  subroutine Print_L2B_Rayleigh_hlos_profile(Rayleigh_MDS,error_flag,Obs_Nr)
    !  #[

    TYPE(L2B_Rayleigh_MDS_type), pointer     :: Rayleigh_MDS
    integer,                     intent(out) :: error_flag       ! output
    integer,                     intent(in)  :: Obs_Nr

    call AsciiWrite_L2B_Rayl_hlos_prof(Rayleigh_MDS,error_flag,&
                                       Obs_Nr, fileunit_stdout)

  end subroutine Print_L2B_Rayleigh_hlos_profile
    !  #]
  !---------------------------------------------------------------------
  subroutine AsciiWrite_L2B_Rayl_hlos_prof(Rayleigh_MDS,error_flag,&
                                           Obs_Nr, Fileunit)
    !  #[
    TYPE(L2B_Rayleigh_MDS_type), pointer     :: Rayleigh_MDS
    integer,                     intent(out) :: error_flag       ! output
    integer,                     intent(in)  :: Obs_Nr
    integer,                     intent(in)  :: Fileunit

    ! local variables
    integer                      :: Bin, Nr_Bins
    character(len=maxlen_IntAs)  :: str_IntAs
    character(len=1)             :: Validity_Flag

    error_flag = no_error

    IF (.not. associated(Rayleigh_MDS)) THEN
       print *,"ERROR: Rayleigh_MDS [seems not associated]"
       error_flag = error_programming
       return
    END IF

    IF (.not. associated(Rayleigh_MDS%Rayleigh_Profile)) THEN
       print *,"ERROR: Rayleigh_MDS%Rayleigh_Profile [seems not associated]"
       error_flag = error_programming
       return
    END IF
   
    write(Fileunit,"(a,i2)") " ==> N_Obs_Actual = ",&
         Rayleigh_MDS%N_Obs_Rayleigh_Actual
    IF (Obs_Nr .gt. Rayleigh_MDS%N_Obs_Rayleigh_Actual) THEN
       write(Fileunit,*) "Sorry, Obs_Nr ",Obs_Nr," was not (yet) defined"
       write(Fileunit,*) "Nothing to print."
       return
    END IF

    write(Fileunit,*) "Rayleigh hlos-profile for Obs. ",Obs_Nr
    Nr_Bins = size(Rayleigh_MDS%Rayleigh_Profile(Obs_Nr)%Height_Bin)
    write(Fileunit,*) "nr height bins = ",Nr_Bins
    
    DO Bin=1,Nr_Bins
       call SetIntAs(Rayleigh_MDS%Rayleigh_Profile(Obs_Nr)%&
            Height_Bin(Bin)%Rayleigh_Wind_Velocity,&
            str_IntAS, error_flag)
       if (error_flag .ne. no_error) return
       IF (Rayleigh_MDS%Rayleigh_Profile(Obs_Nr)%&
                        Height_Bin(Bin)%Validity_Flag) THEN
          Validity_Flag="1"
       ELSE
          Validity_Flag="0"
       ENDIF
       write(Fileunit,*) "hlos = ",str_IntAs,"  ",Validity_Flag
    END DO

  end subroutine AsciiWrite_L2B_Rayl_hlos_prof
    !  #]
  !---------------------------------------------------------------------
end module L2B_Rayleigh_MDS
