module L2BC_DataStructure
  !  #[ Description

  !     a module to define the datastructure used by the L2B processor,
  !     and the routines needed for initialisation, allocation
  !     and deallocation of all data structs inside it.
  !     
  !     created:     22-Apr-2005, by: J. de Kloe, KNMI
  !
  !     Modifications
  !      02-Dec-2005 P. Poli    Added routine Handle_L2BHeaders
  !      11-Jan-2006 J. de Kloe Added the new DateTime type 
  !      22-Feb-2006 J. de Kloe Precision r8_ for lat1, lon1, lat2, lon2
  !      06-Mar-2006 J. de Kloe removed program_abort
  !      12-May-2006 D. Tan     Null use of ParKind1 module
  !      27-Jun-2006 D. Tan     More geolocation params in L2B_ObsRangeGateType
  !      23-Oct-2006 D. Tan     Rename to L2BC_DataStructure
  !      24-Oct-2006 D. Tan     Working elements transferred to WorkingDataType
  !      27-Oct-2006 D. Tan     Generalize for L2B/L2C alternatives
  !      17-Nov-2006 J. de Kloe allow initialisation now without
  !                             allocation of datasets
  !      15-Feb-2007 J. de Kloe created Fill_L2B_FH/MPH/SPH as a simplification
  !                             of Handle_L2BHeaders
  !      22-Feb-2007 J. de Kloe added routine Empty_L2BC_BinaryDatasets
  !                             added some AsciiWrite code
  !      02-Mar-2007 J. de Kloe added some parameters to define Num_DSD
  !      15-Mar-2007 J. de Kloe removed use of SummaryVars
  !      15-Mar-2007 J. de Kloe activated AsciiWrite for L2B_PCD
  !      15-Mar-2007 J. de Kloe some tidying (defined GetStatistics)
  !      19-Mar-2007 J. de Kloe activated AsciiWrite for L2C datasets
  !      27-Mar-2007 J. de Kloe change string to use L2C when L2C_flag is used
  !                             in Fill_L2B_SPH, rename Fill_L2B_SPH to 
  !                             Fill_L2BC_SPH, Fill_L2B_FH to Fill_L2BC_FH,
  !                             and Fill_L2B_MPH to Fill_L2BC_MPH
  !      04-Apr-2007 J. de Kloe adapted to use file_type_L2B_N/file_type_L2C_N
  !      24-Apr-2007 D. Tan     Remove L2B_PreProcessedAuxMetDataType.
  !      24-Apr-2007 J. de Kloe phase out the SummaryVars code
  !      27-Jul-2007 J. de Kloe corrected handling of optional switches
  !                             in AsciiWrite_L2BC_DataStructure
  !      30-Jul-2007 J. de Kloe added Init_Headers switch to 
  !                             subroutine Init_L2BC_DataStructure()
  !      30-Jul-2007 J. de Kloe renamed Fill_L2BC_MPH to 
  !                             Fill_Remaining_L2BC_MPH_Items
  !                             and removed L1B dependancy
  !      03-Sep-2007 P. Poli    Introduced L2BC_Headers.
  !                             Modified routines Fill_L2BC_FH, 
  !                             Fill_Remaining_L2BC_MPH_Items, Fill_L2BC_SPH, 
  !                             AsciiWrite_L2BC_DataStructure, 
  !                             Init_L2BC_DataStructure split in two 
  !                             (Init_L2BC_HDR_DataStructure and 
  !                             Init_L2BC_BRC_DataStructure), 
  !                             Delete_L2BC_DataStructure split in two 
  !                             (Delete_L2BC_HDR_DataStructure and 
  !                             Delete_L2BC_BRC_DataStructure). 
  !                             Allocation of FH,MPH,SPH now done in 
  !                             Init_L2BC_HDR_DataStructure, to ensure 
  !                             symmetry with Delete_L2BC_HDR_DataStructure. 
  !                             Added routines Print_L2BC_HDR_DataStructure 
  !                             and Print_L2BC_BRC_DataStructure 
  !                             (replacing Print_L2BC_DataStructure). 
  !                             Split L2BC_DataType in two parts: 
  !                             L2BC_HDR_DataType and L2BC_BRC_DataType
  !      10-Oct-2007 J. de Kloe adapted subroutine Fill_L2BC_FH() to use
  !                             the FH_Default_Fields defined in AuxPar
  !                             and reorganised the print/AsciiWrite routines.
  !      18-Oct-2007 J. de Kloe adapted Fill_Remaining_L2BC_MPH_Items to use
  !                             the MPH_Default_Fields defined in AuxPar
  !      19-Oct-2007 P. Poli    removed FH,MPH,SPH from L2BC_BRC_DataType
  !      06-Dec-2007 J. de Kloe changed systemstr/running_at parameters
  !                             to reflect allowed values, see SPR 28
  !      06-Dec-2007 J. de Kloe changed Fill_Remaining_L2BC_MPH_Items
  !                             to allow also filling of sensing_start/stop
  !                             and to allow filling MPH.Proc_Center with
  !                             the JobOrder.Processing_Station field
  !      16-Jan-2008 J. de Kloe phase out integer kind i_ 
  !      17-Jan-2008 J. de Kloe removed a last few 64bit incompatibilities
  !
  !  #]
  !  #[ modules used

  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE FixedHeader, only: FH_type, Init_FH, Print_FH, Fill_FH_items, &
                         file_type_L2B_N, file_type_L2C_N, AsciiWrite_FH, &
                         maxlen_file_type, &
                         systemstr_ECMWF_IFS,      systemstr_ESRIN_REPR, &
                         systemstr_LOCAL_MET_INST, systemstr_ACMF
  USE MainProductHeader, only: MPH_type, Fill_MPH_items, AsciiWrite_MPH, &
                               Init_MPH, Print_MPH, maxlen_Ref_Doc, &
                               maxlen_Product, maxlen_Proc_Stage, &
                               maxlen_Proc_Center, maxlen_Software_Ver
  USE L2BC_DS_Settings, only: set_L2BC_string
  USE L2BC_SPH, only: L2BC_SPH_type, Init_L2BC_SPH, Fill_L2BC_SPH_items, &
                      Delete_L2BC_SPH, AsciiWrite_L2BC_SPH, &
                      Num_DSD_L2B, Num_DSD_L2C, Num_DS_L2B, Num_DS_L2C, &
                      Print_L2BC_SPH
  USE L2B_Geoloc_ADS, only: L2B_Geoloc_ADS_type, Init_L2B_Geoloc_ADS, &
                         Init_L2B_Geoloc_ADS_module, &
                         Delete_L2B_Geoloc_ADS, &
                         AsciiWrite_L2B_Geoloc_ADS
  USE L2B_PCD_ADS, only: L2B_PCD_ADS_type, Init_L2B_PCD_ADS, &
                         Delete_L2B_PCD_ADS, Print_L2B_PCD_ADS, &
                         AsciiWrite_L2B_PCD_ADS, &
                         L1B_Mie_Meas_Screening_Type, &
                         L1B_Mie_Bin_Contains_Surface, &
                         L1B_Mie_Contains_Surface_DEM, &
                         L1B_Mie_Bin_Below_Surface,    &
                         L1B_Mie_Below_Surface_DEM
  USE L2B_Mie_MDS, only: L2B_Mie_MDS_type, Init_L2B_Mie_MDS, &
                         Delete_L2B_Mie_MDS, Print_L2B_Mie_MDS, &
                         AsciiWrite_L2B_Mie_hlos_profile, &
                         AsciiWrite_L2B_Mie_MDS, Mie_Profile_type
  USE L2B_Rayleigh_MDS, only: L2B_Rayleigh_MDS_type, Init_L2B_Rayleigh_MDS, &
                         Delete_L2B_Rayleigh_MDS, Print_L2B_Rayleigh_MDS,&
                         AsciiWrite_L2B_Rayl_hlos_prof, &
                         AsciiWrite_L2B_Rayleigh_MDS, Rayleigh_Profile_type

  ! to be moved into the Rayleigh_MDS and Mie_MDS
  USE L2B_parameters, only: not_yet_classified, Obs_Type_undefined, &
                            Obs_Type_cloudy_returns, Obs_Type_clear_returns

  ! L2C-related modules
  USE L2C_AssimPCD_ADS, only: L2C_AssimPCD_ADS_type, Init_L2C_AssimPCD_ADS, &
                              Delete_L2C_AssimPCD_ADS, Print_L2C_AssimPCD_ADS,&
                              AsciiWrite_L2C_AssimPCD_ADS
  USE L2C_MieVec_MDS, only:   L2C_MieVec_MDS_type, Init_L2C_MieVec_MDS, &
                              Delete_L2C_MieVec_MDS, Print_L2C_MieVec_MDS, &
                              AsciiWrite_L2C_MieVec_MDS
  USE L2C_RayleighVec_MDS, only: L2C_RayleighVec_MDS_type,   &
                                 Init_L2C_RayleighVec_MDS,   &
                                 Delete_L2C_RayleighVec_MDS, &
                                 Print_L2C_RayleighVec_MDS,  &
                                 AsciiWrite_L2C_RaylVec_MDS
  USE Meas_Map, only: Meas_Map_Type
  USE L2B_AuxPar_DataStructure, only: FH_Default_Fields_Type, &
                                      MPH_Default_Fields_Type
  USE L2B_Proc_Settings, only: L2BP_Settings_type

  ! support modules
  USE EE_CFI_Datatypes, only: DateTimeType,& 
                              InitDateTime, SetDateTime, PrintDateTime, &
                              IntAuc, IntAs, IntAd, IntAl, IntAus
  USE numerics, only: r8_, missing_indicator_real
  USE ErrorHandler, only: no_error, error_programming, error_allocate, &
                          error_reading_file
  USE Compiler_Features, only: newline
  USE Logging, only: logmsg,log_debug,log_info,log_progress,log_warn,log_error
  USE AeolusConstants, only: N_range_gates
  USE AeolusConstants, only: software_version_L2BP => software_version
  USE LunManager, only: fileunit_stdout

  !  #]
  !  #[ Parameter and data type definitions
  implicit none

! these working arrays should be renamed: IntermediateWorkingArrays
! or something like that, and placed in a separated module.
! They should also be merged with the L1B working arrays
! since these also are similar intermediate values.

  TYPE L2B_MieRangeGateType
     !  #[

     ! temporarily added for the sensitivity study at KNMI:
     real(r8_) :: p ! [Pa]  value in the centre of the range-bin
     real(r8_) :: T ! [K]   value in the centre of the range-bin
     real(r8_) :: u ! [m/s] value in the centre of the range-bin
     real(r8_) :: v ! [m/s] value in the centre of the range-bin

     ! this type packs all elements of data that are added by the L2B processor
     ! to a single Mie range-gate for a measurement 
     ! (i.e. a 3.5 km accumulation of laserpulses)
     ! (reminder: Mie = the thin spectral line caused by aerosols)
     !
     real(r8_) :: alt_vcog   
     ! altitude for the vertical center of gravity for this range gate
     !
     ! extinction: ...
     ! 
     ! The calculation of extinction, Rayleigh-A, Rayleigh-B, and Mie signals
     ! depends on the L2A processing code to be written by DLR 
     ! from Pierre's algorithm
     !
     ! for our own implementation we need the following fields:
     ! but not for mie, just for rayleigh
     real(r8_) :: attenuation_above
     real(r8_) :: attenuation_this_bin
     real(r8_) :: extinction
     !
     real(r8_) :: backscatter_ratio
     !
     ! the following are the classification outputs
     !
     ! can be: cloud, no_cloud (later also: dust? or others?)
     integer :: class 
     ! note, since at this point, it is not yet sure that the Mie and
     ! Rayleigh range-bin definitions will be identical, both should
     ! be classified separately.
     ! 
     ! more precise value then the one defined in the MeasMap
     real(r8_) :: weight 
  END TYPE L2B_MieRangeGateType

     !  #]

  TYPE L2B_MieMeasurementType
     !  #[
     ! this type packs all elements of data that describe a Mie measurement,
     ! i.e. a vertical profile of MieRangeGateType elements
     !
     integer :: nr_of_range_gates
     TYPE(L2B_MieRangeGateType), dimension(:), pointer :: RangeGate
  END TYPE L2B_MieMeasurementType
     !  #]

  TYPE L2B_RayleighRangeGateType
     !  #[
     ! this type packs all elements of data that are added by the L2B processor
     ! to a single Rayleigh range-gate for a measurement
     ! (i.e. a 3.5 km accumulation of laserpulses)
     ! (reminder: Rayleigh = the wide spectral line caused by molecules)
     !
     ! NWP data for this meas. range bin
     ! P and T may be needed to calculate the rayleigh backscatter 
     ! This cannot be established until the L2A code has been examined.
     ! If required, these should be obtained by requesting a 
     ! weighted average of the pre-processed auxiliary meteorological data
     !
! REMARK: these will be the horizontally interpolated
!         values, to be added in a later version
     ! real(r8_) :: P ! pressure in centre of the range-bin
     ! real(r8_) :: T ! temperature in centre of the range-bin

     ! temporarily added for the sensitivity study at KNMI:
     real(r8_) :: p ! [Pa]  value in the centre of the range-bin
     real(r8_) :: T ! [K]   value in the centre of the range-bin
     real(r8_) :: u ! [m/s] value in the centre of the range-bin
     real(r8_) :: v ! [m/s] value in the centre of the range-bin
     !
     real(r8_) :: alt_vcog   
     ! altitude for the vertical center of gravity for this range gate
     !
     real(r8_) :: calc_mol_backscatter ! can be calculated from P and T
     real(r8_) :: calc_mol_extinction  ! can be calculated from P and T
     !
     ! Remarks: following Ad's email of 08-04-2005 the following changes
     !          will probably be needed in the future:
     ! P and T: to be interpolated or copied from the auxiliary data 
     !          needed to calculate mol_backscatter
     ! Rayleigh-A, Rayleigh-B: from this we will have to calculate 
     !          the obs_rayleigh_signal_level. Maybe a simple
     !          summation or averaging will do?
     !
     ! extinction: ...
     ! 
     ! The calculation of extinction, Rayleigh-A, Rayleigh-B, and Mie signals
     ! depends on the L2A processing code to be written by DLR 
     ! from Pierre's algorithm
     !
     ! for our own implementation we need the following fields:
     real(r8_) :: attenuation_above
     real(r8_) :: attenuation_this_bin
     real(r8_) :: extinction
     !
     real(r8_) :: backscatter_ratio
     ! 
     ! the following are the classification outputs
     !
     ! can be: cloud, no_cloud (later also: dust? or others?)
     integer :: class 
     ! note, since at this point, it is not yet sure that the Mie and
     ! Rayleigh range-bin definitions will be identical, both should
     ! be classified separately.
     ! 
     ! more precise value then the one defined in the MeasMap
     real(r8_) :: weight 
  END TYPE L2B_RayleighRangeGateType
     !  #]

  TYPE L2B_RayleighMeasurementType
     !  #[
     ! this type packs all elements of data that describe a Rayleigh 
     ! measurement, i.e. a vertical profile of RayleighRangeGateType elements
     !
     integer :: nr_of_range_gates
     TYPE(L2B_RayleighRangeGateType), dimension(:), pointer :: RangeGate
  END TYPE L2B_RayleighMeasurementType
     !  #]

  TYPE L2B_ObsRangeGateType
     !  #[
     ! this type packs all elements of data that describe a single range-gate
     !
     real(r8_) :: alt_start ! bottom altitude of this range gate
     real(r8_) :: alt_cog   ! altitude for the center of gravity 
                            ! for this range gate
     real(r8_) :: alt_end   ! top altitude fo this range gate
     ! nr_of_measurements used to construct this observation
     integer  :: nr_of_measurements
     ! elements describing the start/stop geolocation, for which
     ! the hlos for this observation is determined
     real(r8_) :: lat_start
     real(r8_) :: lat_stop
     real(r8_) :: lon_start
     real(r8_) :: lon_stop
     ! elements describing the centre of gravity, for which
     ! the hlos for this observation is determined
     real(r8_) :: lat_cog   
     real(r8_) :: lon_cog   
!     real(r8_) :: date_cog
!     real(r8_) :: time_cog
     ! elements describing the reference temperature and pressure,
     ! as used for Rayleigh-Brillouin correction of the hlos
     real(r8_) :: T_ref
     real(r8_) :: p_ref
     ! the final result of the L2B processor: the hlos wind component
     real(r8_) :: hlos_cog
  END TYPE L2B_ObsRangeGateType
     !  #]

  TYPE L2B_ObservationType
     !  #[
     ! this type packs all elements of data that describe an observation,
     ! i.e. a vertical profile of RangeGateType elements, constructuted
     ! from the array of measurement profiles, for a single class
     ! like no_cloud, or cloud-points
     !
     integer :: observation_type
     integer :: nr_of_range_gates
     TYPE(L2B_ObsRangeGateType), dimension(:), pointer :: RangeGate
  END TYPE L2B_ObservationType
     !  #]

! end of working arrays definition

  TYPE L2BC_HDR_DataType
     !  #[

     ! datastructures used to construct the L2B output file
     logical :: FH_is_available 
     logical :: MPH_is_available 
     logical :: SPH_is_available 
     TYPE(FH_type),       pointer :: FH
     TYPE(MPH_type),      pointer :: MPH
     TYPE(L2BC_SPH_type), pointer :: SPH
     !
  END TYPE L2BC_HDR_DataType

     !  #]

  TYPE L2BC_BRC_DataType
     !  #[

     ! this defines the nr of measurements in the current BRC
     integer :: nr_of_measurements
     !
     ! this defines all observations in the current BRC
     integer :: nr_of_observations
     TYPE(L2B_ObservationType), dimension(:), pointer :: Obs
     !
     !
     ! the 4 attached L2B-datasets
     TYPE(L2B_Geoloc_ADS_type),   pointer :: Geoloc_ADS
     TYPE(L2B_PCD_ADS_type),      pointer :: PCD_ADS
     TYPE(L2B_Mie_MDS_type),      pointer :: Mie_MDS
     TYPE(L2B_Rayleigh_MDS_type), pointer :: Rayleigh_MDS
     !
     ! the 5 external L2B-references are stored as DSDs only
     !TYPE(L1B_type),              pointer :: Aeolus_Level_1B_Product
     !TYPE(L2B_Met_MDS_type),      pointer :: Aux_Met_Product
     !TYPE(L2B_RBC_ADS_type),      pointer :: Aeolus_RBC_Product
     !TYPE(L2B_CLM_ADS_type),      pointer :: Aux_CLM_Product
     !TYPE(L2B_Param_ADS_type),    pointer :: Level_2B_Proc_Params
     !
     ! DataSets only present for L2C
     !
     ! the 2 external L2C-references are stored as DSDs only
     !TYPE(L2B_type),              pointer :: Aeolus_Level_2B_Product
     !TYPE(L2C_Param_ADS_type),    pointer :: Level_2C_Proc_Params
     !
     ! the 3 attached L2C-datasets
     TYPE(L2C_AssimPCD_ADS_type),    pointer :: Assim_PCD_ADS
     TYPE(L2C_MieVec_MDS_type),      pointer :: Mie_VecWind_MDS
     TYPE(L2C_RayleighVec_MDS_type), pointer :: Rayleigh_VecWind_MDS

  END TYPE L2BC_BRC_DataType
     !  #]

  !  #]
  !  #[ Module interface
  !     everything is public in this module
  !  #]
contains
  !---------------------------------------------
  subroutine Init_L2B_ObsRangeGate(ObsRangegate)
    !  #[
    TYPE(L2B_ObsRangeGateType), intent(out) :: ObsRangegate
    ObsRangegate%alt_start          = missing_indicator_real
    ObsRangegate%alt_cog            = missing_indicator_real
    ObsRangegate%alt_end            = missing_indicator_real
    ObsRangegate%nr_of_measurements = 0
    !
    ObsRangegate%lat_start          = missing_indicator_real
    ObsRangegate%lat_stop           = missing_indicator_real
    ObsRangegate%lon_start          = missing_indicator_real
    ObsRangegate%lon_stop           = missing_indicator_real
    !
    ObsRangegate%lat_cog            = missing_indicator_real
    ObsRangegate%lon_cog            = missing_indicator_real
!    ObsRangegate%date_cog           = missing_date
!    ObsRangegate%time_cog           = missing_time
    ObsRangegate%T_ref              = missing_indicator_real
    ObsRangegate%p_ref              = missing_indicator_real
    ObsRangegate%hlos_cog           = missing_indicator_real

  end subroutine Init_L2B_ObsRangeGate
    !  #]
  subroutine Init_L2B_Observation(Obs,nr_of_range_gates,error_flag)
    !  #[
    TYPE(L2B_ObservationType), intent(out) :: Obs     ! output
    integer, intent(in)  :: nr_of_range_gates ! input
    integer, intent(out) :: error_flag        ! output

    ! local variables
    integer :: i, AllocStatus

    error_flag = no_error
    Obs%observation_type = Obs_Type_undefined

    Obs%nr_of_range_gates = nr_of_range_gates
    allocate(Obs%RangeGate(nr_of_range_gates),STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation failure in subroutine InitObservations"
       error_flag = error_allocate
       return
    END IF

    DO i=1,Obs%nr_of_range_gates
       call Init_L2B_ObsRangeGate(Obs%RangeGate(i))
    END DO
  end subroutine Init_L2B_Observation
    !  #]
  subroutine Delete_L2B_Observation(Obs)
    !  #[
    TYPE(L2B_ObservationType), intent(inout) :: Obs
    IF (associated(Obs%RangeGate)) deallocate(Obs%RangeGate)

  end subroutine Delete_L2B_Observation
    !  #]
  !---------------------------------------------
  subroutine Init_L2BC_HDR_DataStructure(L2BC_Headers, error_flag)
    !  #[
    TYPE(L2BC_HDR_DataType),  intent(out) :: L2BC_Headers      ! output
    integer,                  intent(out) :: error_flag        ! output

    ! local variables
    integer :: AllocStatus

    ! USAGE of this init subroutine:
    ! The main L2B_processor should call this init routine only once!

    ! init
    error_flag = no_error

    L2BC_Headers%FH_is_available = .false.
    nullify(L2BC_Headers%FH)
    allocate(L2BC_Headers%FH ,STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation failure in subroutine Init_L2BC_HDR_DataStructure"
       error_flag = error_allocate
       return
    END IF
    call Init_FH(L2BC_Headers%FH,error_flag)
    IF (error_flag .ne. no_error) return

    L2BC_Headers%MPH_is_available = .false.
    nullify(L2BC_Headers%MPH)
    allocate(L2BC_Headers%MPH,STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation failure in subroutine Init_L2BC_HDR_DataStructure"
       error_flag = error_allocate
       return
    END IF
    call Init_MPH(L2BC_Headers%MPH,error_flag)
    IF (error_flag .ne. no_error) return

    L2BC_Headers%SPH_is_available = .false.
    nullify(L2BC_Headers%SPH)
    allocate(L2BC_Headers%SPH,STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation failure in subroutine Init_L2BC_HDR_DataStructure"
       error_flag = error_allocate
       return
    END IF
    call Init_L2BC_SPH(L2BC_Headers%SPH,error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine Init_L2BC_HDR_DataStructure
    !  #]
  !---------------------------------------------
  subroutine Init_L2BC_BRC_DataStructure(L2BC, N_Meas0, &
              N_Obs_Mie0, N_Obs_Rayleigh0, N_Range_Gates0, &
              error_flag, L2C_flag)
    !  #[
    TYPE(L2BC_BRC_DataType), intent(out) :: L2BC            ! output
    integer,         intent(in)  :: N_Meas0         ! input
    integer(IntAuc), intent(in)  :: N_Obs_Mie0      ! input
    integer(IntAuc), intent(in)  :: N_Obs_Rayleigh0 ! input
    integer,         intent(in)  :: N_Range_Gates0  ! input
    integer,         intent(out) :: error_flag      ! output
    logical,optional,intent(in)  :: L2C_flag        ! optional input

    ! local variable
    integer :: i, AllocStatus, N_Obs0
    logical :: L2C_flag_local

    ! USAGE of this init subroutine:
    ! The main L2B_processor calls this init routine twice!
    ! 1) At the first call it nullifies all pointers. This is needed
    !    even if a BRC is not processed, because otherwise the deallocate
    !    subroutine will give a segmentation fault and the program will crash
    !    In this call N_Meas0, N_Obs_Mie0, N_Obs_Rayleigh0, N_Range_Gates0
    !    should all be set to zero.
    ! 2) In L2BP_primary the dimensioning of all substructures is done
    !    by a second call to this init routine. Now N_Meas0, N_Obs_Mie0, 
    !    N_Obs_Rayleigh0, N_Range_Gates0 should all have proper values.

    ! interpret optional arugment(s)
    L2C_flag_local=.false.
    IF (present(L2C_flag)) L2C_flag_local=L2C_flag

    ! init
    error_flag = no_error

    N_Obs0 = N_Obs_Mie0 + N_Obs_Rayleigh0

    ! just copy this to the saved variable
    L2BC%nr_of_measurements = N_Meas0

    ! just copy this to the saved variable
    L2BC%nr_of_observations =  N_Obs0

    ! init the module datastructure

    IF (N_Obs0 .gt. 0) THEN
       allocate(L2BC%Obs(N_Obs0),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L2BC_DataStructure"
          error_flag = error_allocate
          return
       END IF
    ELSE
       nullify(L2BC%Obs)
    END IF

    IF (N_Obs0 .gt. 0) THEN
       DO i=1,N_Obs0
          call Init_L2B_Observation(L2BC%Obs(i),N_Range_Gates0,error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    END IF

    ! the 4 attached datasets
    NullifyDataSets: IF ( (N_Meas0         .eq. 0) .and. &
                          (N_Obs_Mie0      .eq. 0) .and. &
                          (N_Obs_Rayleigh0 .eq. 0) .and. &
                          (N_Range_Gates0  .eq. 0) ) THEN
       nullify(L2BC%Geoloc_ADS)
       nullify(L2BC%PCD_ADS)
       nullify(L2BC%Mie_MDS)
       nullify(L2BC%Rayleigh_MDS)
       nullify(L2BC%Assim_PCD_ADS)
       nullify(L2BC%Mie_VecWind_MDS)
       nullify(L2BC%Rayleigh_VecWind_MDS)
    ELSE
       ! allocation of these datasets will be moved into the readl2bcdata.F90
       ! module, but for portability, allow the old method as well
       ! (i.e. giving values for N_Meas, N_Obs_Mie0, etc. )

       ! Poli: moved all the initialization steps in this procedure
       allocate(L2BC%Rayleigh_MDS,&
                L2BC%Mie_MDS, &
                L2BC%PCD_ADS, &
                L2BC%Geoloc_ADS, &
                stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine "//&
               "Init_L2BC_DataStructure for L2B-datasets"
          error_flag = error_allocate
          return
       ENDIF

       call Init_L2B_Rayleigh_MDS(L2BC%Rayleigh_MDS, &
            int(N_Meas0,IntAs), int(N_Obs_Rayleigh0,IntAs), &
            int(N_Range_Gates0,IntAs), error_flag)
       IF (error_flag .ne. no_error) return

       call Init_L2B_Mie_MDS(L2BC%Mie_MDS, &
            int(N_Meas0,IntAs), int(N_Obs_Mie0,IntAs), &
            int(N_Range_Gates0,IntAs), error_flag)
       IF (error_flag .ne. no_error) return
       
       call Init_L2B_PCD_ADS(L2BC%PCD_ADS, &
            int(N_Meas0,IntAs), &
            int(N_Obs_Mie0,IntAs), &
            int(N_Obs_Rayleigh0,IntAs), &
            int(N_Range_Gates0,IntAs), error_flag)
       IF (error_flag .ne. no_error) return
       
       call Init_L2B_Geoloc_ADS(L2BC%Geoloc_ADS, &
            int(N_Obs_Mie0,IntAs), &
            int(N_Obs_Rayleigh0,IntAs), &
            int(N_Range_Gates0,IntAs), error_flag)
       IF (error_flag .ne. no_error) return
       
       L2C_Case: IF (L2C_flag_local) THEN
          ! print *,'DT-Debug: initializing L2C-related datasets'

          allocate(L2BC%Assim_PCD_ADS,       &
               L2BC%Mie_VecWind_MDS,     &
               L2BC%Rayleigh_VecWind_MDS,&
               stat=AllocStatus)
          IF (AllocStatus .ne. 0) THEN
             print *,"Allocation failure in subroutine Init_L2BC_DataStructure"//&
                  " for L2C-datasets"
             error_flag = error_allocate
             return
          ENDIF

          call Init_L2C_AssimPCD_ADS(L2BC%Assim_PCD_ADS, &
               int(N_Meas0,IntAs), &
               int(N_Obs_Mie0,IntAs), &
               int(N_Obs_Rayleigh0,IntAs), &
               int(N_Range_Gates0,IntAs), error_flag)
          IF (error_flag .ne. no_error) return
          
          call Init_L2C_MieVec_MDS(L2BC%Mie_VecWind_MDS, &
               int(N_Meas0,IntAs), int(N_Obs_Mie0,IntAs), &
               int(N_Range_Gates0,IntAs), error_flag)
          IF (error_flag .ne. no_error) return
          
          call Init_L2C_RayleighVec_MDS(L2BC%Rayleigh_VecWind_MDS, &
               int(N_Meas0,IntAs), int(N_Obs_Rayleigh0,IntAs), &
               int(N_Range_Gates0,IntAs), error_flag)
          IF (error_flag .ne. no_error) return
       ELSE
          nullify(L2BC%Assim_PCD_ADS)
          nullify(L2BC%Mie_VecWind_MDS)
          nullify(L2BC%Rayleigh_VecWind_MDS)
       END IF L2C_Case ! L2C_flag_local
    END IF NullifyDataSets

  end subroutine Init_L2BC_BRC_DataStructure
    !  #]
  subroutine Delete_L2BC_HDR_DataStructure(L2BC_Headers)
    !  #[
    TYPE(L2BC_HDR_DataType), intent(inout) :: L2BC_Headers ! input/output
    ! (contents will be deallocated after this routine)

    ! clean up the header datastructures

    IF (associated(L2BC_Headers%FH)) THEN
       deallocate(L2BC_Headers%FH )
    ENDIF
    IF (associated(L2BC_Headers%MPH)) THEN
       deallocate(L2BC_Headers%MPH)
    ENDIF
    IF (associated(L2BC_Headers%SPH)) THEN
       call Delete_L2BC_SPH(L2BC_Headers%SPH)
       deallocate(L2BC_Headers%SPH)
    ENDIF

  end subroutine Delete_L2BC_HDR_DataStructure
    !  #]
  subroutine Delete_L2BC_BRC_DataStructure(L2BC, L2C_flag)
    !  #[

    TYPE(L2BC_BRC_DataType), intent(inout) :: L2BC ! input/output
    ! (contents will be deallocated after this routine)

    ! this flag is not really necessary, and could be removed. JK
    logical,     optional, intent(in)  :: L2C_flag

    ! local variables
    integer :: i
    logical :: L2C_flag_local

    ! interpret optional arugment(s)
    L2C_flag_local=.false.
    IF (present(L2C_flag)) L2C_flag_local=L2C_flag

    ! Delete binary data sub-structures

    IF (associated(L2BC%Obs)) THEN
       IF (L2BC%nr_of_observations > 0) THEN
          DO i=1,L2BC%nr_of_observations
             call Delete_L2B_Observation(L2BC%Obs(i))
          END DO
          deallocate(L2BC%Obs)
       ENDIF
    END IF

    ! clean up the L2B-related datasets
    IF (associated(L2BC%Geoloc_ADS)) &
         call Delete_L2B_Geoloc_ADS(L2BC%Geoloc_ADS)
    IF (associated(L2BC%PCD_ADS)) &
         call Delete_L2B_PCD_ADS(L2BC%PCD_ADS)
    IF (associated(L2BC%Mie_MDS)) &
         call Delete_L2B_Mie_MDS(L2BC%Mie_MDS)
    IF (associated(L2BC%Rayleigh_MDS)) &
         call Delete_L2B_Rayleigh_MDS(L2BC%Rayleigh_MDS)

    IF (associated(L2BC%Geoloc_ADS))   deallocate(L2BC%Geoloc_ADS)
    IF (associated(L2BC%PCD_ADS))      deallocate(L2BC%PCD_ADS)
    IF (associated(L2BC%Mie_MDS))      deallocate(L2BC%Mie_MDS)
    IF (associated(L2BC%Rayleigh_MDS)) deallocate(L2BC%Rayleigh_MDS)

    ! Potential extra test
    ! associated_L2C_dataset=associated(L2BC%Assim_PCD_ADS).or.&
    !                        associated(L2BC%Mie_VecWind_ADS).or.&
    !                        associated(L2BC%Rayleigh_VecWind_ADS)
    !
    ! IF ( (.not.L2C_flag_local) .and.  associated_L2C_dataset) THEN
    !      print *,"Error: L2C-datasets should not exist for"//&
    !              " L2B-datastructure in subroutine Delete_L2BC_DataStructure"
    !      error_flag = error_programming
    !      return
    ! ENDIF

    ! clean up the L2C-related datasets
    IF (L2C_flag_local) THEN
       ! print *,'DT-Debug: deleting L2C-related datasets'

       IF (associated(L2BC%Assim_PCD_ADS)) &
            call Delete_L2C_AssimPCD_ADS(L2BC%Assim_PCD_ADS)
       IF (associated(L2BC%Mie_VecWind_MDS)) &
            call Delete_L2C_MieVec_MDS(L2BC%Mie_VecWind_MDS)
       IF (associated(L2BC%Rayleigh_VecWind_MDS)) &
            call Delete_L2C_RayleighVec_MDS(L2BC%Rayleigh_VecWind_MDS)

       IF (associated(L2BC%Assim_PCD_ADS)) &
           deallocate(L2BC%Assim_PCD_ADS)
       IF (associated(L2BC%Mie_VecWind_MDS)) &
           deallocate(L2BC%Mie_VecWind_MDS)
       IF (associated(L2BC%Rayleigh_VecWind_MDS)) &
           deallocate(L2BC%Rayleigh_VecWind_MDS)
    END IF ! L2C_flag_local

  end subroutine Delete_L2BC_BRC_DataStructure

    !  #]
  subroutine Empty_L2BC_BinaryDatasets(L2BC)
    !  #[
    TYPE(L2BC_BRC_DataType), intent(inout) :: L2BC ! input/output
    ! (contents will be deallocated after this routine)

    ! do not clean up the header datastructures

    ! clean up the L2B-related datasets
    IF (associated(L2BC%Geoloc_ADS)) &
         call Delete_L2B_Geoloc_ADS(L2BC%Geoloc_ADS)
    IF (associated(L2BC%PCD_ADS)) &
         call Delete_L2B_PCD_ADS(L2BC%PCD_ADS)
    IF (associated(L2BC%Mie_MDS)) &
         call Delete_L2B_Mie_MDS(L2BC%Mie_MDS)
    IF (associated(L2BC%Rayleigh_MDS)) &
         call Delete_L2B_Rayleigh_MDS(L2BC%Rayleigh_MDS)

    IF (associated(L2BC%Geoloc_ADS))   deallocate(L2BC%Geoloc_ADS)
    IF (associated(L2BC%PCD_ADS))      deallocate(L2BC%PCD_ADS)
    IF (associated(L2BC%Mie_MDS))      deallocate(L2BC%Mie_MDS)
    IF (associated(L2BC%Rayleigh_MDS)) deallocate(L2BC%Rayleigh_MDS)

    IF (associated(L2BC%Assim_PCD_ADS)) &
         call Delete_L2C_AssimPCD_ADS(L2BC%Assim_PCD_ADS)
    IF (associated(L2BC%Mie_VecWind_MDS)) &
         call Delete_L2C_MieVec_MDS(L2BC%Mie_VecWind_MDS)
    IF (associated(L2BC%Rayleigh_VecWind_MDS)) &
         call Delete_L2C_RayleighVec_MDS(L2BC%Rayleigh_VecWind_MDS)
    
    IF (associated(L2BC%Assim_PCD_ADS)) &
         deallocate(L2BC%Assim_PCD_ADS)
    IF (associated(L2BC%Mie_VecWind_MDS)) &
         deallocate(L2BC%Mie_VecWind_MDS)
    IF (associated(L2BC%Rayleigh_VecWind_MDS)) &
         deallocate(L2BC%Rayleigh_VecWind_MDS)

  end subroutine Empty_L2BC_BinaryDatasets
    !  #]
  subroutine Print_L2BC_HDR_DataStructure(L2BC_Headers, prefix, &
                                          error_flag, L2C_flag)
    !  #[ ascii output to screen
    TYPE(L2BC_HDR_DataType), intent(in)  :: L2BC_Headers  ! input
    character(len=*),        intent(in)  :: prefix        ! input
    integer,                 intent(out) :: error_flag    ! output
    logical,   optional,     intent(in)  :: L2C_flag      ! optional input

    call AsciiWrite_L2BC_HDR_DataStruct(L2BC_Headers, &
                                        fileunit_stdout, prefix, &
                                        error_flag, L2C_flag)

  end subroutine Print_L2BC_HDR_DataStructure
    !  #]
  subroutine Print_L2BC_BRC_DataStructure(L2BC, prefix, &
                     PrintAll,error_flag, &
                     WriteGeoloc,WritePCD,WriteMie,WriteRayleigh,&
                     L2C_flag,&
                     WriteAssimPCD,WriteMieVecWind,WriteRaylVecWind)
    !  #[ ascii output to screen
    TYPE(L2BC_BRC_DataType), intent(in)  :: L2BC ! input
    character(len=*),    intent(in)  :: prefix        ! input
    logical,             intent(in)  :: PrintAll      ! input
    integer,             intent(out) :: error_flag    ! output
    logical,             intent(in)  :: WriteGeoloc   ! input
    logical,             intent(in)  :: WritePCD      ! input
    logical,             intent(in)  :: WriteMie      ! input
    logical,             intent(in)  :: WriteRayleigh ! input
    logical,   optional, intent(in)  :: L2C_flag         ! optional input
    logical,   optional, intent(in)  :: WriteAssimPCD    ! optional input
    logical,   optional, intent(in)  :: WriteMieVecWind  ! optional input
    logical,   optional, intent(in)  :: WriteRaylVecWind ! optional input

    call AsciiWrite_L2BC_BRC_DataStruct(L2BC, fileunit_stdout, prefix, &
                     PrintAll,error_flag, &
                     WriteGeoloc,WritePCD,WriteMie,WriteRayleigh,&
                     L2C_flag,&
                     WriteAssimPCD,WriteMieVecWind,WriteRaylVecWind)

  end subroutine Print_L2BC_BRC_DataStructure
    !  #]
  subroutine AsciiWrite_L2BC_HDR_DataStruct(L2BC_Headers, fileunit, prefix, &
                     error_flag, L2C_flag)
    !  #[ ascii output to a file
    TYPE(L2BC_HDR_DataType), intent(in)  :: L2BC_Headers  ! input
    integer,             intent(in)  :: fileunit      ! input
    character(len=*),    intent(in)  :: prefix        ! input
    integer,             intent(out) :: error_flag    ! output
    logical,   optional, intent(in)  :: L2C_flag         ! optional input

    ! init
    error_flag = no_error

    call AsciiWrite_FH(L2BC_Headers%FH, fileunit,&
                       prefix=trim(prefix)//"%FH",error_flag=error_flag)
    IF(error_flag .ne. no_error) return
    call AsciiWrite_MPH(L2BC_Headers%MPH,fileunit,&
                        prefix=trim(prefix)//"%MPH",error_flag=error_flag)
    IF(error_flag .ne. no_error) return
    call AsciiWrite_L2BC_SPH(L2BC_Headers%SPH,fileunit,&
                             prefix=trim(prefix)//"%SPH",&
                             error_flag=error_flag,L2C_flag=L2C_flag)
    IF(error_flag .ne. no_error) return

  end subroutine AsciiWrite_L2BC_HDR_DataStruct
    !  #]
  subroutine AsciiWrite_L2BC_BRC_DataStruct(L2BC, fileunit, prefix, &
                     PrintAll,error_flag, &
                     WriteGeoloc,WritePCD,WriteMie,WriteRayleigh,&
                     L2C_flag,&
                     WriteAssimPCD,WriteMieVecWind,WriteRaylVecWind)
    !  #[ ascii output to a file
    TYPE(L2BC_BRC_DataType), intent(in)  :: L2BC ! input
    integer,             intent(in)  :: fileunit      ! input
    character(len=*),    intent(in)  :: prefix        ! input
    logical,             intent(in)  :: PrintAll      ! input
    integer,             intent(out) :: error_flag    ! output
    logical,             intent(in)  :: WriteGeoloc   ! input
    logical,             intent(in)  :: WritePCD      ! input
    logical,             intent(in)  :: WriteMie      ! input
    logical,             intent(in)  :: WriteRayleigh ! input
    logical,   optional, intent(in)  :: L2C_flag         ! optional input
    logical,   optional, intent(in)  :: WriteAssimPCD    ! optional input
    logical,   optional, intent(in)  :: WriteMieVecWind  ! optional input
    logical,   optional, intent(in)  :: WriteRaylVecWind ! optional input
    
    ! NOTE that the PrintAll switch is used to decide whether the arrays
    ! should be printed completely, or just first and last element.

    ! local variables
    logical :: L2C_flag_local
    logical :: do_WriteAssimPCD
    logical :: do_WriteMieVecWind
    logical :: do_WriteRaylVecWind

    ! init
    error_flag = no_error

    L2C_flag_local = .false.
    IF (present(L2C_flag)) L2C_flag_local = L2C_flag

    do_WriteAssimPCD    = .false.
    do_WriteMieVecWind  = .false.
    do_WriteRaylVecWind = .false.
    if (present(WriteAssimPCD))    do_WriteAssimPCD    = WriteAssimPCD
    if (present(WriteMieVecWind))  do_WriteMieVecWind  = WriteMieVecWind
    if (present(WriteRaylVecWind)) do_WriteRaylVecWind = WriteRaylVecWind

    ! the 4 attached L2B datasets
    IF (WriteGeoloc) THEN
       call AsciiWrite_L2B_Geoloc_ADS(L2BC%Geoloc_ADS, fileunit, &
                                      prefix=trim(prefix)//"%Geoloc_ADS",&
                                      PrintAll=PrintAll, error_flag=error_flag)
       IF(error_flag .ne. no_error) return
    END IF

    IF (WritePCD) THEN
       call AsciiWrite_L2B_PCD_ADS(L2BC%PCD_ADS, fileunit, &
                                   prefix=trim(prefix)//"%PCD_ADS",&
                                   PrintAll=PrintAll, error_flag=error_flag)
       IF(error_flag .ne. no_error) return
    END IF

    IF (WriteMie) THEN
       call AsciiWrite_L2B_Mie_MDS(L2BC%Mie_MDS, fileunit, &
                                   prefix=trim(prefix)//"%Mie_MDS",&
                                   PrintAll=PrintAll, error_flag=error_flag)
       IF(error_flag .ne. no_error) return
    END IF

    IF (WriteRayleigh) THEN
       call AsciiWrite_L2B_Rayleigh_MDS(L2BC%Rayleigh_MDS, fileunit, &
                                      prefix=trim(prefix)//"%Rayleigh_MDS",&
                                      PrintAll=PrintAll,error_flag=error_flag)
       IF(error_flag .ne. no_error) return
    END IF

    IF (do_WriteAssimPCD) THEN
       IF (L2C_flag_local) THEN
          call AsciiWrite_L2C_AssimPCD_ADS(L2BC%Assim_PCD_ADS,fileunit,&
                                     prefix=trim(prefix)//"%Assim_PCD_ADS",&
                                     PrintAll=PrintAll,error_flag=error_flag)
       ELSE
          call logmsg(log_error,"ERROR in AsciiWrite_L2BC_DataStructure: "//&
               " switch WriteAssimPCD=.true. may only be used in "//&
               "combination with L2C_flag=.true., but currently it seems "//&
               "L2C_flag=.false. !!!")
          error_flag = error_programming
          return
       END IF
    END IF

    IF (do_WriteMieVecWind) THEN
       IF (L2C_flag_local) THEN
          call AsciiWrite_L2C_MieVec_MDS(L2BC%Mie_VecWind_MDS,fileunit,&
                                prefix=trim(prefix)//"%Mie_VecWind_MDS",&
                                PrintAll=PrintAll,error_flag=error_flag)
       ELSE
          call logmsg(log_error,"ERROR in AsciiWrite_L2BC_DataStructure: "//&
               " switch WriteMieVecWind=.true. may only be used in "//&
               "combination with L2C_flag=.true., but currently it seems "//&
               "L2C_flag=.false. !!!")
          error_flag = error_programming
          return
       END IF
    END IF

    IF (do_WriteRaylVecWind) THEN
       IF (L2C_flag_local) THEN
          call AsciiWrite_L2C_RaylVec_MDS(L2BC%Rayleigh_VecWind_MDS,fileunit,&
                           prefix=trim(prefix)//"%Rayleigh_VecWind_MDS",&
                           PrintAll=PrintAll,error_flag=error_flag)
       ELSE
          call logmsg(log_error,"ERROR in AsciiWrite_L2BC_DataStructure: "//&
               " switch WriteRaylVecWind=.true. may only be used in "//&
               "combination with L2C_flag=.true., but currently it seems "//&
               "L2C_flag=.false. !!!")
          error_flag = error_programming
          return
       END IF
    END IF

  end subroutine AsciiWrite_L2BC_BRC_DataStruct
    !  #]
  subroutine AsciiWrite_L2BC_hlos_Data(L2BC, Fileunit, error_flag, &
                                       WriteMie, WriteRayleigh, L2C_flag)
    !  #[ ascii output to a file

    TYPE(L2BC_BRC_DataType), intent(in) :: L2BC          ! input
    integer,             intent(in)  :: Fileunit      ! input
    integer,             intent(out) :: error_flag    ! output
    logical,             intent(in)  :: WriteMie      ! input
    logical,             intent(in)  :: WriteRayleigh ! input
    logical,  optional,  intent(in)  :: L2C_flag      ! optional input

    ! local variables
    integer            :: o
    logical            :: L2C_flag_local
    !character(len=3)   :: L2BC_string ! not yet used
    logical            :: associated_L2C_dataset

    ! Interpret optional argument(s)
    L2C_flag_local=.false.
    IF (present(L2C_flag)) L2C_flag_local=L2C_flag

    ! not yet used
    !L2BC_string=set_L2BC_string(L2C_flag_local)

    ! init
    error_flag = no_error

    write(Fileunit,*) "=================================="
    write(Fileunit,*) "summary of the hlos results:"
    write(Fileunit,*) "=================================="

    IF (WriteMie) THEN
       IF (associated(L2BC%Mie_MDS)) THEN
          IF (L2BC%Mie_MDS%N_Obs_Mie_Actual .eq. 0) THEN
             write(Fileunit,*) "L2BC%Mie_MDS%N_Obs_Mie_Actual = ",&
                  L2BC%Mie_MDS%N_Obs_Mie_Actual
             write(Fileunit,*) "No Mie hlos results available !"
          ELSE
             DO o=1,L2BC%Mie_MDS%N_Obs_Mie_Actual
                call AsciiWrite_L2B_Mie_hlos_profile(L2BC%Mie_MDS,&
                                                     error_flag, o, Fileunit)
             END DO
          END IF
       ELSE
          print *,"ERROR in AsciiWrite_L2BC_hlos_Data:"
          print *,"the Mie_MDS seems not associated ..."
          error_flag = error_programming
          return
       END IF
    END IF

    IF (WriteRayleigh) THEN
       IF (associated(L2BC%Rayleigh_MDS)) THEN
          IF (L2BC%Rayleigh_MDS%N_Obs_Rayleigh_Actual .eq. 0) THEN
             write(Fileunit,*) "L2BC%Rayleigh_MDS%N_Obs_Rayleigh_Actual = ",&
                  L2BC%Rayleigh_MDS%N_Obs_Rayleigh_Actual
             write(Fileunit,*) "No Rayleigh hlos results available !"
          ELSE
             DO o=1,L2BC%Rayleigh_MDS%N_Obs_Rayleigh_Actual
                call AsciiWrite_L2B_Rayl_hlos_prof(L2BC%Rayleigh_MDS,&
                                                   error_flag, o, Fileunit)
             END DO
          END IF
       ELSE
          print *,"ERROR in AsciiWrite_L2BC_hlos_Data:"
          print *,"the Rayleigh_MDS seems not associated ..."
          error_flag = error_programming
          return
       END IF
    END IF

    ! L2C-related datasets if appropriate
    associated_L2C_dataset=associated(L2BC%Assim_PCD_ADS).or.&
                           associated(L2BC%Mie_VecWind_MDS).or.&
                           associated(L2BC%Rayleigh_VecWind_MDS)

    IF ( (.not.L2C_flag_local) .and.  associated_L2C_dataset) THEN
      print *,"Error: L2C-datasets should not exist for"//&
              " L2B-datastructure in subroutine AsciiWrite_L2BC_hlos_Data"
      error_flag = error_programming
      return
   END IF
   IF (L2C_flag_local) THEN
       ! TODO: AsciiWrite for 3 L2C-datasets
       write(Fileunit,*) 'L2C-ascii writing not yet invoked'
    ENDIF

  end subroutine AsciiWrite_L2BC_hlos_Data
  !  #]
  !---------------------------------------------
  subroutine GetStatisticsL2BMieDS(L2B,Num_L2B_BRCs,&
                          M_Mie0,&
                          Num_Valid_Mie_Profiles,        &
                          Num_Clear_Mie_Profiles,        &
                          Num_Cloud_Mie_Profiles,        &
                          Num_Mie_Prof_Warm_Pulses,      &
                          Num_Profiles_Surface_Mie,      &
                          Num_Invalid_Meas_Prof_L2B_Mie, &
                          Num_Invalid_Meas_Bins_L2B_Mie, &
                          Num_Valid_Obs_L2B_Mie,         &
                          error_flag)
    !  #[ 
    ! *** gather some statistics from the Mie DS
    TYPE(L2BC_BRC_DataType), dimension(:), pointer :: L2B ! input
    integer,         intent(in)  :: Num_L2B_BRCs          ! input
    integer(IntAuc), intent(out) :: M_Mie0                        ! output
    integer(IntAus), intent(out) :: Num_Valid_Mie_Profiles        ! output
    integer(IntAus), intent(out) :: Num_Clear_Mie_Profiles        ! output
    integer(IntAus), intent(out) :: Num_Cloud_Mie_Profiles        ! output
    integer(IntAus), intent(out) :: Num_Mie_Prof_Warm_Pulses      ! output
    integer(IntAus), intent(out) :: Num_Profiles_Surface_Mie      ! output
    integer(IntAus), intent(out) :: Num_Invalid_Meas_Prof_L2B_Mie ! output
    integer(IntAus), intent(out) :: Num_Invalid_Meas_Bins_L2B_Mie ! output
    integer(IntAus), intent(out) :: Num_Valid_Obs_L2B_Mie         ! output
    integer,         intent(out) :: error_flag                    ! output

    ! local variables
    integer :: iBRC, Obs_Type, o, m, r, N_Meas
    logical :: valid, ground_present 

    TYPE(Mie_Profile_type),            pointer :: Mie_Profile
    TYPE(Meas_Map_Type),               pointer :: Map
    TYPE(L1B_Mie_Meas_Screening_Type), pointer :: L1B_Mie_Meas_Scr
    integer(IntAuc) :: Bin_QC

    ! init 
    error_flag = no_error

    ! max nr of mie profiles present in any BRC in the L2B output
    M_Mie0 = 0_IntAuc

    ! total nr of valid mie profiles in the L2B output
    Num_Valid_Mie_Profiles = 0_IntAus

    ! nr of clear (no cloud) mie profiles in the L2B output
    Num_Clear_Mie_Profiles = 0_IntAus

    ! nr of cloudy mie profiles in the L2B output
    Num_Cloud_Mie_Profiles = 0_IntAus

    ! nr of mie profiles using warm-up pulses
    ! NOTE: I need to add some WorkingData fields to determine this one
    Num_Mie_Prof_Warm_Pulses = 0_IntAus ! todo

    ! nr of mie profiles in L2B output extending down to the surface
    Num_Profiles_Surface_Mie = 0_IntAus

    ! nr of mie profiles valid in L1B input but flagged invalid
    ! at measurement level during L2B processing
    ! NOTE: I need to add some WorkingData fields to determine this one
    Num_Invalid_Meas_Prof_L2B_Mie = 0_IntAus ! todo

    ! nr of mie range-bins valid in L1B input but flagged invalid
    ! at measurement level during L2B processing
    ! NOTE: I need to add some WorkingData fields to determine this one
    Num_Invalid_Meas_Bins_L2B_Mie = 0_IntAus ! todo

    ! nr of valid rangebins in all valid L2B mie profiles
    ! NOTE that the meaning of this word "Obs" is from the viewpoint
    ! of the NWP user, i.e. nr. of hlos winds available,
    ! and differs from what we are used to within the ADM-Aeolus project
    Num_Valid_Obs_L2B_Mie = 0_IntAus

    ! loop over all BRC's and all observations per BRC
    ! to determine nr of valid clear and cloud profiles
    BrcLoop: DO iBRC=1,Num_L2B_BRCs
       MieAss: IF (associated(L2B(iBRC)%Mie_MDS)) THEN
          Map =>L2B(iBRC)%Mie_MDS%Map_of_L1B_Measurements_Used
          N_Meas = size(Map%Row)
          ObsLoop: DO o=1,L2B(iBRC)%Mie_MDS%N_Obs_Mie_Actual
             !  #[
             Mie_Profile => L2B(iBRC)%Mie_MDS%Mie_Profile(o)
             Obs_Type = Mie_Profile%Obs_Type
             select case (Obs_Type)
             case(Obs_Type_clear_returns)
                Num_Clear_Mie_Profiles = Num_Clear_Mie_Profiles + 1_IntAus
             case(Obs_Type_cloudy_returns)
                Num_Cloud_Mie_Profiles = Num_Cloud_Mie_Profiles + 1_IntAus
             end select

             ! now loop over the rangebins of this observation
             ! and count howmany have a valid result
             ! and datect which profiles extend down to the ground
             ground_present = .false.
             DO r=1,N_range_gates
                valid = Mie_Profile%Height_Bin(r)% Validity_Flag
                IF (valid) THEN
                   Num_Valid_Obs_L2B_Mie = Num_Valid_Obs_L2B_Mie + 1_IntAus
                END IF

                ! loop over the L1B bins used to construct this Obs.bin
                GroundDetLoop: DO m=1,N_Meas
                   IF (Map%Row(m)%Bin(r)%Which_L2B_Obs .eq. o) then
                      ! ok, this L1B bin was used, no see if it had
                      ! ground in it
                      L1B_Mie_Meas_Scr => &
                           L2B(iBRC)%PCD_ADS%L1B_Input_Screening%Mie_Meas(m)
                      Bin_QC = L1B_Mie_Meas_Scr%Bin_Screening(r)%Bin_QC
                      ! WARNING: this will not always work as expected
                      ! We fill Bin_QC with the first problem encountered,
                      ! so when scat-ratio is bas for example, no information
                      ! on ground detection is stored here.
                      ! TODO: We should define one of the flags in Bin_QC_Flags
                      ! to store the ground detection information:
                      !L1B_Mie_Meas_Scr%Bin_Screening(r)%Bin_QC_Flags
! note that the working data is not suitable to store this information because
! is is only kept for the duration of the processing of one BRC, while at
! this point in the code we need the information for all processed BRC's
                      select case (Bin_QC)
                      case(L1B_Mie_Bin_Contains_Surface, &
                           L1B_Mie_Contains_Surface_DEM, &
                           L1B_Mie_Bin_Below_Surface,    &
                           L1B_Mie_Below_Surface_DEM     )
                         ground_present = .true.
                         exit GroundDetLoop
                      end select
                   END IF
                END DO GroundDetLoop
             END DO
             IF (ground_present) THEN
                Num_Profiles_Surface_Mie = Num_Profiles_Surface_Mie + 1_IntAus
             END IF
             !  #]
          END DO ObsLoop
          MeasLoop: DO m=1,N_Meas
             ! check if this L1B Meas was valid
             L1B_Mie_Meas_Scr => &
                  L2B(iBRC)%PCD_ADS%L1B_Input_Screening%Mie_Meas(m)

             ! this checks whether the L2B processing lets this meas. pass
             !IF (L1B_Mie_Meas_Scr%Meas_QC .eq. L1B_Mie_Meas_OK)

             ! there seems to be no general flag to indicate that a L1B 
             ! measurement was valid or not, so we have to scan all rangebins
             ! and only if all of them are invalid the measurement itself
             ! also will be invalid:
             !   fields to check: Mie_Measurement_Invalid
             !                    Mie_Peak_Invalid?????

             ! not yet used
             ! L1B_Mie_Meas_Scr%Meas_QC_Flags

             ! check if the L2B processing did flag any of them as invalid
          END DO MeasLoop

          ! WARNING: using maxval() on 
          ! (/(M_Mie0,L2B(iBRC)%Mie_MDS%N_Obs_Mie_Actual/))
          ! results in a runtime error in my ifort compiler (v8.1)!
          ! This is a bug in this compiler, but for now please don't
          ! use that construct if just calling max() is possible as well
          M_Mie0 = max(M_Mie0,L2B(iBRC)%Mie_MDS%N_Obs_Mie_Actual)
       END IF MieAss
    END DO BrcLoop
    Num_Valid_Mie_Profiles = Num_Clear_Mie_Profiles + &
                             Num_Cloud_Mie_Profiles
        
    ! This is just a filler for now. To do: replace by adequate value.
    Num_Mie_Prof_Warm_Pulses      = 0

  end subroutine GetStatisticsL2BMieDS
    !  #]
  subroutine GetStatisticsL2BRayleighDS(L2B,Num_L2B_BRCs,&
                          M_Rayleigh0,                   &
                          Num_Valid_Rayleigh_Profiles,   &
                          Num_Clear_Rayleigh_Profiles,   &
                          Num_Cloud_Rayleigh_Profiles,   &
                          Num_Rayleigh_Prof_Warm_Pulses, &
                          Num_Profiles_Surface_Ray,      &
                          Num_Invalid_Meas_Prof_L2B_Ray, &
                          Num_Invalid_Meas_Bins_L2B_Ray, &
                          Num_Valid_Obs_L2B_Ray,         &
                          error_flag)
    !  #[ 
    ! *** gather some statistics from the Rayleigh DS
    TYPE(L2BC_BRC_DataType), dimension(:), pointer :: L2B ! input
    integer,         intent(in)  :: Num_L2B_BRCs                  ! input
    integer(IntAuc), intent(out) :: M_Rayleigh0                   ! output
    integer(IntAus), intent(out) :: Num_Valid_Rayleigh_Profiles   ! output
    integer(IntAus), intent(out) :: Num_Clear_Rayleigh_Profiles   ! output
    integer(IntAus), intent(out) :: Num_Cloud_Rayleigh_Profiles   ! output
    integer(IntAus), intent(out) :: Num_Rayleigh_Prof_Warm_Pulses ! output
    integer(IntAus), intent(out) :: Num_Profiles_Surface_Ray      ! output
    integer(IntAus), intent(out) :: Num_Invalid_Meas_Prof_L2B_Ray ! output
    integer(IntAus), intent(out) :: Num_Invalid_Meas_Bins_L2B_Ray ! output
    integer(IntAus), intent(out) :: Num_Valid_Obs_L2B_Ray         ! output
    integer,         intent(out) :: error_flag                    ! output
    
    ! local variables
    integer :: iBRC, Obs_Type, o

    ! init 
    error_flag = no_error

    ! these lines replace the Update_L2B_Rayleigh_SummaryVars() routine
    Num_Clear_Rayleigh_Profiles = 0_IntAus
    Num_Cloud_Rayleigh_Profiles = 0_IntAus
    M_Rayleigh0                 = 0_IntAuc

    ! Note: profiles without any valid results will probably not be reported
    ! in the L2B product, so counting them in a do loop is not possible
    ! Probably we will have to look at the difference between the nr of 
    ! observations in the L1B input file and in the L2B output
    ! I will look at this again after I determined the L1B statistics
    Num_Invalid_Meas_Prof_L2B_Ray = 0_IntAus

    ! Note: we can only partially determine this number here.
    ! For all observations not reported in the L2B product the
    ! same problem arises as for Num_Invalid_Meas_Prof_L2B_Mie above
    ! I will look at this again after I determined the L1B statistics
    Num_Invalid_Meas_Bins_L2B_Ray = 0_IntAus

    ! nr of mie profiles in L2B output extending down to the surface
    Num_Profiles_Surface_Ray = 0_IntAus ! todo

    ! this should report the nr of valid rangebins in all valid profiles
    ! the name Num_Valid_Obs_Bins_L2B_Ray would be more descriptive for that
    Num_Valid_Obs_L2B_Ray = 0_IntAus

    ! This is just a filler for now. To do: replace by adequate value.
    DO iBRC=1,Num_L2B_BRCs
       IF (associated(L2B(iBRC)%Rayleigh_MDS)) THEN
          DO o=1,L2B(iBRC)%Rayleigh_MDS%N_Obs_Rayleigh_Actual
             Obs_Type = L2B(iBRC)%Rayleigh_MDS%Rayleigh_Profile(o)%Obs_Type
             select case (Obs_Type)
             case(Obs_Type_clear_returns)
                Num_Clear_Rayleigh_Profiles = &
                     Num_Clear_Rayleigh_Profiles + 1_IntAus
             case(Obs_Type_cloudy_returns)
                Num_Cloud_Rayleigh_Profiles = &
                     Num_Cloud_Rayleigh_Profiles + 1_IntAus
             end select
          END DO
          M_Rayleigh0 = max(M_Rayleigh0,&
                            L2B(iBRC)%Rayleigh_MDS%N_Obs_Rayleigh_Actual)
       END IF
    END DO
    Num_Valid_Rayleigh_Profiles = Num_Clear_Rayleigh_Profiles + &
                                  Num_Cloud_Rayleigh_Profiles
  
    ! This is just a filler for now. To do: replace by adequate value.
    Num_Rayleigh_Prof_Warm_Pulses = 0

  end subroutine GetStatisticsL2BRayleighDS
    !  #]
  subroutine Fill_L2BC_FH(L2BC_Headers, FH_Default_Fields, &
                          L2BP_Settings, DateTimeStart, DateTimeStop, &
                          error_flag, L2C_flag)
  !  #[
    TYPE(L2BC_HDR_DataType),      intent(inout) :: L2BC_Headers ! updated
    TYPE(FH_Default_Fields_Type), intent(in) :: FH_Default_Fields ! input
    TYPE(L2BP_Settings_type),     intent(in) :: L2BP_Settings
    TYPE(DateTimeType), intent(in)  :: DateTimeStart ! struct(YYYY,MM,DD)
    TYPE(DateTimeType), intent(in)  :: DateTimeStop  ! struct(YYYY,MM,DD)
    integer,            intent(out) :: error_flag    ! output
    logical,  optional, intent(in)  :: L2C_flag

    ! this routine allocates and fills fixed header
    ! for the first L2BC datastructure in the array of L2BC datastructures
    
    ! local variables
    logical            :: L2C_flag_local
    integer            :: int_File_Version

    ! used to fill the FH
    character(len=256) :: file_type
! no longer used
    !character(len=256) :: file_description, notes, file_class

! this one is now imported from FH_Default_Fields
!    integer, parameter :: file_version=1,
! this one is filled used the JobOrder processing_station field
!    integer, parameter :: which_system

    error_flag = no_error
    
    L2C_flag_local=.false.
    IF (present(L2C_flag)) L2C_flag_local=L2C_flag

! this one is now imported from FH_Default_Fields   
!    IF (L2C_flag_local) THEN
!       file_description = "Level 2C Product"
!    ELSE
!       file_description = "Level 2B Product"
!    END IF

! these 2 are now imported from FH_Default_Fields   
!    notes            = "Room for some"//newline//"additional remarks"
!    file_class       = "TEST"

    IF (L2C_flag_local) THEN
       file_type        = file_type_L2C_N ! N=Nominal
    ELSE
       file_type        = file_type_L2B_N ! N=Nominal
    END IF

    ! convert this string to an integer, since the fill routine
    ! needs an integer as input 
    read(FH_Default_Fields%File_Version,"(i4)",err=999) int_File_Version

! obsolete. See SPR 28
    ! double check to make sure the software version defined in the 
    ! aeolusconstants.F90 is identical to the one used in the AuxPar
    ! FH_Default_Fields.
!    IF (software_version_L2BP .ne. FH_Default_Fields%Creator_Version) THEN
!       call logmsg(log_warn,"WARNING in Fill_L2BC_FH: "//&
!            "the AuxPar input file specifies the software_version in "//&
!            "FH_Default_Fields%Creator_Version as "//&
!            trim(FH_Default_Fields%Creator_Version)//&
!            " while the current software "//&
!            "has version "//trim(software_version_L2BP)//". Ignoring the "//&
!            "AuxPar value, and using "//trim(software_version_L2BP)//&
!            " instead")
!    END IF

    call  Fill_FH_items(L2BC_Headers%FH, &
              file_description    = FH_Default_Fields%File_Description, &
              notes               = L2BP_Settings%Notes, &
              file_class          = L2BP_Settings%File_Class, &
              file_type           = file_type, &
              ValidityPeriodStart = DateTimeStart, &
              ValidityPeriodStop  = DateTimeStop, &
              file_version        = int_File_Version, &
              system              = L2BP_Settings%System, &
              creator             = L2BP_Settings%Creator, &
              creator_version     = L2BP_Settings%Creator_Version, &
              error_flag          = error_flag)
    IF (error_flag .ne. no_error) return
    print *,'Fill_L2BC_Headers: Filled FH'

    return

999 call logmsg(log_error,"Error in Fill_L2BC_FH: "//&
                "Could not convert File_Version string to an integer")
    error_flag = error_reading_file
    return

  end subroutine Fill_L2BC_FH
  !  #]
  subroutine Fill_Remaining_L2BC_MPH_Items(MPH, FH, L2BP_Settings, &
                                           MPH_Default_Fields,&
                                           DateTimeStart, DateTimeStop, &
                                           error_flag, L2C_flag)
  !  #[

    TYPE(MPH_type),                pointer     :: MPH ! struct to be filled
    TYPE(FH_type),                 intent(in)  :: FH  ! input
    TYPE(L2BP_Settings_type),      intent(in)  :: L2BP_Settings
    TYPE(MPH_Default_Fields_Type), intent(in)  :: MPH_Default_Fields
    TYPE(DateTimeType), intent(in)  :: DateTimeStart ! struct(YYYY,MM,DD)
    TYPE(DateTimeType), intent(in)  :: DateTimeStop  ! struct(YYYY,MM,DD)
    integer,                       intent(out) :: error_flag 
    logical,             optional, intent(in)  :: L2C_flag

    ! this routine fills the MPH items that have not yet been copied
    ! from the L1B MPH (by subroutine Fill_L2B_MPH_From_L1B_MPH)
    ! for the first L2BC datastructure in the array of L2BC datastructures
    
    ! local variables
    logical            :: L2C_flag_local
    
    ! used to fill the MPH
    character(len=maxlen_Product)      :: product
    character(len=maxlen_Proc_Stage)   :: proc_stage
    character(len=maxlen_Ref_Doc)      :: ref_doc
    character(len=maxlen_Proc_Center)  :: proc_center
    character(len=maxlen_Software_Ver) :: software_version_auxpar
    logical            :: product_err
    integer(IntAd)     :: tot_size
    integer(IntAl)     :: sph_size
    integer(IntAl)     :: num_DSD
    integer(IntAl)     :: num_data_sets
    
    error_flag = no_error
    
    L2C_flag_local=.false.
    IF (present(L2C_flag)) L2C_flag_local=L2C_flag
   
    product          = FH%File_Name
    proc_stage       = "T"     ! to be moved into L2B_Proc_Settings.F90
    ! proc_stage might be: N (near-real-time), R (reprocessing) or T (test)
    ref_doc          = MPH_Default_Fields%Ref_Doc

    !Acquisition_Station is filled by copying the L1B value
    ! Spare 1

    ! use the JobOrder.Processing_Station field as required by SPR 28
    proc_center             = L2BP_Settings%System
    ! Proc_Time ! is filled using computer clock
    software_version_auxpar = MPH_Default_Fields%Software_Ver
    ! Spare 2

    ! Sensing_Start/Stop are copied from the L1B values
    ! but we need to redefine this in case only a selection
    ! of the L1B BRC's is processed.
    
    ! Spare 3

    ! all items between spare 3 and 6 are copied over from L1B

    ! Spare 6

    product_err   = .false. ! assume no errors are reported
! TODO: investigate what type of errors could be reported here
    tot_size      = 0_IntAd ! to be filled later
    sph_size      = 0_IntAl ! to be filled later
    
    ! Pre-set the number of DSDs and attached datasets
    IF (L2C_flag_local) THEN
       num_DSD       = int(Num_DSD_L2C,IntAl)
       ! Dsd_Size is a constant for EE files
       num_data_sets = int(Num_DS_L2C, IntAl)
    ELSE
       num_DSD       = int(Num_DSD_L2B,IntAl)
       ! Dsd_Size is a constant for EE files
       num_data_sets = int(Num_DS_L2B, IntAl)
    ENDIF

    ! double check to make sure the software_version specified in the
    ! AuxPar MPH_Default_Fields%Software_Ver field is equal to
    ! the version defined in AeolusConstants
    if (trim(software_version_auxpar) .ne. trim(software_version_L2BP)) then
       call logmsg(log_warn,"WARNING in Fill_Remaining_L2BC_MPH_Items: "//&
            "the AuxPar input file specifies the software_version in "//&
            "MPH_Default_Fields%Software_Ver as "//&
            trim(software_version_auxpar)//" while the current software "//&
            "has version "//trim(software_version_L2BP)//". Ignoring the "//&
            "AuxPar value, and using "//trim(software_version_L2BP)//&
            " instead")
    end if

    call Fill_MPH_items(MPH, product, proc_stage, ref_doc, proc_center, &
                        software_version_L2BP, product_err, &
                        tot_size, sph_size, num_DSD, num_data_sets, &
                        DateTimeStart, DateTimeStop, error_flag)
    IF (error_flag .ne. no_error) return
    print *,'Fill_L2BC_Headers: Filled MPH'

  end subroutine Fill_Remaining_L2BC_MPH_Items

  !  #]
  subroutine Fill_L2BC_SPH(L2BC_Headers, Num_L2B_BRCs, &
       lat_start, lat_stop, lon_start, lon_stop, &
       Sat_Track, M_Mie0, M_Rayleigh0, M_Meas0, &
       Num_Valid_Mie_Profiles,        Num_Valid_Rayleigh_Profiles,   &
       Num_Clear_Mie_Profiles,        Num_Clear_Rayleigh_Profiles,   &
       Num_Cloud_Mie_Profiles,        Num_Cloud_Rayleigh_Profiles,   &
       Num_Mie_Prof_Warm_Pulses,      Num_Rayleigh_Prof_Warm_Pulses, &
       Num_Profiles_Surface_Mie,      Num_Profiles_Surface_Ray,      &
       Num_Valid_Obs_Prof_L1B_Mie,    Num_Valid_Obs_Prof_L1B_Ray,    &
       Num_Valid_Meas_Prof_L1B_Mie,   Num_Valid_Meas_Prof_L1B_Ray,   &
       Num_Valid_Obs_Bins_L1B_Mie,    Num_Valid_Obs_Bins_L1B_Ray,    &
       Num_Valid_Meas_Bins_L1B_Mie,   Num_Valid_Meas_Bins_L1B_Ray,   &
       Num_Invalid_Meas_Prof_L1B_Mie, Num_Invalid_Meas_Prof_L1B_Ray, &
       Num_Invalid_Meas_Prof_L2B_Mie, Num_Invalid_Meas_Prof_L2B_Ray, &
       Num_Invalid_Meas_Bins_L1B_Mie, Num_Invalid_Meas_Bins_L1B_Ray, &
       Num_Invalid_Meas_Bins_L2B_Mie, Num_Invalid_Meas_Bins_L2B_Ray, &
       Num_Valid_Obs_L2B_Mie,         Num_Valid_Obs_L2B_Ray,         &
       error_flag, &
       L2C_flag,&
       Num_Invalid_Obs_L2C_Mie,       Num_Invalid_Obs_L2C_Ray)
  !  #[
    TYPE(L2BC_HDR_DataType), intent(inout) :: L2BC_Headers  ! input/output
    integer,         intent(in)  :: Num_L2B_BRCs        ! input
    integer(IntAl),  intent(in)  :: lat_start, lat_stop ! input
    integer(IntAl),  intent(in)  :: lon_start, lon_stop ! input
    real(r8_),       intent(in)  :: Sat_Track           ! input
    integer(IntAuc), intent(in)  :: M_Mie0                   ! input
    integer(IntAuc), intent(in)  :: M_Rayleigh0              ! input
    integer(IntAuc), intent(in)  :: M_Meas0                  ! input
    integer(IntAus), intent(in)  :: Num_Valid_Mie_Profiles        ! input
    integer(IntAus), intent(in)  :: Num_Valid_Rayleigh_Profiles   ! input
    integer(IntAus), intent(in)  :: Num_Clear_Mie_Profiles        ! input
    integer(IntAus), intent(in)  :: Num_Clear_Rayleigh_Profiles   ! input
    integer(IntAus), intent(in)  :: Num_Cloud_Mie_Profiles        ! input
    integer(IntAus), intent(in)  :: Num_Cloud_Rayleigh_Profiles   ! input
    integer(IntAus), intent(in)  :: Num_Mie_Prof_Warm_Pulses      ! input
    integer(IntAus), intent(in)  :: Num_Rayleigh_Prof_Warm_Pulses ! input
    integer(IntAus), intent(in)  :: Num_Profiles_Surface_Mie
    integer(IntAus), intent(in)  :: Num_Profiles_Surface_Ray
    integer(IntAus), intent(in)  :: Num_Valid_Obs_Prof_L1B_Mie
    integer(IntAus), intent(in)  :: Num_Valid_Obs_Prof_L1B_Ray
    integer(IntAus), intent(in)  :: Num_Valid_Meas_Prof_L1B_Mie
    integer(IntAus), intent(in)  :: Num_Valid_Meas_Prof_L1B_Ray
    integer(IntAus), intent(in)  :: Num_Valid_Obs_Bins_L1B_Mie
    integer(IntAus), intent(in)  :: Num_Valid_Obs_Bins_L1B_Ray
    integer(IntAus), intent(in)  :: Num_Valid_Meas_Bins_L1B_Mie
    integer(IntAus), intent(in)  :: Num_Valid_Meas_Bins_L1B_Ray
    integer(IntAus), intent(in)  :: Num_Invalid_Meas_Prof_L1B_Mie
    integer(IntAus), intent(in)  :: Num_Invalid_Meas_Prof_L1B_Ray
    integer(IntAus), intent(in)  :: Num_Invalid_Meas_Prof_L2B_Mie
    integer(IntAus), intent(in)  :: Num_Invalid_Meas_Prof_L2B_Ray
    integer(IntAus), intent(in)  :: Num_Invalid_Meas_Bins_L1B_Mie
    integer(IntAus), intent(in)  :: Num_Invalid_Meas_Bins_L1B_Ray
    integer(IntAus), intent(in)  :: Num_Invalid_Meas_Bins_L2B_Mie
    integer(IntAus), intent(in)  :: Num_Invalid_Meas_Bins_L2B_Ray
    integer(IntAus), intent(in)  :: Num_Valid_Obs_L2B_Mie
    integer(IntAus), intent(in)  :: Num_Valid_Obs_L2B_Ray
    integer,         intent(out) :: error_flag                    ! output
    ! L2C items
    logical,         optional, intent(in) :: L2C_flag       ! optional input
    integer(IntAus), optional, intent(in) :: Num_Invalid_Obs_L2C_Mie ! dito
    integer(IntAus), optional, intent(in) :: Num_Invalid_Obs_L2C_Ray ! dito

    ! this routine allocates and fills the SPH
    ! for the first L2BC datastructure in the array of L2BC datastructures
    ! The routine loops along this L2BC array to gather some statistical
    ! information to store in the SPH. It also copies some data from one
    ! dataset to some others (i.e. the N_Obs_*_Actual field)
    
    ! note that we should take the possibility into account
    ! that not a single BRC was processed, in which case we
    ! still need to produce a L2BC product with valid headers.
    ! Therefore we need to do associated tests on every dataset
    ! before we can use it.

    ! 30-Nov-2005  P. Poli    Original code. This routine fills in the FH, MPH,
    !                         and SPH as the L2BC processor makes it way
    !                         through the succesive BRCs.
    ! 30-Jan-2006  P. Poli    Added calls to Update_[PCD/Geoloc]_N_Obs_Actual
    ! 15-Feb-2007  J. de Kloe transformed to a routine that only needs to
    !                         be called once, which makes the L2B_SummaryVars
    !                         obsolete. Renamed from Handle_L2BHeaders
    !                         to Fill_L2B_Headers.
    !                         Also moved the FH and MPH filling to separate
    !                         routines, because these need a filled L1B
    !                         datastructure in memory, while these have all
    !                         been deallocated when the current routine is
    !                         called. Then renamed this routine again
    !                         now to Fill_L2BC_SPH.
    
    ! local variables
    logical            :: L2C_flag_local

    ! used to fill the SPH
    character(len=256) :: SPH_Descriptor
    
    ! init
    error_flag = no_error
    L2C_flag_local=.false.
    IF (present(L2C_flag)) L2C_flag_local=L2C_flag

    IF (L2C_flag_local) THEN
       SPH_Descriptor = "AEOLUS_L2C_SPECIFIC_HEADER"
    ELSE
       SPH_Descriptor = "AEOLUS_L2B_SPECIFIC_HEADER"
    END IF

    ! there is just one call now for this fill function,
    ! so all optional inputs should be converted back now into
    ! normal input parameters.
    call Fill_L2BC_SPH_items(L2BC_Headers%SPH,     &
         SPH_Descriptor, int(Num_L2B_BRCs,IntAus), &
         lat_start, lon_start, lat_stop, lon_stop, &
         Sat_Track, M_Mie0, M_Rayleigh0, M_Meas0,  &
         Num_Valid_Mie_Profiles,        Num_Valid_Rayleigh_Profiles,   &
         Num_Clear_Mie_Profiles,        Num_Clear_Rayleigh_Profiles,   &
         Num_Cloud_Mie_Profiles,        Num_Cloud_Rayleigh_Profiles,   &
         Num_Mie_Prof_Warm_Pulses,      Num_Rayleigh_Prof_Warm_Pulses, &
         Num_Profiles_Surface_Mie,      Num_Profiles_Surface_Ray,      &
         Num_Valid_Obs_Prof_L1B_Mie,    Num_Valid_Obs_Prof_L1B_Ray,    &
         Num_Valid_Meas_Prof_L1B_Mie,   Num_Valid_Meas_Prof_L1B_Ray,   &
         Num_Valid_Obs_Bins_L1B_Mie,    Num_Valid_Obs_Bins_L1B_Ray,    &
         Num_Valid_Meas_Bins_L1B_Mie,   Num_Valid_Meas_Bins_L1B_Ray,   &
         Num_Invalid_Meas_Prof_L1B_Mie, Num_Invalid_Meas_Prof_L1B_Ray, &
         Num_Invalid_Meas_Prof_L2B_Mie, Num_Invalid_Meas_Prof_L2B_Ray, &
         Num_Invalid_Meas_Bins_L1B_Mie, Num_Invalid_Meas_Bins_L1B_Ray, &
         Num_Invalid_Meas_Bins_L2B_Mie, Num_Invalid_Meas_Bins_L2B_Ray, &
         Num_Valid_Obs_L2B_Mie,         Num_Valid_Obs_L2B_Ray,         &
         error_flag, &
         L2C_flag                     = L2C_flag,                &
         Num_Invalid_Obs_L2C_Mie      = Num_Invalid_Obs_L2C_Mie, &
         Num_Invalid_Obs_L2C_Ray      = Num_Invalid_Obs_L2C_Ray   )

    IF (error_flag .ne. no_error) return
    print *,"Fill_L2BC_Headers: Filled SPH"

  end subroutine Fill_L2BC_SPH
  !  #]
  !---------------------------------------------
end module L2BC_DataStructure
