module L2BC_DS_Settings
  !  #[ Description
  !     a module to define dataset settings that
  !     allow distinction between L2B and L2C datsets.
  !     
  !     created:     26-Oct-2006, by: D. Tan, ECMWF
  !
  !     Modifications
  !      26-Oct-2006  D. Tan       Import from l2bc_datastructure
  !      15-Mar-2007  J. de Kloe   added -ZeroObs flag
  !      16-Mar-2007  J. de Kloe   added -Empty flag
  !      27-Mar-2007  J. de Kloe   added -200BRC flag
  !      16-Jan-2008  J. de Kloe   phase out integer kind i_ 
  !
  !  #]
  !  #[ modules used

  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE ErrorHandler, only: no_error, error_cmdline_options 
  USE Compiler_Features, only: iargc_aeolus,getarg_aeolus

  !  #]
  !  #[ Parameter and data type definitions
  implicit none

  ! strings for messages that could be either L2B or L2C
  character(len=3), parameter :: L2B_string    = "L2B"
  character(len=3), parameter :: L2C_string    = "L2C"

  !  #]
  !  #[ Variables
  !  #]
  !  #[ Module interface
  !     everything is public in this module
  !  #]
contains
  !---------------------------------------------
  subroutine Set_Default_L2BCds_Settings(L2C_flag,ZeroObs_flag,Empty_flag,&
                                         BRC200_flag)
    !  #[
    logical,           intent(out) :: L2C_flag     ! output
    logical, optional, intent(out) :: ZeroObs_flag ! output
    logical, optional, intent(out) :: Empty_flag   ! output
    logical, optional, intent(out) :: BRC200_flag  ! output

    ! these settings can be changed by setting the different
    ! commandline options

    L2C_flag     = .false.
    IF (present(ZeroObs_flag)) ZeroObs_flag = .false.
    IF (present(Empty_flag))   Empty_flag   = .false.
    IF (present(BRC200_flag))  BRC200_flag  = .false.

  end subroutine Set_Default_L2BCds_Settings
    !  #]
  subroutine Get_Cmdline_Options_L2BCds(L2C_flag,error_flag,&
                                        ZeroObs_flag,Empty_flag,BRC200_flag)
    !  #[
    logical,           intent(out) :: L2C_flag     ! output
    integer,           intent(out) :: error_flag   ! output
    logical, optional, intent(out) :: ZeroObs_flag ! output
    logical, optional, intent(out) :: Empty_flag   ! output
    logical, optional, intent(out) :: BRC200_flag  ! output

    ! local variables
    integer            :: nr_of_arguments, current_argument_nr
    character(len=256) :: current_argument

    ! init
    error_flag = no_error
    IF (present(ZeroObs_flag)) ZeroObs_flag = .false.
    IF (present(Empty_flag))   Empty_flag   = .false.
    IF (present(BRC200_flag))  BRC200_flag  = .false.

    nr_of_arguments = iargc_aeolus()
    current_argument_nr = 0  
        argloop: DO WHILE (current_argument_nr .lt. nr_of_arguments)
       !  #[ process the given commandline arguments
       current_argument_nr = current_argument_nr+1
       call getarg_aeolus(current_argument_nr,current_argument)

       SELECT CASE (trim(current_argument))
       CASE("-L2B")
          !  #[
          L2C_flag = .false.
          print *,'   option -L2B'
          !  #]
       CASE("-L2C")
          !  #[
          L2C_flag = .true.
          print *,'   option -L2C'
          !  #]
       CASE("-ZeroObs")
          !  #[
          IF (present(ZeroObs_flag)) THEN
             ZeroObs_flag = .true.
          ELSE
             CALL Print_Cmdline_Options_L2BCds;
             print *,"============================================"
             print *,"forbidden commandline option: ",trim(current_argument)
             print *,"This option is only allowed when the ZeroObs_flag is"
             print *,"present in the call of Get_Cmdline_Options_L2BCds()"
             print *,"============================================"
             error_flag = error_cmdline_options
             return
          END IF
          !  #]
       CASE("-Empty")
          !  #[
          IF (present(Empty_flag)) THEN
             Empty_flag = .true.
          ELSE
             CALL Print_Cmdline_Options_L2BCds;
             print *,"============================================"
             print *,"forbidden commandline option: ",trim(current_argument)
             print *,"This option is only allowed when the Empty_flag is"
             print *,"present in the call of Get_Cmdline_Options_L2BCds()"
             print *,"============================================"
             error_flag = error_cmdline_options
             return
          END IF
          !  #]
       CASE("-200BRC")
          !  #[
          IF (present(BRC200_flag)) THEN
             BRC200_flag = .true.
          ELSE
             CALL Print_Cmdline_Options_L2BCds;
             print *,"============================================"
             print *,"forbidden commandline option: ",trim(current_argument)
             print *,"This option is only allowed when the BRC200_flag is"
             print *,"present in the call of Get_Cmdline_Options_L2BCds()"
             print *,"============================================"
             error_flag = error_cmdline_options
             return
          END IF
          !  #]
       CASE DEFAULT
          !  #[
          CALL Print_Cmdline_Options_L2BCds;
          print *,"============================================"
          print *,"unrecognised commandline option: ",trim(current_argument)
          IF (current_argument(1:1) .ne. '-') THEN
             print *,"This argument does not start with: -"
             print *,"Maybe you forgot the - character to indicate that this"
             print *,"is an option, or maybe you supplied the wrong number"
             print *,"of items to an option."
          END IF
          print *,"============================================"
          error_flag = error_cmdline_options
          return
          !  #]
       END SELECT
       !  #]
    END DO argloop

    IF (present(ZeroObs_flag) .and. present(Empty_flag)) THEN
       IF (ZeroObs_flag .and. Empty_flag) THEN
          CALL Print_Cmdline_Options_L2BCds;
          print *,"============================================"
          print *,"forbidden combination of commandline options: "
          print *,"Using both -ZeroObs and -Empty flag is not possible."
          print *,"You have to choose:"
          print *,"1-generate a testfile with valid L1B input screening and"
          print *,"  pcd data, but without any observations."
          print *,"  use -ZeroObs to get this"
          print *,"2-generate a testfile without any dataset in it,"
          print *,"  this is what happens if all BRC's are rejected due to"
          print *,"  the Time_Interval setting in the JobOrder."
          print *,"  use -Empty to get this"
          print *,"============================================"
          error_flag = error_cmdline_options
          return
       END IF
    END IF

  end subroutine Get_Cmdline_Options_L2BCds
    !  #]
  subroutine Print_L2BCds_Settings(L2C_flag, ZeroObs_flag, Empty_flag, &
                                   BRC200_flag)
    !  #[
    logical,           intent(in) :: L2C_flag     ! input
    logical, optional, intent(in) :: ZeroObs_flag ! input
    logical, optional, intent(in) :: Empty_flag   ! input
    logical, optional, intent(in) :: BRC200_flag  ! input

    print *,"L2C_flag     = ", L2C_flag
    IF (present(ZeroObs_flag)) print *,"ZeroObs_flag = ",ZeroObs_flag
    IF (present(Empty_flag))   print *,"Empty_flag   = ",Empty_flag
    IF (present(BRC200_flag))  print *,"BRC200_flag  = ",BRC200_flag

  end subroutine Print_L2BCds_Settings
    !  #]
  subroutine Print_Cmdline_Options_L2BCds
    !  #[
    print *,""
    print *,"Possible commandline options for the main program:"
    print *,""
    print *,"<main_program>"
    print *,"   [-L2B|-L2C]"
    print *,"   [-ZeroObs]"
    print *,"   [-Empty]"
    print *,"   [-200BRC]"
    print *,""

  end subroutine Print_Cmdline_Options_L2BCds
    !  #]
  function set_L2BC_string(L2C_flag) result(L2BC_string)
    !  #[
    logical, intent(in)     :: L2C_flag
    character(len=3)        :: L2BC_string

    IF (L2C_flag) THEN
       L2BC_string=L2C_string
    ELSE
       L2BC_string=L2B_string
    ENDIF
  end function set_L2BC_string
    !  #]
  !---------------------------------------------
end module L2BC_DS_Settings
