! todo: remove all error prints, replace by logmsg()

module L2C_AssimPCD_ADS
  !  #[ Documentation
  !
  !     a module to handle the 
  !     L2C AssimilationProduct Confidence Data Annotation dataset
  !     as defined in AE-IF_ECMWF-L2BP-001 (L2B/L2C IODD), version ... 
  !     sec. ..., 
  !
  !     Written by D. Tan, ECMWF 
  !     created:     17-Oct-2006
  !     based on L2B_PCD_ADS by J. de Kloe, KNMI 
  !
  !     History
  !     17-Oct-2006 D. Tan      First version
  !     20-Oct-2006 D. Tan      Move to use of Set[IntA*s,FAdoFull]_or_Missing
  !     02-Mar-2007 J. de Kloe  adapted to parameter type change in call to 
  !                             GetObsTypeText()
  !     05-Mar-2007 J. de Kloe  adapted to IODD change
  !     15-Mar-2007 J. de Kloe  removed use of SummaryVars
  !     23-Apr-2007 J. de Kloe  phase out fillwithzero code
  !     26-Jun-2007 J. de Kloe  remove some unused variables
  !     16-Jan-2008 J. de Kloe  phase out integer kind i_ 
  !
  !  #]
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE L2B_parameters, only: Obs_Type_undefined, GetObsTypeText
  USE EE_CFI_Datatypes, only: IntAc, &
       IntAl, &
       IntAs,  c_bytelen_IntAs,  maxlen_IntAs,  SetIntAs_or_Missing,  &
       missing_indicator_IntAs,  &
       IntAuc, c_bytelen_IntAuc, maxlen_IntAuc, SetIntAuc, &
       IntAus, c_bytelen_IntAus, maxlen_IntAus, SetIntAus_or_Missing, &
       missing_indicator_IntAus, &
       FAdoxy, c_bytelen_FAdoxy, SetFAdoFull_or_Missing, &
       missing_indicator_FAdoxy, &
       write_FAdoxy_to_bin_file,  read_FAdoxy_from_bin_file, &
       write_IntAuc_to_bin_file,  read_IntAuc_from_bin_file, &
       write_IntAus_to_bin_file,  read_IntAus_from_bin_file, &
       write_IntAs_to_bin_file,   read_IntAs_from_bin_file, &
       write_SpareBytes_to_bin_file, read_SpareBytes_from_bin_file

  USE LunManager, only: fileunit_stdout
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE ErrorHandler, only: no_error, error_programming, error_allocate, &
                          error_writing_file
  !  #]
  !  #[ parameter and type definitions
  implicit none

  !---------------------------------------------------------------------
  !  #[ datastructures for L2B_Input_Screening
  !---------------------------------------------------------------------

  TYPE L2B_Input_Screening_Type
     !  #[
     integer(IntAuc) :: L2B_Obs_Screening
     integer(IntAuc) :: L2B_Obs_Screening_Flags1 ! to be defined later
     integer(IntAuc) :: L2B_Obs_Screening_Flags2 ! to be defined later
     integer(IntAuc) :: L2B_Obs_Screening_Flags3 ! to be defined later
     integer(IntAuc) :: L2B_Obs_Screening_Flags4 ! to be defined later
     integer(IntAuc) :: L2B_Obs_Screening_Flags5 ! to be defined later
     ! 20 spare bytes
  END TYPE L2B_Input_Screening_Type

  ! possible values to be used inside the above enum type
  ! file reading related
  integer(IntAuc), parameter :: L2B_File_OK       = 0
  integer(IntAuc), parameter :: L2B_HDR_Corrupted = 1
  integer(IntAuc), parameter :: L2B_DBL_Corrupted = 2
  ! etc 
    !  #]
  !  #]
  !---------------------------------------------------------------------
  !  #[ datastructures common to L2C_Mie_Quality_Params and 
  !                              L2C_Rayleigh_Qual_Params
  !---------------------------------------------------------------------
  TYPE HLOS_Observation_Errors_Type
     !  #[
     integer(IntAus) :: persistence_error           ! [cm/s]
     integer(IntAus) :: representativity_error      ! [cm/s]
     integer(IntAus) :: final_error                 ! [cm/s]
     integer(IntAs)  :: estimated_obs_bias          ! [cm/s]
     ! Spare_1: 20 spare bytes
  END TYPE HLOS_Observation_Errors_Type
     !  #]

  TYPE Assimilation_Model_PCD_Type
     !  #[
     TYPE(HLOS_Observation_Errors_Type) :: HLOS_Observation_Errors
     integer(IntAs)  :: background_hlos                  ! [cm/s]
     integer(IntAus) :: background_hlos_error            ! [cm/s]
     real(FAdoxy)    :: L2B_hlos_reliability             ! [Between zero and one]
     integer(IntAs)  :: analysis_hlos                    ! [cm/s]
     integer(IntAus) :: zonal_wind_background_error      ! [cm/s]
     integer(IntAus) :: meridional_wind_background_err   ! [cm/s]
     ! Spare_1: 16 spare bytes
  END TYPE Assimilation_Model_PCD_Type
     !  #]
  !  #]
  !---------------------------------------------------------------------
  !  #[ datastructures for L2C_Mie_Quality_Params
  !---------------------------------------------------------------------

  TYPE L2B_Mie_Obs_Screening_Type
     !  #[
     integer(IntAuc) :: L2B_Mie_Obs_QC ! enum field, see possible values below
     integer(IntAuc) :: L2B_Mie_Obs_QC_Flags
  END TYPE L2B_Mie_Obs_Screening_Type

  ! possible values to be used inside the above enum type
  integer(IntAuc), parameter :: L2B_Mie_Obs_OK               = 0
     !  #]

  TYPE L2C_Mie_Bin_QC_Type
     !  #[
     ! see sec. 4.3.5, table 63 of L2B/L2C-IODD, vers. 1.1, 21-Aug-2006
     TYPE(L2B_Mie_Obs_Screening_Type)  :: L2B_Mie_Obs_Screening
     TYPE(Assimilation_Model_PCD_Type) :: Assimilation_Model_PCD
     ! 10 spare bytes
!         1234567890123456789012345678901
  END TYPE L2C_Mie_Bin_QC_Type
     !  #]

  TYPE L2C_Mie_Quality_Params_Type
     !  #[
     integer(IntAuc) ::Obs_Type
     !
     ! Spare 1 (36 bytes reserved for Obs_Type_String)
     !
     TYPE(L2C_Mie_Bin_QC_Type), dimension(:), pointer :: L2C_Mie_Bin_QC
  END TYPE L2C_Mie_Quality_Params_Type
     !  #]

  !  #]
  !---------------------------------------------------------------------
  !  #[ datastructures for L2C_Rayleigh_Quality_Params
  !---------------------------------------------------------------------

  TYPE L2B_Rayleigh_Obs_Screening_Type
     !  #[
     integer(IntAuc) :: L2B_Rayleigh_Obs_QC ! enum field, see possible values below
     integer(IntAuc) :: L2B_Rayleigh_Obs_QC_Flags
  END TYPE L2B_Rayleigh_Obs_Screening_Type

  ! possible values to be used inside the above enum type
  integer(IntAuc), parameter :: L2B_Rayleigh_Obs_OK               = 0
     !  #]

  TYPE L2C_Rayleigh_Bin_QC_Type
     !  #[
     ! see sec. 4.3.5, table 69 of L2B/L2C-IODD, vers. 1.1, 21-Aug-2006
     TYPE(L2B_Rayleigh_Obs_Screening_Type)  :: L2B_Rayleigh_Obs_Screening
     TYPE(Assimilation_Model_PCD_Type) :: Assimilation_Model_PCD
     ! 10 spare bytes
!         1234567890123456789012345678901
  END TYPE L2C_Rayleigh_Bin_QC_Type
     !  #]

  TYPE L2C_Rayl_Quality_Params_Type
     !  #[
     integer(IntAuc) ::Obs_Type
     !
     ! Spare 1 (36 bytes reserved for Obs_Type_String)
     !
     TYPE(L2C_Rayleigh_Bin_QC_Type), dimension(:), pointer :: L2C_Rayleigh_Bin_QC
  END TYPE L2C_Rayl_Quality_Params_Type
     !  #]

  !  #]
  !---------------------------------------------------------------------

  TYPE L2C_AssimPCD_ADS_type 
     !  #[
! Poli: following telecon of 25-01-2006, it is decided that N_Obs_*_Actual
! refers to the number of Mie and Rayleigh profiles with meaningful values.
! Each BRC contains M_Mie and M_Rayleigh profiles (these two numbers are fixed
! for a given file), but in some BRCs there might be less than M_Mie profiles;
! in such case N_Obs_Mie_Actual is less than M_Mie. Same story for Rayleigh.
!   - when initialising, N_Obs_Mie_Actual is set to zero. The structure
!     is dimensioned using M_Mie.
!   - after each BRC is processed, a call to Handle_L2BHeaders will set
!     N_Obs_Mie_Actual to the proper value
!   - when writing, M_Mie profiles are written. The number N_Obs_Mie_Actual
!     is saved so that the user will know when reading the BRC that
!     only the first N_Obs_Mie_Actual profiles are meaningful.
!
!   - when reading, the structure should be dimensioned first to M_Mie.
!  Same story for Rayleigh.

! REMARK:
! N_Meas is not very usefull to have in this dataset, since all
! items are reported on observation scale!
! maybe its better to remove it in the next IODD update?
! JK, 14-mar-2007
     integer(IntAs)      :: N_Meas             ! number of measurements for this BRC
     integer(IntAs)      :: N_Obs_Mie_Actual   ! actual nr of Obs. for this BRC
     integer(IntAs)      :: N_Obs_Rayleigh_Actual ! ditto
     TYPE(L2B_Input_Screening_Type)    :: L2B_Input_Screening
     TYPE(L2C_Mie_Quality_Params_Type), &
                 dimension(:), pointer :: L2C_Mie_Quality_Params
     TYPE(L2C_Rayl_Quality_Params_Type), &
                 dimension(:), pointer :: L2C_Rayleigh_Quality_Params
     ! 20 spare bytes
  END TYPE L2C_AssimPCD_ADS_type
     !  #]

  ! the default verbosity setting is Err (as silent as possible)
  integer, private, save :: LocalVerbosity = VerbosityError

  !  #]
contains
  !---------------------------------------------
  subroutine Init_L2C_AssimPCD_ADS_Module(ChoosenVerbosity)
    !  #[
    integer, intent(in), optional :: ChoosenVerbosity

    IF (present(ChoosenVerbosity)) LocalVerbosity = ChoosenVerbosity

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"L2C_AssimPCD_ADS module has been initialised"
       call print_verbosity_setting(LocalVerbosity,&
                                    ModuleName="L2C_AssimPCD_ADS")
    END IF

  end subroutine Init_L2C_AssimPCD_ADS_Module
    !  #]
  !---------------------------------------------------------------------
  ! subroutines for the L2B_Input_Screening sub-datastructure    
  !---------------------------------------------------------------------
  !  #[

  !---------------------------------------------------------------------
  subroutine Init_L2B_Input_Screening(L2B_Input_Scr, error_flag)
    !  #[
    TYPE(L2B_Input_Screening_Type), intent(out) :: L2B_Input_Scr ! output
    integer,                        intent(out) :: error_flag    ! output

    error_flag = no_error

    L2B_Input_Scr%L2B_Obs_Screening        = L2B_File_OK
    L2B_Input_Scr%L2B_Obs_Screening_Flags1 = 0
    L2B_Input_Scr%L2B_Obs_Screening_Flags2 = 0
    L2B_Input_Scr%L2B_Obs_Screening_Flags3 = 0
    L2B_Input_Scr%L2B_Obs_Screening_Flags4 = 0
    L2B_Input_Scr%L2B_Obs_Screening_Flags5 = 0

  end subroutine Init_L2B_Input_Screening
    !  #]
  subroutine Delete_L2B_Input_Screening(L2B_Input_Scr)
    !  #[
    TYPE(L2B_Input_Screening_Type), intent(inout) :: L2B_Input_Scr

    ! No pointers so nothing to do

  end subroutine Delete_L2B_Input_Screening
    !  #]
  subroutine AsciiWrite_L2B_Input_Screening(&
                                      L2B_Input_Screening,&
                                      fileunit,prefix,&
                                      PrintAll,error_flag)
    !  #[
    TYPE(L2B_Input_Screening_type), intent(in)  :: L2B_Input_Screening ! input
    integer,                        intent(in)  :: fileunit            ! input
    character(len=*),               intent(in)  :: prefix              ! input
    logical,                        intent(in)  :: PrintAll            ! input
    integer,                        intent(out) :: error_flag          ! output

    ! local variables
    character(len=maxlen_IntAuc)    :: str_IntAuc

    call SetIntAuc(L2B_Input_Screening%L2B_Obs_Screening,str_IntAuc,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L2B_Obs_Screening = "//trim(str_IntAuc)

    call SetIntAuc(L2B_Input_Screening%L2B_Obs_Screening_Flags1,str_IntAuc,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L2B_Obs_Screening_Flags1 = "//trim(str_IntAuc)

    call SetIntAuc(L2B_Input_Screening%L2B_Obs_Screening_Flags2,str_IntAuc,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L2B_Obs_Screening_Flags2 = "//trim(str_IntAuc)

    call SetIntAuc(L2B_Input_Screening%L2B_Obs_Screening_Flags3,str_IntAuc,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L2B_Obs_Screening_Flags3 = "//trim(str_IntAuc)

    call SetIntAuc(L2B_Input_Screening%L2B_Obs_Screening_Flags4,str_IntAuc,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L2B_Obs_Screening_Flags4 = "//trim(str_IntAuc)

    call SetIntAuc(L2B_Input_Screening%L2B_Obs_Screening_Flags5,str_IntAuc,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L2B_Obs_Screening_Flags5 = "//trim(str_IntAuc)

    return

999 print *,"ERROR in AsciiWrite_L2B_Input_Screening:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2B_Input_Screening
    !  #]
  function GetByteSize_L2B_Input_Screening() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result

    c_bytesize =   c_bytelen_IntAuc + &
                 5*c_bytelen_IntAuc + &
                 20_IntAl ! spare_bytes

    ! typical sizes
    ! memory size = 1+5    =  6 bytes
    ! file   size = 1+5+20 = 26 bytes
  end function GetByteSize_L2B_Input_Screening
    !  #]
  subroutine Write_L2B_Inp_Screening(fileunit, &
             L2B_Input_Scr, error_flag)
    !  #[
    !  Writes the L2B_Input_Screening in binary format
    
    integer,                        intent(in)  :: fileunit      ! input
    TYPE(L2B_Input_Screening_Type), intent(in)  :: L2B_Input_Scr ! input
    integer,                        intent(out) :: error_flag    ! output

    ! local variable
    integer :: nr_of_zero_bytes

    !-----------------------------------
    ! writing the L2B_Input_Scr elements
    !-----------------------------------

    error_flag = no_error
    
    ! fill with actual values
    
    ! L2B_Obs_Screening
    ! integer(IntAuc) :: L2B_Obs_Screening
    call write_IntAuc_to_bin_file(fileunit, &
         L2B_Input_Scr%L2B_Obs_Screening,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! L2B_Obs_Screening_Flags1
    ! integer(IntAuc) :: L2B_Obs_Screening_Flags1 ! to be defined later
    call write_IntAuc_to_bin_file(fileunit, &
         L2B_Input_Scr%L2B_Obs_Screening_Flags1, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! L2B_Obs_Screening_Flags2
    ! integer(IntAuc) :: L2B_Obs_Screening_Flags2 ! to be defined later
    call write_IntAuc_to_bin_file(fileunit, &
         L2B_Input_Scr%L2B_Obs_Screening_Flags2, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! L2B_Obs_Screening_Flags3
    ! integer(IntAuc) :: L2B_Obs_Screening_Flags3 ! to be defined later
    call write_IntAuc_to_bin_file(fileunit, &
         L2B_Input_Scr%L2B_Obs_Screening_Flags3, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! L2B_Obs_Screening_Flags4
    ! integer(IntAuc) :: L2B_Obs_Screening_Flags4 ! to be defined later
    call write_IntAuc_to_bin_file(fileunit, &
         L2B_Input_Scr%L2B_Obs_Screening_Flags4, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! L2B_Obs_Screening_Flags5
    ! integer(IntAuc) :: L2B_Obs_Screening_Flags5 ! to be defined later
    call write_IntAuc_to_bin_file(fileunit, &
         L2B_Input_Scr%L2B_Obs_Screening_Flags5, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! 20 spare bytes
    nr_of_zero_bytes = 20
    call write_SpareBytes_to_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    return
 
998 print *,"ABORTING subroutine Write_L2B_Inp_Screening"
    return

  end subroutine Write_L2B_Inp_Screening
    !  #]
  subroutine Read_L2B_Inp_Screening(fileunit, &
             L2B_Input_Scr, error_flag)
    !  #[
    integer,                        intent(in)  :: fileunit
    TYPE(L2B_Input_Screening_Type), intent(inout):: L2B_Input_Scr
    integer,                        intent(out) :: error_flag

    ! local variable
    integer :: nr_of_zero_bytes

    !-----------------------------------
    ! reading the L2B_Input_Scr elements
    !-----------------------------------

    error_flag = no_error
    
    ! fill with actual values
    
    ! L2B_Obs_Screening
    ! integer(IntAuc) :: L2B_Obs_Screening
    call read_IntAuc_from_bin_file(fileunit, &
         L2B_Input_Scr%L2B_Obs_Screening,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! L2B_Obs_Screening_Flags1
    ! integer(IntAuc) :: L2B_Obs_Screening_Flags1 ! to be defined later
    call read_IntAuc_from_bin_file(fileunit, &
         L2B_Input_Scr%L2B_Obs_Screening_Flags1, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! L2B_Obs_Screening_Flags2
    ! integer(IntAuc) :: L2B_Obs_Screening_Flags2 ! to be defined later
    call read_IntAuc_from_bin_file(fileunit, &
         L2B_Input_Scr%L2B_Obs_Screening_Flags2, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! L2B_Obs_Screening_Flags3
    ! integer(IntAuc) :: L2B_Obs_Screening_Flags3 ! to be defined later
    call read_IntAuc_from_bin_file(fileunit, &
         L2B_Input_Scr%L2B_Obs_Screening_Flags3, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! L2B_Obs_Screening_Flags4
    ! integer(IntAuc) :: L2B_Obs_Screening_Flags4 ! to be defined later
    call read_IntAuc_from_bin_file(fileunit, &
         L2B_Input_Scr%L2B_Obs_Screening_Flags4, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! L2B_Obs_Screening_Flags5
    ! integer(IntAuc) :: L2B_Obs_Screening_Flags5 ! to be defined later
    call read_IntAuc_from_bin_file(fileunit, &
         L2B_Input_Scr%L2B_Obs_Screening_Flags5, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! 20 spare bytes
    nr_of_zero_bytes = 20
    call read_SpareBytes_from_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 print *,"ABORTING subroutine Read_L2B_Inp_Screening"
    return

  end subroutine Read_L2B_Inp_Screening
    !  #]

  !  #]
  !---------------------------------------------------------------------
  ! subroutines for the Assimilation_Model_PCD sub-datastructure, which
  ! is common to both L2C_Mie_Quality_Params and L2C_Rayleigh_Quality_Params
  !---------------------------------------------------------------------
  !  #[

  subroutine Init_HLOS_Observation_Errors(HLOS_Observation_Errors)
    !  #[
    TYPE(HLOS_Observation_Errors_Type), intent(out) :: HLOS_Observation_Errors ! output

    HLOS_Observation_Errors%persistence_error      = missing_indicator_IntAus
    HLOS_Observation_Errors%representativity_error = missing_indicator_IntAus
    HLOS_Observation_Errors%final_error            = missing_indicator_IntAus
    HLOS_Observation_Errors%estimated_obs_bias     = missing_indicator_IntAs

  end subroutine Init_HLOS_Observation_Errors
    !  #]
  subroutine Delete_HLOS_Observation_Errors(HLOS_Observation_Errors)
    !  #[
    TYPE(HLOS_Observation_Errors_Type), intent(inout) :: HLOS_Observation_Errors

    ! No pointers so nothing to do

  end subroutine Delete_HLOS_Observation_Errors
    !  #]
  subroutine AsciiWrite_HLOS_Obs_Errors(&
                                      HLOS_Observation_Errors,&
                                      fileunit,prefix,&
                                      PrintAll,error_flag)
    !  #[
    TYPE(HLOS_Observation_Errors_type), intent(in) :: &
                                             HLOS_Observation_Errors ! input
    integer,                  intent(in)  :: fileunit         ! input
    character(len=*),         intent(in)  :: prefix           ! input
    logical,                  intent(in)  :: PrintAll         ! input
    integer,                  intent(out) :: error_flag       ! output

    ! local variables
    character(len=maxlen_IntAs )    :: str_IntAs
    character(len=maxlen_IntAus)    :: str_IntAus

    error_flag = no_error

    call SetIntAus_or_Missing(HLOS_Observation_Errors%persistence_error,&
                              str_IntAus,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%persistence_error      = "//trim(str_IntAus)//" [cm/s]"

    call SetIntAus_or_Missing(HLOS_Observation_Errors%representativity_error,&
                              str_IntAus,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%representativity_error = "//trim(str_IntAus)//" [cm/s]"

    call SetIntAus_or_Missing(HLOS_Observation_Errors%final_error,&
                              str_IntAus,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%final_error            = "//trim(str_IntAus)//" [cm/s]"

    call SetIntAs_or_Missing(HLOS_Observation_Errors%estimated_obs_bias,&
                             str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%estimated_obs_bias     = "//trim(str_IntAs)//" [cm/s]"

    return

999 print *,"ERROR in AsciiWrite_HLOS_Obs_Errors:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_HLOS_Obs_Errors
    !  #]
  function GetByteSize_HLOS_Obs_Errors() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result

    c_bytesize = 0
    c_bytesize = c_bytesize + 3_IntAl * c_bytelen_IntAus ! 3 types of error
    c_bytesize = c_bytesize +           c_bytelen_IntAs  ! estimated_obs_bias
    c_bytesize = c_bytesize + 20_IntAl                   ! Spare_1

  end function GetByteSize_HLOS_Obs_Errors
    !  #]
  subroutine Write_HLOS_Observation_Errors(fileunit, &
             HLOS_Observation_Errors, error_flag)
    !  #[
    !  Writes the HLOS_Observation_Errors in binary format
    
    integer,     intent(in)  :: fileunit      ! input
    TYPE(HLOS_Observation_Errors_Type), &
                 intent(in)  :: HLOS_Observation_Errors ! input
    integer,     intent(out) :: error_flag    ! output

    ! local variable
    integer :: nr_of_zero_bytes

    !-----------------------------------
    ! writing the HLOS_Observation_Errors elements
    !-----------------------------------

    error_flag = no_error
    
    ! fill with actual values
    
    ! persistence_error
    ! integer(IntAus) :: persistence_error
    call write_IntAus_to_bin_file(fileunit, &
         HLOS_Observation_Errors%persistence_error,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! representativity_error
    ! integer(IntAus) :: representativity_error
    call write_IntAus_to_bin_file(fileunit, &
         HLOS_Observation_Errors%representativity_error,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! final_error
    ! integer(IntAus) :: final_error
    call write_IntAus_to_bin_file(fileunit, &
         HLOS_Observation_Errors%final_error,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! estimated_obs_bias
    ! integer(IntAs) :: estimated_obs_bias
    call write_IntAs_to_bin_file(fileunit, &
         HLOS_Observation_Errors%estimated_obs_bias,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! 20 spare bytes
    nr_of_zero_bytes = 20
    call write_SpareBytes_to_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    return

998 print *,"ABORTING subroutine Write_HLOS_Observation_Errors"
    return

  end subroutine Write_HLOS_Observation_Errors
    !  #]
  subroutine Read_HLOS_Observation_Errors(fileunit, &
             HLOS_Observation_Errors, error_flag)
    !  #[
    integer,     intent(in)    :: fileunit
    TYPE(HLOS_Observation_Errors_Type), &
                 intent(inout) :: HLOS_Observation_Errors
    integer,     intent(out)   :: error_flag

    ! local variable
    integer :: nr_of_zero_bytes

    !-----------------------------------
    ! reading the HLOS_Observation_Errors elements
    !-----------------------------------

    error_flag = no_error
    
    ! fill with actual values
    
    ! persistence_error
    ! integer(IntAus) :: persistence_error
    call read_IntAus_from_bin_file(fileunit, &
         HLOS_Observation_Errors%persistence_error,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! representativity_error
    ! integer(IntAus) :: representativity_error
    call read_IntAus_from_bin_file(fileunit, &
         HLOS_Observation_Errors%representativity_error,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! final_error
    ! integer(IntAus) :: final_error
    call read_IntAus_from_bin_file(fileunit, &
         HLOS_Observation_Errors%final_error,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! estimated_obs_bias
    ! integer(IntAs) :: estimated_obs_bias
    call read_IntAs_from_bin_file(fileunit, &
         HLOS_Observation_Errors%estimated_obs_bias,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! 20 spare bytes
    nr_of_zero_bytes = 20
    call read_SpareBytes_from_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    return

998 print *,"ABORTING subroutine Read_HLOS_Observation_Errors"
    return

  end subroutine Read_HLOS_Observation_Errors
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_Assimilation_Model_PCD(Assim_Model_PCD)
    !  #[
    TYPE(Assimilation_Model_PCD_Type), intent(out) :: Assim_Model_PCD ! output

    call Init_HLOS_Observation_Errors(Assim_Model_PCD%HLOS_Observation_Errors)

    Assim_Model_PCD%Background_hlos                = missing_indicator_IntAs
    Assim_Model_PCD%Background_hlos_error          = missing_indicator_IntAus
    Assim_Model_PCD%L2B_hlos_reliability           = missing_indicator_FAdoxy
    Assim_Model_PCD%Analysis_hlos                  = missing_indicator_IntAs
    Assim_Model_PCD%Zonal_Wind_Background_error    = missing_indicator_IntAus
    Assim_Model_PCD%Meridional_Wind_Background_err = missing_indicator_IntAus

  end subroutine Init_Assimilation_Model_PCD
    !  #]
  subroutine Delete_Assimilation_Model_PCD(Assim_Model_PCD)
    !  #[
    TYPE(Assimilation_Model_PCD_Type), intent(inout) :: Assim_Model_PCD

    ! No pointers so nothing to do

  end subroutine Delete_Assimilation_Model_PCD
    !  #]
  subroutine AsciiWrite_Assim_Model_PCD(&
                                      Assimilation_Model_PCD,&
                                      fileunit,prefix,&
                                      PrintAll,error_flag)
    !  #[
    TYPE(Assimilation_Model_PCD_type), intent(in) :: &
                                             Assimilation_Model_PCD ! input
    integer,                  intent(in)  :: fileunit         ! input
    character(len=*),         intent(in)  :: prefix           ! input
    logical,                  intent(in)  :: PrintAll         ! input
    integer,                  intent(out) :: error_flag       ! output

    ! local variables
    character(len=25)               :: str_FAdoxy
    character(len=maxlen_IntAs )    :: str_IntAs
    character(len=maxlen_IntAus)    :: str_IntAus

    error_flag = no_error

    call AsciiWrite_HLOS_Obs_Errors(&
         Assimilation_Model_PCD%HLOS_Observation_Errors, fileunit, &
         trim(prefix)//"%HLOS_Observation_Errors",&
         PrintAll,error_flag)
    IF (error_flag .ne. no_error) goto 999

    call SetIntAs_or_Missing(Assimilation_Model_PCD%Background_hlos,&
                             str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Background_hlos       = "//trim(str_IntAs)//" [cm/s]"

    call SetIntAus_or_Missing(Assimilation_Model_PCD%Background_hlos_error,&
                              str_IntAus,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Background_hlos_error = "//trim(str_IntAus)//" [cm/s]"

    call SetFAdoFull_or_Missing(Assimilation_Model_PCD%L2B_hlos_reliability,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) "REALACC(12)"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L2B_hlos_reliability  = "//trim(str_FAdoxy)
    write(fileunit,"(a)",err=999) "ENDREALACC"

    call SetIntAs_or_Missing(Assimilation_Model_PCD%Analysis_hlos,&
                             str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Analysis_hlos         = "//trim(str_IntAs)//" [cm/s]"

    call SetIntAus_or_Missing(Assimilation_Model_PCD%&
                              Zonal_Wind_Background_error,&
                              str_IntAus,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Zonal_Wind_background_error      = "//&
                     trim(str_IntAus)//" [cm/s]"

    call SetIntAus_or_Missing(Assimilation_Model_PCD%&
                              Meridional_Wind_Background_err,&
                              str_IntAus,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Meridional_Wind_background_error = "//&
                     trim(str_IntAus)//" [cm/s]"

    return

999 print *,"ERROR in AsciiWrite_Assim_Model_PCD:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Assim_Model_PCD
    !  #]
  function GetByteSize_Assim_Model_PCD() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result

    c_bytesize = 0
    c_bytesize = c_bytesize + GetByteSize_HLOS_Obs_Errors()
    c_bytesize = c_bytesize + c_bytelen_IntAs  ! Background_hlos
    c_bytesize = c_bytesize + c_bytelen_IntAus ! Background_hlos_error
    c_bytesize = c_bytesize + c_bytelen_FAdoxy ! L2B_hlos_reliability
    c_bytesize = c_bytesize + c_bytelen_IntAs  ! Analysis_hlos
    c_bytesize = c_bytesize + c_bytelen_IntAus ! Zonal_Wind_brackground_error
    c_bytesize = c_bytesize + c_bytelen_IntAus ! Meridional_Wind_brackground_error
    c_bytesize = c_bytesize + 20_IntAl         ! Spare_1

  end function GetByteSize_Assim_Model_PCD
    !  #]
  subroutine Write_Assimilation_Model_PCD(fileunit, &
             Assimilation_Model_PCD, error_flag)
    !  #[
    !  Writes the Assimilation_Model_PCD in binary format
    
    integer,     intent(in)  :: fileunit      ! input
    TYPE(Assimilation_Model_PCD_Type), &
                 intent(in)  :: Assimilation_Model_PCD ! input
    integer,     intent(out) :: error_flag    ! output

    ! local variable
    integer :: nr_of_zero_bytes

    !-----------------------------------
    ! writing the Assimilation_Model_PCD elements
    !-----------------------------------

    error_flag = no_error
    
    ! fill with actual values
    
    call write_HLOS_observation_errors(fileunit, &
         Assimilation_Model_PCD%HLOS_observation_errors,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! Background_hlos
    ! integer(IntAs) :: Background_hlos
    call write_IntAs_to_bin_file(fileunit, &
         Assimilation_Model_PCD%Background_hlos,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! Background_hlos_error
    ! integer(IntAus) :: Background_hlos_error
    call write_IntAus_to_bin_file(fileunit, &
         Assimilation_Model_PCD%Background_hlos_error,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! L2B_hlos_reliability
    ! real(FAdoxy) :: L2B_hlos_reliability
    call write_FAdoxy_to_bin_file(fileunit, &
         Assimilation_Model_PCD%L2B_hlos_reliability,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! Analysis_hlos
    ! integer(IntAs) :: Analysis_hlos
    call write_IntAs_to_bin_file(fileunit, &
         Assimilation_Model_PCD%Analysis_hlos,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! Zonal_Wind_Background_error
    ! integer(IntAus) :: Zonal_Wind_Background_error
    call write_IntAus_to_bin_file(fileunit, &
         Assimilation_Model_PCD%Zonal_Wind_Background_error,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! Meridional_Wind_Background_error
    ! integer(IntAus) :: Meridional_Wind_Background_err
    call write_IntAus_to_bin_file(fileunit, &
         Assimilation_Model_PCD%Meridional_Wind_Background_err,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! 20 spare bytes
    nr_of_zero_bytes = 20
    call write_SpareBytes_to_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    return

998 print *,"ABORTING subroutine Write_Assimilation_Model_PCD"
    return

  end subroutine Write_Assimilation_Model_PCD
    !  #]
  subroutine Read_Assimilation_Model_PCD(fileunit, &
             Assimilation_Model_PCD, error_flag)
    !  #[
    integer,     intent(in)   :: fileunit
    TYPE(Assimilation_Model_PCD_Type), &
                 intent(inout):: Assimilation_Model_PCD
    integer,     intent(out)  :: error_flag

    ! local variable
    integer :: nr_of_zero_bytes

    !-----------------------------------
    ! reading the Assimilation_Model_PCD elements
    !-----------------------------------

    error_flag = no_error
    
    ! fill with actual values
    
    call read_HLOS_observation_errors(fileunit, &
         Assimilation_Model_PCD%HLOS_observation_errors,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! Background_hlos
    ! integer(IntAs) :: Background_hlos
    call read_IntAs_from_bin_file(fileunit, &
         Assimilation_Model_PCD%Background_hlos,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! Background_hlos_error
    ! integer(IntAus) :: Background_hlos_error
    call read_IntAus_from_bin_file(fileunit, &
         Assimilation_Model_PCD%Background_hlos_error,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! L2B_hlos_reliability
    ! real(FAdoxy) :: L2B_hlos_reliability
    call read_FAdoxy_from_bin_file(fileunit, &
         Assimilation_Model_PCD%L2B_hlos_reliability,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! Analysis_hlos
    ! integer(IntAs) :: Analysis_hlos
    call read_IntAs_from_bin_file(fileunit, &
         Assimilation_Model_PCD%Analysis_hlos,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! Zonal_Wind_Background_error
    ! integer(IntAus) :: Zonal_Wind_Background_error
    call read_IntAus_from_bin_file(fileunit, &
         Assimilation_Model_PCD%Zonal_Wind_Background_error,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! Meridional_Wind_Background_error
    ! integer(IntAus) :: Meridional_Wind_Background_err
    call read_IntAus_from_bin_file(fileunit, &
         Assimilation_Model_PCD%Meridional_Wind_Background_err,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! 20 spare bytes
    nr_of_zero_bytes = 20
    call read_SpareBytes_from_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    return

998 print *,"ABORTING subroutine Read_Assimilation_Model_PCD"
    return

  end subroutine Read_Assimilation_Model_PCD
    !  #]

  !  #]
  !---------------------------------------------------------------------
  ! subroutines for the L2C_Mie_Quality_Params sub-datastructure
  !---------------------------------------------------------------------
  !  #[

  !---------------------------------------------------------------------
  subroutine Init_L2B_Mie_Obs_Screening(L2B_Mie_Obs_Screening)
    !  #[
    TYPE(L2B_Mie_Obs_Screening_Type), intent(out) :: L2B_Mie_Obs_Screening ! output

    L2B_Mie_Obs_Screening%L2B_Mie_Obs_QC       = L2B_Mie_Obs_OK
    L2B_Mie_Obs_Screening%L2B_Mie_Obs_QC_Flags = 0

  end subroutine Init_L2B_Mie_Obs_Screening
    !  #]
  subroutine Delete_L2B_Mie_Obs_Screening(L2B_Mie_Obs_Screening)
    !  #[
    TYPE(L2B_Mie_Obs_Screening_Type), intent(inout) :: L2B_Mie_Obs_Screening

    ! No pointers so nothing to do

  end subroutine Delete_L2B_Mie_Obs_Screening
    !  #]
  subroutine AsciiWrite_L2B_Mie_Obs_Screen(&
                                      L2B_Mie_Obs_Screening,&
                                      fileunit,prefix,&
                                      PrintAll,error_flag)
    !  #[
    TYPE(L2B_Mie_Obs_Screening_type), intent(in)  :: L2B_Mie_Obs_Screening ! input
    integer,                  intent(in)  :: fileunit         ! input
    character(len=*),         intent(in)  :: prefix           ! input
    logical,                  intent(in)  :: PrintAll         ! input
    integer,                  intent(out) :: error_flag       ! output

    ! local variables
    character(len=maxlen_IntAuc)    :: str_IntAuc

    call SetIntAuc(L2B_Mie_Obs_Screening%L2B_Mie_Obs_QC,&
                   str_IntAuc,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L2B_Mie_Obs_QC = "//trim(str_IntAuc)

    call SetIntAuc(L2B_Mie_Obs_Screening%L2B_Mie_Obs_QC_Flags,&
                   str_IntAuc,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L2B_Mie_Obs_QC_Flags = "//trim(str_IntAuc)

    return

999 print *,"ERROR in AsciiWrite_L2B_Mie_Obs_Screen:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2B_Mie_Obs_Screen
    !  #]
  function GetByteSize_L2B_Mie_Obs_Scr() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result

    c_bytesize = 0
    c_bytesize = c_bytesize + c_bytelen_IntAuc ! L2B_Mie_Obs_QC
    c_bytesize = c_bytesize + c_bytelen_IntAuc ! L2B_Mie_Obs_QC_Flags
    c_bytesize = c_bytesize + 16_IntAl         ! Spare_1

  end function GetByteSize_L2B_Mie_Obs_Scr
    !  #]
  subroutine Write_L2B_Mie_Obs_Screening(fileunit, &
             L2B_Mie_Obs_Screening, error_flag)
    !  #[
    !  Writes the L2B_Mie_Obs_Screening in binary format
    
    integer,     intent(in)  :: fileunit              ! input
    TYPE(L2B_Mie_Obs_Screening_Type), &
                 intent(in)  :: L2B_Mie_Obs_Screening ! input
    integer,     intent(out) :: error_flag            ! output

    ! local variable
    integer :: nr_of_zero_bytes

    !-----------------------------------
    ! writing the L2B_Mie_Obs_Screening elements
    !-----------------------------------

    error_flag = no_error
    
    ! fill with actual values
    
    ! L2B_Mie_Obs_QC
    ! integer(IntAuc) :: L2B_Mie_Obs_QC
    call write_IntAuc_to_bin_file(fileunit, &
         L2B_Mie_Obs_Screening%L2B_Mie_Obs_QC,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! L2B_Mie_Obs_QC_Flags1
    ! integer(IntAuc) :: L2B_Mie_Obs_QC_Flags ! to be defined later
    call write_IntAuc_to_bin_file(fileunit, &
         L2B_Mie_Obs_Screening%L2B_Mie_Obs_QC_Flags, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! 16 spare bytes
    nr_of_zero_bytes = 16
    call write_SpareBytes_to_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    return

998 print *,"ABORTING subroutine Write_L2B_Mie_Obs_Screening"
    return

  end subroutine Write_L2B_Mie_Obs_Screening
    !  #]
  subroutine Read_L2B_Mie_Obs_Screening(fileunit, &
             L2B_Mie_Obs_Screening, error_flag)
    !  #[
    integer,     intent(in)   :: fileunit
    TYPE(L2B_Mie_Obs_Screening_Type), &
                 intent(inout):: L2B_Mie_Obs_Screening
    integer,     intent(out)  :: error_flag

    ! local variable
    integer :: nr_of_zero_bytes

    !-----------------------------------
    ! reading the L2B_Mie_Obs_Screening elements
    !-----------------------------------

    error_flag = no_error
    
    ! fill with actual values
    
    ! L2B_Mie_Obs_QC
    ! integer(IntAuc) :: L2B_Mie_Obs_QC
    call read_IntAuc_from_bin_file(fileunit, &
         L2B_Mie_Obs_Screening%L2B_Mie_Obs_QC,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! L2B_Mie_Obs_QC_Flags
    ! integer(IntAuc) :: L2B_Mie_Obs_QC_Flags ! to be defined later
    call read_IntAuc_from_bin_file(fileunit, &
         L2B_Mie_Obs_Screening%L2B_Mie_Obs_QC_Flags, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! 16 spare bytes
    nr_of_zero_bytes = 16
    call read_SpareBytes_from_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 print *,"ABORTING subroutine Read_L2B_Mie_Obs_Screening"
    return

  end subroutine Read_L2B_Mie_Obs_Screening
    !  #]

  !---------------------------------------------------------------------
  subroutine Init_L2C_Mie_Bin_QC(L2C_Mie_Bin_QC)
    !  #[
    TYPE(L2C_Mie_Bin_QC_Type), intent(out) :: L2C_Mie_Bin_QC ! output

    call Init_L2B_Mie_Obs_Screening(L2C_Mie_Bin_QC%L2B_Mie_Obs_Screening)
    call Init_Assimilation_Model_PCD(L2C_Mie_Bin_QC%Assimilation_Model_PCD)

  end subroutine Init_L2C_Mie_Bin_QC
    !  #]
  subroutine Delete_L2C_Mie_Bin_QC(L2C_Mie_Bin_QC)
    !  #[
    TYPE(L2C_Mie_Bin_QC_Type), intent(inout) :: L2C_Mie_Bin_QC

    ! No pointers so nothing to do.

  end subroutine Delete_L2C_Mie_Bin_QC
    !  #]
  subroutine AsciiWrite_L2C_Mie_Bin_QC(&
                                      L2C_Mie_Bin_QC,&
                                      fileunit,prefix,&
                                      PrintAll,error_flag)
    !  #[
    TYPE(L2C_Mie_Bin_QC_type),intent(in)  :: L2C_Mie_Bin_QC   ! input
    integer,                  intent(in)  :: fileunit         ! input
    character(len=*),         intent(in)  :: prefix           ! input
    logical,                  intent(in)  :: PrintAll         ! input
    integer,                  intent(out) :: error_flag       ! output

    ! local variables
    ! none for now

    error_flag = no_error

    call AsciiWrite_L2B_Mie_Obs_Screen(&
         L2C_Mie_Bin_QC%L2B_Mie_Obs_Screening, fileunit, &
         trim(prefix)//"%L2B_Mie_Obs_Screening",&
         PrintAll,error_flag)
    IF (error_flag .ne. no_error) goto 999

    call AsciiWrite_Assim_Model_PCD(&
         L2C_Mie_Bin_QC%Assimilation_Model_PCD, fileunit, &
         trim(prefix)//"%Assimilation_Model_PCD",&
         PrintAll,error_flag)
    IF (error_flag .ne. no_error) goto 999

    return

999 print *,"ERROR in AsciiWrite_L2C_Mie_Bin_QC:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2C_Mie_Bin_QC
    !  #]
  function GetByteSize_L2C_Mie_Bin_QC() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result

    c_bytesize = 0
    c_bytesize = c_bytesize + GetByteSize_L2B_Mie_Obs_Scr()
    c_bytesize = c_bytesize + GetByteSize_Assim_Model_PCD()
    c_bytesize = c_bytesize + 10_IntAl ! spare_bytes

    ! typical sizes
    ! memory size = 18+64   =  82 bytes
    ! file   size = 18+64+10 = 92 bytes

  end function GetByteSize_L2C_Mie_Bin_QC
    !  #]
  subroutine Write_L2C_Mie_Bin_QC(fileunit, &
             L2C_Mie_Bin_QC, error_flag)
    !  #[
    integer,                      intent(in)  :: fileunit      ! input
    TYPE(L2C_Mie_Bin_QC_Type), intent(in)  :: L2C_Mie_Bin_QC ! input
    integer,                      intent(out) :: error_flag    ! output

    ! local variables
    integer :: nr_of_zero_bytes

    error_flag = no_error

    call Write_L2B_Mie_Obs_Screening(fileunit,L2C_Mie_Bin_QC%L2B_Mie_Obs_Screening,&
                                     error_flag)
    IF (error_flag .ne. no_error) goto 998

    call Write_Assimilation_Model_PCD(fileunit,L2C_Mie_Bin_QC%Assimilation_Model_PCD,&
                                      error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 10 spare bytes
    nr_of_zero_bytes = 10
    call write_SpareBytes_to_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 print *,"ABORTING subroutine Write_L2C_Mie_Bin_QC"
    return

  end subroutine Write_L2C_Mie_Bin_QC
    !  #]
  subroutine Read_L2C_Mie_Bin_QC(fileunit, &
             L2C_Mie_Bin_QC, error_flag)
    !  #[
    integer,                      intent(in)  :: fileunit
    TYPE(L2C_Mie_Bin_QC_Type), intent(inout):: L2C_Mie_Bin_QC
    integer,                      intent(out) :: error_flag

    ! local variables
    integer :: nr_of_zero_bytes

    error_flag = no_error

    call Read_L2B_Mie_Obs_Screening(fileunit,L2C_Mie_Bin_QC%L2B_Mie_Obs_Screening,&
                                    error_flag)
    IF (error_flag .ne. no_error) goto 998

    call Read_Assimilation_Model_PCD(fileunit,L2C_Mie_Bin_QC%Assimilation_Model_PCD,&
                                     error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 10 spare bytes
    nr_of_zero_bytes = 10
    call read_SpareBytes_from_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    !print *,"done with Read_L2C_Mie_Bin_QC"
    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Read_L2C_Mie_Bin_QC"
    return

  end subroutine Read_L2C_Mie_Bin_QC
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_L2C_Mie_Quality_Params(L2C_Mie_Quality_Params, &
                                    N_Range_Gates0, error_flag)
    !  #[
    TYPE(L2C_Mie_Quality_Params_Type), intent(out) :: L2C_Mie_Quality_Params ! output
    integer(IntAs), intent(in)  :: N_Range_Gates0 ! input
    integer,        intent(out) :: error_flag    ! output

    ! local variable
    integer :: r, AllocStatus

    error_flag = no_error

    L2C_Mie_Quality_Params%Obs_Type = Obs_Type_undefined

    IF (N_Range_Gates0 .eq. 0) THEN
       ! init them to be empty
       nullify(L2C_Mie_Quality_Params%L2C_Mie_Bin_QC)
    ELSE
       allocate(L2C_Mie_Quality_Params%L2C_Mie_Bin_QC(N_Range_Gates0),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L2C_Mie_Quality_Params"
          error_flag = error_allocate
          return
       END IF
       
       DO r=1,N_Range_Gates0
          call Init_L2C_Mie_Bin_QC(L2C_Mie_Quality_Params%L2C_Mie_Bin_QC(r))
       END DO
    END IF

  end subroutine Init_L2C_Mie_Quality_Params
    !  #]
  subroutine Delete_L2C_Mie_Quality_Params(L2C_Mie_Quality_Params)
    !  #[
    TYPE(L2C_Mie_Quality_Params_Type), intent(inout) :: L2C_Mie_Quality_Params

    ! local variable
    integer :: i

    IF (associated(L2C_Mie_Quality_Params%L2C_Mie_Bin_QC)) THEN
       DO i=1,size(L2C_Mie_Quality_Params%L2C_Mie_Bin_QC)
          call Delete_L2C_Mie_Bin_QC(&
               L2C_Mie_Quality_Params%L2C_Mie_Bin_QC(i))
       END DO
       deallocate(L2C_Mie_Quality_Params%L2C_Mie_Bin_QC)
    END IF

  end subroutine Delete_L2C_Mie_Quality_Params
    !  #]
  subroutine AsciiWrite_L2C_Mie_Qual_Params(&
                                      L2C_Mie_Quality_Params,&
                                      fileunit,prefix,&
                                      PrintAll,error_flag)
    !  #[
    TYPE(L2C_Mie_Quality_Params_type), intent(in)  :: L2C_Mie_Quality_Params ! input
    integer,                  intent(in)  :: fileunit         ! input
    character(len=*),         intent(in)  :: prefix           ! input
    logical,                  intent(in)  :: PrintAll         ! input
    integer,                  intent(out) :: error_flag       ! output

    ! local variables
    integer :: i,s,step
    character(len=maxlen_IntAuc)   :: str_IntAuc
    character(len=2)               :: str_cntr

    call SetIntAuc(L2C_Mie_Quality_Params%Obs_Type,str_IntAuc,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Obs_Type = "//trim(str_IntAuc)//" ("//&
         trim(GetObsTypeText(int(L2C_Mie_Quality_Params%Obs_Type,IntAuc)))//")"

    IF (associated(L2C_Mie_Quality_Params%L2C_Mie_Bin_QC)) THEN
       s=size(L2C_Mie_Quality_Params%L2C_Mie_Bin_QC)
       IF (PrintAll .or. (s .le. 1)) THEN
          step=1
       ELSE
          step = s - 1
       END IF
       DO i=1,s,step
          write(str_cntr,"(i2.2)",err=998) i
          call AsciiWrite_L2C_Mie_Bin_QC(&
               L2C_Mie_Quality_Params%L2C_Mie_Bin_QC(i), fileunit, &
               trim(prefix)//"%L2C_Mie_Bin_QC("//trim(str_cntr)//")",&
               PrintAll,error_flag)
          IF (error_flag .ne. no_error) goto 999
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%L2C_Mie_Bin_QC(:) [seems not associated]"
    END IF

    return

998 print *,"conversion error in AsciiWrite_L2C_Mie_Qual_Params:"
    print *,"could not convert the counter ",i," to a string .."
    error_flag = error_writing_file
    return

999 print *,"ERROR in AsciiWrite_L2C_Mie_Qual_Params:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2C_Mie_Qual_Params
    !  #]
  function GetByteSize_L2C_Mie_Qual_Params(L2C_Mie_Quality_Params, &
              M_Mie0) result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result
    TYPE(L2C_Mie_Quality_Params_Type), intent(in) :: L2C_Mie_Quality_Params
    integer,     optional,        intent(in) :: M_Mie0

    ! local variables
    integer :: s

    c_bytesize = 1*c_bytelen_IntAuc + &
                 36_IntAl ! 36 spare bytes

    IF (associated(L2C_Mie_Quality_Params%L2C_Mie_Bin_QC)) THEN
       IF (present(M_Mie0)) THEN
          s = M_Mie0
       ELSE
          s = size(L2C_Mie_Quality_Params%L2C_Mie_Bin_QC)
       ENDIF
       c_bytesize = c_bytesize + int(s*GetByteSize_L2C_Mie_Bin_QC(),IntAl)
    END IF

    ! typical sizes
    ! memory size = 1+24*?? = ???? bytes
    ! file   size = 1+24*92 = 2257 bytes
  end function GetByteSize_L2C_Mie_Qual_Params
    !  #]
  subroutine Write_L2C_Mie_Quality_Params(fileunit, &
             L2C_Mie_Quality_Params,error_flag)
    !  #[
    integer,     intent(in)  :: fileunit               ! input
    TYPE(L2C_Mie_Quality_Params_Type), &
                 intent(in)  :: L2C_Mie_Quality_Params ! input
    integer,     intent(out) :: error_flag             ! output

    ! local variables
    integer :: r, s, nr_of_zero_bytes

    !---------------------------------------
    ! writing the L2C_Mie_Quality_Params elements
    !---------------------------------------

    error_flag = no_error
    IF (.not.(associated(L2C_Mie_Quality_Params%L2C_Mie_Bin_QC))) THEN
       print *,"ERROR in Write_L2C_Mie_Quality_Params():"
       print *,"The pointer to this L2C_Mie_Quality_Params seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! Obs_Type
    call write_IntAuc_to_bin_file(fileunit, L2C_Mie_Quality_Params%Obs_Type,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 36 spare bytes                                                            
    nr_of_zero_bytes = 36
    call write_SpareBytes_to_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! get the size of the array
    s = size(L2C_Mie_Quality_Params%L2C_Mie_Bin_QC)
    print *,"Write_L2C_Mie_Quality_Params: size(L2C_Mie_Quality_Params%L2C_Mie_Bin_QC) =",s
    DO r=1,s
       call Write_L2C_Mie_Bin_QC(fileunit, &
            L2C_Mie_Quality_Params%L2C_Mie_Bin_QC(r), &
            error_flag)
       if (error_flag .ne. no_error) goto 998
    END DO

    !print *,"done with Write_L2C_Mie_Quality_Params"
    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Write_L2C_Mie_Quality_Params"
    return

  end subroutine Write_L2C_Mie_Quality_Params
    !  #]
  subroutine Read_L2C_Mie_Quality_Params(fileunit, &
             L2C_Mie_Quality_Params, error_flag)
    !  #[
    integer,                      intent(in)  :: fileunit
    TYPE(L2C_Mie_Quality_Params_Type), intent(inout):: L2C_Mie_Quality_Params
    integer,                      intent(out) :: error_flag

    ! local variables
    integer :: nr_of_zero_bytes
    integer :: r, s

    !---------------------------------------
    ! reading the L2C_Mie_Quality_Params elements
    !---------------------------------------

    error_flag = no_error
    IF (.not.(associated(L2C_Mie_Quality_Params%L2C_Mie_Bin_QC))) THEN
       print *,"ERROR in Read_L2C_Mie_Quality_Params():"
       print *,"The pointer to this L2C_Mie_Quality_Params seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! Obs_Type
    call read_IntAuc_from_bin_file(fileunit, L2C_Mie_Quality_Params%Obs_Type,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 36 spare bytes                                                            
    nr_of_zero_bytes = 36
    call read_SpareBytes_from_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! get the size of the array
    s = size(L2C_Mie_Quality_Params%L2C_Mie_Bin_QC)
    print *,"Read_L2C_Mie_Quality_Params: size(L2C_Mie_Quality_Params%L2C_Mie_Bin_QC) =",s
    DO r=1,s
       call Read_L2C_Mie_Bin_QC(fileunit, &
            L2C_Mie_Quality_Params%L2C_Mie_Bin_QC(r), &
            error_flag)
       if (error_flag .ne. no_error) goto 998
    END DO

    !print *,"done with Read_L2C_Mie_Quality_Params"
    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Read_L2C_Mie_Quality_Params"
    return

  end subroutine Read_L2C_Mie_Quality_Params
    !  #]

  !  #]
  !---------------------------------------------------------------------
  ! subroutines for the Rayleigh_Processing_QC sub-datastructure
  !---------------------------------------------------------------------
  !  #[

  !---------------------------------------------------------------------
  subroutine Init_L2B_Rayleigh_Obs_Screening(L2B_Rayleigh_Obs_Screening)
    !  #[
    TYPE(L2B_Rayleigh_Obs_Screening_Type), intent(out) :: L2B_Rayleigh_Obs_Screening ! output

    L2B_Rayleigh_Obs_Screening%L2B_Rayleigh_Obs_QC       = L2B_Rayleigh_Obs_OK
    L2B_Rayleigh_Obs_Screening%L2B_Rayleigh_Obs_QC_Flags = 0

  end subroutine Init_L2B_Rayleigh_Obs_Screening
    !  #]
  subroutine Delete_L2B_Rayl_Obs_Screening(L2B_Rayleigh_Obs_Screening)
    !  #[
    TYPE(L2B_Rayleigh_Obs_Screening_Type), intent(inout) :: L2B_Rayleigh_Obs_Screening

    ! No pointers so nothing to do

  end subroutine Delete_L2B_Rayl_Obs_Screening
    !  #]
  subroutine AsciiWrite_L2B_Rayl_Obs_Screen(&
                                      L2B_Rayleigh_Obs_Screening,&
                                      fileunit,prefix,&
                                      PrintAll,error_flag)
    !  #[
    TYPE(L2B_Rayleigh_Obs_Screening_type), intent(in)  :: L2B_Rayleigh_Obs_Screening ! input
    integer,                  intent(in)  :: fileunit         ! input
    character(len=*),         intent(in)  :: prefix           ! input
    logical,                  intent(in)  :: PrintAll         ! input
    integer,                  intent(out) :: error_flag       ! output

    ! local variables
    character(len=maxlen_IntAuc)    :: str_IntAuc

    call SetIntAuc(L2B_Rayleigh_Obs_Screening%L2B_Rayleigh_Obs_QC,&
                   str_IntAuc,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L2B_Rayleigh_Obs_QC = "//trim(str_IntAuc)

    call SetIntAuc(L2B_Rayleigh_Obs_Screening%L2B_Rayleigh_Obs_QC_Flags,&
                   str_IntAuc,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L2B_Rayleigh_Obs_QC_Flags = "//trim(str_IntAuc)

    return

999 print *,"ERROR in AsciiWrite_L2B_Rayl_Obs_Screen:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2B_Rayl_Obs_Screen
    !  #]
  function GetByteSize_L2B_Rayl_Obs_Scr() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result

    c_bytesize = 0
    c_bytesize = c_bytesize + c_bytelen_IntAuc ! L2B_Rayleigh_Obs_QC
    c_bytesize = c_bytesize + c_bytelen_IntAuc ! L2B_Rayleigh_Obs_QC_Flags
    c_bytesize = c_bytesize + 16_IntAl         ! Spare_1

  end function GetByteSize_L2B_Rayl_Obs_Scr
    !  #]
  subroutine Write_L2B_Rayl_Obs_Screening(fileunit, &
             L2B_Rayleigh_Obs_Screening, error_flag)
    !  #[
    !  Writes the L2B_Rayleigh_Obs_Screening in binary format
    
    integer,     intent(in)  :: fileunit                   ! input
    TYPE(L2B_Rayleigh_Obs_Screening_Type), &
                 intent(in)  :: L2B_Rayleigh_Obs_Screening ! input
    integer,     intent(out) :: error_flag                 ! output

    ! local variable
    integer :: nr_of_zero_bytes

    !-----------------------------------
    ! writing the L2B_Rayleigh_Obs_Screening elements
    !-----------------------------------

    error_flag = no_error
    
    ! fill with actual values
    
    ! L2B_Rayleigh_Obs_QC
    ! integer(IntAuc) :: L2B_Rayleigh_Obs_QC
    call write_IntAuc_to_bin_file(fileunit, &
         L2B_Rayleigh_Obs_Screening%L2B_Rayleigh_Obs_QC,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! L2B_Rayleigh_Obs_QC_Flags1
    ! integer(IntAuc) :: L2B_Rayleigh_Obs_QC_Flags ! to be defined later
    call write_IntAuc_to_bin_file(fileunit, &
         L2B_Rayleigh_Obs_Screening%L2B_Rayleigh_Obs_QC_Flags, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! 16 spare bytes
    nr_of_zero_bytes = 16
    call write_SpareBytes_to_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 print *,"ABORTING subroutine Write_L2B_Rayl_Obs_Screening"
    return

  end subroutine Write_L2B_Rayl_Obs_Screening
    !  #]
  subroutine Read_L2B_Rayleigh_Obs_Screening(fileunit, &
             L2B_Rayleigh_Obs_Screening, error_flag)
    !  #[
    integer,     intent(in)   :: fileunit
    TYPE(L2B_Rayleigh_Obs_Screening_Type), &
                 intent(inout):: L2B_Rayleigh_Obs_Screening
    integer,     intent(out)  :: error_flag

    ! local variable
    integer :: nr_of_zero_bytes

    !-----------------------------------
    ! reading the L2B_Rayleigh_Obs_Screening elements
    !-----------------------------------

    error_flag = no_error
    
    ! fill with actual values
    
    ! L2B_Rayleigh_Obs_QC
    ! integer(IntAuc) :: L2B_Rayleigh_Obs_QC
    call read_IntAuc_from_bin_file(fileunit, &
         L2B_Rayleigh_Obs_Screening%L2B_Rayleigh_Obs_QC,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! L2B_Rayleigh_Obs_QC_Flags
    ! integer(IntAuc) :: L2B_Rayleigh_Obs_QC_Flags ! to be defined later
    call read_IntAuc_from_bin_file(fileunit, &
         L2B_Rayleigh_Obs_Screening%L2B_Rayleigh_Obs_QC_Flags, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! 16 spare bytes
    nr_of_zero_bytes = 16
    call read_SpareBytes_from_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    return

998 print *,"ABORTING subroutine Read_L2B_Rayleigh_Obs_Screening"
    return

  end subroutine Read_L2B_Rayleigh_Obs_Screening
    !  #]

  !---------------------------------------------------------------------
  subroutine Init_L2C_Rayleigh_Bin_QC(L2C_Rayleigh_Bin_QC)
    !  #[
    TYPE(L2C_Rayleigh_Bin_QC_Type), intent(out) :: L2C_Rayleigh_Bin_QC ! output

    call Init_L2B_Rayleigh_Obs_Screening(L2C_Rayleigh_Bin_QC%L2B_Rayleigh_Obs_Screening)
    call Init_Assimilation_Model_PCD(L2C_Rayleigh_Bin_QC%Assimilation_Model_PCD)

  end subroutine Init_L2C_Rayleigh_Bin_QC
    !  #]
  subroutine Delete_L2C_Rayleigh_Bin_QC(L2C_Rayleigh_Bin_QC)
    !  #[
    TYPE(L2C_Rayleigh_Bin_QC_Type), intent(inout) :: L2C_Rayleigh_Bin_QC

    ! No pointers so nothing to do.

  end subroutine Delete_L2C_Rayleigh_Bin_QC
    !  #]
  subroutine AsciiWrite_L2C_Rayleigh_Bin_QC(&
                                      L2C_Rayleigh_Bin_QC,&
                                      fileunit,prefix,&
                                      PrintAll,error_flag)
    !  #[
    TYPE(L2C_Rayleigh_Bin_QC_type),intent(in)  :: L2C_Rayleigh_Bin_QC   ! input
    integer,                  intent(in)  :: fileunit         ! input
    character(len=*),         intent(in)  :: prefix           ! input
    logical,                  intent(in)  :: PrintAll         ! input
    integer,                  intent(out) :: error_flag       ! output

    ! local variables
    ! none for now

    error_flag = no_error

    call AsciiWrite_L2B_Rayl_Obs_Screen(&
         L2C_Rayleigh_Bin_QC%L2B_Rayleigh_Obs_Screening, fileunit, &
         trim(prefix)//"%L2B_Rayleigh_Obs_Screening",&
         PrintAll,error_flag)
    IF (error_flag .ne. no_error) goto 999

    call AsciiWrite_Assim_Model_PCD(&
         L2C_Rayleigh_Bin_QC%Assimilation_Model_PCD, fileunit, &
         trim(prefix)//"%Assimilation_Model_PCD",&
         PrintAll,error_flag)
    IF (error_flag .ne. no_error) goto 999

    return

999 print *,"ERROR in AsciiWrite_L2C_Rayleigh_Bin_QC:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2C_Rayleigh_Bin_QC
    !  #]
  function GetByteSize_L2C_Rayleigh_Bin_QC() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result

    c_bytesize = 0
    c_bytesize = c_bytesize + GetByteSize_L2B_Rayl_Obs_Scr()
    c_bytesize = c_bytesize + GetByteSize_Assim_Model_PCD()
    c_bytesize = c_bytesize + 10_IntAl ! spare_bytes

    ! typical sizes
    ! memory size = 18+64   =  82 bytes
    ! file   size = 18+64+10 = 92 bytes

  end function GetByteSize_L2C_Rayleigh_Bin_QC
    !  #]
  subroutine Write_L2C_Rayleigh_Bin_QC(fileunit, &
             L2C_Rayleigh_Bin_QC, error_flag)
    !  #[
    integer,                      intent(in)  :: fileunit      ! input
    TYPE(L2C_Rayleigh_Bin_QC_Type), intent(in)  :: L2C_Rayleigh_Bin_QC ! input
    integer,                      intent(out) :: error_flag    ! output

    ! local variables
    integer :: nr_of_zero_bytes

    error_flag = no_error

    call Write_L2B_Rayl_Obs_Screening(&
                       fileunit,&
                       L2C_Rayleigh_Bin_QC%L2B_Rayleigh_Obs_Screening,&
                       error_flag)
    IF (error_flag .ne. no_error) goto 998

    call Write_Assimilation_Model_PCD(&
                       fileunit,&
                       L2C_Rayleigh_Bin_QC%Assimilation_Model_PCD,&
                       error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 10 spare bytes
    nr_of_zero_bytes = 10
    call write_SpareBytes_to_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 print *,"ABORTING subroutine Write_L2C_Rayleigh_Bin_QC"
    return

  end subroutine Write_L2C_Rayleigh_Bin_QC
    !  #]
  subroutine Read_L2C_Rayleigh_Bin_QC(fileunit, &
             L2C_Rayleigh_Bin_QC, error_flag)
    !  #[
    integer,                      intent(in)  :: fileunit
    TYPE(L2C_Rayleigh_Bin_QC_Type), intent(inout):: L2C_Rayleigh_Bin_QC
    integer,                      intent(out) :: error_flag

    ! local variables
    integer :: nr_of_zero_bytes

    error_flag = no_error

    call Read_L2B_Rayleigh_Obs_Screening(fileunit,L2C_Rayleigh_Bin_QC%L2B_Rayleigh_Obs_Screening,&
                                    error_flag)
    IF (error_flag .ne. no_error) goto 998

    call Read_Assimilation_Model_PCD(fileunit,L2C_Rayleigh_Bin_QC%Assimilation_Model_PCD,&
                                     error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 10 spare bytes
    nr_of_zero_bytes = 10
    call read_SpareBytes_from_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    !print *,"done with Read_L2C_Rayleigh_Bin_QC"
    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Read_L2C_Rayleigh_Bin_QC"
    return

  end subroutine Read_L2C_Rayleigh_Bin_QC
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_L2C_Rayl_Quality_Params(L2C_Rayleigh_Quality_Params, &
                                    N_Range_Gates0, error_flag)
    !  #[
    TYPE(L2C_Rayl_Quality_Params_Type), intent(out) :: L2C_Rayleigh_Quality_Params ! output
    integer(IntAs), intent(in)  :: N_Range_Gates0 ! input
    integer,        intent(out) :: error_flag    ! output

    ! local variable
    integer :: r, AllocStatus

    error_flag = no_error

    L2C_Rayleigh_Quality_Params%Obs_Type = Obs_Type_undefined

    IF (N_Range_Gates0 .eq. 0) THEN
       ! init them to be empty
       nullify(L2C_Rayleigh_Quality_Params%L2C_Rayleigh_Bin_QC)
    ELSE
       allocate(L2C_Rayleigh_Quality_Params%L2C_Rayleigh_Bin_QC(N_Range_Gates0),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L2C_Rayl_Quality_Params"
          error_flag = error_allocate
          return
       END IF
       
       DO r=1,N_Range_Gates0
          call Init_L2C_Rayleigh_Bin_QC(L2C_Rayleigh_Quality_Params%L2C_Rayleigh_Bin_QC(r))
       END DO
    END IF

  end subroutine Init_L2C_Rayl_Quality_Params
    !  #]
  subroutine Delete_L2C_Rayl_Quality_Params(L2C_Rayleigh_Quality_Params)
    !  #[
    TYPE(L2C_Rayl_Quality_Params_Type), intent(inout) :: L2C_Rayleigh_Quality_Params

    ! local variable
    integer :: i

    IF (associated(L2C_Rayleigh_Quality_Params%L2C_Rayleigh_Bin_QC)) THEN
       DO i=1,size(L2C_Rayleigh_Quality_Params%L2C_Rayleigh_Bin_QC)
          call Delete_L2C_Rayleigh_Bin_QC(&
               L2C_Rayleigh_Quality_Params%L2C_Rayleigh_Bin_QC(i))
       END DO
       deallocate(L2C_Rayleigh_Quality_Params%L2C_Rayleigh_Bin_QC)
    END IF

  end subroutine Delete_L2C_Rayl_Quality_Params
    !  #]
  subroutine AsciiWrite_L2C_Rayl_Qual_Params(&
                                      L2C_Rayleigh_Quality_Params,&
                                      fileunit,prefix,&
                                      PrintAll,error_flag)
    !  #[
    TYPE(L2C_Rayl_Quality_Params_Type), intent(in)  :: L2C_Rayleigh_Quality_Params ! input
    integer,                  intent(in)  :: fileunit         ! input
    character(len=*),         intent(in)  :: prefix           ! input
    logical,                  intent(in)  :: PrintAll         ! input
    integer,                  intent(out) :: error_flag       ! output

    ! local variables
    integer                        :: i,s,step
    character(len=maxlen_IntAuc)   :: str_IntAuc
    character(len=2)               :: str_cntr

    call SetIntAuc(L2C_Rayleigh_Quality_Params%Obs_Type,str_IntAuc,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Obs_Type = "//trim(str_IntAuc)//" ("//&
         trim(GetObsTypeText(int(L2C_Rayleigh_Quality_Params%Obs_Type,IntAuc)))//")"

    IF (associated(L2C_Rayleigh_Quality_Params%L2C_Rayleigh_Bin_QC)) THEN
       s=size(L2C_Rayleigh_Quality_Params%L2C_Rayleigh_Bin_QC)
       IF (PrintAll .or. (s .le. 1)) THEN
          step=1
       ELSE
          step = s - 1
       END IF
       DO i=1,s,step
          write(str_cntr,"(i2.2)",err=998) i
          call AsciiWrite_L2C_Rayleigh_Bin_QC(&
               L2C_Rayleigh_Quality_Params%L2C_Rayleigh_Bin_QC(i), fileunit, &
               trim(prefix)//"%L2C_Rayleigh_Bin_QC("//trim(str_cntr)//")",&
               PrintAll,error_flag)
          IF (error_flag .ne. no_error) goto 999
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%L2C_Rayleigh_Bin_QC(:) [seems not associated]"
    END IF

    return

998 print *,"conversion error in AsciiWrite_L2C_Rayl_Qual_Params:"
    print *,"could not convert the counter ",i," to a string .."
    error_flag = error_writing_file
    return

999 print *,"ERROR in AsciiWrite_L2C_Rayl_Qual_Params:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2C_Rayl_Qual_Params
    !  #]
  function GetByteSize_L2C_Ray_Qual_Params(L2C_Rayleigh_Quality_Params, &
              M_Rayleigh0) result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result
    TYPE(L2C_Rayl_Quality_Params_Type), intent(in) :: L2C_Rayleigh_Quality_Params
    integer,     optional,        intent(in) :: M_Rayleigh0

    ! local variables
    integer :: s

    c_bytesize = 1*c_bytelen_IntAuc + &
                 36_IntAl ! 36 spare bytes

    IF (associated(L2C_Rayleigh_Quality_Params%L2C_Rayleigh_Bin_QC)) THEN
       IF (present(M_Rayleigh0)) THEN
          s = M_Rayleigh0
       ELSE
          s = size(L2C_Rayleigh_Quality_Params%L2C_Rayleigh_Bin_QC)
       ENDIF
       c_bytesize = c_bytesize + int(s*GetByteSize_L2C_Rayleigh_Bin_QC(),IntAl)
    END IF

    ! typical sizes
    ! memory size = 1+24*?? = ???? bytes
    ! file   size = 1+24*92 = 2257 bytes
  end function GetByteSize_L2C_Ray_Qual_Params
    !  #]
  subroutine Write_L2C_Rayl_Quality_Params(fileunit, &
             L2C_Rayleigh_Quality_Params, error_flag)
    !  #[
    integer,     intent(in)  :: fileunit                    ! input
    TYPE(L2C_Rayl_Quality_Params_Type), &
                 intent(in)  :: L2C_Rayleigh_Quality_Params ! input
    integer,     intent(out) :: error_flag                  ! output

    ! local variables
    integer :: r, s, nr_of_zero_bytes

    !---------------------------------------
    ! writing the L2C_Rayleigh_Quality_Params elements
    !---------------------------------------

    error_flag = no_error
    IF (.not.(associated(L2C_Rayleigh_Quality_Params%L2C_Rayleigh_Bin_QC))) THEN
       print *,"ERROR in Write_L2C_Rayl_Quality_Params():"
       print *,"The pointer to this L2C_Rayleigh_Quality_Params seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! Obs_Type
    call write_IntAuc_to_bin_file(fileunit, L2C_Rayleigh_Quality_Params%Obs_Type,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 36 spare bytes                                                            
    nr_of_zero_bytes = 36
    call write_SpareBytes_to_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! get the size of the array
    s = size(L2C_Rayleigh_Quality_Params%L2C_Rayleigh_Bin_QC)
    print *,"Write_L2C_Rayl_Quality_Params: size(L2C_Rayleigh_Quality_Params%L2C_Rayleigh_Bin_QC) =",s
    DO r=1,s
       call Write_L2C_Rayleigh_Bin_QC(fileunit, &
            L2C_Rayleigh_Quality_Params%L2C_Rayleigh_Bin_QC(r), &
            error_flag)
       if (error_flag .ne. no_error) goto 998
    END DO

    !print *,"done with Write_L2C_Rayl_Quality_Params"
    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Write_L2C_Rayl_Quality_Params"
    return

  end subroutine Write_L2C_Rayl_Quality_Params
    !  #]
  subroutine Read_L2C_Rayl_Quality_Params(fileunit, &
             L2C_Rayleigh_Quality_Params, error_flag)
    !  #[
    integer,                      intent(in)  :: fileunit
    TYPE(L2C_Rayl_Quality_Params_Type), intent(inout):: L2C_Rayleigh_Quality_Params
    integer,                      intent(out) :: error_flag

    ! local variables
    integer :: nr_of_zero_bytes
    integer :: r, s

    !---------------------------------------
    ! reading the L2C_Rayleigh_Quality_Params elements
    !---------------------------------------

    error_flag = no_error
    IF (.not.(associated(L2C_Rayleigh_Quality_Params%L2C_Rayleigh_Bin_QC))) THEN
       print *,"ERROR in Read_L2C_Rayl_Quality_Params():"
       print *,"The pointer to this L2C_Rayleigh_Quality_Params seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! Obs_Type
    call read_IntAuc_from_bin_file(fileunit, L2C_Rayleigh_Quality_Params%Obs_Type,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 36 spare bytes                                                            
    nr_of_zero_bytes = 36
    call read_SpareBytes_from_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! get the size of the array
    s = size(L2C_Rayleigh_Quality_Params%L2C_Rayleigh_Bin_QC)
    print *,"Read_L2C_Rayleigh_Quality_Params: size(L2C_Rayleigh_Quality_Params%L2C_Rayleigh_Bin_QC) =",s
    DO r=1,s
       call Read_L2C_Rayleigh_Bin_QC(fileunit, &
            L2C_Rayleigh_Quality_Params%L2C_Rayleigh_Bin_QC(r), &
            error_flag)
       if (error_flag .ne. no_error) goto 998
    END DO

    !print *,"done with Read_L2C_Rayl_Quality_Params"
    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Read_L2C_Rayl_Quality_Params"
    return

  end subroutine Read_L2C_Rayl_Quality_Params
    !  #]

  !  #]
  !---------------------------------------------------------------------
  ! subroutines for the L2C_AssimPCD_ADS datastructure    
  !---------------------------------------------------------------------
  !  #[
  subroutine Init_L2C_AssimPCD_ADS(L2C_AssimPCD_ADS, N_Meas0, &
             N_Obs_Mie_Max0, N_Obs_Rayleigh_Max0, N_Range_Gates0, error_flag)
    !  #[
    TYPE(L2C_AssimPCD_ADS_type), pointer:: L2C_AssimPCD_ADS  ! input/output
    integer(IntAs),         intent(in)  :: N_Meas0        ! input
    integer(IntAs),         intent(in)  :: N_Obs_Mie_Max0     ! input
    integer(IntAs),         intent(in)  :: N_Obs_Rayleigh_Max0     ! input
    integer(IntAs),         intent(in)  :: N_Range_Gates0 ! input
    integer,                intent(out) :: error_flag    ! output

    ! local variables
    integer :: o, AllocStatus

    error_flag = no_error

    IF (.not. associated(L2C_AssimPCD_ADS)) THEN
       print *,"ERROR in Init_L2C_AssimPCD_ADS: "
       print *,"pointer L2C_AssimPCD_ADS seems not associated !!!"
       error_flag = error_programming
       return
    END IF

    L2C_AssimPCD_ADS%N_Meas                = N_Meas0

    ! these are filled during processing, so init them to zero
    L2C_AssimPCD_ADS%N_Obs_Mie_Actual      = 0
    L2C_AssimPCD_ADS%N_Obs_Rayleigh_Actual = 0

    call Init_L2B_Input_Screening(L2C_AssimPCD_ADS%L2B_Input_Screening,&
                                  error_flag)
    IF (error_flag .ne. no_error) return

    IF (N_Obs_Mie_Max0 .eq. 0) THEN
       ! init them to be empty
       nullify(L2C_AssimPCD_ADS%L2C_Mie_Quality_Params)
    ELSE
       allocate(L2C_AssimPCD_ADS%L2C_Mie_Quality_Params(N_Obs_Mie_Max0),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure for L2C_Mie_Quality_Params in subroutine Init_L2C_AssimPCD_ADS()"
          error_flag = error_allocate
          return
       END IF
       DO o=1,N_Obs_Mie_Max0
          call Init_L2C_Mie_Quality_Params(L2C_AssimPCD_ADS%L2C_Mie_Quality_Params(o),&
                               N_Range_Gates0,error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    END IF

    IF (N_Obs_Rayleigh_Max0 .eq. 0) THEN
       ! init them to be empty
       nullify(L2C_AssimPCD_ADS%L2C_Rayleigh_Quality_Params)
    ELSE
       allocate(L2C_AssimPCD_ADS%L2C_Rayleigh_Quality_Params(N_Obs_Rayleigh_Max0),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure for L2C_Rayleigh_Quality_Params in subroutine Init_L2C_AssimPCD_ADS()"
          error_flag = error_allocate
          return
       END IF
       DO o=1,N_Obs_Rayleigh_Max0
          call Init_L2C_Rayl_Quality_Params(L2C_AssimPCD_ADS%L2C_Rayleigh_Quality_Params(o),&
                               N_Range_Gates0,error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    END IF

  end subroutine Init_L2C_AssimPCD_ADS
    !  #]
  subroutine Delete_L2C_AssimPCD_ADS(L2C_AssimPCD_ADS)
    !  #[
    TYPE(L2C_AssimPCD_ADS_type), pointer :: L2C_AssimPCD_ADS

    ! local variables
    integer :: i

    IF (.not. associated(L2C_AssimPCD_ADS)) return

    call Delete_L2B_Input_Screening(L2C_AssimPCD_ADS%L2B_Input_Screening)

    IF (associated(L2C_AssimPCD_ADS%L2C_Mie_Quality_Params)) THEN
       DO i=1,size(L2C_AssimPCD_ADS%L2C_Mie_Quality_Params)
          call Delete_L2C_Mie_Quality_Params(&
               L2C_AssimPCD_ADS%L2C_Mie_Quality_Params(i))
       END DO
       deallocate(L2C_AssimPCD_ADS%L2C_Mie_Quality_Params)
    END IF

    IF (associated(L2C_AssimPCD_ADS%L2C_Rayleigh_Quality_Params)) THEN
       DO i=1,size(L2C_AssimPCD_ADS%L2C_Rayleigh_Quality_Params)
          call Delete_L2C_Rayl_Quality_Params(&
               L2C_AssimPCD_ADS%L2C_Rayleigh_Quality_Params(i))
       END DO
       deallocate(L2C_AssimPCD_ADS%L2C_Rayleigh_Quality_Params)
    END IF

  end subroutine Delete_L2C_AssimPCD_ADS
    !  #]
  subroutine Print_L2C_AssimPCD_ADS(L2C_AssimPCD_ADS,prefix,&
                                    PrintAll,error_flag)
    !  #[
    TYPE(L2C_AssimPCD_ADS_Type), pointer:: L2C_AssimPCD_ADS
    character(len=*),       intent(in)  :: prefix           ! input
    logical,                intent(in)  :: PrintAll         ! input
    integer,                intent(out) :: error_flag       ! output

    call AsciiWrite_L2C_AssimPCD_ADS(L2C_AssimPCD_ADS,fileunit_stdout,prefix,&
                                       PrintAll,error_flag)

  end subroutine Print_L2C_AssimPCD_ADS
    !  #]
  subroutine AsciiWrite_L2C_AssimPCD_ADS(L2C_AssimPCD_ADS,fileunit,prefix,&
                                         PrintAll,error_flag)
    !  #[
    TYPE(L2C_AssimPCD_ADS_type), pointer:: L2C_AssimPCD_ADS
    integer,                intent(in)  :: fileunit         ! input
    character(len=*),       intent(in)  :: prefix           ! input
    logical,                intent(in)  :: PrintAll         ! input
    integer,                 intent(out) :: error_flag       ! output

    ! local variables
    integer :: i, s, step
    character(len=maxlen_IntAs)    :: str_IntAs
    character(len=2)               :: str_cntr

    error_flag = no_error

    IF (.not. associated(L2C_AssimPCD_ADS)) THEN
       write(fileunit,"(a)",err=999) trim(prefix)//&
            " [seems not associated]"
       error_flag = error_programming
       return
    END IF

    call SetIntAs_or_Missing(L2C_AssimPCD_ADS%N_Meas,&
                             str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%N_Meas = "//trim(str_IntAs)

    call SetIntAs_or_Missing(L2C_AssimPCD_ADS%N_Obs_Mie_Actual,&
                             str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%N_Obs_Mie_Actual = "//trim(str_IntAs)

    call SetIntAs_or_Missing(L2C_AssimPCD_ADS%N_Obs_Rayleigh_Actual,&
                             str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%N_Obs_Rayleigh_Actual = "//trim(str_IntAs)

    call AsciiWrite_L2B_Input_Screening(&
         L2C_AssimPCD_ADS%L2B_Input_Screening, fileunit, &
         trim(prefix)//"%L2B_Input_Screening",&
         PrintAll,error_flag)
    IF (error_flag .ne. no_error) goto 999

    IF (associated(L2C_AssimPCD_ADS%L2C_Mie_Quality_Params)) THEN
       s=size(L2C_AssimPCD_ADS%L2C_Mie_Quality_Params)
       IF (PrintAll .or. (s .le. 1)) THEN
          step=1
       ELSE
          step = s - 1
       END IF
       DO i=1,s,step
          write(str_cntr,"(i2.2)",err=998) i
          call AsciiWrite_L2C_Mie_Qual_Params(&
               L2C_AssimPCD_ADS%L2C_Mie_Quality_Params(i), fileunit, &
               trim(prefix)//"%L2C_Mie_Quality_Params("//trim(str_cntr)//")",&
               PrintAll,error_flag)
          IF (error_flag .ne. no_error) goto 999
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%L2C_Mie_Quality_Params(:) [seems not associated]"
    END IF

    IF (associated(L2C_AssimPCD_ADS%L2C_Rayleigh_Quality_Params)) THEN
       s=size(L2C_AssimPCD_ADS%L2C_Rayleigh_Quality_Params)
       IF (PrintAll .or. (s .le. 1)) THEN
          step=1
       ELSE
          step = s - 1
       END IF
       DO i=1,s,step
          write(str_cntr,"(i2.2)",err=998) i
          call AsciiWrite_L2C_Rayl_Qual_Params(&
               L2C_AssimPCD_ADS%L2C_Rayleigh_Quality_Params(i), fileunit, &
               trim(prefix)//"%L2C_Rayleigh_Quality_Params("//trim(str_cntr)//")",&
               PrintAll,error_flag)
          IF (error_flag .ne. no_error) goto 999
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%L2C_Rayleigh_Quality_Params(:) [seems not associated]"
    END IF

    return

998 print *,"conversion error in AsciiWrite_L2C_AssimPCD_ADS:"
    print *,"could not convert the counter ",i," to a string .."
    error_flag = error_writing_file
    return

999 print *,"ERROR in AsciiWrite_L2C_AssimPCD_DS:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2C_AssimPCD_ADS
    !  #]
  function GetByteSize_L2C_AssimPCD_ADS(L2C_AssimPCD_ADS,&
                                        M_Mie0,M_Rayleigh0) result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result
    TYPE(L2C_AssimPCD_ADS_type), pointer    :: L2C_AssimPCD_ADS
    integer(IntAuc), optional,   intent(in) :: M_Mie0
    integer(IntAuc), optional,   intent(in) :: M_Rayleigh0

    ! local variables
    integer :: s

    c_bytesize = 0
    IF (.not. associated(L2C_AssimPCD_ADS)) return
       
    c_bytesize = c_bytesize + c_bytelen_IntAs ! N_Meas

    c_bytesize = c_bytesize + c_bytelen_IntAs ! N_Obs_Mie_Actual

    c_bytesize = c_bytesize + c_bytelen_IntAs ! N_Obs_Rayleigh_Actual

    c_bytesize = c_bytesize + GetByteSize_L2B_Input_Screening()

    IF (associated(L2C_AssimPCD_ADS%L2C_Mie_Quality_Params)) THEN
       IF (present(M_Mie0)) THEN
          s = M_Mie0
       ELSE
          s = size(L2C_AssimPCD_ADS%L2C_Mie_Quality_Params)
       ENDIF
       c_bytesize = c_bytesize + int(s,IntAl) *&
            GetByteSize_L2C_Mie_Qual_Params(&
                        L2C_AssimPCD_ADS%L2C_Mie_Quality_Params(1))
    END IF

    IF (associated(L2C_AssimPCD_ADS%L2C_Rayleigh_Quality_Params)) THEN
       IF (present(M_Rayleigh0)) THEN
          s = M_Rayleigh0
       ELSE
          s = size(L2C_AssimPCD_ADS%L2C_Rayleigh_Quality_Params)
       ENDIF
       c_bytesize = c_bytesize + int(s,IntAl) *&
            GetByteSize_L2C_Ray_Qual_Params(&
                        L2C_AssimPCD_ADS%L2C_Rayleigh_Quality_Params(1))
    END IF

    c_bytesize = c_bytesize + 20_IntAl ! spare_bytes

    ! typical sizes
    ! memory size = ?+?+?+??+?????+?????+ 0 = ????? bytes
    ! file   size = 2+2+2+26+11285+11285+20 = 22622 bytes
  end function GetByteSize_L2C_AssimPCD_ADS
    !  #]
  subroutine Write_L2C_AssimPCD_ADS(fileunit,L2C_AssimPCD_ADS,&
                         N_Obs_Mie_to_write, N_Obs_Rayleigh_to_write, &
                         error_flag)
    !  #[
    !   Jos de Kloe    January 2006   Original code
    !
    ! History:
    !   P. Poli    10-01-2006   Replaced placeholders by actual writing
    !   P. Poli    30-01-2006   Added N_Meas, N_Obs_*_Actual

    integer,                 intent(in)  :: fileunit                ! input
    TYPE(L2C_AssimPCD_ADS_type), pointer :: L2C_AssimPCD_ADS        ! input
    integer(IntAuc),         intent(in)  :: N_Obs_Mie_to_write      ! input
    integer(IntAuc),         intent(in)  :: N_Obs_Rayleigh_to_write ! input
    integer,                 intent(out) :: error_flag     ! output

    ! local variables
    integer :: nr_of_zero_bytes, o

    error_flag = no_error
    IF (.not. associated(L2C_AssimPCD_ADS)) THEN
       print *,"ERROR in Write_L2C_AssimPCD_ADS():"
       print *,"The pointer to this L2C_AssimPCD_ADS seems not associated !!"
       error_flag = error_programming
       return
    END IF

    call write_IntAs_to_bin_file(fileunit, &
         L2C_AssimPCD_ADS%N_Meas,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call write_IntAs_to_bin_file(fileunit, &
         L2C_AssimPCD_ADS%N_Obs_Mie_Actual,error_flag)
    IF (error_flag .ne. no_error) goto 998
       
    call write_IntAs_to_bin_file(fileunit, &
         L2C_AssimPCD_ADS%N_Obs_Rayleigh_Actual,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! L2B_Input_Screening
    call Write_L2B_Inp_Screening(fileunit, &
         L2C_AssimPCD_ADS%L2B_Input_Screening, &
         error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! L2C_Mie_Quality_Params
    DO o=1,N_Obs_Mie_to_write
       call Write_L2C_Mie_Quality_Params(fileunit, &
            L2C_AssimPCD_ADS%L2C_Mie_Quality_Params(o), error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    ! L2C_Rayleigh_Quality_Params
    DO o=1,N_Obs_Rayleigh_to_write
       call Write_L2C_Rayl_Quality_Params(fileunit, &
            L2C_AssimPCD_ADS%L2C_Rayleigh_Quality_Params(o), &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    ! 20 spare bytes
    nr_of_zero_bytes = 20
    call write_SpareBytes_to_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 print *,"ABORTING subroutine Write_L2C_AssimPCD_ADS"
    return

  end subroutine Write_L2C_AssimPCD_ADS
    !  #]
  subroutine Read_L2C_AssimPCD_ADS(fileunit,L2C_AssimPCD_ADS,error_flag)
    !  #[
    integer,                intent(in)  :: fileunit
    TYPE(L2C_AssimPCD_ADS_type), pointer:: L2C_AssimPCD_ADS
    integer,                intent(out) :: error_flag

    ! local variables
    integer :: nr_of_zero_bytes, o

    error_flag = no_error
    IF (.not. associated(L2C_AssimPCD_ADS)) THEN
       print *,"ERROR in Read_L2C_AssimPCD_ADS():"
       print *,"The pointer to this L2C_AssimPCD_ADS seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! Reading N_Meas, N_Obs_*_Actual
    call read_IntAs_from_bin_file(fileunit, &
         L2C_AssimPCD_ADS%N_Meas,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call read_IntAs_from_bin_file(fileunit, &
         L2C_AssimPCD_ADS%N_Obs_Mie_Actual,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call read_IntAs_from_bin_file(fileunit, &
         L2C_AssimPCD_ADS%N_Obs_Rayleigh_Actual,error_flag)
    IF (error_flag .ne. no_error) goto 998
  
    ! L2B_Input_Screening
    call Read_L2B_Inp_Screening(fileunit, &
         L2C_AssimPCD_ADS%L2B_Input_Screening, &
         error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! for N_Obs=0 this pointer may be un-associated
    ! so we have to check for that before using its size
    IF (associated(L2C_AssimPCD_ADS%L2C_Mie_Quality_Params)) THEN
       ! L2C_Mie_Quality_Params
       DO o=1,size(L2C_AssimPCD_ADS%L2C_Mie_Quality_Params)
          call Read_L2C_Mie_Quality_Params(fileunit, &
               L2C_AssimPCD_ADS%L2C_Mie_Quality_Params(o), &
               error_flag)
          IF (error_flag .ne. no_error) goto 998
       END DO
    END IF

    ! for N_Obs=0 this pointer may be un-associated
    ! so we have to check for that before using its size
    IF (associated(L2C_AssimPCD_ADS%L2C_Rayleigh_Quality_Params)) THEN
       ! L2C_Rayleigh_Quality_Params
       DO o=1,size(L2C_AssimPCD_ADS%L2C_Rayleigh_Quality_Params)
          call Read_L2C_Rayl_Quality_Params(fileunit, &
               L2C_AssimPCD_ADS%L2C_Rayleigh_Quality_Params(o), &
               error_flag)
          IF (error_flag .ne. no_error) goto 998
       END DO
    END IF

    ! 20 spare bytes
    nr_of_zero_bytes = 20
    call read_SpareBytes_from_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 print *,"ABORTING subroutine Read_L2C_AssimPCD_ADS"
    return

  end subroutine Read_L2C_AssimPCD_ADS
    !  #]
  !  #]
  !---------------------------------------------------------------------
end module L2C_AssimPCD_ADS
