! todo: remove all error prints, replace by logmsg()

module L2C_MieVec_MDS
  !  #[ Documentation
  !
  !     a module to handle the L2C Mie VecWind Measurement dataset
  !     as defined in L2B/L2C-IODD, version 1.1 (draft, 16-Jun-2006).
  !
  !     Written by David Tan, ECMWF
  !     Based on l2b_mie_mds.F90 by  J. de Kloe, KNMI 
  !                              and P. Poli,    Meteo-France
  !     Modifications:
  !     04-Jul-2006 David Tan   Original version
  !     02-Mar-2007 J. de Kloe  adapted to parameter type change in call to 
  !                             GetObsTypeText()
  !     05-Mar-2007 J. de Kloe  adapted to IODD change, 
  !                             and added getBytesize functions
  !     15-Mar-2007 J. de Kloe  removed use of SummaryVars
  !     19-Mar-2007 J. de Kloe  added AsciiWrite routines
  !     23-Apr-2007 J. de Kloe  phase out fillwithzero code
  !     09-Jan-2007 J. de Kloe  added check for missing_DateTime_KVT
  !     16-Jan-2008 J. de Kloe  phase out integer kind i_ 
  !
  !  #]
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE EE_CFI_Datatypes, only: &
       IntAc,  IntAuc, IntAl, &
       IntAs,  maxlen_IntAs,  SetIntAs,  &
       missing_indicator_IntAs, &
       DateTimeType,Construct_KVT_DateTime, &
       maxlen_DateTime_KVT, &
       read_IntAs_from_bin_file, &
       write_IntAs_to_bin_file, &
       read_IntAc_from_bin_file, &
       write_IntAc_to_bin_file, &
       InitDateTime, DateTimeMissing, &
       write_DateTime_to_bin_file, &
       read_DateTime_from_bin_file, &
       c_bytelen_Bool, c_bytelen_IntAs, &
       c_bytelen_IntAc, c_bytelen_datetime, &
       write_SpareBytes_to_bin_file, &
       read_SpareBytes_from_bin_file

  USE L2B_parameters, only: Obs_Type_undefined, GetObsTypeText
  USE ErrorHandler, only: no_error, error_programming, error_allocate, &
                          error_writing_file
  USE DirectBinaryIO, only: write_logical1_to_bin_file, &
                            read_logical1_from_bin_file
  !  #]
  !  #[ type definitions

  implicit none

  TYPE L2C_Height_Bin_VecWind_type
     !  #[
     ! see sec. 4.3.6, L2B/L2C-IODD, vers. 1.1, (draft, 16-Jun-2006)
     logical         :: Validity_Flag                   ! .true. = valid = 1
     integer(IntAs)  :: Background_Zonal_WindVeloc      ! [cm/s]
     integer(IntAs)  :: Background_Meridional_WindVeloc ! [cm/s]
     integer(IntAs)  :: Analysis_Zonal_WindVeloc        ! [cm/s]
     integer(IntAs)  :: Analysis_Meridional_WindVeloc   ! [cm/s]
     ! Spare_1 (20 bytes)
!                       1234567890123456789012345678901
  END TYPE L2C_Height_Bin_VecWind_type
     !  #]

  TYPE MieVec_Profile_type
     !  #[
     !
     ! An integer (enumerated type) to remember if this was
     ! a cloud or no-cloud profile (or any other type we define)
     integer(IntAc) ::Obs_Type
     !
     ! Spare 1 (36 bytes reserved for Obs_Type_String)
     !
     ! see sec. 4.3.6, L2B/L2C-IODD, vers. 1.1, (draft, 16-Jun-2006)
     ! this profile is an array of size Num_Rangegates
     TYPE(L2C_Height_Bin_VecWind_type), dimension(:), pointer :: Height_Bin_VecWind
  END TYPE MieVec_Profile_type
     !  #]

  TYPE L2C_MieVec_MDS_type 
     !  #[
     ! see sec. 4.3.6, L2B/L2C-IODD, vers. 1.1 (draft, 16-Jun-2006)
     type(DateTimeType)  :: Start_of_Obs_DateTime    ! UTC
     integer(IntAs)      :: N_Meas       ! nr of Measurements in this product
     integer(IntAs)      :: N_Obs_Mie_Actual ! actual nr of Obs. for this BRC
! Poli: following telecon of 25-01-2006, it is decided that N_Obs_*_Actual
! refers to the number of Mie and Rayleigh profiles with meaningful values.
! Each BRC contains M_Mie and M_Rayleigh profiles (these two numbers are fixed
! for a given file), but in some BRCs there might be less than M_Mie profiles;
! in such case N_Obs_Mie_Actual is less than M_Mie. Same story for Rayleigh.
!   - when initialising, N_Obs_Mie_Actual is set to zero. The structure
!     is dimensioned using M_Mie.
!   - after each BRC is processed, a call to Handle_L2BHeaders will set
!     N_Obs_Mie_Actual to the proper value
!   - when writing, M_Mie profiles are written. The number N_Obs_Mie_Actual
!     is saved so that the user will know when reading the BRC that
!     only the first N_Obs_Mie_Actual profiles are meaningful.
!
!   - when reading, the structure should be dimensioned first to M_Mie.
!
!     ! Remark: the nr of generated observations may vary from one BRC to 
!     ! the next but it was agreed that for one product file the maximum
!     ! N_Obs_max should be used for dimensioning the datastructures 
!     ! that are written to the output file:
!     !         N_Obs_max = max(N_Obs_Actual(first_BRC:last_BRC))
!     ! (see email of David Tan, 20051012, 19.44 utc)
!     ! So: -when initialising N_Obs_Max is used, N_Obs_Actual is set to 0
!     !     -when filling the datastructure N_Obs_Actual is updated
!     !      everytime an observation is generated
!     !     -when printing only N_Obs_Actual is used
!     !     -when writing to file N_Obs_Actual is used, 
!     !      and also all not used Observations are written to file
!
     !
     ! an array for N_Obs_Max profiles for this BRC
     TYPE(MieVec_Profile_type), dimension(:), &
          pointer :: MieVec_Profile
     !              1234567890123456789012345678901
  END TYPE L2C_MieVec_MDS_type
     !  #]

  !  #]
contains
  !---------------------------------------------------------------------
  subroutine Init_L2C_Height_Bin_VecWind(L2C_Height_Bin_VecWind)
    !  #[
    TYPE(L2C_Height_Bin_VecWind_type), intent(out) :: L2C_Height_Bin_VecWind

    L2C_Height_Bin_VecWind%Validity_Flag                   = .false.
    L2C_Height_Bin_VecWind%Background_Zonal_WindVeloc      = missing_indicator_IntAs
    L2C_Height_Bin_VecWind%Background_Meridional_WindVeloc = missing_indicator_IntAs
    L2C_Height_Bin_VecWind%Analysis_Zonal_WindVeloc        = missing_indicator_IntAs
    L2C_Height_Bin_VecWind%Analysis_Meridional_WindVeloc   = missing_indicator_IntAs

  end subroutine Init_L2C_Height_Bin_VecWind
    !  #]
  subroutine Print_L2C_Height_Bin_VecWind(L2C_Height_Bin_VecWind, error_flag)
    !  #[
    TYPE(L2C_Height_Bin_VecWind_type), intent(in)  :: L2C_Height_Bin_VecWind
    integer,                           intent(out) :: error_flag    ! output

    ! local variables
    character(len=maxlen_IntAs)  :: str_IntAs

    error_flag = no_error

    if (L2C_Height_Bin_VecWind%Validity_Flag) then
       print *,"    ...%Validity_Flag                   = Valid"
    else
       print *,"    ...%Validity_Flag                   = Not Valid"
    endif

    call SetIntAs(L2C_Height_Bin_VecWind%Background_Zonal_WindVeloc,&
                  str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%Background_Zonal_WindVeloc      = ",str_IntAs

    call SetIntAs(L2C_Height_Bin_VecWind%Background_Meridional_WindVeloc,&
                  str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%Background_Meridional_WindVeloc = ",str_IntAs

    call SetIntAs(L2C_Height_Bin_VecWind%Analysis_Zonal_WindVeloc,&
                  str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%Analysis_Zonal_WindVeloc        = ",str_IntAs

    call SetIntAs(L2C_Height_Bin_VecWind%Analysis_Meridional_WindVeloc,&
                  str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    print *,"    ...%Analysis_Meridional_WindVeloc   = ",str_IntAs

  end subroutine Print_L2C_Height_Bin_VecWind
    !  #]
  subroutine AsciiWrite_L2C_HghtBin_VecWind(L2C_Height_Bin_VecWind, &
                                            fileunit, prefix, error_flag)
    !  #[
    TYPE(L2C_Height_Bin_VecWind_type), &
                      intent(in)  :: L2C_Height_Bin_VecWind ! input
    integer,          intent(in)  :: fileunit               ! input
    character(len=*), intent(in)  :: prefix                 ! input
    integer,          intent(out) :: error_flag             ! output

    ! local variables
    character(len=maxlen_IntAs)  :: str_IntAs
    character(len=10)            :: str_Valid

    error_flag = no_error

    if (L2C_Height_Bin_VecWind%Validity_Flag) then
       str_Valid = "Valid"
    else
       str_Valid = "Not Valid"
    endif
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Validity_Flag                   = "//trim(str_Valid)

    call SetIntAs(L2C_Height_Bin_VecWind%Background_Zonal_WindVeloc,&
                  str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Background_Zonal_WindVeloc      = "//&
         trim(str_IntAs)//" [cm/s]"

    call SetIntAs(L2C_Height_Bin_VecWind%Background_Meridional_WindVeloc,&
                  str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Background_Meridional_WindVeloc = "//&
         trim(str_IntAs)//" [cm/s]"

    call SetIntAs(L2C_Height_Bin_VecWind%Analysis_Zonal_WindVeloc,&
                  str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Analysis_Zonal_WindVeloc        = "//&
         trim(str_IntAs)//" [cm/s]"

    call SetIntAs(L2C_Height_Bin_VecWind%Analysis_Meridional_WindVeloc,&
                  str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Analysis_Meridional_WindVeloc   = "//&
         trim(str_IntAs)//" [cm/s]"

    return

999 print *,"ERROR in AsciiWrite_L2C_HghtBin_VecWind:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2C_HghtBin_VecWind
    !  #]
  subroutine Write_L2C_Height_Bin_VecWind(fileunit, &
             L2C_Height_Bin_VecWind, error_flag)
    !  #[
    integer,                           intent(in)  :: fileunit
    TYPE(L2C_Height_Bin_VecWind_type), intent(in)  :: L2C_Height_Bin_VecWind
    integer,                           intent(out) :: error_flag

    ! local variables
    integer :: nr_of_zero_bytes

    error_flag = no_error

    call write_logical1_to_bin_file(fileunit, &
         L2C_Height_Bin_VecWind%Validity_Flag, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAs_to_bin_file(fileunit, &
         L2C_Height_Bin_VecWind%Background_Zonal_WindVeloc, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAs_to_bin_file(fileunit, &
         L2C_Height_Bin_VecWind%Background_Meridional_WindVeloc, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAs_to_bin_file(fileunit, &
         L2C_Height_Bin_VecWind%Analysis_Zonal_WindVeloc, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAs_to_bin_file(fileunit, &
         L2C_Height_Bin_VecWind%Analysis_Meridional_WindVeloc, error_flag)
    IF (error_flag .ne. no_error) goto 998
    nr_of_zero_bytes = 20 ! Spare_1
    call write_SpareBytes_to_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Write_L2C_Height_Bin_VecWind"
    return

    end subroutine Write_L2C_Height_Bin_VecWind
    !  #]
  subroutine Read_L2C_Height_Bin_VecWind(fileunit, &
             L2C_Height_Bin_VecWind, error_flag)
    !  #[

    integer,                           intent(in)  :: fileunit
    TYPE(L2C_Height_Bin_VecWind_type), intent(inout):: L2C_Height_Bin_VecWind
    integer,                           intent(out) :: error_flag

    ! local variables
    integer :: nr_of_zero_bytes

    error_flag = no_error

    call read_logical1_from_bin_file(fileunit, &
         L2C_Height_Bin_VecWind%Validity_Flag, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_IntAs_from_bin_file(fileunit, &
         L2C_Height_Bin_VecWind%Background_Zonal_WindVeloc, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_IntAs_from_bin_file(fileunit, &
         L2C_Height_Bin_VecWind%Background_Meridional_WindVeloc, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_IntAs_from_bin_file(fileunit, &
         L2C_Height_Bin_VecWind%Analysis_Zonal_WindVeloc, error_flag)
    IF (error_flag .ne. no_error) goto 998
    call read_IntAs_from_bin_file(fileunit, &
         L2C_Height_Bin_VecWind%Analysis_Meridional_WindVeloc, error_flag)
    IF (error_flag .ne. no_error) goto 998
    nr_of_zero_bytes = 20 ! Spare_1
    call read_SpareBytes_from_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Read_L2C_Height_Bin_VecWind"
    return

    end subroutine Read_L2C_Height_Bin_VecWind

    !  #]
  function Get_ByteSize_L2C_Height_Bin_VW() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result

    c_bytesize = 1*c_bytelen_Bool   + &
                 4*c_bytelen_IntAs  + &
                 20 ! Spare 1: 20 bytes

    ! typical sizes
    ! memory size = 4+4*2    = 12 bytes
    ! file   size = 1+4*2+20 = 29 bytes

  end function Get_ByteSize_L2C_Height_Bin_VW
    !  #]
  !---------------------------------------------------------------------
! What is the use of this extra layer of subroutines?
! They just copy the L2C_Height_Bin_VecWind functionality form the 
! routines above.
! Even if the L2C_RayleighVec_MDS module also uses the above 5 routines,
! renaming them first to something with "Mie" in the name does not add
! much to the readability of the code.
! I would just delete the next 5 routines.
! JK, 5-mar-2007
  subroutine Init_MieVec_Height_Bin(MieVec_Height_Bin)
    !  #[
    TYPE(L2C_Height_Bin_VecWind_type), intent(out) :: MieVec_Height_Bin

    call Init_L2C_Height_Bin_VecWind(MieVec_Height_Bin)

  end subroutine Init_MieVec_Height_Bin
    !  #]
  subroutine Print_MieVec_Height_Bin(MieVec_Height_Bin, error_flag)
    !  #[
    TYPE(L2C_Height_Bin_VecWind_type), intent(in)  :: MieVec_Height_Bin
    integer,                           intent(out) :: error_flag    ! output

    call Print_L2C_Height_Bin_VecWind(MieVec_Height_Bin, error_flag)

  end subroutine Print_MieVec_Height_Bin
    !  #]
  subroutine AsciiWrite_MieVec_Height_Bin(MieVec_Height_Bin, &
                                          fileunit, prefix, error_flag)
    !  #[
    TYPE(L2C_Height_Bin_VecWind_type), &
                      intent(in)  :: MieVec_Height_Bin ! input
    integer,          intent(in)  :: fileunit          ! input
    character(len=*), intent(in)  :: prefix            ! input
    integer,          intent(out) :: error_flag        ! output

    call AsciiWrite_L2C_HghtBin_VecWind(MieVec_Height_Bin, &
                                        fileunit, prefix, error_flag)

  end subroutine AsciiWrite_MieVec_Height_Bin
    !  #]
  subroutine Write_L2C_MieVec_HeightBin(fileunit, &
             MieVec_Height_Bin, error_flag)
    !  #[
    integer,                           intent(in)  :: fileunit
    TYPE(L2C_Height_Bin_VecWind_type), intent(in)  :: MieVec_Height_Bin
    integer,                           intent(out) :: error_flag

    call  Write_L2C_Height_Bin_VecWind(fileunit, &
             MieVec_Height_Bin, error_flag)

    end subroutine Write_L2C_MieVec_HeightBin
    !  #]
  subroutine Read_L2C_MieVec_Height_Bin(fileunit, &
             MieVec_Height_Bin, error_flag)
    !  #[
    integer,                           intent(in)  :: fileunit
    TYPE(L2C_Height_Bin_VecWind_type), intent(inout):: MieVec_Height_Bin
    integer,                           intent(out) :: error_flag

    call  Read_L2C_Height_Bin_VecWind(fileunit, &
             MieVec_Height_Bin, error_flag)

    end subroutine Read_L2C_MieVec_Height_Bin
    !  #]
  function Get_ByteSize_L2B_MieVec_H_Bin() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result

    c_bytesize = Get_ByteSize_L2C_Height_Bin_VW()

    ! memory size = 4+4*2    = 12 bytes
    ! file   size = 1+4*2+20 = 29 bytes

  end function Get_ByteSize_L2B_MieVec_H_Bin
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_MieVec_Profile(MieVec_Profile, N_Range_Gates0, error_flag)
    !  #[
    TYPE(MieVec_Profile_type), intent(out) :: MieVec_Profile ! output
    integer(IntAs),            intent(in)  :: N_Range_Gates0 ! input
    integer,                   intent(out) :: error_flag     ! output

    ! local variable
    integer :: r, AllocStatus

    error_flag = no_error

    MieVec_Profile%Obs_Type = Obs_Type_undefined

    IF (N_Range_Gates0 .eq. 0) THEN
       ! init them to be empty
       nullify(MieVec_Profile%Height_Bin_VecWind)
    ELSE
       allocate(MieVec_Profile%Height_Bin_VecWind(N_Range_Gates0),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_MieVec_Profile"
          error_flag = error_allocate
          return
       END IF
       
       DO r=1,N_Range_Gates0
          call Init_MieVec_Height_Bin(MieVec_Profile%Height_Bin_VecWind(r))
       END DO
    END IF

  end subroutine Init_MieVec_Profile
    !  #]
  subroutine Delete_MieVec_Profile(MieVec_Profile)
    !  #[
    TYPE(MieVec_Profile_type), intent(inout) :: MieVec_Profile

    IF (associated(MieVec_Profile%Height_Bin_VecWind)) &
         deallocate(MieVec_Profile%Height_Bin_VecWind)

  end subroutine Delete_MieVec_Profile
    !  #]
! to add: PrintAll
  subroutine Print_MieVec_Profile(MieVec_Profile, error_flag)
    !  #[
    TYPE(MieVec_Profile_type), intent(in)  :: MieVec_Profile ! input
    integer,                   intent(out) :: error_flag     ! output

    ! local variables
    integer :: r, s

    error_flag = no_error

    print *,"MieVec_Profile%Obs_Type = ",&
         trim(GetObsTypeText(int(MieVec_Profile%Obs_Type,IntAuc)))

    IF (associated(MieVec_Profile%Height_Bin_VecWind)) THEN
       ! get the sizes of the array
       s = size(MieVec_Profile%Height_Bin_VecWind)
       print "(a,i2,a)","    ==> size(MieVec_Profile%Height_Bin_VecWind) = ",s,&
                        " [printing first and last one]"
       DO r=1,s,s-1
          print "('    MieVec_Profile%Height_Bin_VecWind(',i2.2,')%...')",r
          call Print_MieVec_Height_Bin(&
               MieVec_Profile%Height_Bin_VecWind(r),error_flag)
          if (error_flag .ne. no_error) return
       END DO
    ELSE
       print *,"    MieVec_Profile%Height_Bin_VecWind(:) [seems not associated]"
    END IF

  end subroutine Print_MieVec_Profile
    !  #]
  subroutine AsciiWrite_MieVec_Profile(MieVec_Profile, fileunit, &
                                       prefix, PrintAll, error_flag)
    !  #[
    TYPE(MieVec_Profile_type), intent(in)  :: MieVec_Profile ! input
    integer,                   intent(in)  :: fileunit       ! input
    character(len=*),          intent(in)  :: prefix         ! input
    logical,                   intent(in)  :: PrintAll       ! input
    integer,                   intent(out) :: error_flag     ! output

    ! local variables
    integer            :: r, s, step
    character(len=256) :: ObsTypeText
    character(len=2)   :: str_cntr

    error_flag = no_error

    ObsTypeText = GetObsTypeText(int(MieVec_Profile%Obs_Type,IntAuc))
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Obs_Type = "//trim(ObsTypeText)

    IF (associated(MieVec_Profile%Height_Bin_VecWind)) THEN
       ! get the sizes of the array
       s = size(MieVec_Profile%Height_Bin_VecWind)
       IF (PrintAll .or. (s .le. 1) ) THEN
          step = 1
       ELSE
          step = s-1
       END IF
       DO r=1,s,step
          write(str_cntr,"(i2.2)",err=998) r
          call  AsciiWrite_MieVec_Height_Bin(&
                         MieVec_Profile%Height_Bin_VecWind(r),&
                         fileunit, trim(prefix)//"%Height_Bin_VecWind("//&
                         trim(str_cntr)//")", error_flag)
          if (error_flag .ne. no_error) return
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%Height_Bin_VecWind(:) [seems not associated]"
    END IF

    return

998 print *,"conversion error in AsciiWrite_MieVec_Profile:"
    print *,"could not convert the counter ",r," to a string .."
    error_flag = error_writing_file
    return

999 print *,"ERROR in AsciiWrite_MieVec_Profile:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_MieVec_Profile
    !  #]
  subroutine Write_L2C_MieVec_Profile(fileunit, &
             MieVec_Profile, error_flag)
    !  #[
    integer,                   intent(in)  :: fileunit
    TYPE(MieVec_Profile_type), intent(in)  :: MieVec_Profile
    integer,                   intent(out) :: error_flag

    ! local variables
    integer :: r, s, nr_of_zero_bytes

    error_flag = no_error
    IF (.not.(associated(MieVec_Profile%Height_Bin_VecWind))) THEN
       print *,"ERROR in Write_L2C_MieVec_Profile():"
       print *,"The pointer to this MieVec_Profile seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! Obs_Type
    call write_IntAc_to_bin_file(fileunit, MieVec_Profile%Obs_Type,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 36 spare bytes                                                            
    nr_of_zero_bytes = 36
    call write_SpareBytes_to_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! get the size of the array
    s = size(MieVec_Profile%Height_Bin_VecWind)
    print *,"Write_L2C_MieVec_Profile: size(MieVec_Profile%Height_Bin_VecWind) =",s
    DO r=1,s
       call Write_L2C_MieVec_HeightBin(fileunit, &
            MieVec_Profile%Height_Bin_VecWind(r), error_flag)
       if (error_flag .ne. no_error) goto 998
    END DO

    !print *,"done with Write_L2C_MieVec_Profile"
    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Write_L2C_MieVec_Profile"
    return

  end subroutine Write_L2C_MieVec_Profile
    !  #]
  subroutine Read_L2C_MieVec_Profile(fileunit, &
             MieVec_Profile, error_flag)
    !  #[

    integer,                   intent(in)  :: fileunit
    TYPE(MieVec_Profile_type), intent(inout):: MieVec_Profile
    integer,                   intent(out) :: error_flag

    ! local variables
    integer :: r, s, nr_of_zero_bytes

    error_flag = no_error
    IF (.not.(associated(MieVec_Profile%Height_Bin_VecWind))) THEN
       print *,"ERROR in Read_L2C_MieVec_Profile():"
       print *,"The pointer to this MieVec_Profile seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! Obs_Type
    call read_IntAc_from_bin_file(fileunit, MieVec_Profile%Obs_Type,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 36 spare bytes
    nr_of_zero_bytes = 36
    call read_SpareBytes_from_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! get the size of the array
    s = size(MieVec_Profile%Height_Bin_VecWind)
    print *,"Read_L2C_MieVec_Profile: size(MieVec_Profile%Height_Bin_VecWind) =",s
    DO r=1,s
       call Read_L2C_MieVec_Height_Bin(fileunit, &
            MieVec_Profile%Height_Bin_VecWind(r), error_flag)
       if (error_flag .ne. no_error) goto 998
    END DO

    !print *,"done with Read_L2C_MieVec_Profile"
    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Read_L2C_MieVec_Profile"
    return

  end subroutine Read_L2C_MieVec_Profile

    !  #]
  function Get_ByteSize_L2B_MieVec_Profile(MieVec_Profile) result(c_bytesize)
    !  #[
    integer(IntAl)                       :: c_bytesize     ! result
    TYPE(MieVec_Profile_type), intent(in):: MieVec_Profile ! input

    c_bytesize = 1*c_bytelen_IntAc + &
                 36_IntAl ! 36 spare bytes
    IF (associated(MieVec_Profile%Height_Bin_VecWind)) THEN
       c_bytesize = c_bytesize + &
            int(size(MieVec_Profile%Height_Bin_VecWind)*&
                Get_ByteSize_L2B_MieVec_H_Bin(),IntAl)
    END IF

    ! typical sizes
    ! memory size = 1+36+24*12 = 325 bytes
    ! file   size = 1+36+24*29 = 733 bytes
  end function Get_ByteSize_L2B_MieVec_Profile
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_L2C_MieVec_MDS(MieVec_MDS, & 
                      N_Meas0, N_Obs_Max_Mie0, N_Range_Gates0, error_flag)
    !  #[
    TYPE(L2C_MieVec_MDS_type), pointer     :: MieVec_MDS     ! input/output
    integer(IntAs),            intent(in)  :: N_Meas0        ! input
    integer(IntAs),            intent(in)  :: N_Obs_Max_Mie0 ! input
    integer(IntAs),            intent(in)  :: N_Range_Gates0 ! input
    integer,                   intent(out) :: error_flag     ! output

    ! local variable
    integer :: p, AllocStatus

    error_flag = no_error

    IF (.not. associated(MieVec_MDS)) THEN
       print *,"ERROR in Init_L2C_MieVec_MDS: "
       print *,"pointer MieVec_MDS seems not associated !!!"
       error_flag = error_programming
       return
    END IF
    
    call InitDateTime(MieVec_MDS%Start_of_Obs_DateTime)

    ! the first 2 are used for allocating/building the datastructure
    ! so I'll keep a copy for convenience
    MieVec_MDS%N_Meas       = N_Meas0
    ! the following should be filled during processing, so init to zero
    MieVec_MDS%N_Obs_Mie_Actual = 0

    IF (N_Obs_Max_Mie0 .eq. 0) THEN
       ! init it to be empty
       nullify(MieVec_MDS%MieVec_Profile)
    ELSE
       allocate(MieVec_MDS%MieVec_Profile(N_Obs_Max_Mie0),&
                STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L2C_MieVec_MDS"
          error_flag = error_allocate
          return
       END IF
       
       DO p=1,N_Obs_Max_Mie0
          call Init_MieVec_Profile(MieVec_MDS%MieVec_Profile(p),&
                                     N_Range_Gates0, error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    END IF

  end subroutine Init_L2C_MieVec_MDS
    !  #]
  subroutine Delete_L2C_MieVec_MDS(MieVec_MDS)
    !  #[
    TYPE(L2C_MieVec_MDS_type), pointer :: MieVec_MDS

    ! local variable
    integer :: p

    IF (.not. associated(MieVec_MDS)) return
    
    IF (associated(MieVec_MDS%MieVec_Profile)) THEN
       DO p=1,size(MieVec_MDS%MieVec_Profile)
          call Delete_MieVec_Profile(MieVec_MDS%MieVec_Profile(p))
       END DO
       deallocate(MieVec_MDS%MieVec_Profile)
    END IF

  end subroutine Delete_L2C_MieVec_MDS
    !  #]
  function Get_ByteSize_L2C_MieVec_MDS(MieVec_MDS, &
                                       N_Mie_Obs_to_write) result(c_bytesize)
    !  #[
    integer(IntAl)                        :: c_bytesize         ! result
    TYPE(L2C_MieVec_MDS_type), pointer    :: MieVec_MDS         ! input
    integer(IntAuc),           intent(in) :: N_Mie_Obs_to_write ! input

    ! local variable
    integer :: s

    c_bytesize = 0
    IF (.not. associated(MieVec_MDS)) return
       
    c_bytesize = c_bytesize + c_bytelen_DateTime
    c_bytesize = c_bytesize + c_bytelen_IntAs ! N_Meas_Actual
    c_bytesize = c_bytesize + c_bytelen_IntAs ! N_Obs_Mie_Actual

    IF (associated(MieVec_MDS%MieVec_Profile)) THEN
       s = size(MieVec_MDS%MieVec_Profile)
       IF (N_Mie_Obs_to_write .gt. s) THEN
          print *,"WARNING in Get_ByteSize_L2C_MieVec_MDS()"
          print *,"invalid value for N_Mie_Obs_to_write: ",N_Mie_Obs_to_write
          print *,"max. possible value is: size(MieVec_MDS%MieVec_Profile) = ",s
          print *,"Get_ByteSize result will be useless...."
       END IF
       c_bytesize = c_bytesize + N_Mie_Obs_to_write * &
            Get_ByteSize_L2B_MieVec_Profile(MieVec_MDS%MieVec_Profile(1))
    END IF

    ! typical sizes
    ! memory size = 20 + 2*2 + 5* 325 + = 1649 bytes
    ! file   size = 12 + 2*2 + 5* 733 + = 3681 bytes
  end function Get_ByteSize_L2C_MieVec_MDS
    !  #]
  !---------------------------------------------------------------------
  ! functions?
  !---------------------------------------------------------------------
  subroutine Print_L2C_MieVec_MDS(MieVec_MDS,error_flag,PrintAll)
    !  #[
    TYPE(L2C_MieVec_MDS_type), pointer     :: MieVec_MDS
    integer,                   intent(out) :: error_flag       ! output
    logical, optional,         intent(in)  :: PrintAll

    ! local variables
    integer                            :: p, s, step
    logical                            :: do_PrintAll
    character(len=maxlen_DateTime_KVT) :: DateTimeString

    error_flag = no_error

    do_PrintAll = .false.
    IF (present(PrintAll)) do_PrintAll = PrintAll

    IF (associated(MieVec_MDS)) THEN
       print *,"MieVec_MDS%..."
    ELSE
       print *,"MieVec_MDS [seems not associated]"
       return
    END IF

    ! print the date and time
    if (DateTimeMissing(MieVec_MDS%start_of_obs_datetime)) then
       DateTimeString = "[missing]"
    else
       DateTimeString = &
            Construct_KVT_DateTime(MieVec_MDS%start_of_obs_datetime)
    end if
    print *,"...%Start_of_Obs_DateTime: "//trim(DateTimeString)
    print *,"...%N_Meas       = ",MieVec_MDS%N_Meas
    ! Poli: keeping the following line otherwise all expected outputs have to be
    ! changed....
    print *,"...%N_Obs_Max    = ",size(MieVec_MDS%MieVec_Profile)
    print *,"...%N_Obs_Actual = ",MieVec_MDS%N_Obs_Mie_Actual

    IF (associated(MieVec_MDS%MieVec_Profile)) THEN
       s = MieVec_MDS%N_Obs_Mie_Actual

       IF (do_PrintAll .or. (s .le. 1) ) THEN
          step=1
          print "(a,i2)"," ==> N_Obs_Actual = ",s
       ELSE
          step = s-1
          print "(a,i2,a)",&
               " ==> N_Obs_Actual = ",s," [printing first and last one]"
       END IF

       DO p=1,s,step
          print "(' ...%MieVec_Profile(',i2.2,')%...')",p
          call Print_MieVec_Profile(&
                   MieVec_MDS%MieVec_Profile(p), error_flag)
       END DO
    ELSE
       print *,"...%MieVec_Profile(:) = [NOT ASSOCIATED]"
    END IF

  end subroutine Print_L2C_MieVec_MDS
    !  #]
  subroutine AsciiWrite_L2C_MieVec_MDS(MieVec_MDS,fileunit,prefix,PrintAll,&
                                       error_flag)
    !  #[
    TYPE(L2C_MieVec_MDS_type), pointer     :: MieVec_MDS ! input
    integer,                   intent(in)  :: fileunit   ! input
    character(len=*),          intent(in)  :: prefix     ! input
    logical,                   intent(in)  :: PrintAll   ! input
    integer,                   intent(out) :: error_flag ! output

    ! local variables
    integer                            :: p, s, step
    character(len=maxlen_DateTime_KVT) :: str_DateTime
    character(len=maxlen_IntAs)        :: str_IntAs
    character(len=2)                   :: str_cntr

    error_flag = no_error

    IF (.not. associated(MieVec_MDS)) THEN
       print *,"Error in AsciiWrite_L2C_MieVec_MDS():"
       print *,"MieVec_MDS [seems not associated]"
       error_flag = error_programming
       return
    END IF

    if (DateTimeMissing(MieVec_MDS%start_of_obs_datetime)) then
       str_DateTime = "[missing]"
    else
       str_DateTime = Construct_KVT_DateTime(MieVec_MDS%start_of_obs_datetime)
    end if
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%start_of_obs_datetime = "//trim(str_DateTime)

    call SetIntAs(MieVec_MDS%N_Meas,str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%N_Meas                = "//trim(str_IntAs)

    call SetIntAs(MieVec_MDS%N_Obs_Mie_Actual,str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%N_Obs_Mie_Actual      = "//trim(str_IntAs)

    IF (associated(MieVec_MDS%MieVec_Profile)) THEN
       s = MieVec_MDS%N_Obs_Mie_Actual
       IF (PrintAll .or. (s .le. 1) ) THEN
          step=1
       ELSE
          step = s-1
       END IF
       DO p=1,s,step
          write(str_cntr,"(i2.2)",err=998) p
          call AsciiWrite_MieVec_Profile(MieVec_MDS%MieVec_Profile(p), &
                      fileunit, &
                      trim(prefix)//"%MieVec_Profile("//trim(str_cntr)//")", &
                      PrintAll, error_flag)
          IF (error_flag .ne. no_error) goto 999
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%MieVec_Profile(:) [seems not associated]"
    END IF

    return

998 print *,"conversion error in AsciiWrite_L2C_MieVec_MDS:"
    print *,"could not convert the counter ",p," to a string .."
    error_flag = error_writing_file
    return

999 print *,"ERROR in AsciiWrite_L2C_MieVec_MDS:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2C_MieVec_MDS
    !  #]
!  subroutine Print_L2C_MieVec_wind_profile(MieVec_MDS,error_flag,Obs_Nr)
    !  #[
!
!    TYPE(L2C_MieVec_MDS_type), pointer     :: MieVec_MDS
!    integer,                   intent(out) :: error_flag       ! output
!    integer,                   intent(in)  :: Obs_Nr
!
!    ! local variables
!    integer                     :: Bin, Nr_Bins
!    character(len=maxlen_IntAs) :: str_IntAs
!
!    error_flag = no_error
!
!  end subroutine Print_L2C_MieVec_wind_profile
    !  #]
  !---------------------------------------------------------------------
  subroutine Write_L2C_MieVec_MDS(fileunit,MieVec_MDS,&
                                  N_Obs_Mie_to_write,error_flag)
    !  #[

    integer,                   intent(in)  :: fileunit           ! input
    TYPE(L2C_MieVec_MDS_type), pointer     :: MieVec_MDS         ! input
    integer(IntAuc),           intent(in)  :: N_Obs_Mie_to_write ! input
    integer,                   intent(out) :: error_flag         ! output

    ! local variables
    integer :: p, s

    !--------------------------------
    ! writing the MieVec MDS elements
    !--------------------------------

    error_flag = no_error
    IF (.not. associated(MieVec_MDS)) THEN
       print *,"ERROR in Write_L2C_MieVec_MDS():"
       print *,"The pointer to this MieVec_MDS seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! Start_of_Obs_DateTime
    call write_DateTime_to_bin_file(fileunit, &
         MieVec_MDS%Start_of_Obs_DateTime,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call write_IntAs_to_bin_file(fileunit,MieVec_MDS%N_Meas,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call write_IntAs_to_bin_file(fileunit,MieVec_MDS%N_Obs_Mie_Actual,error_flag)
    IF (error_flag .ne. no_error) goto 998

    print *,"Write_L2C_MieVec_MDS: MieVec_MDS%N_Obs_Actual = ",&
                                   MieVec_MDS%N_Obs_Mie_Actual
    print *,"Write_L2C_MieVec_MDS: N_Obs_Mie_to_write      = ",&
                                   N_Obs_Mie_to_write

    s = N_Obs_Mie_to_write
    IF (s.gt.0) THEN
       DO p=1,s
          call Write_L2C_MieVec_Profile(fileunit,MieVec_MDS%MieVec_Profile(p),&
                                        error_flag)
          IF (error_flag .ne. no_error) goto 998
       ENDDO
    ENDIF

    !print *,"Write_L2C_MieVec_MDS: done with Write_L2C_MieVec_MDS"

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Write_L2C_MieVec_MDS"
    return

  end subroutine Write_L2C_MieVec_MDS

    !  #]
  subroutine Read_L2C_MieVec_MDS(fileunit,MieVec_MDS,error_flag)
    !  #[
    integer,                   intent(in)  :: fileunit
    TYPE(L2C_MieVec_MDS_type), pointer     :: MieVec_MDS
    integer,                   intent(out) :: error_flag

    ! local variables
    integer                            :: p, s
    character(len=maxlen_DateTime_KVT) :: DateTimeString

    !--------------------------------
    ! reading the MieVec MDS elements
    !--------------------------------

    error_flag = no_error
    IF (.not. associated(MieVec_MDS)) THEN
       print *,"ERROR in Read_L2C_MieVec_MDS():"
       print *,"The pointer to this MieVec_MDS seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! Start_of_Obs_DateTime
    call read_DateTime_from_bin_file(fileunit,&
                      MieVec_MDS%Start_of_Obs_DateTime,error_flag)

    print *,"Read_L2C_MieVec_MDS: Start_of_Obs_DateTime is: "
    if (DateTimeMissing(MieVec_MDS%Start_of_Obs_DateTime)) then
       DateTimeString = "[missing]"
    else
       DateTimeString = &
            Construct_KVT_DateTime(MieVec_MDS%Start_of_Obs_DateTime)
    end if
    print *,DateTimeString

    call read_IntAs_from_bin_file(fileunit,MieVec_MDS%N_Meas,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call read_IntAs_from_bin_file(fileunit,MieVec_MDS%N_Obs_Mie_Actual,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! for N_Obs=0 this pointer may be un-associated
    ! so we have to check for that before requesting its size
    IF (associated(MieVec_MDS%MieVec_Profile)) THEN
       s = size(MieVec_MDS%MieVec_Profile)
       print *,"Read_L2C_MieVec_MDS: MieVec_MDS%N_Obs_Actual =",&
            MieVec_MDS%N_Obs_Mie_Actual
       DO p=1,s
          call Read_L2C_MieVec_Profile(fileunit,&
                                       MieVec_MDS%MieVec_Profile(p),error_flag)
          IF (error_flag .ne. no_error) goto 998
       ENDDO
    END IF

    !print *,"Read_L2C_MieVec_MDS: done with Read_L2C_MieVec_MDS"

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Read_L2C_MieVec_MDS"
    return

  end subroutine Read_L2C_MieVec_MDS
    !  #]
  !---------------------------------------------------------------------
end module L2C_MieVec_MDS
