! todo: remove all error prints, replace by logmsg()

module L2C_RayleighVec_MDS
  !  #[ Documentation
  !
  !     a module to handle the L2C Rayleigh VecWind Measurement dataset
  !     as defined in L2B/L2C-IODD, version 1.1 (draft, 16-Jun-2006).
  !
  !     Written by David Tan, ECMWF
  !     Based on l2b_mie_mds.F90 by  J. de Kloe, KNMI 
  !                              and P. Poli,    Meteo-France
  !     Modifications:
  !     04-Jul-2006 David Tan   Original version
  !     02-Mar-2007 J. de Kloe  adapted to parameter type change in call to 
  !                             GetObsTypeText()
  !     05-Mar-2007 J. de Kloe  adapted to IODD change, 
  !                             and added getBytesize functions
  !     15-Mar-2007 J. de Kloe  removed use of SummaryVars
  !     19-Mar-2007 J. de Kloe  added AsciiWrite routines
  !     23-Apr-2007 J. de Kloe  phase out fillwithzero code
  !     09-Jan-2008 J. de Kloe  added check for missing_DateTime_KVT
  !     16-Jan-2008 J. de Kloe  phase out integer kind i_ 
  !
  !  #]
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE L2C_MieVec_MDS, only :  L2C_Height_Bin_VecWind_type,  &
                              Init_L2C_Height_Bin_VecWind,  &
                              Print_L2C_Height_Bin_VecWind, &
                              Write_L2C_Height_Bin_VecWind, &
                              Read_L2C_Height_Bin_VecWind, &
                              Get_ByteSize_L2C_Height_Bin_VW, &
                              AsciiWrite_L2C_HghtBin_VecWind
  USE EE_CFI_Datatypes, only: &
       IntAc,  IntAuc, IntAl, &
       IntAs,  maxlen_IntAs,  SetIntAs,  &
       missing_indicator_IntAs, &
       DateTimeType,Construct_KVT_DateTime, &
       read_IntAs_from_bin_file, &
       write_IntAs_to_bin_file, &
       read_IntAc_from_bin_file, &
       write_IntAc_to_bin_file, &
       InitDateTime, maxlen_DateTime_KVT, &
       write_DateTime_to_bin_file, &
       read_DateTime_from_bin_file, &
       c_bytelen_Bool, c_bytelen_IntAs, &
       c_bytelen_IntAc, c_bytelen_datetime, &
       write_SpareBytes_to_bin_file, &
       read_SpareBytes_from_bin_file, &
       DateTimeMissing
  
  USE L2B_parameters, only: Obs_Type_undefined, GetObsTypeText
  USE ErrorHandler, only: no_error, error_programming, error_allocate, &
                          error_writing_file
  USE DirectBinaryIO, only: write_logical1_to_bin_file, &
                            read_logical1_from_bin_file
  !  #]
  !  #[ type definitions

  implicit none

  TYPE RayleighVec_Profile_type
     !  #[

     !
     ! An integer (enumerated type) to remember if this was
     ! a cloud or no-cloud profile (or any other type we define)
     integer(IntAc) ::Obs_Type
     !
     ! Spare 1 (36 bytes reserved for Obs_Type_String)
     !
     ! see sec. 4.3.6, L2B/L2C-IODD, vers. 1.1, (draft, 16-Jun-2006)
     ! this profile is an array of size Num_Rangegates
     TYPE(L2C_Height_Bin_VecWind_type), dimension(:), pointer :: Height_Bin_VecWind
  END TYPE RayleighVec_Profile_type

     !  #]

  TYPE L2C_RayleighVec_MDS_type 
     !  #[

     ! see sec. 4.3.6, L2B/L2C-IODD, vers. 1.1 (draft, 16-Jun-2006)
     type(DateTimeType)  :: Start_of_Obs_DateTime    ! UTC
     integer(IntAs)      :: N_Meas       ! nr of Measurements in this product
     integer(IntAs)      :: N_Obs_Rayleigh_Actual ! actual nr of Obs. for this BRC
! Poli: following telecon of 25-01-2006, it is decided that N_Obs_*_Actual
! refers to the number of Mie and Rayleigh profiles with meaningful values.
! Each BRC contains M_Mie and M_Rayleigh profiles (these two numbers are fixed
! for a given file), but in some BRCs there might be less than M_Mie profiles;
! in such case N_Obs_Mie_Actual is less than M_Mie. Same story for Rayleigh.
!   - when initialising, N_Obs_Rayleigh_Actual is set to zero. The structure
!     is dimensioned using M_Rayleigh.
!   - after each BRC is processed, a call to Handle_L2BHeaders will set
!     N_Obs_Rayleigh_Actual to the proper value
!   - when writing, M_Rayleigh profiles are written. The number N_Obs_Rayleigh_Actual
!     is saved so that the user will know when reading the BRC that
!     only the first N_Obs_Rayleigh_Actual profiles are meaningful.
!
!   - when reading, the structure should be dimensioned first to M_Rayleigh.
!
!     ! Remark: the nr of generated observations may vary from one BRC to 
!     ! the next but it was agreed that for one product file the maximum
!     ! N_Obs_max should be used for dimensioning the datastructures 
!     ! that are written to the output file:
!     !         N_Obs_max = max(N_Obs_Actual(first_BRC:last_BRC))
!     ! (see email of David Tan, 20051012, 19.44 utc)
!     ! So: -when initialising N_Obs_Max is used, N_Obs_Actual is set to 0
!     !     -when filling the datastructure N_Obs_Actual is updated
!     !      everytime an observation is generated
!     !     -when printing only N_Obs_Actual is used
!     !     -when writing to file N_Obs_Actual is used, 
!     !      and also all not used Observations are written to file
!
     !
     ! an array for N_Obs_Max profiles for this BRC
     TYPE(RayleighVec_Profile_type), dimension(:), &
          pointer :: RayleighVec_Profile
     !              1234567890123456789012345678901
  END TYPE L2C_RayleighVec_MDS_type

     !  #]

  !  #]
contains
  !---------------------------------------------------------------------
  subroutine Init_RayleighVec_Height_Bin(RayleighVec_Height_Bin)
    !  #[
    TYPE(L2C_Height_Bin_VecWind_type), intent(out) :: RayleighVec_Height_Bin

    call Init_L2C_Height_Bin_VecWind(RayleighVec_Height_Bin)

  end subroutine Init_RayleighVec_Height_Bin
    !  #]
  subroutine Print_RayleighVec_Height_Bin(RayleighVec_Height_Bin, error_flag)
    !  #[
    TYPE(L2C_Height_Bin_VecWind_type), intent(in)  :: RayleighVec_Height_Bin
    integer,                           intent(out) :: error_flag    ! output

    call Print_L2C_Height_Bin_VecWind(RayleighVec_Height_Bin, error_flag)

  end subroutine Print_RayleighVec_Height_Bin
    !  #]
  subroutine AsciiWrite_RaylVec_Height_Bin(RayleighVec_Height_Bin, &
                                           fileunit, prefix, error_flag)
    !  #[
    TYPE(L2C_Height_Bin_VecWind_type), &
                      intent(in)  :: RayleighVec_Height_Bin ! input
    integer,          intent(in)  :: fileunit               ! input
    character(len=*), intent(in)  :: prefix                 ! input
    integer,          intent(out) :: error_flag             ! output

    call AsciiWrite_L2C_HghtBin_VecWind(RayleighVec_Height_Bin, &
                                        fileunit, prefix, error_flag)

  end subroutine AsciiWrite_RaylVec_Height_Bin
    !  #]
  subroutine Write_L2C_RayleighVec_HeightBin(fileunit, &
             RayleighVec_Height_Bin, error_flag)
    !  #[
    integer,                           intent(in)  :: fileunit
    TYPE(L2C_Height_Bin_VecWind_type), intent(in)  :: RayleighVec_Height_Bin
    integer,                           intent(out) :: error_flag

    call  Write_L2C_Height_Bin_VecWind(fileunit, &
             RayleighVec_Height_Bin, error_flag)

    end subroutine Write_L2C_RayleighVec_HeightBin
    !  #]
  subroutine Read_L2C_RayleighVec_Height_Bin(fileunit, &
             RayleighVec_Height_Bin, error_flag)
    !  #[
    integer,                   intent(in)  :: fileunit
    TYPE(L2C_Height_Bin_VecWind_type), intent(inout):: RayleighVec_Height_Bin
    integer,                   intent(out) :: error_flag

    call  Read_L2C_Height_Bin_VecWind(fileunit, &
             RayleighVec_Height_Bin, error_flag)

    end subroutine Read_L2C_RayleighVec_Height_Bin
    !  #]
  function Get_ByteSize_L2B_RaylVec_H_Bin() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result

    c_bytesize = Get_ByteSize_L2C_Height_Bin_VW()

    ! memory size = 4+4*2    = 12 bytes
    ! file   size = 1+4*2+20 = 29 bytes

  end function Get_ByteSize_L2B_RaylVec_H_Bin
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_RayleighVec_Profile(RayleighVec_Profile, N_Range_Gates0, error_flag)
    !  #[
    TYPE(RayleighVec_Profile_type), intent(out) :: RayleighVec_Profile ! output
    integer(IntAs),            intent(in)  :: N_Range_Gates0 ! input
    integer,                   intent(out) :: error_flag     ! output

    ! local variable
    integer :: r, AllocStatus

    error_flag = no_error

    RayleighVec_Profile%Obs_Type = Obs_Type_undefined

    IF (N_Range_Gates0 .eq. 0) THEN
       ! init them to be empty
       nullify(RayleighVec_Profile%Height_Bin_VecWind)
    ELSE
       allocate(RayleighVec_Profile%Height_Bin_VecWind(N_Range_Gates0),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_RayleighVec_Profile"
          error_flag = error_allocate
          return
       END IF
       
       DO r=1,N_Range_Gates0
          call Init_RayleighVec_Height_Bin(RayleighVec_Profile%Height_Bin_VecWind(r))
       END DO
    END IF

  end subroutine Init_RayleighVec_Profile
    !  #]
  subroutine Delete_RayleighVec_Profile(RayleighVec_Profile)
    !  #[
    TYPE(RayleighVec_Profile_type), intent(inout) :: RayleighVec_Profile

    IF (associated(RayleighVec_Profile%Height_Bin_VecWind)) &
         deallocate(RayleighVec_Profile%Height_Bin_VecWind)

  end subroutine Delete_RayleighVec_Profile
    !  #]
! to add: PrintAll
  subroutine Print_RayleighVec_Profile(RayleighVec_Profile, error_flag)
    !  #[
    TYPE(RayleighVec_Profile_type), intent(in)  :: RayleighVec_Profile ! input
    integer,                   intent(out) :: error_flag     ! output

    ! local variables
    integer :: r, s

    error_flag = no_error

    print *,"RayleighVec_Profile%Obs_Type = ",&
         trim(GetObsTypeText(int(RayleighVec_Profile%Obs_Type,IntAuc)))

    IF (associated(RayleighVec_Profile%Height_Bin_VecWind)) THEN
       ! get the sizes of the array
       s = size(RayleighVec_Profile%Height_Bin_VecWind)
       print "(a,i2,a)","    ==> size(RayleighVec_Profile%Height_Bin_VecWind) = ",s,&
                        " [printing first and last one]"
       DO r=1,s,s-1
          print "('    RayleighVec_Profile%Height_Bin_VecWind(',i2.2,')%...')",r
          call Print_RayleighVec_Height_Bin(&
               RayleighVec_Profile%Height_Bin_VecWind(r),error_flag)
          if (error_flag .ne. no_error) return
       END DO
    ELSE
       print *,"    RayleighVec_Profile%Height_Bin_VecWind(:) [seems not associated]"
    END IF

  end subroutine Print_RayleighVec_Profile
    !  #]
  subroutine AsciiWrite_RaylVec_Profile(RayleighVec_Profile, fileunit, &
                                        prefix, PrintAll, error_flag)
    !  #[
    TYPE(RayleighVec_Profile_type), intent(in) :: RayleighVec_Profile ! input
    integer,                   intent(in)  :: fileunit       ! input
    character(len=*),          intent(in)  :: prefix         ! input
    logical,                   intent(in)  :: PrintAll       ! input
    integer,                   intent(out) :: error_flag     ! output

    ! local variables
    integer :: r, s, step
    character(len=256) :: ObsTypeText
    character(len=2)   :: str_cntr

    error_flag = no_error

    ObsTypeText = GetObsTypeText(int(RayleighVec_Profile%Obs_Type,IntAuc))
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Obs_Type = "//trim(ObsTypeText)

    IF (associated(RayleighVec_Profile%Height_Bin_VecWind)) THEN
       ! get the sizes of the array
       s = size(RayleighVec_Profile%Height_Bin_VecWind)
       IF (PrintAll .or. (s .le. 1) ) THEN
          step = 1
       ELSE
          step = s-1
       END IF
       DO r=1,s,step
          write(str_cntr,"(i2.2)",err=998) r
          call  AsciiWrite_RaylVec_Height_Bin(&
                         RayleighVec_Profile%Height_Bin_VecWind(r),&
                         fileunit, trim(prefix)//"%Height_Bin_VecWind("//&
                         trim(str_cntr)//")", error_flag)
          if (error_flag .ne. no_error) return
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%Height_Bin_VecWind(:) [seems not associated]"
    END IF

    return

998 print *,"conversion error in AsciiWrite_RaylVec_Profile:"
    print *,"could not convert the counter ",r," to a string .."
    error_flag = error_writing_file
    return

999 print *,"ERROR in AsciiWrite_RaylVec_Profile:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_RaylVec_Profile
    !  #]
  subroutine Write_L2C_RayleighVec_Profile(fileunit, &
             RayleighVec_Profile, error_flag)
    !  #[

    integer,                   intent(in)  :: fileunit
    TYPE(RayleighVec_Profile_type), intent(in)  :: RayleighVec_Profile
    integer,                   intent(out) :: error_flag

    ! local variables
    integer :: r, s, nr_of_zero_bytes

    error_flag = no_error
    IF (.not.(associated(RayleighVec_Profile%Height_Bin_VecWind))) THEN
       print *,"ERROR in Write_L2C_RayleighVec_Profile():"
       print *,"The pointer to this RayleighVec_Profile seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! Obs_Type
    call write_IntAc_to_bin_file(fileunit, RayleighVec_Profile%Obs_Type,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 36 spare bytes                                                            
    nr_of_zero_bytes = 36
    call write_SpareBytes_to_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! get the size of the array
    s = size(RayleighVec_Profile%Height_Bin_VecWind)
    print *,"Write_L2C_RayleighVec_Profile: size(RayleighVec_Profile%Height_Bin_VecWind) =",s
    DO r=1,s
       call Write_L2C_RayleighVec_HeightBin(fileunit, &
            RayleighVec_Profile%Height_Bin_VecWind(r), error_flag)
       if (error_flag .ne. no_error) goto 998
    END DO

    !print *,"done with Write_L2C_RayleighVec_Profile"
    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Write_L2C_RayleighVec_Profile"
    return

  end subroutine Write_L2C_RayleighVec_Profile

    !  #]
  subroutine Read_L2C_RayleighVec_Profile(fileunit, &
             RayleighVec_Profile, error_flag)
    !  #[

    integer,                   intent(in)  :: fileunit
    TYPE(RayleighVec_Profile_type), intent(inout):: RayleighVec_Profile
    integer,                   intent(out) :: error_flag

    ! local variables
    integer :: r, s, nr_of_zero_bytes

    error_flag = no_error
    IF (.not.(associated(RayleighVec_Profile%Height_Bin_VecWind))) THEN
       print *,"ERROR in Read_L2C_RayleighVec_Profile():"
       print *,"The pointer to this RayleighVec_Profile seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! Obs_Type
    call read_IntAc_from_bin_file(fileunit, RayleighVec_Profile%Obs_Type,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 36 spare bytes                                                            
    nr_of_zero_bytes = 36
    call read_SpareBytes_from_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! get the size of the array
    s = size(RayleighVec_Profile%Height_Bin_VecWind)
    print *,"Read_L2C_RayleighVec_Profile: size(RayleighVec_Profile%Height_Bin_VecWind) =",s
    DO r=1,s
       call Read_L2C_RayleighVec_Height_Bin(fileunit, &
            RayleighVec_Profile%Height_Bin_VecWind(r), error_flag)
       if (error_flag .ne. no_error) goto 998
    END DO

    !print *,"done with Read_L2C_RayleighVec_Profile"
    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Read_L2C_RayleighVec_Profile"
    return

  end subroutine Read_L2C_RayleighVec_Profile

    !  #]
  function Get_ByteSize_L2B_RaylVec_Prof(RayleighVec_Profile) result(c_bytesize)
    !  #[
    integer(IntAl)                            :: c_bytesize          ! result
    TYPE(RayleighVec_Profile_type), intent(in):: RayleighVec_Profile ! input

    c_bytesize = 1*c_bytelen_IntAc + &
                 36_IntAl ! 36 spare bytes
    IF (associated(RayleighVec_Profile%Height_Bin_VecWind)) THEN
       c_bytesize = c_bytesize + &
            int(size(RayleighVec_Profile%Height_Bin_VecWind)*&
                Get_ByteSize_L2B_RaylVec_H_Bin(),IntAl)
    END IF

    ! typical sizes
    ! memory size = 1+36+24*12 = 325 bytes
    ! file   size = 1+36+24*29 = 733 bytes
  end function Get_ByteSize_L2B_RaylVec_Prof
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_L2C_RayleighVec_MDS(RayleighVec_MDS, & 
                      N_Meas0, N_Obs_Max_Rayleigh0, N_Range_Gates0, error_flag)
    !  #[
    TYPE(L2C_RayleighVec_MDS_type), pointer  :: RayleighVec_MDS  ! input/output
    integer(IntAs),              intent(in)  :: N_Meas0        ! input
    integer(IntAs),              intent(in)  :: N_Obs_Max_Rayleigh0     ! input
    integer(IntAs),              intent(in)  :: N_Range_Gates0 ! input
    integer,                     intent(out) :: error_flag    ! output

    ! local variable
    integer :: p, AllocStatus

    error_flag = no_error

    IF (.not. associated(RayleighVec_MDS)) THEN
       print *,"ERROR in Init_L2C_RayleighVec_MDS: "
       print *,"pointer RayleighVec_MDS seems not associated !!!"
       error_flag = error_programming
       return
    END IF
    
    call InitDateTime(RayleighVec_MDS%Start_of_Obs_DateTime)

    ! the first 2 are used for allocating/building the datastructure
    ! so I'll keep a copy for convenience
    RayleighVec_MDS%N_Meas       = N_Meas0
    ! the following should be filled during processing, so init to zero
    RayleighVec_MDS%N_Obs_Rayleigh_Actual = 0

    IF (N_Obs_Max_Rayleigh0 .eq. 0) THEN
       ! init it to be empty
       nullify(RayleighVec_MDS%RayleighVec_Profile)
    ELSE
       allocate(RayleighVec_MDS%RayleighVec_Profile(N_Obs_Max_Rayleigh0),&
                STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_L2C_RayleighVec_MDS"
          error_flag = error_allocate
          return
       END IF
       
       DO p=1,N_Obs_Max_Rayleigh0
          call Init_RayleighVec_Profile(RayleighVec_MDS%RayleighVec_Profile(p),&
                                     N_Range_Gates0, error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    END IF

  end subroutine Init_L2C_RayleighVec_MDS
    !  #]
  subroutine Delete_L2C_RayleighVec_MDS(RayleighVec_MDS)
    !  #[
    TYPE(L2C_RayleighVec_MDS_type), pointer :: RayleighVec_MDS

    ! local variable
    integer :: p

    IF (.not. associated(RayleighVec_MDS)) return
    
    IF (associated(RayleighVec_MDS%RayleighVec_Profile)) THEN
       DO p=1,size(RayleighVec_MDS%RayleighVec_Profile)
          call Delete_RayleighVec_Profile(RayleighVec_MDS%RayleighVec_Profile(p))
       END DO
       deallocate(RayleighVec_MDS%RayleighVec_Profile)
    END IF

  end subroutine Delete_L2C_RayleighVec_MDS
    !  #]
  function Get_ByteSize_L2C_RaylVec_MDS(RayleighVec_MDS, &
                                        N_Rayleigh_Obs_to_write) result(c_bytesize)
    !  #[
    integer(IntAl)                             :: c_bytesize              ! result
    TYPE(L2C_RayleighVec_MDS_type), pointer    :: RayleighVec_MDS         ! input
    integer(IntAuc),                intent(in) :: N_Rayleigh_Obs_to_write ! input

    ! local variable
    integer :: s

    c_bytesize = 0
    IF (.not. associated(RayleighVec_MDS)) return
       
    c_bytesize = c_bytesize + c_bytelen_DateTime
    c_bytesize = c_bytesize + c_bytelen_IntAs ! N_Meas_Actual
    c_bytesize = c_bytesize + c_bytelen_IntAs ! N_Obs_Mie_Actual

    IF (associated(RayleighVec_MDS%RayleighVec_Profile)) THEN
       s = size(RayleighVec_MDS%RayleighVec_Profile)
       IF (N_Rayleigh_Obs_to_write .gt. s) THEN
          print *,"WARNING in Get_ByteSize_L2C_RaylVec_MDS()"
          print *,"invalid value for N_Rayleigh_Obs_to_write: ",&
               N_Rayleigh_Obs_to_write
          print *,"max. possible value is: "//&
               "size(RayleighVec_MDS%RayleighVec_Profile) = ",s
          print *,"Get_ByteSize result will be useless...."
       END IF
       c_bytesize = c_bytesize + N_Rayleigh_Obs_to_write * &
            Get_ByteSize_L2B_RaylVec_Prof(RayleighVec_MDS%RayleighVec_Profile(1))
    END IF

    ! typical sizes
    ! memory size = 20 + 2*2 + 5* 325 + = 1649 bytes
    ! file   size = 12 + 2*2 + 5* 733 + = 3681 bytes
  end function Get_ByteSize_L2C_RaylVec_MDS
    !  #]
  !---------------------------------------------------------------------

  ! functions?

  !---------------------------------------------------------------------
  subroutine Print_L2C_RayleighVec_MDS(RayleighVec_MDS,error_flag,PrintAll)
    !  #[

    TYPE(L2C_RayleighVec_MDS_type), pointer  :: RayleighVec_MDS
    integer,                     intent(out) :: error_flag       ! output
    logical, optional,           intent(in)  :: PrintAll

    ! local variables
    integer                            :: p, s, step
    logical                            :: do_PrintAll
    character(len=maxlen_DateTime_KVT) :: DateTimeString

    error_flag = no_error

    do_PrintAll = .false.
    IF (present(PrintAll)) do_PrintAll = PrintAll

    IF (associated(RayleighVec_MDS)) THEN
       print *,"RayleighVec_MDS%..."
    ELSE
       print *,"RayleighVec_MDS [seems not associated]"
       return
    END IF

    ! print the date and time
    if (DateTimeMissing(RayleighVec_MDS%start_of_obs_datetime)) then
       DateTimeString = "[missing]"
    else
       DateTimeString = &
            Construct_KVT_DateTime(RayleighVec_MDS%start_of_obs_datetime)
    end if
    print *,"...%Start_of_Obs_DateTime: "//trim(DateTimeString)
    print *,"...%N_Meas       = ",RayleighVec_MDS%N_Meas
    ! Poli: keeping the following line otherwise all expected outputs have to be
    ! changed....
    print *,"...%N_Obs_Max    = ",size(RayleighVec_MDS%RayleighVec_Profile)
    print *,"...%N_Obs_Actual = ",RayleighVec_MDS%N_Obs_Rayleigh_Actual

    IF (associated(RayleighVec_MDS%RayleighVec_Profile)) THEN
       s = RayleighVec_MDS%N_Obs_Rayleigh_Actual

       IF (do_PrintAll .or. (s .le. 1) ) THEN
          step=1
          print "(a,i2)"," ==> N_Obs_Actual = ",s
       ELSE
          step = s-1
          print "(a,i2,a)",&
               " ==> N_Obs_Actual = ",s," [printing first and last one]"
       END IF

       DO p=1,s,step
          print "(' ...%RayleighVec_Profile(',i2.2,')%...')",p
          call Print_RayleighVec_Profile(&
                   RayleighVec_MDS%RayleighVec_Profile(p), error_flag)
       END DO
    ELSE
       print *,"...%RayleighVec_Profile(:) = [NOT ASSOCIATED]"
    END IF

  end subroutine Print_L2C_RayleighVec_MDS

    !  #]
  subroutine AsciiWrite_L2C_RaylVec_MDS(RayleighVec_MDS,fileunit,prefix,&
                                        PrintAll,error_flag)
    !  #[

    TYPE(L2C_RayleighVec_MDS_type), pointer :: RayleighVec_MDS ! input
    integer,                   intent(in)  :: fileunit   ! input
    character(len=*),          intent(in)  :: prefix     ! input
    logical,                   intent(in)  :: PrintAll   ! input
    integer,                   intent(out) :: error_flag ! output

    ! local variables
    integer                            :: p, s, step
    character(len=maxlen_DateTime_KVT) :: str_DateTime
    character(len=maxlen_IntAs)        :: str_IntAs
    character(len=2)                   :: str_cntr

    error_flag = no_error

    IF (.not. associated(RayleighVec_MDS)) THEN
       print *,"Error in AsciiWrite_L2C_RaylVec_MDS():"
       print *,"RayleighVec_MDS [seems not associated]"
       error_flag = error_programming
       return
    END IF

    if (DateTimeMissing(RayleighVec_MDS%start_of_obs_datetime)) then
       str_DateTime = "[missing]"
    else
       str_DateTime = Construct_KVT_DateTime(&
                               RayleighVec_MDS%start_of_obs_datetime)
    end if
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%start_of_obs_datetime = "//trim(str_DateTime)

    call SetIntAs(RayleighVec_MDS%N_Meas,str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%N_Meas                = "//trim(str_IntAs)

    call SetIntAs(RayleighVec_MDS%N_Obs_Rayleigh_Actual,str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%N_Obs_Rayleigh_Actual = "//trim(str_IntAs)

    IF (associated(RayleighVec_MDS%RayleighVec_Profile)) THEN
       s = RayleighVec_MDS%N_Obs_Rayleigh_Actual
       IF (PrintAll .or. (s .le. 1) ) THEN
          step=1
       ELSE
          step = s-1
       END IF
       DO p=1,s,step
          write(str_cntr,"(i2.2)",err=998) p
          call AsciiWrite_RaylVec_Profile(&
                  RayleighVec_MDS%RayleighVec_Profile(p), &
                  fileunit, &
                  trim(prefix)//"%RayleighVec_Profile("//trim(str_cntr)//")", &
                  PrintAll, error_flag)
          IF (error_flag .ne. no_error) goto 999
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%RayleighVec_Profile(:) [seems not associated]"
    END IF

    return

998 print *,"conversion error in AsciiWrite_L2C_RaylVec_MDS:"
    print *,"could not convert the counter ",p," to a string .."
    error_flag = error_writing_file
    return

999 print *,"ERROR in AsciiWrite_L2C_RaylVec_MDS:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2C_RaylVec_MDS

    !  #]

!  subroutine Print_L2C_RayleighVec_wind_profile(RayleighVec_MDS,error_flag,Obs_Nr)
    !  #[
!
!    TYPE(L2C_RayleighVec_MDS_type), pointer       :: RayleighVec_MDS
!    integer,                     intent(out) :: error_flag       ! output
!    integer,                     intent(in)  :: Obs_Nr
!
!    ! local variables
!    integer                      :: Bin, Nr_Bins
!    character(len=maxlen_IntAs)  :: str_IntAs
!
!    error_flag = no_error
!
!  end subroutine Print_L2C_RayleighVec_wind_profile
    !  #]
  !---------------------------------------------------------------------
  subroutine Write_L2C_RayleighVec_MDS(fileunit,RayleighVec_MDS,&
                                       N_Obs_Rayleigh_to_write,error_flag)
    !  #[

    integer,                   intent(in)  :: fileunit                ! input
    TYPE(L2C_RayleighVec_MDS_type), pointer:: RayleighVec_MDS         ! input
    integer(IntAuc),           intent(in)  :: N_Obs_Rayleigh_to_write ! input
    integer,                   intent(out) :: error_flag              ! output

    ! local variables
    integer :: p, s

    !--------------------------------
    ! writing the RayleighVec MDS elements
    !--------------------------------

    error_flag = no_error
    IF (.not. associated(RayleighVec_MDS)) THEN
       print *,"ERROR in Write_L2C_RayleighVec_MDS():"
       print *,"The pointer to this RayleighVec_MDS seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! Start_of_Obs_DateTime
    call write_DateTime_to_bin_file(fileunit, &
         RayleighVec_MDS%Start_of_Obs_DateTime,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call write_IntAs_to_bin_file(fileunit,RayleighVec_MDS%N_Meas,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call write_IntAs_to_bin_file(fileunit,RayleighVec_MDS%N_Obs_Rayleigh_Actual,error_flag)
    IF (error_flag .ne. no_error) goto 998

    s = N_Obs_Rayleigh_to_write

! this is wrong: s = RayleighVec_MDS%N_Obs_Rayleigh_Actual
! since N_Obs_Rayleigh_to_write = max(RayleighVec_MDS%N_Obs_Rayleigh_Actual)
! taken over all BRC's. 

    print *,"Write_L2C_RayleighVec_MDS: RayleighVec_MDS%N_Obs_Actual =",&
                                        RayleighVec_MDS%N_Obs_Rayleigh_Actual
    print *,"Write_L2C_RayleighVec_MDS: N_Obs_Rayleigh_to_write      =",&
                                        N_Obs_Rayleigh_to_write
    IF (s.gt.0) THEN
       DO p=1,s
          call Write_L2C_RayleighVec_Profile(fileunit,&
                      RayleighVec_MDS%RayleighVec_Profile(p),error_flag)
          IF (error_flag .ne. no_error) goto 998
       ENDDO
    ENDIF

    !print *,"Write_L2C_RayleighVec_MDS: done with Write_L2C_RayleighVec_MDS"

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Write_L2C_RayleighVec_MDS"
    return

  end subroutine Write_L2C_RayleighVec_MDS

    !  #]
  subroutine Read_L2C_RayleighVec_MDS(fileunit,RayleighVec_MDS,error_flag)
    !  #[

    integer,                        intent(in)  :: fileunit
    TYPE(L2C_RayleighVec_MDS_type), pointer     :: RayleighVec_MDS
    integer,                        intent(out) :: error_flag

    ! local variables
    integer                            :: p, s
    character(len=maxlen_DateTime_KVT) :: DateTimeString

    !--------------------------------
    ! reading the RayleighVec MDS elements
    !--------------------------------

    error_flag = no_error
    IF (.not. associated(RayleighVec_MDS)) THEN
       print *,"ERROR in Read_L2C_RayleighVec_MDS():"
       print *,"The pointer to this RayleighVec_MDS seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! Start_of_Obs_DateTime
    call read_DateTime_from_bin_file(fileunit,&
                      RayleighVec_MDS%Start_of_Obs_DateTime,error_flag)

    print *,"Read_L2C_RayleighVec_MDS: Start_of_Obs_DateTime is: "
    if (DateTimeMissing(RayleighVec_MDS%Start_of_Obs_DateTime)) then
       DateTimeString = "[missing]"
    else
       DateTimeString = &
            Construct_KVT_DateTime(RayleighVec_MDS%Start_of_Obs_DateTime)
    end if
    print *,DateTimeString

    call read_IntAs_from_bin_file(fileunit,RayleighVec_MDS%N_Meas,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call read_IntAs_from_bin_file(fileunit,&
                            RayleighVec_MDS%N_Obs_Rayleigh_Actual,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! for N_Obs=0 this pointer may be un-associated
    ! so we have to check for that before requesting its size
    IF (associated(RayleighVec_MDS%RayleighVec_Profile)) THEN
       s = size(RayleighVec_MDS%RayleighVec_Profile)
       print *,"Read_L2C_RayleighVec_MDS: RayleighVec_MDS%N_Obs_Actual =",&
            RayleighVec_MDS%N_Obs_Rayleigh_Actual
       DO p=1,s
          call Read_L2C_RayleighVec_Profile(fileunit,&
                      RayleighVec_MDS%RayleighVec_Profile(p),error_flag)
          IF (error_flag .ne. no_error) goto 998
       ENDDO
    END IF

    !print *,"Read_L2C_RayleighVec_MDS: done with Read_L2C_RayleighVec_MDS"

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Read_L2C_RayleighVec_MDS"
    return

  end subroutine Read_L2C_RayleighVec_MDS

    !  #]
  !---------------------------------------------------------------------
end module L2C_RayleighVec_MDS
