module MainProductHeader
  !  #[ Documentation

  !
  ! this module defines the Main Product Header for the ee_cfi fileformat
  !
  ! the next type definitions are generic for aeolus files
  ! and are used for both the L1B and L2B file headers
  !
  ! for a description, see section 3.3.2 of the L1B_iodd_2.1
  ! or L2B TN2.1, version 1.2, (6-9-2005), sec.4.3, p.15, table 9
  !
  ! Written by J. de Kloe (KNMI)
  ! Created:     08-Sep-2005
  !
  ! History:
  !    28-Nov-2005 P. Poli    Added routine Write_MPH_to_KVT
  !                           Added routine Fill_L2B_MPH_From_L1B_MPH
  !    01-Dec-2005 P. Poli    Added routine Fill_L2B_MPH_Sizes
  !    01-Dec-2005 P. Poli    Added routine Fill_AMD_MPH_Sizes
  !    11-Jan-2006 J. de Kloe Removed a not used USE statement
  !    02-Feb-2006 J. de Kloe corrected handling of Leap_err and Product_Err 
  !                           fields (which differ for XML and KVT) after
  !                           update of their definition in MDA IODD v2.2
  !    16-Feb-2006 J. de Kloe added extra inputs for Fill_MPH_items
  !    17-Feb-2006 J. de Kloe Reverted the change from type IntAl to IntAul
  !                           for some items inside the MPH for now.
  !                           (we need to discuss this first)
  !    24-Feb-2006 J. de Kloe Added GetByteSize_MPH_KVT() function
  !    28-Feb-2006 J. de Kloe Added prefix parameter to print routine
  !    02-Mar-2006 J. de Kloe Adapted to changes in XML_module
  !    13-Mar-2006 J. de Kloe added routine AsciiWrite_MPH
  !    28-Mar-2006 J. de Kloe Some dummy values for Fill_MPH_Items.
  !    13-Apr-2006 D. Tan     Move inits from Fill_MPH_items to Init_MPH
  !    12-May-2006 D. Tan     Null use of ParKind1 module
  !    12-May-2006 J. de Kloe made Write_MPH_to_KVT generic
  !    15-Sep-2006 J. de Kloe rephrased an error message (when the
  !                           get_filesize command fails)
  !    03-Nov-2006 P. Poli    upgrade to L1BP v1.04
  !    09-Mar-2007 D. Tan     Use ee_mph_name.
  !    15-Mar-2007 J. de Kloe removed use of SummaryVars for L2B
  !    23-Mar-2007 J. de Kloe introduced workaround_in parameter in
  !                           subroutine Read_MPH_KVT()
  !    24-Apr-2007 J. de Kloe removed use of SummaryVars for AMD
  !    07-May-2007 J. de Kloe added routine GetFileVersion
  !    07-Jun-2007 J. de Kloe adapted to changed xml interface by adding
  !                           a path_in parameter to the GetFrom/CopyTo xml 
  !                           subroutines
  !    15-Jun-2007 J. de Kloe changed GetFileVersion() to return integer
  !                           version numbers as well
  !    31-Jul-2007 J. de Kloe renamed Get_Current_DateTime to 
  !                           GetCurrentDateTime 
  !    18-Oct-2007 J. de Kloe updated fill_MPH_items to enable use the 
  !                           MPH_Default_Fields
  !    14-Nov-2007 J. de Kloe updated to fill all datetime fields with the
  !                           new missing_DateTime_KVT string parameter
  !    06-Dec-2007 J. de Kloe adapted to no longer copy sensing_start/stop
  !                           from the L1B file, but allow use of proper
  !                           values in stead (needed in case only a selection
  !                           of BRC's has been processed)
  !    08-Jan-2007 J. de Kloe added extra checks on the validity of the input
  !                           XML numbers (should prevent the problem reported
  !                           by Finn due to the out-of-range Abs_Orbit value
  !    08-Jan-2007 J. de Kloe converted most warnings and errors to logmsg
  !                           calls, and removed the old print routine
  !    09-Jan-2008 J. de Kloe added check for missing_DateTime_KVT
  !                           in AsciiWrite_MPH routine
  !    16-Jan-2008 J. de Kloe phase out integer kind i_ 
  !    18-Jan-2008 J. de Kloe changed the type for MPH%ABS_ORBIT from IntAs
  !                           to IntAus, anticipating an IODD change needed
  !                           to accomodate orbit numbers>32767 needed for
  !                           the GSOV testing. (confirmed now by the draft
  !                           IODD v1.08, JK)
  !    21-Feb-2008 J. de Kloe adapted to L1B v1.09 and made workarounds that
  !                           we have for xml handling explicit
  !  #] 
  !  #[ Modules used

  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE XML_module, only: max_str_len, xml_node_type, xml_node_get_string, &
                        xml_tree_add_child, xml_tree_set_string_node_value, &
                        ee_vh_name, ee_mph_name, check_length_xml_item, &
                        delete_xml_tree, xml_check_node_exists, xml_load_file
  USE KVT_module, only: Init_KVT_module, &
                        read_kvt_string, read_spare_kvt_string, &
                        write_kvt_string, write_spare_kvt_string
  USE DirectBinaryIO, only: DirectBinaryIO_fileunit_type, &
                            GetFilePointer, SetFilePointer, &
                            Init_DirectBinaryIO_fileunit
  USE Buffering, only: BufferedFile_type
  USE EE_CFI_Datatypes, only: &
       DateTimeType, maxlen_DateTime_KVT, maxlen_DateTime_XML, &
       TestDateTime, missing_DateTime_KVT, &
       Convert_XML_to_KVT_DateTime, Convert_KVT_to_XML_DateTime,&
       GetCurrentDateTime, Construct_KVT_DateTime, missing_DateTime_KVT, &
       IntAuc, SetIntAuc, maxlen_IntAuc, maxval_IntAuc, GetIntAuc, &
       IntAc,  SetIntAc,  maxlen_IntAc,  maxval_IntAc,  GetIntAc,  &
       IntAs,  SetIntAs,  maxlen_IntAs,  maxval_IntAs,  GetIntAs,  &
       IntAus, SetIntAus, maxlen_IntAus, maxval_IntAus, GetIntAus, &
       IntAul, SetIntAul, maxlen_IntAul, maxval_IntAul, GetIntAul, &
       IntAl,  SetIntAl,  maxlen_IntAl,  maxval_IntAl,  GetIntAl,  &
       IntAd,  SetIntAd,  maxlen_IntAd,  maxval_IntAd,  GetIntAd,  &
       FAdoxy, SetFAdoxy, maxlen_FAdo06, maxlen_FAdo46, maxlen_FAdo73, &
       maxval_FAdo06, maxval_FAdo46, maxval_FAdo73, GetFAdoxy, &
       maxlen_BoolKVT, maxlen_BoolXML, SetBoolKVT, &       
       Remove_Plus_Sign
  USE DataSetDescriptor, only: GetByteSize_DSD_KVT

  ! support modules
  USE Logging, only: logmsg,log_debug,log_info,log_progress,log_warn,log_error
  USE c_support, only: get_filesize
  USE ErrorHandler, only: no_error, error_allocate, error_programming, &
                          error_reading_file, error_writing_file
  USE Numerics, only: i8_, i4_
  USE LunManager, only: fileunit_stdout
  USE Compiler_Features, only: newline
  USE StringTools, only: remove_eol, to_lowercase
  !  #]
  !  #[ Type definitions and parameters

  implicit none
  !                                          L2B    L1B
  integer, parameter :: maxlen_Product      = 62!56! 62  ! string
  integer, parameter :: maxlen_Proc_Stage   =  1!4 !  1  ! enum
  integer, parameter :: maxlen_Ref_Doc      = 23!32! 23  ! string
  integer, parameter :: maxlen_Acq_Stat     = 20   !dito ! string
! this was changed for L1Bv1.08 without any communication to us
! it seems this was a mistake, and has been reverted for L1Bv1.09
  integer, parameter :: maxlen_Acq_Stat_108 = 24   !dito ! string


  integer, parameter :: maxlen_Proc_Center  =  6   !dito ! string
  integer, parameter :: maxlen_Software_Ver = 14   !dito ! string
  integer, parameter :: maxlen_Phase        =  1   !dito ! enum
  integer, parameter :: maxlen_Vecor_Src    =  2   !dito ! string ! not used 

  TYPE MPH_type
    !  #[

     ! GENERAL REMARK:
     ! we decided to store the KVT format if an item has different format
     ! definitions for XML and KVT. The code is written to convert 
     ! automatically to the XML definitions when needed.

     character(len=maxlen_Product)     :: Product
     character(len=maxlen_Proc_Stage)  :: Proc_Stage
     character(len=maxlen_Ref_Doc)     :: Ref_Doc
     ! Spare 1
     character(len=maxlen_Acq_Stat_108):: Acquisition_Station
     character(len=maxlen_Proc_Center) :: Proc_Center
     character(len=maxlen_DateTime_KVT):: Proc_Time
     character(len=maxlen_Software_Ver):: Software_Ver
     ! Spare 2
     character(len=maxlen_DateTime_KVT):: Sensing_Start
     character(len=maxlen_DateTime_KVT):: Sensing_Stop
     ! Spare 3
     character(len=maxlen_Phase)       :: Phase
     character(len=maxlen_IntAuc)      :: Cycle
     character(len=maxlen_IntAs)       :: Rel_Orbit
     character(len=maxlen_IntAus)      :: Abs_Orbit
     character(len=maxlen_DateTime_KVT):: State_Vector_Time
     character(len=maxlen_FAdo06)      :: Delta_UT1
     character(len=maxlen_FAdo73)      :: X_Position
     character(len=maxlen_FAdo73)      :: Y_Position
     character(len=maxlen_FAdo73)      :: Z_Position
     character(len=maxlen_FAdo46)      :: X_Velocity
     character(len=maxlen_FAdo46)      :: Y_Velocity
     character(len=maxlen_FAdo46)      :: Z_Velocity
     character(len=maxlen_Vecor_Src)   :: Vector_Source! not used by ADM_Aeolus
     ! Spare 4
     character(len=maxlen_DateTime_KVT):: Utc_Sbt_Time ! not used by ADM_Aeolus
     character(len=maxlen_IntAul)   :: Sat_Binary_Time ! not used by ADM_Aeolus
     character(len=maxlen_IntAul)   :: Clock_Step      ! not used by ADM_Aeolus
     ! Spare 5
     character(len=maxlen_DateTime_KVT) :: Leap_Utc
     character(len=maxlen_IntAc)        :: Leap_Sign
     character(len=maxlen_BoolKVT)      :: Leap_Err
     ! Spare 6
     character(len=maxlen_BoolKVT)      :: Product_Err
     character(len=maxlen_IntAd)        :: Tot_Size
     character(len=maxlen_IntAl)        :: Sph_Size
     character(len=maxlen_IntAl)        :: Num_Dsd
     character(len=maxlen_IntAl)        :: Dsd_Size
     character(len=maxlen_IntAl)        :: Num_Data_Sets
     ! spare 7
  END TYPE MPH_type

     !  #]

  ! temporary definitions.
  ! TODO:
  ! these should be moved into EE_CFI_Datatypes.F90 once we agree on using them
  integer(IntAuc), parameter :: missing_indicator_IntAuc = maxval_IntAuc
  integer(IntAuc), parameter :: missing_indicator_IntAc  = maxval_IntAc
  integer(IntAs),  parameter :: missing_indicator_IntAs  = maxval_IntAs
  integer(IntAus), parameter :: missing_indicator_IntAus = maxval_IntAus
  integer(IntAul), parameter :: missing_indicator_IntAul = maxval_IntAul
  real(FAdoxy),    parameter :: missing_indicator_FAdo06 = maxval_FAdo06
  real(FAdoxy),    parameter :: missing_indicator_FAdo46 = maxval_FAdo46
  real(FAdoxy),    parameter :: missing_indicator_FAdo73 = maxval_FAdo73
  !  #]
  !  #[ interfaces for overloading functions
  INTERFACE Write_MPH_to_KVT
     module procedure Write_MPH_to_KVT_unbuff, & ! old direct type
                      Write_MPH_to_KVT_buff,   & ! old buffered type
                      Write_MPH_to_KVT_gen       ! new generic type
  END INTERFACE

  !  #]
  !  #[ Module interface
  ! Type definitions
  public :: MPH_type  
  ! subroutines
  public :: Init_MPH, Print_MPH, Fill_MPH_items, Copy_MPH_to_xml_tree, &
            Fill_L2B_MPH_From_L1B_MPH, Fill_MPH_Sizes
  !  #]
contains
  !---------------------------------------------
  subroutine Init_MPH(MPH,error_flag)
    !  #[

    TYPE(MPH_type), pointer :: MPH
    integer, intent(out) :: error_flag
    
    error_flag = no_error
    IF (.not. associated(MPH)) THEN
       call logmsg(log_error,"ERROR in Init_MPH(): "//&
            "The pointer to this MPH seems not associated !!")
       error_flag = error_programming
       return
    END IF

! WARNING: the pgf90 compiler version 5.2-4 has a bug (or is it?).
! it fails to initialise strings of length 1 when assigning to ""
! Therefore explicitely putting a single space in it is needed.
!
! REMARK: however, I am not sure whether assigning to an empty string
!         is a valid initialisation. Following the strict fortran standard
!         the "" has length zero. And assigning a shorter string to a
!         longer one should not change the extra characters in the longer
!         one. Therefore chances are that nothing happens at all.
!         (although with most compilers initialisation seems to happen
!          anyway when you say x="", this seems a fortran extension?)
!         To force initialisation I will use implicit do loops from now
!         on, like this: x(:)=' ' in stead of x=""

    MPH%Product(:)             = ' '
    MPH%Proc_Stage(:)          = ' '
    MPH%Ref_Doc(:)             = ' '
    ! Spare_1
    MPH%Acquisition_Station(:) = ' '
    MPH%Proc_Center(:)         = ' '
    MPH%Proc_Time              = missing_DateTime_KVT
    MPH%Software_Ver(:)        = ' '
    ! Spare_2
    MPH%Sensing_Start          = missing_DateTime_KVT
    MPH%Sensing_Stop           = missing_DateTime_KVT
    ! Spare_3
    MPH%Phase(:)               = ' '
    MPH%Cycle(:)               = ' '
    MPH%Rel_Orbit(:)           = ' '
    MPH%Abs_Orbit(:)           = ' '
    MPH%State_Vector_Time      = missing_DateTime_KVT
    MPH%Delta_UT1(:)           = ' '
    MPH%X_Position(:)          = ' '
    MPH%Y_Position(:)          = ' '
    MPH%Z_Position(:)          = ' '
    MPH%X_Velocity(:)          = ' ' 
    MPH%Y_Velocity(:)          = ' '
    MPH%Z_Velocity(:)          = ' '
    MPH%Vector_Source(:)       = ' '
    ! Spare_4
    MPH%UTC_SBT_Time           = missing_DateTime_KVT
    MPH%SAT_Binary_Time(:)     = ' '
    MPH%Clock_Step(:)          = ' '
    ! Spare_5
    MPH%Leap_Utc               = missing_DateTime_KVT
    MPH%Leap_Sign (:)          = ' '
    MPH%Leap_Err(:)            = ' '
    ! Spare_6
    MPH%Product_Err(:)         = ' '
    MPH%Tot_Size(:)            = ' '
    MPH%Sph_Size(:)            = ' '
    MPH%Num_Dsd(:)             = ' '
    MPH%Dsd_Size(:)            = ' '
    MPH%Num_Data_Sets(:)       = ' '
    ! Spare_7

!
! Now fill those that need a non-blank numerical string for comparison with KVT
! in the routine Read_MPH_from_KVT.

    ! Spare_3
    MPH%PHASE               = "X"
    MPH%CYCLE               = "+255"
    MPH%REL_ORBIT           = "+32767"
    MPH%ABS_ORBIT           = "+65535"
!    MPH%STATE_VECTOR_TIME   = 
    MPH%DELTA_UT1           = "+.999999"
    MPH%X_POSITION          = "+9999999.999"
    MPH%Y_POSITION          = "+9999999.999"
    MPH%Z_POSITION          = "+9999999.999"
    MPH%X_VELOCITY          = "+9999.999999"
    MPH%Y_VELOCITY          = "+9999.999999"
    MPH%Z_VELOCITY          = "+9999.999999"
!    MPH%VECTOR_SOURCE       = 
    ! Spare 4
!    MPH%UTC_SBT_TIME        = 
    MPH%SAT_BINARY_TIME     = "+4294967295"
    MPH%CLOCK_STEP          = "+4294967295"
    ! Spare 5
!    MPH%LEAP_Utc            = 
    MPH%LEAP_Sign           = "+127"
    MPH%LEAP_ERR            = "0"

  end subroutine Init_MPH

    !  #]
  subroutine Print_MPH(MPH,error_flag,prefix)
    !  #[
    TYPE(MPH_type),             pointer     :: MPH        ! input
    integer,                    intent(out) :: error_flag ! input
    character(len=*), optional, intent(in)  :: prefix

    ! local variable
    character(len=256) :: tmp_prefix

    tmp_prefix = "..."
    IF (present(prefix)) tmp_prefix = prefix

    call AsciiWrite_MPH(MPH,fileunit_stdout,tmp_prefix,error_flag)

  end subroutine Print_MPH
    !  #]
  subroutine AsciiWrite_MPH(MPH,fileunit,prefix,error_flag)
    !  #[
    TYPE(MPH_type),   pointer     :: MPH        ! input
    integer,          intent(in)  :: fileunit   ! input
    character(len=*), intent(in)  :: prefix     ! input
    integer,          intent(out) :: error_flag ! input

    ! local variable
    character(len=256) :: txt

    error_flag = no_error

    IF (.not. associated(MPH)) THEN
       write(fileunit,"(a)",err=999) trim(prefix)//" [seems not associated]"
       return
    END IF

    txt = MPH%Product
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Product             = "//trim(txt)

    txt = MPH%Proc_Stage
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Proc_Stage          = "//trim(txt)

    txt = MPH%Ref_Doc
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Ref_Doc             = "//trim(txt)
    ! Spare_1

    txt = MPH%Acquisition_Station
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Acquisition_Station = "//trim(txt)

    txt = MPH%Proc_Center
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Proc_Center         = "//trim(txt)

    txt = MPH%Proc_Time
    IF (trim(txt) .eq. trim(missing_DateTime_KVT)) THEN
       txt = "[missing]"
    ELSE
       IF (len_trim(txt).eq.0) txt = "[empty]"
    END IF
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Proc_Time           = "//trim(txt)

    txt = MPH%Software_Ver
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Software_Ver        = "//trim(txt)

    ! Spare_2

    txt = MPH%Sensing_Start
    IF (trim(txt) .eq. trim(missing_DateTime_KVT)) THEN
       txt = "[missing]"
    ELSE
       IF (len_trim(txt).eq.0) txt = "[empty]"
    END IF
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Sensing_Start       = "//trim(txt)

    txt = MPH%Sensing_Stop
    IF (trim(txt) .eq. trim(missing_DateTime_KVT)) THEN
       txt = "[missing]"
    ELSE
       IF (len_trim(txt).eq.0) txt = "[empty]"
    END IF
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Sensing_Stop        = "//trim(txt)

    ! Spare_3

    txt = MPH%Phase
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Phase               = "//trim(txt)

    txt = MPH%Cycle
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Cycle               = "//trim(txt)

    txt = MPH%Rel_Orbit
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Rel_Orbit           = "//trim(txt)

    txt = MPH%Abs_Orbit
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Abs_Orbit           = "//trim(txt)

    txt = MPH%State_Vector_Time
    IF (trim(txt) .eq. trim(missing_DateTime_KVT)) THEN
       txt = "[missing]"
    ELSE
       IF (len_trim(txt).eq.0) txt = "[empty]"
    END IF
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%State_Vector_Time   = "//trim(txt)

    txt = MPH%Delta_UT1
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Delta_UT1           = "//trim(txt)

    txt = MPH%X_Position
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%X_Position          = "//trim(txt)

    txt = MPH%Y_Position
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Y_Position          = "//trim(txt)

    txt = MPH%Z_Position
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Z_Position          = "//trim(txt)

    txt = MPH%X_Velocity
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%X_Velocity          = "//trim(txt)

    txt = MPH%Y_Velocity
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Y_Velocity          = "//trim(txt)

    txt = MPH%Z_Velocity
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Z_Velocity          = "//trim(txt)

    txt = MPH%Vector_Source
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Vector_Source       = "//trim(txt)

    ! Spare 4

    txt = MPH%UTC_SBT_Time
    IF (trim(txt) .eq. trim(missing_DateTime_KVT)) THEN
       txt = "[missing]"
    ELSE
       IF (len_trim(txt).eq.0) txt = "[empty]"
    END IF
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%UTC_SBT_Time        = "//trim(txt)

    txt = MPH%Sat_Binary_Time
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Sat_Binary_Time     = "//trim(txt)

    txt = MPH%Clock_Step
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Clock_Step          = "//trim(txt)

    ! Spare 5
    txt = MPH%Leap_Utc
    IF (trim(txt) .eq. trim(missing_DateTime_KVT)) THEN
       txt = "[missing]"
    ELSE
       IF (len_trim(txt).eq.0) txt = "[empty]"
    END IF
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Leap_Utc            = "//trim(txt)

    txt = MPH%Leap_Sign
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Leap_Sign           = "//trim(txt)

    txt = MPH%Leap_Err
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Leap_Err            = "//trim(txt)

    ! Spare 6
    txt = MPH%Product_Err
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Product_Err         = "//trim(txt)

    txt = MPH%Tot_Size
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Tot_Size            = "//trim(txt)

    txt = MPH%Sph_Size
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Sph_Size            = "//trim(txt)

    txt = MPH%Num_Dsd
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Num_Dsd             = "//trim(txt)

    txt = MPH%Dsd_Size
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Dsd_Size            = "//trim(txt)

    txt = MPH%Num_Data_Sets
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Num_Data_Sets       = "//trim(txt)

    ! Spare 7

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_MPH "//&
                          "error while writing to outputfile")
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_MPH
    !  #]
  function GetByteSize_MPH_KVT(workaround) result(KVT_bytesize)
    !  #[

    logical, optional, intent(in) :: workaround   ! optional input
    integer(IntAl)                :: KVT_bytesize ! result

    ! KVT sizes have size(key)+size(=)+size(value)+size(terminator)
    KVT_bytesize = &
         ! len(key)               len(=)  len(value)  len(unit) len(terminator)
         !                                         or len(quotes)
         len("Product")             + 1 + maxlen_Product      + 2    + 1 + &
         len("Proc_Stage")          + 1 + maxlen_Proc_Stage          + 1 + &
         len("Ref_Doc")             + 1 + maxlen_Ref_Doc      + 2    + 1 + &
         40+1+& ! spare_1
         len("Acquisition_Station") + 1 + maxlen_Acq_Stat     + 2    + 1 + &
         len("Proc_Center")         + 1 + maxlen_Proc_Center  + 2    + 1 + &
         len("Proc_Time")           + 1 + maxlen_DateTime_KVT + 2    + 1 + &
         len("Software_Ver")        + 1 + maxlen_Software_Ver + 2    + 1 + &
         40+1+& ! spare_2
         len("Sensing_Start")       + 1 + maxlen_DateTime_KVT + 2    + 1 + &
         len("Sensing_Stop")        + 1 + maxlen_DateTime_KVT + 2    + 1 + &
         40+1+& ! spare_3
         len("Phase")             + 1 + maxlen_Phase                 + 1 + &
         len("Cycle")             + 1 + maxlen_IntAuc                + 1 + &
         len("Rel_Orbit")         + 1 + maxlen_IntAs                 + 1 + &
         len("Abs_Orbit")         + 1 + maxlen_IntAus                + 1 + &
         len("State_Vector_Time") + 1 + maxlen_DateTime_KVT + 2      + 1 + &
         len("Delta_UT1")         + 1 + maxlen_FAdo06 + len("<s>")   + 1 + &
         len("X_Position")        + 1 + maxlen_FAdo73 + len("<m>")   + 1 + &
         len("Y_Position")        + 1 + maxlen_FAdo73 + len("<m>")   + 1 + &
         len("Z_Position")        + 1 + maxlen_FAdo73 + len("<m>")   + 1 + &
         len("X_Velocity")        + 1 + maxlen_FAdo46 + len("<m/s>") + 1 + &
         len("Y_Velocity")        + 1 + maxlen_FAdo46 + len("<m/s>") + 1 + &
         len("Z_Velocity")        + 1 + maxlen_FAdo46 + len("<m/s>") + 1 + &
         len("Vector_Source")     + 1 + maxlen_Vecor_Src    + 2      + 1 + &
         40+1+& ! spare_4
         len("Utc_Sbt_Time")      + 1 + maxlen_DateTime_KVT + 2      + 1 + &
         len("Sat_Binary_Time")   + 1 + maxlen_IntAul                + 1 + &
         len("Clock_Step")        + 1 + maxlen_IntAul + len("<ps>")  + 1 + &
         32+1+& ! spare_5
         len("Leap_Utc")          + 1 + maxlen_DateTime_KVT + 2      + 1 + &
         len("Leap_Sign")         + 1 + maxlen_IntAc                 + 1 + &
         len("Leap_Err")          + 1 + maxlen_BoolKVT               + 1 + &
         40+1+& ! spare_4
         len("Product_Err")    + 1 + maxlen_BoolKVT                  + 1 + &
         len("Tot_Size")       + 1 + maxlen_IntAd   + len("<bytes>") + 1 + &
         len("Sph_Size")       + 1 + maxlen_IntAl   + len("<bytes>") + 1 + &
         len("Num_Dsd")        + 1 + maxlen_IntAl                    + 1 + &
         len("Dsd_Size")       + 1 + maxlen_IntAl   + len("<bytes>") + 1 + &
         len("Num_Data_Sets")  + 1 + maxlen_IntAl                    + 1 + &
         40+1   ! spare_4

    if (present(workaround)) then
       if (workaround) then
          KVT_bytesize = KVT_bytesize + maxlen_Acq_Stat_108 - maxlen_Acq_Stat
       end if
    end if

  end function GetByteSize_MPH_KVT

    !  #]
  subroutine Get_MPH_from_xml_tree(xml_tree, path_in, MPH, error_flag, &
                                   workaround1,workaround2)
    !  #[

    TYPE(xml_node_type), pointer     :: xml_tree    ! input
    character(len=*),    intent(in)  :: path_in     ! input
    TYPE(MPH_type),      pointer     :: MPH         ! result
    integer,             intent(out) :: error_flag  ! output
    logical, optional,   intent(in)  :: workaround1 ! optional input
    logical, optional,   intent(in)  :: workaround2 ! optional input

    ! REMARK:
    ! workaround1 enables the use of 24 characters for the Acq_Stat
    ! field which was (by mistake) used for L1B v.1.08
    ! workaround2 enables the old style xml datetime format
    ! (with only second precision) which was used for L1B v1.07 and earlier

    ! local variables
    character(len=max_str_len)  :: path
    character(len=max_str_len)  :: text
    character(len=max_str_len)  :: path_MPH
    integer(IntAc)  :: value_IntAc
    integer(IntAuc) :: value_IntAuc
    integer(IntAs)  :: value_IntAs
    integer(IntAus) :: value_IntAus
    integer(IntAl)  :: value_IntAl
    integer(IntAul) :: value_IntAul
    integer(IntAd)  :: value_IntAd
    real(FAdoxy)    :: value_FAdoxy
    logical         :: use_workaround1
    logical         :: use_workaround2

    ! init
    error_flag = no_error
    path_MPH = trim(path_in)//"/"//ee_vh_name//"/"//ee_mph_name

    use_workaround1 = .false.
    if (present(workaround1)) use_workaround1 = workaround1

    use_workaround2 = .false.
    if (present(workaround2)) use_workaround2 = workaround2

    IF (.not. associated(MPH)) THEN
       call logmsg(log_error,"ERROR in Get_MPH_from_xml_tree(): "//&
                   "the pointer MPH seems not associated !!!")
       error_flag = error_programming
       return
    END IF

    IF (.not. associated(xml_tree)) THEN
       call logmsg(log_error,"ERROR in Get_MPH_from_xml_tree(): "//&
                   "the pointer xml_tree seems not associated !!!")
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! copying the MPH XML elements
    !------------------------------

    path  =trim(path_MPH)//"/Product"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Product = check_length_xml_item(text,path,maxlen_Product,&
         "Get_MPH_from_xml_tree")

    path  =trim(path_MPH)//"/Proc_Stage"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Proc_Stage = check_length_xml_item(text,path,maxlen_Proc_Stage,&
         "Get_MPH_from_xml_tree")

    path  =trim(path_MPH)//"/Ref_Doc"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Ref_Doc = check_length_xml_item(text,path,maxlen_Ref_Doc,&
         "Get_MPH_from_xml_tree")

    ! Spare 1

    path  =trim(path_MPH)//"/Acquisition_Station"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return

    if (use_workaround1) then
       MPH%Acquisition_Station = &
            check_length_xml_item(text,path,maxlen_Acq_Stat_108,&
                                  "Get_MPH_from_xml_tree")
    else
       MPH%Acquisition_Station = &
            check_length_xml_item(text,path,maxlen_Acq_Stat,&
                                  "Get_MPH_from_xml_tree")
    end if

    path  =trim(path_MPH)//"/Proc_Center"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Proc_Center = check_length_xml_item(text,path,maxlen_Proc_Center,&
         "Get_MPH_from_xml_tree")

    path  =trim(path_MPH)//"/Proc_Time"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,maxlen_DateTime_XML,&
         "Get_MPH_from_xml_tree")
    MPH%Proc_Time = Convert_XML_to_KVT_DateTime(text,use_workaround2)

    path  =trim(path_MPH)//"/Software_Ver"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Software_Ver = check_length_xml_item(text,path,maxlen_Software_Ver,&
         "Get_MPH_from_xml_tree")

    ! Spare 2

    path  =trim(path_MPH)//"/Sensing_Start"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,maxlen_DateTime_XML,&
         "Get_MPH_from_xml_tree")
    MPH%Sensing_Start = Convert_XML_to_KVT_DateTime(text,use_workaround2)

    path  =trim(path_MPH)//"/Sensing_Stop"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,maxlen_DateTime_XML,&
         "Get_MPH_from_xml_tree")
    MPH%Sensing_Stop = Convert_XML_to_KVT_DateTime(text,use_workaround2)

    ! Spare 3

    path  =trim(path_MPH)//"/Phase"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Phase = check_length_xml_item(text,path,maxlen_Phase,&
         "Get_MPH_from_xml_tree")

    path  =trim(path_MPH)//"/Cycle"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Cycle = check_length_xml_item(text,path,maxlen_IntAuc,&
         "Get_MPH_from_xml_tree")
    ! try to convert the number, to check its validity
    call GetIntAuc(MPH%Cycle,value_IntAuc,error_flag)
    IF (error_flag .ne. no_error) return

    path  =trim(path_MPH)//"/Rel_Orbit"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Rel_Orbit = check_length_xml_item(text,path,maxlen_IntAs,&
         "Get_MPH_from_xml_tree")
    ! try to convert the number, to check its validity
    call GetIntAs(MPH%Rel_Orbit,value_IntAs,error_flag)
    IF (error_flag .ne. no_error) return

    path  =trim(path_MPH)//"/Abs_Orbit"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Abs_Orbit = check_length_xml_item(text,path,maxlen_IntAus,&
         "Get_MPH_from_xml_tree")
    ! try to convert the number, to check its validity
    call GetIntAus(MPH%Abs_Orbit,value_IntAus,error_flag)
    IF (error_flag .ne. no_error) return

    path  =trim(path_MPH)//"/State_Vector_Time"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,maxlen_DateTime_XML,&
         "Get_MPH_from_xml_tree")
    MPH%State_Vector_Time = Convert_XML_to_KVT_DateTime(text,use_workaround2)

    path  =trim(path_MPH)//"/Delta_UT1"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Delta_UT1 = check_length_xml_item(text,path,maxlen_FAdo06,&
         "Get_MPH_from_xml_tree")
    ! try to convert the number, to check its validity
    call GetFAdoxy(MPH%Delta_UT1,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return

    path  =trim(path_MPH)//"/X_Position"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%X_Position = check_length_xml_item(text,path,maxlen_FAdo73,&
         "Get_MPH_from_xml_tree")
    ! try to convert the number, to check its validity
    call GetFAdoxy(MPH%X_Position,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return

    path  =trim(path_MPH)//"/Y_Position"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Y_Position = check_length_xml_item(text,path,maxlen_FAdo73,&
         "Get_MPH_from_xml_tree")
    ! try to convert the number, to check its validity
    call GetFAdoxy(MPH%Y_Position,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return

    path  =trim(path_MPH)//"/Z_Position"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Z_Position = check_length_xml_item(text,path,maxlen_FAdo73,&
         "Get_MPH_from_xml_tree")
    ! try to convert the number, to check its validity
    call GetFAdoxy(MPH%Z_Position,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return

    path  =trim(path_MPH)//"/X_Velocity"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%X_Velocity = check_length_xml_item(text,path,maxlen_FAdo46,&
         "Get_MPH_from_xml_tree")
    ! try to convert the number, to check its validity
    call GetFAdoxy(MPH%X_Velocity,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return

    path  =trim(path_MPH)//"/Y_Velocity"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Y_Velocity = check_length_xml_item(text,path,maxlen_FAdo46,&
         "Get_MPH_from_xml_tree")
    ! try to convert the number, to check its validity
    call GetFAdoxy(MPH%Y_Velocity,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return

    path  =trim(path_MPH)//"/Z_Velocity"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Z_Velocity = check_length_xml_item(text,path,maxlen_FAdo46,&
         "Get_MPH_from_xml_tree")
    ! try to convert the number, to check its validity
    call GetFAdoxy(MPH%Z_Velocity,value_FAdoxy,error_flag)
    IF (error_flag .ne. no_error) return

    path  =trim(path_MPH)//"/Vector_Source"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Vector_Source = check_length_xml_item(text,path,maxlen_Vecor_Src,&
         "Get_MPH_from_xml_tree")

    ! Spare 4

    path  =trim(path_MPH)//"/Utc_Sbt_Time" ! not used by ADM_Aeolus
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,maxlen_DateTime_XML,&
         "Get_MPH_from_xml_tree")
    MPH%Utc_Sbt_Time = Convert_XML_to_KVT_DateTime(text,use_workaround2)

    path  =trim(path_MPH)//"/Sat_Binary_Time" ! not used by ADM_Aeolus
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Sat_Binary_Time = check_length_xml_item(text,path,maxlen_IntAul,&
         "Get_MPH_from_xml_tree")
    ! try to convert the number, to check its validity
    call GetIntAul(MPH%Sat_Binary_Time,value_IntAul,error_flag)
    IF (error_flag .ne. no_error) return

    path  =trim(path_MPH)//"/Clock_Step" ! not used by ADM_Aeolus
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Clock_Step = check_length_xml_item(text,path,maxlen_IntAul,&
         "Get_MPH_from_xml_tree")
    ! try to convert the number, to check its validity
    call GetIntAul(MPH%Clock_Step,value_IntAul,error_flag)
    IF (error_flag .ne. no_error) return

    ! Spare 5

    path  =trim(path_MPH)//"/Leap_Utc"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,path,maxlen_DateTime_XML,&
         "Get_MPH_from_xml_tree")
    MPH%Leap_Utc = Convert_XML_to_KVT_DateTime(text,use_workaround2)

    path  =trim(path_MPH)//"/Leap_Sign"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Leap_Sign = check_length_xml_item(text,path,maxlen_IntAc,&
         "Get_MPH_from_xml_tree")
    ! try to convert the number, to check its validity
    call GetIntAc(MPH%Leap_Sign,value_IntAc,error_flag)
    IF (error_flag .ne. no_error) return

    path  =trim(path_MPH)//"/Leap_Err"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return

    text = check_length_xml_item(text,path,maxlen_BoolXML,&
         "Get_MPH_from_xml_tree")

    ! convert from XML to KVT 
    select case (trim(to_lowercase(text)))
    case("")
       ! also allow missing for now
       text = ""
    case("false")
       MPH%Leap_Err = "0"
    case("true")
       MPH%Leap_Err = "1"
    case default
       call logmsg(log_error,"ERROR in Get_MPH_from_xml_tree(): "//&
                   "Incorrect value for XML field Leap_Err: ["//trim(text)//"] "//&
                   "possible values are 'true' or 'false'")
       error_flag = error_reading_file
       return
    end select

    ! Spare 6

    path  =trim(path_MPH)//"/Product_Err"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return

    text = check_length_xml_item(text,path,maxlen_BoolXML,&
         "Get_MPH_from_xml_tree")

    ! convert from XML to KVT 
    select case (trim(to_lowercase(text)))
    case("")
       ! also allow missing for now
       text = ""
    case("false")
       MPH%Product_Err = "0"
    case("true")
       MPH%Product_Err = "1"
    case default
       call logmsg(log_error,"ERROR in Get_MPH_from_xml_tree(): "//&
                   "Incorrect value for XML field Product_Err: ["//trim(text)//"] "//&
                   "possible values are 'true' or 'false'")
       error_flag = error_reading_file
       return
    end select

    path  =trim(path_MPH)//"/Tot_Size"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Tot_Size = check_length_xml_item(text,path,maxlen_IntAd,&
         "Get_MPH_from_xml_tree")
    ! try to convert the number, to check its validity
    call GetIntAd(MPH%Tot_Size,value_IntAd,error_flag)
    IF (error_flag .ne. no_error) return

    path  =trim(path_MPH)//"/Sph_Size"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Sph_Size = check_length_xml_item(text,path,maxlen_IntAl,&
         "Get_MPH_from_xml_tree")
    ! try to convert the number, to check its validity
    call GetIntAl(MPH%Sph_Size,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return

    path  =trim(path_MPH)//"/Num_Dsd"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Num_Dsd = check_length_xml_item(text,path,maxlen_IntAl,&
         "Get_MPH_from_xml_tree")
    ! try to convert the number, to check its validity
    call GetIntAl(MPH%Num_Dsd,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return

    path  =trim(path_MPH)//"/Dsd_Size"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Dsd_Size = check_length_xml_item(text,path,maxlen_IntAl,&
         "Get_MPH_from_xml_tree")
    ! try to convert the number, to check its validity
    call GetIntAl(MPH%Dsd_Size,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return

    path  =trim(path_MPH)//"/Num_Data_Sets"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Num_Data_Sets = check_length_xml_item(text,path,maxlen_IntAl,&
         "Get_MPH_from_xml_tree")
    ! try to convert the number, to check its validity
    call GetIntAl(MPH%Num_Data_Sets,value_IntAl,error_flag)
    IF (error_flag .ne. no_error) return

    ! Spare 7

    return

  end subroutine Get_MPH_from_xml_tree

    !  #]
  subroutine Fill_MPH_items(MPH, product, proc_stage, ref_doc, proc_center, &
                            software_version, product_err, tot_size, &
                            sph_size, num_DSD, num_data_sets, &
                            DateTimeStart, DateTimeStop, error_flag)
     !  #[
    TYPE(MPH_type),   pointer     :: MPH ! structure to be filled
    character(len=*), intent(in)  :: product    ! actually this is the filename
    character(len=1), intent(in)  :: proc_stage ! possible values: N/R/T
    character(len=*), intent(in)  :: ref_doc
    character(len=*), intent(in)  :: proc_center
    character(len=*), intent(in)  :: software_version
    logical,          intent(in)  :: product_err
    integer(IntAd),   intent(in)  :: tot_size
    integer(IntAl),   intent(in)  :: sph_size
    integer(IntAl),   intent(in)  :: num_DSD
    integer(IntAl),   intent(in)  :: num_data_sets
    TYPE(DateTimeType), intent(in)  :: DateTimeStart ! struct(YYYY,MM,DD)
    TYPE(DateTimeType), intent(in)  :: DateTimeStop  ! struct(YYYY,MM,DD)
    integer,          intent(out) :: error_flag

    ! local variables
    character(len=maxlen_BoolKVT)      :: text_BoolKVT
    character(len=maxlen_IntAl)        :: text_IntAl
    character(len=maxlen_IntAd)        :: text_IntAd
    character(len=maxlen_DateTime_KVT) :: DateTime_KVT
    integer(IntAl)                     :: Dsd_Size

    error_flag = no_error
    IF (.not. associated(MPH)) THEN
       call logmsg(log_error,"ERROR in Fill_MPH_items(): "//&
                   "The pointer to this MPH seems not associated !!")
       error_flag = error_programming
       return
    END IF

    ! this is the filename
    MPH%Product             = check_length_xml_item(product,"Product",&
                                 maxlen_Product,"Fill_MPH_items")

    ! this may be N/R/T (near-real-time, reprocess or test)
    MPH%Proc_Stage          = check_length_xml_item(proc_stage,"Proc_Stage",&
                                 maxlen_Proc_Stage,"Fill_MPH_items")

    ! for L2B this reference is imported from the AuxPar default MPH fields
    MPH%Ref_Doc             = check_length_xml_item(ref_doc,"Ref_Doc",&
                                 maxlen_Ref_Doc,"Fill_MPH_items")

    ! Spare 1

    ! this will be copied from L1B by the subroutine
    ! Fill_L2B_MPH_From_L1B_MPH() below, and is not usefull
    ! for aux files, so leave it empty for now
    !MPH%ACQUISITION_STATION = ...

    ! the location where this file was produced
    MPH%Proc_Center         = check_length_xml_item(proc_center,"Proc_Center",&
                                 maxlen_Proc_Center,"Fill_MPH_items")

    ! the current time at which this file was produced
    DateTime_KVT            = Construct_KVT_DateTime(GetCurrentDateTime())

    ! the check for this item is a bit weird, because the allowed length
    ! is different in the XML and the KVT part.
    ! Since we use the KVT part for storage (which has higher accuracy)
    ! we must check the length as expected for the KVT DateTimes, but
    ! we use the xml check routine to do this for now .....
    MPH%Proc_Time           = check_length_xml_item(DateTime_KVT, &
                         "Proc_Time",maxlen_DateTime_KVT,"Fill_MPH_items")

    MPH%Software_Ver         = check_length_xml_item(software_version,&
                         "Software_ver",maxlen_Software_Ver,"Fill_MPH_items")

    ! Spare_2

    ! test the validity if the provided datetimes
    call TestDateTime(DateTimeStart,error_flag)
    IF (error_flag .ne. no_error) return
    call TestDateTime(DateTimeStop,error_flag)
    IF (error_flag .ne. no_error) return

    MPH%SENSING_START       = Construct_KVT_DateTime(DateTimeStart)
    MPH%SENSING_STOP        = Construct_KVT_DateTime(DateTimeStop)

    ! Spare_3

    ! all following items will be copied from L1B by the subroutine
    ! Fill_L2B_MPH_From_L1B_MPH() below, and are not usefull
    ! for aux files, so leave them empty for now

!    MPH%PHASE               = 
!    MPH%CYCLE               = 
!    MPH%REL_ORBIT           = 
!    MPH%ABS_ORBIT           = 
!    MPH%STATE_VECTOR_TIME   = 
!    MPH%DELTA_UT1           = 
!    MPH%X_POSITION          = 
!    MPH%Y_POSITION          = 
!    MPH%Z_POSITION          = 
!    MPH%X_VELOCITY          = 
!    MPH%Y_VELOCITY          = 
!    MPH%Z_VELOCITY          = 
!    MPH%VECTOR_SOURCE       = 
    ! Spare 4
!    MPH%UTC_SBT_TIME        = 
!    MPH%SAT_BINARY_TIME     = 
!    MPH%CLOCK_STEP          = 
    ! Spare 5
!    MPH%LEAP_Utc            = 
!    MPH%LEAP_Sign           = 
!    MPH%LEAP_ERR            = 

    ! Spare 6

    ! these next items can only be filled when all data sets 
    ! are available, the calling program should take care of this
    call SetBoolKVT(product_err,text_BoolKVT)
    IF (error_flag .ne. no_error) return
    MPH%PRODUCT_ERR         = text_BoolKVT

    call SetIntAd(tot_size,text_IntAd,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Tot_Size            = text_IntAd

    ! these are constant for a given filetype, but since this is a generic
    ! piece of code for all filetypes, we need these to be inputs
    call SetIntAl(sph_size,text_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Sph_Size            = text_IntAl

    call SetIntAl(num_DSD,text_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%NUM_Dsd             = text_IntAl

    ! this is a constant. See TN2.1, sec. 4.6, table 11.
    Dsd_Size = GetByteSize_DSD_KVT() ! Dsd_Size = 288_IntAl
    call SetIntAl(Dsd_Size,text_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%Dsd_SIZE            = text_IntAl

    ! this is constant for a given filetype, but since this is a generic
    ! piece of code for all filetypes, we need it to be an input
    call SetIntAl(num_data_sets,text_IntAl,error_flag)
    IF (error_flag .ne. no_error) return
    MPH%NUM_DATA_SETS       = text_IntAl

    ! Spare 7

  end subroutine Fill_MPH_items
    !  #]
  subroutine Copy_MPH_to_xml_tree(MPH, path_in, xml_tree, error_flag)
    !  #[

    !  Modifications
    !   01-12-2005  P. Poli    Added a few calls to "Remove_Plus_Sign"

    TYPE(MPH_type),      pointer    :: MPH        ! input
    character(len=*),    intent(in) :: path_in    ! input
    TYPE(xml_node_type), pointer    :: xml_tree   ! points to updated struct
    integer,             intent(out):: error_flag ! result

    ! local variable
    character(len=max_str_len)  :: text
    character(len=max_str_len)  :: path

    error_flag = no_error

    ! add the <MPH></MPH> tags
    path = trim(path_in)//"/"//ee_vh_name
    call xml_tree_add_child(xml_tree,path,ee_mph_name,error_flag)
    IF (error_flag .ne. no_error) return

    ! fill all items inside the MainProductHeader
    call add_MPH_item(xml_tree, path_in, "Product",    MPH%Product,    error_flag )
    IF (error_flag .ne. no_error) return
    call add_MPH_item(xml_tree, path_in, "Proc_Stage", MPH%Proc_Stage, error_flag )
    IF (error_flag .ne. no_error) return
    call add_MPH_item(xml_tree, path_in, "Ref_Doc",    MPH%Ref_Doc,    error_flag )
    IF (error_flag .ne. no_error) return

    ! Spare 1
    call add_MPH_item(xml_tree,path_in, "Spare_1",newline,error_flag)
    IF (error_flag .ne. no_error) return

    call add_MPH_item(xml_tree, path_in, "Acquisition_Station", MPH%Acquisition_Station, error_flag )
    IF (error_flag .ne. no_error) return
    call add_MPH_item(xml_tree, path_in, "Proc_Center",         MPH%Proc_Center,         error_flag )
    IF (error_flag .ne. no_error) return
    ! convert DateTime from KVT to XML format
    text = Convert_KVT_to_XML_DateTime(MPH%Proc_Time)
    call add_MPH_item(xml_tree, path_in, "Proc_Time",           text,                    error_flag )
    IF (error_flag .ne. no_error) return
    call add_MPH_item(xml_tree, path_in, "Software_Ver",        MPH%Software_Ver,        error_flag )
    IF (error_flag .ne. no_error) return

    ! Spare 2
    call add_MPH_item(xml_tree, path_in, "Spare_2", newline, error_flag)
    IF (error_flag .ne. no_error) return

    ! convert DateTime from KVT to XML format
    text = Convert_KVT_to_XML_DateTime(MPH%Sensing_Start)
    call add_MPH_item(xml_tree, path_in, "Sensing_Start",       text,                    error_flag )
    IF (error_flag .ne. no_error) return
    ! convert DateTime from KVT to XML format
    text = Convert_KVT_to_XML_DateTime(MPH%Sensing_Stop)
    call add_MPH_item(xml_tree, path_in, "Sensing_Stop",        text,                    error_flag )
    IF (error_flag .ne. no_error) return

    ! Spare 3
    call add_MPH_item(xml_tree, path_in, "Spare_3",newline,error_flag)
    IF (error_flag .ne. no_error) return

    call add_MPH_item(xml_tree, path_in, "Phase",               MPH%Phase,               error_flag )
    IF (error_flag .ne. no_error) return
    call add_MPH_item(xml_tree, path_in, "Cycle",               Remove_Plus_Sign(MPH%Cycle),      error_flag )
    IF (error_flag .ne. no_error) return
    call add_MPH_item(xml_tree, path_in, "Rel_Orbit",           Remove_Plus_Sign(MPH%Rel_Orbit),  error_flag )
    IF (error_flag .ne. no_error) return
    call add_MPH_item(xml_tree, path_in, "Abs_Orbit",           Remove_Plus_Sign(MPH%Abs_Orbit),  error_flag )
    IF (error_flag .ne. no_error) return
    ! convert DateTime from KVT to XML format
    text = Convert_KVT_to_XML_DateTime(MPH%State_Vector_Time)
    call add_MPH_item(xml_tree, path_in, "State_Vector_Time",   text,                             error_flag )
    IF (error_flag .ne. no_error) return
    call add_MPH_item(xml_tree, path_in, "Delta_UT1",           Remove_Plus_Sign(MPH%Delta_UT1),  error_flag )
    IF (error_flag .ne. no_error) return
    call add_MPH_item(xml_tree, path_in, "X_Position",          Remove_Plus_Sign(MPH%X_Position), error_flag )
    IF (error_flag .ne. no_error) return
    call add_MPH_item(xml_tree, path_in, "Y_Position",          Remove_Plus_Sign(MPH%Y_Position), error_flag )
    IF (error_flag .ne. no_error) return
    call add_MPH_item(xml_tree, path_in, "Z_Position",          Remove_Plus_Sign(MPH%Z_Position), error_flag )
    IF (error_flag .ne. no_error) return
    call add_MPH_item(xml_tree, path_in, "X_Velocity",          Remove_Plus_Sign(MPH%X_Velocity), error_flag )
    IF (error_flag .ne. no_error) return
    call add_MPH_item(xml_tree, path_in, "Y_Velocity",          Remove_Plus_Sign(MPH%Y_Velocity), error_flag )
    IF (error_flag .ne. no_error) return
    call add_MPH_item(xml_tree, path_in, "Z_Velocity",          Remove_Plus_Sign(MPH%Z_Velocity), error_flag )
    IF (error_flag .ne. no_error) return
    call add_MPH_item(xml_tree, path_in, "Vector_Source",       MPH%Vector_Source,                error_flag )
    IF (error_flag .ne. no_error) return

    ! Spare 4
    call add_MPH_item(xml_tree, path_in, "Spare_4",newline,error_flag)
    IF (error_flag .ne. no_error) return

    ! convert DateTime from KVT to XML format
    text = Convert_KVT_to_XML_DateTime(MPH%Utc_Sbt_Time)
    call add_MPH_item(xml_tree, path_in, "Utc_Sbt_Time",        text,                                  error_flag )
    IF (error_flag .ne. no_error) return 
    call add_MPH_item(xml_tree, path_in, "Sat_Binary_Time",     Remove_Plus_Sign(MPH%Sat_Binary_Time), error_flag )
    IF (error_flag .ne. no_error) return
    call add_MPH_item(xml_tree, path_in, "Clock_Step",          Remove_Plus_Sign(MPH%Clock_Step),      error_flag )
    IF (error_flag .ne. no_error) return

    ! Spare 5
    call add_MPH_item(xml_tree, path_in, "Spare_5",newline,error_flag)
    IF (error_flag .ne. no_error) return

    ! convert DateTime from KVT to XML format
    text = Convert_KVT_to_XML_DateTime(MPH%Leap_Utc)
    call add_MPH_item(xml_tree, path_in, "Leap_Utc",            text,                            error_flag )
    IF (error_flag .ne. no_error) return
    call add_MPH_item(xml_tree, path_in, "Leap_Sign",           Remove_Plus_Sign(MPH%Leap_Sign), error_flag )
    IF (error_flag .ne. no_error) return

    ! convert KVT to XML format for this ENUM error type
    ! see MDA IODD v2.2 (file 521666_IODD_2_2.pdf), sec.3.3.1, table 3-7
    select case(trim(MPH%Leap_Err))
    case("")
       ! also allow missing for now
       text = ""
    case("0")
       text = "False"
    case("1")
       text = "True"
    case default
       call logmsg(log_error,"ERROR in Copy_MPH_to_xml_tree(): "//&
                   "Incorrect value for KVT field: Leap_Err = "//MPH%Leap_Err//&
                   " possible values are '0' or '1'")
       error_flag = error_programming
       return
    end select

    call add_MPH_item(xml_tree, path_in, "Leap_Err",            text, error_flag )
    IF (error_flag .ne. no_error) return

    ! Spare 6
    call add_MPH_item(xml_tree,path_in, "Spare_6",newline,error_flag)
    IF (error_flag .ne. no_error) return

    ! convert KVT to XML format for this ENUM error type
    ! see MDA IODD v2.2 (file 521666_IODD_2_2.pdf), sec.3.3.1, table 3-7
    ! This is needed since we store the MPH in KVT formatted strings
    ! which is different from the XML formatted strings !!!
    select case(trim(MPH%Product_Err))
    case("")
       ! also allow missing for now
       text = ""
    case("0")
       text = "False"
    case("1")
       text = "True"
    case default
       call logmsg(log_error,"ERROR in Copy_MPH_to_xml_tree(): "//&
                   "Incorrect value for KVT field: Product_Err = "//MPH%Product_Err//&
                   " possible values are '0' or '1'")
       error_flag = error_programming
       return
    end select

    call add_MPH_item(xml_tree, path_in, "Product_Err",         text, error_flag )
    IF (error_flag .ne. no_error) return

    call add_MPH_item(xml_tree, path_in, "Tot_Size",            Remove_Plus_Sign(MPH%Tot_Size),      error_flag )
    IF (error_flag .ne. no_error) return
    call add_MPH_item(xml_tree, path_in, "Sph_Size",            Remove_Plus_Sign(MPH%Sph_Size),      error_flag )
    IF (error_flag .ne. no_error) return
    call add_MPH_item(xml_tree, path_in, "Num_Dsd",             Remove_Plus_Sign(MPH%Num_Dsd),       error_flag )
    IF (error_flag .ne. no_error) return
    call add_MPH_item(xml_tree, path_in, "Dsd_Size",            Remove_Plus_Sign(MPH%Dsd_Size),      error_flag )
    IF (error_flag .ne. no_error) return
    call add_MPH_item(xml_tree, path_in, "Num_Data_Sets",       Remove_Plus_Sign(MPH%Num_Data_Sets), error_flag )
    IF (error_flag .ne. no_error) return

    ! Spare 7
    call add_MPH_item(xml_tree,path_in, "Spare_7",newline,error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine Copy_MPH_to_xml_tree

  !  #]
  !---------------------------------------------
  subroutine GetFileVersion(filename_hdr, Software_Version_String, &
                            Software_Version_Major_Number, &
                            Software_Version_Minor_Number, &
                            error_flag)
    !  #[ 

    ! retrieves the Software_Ver field for several definitions that exist
    ! for the header file (tested on L1B v.1.01, 1.04,1.05,1.06, 1.07 
    ! 1.08 and L2B)
    ! Note that this routine should be called before entering the actual
    ! file reading (it does its own file reading!)
    ! Its results may be used to switch on/off workarounds needed
    ! to read a particular version of a file.
    character(len=*), intent(in)  :: filename_hdr
    character(len=maxlen_Software_Ver), &
                      intent(out) :: Software_Version_String
    integer,          intent(out) :: Software_Version_Major_Number
    integer,          intent(out) :: Software_Version_Minor_Number
    integer,          intent(out) :: error_flag

    ! local variables
    type(xml_node_type), pointer :: xml_tree
    character(len=max_str_len)   :: path
    character(len=max_str_len)   :: text
    logical                      :: exists
    integer                      :: i
    character(len=2)             :: MajorNumberText
    character(len=2)             :: MinorNumberText

    integer, parameter :: npaths = 3
    character(len=max_str_len), dimension(npaths) :: possible_header_path

    ! a number of different header file structures have been detected:
    ! L1B,v1.01 upto v1.03 use: /Header_File
    ! L1B,v1.04 and higher use: /Earth_Explorer_Header/Variable_Header
    ! L2B upto v1.30 uses: 
    ! /Earth_Explorer_File/Earth_Explorer_Header/Variable_Header
    ! both for *.HDR and *.EEF files.
    ! L2B upto v1.31 and higher uses: /Earth_Explorer_Header/Variable_Header
    ! for *.HDR files and 
    ! /Earth_Explorer_File/Earth_Explorer_Header/Variable_Header
    ! for *.EEF files

    data (possible_header_path(i),i=1,npaths) / &
         "/Earth_Explorer_File/Earth_Explorer_Header/Variable_Header",&
         "/Earth_Explorer_Header/Variable_Header",&
         "/Header_File" /

    ! init
    error_flag = no_error
    Software_Version_String(:) = ' '

    ! loading xml file
    nullify(xml_tree)
    call xml_load_file(filename_hdr, xml_tree, error_flag)
    IF (error_flag .ne. no_error) return

    ! remark:
    ! in this case checking the xmlns attribute is not added, 
    ! because this GetFileVersion should be compatible with all
    ! xml files, also older types, and because this checking is done
    ! already when the full xml file is parsed later.

    path_check: do i=1,npaths
       ! check whether the header tag pair is present
       path = possible_header_path(i)
       ! testing this path
       call xml_check_node_exists(xml_tree, path, exists, error_flag)
       if (error_flag .ne. no_error) return
       IF (exists) exit path_check
    end do path_check
    
    IF (.not. exists) THEN
       error_flag = error_reading_file
       call logmsg(log_error,"Error in GetFileVersion: "//&
                   "headerfile has an unrecognised structure; "//&
                   "could not find Software_Ver element ...")
       return
    END IF

    ! retrieve the MPH/FileVersion element from the header
    path = trim(path)//"/Main_Product_Header/Software_Ver"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) then
       call logmsg(log_error,"ERROR in GetFileVersion: "//&
                   "could not a tag named "//&
                   trim(path)//"/Main_Product_Header/Software_Ver")
       return
    END IF
    
    Software_Version_String = text(1:maxlen_Software_Ver)

    ! possible results are:
    ! "ADM_L1bP/1.04"
    ! "ADM_L1bP/1.05"
    ! "ADM_L1bP/1.06"
    ! "ADM_L1bP/1.07"
    ! "ADM_L1bP/1.08"
    ! etc. for L1BP product files
    ! 
    ! and:
    ! 
    ! "01.10"
    ! "01.20"
    ! "01.30"
    ! "01.31"
    ! etc. for L2BP product files

    ! retrieve the numerical values of of the version number
    if (Software_Version_String(1:9) .eq. "ADM_L1bP/") then
       ! handle the L1B case
       MajorNumberText(1:1) = '0'
       MajorNumberText(2:2) = Software_Version_String(10:10)
       MinorNumberText = Software_Version_String(12:13)
    else
       ! handle the L2B case (not yet used at the moment)
       MajorNumberText = Software_Version_String(1:2)
       MinorNumberText = Software_Version_String(4:5)
    end if
    read(MajorNumberText,"(i2)",err=998) Software_Version_Major_Number
    read(MinorNumberText,"(i2)",err=999) Software_Version_Minor_Number

    call delete_xml_tree(xml_tree, error_flag)
    IF (error_flag .ne. no_error) return

    return

998 call logmsg(log_error,"ERROR in GetFileVersion: could not retrieve "//&
         "Software_Version_Major_Number from Software_Verstion_String: "//&
         trim(Software_Version_String))
    error_flag = error_reading_file
    return

999 call logmsg(log_error,"ERROR in GetFileVersion: could not retrieve "//&
         "Software_Version_Minor_Number from Software_Verstion_String: "//&
         trim(Software_Version_String))
    error_flag = error_reading_file
    return

  end subroutine GetFileVersion

    !  #]
  !---------------------------------------------
  subroutine Read_MPH_from_KVT(MPH, fileunit, error_flag, CompareOnly, &
                               Compare, workaround1, workaround2)
    !  #[ 

    ! read the main product header (KVT values)

    ! implemented workarounds:
    ! ==>workaround1
    ! allow the test (actual_product_size .ne. total_product_size) to fail
    ! This is needed to be able to use L1B v1.05/v1.06 files
    !
    ! ==>workaround2
    ! allow the <Acquisition_Station> field to have 24 characters
    ! in stead of 20 characters
    ! This is an IODD change occurred in L1B v1.08, but was a mistake
    ! and has been reverted in L1B v1.09

    ! Note that I cannot number the workarounds 105/106/108 here because
    ! this is common MPH code for all files, and these numbers refer
    ! to the L1B IODD only.
    
    ! if CompareOnly=.true. then
    ! assume the values are already read from the XML file
    ! so read and compare them but don't fill them again
    ! Modification: P. Poli: except for dates/times which are replaced
    ! because they are more accurate in KVT than in XML (29-11-2005).
    ! Modification: P. Poli: added flag Compare. Compares values read 
    ! with those in memory, and then replaces the ones in memory by
    ! those read, because those read are in KVT format, which is the
    ! convention of choice.

    TYPE(MPH_type),    pointer     :: MPH
    integer,           intent(in)  :: fileunit
    integer,           intent(out) :: error_flag
    logical, optional, intent(in)  :: CompareOnly
    logical, optional, intent(in)  :: Compare
    logical, optional, intent(in)  :: workaround1
    logical, optional, intent(in)  :: workaround2

    ! local variables
    character(len=max_str_len) :: exp_name
    character(len=max_str_len) :: value
    character(len=max_str_len) :: exp_phys_unit
    integer                    :: exp_val_length, pos, ios
    integer(i8_)               :: total_product_size, actual_product_size
    logical                    :: do_CompareOnly
    logical                    :: do_Compare
    character(len=256)         :: actual_filename
    logical                    :: nmd
    logical                    :: use_workaround1
    logical                    :: use_workaround2
    character(len=20)          :: actual_product_size_txt
    character(len=20)          :: total_product_size_txt

    !------------------------------
    ! reading the MPH KVT elements
    !------------------------------

    ! init
    error_flag = no_error

    use_workaround1 = .false.
    IF (present(workaround1)) use_workaround1 = workaround1

    use_workaround2 = .false.
    IF (present(workaround2)) use_workaround2 = workaround2

    IF (.not. associated(MPH)) THEN
       call logmsg(log_error,"ERROR in Read_MPH_KVT(): "//&
                   "The pointer to this MPH seems not associated !!")
       error_flag = error_programming
       return
    END IF

    do_CompareOnly = .false.
    IF (present(CompareOnly)) do_CompareOnly = CompareOnly
    do_Compare = .false.
    IF (present(Compare)) do_Compare = Compare


    IF (do_CompareOnly) THEN
       print *,"Read_MPH_from_KVT: Reading MPH-KVT items from file and comparing with XML result"
    ELSE
       print *,"Read_MPH_from_KVT: Reading MPH-KVT items from file and filling MPH structure"
    END IF

    IF (do_Compare) THEN
       print *,"Read_MPH_from_KVT: Comparing MPH-KVT items with XML items"
    END IF

    ! read the product name
    exp_val_length = len(MPH%Product)
    exp_name       = "PRODUCT"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expect_quotes=.true.)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Product, value, &
         "MPH%Product", do_CompareOnly, do_Compare, error_flag)
    IF (error_flag .ne. no_error) return

    ! read the processing stage flag
    exp_val_length = len(MPH%Proc_Stage)
    exp_name       = "PROC_STAGE"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length)
    IF (error_flag .ne. no_error) return

    ! check the result
    pos = verify(trim(value),"NRBTS")
    IF (pos .ne. 0) THEN
       call logmsg(log_error,"ERROR in Read_MPH_KVT(): "//&
                   "item: PROC_STAGE is only allowed to have the values "//&
                   "N, S, R, B or T (Normal or Standard processing, Reprocessing, "//&
                   "Backlog-processing, or Test) "//&
                   "but is seems to contain: "//trim(value))
       error_flag = error_reading_file
       return
    END IF

    call Fill_or_Compare_MPH_item(MPH%Proc_Stage, value, &
         "MPH%Proc_Stage", do_CompareOnly, do_Compare, error_flag)
    IF (error_flag .ne. no_error) return

    ! read the reference document name
    exp_val_length = len(MPH%Ref_Doc)
    exp_name       = "REF_DOC"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expect_quotes=.true.)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Ref_Doc, value, &
         "MPH%Ref_Doc", do_CompareOnly, do_Compare, error_flag)
    IF (error_flag .ne. no_error) return

    ! Spare_1: this is a spare value without content
    exp_val_length = 40
    call read_spare_kvt_string(fileunit,exp_val_length, error_flag)
    IF (error_flag .ne. no_error) return

    ! read the acquisition station ID
    if (use_workaround2) then
       exp_val_length = maxlen_Acq_Stat_108
    else
       exp_val_length = maxlen_Acq_Stat
    end if
    exp_name       = "ACQUISITION_STATION"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expect_quotes=.true.)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Acquisition_Station, value, &
         "MPH%Acquisition_Station",do_CompareOnly, do_Compare, error_flag)
    IF (error_flag .ne. no_error) return

    ! read the processing center ID
    exp_val_length = len(MPH%Proc_Center)
    exp_name       = "PROC_CENTER"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expect_quotes=.true.)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Proc_Center, value, &
         "MPH%Proc_Center",do_CompareOnly, do_Compare, error_flag)
    IF (error_flag .ne. no_error) return

    ! read the processing time
    exp_val_length = len(MPH%Proc_Time)
    exp_name       = "PROC_TIME"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expect_quotes=.true.)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Proc_Time, value, &
         "MPH%Proc_Time",do_CompareOnly, do_Compare, error_flag,&
         CompareAsDate=.true.)
    IF (error_flag .ne. no_error) return

    ! read the software version number
    exp_val_length = len(MPH%Software_Ver)
    exp_name       = "SOFTWARE_VER"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expect_quotes=.true.)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Software_Ver, value, &
         "MPH%Software_Ver",do_CompareOnly, do_Compare, error_flag)
    IF (error_flag .ne. no_error) return

    ! Spare_2: this is a spare value without content
    exp_val_length = 40
    call read_spare_kvt_string(fileunit,exp_val_length,error_flag)
    IF (error_flag .ne. no_error) return

    ! read the start time of sensing
    exp_val_length = len(MPH%Sensing_Start)
    exp_name       = "SENSING_START"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expect_quotes=.true.)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Sensing_Start, value, &
         "MPH%Sensing_Start",do_CompareOnly, do_Compare, error_flag, &
         CompareAsDate=.true.)
    IF (error_flag .ne. no_error) return

    ! read the stop time of sensing
    exp_val_length = len(MPH%Sensing_Stop)
    exp_name       = "SENSING_STOP"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expect_quotes=.true.)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Sensing_Stop, value, &
         "MPH%Sensing_Stop",do_CompareOnly, do_Compare, error_flag, &
         CompareAsDate=.true.)
        IF (error_flag .ne. no_error) return

    ! Spare_3: this is a spare value without content
    exp_val_length = 40
    call read_spare_kvt_string(fileunit,exp_val_length,error_flag)
    IF (error_flag .ne. no_error) return

    ! read the phase number
    exp_val_length = len(MPH%Phase)
    exp_name       = "PHASE"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Phase, value, &
         "MPH%Phase",do_CompareOnly, do_Compare, error_flag)
    IF (error_flag .ne. no_error) return

    ! read the cycle number
    exp_val_length = len(MPH%Cycle)
    exp_name       = "CYCLE"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Cycle, value, &
         "MPH%Cycle",do_CompareOnly, do_Compare, error_flag,&
         CompareAsInteger=.true.)
    IF (error_flag .ne. no_error) return

    ! read the start relative orbit number
    exp_val_length = len(MPH%Rel_Orbit)
    exp_name       = "REL_ORBIT"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Rel_Orbit, value, &
         "MPH%Rel_Orbit",do_CompareOnly, do_Compare, error_flag, &
         CompareAsInteger=.true.)
    IF (error_flag .ne. no_error) return

    ! read the start absolute orbit number
    exp_val_length = len(MPH%Abs_Orbit)
    exp_name       = "ABS_ORBIT"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Abs_Orbit, value, &
         "MPH%Abs_Orbit",do_CompareOnly, do_Compare, error_flag, &
         CompareAsInteger=.true.)
    IF (error_flag .ne. no_error) return

    ! read the time of state vector
    exp_val_length = len(MPH%State_Vector_Time)
    exp_name       = "STATE_VECTOR_TIME"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length, expect_quotes=.true.)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%State_Vector_Time, value, &
         "MPH%State_Vector_Time",do_CompareOnly, do_Compare, error_flag, &
         CompareAsDate=.true.)
    IF (error_flag .ne. no_error) return

    ! read the Delta_UT1 = UT1-UTC
    exp_val_length = len(MPH%Delta_UT1)
    exp_name       = "DELTA_UT1"
    exp_phys_unit  = "s"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.,&
         expected_phys_unit=exp_phys_unit)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Delta_UT1, value, &
         "MPH%Delta_UT1",do_CompareOnly, do_Compare, error_flag, &
         CompareAsReal=.true.)
    IF (error_flag .ne. no_error) return

    ! read the X position in Earth-fixed-reference
    exp_val_length = len(MPH%X_Position)
    exp_name       = "X_POSITION"
    exp_phys_unit  = "m"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.,&
         expected_phys_unit=exp_phys_unit)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%X_Position, value, &
         "MPH%X_Position",do_CompareOnly, do_Compare, error_flag, &
         CompareAsReal=.true.)
    IF (error_flag .ne. no_error) return

    ! read the Y position in Earth-fixed-reference
    exp_val_length = len(MPH%Y_Position)
    exp_name       = "Y_POSITION"
    exp_phys_unit  = "m"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.,&
         expected_phys_unit=exp_phys_unit)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Y_Position, value, &
         "MPH%Y_Position",do_CompareOnly, do_Compare, error_flag, &
         CompareAsReal=.true.)
    IF (error_flag .ne. no_error) return

    ! read the Z position in Earth-fixed-reference
    exp_val_length = len(MPH%Z_Position)
    exp_name       = "Z_POSITION"
    exp_phys_unit  = "m"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.,&
         expected_phys_unit=exp_phys_unit)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Z_Position, value, &
         "MPH%Z_Position",do_CompareOnly, do_Compare, error_flag, &
         CompareAsReal=.true.)
    IF (error_flag .ne. no_error) return

    ! read the X velocity in Earth-fixed-reference
    exp_val_length = len(MPH%X_Velocity)
    exp_name       = "X_VELOCITY"
    exp_phys_unit  = "m/s"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.,&
         expected_phys_unit=exp_phys_unit)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%X_Velocity, value, &
         "MPH%X_Velocity",do_CompareOnly, do_Compare, error_flag, &
         CompareAsReal=.true.)
    IF (error_flag .ne. no_error) return

    ! read the Y velocity in Earth-fixed-reference
    exp_val_length = len(MPH%Y_Velocity)
    exp_name       = "Y_VELOCITY"
    exp_phys_unit  = "m/s"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.,&
         expected_phys_unit=exp_phys_unit)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Y_Velocity, value, &
         "MPH%Y_Velocity",do_CompareOnly, do_Compare, error_flag, &
         CompareAsReal=.true.)
    IF (error_flag .ne. no_error) return

    ! read the Z velocity in Earth-fixed-reference
    exp_val_length = len(MPH%Z_Velocity)
    exp_name       = "Z_VELOCITY"
    exp_phys_unit  = "m/s"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.,&
         expected_phys_unit=exp_phys_unit)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Z_Velocity, value, &
         "MPH%Z_Velocity",do_CompareOnly, do_Compare, error_flag,&
         CompareAsReal=.true.)
    IF (error_flag .ne. no_error) return

    ! read the Source of orbit vectors
    exp_val_length = len(MPH%Vector_Source)
    exp_name       = "VECTOR_SOURCE"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length, expect_quotes=.true.)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Vector_Source, value, &
         "MPH%Vector_Source",do_CompareOnly, do_Compare, error_flag)
    IF (error_flag .ne. no_error) return

    ! Spare_4: this is a spare value without content
    exp_val_length = 40
    call read_spare_kvt_string(fileunit,exp_val_length,error_flag)
    IF (error_flag .ne. no_error) return

    ! read the time corresponding to Sat. Binary time (not used by ADM_Aeolus)
    exp_val_length = len(MPH%UTC_SBT_Time)
    exp_name       = "UTC_SBT_TIME"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length, expect_quotes=.true.)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%UTC_SBT_Time, value, &
         "MPH%UTC_SBT_Time",do_CompareOnly, do_Compare, error_flag,&
         CompareAsDate=.true.)
    IF (error_flag .ne. no_error) return
    ! not used by ADM_Aeolus

    ! read the Satellite binary time (not used by ADM_Aeolus)
    exp_val_length = len(MPH%Sat_Binary_Time)
    exp_name       = "SAT_BINARY_TIME"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Sat_Binary_Time, value, &
         "MPH%Sat_Binary_Time",do_CompareOnly, do_Compare, error_flag,&
         CompareAsInteger=.true.)
    IF (error_flag .ne. no_error) return
    ! not used by ADM_Aeolus

    ! read the Clock step size  (not used by ADM_Aeolus)
    exp_val_length = len(MPH%Clock_Step)
    exp_name       = "CLOCK_STEP"
    exp_phys_unit  = "ps"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.,&
         expected_phys_unit=exp_phys_unit)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Clock_Step, value, &
         "MPH%Clock_Step",do_CompareOnly, do_Compare, error_flag,&
         CompareAsInteger=.true.)
    IF (error_flag .ne. no_error) return
    ! not used by ADM_Aeolus

    ! Spare_5: this is a spare value without content
    exp_val_length = 32
    call read_spare_kvt_string(fileunit,exp_val_length,error_flag)
    IF (error_flag .ne. no_error) return

    ! read the time of occurrence of the leap second
    exp_val_length = len(MPH%Leap_Utc)
    exp_name       = "LEAP_UTC"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length, expect_quotes=.true.)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Leap_Utc, value, &
         "MPH%Leap_Utc",do_CompareOnly, do_Compare, error_flag,&
         CompareAsDate=.true.)
    IF (error_flag .ne. no_error) return

    ! read the Leap second sign
    exp_val_length = len(MPH%Leap_Sign)
    exp_name       = "LEAP_SIGN"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Leap_Sign, value, &
         "MPH%Leap_Sign",do_CompareOnly, do_Compare, error_flag,&
         CompareAsInteger=.true.)
    IF (error_flag .ne. no_error) return

    ! read the Leap second error (flag)
    exp_val_length = len(MPH%Leap_Err)
    exp_name       = "LEAP_ERR"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Leap_Err, value, &
         "MPH%Leap_Err",do_CompareOnly, do_Compare, error_flag)
    IF (error_flag .ne. no_error) return
 
    ! extra check, allow only values '0' or '1'
    ! also allow missing for now for testing purposes
    IF ( (trim(MPH%LEAP_ERR) .ne.'0') .and. &
         (trim(MPH%LEAP_ERR) .ne.'1') .and. &
         (trim(MPH%LEAP_ERR) .ne.'') ) THEN
       call logmsg(log_error,"ERROR in Read_MPH_KVT(): "//&
                   "The LEAP_ERR field should contain 0 or 1, but it seems "//&
                   "to contain: ["//trim(MPH%LEAP_ERR)//"]")
       error_flag = error_programming
       return
    END IF

    ! Spare_6: this is a spare value without content
    exp_val_length = 40
    call read_spare_kvt_string(fileunit,exp_val_length,error_flag)
    IF (error_flag .ne. no_error) return

    ! read the product error
    exp_val_length = len(MPH%Product_Err)
    exp_name       = "PRODUCT_ERR"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Product_Err, value, &
         "MPH%Product_Err",do_CompareOnly, do_Compare, error_flag)
    IF (error_flag .ne. no_error) return
        
    ! extra check, allow only values '0' or '1'
    ! also allow missing for now for testing purposes
    IF ( (trim(MPH%PRODUCT_ERR) .ne.'0') .and. &
         (trim(MPH%PRODUCT_ERR) .ne.'1') .and. &
         (trim(MPH%PRODUCT_ERR) .ne.'') ) THEN
       call logmsg(log_warn,"WARNING in Read_MPH_KVT(): "//&
                   "The PRODUCT_ERR field should contain 0 or 1, but it seems "//&
                   "to contain: ["//trim(MPH%PRODUCT_ERR)//"]")
       !error_flag = error_programming
       !return
    END IF

    ! read the total product size ( = size(DSR+SPH+MPH) )
    exp_val_length = len(MPH%Tot_Size)
    exp_name       = "TOT_SIZE"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.,&
         expected_phys_unit="bytes")
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Tot_Size, value, &
         "MPH%Tot_Size",do_CompareOnly, do_Compare, error_flag,&
         CompareAsInteger=.true.)
        IF (error_flag .ne. no_error) return

    ! convert to a numerical (integer) value
    read(MPH%TOT_SIZE,"(i21)",iostat=ios) total_product_size
    IF (ios .ne. 0) goto 999

    ! request the actual filesize, to be able to compare the two
    actual_product_size = get_filesize(fileunit)

    ! check the actual filesize
    IF (actual_product_size .le. 0) THEN
       actual_filename(:) = ' '
       inquire(unit=fileunit,named=nmd)
       if (nmd) inquire(unit=fileunit,name=actual_filename)
       call logmsg(log_error,"ERROR in Read_MPH_KVT(): "//&
                   "could not open the file: "//trim(actual_filename)//&
                   " to retrieve its size (or size equals zero)")
       error_flag = error_reading_file
       return
    END IF

    ! check the actual and the expected filesize
    IF (actual_product_size .ne. total_product_size) THEN
       write(actual_product_size_txt,"(i20)",err=997) actual_product_size
       write(total_product_size_txt, "(i20)",err=998) total_product_size
       IF (use_workaround1) THEN
          call logmsg(log_warn,"WARNING in Read_MPH_KVT(): "//&
               "Actual filesize for L1B DBL file does not match the "//&
               "expected value as given inside this file!!! "//&
               "Found actual_product_size = "//&
               trim(adjustl(actual_product_size_txt))//&
               " Expected total_product_size = "//&
               trim(adjustl(total_product_size_txt))//&
               " (ignoring this problem for now)")
       ELSE
          call logmsg(log_error,"ERROR in Read_MPH_KVT(): "//&
               "Actual filesize for L1B DBL file does not match the "//&
               "expected value as given inside this file!!! "//&
               "Found actual_product_size = "//&
               trim(adjustl(actual_product_size_txt))//&
               " Expected total_product_size = "//&
               trim(adjustl(total_product_size_txt)))
          error_flag = error_reading_file
          return
       END IF
    END IF

    ! read the length of the SPH
    exp_val_length = len(MPH%Sph_Size)
    exp_name       = "SPH_SIZE"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.,&
         expected_phys_unit="bytes")
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Sph_Size, value, &
         "MPH%Sph_Size",do_CompareOnly, do_Compare, error_flag,&
         CompareAsInteger=.true.)
    IF (error_flag .ne. no_error) return

    ! read the number of DSD's
    exp_val_length = len(MPH%Num_Dsd)
    exp_name       = "NUM_DSD"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Num_Dsd, value, &
         "MPH%Num_Dsd",do_CompareOnly, do_Compare, error_flag,&
         CompareAsInteger=.true.)
    IF (error_flag .ne. no_error) return

    ! read the length of each DSD (all DSD's shall have the same length)
    exp_val_length = len(MPH%Dsd_Size)
    exp_name       = "DSD_SIZE"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.,&
         expected_phys_unit="bytes")
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Dsd_Size, value, &
         "MPH%Dsd_Size",do_CompareOnly, do_Compare,error_flag,&
         CompareAsInteger=.true.)
    IF (error_flag .ne. no_error) return

    ! read the number of DS attached (not all DSD's have a DS attached!)
    exp_val_length = len(MPH%Num_Data_Sets)
    exp_name       = "NUM_DATA_SETS"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) return

    call Fill_or_Compare_MPH_item(MPH%Num_Data_Sets, value, &
         "MPH%Num_Data_Sets",do_CompareOnly, do_Compare, error_flag,&
         CompareAsInteger=.true.)
    IF (error_flag .ne. no_error) return

    ! Spare_7: this is a spare value without content
    exp_val_length = 40
    call read_spare_kvt_string(fileunit,exp_val_length,error_flag)
    IF (error_flag .ne. no_error) return

    !------------------------------
    ! done
    !------------------------------
    
    return

997 call logmsg(log_error,"ERROR in Read_MPH_from_KVT: "//&
         "could not write integer actual_product_size to a string")
    error_flag = error_writing_file
    return

998 call logmsg(log_error,"ERROR in Read_MPH_from_KVT: "//&
         "could not write integer total_product_size to a string")
    error_flag = error_writing_file
    return

999 call logmsg(log_error,"ERROR in Read_MPH_from_KVT: "//&
         "could not read integer total_product_size from the string "//&
         "MPH%TOT_SIZE: "//trim(MPH%TOT_SIZE))
    error_flag = error_reading_file
    return

  end subroutine Read_MPH_from_KVT

    !  #]
  subroutine Write_MPH_to_KVT_gen(MPH, fileunit, error_flag, workaround)
    !  #[ 

    ! Writes the main product header (KVT values)
    ! 28-11-2005 P. Poli Original code. From L1B IODD v2.2 dated 09-09-2005
    ! 09-02-2006 P. Poli Added switch_AMD to write AMD MPH
    ! 21-02-2006 J. de Kloe removed switch_AMD
    ! reason: the MPH has a fixed format for all ee_cfi files, so there should
    !         not be exceptions in it for specific file types.
    !         If we need an offset to be exported to the calling routine
    !         it should be added to the parameter list.

    TYPE(MPH_type),    pointer       :: MPH
    TYPE(DirectBinaryIO_fileunit_type), intent(in) :: fileunit
    integer,           intent(out)   :: error_flag
    logical, optional, intent(in)    :: workaround

    ! local variables
    character(len=max_str_len) :: varname
    character(len=max_str_len) :: value
    character(len=max_str_len) :: phys_unit
    integer                    :: max_value_length, pos, ios
    integer(IntAuc)            :: Cycle1
    integer(i8_)               :: filepos_end,filepos_start
    integer(i8_)               :: nbytes_written, nbytes_expected

    ! a switch that allows using the missing indicator for a given
    ! numerical type, in case the MPH item is still empty
    ! Remind: for Boolean and DateTime types no exceptions are yet
    ! implemented for missing input, so these are just filled
    ! with blanks
    logical, parameter :: use_missing_indicator = .true. !.false.

    integer(IntAc)             :: Leap_Sign1
    integer(IntAs)             :: Rel_Orbit1
    integer(IntAus)            :: Abs_Orbit1
    real(FAdoxy)               :: Delta_UT11,&
                                  X_Position1,Y_Position1,Z_Position1, &
                                  X_Velocity1,Y_Velocity1, Z_Velocity1
    integer(IntAul)            :: Sat_Binary_Time1,Clock_Step1
    character(len=25)          :: txt_nbytes_written, txt_nbytes_expected
    logical                    :: use_workaround

    !------------------------------
    ! writing the MPH KVT elements
    !------------------------------

    error_flag = no_error

    use_workaround = .false.
    IF (present(workaround)) use_workaround = workaround

    IF (.not. associated(MPH)) THEN
       call logmsg(log_error,"ERROR in Write_MPH_to_KVT_gen(): "//&
                   "The pointer to this MPH seems not associated !!")
       error_flag = error_programming
       return
    END IF

    ! as double check, compare the number of bytes written
    ! to what was expected from the GetByteSize_MPH_KVT() function
    ! so store the filepointer before we write anything
    call GetFilePointer(fileunit,filepos_start,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    varname="PRODUCT"
    value=MPH%Product
    max_value_length=len(MPH%Product)
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag, use_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998
    
    varname="PROC_STAGE"
    value=MPH%Proc_Stage
    pos = verify(trim(value),"NRBT")
    IF (pos .ne. 0) THEN
       call logmsg(log_error,"ERROR in Write_MPH_to_KVT_gen(): "//&
                   "item: PROC_STAGE is only allowed to have the values "//&
                   "N, R, T, or B, but is seems to contain: "//trim(value)//&
                   "; Check program.")
       error_flag = error_programming
       return
    END IF
    max_value_length=len(MPH%Proc_Stage)
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    varname="REF_DOC"
    value=MPH%Ref_Doc
    max_value_length=len(MPH%Ref_Doc)
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag,use_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998

    ! Spare_1: this is a spare value without content
    max_value_length = 40
    call write_spare_kvt_string(fileunit,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    varname="ACQUISITION_STATION"
    value=MPH%Acquisition_Station

    if (use_workaround) then
       max_value_length = maxlen_Acq_Stat_108
    else
       max_value_length = maxlen_Acq_Stat
    end if
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag,use_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998
    
    varname="PROC_CENTER"
    value=MPH%Proc_Center
    max_value_length=len(MPH%Proc_Center)
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag,use_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998
    
    varname="PROC_TIME"
    value=MPH%Proc_Time
    max_value_length=len(MPH%Proc_Time)
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag,use_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998
    
    varname="SOFTWARE_VER"
    value=MPH%Software_Ver
    max_value_length=len(MPH%Software_Ver)
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag,use_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998
    
    ! Spare_2: this is a spare value without content
    max_value_length = 40
    call write_spare_kvt_string(fileunit,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    varname="SENSING_START"
    value=MPH%Sensing_Start
    max_value_length=len(MPH%Sensing_Start)
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag,use_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998
    
    varname="SENSING_STOP"
    value=MPH%Sensing_Stop
    max_value_length=len(MPH%Sensing_Stop)
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag,use_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998
    
    ! Spare_3: this is a spare value without content
    max_value_length = 40
    call write_spare_kvt_string(fileunit,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    varname="PHASE"
    value=MPH%Phase
    max_value_length=len(MPH%Phase)
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag)
    IF (error_flag .ne. no_error) goto 998

    varname="CYCLE"
    value=MPH%Cycle
    ! WARNING: this if-then construct is needed if MPH%Cycle can have an
    ! empty string! Otherwise the KVT element for Cycle might not be
    ! properly filled with a numerical value
    !
    ! TODO: remove this exception
    !
    IF (len_trim(MPH%Cycle) .gt. 0) THEN
       read(value,*,iostat=ios) Cycle1
       IF (ios .ne. 0) goto 990
       call SetIntAuc(Cycle1,value(1:maxlen_IntAuc),error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       IF (use_missing_indicator) THEN
          Cycle1 = missing_indicator_IntAuc
          call SetIntAuc(Cycle1,value(1:maxlen_IntAuc),error_flag)
          IF (error_flag .ne. no_error) goto 998
       ELSE
          ! filling with spaces is not a legal KVT numerical value
          ! but is allowed for compatibility reasons for now...
          !
          ! TODO: remove this exception
          !
          value(:) = ' '
       END IF
    END IF
    max_value_length=len(MPH%Cycle)
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    varname="REL_ORBIT"
    value=MPH%Rel_Orbit
    IF (len_trim(MPH%Rel_Orbit) .gt. 0) THEN
       read(value,*,iostat=ios) Rel_Orbit1
       IF (ios .ne. 0) goto 990
       call SetIntAs(Rel_Orbit1,value(1:maxlen_IntAs),error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       IF (use_missing_indicator) THEN
          Rel_Orbit1 = missing_indicator_IntAs
          call SetIntAs(Rel_Orbit1,value(1:maxlen_IntAs),error_flag)
          IF (error_flag .ne. no_error) goto 998
       ELSE
          ! filling with spaces is not a legal KVT numerical value
          ! but is allowed for compatibility reasons for now...
          !
          ! TODO: remove this exception
          !
          value(:) = ' '
       END IF
    END IF
    max_value_length=len(MPH%Rel_Orbit)
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    varname="ABS_ORBIT"
    value=MPH%Abs_Orbit
    IF (len_trim(MPH%Abs_Orbit) .gt. 0) THEN
       read(value,*,iostat=ios) Abs_Orbit1
       IF (ios .ne. 0) goto 990
       call SetIntAus(Abs_Orbit1,value(1:maxlen_IntAus),error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       IF (use_missing_indicator) THEN
          Abs_Orbit1 = missing_indicator_IntAus
          call SetIntAus(Abs_Orbit1,value(1:maxlen_IntAus),error_flag)
          IF (error_flag .ne. no_error) goto 998
       ELSE
          ! filling with spaces is not a legal KVT numerical value
          ! but is allowed for compatibility reasons for now...
          !
          ! TODO: remove this exception
          !
          value(:) = ' '
       END IF
    END IF
    max_value_length=len(MPH%Abs_Orbit)
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    varname="STATE_VECTOR_TIME"
    value=MPH%State_Vector_Time
    max_value_length=len(MPH%State_Vector_Time)
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag,use_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998
    
    varname="DELTA_UT1"
    value=MPH%Delta_UT1
    IF (len_trim(MPH%Delta_UT1) .gt. 0) THEN
       read(value,*,iostat=ios) Delta_UT11
       IF (ios .ne. 0) goto 991
       call SetFAdoxy(Delta_UT11,0,6,value(1:maxlen_FAdo06),error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       IF (use_missing_indicator) THEN
          Delta_UT11 = missing_indicator_FAdo06
          call SetFAdoxy(Delta_UT11,0,6,value(1:maxlen_FAdo06),error_flag)
          IF (error_flag .ne. no_error) goto 998
       ELSE
          ! filling with spaces is not a legal KVT numerical value
          ! but is allowed for compatibility reasons for now...
          !
          ! TODO: remove this exception
          !
          value(:) = ' '
       END IF
    END IF
    max_value_length=len(MPH%Delta_UT1)
    phys_unit="s"
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag,phys_unit=phys_unit)
    IF (error_flag .ne. no_error) goto 998
    
    varname="X_POSITION"
    value=MPH%X_Position
    IF (len_trim(MPH%X_Position) .gt. 0) THEN
       read(value,*,iostat=ios) X_Position1
       IF (ios .ne. 0) goto 991
       call SetFAdoxy(X_Position1,7,3,value(1:maxlen_FAdo73),error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       IF (use_missing_indicator) THEN
          X_Position1 = missing_indicator_FAdo73
          call SetFAdoxy(X_Position1,7,3,value(1:maxlen_FAdo73),error_flag)
          IF (error_flag .ne. no_error) goto 998
       ELSE
          ! filling with spaces is not a legal KVT numerical value
          ! but is allowed for compatibility reasons for now...
          !
          ! TODO: remove this exception
          !
          value(:) = ' '
       END IF
    END IF
    max_value_length=len(MPH%X_Position)
    phys_unit="m"
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag,phys_unit=phys_unit)
    IF (error_flag .ne. no_error) goto 998
    
    varname="Y_POSITION"
    value=MPH%Y_Position
    IF (len_trim(MPH%Y_Position) .gt. 0) THEN
       read(value,*,iostat=ios) Y_Position1
       IF (ios .ne. 0) goto 991
       call SetFAdoxy(Y_Position1,7,3,value(1:maxlen_FAdo73),error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       IF (use_missing_indicator) THEN
          Y_Position1 = missing_indicator_FAdo73
          call SetFAdoxy(Y_Position1,7,3,value(1:maxlen_FAdo73),error_flag)
          IF (error_flag .ne. no_error) goto 998
       ELSE
          ! filling with spaces is not a legal KVT numerical value
          ! but is allowed for compatibility reasons for now...
          !
          ! TODO: remove this exception
          !
          value(:) = ' '
       END IF
    END IF
    max_value_length=len(MPH%Y_Position)
    phys_unit="m"
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag,phys_unit=phys_unit)
    IF (error_flag .ne. no_error) goto 998

    varname="Z_POSITION"
    value=MPH%Z_Position
    IF (len_trim(MPH%Z_Position) .gt. 0) THEN
       read(value,*,iostat=ios) Z_Position1
       IF (ios .ne. 0) goto 991
       call SetFAdoxy(Z_Position1,7,3,value(1:maxlen_FAdo73),error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       IF (use_missing_indicator) THEN
          Z_Position1 = missing_indicator_FAdo73
          call SetFAdoxy(Z_Position1,7,3,value(1:maxlen_FAdo73),error_flag)
          IF (error_flag .ne. no_error) goto 998
       ELSE
          ! filling with spaces is not a legal KVT numerical value
          ! but is allowed for compatibility reasons for now...
          !
          ! TODO: remove this exception
          !
          value(:) = ' '
       END IF
    END IF
    max_value_length=len(MPH%Z_Position)
    phys_unit="m"
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag, phys_unit=phys_unit)
    IF (error_flag .ne. no_error) goto 998
    
    varname="X_VELOCITY"
    value=MPH%X_Velocity
    IF (len_trim(MPH%X_Velocity) .gt. 0) THEN
       read(value,*,iostat=ios) X_Velocity1
       IF (ios .ne. 0) goto 991
       call SetFAdoxy(X_Velocity1,4,6,value(1:maxlen_FAdo46),error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       IF (use_missing_indicator) THEN
          X_Velocity1 = missing_indicator_FAdo46
          call SetFAdoxy(X_Velocity1,4,6,value(1:maxlen_FAdo46),error_flag)
          IF (error_flag .ne. no_error) goto 998
       ELSE
          ! filling with spaces is not a legal KVT numerical value
          ! but is allowed for compatibility reasons for now...
          !
          ! TODO: remove this exception
          !
          value(:) = ' '
       END IF
    END IF
    max_value_length=len(MPH%X_Velocity)
    phys_unit="m/s"
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag, phys_unit=phys_unit)
    IF (error_flag .ne. no_error) goto 998
    
    varname="Y_VELOCITY"
    value=MPH%Y_Velocity
    IF (len_trim(MPH%Y_Velocity) .gt. 0) THEN
       read(value,*,iostat=ios) Y_Velocity1
       IF (ios .ne. 0) goto 991
       call SetFAdoxy(Y_Velocity1,4,6,value(1:maxlen_FAdo46),error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       IF (use_missing_indicator) THEN
          Y_Velocity1 = missing_indicator_FAdo46
          call SetFAdoxy(Y_Velocity1,4,6,value(1:maxlen_FAdo46),error_flag)
          IF (error_flag .ne. no_error) goto 998
       ELSE
          ! filling with spaces is not a legal KVT numerical value
          ! but is allowed for compatibility reasons for now...
          !
          ! TODO: remove this exception
          !
          value(:) = ' '
       END IF
    END IF
    max_value_length=len(MPH%Y_Velocity)
    phys_unit="m/s"
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag,phys_unit=phys_unit)
    IF (error_flag .ne. no_error) goto 998

    varname="Z_VELOCITY"
    value=MPH%Z_Velocity
    IF (len_trim(MPH%Z_Velocity) .gt. 0) THEN
       read(value,*,iostat=ios) Z_Velocity1
       IF (ios .ne. 0) goto 991
       call SetFAdoxy(Z_Velocity1,4,6,value(1:maxlen_FAdo46),error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       IF (use_missing_indicator) THEN
          Z_Velocity1 = missing_indicator_FAdo46
          call SetFAdoxy(Z_Velocity1,4,6,value(1:maxlen_FAdo46),error_flag)
          IF (error_flag .ne. no_error) goto 998
       ELSE
          ! filling with spaces is not a legal KVT numerical value
          ! but is allowed for compatibility reasons for now...
          !
          ! TODO: remove this exception
          !
          value(:) = ' '
       END IF
    END IF
    max_value_length=len(MPH%Z_Velocity)
    phys_unit="m/s"
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag,phys_unit=phys_unit)
    IF (error_flag .ne. no_error) goto 998

    varname="VECTOR_SOURCE"
    value=MPH%Vector_Source
    max_value_length=len(MPH%Vector_Source)
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag,use_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998
    
    ! Spare_4: this is a spare value without content
    max_value_length = 40
    call write_spare_kvt_string(fileunit,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    varname="UTC_SBT_TIME"
    value=MPH%UTC_SBT_Time
    max_value_length=len(MPH%UTC_SBT_Time)
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag,use_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998
    
    varname="SAT_BINARY_TIME"
    value=MPH%Sat_Binary_Time
    IF (len_trim(MPH%Sat_Binary_Time) .gt. 0) THEN
       read(value,*,iostat=ios) Sat_Binary_Time1
       IF (ios .ne. 0) goto 990
       call SetIntAul(Sat_Binary_Time1,value(1:maxlen_IntAul),error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       IF (use_missing_indicator) THEN
          Sat_Binary_Time1 = missing_indicator_IntAul
          call SetIntAul(Sat_Binary_Time1,value(1:maxlen_IntAul),error_flag)
          IF (error_flag .ne. no_error) goto 998
       ELSE
          ! filling with spaces is not a legal KVT numerical value
          ! but is allowed for compatibility reasons for now...
          !
          ! TODO: remove this exception
          !
          value(:) = ' '
       END IF
    END IF
    max_value_length=len(MPH%Sat_Binary_Time)
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    varname="CLOCK_STEP"
    value=MPH%Clock_Step
    IF (len_trim(MPH%Clock_Step) .gt. 0) THEN
       read(value,*,iostat=ios) Clock_Step1
       IF (ios .ne. 0) goto 990
       call SetIntAul(Clock_Step1,value(1:maxlen_IntAul),error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       IF (use_missing_indicator) THEN
          Clock_Step1 = missing_indicator_IntAul
          call SetIntAul(Clock_Step1,value(1:maxlen_IntAul),error_flag)
          IF (error_flag .ne. no_error) goto 998
       ELSE
          ! filling with spaces is not a legal KVT numerical value
          ! but is allowed for compatibility reasons for now...
          !
          ! TODO: remove this exception
          !
          value(:) = ' '
       END IF
    END IF
    max_value_length=len(MPH%Clock_Step)
    phys_unit="ps"
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag,phys_unit=phys_unit)
    IF (error_flag .ne. no_error) goto 998
    
    ! Spare_5: this is a spare value without content
    max_value_length = 32
    call write_spare_kvt_string(fileunit,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    varname="LEAP_UTC"
    value=MPH%Leap_Utc
    max_value_length=len(MPH%Leap_Utc)
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag,use_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998
       
    varname="LEAP_SIGN"
    value=MPH%Leap_Sign
    IF (len_trim(MPH%Leap_Sign) .gt. 0) THEN
       read(value,*,iostat=ios) Leap_Sign1
       IF (ios .ne. 0) goto 990
       call SetIntAc(Leap_Sign1,value(1:maxlen_IntAc),error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       IF (use_missing_indicator) THEN
          Leap_Sign1 = missing_indicator_IntAc
          call SetIntAc(Leap_Sign1,value(1:maxlen_IntAc),error_flag)
          IF (error_flag .ne. no_error) goto 998
       ELSE
          ! filling with spaces is not a legal KVT numerical value
          ! but is allowed for compatibility reasons for now...
          !
          ! TODO: remove this exception
          !
          value(:) = ' '
       END IF
    END IF
    max_value_length=len(MPH%Leap_Sign)
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    varname="LEAP_ERR"
    value=MPH%Leap_Err
    ! extra check, allow only values '0' or '1'
    ! also allow missing for now for testing purposes
    IF ( (trim(MPH%LEAP_ERR) .ne.'0') .and. &
         (trim(MPH%LEAP_ERR) .ne.'1') .and. &
         (trim(MPH%LEAP_ERR) .ne.'') ) THEN
       call logmsg(log_error,"WARNING in Write_MPH_to_KVT_gen(): "//&
                   "The LEAP_ERR field should contain 0 or 1, but it seems "//&
                   "to contain: ["//trim(MPH%LEAP_ERR)//"]")
       error_flag = error_programming
       return
    END IF
    max_value_length=len(MPH%Leap_Err)
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! Spare_6: this is a spare value without content
    max_value_length = 40
    call write_spare_kvt_string(fileunit,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    varname="PRODUCT_ERR"
    value=MPH%Product_Err
    ! extra check, allow only values '0' or '1'
    ! also allow missing for now for testing purposes
    IF ( (trim(MPH%PRODUCT_ERR) .ne.'0') .and. &
         (trim(MPH%PRODUCT_ERR) .ne.'1') .and. &
         (trim(MPH%PRODUCT_ERR) .ne.'') ) THEN
       call logmsg(log_error,"WARNING in Write_MPH_to_KVT_gen(): "//&
                   "The PRODUCT_ERR field should contain 0 or 1, but it seems "//&
                   "to contain: ["//trim(MPH%PRODUCT_ERR)//"]")
       error_flag = error_programming
       return
    END IF
    max_value_length=len(MPH%Product_Err)
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    varname="TOT_SIZE"
    value=MPH%Tot_Size
    max_value_length=len(MPH%Tot_Size)
    phys_unit="bytes"
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag,&
                          phys_unit=phys_unit)
    IF (error_flag .ne. no_error) goto 998

    varname="SPH_SIZE"
    value=MPH%Sph_Size
    max_value_length=len(MPH%Sph_Size)
    phys_unit="bytes"
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag,&
                          phys_unit=phys_unit)
    IF (error_flag .ne. no_error) goto 998

    varname="NUM_DSD"
    value=MPH%Num_Dsd
    max_value_length=len(MPH%Num_Dsd)
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    varname="DSD_SIZE"
    value=MPH%Dsd_Size
    max_value_length=len(MPH%Dsd_Size)
    phys_unit="bytes"
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag,&
                          phys_unit=phys_unit)
    IF (error_flag .ne. no_error) goto 998

    varname="NUM_DATA_SETS"
    value=MPH%Num_Data_Sets
    max_value_length=len(MPH%Num_Data_Sets)
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Spare_7: this is a spare value without content
    max_value_length = 40
    call write_spare_kvt_string(fileunit,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! as double check, compare the number of bytes written
    ! to what was expected from the GetByteSize_MPH_KVT() function
    ! this only makes sence if teh full MPH is written !!
    call GetFilePointer(fileunit,filepos_end,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    nbytes_written  = filepos_end-filepos_start
    nbytes_expected = GetByteSize_MPH_KVT(workaround)
    
    IF (nbytes_written .ne. nbytes_expected) THEN
       write(txt_nbytes_written,*,err=980) nbytes_written
       write(txt_nbytes_expected,*,err=981) nbytes_expected
       call logmsg(log_error,"ERROR in Write_MPH_to_KVT_gen():"//&
                   " nbytes_written  = "//trim(txt_nbytes_written)//&
                   " nbytes_expected = "//trim(txt_nbytes_expected)//&
                   " something went wrong when writing the MPH to the KVT"//&
                   " part of the binary file, probably a programming error ...")
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! done
    !------------------------------
    return

980 call logmsg(log_error,"ERROR in Write_MPH_to_KVT_gen: "//&
                "could not convert nbytes_written into a string")
    error_flag = error_writing_file
    return

981 call logmsg(log_error,"ERROR in Write_MPH_to_KVT_gen: "//&
                "could not convert nbytes_expected into a string")
    error_flag = error_writing_file
    return

990 call logmsg(log_error,"ERROR in Write_MPH_to_KVT_gen(): "//&
                "Could not convert MPH value into KVT: "//trim(value)//&
                "into an integer number ....")
    error_flag = error_writing_file
    return

991 call logmsg(log_error,"ERROR in Write_MPH_to_KVT_gen(): "//&
                "Could not convert MPH value into KVT: "//trim(value)//&
                "into a real number ....")
    error_flag = error_writing_file
    return

998 call logmsg(log_error,"ABORTING subroutine Write_MPH_to_KVT_gen")
    return

  end subroutine Write_MPH_to_KVT_gen
    !  #]
  subroutine Write_MPH_to_KVT_unbuff(MPH, fileunit, error_flag, workaround)
    !  #[ 
    TYPE(MPH_type),    pointer       :: MPH
    integer,           intent(in)    :: fileunit
    integer,           intent(out)   :: error_flag
    logical, optional, intent(in)    :: workaround

    ! local variable
    TYPE(DirectBinaryIO_fileunit_type) :: local_fileunit

    error_flag = no_error

    call Init_DirectBinaryIO_fileunit(local_fileunit)
    local_fileunit%use_buffering = .false.
    local_fileunit%fileunit      = fileunit

    call Write_MPH_to_KVT_gen(MPH, local_fileunit, error_flag, workaround)

  end subroutine Write_MPH_to_KVT_unbuff
    !  #]
  subroutine Write_MPH_to_KVT_buff(MPH, fileunit, error_flag, workaround)
    !  #[ 
    TYPE(MPH_type),    pointer       :: MPH
    TYPE(BufferedFile_type), pointer :: fileunit
    integer,           intent(out)   :: error_flag
    logical, optional, intent(in)    :: workaround

    ! local variable
    TYPE(DirectBinaryIO_fileunit_type) :: local_fileunit

    error_flag = no_error

    call Init_DirectBinaryIO_fileunit(local_fileunit)
    local_fileunit%use_buffering = .true.
    local_fileunit%buffered_fileunit => fileunit

     call Write_MPH_to_KVT_gen(MPH, local_fileunit, error_flag, workaround)

   end subroutine Write_MPH_to_KVT_buff
    !  #]
  !---------------------------------------------
  subroutine Fill_L2B_MPH_From_L1B_MPH(L1B_MPH,L2B_MPH,error_flag)
    !  #[

    TYPE(MPH_type), pointer :: L1B_MPH,L2B_MPH
    integer,    intent(out) :: error_flag

    error_flag = no_error
    IF (.not. associated(L1B_MPH)) THEN
       call logmsg(log_error,"ERROR in Fill_L2B_MPH_From_L1B_MPH(): "//&
                   "The pointer to this L1B_MPH seems not associated !!")
       error_flag = error_programming
       return
    END IF
    IF (.not. associated(L2B_MPH)) THEN
       call logmsg(log_error,"ERROR in Fill_L2B_MPH_From_L1B_MPH(): "//&
                   "The pointer to this L2B_MPH seems not associated !!")
       error_flag = error_programming
       return
    END IF

    !L2B_MPH%Product(:)             = ' ' ! copied from L2B FH
    !L2B_MPH%Proc_Stage(:)          = ' ' ! fixed value "T" for now
    !L2B_MPH%Ref_Doc(:)             = ' ' ! is set by MPH_Default_Fields
    ! Spare 1
    L2B_MPH%Acquisition_Station(:) = L1B_MPH%Acquisition_Station(:)
    !L2B_MPH%Proc_Center(:)         = ' ' ! is set by MPH_Default_Fields
    !L2B_MPH%Proc_Time(:)           = ' ' ! filled using computer clock
    !L2B_MPH%Software_Ver(:)        = ' ' ! is set by MPH_Default_Fields
    ! Spare_2

    ! these 2 cannot be copied from L1B, because the L2BP may not 
    ! process all BRC's, in which case the start/stop times will differ
    !L2B_MPH%Sensing_Start(:)       = L1B_MPH%Sensing_Start(:)
    !L2B_MPH%Sensing_Stop(:)        = L1B_MPH%Sensing_Stop(:)

    ! Spare_3
    L2B_MPH%Phase(:)               = L1B_MPH%Phase(:)
    L2B_MPH%Cycle(:)               = L1B_MPH%Cycle(:)
    L2B_MPH%Rel_Orbit(:)           = L1B_MPH%Rel_Orbit(:)
    L2B_MPH%Abs_Orbit(:)           = L1B_MPH%Abs_Orbit(:)
    L2B_MPH%State_Vector_Time(:)   = L1B_MPH%State_Vector_Time(:)
    L2B_MPH%Delta_UT1(:)           = L1B_MPH%Delta_UT1(:)
    L2B_MPH%X_Position(:)          = L1B_MPH%X_Position(:)
    L2B_MPH%Y_Position(:)          = L1B_MPH%Y_Position(:)
    L2B_MPH%Z_Position(:)          = L1B_MPH%Z_Position(:)
    L2B_MPH%X_Velocity(:)          = L1B_MPH%X_Velocity(:)
    L2B_MPH%Y_Velocity(:)          = L1B_MPH%Y_Velocity(:)
    L2B_MPH%Z_Velocity(:)          = L1B_MPH%Z_Velocity(:)
    L2B_MPH%Vector_Source(:)       = L1B_MPH%Vector_Source(:)
    ! Spare 4
    L2B_MPH%UTC_SBT_Time(:)        = L1B_MPH%UTC_SBT_Time(:)
    L2B_MPH%Sat_Binary_Time(:)     = L1B_MPH%Sat_Binary_Time(:)
    L2B_MPH%Clock_Step(:)          = L1B_MPH%Clock_Step(:)
    ! Spare 5
    L2B_MPH%Leap_Utc(:)            = L1B_MPH%Leap_Utc(:)
    L2B_MPH%Leap_Sign(:)           = L1B_MPH%Leap_Sign(:)
    L2B_MPH%Leap_Err(:)            = L1B_MPH%Leap_Err(:)
    ! Spare 6
!    L2B_MPH%Product_Err(:)         = ' '
!    L2B_MPH%Tot_Size(:)            = ' ' ! calculated
!    L2B_MPH%Sph_Size(:)            = ' ' ! calculated
!    L2B_MPH%Num_Dsd(:)             = ' ' ! fixed per filetype
!    L2B_MPH%Dsd_Size(:)            = ' ' ! is a constant for EE files
!    L2B_MPH%Num_Data_Sets(:)       = ' ' ! fixed per filetype
    ! Spare 7

    return

  end subroutine Fill_L2B_MPH_From_L1B_MPH
    !  #]
  subroutine Fill_MPH_Sizes (L2B_MPH,Tot_Size, Sph_Size, Num_Dsd, &
                             Dsd_Size, Num_Data_Sets, error_flag)
    !  #[
    !  01-Dec-2005 P. Poli   Original code.
    !  13-Mar-2007 J.de Kloe  removed use of SummaryVars
    TYPE(MPH_type), pointer     :: L2B_MPH ! output, some fields will be filled
    integer(IntAd), intent(in)  :: Tot_Size      ! input
    integer(IntAl), intent(in)  :: Sph_Size      ! input
    integer(IntAl), intent(in)  :: Num_Dsd       ! input
    integer(IntAl), intent(in)  :: Dsd_Size      ! input
    integer(IntAl), intent(in)  :: Num_Data_Sets ! input
    integer,        intent(out) :: error_flag    ! output

    error_flag = no_error

    IF (.not. associated(L2B_MPH)) THEN
       call logmsg(log_error,"ERROR in Fill_L2B_MPH_From_L1B_MPH(): "//&
                   "The pointer to this L2B_MPH seems not associated !!")
       error_flag = error_programming
       return
    END IF

    ! Tot_Size
    call SetIntAd(Tot_Size,L2B_MPH%Tot_Size,error_flag)
    IF (error_flag .ne. no_error) return
    IF (Tot_Size .le. 0) THEN
       call logmsg(log_warn,"Fill_L2B_MPH_Sizes: WARNING Tot_Size ="//&
                   trim(L2B_MPH%Tot_Size))
    END IF

    ! Sph_Size
    call SetIntAl(Sph_Size,L2B_MPH%Sph_Size,error_flag)
    IF (error_flag .ne. no_error) return
    IF (Sph_Size .le. 0) THEN
       call logmsg(log_warn,"Fill_L2B_MPH_Sizes: WARNING Sph_Size ="//&
                   trim(L2B_MPH%Sph_Size))
    END IF

    ! Num_Dsd
    call SetIntAl(Num_Dsd ,L2B_MPH%Num_Dsd ,error_flag)
    IF (error_flag .ne. no_error) return
    IF (Num_Dsd  .le. 0) THEN
       call logmsg(log_warn,"Fill_L2B_MPH_Sizes: WARNING Num_Dsd ="//&
                   trim(L2B_MPH%Num_Dsd))
    END IF

    ! Dsd_Size
    call SetIntAl(Dsd_Size,L2B_MPH%Dsd_Size,error_flag)
    IF (error_flag .ne. no_error) return
    IF (Dsd_Size .le. 0) THEN
       call logmsg(log_warn,"Fill_L2B_MPH_Sizes: WARNING Dsd_Size ="//&
                   trim(L2B_MPH%Dsd_Size))
    END IF

    ! Num_Data_Sets
    call SetIntAl(Num_Data_Sets,L2B_MPH%Num_Data_Sets,error_flag)
    IF (error_flag .ne. no_error) return
    IF (Num_Data_Sets .le. 0) THEN
       call logmsg(log_warn,"Fill_L2B_MPH_Sizes: WARNING Num_Data_Sets ="//&
                   trim(L2B_MPH%Num_Data_Sets))
    END IF

  end subroutine Fill_MPH_Sizes
  !  #]
  !---------------------------------------------
  ! local private subroutines
  subroutine add_MPH_item(xml_tree, path_in, item_name, item, error_flag )
    !  #[
    TYPE(xml_node_type), pointer     :: xml_tree   ! points to updated struct
    character(len=*),    intent(in)  :: path_in    ! input
    character(len=*),    intent(in)  :: item_name  ! input
    character(len=*),    intent(in)  :: item       ! input
    integer,             intent(out) :: error_flag

    ! local variables
    character(len=max_str_len) :: path

    error_flag = no_error
    path = trim(path_in)//"/"//ee_vh_name//"/"//ee_mph_name

    call xml_tree_add_child(xml_tree,path,item_name,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(path)//"/"//trim(item_name)
    call xml_tree_set_string_node_value(xml_tree,path,item,error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine add_MPH_item
    !  #]
  subroutine Fill_or_Compare_MPH_item(MPH_value,KVT_value,Item,&
             do_CompareOnly,do_Compare, error_flag, &
             CompareAsInteger,CompareAsReal,CompareAsDate)
    !  #[
    ! Modifications
    !   P. Poli   29-11-2005  When do_CompareOnly and CompareAsDate are true,
    !                         compare times but REPLACE the times
    !                         in memory by those read from KVT
    !                         because they are accurate to 1 microsecond
    !                         while those read previously might
    !                         come from XML (i.e. accurate to the second only)
    !   P. Poli   02-12-2005  Added flag do_Compare to compare and replace.

    character(len=*), intent(inout) :: MPH_value ! filled or compared
    character(len=*), intent(in)    :: KVT_value ! input
    character(len=*), intent(in)    :: Item      ! input
    logical,          intent(in)    :: do_CompareOnly    ! switch
    logical,          intent(in)    :: do_Compare        ! switch
    integer,          intent(out)   :: error_flag
    logical,optional, intent(in)    :: CompareAsInteger  ! switch
    logical,optional, intent(in)    :: CompareAsReal     ! switch
    logical,optional, intent(in)    :: CompareAsDate     ! switch

    ! local variables
    logical :: do_CompareAsInteger
    logical :: do_CompareAsReal
    logical :: do_CompareAsDate
    logical :: ValuesDiffer
    integer        :: ios
    integer(IntAd) :: Temp_KVT_IntAd  ! temporary int*8
    integer(IntAd) :: Temp_MPH_IntAd  ! temporary int*8
    real(FAdoxy)   :: Temp_KVT_FAdoxy ! temporary real*8
    real(FAdoxy)   :: Temp_MPH_FAdoxy ! temporary real*8

    ! initialise, to prevent the from being uninitialised
    ! when a read statement fails without reporting an error.
    ! (seems to happen with g95)
    Temp_KVT_IntAd = 0
    Temp_MPH_IntAd = 0
    Temp_KVT_FAdoxy = 0.
    Temp_MPH_FAdoxy = 0.
    error_flag = no_error

    do_CompareAsInteger = .false.
    IF (present(CompareAsInteger)) do_CompareAsInteger = CompareAsInteger

    do_CompareAsReal = .false.
    IF (present(CompareAsReal)) do_CompareAsReal = CompareAsReal

    do_CompareAsDate = .false.
    IF (present(CompareAsDate)) do_CompareAsDate = CompareAsDate

    IF (do_CompareOnly .or. do_Compare) THEN
       ! compare the values
       ValuesDiffer = .false.
       IF (do_CompareAsInteger) THEN
          ! convert to a numerical (integer) value
          read(KVT_value,*,iostat=ios) Temp_KVT_IntAd
          IF (ios .ne. 0) goto 990
          read(MPH_value,*,iostat=ios) Temp_MPH_IntAd
          IF (ios .ne. 0) goto 991

          IF ((Temp_KVT_IntAd .ne. Temp_MPH_IntAd)) THEN
             ! this is an exception at the moment that occurs when
             ! an attempt is made to read a value from an empty MPH item
             IF (Temp_MPH_IntAd .ne. 0) ValuesDiffer = .true.
          END IF
       ELSE IF (do_CompareAsReal) THEN
          ! convert to a numerical (real) value
          ! should I add a delta here ?
          read(KVT_value,*,iostat=ios) Temp_KVT_FAdoxy
          IF (ios .ne. 0) goto 992
          read(MPH_value,*,iostat=ios) Temp_MPH_FAdoxy
          IF (ios .ne. 0) goto 993

          IF (Temp_KVT_FAdoxy .ne. Temp_MPH_FAdoxy) THEN
             ! this is an exception at the moment that occurs when
             ! an attempt is made to read a value from an empty MPH item
             IF (Temp_MPH_FAdoxy .ne. 0.) ValuesDiffer = .true.
          END IF
       ELSE IF (do_CompareAsDate) THEN
          ! real KVT dates have the precision for seconds upto the microsecond
          ! XML-converted to-KVT dates have just integer seconds, 
          ! so when comparing just throw away the last 6 digits of the second
          IF (KVT_value(1:21) .ne. MPH_value(1:21)) THEN
             ValuesDiffer = .true.
          ELSE
             ! P. Poli: replace MPH_Value by KVT_Value becase time is more accurate
             ! in KVT than in XML
             MPH_value = KVT_value
          ENDIF
       ELSE
          if (trim(KVT_value) .ne. trim(MPH_value)) ValuesDiffer = .true.
       END IF

       IF (ValuesDiffer) THEN
          call logmsg(log_warn,"WARNING in Read_MPH_from_KVT():"//&
                      " XML-MPH and KVT-MPH values for the same item differ !!! "//&
                      " XML: "//trim(Item)//": "//trim(MPH_value)//&
                      " KVT: "//trim(Item)//": "//trim(KVT_value)//&
                      " keeping value from XML file for now")
       ELSE
          ! Poli: added flag do_Compare
          IF (do_Compare) MPH_value = KVT_value
       END IF
    ELSE
       ! store the result
       MPH_value = KVT_value
    END IF

    return

    ! error handlers
990 call logmsg(log_error,"ERROR in Fill_or_Compare_MPH_item():"//&
                " Item name: "//trim(Item)//&
                " Could not convert KVT-MPH value: "//trim(KVT_value)//&
                " into an integer number ....")
    error_flag = error_writing_file
    return

991 call logmsg(log_error,"ERROR in Fill_or_Compare_MPH_item():"//&
                " Item name: "//trim(Item)//&
                " Could not convert XML-MPH value: "//trim(MPH_value)//&
                " into an integer number ....")
    error_flag = error_writing_file
    return

992 call logmsg(log_error,"ERROR in Fill_or_Compare_MPH_item():"//&
                " Item name: "//trim(Item)//&
                " Could not convert KVT-MPH value: "//trim(KVT_value)//&
                " into a real number ....")
    error_flag = error_writing_file
    return

993 call logmsg(log_error,"ERROR in Fill_or_Compare_MPH_item():"//&
                " Item name: "//trim(Item)//&
                " Could not convert XML-MPH value: "//trim(MPH_value)//&
                "into a real number ....")
    error_flag = error_writing_file
    return

  END subroutine Fill_or_Compare_MPH_item
  !  #]
  !---------------------------------------------

end module MainProductHeader
