module Meas_Map
  !  #[ Documentation
  !
  !     a module to handle the Measurement Map datatype
  !     that records which measurements from the L1B data
  !     were used to produce which L2B observations.
  !     
  !     as defined in TN2.1, version 1.1 (14-2-2005), 
  !     sec...., p...., tables ...
  !
  !     Written by J. de Kloe, KNMI 
  !     created:     13-Nov-2005
  !
  !     History:
  !     28-Nov-2005 P. Poli    Added routine Write_Meas_Map
  !     10-Jan-2006 P. Poli    Added routine Read_Meas_Map
  !     30-Mar-2006 J. de Kloe Added AsciiWrite routines and added
  !                            N_Meas_to_write parameter to Write_Meas_Map
  !     12-May-2006 D. Tan     Null use of ParKind1 module
  !     16-Jan-2008 J. de Kloe phase out integer kind i_ 
  !
  !  #]
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.

  USE EE_CFI_Datatypes, only: &
       IntAuc, c_bytelen_IntAuc, maxlen_IntAuc, SetIntAuc, &
       IntAus, c_bytelen_IntAus, maxlen_IntAus, SetIntAus, &
       IntAs,  c_bytelen_IntAs,  &
       IntAl,&
       write_IntAuc_to_bin_file, read_IntAuc_from_bin_file, &
       write_IntAus_to_bin_file, read_IntAus_from_bin_file

  USE ErrorHandler, only: no_error, error_programming, error_allocate, &
                          error_writing_file
  USE LunManager, only: fileunit_stdout
  !  #]
  !  #[ type definitions

  implicit none

  TYPE Meas_Map_Bin_type
     !  #[
     ! see sec. 5.5, table 16, p. 21, of TN2.1, vers. 1.2, 06-09-2005
     ! store for each Bin of a L1B Measurement to which L2B Observation
     ! it was assigned, or 0 if it was not used. 
     integer(IntAuc) :: Which_L2B_Obs
     ! see sec. 5.5, table 17, p. 21, of TN2.1, vers. 1.2, 06-09-2005
     ! store for each Bin of a L1B Measurement to how much it contributed
     ! to the L2B Observation to which it was assigned, 
     ! This is the weight, and is a value between 0-1000.
     ! (or 0 if it was not assigned)
     integer(IntAus) :: Weight
     ! this will be floor(1000/N) of the nr of cases for this class
     ! at this level
     ! REMARK: for now we don't use this weight within the processor.
     !         We rather use a Weight stored as real in the working
     !         arrays inside the L2B_DataStructure, to prevent 
     !         any precision loss that might occur due to rounding.
     !
     ! (since each range-bin of each measurement can only be classified
     !  in a single unique way, only one weight needs to be stored,
     !  all other weights for the other possible classes are to be
     !  taken as zero)
  END TYPE Meas_Map_Bin_type
     !  #]

  TYPE Meas_Map_Row_type
     !  #[
     ! see sec. 5.5, table 21, p. 22, of TN2.1, vers. 1.2, 06-09-2005
     TYPE(Meas_Map_Bin_Type), dimension(:), pointer :: Bin
  END TYPE Meas_Map_Row_type
     !  #]

  TYPE Meas_Map_type
     !  #[
     ! see sec. 5.5, table 21, p. 22, of TN2.1, vers. 1.2, 06-09-2005
     TYPE(Meas_Map_Row_type), dimension(:), pointer :: Row
  END TYPE Meas_Map_type
     !  #]
  !  #]
contains
  !---------------------------------------------------------------------
  subroutine Init_Meas_Map_Bin(Bin)
    !  #[
    TYPE(Meas_Map_Bin_type), intent(out) :: Bin  ! output

    Bin%which_L2B_Obs = 0_IntAuc
    Bin%Weight        = 0_IntAus

  end subroutine Init_Meas_Map_Bin
    !  #]
  subroutine AsciiWrite_Meas_Map_Bin(Bin,fileunit,prefix,error_flag)
    !  #[
    TYPE(Meas_Map_Bin_type), intent(in)  :: Bin        ! input
    integer,                 intent(in)  :: fileunit   ! input
    character(len=*),        intent(in)  :: prefix     ! input
    integer,                 intent(out) :: error_flag ! output
                            
   ! local variables
    character(len=maxlen_IntAuc) :: str_IntAuc
    character(len=maxlen_IntAus) :: str_IntAus

    call SetIntAuc(Bin%which_L2B_Obs,str_IntAuc,error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%which_L2B_Obs = "//trim(str_IntAuc)

    call SetIntAus(Bin%Weight,str_IntAus,error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Weight        = "//trim(str_IntAus)
    
    return

999 print *,"ERROR in AsciiWrite_Meas_Map_Bin"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Meas_Map_Bin
    !  #]
  subroutine Meas_Map_Bin_to_Char(Bin,c)
    !  #[
    TYPE(Meas_Map_Bin_type), intent(in)  :: Bin  ! input
    character(len=1),        intent(out) :: c    ! output

    ! copy the numerical value to a 1 digit string

    ! this only works for values between 0-9, but then we don't expect
    ! to make much more than maximum 5 observations per BRC
    ! so this should not be a problem. 
    ! However, to keep the code nice and clean, I test anyway

    IF ( (Bin%which_L2B_Obs .ge. 0) .and. &
         (Bin%which_L2B_Obs .le. 9)       ) THEN
       write(c,"(i1)") Bin%which_L2B_Obs
    ELSE
       c = '?'
    END IF

  end subroutine Meas_Map_Bin_to_Char
    !  #]
  function Get_ByteSize_Meas_Map_Bin() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result
    
    c_bytesize = 1*c_bytelen_IntAuc + &
                 1*c_bytelen_IntAus     ! 2 items in total

    ! typical sizes:
    ! fortran size = 6 bytes
    ! c       size = 3 bytes

  end function Get_ByteSize_Meas_Map_Bin
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_Meas_Map_Row(Row, N_Range_Gates0, error_flag)
    !  #[
    TYPE(Meas_Map_Row_type), intent(out) :: Row            ! output
    integer(IntAs),          intent(in)  :: N_Range_Gates0 ! input
    integer,                 intent(out) :: error_flag     ! output

    ! local variable
    integer :: r, AllocStatus

    error_flag = no_error

    IF (N_Range_Gates0 .eq. 0) THEN
       ! init them to be empty
       nullify(Row%Bin)
    ELSE
       allocate(Row%Bin(N_Range_Gates0),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_Meas_Map_Row"
          error_flag = error_allocate
          return
       END IF
       
       DO r=1,N_Range_Gates0
          call Init_Meas_Map_Bin(Row%Bin(r))
       END DO
    END IF

  end subroutine Init_Meas_Map_Row
    !  #]
  subroutine AsciiWrite_Meas_Map_Row(Row,fileunit,prefix,PrintAll,error_flag)
    !  #[
    TYPE(Meas_Map_Row_type), intent(in)  :: Row        ! input
    integer,                 intent(in)  :: fileunit   ! input
    character(len=*),        intent(in)  :: prefix     ! input
    logical,                 intent(in)  :: PrintAll   ! input
    integer,                 intent(out) :: error_flag ! output

    ! local variables
    integer          :: r,s,step
    character(len=2) :: str_cntr

    error_flag = no_error

    IF (associated(Row%Bin)) THEN
       ! get the size of the array
       s = size(Row%Bin)
       IF (PrintAll .or. (s .le. 1) ) THEN
          step=1
       ELSE
          step=s-1
       END IF
       DO r=1,s,step
          write(str_cntr,"(i2.2)",err=998) r
          call AsciiWrite_Meas_Map_Bin(Row%Bin(r),fileunit,&
               trim(prefix)//"%Bin("//trim(str_cntr)//")",error_flag)
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%Bin(:) [seems not associated]"
    END IF

    return

998 print *,"conversion error in AsciiWrite_Meas_Map_Row:"
    print *,"could not convert the counter ",r," to a string .."
    error_flag = error_writing_file
    return

999 print *,"ERROR in AsciiWrite_Meas_Map_Row:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Meas_Map_Row
    !  #]
  subroutine Delete_Meas_Map_Row(Row)
    !  #[
    TYPE(Meas_Map_Row_type), intent(inout) :: Row ! deleted 

    IF (associated(Row%Bin)) deallocate(Row%Bin)

  end subroutine Delete_Meas_Map_Row
    !  #]
  function Get_ByteSize_Meas_Map_Row(Row) result(c_bytesize)
    !  #[
    TYPE(Meas_Map_Row_type), intent(in) :: Row
    integer(IntAl) :: c_bytesize ! result
    
    c_bytesize = &
         int(size(Row%Bin)*Get_ByteSize_Meas_Map_Bin(),IntAl)
    
    ! typical sizes:
    ! fortran size = 6*24 = 144 bytes
    ! c       size = 3*24 =  72 bytes

  end function Get_ByteSize_Meas_Map_Row
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_Meas_Map(Map, N_Meas0, N_Range_Gates0, error_flag)
    !  #[
    TYPE(Meas_Map_type), intent(out) :: Map           ! output
    integer(IntAs),      intent(in)  :: N_Meas0        ! input
    integer(IntAs),      intent(in)  :: N_Range_Gates0 ! input
    integer,             intent(out) :: error_flag    ! output

    ! local variable
    integer :: m, AllocStatus

    error_flag = no_error

    IF (N_Meas0 .eq. 0) THEN
       ! init them to be empty
       nullify(Map%Row)
    ELSE
       allocate(Map%Row(N_Meas0),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_Meas_Map"
          error_flag = error_allocate
          return
       END IF
       
       DO m=1,N_Meas0
          call Init_Meas_Map_Row(Map%Row(m),N_Range_Gates0, error_flag)
       END DO
    END IF

  end subroutine Init_Meas_Map
    !  #]
  subroutine Delete_Meas_Map(Map)
    !  #[
    TYPE(Meas_Map_type), intent(inout) :: Map ! deleted 

    ! local variable
    integer :: m
    
    IF (associated(Map%Row)) THEN
       DO m=1,size(Map%Row)
          call Delete_Meas_Map_Row(Map%Row(m))
       END DO
       deallocate(Map%Row)
    END IF

  end subroutine Delete_Meas_Map
    !  #]
  subroutine Print_Meas_Map(Map)
    !  #[
    TYPE(Meas_Map_type), intent(in)  :: Map ! input

    ! local variables
    integer, parameter        :: N_Meas_Max = 67
    character(len=N_Meas_Max) :: BinRowString
    character(len=N_Meas_Max) :: MeasString1
    character(len=N_Meas_Max) :: MeasString2
    character(len=N_Meas_Max) :: MeasString3
    character(len=1)          :: c
    integer                   :: r,m,a
    integer                   :: N_Meas0
    integer                   :: N_Range_Gates0

    ! init
    BinRowString(:)=' '
    MeasString1(:)=' '
    MeasString2(:)=' '
    MeasString3(:)=' '

    N_Meas0       = size(Map%Row)
    N_Range_Gates0 = size(Map%Row(1)%Bin)

    ! extra check
    IF (N_Meas0 .gt. N_Meas_Max) THEN
       print *,"Warning in Print_Meas_Map():"
       print *,"N_Meas0 in this Map seems larger than ",N_Meas_Max," !!!"
       print *,"Only printing the first ",N_Meas_Max," measurements ..."
       N_Meas0 = N_Meas_Max
    END IF
    
    DO m=1,N_Meas0 
       a=m/10
       write(MeasString1(m:m),"(i1)") a
       a=m-10*a
       write(MeasString2(m:m),"(i1)") a
       write(MeasString3(m:m),"(a1)") '-'
    END DO

    print "(7x,a)",       MeasString3(1:N_Meas0)
    print "('Meas:  ',a)",MeasString1(1:N_Meas0)
    print "(7x,a)",       MeasString2(1:N_Meas0)
    print "(7x,a)",       MeasString3(1:N_Meas0)

    DO r=1,N_Range_Gates0
       DO m=1,N_Meas0
          call Meas_Map_Bin_to_Char(Map%Row(m)%Bin(r),c)
          BinRowString(m:m) = c
       END DO
       print "('bin ',i2,':',a)",r,BinRowString(1:N_Meas0)
    END DO

    print "(7x,a)",       MeasString3(1:N_Meas0)
    print "('Meas:  ',a)",MeasString1(1:N_Meas0)
    print "(7x,a)",       MeasString2(1:N_Meas0)
    print "(7x,a)",       MeasString3(1:N_Meas0)
    
  end subroutine Print_Meas_Map
    !  #]
  subroutine AsciiWrite_Meas_Map(Map,fileunit,prefix,PrintAsMap,PrintAll,&
                                 error_flag)
    !  #[
    TYPE(Meas_Map_type), intent(in)  :: Map         ! input
    integer,             intent(in)  :: fileunit    ! input
    character(len=*),    intent(in)  :: prefix      ! input
    logical,             intent(in)  :: PrintAsMap  ! input
    logical,             intent(in)  :: PrintAll    ! input
    integer,             intent(out) :: error_flag  ! output

    ! local variables
    integer, parameter        :: N_Meas_Max = 67
    character(len=N_Meas_Max) :: BinRowString
    character(len=N_Meas_Max) :: MeasString1
    character(len=N_Meas_Max) :: MeasString2
    character(len=N_Meas_Max) :: MeasString3
    character(len=1)          :: c
    character(len=2)          :: str_cntr
    integer                   :: r,m,a,step,s
    integer                   :: N_Meas0
    integer                   :: N_Range_Gates0

    ! init
    error_flag = no_error

    IF (PrintAsMap) THEN
       ! print the map in 2 dimensional layout

       ! init
       BinRowString(:)=' '
       MeasString1(:)=' '
       MeasString2(:)=' '
       MeasString3(:)=' '
       N_Meas0        = size(Map%Row)
       N_Range_Gates0 = size(Map%Row(1)%Bin)
       
       ! extra check
       IF (N_Meas0 .gt. N_Meas_Max) THEN
          print *,"ERROR in Print_Meas_Map():"
          print *,"N_Meas0 in this Map seems larger than ",N_Meas_Max," !!!"
          error_flag = error_programming
          return
       END IF
    
       DO m=1,N_Meas0 
          a=m/10
          write(MeasString1(m:m),"(i1)",err=998) a
          a=m-10*a
          write(MeasString2(m:m),"(i1)",err=998) a
          write(MeasString3(m:m),"(a1)",err=998) '-'
       END DO
       
       write(fileunit,"(a)",          err=999) trim(prefix)//":"
       write(fileunit,"(7x,a)",       err=999) MeasString3(1:N_Meas0)
       write(fileunit,"('Meas:  ',a)",err=999) MeasString1(1:N_Meas0)
       write(fileunit,"(7x,a)",       err=999) MeasString2(1:N_Meas0)
       write(fileunit,"(7x,a)",       err=999) MeasString3(1:N_Meas0)
       
       DO r=1,N_Range_Gates0
          DO m=1,N_Meas0
             call Meas_Map_Bin_to_Char(Map%Row(m)%Bin(r),c)
             BinRowString(m:m) = c
          END DO
          write(fileunit,"('bin ',i2,':',a)",err=999) r,BinRowString(1:N_Meas0)
       END DO
       
       write(fileunit,"(7x,a)",       err=999) MeasString3(1:N_Meas0)
       write(fileunit,"('Meas:  ',a)",err=999) MeasString1(1:N_Meas0)
       write(fileunit,"(7x,a)",       err=999) MeasString2(1:N_Meas0)
       write(fileunit,"(7x,a)",       err=999) MeasString3(1:N_Meas0)
    ELSE
       ! just do a flat print of all elements
       IF (associated(Map%Row)) THEN
          ! get the size of the array
          s = size(Map%Row)
          IF (PrintAll .or. (s .le. 1) ) THEN
             step=1
          ELSE
             step=s-1
          END IF
          DO m=1,s,step
             write(str_cntr,"(i2.2)",err=998) m
             call AsciiWrite_Meas_Map_Row(Map%Row(m),fileunit,&
                  trim(prefix)//"%Row("//trim(str_cntr)//")",&
                  PrintAll,error_flag)
          END DO
       ELSE
          write(fileunit,"(a)",err=999) trim(prefix)//&
               "%Row(:) [seems not associated]"
       END IF
    END IF

    return

998 print *,"conversion error in AsciiWrite_Meas_Map:"
    print *,"could not write data to the MeasString variable ..."
    error_flag = error_writing_file
    return

999 print *,"ERROR in AsciiWrite_Meas_Map:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Meas_Map
    !  #]
  function Get_ByteSize_Meas_Map(Map,N_Meas_to_write) result(c_bytesize)
    !  #[
    TYPE(Meas_Map_type), intent(in) :: Map
    integer(IntAuc),     intent(in) :: N_Meas_to_write
    integer(IntAl)                  :: c_bytesize ! result

    ! local variable
    integer :: s
    
    c_bytesize = 0
    IF (associated(Map%Row)) THEN
       s=size(Map%Row)
       IF (N_Meas_to_write .gt. s) THEN
          print *,"WARNING in Get_ByteSize_Meas_Map()"
          print *,"invalid value for N_Meas_to_write: ",N_Meas_to_write
          print *,"max. possible value is: size(Map%Row) = ",s
          print *,"Get_ByteSize result will be useless...."
       END IF

       c_bytesize = &
            N_Meas_to_write*Get_ByteSize_Meas_Map_Row(Map%Row(1))
    END IF

    ! typical sizes:
    ! fortran size = 144*67 = 9648 bytes
    ! c       size =  72*67 = 4824 bytes

  end function Get_ByteSize_Meas_Map
    !  #]
  function Count_Nr_Points_On_Level(Map,r,Obs_Number) result(nr_of_points)
    !  #[ 
    ! Modifications
    !    P. Poli    30-11-2005 Changed variable name Obs_Type --> Obs_Number
    TYPE(Meas_Map_type), intent(in) :: Map        ! input
    integer,             intent(in) :: r          ! input
    integer,             intent(in) :: Obs_Number ! input
    integer                         :: nr_of_points ! result
 
    ! local variables
    integer :: m

    ! the counts are not stored as such, so use a little loop 
    ! to construct the numbers
    nr_of_points = 0
    do m=1,size(Map%Row)
       ! count the nr of rangebins on this level that contributed
       ! to the requested observation type
       if (Map%Row(m)%Bin(r)%Which_L2B_Obs .eq. Obs_Number) &
            nr_of_points = nr_of_points + 1
    end do

  end function Count_Nr_Points_On_Level
  !  #]
  !---------------------------------------------------------------------
  subroutine Write_Meas_Map(fileunit,Map,N_Meas_to_write,error_flag)
    !  #[
    ! 28-11-2005 P. Poli    Original code. From TN2.1 v1.2 06-09-2005
    ! 30-03-2006 J. de Kloe Added N_Meas_to_write
    integer,             intent(in)  :: fileunit
    TYPE(Meas_Map_type), intent(in)  :: Map
    integer(IntAuc),     intent(in)  :: N_Meas_to_write
    integer,             intent(out) :: error_flag

    ! local variables 
    integer, parameter :: N_Meas_Max = 67
    integer            :: N_Meas0,N_Range_Gates0,m,r

    error_flag = no_error
    IF (.not.(associated(Map%Row))) THEN
       print *,"ERROR in Write_Meas_Map():"
       print *,"The pointer to this Map seems not associated !!"
       error_flag = error_programming
       goto 998
    END IF

    N_Meas0       = size(Map%Row)
    N_Range_Gates0 = size(Map%Row(1)%Bin)

    print *,"Write_Meas_Map: N_Meas=",N_Meas0
    print *,"Write_Meas_Map: N_Range_Bins=",N_Range_Gates0

    ! extra check
    IF (N_Meas0 .gt. N_Meas_Max) THEN
       print *,"Error in Write_Meas_Map():"
       print *,"N_Meas0 in this Map seems larger than ",N_Meas_Max," !!!"
       print *,"N_Meas0 = ",N_Meas0
       error_flag = error_programming
       goto 998
    END IF
  
    IF (N_Meas_to_write .gt. N_Meas0) THEN
       print *,"ERROR in Write_Meas_Map():"
       print *,"invalid value for N_Meas_to_write: ",N_Meas_to_write
       print *,"max. possible value is: size(Map%Row) = ",N_Meas0
       error_flag = error_programming
       goto 998
    END IF


    ! write Map_of_L1_Measurements_Used 
    DO r=1,N_Range_Gates0
       DO m=1,N_Meas_to_write
          call write_IntAuc_to_bin_file(fileunit, &
               Map%Row(m)%Bin(r)%Which_L2B_Obs,error_flag)
          IF (error_flag .ne. no_error) goto 998
       END DO
    END DO

    ! write List_of_L1_Measurements_Weight
    DO r=1,N_Range_Gates0
       DO m=1,N_Meas_to_write
          call write_IntAus_to_bin_file(fileunit, &
               Map%Row(m)%Bin(r)%Weight,error_flag)
          IF (error_flag .ne. no_error) goto 998
       END DO
    END DO

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Write_Meas_Map"
    return


  end subroutine Write_Meas_Map
  !  #]
  subroutine Read_Meas_Map(fileunit,Map,error_flag)
    !  #[
    ! 10-01-2006 P. Poli  Original code. From TN2.1 v1.2 06-09-2005
    integer,             intent(in)    :: fileunit
    TYPE(Meas_Map_type), intent(inout) :: Map
    integer,             intent(out)   :: error_flag

    ! local variables 
    integer, parameter :: N_Meas_Max = 67
    integer            :: N_Meas0,N_Range_Gates0,m,r

    error_flag = no_error
    IF (.not.(associated(Map%Row))) THEN
       print *,"ERROR in Read_Meas_Map():"
       print *,"The pointer to this Map seems not associated !!"
       error_flag = error_programming
       goto 998
    END IF

    N_Meas0       = size(Map%Row)
    N_Range_Gates0 = size(Map%Row(1)%Bin)

    print *,"Read_Meas_Map: N_Meas=",N_Meas0
    print *,"Read_Meas_Map: N_Range_Bins=",N_Range_Gates0

    ! extra check
    IF (N_Meas0 .gt. N_Meas_Max) THEN
       print *,"Warning in Read_Meas_Map():"
       print *,"N_Meas0 in this Map seems larger than ",N_Meas_Max," !!!"
       print *,"N_Meas0 = ",N_Meas0
       error_flag = error_programming
       goto 998
    END IF
  
    ! write Map_of_L1_Measurements_Used 
    DO r=1,N_Range_Gates0
       DO m=1,N_Meas0
          call read_IntAuc_from_bin_file(fileunit, &
               Map%Row(m)%Bin(r)%Which_L2B_Obs,error_flag)
          IF (error_flag .ne. no_error) goto 998
       END DO
    END DO

    ! write List_of_L1_Measurements_Weight
    DO r=1,N_Range_Gates0
       DO m=1,N_Meas0
          call read_IntAus_from_bin_file(fileunit, &
               Map%Row(m)%Bin(r)%Weight,error_flag)
          IF (error_flag .ne. no_error) goto 998
       END DO
    END DO

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Read_Meas_Map"
    return


  end subroutine Read_Meas_Map
  !  #]
  !---------------------------------------------------------------------
end module Meas_Map
