module RBC_DataStructure
  !  #[ Description
  !     a module to define the datastructure used by the L2B processor,
  !     top store the RayleighBrillouin correction look-up tables,
  !     and the routines needed for initialisation, allocation
  !     and deallocation of all data structs inside it.
  !     
  !     created:    15-Feb-2006, by: J. de Kloe
  !
  !     Modifications
  !     16-Feb-2006 J. de Kloe  extended to use MPH
  !     22-Feb-2006 J. de Kloe  extended to use SPH
  !     27-Feb-2006 J. de Kloe  extended with RBC_DS
  !     01-Mar-2006 J. de Kloe  added writing subroutines
  !     02-Mar-2006 J. de Kloe  added reading subroutines
  !     03-Mar-2006 J. de Kloe  added buffered reading subroutines
  !     13-Mar-2006 J. de Kloe  added error_flag to Print_RBC_SPH routine
  !     24-Mar-2006 D. Tan      Add spare for error parameter (IODD V1.0)
  !     03-May-2006 D. Tan      Rename T1_FP to TA_FP, T2_FP to TB_FP.
  !     12-May-2006 D. Tan      Null use of ParKind1 module
  !     16-Jun-2006 J. de Kloe  added some REALACC/ENDREALACC commands
  !     22-Jun-2006 P. Poli     Adding Na_Fd and Nb_Fd, Fint_R
  !     22-Aug-2006 D. Tan      IODD order for Na_Fd and Nb_Fd i/o
  !     04-Apr-2007 J. de Kloe  clean up of USE FixedHeader
  !     16-Jan-2008 J. de Kloe  phase out integer kind i_
  !     21-Jan-2008 J. de kloe  added workaround in call to GetByteSize_MPH_KVT
  !                             to not change the fileformat for now.
  !     21-Feb-2008 J. de Kloe  removed this workaround agian, since it was
  !                             caused by a misunderstanding on L1B side
  !
  !  #]
  !  #[ modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE FixedHeader, only: FH_type, Init_FH, Print_FH, Fill_FH_items
  USE MainProductHeader, only: MPH_type, Init_MPH, Print_MPH, &
                               Fill_MPH_items, GetByteSize_MPH_KVT
  USE RBC_SPH, only: RBC_SPH_type, Init_RBC_SPH, Fill_RBC_SPH_items, &
                     Delete_RBC_SPH, Print_RBC_SPH, GetByteSize_RBC_SPH_KVT

  ! support modules
  USE EE_CFI_Datatypes, only: DateTimeType, & !Construct_KVT_DateTime, &
       InitDateTime, SetDateTime, PrintDateTime, &
       FAdoxy, c_bytelen_fadoxy, &
       IntAl,  IntAs, &
       IntAus, maxval_IntAus, maxlen_IntAus, SetIntAus, c_bytelen_IntAus, &
       IntAul, maxval_IntAul, maxlen_IntAul, SetIntAul, c_bytelen_IntAul, &
       IntAd,  maxval_IntAd,  maxlen_IntAd,  SetIntAd,  c_bytelen_IntAd,  &
       write_IntAus_to_bin_file, read_IntAus_from_bin_file, &
       write_IntAul_to_bin_file, read_IntAul_from_bin_file, &
       write_IntAd_to_bin_file,  read_IntAd_from_bin_file, &
       write_FAdoxy_to_bin_file, read_FAdoxy_from_bin_file, &
       BufferedFile_type,&
       read_IntAus_from_buffbin_file, &
       read_IntAul_from_buffbin_file, &
       read_IntAd_from_buffbin_file,  &
       read_FAdoxy_from_buffbin_file
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE numerics, only: missing_indicator_real_r8
  USE ErrorHandler, only: no_error, error_programming, error_allocate
  USE Compiler_Features, only: newline

  !  #]
  !  #[ Parameter and data type definitions
  implicit none

  integer(IntAul), parameter :: missing_indicator_IntAul = maxval_IntAul
  integer(IntAus), parameter :: missing_indicator_IntAus = maxval_IntAus
  integer(IntAd),  parameter :: missing_indicator_IntAd  = maxval_IntAd
  real(FAdoxy),    parameter :: missing_indicator_FAdoxy = &
                                missing_indicator_real_r8

  ! see TN2.1, sec. 6.3.2, table 67
  TYPE Spec_Grid_TF_type
     real(FAdoxy), dimension(:), pointer :: Spec_Grid_F
  END TYPE Spec_Grid_TF_type

  ! see TN2.1, sec. 6.3.2, table 66
  TYPE Spec_Grid_PTF_type
     TYPE(Spec_Grid_TF_type), dimension(:), pointer :: Spec_Grid_TF
  END TYPE Spec_Grid_PTF_type

  ! see TN2.1, sec. 6.3.2, table 69
  TYPE Fcalib_TR_type
     real(FAdoxy), dimension(:), pointer :: Fcalib_R
     real(FAdoxy), dimension(:), pointer :: Fcalib_R_error
  END TYPE Fcalib_TR_type

  ! see TN2.1, sec. 6.3.2, table 68
  TYPE Fcalib_PTR_type
     TYPE(Fcalib_TR_type), dimension(:), pointer :: Fcalib_TR
  END TYPE Fcalib_PTR_type

  ! PP: addition
  TYPE Nab_TFd_type
     integer(IntAul), dimension(:), pointer :: Na_Fd
     integer(IntAul), dimension(:), pointer :: Nb_Fd
  END TYPE Nab_TFd_type

  ! PP: addition
  TYPE Nab_PTFd_type
     TYPE(Nab_TFd_type), dimension(:), pointer :: Nab_TFd
  END TYPE Nab_PTFd_type

  ! see TN2.1, sec. 6.3.2, table 65
  TYPE RBC_DS_Type
     integer(IntAul),          dimension(:), pointer :: P_Grid
     integer(IntAus),          dimension(:), pointer :: T_Grid
     integer(IntAd),           dimension(:), pointer :: F_Gridtmp
     TYPE(Spec_Grid_PTF_type), dimension(:), pointer :: Spec_Grid_PTF
     integer(IntAd),           dimension(:), pointer :: F_FP
     real(FAdoxy),             dimension(:), pointer :: TA_FP
     real(FAdoxy),             dimension(:), pointer :: TB_FP
     integer(IntAd),           dimension(:), pointer :: fd
     real(FAdoxy),             dimension(:), pointer :: RR
     TYPE(Fcalib_PTR_type),    dimension(:), pointer :: Fcalib_PTR
     TYPE(Nab_PTFd_type),      dimension(:), pointer :: Nab_PTFd
     integer(IntAd),           dimension(:), pointer :: Fint_R
  END TYPE RBC_DS_Type

  TYPE RBC_DataType
     ! default headers needed for all EE_CFI type files
     TYPE(FH_type),      pointer :: FH
     TYPE(MPH_type),     pointer :: MPH
     TYPE(RBC_SPH_type), pointer :: SPH
     !
     ! the attached datasets
     TYPE(RBC_DS_type), pointer :: RBC_DS
  END TYPE RBC_DataType

  ! some often used strings
  character(len=*), parameter:: firstlast = " [printing first and last one]"
  character(len=*), parameter:: not_associated = "( : ) [seems not associated]"
  character(len=*), parameter:: not_associated2 = " [seems not associated]"

  ! the default verbosity setting is Err (as silent as possible)
  integer, private, save :: LocalVerbosity = VerbosityError

  !  #]
  !  #[ Module interface
  !     everything is public in this module,
  !     unless explicitely given the private keyword below
  private :: PrintSizeAndGetStep
  !  #]
contains
  !---------------------------------------------
  subroutine Init_RBC_DataStruct_Module(ChoosenVerbosity)
    !  #[
    integer, intent(in), optional :: ChoosenVerbosity

    IF (present(ChoosenVerbosity)) LocalVerbosity = ChoosenVerbosity

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"RBC_DataStructure module has been initialised"
       call print_verbosity_setting(LocalVerbosity,&
                                    ModuleName="RBC_DataStructure")
    END IF

  end subroutine Init_RBC_DataStruct_Module
    !  #]
  !---------------------------------------------
  subroutine Init_Spec_Grid_TF(Spec_Grid_TF, Num_F, error_flag)
    !  #[
    TYPE(Spec_Grid_TF_type), intent(out) :: Spec_Grid_TF ! result
    integer(IntAs),          intent(in)  :: Num_F
    integer,                 intent(out) :: error_flag

    ! local variable
    integer :: AllocStatus

    error_flag = no_error
    IF (Num_F .eq. 0) THEN
       ! init the struct to hold no data
       nullify(Spec_Grid_TF%Spec_Grid_F)
    ELSE
       allocate(Spec_Grid_TF%Spec_Grid_F(1:Num_F),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_Spec_Grid_TF()"
          error_flag = error_allocate
          return
       END IF
       Spec_Grid_TF%Spec_Grid_F(:) = missing_indicator_real_r8
    END IF

  end subroutine Init_Spec_Grid_TF
    !  #]
  subroutine Delete_Spec_Grid_TF(Spec_Grid_TF)
    !  #[
    TYPE(Spec_Grid_TF_type), intent(inout) :: Spec_Grid_TF

    IF (associated(Spec_Grid_TF%Spec_Grid_F)) &
         deallocate(Spec_Grid_TF%Spec_Grid_F)

  end subroutine Delete_Spec_Grid_TF
    !  #]
  subroutine Print_Spec_Grid_TF(Spec_Grid_TF,prefix,error_flag,PrintAll)
    !  #[
    TYPE(Spec_Grid_TF_type), intent(in) :: Spec_Grid_TF
    character(len=*),  intent(in)    :: prefix
    integer,           intent(out)   :: error_flag
    logical, optional, intent(in)    :: PrintAll

    ! local variable
    integer            :: s, step, i
    character(len=256) :: text
    character(len=15)  :: name

    error_flag = no_error

    name = "%Spec_Grid_F"
    IF (associated(Spec_Grid_TF%Spec_Grid_F)) THEN
       ! get the size of the array
       s = size(Spec_Grid_TF%Spec_Grid_F)
       call PrintSizeAndGetStep(prefix,name,s,step,PrintAll)
       print *,"REALACC(12)"
       DO i=1,s,step
          write(text,"(3a,i3,a)") trim(prefix),trim(name),"(",i,") = "
          print *,trim(text),Spec_Grid_TF%Spec_Grid_F(i)
       END DO
       print *,"ENDREALACC"
    ELSE
       print *,trim(prefix)//trim(name)//trim(not_associated)
    END IF

  end subroutine Print_Spec_Grid_TF
    !  #]
  function GetByteSize_Spec_Grid_TF(Spec_Grid_TF) result(c_bytesize)
    !  #[
    TYPE(Spec_Grid_TF_type), intent(in) :: Spec_Grid_TF
    integer(IntAl)                      :: c_bytesize ! result

    c_bytesize = 0
    IF (associated(Spec_Grid_TF%Spec_Grid_F)) THEN
       c_bytesize = c_bytesize + &
            int(size(Spec_Grid_TF%Spec_Grid_F) * c_bytelen_FAdoxy,IntAl)
    ELSE
       print *,"WARNING in GetByteSize_Spec_Grid_TF():"
       print *,"Spec_Grid_TF%Spec_Grid_F [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

  end function GetByteSize_Spec_Grid_TF
    !  #]
  subroutine write_Spec_Grid_TF_to_bin_file(fileunit,Spec_Grid_TF,error_flag)
    !  #[
    integer,                 intent(in)  :: fileunit
    TYPE(Spec_Grid_TF_type), intent(in)  :: Spec_Grid_TF
    integer,                 intent(out) :: error_flag

    ! local variables
    integer :: i,s
    
    error_flag = no_error

    ! write the Spec_Grid_F array 
    s = size(Spec_Grid_TF%Spec_Grid_F)
    DO i=1,s
       call write_FAdoxy_to_bin_file(fileunit,&
                  Spec_Grid_TF%Spec_Grid_F(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine write_Spec_Grid_TF_to_bin_file
    !  #]
  subroutine read_Spec_Grid_TF_from_bin_file(fileunit,Spec_Grid_TF,error_flag)
    !  #[
    integer,                 intent(in)    :: fileunit
    TYPE(Spec_Grid_TF_type), intent(inout) :: Spec_Grid_TF
    integer,                 intent(out)   :: error_flag

    ! local variables
    integer :: i,s
    
    error_flag = no_error

    ! read the Spec_Grid_F array 
    s = size(Spec_Grid_TF%Spec_Grid_F)
    DO i=1,s
       call read_FAdoxy_from_bin_file(fileunit,&
                  Spec_Grid_TF%Spec_Grid_F(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine read_Spec_Grid_TF_from_bin_file
    !  #]
  subroutine read_Spec_Grid_TF_buffbin_file(BufferedFile,Spec_Grid_TF,&
                                            error_flag)
    !  #[
    TYPE(BufferedFile_type), pointer       :: BufferedFile
    TYPE(Spec_Grid_TF_type), intent(inout) :: Spec_Grid_TF
    integer,                 intent(out)   :: error_flag

    ! local variables
    integer :: i,s
    
    error_flag = no_error

    ! read the Spec_Grid_F array 
    s = size(Spec_Grid_TF%Spec_Grid_F)
    DO i=1,s
       call read_FAdoxy_from_buffbin_file(BufferedFile,&
                  Spec_Grid_TF%Spec_Grid_F(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine read_Spec_Grid_TF_buffbin_file
    !  #]
  !---------------------------------------------
  subroutine Init_Spec_Grid_PTF(Spec_Grid_PTF, Num_T, Num_F, error_flag)
    !  #[
    TYPE(Spec_Grid_PTF_type), intent(out) :: Spec_Grid_PTF ! result
    integer(IntAs),           intent(in)  :: Num_T, Num_F
    integer,                  intent(out) :: error_flag

    ! local variable
    integer :: i, AllocStatus

    error_flag = no_error
    IF (Num_T .eq. 0) THEN
       ! init the struct to hold no data
       nullify(Spec_Grid_PTF%Spec_Grid_TF)
    ELSE
       allocate(Spec_Grid_PTF%Spec_Grid_TF(1:Num_T),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_Spec_Grid_PTF()"
          error_flag = error_allocate
          return
       END IF
       DO i=1,Num_T
          call Init_Spec_Grid_TF(Spec_Grid_PTF%Spec_Grid_TF(i),&
                                 Num_F, error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    END IF

  end subroutine Init_Spec_Grid_PTF
    !  #]
  subroutine Delete_Spec_Grid_PTF(Spec_Grid_PTF)
    !  #[
    TYPE(Spec_Grid_PTF_type), intent(inout) :: Spec_Grid_PTF

    ! local variable
    integer :: i

    IF (associated(Spec_Grid_PTF%Spec_Grid_TF)) THEN
       DO i=1,size(Spec_Grid_PTF%Spec_Grid_TF)
          call Delete_Spec_Grid_TF(Spec_Grid_PTF%Spec_Grid_TF(i))
       END DO
       deallocate(Spec_Grid_PTF%Spec_Grid_TF)
    END IF
  
  end subroutine Delete_Spec_Grid_PTF
    !  #]
  subroutine Print_Spec_Grid_PTF(Spec_Grid_PTF,prefix,error_flag,PrintAll)
    !  #[
    TYPE(Spec_Grid_PTF_type), intent(in) :: Spec_Grid_PTF
    character(len=*),  intent(in)    :: prefix
    integer,           intent(out)   :: error_flag
    logical, optional, intent(in)    :: PrintAll

    ! local variable
    integer            :: s, step, i
    character(len=256) :: text
    character(len=15)  :: name

    error_flag = no_error

    name = "%Spec_Grid_TF"
    IF (associated(Spec_Grid_PTF%Spec_Grid_TF)) THEN
       ! get the size of the array
       s = size(Spec_Grid_PTF%Spec_Grid_TF)
       call PrintSizeAndGetStep(prefix,name,s,step,PrintAll)
       DO i=1,s,step
          write(text,"(3a,i3,a)") trim(prefix),trim(name),"(",i,")"
          call Print_Spec_Grid_TF(Spec_Grid_PTF%Spec_Grid_TF(i),text,&
                                  error_flag,PrintAll)
       END DO
    ELSE
       print *,trim(prefix)//trim(name)//trim(not_associated)
    END IF

  end subroutine Print_Spec_Grid_PTF
    !  #]
  function GetByteSize_Spec_Grid_PTF(Spec_Grid_PTF) result(c_bytesize)
    !  #[
    TYPE(Spec_Grid_PTF_type), intent(in) :: Spec_Grid_PTF
    integer(IntAl)                       :: c_bytesize ! result

    c_bytesize = 0
    IF (associated(Spec_Grid_PTF%Spec_Grid_TF)) THEN
       c_bytesize = c_bytesize + &
            int(size(Spec_Grid_PTF%Spec_Grid_TF) * &
                GetByteSize_Spec_Grid_TF(Spec_Grid_PTF%Spec_Grid_TF(1)),IntAl)
    ELSE
       print *,"WARNING in GetByteSize_Spec_Grid_PTF():"
       print *,"Spec_Grid_PTF%Spec_Grid_TF [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

  end function GetByteSize_Spec_Grid_PTF
    !  #]
  subroutine write_Spec_Grid_PTF_to_bin_file(fileunit,Spec_Grid_PTF,error_flag)
    !  #[
    integer,                  intent(in)  :: fileunit
    TYPE(Spec_Grid_PTF_type), intent(in)  :: Spec_Grid_PTF
    integer,                  intent(out) :: error_flag

    ! local variables
    integer :: i,s
    
    error_flag = no_error

    ! write the Spec_Grid_TF array of structs
    s = size(Spec_Grid_PTF%Spec_Grid_TF)
    DO i=1,s
       call write_Spec_Grid_TF_to_bin_file(fileunit,&
                  Spec_Grid_PTF%Spec_Grid_TF(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine write_Spec_Grid_PTF_to_bin_file
    !  #]
  subroutine read_SpecGridPTF_from_bin_file(fileunit,Spec_Grid_PTF,error_flag)
    !  #[
    integer,                  intent(in)    :: fileunit
    TYPE(Spec_Grid_PTF_type), intent(inout) :: Spec_Grid_PTF
    integer,                  intent(out)   :: error_flag

    ! local variables
    integer :: i,s
    
    error_flag = no_error

    ! read the Spec_Grid_TF array of structs
    s = size(Spec_Grid_PTF%Spec_Grid_TF)
    DO i=1,s
       call read_Spec_Grid_TF_from_bin_file(fileunit,&
            Spec_Grid_PTF%Spec_Grid_TF(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine read_SpecGridPTF_from_bin_file
    !  #]
  subroutine read_SpecGridPTF_buffbin_file(BufferedFile,Spec_Grid_PTF,&
                                           error_flag)
    !  #[
    TYPE(BufferedFile_type),  pointer       :: BufferedFile
    TYPE(Spec_Grid_PTF_type), intent(inout) :: Spec_Grid_PTF
    integer,                  intent(out)   :: error_flag

    ! local variables
    integer :: i,s
    
    error_flag = no_error

    ! write the Spec_Grid_TF array of structs
    s = size(Spec_Grid_PTF%Spec_Grid_TF)
    DO i=1,s
       call read_Spec_Grid_TF_buffbin_file(BufferedFile,&
            Spec_Grid_PTF%Spec_Grid_TF(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine read_SpecGridPTF_buffbin_file
    !  #]
  !---------------------------------------------
  subroutine Init_Nab_TFd(Nab_TFd, Num_fd, error_flag)
    !  #[
    TYPE(Nab_TFd_type),      intent(out) :: Nab_TFd ! result
    integer(IntAs),          intent(in)  :: Num_fd
    integer,                 intent(out) :: error_flag

    ! local variable
    integer :: AllocStatus

    error_flag = no_error
    IF (Num_fd .eq. 0) THEN
       ! init the struct to hold no data
       nullify(Nab_TFd%Na_Fd,Nab_TFd%Nb_Fd)
    ELSE
       allocate(Nab_TFd%Na_Fd(1:Num_fd),&
                Nab_TFd%Nb_Fd(1:Num_fd),&
                STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_Nab_TFd()"
          error_flag = error_allocate
          return
       END IF
       Nab_TFd%Na_Fd(:) = missing_indicator_IntAul
       Nab_TFd%Nb_Fd(:) = missing_indicator_IntAul
    END IF

  end subroutine Init_Nab_TFd
    !  #]
  subroutine Delete_Nab_TFd(Nab_TFd)
    !  #[
    TYPE(Nab_TFd_type), intent(inout) :: Nab_TFd

    IF (associated(Nab_TFd%Na_Fd)) &
         deallocate(Nab_TFd%Na_Fd)
    IF (associated(Nab_TFd%Nb_Fd)) &
         deallocate(Nab_TFd%Nb_Fd)

  end subroutine Delete_Nab_TFd
    !  #]
  subroutine Print_Nab_TFd(Nab_TFd,prefix,error_flag,PrintAll)
    !  #[
    TYPE(Nab_TFd_type), intent(in)   :: Nab_TFd
    character(len=*),  intent(in)    :: prefix
    integer,           intent(out)   :: error_flag
    logical, optional, intent(in)    :: PrintAll

    ! local variable
    integer            :: s, step, i
    character(len=256) :: text
    character(len=15)  :: name

    error_flag = no_error

    name = "%Na_Fd"
    IF (associated(Nab_TFd%Na_Fd)) THEN
       ! get the size of the array
       s = size(Nab_TFd%Na_Fd)
       call PrintSizeAndGetStep(prefix,name,s,step,PrintAll)
       print *,"REALACC(12)"
       DO i=1,s,step
          write(text,"(3a,i3,a)") trim(prefix),trim(name),"(",i,") = "
          print *,trim(text),Nab_TFd%Na_Fd(i)
       END DO
       print *,"ENDREALACC"
    ELSE
       print *,trim(prefix)//trim(name)//trim(not_associated)
    END IF

    name = "%Nb_Fd"
    IF (associated(Nab_TFd%Nb_Fd)) THEN
       ! get the size of the array
       s = size(Nab_TFd%Nb_Fd)
       call PrintSizeAndGetStep(prefix,name,s,step,PrintAll)
       print *,"REALACC(12)"
       DO i=1,s,step
          write(text,"(3a,i3,a)") trim(prefix),trim(name),"(",i,") = "
          print *,trim(text),Nab_TFd%Nb_Fd(i)
       END DO
       print *,"ENDREALACC"
    ELSE
       print *,trim(prefix)//trim(name)//trim(not_associated)
    END IF

  end subroutine Print_Nab_TFd
    !  #]
  function GetByteSize_Nab_TFd(Nab_TFd) result(c_bytesize)
    !  #[
    TYPE(Nab_TFd_type), intent(in) :: Nab_TFd
    integer(IntAl)                 :: c_bytesize ! result

    c_bytesize = 0
    IF (associated(Nab_TFd%Na_Fd)) THEN
       c_bytesize = c_bytesize + &
            int(size(Nab_TFd%Na_Fd) * c_bytelen_IntAul,IntAl)
    ELSE
       print *,"WARNING in GetByteSize_Nab_TFd():"
       print *,"Nab_TFd%Na_Fd [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

    IF (associated(Nab_TFd%Nb_Fd)) THEN
       c_bytesize = c_bytesize + &
            int(size(Nab_TFd%Nb_Fd) * c_bytelen_IntAul,IntAl)
    ELSE
       print *,"WARNING in GetByteSize_Nab_TFd():"
       print *,"Nab_TFd%Nb_Fd [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

  end function GetByteSize_Nab_TFd
    !  #]
  subroutine write_Nab_TFd_to_bin_file(fileunit,Nab_TFd,error_flag)
    !  #[
    integer,                 intent(in)  :: fileunit
    TYPE(Nab_TFd_type), intent(in)  :: Nab_TFd
    integer,                 intent(out) :: error_flag

    ! local variables
    integer :: i,s
    
    error_flag = no_error

    ! write the Na_Fd array 
    s = size(Nab_TFd%Na_Fd)
    DO i=1,s
       call write_IntAul_to_bin_file(fileunit,&
                  Nab_TFd%Na_Fd(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO
    ! write the Nb_Fd array 
    DO i=1,s
       call write_IntAul_to_bin_file(fileunit,&
                  Nab_TFd%Nb_Fd(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine write_Nab_TFd_to_bin_file
    !  #]
  subroutine read_Nab_TFd_from_bin_file(fileunit,Nab_TFd,error_flag)
    !  #[
    integer,                 intent(in)    :: fileunit
    TYPE(Nab_TFd_type), intent(inout) :: Nab_TFd
    integer,                 intent(out)   :: error_flag

    ! local variables
    integer :: i,s
    
    error_flag = no_error

    ! read the Na_Fd array 
    s = size(Nab_TFd%Na_Fd)
    DO i=1,s
       call read_IntAul_from_bin_file(fileunit,&
                  Nab_TFd%Na_Fd(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO
    ! read the Nb_Fd array 
    DO i=1,s
       call read_IntAul_from_bin_file(fileunit,&
                  Nab_TFd%Nb_Fd(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine read_Nab_TFd_from_bin_file
    !  #]
  subroutine read_Nab_TFd_buffbin_file(BufferedFile,Nab_TFd,&
                                            error_flag)
    !  #[
    TYPE(BufferedFile_type), pointer       :: BufferedFile
    TYPE(Nab_TFd_type),      intent(inout) :: Nab_TFd
    integer,                 intent(out)   :: error_flag

    ! local variables
    integer :: i,s
    
    error_flag = no_error

    ! read the Na_Fd array 
    s = size(Nab_TFd%Na_Fd)
    DO i=1,s
       call read_IntAul_from_buffbin_file(BufferedFile,&
                  Nab_TFd%Na_Fd(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO
    ! read the Nb_Fd array 
    DO i=1,s
       call read_IntAul_from_buffbin_file(BufferedFile,&
                  Nab_TFd%Nb_Fd(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine read_Nab_TFd_buffbin_file
    !  #]
  !---------------------------------------------
  subroutine Init_Nab_PTFd(Nab_PTFd, Num_T, Num_fd, error_flag)
    !  #[
    TYPE(Nab_PTFd_type),      intent(out) :: Nab_PTFd ! result
    integer(IntAs),           intent(in)  :: Num_T, Num_fd
    integer,                  intent(out) :: error_flag

    ! local variable
    integer :: i, AllocStatus

    error_flag = no_error
    IF (Num_T .eq. 0) THEN
       ! init the struct to hold no data
       nullify(Nab_PTFd%Nab_TFd)
    ELSE
       allocate(Nab_PTFd%Nab_TFd(1:Num_T),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_Nab_PTFd()"
          error_flag = error_allocate
          return
       END IF
       DO i=1,Num_T
          call Init_Nab_TFd(Nab_PTFd%Nab_TFd(i),&
                                 Num_fd, error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    END IF

  end subroutine Init_Nab_PTFd
    !  #]
  subroutine Delete_Nab_PTFd(Nab_PTFd)
    !  #[
    TYPE(Nab_PTFd_type), intent(inout) :: Nab_PTFd

    ! local variable
    integer :: i

    IF (associated(Nab_PTFd%Nab_TFd)) THEN
       DO i=1,size(Nab_PTFd%Nab_TFd)
          call Delete_Nab_TFd(Nab_PTFd%Nab_TFd(i))
       END DO
       deallocate(Nab_PTFd%Nab_TFd)
    END IF
  
  end subroutine Delete_Nab_PTFd
    !  #]
  subroutine Print_Nab_PTFd(Nab_PTFd,prefix,error_flag,PrintAll)
    !  #[
    TYPE(Nab_PTFd_type), intent(in) :: Nab_PTFd
    character(len=*),  intent(in)    :: prefix
    integer,           intent(out)   :: error_flag
    logical, optional, intent(in)    :: PrintAll

    ! local variable
    integer            :: s, step, i
    character(len=256) :: text
    character(len=15)  :: name

    error_flag = no_error

    name = "%Nab_TFd"
    IF (associated(Nab_PTFd%Nab_TFd)) THEN
       ! get the size of the array
       s = size(Nab_PTFd%Nab_TFd)
       call PrintSizeAndGetStep(prefix,name,s,step,PrintAll)
       DO i=1,s,step
          write(text,"(3a,i3,a)") trim(prefix),trim(name),"(",i,")"
          call Print_Nab_TFd(Nab_PTFd%Nab_TFd(i),text,&
                                  error_flag,PrintAll)
       END DO
    ELSE
       print *,trim(prefix)//trim(name)//trim(not_associated)
    END IF

  end subroutine Print_Nab_PTFd
    !  #]
  function GetByteSize_Nab_PTFd(Nab_PTFd) result(c_bytesize)
    !  #[
    TYPE(Nab_PTFd_type), intent(in) :: Nab_PTFd
    integer(IntAl)                  :: c_bytesize ! result

    c_bytesize = 0
    IF (associated(Nab_PTFd%Nab_TFd)) THEN
       c_bytesize = c_bytesize + &
            int(size(Nab_PTFd%Nab_TFd) * &
                GetByteSize_Nab_TFd(Nab_PTFd%Nab_TFd(1)),IntAl)
    ELSE
       print *,"WARNING in GetByteSize_Nab_PTFd():"
       print *,"Nab_PTFd%Nab_TFd [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

  end function GetByteSize_Nab_PTFd
    !  #]
  subroutine write_Nab_PTFd_to_bin_file(fileunit,Nab_PTFd,error_flag)
    !  #[
    integer,                  intent(in)  :: fileunit
    TYPE(Nab_PTFd_type), intent(in)  :: Nab_PTFd
    integer,                  intent(out) :: error_flag

    ! local variables
    integer :: i,s
    
    error_flag = no_error

    ! write the Nab_TFd array of structs
    s = size(Nab_PTFd%Nab_TFd)
    DO i=1,s
       call write_Nab_TFd_to_bin_file(fileunit,&
                  Nab_PTFd%Nab_TFd(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine write_Nab_PTFd_to_bin_file
    !  #]
  subroutine read_NabPTFd_from_bin_file(fileunit,Nab_PTFd,error_flag)
    !  #[
    integer,                  intent(in)    :: fileunit
    TYPE(Nab_PTFd_type), intent(inout) :: Nab_PTFd
    integer,                  intent(out)   :: error_flag

    ! local variables
    integer :: i,s
    
    error_flag = no_error

    ! write the Nab_TFd array of structs
    s = size(Nab_PTFd%Nab_TFd)
    DO i=1,s
       call read_Nab_TFd_from_bin_file(fileunit,&
            Nab_PTFd%Nab_TFd(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine read_NabPTFd_from_bin_file
    !  #]
  subroutine read_NabPTFd_buffbin_file(BufferedFile,Nab_PTFd,&
                                           error_flag)
    !  #[
    TYPE(BufferedFile_type),  pointer       :: BufferedFile
    TYPE(Nab_PTFd_type), intent(inout) :: Nab_PTFd
    integer,                  intent(out)   :: error_flag

    ! local variables
    integer :: i,s
    
    error_flag = no_error

    ! write the Nab_TFd array of structs
    s = size(Nab_PTFd%Nab_TFd)
    DO i=1,s
       call read_Nab_TFd_buffbin_file(BufferedFile,&
            Nab_PTFd%Nab_TFd(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine read_NabPTFd_buffbin_file
    !  #]
  !---------------------------------------------
  subroutine Init_Fcalib_TR(Fcalib_TR, Num_RR, error_flag)
    !  #[
    TYPE(Fcalib_TR_type), intent(out) :: Fcalib_TR ! result
    integer(IntAs),       intent(in)  :: Num_RR
    integer,              intent(out) :: error_flag

    ! local variable
    integer :: AllocStatus

    error_flag = no_error
    IF (Num_RR .eq. 0) THEN
       ! init the struct to hold no data
       nullify(Fcalib_TR%Fcalib_R)
       nullify(Fcalib_TR%Fcalib_R_error)
    ELSE
       allocate(Fcalib_TR%Fcalib_R(1:Num_RR),&
                Fcalib_TR%Fcalib_R_error(1:Num_RR),&
                STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_Fcalib_TR()"
          error_flag = error_allocate
          return
       END IF
       Fcalib_TR%Fcalib_R(:) = missing_indicator_real_r8
       Fcalib_TR%Fcalib_R_error(:) = missing_indicator_real_r8
    END IF

  end subroutine Init_Fcalib_TR
    !  #]
  subroutine Delete_Fcalib_TR(Fcalib_TR)
    !  #[
    TYPE(Fcalib_TR_type), intent(inout) :: Fcalib_TR

    IF (associated(Fcalib_TR%Fcalib_R)) &
         deallocate(Fcalib_TR%Fcalib_R)

    IF (associated(Fcalib_TR%Fcalib_R_error)) &
         deallocate(Fcalib_TR%Fcalib_R_error)

  end subroutine Delete_Fcalib_TR
    !  #]
  subroutine Print_Fcalib_TR(Fcalib_TR,prefix,error_flag,PrintAll)
    !  #[
    TYPE(Fcalib_TR_type), intent(in)  :: Fcalib_TR
    character(len=*),     intent(in)  :: prefix
    integer,              intent(out) :: error_flag
    logical,    optional, intent(in)  :: PrintAll

    ! local variable
    integer            :: s, step, i
    character(len=256) :: text
    character(len=15)  :: name

    error_flag = no_error

    name = "%Fcalib_R"
    IF (associated(Fcalib_TR%Fcalib_R)) THEN
        ! get the size of the array
       s = size(Fcalib_TR%Fcalib_R)
       call PrintSizeAndGetStep(prefix,name,s,step,PrintAll)
       print *,"REALACC(12)"
       DO i=1,s,step
          write(text,"(3a,i3,a)") trim(prefix),trim(name),"(",i,") = "
          print *,trim(text),Fcalib_TR%Fcalib_R(i)
       END DO
       print *,"ENDREALACC"
    ELSE
       print *,trim(prefix)//trim(name)//trim(not_associated)
    END IF
    
    name = "%Fcalib_R_error"
    IF (associated(Fcalib_TR%Fcalib_R_error)) THEN
        ! get the size of the array
       s = size(Fcalib_TR%Fcalib_R_error)
       call PrintSizeAndGetStep(prefix,name,s,step,PrintAll)
       print *,"REALACC(12)"
       DO i=1,s,step
          write(text,"(3a,i3,a)") trim(prefix),trim(name),"(",i,") = "
          print *,trim(text),Fcalib_TR%Fcalib_R_error(i)
       END DO
       print *,"ENDREALACC"
    ELSE
       print *,trim(prefix)//trim(name)//trim(not_associated)
    END IF

  end subroutine Print_Fcalib_TR
    !  #]
  function GetByteSize_Fcalib_TR(Fcalib_TR) result(c_bytesize)
    !  #[
    TYPE(Fcalib_TR_type), intent(in) :: Fcalib_TR
    integer(IntAl)                   :: c_bytesize ! result

    c_bytesize = 0
    IF (associated(Fcalib_TR%Fcalib_R)) THEN
       c_bytesize = c_bytesize + &
            int(size(Fcalib_TR%Fcalib_R) * c_bytelen_FAdoxy,IntAl)
    ELSE
       print *,"WARNING in GetByteSize_Fcalib_TR():"
       print *,"Fcalib_TR%Fcalib_R [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

    IF (associated(Fcalib_TR%Fcalib_R_error)) THEN
       c_bytesize = c_bytesize + &
            int(size(Fcalib_TR%Fcalib_R_error) * c_bytelen_FAdoxy,IntAl)
    ELSE
       print *,"WARNING in GetByteSize_Fcalib_TR():"
       print *,"Fcalib_TR%Fcalib_R_error [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

  end function GetByteSize_Fcalib_TR
    !  #]
  subroutine write_Fcalib_TR_to_bin_file(fileunit,Fcalib_TR,error_flag)
    !  #[
    integer,              intent(in)  :: fileunit
    TYPE(Fcalib_TR_type), intent(in)  :: Fcalib_TR
    integer,              intent(out) :: error_flag

    ! local variables
    integer :: i,s
    
    error_flag = no_error

    ! write the Fcalib_R array
    s = size(Fcalib_TR%Fcalib_R)
    DO i=1,s
       call write_FAdoxy_to_bin_file(fileunit,Fcalib_TR%Fcalib_R(i),&
                                     error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! write the Fcalib_R_error array
    s = size(Fcalib_TR%Fcalib_R_error)
    DO i=1,s
       call write_FAdoxy_to_bin_file(fileunit,Fcalib_TR%Fcalib_R_error(i),&
                                     error_flag)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine write_Fcalib_TR_to_bin_file
    !  #]
  subroutine read_Fcalib_TR_from_bin_file(fileunit,Fcalib_TR,error_flag)
    !  #[
    integer,              intent(in)    :: fileunit
    TYPE(Fcalib_TR_type), intent(inout) :: Fcalib_TR
    integer,              intent(out)   :: error_flag

    ! local variables
    integer :: i,s
    
    error_flag = no_error

    ! read the Fcalib_R array
    s = size(Fcalib_TR%Fcalib_R)
    DO i=1,s
       call read_FAdoxy_from_bin_file(fileunit,Fcalib_TR%Fcalib_R(i),&
                                      error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! read the Fcalib_R_error array
    s = size(Fcalib_TR%Fcalib_R_error)
    DO i=1,s
       call read_FAdoxy_from_bin_file(fileunit,Fcalib_TR%Fcalib_R_error(i),&
                                      error_flag)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine read_Fcalib_TR_from_bin_file
    !  #]
  subroutine read_Fcalib_TR_buffbin_file(BufferedFile,Fcalib_TR,error_flag)
    !  #[
    TYPE(BufferedFile_type), pointer       :: BufferedFile
    TYPE(Fcalib_TR_type),    intent(inout) :: Fcalib_TR
    integer,                 intent(out)   :: error_flag

    ! local variables
    integer :: i,s
    
    error_flag = no_error

    ! read the Fcalib_R array
    s = size(Fcalib_TR%Fcalib_R)
    DO i=1,s
       call read_FAdoxy_from_buffbin_file(BufferedFile,Fcalib_TR%Fcalib_R(i),&
                                          error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! read the Fcalib_R_error array
    s = size(Fcalib_TR%Fcalib_R_error)
    DO i=1,s
       call read_FAdoxy_from_buffbin_file(BufferedFile,Fcalib_TR%Fcalib_R_error(i),&
                                          error_flag)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine read_Fcalib_TR_buffbin_file
    !  #]
  !---------------------------------------------
  subroutine Init_Fcalib_PTR(Fcalib_PTR, Num_T, Num_RR, error_flag)
    !  #[
    TYPE(Fcalib_PTR_type), intent(out) :: Fcalib_PTR ! result
    integer(IntAs),        intent(in)  :: Num_T, Num_RR
    integer,               intent(out) :: error_flag

    ! local variable
    integer :: i, AllocStatus

    error_flag = no_error
    IF (Num_T .eq. 0) THEN
       ! init the struct to hold no data
       nullify(Fcalib_PTR%Fcalib_TR)
    ELSE
       allocate(Fcalib_PTR%Fcalib_TR(1:Num_T),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_Fcalib_PTR()"
          error_flag = error_allocate
          return
       END IF
       DO i=1,Num_T
          call Init_Fcalib_TR(Fcalib_PTR%Fcalib_TR(i), Num_RR, error_flag)
       END DO
    END IF

  end subroutine Init_Fcalib_PTR
    !  #]
  subroutine Delete_Fcalib_PTR(Fcalib_PTR)
    !  #[
    TYPE(Fcalib_PTR_type), intent(inout) :: Fcalib_PTR

    ! local variable
    integer :: i

    IF (associated(Fcalib_PTR%Fcalib_TR)) THEN
       DO i=1,size(Fcalib_PTR%Fcalib_TR)
          call Delete_Fcalib_TR(Fcalib_PTR%Fcalib_TR(i))
       END DO
       deallocate(Fcalib_PTR%Fcalib_TR)
    END IF

  end subroutine Delete_Fcalib_PTR
    !  #]
  subroutine Print_Fcalib_PTR(Fcalib_PTR,prefix,error_flag,PrintAll)
    !  #[
    TYPE(Fcalib_PTR_type), intent(in)  :: Fcalib_PTR
    character(len=*),      intent(in)  :: prefix
    integer,               intent(out) :: error_flag
    logical,     optional, intent(in)  :: PrintAll

    ! local variable
    integer :: s, step, i
    character(len=256) :: text
    character(len=15)  :: name

    name = "%Fcalib_TR"
    IF (associated(Fcalib_PTR%Fcalib_TR)) THEN
        ! get the size of the array
       s = size(Fcalib_PTR%Fcalib_TR)
       call PrintSizeAndGetStep(prefix,name,s,step,PrintAll)
       DO i=1,s,step
          write(text,"(3a,i3,a)") trim(prefix),trim(name),"(",i,")"
          call Print_Fcalib_TR(Fcalib_PTR%Fcalib_TR(i),text,&
                               error_flag,PrintAll)
       END DO
    ELSE
       print *,trim(prefix)//trim(name)//trim(not_associated)
    END IF

  end subroutine Print_Fcalib_PTR
    !  #]
  function GetByteSize_Fcalib_PTR(Fcalib_PTR) result(c_bytesize)
    !  #[
    TYPE(Fcalib_PTR_type), intent(in) :: Fcalib_PTR
    integer(IntAl)                    :: c_bytesize ! result

    c_bytesize = 0
    IF (associated(Fcalib_PTR%Fcalib_TR)) THEN
       c_bytesize = c_bytesize + &
            int(size(Fcalib_PTR%Fcalib_TR) * &
                GetByteSize_Fcalib_TR(Fcalib_PTR%Fcalib_TR(1)),IntAl)
    ELSE
       print *,"WARNING in GetByteSize_Fcalib_PTR():"
       print *,"Fcalib_PTR%Fcalib_TR [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

  end function GetByteSize_Fcalib_PTR
    !  #]
  subroutine write_Fcalib_PTR_to_bin_file(fileunit,Fcalib_PTR,error_flag)
    !  #[
    integer,               intent(in)  :: fileunit
    TYPE(Fcalib_PTR_type), intent(in)  :: Fcalib_PTR
    integer,               intent(out) :: error_flag

    ! local variables
    integer :: i,s
    
    error_flag = no_error

    ! write the Fcalib_TR array of structs
    s = size(Fcalib_PTR%Fcalib_TR)
    DO i=1,s
       call write_Fcalib_TR_to_bin_file(fileunit,Fcalib_PTR%Fcalib_TR(i),&
                                        error_flag)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine write_Fcalib_PTR_to_bin_file
    !  #]
  subroutine read_Fcalib_PTR_from_bin_file(fileunit,Fcalib_PTR,error_flag)
    !  #[
    integer,               intent(in)    :: fileunit
    TYPE(Fcalib_PTR_type), intent(inout) :: Fcalib_PTR
    integer,               intent(out)   :: error_flag

    ! local variables
    integer :: i,s
    
    error_flag = no_error

    ! read the Fcalib_TR array of structs
    s = size(Fcalib_PTR%Fcalib_TR)
    DO i=1,s
       call read_Fcalib_TR_from_bin_file(fileunit,Fcalib_PTR%Fcalib_TR(i),&
                                        error_flag)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine read_Fcalib_PTR_from_bin_file
    !  #]
  subroutine read_Fcalib_PTR_buffbin_file(BufferedFile,Fcalib_PTR,error_flag)
    !  #[
    TYPE(BufferedFile_type), pointer       :: BufferedFile
    TYPE(Fcalib_PTR_type),   intent(inout) :: Fcalib_PTR
    integer,                 intent(out)   :: error_flag

    ! local variables
    integer :: i,s
    
    error_flag = no_error

    ! read the Fcalib_TR array of structs
    s = size(Fcalib_PTR%Fcalib_TR)
    DO i=1,s
       call read_Fcalib_TR_buffbin_file(BufferedFile,Fcalib_PTR%Fcalib_TR(i),&
                                        error_flag)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine read_Fcalib_PTR_buffbin_file
    !  #]
  !---------------------------------------------
  subroutine Init_RBC_DS(RBC_DS, Num_P, Num_T, Num_F, Num_FP, Num_fd, Num_RR, &
                         error_flag)
    !  #[
    TYPE(RBC_DS_type), intent(out) :: RBC_DS ! result
    integer(IntAs),    intent(in)  :: Num_P, Num_T, Num_F
    integer(IntAs),    intent(in)  :: Num_FP, Num_fd, Num_RR
    integer,           intent(out) :: error_flag

    ! local variable
    integer :: i, AllocStatus

    error_flag = no_error

    !  #[ init the P_Grid array
    IF (Num_P .eq. 0) THEN
       ! init the struct to hold no data
       nullify(RBC_DS%P_Grid)
    ELSE
       allocate(RBC_DS%P_Grid(1:Num_P),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_RBC_DS()"
          error_flag = error_allocate
          return
       END IF
       RBC_DS%P_Grid(:) = missing_indicator_IntAul
    END IF
    !  #]
    !  #[ init the T_Grid array
    IF (Num_T .eq. 0) THEN
       ! init the struct to hold no data
       nullify(RBC_DS%T_Grid)
    ELSE
       allocate(RBC_DS%T_Grid(1:Num_T),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_RBC_DS()"
          error_flag = error_allocate
          return
       END IF
       RBC_DS%T_Grid(:) = missing_indicator_IntAus
    END IF
    !  #]
    !  #[ init the F_Gridtmp array
    IF (Num_F .eq. 0) THEN
       ! init the struct to hold no data
       nullify(RBC_DS%F_Gridtmp)
    ELSE
       allocate(RBC_DS%F_Gridtmp(1:Num_F),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_RBC_DS()"
          error_flag = error_allocate
          return
       END IF
       RBC_DS%F_Gridtmp(:) = missing_indicator_IntAd
    END IF
    !  #]
    !  #[ init the Spec_Grid_PTF, Fcalib_PTR, Nab_PTFd arrays of structs
    IF (Num_P .eq. 0) THEN
       ! init the struct to hold no data
       nullify(RBC_DS%Spec_Grid_PTF, &
               RBC_DS%Fcalib_PTR, &
               RBC_DS%Nab_PTFd)
    ELSE
       allocate(RBC_DS%Spec_Grid_PTF(1:Num_P),&
                RBC_DS%Fcalib_PTR(1:Num_P), &
                RBC_DS%Nab_PTFd(1:Num_P), &
                STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_RBC_DS()"
          error_flag = error_allocate
          return
       END IF
       DO i=1,Num_P
          call Init_Spec_Grid_PTF(RBC_DS%Spec_Grid_PTF(i), Num_T, Num_F,&
                                  error_flag)
          IF (error_flag .ne. no_error) return
          call Init_Fcalib_PTR(RBC_DS%Fcalib_PTR(i), Num_T, Num_RR,&
                               error_flag)
          IF (error_flag .ne. no_error) return
          call Init_Nab_PTFd(RBC_DS%Nab_PTFd(i), Num_T, Num_fd,&
                                  error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    END IF
    !  #]
    !  #[ init the F_FP, TA_FP, TB_FP arrays
    IF (Num_FP .eq. 0) THEN
       ! init the struct to hold no data
       nullify(RBC_DS%F_FP,RBC_DS%TA_FP,RBC_DS%TB_FP)
    ELSE
       allocate(RBC_DS%F_FP( 1:Num_FP),&
                RBC_DS%TA_FP(1:Num_FP),&
                RBC_DS%TB_FP(1:Num_FP),&
                STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_RBC_DS()"
          error_flag = error_allocate
          return
       END IF
       RBC_DS%F_FP(:)  = missing_indicator_IntAd
       RBC_DS%TA_FP(:) = missing_indicator_real_r8
       RBC_DS%TB_FP(:) = missing_indicator_real_r8
    END IF
    !  #]
    !  #[ init the fd array
    IF (Num_fd .eq. 0) THEN
       ! init the struct to hold no data
       nullify(RBC_DS%fd)
    ELSE
       allocate(RBC_DS%fd(1:Num_fd),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_RBC_DS()"
          error_flag = error_allocate
          return
       END IF
       RBC_DS%fd(:) = missing_indicator_IntAd
    END IF
    !  #]
    !  #[ init the RR array
    IF (Num_RR .eq. 0) THEN
       ! init the struct to hold no data
       nullify(RBC_DS%RR)
    ELSE
       allocate(RBC_DS%RR(1:Num_RR),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_RBC_DS()"
          error_flag = error_allocate
          return
       END IF
       RBC_DS%RR(:) = missing_indicator_real_r8
    END IF
    !  #]
    !  #[ init the Fint_R array
    IF (Num_fd .eq. 0) THEN
       ! init the struct to hold no data
       nullify(RBC_DS%Fint_R)
    ELSE
       allocate(RBC_DS%Fint_R(1:Num_RR),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Init_RBC_DS()"
          error_flag = error_allocate
          return
       END IF
       RBC_DS%Fint_R(:) = missing_indicator_IntAd
    END IF
    !  #]

  end subroutine Init_RBC_DS
    !  #]
  subroutine Delete_RBC_DS(RBC_DS)
    !  #[
    TYPE(RBC_DS_type), intent(inout) :: RBC_DS

    ! local variable
    integer :: i

    IF (associated(RBC_DS%P_Grid))    deallocate(RBC_DS%P_Grid)
    IF (associated(RBC_DS%T_Grid))    deallocate(RBC_DS%T_Grid)
    IF (associated(RBC_DS%F_Gridtmp)) deallocate(RBC_DS%F_Gridtmp)
    
    IF (associated(RBC_DS%Spec_Grid_PTF)) THEN
       DO i=1,size(RBC_DS%Spec_Grid_PTF)
          call Delete_Spec_Grid_PTF(RBC_DS%Spec_Grid_PTF(i))
       END DO
       deallocate(RBC_DS%Spec_Grid_PTF)
    END IF

    IF (associated(RBC_DS%F_FP))  deallocate(RBC_DS%F_FP)
    IF (associated(RBC_DS%TA_FP)) deallocate(RBC_DS%TA_FP)
    IF (associated(RBC_DS%TB_FP)) deallocate(RBC_DS%TB_FP)
    IF (associated(RBC_DS%fd))    deallocate(RBC_DS%fd)
    IF (associated(RBC_DS%RR))    deallocate(RBC_DS%RR)

    IF (associated(RBC_DS%Fcalib_PTR)) THEN
       DO i=1,size(RBC_DS%Fcalib_PTR)
          call Delete_Fcalib_PTR(RBC_DS%Fcalib_PTR(i))
       END DO
       deallocate(RBC_DS%Fcalib_PTR)
    END IF

    IF (associated(RBC_DS%Nab_PTFd)) THEN
       DO i=1,size(RBC_DS%Nab_PTFd)
          call Delete_Nab_PTFd(RBC_DS%Nab_PTFd(i))
       END DO
       deallocate(RBC_DS%Nab_PTFd)

    IF (associated(RBC_DS%Fint_R))    deallocate(RBC_DS%Fint_R)
    END IF

  end subroutine Delete_RBC_DS
    !  #]
  subroutine Print_RBC_DS(RBC_DS,prefix,error_flag,PrintAll)
    !  #[
    ! WARNING: PrintAll=.true. will generate about 274 MB ascii output !!!
    TYPE(RBC_DS_type), intent(in)  :: RBC_DS
    character(len=*),  intent(in)  :: prefix
    integer,           intent(out) :: error_flag
    logical, optional, intent(in)  :: PrintAll

    ! local variables
    integer                      :: s, step, i
    character(len=256)           :: text
    character(len=maxlen_IntAus) :: str_IntAus
    character(len=maxlen_IntAul) :: str_IntAul
    character(len=maxlen_IntAd)  :: str_IntAd
    character(len=15)            :: name

    error_flag = no_error

    !  #[ P_Grid
    name = "%P_Grid"
    IF (associated(RBC_DS%P_Grid)) THEN
       ! get the size of the array
       s = size(RBC_DS%P_Grid)
       call PrintSizeAndGetStep(prefix,name,s,step,PrintAll)
       DO i=1,s,step
          call SetIntAul(RBC_DS%P_Grid(i),str_IntAul,error_flag)
          if (error_flag .ne. no_error) return
          write(text,"(3a,i3,3a)") trim(prefix),trim(name),"(",i,") = ",&
               trim(str_IntAul)," [Pa]"
          print *,trim(text)
       END DO
    ELSE
       print *,trim(prefix)//trim(name)//trim(not_associated)
    END IF
    !  #]
    !  #[ T_Grid
    name = "%T_Grid"
    IF (associated(RBC_DS%T_Grid)) THEN
       ! get the size of the array
       s = size(RBC_DS%T_Grid)
       call PrintSizeAndGetStep(prefix,name,s,step,PrintAll)
       DO i=1,s,step
          call SetIntAus(RBC_DS%T_Grid(i),str_IntAus,error_flag)
          if (error_flag .ne. no_error) return
          write(text,"(3a,i3,3a)") trim(prefix),trim(name),"(",i,") = ",&
               trim(str_IntAus)," [10-2K]"
          print *,trim(text)
       END DO
    ELSE
       print *,trim(prefix)//trim(name)//trim(not_associated)
    END IF
    !  #]
    !  #[ F_Gridtmp
    name = "%F_Gridtmp"
    IF (associated(RBC_DS%F_Gridtmp)) THEN
       ! get the size of the array
       s = size(RBC_DS%F_Gridtmp)
       call PrintSizeAndGetStep(prefix,name,s,step,PrintAll)
       DO i=1,s,step
          call SetIntAd(RBC_DS%F_Gridtmp(i),str_IntAd,error_flag)
          if (error_flag .ne. no_error) return
          write(text,"(3a,i3,3a)") trim(prefix),trim(name),"(",i,") = ",&
               trim(str_IntAd)," [Hz]"
          print *,trim(text)
       END DO
    ELSE
       print *,trim(prefix)//trim(name)//trim(not_associated)
    END IF
    !  #]
    !  #[ Spec_Grid_PTF
    name = "%Spec_Grid_PTF"
    IF (associated(RBC_DS%Spec_Grid_PTF)) THEN
       ! get the size of the array
       s = size(RBC_DS%Spec_Grid_PTF)
       call PrintSizeAndGetStep(prefix,name,s,step,PrintAll)
       DO i=1,s,step
          write(text,"(3a,i3,a)") trim(prefix),trim(name),"(",i,")"
          call Print_Spec_Grid_PTF(RBC_DS%Spec_Grid_PTF(i),text,error_flag,&
                                   PrintAll)
       END DO
    ELSE
       print *,trim(prefix)//trim(name)//trim(not_associated)
    END IF
    !  #]
    !  #[ F_FP
    name = "%F_FP"
    IF (associated(RBC_DS%F_FP)) THEN
       ! get the size of the array
       s = size(RBC_DS%F_FP)
       call PrintSizeAndGetStep(prefix,name,s,step,PrintAll)
       DO i=1,s,step
          call SetIntAd(RBC_DS%F_FP(i),str_IntAd,error_flag)
          if (error_flag .ne. no_error) return
          write(text,"(3a,i3,3a)") trim(prefix),trim(name),"(",i,") = ",&
               trim(str_IntAd)," [Hz]"
          print *,trim(text)
       END DO
    ELSE
        print *,trim(prefix)//trim(name)//trim(not_associated)
    END IF
    !  #]
    !  #[ TA_FP
    name = "%TA_FP"
    IF (associated(RBC_DS%TA_FP)) THEN
       ! get the size of the array
       s = size(RBC_DS%TA_FP)
       call PrintSizeAndGetStep(prefix,name,s,step,PrintAll)
       print *,"REALACC(12)"
       DO i=1,s,step
          write(text,"(3a,i3,a)") trim(prefix),trim(name),"(",i,") = "
          print *,trim(text),RBC_DS%TA_FP(i)
       END DO
       print *,"ENDREALACC"
    ELSE
       print *,trim(prefix)//trim(name)//trim(not_associated)
    END IF
    !  #]
    !  #[ TB_FP
    name = "%TB_FP"
    IF (associated(RBC_DS%TB_FP)) THEN
       ! get the size of the array
       s = size(RBC_DS%TB_FP)
       call PrintSizeAndGetStep(prefix,name,s,step,PrintAll)
       print *,"REALACC(12)"
       DO i=1,s,step
          write(text,"(3a,i3,a)") trim(prefix),trim(name),"(",i,") = "
          print *,trim(text),RBC_DS%TB_FP(i)
       END DO
       print *,"ENDREALACC"
    ELSE
       print *,trim(prefix)//trim(name)//trim(not_associated)
    END IF
    !  #]
    !  #[ fd
    name = "%fd"
    IF (associated(RBC_DS%fd)) THEN
       ! get the size of the array
       s = size(RBC_DS%fd)
       call PrintSizeAndGetStep(prefix,name,s,step,PrintAll)
       DO i=1,s,step
          call SetIntAd(RBC_DS%fd(i),str_IntAd,error_flag)
          if (error_flag .ne. no_error) return
          write(text,"(3a,i3,3a)") trim(prefix),trim(name),"(",i,") = ",&
               trim(str_IntAd)," [Hz]"
          print *,trim(text)
       END DO
    ELSE
        print *,trim(prefix)//trim(name)//trim(not_associated)
    END IF
    !  #]
    !  #[ RR
    name = "%RR"
    IF (associated(RBC_DS%RR)) THEN
       ! get the size of the array
       s = size(RBC_DS%RR)
       call PrintSizeAndGetStep(prefix,name,s,step,PrintAll)
       print *,"REALACC(12)"
       DO i=1,s,step
          write(text,"(3a,i3,a)") trim(prefix),trim(name),"(",i,") = "
          print *,trim(text),RBC_DS%RR(i)
       END DO
       print *,"ENDREALACC"
    ELSE
       print *,trim(prefix)//trim(name)//trim(not_associated)
    END IF
    !  #]
    !  #[ Fcalib_PTR
    name = "%Fcalib_PTR"
    IF (associated(RBC_DS%Fcalib_PTR)) THEN
       ! get the size of the array
       s = size(RBC_DS%Fcalib_PTR)
       call PrintSizeAndGetStep(prefix,name,s,step,PrintAll)
       DO i=1,s,step
          write(text,"(3a,i3,a)") trim(prefix),trim(name),"(",i,")"
          call Print_Fcalib_PTR(RBC_DS%Fcalib_PTR(i),text,error_flag,PrintAll)
       END DO
    ELSE
       print *,trim(prefix)//trim(name)//trim(not_associated)
    END IF
    !  #]

    !  #[ Nab_PTFd
    name = "%Nab_PTFd"
    IF (associated(RBC_DS%Nab_PTFd)) THEN
       ! get the size of the array
       s = size(RBC_DS%Nab_PTFd)
       call PrintSizeAndGetStep(prefix,name,s,step,PrintAll)
       DO i=1,s,step
          write(text,"(3a,i3,a)") trim(prefix),trim(name),"(",i,")"
          call Print_Nab_PTFd(RBC_DS%Nab_PTFd(i),text,error_flag,&
                                   PrintAll)
       END DO
    ELSE
       print *,trim(prefix)//trim(name)//trim(not_associated)
    END IF
    !  #]
    !  #[ Fint_R
    name = "%Fint_R"
    IF (associated(RBC_DS%Fint_R)) THEN
       ! get the size of the array
       s = size(RBC_DS%Fint_R)
       call PrintSizeAndGetStep(prefix,name,s,step,PrintAll)
       DO i=1,s,step
          call SetIntAd(RBC_DS%Fint_R(i),str_IntAd,error_flag)
          if (error_flag .ne. no_error) return
          write(text,"(3a,i3,3a)") trim(prefix),trim(name),"(",i,") = ",&
               trim(str_IntAd)," [Hz]"
          print *,trim(text)
       END DO
    ELSE
        print *,trim(prefix)//trim(name)//trim(not_associated)
    END IF
    !  #]

  end subroutine Print_RBC_DS
    !  #]
  function GetByteSize_RBC_DS(RBC_DS) result(c_bytesize)
    !  #[
    TYPE(RBC_DS_type), intent(in) :: RBC_DS
    integer(IntAl)                :: c_bytesize ! result

    c_bytesize = 0

    IF (associated(RBC_DS%P_Grid)) THEN
       c_bytesize = c_bytesize + &
                    int(size(RBC_DS%P_Grid) * c_bytelen_IntAul,IntAl)
    ELSE
       print *,"WARNING in GetByteSize_RBC_DS():"
       print *,"RBC_DS%P_Grid [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

    IF (associated(RBC_DS%T_Grid)) THEN
       c_bytesize = c_bytesize + &
                    int(size(RBC_DS%T_Grid) * c_bytelen_IntAus,IntAl)
    ELSE
       print *,"WARNING in GetByteSize_RBC_DS():"
       print *,"RBC_DS%T_Grid [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

    IF (associated(RBC_DS%F_Gridtmp)) THEN
       c_bytesize = c_bytesize + &
                    int(size(RBC_DS%F_Gridtmp) * c_bytelen_IntAd,IntAl)
    ELSE
       print *,"WARNING in GetByteSize_RBC_DS():"
       print *,"RBC_DS%F_Gridtmp [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

    IF (associated(RBC_DS%Spec_Grid_PTF)) THEN
       c_bytesize = c_bytesize + &
            int(size(RBC_DS%Spec_Grid_PTF) * &
                GetByteSize_Spec_Grid_PTF(RBC_DS%Spec_Grid_PTF(1)),IntAl)
    ELSE
       print *,"WARNING in GetByteSize_RBC_DS():"
       print *,"RBC_DS%Spec_Grid_PTF [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

    IF (associated(RBC_DS%F_FP)) THEN
       c_bytesize = c_bytesize + &
                    int(size(RBC_DS%F_FP) * c_bytelen_IntAd,IntAl)
    ELSE
       print *,"WARNING in GetByteSize_RBC_DS():"
       print *,"RBC_DS%F_FP [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

    IF (associated(RBC_DS%TA_FP)) THEN
       c_bytesize = c_bytesize + &
                    int(size(RBC_DS%TA_FP) * c_bytelen_FAdoxy,IntAl)
    ELSE
       print *,"WARNING in GetByteSize_RBC_DS():"
       print *,"RBC_DS%TA_FP [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

    IF (associated(RBC_DS%TB_FP)) THEN
       c_bytesize = c_bytesize + &
                    int(size(RBC_DS%TB_FP) * c_bytelen_FAdoxy,IntAl)
    ELSE
       print *,"WARNING in GetByteSize_RBC_DS():"
       print *,"RBC_DS%TB_FP [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

    IF (associated(RBC_DS%fd)) THEN
       c_bytesize = c_bytesize + &
                    int(size(RBC_DS%fd) * c_bytelen_IntAd,IntAl)
    ELSE
       print *,"WARNING in GetByteSize_RBC_DS():"
       print *,"RBC_DS%fd [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

    IF (associated(RBC_DS%RR)) THEN
       c_bytesize = c_bytesize + &
                    int(size(RBC_DS%RR) * c_bytelen_FAdoxy,IntAl)
    ELSE
       print *,"WARNING in GetByteSize_RBC_DS():"
       print *,"RBC_DS%RR [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

    IF (associated(RBC_DS%Fcalib_PTR)) THEN
       c_bytesize = c_bytesize + &
            int(size(RBC_DS%Fcalib_PTR) * &
                GetByteSize_Fcalib_PTR(RBC_DS%Fcalib_PTR(1)),IntAl)
    ELSE
       print *,"WARNING in GetByteSize_RBC_DS():"
       print *,"RBC_DS%Fcalib_PTR [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

    IF (associated(RBC_DS%Nab_PTFd)) THEN
       c_bytesize = c_bytesize + &
            int(size(RBC_DS%Nab_PTFd) * &
                GetByteSize_Nab_PTFd(RBC_DS%Nab_PTFd(1)),IntAl)
    ELSE
       print *,"WARNING in GetByteSize_RBC_DS():"
       print *,"RBC_DS%Nab_PTFd [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

    IF (associated(RBC_DS%Fint_R)) THEN
       c_bytesize = c_bytesize + &
                    int(size(RBC_DS%Fint_R) * c_bytelen_IntAd,IntAl)
    ELSE
       print *,"WARNING in GetByteSize_RBC_DS():"
       print *,"RBC_DS%Fint_R [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

  end function GetByteSize_RBC_DS
    !  #]
  subroutine Write_RBC_DS(fileunit,RBC_DS,error_flag)
    !  #[
    integer,           intent(in)  :: fileunit
    TYPE(RBC_DS_type), intent(in)  :: RBC_DS
    integer,           intent(out) :: error_flag

    ! local variables
    integer :: i,s
    
    error_flag = no_error
    
    ! write the P_Grid array
    s = size(RBC_DS%P_Grid)
    DO i=1,s
       call write_IntAul_to_bin_file(fileunit,RBC_DS%P_Grid(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! write the T_Grid array
    s = size(RBC_DS%T_Grid)
    DO i=1,s
       call write_IntAus_to_bin_file(fileunit,RBC_DS%T_Grid(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! write the F_Gridtmp array
    s = size(RBC_DS%F_Gridtmp)
    DO i=1,s
       call write_IntAd_to_bin_file(fileunit,RBC_DS%F_Gridtmp(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! write the Spec_Grid_PTF array of structs
    s = size(RBC_DS%Spec_Grid_PTF)
    DO i=1,s
       call write_Spec_Grid_PTF_to_bin_file(fileunit,RBC_DS%Spec_Grid_PTF(i),&
                                            error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! write the F_FP array
    s = size(RBC_DS%F_FP)
    DO i=1,s
       call write_IntAd_to_bin_file(fileunit,RBC_DS%F_FP(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! write the TA_FP array
    s = size(RBC_DS%TA_FP)
    DO i=1,s
       call write_FAdoxy_to_bin_file(fileunit,RBC_DS%TA_FP(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! write the TB_FP array
    s = size(RBC_DS%TB_FP)
    DO i=1,s
       call write_FAdoxy_to_bin_file(fileunit,RBC_DS%TB_FP(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! write the fd array
    s = size(RBC_DS%fd)
    DO i=1,s
       call write_IntAd_to_bin_file(fileunit,RBC_DS%fd(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! write the RR array
    s = size(RBC_DS%RR)
    DO i=1,s
       call write_FAdoxy_to_bin_file(fileunit,RBC_DS%RR(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! write the Fcalib_PTR array of structs
    s = size(RBC_DS%Fcalib_PTR)
    DO i=1,s
       call write_Fcalib_PTR_to_bin_file(fileunit,RBC_DS%Fcalib_PTR(i),&
                                         error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! write the Nab_PTFd array of structs
    s = size(RBC_DS%Nab_PTFd)
    DO i=1,s
       call write_Nab_PTFd_to_bin_file(fileunit,RBC_DS%Nab_PTFd(i),&
                                            error_flag)
       IF (error_flag .ne. no_error) return
    END DO

    ! write the Fint_R array
    s = size(RBC_DS%Fint_R)
    DO i=1,s
       call write_IntAd_to_bin_file(fileunit,RBC_DS%Fint_R(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO

  end subroutine Write_RBC_DS
    !  #]
  subroutine Read_RBC_DS(fileunit,RBC_DS,error_flag)
    !  #[
    integer,           intent(in)    :: fileunit
    TYPE(RBC_DS_type), intent(inout) :: RBC_DS
    integer,           intent(out)   :: error_flag

    ! local variables
    integer :: i,s
    
    error_flag = no_error
    
    ! read the P_Grid array
    s = size(RBC_DS%P_Grid)
    DO i=1,s
       call read_IntAul_from_bin_file(fileunit,RBC_DS%P_Grid(i),error_flag)
       IF (error_flag .ne. no_error) return

    END DO
    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"P_Grid succesfully read, s=",s
       print *,"RBC_DS%P_Grid(:) = ",RBC_DS%P_Grid(:)
    END IF

    ! read the T_Grid array
    s = size(RBC_DS%T_Grid)
    DO i=1,s
       call read_IntAus_from_bin_file(fileunit,RBC_DS%T_Grid(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"T_Grid succesfully read, s=",s
       print *,"RBC_DS%T_Grid(:) = ",RBC_DS%T_Grid(:)
    END IF

    ! read the F_Gridtmp array
    s = size(RBC_DS%F_Gridtmp)
    DO i=1,s
       call read_IntAd_from_bin_file(fileunit,RBC_DS%F_Gridtmp(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"F_Gridtmp succesfully read, s=",s
       print *,"RBC_DS%F_Gridtmp(1:10) = ",RBC_DS%F_Gridtmp(1:10)
       print *,"RBC_DS%F_Gridtmp(s-9:s) = ",RBC_DS%F_Gridtmp(s-9:s)
    END IF

    ! read the Spec_Grid_PTF array of structs
    s = size(RBC_DS%Spec_Grid_PTF)
    DO i=1,s
       call read_SpecGridPTF_from_bin_file(fileunit,RBC_DS%Spec_Grid_PTF(i),&
                                           error_flag)
       IF (error_flag .ne. no_error) return
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"==>Spec_Grid_PTF succesfully read, i=",i," out of ",s
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"Spec_Grid_PTF succesfully read, s=",s

    ! read the F_FP array
    s = size(RBC_DS%F_FP)
    DO i=1,s
       call read_IntAd_from_bin_file(fileunit,RBC_DS%F_FP(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"F_FP succesfully read, s=",s
       print *,"RBC_DS%F_FP(1:10) = ",RBC_DS%F_FP(1:10)
       print *,"RBC_DS%F_FP(s-9:s) = ",RBC_DS%F_FP(s-9:s)
    END IF

    ! read the TA_FP array
    s = size(RBC_DS%TA_FP)
    DO i=1,s
       call read_FAdoxy_from_bin_file(fileunit,RBC_DS%TA_FP(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"TA_FP succesfully read, s=",s
       print *,"REALACC(12)"
       print *,"RBC_DS%TA_FP(1:10) = ",RBC_DS%TA_FP(1:10)
       print *,"RBC_DS%TA_FP(s-9:s) = ",RBC_DS%TA_FP(s-9:s)
       print *,"ENDREALACC"
    END IF

    ! read the TB_FP array
    s = size(RBC_DS%TB_FP)
    DO i=1,s
       call read_FAdoxy_from_bin_file(fileunit,RBC_DS%TB_FP(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"TB_FP succesfully read, s=",s
       print *,"REALACC(12)"
       print *,"RBC_DS%TB_FP(1:10) = ",RBC_DS%TB_FP(1:10)
       print *,"RBC_DS%TB_FP(s-9:s) = ",RBC_DS%TB_FP(s-9:s)
       print *,"ENDREALACC"
    END IF

    ! read the fd array
    s = size(RBC_DS%fd)
    DO i=1,s
       call read_IntAd_from_bin_file(fileunit,RBC_DS%fd(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"fd succesfully read, s=",s
       print *,"RBC_DS%fd(:) = ",RBC_DS%fd(:)
    END IF

    ! read the RR array
    s = size(RBC_DS%RR)
    DO i=1,s
       call read_FAdoxy_from_bin_file(fileunit,RBC_DS%RR(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"RR succesfully read, s=",s
       print *,"REALACC(12)"
       print *,"RBC_DS%RR(:) = ",RBC_DS%RR(:)
       print *,"ENDREALACC"
    END IF

    ! read the Fcalib_PTR array of structs
    s = size(RBC_DS%Fcalib_PTR)
    DO i=1,s
       call read_Fcalib_PTR_from_bin_file(fileunit,RBC_DS%Fcalib_PTR(i),&
                                          error_flag)
       IF (error_flag .ne. no_error) return
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"==>Fcalib_TR succesfully read, i=",i," out of ",s
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"Fcalib_PTR succesfully read, s=",s

    ! read the Nab_PTFd array of structs
    s = size(RBC_DS%Nab_PTFd)
    DO i=1,s
       call read_NabPTFd_from_bin_file(fileunit,RBC_DS%Nab_PTFd(i),&
                                           error_flag)
       IF (error_flag .ne. no_error) return
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"==>Nab_PTFd succesfully read, i=",i," out of ",s
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"Nab_PTFd succesfully read, s=",s

    ! read the Fint_R array
    s = size(RBC_DS%Fint_R)
    DO i=1,s
       call read_IntAd_from_bin_file(fileunit,RBC_DS%Fint_R(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"Fint_R succesfully read, s=",s
       print *,"RBC_DS%Fint_R(:) = ",RBC_DS%Fint_R(:)
    END IF

  end subroutine Read_RBC_DS
    !  #]
  subroutine Read_RBC_DS_Buff(BufferedFile,RBC_DS,error_flag)
    !  #[
    TYPE(BufferedFile_type), pointer       :: BufferedFile
    TYPE(RBC_DS_type),       intent(inout) :: RBC_DS
    integer,                 intent(out)   :: error_flag

    ! local variables
    integer :: i,s
    
    error_flag = no_error
    
    ! read the P_Grid array
    s = size(RBC_DS%P_Grid)
    DO i=1,s
       call read_IntAul_from_buffbin_file(BufferedFile,RBC_DS%P_Grid(i),&
                                          error_flag)
       IF (error_flag .ne. no_error) return
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"P_Grid succesfully read, s=",s
       print *,"RBC_DS%P_Grid(:) = ",RBC_DS%P_Grid(:)
    END IF

    ! read the T_Grid array
    s = size(RBC_DS%T_Grid)
    DO i=1,s
       call read_IntAus_from_buffbin_file(BufferedFile,RBC_DS%T_Grid(i),&
                                          error_flag)
       IF (error_flag .ne. no_error) return
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"T_Grid succesfully read, s=",s
       print *,"RBC_DS%T_Grid(:) = ",RBC_DS%T_Grid(:)
    END IF

    ! read the F_Gridtmp array
    s = size(RBC_DS%F_Gridtmp)
    DO i=1,s
       call read_IntAd_from_buffbin_file(BufferedFile,RBC_DS%F_Gridtmp(i),&
                                         error_flag)
       IF (error_flag .ne. no_error) return
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"F_Gridtmp succesfully read, s=",s
       print *,"RBC_DS%F_Gridtmp(1:10) = ",RBC_DS%F_Gridtmp(1:10)
       print *,"RBC_DS%F_Gridtmp(s-9:s) = ",RBC_DS%F_Gridtmp(s-9:s)
    END IF

    ! read the Spec_Grid_PTF array of structs
    s = size(RBC_DS%Spec_Grid_PTF)
    DO i=1,s
       call read_SpecGridPTF_buffbin_file(BufferedFile,&
            RBC_DS%Spec_Grid_PTF(i), error_flag)
       IF (error_flag .ne. no_error) return
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"==>Spec_Grid_PTF succesfully read, i=",i," out of ",s
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"Spec_Grid_PTF succesfully read, s=",s

    ! read the F_FP array
    s = size(RBC_DS%F_FP)
    DO i=1,s
       call read_IntAd_from_buffbin_file(BufferedFile,RBC_DS%F_FP(i),&
                                         error_flag)
       IF (error_flag .ne. no_error) return
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"F_FP succesfully read, s=",s
       print *,"RBC_DS%F_FP(1:10) = ",RBC_DS%F_FP(1:10)
       print *,"RBC_DS%F_FP(s-9:s) = ",RBC_DS%F_FP(s-9:s)
    END IF

    ! read the TA_FP array
    s = size(RBC_DS%TA_FP)
    DO i=1,s
       call read_FAdoxy_from_buffbin_file(BufferedFile,RBC_DS%TA_FP(i),&
                                          error_flag)
       IF (error_flag .ne. no_error) return
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"TA_FP succesfully read, s=",s
       print *,"REALACC(12)"
       print *,"RBC_DS%TA_FP(1:10) = ",RBC_DS%TA_FP(1:10)
       print *,"RBC_DS%TA_FP(s-9:s) = ",RBC_DS%TA_FP(s-9:s)
       print *,"ENDREALACC"
    END IF

    ! read the TB_FP array
    s = size(RBC_DS%TB_FP)
    DO i=1,s
       call read_FAdoxy_from_buffbin_file(BufferedFile,RBC_DS%TB_FP(i),&
                                          error_flag)
       IF (error_flag .ne. no_error) return
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"TB_FP succesfully read, s=",s
       print *,"REALACC(12)"
       print *,"RBC_DS%TB_FP(1:10) = ",RBC_DS%TB_FP(1:10)
       print *,"RBC_DS%TB_FP(s-9:s) = ",RBC_DS%TB_FP(s-9:s)
       print *,"ENDREALACC"
    END IF

    ! read the fd array
    s = size(RBC_DS%fd)
    DO i=1,s
       call read_IntAd_from_buffbin_file(BufferedFile,RBC_DS%fd(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"fd succesfully read, s=",s
       print *,"RBC_DS%fd(:) = ",RBC_DS%fd(:)
    END IF

    ! read the RR array
    s = size(RBC_DS%RR)
    DO i=1,s
       call read_FAdoxy_from_buffbin_file(BufferedFile,RBC_DS%RR(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"RR succesfully read, s=",s
       print *,"REALACC(12)"
       print *,"RBC_DS%RR(:) = ",RBC_DS%RR(:)
       print *,"ENDREALACC"
    END IF

    ! read the Fcalib_PTR array of structs
    s = size(RBC_DS%Fcalib_PTR)
    DO i=1,s
       call read_Fcalib_PTR_buffbin_file(BufferedFile,RBC_DS%Fcalib_PTR(i),&
                                         error_flag)
       IF (error_flag .ne. no_error) return
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"==>Fcalib_TR succesfully read, i=",i," out of ",s
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"Fcalib_PTR succesfully read, s=",s

    ! read the Nab_PTFd array of structs
    s = size(RBC_DS%Nab_PTFd)
    DO i=1,s
       call read_NabPTFd_buffbin_file(BufferedFile,&
            RBC_DS%Nab_PTFd(i), error_flag)
       IF (error_flag .ne. no_error) return
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"==>Nab_PTFd succesfully read, i=",i," out of ",s
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"Nab_PTFd succesfully read, s=",s

    ! read the Fint_R array
    s = size(RBC_DS%Fint_R)
    DO i=1,s
       call read_IntAd_from_buffbin_file(BufferedFile,RBC_DS%Fint_R(i),error_flag)
       IF (error_flag .ne. no_error) return
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"Fint_R succesfully read, s=",s
       print *,"RBC_DS%Fint_R(:) = ",RBC_DS%Fint_R(:)
    END IF

  end subroutine Read_RBC_DS_Buff
    !  #]
  subroutine Fill_RBC_DS_with_Dummy_data(RBC_DS,error_flag)
    !  #[
    TYPE(RBC_DS_type), intent(inout) :: RBC_DS
    integer,           intent(out)   :: error_flag

    ! local variables
    integer :: i,j,k,si,sj,sk,c
    
    error_flag = no_error
    
    si = size(RBC_DS%P_Grid)
    DO i=1,si
       RBC_DS%P_Grid(i) = i
    END DO
    si = size(RBC_DS%T_Grid)
    DO i=1,si
       RBC_DS%T_Grid(i) = int(i,IntAus)
    END DO
    si = size(RBC_DS%F_Gridtmp)
    DO i=1,si
       RBC_DS%F_Gridtmp(i)=i
    END DO
    si = size(RBC_DS%Spec_Grid_PTF)
    c=0
    DO i=1,si
       sj=size(RBC_DS%Spec_Grid_PTF(1)%Spec_Grid_TF)
       DO j=1,sj
          sk=size(RBC_DS%Spec_Grid_PTF(1)%Spec_Grid_TF(1)%Spec_Grid_F)
          DO k=1,sk             
             c=c+1
             RBC_DS%Spec_Grid_PTF(i)%Spec_Grid_TF(j)%Spec_Grid_F(k)=c
          END DO
       END DO
    END DO
    si = size(RBC_DS%F_FP)
    DO i=1,si
       RBC_DS%F_FP(i)=i
    END DO
    si = size(RBC_DS%TA_FP)
    DO i=1,si
       RBC_DS%TA_FP(i)=i
    END DO
    si = size(RBC_DS%TB_FP)
    DO i=1,si
       RBC_DS%TB_FP(i)=i
    END DO
    si = size(RBC_DS%fd)
    DO i=1,si
       RBC_DS%fd(i)=i
    END DO
    si = size(RBC_DS%RR)
    DO i=1,si
       RBC_DS%RR(i)=i
    END DO
    si = size(RBC_DS%Fcalib_PTR)
    c=0
    DO i=1,si
       sj=size(RBC_DS%Fcalib_PTR(1)%Fcalib_TR)
       DO j=1,sj
          sk=size(RBC_DS%Fcalib_PTR(1)%Fcalib_TR(1)%Fcalib_R)
          DO k=1,sk
             c=c+1
             RBC_DS%Fcalib_PTR(i)%Fcalib_TR(j)%Fcalib_R(k)=c
          END DO
          sk=size(RBC_DS%Fcalib_PTR(1)%Fcalib_TR(1)%Fcalib_R_error)
          DO k=1,sk
             c=c+1
             RBC_DS%Fcalib_PTR(i)%Fcalib_TR(j)%Fcalib_R_error(k)=c
          END DO
       END DO
    END DO
   
    si = size(RBC_DS%Nab_PTFd)
    c=0
    DO i=1,si
       sj=size(RBC_DS%Nab_PTFd(1)%Nab_TFd)
       DO j=1,sj
          sk=size(RBC_DS%Nab_PTFd(1)%Nab_TFd(1)%Na_Fd)
          DO k=1,sk             
             c=c+1
             RBC_DS%Nab_PTFd(i)%Nab_TFd(j)%Na_Fd(k)=c
          END DO
          DO k=1,sk             
             c=c+1
             RBC_DS%Nab_PTFd(i)%Nab_TFd(j)%Nb_Fd(k)=c
          END DO
       END DO
    END DO
    si = size(RBC_DS%Fint_R)
    DO i=1,si
       RBC_DS%Fint_R(i)=i
    END DO
 
  end subroutine Fill_RBC_DS_with_Dummy_data
    !  #]
  subroutine Check_RBC_DS_with_Dummy_data(RBC_DS,error_flag)
    !  #[
    TYPE(RBC_DS_type), intent(in)  :: RBC_DS
    integer,           intent(out) :: error_flag

    ! local variables
    integer                 :: i,j,k,si,sj,sk,c
    real(FAdoxy), parameter :: eps = 1.e-9
    real(FAdoxy)            :: diff
    error_flag = no_error
    
    si = size(RBC_DS%P_Grid)
    DO i=1,si
       IF (RBC_DS%P_Grid(i) .ne. i) error_flag = error_programming
    END DO
    si = size(RBC_DS%T_Grid)
    DO i=1,si
       IF (RBC_DS%T_Grid(i) .ne. i) error_flag = error_programming
    END DO
    si = size(RBC_DS%F_Gridtmp)
    DO i=1,si
       IF (RBC_DS%F_Gridtmp(i) .ne. i) error_flag = error_programming
    END DO
    si = size(RBC_DS%Spec_Grid_PTF)
    c=0
    DO i=1,si
       sj=size(RBC_DS%Spec_Grid_PTF(1)%Spec_Grid_TF)
       DO j=1,sj
          sk=size(RBC_DS%Spec_Grid_PTF(1)%Spec_Grid_TF(1)%Spec_Grid_F)
          DO k=1,sk             
             c=c+1
             diff = RBC_DS%Spec_Grid_PTF(i)%Spec_Grid_TF(j)%Spec_Grid_F(k)-c
             IF (abs(diff) .gt. eps) error_flag = error_programming
          END DO
       END DO
    END DO
    si = size(RBC_DS%F_FP)
    DO i=1,si
       IF (RBC_DS%F_FP(i) .ne. i) error_flag = error_programming
    END DO
    si = size(RBC_DS%TA_FP)
    DO i=1,si
       IF (RBC_DS%TA_FP(i) .ne. i) error_flag = error_programming
    END DO
    si = size(RBC_DS%TB_FP)
    DO i=1,si
       IF (RBC_DS%TB_FP(i) .ne. i) error_flag = error_programming
    END DO
    si = size(RBC_DS%fd)
    DO i=1,si
       IF (RBC_DS%fd(i) .ne. i) error_flag = error_programming
    END DO
    si = size(RBC_DS%RR)
    DO i=1,si
       IF (RBC_DS%RR(i) .ne. i) error_flag = error_programming
    END DO
    si = size(RBC_DS%Fcalib_PTR)
    c=0
    DO i=1,si
       sj=size(RBC_DS%Fcalib_PTR(1)%Fcalib_TR)
       DO j=1,sj
          sk=size(RBC_DS%Fcalib_PTR(1)%Fcalib_TR(1)%Fcalib_R)
          DO k=1,sk
             c=c+1
             diff = RBC_DS%Fcalib_PTR(i)%Fcalib_TR(j)%Fcalib_R(k) - c
             IF (abs(diff) .gt. eps) error_flag = error_programming
          END DO
          sk=size(RBC_DS%Fcalib_PTR(1)%Fcalib_TR(1)%Fcalib_R_error)
          DO k=1,sk
             c=c+1
             diff = RBC_DS%Fcalib_PTR(i)%Fcalib_TR(j)%Fcalib_R_error(k) - c
             IF (abs(diff) .gt. eps) error_flag = error_programming
          END DO
       END DO
    END DO

    si = size(RBC_DS%Nab_PTFd)
    c=0
    DO i=1,si
       sj=size(RBC_DS%Nab_PTFd(1)%Nab_TFd)
       DO j=1,sj
          sk=size(RBC_DS%Nab_PTFd(1)%Nab_TFd(1)%Na_Fd)
          DO k=1,sk             
             c=c+1
             diff = RBC_DS%Nab_PTFd(i)%Nab_TFd(j)%Na_Fd(k)-c
             IF (abs(diff) .gt. eps) error_flag = error_programming
          END DO
          DO k=1,sk             
             c=c+1
             diff = RBC_DS%Nab_PTFd(i)%Nab_TFd(j)%Nb_Fd(k)-c
             IF (abs(diff) .gt. eps) error_flag = error_programming
          END DO
       END DO
    END DO
    si = size(RBC_DS%Fint_R)
    DO i=1,si
       IF (RBC_DS%Fint_R(i) .ne. i) error_flag = error_programming
    END DO
    
  end subroutine Check_RBC_DS_with_Dummy_data
    !  #]
  !---------------------------------------------
  subroutine Init_RBC_DataStructure(RBC, error_flag)
    !  #[
    TYPE(RBC_DataType), intent(out) :: RBC ! output
    integer,     intent(out) :: error_flag ! output

    ! init
    error_flag = no_error

    ! init the module datastructure
    nullify(RBC%FH)
    nullify(RBC%MPH)
    nullify(RBC%SPH)

    ! the attached dataset
    nullify(RBC%RBC_DS)

  end subroutine Init_RBC_DataStructure
    !  #]
  subroutine Delete_RBC_DataStructure(RBC)
    !  #[
    TYPE(RBC_DataType), intent(inout) :: RBC ! input/output
    ! (contents will be deallocated after this routine)

    ! clean up the header datastructures
    IF (associated(RBC%SPH)) call Delete_RBC_SPH(RBC%SPH)

    IF (associated(RBC%FH)) deallocate(RBC%FH)
    IF (associated(RBC%MPH)) deallocate(RBC%MPH)
    IF (associated(RBC%SPH)) deallocate(RBC%SPH)

    ! clean up the dataset
    IF (associated(RBC%RBC_DS)) THEN
       call Delete_RBC_DS(RBC%RBC_DS)
       deallocate(RBC%RBC_DS)
    END IF

  end subroutine Delete_RBC_DataStructure
    !  #]
  subroutine Print_RBC_DataStructure(RBC, prefix, error_flag, PrintAll)
    !  #[ ascii output to screen
    ! WARNING: PrintAll=.true. will generate about 274 MB ascii output !!!

    TYPE(RBC_DataType), intent(in)  :: RBC        ! input
    character(len=*),   intent(in)  :: prefix     ! input
    integer,            intent(out) :: error_flag ! output
    logical,  optional, intent(in)  :: PrintAll   ! optional input

    ! local variable
    character(len=256) :: text
    character(len=15)  :: name

    ! init
    error_flag = no_error

    name = "%FH"
    IF (associated(RBC%FH)) THEN
       call Print_FH(RBC%FH,error_flag,prefix)
    ELSE
       print *,trim(prefix)//trim(name)//trim(not_associated2)
    END IF

    name = "%MPH"
    IF (associated(RBC%MPH)) THEN
       call Print_MPH(RBC%MPH,error_flag,prefix)
    ELSE
       print *,trim(prefix)//trim(name)//trim(not_associated2)
    END IF

    name = "%SPH"
    IF (associated(RBC%SPH)) THEN
       call Print_RBC_SPH(RBC%SPH,error_flag)
       IF (error_flag .ne. no_error) return
    ELSE
       print *,trim(prefix)//trim(name)//trim(not_associated2)
    END IF

    name = "%RBC_DS"
    IF (associated(RBC%RBC_DS)) THEN
       write(text,"(2a)") trim(prefix),trim(name)
       call Print_RBC_DS(RBC%RBC_DS,text,error_flag,PrintAll)
    ELSE
       print *,trim(prefix)//trim(name)//trim(not_associated2)
    END IF

  end subroutine Print_RBC_DataStructure
    !  #]
  function GetByteSize_RBC_DataStruct_KVT(RBC) result(c_bytesize)
    !  #[
    TYPE(RBC_DataType), intent(in)  :: RBC        ! input
    integer(IntAl)                  :: c_bytesize ! result

    c_bytesize = 0
    c_bytesize = c_bytesize + GetByteSize_MPH_KVT()
    c_bytesize = c_bytesize + GetByteSize_RBC_SPH_KVT()

    IF (associated(RBC%RBC_DS)) THEN
       c_bytesize = c_bytesize + GetByteSize_RBC_DS(RBC%RBC_DS)
    ELSE
       print *,"WARNING in GetByteSize_RBC_DataStruct_KVT():"
       print *,"RBC%RBC_DS [seems not associated]"
       print *,"determining the size of this datatruct is not possible !!"
    END IF

  end function GetByteSize_RBC_DataStruct_KVT
    !  #]
  !---------------------------------------------
  subroutine PrintSizeAndGetStep(prefix,name,s,step,PrintAll)
    !  #[
    character(len=*),  intent(in)  :: prefix
    character(len=*),  intent(in)  :: name
    logical, optional, intent(in)  :: PrintAll
    integer,           intent(in)  :: s
    integer,           intent(out) :: step

    ! local variables
    logical                      :: do_PrintAll
    character(len=256)           :: text

    do_PrintAll = .false.
    IF (present(PrintAll)) do_PrintAll = PrintAll

    write(text,"(3a,i3)") trim(prefix),trim(name),"( : ) > size = ",s    
    IF (do_PrintAll) THEN
       print *,trim(text)
       step = 1
    ELSE
       print *,trim(text),firstlast
       step = s-1
    END IF

  end subroutine PrintSizeAndGetStep
    !  #]
  !---------------------------------------------
end module RBC_DataStructure
