module RBC_SPH
  !  #[ Documentation
  !
  ! this module defines the Specific Product Header for the ee_cfi fileformat
  ! for the L2B processor RBC look-up table files
  ! The SPH format may differ for different filedefinitions, therefore
  ! we use a private definition for this filetype only
  !
  ! for a description, see 
  ! TN2.1, sec.6.3.1, table 64
  !
  ! Written by J. de Kloe
  !
  ! History:
  !    22-Feb-2006 J. de Kloe moved add_RBC_SPH_item into the xml_module
  !    28-Feb-2006 J. de Kloe added prefix parameter to print routine
  !    01-Mar-2006 J. de Kloe added routine Read_RBC_SPH_from_KVT
  !    02-Mar-2006 J. de Kloe adapted to changes in XML_module
  !    13-Mar-2006 J. de Kloe updated Get_RBC_SPH_from_xml_tree, and
  !                           added error_flag to Print_DSD routine
  !    29-Mar-2006 D. Tan     Add FP_DSD + quality params (L2B IODD Version 1.0)
  !    11-Apr-2006 P. Poli    changed Start/Stop laser freq offset to FAdo56
  !    12-May-2006 D. Tan     Null use of ParKind1 module
  !    17-Aug-2006 P. Poli    Fixed bug in assignment of 
  !                           Total_Num_of_Corrupt_*_RefP
  !    09-Mar-2007 D. Tan     Use ee_sph_name.
  !    24-Apr-2007 J. de Kloe phase out the SummaryVars code
  !    07-Jun-2007 J. de Kloe adapted to changed xml interface
  !    03-Dec-2007 D. Tan     String "Sph_Descriptor", not "SPH_..."
  !    16-Jan-2008 J. de Kloe phase out integer kind i_
  !
  !  #] 
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE XML_module, only: xml_node_type, &
                    max_str_len, xml_node_get_string, xml_set_fixed_header, &
                    check_length_xml_item, check_string_length, &
                    xml_tree_set_string_node_value, xml_tree_add_child, &
                    ee_vh_name, ee_sph_name, xml_add_SPH_item, &
                    xml_tree_add_and_set_attribute
  USE EE_CFI_Datatypes, only: IntAs,  maxlen_IntAs,  SetIntAs,  &
                              IntAus, maxlen_IntAus, SetIntAus, &
                              IntAl,  maxlen_IntAl,  SetIntAl,  &
                              IntAul, maxlen_IntAul, SetIntAul, &
                              IntAd, &
                              FAdoxy, maxlen_FAdo36, maxlen_FAdo56, &
                              SetFAdoxy, Remove_Plus_Sign
  USE DataSetDescriptor, only: DSD_type, Init_DSD, Print_DSD, &
                               Add_DSD_to_xml_SPH, Write_DSD_to_KVT, &
                               Get_DSD_from_xml_tree, GetByteSize_DSD_KVT, &
                               Read_DSD_from_KVT, &
                               maxlen_DS_Name, maxlen_DS_Type
  USE KVT_module, only: write_kvt_string, write_spare_kvt_string, &
                        read_kvt_string, read_spare_kvt_string, &
                        Compare_XML_to_KVT_item_Repl
  ! support modules
  USE ErrorHandler, only: no_error, error_allocate, error_programming, &
                          error_reading_file
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE Compiler_Features, only: newline

  !  #]
  !  #[ Type definitions and parameters
  implicit none

  public

  integer, parameter :: maxlen_SPH_Descriptor   = 28
  integer, parameter :: maxlen_Ref_RBC_Suite    = 20

  !
  ! DS_Names for "attached datasets" in the RBC product file
  !
  character(len=maxlen_DS_Name) ,parameter :: RBC_DS_Name_RBC_ADS=&
                                     "Rayleigh_Brillouin_ADS      "
  character(len=maxlen_DS_Name) ,parameter :: RBC_DS_Name_FP_MDS=&
                                     "Fabry_Perot_MDS             "

  ! DS_Types (annotation/measurement) for "attached datasets" in
  ! the RBC product file
  character(len=maxlen_DS_Type) ,parameter :: RBC_DS_Type_RBC_ADS="A"
  ! DS_Types for "reference datasets" in the RBC product file
  character(len=maxlen_DS_Type) ,parameter :: RBC_DS_Type_FP_MDS="R"

  TYPE RBC_SPH_type
     !  #[
     character(len=maxlen_SPH_Descriptor) :: SPH_Descriptor
     ! Spare 1
     character(len=maxlen_Ref_RBC_Suite)  :: Ref_RBC_Suite
     character(len=maxlen_IntAs)          :: Num_P  ! =  23
     character(len=maxlen_IntAs)          :: Num_T  ! = 161
     character(len=maxlen_IntAs)          :: Num_F  ! = 937
     character(len=maxlen_IntAs)          :: Num_FP ! = 877
     character(len=maxlen_IntAs)          :: Num_Fd ! =  61
     character(len=maxlen_IntAs)          :: Num_RR ! = 101
     ! Spare 2
     character(len=maxlen_IntAul)         :: P_min ! in [Pa]
     character(len=maxlen_IntAul)         :: P_max ! in [Pa]
     character(len=maxlen_IntAus)         :: T_min ! in [10^-2 K]
     character(len=maxlen_IntAus)         :: T_max ! in [10^-2 K]
     !
     character(len=maxlen_FAdo56)         :: Laser_Freq_Offset_Start
     character(len=maxlen_FAdo56)         :: Laser_Freq_Offset_Stop
     character(len=maxlen_IntAl)          :: Total_Num_of_Observations
     character(len=maxlen_IntAl)          :: Total_Num_of_Measurements
     character(len=maxlen_IntAl)          :: Total_Num_of_Reference_Pulses
     character(len=maxlen_IntAl)          :: Total_Num_of_Corrupt_Mie_Meas
     character(len=maxlen_IntAl)          :: Total_Num_of_Corrupt_Ray_Meas
     character(len=maxlen_IntAl)          :: Total_Num_of_Corrupt_Mie_RefP
     character(len=maxlen_IntAl)          :: Total_Num_of_Corrupt_Ray_RefP
     character(len=maxlen_FAdo36)         :: Average_Error_FP_Response_A
     character(len=maxlen_FAdo36)         :: Average_Error_FP_Response_B
     ! Spare 3
     TYPE(DSD_type), pointer :: RBC_DSD
     TYPE(DSD_type), pointer :: FP_DSD
     !
  END TYPE RBC_SPH_type
    !  #]

  ! the default verbosity setting is Err (as silent as possible)
  ! (this switch is not yet set by an init routine, and can be set 
  !  to VerbosityReport manually for debugging purposes)
  integer, save :: LocalVerbosity = VerbosityError
!  integer, save :: LocalVerbosity = VerbosityReport
  
  !  #]
  !  #[ Module interface
  ! all is public for the moment, unless specified explicitely otherwise
  !  #]
contains
  !---------------------------------
  subroutine Init_RBC_SPH(RBC_SPH,error_flag)
    !  #[
    TYPE(RBC_SPH_type), pointer :: RBC_SPH
    integer, intent(out) :: error_flag
    
    error_flag = no_error
    IF (.not. associated(RBC_SPH)) THEN
       print *,"ERROR in Init_RBC_SPH():"
       print *,"The pointer to this RBC_SPH seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! see the remark in MainProductHeader.F90 for the reason why I use
    ! the form x(:)=' ' to initialise the strings

    RBC_SPH%SPH_Descriptor(:) = ' '
    ! Spare 1
    RBC_SPH%Ref_RBC_Suite(:)  = ' '
    RBC_SPH%Num_P  = ' '
    RBC_SPH%Num_T  = ' '
    RBC_SPH%Num_F  = ' '
    RBC_SPH%Num_FP = ' '
    RBC_SPH%Num_Fd = ' '
    RBC_SPH%Num_RR = ' '
    ! Spare 2
    RBC_SPH%P_min  = ' '
    RBC_SPH%P_max  = ' '
    RBC_SPH%T_min  = ' '
    RBC_SPH%T_max  = ' '
    !
    RBC_SPH%Laser_Freq_Offset_Start       = ' '
    RBC_SPH%Laser_Freq_Offset_Stop        = ' '
    RBC_SPH%Total_Num_of_Observations     = ' '
    RBC_SPH%Total_Num_of_Measurements     = ' '
    RBC_SPH%Total_Num_of_Reference_Pulses = ' '
    RBC_SPH%Total_Num_of_Corrupt_Mie_Meas = ' '
    RBC_SPH%Total_Num_of_Corrupt_Ray_Meas = ' '
    RBC_SPH%Total_Num_of_Corrupt_Mie_RefP = ' '
    RBC_SPH%Total_Num_of_Corrupt_Ray_RefP = ' '
    RBC_SPH%Average_Error_FP_Response_A   = ' '
    RBC_SPH%Average_Error_FP_Response_B   = ' '
    ! Spare 3
    nullify(RBC_SPH%RBC_DSD)
    nullify(RBC_SPH%FP_DSD)

  end subroutine Init_RBC_SPH
    !  #]
  subroutine Print_RBC_SPH(RBC_SPH,error_flag,prefix)
    !  #[
    TYPE(RBC_SPH_type), pointer     :: RBC_SPH
    integer,            intent(out) :: error_flag
    character(len=*), optional, intent(in) :: prefix

    ! local variable
    character(len=256) :: txt
    character(len=256) :: tmp_prefix

    error_flag = no_error

    tmp_prefix = "..."
    IF (present(prefix)) tmp_prefix = prefix

    IF (.not. associated(RBC_SPH)) THEN
       print *,"[pointer to RBC_SPH is not associated so nothing to print]"
       return
    END IF

    txt = RBC_SPH%SPH_Descriptor
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,trim(tmp_prefix)//"%Sph_Descriptor = ",trim(txt)

    ! Spare 1

    txt = RBC_SPH%Ref_RBC_Suite
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,trim(tmp_prefix)//"%Ref_RBC_Suite  = ",trim(txt)

    txt = RBC_SPH%Num_P
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,trim(tmp_prefix)//"%Num_P          = ",trim(txt)

    txt = RBC_SPH%Num_T
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,trim(tmp_prefix)//"%Num_T          = ",trim(txt)

    txt = RBC_SPH%Num_F
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,trim(tmp_prefix)//"%Num_F          = ",trim(txt)

    txt = RBC_SPH%Num_FP
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,trim(tmp_prefix)//"%Num_FP         = ",trim(txt)

    txt = RBC_SPH%Num_Fd
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,trim(tmp_prefix)//"%Num_Fd         = ",trim(txt)

    txt = RBC_SPH%Num_RR
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,trim(tmp_prefix)//"%Num_RR         = ",trim(txt)

    ! Spare 2

    txt = RBC_SPH%P_min
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,trim(tmp_prefix)//"%P_min          = ",trim(txt)," <Pa>"

    txt = RBC_SPH%P_max
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,trim(tmp_prefix)//"%P_max          = ",trim(txt)," <Pa>"

    txt = RBC_SPH%T_min
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,trim(tmp_prefix)//"%T_min          = ",trim(txt)," <10-2K>"

    txt = RBC_SPH%T_max
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,trim(tmp_prefix)//"%T_max          = ",trim(txt)," <10-2K>"

    !
    txt = RBC_SPH%Laser_Freq_Offset_Start
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,trim(tmp_prefix)//"%Laser_Freq_Offset_Start       = ",trim(txt)," <MHz>"

    txt = RBC_SPH%Laser_Freq_Offset_Stop
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,trim(tmp_prefix)//"%Laser_Freq_Offset_Stop        = ",trim(txt)," <MHz>"

    txt = RBC_SPH%Total_Num_of_Observations
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,trim(tmp_prefix)//"%Total_Num_of_Observations     = ",trim(txt)

    txt = RBC_SPH%Total_Num_of_Measurements
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,trim(tmp_prefix)//"%Total_Num_of_Measurements     = ",trim(txt)

    txt = RBC_SPH%Total_Num_of_Reference_Pulses
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,trim(tmp_prefix)//"%Total_Num_of_Reference_Pulses = ",trim(txt)

    txt = RBC_SPH%Total_Num_of_Corrupt_Mie_Meas
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,trim(tmp_prefix)//"%Total_Num_of_Corrupt_Mie_Meas = ",trim(txt)

    txt = RBC_SPH%Total_Num_of_Corrupt_Ray_Meas
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,trim(tmp_prefix)//"%Total_Num_of_Corrupt_Ray_Meas = ",trim(txt)

    txt = RBC_SPH%Total_Num_of_Corrupt_Mie_RefP
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,trim(tmp_prefix)//"%Total_Num_of_Corrupt_Mie_RefP = ",trim(txt)

    txt = RBC_SPH%Total_Num_of_Corrupt_Ray_RefP
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,trim(tmp_prefix)//"%Total_Num_of_Corrupt_Ray_RefP = ",trim(txt)

    txt = RBC_SPH%Average_Error_FP_Response_A
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,trim(tmp_prefix)//"%Average_Error_FP_Response_A   = ",trim(txt)

    txt = RBC_SPH%Average_Error_FP_Response_B
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,trim(tmp_prefix)//"%Average_Error_FP_Response_B   = ",trim(txt)

    ! Spare 3

    call Print_DSD(RBC_SPH%RBC_DSD,trim(tmp_prefix)//"%RBC_DSD",error_flag)
    if (error_flag .ne. no_error) return
    call Print_DSD(RBC_SPH%FP_DSD,trim(tmp_prefix)//"%FP_DSD",error_flag)
    if (error_flag .ne. no_error) return

  end subroutine Print_RBC_SPH
    !  #]
  subroutine Delete_RBC_SPH(RBC_SPH)
    !  #[
    TYPE(RBC_SPH_type), pointer :: RBC_SPH

    IF (associated(RBC_SPH%RBC_DSD)) &
         deallocate(RBC_SPH%RBC_DSD)

    IF (associated(RBC_SPH%FP_DSD)) &
         deallocate(RBC_SPH%FP_DSD)

  end subroutine Delete_RBC_SPH
    !  #]
  subroutine Fill_RBC_SPH_items(RBC_SPH, error_flag, &
       SPH_Descriptor, Ref_RBC_Suite, Num_P, Num_T, Num_F, Num_FP, Num_Fd,&
       Num_RR, P_min, P_max, T_min, T_max,&
       Laser_Freq_Offset_Start, Laser_Freq_Offset_Stop,&
       Total_Num_of_Observations,&
       Total_Num_of_Measurements,&
       Total_Num_of_Reference_Pulses,&
       Total_Num_of_Corrupt_Mie_Meas, Total_Num_of_Corrupt_Ray_Meas,&
       Total_Num_of_Corrupt_Mie_RefP, Total_Num_of_Corrupt_Ray_RefP,&
       Average_Error_FP_Response_A, Average_Error_FP_Response_B)
    !  #[
    ! Remark: the DSD structures are not yet filled inside this routine
    ! To get the needed sizes the GetByteSize functions are used,
    ! this is done in the WriteRBCdata module.
    ! If we would do it here, the SPH module would become dependant
    ! on the RBC_Datastructure module, which in turn is dependent on this
    ! SPH module, and this cyclic dependency cannot be properly compiled.

    TYPE(RBC_SPH_type), pointer     :: RBC_SPH        ! filled on output
    integer,            intent(out) :: error_flag     ! result
    character(len=*),   intent(in)  :: SPH_Descriptor
    character(len=*),   intent(in)  :: Ref_RBC_Suite
    integer(IntAs),     intent(in)  :: Num_P
    integer(IntAs),     intent(in)  :: Num_T
    integer(IntAs),     intent(in)  :: Num_F
    integer(IntAs),     intent(in)  :: Num_FP
    integer(IntAs),     intent(in)  :: Num_Fd
    integer(IntAs),     intent(in)  :: Num_RR
    integer(IntAul),    intent(in)  :: P_min
    integer(IntAul),    intent(in)  :: P_max
    integer(IntAus),    intent(in)  :: T_min
    integer(IntAus),    intent(in)  :: T_max
    !
    real(FAdoxy),       intent(in)  :: Laser_Freq_Offset_Start
    real(FAdoxy),       intent(in)  :: Laser_Freq_Offset_Stop
    integer(IntAl),     intent(in)  :: Total_Num_of_Observations
    integer(IntAl),     intent(in)  :: Total_Num_of_Measurements
    integer(IntAl),     intent(in)  :: Total_Num_of_Reference_Pulses
    integer(IntAl),     intent(in)  :: Total_Num_of_Corrupt_Mie_Meas
    integer(IntAl),     intent(in)  :: Total_Num_of_Corrupt_Ray_Meas
    integer(IntAl),     intent(in)  :: Total_Num_of_Corrupt_Mie_RefP
    integer(IntAl),     intent(in)  :: Total_Num_of_Corrupt_Ray_RefP
    real(FAdoxy),       intent(in)  :: Average_Error_FP_Response_A
    real(FAdoxy),       intent(in)  :: Average_Error_FP_Response_B

    error_flag = no_error
    IF (.not. associated(RBC_SPH)) THEN
       print *,"ERROR in Fill_RBC_SPH_items():"
       print *,"The pointer to this RBC_SPH seems not associated !!"
       error_flag = error_programming
       return
    END IF

    call  check_string_length(SPH_Descriptor,"Sph_Descriptor",&
         maxlen_SPH_Descriptor,"Fill_RBC_SPH_items") 
    RBC_SPH%SPH_Descriptor = SPH_Descriptor

    call  check_string_length(Ref_RBC_Suite,"Ref_RBC_Suite",&
         maxlen_Ref_RBC_Suite,"Fill_RBC_SPH_items") 
    RBC_SPH%Ref_RBC_Suite = Ref_RBC_Suite

    call SetIntAs(Num_P,RBC_SPH%Num_P,error_flag)
    IF (error_flag .ne. no_error) return

    call SetIntAs(Num_T,RBC_SPH%Num_T,error_flag)
    IF (error_flag .ne. no_error) return

    call SetIntAs(Num_F,RBC_SPH%Num_F,error_flag)
    IF (error_flag .ne. no_error) return

    call SetIntAs(Num_FP,RBC_SPH%Num_FP,error_flag)
    IF (error_flag .ne. no_error) return

    call SetIntAs(Num_Fd,RBC_SPH%Num_Fd,error_flag)
    IF (error_flag .ne. no_error) return

    call SetIntAs(Num_RR,RBC_SPH%Num_RR,error_flag)
    IF (error_flag .ne. no_error) return

    call SetIntAul(P_min,RBC_SPH%P_min,error_flag)
    IF (error_flag .ne. no_error) return

    call SetIntAul(P_max,RBC_SPH%P_max,error_flag)
    IF (error_flag .ne. no_error) return

    call SetIntAus(T_min,RBC_SPH%T_min,error_flag)
    IF (error_flag .ne. no_error) return

    call SetIntAus(T_max,RBC_SPH%T_max,error_flag)
    IF (error_flag .ne. no_error) return

    call SetFAdoxy(Laser_Freq_Offset_Start,5,6,RBC_SPH%Laser_Freq_Offset_Start,error_flag)
    IF (error_flag .ne. no_error) return

    call SetFAdoxy(Laser_Freq_Offset_Stop,5,6,RBC_SPH%Laser_Freq_Offset_Stop,error_flag)
    IF (error_flag .ne. no_error) return

    call SetIntAl(Total_Num_of_Observations,RBC_SPH%Total_Num_of_Observations,error_flag)
    IF (error_flag .ne. no_error) return

    call SetIntAl(Total_Num_of_Measurements,RBC_SPH%Total_Num_of_Measurements,error_flag)
    IF (error_flag .ne. no_error) return

    call SetIntAl(Total_Num_of_Reference_Pulses,RBC_SPH%Total_Num_of_Reference_Pulses,error_flag)
    IF (error_flag .ne. no_error) return

    call SetIntAl(Total_Num_of_Corrupt_Mie_Meas,RBC_SPH%Total_Num_of_Corrupt_Mie_Meas,error_flag)
    IF (error_flag .ne. no_error) return

    call SetIntAl(Total_Num_of_Corrupt_Ray_Meas,RBC_SPH%Total_Num_of_Corrupt_Ray_Meas,error_flag)
    IF (error_flag .ne. no_error) return

    call SetIntAl(Total_Num_of_Corrupt_Mie_RefP,RBC_SPH%Total_Num_of_Corrupt_Mie_RefP,error_flag)
    IF (error_flag .ne. no_error) return

    call SetIntAl(Total_Num_of_Corrupt_Ray_RefP,RBC_SPH%Total_Num_of_Corrupt_Ray_RefP,error_flag)
    IF (error_flag .ne. no_error) return

    call SetFAdoxy(Average_Error_FP_Response_A,3,6,RBC_SPH%Average_Error_FP_Response_A,error_flag)
    IF (error_flag .ne. no_error) return

    call SetFAdoxy(Average_Error_FP_Response_B,3,6,RBC_SPH%Average_Error_FP_Response_B,error_flag)
    IF (error_flag .ne. no_error) return

    ! the DSD is not filled in this subroutine
    
  end subroutine Fill_RBC_SPH_items
    !  #]
  subroutine Copy_RBC_SPH_to_xml_tree(RBC_SPH, path_in, xml_tree, error_flag)
    !  #[
    TYPE(RBC_SPH_type),  pointer    :: RBC_SPH    ! input
    character(len=*),    intent(in) :: path_in    ! input
    TYPE(xml_node_type), pointer    :: xml_tree   ! points to updated struct
    integer,             intent(out):: error_flag ! output

    ! local variables
    character(len=max_str_len) :: path
    character(len=max_str_len) :: path_dsd

    error_flag = no_error

    ! add the <Specific_Product_Header></Specific_Product_Header> tags
    path = trim(path_in)//"/"//ee_vh_name
    call xml_tree_add_child(xml_tree,path,ee_sph_name,error_flag)
    IF (error_flag .ne. no_error) return

    ! fill all items inside the RBC_SPH
    call xml_add_SPH_item(xml_tree, path_in,&
                          "Sph_Descriptor",RBC_SPH%SPH_Descriptor,error_flag)
    IF (error_flag .ne. no_error) return

    ! Spare_1
    call xml_add_SPH_item(xml_tree, path_in, "Spare_1",newline,error_flag)
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Ref_RBC_Suite",&
                          RBC_SPH%Ref_RBC_Suite,error_flag)
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_P",&
                          Remove_Plus_Sign(RBC_SPH%Num_P),error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_T",&
                          Remove_Plus_Sign(RBC_SPH%Num_T),error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_F",&
                          Remove_Plus_Sign(RBC_SPH%Num_F),error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_FP",&
                          Remove_Plus_Sign(RBC_SPH%Num_FP),error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Fd",&
                          Remove_Plus_Sign(RBC_SPH%Num_Fd),error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_RR",&
                          Remove_Plus_Sign(RBC_SPH%Num_RR),error_flag )
    IF (error_flag .ne. no_error) return

    ! Spare_2
    call xml_add_SPH_item(xml_tree, path_in, "Spare_2",newline,error_flag)
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "P_min",&
                          Remove_Plus_Sign(RBC_SPH%P_min),error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "P_max",&
                          Remove_Plus_Sign(RBC_SPH%P_max),error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "T_min",&
                          Remove_Plus_Sign(RBC_SPH%T_min),error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "T_max",&
                          Remove_Plus_Sign(RBC_SPH%T_max),error_flag )
    IF (error_flag .ne. no_error) return

    !
    call xml_add_SPH_item(xml_tree, path_in, "Laser_Freq_Offset_Start",&
                          Remove_Plus_Sign(RBC_SPH%Laser_Freq_Offset_Start),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Laser_Freq_Offset_Stop",&
                          Remove_Plus_Sign(RBC_SPH%Laser_Freq_Offset_Stop),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Total_Num_of_Observations",&
                          Remove_Plus_Sign(RBC_SPH%Total_Num_of_Observations),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Total_Num_of_Measurements",&
                          Remove_Plus_Sign(RBC_SPH%Total_Num_of_Measurements),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Total_Num_of_Reference_Pulses",&
                          Remove_Plus_Sign(RBC_SPH%Total_Num_of_Reference_Pulses),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Total_Num_of_Corrupt_Mie_Meas",&
                          Remove_Plus_Sign(RBC_SPH%Total_Num_of_Corrupt_Mie_Meas),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Total_Num_of_Corrupt_Ray_Meas",&
                          Remove_Plus_Sign(RBC_SPH%Total_Num_of_Corrupt_Ray_Meas),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Total_Num_of_Corrupt_Mie_RefP",&
                          Remove_Plus_Sign(RBC_SPH%Total_Num_of_Corrupt_Mie_RefP),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Total_Num_of_Corrupt_Ray_RefP",&
                          Remove_Plus_Sign(RBC_SPH%Total_Num_of_Corrupt_Ray_RefP),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Average_Error_FP_Response_A",&
                          Remove_Plus_Sign(RBC_SPH%Average_Error_FP_Response_A),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Average_Error_FP_Response_B",&
                          Remove_Plus_Sign(RBC_SPH%Average_Error_FP_Response_B),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    ! Spare_3
    call xml_add_SPH_item(xml_tree, path_in, "Spare_3",newline,error_flag)
    IF (error_flag .ne. no_error) return

    ! add the <List_of_DSDs></List_of_DSDs> tags
    path = trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name
    call xml_tree_add_child(xml_tree,path,"List_of_Dsds",error_flag)
    IF (error_flag .ne. no_error) return

    path_dsd = trim(path)//"/List_of_Dsds"
    call xml_tree_add_and_set_attribute(xml_tree,path_dsd,"count","2",&
                                        error_flag)

    call Add_DSD_to_xml_SPH(xml_tree, path_in, RBC_SPH%RBC_DSD, 1, error_flag)
    IF (error_flag .ne. no_error) return
    
    call Add_DSD_to_xml_SPH(xml_tree, path_in, RBC_SPH%FP_DSD, 2, error_flag)
    IF (error_flag .ne. no_error) return
    
    return

  end subroutine Copy_RBC_SPH_to_xml_tree
    !  #]
  subroutine Get_RBC_SPH_from_xml_tree(xml_tree, path_in, RBC_SPH, error_flag)
    !  #[
    TYPE(xml_node_type), pointer     :: xml_tree   ! input
    character(len=*),    intent(in)  :: path_in    ! input
    TYPE(RBC_SPH_type),  pointer     :: RBC_SPH    ! result
    integer,             intent(out) :: error_flag ! output

    ! local variables
    character(len=max_str_len) :: path
    character(len=max_str_len) :: text
    integer                    :: AllocStatus

    ! init
    error_flag = no_error

    IF (.not. associated(RBC_SPH)) THEN
       print *,"ERROR in Get_RBC_SPH_from_xml_tree():"
       print *,"the pointer RBC_SPH seems not associated !!!"
       error_flag = error_programming
       return
    END IF

    IF (.not. associated(xml_tree)) THEN
       print *,"ERROR in Get_RBC_SPH_from_xml_tree():"
       print *,"the pointer xml_tree seems not associated !!!"
       error_flag = error_programming
       return
    END IF

    !---------------------------------
    ! copying the RBC_SPH XML elements
    !---------------------------------

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//"/Sph_Descriptor"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    RBC_SPH%SPH_Descriptor = &
         check_length_xml_item(text,path,maxlen_SPH_Descriptor,&
                               "Get_RBC_SPH_from_xml_tree")

    ! Spare_1: no need to read this one

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//"/Ref_RBC_Suite"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    RBC_SPH%Ref_RBC_Suite = &
         check_length_xml_item(text,path,maxlen_Ref_RBC_Suite,&
                               "Get_RBC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//"/Num_P"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    RBC_SPH%Num_P = &
         check_length_xml_item(text,path,maxlen_IntAs,&
                               "Get_RBC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//"/Num_T"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    RBC_SPH%Num_T = &
         check_length_xml_item(text,path,maxlen_IntAs,&
                               "Get_RBC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//"/Num_F"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    RBC_SPH%Num_F = &
         check_length_xml_item(text,path,maxlen_IntAs,&
                               "Get_RBC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//"/Num_FP"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    RBC_SPH%Num_FP = &
         check_length_xml_item(text,path,maxlen_IntAs,&
                               "Get_RBC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//"/Num_Fd"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    RBC_SPH%Num_Fd = &
         check_length_xml_item(text,path,maxlen_IntAs,&
                               "Get_RBC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//"/Num_RR"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    RBC_SPH%Num_RR = &
         check_length_xml_item(text,path,maxlen_IntAs,&
                               "Get_RBC_SPH_from_xml_tree")

    ! Spare_2: no need to read this one

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//"/P_min"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    RBC_SPH%P_min = &
         check_length_xml_item(text,path,maxlen_IntAul,&
                               "Get_RBC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//"/P_max"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    RBC_SPH%P_max = &
         check_length_xml_item(text,path,maxlen_IntAul,&
                               "Get_RBC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//"/T_min"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    RBC_SPH%T_min = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_RBC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//"/T_max"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    RBC_SPH%T_max = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_RBC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Laser_Freq_Offset_Start"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    RBC_SPH%Laser_Freq_Offset_Start = &
         check_length_xml_item(text,path,maxlen_FAdo56,&
                               "Get_RBC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Laser_Freq_Offset_Stop"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    RBC_SPH%Laser_Freq_Offset_Stop = &
         check_length_xml_item(text,path,maxlen_FAdo56,&
                               "Get_RBC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Total_Num_of_Observations"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    RBC_SPH%Total_Num_of_Observations = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_RBC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Total_Num_of_Measurements"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    RBC_SPH%Total_Num_of_Measurements = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_RBC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Total_Num_of_Reference_Pulses"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    RBC_SPH%Total_Num_of_Reference_Pulses = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_RBC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Total_Num_of_Corrupt_Mie_Meas"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    RBC_SPH%Total_Num_of_Corrupt_Mie_Meas = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_RBC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Total_Num_of_Corrupt_Ray_Meas"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    RBC_SPH%Total_Num_of_Corrupt_Ray_Meas = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_RBC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Total_Num_of_Corrupt_Mie_RefP"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    RBC_SPH%Total_Num_of_Corrupt_Mie_RefP = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_RBC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Total_Num_of_Corrupt_Ray_RefP"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    RBC_SPH%Total_Num_of_Corrupt_Ray_RefP = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_RBC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Average_Error_FP_Response_A"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    RBC_SPH%Average_Error_FP_Response_A = &
         check_length_xml_item(text,path,maxlen_FAdo36,&
                               "Get_RBC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Average_Error_FP_Response_B"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    RBC_SPH%Average_Error_FP_Response_B = &
         check_length_xml_item(text,path,maxlen_FAdo36,&
                               "Get_RBC_SPH_from_xml_tree")

    ! Spare_3: no need to read this one
    allocate(RBC_SPH%RBC_DSD,STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       error_flag = error_allocate
       return
    END IF

    call Init_DSD(RBC_SPH%RBC_DSD,error_flag)
    IF (error_flag .ne. no_error) return

    ! get the RBC_DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree, path_in, RBC_SPH%RBC_DSD, 1, &
                               RBC_DS_Name_RBC_ADS, RBC_DS_Type_RBC_ADS, &
                               error_flag)
    IF (error_flag .ne. no_error) return

    allocate(RBC_SPH%FP_DSD,STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       error_flag = error_allocate
       return
    END IF

    call Init_DSD(RBC_SPH%FP_DSD,error_flag)
    IF (error_flag .ne. no_error) return

    ! get the FP_DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree, path_in, RBC_SPH%FP_DSD, 2, &
                               RBC_DS_Name_FP_MDS, RBC_DS_Type_FP_MDS, &
                               error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine Get_RBC_SPH_from_xml_tree
    !  #]
  !---------------------------------
  subroutine Write_RBC_SPH_to_KVT(RBC_SPH, fileunit, error_flag)
    !  #[
    ! Writes the RBC Specific Product Header (KVT values).

    TYPE(RBC_SPH_type), pointer     :: RBC_SPH
    integer,            intent(in)  :: fileunit
    integer,            intent(out) :: error_flag

    ! local variables
    character(len=64) :: varname
    character(len=64) :: value
    character(len=64) :: phys_unit
    integer           :: max_value_length
    integer           :: filepos_end,filepos_start
    integer           :: nbytes_written, nbytes_expected

    !---------------------------------
    ! Writing the RBC SPH KVT elements
    !---------------------------------

    error_flag = no_error
    IF (.not. associated(RBC_SPH)) THEN
       print *,"ERROR in Write_RBC_SPH_to_KVT():"
       print *,"The pointer to this RBC_SPH seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! as double check, compare the number of bytes written
    ! to what was expected from the GetByteSize_RBC_SPH_KVT() function
    ! so store the filepointer before we write anything
    inquire(unit=fileunit,nextrec=filepos_start)
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"in Write_RBC_SPH_to_KVT: filepos_start = ",filepos_start

    ! SPH_Descriptor 
    varname          = "Sph_Descriptor"
    value            = RBC_SPH%SPH_Descriptor
    max_value_length = maxlen_SPH_Descriptor
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag,&
                          use_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998
   
    ! Spare_1: this is a spare value without content
    max_value_length = 40
    call write_spare_kvt_string(fileunit,max_value_length,error_flag)
    IF (error_flag .ne. no_error) return

    ! Ref_RBC_Suite
    varname          = "Ref_RBC_Suite"
    value            = RBC_SPH%Ref_RBC_Suite
    max_value_length = maxlen_Ref_RBC_Suite
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag,&
                          use_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998

    varname          = "Num_P"
    value            = RBC_SPH%Num_P
    max_value_length = maxlen_IntAs
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    varname          = "Num_T"
    value            = RBC_SPH%Num_T
    max_value_length = maxlen_IntAs
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    varname          = "Num_F"
    value            = RBC_SPH%Num_F
    max_value_length = maxlen_IntAs
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    varname          = "Num_FP"
    value            = RBC_SPH%Num_FP
    max_value_length = maxlen_IntAs
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    varname          = "Num_Fd"
    value            = RBC_SPH%Num_Fd
    max_value_length = maxlen_IntAs
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    varname          = "Num_RR"
    value            = RBC_SPH%Num_RR
    max_value_length = maxlen_IntAs
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Spare_2: this is a spare value without content
    max_value_length = 40
    call write_spare_kvt_string(fileunit,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    varname          = "P_min"
    value            = RBC_SPH%P_min
    phys_unit        = "Pa"
    max_value_length = maxlen_IntAul
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag, phys_unit=phys_unit)
    IF (error_flag .ne. no_error) goto 998

    varname          = "P_max"
    value            = RBC_SPH%P_max
    phys_unit        = "Pa"
    max_value_length = maxlen_IntAul
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag, phys_unit=phys_unit)
    IF (error_flag .ne. no_error) goto 998

    varname          = "T_min"
    value            = RBC_SPH%T_min
    phys_unit        = "10-2K"
    max_value_length = maxlen_IntAus
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag, phys_unit=phys_unit)
    IF (error_flag .ne. no_error) goto 998

    varname          = "T_max"
    value            = RBC_SPH%T_max
    phys_unit        = "10-2K"
    max_value_length = maxlen_IntAus
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag, phys_unit=phys_unit)
    IF (error_flag .ne. no_error) goto 998

    !
    varname          = "Laser_Freq_Offset_Start"
    value            = RBC_SPH%Laser_Freq_Offset_Start
    phys_unit        = "MHz"
    max_value_length = maxlen_FAdo56
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag, phys_unit=phys_unit)
    IF (error_flag .ne. no_error) goto 998

    varname          = "Laser_Freq_Offset_Stop"
    value            = RBC_SPH%Laser_Freq_Offset_Stop
    phys_unit        = "MHz"
    max_value_length = maxlen_FAdo56
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag, phys_unit=phys_unit)
    IF (error_flag .ne. no_error) goto 998

    varname          = "Total_Num_of_Observations"
    value            = RBC_SPH%Total_Num_of_Observations
    max_value_length = maxlen_IntAl
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    varname          = "Total_Num_of_Measurements"
    value            = RBC_SPH%Total_Num_of_Measurements
    max_value_length = maxlen_IntAl
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    varname          = "Total_Num_of_Reference_Pulses"
    value            = RBC_SPH%Total_Num_of_Reference_Pulses
    max_value_length = maxlen_IntAl
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    varname          = "Total_Num_of_Corrupt_Mie_Meas"
    value            = RBC_SPH%Total_Num_of_Corrupt_Mie_Meas
    max_value_length = maxlen_IntAl
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    varname          = "Total_Num_of_Corrupt_Ray_Meas"
    value            = RBC_SPH%Total_Num_of_Corrupt_Ray_Meas
    max_value_length = maxlen_IntAl
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    varname          = "Total_Num_of_Corrupt_Mie_RefP"
    value            = RBC_SPH%Total_Num_of_Corrupt_Mie_RefP
    max_value_length = maxlen_IntAl
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    varname          = "Total_Num_of_Corrupt_Ray_RefP"
    value            = RBC_SPH%Total_Num_of_Corrupt_Ray_RefP
    max_value_length = maxlen_IntAl
    call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    varname          = "Average_Error_FP_Response_A"
    value            = RBC_SPH%Average_Error_FP_Response_A
!    phys_unit        = "AU"
    max_value_length = maxlen_FAdo36
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag) !, phys_unit=phys_unit)
    IF (error_flag .ne. no_error) goto 998

    varname          = "Average_Error_FP_Response_B"
    value            = RBC_SPH%Average_Error_FP_Response_B
!    phys_unit        = "AU"
    max_value_length = maxlen_FAdo36
    call write_kvt_string(fileunit,varname,value,max_value_length,&
                          error_flag) !, phys_unit=phys_unit)
    IF (error_flag .ne. no_error) goto 998

    ! Spare_3: this is a spare value without content
    max_value_length = 40
    call write_spare_kvt_string(fileunit,max_value_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Write RBC_DSD
    call Write_DSD_to_KVT(RBC_SPH%RBC_DSD,fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Write FP_DSD
    call Write_DSD_to_KVT(RBC_SPH%FP_DSD,fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! as double check, compare the number of bytes written
    ! to what was expected from the GetByteSize_RBC_SPH_KVT() function
    inquire(unit=fileunit,nextrec=filepos_end)
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"in Write_RBC_SPH_to_KVT: filepos_end = ",filepos_end

    nbytes_written  = filepos_end-filepos_start
    nbytes_expected = GetByteSize_RBC_SPH_KVT()

    IF (nbytes_written .ne. nbytes_expected) THEN
       print *,"ERROR in Write_RBC_SPH_to_KVT():"
       print *,"nbytes_written  = ",nbytes_written
       print *,"nbytes_expected = ",nbytes_expected
       print *,"something went wrong when writing the RBC_SPH to the KVT"
       print *,"part of the binary file, probably a programming error ..."
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Write_RBC_SPH_to_KVT"
    return

  end subroutine Write_RBC_SPH_to_KVT
    !  #]
  subroutine Read_RBC_SPH_from_KVT(RBC_SPH, fileunit, error_flag, Compare)
    !  #[
    ! Reads the RBC Specific Product Header (KVT values).

    ! if Compare=.true. then
    ! assume the values are already read from the XML file
    ! so read and compare them, and then replace the ones in memory by
    ! those read, because those read are in KVT format, which is the
    ! convention of choice.

    TYPE(RBC_SPH_type), pointer    :: RBC_SPH
    integer,           intent(in)  :: fileunit
    integer,           intent(out) :: error_flag
    logical, optional, intent(in)  :: Compare

    ! local variables
    character(len=max_str_len) :: exp_name
    character(len=max_str_len) :: value
    character(len=max_str_len) :: exp_phys_unit
    integer                    :: exp_val_length
    integer                    :: AllocStatus
    logical                    :: do_Compare

    !---------------------------------
    ! Reading the RBC SPH KVT elements
    !---------------------------------

    error_flag = no_error
    IF (.not. associated(RBC_SPH)) THEN
       print *,"ERROR in Read_RBC_SPH_from_KVT():"
       print *,"The pointer to this RBC_SPH seems not associated !!"
       error_flag = error_programming
       return
    END IF

    do_Compare = .false.
    IF (present(Compare)) do_Compare = Compare

    IF (do_Compare) THEN
       print *,"Read_RBC_SPH_from_KVT: reading and comparing RBC SPH-KVT items"
       print *,"with the XML items in memory"
    ELSE
       print *,"Read_RBC_MPH_from_KVT: reading RBC SPH-KVT items "
    END IF

    ! read the SPH_Descriptor 
    exp_val_length = len(RBC_SPH%SPH_Descriptor)
    exp_name       = "Sph_Descriptor"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expect_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(RBC_SPH%SPH_Descriptor, value, &
                                    "RBC_SPH%Sph_Descriptor",error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       RBC_SPH%SPH_Descriptor = value
    END IF

    ! Spare_1: this is a spare value without content
    exp_val_length = 40
    call read_spare_kvt_string(fileunit,exp_val_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! read the Ref_RBC_Suite
    exp_val_length = len(RBC_SPH%Ref_RBC_Suite)
    exp_name       = "Ref_RBC_Suite"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expect_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(RBC_SPH%Ref_RBC_Suite, value, &
                                    "RBC_SPH%Ref_RBC_Suite",error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       RBC_SPH%Ref_RBC_Suite = value
    END IF

    ! read the Num_P
    exp_val_length = len(RBC_SPH%Num_P)
    exp_name       = "Num_P"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(RBC_SPH%Num_P, value, &
            "RBC_SPH%Num_P",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       RBC_SPH%Num_P = value
    END IF

    ! read the Num_T
    exp_val_length = len(RBC_SPH%Num_T)
    exp_name       = "Num_T"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(RBC_SPH%Num_T, value, &
            "RBC_SPH%Num_T",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       RBC_SPH%Num_T = value
    END IF

    ! read the Num_F
    exp_val_length = len(RBC_SPH%Num_F)
    exp_name       = "Num_F"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(RBC_SPH%Num_F, value, &
            "RBC_SPH%Num_F",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       RBC_SPH%Num_F = value
    END IF

    ! read the Num_FP
    exp_val_length = len(RBC_SPH%Num_FP)
    exp_name       = "Num_FP"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(RBC_SPH%Num_FP, value, &
            "RBC_SPH%Num_FP",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       RBC_SPH%Num_FP = value
    END IF

    ! read the Num_Fd
    exp_val_length = len(RBC_SPH%Num_Fd)
    exp_name       = "Num_Fd"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(RBC_SPH%Num_Fd, value, &
            "RBC_SPH%Num_Fd",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       RBC_SPH%Num_Fd = value
    END IF

    ! read the Num_RR
    exp_val_length = len(RBC_SPH%Num_RR)
    exp_name       = "Num_RR"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(RBC_SPH%Num_RR, value, &
            "RBC_SPH%Num_RR",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       RBC_SPH%Num_RR = value
    END IF

    ! Spare_2: this is a spare value without content
    exp_val_length = 40
    call read_spare_kvt_string(fileunit,exp_val_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! read the P_min
    exp_val_length = len(RBC_SPH%P_min)
    exp_name       = "P_min"
    exp_phys_unit  = "Pa"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,&
         expected_phys_unit=exp_phys_unit, is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(RBC_SPH%P_min, value, &
            "RBC_SPH%P_min",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       RBC_SPH%P_min = value
    END IF

    ! read the P_max
    exp_val_length = len(RBC_SPH%P_max)
    exp_name       = "P_max"
    exp_phys_unit  = "Pa"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,&
         expected_phys_unit=exp_phys_unit, is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(RBC_SPH%P_max, value, &
            "RBC_SPH%P_max",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       RBC_SPH%P_max = value
    END IF

    ! read the T_min
    exp_val_length = len(RBC_SPH%T_min)
    exp_name       = "T_min"
    exp_phys_unit  = "10-2K"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,&
         expected_phys_unit=exp_phys_unit, is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(RBC_SPH%T_min, value, &
            "RBC_SPH%T_min",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       RBC_SPH%T_min = value
    END IF

    ! read the T_max
    exp_val_length = len(RBC_SPH%T_max)
    exp_name       = "T_max"
    exp_phys_unit  = "10-2K"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,&
         expected_phys_unit=exp_phys_unit, is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(RBC_SPH%T_max, value, &
            "RBC_SPH%T_max",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       RBC_SPH%T_max = value
    END IF

    !
    ! read the Laser_Freq_Offset_Start
    exp_val_length = len(RBC_SPH%Laser_Freq_Offset_Start)
    exp_name       = "Laser_Freq_Offset_Start"
    exp_phys_unit  = "MHz"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,&
         expected_phys_unit=exp_phys_unit, is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(RBC_SPH%Laser_Freq_Offset_Start, value, &
            "RBC_SPH%Laser_Freq_Offset_Start",error_flag,CompareAsReal=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       RBC_SPH%Laser_Freq_Offset_Start = value
    END IF

    ! read the Laser_Freq_Offset_Stop
    exp_val_length = len(RBC_SPH%Laser_Freq_Offset_Stop)
    exp_name       = "Laser_Freq_Offset_Stop"
    exp_phys_unit  = "MHz"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,&
         expected_phys_unit=exp_phys_unit, is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(RBC_SPH%Laser_Freq_Offset_Stop, value, &
            "RBC_SPH%Laser_Freq_Offset_Stop",error_flag,CompareAsReal=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       RBC_SPH%Laser_Freq_Offset_Stop = value
    END IF

    ! read the Total_Num_of_Observations
    exp_val_length = len(RBC_SPH%Total_Num_of_Observations)
    exp_name       = "Total_Num_of_Observations"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(RBC_SPH%Total_Num_of_Observations, value, &
            "RBC_SPH%Total_Num_of_Observations",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       RBC_SPH%Total_Num_of_Observations = value
    END IF

    ! read the Total_Num_of_Measurements
    exp_val_length = len(RBC_SPH%Total_Num_of_Measurements)
    exp_name       = "Total_Num_of_Measurements"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(RBC_SPH%Total_Num_of_Measurements, value, &
            "RBC_SPH%Total_Num_of_Measurements",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       RBC_SPH%Total_Num_of_Measurements = value
    END IF

    ! read the Total_Num_of_Reference_Pulses
    exp_val_length = len(RBC_SPH%Total_Num_of_Reference_Pulses)
    exp_name       = "Total_Num_of_Reference_Pulses"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(RBC_SPH%Total_Num_of_Reference_Pulses, value, &
            "RBC_SPH%Total_Num_of_Reference_Pulses",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       RBC_SPH%Total_Num_of_Reference_Pulses = value
    END IF

    ! read the Total_Num_of_Corrupt_Mie_Meas
    exp_val_length = len(RBC_SPH%Total_Num_of_Corrupt_Mie_Meas)
    exp_name       = "Total_Num_of_Corrupt_Mie_Meas"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(RBC_SPH%Total_Num_of_Corrupt_Mie_Meas, value, &
            "RBC_SPH%Total_Num_of_Corrupt_Mie_Meas",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       RBC_SPH%Total_Num_of_Corrupt_Mie_Meas = value
    END IF

    ! read the Total_Num_of_Corrupt_Ray_Meas
    exp_val_length = len(RBC_SPH%Total_Num_of_Corrupt_Ray_Meas)
    exp_name       = "Total_Num_of_Corrupt_Ray_Meas"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(RBC_SPH%Total_Num_of_Corrupt_Ray_Meas, value, &
            "RBC_SPH%Total_Num_of_Corrupt_Ray_Meas",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       RBC_SPH%Total_Num_of_Corrupt_Ray_Meas = value
    END IF

    ! read the Total_Num_of_Corrupt_Mie_RefP
    exp_val_length = len(RBC_SPH%Total_Num_of_Corrupt_Mie_RefP)
    exp_name       = "Total_Num_of_Corrupt_Mie_RefP"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(RBC_SPH%Total_Num_of_Corrupt_Mie_RefP, value, &
            "RBC_SPH%Total_Num_of_Corrupt_Mie_RefP",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       RBC_SPH%Total_Num_of_Corrupt_Mie_RefP = value
    END IF

    ! read the Total_Num_of_Corrupt_Ray_RefP
    exp_val_length = len(RBC_SPH%Total_Num_of_Corrupt_Ray_RefP)
    exp_name       = "Total_Num_of_Corrupt_Ray_RefP"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(RBC_SPH%Total_Num_of_Corrupt_Ray_RefP, value, &
            "RBC_SPH%Total_Num_of_Corrupt_Ray_RefP",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       RBC_SPH%Total_Num_of_Corrupt_Ray_RefP = value
    END IF

    ! read the Average_Error_FP_Response_A
    exp_val_length = len(RBC_SPH%Average_Error_FP_Response_A)
    exp_name       = "Average_Error_FP_Response_A"
!    exp_phys_unit  = "AU"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,&        ! expected_phys_unit=exp_phys_unit, 
         is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(RBC_SPH%Average_Error_FP_Response_A, value, &
            "RBC_SPH%Average_Error_FP_Response_A",error_flag,CompareAsReal=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       RBC_SPH%Average_Error_FP_Response_A = value
    END IF

    ! read the Average_Error_FP_Response_B
    exp_val_length = len(RBC_SPH%Average_Error_FP_Response_B)
    exp_name       = "Average_Error_FP_Response_B"
!    exp_phys_unit  = "AU"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,&        ! expected_phys_unit=exp_phys_unit, 
         is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(RBC_SPH%Average_Error_FP_Response_B, value, &
            "RBC_SPH%Average_Error_FP_Response_B",error_flag,CompareAsReal=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       RBC_SPH%Average_Error_FP_Response_B = value
    END IF

    ! Spare_3: this is a spare value without content
    exp_val_length = 40
    call read_spare_kvt_string(fileunit,exp_val_length,error_flag)
    IF (error_flag .ne. no_error) goto 998

    IF (.not. associated(RBC_SPH%RBC_DSD)) THEN
       ! allow this, to enable the possibility of reading the dbl file only,
       ! without reading the HDR files.
       allocate(RBC_SPH%RBC_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_RBC_SPH_from_KVT()"
          error_flag = error_allocate
          goto 998
       END IF
       call Init_DSD(RBC_SPH%RBC_DSD,error_flag)
       IF (error_flag .ne. no_error) goto 998

       ! Read the RBC_DSD without comparing
       call Read_DSD_from_KVT(RBC_SPH%RBC_DSD,fileunit,&
                              RBC_DS_Name_RBC_ADS, &
                              RBC_DS_Type_RBC_ADS, &
                              error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the RBC_DSD and compare if needed
       call Read_DSD_from_KVT(RBC_SPH%RBC_DSD,fileunit,&
                              RBC_DS_Name_RBC_ADS, &
                              RBC_DS_Type_RBC_ADS, &
                              error_flag,Compare)
       IF (error_flag .ne. no_error) goto 998
    END IF

    IF (.not. associated(RBC_SPH%FP_DSD)) THEN
       ! allow this, to enable the possibility of reading the dbl file only,
       ! without reading the HDR files.
       allocate(RBC_SPH%FP_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_RBC_SPH_from_KVT()"
          error_flag = error_allocate
          goto 998
       END IF
       call Init_DSD(RBC_SPH%FP_DSD,error_flag)
       IF (error_flag .ne. no_error) goto 998

       ! Read the FP_DSD without comparing
       call Read_DSD_from_KVT(RBC_SPH%FP_DSD,fileunit,&
                              RBC_DS_Name_FP_MDS, &
                              RBC_DS_Type_FP_MDS, error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the FP_DSD and compare if needed
       call Read_DSD_from_KVT(RBC_SPH%FP_DSD,fileunit,&
                              RBC_DS_Name_FP_MDS, &
                              RBC_DS_Type_FP_MDS, &
                              error_flag,Compare)
       IF (error_flag .ne. no_error) goto 998
    END IF

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Read_RBC_SPH_from_KVT"
    return

  end subroutine Read_RBC_SPH_from_KVT
    !  #]
  !---------------------------------
  function GetByteSize_RBC_SPH_KVT() result(KVT_bytesize)
    !  #[
    integer(IntAl) :: KVT_bytesize
    
    ! KVT sizes have size(key)+size(=)+size(value)+size(terminator)
    KVT_bytesize = &
         !     len(key)          len(=)  len(value)     len(unit)   len(terminator)
         !                                           or len(quotes)
         int(len("Sph_Descriptor") + 1 + maxlen_SPH_Descriptor + 2      + 1 + &
             40+1+& ! spare_1
             len("Ref_RBC_Suite")  + 1 + maxlen_Ref_RBC_Suite  + 2      + 1 + &
             len("Num_P")          + 1 + maxlen_IntAs                   + 1 + &
             len("Num_T")          + 1 + maxlen_IntAs                   + 1 + &
             len("Num_F")          + 1 + maxlen_IntAs                   + 1 + &
             len("Num_FP")         + 1 + maxlen_IntAs                   + 1 + &
             len("Num_Fd")         + 1 + maxlen_IntAs                   + 1 + &
             len("Num_RR")         + 1 + maxlen_IntAs                   + 1 + &
             40+1+& ! spare_2
             len("P_min")          + 1 + maxlen_IntAul + len("<Pa>")    + 1 + &
             len("P_max")          + 1 + maxlen_IntAul + len("<Pa>")    + 1 + &
             len("T_min")          + 1 + maxlen_IntAus + len("<10-2K>") + 1 + &
             len("T_max")          + 1 + maxlen_IntAus + len("<10-2K>") + 1   ,IntAl)
!
    KVT_bytesize = KVT_bytesize + &
         int(len("Laser_Freq_Offset_Start")       + 1 + maxlen_FAdo56 + len("<MHz>") + 1 + &
             len("Laser_Freq_Offset_Stop")        + 1 + maxlen_FAdo56 + len("<MHz>") + 1 + &
             len("Total_Num_of_Observations")     + 1 + maxlen_IntAl                 + 1 + &
             len("Total_Num_of_Measurements")     + 1 + maxlen_IntAl                 + 1 + &
             len("Total_Num_of_Reference_Pulses") + 1 + maxlen_IntAl                 + 1 + &
             len("Total_Num_of_Corrupt_Mie_Meas") + 1 + maxlen_IntAl                 + 1 + &
             len("Total_Num_of_Corrupt_Ray_Meas") + 1 + maxlen_IntAl                 + 1 + &
             len("Total_Num_of_Corrupt_Mie_RefP") + 1 + maxlen_IntAl                 + 1 + &
             len("Total_Num_of_Corrupt_Ray_RefP") + 1 + maxlen_IntAl                 + 1 + &
             len("Average_Error_FP_Response_A")   + 1 + maxlen_FAdo36                + 1 + &
             len("Average_Error_FP_Response_B")   + 1 + maxlen_FAdo36                + 1 + &
             !
             40+1+& ! spare_3
             2 * GetByteSize_DSD_KVT()  ,IntAl)

!         print *,"tests lengths of items in RBC_SPH KVT part"
!         print *,len("Sph_Descriptor") + 1 + maxlen_SPH_Descriptor + 2      + 1
!         print *,40+1
!         print *,len("Ref_RBC_Suite")  + 1 + maxlen_Ref_RBC_Suite  + 2      + 1
!         print *,len("Num_P")          + 1 + maxlen_IntAs                   + 1
!         print *,len("Num_T")          + 1 + maxlen_IntAs                   + 1
!         print *,len("Num_F")          + 1 + maxlen_IntAs                   + 1
!         print *,len("Num_FP")         + 1 + maxlen_IntAs                   + 1
!         print *,len("Num_Fd")         + 1 + maxlen_IntAs                   + 1
!         print *,len("Num_RR")         + 1 + maxlen_IntAs                   + 1
!         print *,40+1
!         print *,len("P_min")          + 1 + maxlen_IntAul + len("<Pa>")    + 1
!         print *,len("P_max")          + 1 + maxlen_IntAul + len("<Pa>")    + 1
!         print *,len("T_min")          + 1 + maxlen_IntAus + len("<10-2K>") + 1
!         print *,len("T_max")          + 1 + maxlen_IntAus + len("<10-2K>") + 1
!
!         print *,len("Laser_Freq_Offset_Start")       + 1 + maxlen_FAdo56 + len("<MHz>") + 1
!         print *,len("Laser_Freq_Offset_Stop")        + 1 + maxlen_FAdo56 + len("<MHz>") + 1
!         print *,len("Total_Num_of_Observations")     + 1 + maxlen_IntAl                 + 1
!         print *,len("Total_Num_of_Measurements")     + 1 + maxlen_IntAl                 + 1
!         print *,len("Total_Num_of_Reference_Pulses") + 1 + maxlen_IntAl                 + 1
!         print *,len("Total_Num_of_Corrupt_Mie_Meas") + 1 + maxlen_IntAl                 + 1
!         print *,len("Total_Num_of_Corrupt_Ray_Meas") + 1 + maxlen_IntAl                 + 1
!         print *,len("Total_Num_of_Corrupt_Mie_RefP") + 1 + maxlen_IntAl                 + 1
!         print *,len("Total_Num_of_Corrupt_Ray_RefP") + 1 + maxlen_IntAl                 + 1
!         print *,len("Average_Error_FP_Response_A")   + 1 + maxlen_FAdo36                + 1
!         print *,len("Average_Error_FP_Response_B")   + 1 + maxlen_FAdo36                + 1
!
!         print *,40+1
!         print *,2 * GetByteSize_DSD_KVT()
    
  end function GetByteSize_RBC_SPH_KVT
    !  #]
  !---------------------------------
end module RBC_SPH
