module Working_DataStructure
  !  #[ Description
  !     a module to define a working datastructure used by the L2B processor,
  !     for storing intermediate results
  !     and the routines needed for initialisation, allocation
  !     and deallocation of all data structs inside it.
  !     
  !     split from the L1B_DataStructure/L2B_DataStructure 
  !     on 15-Aug-2006, by: J. de Kloe, KNMI
  !
  !     Modifications:
  !     16-Aug-2006 J. de Kloe  print/AsciiWrite routines added
  !     05-Oct-2006 J. de Kloe  removed nr_of_measurements_max/actual
  !                             and replaced by nr_of_measurements
  !                             inside WorkingDataType
  !     31-Oct-2006 J. de Kloe  added datetime elements to Mie/Rayl.Meas
  !     10-Nov-2006 J. de Kloe  moved Copy_L1B_fields2WorkingArrays from 
  !                             l2bp_module.F90 to this file
  !     02-Mar-2007 J. de Kloe  adapted to parameter type change in call to 
  !                             GetObsTypeText()
  !     28-Aug-2007 D. Tan      Phase out L1B-working elements
  !     03-Sep-2007 P. Poli     Modified arguments of routine 
  !                             Copy_L1B_fields2WorkingArrays
  !                             change L1B_DataType to L1B_BRC_DataType
  !     16-Jan-2008 J. de Kloe  phase out integer kind i_
  !
  !  #]
  !  #[ modules used

  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE L2B_parameters, only: not_yet_classified, GetClassTypeText, &
                            Obs_Type_undefined, GetObsTypeText
  USE EE_CFI_Datatypes, only: missing_indicator_FAdoxy, &
       FAdoxy, maxlen_FAdoxy_full, SetFadoFull_or_Missing, &
       IntAs,  maxlen_IntAs,       SetIntAs_or_Missing, &
       IntAc,  DateTimeType,       maxlen_DateTime_KVT, &
       IntAuc, &
       Construct_KVT_DateTime, InitDateTime
  USE L1B_DataStructure, only: L1B_BRC_DataType
  USE L1B_Geoloc_ADS, only: Rangebin_Meas_Geoloc_type, &
                            Rangebin_Obs_Geoloc_type

  ! support modules
  USE numerics, only: i1_, r8_, missing_indicator_real
  USE ErrorHandler, only: no_error, error_allocate, error_writing_file
  USE LunManager, only: fileunit_stdout
  USE AeolusConstants, only: N_range_gates

  !  #]
  !  #[ Parameter and data type definitions
  implicit none

  ! make all module contents public
  public

  TYPE MieRangeGateType
     !  #[
     ! copied L1B items (or missing if they failed InputScreening)
     !
     real(FAdoxy) :: SatRange  ! [m]   value in the centre of the range-bin
     real(FAdoxy) :: IncAngle  ! [deg] value in the centre of the range-bin
     real(FAdoxy) :: AziAngle  ! [deg] value in the centre of the range-bin
     real(FAdoxy) :: lat       ! [deg] latitude
     real(FAdoxy) :: lon       ! [deg] longitude
     real(FAdoxy) :: alt_start ! [m]   bottom altitude of this range gate
     real(FAdoxy) :: alt_end   ! [m]   top altitude fo this range gate
     real(FAdoxy) :: obs_signal_level ! signal level as observed
     !
     ! copied AMD items (or missing if they failed InputScreening)
     !
     real(FAdoxy) :: p ! [Pa]  value in the centre of the range-bin
     real(FAdoxy) :: T ! [K]   value in the centre of the range-bin
     real(FAdoxy) :: u ! [m/s] value in the centre of the range-bin
     real(FAdoxy) :: v ! [m/s] value in the centre of the range-bin
     !
     ! generated intermediate data
     !
     ! the first and last rayleigh range-bins that overlap 
     ! with this mie range-bin (in case they have different definitions)
     ! (usefull when combining Mie and Rayleigh during the classification)
     integer(IntAs) :: first_rayleigh_rangebin
     integer(IntAs) :: last_rayleigh_rangebin
     !
     real(FAdoxy) :: alt_vcog
     real(FAdoxy) :: attenuation_above
     real(FAdoxy) :: attenuation_this_bin
     real(FAdoxy) :: extinction
     ! should we renamed the next item to rho?
     real(FAdoxy) :: backscatter_ratio 
     ! a flag to store the validity of the backscatter_ratio number
     ! (determined from the Mie spectrum) in the L1B input file. 
     ! Set in inputscreening. Default: .true.
     ! maybe this should be renamed to rho_Mie_ok to make clear
     ! what it means ?
     logical      :: rho_ok
     !
     integer      :: class ! classification result
     ! can be: cloud, no_cloud (later also: dust? or others?)
     real(FAdoxy) :: weight 
     ! note that this real has much more precision then the integer weight
     ! defined in the MeasMap. That one is not to be used for calculations,
     ! only for exporting to the L2B outputfile
  END TYPE MieRangeGateType
     !  #]

  TYPE MieMeasurementType
     !  #[
     integer(IntAs) :: nr_of_range_gates
     TYPE(DateTimeType) :: datetime
     TYPE(MieRangeGateType), dimension(:), pointer :: RangeGate
  END TYPE MieMeasurementType
     !  #]

  TYPE RayleighRangeGateType
     !  #[
     ! copied L1B items (or missing if they failed InputScreening)
     !
     real(FAdoxy) :: SatRange  ! [m]   value in the centre of the range-bin
     real(FAdoxy) :: IncAngle  ! [deg] value in the centre of the range-bin
     real(FAdoxy) :: AziAngle  ! [deg] value in the centre of the range-bin
     real(FAdoxy) :: lat       ! [deg] latitude 
     real(FAdoxy) :: lon       ! [deg] longitude
     real(FAdoxy) :: alt_start ! [m]   bottom altitude of this range gate
     real(FAdoxy) :: alt_end   ! [m]   top altitude fo this range gate
     real(FAdoxy) :: obs_signal_level ! signal level as observed
     !
     ! copied AMD items (or missing if they failed InputScreening)
     !
     real(FAdoxy) :: p ! [Pa]  value in the centre of the range-bin
     real(FAdoxy) :: T ! [K]   value in the centre of the range-bin
     real(FAdoxy) :: u ! [m/s] value in the centre of the range-bin
     real(FAdoxy) :: v ! [m/s] value in the centre of the range-bin
     !
     ! generated intermediate data
     !
     ! the first and last mie range-bins that overlap 
     ! with this rayleigh range-bin (in case they have different definitions)
     ! (usefull during estimation of Mie contamination)
     integer(IntAs) :: first_mie_rangebin
     integer(IntAs) :: last_mie_rangebin
     !
     real(FAdoxy) :: alt_vcog
     real(FAdoxy) :: calc_mol_backscatter ! can be calculated from P and T
     real(FAdoxy) :: calc_mol_extinction  ! can be calculated from P and T
     real(FAdoxy) :: attenuation_above
     real(FAdoxy) :: attenuation_this_bin
     real(FAdoxy) :: extinction
     ! should we renamed the next item to rho?
     real(FAdoxy) :: backscatter_ratio
     ! a flag to store the validity of the backscatter_ratio number
     ! (determined from the Mie spectrum) in the L1B input file. 
     ! Set in inputscreening. Default: .true.
     ! maybe this should be renamed to rho_Mie_ok to make clear
     ! what it means ?
     logical      :: rho_ok
     !
     integer      :: class ! classification result
     ! can be: cloud, no_cloud (later also: dust? or others?)
     real(FAdoxy) :: weight 
     ! note that this real has much more precision then the integer weight
     ! defined in the MeasMap. That one is not to be used for calculations,
     ! only for exporting to the L2B outputfile
     !
  END TYPE RayleighRangeGateType
     !  #]

  TYPE RayleighMeasurementType
     !  #[
     integer(IntAs) :: nr_of_range_gates
     TYPE(DateTimeType) :: datetime
     TYPE(RayleighRangeGateType), dimension(:), pointer :: RangeGate
  END TYPE RayleighMeasurementType
     !  #]

  TYPE MieObsRangeGateType
     !  #[
     ! 
     ! generated intermediate observation scale data
     !
     ! elements describing the start/cog/stop geolocation, for which
     ! the hlos for this observation is determined
     real(FAdoxy) :: alt_start ! [m] bottom altitude of this range gate
     real(FAdoxy) :: alt_cog   ! [m] altitude for the center of gravity 
     real(FAdoxy) :: alt_end   ! [m] top altitude fo this range gate
     real(FAdoxy) :: lat_start ! [deg]
     real(FAdoxy) :: lat_cog   ! [deg]
     real(FAdoxy) :: lat_stop  ! [deg]
     real(FAdoxy) :: lon_start ! [deg]
     real(FAdoxy) :: lon_cog   ! [deg]
     real(FAdoxy) :: lon_stop  ! [deg]
     !TYPE(DateTimeType) :: date_cog
     !TYPE(DateTimeType) :: time_cog

     ! nr used to construct this observation
     integer(IntAs) :: nr_of_measurements 

     ! elements describing the reference temperature and pressure,
     ! as used for Rayleigh-Brillouin correction of the hlos
     real(FAdoxy) :: T_ref
     real(FAdoxy) :: p_ref

     ! the final result of the L2B processor: the hlos wind component
     real(FAdoxy) :: hlos_cog
  END TYPE MieObsRangeGateType
     !  #]

  TYPE MieObservationType
     !  #[
     integer(IntAs) :: observation_type ! cloud/no_cloud etc.
     integer(IntAs) :: nr_of_range_gates
     TYPE(MieObsRangeGateType), dimension(:), pointer :: RangeGate
  END TYPE MieObservationType
     !  #]

  TYPE RayleighObsRangeGateType
     !  #[
     ! 
     ! generated intermediate observation scale data
     !
     ! elements describing the start/cog/stop geolocation, for which
     ! the hlos for this observation is determined
     real(FAdoxy) :: alt_start ! [m] bottom altitude of this range gate
     real(FAdoxy) :: alt_cog   ! [m] altitude for the center of gravity 
     real(FAdoxy) :: alt_end   ! [m] top altitude fo this range gate
     real(FAdoxy) :: lat_start ! [deg]
     real(FAdoxy) :: lat_cog   ! [deg]
     real(FAdoxy) :: lat_stop  ! [deg]
     real(FAdoxy) :: lon_start ! [deg]
     real(FAdoxy) :: lon_cog   ! [deg]
     real(FAdoxy) :: lon_stop  ! [deg]
     !TYPE(DateTimeType) :: date_cog
     !TYPE(DateTimeType) :: time_cog

     ! nr used to construct this observation
     integer(IntAs) :: nr_of_measurements 

     ! elements describing the reference temperature and pressure,
     ! as used for Rayleigh-Brillouin correction of the hlos
     real(FAdoxy) :: T_ref
     real(FAdoxy) :: p_ref

     ! the final result of the L2B processor: the hlos wind component
     real(FAdoxy) :: hlos_cog
  END TYPE RayleighObsRangeGateType
     !  #]
  
  TYPE RayleighObservationType
     !  #[
     integer(IntAs) :: observation_type ! cloud/no_cloud etc.
     integer(IntAs) :: nr_of_range_gates
     TYPE(RayleighObsRangeGateType), dimension(:), pointer :: RangeGate
  END TYPE RayleighObservationType
     !  #]

  TYPE WorkingDataType
     !  #[
     ! to store the actual number of measurements
     integer(IntAs) :: nr_of_range_gates
     integer(IntAs) :: nr_of_measurements
     integer(IntAs) :: nr_of_observations_mie_max
     integer(IntAs) :: nr_of_observations_mie_actual
     integer(IntAs) :: nr_of_observations_rayleigh_max
     integer(IntAs) :: nr_of_observations_rayl_actual
     !

     ! todo: store
     ! n_pulses_per_BRC = n_meas*n_pulses 
     ! if (n_pulses_per_BRC .gt. 700) then
     !    warm-up pulses have been used
     !
     ! not used pulses: nu = 1005-n_pulses_per_BRC
     !
     ! so:
     ! for current meas m
     !
     ! pulses used are: nu + (m-1)*p + 1 ... m*p
     ! so pulse_1 = nu + (m-1)*p + 1
     ! so pulse_2 = nu +  m   *p     = pulse_1+p-1
     !
     ! if pulse_1 .le. 305 warmup_pulses_used = .true. for this meas

     ! these define working arrays to store intermediate results
     ! for which we have no place in the L1B or L2B structures
     TYPE(MieMeasurementType),      dimension(:), pointer :: MieMeas
     TYPE(RayleighMeasurementType), dimension(:), pointer :: RayleighMeas
     TYPE(MieObservationType),      dimension(:), pointer :: MieObs
     TYPE(RayleighObservationType), dimension(:), pointer :: RayleighObs
     !
  END TYPE WorkingDataType
     !  #]

  !  #]
  !  #[ Module interface
  !     everything is public in this module
  !  #]
contains
  !---------------------------------------------
  subroutine InitMieRangeGate(MieRangegate)
    !  #[
    TYPE(MieRangeGateType), intent(out) :: MieRangegate

    MieRangegate%SatRange         = missing_indicator_FAdoxy
    MieRangegate%IncAngle         = missing_indicator_FAdoxy
    MieRangegate%AziAngle         = missing_indicator_FAdoxy
    MieRangegate%lat              = missing_indicator_FAdoxy
    MieRangegate%lon              = missing_indicator_FAdoxy
    MieRangegate%alt_start        = missing_indicator_FAdoxy
    MieRangegate%alt_end          = missing_indicator_FAdoxy
    MieRangegate%obs_signal_level = missing_indicator_FAdoxy
    !
    MieRangegate%p                = missing_indicator_FAdoxy
    MieRangegate%T                = missing_indicator_FAdoxy
    MieRangegate%u                = missing_indicator_FAdoxy
    MieRangegate%v                = missing_indicator_FAdoxy
    !
    MieRangegate%first_rayleigh_rangebin   = -1
    MieRangegate%last_rayleigh_rangebin    = -1
    !
    MieRangegate%alt_vcog             = missing_indicator_FAdoxy
    MieRangegate%attenuation_above    = missing_indicator_FAdoxy
    MieRangegate%attenuation_this_bin = missing_indicator_FAdoxy
    MieRangegate%extinction           = missing_indicator_FAdoxy
    MieRangegate%backscatter_ratio    = missing_indicator_FAdoxy
    MieRangeGate%rho_ok               = .true.
    MieRangegate%class                = not_yet_classified
    MieRangegate%weight               = missing_indicator_FAdoxy

  end subroutine InitMieRangeGate
    !  #]
  subroutine PrintMieRangeGate(MieRangegate,prefix,error_flag)
    !  #[
    TYPE(MieRangeGateType), intent(in)  :: MieRangegate ! input
    character(len=*),       intent(in)  :: prefix       ! input
    integer,                intent(out) :: error_flag   ! output

    call AsciiWriteMieRangeGate(MieRangegate,fileunit_stdout,prefix,error_flag)

  end subroutine PrintMieRangeGate
    !  #]
  subroutine AsciiWriteMieRangeGate(MieRangegate,fileunit,prefix,error_flag)
    !  #[

    TYPE(MieRangeGateType), intent(in)  :: MieRangegate ! input
    integer,                intent(in)  :: fileunit     ! input
    character(len=*),       intent(in)  :: prefix       ! input
    integer,                intent(out) :: error_flag   ! output

    ! local variables
    character(len=maxlen_IntAs)       :: str_IntAs
    character(len=maxlen_FAdoxy_full) :: str_FAdoxy
    character(len=256)                :: class_str
    
    error_flag = no_error

    write(fileunit,"(a)",err=999) "REALACC(12)"

    call SetFAdoFull_or_Missing(MieRangegate%SatRange,str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%SatRange                  = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieRangegate%IncAngle,str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%IncAngle                  = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieRangegate%AziAngle,str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%AziAngle                  = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieRangegate%lat,str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%lat                       = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieRangegate%lon,str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%lon                       = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieRangegate%alt_start,str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%alt_start                 = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieRangegate%alt_end,str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%alt_end                   = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieRangegate%obs_signal_level,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%obs_signal_level          = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieRangegate%p,str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%p                         = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieRangegate%T,str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%T                         = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieRangegate%u,str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%u                         = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieRangegate%v,str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%v                         = "//trim(str_FAdoxy)

    call SetIntAs_or_Missing(MieRangegate%first_rayleigh_rangebin,&
                             str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%first_rayleigh_rangebin   = "//trim(str_IntAs)

    call SetIntAs_or_Missing(MieRangegate%last_rayleigh_rangebin,&
                             str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%last_rayleigh_rangebin    = "//trim(str_IntAs)

    call SetFAdoFull_or_Missing(MieRangegate%alt_vcog,str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%alt_vcog                  = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieRangegate%attenuation_above,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%attenuation_above         = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieRangegate%attenuation_this_bin,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%attenuation_this_bin      = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieRangegate%extinction,str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%extinction                = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieRangegate%backscatter_ratio,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%backscatter_ratio         = "//trim(str_FAdoxy)

    if (MieRangeGate%rho_ok) then
       write(fileunit,"(a)",err=999) &
            trim(prefix)//"%rho_ok                    = .true."
    else
       write(fileunit,"(a)",err=999) &
            trim(prefix)//"%rho_ok                    = .false."
    end if

    class_str = GetClassTypeText(MieRangegate%class)
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%class                     = "//trim(class_str)

    call SetFAdoFull_or_Missing(MieRangegate%weight,str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%weight                    = "//trim(str_FAdoxy)

    write(fileunit,"(a)",err=999) "ENDREALACC"

    return

999 print *,"ERROR in AsciiWriteMieRangeGate"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWriteMieRangeGate
    !  #]
  !---------------------------------------------
  subroutine InitMieMeasurement(MieMeas,nr_of_range_gates,error_flag)
    !  #[
    TYPE(MieMeasurementType), intent(out) :: MieMeas ! output
    integer(IntAs), intent(in)  :: nr_of_range_gates    ! input
    integer,        intent(out) :: error_flag           ! output

    ! local variable
    integer :: i, AllocStatus

    error_flag = no_error
    MieMeas%nr_of_range_gates = nr_of_range_gates
    allocate(MieMeas%RangeGate(nr_of_range_gates),STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation failure in subroutine InitMieMeasurement"
       error_flag = error_allocate
       return
    END IF

    call InitDateTime(MieMeas%datetime)

    DO i=1,MieMeas%nr_of_range_gates
       call InitMieRangeGate(MieMeas%RangeGate(i))
    END DO

  end subroutine InitMieMeasurement
    !  #]
  subroutine DeleteMieMeasurement(MieMeas)
    !  #[
    TYPE(MieMeasurementType), intent(inout) :: MieMeas

    IF (associated(MieMeas%RangeGate)) deallocate(MieMeas%RangeGate)

  end subroutine DeleteMieMeasurement
    !  #]
  subroutine PrintMieMeasurement(MieMeas,prefix,PrintAll,error_flag)
    !  #[
    TYPE(MieMeasurementType), intent(in)  :: MieMeas    ! input
    character(len=*),         intent(in)  :: prefix     ! input
    logical,                  intent(in)  :: PrintAll   ! input
    integer,                  intent(out) :: error_flag ! output

    call AsciiWriteMieMeasurement(MieMeas,fileunit_stdout,prefix,PrintAll,&
                                  error_flag)

  end subroutine PrintMieMeasurement
    !  #]
  subroutine AsciiWriteMieMeasurement(MieMeas,fileunit,prefix,PrintAll,&
                                      error_flag)
    !  #[

    TYPE(MieMeasurementType), intent(in)  :: MieMeas    ! input
    integer,                  intent(in)  :: fileunit   ! input
    character(len=*),         intent(in)  :: prefix     ! input
    logical,                  intent(in)  :: PrintAll   ! input
    integer,                  intent(out) :: error_flag ! output

    ! local variables
    integer :: i,s,step
    character(len=maxlen_IntAs)        :: str_IntAs
    character(len=2)                   :: str_cntr
    character(len=maxlen_DateTime_KVT) :: DateTime_KVT ! result

    error_flag = no_error

    call SetIntAs_or_Missing(MieMeas%nr_of_range_gates,str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%nr_of_range_gates = "//trim(str_IntAs)

    DateTime_KVT = Construct_KVT_DateTime(MieMeas%datetime)
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%datetime          = "//trim(DateTime_KVT)

    IF (associated(MieMeas%RangeGate)) THEN
       s=MieMeas%nr_of_range_gates
       IF (PrintAll .or. (s .le. 1)) THEN
          step=1
       ELSE
          step = s - 1
       END IF
       DO i=1,s,step
          write(str_cntr,"(i2.2)",err=998) i
          call AsciiWriteMieRangeGate(&
               MieMeas%RangeGate(i), fileunit, &
               trim(prefix)//"%RangeGate("//trim(str_cntr)//")",error_flag)
          IF (error_flag .ne. no_error) goto 999
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%RangeGate(:) [seems not associated]"
    END IF


    return

998 print *,"conversion error in AsciiWriteMieMeasurement:"
    print *,"could not convert the counter ",i," to a string .."
    error_flag = error_writing_file
    return

999 print *,"ERROR in AsciiWriteMieMeasurement"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWriteMieMeasurement
    !  #]
  !---------------------------------------------
  subroutine InitRayleighRangeGate(RayleighRangegate)
    !  #[

    TYPE(RayleighRangeGateType), intent(out) :: RayleighRangegate

    RayleighRangegate%SatRange = missing_indicator_FAdoxy
    RayleighRangegate%IncAngle = missing_indicator_FAdoxy
    RayleighRangegate%AziAngle = missing_indicator_FAdoxy
    RayleighRangegate%lat      = missing_indicator_FAdoxy
    RayleighRangegate%lon      = missing_indicator_FAdoxy
    RayleighRangegate%alt_start        = missing_indicator_FAdoxy
    RayleighRangegate%alt_end          = missing_indicator_FAdoxy
    RayleighRangegate%obs_signal_level = missing_indicator_FAdoxy
    !
    RayleighRangegate%p = missing_indicator_FAdoxy
    RayleighRangegate%T = missing_indicator_FAdoxy
    RayleighRangegate%u = missing_indicator_FAdoxy
    RayleighRangegate%v = missing_indicator_FAdoxy
    !
    RayleighRangegate%first_mie_rangebin = -1
    RayleighRangegate%last_mie_rangebin  = -1
    !
    RayleighRangegate%alt_vcog             = missing_indicator_FAdoxy
    RayleighRangegate%calc_mol_backscatter = missing_indicator_FAdoxy
    RayleighRangegate%calc_mol_extinction  = missing_indicator_FAdoxy
    RayleighRangegate%attenuation_above    = missing_indicator_FAdoxy
    RayleighRangegate%attenuation_this_bin = missing_indicator_FAdoxy
    RayleighRangegate%extinction           = missing_indicator_FAdoxy
    RayleighRangegate%backscatter_ratio    = missing_indicator_FAdoxy
    RayleighRangeGate%rho_ok               = .true.
    RayleighRangegate%class                = not_yet_classified
    RayleighRangegate%weight               = missing_indicator_FAdoxy

  end subroutine InitRayleighRangeGate
    !  #]
  subroutine PrintRayleighRangeGate(RayleighRangegate,prefix,error_flag)
    !  #[
    TYPE(RayleighRangeGateType), intent(in)  :: RayleighRangegate ! input
    character(len=*),            intent(in)  :: prefix     ! input
    integer,                     intent(out) :: error_flag ! output

    call AsciiWriteRayleighRangeGate(RayleighRangegate,fileunit_stdout,&
                                     prefix,error_flag)

  end subroutine PrintRayleighRangeGate
    !  #]
  subroutine AsciiWriteRayleighRangeGate(RayleighRangegate,fileunit,prefix,&
                                         error_flag)
    !  #[

    TYPE(RayleighRangeGateType), intent(in)  :: RayleighRangegate ! input
    integer,                     intent(in)  :: fileunit          ! input
    character(len=*),            intent(in)  :: prefix            ! input
    integer,                     intent(out) :: error_flag        ! output

    ! local variables
    character(len=maxlen_IntAs)       :: str_IntAs
    character(len=maxlen_FAdoxy_full) :: str_FAdoxy
    character(len=256)                :: class_str
    
    error_flag = no_error

    write(fileunit,"(a)",err=999) "REALACC(12)"

    call SetFAdoFull_or_Missing(RayleighRangegate%SatRange,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%SatRange             = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RayleighRangegate%IncAngle,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%IncAngle             = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RayleighRangegate%AziAngle,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%AziAngle             = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RayleighRangegate%lat,str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%lat                  = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RayleighRangegate%lon,str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%lon                  = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RayleighRangegate%alt_start,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%alt_start            = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RayleighRangegate%alt_end,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%alt_end              = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RayleighRangegate%obs_signal_level,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%obs_signal_level     = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RayleighRangegate%p,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%p                    = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RayleighRangegate%T,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%T                    = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RayleighRangegate%u,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%u                    = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RayleighRangegate%v,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%v                    = "//trim(str_FAdoxy)

    call SetIntAs_or_Missing(RayleighRangegate%first_mie_rangebin,&
                             str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%first_mie_rangebin   = "//trim(str_IntAs)

    call SetIntAs_or_Missing(RayleighRangegate%last_mie_rangebin,&
                             str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%last_mie_rangebin    = "//trim(str_IntAs)

    call SetFAdoFull_or_Missing(RayleighRangegate%alt_vcog,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%alt_vcog             = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RayleighRangegate%calc_mol_backscatter,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%calc_mol_backscatter = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RayleighRangegate%calc_mol_extinction,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%calc_mol_extinction  = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RayleighRangegate%attenuation_above,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%attenuation_above    = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RayleighRangegate%attenuation_this_bin,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%attenuation_this_bin = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RayleighRangegate%extinction,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%extinction           = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RayleighRangegate%backscatter_ratio,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%backscatter_ratio    = "//trim(str_FAdoxy)

    if (RayleighRangeGate%rho_ok) then
       write(fileunit,"(a)",err=999) &
            trim(prefix)//"%rho_ok                    = .true."
    else
       write(fileunit,"(a)",err=999) &
            trim(prefix)//"%rho_ok                    = .false."
    end if

    class_str = GetClassTypeText(RayleighRangegate%class)
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%class                = "//trim(class_str)

    call SetFAdoFull_or_Missing(RayleighRangegate%weight,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%weight               = "//trim(str_FAdoxy)

    write(fileunit,"(a)",err=999) "ENDREALACC"

    return

999 print *,"ERROR in AsciiWriteRayleighRangeGate"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return
    
  end subroutine AsciiWriteRayleighRangeGate
    !  #]
  !---------------------------------------------
  subroutine InitRayleighMeasurement(RayleighMeas,nr_of_range_gates,error_flag)
    !  #[
    TYPE(RayleighMeasurementType), intent(out) :: RayleighMeas    ! output
    integer(IntAs), intent(in)  :: nr_of_range_gates ! input
    integer,        intent(out) :: error_flag        ! output

    ! local variable
    integer :: i, AllocStatus

    error_flag = no_error
    RayleighMeas%nr_of_range_gates = nr_of_range_gates
    allocate(RayleighMeas%RangeGate(nr_of_range_gates),STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation failure in subroutine InitRayleighMeasurement"
       error_flag = error_allocate
       return
    END IF

    call InitDateTime(RayleighMeas%datetime)

    DO i=1,RayleighMeas%nr_of_range_gates
       call InitRayleighRangeGate(RayleighMeas%RangeGate(i))
    END DO

  end subroutine InitRayleighMeasurement
    !  #]
  subroutine DeleteRayleighMeasurement(RayleighMeas)
    !  #[
    TYPE(RayleighMeasurementType), intent(inout) :: RayleighMeas

    IF (associated(RayleighMeas%RangeGate)) deallocate(RayleighMeas%RangeGate)

  end subroutine DeleteRayleighMeasurement
    !  #]
  subroutine PrintRayleighMeasurement(RayleighMeas,prefix,&
                                      PrintAll,error_flag)
    !  #[
    TYPE(RayleighMeasurementType), intent(in)  :: RayleighMeas    ! input
    character(len=*),              intent(in)  :: prefix     ! input
    logical,                       intent(in)  :: PrintAll   ! input
    integer,                       intent(out) :: error_flag ! output

    call AsciiWriteRayleighMeasurement(RayleighMeas,fileunit_stdout,prefix,&
                                       PrintAll,error_flag)

  end subroutine PrintRayleighMeasurement
    !  #]
  subroutine AsciiWriteRayleighMeasurement(RayleighMeas,fileunit,prefix,&
                                           PrintAll,error_flag)
    !  #[

    TYPE(RayleighMeasurementType), intent(in)  :: RayleighMeas    ! input
    integer,                       intent(in)  :: fileunit   ! input
    character(len=*),              intent(in)  :: prefix     ! input
    logical,                       intent(in)  :: PrintAll   ! input
    integer,                       intent(out) :: error_flag ! output

    ! local variables
    integer :: i,s,step
    character(len=maxlen_IntAs)        :: str_IntAs
    character(len=2)                   :: str_cntr
    character(len=maxlen_DateTime_KVT) :: DateTime_KVT ! result

    error_flag = no_error

    call SetIntAs_or_Missing(RayleighMeas%nr_of_range_gates,str_IntAs,&
                             error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%nr_of_range_gates = "//trim(str_IntAs)

    DateTime_KVT = Construct_KVT_DateTime(RayleighMeas%datetime)
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%datetime          = "//trim(DateTime_KVT)

    IF (associated(RayleighMeas%RangeGate)) THEN
       s=RayleighMeas%nr_of_range_gates
       IF (PrintAll .or. (s .le. 1)) THEN
          step=1
       ELSE
          step = s - 1
       END IF
       DO i=1,s,step
          write(str_cntr,"(i2.2)",err=998) i
          call AsciiWriteRayleighRangeGate(&
               RayleighMeas%RangeGate(i), fileunit, &
               trim(prefix)//"%RangeGate("//trim(str_cntr)//")",error_flag)
          IF (error_flag .ne. no_error) goto 999
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%RangeGate(:) [seems not associated]"
    END IF


    return

998 print *,"conversion error in AsciiWriteRayleighMeasurement:"
    print *,"could not convert the counter ",i," to a string .."
    error_flag = error_writing_file
    return

999 print *,"ERROR in AsciiWriteRayleighMeasurement"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWriteRayleighMeasurement
    !  #]
  !---------------------------------------------
  subroutine InitMieObsRangeGate(MieObsRangegate)
    !  #[
    TYPE(MieObsRangeGateType), intent(out) :: MieObsRangegate

    MieObsRangegate%alt_start = missing_indicator_FAdoxy
    MieObsRangegate%alt_cog   = missing_indicator_FAdoxy
    MieObsRangegate%alt_end   = missing_indicator_FAdoxy
    MieObsRangegate%lat_start = missing_indicator_FAdoxy
    MieObsRangegate%lat_cog   = missing_indicator_FAdoxy
    MieObsRangegate%lat_stop  = missing_indicator_FAdoxy
    MieObsRangegate%lon_start = missing_indicator_FAdoxy
    MieObsRangegate%lon_cog   = missing_indicator_FAdoxy
    MieObsRangegate%lon_stop  = missing_indicator_FAdoxy
    MieObsRangegate%nr_of_measurements = 0
    MieObsRangegate%T_ref     = missing_indicator_FAdoxy
    MieObsRangegate%p_ref     = missing_indicator_FAdoxy
    MieObsRangegate%hlos_cog  = missing_indicator_FAdoxy

  end subroutine InitMieObsRangeGate
    !  #]
  subroutine PrintMieObsRangeGate(MieObsRangegate,prefix,error_flag)
    !  #[
    TYPE(MieObsRangeGateType), intent(in)  :: MieObsRangegate ! input
    character(len=*),          intent(in)  :: prefix          ! input
    integer,                   intent(out) :: error_flag      ! output

    call AsciiWriteMieObsRangeGate(MieObsRangegate,fileunit_stdout,&
                                   prefix,error_flag)

  end subroutine PrintMieObsRangeGate
    !  #]
  subroutine AsciiWriteMieObsRangeGate(MieObsRangegate,fileunit,prefix,&
                                       error_flag)
    !  #[
    TYPE(MieObsRangeGateType), intent(in)  :: MieObsRangegate ! input
    integer,                   intent(in)  :: fileunit        ! input
    character(len=*),          intent(in)  :: prefix          ! input
    integer,                   intent(out) :: error_flag      ! output

    ! local variables
    character(len=maxlen_IntAs)       :: str_IntAs
    character(len=maxlen_FAdoxy_full) :: str_FAdoxy
    
    error_flag = no_error

    write(fileunit,"(a)",err=999) "REALACC(12)"

    call SetFAdoFull_or_Missing(MieObsRangegate%alt_start,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%alt_start          = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieObsRangegate%alt_cog,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%alt_cog            = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieObsRangegate%alt_end,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%alt_end            = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieObsRangegate%lat_start,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%lat_start          = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieObsRangegate%lat_cog,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%lat_cog            = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieObsRangegate%lat_stop,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%lat_stop           = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieObsRangegate%lon_start,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%lon_start          = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieObsRangegate%lon_cog,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%lon_cog            = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieObsRangegate%lon_stop,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%lon_stop           = "//trim(str_FAdoxy)

    call SetIntAs_or_Missing(MieObsRangegate%nr_of_measurements,&
                             str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%nr_of_measurements = "//trim(str_IntAs)

    call SetFAdoFull_or_Missing(MieObsRangegate%T_ref,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%T_ref              = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieObsRangegate%p_ref,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%p_ref              = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(MieObsRangegate%hlos_cog,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%hlos_cog           = "//trim(str_FAdoxy)

    write(fileunit,"(a)",err=999) "ENDREALACC"

    return

999 print *,"ERROR in AsciiWriteMieObsRangeGate"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return
    
  end subroutine AsciiWriteMieObsRangeGate
    !  #]
  !---------------------------------------------
  subroutine InitMieObservation(MieObs,nr_of_range_gates,error_flag)
    !  #[
    TYPE(MieObservationType), intent(out) :: MieObs            ! output
    integer(IntAs),           intent(in)  :: nr_of_range_gates ! input
    integer,                  intent(out) :: error_flag        ! output

    ! local variable
    integer :: i, AllocStatus

    error_flag = no_error

    MieObs%observation_type  = Obs_Type_undefined
    MieObs%nr_of_range_gates = nr_of_range_gates

    allocate(MieObs%RangeGate(nr_of_range_gates),STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation failure in subroutine InitMieObservation"
       error_flag = error_allocate
       return
    END IF

    DO i=1,MieObs%nr_of_range_gates
       call InitMieObsRangeGate(MieObs%RangeGate(i))
    END DO

  end subroutine InitMieObservation
    !  #]
  subroutine DeleteMieObservation(MieObs)
    !  #[
    TYPE(MieObservationType), intent(inout) :: MieObs

    IF (associated(MieObs%RangeGate)) deallocate(MieObs%RangeGate)

  end subroutine DeleteMieObservation
    !  #]
  subroutine PrintMieObservation(MieObs,prefix,PrintAll,error_flag)
    !  #[
    TYPE(MieObservationType), intent(in)  :: MieObs     ! input
    character(len=*),         intent(in)  :: prefix     ! input
    logical,                  intent(in)  :: PrintAll   ! input
    integer,                  intent(out) :: error_flag ! output

    call AsciiWriteMieObservation(MieObs,fileunit_stdout,prefix,PrintAll,&
                                  error_flag)

  end subroutine PrintMieObservation
    !  #]
  subroutine AsciiWriteMieObservation(MieObs,fileunit,prefix,PrintAll,&
                                      error_flag)
    !  #[

    TYPE(MieObservationType), intent(in)  :: MieObs    ! input
    integer,                  intent(in)  :: fileunit   ! input
    character(len=*),         intent(in)  :: prefix     ! input
    logical,                  intent(in)  :: PrintAll   ! input
    integer,                  intent(out) :: error_flag ! output

    ! local variables
    integer                        :: i,s,step
    character(len=maxlen_IntAs)    :: str_IntAs
    character(len=2)               :: str_cntr
    character(len=256)             :: obs_type_str
    
    error_flag = no_error

    obs_type_str = GetObsTypeText(int(MieObs%observation_type,IntAuc))
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%observation_type  = "//trim(obs_type_str)

    call SetIntAs_or_Missing(MieObs%nr_of_range_gates,str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%nr_of_range_gates = "//trim(str_IntAs)

    IF (associated(MieObs%RangeGate)) THEN
       s=MieObs%nr_of_range_gates
       IF (PrintAll .or. (s .le. 1)) THEN
          step=1
       ELSE
          step = s - 1
       END IF
       DO i=1,s,step
          write(str_cntr,"(i2.2)",err=998) i
          call AsciiWriteMieObsRangeGate(&
               MieObs%RangeGate(i), fileunit, &
               trim(prefix)//"%RangeGate("//trim(str_cntr)//")",error_flag)
          IF (error_flag .ne. no_error) goto 999
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%RangeGate(:) [seems not associated]"
    END IF

    return

998 print *,"conversion error in AsciiWriteMieObservation:"
    print *,"could not convert the counter ",i," to a string ..."
    error_flag = error_writing_file
    return

999 print *,"ERROR in AsciiWriteMieObservation:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWriteMieObservation
    !  #]
  !---------------------------------------------
  subroutine InitRaylObsRangeGate(RaylObsRangegate)
    !  #[
    TYPE(RayleighObsRangeGateType), intent(out) :: RaylObsRangegate

    RaylObsRangegate%alt_start = missing_indicator_FAdoxy
    RaylObsRangegate%alt_cog   = missing_indicator_FAdoxy
    RaylObsRangegate%alt_end   = missing_indicator_FAdoxy
    RaylObsRangegate%lat_start = missing_indicator_FAdoxy
    RaylObsRangegate%lat_cog   = missing_indicator_FAdoxy
    RaylObsRangegate%lat_stop  = missing_indicator_FAdoxy
    RaylObsRangegate%lon_start = missing_indicator_FAdoxy
    RaylObsRangegate%lon_cog   = missing_indicator_FAdoxy
    RaylObsRangegate%lon_stop  = missing_indicator_FAdoxy
    RaylObsRangegate%nr_of_measurements = 0
    RaylObsRangegate%T_ref     = missing_indicator_FAdoxy
    RaylObsRangegate%p_ref     = missing_indicator_FAdoxy
    RaylObsRangegate%hlos_cog  = missing_indicator_FAdoxy

  end subroutine InitRaylObsRangeGate
    !  #]
  subroutine PrintRaylObsRangeGate(RaylObsRangegate,prefix,error_flag)
    !  #[
    TYPE(RayleighObsRangeGateType), intent(in)  :: RaylObsRangegate ! input
    character(len=*),               intent(in)  :: prefix           ! input
    integer,                        intent(out) :: error_flag       ! output

    call AsciiWriteRaylObsRangeGate(RaylObsRangegate,fileunit_stdout,&
                                    prefix,error_flag)

  end subroutine PrintRaylObsRangeGate
    !  #]
  subroutine AsciiWriteRaylObsRangeGate(RaylObsRangegate,fileunit,prefix,&
                                        error_flag)
    !  #[
    TYPE(RayleighObsRangeGateType), intent(in)  :: RaylObsRangegate ! input
    integer,                        intent(in)  :: fileunit         ! input
    character(len=*),               intent(in)  :: prefix           ! input
    integer,                        intent(out) :: error_flag       ! output

    ! local variables
    character(len=maxlen_IntAs)       :: str_IntAs
    character(len=maxlen_FAdoxy_full) :: str_FAdoxy
    
    error_flag = no_error

    write(fileunit,"(a)",err=999) "REALACC(12)"

    call SetFAdoFull_or_Missing(RaylObsRangegate%alt_start,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%alt_start          = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RaylObsRangegate%alt_cog,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%alt_cog            = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RaylObsRangegate%alt_end,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%alt_end            = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RaylObsRangegate%lat_start,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%lat_start          = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RaylObsRangegate%lat_cog,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%lat_cog            = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RaylObsRangegate%lat_stop,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%lat_stop           = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RaylObsRangegate%lon_start,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%lon_start          = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RaylObsRangegate%lon_cog,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%lon_cog            = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RaylObsRangegate%lon_stop,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%lon_stop           = "//trim(str_FAdoxy)

    call SetIntAs_or_Missing(RaylObsRangegate%nr_of_measurements,&
                             str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%nr_of_measurements = "//trim(str_IntAs)

    call SetFAdoFull_or_Missing(RaylObsRangegate%T_ref,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%T_ref              = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RaylObsRangegate%p_ref,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%p_ref              = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(RaylObsRangegate%hlos_cog,str_FAdoxy,&
                                error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%hlos_cog           = "//trim(str_FAdoxy)

    write(fileunit,"(a)",err=999) "ENDREALACC"

    return

999 print *,"ERROR in AsciiWriteRaylObsRangeGate"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return
    
  end subroutine AsciiWriteRaylObsRangeGate
    !  #]
  !---------------------------------------------
  subroutine InitRaylObservation(RaylObs,nr_of_range_gates,error_flag)
    !  #[
    TYPE(RayleighObservationType), intent(out) :: RaylObs           ! output
    integer(IntAs),                intent(in)  :: nr_of_range_gates ! input
    integer,                       intent(out) :: error_flag        ! output

    ! local variable
    integer :: i, AllocStatus

    error_flag = no_error

    RaylObs%observation_type  = Obs_Type_undefined
    RaylObs%nr_of_range_gates = nr_of_range_gates

    allocate(RaylObs%RangeGate(nr_of_range_gates),STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation failure in subroutine InitRaylObservation"
       error_flag = error_allocate
       return
    END IF

    DO i=1,RaylObs%nr_of_range_gates
       call InitRaylObsRangeGate(RaylObs%RangeGate(i))
    END DO

  end subroutine InitRaylObservation
    !  #]
  subroutine DeleteRaylObservation(RaylObs)
    !  #[
    TYPE(RayleighObservationType), intent(inout) :: RaylObs

    IF (associated(RaylObs%RangeGate)) deallocate(RaylObs%RangeGate)

  end subroutine DeleteRaylObservation
    !  #]
  subroutine PrintRaylObservation(RaylObs,prefix,PrintAll,error_flag)
    !  #[
    TYPE(RayleighObservationType), intent(in)  :: RaylObs    ! input
    character(len=*),              intent(in)  :: prefix     ! input
    logical,                       intent(in)  :: PrintAll   ! input
    integer,                       intent(out) :: error_flag ! output

    call AsciiWriteRaylObservation(RaylObs,fileunit_stdout,prefix,PrintAll,&
                                  error_flag)

  end subroutine PrintRaylObservation
    !  #]
  subroutine AsciiWriteRaylObservation(RaylObs,fileunit,prefix,PrintAll,&
                                       error_flag)
    !  #[

    TYPE(RayleighObservationType), intent(in)  :: RaylObs    ! input
    integer,                       intent(in)  :: fileunit   ! input
    character(len=*),              intent(in)  :: prefix     ! input
    logical,                       intent(in)  :: PrintAll   ! input
    integer,                       intent(out) :: error_flag ! output

    ! local variables
    integer                        :: i,s,step
    character(len=maxlen_IntAs)    :: str_IntAs
    character(len=2)               :: str_cntr
    character(len=256)             :: obs_type_str
    
    error_flag = no_error

    obs_type_str = GetObsTypeText(int(RaylObs%observation_type,IntAuc))
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%observation_type  = "//trim(obs_type_str)

    call SetIntAs_or_Missing(RaylObs%nr_of_range_gates,str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%nr_of_range_gates = "//trim(str_IntAs)

    IF (associated(RaylObs%RangeGate)) THEN
       s=RaylObs%nr_of_range_gates
       IF (PrintAll .or. (s .le. 1)) THEN
          step=1
       ELSE
          step = s - 1
       END IF
       DO i=1,s,step
          write(str_cntr,"(i2.2)",err=998) i
          call AsciiWriteRaylObsRangeGate(&
               RaylObs%RangeGate(i), fileunit, &
               trim(prefix)//"%RangeGate("//trim(str_cntr)//")",error_flag)
          IF (error_flag .ne. no_error) goto 999
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%RangeGate(:) [seems not associated]"
    END IF

    return

998 print *,"conversion error in AsciiWriteRaylObservation:"
    print *,"could not convert the counter ",i," to a string ..."
    error_flag = error_writing_file
    return

999 print *,"ERROR in AsciiWriteRaylObservation:"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWriteRaylObservation
    !  #]
  !---------------------------------------------
  subroutine InitWorkingData(WorkingData, N_range_gates, N_Meas, &
                             N_Obs_Mie_max, N_Obs_Rayleigh_max, error_flag)
    !  #[
    TYPE(WorkingDataType), intent(out) :: WorkingData        ! output
    integer(IntAs),        intent(in)  :: N_range_gates      ! input
    integer(IntAs),        intent(in)  :: N_Meas             ! input
    integer(IntAs),        intent(in)  :: N_Obs_Mie_max      ! input
    integer(IntAs),        intent(in)  :: N_Obs_Rayleigh_max ! input
    integer,               intent(out) :: error_flag         ! output

    ! local variables
    integer :: m, AllocStatus

    error_flag = no_error

    WorkingData%nr_of_range_gates               = N_Range_Gates
    WorkingData%nr_of_measurements              = N_Meas
    WorkingData%nr_of_observations_mie_max      = N_Obs_Mie_max
    WorkingData%nr_of_observations_rayleigh_max = N_Obs_Rayleigh_max

    WorkingData%nr_of_observations_mie_actual   = 0
    WorkingData%nr_of_observations_rayl_actual  = 0

    IF (N_Meas .ne. 0) THEN
       allocate(WorkingData%MieMeas(     N_Meas), &
                WorkingData%RayleighMeas(N_Meas), STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine InitWorkingData"
          error_flag = error_allocate
          return
       END IF
    
       DO m=1,N_Meas
          call InitMieMeasurement(WorkingData%MieMeas(m),&
                                  N_range_gates,error_flag)
          IF (error_flag .ne. no_error) return
          call InitRayleighMeasurement(WorkingData%RayleighMeas(m),&
                                       N_range_gates, error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    ELSE
       nullify(WorkingData%MieMeas)
       nullify(WorkingData%RayleighMeas)
    END IF

    IF (N_Obs_Mie_max .ne. 0) THEN
       allocate(WorkingData%MieObs(     N_Obs_Mie_max), &
                WorkingData%RayleighObs(N_Obs_Mie_max), STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine InitWorkingData"
          error_flag = error_allocate
          return
       END IF
    
       DO m=1,N_Obs_Mie_max
          call InitMieObservation(WorkingData%MieObs(m),&
                                  N_range_gates,error_flag)
          IF (error_flag .ne. no_error) return
          call InitRaylObservation(WorkingData%RayleighObs(m),&
                                   N_range_gates, error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    ELSE
       nullify(WorkingData%MieObs)
       nullify(WorkingData%RayleighObs)
    END IF

    ! similar code for the RaylObs data

  END subroutine InitWorkingData
  !  #]
  subroutine DeleteWorkingData(WorkingData)
    !  #[

    TYPE(WorkingDataType), intent(out) :: WorkingData ! input/output

    ! local variable
    integer :: m

    WorkingData%nr_of_range_gates               = 0
    WorkingData%nr_of_measurements              = 0
    WorkingData%nr_of_observations_mie_max      = 0
    WorkingData%nr_of_observations_rayleigh_max = 0

    WorkingData%nr_of_observations_mie_actual   = 0
    WorkingData%nr_of_observations_rayl_actual  = 0

    IF ( associated(WorkingData%MieMeas)) THEN
       DO m=1,size(WorkingData%MieMeas)
          call DeleteMieMeasurement(WorkingData%MieMeas(m))
       END DO
       deallocate(WorkingData%MieMeas)
       nullify(WorkingData%MieMeas)
    END IF

    IF (associated(WorkingData%RayleighMeas)) THEN
       DO m=1,size(WorkingData%RayleighMeas)
          call DeleteRayleighMeasurement(WorkingData%RayleighMeas(m))
       END DO
       deallocate(WorkingData%RayleighMeas)
       nullify(WorkingData%RayleighMeas)
    END IF
    
    IF ( associated(WorkingData%MieObs)) THEN
       DO m=1,size(WorkingData%MieObs)
          call DeleteMieObservation(WorkingData%MieObs(m))
       END DO
       deallocate(WorkingData%MieObs)
       nullify(WorkingData%MieObs)
    END IF

    IF ( associated(WorkingData%RayleighObs)) THEN
       DO m=1,size(WorkingData%RayleighObs)
          call DeleteRaylObservation(WorkingData%RayleighObs(m))
       END DO
       deallocate(WorkingData%RayleighObs)
       nullify(WorkingData%RayleighObs)
    END IF

  END subroutine DeleteWorkingData
  !  #]
  subroutine PrintWorkingData(WorkingData,prefix,PrintAll,error_flag)
    !  #[
    TYPE(WorkingDataType), intent(in)  :: WorkingData ! input
    character(len=*),      intent(in)  :: prefix     ! input
    logical,               intent(in)  :: PrintAll   ! input
    integer,               intent(out) :: error_flag ! output

    call AsciiWriteWorkingData(WorkingData,fileunit_stdout,prefix,&
                               PrintAll,error_flag)

  end subroutine PrintWorkingData
    !  #]
  subroutine AsciiWriteWorkingData(WorkingData,fileunit,prefix,&
                                   PrintAll,error_flag)
    !  #[
    TYPE(WorkingDataType), intent(in)  :: WorkingData ! input
    integer,               intent(in)  :: fileunit    ! input
    character(len=*),      intent(in)  :: prefix      ! input
    logical,               intent(in)  :: PrintAll    ! input
    integer,               intent(out) :: error_flag  ! output

    ! local variables
    integer                        :: i,s,step
    character(len=maxlen_IntAs)    :: str_IntAs
    character(len=2)               :: str_cntr
    
    error_flag = no_error

    call SetIntAs_or_Missing(WorkingData%nr_of_range_gates,str_IntAs,&
                             error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%nr_of_range_gates               = "//trim(str_IntAs)

    call SetIntAs_or_Missing(WorkingData%nr_of_measurements,str_IntAs,&
                             error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%nr_of_measurements              = "//trim(str_IntAs)

    call SetIntAs_or_Missing(WorkingData%nr_of_observations_mie_max,str_IntAs,&
                             error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%nr_of_observations_mie_max      = "//trim(str_IntAs)

    call SetIntAs_or_Missing(WorkingData%nr_of_observations_mie_actual,&
                             str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%nr_of_observations_mie_actual   = "//trim(str_IntAs)

    call SetIntAs_or_Missing(WorkingData%nr_of_observations_rayleigh_max,&
                             str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%nr_of_observations_rayleigh_max = "//trim(str_IntAs)

    call SetIntAs_or_Missing(WorkingData%nr_of_observations_rayl_actual,&
                             str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%nr_of_observations_rayl_actual  = "//trim(str_IntAs)

    IF (associated(WorkingData%MieMeas)) THEN
       s=size(WorkingData%MieMeas)
       !s=WorkingData%nr_of_measurements
       IF (PrintAll .or. (s .le. 1)) THEN
          step=1
       ELSE
          step = s - 1
       END IF
       DO i=1,s,step
          write(str_cntr,"(i2.2)",err=998) i
          call AsciiWriteMieMeasurement(&
               WorkingData%MieMeas(i), fileunit, &
               trim(prefix)//"%MieMeas("//trim(str_cntr)//")",&
               PrintAll,error_flag)
          IF (error_flag .ne. no_error) goto 999
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%MieMeas(:) [seems not associated]"
    END IF

    IF (associated(WorkingData%RayleighMeas)) THEN
       s=size(WorkingData%RayleighMeas)
       !s=WorkingData%nr_of_measurements
       IF (PrintAll .or. (s .le. 1)) THEN
          step=1
       ELSE
          step = s - 1
       END IF
       DO i=1,s,step
          write(str_cntr,"(i2.2)",err=998) i
          call AsciiWriteRayleighMeasurement(&
               WorkingData%RayleighMeas(i), fileunit, &
               trim(prefix)//"%RayleighMeas("//trim(str_cntr)//")",&
               PrintAll,error_flag)
          IF (error_flag .ne. no_error) goto 999
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%RayleighMeas(:) [seems not associated]"
    END IF

    IF (associated(WorkingData%MieObs)) THEN
       s=size(WorkingData%MieObs)
       !s=WorkingData%nr_of_observations_mie_max
       !s=WorkingData%nr_of_observations_mie_actual
       IF (PrintAll .or. (s .le. 1)) THEN
          step=1
       ELSE
          step = s - 1
       END IF
       DO i=1,s,step
          write(str_cntr,"(i2.2)",err=998) i
          call AsciiWriteMieObservation(&
               WorkingData%MieObs(i), fileunit, &
               trim(prefix)//"%MieObs("//trim(str_cntr)//")",&
               PrintAll,error_flag)
          IF (error_flag .ne. no_error) goto 999
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%MieObs(:) [seems not associated]"
    END IF

    IF (associated(WorkingData%RayleighObs)) THEN
       s=size(WorkingData%RayleighObs)
       !s=WorkingData%nr_of_observations_rayleigh_max
       !s=WorkingData%nr_of_observations_rayl_actual
       IF (PrintAll .or. (s .le. 1)) THEN
          step=1
       ELSE
          step = s - 1
       END IF
       DO i=1,s,step
          write(str_cntr,"(i2.2)",err=998) i
          call AsciiWriteRaylObservation(&
               WorkingData%RayleighObs(i), fileunit, &
               trim(prefix)//"%RayleighObs("//trim(str_cntr)//")",&
               PrintAll,error_flag)
          IF (error_flag .ne. no_error) goto 999
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%RayleighObs(:) [seems not associated]"
    END IF

    return

998 print *,"conversion error in AsciiWriteWorkingData:"
    print *,"could not convert the counter ",i," to a string .."
    error_flag = error_writing_file
    return

999 print *,"ERROR in AsciiWriteWorkingData"
    print *,"error while writing to outputfile with fileunit: ",fileunit
    error_flag = error_writing_file
    return

  end subroutine AsciiWriteWorkingData
    !  #]
  !---------------------------------------------
  subroutine Copy_L1B_fields2WorkingArrays(L1B,WorkingData)
    !  #[

    TYPE(L1B_BRC_DataType),    intent(inout) :: L1B         ! input
    ! (for now als output, but this is to be moved into the workingdata)
    TYPE(WorkingDataType), intent(inout) :: WorkingData ! input, output

    ! local variables
    TYPE(Rangebin_Meas_Geoloc_type), pointer :: MieGeolocation
    TYPE(Rangebin_Meas_Geoloc_type), pointer :: RayleighGeolocation

    TYPE(Rangebin_Meas_Geoloc_type), pointer :: MieGeolocation_next
    TYPE(Rangebin_Meas_Geoloc_type), pointer :: RayleighGeolocation_next

    TYPE(Rangebin_Obs_Geoloc_type),  pointer :: MieObsGeolocation
    TYPE(Rangebin_Obs_Geoloc_type),  pointer :: RayleighObsGeolocation

    TYPE(MieRangeGateType),          pointer :: WD_MieRangeGate 
    TYPE(RayleighRangeGateType),     pointer :: WD_RayleighRangeGate 

    integer      :: m, r, r_next
    integer(i1_) :: flags
    !character(len=maxlen_DateTime_KVT)  :: txt_DateTime_KVT

    ! this is no longer a parameter, but the actual value, 
    ! read from the L1B file
    integer      :: N_meas

! Poli: comment out:
!    N_meas = L1B%SPH%N_Meas_Max
! and replace by:
    N_meas = size(L1B%Geoloc_ADS%Meas_AOCS)

    print *,"REALACC(6)"
    meas_loop: DO m=1,N_meas

       WorkingData%RayleighMeas(m)%datetime = & ! [UTC]
            L1B%Geoloc_ADS%Meas_AOCS(m)%obs_centroid_datetime
       ! remark: this name is misleading since Meas_AOCS and
       !         Obs_AOCS both use the same type. In the L1B IODD
       !         this element is named meas_centroid_datetime

       WorkingData%MieMeas(     m)%datetime = & ! [UTC]
            L1B%Geoloc_ADS%Meas_AOCS(m)%obs_centroid_datetime
       ! remark: this name is misleading since Meas_AOCS and
       !         Obs_AOCS both use the same type. In the L1B IODD
       !         this element is named meas_centroid_datetime

       !txt_DateTime_KVT = &
       !     Construct_KVT_DateTime(WorkingData%RayleighMeas(m)%datetime)
       !print *,"RayleighMeas(",m,")%datetime: ",trim(txt_DateTime_KVT)," [UTC]"
       !txt_DateTime_KVT = &
       !     Construct_KVT_DateTime(WorkingData%MieMeas(     m)%datetime)
       !print *,"MieMeas(     ",m,")%datetime: ",trim(txt_DateTime_KVT)," [UTC]"

       ! remember the 1st element of Rayleigh_Rangebin_Geoloc(:) contains
       ! the upper edge of the highest rangebin, while elements 2:25
       ! contain the lower edge of rangebins 1:24
       print *,"   Rayleigh  Rayleigh  Rayleigh       "//&
            "Mie       Mie       Mie"
       print *,"r  alt_start alt_end   obs_signal_A+B "//&
            "alt_start alt_end   obs_signal"
       rangegate_loop: DO r=1,N_range_gates
          !  #[
          ! set some pointers for shorter notation to the L1B data
          RayleighGeolocation => &
               L1B%Geoloc_ADS%Meas_Geoloc(m)%Rayleigh_Rangebin_Geoloc(r+1)
          MieGeolocation => &
               L1B%Geoloc_ADS%Meas_Geoloc(m)%Mie_Rangebin_Geoloc(r+1)

          r_next = r
          RayleighGeolocation_next => &
               L1B%Geoloc_ADS%Meas_Geoloc(m)%Rayleigh_Rangebin_Geoloc(r_next)
          MieGeolocation_next => &
               L1B%Geoloc_ADS%Meas_Geoloc(m)%Mie_Rangebin_Geoloc(r_next)

          MieObsGeolocation => &
               L1B%Geoloc_ADS%Obs_Geoloc%Mie_Rangebin_Geoloc(r)
          RayleighObsGeolocation =>  &
               L1B%Geoloc_ADS%Obs_Geoloc%Rayleigh_Rangebin_Geoloc(r)

          ! set some pointers for shorter notation to the working data structs
          WD_RayleighRangeGate => WorkingData%RayleighMeas(m)%RangeGate(r)
          WD_MieRangeGate      => WorkingData%MieMeas(     m)%RangeGate(r)

          !=============================================================
          ! copy the needed Rayleigh data
          !=============================================================

          ! to be replaced by:

          WD_RayleighRangeGate%alt_start = & ! [deg]
               RayleighGeolocation%alt_of_height_bin
          WD_RayleighRangeGate%alt_end = &   ! [deg]
               RayleighGeolocation_next%alt_of_height_bin

          ! copy lat,lon taking the scaling 1.e-6 into account
          WD_RayleighRangeGate%lat = &       ! [deg]
               RayleighGeolocation%lat_of_height_bin * 1.e-6_r8_
          WD_RayleighRangeGate%lon = &       ! [deg]
               RayleighGeolocation%lon_of_height_bin * 1.e-6_r8_

          WD_RayleighRangeGate%SatRange = &  ! [m]
               RayleighObsGeolocation%satellite_range_of_height_bin
          WD_RayleighRangeGate%IncAngle = &  ! [deg]
               RayleighObsGeolocation%topoc_elev_of_height_bin
          WD_RayleighRangeGate%AziAngle = &  ! [deg]
               RayleighObsGeolocation%topoc_azim_of_height_bin

          ! check the validity of this data

          ! bit 1: overall validity. Data invalid 1, otherwise 0
          ! bit 2: Bin-level validity spare. Default 0
          ! bit 3: Data saturation found 1, otherwise 0 
          ! bit 4: Data spike found 1, otherwise 0
          ! bit 5: Measurement-level validity spare. Default 0
          ! bit 6: Source packet invalid 1, otherwise 0
          ! bit 7: Laser frequency not locked 1, otherwise 0
          ! bit 8: Spacecraft attitude not on target 1, otherwise 0
          flags = int(L1B%US_MDS%Meas_US(m)%Rayleigh(r)%Data_Quality_Flag,i1_)
          IF (btest(flags,1 - 1)) THEN
             print *,"Rayleigh data flagged as invalid !!!"
             print *,"for measurement ",m," rangebin ",r
             print *,"TEST: flags = ",flags
             IF (btest(flags,1 - 1)) &
                  print *,"bit 1 set: overall validity"
             IF (btest(flags,2 - 1)) &
                  print *,"bit 2 set: Bin-level validity spare"
             IF (btest(flags,3 - 1)) &
                  print *,"bit 3 set: Data saturation found"
             IF (btest(flags,4 - 1)) &
                  print *,"bit 4 set: Data spike found"
             IF (btest(flags,5 - 1)) &
                  print *,"bit 5 set: Measurement-level validity spare"
             IF (btest(flags,6 - 1)) &
                  print *,"bit 6 set: Source packet invalid"
             IF (btest(flags,7 - 1)) &
                  print *,"bit 7 set: Laser frequency not locked"
             IF (btest(flags,8 - 1)) &
                  print *,"bit 8 set: Spacecraft attitude not on target"
             WD_RayleighRangeGate%obs_signal_level = missing_indicator_real
          ELSE
             WD_RayleighRangeGate%obs_signal_level = &
                  L1B%US_MDS%Meas_US(m)%Rayleigh(r)%Useful_Signal_Channel_A + &
                  L1B%US_MDS%Meas_US(m)%Rayleigh(r)%Useful_Signal_Channel_B
             ! unit is said to be "PixelLevel", what does this mean?
             ! is this a number of detected photons?
             IF (WD_RayleighRangeGate%obs_signal_level .lt. 0.) THEN
                print *,"WARNING: Rayleigh signal below 0 found !!"
                print *,"assuming this is missing data..."
                WD_RayleighRangeGate%obs_signal_level = missing_indicator_real
             END IF
          END IF

          ! this element is not present in the L1B data, and must be 
          ! calculated by the L2B processor:
          !WD_RayleighRangeGate%temp_mol_backscatter = data(3)

          !=============================================================
          ! copy the needed Mie data
          !=============================================================

          WD_MieRangeGate%alt_start = & ! [m]
               MieGeolocation%alt_of_height_bin
          WD_MieRangeGate%alt_end   = & ! [m]
               MieGeolocation_next%alt_of_height_bin

          ! copy lat,lon taking the scaling 1.e-6 into account
          WD_MieRangeGate%lat      = & ! [deg]
               MieGeolocation%lat_of_height_bin * 1.e-6_r8_
          WD_MieRangeGate%lon      = & ! [deg]
               MieGeolocation%lon_of_height_bin * 1.e-6_r8_

          WD_MieRangeGate%SatRange = & ! [m]
               MieObsGeolocation%satellite_range_of_height_bin
          WD_MieRangeGate%IncAngle = & ! [deg]
               MieObsGeolocation%topoc_elev_of_height_bin
          WD_MieRangeGate%AziAngle = & ! [deg]
               MieObsGeolocation%topoc_azim_of_height_bin

!          IF (r.eq.1) THEN 
!             print *,"testjos: numbers for rangebin nr. ",r
!             print *,"testjos: MieRangeGate%SatRange = ",MieRangeGate%SatRange
!             print *,"testjos: MieRangeGate%IncAngle = ",MieRangeGate%IncAngle
!             print *,"testjos: MieRangeGate%AziAngle = ",MieRangeGate%AziAngle
!          END IF

          ! check the validity of this data

          ! bit 1: overall validity. Data invalid 1, otherwise 0
          ! bit 2: Bin-level validity spare. Default 0
          ! bit 3: Data saturation found 1, otherwise 0 
          ! bit 4: Data spike found 1, otherwise 0
          ! bit 5: Measurement-level validity spare. Default 0
          ! bit 6: Source packet invalid 1, otherwise 0
          ! bit 7: Laser frequency not locked 1, otherwise 0
          ! bit 8: Spacecraft attitude not on target 1, otherwise 0
          flags = int(L1B%US_MDS%Meas_US(m)%Mie(r)%Data_Quality_Flag,i1_)
          IF (btest(flags,1 - 1)) THEN
             print *,"Mie data flagged as invalid !!!"
             print *,"for measurement ",m," rangebin ",r
             IF (btest(flags,1 - 1)) &
                  print *,"bit 1 set: overall validity"
             IF (btest(flags,2 - 1)) &
                  print *,"bit 2 set: Bin-level validity spare"
             IF (btest(flags,3 - 1)) &
                  print *,"bit 3 set: Data saturation found"
             IF (btest(flags,4 - 1)) &
                  print *,"bit 4 set: Data spike found"
             IF (btest(flags,5 - 1)) &
                  print *,"bit 5 set: Measurement-level validity spare"
             IF (btest(flags,6 - 1)) &
                  print *,"bit 6 set: Source packet invalid"
             IF (btest(flags,7 - 1)) &
                  print *,"bit 7 set: Laser frequency not locked"
             IF (btest(flags,8 - 1)) &
                  print *,"bit 8 set: Spacecraft attitude not on target"
             WD_MieRangeGate%obs_signal_level = missing_indicator_real
          ELSE
             WD_MieRangeGate%obs_signal_level = &
                  L1B%US_MDS%Meas_US(m)%Mie(r)%Useful_Signal
             ! unit is said to be "PixelLevel", what does this mean?
             ! is this a number of detected photons?
             IF (WD_MieRangeGate%obs_signal_level .lt. 0.) THEN
                print *,"WARNING: Mie signal below 0 found !!"
                print *,"assuming this is missing data..."
                WD_MieRangeGate%obs_signal_level = missing_indicator_real
             END IF
          END IF

          print "(i2,2(2(1x,f9.2),1x,es13.6e2,1x))",r,&
               WD_RayleighRangeGate%alt_start,WD_RayleighRangeGate%alt_end,&
               WD_RayleighRangeGate%obs_signal_level,&
               WD_MieRangeGate%alt_start,WD_MieRangeGate%alt_end,&
               WD_MieRangeGate%obs_signal_level

          !  #]
       END DO rangegate_loop

    END DO meas_loop
    print *,"ENDREALACC"

  end subroutine Copy_L1B_fields2WorkingArrays

  !  #]
  !---------------------------------------------
end module Working_DataStructure
