/*
  a test program to write several integer types to a binary datafile
  to be read by the fortran code.

  inspect the result using: hexdump -C SomeNumbers.bin

  written by:  08-Nov-2005, J. de Kloe, KNMI

  Modifications:
    J. de Kloe, 20-Apr-2006: added a 0 return value to suppress warnings
                             in the test suite
    J. de Kloe, 17-Apr-2007: made compatible with 64 bit systems
*/

#include <stdio.h>
#include <limits.h>
#include <float.h>
int main()
{
  /* see also the file /usr/include/limits.h
     for definitions of these maximum ranges */

  unsigned           char u1[5] = {0,1,2,254,255};
  signed             char s1[5] = {-128,-1,0,1,127};
  unsigned short     int  u2[5] = {0U,1U,2U,65534U,65535U};
  signed   short     int  s2[5] = {-32768,-1,0,1,32767};
  unsigned           int  ui[5] = {0,1,2,4294967294LL,4294967295LL};
  signed             int  si[5] = {-2147483648LL,-1,0,1,2147483647LL};
  unsigned long      int  u4[5] = {0,1,2,4294967294LL,4294967295LL};
  signed   long      int  s4[5] = {-2147483648LL,-1,0,1,2147483647LL};
  unsigned long long int  u8[5] = {0,1,2,3,18446744073709551615ULL}; 
  signed   long long int  s8[5] = {-9223372036854775807LL,-1,0,1,
				   +9223372036854775807LL        };
  float                   r4[5] = {-3.4e+38, -1.2e-38, 0., 1.2e-38,  3.4e+38 };
  double                  r8[5] = {-1.7e+308,-2.5e-308,0., 2.5e-308, 1.7e+308};

  /* REMARK:  
  if I use the lower max. range -9223372036854775808LL for s8 then
  gcc reports the following warnings:
      warning: integer constant is so large that it is unsigned
      warning: this decimal constant is unsigned only in ISO C90
  */
  
  static char filename[22] = "SomeNumbers.bin";
  static char flags[4]     = "w";
  static FILE *fileunit    = NULL;

  /* 1 byte unsigned integer case */
  if (sizeof(u1[0]) == 1)
    {
      printf("1 byte unsigned integer test case:\n");
      printf("max. value = %u\n",UCHAR_MAX);
    }
  else
    {
      printf("ERROR in WriteSomeNumbers:");
      printf("unsigned char does not have length of 1 byte\n");
    }

  /* 1 byte signed integer case */
  if (sizeof(s1[0]) == 1)
    {
      printf("1 byte signed integer test case:\n");
      printf("min. value = %i\n",SCHAR_MIN);
      printf("max. value = %i\n",SCHAR_MAX);
    }
  else
    {
      printf("ERROR in WriteSomeNumbers:");
      printf("signed char does not have length of 1 byte\n");
    }

  /* 2 byte unsigned integer case */
  if (sizeof(u2[0]) == 2)
    {
      printf("2 byte unsigned integer test case:\n");
      printf("max. value = %u\n",USHRT_MAX);
    }
  else
    {
      printf("ERROR in WriteSomeNumbers:");
      printf("unsigned short integer does not have length of 2 bytes\n");
    }

  /* 2 byte signed integer case */
  if (sizeof(s2[0]) == 2)
    {
      printf("2 byte signed integer test case:\n");
      printf("min_value = %i\n",SHRT_MIN);
      printf("max_value = %i\n",SHRT_MAX);
    }
  else
    {
      printf("ERROR in WriteSomeNumbers:");
      printf("signed char does not have length of 1 byte\n");
    }

  /* the next case requires some extra attention. since on 32 bit platforms
     a long integer usually is 4 bytes long, while on 64 bit platforms 
     it is usually 8 bytes long. The default integer is still 4 bytes in 
     that case, so use that one in stead. */

  /* 4 byte unsigned integer case */
  if (sizeof(u4[0]) == 4)
    {
      printf("4 byte unsigned integer test case:\n");
      printf("max. value = %u\n",(unsigned int) ULONG_MAX);
    }
  else
    {
      if (sizeof(ui[0]) == 4)
	{
	  printf("4 byte unsigned integer test case:\n");
	  printf("max. value = %u\n",UINT_MAX);
	}
      else
	{
	  printf("ERROR in WriteSomeNumbers:");
	  printf("unsigned long integer and unsigned integer do not ");
	  printf("have length of 4 bytes\n");
	}
    }

  /* 4 byte signed integer case */
  if (sizeof(s4[0]) == 4)
    {
      printf("4 byte signed integer test case:\n");
      printf("min. value = %i\n",(int) LONG_MIN);
      printf("max. value = %i\n",(int) LONG_MAX);
    }
  else
    {
      if (sizeof(si[0]) == 4)
	{
	  printf("4 byte signed integer test case:\n");
	  printf("min. value = %i\n",INT_MIN);
	  printf("max. value = %i\n",INT_MAX);
	}
      else
	{
	  printf("ERROR in WriteSomeNumbers:");
	  printf("signed long integer and signed integer do not ");
	  printf("have length of 4 bytes\n");
	}
    }

  /* 8 byte unsigned integer case */
  if (sizeof(u8[0]) == 8)
    {
      printf("8 byte unsigned integer test case:\n");
      printf("max. value = %llu\n",u8[4]);
    }
  else
    {
      printf("ERROR in WriteSomeNumbers:");
      printf("unsigned long long integer does not have length of 8 bytes\n");
    }

  /* 8 byte signed integer case */
  if (sizeof(s8[0]) == 8)
    {
      printf("8 byte signed integer test case:\n");
      printf("min_value = %lli\n",s8[0]);
      printf("max_value = %lli\n",s8[4]);
    }
  else
    {
      printf("ERROR in WriteSomeNumbers:");
      printf("signed char does not have length of 1 byte\n");
    }

  /* 4 byte real case */
  if (sizeof(r4[0]) == 4)
    {
      printf("4 byte real test case:\n");
      printf("min. value = %e\n",FLT_MIN);
      printf("max. value = %e\n",FLT_MAX);
    }
  else
    {
      printf("ERROR in WriteSomeNumbers:");
      printf("float does not have length of 4 bytes\n");
    }

  /* 8 byte real case */
  if (sizeof(r8[0]) == 8)
    {
      printf("8 byte real test case:\n");
      printf("min. value = %e\n",DBL_MIN);
      printf("max. value = %e\n",DBL_MAX);
    }
  else
    {
      printf("ERROR in WriteSomeNumbers:");
      printf("double does not have length of 8 bytes\n");
    }
  
  printf("Writing binary file: %s\n",filename);
  fileunit = fopen(filename,flags);

  /* write integer tests to the binary file */
  fwrite(u1,1,5,fileunit);
  fwrite(s1,1,5,fileunit);
  fwrite(u2,2,5,fileunit);
  fwrite(s2,2,5,fileunit);

  if (sizeof(u4[0]) == 4)
    { fwrite(u4,4,5,fileunit); }
  else
    { fwrite(ui,4,5,fileunit); }

  if (sizeof(s4[0]) == 4)
    { fwrite(s4,4,5,fileunit); }
  else
    { fwrite(si,4,5,fileunit); }

  /* unsigned 8byte integers cannot be represented in fortran
     so testing them is useless
  fwrite(u8,8,5,fileunit);  */
  fwrite(s8,8,5,fileunit);

  /* write float/real tests to the binary file */
  fwrite(&r4,4,5,fileunit);
  fwrite(&r8,8,5,fileunit);
  
  fclose(fileunit);
  printf("done\n");

  return 0;
}
