module Buffering
  !  #[ Description:
  !
  !     A module to
  !     -a module to provide buffering in memory to hopefully
  !      increase the speed of file access by the DirectAccess module
  !     -remark: byteswapping is not possible at this level, since 
  !              then you need to be sure that the file only contains
  !              aligned 4 and 8 byte variables. Since there are also 2 byte
  !              ee_cfi variables that need to be byteswapped, and since
  !              we have no guarantee that the 4/8 byte variables are 
  !              correctly aligned on a multiple of 4 bytes in the file,
  !              this is not possible....
  !
  ! created: 03-02-2006 by J. de Kloe
  !
  ! Modifications:
  !   03-Feb-2006 J. de Kloe initial version, not yet functional
  !   09-Feb-2006 J. de Kloe developed further, functioning correct now
  !   10-Feb-2006 J. de Kloe added shrinking of the buffersize at the end
  !                          of the file.
  !   03-Mar-2006 J. de Kloe rewritten the handling of the remainder of the 
  !                          file
  !   04-Mar-2006 J. de Kloe added support for Buffered writing
  !   12-May-2006 D. Tan     Null use of ParKind1 module
  !   27-Jul-2006 J. de Kloe WriteNextBytesBuffered()
  !   31-ul-2006  J. de Kloe added a routine AdvanceFilePointer
  !   16-Jan-2008 J. de Kloe phase out integer kind i_ 
  !
  !  #]
  !  #[ Modules used: 
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE Numerics, only: i8_, i4_, r8_
  USE LunManager, only: get_lun, free_lun
  USE c_support, only: get_filesize
  USE ErrorHandler, only: no_error, error_programming, &
                          error_reading_file, error_writing_file, &
                          error_opening_file, error_allocate
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  !  #]
  !  #[ Parameter/Variable/Type declarations

  implicit none

  TYPE BufferedFile_type
     integer        :: fileunit
     ! this field is not directly used inside this module, but since it
     ! is very usefull to remember the location of the last read/write
     ! the field and some interfacing routines are available
     integer            :: nextbyte
     integer            :: filesize
     integer            :: action ! read or write?
     logical            :: file_is_open
     character(len=256) :: filename
     integer            :: BufferStartPos 
     integer            :: BufferEndPos 
     ! pos. in file that matches with byte 1 of the buffer
     integer            :: LastDataByte
     ! last byte in buffer that contains data from the file
     ! (needed in case the buffer is at the end of the file, and
     !  not completely filled)
     logical            :: BufferFilled
     ! the size of the buffer to be used (in bytes)
     integer            :: BufferSize
     ! counters to check performance
     integer            :: count_reads
     integer            :: count_writes
     ! the actual buffer holding copied data from the file
     character(len=1), dimension(:), pointer :: buffer
  END TYPE BufferedFile_type

  integer, private, parameter :: undefined_action = -1
  integer, private, parameter :: open_for_reading = 1
  integer, private, parameter :: open_for_writing = 2

  ! the default size of the buffer to be used (in bytes)
  integer, parameter :: DefaultBuffersize = 8192

  ! the smallest size of the buffer, before switching to byte-by-byte reading
  integer, parameter :: SmallestBuffersize = 16

  ! the default verbosity setting is Error (as silent as possible)
  integer, private, save :: LocalVerbosity = VerbosityError
  !  #]
  !  #[ Module interface
  !  keep the following routines private to this module
  private :: ReadDataIntoBuffer, &
             InitBufferedFile, &
             PrintBufferedFileProperties, &
             ReadFileRemainderIntoBuffer, &
             Init_New_Write_Buffer
  !  #]
contains
  !------------------------------------------
  subroutine InitBuffering(ChoosenVerbosity)
    !  #[
    integer, optional, intent(in)  :: ChoosenVerbosity

    IF (present(ChoosenVerbosity)) LocalVerbosity = ChoosenVerbosity
    
    IF (LocalVerbosity .ge. VerbosityReport) &
         call print_verbosity_setting(LocalVerbosity,"Buffering")

  end subroutine InitBuffering
    !  #]
  subroutine OpenBufferedFile(BufferedFile,filename,action,error_flag, &
                              BufferSize)
    !  #[

    TYPE(BufferedFile_type), pointer     :: BufferedFile ! result
    character(len=*),        intent(in)  :: filename
    character(len=*),        intent(in)  :: action   ! can be: read/write
    integer,                 intent(out) :: error_flag
    integer,       optional, intent(in)  :: BufferSize

    ! local variables
    integer :: fileunit, filesize, actiontype, AllocStatus, power

    error_flag = no_error
    actiontype = undefined_action
    filesize   = -1

    IF (.not. associated(BufferedFile)) THEN
       print *,"ERROR in OpenBufferedFile:"
       print *,"the pointer BufferedFile seems not associated."
       print *,"It should be pointing to some allocated memory"
       print *,"before trying to open a file with it !!!"
       error_flag = error_programming
       return
    END IF

    ! initialise the BufferedFile struct
    call InitBufferedFile(BufferedFile, error_flag)

    ! handle the optional inputs
    BufferedFile%BufferSize = DefaultBuffersize
    IF (present(BufferSize)) BufferedFile%BufferSize = BufferSize

    ! request a free fileunit number
    fileunit = get_lun()    
    IF (fileunit .lt. 0) THEN
       print *,"ERROR in OpenBufferedFile:"
       print *,"no more free fileunits available."
       print *,"call of get_lun() function failed ...."
       error_flag = error_opening_file
       return
    END IF

    ! handle the action type
    select case(trim(action))
    case("read","READ")
       filesize = get_filesize(filename)
       open(unit=fileunit,file=filename,status="old",&
            action="read",access="direct",&
            recl=BufferedFile%BufferSize,form="UNFORMATTED",ERR=998)
       actiontype = open_for_reading
    case("write","WRITE")
       open(unit=fileunit,file=filename,status="replace",&
            action="write",access="direct",&
            recl=BufferedFile%BufferSize,form="UNFORMATTED",ERR=999)
       actiontype = open_for_writing
       filesize = 0
    case default
       print *,"ERROR in OpenBufferedFile:"
       print *,"invalid action type: ",trim(action)
       print *,"allowed values are: read, or write"
       error_flag = error_programming
       call free_lun(fileunit)
       return
    end select

    IF (BufferedFile%BufferSize .lt. filesize) THEN
       ! check if the buffersize is a power of 2, and issue a warning if it 
       ! isn't. The +0.5 is needed to prevent the test to fail if you end up 
       ! just below the integer, and floor rounds it to the integer below...
       power=floor(log(1._r8_*BufferedFile%BufferSize+0.5_r8_)/log(2._r8_))

       IF (BufferedFile%BufferSize .ne. 2**power) THEN
          print *,"ERROR in OpenBufferedFile():"
          print *,"file buffering works most efficient if you use a buffersize"
          print *,"that is a power of 2 !!! (or if the buffersize matches"
          print *,"the filesize)"
          print *,"However, you supplied a value of ",BufferedFile%BufferSize
          print *,"this may harm the performance of the buffering module ...."
          print *,"the two closest powers of 2 are:"
          print *,2**power," and ",2**(power+1)
          print *,"filesize = ",filesize
          print *,"please use one of these 3 numbers as buffer size"
          error_flag = error_programming
          return
       END IF
    END IF

    ! request a block of memory for the buffer
    allocate(BufferedFile%buffer(BufferedFile%BufferSize),stat=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation problem in OpenBufferedFile()"
       error_flag = error_allocate
       return
    END IF

    ! fill the BufferedFile struct only if the file was opened succesfully
    BufferedFile%fileunit       = fileunit
    BufferedFile%filesize       = filesize ! unly usefull when reading?
    BufferedFile%action         = actiontype
    BufferedFile%file_is_open   = .true.
    BufferedFile%filename       = filename
    BufferedFile%nextbyte       = 1

    ! don't yet fill the buffer with data, because we don't know
    ! yet where the user wants to start reading or writing the file, 
    ! and for large files the first BufferSize bytes might not be
    ! needed.

    ! for debugging
    !call PrintBufferedFileProperties(BufferedFile)
    !stop

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"Successfuly opened buffered file: ",trim(filename)
       print *,"with the following properties:"
       call PrintBufferedFileProperties(BufferedFile)
    END IF
    
    return

998 print *,"ERROR could not open inputfile: ",trim(filename)
    error_flag = error_opening_file
    return

999 print *,"ERROR could not open outputfile: ",trim(filename)
    error_flag = error_opening_file
    return

  end subroutine OpenBufferedFile
  !  #]
  subroutine CloseBufferedFile(BufferedFile, error_flag)
    !  #[

    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer,                 intent(out) :: error_flag

    ! local variable
    character(len=256) :: filename

    error_flag = no_error
    IF (.not. associated(BufferedFile)) THEN
       print *,"ERROR in CloseBufferedFile:"
       print *,"the BufferedFile pointer seems not associated"
       error_flag = error_programming
       return
    END IF

    ! if the buffer was used for writing, flush the changed buffer
    IF (BufferedFile%action .eq. open_for_writing) THEN
       call Flush_Buffer(BufferedFile,error_flag,remainder=.true.)
    END IF
    
    ! copy the name to be able to print it later
    filename = BufferedFile%filename

    ! close the file
    close(unit=BufferedFile%fileunit)

    ! free the fileunit number
    IF (BufferedFile%fileunit .ge. 0) call free_lun(BufferedFile%fileunit)

    ! free the allocated memory used by the buffer
    IF (associated(BufferedFile%buffer)) deallocate(BufferedFile%buffer)

    ! re-initialise the BufferedFile struct
    !call InitBufferedFile(BufferedFile, error_flag)
    ! dont do this for know, to be able to report the
    ! nrs of reads and counts

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"Successfuly closed buffered file: ",trim(filename)
    END IF

  end subroutine CloseBufferedFile

    !  #]
  subroutine SetFilePointer(BufferedFile,position,error_flag)
    !  #[
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer(i8_),            intent(in)  :: position
    integer,                 intent(out) :: error_flag
    
    ! note that position is an 8 byte integer
    ! because the input for this routine is often constructed
    ! from the DSD%Start_of_ds fields, which is also an 8 byte
    ! integer. We don't need large numbers for our purpose
    ! (i.e. don't expect to have 4GB or larger files any time
    ! in the future), so directly convert back to 4 a byte integer

    error_flag = no_error
    IF (.not. associated(BufferedFile)) THEN
       print *,"ERROR in ReadBytesBuffered:"
       print *,"the BufferedFile pointer seems not associated"
       error_flag = error_programming
       return
    END IF

    BufferedFile%nextbyte = int(position,i4_)

  end subroutine SetFilePointer
    !  #]
  subroutine GetFilePointer(BufferedFile,position,error_flag)
    !  #[
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer(i8_),            intent(out) :: position
    integer,                 intent(out) :: error_flag
    
    error_flag = no_error
    position = -1
    IF (.not. associated(BufferedFile)) THEN
       print *,"ERROR in ReadBytesBuffered:"
       print *,"the BufferedFile pointer seems not associated"
       error_flag = error_programming
       return
    END IF

    position = BufferedFile%nextbyte

  end subroutine GetFilePointer
    !  #]
  subroutine AdvanceFilePointer(BufferedFile,amount,error_flag)
    !  #[
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer(i8_),            intent(in)  :: amount
    integer,                 intent(out) :: error_flag
    
    error_flag = no_error
    IF (.not. associated(BufferedFile)) THEN
       print *,"ERROR in ReadBytesBuffered:"
       print *,"the BufferedFile pointer seems not associated"
       error_flag = error_programming
       return
    END IF

    BufferedFile%nextbyte = BufferedFile%nextbyte + int(amount,i4_)

  end subroutine AdvanceFilePointer
    !  #]
  subroutine ReadBytesBuffered(BufferedFile,startpos,endpos,data,error_flag)
    !  #[
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer,                 intent(in)  :: startpos
    integer,                 intent(in)  :: endpos
    character(len=1), dimension(endpos-startpos+1), intent(out) :: data
    integer,                 intent(out) :: error_flag

    ! local variables
    integer :: i, tmp_startpos, tmp_endpos, pos, bufpos

    error_flag = no_error
    IF (.not. associated(BufferedFile)) THEN
       print *,"ERROR in ReadBytesBuffered:"
       print *,"the BufferedFile pointer seems not associated"
       error_flag = error_programming
       return
    END IF

    ! remark: eliminating the if-thens inside the loop below has a
    ! significant effect on the performance, so be very carefull what
    ! you place inside this loop...
    tmp_startpos = startpos
    dataloop: DO

       ! replacing this call by the if-then below saves
       ! almost a full second of processing time on the big TENTI file...

       !IF (LocalVerbosity .ge. VerbosityReport) &
       !     print *,"requesting data"
       !call ReadDataIntoBuffer(BufferedFile,tmp_startpos,error_flag)
       !IF (error_flag .ne. no_error) return

       ! check if we need to read some data or not
       IF (BufferedFile%BufferFilled) THEN
          IF ( (tmp_startpos .lt. BufferedFile%BufferStartPos) .or. &
               (tmp_startpos .gt. BufferedFile%LastDataByte)         ) THEN
             !print *,"reading data into buffer: tmp_startpos = ",tmp_startpos
             call ReadDataIntoBuffer(BufferedFile,tmp_startpos,error_flag)
             IF (error_flag .ne. no_error) return             
          END IF
       ELSE
          !print *,"reading data into buffer: tmp_startpos = ",tmp_startpos
          call ReadDataIntoBuffer(BufferedFile,tmp_startpos,error_flag)
          IF (error_flag .ne. no_error) return
       END IF

       ! then copy the data
       tmp_endpos = min(endpos,BufferedFile%LastDataByte)
       
       !IF (LocalVerbosity .ge. VerbosityReport) &
       !     print *,"tmp_startpos,tmp_endpos = ",tmp_startpos,tmp_endpos
       pos    = tmp_startpos-startpos
       bufpos = tmp_startpos-BufferedFile%BufferStartPos
       DO i=tmp_startpos,tmp_endpos
          pos=pos+1
          bufpos=bufpos+1
          data(pos) = BufferedFile%buffer(bufpos)
          !print *,"copying a byte: i,pos,bufpos,byte = ",&
          !        i,pos,bufpos,ichar(data(pos))
       END DO
       ! is the data complete?
       IF (tmp_endpos .eq. endpos) THEN
          !IF (LocalVerbosity .ge. VerbosityReport) &
          !     print *,"ok, data is complete"
          exit dataloop
       ELSE
          IF (LocalVerbosity .ge. VerbosityReport) &
               print *,"data is not yet complete"
          tmp_startpos = tmp_endpos+1          
       END IF
    END DO dataloop

    ! remember where we are in the file
    BufferedFile%nextbyte = endpos+1

    !print *,"inside ReadBytesBuffered"
    !print *,"BufferedFile%nextbyte = ",BufferedFile%nextbyte
    !print *,"testjos: inside ReadBytesBuffered():"
    !print "('data(1:4) = ',4(i3.3,1x),' filepos = ',i8)",&
    !     ichar(data(1)),ichar(data(2)),&
    !     ichar(data(3)),ichar(data(4)),startpos

  end subroutine ReadBytesBuffered
    !  #]
  subroutine ReadNextBytesBuffered(BufferedFile,nbytes,data,error_flag)
    !  #[
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer,                 intent(in)  :: nbytes
    character(len=1), dimension(nbytes), intent(out) :: data
    integer,                 intent(out) :: error_flag

    ! local variables
    integer :: startpos, endpos

    error_flag = no_error
    IF (.not. associated(BufferedFile)) THEN
       print *,"ERROR in ReadNextBytesBuffered:"
       print *,"the BufferedFile pointer seems not associated"
       error_flag = error_programming
       return
    END IF

    startpos = BufferedFile%nextbyte
    endpos   = startpos+nbytes-1
    data(:) = achar(0)
    call ReadBytesBuffered(BufferedFile,startpos,endpos,data,error_flag)

    !print *,"testjos: inside ReadNextBytesBuffered():"
    !print "('data(1:4) = ',4(i3.3,1x),' filepos = ',i8)",&
    !     ichar(data(1)),ichar(data(2)),&
    !     ichar(data(3)),ichar(data(4)),startpos

  end subroutine ReadNextBytesBuffered
    !  #]
  subroutine WriteBytesBuffered(BufferedFile,startpos,endpos,data,error_flag)
    !  #[
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer,                 intent(in)  :: startpos
    integer,                 intent(in)  :: endpos
    character(len=1), dimension(endpos-startpos+1), intent(in) :: data
    integer,                 intent(out) :: error_flag

    ! local variables
    integer :: i

    error_flag = no_error
    IF (.not. associated(BufferedFile)) THEN
       print *,"ERROR in WriteBytesBuffered:"
       print *,"the BufferedFile pointer seems not associated"
       error_flag = error_programming
       return
    END IF

    dataloop: DO i=startpos,endpos
       ! initialise the buffer on first entry
       IF (.not. BufferedFile%BufferFilled) THEN
          call Init_New_Write_Buffer(BufferedFile,i)
       END IF

       ! switch to a next buffer when we leave one
       IF ( (i .lt. BufferedFile%BufferStartPos) .or. &
            (i .gt. BufferedFile%BufferEndPos)        ) THEN
          call Flush_Buffer(BufferedFile,error_flag)
          IF (error_flag .ne. no_error) return
          call Init_New_Write_Buffer(BufferedFile,i)
       END IF

       ! then copy the data
       BufferedFile%buffer(i-BufferedFile%BufferStartPos+1) = &
            data(i-startpos+1)

    END DO dataloop

    ! remember where we are in the file
    BufferedFile%nextbyte = endpos+1

    !print *,"setting BufferedFile%nextbyte = ",BufferedFile%nextbyte,&
    !     " startpos = ",startpos," endpos = ",endpos

    !print *,"inside WriteBytesBuffered"
    !print *,"BufferedFile%nextbyte = ",BufferedFile%nextbyte
    !print *,"testjos: inside WriteBytesBuffered():"
    !print "('data(1:4) = ',4(i3.3,1x),' filepos = ',i8)",&
    !     ichar(data(1)),ichar(data(2)),&
    !     ichar(data(3)),ichar(data(4)),startpos

  end subroutine WriteBytesBuffered
    !  #]
  subroutine WriteNextBytesBuffered(BufferedFile,nbytes,data,error_flag)
    !  #[
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer,                 intent(in)  :: nbytes
    character(len=1), dimension(nbytes), intent(in) :: data
    integer,                 intent(out) :: error_flag

    ! local variables
    integer :: startpos, endpos

    error_flag = no_error
    IF (.not. associated(BufferedFile)) THEN
       print *,"ERROR in WriteNextBytesBuffered:"
       print *,"the BufferedFile pointer seems not associated"
       error_flag = error_programming
       return
    END IF

    startpos = BufferedFile%nextbyte
    endpos   = startpos+nbytes-1
    call WriteBytesBuffered(BufferedFile,startpos,endpos,data,error_flag)

  end subroutine WriteNextBytesBuffered
    !  #]
  !------------------------------------------
  ! private routines
  !------------------------------------------
  subroutine InitBufferedFile(BufferedFile, error_flag)
    !  #[
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer,                 intent(out) :: error_flag

    error_flag = no_error
    IF (.not. associated(BufferedFile)) THEN
       print *,"ERROR in InitBufferedFile:"
       print *,"the BufferedFile pointer seems not associated"
       error_flag = error_programming
       return
    END IF

    ! initialise the BufferedFile struct
    BufferedFile%fileunit       = -1
    BufferedFile%nextbyte       = -1
    BufferedFile%filesize       = -1
    BufferedFile%action         = undefined_action
    BufferedFile%file_is_open   = .false.
    BufferedFile%filename       = "[undefined]"
    BufferedFile%BufferStartPos = -1
    BufferedFile%BufferEndPos   = -1
    BufferedFile%LastDataByte   = -1
    BufferedFile%BufferFilled   = .false.
    BufferedFile%BufferSize     = -1
    BufferedFile%count_reads    = 0
    BufferedFile%count_writes   = 0
    nullify(BufferedFile%buffer)
       
  end subroutine InitBufferedFile
    !  #]
  subroutine PrintBufferedFileProperties(BufferedFile)
    !  #[
    TYPE(BufferedFile_type), pointer :: BufferedFile ! result

    IF (.not. associated(BufferedFile)) THEN
       print *,"[the BufferedFile pointer seems not associated]"
       return
    END IF

    print *,"fileunit       = ",BufferedFile%fileunit
    print *,"nextbyte       = ",BufferedFile%nextbyte
    print *,"filesize       = ",BufferedFile%filesize
    select case (BufferedFile%action)
    case(open_for_reading)
       print *,"action         = open_for_reading"
    case(open_for_writing)
       print *,"action         = open_for_writing"
    end select
    print *,"file_is_open   = ",BufferedFile%file_is_open
    print *,"filename       = ",trim(BufferedFile%filename)
    print *,"BufferStartPos = ",BufferedFile%BufferStartPos
    print *,"BufferEndPos   = ",BufferedFile%BufferEndPos
    print *,"LastDataByte   = ",BufferedFile%LastDataByte
    print *,"BufferFilled   = ",BufferedFile%BufferFilled
    print *,"BufferSize     = ",BufferedFile%BufferSize
    print *,"count_reads    = ",BufferedFile%count_reads
    print *,"count_writes   = ",BufferedFile%count_writes
    IF (associated(BufferedFile%buffer)) THEN
       print *,"size(buffer)         = ",size(BufferedFile%buffer)
    ELSE
       print *,"buffer               = [not associated]"
    END IF
    !     character(len=1), dimension(:), pointer :: buffer
    !print *,"",BufferedFile%
    
  end subroutine PrintBufferedFileProperties
    !  #]
  subroutine ReadDataIntoBuffer(BufferedFile,startpos,error_flag)
    !  #[
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer,                 intent(in)  :: startpos
    integer,                 intent(out) :: error_flag

    ! local variables
    integer :: record, record_startpos, record_endpos,ios
    
    error_flag = no_error
    IF (.not. associated(BufferedFile)) THEN
       print *,"ERROR in ReadDataIntoBuffer:"
       print *,"the BufferedFile pointer seems not associated"
       error_flag = error_programming
       return
    END IF

    IF (.not. BufferedFile%file_is_open) THEN
       print *,"ERROR in ReadDataIntoBuffer:"
       print *,"the file is not yet open!!!!"
       print *,"please first open a file before requesting data from it..."
       error_flag = error_programming
       return
    END IF

    IF (startpos .gt. BufferedFile%filesize) THEN
       print *,"ERROR in ReadDataIntoBuffer:"
       print *,"the requested startposition for this datablock"
       print *,"is not within the size of this file !!!!"
       print *,"requested startpos = ",startpos
       print *,"while filesize     = ",BufferedFile%filesize
       ! remark: when writing to a file, this should be allowed,
       ! so for that code we need an exception to handle that.
       error_flag = error_reading_file
       return
    END IF

    ! check if we need to read some data or not
    IF (BufferedFile%BufferFilled) THEN
       IF ( (startpos .ge. BufferedFile%BufferStartPos) .and. &
            (startpos .le. BufferedFile%LastDataByte)         ) THEN
          IF (LocalVerbosity .ge. VerbosityReport) &
               print *,"not reading data, since it is already in the buffer"
          return
       END IF
    END IF

    record          = 1+(startpos-1)/BufferedFile%BufferSize
    record_startpos = 1+(record-1)*BufferedFile%BufferSize
    record_endpos   =    record   *BufferedFile%BufferSize

    IF (record_endpos .gt. BufferedFile%filesize) THEN
       call ReadFileRemainderIntoBuffer(BufferedFile,startpos,error_flag)
       IF (error_flag .ne. no_error) return
    ELSE
       !  #[ read blocks of buffersize bytes at once,
       ! since this gives a huge improvement in performance
       ! compared to reading bytes one at a time
       !print *,"Reading data into buffer:"
       read(BufferedFile%fileunit,rec=record,iostat=ios) BufferedFile%buffer
       IF (ios .ne. 0) goto 999
       !print *,"reading record: ",record
       !print *,"startpos = ",startpos
       !print *,"record_startpos = ",record_startpos
       !print *,"record_endpos   = ",record_endpos

       BufferedFile%count_reads = BufferedFile%count_reads + 1

       ! do some bookkeeping
       BufferedFile%BufferStartPos = record_startpos
       BufferedFile%BufferEndPos   = record_startpos+BufferedFile%BufferSize-1
       BufferedFile%LastDataByte   = record_endpos
       BufferedFile%BufferFilled   = .true.

       !  #]
    END IF
    
    return

999 print *,"ERROR in  ReadDataIntoBuffer:"
    print *,"ERROR could not read from BufferedFile, with properties: "
    call PrintBufferedFileProperties(BufferedFile)    
    error_flag = error_reading_file
    return
    
  end subroutine ReadDataIntoBuffer
    !  #]
  subroutine ReadFileRemainderIntoBuffer(BufferedFile,startpos,error_flag) 
    !  #[

    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer,                 intent(in)  :: startpos
    integer,                 intent(out) :: error_flag
    
    ! local variables
    integer     :: i, pos, endpos, record, local_buffersize, ios
    integer     :: AllocStatus, local_offset
    integer     :: remainder_startpos, remainder_endpos
    logical     :: switch_to_byte_reading
    logical     :: switch_to_smaller_buffersize
    character(len=1), dimension(:), pointer :: local_buffer

    ! some initialisations
    local_buffersize = BufferedFile%BufferSize
    nullify(local_buffer)
    BufferedFile%buffer(:) = ' '
    local_offset = 0

    record=1+(startpos-1)/BufferedFile%BufferSize
    remainder_startpos = 1+(record-1)*BufferedFile%BufferSize
    remainder_endpos   = BufferedFile%filesize
    IF ( remainder_endpos .ge. &
         remainder_startpos+BufferedFile%BufferSize-1) then
       print *,"Error in  ReadFileRemainderIntoBuffer()"
       print *,"Wrong use of this function:"
       print *,"the file remainders seems larger than the buffersize !!!"
       error_flag = error_programming
       return
    END IF

    read_file_remainder: DO 

       ! inspect the buffer and the size of the requested datablock
       ! and see if we need to shrink the buffer 
       switch_to_smaller_buffersize = .false.
       switch_to_byte_reading = .false.
       
       endpos = startpos+local_offset+local_BufferSize-1
       !print *,"Reading fileremainder into buffer:"
       !print *,"startpos+local_offset = ",startpos+local_offset,&
       !     " endpos = ",endpos
       !print *,"BufferedFile%filesize = ",BufferedFile%filesize
       !print *,"local_Buffersize = ",local_BufferSize

       ! fill buffer partially, with the remainder of the file
       !  #[ determine new buffersize or bytereading
       shrink_buffersize: DO
          IF (local_buffersize .ge. 2*SmallestBuffersize) THEN
             ! halve the buffersize, and try again if it fits
             local_buffersize = local_buffersize/2
             endpos = startpos+local_offset+local_buffersize-1
             IF (endpos .le. BufferedFile%filesize) THEN
                ! ok, the new buffersize seems to be usable for reading
                ! some of the remaining bytes
                switch_to_smaller_buffersize = .true.
                exit shrink_buffersize
             END IF
          ELSE
             ! switch to reading separate bytes
             endpos = BufferedFile%filesize
             switch_to_byte_reading = .true.
             exit shrink_buffersize
          END IF
       END DO shrink_buffersize
       !  #]

       IF (switch_to_smaller_buffersize) THEN
          !  #[
          !print *,"testjos: switch_to_smaller_buffersize = ",&
          !     switch_to_smaller_buffersize
          
          ! close the file
          close(unit=BufferedFile%fileunit)
          
          ! reopen the file
          open(unit=BufferedFile%fileunit,&
               file=BufferedFile%filename,status="old",&
               action="read",access="direct",&
               recl=local_buffersize,form="UNFORMATTED",ERR=998)

          IF (associated(local_buffer)) THEN
             ! if needed deallocate the local buffer
             deallocate(local_buffer)
          END IF

          ! allocate the smaller buffer
          ! request a block of memory for the buffer
          allocate(local_buffer(local_BufferSize),stat=AllocStatus)
          IF (AllocStatus .ne. 0) THEN
             print *,"Allocation problem in  ReadDataIntoBuffer:"
             error_flag = error_allocate
             return
          END IF
          !  #]
       END IF

       IF (switch_to_byte_reading) THEN
          !  #[
          ! if the remainder of the file is smaller than the minimum
          ! allowed buffersize we read these bytes one by one...
          ! However, to be able to do this, we must reopen the file
          ! with recl=1 
          
          !print *,"testjos: switch_to_byte_reading       = ",&
          !     switch_to_byte_reading

          ! close the file
          close(unit=BufferedFile%fileunit)
          
          ! reopen the file
          open(unit=BufferedFile%fileunit,&
               file=BufferedFile%filename,status="old",&
               action="read",access="direct",recl=1,form="UNFORMATTED",ERR=998)
          
          DO i=startpos+local_offset,endpos
             
             pos=i-startpos+1
             read(BufferedFile%fileunit,rec=i,iostat=ios) &
                  BufferedFile%buffer(pos:pos)
             IF (ios .ne. 0) goto 999
             !print *,"reading a byte: i,pos,byte=",&
             !        i,pos,ichar(BufferedFile%buffer(pos:pos))
             BufferedFile%count_reads = BufferedFile%count_reads + 1
          END DO

          ! this will break the do-loop
          local_offset = endpos+1

          !  #]
       ELSE
          !  #[ read blocks of local_buffersize bytes at once
          ! since this gives a huge improvement in performance
          ! compared to reading bytes one at a time
          record=1+(startpos+local_offset-1)/local_BufferSize
          read(BufferedFile%fileunit,rec=record,iostat=ios) local_buffer
          IF (ios .ne. 0) goto 999
          !print *,"reading record: ",record
          !print *,"startpos = ",startpos+local_offset," endpos = ",&
          !     startpos+local_offset+local_BufferSize-1
          BufferedFile%count_reads = BufferedFile%count_reads + 1

          ! copy the data into the main Buffer
          BufferedFile%buffer(1+local_offset:local_BufferSize+local_offset) = &
               local_buffer(1:local_BufferSize)
          local_offset = local_offset + local_BufferSize
          !print *,"new local_offset = ",local_offset
          !  #]         
       END IF
       
       IF (local_offset .gt. BufferedFile%filesize) exit read_file_remainder

    END DO read_file_remainder

    ! do some bookkeeping
    BufferedFile%BufferStartPos = remainder_startpos 
    BufferedFile%BufferEndPos   = remainder_startpos+BufferedFile%BufferSize-1
    BufferedFile%LastDataByte   = remainder_endpos
    BufferedFile%BufferFilled   = .true.
    
    IF (associated(local_buffer)) THEN
       ! if needed deallocate the local buffer
       deallocate(local_buffer)
    END IF

    ! close the file
    close(unit=BufferedFile%fileunit)
    
    ! reopen it again with recl=buffersize
    ! in case more data is needed earlier in the file
    open(unit=BufferedFile%fileunit,&
         file=BufferedFile%filename,status="old",&
         action="read",access="direct",&
         recl=BufferedFile%BufferSize,form="UNFORMATTED",ERR=998)
    
    return

998 print *,"ERROR in ReadFileRemainderIntoBuffer:"
    print *,"ERROR could not open inputfile: ",trim(BufferedFile%filename)
    error_flag = error_opening_file
    return

999 print *,"ERROR in ReadFileRemainderIntoBuffer:"
    print *,"ERROR could not read from BufferedFile, with properties: "
    call PrintBufferedFileProperties(BufferedFile)    
    error_flag = error_reading_file
    return

  end subroutine ReadFileRemainderIntoBuffer
    !  #]
  subroutine Init_New_Write_Buffer(BufferedFile,pos)
    !  #[
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer,                 intent(in)  :: pos

    ! local variable
    integer :: record

    record                      = 1+(pos-1)/BufferedFile%BufferSize
    BufferedFile%BufferFilled   = .true.
    BufferedFile%BufferStartPos = 1+(record-1)*BufferedFile%BufferSize
    BufferedFile%BufferEndPos   =    record   *BufferedFile%BufferSize

! to add: read the buffer from the file of this location in the file
!         already exists!
!         This may occur if more or less random locations are written.
!         however, since for now we will be doing sequential writing
!         this is not yet implemented ...

    BufferedFile%buffer(:)      = ' '

    !print *,"Initialised writing buffer for record: ",record," pos = ",pos

  end subroutine Init_New_Write_Buffer
    !  #]
  subroutine Flush_Buffer(BufferedFile,error_flag,remainder)
    !  #[
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer,                 intent(out) :: error_flag
    logical, optional,       intent(in)  :: remainder

    ! local variable
    integer :: i,pos,record
    logical :: flush_remainder

    error_flag = no_error

    flush_remainder = .false.
    IF (present(remainder)) flush_remainder = remainder

    IF (flush_remainder) THEN
       !print *,"Flushing partial buffer ..."
       !print *,"BufferedFile%BufferStartPos = ",BufferedFile%BufferStartPos
       !print *,"BufferedFile%nextbyte       = ",BufferedFile%nextbyte
       ! write the remainder of the bytes one by one...
       ! However, to be able to do this, we must reopen the file
       ! with recl=1 

       ! no bytes have been written yet, so nothing to write ...
       ! (to introduce a changed flag for the buffer)
       IF (BufferedFile%nextbyte .eq. 1) return
          
       ! close the file
       close(unit=BufferedFile%fileunit)
          
       ! reopen the file
       open(unit=BufferedFile%fileunit,&
            file=BufferedFile%filename,status="old",&
            action="write",access="direct",recl=1,form="UNFORMATTED",ERR=998)
          
       DO i=BufferedFile%BufferStartPos,BufferedFile%nextbyte-1
          pos=i-BufferedFile%BufferStartPos+1
          write(BufferedFile%fileunit,rec=i,ERR=999) &
                BufferedFile%buffer(pos:pos)
          !print *,"writing a byte: i,pos,byte,char",&
          !        i,pos,ichar(BufferedFile%buffer(pos:pos)),&
          !        BufferedFile%buffer(pos:pos)
          BufferedFile%count_writes = BufferedFile%count_writes + 1
       END DO

       ! close the file
       !close(unit=BufferedFile%fileunit)
       
       ! reopen it again with recl=buffersize
       ! in case more data is needed earlier in the file
       !open(unit=BufferedFile%fileunit,&
       !     file=BufferedFile%filename,status="old",&
       !     action="read",access="direct",&
       !     recl=BufferedFile%BufferSize,form="UNFORMATTED",ERR=998)

    ELSE
       !print *,"Flushing buffer ..."
       record = 1+(BufferedFile%BufferStartPos-1)/BufferedFile%BufferSize
       write(BufferedFile%fileunit,rec=record,ERR=999) BufferedFile%buffer
       BufferedFile%count_writes = BufferedFile%count_writes + 1
    END IF

    return

998 print *,"ERROR in  Flush_Buffer:"
    print *,"ERROR could not open inputfile: ",trim(BufferedFile%filename)
    error_flag = error_opening_file
    return

999 print *,"ERROR in Flush_Buffer:"
    print *,"ERROR could not write to BufferedFile, with properties: "
    call PrintBufferedFileProperties(BufferedFile)    
    error_flag = error_reading_file
    return

  end subroutine Flush_Buffer
    !  #]
  !------------------------------------------
end module Buffering
