module DirectBinaryIO
  !  #[ Description:
  !   a generic module to read/write binary data in c/c++ format
  !   taking byteswapping into account (if needed/wanted)
  !   from a direct access file
  !
  ! split from KVT_module: 12-07-2005 by J. de Kloe
  !
  ! Modifications:
  !    05-Sep-2005  D. Tan     Add read_real8array_from_bin_file.
  !    09-Nov-2005  J. de Kloe Added read routines for all int and real types
  !    29-Nov-2005  P. Poli    Added routine write_logical1_to_bin_file
  !    12-Dec-2005  D. Tan     Declare as public write_int8_to_bin_file
  !    19-Dec-2005  J. de Kloe Change default from private to public
  !    04-Jan-2006  J. de Kloe added routine write_uns_int1_to_bin_file
  !    09-Jan-2006  P. Poli/J. de Kloe
  !                      Added routines write_uns_int1array_to_bin_file,
  !                      write_real8_to_bin_file, write_uns_int2_to_bin_file
  !                      write_uns_int4_to_bin_file, read_zero_bytes_bin_file
  !    23-Jan-2006  J. de Kloe added open/close routines
  !    07-Mar-2006  P. Poli flag_nocheck added to read_logical1_from_bin_file()
  !    07-Mar-2006  J. de Kloe added some interface routines to the new
  !                            Buffering module
  !    12-May-2006  D. Tan     Null use of ParKind1 module
  !    27-Jul-2006  J. de Kloe added a first few generic routines
  !    28-Jul-2006  J. de Kloe added deletion of tempfile and use of pid
  !    31-Jul-2006  J. de Kloe added generic Set/Get/Advance filepointer
  !                            routines and open/close routines
  !    20-Mar-2007  J. de Kloe changed flag_nocheck to workaround_in in 
  !                            read_logical1_from_bin_file, and changed the
  !                            the code to display a warning in stead of an
  !                            error if an exception occurs and the workaround 
  !                            is activated.
  !    11-Apr-2007  D. Tan     Upgrade one log_warn to log_error 
  !    09-Apr-2007  J. de Kloe added some debugging code in 
  !                            read_logical1_from_bin_file()
  !    16-Jan-2008  J. de Kloe phase out integer kind i_ 
  !
  !  #]
  !  #[ Modules used:
  USE Logging, only: logmsg,log_debug,log_info,log_progress,log_warn,log_error

  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.

  ! define the default integer and real to be used by the module
  USE Numerics, only: i1_, i2_, i4_, i8_, r4_, r8_, &
                      nbytes_i1_, nbytes_i2_, nbytes_i4_, &
                      nbytes_i8_, nbytes_r4_, nbytes_r8_, &
                      check_variable_sizes

  USE Buffering, only: BufferedFile_type, InitBuffering, OpenBufferedFile,&
                       CloseBufferedFile, ReadBytesBuffered,&
                       ReadNextBytesBuffered, WriteNextBytesBuffered,&
                       GetFilePointer_Buff     => GetFilePointer, &
                       SetFilePointer_Buff     => SetFilePointer, &
                       AdvanceFilePointer_Buff => AdvanceFilePointer

  USE ErrorHandler, only: no_error, error_programming, error_allocate, &
                          error_reading_file, error_writing_file, &
                          error_opening_file, error_c_interface
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE LunManager, only: get_lun, free_lun
  USE c_support, only: get_filesize, get_process_id, remove_file

  !  #]
  !  #[ Parameter/Variable/Type declarations
  
  implicit none

  ! we use direct access only to circumvent the 4/8 byte head/tail that
  ! fortran usually writes when doing unformatted output. 
  ! Mostly we just step sequentially through the file using commands
  ! like: inquire(fileunit,nextrec=pos); read(unit,rec=pos,iostat=ios)
  !
  ! However, this means that repositioning to the beginning of the file
  ! becomes impossible, because the rewind statement does not work in 
  ! combination with direct access file IO.
  ! Therefore I introduced the next type, that remembers the next
  ! byte to be read from the file. It also remembers the filesize to 
  ! prevent having to call the get_file_size function (which interfaces
  ! with the stat() function in c) many times.

  TYPE DirectBinaryIO_fileunit_type
     logical :: use_buffering
     integer :: fileunit
     integer :: nextbyte
     integer :: filesize
     TYPE(BufferedFile_type), pointer :: buffered_fileunit
  END TYPE DirectBinaryIO_fileunit_type

  ! the default verbosity setting is Err (as silent as possible)
  integer, private, save :: LocalVerbosity = VerbosityError

  ! switch determining whether numerical values should be swapped
  ! before writing to the binairy output file or not
  logical, private, save :: do_byteswap = .false.

  integer, save :: headtailsize_this_machine

  ! save the initialisation status
  logical, private, save :: module_is_initialised = .false.

  integer, parameter :: DefaultByteOrder = 0
  integer, parameter :: LittleEndian     = 1
  integer, parameter :: BigEndian        = 2
  !  #]
  !  #[ interfaces for overloading functions
  INTERFACE OpenDirectBinaryFile
     module procedure OpenDirectBinaryFile_gen, & ! new generic type
                      OpenDirectBinaryFile_old    ! old direct type
  END INTERFACE
  INTERFACE CloseDirectBinaryFile
     module procedure CloseDirectBinaryFile_gen, & ! new generic type
                      CloseDirectBinaryFile_old    ! old direct type
  END INTERFACE
  INTERFACE GetFilePointer
     module procedure GetFilePointer_Buff,   & ! old buffered type
                      GetFilePointer_unBuff, & ! old direct type
                      GetFilePointer_gen       ! new generic type
  END INTERFACE
  INTERFACE SetFilePointer
     module procedure SetFilePointer_Buff,   & ! old buffered type
                      SetFilePointer_unBuff, & ! old direct type
                      SetFilePointer_gen       ! new generic type
  END INTERFACE
  INTERFACE AdvanceFilePointer
     module procedure AdvanceFilePointer_Buff,   & ! old buffered type
                      AdvanceFilePointer_unBuff, & ! old direct type
                      AdvanceFilePointer_gen       ! new generic type
  END INTERFACE
  INTERFACE write_int2_to_bin_file
     module procedure write_int2_to_buffbin_file,   & ! old buffered type
                      write_int2_to_unbuffbin_file, & ! old direct type
                      write_int2_to_genbin_file       ! new generic type
  END INTERFACE
  INTERFACE write_int4_to_bin_file
     module procedure write_int4_to_buffbin_file,   & ! old buffered type
                      write_int4_to_unbuffbin_file, & ! old direct type
                      write_int4_to_genbin_file       ! new generic type
  END INTERFACE
  INTERFACE write_int8_to_bin_file
     module procedure write_int8_to_buffbin_file,   & ! old buffered type
                      write_int8_to_unbuffbin_file, & ! old direct type
                      write_int8_to_genbin_file       ! new generic type
  END INTERFACE
  INTERFACE write_uns_int2_to_bin_file
     !                1234567890123456789012345678901
     module procedure write_uns_int2_to_buffbin_file,  & ! old buffered type
                      write_uns_int2_to_unbuffbinfile, & ! old direct type
                      write_uns_int2_to_genbin_file      ! new generic type
  END INTERFACE
  INTERFACE write_uns_int4_to_bin_file
     module procedure write_uns_int4_to_buffbin_file,  & ! old buffered type
                      write_uns_int4_to_unbuffbinfile, & ! old direct type
                      write_uns_int4_to_genbin_file      ! new generic type
  END INTERFACE
  !  #]
  !  #[ Module interface
  ! keep the following routines private to this module
  private :: get_machine_mem_properties, get_machine_io_properties, &
             int1_to_unsigned, int2_to_unsigned, int4_to_unsigned
  !  #]
contains
  !----------------------------------------------------
  subroutine Init_DirectBinaryIO_module(error_flag, FileByteOrder, &
                                        ChoosenVerbosity)
    !  #[

    integer,           intent(in)  :: FileByteOrder ! Little or BigEndian
    integer,           intent(out) :: error_flag
    integer, optional, intent(in)  :: ChoosenVerbosity

    ! local variables
    logical :: mem_is_little_endian
    logical :: io_is_little_endian

    error_flag = no_error

    IF (present(ChoosenVerbosity)) LocalVerbosity = ChoosenVerbosity

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       call print_verbosity_setting(LocalVerbosity,&
            ModuleName="DirectBinaryIO")
    END IF

    ! this only sets the verbosity setting, so it doesn't do any harm
    ! if buffering is not used
    call InitBuffering!(ChoosenVerbosity=VerbosityReport)       

    ! check if the variable types have the number of bytes that we expect
    ! Some compilers use words in stead of bytes as default recordlength
    ! for direct access, so this catches use of these compilers
    ! with the wrong commandline options. (usually you have to specify
    ! some option like -bytereclen to set the default recordlength to 1 byte)
    call check_variable_sizes(error_flag)
    IF (error_flag .ne. no_error) return

    ! see whether the program uses a Big or Little endian representation
    ! of numbers in memory
    do_byteswap = .false.
    call get_machine_mem_properties(mem_is_little_endian,error_flag)
    if (error_flag .ne. no_error) return

    ! see whether the program uses a Big or Little endian representation
    ! of numbers when writing to file, and if the head/tail fields 
    ! are 4 or 8 bytes long
    call get_machine_io_properties(io_is_little_endian, &
                                   headtailsize_this_machine, &
                                   error_flag)
    if (error_flag .ne. no_error) return

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"mem_is_little_endian      = ",mem_is_little_endian
       print *,"io_is_little_endian       = ",io_is_little_endian
       print *,"headtailsize_this_machine = ",headtailsize_this_machine
    END IF

    ! REMARK1:
    ! according to L1B_IODD version 2.1 (file 521666_IODD_2_1.pdf):
    ! section 3.2.5, bit and byte ordering:
    ! For external interfaces bit 0 shall be the most significant bit,
    ! byte 0 shall be the most significant byte.
    !
    ! This is usually referred to as "big_endian", and is the default
    ! on many unix machines, but not on linux !
    !
    ! So EE_CFI files should always be read/written with the flag
    ! FileByteOrder = BigEndian set

    ! REMARK2:
    ! binary files on linux (i.e. the RayleighBrillouin lookup tables)
    ! are in little endian representation !!!
    ! So these files should always be read/written with the flag
    ! FileByteOrder = LittleEndian set

    select case (FileByteOrder)
    case(DefaultByteOrder)
       ! just use the default byte order for this machine,
       ! do not do any byteswapping.
       ! This is needed to be able to read numbers written by the test
       ! program WriteSomeNumbers.c, since in there no byteswapping
       ! is implemented, and the default byte order for the current 
       ! machine is used.
       ! Remind that ee_cfi formatted files have a forced big_endian
       ! format regardless of the machine on which the software runs
       do_byteswap = .false.
       IF (LocalVerbosity .ge. VerbosityReport) THEN
          print *,"bypassing all byteorder switches and just using the"
          print *,"byte order that is default for this platform ..."
       END IF
    case(BigEndian)
       IF (mem_is_little_endian) THEN
          do_byteswap = .true.
          IF (LocalVerbosity .ge. VerbosityReport) THEN
             print *,"The program uses a little_endian memory representation"
             print *,"so turning byteswapping on for binary reading/writing"
             print *,"which we do byte by byte in fortran."
          END IF
       ELSE
          IF (LocalVerbosity .ge. VerbosityReport) THEN
             print *,"The program uses a big_endian memory representation"
          END IF
       END IF
    case(LittleEndian)
       IF (mem_is_little_endian) THEN
          IF (LocalVerbosity .ge. VerbosityReport) THEN
             print *,"The program uses a little_endian memory representation"
          END IF
       ELSE
          do_byteswap = .true.
          IF (LocalVerbosity .ge. VerbosityReport) THEN
             print *,"The program uses a big_endian memory representation"
             print *,"so turning byteswapping on for binary reading/writing"
             print *,"which we do byte by byte in fortran."
          END IF
       END IF
    case default
       print *,"invalid value for FileByteOrder: ",FileByteOrder
       error_flag = error_programming
       return
    end select

    module_is_initialised = .true.

  end subroutine Init_DirectBinaryIO_module
    !  #]
  !----------------------------------------------------
  !  open/close routines
  !----------------------------------------------------
  subroutine OpenDirectBinaryFile_gen(filename,action,fileunit,&
                                      error_flag,UseBuffering)
    !  #[
    character(len=*), intent(in)  :: filename
    character(len=*), intent(in)  :: action   ! can be: read/write
    TYPE(DirectBinaryIO_fileunit_type), intent(out)  :: fileunit
    integer,          intent(out) :: error_flag
    ! use buffering or not ?
    logical,          optional    :: UseBuffering

    ! local variables
    integer :: AllocStatus

    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in OpenDirectBinaryFile:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    call Init_DirectBinaryIO_fileunit(fileunit)

    fileunit%use_buffering = .false.
    IF (present(UseBuffering)) THEN
       IF (UseBuffering) fileunit%use_buffering = .true.
    END IF

    IF (fileunit%use_buffering) THEN
       print *,"Opening buffered file: ",trim(filename)

       allocate(fileunit%buffered_fileunit,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF
       call OpenBufferedFile(fileunit%buffered_fileunit,filename,action,&
                             error_flag, BufferSize=5000)
       IF (error_flag .ne. no_error) return
    ELSE
       print *,"Opening unbuffered file: ",trim(filename)

       call OpenDirectBinaryFile_old(filename,action,fileunit%fileunit,&
                                     error_flag)
       IF (error_flag .ne. no_error) return
    END IF

  end subroutine OpenDirectBinaryFile_gen
  !  #]

  subroutine OpenDirectBinaryFile_old(filename,action,fileunit,&
                                      error_flag,UseBuffering)
    !  #[
    character(len=*), intent(in)  :: filename
    character(len=*), intent(in)  :: action   ! can be: read/write
    integer,          intent(out) :: fileunit
    integer,          intent(out) :: error_flag
    ! use buffering or not ?
    logical,          optional    :: UseBuffering
    
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in OpenDirectBinaryFile:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    IF (present(UseBuffering)) THEN
       print *,"WARNING in OpenDirectBinaryFile_old:"
       print *,"buffering is not possible if you use the old style integer"
       print *,"fileunit for this file. Please use the new type"
       print *,"DirectBinaryIO_fileunit_type before trying to switch"
       print *,"buffring on ..."
    END IF

    fileunit = get_lun()    
    IF (fileunit .lt. 0) THEN
       print *,"ERROR in OpenDirectBinaryFile:"
       print *,"no more free fileunits available."
       print *,"call of get_lun() function failed ...."
       error_flag = error_opening_file
       return
    END IF
       
    select case(trim(action))
    case("read","READ")
       open(unit=fileunit,file=filename,status="old",&
            action="read",access="direct",recl=1,&
            form="UNFORMATTED",ERR=998)
    case("write","WRITE")
       open(unit=fileunit,file=filename,status="replace",&
            action="write",access="direct",recl=1,&
            form="UNFORMATTED",ERR=999)
    case default
       print *,"ERROR in OpenDirectBinaryFile:"
       print *,"invalid action type: ",trim(action)
       print *,"allowed values are: read, or write"
       error_flag = error_programming
       call free_lun(fileunit)
       return
    end select

    return

998 print *,"ERROR could not open inputfile: ",trim(filename)
    error_flag = error_opening_file
    return

999 print *,"ERROR could not open outputfile: ",trim(filename)
    error_flag = error_opening_file
    return

  end subroutine OpenDirectBinaryFile_old
  !  #]
  subroutine CloseDirectBinaryFile_gen(fileunit,error_flag)
    !  #[
    TYPE(DirectBinaryIO_fileunit_type), intent(inout) :: fileunit
    integer,                            intent(out)   :: error_flag

    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"WARNING in CloseDirectBinaryFile:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
    END IF

    IF (fileunit%use_buffering) THEN
       call CloseBufferedFile(fileunit%buffered_fileunit,error_flag)
       IF (error_flag .ne. no_error) return
       deallocate(fileunit%buffered_fileunit)
    ELSE
       close(unit=fileunit%fileunit)
       call free_lun(fileunit%fileunit)
    END IF

    call Init_DirectBinaryIO_fileunit(fileunit)

  end subroutine CloseDirectBinaryFile_gen
    !  #]
  subroutine CloseDirectBinaryFile_old(fileunit)
    !  #[
    integer, intent(in) :: fileunit

    IF (.not. module_is_initialised) THEN
       print *,"WARNING in CloseDirectBinaryFile:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
    END IF

    close(unit=fileunit)
    call free_lun(fileunit)
    
  end subroutine CloseDirectBinaryFile_old
    !  #]
  function byteswap_needed() result(l)
    !  #[ export this private variable
    logical :: l ! result
    l = do_byteswap
  end function byteswap_needed
    !  #]

  !----------------------------------------------------
  !  generic unbuffered/buffered routines
  !----------------------------------------------------
  subroutine Init_DirectBinaryIO_fileunit(fileunit)
    !  #[
    TYPE(DirectBinaryIO_fileunit_type), intent(out) :: fileunit

    fileunit%use_buffering = .false.
    fileunit%fileunit      = -1
    fileunit%nextbyte      = -1
    fileunit%filesize      = -1
    nullify(fileunit%buffered_fileunit)

  end subroutine Init_DirectBinaryIO_fileunit
    !  #]
  subroutine write_int2_to_genbin_file(fileunit,value,error_flag)
    !  #[

    TYPE(DirectBinaryIO_fileunit_type), intent(in)  :: fileunit
    integer(i2_),                       intent(in)  :: value
    integer,                            intent(out) :: error_flag

    IF (fileunit%use_buffering) THEN
       call write_int2_to_buffbin_file(fileunit%buffered_fileunit,&
                                       value,error_flag)
    ELSE
       call write_int2_to_unbuffbin_file(fileunit%fileunit,value,error_flag)
    END IF

  end subroutine write_int2_to_genbin_file

    !  #]
  subroutine write_int4_to_genbin_file(fileunit,value,error_flag)
    !  #[
    TYPE(DirectBinaryIO_fileunit_type), intent(in)  :: fileunit
    integer(i4_),                       intent(in)  :: value
    integer,                            intent(out) :: error_flag

    IF (fileunit%use_buffering) THEN
       call write_int4_to_buffbin_file(fileunit%buffered_fileunit,&
                                       value,error_flag)
    ELSE
       call write_int4_to_unbuffbin_file(fileunit%fileunit,value,error_flag)
    END IF

  end subroutine write_int4_to_genbin_file
    !  #]
  subroutine write_int8_to_genbin_file(fileunit,value,error_flag)
    !  #[
    TYPE(DirectBinaryIO_fileunit_type), intent(in)  :: fileunit
    integer(i8_),                       intent(in)  :: value
    integer,                            intent(out) :: error_flag

    IF (fileunit%use_buffering) THEN
       call write_int8_to_buffbin_file(fileunit%buffered_fileunit,&
                                       value,error_flag)
    ELSE
       call write_int8_to_unbuffbin_file(fileunit%fileunit,value,error_flag)
    END IF

  end subroutine write_int8_to_genbin_file
    !  #]
  subroutine write_uns_int2_to_genbin_file(fileunit,value,error_flag)
    !  #[
    TYPE(DirectBinaryIO_fileunit_type), intent(in)  :: fileunit
    integer(i4_),                       intent(in)  :: value
    integer,                            intent(out) :: error_flag

    IF (fileunit%use_buffering) THEN
       call write_uns_int2_to_buffbin_file(fileunit%buffered_fileunit,&
                                           value,error_flag)
    ELSE
       call write_uns_int2_to_unbuffbinfile(fileunit%fileunit,value,error_flag)
    END IF

  end subroutine write_uns_int2_to_genbin_file
    !  #]
  subroutine write_uns_int4_to_genbin_file(fileunit,value,error_flag)
    !  #[
    TYPE(DirectBinaryIO_fileunit_type), intent(in)  :: fileunit
    integer(i8_),                       intent(in)  :: value
    integer,                            intent(out) :: error_flag

    IF (fileunit%use_buffering) THEN
       call write_uns_int4_to_buffbin_file(fileunit%buffered_fileunit,&
                                           value,error_flag)
    ELSE
       call write_uns_int4_to_unbuffbinfile(fileunit%fileunit,value,error_flag)
    END IF

  end subroutine write_uns_int4_to_genbin_file
    !  #]
  subroutine GetFilePointer_gen(fileunit,position,error_flag)
    !  #[
    TYPE(DirectBinaryIO_fileunit_type), intent(in)  :: fileunit
    integer(i8_),                       intent(out) :: position
    integer,                            intent(out) :: error_flag

    IF (fileunit%use_buffering) THEN
       call GetFilePointer_Buff(fileunit%buffered_fileunit,position,error_flag)
    ELSE
       call GetFilePointer_unBuff(fileunit%fileunit,position,error_flag)
    END IF

  end subroutine GetFilePointer_gen
    !  #]
  subroutine SetFilePointer_gen(fileunit,position,error_flag)
    !  #[
    TYPE(DirectBinaryIO_fileunit_type), intent(in)  :: fileunit
    integer(i8_),                       intent(in)  :: position
    integer,                            intent(out) :: error_flag

    IF (fileunit%use_buffering) THEN
       call SetFilePointer_Buff(fileunit%buffered_fileunit,position,error_flag)
    ELSE
       call SetFilePointer_unBuff(fileunit%fileunit,position,error_flag)
    END IF

  end subroutine SetFilePointer_gen
    !  #]
  subroutine AdvanceFilePointer_gen(fileunit,amount,error_flag)
    !  #[
    TYPE(DirectBinaryIO_fileunit_type), intent(in)  :: fileunit
    integer(i8_),                       intent(in)  :: amount
    integer,                            intent(out) :: error_flag

    IF (fileunit%use_buffering) THEN
       call AdvanceFilePointer_Buff(fileunit%buffered_fileunit,&
                                    amount,error_flag)
    ELSE
       call AdvanceFilePointer_unBuff(fileunit%fileunit,&
                                      amount,error_flag)
    END IF

  end subroutine AdvanceFilePointer_gen
    !  #]
  !----------------------------------------------------
  !  buffered routines
  !----------------------------------------------------
  subroutine OpenBuffDirectBinaryFile(filename,action,BufferedFile,error_flag,&
                                      BufferSize)
    !  #[
    character(len=*),        intent(in)  :: filename
    character(len=*),        intent(in)  :: action   ! can be: read/write
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer,                 intent(out) :: error_flag
    integer,       optional, intent(in)  :: BufferSize
    
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in OpenBuffDirectBinaryFile:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    call OpenBufferedFile(BufferedFile,filename,action,error_flag, &
                          BufferSize=BufferSize)
    IF (error_flag .ne. no_error) return

  end subroutine OpenBuffDirectBinaryFile
  !  #]
  subroutine CloseBuffDirectBinaryFile(BufferedFile, error_flag)
    !  #[
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer,                 intent(out) :: error_flag

    IF (.not. module_is_initialised) THEN
       print *,"WARNING in CloseBuffDirectBinaryFile:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
    END IF

    call CloseBufferedFile(BufferedFile, error_flag)
    IF (error_flag .ne. no_error) return
    
  end subroutine CloseBuffDirectBinaryFile
    !  #]

  subroutine read_int2_from_buffbin_file(BufferedFile,value,error_flag)
    !  #[ read a 2 byte signed integer
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer(i2_),            intent(out) :: value
    integer,                 intent(out) :: error_flag

    ! local variables
    integer            :: i, j
    character(len=1), dimension(16)  :: byte_array
    character(len=1)   :: tmp_byte

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_int2_from_buffbin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! read the bytes from the input file
    call ReadNextBytesBuffered(BufferedFile,nbytes_i2_,&
                               byte_array(1:nbytes_i2_),error_flag)
    IF (error_flag .ne. no_error) return

    IF (do_byteswap) THEN
       ! convert Little<=>Big endian
       DO i=1,int(nbytes_i2_/2)
          j=nbytes_i2_+1-i
          ! swap bytes i and j
          tmp_byte      = byte_array(i)
          byte_array(i) = byte_array(j)
          byte_array(j) = tmp_byte
       END DO
    END IF

    ! convert the sequence of bytes stored in a chararray to 
    ! a numerical value
    ! (this only swaps the data-type and copies no data ! )
    value = transfer(byte_array(1:nbytes_i2_),value)

    !print *,"read: value = ",value

    return

  end subroutine read_int2_from_buffbin_file
    !  #]
  subroutine read_int4_from_buffbin_file(BufferedFile,value,error_flag)
    !  #[ read a 4 byte signed integer
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer(i4_),            intent(out) :: value
    integer,                 intent(out) :: error_flag

    ! local variables
    integer            :: i, j
    character(len=1), dimension(16)  :: byte_array
    character(len=1)   :: tmp_byte

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_int4_from_buffbin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! read the bytes from the input file
    call ReadNextBytesBuffered(BufferedFile,nbytes_i4_,&
                               byte_array(1:nbytes_i4_),error_flag)
    IF (error_flag .ne. no_error) return

    IF (do_byteswap) THEN
       ! convert Little<=>Big endian
       DO i=1,int(nbytes_i4_/2)
          j=nbytes_i4_+1-i
          ! swap bytes i and j
          tmp_byte      = byte_array(i)
          byte_array(i) = byte_array(j)
          byte_array(j) = tmp_byte
       END DO
    END IF

    ! convert the sequence of bytes stored in a char array to 
    ! a numerical value
    ! (this only swaps the data-type and copies no data ! )
    value = transfer(byte_array(1:nbytes_i4_),value)

    !print *,"read: value = ",value

    return

  end subroutine read_int4_from_buffbin_file
    !  #]
  subroutine read_int8_from_buffbin_file(BufferedFile,value,error_flag)
    !  #[ read a 8 byte signed integer
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer(i8_),            intent(out) :: value
    integer,                 intent(out) :: error_flag

    ! local variables
    integer            :: i, j
    character(len=1), dimension(16)  :: byte_array
    character(len=1)   :: tmp_byte

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_int8_from_buffbin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! read the bytes from the input file
    call ReadNextBytesBuffered(BufferedFile,nbytes_i8_,&
                               byte_array(1:nbytes_i8_),error_flag)
    IF (error_flag .ne. no_error) return

    IF (do_byteswap) THEN
       ! convert Little<=>Big endian
       DO i=1,int(nbytes_i8_/2)
          j=nbytes_i8_+1-i
          ! swap bytes i and j
          tmp_byte      = byte_array(i)
          byte_array(i) = byte_array(j)
          byte_array(j) = tmp_byte
       END DO
    END IF

    ! convert the sequence of bytes stored in a char array to 
    ! a numerical value
    ! (this only swaps the data-type and copies no data ! )
    value = transfer(byte_array(1:nbytes_i8_),value)

    !print *,"read: value = ",value

    return

  end subroutine read_int8_from_buffbin_file
    !  #]
  subroutine read_uns_int2_from_buffbin_file(BufferedFile,value,error_flag)
    !  #[ read a 2 byte unsigned integer
    ! WARNING: an unsigned integer of 2 bytes CANNOT be represented 
    !          in a standard 2 byte fortran integer variable,
    !          therefore an 4 byte integer is used for the result
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer(i4_),            intent(out) :: value
    integer,                 intent(out) :: error_flag

    ! local variables
    integer(i2_)       :: tmp_value

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_uns_int2_from_buffbin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! read the integer as if it where a signed integer
    call read_int2_from_buffbin_file(BufferedFile,tmp_value,error_flag)

    ! now remove the sign:
    IF (tmp_value .ge. 0) THEN
       value = int(tmp_value,i4_)
    ELSE
       value = int2_to_unsigned(tmp_value)
    END IF

    return

  end subroutine read_uns_int2_from_buffbin_file
    !  #]
  subroutine read_uns_int4_from_buffbin_file(BufferedFile,value,error_flag)
    !  #[ read a 4 byte unsigned integer
    ! WARNING: an unsigned integer of 4 bytes CANNOT be represented 
    !          in a standard 4 byte fortran integer variable,
    !          therefore an 8 byte integer is used for the result
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer(i8_),            intent(out) :: value
    integer,                 intent(out) :: error_flag

    ! local variables
    integer(i4_)       :: tmp_value

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_uns_int4_from_buffbin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! read the integer as if it where a signed integer
    call read_int4_from_buffbin_file(BufferedFile,tmp_value,error_flag)

    ! now remove the sign:
    IF (tmp_value .ge. 0) THEN
       value = int(tmp_value,i8_)
    ELSE
       value = int4_to_unsigned(tmp_value)
    END IF

    return

  end subroutine read_uns_int4_from_buffbin_file
    !  #]
  subroutine read_real8_from_buffbin_file(BufferedFile,value,error_flag)
    !  #[ read an 8 byte real from a buffered binary file
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    real(r8_),               intent(out) :: value
    integer,                 intent(out) :: error_flag

    ! local variables
    integer            :: i, j
    character(len=1), dimension(16)  :: byte_array
    character(len=1)   :: tmp_byte

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_real8_from_buffbin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! read the bytes from the input file
    call ReadNextBytesBuffered(BufferedFile,nbytes_r8_,&
                               byte_array(1:nbytes_r8_),error_flag)
    IF (error_flag .ne. no_error) return

    IF (do_byteswap) THEN
       ! convert Little<=>Big endian
       DO i=1,int(nbytes_r8_/2)
          j=nbytes_r8_+1-i
          ! swap bytes i and j
          tmp_byte      = byte_array(i)
          byte_array(i) = byte_array(j)
          byte_array(j) = tmp_byte
       END DO
    END IF

    !print *,"byte values: ",(ichar(byte_array(i:i)),i=1,8)

    ! convert the sequence of bytes stored in a string to 
    ! a numerical value
    ! (this only swaps the data-type and copies no data ! )
    value = transfer(byte_array(1:nbytes_r8_),value)

    !print *,"read: value = ",value

    return

  end subroutine read_real8_from_buffbin_file
    !  #]

  subroutine write_int2_to_buffbin_file(BufferedFile,value,error_flag)
    !  #[ write a 2 byte signed integer to a buffered file
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer(i2_),            intent(in)  :: value
    integer,                 intent(out) :: error_flag

    ! local variables
    integer            :: i, j
    character(len=1), dimension(16)  :: byte_array
    character(len=1)   :: tmp_byte

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in write_int2_to_buffbin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! convert the data to a sequence of bytes stored in a string
    ! (this only swaps the data-type and copies no data ! )
    byte_array(1:nbytes_i2_) = transfer(value,byte_array(1:nbytes_i2_))

    IF (do_byteswap) THEN
       ! convert Little<=>Big endian
       DO i=1,int(nbytes_i2_/2)
          j=nbytes_i2_+1-i
          ! swap bytes i and j
          tmp_byte = byte_array(i)
          byte_array(i) = byte_array(j)
          byte_array(j) = tmp_byte
       END DO
    END IF

    ! write the bytes to the output file
    call WriteNextBytesBuffered(BufferedFile,nbytes_i2_,&
                                byte_array(1:nbytes_i2_),error_flag)
    IF (error_flag .ne. no_error) return

    return

  end subroutine write_int2_to_buffbin_file
    !  #]
  subroutine write_int4_to_buffbin_file(BufferedFile,value,error_flag)
    !  #[ write a 4 byte signed integer to a buffered file
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer(i4_),            intent(in)  :: value
    integer,                 intent(out) :: error_flag

    ! local variables
    integer            :: i, j
    character(len=1), dimension(16)  :: byte_array
    character(len=1)   :: tmp_byte

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in write_int4_to_buffbin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! convert the data to a sequence of bytes stored in a string
    ! (this only swaps the data-type and copies no data ! )
    byte_array(1:nbytes_i4_) = transfer(value,byte_array(1:nbytes_i4_))

    IF (do_byteswap) THEN
       ! convert Little<=>Big endian
       DO i=1,int(nbytes_i4_/2)
          j=nbytes_i4_+1-i
          ! swap bytes i and j
          tmp_byte = byte_array(i)
          byte_array(i) = byte_array(j)
          byte_array(j) = tmp_byte
       END DO
    END IF

    ! write the bytes to the output file
    call WriteNextBytesBuffered(BufferedFile,nbytes_i4_,&
                                byte_array(1:nbytes_i4_),error_flag)
    IF (error_flag .ne. no_error) return

    return

  end subroutine write_int4_to_buffbin_file
    !  #]
  subroutine write_int8_to_buffbin_file(BufferedFile,value,error_flag)
    !  #[ write a 8 byte signed integer to a buffered file
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer(i8_),            intent(in)  :: value
    integer,                 intent(out) :: error_flag

    ! local variables
    integer            :: i, j
    character(len=1), dimension(16)  :: byte_array
    character(len=1)   :: tmp_byte

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in write_int8_to_buffbin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! convert the data to a sequence of bytes stored in a string
    ! (this only swaps the data-type and copies no data ! )
    byte_array(1:nbytes_i8_) = transfer(value,byte_array(1:nbytes_i8_))

    IF (do_byteswap) THEN
       ! convert Little<=>Big endian
       DO i=1,int(nbytes_i8_/2)
          j=nbytes_i8_+1-i
          ! swap bytes i and j
          tmp_byte = byte_array(i)
          byte_array(i) = byte_array(j)
          byte_array(j) = tmp_byte
       END DO
    END IF

    ! write the bytes to the output file
    call WriteNextBytesBuffered(BufferedFile,nbytes_i8_,&
                                byte_array(1:nbytes_i8_),error_flag)
    IF (error_flag .ne. no_error) return

    return

  end subroutine write_int8_to_buffbin_file
    !  #]

  subroutine write_uns_int2_to_buffbin_file(BufferedFile,value,error_flag)
    !  #[ write a 2 byte unsigned integer
    ! WARNING: an unsigned integer of 2 bytes CANNOT be represented 
    !          in a standard 2 byte fortran integer variable,
    !          therefore an 4 byte integer is used as input
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer(i4_), intent(in)  :: value
    integer,      intent(out) :: error_flag

    ! local variables
    integer            :: i, j
    character(len=1), dimension(16)  :: byte_array
    character(len=1)   :: tmp_byte

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in write_uns_int2_to_buffbin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! check if we are inside the valid range
    IF ((value .lt. 0) .or. (value .gt. 256*256-1)) THEN
       print *,"ERROR: this value of: ",value," cannot be written"
       print *,"as a 2 byte unsigned integer, since it is outside the"
       print *,"valid range for this datatype"
       error_flag = error_writing_file
       return
    END IF
    
    ! convert the data to a sequence of bytes stored in a string
    byte_array(1:4) = transfer(value,byte_array(1:4))

    IF (do_byteswap) THEN
       ! convert Little<=>Big endian
       DO i=1,int(nbytes_i4_/2)
          j=nbytes_i4_+1-i
          ! swap bytes i and j
          tmp_byte = byte_array(i)
          byte_array(i) = byte_array(j)
          byte_array(j) = tmp_byte
       END DO
    END IF

    ! then take only the last 2 bytes, which should contain the number
    ! that we are interested in
    byte_array(1:2) = byte_array(3:4)

    ! then write these 2 bytes to the output file
    call WriteNextBytesBuffered(BufferedFile,nbytes_i2_,&
                                byte_array(1:nbytes_i2_),error_flag)
    IF (error_flag .ne. no_error) return

    return

  end subroutine write_uns_int2_to_buffbin_file
    !  #]
  subroutine write_uns_int4_to_buffbin_file(BufferedFile,value,error_flag)
    !  #[ write a 4 byte unsigned integer
    ! WARNING: an unsigned integer of 4 bytes CANNOT be represented 
    !          in a standard 4 byte fortran integer variable,
    !          therefore an 8 byte integer is used as input
    TYPE(BufferedFile_type), pointer     :: BufferedFile
    integer(i8_), intent(in)  :: value
    integer,      intent(out) :: error_flag

    ! local variables
    integer            :: i, j
    character(len=1), dimension(16)  :: byte_array
    character(len=1)   :: tmp_byte

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in write_uns_int4_to_buffbin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! check if we are inside the valid range
    IF ((value .lt. 0) .or. (value .gt. 256_i8_*256*256*256-1)) THEN
       print *,"ERROR: this value of: ",value," cannot be written"
       print *,"as a 4 byte unsigned integer, since it is outside the"
       print *,"valid range for this datatype"
       error_flag = error_writing_file
       return
    END IF
    
    ! convert the data to a sequence of bytes stored in a string
    byte_array(1:8) = transfer(value,byte_array(1:8))

    IF (do_byteswap) THEN
       ! convert Little<=>Big endian
       DO i=1,int(nbytes_i8_/2)
          j=nbytes_i8_+1-i
          ! swap bytes i and j
          tmp_byte = byte_array(i)
          byte_array(i) = byte_array(j)
          byte_array(j) = tmp_byte
       END DO
    END IF

    ! then take only the last 4 bytes, which should contain the number
    ! that we are interested in
    byte_array(1:4) = byte_array(5:8)

    ! then write these 4 bytes to the output file
    call WriteNextBytesBuffered(BufferedFile,nbytes_i4_,&
                                byte_array(1:nbytes_i4_),error_flag)
    IF (error_flag .ne. no_error) return

    return

  end subroutine write_uns_int4_to_buffbin_file
    !  #]

  !----------------------------------------------------
  !  unbuffered writing routines
  !----------------------------------------------------
  subroutine write_int1_to_bin_file(fileunit,value,error_flag)
    !  #[ write a 1 byte signed integer
    integer,          intent(in)  :: fileunit
    integer(i1_),     intent(in)  :: value
    integer,          intent(out) :: error_flag

    ! local variables
    integer            :: filepos ! i,j
    character(len=256) :: filename
    character(len=16)  :: byte_array
    !character(len=1)   :: tmp_byte
    logical            :: nmd

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in write_int1_to_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! convert the data to a sequence of bytes stored in a string
    ! (this only swaps the data-type and copies no data ! )
    byte_array(1:nbytes_i1_) = transfer(value,byte_array(1:nbytes_i1_))

    ! Poli: no need to swap ?
    !IF (do_byteswap) THEN
    !   ! convert Little<=>Big endian
    !   DO i=1,int(nbytes_i1_/2)
    !      j=nbytes_i1_+1-i
    !      ! swap bytes i and j
    !      tmp_byte = byte_array(i:i)
    !      byte_array(i:i) = byte_array(j:j)
    !      byte_array(j:j) = tmp_byte
    !   END DO
    !END IF

    ! write the bytes to the output file
    inquire(unit=fileunit,nextrec=filepos)
    write(fileunit,rec=filepos,ERR=999) byte_array(1:1)
    ! loop is not needed in this case. JK
    !DO i=1,nbytes_i1_
    !   write(fileunit,rec=filepos+i-1,ERR=999) byte_array(i:i)
    !END DO

    return

999 filename = "[undefined]"
    inquire(unit=fileunit,named=nmd)
    if (nmd) inquire(unit=fileunit,name=filename)
    print *,"write_int1_to_bin_file: ERROR could not write to outputfile: ", &
            trim(filename)
    error_flag = error_writing_file
    return

  end subroutine write_int1_to_bin_file
    !  #]
  subroutine write_int2_to_unbuffbin_file(fileunit,value,error_flag)
    !  #[ write a 2 byte signed integer
    integer,          intent(in)  :: fileunit
    integer(i2_),     intent(in)  :: value
    integer,          intent(out) :: error_flag

    ! local variables
    integer            :: i, j, filepos
    character(len=256) :: filename
    character(len=16)  :: byte_array
    character(len=1)   :: tmp_byte
    logical            :: nmd

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in write_int2_to_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! convert the data to a sequence of bytes stored in a string
    ! (this only swaps the data-type and copies no data ! )
    byte_array(1:nbytes_i2_) = transfer(value,byte_array(1:nbytes_i2_))

    IF (do_byteswap) THEN
       ! convert Little<=>Big endian
       DO i=1,int(nbytes_i2_/2)
          j=nbytes_i2_+1-i
          ! swap bytes i and j
          tmp_byte = byte_array(i:i)
          byte_array(i:i) = byte_array(j:j)
          byte_array(j:j) = tmp_byte
       END DO
    END IF

    ! write the bytes to the output file
    inquire(unit=fileunit,nextrec=filepos)
    DO i=1,nbytes_i2_
       write(fileunit,rec=filepos+i-1,ERR=999) byte_array(i:i)
    END DO

    return

999 filename = "[undefined]"
    inquire(unit=fileunit,named=nmd)
    if (nmd) inquire(unit=fileunit,name=filename)
    print *,"write_int2_to_bin_file: ERROR could not write to outputfile: ", &
            trim(filename)
    error_flag = error_writing_file
    return

  end subroutine write_int2_to_unbuffbin_file
    !  #]
  subroutine write_int4_to_unbuffbin_file(fileunit,value,error_flag)
    !  #[ write a 4 byte signed integer

    integer,          intent(in)  :: fileunit
    integer(i4_),     intent(in)  :: value
    integer,          intent(out) :: error_flag

    ! local variables
    integer            :: i, j, filepos
    character(len=256) :: filename
    character(len=16)  :: byte_array
    character(len=1)   :: tmp_byte
    logical            :: nmd

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in write_int4_to_unbuffbin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! convert the data to a sequence of bytes stored in a string
    ! (this only swaps the data-type and copies no data ! )
    byte_array(1:nbytes_i4_) = transfer(value,byte_array(1:nbytes_i4_))

    IF (do_byteswap) THEN
       ! convert Little<=>Big endian
       DO i=1,int(nbytes_i4_/2)
          j=nbytes_i4_+1-i
          ! swap bytes i and j
          tmp_byte = byte_array(i:i)
          byte_array(i:i) = byte_array(j:j)
          byte_array(j:j) = tmp_byte
       END DO
    END IF

    ! write the bytes to the output file
    inquire(unit=fileunit,nextrec=filepos)
    DO i=1,nbytes_i4_
       write(fileunit,rec=filepos+i-1,ERR=999) byte_array(i:i)
    END DO

    return

999 filename = "[undefined]"
    inquire(unit=fileunit,named=nmd)
    if (nmd) inquire(unit=fileunit,name=filename)
    print *,"write_int4_to_unbuffbin_file: "
    print *,"ERROR could not write to outputfile: ",trim(filename)
    error_flag = error_writing_file
    return

  end subroutine write_int4_to_unbuffbin_file

    !  #]
  subroutine write_int8_to_unbuffbin_file(fileunit,value,error_flag)
    !  #[ write a 8 byte signed integer
    integer,          intent(in)  :: fileunit
    integer(i8_),     intent(in)  :: value
    integer,          intent(out) :: error_flag

    ! local variables
    integer            :: i, j, filepos
    character(len=256) :: filename
    character(len=16)  :: byte_array
    character(len=1)   :: tmp_byte
    logical            :: nmd

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in write_int8_to_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! convert the data to a sequence of bytes stored in a string
    ! (this only swaps the data-type and copies no data ! )
    byte_array(1:nbytes_i8_) = transfer(value,byte_array(1:nbytes_i8_))

    IF (do_byteswap) THEN
       ! convert Little<=>Big endian
       DO i=1,int(nbytes_i8_/2)
          j=nbytes_i8_+1-i
          ! swap bytes i and j
          tmp_byte = byte_array(i:i)
          byte_array(i:i) = byte_array(j:j)
          byte_array(j:j) = tmp_byte
       END DO
    END IF

    ! write the bytes to the output file
    inquire(unit=fileunit,nextrec=filepos)
    DO i=1,nbytes_i8_
       write(fileunit,rec=filepos+i-1,ERR=999) byte_array(i:i)
    END DO

    return

999 filename = "[undefined]"
    inquire(unit=fileunit,named=nmd)
    if (nmd) inquire(unit=fileunit,name=filename)
    print *,"write_int8_to_bin_file: ERROR could not write to outputfile: ", &
            trim(filename)
    error_flag = error_writing_file
    return

  end subroutine write_int8_to_unbuffbin_file
    !  #]

  subroutine write_uns_int1_to_bin_file(fileunit,value,error_flag)
    !  #[ write a 1 byte unsigned integer
    ! WARNING: an unsigned integer of 1 bytes CANNOT be represented 
    !          in a standard 1 byte fortran integer variable,
    !          therefore an 2 byte integer is used as input
    integer,      intent(in)  :: fileunit
    integer(i2_), intent(in)  :: value
    integer,      intent(out) :: error_flag

    ! local variables
    integer            :: i, j, filepos
    character(len=256) :: filename
    logical            :: nmd
    character(len=2)   :: byte_array
    character(len=1)   :: tmp_byte

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in write_uns_int1_to_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! check if we are inside the valid range
    IF ((value .lt. 0) .or. (value .gt. 255)) THEN
       print *,"ERROR: this value of: ",value," cannot be written"
       print *,"as a 1 byte unsigned integer, since it is outside the"
       print *,"valid range for this datatype"
       error_flag = error_writing_file
       return
    END IF

    ! convert the data to a sequence of bytes stored in a string
    byte_array(1:2) = transfer(value,byte_array(1:2))

    IF (do_byteswap) THEN
       ! convert Little<=>Big endian
       DO i=1,int(nbytes_i2_/2)
          j=nbytes_i2_+1-i
          ! swap bytes i and j
          tmp_byte = byte_array(i:i)
          byte_array(i:i) = byte_array(j:j)
          byte_array(j:j) = tmp_byte
       END DO
    END IF

    ! then write only the 2nd byte, which should contain the number
    ! that we are interested in to the output file
    inquire(unit=fileunit,nextrec=filepos)
    write(fileunit,rec=filepos,ERR=999) byte_array(2:2)

    return

999 filename = "[undefined]"
    inquire(unit=fileunit,named=nmd)
    if (nmd) inquire(unit=fileunit,name=filename)
    print *,"write_uns_int1_to_bin_file: ERROR could not write to outputfile: ",&
            trim(filename)
    error_flag = error_writing_file
    return

  end subroutine write_uns_int1_to_bin_file
    !  #]
  subroutine write_uns_int2_to_unbuffbinfile(fileunit,value,error_flag)
    !  #[ write a 2 byte unsigned integer
    ! P. Poli   10-01-2006  Original Code.
    !                       Derived from JdK's write_uns_int1_to_file
    ! WARNING: an unsigned integer of 2 bytes CANNOT be represented 
    !          in a standard 2 byte fortran integer variable,
    !          therefore an 4 byte integer is used as input
    integer,      intent(in)  :: fileunit
    integer(i4_), intent(in)  :: value
    integer,      intent(out) :: error_flag

    ! local variables
    integer            :: i, j, filepos
    character(len=256) :: filename
    logical            :: nmd
    character(len=8)   :: byte_array
    character(len=1)   :: tmp_byte

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in write_uns_int2_to_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! check if we are inside the valid range
    IF ((value .lt. 0) .or. (value .gt. 256*256-1)) THEN
       print *,"ERROR: this value of: ",value," cannot be written"
       print *,"as a 2 byte unsigned integer, since it is outside the"
       print *,"valid range for this datatype"
       error_flag = error_writing_file
       return
    END IF   
    
    ! convert the data to a sequence of bytes stored in a string
    byte_array(1:4) = transfer(value,byte_array(1:4))

    IF (do_byteswap) THEN
       ! convert Little<=>Big endian
       DO i=1,int(nbytes_i4_/2)
          j=nbytes_i4_+1-i
          ! swap bytes i and j
          tmp_byte = byte_array(i:i)
          byte_array(i:i) = byte_array(j:j)
          byte_array(j:j) = tmp_byte
       END DO
    END IF

    ! then take only the last 2 bytes, which should contain the number
    ! that we are interested in
    byte_array(1:2) = byte_array(3:4)

    ! then write these 2 bytes to the output file
    inquire(unit=fileunit,nextrec=filepos)
    DO i=1,nbytes_i2_
       write(fileunit,rec=filepos+i-1,ERR=999) byte_array(i:i)
!print "('writing byte ',i1,':',i3)",i,ichar(byte_array(i:i))
    END DO

    return

999 filename = "[undefined]"
    inquire(unit=fileunit,named=nmd)
    if (nmd) inquire(unit=fileunit,name=filename)
    print *,"write_uns_int2_to_bin_file: ERROR could not write to outputfile: ",&
            trim(filename)
    error_flag = error_writing_file
    return

  end subroutine write_uns_int2_to_unbuffbinfile
    !  #]
  subroutine write_uns_int4_to_unbuffbinfile(fileunit,value,error_flag)
    !  #[ write a 4 byte unsigned integer
    ! P. Poli   10-01-2006  Original Code.
    !                       Derived from JdK's write_uns_int1_to_file
    ! WARNING: an unsigned integer of 4 bytes CANNOT be represented 
    !          in a standard 4 byte fortran integer variable,
    !          therefore an 8 byte integer is used as input
    integer,      intent(in)  :: fileunit
    integer(i8_), intent(in)  :: value
    integer,      intent(out) :: error_flag

    ! local variables
    integer            :: i, j, filepos
    character(len=256) :: filename
    logical            :: nmd
    character(len=8)   :: byte_array
    character(len=1)   :: tmp_byte

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in write_uns_int4_to_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! check if we are inside the valid range
    IF ((value .lt. 0) .or. (value .gt. 256_i8_*256*256*256-1)) THEN
       print *,"ERROR: this value of: ",value," cannot be written"
       print *,"as a 4 byte unsigned integer, since it is outside the"
       print *,"valid range for this datatype"
       error_flag = error_writing_file
       return
    END IF   
    
    ! convert the data to a sequence of bytes stored in a string
    byte_array(1:8) = transfer(value,byte_array(1:8))

    IF (do_byteswap) THEN
       ! convert Little<=>Big endian
       DO i=1,int(nbytes_i8_/2)
          j=nbytes_i8_+1-i
          ! swap bytes i and j
          tmp_byte = byte_array(i:i)
          byte_array(i:i) = byte_array(j:j)
          byte_array(j:j) = tmp_byte
       END DO
    END IF

    ! then take only the last 4 bytes, which should contain the number
    ! that we are interested in
    byte_array(1:4) = byte_array(5:8)

    ! then write these 4 bytes to the output file
    inquire(unit=fileunit,nextrec=filepos)
    DO i=1,nbytes_i4_
       write(fileunit,rec=filepos+i-1,ERR=999) byte_array(i:i)
!print "('writing byte ',i1,':',i3)",i,ichar(byte_array(i:i))
    END DO

    return

999 filename = "[undefined]"
    inquire(unit=fileunit,named=nmd)
    if (nmd) inquire(unit=fileunit,name=filename)
    print *,"write_uns_int4_to_bin_file: ERROR could not write to outputfile: ",&
            trim(filename)
    error_flag = error_writing_file
    return

  end subroutine write_uns_int4_to_unbuffbinfile
    !  #]

  subroutine write_int4array_to_bin_file(fileunit,values,error_flag,nvalues)
    !  #[

    integer,                    intent(in)  :: fileunit
    integer(i4_), dimension(:), intent(in)  :: values
    integer,                    intent(out) :: error_flag
    integer,      optional,     intent(in)  :: nvalues

    ! local variables
    integer :: i, nr_values_to_write

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in write_int4array_to_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    nr_values_to_write = size(values)
    IF (present(nvalues)) nr_values_to_write = nvalues
    IF (nr_values_to_write .lt. 1) THEN
       print *,"ERROR in write_int4array_to_bin_file:"
       print *,"invalid nr of values to be written"
       print *,"nr_values_to_write = ",nr_values_to_write
       error_flag = error_programming
       return
    END IF
    IF (nr_values_to_write .gt. size(values)) THEN
       print *,"ERROR in write_int4array_to_bin_file:"
       print *,"input array too small for nr of values to be written"
       print *,"nr_values_to_write   = ",nr_values_to_write
       print *,"size of values array = ",size(values)
       error_flag = error_programming
       return
    END IF

    array_loop: DO i=1,nr_values_to_write
       call write_int4_to_unbuffbin_file(fileunit,values(i),error_flag)
       IF (error_flag .ne. no_error) exit array_loop
    END DO array_loop

    return

  end subroutine write_int4array_to_bin_file

    !  #]

  subroutine write_uns_int1array_to_bin_file(fileunit,values,error_flag,nvalues)
    !  #[
    integer,                    intent(in)  :: fileunit
    integer(i2_), dimension(:), intent(in)  :: values
    integer,                    intent(out) :: error_flag
    integer,      optional,     intent(in)  :: nvalues

    ! local variables
    integer :: i, nr_values_to_write

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in write_uns_int1array_to_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    nr_values_to_write = size(values)
    IF (present(nvalues)) nr_values_to_write = nvalues
    IF (nr_values_to_write .lt. 1) THEN
       print *,"ERROR in write_uns_int1array_to_bin_file:"
       print *,"invalid nr of values to be written"
       print *,"nr_values_to_write = ",nr_values_to_write
       error_flag = error_programming
       return
    END IF
    IF (nr_values_to_write .gt. size(values)) THEN
       print *,"ERROR in write_uns_int1array_to_bin_file:"
       print *,"input array too small for nr of values to be written"
       print *,"nr_values_to_write   = ",nr_values_to_write
       print *,"size of values array = ",size(values)
       error_flag = error_programming
       return
    END IF

    array_loop: DO i=1,nr_values_to_write
       call write_uns_int1_to_bin_file(fileunit,values(i),error_flag)
       IF (error_flag .ne. no_error) exit array_loop
    END DO array_loop

    return

  end subroutine write_uns_int1array_to_bin_file
    !  #]

  subroutine write_real4_to_bin_file(fileunit,value,error_flag)
    !  #[ 
    integer,   intent(in)  :: fileunit
    real(r4_), intent(in)  :: value
    integer,   intent(out) :: error_flag

    ! local variables
    integer            :: i, j, filepos
    character(len=256) :: filename
    character(len=16)  :: byte_array
    character(len=1)   :: tmp_byte
    logical            :: nmd

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in write_real4_to_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! convert the data to a sequence of bytes stored in a string
    ! (this only swaps the data-type and copies no data ! )
    byte_array(1:nbytes_r4_) = transfer(value,byte_array(1:nbytes_r4_))

    IF (do_byteswap) THEN
       ! convert Little<=>Big endian
       DO i=1,int(nbytes_r4_/2)
          j=nbytes_r4_+1-i
          ! swap bytes i and j
          tmp_byte = byte_array(i:i)
          byte_array(i:i) = byte_array(j:j)
          byte_array(j:j) = tmp_byte
       END DO
    END IF

    ! write the bytes to the output file
    inquire(unit=fileunit,nextrec=filepos)
    DO i=1,nbytes_r4_
       write(fileunit,rec=filepos+i-1,ERR=999) byte_array(i:i)
    END DO

    return

999 filename = "[undefined]"
    inquire(unit=fileunit,named=nmd)
    if (nmd) inquire(unit=fileunit,name=filename)
    print *,"write_real4_to_bin_file: ERROR could not write to outputfile: ", &
            trim(filename)
    error_flag = error_writing_file
    return

  end subroutine write_real4_to_bin_file
    !  #]
  subroutine write_real8_to_bin_file(fileunit,value,error_flag)
    !  #[ 
    !  10-01-2006   P. Poli    Original code. Derived from JdK's
    !  write_real8_to_bin_file
    integer,          intent(in)  :: fileunit
    real(r8_),        intent(in)  :: value
    integer,          intent(out) :: error_flag

    ! local variables
    integer            :: i, j, filepos
    character(len=256) :: filename
    character(len=16)  :: byte_array
    character(len=1)   :: tmp_byte
    logical            :: nmd

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in write_real8_to_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! convert the data to a sequence of bytes stored in a string
    ! (this only swaps the data-type and copies no data ! )
    byte_array(1:nbytes_r8_) = transfer(value,byte_array(1:nbytes_r8_))

    IF (do_byteswap) THEN
       ! convert Little<=>Big endian
       DO i=1,int(nbytes_r8_/2)
          j=nbytes_r8_+1-i
          ! swap bytes i and j
          tmp_byte = byte_array(i:i)
          byte_array(i:i) = byte_array(j:j)
          byte_array(j:j) = tmp_byte
       END DO
    END IF

    ! write the bytes to the output file
    inquire(unit=fileunit,nextrec=filepos)
    DO i=1,nbytes_r8_
       write(fileunit,rec=filepos+i-1,ERR=999) byte_array(i:i)
    END DO

    return

999 filename = "[undefined]"
    inquire(unit=fileunit,named=nmd)
    if (nmd) inquire(unit=fileunit,name=filename)
    print *,"write_real8_to_bin_file: ERROR could not write to outputfile: ", &
            trim(filename)
    error_flag = error_writing_file
    return

  end subroutine write_real8_to_bin_file
    !  #]
  subroutine write_real4array_to_bin_file(fileunit,values,error_flag)
    !  #[
    integer,                 intent(in)  :: fileunit
    real(r4_), dimension(:), intent(in)  :: values
    integer,                 intent(out) :: error_flag

    ! local variables
    integer :: i

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in write_real4array_to_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    array_loop: DO i=1,size(values)
       call write_real4_to_bin_file(fileunit,values(i),error_flag)
       IF (error_flag .ne. no_error) exit array_loop
    END DO array_loop

    return

  end subroutine write_real4array_to_bin_file
    !  #]
  subroutine write_real8array_to_bin_file(fileunit,values,error_flag)
    !  #[
    integer,                 intent(in)  :: fileunit
    real(r8_), dimension(:), intent(in)  :: values
    integer,                 intent(out) :: error_flag

    ! local variables
    integer :: i

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in write_real8array_to_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    array_loop: DO i=1,size(values)
       call write_real8_to_bin_file(fileunit,values(i),error_flag)
       IF (error_flag .ne. no_error) exit array_loop
    END DO array_loop

    return

  end subroutine write_real8array_to_bin_file
    !  #]

  subroutine write_zero_bytes_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    !  #[ 
    integer,      intent(in)  :: fileunit
    integer,      intent(in)  :: nr_of_zero_bytes
    integer,      intent(out) :: error_flag

    ! local variables
    integer            :: i, filepos
    character(len=256) :: filename
    logical            :: nmd
    integer(i1_), parameter :: zero_byte = 0_i1_

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in write_zero_bytes_bin_file()"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! write the bytes to the output file
    inquire(unit=fileunit,nextrec=filepos)
    DO i=1,nr_of_zero_bytes
       write(fileunit,rec=filepos+i-1,ERR=999) zero_byte
    END DO

    return

999 filename = "[undefined]"
    inquire(unit=fileunit,named=nmd)
    if (nmd) inquire(unit=fileunit,name=filename)
    print *,"write_zero_bytes_bin_file: ERROR could not write to outputfile: ",&
             trim(filename)
    error_flag = error_writing_file
    return

  end subroutine write_zero_bytes_bin_file

    !  #]

  !----------------------------------------------------
  ! reading routines
  !----------------------------------------------------
  subroutine read_int1_from_bin_file(fileunit,value,error_flag)
    !  #[ read a 1 byte signed integer
    integer,      intent(in)  :: fileunit
    integer(i1_), intent(out) :: value
    integer,      intent(out) :: error_flag

    ! local variables
    integer            :: filepos, ios
    character(len=256) :: filename
    character(len=1)   :: tmp_byte
    logical            :: nmd

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_int1_from_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! read the byte from the input file
    inquire(unit=fileunit,nextrec=filepos)
    read(fileunit,rec=filepos,iostat=ios) tmp_byte
    IF (ios .ne. 0) goto 999

    ! convert the byte to the numerical value
    ! (this only swaps the data-type and copies no data ! )
    value = transfer(tmp_byte,value)

    !print *,"read: value = ",value

    return

999 filename = "[undefined]"
    inquire(unit=fileunit,named=nmd)
    if (nmd) inquire(unit=fileunit,name=filename)
    print *,"ERROR could not read from inputfile: ",trim(filename)
    error_flag = error_reading_file
    return

  end subroutine read_int1_from_bin_file
    !  #]
  subroutine read_int2_from_bin_file(fileunit,value,error_flag)
    !  #[ read a 2 byte signed integer
    integer,      intent(in)  :: fileunit
    integer(i2_), intent(out) :: value
    integer,      intent(out) :: error_flag

    ! local variables
    integer            :: i, j, filepos, ios
    character(len=256) :: filename
    character(len=16)  :: byte_array
    character(len=1)   :: tmp_byte
    logical            :: nmd

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_int2_from_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! read the bytes from the input file
    inquire(unit=fileunit,nextrec=filepos)
    DO i=1,nbytes_i2_
       read(fileunit,rec=filepos+i-1,iostat=ios) byte_array(i:i)
       IF (ios .ne. 0) goto 999
    END DO

    IF (do_byteswap) THEN
       ! convert Little<=>Big endian
       DO i=1,int(nbytes_i2_/2)
          j=nbytes_i2_+1-i
          ! swap bytes i and j
          tmp_byte = byte_array(i:i)
          byte_array(i:i) = byte_array(j:j)
          byte_array(j:j) = tmp_byte
       END DO
    END IF

    ! convert the sequence of bytes stored in a string to 
    ! a numerical value
    ! (this only swaps the data-type and copies no data ! )
    value = transfer(byte_array(1:nbytes_i2_),value)

    !print *,"read: value = ",value

    return

999 filename = "[undefined]"
    inquire(unit=fileunit,named=nmd)
    if (nmd) inquire(unit=fileunit,name=filename)
    print *,"ERROR could not read from inputfile: ",trim(filename)
    error_flag = error_reading_file
    return

  end subroutine read_int2_from_bin_file
    !  #]
  subroutine read_int4_from_bin_file(fileunit,value,error_flag)
    !  #[ read a 4 byte signed integer
    integer,      intent(in)  :: fileunit
    integer(i4_), intent(out) :: value
    integer,      intent(out) :: error_flag

    ! local variables
    integer            :: i, j, filepos, ios
    character(len=256) :: filename
    character(len=16)  :: byte_array
    character(len=1)   :: tmp_byte
    logical            :: nmd

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_int4_from_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! read the bytes from the input file
    inquire(unit=fileunit,nextrec=filepos)
    DO i=1,nbytes_i4_
       read(fileunit,rec=filepos+i-1,iostat=ios) byte_array(i:i)
       IF (ios .ne. 0) goto 999
    END DO

    IF (do_byteswap) THEN
       ! convert Little<=>Big endian
       DO i=1,int(nbytes_i4_/2)
          j=nbytes_i4_+1-i
          ! swap bytes i and j
          tmp_byte = byte_array(i:i)
          byte_array(i:i) = byte_array(j:j)
          byte_array(j:j) = tmp_byte
       END DO
    END IF

    !print *,"testjos: inside read_int4_from_bin_file():"
    !print "('byte_array(1:4) = ',4(i3.3,1x),' filepos = ',i8)",&
    !     ichar(byte_array(1:1)),ichar(byte_array(2:2)),&
    !     ichar(byte_array(3:3)),ichar(byte_array(4:4)),filepos

    ! convert the sequence of bytes stored in a string to 
    ! a numerical value
    ! (this only swaps the data-type and copies no data ! )
    value = transfer(byte_array(1:nbytes_i4_),value)

    !print *,"read: value = ",value

    return

999 filename = "[undefined]"
    inquire(unit=fileunit,named=nmd)
    if (nmd) inquire(unit=fileunit,name=filename)
    print *,"ERROR could not read from inputfile: ",trim(filename)
    error_flag = error_reading_file
    return

  end subroutine read_int4_from_bin_file
    !  #]
  subroutine read_int8_from_bin_file(fileunit,value,error_flag)
    !  #[ read a 8 byte signed integer
    integer,      intent(in)  :: fileunit
    integer(i8_), intent(out) :: value
    integer,      intent(out) :: error_flag

    ! local variables
    integer            :: i, j, filepos, ios
    character(len=256) :: filename
    character(len=16)  :: byte_array
    character(len=1)   :: tmp_byte
    logical            :: nmd

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_int4_from_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! read the bytes from the input file
    inquire(unit=fileunit,nextrec=filepos)
    DO i=1,nbytes_i8_
       read(fileunit,rec=filepos+i-1,iostat=ios) byte_array(i:i)
       IF (ios .ne. 0) goto 999
    END DO

    IF (do_byteswap) THEN
       ! convert Little<=>Big endian
       DO i=1,int(nbytes_i8_/2)
          j=nbytes_i8_+1-i
          ! swap bytes i and j
          tmp_byte = byte_array(i:i)
          byte_array(i:i) = byte_array(j:j)
          byte_array(j:j) = tmp_byte
       END DO
    END IF

    ! convert the sequence of bytes stored in a string to 
    ! a numerical value
    ! (this only swaps the data-type and copies no data ! )
    value = transfer(byte_array(1:nbytes_i8_),value)

    !print *,"read: value = ",value

    return

999 filename = "[undefined]"
    inquire(unit=fileunit,named=nmd)
    if (nmd) inquire(unit=fileunit,name=filename)
    print *,"ERROR could not read from inputfile: ",trim(filename)
    error_flag = error_reading_file
    return

  end subroutine read_int8_from_bin_file
    !  #]

  subroutine read_uns_int1_from_bin_file(fileunit,value,error_flag)
    !  #[ read a 1 byte unsigned integer

    ! WARNING: an unsigned integer of 1 bytes CANNOT be represented 
    !          in a standard 1 byte fortran integer variable,
    !          therefore an 2 byte integer is used for the result
    integer,      intent(in)  :: fileunit
    integer(i2_), intent(out) :: value
    integer,      intent(out) :: error_flag

    ! local variables
    integer(i1_)            :: tmp_value

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_uns_int1_from_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! read the integer as if it where a signed integer
    call read_int1_from_bin_file(fileunit,tmp_value,error_flag)

    ! now remove the sign:
    IF (tmp_value .ge. 0) THEN
       value = int(tmp_value,i2_)
    ELSE
       value = int1_to_unsigned(tmp_value)
    END IF

    return

  end subroutine read_uns_int1_from_bin_file

    !  #]
  subroutine read_uns_int2_from_bin_file(fileunit,value,error_flag)
    !  #[ read a 2 byte unsigned integer
    ! WARNING: an unsigned integer of 2 bytes CANNOT be represented 
    !          in a standard 2 byte fortran integer variable,
    !          therefore an 4 byte integer is used for the result
    integer,      intent(in)  :: fileunit
    integer(i4_), intent(out) :: value
    integer,      intent(out) :: error_flag

    ! local variables
    integer(i2_)       :: tmp_value

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_uns_int2_from_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! read the integer as if it where a signed integer
    call read_int2_from_bin_file(fileunit,tmp_value,error_flag)

    ! now remove the sign:
    IF (tmp_value .ge. 0) THEN
       value = int(tmp_value,i4_)
    ELSE
       value = int2_to_unsigned(tmp_value)
    END IF

    return

  end subroutine read_uns_int2_from_bin_file
    !  #]
  subroutine read_uns_int4_from_bin_file(fileunit,value,error_flag)
    !  #[ read a 4 byte unsigned integer
    ! WARNING: an unsigned integer of 4 bytes CANNOT be represented 
    !          in a standard 4 byte fortran integer variable,
    !          therefore an 8 byte integer is used for the result
    integer,      intent(in)  :: fileunit
    integer(i8_), intent(out) :: value
    integer,      intent(out) :: error_flag

    ! local variables
    integer(i4_)       :: tmp_value

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_uns_int4_from_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! read the integer as if it where a signed integer
    call read_int4_from_bin_file(fileunit,tmp_value,error_flag)

    ! now remove the sign:
    IF (tmp_value .ge. 0) THEN
       value = int(tmp_value,i8_)
    ELSE
       value = int4_to_unsigned(tmp_value)
    END IF

    return

  end subroutine read_uns_int4_from_bin_file
    !  #]

  ! 8 byte unsigned integers cannot be used, because we would need 
  ! a 16 byte signed integer to store it, and this type is not
  ! generally available for Fortran90 compilers.

  subroutine read_int1array_from_bin_file(fileunit,values,error_flag,nvalues)
    !  #[ read an array of 1 byte signed integers
    integer,                    intent(in)  :: fileunit
    integer(i1_), dimension(:), intent(out) :: values
    integer,                    intent(out) :: error_flag
    integer,     optional,      intent(in)  :: nvalues

    ! local variables
    integer :: i, nr_values_to_read

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_int1array_from_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    nr_values_to_read = size(values)
    IF (present(nvalues)) nr_values_to_read = nvalues
    IF (nr_values_to_read .lt. 1) THEN
       print *,"ERROR in read_int1array_from_bin_file:"
       print *,"invalid nr of values to be read"
       print *,"nr_values_to_read = ",nr_values_to_read
       error_flag = error_programming
       return
    END IF
    IF (nr_values_to_read .gt. size(values)) THEN
       print *,"ERROR in read_int1array_from_bin_file:"
       print *,"output array too small for nr of values to be read"
       print *,"nr_values_to_read    = ",nr_values_to_read
       print *,"size of values array = ",size(values)
       error_flag = error_programming
       return
    END IF

    array_loop: DO i=1,nr_values_to_read
       call read_int1_from_bin_file(fileunit,values(i),error_flag)
       IF (error_flag .ne. no_error) exit array_loop
    END DO array_loop

    return

  end subroutine read_int1array_from_bin_file
    !  #]
  subroutine read_int2array_from_bin_file(fileunit,values,error_flag,nvalues)
    !  #[ read an array of 2 byte signed integers
    integer,                    intent(in)  :: fileunit
    integer(i2_), dimension(:), intent(out) :: values
    integer,                    intent(out) :: error_flag
    integer,     optional,      intent(in)  :: nvalues

    ! local variables
    integer :: i, nr_values_to_read

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_int2array_from_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    nr_values_to_read = size(values)
    IF (present(nvalues)) nr_values_to_read = nvalues
    IF (nr_values_to_read .lt. 1) THEN
       print *,"ERROR in read_int2array_from_bin_file:"
       print *,"invalid nr of values to be read"
       print *,"nr_values_to_read = ",nr_values_to_read
       error_flag = error_programming
       return
    END IF
    IF (nr_values_to_read .gt. size(values)) THEN
       print *,"ERROR in read_int2array_from_bin_file:"
       print *,"output array too small for nr of values to be read"
       print *,"nr_values_to_read    = ",nr_values_to_read
       print *,"size of values array = ",size(values)
       error_flag = error_programming
       return
    END IF

    array_loop: DO i=1,nr_values_to_read
       call read_int2_from_bin_file(fileunit,values(i),error_flag)
       IF (error_flag .ne. no_error) exit array_loop
    END DO array_loop

    return

  end subroutine read_int2array_from_bin_file
    !  #]
  subroutine read_int4array_from_bin_file(fileunit,values,error_flag,nvalues)
    !  #[ read an array of 4 byte signed integers
    integer,                    intent(in)  :: fileunit
    integer(i4_), dimension(:), intent(out) :: values
    integer,                    intent(out) :: error_flag
    integer,     optional,      intent(in)  :: nvalues

    ! local variables
    integer :: i, nr_values_to_read

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_int4array_from_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    nr_values_to_read = size(values)
    IF (present(nvalues)) nr_values_to_read = nvalues
    IF (nr_values_to_read .lt. 1) THEN
       print *,"ERROR in read_int4array_from_bin_file:"
       print *,"invalid nr of values to be read"
       print *,"nr_values_to_read = ",nr_values_to_read
       error_flag = error_programming
       return
    END IF
    IF (nr_values_to_read .gt. size(values)) THEN
       print *,"ERROR in read_int4array_from_bin_file:"
       print *,"output array too small for nr of values to be read"
       print *,"nr_values_to_read    = ",nr_values_to_read
       print *,"size of values array = ",size(values)
       error_flag = error_programming
       return
    END IF

    array_loop: DO i=1,nr_values_to_read
       call read_int4_from_bin_file(fileunit,values(i),error_flag)
       IF (error_flag .ne. no_error) exit array_loop
    END DO array_loop

    return

  end subroutine read_int4array_from_bin_file
    !  #]
  subroutine read_int8array_from_bin_file(fileunit,values,error_flag,nvalues)
    !  #[ read an array of 8 byte signed integers
    integer,                    intent(in)  :: fileunit
    integer(i8_), dimension(:), intent(out) :: values
    integer,                    intent(out) :: error_flag
    integer,     optional,      intent(in)  :: nvalues

    ! local variables
    integer :: i, nr_values_to_read

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_int8array_from_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    nr_values_to_read = size(values)
    IF (present(nvalues)) nr_values_to_read = nvalues
    IF (nr_values_to_read .lt. 1) THEN
       print *,"ERROR in read_int8array_from_bin_file:"
       print *,"invalid nr of values to be read"
       print *,"nr_values_to_read = ",nr_values_to_read
       error_flag = error_programming
       return
    END IF
    IF (nr_values_to_read .gt. size(values)) THEN
       print *,"ERROR in read_int8array_from_bin_file:"
       print *,"output array too small for nr of values to be read"
       print *,"nr_values_to_read    = ",nr_values_to_read
       print *,"size of values array = ",size(values)
       error_flag = error_programming
       return
    END IF

    array_loop: DO i=1,nr_values_to_read
       call read_int8_from_bin_file(fileunit,values(i),error_flag)
       IF (error_flag .ne. no_error) exit array_loop
    END DO array_loop

    return

  end subroutine read_int8array_from_bin_file
    !  #]

  subroutine read_uns_int1arr_from_bin_file(fileunit,values,error_flag,nvalues)
    !  #[ read an array of 1 byte unsigned integers
    integer,                    intent(in)  :: fileunit
    integer(i2_), dimension(:), intent(out) :: values
    integer,                    intent(out) :: error_flag
    integer,     optional,      intent(in)  :: nvalues

    ! local variables
    integer :: i, nr_values_to_read

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_uns_int1arr_from_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    nr_values_to_read = size(values)
    IF (present(nvalues)) nr_values_to_read = nvalues
    IF (nr_values_to_read .lt. 1) THEN
       print *,"ERROR in read_uns_int1arr_from_bin_file:"
       print *,"invalid nr of values to be read"
       print *,"nr_values_to_read = ",nr_values_to_read
       error_flag = error_programming
       return
    END IF
    IF (nr_values_to_read .gt. size(values)) THEN
       print *,"ERROR in read_uns_int1arr_from_bin_file:"
       print *,"output array too small for nr of values to be read"
       print *,"nr_values_to_read    = ",nr_values_to_read
       print *,"size of values array = ",size(values)
       error_flag = error_programming
       return
    END IF

    array_loop: DO i=1,nr_values_to_read
       call read_uns_int1_from_bin_file(fileunit,values(i),error_flag)
       IF (error_flag .ne. no_error) exit array_loop
    END DO array_loop

    return

  end subroutine read_uns_int1arr_from_bin_file
    !  #]
  subroutine read_uns_int2arr_from_bin_file(fileunit,values,error_flag,nvalues)
    !  #[ read an array of 2 byte unsigned integers
    integer,                    intent(in)  :: fileunit
    integer(i4_), dimension(:), intent(out) :: values
    integer,                    intent(out) :: error_flag
    integer,     optional,      intent(in)  :: nvalues

    ! local variables
    integer :: i, nr_values_to_read

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_uns_int2arr_from_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    nr_values_to_read = size(values)
    IF (present(nvalues)) nr_values_to_read = nvalues
    IF (nr_values_to_read .lt. 1) THEN
       print *,"ERROR in read_uns_int2arr_from_bin_file:"
       print *,"invalid nr of values to be read"
       print *,"nr_values_to_read = ",nr_values_to_read
       error_flag = error_programming
       return
    END IF
    IF (nr_values_to_read .gt. size(values)) THEN
       print *,"ERROR in read_uns_int2arr_from_bin_file:"
       print *,"output array too small for nr of values to be read"
       print *,"nr_values_to_read    = ",nr_values_to_read
       print *,"size of values array = ",size(values)
       error_flag = error_programming
       return
    END IF

    array_loop: DO i=1,nr_values_to_read
       call read_uns_int2_from_bin_file(fileunit,values(i),error_flag)
       IF (error_flag .ne. no_error) exit array_loop
    END DO array_loop

    return

  end subroutine read_uns_int2arr_from_bin_file
    !  #]
  subroutine read_uns_int4arr_from_bin_file(fileunit,values,error_flag,nvalues)
    !  #[ read an array of 4 byte unsigned integers
    integer,                    intent(in)  :: fileunit
    integer(i8_), dimension(:), intent(out) :: values
    integer,                    intent(out) :: error_flag
    integer,     optional,      intent(in)  :: nvalues

    ! local variables
    integer :: i, nr_values_to_read

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_uns_int4arr_from_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    nr_values_to_read = size(values)
    IF (present(nvalues)) nr_values_to_read = nvalues
    IF (nr_values_to_read .lt. 1) THEN
       print *,"ERROR in read_uns_int4arr_from_bin_file:"
       print *,"invalid nr of values to be read"
       print *,"nr_values_to_read = ",nr_values_to_read
       error_flag = error_programming
       return
    END IF
    IF (nr_values_to_read .gt. size(values)) THEN
       print *,"ERROR in read_uns_int4arr_from_bin_file:"
       print *,"output array too small for nr of values to be read"
       print *,"nr_values_to_read    = ",nr_values_to_read
       print *,"size of values array = ",size(values)
       error_flag = error_programming
       return
    END IF

    array_loop: DO i=1,nr_values_to_read
       call read_uns_int4_from_bin_file(fileunit,values(i),error_flag)
       IF (error_flag .ne. no_error) exit array_loop
    END DO array_loop

    return

  end subroutine read_uns_int4arr_from_bin_file
    !  #]

  ! 8 byte unsigned integers cannot be used, because we would need 
  ! a 16 byte signed integer to store it, and this type is not
  ! generally available for Fortran90 compilers.

  subroutine read_real4_from_bin_file(fileunit,value,error_flag)
    !  #[ read a 4 byte real
    integer,     intent(in)  :: fileunit
    real(r4_),   intent(out) :: value
    integer,     intent(out) :: error_flag

    ! local variables
    integer            :: i, j, filepos, ios
    character(len=256) :: filename
    character(len=16)  :: byte_array
    character(len=1)   :: tmp_byte
    logical            :: nmd

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_real4_from_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! read the bytes from the input file
    inquire(unit=fileunit,nextrec=filepos)
    DO i=1,nbytes_r4_
       read(fileunit,rec=filepos+i-1,iostat=ios) byte_array(i:i)
       IF (ios .ne. 0) goto 999
    END DO

    IF (do_byteswap) THEN
       ! convert Little<=>Big endian
       DO i=1,int(nbytes_r4_/2)
          j=nbytes_r4_+1-i
          ! swap bytes i and j
          tmp_byte = byte_array(i:i)
          byte_array(i:i) = byte_array(j:j)
          byte_array(j:j) = tmp_byte
       END DO
    END IF

    ! convert the sequence of bytes stored in a string to 
    ! a numerical value
    ! (this only swaps the data-type and copies no data ! )
    value = transfer(byte_array(1:nbytes_r4_),value)

    !print *,"read: value = ",value

    return

999 filename = "[undefined]"
    inquire(unit=fileunit,named=nmd)
    if (nmd) inquire(unit=fileunit,name=filename)
    print *,"ERROR could not read from inputfile: ",trim(filename)
    error_flag = error_reading_file
    return

  end subroutine read_real4_from_bin_file
    !  #]
  subroutine read_real4array_from_bin_file(fileunit,values,error_flag,nvalues)
    !  #[ read an array of 4 byte reals

    integer,                   intent(in)  :: fileunit
    real(r4_),   dimension(:), intent(out) :: values
    integer,                   intent(out) :: error_flag
    integer,     optional,     intent(in)  :: nvalues

    ! local variables
    integer :: i, nr_values_to_read

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_real4array_from_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    nr_values_to_read = size(values)
    IF (present(nvalues)) nr_values_to_read = nvalues
    IF (nr_values_to_read .lt. 1) THEN
       print *,"ERROR in read_real4array_from_bin_file:"
       print *,"invalid nr of values to be read"
       print *,"nr_values_to_read = ",nr_values_to_read
       error_flag = error_programming
       return
    END IF
    IF (nr_values_to_read .gt. size(values)) THEN
       print *,"ERROR in read_real4array_from_bin_file:"
       print *,"output array too small for nr of values to be read"
       print *,"nr_values_to_read    = ",nr_values_to_read
       print *,"size of values array = ",size(values)
       error_flag = error_programming
       return
    END IF

    array_loop: DO i=1,nr_values_to_read
       call read_real4_from_bin_file(fileunit,values(i),error_flag)
       IF (error_flag .ne. no_error) exit array_loop
    END DO array_loop

    return

  end subroutine read_real4array_from_bin_file
    !  #]
  subroutine read_real8_from_bin_file(fileunit,value,error_flag)
    !  #[ read an 8 byte real
    integer,     intent(in)  :: fileunit
    real(r8_),   intent(out) :: value
    integer,     intent(out) :: error_flag

    ! local variables
    integer            :: i, j, filepos, ios
    character(len=256) :: filename
    character(len=16)  :: byte_array
    character(len=1)   :: tmp_byte
    logical            :: nmd

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_real8_from_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! read the bytes from the input file
    inquire(unit=fileunit,nextrec=filepos)
    DO i=1,nbytes_r8_
       read(fileunit,rec=filepos+i-1,iostat=ios) byte_array(i:i)
       IF (ios .ne. 0) goto 999
    END DO

    IF (do_byteswap) THEN
       ! convert Little<=>Big endian
       DO i=1,int(nbytes_r8_/2)
          j=nbytes_r8_+1-i
          ! swap bytes i and j
          tmp_byte = byte_array(i:i)
          byte_array(i:i) = byte_array(j:j)
          byte_array(j:j) = tmp_byte
       END DO
    END IF

    !print *,"byte values: ",(ichar(byte_array(i:i)),i=1,8)

    ! convert the sequence of bytes stored in a string to 
    ! a numerical value
    ! (this only swaps the data-type and copies no data ! )
    value = transfer(byte_array(1:nbytes_r8_),value)

    !print *,"read: value = ",value

    return

999 filename = "[undefined]"
    inquire(unit=fileunit,named=nmd)
    if (nmd) inquire(unit=fileunit,name=filename)
    print *,"ERROR could not read from inputfile: ",trim(filename)
    error_flag = error_reading_file
    return

  end subroutine read_real8_from_bin_file
    !  #]
  subroutine read_real8array_from_bin_file(fileunit,values,error_flag,nvalues)
    !  #[ read an array of 8 byte reals

    integer,                   intent(in)  :: fileunit
    real(r8_),   dimension(:), intent(out) :: values
    integer,                   intent(out) :: error_flag
    integer,     optional,     intent(in)  :: nvalues

    ! local variables
    integer :: i, nr_values_to_read

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_real8array_from_bin_file:"
       print *,"the ReadDirectBinary module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    nr_values_to_read = size(values)
    IF (present(nvalues)) nr_values_to_read = nvalues
    IF (nr_values_to_read .lt. 1) THEN
       print *,"ERROR in read_real8array_from_bin_file:"
       print *,"invalid nr of values to be read"
       print *,"nr_values_to_read = ",nr_values_to_read
       error_flag = error_programming
       return
    END IF
    IF (nr_values_to_read .gt. size(values)) THEN
       print *,"ERROR in read_real8array_from_bin_file:"
       print *,"output array too small for nr of values to be read"
       print *,"nr_values_to_read    = ",nr_values_to_read
       print *,"size of values array = ",size(values)
       error_flag = error_programming
       return
    END IF

    array_loop: DO i=1,nr_values_to_read
       call read_real8_from_bin_file(fileunit,values(i),error_flag)
       IF (error_flag .ne. no_error) exit array_loop
    END DO array_loop

    return

  end subroutine read_real8array_from_bin_file
    !  #]

  subroutine read_logical1_from_bin_file(fileunit,value,error_flag, &
                                         workaround_in,description)
    ! Poli: added an optional argument flag_nocheck to bypass
    ! errors when reading a value that is not a logical
    ! JK: renamed flag_nocheck to a temporary optional input named 
    !     workaround_in.
    !     This enables printing a warning, in stead of the ERROR that was
    !     printed before (when we did not use the flag_nocheck input, but
    !     simply ignored the error_flag when a non-logical was detected
    !     by this routine).
    !  #[  read a one byte logical
    integer,     intent(in)  :: fileunit
    logical,     intent(out) :: value
    integer,     intent(out) :: error_flag
    logical,          optional, intent(in) :: workaround_in
    character(len=*), optional, intent(in) :: description ! for debugging !

    ! local variables
    integer             :: filepos, ios
    character(len=256)  :: filename
    character(len=1024) :: txt
    character(len=1)    :: tmp_byte
    logical             :: nmd, workaround
    character(len=6)    :: value_txt
    character(len=10)   :: filepos_txt
    integer, parameter  :: nbytes_logical = 1

    ! init
    error_flag = no_error
    workaround = .false.
    IF (present(workaround_in)) workaround = workaround_in

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_logical1_from_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! read the bytes from the input file
    inquire(unit=fileunit,nextrec=filepos)
    read(fileunit,rec=filepos,iostat=ios) tmp_byte
    IF (ios .ne. 0) goto 999

    ! convert the byte value to the logical
    value = .false.
    IF (ichar(tmp_byte) .eq. 1) THEN
       value = .true.
       return
    END IF


    IF (ichar(tmp_byte) .ne. 0) THEN
       write(value_txt,"(i6)",err=998) &
            int1_to_unsigned(int(ichar(tmp_byte),i1_))
       txt = " in read_logical1_from_bin_file(): "//&
             "a 1byte logical should have a value of 0 or 1, "//&
             "but the current value is: "//trim(value_txt)
       if (present(description)) then
          txt = trim(txt)//" for element: "//trim(description)
       end if
       write(filepos_txt,"(i10)",err=997) filepos

       txt = trim(txt)//" at byte: "//trim(filepos_txt)

       IF (workaround) THEN
          call logmsg(log_warn,"WARNING "//trim(txt))
       ELSE
          call logmsg(log_error,"ERROR "//trim(txt))
          error_flag = error_reading_file
          return
       END IF
    END IF
    
    !print *,"testjos: ichar(tmp_byte) = ",ichar(tmp_byte)

    return

997 call logmsg(log_error,"ERROR in read_logical1_from_bin_file(): "//&
               "could not convert filepos to a string, probably we "//&
               "have a programming error here ...")
    error_flag = error_programming
    return

998 call logmsg(log_error,"ERROR in read_logical1_from_bin_file(): "//&
               "could not convert tmp_byte to a string, probably we "//&
               "have a programming error here ...")
    error_flag = error_programming
    return

999 filename = "[undefined]"
    inquire(unit=fileunit,named=nmd)
    if (nmd) inquire(unit=fileunit,name=filename)
    print *,"ERROR could not read from inputfile: ",trim(filename)
    error_flag = error_reading_file
    return

  end subroutine read_logical1_from_bin_file
    !  #]
  subroutine write_logical1_to_bin_file(fileunit,value,error_flag)
    !  #[  write a one byte logical
    !  29-11-2005  P. Poli    Original code (inspired from read_logical1...)
    integer, intent(in)  :: fileunit
    logical, intent(in)  :: value
    integer, intent(out) :: error_flag

    ! local variables
    integer            :: filepos
    character(len=256) :: filename
    character(len=1)   :: tmp_byte
    logical            :: nmd

    integer, parameter :: nbytes_logical = 1

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in write_logical1_to_bin_file:"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! convert the logical value to the byte value
    IF (value) THEN
       tmp_byte=char(1)
    ELSE
       tmp_byte=char(0)
    ENDIF

    IF (ichar(tmp_byte) .ne. 0 .and. ichar(tmp_byte) .ne. 1) THEN
       print *,"ERROR in write_logical1_to_bin_file()"
       print *,"a 1byte logical should have a value of 0 or 1"
       print *,"but the current value is: ",&
            int1_to_unsigned(int(ichar(tmp_byte),i1_))
       error_flag = error_programming
       return
    END IF
    
    ! write the byte to the output file
    inquire(unit=fileunit,nextrec=filepos)
    write(fileunit,rec=filepos,ERR=999) tmp_byte

    return

999 filename = "[undefined]"
    inquire(unit=fileunit,named=nmd)
    if (nmd) inquire(unit=fileunit,name=filename)
    print *,"write_logical1_to_bin_file: ERROR could not write to outputfile: ",&
            trim(filename)
    error_flag = error_writing_file
    return

  end subroutine write_logical1_to_bin_file
    !  #]
  subroutine read_zero_bytes_bin_file(fileunit,nr_of_zero_bytes,error_flag)
    !  #[ 
    !  P. Poli       10-01-2006     Original code
    integer,      intent(in)  :: fileunit
    integer,      intent(in)  :: nr_of_zero_bytes
    integer,      intent(out) :: error_flag

    ! local variables
    integer            :: i, filepos, ios
    character(len=256) :: filename
    logical            :: nmd
    integer(i1_)            :: byte_read
    integer(i1_), parameter :: zero_byte = 0_i1_

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in read_zero_bytes_bin_file()"
       print *,"the DirectBinaryIO module must be initialised before use !!!"
       error_flag = error_programming
       return
    END IF

    ! read the bytes from the input file
    inquire(unit=fileunit,nextrec=filepos)
    DO i=1,nr_of_zero_bytes
       read(fileunit,rec=filepos+i-1,iostat=ios) byte_read
       IF (ios .ne. 0) goto 999

       IF (byte_read.ne.zero_byte) THEN
          print *,"read_zero_bytes_bin_file: "//&
               "ERROR found non-zero where zero was expected"
          error_flag = error_reading_file
          return
       ENDIF
    END DO

    return
    
999 filename = "[undefined]"
    inquire(unit=fileunit,named=nmd)
    if (nmd) inquire(unit=fileunit,name=filename)
    print *,"read_zero_bytes_bin_file: ERROR could not read from inputfile: ",&
             trim(filename)
    print *,"problem occurred at byte: i=",i
    print *,"byte_read = ",byte_read
    error_flag = error_reading_file
    return

  end subroutine read_zero_bytes_bin_file

    !  #]
  !----------------------------------------------------
  ! positioning routines
  !----------------------------------------------------
  ! general remark: we use the 8 byte integer type to stay as close as
  ! possible to what is used in ESA's earth explorer DSD format
  ! However, not all compilers accept 8 byte file pointers in the inquire()
  ! function (for example g95), so before use we convert to a 4byte integer.
  subroutine SetFilePointer_unbuff(fileunit,position,error_flag)
    !  #[
    integer,      intent(in)  :: fileunit
    integer(i8_), intent(in)  :: position
    integer,      intent(out) :: error_flag

    ! local variable
    character(len=1)   :: dummybyte
    integer            :: filepos, ios, dummyvalue
    character(len=256) :: filename
    logical            :: nmd

    ! init
    error_flag = no_error

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"trying to set filepointer to: ",position
    END IF

    IF (position .lt. 1) THEN
       print *,"ERROR: illegal value for filepointer: position=",position
       print *,"position should always be >0 !"
       goto 999
    ELSEIF (position .eq. 1) THEN
       ! this seems impossible for direct access files !!!
       ! the pgf90 and ifort compilers just ignore this statement.
       ! the g95 and gfortran compilers give a runtime error
       ! when you try this. 
       ! For EE_CFI files this is not very problematic, since
       ! we never try to re-read the file.
       ! However, for reading unformatted binary fortran files it
       ! would have been nice if this would have worked ...
       !rewind(fileunit)
       print *,"Sorry, don't know how to do this ...."
       goto 999
    ELSE ! position .gt. 1
       ! read a dummy byte to set the filepointer at the right place
       read(fileunit,rec=position-1,iostat=ios) dummybyte
       IF (ios .ne. 0) goto 999
       ! actually read position-1, to make sure the nextrec result will 
       ! point to the desired position when the next read_*
       ! routine is used.
    END IF

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       inquire(unit=fileunit,nextrec=filepos)
       print *,"filepointer set to filepos : ",filepos
       ! this print statement is just to prevent warnings
       ! about setting but not using the dummybyte variable
       dummyvalue = ichar(dummybyte)
       ! this hexvalue of "A9" gets printed as 169 on most machines
       ! but on ecgate, for some strange reason, it is printed as 169-256 = -87
       ! so this next if then will compensate for this, to not break our unit 
       ! tests on the TestUnformattedIO module
       if (dummyvalue .lt. 0) dummyvalue = dummyvalue+256
       print *,"read: dummybyte = ",dummyvalue
    END IF

    return

999 filename = "[undefined]"
    inquire(unit=fileunit,named=nmd)
    if (nmd) inquire(unit=fileunit,name=filename)
    print *,"ERROR could not set filepointer for file: ",trim(filename)
    error_flag = error_reading_file
    return

  end subroutine SetFilePointer_unbuff
    !  #]
  subroutine GetFilePointer_unbuff(fileunit,position,error_flag)
    !  #[
    integer,      intent(in)  :: fileunit
    integer(i8_), intent(out) :: position
    integer,      intent(out) :: error_flag

    ! local variable 
    integer :: position_i_

    error_flag = no_error
    inquire(unit=fileunit,nextrec=position_i_)
    position = position_i_

  end subroutine GetFilePointer_unbuff
    !  #]
  subroutine AdvanceFilePointer_unbuff(fileunit,amount,error_flag)
    !  #[
    integer,      intent(in)  :: fileunit
    integer(i8_), intent(in)  :: amount
    integer,      intent(out) :: error_flag

    ! local variable
    integer :: filepos

    ! init
    error_flag = no_error

    ! request the current position
    inquire(unit=fileunit,nextrec=filepos)

    call SetFilePointer_unbuff(fileunit,int(filepos+amount,i8_),error_flag)

    return

  end subroutine AdvanceFilePointer_unbuff
    !  #]
  !subroutine GetFilePointer_buff()
  ! this is defined in buffering.F90, and the use statement in the head of 
  ! this module refers to it.
  !subroutine SetFilePointer_buff()
  ! this is defined in buffering.F90, and the use statement in the head of 
  ! this module refers to it.
  !subroutine AdvanceFilePointer_buff()
  ! this is defined in buffering.F90, and the use statement in the head of 
  ! this module refers to it.

  !----------------------------------------------------
  ! routines private to this module
  !----------------------------------------------------
  subroutine get_machine_mem_properties(little_endian,error_flag)
    !  #[

    logical, intent(out) :: little_endian
    integer, intent(out) :: error_flag

    ! local variables
    character(len=4)   :: byte_array
    integer(i4_)       :: testvalue
!    integer(i8_)       :: testvalue8
    integer(i1_)       :: small_int

    error_flag = no_error
    testvalue  = 1 + 2*256 + 3*256**2 + 4*256**3
!    testvalue8 = 1_i8_ + 2*256 + 3*256**2 + 4*256**3 + &
!                 5_i8_*256_i8_**4 + 6_i8_*256_i8_**5 + &
!                 7_i8_*256_i8_**6 + 8_i8_*256_i8_**7

    ! testvalue8 can be used to see if words are also swapped
    ! between little-big endian (usually they are, but maybe there
    ! are architectures that are an exception ??)
    ! This is not yet implemented

    ! transfer the 4 byte value to a byte array
    byte_array(1:4) = transfer(testvalue,byte_array(1:4))

    ! inspect the result of the endian test
    ! the hex. testvalue is 04.03.02.01 so the least significant byte 
    ! should have the value 1.
    ! the first byte is the most significant for big-endian machines
    ! the last byte is the most significant for little-endian machines
    ! ==> so if byte 1 holds the value 1 the endianness is little
    ! ==> so if byte 4 holds the value 1 the endianness is big
    small_int = transfer(byte_array(1:1),small_int)
    IF (small_int .eq. 1) THEN
       little_endian = .true.
    ELSE
       small_int = transfer(byte_array(4:4),small_int)
       IF (small_int .eq. 1) THEN
          little_endian = .false.
       ELSE
          print *,"ERROR in subroutine machine_mem_is_little_endian:"
          print *,"endianness test failed...."
          print *,"byte 1 holds: ",transfer(byte_array(1:1),small_int)
          print *,"byte 2 holds: ",transfer(byte_array(2:2),small_int)
          print *,"byte 3 holds: ",transfer(byte_array(3:3),small_int)
          print *,"byte 4 holds: ",transfer(byte_array(4:4),small_int)
          print *,"please check this code again."
          error_flag = error_programming
          return
       END IF
    END IF

    return

  end subroutine get_machine_mem_properties

    !  #]
  subroutine get_machine_io_properties(little_endian,&
                                       headtailsize, error_flag)
    !  #[

    logical, intent(out) :: little_endian
    integer, intent(out) :: headtailsize
    integer, intent(out) :: error_flag

    ! REMARK: this endianness was tested using the g95 compiler
    !         by setting the environment variable G95_ENDIAN as follows:
    !         setenv G95_ENDIAN LITTLE
    !         setenv G95_ENDIAN BIG
    !         setenv G95_ENDIAN NATIVE
    ! My Fedore Linux machine has LITTLE endian as NATIVE format

    ! local variables
    character(len=256) :: endian_testfile
    character(len=4)   :: byte_array
    integer            :: fileunit, pid_nr
    integer(i4_)       :: testvalue
    integer(i8_)       :: testvalue8
    integer(i1_)       :: small_int
    integer            :: i, filesize, ios
    character(len=10)  :: pid_str

    error_flag = no_error

    ! get the process ID number to construct a unique filename 
    ! for the temporary file (this prevent clashes if multiple instances
    ! of the software are running in parallel)
    pid_nr = get_process_id()
    write(pid_str,"(i10.10)",iostat=ios) pid_nr
    IF (ios .ne. 0) THEN
       print *," ERROR in get_machine_io_properties(): "
       print *,"could not convert pid nr ",pid_nr," to a proper string ..."
       error_flag = error_c_interface
       return
    END IF

    endian_testfile = "/tmp/DirectBinaryIO.temporary_endian_testfile.bin."//&
                      pid_str

    testvalue  = 1 + 2*256 + 3*256**2 + 4*256**3
    testvalue8 = 1_i8_ + 2*256 + 3*256**2 + 4*256**3 + &
                 5_i8_*256_i8_**4 + 6_i8_*256_i8_**5 + &
                 7_i8_*256_i8_**6 + 8_i8_*256_i8_**7

    ! testvalue8 can be used to see if words are also swapped
    ! between little-big endian (usually they are, but maybe there
    ! are architectures that are an exception ??)
    ! This is not yet implemented

    ! request an unused fileunit number
    fileunit = get_lun()
    IF (fileunit .lt. 0) THEN
       ! no fileunits available, so report an error
       error_flag = error_opening_file
       return
    END IF

    ! open an unformatted binary file 
    ! and write 1 integer of 4 bytes 
    ! and 1 integer of 8 bytes to this file
    ! This also writes the 4 or 8 bytes head/tail fields
    ! before and after both test values. So the filesize for this
    ! unformatted file will be 4+4+4+4+8+4 = 28
    !                       or 8+4+8+8+8+8 = 44
    open(unit=fileunit,file=endian_testfile,status="replace",&
         action="write",form="UNFORMATTED",ERR=998)
    write(fileunit,ERR=996) testvalue
    write(fileunit,ERR=996) testvalue8
    close(unit=fileunit)

    filesize = get_filesize(endian_testfile)
    select case (filesize)
    case(28)
       headtailsize=4
    case(44)
       headtailsize=8
    case default
       print *,"ERROR in get_machine_io_properties()"
       print *,"the size of the testfile: ",trim(endian_testfile)
       print *,"differs from the expected 28 or 44 bytes."
       print *,"determination of headtailsize failed !!!"
       error_flag = error_reading_file
       return
    end select

    ! read the same file with direct access method to inspect it
    open(unit=fileunit,file=endian_testfile,status="old",&
         action="read",access="direct",recl=1,form="UNFORMATTED",ERR=999)
    DO i=1,4
       read(fileunit,rec=headtailsize+i,iostat=ios) byte_array(i:i)
       IF (ios .ne. 0) goto 997
    END DO
    close(unit=fileunit)
    call free_lun(fileunit)

    ! inspect the result of the endian test
    ! the hex. testvalue is 04.03.02.01 so the least significant byte 
    ! should have the value 1.
    ! the first byte is the most significant for big-endian machines
    ! the last byte is the most significant for little-endian machines
    ! ==> so if byte 1 holds the value 1 the endianness is little
    ! ==> so if byte 4 holds the value 1 the endianness is big
    small_int = transfer(byte_array(1:1),small_int)
    IF (small_int .eq. 1) THEN
       little_endian = .true.
    ELSE
       small_int = transfer(byte_array(4:4),small_int)
       IF (small_int .eq. 1) THEN
          little_endian = .false.
       ELSE
          print *,"ERROR in subroutine machine_io_is_little_endian:"
          print *,"endianness test failed...."
          print *,"byte 1 holds: ",transfer(byte_array(1:1),small_int)
          print *,"byte 2 holds: ",transfer(byte_array(2:2),small_int)
          print *,"byte 3 holds: ",transfer(byte_array(3:3),small_int)
          print *,"byte 4 holds: ",transfer(byte_array(4:4),small_int)
          print *,"please check this code again."
          error_flag = error_programming
          return
       END IF
    END IF

    ! delete the temporary test file
    call remove_file(endian_testfile,error_flag)
    IF (error_flag .ne. no_error) return

    return

996 print *,"ERROR could not write to outputfile: ",trim(endian_testfile)
    error_flag = error_writing_file
    return

997 print *,"ERROR could not read from inputfile: ",trim(endian_testfile)
    error_flag = error_reading_file
    return

998 print *,"ERROR could not open outputfile: ",trim(endian_testfile)
    error_flag = error_opening_file
    return

999 print *,"ERROR could not open inputfile: ",trim(endian_testfile)
    error_flag = error_opening_file
    return

  end subroutine get_machine_io_properties

    !  #]
  function int1_to_unsigned(i_in) result(i_out)
    !  #[
    integer(i1_) :: i_in
    integer(i2_) :: i_out
    
    IF (i_in .ge. 0) THEN
       i_out = i_in
    ELSE
       i_out = int(i_in,i2_) + 256_i2_
    END IF
    
  end function int1_to_unsigned
    !  #]
  function int2_to_unsigned(i_in) result(i_out)
    !  #[
    integer(i2_) :: i_in
    integer(i4_) :: i_out
    
    IF (i_in .ge. 0) THEN
       i_out = i_in
    ELSE
       i_out = int(i_in,i4_) + 256_i4_*256_i4_
    END IF
    
  end function int2_to_unsigned
  !  #]
  function int4_to_unsigned(i_in) result(i_out)
    !  #[
    integer(i4_) :: i_in
    integer(i8_) :: i_out
    
    IF (i_in .ge. 0) THEN
       i_out = i_in
    ELSE
       i_out = int(i_in,i8_) + 256_i8_*256*256*256
    END IF
    
  end function int4_to_unsigned
  !  #]
  !----------------------------------------------------

end module DirectBinaryIO
