module UnformattedIO
  !  #[ Description:
  !   a generic module to read unformatted binary fortran datafiles
  !   taking byteswapping into account (if needed/wanted)
  !   and taking different headtail sizes into account
  !
  ! Written by J. de Kloe, 31-01-2006
  !
  ! Modifications:
  !   09-Feb-2006 J. de Kloe adapted to use the Buffering module
  !   12-May-2006 D. Tan     Null use of ParKind1 module
  !   31-Jul-2006 J. de Kloe renamed advance_file_pointer and set_file_pointer
  !                          to AdvanceFilePointer and SetFilePointer
  !   16-Jan-2008 J. de Kloe phase out integer kind i_ 
  !
  !  #]
  !  #[ modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE DirectBinaryIO, only: OpenDirectBinaryFile,CloseDirectBinaryFile,&
       Init_DirectBinaryIO_module, &
       read_int4_from_bin_file, read_int8_from_bin_file, &
       read_real8_from_bin_file, &
       SetFilePointer, AdvanceFilePointer
  USE ErrorHandler, only: no_error, error_programming, &
                          error_reading_file, error_writing_file, &
                          error_opening_file
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE Numerics, only: i4_, i8_, r8_, nbytes_i4_, nbytes_r8_
  !  #]
  !  #[ variables
  implicit none

  ! the default verbosity setting is Error (as silent as possible)
  integer, private, save :: LocalVerbosity = VerbosityError

  !  #]
contains
  !-----------------------------------------------
  subroutine InitUnformattedIO(error_flag, FileByteOrder, ChoosenVerbosity)
    !  #[
    integer,           intent(in)  :: FileByteOrder ! Little or BigEndian
    integer,           intent(out) :: error_flag
    integer, optional, intent(in)  :: ChoosenVerbosity
        
    error_flag = no_error
    IF (present(ChoosenVerbosity)) LocalVerbosity = ChoosenVerbosity

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       call print_verbosity_setting(LocalVerbosity,"UnformattedIO")
    END IF

    call Init_DirectBinaryIO_module(error_flag, FileByteOrder, &
                                    ChoosenVerbosity)

  end subroutine InitUnformattedIO
    !  #]
  subroutine check_unf_headtail(fileunit, headtailsize, arraybytesize, &
                                error_flag)
    !  #[
    integer,                    intent(in)    :: fileunit
    integer,                    intent(in)    :: headtailsize
    integer(i8_),               intent(in)    :: arraybytesize
    integer,                    intent(out)   :: error_flag
    
    ! local
    integer            :: filepos
    integer(i4_)       :: head4, tail4
    integer(i8_)       :: head8, tail8, head, tail

    error_flag = no_error

    ! remember the location in the file where we are now
    inquire(unit=fileunit,nextrec=filepos)

    IF (headtailsize .eq. 4) THEN
       call read_int4_from_bin_file(fileunit,head4,error_flag)
       head=head4
       !print *,"head4 = ",head4
    ELSE IF (headtailsize .eq. 8) THEN
       call read_int8_from_bin_file(fileunit,head8,error_flag)
       head=head8
       !print *,"head8 = ",head8
    ELSE
       print *,"ERROR in check_unf_headtail():"
       print *,"headtailsize should be 4 or 8, but currently"
       print *,"it seems to be set to: ",headtailsize
       error_flag = error_programming
       return
    END IF

    call AdvanceFilePointer(fileunit,arraybytesize,error_flag)
    IF (error_flag .ne. no_error) return

    IF (headtailsize .eq. 4) THEN
       call read_int4_from_bin_file(fileunit,tail4,error_flag)
       tail=tail4
       !print *,"tail4 = ",tail4
    ELSE IF (headtailsize .eq. 8) THEN
       call read_int8_from_bin_file(fileunit,tail8,error_flag)
       tail=tail8
       !print *,"tail8 = ",tail8
    ELSE
       print *,"ERROR in check_unf_headtail():"
       print *,"headtailsize should be 4 or 8, but currently"
       print *,"it seems to be set to: ",headtailsize
       error_flag = error_programming
       return
    END IF
    
    ! check if head and tail are identical 
    IF (head .ne. tail) THEN
       print *,"wrong headtailsize value or wrong datatype/nvalues"
       print *,"checking of binary array failed..."
       error_flag = error_reading_file
       return
    !ELSE
    !   print *,"everything seems ok"
    END IF

    ! check if headtailsize is what we expect
    IF (head .ne. arraybytesize) THEN
       print *,"ERROR head/tail value is ",head
       print *,"but expected value is: ",arraybytesize
       print *,"Probably you need some byteswapping ???"
       error_flag = error_reading_file
    !ELSE
    !   print *,"head/tail value seems ok"
    END IF

    ! if all seems ok, return to the start of the data position in the file,
    ! so the original position plus the head size
    call SetFilePointer(fileunit,int(filepos+headtailsize,i8_),error_flag)
    if (error_flag .ne. no_error) return

  end subroutine check_unf_headtail
    !  #]
  subroutine check_unf_int4arr(fileunit, headtailsize, nvalues, error_flag)
    !  #[
    integer,                    intent(in)    :: fileunit
    integer,                    intent(in)    :: headtailsize
    integer,                    intent(in)    :: nvalues
    integer,                    intent(out)   :: error_flag
    
    ! local
    integer(i8_) :: arraybytesize

    error_flag = no_error
    arraybytesize = nvalues*nbytes_i4_
    call check_unf_headtail(fileunit, headtailsize, arraybytesize, error_flag)
    if (error_flag .ne. no_error) return

  end subroutine check_unf_int4arr
    !  #]
  subroutine check_unf_real8arr(fileunit, headtailsize, nvalues, error_flag)
    !  #[
    integer,                    intent(in)    :: fileunit
    integer,                    intent(in)    :: headtailsize
    integer,                    intent(in)    :: nvalues
    integer,                    intent(out)   :: error_flag
    
    ! local
    integer(i8_) :: arraybytesize

    error_flag = no_error
    arraybytesize = nvalues*nbytes_r8_
    call check_unf_headtail(fileunit, headtailsize, arraybytesize, error_flag)
    if (error_flag .ne. no_error) return

  end subroutine check_unf_real8arr
    !  #]
  subroutine read_unf_int4(fileunit, headtailsize, value, error_flag)
    !  #[
    integer,      intent(in)    :: fileunit
    integer,      intent(in)    :: headtailsize
    integer(i4_), intent(out)   :: value
    integer,      intent(out)   :: error_flag
    
    ! local
    integer(i4_)       :: dummytail

    error_flag = no_error

    ! this routine reads the head and tail, checks if the size and contents
    ! are what is to be expected, and then jumps back to the start of
    ! the actual data ...
    call check_unf_int4arr(fileunit, headtailsize, 1, error_flag)
    if (error_flag .ne. no_error) then
       print *,"ERROR in read_unf_int4():"
       print *,"unformatted array does not have the expected head/tail !!!"
       print *,"probably you have a wrong byteswap or headtailsize setting"
       return
    end if

    call read_int4_from_bin_file(fileunit,value,error_flag)

    ! read the tail again as dummy value, to step over it
    call read_int4_from_bin_file(fileunit,dummytail,error_flag)

  end subroutine read_unf_int4
    !  #]
  subroutine read_unf_int4arr(fileunit, headtailsize, &
                              nvalues, values, error_flag)
    !  #[
    integer,                    intent(in)    :: fileunit
    integer,                    intent(in)    :: headtailsize
    integer,                    intent(in)    :: nvalues
    integer(i4_), dimension(:), intent(out)   :: values
    integer,                    intent(out)   :: error_flag
    
    ! local
    integer(i4_)       :: value, j, dummytail

    error_flag = no_error

    ! this routine reads the head and tail, checks if the size and contents
    ! are what is to be expected, and then jumps back to the start of
    ! the actual data ...
    call check_unf_int4arr(fileunit, headtailsize, nvalues, error_flag)
    if (error_flag .ne. no_error) then
       print *,"ERROR in read_unf_int4arr():"
       print *,"unformatted array does not have the expected head/tail !!!"
       print *,"probably you have a wrong byteswap or headtailsize setting"
       return
    end if

    DO j=1,nvalues
       call read_int4_from_bin_file(fileunit,value,error_flag)
       !print *,"j = ",j," value = ",value
       values(j) = value
    END DO

    ! read the tail again as dummy value, to step over it
    call read_int4_from_bin_file(fileunit,dummytail,error_flag)

  end subroutine read_unf_int4arr
    !  #]
  subroutine read_unf_real8arr(fileunit, headtailsize, &
                               nvalues, values, error_flag)
    !  #[
    integer,                 intent(in)    :: fileunit
    integer,                 intent(in)    :: headtailsize
    integer,                 intent(in)    :: nvalues
    real(r8_), dimension(:), intent(out)   :: values
    integer,                 intent(out)   :: error_flag
    
    ! local
    integer(i4_) :: j, dummytail
    real(r8_)    :: value

    error_flag = no_error

    ! this routine reads the head and tail, checks if the size and contents
    ! are what is to be expected, and then jumps back to the start of
    ! the actual data ...
    call check_unf_real8arr(fileunit, headtailsize, nvalues, error_flag)
    if (error_flag .ne. no_error) then
       print *,"ERROR in read_unf_real8arr():"
       print *,"unformatted array does not have the expected head/tail !!!"
       print *,"probably you have a wrong byteswap or headtailsize setting"
       return
    end if

    DO j=1,nvalues
       call read_real8_from_bin_file(fileunit,value,error_flag)
       !print *,"j = ",j," value = ",value
       values(j) = value
    END DO

    ! read the tail again as dummy value, to step over it
    call read_int4_from_bin_file(fileunit,dummytail,error_flag)

  end subroutine read_unf_real8arr
    !  #]
  subroutine read_unf_real8arr3D(fileunit, headtailsize, &
       nvalues1, nvalues2, nvalues3, values, error_flag)
    !  #[
    integer,                     intent(in)    :: fileunit
    integer,                     intent(in)    :: headtailsize
    integer,                     intent(in)    :: nvalues1
    integer,                     intent(in)    :: nvalues2
    integer,                     intent(in)    :: nvalues3
    real(r8_), dimension(:,:,:), intent(out)   :: values
    integer,                     intent(out)   :: error_flag
    
    ! local
    integer      :: i,j,k,nvalues
    integer(i4_) :: dummytail
    real(r8_)    :: value

    error_flag = no_error

    ! this routine reads the head and tail, checks if the size and contents
    ! are what is to be expected, and then jumps back to the start of
    ! the actual data ...
    nvalues = nvalues1*nvalues2*nvalues3
    call check_unf_real8arr(fileunit, headtailsize, nvalues, error_flag)
    if (error_flag .ne. no_error) then
       print *,"ERROR in read_unf_real8arr():"
       print *,"unformatted array does not have the expected head/tail !!!"
       print *,"probably you have a wrong byteswap or headtailsize setting"
       return
    end if

    DO k=1,nvalues3
       DO j=1,nvalues2
          DO i=1,nvalues1
             call read_real8_from_bin_file(fileunit,value,error_flag)
             !print *,"j = ",j," value = ",value
             values(i,j,k) = value
          END DO
       END DO
    END DO

    ! read the tail again as dummy value, to step over it
    call read_int4_from_bin_file(fileunit,dummytail,error_flag)

  end subroutine read_unf_real8arr3D
    !  #]
  !-----------------------------------------------
end module UnformattedIO
