module HLOS_retrieval
  !  #[ Description:

  !     A module to
  !     - compute HLOS retrievals for the Mie and Rayleigh channels
  !     - compute associated error quantifiers
  !
  ! code extracted from the file ../Classification/classification.F90
  ! on: 30-Aug-2006 by D. Tan, ECMWF
  ! 
  ! Modifications: 
  !   30-Aug-2006 D. Tan       First extraction from another file
  !   22-Sep-2006 J. de Kloe   Add checks for missing values
  !   13-Oct-2006 David Tan    Bug fix in a print statement
  !   23-Oct-2006 D. Tan       Use of WorkingDataStructure for old L2B-working
  !   03-Nov-2006 J. de Kloe   added check to prevent crash when no 
  !                            collocated AMD profile is present
  !   07-Nov-2006 P. Poli     Move definition of numMRCFreqSteps to L1B_Cal_ads
  !   08-Nov-2006 P. Poli     Remove the spectrometer_index_offset that was
  !                           previously required to handle the missing first 2 pixels
  !                           in Mie_reference_pulses and Mie_measurement_data
  !                           spectrometer counts. TODO: check if we still need
  !                           to overwrite pixels 1 and 2 using pixels 3 and 4.
  !   13-Nov-2006 J. de Kloe   replaced the use of L2B%Obs elements by the
  !                            corresponding working data struct elements
  !                            and moved some copy/scaling instructions
  !                            to Format_and_Copy_WD_to_L2B_Outp()
  !   08-Feb-2007 P. Poli     Use proper elevation angles
  !   14-Feb-2007 P. Poli     Remove tobs correction (now done in L1bP)
  !   18-Apr-2007 D. Tan      Use WorkingDataStructure T and p in Rayleigh processing
  !   24-Apr-2007 D. Tan      Remove L2B%PreProcAuxMetData, other commented code, and
  !                           obsolete alt_geoid_from_wgs84.
  !   29-May-2007 D. Tan      Tidy some comments
  !   30-Jul-2007 P. Poli     Make RBC dataflow explicit
  !   04-Sep-2007 P. Poli     change L1B_DataType to L1B_BRC_DataType,
  !                           change L2BC_DataType to L2BC_BRC_DataType
  !   03-Dec-2007 P. Poli     move spectrometer_index_offset to module
  !                           mieresponse.F90 for consistency with where 
  !                           variables num*ACCDCols are defined
  !                           re-introduce tobs correction (no more done in
  !                           L1bP)
  !   16-Jan-2008 J. de Kloe  phase out integer kind i_
  !
  !  #]
  !  #[ Modules used:

  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE L1B_DataStructure,  only: L1B_BRC_DataType
  USE L2BC_DataStructure, only: L2BC_BRC_DataType, &
                               L2B_ObsRangeGateType
  USE RBC_DataStructure, only: RBC_DataType
  USE L2B_AuxPar_DataStructure, only: L2B_AuxPar_Type, &
                                      Mie_Core_Algorithm_Params_type, &
                                      ErrorQuantMethod_Ray_1Bweighted, &
                                      ErrorQuantMethod_Ray_iliad_sens, &
                                      ErrorQuantMethod_Mie_1Bweighted, &
                                      ErrorQuantMethod_Mie_core_sens
  USE RayleighBrillouinProcessing, only: Rayl_Brill_Proc_Single, &
                                         RRmes_threshold
  USE MieResponse, only: ComputeMieResponse, numACCDCols, numUsefulACCDCols, &
                         spectrometer_index_offset
  USE L2B_Geoloc_ADS, only: Geolocation_Bin_type
  USE L2B_PCD_ADS, only: Obs_Rayleigh_Bin_QC_type, Obs_Mie_Bin_QC_type
  USE L2B_Rayleigh_MDS, only: Rayleigh_Height_Bin_type
  USE L2B_Mie_MDS, only: Mie_Height_Bin_type
  USE Meas_Map, only: Meas_Map_type
  USE Working_DataStructure, only: WorkingDataType, &
                                   WD_MieObsRangeGateType  => MieObsRangeGateType, &
                                   WD_RaylObsRangeGateType => RayleighObsRangeGateType
  USE EE_CFI_Datatypes,  only: IntAl, IntAs, IntAus, &
                         FAdoxy_to_IntAl,  FAdoxy_to_IntAs,  &
                         FAdoxy_to_IntAul, FAdoxy_to_IntAus, &
                         FAdoxy, missing_indicator_IntAul, &
                         SetFadoFull_or_Missing
  !support modules
  USE Numerics, only: r8_, &
                      int_to_real8, intarr_to_real8arr, &
                      missing_indicator_real_r8, missing_real,&
                      missing_indicator_integer, missing_int
  USE VerbosityControl, only: VerbosityError,VerbosityReport
  USE ErrorHandler, only: no_error, error_allocate
  USE AeolusConstants,   only: lambda, pi
  USE Interp1,           only: hunt_nearest
  USE StringTools, only :  to_lowercase
  USE Logging, only: logmsg, log_warn
  USE L1B_Cal_ADS,  only: Mie_Freq_Step_Res_type,numMRCFreqSteps
  USE L1B_Meas_ADS, only: Mie_spectrometer_count_type
  USE L1B_US_MDS,  only:    Rangebin_US_Rayleigh_type
  USE L1B_PCD_ADS, only:    Meas_Alt_Bin_PCD_type

  !  #]
  !  #[ Parameter/Variable/Type declarations

  implicit none

  logical, parameter :: skip_Mie_nonlin_correction = .true. ! .false.

  !
  ! Interim type definition, pending new L1B TDS with
  ! MRC and RRC data
  ! TODO: Put array-instances into 
  !       working datastructure for Mie meas, Mie internal reference and
  !       potentially Rayleigh internal reference too
  !
  TYPE nonlinear_correction_data_type
     real(r8_)       :: raw_response
     real(r8_)       :: correction
  END TYPE nonlinear_correction_data_type

  ! a variable to remember if the module was initialised or not.
  logical, save :: hlos_retrieval_initialised = .false.
  !  #]
  !  #[ Module interface

  ! all routines/functions are public unless stated otherwise below
  ! private section empty
  !  #]
contains

  !---------------------------------------------
  ! public subroutines
  !---------------------------------------------
  subroutine Init_HLOS_retrieval(error_flag)
    !  #[
    integer, intent(out)   :: error_flag ! output

    error_flag = no_error

    ! TODO: issue an entry logmsg

    ! TODO: initialize values needed for this module

    ! TODO: issue an exit logmsg

    hlos_retrieval_initialised = .true. 

  end subroutine Init_HLOS_retrieval

    !  #]
  subroutine Close_HLOS_retrieval()
    !  #[
    ! a routine intended to
    ! free any allocated memory

    hlos_retrieval_initialised = .false.

  end subroutine Close_HLOS_retrieval
    !  #]

    ! TODO: wrapper module for 
    !       Process_Mie_Observations and Process_Rayleigh_Observations
    !       Then make all other routines private?

  ! Note:currently other public subroutines exist below.
  ! TODO: group all public subroutines together

  !------------------------------------------
  ! local subroutines, destined for other modules
  ! TODO: decide whether to move these routines or keep them here
  !---------------------------------------------
  function variance_from_us_and_snr( useful_signal, snr ) result(variance)
    !  #[
    real(r8_), intent(in) :: useful_signal
    real(r8_), intent(in) :: snr

    real(r8_) :: variance ! output

    ! local variables
    real(r8_) :: standard_error

    variance = missing_indicator_real_r8

    ! TODO: More robust checks including
    !       snr below threshold
    if ( ( snr /= 0 )                            .and. &
         ( .not. missing_real( snr ) )           .and. &
         ( .not. missing_real( useful_signal ) )       &
       ) then
       standard_error = useful_signal / snr
       variance = standard_error ** 2
    end if

  end function variance_from_us_and_snr
    !  #]
  function EQRab_L1B( Useful_Signal, K2, K3 ) result(hlos_error)! FAdoxy, m/s
    !  #[
    real(r8_), intent(in) :: Useful_Signal
    real(r8_), intent(in) :: K2, K3

    ! remark: intent statements are not allowed for function results
    real(r8_) :: hlos_error ! output

    hlos_error = missing_indicator_real_r8

    ! Check on K3 could be done in L1B_Screening

    if ( ( K3 /= 0 ) .and. ( Useful_Signal /= 0 ) ) then
       hlos_error = ( Useful_Signal + K2 ) ** K3
       hlos_error = hlos_error / Useful_Signal 
    end if

  end function EQRab_L1B
    !  #]
  function ErrorQuantRay_1Bweighted(L1B,weights,range_bin_index) result(hlos_error)! FAdoxy, m/s
    !  #[
    TYPE(L1B_BRC_DataType), intent(in)    :: L1B ! input ! this input is not yet used !
    real(r8_), dimension(:), intent(in) :: weights ! input
    integer, intent(in) :: range_bin_index ! input ! this input is not yet used !

    ! remark: intent statements are not allowed for function results
    real(r8_) :: hlos_error ! output

    ! local variables
    real(r8_) :: Useful_Signal_A, Useful_Signal_B
    real(r8_) :: EQRa, EQRb, EQRx, fac1, fac2, EQR
    real(r8_) :: Ka2, Ka3, Kb2, Kb3
    real(r8_) :: lambda0
    real(r8_) :: elevation
    real(r8_) :: RRSR

    hlos_error = missing_indicator_real_r8

    ! Placeholders, to be replaced by L1B values/computations
    Useful_Signal_A = 1020._r8_
    Useful_Signal_B =  250._r8_
    Ka2 = 1.1_r8_
    Ka3 = 1.2_r8_
    Kb2 = 1.3_r8_
    Kb3 = 1.4_r8_
    lambda0 = 1.5_r8_
    elevation = 1.6_r8_
    RRSR = 1.7_r8_

    EQRa = EQRab_L1B( Useful_Signal_A, Ka2, Ka3 )
    EQRb = EQRab_L1B( Useful_Signal_B, Kb2, Kb3)
    ! TODO/WARNING: EQRa and EQRb might be missing at this point!!!
    ! so handle this before entering the next sqrt()
    EQRx = sqrt( EQRa ** 2 + EQRb ** 2 )
    fac1 = 0.5_r8_ * lambda0 / sin( elevation )
    fac2 = 1._r8_ / RRSR
    EQR  = fac1 * fac2 * EQRx

    hlos_error = EQR

  end function ErrorQuantRay_1Bweighted
    !  #]
  function ErrorQuantRay_ILIAD(L1B,weights,range_bin_index, &
                               dHLOS_dR, dHLOS_dT, dHLOS_dP, dHLOS_drho, &
                               useful_signal_A, useful_signal_B) result(hlos_error)! FAdoxy, m/s
    !  #[
    TYPE(L1B_BRC_DataType), intent(in)  :: L1B ! input
    real(r8_), dimension(:), intent(in) :: weights ! input
    integer,   intent(in) :: range_bin_index ! input
    real(r8_), intent(in) :: dHLOS_dR, dHLOS_dT, dHLOS_dP, dHLOS_drho, useful_signal_A, useful_signal_B ! input
    ! TODO: put useful_signal_A,  useful_signal_B, etc in WorkingDataStructure?

    ! remark: intent statements are not allowed for function results
    real(r8_) :: hlos_error ! output

    ! local variables
    integer :: i
    TYPE(Rangebin_US_Rayleigh_type), pointer :: Meas_Bin_US
    TYPE(Meas_Alt_Bin_PCD_type)    , pointer :: Meas_Bin_PCD
    !
    real(r8_) :: sum_weights
    real(r8_), dimension(size(weights)) :: unit_weights
    real(r8_), dimension(size(weights)) :: squared_weights
    real(r8_), dimension(size(weights)) :: variance_meas_A
    real(r8_), dimension(size(weights)) :: variance_meas_B
    ! real(r8_) :: dHLOS_dR, dHLOS_dT, dHLOS_dP, dHLOS_drho
    real(r8_) :: useful_signal_AB
    real(r8_) :: variance_A,  variance_B
    real(r8_) :: sigma_R,  sigma_T,  sigma_P,  sigma_rho
    real(r8_) :: EQR

    ! See the document: AE-TN-ECMWF-L2P-0023, version 0.2 draft etc,
    ! (Mathematical Description of the Aeolus
    !  Level-2B Processor) for details of the derivation of these equations.

    hlos_error = missing_indicator_real_r8

    ! Working values needed for variance calculations

    do i = 1, size(weights)
       ! Unit weights needed because weights not normalized to unit sum
       unit_weights(i)    = 1._r8_
       !
       ! Variances at measurement level for current range gate
       Meas_Bin_US  => L1B%US_MDS%Meas_US(i)%Rayleigh(range_bin_index)
       Meas_Bin_PCD => L1B%PCD_ADS%Measurement_PCD(i)%Meas_Alt_Bin_PCD(range_bin_index)
       variance_meas_A(i) = variance_from_us_and_snr( &
                               Meas_Bin_US%Useful_Signal_Channel_A, &
                               Meas_Bin_PCD%Rayleigh_snr_channel_A )
       variance_meas_B(i) = variance_from_us_and_snr( &
                               Meas_Bin_US%Useful_Signal_Channel_B, &
                               Meas_Bin_PCD%Rayleigh_snr_channel_B )
       ! note that elements of variance_meas_A/B might be missing at this point,
       ! depending on what inputs we had
    end do

    ! JK: this is a bit misleading
    ! if you sum an array, just do so, and not hide it in a sum_with_weigths
    ! by setting al weigths to 1
    sum_weights = sum_with_weights(weights,unit_weights,normalize = .false. )

    ! determine the square of the normalised weights
    do i = 1, size(weights)
       squared_weights(i) = ( weights(i) / sum_weights ) ** 2
    end do

    ! Comment (DT): the computation of squared_weights would be tidier if
    ! the input weights were already normalized to sum to unity

    ! useful_signal_A  is an input argument
    ! useful_signal_B  is an input argument

    ! check for missing !
    IF (       .not. missing_real(useful_signal_A) &
         .and. .not. missing_real(useful_signal_B) ) THEN
       useful_signal_AB = useful_signal_A + useful_signal_B
    ELSE
       useful_signal_AB = missing_indicator_real_r8
    END IF

    ! Placeholders, replaced by ILIAD parameters in argument list
    ! dHLOS_dR   = 296._r8_ ! m/s/a.u.
    ! dHLOS_dT   = 0._r8_ ! Kelvin
    ! dHLOS_dP   = 0._r8_ ! per Pa or hPa?
    ! dHLOS_drho = 0._r8_

    ! L2B variance computations
    variance_A   = sum_with_weights( variance_meas_A, squared_weights, normalize = .false. )
    variance_B   = sum_with_weights( variance_meas_B, squared_weights, normalize = .false. )
    ! note that variance_A/B might be missing, depending on the inputs

    ! print *,'dt_debug: useful_signal_A, useful_signal_B =', useful_signal_A, useful_signal_B
    ! print *,'dt_debug: variance_A, variance_B =', variance_A, variance_B

    ! Placeholders, to be replaced by aux met data
    sigma_T   = 0._r8_ ! Kelvin
    sigma_P   = 0._r8_ ! Check Pa or hPa for ILIAD factor
    sigma_rho = 0._r8_

    ! check all inputs for missing
    IF (       .not. missing_real(useful_signal_A) &
         .and. .not. missing_real(useful_signal_B) &
         .and. .not. missing_real(variance_A) &
         .and. .not. missing_real(variance_B) &
         .and. .not. missing_real(useful_signal_AB) &
         .and. .not. missing_real(sigma_T) &
         .and. .not. missing_real(sigma_P) &
         .and. .not. missing_real(sigma_rho) &
         .and. .not. missing_real(dHLOS_dR) &
         .and. .not. missing_real(dHLOS_dT) &
         .and. .not. missing_real(dHLOS_dP) &
         .and. .not. missing_real(dHLOS_drho) ) THEN

       ! Standard error on the L2B Rayleigh response
       sigma_R   =           (useful_signal_A**2) * variance_B ! Not variance_A
       sigma_R   = sigma_R + (useful_signal_B**2) * variance_A ! Not variance_B
       sigma_R   = sqrt( sigma_R )
       sigma_R   = 2._r8_/(useful_signal_AB**2) * sigma_R

       EQR =       ( dHLOS_dR   * sigma_R   ) ** 2
       EQR = EQR + ( dHLOS_dT   * sigma_T   ) ** 2
       EQR = EQR + ( dHLOS_dP   * sigma_P   ) ** 2
       EQR = EQR + ( dHLOS_drho * sigma_rho ) ** 2
       
       EQR = sqrt( EQR )
       hlos_error = EQR
    ELSE
       hlos_error = missing_indicator_real_r8
    END IF

    ! print *,'dt_debug: dHLOS_dR, sigma_R, hlos_error =', dHLOS_dR, sigma_R, hlos_error

  end function ErrorQuantRay_ILIAD
    !  #]
  function bin_RayErrQuant(L1B,weights,range_bin_index,&
                           dHLOS_dR, dHLOS_dT, dHLOS_dP, dHLOS_drho, &
                           useful_signal_A, useful_signal_B &
                          ) result(hlos_error)! FAdoxy, m/s
    !  #[
    TYPE(L1B_BRC_DataType), intent(in)  :: L1B ! input
    real(r8_), dimension(:), intent(in) :: weights ! input
    integer,   intent(in) :: range_bin_index ! input
    real(r8_), intent(in) :: dHLOS_dR, dHLOS_dT, dHLOS_dP, dHLOS_drho, useful_signal_A, useful_signal_B ! input

!    integer, intent(in) :: ErrorQuantMethod_Rayleigh ! input
!
!   TODO: REMOVE the following temporary hard-wired method selection
!see:
!L2BP_Settings%L2B_AuxPar%Data%WVM_Params%Error_Quantifier_Params%ErrorQuantMethod_Rayleigh
    integer, parameter :: ErrorQuantMethod_Rayleigh = ErrorQuantMethod_Ray_iliad_sens

    ! remark: intent statements are not allowed for function results
    real(r8_) :: hlos_error ! output

    hlos_error = missing_indicator_real_r8

    select case (ErrorQuantMethod_Rayleigh)
    case(ErrorQuantMethod_Ray_1Bweighted)
       hlos_error = ErrorQuantRay_1Bweighted(L1b,weights,range_bin_index)
    case(ErrorQuantMethod_Ray_iliad_sens)
       hlos_error = ErrorQuantRay_ILIAD(L1b,weights,range_bin_index, &
                                        dHLOS_dR, dHLOS_dT, dHLOS_dP, dHLOS_drho, &
                                        useful_signal_A, useful_signal_B)
    case default
       ! Method not known, should have been intercepted when
       ! parsing processing options
       ! return missing_indicator_real as set above
       call logmsg(log_warn,"WARNING in bin_RayErrQuant: "//&
                   "unknown method, returning with missing_indicator_real")
    end select

  end function bin_RayErrQuant
    !  #]
  function ErrorQuantMie_1Bweighted(L1B, &
  ! weights, & ! never used
  ! range_bin_index,& ! never used
                                    useful_signal_weighted &
                                    ) result(hlos_error)! FAdoxy, m/s
    !  #[
    TYPE(L1B_BRC_DataType), intent(in)  :: L1B ! input
!    real(r8_), dimension(:), intent(in) :: weights ! input ! never used
!    integer,   intent(in) :: range_bin_index ! input ! never used
    real(r8_), intent(in) :: useful_signal_weighted ! input

    ! remark: intent statements are not allowed for function results
    real(r8_) :: hlos_error ! output

    ! local variables
    real(r8_) :: K1, K2, K3
    real(r8_), parameter :: frequency_to_wind_factor = lambda * 0.5_r8_
    real(r8_) :: elevation
    real(r8_) :: Mi, EQMx, fac1, fac2, EQM

    hlos_error = missing_indicator_real_r8

    K1 = L1B%Cal_ADS%L1B_Char_Data%Sat_Char_Data%Err_Quant%Mie_Err_Quant_K1
    K2 = L1B%Cal_ADS%L1B_Char_Data%Sat_Char_Data%Err_Quant%Mie_Err_Quant_K2
    K3 = L1B%Cal_ADS%L1B_Char_Data%Sat_Char_Data%Err_Quant%Mie_Err_Quant_K3
    !
    ! TODO: replace placeholder with L1B value
    elevation = 35._r8_ * pi / 180._r8_

    Mi   = useful_signal_weighted

    EQMx = ( Mi + K2 ) ** K3
    EQMx = EQMx / Mi
    fac1 = frequency_to_wind_factor / sin( elevation )
    fac2 = K1
    EQM  = fac1 * fac2 * EQMx

    hlos_error = EQM

  end function ErrorQuantMie_1Bweighted
    !  #]
  function ErrorQuantMie_core( & 
  ! L1B, & ! never used 
  ! weights, & ! never used
                               sin_incidence_angle, &
                               tobs, &
                               accd, &
                               PeakLoc, &
                               PeakHeight_pixel_index, &
                               FWHM_pixel_index, &
                               mie_response_slope_alt_bin &
                             ) result(hlos_error)! FAdoxy, m/s
    !  #[
!    TYPE(L1B_DataType), intent(in)    :: L1B ! input ! never used
!    real(r8_), dimension(:), intent(in) :: weights ! input ! never used
    real(r8_), intent(in) :: sin_incidence_angle, &
                             PeakLoc, &
                             PeakHeight_pixel_index, &
                             FWHM_pixel_index, &
                             mie_response_slope_alt_bin ! input
    real(r8_), dimension(1:numUsefulACCDCols), intent(in) :: tobs
    real(r8_), dimension(1:numACCDCols), intent(in) :: accd
    ! TODO: put useful_signal_weighted etc in WorkingDataStructure?

    ! remark: intent statements are not allowed for function results
    real(r8_) :: hlos_error ! output
    real(r8_), dimension(1:numACCDCols) :: alpha

    ! local variables
    !real(r8_) :: K1, K2, K3
    real(r8_), parameter :: frequency_to_wind_factor = lambda * 0.5_r8_
    real(r8_) :: alpha_term1, f_m, f_p, term_m, term_p, PeakHeight_freq
    real(r8_) :: lid_term, fac1, EQM
    integer :: j

    hlos_error = missing_indicator_real_r8

    !
    fac1     = frequency_to_wind_factor / sin_incidence_angle

    PeakHeight_freq = PeakHeight_pixel_index * abs(mie_response_slope_alt_bin)

    ! set up the alpha_j array
    DO j=1+spectrometer_index_offset, &
         numUsefulACCDCols+spectrometer_index_offset
       f_m = REAL(j,r8_)
       f_p = REAL(j+1,r8_)
       term_m=1._r8_+4._r8_*(f_m-PeakLoc)**2/FWHM_pixel_index**2
       term_m=1._r8_/term_m
       term_p=1._r8_+4._r8_*(f_p-PeakLoc)**2/FWHM_pixel_index**2
       term_p=1._r8_/term_p
       alpha(j)=tobs(j-spectrometer_index_offset)*PeakHeight_freq*(term_m - term_p)
    END DO

    ! calculate sums for j=3,18
    alpha_term1=0._r8_
    lid_term = 0._r8_
    DO j=1+spectrometer_index_offset,numUsefulACCDCols+spectrometer_index_offset
       lid_term = lid_term + accd(j) * alpha(j)**2._r8_
       alpha_term1 = alpha_term1 + alpha(j)**2._r8_
    END DO
    lid_term = sqrt( lid_term )

    alpha_term1 = 1._r8_ / alpha_term1

    print *,"REALACC(6)"
    print *,'fac1',fac1
    print *,'alpha_term1',alpha_term1
    print *,'lid_term',lid_term
    print *,'mie_response_slope_alt_bin',mie_response_slope_alt_bin
    print *,'accd',accd
    print *,'PeakLoc',PeakLoc
    print *,'FWHM_pixel_index',FWHM_pixel_index
    print *,"ENDREALACC"

    EQM = fac1 * alpha_term1 * lid_term

    hlos_error = EQM

    ! print *,'debug: Mie_core_hlos_error =', hlos_error

  end function ErrorQuantMie_core
    !  #]
  function bin_MieErrQuant(L1B, &
  ! weights, & ! never used
  ! range_bin_index,& ! never used
                           sin_incidence_angle, &
                           tobs, &
                           accd, &
                           PeakLoc, &
                           PeakHeight, &
                           FWHM, &
                           mie_response_slope_alt_bin, &
                           useful_signal_weighted &
                          ) result(hlos_error)! FAdoxy, m/s
    !  #[
    TYPE(L1B_BRC_DataType), intent(in)  :: L1B ! input
!    real(r8_), dimension(:), intent(in) :: weights ! input ! never used
!    integer, intent(in) :: range_bin_index ! input ! never used
    real(r8_), intent(in) :: sin_incidence_angle, &
                             PeakLoc, &
                             PeakHeight, &
                             FWHM, &
                             mie_response_slope_alt_bin, &
                             useful_signal_weighted ! input
    real(r8_), dimension(1:numUsefulACCDCols), intent(in) :: tobs
    real(r8_), dimension(1:numACCDCols), intent(in) :: accd

!    integer, intent(in) :: ErrorQuantMethod_Mie ! input
!
!   TODO: REMOVE the following temporary hard-wired method selection
!see:
!L2BP_Settings%L2B_AuxPar%Data%WVM_Params%Error_Quantifier_Params%ErrorQuantMethod_Mie

    integer, parameter :: ErrorQuantMethod_Mie = ErrorQuantMethod_Mie_core_sens

    ! remark: intent statements are not allowed for function results
    real(r8_) :: hlos_error ! output

    hlos_error = missing_indicator_real_r8

    select case (ErrorQuantMethod_Mie)
    case(ErrorQuantMethod_Mie_1Bweighted)
       hlos_error = ErrorQuantMie_1Bweighted(L1b, &
       ! weights, & ! never used
       ! range_bin_index, & ! never used
                                        useful_signal_weighted)
    case(ErrorQuantMethod_Mie_core_sens)
       hlos_error = ErrorQuantMie_core( &
       ! L1b, & ! never used
       ! weights, & ! never used
                                        sin_incidence_angle, &
                                        tobs, &
                                        accd, &
                                        PeakLoc, &
                                        PeakHeight, &
                                        FWHM, &
                                        mie_response_slope_alt_bin)
    case default
       ! Method not know, should have been intercepted when
       ! parsing processing options
       ! return missing_indicator_real as set above
    end select

  end function bin_MieErrQuant
    !  #]
  subroutine get_index_start_stop(weights, meas_index_start, meas_index_stop)
    !  #[
    real(r8_), dimension(:) :: weights ! input
    integer, intent(out) :: meas_index_start, meas_index_stop ! output

    ! local variables
    integer :: i !, nr_elements

    meas_index_start = missing_indicator_integer
    meas_index_stop  = missing_indicator_integer

    do i = size(weights), 1, -1
       if ( .not.missing_real(weights(i)) ) then
          if ( weights(i) .gt. 0._r8_ ) then
             meas_index_start = i
          end if
       end if
    end do

    do i = 1, size(weights)
       if ( .not.missing_real(weights(i)) ) then
          if ( weights(i) .gt. 0._r8_ ) then
             meas_index_stop  = i
          end if
       end if
    end do

  end subroutine get_index_start_stop
    !  #]
  !------------------------------------------
  ! local subroutines, private to this module
  ! TODO: decide whether to move these routines or keep them here
  !---------------------------------------------
  function sum_with_weights(values,weights,normalize) result(weighted_sum)
    !  #[
    real(r8_), dimension(:) :: values  ! input
    real(r8_), dimension(:) :: weights ! input
    logical,   optional, intent(in)    :: normalize
    real(r8_) :: weighted_sum

    ! local variables
    integer   :: i, nr_elements
    real(r8_) :: sum_of_values
    real(r8_) :: sum_of_weights

    sum_of_values  = 0._r8_
    sum_of_weights = 0._r8_
    nr_elements = size(values)

    ! report missing if we encounter a serious problem,
    ! like input arrays of different size ...
    IF (nr_elements .ne. size(weights)) THEN
       print *,"programming problem in sum_with_weights() !!!!"
       weighted_sum = missing_indicator_real_r8
       return
    END IF

    DO i=1,nr_elements
       ! report missing if we encounter a serious problem,
       ! like negative weights
       IF (weights(i) .lt. 0.0_r8_) THEN
          print *,"programming problem in sum_with_weights() !!!!"
          weighted_sum = missing_indicator_real_r8
          return
       END IF

       IF (.not. missing_real(values(i))) THEN
          sum_of_values  = sum_of_values + values(i)*weights(i)
          sum_of_weights = sum_of_weights + weights(i)
       END IF
    END DO

    IF ( (nr_elements .gt. 0) .and. (sum_of_weights .eq. 0._r8_)) THEN
       ! this happens if all values are missing
       weighted_sum = missing_indicator_real_r8
       return ! no normalisation neeeded in this case!
    ELSE
       ! Return sum_of_values when normalization is not required
       weighted_sum = sum_of_values
    END IF

    ! Check to see whether normalization is required
    IF (present(normalize)) THEN
       IF ( normalize ) THEN
          ! if no data is selected for averaging, return the missing indicator
          IF (sum_of_weights .eq. 0._r8_) THEN
             weighted_sum = missing_indicator_real_r8
             return
          END IF
          weighted_sum = sum_of_values/sum_of_weights
       END IF
    END IF

  end function sum_with_weights
    !  #]
  function sum_with_weights_int(values,weights,normalize) result(weighted_sum)
    !  #[

    integer,   dimension(:) :: values  ! input
    real(r8_), dimension(:) :: weights ! input
    logical,   optional, intent(in)    :: normalize
    integer :: weighted_sum ! result

    ! local variables
    ! integer :: i ! not used anymore
    integer   :: nr_elements
    real(r8_) :: weighted_sum_real

    ! use automatic allocation for this array
    real(r8_), dimension(1:size(values)) :: values_real

    nr_elements = size(values)

    ! report missing if we encounter a serious problem,
    ! like input arrays of different size ...
    IF (nr_elements .ne. size(weights)) THEN
       print *,"programming problem in sum_with_weights_int() !!!!"
       weighted_sum = missing_indicator_integer
       return
    END IF

    ! Copy real values from integer ones
    !DO i=1,nr_elements
       !if (missing_int(values(i))) then
       !   values_real(i) = missing_indicator_real_r8
       !else
       !   values_real(i)=real(values(i),r8_)
       !end if
       ! please use the generic routine from the numerics module for this type of thing
       !values_real(i) = int_to_real8(values(i))
    !END DO

    values_real = intarr_to_real8arr(values)

    ! Real-valued weighted_sum
    weighted_sum_real = sum_with_weights(values_real,weights,normalize)

    ! Integer part
    weighted_sum = floor( weighted_sum_real )

  end function sum_with_weights_int
    !  #]
  function sum_with_weights_accd(values,weights,offset,normalize) result(weighted_sum)
    !  #[
    type(Mie_spectrometer_count_type), dimension(:) :: values  ! input
    real(r8_), dimension(:) :: weights                         ! input
    real(r8_), optional, intent(in)    :: offset               ! input
    logical,   optional, intent(in)    :: normalize            ! input
    real(FAdoxy), dimension(1:numACCDCols) :: weighted_sum     ! output

    ! local variables
    integer :: i, nr_elements
    integer :: spectrometer_index
    real(FAdoxy) :: weighted_sum_real

    ! use automatic allocation for this array
    real(FAdoxy), dimension(1:size(values)) :: values_real

    ! Initialize
    DO i=1,size(weighted_sum)
       weighted_sum(i) = missing_indicator_integer
    END DO

    nr_elements = size(values)

    ! report missing if we encounter a serious problem,
    ! like input arrays of different size ...
    IF (nr_elements .ne. size(weights)) THEN
       print *,"programming problem in sum_with_weights_accd() !!!!"
       return
    END IF

    DO spectrometer_index = spectrometer_index_offset + 1, size(weighted_sum)

       ! Copy one spectrometer line
       DO i=1,nr_elements
          values_real(i)=Real(values(i)%spectrometer_count(spectrometer_index))
          if (present(offset)) values_real(i)=values_real(i)-offset
       END DO

       ! Real-valued weighted_sum
       weighted_sum_real = sum_with_weights(values_real,weights,normalize)

       ! Real part
       weighted_sum(spectrometer_index) = weighted_sum_real

    END DO ! spectrometer_index

    ! TODO: Check with L1B/L2B teams for suitable filling of columns 1 and 2
    DO spectrometer_index = 1, spectrometer_index_offset
       weighted_sum(spectrometer_index) = weighted_sum( numACCDcols + 1 - spectrometer_index )
    END DO ! spectrometer_index

  end function sum_with_weights_accd
    !  #]
  subroutine spectral_nonlinear_correction(raw_response,&
                          nonlinearity_correction_data,corrected_response)
    !  #[

    real(r8_), intent(in)  :: raw_response
    type(nonlinear_correction_data_type), dimension(:) :: nonlinearity_correction_data  ! input
    real(r8_), intent(out) :: corrected_response

    ! local variables
    integer                :: i_step
    integer                :: n_fstep,i_fstep,i_fsteplo,i_fstephi
    real(r8_), dimension(1:size(nonlinearity_correction_data)) :: raw_response_grid
    real(r8_)              :: frac
    real(r8_)              :: response_correction

    n_fstep = size( nonlinearity_correction_data )

!   Load raw response elements of correction data into array for hunt_nearest
    do i_step = 1, n_fstep
       raw_response_grid( i_step ) = nonlinearity_correction_data(i_step)%raw_response
    end do

    call hunt_nearest(n_fstep,raw_response_grid,raw_response,i_fstep,i_fsteplo,i_fstephi)

    ! Exception handling is TBD

    ! response_correction = nonlinearity_correction_data(i_fstep)%correction
    ! More sophisticated is a linear interpolation
    frac = (                 raw_response - raw_response_grid(i_fsteplo) ) / &
           ( raw_response_grid(i_fstephi) - raw_response_grid(i_fsteplo) )
    response_correction =   (1._r8_ - frac ) * nonlinearity_correction_data(i_fsteplo)%correction + &
                                      frac   * nonlinearity_correction_data(i_fstephi)%correction

    corrected_response = raw_response - response_correction 

  end subroutine spectral_nonlinear_correction
    !  #]
  !---------------------------------------------
  ! public subroutines
  !---------------------------------------------

! to add (as discussed with Ad, 23-11-2005, 
!         see also the email conversation on this date on this subject):
! determine lat/lon/alt of DEM intersection for observations by:
! ==> recording an altitude profile by recording the values given
!     in the L1B file for the DEM for each measurement
! ==> report the highest value for the DEM intersection for the selection
!     of measurements used to construct the current observation.
! ==> this way we will be on the safe side since this will probably 
!     mainly be used to flag data that probably is in the boundary layer
! ==> at a later time effort can be put in to refine this further

  subroutine Process_Rayleigh_Observations(L1B, WorkingData, RBC, L2B, error_flag)
    !  #[
    ! process the constructed observation
    ! determine the weighted values from the L1B Measurements
    ! and determine a hlos result for it.

    TYPE(L1B_BRC_DataType), intent(in)    :: L1B         ! input
    TYPE(WorkingDataType),  intent(in)    :: WorkingData ! input
    TYPE(RBC_DataType),     intent(inout) :: RBC         ! input/output
    TYPE(L2BC_BRC_DataType),intent(inout) :: L2B         ! input/output
    integer,                intent(out)   :: error_flag  ! output

    ! local variables
    integer :: m, r, o, N_meas, AllocStatus, Nr_Obs

    TYPE(WD_RaylObsRangeGateType),  pointer :: WD_RaylObsRangeGate 
    TYPE(Meas_Map_type),        pointer :: Map
    TYPE(Rayleigh_Height_Bin_type), pointer :: Bin

    real(r8_), dimension(:), allocatable :: meas_index_data
!    real(r8_), dimension(:), allocatable :: lat_data ! never used
!    real(r8_), dimension(:), allocatable :: lon_data ! never used
    real(r8_), dimension(:), allocatable :: alt_data_upper
    real(r8_), dimension(:), allocatable :: alt_data_lower
    real(r8_), dimension(:), allocatable :: temperature_data_rayleigh
    real(r8_), dimension(:), allocatable :: pressure_data_rayleigh
    real(r8_), dimension(:), allocatable :: useful_signal_data_rayleigh_A
    real(r8_), dimension(:), allocatable :: useful_signal_data_rayleigh_B
    real(r8_), dimension(:), allocatable :: rho_meas
    real(r8_), dimension(:), allocatable :: spacecraft_los_velocity_meas
    real(r8_), dimension(:), allocatable :: weights

!   Variables for centre-of-gravity calculations
    real(r8_)   :: meas_index_cog_real
    integer     :: meas_index_cog_int
    real(r8_)   :: geoid_offset

    real(r8_) :: alt_cog_from_wgs84
    real(r8_) :: alt_cog_from_geoid

!   Rayleigh processing local variables
!    real(r8_) :: raw_response_alt_bin ! not yet used
!    real(r8_) :: nonlin_response_alt_bin
!    real(r8_) :: raw_response_int_ref ! not yet used
    real(r8_) :: nonlin_response_int_ref

    real(r8_) :: useful_signal_weighted_A, useful_signal_weighted_B 
    real(r8_) :: rho_weighted, RRmes_weighted
    real(r8_) :: rayleigh_response_slope_int_ref

    real(r8_) :: iliad_los_velocity
    real(r8_) :: rayleigh_ground_corr_velfactor
    real(r8_) :: rayleigh_ground_corr_weighting
    real(r8_) :: rayleigh_nonlin_offset

    !  Local variables used in both Mie and Rayleigh processing
    real(r8_) :: elevation_of_height_bin, incidence_angle
    real(r8_) :: sin_incidence_angle
    ! TODO: since this parameter is a constant and returns in many places
    !       it should be moved into support/aeolusconstants.F90
    real(r8_), parameter :: frequency_to_wind_factor = lambda * 0.5_r8_
    real(r8_) :: spacecraft_los_velocity
    real(r8_) :: internal_reference_los_velocity
    real(r8_) :: corrected_los_velocity

    ! variables needed to use the Rayleigh-Brillouin processing
    real(r8_) :: Pmod,Tmod,RRmes,rho
    real(r8_) :: corr0,corr0dR,corr0dT,corr0dP,corr0drho

    ! Variables need for Geoloc(ation)
    TYPE(Geolocation_Bin_type), pointer :: L2B_Geoloc_Bin
    integer :: meas_index_start, meas_index_stop

    ! Variables need for PCD
    real(r8_) :: hlos_error
    TYPE(Obs_Rayleigh_Bin_QC_type), pointer :: Rayleigh_PCD_Bin

    ! Variables need for PCD and Rayleigh_MDS
    real(r8_) :: dHLOS_dR, dHLOS_dT, dHLOS_dP, dHLOS_drho

    ! init
    error_flag = no_error

    N_meas = L2B%nr_of_measurements

    ! REMARK/TODO: JK, 13-sep-2006
    ! should we move any/all of these items into the working data structure?
    allocate(& 
    ! lat_data(1:N_meas),&
    ! lon_data(1:N_meas),&
             meas_index_data(1:N_meas),&
             alt_data_upper(1:N_meas),&
             alt_data_lower(1:N_meas),&
             temperature_data_rayleigh(1:N_meas),&
             pressure_data_rayleigh(1:N_meas),&
             useful_signal_data_rayleigh_A(1:N_meas),&
             useful_signal_data_rayleigh_B(1:N_meas),&
             rho_meas(1:N_meas),&
             spacecraft_los_velocity_meas(1:N_meas),&
             weights(1:N_meas),STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       error_flag = error_allocate
       return
    END IF

    ! point to the map of L1B_Measurements_Used for shorter notation
    Map => L2B%Rayleigh_MDS%Map_of_L1B_Measurements_Used

    ! the nr of range gates is constant for Mie and Rayleigh,
    ! even if their definitions might differ, so this should always work
    rangegateloop: DO r=1, L2B%Obs(1)%nr_of_range_gates
       ! construct the centre-of-gravity values
       ! from the measurements, and store them 
       ! in the L2B%Obs(:) working datastruct

       Nr_Obs = L2B%Rayleigh_MDS%N_Obs_Rayleigh_Actual
       obs_type_loop: DO o=1,Nr_Obs

          print *,"processing Rayleigh Obs. nr. ",o

          ! Fill the L2B DEM intersection parameters - but only for rangegate #1
          ! Would be tidier to interchange the rangegateloop and obs_type_loop but
          ! this would change a lot of expected output.  TODO later.
          ! Would also be worth considering DEM_intersection above the
          ! obs_type_level in the IODD, i.e. next to geoid, and fill just once per BRC.
          if ( r == 1 ) then
!WARNING: this item is not properly replaced when using Lite testdata!
! but will contain the unrelated values from the L1B file in stead
             L2B%Geoloc_ADS%Rayleigh_Geolocation(o)%Lat_of_DEM_Intersection = &
                 L1B%Geoloc_ADS%Obs_Geoloc%lat_of_DEM_intersection ! [IntAl, 10^-6 degE]
!WARNING: this item is not properly replaced when using Lite testdata!
! but will contain the unrelated values from the L1B file in stead
             L2B%Geoloc_ADS%Rayleigh_Geolocation(o)%Lon_of_DEM_Intersection = &
                 L1B%Geoloc_ADS%Obs_Geoloc%lon_of_DEM_intersection ! [IntAl, 10^-6 degE]
!WARNING: this item is not properly replaced when using Lite testdata!
! but will contain the unrelated values from the L1B file in stead
             L2B%Geoloc_ADS%Rayleigh_Geolocation(o)%Alt_of_DEM_Intersection = &
                 INT(L1B%Geoloc_ADS%Obs_Geoloc%alt_of_DEM_intersection,IntAl) ! [IntAl, m]
          end if

          ! point to this ObsRangeGate for shorter notation
          WD_RaylObsRangeGate => WorkingData%RayleighObs(o)%RangeGate(r)
       
          ! compose the weights array for this observation number
          meas_loop1: DO m=1,L2B%nr_of_measurements
             IF (map%Row(m)%Bin(r)%Which_L2B_Obs .eq. o) THEN
                !weights(m) = 1.*Map%Row(m)%Bin(r)%Weight/1000._r8_
                ! use the weights stored as reals for now, to have
                ! higher precision
                weights(m) = WorkingData%RayleighMeas(m)%RangeGate(r)%weight/1000._r8_
             ELSE
                weights(m) = 0._r8_
             END IF
          END DO meas_loop1
          !
          ! In the old convention, the sum of the real-valued weights would be 
          ! unity and the sum of the integer-valued weights would be 1000.
          ! In the new convention, the max of the real-valued weights should be 
          ! unity and the max of the integer-valued weights sould be 1000.
          ! Care must be taken when using weights in weighted sums, in
          ! particular whether the Normalize argument should be true or false.

          ! copy the lat,lon measurement indices to a simple array
          ! also the alt and geoid data
          ! and the Rayleigh useful signal data
          meas_loop2: DO m=1,L2B%nr_of_measurements
             !  #[ copy some data to arrays

             ! copy the meas index to a real, to enable using it in sum_with_weights()
             ! but why not use sum_with_weights_int() ?
             meas_index_data(m) = Real( m, r8_ )
             alt_data_upper(m) = WorkingData%RayleighMeas(m)%RangeGate(r)%alt_end
             alt_data_lower(m) = WorkingData%RayleighMeas(m)%RangeGate(r)%alt_start

!TODO: should we check for missing with this item?
!WARNING: this item is not properly replaced when using Lite testdata!
! but will contain the unrelated values from the L1B file in stead
             spacecraft_los_velocity_meas(m) = &
                  L1B%Geoloc_ADS%Meas_Geoloc(m)%AOCS_LOS_Velocity

!TODO: should we check for missing with this item?
!WARNING: this item is not properly replaced when using Lite testdata!
! but will contain the unrelated values from the L1B file in stead
             useful_signal_data_rayleigh_A(m) = &
                  L1B%US_MDS%Meas_US(m)%Rayleigh(r)%Useful_Signal_Channel_A
!TODO: should we check for missing with this item?
!WARNING: this item is not properly replaced when using Lite testdata!
! but will contain the unrelated values from the L1B file in stead
             useful_signal_data_rayleigh_B(m) = &
                  L1B%US_MDS%Meas_US(m)%Rayleigh(r)%Useful_Signal_Channel_B
!TODO: should we check for missing with this item?
! probably yes, therefore the following temporary workaround was added


             !WARNING: this item is not properly filled when using Lite testdata!
             ! but will contain the unrelated values from the L1B file in stead
! TODO: use the workingdate rho here, in stead of hard-wiring to the L1B value
!       since we may get another algorithm to retrieve rho in the future.
             rho_meas(m) = &
                  L1B%PCD_ADS%Measurement_PCD(m)%Meas_Alt_Bin_PCD(r)%scattering_ratio
             ! TODO: it should be replaced by:
             ! because this L2B working data field is filled with a copy
             ! of the L1B item if appropriate, and with a calculated result
             ! when using Lite testdata
             !rho_meas(m) = &
             !     WorkingData%RayleighMeas(m)%RangeGate(r)%backscatter_ratio

             ! temporary workaround for faulty values in the 108 BRC L1B file:
             ! TODO: to be moved to the screening subroutine
             if (rho_meas(m) .lt.   1.0) rho_meas(m) = missing_indicator_real_r8
             if (rho_meas(m) .gt. 100.0) rho_meas(m) = missing_indicator_real_r8

             !  #]
          END DO meas_loop2

          ! Determine the L1B measurement indices for the start and
          ! stop geolocations for the current observation type, both
          ! indices as integers
          call get_index_start_stop(weights, meas_index_start, meas_index_stop)

          ! Take lat and lon from L1B measurement index meas_index_start
          if(.not.missing_int(meas_index_start))then
             WD_RaylObsRangeGate%lat_start = WorkingData%RayleighMeas(meas_index_start)%RangeGate(r)%lat
             WD_RaylObsRangeGate%lon_start = WorkingData%RayleighMeas(meas_index_start)%RangeGate(r)%lon
          end if

          ! Take lat and lon from L1B measurement index meas_index_stop
          if(.not.missing_int(meas_index_stop ))then
             WD_RaylObsRangeGate%lat_stop  = WorkingData%RayleighMeas(meas_index_stop )%RangeGate(r)%lat
             WD_RaylObsRangeGate%lon_stop  = WorkingData%RayleighMeas(meas_index_stop )%RangeGate(r)%lon
          end if

          ! Determine the L1B measurement index for the centre-of-gravity
          ! First as a real number, then as an integer
          meas_index_cog_real = sum_with_weights(meas_index_data,weights,normalize=.true.)
          if(.not.missing_real(meas_index_cog_real))then
             meas_index_cog_int  = nint( meas_index_cog_real )

             ! determine the horizontal centre-of-gravity
             !
             ! Take from L1B measurement index meas_index_cog_int
             WD_RaylObsRangeGate%lat_cog = WorkingData%RayleighMeas(meas_index_cog_int)%RangeGate(r)%lat
             WD_RaylObsRangeGate%lon_cog = WorkingData%RayleighMeas(meas_index_cog_int)%RangeGate(r)%lon

             ! L2B alt_end and alt_start also via copy from L1B measurement index meas_index_cog_int
!WARNING: this item is not properly replaced when using Lite testdata!
! but will contain the unrelated values from the L1B file in stead
             geoid_offset = L1B%Geoloc_ADS%Obs_Geoloc%Geoid_Separation_of_Height_Bin ! 0._r8_ ! Previous Placeholder
             WD_RaylObsRangeGate%alt_end   = WorkingData%RayleighMeas(meas_index_cog_int)%RangeGate(r)%alt_end   - geoid_offset
             WD_RaylObsRangeGate%alt_start = WorkingData%RayleighMeas(meas_index_cog_int)%RangeGate(r)%alt_start - geoid_offset

             ! determine the vertical centre-of-gravity
             !
             ! Average from measurement index meas_index_cog_int
             ! and account for geoid offset
             ! TODO: once we have a reliable extinction for this obs.
             !       use this to get a more precise cog value
             alt_cog_from_wgs84 = 0.5_r8_ * ( alt_data_upper(meas_index_cog_int)   + &
                                          alt_data_lower(meas_index_cog_int) )
             alt_cog_from_geoid = alt_cog_from_wgs84 - geoid_offset

             WD_RaylObsRangeGate%alt_cog = alt_cog_from_geoid
          end if ! missing_real(meas_index_cog_real)

          ! get the temperature & pressure data for the appropriate
          ! altitude for each range bin, and copy it to a simple array
          meas_loop_auxmet: DO m=1,L2B%nr_of_measurements
             !  #[ copy nwp data to arrays

                ! recall that in Copy_AMD_to_WorkingData() in
                ! auxiliary.F90 these elements are already interpolated 
                ! on measurement level using the interpolate() function from 
                ! Interp1.F90

             temperature_data_rayleigh(m) = &
                  WorkingData%RayleighMeas(m)%RangeGate(r)%T
             pressure_data_rayleigh(m)    = &
                  WorkingData%RayleighMeas(m)%RangeGate(r)%p

!print *,"m = ",m
!print *,"alt_data_upper(1)       = ",alt_data_upper(1)
!print *,"alt_data_lower(1)       = ",alt_data_lower(1)
!print *,"alt_cog_from_wgs84      = ",alt_cog_from_wgs84
!print *,"alt_cog_from_geoid      = ",alt_cog_from_geoid
!print *,"temperature_data_rayleigh(m) = ",temperature_data_rayleigh(m)
!print *,"pressure_data_rayleigh(m)    = ",pressure_data_rayleigh(m)
             !  #]
          END DO meas_loop_auxmet

          ! average the NWP data, using the proper weights
          WD_RaylObsRangeGate%T_ref = &
               sum_with_weights(temperature_data_rayleigh,weights,normalize=.true.)
          WD_RaylObsRangeGate%p_ref = &
               sum_with_weights(pressure_data_rayleigh,weights,normalize=.true.)

          ! For Rayleigh channels A & B separately,
          ! accumulate the useful signals, using the proper weights
          useful_signal_weighted_A = &
               sum_with_weights(useful_signal_data_rayleigh_A,weights,&
                                normalize=.true.)
          useful_signal_weighted_B = &
               sum_with_weights(useful_signal_data_rayleigh_B,weights,&
                                normalize=.true.)

          ! calculate scattering ratio
          rho_weighted = sum_with_weights(rho_meas,weights,normalize=.true.)

          ! calculated the RRmes value.
          ! this is the input for the RayleighBrillouin algorithm 
          RRmes_weighted = missing_indicator_real_r8
          IF (       .not. missing_real(useful_signal_weighted_A) &
               .and. .not. missing_real(useful_signal_weighted_B) ) THEN
             ! TODO remark: due to offset subtraction negative A/B signals may occur now and
             ! then. Should we just process these? or check for this (in screening)
             ! and set RRmes_weighted to missing if this happens?
             IF ( (useful_signal_weighted_A + useful_signal_weighted_B) &
                  .ne. 0._r8_) THEN
                RRmes_weighted = &
                     (useful_signal_weighted_A-useful_signal_weighted_B) / &
                     (useful_signal_weighted_A+useful_signal_weighted_B)
             END IF
          END IF
          
          IF (missing_real(RRmes_weighted)) THEN
             ! we had no Rayleigh signal in one or both channels
             ! so nothing to do, except logging it
             ! this is a valid condition and can occur below clouds
             print *,"no Rayleigh signal available"
          ELSE
             ! Test for values beyond expected range for Aeolus instrument
             IF ( abs( RRmes_weighted ) .le. RRmes_threshold ) THEN
                ! do the Rayleigh processing

                ! Rayleigh-Brillouin processing
                ! check all needed inputs for missing, to make sure the 
                ! Rayl_Brill_Proc_Single routine is not called with invalid data
                ! since no missing checks are implemented inside
                IF (       .not. missing_real(WD_RaylObsRangeGate%p_ref) &
                     .and. .not. missing_real(WD_RaylObsRangeGate%T_ref) &
                     .and. .not. missing_real(rho_weighted)           ) THEN
                   !  #[ do the actual Rayleigh processing

                   ! Setup for Iliad correction scheme
                   Pmod  = WD_RaylObsRangeGate%p_ref/100._r8_ ! Pa to hPa
                   Tmod  = WD_RaylObsRangeGate%T_ref
                   rho   = rho_weighted
                   RRmes = RRmes_weighted !determined from L1B Useful_Signal


                   !IF (LocalVerbosity >= VerbosityReport) THEN
                   print *,"REALACC(6)"
                   print *,"Rayl_Brill_Proc_Single: inputs"
                   print *,"Pmod = ",Pmod
                   print *,"Tmod = ",Tmod
                   print *,"RRmes = ",RRmes
                   print *,"rho = ",rho
                   print *,"ENDREALACC"
                   !ENDIF

                   call Rayl_Brill_Proc_Single(RBC, & ! input/output
                         Pmod,Tmod,RRmes,rho,&        ! inputs
                        corr0,corr0dR,corr0dT,corr0dP,corr0drho,error_flag) ! outputs
                   if (error_flag .ne. no_error) return

                   !IF (LocalVerbosity >= VerbosityReport) THEN
                   print *,"REALACC(6)"
                   print *,"Rayl_Brill_Proc_Single: outputs"
                   print *,"corr0     = ",corr0
                   print *,"corr0dR   = ",corr0dR
                   print *,"corr0dT   = ",corr0dT
                   print *,"corr0dP   = ",corr0dP
                   print *,"corr0drho = ",corr0drho            
                   print *,"ENDREALACC"
                   ! TODO: should we do any range-screening, 
                   !       or missing checks on these results?
                   !ENDIF

                   iliad_los_velocity = corr0

                   ! post-Iliad Rayleigh-Brillouin processing

                   ! temporary placeholders
                   nonlin_response_int_ref = 0._r8_
                   rayleigh_nonlin_offset = 0._r8_
                   rayleigh_response_slope_int_ref = 1._r8_
                   internal_reference_los_velocity = 0._r8_

                   ! Update placeholders
                   print *,'dt: WARNING: geoid height                    ',&
                        'is placeholder in Ray-Br processing'
                   print *,'dt: WARNING: scattering ratio (rho)          ',&
                        'is placeholder in Ray-Br processing'
                   elevation_of_height_bin = L1B%Geoloc_ADS%Obs_Geoloc%&
                          Rayleigh_Rangebin_Geoloc(r)%topoc_elev_of_height_bin
                   ! the angle reported in L1b data is elevation (about 53 degrees)
                   incidence_angle = (90._r8_ - abs(elevation_of_height_bin)) * &
                                     & pi/180._r8_
                   sin_incidence_angle = sin(incidence_angle)

!TODO: should we check for missing with this item?
! (sum_with_weights retuerns missing if all inputs are missing which may occur)
                   spacecraft_los_velocity = &
                        sum_with_weights(spacecraft_los_velocity_meas,weights,normalize=.true.)
                   ! these are not weighted, 
                   ! so may be moved to outside the loops
!TODO: should we check for missing with this item?
!WARNING: this item is not properly replaced when using Lite testdata!
! but will contain the unrelated values from the L1B file in stead
                   rayleigh_ground_corr_velfactor = &
                        L1B%GWD_ADS%Rayleigh_ground_corr_velocity
!TODO: should we check for missing with this item?
!WARNING: this item is not properly replaced when using Lite testdata!
! but will contain the unrelated values from the L1B file in stead
                   rayleigh_ground_corr_weighting = &
                        L1B%GWD_ADS%Rayleigh_grnd_corr_wght_factor
                   print *,'dt: WARNING: nonlin_response_int_ref         ',&
                        'is placeholder in Ray-Br processing'
                   print *,'dt: WARNING: rayleigh_nonlin_offset          ',&
                        'is placeholder in Ray-Br processing'
                   print *,'dt: WARNING: rayleigh_response_slope_int_ref ',&
                        'is placeholder in Ray-Br processing'
                   print *,'dt: WARNING: internal_reference_los_velocity ',&
                        'is placeholder in Ray-Br processing'

                   ! dt: Call to Rayleigh spectral response nonlinearity 
                   ! dt: correction for int_ref only
                   print *,'dt: WARNING: spectral_nonlinearity_correction ',&
                        'not yet implemented for Ray-Br processing'
                   ! call spectral_nonlinear_correction(&
                   !            raw_response_int_ref,&
                   !            Rayleigh_spec_calib_int_ref,&
                   !            nonlin_response_int_ref)

                   ! this is a temporary placeholder for the moment, since the result is
                   ! always:  355e-9 * 0.5 * ( 0 - 0)/ 1 = 0
                      ! TODO: add a reference to which document (MAD/DPM?) 
                      ! this expression comes from and which page/eq. nr.
                   internal_reference_los_velocity = &
                        frequency_to_wind_factor * &
                        ( nonlin_response_int_ref - rayleigh_nonlin_offset )/&
                        rayleigh_response_slope_int_ref


                   ! dt
                   !print *, 'dt: los corrections =', spacecraft_los_velocity,&
                   !            rayleigh_ground_corr_velfactor, &
                   !            rayleigh_ground_corr_weighting

                   ! It might be possible for the following to share a common 
                   ! method with Mie processing
                   IF (       .not. missing_real(iliad_los_velocity) &
                        .and. .not. missing_real(spacecraft_los_velocity) &
                        .and. .not. missing_real(rayleigh_ground_corr_velfactor) &
                        .and. .not. missing_real(rayleigh_ground_corr_weighting) &
                        .and. .not. missing_real(internal_reference_los_velocity)  ) THEN
                      ! TODO: add a reference to which document (MAD/DPM?) 
                      ! this expression comes from and which page/eq. nr.
                      corrected_los_velocity = iliad_los_velocity - &
                           spacecraft_los_velocity - &
                           rayleigh_ground_corr_velfactor * &
                           rayleigh_ground_corr_weighting - &
                           internal_reference_los_velocity
                      WD_RaylObsRangeGate%hlos_cog = &
                           corrected_los_velocity / sin_incidence_angle
                   ELSE
                      corrected_los_velocity       = missing_indicator_real_r8
                      WD_RaylObsRangeGate%hlos_cog = missing_indicator_real_r8
                   END IF


                   ! L1B elements and L2B results that do not require scaling 
                   ! are copied directly L2B results that must be scaled before
                   ! writing them to file are written in the WorkingData structure first
                   ! and copied in Format_and_Copy_WD_to_L2B_Outp()
                   ! (I don't like this difference much. I think it would be much easier and
                   !  cleaner to just create working data elements for all results we
                   !  wish to export, and copy them all in Format_and_Copy_WD_to_L2B_Outp()
                   !  This prevents mistakes I think)
                   L2B_Geoloc_Bin => L2B%Geoloc_ADS%Rayleigh_Geolocation(o)%Height_Bin(r)
                   L2B_Geoloc_Bin%LOS_Azimuth      =  &
                        L1B%Geoloc_ADS%Obs_Geoloc%Rayleigh_Rangebin_Geoloc(r)%&
                        topoc_azim_of_height_bin ! [deg]
                   L2B_Geoloc_Bin%LOS_Elevation    = elevation_of_height_bin  ! [deg]
                   L2B_Geoloc_Bin%LOS_Sat_Velocity = spacecraft_los_velocity  ! [m/s]

                   ! To do: Computation of PCD parameters
                   !        and copying to PCD_ADS
                   !        Some may be done in Construct_Rayleigh_PCD?
                   dHLOS_dR   = corr0dR   / sin_incidence_angle
                   dHLOS_dT   = corr0dT   / sin_incidence_angle
                   dHLOS_dP   = corr0dP   / sin_incidence_angle
                   dHLOS_drho = corr0drho / sin_incidence_angle
                   hlos_error = bin_RayErrQuant(L1B,weights,r, &
                                                dHLOS_dR, dHLOS_dT, dHLOS_dP, dHLOS_drho, &
                                                useful_signal_weighted_A, &
                                                useful_signal_weighted_B ) ! FAdoxy, m/s
                   Rayleigh_PCD_Bin => L2B%PCD_ADS%Rayleigh_Processing_QC%&
                                           Obs_Rayleigh_QC(o)%Obs_Rayleigh_Bin_QC(r)
                   Rayleigh_PCD_Bin%Hlos_Error_Estimate       = hlos_error ! FAdoxy, m/s

                   !
                   ! copy all obtained results to the Rayleigh_MDS datastruct
                   Bin => L2B%Rayleigh_MDS%Rayleigh_Profile(o)%Height_Bin(r)
                   Bin%Validity_Flag                   =.true.
                   ! convert from m/s to cm/s and convert to IntAs type
                   
                   ! please use the FAdoxy_to_IntAs, FAdoxy_to_IntAus etc. functions
                   ! from ee_cfi_datatypes.F90, since these take missing into account,
                   ! and do a proper rangecheck.

! todo: move this copy into Format_and_Copy_WD_to_L2B_Outp()
                   Bin%Rayleigh_Wind_Velocity          = &
                        FAdoxy_to_IntAs(WD_RaylObsRangeGate%hlos_cog,scale=100._r8_)
                   !     floor(L2B_ObsRangeGate%hlos_cog*100,IntAs)
                   
                   Bin%Rayleigh_Wind_to_Pressure       = &
                        FAdoxy_to_IntAs((corr0dP/sin_incidence_angle),scale=1000000._r8_)
                   !     floor((corr0dP/sin_incidence_angle)*1000000,IntAs)

                   Bin%Rayleigh_Wind_to_Temperature    = &
                        FAdoxy_to_IntAs((corr0dT/sin_incidence_angle),scale=100._r8_)
                   !     floor((corr0dT/sin_incidence_angle)*100,IntAs)

                   Bin%Rayleigh_Wind_to_Backscat_Ratio = &
                        FAdoxy_to_IntAs((corr0drho/sin_incidence_angle),scale=100._r8_)
                   !     floor((corr0drho/sin_incidence_angle)*100,IntAs)

                   Bin%Reference_Pressure              = &
                        FAdoxy_to_IntAul(Pmod,scale=100._r8_)
                   !     floor(Pmod*100._r8_,IntAul)

                   Bin%Reference_Temperature           = &
                        FAdoxy_to_IntAus(Tmod,scale=100._r8_)
                   !     floor(Tmod*100,IntAus)

                   ! type is IntAul, scale is 10^-6
                   Bin%Reference_Backscatter_Ratio     = &
                        FAdoxy_to_IntAul(rho,scale=1000000._r8_)
                   !     floor(rho*1000000,IntAul)

                   !Bin%Rayleigh_Error_Quantifier       = 0
                   !Bin%Integration_Length              = 0

                   !  #]
                ELSE
                   ! Exception, needed nwp values seem invalid:
                   ! this is a problem in the input data, and should have
                   ! been flagged earlier in the program.
                   ! So just continue and leave this result missing,
                   ! and only report the condition
                   print *,'WARNING: Rayleigh calculation failed due to'
                   print *,'missing p_ref and/or T_ref, and/or rho_weighted'
                   print *,"REALACC(6)"
                   print *,"WD_RaylObsRangeGate%p_ref = ",WD_RaylObsRangeGate%p_ref
                   print *,"WD_RaylObsRangeGate%T_ref = ",WD_RaylObsRangeGate%T_ref
                   print *,"rho_weighted = ",rho_weighted
                   print *,"ENDREALACC"
                END IF ! p_ref/T_ref/rho_weighted check
             ELSE
                ! Exception, Rayleigh signal seems invalid:
                ! this is a problem in the input data, so the program
                ! should just continue and leave this result missing,
                ! so only report the condition
                print *,'WARNING: Rayleigh response out of range'
                print *,"REALACC(6)"
                print *,'RRmes_weighted =', RRmes_weighted
                print *,'useful_signal_weighted_A,B =', &
                     useful_signal_weighted_A, useful_signal_weighted_B
                print *,"useful_signal_data_rayleigh_A(:) = ",&
                     useful_signal_data_rayleigh_A(:)
                print *,"useful_signal_data_rayleigh_B(:) = ",&
                     useful_signal_data_rayleigh_B(:)
                print *,"weights(:) = ",weights(:)
                print *,"ENDREALACC"
             END IF ! threshold check on RRmes_weighted

          END IF ! missing check on RRmes_weighted

          print *,"results for Rayleigh Obs. nr. ",o," rangebin ",r
          print *,"REALACC(6)"
          print *,"lat_cog  = ",WD_RaylObsRangeGate%lat_cog
          print *,"lon_cog  = ",WD_RaylObsRangeGate%lon_cog
          print *,"Ref_T    = ",WD_RaylObsRangeGate%T_ref
          print *,"Ref_p    = ",WD_RaylObsRangeGate%p_ref
          print *,"hlos_cog = ",WD_RaylObsRangeGate%hlos_cog
          print *,"ENDREALACC"

       END DO obs_type_loop
       
    END DO rangegateloop

    ! Fill the TimeStamp and GeoidOffset of the GeolocDSR, even if
    ! it overwrites values from Process_Mie_Observations

!WARNING: this item is not properly replaced when using Lite testdata!
! but will contain the unrelated values from the L1B file in stead
    L2B%Geoloc_ADS%Start_of_Obs_DateTime = &
        L1B%Geoloc_ADS%start_of_obs_datetime
    L2B%Geoloc_ADS%WGS84_to_Geoid_Altitude = &
        FAdoxy_to_IntAl(geoid_offset) ! [IntAl, m]

  end subroutine Process_Rayleigh_Observations
    !  #]
  subroutine Process_Mie_Observations(L1B, WorkingData, L2B_AuxPar, &
                                      L2B, error_flag)
    !  #[

    TYPE(L1B_BRC_DataType), intent(in)    :: L1B         ! input
    TYPE(WorkingDataType),  intent(in)    :: WorkingData ! input
    TYPE(L2B_AuxPar_Type),  intent(in)    :: L2B_AuxPar ! input
    TYPE(L2BC_BRC_DataType),intent(inout) :: L2B ! input/output
    integer,                intent(out)   :: error_flag ! output

    ! local variables
    integer :: m, r, o, N_meas, AllocStatus
    integer :: i_tobs
    integer :: Nr_Obs

    TYPE(WD_MieObsRangeGateType), pointer :: WD_MieObsRangeGate 
    TYPE(Meas_Map_type),        pointer :: Map
    TYPE(Mie_Height_Bin_type),  pointer :: Bin

    real(r8_), dimension(:), allocatable :: meas_index_data
!    real(r8_), dimension(:), allocatable :: lat_data ! never used
!    real(r8_), dimension(:), allocatable :: lon_data ! never used
    real(r8_), dimension(:), allocatable :: alt_data_upper
    real(r8_), dimension(:), allocatable :: alt_data_lower
    real(r8_), dimension(:), allocatable :: spacecraft_los_velocity_meas
    real(r8_), dimension(:), allocatable :: weights
    real(r8_), dimension(:), allocatable :: unit_values

!   Variables for centre-of-gravity calculations
    real(r8_) :: meas_index_cog_real
    integer   :: meas_index_cog_int
    real(r8_) :: geoid_offset

    real(r8_) :: alt_cog_from_wgs84 ! not yet used
    real(r8_) :: alt_cog_from_geoid ! not yet used

!   Mie processing local variables
    logical :: do_MieWind_processing
    real(FAdoxy) :: SNR_Threshold
    real(FAdoxy) :: SNR_alt_bin !, SNR_int_ref
    type(Mie_spectrometer_count_type), dimension(:), &
         allocatable :: spectrometer_count_Mie_alt_bin
    type(Mie_spectrometer_count_type), dimension(:), &
         allocatable :: spectrometer_count_Mie_int_ref
    character(len=50) :: SNR_alt_bin_txt
    real(FAdoxy) :: sum_of_weights

    TYPE(Mie_Core_Algorithm_Params_type), pointer  :: Mie_Core_Algorithm_Params
    !
    ! interface to Mie response Core algorithm
    integer        :: ChoosenVerbosity                      ! input
    real(FAdoxy)   :: StartFWHM,ResErrThresh,FWHMUpThresh   ! input
    real(FAdoxy)   :: FWHMLoThresh, PeakHeightUpThresh      ! input
    real(FAdoxy)   :: PeakHeightLoThresh,PeakLocThresh      ! input
    real(FAdoxy)   :: NonLinOptThresh                       ! input
    integer(IntAs) :: MaxItLorFit                           ! input
    integer(IntAs) :: MaxItNonLinOpt,NumSpecSubSamp         ! input
    real(FAdoxy), dimension(1:numUsefulACCDCols) :: tobs    ! input
    ! real(FAdoxy), dimension(1:numACCDCols) :: accd          ! input
    logical        :: isMeas                                ! input
    real(FAdoxy)   :: PeakLoc_meas,FWHM_meas, &
                      Offset_meas,OffsetSub_meas, &
                      PeakHeight_meas,ResErr_meas ! output
    real(FAdoxy)   :: PeakLoc_ref ,FWHM_ref , &
                      Offset_ref ,OffsetSub_ref , &
                      PeakHeight_ref ,ResErr_ref  ! output
    logical        :: ValFlag_meas                          ! output
    logical        :: ValFlag_ref                           ! output
    integer(IntAs) :: NumIt_meas                            ! output
    integer(IntAs) :: NumIt_ref                             ! output
    integer(IntAs) :: error_flag_mieresponse                ! output ! Not IntAl?
    !
    real(FAdoxy)   :: OffsetSub_single_meas
    real(FAdoxy), dimension(1:numACCDCols) :: accd_meas     ! input
    real(FAdoxy), dimension(1:numACCDCols) :: accd_meas_weighted_offset
    real(FAdoxy), dimension(1:numACCDCols) :: accd_intref   ! input
    
    type(nonlinear_correction_data_type), dimension(numMRCFreqSteps) ::  Mie_spec_calib_alt_bin, Mie_spec_calib_int_ref
    ! TODO: put in Working DataStructure
    type(Mie_Freq_Step_Res_type) :: MRC_Freq_Step_Res
    integer :: i_FreqStep
    
    real(r8_) :: raw_response_alt_bin
    real(r8_) :: nonlin_response_alt_bin, mie_response_slope_alt_bin
    real(r8_) :: raw_response_int_ref
    real(r8_) :: nonlin_response_int_ref, mie_response_slope_int_ref

    real(r8_) :: internal_reference_zero_freq
    real(r8_) :: measurement_zero_freq

    real(r8_) :: mie_ground_corr_velfactor
    real(r8_) :: mie_ground_corr_weighting
    real(r8_) :: mie_los_velocity

    !  Local variables used in both Mie and Rayleigh processing
    real(r8_) :: elevation_of_height_bin, incidence_angle
    real(r8_) :: sin_incidence_angle
    real(r8_), parameter :: frequency_to_wind_factor = lambda * 0.5_r8_
    real(r8_) :: spacecraft_los_velocity
    real(r8_) :: internal_reference_los_velocity
    real(r8_) :: corrected_los_velocity

    ! Variables need for Geoloc(ation)
    TYPE(Geolocation_Bin_type), pointer :: L2B_Geoloc_Bin
    integer   :: meas_index_start, meas_index_stop

    ! Variables needed for PCD
    real(r8_) :: useful_signal_weighted
    real(r8_) :: hlos_error
    TYPE(Obs_Mie_Bin_QC_type), pointer :: Mie_PCD_Bin

    ! init
    error_flag = no_error

    N_meas = L2B%nr_of_measurements

    allocate( &
    ! lat_data(1:N_meas),&
    ! lon_data(1:N_meas),&
             meas_index_data(1:N_meas),&
             alt_data_upper(1:N_meas),&
             alt_data_lower(1:N_meas),&
             spectrometer_count_Mie_alt_bin(1:N_meas),&
             spectrometer_count_Mie_int_ref(1:N_meas),&
             spacecraft_los_velocity_meas(1:N_meas),&
             unit_values(1:N_meas),&
             weights(1:N_meas),STAT=AllocStatus)!,&
    IF (AllocStatus .ne. 0) THEN
       error_flag = error_allocate
       return
    END IF

    ! Initialize the unit_values
    do m = 1, size(unit_values)
       unit_values(m) = 1._r8_
    end do

    ! Load Mie_spec_calib_alt_bin and Mie_spec_calib_int_ref
    ! TODO: Put in working DataStructure
    do i_FreqStep = 1, size( Mie_spec_calib_alt_bin )
       MRC_Freq_Step_Res = L1B%Cal_ADS%L1B_Char_Data%Mie_Resp_Calib_Data%List_Mie_Freq_Step_Res(i_FreqStep)%Mie_Freq_Step_Res
       Mie_spec_calib_alt_bin(i_FreqStep)%raw_response = MRC_Freq_Step_Res%Meas_Resp
       Mie_spec_calib_alt_bin(i_FreqStep)%correction   = MRC_Freq_Step_Res%Meas_Err_Mie_Resp
       Mie_spec_calib_int_ref(i_FreqStep)%raw_response = MRC_Freq_Step_Res%Ref_Pulse_Resp
       Mie_spec_calib_int_ref(i_FreqStep)%correction   = MRC_Freq_Step_Res%Ref_Pulse_Err_Mie_Resp
    end do

    ! frequency_to_wind_factor = 0.9_r8_ ! Should be lambda_0 * 0.5_r8_
    print *,'dt: WARNING: frequency_to_wind_factor is placeholder'
    print *,'in Process_Mie_Observations'

    ! point to the map of L1B_Measurements_Used for shorter notation
    Map => L2B%Mie_MDS%Map_of_L1B_Measurements_Used

    ! the nr of range gates is constant for Mie and Rayleigh,
    ! even if their definitions might differ, so this should always work
    rangegateloop: DO r=1, L2B%Obs(1)%nr_of_range_gates
       ! construct the centre-of-gravity values
       ! from the measurements, and store them in the L2B%Obs(:) datastruct

       Nr_Obs     = L2B%Mie_MDS%N_Obs_Mie_Actual

       obs_loop: DO o=1,Nr_Obs

          ! Fill the L2B DEM intersection parameters - but only for rangegate #1
          ! Would be tidier to interchange the rangegateloop and obs_type_loop but
          ! this would change a lot of expected output.  TODO later.
          ! Would also be worth considering DEM_intersection above the
          ! obs_type_level in the IODD, i.e. next to geoid, and fill just once per BRC.
          if ( r == 1 ) then
             L2B%Geoloc_ADS%Mie_Geolocation(o)%Lat_of_DEM_Intersection = &
                 L1B%Geoloc_ADS%Obs_Geoloc%lat_of_DEM_intersection ! [IntAl, 10^-6 degE]
             L2B%Geoloc_ADS%Mie_Geolocation(o)%Lon_of_DEM_Intersection = &
                 L1B%Geoloc_ADS%Obs_Geoloc%lon_of_DEM_intersection ! [IntAl, 10^-6 degE]
             L2B%Geoloc_ADS%Mie_Geolocation(o)%Alt_of_DEM_Intersection = &
                 INT(L1B%Geoloc_ADS%Obs_Geoloc%alt_of_DEM_intersection,IntAl) ! [IntAl, m]
          end if

          ! point to this ObsRangeGate for shorter notation
          WD_MieObsRangeGate => WorkingData%MieObs(o)%RangeGate(r)
       
          ! compose the weights array for this observation class 
          meas_loop1: DO m=1,L2B%nr_of_measurements
             IF (map%Row(m)%Bin(r)%Which_L2B_Obs .eq. o) THEN
                !weights(m) = 1.*Map%Row(m)%Bin(r)%Weight/1000._r8_
                ! use the weights stored as reals for now, to have
                ! higher precision
                weights(m) = WorkingData%MieMeas(m)%RangeGate(r)%weight/1000._r8_
             ELSE
                weights(m) = 0._r8_
             END IF
          END DO meas_loop1
          !
          ! In the old convention, the sum of the real-valued weights would be 
          ! unity and the sum of the integer-valued weights would be 1000._r8_
          ! In the new convention, the max of the real-valued weights should be 
          ! unity and the max of the integer-valued weights sould be 1000._r8_
          ! Care must be taken when using weights in weighted sums, in
          ! particular whether the Normalize argument should be true or false.

          ! copy the lat,lon measurement indices to a simple array
          ! also the alt and geoid data
          ! And any other Mie data?
          meas_loop2: DO m=1,L2B%nr_of_measurements
             !  #[ copy some data to arrays
             meas_index_data(m) = Real( m, r8_ )
             alt_data_upper(m) = WorkingData%MieMeas(m)%RangeGate(r)%alt_end
             alt_data_lower(m) = WorkingData%MieMeas(m)%RangeGate(r)%alt_start

             spacecraft_los_velocity_meas(m) = &
                  L1B%Geoloc_ADS%Meas_Geoloc(m)%AOCS_LOS_Velocity

             spectrometer_count_Mie_alt_bin(m)%spectrometer_count = &
                 L1B%Meas_ADS%Mie_measurement_data(m)%&
                 Mie_spectra_alt_bin(r)%spectrometer_count
             spectrometer_count_Mie_int_ref(m)%spectrometer_count = &
                 L1B%Meas_ADS%Mie_reference_pulses(m)%spectrometer_count

             !  #]
          END DO meas_loop2

          ! Determine the L1B measurement indices for the start and
          ! stop geolocations for the current observation type, both
          ! indices as integers
          call get_index_start_stop(weights, meas_index_start, meas_index_stop)

          ! Take lat and lon from L1B measurement index meas_index_start
          if(.not.missing_int(meas_index_start))then
             WD_MieObsRangeGate%lat_start = WorkingData%MieMeas(meas_index_start)%RangeGate(r)%lat
             WD_MieObsRangeGate%lon_start = WorkingData%MieMeas(meas_index_start)%RangeGate(r)%lon
          end if

          ! Take lat and lon from L1B measurement index meas_index_stop
          if(.not.missing_int(meas_index_stop ))then
             WD_MieObsRangeGate%lat_stop  = WorkingData%MieMeas(meas_index_stop )%RangeGate(r)%lat
             WD_MieObsRangeGate%lon_stop  = WorkingData%MieMeas(meas_index_stop )%RangeGate(r)%lon
          end if

          ! Determine the L1B measurement index for the centre-of-gravity
          ! First as a real number, then as an integer
          meas_index_cog_real = sum_with_weights(meas_index_data,weights,normalize=.true.)
          if(.not.missing_real(meas_index_cog_real))then
             meas_index_cog_int  = nint( meas_index_cog_real )

             ! determine the horizontal centre-of-gravity
             !
             ! Take from L1B measurement index meas_index_cog_int
             WD_MieObsRangeGate%lat_cog = WorkingData%MieMeas(meas_index_cog_int)%RangeGate(r)%lat
             WD_MieObsRangeGate%lon_cog = WorkingData%MieMeas(meas_index_cog_int)%RangeGate(r)%lon

             ! L2B alt_end and alt_start also via copy from L1B measurement index meas_index_cog_int
             geoid_offset =  L1B%Geoloc_ADS%Obs_Geoloc%Geoid_Separation_of_Height_Bin ! 0._r8_ ! Previous Placeholder
             WD_MieObsRangeGate%alt_end   = WorkingData%MieMeas(meas_index_cog_int)%RangeGate(r)%alt_end   - geoid_offset
             WD_MieObsRangeGate%alt_start = WorkingData%MieMeas(meas_index_cog_int)%RangeGate(r)%alt_start - geoid_offset

             ! determine the vertical centre-of-gravity
             !
             ! Average from measurement index meas_index_cog_int
             ! and account for geoid offset
             alt_cog_from_wgs84 = 0.5_r8_ * ( alt_data_upper(meas_index_cog_int)   + &
                                          alt_data_lower(meas_index_cog_int) )
             alt_cog_from_geoid = alt_cog_from_wgs84 - geoid_offset

             WD_MieObsRangeGate%alt_cog = alt_cog_from_geoid
          end if ! missing_real(meas_index_cog_real)

          ! Mie processing

          ! Should test here for some non-missing weights.
          print *,'dt: WARNING: makeshift test for whether Mie '
          print *,'processing is required'
          do_MieWind_processing = .not.missing_real(meas_index_cog_real)

       if(do_MieWind_processing)then

          ! accumulate the spectra, using the proper weights
          !print *,'size(accd_meas)',size(accd_meas)
          !print *,'size(spectrometer_count_Mie_alt_bin%spectrometer_count)',size(spectrometer_count_Mie_alt_bin(1)%spectrometer_count)
          accd_meas   = sum_with_weights_accd(spectrometer_count_Mie_alt_bin,weights,normalize=.false.)
          accd_intref = sum_with_weights_accd(spectrometer_count_Mie_int_ref,weights,normalize=.false.)
          !
          ChoosenVerbosity   = VerbosityError ! VerbosityReport
          ! Parameters for the Mie Core algorithm should come from
          ! the L2B_AuxPar datastructure.  However, some invocations
          ! of the L2B processor do not define this datastructure so
          ! Mie core parameters must be set by other means in such cases
!          if(.not.use_lite_test_data)then
             Mie_Core_Algorithm_Params => &
                 L2B_AuxPar%Data%WVM_Params%Common_Processing_Params%Mie_Core_Algorithm_Params
             SNR_Threshold = Mie_Core_Algorithm_Params%SNR_Threshold
             StartFWHM          = Mie_Core_Algorithm_Params%Start_FWHM
             ResErrThresh       = Mie_Core_Algorithm_Params%Residual_Error_Threshold
             MaxItLorFit        = Mie_Core_Algorithm_Params%Max_Iterations_Lorentz_Fit
             FWHMUpThresh       = Mie_Core_Algorithm_Params%FWHM_Upper_Threshold
             FWHMLoThresh       = Mie_Core_Algorithm_Params%FWHM_Lower_Threshold
             PeakHeightUpThresh = Mie_Core_Algorithm_Params%Peak_Height_Upper_Threshold
             PeakHeightLoThresh = Mie_Core_Algorithm_Params%Peak_Height_Lower_Threshold
             PeakLocThresh      = Mie_Core_Algorithm_Params%Peak_Location_Threshold
             NonLinOptThresh    = Mie_Core_Algorithm_Params%Nonlinear_Optimization_Thresh
             MaxItNonLinOpt     = Mie_Core_Algorithm_Params%Max_Iter_Nonlinear_Optimization
             NumSpecSubSamp     = Mie_Core_Algorithm_Params%Num_Spectral_Sub_Samples
!          else
!             SNR_Threshold      = 8.0_r8_ ! dB
!             StartFWHM          = 5.1_r8_
!             ResErrThresh       = 0.001_r8_
!             MaxItLorFit        = 30
!             FWHMUpThresh       = 2.0_r8_
!             FWHMLoThresh       = 1.7_r8_
!             PeakHeightUpThresh = 1.6_r8_
!             PeakHeightLoThresh = 0.9_r8_
!             PeakLocThresh      = 1.0_r8_
!             NonLinOptThresh    = 0.01_r8_
!             MaxItNonLinOpt     = 20
!             NumSpecSubSamp     = 5
!          endif

          do i_tobs = 1, numUsefulACCDCols
             tobs(i_tobs) = L1B%Cal_ADS%L1B_Char_Data%Sat_Char_Data%List_Tripod_Obscur_Corr(i_tobs)%Tripod_Obscur_Corr
          end do

! TODO: Compute weighted SNR for alt_bin and possibly for int_ref
          SNR_alt_bin = L1B%PCD_ADS%Observation_PCD%Obs_Alt_Bin_PCD(r)%Mie_SNR
          ! SNR_int_ref = ???
          ! Mie_SNR does not exist in L1B for internal reference pulses.
          ! If needed, it must be computed in L2B
if ( ( SNR_alt_bin .gt. SNR_Threshold ) ) then

          PRINT *,'Mie response for atmospheric measurements ...'
 
          isMeas = .true.  ! Triggers TripodObscurationCorrection
          call ComputeMieResponse(ChoosenVerbosity,StartFWHM, &
                          ResErrThresh,MaxItLorFit, FWHMUpThresh, &
                          FWHMLoThresh, PeakHeightUpThresh, &
                          PeakHeightLoThresh, PeakLocThresh, &
                          NonLinOptThresh, MaxItNonLinOpt, &
                          NumSpecSubSamp, tobs, accd_meas, &
                          PeakLoc_meas, FWHM_meas, &
                          Offset_meas, OffsetSub_meas, &
                          PeakHeight_meas, ValFlag_meas, ResErr_meas, &
                          isMeas, NumIt_meas, error_flag_mieresponse)
          !
          ! For a single measurement the average offset is
          ! the total offset from the accumulated measurements divided
          ! by the sum of the weights
          sum_of_weights = sum_with_weights(unit_values,weights)
          OffsetSub_single_meas = OffsetSub_meas / sum_of_weights

          !print *,'FWHM_meas',FWHM_meas
          !print *,'PeakHeight_meas',PeakHeight_meas
          !print *,'PeakLoc_meas',PeakLoc_meas
          !print *,'Offset_meas',Offset_meas
          !print *,'OffsetSub_meas',OffsetSub_meas
          accd_meas_weighted_offset = sum_with_weights_accd(spectrometer_count_Mie_alt_bin,weights**2., &
                                offset=OffsetSub_single_meas,normalize=.false.)
          if (error_flag_mieresponse.ne.no_error)then
!             error_flag = error_flag_mieresponse
!             return
! Poli: changed the error return so that processor does not stop on Mie
! processing errors: TODO: clean this up!!
              cycle obs_loop
          endif
          raw_response_alt_bin = PeakLoc_meas
          !
          ! Save the useful_signal result for use in error quantifier
          ! TODO: check whether FWHM factor is redundant
          useful_signal_weighted = PeakHeight_meas * FWHM_meas
          !
          ! dt: Call to Mie spectral response nonlinearity correction 
          !     for alt_bin
          print *,'skip_Mie_nonlin_correction',skip_Mie_nonlin_correction
          if ( skip_Mie_nonlin_correction ) then
             print *,'dt: WARNING: spectral_nonlinearity_correction not yet'
             PRINT *,'implemented for alt_bin Mie processing'
             !
             ! Placeholder for output of Mie spectral response nonlinearity 
             ! correction for alt_bin
             nonlin_response_alt_bin = raw_response_alt_bin ! 0.35_r_
          else
             call spectral_nonlinear_correction(raw_response_alt_bin,&
                            Mie_spec_calib_alt_bin,nonlin_response_alt_bin)
          endif
          print *," REALACC(6)"
          print *,'nonlin_response_alt_bin =', nonlin_response_alt_bin
          print *,"ENDREALACC"

          PRINT *,'Mie response for internal reference ...'
!        if ( SNR_int_ref .gt. SNR_Threshold ) then
          isMeas = .false. ! Omits TripodObscurationCorrection
          call ComputeMieResponse(ChoosenVerbosity,StartFWHM, &
                          ResErrThresh,MaxItLorFit, FWHMUpThresh, &
                          FWHMLoThresh, PeakHeightUpThresh, &
                          PeakHeightLoThresh, PeakLocThresh, &
                          NonLinOptThresh, MaxItNonLinOpt, &
                          NumSpecSubSamp, tobs, accd_intref, &
                          PeakLoc_ref, FWHM_ref, &
                          Offset_ref, OffsetSub_ref, &
                          PeakHeight_ref, ValFlag_ref, ResErr_ref, &
                          isMeas, NumIt_ref, error_flag_mieresponse)
          if (error_flag_mieresponse.ne.no_error)then
!             error_flag = error_flag_mieresponse
!             return
! Poli: changed the error return so that processor does not stop on Mie
! processing errors: TODO: clean this up!!
              cycle obs_loop
          endif
          raw_response_int_ref = PeakLoc_ref

          ! dt: Call to Mie spectral response nonlinearity correction 
          !     for int_ref
          if ( skip_Mie_nonlin_correction ) then
             print *,'dt: WARNING: spectral_nonlinearity_correction not yet'
             PRINT *,'implemented for int_ref Mie processing'
             ! Placeholders for output of Mie spectral response nonlinearity 
             ! correction
             nonlin_response_int_ref = raw_response_int_ref ! 0.015_r_
          else
             call spectral_nonlinear_correction(raw_response_int_ref,&
                            Mie_spec_calib_int_ref,nonlin_response_int_ref)
          endif
          print *," REALACC(6)"
          print *,'nonlin_response_int_ref =', nonlin_response_int_ref
          print *,"ENDREALACC"

!        else
!          print *, 'No int_ref Mie response computed because SNR_Threshold not met.'
!          print *,"REALACC(6)"
!          print *, 'SNR_int_ref, SNR_Threshold =', SNR_int_ref, SNR_Threshold
!          print *,"ENDREALACC"
!          print *,'Assuming zero response for internal reference'
!          raw_response_int_ref = 0._r8_
!          nonlin_response_int_ref = 0._r8_
!        end if
!

          ! temporary placeholders
          spacecraft_los_velocity = 0._r8_
          mie_ground_corr_velfactor = 0._r8_
          mie_ground_corr_weighting = 0._r8_
          internal_reference_los_velocity = 0._r8_

          ! It seems sufficient to set the zero frequency terms to
          ! zero in accordance with the L1B Mie algorithm.
          ! Using the L1B parameters introduces a wind error of around
          ! 11 m/s in basic tests conducted in June 2006.
          internal_reference_zero_freq = &
             0._r8_ ! L1B%Cal_ADS%L1B_Char_Data%Mie_Resp_Calib_Data%Mie_Ref_Pulse_Resp_Calib%Ref_Pulse_Zero_Freq
          measurement_zero_freq = &
             0._r8_ ! L1B%Cal_ADS%L1B_Char_Data%Mie_Resp_Calib_Data%Mie_Meas_Resp_Calib%Meas_Zero_Freq

!         Convert response slope from units PixelIndex/GHz to PixelIndex/Hz.
          ! Poli: with L1B v1.04 ALL FREQUENCIES are in GHz
          mie_response_slope_int_ref = 1.e-9_r8_ * &
          & L1B%Cal_ADS%L1B_Char_Data%Mie_Resp_Calib_Data%Mie_Ref_Pulse_Resp_Calib%Ref_Pulse_Mean_Sensitivity
          mie_response_slope_alt_bin = 1.e-9_r8_ * &
          & L1B%Cal_ADS%L1B_Char_Data%Mie_Resp_Calib_Data%Mie_Meas_Resp_Calib%Meas_Mean_Sensitivity

          ! Update placeholders
          elevation_of_height_bin = L1B%Geoloc_ADS%Obs_Geoloc%&
                 Mie_Rangebin_Geoloc(r)%topoc_elev_of_height_bin
          ! the angle reported in L1b data is elevation (about 53 degrees)
          incidence_angle = (90._r8_ - abs(elevation_of_height_bin)) * &
                            & pi/180._r8_
          sin_incidence_angle = sin(incidence_angle)

          spacecraft_los_velocity = &
               sum_with_weights(spacecraft_los_velocity_meas,weights,normalize=.true.)
          mie_ground_corr_velfactor = L1B%GWD_ADS%Mie_ground_corr_velocity
          mie_ground_corr_weighting = L1B%GWD_ADS%Mie_grnd_corr_wght_factor
          internal_reference_los_velocity = frequency_to_wind_factor * &
               ( nonlin_response_int_ref - internal_reference_zero_freq ) / mie_response_slope_int_ref

          mie_los_velocity =  frequency_to_wind_factor * &
               ( nonlin_response_alt_bin - measurement_zero_freq ) / mie_response_slope_alt_bin 

          print *, 'REALACC(6)'
          print *, 'mie_los_velocity = ', mie_los_velocity
          print *, 'spacecraft_los_velocity = ', spacecraft_los_velocity
          print *, 'mie_ground_corr_velfactor = ', mie_ground_corr_velfactor
          print *, 'mie_ground_corr_weighting = ', mie_ground_corr_weighting
          print *, 'internal_reference_los_velocity = ', internal_reference_los_velocity
          print *, 'ENDREALACC'

          ! It might be possible for the following to share a common method 
          ! with Rayleigh processing
          !corrected_los_velocity = mie_los_velocity - &
          !     spacecraft_los_velocity - &
          !     mie_ground_corr_velfactor * &
          !     mie_ground_corr_weighting - &
          !     internal_reference_los_velocity
          
          ! According to ADM-MA-52-1800, Version 1.2, a sign inversion
          ! has been observed and so measurement and internal reference
          ! terms are interchanged!
          !
          corrected_los_velocity = internal_reference_los_velocity - &
               spacecraft_los_velocity - &
               mie_ground_corr_velfactor * &
               mie_ground_corr_weighting - &
               mie_los_velocity

          WD_MieObsRangeGate%hlos_cog = &
               corrected_los_velocity / sin_incidence_angle

            ! L1B elements and L2B results that do not require scaling 
          ! are copied directly L2B results that must be scaled before
          ! writing them to file are written in the WorkingData structure first
          ! and copied in Format_and_Copy_WD_to_L2B_Outp()
          ! (I don't like this difference much. I think it would be much easier and
          !  cleaner to just create working data elements for all results we
          !  wish to export, and copy them all in Format_and_Copy_WD_to_L2B_Outp()
          !  This prevents mistakes I think)

                   ! copy all obtained geolocation results to the Geoloc_ADS datastruct
                   L2B_Geoloc_Bin => L2B%Geoloc_ADS%Mie_Geolocation(o)%Height_Bin(r)
          
                   L2B_Geoloc_Bin%LOS_Azimuth      = L1B%Geoloc_ADS%&
                        Obs_Geoloc%Mie_Rangebin_Geoloc(r)%topoc_azim_of_height_bin ! [deg]
                   L2B_Geoloc_Bin%LOS_Elevation    = elevation_of_height_bin  ! [deg]
                   L2B_Geoloc_Bin%LOS_Sat_Velocity = spacecraft_los_velocity  ! [m/s]

          ! To do: Computation of PCD parameters
          !        and copying to PCD_ADS
          !        Some may be done in Construct_Mie_PCD?
          !         hlos_error = 5. ! Placeholder

                   print *,"REALACC(6)"
                   print *,'PeakHeight',PeakHeight_meas
                   print *,"ENDREALACC"
                   hlos_error = bin_MieErrQuant(L1B, &
!                                                weights, & ! never used
!                                                r, & ! never used
                                                sin_incidence_angle, &
                                                tobs, &
                                                accd_meas_weighted_offset, &
                                                PeakLoc_meas, &
                                                PeakHeight_meas, &
                                                FWHM_meas, &
                                                mie_response_slope_alt_bin, &
                                                useful_signal_weighted ) ! FAdoxy, m/s
                   Mie_PCD_Bin => L2B%PCD_ADS%Mie_Processing_QC%Obs_Mie_QC(o)%Obs_Mie_Bin_QC(r)
                   Mie_PCD_Bin%Hlos_Error_Estimate       = hlos_error ! FAdoxy, m/s

          ! copy all obtained results to the Mie_MDS datastruct
          Bin => L2B%Mie_MDS%Mie_Profile(o)%Height_Bin(r)
          Bin%Validity_Flag                   =.true.
          ! convert from m/s to cm/s and convert to IntAs type
          Bin%Mie_Wind_Velocity               = &
               floor(WD_MieObsRangeGate%hlos_cog*100,IntAs)
          !Bin%Mie_Error_Quantifier            = 0
          !Bin%Integration_Length              = 0          
          
          print *,"Mie Obs. nr. ",o
          print *,"Mie Range Gate nr. ",r
          print *," REALACC(6)"
          print *,"lat_cog = ",WD_MieObsRangeGate%lat_cog
          print *,"lon_cog = ",WD_MieObsRangeGate%lon_cog
          print *,"Mie_hls = ",WD_MieObsRangeGate%hlos_cog
          print *,"Mie_err = ",Mie_PCD_Bin%Hlos_Error_Estimate
          print *,"ENDREALACC"
else

   ! remark: the SNR_alt_bin number comes directly from the L1B input file 
   !         without any processing, so the infinite values we see are really
   !         a L1BP problem, not ours. Even so, it is better to use the
   !         SetFAdoFull_or_Missing() function to convert reals to txt
   !         in stead of the real2txt() function combined with to_lowercase()
          call SetFadoFull_or_Missing(SNR_alt_bin,SNR_alt_bin_txt,error_flag)
          if (error_flag .ne. no_error) return
          print *, 'No alt_bin Mie response computed because SNR_Threshold not met.'
          print *," REALACC(6)"
          print *, 'SNR_alt_bin, SNR_Threshold = ', &
                    trim(SNR_alt_bin_txt), SNR_Threshold
          print *,"ENDREALACC"
end if ! SNR_Threshold exceeded
       end if ! do_MieWind_processing

       END DO obs_loop     
    END DO rangegateloop

    ! Fill the TimeStamp and GeoidOffset of the GeolocDSR, even if
    ! it overwrites values from Process_Rayleigh_Observations

    L2B%Geoloc_ADS%Start_of_Obs_DateTime = &
        L1B%Geoloc_ADS%start_of_obs_datetime
    L2B%Geoloc_ADS%WGS84_to_Geoid_Altitude = &
        FAdoxy_to_IntAl(geoid_offset) ! [IntAl, m]

  end subroutine Process_Mie_Observations

    !  #]
  subroutine determine_vcog(WorkingData, error_flag)
    !  #[  to be added later (this is just a first try)
    !
    ! try to determine whether most of the photons came from
    ! the top or bottom region of this layer, using the extinction
    ! value for this layer.
    ! ==> high extinction: most photons came from the upper part of the layer
    ! ==> low extinction:  photons came from the complete layer,
    !     center of gravity depends on the backscatter profile
    !     (do we have a way to estimate it ? maybe look at surrounding layers?)

    ! for now:
    ! this sets the vcog altitudes halfway in the range-bins.
    ! which is used to determine the altitudes for the NWP preprocessing

    TYPE(WorkingDataType), intent(inout) :: WorkingData ! input/output
    integer,               intent(out)   :: error_flag ! output

    ! local variables
    integer :: i_meas, i_rangegate, nr_rangegates

    error_flag = no_error
    measloop: DO i_meas = 1,WorkingData%nr_of_measurements
       nr_rangegates = WorkingData%RayleighMeas(i_meas)%nr_of_range_gates
       rangegateloop: DO i_rangegate=1, nr_rangegates
          WorkingData%RayleighMeas(i_meas)%Rangegate(i_rangegate)%alt_vcog = &
               0.5_r8_*WorkingData%RayleighMeas(i_meas)%Rangegate(i_rangegate)%alt_start+&
               0.5_r8_*WorkingData%RayleighMeas(i_meas)%Rangegate(i_rangegate)%alt_end
          WorkingData%MieMeas(i_meas)%Rangegate(i_rangegate)%alt_vcog = &
               0.5_r8_*WorkingData%MieMeas(i_meas)%Rangegate(i_rangegate)%alt_start+&
               0.5_r8_*WorkingData%MieMeas(i_meas)%Rangegate(i_rangegate)%alt_end
       END DO rangegateloop
    END DO measloop

  end subroutine determine_vcog
    !  #]
  !-------------------------------------------------
end module HLOS_retrieval
