module screening_l1b_data
  !  #[ Description:
  !
  !     A module to screen all elements of the L1B product
  !     of the Aeolus instrument before actually using them.
  !     
  !     [is the following still true? JK]
  !     This also includes setting the value to the missing_indicator
  !     in case the corresponding flag indicates an invalid value
  !     that should not be used.
  !
  ! created:     08-Jun-2006 by J. de Kloe
  !
  ! Modifications: 
  ! 30-Jun-2006 G.J. Marseille  new generic subroutine ThresholdCheck_IntAl
  !                             new generic subroutine ThresholdCheck_FAdoxy
  ! 04-Jul-2006 G.J. Marseille  addition of a string parameter to the header of
  !                             the generic subroutines (see 30-jun-2006 mod.)
  ! 09-Jul-2006 G.J. Marseille  Implementation of double loop, i.e. for 
  !                             screening over measurement scale and range bin
  !                             scale for Mie and Rayleigh data
  ! 12-Jul-2006 G.J. Marseille  The following parameters only have a lower 
  !                             threshold: Mie_snr, ...........
  !                             To screen parameters with a lower threshold 
  !                             only, the upper threshold, that is passed to 
  !                             the generic routine, is set to  the parameter 
  !                             value since this check always passes.
  ! 14-Jul-2006 G.J. Marseille  1) Checking L1B Mie_invalid and Rayl_invalid flags
  !                             2) Two new parameters, meas_string and bin_string,
  !                             to make logmsg unique for the different tests,
  !                             i.e. logmsg as a function of measurement number 
  !                             and range bin level.
  !                             3) character length of string_param increased from 
  !                             31 to 70
  ! 27-Jul-2006 P. Poli         1) Correct L1B parameter names in logmsg calls
  !                             2) Comment out obs-bin level screening
  ! 08-Aug-2006 D. Tan          1) Add realacc_on for logmsg calls involving FAdoxy
  !                             2) Improve formatting of text strings
  ! 18-Aug-2006 G.J. Marseille  Re-shuffling of the order of tests. The order is now 
  !                             in line with the ECMWF L2BP IODD document:
  !                             AE-IF-ECMWF-L2BP-001_IODD_Iss_ 1.0_draft_060217.pdf 
  ! 24-Jul-2007 J. de Kloe      adapted ThresholdCheck_FAdoxy() to allow it to
  !                             silently accept missing values if needed.
  !                             Also adapted both ThresholdCheck routines to return
  !                             a CheckResultType variable, which allows returning
  !                             the error_flag in a proper way as well.
  !                             Added checking of the error_flag in many places
  !                             Added a number of folding marks
  ! 07-Aug-2007 D. Tan          Fix SPR#20 (extra underscores)
  ! 04-Sep-2007 P. Poli         change L1B_DataType to L1B_BRC_DataType
  ! 16-Jan-2008 J. de Kloe      phase out integer kind i_
  ! 
  !  #]
  !  #[ Modules used:

  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE L1B_DataStructure, only: L1B_BRC_DataType !, L1B_SPH_type
  USE L2B_PCD_ADS, only: L1B_Input_Screening_type, &
                         L1B_File_OK, L1B_Laser_Freq_Unlocked, &
                         L1B_Ref_Pulses_Unlocked,L1B_Sat_Not_on_Target, &
                         L1B_Mie_Invalid_Ref_Pulses,L1B_Mie_Corrupt,&
                         L1B_Rayleigh_Corrupt,L1B_Mie_Ref_Pulses_Corrupt,&
                         L1B_Rayl_Ref_Pulses_Corrupt,L1B_Laser_Freq_Offset,&
                         L1B_Laser_UV_Energy,L1B_Laser_Freq_Offs_Stdev,&
                         L1B_Laser_UV_Energy_Stdev,L1B_Mie_Invalid_Meas,&
                         L1B_Mie_Invalid_Ref_Pulses,L1B_Rayl_Invalid_Meas,&
                         L1B_Rayl_Invalid_Ref_Pulses,L1B_Mie_Mean_Emit_Freq,&
                         L1B_Mie_Emit_Freq_Stdev,L1B_Rayl_Mean_Emit_Freq,&
                         L1B_Rayl_Emit_Freq_Stdev,L1B_Mie_Scattering_Ratio,&
                         L1B_Mie_SNR,L1B_Rayl_SNR_A,L1B_Rayl_SNR_B,&
                         L1B_Mie_Meas_OK,L1B_Rayl_Meas_OK,&
                         L1B_Rayl_Bin_OK,&
                         L1B_Mie_Bin_OK,L1B_Mie_SNR,L1B_Mie_Meas_Invalid_Ref_Pulses,&
                         L1B_Mie_Avg_UV_Energy,L1B_Mie_Vel_of_Att_Uncert_Error,&
                         L1B_Mie_Mean_Emitted_Freq,L1B_Mie_Avg_Laser_Freq_Offset,&
                         L1B_Mie_Laser_Freq_Offset_Stdev,L1B_Mie_UV_Energy_Stdev,&
                         L1B_Rayl_Meas_Invalid_Ref_Pls,L1B_Rayl_Avg_Laser_Freq_Offset,&
                         L1B_Rayl_Avg_UV_Energy,L1B_Rayl_Laser_Freq_Offs_Stdev,&
                         L1B_Rayl_UV_Energy_Stdev,L1B_Rayl_Vel_of_Att_Uncert_Err,&
                         L1B_Rayl_Mean_Emitted_Freq,L1B_Rayl_Emitted_Freq_Stdev,&
                         L1B_Mie_Bin_Invalid,L1B_Rayl_Bin_Invalid, &
                         L1B_Invalid_Num_Mie_Peaks
  USE L2B_AuxPar_DataStructure, only: L1B_Screening_Params_type

  USE L1B_Geoloc_ADS, only: L1B_Geoloc_ADS_type
  USE L1B_PCD_ADS, only: L1B_PCD_ADS_type

! todo: add screening for all these datatypes:
!  USE L1B_GWD_ADS, only: L1B_GWD_ADS_type
!  USE L1B_Meas_ADS, only: L1B_Meas_ADS_type
!  USE L1B_US_MDS, only: L1B_US_MDS_type
!  USE L1B_WV_MDS, only: L1B_WV_MDS_type
!  USE L1B_Cal_ADS, only: L1B_Cal_ADS_type
!  USE FixedHeader, only: FH_type
!  USE MainProductHeader, only: MPH_type

  USE Working_DataStructure, only: WorkingDataType
  USE EE_CFI_Datatypes, only: IntAl, SetIntAl, missing_indicator_IntAl,&
                              FAdoxy, missing_indicator_FAdoxy,&
                              SetFAdoFull_or_Missing,&
                              maxlen_IntAl,&
                              IntAuc
  USE AeolusConstants, only: N_range_gates
  USE Numerics, only: r8_, missing_real8
  USE ErrorHandler, only: no_error, error_programming
  USE Logging, only: logmsg,log_debug,log_info,log_progress,log_warn,log_error

  !  #]  
  !  #[ Parameter/Variable/Type declarations

  implicit none
   
  ! a quick hack to suppress the huge amount of output of this module
  ! in case I want to test with the TLE 
  ! (set it to .true. if you want to suppress the output)
  logical, parameter :: Silent = .false.

  ! this type allows using the ThresholdChecks as function, and makes it
  ! possible to export the 2 values anyway. Note that the previous solution
  ! (i.e. changing one of the input parameters of the function) is bad 
  ! coding practice and should be avoided.
  type CheckResultType
     logical :: CheckOK
     integer :: error_flag
  end type CheckResultType

  ! make all module contents private, unless explicitely stated otherwise
  private

  !  #]
  !  #[ Module interface
  public :: screen_l1b_data
  !  #]
contains
  !-----------------------------
  subroutine InitCheckResult(CheckResult)
    !  #[
    type(CheckResultType), intent(out) :: CheckResult

    CheckResult%CheckOK    = .true.
    CheckResult%error_flag = no_error

  end subroutine InitCheckResult
    !  #]
  !-----------------------------
  subroutine L1B_PCD_ADS_Screening(L1B_PCD_ADS,WorkingData,L1B_Screening,&
                                   L1B_Screening_Params,screening_type,&
                                   error_flag)
    !  #[
    TYPE(L1B_PCD_ADS_type),          intent(inout) :: L1B_PCD_ADS
    TYPE(WorkingDataType),           intent(inout) :: WorkingData
    TYPE(L1B_Input_Screening_type),  intent(inout) :: L1B_Screening
    TYPE(L1B_Screening_Params_type), intent(in)    :: L1B_Screening_Params
    character(31),                   intent(in)    :: screening_type
    integer,                         intent(out)   :: error_flag

    ! local variables
    integer(IntAl)          :: val_IntAl, val_IntAl_thr
    real(FAdoxy)            :: val_FAdoxy, val_FAdoxy_thr
    real(FAdoxy)            :: val_FAdoxy_min, val_FAdoxy_max
    integer                 :: r,nmeas,m
    character(len=70)       :: string_param
    character(len=50)       :: meas_string,bin_string
    logical                 :: val_logical
    type(CheckResultType)   :: CheckResult

    print *,"L1B_PCD_ADS  screening not yet fully implemented"

    ! see l1b_pcd_ads.F90 for the definition of the L1B%PCD_ADS%...
    ! input elements (L1B IODD tables ... to ...)
    ! see l2b_auxpar_datastructure.F90 for the definition of the 
    ! thresholds in L2B_AuxPar%Data%Screening_Params%L1B_Screening_Params%...
    ! (L2B IODD tables ... to ...)
    ! see l2b_pcd_ads.F90 for the definition of the L1B_Screening%...
    ! output fields (L2B IODD tables ... to ...)

    ! init
    error_flag = no_error

    ! ===============  START SCREENING  ========================
    !
    ! ===========  1. Set L1B_Screening%Obs_Screening by screening
    !                 L1B PCD Observation-Scale data
    !
    ! Section 1.1 - Screening of scalar parameters
    !               (Parameters from L2B ECMWF IODD document table 18 are screened)
    !
    ! initialize L1B_Screening%Obs_Screening
    L1B_Screening%Obs_Screening = L1B_File_OK

    ! NOTE: the order of these test is different form the order in
    !       which these items are defined in the L1B IODD.
    !       Leading is the numbering used for the enum type defined
    !       in l2b_pcd_ads.F90 to define the numbers stored in 
    !       L1B_Screening%Obs_Screening

    !  #[ screen: nr_meas_laser_freq_unlocked
    val_IntAl     = L1B_PCD_ADS%Observation_PCD%nr_meas_laser_freq_unlocked
    val_IntAl_thr = L1B_Screening_Params%L1B_Obs_Screening_Params%&
                    L1B_Laser_Freq_Unlocked_Thresh

    string_param  = "Num_Measurement_Laser_Freq_Unlocked, Obs level"
    CheckResult=ThresholdCheck_IntAl(val_IntAl,string_param,screening_type,&
                                     upper_thr=val_IntAl_thr,lower_thr=0_IntAl)
    if (CheckResult%error_flag .ne. no_error) return
    if (.not. CheckResult%CheckOK) then
       ! store the failed check, but only if no other Obs_Level checks failed before
       IF (L1B_Screening%Obs_Screening .eq. L1B_File_OK) &
            L1B_Screening%Obs_Screening = L1B_Laser_Freq_Unlocked
    endif
    !  #]
    !  #[ screen: nr_ref_laser_freq_unlocked
    val_IntAl     = L1B_PCD_ADS%Observation_PCD%nr_ref_laser_freq_unlocked
    val_IntAl_thr = L1B_Screening_Params%L1B_Obs_Screening_Params%&
                    L1B_Ref_Pulses_Unlocked_Thresh

    string_param  = "Num_Reference_Pulse_Laser_Freq_Unlocked, Obs level"
    CheckResult=ThresholdCheck_IntAl(val_IntAl,string_param,screening_type,&
                                     upper_thr=val_IntAl_thr,lower_thr=0_IntAl)
    if (CheckResult%error_flag .ne. no_error) return
    if (.not. CheckResult%CheckOK) then
       ! store the failed check, but only if no other Obs_Level checks failed before
       IF (L1B_Screening%Obs_Screening .eq. L1B_File_OK) &
            L1B_Screening%Obs_Screening = L1B_Ref_Pulses_Unlocked
    endif
    !  #]
    !  #[ screen: avg_laser_freq_offset
    val_FAdoxy     = L1B_PCD_ADS%Observation_PCD%avg_laser_freq_offset
    val_FAdoxy_thr = L1B_Screening_Params%L1B_Obs_Screening_Params%&
                         L1B_Laser_Freq_Offset_Thresh

    string_param  = "Avg_Laser_Frequency_Offset, Obs level"
    CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                      upper_thr=val_FAdoxy_thr,lower_thr=0._FAdoxy)
    if (CheckResult%error_flag .ne. no_error) return
    if (.not. CheckResult%CheckOK) then
       ! store the failed check, but only if no other Obs_Level checks failed before
       IF (L1B_Screening%Obs_Screening .eq. L1B_File_OK) &
            L1B_Screening%Obs_Screening = L1B_Laser_Freq_Offset
    endif
    !  #]
    !  #[ screen: avg_UV_energy
    val_FAdoxy     = L1B_PCD_ADS%Observation_PCD%avg_UV_energy
    val_FAdoxy_thr = L1B_Screening_Params%L1B_Obs_Screening_Params%&
                     L1B_Laser_UV_Energy_Thresh

    string_param  = "Avg_UV_Energy, Obs level"
    CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                      upper_thr=val_FAdoxy_thr,lower_thr=0._FAdoxy)
    if (CheckResult%error_flag .ne. no_error) return
    if (.not. CheckResult%CheckOK) then
       ! store the failed check, but only if no other Obs_Level checks failed before
       IF (L1B_Screening%Obs_Screening .eq. L1B_File_OK) &
            L1B_Screening%Obs_Screening = L1B_Laser_UV_Energy
    endif
    !  #]
    !  #[ screen: laser_freq_offset_std_dev
    val_FAdoxy     = L1B_PCD_ADS%Observation_PCD%laser_freq_offset_std_dev
    val_FAdoxy_thr = L1B_Screening_Params%L1B_Obs_Screening_Params%&
                         L1B_Laser_Freq_Offs_StdevThresh

    string_param  = "Laser_Frequency_Offset_Std_Dev, Obs level"
    CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                      upper_thr=val_FAdoxy_thr,lower_thr=0._FAdoxy)
    if (CheckResult%error_flag .ne. no_error) return
    if (.not. CheckResult%CheckOK) then
       ! store the failed check, but only if no other Obs_Level checks failed before
       IF (L1B_Screening%Obs_Screening .eq. L1B_File_OK) &
            L1B_Screening%Obs_Screening = L1B_Laser_Freq_Offs_Stdev
    endif
    !  #]
    !  #[ screen: UV_energy_std_dev
    val_FAdoxy     = L1B_PCD_ADS%Observation_PCD%UV_energy_std_dev
    val_FAdoxy_thr = L1B_Screening_Params%L1B_Obs_Screening_Params%&
                         L1B_Laser_UV_Energy_StdevThresh

    string_param  = "UV_Energy_Std_Dev, Obs level"
    CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                      upper_thr=val_FAdoxy_thr,lower_thr=0._FAdoxy)
    if (CheckResult%error_flag .ne. no_error) return
    if (.not. CheckResult%CheckOK) then
       ! store the failed check, but only if no other Obs_Level checks failed before
       IF (L1B_Screening%Obs_Screening .eq. L1B_File_OK) &
            L1B_Screening%Obs_Screening = L1B_Laser_UV_Energy_Stdev
    endif
    !  #]
    !  #[ screen: mie_mean_emitted_freq
    val_FAdoxy     = L1B_PCD_ADS%Observation_PCD%mie_mean_emitted_freq
    val_FAdoxy_min = L1B_Screening_Params%L1B_Obs_Screening_Params%&
                         L1B_Mie_Mean_Emit_Freq_Min
    val_FAdoxy_max = L1B_Screening_Params%L1B_Obs_Screening_Params%&
                         L1B_Mie_Mean_Emit_Freq_Max

    string_param  = "Mie_Mean_Emitted_Frequency, Obs level"
    CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                      upper_thr=val_FAdoxy_max,lower_thr=val_FAdoxy_min)
    if (CheckResult%error_flag .ne. no_error) return
    if (.not. CheckResult%CheckOK) then
       ! store the failed check, but only if no other Obs_Level checks failed before
       IF (L1B_Screening%Obs_Screening .eq. L1B_File_OK) &
            L1B_Screening%Obs_Screening = L1B_Mie_Mean_Emit_Freq
    endif
    !  #]
    !  #[ screen: mie_emitted_freq_std_dev
    val_FAdoxy     = L1B_PCD_ADS%Observation_PCD%mie_emitted_freq_std_dev
    val_FAdoxy_thr = L1B_Screening_Params%L1B_Obs_Screening_Params%&
                         L1B_Mie_Emit_Freq_Stdev_Thresh

    string_param  = "Mie_Emitted_Frequency_Std_Dev, Obs level"
    CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                      upper_thr=val_FAdoxy_thr,lower_thr=0._FAdoxy)
    if (CheckResult%error_flag .ne. no_error) return
    if (.not. CheckResult%CheckOK) then
       ! store the failed check, but only if no other Obs_Level checks failed before
       IF (L1B_Screening%Obs_Screening .eq. L1B_File_OK) &
            L1B_Screening%Obs_Screening = L1B_Mie_Emit_Freq_Stdev
    endif
    !  #]
    !  #[ screen: rayleigh_mean_emitted_freq
    val_FAdoxy     = L1B_PCD_ADS%Observation_PCD%rayleigh_mean_emitted_freq
    val_FAdoxy_min = L1B_Screening_Params%L1B_Obs_Screening_Params%&
                         L1B_Rayleigh_Mean_Emit_Freq_Min
    val_FAdoxy_max = L1B_Screening_Params%L1B_Obs_Screening_Params%&
                         L1B_Rayleigh_Mean_Emit_Freq_Max

    string_param  = "Rayleigh_Mean_Emitted_Frequency, Obs level"
    CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                      upper_thr=val_FAdoxy_max,lower_thr=val_FAdoxy_min)
    if (CheckResult%error_flag .ne. no_error) return
    if (.not. CheckResult%CheckOK) then
       ! store the failed check, but only if no other Obs_Level checks failed before
       IF (L1B_Screening%Obs_Screening .eq. L1B_File_OK) &
            L1B_Screening%Obs_Screening = L1B_Rayl_Mean_Emit_Freq
    endif
    !  #]
    !  #[ screen: rayleigh_emitted_freq_std_dev
    val_FAdoxy     = L1B_PCD_ADS%Observation_PCD%rayleigh_emitted_freq_std_dev
    val_FAdoxy_thr = L1B_Screening_Params%L1B_Obs_Screening_Params%&
                         L1B_RayleighEmitFreqStdevThresh

    string_param  = "Rayleigh_Emitted_Frequency_Std_Dev, Obs level"
    CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                      upper_thr=val_FAdoxy_thr,lower_thr=0._FAdoxy)
    if (CheckResult%error_flag .ne. no_error) return
    if (.not. CheckResult%CheckOK) then
       ! store the failed check, but only if no other Obs_Level checks failed before
       IF (L1B_Screening%Obs_Screening .eq. L1B_File_OK) &
            L1B_Screening%Obs_Screening = L1B_Rayl_Emit_Freq_Stdev
    endif
    !  #]
    !  #[ screen: nr_sat_not_on_target_meas
    val_IntAl     = L1B_PCD_ADS%Observation_PCD%nr_sat_not_on_target_meas
    val_IntAl_thr = L1B_Screening_Params%L1B_Obs_Screening_Params%&
                    L1B_Sat_Not_on_Target_Thresh

    string_param  = "Num_Sat_Not_on_Target_Measurements, Obs level"
    CheckResult=ThresholdCheck_IntAl(val_IntAl,string_param,screening_type,&
                                     upper_thr=val_IntAl_thr,lower_thr=0_IntAl)
    if (CheckResult%error_flag .ne. no_error) return
    if (.not. CheckResult%CheckOK) then
       ! store the failed check, but only if no other Obs_Level checks failed before
       IF (L1B_Screening%Obs_Screening .eq. L1B_File_OK) &
            L1B_Screening%Obs_Screening = L1B_Sat_Not_on_Target
    endif
    !  #]
    !  #[ screen: nr_corrupt_mie_meas
    val_IntAl     = L1B_PCD_ADS%Observation_PCD%nr_corrupt_mie_meas
    val_IntAl_thr = L1B_Screening_Params%L1B_Obs_Screening_Params%&
                    L1B_Mie_Corrupt_Thresh

    string_param  = "Num_Corrupt_Mie_Measurements, Obs level"
    CheckResult=ThresholdCheck_IntAl(val_IntAl,string_param,screening_type,&
                                     upper_thr=val_IntAl_thr,lower_thr=0_IntAl)

    if (CheckResult%error_flag .ne. no_error) return
    if (.not. CheckResult%CheckOK) then
       ! store the failed check, but only if no other Obs_Level checks failed before
       IF (L1B_Screening%Obs_Screening .eq. L1B_File_OK) &
            L1B_Screening%Obs_Screening = L1B_Mie_Corrupt
    endif
    !  #]
    !  #[ screen: nr_corrupt_rayleigh_meas
    val_IntAl     = L1B_PCD_ADS%Observation_PCD%nr_corrupt_rayleigh_meas
    val_IntAl_thr = L1B_Screening_Params%L1B_Obs_Screening_Params%&
                       L1B_Rayleigh_Corrupt_Thresh

    string_param  = "Num_Corrupt_Rayleigh_Measurements, Obs level"
    CheckResult=ThresholdCheck_IntAl(val_IntAl,string_param,screening_type,&
                                     upper_thr=val_IntAl_thr,lower_thr=0_IntAl)
    if (CheckResult%error_flag .ne. no_error) return
    if (.not. CheckResult%CheckOK) then
       ! store the failed check, but only if no other Obs_Level checks failed before
       IF (L1B_Screening%Obs_Screening .eq. L1B_File_OK) &
            L1B_Screening%Obs_Screening = L1B_Rayleigh_Corrupt
    endif
    !  #]
    !  #[ screen: nr_corrupt_mie_ref_pulses
    val_IntAl     = L1B_PCD_ADS%Observation_PCD%nr_corrupt_mie_ref_pulses
    val_IntAl_thr = L1B_Screening_Params%L1B_Obs_Screening_Params%&
                       L1B_Mie_Ref_PulsesCorruptThresh

    string_param  = "Num_Corrupt_Mie_Reference_Pulses, Obs level"
    CheckResult=ThresholdCheck_IntAl(val_IntAl,string_param,screening_type,&
                                     upper_thr=val_IntAl_thr,lower_thr=0_IntAl)
    if (CheckResult%error_flag .ne. no_error) return
    if (.not. CheckResult%CheckOK) then
       ! store the failed check, but only if no other Obs_Level checks failed before
       IF (L1B_Screening%Obs_Screening .eq. L1B_File_OK) &
            L1B_Screening%Obs_Screening = L1B_Mie_Ref_Pulses_Corrupt
    endif
    !  #]
    !  #[ screen: nr_corrupt_rayleigh_ref_pulses
    val_IntAl     = L1B_PCD_ADS%Observation_PCD%nr_corrupt_rayleigh_ref_pulses
    val_IntAl_thr = L1B_Screening_Params%L1B_Obs_Screening_Params%&
                       L1B_Rayl_RefPulsesCorruptThresh

    string_param  = "Num_Corrupt_Rayleigh_Reference_Pulses, Obs level"
    CheckResult=ThresholdCheck_IntAl(val_IntAl,string_param,screening_type,&
                                     upper_thr=val_IntAl_thr,lower_thr=0_IntAl)
    if (CheckResult%error_flag .ne. no_error) return
    if (.not. CheckResult%CheckOK) then
       ! store the failed check, but only if no other Obs_Level checks failed before
       IF (L1B_Screening%Obs_Screening .eq. L1B_File_OK) &
            L1B_Screening%Obs_Screening = L1B_Rayl_Ref_Pulses_Corrupt
    endif
    !  #]
    !  #[ screen: nr_mie_invalid_meas
    val_IntAl     = L1B_PCD_ADS%Observation_PCD%nr_mie_invalid_meas
    val_IntAl_thr = L1B_Screening_Params%L1B_Obs_Screening_Params%&
                       L1B_Mie_Invalid_Meas_Thresh

    string_param  = "Num_Mie_Invalid_Measurements, Obs level"
    CheckResult=ThresholdCheck_IntAl(val_IntAl,string_param,screening_type,&
                                     upper_thr=val_IntAl_thr,lower_thr=0_IntAl)
    if (CheckResult%error_flag .ne. no_error) return
    if (.not. CheckResult%CheckOK) then
       ! store the failed check, but only if no other Obs_Level checks failed before
       IF (L1B_Screening%Obs_Screening .eq. L1B_File_OK) &
            L1B_Screening%Obs_Screening = L1B_Mie_Invalid_Meas
    endif
    !  #]
    !  #[ screen: nr_mie_invalid_ref_pulses
    val_IntAl     = L1B_PCD_ADS%Observation_PCD%nr_mie_invalid_ref_pulses
    val_IntAl_thr = L1B_Screening_Params%L1B_Obs_Screening_Params%&
                       L1B_Mie_Invalid_RefPulsesThresh

    string_param  = "Num_Mie_Invalid_Reference_Pulse, Obs level"
    CheckResult=ThresholdCheck_IntAl(val_IntAl,string_param,screening_type,&
                                     upper_thr=val_IntAl_thr,lower_thr=0_IntAl)
    if (CheckResult%error_flag .ne. no_error) return
    if (.not. CheckResult%CheckOK) then
       ! store the failed check, but only if no other Obs_Level checks failed before
       IF (L1B_Screening%Obs_Screening .eq. L1B_File_OK) &
            L1B_Screening%Obs_Screening = L1B_Mie_Invalid_Ref_Pulses
    endif
    !  #]
    !  #[ screen: nr_rayleigh_invalid_meas
    val_IntAl     = L1B_PCD_ADS%Observation_PCD%nr_rayleigh_invalid_meas
    val_IntAl_thr = L1B_Screening_Params%L1B_Obs_Screening_Params%&
                       L1B_Rayl_Invalid_Meas_Thresh

    string_param  = "Num_Rayleigh_Invalid_Measurements, Obs level"
    CheckResult=ThresholdCheck_IntAl(val_IntAl,string_param,screening_type,&
                                     upper_thr=val_IntAl_thr,lower_thr=0_IntAl)
    if (CheckResult%error_flag .ne. no_error) return
    if (.not. CheckResult%CheckOK) then
       ! store the failed check, but only if no other Obs_Level checks failed before
       IF (L1B_Screening%Obs_Screening .eq. L1B_File_OK) &
            L1B_Screening%Obs_Screening = L1B_Rayl_Invalid_Meas
    endif
    !  #]
    !  #[ screen: nr_rayleigh_invalid_ref_pulses
    val_IntAl     = L1B_PCD_ADS%Observation_PCD%nr_rayleigh_invalid_ref_pulses
    val_IntAl_thr = L1B_Screening_Params%L1B_Obs_Screening_Params%&
                       L1B_Rayl_InvalidRefPulsesThresh

    string_param  = "Num_Rayleigh_Invalid_Ref_Pulse, Obs level"
    CheckResult=ThresholdCheck_IntAl(val_IntAl,string_param,screening_type,&
                                     upper_thr=val_IntAl_thr,lower_thr=0_IntAl)
    if (CheckResult%error_flag .ne. no_error) return
    if (.not. CheckResult%CheckOK) then
       ! store the failed check, but only if no other Obs_Level checks failed before
       IF (L1B_Screening%Obs_Screening .eq. L1B_File_OK) &
            L1B_Screening%Obs_Screening = L1B_Rayl_Invalid_Ref_Pulses
    endif
    !  #]
    !  #[ screen: nr_invalid_mie_peak
    ! GJM  3/7/2006
    ! I am not quite sure which parameter to use for val_IntAl_thr
    ! and L1B_Screening%Obs_Screening
    ! JK, 24/7/2007
    ! L1B Observation Mie core results are irrelevant for L2B processing
    ! but screening is possible anyway, since the value should always be
    ! between N_range_gates (no results) and 0 (all bins processed)
    ! If this threshold failes it may indicate a damaged file, or a
    ! bug in the L1B processing.
    val_IntAl     = L1B_PCD_ADS%Observation_PCD%nr_invalid_mie_peak
    val_IntAl_thr = N_range_gates

    string_param  = "nr_invalid_mie_peak, Obs level"
    CheckResult=ThresholdCheck_IntAl(val_IntAl,string_param,screening_type,&
                                     upper_thr=val_IntAl_thr,lower_thr=0_IntAl)
    if (CheckResult%error_flag .ne. no_error) return
    if (.not. CheckResult%CheckOK) then
       ! store the failed check, but only if no other Obs_Level checks failed before
       IF (L1B_Screening%Obs_Screening .eq. L1B_File_OK) &
            L1B_Screening%Obs_Screening = L1B_Invalid_Num_Mie_Peaks
    endif
    !  #]
   
! ? array: Obs_Alt_Bin_PCD


! Poli: the section below seems out-of-place. Applying the tests below to
! each altitude_bin_pcd will mean that the flag L1B_Screening%Obs_Screening
! will be set to non-zero as soon as one of the altitude bins (on the
! L1B observation level) presents a problem. We do not want to 
! reject an entire BRC if just one altitude (even on observation level) has a problem.
!
! However, the information contained in the L1B altitude_bin_pcd at
! observation level could be of some use ? -- so the tests could be
! moved somewhere else ?
!
!    !
!    ! Section 1.2 - Screening of parameters over vertical range bin
!    !
!    ! Note GJM 12/7/2006. The parameters below of type Obs_Alt_Bin_PCD_type
!    ! are not in Table 18 of the ECMWF L2B IODD doc. (Ref: AE_IF_ECMWF_L2BP_001).
!    ! I have not included them in the screening documentation (TN2.3, chapter 6)
!    !
!    
!! Screen parameters of type Obs_Alt_Bin_PCD_type
!! Mie_wind_velocity_std_dev, Rayleigh_wind_velocity_std_dev.
!! Mie_useful_signal_std_dev, Rayleigh_useful_signal_stdev_A and
!! Rayleigh_useful_signal_stdev_B are not screened since they are no input
!! for the L2B processor.
!
!    ! Looping over the number of range bins still at observation-scale
!    do r=1,N_range_gates
!    
!       ! bin value to be added to logmsg
!       write(bin_string, '(I3)') r
!    
!       !
!       ! screen scattering_ratio
!       val_FAdoxy     = L1B_PCD_ADS%Observation_PCD%Obs_Alt_Bin_PCD(r)%&
!                        scattering_ratio
!       val_FAdoxy_min = L1B_Screening_Params%L1B_Mie_Meas_Screening_Params%&
!                        L1B_Scattering_Ratio_Min
!       val_FAdoxy_max = L1B_Screening_Params%L1B_Mie_Meas_Screening_Params%&
!                        L1B_Scattering_Ratio_Max
!
!       string_param  = "scattering_ratio, Obs bin "//trim(bin_string)
!       ThresholdCheckOK=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,&
!            screening_type,error_flag,&
!            upper_thr=val_FAdoxy_max,lower_thr=val_FAdoxy_min)
!       if (.not. ThresholdCheckOK) then
!          IF (L1B_Screening%Obs_Screening .eq. L1B_File_OK) &
!               L1B_Screening%Obs_Screening = L1B_Mie_Scattering_Ratio
!       endif
!       
!       ! screen Mie_snr
!       ! This parameter has a lower threshold only
!       val_FAdoxy     = L1B_PCD_ADS%Observation_PCD%Obs_Alt_Bin_PCD(r)%&
!                        Mie_snr
!       val_FAdoxy_thr = L1B_Screening_Params%L1B_Mie_Meas_Screening_Params%&
!                        L1B_Mie_SNR_Thresh
!
!       string_param  = "Mie_snr, Obs bin "//trim(bin_string)
!       ThresholdCheckOK=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,&
!            screening_type,error_flag,&
!            upper_thr=val_FAdoxy,lower_thr=val_FAdoxy_thr)
!       if (.not. ThresholdCheckOK) then
!          IF (L1B_Screening%Obs_Screening .eq. L1B_File_OK) &
!               L1B_Screening%Obs_Screening = L1B_Mie_SNR
!       endif
!       
!       ! screen Rayleigh_snr_channel_A
!       val_FAdoxy     = L1B_PCD_ADS%Observation_PCD%Obs_Alt_Bin_PCD(r)%&
!                        Rayleigh_snr_channel_A
!       val_FAdoxy_min = L1B_Screening_Params%L1B_Rayl_Meas_Screening_Params%&
!                        L1B_Rayleigh_SNR_Min
!       val_FAdoxy_max = L1B_Screening_Params%L1B_Rayl_Meas_Screening_Params%&
!                        L1B_Rayleigh_SNR_Max
!
!       string_param  = "Rayleigh_snr_channel_A, Obs bin "//trim(bin_string)
!       ThresholdCheckOK=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,&
!            screening_type,error_flag,&
!            upper_thr=val_FAdoxy_max,lower_thr=val_FAdoxy_min)
!       if (.not. ThresholdCheckOK) then
!          IF (L1B_Screening%Obs_Screening .eq. L1B_File_OK) &
!               L1B_Screening%Obs_Screening = L1B_Rayl_SNR_A
!       endif
!       
!       ! screen Rayleigh_snr_channel_B
!       val_FAdoxy     = L1B_PCD_ADS%Observation_PCD%Obs_Alt_Bin_PCD(r)%&
!                        Rayleigh_snr_channel_B
!       val_FAdoxy_min = L1B_Screening_Params%L1B_Rayl_Meas_Screening_Params%&
!                        L1B_Rayleigh_SNR_Min
!       val_FAdoxy_max = L1B_Screening_Params%L1B_Rayl_Meas_Screening_Params%&
!                        L1B_Rayleigh_SNR_Max
!
!       string_param  = "Rayleigh_snr_channel_B, Obs bin "//trim(bin_string)
!       ThresholdCheckOK=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,&
!            screening_type,error_flag,&
!            upper_thr=val_FAdoxy_max,lower_thr=val_FAdoxy_min)
!       if (.not. ThresholdCheckOK) then
!          IF (L1B_Screening%Obs_Screening .eq. L1B_File_OK) &
!               L1B_Screening%Obs_Screening = L1B_Rayl_SNR_B
!       endif
!
!    enddo
!    
!    
!    
!    !do r=1,N_range_gates
!       ! screen .........
!       ! to do: error quantifiers
!       !     real(r8_)   :: error_quant_Mie
!       !     real(r8_)   :: error_quant_Rayleigh
!       !     real(r8_)   :: error_quant_Rayleigh_channel_A
!       !     real(r8_)   :: error_quant_Rayleigh_channel_B
!
!       ! This parameter is not screened; no threshold parameter available !!!
!     !  val_IntAl     = L1B_PCD_ADS%Observation_PCD%Obs_Alt_Bin_PCD(r)&
!     !                  %Mie_wind_velocity_std_dev
!       !val_IntAl_thr = ???
!
!   ! enddo
!   
! Poli: end of out-of-place section.
    
    ! GJM Note : for the Mie/Rayleigh measurement screening a double loop
    !            is needed, one over the number of measurement within 
    !            a basic repeat cycle (BRC), the other over the number of
    !            range bins within the measurement profile.
    !
    ! Number of measurements in BRC
    nmeas = L1B_PCD_ADS%nmeas
    
    ! ===========  2. Set L1B_Screening%Mie_Meas() by screening
    !                 L1B PCD Measurement-Scale data
    
    do m=1,nmeas
    !
       ! Section 2.1. Set L1B_Screening%Mie_Meas(m)%Meas_QC
       !
       !    screening of parameters defined for entire measurement independent 
       !    of range-bin
       !    (Parameters from L2B ECMWF IODD document table 20 are screened in m-loop)
       !
       
       ! measurement value to be added to logmsg
       write(meas_string, '(I3)') m
       
       ! Initialization of L1B_Screening%Mie_Meas(m)%Meas_QC
       L1B_Screening%Mie_Meas(m)%Meas_QC = L1B_Mie_Meas_OK
       
       !
       !  Start screening
       !
       
       !  #[ screen: num_mie_invalid_ref_pulses
       val_IntAl     = L1B_PCD_ADS%Measurement_PCD(m)%num_mie_invalid_ref_pulses
       val_IntAl_thr = L1B_Screening_Params%L1B_Mie_Meas_Screening_Params%&
                       L1B_MieMeasInvalidRefPlssThresh
  
       string_param  = "Num_of_Mie_Invalid_Reference_Pulse, Meas="//trim(meas_string)
       CheckResult=ThresholdCheck_IntAl(val_IntAl,string_param,screening_type,&
                                        upper_thr=val_IntAl_thr,lower_thr=0_IntAl)
       if (CheckResult%error_flag .ne. no_error) return
       if (.not. CheckResult%CheckOK) then
          ! store the failed check, but only if no other Meas_Level checks failed before
          IF (L1B_Screening%Mie_Meas(m)%Meas_QC .eq. L1B_Mie_Meas_OK) &
              L1B_Screening%Mie_Meas(m)%Meas_QC = L1B_Mie_Meas_Invalid_Ref_Pulses
       endif
       !  #]
       !  #[ screen: avg_laser_freq_offset
       ! REMARK: this check is performed twice ! once for Mie and once for Rayleigh
       ! TODO: merge them into one check
       val_FAdoxy     = L1B_PCD_ADS%Measurement_PCD(m)%avg_laser_freq_offset
       val_FAdoxy_min = L1B_Screening_Params%L1B_Mie_Meas_Screening_Params%&
                        L1B_Avg_Laser_Freq_Offset_Min
       val_FAdoxy_max = L1B_Screening_Params%L1B_Mie_Meas_Screening_Params%&
                        L1B_Avg_Laser_Freq_Offset_Max

       string_param  = "Avg_Laser_Frequency_Offset, Meas="//trim(meas_string)
       CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                         upper_thr=val_FAdoxy_max,lower_thr=val_FAdoxy_min)
       if (CheckResult%error_flag .ne. no_error) return
       if (.not. CheckResult%CheckOK) then
          ! store the failed check, but only if no other Meas_Level checks failed before
          IF (L1B_Screening%Mie_Meas(m)%Meas_QC .eq. L1B_Mie_Meas_OK) &
              L1B_Screening%Mie_Meas(m)%Meas_QC = L1B_Mie_Avg_Laser_Freq_Offset
       endif
       !  #]
       !  #[ screen: avg_UV_energy
       ! REMARK: this check is performed twice ! once for Mie and once for Rayleigh
       ! TODO: merge them into one check
       val_FAdoxy     = L1B_PCD_ADS%Measurement_PCD(m)%avg_UV_energy
       val_FAdoxy_min = L1B_Screening_Params%L1B_Mie_Meas_Screening_Params%&
                        L1B_Avg_UV_Energy_Min
       val_FAdoxy_max = L1B_Screening_Params%L1B_Mie_Meas_Screening_Params%&
                        L1B_Avg_UV_Energy_Max

       string_param  = "Avg_UV_Energy, Meas="//trim(meas_string)
       CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                         upper_thr=val_FAdoxy_max,lower_thr=val_FAdoxy_min)
       if (CheckResult%error_flag .ne. no_error) return
       if (.not. CheckResult%CheckOK) then
          ! store the failed check, but only if no other Meas_Level checks failed before
          IF (L1B_Screening%Mie_Meas(m)%Meas_QC .eq. L1B_Mie_Meas_OK) &
              L1B_Screening%Mie_Meas(m)%Meas_QC = L1B_Mie_Avg_UV_Energy
       endif
       !  #]
       !  #[ screen: laser_freq_offset_std_dev
       ! REMARK: this check is performed twice ! once for Mie and once for Rayleigh
       ! TODO: merge them into one check
       val_FAdoxy     = L1B_PCD_ADS%Measurement_PCD(m)%laser_freq_offset_std_dev
       val_FAdoxy_thr = L1B_Screening_Params%L1B_Mie_Meas_Screening_Params%&
                        L1B_Laser_FreqOffsetStdevThresh

       string_param  = "Laser_Frequency_Offset_Std_Dev, Meas="//trim(meas_string)
       CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                         upper_thr=val_FAdoxy_thr,lower_thr=0._FAdoxy)
       if (CheckResult%error_flag .ne. no_error) return
       if (.not. CheckResult%CheckOK) then
          ! store the failed check, but only if no other Meas_Level checks failed before
          IF (L1B_Screening%Mie_Meas(m)%Meas_QC .eq. L1B_Mie_Meas_OK) &
              L1B_Screening%Mie_Meas(m)%Meas_QC = L1B_Mie_Laser_Freq_Offset_Stdev
       endif
       !  #]
       !  #[ screen: UV_energy_std_dev
       ! REMARK: this check is performed twice ! once for Mie and once for Rayleigh
       ! TODO: merge them into one check
       val_FAdoxy     = L1B_PCD_ADS%Measurement_PCD(m)%UV_energy_std_dev
       val_FAdoxy_thr = L1B_Screening_Params%L1B_Mie_Meas_Screening_Params%&
                        L1B_UV_Energy_Stdev_Thresh

       string_param  = "UV_Energy_Std_Dev, Meas="//trim(meas_string)
       CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                         upper_thr=val_FAdoxy_thr,lower_thr=0._FAdoxy)
       if (CheckResult%error_flag .ne. no_error) return
       if (.not. CheckResult%CheckOK) then
          ! store the failed check, but only if no other Meas_Level checks failed before
          IF (L1B_Screening%Mie_Meas(m)%Meas_QC .eq. L1B_Mie_Meas_OK) &
              L1B_Screening%Mie_Meas(m)%Meas_QC = L1B_Mie_UV_Energy_Stdev
       endif
       !  #]
       !  #[ screen: vel_of_att_uncertainty_error
       ! REMARK: this check is performed twice ! once for Mie and once for Rayleigh
       ! TODO: merge them into one check
       val_FAdoxy     = L1B_PCD_ADS%Measurement_PCD(m)%&
                        vel_of_att_uncertainty_error
       val_FAdoxy_min = L1B_Screening_Params%L1B_Mie_Meas_Screening_Params%&
                        L1B_VelofAttUncertaintyErrorMin
       val_FAdoxy_max = L1B_Screening_Params%L1B_Mie_Meas_Screening_Params%&
                        L1B_VelofAttUncertaintyErrorMax

       string_param  = "Velocity_of_Attitude_Uncertainty_Error, Meas="//trim(meas_string)
       CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                         upper_thr=val_FAdoxy_max,lower_thr=val_FAdoxy_min)
       if (CheckResult%error_flag .ne. no_error) return
       if (.not. CheckResult%CheckOK) then
          ! store the failed check, but only if no other Meas_Level checks failed before
          IF (L1B_Screening%Mie_Meas(m)%Meas_QC .eq. L1B_Mie_Meas_OK) &
              L1B_Screening%Mie_Meas(m)%Meas_QC = L1B_Mie_Vel_of_Att_Uncert_Error
       endif
       !  #]
       !  #[ screen: mie_mean_emitted_freq
       val_FAdoxy     = L1B_PCD_ADS%Measurement_PCD(m)%&
                        mie_mean_emitted_freq
       val_FAdoxy_min = L1B_Screening_Params%L1B_Mie_Meas_Screening_Params%&
                        L1B_Mie_Mean_Emitted_Freq_Min
       val_FAdoxy_max = L1B_Screening_Params%L1B_Mie_Meas_Screening_Params%&
                        L1B_Mie_Mean_Emitted_Freq_Max

       string_param  = "Mie_Mean_Emitted_Frequency, Meas="//trim(meas_string)
       CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                         upper_thr=val_FAdoxy_max,lower_thr=val_FAdoxy_min)
       if (CheckResult%error_flag .ne. no_error) return
       if (.not. CheckResult%CheckOK) then
          ! store the failed check, but only if no other Meas_Level checks failed before
          IF (L1B_Screening%Mie_Meas(m)%Meas_QC .eq. L1B_Mie_Meas_OK) &
               L1B_Screening%Mie_Meas(m)%Meas_QC = L1B_Mie_Mean_Emitted_Freq
       endif
       !  #]

! the following items are not yet screened. JK, 24-Jul-2007
! TODO: add these
!? mie_emitted_std_dev
!? reference_pulse_FWHM
!? uv_energy_quality_flag


       ! NOTE GJM 10/7/2006
       !   1) parameter mie_emitted_freq_std_dev is commented in l1b_pcd_ads.F90, but 
       !      its threshold is still available in l2b_pcd_ads.F90. Also enum
       !      L1B_Mie_Emitted_FreqStdevThresh is still defined in
       !      l2b_auxpar_datastructure.F90.
       !   11/7/2006. It was decided by phone with David to leave the thresholds in
       !              the thresholds and enum values in the corresponding files.
       !   2) l1b_pcd_ads.F90 contains parameter reference_pulse_FWHM, but no 
       !      thresholds and enum value available for this one.
       !      No screening is applied to this parameter !!
       !   11/7/2006. Decided by phone with David:
       !              TODO. Implement threshold and enum values for this parameter

       do r=1,N_range_gates
       
          ! Section 2.2. Set L1B_Screening%Mie_Meas(m)%Bin_Screening(r)%Bin_QC
          !
          !    Screening of L1B PCD Measurement-range-bin scale data.
          !    screening of parameters defined for entire measurement and range bin 
          !    (loop over range-bins)
          !    (Parameters from L2B ECMWF IODD document table 22 
          !          are screened in (m,r)-loop)
    
          ! bin value to be added to logmsg
          write(bin_string, '(I3)') r

          ! Initialization of L1B_Screening%Mie_Meas(m)%Bin_Screening(r)%Bin_QC
          L1B_Screening%Mie_Meas(m)%Bin_Screening(r)%Bin_QC = L1B_Mie_Bin_OK

          !  #[ screen: Mie_meas_invalid; 
          ! this logical parameter includes the validity check by L1BP
          val_logical   = L1B_PCD_ADS%Measurement_PCD(m)%&
                          Meas_Alt_Bin_PCD(r)%Mie_meas_invalid

          if (val_logical) then

             ! disable printing this logmsg, because we are just copying
             ! a L1B result here, and don't draw any new conclusion.
             ! Flagged results ARE valid inputs so should not generate
             ! warnings in our log file

             !string_param = "Mie_Measurement_Invalid, (meas,bin)=("//&
             !               trim(meas_string)//","//trim(bin_string)//")"
             !call logmsg(log_warn,"Warning in "//trim(screening_type)//": "//&
             !     "validity check failed for "//trim(string_param)//&
             !     "; val= invalid or missing")

             IF (L1B_Screening%Mie_Meas(m)%Bin_Screening(r)%Bin_QC .eq.&
                  L1B_Mie_Bin_OK) &
                  L1B_Screening%Mie_Meas(m)%Bin_Screening(r)%Bin_QC =&
                     L1B_Mie_Bin_Invalid
          endif
          !  #]
          !  #[ screen: scattering_ratio
          val_FAdoxy     = L1B_PCD_ADS%Measurement_PCD(m)%&
                           Meas_Alt_Bin_PCD(r)%scattering_ratio
          val_FAdoxy_min = L1B_Screening_Params%L1B_Mie_Meas_Screening_Params%&
                           L1B_Scattering_Ratio_Min
          val_FAdoxy_max = L1B_Screening_Params%L1B_Mie_Meas_Screening_Params%&
                           L1B_Scattering_Ratio_Max

! REMARK: we should only set the rho_ok flag here when we actually use the 
!         scattering_ratio from the L1B Mie processing !!!
!         when we use a scattering_ratio that we calculated ourselves
!         then another number must be tested. 
! TODO:   Therefore we should
!         add a separate rho_L1B_Mie_ok element to the working datastructure
!         which only gets copied to the rho_ok flag if we really use
!         this scattering_ratio. This rho_L1B_Mie_ok should then be copied
!         to the rho_ok flag in the Copy_Mie_BackscRatio_from_L1B() 
!         routine defined in opticalproperties.F90.
          string_param  = "Scattering_Ratio_Mie, (meas,bin)=("//&
                          trim(meas_string)//","//trim(bin_string)//")"
          CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                      upper_thr=val_FAdoxy_max,lower_thr=val_FAdoxy_min,&
                                      AllowMissing=.true.)
          if (CheckResult%error_flag .ne. no_error) return
          if (.not. CheckResult%CheckOK) then
             WorkingData%MieMeas(m)%RangeGate(r)%rho_ok = .FALSE.
             ! only set the L2B PCD flag if it was not already set
             ! by a problem with higher priority
             IF ( L1B_Screening%Mie_Meas(m)%Bin_Screening(r)%Bin_QC .eq.L1B_Mie_Bin_OK) &
                  L1B_Screening%Mie_Meas(m)%Bin_Screening(r)%Bin_QC = &
                     L1B_Mie_Scattering_Ratio
          else
             WorkingData%MieMeas(m)%RangeGate(r)%rho_ok = .TRUE.
          endif
          !  #]
          !  #[ screen: Mie_snr
          ! This parameter has a lower threshold only
          val_FAdoxy     = L1B_PCD_ADS%Measurement_PCD(m)%&
                           Meas_Alt_Bin_PCD(r)%Mie_snr
          val_FAdoxy_thr = L1B_Screening_Params%L1B_Mie_Meas_Screening_Params%&
                           L1B_Mie_SNR_Thresh

          string_param  = "Mie_Signal_to_Noise_Ratio, (meas,bin)=("//&
                          trim(meas_string)//","//trim(bin_string)//")"
          CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                            upper_thr=val_FAdoxy,lower_thr=val_FAdoxy_thr,&
                                            AllowMissing=.true.)
          if (CheckResult%error_flag .ne. no_error) return
          if (.not. CheckResult%CheckOK) then
             ! only set the L2B PCD flag if it was not already set
             ! by a problem with higher priority
             IF ( L1B_Screening%Mie_Meas(m)%Bin_Screening(r)%Bin_QC .eq.&
                  L1B_Mie_Bin_OK) &
                 L1B_Screening%Mie_Meas(m)%Bin_Screening(r)%Bin_QC =&
                    L1B_Mie_SNR
          endif
          !  #]
     enddo
   enddo
         
    
    
    ! ===========  3. Set L1B_Screening%Rayl_Meas() by screening
    !                 L1B PCD Measurement-Scale data
    
    do m=1,nmeas
    !
    ! Section 3.1. Set L1B_Screening%Rayl_Meas(m)%Meas_QC
    !
    !    screening of parameters defined for entire measurement independent 
    !    of range-bin
    !    (Parameters from L2B ECMWF IODD document table 20 are screened in m-loop)
    !
       ! measurement value to be added to logmsg
       write(meas_string, '(I3)') m
       
       ! Initialization of L1B_Screening%Rayl_Meas(m)%Meas_QC
       L1B_Screening%Rayl_Meas(m)%Meas_QC = L1B_Rayl_Meas_OK
       
       !
       !  Start screening
       !
       
       !  #[ screen: num_rayleigh_invalid_ref_pulses
       val_IntAl     = L1B_PCD_ADS%Measurement_PCD(m)%num_rayleigh_invalid_ref_pulses
       val_IntAl_thr = L1B_Screening_Params%L1B_Rayl_Meas_Screening_Params%&
                       L1B_RayMeasInvalidRefPlssThresh

       string_param  = "Num_of_Rayleigh_Invalid_Reference_Pulse"
       CheckResult=ThresholdCheck_IntAl(val_IntAl,string_param,screening_type,&
                                        upper_thr=val_IntAl_thr,lower_thr=0_IntAl)
       if (CheckResult%error_flag .ne. no_error) return
       if (.not. CheckResult%CheckOK) then
          ! store the failed check, but only if no other Meas_Level checks failed before
          IF (L1B_Screening%Rayl_Meas(m)%Meas_QC .eq. L1B_Rayl_Meas_OK) &
              L1B_Screening%Rayl_Meas(m)%Meas_QC = L1B_Rayl_Meas_Invalid_Ref_Pls
       endif
       !  #]
       !  #[ screen: avg_laser_freq_offset
       ! REMARK: this is a double of the check performed for Mie above !
       val_FAdoxy     = L1B_PCD_ADS%Measurement_PCD(m)%avg_laser_freq_offset
       val_FAdoxy_min = L1B_Screening_Params%L1B_Rayl_Meas_Screening_Params%&
                        L1B_Avg_Laser_Freq_Offset_Min
       val_FAdoxy_max = L1B_Screening_Params%L1B_Rayl_Meas_Screening_Params%&
                        L1B_Avg_Laser_Freq_Offset_Max

       string_param  = "Avg_Laser_Frequency_Offset"
       CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                         upper_thr=val_FAdoxy_max,lower_thr=val_FAdoxy_min)

       if (CheckResult%error_flag .ne. no_error) return
       if (.not. CheckResult%CheckOK) then
          ! store the failed check, but only if no other Meas_Level checks failed before
          IF (L1B_Screening%Rayl_Meas(m)%Meas_QC .eq. L1B_Rayl_Meas_OK) &
              L1B_Screening%Rayl_Meas(m)%Meas_QC =&
                 L1B_Rayl_Avg_Laser_Freq_Offset
       endif
       !  #]
       !  #[ screen: avg_UV_energy
       ! REMARK: this is a double of the check performed for Mie above !
       val_FAdoxy     = L1B_PCD_ADS%Measurement_PCD(m)%avg_UV_energy
       val_FAdoxy_min = L1B_Screening_Params%L1B_Rayl_Meas_Screening_Params%&
                        L1B_Avg_UV_Energy_Min
       val_FAdoxy_max = L1B_Screening_Params%L1B_Rayl_Meas_Screening_Params%&
                        L1B_Avg_UV_Energy_Max

       string_param  = "Avg_UV_Energy"
       CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                       upper_thr=val_FAdoxy_max,lower_thr=val_FAdoxy_min)
       if (CheckResult%error_flag .ne. no_error) return
       if (.not. CheckResult%CheckOK) then
          ! store the failed check, but only if no other Meas_Level checks failed before
          IF (L1B_Screening%Rayl_Meas(m)%Meas_QC .eq. L1B_Rayl_Meas_OK) &
              L1B_Screening%Rayl_Meas(m)%Meas_QC = L1B_Rayl_Avg_UV_Energy
       endif
       !  #]
       !  #[ screen: laser_freq_offset_std_dev
       ! REMARK: this is a double of the check performed for Mie above !
       val_FAdoxy     = L1B_PCD_ADS%Measurement_PCD(m)%laser_freq_offset_std_dev
       val_FAdoxy_thr = L1B_Screening_Params%L1B_Rayl_Meas_Screening_Params%&
                        L1B_Laser_FreqOffsetStdevThresh

       string_param  = "Laser_Frequency_Offset_Std_Dev"
       CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                         upper_thr=val_FAdoxy_thr,lower_thr=0._FAdoxy)
       if (CheckResult%error_flag .ne. no_error) return
       if (.not. CheckResult%CheckOK) then
          ! store the failed check, but only if no other Meas_Level checks failed before
          IF (L1B_Screening%Rayl_Meas(m)%Meas_QC .eq. L1B_Rayl_Meas_OK) &
              L1B_Screening%Rayl_Meas(m)%Meas_QC =&
                 L1B_Rayl_Laser_Freq_Offs_Stdev
       endif
       !  #]
       !  #[ screen: UV_energy_std_dev
       ! REMARK: this is a double of the check performed for Mie above !
       val_FAdoxy     = L1B_PCD_ADS%Measurement_PCD(m)%UV_energy_std_dev
       val_FAdoxy_thr = L1B_Screening_Params%L1B_Rayl_Meas_Screening_Params%&
                        L1B_UV_Energy_Stdev_Thresh

       string_param  = "UV_Energy_Std_Dev"
       CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                         upper_thr=val_FAdoxy_thr,lower_thr=0._FAdoxy)
       if (CheckResult%error_flag .ne. no_error) return
       if (.not. CheckResult%CheckOK) then
          ! store the failed check, but only if no other Meas_Level checks failed before
          IF (L1B_Screening%Rayl_Meas(m)%Meas_QC .eq. L1B_Rayl_Meas_OK) &
              L1B_Screening%Rayl_Meas(m)%Meas_QC =&
              L1B_Rayl_UV_Energy_Stdev
       endif
       !  #]
       !  #[ screen: vel_of_att_uncertainty_error
       ! REMARK: this is a double of the check performed for Mie above !
       val_FAdoxy     = L1B_PCD_ADS%Measurement_PCD(m)%&
                        vel_of_att_uncertainty_error
       val_FAdoxy_min = L1B_Screening_Params%L1B_Rayl_Meas_Screening_Params%&
                        L1B_VelofAttUncertaintyErrorMin
       val_FAdoxy_max = L1B_Screening_Params%L1B_Rayl_Meas_Screening_Params%&
                        L1B_VelofAttUncertaintyErrorMax

       string_param  = "Velocity_of_Attitude_Uncertainty_Error"
       CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                         upper_thr=val_FAdoxy_max,lower_thr=val_FAdoxy_min)
       if (CheckResult%error_flag .ne. no_error) return
       if (.not. CheckResult%CheckOK) then
          ! store the failed check, but only if no other Meas_Level checks failed before
          IF (L1B_Screening%Rayl_Meas(m)%Meas_QC .eq. L1B_Rayl_Meas_OK) &
              L1B_Screening%Rayl_Meas(m)%Meas_QC =&
                L1B_Rayl_Vel_of_Att_Uncert_Err
       endif
       !  #]
       !  #[ screen: rayleigh_mean_emitted_freq
       val_FAdoxy     = L1B_PCD_ADS%Measurement_PCD(m)%&
                        rayleigh_mean_emitted_freq
       val_FAdoxy_min = L1B_Screening_Params%L1B_Rayl_Meas_Screening_Params%&
                        L1B_Rayl_Mean_Emitted_Freq_Min
       val_FAdoxy_max = L1B_Screening_Params%L1B_Rayl_Meas_Screening_Params%&
                        L1B_Rayl_Mean_Emitted_Freq_Max

       string_param  = "Rayleigh_Mean_Emitted_Frequency"
       CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                         upper_thr=val_FAdoxy_max,lower_thr=val_FAdoxy_min)
       if (CheckResult%error_flag .ne. no_error) return
       if (.not. CheckResult%CheckOK) then
          ! store the failed check, but only if no other Meas_Level checks failed before
          IF (L1B_Screening%Rayl_Meas(m)%Meas_QC .eq. L1B_Rayl_Meas_OK) &
              L1B_Screening%Rayl_Meas(m)%Meas_QC = L1B_Rayl_Mean_Emitted_Freq
       endif
       !  #]
       !  #[ screen: rayleigh_emitted_freq_std_dev
       val_FAdoxy     = L1B_PCD_ADS%Measurement_PCD(m)%&
                        rayleigh_emitted_freq_std_dev
       val_FAdoxy_thr = L1B_Screening_Params%L1B_Rayl_Meas_Screening_Params%&
                        L1B_Rayl_EmittedFreqStdevThresh

       string_param  = "Rayleigh_Emitted_Frequency_Std_Dev"
       CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                         upper_thr=val_FAdoxy_max,lower_thr=0._FAdoxy)
       if (CheckResult%error_flag .ne. no_error) return
       if (.not. CheckResult%CheckOK) then
          ! store the failed check, but only if no other Meas_Level checks failed before
          IF (L1B_Screening%Rayl_Meas(m)%Meas_QC .eq. L1B_Rayl_Meas_OK) &
              L1B_Screening%Rayl_Meas(m)%Meas_QC = L1B_Rayl_Emitted_Freq_Stdev
       endif
       !  #]
       
       do r=1,N_range_gates
       
          ! Section 3.2. Set L1B_Screening%Rayl_Meas(m)%Bin_Screening(r)%Bin_QC
          !
          !    Screening of L1B PCD Measurement-range-bin scale data.
          !    screening of parameters defined for entire measurement and range bin 
          !    (loop over range-bins)
          !    (Parameters from L2B ECMWF IODD document table 22 are 
          !            screened in (m,r)-loop)
          
          ! bin value to be added to logmsg
          write(bin_string, '(I3)') r
       
          ! Initialization of L1B_Screening%Rayl_Meas(m)%Bin_Screening(r)%Bin_QC
          L1B_Screening%Rayl_Meas(m)%Bin_Screening(r)%Bin_QC = L1B_Rayl_Bin_OK

          !  #[ screen: Rayleigh_meas_invalid; 
          ! this logical parameter includes the validity check by L1BP
          val_logical   = L1B_PCD_ADS%Measurement_PCD(m)%&
                          Meas_Alt_Bin_PCD(r)%Rayleigh_meas_invalid

          if (val_logical) then

             ! disable printing this logmsg, because we are just copying
             ! a L1B result here, and don't draw any new conclusion.
             ! Flagged results ARE valid inputs so should not generate
             ! warnings in our log file

             !string_param  = "Rayleigh_Meas_Bin_Invalid, (meas,bin)=("//&
             !                trim(meas_string)//","//trim(bin_string)//")"

             !call logmsg(log_warn,"Warning in "//trim(screening_type)//": "//&
             !     "validity check failed for "//trim(string_param)//&
             !     "; val= invalid or missing")

             IF (L1B_Screening%Rayl_Meas(m)%Bin_Screening(r)%Bin_QC .eq.&
                  L1B_Rayl_Bin_OK) &
                 L1B_Screening%Rayl_Meas(m)%Bin_Screening(r)%Bin_QC =&
                    L1B_Rayl_Bin_Invalid
          endif
          !  #]
          !  #[ screen: Rayleigh_snr_channel_A
          val_FAdoxy     = L1B_PCD_ADS%Measurement_PCD(m)%&
                           Meas_Alt_Bin_PCD(r)%Rayleigh_snr_channel_A
          val_FAdoxy_min = L1B_Screening_Params%L1B_Rayl_Meas_Screening_Params%&
                           L1B_Rayleigh_SNR_Min
          val_FAdoxy_max = L1B_Screening_Params%L1B_Rayl_Meas_Screening_Params%&
                           L1B_Rayleigh_SNR_Max

          string_param  = "Rayleigh_Signal_to_Noise_Ratio_Channel_A, (meas,bin)=("//&
                          trim(meas_string)//","//trim(bin_string)//")"
          CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                    upper_thr=val_FAdoxy_max,lower_thr=val_FAdoxy_min,&
                                    AllowMissing=.true.)
          if (CheckResult%error_flag .ne. no_error) return
          if (.not. CheckResult%CheckOK) then
             ! only set the L2B PCD flag if it was not already set
             ! by a problem with higher priority
             IF ( L1B_Screening%Rayl_Meas(m)%Bin_Screening(r)%Bin_QC .eq.&
                  L1B_Rayl_Bin_OK) &
                  L1B_Screening%Rayl_Meas(m)%Bin_Screening(r)%Bin_QC =&
                  L1B_Rayl_SNR_A
          endif
          !  #]
          !  #[ screen: Rayleigh_snr_channel_B
          val_FAdoxy     = L1B_PCD_ADS%Measurement_PCD(m)%&
                           Meas_Alt_Bin_PCD(r)%Rayleigh_snr_channel_B
          val_FAdoxy_min = L1B_Screening_Params%L1B_Rayl_Meas_Screening_Params%&
                           L1B_Rayleigh_SNR_Min
          val_FAdoxy_max = L1B_Screening_Params%L1B_Rayl_Meas_Screening_Params%&
                           L1B_Rayleigh_SNR_Max

          string_param  = "Rayleigh_Signal_to_Noise_Ratio_Channel_B, (meas,bin)=("//&
                          trim(meas_string)//","//trim(bin_string)//")"

          CheckResult=ThresholdCheck_FAdoxy(val_FAdoxy,string_param,screening_type,&
                                      upper_thr=val_FAdoxy_max,lower_thr=val_FAdoxy_min,&
                                      AllowMissing=.true.)
          if (CheckResult%error_flag .ne. no_error) return
          if (.not. CheckResult%CheckOK) then
             ! only set the L2B PCD flag if it was not already set
             ! by a problem with higher priority
             IF ( L1B_Screening%Rayl_Meas(m)%Bin_Screening(r)%Bin_QC .eq.&
                  L1B_Rayl_Bin_OK) &
                 L1B_Screening%Rayl_Meas(m)%Bin_Screening(r)%Bin_QC =&
                    L1B_Rayl_SNR_B
          endif
          !  #]
      enddo
   enddo
    
    
    ! for testing only [!!]
    !print*,'Stop L1b screening'
    !stop

  end subroutine L1B_PCD_ADS_Screening

    !  #]
  subroutine Screen_L1B_Data(L1B,WorkingData, &
                             L1B_Screening,L1B_Screening_Params,&
                             error_flag)
    !  #[
    ! this routine screens L1B input values and writes
    ! the results to the L2B_PCD_ADS%L1B_Screening_Params data structure
    TYPE(L1B_BRC_DataType),          intent(inout) :: L1B
    TYPE(WorkingDataType),           intent(inout) :: WorkingData
    TYPE(L1B_Input_Screening_type),  intent(inout) :: L1B_Screening
    TYPE(L1B_Screening_Params_type), intent(in)    :: L1B_Screening_Params
    integer,                         intent(out)   :: error_flag
    
    ! local variable
    character(31)   :: screening_type

    ! init
    error_flag = no_error

    call logmsg(log_progress,"L1B inputfile screening: starting")

    ! FH, MPH and SPH are not screened in this place,
    ! the reading routine should report if a problem was encountered
    ! and then the L1B_HDR_Corrupted flag can be set

    screening_type = "L1B_PCD_ADS_Screening"
    call L1B_PCD_ADS_Screening(L1B%PCD_ADS,WorkingData,L1B_Screening,&
                               L1B_Screening_Params,screening_type,error_flag)
    IF (error_flag .ne. no_error) return

    print *,"L1B_GWD_ADS  screening not yet implemented"
    ! TODO: screen all pcd items associated with this dataset
    !       and fill L1B_GWD_ADS items with missing_indicator when invalid
    ! TODO: screen L1B_GWD_ADS items that interest us to be in a sensible range
    !       and fill L1B_Screening items accordingly

    print *,"L1B_Meas_ADS screening not yet implemented"
    ! see todo remarks above

    print *,"L1B_US_MDS   screening not yet implemented"
    ! see todo remarks above

    print *,"L1B_WV_MDS   screening not yet implemented"
    ! see todo remarks above

    print *,"L1B_Cal_ADS  screening not yet implemented"
    ! see todo remarks above

    call logmsg(log_progress,"L1B inputfile screening: done")
    
  end subroutine Screen_L1B_Data
  !  #]
  function ThresholdCheck_IntAl(value,string_param,screening_type,&
                                upper_thr,lower_thr,&
                                AllowMissing) result(CheckResult)
    !
    !  #[
    !  Description:
    !          Check if parameter value is missing. If not check if the 
    !          value fits the lower and upper threshold limits.
    !          The function output is a struct CheckResult containing the
    !          boolean CheckOK, which is set to:
    !            -true if the value passes all checks and 
    !            -false if not all checks are passed.
    !          this struct also contains the usual error_flag
    !          to allow signalling a major problem after which the 
    !          program should be aborted
    
    ! interface
    integer(IntAl),        intent(in)  :: value          ! input
    character(len=*),      intent(in)  :: string_param   ! input
    character(len=*),      intent(in)  :: screening_type ! input
    integer(IntAl),        intent(in)  :: upper_thr      ! input
    integer(IntAl),        intent(in)  :: lower_thr      ! input
    type(CheckResultType)              :: CheckResult    ! output
    logical,     optional, intent(in)   :: AllowMissing   ! optional input    

    ! local variables
    character(maxlen_IntAl)     :: txt_value,txt_upper_thr,txt_lower_thr
    
    ! Initiate threshold check
    call InitCheckResult(CheckResult)
    
    IF (value .eq. missing_indicator_IntAl) THEN

       ! a missing value should not be further processed, so set the flag
       CheckResult%CheckOK = .false.

       ! this return is just to suppress the logmsg call below
       if (present(AllowMissing)) then
          if (AllowMissing) return
       end if

       ! remark: missing indicators are not used at the moment for most
       ! elements of the L1B file. For the few exceptions, the optional
       ! AllowMissing switch should be passed on.
       ! For elements that are not supposed to use a missing indicator
       ! print the next warning, and set the CheckOK flag to false
       IF (.not. Silent) THEN
          call logmsg(log_warn,"Warning in "//trim(screening_type)//": "//&
               trim(string_param)//" is missing! "//&
               "threshold not applied")
       END IF
    ELSE
       !
       IF (value .gt. upper_thr) THEN
          IF (.not. Silent) THEN
             call SetIntAl(value,    txt_value,    CheckResult%error_flag)
             IF (CheckResult%error_flag .ne. no_error) return
             call SetIntAl(upper_thr,txt_upper_thr,CheckResult%error_flag)
             IF (CheckResult%error_flag .ne. no_error) return
             call logmsg(log_warn,"Warning in "//trim(screening_type)//": "//&
                  "Threshold failed for "//&
                  trim(string_param)//": val= "//trim(txt_value)//&
                  " upper threshold= "//trim(txt_upper_thr))
          END IF
          CheckResult%CheckOK = .false.
       ENDIF
       
       IF (value .lt. lower_thr) THEN
          IF (.not. Silent) THEN
             call SetIntAl(value,    txt_value,    CheckResult%error_flag)
             IF (CheckResult%error_flag .ne. no_error) return
             call SetIntAl(lower_thr,txt_lower_thr,CheckResult%error_flag)
             IF (CheckResult%error_flag .ne. no_error) return
             call logmsg(log_warn,"Warning in "//trim(screening_type)//": "//&
                  "Threshold failed for "//&
                  trim(string_param)//": val= "//trim(txt_value)//&
                  " lower threshold= "//trim(txt_lower_thr))
          END IF
          CheckResult%CheckOK = .false.
       ENDIF
    ENDIF ! value .eq. missing_indicator_IntAl

  end function ThresholdCheck_IntAl
    !  #]
  function ThresholdCheck_FAdoxy(value,string_param,screening_type,&
                                 upper_thr,lower_thr,&
                                 AllowMissing)  result(CheckResult)
    !
    !  #[
    !  Description:
    !          Check if parameter value is missing. If not check if the 
    !          value fits the lower and upper threshold limits.
    !          The function output is a struct CheckResult containing the
    !          boolean CheckOK, which is set to:
    !            -true if the value passes all checks and 
    !            -false if not all checks are passed.
    !          this struct also contains the usual error_flag
    !          to allow signalling a major problem after which the 
    !          program should be aborted
    !
    
    ! interface
    real(FAdoxy),          intent(in) :: value          ! input
    character(len=*),      intent(in) :: string_param   ! input
    character(len=*),      intent(in) :: screening_type ! input
    real(FAdoxy),          intent(in) :: upper_thr      ! input
    real(FAdoxy),          intent(in) :: lower_thr      ! input
    type(CheckResultType)             :: CheckResult    ! output
    logical,     optional, intent(in) :: AllowMissing   ! optional input
    
    ! local variables
    character(len=50)           :: txt_value,txt_upper_thr,txt_lower_thr
    
    ! Initiate threshold check
    call InitCheckResult(CheckResult)

    IF (missing_real8(value)) THEN

       ! a missing value should not be further processed, so set the flag
       CheckResult%CheckOK = .false.

       ! this return is just to suppress the logmsg call below
       if (present(AllowMissing)) then
          if (AllowMissing) return
       end if

       ! remark: missing indicators are not used at the moment for most
       ! elements of the L1B file. For the few exceptions, the optional
       ! AllowMissing switch should be passed on.
       ! For elements that are not supposed to use a missing indicator
       ! print the next warning, and set the CheckOK flag to false
       IF (.not.Silent) THEN
          call logmsg(log_warn,"Warning in "//trim(screening_type)//": "//&
               trim(string_param)//" is missing! "//&
               "threshold not applied")
       END IF
    ELSE
       !
       IF (value .gt. upper_thr) THEN
          IF (.not.Silent) THEN
             call SetFAdoFull_or_Missing(value,    txt_value,    CheckResult%error_flag)
             IF (CheckResult%error_flag .ne. no_error) return
             call SetFAdoFull_or_Missing(upper_thr,txt_upper_thr,CheckResult%error_flag)
             IF (CheckResult%error_flag .ne. no_error) return
             call logmsg(log_warn,"Warning in "//trim(screening_type)//": "//&
                  "Threshold failed for "//&
                  trim(string_param)//": val= "//trim(txt_value)//&
                  " upper threshold= "//trim(txt_upper_thr),realacc_on=.true.)
          END IF
          CheckResult%CheckOK = .false.
       ENDIF
       
       IF (value .lt. lower_thr) THEN
          IF (.not.Silent) THEN
             call SetFAdoFull_or_Missing(value,    txt_value,    CheckResult%error_flag)
             IF (CheckResult%error_flag .ne. no_error) return
             call SetFAdoFull_or_Missing(lower_thr,txt_lower_thr,CheckResult%error_flag)
             IF (CheckResult%error_flag .ne. no_error) return
             call logmsg(log_warn,"Warning in "//trim(screening_type)//": "//&
                  "Threshold failed for "//&
                  trim(string_param)//": val= "//trim(txt_value)//&
                  " lower threshold= "//trim(txt_lower_thr),realacc_on=.true.)
          END IF
          CheckResult%CheckOK = .false.
       ENDIF
    ENDIF ! value .eq. missing_indicator_FAdoxy

  end function ThresholdCheck_FAdoxy
    !  #]
  !-----------------------------
end module screening_l1b_data
