module kvt_module
  !  #[ Description:
  ! KVT stands for key-value-terminator
  ! and is one of the formats used for writing Aeolus data
  ! to a binary file
  !
  ! created:     23-06-2005 by J. de Kloe, KNMI
  !
  ! Modifications:
  !  01-Mar-2006 J. de Kloe added a generic Compare_XML_to_KVT_item routine
  !                         that can be used in the several read*from_KVT
  !                         routines that we have now.
  !  13-Mar-3006 J. de Kloe added CompareAsReal option to the 
  !                         Compare_XML_to_KVT_item routine
  !  12-May-2006 D. Tan     Null use of ParKind1 module
  !  31-Jul-2006 J. de Kloe added buffered versions of the write routines
  !  26-Jun-2007 J. de Kloe activated actual comparing of integer and real
  !                         items in Compare_XML_to_KVT_item_Repl()
  !  16-Jan-2008 J. de Kloe phase out integer kind i_ 
  !  21-Jan-2008 J. de Kloe correct a typo in an error message
  !
  !  #]
  !  #[ Modules used:
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE DirectBinaryIO, only: DirectBinaryIO_fileunit_type, &
                            WriteNextBytesBuffered
  USE Buffering, only: BufferedFile_type
  ! define the default integer and real to be used by the module
  USE Numerics, only: i8_, r8_
  USE ErrorHandler, only: no_error, error_programming, &
                          error_opening_file, error_reading_file,&
                          error_writing_file
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE LunManager, only: get_lun, free_lun
  USE Compiler_Features, only: newline
  USE StringTools, only: contains_spaces, string2chararray
  !  #]
  !  #[ Parameter/Variable/Type declarations
  
  implicit none

  ! make all module contents private, unless explicitely stated otherwise
  private


  ! the default verbosity setting is Err (as silent as possible)
  integer, save :: LocalVerbosity = VerbosityError

  integer,          parameter :: max_str_len = 256
  character(len=1), parameter :: space   = ' '
  character(len=1), parameter :: equals  = '='
  character(len=1), parameter :: quote   = '"'
  character(len=1), parameter :: unit_bracket_open  = '<'
  character(len=1), parameter :: unit_bracket_close = '>'
  !  #]
  !  #[ interfaces for overloading functions
  INTERFACE write_kvt_string
     module procedure write_kvt_string_unbuff, & ! old direct type
                      write_kvt_string_buff,   & ! old buffered type
                      write_kvt_string_gen       ! new generic type
  END INTERFACE
  INTERFACE write_spare_kvt_string
     module procedure write_spare_kvt_string_unbuff, & ! old direct type
                      write_spare_kvt_string_buff,   & ! old buffered type
                      write_spare_kvt_string_gen       ! new generic type
  END INTERFACE
  !  #]
  !  #[ Module interface
  ! parameters
  public :: max_str_len
  ! subroutines
  public :: Init_KVT_module, &
            write_kvt_string, write_spare_kvt_string, &
            read_kvt_string, read_spare_kvt_string, &
            Compare_XML_to_KVT_item_Repl
  !  #]
contains
  !----------------------------------------------------
  subroutine Init_KVT_module(ChoosenVerbosity)
    !  #[
    integer, optional, intent(in)  :: ChoosenVerbosity

    IF (present(ChoosenVerbosity)) LocalVerbosity = ChoosenVerbosity

    IF (LocalVerbosity .ge. VerbosityReport) &
         call print_verbosity_setting(LocalVerbosity,ModuleName="KVT_module")

  end subroutine Init_KVT_module
    !  #]
  !----------------------------------------------------
  subroutine write_kvt_string_unbuff(fileunit,name,value,max_value_length,&
                                     error_flag,phys_unit,use_quotes)
    !  #[

    integer,          intent(in)  :: fileunit
    character(len=*), intent(in)  :: name
    character(len=*), intent(in)  :: value
    integer,          intent(in)  :: max_value_length
    integer,          intent(out) :: error_flag
    character(len=*), optional, intent(in) :: phys_unit
    logical    ,      optional, intent(in) :: use_quotes

    ! local variables
    character(len=max_str_len) :: kvt_str
    integer                    :: i, kvt_len, filepos

    ! init
    error_flag = no_error

    call construct_kvt_string(kvt_str,kvt_len,name,value,max_value_length,&
                              error_flag,phys_unit,use_quotes)
    IF (error_flag .ne. no_error) return

    ! write the string char by char to the output file
    inquire(unit=fileunit,nextrec=filepos)
    DO i=1,kvt_len
       write(fileunit,rec=filepos+i-1) kvt_str(i:i)
    END DO

  end subroutine write_kvt_string_unbuff
    !  #]
  subroutine write_kvt_string_buff(fileunit,name,value,max_value_length,&
                                   error_flag,phys_unit,use_quotes)
    !  #[
    TYPE(BufferedFile_type), pointer :: fileunit
    character(len=*), intent(in)  :: name
    character(len=*), intent(in)  :: value
    integer,          intent(in)  :: max_value_length
    integer,          intent(out) :: error_flag
    character(len=*), optional, intent(in) :: phys_unit
    logical,          optional, intent(in) :: use_quotes

    ! local variables
    character(len=max_str_len) :: kvt_str
    integer                    :: kvt_len

    ! init
    error_flag = no_error

    call construct_kvt_string(kvt_str,kvt_len,name,value,max_value_length,&
                              error_flag,phys_unit,use_quotes)
    IF (error_flag .ne. no_error) return

    ! write the string char by char to the buffered output file
    call WriteNextBytesBuffered(fileunit,kvt_len,&
                         string2chararray(kvt_str(1:kvt_len)),error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine write_kvt_string_buff
    !  #]
  subroutine write_kvt_string_gen(fileunit,name,value,max_value_length,&
                                  error_flag,phys_unit,use_quotes)
    !  #[
    TYPE(DirectBinaryIO_fileunit_type), intent(in) :: fileunit
    character(len=*), intent(in)  :: name
    character(len=*), intent(in)  :: value
    integer,          intent(in)  :: max_value_length
    integer,          intent(out) :: error_flag
    character(len=*), optional, intent(in) :: phys_unit
    logical,          optional, intent(in) :: use_quotes

    IF (fileunit%use_buffering) THEN
       call write_kvt_string_buff(fileunit%buffered_fileunit,&
                                  name,value,max_value_length,&
                                  error_flag,phys_unit,use_quotes)
    ELSE
       call write_kvt_string_unbuff(fileunit%fileunit,&
                                    name,value,max_value_length,&
                                    error_flag,phys_unit,use_quotes)
    END IF

  end subroutine write_kvt_string_gen
    !  #]
  subroutine construct_kvt_string(kvt_str,kvt_len,name,value,max_value_length,&
                                  error_flag,phys_unit,use_quotes)
    !  #[
    character(len=max_str_len), intent(out) :: kvt_str
    integer,                    intent(out) :: kvt_len
    character(len=*),           intent(in)  :: name
    character(len=*),           intent(in)  :: value
    integer,                    intent(in)  :: max_value_length
    integer,                    intent(out) :: error_flag
    character(len=*), optional, intent(in) :: phys_unit
    logical,          optional, intent(in) :: use_quotes

    ! local variables
    logical :: do_use_quotes
    character(len=max_str_len) :: tmp_value
    integer :: value_length
    integer :: n  ! nr of characters in kvt_str

    ! init
    error_flag = no_error

    do_use_quotes = .false.
    IF (present(use_quotes)) do_use_quotes = use_quotes    

    ! check for proper use of the optional arguments
    IF (do_use_quotes .and. present(phys_unit)) THEN
       print *,"ERROR in construct_kvt_string:"
       print *,"adding a physics unit is only allowed for numerical values"
       print *,"in which case the use_quotes switch should not be used !!!"
       error_flag = error_programming
       return
    END IF

    ! test to see if spaces are present inside the tag-name
    ! (this is not allowed!)
    IF (contains_spaces(name)) THEN
       print *,"ERROR in construct_kvt_string:"
       print *,"The tag name to be written in KVT format seems to hold spaces"
       print *,"this is not allowed. You could add some underscores to the"
       print *,"tag name to fix this..."
       error_flag = error_programming
       return
    END IF

    ! test to see if spaces are present inside the value string
    ! in which case the "use_quotes" option SHOULD be used !!!
    IF (.not. do_use_quotes) THEN
       IF (contains_spaces(value)) THEN
          print *,"ERROR in construct_kvt_string:"
          print *,"The value to be written in KVT format seems to hold spaces"
          print *,"so the use_quotes option MUST be set, but this seems"
          print *,"not to be the case ..."
          error_flag = error_programming
          return
       END IF
    END IF

    ! test if the size of the value fits in our tmp strings
    IF (max_value_length .gt. max_str_len) THEN
       print *,"ERROR in construct_kvt_string:"
       print *,"max_str_len is too small to hold this kvt value string !!!"
       print *,"max_str_len      = ",max_str_len
       print *,"max_value_length = ",max_value_length
       print *,"if this value is correct then increase the max_str_len"
       print *,"parameter in the KVT_module..."
       error_flag = error_programming
       return
    END IF

    ! check the total string to be constructed ...
    n = len_trim(name)+1+max_value_length+1
    IF (do_use_quotes) n=n+2
    IF (present(phys_unit)) n=n+2+len_trim(phys_unit)
    IF (n .gt. max_str_len) THEN
       print *,"ERROR in construct_kvt_string:"
       print *,"max_str_len is too small to hold this kvt line !!!"
       print *,"max_str_len                      = ",max_str_len
       print *,"length of kvt line to be written = ",n
       print *,"if this value is correct then increase the max_str_len"
       print *,"parameter in the KVT_module..."
       error_flag = error_programming
       return
    END IF

    ! first add the name for this item
    n = len_trim(name)
    kvt_str(1:n) = name(1:n)

    ! then add the equals ('=') character
    n=n+1
    kvt_str(n:n) = equals

    IF (do_use_quotes) THEN
       ! then add a quote character if needed
       n=n+1
       kvt_str(n:n) = quote
    END IF

    ! then add the value with the requested amount of characters
    ! padded with spaces to fill the string if needed
    value_length = len_trim(value)
    IF (value_length .gt. max_value_length) THEN
       IF (LocalVerbosity .ge. VerbosityReport) THEN
          print *,"WARNING: truncating this KVT item since it is to long !!!"
          print *,"max length specified is: ",max_value_length
          print *,"value specified is: ",value(1:value_length)
       END IF
       value_length = max_value_length
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"truncated value is: ",value(1:value_length)
    END IF
    ! copy the characters for value
    tmp_value(1:value_length) = value(1:value_length)
    ! padd the string with spaces if not already filled
    IF (value_length .lt. max_value_length) &
         tmp_value(value_length+1:max_value_length) = ' '

    ! now add the padded or truncated value to the kvt_str
    kvt_str(n+1:n+max_value_length) = tmp_value(1:max_value_length)
    n=n+max_value_length

    IF (do_use_quotes) THEN
       ! then add a second quote character if needed
       n=n+1
       kvt_str(n:n) = quote
    END IF

    IF (present(phys_unit)) THEN
       n=n+1
       kvt_str(n:n) = unit_bracket_open

       kvt_str(n+1:n+len_trim(phys_unit)) = trim(phys_unit)
       n=n+len_trim(phys_unit)

       n=n+1
       kvt_str(n:n) = unit_bracket_close
    END IF

    ! finally end with a newline character
    n=n+1
    kvt_str(n:n) = newline

    ! report nr of characters in the constructed string
    kvt_len = n

  end subroutine construct_kvt_string
    !  #]
  subroutine write_spare_kvt_string_unbuff(fileunit,spare_length,error_flag)
    !  #[

    integer, intent(in)  :: fileunit
    integer, intent(in)  :: spare_length
    integer, intent(out) :: error_flag

    ! local variables
    integer :: i, filepos

    ! init
    error_flag = no_error

    ! write the string char by char to the output file
    inquire(unit=fileunit,nextrec=filepos)
    DO i=1,spare_length
       write(fileunit,rec=filepos+i-1) space
    END DO
    inquire(unit=fileunit,nextrec=filepos)
    write(fileunit,rec=filepos) newline

  end subroutine write_spare_kvt_string_unbuff
    !  #]
  subroutine write_spare_kvt_string_buff(fileunit,spare_length,error_flag)
    !  #[
    TYPE(BufferedFile_type), pointer :: fileunit
    integer, intent(in)  :: spare_length
    integer, intent(out) :: error_flag

    ! local variable
    character(len=1),dimension(max_str_len) :: tmp_str

    ! init
    error_flag = no_error

    tmp_str(:) = ' '
    tmp_str(spare_length+1:spare_length+1) = newline

    ! write the string char by char to the buffered output file
    call WriteNextBytesBuffered(fileunit,spare_length+1,tmp_str,error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine write_spare_kvt_string_buff
    !  #]
  subroutine write_spare_kvt_string_gen(fileunit,spare_length,error_flag)
    !  #[
    TYPE(DirectBinaryIO_fileunit_type), intent(in) :: fileunit
    integer, intent(in)  :: spare_length
    integer, intent(out) :: error_flag


    IF (fileunit%use_buffering) THEN
       call write_spare_kvt_string_buff(fileunit%buffered_fileunit,&
                                        spare_length,error_flag)
    ELSE
       call write_spare_kvt_string_unbuff(fileunit%fileunit,&
                                          spare_length,error_flag)
    END IF

  end subroutine write_spare_kvt_string_gen
    !  #]
  !----------------------------------------------------
  subroutine read_kvt_string(fileunit,expected_name,value,&
       error_flag,expected_value_length,expected_phys_unit,is_numerical,&
       expect_quotes)
    !  #[
    integer,                    intent(in)  :: fileunit
    character(len=*),           intent(in)  :: expected_name
    character(len=*),           intent(out) :: value
    integer,                    intent(out) :: error_flag
    integer,                    intent(in)  :: expected_value_length
    character(len=*), optional, intent(in)  :: expected_phys_unit
    logical,          optional, intent(in)  :: is_numerical
    logical,          optional, intent(in)  :: expect_quotes

    ! local variables
    character(len=max_str_len) :: tmp_string
    character(len=max_str_len) :: name
    character(len=max_str_len) :: phys_unit
    character(len=1) :: tmp_char
    integer :: i, filepos,str_length, pos,ios
    integer :: pos_of_equals_char
    integer :: value_length
    integer :: pos_of_opening_bracket 
    logical :: do_expect_phys_unit
    logical :: do_expect_quotes
    logical :: expect_numerical_value

    ! the list of characters allowed in a numerical KVT item
    character(len=256), parameter :: allowed_char_string = "0123456789+-."

    ! init
    error_flag   = no_error
    name(:)      = ' '
    value(:)     = ' '
    phys_unit(:) = ' '

    ! check the presence of the optional inputs
    do_expect_phys_unit = .false.
    IF (present(expected_phys_unit)) do_expect_phys_unit = .true.

    expect_numerical_value = .false.
    IF (present(is_numerical)) expect_numerical_value = is_numerical

    do_expect_quotes = .false.
    IF (present(expect_quotes)) do_expect_quotes = expect_quotes

    ! sanity check on the optional inputs
    IF (do_expect_phys_unit) THEN
       IF (present(is_numerical)) THEN
          IF (.not. is_numerical) THEN
             print *,"ERROR in read_kvt_string:"
             print *,"the optional switches expected_phys_unit and"
             print *,"is_numerical are conflicting !!!"
             print *,"phys_units may only be used for numerical fields !!!"
             error_flag = error_programming
             return
          END IF
       ELSE
          ! default is numerical in case of presence of phys_unit
          expect_numerical_value = .true.
       END IF
    END IF

    !print *,"testjos: do_expect_phys_unit    = ",do_expect_phys_unit
    !print *,"testjos: expect_numerical_value = ",expect_numerical_value
    !print *,"testjos: do_expect_quotes       = ",do_expect_quotes

    ! sanity check on the optional inputs
    IF (do_expect_quotes .and. do_expect_phys_unit) THEN
       print *,"ERROR in read_kvt_string:"
       print *,"the optional switches expect_quotes and expected_phys_unit"
       print *,"should never be true together. For text values quotes"
       print *,"should be used. For numerical strings a phys_unit may"
       print *,"or may not be used, but quotes should not be used."
       error_flag = error_programming
       return
    END IF

    ! read the string char by char from the input file
    ! untill a newline is found, and give an error if the
    ! max_str_len is exceeded
    inquire(unit=fileunit,nextrec=filepos)
    i=1
    readloop: DO 
       read(fileunit,rec=filepos+i-1,iostat=ios) tmp_char
       IF (ios .ne. 0) goto 999
       IF (tmp_char .eq. newline) exit readloop
       tmp_string(i:i) = tmp_char
       i=i+1
       IF (i .gt. max_str_len) THEN 
          print *,"ERROR in subroutine read_kvt_string():"
          print *,"no newline character was found or this item"
          print *,"is too long to be stored in a string of size"
          print *,"max_str_len = ",max_str_len
          print *,"expected name: [",trim(expected_name),"]"
          print *,"current  item: [",tmp_string,"]"
          print *,"If this datafile is correct please increase the value"
          print *,"of max_str_len in the file KVT_module.F90"
          error_flag = error_reading_file
          return
       END IF
    END DO readloop
    str_length = i-1 ! since pos i contains the newline character !
    !print *,"str_length=",str_length
    !print *,"tmp_string = [",tmp_string(1:str_length),"]"

    ! searching for the location of the equals character
    pos_of_equals_char = -1
    DO i=1,str_length
       IF (tmp_string(i:i) .eq. equals) THEN
          pos_of_equals_char = i
          exit
       END IF
    END DO

    IF (pos_of_equals_char .eq. -1) THEN
       print *,"ERROR in subroutine read_kvt_string():"
       print *,"no equals (=) character was found or this item"
       print *,"A kvt item should always have the form:"
       print *,"name=value\\n"
       print *,"expected name: [",trim(expected_name),"]"
       print *,"current  item: [",tmp_string(1:str_length),"]"

       IF (len_trim(tmp_string(1:str_length)) .eq. 0) THEN
          print *,"==>Since this item consists of only spaces this might be"
          print *,"==>a spare item. You should use the subroutine"
          print *,"==>read_spare_kvt_string() to read such items !!!"
       END IF

       error_flag = error_reading_file
       return
    END IF

    !print *,"before: ", tmp_string(1:pos_of_equals_char-1)
    !print *,"after:  ", tmp_string(pos_of_equals_char+1:str_length)

    ! init with spaces
    name(:) = ' ' 
    value(:) = ' '

    ! copy the result to the output strings
    name(1:pos_of_equals_char-1) = tmp_string(1:pos_of_equals_char-1)
    value_length = str_length-pos_of_equals_char
    value(1:value_length) = &
         tmp_string(pos_of_equals_char+1:str_length)

    !print *,"name  = [",trim(name),"]"
    !print *,"value = [",trim(value),"]"

    ! see if the name equals the expected name
    IF (trim(name) .ne. trim(expected_name)) THEN
       print *,"ERROR in subroutine read_kvt_string():"
       print *,"expected an item with name:  ",trim(expected_name)
       print *,"but found an item with name: ",trim(name)
       error_flag = error_reading_file
       return
    END IF

    ! see if we have quotes, and if this was expected
    IF (do_expect_quotes) THEN
       IF (value(1:1) .eq. quote) THEN ! check on the start quote
          ! see if the closing quote is present
          IF (value(value_length:value_length) .eq. quote) THEN
             ! remove the quotes from the string value
             value(1:value_length-2) = value(2:value_length-1)
             value(value_length-1:value_length) = ' '
             value_length = value_length-2
          ELSE
             print *,"ERROR in subroutine read_kvt_string():"
             print *,"this item was expected to be a quoted string, but"
             print *,"an end quote seems not present"
             print *,"expected name: [",trim(expected_name),"]"
             print *,"item value = [",trim(value),"]"
             error_flag = error_reading_file
             return
          END IF
       ELSE
          ! no start quote found
          print *,"ERROR in subroutine read_kvt_string():"
          print *,"this item was expected to be a quoted string, but no"
          print *,"starting quote seems present"
          print *,"expected name: [",trim(expected_name),"]"
          print *,"item value = [",trim(value),"]"
          error_flag = error_reading_file
          return
       END IF
    ELSE
       ! check if there is no unexpected start quote
       IF (value(1:1) .eq. quote) THEN
          print *,"ERROR in subroutine read_kvt_string():"
          print *,"this item was NOT expected to be a quoted string, but"
          print *,"a start quote seems present"
          print *,"expected name: [",trim(expected_name),"]"
          print *,"item value = [",trim(value),"]"
          error_flag = error_reading_file
          return
       END IF
       ! check if there is no unexpected end quote
       IF (value(value_length:value_length) .eq. quote) THEN
          print *,"ERROR in subroutine read_kvt_string():"
          print *,"this item was NOT expected to be a quoted string, but"
          print *,"an end quote seems present"
          print *,"expected name: [",trim(expected_name),"]"
          print *,"item value = [",trim(value),"]"
          error_flag = error_reading_file
          return
       END IF
    END IF
    
    IF (do_expect_phys_unit) THEN
       ! remark: within a quotes string the <> characters are perfectly
       ! legal, so a physical unit might be added there, but this routine
       ! will not attempt to extract that unit, since quotes strings
       ! are not supposed to be numerical values. 
       
       ! searching for the location of the opening bracket
       pos_of_opening_bracket = -1
       DO i=1,value_length
          IF (value(i:i) .eq. unit_bracket_open) THEN
             pos_of_opening_bracket = i
             exit
          END IF
       END DO

       IF (pos_of_opening_bracket .eq. -1) THEN
          print *,"ERROR in subroutine read_kvt_string():"
          print *,"this item was expected to be have a physical unit, but"
          print *,"an opening bracket seems NOT present"
          print *,"expected name: [",trim(expected_name),"]"
          print *,"item value = [",trim(value),"]"
          error_flag = error_reading_file
          return
       END IF

       ! check for the presence of the closing bracket
       IF (value(value_length:value_length) .ne. unit_bracket_close) THEN
          print *,"ERROR in subroutine read_kvt_string():"
          print *,"this item was expected to be have a physical unit, but"
          print *,"a closing bracket seems NOT present"
          print *,"expected name: [",trim(expected_name),"]"
          print *,"item value = [",trim(value),"]"
          error_flag = error_reading_file
          return
       END IF

       ! extract the unit string
       phys_unit = value(pos_of_opening_bracket+1:value_length-1)

       ! check if this unit is what whas expected
       IF (trim(phys_unit) .ne. trim(expected_phys_unit)) THEN
          print *,"ERROR in subroutine read_kvt_string():"
          print *,"expected phys_unit with name:    <",&
               trim(expected_phys_unit),">"
          print *,"but found a phys_unit with name: <",trim(phys_unit),">"
          print *,"expected name: [",trim(expected_name),"]"
          print *,"current item:  [",trim(value),"]"
          error_flag = error_reading_file
          return
       END IF

       ! remove the phys_unit and the brackets from the value string
       value(pos_of_opening_bracket:value_length) = ' '
       value_length = pos_of_opening_bracket-1
    END IF

    ! check the length of the found string
    IF (expected_value_length .ne. value_length) THEN
       print *,"ERROR in subroutine read_kvt_string():"
       print *,"the length of the value of this item"
       print *,"does not match the expected length"
       print *,"expected value length = ",expected_value_length
       print *,"found    value_length = ",value_length
       print *,"expected name: [",trim(expected_name),"]"
       print *,"current item:  [",trim(value),"]"
       error_flag = error_reading_file
       return
    END IF

    IF (.not. do_expect_quotes) THEN
       IF (expect_numerical_value) THEN
          ! in case of no_quotes there should be a numerical value in
          ! the string, so check for invaluid characters
          pos = verify(trim(value),trim(allowed_char_string))
          IF (pos .ne. 0) THEN
             ! some illegal character appears to be present !!!
             print *,"ERROR in read_kvt_string: an illegal character seems to"
             print *,"be present in this numerical item:"
             print *,"item name  = [",trim(name),"]"
             print *,"item value = [",trim(value),"]"
             print *,"pos = ",pos," value(pos:pos)=",value(pos:pos)
             print *,"allowed_char_string = [",trim(allowed_char_string),"]"
             error_flag = error_reading_file
             return
          END IF
       END IF
    END IF

    return

999 print *,"ERROR in subroutine read_kvt_string():"
    print *,"could not read any more characters for this item"
    error_flag = error_reading_file
    return

  end subroutine read_kvt_string
    !  #]
  subroutine read_spare_kvt_string(fileunit,spare_length,error_flag)
    !  #[
    integer, intent(in)  :: fileunit
    integer, intent(in)  :: spare_length
    integer, intent(out) :: error_flag
    
    ! local variables
    integer :: i, filepos, str_length, ios
    character(len=1) :: tmp_char
    character(len=max_str_len) :: tmp_string

    ! init
    error_flag = no_error

    ! read the string char by char from the input file
    ! untill a newline is found, and give an error if the
    ! max_str_len is exceeded
    inquire(unit=fileunit,nextrec=filepos)
    i=1
    readloop: DO 
       read(fileunit,rec=filepos+i-1,iostat=ios) tmp_char
       IF (ios .ne. 0) goto 999

       IF (tmp_char .eq. newline) exit readloop
       tmp_string(i:i) = tmp_char
       i=i+1
       IF (i .gt. max_str_len) THEN 
          print *,"ERROR in subroutine read_spare_kvt_string():"
          print *,"no newline character was found or this item"
          print *,"is too long to be stored in a string of size"
          print *,"max_str_len = ",max_str_len
          print *,"If this datafile is correct please increase the value"
          print *,"of max_str_len in the file KVT_module.F90"
          error_flag = error_reading_file
          return
       END IF
    END DO readloop
    str_length = i-1 ! since pos i contains the newline character !

    ! check if the string contains only spaces
    IF (len_trim(tmp_string(1:str_length)) .ne. 0) THEN
       print *,"ERROR in subroutine read_spare_kvt_string():"
       print *,"This item should consiste of only spaces, but some other"
       print *,"characters seem present !!!"
       print *,"str_length=",str_length
       print *,"tmp_string = [",tmp_string(1:str_length),"]"
       error_flag = error_reading_file
       return
    END IF

    ! check the length of the found string of spaces
    IF (spare_length .ne. str_length) THEN
       print *,"ERROR in subroutine read_spare_kvt_string():"
       print *,"the length of this spare item"
       print *,"does not match the expected length"
       print *,"expected spare_length = ",spare_length
       print *,"found    str_length   = ",str_length
       print *,"current item: [",tmp_string(1:str_length),"]"
       error_flag = error_reading_file
       return
    END IF

    return

999 print *,"ERROR in subroutine read_kvt_string():"
    print *,"could not read any more characters for this item"
    error_flag = error_reading_file
    return

  end subroutine read_spare_kvt_string
    !  #]
  !----------------------------------------------------
  subroutine Compare_XML_to_KVT_item_Repl(XML_value,KVT_value,Item,error_flag,&
       CompareAsInteger,CompareAsReal,CompareAsDate)
    !  #[ compare with a value in memory (usually XML), and replace 
    character(len=*), intent(inout) :: XML_value  ! filled or compared
    character(len=*), intent(in)    :: KVT_value  ! input
    character(len=*), intent(in)    :: Item       ! input
    integer,          intent(out)   :: error_flag ! output
    logical,optional, intent(in)    :: CompareAsInteger  ! switch
    logical,optional, intent(in)    :: CompareAsReal     ! switch
    logical,optional, intent(in)    :: CompareAsDate     ! switch
    ! local variables
    logical :: do_CompareAsInteger
    logical :: do_CompareAsReal
    logical :: do_CompareAsDate
    logical :: ValuesDiffer
    integer :: ios
    integer(i8_) :: Temp_KVT_Int   ! temporary int*8
    integer(i8_) :: Temp_MPH_Int   ! temporary int*8
    real(r8_)    :: Temp_KVT_Real8 ! temporary real*8
    real(r8_)    :: Temp_MPH_Real8 ! temporary real*8
    real(r8_)    :: delta

    ! initialise, to prevent the from being uninitialised
    ! when a read statement fails without reporting an error.
    ! (seems to happen with g95)
    Temp_KVT_Int = 0
    Temp_MPH_Int = 0
    Temp_KVT_Real8 = 0.
    Temp_MPH_Real8 = 0.

    error_flag = no_error

    do_CompareAsInteger = .false.
    IF (present(CompareAsInteger)) do_CompareAsInteger = CompareAsInteger

    do_CompareAsReal = .false.
    IF (present(CompareAsReal)) do_CompareAsReal = CompareAsReal

    do_CompareAsDate = .false.
    IF (present(CompareAsDate)) do_CompareAsDate = CompareAsDate

    ! compare the values
    ValuesDiffer = .false.

    ! TODO: remove this temporary exception
    !       when all filehandling has been implemented
    IF ( (trim(KVT_value) .eq. "") .and. &
         (trim(XML_value) .eq. "")       ) THEN
       ! store the result
       XML_value = KVT_value
       return
       ! this is an exception at the moment that occurs when
       ! an attempt is made to read a value from an empty XML (MPH/SPH) item
    END IF

    IF (do_CompareAsInteger) THEN
       ! convert to a numerical (integer) value
       read(KVT_value,*,iostat=ios) Temp_KVT_Int
       IF (ios .ne. 0) goto 990
       read(XML_value,*,iostat=ios) Temp_MPH_Int
       IF (ios .ne. 0) goto 991
       ! compare the integers to see if they are identical
       IF (Temp_KVT_Int .ne. Temp_MPH_Int) ValuesDiffer = .true.
    ELSE IF (do_CompareAsReal) THEN
       ! convert to a numerical (real) value
       ! should I add a delta here ?
       read(KVT_value,*,iostat=ios) Temp_KVT_Real8
       IF (ios .ne. 0) goto 992
       read(XML_value,*,iostat=ios) Temp_MPH_Real8
       IF (ios .ne. 0) goto 993
       ! compare the reals to see if they are identical
       ! the string values should be identical, except for the number of
       ! leading or trailing zeros. Therefore the numbers should be identical
       ! upto machine precision. 
       ! Nevertheless, allow a relative delta of 1:1.e-12, except when the numbers
       ! are zero, then use an absolute delta of 1.e-12.
       IF ( (Temp_KVT_Real8 .eq. 0._r8_) .or. (Temp_MPH_Real8 .eq. 0._r8_) ) THEN
          if (abs(Temp_KVT_Real8-Temp_MPH_Real8) .gt. 1.e-12_r8_) &
               ValuesDiffer = .true.
       ELSE
          ! calculate the delta
          delta = 1.e-12_r8_ * max(abs(Temp_KVT_Real8),abs(Temp_MPH_Real8))
          ! compare the numbers
          if (abs(Temp_KVT_Real8-Temp_MPH_Real8) .gt. delta) &
               ValuesDiffer = .true.
       END IF
    ELSE IF (do_CompareAsDate) THEN
       ! real KVT dates have the precision for seconds upto the microsecond
       ! XML-converted to-KVT dates have just integer seconds, 
       ! so when comparing just throw away the last 6 digits of the second
       IF (KVT_value(1:21) .ne. XML_value(1:21)) THEN
          ValuesDiffer = .true.
       ELSE
          ! replace XML_Value by KVT_Value becase time is more accurate
          ! in KVT than in XML
          XML_value = KVT_value
       ENDIF
    ELSE
       ! just do a text compare
       if (trim(KVT_value) .ne. trim(XML_value)) ValuesDiffer = .true.
    END IF
    
    IF (ValuesDiffer) THEN
       print *,"WARNING in Compare_XML_to_KVT_item:"
       print *,"XML-MPH and KVT-MPH values for the same item differ !!!"
       print *,"XML: ",trim(Item),": ",trim(XML_value)
       print *,"KVT: ",trim(Item),": ",trim(KVT_value)
       print *,"keeping value from XML file for now"
       error_flag = error_reading_file
       return
    ELSE
       ! store the result
       XML_value = KVT_value
    END IF
 
    return

    ! error handlers
990 print *,"WARNING in Compare_XML_to_KVT_item():"
    print *,"Item name: ",trim(Item)
    print *,"Could not convert KVT value: ",KVT_value
    print *,"into an integer number ...."
    error_flag = error_reading_file
    return

991 print *,"WARNING in Compare_XML_to_KVT_item():"
    print *,"Item name: ",trim(Item)
    print *,"Could not convert XML value: ",XML_value
    print *,"into an integer number ...."
    error_flag = error_reading_file
    return

992 print *,"WARNING in Compare_XML_to_KVT_item():"
    print *,"Item name: ",trim(Item)
    print *,"Could not convert KVT value: ",KVT_value
    print *,"into a real number ...."
    error_flag = error_reading_file
    return
993 print *,"WARNING in Compare_XML_to_KVT_item():"
    print *,"Item name: ",trim(Item)
    print *,"Could not convert XML value: ",XML_value
    print *,"into a real number ...."
    error_flag = error_reading_file
    return

  END subroutine Compare_XML_to_KVT_item_Repl
  !  #]
  !---------------------------------
end module kvt_module
