program TestReadL1Bdata
  !  #[ Documentation
  !     a test program for the ReadL1Bdata module
  !     written by: J. de Kloe (KNMI)
  !
  !     Modifications:
  !        01-Sep-2005  D. Tan      Upgrade to DLR v1.2 TDS.
  !        28-Sep-2005  P. Poli     Compatibility with NAGware Fortran 
  !                                 compile (f95)
  !        26-Oct-2005  J. de Kloe  added test for allocation of working arrays
  !        12-May-2006  D. Tan      Null use of ParKind1 module
  !        23-Nov-2006  J. de Kloe  added loop to read all BRC's in the file
  !        29-Aug-2007  D. Tan      Phase out L1B-working.
  !        31-Aug-2007  P. Poli     Introduce L1B_Headers
  !        04-Sep-2007  P. Poli     Change L1B_DataType to L1B_BRC_DataType 
  !                                 and L1B_HDR_DataType
  !        16-Jan-2008  J. de Kloe  phase out integer kind i_
  !        25-Jan-2008  J. de Kloe  added a commandline option to reduce
  !                                 number of read BRC's (to reduce runtime
  !                                 and the amount of text output of the
  !                                 test program)
  !  #]
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE L1B_DataStructure, only: L1B_HDR_DataType, &
                               L1B_BRC_DataType, &
                               Init_L1B_HDR_DataStructure, &
                               Init_L1B_BRC_DataStructure, &
                               Delete_L1B_HDR_DataStructure, &
                               Delete_L1B_BRC_DataStructure, &
                               Empty_L1B_BinaryDatasets
  USE ReadL1Bdata, only: ReadL1Bfile, InitReadL1BfileModule
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport
  USE ErrorHandler, only: no_error, program_abort, error_cmdline_options
  USE Compiler_Features, only: iargc_aeolus,getarg_aeolus
  !  #]
  !  #[ variables
  implicit none
  integer :: error_flag, nr_of_arguments
  integer :: iBRC, Num_BRC
  character(len=256) :: filename ! without extension
  character(len=256) :: arg_name, arg_value
  TYPE(L1B_HDR_DataType) :: L1B_Headers
  TYPE(L1B_BRC_DataType) :: L1B
  !  #]
  !  #[ program code
  call Init_L1B_HDR_DataStructure(L1B_Headers, error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"Init_L1B_HDR_DataStructure")  

  call Init_L1B_BRC_DataStructure(L1B, error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"Init_L1B_BRC_DataStructure")  

  call InitReadL1BfileModule(error_flag)!,ChoosenVerbosity=VerbosityReport)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"InitReadL1BfileModule")

  ! read the filename without extension as 1st commandline parameter:
  nr_of_arguments = iargc_aeolus()
  IF ( (nr_of_arguments .ne. 1) .and. &
       (nr_of_arguments .ne. 3) ) THEN
     print *,"ERROR: wrong number of commandline arguments."
     print *,"nr_of_arguments = ",nr_of_arguments
     call PrintUsage
     call program_abort(error_cmdline_options,"TestReadL1Bdata")
  END IF

  call getarg_aeolus(1,filename)
!  filename = "AE_TEST_ALD_U_N_1B_20071002T000027892_000252016_000000_0000"
!  filename = "AE_TEST_ALD_U_N_1B_20071002T000000049_000252022_000000_0000"

  ! read the headers to retrieve the nr of BRC's in the file
  call ReadL1Bfile(L1B_Headers, filename, error_flag, ReadHeaders=.true.)
  IF (error_flag .ne. no_error) call program_abort(error_flag,"ReadL1Bfile")

  ! Set the number of BRCs
  Num_BRC = L1B_Headers%SPH%Geolocation_DSD%nr_of_dataset_records
  print *,"This L1B file contains: ",Num_BRC," BRC's"

  IF (nr_of_arguments .eq. 3) THEN
     ! read NumBRC from the commandline
     call getarg_aeolus(2,arg_name)
     call getarg_aeolus(3,arg_value)
     if (trim(arg_name) .eq. "-NumBRC") then
        print *,"Reducing number of read BRC's from ",Num_BRC
        read(arg_value,*,err=999) Num_BRC
        print *,"to ",Num_BRC
     else
        print *,"ERROR: invalid commandline argument: "//trim(arg_name)
        call PrintUsage
        call program_abort(error_cmdline_options,"TestReadL1Bdata")
     end if
  END IF

  ! loop over all BRCs in the file
  brc_loop: DO iBRC=1,Num_BRC
     !  #[
     ! read all datablocks, except the PCD ADS because the reading of that
     ! one still has a problem (data does not correspond to format given
     ! in the L1B_iodd documentation)
     print *,"Reading L1B file, BRC = ",iBRC
     call ReadL1Bfile(L1B_Headers, filename,error_flag, &
          L1B=L1B, Which_DSR=iBRC, &
          ReadHeaders=.false., ReadGeoloc=.true., ReadPCD=.true., & 
          ReadGWD=.true., ReadMeas=.true., ReadUS=.true., ReadWV=.true., &
          ReadCal=.true.)
     !     ReadAll=.true., ReadPCD=.false.)
     IF (error_flag .ne. no_error) call program_abort(error_flag,"ReadL1Bfile")

     print *,"initialising L1B Working arrays..."
     print *,"done initialising working arrays"
     print *,"N_meas = ",L1B_Headers%SPH%N_Meas_Max
     ! TODO: revise these prints when testoutput changes acceptable

     print *,"Deleting L1B Working arrays..."

     print *,"Emptying L1B binary datasets..."

     call Empty_L1B_BinaryDatasets(L1B)
     !  #]
  END DO brc_loop

  print *,"Deleting L1B datastructure"
  call Delete_L1B_HDR_DataStructure(L1B_Headers)
  call Delete_L1B_BRC_DataStructure(L1B)

  goto 111 ! goto the end of the program without using stop
           ! (that messes up our testsuite for certain compilers)

  ! error handler
999 print *,"ERROR: could not convert NumBRC given on the command line"
    print *,"into an integer ...."
    print *,"argument string value is: "//trim(arg_value)
    call PrintUsage
    call program_abort(error_cmdline_options,"TestReadL1Bdata")

111 continue
  !  #]
contains
  !-------------------
  subroutine PrintUsage
    print *,"Usage: TestReadL1Bdata <filename[without extension]>"
    print *,"                       [optional: -NumBRC <count>]"
  end subroutine PrintUsage
  !-------------------
end program TestReadL1Bdata
