module ReadL1Bdata
  !  #[ Description:
  !
  !     a module to read the 2 files describing the L1B product
  !     of the Aeolus instrument
  !
  ! created:     27-06-2005 by J. de Kloe, KNMI
  !
  ! Modifications:
  !   16-Sep-2005  D. Tan      Upgrade to DLR TDS v1.2 (Aug 2005)
  !   29-Sep-2005  J. de Kloe  MPH/SPH improvements
  !   26-Okt-2005  J. de Kloe  repaired memory leaks
  !   30-Nov-2005  P. Poli     Fixed bug on optnl. arg. in ReadL1Bfile
  !   12-Dec-2005  D. Tan      Change defaults for optnl.arg. in ReadL1Bfile
  !   03-Jan-2006  J. de Kloe  small interface change in xml_module
  !   11-Jan-2006  J. de Kloe  added the new DateTime type
  !   21-Feb-2006  P. Poli     Upgrade to DLR TDS Nov 2005
  !   27-Feb-2006  D. Tan      Restore check on size of ref_pulse arrays.
  !   02-Mar-2006  J. de Kloe  adapted to changed XML_module
  !   28-Mar-2006  P. Poli     Upgrade to DLR IODD AE.TN.DLR.APM-L1B.150206
  !   09-May-2006  D. Tan      Generalize size checks to any N_Meas
  !   12-May-2006  D. Tan      Null use of ParKind1 module
  !   31-Jul-2006  J. de Kloe  renamed set_file_pointer to SetFilePointer
  !                            and advance_file_pointer to AdvanceFilePointer
  !   17-Aug-2006  P. Poli     Removed obsolete variables
  !   03-Nov-2006  P. Poli     upgrade to L1BP v1.04
  !   06-Nov-2006  P. Poli     added mods to handle IODD mistake
  !   20-Nov-2006  D. Tan      upgrade to L1BP v1.04, post-FAT 20061114
  !   23-Mar-2007  J. de Kloe  passing workaround flag to Read_MPH_KVT()
  !                            routine to allow it to report a warning
  !                            in stead of an error.
  !   07-May-2007  J. de Kloe  moved dsd handling into Read_L1B_SPH_KVT
  !                            and adapted to enable reading L1BP v1.06 files
  !   07-Jun-2007  J. de Kloe  adapted to changed xml interface
  !                            and removed xml workarounds
  !   15-Jun-2007  J. de Kloe  changed VersioNumber handling to use integers
  !                            to easier allow a check like version .ge. 1.06
  !   31-Aug-2007  P. Poli     modified ReadL1Bfile to allow using L1B header
  !                            information from another datastructure
  !                            change L1B_DataType to L1B_BRC_DataType and
  !                            L1B_HDR_DataType
  !   18-Oct-2007  J. de Kloe  added a nullify and fixed an intent
  !   13-Nov-2007  J. de Kloe  added check for xmlns attribute in xml file
  !   16-Jan-2008  J. de Kloe  phase out integer kind i_
  !   21-Feb-2008  J. de Kloe  adapted to L1B v1.09, and made the workarounds
  !                            that are used for xml handling explicit
  !  #]
  !  #[ Modules used:

  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE L1B_DataStructure, only: L1B_HDR_DataType, L1B_BRC_DataType
  USE DataSetDescriptor, only: Print_DSD
  USE L1B_Geoloc_ADS, only: Init_L1B_Geoloc_ADS, Read_L1B_Geoloc_ADS
  USE L1B_PCD_ADS, only: Init_L1B_PCD_ADS, Read_L1B_PCD_ADS
  USE L1B_GWD_ADS, only: Init_L1B_GWD_ADS, Read_L1B_GWD_ADS
  USE L1B_Meas_ADS, only: Init_L1B_Meas_ADS, Read_L1B_Meas_ADS
  USE L1B_US_MDS, only: Init_L1B_US_MDS, Read_L1B_US_MDS
  USE L1B_WV_MDS, only: Init_L1B_WV_MDS, Read_L1B_WV_MDS
  USE L1B_Cal_ADS, only: Init_L1B_Cal_ADS, Read_L1B_Cal_ADS
  USE L1B_SPH, only: Print_L1B_SPH, Read_L1B_SPH_KVT, &
                     Get_L1B_SPH_from_xml_tree
  USE FixedHeader, only: Print_FH, FH_type, Get_FH_from_xml_tree,&
                         file_type_L1B_N
  USE MainProductHeader, only: Print_MPH, MPH_type, &
                               Get_MPH_from_xml_tree, &
                               Read_MPH_from_KVT, maxlen_Software_Ver, &
                               GetFileVersion
  USE XML_module, only: Init_XML, xml_load_file, delete_xml_tree, &
                        xml_tree_print, xml_node_type, xml_node_get_string, &
                        xml_check_node_exists, ee_h_name, &
                        xml_tree_get_namespace, uri_ns_header
  USE KVT_module, only: max_str_len, Init_KVT_module, &
                        read_kvt_string, read_spare_kvt_string
  USE DirectBinaryIO, only: Init_DirectBinaryIO_module, &
                            read_int1_from_bin_file, &
                            read_logical1_from_bin_file, &
                            SetFilePointer, AdvanceFilePointer, &
                            BigEndian

  USE EE_CFI_Datatypes, only: GetIntAl, DateTimeType,&
                            read_DateTime_from_bin_file, &
                            read_FAdoxy_from_bin_file, FAdoxy, &
                            read_IntAl_from_bin_file, IntAl, &
                            read_IntAs_from_bin_file, IntAs, &
                            read_IntAuc_from_bin_file, IntAuc, &
                            read_IntAul_from_bin_file, IntAul

  USE ErrorHandler, only: no_error, error_programming, &
                          error_opening_file, error_reading_file,&
                          error_writing_file, error_allocate
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE LunManager, only: get_lun, free_lun
  USE c_support, only: get_filesize
  USE AeolusConstants, only: iodd_version_L1B_N, iodd_version_L1B_N_108
  USE Logging, only: logmsg, &
                     log_debug,log_info,log_progress,log_warn,log_error

  !  #]  
  !  #[ Parameter/Variable/Type declarations

  implicit none

  integer, parameter :: N_reference_pulse_Max = 1003

  ! make all module contents private, unless explicitely stated otherwise
  private

! TODO: phase out use of the Verbosity module
!       since it is outdated and all outputs should be redirected
!       to the logging module anyway
!       The verbosity feature is currently never used anyway in any of our
!       test programs, so not a single output will change probably...

  ! the default verbosity setting is Err (as silent as possible)
  integer, save :: LocalVerbosity = VerbosityError

  ! remember the initialisation state of the module
  logical, save :: module_is_initialised = .false.

  !  #]
  !  #[ Module interface
  public :: ReadL1Bfile, InitReadL1BfileModule
  !  #]
contains
  !---------------------------------
  subroutine InitReadL1BfileModule(error_flag, ChoosenVerbosity)
    !  #[
    integer,           intent(out) :: error_flag
    integer, optional, intent(in)  :: ChoosenVerbosity

    error_flag = no_error

    IF (present(ChoosenVerbosity)) LocalVerbosity = ChoosenVerbosity

    IF (LocalVerbosity .ge. VerbosityReport) &
         call print_verbosity_setting(LocalVerbosity,ModuleName="ReadL1Bfile")

    call Init_KVT_module()

    call Init_DirectBinaryIO_module(error_flag,FileByteOrder=BigEndian)
    if (error_flag .ne. no_error) return

    call Init_XML(error_flag) ! (ChoosenVerbosity=VerbosityReport)
    IF (error_flag .ne. no_error) return

    module_is_initialised = .true.

  end subroutine InitReadL1BfileModule
    !  #]
  subroutine ReadL1Bfile(L1B_Headers, filename, error_flag, L1B, Which_DSR, &
       ReadAll, ReadHeaders, ReadGeoloc, ReadPCD, ReadGWD, ReadMeas, &
       ReadCal, ReadUS, ReadWV)
    !  #[

    TYPE(L1B_HDR_DataType),intent(inout) :: L1B_Headers
    character(len=*),      intent(in)    :: filename ! without extension
    integer,               intent(out)   :: error_flag
    TYPE(L1B_BRC_DataType), optional, intent(inout) :: L1B
    integer,     optional, intent(in)    :: Which_DSR
    logical,     optional, intent(in)    :: ReadAll ! headers and datasets
    logical,     optional, intent(in)    :: ReadHeaders
    logical,     optional, intent(in)    :: ReadGeoloc
    logical,     optional, intent(in)    :: ReadPCD
    logical,     optional, intent(in)    :: ReadGWD
    logical,     optional, intent(in)    :: ReadMeas
    logical,     optional, intent(in)    :: ReadCal
    logical,     optional, intent(in)    :: ReadUS
    logical,     optional, intent(in)    :: ReadWV

    ! local variables
    character(len=max_str_len)         :: filename_hdr
    character(len=max_str_len)         :: filename_dbl
    character(len=maxlen_Software_Ver) :: Software_Version_String
    logical                            :: do_ReadHeaders
    integer                            :: Software_Version_Major_Number
    integer                            :: Software_Version_Minor_Number

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in ReadL1Bfile:"
       print *,"the ReadL1Bdata module should be initialised before use !"
       error_flag = error_programming
       return
    END IF

    filename_hdr = trim(filename)//".HDR"
    filename_dbl = trim(filename)//".DBL"

    do_ReadHeaders = .false.
    IF (present(ReadHeaders)) do_ReadHeaders = ReadHeaders
    IF (present(ReadAll)) THEN
       IF (ReadAll) do_ReadHeaders = .true.
    END IF

    ! retrieve the file version, to enable activating workarounds
    ! needed to read the different L1B products
    ! possible results for L1bP: 
    ! ADM_L1bP/1.04, ADM_L1bP/1.05, ADM_L1bP/1.06
    ! I don't have example files at hand for versions 1.02 and 1.03
    ! note that the L1b product files from L1bP_InstallTest.1.0.1.tar.gz
    ! report version ADM_L1bP/1.00 rather than ADM_L1bP/1.01
    call GetFileVersion(filename_hdr, Software_Version_String,&
         Software_Version_Major_Number, Software_Version_Minor_Number,&
         error_flag)
    IF (error_flag .ne. no_error) return

    if ( (Software_Version_Major_Number .lt. 1)          .or. &
         ( (Software_Version_Major_Number .eq. 1) .and.       &
           (Software_Version_Minor_Number .lt. 4)       )    ) then
       print *,"ERROR: this L1B reader is currently not compatible with the"
       print *,"version of the L1B product file that you are trying to read."
       print *,"L1B Reading is compatible with L1B file versions: "
       print *,"   ADM_L1bP/1.04 and upwards"
       print *,""
       print *,"The current file seems to have"
       print *,"Software_Version: ",trim(Software_Version_String)
       error_flag = error_reading_file
       return
    end if

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"Software_Version_String = ",trim(Software_Version_String)

    IF (do_ReadHeaders) THEN
       ! needs to be called just once per set of HDR/DBL files
       call ReadL1Bfile_HDR(L1B_Headers, filename_hdr, Software_Version_Major_Number,&
                            Software_Version_Minor_Number, error_flag)
       IF (error_flag .ne. no_error) return
! to add: check if the filename is consistent with the file contents
    END IF

    ! needs to be called for every BRC present in a set of HDR/DBL files
    call ReadL1Bfile_DBL(L1B_Headers, filename_dbl, Software_Version_Major_Number, &
         Software_Version_Minor_Number, error_flag, &
         L1B, Which_DSR, &
         ReadAll, ReadHeaders, ReadGeoloc, ReadPCD, ReadGWD, ReadMeas, &
         ReadCal, ReadUS, ReadWV)
    IF (error_flag .ne. no_error) return

    return

  end subroutine ReadL1Bfile
    !  #]
  !---------------------------------
  ! private routines

! TODO: phase out the *_is_available variables.
!       These are unnecessary. A check on the allocated() status 
!       gives the same information. Having two ways to determine
!       the same thing just makes things more confusing.

  subroutine ReadL1Bfile_HDR(L1B_Headers, filename_hdr, Software_Version_Major_Number,&
                             Software_Version_Minor_Number, error_flag)
    !  #[ read all XML values

    TYPE(L1B_HDR_DataType),intent(inout):: L1B_Headers
    character(len=*),   intent(in)    :: filename_hdr
    integer,            intent(in)    :: Software_Version_Major_Number
    integer,            intent(in)    :: Software_Version_Minor_Number
    integer,            intent(out)   :: error_flag

    ! local variables
    character(len=256)           :: path
    type(xml_node_type), pointer :: xml_tree
    character(len=256)           :: xml_namespace_L1B
    character(len=256)           :: xml_namespace_uri
    logical                      :: workaround_pre108, workaround_108

    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in ReadL1Bfile_HDR:"
       print *,"the ReadL1Bdata module should be initialised before use !"
       error_flag = error_programming
       return
    END IF

!    IF (LocalVerbosity .ge. VerbosityReport) THEN
!       print *,"reading data from L1B header file: ",trim(filename_hdr)
!       print *,"trying to read this file for Software_Version: ",&
!            Software_Version
!    END IF

    ! implemented workarounds:
    ! ==>workaround_pre108
    ! allow the use of xml DateTime fields with only second precision
    ! This has been changed to microsecond precision for the MPH
    ! in L1B v1.08 
    ! ==>workaround_108
    ! allow the <Acquisition_Station> field to have 24 characters
    ! in stead of 20 characters
    ! This IODD change is present in L1B v1.08 but this was a mistake
    ! and has been reverted in L1B v1.09
    ! Also allow the xmlns attribute to refer to IODD v3.04 in stead
    ! of the current v3.05 (for L1B v1.09 files)

    workaround_pre108 = .false.
    IF ( (Software_Version_Major_Number .eq. 1) .and. &
         (Software_Version_Minor_Number .lt. 8)       ) THEN
       workaround_pre108 = .true.
    END IF

    workaround_108 = .false.
    IF ( (Software_Version_Major_Number .eq. 1) .and. &
         (Software_Version_Minor_Number .eq. 8)       ) THEN
       workaround_108 = .true.
    END IF

    nullify(xml_tree)
    call xml_load_file(filename_hdr, xml_tree, error_flag)
    IF (error_flag .ne. no_error) return

    ! define which xml namespace should be present
    !example: xmlns="http://www.esa.int/schemas/ae/ALD_U_N_2B_01.01"
    !uri_ns_header    ! defined in xml_module
    !file_type_L2B_N  ! defined in FixedHeader
    !iodd_version_L2B ! defined in AeolusConstants


    xml_namespace_L1B = trim(uri_ns_header)//trim(file_type_L1B_N)//&
                        "_"//trim(iodd_version_L1B_N)

    if (workaround_108) then
       xml_namespace_L1B = trim(uri_ns_header)//trim(file_type_L1B_N)//&
                           "_"//trim(iodd_version_L1B_N_108)
    end if

    ! check whether this namespace is really present in the input file
    call xml_tree_get_namespace(xml_tree,xml_namespace_uri,error_flag)
    if (error_flag .ne. no_error) then
       ! allow this attribute to be missing for now, to allow use
       ! of outdated files.
       ! TODO: remove this exception
       error_flag = no_error
       call logmsg(log_warn,"WARNING in ReadL1Bfile_HDR: "//&
                   "could not find xmlns attribute in the root tag "//&
                   "of file: "//trim(filename_hdr))
    else
       if (trim(xml_namespace_L1B) .ne. trim(xml_namespace_uri)) &
            call logmsg(log_warn,"WARNING in ReadL1Bfile_HDR: "//&
                        "expected an xmlns attribute with content: "//&
                        trim(xml_namespace_L1B)//", but found an "//&
                        "xmlns attribute with content: "//&
                        trim(xml_namespace_uri)//". This input file may "//&
                        "not be compatible with the current software!")
    end if


    call logmsg(log_debug,"ReadL1Bfile_HDR: the xml file was "//&
                          "successfully read")
    
    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"==================="
       call xml_tree_print(xml_tree, error_flag)
       IF (error_flag .ne. no_error) return
       print *,"==================="
    END IF

    !------------------------------
    ! copying the FH XML elements
    !------------------------------
    print *,"copying items to the FH datastructure"
    IF (.not. associated(L1B_Headers%FH)) THEN
       print *,"Error in ReadL1Bfile_DBL:"
       print *,"L1B_Headers%FH seems not associated !"
       print *,"this should have been done in Init_L1B_HDR_DataStructure"
       error_flag = error_programming
       return
    ENDIF
    path = "/"//ee_h_name
    call Get_FH_from_xml_tree(xml_tree, path, L1B_Headers%FH, error_flag)
    IF (error_flag .ne. no_error) return
    L1B_Headers%FH_is_available = .true.

    !------------------------------
    ! copying the MPH XML elements
    !------------------------------
    print *,"copying items to the MPH datastructure"
    IF (.not. associated(L1B_Headers%MPH)) THEN
       print *,"Error in ReadL1Bfile_DBL:"
       print *,"L1B_Headers%MPH seems not associated !"
       print *,"this should have been done in Init_L1B_HDR_DataStructure"
       error_flag = error_programming
       return
    ENDIF
    call Get_MPH_from_xml_tree(xml_tree, path, L1B_Headers%MPH, error_flag,&
                               workaround1=workaround_108,&
                               workaround2=workaround_pre108)
    IF (error_flag .ne. no_error) return
    L1B_Headers%MPH_is_available = .true.

    !------------------------------
    ! copying the SPH XML elements (with DSD's)
    !------------------------------
    print *,"copying items to the SPH datastructure"
    IF (.not. associated(L1B_Headers%SPH)) THEN
       print *,"Error in ReadL1Bfile_DBL:"
       print *,"L1B_Headers%SPH seems not associated !"
       print *,"this should have been done in Init_L1B_HDR_DataStructure"
       error_flag = error_programming
       return
    ENDIF
    call Get_L1B_SPH_from_xml_tree(xml_tree, path, L1B_Headers%SPH, &
                                   Software_Version_Major_Number,&
                                   Software_Version_Minor_Number, error_flag)
    IF (error_flag .ne. no_error) return
    L1B_Headers%SPH_is_available = .true.

    !------------------------------
    ! done
    !------------------------------
!    IF (LocalVerbosity .ge. VerbosityReport) THEN
!       print *,"L1B%FH%..."
!       call Print_FH(L1B_Headers%FH)
!       print *,"L1B%MPH%..."
!       call Print_MPH(L1B_Headers%MPH)
!       print *,"L1B%SPH%..."
!       call Print_L1B_SPH(L1B_Headers%SPH,error_flag)
!    END IF

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"deallocating xml items"
    call delete_xml_tree(xml_tree, error_flag)
    IF (error_flag .ne. no_error) return

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"done with reading L1B header file"

    return

  end subroutine ReadL1Bfile_HDR

    !  #]
  subroutine ReadL1Bfile_DBL(L1B_Headers, filename_dbl, &
       Software_Version_Major_Number, &
       Software_Version_Minor_Number, error_flag,&
       L1B, Which_DSR, &
       ReadAll, ReadHeaders, ReadGeoloc, ReadPCD, ReadGWD, ReadMeas, &
       ReadCal, ReadUS, ReadWV)
    !  #[ read all SVT and binary values

    TYPE(L1B_HDR_DataType),intent(inout) :: L1B_Headers
    character(len=*),      intent(in)    :: filename_dbl
    integer,               intent(in)    :: Software_Version_Major_Number
    integer,               intent(in)    :: Software_Version_Minor_Number
    integer,               intent(out)   :: error_flag
    TYPE(L1B_BRC_DataType), optional, intent(inout) :: L1B
    ! this routine may also be used to read an additional dataset,
    ! while others are already in memory. For example you may want
    ! to read the US dataset, when WV is already available.
    ! Therefore L1B should have intent(inout).
    integer,     optional, intent(in)    :: Which_DSR
    logical,     optional, intent(in)    :: ReadAll ! headers and datasets
    logical,     optional, intent(in)    :: ReadHeaders
    logical,     optional, intent(in)    :: ReadGeoloc
    logical,     optional, intent(in)    :: ReadPCD
    logical,     optional, intent(in)    :: ReadGWD
    logical,     optional, intent(in)    :: ReadMeas
    logical,     optional, intent(in)    :: ReadCal
    logical,     optional, intent(in)    :: ReadUS
    logical,     optional, intent(in)    :: ReadWV

    ! local variables
    integer            :: fileunit, N_Meas_Max, NF_Order_Max, N_range_gates
    integer            :: DataSetRecord_to_read
    integer            :: filepos
    logical            :: size_check_ok

    ! switches to copy the optional input switches to
    logical            :: do_ReadHeaders
    logical            :: do_ReadGeoloc
    logical            :: do_ReadPCD
    logical            :: do_ReadGWD
    logical            :: do_ReadMeas
    logical            :: do_ReadCal
    logical            :: do_ReadUS
    logical            :: do_ReadWV

    ! the current binary dataset definition that has been implemented
    ! is the version described in L1B_iodd version 2/1, dd. 06-may-2005
    !! DT: upgraded to the DLR version of August 2005.
    !!     The N=30 expected sizes have been read from the DSDs in the HDR file.
    !!     The N=67 expected sized have been read from the DSDs in the HDR file.
    !!     Commented values are from AE.TN.DLR.IODD-L1B.100605.
    ! PP: upgraded to the DLR version of Nov 2005 (Dec 2006 for PCD_ADS)
    ! PP: upgraded to the DLR version 1.3 of Feb 2006
    ! PP: upgraded to L1BP v1.04 november 2006
    ! DT: upgraded to L1BP v1.04, post-FAT, 14 november 2006
    

    ! Contributions to dataset sizes independent of L1B%SPH%N_Meas_Max
    !
    integer, parameter :: expected_size_Geoloc_0       =  2244
    integer, parameter :: expected_size_PCD_0          =  3799
    ! for file versions 1.07 or below:
    integer, parameter :: expected_size_GWD_0_old      =   130
    ! for file versions 1.08 or higher:
    integer, parameter :: expected_size_GWD_0          =   138
    integer, parameter :: expected_size_Measurement_0  =   220
    integer, parameter :: expected_size_Cal_0          =  6390
    integer, parameter :: expected_size_UsefulSignal_0 =   662
    integer, parameter :: expected_size_WindVel_0      =   447
    !
    ! Contributions to dataset sizes linearly proportional to L1B%SPH%N_Meas_Max
    !
    integer, parameter :: expected_size_Geoloc_1       =   924
    integer, parameter :: expected_size_PCD_1          =  2022
    integer, parameter :: expected_size_GWD_1          =    20
    integer, parameter :: expected_size_Measurement_1  =  1061
    integer, parameter :: expected_size_Cal_1          =     0
    integer, parameter :: expected_size_UsefulSignal_1 =   650
    integer, parameter :: expected_size_WindVel_1      =   434
    !
    ! Contributions to dataset sizes linearly proportional to L1B%SPH%NF_Order
    !
    integer, parameter :: expected_size_Geoloc_2       =     0
    integer, parameter :: expected_size_PCD_2          =     0
    integer, parameter :: expected_size_GWD_2          =     0
    integer, parameter :: expected_size_Measurement_2  =     0
    integer, parameter :: expected_size_Cal_2          =    32
    integer, parameter :: expected_size_UsefulSignal_2 =     0
    integer, parameter :: expected_size_WindVel_2      =     0

!    integer, parameter :: exp_size_Cal_DSR                   =  9354 ! CHECK DS / MODS MADE
!                             1234567890123456789012345678901

    integer :: expected_size_Geoloc_DSR
    integer :: expected_size_PCD_DSR
    integer :: expected_size_GWD_DSR
    integer :: expected_size_Measurement_DSR
    integer :: expected_size_Cal_DSR != exp_size_Cal_DSR
    integer :: expected_size_UsefulSignal_DSR
    integer :: expected_size_WindVel_DSR
    logical :: workaround_pre106, workaround_pre108, workaround_108

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in ReadL1Bfile_DBL():"
       print *,"the ReadL1Bdata module should be initialised before use !"
       error_flag = error_programming
       return
    END IF

    DataSetRecord_to_read = 1 ! by default just read the 1st dataset
    IF (present(Which_DSR)) THEN
       DataSetRecord_to_read = Which_DSR
    END IF

    do_ReadHeaders = .false.
    do_ReadGeoloc  = .false.
    do_ReadPCD     = .false.
    do_ReadGWD     = .false.
    do_ReadMeas    = .false.
    do_ReadCal     = .false.
    do_ReadUS      = .false.
    do_ReadWV      = .false.
    
    IF (present(ReadAll)) THEN
       do_ReadHeaders = .true.
       do_ReadGeoloc  = .true.
       do_ReadPCD     = .true.
       do_ReadGWD     = .true.
       do_ReadMeas    = .true.
       do_ReadCal     = .true.
       do_ReadUS      = .true.
       do_ReadWV      = .true.
    END IF

    IF (present(ReadHeaders)) do_ReadHeaders = ReadHeaders
    IF (present(ReadGeoloc )) do_ReadGeoloc  = ReadGeoloc
    IF (present(ReadPCD    )) do_ReadPCD     = ReadPCD
    IF (present(ReadGWD    )) do_ReadGWD     = ReadGWD
    IF (present(ReadCal    )) do_ReadCal     = ReadCal
    IF (present(ReadUS     )) do_ReadUS      = ReadUS
    IF (present(ReadMeas   )) do_ReadMeas    = ReadMeas
    IF (present(ReadWV     )) do_ReadWV      = ReadWV

    ! if nothing was selected to be read, just read everything
    IF ( .not. do_ReadHeaders .and. &
         .not. do_ReadGeoloc  .and. &
         .not. do_ReadPCD     .and. &
         .not. do_ReadGWD     .and. &
         .not. do_ReadMeas    .and. &
         .not. do_ReadCal    .and. &
         .not. do_ReadUS     .and. &
         .not. do_ReadWV            ) THEN
       do_ReadHeaders = .true.
       do_ReadGeoloc  = .true.
       do_ReadPCD     = .true.
       do_ReadGWD     = .true.
       do_ReadMeas    = .true.
       do_ReadCal     = .true.
       do_ReadUS      = .true.
       do_ReadWV      = .true.
    ENDIF

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"reading data from L1B datafile: ",trim(filename_dbl)

    ! implemented workarounds:
    ! ==>workaround_pre106
    ! allow the test (actual_product_size .ne. total_product_size) to fail
    ! This is needed to be able to use L1B v1.04/1.05 files
    !
    ! ==>workaround_pre108
    ! use IntAl for the Mie_ch_tot_zero_wind_corr and
    ! Rayl_ch_tot_zero_wind_corr fields in the l1b_gwd datastructure 
    ! in stead of FAdoxy (which is used from 108 upwards)
    !
    ! ==>workaround_108
    ! allow the <Acquisition_Station> field to have 24 characters
    ! in stead of 20 characters
    ! This IODD change is present in L1B v1.08 but this was a mistake
    ! and has been reverted in L1B v1.09

    workaround_pre106 = .false.
    IF ( (Software_Version_Major_Number .eq. 1) .and. &
         (Software_Version_Minor_Number .lt. 6)       ) THEN
       workaround_pre106 = .true.
    END IF
    
    workaround_pre108 = .false.
    IF ( (Software_Version_Major_Number .eq. 1) .and. &
         (Software_Version_Minor_Number .lt. 8)       ) THEN
       workaround_pre108 = .true.
    END IF

    workaround_108 = .false.
    IF ( (Software_Version_Major_Number .eq. 1) .and. &
         (Software_Version_Minor_Number .eq. 8)       ) THEN
       workaround_108 = .true.
    END IF
    
    !------------------------------
    ! open the binaryfile
    !------------------------------
    fileunit = get_lun()
    IF (fileunit .lt. 0) THEN
       error_flag = error_opening_file
       print *,"ERROR in ReadL1Bfile_DBL():"
       print *,"no more free fileunits seem available ..."
       error_flag = error_reading_file
       return
    ENDIF
    
    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"opening file:"
       print *,trim(filename_dbl)
    END IF

    open(unit=fileunit,file=filename_dbl,status="old",&
         action="read",access="direct",recl=1,form="UNFORMATTED",ERR=999)
    
    IF (do_ReadHeaders) THEN
       !  #[

       !------------------------------
       ! read and/or compare MPH elements
       !------------------------------

       IF (.not. L1B_Headers%MPH_is_available) THEN

          ! the MPH is allocated and filled when reading the XML file
          ! so this part of the code is usually skipped.
          
          IF (.not. associated(L1B_Headers%MPH)) THEN
             print *,"Error in ReadL1Bfile_DBL:"
             print *,"L1B_Headers%MPH seems not associated !"
             print *,"this should have been done in Init_L1B_HDR_DataStructure"
             error_flag = error_programming
             return
          END IF

          call Read_MPH_from_KVT(L1B_Headers%MPH, fileunit, error_flag, &
                                 workaround1=workaround_pre106,&
                                 workaround2=workaround_108)
          IF (error_flag .ne. no_error) return
       
          L1B_Headers%MPH_is_available = .true.
       ELSE
          ! Poli: modifed call. Now replace the values in memory by what is read.
          ! read the KVT values anyway, and compare them to what was
          ! read from the XML file
          call Read_MPH_from_KVT(L1B_Headers%MPH, fileunit, error_flag,&
                                 workaround1=workaround_pre106,&
                                 workaround2=workaround_108,&
                                 Compare=.true.)
!                                 CompareOnly=.true.)
          IF (error_flag .ne. no_error) return
       END IF

       ! an extra check
       ! this check is only possible if the FH is already loaded 
       IF (L1B_Headers%FH_is_available) THEN
          IF (associated(L1B_Headers%FH)) THEN
             IF (trim(L1B_Headers%MPH%PRODUCT) .ne. trim(L1B_Headers%FH%File_Name)) THEN
                print *,"ERROR in ReadL1Bfile_DBL:"
                print *,"FH File name seems not to match with MPH productname!"
                print *,"L1B%FH%File_Name = ",trim(L1B_Headers%FH%File_Name)
                print *,"L1B%MPH%PRODUCT  = ",trim(L1B_Headers%MPH%PRODUCT)
                error_flag = error_reading_file
                return
             END IF
          END IF
       END IF
   
       !------------------------------
       ! read and/or compare SPH elements
       !------------------------------
       IF (.not.L1B_Headers%SPH_is_available) THEN

          ! the MPH is allocated and filled when reading the XML file
          ! so this part of the code is usually skipped.

          IF (.not. associated(L1B_Headers%SPH)) THEN
             print *,"Error in ReadL1Bfile_DBL:"
             print *,"L1B_Headers%SPH seems not associated !"
             print *,"this should have been done in Init_L1B_HDR_DataStructure"
             error_flag = error_programming
             return
          END IF
       
          call logmsg(log_debug,"ReadL1Bfile_DBL: calling Read_L1B_SPH_KVT")
          call Read_L1B_SPH_KVT(L1B_Headers%SPH,fileunit,&
                                Software_Version_Major_Number,&
                                Software_Version_Minor_Number,&
                                error_flag)
          IF (error_flag .ne. no_error) return

          L1B_Headers%SPH_is_available = .true.
       ELSE
          call logmsg(log_debug,"ReadL1Bfile_DBL: calling Read_L1B_SPH_KVT")
          call Read_L1B_SPH_KVT(L1B_Headers%SPH,fileunit,&
                                Software_Version_Major_Number,&
                                Software_Version_Minor_Number,&
                                error_flag,Compare=.true.)
          IF (error_flag .ne. no_error) return
       END IF
       
       !  #[ check if the DS's have the expected size

       ! Todo: make this size_checking consistent with the way it is done for 
       ! the L2B product file reading. For this we have to implement 
       ! GetByteSize functions for all datastructures which makes the module 
       ! dependent on the actual dataset,
       ! which can only be allocated after the SPH has been read !!!
       ! therefore this test must stay in the ReadL1Bdata module and cannot be
       ! moved into the l1b_sph.F90 module.
       ! JK. 07-May-2007

       size_check_ok = .true.
       
       ! at this point we know how many measurements each BRC holds, 
       ! the size checks depend on this...
       
       expected_size_Geoloc_DSR       = expected_size_Geoloc_0       + &
           L1B_Headers%SPH%N_Meas_Max * expected_size_Geoloc_1 + &
           L1B_Headers%SPH%NF_Order_Max * expected_size_Geoloc_2
       expected_size_PCD_DSR          = expected_size_PCD_0          + &
           L1B_Headers%SPH%N_Meas_Max * expected_size_PCD_1 + &
           L1B_Headers%SPH%NF_Order_Max * expected_size_PCD_2

       if (workaround_pre108) then
          expected_size_GWD_DSR          = expected_size_GWD_0_old   + &
               L1B_Headers%SPH%N_Meas_Max * expected_size_GWD_1 + &
               L1B_Headers%SPH%NF_Order_Max * expected_size_GWD_2
       else
          expected_size_GWD_DSR          = expected_size_GWD_0          + &
               L1B_Headers%SPH%N_Meas_Max * expected_size_GWD_1 + &
               L1B_Headers%SPH%NF_Order_Max * expected_size_GWD_2
       end if
       expected_size_Measurement_DSR  = expected_size_Measurement_0  + &
           L1B_Headers%SPH%N_Meas_Max * expected_size_Measurement_1 + &
           L1B_Headers%SPH%NF_Order_Max * expected_size_Measurement_2
       expected_size_Cal_DSR          = expected_size_Cal_0          + &
           L1B_Headers%SPH%N_Meas_Max * expected_size_Cal_1 + &
           L1B_Headers%SPH%NF_Order_Max * expected_size_Cal_2
       expected_size_UsefulSignal_DSR = expected_size_UsefulSignal_0 + &
           L1B_Headers%SPH%N_Meas_Max * expected_size_UsefulSignal_1 +&
           L1B_Headers%SPH%NF_Order_Max * expected_size_UsefulSignal_2
       expected_size_WindVel_DSR      = expected_size_WindVel_0      + &
           L1B_Headers%SPH%N_Meas_Max * expected_size_WindVel_1 + &
           L1B_Headers%SPH%NF_Order_Max * expected_size_WindVel_2
       
       IF (L1B_Headers%SPH%Geolocation_DSD%size_of_one_dsr .ne. &
            expected_size_Geoloc_DSR ) THEN
          print *,"WARNING: the size of the Geolocation_ADS records: ",&
              L1B_Headers%SPH%Geolocation_DSD%size_of_one_dsr
          print *,"differs from the expected size of: ",&
               expected_size_Geoloc_DSR
          size_check_ok = .false.
       ELSE
          print *,"the size of the Geolocation_ADS records seems OK "
       END IF
       
       IF (L1B_Headers%SPH%Product_Conf_Data_DSD%size_of_one_dsr .ne. &
            expected_size_PCD_DSR ) THEN
          print *,"WARNING: the size of the PCD_ADS records: ",&
              L1B_Headers%SPH%Product_Conf_Data_DSD%size_of_one_dsr 
          print *,"differs from the expected size of: ",expected_size_PCD_DSR
          size_check_ok = .false.
       ELSE
          print *,"the size of the PCD_ADS records seems OK"
       END IF
       
       IF (L1B_Headers%SPH%Ground_Wind_Detection_DSD%size_of_one_dsr .ne. &
            expected_size_GWD_DSR ) THEN
          print *,"WARNING: the size of the Ground_Wind_Detection_DSD records: ",&
              L1B_Headers%SPH%Ground_Wind_Detection_DSD%size_of_one_dsr 
          print *,"differs from the expected size of: ",expected_size_GWD_DSR
          size_check_ok = .false.
       ELSE
          print *,"the size of the Ground_Wind_Detection_DSD records seems OK"
       END IF
       
       IF (L1B_Headers%SPH%Measurement_DSD%size_of_one_dsr .ne. &
            expected_size_Measurement_DSR ) THEN
          print *,"WARNING: the size of the Measurement_DSD records: ",&
              L1B_Headers%SPH%Measurement_DSD%size_of_one_dsr 
          print *,"differs from the expected size of: ",&
               expected_size_Measurement_DSR
          size_check_ok = .false.
       ELSE
          print *,"the size of the Measurement_DSD records seems OK"
       END IF
       
       IF (L1B_Headers%SPH%Cal_DSD%size_of_one_dsr .ne. &
            expected_size_Cal_DSR ) THEN
          print *,"WARNING: the size of the Calibration_DSD records: ",&
              L1B_Headers%SPH%Cal_DSD%size_of_one_dsr 
          print *,"differs from the expected size of: ",&
               expected_size_Cal_DSR
          size_check_ok = .false.
       ELSE
          print *,"the size of the Calibration_Char_DSD records seems OK"
       END IF
       
       IF (L1B_Headers%SPH%Useful_Signal_DSD%size_of_one_dsr .ne. &
            expected_size_UsefulSignal_DSR ) THEN
          print *,"WARNING: the size of the Useful_Signal_DSD records: ",&
              L1B_Headers%SPH%Useful_Signal_DSD%size_of_one_dsr 
          print *,"differs from the expected size of: ",&
               expected_size_UsefulSignal_DSR
          size_check_ok = .false.
       ELSE
          print *,"the size of the Useful_Signal_DSD records seems OK"
       END IF
       
       IF (L1B_Headers%SPH%Wind_Velocity_DSD%size_of_one_dsr .ne. &
            expected_size_WindVel_DSR ) THEN
          print *,"WARNING: the size of the Wind_Velocity_DSD records: ",&
              L1B_Headers%SPH%Wind_Velocity_DSD%size_of_one_dsr 
          print *,"differs from the expected size of: ",&
               expected_size_WindVel_DSR
          size_check_ok = .false.
       ELSE
          print *,"the size of the Wind_Velocity_DSD records seems OK"
       END IF
       
       ! the next 9 datasets are in external files, checking their
       ! sizes has not yet been implemented

       !  #]

       !  #[ print binary dataset start/end filepointers

       !     if something is wrong (for debugging)
       IF ( .not. size_check_ok          ) THEN
          
          print *,"-------------------------------------------------"
          ! check filepos
          inquire(unit=fileunit,nextrec=filepos)
          print *,"filepos = ",filepos
          
          print *,"L1B%SPH%Geolocation_DSD%start_of_ds           = ",&
               L1B_Headers%SPH%Geolocation_DSD%start_of_ds
          print *,"L1B%SPH%Geolocation_DSD%end_of_ds             = ",&
               L1B_Headers%SPH%Geolocation_DSD%end_of_ds
          
          print *,"L1B%SPH%Product_Conf_Data_DSD%start_of_ds     = ",&
               L1B_Headers%SPH%Product_Conf_Data_DSD%start_of_ds
          print *,"L1B%SPH%Product_Conf_Data_DSD%end_of_ds       = ",&
               L1B_Headers%SPH%Product_Conf_Data_DSD%end_of_ds
          
          print *,"L1B%SPH%Ground_Wind_Detection_DSD%start_of_ds = ",&
               L1B_Headers%SPH%Ground_Wind_Detection_DSD%start_of_ds
          print *,"L1B%SPH%Ground_Wind_Detection_DSD%end_of_ds   = ",&
               L1B_Headers%SPH%Ground_Wind_Detection_DSD%end_of_ds
          
          print *,"L1B%SPH%Measurement_DSD%start_of_ds           = ",&
               L1B_Headers%SPH%Measurement_DSD%start_of_ds
          print *,"L1B%SPH%Measurement_DSD%end_of_ds             = ",&
               L1B_Headers%SPH%Measurement_DSD%end_of_ds
          
          print *,"L1B%SPH%Cal_Char_DSD%start_of_ds              = ",&
               L1B_Headers%SPH%Cal_DSD%start_of_ds
          print *,"L1B%SPH%Cal_Char_DSD%end_of_ds                = ",&
               L1B_Headers%SPH%Cal_DSD%end_of_ds
          
          print *,"L1B%SPH%Useful_Signal_DSD%start_of_ds         = ",&
               L1B_Headers%SPH%Useful_Signal_DSD%start_of_ds
          print *,"L1B%SPH%Useful_Signal_DSD%end_of_ds           = ",&
               L1B_Headers%SPH%Useful_Signal_DSD%end_of_ds
          
          print *,"L1B%SPH%Wind_Velocity_DSD%start_of_ds         = ",&
               L1B_Headers%SPH%Wind_Velocity_DSD%start_of_ds
          print *,"L1B%SPH%Wind_Velocity_DSD%end_of_ds           = ",&
               L1B_Headers%SPH%Wind_Velocity_DSD%end_of_ds
          
          print *,"-------------------------------------------------"
          print *,"Please check whether the datafile definition"
          print *,"has been updated ......"
          print *,"-------------------------------------------------"
          !       error_flag = error_reading_file
          !       return
       END IF

       !  #]

       IF ( L1B_Headers%FH_is_available  .and. L1B_Headers%MPH_is_available .and. &
            L1B_Headers%SPH_is_available .and. L1B_Headers%SPH%DSDs_are_available  ) THEN
          L1B_Headers%headers_are_available = .true.
          print *,"all L1B headers have been read succesfully"
       ELSE
          print *,"ERROR in ReadL1Bfile_DBL()"
          print *,"Not all headers have been read succesfully"
          error_flag = error_reading_file
          return
       END IF

       !  #]
    END IF ! do_ReadHeaders

    !------------------------------
    ! read the binary data sets 
    !------------------------------

    !print *,"test: do_ReadGeoloc = ",do_ReadGeoloc

    ! check that L1B data part is present
    IF (.not.(present(L1B)) .and. &
      & (do_ReadGeoloc  .or.  do_ReadPCD     .or. &
      &  do_ReadGWD     .or.  do_ReadMeas    .or. &
      &  do_ReadCal     .or.  do_ReadUS      .or. &
      &  do_ReadWV)) THEN
        print *,"ERROR in ReadL1Bfile_DBL()"
        print *,"L1B data argument is missing"
        error_flag = error_programming
        return
    ENDIF

    IF (do_ReadGeoloc) THEN
       !  #[
       IF (.not. L1B_Headers%headers_are_available) THEN
          print *,"ERROR in ReadL1Bfile_DBL()==>Read_L1B_Geoloc_ADS:"
          print *,"reading a binary dataset is only possible after reading"
          print *,"the L1B headers, but these seem not yet available ..."
          error_flag = error_programming
          return
       END IF

       ! see L1A Geolocation ADSR in section 5.3.1
       N_Meas_Max = L1B_Headers%SPH%N_Meas_Max
       N_range_gates = 24

       IF (associated(L1B%Geoloc_ADS)) THEN
          ! REMARK: this structure is allocated at this point, but its sub-
          ! structures should still be nullified, and are only allocated in the
          ! init routine below. After calling Init_L1B_DataStructure this is
          ! the case, since Init_L1B_Geoloc_ADS() is called with N_Meas=0
          ! and N_range_gates=0. However, for the 2nd and next BRC's an
          ! explicit call to Empty_L1B_BinaryDatasets() is needed to
          ! restore this condition, before entering this reading code.
          !
          ! REMARK: this structure uses an extra rangegate to store the
          ! top of the upper rangebin, and the regular rangegates to store
          ! the bottom of each rangegate
          call Init_L1B_Geoloc_ADS(L1B%Geoloc_ADS, N_meas_Max, &
                                   N_range_gates+1, error_flag)
          IF (error_flag .ne. no_error) return
       ELSE
          print *,"ERROR in ReadL1Bfile_DBL()==>Read_L1B_Geoloc_ADS:"
          print *,"the L1B%Geoloc_ADS structure seems not allocated !!!"
          error_flag = error_programming
          return
       END IF

       IF ( (DataSetRecord_to_read .ge. 1) .and. &
            (DataSetRecord_to_read .le. &
                     L1B_Headers%SPH%Geolocation_DSD%nr_of_dataset_records) ) THEN
          call Read_L1B_Geoloc_ADS(L1B_Headers%SPH%Geolocation_DSD,fileunit,&
               DataSetRecord_to_read,N_Meas_Max,&
               N_range_gates+1, L1B%Geoloc_ADS,error_flag)
          IF (error_flag .ne. no_error) return
       ELSE
          print *,"ERROR in ReadL1Bfile_DBL()==>Read_L1B_Geoloc_ADS:"
          print *,"DataSetRecord_to_read = ",DataSetRecord_to_read
          print *,"This is outside the available range of datasets !!!"
          print *,"L1B%SPH%Geolocation_DSD%nr_of_dataset_records = ",&
               L1B_Headers%SPH%Geolocation_DSD%nr_of_dataset_records
          error_flag = error_reading_file
          return
       END IF
       print *,"done reading L1B%Geoloc_ADS"
       !  #]
    END IF ! do_ReadGeoloc

    !print *,"test: do_ReadPCD = ",do_ReadPCD
    IF (do_ReadPCD) THEN
       !  #[
       IF (.not. L1B_Headers%headers_are_available) THEN
          print *,"ERROR in ReadL1Bfile_DBL()==>Read_L1B_PCD_ADS:"
          print *,"reading a binary dataset is only possible after reading"
          print *,"the L1B headers, but these seem not yet available ..."
          error_flag = error_programming
          return
       END IF

       N_Meas_Max = L1B_Headers%SPH%N_Meas_Max
       N_range_gates = 24

       IF (associated(L1B%PCD_ADS)) THEN
          ! REMARK: this structure is allocated at this point, but its sub-
          ! structures should still be nullified, and are only allocated in the
          ! init routine below. After calling Init_L1B_DataStructure this is
          ! the case, since Init_L1B_PCD_ADS() is called with N_Meas=0
          ! and N_range_gates=0. However, for the 2nd and next BRC's an
          ! explicit call to Empty_L1B_BinaryDatasets() is needed to
          ! restore this condition, before entering this reading code.
          call Init_L1B_PCD_ADS(L1B%PCD_ADS, N_meas_Max, N_range_gates, &
                                error_flag)
          IF (error_flag .ne. no_error) return
       ELSE
          print *,"ERROR in ReadL1Bfile_DBL()==>Read_L1B_PCD_ADS:"
          print *,"the L1B%PCD_ADS structure seems not allocated !!!"
          error_flag = error_programming
          return
       END IF

       IF ( (DataSetRecord_to_read .ge. 1) .and. &
            (DataSetRecord_to_read .le. &
                     L1B_Headers%SPH%Product_Conf_Data_DSD%nr_of_dataset_records) ) THEN
          call Read_L1B_PCD_ADS(L1B_Headers%SPH%Product_Conf_Data_DSD,fileunit,&
                                DataSetRecord_to_read,&
                            N_Meas_Max,N_range_gates,L1B%PCD_ADS,error_flag)
          IF (error_flag .ne. no_error) return
       ELSE
          print *,"ERROR in ReadL1Bfile_DBL()==>Read_L1B_PCD_ADS:"
          print *,"DataSetRecord_to_read = ",DataSetRecord_to_read
          print *,"This is outside the available range of datasets !!!"
          print *,"L1B%SPH%Product_Conf_Data_DSD%nr_of_dataset_records = ",&
               L1B_Headers%SPH%Product_Conf_Data_DSD%nr_of_dataset_records
          error_flag = error_reading_file
          return
       END IF
       print *,"done reading L1B%PCD_ADS"
       !  #]
    END IF ! do_ReadPCD
    
    !print *,"test: do_ReadGWD = ",do_ReadGWD
    IF (do_ReadGWD) THEN
       !  #[
       IF (.not. L1B_Headers%headers_are_available) THEN
          print *,"ERROR in ReadL1Bfile_DBL()==>Read_L1B_GWD_ADS:"
          print *,"reading a binary dataset is only possible after reading"
          print *,"the L1B headers, but these seem not yet available ..."
          error_flag = error_programming
          return
       END IF
       
       N_Meas_Max = L1B_Headers%SPH%N_Meas_Max

       IF (associated(L1B%GWD_ADS)) THEN
          ! REMARK: this structure is allocated at this point, but its sub-
          ! structures should still be nullified, and are only allocated in the
          ! init routine below. After calling Init_L1B_DataStructure this is
          ! the case, since Init_L1B_GWD_ADS() is called with N_Meas=0
          ! and N_range_gates=0. However, for the 2nd and next BRC's an
          ! explicit call to Empty_L1B_BinaryDatasets() is needed to
          ! restore this condition, before entering this reading code.
          call Init_L1B_GWD_ADS(L1B%GWD_ADS, N_meas_Max, error_flag)
          IF (error_flag .ne. no_error) return
       ELSE
          print *,"ERROR in ReadL1Bfile_DBL()==>Read_L1B_GWD_ADS:"
          print *,"the L1B%GWD_ADS structure seems not allocated !!!"
          error_flag = error_programming
          return
       END IF

       IF ( (DataSetRecord_to_read .ge. 1) .and. &
            (DataSetRecord_to_read .le. &
                L1B_Headers%SPH%Ground_Wind_Detection_DSD%&
                            nr_of_dataset_records) ) THEN
          call Read_L1B_GWD_ADS(L1B_Headers%SPH%Ground_Wind_Detection_DSD,&
                                fileunit,&
                                DataSetRecord_to_read,&
                                N_Meas_Max,L1B%GWD_ADS,error_flag,&
                                workaround_pre108)
          IF (error_flag .ne. no_error) return
       ELSE
          print *,"ERROR in ReadL1Bfile_DBL()==>Read_L1B_GWD_ADS:"
          print *,"DataSetRecord_to_read = ",DataSetRecord_to_read
          print *,"This is outside the available range of datasets !!!"
          print *,"L1B%SPH%Ground_Wind_Detection_DSD%nr_of_dataset_records = ",&
               L1B_Headers%SPH%Ground_Wind_Detection_DSD%nr_of_dataset_records
          error_flag = error_reading_file
          return
       END IF
       print *,"done reading L1B%GWD_ADS"
       !  #]
    END IF ! do_ReadGWD

    !print *,"test: do_ReadMeas = ",do_ReadMeas
    IF (do_ReadMeas) THEN
       !  #[
       IF (.not. L1B_Headers%headers_are_available) THEN
          print *,"ERROR in ReadL1Bfile_DBL()==>Read_L1B_Meas_ADS:"
          print *,"reading a binary dataset is only possible after reading"
          print *,"the L1B headers, but these seem not yet available ..."
          error_flag = error_programming
          return
       END IF

       N_Meas_Max = L1B_Headers%SPH%N_Meas_Max
       N_range_gates = 24

       IF (associated(L1B%Meas_ADS)) THEN
          ! REMARK: this structure is allocated at this point, but its sub-
          ! structures should still be nullified, and are only allocated in the
          ! init routine below. After calling Init_L1B_DataStructure this is
          ! the case, since Init_L1B_Meas_ADS() is called with N_Meas=0
          ! and N_range_gates=0. However, for the 2nd and next BRC's an
          ! explicit call to Empty_L1B_BinaryDatasets() is needed to
          ! restore this condition, before entering this reading code.
          call Init_L1B_Meas_ADS(L1B%Meas_ADS, N_meas_Max, N_range_gates, &
                                error_flag)
          IF (error_flag .ne. no_error) return
       ELSE
          print *,"ERROR in ReadL1Bfile_DBL()==>Read_L1B_Meas_ADS:"
          print *,"the L1B%Meas_ADS structure seems not allocated !!!"
          error_flag = error_programming
          return
       END IF

       IF ( (DataSetRecord_to_read .ge. 1) .and. &
            (DataSetRecord_to_read .le. &
                     L1B_Headers%SPH%Measurement_DSD%nr_of_dataset_records) ) THEN
          call Read_L1B_Meas_ADS(L1B_Headers%SPH%Measurement_DSD,fileunit,&
               DataSetRecord_to_read,N_Meas_Max,&
               N_range_gates,L1B%Meas_ADS,error_flag)
          IF (error_flag .ne. no_error) return
       ELSE
          print *,"ERROR in ReadL1Bfile_DBL()==>Read_L1B_Meas_ADS:"
          print *,"DataSetRecord_to_read = ",DataSetRecord_to_read
          print *,"This is outside the available range of datasets !!!"
          print *,"L1B%SPH%Measurement_DSD%nr_of_dataset_records = ",&
               L1B_Headers%SPH%Measurement_DSD%nr_of_dataset_records
          error_flag = error_reading_file
          return
       END IF
       print *,"done reading L1B%Meas_ADS"
       !  #]
    END IF ! do_ReadMeas

    !print *,"test: do_ReadCal = ",do_ReadCal
    IF (do_ReadCal) THEN
       !  #[
       IF (.not. L1B_Headers%headers_are_available) THEN
          print *,"ERROR in ReadL1Bfile_DBL()==>Read_L1B_Cal_ADS:"
          print *,"reading a binary dataset is only possible after reading"
          print *,"the L1B headers, but these seem not yet available ..."
          error_flag = error_programming
          return
       END IF

       NF_Order_Max = L1B_Headers%SPH%NF_Order_Max

       IF (associated(L1B%Cal_ADS)) THEN
          call Init_L1B_Cal_ADS(L1B%Cal_ADS, &
               error_flag)
          IF (error_flag .ne. no_error) return
       ELSE
          print *,"ERROR in ReadL1Bfile_DBL()==>Read_L1B_Cal_ADS:"
          print *,"the L1B%Cal_ADS structure seems not allocated !!!"
          error_flag = error_programming
          return
       END IF

       !IF ( (DataSetRecord_to_read .ge. 1) .and. &
       !     (DataSetRecord_to_read .le. &
       !              L1B%SPH%Cal_DSD%nr_of_dataset_records) ) THEN
          call Read_L1B_Cal_ADS(L1B_Headers%SPH%Cal_DSD,fileunit,&
              ! DataSetRecord_to_read,&
               NF_Order_Max, &
               L1B%Cal_ADS,error_flag)
          IF (error_flag .ne. no_error) return
       !ELSE
       !   print *,"ERROR in ReadL1Bfile_DBL()==>Read_L1B_Cal_ADS:"
       !   print *,"DataSetRecord_to_read = ",DataSetRecord_to_read
       !   print *,"This is outside the available range of datasets !!!"
       !   print *,"L1B%SPH%Cal_DSD%nr_of_dataset_records = ",&
       !        L1B%SPH%Cal_DSD%nr_of_dataset_records
       !   error_flag = error_reading_file
       !   return
       !END IF
       print *,"done reading L1B%Cal_Char_GADS"
       !  #]
    END IF ! do_ReadCal

    !print *,"test: do_ReadUS = ",do_ReadUS
    IF (do_ReadUS) THEN
       !  #[
       IF (.not. L1B_Headers%headers_are_available) THEN
          print *,"ERROR in ReadL1Bfile_DBL()==>Read_L1B_US_MDS:"
          print *,"reading a binary dataset is only possible after reading"
          print *,"the L1B headers, but these seem not yet available ..."
          error_flag = error_programming
          return
       END IF

       N_Meas_Max = L1B_Headers%SPH%N_Meas_Max
       N_range_gates = 24

       IF (associated(L1B%US_MDS)) THEN
          ! REMARK: this structure is allocated at this point, but its sub-
          ! structures should still be nullified, and are only allocated in the
          ! init routine below. After calling Init_L1B_DataStructure this is
          ! the case, since Init_L1B_Meas_ADS() is called with N_Meas=0
          ! and N_range_gates=0. However, for the 2nd and next BRC's an
          ! explicit call to Empty_L1B_BinaryDatasets() is needed to
          ! restore this condition, before entering this reading code.
          call Init_L1B_US_MDS(L1B%US_MDS,N_meas_Max,N_range_gates,error_flag)
          IF (error_flag .ne. no_error) return
       ELSE
          print *,"ERROR in ReadL1Bfile_DBL()==>Read_L1B_US_MDS:"
          print *,"the L1B%US_MDS structure seems not allocated !!!"
          error_flag = error_programming
          return
       END IF

       IF ( (DataSetRecord_to_read .ge. 1) .and. &
            (DataSetRecord_to_read .le. &
                     L1B_Headers%SPH%Useful_Signal_DSD%nr_of_dataset_records) ) THEN
          call Read_L1B_US_MDS(L1B_Headers%SPH%Useful_Signal_DSD,fileunit,&
               DataSetRecord_to_read,N_Meas_Max,&
               N_range_gates,L1B%US_MDS,error_flag)
          IF (error_flag .ne. no_error) return
       ELSE
          print *,"ERROR in ReadL1Bfile_DBL()==>Read_L1B_US_MDS:"
          print *,"DataSetRecord_to_read = ",DataSetRecord_to_read
          print *,"This is outside the available range of datasets !!!"
          print *,"L1B%SPH%Useful_Signal_DSD%nr_of_dataset_records = ",&
               L1B_Headers%SPH%Useful_Signal_DSD%nr_of_dataset_records
          error_flag = error_reading_file
          return
       END IF
       print *,"done reading L1B%US_MDS"
       !  #]
    END IF ! do_ReadUS

    !print *,"test: do_ReadWV = ",do_ReadWV
    IF (do_ReadWV) THEN
       !  #[
       IF (.not. L1B_Headers%headers_are_available) THEN
          print *,"ERROR in ReadL1Bfile_DBL()==>Read_L1B_WV_MDS:"
          print *,"reading a binary dataset is only possible after reading"
          print *,"the L1B headers, but these seem not yet available ..."
          error_flag = error_programming
          return
       END IF

       N_Meas_Max = L1B_Headers%SPH%N_Meas_Max
       N_range_gates = 24

       IF (associated(L1B%WV_MDS)) THEN
          ! REMARK: this structure is allocated at this point, but its sub-
          ! structures should still be nullified, and are only allocated in the
          ! init routine below. After calling Init_L1B_DataStructure this is
          ! the case, since Init_L1B_Meas_ADS() is called with N_Meas=0
          ! and N_range_gates=0. However, for the 2nd and next BRC's an
          ! explicit call to Empty_L1B_BinaryDatasets() is needed to
          ! restore this condition, before entering this reading code.
          call Init_L1B_WV_MDS(L1B%WV_MDS,N_meas_Max,N_range_gates,error_flag)
          IF (error_flag .ne. no_error) return
       ELSE
          print *,"ERROR in ReadL1Bfile_DBL()==>Read_L1B_WV_MDS:"
          print *,"the L1B%WV_MDS structure seems not allocated !!!"
          error_flag = error_programming
          return
       END IF

       IF ( (DataSetRecord_to_read .ge. 1) .and. &
            (DataSetRecord_to_read .le. &
                     L1B_Headers%SPH%Wind_Velocity_DSD%nr_of_dataset_records) ) THEN
          call Read_L1B_WV_MDS(L1B_Headers%SPH%Wind_Velocity_DSD,fileunit,&
               DataSetRecord_to_read,N_Meas_Max,&
               N_range_gates,L1B%WV_MDS,error_flag)
          IF (error_flag .ne. no_error) return
       ELSE
          print *,"ERROR in ReadL1Bfile_DBL()==>Read_L1B_WV_MDS:"
          print *,"DataSetRecord_to_read = ",DataSetRecord_to_read
          print *,"This is outside the available range of datasets !!!"
          print *,"L1B%SPH%Wind_Velocity_DSD%nr_of_dataset_records = ",&
               L1B_Headers%SPH%Wind_Velocity_DSD%nr_of_dataset_records
          error_flag = error_reading_file
          return
       END IF
       !  #]
    END IF ! do_ReadWV

    close(unit=fileunit)
    call free_lun(fileunit)

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"done with reading L1B datafile"

    !------------------------------
    ! print all datasets as test
    !------------------------------

!print *,"----------------------------------------------------------"
!print *,"--------------- extra test prints ------------------------"
!print *,"----------------------------------------------------------"
!print *,"L1B%SPH%Product_Conf_Data_DSD%..."
!call Print_DSD(L1B%SPH%Wind_Velocity_DSD)
!print *,"----------------------------------------------------------"
!print *,"L1B%WV_MDS%... "
!call Print_L1B_WV_MDS(L1B%WV_MDS)
!print *,"----------------------------------------------------------"

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"----------------------------------------------------------"
       print *,"L1B%MPH%..."
       call Print_MPH(L1B_Headers%MPH,error_flag)
       IF (error_flag .ne. no_error) return
       print *,"----------------------------------------------------------"
       print *,"L1B_Headers%SPH%..."
       call Print_L1B_SPH(L1B_Headers%SPH,error_flag)
       IF (error_flag .ne. no_error) return
       print *,"----------------------------------------------------------"
    END IF

    return

999 print *,"ERROR could not open inputfile: ",trim(filename_dbl)
    error_flag = error_opening_file
    return

  end subroutine ReadL1Bfile_DBL

    !  #]
  !---------------------------------
end module ReadL1Bdata
