module ReadL2BCdata
  !  #[ Description:
  !
  !     a module to read the 2 files describing the L2BC product
  !     of the Aeolus instrument
  !
  !     Remark: although not needed for the actual L2B-processor,
  !             this reading module is implemented anyway because:
  !             -we need to check if we write L2BC files correctly
  !             -we will need it for constructing a L2B=>BUFR conversion tool 
  !             -
  !             -
  !             
  ! created:     02-Aug-2005 by J. de Kloe, KNMI
  ! 
  ! Modifications:
  !   02-Mar-2006 J. de Kloe  adapted to changes in the XML_module
  !   29-Mar-2006 J. de Kloe  completed reading PCD ADS
  !   30-Mar-2006 J. de Kloe  added reading Mie MDS and Rayleigh MDS
  !   12-May-2006 D. Tan      Null use of ParKind1 module
  !   31-Jul-2006 J. de Kloe  renamed set_file_pointer to SetFilePointer
  !   17-Aug-2006 P. Poli     Removed obsolete variables
  !   15-Nov-2006 J. de Kloe  reintroduced the ReadL2Bfile_DBL routine
  !                           originally written for the L2BtoTXT
  !                           branch of 15-Mar-2006 
  !                           (SPH and Geoloc only for now)
  !   30-Nov-2006 D. Tan      Fix typos.
  !   23-Feb-2007 J. de Kloe  reintroduced reading of Mie_MDS and Rayleigh_MDS
  !                           from the L2BtoTXT branch of 15-Mar-2006 
  !   15-Mar-2007 J. de Kloe  added l2b_pcd and all l2c datasets
  !   07-Jun-2007 J. de Kloe  adapted to changed xml interface
  !   03-Sep-2007 P. Poli     modifed ReadL2BCfile to allow using L2BC header
  !                           information from another datastructure
  !   04-Sep-2007 P. Poli     change L2BC_DataType to L2BC_BRC_DataType 
  !                           and L2BC_HDR_DataType
  !   10-Oct-2007 J. de Kloe  added a missing nullify()
  !   12-Nov-2007 J. de Kloe  added check for xmlns attribute in xml file
  !   16-Jan-2008 J. de Kloe  phase out integer kind i_
  !   21-Jan-2008 J. de Kloe  added (temporary) workaround switch in the
  !                           call to Read_MPH_from_KVT
  !   21-Feb-2008 J. de Kloe  removed the workaround again since it was
  !                           caused by a misunderstanding on L1B side
  !
  !  #]
  !  #[ Modules used:

  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE FixedHeader, only: FH_type, Init_FH, Print_FH, get_fh_from_xml_tree, &
                         file_type_L2B_N, file_type_L2C_N
  USE MainProductHeader, only: Print_MPH, &
                               Get_MPH_from_xml_tree, Read_MPH_from_KVT
  USE L2BC_DataStructure, only: L2BC_HDR_DataType, L2BC_BRC_DataType
  USE L2BC_SPH, only: Init_L2BC_SPH, Get_L2BC_SPH_from_xml_tree, &
                      Print_L2BC_SPH, Read_L2BC_SPH_from_KVT, &
                      Get_M_Meas, Get_M_Mie, Get_M_Rayleigh
  USE L2BC_DS_Settings, only: set_L2BC_string
  USE L2B_Geoloc_ADS, only: Read_L2B_Geoloc_ADS, GetByteSize_L2B_Geoloc_ADS, &
                            Init_L2B_Geoloc_ADS
  USE L2B_PCD_ADS, only: Read_L2B_PCD_ADS, GetByteSize_L2B_PCD_ADS, &
                         Init_L2B_PCD_ADS
  USE L2B_Mie_MDS, only: Read_L2B_Mie_MDS, Get_ByteSize_L2B_Mie_MDS, &
                         Init_L2B_Mie_MDS
  USE L2B_Rayleigh_MDS, only: Read_L2B_Rayleigh_MDS, &
                              Get_ByteSize_L2B_Rayleigh_MDS, &
                              Init_L2B_Rayleigh_MDS
  USE L2C_AssimPCD_ADS, only: Init_L2C_AssimPCD_ADS, &
                              GetByteSize_L2C_AssimPCD_ADS, &
                              Read_L2C_AssimPCD_ADS
  USE L2C_MieVec_MDS, only: Init_L2C_MieVec_MDS, &
                            Get_ByteSize_L2C_MieVec_MDS, &
                            Read_L2C_MieVec_MDS
  USE L2C_RayleighVec_MDS, only: Init_L2C_RayleighVec_MDS, &
                                 Get_ByteSize_L2C_RaylVec_MDS, &
                                 Read_L2C_RayleighVec_MDS

  USE XML_module, only: xml_node_type, Init_XML, xml_load_file, &
                        xml_tree_print, delete_xml_tree, ee_h_name, &
                        uri_ns_header, xml_tree_get_namespace

  USE KVT_module, only: max_str_len, Init_KVT_module !, &
  USE DirectBinaryIO, only: Init_DirectBinaryIO_module, BigEndian, &
                            OpenDirectBinaryFile, CloseDirectBinaryFile,&
                            SetFilePointer
  USE EE_CFI_Datatypes, only: IntAuc, IntAus, IntAs, IntAd, &
                              GetIntAuc, GetIntAus
  USE Logging, only: logmsg, & ! for debugging only: init_logging, 
                     log_debug,log_info,log_progress,log_warn,log_error
  USE ErrorHandler, only: no_error, error_xml, error_programming, &
                          error_opening_file, error_reading_file,&
                          error_allocate
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE LunManager, only: get_lun, free_lun
  USE AeolusConstants, only: iodd_version_L2B_N, iodd_version_L2C_N
!  USE c_support, only: get_filesize

  !  #]  
  !  #[ Parameter/Variable/Type declarations

  implicit none

  ! make all module contents private, unless explicitely stated otherwise
  private

  ! the default verbosity setting is Error (as silent as possible)
  integer, save :: LocalVerbosity = VerbosityError

  ! remember the initialisation state of the module
  logical, save :: module_is_initialised = .false.

  ! TODO: move this to support/aeolus_constants.F90
  integer(IntAuc), parameter :: N_range_gates = 24
  ! this is a constant for this project, but it might differ 
  ! for future missions ...

  !  #]
  !  #[ Module interface
  public :: ReadL2BCfile, InitReadL2BCfileModule
  !  #]
contains
  !---------------------------------
  subroutine InitReadL2BCfileModule(error_flag, ChoosenVerbosity)
    !  #[

    integer,           intent(out) :: error_flag
    integer, optional, intent(in)  :: ChoosenVerbosity

    error_flag = no_error

    IF (present(ChoosenVerbosity)) LocalVerbosity = ChoosenVerbosity
    ! to be replaced by the global: Stdout_Log_Level/Stderr_Log_Level
    ! switches defined in the JobOrder file
    ! locallay this can be overridden by using:
    !call Init_Logging(Stdout_Log_Level_txt="DEBUG",&
    !                  Stderr_Log_Level_txt="NOOP",&
    !                  error_flag=error_flag)
 
    IF (LocalVerbosity .ge. VerbosityReport) &
         call print_verbosity_setting(LocalVerbosity,ModuleName="ReadL2BCdata")

    call Init_KVT_module()

    call Init_DirectBinaryIO_module(error_flag,FileByteOrder=BigEndian)
    if (error_flag .ne. no_error) return

    call Init_XML(error_flag) !(ChoosenVerbosity=VerbosityReport)
    if (error_flag .ne. no_error) return

    module_is_initialised = .true.

  end subroutine InitReadL2BCfileModule

    !  #]
  subroutine ReadL2BCfile(L2BC_Headers, filename, error_flag, L2BC, Which_DSR,&
       L2C_flag,&
       ReadAll, ReadHeaders, ReadGeoloc, ReadPCD, ReadMie, ReadRayleigh,&
       ReadAssimPCD, ReadMieVecWind, ReadRaylVecWind)
    !  #[

    type(L2BC_HDR_DataType),intent(inout):: L2BC_Headers
    character(len=*),      intent(in)    :: filename ! without extension
    integer,               intent(out)   :: error_flag
    type(L2BC_BRC_DataType),optional,    intent(inout) :: L2BC
    integer,     optional, intent(in)    :: Which_DSR
    logical,     optional, intent(in)    :: L2C_flag
    logical,     optional, intent(in)    :: ReadAll ! headers and datasets
    logical,     optional, intent(in)    :: ReadHeaders
    logical,     optional, intent(in)    :: ReadGeoloc
    logical,     optional, intent(in)    :: ReadPCD
    logical,     optional, intent(in)    :: ReadMie
    logical,     optional, intent(in)    :: ReadRayleigh
    logical,     optional, intent(in)    :: ReadAssimPCD    ! L2C only
    logical,     optional, intent(in)    :: ReadMieVecWind  ! L2C only
    logical,     optional, intent(in)    :: ReadRaylVecWind ! L2C only

    ! local variables
    character(len=max_str_len) :: filename_hdr
    character(len=max_str_len) :: filename_dbl
    logical                    :: do_ReadHeaders
    logical                    :: L2C_flag_local
    character(len=3)           :: L2BC_string

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       print *,"Error in ReadL2BCfile():"
       print *,"the ReadL2BCdata module must be initialised before use ..."
       error_flag = error_programming
       return
    END IF

! for debugging only
!    call Init_Logging(Stdout_Log_Level_txt="DEBUG",&
!                      Stderr_Log_Level_txt="NOOP",&
!                      error_flag=error_flag)

    filename_hdr = trim(filename)//".HDR"
    filename_dbl = trim(filename)//".DBL"

    L2C_flag_local=.false.
    IF (present(L2C_flag)) L2C_flag_local=L2C_flag
    L2BC_string=set_L2BC_string(L2C_flag_local)
    
    do_ReadHeaders = .false.
    IF (present(ReadHeaders)) do_ReadHeaders = ReadHeaders
    IF (present(ReadAll)) THEN
       IF (ReadAll) do_ReadHeaders = .true.
    END IF
    
    IF (do_ReadHeaders) THEN
       call logmsg(log_debug,"Starting ReadL2BCfile_HDR")

       ! needs to be called just once per set of HDR/DBL files
       call ReadL2BCfile_HDR(L2BC_Headers, filename_hdr, L2C_flag_local, error_flag)
       IF (error_flag .ne. no_error) return

       ! sanity check on the filename
       IF (trim(L2BC_Headers%FH%File_Name) .ne. trim(filename)) THEN
          call logmsg(log_warn,"Warning in ReadL2BCfile: "//&
               trim(L2BC_string)//" filename "//&
               "differs from the name specified in the Fixed Header."//&
               "Possibly this file was renamed?")
       END IF
    END IF

    call logmsg(log_debug,"Starting ReadL2BCfile_DBL")

    ! needs to be called for every BRC present in a set of HDR/DBL files
    call ReadL2BCfile_DBL(L2BC_Headers, filename_dbl, L2C_flag_local, error_flag, &
         L2BC, Which_DSR, &
         ReadAll, ReadHeaders, ReadGeoloc, ReadPCD, ReadMie, ReadRayleigh,&
         ReadAssimPCD, ReadMieVecWind, ReadRaylVecWind)
    IF (error_flag .ne. no_error) return

    return

  end subroutine ReadL2BCfile

    !  #]
  subroutine ReadL2BCfile_HDR(L2BC_Headers, filename_hdr, L2C_flag, error_flag)
    !  #[

    type(L2BC_HDR_DataType), intent(inout) :: L2BC_Headers
    character(len=*),    intent(in)    :: filename_hdr
    logical,             intent(in)    :: L2C_flag
    integer,             intent(out)   :: error_flag

    ! local variables
    character(len=256)           :: path
    TYPE(xml_node_type), pointer :: xml_tree
    character(len=3)             :: L2BC_string
    character(len=256)           :: xml_namespace_L2BC
    character(len=256)           :: xml_namespace_uri

    ! init
    error_flag = no_error
    L2BC_string=set_L2BC_string(L2C_flag)
    nullify(xml_tree)

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in ReadL2BCfile:"
       print *,"the ReadL2BCdata module should be initialised before use !"
       error_flag = error_programming
       return
    END IF

    call logmsg(log_debug,"ReadL2BCfile_HDR: Reading "//trim(L2BC_string)//&
                          " header file...")

    call xml_load_file(filename_hdr, xml_tree, error_flag)
    IF (error_flag .ne. no_error) return

    ! define which xml namespace should be present
    !example: xmlns="http://www.esa.int/schemas/ae/ALD_U_N_2B_01.01"
    !uri_ns_header    ! defined in xml_module
    !file_type_L2B_N  ! defined in FixedHeader
    !iodd_version_L2B ! defined in AeolusConstants
    IF (L2C_flag) THEN
       xml_namespace_L2BC = trim(uri_ns_header)//trim(file_type_L2C_N)//&
                            "_"//trim(iodd_version_L2C_N)
    ELSE
       xml_namespace_L2BC = trim(uri_ns_header)//trim(file_type_L2B_N)//&
                            "_"//trim(iodd_version_L2B_N)
    END IF

    ! check whether this namespace is really present in the input file
    call xml_tree_get_namespace(xml_tree,xml_namespace_uri,error_flag)
    if (error_flag .ne. no_error) then
       ! allow this attribute to be missing for now, to allow use
       ! of outdated files.
       ! TODO: remove this exception
       error_flag = no_error
       call logmsg(log_warn,"WARNING in ReadL2BCfile_HDR: "//&
                   "could not find xmlns attribute in the root tag "//&
                   "of file: "//trim(filename_hdr))
    else
       if (trim(xml_namespace_L2BC) .ne. trim(xml_namespace_uri)) &
            call logmsg(log_warn,"WARNING in ReadL2BCfile_HDR: "//&
                        "expected an xmlns attribute with content: "//&
                        trim(xml_namespace_L2BC)//", but found an "//&
                        "xmlns attribute with content: "//&
                        trim(xml_namespace_uri)//". This input file may "//&
                        "not be compatible with the current software!")
    end if

    call logmsg(log_debug,"ReadL2BCfile_HDR: the xml file was "//&
                          "successfully read")

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"==================="
       call xml_tree_print(xml_tree, error_flag)
       IF (error_flag .ne. no_error) return
       print *,"==================="
    END IF
 
    !------------------------------
    ! copying the FH XML elements
    !------------------------------
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"copying items to the FH datastructure"
    IF (.not. associated(L2BC_Headers%FH)) THEN
       print *,"Error in ReadL2BCfile_DBL:"
       print *,"L2BC_Headers%FH seems not associated !"
       print *,"this should have been done in Init_L2BC_HDR_DataStructure"
       error_flag = error_programming
       return
    ENDIF
    path = "/"//ee_h_name
    call Get_FH_from_xml_tree(xml_tree, path, L2BC_Headers%FH, error_flag)
    IF (error_flag .ne. no_error) return

    !------------------------------
    ! copying the MPH XML elements
    !------------------------------
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"copying items to the MPH datastructure"
    IF (.not. associated(L2BC_Headers%MPH)) THEN
       print *,"Error in ReadL2BCfile_DBL:"
       print *,"L2BC_Headers%MPH seems not associated !"
       print *,"this should have been done in Init_L2BC_HDR_DataStructure"
       error_flag = error_programming
       return
    ENDIF
    call Get_MPH_from_xml_tree(xml_tree, path, L2BC_Headers%MPH, error_flag)
    IF (error_flag .ne. no_error) return

    !------------------------------
    ! copying the SPH XML elements (with DSD's)
    !------------------------------
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"copying items to the SPH datastructure"
    IF (.not. associated(L2BC_Headers%SPH)) THEN
       print *,"Error in ReadL2BCfile_DBL:"
       print *,"L2BC_Headers%SPH seems not associated !"
       print *,"this should have been done in Init_L2BC_HDR_DataStructure"
       error_flag = error_programming
       return
    ENDIF
    call Get_L2BC_SPH_from_xml_tree(xml_tree, path, L2BC_Headers%SPH, &
                                    error_flag, L2C_flag)
    IF (error_flag .ne. no_error) return

    !------------------------------
    ! done                         
    !------------------------------

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,trim(L2BC_string)//"%FH%..."
       call Print_FH(L2BC_Headers%FH,&
                     prefix="...",&
                     error_flag=error_flag)
       IF (error_flag .ne. no_error) return
! TODO: replace by:
!       call Print_FH(L2BC%FH,&
!                     prefix=trim(L2BC_string)//"%FH",&
!                     error_flag=error_flag)
!       IF (error_flag .ne. no_error) return

       print *,trim(L2BC_string)//"%MPH%..."
       call Print_MPH(L2BC_Headers%MPH,&
                      prefix="...",&
                      error_flag=error_flag)
       IF (error_flag .ne. no_error) return
! TODO: replace by:
!       call Print_MPH(L2BC%MPH,&
!                      prefix=trim(L2BC_string)//"%MPH",&
!                      error_flag=error_flag)
!       IF (error_flag .ne. no_error) return

! TODO: activate prefix/error-flag parameters in l2bc-sph.F90
!       and start using them here.
       print *,trim(L2BC_string)//"%SPH%..."
       call Print_L2BC_SPH(L2BC_Headers%SPH,L2C_flag,error_flag)
                           !prefix=trim(L2BC_string)//"%SPH",&
                           !error_flag=error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"deallocating xml items"
    call delete_xml_tree(xml_tree, error_flag)
    IF (error_flag .ne. no_error) return

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"done with reading "//trim(L2BC_string)//" header file"

    return

  end subroutine ReadL2BCfile_HDR
    !  #]
  subroutine ReadL2BCfile_DBL(L2BC_Headers, filename_dbl, L2C_flag, &
       error_flag, L2BC, Which_DSR, &
       ReadAll, ReadHeaders, ReadGeoloc, ReadPCD, ReadMie, ReadRayleigh,&
       ReadAssimPCD, ReadMieVecWind, ReadRaylVecWind)
    !  #[
    TYPE(L2BC_HDR_DataType),intent(inout):: L2BC_Headers
    character(len=*),      intent(in)    :: filename_dbl
    logical,               intent(in)    :: L2C_flag
    integer,               intent(out)   :: error_flag
    TYPE(L2BC_BRC_DataType), optional, intent(inout) :: L2BC
    integer,     optional, intent(in)    :: Which_DSR
    logical,     optional, intent(in)    :: ReadAll     ! headers and datasets
    logical,     optional, intent(in)    :: ReadHeaders
    logical,     optional, intent(in)    :: ReadGeoloc
    logical,     optional, intent(in)    :: ReadPCD
    logical,     optional, intent(in)    :: ReadMie
    logical,     optional, intent(in)    :: ReadRayleigh
    logical,     optional, intent(in)    :: ReadAssimPCD    ! L2C only
    logical,     optional, intent(in)    :: ReadMieVecWind  ! L2C only
    logical,     optional, intent(in)    :: ReadRaylVecWind ! L2C only

    ! local variables
    integer            :: fileunit, AllocStatus
    integer            :: DataSetRecord_to_read
    integer            :: nbytes_read, nbytes_expected, nbytes_given_in_dsd
    integer(IntAuc)    :: N_Obs_Mie
    integer(IntAuc)    :: N_Obs_Rayleigh
    integer(IntAuc)    :: N_meas
    integer            :: pos_start, pos_end
    integer(IntAd)     :: pos
    character(len=3)   :: L2BC_string

    ! switches to copy the optional input switches to
    logical            :: do_ReadHeaders
    logical            :: do_ReadGeoloc
    logical            :: do_ReadPCD
    logical            :: do_ReadMie
    logical            :: do_ReadRayleigh

    logical            :: do_ReadAssimPCD    ! L2C only
    logical            :: do_ReadMieVecWind  ! L2C only
    logical            :: do_ReadRaylVecWind ! L2C only
    
    ! the current binary dataset definition that has been implemented
    ! is the version described in L2BC_iodd version 1.3, dd. 20070223

    ! init
    error_flag = no_error
    L2BC_string=set_L2BC_string(L2C_flag)

    IF (.not. module_is_initialised) THEN
       print *,"ERROR in ReadL2BCfile_DBL():"
       print *,"the ReadL2BCdata module should be initialised before use !"
       error_flag = error_programming
       return
    END IF

    DataSetRecord_to_read = 1 ! by default just read the 1st dataset
    IF (present(Which_DSR)) THEN
       DataSetRecord_to_read = Which_DSR
    END IF

    do_ReadHeaders  = .false.
    do_ReadGeoloc   = .false.
    do_ReadPCD      = .false.
    do_ReadMie      = .false.
    do_ReadRayleigh = .false.
    do_ReadAssimPCD    = .false. ! L2C only
    do_ReadMieVecWind  = .false. ! L2C only
    do_ReadRaylVecWind = .false. ! L2C only

    IF (present(ReadAll)) THEN
       do_ReadHeaders  = .true.
       do_ReadGeoloc   = .true.
       do_ReadPCD      = .true.
       do_ReadMie      = .true.
       do_ReadRayleigh = .true.
       IF (L2C_flag) THEN
          do_ReadAssimPCD    = .true. ! L2C only
          do_ReadMieVecWind  = .true. ! L2C only
          do_ReadRaylVecWind = .true. ! L2C only
       END IF
    END IF

    IF (present(ReadHeaders )) do_ReadHeaders  = ReadHeaders
    IF (present(ReadGeoloc  )) do_ReadGeoloc   = ReadGeoloc
    IF (present(ReadPCD     )) do_ReadPCD      = ReadPCD
    IF (present(ReadMie     )) do_ReadMie      = ReadMie
    IF (present(ReadRayleigh)) do_ReadRayleigh = ReadRayleigh

    IF (present(ReadAssimPCD)) THEN
       IF (L2C_flag) THEN
          do_ReadAssimPCD    = ReadAssimPCD
       ELSE
          IF (ReadAssimPCD) THEN
             call logmsg(log_error,"ERROR in ReadL2BCfile_DBL: "//&
                  "ReadAssimPCD flag is present but L2C_flag is false.  "//&
                  "This is not allowed since this is a L2C dataset !!!")
             error_flag = error_programming
             goto 999
          END IF
       END IF
    END IF

    IF (present(ReadMieVecWind)) THEN
       IF (L2C_flag) THEN
          do_ReadMieVecWind  = ReadMieVecWind
       ELSE
          IF (ReadMieVecWind) THEN
             call logmsg(log_error,"ERROR in ReadL2BCfile_DBL: "//&
                  "ReadMieVecWind flag is present but L2C_flag is false. "//&
                  "This is not allowed since this is a L2C dataset !!!")
             error_flag = error_programming
             goto 999
          END IF
       END IF
    END IF

    IF (present(ReadRaylVecWind)) THEN
       IF (L2C_flag) THEN
          do_ReadRaylVecWind = ReadRaylVecWind
       ELSE
          IF (ReadRaylVecWind) THEN
             call logmsg(log_error,"ERROR in ReadL2BCfile_DBL: "//&
                  "ReadRaylVecWind flag is present but L2C_flag is false. "//&
                  "This is not allowed since this is a L2C dataset !!!")
             error_flag = error_programming
             goto 999
          END IF
       END IF
    END IF

    ! if nothing was selected to be read, just read everything
    IF ( .not. do_ReadHeaders     .and. &
         .not. do_ReadGeoloc      .and. &
         .not. do_ReadPCD         .and. &
         .not. do_ReadMie         .and. &
         .not. do_ReadRayleigh    .and. &
         .not. do_ReadAssimPCD    .and. &
         .not. do_ReadMieVecWind  .and. &
         .not. do_ReadRaylVecWind        ) THEN
       do_ReadHeaders  = .true.
       do_ReadGeoloc   = .true.
       do_ReadPCD      = .true.
       do_ReadMie      = .true.
       do_ReadRayleigh = .true.
       IF (L2C_flag) THEN
          do_ReadAssimPCD    = .true. ! L2C only
          do_ReadMieVecWind  = .true. ! L2C only
          do_ReadRaylVecWind = .true. ! L2C only
       END IF
    END IF

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"Reading "//L2BC_string//" binary file:"
       print *,trim(filename_dbl)
    END IF

    call OpenDirectBinaryFile(filename_dbl,"read",fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 999

    IF (do_ReadHeaders) THEN
       !  #[

       !------------------------------
       ! read and/or compare MPH elements
       !------------------------------
       IF (.not. associated(L2BC_Headers%MPH)) THEN
          ! Poli: one should not go through this piece of code
          print *,"Error in ReadL2BCfile_DBL:"
          print *,"L2BC_Headers%MPH seems not associated !"
          print *,"this should have been done in Init_L2BC_HDR_DataStructure"
          error_flag = error_programming
          return

!          ! allow this, to enable the possibility of reading the dbl file only,
!          ! without reading the HDR files.
!          allocate(L2BC_Headers%MPH,STAT=AllocStatus)
!          IF (AllocStatus .ne. 0) THEN
!             print *,"Allocation failure in subroutine ReadL2BCfile_DBL"
!             error_flag = error_allocate
!             goto 999
!          END IF
!          call Init_MPH(L2BC_Headers%MPH,error_flag)
!          IF (error_flag .ne. no_error) goto 999
!          
!          call logmsg(log_debug,"ReadL2BCfile_DBL: calling Read_MPH_from_KVT")
!          call Read_MPH_from_KVT(L2BC_Headers%MPH, fileunit, error_flag)
!          IF (error_flag .ne. no_error) goto 999
       ELSE
          call logmsg(log_debug,"ReadL2BCfile_DBL: calling Read_MPH_from_KVT")
          ! read the KVT values anyway, and compare them to what was
          ! read from the XML file, then replace the values from XML
           call Read_MPH_from_KVT(L2BC_Headers%MPH,fileunit,error_flag,&
                                 Compare=.true.)
          IF (error_flag .ne. no_error) goto 999
       END IF

       !------------------------------
       ! read and/or compare SPH elements
       !------------------------------
       IF (.not. associated(L2BC_Headers%SPH)) THEN
          ! Poli: one should not go through this piece of code
          print *,"Error in ReadL2BCfile_DBL:"
          print *,"L2BC_Headers%SPH seems not associated !"
          print *,"this should have been done in Init_L2BC_HDR_DataStructure"
          error_flag = error_programming
          return

!          ! allow this, to enable the possibility of reading the dbl file only,
!          ! without reading the HDR files.
!          allocate(L2BC_Headers%SPH,STAT=AllocStatus)
!          IF (AllocStatus .ne. 0) THEN
!             print *,"Allocation failure in subroutine ReadL2BCfile_DBL"
!             error_flag = error_allocate
!             goto 999
!          END IF
!          call Init_L2BC_SPH(L2BC_Headers%SPH,error_flag)
!          IF (error_flag .ne. no_error) goto 999
!          
!          call logmsg(log_debug,&
!               "ReadL2BCfile_DBL: calling Read_L2BC_SPH_from_KVT")
!          call Read_L2BC_SPH_from_KVT(L2BC_Headers%SPH, fileunit, error_flag,&
!                                      L2C_flag=L2C_flag)
!          IF (error_flag .ne. no_error) goto 999
       ELSE
          call logmsg(log_debug,&
               "ReadL2BCfile_DBL: calling Read_L2BC_SPH_from_KVT")
          call Read_L2BC_SPH_from_KVT(L2BC_Headers%SPH,fileunit,error_flag,&
                                      Compare=.true.,L2C_flag=L2C_flag)
          IF (error_flag .ne. no_error) goto 999
       END IF

       !  #]
    END IF ! do_ReadHeaders

    IF ( do_ReadGeoloc .or. do_ReadPCD .or. do_ReadMie .or. &
         do_ReadRayleigh .or. do_ReadAssimPCD .or. do_ReadMieVecWind .or. &
         do_ReadRaylVecWind ) THEN
       !  #[
       IF (.not. associated(L2BC_Headers%SPH)) THEN
          print *,"Error in ReadL2BCfile_DBL:"
          print *,"L2BC%SPH seems not associated !"
          print *,"elements from it are needed for before a binary "//&
               "dataset can be read!!"
          error_flag = error_programming
          return
       END IF

       call Get_M_Meas(    L2BC_Headers%SPH, N_meas,         error_flag)
       IF (error_flag .ne. no_error) return
       call Get_M_Mie(     L2BC_Headers%SPH, N_Obs_Mie,      error_flag)
       IF (error_flag .ne. no_error) return 
       call Get_M_Rayleigh(L2BC_Headers%SPH, N_Obs_Rayleigh, error_flag)
       IF (error_flag .ne. no_error) return 
       
!print *,"testjos: N_Meas         = ",N_Meas
!print *,"testjos: N_Obs_Mie      = ",N_Obs_Mie
!print *,"testjos: N_Obs_Rayleigh = ",N_Obs_Rayleigh

       ! check that L2BC data part is present
       IF (.not.(present(L2BC))) THEN
           print *,"ERROR in ReadL2BCfile_DBL()"
           print *,"L2BC data argument is missing"
           error_flag = error_programming
           return
       ENDIF

      !  #]
    END IF

    IF (do_ReadGeoloc) THEN
       !  #[
       IF (.not. associated(L2BC_Headers%SPH%Geolocation_DSD)) THEN
          print *,"Error in ReadL2BCfile_DBL:"
          print *,"L2BC%SPH%Geolocation_DSD seems not associated !"
          print *,"elements from it are needed for reading the Geoloc_ADS !!"
          error_flag = error_programming
          return
       END  IF

       IF (associated(L2BC%Geoloc_ADS)) THEN
          print *,"Error in ReadL2BCfile_DBL:"
          print *,"the L2BC%Geoloc_ADS seems already allocated."
          print *,"It should not be allocated before trying to read it"
          print *,"because this read routine tries to allocate it..."
          error_flag = error_programming
          return
       END IF

       allocate(L2BC%Geoloc_ADS,stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine ReadL2BCfile_DBL..."
          error_flag = error_allocate
          return
       ENDIF
       
       call Init_L2B_Geoloc_ADS(L2BC%Geoloc_ADS, &
            int(N_Obs_Mie,IntAs), int(N_Obs_Rayleigh,IntAs), &
            int(N_Range_Gates,IntAs), error_flag)
       IF (error_flag .ne. no_error) return

       IF ( (DataSetRecord_to_read .lt. 1) .or. &
            (DataSetRecord_to_read .gt. &
                 L2BC_Headers%SPH%Geolocation_DSD%nr_of_dataset_records)) THEN
          print *,"Error in ReadL2BCfile_DBL:"
          print *,"invalid DSR number: "
          print *,"DataSetRecord_to_read = ",DataSetRecord_to_read
          print *,"L2BC%SPH%Geolocation_DSD%nr_of_dataset_records = ",&
               L2BC_Headers%SPH%Geolocation_DSD%nr_of_dataset_records
          error_flag = error_reading_file
          return
       END IF
       
       ! set the file position to the start of the requested datablock
       ! taking into account which BRC to handle
       pos = L2BC_Headers%SPH%Geolocation_DSD%start_of_ds + &
            (DataSetRecord_to_read-1)*L2BC_Headers%SPH%Geolocation_DSD%size_of_one_dsr
       call SetFilePointer(fileunit,pos,error_flag)
       IF (error_flag .ne. no_error) return

       print *,"trying to read the binary Geoloc_ADS nr. ",&
            DataSetRecord_to_read!," from file position ",pos

       inquire(unit=fileunit,nextrec=pos_start)
!print *,"testjos: pos_start = ",pos_start

       call logmsg(log_debug,"ReadL2BCfile_DBL: calling Read_L2B_Geoloc_ADS")
       call Read_L2B_Geoloc_ADS(fileunit,L2BC%Geoloc_ADS,error_flag)
       IF (error_flag .ne. no_error) goto 999

       inquire(unit=fileunit,nextrec=pos_end)

       nbytes_read         = pos_end-pos_start
       nbytes_expected     = GetByteSize_L2B_Geoloc_ADS(L2BC%Geoloc_ADS,&
                                                 N_Obs_Mie, N_Obs_Rayleigh)
       nbytes_given_in_dsd = L2BC_Headers%SPH%Geolocation_DSD%size_of_one_dsr

       IF (nbytes_read .ne. nbytes_expected) THEN
          print *,"ERROR in ReadL2Bdata: reading L2B_Geoloc_ADS"
          print *,"pos_start           = ",pos_start
          print *,"pos_end             = ",pos_end
          print *,"nbytes_read         = ",nbytes_read
          print *,"nbytes_expected     = ",nbytes_expected
          print *,"something went wrong when reading the L2B_Geoloc_ADS from"
          print *,"the binary file, probably a programming error ..."
          error_flag = error_programming
          goto 999
       END IF
       IF (nbytes_read .ne. nbytes_given_in_dsd) THEN
          print *,"ERROR in ReadL2Bdata: reading L2B_Geoloc_ADS"
          print *,"nbytes_expected     = ",nbytes_expected
          print *,"nbytes_given_in_dsd = ",nbytes_given_in_dsd
          print *,"something went wrong when reading the L2B_Geoloc_ADS from"
          print *,"the binary file, the dataset size given in the DSD"
          print *,"differs from the allocated dataset size."
          print *,"Maybe the L2B file definition changed ?"
          error_flag = error_reading_file
          goto 999
       END IF
       !  #]
    END IF ! do_ReadGeoloc

    IF (do_ReadPCD) THEN
       !  #[
       IF (.not. associated(L2BC_Headers%SPH%PCD_DSD)) THEN
          print *,"Error in ReadL2BCfile_DBL:"
          print *,"L2BC%SPH%PCD_DSD seems not associated !"
          print *,"elements from it are needed for reading the PCD_ADS !!"
          error_flag = error_programming
          return
       END  IF

       IF (associated(L2BC%PCD_ADS)) THEN
          print *,"Error in ReadL2BCfile_DBL:"
          print *,"the L2BC%PCD_ADS seems already allocated."
          print *,"It should not be allocated before trying to read it"
          print *,"because this read routine tries to allocate it..."
          error_flag = error_programming
          return
       END IF

       allocate(L2BC%PCD_ADS,stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine ReadL2BCfile_DBL..."
          error_flag = error_allocate
          return
       ENDIF

       call Init_L2B_PCD_ADS(L2BC%PCD_ADS, int(N_meas,IntAs), &
            int(N_Obs_Mie,IntAs), int(N_Obs_Rayleigh,IntAs), &
            int(N_Range_Gates,IntAs), error_flag)
       IF (error_flag .ne. no_error) return

       IF ( (DataSetRecord_to_read .lt. 1) .or. &
            (DataSetRecord_to_read .gt. &
                 L2BC_Headers%SPH%PCD_DSD%nr_of_dataset_records)) THEN
          print *,"Error in ReadL2BCfile_DBL:"
          print *,"invalid DSR number: "
          print *,"DataSetRecord_to_read = ",DataSetRecord_to_read
          print *,"L2BC%SPH%PCD_DSD%nr_of_dataset_records = ",&
               L2BC_Headers%SPH%PCD_DSD%nr_of_dataset_records
          error_flag = error_reading_file
          return
       END IF
       
       ! set the file position to the start of the requested datablock
       ! taking into account which BRC to handle
       pos = L2BC_Headers%SPH%PCD_DSD%start_of_ds + &
            (DataSetRecord_to_read-1)*L2BC_Headers%SPH%PCD_DSD%size_of_one_dsr
       call SetFilePointer(fileunit,pos,error_flag)
       IF (error_flag .ne. no_error) return

       print *,"trying to read the binary PCD_ADS nr. ",DataSetRecord_to_read

       inquire(unit=fileunit,nextrec=pos_start)
       print *,"pos_start=",pos_start

       call Read_L2B_PCD_ADS(fileunit,L2BC%PCD_ADS,error_flag)
       IF (error_flag .ne. no_error) goto 999

       inquire(unit=fileunit,nextrec=pos_end)

       nbytes_read         = pos_end-pos_start
       nbytes_expected     = GetByteSize_L2B_PCD_ADS(L2BC%PCD_ADS,&
                                 N_Meas_to_write         = N_Meas,&
                                 N_Obs_Mie_to_write      = N_Obs_Mie, &
                                 N_Obs_Rayleigh_to_write = N_Obs_Rayleigh)
       nbytes_given_in_dsd = L2BC_Headers%SPH%PCD_DSD%size_of_one_dsr

       IF (nbytes_read .ne. nbytes_expected) THEN
          print *,"ERROR in ReadL2Bdata: reading L2B_PCD_ADS"
          print *,"pos_start           = ",pos_start
          print *,"pos_end             = ",pos_end
          print *,"nbytes_read         = ",nbytes_read
          print *,"nbytes_expected     = ",nbytes_expected
          print *,"something went wrong when reading the L2B_PCD_ADS from"
          print *,"the binary file, probably a programming error ..."
          error_flag = error_programming
          goto 999
       END IF
       IF (nbytes_read .ne. nbytes_given_in_dsd) THEN
          print *,"ERROR in ReadL2Bdata: reading L2B_PCD_ADS"
          print *,"nbytes_expected     = ",nbytes_expected
          print *,"nbytes_given_in_dsd = ",nbytes_given_in_dsd
          print *,"something went wrong when reading the L2B_PCD_ADS from"
          print *,"the binary file, the dataset size given in the DSD"
          print *,"differs from the allocated dataset size."
          print *,"Maybe the L2B file definition changed ?"
          error_flag = error_reading_file
          goto 999
       END IF
       !  #]
    END IF ! do_ReadPCD
    
    IF (do_ReadMie) THEN
       !  #[
       IF (.not. associated(L2BC_Headers%SPH%Mie_DSD)) THEN
          print *,"Error in ReadL2BCfile_DBL:"
          print *,"L2BC%SPH%Mie_DSD seems not associated !"
          print *,"elements from it are needed for reading the Mie_MDS !!"
          error_flag = error_programming
          return
       END  IF
       IF (associated(L2BC%Mie_MDS)) THEN
          print *,"Error in ReadL2BCfile_DBL:"
          print *,"the L2BC%Mie_MDS seems already allocated."
          print *,"It should not be allocated before trying to read it"
          print *,"because this read routine tries to allocate it..."
          error_flag = error_programming
          return
       END IF

       allocate(L2BC%Mie_MDS,stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine ReadL2BCfile_DBL..."
          error_flag = error_allocate
          return
       ENDIF
       
       call Init_L2B_Mie_MDS(L2BC%Mie_MDS, int(N_meas,IntAs), &
            int(N_Obs_Mie,IntAs), int(N_Range_Gates,IntAs), error_flag)
       IF (error_flag .ne. no_error) return
    
       IF ( (DataSetRecord_to_read .lt. 1) .or. &
            (DataSetRecord_to_read .gt. &
            L2BC_Headers%SPH%Mie_DSD%nr_of_dataset_records)) THEN
          print *,"Error in ReadL2BCfile_DBL:"
          print *,"invalid DSR number: "
          print *,"DataSetRecord_to_read = ",DataSetRecord_to_read
          print *,"L2BC%SPH%Mie_DSD%nr_of_dataset_records = ",&
               L2BC_Headers%SPH%Mie_DSD%nr_of_dataset_records
          error_flag = error_reading_file
          return
       END IF
       
       ! set the file position to the start of the requested datablock
       ! taking into account which BRC to handle
       pos = L2BC_Headers%SPH%Mie_DSD%start_of_ds + &
            (DataSetRecord_to_read-1)*L2BC_Headers%SPH%Mie_DSD%size_of_one_dsr
       call SetFilePointer(fileunit,pos,error_flag)
       IF (error_flag .ne. no_error) return

       print *,"trying to read the binary Mie_MDS nr. ",DataSetRecord_to_read

       inquire(unit=fileunit,nextrec=pos_start)
       print *,"pos_start=",pos_start

       call Read_L2B_Mie_MDS(fileunit,L2BC%Mie_MDS,error_flag)
       IF (error_flag .ne. no_error) goto 999

       inquire(unit=fileunit,nextrec=pos_end)

       nbytes_read         = pos_end-pos_start
       nbytes_expected     = Get_ByteSize_L2B_Mie_MDS(L2BC%Mie_MDS,&
                                                      N_Meas, N_Obs_Mie)
       nbytes_given_in_dsd = L2BC_Headers%SPH%Mie_DSD%size_of_one_dsr

       IF (nbytes_read .ne. nbytes_expected) THEN
          print *,"ERROR in ReadL2Bdata: reading L2B_Mie_MDS"
          print *,"pos_start           = ",pos_start
          print *,"pos_end             = ",pos_end
          print *,"nbytes_read         = ",nbytes_read
          print *,"nbytes_expected     = ",nbytes_expected
          print *,"something went wrong when reading the L2B_Mie_MDS from"
          print *,"the binary file, probably a programming error ..."
          error_flag = error_programming
          goto 999
       END IF
       IF (nbytes_read .ne. nbytes_given_in_dsd) THEN
          print *,"ERROR in ReadL2Bdata: reading L2B_Mie_MDS"
          print *,"nbytes_expected     = ",nbytes_expected
          print *,"nbytes_given_in_dsd = ",nbytes_given_in_dsd
          print *,"something went wrong when reading the L2B_Mie_MDS from"
          print *,"the binary file, the dataset size given in the DSD"
          print *,"differs from the allocated dataset size."
          print *,"Maybe the L2B file definition changed ?"
          error_flag = error_reading_file
          goto 999
       END IF
       !  #]
    END IF ! do_ReadMie
    
    IF (do_ReadRayleigh) THEN
       !  #[
       IF (.not. associated(L2BC_Headers%SPH%Rayleigh_DSD)) THEN
          print *,"Error in ReadL2BCfile_DBL:"
          print *,"L2BC%SPH%Rayleigh_DSD seems not associated !"
          print *,"elements from it are needed for reading the Rayleigh_MDS !!"
          error_flag = error_programming
          return
       END  IF
       IF (associated(L2BC%Rayleigh_MDS)) THEN
          print *,"Error in ReadL2BCfile_DBL:"
          print *,"the L2BC%Rayleigh_MDS seems already allocated."
          print *,"It should not be allocated before trying to read it"
          print *,"because this read routine tries to allocate it..."
          error_flag = error_programming
          return
       END IF

       allocate(L2BC%Rayleigh_MDS,stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine ReadL2BCfile_DBL..."
          error_flag = error_allocate
          return
       ENDIF
       
       call Init_L2B_Rayleigh_MDS(L2BC%Rayleigh_MDS, int(N_meas,IntAs), &
            int(N_Obs_Rayleigh,IntAs), int(N_Range_Gates,IntAs), error_flag)
       IF (error_flag .ne. no_error) return
    
       IF ( (DataSetRecord_to_read .lt. 1) .or. &
            (DataSetRecord_to_read .gt. &
            L2BC_Headers%SPH%Rayleigh_DSD%nr_of_dataset_records)) THEN
          print *,"Error in ReadL2BCfile_DBL:"
          print *,"invalid DSR number: "
          print *,"DataSetRecord_to_read = ",DataSetRecord_to_read
          print *,"L2BC%SPH%Rayleigh_DSD%nr_of_dataset_records = ",&
               L2BC_Headers%SPH%Rayleigh_DSD%nr_of_dataset_records
          error_flag = error_reading_file
          return
       END IF
       
       ! set the file position to the start of the requested datablock
       ! taking into account which BRC to handle
       pos = L2BC_Headers%SPH%Rayleigh_DSD%start_of_ds + &
            (DataSetRecord_to_read-1)*L2BC_Headers%SPH%Rayleigh_DSD%size_of_one_dsr
       call SetFilePointer(fileunit,pos,error_flag)
       IF (error_flag .ne. no_error) return

       print *,"trying to read the binary Rayleigh_MDS nr. ",&
            DataSetRecord_to_read

       inquire(unit=fileunit,nextrec=pos_start)
       print *,"pos_start=",pos_start

       call Read_L2B_Rayleigh_MDS(fileunit,L2BC%Rayleigh_MDS,error_flag)
       IF (error_flag .ne. no_error) goto 999

       inquire(unit=fileunit,nextrec=pos_end)

       nbytes_read         = pos_end-pos_start
       nbytes_expected     = Get_ByteSize_L2B_Rayleigh_MDS(L2BC%Rayleigh_MDS,&
                                                       N_Meas, N_Obs_Rayleigh)
       nbytes_given_in_dsd = L2BC_Headers%SPH%Rayleigh_DSD%size_of_one_dsr

       IF (nbytes_read .ne. nbytes_expected) THEN
          print *,"ERROR in ReadL2Bdata: reading L2B_Rayleigh_MDS"
          print *,"pos_start           = ",pos_start
          print *,"pos_end             = ",pos_end
          print *,"nbytes_read         = ",nbytes_read
          print *,"nbytes_expected     = ",nbytes_expected
          print *,"something went wrong when reading the L2B_Rayleigh_MDS from"
          print *,"the binary file, probably a programming error ..."
          error_flag = error_programming
          goto 999
       END IF
       IF (nbytes_read .ne. nbytes_given_in_dsd) THEN
          print *,"ERROR in ReadL2Bdata: reading L2B_Rayleigh_MDS"
          print *,"nbytes_expected     = ",nbytes_expected
          print *,"nbytes_given_in_dsd = ",nbytes_given_in_dsd
          print *,"something went wrong when reading the L2B_Rayleigh_MDS from"
          print *,"the binary file, the dataset size given in the DSD"
          print *,"differs from the allocated dataset size."
          print *,"Maybe the L2B file definition changed ?"
          error_flag = error_reading_file
          goto 999
       END IF
       !  #]
    END IF ! do_ReadRayleigh

    IF (do_ReadAssimPCD) THEN
       !  #[
       IF (.not. associated(L2BC_Headers%SPH%Assim_PCD_DSD)) THEN
          print *,"Error in ReadL2BCfile_DBL:"
          print *,"L2BC%SPH%Assim_PCD_DSD seems not associated !"
          print *,"elements from it are needed for reading "//&
               "the Assim_PCD_ADS !!"
          error_flag = error_programming
          return
       END  IF
       IF (associated(L2BC%Assim_PCD_ADS)) THEN
          print *,"Error in ReadL2BCfile_DBL:"
          print *,"the L2BC%Assim_PCD_ADS seems already allocated."
          print *,"It should not be allocated before trying to read it"
          print *,"because this read routine tries to allocate it..."
          error_flag = error_programming
          return
       END IF

       allocate(L2BC%Assim_PCD_ADS,stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine ReadL2BCfile_DBL..."
          error_flag = error_allocate
          return
       ENDIF
       
       call Init_L2C_AssimPCD_ADS(L2BC%Assim_PCD_ADS, int(N_meas,IntAs), &
            int(N_Obs_Mie,IntAs), int(N_Obs_Rayleigh,IntAs), &
            int(N_Range_Gates,IntAs), error_flag)
       IF (error_flag .ne. no_error) return
    
       IF ( (DataSetRecord_to_read .lt. 1) .or. &
            (DataSetRecord_to_read .gt. &
            L2BC_Headers%SPH%Assim_PCD_DSD%nr_of_dataset_records)) THEN
          print *,"Error in ReadL2BCfile_DBL:"
          print *,"invalid DSR number: "
          print *,"DataSetRecord_to_read = ",DataSetRecord_to_read
          print *,"L2BC%SPH%Assim_PCD_DSD%nr_of_dataset_records = ",&
               L2BC_Headers%SPH%Assim_PCD_DSD%nr_of_dataset_records
          error_flag = error_reading_file
          return
       END IF
       
       ! set the file position to the start of the requested datablock
       ! taking into account which BRC to handle
       pos = L2BC_Headers%SPH%Assim_PCD_DSD%start_of_ds + &
            (DataSetRecord_to_read-1)*L2BC_Headers%SPH%Assim_PCD_DSD%size_of_one_dsr
       call SetFilePointer(fileunit,pos,error_flag)
       IF (error_flag .ne. no_error) return

       print *,"trying to read the binary Assim_PCD_ADS nr. ",&
            DataSetRecord_to_read

       inquire(unit=fileunit,nextrec=pos_start)
       print *,"pos_start=",pos_start

       call Read_L2C_AssimPCD_ADS(fileunit,L2BC%Assim_PCD_ADS,error_flag)
       IF (error_flag .ne. no_error) goto 999

       inquire(unit=fileunit,nextrec=pos_end)

       nbytes_read         = pos_end-pos_start
       nbytes_expected     = GetByteSize_L2C_AssimPCD_ADS(&
                                L2BC%Assim_PCD_ADS,&
                                N_Obs_Mie, N_Obs_Rayleigh)
       nbytes_given_in_dsd = L2BC_Headers%SPH%Assim_PCD_DSD%size_of_one_dsr

       IF (nbytes_read .ne. nbytes_expected) THEN
          print *,"ERROR in ReadL2Bdata: reading L2C_Assim_PCD_ADS"
          print *,"pos_start           = ",pos_start
          print *,"pos_end             = ",pos_end
          print *,"nbytes_read         = ",nbytes_read
          print *,"nbytes_expected     = ",nbytes_expected
          print *,"something went wrong when reading the L2C_Assim_PCD_ADS"
          print *,"from the binary file, probably a programming error ..."
          error_flag = error_programming
          goto 999
       END IF
       IF (nbytes_read .ne. nbytes_given_in_dsd) THEN
          print *,"ERROR in ReadL2Bdata: reading L2B_Assim_PCD_ADS"
          print *,"nbytes_expected     = ",nbytes_expected
          print *,"nbytes_given_in_dsd = ",nbytes_given_in_dsd
          print *,"something went wrong when reading the L2B_Assim_PCD_ADS"
          print *,"from the binary file, the dataset size given in the DSD"
          print *,"differs from the allocated dataset size."
          print *,"Maybe the L2C file definition changed ?"
          error_flag = error_reading_file
          goto 999
       END IF
       !  #]
    END IF ! do_ReadAssimPCD

    IF (do_ReadMieVecWind) THEN
       !  #[
       IF (.not. associated(L2BC_Headers%SPH%Mie_VecWind_DSD)) THEN
          print *,"Error in ReadL2BCfile_DBL:"
          print *,"L2BC%SPH%Mie_VecWind_DSD seems not associated !"
          print *,"elements from it are needed for reading "//&
               "the Mie_VecWind_MDS !!"
          error_flag = error_programming
          return
       END  IF
       IF (associated(L2BC%Mie_VecWind_MDS)) THEN
          print *,"Error in ReadL2BCfile_DBL:"
          print *,"the L2BC%Mie_VecWind_MDS seems already allocated."
          print *,"It should not be allocated before trying to read it"
          print *,"because this read routine tries to allocate it..."
          error_flag = error_programming
          return
       END IF

       allocate(L2BC%Mie_VecWind_MDS,stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine ReadL2BCfile_DBL..."
          error_flag = error_allocate
          return
       ENDIF
       
       call Init_L2C_MieVec_MDS(L2BC%Mie_VecWind_MDS, int(N_meas,IntAs), &
            int(N_Obs_Mie,IntAs), int(N_Range_Gates,IntAs), error_flag)
       IF (error_flag .ne. no_error) return
    
       IF ( (DataSetRecord_to_read .lt. 1) .or. &
            (DataSetRecord_to_read .gt. &
            L2BC_Headers%SPH%Mie_VecWind_DSD%nr_of_dataset_records)) THEN
          print *,"Error in ReadL2BCfile_DBL:"
          print *,"invalid DSR number: "
          print *,"DataSetRecord_to_read = ",DataSetRecord_to_read
          print *,"L2BC%SPH%Mie_VecWind_DSD%nr_of_dataset_records = ",&
               L2BC_Headers%SPH%Mie_VecWind_DSD%nr_of_dataset_records
          error_flag = error_reading_file
          return
       END IF
       
       ! set the file position to the start of the requested datablock
       ! taking into account which BRC to handle
       pos = L2BC_Headers%SPH%Mie_VecWind_DSD%start_of_ds + &
            (DataSetRecord_to_read-1)*L2BC_Headers%SPH%Mie_VecWind_DSD%size_of_one_dsr
       call SetFilePointer(fileunit,pos,error_flag)
       IF (error_flag .ne. no_error) return

       print *,"trying to read the binary MieVecWind_MDS nr. ",&
            DataSetRecord_to_read

       inquire(unit=fileunit,nextrec=pos_start)
       print *,"pos_start=",pos_start

       call Read_L2C_MieVec_MDS(fileunit,L2BC%Mie_VecWind_MDS,error_flag)
       IF (error_flag .ne. no_error) goto 999

       inquire(unit=fileunit,nextrec=pos_end)

       nbytes_read         = pos_end-pos_start
       nbytes_expected     = Get_ByteSize_L2C_MieVec_MDS(&
                                       L2BC%Mie_VecWind_MDS, N_Obs_Mie)
       nbytes_given_in_dsd = L2BC_Headers%SPH%Mie_VecWind_DSD%size_of_one_dsr

       IF (nbytes_read .ne. nbytes_expected) THEN
          print *,"ERROR in ReadL2Bdata: reading L2C_Mie_VecWind_MDS"
          print *,"pos_start           = ",pos_start
          print *,"pos_end             = ",pos_end
          print *,"nbytes_read         = ",nbytes_read
          print *,"nbytes_expected     = ",nbytes_expected
          print *,"something went wrong when reading the L2C_Mie_VecWind_MDS"
          print *,"from the binary file, probably a programming error ..."
          error_flag = error_programming
          goto 999
       END IF
       IF (nbytes_read .ne. nbytes_given_in_dsd) THEN
          print *,"ERROR in ReadL2Bdata: reading L2B_Mie_VecWind_MDS"
          print *,"nbytes_expected     = ",nbytes_expected
          print *,"nbytes_given_in_dsd = ",nbytes_given_in_dsd
          print *,"something went wrong when reading the L2B_Mie_VecWind_MDS"
          print *,"from the binary file, the dataset size given in the DSD"
          print *,"differs from the allocated dataset size."
          print *,"Maybe the L2C file definition changed ?"
          error_flag = error_reading_file
          goto 999
       END IF
       !  #]
    END IF ! do_ReadMieVecWind

    IF (do_ReadRaylvecWind) THEN
       !  #[
       IF (.not. associated(L2BC_Headers%SPH%Rayleigh_VecWind_DSD)) THEN
          print *,"Error in ReadL2BCfile_DBL:"
          print *,"L2BC%SPH%Rayleigh_VecWind_DSD seems not associated !"
          print *,"elements from it are needed for reading "//&
               "the Rayleigh_VecWind_MDS !!"
          error_flag = error_programming
          return
       END  IF
       IF (associated(L2BC%Rayleigh_VecWind_MDS)) THEN
          print *,"Error in ReadL2BCfile_DBL:"
          print *,"the L2BC%Rayleigh_VecWind_MDS seems already allocated."
          print *,"It should not be allocated before trying to read it"
          print *,"because this read routine tries to allocate it..."
          error_flag = error_programming
          return
       END IF

       allocate(L2BC%Rayleigh_VecWind_MDS,stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine ReadL2BCfile_DBL..."
          error_flag = error_allocate
          return
       ENDIF
       
       call Init_L2C_RayleighVec_MDS(L2BC%Rayleigh_VecWind_MDS, &
                          int(N_meas,IntAs), int(N_Obs_Rayleigh,IntAs), &
                          int(N_Range_Gates,IntAs), error_flag)
       IF (error_flag .ne. no_error) return
    
       IF ( (DataSetRecord_to_read .lt. 1) .or. &
            (DataSetRecord_to_read .gt. &
            L2BC_Headers%SPH%Rayleigh_VecWind_DSD%nr_of_dataset_records)) THEN
          print *,"Error in ReadL2BCfile_DBL:"
          print *,"invalid DSR number: "
          print *,"DataSetRecord_to_read = ",DataSetRecord_to_read
          print *,"L2BC%SPH%Rayleigh_VecWind_DSD%nr_of_dataset_records = ",&
               L2BC_Headers%SPH%Rayleigh_VecWind_DSD%nr_of_dataset_records
          error_flag = error_reading_file
          return
       END IF
       
       ! set the file position to the start of the requested datablock
       ! taking into account which BRC to handle
       pos = L2BC_Headers%SPH%Rayleigh_VecWind_DSD%start_of_ds + &
            (DataSetRecord_to_read-1)*L2BC_Headers%SPH%Rayleigh_VecWind_DSD%size_of_one_dsr
       call SetFilePointer(fileunit,pos,error_flag)
       IF (error_flag .ne. no_error) return

       print *,"trying to read the binary RaylvecWind_MDS nr. ",&
            DataSetRecord_to_read

       inquire(unit=fileunit,nextrec=pos_start)
       print *,"pos_start=",pos_start

       call Read_L2C_RayleighVec_MDS(fileunit,L2BC%Rayleigh_VecWind_MDS,&
                                     error_flag)
       IF (error_flag .ne. no_error) goto 999

       inquire(unit=fileunit,nextrec=pos_end)

       nbytes_read         = pos_end-pos_start
       nbytes_expected     = Get_ByteSize_L2C_RaylVec_MDS(&
                                       L2BC%Rayleigh_VecWind_MDS, N_Obs_Mie)
       nbytes_given_in_dsd = L2BC_Headers%SPH%Rayleigh_VecWind_DSD%size_of_one_dsr

       IF (nbytes_read .ne. nbytes_expected) THEN
          print *,"ERROR in ReadL2Bdata: reading L2C_Rayleigh_VecWind_MDS"
          print *,"pos_start           = ",pos_start
          print *,"pos_end             = ",pos_end
          print *,"nbytes_read         = ",nbytes_read
          print *,"nbytes_expected     = ",nbytes_expected
          print *,"something went wrong when reading the L2C_Rayleigh_VecWind_MDS"
          print *,"from the binary file, probably a programming error ..."
          error_flag = error_programming
          goto 999
       END IF
       IF (nbytes_read .ne. nbytes_given_in_dsd) THEN
          print *,"ERROR in ReadL2Bdata: reading L2B_Rayleigh_VecWind_MDS"
          print *,"nbytes_expected     = ",nbytes_expected
          print *,"nbytes_given_in_dsd = ",nbytes_given_in_dsd
          print *,"something went wrong when reading the L2B_Rayleigh_VecWind_MDS"
          print *,"from the binary file, the dataset size given in the DSD"
          print *,"differs from the allocated dataset size."
          print *,"Maybe the L2C file definition changed ?"
          error_flag = error_reading_file
          goto 999
       END IF
       !  #]
    END IF ! do_ReadRaylvecWind

    call CloseDirectBinaryFile(fileunit)

    return

999 print *,"ABORTING subroutine ReadL2BCfile_DBL"
    call CloseDirectBinaryFile(fileunit)
    return

  end subroutine ReadL2BCfile_DBL
    !  #]
  !---------------------------------
end module ReadL2BCdata
