module Read_L2B_AuxPar_data
  !  #[ Description:
  !
  !     a module to read the file describing the aeolus L2B 
  !     processor auxiliary parameters
  !
  !             
  ! created:     05-04-2006 by J. de Kloe, KNMI
  ! 
  ! Modifications:
  !     12-May-2006 D. Tan     Null use of ParKind1 module
  !     03-Apr-2007 J. de Kloe Updated the FH filename consistency check
  !     07-Jun-2007 J. de Kloe adapted to changed xml interface
  !     19-Oct-2007 J. de Kloe added a nullify statement
  !     16-Jan-2008 J. de Kloe phase out integer kind i_
  !
  !  #]
  !  #[ Modules used:

  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE FixedHeader, only: FH_type, Init_FH, Print_FH, get_FH_from_xml_tree, &
                         file_type_AuxPar
  USE MainProductHeader, only: MPH_type, Init_MPH, Print_MPH, &
                               Get_MPH_from_xml_tree

  USE L2B_AuxPar_SPH, only: Init_L2B_AuxPar_SPH, &
                            Get_L2B_AuxPar_SPH_from_xmltree, &
                            Print_L2B_AuxPar_SPH, &
                            AsciiWrite_L2B_AuxPar_SPH

  USE L2B_AuxPar_DataStructure, only: Init_L2B_AuxPar, L2B_AuxPar_Type, &
       Init_L2B_AuxPar_Data, Get_L2BAuxParData_from_xml_tree

  USE XML_module, only: xml_node_type, Init_XML, xml_load_file, &
                        xml_tree_print, delete_xml_tree,&
                        ee_root_name, ee_h_name, &
                        xml_tree_get_namespace, uri_ns_header

!  USE EE_CFI_Datatypes, only: GetIntAuc, GetIntAus, IntAuc, IntAus, IntAs

  USE Logging, only: logmsg,log_debug,log_info,log_progress,log_warn,log_error
  USE ErrorHandler, only: no_error, error_xml, error_programming, &
                          error_opening_file, error_reading_file,&
                          error_allocate
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE LunManager, only: get_lun, free_lun
  USE StringTools, only: Split_path_and_filename
  USE AeolusConstants, only: iodd_version_AuxPar

  !  #]  
  !  #[ Parameter/Variable/Type declarations

  implicit none

  ! make all module contents private, unless explicitely stated otherwise
  private

  ! the default verbosity setting is Error (as silent as possible)
  integer, save :: LocalVerbosity = VerbosityError

  ! remember the initialisation state of the module
  logical, save :: module_is_initialised = .false.

  !  #]
  !  #[ Module interface
  public :: InitRead_L2B_AuxPar_fileModule, Read_L2B_AuxPar_file
  !  #]
contains
  !---------------------------------
  subroutine InitRead_L2B_AuxPar_fileModule(error_flag, ChoosenVerbosity)
    !  #[
    integer,           intent(out) :: error_flag
    integer, optional, intent(in)  :: ChoosenVerbosity

    error_flag = no_error

    IF (present(ChoosenVerbosity)) LocalVerbosity = ChoosenVerbosity

    IF (LocalVerbosity .ge. VerbosityReport) &
         call print_verbosity_setting(LocalVerbosity,&
         ModuleName="Read_L2B_AuxPar_file")

    call Init_XML(error_flag) !(ChoosenVerbosity=VerbosityReport)
    if (error_flag .ne. no_error) return

    module_is_initialised = .true.

  end subroutine InitRead_L2B_AuxPar_fileModule
    !  #]
  subroutine Read_L2B_AuxPar_file(L2B_AuxPar, filename, error_flag)
    !  #[
    type(L2B_AuxPar_Type), intent(inout)  :: L2B_AuxPar
    character(len=*),      intent(in)     :: filename ! without extension
    integer,               intent(out)    :: error_flag

    ! local variables
    integer                      :: AllocStatus
    TYPE(xml_node_type), pointer :: xml_tree
    character(len=256)           :: filename_without_path, path
    character(len=256)           :: xml_namespace_AuxPar
    character(len=256)           :: xml_namespace_uri

    ! init
    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       call logmsg(log_error,&
            "Error in Read_L2B_AuxPar_file(): "//&
            "the Read_L2B_AuxPar_file module must be initialised"//&
            "before use ...")
       error_flag = error_programming
       return
    END IF
    
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"Reading L2B_AuxPar input file..."

    nullify(xml_tree)
    call xml_load_file(filename, xml_tree, error_flag)
    IF (error_flag .ne. no_error) return

    ! define which xml namespace should be present
    !example: xmlns="http://www.esa.int/schemas/ae/ALD_U_N_2B_01.01"
    !uri_ns_header    ! defined in xml_module
    !file_type_L2B_N  ! defined in FixedHeader
    !iodd_version_L2B ! defined in AeolusConstants
    xml_namespace_AuxPar = trim(uri_ns_header)//trim(file_type_AuxPar)//&
                           "_"//trim(iodd_version_AuxPar)

    ! check whether this namespace is really present in the input file
    call xml_tree_get_namespace(xml_tree,xml_namespace_uri,error_flag)
    if (error_flag .ne. no_error) then
       ! allow this attribute to be missing for now, to allow use
       ! of outdated files.
       ! TODO: remove this exception
       error_flag = no_error
       call logmsg(log_warn,"WARNING in Read_L2B_AuxPar_file: "//&
                   "could not find xmlns attribute in the root tag "//&
                   "of file: "//trim(filename))
    else
       if (trim(xml_namespace_AuxPar) .ne. trim(xml_namespace_uri)) &
            call logmsg(log_warn,"WARNING in Read_L2B_AuxPar_file: "//&
                        "expected an xmlns attribute with content: "//&
                        trim(xml_namespace_AuxPar)//", but found an "//&
                        "xmlns attribute with content: "//&
                        trim(xml_namespace_uri)//". This input file may "//&
                        "not be compatible with the current software!")
    end if


    call logmsg(log_debug,"Read_L2B_AuxPar_file: the xml file was "//&
                          "successfully read")

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"==================="
       call xml_tree_print(xml_tree, error_flag)
       IF (error_flag .ne. no_error) return
       print *,"==================="
    END IF
 
    !------------------------------
    ! allocate memory to hold the FH, MPH and SPH elements
    !------------------------------
    allocate(L2B_AuxPar%FH,L2B_AuxPar%MPH,L2B_AuxPar%SPH,&
             L2B_AuxPar%Data, STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       call logmsg(log_error,&
                   "Allocation failure in subroutine Read_L2B_AuxPar_file")
       error_flag = error_allocate
       return
    END IF

    call Init_FH(L2B_AuxPar%FH,error_flag)
    IF (error_flag .ne. no_error) return

    call Init_MPH(L2B_AuxPar%MPH,error_flag)
    IF (error_flag .ne. no_error) return

    call Init_L2B_AuxPar_SPH(L2B_AuxPar%SPH,error_flag)
    IF (error_flag .ne. no_error) return

    call Init_L2B_AuxPar_Data(L2B_AuxPar%Data,error_flag)
    IF (error_flag .ne. no_error) return

    !------------------------------
    ! copying the FH XML elements
    !------------------------------
    print *,"copying items to the FH datastructure"
    path = "/"//ee_root_name//"/"//ee_h_name
    call Get_FH_from_xml_tree(xml_tree, path, L2B_AuxPar%FH, error_flag)
    IF (error_flag .ne. no_error) return

    call Split_path_and_filename(filename,path,filename_without_path)

    ! check if: FH%File_Name .eq. filename
    IF (trim(L2B_AuxPar%FH%File_Name)//".EEF" .ne. filename_without_path) THEN
       call logmsg(log_warn,"inconsistent name of L2B_AuxPar%FH%File_Name:"//&
            " in Read_L2B_AuxPar_file: current filename = "//&
            trim(filename_without_path)//&
            " but inside the FixedHeader the File_Name field holds: "//&
            trim(L2B_AuxPar%FH%File_Name)//" [without .EEF extention]"//&
            " maybe the current file was renamed ???")
    END IF

    !------------------------------
    ! copying the MPH XML elements
    !------------------------------
    print *,"copying items to the MPH datastructure"
    path = "/"//ee_root_name//"/"//ee_h_name
    call Get_MPH_from_xml_tree(xml_tree, path, L2B_AuxPar%MPH, error_flag)
    IF (error_flag .ne. no_error) return

    !------------------------------
    ! copying the SPH XML elements (with DSD's)
    !------------------------------
    print *,"copying items to the SPH datastructure"
    call Get_L2B_AuxPar_SPH_from_xmltree(xml_tree, path, L2B_AuxPar%SPH, error_flag)
    IF (error_flag .ne. no_error) return

    !------------------------------
    ! copying the L2B_AuxPar_Data XML elements
    !------------------------------
    print *,"copying items to the Data datastructure"
    call Get_L2BAuxParData_from_xml_tree(xml_tree,&
         "/Earth_Explorer_File/Data_Block/Level_2B_Proc_Params",&
         L2B_AuxPar%Data,error_flag)
    IF (error_flag .ne. no_error) return

    !------------------------------
    ! done                         
    !------------------------------

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"deallocating xml items"
    call delete_xml_tree(xml_tree, error_flag)
    IF (error_flag .ne. no_error) return

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"done with reading L2B_AuxPar file"

    return

  end subroutine Read_L2B_AuxPar_file
    !  #]
  !---------------------------------
end module Read_L2B_AuxPar_data
