program TestLiteDataModule
  !  #[ Description:
  !     A test program for the LiteDataModule module
  !   
  !     created:     17-11-2005, by: J. de Kloe, KNMI
  !
  !     Modifications:
  !     12-May-2006 D. Tan       Null use of ParKind1 module
  !     20-Jul-2006 J. de Kloe   added test for Get_Nr_available_BRCs()
  !     05-Oct-2006 J. de Kloe   adapted to the new WorkingData structure
  !     03-Nov-2006 J. de Kloe   added call to AddDummyLatLonData()
  !                              and testprint for DateTime
  !     16-Jan-2008 J. de Kloe   phase out integer kind i_
  !
  !  #]
  !  #[ Modules used:
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE LiteDataModule, only: Fill_WorkingData_with_Lite_Data,&
                            Get_Nr_available_BRCs, AddDummyLatLonData
  USE MolScat, only: InitMolScatModule, get_ref_mol_scat
  ! a routine to define the globally used L2B data struct
  USE Working_DataStructure, only: WorkingDataType, &
       InitWorkingData, DeleteWorkingData
  USE EE_CFI_Datatypes, only: IntAs, maxlen_datetime_kvt,construct_kvt_datetime
  ! support modules
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, VerbosityReport
  USE numerics, only: r8_
  USE ErrorHandler, only: no_error, error_allocate, program_abort
  !  #]
  !  #[ Parameter/Variable/Type declarations

  implicit none
  character(len=256) :: filename
  integer            :: error_flag
  integer            :: nr_BRCs_available
  integer            :: read_which_BRC
  integer            :: m, r
  real(r8_)          :: alt_mid, beta_m
  character(len=maxlen_DateTime_KVT) :: txt_DateTime_KVT

  ! define how the measurements are organised
  ! (in 67 profiles consisting of 24 range gates each)
  integer(IntAs), parameter :: N_Obs_Mie_Max      = 5
  integer(IntAs), parameter :: N_Obs_Rayleigh_Max = 5
  integer(IntAs), parameter :: N_Meas             = 67
  integer(IntAs), parameter :: N_Range_Gates      = 24

  ! a struct to hold all necessary data for the processor
  TYPE(WorkingDataType), save :: WorkingData

  !  #]
  !  #[ program code
  print *,"running testprogram: TestLiteDataModule"

  ! initialise the modules used by the program
  call InitWorkingData(WorkingData, N_range_gates, N_meas, &
                       N_Obs_Mie_Max, N_Obs_Rayleigh_Max, error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"InitWorkingData")

  call InitMolScatModule()

  filename = "backscatterdata.dat"

  print *,"testing subroutine: Get_Nr_available_BRCs"
  call Get_Nr_available_BRCs(filename,nr_BRCs_available,error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"Get_Nr_available_BRCs")
  print *,"nr_BRCs_available = ",nr_BRCs_available

  ! loop over all BRCs in the file
  print *,"testing subroutine: Fill_WorkingData_With_Lite_Data"
  read_which_BRC = 1
  call Fill_WorkingData_With_Lite_Data(WorkingData, filename, read_which_BRC, &
                                       nr_BRCs_available, error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"Fill_WorkingData_With_Lite_Data")

  print *,"Lite data was successfully read"

  call AddDummyLatLonData(WorkingData)
  print *,"Dummy latlon and datetime added"

  ! printing data for 1st and last Rayleigh measurement
  print *,"Printing molecular data (Rayleigh):"
  print *,"REALACC(6)"
  DO m=1,WorkingData%nr_of_measurements,WorkingData%nr_of_measurements-1
     txt_DateTime_KVT = Construct_KVT_DateTime(WorkingData%MieMeas(m)%datetime)
     print *,"dummy Lite datetime Rayleigh = ",trim(txt_DateTime_KVT)," [UTC]"
     print *,"alt_start alt_end beta_m     obs_signal"
     DO r=1,WorkingData%RayleighMeas(m)%nr_of_range_gates
        alt_mid = 0.5_r8_*(WorkingData%RayleighMeas(m)%RangeGate(r)%alt_start+&
                           WorkingData%RayleighMeas(m)%RangeGate(r)%alt_end   )
        beta_m = get_ref_mol_scat(alt_mid,two_digits_only=.true.)
        print "(1x,i5,5x,i5,1x,2(1x,es10.3E2))",&
             nint(WorkingData%RayleighMeas(m)%RangeGate(r)%alt_start),&
             nint(WorkingData%RayleighMeas(m)%RangeGate(r)%alt_end),&
             beta_m,&
             WorkingData%RayleighMeas(m)%RangeGate(r)%obs_signal_level
     END DO
  END DO
  print *,"ENDREALACC"

  ! printing data for 1st and last Mie measurement
  print *,"Printing Lite aerosol data (Mie):"
  print *,"REALACC(6)"
  DO m=1,WorkingData%nr_of_measurements,WorkingData%nr_of_measurements-1
     txt_DateTime_KVT = &
          Construct_KVT_DateTime(WorkingData%RayleighMeas(m)%datetime)
     print *,"dummy Lite datetime Mie = ",trim(txt_DateTime_KVT)," [UTC]"
     print *,"alt_start alt_end obs_signal"
     DO r=1,WorkingData%MieMeas(m)%nr_of_range_gates
        print "(1x,i5,5x,i5,2x,es10.3E2)",&
             nint(WorkingData%MieMeas(m)%RangeGate(r)%alt_start),&
             nint(WorkingData%MieMeas(m)%RangeGate(r)%alt_end),&
             WorkingData%MieMeas(m)%RangeGate(r)%obs_signal_level
     END DO
  END DO
  print *,"ENDREALACC"

  ! free allocated memory when done
  call DeleteWorkingData(WorkingData)

  !  #]
end program TestLiteDataModule
