module LiteDataModule
  !  #[ Description:
  !     A module to
  !     -read ascii files holding Lite data, to be used for testing the L2BP 
  !     -the file holds the following data:
  !           lvl_start lvl_end    : rangebin definition 
  !           beta_mol signal      : mol. backscatter and local laser signal
  !           obs_mie obs_rayleigh : observed molecular and aerosol backscatter
  !           layer_extinction     : extinciton for this rangebin
  !   
  ! created:       17-11-2005, by: J. de Kloe, KNMI
  !
  ! Modifications:
  !     26-Jan-2006 J. de Kloe   added abs() to suppress the -0.00 values 
  !                              for obs_signal that are in the test datafile
  !     12-May-2006 D. Tan       Null use of ParKind1 module
  !     20-Jul-2006 J. de Kloe   added subr. Get_Nr_available_BRCs()
  !     05-Oct-2006 J. de Kloe   adapted to the new WorkingData structure
  !     03-Nov-2006 J. de Kloe   added dummy filling for datetime fields
  !     26-Jun-2007 J. de Kloe   added some prints for dummy variables that
  !                              are read from file but never used, to 
  !                              suppress warnings issued by g95
  !     16-Jan-2008 J. de Kloe   phase out integer kind i_
  !
  !  #]
  !  #[ Modules used:
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  ! a routine to define the globally used L2B data struct
  USE Working_DataStructure, only: WorkingDataType, &
       WD_MieRangeGateType       => MieRangeGateType, &
       WD_RayleighRangeGateType  => RayleighRangeGateType 
  USE EE_CFI_Datatypes, only: SetDateTime
  ! support modules
  USE numerics, only: r8_, missing_indicator_real
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, VerbosityReport
  USE LunManager, only: get_lun, free_lun
  USE ErrorHandler, only: no_error, error_opening_file, error_reading_file
  !  #]
  !  #[ Parameter/Variable/Type declarations
  implicit none

  ! make all module contents private, unless explicitely stated otherwise
  private

  ! the default verbosity setting is Error (as silent as possible)
  integer, private, save :: LocalVerbosity = VerbosityError

  !  #]
  !  #[ Module interface
  ! specify which routines are public, so are part of the interface
  ! to this module
  ! public parameters
  ! -
  ! public types
  ! -
  ! public functions/subroutines
  public :: Fill_WorkingData_With_Lite_Data, AddDummyLatLonData, &
            Get_Nr_available_BRCs
  !  #]
contains
  !------------------
  subroutine Get_Nr_available_BRCs(filename,nr_BRCs_available,error_flag)
    !  #[
    character(len=*),   intent(in)    :: filename          ! input
    integer,        intent(out)   :: nr_BRCs_available ! output
    integer,        intent(out)   :: error_flag        ! output

    ! local variables
    integer            :: nr_of_BRCs_in_file
    integer            :: nr_of_measurements_in_file
    integer            :: nr_of_range_gates_in_file
    character(len=10)  :: dummyline
    integer            :: ios, fileunit

    ! init
    error_flag = no_error

    ! open the input file
    if (LocalVerbosity >= VerbosityReport) &
         print *,"opening file: ",trim(filename)

    fileunit = get_lun()
    open(unit=fileunit,file=filename,status="old",&
         action="read",form="FORMATTED",ERR=99)
    
    ! read the first 2 lines to extract the nr_BRCs_available

    ! the 1st line contains some comments
    read(fileunit,"(a)",iostat=ios) dummyline
    IF (ios .ne. 0) goto 98

    ! the 2nd line contains the nr of measurements and
    ! the nr of ranges per measurement
    read(fileunit,*,iostat=ios) nr_of_BRCs_in_file, &
                                nr_of_measurements_in_file, &
                                nr_of_range_gates_in_file
    IF (ios .ne. 0) goto 98

    nr_BRCs_available = nr_of_BRCs_in_file

    close(unit=fileunit)
    call free_lun(fileunit)

    if (LocalVerbosity >= VerbosityReport) THEN
       print *,"Successfully retrieved nr_BRCs_available from file ",&
            trim(filename)
       print *,"nr_BRCs_available = ",nr_BRCs_available
    END if

    return

98  print *,"ERROR: while reading input file: ",trim(filename)
    print *,"dummyline = ",trim(dummyline)
    print *,"nr_of_measurements_in_file = ",nr_of_measurements_in_file
    print *,"nr_of_range_gates_in_file  = ",nr_of_range_gates_in_file
    error_flag = error_reading_file
    return

99  print *,"ERROR: unable to open input file: ",trim(filename)
    error_flag = error_opening_file
    return

  end subroutine Get_Nr_available_BRCs
    !  #]
  subroutine Fill_WorkingData_With_Lite_Data(WorkingData, filename, &
                      read_which_BRC, nr_BRCs_available, error_flag)
    !  #[

    TYPE(WorkingDataType), intent(inout) :: WorkingData       ! input/output
    character(len=*),      intent(in)    :: filename          ! input
    integer,               intent(in)    :: read_which_BRC    ! input
    integer,               intent(out)   :: nr_BRCs_available ! output
    integer,               intent(out)   :: error_flag        ! output

    ! local variables
    integer                 :: brc,p,a_r,a_m,N_meas,ios
    integer                 :: nr_of_BRCs_in_file
    integer                 :: nr_of_measurements_in_file
    integer                 :: nr_of_range_gates_in_file
    character(len=10)       :: dummyline
    real(r8_), dimension(7) :: data
    TYPE(WD_MieRangeGateType),      pointer :: WD_MieRangeGate 
    TYPE(WD_RayleighRangeGateType), pointer :: WD_RayleighRangeGate 
    integer                 :: fileunit

    ! init
    error_flag = no_error

    ! open the input file
    if (LocalVerbosity >= VerbosityReport) &
         print *,"opening file: ",trim(filename)

    fileunit = get_lun()
    open(unit=fileunit,file=filename,status="old",&
         action="read",form="FORMATTED",ERR=99)
    
    ! read the first 2 lines to extract the nr_BRCs_available

    ! the 1st line contains some comments
    read(fileunit,"(a)",iostat=ios) dummyline
    IF (ios .ne. 0) goto 98

    ! the 2nd line contains the nr of measurements and
    ! the nr of ranges per measurement
    read(fileunit,*,iostat=ios) nr_of_BRCs_in_file, &
                                nr_of_measurements_in_file, &
                                nr_of_range_gates_in_file
    IF (ios .ne. 0) goto 98

    nr_BRCs_available = nr_of_BRCs_in_file

    rewind(fileunit)

    !  #[ sanity check on the input data

    IF (read_which_BRC .gt. nr_BRCs_available) THEN
       print *,"ERROR reading backscatter data:"
       print *,"it is not possible to read the ",read_which_BRC,"th BRC"
       print *,"since this file only holds ",nr_BRCs_available," BRCs."
       error_flag = error_reading_file
       return
    END IF

    IF (read_which_BRC .lt. 1) THEN
       print *,"ERROR reading backscatter data:"
       print *,"it is not possible to read the ",read_which_BRC,"th BRC"
       print *,"This number should always above 0"
       error_flag = error_reading_file
       return
    END IF

    ! check if they coincide with the values set by the initialisation
    IF (nr_of_measurements_in_file .ne. WorkingData%nr_of_measurements) THEN
       IF (nr_of_measurements_in_file .gt. WorkingData%nr_of_measurements) THEN
          print *,"WARNING in ReadMeasData:"
          print *,"the WorkingData datastructure was initialised for using"
          print *,"nr_of_measurements = ",WorkingData%nr_of_measurements
          print *,"measurement profiles, but this file "
          print *,"contains: ",nr_of_measurements_in_file," measurements !!!"
          print *,"excess measurement data will be skipped ..."
       ELSE
          print *,"ERROR in ReadMeasData:"
          print *,"the WorkingData datastructure was initialised for using"
          print *,"nr_of_measurements = ",WorkingData%nr_of_measurements
          print *,"measurement profiles, but this file only seems"
          print *,"to contain: ",nr_of_measurements_in_file," measurements !!!"
          
          error_flag = error_reading_file
          return
       END IF
    END IF

    IF ( nr_of_range_gates_in_file .lt. &
         WorkingData%MieMeas(1)%nr_of_range_gates) THEN
       IF (LocalVerbosity >= VerbosityWarn) THEN
          print *,"WARNING in ReadMeasData:"
          print *,"the Mie measurement datastructure was initialised for using"
          print *,"nr_of_range_gates = ",&
               WorkingData%MieMeas(1)%nr_of_range_gates
          print *,"range gates per profile, but this file only seems"
          print *,"to contain: ",nr_of_range_gates_in_file,&
               " range gates per profile !!!"
          print *,"the remaining range gates will contain missing values !"
       END IF
    ELSE
       IF ( nr_of_range_gates_in_file .gt. &
            WorkingData%MieMeas(1)%nr_of_range_gates) THEN
          print *,"ERROR in ReadMeasData:"
          print *,"the Mie measurement datastructure was initialised for using"
          print *,"nr_of_range_gates = ",&
               WorkingData%MieMeas(1)%nr_of_range_gates
          print *,"range gates per profile, but this file seems"
          print *,"to contain: ",nr_of_range_gates_in_file,&
               " range gates per profile !!!"
          print *,"There is not enough room to store all this data in "
          print *,"the WorkingData data structure !!!"
          error_flag = error_reading_file
          return
       END IF
    END IF

    IF ( nr_of_range_gates_in_file .lt. &
         WorkingData%RayleighMeas(1)%nr_of_range_gates) THEN
       IF (LocalVerbosity >= VerbosityWarn) THEN
          print *,"WARNING in ReadMeasData:"
          print *,"the Rayleigh measurement datastructure was initialised for"
          print *,"using nr_of_range_gates = ",&
               WorkingData%RayleighMeas(1)%nr_of_range_gates
          print *,"range gates per profile, but this file only seems"
          print *,"to contain: ",nr_of_range_gates_in_file,&
               " range gates per profile !!!"
          print *,"the remaining range gates will contain missing values !"
       END IF
    ELSE
       IF ( nr_of_range_gates_in_file .gt. &
            WorkingData%RayleighMeas(1)%nr_of_range_gates) THEN
          print *,"ERROR in ReadMeasData:"
          print *,"the Rayleigh measurement datastructure was initialised for"
          print *,"using nr_of_range_gates = ",&
               WorkingData%RayleighMeas(1)%nr_of_range_gates
          print *,"range gates per profile, but this file seems"
          print *,"to contain: ",nr_of_range_gates_in_file,&
               " range gates per profile !!!"
          print *,"There is not enough room to store all this data in "
          print *,"the WorkingData data structure !!!"
          error_flag = error_reading_file
          return
       END IF
    END IF

    !  #]

    brc_loop: DO brc=1,nr_BRCs_available

       print *,"reading backscatter data for brc: ",brc
    
       ! the 1st line contains some comments
       read(fileunit,"(a)",iostat=ios) dummyline
       IF (ios .ne. 0) goto 98

       ! the 2nd line contains the nr of measurements and
       ! the nr of ranges per measurement
       read(fileunit,*,iostat=ios) nr_of_BRCs_in_file, &
                                   nr_of_measurements_in_file, &
                                   nr_of_range_gates_in_file
       IF (ios .ne. 0) goto 98

       ! the 1st line in the brc-block contains some comments
       read(fileunit,"(a)",iostat=ios) dummyline
       IF (ios .ne. 0) goto 98

       N_meas = WorkingData%nr_of_measurements

       meas_loop: DO p=1,nr_of_measurements_in_file 
          !  #[ read the Rayleigh data
          rangegate_loop1: DO a_r=1,nr_of_range_gates_in_file
             read(fileunit,*,iostat=ios) data(1:7)
             IF (ios .ne. 0) goto 98

             ! only use this data for measurements that are defined
             ! in the WorkingData datastructure.
             ! Remind that N_meas can be 30 or 67
             ! while this temporary datafile has 67 measurements,
             ! the 37 profiles are skipped if N_meas=30
             IF (p .le. N_meas) THEN
                ! point to this RangeGate for shorter notation
                WD_RayleighRangeGate => WorkingData%RayleighMeas(p)%RangeGate(a_r)

                ! copy the data into the "WorkingData" datastructure and
                ! round the altitudes to integer numbers to prevent the checks
                ! below to fail in case of numerical un-accuracies
                WD_RayleighRangeGate%alt_start        = 1._r8_*nint(data(1))
                WD_RayleighRangeGate%alt_end          = 1._r8_*nint(data(2))
                WD_RayleighRangeGate%obs_signal_level = data(6)

                ! take missing values into account
                ! (typically happens if we are below the orography)
                IF (WD_RayleighRangeGate%obs_signal_level .le. -0.5_r8_) &
                     WD_RayleighRangeGate%obs_signal_level = &
                                          missing_indicator_real

                ! suppress the -0.00 values that are in the test datafile
                WD_RayleighRangeGate%obs_signal_level = &
                     abs(WD_RayleighRangeGate%obs_signal_level)

             END IF
          END DO rangegate_loop1
          !  #]
          !  #[ read the Mie data
          rangegate_loop2: DO a_m=1,nr_of_range_gates_in_file
             read(fileunit,*,iostat=ios) data(1:7)
             IF (ios .ne. 0) goto 98

             ! only use this data for measurements that are defined
             ! in the WorkingData datastructure. 
             ! Remind that N_meas can be 30 or 67
             ! while this temporary datafile has 67 measurements,
             ! to 37 profiles are skipped if N_meas=30
             IF (p .le. N_meas) THEN
                ! point to this RangeGate for shorter notation
                WD_MieRangeGate      => WorkingData%MieMeas(p)%RangeGate(a_m)

                ! copy the data into the "WorkingData" datastructure and
                ! round the altitudes to integer numbers to prevent the checks
                ! below to fail in case of numerical un-accuracies
                WD_MieRangeGate%alt_start            = 1.*nint(data(1))
                WD_MieRangeGate%alt_end              = 1.*nint(data(2))
                WD_MieRangeGate%obs_signal_level     = data(5)
                
                ! take missing values into account
                ! (typically happens if we are below the orography)
                IF (WD_MieRangeGate%obs_signal_level .le. -0.5) &
                     WD_MieRangeGate%obs_signal_level = missing_indicator_real

                ! suppress the -0.00 values that are in the test datafile
                WD_MieRangeGate%obs_signal_level = &
                     abs(WD_MieRangeGate%obs_signal_level)

             END IF
          END DO rangegate_loop2
          !  #]
       END DO meas_loop

       IF (brc .eq. read_which_BRC) exit brc_loop
       
       if (LocalVerbosity >= VerbosityReport) &
            print *,"requested BRC not yet reached, "//&
            "reading the next from the input file..."
    END DO brc_loop

    close(unit=fileunit)
    call free_lun(fileunit)

    if (LocalVerbosity >= VerbosityReport) THEN
       print *,"Successfully loaded BRC ",read_which_BRC," from file ",trim(filename)
       print *,"done reading data"
    END if

    return

98  print *,"ERROR: while reading input file: ",trim(filename)
    print *,"dummyline = ",trim(dummyline)
    error_flag = error_reading_file
    return

99  print *,"ERROR: unable to open input file: ",trim(filename)
    error_flag = error_opening_file
    return

  end subroutine Fill_WorkingData_With_Lite_Data
    !  #]
  subroutine AddDummyLatLonData(WorkingData)
    !  #[
    TYPE(WorkingDataType), intent(inout) :: WorkingData ! input/output

    ! local variables
    integer   :: p,a
    TYPE(WD_MieRangeGateType), pointer      :: MieRangeGate 
    TYPE(WD_RayleighRangeGateType), pointer :: RayleighRangeGate 
    real(r8_) :: lat_start, lat_step
    real(r8_) :: lon_start, lon_step

    ! init
    lat_start = 55.0_r8_
    lon_start = 5.0_r8_
    lat_step = 0.6_r8_ * 1._r8_/111._r8_ ! 111 km/degree, so 1./111. degree/km
    lon_step = 0.8_r8_ * 1._r8_/111._r8_ ! 111 km/degree, so 1./111. degree/km
    DO p=1,WorkingData%nr_of_measurements
       ! set the datetime stamp such that collocation will be successfull
       call SetDateTime(WorkingData%MieMeas(p)%datetime,&
                        2004,12,01,0,0,0._r8_)
       ! 01-DEC-2004 00:00:00.000000
       DO a=1,WorkingData%MieMeas(1)%nr_of_range_gates
          ! point to this RangeGate for shorter notation
          MieRangeGate      => WorkingData%MieMeas(p)%RangeGate(a)
          ! copy the data into the "L2B" datastructure
          MieRangeGate%lat      = lat_start + (p-1)*lat_step
          MieRangeGate%lon      = lon_start + (p-1)*lon_step
       END DO
       ! set the datetime stamp such that collocation will be successfull
       call SetDateTime(WorkingData%RayleighMeas(p)%datetime,&
                        2004,12,01,0,0,0._r8_)
       ! 01-DEC-2004 00:00:00.000000
       DO a=1,WorkingData%RayleighMeas(1)%nr_of_range_gates
          ! point to this RangeGate for shorter notation
          RayleighRangeGate => WorkingData%RayleighMeas(p)%RangeGate(a)
          ! copy the data into the "L2B" datastructure
          RayleighRangeGate%lat = lat_start + (p-1)*lat_step
          RayleighRangeGate%lon = lon_start + (p-1)*lon_step
       END DO

    END DO

    return

  end subroutine AddDummyLatLonData
    !  #]  
  !------------------
end module LiteDataModule

