module Match_AMD_Module
  !  #[ Description
  !
  !     this module contains the subroutines used to Match the 
  !     BRC locations to the available NWP profiles
  !     
  !     created 10-Nov-2006 by: J. de Kloe
  !
  !     Modifications:
  !        10-Nov-2006 J. de Kloe  defined a new directory for this module
  !        15-Feb-2007 J. de Kloe  fixed a bug for the dummy method
  !        12-Mar-2007 D. Tan      fixed a bug for the dummy method (lt/le)
  !        15-Mar-2007 J. de Kloe  removed some unused lines
  !        17-Apr-2007 D. Tan      Remove use_ECF switches.
  !        24-Apr-2007 J. de Kloe  phase out the SummaryVars code
  !        29-May-2007 D. Tan      Remove some obsolete commented code
  !        28-Aug-2007 D. Tan      Remove references to obsolete routines
  !        03-Sep-2007 P. Poli     Remove references to obsolete routines II
  !        19-Oct-2007 P. Poli     split AMD_DataType in two parts:
  !                                AMD_HDR_DataType and AMD_BRC_DataType
  !        16-Jan-2008 J. de Kloe  phase out integer kind i_
  !  #]
  !  #[ modules used

  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.

  ! modules defining L2B algoritms

  ! define types and functions for auxiliary data
  ! rename the profile_type, to prevent confusion with the other profile_type 
  ! defined in profileinterpolate.F90
!  USE AuxiliaryModule, only: aux_profile_type => profile_type

  USE AMD_DataStructure, only: AMD_BRC_DataType
  USE ReadAMDdata, only: ReadAMDfile

  ! a routine to define the globally used L1B data struct
  USE L1B_Geoloc_ADS, only: Rangebin_Meas_Geoloc_type, &
                            Rangebin_Obs_Geoloc_type
  USE L2B_AuxPar_DataStructure, only: L2B_AuxPar_Type, &
                            AMD_Matchup_Params_type, &
                            Matchup_Dummy, Matchup_Nearest_Neighbour

  ! a routine to define the globally used L2B data struct
  USE L2B_PCD_ADS, only: L2B_AMD_Collocations_Type, &
                         L2B_AMD_Coll_OK, L2B_AMD_Coll_Failed
  USE L2B_Proc_Settings, only: L2BP_Settings_type
  USE Working_DataStructure, only: WorkingDataType, &
                      WD_MieRangeGateType      => MieRangeGateType, &
                      WD_RayleighRangeGateType => RayleighRangeGateType  
  USE EE_CFI_Datatypes, only: IntAs, IntAuc, IntAl, GetIntAl,&
                              SetFAdoFull_or_Missing, &
                              DateTimeType, &
                              get_averaged_datetime, get_time_difference
!                              Construct_KVT_DateTime, maxlen_DateTime_KVT, &
  USE Logging, only: logmsg, &
                     log_debug,log_info,log_progress,log_warn,log_error

  USE ErrorHandler, only: no_error, error_programming, &
       error_writing_file, error_reading_file, &
       error_opening_file, error_allocate
  USE ReadJobOrderData, only: JobOrderData_type
  USE Screening_L1B_Data, only: Screen_L1B_Data

  ! support modules

  ! define the default integer and real to be used by the program
  USE Numerics, only: r8_, &
                      missing_indicator_real, missing_real
  ! define some parameters for verbosity control
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, VerbosityReport
  ! define error parameters, and a program abort routine
  USE ErrorHandler, only: no_error, error_programming, error_allocate, &
                          error_writing_file, error_opening_file
  USE LunManager, only: get_lun, free_lun
  USE LatLon_Handling, only: get_midpoint, get_distance

  !  #]
  !  #[ Variables

  implicit none

  ! define how the measurements are organised
  ! (in 67 profiles consisting of 24 range gates each)
  !integer, parameter :: N_meas        = 67 ! current used nr of meas.
  integer, parameter :: N_range_gates = 24 ! current used nr of gates
  !integer, parameter :: N_obs         = 2  ! actual nr. of observations 
  integer(IntAuc), parameter :: N_Obs_Max_Mie      = 2 ! max. nr. of obs. Mie
  integer(IntAuc), parameter :: N_Obs_Max_Rayleigh = 2 ! max. nr. of obs. Rayl.
  ! REMARK: the measurements may also be split in 2 sets if there
  ! is for example a strong wind shear (or other phenomenon that 
  ! would justify it). So multiple observations with the same
  ! observation type are possible. Therefore nr_of_obs_types
  ! is not identical to the max nr. of allowed observations!
  
  ! this is no longer a parameter, but the actual value, read from the L1B file
  integer :: N_meas

  !  #]
contains
  !---------------------------------------------------------
  subroutine Find_Matching_AMD_Profile(AMD, WorkingData, AMD_Matchup_Params, &
       L2B_AMD_Collocations, which_AMD_DSR, iBRC, error_flag)
    !  #[ select the proper AMD struct for this location
    TYPE(AMD_BRC_DataType), dimension(:), pointer :: AMD            ! input
    TYPE(WorkingDataType), intent(in)         :: WorkingData        ! input
    TYPE(AMD_Matchup_Params_type), intent(in) :: AMD_Matchup_Params ! input
    TYPE(L2B_AMD_Collocations_Type), &
              dimension(:), pointer ::L2B_AMD_Collocations     ! pcd result
    integer,           intent(out)        :: which_AMD_DSR ! result    
    integer,           intent(in)         :: iBRC          ! temp. input
    integer,           intent(out)        :: error_flag    ! output
    
    ! method:
    ! =>take the lat,lon values for the midpoint of the BRC
    ! =>walk along all available NWP profiles to see which one is nearest
    ! =>make sure the timedifference is within the allowed range

    ! REMARK: the iBRC parameter is included for compatibility with the
    ! previous code versions. As soon as a proper matching is implemented
    ! it should be removed from this interface.

    ! This routine combines AMD and L1B data, so hard to tell in which
    ! directory it should be. My preference is to keep it here. JK.

    ! local variables
    ! variables to hold lat, lon, datetime for the L1B data
    real(r8_)          :: L1B_lat1,    L1B_lon1    ! in degrees
    real(r8_)          :: L1B_lat2,    L1B_lon2    ! in degrees
    real(r8_)          :: L1B_lat_mid, L1B_lon_mid ! in degrees
    TYPE(DateTimeType) :: L1B_datetime1      ! struct
    TYPE(DateTimeType) :: L1B_datetime2      ! struct
    TYPE(DateTimeType) :: L1B_datetime_mid   ! struct
    
    ! variables to hold lat, lon, datetime for the AMD data
    real(r8_)          :: AMD_lat, AMD_lon   ! in degrees
    TYPE(DateTimeType) :: AMD_datetime       ! struct

    ! other variables
    integer        :: amd_profile
    integer        :: nr_of_AMD_profiles
    integer        :: closest_AMD_profile
    real(r8_)      :: distance, closest_distance_all
    real(r8_)      :: closest_distance_in_timewindow
    real(r8_)      :: time_diff, closest_time_diff_all
    real(r8_)      :: closest_time_diff_in_range

    !character(len=maxlen_DateTime_KVT) :: txt_DateTime_KVT
    character(len=50)                  :: txt_val1, txt_val2
    character(len=1024)                :: txt
    character(len=20)                  :: txt_BRC

    ! init
    error_flag = no_error
    which_AMD_DSR = -1

    ! init the pcd outputs
    L2B_AMD_Collocations(:)%L2B_AMD_Collocation_QC = L2B_AMD_Coll_Failed
    L2B_AMD_Collocations(:)%L2B_AMD_Collocation_QC_Flags = 0_IntAuc
    !L2B_AMD_Coll_Failed

    IF ( (AMD_Matchup_Params%Matchup_Method .ne. Matchup_Dummy) .and. &
         (AMD_Matchup_Params%Matchup_Method .ne. Matchup_Nearest_Neighbour)) &
         THEN
       error_flag = error_reading_file
       call logmsg(log_error,"Error in Find_Matching_AMD_Profile: "//&
            "unknown or undefined Matchup_Method method. ")
       ! the selected AMD profile
       which_AMD_DSR = -1
       ! the pcd outputs
       L2B_AMD_Collocations(:)%L2B_AMD_Collocation_QC = L2B_AMD_Coll_Failed
       L2B_AMD_Collocations(:)%L2B_AMD_Collocation_QC_Flags = 0_IntAuc
       ! this condition is an error in the aux_par input file so
       ! no flags are defined in L2B_AMD_Collocation_QC_Flags to signal this
       return
    END IF

    IF (AMD_Matchup_Params%Matchup_Method .eq. Matchup_Dummy) THEN
       nr_of_AMD_profiles = size(AMD)

       ! the selected AMD profile, dummy result
       IF (iBRC .le. nr_of_AMD_profiles) THEN
          which_AMD_DSR = iBRC

          ! the pcd outputs
          L2B_AMD_Collocations(:)%L2B_AMD_Collocation_QC = L2B_AMD_Coll_OK
          L2B_AMD_Collocations(:)%L2B_AMD_Collocation_QC_Flags = 0_IntAuc
          ! a test print
          print *,"Find_Matching_AMD_Profile dummy result:"
          print *,"which_AMD_DSR = ",which_AMD_DSR
          return
       ELSE
          error_flag = error_reading_file
          call logmsg(log_error,"Error in Find_Matching_AMD_Profile: "//&
               "nr of BRCs exceeds nr of available AMD profiles. "//&
               "No match found for dummy match method.")
          ! the selected AMD profile
          which_AMD_DSR = -1
          ! the pcd outputs
          L2B_AMD_Collocations(:)%L2B_AMD_Collocation_QC = L2B_AMD_Coll_Failed
          L2B_AMD_Collocations(:)%L2B_AMD_Collocation_QC_Flags = 0_IntAuc
          ! this condition is an error in the aux_par input file so
          ! no flags are defined in L2B_AMD_Collocation_QC_Flags to signal this
          return
       END IF
    END IF

    IF (AMD_Matchup_Params%Matchup_Method .eq. Matchup_Nearest_Neighbour) THEN
       ! get the location+time for this BRC

       ! the Mie and Rayleigh channel might differ slightly in location
       ! when they have different rangebin definitions, but the difference
       ! is so small, combined to the BRC size, that I will ignore this
       ! for now. Also locations for upper and lower rangebins differs a bit,
       ! but again I will ignore this for now,
       ! so just take a copy of the lat,lon values for a
       ! Mie rangebin halfway the profile
       N_Meas = WorkingData%nr_of_measurements

       L1B_lat1     = WorkingData%MieMeas(1     )%RangeGate(12)%lat
       L1B_lat2     = WorkingData%MieMeas(N_Meas)%RangeGate(12)%lat
       L1B_lon1     = WorkingData%MieMeas(1     )%RangeGate(12)%lon
       L1B_lon2     = WorkingData%MieMeas(N_Meas)%RangeGate(12)%lon

       ! the Mie and Rayleigh channel have the same datetime
       ! so just take a copy of the Mie datetime
       L1B_datetime1 = WorkingData%MieMeas(1     )%datetime
       L1B_datetime2 = WorkingData%MieMeas(N_Meas)%datetime
       
       call get_midpoint(L1B_lat1,L1B_lon1,L1B_lat2,L1B_lon2,& ! inputs
                         L1B_lat_mid,L1B_lon_mid)              ! outputs
       if (missing_real(L1B_lat_mid) .or. missing_real(L1B_lon_mid)) then
          error_flag = error_programming
          call logmsg(log_error,"Error in Find_Matching_AMD_Profile: "//&
               "Could not find midpoint between 2 lat-lon pairs. "//&
               "Probably this poitns to a programming error.")
          which_AMD_DSR = -1
          return          
       end if

       L1B_datetime_mid = get_averaged_datetime(L1B_datetime1,L1B_datetime2)

       ! some test prints
       !print *,"================= some test prints ===================="
       !print *,"iBRC = ",iBRC
       !print *,"L1B_lat1,L1B_lat2,L1B_lon1,L1B_lon2 = ",&
       !     L1B_lat1,L1B_lat2,L1B_lon1,L1B_lon2
       !txt_DateTime_KVT = Construct_KVT_DateTime(L1B_datetime1)
       !print *,"L1B_datetime1 = ",trim(txt_DateTime_KVT)," [UTC]"
       !txt_DateTime_KVT = Construct_KVT_DateTime(L1B_datetime2)
       !print *,"L1B_datetime2 = ",trim(txt_DateTime_KVT)," [UTC]"
       !txt_DateTime_KVT = Construct_KVT_DateTime(L1B_datetime_mid)
       !print *,"L1B_datetime_mid = ",trim(txt_DateTime_KVT)," [UTC]"
       !print *,"L1B_lat_mid,L1B_lon_mid = ",L1B_lat_mid,L1B_lon_mid
       !print *,"================= end of test prints =================="
       
       ! initialise
       closest_AMD_profile = -1
       closest_distance_all            = -1._r8_
       closest_distance_in_timewindow  = -1._r8_
       closest_time_diff_all           = -1._r8_
       closest_time_diff_in_range      = -1._r8_

       AMD_loop: do amd_profile=1,size(AMD)
          !  #[

          ! no missing checks on AMD_lat, AMD_lon, AMD_datetime
          ! have been added yet (do we need them?)
          
          ! calculate the time difference between the BRC and this AMD profile
          AMD_datetime = AMD(amd_profile)%Geoloc_ADS_off_nadir%DateTime ! [UTC]
          time_diff = get_time_difference(AMD_datetime,L1B_datetime_mid)
          
          ! calculate the distance between the BRC and this AMD profile
          AMD_lat = AMD(amd_profile)%Geoloc_ADS_off_nadir%Latitude * &
                    1.e-6_r8_ ! [convert 10^-6 degN to degN] 
          AMD_lon = AMD(amd_profile)%Geoloc_ADS_off_nadir%Longitude * &
                    1.e-6_r8_ ! [convert 10^-6 degE to degE]
          distance = get_distance(L1B_lat_mid,L1B_lon_mid,AMD_lat,AMD_lon)

          ! test prints
          !print *,"AMD profile nr. ",amd_profile,&
          !     " lat,lon: ",AMD_lat,AMD_lon," [deg]"
          !txt_DateTime_KVT = Construct_KVT_DateTime(AMD_datetime)
          !print *,"AMD datetime: ",trim(txt_DateTime_KVT)," [UTC]"
          !print *,"distance  = ",distance," [km]"
          !print *,"time_diff = ",time_diff," [s]"

          ! remember overal closest timestep for all AMD profiles
          ! usefull for composing a logmsg afterwards
          ! when collocation fails
          if ( (abs(time_diff) .lt. closest_time_diff_all) .or. &
               (closest_time_diff_all .lt. 0._r8_)              ) then
             closest_time_diff_all   = time_diff
          end if
          
          ! remember overal closest location for all AMD profiles
          ! usefull for composing a logmsg afterwards
          ! when collocation fails
          if ( (distance .lt. closest_distance_all) .or. &
               (closest_distance_all .lt. 0._r8_)        ) then
             closest_distance_all = distance
          end if

          ! now only consider AMD profiles inside the allowed range
          if ( distance .le. &
               AMD_Matchup_Params%Max_Allowed_Distance) then
             ! remember closest timestep for AMD profiles within range
             ! usefull for composing a logmsg afterwards
             ! when collocation fails
             if ( (abs(time_diff) .lt. closest_time_diff_in_range) .or. &
                  (closest_time_diff_in_range .lt. 0._r8_)              ) then
                closest_time_diff_in_range = time_diff
             end if
          end if

          ! now only consider AMD profiles inside the allowed timewindow
          if ( abs(time_diff) .le. &
               AMD_Matchup_Params%Max_Allowed_Time_Diff) then

             ! remember overal closest location inside timewindow
             ! usefull for composing a logmsg afterwards
             ! when collocation fails
             if ( (distance .lt. closest_distance_in_timewindow) .or. &
                  (closest_distance_in_timewindow .lt. 0._r8_)        ) then
                closest_distance_in_timewindow = distance
             end if

             ! remember the closest AMD profile in distance and time,
             ! but only if both conditions are met
             if ( distance .le. &
                  AMD_Matchup_Params%Max_Allowed_Distance) then
                !print *,"switching to AMD profile: ",amd_profile
                closest_AMD_profile = amd_profile
             end if
          end if
          !  #]
       end do AMD_loop

       print *,"Find_Matching_AMD_Profile result:"
       print *,"closest_AMD_profile = ",closest_AMD_profile
       !print *,"closest_distance_in_timewindow = ",&
       !         closest_distance_in_timewindow
       !print *,"closest_distance_all           = ",&
       !         closest_distance_all
       !print *,"closest_time_diff_in_range     = ",&
       !         closest_time_diff_in_range
       !print *,"closest_time_diff_all          = ",&
       !         closest_time_diff_all

       if (closest_AMD_profile .eq. -1) then
          ! compose a proper error message to describe what went wrong
          txt = ""
          print *,"AMD_Matchup_Params%Max_Allowed_Distance = ",&
                   AMD_Matchup_Params%Max_Allowed_Distance
          print *,"AMD_Matchup_Params%Max_Allowed_Time_Diff = ",&
                   AMD_Matchup_Params%Max_Allowed_Time_Diff

          if ( closest_distance_all .gt. &
               AMD_Matchup_Params%Max_Allowed_Distance ) then
             call SetFAdoFull_or_Missing(&
                           closest_distance_all,txt_val1,error_flag)
             call SetFAdoFull_or_Missing(&
                           1._r8_*AMD_Matchup_Params%Max_Allowed_Distance,&
                           txt_val2,error_flag)
             txt = trim(txt)//&
                  " Distance criterium failed. Closest found AMD profile"//&
                  " is located at a distance of "//trim(txt_val1)//&
                  " [km] while the maximum allowed distance is "//&
                  trim(txt_val2)//" [km]"
          end if

          if ( abs(closest_time_diff_all) .gt. &
               AMD_Matchup_Params%Max_Allowed_Time_Diff) then 
             call SetFAdoFull_or_Missing(&
                           closest_time_diff_all,txt_val1,error_flag)
             if (error_flag .ne. no_error) return
             call SetFAdoFull_or_Missing(&
                           1._r8_*AMD_Matchup_Params%Max_Allowed_Time_Diff,&
                           txt_val2,error_flag)
             if (error_flag .ne. no_error) return
             txt = trim(txt)//&
                  " Timing criterium failed. AMD profile with smallest"//&
                  " time difference is separated by "//trim(txt_val1)//&
                  " [s] while the maximum allowed separation is "//&
                  trim(txt_val2)//" [s]"
          end if

          if (len_trim(txt) .eq. 0) txt = "[not yet defined]"

          write(txt_BRC,"(i8)", err=999) iBRC

          !error_flag = error_reading_file
          ! the error flag is not set for this condition.
          ! When using for example a local area NWP model it might 
          ! happen that there are no nearby profiles. 
          ! in this case still this BRC needs to be added to the
          ! output file but without results and with the proper pcd flags set
          call logmsg(log_warn,"Warning in Find_Matching_AMD_Profile: "//&
               "Could not find matching AMD profile for BRC:"//&
               trim(txt_BRC)//". "//trim(txt),realacc_precision=12)
          ! the selected AMD profile          
          which_AMD_DSR = -1
          ! the pcd outputs. For the moment all measurements in this BRC
          ! are flagged alike. If it is needed, we could flag on the
          ! measurement level as well using these pcd elements
          L2B_AMD_Collocations(:)%L2B_AMD_Collocation_QC = L2B_AMD_Coll_Failed
          L2B_AMD_Collocations(:)%L2B_AMD_Collocation_QC_Flags = 0_IntAuc
          return
       end if
       ! the selected AMD profile
       which_AMD_DSR = closest_AMD_profile
       L2B_AMD_Collocations(:)%L2B_AMD_Collocation_QC = L2B_AMD_Coll_OK
       L2B_AMD_Collocations(:)%L2B_AMD_Collocation_QC_Flags = 0_IntAuc
    END IF

    return

999 print *,"Find_Matching_AMD_Profile:"
    print *,"Could not convert iBRC to a string, iBRC =  ",iBRC
    error_flag = error_writing_file
    return

  end subroutine Find_Matching_AMD_Profile
    !  #]
  !---------------------------------------------------------
end module Match_AMD_Module
