function [startTimeKnmi,nbProfile] = ...
    ConvertTxtScenarioToXML(DataDir,scenarioPath,mieflag,removemissing)
% CONVERTTXTSCENARIOTOXML convert an atmospheric scenario file in KNMI ASCII
%    format to the xml format required for input by the E2S.
%    Note that not all items can be converted, because of the limited number
%    of fields available in the E2S xml format.
%    Supported fields are:
%      hs, T, windcomp_u, windcomp_v, windcomp_w, beta_a, sigma_a, beta_c, sigma_c, beta_m

% Input arguments :
%
% - DATADIR is the name of the directory containing the ascii data 
%
% - MIEFLAG is a variable selecting the mie contribution from the knmi
%   ascii file. MIEFLAG can take the values : 'all' (aerosol + cloud
%   contribution), 'cloud' (only cloud contribution), 'aerosol' (only
%   aerosol contribution), 'none' (no mie contribution)
%
% - REMOVEMISSING is a flag to choose whether missing levels should be
%   exported to the conversion result, or just removed.
%   REMOVEMISSING can take the values 'true' and 'false'
%
%   Revision 1.1    6 Jan 2006  Marie-Laure Denneulin
%   Initial Revision
%
%   Revision 1.2    6 Jan 2006  Marie-Laure Denneulin
%
%   Revision 1.2    11 Oct 2007  Jos de Kloe
%   -added REMOVEMISSING flag, and force read permission for all users
%    on the output file.

% define the locations/versions of the E2S/L1B to be used
DefinePaths;

% handle the 4th parameter which might be missing,
% by setting it to the default value 'true'.
if nargin < 4, removemissing = 'true'; end;

% check whether the input directory exists
TestDataKNMI = exist(fullfile(DataPath,DataDir),'dir');
TestDataOther = exist(fullfile(DataPathNew,DataDir),'dir');
if TestDataKNMI == 7,
    pathDataLocal = DataPath;
elseif TestDataOther == 7,
    pathDataLocal = DataPathNew;
else fprintf('Error : the specified scenario does not exist\n');
end

% get the name of the *.asc file from this directory
DataNameFull = dir(fullfile(pathDataLocal,DataDir,'*.asc'));
DataNameFull = DataNameFull.name;

% open the *.asc file
fid=fopen(fullfile(pathDataLocal,DataDir,DataNameFull));

% skip 36 headerlines, and use ',' as fields separator
% then read 2 doubles from 1 line (why not integers?)
tmp = textscan(fid, '%f%f',1,'headerlines',36,'delimiter',',');
%tried this, and it works just as well
%tmp = textscan(fid, '%d8%d8',1,'headerlines',36,'delimiter',',');

% split the result in 2 separate variables
nbProfile = tmp{1};
nbHeightBin = tmp{2};

startTimeKnmi=[];

hlos_reported = 'false';
for kprofile= 1: nbProfile,
    % read 13 doubles from one text line
    tmp2{kprofile} = textscan(fid,'%f%f%f%f%f%f%f%f%f%f%f%f%f',1,'delimiter',',');

    % split the result in separate arrays
    startTimeKnmi(kprofile,1) = tmp2{kprofile}{8}; %YY
    startTimeKnmi(kprofile,2) = tmp2{kprofile}{9}; %MM
    startTimeKnmi(kprofile,3) = tmp2{kprofile}{10}; %DD
    startTimeKnmi(kprofile,4) = tmp2{kprofile}{11}; %h
    startTimeKnmi(kprofile,5) = tmp2{kprofile}{12}; %m
    startTimeKnmi(kprofile,6)= tmp2{kprofile}{13}; %s
    
    %  read 14 doubles from one text line
    tmp3{kprofile} = textscan(fid,'%f%f%f%f%f%f%f%f%f%f%f%f%f%f',nbHeightBin,'delimiter',',');
    hs = tmp3{kprofile}{1};
    T = tmp3{kprofile}{4};
    windcomp_u = tmp3{kprofile}{5};
    windcomp_v = tmp3{kprofile}{6};
    windcomp_w = tmp3{kprofile}{7};
    beta_a = tmp3{kprofile}{9};
    sigma_a = tmp3{kprofile}{10};
    beta_c = tmp3{kprofile}{11};
    sigma_c = tmp3{kprofile}{12};
    beta_m = tmp3{kprofile}{13};
    
    docNode = com.mathworks.xml.XMLUtils.createDocument('atmosphereProfileParameters');
    docRootNode = docNode.getDocumentElement;
    for i=1:nbHeightBin,

%        switch removemissing
%           case 'true'
%           case 'false'
%        end;

        % check for missing in the Temperature field
	% note that & acts on single elements, while && may act on a whole array
        if ( (T(i)>-99.1) & (T(i)<-98.9) & strcmp(removemissing,'true') )
           fprintf('skipping level %i due to missing value in temperature\n',i);
           continue; % step to next iteration in this for loop
        end;

        % check for missing in the windcomp_u field
        if ( (windcomp_u(i)>-999.1) && (windcomp_u(i)<-998.9) & strcmp(removemissing,'true') ) 
           fprintf('skipping level %i due to missing value in windcomp_u\n',i);
           continue;  % step to next iteration in this for loop
        end;

	% at this point we know windcomp_u was not missing.
        % The convention used in the KNMI ASCII format is that when u is filled
        % and v is missing, u should be interpreted as being hlos.
        use_u_as_hlos = 'false';
	if ( (windcomp_v(i)>-999.1) && (windcomp_v(i)<-998.9 ) )
           use_u_as_hlos = 'true';
           fprintf('debug: u=%f7.3\n',windcomp_u(i));
           fprintf('debug: v=%f7.3\n',windcomp_v(i));
           fprintf('debug: w=%f7.3\n',windcomp_w(i));
        end;

        % check for missing in the windcomp_w field
        if ( (windcomp_w(i)>-999.1) && (windcomp_w(i)<-998.9) & strcmp(removemissing,'true') ) 
           fprintf('skipping level %i due to missing value in windcomp_w\n',i);
           continue;  % step to next iteration in this for loop
        end;

        % check for missing in the beta_a field
        if ( (beta_a(i)>-1.01) && (beta_a(i)<-0.99) & strcmp(removemissing,'true') ) 
           fprintf('skipping level %i due to missing value in beta_a\n',i);
           continue;  % step to next iteration in this for loop
        end;

        % check for missing in the sigma_a field
        if ( (sigma_a(i)>-1.01) && (sigma_a(i)<-0.99) & strcmp(removemissing,'true') ) 
           fprintf('skipping level %i due to missing value in sigma_a\n',i);
           continue;  % step to next iteration in this for loop
        end;

        % check for missing in the beta_c field
        if ( (beta_c(i)>-1.01) && (beta_c(i)<-0.99) & strcmp(removemissing,'true') ) 
           fprintf('skipping level %i due to missing value in beta_c\n',i);
           continue;  % step to next iteration in this for loop
        end;

        % check for missing in the sigma_c field
        if ( (sigma_c(i)>-1.01) && (sigma_c(i)<-0.99) & strcmp(removemissing,'true') ) 
           fprintf('skipping level %i due to missing value in sigma_c\n',i);
           continue;  % step to next iteration in this for loop
        end;

        % check for missing in the beta_m field
        if ( (beta_m(i)>-1.01) && (beta_m(i)<-0.99) & strcmp(removemissing,'true') ) 
           fprintf('skipping level %i due to missing value in beta_m\n',i);
           continue;  % step to next iteration in this for loop
        end;

        % if we still have missing beta_m at this stage, molExtinction cannot
        % be calculated, so set that one to missing too
        if ( (beta_m(i)>-1.01) && (beta_m(i)<-0.99) )
           molExtinction(i)=-1.0;
        else
           % calculate molecular extinction from molecular backscatter
           molExtinction(i)=beta_m(i)*(8*pi/3);
        end;

        % calculate the total particle extinction and backscatter
        % the summation is not possible if one of the two parts is missing,
        % so check for that. The copying of just aerosol or just cloud should be fine,
        % when they are missing, the missing value is just copied along
        switch mieflag
            case 'all'
                if ( ( (sigma_a(i)>-1.01) && (sigma_a(i)<-0.99) ) || ...
                     ( (sigma_c(i)>-1.01) && (sigma_c(i)<-0.99) )  )
                   sigma_sum = -1.0;
                   beta_sum  = -1.0;
                else
                   sigma_sum = sigma_a(i)+sigma_c(i);
                   beta_sum  = beta_a(i)+beta_c(i);
                end;
            case 'aerosol'
                sigma_sum = sigma_a(i);
                beta_sum  = beta_a(i);
            case 'cloud'
                sigma_sum = sigma_c(i);
                beta_sum  = beta_c(i);
            case 'none'
                sigma_sum = 0;
                beta_sum  = 0;
        end;

        % convert the numbers to strings, and do unit conversions
        % needed to store them with proper precision in the xml file
        txt_hs            = num2str(hs(i)/1000);
        txt_molExtinction = num2str(molExtinction(i)*1e6,'%10.3f');
        txt_beta_m        = num2str(beta_m(i)*1e6,'%10.3f');
        txt_sigma_sum     = num2str(sigma_sum*1e6,'%10.3f');
        txt_beta_sum      = num2str(beta_sum*1e6,'%10.3f');

        % temperature is an exception, because the unit conversion is a subtraction
        % this messes up the missing indicator, so make sure missing temperature
        % are still -99.0 even after the conversion
        if ( (T(i)>-99.1) & (T(i)<-98.9) )
           txt_T = num2str(-99.0)
        else
           txt_T = num2str(T(i)-273.15);
        end;

        % now handle the conversion to hlos.
        % When u if filled and v is missing, use u as hlos
        % otherwise a proper hlos calculation is needed (not yet implemented)

        if ( strcmp(use_u_as_hlos,'true') & strcmp(hlos_reported,'false') )
           fprintf('using u as hlos\n');
           hlos_reported = 'true';
        end;

        if strcmp(use_u_as_hlos,'true')
           txt_hlos = num2str(windcomp_u(i));
        else
           txt_hlos = num2str(windcomp_u(i));
% disable this check for now, untill we have found a proper solution for this
%           txt_hlos = 0.
%           error('proper conversion of u,v,w to hlos is not yet implemented')
        end;

        % ok, all missing handling, calculation and unit conversions are done now,
        % the remained of this loop does the xml handling

        % create the profile xml elements
        thisElement = docNode.createElement('profileData');

        alt = docNode.createElement('altitude');
        alt.setAttribute('unit','kilometers');
        hlos = docNode.createElement('hlosWindVelocity');
        hlos.setAttribute('unit','meters/second');
        temp = docNode.createElement('temperature');
        temp.setAttribute('unit','celcius');
        molext = docNode.createElement('molExtinction');
        molext.setAttribute('unit','1E-6/meter');
        molback = docNode.createElement('molBackscatter');
        molback.setAttribute('unit','1E-6/meter/steradian');
        aerext = docNode.createElement('aerExtinction');
        aerext.setAttribute('unit','1E-6/meter');
        aerback = docNode.createElement('aerBackscatter');
        aerback.setAttribute('unit','1E-6/meter/steradian');

        alt.appendChild(docNode.createTextNode(txt_hs));
        hlos.appendChild(docNode.createTextNode(txt_hlos));
        temp.appendChild(docNode.createTextNode(txt_T));
        molext.appendChild(docNode.createTextNode(txt_molExtinction));
        molback.appendChild(docNode.createTextNode(txt_beta_m));
        aerext.appendChild(docNode.createTextNode(txt_sigma_sum));
        aerback.appendChild(docNode.createTextNode(txt_beta_sum));

        thisElement.appendChild(alt);
        thisElement.appendChild(hlos);
        thisElement.appendChild(temp);
        thisElement.appendChild(molext);
        thisElement.appendChild(molback);
        thisElement.appendChild(aerext);
        thisElement.appendChild(aerback);

        docRootNode.appendChild(thisElement);
    end %for nbHeightBin

    Flag1=docNode.createElement('interpolateTemperatureFlag');
    Flag2=docNode.createElement('interpolateHlosWindVelocityFlag');
    Flag3=docNode.createElement('interpolateMolExtinctionFlag');
    Flag4=docNode.createElement('interpolateMolBackscatterFlag');
    Flag5=docNode.createElement('interpolateAerExtinctionFlag');
    Flag6=docNode.createElement('interpolateAerBackscatterFlag');

    Flag1.appendChild(docNode.createTextNode('true'));
    Flag2.appendChild(docNode.createTextNode('true'));
    Flag3.appendChild(docNode.createTextNode('true'));
    Flag4.appendChild(docNode.createTextNode('true'));
    Flag5.appendChild(docNode.createTextNode('true'));
    Flag6.appendChild(docNode.createTextNode('true'));

    docRootNode.appendChild(Flag1);
    docRootNode.appendChild(Flag2);
    docRootNode.appendChild(Flag3);
    docRootNode.appendChild(Flag4);
    docRootNode.appendChild(Flag5);
    docRootNode.appendChild(Flag6);

    xmlFileName = fullfile(scenarioPath,sprintf('Segment%02d_atmosphereProfileParameters.xml',kprofile));
    xmlwrite(xmlFileName,docNode);

    % change the read permission, to allow other users to read the file as well
    status = system(sprintf('%s %s','chmod a+r ',xmlFileName));
end %for nbProfile
fclose(fid);

