function WriteE2SInput(ficname,surf,prof,header)
%
% FUNCTION WriteE2SInput(FICNAME,SURF,PROF,[HEADER])
%
% Routine writing a KNMI type of E2S Input data.
% The input surface data and profiles are contained
% in structures SURF and PROF respectively. The HEADER
% is optional. If it is given, it must contain 36 lines
% of comments. If not given, a default header is written
% in the output file FICNAME.
%
% A. Dabas, Météo-France, 6 December 2005
%

% Open output file
fid=fopen(ficname,'wt');

% Write header
if nargin==4,
	for k=1:length(header),fprintf(fid,'%s\n',header{k});end
elseif nargin==3,
	fprintf(fid,'# file format version number: 0.2\n');
	fprintf(fid,'# scene description: ');
	fprintf(fid,'# creation date and time: %s\n',datestr(now));
	[status,author]=system('whoami');
	fprintf(fid,'# creator: %s',author);
	fprintf(fid,'# purpose: line1\n');
	fprintf(fid,'# purpose: line2\n');
	fprintf(fid,'# purpose: line3\n');
	fprintf(fid,'#\n');
	fprintf(fid,'# line definition 1: nr of profiles and nr of height steps per profile\n');
	fprintf(fid,'# nr. of items:   2\n');
	fprintf(fid,'# items: nprofiles, nsteps\n'); 
	fprintf(fid,'# unit: nr, nr\n'); 
	fprintf(fid,'# format: (i3,1x,i3)\n');
	fprintf(fid,'# missing values:  -1,  -1 \n');
	fprintf(fid,'# separator character: '',''\n');
	fprintf(fid,'#\n');
	fprintf(fid,'# nr of profiles in this track: %d\n',length(prof));
	fprintf(fid,'#\n');
	fprintf(fid,'# line definition 2: surface and ancillary information\n');
	fprintf(fid,'# nr. of items:  13\n');
	fprintf(fid,'# items: Lat, Lon, orogr, beta_g, B, albedo, mask, CCYY, MM, DD, h, m, s\n'); 
	fprintf(fid,'# unit: 1 deg, 1 deg, 1 m, 1/sr, 1 W/m^2, fraction, flag, 1 year, 1 month, 1 day, 1 h (utc), 1 min, 1 sec\n'); 
	fprintf(fid,'# format: (f8.2,1x,f8.2,1x,f6.1,1x,es10.3e2,1x,f6.3,1x,f6.3,1x,i1,1x,i4,1x,i2,1x,i2,1x,i2,1x,i2,1x,f5.2)\n');
	fprintf(fid,'# missing values:   999.00,   999.00, -999.0, -1.000E+00, -1.000, -1.000, 9, -999, 99, 99, 99, 99, -9.99\n'); 
	fprintf(fid,'# separator character: ','\n');
	fprintf(fid,'#\n');
	fprintf(fid,'# nr of bins in this profile:  %d\n',length(prof(1).zmin));
	fprintf(fid,'#\n');
	fprintf(fid,'# line definition 3: profile definition\n');
	fprintf(fid,'# nr. of items:  14\n');
	fprintf(fid,'# items: hs, he, p, T, u, v, w, w_p, beta_a, sigma_a, beta_c, sigma_c, beta_m, beta_pr\n'); 
	fprintf(fid,'# unit: 1 m, 1 m, 1 hPa, 1 K, 1 m/s, 1 m/s, 1 m/s, 1 m/s, 1/(m.sr), 1/m, 1/(m.sr), 1/m, 1/(m.sr), 1/(m.sr)\n'); 
	fprintf(fid,'# format: (i6,1x,i6,1x,f6.1,1x,f5.1,1x,f5.1,1x,f5.1,1x,f5.1,1x,f5.1,1x,es10.3e2,1x,es10.3e2,1x,es10.3e2,1x,es10.3e2,1x,es10.3e2,1x,es10.3e2)\n');
	fprintf(fid,'# missing values:  -9999,  -9999, -999.0, -99.0, 999.0, 999.0, 999.0, 999.0, -1.000E+00, -1.000E+00, -1.000E+00, -1.000E+00, -1.000E+00, -1.000E+00 \n');
	fprintf(fid,'# separator character: ','\n');
	fprintf(fid,'#\n');
else
	error('WriteE2SInput: Wrong number of input arguments.');
end

% Write number of profiles and number of levels
nprof=length(prof);
nlev=length(prof(1).zmin);
fprintf(fid,'%3d,%3d\n',nprof,nlev);

% Loop on profiles
for kprof=1:nprof,

	% Write surface data
	fprintf(fid,'%8.2f,%8.2f,%6.1f,%10.3E,%6.3f,%6.3f,%1d,%4d,%2d,%2d,%2d,%2d,%5.2f\n',surf(kprof).lat,surf(kprof).lon,surf(kprof).oro,surf(kprof).beta_g,...
	surf(kprof).B,surf(kprof).albedo,surf(kprof).mask,surf(kprof).year,surf(kprof).month,surf(kprof).day,surf(kprof).hr,surf(kprof).min,surf(kprof).sec);

	% Write profile data
	for klev=1:nlev,
	fprintf(fid,'%6d,%6d,%6.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%10.3E,%10.3E,%10.3E,%10.3E,%10.3E,%10.3E\n',...
		prof(kprof).zmin(klev),prof(kprof).zmax(klev),prof(kprof).P(klev),prof(kprof).T(klev),prof(kprof).u(klev),prof(kprof).v(klev),prof(kprof).w(klev),prof(kprof).wp(klev),...
		prof(kprof).beta_a(klev),prof(kprof).sigma_a(klev),prof(kprof).beta_c(klev),prof(kprof).sigma_c(klev),prof(kprof).beta_m(klev),prof(kprof).beta_pr(klev));
	end
end

% Close output file
fclose(fid);
