function l2b=RBCorrection(varargin)
% Rayleigh-Brillouin correction scheme
%   
%   L2B = RBCORRECTION(JobOrder,[ScenarioName]) or L2B = RBCORRECTION(l1bData,AuxMetData,[LookUpTable])
%
%   Apply the Ryaleigh-Brillouin correction scheme to the L1B data contained in the L1B struture l1BData or produced by the job
%   order JobOrder. The atmospheric conditions are contained in the AuxMetData structure or the AuxMetData file contained in the
%   scenario ScenarioName. If the variable ScenarioName is missing, it is assumed that ScenarioName = JobOrder. The
%   Rayleigh-Brillouin look-up table is contained in the LookUpTable structure. If this structure is not among the input
%   variables, the default look-up table is read.
%
%   Note that at present time AuxMetData are extracted from SLS_Input_brc_nnn.m files produced by the E2S. The scattering ratio
%   information used for step 2 of the correction is computed from these data.
%
%   Alain DABAS, Meteo-France, 20 January 2006
%


%-------------------------------------------------------------------------------------------------------------------------------
% Parse input arguments.
%-------------------------------------------------------------------------------------------------------------------------------
if isequal(nargin,1)&ischar(varargin{1}),
    JobOrder=varargin{1};
    ScnearioName=JobOrder;
elseif isequal(nargin,2),
    if ischar(varargin{1})&ischar(varargin{2}),
        JobOrder=varargin{1};
        ScenarioName=varargin{2};
    else
        l1b=varargin{1};if ~isstruct(l1b), error('RBCorrection : Input 1 must be a L1B data structure.'); end
        brc=varargin{2};if ~isstruct(brc), error('RBCorrection : Input 2 must be an AuxMet data structure.'); end
    end
elseif isequal(nargin,3),
    l1b=varargin{1};if ~isstruct(l1b), error('RBCorrection : Input 1 must be a L1B data structure.'); end
    brc=varargin{2};if ~isstruct(brc), error('RBCorrection : Input 2 must be an AuxMet data structure.'); end
    rbtables=varargin{3};if ~isstruct(rbtables), error('RBCorrection : Input 3 must be a ILIAD structure.'); end
else
    error('RBCorrection : Wrong number or type of input variables.');
end
    

%-------------------------------------------------------------------------------------------------------------------------------
% Read L1B data file if necessary and organize useful l1b information in 1D or 2D arrays.
%-------------------------------------------------------------------------------------------------------------------------------
if ~isequal(exist('l1b'),1),
    l1b=ReadL1BData(JobOrder);
end

nobs=length(l1b.GeoDSR);
for kobs=1:nobs, 
    
    nbins=length(l1b.GeoDSR(kobs).ObsGeolocation.RayGeolocation)-1;
    SatLosVel(kobs)=l1b.GeoDSR(kobs).ObsGeolocation.AOCSLosVelocity;
    Elevation(kobs)=pi*l1b.GeoDSR(kobs).ObsGeolocation.DemGeolocation.TopoElevDemIntersec/180;
    
    for kbin=1:nbins,
        altmax(kobs,kbin) = l1b.GeoDSR(kobs).ObsGeolocation.RayGeolocation(kbin  ).AltHeightBin;
        altmin(kobs,kbin) = l1b.GeoDSR(kobs).ObsGeolocation.RayGeolocation(kbin+1).AltHeightBin;
        RayWnd(kobs,kbin) = l1b.WvDSR(kobs).ObsWindProfile.ListRayAltBinWind(kbin).WindVelocity;
        USRayA(kobs,kbin) = l1b.UsDSR(kobs).ObsUsefulSignal.ListRayAltBinUS(kbin).UsefulSignalChanA;
        USRayB(kobs,kbin) = l1b.UsDSR(kobs).ObsUsefulSignal.ListRayAltBinUS(kbin).UsefulSignalChanB;
    end % End loop on height bins
    
end % End loop on observations

%-------------------------------------------------------------------------------------------------------------------------------
% Open AuxMetData file and average useful information over the measurements of the BRCs. Keep the vertical resolution of
% AuxMetData.
%-------------------------------------------------------------------------------------------------------------------------------
if ~isequal(exist('brc'),1),
    brc=ReadAuxMet(JobOrder);
end

nbrc=length(brc);
numAuxMetAlt=length(brc(1).Atmosphere(1).Altitude);

% Check nobs = nbrc. In current version of AuxMetData, this is an obligation.
if ~isequal(nobs,nbrc),
    error('RBCorrection : Number of L1B obs does not match number of BRCs in AuxMetData');
end

altitude=zeros(nbrc,numAuxMetAlt);
temperature=zeros(nbrc,numAuxMetAlt);
Miebackscatter=zeros(nbrc,numAuxMetAlt);
Raybackscatter=zeros(nbrc,numAuxMetAlt);
ScatRatio=zeros(nbrc,numAuxMetAlt);
for kbrc=1:nbrc,
    altitude(kbrc,:)=mean(vertcat(brc(kbrc).Atmosphere(:).Altitude));
    temperature(kbrc,:)=mean(vertcat(brc(kbrc).Atmosphere(:).Temperature));
    Miebackscatter(kbrc,:)=mean(vertcat(brc(kbrc).Atmosphere(:).MieBackScattCoef));
    Raybackscatter(kbrc,:)=mean(vertcat(brc(kbrc).Atmosphere(:).RayBackScattCoef));
end
ScatRatio=1+Miebackscatter./Raybackscatter;

%-------------------------------------------------------------------------------------------------------------------------------
% Open RB look-up table if necessary
%-------------------------------------------------------------------------------------------------------------------------------
if ~isequal(exist('rbtables'),1),,
    rbtables=load('./AuxData/ILIADTAbles.mat');
end

% Check rbtables not empty
if isempty(rbtables),
    error('RBCorrection : Cannot find ILIAD Tables.');
end

%-------------------------------------------------------------------------------------------------------------------------------
% ILIAD Correction 
%-------------------------------------------------------------------------------------------------------------------------------

% Compute measured responses from useful Rayleigh signals A & B
RayResp=-(USRayA-USRayB)./(USRayA+USRayB);

% Reserve memory for arrays
Scatbin=zeros(nobs,nbins);
Tbin=zeros(nobs,nbins);
Pbin=zeros(nobs,nbins);
Doppler0=zeros(nobs,nbins);
Doppler1=zeros(nobs,nbins);
Doppler2=zeros(nobs,nbins);
dfdR=zeros(nobs,nbins);
dfdT=zeros(nobs,nbins);
dfdP=zeros(nobs,nbins);
dfdrho=zeros(nobs,nbins);

% Start the correction
for kobs=1:nobs,
    
    for kbin=1:nbins,
    
        % Interpolate temperature profile of the BRC to the mid-altitude of height bins. For pressure, we take the constant value
        % P=1000hPa for all height bins. This is because there is presently no pressure indication in AuxMetData as there is no
        % pressure effect in the current versio  of the E2S
        Scatbin(kobs,kbin)=interp1(altitude(kobs,:),ScatRatio(kobs,:),0.5*altmax(kobs,kbin)+0.5*altmin(kobs,kbin),'spline');
        Tbin(kobs,kbin)=interp1(altitude(kobs,:),temperature(kobs,:),0.5*altmax(kobs,kbin)+0.5*altmin(kobs,kbin),'spline');
        Pbin(kobs,kbin)=1000;

        % Start of step 1 of the correction consisting in linearly interpolating the Doppler shifts in the LUT tables. For this,
        % we first find the indices of the largest table pressure, temperature and response smaller than the actual pressure,
        % temperature and response.
        indexT=max(find(rbtables.T <= Tbin(kobs,kbin)));
        indexR=max(find(rbtables.R <= RayResp(kobs,kbin)));
        indexP=max(find(rbtables.P <= Pbin(kobs,kbin)));

        % Doppler shift in LUT at indices indexP, indexT, indexR
        Doppler0(kobs,kbin)=rbtables.LUT(indexP,indexT,indexR);

        % Compute first order derivatives of Doppler shift wrt to P, R and T.
        dfdP(kobs,kbin)=(rbtables.LUT(indexP+1,indexT,indexR)-Doppler0(kobs,kbin))/(rbtables.P(indexP+1)-rbtables.P(indexP));
        dfdR(kobs,kbin)=(rbtables.LUT(indexP,indexT,indexR+1)-Doppler0(kobs,kbin))/(rbtables.R(indexR+1)-rbtables.R(indexR));
        dfdT(kobs,kbin)=(rbtables.LUT(indexP,indexT+1,indexR)-Doppler0(kobs,kbin))/(rbtables.T(indexT+1)-rbtables.T(indexT));

        % Compute first estimate of Doppler shift
        Doppler1(kobs,kbin)=Doppler0(kobs,kbin)+...
            dfdR(kobs,kbin).*(RayResp(kobs,kbin)-rbtables.R(indexR))+...
            dfdT(kobs,kbin).*(Tbin(kobs,kbin)-rbtables.T(indexT))+...
            dfdP(kobs,kbin).*(Pbin(kobs,kbin)-rbtables.P(indexP));

        % Interpolate Na and Nb at frequencies in Doppler1
        Na=interp1(rbtables.fdoppler,squeeze(rbtables.Na(indexP,indexT,:)),Doppler1(kobs,kbin),'linear');
        Nb=interp1(rbtables.fdoppler,squeeze(rbtables.Nb(indexP,indexT,:)),Doppler1(kobs,kbin),'linear');
        Fpa=interp1(rbtables.fFP,rbtables.FPa,Doppler1(kobs,kbin),'linear');
        Fpb=interp1(rbtables.fFP,rbtables.FPb,Doppler1(kobs,kbin),'linear');
        
        % Compute dRdrho
        dRdrho(kobs,kbin)=((Fpa-Fpb)-RayResp(kobs,kbin)*(Fpa+Fpb))/(Na+Nb+(Scatbin(kobs,kbin)-1)*(Fpa+Fpb));
        
        % Apply correction for Mie contamination
        Doppler2(kobs,kbin)=Doppler1(kobs,kbin)-(Scatbin(kobs,kbin)-1)*dfdR(kobs,kbin)*dRdrho(kobs,kbin);
    
    end
end

% Compute final HLOS wind
l2b.Wind1=(-0.1775*Doppler1-repmat(SatLosVel',1,nbins))./cos(repmat(Elevation',1,nbins));
l2b.Wind2=(-0.1775*Doppler2-repmat(SatLosVel',1,nbins))./cos(repmat(Elevation',1,nbins));

% Store useful information to output structure l2b
l2b.altmin=altmin;
l2b.altmax=altmax;
l2b.Scatbin=Scatbin;
l2b.Tbin=Tbin;
l2b.Pbin=Pbin;
l2b.dfdT=dfdT;
l2b.dfdP=dfdP;
l2b.dfdR=dfdR;
l2b.dRdrho=dRdrho;

    
