%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% ReadL2BLoop.m              version 1.3
%
% This program generates plots of E2S data, L1B data, and L2B data
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% History
%
% Paul Poli, METEO-FRANCE     April 2007    Original code in final form
% Paul Poli, METEO-FRANCE     July 2007     Mods following ESA comments:
%                    - in 2D plots:
%                      - indicate range of selected BRCs
%                    - fix legend for the result of the L1B_Input_Screening
%                    - in 1D HLOS wind plots:
%                      - indicate NO DATA in legend if there are no retrieval
%                      - indicate number of out-of-range values
%                           only it is when non-zero (i.e. remove 'excld:0)
%                    - ability to generate plots for L1B and/or L2B and/or E2S
%                    - addition of a configuration file:
%                      - to select particular plots
%                      - to select a range of BRCs
% Paul Poli, METEO-FRANCE     October 2007 Mods following ESA comments:
%                    - allow for a different numbers of BRCs for L1B and L2B
%                    - addition of error bars for the L1B Mie and Rayleigh data
%                    - fixing definition of scenario name
% Paul Poli, METEO-FRANCE     December 2007 Add plot of lat/lon tracks 
%                      for L1B observations and L2B observations
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Pre-requisites: YOU NEED TO HAVE LOADED AT LEAST ONE THE FOLLOWING IN MEMORY:
%   e2sData=ReadE2SData(e2s_instrumentdata_path);
%   l1bData=ReadBinaryData([pathfic l1bfic],'ALL');
%   l2bData=ReadBinaryData([pathfic l2bfic],'ALL');
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if (exist('scenario'))
	plotconfig=Set_PlotConfig(scenario);
else
	plotconfig=Set_PlotConfig;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                              CORE SETTINGS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% core actions to perform
trsfdatal2b=true;    % transfer L2B data into arrays
trsfdatal1b=true;    % transfer L1B data into arrays
trsfdatae2s=true;    % transfer E2S data into arrays
check_timestamps=true; 
check_latlonstamps=true; 
plotdata=true;

% L2B classification types
icloudy=1;
iclear=2;
chproftype=['Cloudy';'Clear '];

% plotting symbols and colors
colproftype=['r','b'];
miesymproftype='d';
raysymproftype='o';
e2ssym='-';
e2scol='k';

ibrcrange=eval(plotconfig.BRCrange_To_Plot);
NbBRC_Plot=length(ibrcrange);

NbBIGPlot=1;
ylab_alt='Altitude [km]';
ylab_bin='25 minus Height Bin #';

inputscreeningflag{1}='Bin OK';
inputscreeningflag{2}='Bin_Invalid';
inputscreeningflag{3}='SNR_A';
inputscreeningflag{4}='SNR_B';

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                           SEE WHAT'S IN MEMORY
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

l2bpresent=exist('l2bData');
l1bpresent=exist('l1bData');
e2spresent=exist('e2sData');

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%               TRANSFER L2B DATA INTO ARRAYS
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if (trsfdatal2b & l2bpresent)
N_meas=l2bData.Nmax;
NbBRC_2b=l2bData.nbDSR;

mieinpscr2b=[];
miewind2b=[];
altmie2b=[];
miehloserror2b=[];
mapray2b=[];
rayinpscr2b=[];
raywind2b=[];
altray2b=[];
rayhloserror2b=[];
mapmie2b=[];
nrayobs2b=[];
nmieobs2b=[];
lat_losdem_mie_2b=[];
lon_losdem_mie_2b=[];
lat_start_mie_2b=[];
lon_start_mie_2b=[];
lat_stop_mie_2b=[];
lon_stop_mie_2b=[];
lat_cog_mie_2b=[];
lon_cog_mie_2b=[];
lat_losdem_ray_2b=[];
lon_losdem_ray_2b=[];
lat_start_ray_2b=[];
lon_start_ray_2b=[];
lat_stop_ray_2b=[];
lon_stop_ray_2b=[];
lat_cog_ray_2b=[];
lon_cog_ray_2b=[];

NbBRC_Plot_2b=0;
for ibrc=1:NbBRC_Plot
	jbrc=ibrcrange(ibrc);
	if (jbrc <= NbBRC_2b)
		nmieobs2b(ibrc)=l2bData.Geolocation_ADS.DSR(jbrc).N_Obs_Mie_Actual;
		for iobs=1:nmieobs2b(ibrc)
			proftype=l2bData.Mie_HLOSwind_MDS.DSR(jbrc). ...
			    List_of_Mie_Profiles(iobs).Obs_Type;
			mieproftype2b(ibrc,iobs)=proftype;
			lat_losdem_mie_2b(ibrc,iobs)=l2bData.Geolocation_ADS.DSR(jbrc). ...
			    Mie_Geolocation(iobs).Latitude_of_DEM_Intersection;
			lon_losdem_mie_2b(ibrc,iobs)=l2bData.Geolocation_ADS.DSR(jbrc). ...
			    Mie_Geolocation(iobs).Longitude_of_DEM_Intersection;
			for ibin=1:24
				lat_start_mie_2b(ibrc,ibin,iobs)=l2bData.Geolocation_ADS.DSR(jbrc). ...
				    Mie_Geolocation(iobs).List_of_Profile_Height_Bin_Geolocation(ibin).Latitude_Start;
				lon_start_mie_2b(ibrc,ibin,iobs)=l2bData.Geolocation_ADS.DSR(jbrc). ...
				    Mie_Geolocation(iobs).List_of_Profile_Height_Bin_Geolocation(ibin).Longitude_Start;
				lat_stop_mie_2b(ibrc,ibin,iobs)=l2bData.Geolocation_ADS.DSR(jbrc). ...
				    Mie_Geolocation(iobs).List_of_Profile_Height_Bin_Geolocation(ibin).Latitude_Stop;
				lon_stop_mie_2b(ibrc,ibin,iobs)=l2bData.Geolocation_ADS.DSR(jbrc). ...
				    Mie_Geolocation(iobs).List_of_Profile_Height_Bin_Geolocation(ibin).Longitude_Stop;
				lat_cog_mie_2b(ibrc,ibin,iobs)=l2bData.Geolocation_ADS.DSR(jbrc). ...
				    Mie_Geolocation(iobs).List_of_Profile_Height_Bin_Geolocation(ibin).Latitude_COG;
				lon_cog_mie_2b(ibrc,ibin,iobs)=l2bData.Geolocation_ADS.DSR(jbrc). ...
				    Mie_Geolocation(iobs).List_of_Profile_Height_Bin_Geolocation(ibin).Longitude_COG;
				altmie2b(ibrc,ibin,iobs)=l2bData.Geolocation_ADS.DSR(jbrc). ...
				    Mie_Geolocation(iobs).List_of_Profile_Height_Bin_Geolocation(ibin).Altitude_Top;
				miewind2b(ibrc,ibin,iobs)=l2bData.Mie_HLOSwind_MDS.DSR(jbrc). ...
				    List_of_Mie_Profiles(iobs).List_of_Mie_Height_Bin_Winds(ibin).Mie_Wind_Velocity;
				miehloserror2b(ibrc,ibin,iobs)=l2bData.Product_Confidence_Data_ADS.DSR(jbrc). ...
				    L2B_Mie_Processing_QC(1).L2B_Obs_Mie_QC(iobs).L2B_Obs_Mie_Bin_QC(ibin).Hlos_Error_Estimate;

				if (~l2bData.Mie_HLOSwind_MDS.DSR(jbrc).List_of_Mie_Profiles(iobs). ...
					List_of_Mie_Height_Bin_Winds(ibin).Validity_Flag || ...
					altmie2b(ibrc,ibin,iobs) == 2147483647 || ...
					miewind2b(ibrc,ibin,iobs) == 2147483647 || ...
					miehloserror2b(ibrc,ibin,iobs) == 2147483647)
					miewind2b(ibrc,ibin,iobs)=NaN;
					altmie2b(ibrc,ibin,iobs)=NaN;
					miehloserror2b(ibrc,ibin,iobs)=NaN;
				end
				if (lat_start_mie_2b(ibrc,ibin,iobs) == 2147483647)
					lat_start_mie_2b(ibrc,ibin,iobs)=NaN;
				else
					lat_start_mie_2b(ibrc,ibin,iobs)=lat_start_mie_2b(ibrc,ibin,iobs)*1.e-6;
				end
				if (lon_start_mie_2b(ibrc,ibin,iobs) == 2147483647)
					lon_start_mie_2b(ibrc,ibin,iobs)=NaN;
				else
					lon_start_mie_2b(ibrc,ibin,iobs)=lon_start_mie_2b(ibrc,ibin,iobs)*1.e-6;
				end
				if (lat_stop_mie_2b(ibrc,ibin,iobs) == 2147483647)
					lat_stop_mie_2b(ibrc,ibin,iobs)=NaN;
				else
					lat_stop_mie_2b(ibrc,ibin,iobs)=lat_stop_mie_2b(ibrc,ibin,iobs)*1.e-6;
				end
				if (lon_stop_mie_2b(ibrc,ibin,iobs) == 2147483647)
					lon_stop_mie_2b(ibrc,ibin,iobs)=NaN;
				else
					lon_stop_mie_2b(ibrc,ibin,iobs)=lon_stop_mie_2b(ibrc,ibin,iobs)*1.e-6;
				end
				if (lat_cog_mie_2b(ibrc,ibin,iobs) == 2147483647)
					lat_cog_mie_2b(ibrc,ibin,iobs)=NaN;
				else
					lat_cog_mie_2b(ibrc,ibin,iobs)=lat_cog_mie_2b(ibrc,ibin,iobs)*1.e-6;
				end
				if (lon_cog_mie_2b(ibrc,ibin,iobs) == 2147483647)
					lon_cog_mie_2b(ibrc,ibin,iobs)=NaN;
				else
					lon_cog_mie_2b(ibrc,ibin,iobs)=lon_cog_mie_2b(ibrc,ibin,iobs)*1.e-6;
				end
			end
			if (lat_losdem_mie_2b(ibrc,iobs) == 2147483647)
				lat_losdem_mie_2b(ibrc,iobs)=NaN;
			else
				lat_losdem_mie_2b(ibrc,iobs)=lat_losdem_mie_2b(ibrc,iobs)*1.e-6;
			end
			if (lon_losdem_mie_2b(ibrc,iobs) == 2147483647)
				lon_losdem_mie_2b(ibrc,iobs)=NaN;
			else
				lon_losdem_mie_2b(ibrc,iobs)=lon_losdem_mie_2b(ibrc,iobs)*1.e-6;
			end
		end
		for imeas=1:N_meas
			for ibin=1:24
				mapmie2b(ibin,((ibrc-1)*N_meas)+imeas)=l2bData.Mie_HLOSwind_MDS.DSR(jbrc).Map_of_L1_Measurements_Used(ibin,imeas).L1_Measurement_Used;
				if (mapmie2b(ibin,((ibrc-1)*N_meas)+imeas) ~= 0)
					mapmie2b(ibin,((ibrc-1)*N_meas)+imeas)=mieproftype2b(ibrc,mapmie2b(ibin,((ibrc-1)*N_meas)+imeas));
				end
				mieinpscr2b(ibin,((ibrc-1)*N_meas)+imeas)=l2bData.Product_Confidence_Data_ADS.DSR(jbrc).L1B_Input_Screening.L1B_Mie_Meas_Screening(imeas).L1B_Mie_Meas_Bin_Screening(ibin).L1B_Mie_Meas_Bin_QC;
			end
		end
		NbBRC_Plot_2b=NbBRC_Plot_2b+1;
	end
end
str_NbBRC_Plot_2b=int2str(NbBRC_Plot_2b);

rho2b=[];

for ibrc=1:NbBRC_Plot_2b
	jbrc=ibrcrange(ibrc);
	if (jbrc <= NbBRC_2b)
		nrayobs2b(ibrc)=l2bData.Geolocation_ADS.DSR(jbrc).N_Obs_Rayleigh_Actual;
		for iobs=1:nrayobs2b(ibrc)
			proftype=l2bData.Rayleigh_HLOSwind_MDS.DSR(jbrc). ...
			    List_of_Rayleigh_Profiles(iobs).Obs_Type;
			rayproftype2b(ibrc,iobs)=proftype;
			lat_losdem_ray_2b(ibrc,iobs)=l2bData.Geolocation_ADS.DSR(jbrc). ...
			    Rayleigh_Geolocation(iobs).Latitude_of_DEM_Intersection;
			lon_losdem_ray_2b(ibrc,iobs)=l2bData.Geolocation_ADS.DSR(jbrc). ...
			    Rayleigh_Geolocation(iobs).Longitude_of_DEM_Intersection;
			for ibin=1:24
				lat_start_ray_2b(ibrc,ibin,iobs)=l2bData.Geolocation_ADS.DSR(jbrc). ...
				    Rayleigh_Geolocation(iobs).List_of_Profile_Height_Bin_Geolocation(ibin).Latitude_Start;
				lon_start_ray_2b(ibrc,ibin,iobs)=l2bData.Geolocation_ADS.DSR(jbrc). ...
				    Rayleigh_Geolocation(iobs).List_of_Profile_Height_Bin_Geolocation(ibin).Longitude_Start;
				lat_stop_ray_2b(ibrc,ibin,iobs)=l2bData.Geolocation_ADS.DSR(jbrc). ...
				    Rayleigh_Geolocation(iobs).List_of_Profile_Height_Bin_Geolocation(ibin).Latitude_Stop;
				lon_stop_ray_2b(ibrc,ibin,iobs)=l2bData.Geolocation_ADS.DSR(jbrc). ...
				    Rayleigh_Geolocation(iobs).List_of_Profile_Height_Bin_Geolocation(ibin).Longitude_Stop;
				lat_cog_ray_2b(ibrc,ibin,iobs)=l2bData.Geolocation_ADS.DSR(jbrc). ...
				    Rayleigh_Geolocation(iobs).List_of_Profile_Height_Bin_Geolocation(ibin).Latitude_COG;
				lon_cog_ray_2b(ibrc,ibin,iobs)=l2bData.Geolocation_ADS.DSR(jbrc). ...
				    Rayleigh_Geolocation(iobs).List_of_Profile_Height_Bin_Geolocation(ibin).Longitude_COG;
				altray2b(ibrc,ibin,iobs)=l2bData.Geolocation_ADS.DSR(jbrc). ...
				    Rayleigh_Geolocation(iobs).List_of_Profile_Height_Bin_Geolocation(ibin).Altitude_Top;
				raywind2b(ibrc,ibin,iobs)=l2bData.Rayleigh_HLOSwind_MDS.DSR(jbrc). ...
				    List_of_Rayleigh_Profiles(iobs).List_of_Rayleigh_Height_Bin_Winds(ibin).Rayleigh_Wind_Velocity;
				rayhloserror2b(ibrc,ibin,iobs)=l2bData.Product_Confidence_Data_ADS.DSR(jbrc). ...
				    L2B_Rayleigh_Processing_QC(1).L2B_Obs_Rayleigh_QC(iobs).L2B_Obs_Rayleigh_Bin_QC(ibin).Hlos_Error_Estimate;

				rhotmp=double(l2bData.Rayleigh_HLOSwind_MDS.DSR(jbrc). ...
				    List_of_Rayleigh_Profiles(iobs).List_of_Rayleigh_Height_Bin_Winds(ibin).Reference_Backscatter_Ratio)*1.e-6;
				if (rhotmp > 0.)
				    rho2b(ibrc,ibin,iobs)=rhotmp;
				else
				    rho2b(ibrc,ibin,iobs)=NaN;
				end
%for j=1:N_meas
%tabint(j)=l1bData.Product_Confidence_Data_ADS.DSR(jbrc).List_of_Measurement_PCDs(1). ...
%          Measurement_PCD(j).List_of_Meas_Alt_Bin_PCDs(1). ...
%	  Meas_Alt_Bin_PCD(ibin).Scattering_Ratio_Mie;
%end
%rho1b(ibrc,ibin)=double(mean(tabint));
				if (~l2bData.Rayleigh_HLOSwind_MDS.DSR(jbrc). ...
					    List_of_Rayleigh_Profiles(iobs).List_of_Rayleigh_Height_Bin_Winds(ibin).Validity_Flag ||  ...
					    altray2b(ibrc,ibin,iobs) == 2147483647 ||  ...
					    raywind2b(ibrc,ibin,iobs) == 2147483647 ||  ...
					    rayhloserror2b(ibrc,ibin,iobs) == 2147483647)
					raywind2b(ibrc,ibin,iobs)=NaN;
					altray2b(ibrc,ibin,iobs)=NaN;
					rayhloserror2b(ibrc,ibin,iobs)=NaN;
				end
				if (lat_start_ray_2b(ibrc,ibin,iobs) == 2147483647)
					lat_start_ray_2b(ibrc,ibin,iobs)=NaN;
				else
					lat_start_ray_2b(ibrc,ibin,iobs)=lat_start_ray_2b(ibrc,ibin,iobs)*1.e-6;
				end
				if (lon_start_ray_2b(ibrc,ibin,iobs) == 2147483647)
					lon_start_ray_2b(ibrc,ibin,iobs)=NaN;
				else
					lon_start_ray_2b(ibrc,ibin,iobs)=lon_start_ray_2b(ibrc,ibin,iobs)*1.e-6;
				end
				if (lat_stop_ray_2b(ibrc,ibin,iobs) == 2147483647)
					lat_stop_ray_2b(ibrc,ibin,iobs)=NaN;
				else
					lat_stop_ray_2b(ibrc,ibin,iobs)=lat_stop_ray_2b(ibrc,ibin,iobs)*1.e-6;
				end
				if (lon_stop_ray_2b(ibrc,ibin,iobs) == 2147483647)
					lon_stop_ray_2b(ibrc,ibin,iobs)=NaN;
				else
					lon_stop_ray_2b(ibrc,ibin,iobs)=lon_stop_ray_2b(ibrc,ibin,iobs)*1.e-6;
				end
				if (lat_cog_ray_2b(ibrc,ibin,iobs) == 2147483647)
					lat_cog_ray_2b(ibrc,ibin,iobs)=NaN;
				else
					lat_cog_ray_2b(ibrc,ibin,iobs)=lat_cog_ray_2b(ibrc,ibin,iobs)*1.e-6;
				end
				if (lon_cog_ray_2b(ibrc,ibin,iobs) == 2147483647)
					lon_cog_ray_2b(ibrc,ibin,iobs)=NaN;
				else
					lon_cog_ray_2b(ibrc,ibin,iobs)=lon_cog_ray_2b(ibrc,ibin,iobs)*1.e-6;
				end
			end
			if (lat_losdem_ray_2b(ibrc,iobs) == 2147483647)
				lat_losdem_ray_2b(ibrc,iobs)=NaN;
			else
				lat_losdem_ray_2b(ibrc,iobs)=lat_losdem_ray_2b(ibrc,iobs)*1.e-6;
			end
			if (lon_losdem_ray_2b(ibrc,iobs) == 2147483647)
				lon_losdem_ray_2b(ibrc,iobs)=NaN;
			else
				lon_losdem_ray_2b(ibrc,iobs)=lon_losdem_ray_2b(ibrc,iobs)*1.e-6;
			end
		end
		for imeas=1:N_meas
			for ibin=1:24
				mapray2b(ibin,((ibrc-1)*N_meas)+imeas)=l2bData.Rayleigh_HLOSwind_MDS.DSR(jbrc).Map_of_L1_Measurements_Used(ibin,imeas).L1_Measurement_Used;
				if (mapray2b(ibin,((ibrc-1)*N_meas)+imeas) ~= 0)
					mapray2b(ibin,((ibrc-1)*N_meas)+imeas)=rayproftype2b(ibrc,mapray2b(ibin,((ibrc-1)*N_meas)+imeas));
				end
				rayinpscr2b(ibin,((ibrc-1)*N_meas)+imeas)=l2bData.Product_Confidence_Data_ADS.DSR(jbrc).L1B_Input_Screening.L1B_Rayleigh_Meas_Screening(imeas).L1B_Rayleigh_Meas_Bin_Screening(ibin).L1B_Rayleigh_Meas_Bin_QC;
			end
		end
	else
		nrayobs2b(ibrc)=0;
		for imeas=1:N_meas
			for ibin=1:24
				mapray2b(ibin,((ibrc-1)*N_meas)+imeas)=NaN;
				rayinpscr2b(ibin,((ibrc-1)*N_meas)+imeas)=NaN;
			end
		end
	end
end

end % trsfdatal2b & l2bpresent

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%               TRANSFER L1B DATA INTO ARRAYS
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if (trsfdatal1b & l1bpresent)

NbBRC_1b=l1bData.nbDSR;

if (exist('N_meas'))
        if (l1bData.Nmax ~= N_meas)
		error(sprintf('Inconsistent number of measurements between L1B (%i) and L2B (%i)', ...
		l1bData.Nmax, N_meas));
	end
else
	N_meas=l1bData.Nmax;
end
if (exist('NbBRC_2b'))
	if (NbBRC_2b ~= NbBRC_1b)
		fprintf('Warning: Inconsistent number of BRCs between L1B (%i) and L2B (%i).\n', ...
		NbBRC_1b,NbBRC_2b)
	end
end

mie1bwindobs=[];
mie1bhloserrorobs=[];
altmie1bobs=[];
ray1bwindobs=[];
ray1bhloserrorobs=[];
altray1bobs=[];
mie1bwindmeas=[];
altmie1bmeas=[];
ray1bwindmeas=[];
altray1bmeas=[];
lat_losdem_obs_1b=[];
lon_losdem_obs_1b=[];
lat_obs_mie_1b=[];
lon_obs_mie_1b=[];
lat_obs_ray_1b=[];
lon_obs_ray_1b=[];

NbBRC_Plot_1b=0;
for ibrc=1:NbBRC_Plot
	jbrc=ibrcrange(ibrc);
	if (jbrc <= NbBRC_1b)
		lat_losdem_obs_1b(ibrc)=1.e-6*double(l1bData.Geolocation_ADS.DSR(jbrc). ...
		    Observation_Geolocation(1).Geolocation_of_DEM_Intersection(1).Latitude_of_DEM_Intersection);
		lon_losdem_obs_1b(ibrc)=1.e-6*double(l1bData.Geolocation_ADS.DSR(jbrc). ...
		    Observation_Geolocation(1).Geolocation_of_DEM_Intersection(1).Longitude_of_DEM_Intersection);
		for imeas=1:N_meas
			for ibin=1:24
				ixslot=(ibrc-1)*N_meas+imeas;
				iyslot=ibin;
				lat_obs_mie_1b(ibrc,ibin)=1.e-6*double(l1bData.Geolocation_ADS.DSR(jbrc). ...
				    Observation_Geolocation.Observation_Mie_Geolocation(1). ...
				    List_of_Observation_Geolocation_of_Height_Bins(1). ...
				    Observation_Geolocation_of_Height_Bin(ibin).Latitude_of_Height_Bin);
				lon_obs_mie_1b(ibrc,ibin)=1.e-6*double(l1bData.Geolocation_ADS.DSR(jbrc). ...
				    Observation_Geolocation.Observation_Mie_Geolocation(1). ...
				    List_of_Observation_Geolocation_of_Height_Bins(1). ...
				    Observation_Geolocation_of_Height_Bin(ibin).Longitude_of_Height_Bin);
				lat_obs_ray_1b(ibrc,ibin)=1.e-6*double(l1bData.Geolocation_ADS.DSR(jbrc). ...
				    Observation_Geolocation.Observation_Rayleigh_Geolocation(1). ...
				    List_of_Observation_Geolocation_of_Height_Bins(1). ...
				    Observation_Geolocation_of_Height_Bin(ibin).Latitude_of_Height_Bin);
				lon_obs_ray_1b(ibrc,ibin)=1.e-6*double(l1bData.Geolocation_ADS.DSR(jbrc). ...
				    Observation_Geolocation.Observation_Rayleigh_Geolocation(1). ...
				    List_of_Observation_Geolocation_of_Height_Bins(1). ...
				    Observation_Geolocation_of_Height_Bin(ibin).Longitude_of_Height_Bin);
				mie1bwindmeas(ixslot,iyslot)=l1bData.Wind_Velocity_MDS.DSR(jbrc).List_of_Measurement_Wind_Profile(1).Measurement_Wind_Profile(imeas).List_of_Mie_Altitude_Bin_Wind_Infos(1).Mie_Altitude_Bin_Wind_Info(ibin).Wind_Velocity;
				altmie1bmeas (ixslot,iyslot)=l1bData.Geolocation_ADS.DSR(jbrc).List_of_Measurement_Geolocations(1).Measurement_Geolocation(imeas).Mie_Geolocation(1).List_of_Geolocation_Height_Bins(1).Geolocation_Height_Bin(ibin).Altitude_of_Height_Bin;
				if (max(l1bData.Wind_Velocity_MDS.DSR(jbrc).List_of_Measurement_Wind_Profile(1).Measurement_Wind_Profile(imeas).Bin_Quality_Flag(ibin).data) ~= 0 | mie1bwindmeas(ixslot,iyslot)==0.)
					mie1bwindmeas(ixslot,iyslot)=NaN;
				end
				ray1bwindmeas(ixslot,iyslot)=l1bData.Wind_Velocity_MDS.DSR(jbrc).List_of_Measurement_Wind_Profile(1).Measurement_Wind_Profile(imeas).List_of_Rayleigh_Altitude_Bin_Wind_Infos(1).Rayleigh_Altitude_Bin_Wind_Info(ibin).Wind_Velocity;
				altray1bmeas (ixslot,iyslot)=l1bData.Geolocation_ADS.DSR(jbrc).List_of_Measurement_Geolocations(1).Measurement_Geolocation(imeas).Rayleigh_Geolocation(1).List_of_Geolocation_Height_Bins(1).Geolocation_Height_Bin(ibin).Altitude_of_Height_Bin;
				if (max(l1bData.Wind_Velocity_MDS.DSR(jbrc).List_of_Measurement_Wind_Profile(1).Measurement_Wind_Profile(imeas).Bin_Quality_Flag(ibin).data) ~= 0 | ray1bwindmeas(ixslot,iyslot)==0.)
					ray1bwindmeas(ixslot,iyslot)=NaN;
				end
			end
		end
		for ibin=1:24
			ixslot=ibrc;
			iyslot=ibin;
			mie1bwindobs(ixslot,iyslot)=l1bData.Wind_Velocity_MDS.DSR(jbrc).Observation_Wind_Profile(1).List_of_Mie_Altitude_Bin_Wind_Infos(1).Mie_Altitude_Bin_Wind_Info(ibin).Wind_Velocity;
			mie1bhloserrorobs(ixslot,iyslot)=l1bData.Product_Confidence_Data_ADS.DSR(jbrc).Observation_PCD(1).List_of_Observation_Alt_Bin_PCDs(1).Observation_Alt_Bin_PCD(ibin).ErrorQuantifier.ErrorQuantifierMie;
			altmie1bobs (ixslot,iyslot)=l1bData.Geolocation_ADS.DSR(jbrc).Observation_Geolocation(1).Observation_Mie_Geolocation(1).List_of_Observation_Geolocation_of_Height_Bins(1).Observation_Geolocation_of_Height_Bin(ibin).Altitude_of_Height_Bin;
			if (max(l1bData.Wind_Velocity_MDS.DSR(jbrc).Observation_Wind_Profile(1).List_of_Mie_Altitude_Bin_Wind_Infos(1).Mie_Altitude_Bin_Wind_Info(ibin).Bin_Quality_Flag) ~= 0 || ...
				mie1bwindobs(ixslot,iyslot) == 0.)
				mie1bwindobs(ixslot,iyslot)=NaN;
				mie1bhloserrorobs(ixslot,iyslot)=NaN;
			end
			ray1bwindobs(ixslot,iyslot)=l1bData.Wind_Velocity_MDS.DSR(jbrc).Observation_Wind_Profile(1).List_of_Rayleigh_Altitude_Bin_Wind_Infos(1).Rayleigh_Altitude_Bin_Wind_Info(ibin).Wind_Velocity;
			ray1bhloserrorobs(ixslot,iyslot)=l1bData.Product_Confidence_Data_ADS.DSR(jbrc).Observation_PCD(1).List_of_Observation_Alt_Bin_PCDs(1).Observation_Alt_Bin_PCD(ibin).ErrorQuantifier.ErrorQuantifierRay;
			altray1bobs (ixslot,iyslot)=l1bData.Geolocation_ADS.DSR(jbrc).Observation_Geolocation(1).Observation_Rayleigh_Geolocation(1).List_of_Observation_Geolocation_of_Height_Bins(1).Observation_Geolocation_of_Height_Bin(ibin).Altitude_of_Height_Bin;
			if (max(l1bData.Wind_Velocity_MDS.DSR(jbrc).Observation_Wind_Profile(1).List_of_Rayleigh_Altitude_Bin_Wind_Infos(1).Rayleigh_Altitude_Bin_Wind_Info(ibin).Bin_Quality_Flag) ~= 0 || ...
				ray1bwindobs(ixslot,iyslot) == 0.)
				ray1bwindobs(ixslot,iyslot)=NaN;
				ray1bhloserrorobs(ixslot,iyslot)=NaN;
			end
		end
		ibin=25;
		iyslot=ibin;
		altray1bobs (ixslot,iyslot)=l1bData.Geolocation_ADS.DSR(jbrc).Observation_Geolocation(1).Observation_Rayleigh_Geolocation(1).List_of_Observation_Geolocation_of_Height_Bins(1).Observation_Geolocation_of_Height_Bin(ibin).Altitude_of_Height_Bin;
		altmie1bobs (ixslot,iyslot)=l1bData.Geolocation_ADS.DSR(jbrc).Observation_Geolocation(1).Observation_Mie_Geolocation(1).List_of_Observation_Geolocation_of_Height_Bins(1).Observation_Geolocation_of_Height_Bin(ibin).Altitude_of_Height_Bin;
		NbBRC_Plot_1b=NbBRC_Plot_1b+1;
	end
end
str_NbBRC_Plot_1b=int2str(NbBRC_Plot_1b);

nprofray1b=size(altray1bobs,1);
nprofmie1b=size(altmie1bobs,1);
altrayavg1b=[];
altmieavg1b=[];
for ibin=1:25
	altrayavg1b(ibin)=nanmean(altray1bobs(1:nprofray1b,ibin))/1000.;
	altmieavg1b(ibin)=nanmean(altmie1bobs(1:nprofmie1b,ibin))/1000.;
end

rho1bmeas=[];
rho1b=[];
for ibin=1:24
kray1bmeas=0;
kray1b=0;
   for ibrc=1:NbBRC_Plot_1b
      jbrc=ibrcrange(ibrc);
      kray1b=kray1b+1;
      for j=1:N_meas
         kray1bmeas=kray1bmeas+1;
         rho1bmeas(ibin,kray1bmeas)=l1bData.Product_Confidence_Data_ADS.DSR(jbrc). ...
         List_of_Measurement_PCDs(1).Measurement_PCD(j).List_of_Meas_Alt_Bin_PCDs(1). ...
         Meas_Alt_Bin_PCD(ibin).Scattering_Ratio_Mie;
         tabint(j)=rho1bmeas(ibin,kray1bmeas);
         if (rho1bmeas(ibin,kray1bmeas) <= 0),
            rho1bmeas(ibin,kray1bmeas)=NaN;
         end
      end
      rho1b(kray1b,ibin)=double(mean(tabint));
   end
end

end % trsfdatal1b & l1bpresent

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%               TRANSFER E2S DATA INTO ARRAYS
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if (trsfdatae2s & e2spresent)

winde2s=[];
hloswinde2s=[];
alte2s=[];
raye2sbackscat=[];
miee2sbackscat=[];
rhoe2s=[];
alte2savg=[];
alte2sobs=[];
hloswinde2sobs=[];
hloswinde2sobsstd=[];
hloswinde2sobsmin=[];
hloswinde2sobsmax=[];

nbine2s=length(e2sData.DSR(1).Atmosphere(1).Altitude);
NbBRC_e2s=length(e2sData.DSR);

if (NbBRC_1b ~= NbBRC_e2s)
	fprintf('Inconsistent number of BRC between E2S input (%i) and L1B (%i).\n',NbBRC_e2s,NbBRC_1b)
end

NbBRC_Plot_e2s=0;
for ibrc=1:NbBRC_Plot
	jbrc=ibrcrange(ibrc);
	if (jbrc <= NbBRC_e2s)
		if (length(e2sData.DSR(jbrc).Atmosphere) ~= N_meas)
			error(sprintf('Inconsistent number of measurements between E2S input (%i) and L1B (%i)',length(e2sData.DSR(jbrc).Atmosphere),N_meas))
		end
		for imeas=1:N_meas
			for ibin=1:nbine2s
				ixslot=(ibrc-1)*N_meas+imeas;
				iyslot=ibin;
				if (ibin <= length(e2sData.DSR(jbrc).Atmosphere(imeas).Altitude)),
					alte2s(ixslot,iyslot)=e2sData.DSR(jbrc).Atmosphere(imeas).Altitude(ibin);
					winde2s(ixslot,iyslot)=e2sData.DSR(jbrc).Atmosphere(imeas).Wind(ibin);
					hloswinde2s(ixslot,iyslot)=e2sData.DSR(jbrc).Atmosphere(imeas).HlosWind(ibin);
					miee2sbackscat(ixslot,iyslot)=e2sData.DSR(jbrc).Atmosphere(imeas).MieBackScattCoef(ibin);
					raye2sbackscat(ixslot,iyslot)=e2sData.DSR(jbrc).Atmosphere(imeas).RayBackScattCoef(ibin);
					rhoe2s(ixslot,iyslot)=1.+miee2sbackscat(ixslot,iyslot)/raye2sbackscat(ixslot,iyslot);
				else
					alte2s(ixslot,iyslot)=NaN;
					winde2s(ixslot,iyslot)=NaN;
					hloswinde2s(ixslot,iyslot)=NaN;
					miee2sbackscat(ixslot,iyslot)=NaN;
					raye2sbackscat(ixslot,iyslot)=NaN;
					rhoe2s(ixslot,iyslot)=NaN;
				end
			end
		end
		NbBRC_Plot_e2s=NbBRC_Plot_e2s+1;
	end
end
str_NbBRC_Plot_e2s=int2str(NbBRC_Plot_e2s);

for ibrc=1:NbBRC_Plot_e2s
	jbrc=ibrcrange(ibrc);
	xrangebrc=(ibrc-1)*N_meas+1:ibrc*N_meas;
	for ibin=1:nbine2s
		ixslot=ibrc;
		iyslot=ibin;
		alte2sobs(ixslot,iyslot)=mean(alte2s(xrangebrc,ibin));
		hloswinde2sobs(ixslot,iyslot)=mean(hloswinde2s(xrangebrc,ibin));
		hloswinde2sobsstd(ixslot,iyslot)=std(hloswinde2s(xrangebrc,ibin));
		hloswinde2sobsmin(ixslot,iyslot)=min(hloswinde2s(xrangebrc,ibin));
		hloswinde2sobsmax(ixslot,iyslot)=max(hloswinde2s(xrangebrc,ibin));
	end
end
for ibin=1:nbine2s
	alte2savg(ibin)=mean(alte2sobs(1:NbBRC_Plot,ibin));
end

end % trsfdatae2s & e2spresent


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%       GENERATE REPORT ON TIME MATCHING
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if (check_timestamps)
	str_match_1b_2b=[];
	fprintf('=========== TIME MATCHING REPORT ===========\n')
	fprintf('BRC# |1b present|2b present|1b & 2b matching\n')
	fprintf('     |          |          |day sec microsec\n')
	fprintf('-----+----------+----------+----------------\n')
	for ibrc=1:NbBRC_Plot
		jbrc=ibrcrange(ibrc);
		e2spresent_ibrc=false;
		if (e2spresent)
			e2spresent_ibrc=jbrc <= NbBRC_Plot_e2s;
		end
		l1bpresent_ibrc=false;
		if (l1bpresent)
			l1bpresent_ibrc=jbrc <= NbBRC_Plot_1b;
		end
		l2bpresent_ibrc=false;
		if (l2bpresent)
			l2bpresent_ibrc=jbrc <= NbBRC_Plot_2b;
		end
		if (e2spresent_ibrc)
			str_e2spresent_ibrc='YES';
		else
			str_e2spresent_ibrc='NO ';
		end
		if (l1bpresent_ibrc)
			str_l1bpresent_ibrc='YES';
		else
			str_l1bpresent_ibrc='NO ';
		end
		if (l2bpresent_ibrc)
			str_l2bpresent_ibrc='YES';
		else
			str_l2bpresent_ibrc='NO ';
		end
		if (l1bpresent_ibrc & l2bpresent_ibrc)
			for it=1:3
				if (l1bData.Geolocation_ADS.DSR(jbrc).Start_of_Observation_Time(it) == l2bData.Geolocation_ADS.DSR(jbrc).Start_of_Obs_Time(it))
					str_match_1b_2b{ibrc,it}='YES';
				else
					str_match_1b_2b{ibrc,it}='NO ';
				end
			end
		else
			str_match_1b_2b{ibrc,1}='N/A';
			str_match_1b_2b{ibrc,2}='N/A';
			str_match_1b_2b{ibrc,3}='N/A';
		end
		fprintf('%4d |    %s   |    %s   |%s %s %s\n',jbrc,str_l1bpresent_ibrc,str_l2bpresent_ibrc, ...
		      str_match_1b_2b{ibrc,1},str_match_1b_2b{ibrc,2},str_match_1b_2b{ibrc,3})
	end
	fprintf('======== END OF TIME MATCHING REPORT =======\n')
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%       GENERATE REPORT ON TIME MATCHING
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if (check_latlonstamps)
	str_match_1b_2b=[];
	fprintf('========= LAT/LON MATCHING REPORT ==========\n')
	for ibrc=1:NbBRC_Plot
		jbrc=ibrcrange(ibrc);
		l1bpresent_ibrc=false;
		if (l1bpresent)
			l1bpresent_ibrc=jbrc <= NbBRC_Plot_1b;
		end
		l2bpresent_ibrc=false;
		if (l2bpresent)
			l2bpresent_ibrc=jbrc <= NbBRC_Plot_2b;
		end
		if (l1bpresent_ibrc)
			str_l1bpresent_ibrc='YES';
		else
			str_l1bpresent_ibrc='NO ';
		end
		if (l2bpresent_ibrc)
			str_l2bpresent_ibrc='YES';
		else
			str_l2bpresent_ibrc='NO ';
		end
		if (l1bpresent_ibrc & l2bpresent_ibrc)
			fprintf('BRC#%i 1b and 2b data present                                              SEEMS OK? 1b OK?  2b OK? (blank=YES)\n',jbrc);
			% compare LOSDEM Mie L2B with LOSDEM L1B
			if (nmieobs2b(ibrc) > 0)
				for iobs=1:nmieobs2b(ibrc)
					d1=print_dist(lat_losdem_obs_1b(ibrc),lon_losdem_obs_1b(ibrc), ...
					    lat_losdem_mie_2b(ibrc,iobs),lon_losdem_mie_2b(ibrc,iobs), ...
					    sprintf('BRC#%i distance between Mie 2b ondem obs#%1i and Mie obs 1b       =',ibrc,iobs));
					%fprintf('BRC#%i distance between Mie 2b ondem obs#%1i and Mie obs 1b           =%5.0f km\n', ...
					%    ibrc,iobs,d1);
					for ibin=1:24
						if (~isnan(altmie2b(ibrc,ibin,iobs)))
						d1=print_dist(lat_obs_mie_1b(ibrc,ibin),lon_obs_mie_1b(ibrc,ibin), ...
						    lat_start_mie_2b(ibrc,ibin,iobs),lon_start_mie_2b(ibrc,ibin,iobs), ...
						    sprintf('BRC#%i distance between Mie 2b start obs#%1i and Mie obs 1b bin#%2i=',ibrc,iobs,ibin));
						d1=print_dist(lat_obs_mie_1b(ibrc,ibin),lon_obs_mie_1b(ibrc,ibin), ...
						    lat_stop_mie_2b(ibrc,ibin,iobs),lon_stop_mie_2b(ibrc,ibin,iobs), ...
						    sprintf('BRC#%i distance between Mie 2b stop  obs#%1i and Mie obs 1b bin#%2i=',ibrc,iobs,ibin));
						d1=print_dist(lat_obs_mie_1b(ibrc,ibin),lon_obs_mie_1b(ibrc,ibin), ...
						    lat_cog_mie_2b(ibrc,ibin,iobs),lon_cog_mie_2b(ibrc,ibin,iobs), ...
						    sprintf('BRC#%i distance between Mie 2b cog   obs#%1i and Mie obs 1b bin#%2i=',ibrc,iobs,ibin));
						end
					end
				end
			end
		else
			fprintf('BRC#%i either 1b or 2b data not present\n',jbrc);
		end
	end
	fprintf('====== END OF LAT/LON MATCHING REPORT ======\n')
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%               GENERATE PLOTS
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if (plotdata)

ifig=0;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 2D plot of E2S and L1B scattering ratio at measurement level
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if (plotconfig.plot_rho & (l1bpresent | e2spresent) )

ifig=ifig+1;
figure(ifig)
clf

if (l1bpresent)
	rhomax=mean(rho1b(rho1b>1. & rho1b < 5.))+4.*std(rho1b(rho1b>1. & rho1b < 5.));
	zmin=min(altrayavg1b);
	zmax=max(altrayavg1b);
else
	zmin=min(min(alte2s*1.e-3));
	zmax=max(max(alte2s*1.e-3));
	rhomax=mean(rhoe2s(rhoe2s>1. & rhoe2s < 5.))+4.*std(rhoe2s(rhoe2s>1. & rhoe2s < 5.));
end % l1bpresent

for ibrc=1:NbBIGPlot
	stribrc=int2str(ibrc);
	subplot(2,NbBIGPlot,ibrc)

	if (e2spresent)
		x=1:N_meas*NbBRC_Plot_e2s/NbBIGPlot;
		y=alte2savg*1.e-3;
		pretabplot=rhoe2s((ibrc-1)*max(x)+1:ibrc*max(x),nbine2s:-1:1)';
		x=[x,x(end)+x(end)-x(end-1)];
		tabplot=[[pretabplot,pretabplot(:,end)];[pretabplot(end,:),pretabplot(end,end)]];
		h=pcolor(x,[y(nbine2s:-1:1) y(1)+(y(1)-y(2))],tabplot);
		set(h,'EdgeAlpha',0)
		set(gca,'XLim',[1 max(x)]);
		set(gca,'YLim',[zmin zmax]);
		for imeas=1:N_meas*NbBRC_Plot_e2s
			if (imeas > 1 & mod(imeas,N_meas) == 0)
				hline=line([imeas+1,imeas+1],[zmin,zmax]);
				set(hline,'Color',[0 0 0])
			end
		end
		if (ibrc==1)
			ylabel(ylab_alt)
		end
		xlabel(['Measurement # out of the first ' str_NbBRC_Plot_e2s ' BRCs ' plotconfig.BRCrange_To_Plot])
		caxis([1. rhomax]);
		hc=colorbar;
		title([ plotconfig.scenario ' : E2S meas. scattering ratio'],'Interpreter','none')
	else
		title([ plotconfig.scenario ' : NO E2S DATA AVAILABLE'],'Interpreter','none')
	end % e2spresent
end % ibrc

for ibrc=1:NbBIGPlot
	stribrc=int2str(ibrc);
	subplot(2,NbBIGPlot,NbBIGPlot+ibrc)

	if (l1bpresent)
		x=1:N_meas*NbBRC_Plot_1b/NbBIGPlot;
		y=altrayavg1b;
		pretabplot=rho1bmeas(24:-1:1,(ibrc-1)*max(x)+1:ibrc*max(x));
		x=[x,x(end)+x(end)-x(end-1)];
		tabplot=[[pretabplot,pretabplot(:,end)];[pretabplot(end,:),pretabplot(end,end)]];
		h=pcolor(x,y(25:-1:1),tabplot);
		set(h,'EdgeAlpha',0)
		set(gca,'XLim',[1 max(x)]);
		set(gca,'YLim',[zmin zmax]);
		for imeas=1:N_meas*NbBRC_Plot_1b
			if (imeas > 1 & mod(imeas,N_meas) == 0)
				hline=line([imeas+1,imeas+1],[zmin,zmax]);
				set(hline,'Color',[0 0 0])
			end
		end
		for ibin=2:24
			hline=line([x(1) x(end)], [y(ibin) y(ibin)]);
			set(hline,'Color',[0 0 0])
		end
		if (ibrc==1)
			ylabel(ylab_alt)
		end
		xlabel(['Measurement # out of the first ' str_NbBRC_Plot_1b ' BRCs ' plotconfig.BRCrange_To_Plot])
		caxis([1. rhomax]);
		hc=colorbar;
		title([ plotconfig.scenario ' : L1b meas. scattering ratio'],'Interpreter','none')
	else
		title([ plotconfig.scenario ' : NO L1B DATA AVAILABLE'],'Interpreter','none')
	end % l1bpresent
end % ibrc

if (isfield(plotconfig,'printext'))
	fic=[ plotconfig.scenario '_rho' plotconfig.printext ]
	set(gcf,'PaperUnits','centimeters','PaperPosition',[2. 4. 18. 24.])
	print(fic,plotconfig.printcmd);
end
if (isfield(plotconfig,'figext'))
	fic=[ plotconfig.scenario '_rho' plotconfig.figext ]
	saveas(gcf,fic);
end
end % l1bpresent or e2spresent

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 2D plot of L2B scattering ratio at observation level 
% at the measurement locations of the clear and cloudy observations
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if (plotconfig.plot_rho2b & l2bpresent)
ifig=ifig+1;
figure(ifig)
clf
%if (l1bpresent)
%	rhomax=mean(rho1b(rho1b>1. & rho1b < 5.))+4.*std(rho1b(rho1b>1. & rho1b < 5.));
%	y=altrayavg1b;
%	ylab=ylab_alt;
%else
	rhomax=mean(rho2b(rho2b>1. & rho2b < 5.))+4.*std(rho2b(rho2b>1. & rho2b < 5.));
	y=[25:-1:1];
	ylab=ylab_bin;
%end
zmin=min(y);
zmax=max(y);
NbBIGPlot=1;

for ibrc=1:NbBIGPlot

	subplot(2,NbBIGPlot,1)
	x=1:N_meas*NbBRC_Plot_2b/NbBIGPlot;
	x=[x,x(end)+x(end)-x(end-1)];
	pretabplot=[];
	for idsr=1:NbBRC_Plot_2b
		pretabplot((idsr-1)*N_meas+1:idsr*N_meas,1:24)=NaN;
		for iobs=1:nrayobs2b(idsr),
			if (rayproftype2b(idsr,iobs) == iclear),
			for ibin=1:24,
				for imeas=1:N_meas,
					iimeas0=(idsr-1)*N_meas+imeas;
					if (mapray2b(ibin,iimeas0) == iclear)
						pretabplot(iimeas0,24-ibin+1)=rho2b(idsr,ibin,iobs);
					end
				end
			end
			end
		end
	end
	pretabplot=pretabplot';
	tabplot=[[pretabplot,pretabplot(:,end)];[pretabplot(end,:),pretabplot(end,end)]];
	h=pcolor(x,y(25:-1:1),tabplot);
	set(h,'EdgeAlpha',0)
	set(gca,'XLim',[1 max(x)]);
	set(gca,'YLim',[zmin zmax]);
	for imeas=1:N_meas*NbBRC_Plot_2b
		if (imeas > 1 & mod(imeas,N_meas) == 0)
			hline=line([imeas+1,imeas+1],[zmin,zmax]);
			set(hline,'Color',[0 0 0])
		end
	end
	for ibin=2:24
		hline=line([x(1) x(end)], [y(ibin) y(ibin)]);
		set(hline,'Color',[0 0 0])
	end
	if (ibrc==1)
		ylabel(ylab)
	end
	xlabel(['Measurement # out of the first ' str_NbBRC_Plot_2b ' BRCs ' plotconfig.BRCrange_To_Plot])
	caxis([1. rhomax]);
	hc=colorbar;
	title([ plotconfig.scenario ' : L2b obs. scattering ratio [Clear]'],'Interpreter','none')
	
	subplot(2,NbBIGPlot,NbBIGPlot+1)
	x=1:N_meas*NbBRC_Plot_2b/NbBIGPlot;
	x=[x,x(end)+x(end)-x(end-1)];
	pretabplot=[];
	for idsr=1:NbBRC_Plot_2b
		pretabplot((idsr-1)*N_meas+1:idsr*N_meas,1:24)=NaN;
		for iobs=1:nrayobs2b(idsr),
			if (rayproftype2b(idsr,iobs) == icloudy),
			for ibin=1:24,
				for imeas=1:N_meas,
					iimeas0=(idsr-1)*N_meas+imeas;
					if (mapray2b(ibin,iimeas0) == icloudy)
						pretabplot(iimeas0,24-ibin+1)=rho2b(idsr,ibin,iobs);
					end
				end
			end
			end
		end
	end
	pretabplot=pretabplot';
	tabplot=[[pretabplot,pretabplot(:,end)];[pretabplot(end,:),pretabplot(end,end)]];
	h=pcolor(x,y(25:-1:1),tabplot);
	set(h,'EdgeAlpha',0)
	set(gca,'XLim',[1 max(x)]);
	set(gca,'YLim',[zmin zmax]);
	for imeas=1:N_meas*NbBRC_Plot_2b
		if (imeas > 1 & mod(imeas,N_meas) == 0)
			hline=line([imeas+1,imeas+1],[zmin,zmax]);
			set(hline,'Color',[0 0 0])
		end
	end
	for ibin=2:24
		hline=line([x(1) x(end)], [y(ibin) y(ibin)]);
		set(hline,'Color',[0 0 0])
	end
	if (ibrc==1)
		ylabel(ylab)
	end
	xlabel(['Measurement # out of the first ' str_NbBRC_Plot_2b ' BRCs ' plotconfig.BRCrange_To_Plot])
	caxis([1. rhomax]);
	hc=colorbar;
	title([ plotconfig.scenario ' : L2b obs. scattering ratio [Cloudy]'],'Interpreter','none')
	
end % ibrc

if (isfield(plotconfig,'printext'))
	fic=[ plotconfig.scenario '_rho2b' plotconfig.printext ]
	set(gcf,'PaperUnits','centimeters','PaperPosition',[2. 4. 18. 24.])
	print(fic,plotconfig.printcmd);
end
if (isfield(plotconfig,'figext'))
	fic=[ plotconfig.scenario '_rho2b' plotconfig.figext  ]
	saveas(gcf,fic);
end

end % l2bpresent

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 2D plot of L2B classification map result
% and 2D plot of the result of the L1B_Input_Screening by the L2BP
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if (plotconfig.plot_rayuse & l2bpresent)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% for Rayleigh
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
ifig=ifig+1;
figure(ifig)
clf
NbBIGPlot=1;
%inpscrmax=max(max(rayinpscr2b));
inpscrmax=3;

for ibrc=1:NbBIGPlot
	stribrc=int2str(ibrc);
	subplot(2,NbBIGPlot,ibrc)
	x=1:N_meas*NbBRC_Plot_2b/NbBIGPlot;
%	if (l1bpresent)
%		y=altrayavg1b;
%		ylab=ylab_alt;
%	else
		y=[25:-1:1];
		ylab=ylab_bin;
%	end
	zmin=min(y);
	zmax=max(y);
	pretabplot=mapray2b(24:-1:1,(ibrc-1)*max(x)+1:ibrc*max(x));
	x=[x,x(end)+x(end)-x(end-1)];
	tabplot=[[pretabplot,pretabplot(:,end)];[pretabplot(end,:),pretabplot(end,end)]];
	h=pcolor(x,y(25:-1:1),tabplot);
	set(h,'EdgeAlpha',0)
	set(gca,'XLim',[1 max(x)]);
	set(gca,'YLim',[zmin zmax]);
	for imeas=1:N_meas*NbBRC_Plot_2b
		if (imeas > 1 & mod(imeas,N_meas) == 0)
			hline=line([imeas+1,imeas+1],[zmin,zmax]);
			set(hline,'Color',[0 0 0])
		end
	end
	for ibin=2:24
		hline=line([x(1) x(end)], [y(ibin) y(ibin)]);
		set(hline,'Color',[0 0 0])
	end
	if (ibrc==1)
		ylabel(ylab)
	end
	xlabel(['Measurement # out of the first ' str_NbBRC_Plot_2b ' BRCs ' plotconfig.BRCrange_To_Plot])
	caxis([-0.1 2.1]);
	hbar=colorbar;
	set(hbar,'YTick',[0,1,2]);
	set(hbar,'YTickLabel',['Not used';'Cloudy  ';'Clear   '])
	title([ plotconfig.scenario ' : L2b classification map Rayleigh'],'Interpreter','none')
	
	stribrc=int2str(ibrc);
	subplot(2,NbBIGPlot,NbBIGPlot+ibrc)
	x=1:N_meas*NbBRC_Plot_2b/NbBIGPlot;
%	if (l1bpresent)
%		y=altrayavg1b;
%		ylab=ylab_alt;
%	else
		y=[25:-1:1];
		ylab=ylab_bin;
%	end
	pretabplot=rayinpscr2b(24:-1:1,(ibrc-1)*max(x)+1:ibrc*max(x));
	x=[x,x(end)+x(end)-x(end-1)];
	tabplot=[[pretabplot,pretabplot(:,end)];[pretabplot(end,:),pretabplot(end,end)]];
	h=pcolor(x,y(25:-1:1),tabplot);
	set(h,'EdgeAlpha',0)
	set(gca,'XLim',[1 max(x)]);
	set(gca,'YLim',[zmin zmax]);
	for imeas=1:N_meas*NbBRC_Plot_2b
		if (imeas > 1 & mod(imeas,N_meas) == 0)
			hline=line([imeas+1,imeas+1],[zmin,zmax]);
			set(hline,'Color',[0 0 0])
		end
	end
	for ibin=2:24
		hline=line([x(1) x(end)], [y(ibin) y(ibin)]);
		set(hline,'Color',[0 0 0])
	end
	if (ibrc==1)
		ylabel(ylab)
	end
	xlabel(['Measurement # out of the first ' str_NbBRC_Plot_2b ' BRCs ' plotconfig.BRCrange_To_Plot])
	caxis([-0.1 inpscrmax+0.1]);
	hbar=colorbar;
	yticks=1:inpscrmax+1;
	yticks=[];
	yticks=0:inpscrmax;
	set(hbar,'YTick',yticks);
	set(hbar,'YTickLabel',inputscreeningflag)
	title([ plotconfig.scenario ' : L1b input screening Rayleigh'],'Interpreter','none')
end % ibrc

if (isfield(plotconfig,'printext'))
	fic=[ plotconfig.scenario '_rayuse' plotconfig.printext ]
	set(gcf,'PaperUnits','centimeters','PaperPosition',[2. 4. 18. 24.])
	print(fic,plotconfig.printcmd);
end
if (isfield(plotconfig,'figext'))
	fic=[ plotconfig.scenario '_rayuse' plotconfig.figext  ]
	saveas(gcf,fic);
end

end % l2bpresent

if (plotconfig.plot_mieuse & l2bpresent)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% for Mie
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
ifig=ifig+1;
figure(ifig)
clf
NbBIGPlot=1;
%inpscrmax=max(max(mieinpscr2b));
inpscrmax=3;

for ibrc=1:NbBIGPlot
	stribrc=int2str(ibrc);
	subplot(2,NbBIGPlot,ibrc)
	x=1:N_meas*NbBRC_Plot_2b/NbBIGPlot;
	if (l1bpresent)
		y=altmieavg1b;
		ylab=ylab_alt;
	else
		y=[25:-1:1];
		ylab=ylab_bin;
	end
	zmin=min(y);
	zmax=max(y);
	pretabplot=mapmie2b(24:-1:1,(ibrc-1)*max(x)+1:ibrc*max(x));
	x=[x,x(end)+x(end)-x(end-1)];
	tabplot=[[pretabplot,pretabplot(:,end)];[pretabplot(end,:),pretabplot(end,end)]];
	h=pcolor(x,y(25:-1:1),tabplot);
	set(h,'EdgeAlpha',0)
	set(gca,'XLim',[1 max(x)]);
	set(gca,'YLim',[zmin zmax]);
	for imeas=1:N_meas*NbBRC_Plot_2b
		if (imeas > 1 & mod(imeas,N_meas) == 0)
			hline=line([imeas+1,imeas+1],[zmin,zmax]);
			set(hline,'Color',[0 0 0])
		end
	end
	for ibin=2:24
		hline=line([x(1) x(end)], [y(ibin) y(ibin)]);
		set(hline,'Color',[0 0 0])
	end
	if (ibrc==1)
		ylabel(ylab);
	end
	xlabel(['Measurement # out of the first ' str_NbBRC_Plot_2b ' BRCs ' plotconfig.BRCrange_To_Plot])
	caxis([-0.1 2.1]);
	hbar=colorbar;
	set(hbar,'YTick',[0,1,2]);
	set(hbar,'YTickLabel',['Not used';'Cloudy  ';'Clear   '])
	title([ plotconfig.scenario ' : L2b classification map Mie'],'Interpreter','none')
	
	stribrc=int2str(ibrc);
	subplot(2,NbBIGPlot,NbBIGPlot+ibrc)
	x=1:N_meas*NbBRC_Plot_2b/NbBIGPlot;
	if (l1bpresent)
		y=altmieavg1b;
		ylab=ylab_alt;
	else
		y=[25:-1:1];
		ylab=ylab_bin;
	end
	pretabplot=mieinpscr2b(24:-1:1,(ibrc-1)*max(x)+1:ibrc*max(x));
	x=[x,x(end)+x(end)-x(end-1)];
	tabplot=[[pretabplot,pretabplot(:,end)];[pretabplot(end,:),pretabplot(end,end)]];
	h=pcolor(x,y(25:-1:1),tabplot);
	set(h,'EdgeAlpha',0)
	set(gca,'XLim',[1 max(x)]);
	set(gca,'YLim',[zmin zmax]);
	for imeas=1:N_meas*NbBRC_Plot_2b
		if (imeas > 1 & mod(imeas,N_meas) == 0)
			hline=line([imeas+1,imeas+1],[zmin,zmax]);
			set(hline,'Color',[0 0 0])
		end
	end
	for ibin=2:24
		hline=line([x(1) x(end)], [y(ibin) y(ibin)]);
		set(hline,'Color',[0 0 0])
	end
	if (ibrc==1)
		ylabel(ylab);
	end
	xlabel(['Measurement # out of the first ' str_NbBRC_Plot_2b ' BRCs ' plotconfig.BRCrange_To_Plot])
	caxis([-0.1 inpscrmax+.1]);
	hbar=colorbar;
	yticks=1:inpscrmax+1;
	yticks=[];
	yticks=0:inpscrmax;
	set(hbar,'YTick',yticks);
	set(hbar,'YTickLabel',inputscreeningflag);
	title([ plotconfig.scenario ' : L1b input screening Mie'],'Interpreter','none')
end % ibrc

if (isfield(plotconfig,'printext'))
	fic=[ plotconfig.scenario '_mieuse' plotconfig.printext ]
	set(gcf,'PaperUnits','centimeters','PaperPosition',[2. 4. 18. 24.])
	print(fic,plotconfig.printcmd);
end
if (isfield(plotconfig,'figext'))
	fic=[ plotconfig.scenario '_mieuse' plotconfig.figext  ]
	saveas(gcf,fic);
end

end % l2bpresent

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 2D plot of L1B HLOS wind at measurement level
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if (plotconfig.plot_1bwind & l1bpresent)
ifig=ifig+1;
figure(ifig)
clf
w1bmax=mean(ray1bwindmeas(~isnan(ray1bwindmeas)))+4.*std(ray1bwindmeas(~isnan(ray1bwindmeas)));
w1bmin=mean(ray1bwindmeas(~isnan(ray1bwindmeas)))-4.*std(ray1bwindmeas(~isnan(ray1bwindmeas)));
NbBIGPlot=1;
zmin=min(altrayavg1b);
zmax=max(altrayavg1b);
for ibrc=1:NbBIGPlot
	stribrc=int2str(ibrc);
	subplot(2,NbBIGPlot,ibrc)
	x=1:N_meas*NbBRC_Plot_1b/NbBIGPlot;
	y=altrayavg1b;
	pretabplot=ray1bwindmeas((ibrc-1)*max(x)+1:ibrc*max(x),24:-1:1)';
	x=[x,x(end)+x(end)-x(end-1)];
	tabplot=[[pretabplot,pretabplot(:,end)];[pretabplot(end,:),pretabplot(end,end)]];
	h=pcolor(x,y(25:-1:1),tabplot);
	set(h,'EdgeAlpha',0)
	set(gca,'XLim',[1 max(x)]);
	set(gca,'YLim',[zmin zmax]);
	for imeas=1:N_meas*NbBRC_Plot_1b
		if (imeas > 1 & mod(imeas,N_meas) == 0)
			hline=line([imeas+1,imeas+1],[zmin,zmax]);
			set(hline,'Color',[0 0 0])
		end
	end
	for ibin=2:24
		hline=line([x(1) x(end)], [y(ibin) y(ibin)]);
		set(hline,'Color',[0 0 0])
	end
	if (ibrc==1)
		ylabel(ylab_alt);
	end
	xlabel(['Measurement # out of the first ' str_NbBRC_Plot_1b ' BRCs ' plotconfig.BRCrange_To_Plot])
	caxis([w1bmin w1bmax]);
	hc=colorbar;
	title([ plotconfig.scenario ' : L1b meas. wind Rayleigh'],'Interpreter','none')
end % ibrc

for ibrc=1:NbBIGPlot
	stribrc=int2str(ibrc);
	subplot(2,NbBIGPlot,NbBIGPlot+ibrc)
	x=1:N_meas*NbBRC_Plot_1b/NbBIGPlot;
	y=altmieavg1b;
	pretabplot=mie1bwindmeas((ibrc-1)*max(x)+1:ibrc*max(x),24:-1:1)';
	x=[x,x(end)+x(end)-x(end-1)];
	tabplot=[[pretabplot,pretabplot(:,end)];[pretabplot(end,:),pretabplot(end,end)]];
	h=pcolor(x,y(25:-1:1),tabplot);
	set(h,'EdgeAlpha',0)
	set(gca,'XLim',[1 max(x)]);
	set(gca,'YLim',[zmin zmax]);
	for imeas=1:N_meas*NbBRC_Plot_1b
		if (imeas > 1 & mod(imeas,N_meas) == 0)
			hline=line([imeas+1,imeas+1],[zmin,zmax]);
			set(hline,'Color',[0 0 0])
		end
	end
	for ibin=2:24
		hline=line([x(1) x(end)], [y(ibin) y(ibin)]);
		set(hline,'Color',[0 0 0])
	end
	if (ibrc==1)
		ylabel(ylab_alt);
	end
	xlabel(['Measurement # out of the first ' str_NbBRC_Plot_1b ' BRCs ' plotconfig.BRCrange_To_Plot])
	caxis([w1bmin w1bmax]);
	hc=colorbar;
	title([ plotconfig.scenario ' : L1b meas. wind Mie'],'Interpreter','none')
end % ibrc

if (isfield(plotconfig,'printext'))
	fic=[ plotconfig.scenario '_1bwind' plotconfig.printext ]
	set(gcf,'PaperUnits','centimeters','PaperPosition',[2. 4. 18. 24.])
	print(fic,plotconfig.printcmd);
end
if (isfield(plotconfig,'figext'))
	fic=[ plotconfig.scenario '_1bwind' plotconfig.figext  ]
	saveas(gcf,fic);
end

end % plotconfig.plot_1bwind & l1bpresent

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 1D plot of HLOS winds:
% - the min/mean/max E2S input winds for the observation
% - the L1B observation wind, for Mie and Rayleigh
% - the L2B observation wind, for clear and cloudy Mie and Rayleigh,
%   with error bars indicating the L2B error estimate
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if (plotconfig.plot_brc & (e2spresent | l1bpresent | l2bpresent))
nbp=strread(num2str(sqrt(NbBRC_Plot)),'%4.0f',1);
if (nbp^2 ~= NbBRC_Plot)
	nbp=nbp+1;
end
for ibrc=1:NbBRC_Plot
	ifig=ifig+1;
	figure(ifig)
	clf
	hold on
	nray=[];
	nrayout=[];
	nmie=[];
	nmieout=[];
	ip=0;
	leg=[];
	ip_mis=0;
	leg_mis=[];
	if (e2spresent)
		ne2s=[];
		ne2sout=[];
		zmine2s=min(alte2sobs(ibrc,:));
		e2spresent_ibrc=ibrc <= NbBRC_Plot_e2s;
	else
		zmine2s=[];
		e2spresent_ibrc=false;
	end % e2spresent
	if (l1bpresent)
		l1bpresent_ibrc=ibrc <= NbBRC_Plot_1b;
		if (l1bpresent_ibrc)
			zminl1b=[min(min(altray1bobs(ibrc,:))) min(min(altmie1bobs(ibrc,:)))];
			zmaxl1b=[max(max(altray1bobs(ibrc,:))) max(max(altmie1bobs(ibrc,:)))];
		end
	else
		zminl1b=[];
		zmaxl1b=[];
		l1bpresent_ibrc=false;
	end % l1bpresent
	if (l2bpresent)
		l2bpresent_ibrc=ibrc <= NbBRC_Plot_2b;
		if (l2bpresent_ibrc)
			zminl2b=[min(min(altray2b(ibrc,:,:))) min(min(altmie2b(ibrc,:,:)))];
			zmaxl2b=[max(max(altray2b(ibrc,:,:))) max(max(altmie2b(ibrc,:,:)))];
		end
	else
		zminl2b=[];
		zmaxl2b=[];
		l2bpresent_ibrc=false;
	end % l2bpresent
	zmin=min([zmine2s zminl2b zminl1b])/1.e3-.2;
	zmax=max([        zmaxl2b zmaxl1b])/1.e3+.2;
	xmin=-100;
	xmax=100;
	
	if (e2spresent_ibrc)
		tab=hloswinde2sobs(ibrc,1:nbine2s);
		ne2s=int2str(length(find(~isnan(tab))));
		ne2sout=int2str(length(find(tab < xmin | tab > xmax)));
		plot(tab,alte2sobs(ibrc,1:nbine2s)/1000., [e2ssym e2scol]);
		sexcl='';
		if (length(find(~isnan(tab))) ~= 0)
			ip=ip+1;
			if (ne2sout~='0')
				sexcl=[' excld:' ne2sout];
			end
			leg{ip}=['E2S input (' ne2s sexcl ')' ];
		else
			ip_mis=ip_mis+1;
			leg_mis{ip_mis}=['E2S input: NO VALID DATA' ];
		end
	else
		ip_mis=ip_mis+1;
		leg_mis{ip_mis}=['E2S input: NO DATA AVAILABLE' ];
	end % e2spresent_ibrc
	
	if (l2bpresent_ibrc)
	for iobs=1:nrayobs2b(ibrc)
	    tab=raywind2b(ibrc,1:24,iobs);
	    nray{iobs}=int2str(length(find(~isnan(tab))));
	    nrayout{iobs}=int2str(length(find(tab/100. < xmin | tab/100. > xmax)));
	    plot(tab/100,altray2b(ibrc,1:24,iobs)/1000., [raysymproftype colproftype(rayproftype2b(ibrc,iobs))],'MarkerSize',8)
	    sexcl='';
	    if (length(find(~isnan(tab))) ~= 0)
		    ip=ip+1;
		    if (nrayout{iobs}~='0')
			sexcl=[' excld:' nrayout{iobs}];
		    end
		    leg{ip}=['2b Rayleigh ' chproftype(rayproftype2b(ibrc,iobs),:) ' (' nray{iobs} sexcl ')' ];
	    else
		    ip_mis=ip_mis+1;
		    leg_mis{ip_mis}=['2b Rayleigh ' chproftype(rayproftype2b(ibrc,iobs),:) ': NO VALID DATA' ];
	    end
	end
	else
		ip_mis=ip_mis+1;
		leg_mis{ip_mis}=['2b Retrievals: NO DATA AVAILABLE'];
	end % l2bpresent_ibrc

	if (l1bpresent_ibrc)
		tab=ray1bwindobs(ibrc,1:24);
		nray1b=int2str(length(find(~isnan(tab))));
		nray1bout=int2str(length(find(tab < xmin | tab > xmax)));
		plot(tab,altray1bobs(ibrc,1:24)/1000,'+m','MarkerSize',8)
		sexcl='';
		if (length(find(~isnan(tab))) ~= 0)
			ip=ip+1;
			if (nray1bout~='0')
				sexcl=[' excld: ' nray1bout];
			end
			leg{ip}=['1b Rayleigh obs (' nray1b sexcl ')' ];
		else
			ip_mis=ip_mis+1;
			leg_mis{ip_mis}=['1b Rayleigh obs: NO VALID DATA' ];
		end
	else
		ip_mis=ip_mis+1;
		leg_mis{ip_mis}=['1b Retrievals: NO DATA AVAILABLE'];
	end % l1bpresent_ibrc

	if (l2bpresent_ibrc)
	for iobs=1:nmieobs2b(ibrc)
	    tab=miewind2b(ibrc,1:24,iobs);
	    nmie{iobs}=int2str(length(find(~isnan(tab))));
	    nmieout{iobs}=int2str(length(find(tab/100. < xmin | tab/100. > xmax)));
	    plot(tab/100,altmie2b(ibrc,1:24,iobs)/1000., [miesymproftype colproftype(mieproftype2b(ibrc,iobs))],'MarkerSize',8)
	    sexcl='';
	    if (length(find(~isnan(tab))) ~= 0)
		    ip=ip+1;
		    if (nmieout{iobs}~='0')
			    sexcl=[' excld: ' nmieout{iobs}];
		    end
		    leg{ip}=['2b Mie ' chproftype(mieproftype2b(ibrc,iobs),:) ' (' nmie{iobs} sexcl ')' ];
	    else
		    ip_mis=ip_mis+1;
		    leg_mis{ip_mis}=['2b Mie ' chproftype(mieproftype2b(ibrc,iobs),:) ': NO VALID DATA'];
	    end
	end
	end % l2bpresent_ibrc
	
	if (l1bpresent_ibrc)
		tab=mie1bwindobs(ibrc,1:24);
		nmie1b=int2str(length(find(~isnan(tab))));
		nmie1bout=int2str(length(find(tab < xmin | tab > xmax)));
		plot(tab,altmie1bobs(ibrc,1:24)/1000,'.k','MarkerSize',8)
		sexcl='';
		if (length(find(~isnan(tab))) ~= 0)
			ip=ip+1;
			if (nmie1bout~='0')
				sexcl=[' excld: ' nmie1bout];
			end
			leg{ip}=['1b Mie obs (' nmie1b sexcl ')' ];
		else
			ip_mis=ip_mis+1;
			leg_mis{ip_mis}=['1b Mie obs: NO VALID DATA'];
		end
	end % l1bpresent_ibrc
	
	set(gca,'XLim',[xmin xmax])
	set(gca,'YLim',[zmin zmax])
	if (ip>0)
		hleg=legend(leg,'Location','Best');
		leg_pos=get(hleg,'Position');
		legend('boxoff')
	end % ip>0
	if (ip_mis > 0)
		htext=text(1,1,leg_mis);
		set(htext,'Units','normalized');
		set(htext,'Position',[leg_pos(1),leg_pos(2)-leg_pos(4)/(ip+0.1),0]);
	end % ip_mis>0
	
	if (e2spresent_ibrc)
	% plot hlos e2s input min/max
		tab=hloswinde2sobsmin(ibrc,1:nbine2s);
		plot(tab,alte2sobs(ibrc,1:nbine2s)/1000., [e2ssym e2scol]);  %,'MarkerSize',3)
		tab=hloswinde2sobsmax(ibrc,1:nbine2s);
		plot(tab,alte2sobs(ibrc,1:nbine2s)/1000., [e2ssym e2scol]);  %,'MarkerSize',3)
	end % e2spresent_ibrc

	% plot 2b and 1b hlos error estimates
	for ibin=1:24
	   if (l2bpresent_ibrc)
	       for iobs=1:nrayobs2b(ibrc)
	           if (~isnan(raywind2b(ibrc,ibin,iobs)))
	               plot([raywind2b(ibrc,ibin,iobs)/100.-rayhloserror2b(ibrc,ibin,iobs) ...
	                     raywind2b(ibrc,ibin,iobs)/100.+rayhloserror2b(ibrc,ibin,iobs)], ...
			     [altray2b(ibrc,ibin,iobs)/1000. altray2b(ibrc,ibin,iobs)/1000.], ...
			     ['-' colproftype(rayproftype2b(ibrc,iobs))])
	           end
	       end
	       for iobs=1:nmieobs2b(ibrc)
	           if (~isnan(miewind2b(ibrc,ibin,iobs)))
	               plot([miewind2b(ibrc,ibin,iobs)/100.-miehloserror2b(ibrc,ibin,iobs) ...
	                     miewind2b(ibrc,ibin,iobs)/100.+miehloserror2b(ibrc,ibin,iobs)], ...
			     [altmie2b(ibrc,ibin,iobs)/1000. altmie2b(ibrc,ibin,iobs)/1000.], ...
			     ['-' colproftype(mieproftype2b(ibrc,iobs))])
	           end
	       end
	   end % l2bpresent_ibrc
	   if (l1bpresent_ibrc)
	       if (~isnan(ray1bwindobs(ibrc,ibin)))
	           plot([ray1bwindobs(ibrc,ibin)-ray1bhloserrorobs(ibrc,ibin) ...
	                 ray1bwindobs(ibrc,ibin)+ray1bhloserrorobs(ibrc,ibin)], ...
		         [altray1bobs(ibrc,ibin)/1000. altray1bobs(ibrc,ibin)/1000.], ...
		         ['-m'] )
	       end
	       if (~isnan(mie1bwindobs(ibrc,ibin)))
	           plot([mie1bwindobs(ibrc,ibin)-mie1bhloserrorobs(ibrc,ibin) ...
	                 mie1bwindobs(ibrc,ibin)+mie1bhloserrorobs(ibrc,ibin)], ...
		         [altmie1bobs(ibrc,ibin)/1000. altmie1bobs(ibrc,ibin)/1000.], ...
		         ['-k'] )
	       end
	   end % l1bpresent_ibrc
	end
	
	xlabel('HLOS wind [m/s]')
	ylabel(ylab_alt);
	jbrc=ibrcrange(ibrc);
	title(['BRC #' int2str(jbrc) ])
	if (isfield(plotconfig,'printext'))
		fic=[ plotconfig.scenario '_brc' int2str(jbrc) plotconfig.printext ]
		print(fic,plotconfig.printcmd);
	end
	if (isfield(plotconfig,'figext'))
		fic=[ plotconfig.scenario '_brc' int2str(jbrc) plotconfig.figext  ]
		saveas(gcf,fic);
	end

end % for ibrc

end % plotobswind & (e2spresent | l1bpresent | l2bpresent)

end % plotdata

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

