module Select_And_Weigh_Measurements
  !  #[ Description:

  !     A module to
  !     -select measurements by assigning a weight to them.
  !      Using this method, measurements with problems indicated
  !      by the Bin_QC flag get weight zero and are discarded
  !      by all following routines.
  !     -later we may also vary the weights according to the SNR
  !      ratio reported for each measurement
  !     -note that if any normalisation is needed of these weights
  !      this normalisation should be done just before accumulating
  !      the measurements into an observation.
  !      Reasons for doing it like this are:
  !      ==>normalisation should be done separately for each
  !         observation profile, and the assignment to observations
  !         is only done in Construct_Mie/Rayleigh_Obs which is 
  !         called after this selection module)
  !      ==>we may decide to add switches to allow combining several
  !         classes into one observation, but with different weight,
  !         for example: classified_as_cloud and classified_as_maybe_cloud
  !         (the latter is not yet defined). This combining would also
  !         be done in Construct_Mie/Rayleigh_Obs, so again doing a
  !         normalisation is not possible in this module
  !
  ! created:      04-Oct-2007, by: J. de Kloe (KNMI)
  ! 
  ! Modifications: 
  !   04-Oct-2007 J. de Kloe, initial version, using code removed
  !                           from the Classification module
  !  #]
  !  #[ Modules used:
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE L2BC_DataStructure, only: L2BC_BRC_DataType
  USE Meas_Map, only: Meas_Map_type
  USE Working_DataStructure, only: WorkingDataType
  USE L2B_Parameters, only: no_classification_possible, &
                            classified_as_cloud, classified_as_no_cloud, &
                            not_yet_classified
  USE L2B_PCD_ADS, only: L1B_Rayl_Bin_OK, L1B_Mie_Bin_OK
  !support modules
  USE Numerics, only: r8_
  USE ErrorHandler, only: no_error, error_programming
  USE Logging, only: logmsg,log_debug,log_info,log_progress,log_warn,log_error

  !  #]
  !  #[ Parameter/Variable/Type declarations
  implicit none
  !  #]
  !  #[ Module interface

  ! all routines/functions are public unless stated otherwise below
  !private 
  !  #]
contains
  !---------------------------------------------
  ! public subroutines
  !---------------------------------------------
  subroutine Select_And_Weigh_Rayleigh_Meas(WorkingData, L2B, error_flag)
    !  #[
    TYPE(WorkingDataType),   intent(inout) :: WorkingData ! input
    TYPE(L2BC_BRC_DataType), intent(inout) :: L2B         ! input/output
    integer,                 intent(out)   :: error_flag  ! output

    ! local variables
    TYPE(Meas_Map_type),    pointer :: Map
    integer   :: r,m
    real(r8_) :: weight

    ! init
    error_flag = no_error
    ! point to the map for shorter notation
    Map => L2B%Rayleigh_MDS%Map_of_L1B_Measurements_Used

    ! now fill  the weights used for creating the observations
    measloop: DO m=1, WorkingData%nr_of_measurements
       rangegateloop: DO r=1,WorkingData%RayleighMeas(1)%nr_of_range_gates
          !  #[

          ! All measurements get a weight of 1000 for now.
          ! It is foreseen that these weights might be adjusted in
          ! future versions of the L2B Processor, e.g depending on
          ! the SNR of the different measurements.
          weight = 1000._r8_

          ! do not use measurements for which classification
          ! was not possible
          IF ( WorkingData%RayleighMeas(m)%RangeGate(r)%class .eq. &
               no_classification_possible) weight = 0._r8_

          ! a sanity check
          IF ( WorkingData%RayleighMeas(m)%RangeGate(r)%class .eq. &
               not_yet_classified) THEN
             ! this should never happen !
             call logmsg(log_error,"Error in Select_And_Weigh_Rayleigh_Meas: "//&
                  "no classification seems to be performed for this "//&
                  "measurement. This should never happen !")
             error_flag = error_programming
             return
          END IF

          ! reject all measurements that have Bin_QC not OK
          IF ( L2B%PCD_ADS%L1B_Input_Screening%Rayl_Meas(m)%&
                   Bin_Screening(r)%Bin_QC .ne. L1B_Rayl_Bin_OK) weight = 0._r8_

          ! copy the weight to the appropriate the L2B element 
          ! see sec. 5.5, table 17, p. 21, of TN2.1, vers. 1.2, 06-09-2005
          ! store for each Bin of a L1B Measurement to how much it 
          ! contributed to the L2B Observation to which it will be assigned, 
          ! This is the weight, and is a value between 0-1000.
          ! (or 0 if it was not assigned)
          Map%Row(m)%Bin(r)%Weight = floor(weight)
          ! store also the real value, since we cannot be sure yet if 
          ! we can afford the precision loss caused by the above 
          ! floor() statement
          WorkingData%RayleighMeas(m)%RangeGate(r)%weight = weight

          !  #]
       END DO rangegateloop
    END DO measloop

  end subroutine Select_And_Weigh_Rayleigh_Meas
    !  #]
  subroutine Select_And_Weigh_Mie_Meas(WorkingData, L2B, error_flag)
    !  #[

    TYPE(WorkingDataType),   intent(inout) :: WorkingData ! input
    TYPE(L2BC_BRC_DataType), intent(inout) :: L2B         ! input/output
    integer,                 intent(out)   :: error_flag  ! output

    ! local variables
    TYPE(Meas_Map_type),    pointer :: Map
    integer   :: r,m
    real(r8_) :: weight

    ! init
    error_flag = no_error
    ! point to the map for shorter notation
    Map => L2B%Mie_MDS%Map_of_L1B_Measurements_Used
    
    ! now fill  the weights used for creating the observations
    measloop: DO m=1, WorkingData%nr_of_measurements
       rangegateloop: DO r=1,WorkingData%MieMeas(1)%nr_of_range_gates
          !  #[

          ! All measurements get a weight of 1000 for now.
          ! It is foreseen that these weights might be adjusted in
          ! future versions of the L2B Processor, e.g depending on
          ! the SNR of the different measurements.
          weight = 1000._r8_

          ! do not use measurements for which classification
          ! was not possible
          IF ( WorkingData%MieMeas(m)%RangeGate(r)%class .eq. &
               no_classification_possible) weight = 0._r8_

          ! a sanity check
          IF ( WorkingData%MieMeas(m)%RangeGate(r)%class .eq. &
               not_yet_classified) THEN
             ! this should never happen !
             call logmsg(log_error,"Error in Select_And_Weigh_Mie_Meas: "//&
                  "no classification seems to be performed for this "//&
                  "measurement. This should never happen !")
             error_flag = error_programming
             return
          END IF

          ! reject all measurements that have Bin_QC not OK
          IF ( L2B%PCD_ADS%L1B_Input_Screening%Mie_Meas(m)%&
                   Bin_Screening(r)%Bin_QC .ne. L1B_Mie_Bin_OK) weight = 0._r8_

          ! copy the weight to the appropriate the L2B element 
          ! see sec. 5.5, table 17, p. 21, of TN2.1, vers. 1.2, 06-09-2005
          ! store for each Bin of a L1B Measurement to how much it 
          ! contributed to the L2B Observation to which it will be assigned, 
          ! This is the weight, and is a value between 0-1000.
          ! (or 0 if it was not assigned)
          Map%Row(m)%Bin(r)%Weight = floor(weight)
          ! store also the real value, since we cannot be sure yet if 
          ! we can afford the precision loss caused by the above 
          ! floor() statement
          WorkingData%MieMeas(m)%RangeGate(r)%weight = weight

          !  #]
       END DO rangegateloop
    END DO measloop

  end subroutine Select_And_Weigh_Mie_Meas

    !  #]
  !---------------------------------------------
end module Select_And_Weigh_Measurements
