program TestMieResponse
  !  #[ Description
  !---------------------------------------------------
  ! purpose: a test program for the MieResponse module
  !          
  !    Written by:  Jos de Kloe.
  !    created:     April-2006
  !---------------------------------------------------
  !  #]
  !  #[ Modifications:
  !       23-May-2006  D. Tan     Null use of ParKind1 module
  !       24-May-2006  D. Tan     Upgrade datatypes for full Mie core algorithm
  !       25-May-2006  D. Tan     Simple input params, tobs and accd values
  !       14-Feb-2007  P. Poli    Remove tobs input
  !       03-Dec-2007  P. Poli    Re-introduce tobs input
  !       28-Jan-2008  J. de Kloe added optional input of test datafile
  !  #]
  !  #[ modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE MieResponse, only: ComputeMieResponse, numACCDCols, numUsefulACCDCols
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, VerbosityReport
  USE ErrorHandler, only: no_error, error_programming, &
                          error_cmdline_options, error_reading_file, &
                          program_abort
  USE EE_CFI_Datatypes, only: IntAs, IntAuc, FAdoxy
  USE Compiler_Features, only: iargc_aeolus,getarg_aeolus
  USE LunManager, only: get_lun, free_lun
  !  #]
  !  #[ variables
  implicit none 

  ! variables needed for commandline handling
  integer            :: nr_of_arguments, fileunit, ios
  character(len=256) :: testfile_name
  character(len=256) :: case_text
  real(FAdoxy), dimension(1:numACCDCols) :: accdInt

  ! variables needed to interface to ComputeMieResponse
  integer        :: ChoosenVerbosity                      ! input
  real(FAdoxy)   :: StartFWHM,ResErrThresh,FWHMUpThresh   ! input
  real(FAdoxy)   :: FWHMLoThresh, PeakHeightUpThresh      ! input
  real(FAdoxy)   :: PeakHeightLoThresh,PeakLocThresh      ! input
  real(FAdoxy)   :: NonLinOptThresh                       ! input
  integer(IntAs) :: MaxItLorFit                           ! input
  integer(IntAs) :: MaxItNonLinOpt,NumSpecSubSamp         ! input
  real(FAdoxy), dimension(1:numUsefulACCDCols) :: tobs    ! input
  real(FAdoxy), dimension(1:numACCDCols) :: accd          ! input
  logical        :: isMeas                                ! input
  real(FAdoxy)   :: PeakLoc,FWHM,Offset,OffsetSub, &
                    PeakHeight,ResErr ! output
  logical        :: ValFlag                               ! output
  integer(IntAs) :: NumIt                                 ! output
  integer(IntAs) :: error_flag                            ! output
  !  #]
  !  #[ algorithm settings
  print *,"Testing subroutine: ComputeMieResponse()"
  
  ChoosenVerbosity   = VerbosityReport
  StartFWHM          = 5.1_FAdoxy
  ResErrThresh       = 0.001_FAdoxy
  MaxItLorFit        = 30
  FWHMUpThresh       = 2.0_FAdoxy
  FWHMLoThresh       = 1.7_FAdoxy
  PeakHeightUpThresh = 1.6_FAdoxy
  PeakHeightLoThresh = 0.9_FAdoxy
  PeakLocThresh      = 1.0_FAdoxy
  NonLinOptThresh    = 0.01_FAdoxy
  MaxItNonLinOpt     = 20
  NumSpecSubSamp     = 5
  tobs(:)            = 1.05_FAdoxy

  ! this switch enables tripod-obscuration correction for measurements
  ! (it is not needed for internal reference pulses!)
  isMeas             = .false.
  !  #]
  !  #[ define test data
  nr_of_arguments = iargc_aeolus()
  IF (nr_of_arguments .gt. 1) THEN
     print *,"ERROR: wrong number of commandline arguments."
     print *,"nr_of_arguments = ",nr_of_arguments
     call PrintUsage
     call program_abort(error_cmdline_options,"TestMieResponse")
  END IF

  accd(:) =     0.0_FAdoxy
  case_text = "[undefined]"
  IF (nr_of_arguments .eq. 1) THEN
      call getarg_aeolus(1,testfile_name)

      fileunit = get_lun()
      open(unit=fileunit, file=testfile_name,status="old",&
           action="read",form="formatted",err=99)
      read(fileunit,"(A)",iostat=ios) case_text
      IF (ios .ne. 0) goto 98
      read(fileunit,*,iostat=ios) accdInt(:)
      IF (ios .ne. 0) goto 98

      close(unit=fileunit)
      call free_lun(fileunit)

      accd(:) = real(accdInt(:), FAdoxy)
  ELSE
     ! use the default test case
     accd(4) =  6000.0_FAdoxy
     accd(5) = 10000.0_FAdoxy
     accd(6) =  6000.0_FAdoxy
  END IF
  !  #]
  !  #[ do the test
  call ComputeMieResponse(ChoosenVerbosity,StartFWHM, &
                          ResErrThresh,MaxItLorFit, FWHMUpThresh, &
                          FWHMLoThresh, PeakHeightUpThresh, &
                          PeakHeightLoThresh, PeakLocThresh, &
                          NonLinOptThresh, MaxItNonLinOpt, &
                          NumSpecSubSamp, tobs, accd, PeakLoc, FWHM, &
                          Offset, OffsetSub, PeakHeight, ValFlag, ResErr, &
                          isMeas, NumIt, error_flag)

  print *,"results:"
  print *,"REALACC(12)"
  print *,"PeakLoc    = ",PeakLoc
  print *,"FWHM       = ",FWHM
  print *,"Offset     = ",Offset
  print *,"OffsetSub  = ",OffsetSub
  print *,"PeakHeight = ",PeakHeight
  print *,"ResErr     = ",ResErr
  print *,"ENDREALACC"
  print *,"ValFlag    = ",ValFlag
  print *,"NumIt      = ",NumIt
  print *,"error_flag = ",error_flag 

  print *,"test done"
  !  #]
  !  #[ error handlers
  goto 111

98 print *,"error reading data from datafile: "//trim(testfile_name)
   print *,"for testcase: "//trim(case_text)
   call program_abort(error_reading_file,"TestMieResponse")

99 print *,"error opening datafile: "//trim(testfile_name)  
   call program_abort(error_reading_file,"TestMieResponse")

111 continue
   !  #]
contains
  !-------------------
  subroutine PrintUsage
    print *,"Usage: TestMieResponse [optional: <testdatafile.dat>]"
  end subroutine PrintUsage
  !-------------------
end program TestMieResponse
