module MieResponse
  !  #[ Description
  !---------------------------------------------------
  ! purpose: a module to calculate the peak location,
  !          FWHM, peak height, and offset of one spectrum 
  !          stored in a single read out of one ACCD line  
  !          of the Mie channel.
  !          The data will be processed using a Downhill Simplex 
  !          method, returning the above mentioned parameters
  !          and additionally a residual error and a validity flag.
  !          Processing is described in AE.TN.DLR.ACS-L1B.100306.
  !
  !    Written by:  Dorit Huber, DLR
  !    created:     27-03-2006
  !    Modification:
  !       
  !       27-mar-2006 D. Huber,   initial version, only empty frame
  !       22-may-2006 D. Huber,   all subroutines coded
  !       24-may-2006 D. Tan,     Uncomment null use of ParKind1 module
  !       16-jun-2006 J. de Kloe, added REALACC/ENDREALACC commands
  !       10-jul-2006 P. Poli     added output OffsetSub
  !       17-Aug-2006 P. Poli     Removed obsolete variables and fixed precision
  !                               loss problems with explicit type conversions
  !       08-Nov-2006 P. Poli     Added printouts when error_flag/=0
  !       14-Feb-2007 P. Poli     Remove tobs correction (now done in L1bP)
  !       03-Dec-2007 P. Poli     Re-introduce tobs correction (not done in L1bP)
  !       17-Jan-2008 J. de Kloe  changed all real constants to FAdoxy type
  !       
  !---------------------------------------------------
  !  #]
  !  #[ Modules used
  !
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, VerbosityReport
  USE ErrorHandler, only: no_error, error_programming
  USE EE_CFI_Datatypes, only: IntAs, IntAuc, FAdoxy
  !  #]
  !  #[ module parameters and variables
  implicit none

  ! module parameters
  real(FAdoxy),    parameter :: alpha =  1._FAdoxy
  real(FAdoxy),    parameter :: beta =  0.5_FAdoxy
  real(FAdoxy),    parameter :: gamma =  2._FAdoxy
  integer(IntAuc), parameter :: numACCDCols = 20
  integer(IntAuc), parameter :: numUsefulACCDCols = 16
  integer(IntAuc), parameter :: spectrometer_index_offset = 2
  real(FAdoxy),    parameter :: ZeroEpsilon = 0.0000000000001_FAdoxy

  ! module variables
  real(FAdoxy), save, dimension(1:numACCDCols) :: accdLocal
  real(FAdoxy), save                           :: miny = 0.0_FAdoxy
  real(FAdoxy), save                           :: maxy
  integer(IntAuc), save, dimension(1)          :: maxj
  real(FAdoxy), save, dimension(1)             :: com
  real(FAdoxy), save                           :: peakHeightLocal, offsetLocal 
                    
  !  #]
contains 
  !-----------------------------------
  subroutine ComputeMieResponse(ChoosenVerbosity,StartFWHM, &
                                ResErrThresh,MaxItLorFit, FWHMUpThresh, &
                                FWHMLoThresh, PeakHeightUpThresh, &
                                PeakHeightLoThresh, PeakLocThresh, &
                                NonLinOptThresh, MaxItNonLinOpt, &
                                NumSpecSubSamp, tobs, accd, PeakLoc, FWHM, &
                                Offset, OffsetSub, &
                                PeakHeight, ValFlag, ResErr, &
                                isMeas, NumIt, error_flag)
    !  #[

    ! A routine to process a single accd line in 8 steps as described
    ! in AE.TN.DLR.ACS-L1B.100306. 
    
    ! interface
    integer,        intent(in) :: ChoosenVerbosity                       ! input
    real(FAdoxy),   intent(in) :: StartFWHM,ResErrThresh,FWHMUpThresh    ! input
    real(FAdoxy),   intent(in) :: FWHMLoThresh, PeakHeightUpThresh       ! input
    real(FAdoxy),   intent(in) :: PeakHeightLoThresh,PeakLocThresh       ! input
    real(FAdoxy),   intent(in) :: NonLinOptThresh                        ! input
    integer(IntAs), intent(in) :: MaxItLorFit                            ! input
    integer(IntAs), intent(in) :: MaxItNonLinOpt,NumSpecSubSamp          ! input
    real(FAdoxy), intent(in), dimension(1:numUsefulACCDCols) :: tobs     ! input
    real(FAdoxy), intent(in), dimension(1:numACCDCols) :: accd           ! input
    logical,        intent(in) :: isMeas                                 ! input
    real(FAdoxy),   intent(out) :: PeakLoc,FWHM,Offset,OffsetSub, &
                                   PeakHeight,ResErr                     ! output
    logical,        intent(out) :: ValFlag                               ! output
    integer(IntAs), intent(out) :: NumIt                                 ! output
    integer(IntAs), intent(out) :: error_flag                            ! output

    ! local variables
    real(FAdoxy)  :: PeakDiff

    ! Check input
    IF (ChoosenVerbosity .ge. VerbosityReport) THEN
       print *, "Input accd line"
       print *, "REALACC(12)"
       print *, "accdLocal           = ", accd
       print *, "ENDREALACC"
    END IF    
    IF ( isNearZero(SUM(accd)) .OR. &
         isNearZero(SUM(tobs)) .OR. &
         (NumSpecSubSamp .LE. 0)) THEN
       print *,"ERROR:mieresponse: sum(accd) or sum(tobs) is near zero or NumSpecSubSamp<0"
       error_flag = error_programming
       return
    END IF
    
    ValFlag = .false.

    ! step 1
    call OffsetSubtraction(accd,OffsetSub)
    ! Check step 1
    IF (ChoosenVerbosity .ge. VerbosityReport) THEN
       print *, "step 1"
       print *, "REALACC(12)"
       print *, "accdLocal           = ", accdLocal
       print *, "ENDREALACC"
    END IF        

    ! step 2 : tobs correction
    IF (isMeas) THEN
       call ObscurationCorrection(tobs)

       ! Check step 2
       IF (ChoosenVerbosity .ge. VerbosityReport) THEN
          print *, "step 2"
          print *, "REALACC(12)"
          print *, "accdLocal           = ", accdLocal
          print *, "ENDREALACC"
       END IF
    END IF

    ! step 3
    call CorrectBackground()
    ! Check step 3
    IF (ChoosenVerbosity .ge. VerbosityReport) THEN
       print *, "step 3"
       print *, "REALACC(12)"
       print *, "accdLocal           = ", accdLocal
       print *, "MINY                = ", miny
       print *, "ENDREALACC"
    END IF 

    ! Check if steps 1 to 3 have 
    ! ruined our input
    IF ( isNearZero(SUM(accdLocal)) ) THEN
       print *,"ERROR:mieresponse: sum(accdLocal) is near zero"
       error_flag = error_programming
       return 
    END IF  

    ! step 4
    call DivideByLargest(error_flag)
    ! Check step 4
    IF (ChoosenVerbosity .ge. VerbosityReport) THEN
       print *, "step 4"
       print *, "REALACC(12)"
       print *, "accdLocal           = ", accdLocal
       print *, "MAXY                = ", maxy
       print *, "MAXJ                = ", maxj
       print *, "ENDREALACC"
    END IF 

    ! step 5
    call FindStartPeakLoc()
    ! Check step 5
    IF (ChoosenVerbosity .ge. VerbosityReport) THEN
       print *, "step 5"
       print *, "REALACC(12)"
       print *, "accdLocal           = ", accdLocal
       print *, "COM                 = ", com
       print *, "ENDREALACC"
    END IF 

    ! step 6
    IF (ChoosenVerbosity .ge. VerbosityReport) THEN
       print *, "step 6"
    END IF
    call IterativeOptSearch(StartFWHM, ResErrThresh, MaxItNonLinOpt, & 
                            NumSpecSubSamp, MaxItLorFit, NonLinOptThresh, &
                            NumIt, FWHM, ResErr, ChoosenVerbosity)

    ! Again compatibility with C++ Mie Core we should only add 2.0 here. 
    ! But as the L2B team needs this algorithm, we add 3.0 add correct for -1.0
    ! outside.

    PeakLoc = SUM(com) + 3.0_FAdoxy


    ! step 7
    PeakDiff = ABS(PeakLoc - SUM(maxj))
    IF ( (PeakHeightLoThresh .LT.  peakHeightLocal)    .AND. &
         (peakHeightLocal .LT. PeakHeightUpThresh)     .AND. &
         (FWHMLoThresh .LT. FWHM)                      .AND. & 
         (FWHM .LT. FWHMUpThresh)                      .AND. &
         (PeakDiff .LT. PeakLocThresh)                        ) THEN

       ValFlag = .true.
    END IF

    IF (ChoosenVerbosity .ge. VerbosityReport) THEN
       print *, "step 7"
       print *, "peak height lower threshold: ", (PeakHeightLoThresh .LT.  peakHeightLocal)
       print *, "peak height upper threshold: ", (peakHeightLocal .LT. PeakHeightUpThresh)
       print *, "FWHM lower threshold       : ", (FWHMLoThresh .LT. FWHM)
       print *, "FWHM upper threshold       : ", (FWHM .LT. FWHMUpThresh)
       print *, "peak location threshold    : ", (ABS(PeakLoc - maxj) .LT. PeakLocThresh)
    END IF 

    ! step 8
    Offset     = offsetLocal + miny
    PeakHeight = peakHeightLocal * maxy
    error_flag = no_error

  end subroutine ComputeMieResponse

    !  #] 

  function isNearZero(value) result(isZero)
    !  #[
    ! A function that checks, if value is close to zero,
    ! using defined module variable ZeroEpsilon. 
    
    ! interface 
    real(FAdoxy), intent(in) :: value     ! input
    logical                  :: isZero    ! output

    isZero = .false.

    IF ((value .EQ. 0.0_FAdoxy) .OR. &
        (ABS(value) .LT. ZeroEpsilon)) THEN
      isZero = .true.
    END IF

  end function isNearZero
    !  #] 

  subroutine OffsetSubtraction(accd,offsetSub)
    !  #[
    ! A routine to subtract the offset as described 
    ! in DPM, ADM-MA-52-1800.
    
    ! interface 
    real(FAdoxy), intent(in), dimension(1:numACCDCols) :: accd      ! input
    real(FAdoxy), intent(out)                          :: offsetSub ! output

    ! no local variables

    ! Calculate offset using pixels 19, 20 
    offsetSub = (accd(19) + accd(20)) / 2.0_FAdoxy

    ! Subtract offset and fill local accd
    accdLocal = accd - offsetSub
    ! Note that accdLocal is a global module parameter

  end subroutine OffsetSubtraction
    !  #]

  subroutine ObscurationCorrection(tobs)
    !  #[
    ! A routine to correct the tripod obscuration as described 
    ! in DPM, ADM-MA-52-1800.
    
    ! interface 
    real(FAdoxy), intent(in), dimension(1:numUsefulACCDCols) :: tobs      ! input 

    ! Correct obscuration
    accdLocal(3:18) = accdLocal(3:18)/tobs
    ! Note that accdLocal is a global module parameter
 
  end subroutine ObscurationCorrection 
    !  #]

  subroutine CorrectBackground()
    !  #[
    ! A routine to correct for the background as described 
    ! in DPM, ADM-MA-52-1800.


    ! Find minimum
    miny = MINVAL(accdLocal(3:18))

    ! Correct background
    accdLocal(3:18) = accdLocal(3:18) - miny    
    ! Note that accdLocal is a global module parameter
 
  end subroutine CorrectBackground
    !  #]

  subroutine DivideByLargest(error_flag)   
    !  #[
    ! A routine to divide by the largest entry in the array.
    ! Save largest value and its position.

    ! interface 
    integer(IntAs), intent(out) :: error_flag                            ! output

    ! init
    error_flag = no_error

    ! Find maximum
    maxy = MAXVAL(accdLocal(3:18))
    maxj = INT(MAXLOC(accdLocal(3:18)),IntAuc) + 2_IntAuc 

    ! Check maxy 
    IF ( isNearZero(maxy) ) THEN
       error_flag = error_programming
       return 
    END IF 

    ! Correct background
    accdLocal(3:18) = accdLocal(3:18) / maxy    
    ! Note that accdLocal is a global module parameter
 
  end subroutine DivideByLargest
    !  #]

  subroutine FindStartPeakLoc()
     !  #[
    ! A routine to find the starting peak location as described
    ! in AE.TN.DLR.ACS-L1B.100306.

    ! Initialise accdLocal(2) & accdLocal(19) in case 
    ! maxj is at position 3 or 18
    accdLocal(2)  = accdLocal(18)
    accdLocal(19) = accdLocal(3)
    ! Note that accdLocal is a global module parameter

    ! Calculate weighted mean
    com =  ( (maxj-1) * accdLocal(maxj-1) + &
              maxj    * accdLocal(maxj)   + &
             (maxj+1) * accdLocal(maxj+1)    ) /  &
         (accdLocal(maxj-1) + accdLocal(maxj) + accdLocal(maxj+1) )
 
  end subroutine FindStartPeakLoc
    !  #] 

  function SubSamplesf_j(x_i, peakHeight, FWHM, offset, y_j, NumSpecSubSamp, j , &
                         ChoosenVerbosity) result (f_j)
    !  #[
    ! Function to calculate a single f_j as described in step 6
    ! of AE.TN.DLR.ACS-L1B.100306.

    
    ! interface
    real(FAdoxy),    intent(in) :: x_i, peakHeight, FWHM, offset, y_j ! input
    integer(IntAs),  intent(in) :: NumSpecSubSamp                     ! input
    integer(IntAuc), intent(in) :: j                                  ! input 
    integer,         intent(in) :: ChoosenVerbosity                   ! input  
    real(FAdoxy)                :: f_j                                ! result  

    ! local variables 
    integer(IntAuc)                           :: k
    real(FAdoxy)                              :: numerator, denominator, denominator_base
    real(FAdoxy), dimension(1:NumSpecSubSamp) :: subsamples  
    real(FAdoxy)                              :: k_real, NumSpecSubSamp_real

    ! for every subsample do
    NumSpecSubSamp_real = NumSpecSubSamp
    numerator           = peakHeight * FWHM * FWHM
    denominator_base    = x_i  - j + 0.5_FAdoxy + &
                          1.0_FAdoxy/(2.0_FAdoxy * NumSpecSubSamp)

    DO k = 1,NumSpecSubSamp 
       ! calculate fraction        
       k_real = k
       denominator   =  denominator_base - (k_real/NumSpecSubSamp_real) 

       subsamples(k) = numerator / &
                       (4.0_FAdoxy * denominator * denominator + FWHM * FWHM)  

       ! add & subtract
       subsamples(k) = subsamples(k) + offset - y_j 
    END DO

    f_j =  (1.0_FAdoxy / NumSpecSubSamp) *  SUM(subsamples)
     
  end function SubSamplesf_j 
    !  #] 

  function TargetFunc(x_i, StartFWHM, NumSpecSubSamp, ChoosenVerbosity) result (residual)
    !  #[
    ! Function to be optimized by simplex as described
    ! in AE.TN.DLR.ACS-L1B.100306.
    
    ! interface
    real(FAdoxy),   intent(in) :: x_i, StartFWHM             ! input 
    integer(IntAs), intent(in) :: NumSpecSubSamp             ! input 
    integer,        intent(in) :: ChoosenVerbosity           ! input  
    real(FAdoxy)               :: residual                   ! result 

    ! local variables
    real(FAdoxy)                              :: x_i_local
    real(FAdoxy), dimension(1:numACCDCols)    :: squares
    integer(IntAuc)                           :: j
    ! To be able to compare results with the C++ version, in which arrays
    ! start with index 0, the com is shifted to the left
    ! by three pixels before TargetFunc is called. But this Fortran routine
    ! calculates starting from position 3, so now let's shift it back 
    ! by three pixels
    x_i_local = x_i + 3.0_FAdoxy

    !loop over array position j in accdLocal
    DO j = 3,18
       squares(j) = SubSamplesf_j(x_i_local, peakHeightLocal, StartFWHM, offsetLocal, &
                                  accdLocal(j), NumSpecSubSamp, j,ChoosenVerbosity )
    END DO
    ! do square of addends
    squares(3:18) = squares(3:18) * squares(3:18) 
    ! final sum
    residual = SUM(squares(3:18))    
 
  end function TargetFunc
    !  #] 

  subroutine IterativeOptSearch(StartFWHM, ResErrThresh, MaxItNonLinOpt, & 
                                NumSpecSubSamp, MaxItLorFit, NonLinOptThresh, &
                                NumIt, FWHM, ResErr, ChoosenVerbosity)
    !  #[
    ! Iterative optimization search as described in step 6
    ! of AE.TN.DLR.ACS-L1B.100306.
    
    ! interface
    real(FAdoxy),   intent(in)  :: ResErrThresh,StartFWHM, NonLinOptThresh ! input 
    integer(IntAs), intent(in)  :: MaxItNonLinOpt, NumSpecSubSamp          ! input  
    integer(IntAs), intent(in)  :: MaxItLorFit                             ! input 
    integer,        intent(in)  :: ChoosenVerbosity                        ! input
    real(FAdoxy),   intent(out) :: FWHM,ResErr                             ! output
    integer(IntAs), intent(out) :: NumIt                                   ! output

    ! local variables
    real(FAdoxy), dimension(1:numACCDCols)    :: fieldf_j, fieldf_jsquare, f_jy_j
    real(FAdoxy)                              :: sumf_j, sumf_jsquare, sumy_j, sumy_jf_j
    integer(IntAuc)                           :: j, good, worse, worst
    real(FAdoxy)                              :: dummy1, dummy0
    real(FAdoxy), dimension(1:3,1:2)          :: vertices
    real(FAdoxy), dimension(1:3)              :: funcvalues 
!    real(FAdoxy), dimension(1:2)              :: xmin  ! never used
    real(FAdoxy)                              :: tmpResError, tmpFWHM, tmpcom                    

 
    dummy1 = 1.0_FAdoxy
    dummy0 = 0.0_FAdoxy
    DO j = 3,18
      fieldf_j(j)  = SubSamplesf_j(SUM(com), dummy1, StartFWHM, dummy0, dummy0, &
                                   NumSpecSubSamp, j, ChoosenVerbosity)
    END DO

    ! Loop over alternating linear/non-linear approach
    tmpResError = 1.0_FAdoxy
    NumIt = 0_IntAs
    tmpFWHM = StartFWHM


    ! To be able to compare reults with the C++ version, in which arrays
    ! start with index 0, the com is shifted to the left
    ! by three pixels.
    com = com - 3.0_FAdoxy

    DO WHILE ((tmpResError .GT. ResErrThresh) .AND. &
              (NumIt .LE. MaxItLorFit))

      ! calculate f_j for every pixel; f_j is coded to
      ! work on original com position, so (locally) add 3.0  
      tmpcom = SUM(com)  + 3.0_FAdoxy
      DO j = 3,18
      fieldf_j(j)  = SubSamplesf_j(tmpcom, dummy1, tmpFWHM, dummy0, dummy0, &
                                   NumSpecSubSamp, j, ChoosenVerbosity)
      END DO
      ! prepare linear equation system
      fieldf_jsquare(3:18) = fieldf_j(3:18) * fieldf_j(3:18)
      f_jy_j(3:18) = accdLocal(3:18) * fieldf_j(3:18)
      sumf_j = SUM(fieldf_j(3:18))
      sumf_jsquare = SUM(fieldf_jsquare(3:18))
      sumy_j = SUM(accdLocal(3:18))
      sumy_jf_j = SUM(f_jy_j(3:18)) 

    
      ! solve linear equation
      peakHeightLocal = (16.0_FAdoxy * sumy_jf_j - sumf_j * sumy_j) / &
                        (16.0_FAdoxy * sumf_jsquare - sumf_j *  sumf_j)
    
      offsetLocal = (sumy_j / 16.0_FAdoxy) - (sumf_j / 16.0_FAdoxy)*peakHeightLocal 
      IF (ChoosenVerbosity .ge. VerbosityReport) THEN
         print *, "Solution linear equation "
         print *, "REALACC(12)"
         print *, "peakHeightLocal           = ", peakHeightLocal
         print *, "offsetLocal               = ", offsetLocal
         print *, "ENDREALACC"
      END IF  

      ! Prepare input vertices and function values

      vertices(1,1) = SUM(com)
      vertices(1,2) = tmpFWHM
      vertices(2,1) = 1.1_FAdoxy * SUM(com)
      vertices(2,2) = tmpFWHM
      vertices(3,1) = SUM(com)
      vertices(3,2) = 1.1_FAdoxy * tmpFWHM

      funcvalues(1) = TargetFunc(vertices(1,1), vertices(1,2), NumSpecSubSamp, ChoosenVerbosity)
      funcvalues(2) = TargetFunc(vertices(2,1), vertices(2,2), NumSpecSubSamp, ChoosenVerbosity)
      funcvalues(3) = TargetFunc(vertices(3,1), vertices(3,2), NumSpecSubSamp, ChoosenVerbosity)
    
      IF (ChoosenVerbosity .ge. VerbosityReport) THEN
         print *, " input vertices and function values"
         print *, "REALACC(12)"
         print *, " funcvalues         = ", funcvalues
         print *, " vertices          = ", vertices
         print *, "ENDREALACC"
      END IF     

      ! Call simplex routine

      call simplex(vertices, funcvalues, NonLinOptThresh, NumSpecSubSamp, &
                   MaxItNonLinOpt, ChoosenVerbosity) 

      ! Find best new vertice in array of vertices
      call findGoodWorst(funcvalues, good, worse, worst)
      com = vertices(good ,1)
      tmpFWHM = vertices(good ,2)
      ! set new residual error
      tmpResError = funcvalues(good)    
      NumIt = NumIt + 1_IntAs
      IF (ChoosenVerbosity .ge. VerbosityReport) THEN
         print *, " output of simplex iteration"
         print *, "REALACC(12)"
         print *, " tmpResError        = ", tmpResError
         print *, " NumIt              = ", NumIt
         print *, "ENDREALACC"
      END IF     
    END DO
  FWHM = tmpFWHM
  ResErr = tmpResError 
  end subroutine IterativeOptSearch
    !  #] 

  subroutine simplex(vertices, funcvalues, NonLinOptThresh, NumSpecSubSamp, &
                     MaxItNonLinOpt, ChoosenVerbosity) 
    !  #[
    ! Nelder/Mead algorithm
    
    ! interface 
    integer(IntAs), intent(in) :: MaxItNonLinOpt, NumSpecSubSamp  ! input 
    real(FAdoxy),   intent(in) :: NonLinOptThresh                 ! input  
    integer,        intent(in) :: ChoosenVerbosity                ! input 
    real(FAdoxy), dimension(1:3,1:2), intent(inout) :: vertices   ! input/output
    real(FAdoxy), dimension(1:3),     intent(inout) :: funcvalues ! input/output

    ! local variables 
    integer(IntAuc)               :: good, worse, worst, conv, it 
    real(FAdoxy), dimension(1:2)  :: reflected, extended, contracted
    real(FAdoxy)                  :: residual


    conv = 0_IntAuc
    it = 0_IntAuc

    DO WHILE (conv .EQ. 0)

      ! sort vertices & check convergence
      call findGoodWorst(funcvalues, good, worse, worst)
      conv = checkConvergence(funcvalues, good, worst, NonLinOptThresh)
      IF((conv .EQ. 1) .OR. (it .GE. MaxItNonLinOpt)) THEN
        EXIT
      END IF 
      
      ! Reflect vertice and calculate function value at this new vertice      
      
      call reflectVertice(vertices, good, worse, worst, reflected)

      residual = TargetFunc(reflected(1), reflected(2), NumSpecSubSamp, ChoosenVerbosity)
      it = it + 1_IntAuc
      IF (residual .LE. funcvalues(worst)) THEN     

         ! Reflection is good, keep new vertice
         vertices(worst,:) = reflected
         funcvalues(worst) = residual
      END IF ! reflection done
         
      ! Let's try extension or contraction
      IF (residual .LE. funcvalues(good)) THEN

         call extendVertice(vertices, good, worse, worst, extended)

         residual = TargetFunc(extended(1), extended(2), NumSpecSubSamp, ChoosenVerbosity)
         it = it + 1_IntAuc
         ! Check if extensions is even better than reflection
         IF (residual .LE. funcvalues(worst)) THEN

            ! Extension is good, keep new vertice
            vertices(worst,:) = extended
            funcvalues(worst) = residual
         END IF
      ELSE IF (residual .GE. funcvalues(worse)) THEN  

        call contractVertice(vertices, good, worse, worst, contracted) 

        residual = TargetFunc(contracted(1), contracted(2), NumSpecSubSamp, ChoosenVerbosity)
        it = it + 1_IntAuc
        IF (residual .LE. funcvalues(worst)) THEN

           ! Contraction is good, keep new vertice
           vertices(worst,:) = contracted
           funcvalues(worst) = residual
        ELSE

           ! Replace worse and worst vertice of the start simplex
           vertices(worse,:) = 0.5_FAdoxy * (vertices(worse,:) + vertices(good,:))
           vertices(worst,:) = 0.5_FAdoxy * (vertices(worst,:) + vertices(good,:))
           ! Calculate new function values 

           funcvalues(worse) = TargetFunc(vertices(worse,1), vertices(worse,2), & 
                                          NumSpecSubSamp, ChoosenVerbosity) 

           funcvalues(worst) = TargetFunc(vertices(worst,1), vertices(worst,2), &
                                          NumSpecSubSamp, ChoosenVerbosity)
           it = it + 2_IntAuc
        END IF
      END IF
    END DO 
           
  end subroutine simplex
    !  #]


  subroutine findGoodWorst(funcvalues, good, worse, worst)
    !  #[
    ! Find good is position of lowest value in funcvalues, worst is highest,
    ! worse is the remaining position
    
    ! interface
    real(FAdoxy), dimension(1:3), intent(in)  :: funcvalues           ! input
    integer(IntAuc),              intent(out) :: good, worse, worst   ! output
   
    good  = SUM(INT(MINLOC(funcvalues),IntAuc))
    worst = SUM(INT(MAXLOC(funcvalues),IntAuc))
    worse = 1
    IF ((worse .EQ. good) .OR. (worse .EQ. worst)) THEN
       worse = 2
       IF ((worse .EQ. good) .OR. (worse .EQ. worst)) THEN
       worse = 3
       END IF
    END IF
    
  end subroutine findGoodWorst
    !  #]  

  function checkConvergence(funcvalues, good, worst, NonLinOptThresh) result(conv)
    !  #[
    ! Check convergence of simplex
    
    ! interface
    real(FAdoxy), dimension(1:3), intent(in) :: funcvalues       ! input 
    integer(IntAuc),              intent(in) :: good, worst      ! input 
    real(FAdoxy),                 intent(in) :: NonLinOptThresh  ! input    
    integer(IntAuc)                          :: conv             ! result 

    ! local variables
    real(FAdoxy)                             :: optimized

    optimized = 2.0_FAdoxy * ABS(funcvalues(worst)  -     funcvalues(good))/ &
                            (ABS(funcvalues(worst)) + ABS(funcvalues(good)))
    IF (optimized .LT. NonLinOptThresh) THEN
       conv = 1
    ELSE
       conv = 0
    END IF
 
  end function checkConvergence
    !  #]   

  subroutine reflectVertice(vertices, good, worse, worst, reflected)
    !  #[
    ! Reflect vertice through the side connecting the two
    ! better vertices.
    
    ! interface
    real(FAdoxy), dimension(1:3,1:2), intent(in)  :: vertices           ! input
    integer(IntAuc),                  intent(in)  :: good, worse, worst ! input
    real(FAdoxy), dimension(1:2),     intent(out) :: reflected          ! output

    reflected =  (vertices(worse,:) + vertices(good,:)) / 2.0_FAdoxy
    reflected =  reflected + ( reflected - vertices(worst,:))
    
  end subroutine reflectVertice
    !  #]    

  subroutine extendVertice(vertices, good, worse, worst, extended)
    !  #[
    ! Reflect vertice through the side connecting the two
    ! better vertices.
    
    ! interface
    real(FAdoxy), dimension(1:3,1:2), intent(in)  :: vertices           ! input
    integer(IntAuc),                  intent(in)  :: good, worse, worst ! input
    real(FAdoxy), dimension(1:2),     intent(out) :: extended           ! output
 
    extended = (vertices(worse,:) + vertices(good,:)) / 2.0_FAdoxy
    extended = extended + 2.0_FAdoxy * ( vertices(worst,:) - extended)
    
  end subroutine extendVertice
    !  #] 

  subroutine contractVertice(vertices, good, worse, worst, contracted)
    !  #[
    ! Reflect vertice through the side connecting the two
    ! better vertices.
    
    ! interface
    real(FAdoxy), dimension(1:3,1:2), intent(in)  :: vertices           ! input
    integer(IntAuc),                  intent(in)  :: good, worse, worst ! input
    real(FAdoxy), dimension(1:2),     intent(out) :: contracted         ! output
   
    contracted = (vertices(worse,:) + vertices(good,:)) / 2.0_FAdoxy
    contracted = contracted + 0.5_FAdoxy * (vertices(worst,:) - contracted)
    
  end subroutine contractVertice
    !  #] 
  !-----------------------------------
end module MieResponse

