module MolScat
  !  #[ Description
  !---------------------------------------------------
  ! purpose: a module to calculate molecular extinction and
  !          backscatter, as a function of atmospheric
  !          temperature and pressure.
  !          (based on the report by Wiel Wauben:
  !           KNMI, Scientific Report; WR 96-01, by W.M.F. Wauben
  !           "a new algorithm for total ozone retrieval from direct
  !            sun measurements with a filter instrument"
  !          (see my collection, article nr. 56)
  !
  !    Written by:  Jos de Kloe, KNMI
  !    created:     03-03-2005
  !
  !    Modifications:
  !     12-May-2006  D. Tan       Null use of ParKind1 module
  !
  !---------------------------------------------------
  !  #]
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE AeolusConstants, only: pi
  USE AeolusConstants, only: adm_laser_wavelength => lambda
  USE AeolusConstants, only: kb ! the Boltzmann constant in [J/s]

  USE VerbosityControl, only: VerbosityError, VerbosityWarn, VerbosityReport
  USE numerics, only: r8_,missing_real, missing_indicator_real
  !  #]
  !  #[ module parameters and variables
  implicit none

  ! module parameters

  ! the Boltzmann constant in [J/K]
  real(r8_), parameter :: k = kb*1.e4 ! convert from si to cgs units
  ! remember: p = n.k.T 
  !   in s.i.:   in cgs:
  ! p [Pa]       [hPa]
  ! n [m^-3]     [cm^-3]
  ! T [K]        [K]
  ! so in s.i.:
  ! k = p/nT = [Pa.m^3/K] = [N.m/K] = [J/K]
  ! and in c.g.s.:
  ! k = p/nT = [hPa.cm^3/K] = [10^-4.N.m/K] = 10^-4.[J/K]

  real(r8_), parameter :: n_s = 2.54743e19  ! numberdensity at standard P and T
  real(r8_), parameter :: ns2 = n_s*n_s     ! n_s^2
  real(r8_), parameter :: rho_n = 0.03178   ! near 306 nm
  real(r8_), parameter :: dz = 100.         ! cm per meter
  real(r8_), parameter :: pi3 = pi*pi*pi    ! pi^3
  real(r8_), parameter :: factor3 = ((6.+3.*rho_n)/(6.-7.*rho_n)) ! eq 4.5
                                ! = 1.05500611126177
  ! module variables
  integer, public, save :: Verbosity = VerbosityError 
  logical,   save :: is_initialised = .false.
  real(r8_), save :: lambda_mu,  lambda_cm
  real(r8_), save :: lmt, m_s, l4, sigma_z_mol
  real(r8_), save :: factor1, factor2
  !  #]
contains
  !-----------------------------------
  subroutine InitMolScatModule(ChoosenVerbosity)
    !  #[
    integer, optional, intent(in) :: ChoosenVerbosity

    print *,"initialising MolScat Module"
    is_initialised = .true.

    ! set verbosity if needed
    IF (present(ChoosenVerbosity)) Verbosity = ChoosenVerbosity

    lambda_mu = adm_laser_wavelength*1e6 ! laser wavelength in micro-meter
    lambda_cm = adm_laser_wavelength*1e2 ! laser wavelength in [cm]

    ! refractive index of air:       ! see eq. 4.7
    lmt = 1./(lambda_mu*lambda_mu)
    m_s = 1. + 1.e-8 * (5791817./(238.0185 - lmt) + 167909./( 57.362  - lmt))

    ! brekingsindex van lucht bij 589 nm is 1.00029 volgens BINAS, tabel 18c
    ! deze bovenstaande berekening geeft: 
    !         1.00028569773896 voor 355 nm
    ! of      1.00027713833779 voor 589 nm     dus dat klopt aardig

    l4  = lambda_cm**4                       
    factor1 = (24.*pi3/(l4*ns2))               ! eq 4.5
    factor2 = ((m_s*m_s-1.)**2/(m_s*m_s+2)**2) ! eq 4.5
    sigma_z_mol = factor1*factor2*factor3      ! eq 4.5

    IF (Verbosity .ge. VerbosityReport) THEN
                                            ! values for 355 nm
       print *,"m_s         = ",m_s         ! 1.00028569773896
       print *,"l4          = ",l4          ! 1.58823001594309E-18
       print *,"factor1     = ",factor1     ! 7.22009605405709E-19
       print *,"factor2     = ",factor2     ! 3.62735209265506E-8
       print *,"factor3     = ",factor3     ! 1.05500611126177
       print *,"sigma_z_mol = ",sigma_z_mol ! 2.76304312629617E-26
    END IF

  end subroutine InitMolScatModule
    !  #]
  !-----------------------------------
  ! real calculation using p and T
  ! copied from: ../adm/src/MolScat
  function calc_molecular_extinction(p,t) result(sigma_m)
    !  #[
    real(r8_) :: p, t    ! input
    real(r8_) :: sigma_m ! result
    ! local variable
    real(r8_) :: n_z

    IF (.not. is_initialised) THEN
       print *,"ERROR in calc_molecular_extinction:"
       print *,"the MolScat module should be initialised before use !!"
       stop 1
    END IF

    IF (missing_real(p) .or. missing_real(t)) THEN
       sigma_m = missing_indicator_real
       return
    END IF
    
    n_z = p/(k*t)
    sigma_m = sigma_z_mol * n_z * dz ! eq. 4.1

  end function calc_molecular_extinction
    !  #]
  function calc_molecular_backscatter(p,t) result(beta_m)
    !  #[
    real(r8_) :: p, t   ! input
    real(r8_) :: beta_m ! result
    ! local variable
    real(r8_) :: sigma_m

    !      sigma_m = 8*pi*beta_m/3
    ! so   beta_m  = 3*sigma_m/(8*pi)
    ! see [RD1] (wales), p.A3, or [RD3] (dera), p. 50.
    ! Note that in these references the symbol alpha is
    ! used for extinction

    IF (.not. is_initialised) THEN
       print *,"ERROR in calc_molecular_backscatter:"
       print *,"the MolScat module should be initialised before use !!"
       stop 1
    END IF

    IF (missing_real(p) .or. missing_real(t)) THEN
       beta_m = missing_indicator_real
       return
    END IF

    sigma_m = calc_molecular_extinction(p,t)
    beta_m  = 3.*sigma_m/(8.*pi)
  end function calc_molecular_backscatter
    !  #]
  !-----------------------------------
  ! climatological only based on z
  ! copied from: ../adm/src/AtmState
  function get_ref_mol_scat(wanted_z,two_digits_only) result(mol_scat)
    !  #[
    ! inputs
    real(r8_)         :: wanted_z        ! height in [m] 
    logical, optional :: two_digits_only ! optional switch

    ! output
    real(r8_) :: mol_scat
    
    ! local variable
    character(15)      :: txt_beta_m
    integer            :: ios

    if (missing_real(wanted_z)) then
       mol_scat = missing_indicator_real
       return
    end if

    mol_scat = AtmScatMol(adm_laser_wavelength,wanted_z)

    ! this rather dirty workaround trick is needed to remain compatible
    ! with the lite ascii datafiles, in which this number was written with
    ! a precision of only 2 digits after the period.
    if (present(two_digits_only)) then
       if (two_digits_only) then
          write (txt_beta_m,"(es10.2E2)",iostat=ios) mol_scat
          IF (ios .ne. 0) goto 999
          read(txt_beta_m,*,iostat=ios) mol_scat
          IF (ios .ne. 0) goto 999
       end if
    end if
    return

999 mol_scat = missing_indicator_real

  end function get_ref_mol_scat
    !  #]
  function AtmScatMol(rLamb,z) result(mol_scat)
    !  #[

    ! SUBROUTINE AtmScatMol
    !
    ! G.J. Marseille
    ! KNMI
    ! August 2002
    !
    !   PURPOSE
    !   -------
    !   Molecular backscatter as a function of height and laser wavelength.
    !         
    !   INTERFACE
    !   ---------
    !   call AtmScatMol
    !
    !   AtmScatMol is called by AtmScatRef
    !
    !   METHOD
    !   ------
    !   Parameterization of DERA (Vaughan et al.)
    !   see as source for this formula and its parameters
    !   the DIAL report [RD2], p. 82, zie map 2, artikel 52
    !
    !   INPUT
    !   -----
    !   rLamb   - laser wavelength
    !
    !   OUTPUT
    !   -----
    !   Reference molecular backscatter profile
    !         
    ! =============================================================

    ! input
    real(r8_) :: rLamb
    real(r8_) :: z

    ! output
    real(r8_) :: mol_scat

    ! local parameters and variables
    real(r8_), parameter :: rLamb0 = 1.06E-6_r8_ ! ref. wavelength (1 micron)
    real(r8_), parameter :: rZmol  = 8000._r8_   ! ??? length in [m]

    ! used: nLevsRaw, rzp(*,1) = heigth
    !       nbRgate= 3
    !       rlamb = laser wavelength (input)
    
    mol_scat = 1.0E-7_r8_ * (rLamb0/rLamb)**4  * EXP( -z/rZmol )

    RETURN
  END function AtmScatMol
  !  #]
  !-----------------------------------
end module MolScat
