program TestOpticalProperties
  !  #[ Description:
  !     A test program for the OpticalProperties module
  !   
  !     created:     05-01-2006, by: J. de Kloe
  !
  !     Modifications:
  !     12-May-2006 D. Tan       Null use of ParKind1 module
  !     23-Jun-2006 J. de Kloe   updated to new interface of Opt.Prop. module
  !     05-Oct-2006 J. de Kloe   added use of WorkingDataType
  !     23-Oct-2006 D. Tan       Use of WorkingDataType for old L2B-working
  !     24-Apr-2007 J. de Kloe   phase out the SummaryVars code
  !     28-Aug-2007 D. Tan       Phase out obsolete init routine.
  !     03-Sep-2007 P. Poli      Modified calls to Init_L1B_ and Delete_L1B_
  !     04-Sep-2007 P. Poli      change L1B_DataType to L1B_BRC_DataType
  !     16-Jan-2008 J. de Kloe   phase out integer kind i_
  !
  !  #]
  !  #[ Modules used:
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE OpticalProperties, only: InitOpticalProperties, &
                               PrintBackscatterRatio, &
                               Match_Mie_And_Rayleigh_Bins, &
                               Calc_Extinction, Calc_BackscatterRatio,&
                               Calc_Extinction_Old, &
                               Copy_Extinction_Rayl_to_Mie, &
                               CloseOpticalProperties
  USE L2B_Proc_Settings, only: PartExt_Method_Old, PartExt_Method_Iterative,&
                               MolExt_Method_Parametrised, MolExt_Method_Proper

  USE L2B_AuxPar_DataStructure, only: Scat_Ratio_Undefined, &
                               Scat_Ratio_from_L1B, Scat_Ratio_from_MieRayl,&
                               Scat_Ratio_from_RaylOnly, Scat_Ratio_Dont_Use

  USE LiteDataModule, only: Fill_WorkingData_With_Lite_Data
  ! a routine to define the globally used L1B data struct
  USE L1B_DataStructure, only: L1B_BRC_DataType, Init_L1B_BRC_DataStructure, &
                               Delete_L1B_BRC_DataStructure
  USE L1B_PCD_ADS, only: Init_L1B_PCD_ADS, Delete_L1B_PCD_ADS
  USE Working_Datastructure, only: WorkingDataType,InitWorkingData, &
                                   DeleteWorkingData !, PrintWorkingData
  USE EE_CFI_Datatypes, only: IntAuc, IntAs
  USE MolScat,  only: InitMolScatModule, get_ref_mol_scat
  ! support modules
  USE Compiler_Features, only: iargc_aeolus,getarg_aeolus
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, VerbosityReport
  USE numerics, only: r8_
  USE ErrorHandler, only: no_error, error_allocate, error_cmdline_options, &
                          program_abort

  !  #]
  !  #[ Parameter/Variable/Type declarations

  implicit none

  ! variables
  character(len=256) :: filename
  integer            :: error_flag
  integer            :: nr_BRCs_available
  integer            :: read_which_BRC
!  integer            :: AllocStatus
  integer            :: N_meas, r, m
  integer            :: nr_of_arguments, current_argument_nr
  character(len=256) :: current_argument
  character(len=256) :: scatteringratio_type
  integer(IntAuc)    :: ScatRatio_Method, ScatRatio_Method2
  character(len=256) :: molext_type
  integer            :: MolExt_Method
  character(len=256) :: partext_type
  integer            :: PartExt_Method
  real(r8_)          :: alt_mid

  ! define how the measurements are organised
  ! (in 67 profiles consisting of 24 range gates each)
  integer,         parameter :: N_meas_max    = 67 ! max used nr of meas.
  integer,         parameter :: N_range_gates = 24 ! current used nr of gates
  integer(IntAuc), parameter :: N_Obs_Mie_Max      = 4
  integer(IntAuc), parameter :: N_Obs_Rayleigh_Max = 4

  ! a struct to hold all necessary data for the processor
  TYPE(L1B_BRC_DataType), save   :: L1B
  TYPE(WorkingDataType)      :: WorkingData

  !  #]
  !  #[ program code
  print *,"running testprogram: TestClassification"

  ! init
  N_meas = N_meas_max
  error_flag = no_error
  call InitMolScatModule()

  ! set defaults
  ScatRatio_Method  = Scat_Ratio_from_L1B
  ScatRatio_Method2 = Scat_Ratio_Dont_Use
  MolExt_Method     = MolExt_Method_Proper
  PartExt_Method    = PartExt_Method_Old

  ! read commandline arguments
  nr_of_arguments = iargc_aeolus()
  current_argument_nr = 0  
  DO WHILE (current_argument_nr .lt. nr_of_arguments)
     !  #[ process the given commandline arguments

     current_argument_nr = current_argument_nr+1
     call getarg_aeolus(current_argument_nr,current_argument)
     SELECT CASE (trim(current_argument))
     CASE ("-30") 
        ! 
        print *,"using only 30 measurements"
        N_meas = 30
     CASE("-ScatteringRatio")
        !  #[
        current_argument_nr = current_argument_nr+1
        IF (current_argument_nr .gt. nr_of_arguments) THEN
           print *,"type is missing after -ScatteringRatio option !!!"
           error_flag = error_cmdline_options
           call program_abort(error_flag,"TestOpticalProperties")
        END IF
        call getarg_aeolus(current_argument_nr,scatteringratio_type)
        select case (scatteringratio_type)
        case("L1B")
           ScatRatio_Method = Scat_Ratio_from_L1B
        case("MieRayl")
           ScatRatio_Method = Scat_Ratio_from_MieRayl
        case("RaylOnly")
           ScatRatio_Method = Scat_Ratio_from_RaylOnly
        case("DontUse")
           print *,"ERROR: the DontUse switch is only usefull for the"
           print *,"fall-back method, not for the 1st choice method !!"
           print *,"Update the L2B_Proc_Settings module if you really need"
           print *,"the option to switch off this calculation..."
           error_flag = error_cmdline_options
           call program_abort(error_flag,"TestOpticalProperties")
        case default
           print *,"unrecognised scatteringratio type: ",&
                trim(scatteringratio_type)
           error_flag = error_cmdline_options
           call program_abort(error_flag,"TestOpticalProperties")
        end select
        print *,'using scatteringratio type: ',trim(scatteringratio_type)
        !  #]
     CASE("-ScatteringRatio2")
        !  #[
        current_argument_nr = current_argument_nr+1
        IF (current_argument_nr .gt. nr_of_arguments) THEN
           print *,"type is missing after -ScatteringRatio2 option !!!"
           error_flag = error_cmdline_options
           call program_abort(error_flag,"TestOpticalProperties")
        END IF
        call getarg_aeolus(current_argument_nr,scatteringratio_type)
        select case (scatteringratio_type)
        case("L1B")
           ScatRatio_Method2 = Scat_Ratio_from_L1B
        case("MieRayl")
           ScatRatio_Method2 = Scat_Ratio_from_MieRayl
        case("RaylOnly")
           ScatRatio_Method2 = Scat_Ratio_from_RaylOnly
        case("DontUse")
           ScatRatio_Method2 = Scat_Ratio_Dont_Use
        case default
           print *,"unrecognised scatteringratio2 type: ",&
                trim(scatteringratio_type)
           error_flag = error_cmdline_options
           call program_abort(error_flag,"TestOpticalProperties")
        end select
        print *,'using scatteringratio2 type: ',trim(scatteringratio_type)
        !  #]
     CASE("-MolExt")
        !  #[
        current_argument_nr = current_argument_nr+1
        IF (current_argument_nr .gt. nr_of_arguments) THEN
           print *,"type is missing after -MolExt option !!!"
           error_flag = error_cmdline_options
           call program_abort(error_flag,"TestOpticalProperties")
        END IF
        call getarg_aeolus(current_argument_nr,molext_type)
        select case (molext_type)
        case("Parametrised")
           MolExt_Method = MolExt_Method_Parametrised
        case("Proper")
           MolExt_Method = MolExt_Method_Proper
        case default
           print *,"unrecognised MolExt method: ",&
                trim(molext_type)
           error_flag = error_cmdline_options
           call program_abort(error_flag,"TestOpticalProperties")
        end select
        print *,'using MolExt method: ',trim(molext_type)
        !  #]
     CASE("-PartExt")
        !  #[
        current_argument_nr = current_argument_nr+1
        IF (current_argument_nr .gt. nr_of_arguments) THEN
           print *,"type is missing after -PartExt option !!!"
           error_flag = error_cmdline_options
           call program_abort(error_flag,"TestOpticalProperties")
        END IF
        call getarg_aeolus(current_argument_nr,partext_type)
        select case (partext_type)
        case("Old")
           PartExt_Method = PartExt_Method_Old
        case("Iterative")
           PartExt_Method = PartExt_Method_Iterative
        case default
           print *,"unrecognised PartExt method: ",&
                trim(partext_type)
           error_flag = error_cmdline_options
           call program_abort(error_flag,"TestOpticalProperties")
        end select
        print *,'using PartExt method: ',trim(partext_type)
        !  #]
     CASE DEFAULT
        print *,"commandline argument not recognised: ",trim(current_argument)
        error_flag = error_cmdline_options
        call program_abort(error_flag,"TestOpticalProperties")
     END SELECT
     !  #]
  END DO

  ! initialise the modules used by the program
  call Init_L1B_BRC_DataStructure(L1B, error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"Init_L1B_BRC_DataStructure")

  call InitOpticalProperties(PartExt_Method, error_flag,&
                             ChoosenVerbosity=VerbosityError)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"InitOpticalProperties")

  ! allocate the L1B PCD and fill the backscatter ratio field
  ! with dummy values, to be able to test the copy routine
  call Init_L1B_PCD_ADS(L1B%PCD_ADS, N_meas, N_range_gates, error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"Init_L1B_PCD_ADS")

  ! fill the L1B scattering_ratio field with dummy data, to be able 
  ! to run the test on it
  DO m=1,N_meas
     DO r=1,N_range_gates
        L1B%PCD_ADS%Measurement_PCD(m)%&
             Meas_Alt_Bin_PCD(r)%scattering_ratio = 0.01*m + 0.0001*r
     END DO
  END DO

  filename = "backscatterdata.dat"

  ! loop over all BRCs in the file
  read_which_BRC = 1
  brc_loop: DO

     ! WorkingData must be initialized/deleted every BRC so that
     ! working values from a previous BRC are not retained 
     call InitWorkingData(WorkingData, &
                 int(N_range_gates,IntAs), int(N_Meas,IntAs), &
                 int(N_Obs_Mie_Max,IntAs), int(N_Obs_Rayleigh_Max,IntAs),&
                 error_flag)
     IF (error_flag .ne. no_error) &
          call program_abort(error_flag,"InitWorkingData")

     !call PrintWorkingData(WorkingData,prefix="testjos",PrintAll=.false.,&
     !                      error_flag=error_flag)

     ! GJM, change this routine. Read L1B instead of LITE
     call Fill_WorkingData_With_Lite_Data(WorkingData, filename, &
                      read_which_BRC, nr_BRCs_available, error_flag)
     IF (error_flag .ne. no_error) &
          call program_abort(error_flag,"Fill_WorkingData_With_Lite_Data")

     call Match_Mie_And_Rayleigh_Bins(WorkingData, error_flag)
     IF (error_flag .ne. no_error) &
          call program_abort(error_flag,"Match_Mie_And_Rayleigh_Bins")

     DO m=1,N_meas
        DO r=1,WorkingData%RayleighMeas(m)%nr_of_range_gates
           alt_mid = &
                (WorkingData%RayleighMeas(m)%RangeGate(r)%alt_start + &
                 WorkingData%RayleighMeas(m)%RangeGate(r)%alt_end    )*0.5_r8_
           ! the two_digits_only workaround trick is needed to remain 
           ! compatible with the previous code in which this number was 
           ! read from the lite ascii datafile with a precision of just 
           ! 2 digits after the period
           WorkingData%RayleighMeas(m)%RangeGate(r)%calc_mol_backscatter = &
                get_ref_mol_scat(alt_mid,two_digits_only=.true.)
        END DO
     END DO

     call Calc_Extinction(L1B, WorkingData, &
                          MolExt_Method, PartExt_Method, error_flag, &
                          calibr_on_upper_bin=.true.,&
                          use_lite_test_data=.true.)
     IF (error_flag .ne. no_error) &
          call program_abort(error_flag,"Calc_Extinction")

     call Copy_Extinction_Rayl_to_Mie(WorkingData)

     call Calc_BackscatterRatio(L1B, WorkingData, error_flag, &
                                Method         = ScatRatio_Method, &
                                FallbackMethod = ScatRatio_Method2 )
     IF (error_flag .ne. no_error) &
          call program_abort(error_flag,"Calc_BackscatterRatio")

     ! print BackscatterRatio, and Extinction
     ! result for a single Measurement
     call PrintBackscatterRatio(WorkingData,Meas_Nr=1)

     ! delete the WorkingData structure, it will be allocated again for
     ! the next BRC after reading the next BRC from the L1B file.
     call DeleteWorkingData(WorkingData)

     ! as test, just loop once
     ! exit brc_loop

     read_which_BRC =  read_which_BRC + 1
     IF (read_which_BRC .gt. nr_BRCs_available) exit brc_loop
  END DO brc_loop

  print *,"=================================="
  print *,"nr of processed BRC's was: ",nr_BRCs_available
  print *,"=================================="

  ! free allocated memory when done
  call Delete_L1B_PCD_ADS(L1B%PCD_ADS)
  call Delete_L1B_BRC_DataStructure(L1B)
  call DeleteWorkingData(WorkingData)
  call CloseOpticalProperties

  !  #]
end program TestOpticalProperties

