MODULE LUT_RayCalib_handling
  !  #[ Description:
  !
  ! Modifications: 
  !   26-Apr-2006  G.J. Marseille, initial version
  !   30-Jun-2006  D. Tan,         Null use of ParKind1 module
  !   19-Sep-2006  J. de Kloe,     added TableLoaded switch,
  !                                removed double code from Read_LUT_RayCalib
  !   02-Oct-2007  J. de Kloe      added code to use RBC file in stead of the
  !                                custom binary RayCalib file
  !   05-Oct-2007  J. de Kloe      added some missing intents
  !   16-Jan-2008  J. de Kloe      phase out integer kind i_
  !
  !  #]
  !  #[ Modules used:
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE numerics    , only: r8_, missing_indicator_real
  USE LunManager  , only: get_lun, free_lun
  USE ErrorHandler, only: no_error,error_opening_file,error_allocate, &
                          error_programming
  USE interp1     , only: HUNT
  USE rbc_datastructure, only: RBC_DataType
  USE EE_CFI_Datatypes, only: GetIntAs, maxlen_IntAs, IntAs
  !  #]
  !  #[ variables

  implicit none
  
  ! local variable
  !integer, parameter :: nPmx =104  ! LUT number of pressure values
  !integer, parameter :: nTmx =201  ! LUT number of temperature values
  !integer, parameter :: nFmx = 61  ! LUT number of frequency values
  
  
  logical, save :: TableLoaded = .false.
  integer, save :: nP  ! LUT number of pressure values
  integer, save :: nT  ! LUT number of temperature values
  integer, save :: nF  ! LUT number of frequency values

  real(r8_), allocatable, dimension (:), save     :: rPtab,rTtab,rFtab
  real(r8_), allocatable, dimension (:,:,:), save :: rNa,rNb

  ! filename to be used for the LUT
  character(len=*), parameter ::filename="iliadRayCalTable.bin"

  !  #]
contains
  !---------------------------
  SUBROUTINE Load_LUT_RayCalib(error_flag)
    !  #[

    ! Description:
    !    This routine opens the binary file that contains the lookup table 
    !    of the Rayleigh calibration factor (C1). The lookup table 
    !    inlcudes the number of counts on the
    !    Rayleigh channels A and B as a function of pressure, temperature and
    !    Doppler shifted frequency.
    !    The lookup table has been generated by Alain Dabas 
    !    based on the ILIAD report, using MATLAB. The resulting tables were
    !
    ! Input
    ! =====
    !    nothing
    !
    ! Output
    ! ======
    !    nP        : LUT dimension of pressure
    !    nT        : LUT dimension of temprature
    !    nF        : LUT dimension of Doppler shifted frequency
    !    rPtab     : array of pressure values (hPa)
    !    rTtab     : array of temperature values (K)
    !    rFtab     : array of frequency values (MHz)
    !    rNa       : Rayleigh channel A photon counts
    !    rNb       : Rayleigh channel B photon counts
    !
    ! Method
    ! ======
    !    Open binary file and read parameters
    !
    ! Modifications:
    ! =============
    !    08-05-2006 G.J. Marseille, initial version
    ! 
    ! ===========================================================
    ! 
    ! I/O variables
    integer, intent(out) :: error_flag
    
    ! local variables
    integer :: fileunit,allocstat

    ! ========================================================
    !
    ! init
    error_flag = no_error
    
    ! Open binary file containing lookup table
    fileunit = get_lun()
    open(unit=fileunit,file=filename,form='UNFORMATTED',&
         status='old',action="read",ERR=96)
    read(fileunit,ERR=98,END=97) nP,nT,nF
    rewind(fileunit)
    
    allocate(rPtab(nP),stat=allocstat)
    if (allocstat .ne. 0) goto 99
    allocate(rTtab(nT),stat=allocstat)
    if (allocstat .ne. 0) goto 99
    allocate(rFtab(nF),stat=allocstat)
    if (allocstat .ne. 0) goto 99
    allocate(rNa(nP,nT,nF),stat=allocstat)
    if (allocstat .ne. 0) goto 99
    allocate(rNb(nP,nT,nF),stat=allocstat)
    if (allocstat .ne. 0) goto 99
    
    read(fileunit,ERR=98,END=97) nP,nT,nF,rPtab,rTtab,rFtab,rNa,rNb
    close(unit=fileunit)
    call free_lun(fileunit)

    ! remember that we have the table in memory
    TableLoaded = .true.
    
    return
    
96  print *,"ERROR: Load_LUT_RayCalib"
    print *,"unable to open input file: ",trim(filename)
    error_flag = error_opening_file
    return

97  print *,"ERROR: Load_LUT_RayCalib"
    print *,"unexpected end of input file: ",trim(filename)
    error_flag = error_opening_file
    return
    
98  print *,"ERROR: Load_LUT_RayCalib"
    print *,"reading failed from input file: ",trim(filename)
    error_flag = error_opening_file
    return
    
99  print *,"ALLOCATION ERROR Load_LUT_RayCalib"
    error_flag = error_allocate
    return
    
  END SUBROUTINE Load_LUT_RayCalib
  !  #]
  SUBROUTINE LUT_RayCalib_Intpol(Temp,PresshPa,Freq_wind,RayCal,error_flag)
    !  #[
    !  Description:
    !    This routine retrieves the Rayleigh calibration factor (C1) from
    !    a lookup table. The lookup table inlcudes the number of counts on the
    !    Rayleigh channels A and B as a function of pressure, temperature and
    !    Doppler shifted frequency.
    !    The lookup table has been generated by Alain Dabas 
    !    based on the ILIAD report, using MATLAB. The resulting tables were
    !    saved in ASCII format, next imported in FORTRAN and saved in binary 
    !    format.
    !
    ! Input
    ! =====
    !    Temp       ! Temperature (K)
    !    PresshPa   ! Pressure (hPa)
    !    Freq_wind  ! Frequency of Doppler shift (MHz)
    !    Lookup table in binary format
    !
    ! Output
    ! ======
    !    RayCal     ! Rayleigh channel calibration value
    !    error_flag ! success or not
    !
    ! Method
    ! ======
    !    Interpolate table to input values    
    !
    ! Modifications:
    ! =============
    !    26-apr-2006 G.J. Marseille, initial version
    !    19-sep-2006 J. de kloe, added error_flag and TableLoaded switch
    !           and changed press unit to hPa to prevent double conversion
    ! 
    ! ===========================================================

    ! I/O variables
    real(r8_), intent(in)  :: Temp,PresshPa,Freq_wind
    real(r8_), intent(out) :: RayCal ! result
    integer,   intent(out) :: error_flag
    
    ! local variables
    real(r8_)   :: rNa_int,rNb_int

    ! ========================================================
    !
    ! init

    error_flag = no_error
    RayCal = missing_indicator_real
    
    ! binary file containing lookup table should already have been
    ! loaded by the Load_LUT_RayCalib subroutine
    ! to be sure, check the TableLoaded switch
    IF (.not. TableLoaded) THEN
       error_flag = error_programming
       return
    END IF
    
    ! ============  Interpolate calibration table  ==============
    call INTERP_3D(rPtab,nP,PresshPa,rTtab,nT,Temp,rFtab,nF,Freq_wind, &
                   rNa,rNa_int)
    call INTERP_3D(rPtab,nP,PresshPa,rTtab,nT,Temp,rFtab,nF,Freq_wind, &
                   rNb,rNb_int)
    !print*,rNa_int,rNb_int

    ! Add counts in Rayleigh channels A and B
    RayCal=rNa_int+rNb_int
    
    return
    
  END SUBROUTINE LUT_RayCalib_Intpol
  !  #]
  SUBROUTINE Unload_LUT_RayCalib
    !  #[
    if (allocated(rPtab)) deallocate(rPtab)
    if (allocated(rTtab)) deallocate(rTtab)
    if (allocated(rFtab)) deallocate(rFtab)
    if (allocated(rNa))   deallocate(rNa)
    if (allocated(rNb))   deallocate(rNb)

    ! remember that we have the table no longer in memory
    TableLoaded = .false.
    
  END SUBROUTINE Unload_LUT_RayCalib
    !  #]
  !---------------------------
  SUBROUTINE copy_RBC_AB_data(RBC, error_flag)
    !  #[
    ! do the copying only once, to prevent a lot of double work.
 
    ! I/O variables
    TYPE(RBC_DataType), intent(in)  :: RBC
    integer,            intent(out) :: error_flag

    ! local variables
    integer    :: i,j,k,AllocStatus
    character(len=maxlen_IntAs) :: str_IntAs
    integer(IntAs)              :: value_IntAs

    error_flag = no_error

    str_IntAs = RBC%SPH%Num_P
    call GetIntAs(str_IntAs,value_IntAs,error_flag)
    IF (error_flag .ne. no_error) return
    nP=INT(value_IntAs)
    
    str_IntAs = RBC%SPH%Num_T
    call GetIntAs(str_IntAs,value_IntAs,error_flag)
    IF (error_flag .ne. no_error) return
    nT=INT(value_IntAs)
    
    str_IntAs = RBC%SPH%Num_fd
    call GetIntAs(str_IntAs,value_IntAs,error_flag)
    IF (error_flag .ne. no_error) return
    nF=INT(value_IntAs)
    
    allocate(rPtab(nP), rTtab(nT), rFtab(nF), STAT=AllocStatus)
    if (AllocStatus .ne. 0) then
       error_flag = error_allocate
       return
    end if
    
    allocate(rNa(nP,nT,nF), rNb(nP,nT,nF), STAT=AllocStatus)
    if (AllocStatus .ne. 0) then
       error_flag = error_allocate
       return
    end if
    
    rPtab  =REAL(RBC%RBC_DS%P_Grid/100)  ! need Pressures in hPa
    rTtab  =REAL(RBC%RBC_DS%T_Grid/100)  ! need Temperatures in K
    rFtab  =REAL(RBC%RBC_DS%Fd)          ! in Hz
       
    !Na_F(nP,nT,nfd)
    do i=1,nP
       do j=1,nT
          do k=1,nF
             rNa(i,j,k) =REAL(RBC%RBC_DS%Nab_PTFd(i)%Nab_TFd(j)%Na_Fd(k)) ! in Hz
             rNb(i,j,k) =REAL(RBC%RBC_DS%Nab_PTFd(i)%Nab_TFd(j)%Nb_Fd(k)) ! in Hz
          end do
       end do
    end do

    ! remember that we have the table in memory
    TableLoaded = .true.
 
  END SUBROUTINE Copy_RBC_AB_data
    !  #]
  SUBROUTINE LUT_RBC_Intpol(Temp,PresshPa,Freq_wind,RayCal,error_flag)
    !  #[

    !  Description:
    !    This routine retrieves the Rayleigh calibration factor (C1) from
    !    lookup tables in the RBC datastructure. 
    !    The lookup table inlcudes the number of counts on the
    !    Rayleigh channels A and B as a function of pressure, temperature and
    !    Doppler shifted frequency.
    !
    ! Input
    ! =====
    !    Temp       ! Temperature (K)
    !    PresshPa   ! Pressure (hPa)
    !    Freq_wind  ! Frequency of Doppler shift (MHz)
    !    Lookup table in binary format
    !
    ! Output
    ! ======
    !    RayCal     ! Rayleigh channel calibration value
    !    error_flag ! success or not
    !
    ! Implicit imput:
    !    copied values from the RBC datastructure
    !    in rPtab,rTtab,rFtab, rNa, rNb arrays
    !    using dimensions nT, nP, nfd
    !
    ! Method
    ! ======
    !    Interpolate table to input values    
    !
    ! Modifications:
    ! =============
    !    26-apr-2006 G.J. Marseille, initial version
    !    19-sep-2006 J. de kloe, added error_flag and TableLoaded switch
    !           and changed press unit to hPa to prevent double conversion
    ! 
    ! ===========================================================

    ! I/O variables
    real(r8_),          intent(in)  :: Temp,PresshPa,Freq_wind
    real(r8_),          intent(out) :: RayCal ! result
    integer,            intent(out) :: error_flag
    
    ! local variables
    real(r8_)  :: rNa_int,rNb_int

    ! ========================================================
    !
    ! init

    error_flag = no_error
    RayCal = missing_indicator_real

    ! RBC data should have been copied by the init routine
    ! to be sure, check the TableLoaded switch
    IF (.not. TableLoaded) THEN
       error_flag = error_programming
       return
    END IF

    ! ============  Interpolate calibration table  ==============
    call INTERP_3D(rPtab,nP,PresshPa,rTtab,nT,Temp,rFtab,nF,Freq_wind, &
                   rNa,rNa_int)
    call INTERP_3D(rPtab,nP,PresshPa,rTtab,nT,Temp,rFtab,nF,Freq_wind, &
                   rNb,rNb_int)
    !print*,Na_int,Nb_int

    ! Add counts in Rayleigh channels A and B
    RayCal=rNa_int+rNb_int
    
    return
    
  END SUBROUTINE LUT_RBC_Intpol

  !  #]
  SUBROUTINE Unload_RBC_AB_data
    !  #[
    if (allocated(rPtab)) deallocate(rPtab)
    if (allocated(rTtab)) deallocate(rTtab)
    if (allocated(rFtab)) deallocate(rFtab)
    if (allocated(rNa))   deallocate(rNa)
    if (allocated(rNb))   deallocate(rNb)

    ! remember that we have the table no longer in memory
    TableLoaded = .false.
    
  END SUBROUTINE Unload_RBC_AB_data
    !  #]
  !---------------------------
  SUBROUTINE Read_LUT_RayCalib(Temp,PresshPa,Freq_wind,RayCal,error_flag)
    !  #[
    !  Description:
    !    This routine loads the LUT, then gets a value from it
    !    by calling LUT_RayCalib_Intpol()
    !
    ! Input
    ! =====
    !    Temp       ! Temperature (K)
    !    PresshPa   ! Pressure (hPa)
    !    Freq_wind  ! Frequency of Doppler shift (MHz)
    !    Lookup table in binary format
    !
    ! Output
    ! ======
    !    RayCal     ! Rayleigh channel calibration value
    !
    ! Method
    ! ======
    !    Read lookup table from binary files
    !    Interpolate table to input values    
    !
    ! Modifications:
    ! =============
    !    26-Apr-2006 G.J. Marseille, initial version
    !    19-Sep-2006 J. de Kloe, removed double code
    ! 
    ! ===========================================================
    
    ! I/O variables
    real(r8_),   intent(in)  :: Temp,PresshPa,Freq_wind
    real(r8_),   intent(out) :: RayCal
    integer,     intent(out) :: error_flag
    
    ! init
    error_flag = no_error
    
    ! if needed, open the binary file containing lookup table
    if (.not. TableLoaded) then
       call Load_LUT_RayCalib(error_flag)
       if (error_flag .ne. no_error) return
    end if
    
    ! ============  Interpolate calibration table  ==============
    call LUT_RayCalib_Intpol(Temp,PresshPa,Freq_wind,RayCal,error_flag)
    if (error_flag .ne. no_error) return
    
    return
    
  END SUBROUTINE Read_LUT_RayCalib
  !  #]
  SUBROUTINE INTERP_3D(x1,n1,v1,x2,n2,v2,x3,n3,v3,y3d,yint)
    !  #[
    ! This routine computes the interpolated value in a 3-dimensional array.
    ! 
    ! Input
    ! =====
    !    for i=1,2,3:
    !    xi(1:ni) : array of input values
    !    ni       : array size of x1
    !    vi       : interpolation value for array x1
    !    y3d      : 3-dimensional table
    !
    !  Output
    !  ======
    !    yint     : interpolated table value
    !    
    ! Method
    ! ======
    !    For each input array the locations enclosing the interpolation
    !    value (vi) are found, using the HUNT routine, followed by 
    !    linear interpolation.
    !
    !
    ! ===================================================================
  
    ! I/O variables
    integer,   intent(in)  :: n1,n2,n3
    real(r8_), intent(in)  :: v1,v2,v3
    real(r8_), intent(in)  :: x1(1:n1),x2(1:n2),x3(1:n3),y3d(n1,n2,n3)
    real(r8_), intent(out) :: yint
    !integer,   intent(out) :: error_flag

    ! local variables
    integer   :: i2,i3
    integer   :: iLoc,ilow(3),iHigh(3)
    real(r8_) :: y1,dy,dz
    real(r8_) :: yy(2,2,2),xy(2),zz(2),ddx(3)
  
    ! =======   Find the locations in the input arrays 
    ! =======   that enclose the interpolation value
    !
    ! checking for table boundaries is done inside the 
    ! hunt_table routine defined below

    ! 1) 1st input array
    call hunt_table(n1,x1,v1,iLoc)
    ilow(1)  = iLoc
    iHigh(1) = iLoc+1
    !
    ! 2) 2nd input array
    call hunt_table(n2,x2,v2,iLoc)
    ilow(2)  = iLoc
    iHigh(2) = iLoc+1
    !
    ! 3) 3rd input array
    call hunt_table(n3,x3,v3,iLoc)
    ilow(3)  = iLoc
    iHigh(3) = iLoc+1
    
    !  print*,ilow(1),iHigh(1),ilow(2),iHigh(2),ilow(3),iHigh(3)
    
    ! =======   Linear Interpolation
    
    ddx(1) = (v1-x1(ilow(1)))/(x1(ihigh(1))-x1(ilow(1)))
    ddx(2) = (v2-x2(ilow(2)))/(x2(ihigh(2))-x2(ilow(2)))
    ddx(3) = (v3-x3(ilow(3)))/(x3(ihigh(3))-x3(ilow(3)))
    
    yy(1,1,1) = y3d(ilow(1) ,ilow(2) ,ilow(3))
    yy(1,1,2) = y3d(ilow(1) ,ilow(2) ,ihigh(3))
    yy(1,2,1) = y3d(ilow(1) ,ihigh(2),ilow(3))
    yy(1,2,2) = y3d(ilow(1) ,ihigh(2),ihigh(3))
    yy(2,1,1) = y3d(ihigh(1),ilow(2) ,ilow(3))
    yy(2,1,2) = y3d(ihigh(1),ilow(2) ,ihigh(3))
    yy(2,2,1) = y3d(ihigh(1),ihigh(2),ilow(3))
    yy(2,2,2) = y3d(ihigh(1),ihigh(2),ihigh(3))
    
    do i3=1,2
       do i2=1,2
          y1=yy(1,i2,i3)
          dy=yy(2,i2,i3)-y1
          xy(i2)=y1+dy*ddx(1)
       enddo
     
       dy=xy(2)-xy(1)
       zz(i3)=xy(1)+dy*ddx(2)
    enddo
  
    dz=zz(2)-zz(1)
    yint=zz(1)+dz*ddx(3)
  
    !  write(*,*) yy(1,1,1),yy(1,1,2),yy(1,2,1),yy(1,2,2)
    !  write(*,*) yy(2,1,1),yy(2,1,2),yy(2,2,1),yy(2,2,2)
    !  print*,''
    !  print*,'yint ',yint  
    
  END SUBROUTINE INTERP_3D
  !  #]
  SUBROUTINE HUNT_TABLE(n,xx,x,iLoc)
    !  #[
    ! Extension of rount HUNT:
    !   1) to provide an initial guess
    !   2) To give warnings if values are out of array range
    ! 
    ! Input
    ! =====
    !    xx       : array of input values
    !    n        : array size of x 
    !    x        : interpolation value for array x
    !    iLoc     : initial guess
    !
    !  Output
    !  ======
    !    iLoc     : value such that x is between xx(i0) and xx(i0+1)
    !    
    !  
    !
    ! ===================================================================

    ! I/O variables
    integer,   intent(in)    :: n
    integer,   intent(inout) :: iLoc
    real(r8_), intent(in)    :: x,xx(1:n)
    
    ! local variables
    real(r8_) :: res
  
    ! Estimate array resolution
    res=1._r8_
    if (n .gt. 1) then
       res=abs(xx(n)-xx(1))/(n-1)
    endif
    if (res .eq. 0._r8_ ) res=1._r8_
    iLoc=INT(x/res)                  ! initial estimate
    call hunt(n,xx,x,iLoc)
    if (iLoc .eq. 0) then
       write(*,*) 'INTERP_3D: WARNING. value out of table range '
       iLoc=1
    endif
    if ( iLoc .eq. n) then
       write(*,*) 'INTERP_3D: Warning. value out of table range '
       iLoc=n-1
    endif
  
  END SUBROUTINE HUNT_TABLE
  !  #]
  !---------------------------
END MODULE LUT_RayCalib_handling
