module OpticalProperties
  !  #[ Description:
  !     A module to
  !     -determine extinction needed to determine vcog of the rangebins
  !      and optionally for classification and backscatter ratio calculation
  !     -determine/copy the Backscatter Ratio values needed
  !      by the processor for RayleighBrillouin processing
  !      and optionally for classification.
  !
  ! created: 05-Jan-2006 (based on routines copied from Classification.F90)
  !
  ! Modifications: 
  !   05-Jan-2006 J. de Kloe     initial version
  !   24-Jan-2006 J. de Kloe      limited nr of printed errors
  !   12-May-2006 D. Tan          Null use of ParKind1 module
  !   30-May-2006 G.J. Marseille  modification to subroutine Calc_Extinction
  !      1. New subroutine to estimate the Fabry-Perot (FP) signal for an 
  !         atmosphere with molecules only: calc_FP_Molecules_Only
  !         This routine includes FP channel calibration from the upper bin.
  !      2. Detection of particle (aerosol or cloud) layers through cross-talk
  !         on the FP channel by comparing the measured and expected (no layer
  !         assumed) FP signals. Computation of the particle layer 
  !         transmission.
  !      3. Computation of aerosol transmission (other than layers) from
  !         the measured and molecules-only FP signals, see TN2.3, section 7.
  !   22-Jun-2006 G.J. Marseille
  !     Addition of a power k in the FP particle transmission retrieval 
  !     scheme, in subroutine Calc_Extinction
  !   23-Jun-2006 J. de Kloe 
  !         reinserted the old extinction calculation to enable
  !         adding this file to the L2BP without breaking tests
  !         Also wired to a number of switches defined in 
  !         l2b_proc_settings to enable use of the old code by default
  !         and use of the new code if needed by setting proper command
  !         line switches.
  !   09-Aug-2006 J. de Kloe some small interface changes
  !   17-Aug-2006 P. Poli    Removed obsolete variables
  !   15-Sep-2006 J. de Kloe adapted Match_Mie_And_Rayleigh_Bins, and the
  !                          copy*from_Rayl_to_Mie, copy*from_Mie_to_Rayl
  !                          routines to unlikely rangebin definitions
  !   20-Sep-2006 J. de Kloe added a large number of missing tests and related
  !                          code to Calc_Extinction_Iterative and
  !                          calc_FP_Molecules_Only to make the code more
  !                          robust for missing or invalid data.
  !   22-Sep-2006 P. Poli    set flag rho_ok to TRUE when calculating 
  !                          scattering ratio from data
  !   05-Oct-2006 J. de Kloe added use of the WorkingDataStructure
  !   23-Oct-2006 D. Tan     Use of WorkingDataStructure for old L2B-working
  !   28-Aug-2007 D. Tan     Use of WorkingDataStructure for old L1B-working
  !   04-Sep-2007 P. Poli    change L1B_DataType to L1B_BRC_DataType
  !   02-Oct-2007 J. de Kloe added code to use RBC file in stead of 
  !                          the custom binary RayCalib file
  !   16-Jan-2008 J. de Kloe phase out integer kind i_
  !  #]
  !  #[ Modules used:

  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE L1B_DataStructure, only: L1B_BRC_DataType
  USE L2B_Proc_Settings, only: PartExt_Method_Undefined, &
                               PartExt_Method_Old, PartExt_Method_Iterative, &
                               MolExt_Method_Undefined,&
                               MolExt_Method_Parametrised, &
                               MolExt_Method_Proper,&
                               use_lite_test_data
  USE L2B_AuxPar_DataStructure, only: Scat_Ratio_Undefined, &
                               Scat_Ratio_from_L1B, Scat_Ratio_from_MieRayl,&
                               Scat_Ratio_from_RaylOnly, Scat_Ratio_Dont_Use
  USE Working_DataStructure, only: WorkingDataType,&
                           WD_MieRangeGateType      => MieRangeGateType,&
                           WD_RayleighRangeGateType => RayleighRangeGateType

  !support modules 
  USE Logging, only: logmsg,log_debug,log_info,log_progress,log_warn,log_error
  USE AeolusConstants, only: lambda,pi
  USE LunManager, only: get_lun, free_lun
  USE Numerics, only: r8_, real2int, &
                      missing_int, missing_real, &
                      missing_indicator_real, missing_indicator_integer
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE ErrorHandler, only: no_error, error_programming, error_program_usage
  
  USE MolScat,  only: InitMolScatModule,calc_molecular_extinction, &
                      calc_molecular_backscatter, get_ref_mol_scat
        
!  USE LUT_RayCalib_handling, only: Load_LUT_RayCalib, Unload_LUT_RayCalib
!  USE LUT_RayCalib_handling, only: LUT_RayCalib_Intpol
  USE LUT_RayCalib_handling, only: copy_RBC_AB_data, Unload_RBC_AB_data
  USE LUT_RayCalib_handling, only: LUT_RBC_Intpol
  USE EE_CFI_Datatypes, only: IntAuc, IntAs
  USE rbc_datastructure, only: RBC_DataType

  !  #]
  !  #[ Parameter/Variable/Type declarations

  implicit none

  ! this parameter is just to prevent dividing by zero, so its 
  ! actual value is not very important
  real(r8_), parameter :: Rayleigh_signal_level_too_low = 1.0e-30_r8_

  ! the default verbosity setting is Error (as silent as possible)
  integer, private, save :: LocalVerbosity = VerbosityError

  !  #]
contains
  !------------------------------------------
  subroutine InitOpticalProperties(PartExt_Method, &
                                   error_flag,ChoosenVerbosity,RBC)
    !  #[
    integer,            intent(in)  :: PartExt_Method   ! input
    integer,            intent(out) :: error_flag       ! output
    integer,  optional, intent(in)  :: ChoosenVerbosity ! optional input
    TYPE(RBC_DataType), optional, intent(in)  :: RBC    ! optional input

    IF (present(ChoosenVerbosity)) LocalVerbosity = ChoosenVerbosity
    
    call InitMolScatModule()

    ! only load this LUT when the algorithm needs it
    IF (PartExt_Method .eq. PartExt_Method_Iterative) THEN
       ! this loads the C1 calibration LUT, needed to calculate the
       ! expected FP signal from p and T
       !call Load_LUT_RayCalib(error_flag)
       !if (error_flag .ne. no_error) return

       ! replaced by use of the RBC table
       IF (present(RBC)) THEN
          call copy_RBC_AB_data(RBC,error_flag)
          if (error_flag .ne. no_error) return
       ELSE
          call logmsg(log_error,"ERROR in InitOpticalProperties: "//&
               "the optional RBC parameter should be supplied when "//&
               "PartExt_Method_Iterative is used")
          error_flag = error_programming
          return
       END IF
    END IF

  end subroutine InitOpticalProperties
      !  #]
  subroutine CloseOpticalProperties
    !  #[
!    ! free the memory used by the C1 calibration LUT
!    call Unload_LUT_RayCalib

    call Unload_RBC_AB_data

  end subroutine CloseOpticalProperties
    !  #]
  subroutine Match_Mie_And_Rayleigh_Bins(WorkingData, error_flag)
    !  #[
    TYPE(WorkingDataType), intent(inout) :: WorkingData    ! input/output
    integer,               intent(out)   :: error_flag     ! output

    ! local variables
    integer(IntAs)          :: m,a_r,a_m !, N_meas
    !integer                :: nr_of_BRCs_in_file
    !integer                :: nr_of_measurements_in_file
    !integer                :: nr_of_range_gates_in_file
    !character(len=10)      :: dummyline
    !real(r_), dimension(7) :: data
    TYPE(WD_MieRangeGateType), pointer      :: WD_MieRangeGate
    TYPE(WD_RayleighRangeGateType), pointer :: WD_RayleighRangeGate
    real(r8_)   :: Rayleigh_start, Rayleigh_end, Mie_start, Mie_end, eps
    real(r8_)   :: Rayleigh_rangegate_size, Mie_rangegate_size
    integer     :: bin1, bin2
    logical     :: both_tests_should_pass, test1_passed, test2_passed, &
                   tests_passed

    ! limit the amount of error/warning prints in case of trouble
    integer,     parameter :: Max_Num_Printed_Errors = 25
    integer,     save      :: Num_Printed_Errors = 0

    error_flag = no_error
    eps = 1. ! allow for 1 [m] inaccuracy in rangebin definition
    
    meas_loop: DO m=1,WorkingData%nr_of_measurements

       !  #[ set the pointers from Rayleigh=>Mie range-gates

       ! for each Rayleigh Rangegate
       rangegate_loop1: DO a_r=1,WorkingData%RayleighMeas(m)%nr_of_range_gates
          
          ! point to this RangeGate for shorter notation
          WD_RayleighRangeGate => WorkingData%RayleighMeas(m)%RangeGate(a_r)
          Rayleigh_start = WD_RayleighRangeGate%alt_start
          Rayleigh_end   = WD_RayleighRangeGate%alt_end
          
          ! init to -1
          WD_RayleighRangeGate%first_mie_rangebin = -1
          WD_RayleighRangeGate%last_mie_rangebin  = -1
          
          ! loop along all Mie range gates and mark the overlapping
          rangegate_loop2: DO a_m=1,WorkingData%MieMeas(m)%nr_of_range_gates
             
             ! if this rayleigh bin has overlap with this mie bin
             
             ! point to this RangeGate for shorter notation
             WD_MieRangeGate => WorkingData%MieMeas(m)%RangeGate(a_m)
             Mie_start = WD_MieRangeGate%alt_start
             Mie_end   = WD_MieRangeGate%alt_end
             
             ! there is overlap if:
             ! -Rayleigh_start =< Mie_start AND Rayleigh_end > Mie_start
             ! (i.e. Mie_start falls inside the rayleigh bin)
             ! or
             ! -Rayleigh_start < Mie_end AND Rayleigh_end >= Mie_end
             ! (i.e. Mie_end falls inside the rayleigh bin)
             ! or
             ! -Rayleigh_start > Mie_start AND Rayleigh_end < Mie_end
             ! (i.e. the full rayleigh bin falls inside the mie bin)
             !
             IF ( ( (Rayleigh_start .le. Mie_start) .and. &
                    (Rayleigh_end   .gt. Mie_start)       ) .or. &
                  ( (Rayleigh_start .lt. Mie_end  ) .and. &
                    (Rayleigh_end   .ge. Mie_end  )       ) .or. &
                  ( (Rayleigh_start .gt. Mie_start) .and. &
                    (Rayleigh_end   .lt. Mie_end  )       )       ) THEN
                
                ! set the first pointer only at the first match
                IF (WD_RayleighRangeGate%first_mie_rangebin .eq. -1) &
                     WD_RayleighRangeGate%first_mie_rangebin = a_m
                
                ! allways set the last pointer
                WD_RayleighRangeGate%last_mie_rangebin = a_m
                
             END IF
             
          END DO rangegate_loop2
          
          IF (LocalVerbosity .ge. VerbosityReport) THEN
             print *,"test: Rayleigh_start=",Rayleigh_start,&
                  " Rayleigh_end=",Rayleigh_end
             print *,"first_mie_rangebin = ",&
                  WD_RayleighRangeGate%first_mie_rangebin,&
                  " last_mie_rangebin = ",&
                  WD_RayleighRangeGate%last_mie_rangebin
          END IF
          
       END DO rangegate_loop1
       !  #]
       
       !  #[ set the pointers from Mie=>Rayleigh range-gates

       ! for each Mie Rangegate
       rangegate_loop3: DO a_m=1,WorkingData%MieMeas(m)%nr_of_range_gates
          
          ! point to this RangeGate for shorter notation
          WD_MieRangeGate => WorkingData%MieMeas(m)%RangeGate(a_m)
          Mie_start = WD_MieRangeGate%alt_start
          Mie_end   = WD_MieRangeGate%alt_end
          
          ! init to -1
          WD_MieRangeGate%first_rayleigh_rangebin = -1
          WD_MieRangeGate%last_rayleigh_rangebin  = -1
          
          ! loop along all Rayleigh range gates and mark the overlapping
          rangegate_loop4: DO a_r=1,WorkingData%RayleighMeas(m)%nr_of_range_gates
             
             ! if this mie bin has overlap with this rayleigh bin
             
             ! point to this RangeGate for shorter notation
             WD_RayleighRangeGate => WorkingData%RayleighMeas(m)%RangeGate(a_r)
             Rayleigh_start = WD_RayleighRangeGate%alt_start
             Rayleigh_end   = WD_RayleighRangeGate%alt_end
             
             ! there is overlap if:
             ! -Mie_start =< Rayleigh_start AND Mie_end > Rayleigh_start
             ! (i.e. Rayleigh_start falls inside the mie bin)
             ! or
             ! -Mie_start < Rayleigh_end AND Mie_end >= Rayleigh_end
             ! (i.e. Rayleigh_end falls inside the mie bin)
             ! or
             ! -Mie_start > Rayleigh_start AND Mie_end < Rayleigh_end
             ! (i.e. the full mie bin falls inside the rayleigh bin)
             !
             IF ( ( (Mie_start .le. Rayleigh_start) .and. &
                    (Mie_end   .gt. Rayleigh_start)       ) .or. &
                  ( (Mie_start .lt. Rayleigh_end  ) .and. &
                    (Mie_end   .ge. Rayleigh_end  )       ) .or. &
                  ( (Mie_start .gt. Rayleigh_start) .and. &
                    (Mie_end   .lt. Rayleigh_end  )       )      ) THEN
                
                ! set the first pointer only at the first match
                IF (WD_MieRangeGate%first_rayleigh_rangebin .eq. -1) &
                     WD_MieRangeGate%first_rayleigh_rangebin = a_r
                
                ! allways set the last pointer
                WD_MieRangeGate%last_rayleigh_rangebin = a_r
                
             END IF
             
          END DO rangegate_loop4
          
          IF (LocalVerbosity .ge. VerbosityReport) THEN
             print *,"test: Mie_start=",Mie_start," Mie_end=",Mie_end
             print *,"first_rayleigh_rangebin = ",&
                  WD_MieRangeGate%first_rayleigh_rangebin,&
                  " last_rayleigh_rangebin = ",&
                  WD_MieRangeGate%last_rayleigh_rangebin
          END IF
          
       END DO rangegate_loop3
       !  #]
       
       IF (LocalVerbosity .ge. VerbosityReport) print *,"1st test"
       !  #[ check if Rayleigh => Mie rangebins are properly aligned
       
       ! in the classification algorithm we need the rangebins to
       ! be aligned, i.e. a Mie rangebin must contain 1 or more complete
       ! Rayleigh rangebins, or vice versa.
       
       ! for each Rayleigh Rangegate
       rangegate_loop5: DO a_r=1,WorkingData%RayleighMeas(m)%nr_of_range_gates
          
          ! point to this RangeGate for shorter notation
          WD_RayleighRangeGate => WorkingData%RayleighMeas(m)%RangeGate(a_r)
          Rayleigh_start = WD_RayleighRangeGate%alt_start
          Rayleigh_end   = WD_RayleighRangeGate%alt_end
          
          ! init to -1
          bin1 = WD_RayleighRangeGate%first_mie_rangebin
          bin2 = WD_RayleighRangeGate%last_mie_rangebin
          
          Mie_start = missing_indicator_real
          Mie_end   = missing_indicator_real
          
          ! if the sizes differ, it is sufficient for the moment to see
          ! if either the start or end of the rangegate is aligned.
          ! At the moment we can have multiple Mie rangegates in one
          ! Rayleigh rangegate, but not vice-versa, so this is automatically
          ! taken care of if the first/last pointers are set correctly.
          
          ! remember the profiles are top-down !
          IF (bin2 .gt. 0) &
               Mie_start = WorkingData%MieMeas(m)%RangeGate(bin2)%alt_start
          IF (bin1 .gt. 0) &
               Mie_end   = WorkingData%MieMeas(m)%RangeGate(bin1)%alt_end
          
          IF (.not. missing_real(Mie_start)) THEN
             IF ( abs(Rayleigh_start - Mie_start) .gt. eps) THEN
                IF (Num_Printed_Errors .lt. Max_Num_Printed_Errors) THEN
                   Num_Printed_Errors = Num_Printed_Errors+1
                   print *,"ERROR: rangebins seem not aligned !!!"
                   print *,"REALACC(6)"
                   print *,"Rayleigh_start=",Rayleigh_start,&
                        " Rayleigh_end=",Rayleigh_end
                   print *,"first_mie_rangebin = ",&
                        WD_RayleighRangeGate%first_mie_rangebin,&
                        " last_mie_rangebin = ",&
                        WD_RayleighRangeGate%last_mie_rangebin
                   print *,"Mie_start = ",Mie_start," Mie_end = ",Mie_end
                   print *,"ENDREALACC"
                   
                   ! our L1B testfile seems to fail this test, so disable it
                   print *,"ignoring this problem for now ..."
                   !error_flag = error_reading_file
                   !return
                END IF
             END IF
          END IF
          
          IF (.not. missing_real(Mie_end)) THEN
             IF ( abs(Rayleigh_end - Mie_end) .gt. eps) THEN
                IF (Num_Printed_Errors .lt. Max_Num_Printed_Errors) THEN
                   Num_Printed_Errors = Num_Printed_Errors+1
                   print *,"ERROR: rangebins seem not aligned !!!"
                   print *,"REALACC(6)"
                   print *,"Rayleigh_start=",Rayleigh_start,&
                        " Rayleigh_end=",Rayleigh_end
                   print *,"first_mie_rangebin = ",&
                        WD_RayleighRangeGate%first_mie_rangebin,&
                        " last_mie_rangebin = ",&
                        WD_RayleighRangeGate%last_mie_rangebin
                   print *,"Mie_start = ",Mie_start," Mie_end = ",Mie_end
                   print *,"ENDREALACC"
                   
                   ! our L1B testfile seems to fail this test, so disable it
                   print *,"ignoring this problem for now ..."
                   !error_flag = error_reading_file
                   !return
                END IF
             END IF
          END IF
          
       end DO rangegate_loop5
       !  #]
          
       IF (LocalVerbosity .ge. VerbosityReport) print *,"2nd test"
       !  #[ check if Mie => Rayleigh rangebins are properly aligned
       
       ! we the classification algorithm we need the rangebins to
       ! be aligned, i.e. a Mie rangebin must contain 1 or more complete
       ! Rayleigh rangebins, or vice versa.
       
       ! for each Mie Rangegate
       rangegate_loop6: DO a_m=1,WorkingData%MieMeas(m)%nr_of_range_gates
          
          ! point to this RangeGate for shorter notation
          WD_MieRangeGate => WorkingData%MieMeas(m)%RangeGate(a_m)
          Mie_start = WD_MieRangeGate%alt_start
          Mie_end   = WD_MieRangeGate%alt_end
          
          ! 
          bin1 = WD_MieRangeGate%first_rayleigh_rangebin
          bin2 = WD_MieRangeGate%last_rayleigh_rangebin
          
          Rayleigh_start = missing_indicator_real
          Rayleigh_end   = missing_indicator_real
          
          Rayleigh_rangegate_size = Rayleigh_end - Rayleigh_start
          Mie_rangegate_size      = Mie_end      - Mie_start
          
          IF ( abs(Rayleigh_rangegate_size - Mie_rangegate_size) .lt. eps) THEN
             ! if Mie_rangegate_size .eq. Rayleigh_rangegate_size
             ! both tests below should be passed successfully, since the start
             ! and the end of the rangegate should be aligned
             both_tests_should_pass = .true.
          ELSE
             ! if the sizes differ, it is sufficient for the moment to see
             ! if either the start or end of the rangegate is aligned, since
             ! we never have more than 2 Mie rangegates in 1 Rayleigh rangegate
             ! (and never multiple Rayleigh rangegates in 1 Mie rangegate)
             both_tests_should_pass = .false.
          END IF
          
          ! remember the profiles are top-down !
          IF (bin2 .gt. 0) Rayleigh_start = &
               WorkingData%RayleighMeas(m)%RangeGate(bin2)%alt_start
          IF (bin1 .gt. 0) Rayleigh_end   = &
               WorkingData%RayleighMeas(m)%RangeGate(bin1)%alt_end
          
          test1_passed = .true.
          IF (.not. missing_real(Rayleigh_start)) THEN
             IF ( abs(Rayleigh_start - Mie_start) .gt. eps) &
                  test1_passed = .false.
          END IF
          
          test2_passed = .true.
          IF (.not. missing_real(Rayleigh_end)) THEN
             IF ( abs(Rayleigh_end - Mie_end) .gt. eps) test2_passed = .false.
          END IF
          
          tests_passed = .false.
          IF (both_tests_should_pass) THEN
             IF (test1_passed .and. test2_passed) tests_passed = .true.
          ELSE
             IF (test1_passed .or.  test2_passed) tests_passed = .true.
          END IF
          
          IF (.not. tests_passed) THEN
             IF (Num_Printed_Errors .lt. Max_Num_Printed_Errors) THEN
                Num_Printed_Errors = Num_Printed_Errors+1
                print *,"ERROR: rangebins seem not aligned !!!"
                print *,"REALACC(6)"
                print *,"Mie_start = ",Mie_start," Mie_end = ",Mie_end
                print *,"first_rayleigh_rangebin = ",&
                     WD_MieRangeGate%first_rayleigh_rangebin,&
                     " last_rayleigh_rangebin = ",&
                     WD_MieRangeGate%last_rayleigh_rangebin
                print *,"Rayleigh_start=",Rayleigh_start,&
                     " Rayleigh_end=",Rayleigh_end
                print *,"ENDREALACC"
                
                ! our L1B testfile seems to fail this test, so disable it
                print *,"ignoring this problem for now ..."
                !error_flag = error_reading_file
                !return
             END IF
          END IF
          
       end DO rangegate_loop6
       !  #]
          
       ! REMARK: it is not yet clear if gaps in between rangebins will
       ! or will not occur. A single Mie rangebin above the highest
       ! Rayleigh rangebin might still be usefull. This would mean a large
       ! gap between rangebin 1 and 2 in the Mie channel
       ! Therefore the next 2 tests only generate warnings.
       ! Overlaps however should never occur, so the tests below might be
       ! adapted in a later stage to let the processing fail when this occurs.
! REMARK: the current L1B datastructure does not allow such flexibility
!         in the rangebin definitions !!!!

! for L1B data this will allways be the case, due to the way
! the datastructure is defined, so this test could be removed
! when we stop using lite test data
       IF (LocalVerbosity .ge. VerbosityReport) print *,"3rd test"
       !  #[ test if Mie-rangebins are following each other without gaps

       rangegate_loop7: DO a_m=1,WorkingData%MieMeas(m)%nr_of_range_gates-1
          IF ( .not. missing_real(WorkingData%MieMeas(m)%RangeGate(a_m)%alt_start) .and.&
               .not. missing_real(WorkingData%MieMeas(m)%RangeGate(a_m+1)%alt_end) ) THEN
             ! remember that the rangebins are top-down at the moment
             IF (abs(WorkingData%MieMeas(m)%RangeGate(a_m)%alt_start - &
                  WorkingData%MieMeas(m)%RangeGate(a_m+1)%alt_end   ) .gt. eps) THEN
                print *,"WARNING: the Mie-rangebins seem to have a definition"
                print *,"problem"
                print *,"rangebin ",a_m," does not connect to rangebin ",a_m+1
                print *,"they either have a gap in between, or they overlap..."
             END IF
          END IF
       END DO rangegate_loop7
       !  #]
          
! for L1B data this will allways be the case, due to the way
! the datastructure is defined, so this test could be removed
! when we stop using lite test data
       IF (LocalVerbosity .ge. VerbosityReport) print *,"4th test"
       !  #[ test if Rayleigh-rangebins are following eachother without gaps

       rangegate_loop8: DO a_r=1,WorkingData%RayleighMeas(m)%nr_of_range_gates-1
          IF ( .not. missing_real(WorkingData%RayleighMeas(m)%&
               RangeGate(a_r)%alt_start) .and.&
               .not. missing_real(WorkingData%RayleighMeas(m)%&
               RangeGate(a_r+1)%alt_end) ) THEN
             ! remember that the rangebins are top-down at the moment
             IF (abs(WorkingData%RayleighMeas(m)%RangeGate(a_r)%alt_start - &
                  WorkingData%RayleighMeas(m)%RangeGate(a_r+1)%alt_end   ) .gt. eps) THEN
                print *,"WARNING: the Rayleigh-rangebins seem to have a"
                print *,"definition problem:"
                print *,"rangebin ",a_r," does not connect to rangebin ",a_r+1
                print *,"they either have a gap in between, or they overlap..."
             END IF
          END IF
       END DO rangegate_loop8
       !  #]

    END DO meas_loop

!    ! copy rangegate pointers from meas 1 to all next measurements
!    meas_loop: DO m=2,WorkingData%nr_of_measurements
!       rangegate_loop9: DO a_r=1,L1B%RayleighMeas(m)%nr_of_range_gates
          !  #[ check and copy Rayleigh pointers
!          RayleighRangeGate1 => L1B%RayleighMeas(1)%RangeGate(a_r)
!          RayleighRangeGate  => L1B%RayleighMeas(m)%RangeGate(a_r)
!          tests_passed = .true.
!          IF (abs(RayleighRangeGate1%alt_start - &
!                  RayleighRangeGate%alt_start) .gt. eps) tests_passed = .false.
!          IF (abs(RayleighRangeGate1%alt_end - &
!                  RayleighRangeGate%alt_end) .gt. eps) tests_passed = .false.
!          IF (.not. tests_passed) THEN
!             print *,"ERROR: rangebin definition for different L1B"
!             print *,"measurements within the same BRC seem different !!!"
!             print *,"m=1, r=",a_r," Rayleigh: alt_start = ",&
!                  RayleighRangeGate1%alt_start," alt_end = ",&
!                  RayleighRangeGate1%alt_end
!             print *,"m=",m,", r=",a_r," Rayleigh: alt_start = ",&
!                  RayleighRangeGate%alt_start," alt_end = ",&
!                  RayleighRangeGate%alt_end
!             error_flag = error_reading_file
!             return
!          END IF
!          ! all seems fine for this Rayleigh rangegate, so copy the pointers
!          RayleighRangeGate%first_mie_rangebin = &
!               RayleighRangeGate1%first_mie_rangebin
!          RayleighRangeGate%last_mie_rangebin = &
!               RayleighRangeGate1%last_mie_rangebin
          !  #]
!       END DO rangegate_loop9
!       rangegate_loop10: DO a_m=1,L1B%MieMeas(m)%nr_of_range_gates
          !  #[ check and copy Mie pointers
!          MieRangeGate1 => L1B%MieMeas(1)%RangeGate(a_m)
!          MieRangeGate  => L1B%MieMeas(m)%RangeGate(a_m)
!          tests_passed = .true.
!          IF (abs(MieRangeGate1%alt_start - &
!                  MieRangeGate%alt_start) .gt. eps) tests_passed = .false.
!          IF (abs(MieRangeGate1%alt_end - &
!                  MieRangeGate%alt_end) .gt. eps) tests_passed = .false.
!          IF (.not. tests_passed) THEN
!             print *,"ERROR: rangebin definition for different L1B"
!             print *,"measurements within the same BRC seem different !!!"
!             print *,"m=1, r=",a_r," Mie: alt_start = ",&
!                  MieRangeGate1%alt_start," alt_end = ",&
!                  MieRangeGate1%alt_end
!             print *,"m=",m,", r=",a_r," Mie: alt_start = ",&
!                  MieRangeGate%alt_start," alt_end = ",&
!                  MieRangeGate%alt_end
!             error_flag = error_reading_file
!             return
!          END IF
!          ! all seems fine for this Rayleigh rangegate, so copy the pointers
!          MieRangeGate%first_rayleigh_rangebin = &
!               MieRangeGate1%first_rayleigh_rangebin
!          MieRangeGate%last_rayleigh_rangebin = &
!               MieRangeGate1%last_rayleigh_rangebin
          !  #]
!       END DO rangegate_loop10
!    END DO meas_loop

  end subroutine Match_Mie_And_Rayleigh_Bins
  !  #]
  !------------------------------------------
  ! calculation from p,T
  !beta_m = calc_molecular_backscatter(pressure(z),temperature(z))
  ! climatological
  !beta_m = get_ref_mol_scat(wanted_z)
  subroutine Calc_Extinction(L1B, WorkingData, &
                             MolExt_Method, PartExt_Method, error_flag, &
                             use_lite_test_data, calibr_on_upper_bin)
    !  #[
    TYPE(L1B_BRC_DataType),   intent(in):: L1B                 ! input
    TYPE(WorkingDataType),intent(inout) :: WorkingData         ! input/output
    integer,              intent(in)    :: MolExt_Method       ! input
    integer,              intent(in)    :: PartExt_Method      ! input
    integer,              intent(out)   :: error_flag          ! output
    logical,optional,     intent(in)    :: use_lite_test_data  ! optional input
    logical,optional,     intent(in)    :: calibr_on_upper_bin ! optional input

    error_flag = no_error

    select case(PartExt_Method)
    case(PartExt_Method_Undefined)
       print *,"WARNING: the Method for determining particle extinction"
       print *,"WARNING: is still set to undefined !"
       print *,"WARNING: this probably points to a programming error !!!"
       error_flag = error_programming
       return
    case(PartExt_Method_Old)
       ! use the old algorithm (only to be used with
       ! tests using the Lite data, not to be used when doing processing
       ! of real aeolus data)
       call Calc_Extinction_Old(WorkingData, error_flag, &
                                use_lite_test_data, calibr_on_upper_bin)
       IF (error_flag .ne. no_error) return
       call Copy_Extinction_Rayl_to_Mie(WorkingData)
    case(PartExt_Method_Iterative)
       ! use Gert-Jan's new iterative algorithm
       call Calc_Extinction_Iterative(L1B, WorkingData, error_flag, &
                                      MolExt_Method, calibr_on_upper_bin) 
       IF (error_flag .ne. no_error) return
       call Copy_Extinction_Rayl_to_Mie(WorkingData)
    case default
       print *,"WARNING: this Method for determining particle extinction"
       print *,"WARNING: is not defined !"
       print *,"WARNING: this probably points to a programming error !!!"
       error_flag = error_programming
       return
    end select

  end subroutine Calc_Extinction
    !  #]
  subroutine Calc_Extinction_Iterative(L1B, WorkingData, error_flag, &
                                       MolExt_Method, calibr_on_upper_bin)
    !  #[
    TYPE(L1B_BRC_DataType), intent(in):: L1B                 ! input
    TYPE(WorkingDataType), intent(inout) :: WorkingData      ! input/output
    integer,            intent(out)   :: error_flag          ! output
    integer,            intent(in)    :: MolExt_Method       ! input
    logical,optional,   intent(in)    :: calibr_on_upper_bin ! input

    ! local variables
    logical     :: do_calibr_on_upper_bin
    logical     :: cross_talk
!    integer     :: nr_of_measurements ! not used at the moment
    integer     :: m,r,rr,cross_talk_rangegate,nbins
    integer     :: fileunit1, fileunit2
    real(r8_)   :: FP_molOnly_exp_signal_level ! FP (mol-only)expected value
    real(r8_)   :: FP_exp_signal_level ! FP expected value
    real(r8_)   :: FP_obs_signal_level ! FP observed value
    real(r8_)   :: cos_incidence_angle
    real(r8_)   :: rangegate_size
    real(r8_)   :: sigma_part
    real(r8_)   :: rangeGateTop,rangeGateBot,rangeGateCenter
    real(r8_)   :: incAngle,scatratio
    real(r8_)   :: cross_talk_thresh,tau_bin,tau_prev_bins,trans_part_above
    real(r8_)   :: trans_part_layers_above
    real(r8_)   :: kPow
    real(r8_),dimension(24)   :: tau_particle_rangegate
    real(r8_),dimension(24)   :: FP_Molecules_Only
    real(r8_)   :: trans_aer_above, trans_aer_prevbin
    real(r8_)   :: trans_aer_thisbin, trans_aer_thisbin_saved
    real(r8_)   :: trans_aer_and_cloud_prevbin

    ! this should be imported from the calibration datasets
    ! but for now calculate it in here
    real(r8_) :: calibration_and_solid_angle

    ! init
    error_flag = no_error

    if (use_lite_test_data) then
       print *,"ERROR in Calc_Extinction_Iterative():"
       print *,"This method should not be used on Lite data,"
       print *,"only on proper L1B data!"
       print *,"fix the rangebin size bug in the signal of the Lite data first"
       print *,"if you really need this ..."
       error_flag = error_program_usage
       return
    end if

    ! open ASCII file to save intermediate results
    fileunit1 = get_lun()
    open(unit=fileunit1,file='optProp.dat')

    ! Power used in the iterative particle transmission retrieval scheme:
    ! 0 < kPow < 1
    kPow=1.0_r8_

    ! cross talk transmission threshold
    cross_talk_thresh = 1.03_r8_**kPow

    ! init
    calibration_and_solid_angle = missing_indicator_real

    do_calibr_on_upper_bin = .false.
    IF (present(calibr_on_upper_bin)) &
         do_calibr_on_upper_bin = calibr_on_upper_bin
          
!print *,"r  size beta_mol  exp_rayl_signal obs_rayl_signal extinction exp_ext"

!Poli: this line is not used because the following loop from 1 to 1. Is this
!intentional?
! JK: no, this is just a hack to make runtime shorter during testing

!    nr_of_measurements = WorkingData%nr_of_measurements
    measloop: DO m=1, 1 !nr_of_measurements
       !  #[

       ! Calculate the expected Fabry-Perot signal for an atmosphere with 
       ! molecules only. Resulting profile is in FP_Molecules_Only
       call calc_FP_Molecules_Only(L1B, WorkingData, error_flag, m, &
                                   calibration_and_solid_angle,&
                                   FP_Molecules_Only,&
                                   MolExt_Method, calibr_on_upper_bin)
       IF (error_flag .ne. no_error) return

       IF (calibration_and_solid_angle .le. 0.) THEN
           print *,"error in Calc_Extinction():"
           print *,"invalid value for calibration_and_solid_angle: ",&
                     calibration_and_solid_angle
           error_flag = error_programming
           return
       END IF
       
       !print *,"result of calc_FP_Molecules_Only:"
       !print *,"FP_Molecules_Only = ",FP_Molecules_Only
    
       ! Initiate no cross-talk in first bin and no particle layers above 
       ! the first bin.
       cross_talk = .false.

       ! Poli: set cross_talk_rangegate to zero to make sure it is initialized
       cross_talk_rangegate = 0._r8_

       trans_part_layers_above = 1._r8_

       ! init to missing, to take care of the possibility that the
       ! Rangegateloop1 below stops halfway the profile due to missing data
       tau_particle_rangegate(:) = missing_indicator_real
       
       Rangegateloop1: DO r=1, WorkingData%RayleighMeas(m)%nr_of_range_gates
          !  #[

          !  Fabry-Perot expected signal
          !  this was determined by calling calc_FP_Molecules_Only()
          FP_molOnly_exp_signal_level = FP_Molecules_Only(r)
          IF (FP_molOnly_exp_signal_level .le. 0.) THEN
              print *,"error in Calc_Extinction():"
              print *,"invalid value for FP_molOnly_exp_signal_level: ",&
                     FP_molOnly_exp_signal_level
              error_flag = error_programming
              return
          END IF
        
          ! Initialize each rangegate as free of particles
          tau_particle_rangegate(r) = 1.
   
          ! Fabry-Perot measured signal (sum of channels A and B from L1B)
          FP_obs_signal_level = &
               WorkingData%RayleighMeas(m)%RangeGate(r)%obs_signal_level

          ! this statement is just to be able to export missing to the
          ! temp. outputfile in case FP_molOnly_exp_signal_level is missing
          FP_exp_signal_level = FP_molOnly_exp_signal_level

          IF (missing_real(FP_obs_signal_level)) &
               goto 111 ! cycle Rangegateloop1
          IF (missing_real(FP_molOnly_exp_signal_level)) &
               goto 111 ! cycle Rangegateloop1

          ! =====================================================
          !
          ! Now we have the expected signal level, 
          ! FP_molOnly_exp_signal_level (fp_mol), for the center of the
          ! current range bin for an atmosphere with molecules only.
          ! Assuming no contamination of aerosol backscatter into the FP
          ! channel through cross-talk, the expected signal is obtained
          ! from fp_mol and aerosol attenuation in the layers above:
          !
          !  fp_exp(r) = fp_mol(r)*trans_aer_above(r-1)**2
          !
          ! with trans_aer_above(i-1) the one-way aerosol transmission 
          ! between the satellite and the bottom of bin i-1.
          !
          ! Two types of particles are discriminated that are treated
          ! separately:
          !
          !    1) background aerosol
          !    2) particle layer (either cloud or aerosol).
          !
          ! The latter is characterized by large particle backscatter
          ! resulting into cross-talk on the Fabry-Perot channels.
          ! A cross-talk check on the FP signal is applied by comparing the
          ! measured, fp_meas, and expected, fp_exp, FP signal. 
          ! If fp_meas > fp_exp + eps, for threshold eps, cross-talk is
          ! detected. The bottom of the layer is searched and the layer
          ! 2-way transmission follows from the quotient of the observed and
          ! expected signal below the layer.
          !
          ! An estimate of the background aerosol transmission in the 
          ! current bin r, trans_aer_thisbin, is obtained from (see TN 2.3,
          ! section 7)
          !
          !                                 fp_meas(r) k
          !    trans_aer_thisbin(r) = [---------- ]
          !                                 fp_exp(r)
          !
          ! =====================================================
          
          IF (r .eq. 1) THEN
             ! Initiate one-way aerosol transmission of layers above, i.e.
             ! between satellite and top of first DWL range bin.
             trans_aer_above = 1.
          ELSE
             ! 1-way aerosol transmission from top of atmosphere 
             ! to the top of the current bin.
             trans_aer_above = trans_aer_above * trans_aer_prevbin
             !print *,"trans_aer_above,trans_aer_prevbin = ",&
             !     trans_aer_above,trans_aer_prevbin
             !print *,"updated trans_aer_above = ",trans_aer_above
          END IF

          IF (trans_aer_above .gt. 0.) THEN
             !  #[
             ! Expected signal on Fabry-Perot, given background aerosol 
             ! and (cloud) layer transmission aloft and asuming no 
             ! particle transmission in the current layer.
             ! This approximation is valid in the upper-troposphere and
             ! lower stratosphere and is used to detect particle layers
             ! and for aerosol optical properties computation.
             
             FP_exp_signal_level = FP_molOnly_exp_signal_level*&
                                   trans_part_layers_above**2*&
                                   trans_aer_above**2
             
             ! Parameter x in retrieval scheme (mimicing aerosol
             ! transmission in this bin). This parameter is also used for 
             ! particle layer detection
             trans_aer_thisbin=(FP_obs_signal_level/&
                                FP_exp_signal_level)**kPow
     
             !print *,"FP_obs_signal_level, FP_exp_signal_level = ",&
             !     FP_obs_signal_level, FP_exp_signal_level
             !print *,"testjos: trans_aer_above = ",trans_aer_above,&
             !     " trans_aer_thisbin = ",trans_aer_thisbin
          
             ! save the total cloud and aerosol transmission for this bin
             ! this is needed when trans_aer_thisbin in the next bin turns
             ! out to be zero, because the cross-talk code below deletes
             ! the trans_aer_thisbin value and sets it to 1 if it thinks
             ! it is inside a cloud layer....
             trans_aer_thisbin_saved = trans_aer_thisbin

             !In case of noisy data set particle transmission to 1
             !if ((trans_aer_thisbin .lt. 1.02_r8_) .and.&
             !    (trans_aer_thisbin .gt. 0.98_r8_)) &
             !   trans_aer_thisbin = 1._r8_
  
             !
             ! ***  Check for Cross-talk
             !
             if (trans_aer_thisbin .gt. cross_talk_thresh) then
                !print*,'cross talk detected ',r,trans_aer_thisbin
                if (.not. cross_talk) then
                   ! Remember cross talk range gate
                   cross_talk_rangegate = r
                endif
                cross_talk = .true.

                ! Discriminate between aerosol and particle layer (cloud).
                ! Set aerosol transmission equal to one. Particle layer
                ! transmission is computed later, when the layer bottom has
                ! been detected, by the quotient of the 
                ! measured and expected signal below the particle layer.
                trans_aer_thisbin = 1._r8_
             endif

             if (cross_talk) then
                !  #[
                !print*,'cross_talk ',trans_aer_thisbin
                if (trans_aer_thisbin .lt. 1._r8_/cross_talk_thresh) then
       
                   ! We are below the particle layer now. 
                   ! trans_aer_thisbin corresponds to the one-way particle  
                   ! transmission in the detected particle layer aloft.
                   ! Since the exact location of the layer is unknown, 
                   ! distribute the particle layer transmission equally over
                   ! the range bins that cover the particle layer.
                   ! Assume a homogeneous particle distribution.
                   
                   ! number of range bins that cover the particle layer
                   nbins=r-cross_talk_rangegate
        
                   ! Validity check        
                   if (nbins .lt. 1) then
                      print *,"error in Calc_Extinction():"
                      print *,"invalid number of particle layers : ",nbins
                      error_flag = error_programming
                      return
                   endif
        
                   ! one-way transmission per bin
                   ! remind that all bins above this one inside the cloud
                   ! layer have trans_aer set to 1, so the complete
                   ! tau is available in trans_aer_thisbin
                   if (trans_aer_thisbin .gt. 0._r8_) then
                      ! this breaks down when trans_aer_thisbin .eq. 0 !!!
                      tau_bin=trans_aer_thisbin**(1._r8_/nbins)
                      do rr=cross_talk_rangegate,r-1
                         !print*,"rr,tau_particle_rangegate(rr),tau_prev_bins"
                         !print*,rr,tau_particle_rangegate(rr),tau_prev_bins
                         tau_particle_rangegate(rr) = tau_bin*&
                              tau_particle_rangegate(rr)
                      enddo
                   else
                      tau_prev_bins = trans_aer_and_cloud_prevbin**(1._r8_/nbins)
                      do rr=cross_talk_rangegate,r-1
                         !print*,"rr,tau_particle_rangegate(rr),tau_prev_bins"
                         !print*,rr,tau_particle_rangegate(rr),tau_prev_bins
                         tau_particle_rangegate(rr) = tau_prev_bins*&
                              tau_particle_rangegate(rr)
                      enddo
                   end if

    
                   ! Update the transmission of all particle layers above
                   trans_part_layers_above = trans_part_layers_above*&
                               trans_aer_thisbin
                   !print*,r,cross_talk_rangegate,trans_aer_thisbin

                   ! The signal has passed the particle layer.
                   ! reset cross_talk
                   cross_talk = .false.
                else
                   ! As long as we are in the particle (cloud) layer, no other 
                   ! particle type (aerosol) is assumed.
                   trans_aer_thisbin = 1.
                endif     
                !  #]
             endif ! if (cross_talk)
  
             !print*,r,cross_talk,trans_aer_thisbin   
             ! Save particle transmission
             tau_particle_rangegate(r) = trans_aer_thisbin
             !  #]
          ELSE
             ! no signal available, so no result
             trans_aer_thisbin         = missing_indicator_real
             tau_particle_rangegate(r) = missing_indicator_real
          END IF ! trans_aer_above .gt. 0.
      
          ! save for use in the next lower rangebin
          trans_aer_prevbin           = trans_aer_thisbin
          trans_aer_and_cloud_prevbin = trans_aer_thisbin_saved

          ! jump point for end of rangegate loop 1
          ! needed to make sure the elements of the test output file are
          ! written, even in case of missing data
111       continue

          ! L1B backscatter ratio
          ! WARNING: this number is derived from the Mie signal, so if the
          ! rangebins do not match, this number does not represent
          ! the value in this rayleigh rangebin !!
          scatratio = L1B%PCD_ADS%Measurement_PCD(m)%&
                      Meas_Alt_Bin_PCD(r)%scattering_ratio
      
          !print *,"REALACC(6)"
          ! check if current layer is missing
          !if ( .not. missing_real(Temp)) then
          !   write(31,'(i2,1x,i5,1x,F8.1,1x,F5.1,1x,2(e10.4,1x),F8.2,1x,'//&
          !            '2(F8.3,1x),3(F5.3,1x!),1x,F6.3)') &
          !         r,NINT(rangeGateCenter),Press,Temp,mol_beta,mol_sigma,&
          !         RayCal,FP_obs_signal_level,&
          !         FP_exp_signal_level,&
          !         tau_this_rangegate,tau_this_rangegate_mol,&
          !         tau_this_rangegate_part,scatratio

             ! rangeGateCenter is only used in the writestatement below
             ! and nowhere else inside this loop.
             rangeGateTop    = WorkingData%RayleighMeas(m)%RangeGate(r)%alt_end
             rangeGateBot    = WorkingData%RayleighMeas(m)%RangeGate(r)%alt_start
             rangeGateCenter = (rangeGateTop+rangeGateBot)/2.

             write(fileunit1,'(i2,1x,i5,1x,2(F8.3,1x),2(E10.4,1x))') &
!          write(fileunit2,'(i2,1x,i5,2(F10.4,1x),2(E10.4,1x))') &
!             write(fileunit1,*) &
                  r,NINT(rangeGateCenter),&
                  FP_obs_signal_level,&
                  FP_exp_signal_level,&
                  trans_aer_thisbin,scatratio
          !endif
          !print *,"ENDREALACC"
          !print*,'Press,Temp ',Press,Temp

          !IF (.not. missing_real(tau_this_rangegate)) THEN
             ! warning: tau_this_rangegate can become 0 if 
             ! the FP_obs_signal_level=0 !!!
             ! warning: tau_this_rangegate can become larger than 1 if 
             ! the FP_obs_signal_level>FP_exp_signal_level
             ! (which can happen due to noise, inaccurate p,T, etc.)

             ! safety catches - GJM; > 1 also possible, include explanation
             ! GJM   IF (tau_this_rangegate .ge. 1.) tau_this_rangegate = 1.
             !IF (tau_this_rangegate .le. 0.) tau_this_rangegate = 0.

             ! store the result for later use
             !WorkingData%RayleighMeas(m)%RangeGate(r)%attenuation_above = &
             !                     attenuation_above
             !WorkingData%RayleighMeas(m)%RangeGate(r)%attenuation_this_bin = &
             !                     tau_this_rangegate

             !! to be replaced by a threshold on tau_this_rangegate ?
             !IF (FP_obs_signal_level .lt. &
             !     Rayleigh_signal_level_too_low) THEN
             !   WorkingData%RayleighMeas(m)%RangeGate(r)%extinction = &
             !        missing_indicator_real
             !ELSE
             !   IF (.not.(do_calibr_on_upper_bin .and. (r .eq. 1))) THEN 
             !      WorkingData%RayleighMeas(m)%RangeGate(r)%extinction = sigma
             !   ELSE
             !      ! the upper rangebin was used for calibration, 
             !      ! so no usefull extinction can be derived from it
             !      WorkingData%RayleighMeas(m)%RangeGate(r)%extinction = &
             !           missing_indicator_real
             !   END IF
             ! 
          !END IF

          ! warning: rangegate_size is undefined at this point !
          !print *,"rangegate_size=",rangegate_size 
          !
          !print "(i2,1x,i5,1x,5(es10.3e2,1x))",r,nint(rangegate_size),&
          !     WorkingData%RayleighMeas(m)%RangeGate(r)%calc_mol_backscatter,&
          !     FP_exp_signal_level,FP_obs_signal_level, &
          !     WorkingData%RayleighMeas(m)%RangeGate(r)%extinction,&
          !     (8.*pi/3)*WorkingData%RayleighMeas(m)%RangeGate(r)%calc_mol_backscatter

          !END IF ! missing_real(tau_this_rangegate)

          !  #]
       END DO Rangegateloop1

       !
       !  Computation of particle extinction from particle transmission 
       !
       fileunit2 = get_lun()
       open(unit=fileunit2,file='extinction.dat')
       trans_part_above = 1.
       WorkingData%RayleighMeas(m)%RangeGate(:)%extinction = missing_indicator_real

       Rangegateloop2: DO r=1, WorkingData%RayleighMeas(m)%nr_of_range_gates
          !  #[
          ! why check this element? it is not used here ...
          !IF (missing_real(WorkingData%RayleighMeas(m)%RangeGate(r)%P))goto 222
       
          incAngle = WorkingData%RayleighMeas(m)%RangeGate(r)%IncAngle
          !  Check the value
          if (incAngle .eq. 0.) then
             !print*,'Assuming incidence angle of 37.56 degrees'
             incAngle = 37.56*pi/180.
          endif

! todo: move this cos_incidence_angle into the working data structure
!       and fill it only once (in Copy_L1B_fields2WorkingArrays)
          cos_incidence_angle = cos(incAngle)
   
          rangeGateTop    = WorkingData%RayleighMeas(m)%RangeGate(r)%alt_end
          rangeGateBot    = WorkingData%RayleighMeas(m)%RangeGate(r)%alt_start
          rangeGateCenter = (rangeGateTop+rangeGateBot)/2.
       
          rangegate_size = (WorkingData%RayleighMeas(m)%RangeGate(r)%alt_end - &
                            WorkingData%RayleighMeas(m)%RangeGate(r)%alt_start)/&
                           cos_incidence_angle
          
          !tau_this_rangegate=&
          !      WorkingData%RayleighMeas(m)%RangeGate(r)%attenuation_this_bin
   
          ! Fabry-Perot observed signal
          FP_obs_signal_level = &
               WorkingData%RayleighMeas(m)%RangeGate(r)%obs_signal_level
   
          IF (missing_real(trans_part_above)          .or. &
              missing_real(tau_particle_rangegate(r)) .or. &
              missing_real(FP_Molecules_Only(r))            ) goto 222

          ! Fabry-Perot estimated signal
          trans_part_above=trans_part_above*tau_particle_rangegate(r)

          FP_exp_signal_level = FP_Molecules_Only(r)*trans_part_above**2

          ! Obtain particle extinction
          IF (tau_particle_rangegate(r) .gt. 0._r8_) THEN
             sigma_part = -log(tau_particle_rangegate(r))/rangegate_size
          ELSE
             ! TODO: actually for the first bin without signal this should be 
             ! infinite and for all others it should be missing.
             ! Should we define allow an infinite value here
             ! (and define a way to report it in the outputfile?)
             ! or should we just keep it missing?
             sigma_part = missing_indicator_real
          END IF

          ! Save particle extinction
          IF (.not.(do_calibr_on_upper_bin .and. (r .eq. 1))) THEN 
             WorkingData%RayleighMeas(m)%RangeGate(r)%extinction = sigma_part
          ELSE
             ! the upper rangebin was used for calibration, 
             ! so no usefull extinction can be derived from it
             WorkingData%RayleighMeas(m)%RangeGate(r)%extinction = &
                  missing_indicator_real
          END IF

          ! jump point for end of rangegate loop 2
          ! needed to make sure the elements of the test output file are
          ! written, even in case of missing data
222       continue

!          print*,'sigma ',r,tau_particle_rangegate(r),rangegate_size,&
!                 sigma_part 
   
          write(fileunit2,'(i2,1x,i5,2(F10.4,1x),2(E10.4,1x))') &
               r,NINT(rangeGateCenter),&
               FP_obs_signal_level,FP_exp_signal_level,&
               tau_particle_rangegate(r),&
               WorkingData%RayleighMeas(m)%RangeGate(r)%extinction

          !  #]
       END DO Rangegateloop2

stop 1

       ! activated for debugging only:
       Rangegateloop3: DO r=1, WorkingData%RayleighMeas(m)%nr_of_range_gates
          print "(a,i2,2(a,es10.3e2))","r=",r,&
               " beta_mol=",&
               WorkingData%RayleighMeas(m)%RangeGate(r)%calc_mol_backscatter,&
               " extinction= ",&
               WorkingData%RayleighMeas(m)%RangeGate(r)%extinction
       END DO Rangegateloop3
       !stop 1

       !  #]
    END DO measloop

    close(fileunit1)
    call free_lun(fileunit1)
    close(fileunit2)
    call free_lun(fileunit2)
    
  end subroutine Calc_Extinction_Iterative
    !  #]
  subroutine calc_FP_Molecules_Only(L1B, WorkingData, error_flag, m,&
                                    calibration_and_solid_angle,&
                                    FP_Molecules_Only,&
                                    MolExt_Method, &
                                    calibr_on_upper_bin)
    !  #[ 

    ! Description:
    !     Calculate the Fabry-Perot (FP) signal for an atmosphere that 
    !     contains molecules only (no particles assumed)
    !     
    !   * Auxiliary Meteorological Data input is used to determine molecular
    !     backscatter and extinction
    !
    !   * Temperature, pressure and frequency dependence of the fraction
    !     of the atmospheric signal detected by FP is implemented through
    !     parameter RayCal.
    !
    !   * Calibration on the first bin is applied.
    !
    !   * side-effects: besides FP_Molecules_Only,
    !     calibration_and_solid_angle and the error_flag also the
    !     following L2B/workingdata fields are set:
    !          WorkingData%RayleighMeas(m)%RangeGate(r)%attenuation_above
    !          WorkingData%RayleighMeas(m)%RangeGate(r)%attenuation_this_bin
    !
    ! created: 23-06-2006
    !
    ! Modifications: 
    !   23-06-2006 G.J. Marseille, initial version
    !

    USE AeolusConstants, only: lambda,pi
!    USE MolScat, only: get_ref_mol_scat
    
    TYPE(L1B_BRC_DataType), intent(in)    :: L1B ! input
    TYPE(WorkingDataType), intent(inout)  :: WorkingData ! input/output
    integer,            intent(in)        :: m   ! measurement number
    integer,            intent(out)       :: error_flag ! output
    real(r8_),dimension(24),intent(out)   :: FP_Molecules_Only
    real(r8_),          intent(out)       :: &
         calibration_and_solid_angle ! FP Calibration constant
    integer,            intent(in)        :: MolExt_Method       ! input
    logical,optional,   intent(in)        :: calibr_on_upper_bin ! input

    ! local variables
    logical :: do_calibr_on_upper_bin
    integer     :: r
    real(r8_)   :: FP_obs_signal_level ! FP observed value
    real(r8_)   :: FP_exp_signal_level ! FP expected value
    real(r8_)   :: cos_incidence_angle,sin_incidence_angle
    real(r8_)   :: rangegate_size
    real(r8_)   :: attenuation_above
    real(r8_)   :: tau_this_rangegate
    real(r8_)   :: Temp,Press,mol_sigma,mol_beta
    real(r8_)   :: RayCal,PresshPa
    real(r8_)   :: rangeGateTop,rangeGateBot,rangeGateCenter
    real(r8_)   :: uwind,vwind,wwind,hlos_wind,los_wind
    real(r8_)   :: freq_wind
    real(r8_)   :: satRange
    real(r8_)   :: tau_this_rangegate_mol
    real(r8_)   :: incAngle,aziAngle
    
    ! ========================================================

    ! init
    error_flag = no_error
    calibration_and_solid_angle = missing_indicator_real
    attenuation_above           = missing_indicator_real
    
    Rangegateloop1: DO r=1, WorkingData%RayleighMeas(m)%nr_of_range_gates
       ! init on the rangegate level
       FP_Molecules_Only(r) = missing_indicator_real
       WorkingData%RayleighMeas(m)%RangeGate(r)%attenuation_above    = &
            missing_indicator_real
       WorkingData%RayleighMeas(m)%RangeGate(r)%attenuation_this_bin = &
            missing_indicator_real
    END DO Rangegateloop1

    do_calibr_on_upper_bin = .false.
    IF (present(calibr_on_upper_bin)) &
         do_calibr_on_upper_bin = calibr_on_upper_bin

    Rangegateloop2: DO r=1, WorkingData%RayleighMeas(m)%nr_of_range_gates
       !  #[ calculate the extinction for the Rayleigh rangebins

       ! Estimate p and T from simple model given range gate center altitude
       ! Next, determine molecular backscatter and extinction
       !          WorkingData%RayleighMeas(m)%RangeGate(r)%calc_mol_backscatter = &
       !               L1B%RayleighMeas(m)%RangeGate(r)%temp_mol_backscatter
        
       rangeGateTop    = WorkingData%RayleighMeas(m)%RangeGate(r)%alt_end
       rangeGateBot    = WorkingData%RayleighMeas(m)%RangeGate(r)%alt_start
       rangeGateCenter = (rangeGateTop+rangeGateBot)/2._r8_
   
       satRange = WorkingData%RayleighMeas(m)%RangeGate(r)%satRange
       incAngle = WorkingData%RayleighMeas(m)%RangeGate(r)%IncAngle
       aziAngle = WorkingData%RayleighMeas(m)%RangeGate(r)%AziAngle
       !
       ! GJM 26/4/2006
       ! Current L1B contains zeros for incidence and azimuth angles
       ! so assume a value to be able to continue

       ! Check the input value
       if (incAngle .eq. 0._r8_) then
          !print*,'Assuming incidence angle of 37.56 degrees'
          incAngle = 37.56_r8_*pi/180._r8_
       endif
       cos_incidence_angle = cos(incAngle)
       sin_incidence_angle = sin(incAngle) 
   
       ! Check the input value
       if (aziAngle .eq. 0._r8_) then
          !print*,'Assuming azimuth angle of 90 degrees'
          aziAngle = 90._r8_*pi/180._r8_
       endif
   
       Press = WorkingData%RayleighMeas(m)%RangeGate(r)%P
       Temp  = WorkingData%RayleighMeas(m)%RangeGate(r)%T
       ! to be renamed to:
       !     WorkingData%RayleighMeas(m)%RangeGate(r)%T
       !     WorkingData%RayleighMeas(m)%RangeGate(r)%P   

       ! remind that not all methods below depend on pressure and temperature,
       ! so one of them being  missing is not a sufficient condition
       ! to abort processing the rest of the rangebins. JK.
       ! Maybe we could consider automatically switching to 
       ! the MolExt_Method_Parametrised method in case the AuxMetData 
       ! is missing ? JK
       IF (.not. missing_real(Press)) THEN
          ! Convert pressure from Pa -> hPa
          PresshPa=Press/100._r8_
       ELSE
          PresshPa = missing_indicator_real
       END IF

       select case (MolExt_Method)
       case(MolExt_Method_Undefined)
          print *,"ERROR in calc_FP_Molecules_Only:"
          print *,"MolExt_Method is still set to MolExt_Method_Undefined!"
          print *,"don't know what to do ....."
          error_flag = error_programming
          return
       case(MolExt_Method_Parametrised)
          ! If no AUX. MET. Data available, or if you don't wish to use it,
          ! then you may get the molecular 
          ! scattering and extinction from (climatological) parameterization
          mol_beta  = get_ref_mol_scat(rangeGateCenter)
          mol_sigma = mol_beta*(8._r8_*pi)/3._r8_

          ! should we in this case also fill PresshPa and Temp
          ! with a climatological profile to be able to run
          ! LUT_RayCalib_Intpol() anyway ?
       case(MolExt_Method_Proper)
          IF (       .not. missing_real(PresshPa) &
               .and. .not. missing_real(Temp)   ) THEN
             ! Compute molecular backscatter (1/m.sr) and extinction (1/m).
             mol_beta  = calc_molecular_backscatter(PresshPa,Temp)
             mol_sigma = calc_molecular_extinction(PresshPa,Temp)
          ELSE
             mol_beta  = missing_indicator_real
             mol_sigma = missing_indicator_real
          END IF
       case default
          print *,"ERROR in calc_FP_Molecules_Only:"
          print *,"MolExt_Method is set to an unknown type!!!"
          print *,"please update the calc_FP_Molecules_Only if you just"
          print *,"added a new way of calculation molecular extinction."
          print *,"If you didn't add a new algorithm this points to"
          print *,"a programming error ....."
          error_flag = error_programming
          return
       end select
       
       ! HLOS wind and Doppler shifted frequency
       ! to be calculated from NWP u,v,w windcomponents
       ! and the known measurement geometry
       uwind=0._r8_ !WorkingData%RayleighMeas(m)%RangeGate(r)%u  ! (m/s) zonal wind
       vwind=0._r8_ !WorkingData%RayleighMeas(m)%RangeGate(r)%v  ! (m/s) meridional wind
       ! vertical wind needs to be done
       wwind=0._r8_ !0.1         ! (m/s) vertical wind, not available in AMD
       hlos_wind=-uwind*sin(aziAngle)-vwind*cos(aziAngle)
       los_wind =hlos_wind*sin_incidence_angle+wwind*cos_incidence_angle
       freq_wind=((2._r8_*los_wind)/lambda)*1.e-6_r8_ ! (MHz)
   
       ! Interpolate lookup table to actual T,P,Doppler shift to get
       ! Rayleigh channel calibration value.
       IF (       .not. missing_real(PresshPa)  &
            .and. .not. missing_real(freq_wind) &
            .and. .not. missing_real(Temp)       ) THEN
!          call LUT_RayCalib_Intpol(Temp,PresshPa,freq_wind,RayCal,error_flag)
          call LUT_RBC_Intpol(Temp,PresshPa,freq_wind,RayCal,error_flag)
          if (error_flag .ne. no_error) return
       ELSE
          RayCal = missing_indicator_real
       END IF

       !print *,"PresshPa,Temp,freq_wind,lambda,mol_sigma,mol_beta,RayCal = "
       !print *,PresshPa,Temp,freq_wind,lambda,mol_sigma,mol_beta,RayCal

       ! To get the expected Rayleigh signal from the local mol_backscatter:
       ! =>multiply with rangebinsize
       ! =>calibrate the instrument by assuming the
       !   observed and expected signals in the upper rangebin
       !   should be identical
       
       ! this happens if a rangebin contains no data
       ! (only for the 23 layer testfile for now)
       IF (missing_real(WorkingData%RayleighMeas(m)%RangeGate(r)%alt_end)) &
            cycle Rangegateloop2

       ! we don't check the other geolocation data yet, and assume they will
       ! always be valid when using a proper L1B input file.

       rangegate_size = (WorkingData%RayleighMeas(m)%RangeGate(r)%alt_end - &
                         WorkingData%RayleighMeas(m)%RangeGate(r)%alt_start)/&
                        cos_incidence_angle

       IF (missing_real(mol_sigma)) cycle Rangegateloop2
       IF (missing_real(mol_beta))  cycle Rangegateloop2
       IF (missing_real(RayCal))    cycle Rangegateloop2       
       
       ! one-way molecular transmission
       tau_this_rangegate_mol = exp(-mol_sigma*rangegate_size)

       ! Expected Rayleigh channel signal, excluding transmission
       FP_exp_signal_level  = &
            (RayCal*mol_beta*rangegate_size)/satRange**2

       ! Measured FP signal from L1B
       FP_obs_signal_level = &
            WorkingData%RayleighMeas(m)%RangeGate(r)%obs_signal_level

       IF (missing_real(FP_obs_signal_level)) cycle Rangegateloop2

       ! Some extra sanity checks.
       IF (FP_exp_signal_level .le. 0.) THEN
          print *,"error in calc_FP_Molecules_Only():"
          print *,"invalid value for FP_exp_signal_level: ",&
               FP_exp_signal_level
          error_flag = error_programming
          return
       END IF

       IF (rangegate_size .le. 0._r8_) THEN
          print *,"error in calc_FP_Molecules_Only():"
          print *,"invalid value for rangegate_size: ",rangegate_size
          error_flag = error_programming
          return
       END IF

       ! set the attenuation above this layer
       IF (r .eq. 1) THEN
          attenuation_above = 1._r8_
          IF (do_calibr_on_upper_bin) THEN
             ! K_ray
             ! GJM 9/5/2006; expected Rayleigh signal must include
             !               molecular transmission loss in first bin
             calibration_and_solid_angle = &
                  FP_obs_signal_level/ &
                  (FP_exp_signal_level*attenuation_above**2*&
                  tau_this_rangegate_mol)
             !print *,"Rayleigh calibration = ",&
             !     calibration_and_solid_angle
          ELSE
             ! this should be copied from the instrument calibration
             ! but set it to 1 for now
             calibration_and_solid_angle = 1._r8_
             !print *,"JK: calibration placeholder for: ",&
             !     "calibration_and_solid_angle"
          END IF
       ELSE
          IF (missing_real(WorkingData%RayleighMeas(m)%&
               RangeGate(r-1)%attenuation_above)) THEN
             cycle rangegateloop2
          ELSE
             attenuation_above = &
                  WorkingData%RayleighMeas(m)%RangeGate(r-1)%attenuation_above*&
                  WorkingData%RayleighMeas(m)%RangeGate(r-1)%attenuation_this_bin
          END IF
       END IF
       
       ! catch div by zero errors
       ! probably not needed, but you never know ...
       IF (attenuation_above .lt. 0._r8_) THEN
          print *,"error in Calc_Extinction():"
          print *,"invalid value for attenuation_above: ",&
               attenuation_above
          error_flag = error_programming
          return
       END IF
       IF (calibration_and_solid_angle .le. 0._r8_) THEN
          print *,"error in Calc_Extinction():"
          print *,"invalid value for calibration_and_solid_angle: ",&
               calibration_and_solid_angle
          error_flag = error_programming
          return
       END IF

       ! apply the calibration and molecular transmission in layers 
       ! above (two-way) and molecular transmission in current 
       ! layer (one-way)
       FP_exp_signal_level = &
            FP_exp_signal_level * calibration_and_solid_angle *&
            attenuation_above**2 * tau_this_rangegate_mol
       
       ! Save result
       FP_Molecules_Only(r) = FP_exp_signal_level
       
       ! Save one-way molecular transmission
       IF (attenuation_above .gt. 0._r8_) THEN
          tau_this_rangegate = tau_this_rangegate_mol  
       ELSE
          print *,"warning in calc_FP_Molecules_Only()"
          print *,"attenuation_above seems 0, i.e. no signal left here ...."
          print *,"so no information available on this rangegate"
          tau_this_rangegate = missing_indicator_real
       END IF
       
       IF (.not. missing_real(tau_this_rangegate)) THEN
          ! warning: tau_this_rangegate can become 0 if 
          ! the FP_obs_signal_level=0 !!!
          ! warning: tau_this_rangegate can become larger than 1 if 
          ! the FP_obs_signal_level>FP_exp_signal_level
          ! (which can happen due to noise, inaccurate p,T, etc.)
          
          IF (tau_this_rangegate .le. 0._r8_) tau_this_rangegate = 0._r8_
          
          ! store the result for later use
          WorkingData%RayleighMeas(m)%RangeGate(r)%attenuation_above = &
               attenuation_above
          WorkingData%RayleighMeas(m)%RangeGate(r)%attenuation_this_bin = &
               tau_this_rangegate
          
       END IF ! missing_real(tau_this_rangegate)

          !  #]
    END DO Rangegateloop2

  end subroutine calc_FP_Molecules_Only

    !  #]
  subroutine Copy_Extinction_Rayl_to_Mie(WorkingData)
    !  #[

    TYPE(WorkingDataType), intent(inout) :: WorkingData ! input/output

    ! local variables
    integer     :: nr_of_measurements, m,r
    integer     :: rrb1, rrb2
!    real(r8_)   :: eps
    real(r8_)   :: rayleigh_part_above_mie
    real(r8_)   :: rayleigh_part_below_mie
    real(r8_)   :: mie_part_above_rayleigh
    real(r8_)   :: mie_part_below_rayleigh
    real(r8_)   :: cos_incidence_angle
    real(r8_)   :: tau_above_rayleigh, tau_above_mie
    real(r8_)   :: sigma_rayl, sigma_mie
    real(r8_)   :: tau_in_and_above_rayleigh
    real(r8_)   :: tau_in_and_above_mie
    real(r8_)   :: tau_this_rayl_bin
    real(r8_)   :: tau_this_mie_bin
    real(r8_)   :: mie_alt_bottom,mie_alt_top
    real(r8_)   :: rayleigh_alt_bottom,rayleigh_alt_top
    real(r8_)   :: size_mie_bin ! overlap, size_rayleigh_bins, 
    TYPE(WD_MieRangeGateType),  pointer :: WD_MieRangeGate 

    ! workaround to let the old testcases pass
    logical, parameter :: approximated_log = .true.

    ! Temporary placeholder
    ! (to be copied from the L1B Geoloc_ADS)
    cos_incidence_angle = 0.819152_r8_ ! cos(35 degrees)

    ! allow an inaccuracy in comparing the rangegate sizes of 1 m/s
!    eps = 1._r8_

    nr_of_measurements = WorkingData%nr_of_measurements
    measloop: DO m=1, nr_of_measurements
       Rangegateloop: DO r=1, WorkingData%MieMeas(m)%nr_of_range_gates
          !  #[ copy the extinction from Rayleigh to Mie rangebins

          ! point to this RangeGate for shorter notation
          WD_MieRangeGate => WorkingData%MieMeas(m)%RangeGate(r)
 
          rrb1 = WD_MieRangeGate%first_rayleigh_rangebin
          rrb2 = WD_MieRangeGate%last_rayleigh_rangebin

          !print *,"testjos, r_mie, rrb1, rrb2 = ",r,rrb1,rrb2

          OnlyMie: IF ( (rrb1 .eq. -1) .or. (rrb2 .eq. -1) ) THEN
             ! we don't have a Rayleigh rangebin corresponding to this 
             ! Mie rangebin. Nothing to do, just keep the value missing
             cycle Rangegateloop
          ELSE

             mie_alt_top      = WorkingData%MieMeas(m)%RangeGate(r)%alt_end
             mie_alt_bottom   = WorkingData%MieMeas(m)%RangeGate(r)%alt_start

             !print *,"testjos: mie bottom, top = ",mie_alt_bottom,mie_alt_top

             rayleigh_alt_top    = WorkingData%RayleighMeas(m)%RangeGate(rrb1)%alt_end
             rayleigh_alt_bottom = WorkingData%RayleighMeas(m)%RangeGate(rrb2)%alt_start

             !print *,"testjos: rayleigh bottom, top = ",&
             !     rayleigh_alt_bottom,rayleigh_alt_top

             ! determine rayleigh part above this mie bin:
             if (rayleigh_alt_top .gt. mie_alt_top) then
                rayleigh_part_above_mie = rayleigh_alt_top - mie_alt_top
             else
                rayleigh_part_above_mie = 0._r8_
             end if

             ! determine rayleigh part below this mie bin
             if (rayleigh_alt_bottom .lt. mie_alt_bottom) then
                rayleigh_part_below_mie = mie_alt_bottom - rayleigh_alt_bottom
             else
                rayleigh_part_below_mie = 0._r8_
             end if

             ! determine mie part above this rayleigh bin:
             if (mie_alt_top .gt. rayleigh_alt_top) then
                mie_part_above_rayleigh = mie_alt_top - rayleigh_alt_top
             else
                mie_part_above_rayleigh = 0._r8_
             end if

             ! determine mie part below this rayleigh bin
             if (mie_alt_bottom .lt. rayleigh_alt_bottom) then
                mie_part_below_rayleigh = rayleigh_alt_bottom - mie_alt_bottom
             else
                mie_part_below_rayleigh = 0._r8_
             end if

             ! determine overlapping part
             !size_rayleigh_bins = rayleigh_alt_top-rayleigh_alt_bottom
             size_mie_bin       = mie_alt_top     -mie_alt_bottom
             !overlap = size_rayleigh_bins - rayleigh_part_above_mie - &
             !                               rayleigh_part_below_mie

             !print *,"testjos: rayleigh_part_above_mie  = ",&
             !                  rayleigh_part_above_mie
             !print *,"testjos: rayleigh_part_below_mie  = ",&
             !                  rayleigh_part_below_mie
             !print *,"testjos: mie_part_above_rayleigh  = ",&
             !                  mie_part_above_rayleigh
             !print *,"testjos: mie_part_below_rayleigh   = ",&
             !                  mie_part_below_rayleigh

             !print *,"testjos: overlap            = ",overlap
             !print *,"testjos: size_rayleigh_bins = ",size_rayleigh_bins

             ! determine transmission to top of mie bin
             sigma_rayl = WorkingData%RayleighMeas(m)%RangeGate(rrb1)%extinction
             tau_above_rayleigh = &
                  WorkingData%RayleighMeas(m)%RangeGate(rrb1)%attenuation_above

             ! remember that the upper rangebin has missing sigma
             ! if you use this bin for calibration! so checking for
             ! missing is always a good idea
             IF (       .not. missing_real(sigma_rayl)        &
                  .and. .not. missing_real(tau_above_rayleigh) ) THEN

                IF (rayleigh_part_above_mie .gt. 0._r8_) THEN
                   IF (approximated_log) THEN
                      tau_above_mie = tau_above_rayleigh * &
                           (1._r8_ - sigma_rayl * rayleigh_part_above_mie/&
                               cos_incidence_angle)
                   ELSE
                      tau_above_mie = tau_above_rayleigh * &
                           exp(- sigma_rayl * rayleigh_part_above_mie/&
                               cos_incidence_angle)
                   END IF
                ELSE
                   ! the mie bin extends above the highest rayleigh bin.
                   ! In this case, just assume the extinction remains constant
                   ! and equal to the highest value determined in the rayleigh
                   ! channel (what else can you do?)
                   IF (approximated_log) THEN
                      tau_above_mie = tau_above_rayleigh / &
                           (1._r8_ -  sigma_rayl * mie_part_above_rayleigh/&
                               cos_incidence_angle)
                   ELSE
                      tau_above_mie = tau_above_rayleigh / &
                           exp(- sigma_rayl * mie_part_above_rayleigh/&
                               cos_incidence_angle)
                   END IF
                END IF
             ELSE
                tau_above_mie = missing_indicator_real
             END IF
 
             !print *,"testjos: sigma_rayl(rrb1)         = ",sigma_rayl
             !print *,"testjos: tau_above_rayleigh(rrb1) = ",tau_above_rayleigh
             !print *,"testjos: tau_above_mie            = ",tau_above_mie

             ! take transmission in this mie bin into account
             sigma_rayl = WorkingData%RayleighMeas(m)%RangeGate(rrb2)%extinction

             ! note that we are looking relative to the lower rayleigh
             ! bin now (only important if the mie bin overlaps with more
             ! than 1 rayleigh bin, which we would not expect in a
             ! normal situation)
             tau_above_rayleigh = &
                  WorkingData%RayleighMeas(m)%RangeGate(rrb2)%attenuation_above
             tau_this_rayl_bin       = &
                  WorkingData%RayleighMeas(m)%RangeGate(rrb2)%attenuation_this_bin

             !print *,"testjos: sigma_rayl(rrb2)         = ",sigma_rayl
             !print *,"testjos: tau_above_rayleigh(rrb2) = ",tau_above_rayleigh
             !print *,"testjos: tau_this_rayl_bin(rrb2)  = ",tau_this_rayl_bin

             IF (       .not. missing_real(tau_above_rayleigh) &
                  .and. .not. missing_real(tau_this_rayl_bin)  ) THEN
                tau_in_and_above_rayleigh = &
                     tau_above_rayleigh*tau_this_rayl_bin
                ! should be identical to
                ! WorkingData%RayleighMeas(m)%RangeGate(rrb2-1)%attenuation_above
                ! provided that it exists, so test for this.
             ELSE
                tau_in_and_above_rayleigh = missing_indicator_real
             END IF

             ! correct for transmission between bottom of mie 
             ! and bottom of rayleigh range bin
             IF (       .not. missing_real(tau_in_and_above_rayleigh) &
                  .and. .not. missing_real(sigma_rayl)                 ) THEN

                IF (rayleigh_part_below_mie .ge. 0._r8_) THEN
                   ! ok, rayleigh reached below the lower part of this Mie bin,
                   ! so just remove the extinction in the part_below,
                   ! according to the rayleigh extinction we found,
                   ! and we are done
                   IF (approximated_log) THEN
                      tau_in_and_above_mie = tau_in_and_above_rayleigh/&
                           (1._r8_ - sigma_rayl * &
                               rayleigh_part_below_mie/cos_incidence_angle)
                   ELSE
                      tau_in_and_above_mie = tau_in_and_above_rayleigh/&
                           exp(- sigma_rayl * &
                               rayleigh_part_below_mie/cos_incidence_angle)
                   END IF
                ELSE
                   ! the mie bin extends below the lowest rayleigh bin.
                   ! In this case, just assume the extinction remains constant
                   ! and equal to the lowest value determined in the rayleigh
                   ! channel (what else can you do?)
                   IF (approximated_log) THEN
                      tau_in_and_above_mie = tau_in_and_above_rayleigh*&
                           (1._r8_ - sigma_rayl * &
                            mie_part_below_rayleigh/cos_incidence_angle)
                   ELSE
                      tau_in_and_above_mie = tau_in_and_above_rayleigh*&
                           exp(- sigma_rayl * &
                               mie_part_below_rayleigh/cos_incidence_angle)
                   END IF
                END IF
             ELSE
                tau_in_and_above_mie = missing_indicator_real
             END IF

             !print *,"testjos: sigma_rayl                = ",sigma_rayl
             !print *,"testjos: tau_in_and_above_rayleigh = ",&
             !                  tau_in_and_above_rayleigh
             !print *,"testjos: tau_in_and_above_mie      = ",&
             !                  tau_in_and_above_mie

             
             ! store the results
             IF (       .not. missing_real(tau_in_and_above_mie) &
                  .and. .not. missing_real(tau_above_mie)        &
                  .and. .not. (tau_in_and_above_mie .eq. 0._r8_) &
                  .and. .not. (tau_above_mie        .eq. 0._r8_)  ) THEN
                
                IF (approximated_log) THEN
 
! this would be the correct approximation but gives different
! results due to rounding differences, than just copying
! sigma from rayleigh to mie, therefore it is disables for now.
!                   tau_this_mie_bin = tau_this_rayl_bin
!                   if (rayleigh_part_above_mie .gt. 0._r8_) then
!                      tau_this_mie_bin = tau_this_mie_bin/&
!                           (1._r8_ - sigma_rayl * &
!                           rayleigh_part_above_mie/cos_incidence_angle)
!                   end if
!                   if (rayleigh_part_below_mie .gt. 0._r8_) then
!                      tau_this_mie_bin = tau_this_mie_bin/&
!                           (1._r8_ - sigma_rayl * &
!                           rayleigh_part_below_mie/cos_incidence_angle)
!                   end if

!                   sigma_mie = (1._r8_ - tau_this_mie_bin)/&
!                                 (size_mie_bin/cos_incidence_angle)

! this gives the old results (and is wrong if the rangebin boundaries
! do not nicely match):
                   sigma_mie = sigma_rayl
                ELSE
                   tau_this_mie_bin = tau_in_and_above_mie/tau_above_mie
                   sigma_mie = - log(tau_this_mie_bin)/&
                                 (size_mie_bin/cos_incidence_angle)
                END IF
             ELSE
                tau_this_mie_bin = missing_indicator_real
                sigma_mie        = missing_indicator_real
             END IF

             WD_MieRangeGate%attenuation_above    = tau_above_mie
             WD_MieRangeGate%attenuation_this_bin = tau_this_mie_bin
             WD_MieRangeGate%extinction           = sigma_mie
                  
             !print *,"testjos: tau_above_mie    = ",tau_above_mie
             !print *,"testjos: tau_this_mie_bin = ",tau_this_mie_bin
             !print *,"testjos: sigma_mie        = ",sigma_mie

          END IF OnlyMie
       !  #]
       END DO Rangegateloop

    END DO measloop

  end subroutine Copy_Extinction_Rayl_to_Mie
    !  #]
  !------------------------------------------
  ! this is a copy of the old extinction routine, 
  ! and is only kept here to prevent breaking tests.
  ! Once we all are happy with the above new implementation it can be deleted.
  ! (this copy was taken from version 1.8 from the KNMI cvs system)
  subroutine Calc_Extinction_Old(WorkingData, error_flag, &
                                 use_lite_test_data, calibr_on_upper_bin)
    !  #[

    TYPE(WorkingDataType), intent(inout) :: WorkingData ! input/output
    integer,               intent(out)   :: error_flag ! output
    logical,optional,      intent(in)    :: use_lite_test_data  ! input
    logical,optional,      intent(in)    :: calibr_on_upper_bin ! input

    ! local variables
    logical   :: do_use_lite_test_data
    logical   :: do_calibr_on_upper_bin
    integer   :: nr_of_measurements, m,r
    real(r8_) :: Rayleigh_exp_signal_level ! expected value
    real(r8_) :: Rayleigh_obs_signal_level ! observed value
    real(r8_) :: signal_fraction ! correction for spectrometer
    real(r8_) :: cos_incidence_angle
    real(r8_) :: rangegate_size
    real(r8_) :: attenuation_above
    real(r8_) :: tau_this_rangegate
    real(r8_) :: sigma, alt_mid

    ! this should be imported from the calibration datasets
    ! but for now calculate it in here
    real(r8_) :: calibration_and_solid_angle

    ! init
    error_flag = no_error
    calibration_and_solid_angle = missing_indicator_real
    attenuation_above           = missing_indicator_real

    !print *,"testjos: starting Calc_Extinction_Old():"

    do_use_lite_test_data = .false.
    IF (present(use_lite_test_data)) &
         do_use_lite_test_data = use_lite_test_data
    do_calibr_on_upper_bin = .false.
    IF (present(calibr_on_upper_bin)) &
         do_calibr_on_upper_bin = calibr_on_upper_bin

    IF (.not. do_calibr_on_upper_bin) THEN
       ! extra check, added 20060623 JK
       IF (.not. do_use_lite_test_data) THEN
          print *,"ERROR: the old extinction calculation contains no calibration"
          print *,"so should only be used on Lite test data. However, it seems you"
          print *,"are trying to use it on proper aeolus data!!!"
          print *,"this is probably due to a programming error in the test suite..."
          error_flag = error_programming
          return
       END IF
    END IF

    ! Temporary placeholder
    ! (to be copied from the L1B Geoloc_ADS)
    cos_incidence_angle = 0.819152 ! cos(35 degrees)

    ! for real ADM-Aeolus data only about 25% of all photons can be
    ! detected in the channels A and B (after proper calibration)
    ! but for Lite testdata this should be 100%.
    ! Therefore we have the following switch:
    IF (do_use_lite_test_data) THEN
       signal_fraction = 1.0
    ELSE
       signal_fraction = 0.26
    END IF
          
    nr_of_measurements = WorkingData%nr_of_measurements
    measloop: DO m=1, nr_of_measurements
       Rangegateloop1: DO r=1, WorkingData%RayleighMeas(m)%nr_of_range_gates
          ! use the parametrised mol_backscatter for now 
          ! this is to be replaced by a call to a function that calculates 
          ! it from the given p and T at this altitude

          !  #[ calculate the extinction for the Rayleigh rangebins
          alt_mid = &
               (WorkingData%RayleighMeas(m)%RangeGate(r)%alt_start + &
                WorkingData%RayleighMeas(m)%RangeGate(r)%alt_end   )*0.5_r8_

          ! the two_digits_only workaround trick is needed to remain 
          ! compatible with the previous code in which this number was 
          ! read from the lite ascii datafile with a precision of just 
          ! 2 digits after the period
          
          WorkingData%RayleighMeas(m)%RangeGate(r)%calc_mol_backscatter = &
               get_ref_mol_scat(alt_mid,two_digits_only=use_lite_test_data)

          ! To get the expected Rayleigh signal from the local mol_backscatter:
          ! =>multiply with rangebinsize
          ! correct for fraction of spectrum detected by the spectrometer:
          ! ==>signal_fraction
          ! =>calibrate the instrument by assuming the
          !   observed and expected signals in the upper rangebin
          !   shoud be identical
          ! =>multiply with solid angle of the instrument 
          !   (I assume this is part of the calibration constant)

          ! this happens if a rangebin contains no data
          ! (only for the 23 layer testfile for now)
          IF (missing_real(WorkingData%RayleighMeas(m)%RangeGate(r)%alt_end)) &
               cycle Rangegateloop1

          rangegate_size = &
               (WorkingData%RayleighMeas(m)%RangeGate(r)%alt_end - &
                WorkingData%RayleighMeas(m)%RangeGate(r)%alt_start)/&
                           cos_incidence_angle

          !  calculate the expected signal level
          Rayleigh_exp_signal_level  = &
               rangegate_size*&
               WorkingData%RayleighMeas(m)%RangeGate(r)%calc_mol_backscatter*&
               signal_fraction

          Rayleigh_obs_signal_level = &
               WorkingData%RayleighMeas(m)%RangeGate(r)%obs_signal_level

          ! for lite data apply the rangebin size (this was not yet done
          ! in the datafile)
          IF (do_use_lite_test_data) &
               Rayleigh_obs_signal_level = &
               rangegate_size*Rayleigh_obs_signal_level 

          IF (.not. missing_real(Rayleigh_obs_signal_level)) THEN

             ! now we have the exp_signal_level s_exp that should reach 
             ! the instrument if there were no extinction, and we
             ! have the measured signal level s_meas.
             ! so using the following:
             !   s_meas = tau_above^2 * tau_this_rangegate * s_exp
             ! we can calculate the attenuation in this rangegate: 
             ! tau_this_rangegate. For this we have to assume
             ! a value for the upper rangebin. (assume no attenuation
             ! for now above the upper rangebin. The error will be small
             ! and constant for all rangebins).
             ! tau_this_rangegate = (s_meas/s_exp) * (1/tau_above^2)
             ! Then devide 1 - the found attenuation of the rangegate with
             ! the rangebin size to get the value of the extinction in [1/m]
             ! sigma = (1-tau)/dx
             
             !  #[ catch div by zero errors
             ! probably not needed, but you never know ...
             IF (Rayleigh_exp_signal_level .le. 0.) THEN
                print *,"error in Calc_Extinction():"
                print *,"invalid value for Rayleigh_exp_signal_level: ",&
                     Rayleigh_exp_signal_level
                error_flag = error_programming
                return
             END IF
             IF (rangegate_size .le. 0.) THEN
                print *,"error in Calc_Extinction():"
                print *,"invalid value for rangegate_size: ",rangegate_size
                error_flag = error_programming
                return
             END IF
             !  #]

             ! set the attenuation above this layer
             IF (r .eq. 1) THEN
                attenuation_above = 1.0
                IF (do_calibr_on_upper_bin) THEN
                   calibration_and_solid_angle = &
                        Rayleigh_obs_signal_level/Rayleigh_exp_signal_level
                   !print *,"Rayleigh calibration = ",&
                   !     calibration_and_solid_angle
                ELSE
                   ! this should be copied from the instrument calibration
                   ! but set it to 1 for now
                   calibration_and_solid_angle = 1.0
                   !print *,"JK: calibration placeholder for: ",&
                   !     "calibration_and_solid_angle"
                END IF
             ELSE
                IF (missing_real(WorkingData%RayleighMeas(m)%&
                                 RangeGate(r-1)%attenuation_above)) THEN
                   cycle rangegateloop1
                ELSE
                   attenuation_above = &
                        WorkingData%RayleighMeas(m)%RangeGate(r-1)%attenuation_above*&
                        WorkingData%RayleighMeas(m)%RangeGate(r-1)%attenuation_this_bin
                END IF
             END IF
             
             !  #[ catch div by zero errors
             ! probably not needed, but you never know ...
             IF (attenuation_above .lt. 0.) THEN
                print *,"error in Calc_Extinction():"
                print *,"invalid value for attenuation_above: ",&
                     attenuation_above
                error_flag = error_programming
                return
             END IF
             IF (calibration_and_solid_angle .le. 0.) THEN
                print *,"error in Calc_Extinction():"
                print *,"invalid value for calibration_and_solid_angle: ",&
                     calibration_and_solid_angle
                error_flag = error_programming
                return
             END IF
             !  #]

             ! apply the calibration
             Rayleigh_obs_signal_level = &
                  Rayleigh_obs_signal_level/calibration_and_solid_angle

             IF (attenuation_above .gt. 0.) THEN
                ! see eq. 4 from TN2.3 section 7.2 
                ! tau_i=(S_obs/S_exp)*(1/tau_above**2)
                tau_this_rangegate = &
                     (Rayleigh_obs_signal_level/Rayleigh_exp_signal_level)/&
                     (attenuation_above**2)
             ELSE
                tau_this_rangegate = missing_indicator_real
             END IF

             IF (.not. missing_real(tau_this_rangegate)) THEN
                ! warning: tau_this_rangegate can become 0 if 
                ! the Rayleigh_obs_signal_level=0 !!!
                ! warning: tau_this_rangegate can become larger than 1 if 
                ! the Rayleigh_obs_signal_level>Rayleigh_exp_signal_level
                ! (which can happen due to noise, inaccurate p,T, etc.)

                ! safety catches
                IF (tau_this_rangegate .ge. 1.) tau_this_rangegate = 1.
                IF (tau_this_rangegate .le. 0.) tau_this_rangegate = 0.

                sigma = (1.-tau_this_rangegate)/rangegate_size

                ! store the result for later use
                WorkingData%RayleighMeas(m)%RangeGate(r)%attenuation_above = &
                     attenuation_above
                WorkingData%RayleighMeas(m)%RangeGate(r)%attenuation_this_bin = &
                     tau_this_rangegate

                ! to be replaced by a threshold on tau_this_rangegate ?
                IF (Rayleigh_obs_signal_level .lt. &
                     Rayleigh_signal_level_too_low) THEN
                   WorkingData%RayleighMeas(m)%RangeGate(r)%extinction = &
                        missing_indicator_real
                ELSE
                   IF (.not.(do_calibr_on_upper_bin .and. (r .eq. 1))) THEN 
                      WorkingData%RayleighMeas(m)%RangeGate(r)%extinction = sigma
                   ELSE
                      ! the upper rangebin was used for calibration, 
                      ! so no usefull extinction can be derived from it
                      WorkingData%RayleighMeas(m)%RangeGate(r)%extinction = &
                           missing_indicator_real
                   END IF
                END IF

             END IF ! missing_real(tau_this_rangegate)
          END IF
          !  #]
       END DO Rangegateloop1
    END DO measloop

  end subroutine Calc_Extinction_Old

    !  #]
  !------------------------------------------
! to add:
! a routine to copy backscatter ratio from the E2S input
! a routine to determine backscatter ratio from Rayleigh only
! a routine to determine backscatter ratio from Mie only
!
  subroutine Calc_BackscatterRatio(L1B,WorkingData,error_flag,&
                                   Method,FallbackMethod)
    !  #[ generic routine to get backscatter
    TYPE(L1B_BRC_DataType),    intent(in):: L1B            ! input
    TYPE(WorkingDataType), intent(inout) :: WorkingData    ! input/output
    integer,               intent(out)   :: error_flag     ! output
    integer(IntAuc),       intent(in)    :: Method         ! input
    integer(IntAuc),       intent(in)    :: FallbackMethod ! input

    select case(Method)
    case(Scat_Ratio_Undefined)
       print *,"WARNING: the Method for determining BackscatterRatio"
       print *,"WARNING: is still set to undefined !"
       print *,"WARNING: this probably points to a programming error !!!"
    case(Scat_Ratio_from_L1B)
       ! get the value from the L1B file
       call Copy_Mie_BackscRatio_from_L1B(L1B, WorkingData, error_flag)
    case(Scat_Ratio_from_MieRayl)
       ! calc. the value by comparing Mie and Rayleigh Usefull Signal
       call Calc_BackscRatio_Comp_US(WorkingData, error_flag)
    case(Scat_Ratio_from_RaylOnly)
       ! calc. the value by assuming a backscatter to extinction 
       ! ratio for aerosols
       print *,"Method: Scat_Ratio_from_RaylOnly is not yet implemented"
       stop 1
    case(Scat_Ratio_Dont_Use)
       ! do nothing
    case default
       print *,"WARNING: this Method for determining BackscatterRatio"
       print *,"WARNING: is not defined !"
       print *,"WARNING: this probably points to a programming error !!!"
    end select

    ! REMARK: The fallback method only fills BackscatterRatio fields
    !         that could not be filled by the method of first choice above
    !         (i.e. which are still set to missing at this point).
    !         This may occur for example for Rayleigh rangebins without
    !         corresponding Mie rangebin, if the above method is set
    !         to Scat_Ratio_from_L1B.

    ! sanity check
    IF (Method .eq. FallbackMethod) THEN
       print *,"WARNING: Method and FallbackMethod are both set to the"
       print *,"WARNING: same value. This is not usefull to fill missing"
       print *,"WARNING: values, so skipping the FallbackMethod to"
       print *,"WARNING: to prevent double calculations"
       return
    END IF

    select case(FallbackMethod)
    case(Scat_Ratio_Undefined)
       print *,"WARNING: the FallbackMethod for determining BackscatterRatio"
       print *,"WARNING: is still set to undefined !"
       print *,"WARNING: this probably points to a programming error !!!"
    case(Scat_Ratio_from_L1B)
       ! get the value from the L1B file
       call Copy_Mie_BackscRatio_from_L1B(L1B,WorkingData,&
                                          error_flag,fallback=.true.)
    case(Scat_Ratio_from_MieRayl)
       ! calc. the value by comparing Mie and Rayleigh Usefull Signal
       call Calc_BackscRatio_Comp_US(WorkingData, &
                                     error_flag, fallback=.true.)
    case(Scat_Ratio_from_RaylOnly)
       ! calc. the value by assuming a backscatter to extinction 
       ! ratio for aerosols
       print *,"FallbackMethod:Scat_Ratio_from_RaylOnly is not yet implemented"
       stop 1
    case(Scat_Ratio_Dont_Use)
       ! do nothing
    case default
       print *,"WARNING: this FallbackMethod for determining BackscatterRatio"
       print *,"WARNING: is not defined !"
       print *,"WARNING: this probably points to a programming error !!!"
    end select

  end subroutine Calc_BackscatterRatio
    !  #]
  subroutine Calc_BackscRatio_Comp_US(WorkingData, error_flag, fallback)
    !  #[ get backscatter by comparing Mie/Rayl.Usefull Signal
    TYPE(WorkingDataType), intent(inout) :: WorkingData ! input/output
    integer,               intent(out)   :: error_flag  ! output
    logical, optional,     intent(in)    :: fallback    ! optional input

    ! local variables
    integer   :: nr_of_measurements, m, r, r_m
    integer   :: mrb1, mrb2
    real(r8_) :: FP_obs_signal_level
    real(r8_) :: Mie_obs_signal_level
    real(r8_) :: Cal_Factor_Mie
    real(r8_) :: Cal_Factor_Rayleigh
    logical   :: fill_missing_only

    fill_missing_only = .false.
    IF (present(fallback)) THEN
       IF (fallback) fill_missing_only = .true.
    END IF

    error_flag = no_error
    nr_of_measurements = WorkingData%nr_of_measurements

    Cal_Factor_Mie      = 1.
    Cal_Factor_Rayleigh = 1.
    ! how to obtain these numbers ??
    !
    ! Possible inputs are:
    ! MRC => MRC-GADS => Measurement_Mean_Sensitivity 
    ! (table 7.42 of L1B-IODD)
    ! MRC => MRC-GADS => Min/Max_Mie_Measurement_Mean_Sensitivity 
    ! (table 7.45 of L1B-IODD)
    ! (see also table 8-52 what is the difference ?????)
    !
    ! these numbers have a unit [PixelIndex/MHz]
    ! but I don't yet know what this means/how to use this
    !
    ! RRC => RRC-GADS => Measurement_Response (unit is [AU] !!)
    ! (table 7.50 of L1B-IODD)
    ! RRC => RRC-GADS => Measurement_Mean_Sensitivity (unit is [AU/MHz] !!)
    ! (table 7.51 of L1B-IODD)
    ! RRC => RRC-GADS => Min/Max_Rayleigh_Measurement_Mean_Sensitivity 
    ! (unit is [AU/MHz] !!)
    ! (table 7.57 of L1B-IODD)
    ! (see also table 8-53 what is the difference ?????)
    !
    ! also this unit [AU/MHz] is not very clear to me.

    measloop: DO m=1, nr_of_measurements
       Rangegateloop: DO r=1, WorkingData%RayleighMeas(m)%nr_of_range_gates
          !  #[ calculate the backscatter ratio for the Rayleigh rangebins

          IF (fill_missing_only) THEN
             IF (missing_real(WorkingData%RayleighMeas(m)%RangeGate(r)%&
                                               backscatter_ratio)) THEN
                ! print *,"TESTJOS: fallback used for m=",m," r=",r
             ELSE
                cycle Rangegateloop
             END IF
          END IF

          ! we have 2 cases here:
          ! -we have both Mie and Rayleigh data
          ! -we have Rayleigh but no Mie data

          mrb1 = WorkingData%RayleighMeas(m)%RangeGate(r)%first_mie_rangebin
          mrb2 = WorkingData%RayleighMeas(m)%RangeGate(r)%last_mie_rangebin

          OnlyRayleigh: IF ( (mrb1 .eq. -1) .or. (mrb2 .eq. -1) ) THEN
             ! we don't have a Mie rangebin corresponding to this Rayleigh 
             ! rangebin. This case is not yet implemented,
             ! (we could maybe use an assumed backscatter-to-extinction ratio?)
          ELSE
             matching_rangebins2: IF (mrb1 .eq. mrb2) THEN 
                ! we have identical rangegates for Mie and Rayleigh
                Mie_obs_signal_level = &
                     WorkingData%MieMeas(m)%RangeGate(mrb1)%obs_signal_level
             ELSE
                ! we have several Mie rangegates for this Rayleigh Rangegate
                ! so just add the signals for the (usually 2) Mie rangebins
                !  #[
                Mie_obs_signal_level = 0.
                AddMieBins: DO r_m = mrb1,mrb2
                   IF (.not. missing_real(WorkingData%MieMeas(m)%&
                                       RangeGate(r_m)%obs_signal_level)) THEN
                      Mie_obs_signal_level = Mie_obs_signal_level + &
                           WorkingData%MieMeas(m)%RangeGate(r_m)%&
                           obs_signal_level
                   ELSE
                      IF (LocalVerbosity .ge. VerbosityReport) THEN
                         print *,"WARNING in Calc_BackscatterRatio():"
                         print *,"signal for Mie range bin nr. ",r_m
                         print *,"seems missing for measurement nr. ",m
                      END IF
                      Mie_obs_signal_level = missing_indicator_real
                      exit AddMieBins
                   END IF
                END DO AddMieBins
                !  #]
             END IF matching_rangebins2

             FP_obs_signal_level = &
                  WorkingData%RayleighMeas(m)%RangeGate(r)%obs_signal_level
             IF ( .not. missing_real(Mie_obs_signal_level) .and. &
                  .not. missing_real(FP_obs_signal_level)  ) THEN
                IF ( FP_obs_signal_level .gt. &
                     Rayleigh_signal_level_too_low) THEN
                   ! see eq.10 from TN2.3 section 7.2
                   ! rho=1+C_rayl*S_mie/(C_mie*S_rayl)
                   WorkingData%RayleighMeas(m)%RangeGate(r)%backscatter_ratio = &
                        1 + (Cal_Factor_Rayleigh*Mie_obs_signal_level     )/&
                            (Cal_Factor_Mie     *FP_obs_signal_level)
                   WorkingData%RayleighMeas(m)%RangeGate(r)%rho_ok = .TRUE.
                END IF
             END IF
          END IF OnlyRayleigh
          !  #]
       END DO Rangegateloop
    END DO measloop
    
    ! the above method writes its results to the Rayleigh rangebins
    ! so we still need to fill the corresponding Mie rangebins
    call Copy_Backsc_Ratio_Rayl_to_Mie(WorkingData)

  end subroutine Calc_BackscRatio_Comp_US
    !  #]
  subroutine Copy_Mie_BackscRatio_from_L1B(L1B, WorkingData, &
                                           error_flag, fallback)
    !  #[ copy the Mie backscatter from the L1B result
    TYPE(L1B_BRC_DataType),    intent(in):: L1B         ! input
    TYPE(WorkingDataType), intent(inout) :: WorkingData ! input/output
    integer,               intent(out)   :: error_flag  ! output
    logical, optional,     intent(in)    :: fallback    ! optional input

    ! local variables
    integer :: nr_of_measurements, m, r
    logical :: fill_missing_only

    fill_missing_only = .false.
    IF (present(fallback)) THEN
       IF (fallback) fill_missing_only = .true.
    END IF

    error_flag = no_error
    nr_of_measurements = WorkingData%nr_of_measurements
    measloop: DO m=1, nr_of_measurements
       Rangegateloop: DO r=1, WorkingData%MieMeas(m)%nr_of_range_gates
          !  #[ copy backscatter ratio for the Mie rangebins

          IF (fill_missing_only) THEN
             IF (missing_real(WorkingData%MieMeas(m)%RangeGate(r)%&
                                             backscatter_ratio)) THEN
                ! print *,"TESTJOS: fallback used for m=",m," r=",r
             ELSE
                cycle Rangegateloop
             END IF
          END IF

          WorkingData%MieMeas(m)%RangeGate(r)%backscatter_ratio = &
               L1B%PCD_ADS%Measurement_PCD(m)%&
                   Meas_Alt_Bin_PCD(r)%scattering_ratio
          ! TO DO: activate this when the rho_L1B_Mie_ok has been
          !        added to the working datastructure
          !        For the moment this item is set in screening_l1b_data.F90
          !WorkingData%MieMeas(m)%RangeGate(r)%rho_ok = &
          !     WorkingData%MieMeas(m)%RangeGate(r)%rho_L1B_Mie_ok
          !  #]
       END DO Rangegateloop
    END DO measloop
    
    ! the above method writes its results to the Mie rangebins
    ! so we still need to fill the corresponding Rayleigh rangebins
    call Copy_Backsc_Ratio_Mie_to_Rayl(WorkingData)

  end subroutine Copy_Mie_BackscRatio_from_L1B
    !  #]
  subroutine Copy_Backsc_Ratio_Rayl_to_Mie(WorkingData)
    !  #[
    ! copy the backscatter ratio from Rayleigh to Mie rangebins
    ! but only if missing, to not overwrite any results
    ! if we are using a fallback method
    TYPE(WorkingDataType), intent(inout)    :: WorkingData ! input/output

    ! local variables
    integer   :: nr_of_measurements, m, r, r2
    integer   :: rrb1, rrb2, count
    real(r8_) :: sum

    nr_of_measurements = WorkingData%nr_of_measurements
    measloop: DO m=1, nr_of_measurements
       Rangegateloop: DO r=1, WorkingData%MieMeas(m)%nr_of_range_gates
          !  #[
          ! a value is present already, so don't copy
          ! to not overwrite any results if we are using a fallback method 
          IF (.not. missing_real(WorkingData%MieMeas(m)%RangeGate(r)%&
                                 backscatter_ratio)) cycle Rangegateloop

          ! we have 2 cases here:
          ! -we have both Mie and Rayleigh data
          ! -we have Mie but no Rayleigh data

          rrb1 = WorkingData%MieMeas(m)%RangeGate(r)%first_rayleigh_rangebin
          rrb2 = WorkingData%MieMeas(m)%RangeGate(r)%last_rayleigh_rangebin

          OnlyMie: IF ( (rrb1 .eq. -1) .or. (rrb2 .eq. -1) ) THEN
             ! we don't have a Rayleigh rangebin corresponding to this 
             ! Mie rangebin. So nothing to do/copy, the Mie value will
             ! remain missing
          ELSE
             matching_rangebins1: IF (rrb1 .eq. rrb2) THEN 
                ! This Mie range bin overlaps with just 1 Rayleigh range bin,
                ! and might be just part of it. So just copy the 
                ! Rayleigh value

                ! The Rayleigh range bin might be larger than the Mie range bin
                ! or may partially overlap, although we
                ! don't expect to use such rangebin definitions. 
                ! However, we have done all warning/error generation already
                ! in Match_Mie_And_Rayleigh_Bins(), so if this condition
                ! occurs we try to process it anyway here (assuming there
                ! is a good reason why this strange case happens).

                ! the Mie and Rayleigh rangebins match 1-to-1, so just copy
                WorkingData%MieMeas(m)%RangeGate(r)%backscatter_ratio = &
                     WorkingData%RayleighMeas(m)%RangeGate(rrb1)%backscatter_ratio
                WorkingData%MieMeas(m)%RangeGate(r)%rho_ok = &
                     WorkingData%RayleighMeas(m)%RangeGate(r)%rho_ok
             ELSE
                ! we have several Rayleigh rangegates for this Mie Rangegate
                ! this situation is not present in the rangebin definitions 
                ! that we expect to see.
                ! for now take the average over the available Rayleigh bins
                ! (only makes sense for slowly varying profiles)
                sum   = 0.0
                count = 0
                DO r2 = rrb1,rrb2
                   IF (.not. missing_real(WorkingData%RayleighMeas(m)%RangeGate(r2)%&
                                              backscatter_ratio)) THEN
                      count = count + 1
                      sum = sum + &
                           WorkingData%RayleighMeas(m)%RangeGate(r2)%backscatter_ratio
                   END IF
                END DO
                
                IF (count .gt. 0) THEN
                   WorkingData%MieMeas(m)%RangeGate(r)%backscatter_ratio = &
                        sum/count
                   WorkingData%MieMeas(m)%RangeGate(r)%rho_ok = .true.
                ELSE
                   ! no valid results, so don't update the Mie value
                   ! (it stays missing)
                END IF
                
                !print *,"m,r=",m,r
                !print *,"rrb1,rrb2=",rrb1,rrb2
                !print *,"count = ", count," sum = ",sum
                !print *,"averaged backscatter ratios: ",&
                !     WorkingData%RayleighMeas(m)%RangeGate(rrb1:rrb2)%backscatter_ratio
                !print *,"from rayleigh range bins ",rrb1," upto",rrb2
                !print *,"to mie rangebin ",r
                !print *,"resulting value = ",sum/count

             END IF matching_rangebins1
          END IF OnlyMie
       !  #]
       END DO Rangegateloop
    END DO measloop

  end subroutine Copy_Backsc_Ratio_Rayl_to_Mie
    !  #]
  subroutine Copy_Backsc_Ratio_Mie_to_Rayl(WorkingData)
    !  #[
    ! copy the backscatter ratio from Mie to Rayleigh rangebins
    ! but only if missing, to not overwrite any results
    ! if we are using a fallback method
    TYPE(WorkingDataType), intent(inout) :: WorkingData ! input/output

    ! local variables
    integer   :: nr_of_measurements, m, m2, r
    integer   :: mrb1, mrb2, count
    real(r8_) :: sum

    nr_of_measurements = WorkingData%nr_of_measurements
    measloop: DO m=1, nr_of_measurements
       Rangegateloop: DO r=1, WorkingData%RayleighMeas(m)%nr_of_range_gates
          !  #[
          ! a value is present already, so don't copy
          ! to not overwrite any results if we are using a fallback method 
          IF (.not. missing_real(WorkingData%RayleighMeas(m)%RangeGate(r)%&
                                 backscatter_ratio)) cycle Rangegateloop

          ! we have several cases here:
          ! -Mie and Rayleigh rangebins match one to one (just copy)
          ! -multiple Mie match to one Rayleigh rangbin (average?)
          ! -Mie and Rayleigh are not properly aligned (average?)
          ! -we have Mie but no Rayleigh data (nothing to copy)

          mrb1 = WorkingData%RayleighMeas(m)%RangeGate(r)%first_mie_rangebin
          mrb2 = WorkingData%RayleighMeas(m)%RangeGate(r)%last_mie_rangebin

          OnlyRayleigh: IF ( (mrb1 .eq. -1) .or. (mrb2 .eq. -1) ) THEN
             ! we don't have a Mie rangebin corresponding to this 
             ! Rayleigh rangebin. So nothing to do/copy, the Rayleigh
             ! value will remain missing.
          ELSE
             matching_rangebins1: IF (mrb1 .eq. mrb2) THEN 
                ! This Rayleigh range bin overlaps with just 1 Mie range bin.
                ! The Mie range bin might be just part of the Rayleigh bin,
                ! or may be larger than the Rayleigh bin, although we
                ! don't expect to use such rangebin definitions. 
                ! However, we have done all warning/error generation already
                ! in Match_Mie_And_Rayleigh_Bins(), so if this condition
                ! occurs we try to process it anyway here (assuming there
                ! is a good reason why this strange case happens).

                ! the Mie and Rayleigh rangebins match 1-to-1, so just copy
                 WorkingData%RayleighMeas(m)%RangeGate(r)%backscatter_ratio = &
                      WorkingData%MieMeas(m)%RangeGate(mrb1)%backscatter_ratio

                 WorkingData%RayleighMeas(m)%RangeGate(r)%rho_ok = &
                      WorkingData%MieMeas(m)%RangeGate(r)%rho_ok
                 !print *,"m,r = ",m,r," mrb1,mrb2= ",mrb1,mrb2
                 !print *,"copied backscatter ratio: ",&
                 !     WorkingData%MieMeas(m)%RangeGate(mrb1)%backscatter_ratio
                 !print *,"from mie range bin ",mrb1," to rayleigh rangebin ",r
                 !print *,"and copied rho_ok flag: ",
                 !     WorkingData%MieMeas(m)%RangeGate(r)%rho_ok
             ELSE
                ! several Mie rangebins are present in this rayleigh bin,
                ! for now take the average over the available mie bins
                ! (only makes sense for slowly varying profiles)
                !
                ! a more elaborate method would maybe be to take a waited
                ! average and taking the weights proportional to the
                ! size of the overlap between the rangebins.
                ! 
                ! an better option might be to just take the maximum
                ! backscatter ratio that occurs within this rayleigh rangebin
                ! It depends on what we will use this number for.
                ! If we would do classification on it, the maximum would
                ! probably be most appropriate, since that indicates
                ! whether we have a cloud/aerosol layer or not.
                ! After averaging this information might be lost.
                ! When doing decontamination of the rayleigh signal we
                ! should use each mie backscatter value combined with
                ! the mie signal, and not use a rayleigh-bin averaged
                ! value of the backscatter ratio.
                sum   = 0.0
                count = 0
                DO m2 = mrb1,mrb2
                   IF (.not. missing_real(WorkingData%MieMeas(m)%RangeGate(m2)%&
                                              backscatter_ratio)) THEN
                      count = count + 1
                      sum = sum + &
                           WorkingData%MieMeas(m)%RangeGate(m2)%backscatter_ratio
                   END IF
                END DO
                
                IF (count .gt. 0) THEN
                   WorkingData%RayleighMeas(m)%RangeGate(r)%backscatter_ratio = &
                        sum/count
                   WorkingData%RayleighMeas(m)%RangeGate(r)%rho_ok = .true.
                ELSE
                   ! no valid results, so don't update the Rayleigh value
                   ! (it stays missing)
                END IF

                !print *,"m,r=",m,r
                !print *,"mrb1,mrb2=",mrb1,mrb2
                !print *,"count = ", count," sum = ",sum
                !print *,"averaged backscatter ratios: ",&
                !     WorkingData%MieMeas(m)%RangeGate(mrb1:mrb2)%backscatter_ratio
                !print *,"from mie range bins ",mrb1," upto",mrb2
                !print *,"to rayleigh rangebin ",r
                !print *,"resulting value = ",sum/count

             END IF matching_rangebins1
          END IF OnlyRayleigh
       !  #]
       END DO Rangegateloop
    END DO measloop

  end subroutine Copy_Backsc_Ratio_Mie_to_Rayl
    !  #]
  !------------------------------------------
  ! 2 helper functions for the PrintBackscatterRatio routine below
  function ConstructRayleighLevelString(WorkingData,&
                                        m,r,PrintClass) result(txt)
    !  #[
    TYPE(WorkingDataType), intent(in) :: WorkingData ! input
    integer,               intent(in) :: m           ! input
    integer,               intent(in) :: r           ! input
    logical,  optional,    intent(in) :: PrintClass  ! optional input
    character(len=38)              :: txt ! result
 
    ! local variables
    integer   :: rayl_class
    real(r8_) :: rayl_ext, rayl_backsc
    integer   :: rayl_alt_start, rayl_alt_end
    logical   :: do_PrintClass 

    do_PrintClass = .false.
    IF (present(PrintClass)) do_PrintClass = PrintClass

    rayl_alt_start = real2int(WorkingData%RayleighMeas(m)%RangeGate(r)%alt_start)
    rayl_alt_end   = real2int(WorkingData%RayleighMeas(m)%RangeGate(r)%alt_end)
    rayl_ext       = WorkingData%RayleighMeas(m)%RangeGate(r)%extinction
    rayl_backsc    = WorkingData%RayleighMeas(m)%RangeGate(r)%backscatter_ratio
    rayl_class     = WorkingData%RayleighMeas(m)%RangeGate(r)%class
    
    if (missing_int(rayl_alt_start)) then
       write(txt,"(i2,1x,'missing',28x)") r
    else
       IF (do_PrintClass) THEN
          write(txt,"(i2,1x,i5,'-',i5,1x,2(es10.3e2,1x),i1)") &
               r,rayl_alt_start,rayl_alt_end,rayl_ext,rayl_backsc,rayl_class
       ELSE
          write(txt,"(i2,1x,i5,'-',i5,1x,2(es10.3e2,1x),1x)") &
               r,rayl_alt_start,rayl_alt_end,rayl_ext,rayl_backsc
       END IF
    end if

  end function ConstructRayleighLevelString
  !  #]
  function ConstructMieLevelString(WorkingData,m,r,PrintClass) result(txt)
    !  #[
    TYPE(WorkingDataType), intent(in) :: WorkingData ! input
    integer,               intent(in) :: m           ! input
    integer,               intent(in) :: r           ! input
    logical,  optional,    intent(in) :: PrintClass  ! optional input
    character(len=38)              :: txt ! result
 
    ! local variables
    integer   :: mie_class
    real(r8_) :: mie_ext, mie_backsc
    integer   :: mie_alt_start, mie_alt_end
    logical   :: do_PrintClass 

    do_PrintClass = .false.
    IF (present(PrintClass)) do_PrintClass = PrintClass

    mie_alt_start = real2int(WorkingData%MieMeas(m)%RangeGate(r)%alt_start)
    mie_alt_end   = real2int(WorkingData%MieMeas(m)%RangeGate(r)%alt_end)
    mie_ext       = WorkingData%MieMeas(m)%RangeGate(r)%extinction
    mie_backsc    = WorkingData%MieMeas(m)%RangeGate(r)%backscatter_ratio
    mie_class     = WorkingData%MieMeas(m)%RangeGate(r)%class

    if (missing_int(mie_alt_start)) then
       write(txt,"(i2,1x,'missing',28x)") r
    else
       IF (do_PrintClass) THEN
          write(txt,"(i2,1x,i5,'-',i5,1x,2(es10.3e2,1x),i1)") &
               r,mie_alt_start,mie_alt_end,mie_ext,mie_backsc,mie_class
       ELSE
          write(txt,"(i2,1x,i5,'-',i5,1x,2(es10.3e2,1x),1x)") &
               r,mie_alt_start,mie_alt_end,mie_ext,mie_backsc
       END IF
    end if

  end function ConstructMieLevelString
  !  #]
  ! print extinction/backscatratio/class.result/rangebins for 1 measurement  
  subroutine PrintBackscatterRatio(WorkingData,Meas_Nr,PrintClass)
    !  #[
    TYPE(WorkingDataType), intent(in) :: WorkingData ! input
    integer,               intent(in) :: Meas_Nr     ! input
    logical,  optional,    intent(in) :: PrintClass  ! optional input

    ! local variables
    integer           :: mrb1, mrb2, r_m, r, last_mie_rangebin_printed
    character(len=38) :: txt_Rayleigh
    character(len=38) :: txt_Mie
    logical           :: do_PrintClass 

    ! remember which Mie rangebin was last printed along with a corresponding
    ! Rayleigh rangebin, in order to be able to print excess Mie rangebins
    ! below the lowest Rayleigh rangebin.
    last_mie_rangebin_printed = 0

    do_PrintClass = .false.
    IF (present(PrintClass)) do_PrintClass = PrintClass

    ! the numbers are printed as x.xxxexxx so only 4 significant
    ! digits, so we have to compare with REALACC(4)
    write(*,"(a)") "REALACC(4)"
    write(*,"(a,i2)") "Measurement = ",Meas_Nr
    write(*,"(2a)") "Rayleigh.............................. ",&
                    "Mie................................... "
    IF (do_PrintClass) THEN
       write(*,"(2a)") "nr alt_range    extinction backsc_r  c ",&
                       "nr alt_range    extinction backsc_r  c "
    ELSE
       write(*,"(2a)") "nr alt_range    extinction backsc_r    ",&
                       "nr alt_range    extinction backsc_r    "
    END IF

    ! loop along all Rayleigh rangebins first, since they normally
    ! starts highest up in the atmosphere
    do r=1,WorkingData%RayleighMeas(Meas_Nr)%nr_of_range_gates
       ! copy some Rayleigh rangebin fields for shorter notation
       mrb1 = WorkingData%RayleighMeas(Meas_Nr)%RangeGate(r)%first_mie_rangebin
       mrb2 = WorkingData%RayleighMeas(Meas_Nr)%RangeGate(r)%last_mie_rangebin

       txt_Rayleigh = &
            ConstructRayleighLevelString(WorkingData,Meas_Nr,r,PrintClass)

       if ((mrb1 .eq. -1) .or. (mrb2 .eq. -1)) then
          ! we have only Rayleigh data at this level
          print "(a38,1x,38('-'))",txt_Rayleigh
       elseif (mrb1 .eq. mrb2) then
          txt_Mie = ConstructMieLevelString(WorkingData,&
                                            Meas_Nr,mrb1,PrintClass)
          print "(a,1x,a)",txt_Rayleigh,txt_Mie
          last_mie_rangebin_printed = mrb1
       else      
          txt_Mie = ConstructMieLevelString(WorkingData,&
                                            Meas_Nr,mrb1,PrintClass)
          print "(a,1x,a)",txt_Rayleigh,txt_Mie
          DO r_m=mrb1+1,mrb2
             txt_Mie = ConstructMieLevelString(WorkingData,&
                                               Meas_Nr,r_m,PrintClass)
             print "(39x,a)",txt_Mie
          END DO
          last_mie_rangebin_printed = mrb2
       end if
    end do

    do r=last_mie_rangebin_printed+1,WorkingData%MieMeas(Meas_Nr)%nr_of_range_gates
       txt_Mie = ConstructMieLevelString(WorkingData,Meas_Nr,r,PrintClass)
       print "(38('-'),1x,a)",txt_Mie
    end do
    write(*,"(a)") "ENDREALACC"

  end subroutine PrintBackscatterRatio
    !  #]
  !------------------------------------------

end module OpticalProperties
