program TestWriteRBCdata
  !  #[ Documentation
  !     a test program for the ReadRBCdata module
  !     written by: J. de Kloe
  !     created:    15-02-2006
  !
  !     Modifications:
  !     16-Feb-2006  J. de Kloe  Added MPH handling
  !     22-Feb-2006  J. de Kloe  Added SPH handling
  !     27-Mar-2006  D. Tan      Added quality parameters
  !     12-May-2006  D. Tan      Null use of ParKind1 module
  !     18-Apr-2007  J. de Kloe  removed an io-recursion problem
  !     30-Jul-2007  J. de Kloe  adapt to changed interface of Fill_FH_Items
  !     14-Nov-2007  J. de Kloe  decrease validity_period with 1 ms
  !     07-Dec-2007  J. de Kloe  adapted to changed FH/MPH interface
  !     16-Jan-2008  J. de Kloe  phase out integer kind i_
  !     25-Feb-2008  D. Tan      Proc_Center from FixedHeader.System
  !     25-Feb-2008  J. de Kloe  merged in a comment
  !
  !  #]
  !  #[ Modules used

  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE WriteRBCdata, only: InitWriteRBCfileModule,WriteRBCfile,Fill_RBC_Headers
  USE RBC_DataStructure, only: RBC_DataType, Init_RBC_DataStructure, &
       Delete_RBC_DataStructure, Init_RBC_DS, Fill_RBC_DS_with_Dummy_data
  USE FixedHeader,  only: systemstr_LOCAL_MET_INST, maxlen_file_class, &
                          maxlen_system
  USE MainProductHeader, only: maxlen_Proc_Stage, maxlen_Proc_Center

  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE Compiler_Features, only: newline
  USE ErrorHandler, only: no_error, program_abort, error_allocate
  USE Numerics, only: r8_
  USE EE_CFI_Datatypes, only: IntAs, PrintDateTime, SetDateTime, &
                              IntAl, &
                              FAdoxy, &
                              DateTimeType,Construct_KVT_DateTime, &
                              maxlen_datetime_kvt, IncrementDateTime
  !  #]
  !  #[ local variables and parameters

  implicit none

  ! used to fill the FH
  character(len=maxlen_file_class) :: file_class
  TYPE(DateTimeType)               :: validity_start_datetime
  TYPE(DateTimeType)               :: validity_stop_datetime
  character(maxlen_datetime_kvt)   :: text_datetime
  real(r8_)                        :: validity_period 
  integer                          :: file_version
  character(len=maxlen_system)     :: system

  ! used to fill the MPH
  character(len=maxlen_Proc_Stage)    :: proc_stage
  character(len=maxlen_Proc_Center)   :: proc_center
  logical            :: product_err
  ! used to fill the SPH
  integer(IntAs)   :: Num_P  ! =  23
  integer(IntAs)   :: Num_T  ! = 161
  integer(IntAs)   :: Num_F  ! = 937
  integer(IntAs)   :: Num_FP ! = 877
  integer(IntAs)   :: Num_fd ! =  61
  Integer(IntAs)   :: Num_RR ! = 101
  real(r8_)        :: P_min  ! in [Pa]
  real(r8_)        :: P_max  ! in [Pa]
  real(r8_)        :: T_min  ! in [K]
  real(r8_)        :: T_max  ! in [K]
  !
  real(FAdoxy)     :: Laser_Freq_Offset_Start
  real(FAdoxy)     :: Laser_Freq_Offset_Stop
  integer(IntAl)   :: Total_Num_of_Observations
  integer(IntAl)   :: Total_Num_of_Measurements
  integer(IntAl)   :: Total_Num_of_Reference_Pulses
  integer(IntAl)   :: Total_Num_of_Corrupt_Mie_Meas
  integer(IntAl)   :: Total_Num_of_Corrupt_Ray_Meas
  integer(IntAl)   :: Total_Num_of_Corrupt_Mie_RefP
  integer(IntAl)   :: Total_Num_of_Corrupt_Ray_RefP
  real(FAdoxy)     :: Average_Error_FP_Response_A
  real(FAdoxy)     :: Average_Error_FP_Response_B

  integer            :: error_flag, AllocStatus
  TYPE(RBC_DataType) :: RBC

  !  #]
  !  #[ program code
  print *,"starting TestWriteRBCdata"

  !*** initialisations ***
  error_flag = no_error

!  call InitWriteRBCfileModule(error_flag)
  call InitWriteRBCfileModule(error_flag, ChoosenVerbosity=VerbosityReport)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"InitWriteRBCfileModule()")

  call Init_RBC_DataStructure(RBC,error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"Init_RBC_DataStructure()")

  allocate(RBC%RBC_DS,STAT=AllocStatus)
  IF (AllocStatus .ne. 0) &
       call program_abort(error_allocate,"TestWriteRBCdata")

  ! specify the dimensions of the look-up table
  Num_P  =  23
  Num_T  = 161
  Num_F  = 937
  Num_FP = 877
  Num_fd =  61
  Num_RR = 101
  ! allocate and initialise the whole structure
  call Init_RBC_DS(RBC%RBC_DS, Num_P, Num_T, Num_F, Num_FP, Num_fd, Num_RR, &
                   error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"Init_RBC_DS")

  ! fill the datastruct with dummy data
  call Fill_RBC_DS_with_Dummy_data(RBC%RBC_DS,error_flag)

  ! set some test values for FH input
  file_class   = "TEST"
  file_version = 1
  system       = systemstr_LOCAL_MET_INST
  call SetDateTime(validity_start_datetime,2001,03,12,12,34,56.0_r8_)
  validity_period = 1._r8_*365*24*60*60 - 0.001_r8_ ! 1 year in seconds
  ! NOTE: subtract 1 millisecond to make sure the ceil() rounds
  ! in the intended way. The ceil() used to get
  ! FH%Validity_Period%Validity_Stop might return with 1 second more
  ! due to machine inaccuracy ! (And this makes the result not portable!)

  validity_stop_datetime = validity_start_datetime
  call IncrementDateTime(validity_stop_datetime,validity_period)

  ! set some test values for MPH input
  ! proc_stage might be: N (near-real-time), R (reprocessing) or T (test)
  proc_stage  = "T"     ! indicating this is a only testfile
  proc_center = system(1:maxlen_Proc_Center)
                        ! the location where this file was produced
                        ! Copy (possibly truncated) of FixedHeader system
                        ! new requirement in spr 28
  product_err = .false. ! no errors in this test product

  ! set some test values for SPH input
  P_min          = 1013._r8_! in [Pa]
  P_max          = 1013._r8_! in [Pa]
  T_min          =  273._r8_! in [K]
  T_max          =  273._r8_! in [K]
  !
  Laser_Freq_Offset_Start       = 0._r8_ ! in [Mhz]
  Laser_Freq_Offset_Stop        = 0._r8_ ! in [Mhz]
  Total_Num_of_Observations     = 0
  Total_Num_of_Measurements     = 0
  Total_Num_of_Reference_Pulses = 0
  Total_Num_of_Corrupt_Mie_Meas = 0
  Total_Num_of_Corrupt_Ray_Meas = 0
  Total_Num_of_Corrupt_Mie_RefP = 0
  Total_Num_of_Corrupt_Ray_RefP = 0
  Average_Error_FP_Response_A   = 0._r8_
  Average_Error_FP_Response_B   = 0._r8_

  ! some test prints
  call PrintDateTime(validity_start_datetime)
  text_datetime = Construct_KVT_DateTime(validity_start_datetime)
  print *,"validity_start_datetime:KVT formatted = ",trim(text_datetime)

  call Fill_RBC_Headers(RBC, &
       file_class, &                                       ! FH inputs
       validity_start_datetime, validity_stop_datetime,  & ! FH inputs
       file_version, system, &                             ! FH inputs
       proc_stage, proc_center, product_err, &             ! MPH inputs
       Num_P, Num_T, Num_F, Num_FP, Num_fd, Num_RR, &      ! SPH inputs
       P_min, P_max, T_min, T_max, &                       ! SPH inputs
       Laser_Freq_Offset_Start, Laser_Freq_Offset_Stop,&   ! Ditto
       Total_Num_of_Observations,&
       Total_Num_of_Measurements,&
       Total_Num_of_Reference_Pulses,&
       Total_Num_of_Corrupt_Mie_Meas, Total_Num_of_Corrupt_Ray_Meas,&
       Total_Num_of_Corrupt_Mie_RefP, Total_Num_of_Corrupt_Ray_RefP,&
       Average_Error_FP_Response_A,   Average_Error_FP_Response_B,&
       error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"Fill_RBC_Headers()")

  !call Init_L2B_PCD_ADS(L2B(iBRC)%PCD_ADS, &
  !        int(N_Meas,IntAs), int(N_Obs_Max_Rayleigh,IntAs), &
  !        int(N_Range_Gates,IntAs), error_flag)

  print *,'TestWriteRBCdata: calling WriteRBCfile'
  call WriteRBCfile(RBC, error_flag)
  IF (error_flag .ne. no_error) call program_abort(error_flag,"WriteRBCfile")

  call Delete_RBC_DataStructure(RBC)

  print *,"done writing RBC test datafile"
  !  #]
end program TestWriteRBCdata
