module ReadRBCdata
  !  #[ Description:
  !
  !     a module to read the 2 files describing the RBC look-up table
  !     of the Aeolus instrument
  !             
  ! created: 16-02-2006 by J. de Kloe
  !
  ! Modifications:
  !     02-Mar-2006  J. de Kloe  adapted to changes in the XML_module
  !                              and implemented reading the binary file
  !     12-May-2006  D. Tan      Null use of ParKind1 module
  !     20-Jul-2006  J. de Kloe  switched use_buffered_io to .true.
  !                              this reduces runtime on my laptop for test1 
  !                              in main from 7m16s to 1m12s !!!! (I have not 
  !                              yet tested the other tests using the RBC file,
  !                              but expect similar results)
  !     31-Jul-2006 J. de Kloe   changed pos parameter in GetFilePointer to i8_
  !     07-Jun-2007 J. de Kloe   adapted to changed xml interface
  !     18-Oct-2007 J. de Kloe   added 2 nullify statements
  !     12-Nov-2007 J. de Kloe   added check for xmlns attribute in xml file
  !     16-Jan-2008 J. de Kloe   phase out integer kind i_
  !     21-Jan-2008 J. de Kloe   added (temporary) workaround switch in the
  !                              call to Read_MPH_from_KVT
  !     21-Feb-2008 J. de Kloe   removed the workaround again
  !
  !  #]
  !  #[ Modules used:
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE FixedHeader, only: Init_FH, Get_FH_from_xml_tree, Print_FH, &
                         file_type_AuxRBC
  USE MainProductHeader, only: Init_MPH, Get_MPH_from_xml_tree, Print_MPH, &
                               Read_MPH_from_KVT
  USE RBC_SPH, only: Init_RBC_SPH, Get_RBC_SPH_from_xml_tree, Print_RBC_SPH,&
                     Read_RBC_SPH_from_KVT
  USE RBC_DataStructure, only: RBC_DataType, Init_RBC_DS, Read_RBC_DS, &
                               Read_RBC_DS_Buff, GetByteSize_RBC_DS, &
                               Init_RBC_DataStruct_Module
  USE EE_CFI_DataTypes, only: IntAs, GetIntAs, IntAd
  USE XML_module, only: xml_node_type, Init_XML, xml_load_file, &
                        xml_tree_print, delete_xml_tree, ee_h_name,&
                        xml_tree_get_namespace, uri_ns_header
  USE KVT_module, only: max_str_len, Init_KVT_module
  USE DirectBinaryIO, only: Init_DirectBinaryIO_module, BigEndian , &
                            OpenDirectBinaryFile,CloseDirectBinaryFile, &
                            OpenBuffDirectBinaryFile,&
                            CloseBuffDirectBinaryFile,BufferedFile_type,&
                            SetFilePointer,GetFilePointer
  USE ErrorHandler, only: no_error, error_xml, error_programming, &
                          error_opening_file, error_reading_file,&
                          error_allocate
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE AeolusConstants, only: iodd_version_AuxRBC
  USE Logging, only: logmsg,log_debug,log_info,log_progress,log_warn,log_error
  !  #]  
  !  #[ Parameter/Variable/Type declarations

  implicit none

  ! make all module contents private, unless explicitely stated otherwise
  private

  ! the default verbosity setting is Error (as silent as possible)
  integer, save :: LocalVerbosity = VerbosityError

  ! remember the initialisation state of the module
  logical, save :: module_is_initialised = .false.

  !  #]
  !  #[ Module interface
  public :: ReadRBCfile, InitReadRBCfileModule
  !  #]
contains
  !---------------------------------
  subroutine InitReadRBCfileModule(error_flag, ChoosenVerbosity)
    !  #[

    integer,           intent(out) :: error_flag
    integer, optional, intent(in)  :: ChoosenVerbosity

    error_flag = no_error

    IF (present(ChoosenVerbosity)) LocalVerbosity = ChoosenVerbosity

    IF (LocalVerbosity .ge. VerbosityReport) &
         call print_verbosity_setting(LocalVerbosity,ModuleName="ReadRBCdata")

    call Init_KVT_module()

    call Init_DirectBinaryIO_module(error_flag,FileByteOrder=BigEndian)
    if (error_flag .ne. no_error) return

    call Init_XML(error_flag) !(ChoosenVerbosity=VerbosityReport)
    if (error_flag .ne. no_error) return

    call Init_RBC_DataStruct_Module(ChoosenVerbosity)

    module_is_initialised = .true.

  end subroutine InitReadRBCfileModule
    !  #]
  subroutine ReadRBCfile(RBC, filename, error_flag,Only_DBL)
    !  #[
    type(RBC_DataType), intent(inout)  :: RBC
    character(len=*),   intent(in)     :: filename ! without extension
    integer,            intent(out)    :: error_flag
    logical,  optional, intent(in)     :: Only_DBL

    ! local variables
    character(len=256) :: filename_hdr
    character(len=256) :: filename_dbl
    logical            :: Read_Only_DBL 

    error_flag = no_error
    IF (.not. module_is_initialised) THEN
       print *,"Error in ReadRBCfile():"
       print *,"the ReadRBCdata module must be initialised before use ..."
       error_flag = error_programming
       return
    END IF

    Read_Only_DBL = .false.
    IF (present(Only_DBL)) Read_Only_DBL = Only_DBL

    filename_hdr = trim(filename)//".HDR"
    filename_dbl = trim(filename)//".DBL"

    IF (.not. Read_Only_DBL) THEN
       call ReadRBCfile_HDR(RBC, filename_hdr, error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    call ReadRBCfile_DBL(RBC, filename_dbl, error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine ReadRBCfile
    !  #]
  subroutine ReadRBCfile_HDR(RBC, filename_hdr, error_flag)
    !  #[
    type(RBC_DataType), intent(inout)  :: RBC
    character(len=*),   intent(in)     :: filename_hdr
    integer,            intent(out)    :: error_flag

    ! local variables
    integer                      :: AllocStatus
    TYPE(xml_node_type), pointer :: xml_tree
    character(len=256)           :: path
    character(len=256)           :: xml_namespace_RBC
    character(len=256)           :: xml_namespace_uri

    ! init
    error_flag = no_error

    ! to check later if: RBC%FH%File_Name .eq. filename

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"Reading RBC header file..."

    nullify(xml_tree)
    call xml_load_file(filename_hdr, xml_tree, error_flag)
    IF (error_flag .ne. no_error) return

    ! define which xml namespace should be present
    !example: xmlns="http://www.esa.int/schemas/ae/ALD_U_N_2B_01.01"
    !uri_ns_header    ! defined in xml_module
    !file_type_L2B_N  ! defined in FixedHeader
    !iodd_version_L2B ! defined in AeolusConstants
    xml_namespace_RBC = trim(uri_ns_header)//trim(file_type_AuxRBC)//&
                        "_"//trim(iodd_version_AuxRBC)

    ! check whether this namespace is really present in the input file
    call xml_tree_get_namespace(xml_tree,xml_namespace_uri,error_flag)
    if (error_flag .ne. no_error) then
       ! allow this attribute to be missing for now, to allow use
       ! of outdated files.
       ! TODO: remove this exception
       error_flag = no_error
       call logmsg(log_warn,"WARNING in ReadRBCfile_HDR: "//&
                   "could not find xmlns attribute in the root tag "//&
                   "of file: "//trim(filename_hdr))
    else
       if (trim(xml_namespace_RBC) .ne. trim(xml_namespace_uri)) &
            call logmsg(log_warn,"WARNING in ReadRBCfile_HDR: "//&
                        "expected an xmlns attribute with content: "//&
                        trim(xml_namespace_RBC)//", but found an "//&
                        "xmlns attribute with content: "//&
                        trim(xml_namespace_uri)//". This input file may "//&
                        "not be compatible with the current software!")
    end if


    call logmsg(log_debug,"ReadRBCfile_HDR: the xml file was "//&
                          "successfully read")

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"==================="
       call xml_tree_print(xml_tree, error_flag)
       IF (error_flag .ne. no_error) return
       print *,"==================="
    END IF
 
    !------------------------------
    ! allocate memory to hold the FH, MPH and SPH elements
    !------------------------------
    allocate(RBC%FH,RBC%MPH,RBC%SPH,STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation failure in subroutine ReadRBCfile_HDR"
       error_flag = error_allocate
       return
    END IF
    call Init_FH(RBC%FH,error_flag)
    IF (error_flag .ne. no_error) return
    call Init_MPH(RBC%MPH,error_flag)
    IF (error_flag .ne. no_error) return
    call Init_RBC_SPH(RBC%SPH,error_flag)
    IF (error_flag .ne. no_error) return

    !------------------------------
    ! copying the FH XML elements
    !------------------------------
    print *,"copying items to the FH datastructure"
    path = "/"//ee_h_name
    call Get_FH_from_xml_tree(xml_tree, path, RBC%FH, error_flag)
    IF (error_flag .ne. no_error) return

    !------------------------------
    ! copying the MPH XML elements
    !------------------------------
    print *,"copying items to the MPH datastructure"
    call Get_MPH_from_xml_tree(xml_tree, path, RBC%MPH, error_flag)
    IF (error_flag .ne. no_error) return

    !------------------------------
    ! copying the SPH XML elements (with DSD's)
    !------------------------------
    print *,"copying items to the SPH datastructure"
    call Get_RBC_SPH_from_xml_tree(xml_tree, path, RBC%SPH, error_flag)
    IF (error_flag .ne. no_error) return

    !------------------------------
    ! done                         
    !------------------------------

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       call Print_FH(RBC%FH,prefix="RBC%FH",error_flag=error_flag)
       IF (error_flag .ne. no_error) return
       call Print_MPH(RBC%MPH,prefix="RBC%MPH",error_flag=error_flag)
       IF (error_flag .ne. no_error) return
       call Print_RBC_SPH(RBC%SPH,error_flag,prefix="RBC%SPH")
       IF (error_flag .ne. no_error) return
    END IF

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"deallocating xml items"
    call delete_xml_tree(xml_tree, error_flag)
    IF (error_flag .ne. no_error) return

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"done with reading RBC header file"

    return

  end subroutine ReadRBCfile_HDR
    !  #]
  subroutine ReadRBCfile_DBL(RBC, filename_dbl, error_flag)
    !  #[
    type(RBC_DataType), intent(inout)  :: RBC
    character(len=*),   intent(in)     :: filename_dbl
    integer,            intent(out)    :: error_flag

    ! local variables
    integer        :: fileunit, AllocStatus
    integer(IntAs) :: Num_P, Num_T, Num_F, Num_FP, Num_fd, Num_RR
    integer(IntAd) :: pos_start, pos_end
    integer        :: nbytes_read, nbytes_expected

    TYPE(BufferedFile_type), pointer :: BufferedFile

    ! on my laptop at home, runtime of TestReadRBCdata:
    ! without buffering: 170 s
    ! with buffering:     25 s

!    logical, parameter :: use_buffered_io = .false.
    logical, parameter :: use_buffered_io = .true.

    ! init
    error_flag = no_error
    nullify(BufferedFile)

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"Reading RBC binary file:"
       print *,trim(filename_dbl)
    END IF

    call OpenDirectBinaryFile(filename_dbl,"read",fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 999

    !------------------------------
    ! read and/or compare MPH elements
    !------------------------------
    IF (.not. associated(RBC%MPH)) THEN
       ! allow this, to enable the possibility of reading the dbl file only,
       ! without reading the HDR files.
       allocate(RBC%MPH,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine ReadRBCfile_DBL"
          error_flag = error_allocate
          goto 999
       END IF
       call Init_MPH(RBC%MPH,error_flag)
       IF (error_flag .ne. no_error) goto 999

       call Read_MPH_from_KVT(RBC%MPH, fileunit, error_flag)
       IF (error_flag .ne. no_error) goto 999
    ELSE
       ! read the KVT values anyway, and compare them to what was
       ! read from the XML file, then replace the values from XML
        call Read_MPH_from_KVT(RBC%MPH, fileunit, error_flag,Compare=.true.)
       IF (error_flag .ne. no_error) goto 999

    END IF

    !------------------------------
    ! read and/or compare SPH elements
    !------------------------------
    IF (.not. associated(RBC%SPH)) THEN
       ! allow this, to enable the possibility of reading the dbl file only,
       ! without reading the HDR files.
       allocate(RBC%SPH,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine ReadRBCfile_DBL"
          error_flag = error_allocate
          goto 999
       END IF
       call Init_RBC_SPH(RBC%SPH,error_flag)
       IF (error_flag .ne. no_error) goto 999

       call Read_RBC_SPH_from_KVT(RBC%SPH, fileunit, error_flag)
       IF (error_flag .ne. no_error) goto 999
    ELSE
       call Read_RBC_SPH_from_KVT(RBC%SPH,fileunit,error_flag,Compare=.true.)
       IF (error_flag .ne. no_error) goto 999
    END IF

    !IF (LocalVerbosity .ge. VerbosityReport) THEN
    !   call Print_FH(RBC%FH,prefix="RBC%FH")
    !   call Print_MPH(RBC%MPH,prefix="RBC%MPH")
    !   call Print_RBC_SPH(RBC%SPH,prefix="RBC%SPH")
    !END IF

    allocate(RBC%RBC_DS,STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation failure in subroutine ReadRBCfile_DBL"
       error_flag = error_allocate
       goto 999
    END IF

    call GetIntAs(RBC%SPH%Num_P, Num_P, error_flag)
    IF (error_flag .ne. no_error) goto 999
    call GetIntAs(RBC%SPH%Num_T, Num_T, error_flag)
    IF (error_flag .ne. no_error) goto 999
    call GetIntAs(RBC%SPH%Num_F, Num_F, error_flag)
    IF (error_flag .ne. no_error) goto 999
    call GetIntAs(RBC%SPH%Num_FP,Num_FP,error_flag)
    IF (error_flag .ne. no_error) goto 999
    call GetIntAs(RBC%SPH%Num_fd,Num_fd,error_flag)
    IF (error_flag .ne. no_error) goto 999
    call GetIntAs(RBC%SPH%Num_RR,Num_RR,error_flag)
    IF (error_flag .ne. no_error) goto 999

    ! allocate the needed datatructure using the sizes from  the SPH
    call Init_RBC_DS(RBC%RBC_DS,Num_P, Num_T, Num_F, Num_FP, &
                     Num_fd, Num_RR, error_flag)
    IF (error_flag .ne. no_error) goto 999
    
    call GetFilePointer(fileunit,pos_start,error_flag)
    IF (error_flag .ne. no_error) goto 998

    IF (use_buffered_io) THEN
       allocate(BufferedFile,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine ReadRBCfile_DBL"
          error_flag = error_allocate
          goto 998
       END IF

       call CloseDirectBinaryFile(fileunit)

       call OpenBuffDirectBinaryFile(filename_dbl,"read",&
            BufferedFile,error_flag,BufferSize=8192)
       IF (error_flag .ne. no_error) goto 998

       call SetFilePointer(BufferedFile,pos_start,error_flag)
       IF (error_flag .ne. no_error) goto 998

       call Read_RBC_DS_Buff(BufferedFile,RBC%RBC_DS,error_flag)
       IF (error_flag .ne. no_error) goto 999

       call GetFilePointer(BufferedFile,pos_end,error_flag)
       IF (error_flag .ne. no_error) goto 998

       call CloseBuffDirectBinaryFile(BufferedFile, error_flag)
       IF (error_flag .ne. no_error) goto 998

       deallocate(BufferedFile)
    ELSE
       call Read_RBC_DS(fileunit,RBC%RBC_DS,error_flag)
       IF (error_flag .ne. no_error) goto 999

       call GetFilePointer(fileunit,pos_end,error_flag)
       IF (error_flag .ne. no_error) goto 998

       call CloseDirectBinaryFile(fileunit)
    END IF

    nbytes_read     = int(pos_end-pos_start)
    nbytes_expected = GetByteSize_RBC_DS(RBC%RBC_DS)

    IF (nbytes_read .ne. nbytes_expected) THEN
       print *,"ERROR in ReadRBCfile_DBL: reading RBC_DS"
       print *,"pos_start       = ",pos_start
       print *,"pos_end         = ",pos_end
       print *,"nbytes_read     = ",nbytes_read
       print *,"nbytes_expected = ",nbytes_expected
       print *,"something went wrong when reading the RBC_DS from the"
       print *,"binary file, probably a programming error ..."
       error_flag = error_programming
       IF (use_buffered_io) THEN
          goto 998
       ELSE
          goto 999
       END IF
    END IF

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"done with reading RBC binary file"

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine ReadRBCfile_DBL (buffered_io problem)"
    call CloseBuffDirectBinaryFile(BufferedFile, error_flag)
    return

999 print *,"ABORTING subroutine ReadRBCfile_DBL"
    call CloseDirectBinaryFile(fileunit)
    return

  end subroutine ReadRBCfile_DBL
    !  #]
  !---------------------------------
end module ReadRBCdata
