PROGRAM TEST_CORRECTION

  !
  ! This program estimates the wind velocity according to the values
  ! of pressures Pmod, temperatures Tmod, measured Rayleigh channel
  ! responses RRmes, and scattering ratios rho.
  !
  ! Three tables are required in input:
  ! binary file 'CALIB_GRID.dat'
  ! binary file 'TENTI_SPECTRUM.dat'
  ! ASCII file 'TransFPinterp.dat'
  !
  ! SUBROUTINES CALLED: 1
  ! subroutine correction
  !
  ! History
  !     13-Jun-2005 Paul Poli, METEO FRANCE   Original code
  !     28-Sep-2005 Paul Poli  Added dwind/dT, dwind/dP
  !     04-Nov-2005 J. de Kloe simplified by adding all subroutines together
  !                            in a single module named TentiSpectrum
  !     23-Dec-2005 D. Tan     Add dwind/dR
  !     12-May-2006 D. Tan     Null use of ParKind1 module
  !     16-Jan-2008 J. de Kloe phase out integer kind i_

  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE Numerics,      only: r8_,missing_indicator_real_r8
  USE ErrorHandler,  only: no_error, program_abort
  USE TentiSpectrum, only: READ_TENTISPECTRUM, READ_TRANS, READ_GRID, &
                           CORRECTION

  implicit none

  ! variables needed to load up the tables
  character(len=100) :: filename_transi,filename_spec,filename_grid
  integer                                  :: nfreq,nP,nT,nfr,nRR,nfd
  real(r8_), dimension(:), allocatable     :: freq,T1,T2, &
                                              Pgrid,Tgrid,freqtmp, &
                                              RR0,fd
  real(r8_), dimension(:,:,:), allocatable :: spec,Fcalib
  ! variables needed to test the subroutine correction
  integer                                  :: i,n
  real(r8_), allocatable, dimension(:)     :: Pmod,Tmod,RRmes,rho,&
                                              corr0,corr0dR,corr0dT,&
                                              corr0dP,corr0drho
  integer                                  :: error_flag

  ! load up tables

  filename_transi=''
  filename_transi='TransFPinterp.dat'
  call read_trans(nfreq,.true.,error_flag,filename=filename_transi)
  if (error_flag .ne. no_error) &
       call program_abort(error_flag,"read_trans")
  allocate (freq(nfreq),T1(nfreq),T2(nfreq))
  call read_trans(nfreq,.false.,error_flag,&
                  T1=T1,T2=T2,f=freq,filename=filename_transi)
  if (error_flag .ne. no_error) &
       call program_abort(error_flag,"read_trans")
  print *,'Test_correction: loaded TransFPinterp.dat'
  
  filename_spec=''
  filename_spec='TENTI_SPECTRUM.dat'
  call read_TentiSpectrum(nP,nT,nfr,.true.,error_flag,filename=filename_spec)
  if (error_flag .ne. no_error) &
        call program_abort(error_flag,"read_TentiSpectrum")
  allocate(Pgrid(nP),Tgrid(nT),freqtmp(nfr),spec(nP,nT,nfr))
  call read_TentiSpectrum(nP,nT,nfr,.false., error_flag,&
       Pgrid=Pgrid,Tgrid=Tgrid,freqtmp=freqtmp,spec=spec, &
       filename=filename_spec)
  if (error_flag .ne. no_error) &
        call program_abort(error_flag,"read_TentiSpectrum")
  print *,'Test_correction: loaded TENTI_SPECTRUM.dat'

  filename_grid=''
  filename_grid='CALIB_GRID.dat'
  call read_grid(nP,nT,nRR,nfd,nfreq,.true.,error_flag,filename=filename_grid)
  if (error_flag .ne. no_error) &
        call program_abort(error_flag,"read_grid")
  allocate (RR0(nRR),fd(nfd),Fcalib(nP,nT,nRR))
  call read_grid(nP,nT,nRR,nfd,nfreq,.false., error_flag,&
       Pgrid=Pgrid,Tgrid=Tgrid,RR0=RR0,fd=fd,freq=freq,Fcalib=Fcalib, &
       filename=filename_grid)
  if (error_flag .ne. no_error) &
        call program_abort(error_flag,"read_grid")
  print *,'Test_correction: loaded CALIB_GRID.dat'

  ! setup input profiles

  n=3
  allocate(Pmod(n),Tmod(n),RRmes(n),rho(n))
  allocate(corr0(n),corr0dR(n),corr0dT(n),corr0dP(n),corr0drho(n))
  Pmod=(/10._r8_,510._r8_,1010._r8_/)
  Tmod=(/200._r8_,250._r8_,300._r8_/)
  RRmes=(/-0.3_r8_,0._r8_,0.4_r8_/)
  rho=(/10._r8_,10._r8_,10._r8_/)
  print *,'Test_correction: ',n,' points in input for function correction'
  print *,"REALACC(6)"
  print *,' Pmod Tmod   RRmes  rho'
  do i=1,n
    print '(F6.1,1X,F5.1,1X,F6.3,1X,F7.3)',Pmod(i),Tmod(i),RRmes(i),rho(i)
  enddo
  print *,"ENDREALACC"

  ! call the subroutine correction

  call correction &
  (nfreq,nP,nT,nfr,nRR,nfd,freq,T1,T2,Pgrid,Tgrid,freqtmp,RR0,fd,spec,Fcalib, &
   n,Pmod,Tmod,RRmes,rho,corr0,corr0dR,corr0dT,corr0dP,corr0drho,error_flag)
  if (error_flag .ne. no_error) &
       call program_abort(error_flag,"correction")

  print *,"REALACC(6)"
  print *,"corr0(1:3)  =",corr0(1:3)
  print *,"corr0dR(1:3)=",corr0dR(1:3)
  print *,"corr0dT(1:3)=",corr0dT(1:3)
  print *,"corr0dP(1:3)=",corr0dP(1:3)
  print *,"corr0drho(1:3)=",corr0drho(1:3)
  print *,"ENDREALACC"

  ! free up memory

  deallocate(freq,T1,T2,Pgrid,Tgrid,freqtmp,spec,RR0,fd,Fcalib)
  deallocate(Pmod,Tmod,RRmes,rho,corr0,corr0dR,corr0dT,corr0dP,corr0drho)

END PROGRAM TEST_CORRECTION

