PROGRAM CALIB_GRID

  !
  ! This program computes the frequency table used later 
  ! as a reference to perform the corrections.
  !
  ! It requires two input tables:
  ! - the binary file 'TENTI_SPECTRUM.dat' containing
  !       the 3D table calculated by calib_tenti.x
  ! - the ASCII file 'TransFPISR.dat' containing the Fabry-Perot
  !       spectral responses to be provided by the instrument
  !       team. For now, use file derived from Matlab simulations.
  !
  ! The output table is written in the binary file 'CALIB_GRID.dat'.
  ! The interpolated Fabry-Perot spectral responses are 
  ! saved in the ASCII file 'TransFPinterp.dat'.
  !
  ! This program needs to be rerun after each calibration procedure.
  !
  ! SUBROUTINES CALLED: 7
  ! subroutines read_TentiSpectrum,read_trans,write_grid,write_transinterp
  ! functions fcalcn,farray1,finterp1
  !
  ! History
  !  13-Jun-2005  Paul Poli, METEO FRANCE   Original F90 code from 
  !                          Marie Laure Denneulin's Matlab code
  !  12-May-2006  D. Tan     Null use of ParKind1 module
  !  30-Jan-2007  P. Poli    Added option to select Gauss or Tenti model as
  !                          input/output
  !  16-Jan-2008  J. de Kloe phase out integer kind i_
  !
  
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE Numerics, only: r8_
  USE ErrorHandler, only: no_error, program_abort
  USE Arraytools, only: fcalcn,farray1
  USE Interp1, only: finterp1
  USE AeolusConstants, only: FSR,USR,df,dRR

  USE TentiSpectrum, only: READ_TENTISPECTRUM, READ_TRANS, WRITE_TRANSINTERP,&
                           WRITE_GRID

  implicit none

  integer :: i,j,nP,nT,nfr,nisr,nfreq,nfd,iP,iT,nRR, error_flag
  real(r8_), dimension(:),     allocatable :: Pgrid,Tgrid,freqtmp,T1isr,T2isr,&
                                              fisr,freq,ff,T1,T2, &
                                              fd,RR0,Na,Nb,RR,Fcalib0,Na1,Nb1, &
                                              Fint_R,Rint_FSR,Rint_USR
  real(r8_), dimension(:,:),   allocatable :: K,inctab,K1,onetab,spec2,S1
  real(r8_), dimension(:,:,:), allocatable :: spec,Fcalib,Na_F,Nb_F
  character(len=100)                       :: filename_spec,filename_trans, &
                                              filename_grid,filename_transi
  real(r8_) :: tab1(1), RRmin, RRmax
  
  logical :: Model_Tenti ! Switch to select Tenti or Gauss model
  
! never used
!  integer,dimension(:),    allocatable :: invmask

  Model_Tenti = .false. ! .true. for Tenti, .false. for Gauss
                             ! TODO: accept as command line option?

  error_flag = no_error

!  filename_spec=''
  if(Model_Tenti) then
      filename_spec='TENTI_SPECTRUM.dat'
  else
      filename_spec='GAUSS_SPECTRUM.dat'
  endif
  call read_TentiSpectrum(nP,nT,nfr,.true.,error_flag,filename=filename_spec)
  if (error_flag .ne. no_error) &
       call program_abort(error_flag,"read_TentiSpectrum")
  allocate(Pgrid(nP),Tgrid(nT),freqtmp(nfr),spec(nP,nT,nfr))
  call read_TentiSpectrum(nP,nT,nfr,.false.,error_flag,Pgrid=Pgrid,&
       Tgrid=Tgrid, freqtmp=freqtmp,spec=spec,filename=filename_spec)
  if (error_flag .ne. no_error) &
       call program_abort(error_flag,"read_TentiSpectrum")
  !print *,'Pgrid',Pgrid
  !print *,'Tgrid',Tgrid
  !print *,'freqtmp',freqtmp
  !print *,'spec',spec

  filename_trans=''
  if(Model_Tenti) then
     filename_trans='TransFPISR.dat'
  else
     filename_trans='TransFPISR_Gauss.dat'
  endif
  call read_trans(nisr,.true.,error_flag,filename=filename_trans)
  if (error_flag .ne. no_error) &
       call program_abort(error_flag,"read_trans")
  allocate(T1isr(nisr),T2isr(nisr),fisr(nisr))
  call read_trans(nisr,.false.,error_flag,TA=T1isr,TB=T2isr,f=fisr, &
       filename=filename_trans)
  if (error_flag .ne. no_error) &
       call program_abort(error_flag,"read_trans")

  nfreq=fcalcn(-FSR,df,FSR+df*0.9_r8_)
  allocate(freq(nfreq),ff(nfreq))
  freq=farray1(-FSR,df,nfreq)
  ff=freq-FSR*floor(freq/FSR+0.5_r8_)

  allocate(T1(nfreq),T2(nfreq))
  T1=finterp1(fisr,T1isr,ff,flag_extrapolate=.TRUE.)
  T2=finterp1(fisr,T2isr,ff,flag_extrapolate=.TRUE.)

  nfd=fcalcn(-USR/2._r8_,df,USR/2._r8_+df*0.9_r8_)
  print *,'nfd',nfd

  allocate(Rint_FSR(nfreq),Rint_USR(nfd))
  Rint_FSR=(T1-T2)/(T1+T2)
  j=0
  do i=1,nfreq
    IF (freq(i) .ge. -USR/2._r8_-df*0.1_r8_ .and. &
      & freq(i) .le.  USR/2._r8_+df*0.1_r8_) THEN
      j=j+1
      Rint_USR(j)=Rint_FSR(i)
    ENDIF
  enddo
  if (j.ne.nfd) then
     print *,'problem!!! j <> nfd',j,nfd
     stop
  endif

  filename_transi=''
  if(Model_Tenti) then
     filename_transi='TransFPinterp.dat'
  else
     filename_transi='TransFPinterp_Gauss.dat'
  endif
  call write_transinterp(nfreq,freq,T1,T2,error_flag,filename=filename_transi)
  if (error_flag .ne. no_error) &
       call program_abort(error_flag,"write_transinterp")

  !print '(f10.4)',T2

  allocate(fd(nfd))
  fd=farray1(-USR/2._r8_,df,nfd)
  allocate(K(nfd,1),inctab(1,nfreq),K1(nfd,nfreq),onetab(1,nfreq), &
           Na_F(nP,nT,nfd),Nb_F(nP,nT,nfd))
  K=reshape(farray1(1._r8_,1._r8_,nfd),shape(K))
  inctab=reshape(farray1(0._r8_,1._r8_,nfreq),shape(inctab))
  onetab(:,:)=1._r8_
  K1=matmul(K,onetab)+ &
   & matmul(reshape(onetab(1,1:nfd),(/nfd,1/)),inctab)
  !print '(I6)',int(K1)
!  K=1+[0:length(fd)-1];
!  K1=K'*ones(1,length(freq))+ones(length(fd),1)*[0:length(freq)-1];

  nRR=int(fcalcn(-0.5_r8_,dRR,0.5_r8_))
  allocate(spec2(nfd,nfreq),Na(nfd),Nb(nfd),RR(nfd),S1(nfd,nfreq), &
     Na1(nfd),Nb1(nfd))
  allocate(RR0(nRR),Fcalib0(nRR))
  RR0=farray1(-0.5_r8_,dRR,nRR)  
  allocate(Fcalib(nP,nT,nRR))
! never used
!  allocate(invmask(nfd))
!  do i=1,nfd
!    invmask(i)=nfd-i+1
!  enddo

  allocate(Fint_R(nRR))
  RRmin=minval(Rint_USR)
  RRmax=maxval(Rint_USR)
  do i=1,nRR
  print *,'RRmin/max',RRmin,RRmax,RR0(i),i,size(Fint_R)
     IF (RR0(i) .lt. RRmin) THEN
        Fint_R(i)=fd(1)
     ELSEIF (RR0(i) .gt. RRmax) THEN
        Fint_R(i)=fd(nfd)
     ELSE
        tab1=finterp1(Rint_USR,fd,(/RR0(i)/),flag_extrapolate=.TRUE.)
        Fint_R(i)=tab1(1)
     ENDIF
  enddo
  print *,'Rint_USR',j,nfd,Rint_USR
  print *,'Fint_R',Fint_R

  do iP=1,nP
    do iT=1,nT
      print *,'iP,iT',iP,iT
      do i=1,nfd
         do j=1,nfreq
            S1(i,j)=spec(iP,iT,i+j-1)
         enddo
      enddo
      Na1=reshape(matmul(S1,reshape(T1,(/nfreq,1/))),shape(Na1))*df
      Nb1=reshape(matmul(S1,reshape(T2,(/nfreq,1/))),shape(Nb1))*df
      do i=1,nfd
         Na_F(iP,iT,i)=Na1(i)
         Nb_F(iP,iT,i)=Nb1(i)
      enddo

      do i=1,nfd
        do j=1,nfreq
          spec2(i,j)=spec(iP,iT,int(K1(i,j)))
        enddo
      enddo
      Na=reshape(matmul(spec2,reshape(T1,(/nfreq,1/))),shape(Na))
      Nb=reshape(matmul(spec2,reshape(T2,(/nfreq,1/))),shape(Nb))
      RR=(Na-Nb)/(Na+Nb)
      !print '(F10.4)',RR
      Fcalib0=finterp1(RR,-fd,RR0,flag_extrapolate=.TRUE.)
      Fcalib(iP,iT,:)=Fcalib0
      !print *,'calib0',Fcalib0
      !print '(F10.4)',1.e-8_r8_*Fcalib0
      !stop
      !SPEC2=reshape(squeeze(SPEC(p,t,K1)),length(fd),length(freq));
      !Na=SPEC2*T1;
      !Nb=SPEC2*T2;
      !RR=(Na'-Nb')./(Na'+Nb');
      !Fcalib(p,t,:)=interp1(RR,-fd,RR0,'spline',NaN);
    enddo
  enddo

!save('CALIB_GRID.mat','Fcalib','Pgrid','Tgrid','RR0','fd','freq','-mat')
  filename_grid=''
  if(Model_Tenti) then
     filename_grid='CALIB_GRID.dat'
  else
     filename_grid='CALIB_GRID_Gauss.dat'
  endif
  call write_grid(nP,nT,nRR,nfd,nfreq, &
                  Pgrid,Tgrid,RR0,fd,freq,Fcalib, &
                  Na_F,Nb_F,Fint_R,error_flag,&
                  filename=filename_grid)
  if (error_flag .ne. no_error) &
       call program_abort(error_flag,"write_grid")

  ! never used
!  deallocate(invmask)

  deallocate(Pgrid,Tgrid,freqtmp,spec,Fcalib)
  deallocate(spec2,Na,Nb,RR,RR0,Fcalib0,K,inctab,K1,onetab,fd,T1,T2,freq,ff)
  deallocate(T1isr,T2isr,fisr)
  deallocate(Na_F,Nb_F,Fint_R)

END PROGRAM CALIB_GRID

