module RayleighBrillouinProcessing
  !  #[ documentation
  ! 
  ! This routine is an interface layer between the L2BP and the Iliad code
  ! it: -simplifies the interface, by only exporting the correction()
  !      routine, and automatically calls all needed file loading code
  !      at the first call of correction.
  !     -it allowes easy replacement by another or updated
  !      RayleighBrillouinProcessing, might we need that.
  !
  !     code written by PaulPoli/David Tan
  !     reformatted as a Fortran90 module
  !     by J. de Kloe, KNMI, 09-11-2005
  !
  !     Modifications:
  !     23-Dec-2005 David Tan  Add in output derivative of wind w.r.t R.
  !     11-Apr-2006 Paul Poli  Added routine to load RBC data into memory,
  !                            removed previous variables T1,T2 etc..
  !                            and replaced them by variable RBC
  !     12-May-2006 D. Tan     Null use of ParKind1 module
  !     17-Aug-2006 P. Poli    Removed obsolete variables
  !     30-Jul-2007 P. Poli    Remove variable RBC to make dataflow explicit
  !     16-Oct-2007 P. Poli    Change RBC default date
  !     18-Oct-2007 J. de Kloe added a nullify statement and fixed some intents
  !     16-Jan-2008 J. de Kloe phase out integer kind i_
  !  #]
  !  #[ modules used

  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE Numerics,        only: r8_
!  USE AeolusConstants, only:
!  USE Arraytools,      only:
!  USE LunManager,      only:
!  USE Interp1,         only: 
  USE ErrorHandler,    only: program_abort, no_error, error_allocate, error_programming

  ! use the iliad code for Rayleigh-Brillouin processing
  USE TentiSpectrum, only: read_tentispectrum, read_trans, read_grid,&
                           correction
  USE RBC_DataStructure, only: RBC_DataType, Init_RBC_DataStructure, &
                               Delete_RBC_DataStructure
  USE ReadRBCdata, only: ReadRBCfile, InitReadRBCfileModule
  USE VerbosityControl, only: VerbosityError

  !  #]
  !  #[ variables, parameters
  implicit none

!  ! the following variables are used in the call to correction() 
!  ! and are not very usefull to the user, so I hide them inhere
!  real(r8_), dimension(:), allocatable, save :: freq,T1,T2, RR0,fd
!  real(r8_), dimension(:), allocatable, save :: Pgrid,Tgrid, freqtmp
!  real(r8_), dimension(:,:,:), allocatable, save :: spec,Fcalib
!
!  ! the following variables are used in the call to correction() 
!  ! and are not very usefull to the user, so I hide them inhere
!  integer, save :: nfreq,nP,nT,nfr,nRR,nfd

  !  define what the valid range is for the RRmes input
  real(r8_), parameter ::  RRmes_threshold = 0.5_r8_
  
  !  #]
contains
  !-----------------------------------
  subroutine Rayl_Brill_Proc_Single(RBC, Pmod_in,Tmod_in,RRmes_in,rho_in,&
                                    corr0_out,corr0dR_out,corr0dT_out,&
                                    corr0dP_out,corr0drho_out,error_flag)
    !  #[
    ! a routine to process just one single observation, instead of
    ! an array of observations, which makes the interface even more simple

    ! interface
    TYPE(RBC_DataType), intent(inout) :: RBC ! input/output
    real(r8_), intent(in)  :: Pmod_in,Tmod_in,RRmes_in,rho_in     ! input
    real(r8_), intent(out) :: corr0_out, corr0dR_out, corr0dT_out ! output
    real(r8_), intent(out) :: corr0dP_out,corr0drho_out           ! output
    integer,   intent(out) :: error_flag 

    ! local variables
    integer :: AllocStatus
    real(r8_), dimension(:), pointer :: Pmod,Tmod,RRmes,rho     ! input
    real(r8_), dimension(:), pointer :: corr0, corr0dR, corr0dT ! output
    real(r8_), dimension(:), pointer :: corr0dP,corr0drho       ! output

    ! init and check
    error_flag = no_error

    ! remark: these nullify statements suppress warnings like
    !     Variable 'corr0' at (1) is used but not set
    ! which are produced by recent g95 versions on this code.
    ! JK. 18-10-2007

    nullify(Pmod,Tmod,RRmes,rho)
    nullify(corr0, corr0dR, corr0dT)
    nullify(corr0dP,corr0drho)

    ! Setup the datastructures to be able to call the Rayl_Brill_Proc()
    ! subroutine defined below
    allocate(Pmod(1), Tmod(1), RRmes(1), rho(1), &
             corr0(1), corr0dR(1), corr0dT(1), &
             corr0dP(1), corr0drho(1),STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       error_flag = error_allocate
       return
    END IF

    ! copy the input data
    Pmod(1)  = Pmod_in
    Tmod(1)  = Tmod_in
    RRmes(1) = RRmes_in
    rho(1)   = rho_in

    ! call the multiple measurement version
    call Rayl_Brill_Proc(RBC, Pmod,Tmod,RRmes,rho,corr0,corr0dR,&
                         corr0dT,corr0dP,corr0drho,error_flag)

    ! copy the output data
    corr0_out     = corr0(1)
    corr0dR_out   = corr0dR(1)
    corr0dT_out   = corr0dT(1)
    corr0dP_out   = corr0dP(1)
    corr0drho_out = corr0drho(1)

    ! clean up the used memory
    deallocate(Pmod, Tmod, RRmes, rho, corr0, corr0dR, &
               corr0dT, corr0dP, corr0drho)

  end subroutine Rayl_Brill_Proc_Single
  !  #]
  subroutine Rayl_Brill_Proc(RBC,Pmod,Tmod,RRmes,rho,corr0,corr0dR,&
                             corr0dT,corr0dP,corr0drho,error_flag)
    !  #[
    ! interface
    TYPE(RBC_DataType), intent(inout) :: RBC           ! input/output
    real(r8_), dimension(:), pointer :: Pmod,Tmod,RRmes,rho     ! input
    real(r8_), dimension(:), pointer :: corr0, corr0dR, corr0dT ! output
    real(r8_), dimension(:), pointer :: corr0dP,corr0drho       ! output
    integer, intent(out) :: error_flag 

    ! local variables
    character(len=256) :: filename ! without extension
!    character(len=100) :: filename_transi,filename_spec,filename_grid
    integer :: n_RayleighBrillouin

    ! init and check
    error_flag = no_error
    n_RayleighBrillouin = size(Pmod)
    if (n_RayleighBrillouin .ne. size(Tmod)) error_flag = error_programming
    if (n_RayleighBrillouin .ne. size(RRmes)) error_flag = error_programming
    if (n_RayleighBrillouin .ne. size(rho)) error_flag = error_programming
    if (n_RayleighBrillouin .ne. size(corr0)) error_flag = error_programming
    if (n_RayleighBrillouin .ne. size(corr0dR)) error_flag = error_programming
    if (n_RayleighBrillouin .ne. size(corr0dT)) error_flag = error_programming
    if (n_RayleighBrillouin .ne. size(corr0dP)) error_flag = error_programming
    if (n_RayleighBrillouin .ne. size(corr0drho)) error_flag = error_programming
    if (error_flag .ne. no_error) then
       print *,"Error inside Rayl_Brill_Proc():"
       print *,"one of the input arrays does not have the expected size"
       print *,"of: ",n_RayleighBrillouin,"elements !!!"
       return
    end if

    ! load the datafiles at the first call 
    ! this should be removed later on -- kept if for now 
    ! to prevent many breaks
    IF (.not.associated(RBC%RBC_DS)) THEN

       filename(:) = ' '
       filename = 'AE_TEST_AUX_RBC_L2_20071003T123456_20081002T123456_0001'
       call Load_Rayl_Brill_Tables(RBC, filename, error_flag)
       if (error_flag .ne. no_error) return

    end IF

    ! call the routine that does the actual work
!    call correction(nfreq,nP,nT,nfr,nRR,nfd,freq,T1,T2,&
!                    Pgrid,Tgrid,freqtmp,RR0,fd,spec,Fcalib, &
!                    n_RayleighBrillouin,Pmod,Tmod,RRmes,rho,corr0, &
!                    corr0dR, corr0dT,&
!                    corr0dP,corr0drho,error_flag)
    call correction(RBC, &
                    n_RayleighBrillouin,Pmod,Tmod,RRmes,rho,corr0, &
                    corr0dR, corr0dT,&
                    corr0dP,corr0drho,error_flag)

  end subroutine Rayl_Brill_Proc
  !  #]
  subroutine Load_Rayl_Brill_Tables(RBC, filename, error_flag)
  !  #[

  TYPE(RBC_DataType), intent(out) :: RBC
  character(len=*),   intent(in) :: filename ! without extension
  integer,            intent(out):: error_flag

    error_flag = no_error

    call Init_RBC_DataStructure(RBC, error_flag)
    IF (error_flag .ne. no_error) &
         call program_abort(error_flag,"Init_RBC_DataStructure")  
    print *,"Load_Rayl_Brill_Tables: Initialized RBC_DS"

    call InitReadRBCfileModule(error_flag,ChoosenVerbosity=VerbosityError)
    IF (error_flag .ne. no_error) &
         call program_abort(error_flag,"InitReadRBCfileModule")

    ! read all datablocks
    print *,"Load_Rayl_Brill_Tables: Attempting to read RBC_DS from file"
    print *,trim(filename)
    call ReadRBCfile(RBC, filename, error_flag)
    IF (error_flag .ne. no_error) call program_abort(error_flag,"ReadRBCfile")
    print *,"Load_Rayl_Brill_Tables: Read RBC_DS"

  end subroutine Load_Rayl_Brill_Tables
  !  #]
  subroutine Unload_Rayl_Brill_Tables(RBC)
    TYPE(RBC_DataType), intent(inout) :: RBC
    !  #[
    ! deallocate the memory used by the tables

    call Delete_RBC_DataStructure(RBC)

!    if (allocated(freq))    deallocate(freq)
!    if (allocated(T1))      deallocate(T1)
!    if (allocated(T2))      deallocate(T2)
!    if (allocated(RR0))     deallocate(RR0)
!    if (allocated(fd))      deallocate(fd)
!    if (allocated(Pgrid))   deallocate(Pgrid)
!    if (allocated(Tgrid))   deallocate(Tgrid)
!    if (allocated(freqtmp)) deallocate(freqtmp)
!    if (allocated(spec))    deallocate(spec)
!    if (allocated(Fcalib))  deallocate(Fcalib)

  end subroutine Unload_Rayl_Brill_Tables
    !  #]
  !-----------------------------------
end module RayleighBrillouinProcessing
     
