module TentiSpectrum
  !  #[ documentation
  !     code written by Paul Poli,
  !     reformatted as a Fortran90 module
  !     by J. de Kloe, KNMI, 02-11-2005
  !
  !     See individual subroutines for modification history.
  !
  !     Modifications (global):
  !     03-May-2006  D. Tan       Rename T1 to TA, T2 to TB.
  !     12-May-2006  D. Tan       Null use of ParKind1 module
  !     03-Jul-2006  P. Poli      Adapt to NewRBC interfaces (Nab_F/Fint_R)
  !     17-Aug-2006  P. Poli      Removed obsolete variables
  !     28-Nov-2006  J. de Kloe   bugfix in READ_GRID
  !     16-Jan-2008  J. de Kloe   phase out integer kind i_
  !  #]
  !  #[ modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  ! these support modules are used inside the routines below and are repeated
  ! here to make the dependencies more visible.
  ! The "only:" without arguments lets them do nothing,
  ! except the dependency check.
  USE DirectBinaryIO,  only: Init_DirectBinaryIO_module, &
                             LittleEndian, BigEndian, &
                             OpenDirectBinaryFile, CloseDirectBinaryFile, &
                             byteswap_needed, headtailsize_this_machine
  USE UnformattedIO,   only: read_unf_int4arr, read_unf_real8arr, &
                             read_unf_real8arr3D, read_unf_int4
  USE RBC_DataStructure, only: RBC_DataType
  USE EE_CFI_Datatypes, only: GetIntAs, maxlen_IntAs, IntAs

  USE Numerics,        only: i4_
  USE AeolusConstants, only:
  USE Arraytools,      only:
  USE LunManager,      only:
  USE Interp1,         only: 

  USE ErrorHandler,    only: no_error, error_programming, error_opening_file,&
                             error_reading_file, error_writing_file, &
                             error_allocate
  !  #]
  !  #[ variables, parameters
  implicit none
  !  #]
contains
  !-----------------------------------
  SUBROUTINE CORRECTION &
       (RBC, &
       !nfreq,nP,nT,nfr,nRR,nfd,freq,TA,TB,Pgrid,Tgrid,freqtmp,RR0,fd,spec,Fcalib, &
       n,Pmod,Tmod,RRmes,rho,wind,alpha_R,alpha_T,alpha_P,alpha_rho,error_flag)
    !  #[
    !
    ! Compute the wind velocity given Pressure Pmod, Temperature Tmod,
    ! measured Rayleigh channel Responses RRmes, and scattering ratio RHO
    !
    ! All the input arguments are row vectors of same length
    !
    ! Original code in matlab by Marie-Laure Denneulin
    !
    !     Modifications:
    !        07-Jun-2005 Paul Poli      First Fortran version from matlab code
    !        15-Sep-2005 Paul Poli      Added in output derivatives of wind w.r.t
    !                                   T,P,rho
    !        23-Dec-2005 David Tan      Add in output derivative of wind w.r.t R.
    !        10-Apr-2006 Paul Poli      Use RBC datastructure.
    !        03-Jul-2006 David Tan      Remove print of "error"
    !        06-Jun-2007 Paul Poli      Introduce logical flag_apply_baseline_algo
    !                                   (the baseline algorithm is only applied
    !                                   if this flag is set to .TRUE., in which
    !                                   case the error is calculated)
    !                                   Introduce logical flag_display_error
    !                                   (the error is displayed if this flag is
    !                                   true)
    ! Units:
    !     All pressures    in hPa
    !     All temperatures in K
    !     All frequencies  in Hz
    !
    ! SUBROUTINES CALLED: 5
    ! subroutines hunt_nearest
    ! functions FSPEC_TENTI4grid, farray1, fsign, finterp1
    !
    
    USE Numerics, only: r8_
    USE Arraytools, only: farray1,fsign
    USE Interp1, only: finterp1,hunt_nearest
    USE AeolusConstants, only: lambda,kb,M,c,pi,df
    
    implicit none
    
    ! I/O variables
    
!    integer,                 intent(in) :: nfreq,nP,nT,nfr,nRR,nfd
!    real(r8_),               intent(in) :: freq(nfreq),TA(nfreq),TB(nfreq), &
!         Pgrid(nP),Tgrid(nT),freqtmp(nfr), &
!         spec(nP,nT,nfr), &
!         RR0(nRR),fd(nfd),Fcalib(nP,nT,nRR)
    TYPE(RBC_DataType),      intent(in) :: RBC
    
    integer,                 intent(in) :: n
    real(r8_), dimension(n), intent(in) :: Pmod,Tmod,RRmes,rho
    real(r8_), dimension(n), intent(out):: wind,alpha_R,alpha_T,alpha_P,alpha_rho
    integer,                 intent(out):: error_flag
    
    ! local variables
    
    integer :: tpr,iT,iP,iR,iTlo,iPlo,iRlo,iThi,iPhi,iRhi,i,j
    integer :: AllocStatus
    real(r8_)                            :: fwhmm,ctemie,dfdT,dfdP,dfdR, &
                                            fRAY1,TAtemp1,TBtemp1, &
                                            dRdrho2,dRdf2, &
                                            fRAY2,natemp2,nbtemp2
    real(r8_), dimension(:), allocatable :: Mie,NaRAY1,NbRAY1,RRRAY1,&
                                            tab1,NaRAY2,NbRAY2,RRRAY2,&
                                            tab2,fmod,dfdrho2,fmodele
    real(r8_), dimension(:,:), allocatable :: K,inctab,K1,onetab, &
                                              SPECtmp1,SPECtmp2
    integer,    dimension(:),  allocatable :: invmask

    ! used only by the baseline algorithm:
    real(r8_), dimension(:,:,:), allocatable :: SPECtmp3
    real(r8_), dimension(:),     allocatable :: NaMRB,NbMRB,RRMRB,fMRB,error
    real(r8_), dimension(:,:),   allocatable :: SPECtmp4,SPECMRB

    integer :: nfreq,nP,nT,nfr,nRR,nfd
    real(r8_), dimension(:), allocatable :: Pgrid,Tgrid,RR0,TA,TB,freq,fd,freqtmp
    character(len=maxlen_IntAs) :: str_IntAs
    integer(IntAs)              :: value_IntAs
    logical                     :: flag_apply_baseline_algo=.TRUE., &
                                   flag_display_error=.FALSE.

    error_flag = no_error
   
    str_IntAs = RBC%SPH%Num_P
    call GetIntAs(str_IntAs,value_IntAs,error_flag)
    IF (error_flag .ne. no_error) return
    nP=INT(value_IntAs)
    str_IntAs = RBC%SPH%Num_T
    call GetIntAs(str_IntAs,value_IntAs,error_flag)
    IF (error_flag .ne. no_error) return
    nT=INT(value_IntAs)
    str_IntAs = RBC%SPH%Num_F
    call GetIntAs(str_IntAs,value_IntAs,error_flag)
    IF (error_flag .ne. no_error) return
    nfr=INT(value_IntAs)
    str_IntAs = RBC%SPH%Num_FP
    call GetIntAs(str_IntAs,value_IntAs,error_flag)
    IF (error_flag .ne. no_error) return
    nfreq=INT(value_IntAs)
    str_IntAs = RBC%SPH%Num_RR
    call GetIntAs(str_IntAs,value_IntAs,error_flag)
    IF (error_flag .ne. no_error) return
    nRR=INT(value_IntAs)
    str_IntAs = RBC%SPH%Num_fd
    call GetIntAs(str_IntAs,value_IntAs,error_flag)
    IF (error_flag .ne. no_error) return
    nfd=INT(value_IntAs)
    print '(A,6I6)','correction: nP,nT,nfr,nfreq,nRR,nfd',nP,nT,nfr,nfreq,nRR,nfd
   
    !print *,'Pgrid',Pgrid
    
    ! Mie spectrum where FWHM is fixed equal to the laser FWHM
    
    ! never used 
    !allocate(v0(n),stat=AllocStatus)
    !v0=sqrt(2._r8_*kb*Tmod/M)
    
    fwhmm=c*0.020e-12_r8_/lambda**2._r8_
    ctemie=fwhmm/(2._r8_*sqrt(log(2._r8_)))

    ! process entry
    
    allocate(Mie(nfr),K(nfd,1),inctab(1,nfreq),K1(nfd,nfreq),onetab(1,nfreq),&
             SPECtmp1(nfd,nfreq),NaRAY1(nfd),NbRAY1(nfd),RRRAY1(nfd),&
             invmask(nfd),tab1(1),SPECtmp2(nfd,nfreq),&
             NaRAY2(nfd),NbRAY2(nfd),RRRAY2(nfd),tab2(2),&
             fmod(n),dfdrho2(n),&
             Pgrid(nP),Tgrid(nT),RR0(nRR), &
             TA(nfreq),TB(nfreq),freq(nfreq),fd(nfd), &
             freqtmp(nfr), &
             stat=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation failure in subroutine CORRECTION (part 1)"
       error_flag = error_allocate
       return
    END IF
    if (flag_apply_baseline_algo) then
      allocate(SPECtmp3(1,1,nfr),SPECtmp4(nfd,nfreq),SPECMRB(nfd,nfreq),&
               NaMRB(nfd),NbMRB(nfd),RRMRB(nfd),fMRB(n),error(n), &
               stat=AllocStatus)
      IF (AllocStatus .ne. 0) THEN
         print *,"Allocation failure in subroutine CORRECTION (part 2)"
         error_flag = error_allocate
         return
      END IF
    endif

    ! these intermediate arrays are required because arguments which are
    ! of type RBC%RBC_DS%array cannot be passed directly to subroutines
    ! expecting arrays.
    Tgrid  =REAL(RBC%RBC_DS%T_Grid/100)  ! need Temperatures in K
    Pgrid  =REAL(RBC%RBC_DS%P_Grid/100)  ! need Pressures in hPa
    RR0    =RBC%RBC_DS%RR                ! between -0.5 and +0.5 (approx.)
    TA     =RBC%RBC_DS%TA_FP
    TB     =RBC%RBC_DS%TB_FP
    freq   =REAL(RBC%RBC_DS%F_FP)        ! in Hz
    fd     =REAL(RBC%RBC_DS%Fd)          ! in Hz
    freqtmp=REAL(RBC%RBC_DS%F_Gridtmp)   ! in Hz
    ! print intermediate arrays -- TODO according to VerbosityLevel
    !print '(A,2G20.12)','Pgrid  :',minval(Pgrid),maxval(Pgrid) 
    !print '(A,2G20.12)','Tgrid  :',minval(Tgrid),maxval(Tgrid) 
    !print '(A,2G20.12)','RR0    :',minval(RR0),maxval(RR0) 
    !print '(A,2G20.12)','TA     :',minval(TA),maxval(TA) 
    !print '(A,2G20.12)','TB     :',minval(TB),maxval(TB) 
    !print '(A,2G20.12)','freq   :',minval(freq),maxval(freq) 
    !print '(A,2G20.12)','fd     :',minval(fd),maxval(fd) 
    !print '(A,2G20.12)','freqtmp:',minval(freqtmp),maxval(freqtmp)

    Mie=1._r8_/(ctemie*sqrt(pi))*exp(-freqtmp*freqtmp/(ctemie*ctemie))
    K=reshape(farray1(1._r8_,1._r8_,nfd),shape(K))
    inctab=reshape(farray1(0._r8_,1._r8_,nfreq),shape(inctab))
    onetab(:,:)=1._r8_
    K1=matmul(K,onetab)+ &
         & matmul(reshape(onetab(1,1:nfd),(/nfd,1/)),inctab)
    
    do i=1,nfd
       invmask(i)=nfd-i+1
    enddo
    
    iT=0
    iP=0
    iR=0
   
    do tpr=1,n
       
       !    
       !    % CORRECTION OF TEMPERATURE AND PRESSURE EFFECT
       !    %
       !    % First Step : find the nearest point of the reference grid
       !    
     
       call hunt_nearest(nT,Tgrid,Tmod(tpr),iT,iTlo,iThi)
       call hunt_nearest(nP,Pgrid,Pmod(tpr),iP,iPlo,iPhi)
       call hunt_nearest(nRR,RR0,RRmes(tpr),iR,iRlo,iRhi)
       
       ! exception handling for values out of range
       if (iT .eq. -1) then
          if (Tmod(tpr) .lt. min(Tgrid(1),Tgrid(nT))) then
             ! extrapolate at the low end
             iT   = 1
             iTlo = 1
             iThi = 2
          else
             ! extrapolate at the high end
             iT   = nT
             iTlo = nT-1
             iThi = nT
          end if
       end if
       if (iP .eq. -1) then
          if (Pmod(tpr) .lt. min(Pgrid(1),Pgrid(nP))) then
             ! extrapolate at the low end
             iP   = 1
             iPlo = 1
             iPhi = 2
          else
             ! extrapolate at the high end
             iP   = nP
             iPlo = nP-1
             iPhi = nP
          end if
       end if
       if (iR .eq. -1) then
          if (RRmes(tpr) .lt. min(RR0(1),RR0(nRR))) then
             ! extrapolate at the low end
             iR   = 1
             iRlo = 1
             iRhi = 2
          else
             ! extrapolate at the high end
             iR   = nRR
             iRlo = nRR-1
             iRhi = nRR
          end if
       end if

       !print *,'found iTlo,iPlo,iRlo',iTlo,iPlo,iRlo
       !print *,'found iT,iP,iR      ',iT,iP,iR
       !print *,'found iThi,iPhi,iRhi',iThi,iPhi,iRhi
       !print *,'RRmes(tpr)',RRmes(tpr)
       !print *,'RR0',RR0
       
       !    
       !    % Second step : Compute the derivative of the reference 
       !    % grid nearest point
       !   
       
       dfdT=(RBC%RBC_DS%Fcalib_PTR(iP)%Fcalib_TR(iThi)%Fcalib_R(iR)- &
          &  RBC%RBC_DS%Fcalib_PTR(iP)%Fcalib_TR(iTlo)%Fcalib_R(iR))/ &
          & (Tgrid(iThi)-Tgrid(iTlo))
       dfdP=(RBC%RBC_DS%Fcalib_PTR(iPhi)%Fcalib_TR(iT)%Fcalib_R(iR)- &
          &  RBC%RBC_DS%Fcalib_PTR(iPlo)%Fcalib_TR(iT)%Fcalib_R(iR))/ &
         &  (Pgrid(iPhi)-Pgrid(iPlo))
       dfdR=(RBC%RBC_DS%Fcalib_PTR(iP)%Fcalib_TR(iT)%Fcalib_R(iRhi)- &
          &  RBC%RBC_DS%Fcalib_PTR(iP)%Fcalib_TR(iT)%Fcalib_R(iRlo))/ &
          & (RR0(iRhi)-RR0(iRlo))
       
       !
       !    % Third step : computation of the pressure and temperature
       !    % corrected frequency 
       !
       
       fmod(tpr)=RBC%RBC_DS%Fcalib_PTR(iP)%Fcalib_TR(iT)%Fcalib_R(iR) + &
            & dfdT*(maxval((/Tmod(tpr),Tgrid(iT)/))- &
            &minval((/Tmod(tpr),Tgrid(iT)/)))*fsign(Tmod(tpr)-Tgrid(iT)) + &
            & dfdP*(maxval((/Pmod(tpr),Pgrid(iP)/))- &
            &minval((/Pmod(tpr),Pgrid(iP)/)))*fsign(Pmod(tpr)-Pgrid(iP)) + &
            & dfdR*(maxval((/RRmes(tpr),RR0(iR)/))- &
            &minval((/RRmes(tpr),RR0(iR)/)))*fsign(RRmes(tpr)-RR0(iR))
       
       alpha_T(tpr)=dfdT
       alpha_P(tpr)=dfdP
       alpha_R(tpr)=dfdR
       
       !    
       !    % CORRECTION OF MIE CONTRIBUTION
       !    % 
       !    % Computation of the response of a Tenti spectrum without Mie
       !    % contribution
       !    
       
       do i=1,nfd
          do j=1,nfreq
             SPECtmp1(i,j)=RBC%RBC_DS%Spec_Grid_PTF(iP)%Spec_Grid_TF(iT)%Spec_Grid_F(int(K1(i,j)))
          enddo
       enddo
       NaRAY1=reshape(matmul(SPECtmp1,reshape(TA,(/nfreq,1/))),shape(NaRAY1))*df
       NbRAY1=reshape(matmul(SPECtmp1,reshape(TB,(/nfreq,1/))),shape(NbRAY1))*df
       RRRAY1=(NaRAY1-NbRAY1)/(NaRAY1+NbRAY1)
       
       tab1=finterp1(RRRAY1,-fd,(/RRmes(tpr)/),flag_extrapolate=.TRUE.)
       fRAY1=tab1(1)
       
       tab1=finterp1(freq,TA,(/fRAY1/),flag_extrapolate=.TRUE.)
       TAtemp1=tab1(1)
       
       tab1=finterp1(freq,TB,(/fRAY1/),flag_extrapolate=.TRUE.)
       TBtemp1=tab1(1)
       
       do i=1,nfd
          do j=1,nfreq
             SPECtmp2(i,j)=SPECtmp1(i,j)+(rho(tpr)-1._r8_)*Mie(int(K1(i,j)))
          enddo
       enddo
       
       NaRAY2=reshape(matmul(SPECtmp2,reshape(TA,(/nfreq,1/))),shape(NaRAY2))*df
       NbRAY2=reshape(matmul(SPECtmp2,reshape(TB,(/nfreq,1/))),shape(NbRAY2))*df
       RRRAY2=(NaRAY2-NbRAY2)/(NaRAY2+NbRAY2)
       
       tab1=finterp1(RRRAY2,-fd,(/RRmes(tpr)/),flag_extrapolate=.TRUE.)
       fRAY2=tab1(1)
       
       tab1=finterp1(fd,NaRAY2,(/fRAY2/),flag_extrapolate=.TRUE.)
       natemp2=tab1(1)
       
       tab1=finterp1(fd,NbRAY2,(/fRAY2/),flag_extrapolate=.TRUE.)
       nbtemp2=tab1(1)
       
       !    % Computation of the derivative of frequency wrt scattering ratio
       
       dRdrho2=(TAtemp1-TBtemp1-RRmes(tpr)*(TAtemp1+TBtemp1))/(natemp2+nbtemp2)
       tab2=finterp1(fd(invmask),RRRAY2,(/fRAY2,fRAY2+1._r8_/),flag_extrapolate=.TRUE.)
       dRdf2=tab2(2)-tab2(1)
       dfdrho2(tpr)=dRdrho2/dRdf2
       
       alpha_rho(tpr)=-dfdrho2(tpr)
      
       if (flag_apply_baseline_algo) then 
       !    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
       !    % Compute the baseline correction for evaluating the 
       !    % error of the operational correction scheme.
       !    % This part will be removed in operational running
       !    % mode.
       call FSPEC_TENTI4grid(1,1,nfr,freqtmp,(/Pmod(tpr)/),(/Tmod(tpr)/),&
                             SPECtmp3,error_flag)
       IF (error_flag .ne. no_error) return

       do i=1,nfd
          do j=1,nfreq
             SPECtmp4(i,j)=SPECtmp3(1,1,int(K1(i,j)))
             SPECMRB(i,j)=SPECtmp4(i,j)+(rho(tpr)-1._r8_)*Mie(int(K1(i,j)))
          enddo
       enddo
       NaMRB=reshape(matmul(SPECMRB,reshape(TA,(/nfreq,1/))),shape(NaMRB))
       NbMRB=reshape(matmul(SPECMRB,reshape(TB,(/nfreq,1/))),shape(NbMRB))
       RRMRB=(NaMRB-NbMRB)/(NaMRB+NbMRB)
       tab1=finterp1(RRMRB,-fd,(/RRmes(tpr)/),flag_extrapolate=.TRUE.)
       fMRB(tpr)=tab1(1)
       endif ! flag_apply_baseline_algo
       
    enddo
    
    !% Computation of the frequency corrected of Pressure, Temperature and Mie 
    
    allocate(fmodele(n),stat=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation failure in subroutine CORRECTION"
       error_flag = error_allocate
       return
    END IF
    
    fmodele=fmod+(1._r8_-rho)*dfdrho2
    
    !% Computation of wind velocity. 
    
    wind=-fmodele*lambda/2._r8_
    alpha_R=-alpha_R*lambda/2._r8_
    alpha_T=-alpha_T*lambda/2._r8_
    alpha_P=-alpha_P*lambda/2._r8_
    alpha_rho=-alpha_rho*lambda/2._r8_
    
    !% Display the estimated wind velocity and the error of the operationnal
    !% correction scheme. This part will be removed in operationnal running.
   
    if (flag_apply_baseline_algo) error=-(fmodele-fMRB)*lambda/2._r8_
    
    ! REMARK: the differences between the numbers found by executables
    !         compiled by different compilers can be substantial.
    !         Especially the error(i) below does only reproduce
    !         in the first 2(!) significant numbers, when the g95 and the
    !         pgf90 compilers are compared, even though 
    !         all calculations are done with 8 byte reals. JdK.
    !         (is this an indication of some numerical problem here?)
    !         Therefore the REALACC keyword below (used by the Test
    !         system) is set to a reduced accuray for the error(i).
    !
    print *,'correction: Estimated wind (m/s)'
    do i=1,n
       print *,"REALACC(6)",wind(i),"ENDREALACC"
       if (flag_apply_baseline_algo .and. flag_display_error) &
         print '(A,2F10.4,A)',"REALACC(2)",error(i),"ENDREALACC"
    enddo
    
    !deallocate(v0) ! never used
    deallocate(Mie)
    deallocate(K,inctab,K1,onetab,SPECtmp1,NaRAY1,NbRAY1,RRRAY1)
    deallocate(invmask,tab1,SPECtmp2,NaRAY2,NbRAY2,RRRAY2,tab2)
    deallocate(fmod,dfdrho2)
    deallocate(fmodele)
    deallocate(Tgrid,Pgrid,RR0,TA,TB,freq,fd,freqtmp)
    if (flag_apply_baseline_algo) then
      deallocate(SPECtmp3,SPECtmp4,SPECMRB,NaMRB,NbMRB,RRMRB,fMRB)
      deallocate(error)
    endif
    
  END SUBROUTINE CORRECTION
    !  #] 
  SUBROUTINE FPLASMA(z,w0,error_flag)
    !  #[
    !
    ! History
    !  xx/xx/2005  Marie-Laure Denneulin   Original code, in Matlab
    !  13/06/2005  Paul Poli               Code translated to F90
    !
    ! SUBROUTINES USED: 3
    ! function farray1
    ! function fcumprod_c
    ! function fdiff_r
    !

    USE Numerics, only: r8_,missing_indicator_real_r8
    USE AeolusConstants, only: pi
    USE Arraytools, only: farray1,fcumprod_c,fdiff_r
    
    implicit none
    
    ! I/O variables
    
    complex(r8_), intent(in)  :: z          ! input
    complex(r8_), intent(out) :: w0         ! result
    integer,      intent(out) :: error_flag ! output
    
    ! local variables
    
    real(r8_) :: x,y
    real(r8_), allocatable, dimension(:) :: xy
    
    integer :: i,idf,nidf1,nidf2, AllocStatus
    integer, parameter :: n100=100,n101=101
    
    complex(r8_), allocatable, dimension(:) :: n,N0,term
    complex(r8_) :: zsq
    logical, allocatable, dimension(:) :: mask1,mask2
    
    error_flag = no_error
    ! JK: fill it with the missing value, to make sure it is defined
    !     even when the routine returns after an error
    w0 = cmplx(missing_indicator_real_r8,missing_indicator_real_r8)

    x=REAL(z)
    y=AIMAG(z)
        
    !print *,'x,y,z',x,y,z
    
    IF (x .lt. 0. .or. y .lt. 0.) THEN
       print *,'plasma x<0 or y<0'
       error_flag = error_programming
       return
    ENDIF
    
    IF (x.le.5. .and. y.le.1.5) THEN
       zsq=-cmplx(2._r8_)*z*z

       allocate(n(n101),N0(n101),term(n101),mask1(n101),xy(n101),&
                stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine FPLASMA"
          error_flag = error_allocate
          return
       END IF

       n=cmplx(farray1(0._r8_,1._r8_,n101))
       N0=fcumprod_c(n101,cmplx(2._r8_)*n+cmplx(1._r8_))
       do i=1,n101
          term(i)=-cmplx(2._r8_)*z*(zsq**n(i))/N0(i)
       enddo
       xy(:)=abs(real(term(:)))+abs(aimag(term(:)))
       
       mask1(1:n101) = xy(1:n101) .lt. 1.E-15_r8_
       
       nidf1=count(mask1)
       if (nidf1 .gt. 0) then
          idf=minval((/ minloc(real(n),mask1), n101 /))
       else
          idf=n101
       endif
       deallocate(mask1)
       w0=cmplx(0.,sqrt(pi))*exp(zsq/cmplx(2._r8_))+sum(term(1:idf))
       deallocate(xy,term,N0,n)
    ELSE
       IF (y.gt.1.5_r8_) THEN
          print *,'plasma y>1.5'
          error_flag = error_programming
          return
       ENDIF
       zsq=z*z

       allocate(n(n100),N0(n100),term(n101),xy(n101),&
                mask1(n101),mask2(n100),stat=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine FPLASMA"
          error_flag = error_allocate
          return
       END IF

       n=cmplx(farray1(1._r8_,1._r8_,n100))
       !print *,'n',n
       N0=fcumprod_c(n100,cmplx(2._r8_)*n-cmplx(1._r8_))
       !print *,'N0',N0
       term(1)=-cmplx(1._r8_)
       do i=1,n100
          term(i+1)=-N0(i)/((cmplx(2._r8_)*zsq)**n(i))
       enddo
       term(:)=term(:)/z
       !print *,'term',term
       xy(:)=abs(real(term(:)))+abs(aimag(term(:)))
       !print *,'xy',xy
       mask1(1:n101)=xy(1:n101) .lt. 1.E-15_r8_
       mask2(1:n100)=fdiff_r(n100,xy) .gt. 0.
       nidf1=count(mask1)
       nidf2=count(mask2)
       !print *,'nidf1,nidf2',nidf1,nidf2
       if (nidf1 .gt. 0) then
          if (nidf2 .gt. 0) then
             idf=minval((/ minloc(real(n),mask1), minloc(real(n),mask2), n100 /))
          else
             idf=minval((/ minloc(real(n),mask1), n100 /))
          endif
       else
          if (nidf2 .gt. 0) then
             idf=minval((/ minloc(real(n),mask2), n100 /))
          else
             idf=n100
          endif
       endif
       !print *,'idf',idf
       deallocate(mask1,mask2)
       w0=sum(term(1:idf))
       if (y.eq.0.) w0=w0+cmplx(sqrt(pi))*exp(-zsq)
       deallocate(xy,term,N0,n)        
    ENDIF
    
  END SUBROUTINE FPLASMA
    !  #] 
  SUBROUTINE FSPEC_TENTI4grid(nP,nT,nfr,dfrequency,P,T,spec,error_flag)
    !  #[
    
    ! FUNCTION SPEC=TENTI(FREQ,PRESSURE,TEMPERTURE)
    !
    ! Compute spectrum of UV light (355nm) scattered by air molecules
    ! at pressure P (in hPa) and temperature T (in K).
    ! DFREQUENCY is a row vector containing the input frequencies in Hz.
    ! P and T are row vectors.
    !
    ! The computation is based on Tenti S6 model.
    !
    ! Matlab version of the Fortran code was written by Marie-Laure Denneulin 
    ! on 22/03/05 
    ! It is based on the following documents :
    !   - the Original Fortran code written by R. Seasholtz, NASA Lewis 
    !   - Pan, Shneider and Miles, Coherent Rayleigh-Brillouin Scattering in gases
    !     in the kinetic regime, 2003. This article can be found at
    !     http://molspect.chemistry.ohio-state.edu/institute/2005/praCRBS.pdf
    ! 
    !   - Tenti and Desai, Kinetic Theory of Molecular Gases I, Can.J.Phys.,53,1266
    !   - Tenti and Desai, Kinetic Theory of Molecular Gases II,Can.J.Phys.,53,1279
    !   - Boley, Desai, Tenti, Kinetic Models and Brillouin Scattering in a
    !     Molecular Gas, Can.J.Phys.,50,2158
    !
    ! Modifications
    !  10/06/2005  Paul Poli   Wrote F90 from Matlab code, retained all vars names
    !                          and in-file documentation
    !
    ! SUBROUTINES CALLED: 2
    ! function fplasma
    ! function finvert
    !
    
    USE Numerics, only: r8_
    USE Arraytools, only: finvert
    USE AeolusConstants, only: M,kb,lambda,pi
    
    implicit none

    ! I/O variables

    integer,     intent(in)  :: nP,nT,nfr                   ! input
    real(r8_),   intent(in)  :: dfrequency(nfr),P(nP),T(nT) ! input
    real(r8_),   intent(out) :: spec(nP,nT,nfr)             ! result
    integer,     intent(out) :: error_flag                  ! output
    
    ! local variables
    
    real(r8_) :: etalam,cv,C0,J011,DEN,J100,J110,J1,J2,J001, &
         y,A0(6,6),A1(6,6),A3(12,12),BB(12),XX(12)
    ! never used
    !  real(r8_) :: fu
    real(r8_), allocatable, dimension(:)   :: Patm,etazet,Rint,v0,deltanu
    real(r8_), allocatable, dimension(:,:) :: x,Y0
    integer  :: i,iP,iT,ifr, AllocStatus
    complex(r8_) :: z,w0p,zlam,w0,w1,w2,w3,w4,w5,w6, &
         gam(6,6),AA(6,6),B(6), &
         i00_00, i00_01, i01_01, i00_10, i00_11, i01_10, i01_11, &
         i10_11, i10_10, i11_11
    complex(r8_) :: ci
    complex(r8_) :: c_null

    error_flag = no_error
    
    ci=cmplx(0._r8_,1._r8_)
    c_null=cmplx(0._r8_,0._r8_)
    
    print *,'fspec_tenti4grid: nP,nT,nfr',nP,nT,nfr
    !print *,'P,T,dfrequency',nP,nT,nfr,P,T,dfrequency
    
    ! Definition of atmopsheric and spectrum parameters
    
    allocate(Patm(nP),etazet(nP),Rint(nP),v0(nT),deltanu(nT),&
             stat=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation failure in subroutine FSPEC_TENTI4grid"
       error_flag = error_allocate
       return
    END IF


    Patm(:)=P(:)/1013.25_r8_      ! Pressure in atmospheres
    etalam=0.198_r8_              ! Ratio of Shear viscosity to thermal conductivity
    etazet(:)=1.387_r8_/Patm(:)   ! Ratio of Shear viscoity to bulk viscosity
    cv=1._r8_                     ! Internal specific heat/Boltzman's constant
    Rint(:)=1.5_r8_*(1.5_r8_+cv)/(etazet(:)*cv) ! internal relaxation number
    
    ! never used
    !  fu=M/(etalam*kb*(1.5_r8_+cv)) ! Eucken factor
    
    v0(:)=sqrt(2._r8_*kb*T(:)/M)
    deltanu(:)=2*v0(:)/lambda    !1/e width of the spectrum
    
    ! Definition of Tenti S6 variables
    allocate(x(nT,nfr),Y0(nP,nT), stat=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation failure in subroutine FSPEC_TENTI4grid"
       error_flag = error_allocate
       return
    END IF

    x=matmul(reshape(1./deltanu,(/nT,1/)),reshape(abs(dfrequency),(/1,nfr/)))
    !  print *,'x(1,1)',x(1,1)
    Y0=500._r8_*lambda*1.E6_r8_* &
         & matmul(reshape(Patm,(/nP,1/)),reshape((T+111._r8_)/(2.17_r8_*T*T),(/1,nT/)))
    !  print *,'Y0',Y0(1,1)
    
    !print *,'entering pressure loop'
    do iP=1,nP
       !Computation of the part of the matrix A dependant of the pressure only
       !J_{rln}^{r'l'n'} are linearized collision operators
       C0=(cv/(1.5_r8_+cv))*(cv/(1.5_r8_+cv))
       J011=0.4_r8_*(1.5_r8_+cv)*(1.5_r8_+cv)+cv*(1._r8_+cv/3._r8_)* &
            & etazet(iP)+C0*etazet(iP)**2._r8_/(6._r8_*etalam)
       DEN=4._r8_/15._r8_-etalam+(2._r8_/9._r8_)*etazet(iP)*C0
       J011=J011/DEN
       J011=J011*2._r8_*cv*etalam/(3._r8_*(1.5_r8_+cv)**2._r8_)
       J100=cv/(Rint(iP)*(1.5_r8_+cv))
       J110=2._r8_/3._r8_+5._r8_/6._r8_*cv/(Rint(iP)*(1.5+cv))
       J1=J100*sqrt(3._r8_/2._r8_/cv)   !J_{100}^{001}
       J2=sqrt(5._r8_/8._r8_/cv)*cv/(Rint(iP)*(1.5_r8_+cv))    !J_{011}^{110}
       J001=3._r8_*J100/2._r8_/cv
       
       A0(1,:)=(/1._r8_, 1._r8_, (1._r8_-J110), (1._r8_-J100),      -J1     ,        -J2   /)
       A0(2,:)=(/1._r8_, 1._r8_, (1._r8_-J110), (1._r8_-J100),      -J1     ,        -J2   /)
       A0(3,:)=(/1._r8_, 1._r8_, (1._r8_-J110), (1._r8_-J100),      -J1     ,        -J2   /)
       A0(4,:)=(/1._r8_, 1._r8_, (1._r8_-J110), (1._r8_-J100),      -J1     ,        -J2   /)
       A0(5,:)=(/0._r8_, 0._r8_,     -J2     ,     -J1     , -(J001-1._r8_), -(J011-1._r8_)/)
       A0(6,:)=(/0._r8_, 0._r8_,     -J2     ,     -J1     , -(J001-1._r8_), -(J011-1._r8_)/)
       
       do iT=1,nT
          print *,"REALACC(6)"
          print *,'fspec_tenti4grid P=',P(iP),'T=',T(iT)
          print *,"ENDREALACC"
          y=Y0(iP,iT)
          !        print *,'y',y
          A1(:,:)=-y*A0(:,:)
          !        print *,'A1(1,:)',A1(1,:)
          !        print *,'A1(2,:)',A1(2,:)
          !        print *,'A1(3,:)',A1(3,:)
          !        print *,'A1(4,:)',A1(4,:)
          !        print *,'A1(5,:)',A1(5,:)
          !        print *,'A1(6,:)',A1(6,:)

          do ifr=1,nfr
             z=cmplx(x(iT,ifr),y)
             !print *,'z',z
             !we define :w_k(Z)=sqrt(pi)*integ(t^k * exp(-t^2)/(t-z),t,-inf,+inf)
             !w_0 corresponds to the Plasma Dispersion Function
             ! w0p=FPLASMA(z) replaced by:
             call FPLASMA(z,w0p,error_flag)
             IF (error_flag .ne. no_error) return
             !print *,'w0P',w0P
             
             zlam=ci*cmplx(sqrt(2._r8_))*z
             !print *,'zlam',zlam
             
             !Simple relations exists between W_(k+1) and W_{k}
             W0=ci*w0p/cmplx(sqrt(2._r8_))
             W1=ci*(cmplx(1._r8_)-zlam*W0)
             W2=ci*(-zlam*W1)
             W3=ci*(cmplx(1._r8_)-zlam*W2)
             W4=ci*(-zlam*W3)
             W5=ci*(cmplx(3._r8_)-zlam*W4)
             W6=ci*(-zlam*W5)
             
             !I_{rl}^{r'l'} are integrals which can be expressed in terms of the
             !plasma dispersion function
             I00_00=W0*cmplx(sqrt(2._r8_))
             I00_01=W1*cmplx(sqrt(2._r8_))
             I01_01=W2*cmplx(2._r8_/sqrt(2._r8_))
             I00_10=(-W0+W2)/cmplx(sqrt(3._r8_))
             I00_11=(-cmplx(3._r8_)*W1+W3)/cmplx(sqrt(5._r8_))
             I01_10=(-W1+W3)/cmplx(sqrt(3._r8_))
             I01_11=(-cmplx(3._r8_)*W2+W4)/cmplx(sqrt(5._r8_))
             I10_11=(cmplx(7._r8_)*W1-cmplx(4._r8_)*W3+W5)/cmplx(sqrt(30._r8_))
             I10_10=(cmplx(5._r8_)*W0-cmplx(2._r8_)*W2+W4)/cmplx(sqrt(18._r8_))
             I11_11=(cmplx(13._r8_)*W2-cmplx(6._r8_)*W4+W6)/cmplx(sqrt(50._r8_))
             
             gam(1,:)=(/I00_00, I00_01, I00_11, I00_10, I00_10, I00_11/)
             gam(2,:)=(/I00_01, I01_01, I01_11, I01_10, I01_10, I01_11/)
             gam(3,:)=(/I00_11, I01_11, I11_11, I10_11, I10_11, I11_11/)
             gam(4,:)=(/I00_10, I01_10, I10_11, I10_10, I10_10, I10_11/)
             
             ! I am not yet sure what is the problem here, but my g95
             ! compiler refuses to compile the next 2 statements
             !(might be a compiler bug)
             !            gam(5,:)=(/ cmplx(0._r8_),cmplx(0._r8_), &
             !                        I00_01, I00_00, I00_00, I00_01/)
             !            gam(6,:)=(/ cmplx(0._r8_),cmplx(0._r8_), &
             !                        I01_01, I00_01, I00_01, I01_01/)
             
             ! using this c_null parameter solves the problem....
             !c_null=cmplx(0._r8_)
             gam(5,:)=(/ c_null, c_null, I00_01, I00_00, I00_00, I00_01/)
             gam(6,:)=(/ c_null, c_null, I01_01, I00_01, I00_01, I01_01/)
             
             !            print *,'gam(1,:)',aimag(gam(1,:))
             !            print *,'gam(2,:)',aimag(gam(2,:))
             !            print *,'gam(3,:)',aimag(gam(3,:))
             !            print *,'gam(4,:)',aimag(gam(4,:))
             !            print *,'gam(5,:)',aimag(gam(5,:))
             !            print *,'gam(6,:)',aimag(gam(6,:))
             
             !matrix AA is the matrix of model operator coefficients
             AA=cmplx(A1(:,:))*gam(:,:)
             do i=1,6
                AA(i,i)=AA(i,i)-cmplx(1._r8_)
             enddo
             
             !            print *,'AA (1,:)',aimag(AA (1,:))
             !            print *,'AA (2,:)',aimag(AA (2,:))
             !            print *,'AA (3,:)',aimag(AA (3,:))
             !            print *,'AA (4,:)',aimag(AA (4,:))
             !            print *,'AA (5,:)',aimag(AA (5,:))
             !            print *,'AA (6,:)',aimag(AA (6,:))
             
             !            print *,'AA (1,:)',real(AA (1,:))
             !            print *,'AA (2,:)',real(AA (2,:))
             !            print *,'AA (3,:)',real(AA (3,:))
             !            print *,'AA (4,:)',real(AA (4,:))
             !            print *,'AA (5,:)',real(AA (5,:))
             !            print *,'AA (6,:)',real(AA (6,:))
             
             !vector B is the vector of initial values
             
             ! I am not yet sure what is the problem here, but my g95
             ! compiler refuses to compile the next statement
             !(might be a compiler bug)
             !            B=(/ gam(1,1), gam(2,1), gam(3,1), gam(4,1), &
             !                 cmplx(0._r8_), cmplx(0._r8_) /)
             
             ! using this c_null parameter solves the problem....
             !c_null=cmplx(0._r8_)
             B=(/gam(1,1), gam(2,1), gam(3,1), gam(4,1), c_null, c_null /)
             !the resolution of the equation : A*X=B
             !leads to : X=A^(-1)*B
             A3( 1:6, 1:6)= real(AA)
             A3( 1:6,7:12)=-aimag(AA)
             A3(7:12, 1:6)= aimag(AA)
             A3(7:12,7:12)= real(AA)
             
             !            do i=1,12
             !              print 1000,A3(i,1:12)
             !            enddo
             !  1000    format(12G10.3)
             
             BB( 1:6)= real(B)
             BB(7:12)= aimag(B)
             
             !            print *,'inv',finvert(A3)
             
             XX=reshape(matmul(finvert(A3),reshape(BB,(/12,1/))),shape(XX))
             !            print *,'XX',XX
             !the first component of XX is the result necessary for the
             !computation of the spectrum
             spec(iP,iT,ifr)=XX(1)/pi
          enddo
          !normalisation          
          spec(iP,iT,:)=spec(iP,iT,:)/deltanu(iT)
!                 print *,'spec(',iP,',',iT,',1:',nfr,')=',minval(spec(iP,iT,:)),maxval(spec(iP,iT,:)),SUM(spec(iP,iT,:))
       enddo
    enddo
    
    deallocate(Patm,etazet,Rint,v0,deltanu,x,Y0)
    
  END SUBROUTINE FSPEC_TENTI4grid
  !  #]
  SUBROUTINE FSPEC_GAUSS4grid(nP,nT,nfr,dfrequency,P,T,spec,error_flag)
    !  #[
     USE Numerics, only: r8_
     USE Arraytools, only: finvert
     USE AeolusConstants, only: M,kb,lambda,pi
     
     implicit none
 
     ! I/O variables
 
     integer,     intent(in)  :: nP,nT,nfr                   ! input
     real(r8_),   intent(in)  :: dfrequency(nfr),P(nP),T(nT) ! input
     real(r8_),   intent(out) :: spec(nP,nT,nfr)             ! result
     integer,     intent(out) :: error_flag                  ! output
     
     ! local variables
     
     real(r8_), allocatable, dimension(:)   :: v0,deltanu
     integer  :: iP,iT,ifr
 
     error_flag = no_error
     
     print *,'fspec_gauss4grid: nP,nT,nfr',nP,nT,nfr
     !print *,'P,T,dfrequency',nP,nT,nfr,P,T,dfrequency
     
     allocate(v0(nT),deltanu(nT))  
     v0(:)=sqrt(2._r8_*kb*T(:)/M)
     deltanu(:)=2*v0(:)/lambda    !1/e width of the spectrum
     
     !print *,'entering pressure loop'
     do iP=1,nP
        
        ! print *,'entering temperature loop'
        do iT=1,nT
           print *,"REALACC(6)"
           print *,'fspec_gauss4grid P=',P(iP),'T=',T(iT)
           print *,"ENDREALACC"
 
           do ifr=1,nfr
              spec(iP,iT,ifr)=exp(-dfrequency(ifr)*dfrequency(ifr)/(deltanu(iT)*deltanu(iT)))
           enddo
           !normalisation          
           spec(iP,iT,:)=spec(iP,iT,:)/(sqrt(pi)*deltanu(iT))
           !        print *,'spec(',iP,',',iT,',1:',nfr,')=',spec(iP,iT,:)
        enddo
     enddo
     
     deallocate(v0,deltanu)
     
   END SUBROUTINE FSPEC_GAUSS4grid
   !  #]
   SUBROUTINE WRITE_TENTISPECTRUM(nP,nT,nfr,Pgrid,Tgrid,freqtmp,spec,&
                                 error_flag,filename)
    !  #[
    USE Numerics, only: r8_
    USE LunManager, only: get_lun, free_lun
    
    implicit none
    
    !
    ! Writes out the spec matrix to a binary file
    !
    ! 13/06/2005  Paul Poli, METEO FRANCE    Original code
    !
    
    ! I/O variables
    
    integer,           intent(in)          :: nP,nT,nfr
    real(r8_),         intent(in)          :: Pgrid(nP), &
         Tgrid(nT), &
         freqtmp(nfr), &
         spec(nP,nT,nfr)
    integer,           intent(out)         :: error_flag
    character(len=100),intent(in),optional :: filename
    
    ! local variables
    
    integer            :: iunit
    character(len=100) :: ficname

    error_flag = no_error

    ! this will return -1 if no more free units are available
    ! and this triggers an error in the open statement
    ! which is catched and handled after label 999
    iunit = get_lun()

    IF (present(filename)) THEN
       ficname=trim(filename)
    ELSE
       ficname='TENTI_SPECTRUM.dat'
    ENDIF
    
    open (iunit,FILE=ficname,STATUS='NEW', &
         FORM='UNFORMATTED',ACCESS='SEQUENTIAL',err=999)
    write (iunit,err=998) nP,nT,nfr
    write (iunit,err=998) Pgrid
    write (iunit,err=998) Tgrid
    write (iunit,err=998) freqtmp
    write (iunit,err=998) spec
    close (iunit)
    call free_lun(iunit)
    
    return

998 print *,"could not write file: ",trim(ficname)
    error_flag = error_writing_file
    return

999 print *,"could not open file: ",trim(ficname)
    error_flag = error_opening_file
    return

  END SUBROUTINE WRITE_TENTISPECTRUM
  !  #]
  SUBROUTINE READ_TENTISPECTRUM(nP,nT,nfr,lsizeonly,error_flag,Pgrid,Tgrid,&
                                freqtmp,spec,filename)
    !  #[
    USE Numerics, only: r8_
    USE LunManager, only: get_lun, free_lun
    
    implicit none
    
    !
    ! Reads the spec matrix from binary file
    !
    ! 13/06/2005  Paul Poli, METEO FRANCE    Original code
    !
    
    ! I/O variables
    
    integer,           intent(inout)         :: nP,nT,nfr
    logical,           intent(in)            :: lsizeonly
    integer,           intent(out)           :: error_flag
    real(r8_),         intent(out), optional :: Pgrid(nP),Tgrid(nT),&
                                                freqtmp(nfr),spec(nP,nT,nfr)
    character(len=100),intent(in),  optional :: filename

    ! local variables    
    integer            :: iunit, headtailsize_this_file
    integer(i4_), dimension(3) :: arr
    character(len=100) :: ficname
    character(len=100) :: alt_filename
    logical            :: file_exists
    integer(i4_)       :: nP_tmp, nT_tmp, nfr_tmp

    error_flag = no_error

    IF (present(filename)) THEN
       ficname=trim(filename)
    ELSE
       ficname='TENTI_SPECTRUM.dat'
    ENDIF

    !print *,"Reading file: ",trim(ficname)
    alt_filename = trim(ficname)//".alt"

    ! sanity check
    IF (.not. lsizeonly) THEN
       !print *,"Reading complete file"
       IF (.not. present(Pgrid)) THEN
          print *,"ERROR: optional argument Pgrid missing"
          goto 998
       END IF
       IF (.not. present(Tgrid)) THEN
          print *,"ERROR: optional argument Tgrid missing"
          goto 998
       END IF
       IF (.not. present(freqtmp)) THEN
          print *,"ERROR: optional argument freqtmp missing"
          goto 998
       END IF
       IF (.not. present(spec)) THEN
          print *,"ERROR: optional argument spec missing"
          goto 998
       END IF
    !ELSE
    !   print *,"Reading dimensions only" 
    END IF

    ! remark: a normal unformatted read fails for the gfortran compiler
    ! because it uses 8 byte head/tail marks in stead of the usual
    ! 4 byte head/tail marks that most other compilers on 32-bit platforms use.
    ! This seems to be a deliberate choice, to enable use of
    ! multi gigabyte datafiles ... So we have to make a workaround here
    ! untill the RBC-lookuptable format is converted to ee_cfi format.

    call Init_DirectBinaryIO_module(error_flag, FileByteOrder=LittleEndian)
    if (error_flag .ne. no_error) return

    call OpenDirectBinaryFile(ficname,"read",iunit,error_flag)
    if (error_flag .ne. no_error) goto 999

    !print *,"trying headtailsize_this_file=4"
    headtailsize_this_file=4
    call read_unf_int4arr(iunit,headtailsize_this_file,3,arr,error_flag)
    IF (error_flag .ne. no_error) THEN
       error_flag = no_error
       !print *,"trying headtailsize_this_file=8, since 4 seems wrong ..."
       ! headtailsize of 4 might be wrong, so try again with 8
       ! the filepos is shifted by 4 bytes now, so reopen the file
       ! to correct for this.
       call CloseDirectBinaryFile(iunit)
       call OpenDirectBinaryFile(filename,"read",iunit,error_flag)
       if (error_flag .ne. no_error) goto 999
       headtailsize_this_file=8
       call read_unf_int4arr(iunit,headtailsize_this_file,3,arr,error_flag)
       if (error_flag .ne. no_error) THEN
          print *,"ERROR reading 8 byte head field"
          goto 998
       end if
    END IF

    nP  = arr(1)
    nT  = arr(2)
    nfr = arr(3)

    !print *,"nP,nT,nfr = ",nP,nT,nfr
    !print *,"headtailsize_this_file    =",headtailsize_this_file
    !print *,"headtailsize_this_machine =",headtailsize_this_machine

    IF (.NOT.(lsizeonly)) THEN
       IF ( (headtailsize_this_file .eq. headtailsize_this_machine) .and. &
            .not. byteswap_needed() ) THEN
          ! to a direct unformatted read only if the head/tail size corresponds
          ! to what this platform/compiler combination uses, 
          ! and if no byteswapping is needed.
          call CloseDirectBinaryFile(iunit)
          iunit = get_lun()
          open (iunit,FILE=ficname,STATUS='OLD', &
                FORM='UNFORMATTED',ACCESS='SEQUENTIAL', ERR=999)
          read (iunit,err=998) nP_tmp,nT_tmp,nfr_tmp
          nP  = int(nP_tmp)
          nT  = int(nT_tmp)
          nfr = int(nfr_tmp)
          !print *,"nP,nT,nfr = ",nP,nT,nfr
          read (iunit,err=998) Pgrid
          read (iunit,err=998) Tgrid
          read (iunit,err=998) freqtmp
          read (iunit,err=998) spec
          close(iunit)
          call free_lun(iunit)
       ELSE
          ! in all other cases use the UnformattedIO module to take care
          ! of the different headtailsize and/or byteswapping and write a native
          ! unformatted file version when done
          
          ! so first check if this converted file already exists:
          inquire(file=alt_filename,exist=file_exists)
          IF (file_exists) THEN
             ! a converted file seems to exist, so do a regular
             ! unformatted read
             !print *,"Reading converted file: ",trim(alt_filename)
             call CloseDirectBinaryFile(iunit)
             iunit = get_lun()
             ficname = alt_filename
             open (iunit,FILE=ficname,STATUS='OLD', &
                  FORM='UNFORMATTED',ACCESS='SEQUENTIAL', ERR=999)
             read (iunit,err=998) nP,nT,nfr
             !print *,"nP,nT,nfr = ",nP,nT,nfr
             read (iunit,err=998) Pgrid
             read (iunit,err=998) Tgrid
             read (iunit,err=998) freqtmp
             read (iunit,err=998) spec
             close(iunit)
             call free_lun(iunit)
          ELSE
             !print *,"Reading file using UnformattedIO portability code"
             call read_unf_real8arr(iunit,headtailsize_this_file,&
                                    nP,Pgrid,error_flag)
             if (error_flag .ne. no_error) goto 998
             call read_unf_real8arr(iunit,headtailsize_this_file,&
                                    nT,Tgrid,error_flag)
             if (error_flag .ne. no_error) goto 998
             call read_unf_real8arr(iunit,headtailsize_this_file,&
                                    nfr,freqtmp,error_flag)
             if (error_flag .ne. no_error) goto 998

             ! WARNING: especially this call is VERY slow (38 seconds on my
             !          workstation) since it does 3469711 function calls 
             !          to read_real8_from_bin_file()
             ! However, I don't know any other portable way to do this,
             ! since reading from a directaccess file with recl=1 allways
             ! involves a large amount of looping for the reading and possibly
             ! the byteswapping.
             ! Therefore I let the code write the look-up-table in unformatted
             ! form again, so this is only needed once during installation
             ! of the software....
             call read_unf_real8arr3D(iunit,headtailsize_this_file,&
                                      nP,nT,nfr,spec,error_flag)
             if (error_flag .ne. no_error) goto 998
             call CloseDirectBinaryFile(iunit)
             
             ! write the table in native format for this platform/compiler
             ! combination:
             
             call WRITE_TENTISPECTRUM(nP,nT,nfr,Pgrid,Tgrid,freqtmp,spec,&
                                      error_flag,alt_filename)
          END IF
       END IF
    ELSE
       call CloseDirectBinaryFile(iunit)
    ENDIF

    ! restore the big-endian setting that we need for handling the
    ! EE_CFI formatted files
    call Init_DirectBinaryIO_module(error_flag, FileByteOrder=BigEndian)
    if (error_flag .ne. no_error) return
    
    return

998 print *,"ERROR in READ_TENTISPECTRUM:"
    print *,"could not read file: ",trim(ficname)
    error_flag = error_reading_file
    return

999 print *,"ERROR in READ_TENTISPECTRUM:"
    print *,"could not open file: ",trim(ficname)
    error_flag = error_opening_file
    return
    
  END SUBROUTINE READ_TENTISPECTRUM
  !  #]
  SUBROUTINE READ_TRANS(n,lsizeonly,error_flag,TA,TB,f,filename)
    !  #[
    USE Numerics, only: r8_
    USE LunManager, only: get_lun, free_lun

    implicit none
    
    !
    ! Reads the instrument Fabry Perot transfer 
    ! functions measured during calibration from ASCII file
    !
    ! 13/06/2005  Paul Poli, METEO FRANCE    Original code
    !
    
    ! I/O variables
    
    integer,           intent(inout)         :: n
    logical,           intent(in)            :: lsizeonly
    integer,           intent(out)           :: error_flag
    real(r8_),         intent(out), optional :: TA(n),TB(n),f(n)
    character(len=100),intent(in),  optional :: filename
    
    ! local variables
    
    integer            :: iunit
    character(len=100) :: ficname

    error_flag = no_error

    IF (present(filename)) THEN
       ficname=trim(filename)
    ELSE
       ficname='TransFPISR.dat'
    ENDIF

    !print *,"Reading file: ",trim(ficname)

     ! sanity check
    IF (.not. lsizeonly) THEN
       IF (.not. present(TA)) THEN
          print *,"ERROR: optional argument TA missing"
          goto 998
       END IF
       IF (.not. present(TB)) THEN
          print *,"ERROR: optional argument TB missing"
          goto 998
       END IF
       IF (.not. present(f)) THEN
          print *,"ERROR: optional argument f missing"
          goto 998
       END IF
    END IF
 
    iunit = get_lun()
    
    open (iunit,FILE=trim(ficname),STATUS='old',FORM='FORMATTED', ERR=999)
    read (iunit,*,err=998) n
    IF (.not.(lsizeonly)) THEN
       read (iunit,*,err=998) TA
       read (iunit,*,err=998) TB
       read (iunit,*,err=998) f
    ENDIF
    close (iunit)
    call free_lun(iunit)
    
    return
    
998 print *,"could not read file: ",trim(ficname)
    error_flag = error_reading_file
    return

999 print *,"could not open file: ",trim(ficname)
    error_flag = error_opening_file
    return
    
  END SUBROUTINE READ_TRANS
  !  #]
  SUBROUTINE WRITE_TRANSINTERP(nfreq,freq,TA,TB,error_flag,filename)
    !  #[
    USE Numerics, only: r8_
    USE LunManager, only: get_lun, free_lun

    implicit none

    !
    ! Writes out the interpolated Fabry Perot response in ASCII file
    !
    ! 15/06/2005  Paul Poli, METEO FRANCE    Original code
    !
    
    ! I/O variables
    
    integer,           intent(in)          :: nfreq
    real(r8_),         intent(in)          :: freq(nfreq),TA(nfreq),TB(nfreq)
    integer,           intent(out)         :: error_flag
    character(len=100),intent(in),optional :: filename
    
    ! local variables
    
    integer            :: iunit
    character(len=100) :: ficname
    
    iunit = get_lun()
    
    IF (present(filename)) THEN
       ficname=trim(filename)
    ELSE
       ficname='TransFPinterp.dat'
    ENDIF
    
    open (iunit,FILE=ficname,STATUS='NEW', &
         FORM='FORMATTED',err=999)
    write (iunit,*,err=998) nfreq
    write (iunit,*,err=998) TA
    write (iunit,*,err=998) TB
    write (iunit,*,err=998) freq
    close (iunit)
    call free_lun(iunit)

    return

998 print *,"could not write file: ",trim(ficname)
    error_flag = error_writing_file
    return

999 print *,"could not open file: ",trim(ficname)
    error_flag = error_opening_file
    return
    
  END SUBROUTINE WRITE_TRANSINTERP
  !  #]
  SUBROUTINE READ_GRID(nP,nT,nRR,nfd,nfreq,lsizeonly, error_flag, &
                       Pgrid,Tgrid,RR0,fd,freq,Fcalib,&
                       Na_F,Nb_F,Fint_R,filename)
    !  #[
    USE Numerics, only: r8_
    USE LunManager, only: get_lun, free_lun

    implicit none
    
    !
    ! Reads the calibration matrix from binary file
    !
    ! 16/06/2005  Paul Poli, METEO FRANCE    Original code
    !
    
    ! I/O variables
    
    integer,           intent(inout)         :: nP,nT,nRR,nfd,nfreq
    logical,           intent(in)            :: lsizeonly
    integer,           intent(out)           :: error_flag
    real(r8_),         intent(out), optional :: Pgrid(nP),Tgrid(nT),RR0(nRR), &
                                                fd(nfd),freq(nfreq), &
                                                Fcalib(nP,nT,nRR), &
                                                Na_F(nP,nT,nfd), &
                                                Nb_F(nP,nT,nfd), &
                                                Fint_R(nRR)
    character(len=100),intent(in),  optional :: filename
    
    ! local variables
    
    integer            :: iunit, headtailsize_this_file
    character(len=100) :: ficname
    character(len=100) :: alt_filename
    logical            :: file_exists
    integer(i4_), dimension(5) :: arr

    error_flag = no_error

    IF (present(filename)) THEN
       ficname=trim(filename)
    ELSE
       ficname='CALIB_GRID.dat'
    ENDIF
    
    !print *,"Reading file: ",trim(ficname)
    alt_filename = trim(ficname)//".alt"
 
    ! sanity check
    IF (.not. lsizeonly) THEN
       !print *,"Reading complete file"
       IF (.not. present(Pgrid)) THEN
          print *,"ERROR: optional argument Pgrid missing"
          goto 998
       END IF
       IF (.not. present(Tgrid)) THEN
          print *,"ERROR: optional argument Tgrid missing"
          goto 998
       END IF
       IF (.not. present(RR0)) THEN
          print *,"ERROR: optional argument RR0 missing"
          goto 998
       END IF
       IF (.not. present(fd)) THEN
          print *,"ERROR: optional argument fd missing"
          goto 998
       END IF
       IF (.not. present(freq)) THEN
          print *,"ERROR: optional argument freq missing"
          goto 998
       END IF
       IF (.not. present(Fcalib)) THEN
          print *,"ERROR: optional argument Fcalib missing"
          goto 998
       END IF
       IF (.not. present(Na_F)) THEN
          print *,"ERROR: optional argument Na_F missing"
          goto 998
       END IF
       IF (.not. present(Nb_F)) THEN
          print *,"ERROR: optional argument Nb_F missing"
          goto 998
       END IF
       IF (.not. present(Fint_R)) THEN
          print *,"ERROR: optional argument Fint_R missing"
          goto 998
       END IF
    !ELSE
    !   print *,"Reading dimensions only" 
    END IF

    ! remark: a normal unformatted read fails for the gfortran compiler
    ! because it uses 8 byte head/tail marks in stead of the usual
    ! 4 byte head/tail marks that most other compilers on 32-bit platforms use.
    ! This seems to be a deliberate choice, to enable use of
    ! multi gigabyte datafiles ... So we have to make a workaround here
    ! untill the RBC-lookuptable format is converted to ee_cfi format.

    call Init_DirectBinaryIO_module(error_flag, FileByteOrder=LittleEndian)
    if (error_flag .ne. no_error) return
  
    call OpenDirectBinaryFile(ficname,"read",iunit,error_flag)
    if (error_flag .ne. no_error) goto 999

    !print *,"trying headtailsize_this_file=4"
    headtailsize_this_file=4
    call read_unf_int4arr(iunit,headtailsize_this_file,5,arr,error_flag)
    IF (error_flag .ne. no_error) THEN
       error_flag = no_error
       !print *,"trying headtailsize_this_file=8, since 4 seems wrong ..."
       ! headtailsize of 4 might be wrong, so try again with 8
       ! the filepos is shifted by 4 bytes now, so reopen the file
       ! to correct for this.
       call CloseDirectBinaryFile(iunit)
       call OpenDirectBinaryFile(filename,"read",iunit,error_flag)
       if (error_flag .ne. no_error) goto 999
       headtailsize_this_file=8
       call read_unf_int4arr(iunit,headtailsize_this_file,5,arr,error_flag)
       if (error_flag .ne. no_error) goto 998
    END IF

    nP    = arr(1)
    nT    = arr(2)
    nRR   = arr(3)
    nfd   = arr(4)
    nfreq = arr(5)

    !print *,"nP,nT,nRR,nfd,nfreq = ",nP,nT,nRR,nfd,nfreq
    !print *,"headtailsize_this_file    =",headtailsize_this_file
    !print *,"headtailsize_this_machine =",headtailsize_this_machine

    IF (.NOT.(lsizeonly)) THEN
       IF ( (headtailsize_this_file .eq. headtailsize_this_machine) .and. &
            .not. byteswap_needed() ) THEN
          ! to a direct unformatted read only if the head/tail size corresponds
          ! to what this platform/compiler combination uses, 
          ! and if no byteswapping is needed.
          call CloseDirectBinaryFile(iunit)
          iunit = get_lun()
          open (iunit,FILE=trim(ficname),STATUS='old',FORM='UNFORMATTED',ERR=999)
          read (iunit,err=998) nP,nT,nRR,nfd,nfreq
          !print *,"nP,nT,nRR,nfd,nfreq = ",nP,nT,nRR,nfd,nfreq
          read (iunit,err=998) Pgrid
          read (iunit,err=998) Tgrid
          read (iunit,err=998) RR0
          read (iunit,err=998) fd
          read (iunit,err=998) freq
          read (iunit,err=998) Fcalib
          read (iunit,err=998) Na_F
          read (iunit,err=998) Nb_F
          read (iunit,err=998) Fint_R
          close (iunit)
          call free_lun(iunit)
       ELSE
          ! in all other cases use the UnformattedIO module to take care
          ! of the different headtailsize and/or byteswapping and write a native
          ! unformatted file version when done

          ! so first check if this converted file already exists:
          inquire(file=alt_filename,exist=file_exists)
          IF (file_exists) THEN
             ! a converted file seems to exist, so do a regular
             ! unformatted read
             !print *,"Reading converted file: ",trim(alt_filename)
             call CloseDirectBinaryFile(iunit)
             iunit = get_lun()
             open (iunit,FILE=trim(alt_filename),STATUS='old',&
                  FORM='UNFORMATTED',ERR=999)
             read (iunit,err=998) nP,nT,nRR,nfd,nfreq
             !print *,"nP,nT,nRR,nfd,nfreq = ",nP,nT,nRR,nfd,nfreq
             read (iunit,err=998) Pgrid
             read (iunit,err=998) Tgrid
             read (iunit,err=998) RR0
             read (iunit,err=998) fd
             read (iunit,err=998) freq
             read (iunit,err=998) Fcalib
             read (iunit,err=998) Na_F
             read (iunit,err=998) Nb_F
             read (iunit,err=998) Fint_R
             close (iunit)
             call free_lun(iunit)
          ELSE
             !print *,"Reading file using UnformattedIO portability code"
             call read_unf_real8arr(iunit,headtailsize_this_file,&
                                    nP,Pgrid,error_flag)
             if (error_flag .ne. no_error) goto 998
             call read_unf_real8arr(iunit,headtailsize_this_file,&
                                    nT,Tgrid,error_flag)
             if (error_flag .ne. no_error) goto 998
             call read_unf_real8arr(iunit,headtailsize_this_file,&
                                    nRR,RR0,error_flag)
             if (error_flag .ne. no_error) goto 998
             call read_unf_real8arr(iunit,headtailsize_this_file,&
                                    nfd,fd,error_flag)
             if (error_flag .ne. no_error) goto 998
             call read_unf_real8arr(iunit,headtailsize_this_file,&
                                    nfreq,freq,error_flag)
             if (error_flag .ne. no_error) goto 998

             ! WARNING: especially this call is VERY slow (38 seconds on my
             !          workstation) since it does 3469711 function calls 
             !          to read_real8_from_bin_file()
             ! However, I don't know any other portable way to do this,
             ! since reading from a directaccess file with recl=1 allways
             ! involves a large amount of looping for the reading and possibly
             ! the byteswapping.
             ! Therefore I let the code write the look-up-table in unformatted
             ! form again, so this is only needed once during installation
             ! of the software....
             call read_unf_real8arr3D(iunit,headtailsize_this_file,&
                                      nP,nT,nRR,Fcalib,error_flag)
             if (error_flag .ne. no_error) goto 998
             call read_unf_real8arr3D(iunit,headtailsize_this_file,&
                                      nP,nT,nfd,Na_F,error_flag)
             if (error_flag .ne. no_error) goto 998
             call read_unf_real8arr3D(iunit,headtailsize_this_file,&
                                      nP,nT,nfd,Nb_F,error_flag)
             if (error_flag .ne. no_error) goto 998
             call read_unf_real8arr(iunit,headtailsize_this_file,&
                                    nRR,Fint_R,error_flag)
             if (error_flag .ne. no_error) goto 998
             call CloseDirectBinaryFile(iunit)
             
             ! write the table in native format for this platform/compiler
             ! combination:
             call WRITE_GRID(nP,nT,nRR,nfd,nfreq, &
                  Pgrid,Tgrid,RR0,fd,freq,Fcalib,&
                  Na_F,Nb_F,Fint_R,error_flag,alt_filename)

          END IF
       END IF
    ELSE
       call CloseDirectBinaryFile(iunit)
    ENDIF

    ! restore the big-endian setting that we need for handling the
    ! EE_CFI formatted files
    call Init_DirectBinaryIO_module(error_flag, FileByteOrder=BigEndian)
    if (error_flag .ne. no_error) return

    return
    
998 print *,"ERROR in READ_GRID"
    print *,"could not read file: ",trim(ficname)
    error_flag = error_reading_file
    return

999 print *,"ERROR in READ_GRID"
    print *,"could not open file: ",trim(ficname)
    error_flag = error_opening_file
    return
    
  END SUBROUTINE READ_GRID  
    !  #]
  SUBROUTINE WRITE_GRID(nP,nT,nRR,nfd,nfreq, &
                        Pgrid,Tgrid,RR0,fd,freq,Fcalib, &
                        Na_F,Nb_F,Fint_R,error_flag,filename)
    !  #[
    USE Numerics, only: r8_
    USE LunManager, only: get_lun, free_lun
    
    implicit none
    
    !
    ! Writes out the calibration grid matrix in binary file
    !
    ! 15/06/2005  Paul Poli, METEO FRANCE    Original code
    !
    
    ! I/O variables
    
    integer,           intent(in)          :: nP,nT,nRR,nfd,nfreq
    real(r8_),         intent(in)          :: Pgrid(nP),Tgrid(nT),RR0(nRR), &
                                              fd(nfd),freq(nfreq), &
                                              Fcalib(nP,nT,nRR), &
                                              Na_F(nP,nT,nfd), &
                                              Nb_F(nP,nT,nfd), &
                                              Fint_R(nRR)
    integer,           intent(out)         :: error_flag
    character(len=100),intent(in),optional :: filename
    
    ! local variables
    
    integer            :: iunit
    character(len=100) :: ficname
    
    error_flag = no_error

    iunit = get_lun()
    
    IF (present(filename)) THEN
       ficname=trim(filename)
    ELSE
       ficname='CALIB_GRID.dat'
    ENDIF
    
    open (iunit,FILE=ficname,STATUS='NEW', &
         FORM='UNFORMATTED',err=999)
    write (iunit,err=998) nP,nT,nRR,nfd,nfreq
    write (iunit,err=998) Pgrid
    write (iunit,err=998) Tgrid
    write (iunit,err=998) RR0
    write (iunit,err=998) fd
    write (iunit,err=998) freq
    write (iunit,err=998) Fcalib
    write (iunit,err=998) Na_F
    write (iunit,err=998) Nb_F
    write (iunit,err=998) Fint_R
    close (iunit)
    call free_lun(iunit)
    
    return

998 print *,"could not write file: ",trim(ficname)
    error_flag = error_writing_file
    return

999 print *,"could not open file: ",trim(ficname)
    error_flag = error_opening_file
    return
    
  END SUBROUTINE WRITE_GRID
  !  #]
  !-----------------------------------
end module TentiSpectrum
