#!/usr/bin/env python

#  #[ documentation
# a little script to automatically generate the WhoCallsWho list
# 
# written by: J. de Kloe, KNMI
# modifications:
#   09-Jan-2008 original version (not yet in working order)
#   21-Jan-2008 produces correct dependencies for directories now
#               but nice textual output is still to be implemented
#
#  #]
#  #[ known limitations
# line continuation characters "&" between for example
# "use" and "<modulename>" or between "module" and "<modulename>"
# are not yet taken into account
#  #]
#  #[ imported modules
import os, string
#  #]
#  #[ some constants used by the code
StartDir = "./"
WhoCallsWhoFile = "whocallswho_generated.txt"
FileType    = 1
DirType     = 2
NotFound    = 3
SymLinkType = 4
CodeExtensions = [".F90",".c"]
AllowedDoubles = ["compiler_features","latlon_handling"]
IngoredModules = ["f90_unix_env"]
ExcludedDirs   = ["simple_xml",
                  "ee_cfi_wrapper_module",
                  "ee_cfi_software_3.7.1",
                  "ee_cfi_software",
                  "src_ecmwf"]
# note: support/compiler_features_f95.F90 uses a use of the F90_UNIX_ENV
# to implement iargc and getarg. This is a specific Fortran90 extension
# by the f95 compiler and is not to be included in the dependency resolution
#  #]
#-------------
# subroutines/functions
#-------------
def GetObjectType(Object):
    #  #[
    ObjectType = NotFound
    if (os.path.exists(Object)):
        if (os.path.islink(Object)):
            ObjectType = SymLinkType
        elif (os.path.isdir(Object)):
            ObjectType = DirType
        elif (os.path.isfile(Object)):
            ObjectType = FileType

    return ObjectType
    #  #]                    
def ConstructDirList(Dir):
    #  #[
    FileList = os.listdir(Dir)
    DirList=[]
    for File in FileList:
        if (not File in ExcludedDirs):
            # FullName = os.path.abspath(os.path.join(Dir,File))
            FullName = os.path.join(Dir,File)
            if (GetObjectType(FullName) == DirType):
                DirList.append(FullName)
            if (GetObjectType(FullName) == SymLinkType):
                # follow the symlink
                RealName = os.path.realpath(FullName)
                if (GetObjectType(RealName) == DirType):
                    # print "Dir symlink: "+FullName+" points to: "+RealName
                    # remove the common part and replace by a relative
                    # directory if possible
                    name1 = os.path.abspath(FullName)
                    name2 = os.path.abspath(RealName)
                    prefix = os.path.commonprefix([name1,name2])
                    # print "name1 = "+name1
                    # print "name2 = "+name2
                    (dir1,file1) = os.path.split(name1)
                    (dir2,file2) = os.path.split(name2)
                    # print "prefix = "+prefix
                    if ( (dir1 == prefix) and (dir2 == prefix)):
                        name2 = "./"+file2
                        DirList.append(name2)
                    else:
                        DirList.append(RealName)
    return DirList
    #  #]
def GetSourceFiles(Dir):
    #  #[
    SourceFiles = []
    FileList = os.listdir(Dir)
    for File in FileList:
        FullName = os.path.abspath(os.path.join(Dir,File))
        if (GetObjectType(FullName) == FileType):
            Ext = os.path.splitext(File)
            if (Ext[1] in CodeExtensions):
                #print "Ext: "+Ext[1]+" file: "+File
                SourceFiles.append(File)
    return SourceFiles
    #  #]
def RemoveDoubleModules(Modules):
    #  #[
    UniqueModules = []
    for module in Modules:
        if (not module in UniqueModules):
            UniqueModules.append(module)
    return UniqueModules
    #  #]
def FileGetProvidedModules(Dir,File):
    #  #[
    ProvidedModules = []
    FullName = os.path.join(Dir,File)
    # parse the file
    for Line in file(FullName):
        # if the first 7 non-space characters of the line are "module "
        # or "MODULE " then we found a new module definition
        if (string.lower(string.lstrip(Line)[0:7]) == "module "):
            # do this is 2 steps because of multiple spaces are present
            # the split function will generate words from single spaces
            # (this is the reason I use maxsplit=1 here)
            parts1 = string.split(string.lstrip(Line),sep=' ',maxsplit=1)
            parts2 = string.split(string.lstrip(parts1[1]),sep=' ',maxsplit=1)
            ModuleName = string.lower(string.strip(parts2[0]))

            # note that the sequence "module procedure" may occur inside
            # an interface block. We have to skip these since these
            # are not module definitions
            if (ModuleName != "procedure"):
                #print "module name seems to be: ",ModuleName
                if (not ModuleName in ProvidedModules):
                    ProvidedModules.append(ModuleName)
                else:
                    print "WARNING: parsing problem for line:"
                    print "Line: ["+string.strip(Line)+"]"
                    print "MODULE "+ModuleName+\
                          " seems defined more then once ..."

    return ProvidedModules
    #  #]
def FileGetUsedModules(Dir,File):
    #  #[
    UsedModules = []
    FullName = os.path.join(Dir,File)
    # parse the file
    for Line in file(FullName):
        # if the first 3 non-space characters of the line are "use "
        # or "USE " then we found a USE statement
        if (string.lower(string.lstrip(Line)[0:4]) == "use "):
            # get the first word after the use statement
            # (this is the module name)
            # use either ',' or ' ' as delimiter
            split1 = string.split(string.lstrip(Line),sep=' ')
            used_name = ""
            for part in split1:
                if (used_name == ""):
                    if (string.lower(part) != 'use'):
                        used_name = string.lower(part)
            if (used_name == ""):
                print "WARNING: parsing problem for line:"
                print "Line: ["+string.strip(Line)+"]"
                print "No name found after USE statement ..."

            # note that the name still may have a comma attached to it,
            # so remove it
            split2 = string.split(used_name,",")
            used_name = split2[0]
            # also note that if no "only" keyword was used, a leading
            # newline may be present, so remove that as well
            used_name = used_name.replace("\n","")
            #print "used_name: ["+used_name+"]"

            if (not used_name in IngoredModules):
                UsedModules.append(used_name)

    return UsedModules
    #  #]
def DirGetUsedModules(Dir):
    #  #[
    DirUsedModules = []
    SourceFiles = GetSourceFiles(Dir)
    #print "SourceFiles to inspect:"
    for SFile in SourceFiles:
        #print "==>"+SFile
        UsedModules     = FileGetUsedModules(Dir,SFile)
        UsedModules     = RemoveDoubleModules(UsedModules)
        for module in UsedModules:
            DirUsedModules.append(module)
    # note that doubles may occur, so remove them before
    # reporting the result
    DirUsedModules = RemoveDoubleModules(DirUsedModules)
    return DirUsedModules
    #  #]
def DirGetProvidedModules(Dir):
    #  #[
    DirProvidedModules = []
    SourceFiles = GetSourceFiles(Dir)
    #print "SourceFiles to inspect:"
    for SFile in SourceFiles:
        #print "==>"+SFile
        ProvidedModules = FileGetProvidedModules(Dir,SFile)
        for ModuleName in ProvidedModules:
            if (not ModuleName in DirProvidedModules):
                DirProvidedModules.append(ModuleName)
            else:
                if (not ModuleName in AllowedDoubles):
                    print "WARNING:"
                    print "MODULE "+ModuleName
                    print "seems to be defined more then once"
                    print "in directory: "+Dir+" !!!"
                #else:
                #    print "found an allowed double: "+ModuleName

    # note that doubles should not occur in this list!
    return DirProvidedModules
    #  #]
def GetExtModulesNeeded(DirUsedModules,DirProvidedModules):
    #  #[
    # check all needed modules in this directory, and
    # If they are not provided by any of the files inside this
    # directory, add them to the list of ExternalModulesNeeded
    ExternalModulesNeeded = []
    for module in DirUsedModules:
        if (not module in DirProvidedModules):
            ExternalModulesNeeded.append(module)
    return ExternalModulesNeeded
    #  #]
#-------------
# main program
#-------------
# construct the list of directories to be inspected

DirInfo = []
#DirList=["./support","./simple_xml"]
DirList = ConstructDirList(StartDir)

print "Directories to inspect:"
for Dir in DirList:
    #print "==>"+Dir
    DirUsedModules = DirGetUsedModules(Dir)
    #print "DirUsedModules = ",DirUsedModules
    DirProvidedModules = DirGetProvidedModules(Dir)
    #print "DirProvidedModules = ",DirProvidedModules
    ExternalModulesNeeded = GetExtModulesNeeded(DirUsedModules,
                                                DirProvidedModules)
    #print "ExternalModulesNeeded = ",ExternalModulesNeeded
    # store the results for use in the next loop
    if (len(DirUsedModules)        != 0 or
        len(DirProvidedModules)    != 0 or
        len(ExternalModulesNeeded) != 0   ):
        DirInfo.append((Dir,
                        DirUsedModules,
                        DirProvidedModules,
                        ExternalModulesNeeded))
    else:
        FileList = GetSourceFiles(Dir)
        if (len(FileList) == 0):
            print "skipping dir: "+Dir+\
                  " [it seems not to contains any F90 code]"
        else:
            print "programming error during skipping of dirs ..."
            os._exit(1)

for DirStruct in DirInfo:
    DirName               = DirStruct[0]
    DirUsedModules        = DirStruct[1]
    DirProvidedModules    = DirStruct[2]
    ExternalModulesNeeded = DirStruct[3]
    print "Dir: "+DirName
    #print "==>provides: "
    #print "".join("            [%s]\n" % (m) for m in DirProvidedModules)
    #print "==>external uses: "
    DirDependencies = []
    for UsedModule in ExternalModulesNeeded:
        #  #[ find the directory in which it is defined
        DefDir = ""
        for ExtDirStruct in DirInfo:
            ExtDirName            = ExtDirStruct[0]
            ExtDirProvidedModules = ExtDirStruct[2]
            if (UsedModule in ExtDirProvidedModules):
                if (DefDir == ""):
                    DefDir = ExtDirName
                else:
                    print "ERROR: UsedModule seems to be defined in 2 "+\
                          "different modules !"
                    print "ModuleName: "+UsedModule
                    print "1: "+DefDir
                    print "2: "+ExtDirName
                
        # print the result
        # print "            [%s] :defined in: [%s]" % (UsedModule,DefDir)
        if (not DefDir in DirDependencies):
            DirDependencies.append(DefDir)
        #  #]
    print "DirDependencies: ",DirDependencies
    
#Dir      = "./support"
#TestFile = "interp1.F90"
#ProvidedModules = FileGetProvidedModules(Dir,TestFile)
#print "ProvidedModules",ProvidedModules

os._exit(1)

# write the output file
fd = open(WhoCallsWhoFile,'a')
fd.write("---------------------------------------------------\n")
fd.write("This file describes for the code in each directory,\n"+
         "on which other directory the code depends\n")
fd.write("This file was auto-generated by the tool: \n")
fd.write("==>./Scripts/GenerateWhoCallsWhoList.py\n")
fd.write("---------------------------------------------------\n")
fd.write("\n\n")
fd.write("main\n")
fd.close()
print "done"



