#!/usr/bin/env python
# if the above line does not work for you, try replacing it with:
#!/usr/bin/python
#  #[ documentation

#  a little python report generation script
#
# Written by: J. de Kloe, KNMI
#
# Modifications:
#   03-Oct-2006 J. de Kloe   Initial version
#   25-Oct-2006 J. de Kloe   corrected the regular expression to work properly
#                            even if the text contains pairs of [] chars.
#   25-Oct-2006 J. de Kloe   changed the File_Name entry in the report to
#                            hold the name of the report rather than the name
#                            of the log file. Also added some extra
#                            checks and comments.
#   14-Nov-2006 J. de Kloe   reorganised to do the actual work outside the
#                            ProcList loop, and added the commandline option
#                            '-l' to specify a logfile directly.
#   31-Jan-2007 D. Huber     addition of FixedHeader to the XML output 
#   07-Mar-2007 J. de Kloe   changed to construct needed EE headers, in stead
#                            of copying the lines from another EE file.
#   09-Mar-2007 D. Huber     -deleted the unit='UTC' part for date times.
#                            -generated an empty text node for the
#                            Variable_Header tag and now it prints
#                            the EE compliant format.
#                            -generated a new name for the L2A product
#                            report file.
#   16-Mar-2007 J. de Kloe   added a proper error in case JobOrder fileformat
#                            is not valid xml. Changed datetime handling to
#                            use GMT time. Fixed a small bug for the -l option.
#                            Set ValidityStart/Stop to the special datetime
#                            codes for begin/end mission.
#   02-Apr-2007 J. de Kloe   changed naming convention for L2B report files
#   10-Apr-2007 J. de Kloe   added some python trickery to redefine the
#                            writexml functions in xml.dom.minidom, which is
#                            usefull on platforms which have no xml.dom.ext
#                            package installed ...
#   02-May-2007 J. de Kloe   added some logic to circumvent use of the
#                            minicompat module when it is not installed
#   03-May-2007 J. de Kloe   added a few dummy init routines in an attempt to
#                            solve Finn's problems.
#   04-May-2007 J. de Kloe   changed filetype from RP_PRO_L2A/B to ALD_REP_2A/B
#   15-May-2007 J. de Kloe   added an inspect.getargspec() call to prevent the
#                            trouble reported by Dorit when using toprettyxml()
#   24-Jun-2007 J. de Kloe   disabled PrettyPrint, and updated the custom 
#                            to_xml routines to produce similar output.
#                            Then added a workaround to circumvent the huge
#                            memory usage of this script for large log files.
#   27-Jun-2007 J. de Kloe   removed use of the inspect module to prevent
#                            portability problems on Finns machine
#   12-Nov-2007 J. de Kloe   bugfix. The script crashed when called with a
#                            relative path in front of the JobOrderFilename
#                            like: ./L2B_ReportGenerator.py ./JobOrder.1.xml
#                            This is taken into account now.
#   20-Nov-2007 J. de Kloe   get the L2B/L2A product filename from the LIST
#                            file, and use this name to construct the report
#                            filename, and fill the validity start/stop times
#   27-Nov-2007 J. de Kloe   added check for the existence of the file
#                            ProductListFileName before trying to read it.
#   17-Dec-2007 J. de Kloe   Implemented File_Class handling and introduced
#                            new RaiseError function
#   19-Feb-2008 J. de Kloe   added namespace attribute in document root
#                            adressed problems reported in L2B SPR 28
#                            moved a lot of code to smaller subroutines
#                            to make it easier to maintain
#   20-Feb-2008 J. de Kloe   removed use of the minicompat module since it
#                            is not necessary, and gives problems for Finn.
#                            added consistency check for TaskVersion
#   26-Feb-2008 J. de Kloe   changed xmlns reference to point to the
#                            Thinlayer Generic Interface Guidelines doc
#
# General remark:
#         the "#  #[" and "#  #]" marks are used by the folding feature of the
#         emacs editor that I use. If anyone changes this document please
#         leave them in place. If you are not already using emacs 
#         you might want to consider starting to use it now, because of
#         this (extremely usefull) folding option (and many other reasons).
#         If you wonder how to install/activate this feature, feel free
#         to contact me.                   Jos de Kloe, 12-Nov-2007
#

#  #]
#  #[ imported modules

import sys                  # system functions
import os                   # operating system specific functions
import re                   # regular expression functions
import time                 # access to current date and time
import math                 # access to math functions like ceil()
from xml.dom import minidom # xml document object model functions

#use_minicompat = True
#try:
#    from xml.dom.minicompat import * # defines: StringTypes
#except:
#    # the xml.dom extention xml.dom.minicompat seems not installed on this
#    # machine so set a switch to signal that the software should disable
#    # a few tests using this package. This should not change the functionality
#    # for correct inputs ....
#    use_minicompat = False
#
## for testing, force this to be false
#use_minicompat = False
    
use_xml_dom_ext = True
try:
    import xml.dom.ext      # extended xml document object model functions
                            # used to get nice indentation of the xml file
except:
    # the xml.dom extention xml.dom.ext seems not installed on this machine
    # (this is part of the pyXml package, which extends the default xml
    #  package for python)
    # so set a switch to signal that the software should fallback to
    # using the default xml.dom package ....
    use_xml_dom_ext = False

# for testing, force this to be false
use_xml_dom_ext=False

#  #]
#  #[ a custom error for use in this script
class RepGenError(Exception): pass
#  #]
#---------- classes and functions -------------
#  #[ some python hockus-pockus:

#     redefine some functions inside xml.dom.minidom to avoid use of the
#     xml.dom.ext module which seems not available on all platforms
#     (for now, do this only in case it is not available)

# subclass Element, to enable changing the writexml function
class MyElement(minidom.Element):
    #  #[
    # the minidom.Element class defines its own init routine, so use that one
    def writexml(self, writer, indent="", addindent="", newl=""):
        # indent = current indentation
        # addindent = indentation to add to higher levels
        # newl = newline string
        writer.write(indent+"<" + self.tagName)

        attrs = self._get_attributes()
        a_names = attrs.keys()
        a_names.sort()

        for a_name in a_names:
            # write the attributes
            writer.write(" %s='" % a_name)
            minidom._write_data(writer, attrs[a_name].value)
            writer.write("'")

        if self.childNodes:
            if (self.childNodes[0].nodeType == minidom.Node.TEXT_NODE):
                writer.write(">") # no newline in this case!
            else:
                writer.write(">%s"%(newl))

            #if (self.childNodes[0].nodeType == minidom.Node.CDATA_SECTION_NODE):
            #    print "test van Jos: cdata section found !"
            #    print "indent    = ["+indent+"]"
            #    print "addindent = ["+addindent+"]"
            #    print "newl      = ["+newl+"]"
            #    os._exit(1)

            for node in self.childNodes:
                node.writexml(writer,indent+addindent,addindent,newl)

            if (self.childNodes[0].nodeType == minidom.Node.TEXT_NODE):
                writer.write("</%s>%s" % (self.tagName,newl)) # no indentation
            else:
                writer.write("%s</%s>%s" % (indent,self.tagName,newl))
        else:
            # for empty elements, use ESA's preferred style <tag></tag>
            #writer.write("></%s>%s"%(self.tagName,newl))
            # for now, replace this by <tag>\n[indent]</tag>
            # to not break any existing tests in the test suite
            writer.write(">%s"%(newl))
            writer.write("%s</%s>%s"%(indent,self.tagName,newl))
    #  #]
    
# subclass Text, to enable changing the toprettyxml function
class MyText(minidom.Text):
    #  #[
    # the minidom.Text class seems not to define an init routine, which gives
    # trouble when Finn tries to use this program, so define a dummy one
    def __init__(self):
        pass
    def writexml(self, writer, indent="", addindent="", newl=""):
        # my version has no indent, and no newl
        minidom._write_data(writer, "%s"%(self.data))
        # original was:
        #_write_data(writer, "%s%s%s"%(indent, self.data, newl))
    #  #]

# subclass CDATASection,  to enable changing the writexml function
class MyCDATASection(minidom.CDATASection):
    #  #[
    # the minidom.CDATASection class is derived from the minidom.Text class
    # which seems not to define an init routine, which gives trouble
    # when Finn tries to use this program, so define a dummy one
    def __init__(self):
        pass
    def writexml(self, writer, indent="", addindent="", newl=""):
        if self.data.find("]]>") >= 0:
            raise ValueError("']]>' not allowed in a CDATA section")
        # this version uses indentation and newlines (the original doesn't)
        writer.write("%s<![CDATA[%s]]>%s" % (indent,self.data,newl))
    #  #]
    
# subclass Document, to enable changing the writexml function
class MyMinidomDocument(minidom.Document):
    #  #[ 

    def __init__(self):
        # init the new class by calling init for the parent class
        minidom.Document.__init__(self)
    def createTextNode(self, data):
        # override this function, into using an object of
        # the MyText class, rather than the Text class
        #if (use_minicompat):
        #    if not isinstance(data, StringTypes):
        #        raise TypeError, "node contents must be a string"
        t = MyText()
        t.data = data
        t.ownerDocument = self
        return t
    def createElement(self, tagName):
        # override this function, into using an object of
        # the MyElement class, rather than the Element class
        e = MyElement(tagName)
        e.ownerDocument = self
        return e
    def createCDATASection(self, data):
        # override this function, into using an object of
        # the MyCDATASection class, rather than the CDATASection class        
        #if (use_minicompat):
        #    if not isinstance(data, StringTypes):
        #        raise TypeError, "node contents must be a string"
        c = MyCDATASection()
        c.data = data
        c.ownerDocument = self
        return c
    def writexml(self, writer, indent="", addindent="", newl="",
                 encoding = None):
        # override this function, into using single quotes around the
        # attributes in stead of double quotes, which makes the result
        # more similar to the result of xml.dom.ext.PrettyPrint()

        #if encoding is None:
        #    writer.write("<?xml version='1.0' ?>\n")
        #else:
        #    writer.write("<?xml version='1.0' encoding='%s'?>\n" % encoding)

        # ignore the encoding settings, and just always use 'UTF-8'
        # because this encoding causes compatibility problems when
        # used with python 2.2
        writer.write("<?xml version='1.0' encoding='UTF-8'?>\n")
        for node in self.childNodes:
            node.writexml(writer, indent, addindent, newl)

    #  #]

#  #]
def PrintUsage():
    #  #[ usage function
    ScriptPathAndName = sys.argv[0]
    (ScriptPath,ScriptName) = os.path.split(ScriptPathAndName)
    print "nr. of commandline arguments differs from what was expected"
    print "usage:"
    print "  "+ScriptName+" JobOrder.xml"
    print "or for manual testing:"
    print "  "+ScriptName+" -l LogFile"
    #  #]
def RaiseError(msg):
    #  #[
    ScriptPathAndName = sys.argv[0]
    (ScriptPath,ScriptName) = os.path.split(ScriptPathAndName)
    # print "My name is: ",ScriptName
    print "ERROR in "+ScriptName+":\n"+msg+"\n"+\
          "Aborting ReportGenerator..."
    raise RepGenError
    #  #]
def GetUndefinedDateTime():
    #  #[
    # fill with missing (undefined strings)
    Year   = "[undefined]"
    Month  = "[undefined]"
    Day    = "[undefined]"
    Hour   = "[undefined]"
    Minute = "[undefined]"
    Second = "[undefined]"
    DateTime = (Year, Month, Day, Hour, Minute, Second)
    return DateTime
    #  #]
def GetCurrentDateTime():
    #  #[
    # get the current gmt date-time value
    CurrentDateTimeValues = time.gmtime()
    CurrentYear   = "%4.4i" % CurrentDateTimeValues[0]
    CurrentMonth  = "%2.2i" % CurrentDateTimeValues[1]
    CurrentDay    = "%2.2i" % CurrentDateTimeValues[2]
    CurrentHour   = "%2.2i" % CurrentDateTimeValues[3]
    CurrentMinute = "%2.2i" % CurrentDateTimeValues[4]
    CurrentSecond = "%2.2i" % CurrentDateTimeValues[5]
    CurrentDateTime = (CurrentYear, CurrentMonth,  CurrentDay,
                       CurrentHour, CurrentMinute, CurrentSecond)
    return CurrentDateTime
    #  #]
def DecodeStartStopDateTime(StartStopDateTimeStr):
    #  #[
    # assume DateTime has a format like:
    # 20071002T000002_20071002T000040 (at least for L2B this is the case)
    # 0123456789012345678901234567890
    # 0         1         2         3
    # Its format is fairly simple:
    #
    # YYYYMMDD.HHMMSS.YYYYMMDD.HHMMSS
    # (validity start/(validity stop)
    #
    # L2A (as well as L1B) use a datetime field like so:
    # 20071010T080005036_005376000
    # 0123456789012345678901234567
    # 0         1         2       
    # This gives the start datetime, the duration in ms and an orbit number:
    # Its format is:
    #
    # YYYYMMDD.HHMMSSzzz.uuuuuuuuu
    # (validity start    duration in ms)
    #
    # in which zzz are the milliseconds of the ValidityStart time 
    # and uuuuuuuuu is the duration in milliseconds.
    #
    StartYear   = StartStopDateTimeStr[0:4]
    StartMonth  = StartStopDateTimeStr[4:6]
    StartDay    = StartStopDateTimeStr[6:8]
    StartHour   = StartStopDateTimeStr[9:11]
    StartMinute = StartStopDateTimeStr[11:13]
    StartSecond = StartStopDateTimeStr[13:15]
    if (StartStopDateTimeStr[15:16] == "_"):
        # ok, this is the L2B datetime format
        StopYear    = StartStopDateTimeStr[16:20]
        StopMonth   = StartStopDateTimeStr[20:22]
        StopDay     = StartStopDateTimeStr[22:24]
        StopHour    = StartStopDateTimeStr[25:27]
        StopMinute  = StartStopDateTimeStr[27:29]
        StopSecond  = StartStopDateTimeStr[29:31]
    else:
        # ok, this is the L2A/L1B datetime format
        StartMilliSeconds = StartStopDateTimeStr[15:18]
        Duration          = StartStopDateTimeStr[19:28]
        DateTimeFormat    = "%Y%m%dT%H%M%S"
        StartDateTimeString    = StartStopDateTimeStr[0:15]
        # convert the StartDateTime to pythons internal
        # date-time structure (tuple)
        StartDateTimeTuple = time.strptime(StartDateTimeString,DateTimeFormat)
        # convert to seconds since the Epoch (in local time not gmt/utc!).
        StartDateTimeFloat = time.mktime(StartDateTimeTuple)
        # convert to a float in seconds
        DurationFloat     = float(Duration)*1.e-3
        # calculate Stop datetime
        StopDateTimeFloat = StartDateTimeFloat + DurationFloat
        # round upwards to whole seconds, as agreed for the ValidityStop time
        StopDateTimeFloat = math.ceil(StopDateTimeFloat)
        # convert from seconds to pythons internal date-time structure (tuple)
        # again use localtime, since mktime() above did this also.
        StopDateTimeTuple = time.localtime(StopDateTimeFloat)
        # convert to a string 
        StopDateTime      = time.strftime(DateTimeFormat,StopDateTimeTuple)
        StopYear    = StopDateTime[0:4]
        StopMonth   = StopDateTime[4:6]
        StopDay     = StopDateTime[6:8]
        StopHour    = StopDateTime[9:11]
        StopMinute  = StopDateTime[11:13]
        StopSecond  = StopDateTime[13:15]

    StartDateTime = (StartYear,StartMonth,StartDay,
                     StartHour,StartMinute,StartSecond)
    StopDateTime  = (StopYear, StopMonth, StopDay,
                     StopHour, StopMinute, StopSecond)
    return (StartDateTime, StopDateTime)
    #  #]
def EncodeStartStopDateTime(StartDateTime,StopDateTime):
    #  #[
    # always return the L2B format:
    # YYYYMMDD.HHMMSS.YYYYMMDD.HHMMSS
    # 20071002T000002_20071002T000040 (at least for L2B this is the case)
    # (validity start/(validity stop)
    (StartYear,StartMonth, StartDay,
     StartHour,StartMinute,StartSecond) = StartDateTime
    (StopYear, StopMonth,  StopDay,
     StopHour, StopMinute, StopSecond ) = StopDateTime
    
    StartDateTimeStr = StartYear+StartMonth +StartDay+"T"+\
                       StartHour+StartMinute+StartSecond
    StopDateTimeStr  = StopYear +StopMonth  +StopDay+"T"+\
                       StopHour +StopMinute +StopSecond

    StartStopDateTimeStr = StartDateTimeStr+"_"+StopDateTimeStr

    return StartStopDateTimeStr
    #  #]
def XmlEncodeDateTime(DateTime):
    #  #[
    # only to be used in the FixedHeader !
    # The Date_Time in the log messages have YET another format ...
    # (but we don't have to handle that one, since we just copy
    #  it fromt he log file)
    
    #(Year,Month,Day,Hour,Minute,Second) = DateTime
    XmlDateTimeString = "UTC=%s-%s-%sT%s:%s:%s" % DateTime
    return XmlDateTimeString
    #  #]
def GetJobOrderFilenameCenterpart(JobOrderFile):
    #  #[

    # the string JobOrderFile variable may contain a path
    # (possibly containing dots) in front of the JobOrder filename.
    # This path needs to be removed
    # first before inspecting the name itself (since the dots inside the
    # filename are used to extract the center string of the JobOrderFileName).
    (JobOrderPath,JobOrderFileName) = os.path.split(JobOrderFile)

    JobOrderFileNameParts = JobOrderFileName.split(".")
    if (not (len(JobOrderFileNameParts) == 3)):
        RaiseError("The JobOrder filename is expected to have the format: "+
                   "JobOrder.Id.xml\n"+
                   "The current name: ["+JobOrderFileName+"] seems different!")

    JobOrderFilenameCenterpart = JobOrderFileName.split(".")[1]
    return JobOrderFilenameCenterpart

    #  #]
def CheckInputFile(File,txt):
    #  #[
    # check to see if it exists
    if (not os.path.exists(File)):
        RaiseError("could not find "+txt+": "+File+"\n"+\
                   "Please make sure this file exists before calling the ReportGenerator.")
    # check to see if it is a proper file
    if (not os.path.isfile(File)):
        RaiseError(txt+": "+File+"\nseems not a regular file!\n"+\
                   "Please make sure this file exists and is readable\n"+\
                   "before calling the ReportGenerator.")
    # check for proper read permissions
    if (not os.access(File,os.R_OK)):
        RaiseError(txt+": "+LogFile+"\nseems not readable!\n"+\
                   "Please make sure this file exists and is readable\n"+\
                   "before calling the ReportGenerator.")
    #  #]
def ParseJobOrder(JobOrderFile):
    #  #[
    try:
        xmldoc = minidom.parse(JobOrderFile).documentElement
    except:
        RaiseError("error while parsing JobOrderFile: "+JobOrderFile+"\n"+
                   "This file seems not to have proper xml formatting!")
    return xmldoc
    #  #]
def JobOrderExtractConfSettings(xmldoc):
    #  #[
    # there should be only one Ipf_Conf block in a JobOrder,
    # so no need for looping here
    IpfConf       = xmldoc.getElementsByTagName('Ipf_Conf')[0]
    
    JO_ProcessorName = IpfConf.getElementsByTagName('Processor_Name')[0].firstChild.data
    JO_Version       = IpfConf.getElementsByTagName('Version')[0].firstChild.data
    JO_OrderType     = IpfConf.getElementsByTagName('Order_Type')[0].firstChild.data
    JO_TestSwitch    = IpfConf.getElementsByTagName('Test')[0].firstChild.data
    JO_ProcStation   = IpfConf.getElementsByTagName('Processing_Station')[0].firstChild.data

    return (JO_ProcessorName, JO_Version, JO_OrderType, JO_TestSwitch, JO_ProcStation)
    #  #]
def JobOrderExtract(xmldoc):
    #  #[
    InputDir  = "[undefined]"
    OutputDir = "[undefined]"
    L2AorL2B  = "[undefined]"
    LogFile   = "[undefined]"
    ProductReportFileType     = "[undefined]"
    ProductReportFileName     = "[undefined]"
    ProductReportFileNameType = "[undefined]"
     
    ProcList = xmldoc.getElementsByTagName('Ipf_Proc')
    for Proc in ProcList:
        TaskName = Proc.getElementsByTagName('Task_Name')[0].firstChild.data
        TaskVersion = Proc.getElementsByTagName('Task_Version')[0].firstChild.data

        if (TaskName=='IPF2_REP_GEN'):  L2AorL2B = "L2A"
        if (TaskName=='IPF2B_REP_GEN'): L2AorL2B = "L2B"

        # extract the name of the ProductDir and ListDir 
        # and the LogFile, ProductReportFileName, ProductReportFileNameType
        if ((TaskName=='IPF2_REP_GEN') | (TaskName=='IPF2B_REP_GEN')):
            #  #[ check the task version number
            if (TaskVersion != ScriptVersion):
                print "WARNING: RepGen ScriptVersion is not equal to TaskVersion"
                print "WARNING: as specified in the JobOrder!"
                print "WARNING: Inconsistent version numbering may indicate that"
                print "WARNING: an outdated RepGen script is being used"
                print "WARNING: TaskVersion   = "+TaskVersion
                print "WARNING: ScriptVersion = "+ScriptVersion
            #  #]
            #  #[ get the InputDir
            InputList = Proc.getElementsByTagName('Input')
            FirstInput = InputList[0]
            FileNameType = FirstInput.\
                           getElementsByTagName('File_Name_Type')[0].\
                           firstChild.data
            if (FileNameType == "Physical"):
                InputFileAndDir = FirstInput.getElementsByTagName('File_Name')[0].\
                                  firstChild.data
                (InputDir,InputFile) = os.path.split(InputFileAndDir)
            else:
                RaiseError("Input specification used by ReportGenerator is\n"+\
                           "expected to be Physical, but currently it is: ["+\
                           FileNameType+"]\n"+\
                           "Sorry, this case is not yet defined")
            #  #]
            #  #[ get the OutputDir
            OutputList = Proc.getElementsByTagName('Output')
            for Output in OutputList:
                FileType = Output.getElementsByTagName('File_Type')[0].firstChild.data
                if ((FileType == "ALD_REP_2A") |
                    (FileType == "RP_PRO_2A")  |
                    (FileType == "ALD_REP_2B")  ):
                    FileNameType = Output.\
                                   getElementsByTagName('File_Name_Type')[0].\
                                   firstChild.data
                    if (FileNameType == "Directory"):
                        OutputDir = Output.getElementsByTagName('File_Name')[0].\
                                    firstChild.data
                    else:
                        RaiseError("Output specification used by ReportGenerator is\n"+\
                                   "expected to be Directory, but currently it is: ["+\
                                   FileNameType+"]\n"+\
                                   "Sorry, this case is not yet defined")
            #  #]
            #  #[ get the LogFile name
            InputList = Proc.getElementsByTagName('Input')
            for Input in InputList:
                FileType = Input.getElementsByTagName('File_Type')[0].firstChild.data
                if (FileType == "LOG"):
                    FileNameType =  Input.getElementsByTagName('File_Name_Type')[0].\
                                   firstChild.data
                    if (FileNameType == "Physical"):
                        LogFile = Input.getElementsByTagName('File_Name')[0].firstChild.data
                    else:
                        RaiseError("File_Name_Type specification for LOG file used by "+\
                                   "ReportGenerator is\n"+\
                                   "expected to be Physical, but currently it is: ["+\
                                   FileNameType+"]\n"+\
                                   "Sorry, this case is not yet defined") 
        #  #]
            #  #[ get the ProductReportFileName/ProductReportFileNameType
            OutputList = Proc.getElementsByTagName('Output')
            for Output in OutputList:
                FileType = Output.getElementsByTagName('File_Type')[0].firstChild.data
                if ((FileType == "ALD_REP_2A") |
                    (FileType == "RP_PRO_2A")  |
                    (FileType == "ALD_REP_2B")  ):
                    ProductReportFileType = FileType
                    ProductReportFileNameType = \
                        Output.getElementsByTagName('File_Name_Type')[0].\
                               firstChild.data
                    ProductReportFileName = \
                        Output.getElementsByTagName('File_Name')[0].\
                               firstChild.data
        #  #]

    #  #[ a number of sanity checks
    
    # note that InputDir may be an empty string, if the input gives only
    # a physical filename without any prepending path, so test for this:
    if (InputDir  == ""): InputDir  = "./"
    if (OutputDir == ""): OutputDir = "./"

    if ( (InputDir  == "[undefined]") |
         (OutputDir == "[undefined]")  ):
        RaiseError("InputDir or OutputDir to be used by ReportGenerator are\n"+\
                   "not properly defined in the JobOrder file. \n"+\
                   "InputDir =["+InputDir +"]\n"+\
                   "OutputDir=["+OutputDir+"]\n"+\
                   "for one of the tasks 'IPF2_L1B_L2A' or 'IPF2B_REP_GEN'\n"+\
                   "Something seems wrong in this JobOrder file, please fix it ...")

    if (os.path.abspath(InputDir) != os.path.abspath(OutputDir)):
        RaiseError("InputDir and OutputDir to be used by ReportGenerator differ\n"+\
                   "This is not how we expected the ReportGenerator would be used. \n"+\
                   "InputDir =["+InputDir +"]\n"+\
                   "OutputDir=["+OutputDir+"]\n"+\
                   "Something seems wrong in this JobOrder file, please fix it ...")

    if (ProductReportFileName == "[undefined]"):
        RaiseError("no task named IPF2B_REP_GEN or IPF2_REP_GEN found\n"+\
                   "with an outputfile type of ALD_REP_2A or RP_PRO_2A or ALD_REP_2B\n"+\
                   "Something seems wrong in this JobOrder file, please fix it ...")

    #  #]

    # the ProductList should be written in the same directory as
    # the Product itself, so retrieving the ListDir is not needed
    # (see PDS-IPF-ICD generic interface guidelines, sec. 4.8, p.68)
    ProductDir = OutputDir
    ListDir    = ProductDir

    return (ProductDir,ListDir,LogFile,L2AorL2B,
            ProductReportFileType,ProductReportFileName,ProductReportFileNameType)
    #  #]
def ExtractProductDateTimes(ProductListFileName):
    #  #[

    # note that it is a valid possibility to NOT have a LIST file
    # this happens in our test suite, and when testing manually
    # A LIST file is only required in the TLE environment.
    # therefore an existence test is needed before opening the LIST file

    # some constants
    FileType_L2A     = "ALD_U_N_2A"
    FileType_L2B     = "ALD_U_N_2B"

    # initialisations
    Product1DateTime = "[undefined]"
    Product2DateTime = "[undefined]"
    StartDateTime = GetUndefinedDateTime()
    StopDateTime  = GetUndefinedDateTime()
    
    if (os.path.exists(ProductListFileName)):
        #  #[
        # try to read the ProductListFile to extract the datetime of the
        # L2A or L2B product, which is needed to construct a product report
        # with the same datetime string.
        # (see email 05-04-07 15:33 by Christophe)
        for line in file(ProductListFileName):
            #print "line = ["+line+"]"
            pos = line.find(FileType_L2B)
            if (pos != -1):
                #print "string contains "+FileType_L2B+"..."
                # expected filename format:
                # AE_OPER_ALD_U_N_2B_20071002T000002_20071002T000040_0001.HDR
                # 012345678901234567890123456789012345678901234567890123456789
                # 0         1         2         3         4         5
                #print line[19:50]
                #
                # additional remark: 
                # it is possible for the L2BP to generate 2 product files
                # based on a single L1B input file (in case 2 AMD files are
                # provided that each partially overlap the L1B file)
                # Both product files should be in the product list.
                # To get a proper report start/stop time, take the start of the
                # first file and the stop of the second file.
                # ASSUMPTION is that the product files are in the proper order
                # in the product list file !!!!
                if (Product1DateTime == "[undefined]"):
                    Product1DateTime = line[19:50]
                Product2DateTime = line[19:50]
            pos = line.find(FileType_L2A)
            if (pos != -1):
                #print "string contains "+FileType_L2A+"..."
                # expected filename format:
                # AE_OPER_ALD_U_N_2A_20071010T080005036_005376000_000123_0001.HDR
                # 012345678901234567890123456789012345678901234567890123456789
                # 0         1         2         3         4         5
                #print line[19:47]
                # see remark on multiple files for L2B
                # (does this also apply for L2A?)
                if (Product1DateTime == "[undefined]"):
                    Product1DateTime = line[19:47]
                Product2DateTime = line[19:47]
        #  #]
    if (Product1DateTime != "[undefined]"):
        #  #[
        # get ValidityStartDateTime from the first product
        (StartDateTime, DummyStopDateTime) = \
                        DecodeStartStopDateTime(Product1DateTime)

        # get ValidityStopDateTime from the last product
        # the second DateTime is always filled, if the first one is filled,
        # so no additional undefined check is needed.
        (DummyStartDateTime, StopDateTime) = \
                             DecodeStartStopDateTime(Product2DateTime)
        #  #]

    return (StartDateTime, StopDateTime)

    #  #]
def ConstructProductReportFileName(StartDateTime, StopDateTime,CurrentDateTime,
                                   FileClass,ProductReportFileType):
    #  #[

    # construct the first part,
    # for example: AE_RPRO_ALD_REP_2B
    ProductReportFileName = "AE_"+FileClass+"_" + ProductReportFileType

    StartYear = StartDateTime[0]
    if (StartYear=="[undefined]"):
        # backup method. No L2A or L2B file was found in the
        # ProductList file, so use current date and time to
        # construct the filename for the report
        StartStopDateTimeString = EncodeStartStopDateTime(CurrentDateTime,
                                                          CurrentDateTime)
    else:
        StartStopDateTimeString = EncodeStartStopDateTime(StartDateTime,
                                                          StopDateTime)
            
    ProductReportFileName = ProductReportFileName + \
                            "_" + StartStopDateTimeString  + \
                            "_0001.EEF"
            
    return ProductReportFileName 
    #  #]
def ConstructMessage(doc, DateTime,NodeName,ProcName,ProcVersion,ProcID,
                     MsgType,LogText):
    #  #[ 

    # create the Message tag pair
    Message = doc.createElement("Message")
                
    # create and fill the Date_Time element
    XmlDateTime  = doc.createElement("Date_Time") # create the tag
    Message.appendChild(XmlDateTime) # make it a child of Message
    XmlDateTime.appendChild(doc.createTextNode(DateTime)) # fill it
        
    # create and fill the Node_Name element
    XmlNodeName = doc.createElement("Node_Name")
    Message.appendChild(XmlNodeName)
    XmlNodeName.appendChild(doc.createTextNode(NodeName))
    
    # create and fill the Processor_Name element
    XmlProcName = doc.createElement("Processor_Name")
    Message.appendChild(XmlProcName)
    XmlProcName.appendChild(doc.createTextNode(ProcName))
    
    # create and fill the Processor_Version element
    XmlProcVersion = doc.createElement("Processor_Version")
    Message.appendChild(XmlProcVersion)
    XmlProcVersion.appendChild(doc.createTextNode(ProcVersion))
    
    # create and fill the PID element
    XmlProcID = doc.createElement("PID")
    Message.appendChild(XmlProcID)
    XmlProcID.appendChild(doc.createTextNode(ProcID))
    
    # create and fill the Type element
    XmlMsgType = doc.createElement("Type")
    Message.appendChild(XmlMsgType)
    XmlMsgType.appendChild(doc.createTextNode(MsgType))
    
    # create and fill the Text element
    # this text is packed in a special !CDATA command
    # to prevent problems when a '<' or '>' character
    # might occur in the log message
    XmlMsgText = doc.createElement("Text")
    Message.appendChild(XmlMsgText)
    XmlMsgText.appendChild(doc.createCDATASection(LogText))

    return Message

    #  #]
def WriteToXmlFile(node, file_object, indent="  "):
    #  #[

    if (use_xml_dom_ext):
        xml.dom.ext.PrettyPrint(node, file_object)
    else:
        # this toprettyxml uses the writexml functions which have been
        # adapted above

        #import inspect
        #if(inspect.getargspec(node.toprettyxml)[0].count("encoding") == 1):
        #    file_object.write(node.toprettyxml(indent,
        #                                       newl="\n",encoding='UTF-8'))
        #else:
        #    file_object.write(node.toprettyxml(indent,newl="\n"))

        # ignore the encoding settings, and just always use 'UTF-8'
        # because this encoding causes compatibility problems when
        # used with python 2.2
        file_object.write(node.toprettyxml(indent,newl="\n"))
            
    # alternative without any indentations or newlines
    #file_object.write(node.toxml())

    #  #]
    
#---------- main program ----------------------
#  #[ initialisations

# According to L2B SPR 28 the Creator_Version, to be used when the script
# is run without JobOrder should correspond to the delivery version
# (so 01.33 for the L2B and 02.01? for the L2A) but since ESA should only use 
# this script with a JobOrder to start with, this is not very relevant 
# and I'll keep a separate version number for this RepGen script for now
# Note that the Task_Version in the JobOrder file should correspond to this
# version number, otherwise a warning will be issued
ScriptVersion = "00.34" # only used in case the script is run without JobOrder
FileClass     = "TEST"  # only used in case the script is run without JobOrder
L2AorL2B      = "L2B"   # only used in case the script is run without JobOrder

# some initilisations
JobOrderFile          = "[undefined]"
LogFile               = "[undefined]"
ProductDir            = "[undefined]"
ProductReportFileName = "[undefined]"
ProductListFileName   = "[undefined]"
ProductReportFileType = "[undefined]"
ListDir               = "[undefined]"
StartDateTime = GetUndefinedDateTime()
StopDateTime  = GetUndefinedDateTime()

# this version number now refers to the version of the PDS-IPF-ICD
# Generic Interface Guidelines document, in which the format of the
# product report is defined. This way it is generic, and can be used for
# both L2A and L2B reports without customisation.
nameSpace = "http://www.esa.int/schemas/ae/Product_Report_02.20"

# some JobOrder (JO) items that we need
JO_ProcessorName      = "[undefined]"
JO_Version            = "[undefined]"
JO_OrderType          = "[undefined]"
JO_TestSwitch         = "[undefined]"
JO_ProcStation        = "[undefined]"

# get the current gmt date-time value
CurrentDateTime = GetCurrentDateTime()
#  #]
#  #[ define the regexp pattern to use for splitting and filtering

# pattern to be used for parsing the loglines
LogPattern = re.compile(r"""
^                    # beginning of the line
(20[0-1][0-9]-       # year    (yyyy)
 [0-1][0-9]-         # month   (mm)
 [0-3][0-9]T         # day     (dd)
 [0-2][0-9]:         # hour    (hh)
 [0-5][0-9]:         # minute  (mm)
 [0-6][0-9].         # seconds (ss)
 [0-9][0-9][0-9]     # milliseconds (nnn)
 [0-9][0-9][0-9])    # microseconds (nnn)
[ ]                  # a single space
(\S+)                # the node name (1 or more non-white-space characters)
[ ]                  # a single space
(\S+)                # the processor name (1 or more non-white-space chars)
[ ]                  # a single space
(\d{2}.\d{2})        # the processor version (2 digits, a dot, and 2 digits)
[ ]                  # a single space
\[(\d+)\]            # the process ID (1 or more digits)
:[ ]                 # a semicolon and a single space
\[([D,I,P,W,E])\]    # the message type (a single D,I,P,W or E character)
[ ]                  # a single space
(.*)                 # all that remains (i.e. the message text itself)
$                    # end of the line
""",re.VERBOSE)

#  #]
#  #[ commandline handling
if ( (len(sys.argv) != 2) and (len(sys.argv) != 3)):
    PrintUsage()
    os._exit(1)

if (len(sys.argv) == 2):
    UseJobOrder  = True
    JobOrderFile = sys.argv[1]

if (len(sys.argv) == 3):
    if (sys.argv[1] != '-l'):
        PrintUsage()
        os._exit(1)
    UseJobOrder   = False
    LogFile = sys.argv[2] 
#  #]

if (UseJobOrder):
    #  #[ extract some settings from the JobOrder file
    JobOrderFilenameCenterpart = GetJobOrderFilenameCenterpart(JobOrderFile)
    # do some sanity tests on the input file (existence, read permission etc.)
    CheckInputFile(JobOrderFile,"JobOrderFile")
    # parse the xml
    xmldoc = ParseJobOrder(JobOrderFile)

    (JO_ProcessorName,
     JO_Version, 
     JO_OrderType,
     JO_TestSwitch,
     JO_ProcStation) = JobOrderExtractConfSettings(xmldoc)
    

    if (JO_TestSwitch.lower() == "true"):
        FileClass = "TEST"
    else:
        FileClass = JO_OrderType.upper()

    # remark: the ProductReportFileName field in the Output block of the
    # report generator should actually contain the specification of the
    # output directory where the product report should be written.
    # This setting is returned in ProductDir.
    # Therefore the result is stored in the dummy variable
    # DummyProductReportFileName is not used at the moment
    (ProductDir,ListDir,LogFile,L2AorL2B,
     ProductReportFileType,DummyProductReportFileName,
     ProductReportFileNameType) = JobOrderExtract(xmldoc)

    if (L2AorL2B == "L2A"):
        # construct the name of the LogFile since it is not (yet)
        # present in the L2A JobOrder file
        # (it should be added there I think, JK)
        LogFile = os.path.join(ProductDir,
                               "l2ap_"+JobOrderFilenameCenterpart+".log")

    if (LogFile == "[undefined]"):
        RaiseError("Sorry, could not find any LogFile in JobOrder file:"+\
                   JobOrderFile+"\n"+
                   "Something seems wrong in this JobOrder file, "+\
                   "please fix it ...")

    # this one is constructed using the name of the JobOrder file,
    # even though it is given in the L2A JobOrder file
    # (it should be removed there I think, JK)
    ProductListFileName = os.path.join(ListDir,
                                       JobOrderFilenameCenterpart+".LIST")    

    (StartDateTime, StopDateTime) = \
                    ExtractProductDateTimes(ProductListFileName)

    if (ProductReportFileNameType == "Directory"):
        ProductReportFileName = \
           ConstructProductReportFileName(StartDateTime, StopDateTime,
                                          CurrentDateTime,
                                          FileClass,ProductReportFileType)
    else:
        RaiseError("he ProductReportFileNameType is found to be "+\
                   ProductReportFileNameType+"\n"+\
                   "This has not yet been implemented")

    # combine the constructed Product Report name with the ProductDir
    ProductReportFileName = os.path.join(ProductDir,ProductReportFileName)
    #  #]
else:
    #  #[ in this case the LogFile has been read from the commandline
    # (intended for testing purposes only)
    ProductDir = "./"
    ListDir    = ProductDir
    ProductReportFileName = os.path.join(ProductDir,"DefaultProduct.REPORT")
    ProductListFileName = os.path.join(ListDir,"DefaultProduct.LIST")
    #  #]

#  #[ set ProductReportFileBaseName and ProductListFileName
(PRpath,PRname) = os.path.split(ProductReportFileName)
(ProductReportFileBaseName,PRextension) = os.path.splitext(PRname)

print "Current settings for the ReportGenerator:"
print "ProductDir                = ",ProductDir
print "LogFile                   = ",LogFile
print "ListDir                   = ",ListDir
print "ProductListFileName       = ",ProductListFileName
print "ProductReportFileName     = ",ProductReportFileName
print "ProductReportFileBaseName = ",ProductReportFileBaseName
print "L2AorL2B                  = ",L2AorL2B

# do some sanity tests on the input file (existence, read permission etc.)
CheckInputFile(LogFile,"logfile")
#  #]
#  #[ start a new xml document for the output xml file
if (use_xml_dom_ext):
    doc = minidom.Document()
else:
    doc = MyMinidomDocument()
    
EERoot = doc.createElement("Earth_Explorer_File")
doc.appendChild(EERoot)
EERoot.setAttribute("xmlns",nameSpace)
#  #]
#  #[ add empty Earth Explorer Headers (FH and VH)
EEheader = doc.createElement("Earth_Explorer_Header")
EERoot.appendChild(EEheader)
Fixedheader = doc.createElement("Fixed_Header")
EEheader.appendChild(Fixedheader)
VariableHeader = doc.createElement("Variable_Header")
EEheader.appendChild(VariableHeader)
if (use_xml_dom_ext):
    # Dorit: create empty variable header text node to please EE standard
    # JK: this is not needed when using the adapted minidom functions
    VariableHeader.appendChild(doc.createTextNode("             "))
#  #]
#  #[ define and fill the Fixed Header

# define the Fixed Header items
FHfileName        = doc.createElement("File_Name")
FHfileDescription = doc.createElement("File_Description")
FHnotes           = doc.createElement("Notes")
FHmission         = doc.createElement("Mission")
FHfileClass       = doc.createElement("File_Class")
FHfileType        = doc.createElement("File_Type")
FHvalidityPeriod  = doc.createElement("Validity_Period")
FHsource          = doc.createElement("Source")

# append the Fixed Header items
Fixedheader.appendChild(FHfileName)
Fixedheader.appendChild(FHfileDescription)
Fixedheader.appendChild(FHnotes)
Fixedheader.appendChild(FHmission)
Fixedheader.appendChild(FHfileClass)
Fixedheader.appendChild(FHfileType)
Fixedheader.appendChild(FHvalidityPeriod)
Fixedheader.appendChild(FHsource)

# fill the Fixed Header items
ScriptPathAndName = sys.argv[0]
(ScriptPath,ScriptName) = os.path.split(ScriptPathAndName)

FHfileDescriptionValue   = L2AorL2B+" Product Report"
FHnotesValue             = "This product report was generated by the script: "+\
                           ScriptName+", version nr: "+ScriptVersion
DefaultFHmission         = "Aeolus"
FHfileClassValue         = FileClass
FHfileName.appendChild(       doc.createTextNode(ProductReportFileBaseName))
FHfileDescription.appendChild(doc.createTextNode(FHfileDescriptionValue))
FHnotes.appendChild(          doc.createTextNode(FHnotesValue))
FHmission.appendChild(        doc.createTextNode(DefaultFHmission))
FHfileClass.appendChild(      doc.createTextNode(FHfileClassValue))
FHfileType.appendChild(       doc.createTextNode(ProductReportFileType))

# define the Validity_Period items
ValidityStart = doc.createElement("Validity_Start")
ValidityStop  = doc.createElement("Validity_Stop")

# append the Validity_Period items
FHvalidityPeriod.appendChild(ValidityStart)
FHvalidityPeriod.appendChild(ValidityStop)

# fill the Validity_Period items with special values
# that signal begin and end of mission
# Note that this is only allowed in the FixedHeader and only
# for the ValidityStart/Stop tags.
# See Herberts mail of 9-mar-2007
DefaultValidityStart = "UTC=0000-00-00T00:00:00"
DefaultValidityStop  = "UTC=9999-99-99T99:99:99"

StartYear = StartDateTime[0]
if (StartYear == "[undefined]"):
    ValidityStartString = DefaultValidityStart
    ValidityStopString  = DefaultValidityStop
else:
    ValidityStartString = XmlEncodeDateTime(StartDateTime)
    ValidityStopString  = XmlEncodeDateTime(StopDateTime)

ValidityStart.appendChild(doc.createTextNode(ValidityStartString))
ValidityStop.appendChild(doc.createTextNode(ValidityStopString))

# define the Source items
SourceSystem         = doc.createElement("System")
SourceCreator        = doc.createElement("Creator")
SourceCreatorVersion = doc.createElement("Creator_Version")
SourceCreationDate   = doc.createElement("Creation_Date")

# append the Source items
FHsource.appendChild(SourceSystem)
FHsource.appendChild(SourceCreator)
FHsource.appendChild(SourceCreatorVersion)
FHsource.appendChild(SourceCreationDate)

# fill the Source items
SourceSystemValue  = JO_ProcStation
# remark: the TaskName is currently different for l2A and L2B
# todo: make them equal
if   (L2AorL2B=="L2A"):
    TaskName = "IPF2_REP_GEN"
elif (L2AorL2B=="L2B"):
    TaskName = "IPF2B_REP_GEN"
else:
    RaiseError("L2AorL2B switch seems not properly set. ")
    
SourceCreatorValue        = JO_ProcessorName+"/"+TaskName
if (JO_Version=="[undefined]"):
    SourceCreatorVersionValue = ScriptVersion
else:
    SourceCreatorVersionValue = JO_Version    

CreationDateTime = XmlEncodeDateTime(CurrentDateTime)
SourceSystem.appendChild(doc.createTextNode(SourceSystemValue))
SourceCreator.appendChild(doc.createTextNode(SourceCreatorValue))
SourceCreatorVersion.appendChild(doc.createTextNode(SourceCreatorVersionValue))
SourceCreationDate.appendChild(doc.createTextNode(CreationDateTime))

#  #]
#  #[ add the Product_Report in a Data block
DataBlock = doc.createElement("Data_Block")
DataBlock.setAttribute("type","xml")
EERoot.appendChild(DataBlock)
ReportRoot = doc.createElement("Product_Report")
DataBlock.appendChild(ReportRoot)
#  #]
#  #[ the actual work

# fill the report filename in the output report
reportFileName = doc.createElement("File_Name")
text = doc.createTextNode(ProductReportFileBaseName)
reportFileName.appendChild(text)
ReportRoot.appendChild(reportFileName)

# create the list of messages
# the count attribute will only be set after we are done
# with the next for loop
ListOfMessages = doc.createElement("List_of_Messages")
ReportRoot.appendChild(ListOfMessages)

file_object_tmp1 = open(ProductReportFileName+".temp1", "w")

n_ok     = 0
n_not_ok = 0
for line in file(LogFile):
    #  #[ parse the log file and convert to xml

    result = LogPattern.match(line)
    if (result):
        n_ok+=1
        #print "n_ok = ",n_ok
        #print result.groups()
        DateTime    = result.groups()[0]
        NodeName    = result.groups()[1]
        ProcName    = result.groups()[2]
        ProcVersion = result.groups()[3]
        ProcID      = result.groups()[4]
        MsgType     = result.groups()[5]
        LogText     = result.groups()[6]
        
        Message = ConstructMessage(doc,DateTime,NodeName,ProcName,ProcVersion,
                                   ProcID,MsgType,LogText)
        # MsgToXml
        #ListOfMessages.appendChild(Message)

        # print to a temporary file and dispose of this
        # message datastruct. This is necessary because otherwise
        # memoryuse will grow beyond reasonable proportions and
        # this script will bring an average size computer on its knees
        # when nr of log messages is in the order of 100.000 or more.

        WriteToXmlFile(Message, file_object_tmp1)

        #Message.delete()
    else:
        n_not_ok+=1

    #  #]

file_object_tmp1.close()        
print "parsed:    %i lines" % n_ok
print "discarded: %i lines" % n_not_ok

# ok, the looping is now done, so we know the nr of logmsgs
NrMsgs ="%i" % n_ok
ListOfMessages.setAttribute("count",NrMsgs)

# write the result to a temp file
file_object_tmp2 = open(ProductReportFileName+".temp2", "w")
WriteToXmlFile(doc, file_object_tmp2)
file_object_tmp2.close()

# combine both temp files and write the results to the report file
file_object = open(ProductReportFileName, "w")
line_count=0
for line in file(ProductReportFileName+".temp2"):
    line_count=line_count+1
    if (line_count <= 28):
        file_object.write(line)
for line in file(ProductReportFileName+".temp1"):
    file_object.write("        "+line)
line_count=0
for line in file(ProductReportFileName+".temp2"):
    line_count=line_count+1
    if (line_count > 28):
        file_object.write(line)
file_object.close()

# remove the 2 temporary files
os.remove(ProductReportFileName+".temp1")
os.remove(ProductReportFileName+".temp2")

# write the report file into the product list file (append it)
file_object = open(ProductListFileName, "a")
file_object.write(os.path.split(ProductReportFileName)[1])
file_object.write("\n")
file_object.close()

#  #]

print "done"
