#!/bin/csh
# the -v options is very usefull for debugging ...
#!/bin/csh -v

##############################################################################
# Documentation:
#
# This script automatically builds the full  L2BP
# and all test programs coming with it.

# The user doing the install should provide the following
# commandline inputs:
# 1. dir in which the source code can be found
# 2. dir in which the binary L2B_Processor should be installed
# 3. dir in which the ee_cfi_software can be found
# 4. dir in which the libxml2 library can be found
# 5. full path and filename of the binary datapack
#
# all path names should be given as full paths, not relative
#
# Note: Directory3 ($EECFIDIR) must end with "/" !!
#
# This script will then:
# 1. compile the software
# 2. copy the main program and all tools to the install/ directory
# 3  copy the BUFR tables needed to run the ee2bufr tool to the
#    install/BUFR_tables directory
# 4. copy all data files needed to test the software in the ThinLayerEmulator
#    using the provided order.5.xml test file, to the install/TLE_testfiles/ 
#    directory 
# 5. run the automated test suite, and report on the results
#

# Example:
# $SOURCEDIR/Scripts/build_L2BP.sc \
#        $SOURCEDIR $INSTALDIR $EECFIDIR $EE_XML_DIR $DATAPACKFILE
# 
# or with variables expanded:
# ./Scripts/build_L2BP.sc 
#   /nobackup/users/kloedej/temp_ectrans/dat_CY30R1_solstice/aeolus/  \
#   /nobackup/users/kloedej/temp_ectrans/dat_CY30R1_solstice/install/ \
#   /nobackup/users/kloedej/versiebeheer/adm/L2B_processor/ee_cfi_software_3.7.1/ \
#   /nobackup/users/kloedej/versiebeheer/adm/L2B_processor/ee_cfi_software_3.7.1/aux_tools/libxml \
#   /nobackup/users/kloedej/temp_ectrans/L2BP_Release1.32_datapack_20080118.tar.gz
#
# This call may ofcourse be stored in a little script for convenience of the usr,
# for an example, see the file:   ./Scripts/build_L2BP.KNMI.sc

# Optionally, building the software without the ee_cfi is possible.
# in this case, the 3rd and 4th commandline option should be
# set to "dontuse". This is especially usefull when the gfortran compiler
# has to be used, which currently seems incompatible with the ee_cfi library.

# commandline inputs:
# 1. dir in which the source code can be found
# 2. dir in which the binary L2B_Processor should be installed
# 3. set to "dontuse"
# 4. set to "dontuse"
# 5. full path and filename of the binary datapack
#
##############################################################################
# Author:  J. de Kloe
#
# Modifications:
# J. de Kloe   27-Jun-2006 initial version
# D. Tan       29-Jun-2006 Cater for cases where "." is not in PATH, second search for libxml2.a
# D. Tan       21-Jul-2006 Comment out second make
# D. Tan       08-Dec-2006 Copy other executables to $INSTALDIR, plus symlinks
# D. Tan       25-Jan-2007 Add L1B EE2BUFR executables
# J. de Kloe   13-Feb-2007 remove a csh incompatibility between linux/unix
# D. Tan       16-Feb-2007 Add L1B EE2BUFR compiled Bufr tables in a sub-dir
# D. Tan       28-Mar-2007 Copy TestWriteL2BCdata to $INSTALDIR
# J. de Kloe   18-Apr-2007 added some code to circumvent the error that
#                          always occurred with the libxml2.a testing
# J. de Kloe   10-May-2007 added some code to create a install/TLE_testfiles/ 
#                          directory, holding all files needed to test the 
#                          software with the ThinLayerEmulator using 
#                          the provided order.5.xml test file.
# J. de Kloe   04-Jun-2007 enabled overwriting files in the install dir
#                          and added checks to warn if this occurs
# J. de Kloe   13-Jun-2007 added new L1B testfile to the TLE test files
# D. Tan       22-Oct-2007 new RBC dates in filename
# J. de Kloe   28-Jan-2008 added ConvertKnmiAscToAMD to tools copied to the build directory
#                          added possibility to use simple_xml in stead of ee_cfi
#                          and updated the usage description above
#                          
##############################################################################

##############################################################################
# 1.  inspect the given commandline inputs
##############################################################################

set SOURCE_DIR  = $1
set INSTALL_DIR = $2
set EECFI_DIR   = $3
set EE_XML_DIR  = $4
set DATAPACK    = $5
set N_ARGS      = $#argv

#echo "arg(1) = SOURCE_DIR  = "${SOURCE_DIR}
#echo "arg(2) = INSTALL_DIR = "${INSTALL_DIR}
#echo "arg(3) = EECFI_DIR   = "${EECFI_DIR}
#echo "arg(4) = EE_XML_DIR  = "${EE_XML_DIR}
#echo "arg(5) = DATAPACK    = "${DATAPACK}
#echo "nargs  = "${N_ARGS}

##############################################################################
# 1a. check if the nr. of parameters is correct
##############################################################################

if ( "${N_ARGS}" != "5" ) then
  echo "ERROR: nr. of arguments differs from the expected 5 \!\!\!"
  goto usage
endif

echo "argcheck ok"

##############################################################################
# issue a message to acknowledge recognising the "dontuse" switch
##############################################################################
if ( "${EECFI_DIR}" == "dontuse" ) then
  echo "Not using the ee_cfi library for L2Bp building."
endif
if ( "${EE_XML_DIR}" == "dontuse" ) then
  echo "Not using the libxml2 library for L2Bp building."
endif

##############################################################################
# 1b. check if the sources are in the indicated place
#     by looking at some key files/subdirs
##############################################################################

if (!(-e "${SOURCE_DIR}")) then
  echo "ERROR: source dir does not exist or is not accessible...."
  goto usage
endif

set TEST_DIR = MieCoreProcessing
if (!(-e "${SOURCE_DIR}/${TEST_DIR}")) then
  echo "ERROR: the ${TEST_DIR} dir does not exist or is not accessible in"
  echo "the given source dir. Something seems wrong with this installation..."
  goto usage
endif

set TEST_DIR = RayleighBrillouinProcessing
if (!(-e "${SOURCE_DIR}/${TEST_DIR}")) then
  echo "ERROR: the ${TEST_DIR} dir does not exist or is not accessible in"
  echo "the given source dir. Something seems wrong with this installation..."
  goto usage
endif

set TEST_FILE = Set_Makeoptions
if (!(-e "${SOURCE_DIR}/${TEST_FILE}")) then
  echo "ERROR: the ${TEST_FILE} file does not exist or is not accessible in"
  echo "the given SOURCE_DIR. Something seems wrong with this installation..."
  goto usage
endif

echo "sources ok"

##############################################################################
# 1c. check if the install dir exists, and if not create it.
##############################################################################

if (-e "${INSTALL_DIR}") then
  echo "WARNING: install dir already exists..."
else
  echo "install dir does not yet exist..."
  # create the install directory, with any parents if needed
  echo "creating installation directory for L2BP...."
  mkdir -p ${INSTALL_DIR} 
  # make this directory accessibale to other users as well 
  # remind that this does not set permissions correct if the above command
  # created multiple subdirectories...
  chmod a+rx ${INSTALL_DIR}
endif

echo "install dir ok"

##############################################################################
# 1c.1. check if the install/BUFR_tables sub-dir exists, and if not create it.
##############################################################################
set BUFR_TABLES_DIR = ${INSTALL_DIR}/BUFR_tables
if (-e "${BUFR_TABLES_DIR}") then
  echo "WARNING: install/BUFR_tables sub-dir already exists..."
else
  echo "install/BUFR_tables sub-dir does not yet exist..."
  # create the install BUFR_tables sub-directory
  echo "creating installation directory for L1B ee2bufr BUFR tables...."
  mkdir ${BUFR_TABLES_DIR} 
  # make this directory accessibale to other users as well 
  chmod a+rx ${BUFR_TABLES_DIR}
endif

echo "install/BUFR_tables sub-dir ok"

##############################################################################
# 1c.2. check if the install/TLE_testfiles/ sub-dir exists, if not create it.
##############################################################################
set TLE_TEST_DIR = ${INSTALL_DIR}/TLE_testfiles
if (-e "${TLE_TEST_DIR}") then
  echo "WARNING: install/TLE_testfiles sub-dir already exists..."
else
  echo "install/TLE_testfiles sub-dir does not yet exist..."
  # create the install/TLE_testfiles sub-directory
  echo "creating installation directory for TLE test files ...."
  mkdir -p ${TLE_TEST_DIR} 
  # make this directory accessibale to other users as well 
  chmod a+rx ${TLE_TEST_DIR}
endif

echo "install/TLE_TEST_DIR sub-dir ok"

##############################################################################
# 1d. check if the ee_cfi_software is present, and has the expected version
##############################################################################

if ( "${EECFI_DIR}" == "dontuse" ) then
  echo "skipping presence check for ee_cfi library files"
else

  if (!(-e "${EECFI_DIR}")) then
    echo "ERROR: EECFI dir does not exist or is not accessible...."
    goto usage
  endif

  set TEST_DIR = explorer_file_handling
  if (!(-e "${EECFI_DIR}/${TEST_DIR}")) then
    echo "ERROR: the ${TEST_DIR} dir does not exists or is not accessible in"
    echo "the given EECFI_DIR. The ee_cfi software seems not correctly installed"
    goto usage
  endif

  set TEST_FILE = explorer_file_handling/include/explorer_file_handling.h
  if (!(-e "${EECFI_DIR}/${TEST_FILE}")) then
    echo "ERROR: the file ${TEST_FILE} does"
    echo "not exists or is not accessible in te given EECFI_DIR."
    echo "The ee_cfi software seems not correctly installed"
    goto usage
  endif

  # this usually is a symbolic link to the file checked above
  set TEST_FILE = include/explorer_file_handling.h
  if (!(-e "${EECFI_DIR}/${TEST_FILE}")) then
    echo "ERROR: the file ${TEST_FILE} does not exists or is not"
    echo "accessible in the given EECFI_DIR."
    echo "The ee_cfi software seems not correctly installed"
    goto usage
  endif

  # remark: if the file does not exist, the TEST_FILE variable remains empty
  # (due to the wildcard) so it cannot be used in the echo commands below ...
  set TEST_FILE = `ls ${EECFI_DIR}/explorer_file_handling/lib/*/libexplorer_file_handling.a`
  if (!(-e "${TEST_FILE}")) then
    echo "ERROR: the file explorer_file_handling/lib/*/libexplorer_file_handling.a"
    echo "does not exists or is not accessible in the given EECFI_DIR."
    echo "The ee_cfi software seems not correctly installed"
    goto usage
  endif

  # remark: if the file does not exist, the TEST_FILE variable remains empty
  # (due to the wildcard) so it cannot be used in the echo commands below ...
  set TEST_FILE = `ls ${EECFI_DIR}/lib/*/libexplorer_file_handling.a`
  if (!(-e "${TEST_FILE}")) then
    echo "ERROR: the file lib/*/libexplorer_file_handling.a does not exists or"
    echo "is not accessible in the given EECFI_DIR."
    echo "The ee_cfi software seems not correctly installed"
    goto usage
  endif

  # ok, all seems alright, so make the symbolic link for our own use
  # to the default name (installation dir of ee_cfi_software might have
  # a different name, like ee_cfi_software_3.7.1)
  # use the -f option to force overwriting any old links that might exist
  ln -fs ${EECFI_DIR} ${SOURCE_DIR}/ee_cfi_software

  echo "ee_cfi ok"

endif

##############################################################################
# 1f. check if the libxml2.a library can be found
##############################################################################

# remark:
# expected below EE_XML_DIR dir are the paths:
# include/libxml2/libxml/
# LINUX/lib/

# on my system this is installed in: ee_cfi_software/aux_tools/libxml/

if ( "${EE_XML_DIR}" == "dontuse" ) then
  echo "skipping presence check for libxml2 library files"
else

  if (!(-e "${EE_XML_DIR}")) then
    echo "ERROR: the file ${EE_XML_DIR} does not exists or is not accessible" 
    goto usage
  endif

  set EE_XML_INC = ${EE_XML_DIR}/include/libxml2/
  if (!(-e "${EE_XML_INC}")) then
    echo "ERROR: the ${EE_XML_INC} dir does not exists or is not accessible" 
    goto usage
  endif

  set TEST_FILE = ${EE_XML_INC}/libxml/globals.h
  if (!(-e "${TEST_FILE}")) then
    echo "ERROR: the file ${TEST_FILE} does not exists or is not accessible" 
    goto usage
  endif

  # init to a nonsense value
  set TEST_FILE = "[undefined]"

  # try the ls command on the first possible location
  ls ${EE_XML_DIR}/*/lib/libxml2.a >& /dev/null
  if ( $status == 0 ) then
    set TEST_FILE = `ls ${EE_XML_DIR}/*/lib/libxml2.a`
  endif

  # try the ls command on the second possible location
  ls ${EE_XML_DIR}/lib/libxml2.a >& /dev/null
  if ( $status == 0 ) then
    set TEST_FILE = `ls ${EE_XML_DIR}/lib/libxml2.a`
  endif

  set EE_XML_LIB = ${TEST_FILE:h}/
  if (!(-e "${EE_XML_LIB}")) then
    echo "ERROR: the ${EE_XML_LIB}/*/lib/ dir does not exists or is not accessible" 
    goto usage
  endif

  echo "libxml2.a ok"

endif

##############################################################################
# 1f. check if the binary datapack can be found
##############################################################################

if (!(-e ${DATAPACK})) then
  echo "ERROR: cannot acces the given binary datapack file ..."
  echo "the file does not exists or is not accessible ..."
  goto usage
endif

# ok, all the datapack is there, so make the symbolic link for our own use
# use the -f option to force overwriting any old links that might exist
set DATAPACK_WITHOUT_PATH = ${DATAPACK:t}
ln -fs ${DATAPACK} ${SOURCE_DIR}/${DATAPACK_WITHOUT_PATH}

echo "datapack ok"

##############################################################################
# 2. system settings
##############################################################################

# remark: I don't think it is wise to include the compiler choice here
# this still will require manual intervention by either
# sourcing any of trhe Scripts/use_* files, 
# or by setting the needed environment switches.

# run a little script to make sure all other needed scripts are executable
chmod a+x ${SOURCE_DIR}/Scripts/Set_Permissions
cd ${SOURCE_DIR}
Scripts/Set_Permissions

# use the ee_cfi software as default when this script is used:
# a later addition may be to use simple_xml in case the ee_cfi_software
# installation seems missing...

if ( "${EECFI_DIR}" == "dontuse" ) then
  echo "using simple_xml for xml handling"
  ./use_simple_xml
  ./Scripts/Set_Systemsettings
else
  echo "using ee_cfi library for xml handling"
  ./use_ee_cfi_software
  ./Scripts/Set_Systemsettings ${EECFI_DIR} ${EE_XML_INC} ${EE_XML_LIB}
endif

echo "system settings ok"

##############################################################################
# 3. install binary datapack
##############################################################################

# install the files from the binary datapack
./install_binary_datapack > /dev/null

echo "datapack installed"

##############################################################################
# 3. run the make system
##############################################################################

\rm -f build_L2BP.log
make -f Makefile.aeolus > build_L2BP.log

echo "software compiled"

##############################################################################
# 4. copy the executable to the install dir
##############################################################################

set EXECUTABLE = L2B_processor
if (!(-x main/${EXECUTABLE})) then
  echo "ERROR: executable main/${EXECUTABLE} is not executable,"
  echo "does not exist, or is not accessible ..."
  goto usage
endif

set OVERWRITTEN = "FALSE"
if (-e "${INSTALL_DIR}/${EXECUTABLE}") then
  set OVERWRITTEN = "TRUE"
endif

cp -f ${SOURCE_DIR}/main/${EXECUTABLE} ${INSTALL_DIR}

# make this executable usable for all users on the system
chmod a+rx ${INSTALL_DIR}/${EXECUTABLE}

if (${OVERWRITTEN} == "TRUE") then
  echo "WARNING: main program ${EXECUTABLE} overwritten in install directory"
else
  echo "main program copied to install directory"
endif

##############################################################################
# 4.1 copy other executables to the install dir
##############################################################################

# The BUFR executables cannot be included in the list below because
# they are not yet part of the main build.
#  BUFR_file_handling/L1B_ee2bufr      \
#  BUFR_file_handling/TestL1B_ee2bufr  \

foreach PATH_AND_EXECUTABLE ( \
  Scripts/L2B_ReportGenerator.py      \
  Scripts/run_L2BP.sc                 \
  Test/compare_result                 \
  Test/compare_result_hexdump         \
  Test/compare_result_reals           \
  Test/simplify_logging_lines         \
  AMD_file_handling/TestReadAMDdata   \
  AMD_file_handling/ConvertKnmiAscToAMD  \
  BUFR_file_handling/L1B_ee2bufr         \
  BUFR_file_handling/TestL1B_ee2bufr     \
  L1B_file_handling/TestReadL1Bdata      \
  L2BC_file_handling/TestReadL2BCdata    \
  L2BC_file_handling/TestWriteL2BCdata               \
  L2BC_file_handling/Write_L2BC_Text_Product         \
  L2B_AuxPar_file_handling/Test_Read_L2B_AuxPar_file \
  RayleighBrillouinProcessing/GenerateRBCdata        \
  RBC_FileHandling/TestReadRBCdata    \
  ThinLayer/TestReadJobOrderData      \
  Tools/difftool                      \
 ) \

  if (!(-x ${PATH_AND_EXECUTABLE})) then
    echo "ERROR: executable ${PATH_AND_EXECUTABLE} is not executable,"
    echo "does not exist, or is not accessible ..."
    goto usage
  endif

  #EXECUTABLE=`echo ${PATH_AND_EXECUTABLE} | awk -F/ '{print $2}'`
  set EXECUTABLE_WITHOUT_PATH = ${PATH_AND_EXECUTABLE:t}

  set OVERWRITTEN = "FALSE"
  if (-e "${INSTALL_DIR}/${EXECUTABLE_WITHOUT_PATH}") then
    set OVERWRITTEN = "TRUE"
  endif

  cp -fr ${SOURCE_DIR}/${PATH_AND_EXECUTABLE} ${INSTALL_DIR}

  # make this executable usable for all users on the system
  chmod a+rx ${INSTALL_DIR}/${EXECUTABLE_WITHOUT_PATH}

  if (${OVERWRITTEN} == "TRUE") then
    echo "WARNING: ${EXECUTABLE_WITHOUT_PATH} overwritten in install directory"
  else
    echo "${EXECUTABLE_WITHOUT_PATH} copied to install directory"
  endif

end

##############################################################################
# 4.2 make some symlinks in the install dir, for alternative command names
##############################################################################

# LTB_Level2BProcMain mentioned in SUM
ln -sf ${INSTALL_DIR}/L2B_processor ${INSTALL_DIR}/LTB_Level2BProcMain

# RGC_RGProcMain specified in TaskTable
ln -sf ${INSTALL_DIR}/L2B_ReportGenerator.py ${INSTALL_DIR}/RGC_RGProcMain

##############################################################################
# 4.3 Copy the compiled BUFR tables to a sub-directory of the install dir
##############################################################################

foreach TEST_FILE ( \
  BUFR_tables/B0000000000098015001 \
  BUFR_tables/D0000000000098015001 \
 ) \

  if (!(-e "${SOURCE_DIR}/${TEST_FILE}")) then
    echo "ERROR: the ${TEST_FILE} file does not exist or is not accessible in"
    echo "the given SOURCE_DIR. Something seems wrong with this installation..."
    goto usage
  endif

  set OVERWRITTEN = "FALSE"
  if (-e "${BUFR_TABLES_DIR}/${TEST_FILE}") then
    set OVERWRITTEN = "TRUE"
  endif

  cp -f ${SOURCE_DIR}/${TEST_FILE} ${BUFR_TABLES_DIR}

  if (${OVERWRITTEN} == "TRUE") then
    echo "WARNING: ${TEST_FILE} overwritten in install/BUFR_tables sub-directory"
  else
    echo "${TEST_FILE} copied to install/BUFR_tables sub-directory"
  endif

end

# make bufr tables readable for all users on the system
chmod a+r ${BUFR_TABLES_DIR}/*


##############################################################################
# 4.4 Copy the datafiles needed for TLE testing to 
#     a sub-directory of the install dir
##############################################################################

#  input_data/l1bv105_example02_20070309_cloudlayers/AE_TEST_ALD_U_N_1B_20080101T000001986_000084000_001421_0001.HDR \
#  input_data/l1bv105_example02_20070309_cloudlayers/AE_TEST_ALD_U_N_1B_20080101T000001986_000084000_001421_0001.DBL \

foreach TEST_FILE ( \
  input_data/l1bv106_test_12345_ref_N_20/AE_TEST_ALD_U_N_1B_20071002T001842036_003696000_000005_0001.HDR \
  input_data/l1bv106_test_12345_ref_N_20/AE_TEST_ALD_U_N_1B_20071002T001842036_003696000_000005_0001.DBL \
  input_data/l1bv105_example02_20070309_cloudlayers/AE_TEST_AUX_MET_12_20060919T000000_20060919T003000_0001.HDR     \
  input_data/l1bv105_example02_20070309_cloudlayers/AE_TEST_AUX_MET_12_20060919T000000_20060919T003000_0001.DBL     \
  RayleighBrillouinProcessing/AE_TEST_AUX_RBC_L2_20071003T123456_20081002T123456_0001.HDR                           \
  RayleighBrillouinProcessing/AE_TEST_AUX_RBC_L2_20071003T123456_20081002T123456_0001.DBL                           \
  Test/AuxClim_file_handling/AE_TEST_AUX_CLM_L2_20010312T123456_20110312T123456_0001.HDR.expected                   \
  Test/AuxClim_file_handling/AE_TEST_AUX_CLM_L2_20010312T123456_20110312T123456_0001.DBL.expected.LINUX_pgf90       \
  ThinLayer/AE_TEST_AUX_PAR_2B_20050331T000000_20111231T000000_0000.EEF \
  ThinLayer/AE_TEST_AUX_CFG_2B_20050331T000000_20111231T000000_0000.EEF \
  ThinLayer/order.5.xml                      \
  ThinLayer/TaskTable.AE_L1B_L2B_WIND.xml    \
  ThinLayer/WorkstationConfigurationFile.xml \
 ) \

  if (!(-e "${SOURCE_DIR}/${TEST_FILE}")) then
    echo "ERROR: the ${TEST_FILE} file does not exist or is not accessible in"
    echo "the given SOURCE_DIR. Something seems wrong with this installation..."
    goto usage
  endif

  set FILE_WITHOUT_PATH = ${TEST_FILE:t}

  set OVERWRITTEN = "FALSE"
  if (-e "${TLE_TEST_DIR}/${FILE_WITHOUT_PATH}") then
    set OVERWRITTEN = "TRUE"
  endif

  cp -f ${SOURCE_DIR}/${TEST_FILE} ${TLE_TEST_DIR}

  if (${OVERWRITTEN} == "TRUE") then
    echo "WARNING: ${FILE_WITHOUT_PATH} overwritten in install/TLE_testfiles sub-directory"
  else
    echo "${FILE_WITHOUT_PATH} copied to install/TLE_testfiles sub-directory"
  endif

end

# rename the 2 expected files
set AUX_CLM_BASE  = AE_TEST_AUX_CLM_L2_20010312T123456_20110312T123456_0001

set OVERWRITTEN = "FALSE"
if (-e ${TLE_TEST_DIR}/${AUX_CLM_BASE}.HDR) then
  set OVERWRITTEN = "TRUE"
endif

mv -f ${TLE_TEST_DIR}/${AUX_CLM_BASE}.HDR.expected \
      ${TLE_TEST_DIR}/${AUX_CLM_BASE}.HDR

echo "renamed: ${TLE_TEST_DIR}/${AUX_CLM_BASE}.HDR.expected"
echo "to:      ${TLE_TEST_DIR}/${AUX_CLM_BASE}.HDR"
if (${OVERWRITTEN} == "TRUE") then
  echo "WARNING: the old file with this name was overwritten!"
endif

set OVERWRITTEN = "FALSE"
if (-e ${TLE_TEST_DIR}/${AUX_CLM_BASE}.DBL) then
  set OVERWRITTEN = "TRUE"
endif

mv -f ${TLE_TEST_DIR}/${AUX_CLM_BASE}.DBL.expected.LINUX_pgf90 \
      ${TLE_TEST_DIR}/${AUX_CLM_BASE}.DBL

echo "renamed: ${TLE_TEST_DIR}/${AUX_CLM_BASE}.DBL.expected.LINUX_pgf90"
echo "to:      ${TLE_TEST_DIR}/${AUX_CLM_BASE}.DBL"
if (${OVERWRITTEN} == "TRUE") then
  echo "WARNING: the old file with this name was overwritten!"
endif

# make these test datafiles readable for all users on the system
chmod a+r ${TLE_TEST_DIR}/*

##############################################################################
# 5. run the test suite
##############################################################################

echo "running test suite"

\rm -f build_L2BP.testlog
make -f Makefile.aeolus test > build_L2BP.testlog

set npassed = `cat ${SOURCE_DIR}/Test/testresults.accumulated | grep -i passed | wc -l`
set nfailed = `cat ${SOURCE_DIR}/Test/testresults.accumulated | grep -i failed | wc -l`
set nskipped = `cat ${SOURCE_DIR}/Test/testresults.accumulated | grep -i skipped | wc -l`

echo "-------------------------------"
echo "software tests run successfully"
echo "nr of passed  tests: "${npassed}
echo "nr of failed  tests: "${nfailed}
echo "nr of skipped tests: "${nskipped}
echo "-------------------------------"
echo ""
echo "inspect the file: ${SOURCE_DIR}/Test/testresults.accumulated"
echo "for details on which tests passed successfuly, failed, or where skipped"

##############################################################################
# 6. print on "installation succeeded" message
##############################################################################

echo "software installed successfully"

# end of this script in case everything is ok
exit

##############################################################################
# usage text to be printed in case of problems
##############################################################################
usage:
  echo "usage:"
  echo "build_L2BP.sc [SOURCE_DIR] [INSTALL_DIR] [EECFI_DIR] [EE_XML_DIR] [DATAPACK]"
  echo " or "
  echo "build_L2BP.sc [SOURCE_DIR] [INSTALL_DIR] dontuse dontuse [DATAPACK]"
  echo "aborting script"
  exit
