module ReadJobOrderData
  !  #[ Description:
  !
  !     a module to read the JobOrder file defining what Job the aeolus L2B 
  !     processor should perform
  !             
  ! created:     24-Apr-2006 by J. de Kloe, KNMI
  ! 
  ! Modifications:
  !     27-Apr-2006 J. de Kloe finished Global_Configuration section
  !     12-May-2006 D. Tan     Null use of ParKind1 module
  !     07-Aug-2006 J. de Kloe increased maxlen_log_level to 8
  !                            added a check in Check_Output and added a
  !                            forgotten item in Get_Output_from_xml_tree
  !     15-Aug-2006 J. de Kloe moved datastruct definition and associated
  !                            routines to the file
  !                            DataStructures/joborder_datastructure.F90
  !     07-Sep-2006 J. de Kloe activated workaround needed for the 
  !                            ThinLayer emulator
  !     18-Oct-2007 J. de Kloe added a nullify statement
  !     13-Nov-2007 J. de Kloe added a remark on the checking of xmlns
  !
  !  #]
  !  #[ Modules used:

  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE JobOrder_DataStructure, only: JobOrderData_type,Get_JobOrderData_from_xml
  USE XML_module, only: xml_node_type, Init_XML, xml_load_file, delete_xml_tree
  USE Logging, only: logmsg,log_debug,log_info,log_progress,log_warn,log_error
  USE ErrorHandler, only: no_error
  USE LunManager, only: get_lun, free_lun
  !  #]
  !  #[ Parameter/Variable/Type declarations

  implicit none

  ! make all module contents public, unless explicitely statet otherwise
  public

  !  #]
  !  #[ Module interface
  ! all is public for now
  !  #]
contains
  !-------------------------------------------------
  subroutine ReadJobOrderFile(JobOrderData,filename_JobOrder,&
                              Task_Number, error_flag, workaround)
    !  #[

    TYPE(JobOrderData_type), intent(out) :: JobOrderData
    character(len=*),        intent(in)  :: filename_JobOrder
    integer,                 intent(out) :: Task_Number
    integer,                 intent(out) :: error_flag
    logical, optional,       intent(in)  :: workaround

    ! local variables
    TYPE(xml_node_type), pointer :: xml_tree

    ! init
    error_flag = no_error
    
    ! init the xml handling module
    call Init_XML(error_flag)
    IF (error_flag .ne. no_error) return
    
    call logmsg(log_progress,"ReadJobOrderFile: loading JobOrder xml file: "//&
                trim(filename_JobOrder))

    nullify(xml_tree)
    call xml_load_file(filename_JobOrder, xml_tree, error_flag)
    IF (error_flag .ne. no_error) return

    ! NOTE that the JobOrder files, as generated by the ThinLayer Emulator
    ! do not use the ESA Earth Explorer convention form xmlns.
    ! The files use 2 different namespaces, and specify a schemalocation
    ! of the form:
    !     xmlns:a="http://www.acsys.it/schemas/IPF" 
    !     xmlns:xsi="http://www.w3.org/2000/10/XMLSchema-instance" 
    !     xsi:schemaLocation="http://www.acsys.it/schemas/IPF JobOrder.xsd"
    ! 
    ! It is not clear to me if this will be changed, since the TLE
    ! seems to have some exceptional position anyway, and is not required 
    ! to always follow EE standards for its xml files.
    ! Therefore I have not yet activated the following xmlns check, only
    ! copied and pasted it here for possible future use.

    ! define which xml namespace should be present
    !example: xmlns="http://www.esa.int/schemas/ae/ALD_U_N_2B_01.01"
    !uri_ns_header    ! defined in xml_module
    !file_type_L2B_N  ! defined in FixedHeader
    !iodd_version_L2B ! defined in AeolusConstants
    !xml_namespace_JobOrder = trim(uri_ns_header)//trim(file_type_JobOrder)//&
    !                         "_"//trim(iodd_version_JobOrder)
    !
    ! check whether this namespace is really present in the input file
    !call xml_tree_get_namespace(xml_tree,xml_namespace_uri,error_flag)
    !if (error_flag .ne. no_error) then
    !   ! allow this attribute to be missing for now, to allow use
    !   ! of outdated files.
    !   ! TODO: remove this exception
    !   error_flag = no_error
    !   call logmsg(log_warn,"WARNING in ReadJobOrderFile: "//&
    !               "could not find xmlns attribute in the root tag "//&
    !               "of file: "//trim(filename_hdr))
    !else
    !   if (trim(xml_namespace_JobOrder) .ne. trim(xml_namespace_uri)) &
    !        call logmsg(log_warn,"WARNING in ReadJobOrderFile: "//&
    !                    "expected an xmlns attribute with content: "//&
    !                    trim(xml_namespace_JobOrder)//", but found an "//&
    !                    "xmlns attribute with content: "//&
    !                    trim(xml_namespace_uri)//". This input file may "//&
    !                    "not be compatible with the current software!")
    !end if

    call logmsg(log_debug,"ReadJobOrderFile: the xml file was "//&
                          "successfully read")

    ! copy and check xml items

    call Get_JobOrderData_from_xml(xml_tree, JobOrderData, Task_Number, &
                                   error_flag, workaround)
    IF (error_flag .ne. no_error) return

    ! cleaning up
    call delete_xml_tree(xml_tree, error_flag)
    IF (error_flag .ne. no_error) return

    call logmsg(log_progress,"ReadJobOrderFile: "//&
                             "finished loading JobOrder xml file")

    return

  end subroutine ReadJobOrderFile
    !  #]
  !-------------------------------------------------
end module ReadJobOrderData
