module diff_module
  !  #[ documentation
  !     a collection of routines used by the difftool command
  !     
  !     written by: J. de Kloe, KNMI
  !
  !     Modifications:
  !     12-May-2006 D. Tan      Null use of ParKind1 module
  !     18-Jul-2006 J. de Kloe  added reporting of linenumbers
  !     11-Aug-2006 J. de Kloe  repaired a memory leak
  !     17-Aug-2006 P. Poli     changed linenumbers* from pointers to arrays
  !     15-Sep-2006 J. de Kloe  change text in error in case file cannot
  !                             bee opened.
  !     11-Dec-2006 J. de Kloe  added -i1/--ifort1  -i2/--ifort2 switches
  !     12-Dec-2006 J. de Kloe  implemented exception for empty files
  !     19-Feb-2007 J. de Kloe  using error_diff_tool_syntax_error imported
  !                             from the errorhandler now.
  !     19-Feb-2007 J. de Kloe  added a note on a newly found known problem
  !     18-Apr-2007 J. de Kloe  added use of Compiler_Features
  !     01-Jun-2007 J. de Kloe  fixed a small bug in the linenumber print
  !     03-Aug-2007 J. de Kloe  fixed a bug due to the double call to 
  !                             compare_words, which failed when one of the
  !                             words was a REALACC/ENDREALACC command
  !                             and the other was not.
  !     19-Oct-2007 J. de Kloe  added several nullify statements to
  !                             suppress g95 warnings
  !     16-Jan-2008  J. de Kloe phase out integer kind i_ 
  !     26-Feb-2008 P. Poli     add --necsx1 switch
  !  #]
  !  #[ known problems:
  !
  !     -Comparing of complex numbers fails if the complex number
  !      contains any spaces. (this happens for example for the
  !      gfortran compiler). JK. 17-01-2006
  !      Example:
  !      most compilers print a complex number as follows:
  !           (1.00000000000000,0.500000000000000)
  !      however, gfortran currently prints it as:
  !           (  1.00000000000000     , 0.500000000000000     )
  !      ==>proposed solution:
  !         use the write_complex_number() function available
  !         in the support/stringtools.F90 module, which prints
  !         complex numbers in a uniform way on all platforms
  !
  !  #]
  !  #[ modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE ErrorHandler, only: no_error, program_abort, error_cmdline_options, &
                          error_reading_file, error_opening_file, &
                          error_programming, error_allocate, &
                          error_diff_tool_syntax_error
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE LunManager, only: get_lun, free_lun
  USE numerics,   only: r8_
  USE c_support,  only: get_filesize
  USE Compiler_Features, only: tabchar, retchar, newline

  !  #]
  !  #[ variables and parameters
  implicit none

  ! parameters
  character(len=7),  parameter :: real_command_start = "REALACC"
  character(len=10), parameter :: real_command_end   = "ENDREALACC"
    
  character, parameter :: space   = ' '

  ! a value to remember the module status
  logical, save :: module_is_initialised = .false.
  ! the default verbosity setting is Error (as silent as possible)
  integer, private, save :: LocalVerbosity = VerbosityError

  !  #]
contains
  !-----------------------------------------------------
  subroutine Init_diff_module(ChoosenVerbosity)
    !  #[
    integer, optional, intent(in) :: ChoosenVerbosity

    IF (present(ChoosenVerbosity)) LocalVerbosity = ChoosenVerbosity

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"diff_module module has been initialised"
       call print_verbosity_setting(LocalVerbosity,ModuleName="diff_module")
    END IF

    module_is_initialised = .true.

  end subroutine Init_diff_module
    !  #]
  subroutine load_file(buffer,filename,error_flag)
    !  #[
    character(len=1), dimension(:), pointer :: buffer
    character(len=*), intent(in)  :: filename
    integer,      intent(out) :: error_flag

    ! local variables
    integer :: filesize, fileunit
    integer :: status, ios

    error_flag = no_error
    
    IF (associated(buffer)) THEN
       print *,"Error in load_file():"
       print *,"the pointer buffer should be nullified when entering"
       print *,"this subroutine, but it isn't"
       error_flag = error_programming
       return
    END IF

    filesize = get_filesize(filename)
    IF (filesize .lt. 0) THEN
       print *,"could not open file: ",trim(filename)
       error_flag = error_reading_file
       return
    END IF

    IF (filesize .eq. 0) THEN
       ! exception for zero length files
       nullify(buffer)
       return
    END IF

    allocate(buffer(filesize),STAT=status)
    IF (status .ne. 0) THEN
       print *,"allocating problem in load_file():"
       print *,"could not get memory to store the complete file"
       print *,"if this file is very big, increase the limit settings ..."
       error_flag = error_allocate
       return
    END IF

    fileunit = get_lun()
    open(unit=fileunit,file=filename,status="old",&
         action="read",access="direct",form="unformatted",&
         recl=filesize,ERR=999)
    read(fileunit,rec=1,iostat=ios) buffer
    IF (ios .ne. 0) goto 998
    close(unit=fileunit)
    call free_lun(fileunit)

!    IF (LocalVerbosity .ge. VerbosityReport) &
!         print *,"buffer1 contains ",linecount1," lines"
!    END IF

    return

    ! Error handlers
998 print *,"ERROR could not read from inputfile: ",trim(filename)
    deallocate(buffer)
    error_flag = error_reading_file
    return
999 print *,"ERROR could not open inputfile: ",trim(filename)
    deallocate(buffer)
    error_flag = error_opening_file
    return

  end subroutine load_file
    !  #]
  subroutine compare_buffers_binary(buffer1,buffer2,buffers_differ,error_flag)
    !  #[
    character(len=1), dimension(:), pointer :: buffer1
    character(len=1), dimension(:), pointer :: buffer2
    logical,          intent(out) :: buffers_differ
    integer,          intent(out) :: error_flag

    ! local variables
    integer :: i

    error_flag = no_error
    buffers_differ = .false.
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"starting binary compare"

    ! this exception happens if both files have zero bytes
    IF ( .not. associated(buffer1) .and. &
         .not. associated(buffer2)       ) return

    ! this exception happens if file1 has zero bytes
    IF (.not. associated(buffer1)) THEN
       buffers_differ = .true.       
       return
    END IF

    ! this exception happens if file2 has zero bytes
    IF (.not. associated(buffer2)) THEN
       buffers_differ = .true.
       return
    END IF

    IF (size(buffer1) .ne. size(buffer2)) THEN
       buffers_differ = .true.
       return
    END IF

    DO i=1,size(buffer1)
       IF (buffer1(i) .ne. buffer2(i)) THEN
          buffers_differ = .true.
          return
       END IF
    END DO

  end subroutine compare_buffers_binary
    !  #]
  subroutine compare_buffers_word_by_word(buffer1, buffer2,&
                                          compare_ifort1, compare_ifort2,&
                                          compare_necsx1,&
                                          buffers_differ, error_flag)
    !  #[
    character(len=1), dimension(:), pointer :: buffer1
    character(len=1), dimension(:), pointer :: buffer2
    logical,          intent(in)  :: compare_ifort1
    logical,          intent(in)  :: compare_ifort2
    logical,          intent(in)  :: compare_necsx1
    logical,          intent(out) :: buffers_differ
    integer,          intent(out) :: error_flag

    ! local variables
    integer :: i,AllocStatus
    integer :: linecount1, linecount2
    integer :: size_buffer1, size_buffer2
    integer :: current_pos_buffer1, current_pos_buffer2
    integer :: line_start, line_end, len_w, len_r
    logical :: words_differ
    character(len=1), dimension(:), pointer :: word1
    character(len=1), dimension(:), pointer :: word2
    ! the next 2 are only used when compare_ifort1 is set
    character(len=1), dimension(:), pointer :: word1_remainder
    character(len=1), dimension(:), pointer :: word1_reconstructed
    ! the next 2 are only used when compare_ifort2 is set
    character(len=1), dimension(:), pointer :: word2_remainder
    character(len=1), dimension(:), pointer :: word2_reconstructed
    integer,          dimension(:), allocatable :: linenumbers1
    integer,          dimension(:), allocatable :: linenumbers2
    logical :: msg_already_printed

    ! init
    error_flag = no_error
    buffers_differ = .false.
    nullify(word1,word2)
    nullify(word1_remainder,word1_reconstructed)
    nullify(word2_remainder,word2_reconstructed)

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"starting ascii word-by-word compare"

    ! this exception happens if both files have zero bytes
    IF ( .not. associated(buffer1) .and. &
         .not. associated(buffer2)       ) return

    ! this exception happens if file1 has zero bytes
    IF (.not. associated(buffer1)) THEN
       buffers_differ = .true. 
       return
    END IF

    ! this exception happens if file2 has zero bytes
    IF (.not. associated(buffer2)) THEN
       buffers_differ = .true.
       return
    END IF

    size_buffer1 = size(buffer1)
    size_buffer2 = size(buffer2)
    ! remark: it is not obvious that the 2 buffers are of equal size !
    ! they may differ due to differences in whitespace, etc.

    allocate(linenumbers1(size_buffer1),&
             linenumbers2(size_buffer2),stat=AllocStatus)
    if (AllocStatus .ne. 0) then
       print *,"Allocation problem in compare_buffers_word_by_word()"
       error_flag = error_allocate
       return
    end if

    linecount1=1
    line_start =  1 ! only used when compare_ifort1 is set
    line_end   = -1 ! only used when compare_ifort1 is set
    DO i=1,size_buffer1
       !  #[

       ! assign a linenumber to each byte in the buffer
       ! (not very efficient, but simplest way to implement it for 
       !  the moment)
       linenumbers1(i) = linecount1
       IF (buffer1(i) .eq. newline) linecount1 = linecount1 + 1

       IF (compare_ifort1) THEN
          ! exception for lines broken at the 80th character by ifort
          ! compiled executables. Leave the newline in for this case
          ! to enable glueing the words together later
          !
          ! use the property that default formatted text output always
          ! starts with a space, and is truncated at the 80th character
          ! when compiling with ifort.....

          IF (buffer1(i) .eq. newline) THEN
             line_end =i
             IF (buffer1(line_start) .eq. ' ') THEN
                ! test print
                !print *,"line length = ",line_end-line_start+1
                IF (line_end-line_start+1 .eq. 81) THEN
                   !print *,"leaving unwanted newline for now ..."
                ELSE
                   ! for lines not equal to 80 chars remove the newline
                   IF (buffer1(i) .eq. newline) buffer1(i) = space
                END IF
             ELSE
                ! for lines not starting with a space remove the newline
                IF (buffer1(i) .eq. newline) buffer1(i) = space
             END IF
             ! reset counters for next line
             line_start =  i+1
             line_end   = -1
          END IF
       ELSE
          IF (buffer1(i) .eq. newline) buffer1(i) = space
       END IF

       ! convert all whitespaces (tabs, newlines etc) to real spaces
       IF (buffer1(i) .eq. tabchar) buffer1(i) = space
       IF (buffer1(i) .eq. retchar) buffer1(i) = space

       !  #]
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"buffer1 contains ",linecount1," lines"

    ! dito for buffer2
    linecount2=1
    line_start =  1 ! only used when compare_ifort2 is set
    line_end   = -1 ! only used when compare_ifort2 is set
    DO i=1,size_buffer2
       !  #[

       ! assign a linenumber to each byte in the buffer
       linenumbers2(i) = linecount2
       IF (buffer2(i) .eq. newline) linecount2 = linecount2 + 1

       IF (compare_ifort2) THEN
          ! exception for lines broken at the 80th character by ifort
          ! compiled executables. Leave the newline in for this case
          ! to enable glueing the words together later
          !
          ! use the property that default formatted text output always
          ! starts with a space, and is truncated at the 80th character
          ! when compiling with ifort....

          IF (buffer2(i) .eq. newline) THEN
             line_end =i
             IF (buffer2(line_start) .eq. ' ') THEN
                ! test print
                !print *,"line length = ",line_end-line_start+1
                IF (line_end-line_start+1 .eq. 81) THEN
                   !print *,"leaving unwanted newline for now ..."
                ELSE
                   ! for lines not equal to 80 chars remove the newline
                   IF (buffer2(i) .eq. newline) buffer2(i) = space
                END IF
             ELSE
                ! for lines not starting with a space remove the newline
                IF (buffer2(i) .eq. newline) buffer2(i) = space
             END IF
             ! reset counters for next line
             line_start =  i+1
             line_end   = -1
          END IF
       ELSE
          IF (buffer2(i) .eq. newline) buffer2(i) = space
       END IF

       ! convert all whitespaces (tabs, newlines etc) to real spaces
       IF (buffer2(i) .eq. tabchar) buffer2(i) = space
       IF (buffer2(i) .eq. retchar) buffer2(i) = space

       !  #]
    END DO
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"buffer2 contains ",linecount2," lines"

    ! walk through the buffers, word by word, and compare the words
    current_pos_buffer1 = 1
    current_pos_buffer2 = 1
    msg_already_printed = .false.
    word_loop: DO
       call get_next_word(buffer1,current_pos_buffer1,word1)
       call get_next_word(buffer2,current_pos_buffer2,word2)

       IF (LocalVerbosity .ge. VerbosityReport) THEN
          print *,"buffer1, word: ",trim(chararray2string(word1))
          print *,"buffer2, word: ",trim(chararray2string(word2))
       END IF

       call compare_words(chararray2string(word1),&
                          chararray2string(word2),&
                          words_differ,error_flag,silent=.true., &
                          flag_compare_necsx1=compare_necsx1)
       IF (error_flag .ne. no_error) return

!IF (words_differ) THEN
!   print *,words_differ,compare_ifort1,(word1(size(word1)) .eq. newline) 
!   print *,"ichar(word1(size(word1)-2)) = ",ichar(word1(size(word1)-2)),word1(size(word1)-2)
!   print *,"ichar(word1(size(word1)-1)) = ",ichar(word1(size(word1)-2)),word1(size(word1)-1)
!   print *,"ichar(word1(size(word1)  )) = ",ichar(word1(size(word1)  )),word1(size(word1)  )
       !END IF

       ! exception for ifort: the output might be broken at the 80th 
       ! character, so glue 2 words from buffer1 together and try again
       IF ( words_differ .and. compare_ifort1 .and. &
            (word1(size(word1)) .eq. newline)      ) THEN
          !  #[
          call get_next_word(buffer1,current_pos_buffer1,word1_remainder)
          len_w = size(word1)-1 ! -1 to skip the newline character!
          len_r = size(word1_remainder)
          allocate(word1_reconstructed(len_w+len_r))
          word1_reconstructed(1      :len_w      ) = word1(1:len_w)
          word1_reconstructed(len_w+1:len_w+len_r) = word1_remainder(1:len_r)
          IF (LocalVerbosity .ge. VerbosityReport) THEN
             print *,"word1               = ["//&
                  chararray2string(word1)//"]"
             print *,"word1_remainder     = ["//&
                  chararray2string(word1_remainder)//"]"
             print *,"word1_reconstructed = ["//&
                  chararray2string(word1_reconstructed)//"]"
          END IF

          call compare_words(chararray2string(word1_reconstructed),&
                             chararray2string(word2),&
                             words_differ,error_flag, &
                             flag_compare_necsx1=compare_necsx1)
          IF (error_flag .ne. no_error) return
          
          IF (words_differ) THEN
             msg_already_printed = .false.
          ELSE
             IF (LocalVerbosity .ge. VerbosityReport) THEN
                print *,"ifort word reconstruction seems succesfull !!!"
             END IF
          END IF
          deallocate(word1_remainder)
          deallocate(word1_reconstructed)
          !  #]
       END IF

       ! exception for ifort: the output might be broken at the 80th 
       ! character, so glue 2 words from buffer2 together and try again
       IF ( words_differ .and. compare_ifort2 .and. &
            (word2(size(word2)) .eq. newline)      ) THEN
          !  #[
          call get_next_word(buffer2,current_pos_buffer2,word2_remainder)
          len_w = size(word2)-1 ! -1 to skip the newline character!
          len_r = size(word2_remainder)
          allocate(word2_reconstructed(len_w+len_r))
          word2_reconstructed(1      :len_w      ) = word2(1:len_w)
          word2_reconstructed(len_w+1:len_w+len_r) = word2_remainder(1:len_r)
          IF (LocalVerbosity .ge. VerbosityReport) THEN
             print *,"word2               = ["//&
                  chararray2string(word2)//"]"
             print *,"word2_remainder     = ["//&
                  chararray2string(word2_remainder)//"]"
             print *,"word2_reconstructed = ["//&
                  chararray2string(word2_reconstructed)//"]"
          END IF

       call compare_words(chararray2string(word1),&
                             chararray2string(word2_reconstructed),&
                             words_differ,error_flag, &
                             flag_compare_necsx1=compare_necsx1)
          IF (error_flag .ne. no_error) return
          
          IF (words_differ) THEN
             msg_already_printed = .false.
          ELSE
             IF (LocalVerbosity .ge. VerbosityReport) THEN
                print *,"ifort word reconstruction seems succesfull !!!"
             END IF
          END IF
          deallocate(word2_remainder)
          deallocate(word2_reconstructed)
          !  #]
       END IF

       ! not that the ifort exception should not be necessary if both
       ! files are produced using ifort compiled executables, because
       ! then the linebreaks should be reproducable, and comparing should
       ! succeed anyway.

       IF (words_differ) THEN
          IF (.not. msg_already_printed) THEN

             ! compare again, but now allow to print the message
             call compare_words(chararray2string(word1),&
                          chararray2string(word2),&
                          words_differ,error_flag,&
                          dont_switch_mode=.true., &
                          flag_compare_necsx1=compare_necsx1)
             IF (error_flag .ne. no_error) return
          END IF

          print *,"words differ !!!"

          ! print linenumbers of found difference
          print *,"this difference is located at:"
          IF (current_pos_buffer1 .le. size_buffer1) THEN
             print *,"line ",linenumbers1(current_pos_buffer1)," of buffer 1"
          ELSE
             print *,"beyond the end of buffer 1"
          END IF
          IF (current_pos_buffer2 .le. size_buffer2) THEN
             print *,"line ",linenumbers2(current_pos_buffer2)," of buffer 2"
          ELSE
             print *,"beyond the end of buffer 2"
          END IF

          ! print context of found difference
          print *,"context buffer 1:"
          call print_context(buffer1, current_pos_buffer1 )
          print *,"context buffer 2:"
          call print_context(buffer2, current_pos_buffer2 )

          buffers_differ = .true.

          deallocate(word1)
          deallocate(word2)
          deallocate(linenumbers1)
          deallocate(linenumbers2)

          return
       END IF
       
       deallocate(word1)
       deallocate(word2)

       IF (current_pos_buffer1 .gt. size_buffer1) exit word_loop
       IF (current_pos_buffer2 .gt. size_buffer2) exit word_loop
    END DO word_loop

    deallocate(linenumbers1)
    deallocate(linenumbers2)

  end subroutine compare_buffers_word_by_word
    !  #]
  subroutine get_next_word(buffer,current_pos,word)
    !  #[
    character(len=1), dimension(:), pointer :: buffer ! input
    character(len=1), dimension(:), pointer :: word   ! result
    integer, intent(inout)  :: current_pos ! to be updated

    ! local variables
    integer :: i
    integer :: word_start, word_end, word_size
    integer :: size_buffer

    size_buffer = size(buffer)

    ! scan, start at current_pos, untill a non-space character is found
    i = current_pos

    ! safety catches
    IF (i .gt. size(buffer)) i = size(buffer)
    IF (i .lt. 1) i = 1

    search_word_begin: DO
       IF (buffer(i) .ne. space) exit search_word_begin ! word_start found
       IF (i .ge. size_buffer)   exit search_word_begin ! safety catch
       i=i+1
    END DO search_word_begin
    word_start = i

    ! scan, start at current_pos, untill a space character is found
    search_word_end: DO
       IF (buffer(i) .eq. space) exit search_word_end ! word_end found
       IF (i .ge. size_buffer)   exit search_word_end ! safety catch
       i=i+1
    END DO search_word_end
    word_end = i-1

    ! safety catch 
    IF (word_end .lt. word_start) word_end=word_start

    word_size = word_end-word_start+1
    allocate(word(word_size))
    word(1:word_size) = buffer(word_start:word_end)

    !print *,"inside get_next_word: word = ",chararray2string(word)

    current_pos = word_end+1

  end subroutine get_next_word
    !  #]
  subroutine get_prev_word(buffer,current_pos,word)
    !  #[
    character(len=1), dimension(:), pointer :: buffer ! input
    character(len=1), dimension(:), pointer :: word   ! result
    integer, intent(inout)  :: current_pos ! to be updated

    ! local variables
    integer :: i
    integer :: word_start, word_end, word_size

    ! scan, start at current_pos, untill a non-space character is found
    i = current_pos

    ! safety catches
    IF (i .gt. size(buffer)) i = size(buffer)
    IF (i .lt. 1) i = 1

    ! scan, start at current_pos, untill a non-space character is found
    search_word_end: DO
       IF (buffer(i) .ne. space) exit search_word_end ! word_end found
       IF (i .le. 1) exit search_word_end ! safety catch
       i=i-1
    END DO search_word_end
    word_end = i

    search_word_begin: DO
       IF (buffer(i) .eq. space) exit search_word_begin ! word_start found
       IF (i .le. 1)   exit search_word_begin ! safety catch
       i=i-1
    END DO search_word_begin
    word_start = i+1

    ! safety catch 
    IF (word_end .lt. word_start) word_end=word_start

    word_size = word_end-word_start+1
    allocate(word(word_size))
    word(1:word_size) = buffer(word_start:word_end)

    !print *,"inside get_prev_word: word = ",chararray2string(word)

    current_pos = word_start-1

  end subroutine get_prev_word
    !  #]
  subroutine compare_words(word1_input,word2_input,words_differ,error_flag,silent,&
                           dont_switch_mode,flag_compare_necsx1)
    !  #[
    character(len=*),  intent(in)  :: word1_input ! input
    character(len=*),  intent(in)  :: word2_input ! input
    logical,           intent(out) :: words_differ ! result
    integer,       intent(out) :: error_flag   ! result
    logical, optional, intent(in)  :: silent           ! optional input
    logical, optional, intent(in)  :: dont_switch_mode ! optional input
    logical, optional, intent(in)  :: flag_compare_necsx1 ! optional input

    ! local variables and parameters
    integer, parameter :: comparing_reals = 1
    integer, parameter :: comparing_text  = 2
    integer, save      :: comparing_mode = comparing_text

    integer,   save   :: acc
    real(r8_), save   :: allowed_rel_diff

    integer           :: n
    character(len=20) :: command
    logical           :: do_not_switch_mode
    logical           :: is_real_word1, is_real_word2
    logical           :: is_complex_word1, is_complex_word2

    character(len=len(word1_input)+1) :: word1_local
    character(len=len(word1_input)+1) :: word1_local_part1
    character(len=len(word1_input)+1) :: word1_local_part2
    character(len=len(word2_input)+1) :: word2_local
    character(len=len(word2_input)+1) :: word2_local_part1
    character(len=len(word2_input)+1) :: word2_local_part2

    ! init
    words_differ      = .false.
    error_flag        = no_error

    ! dont_switch_mode should be set when this compare_words() routine
    ! is called twice on the same pair of words 
    ! (for example first in silent mode, then in non-silent mode).
    do_not_switch_mode = .false.
    if (present(dont_switch_mode)) &
         do_not_switch_mode = dont_switch_mode

    command = word1_input
    n = len_trim(word1_input)
    IF (n .ge. 7) THEN
       IF (word1_input(1:7) .eq. real_command_start) &
            command = real_command_start
    END IF
    IF (n .ge. 10) THEN
       IF (word1_input(1:10) .eq. real_command_end) &
            command = real_command_end
    END IF

    IF (.not. do_not_switch_mode) THEN
       select case (comparing_mode)
       case(comparing_text)
          !  #[
          select case (command)
          case(real_command_start) 
             IF (LocalVerbosity .ge. VerbosityReport) &
                  print *,"Switching to comparing reals"
             comparing_mode = comparing_reals
             ! get the accuracy
             call get_real_acc(word1_input,acc,error_flag)
             IF (error_flag .ne. no_error) return
             
             allowed_rel_diff = 1./(10.**acc)
             IF (LocalVerbosity .ge. VerbosityReport) &
                  print *,"REALACC = ",acc," Allowed_rel_diff = ",&
                          allowed_rel_diff
             
          case(real_command_end) 
             print *,"ERROR in compare_words()"
             print *,"the command ",real_command_end," should always follow"
             print *,"a ",real_command_start," but the order seems wrong here"
             print *,"word1 = ",trim(word1_input)
             print *,"word2 = ",trim(word2_input)
             error_flag = error_diff_tool_syntax_error
             return
          end select
          
          !  #]
       case(comparing_reals)
          !  #[
          select case (command)
          case(real_command_start) 
             print *,"ERROR in compare_words()"
             print *,"the command ",real_command_start," should always be"
             print *,"terminated by a ",real_command_end," command,"
             print *,"but the order seems wrong here"
             print *,"word1 = ",trim(word1_input)
             print *,"word2 = ",trim(word2_input)
             error_flag = error_diff_tool_syntax_error
             return
          case(real_command_end) 
             comparing_mode = comparing_text
             IF (LocalVerbosity .ge. VerbosityReport) &
                  print *,"Switching to comparing text"
          end select
          !  #]
       case default
          !  #[ this should never happen !
          print *,"ERROR in compare_words(): impossible case !!!"
          error_flag = error_programming
          return
          !  #]
       end select
    END IF

    select case (comparing_mode)
    case(comparing_text)
       call textcompare_words(word1_input,word2_input,words_differ,silent)
    case(comparing_reals)
       call determine_is_real(word1_input,word1_local,is_real_word1,flag_compare_necsx1)
       call determine_is_real(word2_input,word2_local,is_real_word2)
       IF (is_real_word1 .and. is_real_word2) THEN
          ! comparing the words as real numbers, allowing differences
          ! in real numbers upto the specified accuracy
          call realcompare_words(word1_local,word2_local,allowed_rel_diff,acc,&
                                 words_differ,silent)
       ELSE 
          ! check if we have a complex number
          call determine_is_complex(word1_input,word1_local_part1,word1_local_part2,is_complex_word1,flag_compare_necsx1)
          call determine_is_complex(word2_input,word2_local_part1,word2_local_part2,is_complex_word2)
          IF (is_complex_word1 .and. is_complex_word2) THEN
             !print *,"complex number found !!!"
             ! compare the real parts
             call realcompare_words(word1_local_part1,word2_local_part1,&
                                    allowed_rel_diff,acc,words_differ,silent)
             IF (words_differ) return
             ! compare the imaginary parts
             call realcompare_words(word1_local_part2,word2_local_part2,&
                                    allowed_rel_diff,acc,words_differ,silent)
             IF (words_differ) return
          ELSE
             ! if the words are not reals, and also not complex numbers
             ! fall back to normal text comparison
             call textcompare_words(word1_input,word2_input,words_differ,silent)
          END IF
       END IF
    end select

  end subroutine compare_words
    !  #]
  subroutine textcompare_words(word1,word2,words_differ,silent)
    !  #[ just compare the words as text
    character(len=*), intent(in)  :: word1 ! input
    character(len=*), intent(in)  :: word2 ! input
    logical,          intent(out) :: words_differ ! result
    logical, optional, intent(in)  :: silent ! optional input

    ! local variable
    logical :: be_silent
    integer :: len_w1, len_w2

    be_silent = .false.
    IF (present(silent)) be_silent = silent

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"textcompare for:"
       print *,"     word1 = ["//word1//"]"
       print *," and word2 = ["//word2//"]"
    END IF

    ! a newline is possible because of the new ifort reconstruction code
    len_w1 = len(word1)
    IF (word1(len_w1:len_w1) .eq. newline) len_w1 = len_w1-1
    len_w2 = len(word2)
    IF (word2(len_w2:len_w2) .eq. newline) len_w2 = len_w2-1

    words_differ      = .false.
    IF (word1(1:len_w1) .ne. word2(1:len_w2)) THEN
       IF (.not. be_silent) THEN
          print *,"difference found while comparing "
          print *,"words: ["//word1(1:len_w1)//"]"
          print *,"  and: ["//word2(1:len_w2)//"]"
       END IF
       words_differ = .true.
       return
    END IF

  end subroutine textcompare_words
  !  #]
  subroutine realcompare_words(word1,word2,allowed_rel_diff,acc,&
                               words_differ,silent)
    !  #[ compare the words as reals
    character(len=*),  intent(in)  :: word1 ! input
    character(len=*),  intent(in)  :: word2 ! input
    real(r8_),         intent(in)  :: allowed_rel_diff
    integer,           intent(in)  :: acc
    logical,           intent(out) :: words_differ ! result
    logical, optional, intent(in)  :: silent ! optional input

    ! local variables
    logical :: be_silent

    ! exceptions to be able to compare inf for one compiler
    ! with huge() for another compiler
    logical          :: nr1_ge_maxreal
    logical          :: nr2_ge_maxreal
    logical          :: nr1_le_minreal
    logical          :: nr2_le_minreal

    real(r8_)        :: nr1, nr2, max_abs_val, allowed_abs_diff
    character(len=2) :: acc_str
    integer          :: ios

    be_silent = .false.
    IF (present(silent)) be_silent = silent

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"realcompare for word1 = ",word1," and word2 = ",word2

    nr1_ge_maxreal = .false.
    nr2_ge_maxreal = .false.
    nr1_le_minreal = .false.
    nr2_le_minreal = .false.
    
    read(word1,*,iostat=ios) nr1
    
    ! catch the read conversion errors in word1
    IF (ios .ne. 0) THEN
       IF (LocalVerbosity .ge. VerbosityReport) THEN
          print *,"problem interpreting word1, maybe it is out of range?"
          print *,"nr1 = ",nr1," word1 = ",word1
          IF (word1(1:1) .eq. '-') THEN
             print *,"setting this number to -huge()"
          ELSE
             print *,"setting this number to huge()"
          END IF
       END IF
       IF (word1(1:1) .eq. '-') THEN
          nr1 = -huge(nr1)
       ELSE
          nr1 = huge(nr1)
       END IF
    END IF
    
    read(word2,*,iostat=ios) nr2
    
    ! catch the read conversion errors in word2
    IF (ios .ne. 0) THEN
       IF (LocalVerbosity .ge. VerbosityReport) THEN
          print *,"problem interpreting word2, maybe it is out of range?"
          print *,"testjos: nr2 = ",nr2," word2 = ",word2
          print *,"setting this number to huge()"
          IF (word2(1:1) .eq. '-') THEN
             print *,"setting this number to -huge()"
          ELSE
             print *,"setting this number to huge()"
          END IF
       END IF
       IF (word2(1:1) .eq. '-') THEN
          nr2 = -huge(nr2)
       ELSE
          nr2 = huge(nr2)
       END IF
    END IF
    
    IF (nr1 .ge.  huge(nr1) ) nr1_ge_maxreal = .true.
    IF (nr2 .ge.  huge(nr2) ) nr2_ge_maxreal = .true.
    IF (nr1 .le. -huge(nr1) ) nr1_le_minreal = .true.
    IF (nr2 .le. -huge(nr2) ) nr2_le_minreal = .true.
    
    IF ( nr1_ge_maxreal .and. nr2_ge_maxreal) THEN
       IF (LocalVerbosity .ge. VerbosityReport) THEN
          print *,"nr1 = ",nr1," word1 = ",word1
          print *,"nr2 = ",nr2," word2 = ",word2
          print *,"these huge/inf numbers are assumed to be equal"
       ENDIF
       return
    ENDIF
    
    IF ( nr1_le_minreal .and. nr2_le_minreal) THEN
       IF (LocalVerbosity .ge. VerbosityReport) THEN
          print *,"nr1 = ",nr1," word1 = ",word1
          print *,"nr2 = ",nr2," word2 = ",word2
          print *,"these -huge/-inf numbers are assumed to be equal"
       END IF
       return
    ENDIF
    
    max_abs_val = max(abs(nr1),abs(nr2))
          
    ! exception in case of zero, because a relative 
    ! diff makes no sence in that case
    IF ( (nr1 .eq. 0.0_r8_) .or. (nr2 .eq. 0.0_r8_) ) THEN
       allowed_abs_diff = allowed_rel_diff*1.0_r8_
    ELSE
       allowed_abs_diff = allowed_rel_diff*max_abs_val
    END IF
    
    IF (abs(nr1-nr2) .gt. allowed_abs_diff) THEN
       IF (.not. be_silent) THEN
       print *,"difference found while ",&
            "comparing reals: [",trim(word1),&
            "] and [",trim(word2),"]"
       write(acc_str,"(i2.2)") acc
       print *,"REALACC(",acc_str,") allowed_rel_diff = ",allowed_rel_diff,&
            " ENDREALACC"
       print *,"REALACC(",acc_str,") allowed_abs_diff = ",allowed_abs_diff,&
            " ENDREALACC"
       print *,"REALACC(",acc_str,") nr1 = ",nr1," nr2 = ",nr2," ENDREALACC"
       END IF
       words_differ=.true.
       return
    END IF
    
  end subroutine realcompare_words
  !  #]
  subroutine print_context(buffer, current_pos_buffer)
    !  #[ 
    character(len=1), dimension(:), pointer    :: buffer             ! input
    integer,                        intent(in) :: current_pos_buffer ! input

    ! local variables
    integer, parameter :: nr_context_words = 5
    integer :: copied_current_pos
    integer :: context_start
    integer :: context_end
    integer :: w, size_buffer
    character(len=1), dimension(:), pointer :: word   ! result

    nullify(word)

    copied_current_pos = current_pos_buffer
    size_buffer        = size(buffer)

    ! loop back a given nr of words
    backwardloop: DO w=1,nr_context_words
       call get_prev_word(buffer,copied_current_pos,word)
       !print *,"word ",w," : [",chararray2string(word),"]"
       deallocate(word)
       IF (copied_current_pos .le. 1) exit backwardloop
    END DO backwardloop
    context_start = copied_current_pos
    IF (context_start .lt. 1) context_start = 1

    copied_current_pos = current_pos_buffer
    ! loop forward a given nr of words
    forwardloop: DO w=1,nr_context_words
       call get_next_word(buffer,copied_current_pos,word)
       !print *,"word ",w," : [",chararray2string(word),"]"
       deallocate(word)
       IF (copied_current_pos .ge. size_buffer) exit forwardloop
    END DO forwardloop
    context_end = copied_current_pos
    IF (context_end .gt. size_buffer) context_end = size_buffer

    print *,chararray2string(buffer(context_start:context_end))

  end subroutine print_context
    !  #]
  function chararray2string(chararray,remove_newline) result(text)
    !  #[
    character(len=1), dimension(:) :: chararray      ! input
    character(len=size(chararray)) :: text           ! output
    logical, optional, intent(in)  :: remove_newline ! optional input

    integer :: n,i
    n = size(chararray)
    text = transfer(chararray,text)

    IF (present(remove_newline)) THEN
       IF (remove_newline) THEN
          DO i=1,n
             IF (text(i:i) .eq. newline) text(i:i) = space
          END DO
       END IF
    END IF

  end function chararray2string
    !  #]
  subroutine determine_is_real(word_input,word_output,string_is_a_real,flag_necsx)
    !  #[

    character(len=*),  intent(in)  :: word_input
    character(len=len(word_input)+1),  intent(out) :: word_output
    logical,           intent(out) :: string_is_a_real
    logical, optional, intent(in)  :: flag_necsx ! optional input

    ! local parameters
    character(len=*), parameter :: allowed_real_characters = &
         "0123456789-+Ee."
    character(len=*), parameter :: allowed_real_characters_necsx = &
         "0123456789-+ED."
     character(len=*), parameter :: allowed_digits = &
         "0123456789"

    ! local variable
    integer :: result
    integer :: pos_E1, pos_E2, pos_min1, pos_min2, pos_plus1, pos_plus2
    integer :: pos_D1, pos_D2, pos_sign1, pos_sign2
    logical :: do_necsx
    character(len=len(word_input)) :: word_copy

    do_necsx=.FALSE.
    if (present(flag_necsx)) do_necsx=flag_necsx

    ! assume we have a real number
    string_is_a_real = .true.
    word_output(:)=" "
    word_output(1:len_trim(word_input))=word_input(1:len_trim(word_input))

    ! first see if we have any characters that cannot be part of a real
    IF (do_necsx) THEN
       result = verify(trim(word_output),allowed_real_characters_necsx)
    ELSE
       result = verify(trim(word_output),allowed_real_characters)
    ENDIF
    IF (result .ne. 0) string_is_a_real = .false.

    ! then see if there is a digit in it, to catch cases
    ! like a string of dashes: "-----" to be interpreted as a number
    result =  scan(trim(word_output),allowed_digits)
    IF (result .eq. 0) string_is_a_real = .false.

    ! for the NEC-SX:
    ! - replace exponent indicator "D" by "E"
    ! - add exponent indicator "E" if there isn't any before the exponent digits
    IF (do_necsx) THEN
       pos_D1 = scan(trim(word_output),"D")
       pos_D2 = scan(trim(word_output),"D",back=.true.)
       ! see if a "D" character is present
       IF (pos_D1 .ne. 0) THEN
          ! multiple "D" chars seem present
          IF (pos_D1 .ne. pos_D2) THEN
             string_is_a_real = .false.
          ELSE ! replace "D" by "E"
             word_output(pos_D1:pos_D1)="E"
          ENDIF
       ENDIF
       pos_E1 = scan(trim(word_output),"E")
       pos_E2 = scan(trim(word_output),"E",back=.true.)
       ! see if a "E" character is present
       IF (pos_E1 .ne. 0) THEN
          ! multiple "E" chars seem present
          IF (pos_E1 .ne. pos_E2) string_is_a_real = .false.
       ELSEIF (len_trim(word_output) > 1) THEN ! there is no "E"
          pos_sign1 = scan(word_output(2:len_trim(word_output)),"+-")
          pos_sign2 = scan(word_output(2:len_trim(word_output)),"+-",back=.true.)
          ! see if a + or a - sign is present
          IF (pos_sign1 .ne. 0) THEN
             ! multiple signs seem present
             IF (pos_sign1 .ne. pos_sign2) THEN
                string_is_a_real = .false.
             ELSE ! one sign is present, and "E" is missing: insert "E"
                word_copy(:)=" "
                word_copy=trim(word_output)
                word_output(:)=" "
                word_output(1:pos_sign1)=word_copy(1:pos_sign1)
                word_output(pos_sign1+1:pos_sign1+1)='E'
                word_output(pos_sign1+2:len_trim(word_copy)+1)=word_copy(pos_sign1+1:len_trim(word_copy))
             ENDIF
          ENDIF
       ENDIF
    ENDIF

    ! if a -/+ character is present,
    ! then make sure it is either the first character,
    ! or it follows directly the e/E character.
    IF (scan(trim(word_output),"-+") .ne. 0) THEN
       pos_plus1 = scan(trim(word_output),"+")
       pos_plus2 = scan(trim(word_output),"+",back=.true.)
       pos_min1  = scan(trim(word_output),"-")
       pos_min2  = scan(trim(word_output),"-",back=.true.)
       pos_E1    = scan(trim(word_output),"eE")
       pos_E2    = scan(trim(word_output),"eE",back=.true.)
       !print *,"pos_E1/2    = ",pos_E1,   pos_E2
       !print *,"pos_plus1/2 = ",pos_plus1,pos_plus2
       !print *,"pos_min1/2  = ",pos_min1, pos_min2

       ! see if an "E" character is present
       IF (pos_E1 .ne. 0) THEN
          ! multiple "E" chars seem present
          IF (pos_E1 .ne. pos_E2) string_is_a_real = .false.
       END IF

       ! a "+" character is present
       IF (pos_plus1 .ne. 0) THEN
          !  #[ see if it is in a valid position

          IF (LocalVerbosity .ge. VerbosityReport) &
               print *,"a '+' sign seems to be present inside this string"

          IF (pos_plus1 .eq. pos_plus2) THEN
             ! just one "+" is present
             IF (pos_plus1 .ne. 1) THEN
                IF (pos_E1 .eq. 0) THEN
                   ! no "E" char present, so no valid number
                   string_is_a_real = .false.
                ELSE
                   IF (pos_plus1 .ne. pos_E1+1) THEN
                      ! the "+" char does not follow the "E" char
                      ! so no valid number
                       string_is_a_real = .false.
                   END IF
                END IF
             END IF
          ELSE
             ! at least 2 "+" chars are present
             IF (pos_plus1 .ne. 1)  string_is_a_real = .false.

             IF (pos_E1 .eq. 0) THEN
                ! no "E" char present, so no valid number
                string_is_a_real = .false.
             ELSE
                IF (pos_plus2 .ne. pos_E1+1) THEN
                   ! the "+" char does not follow the "E" char
                   ! so no valid number
                   string_is_a_real = .false.
                END IF
                ! at this point I could check if there is a third "E"
                ! in the string (not yet implemented)
             END IF
             
          END IF
          !  #]
       END IF

       ! a "-" character is present
       IF (pos_min1 .ne. 0) THEN
          !  #[ see if it is in a valid position

          IF (LocalVerbosity .ge. VerbosityReport) &
               print *,"a '-' sign seems to be present inside this string"

          IF (pos_min1 .eq. pos_min2) THEN
             ! just one "-" is present
             IF (pos_min1 .ne. 1) THEN
                IF (pos_E1 .eq. 0) THEN
                   ! no "E" char present, so no valid number
                   string_is_a_real = .false.
                ELSE
                   IF (pos_min1 .ne. pos_E1+1) THEN
                      ! the "-" char does not follow the "E" char
                      ! so no valid number
                       string_is_a_real = .false.
                   END IF
                END IF
             END IF
          ELSE
             ! at least 2 "-" chars are present
             IF (pos_min1 .ne. 1)  string_is_a_real = .false.

             IF (pos_E1 .eq. 0) THEN
                ! no "E" char present, so no valid number
                string_is_a_real = .false.
             ELSE
                IF (pos_min2 .ne. pos_E1+1) THEN
                   ! the "-" char does not follow the "E" char
                   ! so no valid number
                   string_is_a_real = .false.
                END IF
                ! at this point I could check if there is a third "E"
                ! in the string (not yet implemented)
             END IF
             
          END IF
          !  #]
       END IF

    END IF

  end subroutine determine_is_real

    !  #]
  subroutine determine_is_complex(word_input,word_output_part1,word_output_part2, &
                                  string_is_a_complex_nr,flag_necsx)
    !  #[
    character(len=*),  intent(in)  :: word_input
    character(len=len(word_input)+1),  intent(out) :: word_output_part1
    character(len=len(word_input)+1),  intent(out) :: word_output_part2
    logical,           intent(out) :: string_is_a_complex_nr
    logical, optional, intent(in)  :: flag_necsx ! optional input

    ! REMARK: complex real numbers seem to be printed by default
    !         as (real,real) without any intermediate spaces

    ! local parameter
    character(len=*), parameter :: allowed_cmplx_characters = &
         "0123456789-+Ee.(,)"
    character(len=*), parameter :: allowed_cmplx_characters_necsx = &
         "0123456789-+ED.(,)"

    ! local variable
    integer                  :: result
    character(len=len(word_input)) :: word_part1
    character(len=len(word_input)) :: word_part2
    logical                  :: succes
    logical                  :: is_real_part1, is_real_part2
    logical                  :: do_necsx

    !print *,"testjos inside is_complex()"
    !print *,"complex test for word = ",word

    do_necsx=.FALSE.
    if (present(flag_necsx)) do_necsx=flag_necsx

    ! first inspect the whole word
    string_is_a_complex_nr = .true.
    IF (do_necsx) THEN
       result = verify(trim(word_input),allowed_cmplx_characters_necsx)
    ELSE
       result = verify(trim(word_input),allowed_cmplx_characters)
    ENDIF

    IF (result .ne. 0) string_is_a_complex_nr = .false.

    call split_complex_word(word_input,word_part1,word_part2,succes)
    IF (.not. succes)  then
       string_is_a_complex_nr = .false.
       return
    end if

    call determine_is_real(word_part1,word_output_part1,is_real_part1,flag_necsx)
    call determine_is_real(word_part2,word_output_part2,is_real_part2,flag_necsx)
    if (.not. is_real_part1 .or. .not. is_real_part2) then
       string_is_a_complex_nr = .false.
       return
    end if

  end subroutine determine_is_complex
    !  #]
  subroutine split_complex_word(word,part1,part2,succes)
    !  #[
    character(len=*),         intent(in)  :: word
    character(len=len(word)), intent(out) :: part1
    character(len=len(word)), intent(out) :: part2
    logical,                  intent(out) :: succes

    ! local variables
    integer :: pos_open_bracket
    integer :: pos_close_bracket
    integer :: pos_comma

    ! init 
    succes = .true.
    part1(:) = ' '
    part2(:) = ' '

    ! then inspect the 2 real parts
    pos_open_bracket  = scan(word,'(')
    pos_comma         = scan(word,',')
    pos_close_bracket = scan(word,')')

    ! check for the presence and position of the 2 brackets. 
    if ( (pos_open_bracket .ne. 1)  .or. &
         (pos_close_bracket .ne. len_trim(word)) ) then
       !print *,"bracket test failed"
       succes = .false.
       return
    end if

    part1 = word(pos_open_bracket+1:pos_comma-1)
    part2 = word(pos_comma+1:pos_close_bracket-1)

    !print *,"part1 = ",trim(part1)
    !print *,"part2 = ",trim(part2)

  end subroutine split_complex_word
    !  #]
  subroutine get_real_acc(word,acc,error_flag)
    !  #[
    character(len=*), intent(in)  :: word
    integer,          intent(out) :: acc
    integer,          intent(out) :: error_flag

    ! local variable
    integer :: ios

    error_flag = no_error
    acc = -1

    ! check if we really have a real_command in this word
    IF (word(1:7) .ne. real_command_start) THEN    
       error_flag = error_programming
       return
    END IF

    ! this word starts with the string real_command
    ! so at position 8 we expect the "(" character
    IF (word(8:8) .ne. '(') THEN
       error_flag = error_diff_tool_syntax_error
       return
    END IF
    
    IF (word(10:10) .eq. ')') THEN
       ! acc is in digit 9
       read(word(9:9),"(i1)",iostat=ios) acc
       IF (ios .ne. 0) goto 999
    ELSE
       IF (word(11:11) .eq. ')') THEN
          ! acc is in digits 9 and 10
          read(word(9:10),"(i2)",iostat=ios) acc
       IF (ios .ne. 0) goto 999
       ELSE
          error_flag = error_diff_tool_syntax_error
          return
       END IF
    END IF

    return

999 print *,"ERROR in get_real_acc()"
    print *,"could not read an integer from string: ",trim(word)
    error_flag = error_diff_tool_syntax_error
    return

  end subroutine get_real_acc
    !  #]
  !-----------------------------------------------------
end module diff_module
