program Difftool
  !  #[ Documentation
  ! 
  !     written by: J. de Kloe, KNMI
  !
  !     Modifications:
  !     12-May-2006  D. Tan       Null use of ParKind1 module
  !     11-Aug-2006  J. de Kloe removed the stop statement at the end 
  !                             and deallocate the buffers when done
  !     11-Dec-2006  J. de Kloe added -i1/--ifort1  -i2/--ifort2 switches
  !     19-Feb-2007  J. de Kloe added a forgotten error check
  !     16-Jan-2008  J. de Kloe phase out integer kind i_ 
  !     26-Feb-2008  P. Poli    added -n1/--necsx1
  !
  ! a program that basically compares 2 files, and
  ! only prints a warning in case they are not found to be equal.
  !
  ! Contrary to the default unix/linux diff command, comparing 
  ! is not only possible on the text level, but special commands
  ! can be inserted to request numerical (real) comparisons with 
  ! a given accuracy.
  !
  ! example:
  ! file1: 1.23456
  ! file2: 1.23455
  ! 
  ! the default diff would find these 2 numbers different, which makes
  ! it rather difficult to compare files produced with different compilers,
  ! or on different platforms (with different numerical implementation)
  !
  ! With this difftool you can write the numerical output as follows
  !
  ! example:
  ! file1: REALACC(4): 1.23456 ENDREALACC
  ! file2: REALACC(4): 1.23455 ENDREALACC
  !
  ! the diff tool will then use a precision of 1:10^-4 when comparing
  ! the numbers, so a difference of 0.0001 or smaller is discarded.
  ! Thus, the numbers in the given example will be found to be equal
  ! even if they differ by 1 in the last digit.
  ! 
  ! The difftool will start interpreting numbers as reals after the REALACC 
  ! keyword, and stop doing so after the ENDREALACC keyword.
  !
  ! Furthermore, the tool discards all linebreaks.
  ! This is usefull because (especially with default print statements
  ! like print *,a,b,c,d,e and write *,* a,b,c,d,e) the output might be
  ! continued on a next line if the output it too long, but the exact point
  ! at which the linebreak occurs is system/compiler dependent. So if you
  ! wish to compare numerical output from different systems/compilers
  ! you want to get rid of these unpredictable linebreaks. 
  !
  !  #]
  !  #[ modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE diff_module, only: Init_diff_module, &
                         load_file, compare_buffers_binary, &
                         compare_buffers_word_by_word
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE ErrorHandler, only: no_error, program_abort, error_cmdline_options
  USE Compiler_Features, only: iargc_aeolus,getarg_aeolus
  !  #]
  !  #[ Variables
  implicit none

  integer :: error_flag, LocalVerbosity
  integer :: nr_of_arguments, current_argument_nr
  character(len=256) :: current_argument
  character(len=256) :: filename1
  character(len=256) :: filename2
  logical :: filename1_is_defined
  logical :: filename2_is_defined
  logical :: compare_binary
  logical :: compare_ifort1
  logical :: compare_ifort2
  logical :: compare_necsx1
  logical :: buffers_differ
  character(len=1), dimension(:), pointer :: buffer1
  character(len=1), dimension(:), pointer :: buffer2
  !  #]
  !  #[ Initialise
  nullify(buffer1)
  nullify(buffer2)
  compare_binary = .false.
  compare_ifort1 = .false.
  compare_ifort2 = .false.
  compare_necsx1 = .false.

  filename1_is_defined = .false.
  filename2_is_defined = .false.
  filename1 = "[undefined]"
  filename2 = "[undefined]"
  LocalVerbosity = VerbosityError

  call Init_diff_module(ChoosenVerbosity=VerbosityError)
  !  #]
  !  #[ Command line handling
  nr_of_arguments = iargc_aeolus()
  current_argument_nr = 0
  DO WHILE (current_argument_nr .lt. nr_of_arguments)
     !  #[ process the given commandline arguments
     current_argument_nr = current_argument_nr+1
     call getarg_aeolus(current_argument_nr,current_argument)
     SELECT CASE (trim(current_argument))
     CASE ("-b","--binary")
        IF (LocalVerbosity .ge. VerbosityReport) &
             print *,"assume files are binary, so compare byte by byte"
        compare_binary = .true.
     CASE ("-i1","--ifort1")
        IF (LocalVerbosity .ge. VerbosityReport) &
             print *,"experimental workaround for output by ifort compiled "//&
                     "executables for file 1 activated ...."
        compare_ifort1  = .true.
     CASE ("-i2","--ifort2")
        IF (LocalVerbosity .ge. VerbosityReport) &
             print *,"experimental workaround for output by ifort compiled "//&
                     "executables for file 2 activated...."
        compare_ifort2  = .true.
     CASE ("-n1","--necsx1")
        IF (LocalVerbosity .ge. VerbosityReport) &
             print *,"experimental workaround for output on NEC-SX "//&
                     "executables for file 1 activated ...."
        compare_necsx1  = .true.
     CASE ("-v","--verbose")
        LocalVerbosity = VerbosityReport
        call Init_diff_module(ChoosenVerbosity=VerbosityReport)
     CASE ("-h","--help")
        CALL print_usage;
        stop
     CASE DEFAULT
        IF (current_argument(1:1) .eq. '-') THEN
           CALL print_usage;
           print *,"============================================"
           print *,"unrecognised commandline option: ",trim(current_argument)
           print *,"============================================"
           call program_abort(error_cmdline_options,"difftool")
        ELSE ! read the input filenames
           IF (.not. filename1_is_defined) THEN
              filename1 = current_argument
              filename1_is_defined = .true.
           ELSE
              IF (.not. filename2_is_defined) THEN
                 filename2 = current_argument
                 filename2_is_defined = .true.
              ELSE
                 print *,"ERROR: you seem to have specified 3 filenames"
                 print *,"after the difftool command, in stead of 2..."
                  call program_abort(error_cmdline_options,"difftool")
              END IF
           END IF
        END IF
     END SELECT
     !  #]
  END DO
  !  #]
  !  #[ Load the files and compare them
  IF (LocalVerbosity .ge. VerbosityReport) THEN
     print *,"Starting difftool:"
     print *,"file1: ",trim(filename1)
     print *,"file2: ",trim(filename2)
  END IF

  call load_file(buffer1,filename1,error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"load_file")
  call load_file(buffer2,filename2,error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"load_file")

  IF (LocalVerbosity .ge. VerbosityReport) &
       print *,"TEST: files succesfully loaded"

  IF (compare_binary) THEN
     call compare_buffers_binary(buffer1, buffer2,&
                                 buffers_differ, error_flag)
     IF (error_flag .ne. no_error) &
          call program_abort(error_flag,"compare_buffers_binary")
  ELSE
     call compare_buffers_word_by_word(buffer1, buffer2,&
                                       compare_ifort1, compare_ifort2,&
                                       compare_necsx1,&
                                       buffers_differ, error_flag)
     IF (error_flag .ne. no_error) &
          call program_abort(error_flag,"compare_buffers_word_by_word")
  END IF

  IF (buffers_differ) THEN
     print *,"files seem different"
  ELSE
     IF (LocalVerbosity .ge. VerbosityReport) &
          print *,"files are identical"
  END IF

  IF (associated(buffer1)) deallocate(buffer1)
  IF (associated(buffer2)) deallocate(buffer2)
  !  #]
end program Difftool

subroutine print_usage
  !  #[

  implicit none

  print *,""
  print *,"Usage of the difftool program: "
  print *,""
  print *,"difftool [options] file1 file2"
  print *,""
  print *,"the optional [options] may be:"  
  print *,"   [-b/--binary    ]"
  print *,"   [-i1/--ifort1   ]"
  print *,"   [-i2/--ifort2   ]"
  print *,"   [-n1/--necsx1   ]"
  print *,"   [-v/--verbose   ]"
  print *,"   [-h/--help      ]"
  print *,""
  !  #]
end subroutine print_usage

