module DummyAuxiliaryModule
  !  #[ Description:
  !     A module to
  !     -temporary replace the module at ECMWF
  !      which retrieves the real NWP and auxilliary data
  !     OR
  !     -to serve as a template for a special local module to be
  !      defined by any MET-center that wishes to run the L2B processor 
  !      using their own model data.
  !  #]
  !  #[ Modifications
  !       Written by: J. de Kloe, 01-04-2005
  !
  !     Modified:
  !      J. de Kloe  11-Jun-2006 added the new DateTime type
  !      D. Tan      12-May-2006 Null use of ParKind1 module
  !      J. de Kloe  26-Oct-2007 made DateTimeType public to allow compilation
  !                              with g95 -std=f95
  !      16-Jan-2008 J. de Kloe  phase out integer kind i_ and real kind r_
  !
  !  #]
  !  #[ Modules used:
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  ! define the datatypes needed for the ee_cfi fileformat
  USE EE_CFI_Datatypes, only: DateTimeType,IncrementDateTime,TestDateTime
  ! define the default integer and real to be used by the module
  USE Numerics, only: r8_
  ! define the date and time types
!  USE DateTimeMod, only: date_type, time_type, increment_datetime, &
!                         test_date, test_time
  USE ErrorHandler, only: no_error, error_allocate
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  !  #]
  !  #[ Parameter/Variable/Type declarations
  implicit none

  ! make all module contents private, unless explicitely stated otherwise
  private

  TYPE profile_type
     real(r8_)                        :: lat
     real(r8_)                        :: lon
     TYPE(DateTimeType)               :: datetime
     integer                          :: nlevels
     ! will we define z compared to the geoid ? In that way we can
     ! just copy it from the model output (if I remember correctly).
     real(r8_), dimension(:), pointer :: z
     real(r8_), dimension(:), pointer :: T_nom
     real(r8_), dimension(:), pointer :: P_nom
     real(r8_)                        :: geoid_ground_level
     real(r8_)                        :: ellipsoid_ground_level
  END TYPE profile_type

  ! the default verbosity setting is Err (as silent as possible)
  integer, private, save :: LocalVerbosity = VerbosityError

  !  #]
  !  #[ Module interface
  ! specify which routines are public, so are part of the interface
  ! to this module
  public :: InitAuxiliaryModule, GetAuxiliaryData, &
            TestPrintAuxiliaryData, DeallocAuxiliaryData,&
            profile_type,DateTimeType
  !  #]
contains
  !------------------------------------------
  subroutine InitAuxiliaryModule(ChoosenVerbosity)
    !  #[
    integer, optional, intent(in) :: ChoosenVerbosity

    IF (present(ChoosenVerbosity)) LocalVerbosity = ChoosenVerbosity

    call print_verbosity_setting(LocalVerbosity,ModuleName="AuxiliaryModule")

  end subroutine InitAuxiliaryModule
    !  #]
  subroutine GetAuxiliaryData( lat1,lon1,datetime,  & ! input
                               lat2,lon2,timestep,  & ! input
                               nprofiles,profiles,  & ! output
                               error_flag            )! output
    !  #[
    ! inputs
    real(r8_),          intent(in) :: lat1, lon1 ! in degrees
    TYPE(DateTimeType), intent(in) :: datetime   ! struct
    real(r8_),          intent(in) :: lat2, lon2 ! in degrees
    real(r8_),          intent(in) :: timestep   ! in seconds

    ! outputs
    integer,                      intent(out) :: nprofiles 
    TYPE(profile_type), dimension(:), pointer :: profiles
    integer,                      intent(out) :: error_flag

    ! local variables
    TYPE(DateTimeType) :: datetime_mid
    integer            :: AllocStatus
    integer            :: nlevels

    ! init
    error_flag = no_error

    ! allocate and fill 1 profile as test
    nprofiles = 1
    allocate(profiles(1),STAT=AllocStatus)
    IF (AllocStatus .ne. 0) goto 999

    ! set the lat,lon position to the middle of the track
    ! (assume a simple average will do for now, if we need a
    !  more fancy method we can make a special module for it later)
    profiles(1)%lat = (lat1+lat2)/2
    profiles(1)%lon = (lon1+lon2)/2

    ! test the validity if the provided datetime 
    call TestDateTime(datetime,error_flag)
    IF (error_flag .ne. no_error) return

    ! construct a date-time stamp in the middle of
    ! (date,time) and (date,time+timestep)
    datetime_mid = datetime
    call IncrementDateTime(datetime_mid,timestep)

    profiles(1)%datetime = datetime_mid

    ! fill the atmospheric profiles with some dummy data (5 levels only)
    nlevels = 5
    profiles(1)%nlevels = nlevels

    allocate(profiles(1)%z(nlevels),STAT=AllocStatus)
    IF (AllocStatus .ne. 0) goto 999
    profiles(1)%z = (/100.,200.,300.,400.,500./)

    allocate(profiles(1)%T_nom(nlevels),STAT=AllocStatus)
    IF (AllocStatus .ne. 0) goto 999
    profiles(1)%T_nom = (/273.1,273.2,273.3,273.4,273.5/)

    allocate(profiles(1)%P_nom(nlevels),STAT=AllocStatus)
    IF (AllocStatus .ne. 0) goto 999
    profiles(1)%P_nom = (/1013.,1012.,1011.,1010.,1009./)

    ! fill the geoid/ellipsoid fields
    profiles(1)%geoid_ground_level     = 123.45
    profiles(1)%ellipsoid_ground_level = 456.78

    return

    ! print a warning and report the error in case of allocation problems
999 print *,"ERROR: allocation failure in: get_auxiliary_data"
    error_flag = error_allocate
    return

  end subroutine GetAuxiliaryData
    !  #]
  subroutine TestPrintAuxiliaryData(profiles)
    !  #[
    ! inputs
    TYPE(profile_type), dimension(:), pointer :: profiles
    ! outputs none

    ! local variables
    integer :: i,j,nelements

    IF (.not. associated(profiles)) THEN
       print *,"WARNING in test_print_auxiliary_data:"
       print *,"==> the profiles array seems not associated to any memory"
       print *,"nothing to print ..."
       return
    END IF

    ! print the z, T_nom and P_nom arrays inside the profiles-struct
    DO i=1,size(profiles)
       print *,"contents of profile ",i

       IF (associated(profiles(i)%z)) THEN
          nelements = size(profiles(i)%z)
          write(*,*) "REALACC(5) z = ",&
               (profiles(i)%z(j),j=1,nelements),&
               " ENDREALACC"
       ELSE
          print *,"z = [not associated]"
       END IF

       IF (associated(profiles(i)%T_nom)) THEN
          nelements = size(profiles(i)%T_nom)
          write(*,*) "REALACC(5) T_nom = ",&
               (profiles(i)%T_nom(j),j=1,nelements),&
               " ENDREALACC"
       ELSE
          print *,"T_nom = [not associated]"
       END IF

       IF (associated(profiles(i)%P_nom)) THEN
          nelements = size(profiles(i)%P_nom)
          write(*,*) "REALACC(5) P_nom = ",&
               (profiles(i)%P_nom(j),j=1,nelements),&
               " ENDREALACC"
       ELSE
          print *,"P_nom = [not associated]"
       END IF

    END DO
  end subroutine TestPrintAuxiliaryData
    !  #]
  subroutine DeallocAuxiliaryData(profiles) ! input
    !  #[
    ! inputs
    TYPE(profile_type), dimension(:), pointer :: profiles

    ! outputs none
    ! side effect: the profiles will be deallocated after this call

    ! local variables
    integer :: i

    IF (.not. associated(profiles)) THEN
       print *,"WARNING in dealloc_auxiliary_data:"
       print *,"==> the profiles array seems not associated to any memory"
       print *,"nothing to do ..."
       return
    END IF

    ! first deallocate the z, T_nom and P_nom arrays inside the profiles-struct
    DO i=1,size(profiles)
       IF (associated(profiles(i)%z)) THEN
          deallocate(profiles(i)%z)
       ELSE
          print *,"WARNING in dealloc_auxiliary_data:"
          print *,"for profile ",i
          print *,"==> the z array seems not associated to any memory"
          print *,"nothing to do, continuing ..."
       END IF

       IF (associated(profiles(i)%T_nom)) THEN
          deallocate(profiles(i)%T_nom)
       ELSE
          print *,"WARNING in dealloc_auxiliary_data:"
          print *,"for profile ",i
          print *,"==> the T_nom array seems not associated to any memory"
          print *,"nothing to do, continuing ..."
       END IF

       IF (associated(profiles(i)%P_nom)) THEN
          deallocate(profiles(i)%P_nom)
       ELSE
          print *,"WARNING in dealloc_auxiliary_data:"
          print *,"for profile ",i
          print *,"==> the P_nom array seems not associated to any memory"
          print *,"nothing to do, continuing ..."
       END IF
    END DO

    ! the deallocate the struct itself
    deallocate(profiles)

  end subroutine DeallocAuxiliaryData
    !  #]
  !------------------------------------------
end module DummyAuxiliaryModule
