module AuxiliaryModule
  !  #[ Description:

  !     A module to
  !     -temporary replace the module at ECMWF
  !      which retrieves the real NWP and auxiliary data
  !     OR
  !     -to serve as a template for a special local module to be
  !      defined by any MET-center that wishes to run the L2B processor 
  !      using their own model data.
  !
  !     This ECMWF temporary module assumes that odbviewer has 
  !     produced a report file from the following SQL
  !
  !  CREATE VIEW myview AS
  !    SELECT seqno@hdr, lat@hdr, lon@hdr, lat@aeolus_auxmet, \\
  !           lon@aeolus_auxmet, lev@aeolus_auxmet, \\
  !           ptop@aeolus_auxmet, pnom@aeolus_auxmet, \\
  !           ztop@aeolus_auxmet, znom@aeolus_auxmet, \\
  !           u@aeolus_auxmet, v@aeolus_auxmet, t@aeolus_auxmet, \\
  !           rh@aeolus_auxmet, q@aeolus_auxmet, \\
  !           cc@aeolus_auxmet, clwc@aeolus_auxmet, ciwc@aeolus_auxmet, \\
  !           timeslot@timeslot_index, tslot@index, \\
  !           date@hdr, time@hdr, \\
  !           andate@desc, antime@desc, \\
  !           error_t@aeolus_auxmet, error_rh@aeolus_auxmet, \\
  !           error_p@aeolus_auxmet, aeolus_hdrflag@aeolus_hdr
  !      FROM hdr, sat, aeolus_hdr, aeolus_auxmet, \\
  !           timeslot_index, index, desc
  !  ORDERBY tslot, seqno, ptop
  ! ;

  !  #]
  !  #[ Modifications:
  !      25-Feb-2005 J. de Kloe  Original version.
  !      17-Mar-2005 D. Tan      Read from aux data from ODB rpt.
  !      22-Apr-2005 J. de Kloe  added the error handling module
  !      03-May-2005 D. Tan      Print lat and lon for each profile
  !      09-May-2005 D. Tan      Limit which profiles are printed
  !      20-May-2005 J. de Kloe  added the write_aux_met_data routine
  !      31-May-2005 J. de Kloe  added simple_xml module
  !      21-Jun-2005 D. Tan      Convert geopotential to geop.height
  !      21-Jun-2005 D. Tan      Take profile lat/lon from data file.
  !      22-Jun-2005 D. Tan      Calculate profile date/time from tslot@index.
  !      05-Aug-2005 D. Tan      Read AMD from EE binary file (one profile).
  !      26-Aug-2005 D. Tan      Allow skipping of profiles when reading 
  !                              ODB rpt.
  !      30-Aug-2005 J. de Kloe  added REALACC/ENDREALACC to the output
  !      27-Sep-2005 J. de Kloe  moved L2B_AMD_SPH to its own module
  !      19-Oct-2005 D. Tan      No workaround for AMD_FH.
  !      24-Nov-2005 J. de Kloe  changed timestep to r8_ type
  !      03-Jan-2006 J. de Kloe  small xml-module interface change
  !      11-Jan-2006 J. de Kloe  added use of the new DateTime type
  !      25-Jan-2006 P. Poli     modified GetAuxiliaryData: added KNMI 
  !                              metdb reading
  !      31-Jan-2006 D. Tan      Additional parameters in Profile_Type.
  !      03-Feb-2006 D. Tan      Convertor RawProfile_to_AMD_DataStructure
  !      10-Feb-2006 D. Tan      Merge fixes
  !      14-Feb-2006 D. Tan,     Rename file_type_AMD with _20060131
  !      20-Feb-2006 J. de Kloe  Corrected handling of missing numbers
  !                              and adapted to changed Fill_MPH_items()
  !      27-Feb-2006 D. Tan      Trap negative rh and cc
  !      02-Mar-2006 J. de Kloe  adapted to changed XML_module
  !      06-Mar-2006 D. Tan      Profile_type r8_ precision
  !      10-Mar-2006 D. Tan      Read ODB variables into r8_ precision
  !      30-Mar-2006 D. Tan      Introduce fill_off_nadir flag.
  !      07-Apr-2006 D. Tan      Surface geopotential converted to metres
  !                              + 1st Convertor AMD_DataStructure_to_RawProfile
  !      12-May-2006 D. Tan      Null use of ParKind1 module
  !      31-Jul-2006 J. de Kloe  renamed set_file_pointer to SetFilePointer
  !      07-Aug-2006 J. de Kloe  added delete routine for L2B_AMD_SPH
  !      17-Aug-2006 P. Poli     Removed obsolete variables
  !      04-Sep-2006 P. Poli     Fixed inconsistencies in (P,z)nom/base/top
  !                              in routine GetAuxiliaryData
  !      17-Apr-2007 D. Tan      Remove obsolete routines WriteAuxMetData,
  !                              WriteAMDtoBinaryFile (non-public),
  !                              ReadRawAMD_DBL,
  !                              GetRawAuxiliaryDataEEformat,
  !                              and obsolete file_type_AMD_20060131.
  !      03-May-2007 D. Tan      TestPrintAuxiliaryData with non-pointer 
  !                              argument
  !      29-May-2007 D. Tan      Move Copy_AMD_to_WorkingData() and
  !                              subsidiary routines from l2bp_module, and
  !                              hence remove AMD_DataStructure_to_RawProfile.
  !      06-Jun-2007 P. Poli     Fix memory leak: add routine Delete_Profile 
  !                              Remove variable knmi_z_1 (not set, not used)
  !      26-Jun-2007 J. de Kloe  added some prints for dummy variables that are
  !                              set but never used (except for stepping over 
  !                              unused parts of the input file), to prevent
  !                              the g95 compiler from printing
  !                              warnings about them.
  !      28-Jun-2007 J. de kloe  Split GetAuxiliaryData in 2 parts. One 
  !                              containing code to handle ECMWF's ODB input 
  !                              (name remains GetAuxiliaryData)
  !                              and one to handle KNMI ascii input named 
  !                              GetAuxiliaryDataKnmi(). Removed the conversion
  !                              to the 60 levels used by the ODB output, since
  !                              our code should be robust enough to use 
  !                              different numbers of vertical levels
  !                              (not yet finished due to other priorities)
  !      02-Aug-2007 J. de Kloe  finished the above splitting of the 
  !                              GetAuxiliaryData
  !                              routine. Code is in working order again.
  !      06-Aug-2007 D. Tan      Revert p and T interpolation pending further
  !                              study
  !      07-Aug-2007 D. Tan      Tidy column indexing in GetAuxiliaryData
  !      07-Aug-2007 D. Tan      New interface to GetAuxiliaryData which
  !                              returns multiple profiles
  !      09-Aug-2007 D. Tan      Extended format for ODB rpt files - 4 new 
  !                              columns
  !      10-Aug-2007 D. Tan      Pass num_model_layers as GetAuxiliaryData arg.
  !      04-Sep-2007 P. Poli     change L1B_DataType to L1B_BRC_DataType
  !      19-Oct-2007 P. Poli     split AMD_DataType in two parts:
  !                              AMD_HDR_DataType and AMD_BRC_DataType
  !      26-Oct-2007 J. de Kloe  made DateTimeType public to allow compilation
  !                              with g95 -std=f95
  !      12-Nov-2007 J. de Kloe  removed several not needed items from 
  !                              the line: use xml_module 
  !      16-Jan-2008 J. de Kloe  phase out integer kind i_ and real kind r_
  !      17-Jan-2008 J. de Kloe  removed a last few 64bit incompatibilities
  !  #]
  !  #[ Modules used:

  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE FixedHeader, only: FH_type
  USE MainProductHeader, only: MPH_type
  USE L2B_AMD_SPH, only: L2B_AMD_SPH_type
  USE EE_CFI_Datatypes, only: DateTimeType,SetDateTime, IncrementDateTime, &
                         InitDateTime,&
                         TestDateTime, write_DateTime_to_bin_file, &
                         read_DateTime_from_bin_file, maxval_IntAl, &
                         IntAuc, IntAul,IntAl, IntAus, IntAs, FAdoxy, &
                         GetIntAus, &
                         format_IntAl, format_IntAus, format_IntAs, &
                         FAdoxy_to_IntAuc, FAdoxy_to_IntAs, FAdoxy_to_IntAus, &
                         FAdoxy_to_IntAl, FAdoxy_to_IntAul, IntAd

 ! define the default integer and real to be used by the module
  USE Compiler_Features, only: newline
  USE Numerics, only: i8_, i4_, r8_, missing_indicator_real
  USE LunManager, only: get_lun, free_lun
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE ErrorHandler, only: no_error, error_reading_file,&
                          error_opening_file, error_allocate, &
                          error_programming
!  USE Interp1, only: finterp1, interpolate

  USE ProfileInterpolate,only: ProfileInterpolate_type => profile_type, &
       ProfileInterpolate_Init_Profile => init_profile, &
       interpolate, &
       ProfileInterpolateDeleteProfile => delete_profile, &
       interp_type_log, extrap_type_log, interp_type_spline, &
       extrap_type_edgeval

  USE XML_module, only: Init_XML

  USE KVT_module, only: write_kvt_string, Init_KVT_module
  USE DirectBinaryIO, only: Init_DirectBinaryIO_module, &
       write_int4_to_bin_file, write_int4array_to_bin_file, &
       write_real4_to_bin_file, write_real4array_to_bin_file, &
       write_real8_to_bin_file, write_real8array_to_bin_file, &
                        read_real4_from_bin_file, &
                        read_real4array_from_bin_file, &
                        read_real8_from_bin_file, &
                        read_real8array_from_bin_file, &
                        read_int4_from_bin_file, &
                        read_int4array_from_bin_file, &
                        SetFilePointer, & ! AdvanceFilePointer,
                        BigEndian

  ! define the AMD_DataStructure
  USE AMD_DataStructure, only: AMD_BRC_DataType
  USE AMD_Geoloc_ADS,    only: AMD_Geoloc_ADS_Type
  USE AMD_Met_MDS,       only: AMD_met_MDS_Type
  USE L2B_AMD_SPH,       only: Delete_L2B_AMD_SPH

  ! define the L1B_DataStructure
  USE L1B_DataStructure, only: L1B_BRC_DataType

  ! define the Working_DataStructure
  USE Working_DataStructure, only: WorkingDataType, &
                      WD_MieRangeGateType      => MieRangeGateType, &
                      WD_RayleighRangeGateType => RayleighRangeGateType

  !  #]
  !  #[ Parameter/Variable/Type declarations

  implicit none

  ! make all module contents private, unless explicitely stated otherwise

  private

  TYPE profile_type
     real(r8_)                        :: lat
     real(r8_)                        :: lon
     TYPE(DateTimeType)               :: datetime

     ! Meteorological surface parameters
     real(r8_)                        :: us
     real(r8_)                        :: vs
     real(r8_)                        :: ps
     real(r8_)                        :: err_ps
     real(r8_)                        :: zs

     integer                          :: nlevels
     ! will we define z compared to the geoid ? In that way we can
     ! just copy it from the model output (if I remember correctly).
     real(r8_), dimension(:), pointer :: z
     real(r8_), dimension(:), pointer :: T_nom ! was T_nwp
     real(r8_), dimension(:), pointer :: P_nom ! was P_nwp

     ! Further profile parameters defined in AMD IODD
     ! Available for use in baseline L2Bp.
     real(r8_), dimension(:), pointer :: P_base
     real(r8_), dimension(:), pointer :: P_top
     real(r8_), dimension(:), pointer :: z_base
     real(r8_), dimension(:), pointer :: z_top
     real(r8_), dimension(:), pointer :: err_T

     ! Further profile parameters defined in AMD IODD
     ! Not for use in baseline L2Bp.
     real(r8_), dimension(:), pointer :: u
     real(r8_), dimension(:), pointer :: v
     real(r8_), dimension(:), pointer :: rh
     real(r8_), dimension(:), pointer :: err_rh
     real(r8_), dimension(:), pointer :: q
     real(r8_), dimension(:), pointer :: cc
     real(r8_), dimension(:), pointer :: clwc
     real(r8_), dimension(:), pointer :: ciwc

     ! Provisional parameters.
     ! Geoid may be removed if present in L1B.

     real(r8_)                       :: geoid_ground_level
     real(r8_)                       :: ellipsoid_ground_level

     logical                         :: Computed_L1B_Location
     logical                         :: OffNadir

  END TYPE profile_type

  ! the default verbosity setting is Err (as silent as possible)
  integer, private, save :: LocalVerbosity = VerbosityError

  TYPE L2B_AMD_HEADER_TYPE
     TYPE(FH_type),          pointer :: FH
     TYPE(MPH_type),         pointer :: MPH
     TYPE(L2B_AMD_SPH_type), pointer :: SPH
  END TYPE L2B_AMD_HEADER_TYPE

  !  #]
  !  #[ Module interface
  ! specify which routines are public, so are part of the interface
  ! to this module
  ! public parameters
  ! public types
  public :: profile_type,DateTimeType
  ! public functions/subroutines
  public :: InitAuxiliaryModule, GetAuxiliaryData, GetAuxiliaryDataKnmi, &
            TestPrintAuxiliaryData, DeallocAuxiliaryData, &
            RawProfile_to_AMD_DataStructure, Copy_AMD_to_RawProfile,&
            Init_Profile, Delete_Profile, Copy_AMD_to_WorkingData
  !  #]
contains
  !------------------------------------------
  subroutine InitAuxiliaryModule(error_flag,ChoosenVerbosity)
    !  #[
    integer,           intent(out) :: error_flag
    integer, optional, intent(in)  :: ChoosenVerbosity

    error_flag = no_error

    IF (present(ChoosenVerbosity)) LocalVerbosity = ChoosenVerbosity

    IF (LocalVerbosity .ge. VerbosityReport) &
         call print_verbosity_setting(LocalVerbosity,&
                                      ModuleName="AuxiliaryModule")

    call Init_XML(error_flag) !(ChoosenVerbosity=LocalVerbosity)
    IF (error_flag .ne. no_error) return

    call Init_KVT_module ! (ChoosenVerbosity=VerbosityReport)

    call Init_DirectBinaryIO_module(error_flag,FileByteOrder=BigEndian)
    IF (error_flag .ne. no_error) return

  end subroutine InitAuxiliaryModule
    !  #]
  subroutine update_odbrpt_read_parameters(col_start, col_end, col_width)
    !  #[
    integer,           intent(inout) :: col_start
    integer,           intent(inout) :: col_end
    integer,           intent(in)    :: col_width

    col_start = col_end + 1
    col_end   = col_end + col_width

  end subroutine update_odbrpt_read_parameters
    !  #]
  subroutine GetAuxiliaryData( in_file,             & ! input
                               num_model_layers,    & ! input
                               num_ecmwf_profiles,  & ! input
                               nprofiles, profiles, & ! output
                               error_flag           ) ! output
    !  #[

    ! TODO:    Clearly we need to do some cleaning-up here

    ! reads NWP data from ODB ascii file
    ! WARNING: u,v elements are not yet available in the ODB file
    ! (are set to NULL) so these remain missing for now.

    ! filename used for reading input ODB data
    character(len=*),intent(in) :: in_file
    integer,         intent(in) :: num_model_layers
    integer,         intent(in) :: num_ecmwf_profiles

    ! outputs
    integer,                          intent(out) :: nprofiles
    TYPE(profile_type), dimension(:), pointer     :: profiles
    integer,                          intent(out) :: error_flag
    
    ! local variables
    integer     :: AllocStatus
    integer     :: nlevels
    integer     :: ios

    integer     :: i

    ! For reading ODB report file
    integer, parameter          :: len_dataline = 525
    character(len=len_dataline) :: data_line
    integer          :: in_unit
    character        :: dummy_character
    
    integer          :: col_start
    integer          :: col_end
    integer          :: col_width

    integer          :: ncols, dummy_integer, nrows
    integer          :: ilevel
    
    integer          :: seqno
    real(r8_)        :: lat_at_hdr, lon_at_hdr
    real(r8_)        :: lat_at_auxmet, lon_at_auxmet
    integer          :: model_level
    real(r8_)        :: ptop, pnom
    real(r8_)        :: ztop, znom
    real(r8_)        :: u, v
    real(r8_)        :: t
    real(r8_)        :: err_t
    real(r8_)        :: rh
    real(r8_)        :: err_rh
    real(r8_)        :: q
    real(r8_)        :: cc, clwc, ciwc
    
    real(r8_)        :: ps, zs
    real(r8_)        :: u10, v10

    real(r8_)        :: err_ps

    integer, parameter                :: len_hdrflag_string = 32
    character(len=len_hdrflag_string) :: hdrflag_string_ref
    character(len=len_hdrflag_string) :: hdrflag_string
    
    integer          :: print_level, read_statement_number 
    
    real(r8_), parameter :: Rg = 9.80665 ! Would be appended with _JPRB in IFS
                                 ! Later, move to a universal parameter module

    TYPE(DateTimeType) :: datetime_analysis ! struct
    TYPE(DateTimeType) :: datetime_timeslot ! struct
    real(r8_)          :: time_change  ! in seconds    
    integer            :: slot_seconds ! in seconds    
    integer            :: tslot_at_index
    integer            :: tslot_analysis
    character(len=12)  :: an_date_string
    character(len=12)  :: an_time_string
    integer            :: an_year, an_month, an_day
    integer            :: an_hour, an_minute
    real(r8_)          :: an_second

!   The following values should be set in a setup routine.
!   They are assimilation dependent.

    slot_seconds   = 1800 ! 30 minute timeslots
    tslot_analysis = 7 ! assuming 12 hour 4d-var, early delivery system 
                       ! and 30 minute timeslots

    ! init
    error_flag = no_error
    ! init to suppres a g95 warning about uninitialised use of this var
    nlevels = num_model_layers 

    ! define for which model levels some test data should be printed
    ! when the verbosity settings allow it (0=surface, 1=upper air)
    
    print_level = 1
!    print_level = 0

    !------------------
    ! Open input files
    !------------------

    ! open ODB file
    in_unit = get_lun()
    IF (in_unit .lt. 0) THEN
       ! no fileunits available, so report an error
       error_flag = error_opening_file
       return
    END IF
    print *,"opening aux met datafile: ",trim(in_file)
    open( UNIT=in_unit, FILE=in_file, FORM='formatted', status="old", err=998)

    !---------------------------
    ! read ODB file first lines
    !---------------------------

    ! read the 1st header line and extract ncols, nrows from it
    read_statement_number = 1 !used as index for the error handler
    read( in_unit, '(A1,1X,I2,1X,I1,1X,I4)', iostat=ios) &
         dummy_character, ncols, dummy_integer, nrows
    IF (ios .ne. 0) goto 997

    IF (LocalVerbosity .ge. VerbosityReport) &
         write(6,*) 'AMD report, ncols = ', ncols, ', nrows = ', nrows
    
    ! read the next 5 header lines without using them
    read_statement_number = 2
    read( in_unit, '(A)', iostat=ios ) dummy_character
    IF (ios .ne. 0) goto 997
    read_statement_number = 3
    read( in_unit, '(A)', iostat=ios ) dummy_character
    IF (ios .ne. 0) goto 997
    read_statement_number = 4
    read( in_unit, '(A)', iostat=ios ) dummy_character
    IF (ios .ne. 0) goto 997
    read_statement_number = 5
    read( in_unit, '(A)', iostat=ios ) dummy_character
    IF (ios .ne. 0) goto 997
    read_statement_number = 6
    read( in_unit, '(A)', iostat=ios ) dummy_character
    IF (ios .ne. 0) goto 997
       
    !-------------------------
    ! allocate profile
    !-------------------------
    !  #[ initialisations and allocations

    ! allocate multiple profile
    nprofiles = num_ecmwf_profiles
    allocate(profiles(nprofiles),STAT=AllocStatus)
    IF (AllocStatus .ne. 0) goto 999

    ! allocate profiles needed to fill all AMD file fields
    DO i=1,nprofiles

    ! The date and time for the profile are deduced below from the
    ! data file.
    ! Any mismatch between the input position and the parameters
    ! date_mid, time_mid should be tested elsewhere.
         
    ! fill the atmospheric profiles with data
    nlevels = num_model_layers 
    profiles(i)%nlevels = nlevels
    
    allocate(profiles(i)%z(     nlevels), &
             profiles(i)%T_nom( nlevels), &
             profiles(i)%P_nom( nlevels), &
             profiles(i)%P_base(nlevels), &
             profiles(i)%P_top( nlevels), &
             profiles(i)%z_base(nlevels), &
             profiles(i)%z_top( nlevels), &
             profiles(i)%err_T( nlevels), &
             profiles(i)%u(     nlevels), &
             profiles(i)%v(     nlevels), &
             profiles(i)%rh(    nlevels), &
             profiles(i)%err_rh(nlevels), &
             profiles(i)%q(     nlevels), &
             profiles(i)%cc(    nlevels), &
             profiles(i)%clwc(  nlevels), &
             profiles(i)%ciwc(  nlevels), STAT=AllocStatus)
    IF (AllocStatus .ne. 0) goto 999
    
    ! fill the geoid/ellipsoid fields
    profiles(i)%geoid_ground_level     = 123.45
    profiles(i)%ellipsoid_ground_level = 456.78

    END DO
    !  #]

    profile_loop: DO i = 1, num_ecmwf_profiles
       write(6,*) 'Reading profile number ', i

       DO ilevel = 0, nlevels

          ! read one line in ODB file
          read_statement_number = 7
          read( in_unit, '(A)', iostat=ios ) data_line
          IF (ios .ne. 0) goto 997

             !  #[ read a profile layer from data_line

             ! Initialize the column boundaries
             col_start = 0
             col_end   = 0

             ! split the line in substrings, and convert these to numbers
             read_statement_number = 8
             col_width = 14
             call update_odbrpt_read_parameters(col_start, col_end, col_width)
             read( data_line(col_start:col_end), *, iostat=ios ) seqno
             IF (ios .ne. 0) goto 997

             read_statement_number = 9
             col_width = 20
             call update_odbrpt_read_parameters(col_start, col_end, col_width)
             read( data_line(col_start:col_end), *, iostat=ios ) lat_at_hdr
             IF (ios .ne. 0) goto 997

             read_statement_number = 10
             col_width = 20
             call update_odbrpt_read_parameters(col_start, col_end, col_width)
             read( data_line(col_start:col_end), *, iostat=ios ) lon_at_hdr
             IF (ios .ne. 0) goto 997

             read_statement_number = 11
             col_width = 20
             call update_odbrpt_read_parameters(col_start, col_end, col_width)
             read( data_line(col_start:col_end), *, iostat=ios ) lat_at_auxmet
             IF (ios .ne. 0) goto 997

             read_statement_number = 12
             col_width = 20
             call update_odbrpt_read_parameters(col_start, col_end, col_width)
             read( data_line(col_start:col_end), *, iostat=ios ) lon_at_auxmet
             IF (ios .ne. 0) goto 997

             read_statement_number = 13
             col_width = 20
             call update_odbrpt_read_parameters(col_start, col_end, col_width)
             read( data_line(col_start:col_end), *, iostat=ios ) model_level
             IF (ios .ne. 0) goto 997
          
             ! a test print
             IF (LocalVerbosity .ge. VerbosityReport) THEN
                if ( ( ( print_level .ge. 1 ) .and. ( model_level .gt. 0 ) ) .or. &
                     ( ( print_level .ge. 0 ) .and. ( model_level .eq. 0 ) ) ) then
                   write(6,*) 'Seqno = ',seqno
                   write(6,*) ' Model level = ',model_level
                   write(6,*) '  lat_at_hdr    = ', lat_at_hdr
                   write(6,*) '  lon_at_hdr    = ', lon_at_hdr
                end if
             END IF
            
             if ( model_level .gt. 0 ) then
                !  #[ Proper layers, i.e. not surface
             
                read_statement_number = 14
                col_width = 20
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
                read( data_line(col_start:col_end), *, iostat=ios ) ptop
                IF (ios .ne. 0) goto 997

                read_statement_number = 15
                col_width = 20
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
                read( data_line(col_start:col_end), *, iostat=ios ) pnom
                IF (ios .ne. 0) goto 997

                read_statement_number = 16
                col_width = 20
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
                read( data_line(col_start:col_end), *, iostat=ios ) ztop
                IF (ios .ne. 0) goto 997

                read_statement_number = 17
                col_width = 20
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
                read( data_line(col_start:col_end), *, iostat=ios ) znom
                IF (ios .ne. 0) goto 997

!       u and v are NULL
!                read_statement_number = 18
                col_width = 20
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
!                read( data_line(col_start:col_end), *, iostat=ios ) u
!                IF (ios .ne. 0) goto 997
!                read_statement_number = 19
                 col_width = 20
                 call update_odbrpt_read_parameters(col_start, col_end, col_width)
!                read( data_line(col_start:col_end), *, iostat=ios ) v
!                IF (ios .ne. 0) goto 997
!                u=0.
!                v=0.
                u = missing_indicator_real
                v = missing_indicator_real

                read_statement_number = 20
                col_width = 20
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
                read( data_line(col_start:col_end), *, iostat=ios ) t
                IF (ios .ne. 0) goto 997

!      
!       RH is NULL
!       for elwl.  TODO: read non-null when input_rpt upgraded to etpm or later.
!                read_statement_number = 21
                col_width = 20
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
!                read( data_line(col_start:col_end), *, iostat=ios ) rh
!                IF (ios .ne. 0) goto 997
                rh = missing_indicator_real

                read_statement_number = 22
                col_width = 20
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
                read( data_line(col_start:col_end), *, iostat=ios ) q
                IF (ios .ne. 0) goto 997

                read_statement_number = 23
                col_width = 20
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
                read( data_line(col_start:col_end), *, iostat=ios ) cc
                IF (ios .ne. 0) goto 997

                read_statement_number = 24
                col_width = 20
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
                read( data_line(col_start:col_end), *, iostat=ios ) clwc
                IF (ios .ne. 0) goto 997

                read_statement_number = 25
                col_width = 20
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
                read( data_line(col_start:col_end), *, iostat=ios ) ciwc
                IF (ios .ne. 0) goto 997
   
                ! Skip columns of 16 + 13 + 4 * 12
                col_width = 16 + 13 + 4 * 12
                call update_odbrpt_read_parameters(col_start, col_end, col_width)

                read_statement_number = 251
                col_width = 20
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
                read( data_line(col_start:col_end), *, iostat=ios ) err_T
                IF (ios .ne. 0) goto 997

                read_statement_number = 252
                col_width = 20
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
                read( data_line(col_start:col_end), *, iostat=ios ) err_rh
                IF (ios .ne. 0) goto 997
             
!      
!       err_ps is NULL for non-surface layers
                read_statement_number = 253
                col_width = 20
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
!                read( data_line(col_start:col_end), *, iostat=ios ) err_ps
!                IF (ios .ne. 0) goto 997
                err_ps = missing_indicator_real

                ! Skip 2 blank characters
                read_statement_number = 254
                col_width = 2
                call update_odbrpt_read_parameters(col_start, col_end, col_width)

                ! Take the aeolus_hdrflag as a character string
                col_width = len_hdrflag_string
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
                hdrflag_string = data_line(col_start:col_end)

                if ( ilevel .eq. 0 ) then
                   hdrflag_string_ref = hdrflag_string
                else
                   if (hdrflag_string_ref .ne. hdrflag_string) then
                      print *,'GetAuxiliaryData: unexpected change in '//&
                              'aeolus_hdrflag at model_level ', model_level
                      print *,'GetAuxiliaryData: hdrflag expected '//hdrflag_string_ref
                      print *,'GetAuxiliaryData: hdrflag found    '//hdrflag_string
                   end if
                end if

                ! Convert from geopotential to geopotential height
                znom = znom / Rg

               if( print_level .ge. 1 ) then
                   IF (LocalVerbosity .ge. VerbosityReport) THEN
                      ! a test print
                      write(6,*) '  ptop = ', ptop
                      write(6,*) '  pnom = ', pnom
                      write(6,*) '  ztop = ', ztop
                      write(6,*) '  znom = ', znom
!                      write(6,*) '  u    = ', u
!                      write(6,*) '  v    = ', v
                      write(6,*) '  t    = ', t
!                      write(6,*) '  rh   = ', rh
                      write(6,*) '  q    = ', q
                      write(6,*) '  cc   = ', cc
                      write(6,*) '  clwc = ', clwc
                      write(6,*) '  ciwc = ', ciwc
                   END IF
                endif
             
                ! copy the data to the profile struct for later use
     
                profiles(i)%z(     model_level ) = znom
                profiles(i)%T_nom( model_level ) = t
                profiles(i)%P_nom( model_level ) = pnom
             
                if ( model_level .gt. 1 ) profiles(i)%P_base( model_level - 1 ) = ptop
                profiles(i)%P_top ( model_level ) = ptop

                if ( model_level .gt. 1 ) profiles(i)%z_base( model_level - 1 ) = ztop / Rg
                profiles(i)%z_top ( model_level ) = ztop / Rg
          
                profiles(i)%err_T( model_level ) = err_T
                profiles(i)%u( model_level ) = u
                profiles(i)%v( model_level ) = v
                profiles(i)%rh( model_level ) = rh
                profiles(i)%err_rh( model_level ) = err_rh
                profiles(i)%q( model_level ) = q
                profiles(i)%cc( model_level ) = cc
                profiles(i)%clwc( model_level ) = clwc
                profiles(i)%ciwc( model_level ) = ciwc
                !  #]
             else
                !  #[ model level eq 0, so we are on the surface 
   
                read_statement_number = 26
                col_width = 20
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
                read( data_line(col_start:col_end), *, iostat=ios ) ps
                IF (ios .ne. 0) goto 997

                ! Skip pnom column
                col_width = 20
                call update_odbrpt_read_parameters(col_start, col_end, col_width)

                read_statement_number = 27
                col_width = 20
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
                read( data_line(col_start:col_end), *, iostat=ios ) zs
                IF (ios .ne. 0) goto 997

                ! Skip znom column
                col_width = 20
                call update_odbrpt_read_parameters(col_start, col_end, col_width)

                read_statement_number = 28
                col_width = 20
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
                read( data_line(col_start:col_end), *, iostat=ios ) u10
                IF (ios .ne. 0) goto 997

                read_statement_number = 29
                col_width = 20
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
                read( data_line(col_start:col_end), *, iostat=ios ) v10
                IF (ios .ne. 0) goto 997

                ! Skip 6 columns of 20 = 120
                col_width = 20 * 6
                call update_odbrpt_read_parameters(col_start, col_end, col_width)

!                read_statement_number = 30
                col_width = 16
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
!                read( data_line(col_start:col_end), *, iostat=ios ) timeslot_at_timeslotindex
!                IF (ios .ne. 0) goto 997
   
                read_statement_number = 31
                col_width = 13
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
                read( data_line(col_start:col_end), *, iostat=ios ) tslot_at_index
                IF (ios .ne. 0) goto 997

                col_width = 12
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
!                obs_date_string = data_line(col_start:col_end)
                col_width = 12
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
!                obs_time_string = data_line(col_start:col_end)

                col_width = 12
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
                an_date_string = data_line(col_start:col_end)

                read_statement_number = 36
                read( an_date_string(1:col_width), '(4X,I4,I2,I2)', iostat=ios ) &
                   an_year, an_month, an_day
                IF (ios .ne. 0) goto 997

                col_width = 12
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
                an_time_string = data_line(col_start:col_end)

                read_statement_number = 39
                read( an_time_string(1:col_width), '(6X,I2,I2,F2.0)', iostat=ios ) &
                   an_hour, an_minute, an_second
                IF (ios .ne. 0) goto 997

                call SetDateTime(datetime_analysis,&
                                 an_year,an_month,an_day,&
                                 an_hour,an_minute,an_second)

!       err_T is NULL for surface layers
                read_statement_number = 451
                col_width = 20
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
!                read( data_line(col_start:col_end), *, iostat=ios ) err_T
!                IF (ios .ne. 0) goto 997

!       err_rh is NULL for surface layers
                read_statement_number = 452
                col_width = 20
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
!                read( data_line(col_start:col_end), *, iostat=ios ) err_rh
!                IF (ios .ne. 0) goto 997

                read_statement_number = 453
                col_width = 20
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
                read( data_line(col_start:col_end), *, iostat=ios ) err_ps
                IF (ios .ne. 0) goto 997

                ! Skip 2 blank characters
                read_statement_number = 454
                col_width = 2
                call update_odbrpt_read_parameters(col_start, col_end, col_width)

                ! Take the aeolus_hdrflag as a character string
                col_width = len_hdrflag_string
                call update_odbrpt_read_parameters(col_start, col_end, col_width)
                hdrflag_string = data_line(col_start:col_end)

                if ( ilevel .eq. 0 ) then
                   hdrflag_string_ref = hdrflag_string
                else
                   if (hdrflag_string_ref .ne. hdrflag_string) then
                      print *,'GetAuxiliaryData: unexpected change in '//&
                              'aeolus_hdrflag at model_level ', model_level
                      print *,'GetAuxiliaryData: hdrflag expected '//hdrflag_string_ref
                      print *,'GetAuxiliaryData: hdrflag found    '//hdrflag_string
                   end if
                end if

                ! Display these even though we do not save them in profiles
                ! for debugging purposes
                IF (LocalVerbosity .ge. VerbosityReport) THEN
                   if( print_level .ge. 0 ) then
                      write(6,*) '  ps  = ', ps
                      write(6,*) '  zs  = ',zs
                      write(6,*) '  u10 = ',u10
                      write(6,*) '  v10 = ',v10
                      write(6,*) '  tslot  = ', tslot_at_index
                      write(6,*) '  andate = ', an_date_string
                      write(6,*) '  antime = ', an_time_string
                  endif
                END IF
                
                profiles(i)%lat = lat_at_hdr
                profiles(i)%lon = lon_at_hdr
   
   
                ! construct a date-time stamp from the analysis and timeslot 
                ! information present in the IFS ODB (ascii input data file)
                datetime_timeslot = datetime_analysis
                time_change   = ( tslot_at_index - tslot_analysis ) * slot_seconds
                
                call TestDateTime(datetime_timeslot,error_flag)
                IF (error_flag .ne. no_error) return
   
                call IncrementDateTime(datetime_timeslot,time_change)
      
                profiles(i)%datetime = datetime_timeslot

                ! Meteorological surface parameters
                profiles(i)%us = u10
                profiles(i)%vs = v10
                profiles(i)%ps = ps
                profiles(i)%err_ps = err_ps
                profiles(i)%zs = zs / Rg ! geopotential converted to metres

                profiles(i)%P_base(nlevels) = ps
                profiles(i)%z_base(nlevels) = zs / Rg
             !  #]
             endif
             !  #]

       END DO ! ilevel

       ! Use hdrflag_string_ref to set flags in profiles(i)
       ! Bit settings defined in geodetic2simulobs.ksh
       profiles(i)%Computed_L1B_Location = ( hdrflag_string_ref(31:31) .eq. '1' )
       profiles(i)%OffNadir              = ( hdrflag_string_ref(32:32) .eq. '0' )

    END DO profile_loop

    ! Close input file
    close( in_unit )
    call free_lun(in_unit)

    return
    !  #[ error handlers 

    ! print warning and report the error in case a read statement failed
997 print *,"Could not properly read inputfile: ",trim(in_file)
    print *,"this is probably due to a problem reading the format at"
    print *,"read_statement_number: ",read_statement_number
    ! printing of these dummy variables suppresses several warnings about 
    ! variables that are set but never used, when compiling with g95, and
    ! since these print may also be usefull for debugging, I added them here
    print *,"dummy_character = ",dummy_character
    print *,"dummy_integer   = ",dummy_integer
    print *,"lat_at_auxmet   = ",lat_at_auxmet
    print *,"lon_at_auxmet   = ",lon_at_auxmet
    
    error_flag = error_reading_file
    return

    ! print warning and report the error in case the file could not be opened
998 print *,"Could not open inputfile: ",trim(in_file)
    error_flag = error_opening_file
    return

    ! print a warning and report the error in case of allocation problems
999 print *,"ERROR: allocation failure in: get_auxiliary_data"
    error_flag = error_allocate
    return

    !  #]
  end subroutine GetAuxiliaryData
    !  #]
  subroutine GetAuxiliaryDataKnmi(knmi_file,           & ! input
                                  nprofiles, profiles, & ! output
                                  error_flag)            ! output
    !  #[
    ! reads NWP data from a KNMI ascii file

    ! inputs
    ! filename for reading input KNMI data
    character(len=*),intent(in) :: knmi_file  

    ! outputs
    integer,                      intent(out) :: nprofiles 
    TYPE(profile_type), dimension(:), pointer :: profiles
    integer,                      intent(out) :: error_flag
    
    ! local variables
    integer   :: AllocStatus, ios, i, file_unit
    integer   :: n_knmi_levels, n_knmi_profiles, prof, l
    integer   :: hs,he
    real(r8_) :: P,T,u,v,w
    integer   :: mask,year,month,day,hour,minute   
    real(r8_) :: lat,lon,orogr,beta_g, B, albedo, second
    real(r8_) :: wanted_z_base, wanted_z_mid, wanted_z_top
    real(r8_) :: wanted_z_surface
    TYPE(DateTimeType) :: datetime_profile
    character(len=150) :: astuff150
    character(len=150) :: test150

    ! to enable interpolation using the ProfileInterpolate module
    type(ProfileInterpolate_type), dimension(:), pointer :: Interp_P
    type(ProfileInterpolate_type), dimension(:), pointer :: Interp_T
    type(ProfileInterpolate_type), dimension(:), pointer :: Interp_u
    type(ProfileInterpolate_type), dimension(:), pointer :: Interp_v
    type(ProfileInterpolate_type), dimension(:), pointer :: Interp_w

    real(r8_), allocatable, dimension(:) :: knmi_z_base, knmi_z_mid, knmi_z_top
    real(r8_), allocatable, dimension(:) :: knmi_p
    real(r8_), allocatable, dimension(:) :: knmi_T
    real(r8_), allocatable, dimension(:) :: knmi_u
    real(r8_), allocatable, dimension(:) :: knmi_v
    real(r8_), allocatable, dimension(:) :: knmi_w

    ! init
    error_flag = no_error
    nprofiles = 0
    nullify(Interp_P)
    nullify(Interp_T)
    nullify(Interp_u)
    nullify(Interp_v)
    nullify(Interp_w)

    !  #[ open KNMI file
    file_unit = get_lun()
    IF (file_unit .lt. 0) THEN
       ! no fileunits available, so report an error
       error_flag = error_opening_file
       return
    END IF
    open(unit=file_unit,file=knmi_file,form='formatted',status="old", err=998)
    !  #]
    ! read KNMI file first lines
    ! read the comment lines and verify that the file format is as expected
    !  #[ check format for line definition 1
    read( file_unit, '(A150)', iostat=ios ) astuff150
    IF (ios .ne. 0) goto 997
    DO WHILE (astuff150(1:9) .ne. '# format:')
       read( file_unit, '(A150)', iostat=ios ) astuff150
       IF (ios .ne. 0) goto 997
    END DO

    ! note that the format changed 27-jun-2007 to accomodate
    ! use of more than 999 profiles per scene (needed for Lite scenes
    ! at 0.5km hor. resolution, which have 1600 profiles !)
    IF ( (astuff150(1:20) .ne. '# format: (i3,1x,i3)') .and. &
         (astuff150(1:20) .ne. '# format: (i4,1x,i3)')       ) THEN
       print *,"ERROR in GetAuxiliaryDataKnmi: "
       print *,"expected format: (i3,1x,i3) or"
       print *,"expected format: (i4,1x,i3)"
       print *,"but found       "//astuff150(3:len_trim(astuff150))
       error_flag = error_reading_file
       return
    ENDIF
    !  #]
    !  #[ check format for line definition 2
    read( file_unit, '(A150)', iostat=ios ) astuff150
    IF (ios .ne. 0) goto 997
    DO WHILE (astuff150(1:9) .ne. '# format:')
       read( file_unit, '(A150)', iostat=ios ) astuff150
       IF (ios .ne. 0) goto 997
    END DO
    test150 = "# format: (f8.2,1x,f8.2,1x,f6.1,1x,es10.3e2,1x,f6.3,"//&
              "1x,f6.3,1x,i1,1x,i4,1x,i2,1x,i2,1x,i2,1x,i2,1x,f5.2)"
    IF (astuff150(1:104) .ne. test150(1:104) ) THEN
       print *,"ERROR in GetAuxiliaryData:"
       print *,"expected format: "//trim(test150)
       print *,"but found:       "//trim(astuff150(3:))
       error_flag = error_reading_file
       return
    ENDIF
    !  #]
    !  #[ check format for line definition 3

    read( file_unit, '(A150)', iostat=ios ) astuff150
    IF (ios .ne. 0) goto 997
    DO WHILE (astuff150(1:9) .ne. '# format:')
       read( file_unit, '(A150)', iostat=ios ) astuff150
       IF (ios .ne. 0) goto 997
    END DO
    test150 = "# format: (i6,1x,i6,1x,f6.1,1x,f5.1,1x,f5.1,1x,f5.1,1x,"//&
              "f5.1,1x,f5.1,1x,es10.3e2,1x,es10.3e2,1x,es10.3e2,1x,"//&
              "es10.3e2,1x,es10.3e2,1x,es10.3e2)"
    IF (astuff150(1:140) .ne. test150(1:140)) THEN
       print *,"ERROR in GetAuxiliaryData:"
       print *,"expected format: "//trim(test150)
       print *,"but found:       "//trim(astuff150(3:))
       error_flag = error_reading_file
       return
    ENDIF

    !  #]
    !  #[ skip all remaining commented lines, and read the first data line
    DO WHILE (astuff150(1:1) .eq. '#')
       read( file_unit, '(A150)', iostat=ios ) astuff150
       IF (ios .ne. 0) goto 997
    END DO
    !  #]
    !  #[ read definition line 1
    ! note that astuff150 now contains data, read by the list 
    ! read statement executed in the above loop
    read( astuff150, *, iostat=ios ) n_knmi_profiles, n_knmi_levels
    IF (ios .ne. 0) goto 997
    print *,"GetAuxiliaryDataKnmi: n_knmi_profiles =",n_knmi_profiles
    print *,"GetAuxiliaryDataKnmi: n_knmi_levels   =",n_knmi_levels
    !  #]
    !  #[ initialisations and allocations of the temporary arrays

    ! allocate profiles to hold the available knmi ascii data
    allocate (knmi_z_base(n_knmi_levels), &
              knmi_z_mid( n_knmi_levels), &
              knmi_z_top( n_knmi_levels), &
              knmi_p(     n_knmi_levels), &
              knmi_T(     n_knmi_levels), &
              knmi_u(     n_knmi_levels), &
              knmi_v(     n_knmi_levels), &
              knmi_w(     n_knmi_levels), &
              STAT=AllocStatus)
    IF (AllocStatus .ne. 0) goto 999
    
    allocate(Interp_P(n_knmi_profiles),&
             Interp_T(n_knmi_profiles),&
             Interp_u(n_knmi_profiles),&
             Interp_v(n_knmi_profiles),&
             Interp_w(n_knmi_profiles), &
             STAT=AllocStatus)
    IF (AllocStatus .ne. 0) goto 999
    !  #]
    
    readloop: DO prof=1, n_knmi_profiles
       !  #[  read a KNMI ascii profiles
       print *,"GetAuxiliaryDataKnmi: reading KNMI profile: ",prof
       !  #[ read surface and ancillary information line 
       astuff150(:) = ' '
       read(file_unit, '(A150)', iostat=ios ) astuff150
       IF (ios .ne. 0) goto 997
       read(astuff150,'(f8.2,1x,f8.2,1x,f6.1,1x,es10.3e2,'//&
            '1x,f6.3,1x,f6.3,1x,i1,1x,i4,1x,i2,1x,i2,1x,i2,'//&
            '1x,i2,1x,f5.2)',iostat=ios) &
            lat, lon, orogr, beta_g, B, albedo,&
            mask, year, month, day, hour, minute, second
       IF (ios .ne. 0) goto 997
       !  #]
       !  #[ read entire profile
       DO l=1,n_knmi_levels
          astuff150(:) = ' '
          read(file_unit, '(A150)', iostat=ios ) astuff150
          IF (ios .ne. 0) goto 997
          read(astuff150,'(2(i6,1x),f6.1,1x,4(f5.1,1x))',iostat=ios) &
               hs,he,P,T,u,v,w
          IF (ios .ne. 0) goto 997

          knmi_z_base(l) = hs
          knmi_z_top(l)  = he
          knmi_z_mid(l)  = (hs+he)*0.5_r8_ ! take the bin center
          knmi_P(l)      = P*100._r8_  ! convert Pa to hPa
          knmi_T(l)      = T
          knmi_u(l)      = u
          knmi_v(l)      = v
          knmi_w(l)      = w

          ! handle missing values
          ! hs,he should never be missing
          if (knmi_P(l) .lt. -998.0_r8_) knmi_P(l) = missing_indicator_real
          if (knmi_T(l) .lt.  -98.0_r8_) knmi_T(l) = missing_indicator_real
          if (knmi_u(l) .gt.  998.0_r8_) knmi_u(l) = missing_indicator_real
          if (knmi_v(l) .gt.  998.0_r8_) knmi_v(l) = missing_indicator_real
          if (knmi_w(l) .gt.  998.0_r8_) knmi_w(l) = missing_indicator_real
          
       ENDDO
       !  #]
       !  #[ convert knmi data to ProfileInterpolate datastructures

       call ProfileInterpolate_Init_Profile(Interp_P(prof),&
            knmi_z_mid,knmi_P,n_knmi_levels,&
            "pressure profile",&
            error_flag,SkipMissingData=.true.,&
            interpolation_type=interp_type_log,&
            extrapolation_type_down=extrap_type_log,&
            extrapolation_type_up=extrap_type_log)
       if (error_flag .ne. no_error) then
          print *,"GetAuxiliaryDataKnmi: initialising Interp_P profile failed"
          return
       end if

       call ProfileInterpolate_Init_Profile(Interp_T(prof),&
            knmi_z_mid,knmi_T,n_knmi_levels,&
            "temperature profile",&
            error_flag,SkipMissingData=.true.,&
            interpolation_type=interp_type_spline,&
            extrapolation_type_down=extrap_type_edgeval,&
            extrapolation_type_up=extrap_type_edgeval)
       if (error_flag .ne. no_error) then
          print *,"GetAuxiliaryDataKnmi: initialising Interp_T profile failed"
          return
       end if

       call ProfileInterpolate_Init_Profile(Interp_u(prof),&
            knmi_z_mid,knmi_u,n_knmi_levels,&
            "wind u-component profile",&
            error_flag,SkipMissingData=.true.,&
            interpolation_type=interp_type_spline,&
            extrapolation_type_down=extrap_type_edgeval,&
            extrapolation_type_up=extrap_type_edgeval)
       if (error_flag .ne. no_error) then
          print *,"GetAuxiliaryDataKnmi: initialising Interp_u profile failed"
          return
       end if

       call ProfileInterpolate_Init_Profile(Interp_v(prof),&
            knmi_z_mid,knmi_v,n_knmi_levels,&
            "wind v-component profile",&
            error_flag,SkipMissingData=.true.,&
            interpolation_type=interp_type_spline,&
            extrapolation_type_down=extrap_type_edgeval,&
            extrapolation_type_up=extrap_type_edgeval)
       if (error_flag .ne. no_error) then
          print *,"GetAuxiliaryDataKnmi: initialising Interp_v profile failed"
          return
       end if

       call ProfileInterpolate_Init_Profile(Interp_w(prof),&
            knmi_z_mid,knmi_w,n_knmi_levels,&
            "wind w-component profile",&
            error_flag,SkipMissingData=.true.,&
            interpolation_type=interp_type_spline,&
            extrapolation_type_down=extrap_type_edgeval,&
            extrapolation_type_up=extrap_type_edgeval)
       if (error_flag .ne. no_error) then
          print *,"GetAuxiliaryDataKnmi: initialising Interp_w profile failed"
          return
       end if

       !  #]

       !  #]
    END DO readloop

    ! Close input file
    close(file_unit)
    call free_lun(file_unit)

    !  #[ initialisations and allocations of the output arrays

    ! allocate 1 profile
    nprofiles = n_knmi_profiles
    allocate(profiles(nprofiles),STAT=AllocStatus)
    IF (AllocStatus .ne. 0) goto 999

    ! allocate profiles needed to fill all AMD file fields
    DO i=1,nprofiles
       allocate(profiles(i)%z(     n_knmi_levels), &
                profiles(i)%z_base(n_knmi_levels), &
                profiles(i)%z_top( n_knmi_levels), &
                profiles(i)%T_nom( n_knmi_levels), &
                profiles(i)%P_nom( n_knmi_levels), &
                profiles(i)%P_base(n_knmi_levels), &
                profiles(i)%P_top( n_knmi_levels), &
                profiles(i)%err_T( n_knmi_levels), &
                profiles(i)%u(     n_knmi_levels), &
                profiles(i)%v(     n_knmi_levels), &
                profiles(i)%rh(    n_knmi_levels), &
                profiles(i)%err_rh(n_knmi_levels), &
                profiles(i)%q(     n_knmi_levels), &
                profiles(i)%cc(    n_knmi_levels), &
                profiles(i)%clwc(  n_knmi_levels), &
                profiles(i)%ciwc(  n_knmi_levels), STAT=AllocStatus)
       IF (AllocStatus .ne. 0) goto 999
    END DO
    !  #]

    ! fill the profile structure
    fillloop: DO prof=1, n_knmi_profiles
       level_loop: DO l=1,n_knmi_levels
          !  #[
          wanted_z_base = knmi_z_base(l)
          wanted_z_mid  = knmi_z_mid(l)
          wanted_z_top  = knmi_z_top(l)

          profiles(prof)%z(     l) = wanted_z_mid
          profiles(prof)%T_nom( l) = interpolate(Interp_T(prof),wanted_z_mid)
          profiles(prof)%P_nom( l) = interpolate(Interp_P(prof),wanted_z_mid)

          profiles(prof)%P_base(l) = interpolate(Interp_P(prof),wanted_z_base)
          profiles(prof)%P_top( l) = interpolate(Interp_P(prof),wanted_z_top)
          profiles(prof)%z_base(l) = wanted_z_base
          profiles(prof)%z_top( l) = wanted_z_top
          profiles(prof)%err_T( l) = 0.8_r8_ ! Kelvin

          profiles(prof)%u(     l) = interpolate(Interp_u(prof),wanted_z_mid)
          profiles(prof)%v(     l) = interpolate(Interp_v(prof),wanted_z_mid)
          profiles(prof)%rh(    l) = missing_indicator_real
          profiles(prof)%err_rh(l) = missing_indicator_real
          profiles(prof)%q(     l) = missing_indicator_real
          profiles(prof)%cc(    l) = missing_indicator_real
          profiles(prof)%clwc(  l) = missing_indicator_real
          profiles(prof)%ciwc(  l) = missing_indicator_real
          !  #]
       END DO level_loop
       !  #[ fill the surface items

       call SetDateTime(datetime_profile,&
                        year,month,day,hour,minute,second)
       wanted_z_surface     = orogr
       profiles(prof)%lat      = lat
       profiles(prof)%lon      = lon
       profiles(prof)%datetime = datetime_profile
       profiles(prof)%us       = interpolate(Interp_u(prof),wanted_z_surface)
       profiles(prof)%vs       = interpolate(Interp_v(prof),wanted_z_surface)
       profiles(prof)%ps       = interpolate(Interp_P(prof),wanted_z_surface)
       profiles(prof)%err_ps   = missing_indicator_real
       profiles(prof)%zs       = wanted_z_surface
       profiles(prof)%nlevels  = n_knmi_levels
       ! fill the geoid/ellipsoid fields
       profiles(prof)%geoid_ground_level     = 123.45 ! dummy values
       profiles(prof)%ellipsoid_ground_level = 456.78 ! dummy values
       
    !  #]
    END DO fillloop

    !  #[ clean up after the work is done
    deallocate(knmi_z_base,knmi_z_mid,knmi_z_top)
    deallocate(knmi_p,knmi_T)
    deallocate(knmi_u,knmi_v,knmi_w)

    IF (associated(Interp_P)) THEN
       DO i=1,size(Interp_P)
          call ProfileInterpolateDeleteProfile(Interp_P(i))
       END DO
       deallocate(Interp_P)
    END IF
    IF (associated(Interp_T)) THEN
       DO i=1,size(Interp_T)
          call ProfileInterpolateDeleteProfile(Interp_T(i))
       END DO
       deallocate(Interp_T)
    END IF
    IF (associated(Interp_u)) THEN
       DO i=1,size(Interp_u)
          call ProfileInterpolateDeleteProfile(Interp_u(i))
       END DO
       deallocate(Interp_u)
    END IF
    IF (associated(Interp_v)) THEN
       DO i=1,size(Interp_v)
          call ProfileInterpolateDeleteProfile(Interp_v(i))
       END DO
       deallocate(Interp_v)
    END IF
    IF (associated(Interp_w)) THEN
       DO i=1,size(Interp_w)
          call ProfileInterpolateDeleteProfile(Interp_w(i))
       END DO
       deallocate(Interp_w)
    END IF
    !  #]
    return
    !  #[ error handlers 

    ! print warning and report the error in case a read statement failed
997 print *,"Could not properly read inputfile: ",trim(knmi_file)
    ! printing of these dummy variables suppresses several warnings about 
    ! variables that are set but never used, when compiling with g95, and
    ! since these print may also be usefull for debugging, I added them here
    print *,"beta_g, B, albedo, mask = ",beta_g, B, albedo, mask
    print *,"astuff150 = ["//trim(astuff150)//"]"

    error_flag = error_reading_file
    return

    ! print warning and report the error in case the file could not be opened
998 print *,"Could not open inputfile: ",trim(knmi_file)
    error_flag = error_opening_file
    return

    ! print a warning and report the error in case of allocation problems
999 print *,"ERROR: allocation failure in: GetAuxiliaryDataKnmi"
    error_flag = error_allocate
    return
    !  #]
  end subroutine GetAuxiliaryDataKnmi
    !  #]
  subroutine TestPrintAuxiliaryData(RawProfile)
    !  #[
    ! inputs
    TYPE(profile_type) :: RawProfile
    ! outputs none

    ! local variables
    integer :: i,j,nelements

    ! generate some testprints

       ! TODO: remove variable i and print when
       !       testoutput changes are acceptable
       i = 1
       print *,"contents of profile ",i

       print *,"REALACC(6) profile, lat, lon = ", &
                i, ", ",  RawProfile%lat, ", "
       print *, RawProfile%lon," ENDREALACC"

       IF (associated(RawProfile%z)) THEN
          nelements = size(RawProfile%z)
          print *,"z = "
          print *,"REALACC(5) ",(RawProfile%z(j),j=1,nelements),"ENDREALACC"
       ELSE
          print *,"z = [not associated]"
       END IF

       IF (associated(RawProfile%T_nom)) THEN
          nelements = size(RawProfile%T_nom)
          print *,"T_nom = "
          print *,"REALACC(5) ",(RawProfile%T_nom(j),j=1,nelements)," ENDREALACC"
       ELSE
          print *,"T_nom = [not associated]"
       END IF

       IF (associated(RawProfile%P_nom)) THEN
          nelements = size(RawProfile%P_nom)
          print *,"P_nom = "
          print *,"REALACC(5) ",(RawProfile%P_nom(j),j=1,nelements)," ENDREALACC"
       ELSE
          print *,"P_nom = [not associated]"
       END IF

  end subroutine TestPrintAuxiliaryData
    !  #]
  subroutine Init_Profile(profile,n_model_layers,error_flag)
    !  #[

    TYPE(profile_type), intent(out) :: profile ! output
    integer,            intent(in)  :: n_model_layers
    integer,            intent(out) :: error_flag

    ! local variables
    integer :: AllocStatus

    error_flag = no_error

    profile%lat     = missing_indicator_real
    profile%lon     = missing_indicator_real
    call InitDateTime(profile%datetime)
    profile%us      = missing_indicator_real
    profile%vs      = missing_indicator_real
    profile%ps      = missing_indicator_real
    profile%err_ps  = missing_indicator_real
    profile%zs      = missing_indicator_real
    profile%nlevels = -1
    nullify(profile%z)
    nullify(profile%T_nom)
    nullify(profile%P_nom)
    nullify(profile%P_base)
    nullify(profile%P_top)
    nullify(profile%z_base)
    nullify(profile%z_top)
    nullify(profile%err_T)
    nullify(profile%u)
    nullify(profile%v)
    nullify(profile%rh)
    nullify(profile%err_rh)
    nullify(profile%q)
    nullify(profile%cc)
    nullify(profile%clwc)
    nullify(profile%ciwc)
    profile%geoid_ground_level     = missing_indicator_real
    profile%ellipsoid_ground_level = missing_indicator_real
    profile%Computed_L1B_Location  = .false.
    profile%OffNadir               = .false.

    profile%nlevels = n_model_layers

    !---------------------------
    ! Allocate profile arrays
    !---------------------------
    ! TODO: check that err_T is not needed in the baseline L2BP    
    ! (Poli, April 28 2006)
    allocate(profile%z(n_model_layers)    ,&
             profile%P_nom(n_model_layers),&
             profile%T_nom(n_model_layers),&
             profile%u(n_model_layers)    ,&
             profile%v(n_model_layers)    ,&
    !        profile%err_T(n_model_layers),&
             STAT=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       print *,"Allocation problem for profile in Init_Profile()"
       error_flag = error_allocate
       return
    END IF

  end subroutine Init_Profile

    !  #]
  subroutine Delete_Profile(profile)
    !  #[
    TYPE(profile_type), intent(inout) :: profile

    IF (associated(profile%z)) deallocate(profile%z)
    IF (associated(profile%P_nom)) deallocate(profile%P_nom)
    IF (associated(profile%T_nom)) deallocate(profile%T_nom)
    IF (associated(profile%P_base)) deallocate(profile%P_base)
    IF (associated(profile%P_top)) deallocate(profile%P_top)
    IF (associated(profile%z_base)) deallocate(profile%z_base)
    IF (associated(profile%z_top)) deallocate(profile%z_top)
    IF (associated(profile%err_T)) deallocate(profile%err_T)
    IF (associated(profile%u)) deallocate(profile%u)
    IF (associated(profile%v)) deallocate(profile%v)
    IF (associated(profile%rh)) deallocate(profile%rh)
    IF (associated(profile%err_rh)) deallocate(profile%err_rh)
    IF (associated(profile%q)) deallocate(profile%q)
    IF (associated(profile%cc)) deallocate(profile%cc)
    IF (associated(profile%clwc)) deallocate(profile%clwc)
    IF (associated(profile%ciwc)) deallocate(profile%ciwc)

  end subroutine Delete_Profile
    !  #]
  subroutine DeallocAuxiliaryData(profiles) ! input
    !  #[

    ! inputs
    TYPE(profile_type), dimension(:), pointer :: profiles

    ! outputs none
    ! side effect: the profiles will be deallocated after this call

    ! local variables
    integer :: i

    IF (.not. associated(profiles)) THEN
       IF (LocalVerbosity .ge. VerbosityWarn) THEN
          print *,"WARNING in dealloc_auxiliary_data:"
          print *,"==> the profiles array seems not associated to any memory"
          print *,"nothing to do ..."
       END IF
       return
    END IF

    if (associated(profiles)) then
       DO i=1,size(profiles)
          call Delete_Profile(profiles(i))
       END DO

       ! then deallocate the struct itself
       deallocate(profiles)
       nullify(profiles)
    end if

  end subroutine DeallocAuxiliaryData

    !  #]
  subroutine RawProfile_to_AMD_DataStructure( &
                               fill_off_nadir_flag, & ! input 
                               RawProfile,          & ! input
                               AMD,                 & ! output
                               error_flag           ) ! output

    !  #[ transfer values from RawProfile to AMD
    
    ! inputs
    logical           , intent(in) :: fill_off_nadir_flag
    TYPE(profile_type), intent(in) :: RawProfile

    ! outputs
    TYPE(AMD_BRC_DataType), intent(inout) :: AMD
    integer,            intent(out)   :: error_flag

    ! local variables
    TYPE(AMD_Geoloc_ADS_Type), pointer :: Geoloc_ADS_selected
    TYPE(AMD_Met_MDS_Type),    pointer ::    Met_MDS_selected
    integer                            :: i_layer
    real(r8_)                          :: rh_positive, cc_positive

    error_flag = no_error

    if ( fill_off_nadir_flag ) then
       Geoloc_ADS_selected => AMD%Geoloc_ADS_off_nadir
       Met_MDS_selected    => AMD%Met_MDS_off_nadir
    else
       Geoloc_ADS_selected => AMD%Geoloc_ADS_nadir
       Met_MDS_selected    => AMD%Met_MDS_nadir
    end if

! TO DO: some logic to select nadir instead.
!        Needs an input argument.

    Geoloc_ADS_selected%DateTime  = RawProfile%DateTime
    Geoloc_ADS_selected%Latitude  = FAdoxy_to_IntAl(RawProfile%lat,scale=1000000._r8_)
    Geoloc_ADS_selected%Longitude = FAdoxy_to_IntAl(RawProfile%lon,scale=1000000._r8_)

!! TODO: change start_of_obs_datetime to DateTime
!    Met_MDS_selected%Start_of_Obs_DateTime     = RawProfile%DateTime
! Num_of_Model_Layers?
    Met_MDS_selected%us           = FAdoxy_to_IntAs( RawProfile%us,scale=100._r8_)
    Met_MDS_selected%vs           = FAdoxy_to_IntAs( RawProfile%vs,scale=100._r8_)
    Met_MDS_selected%ps           = FAdoxy_to_IntAul(RawProfile%ps)
    Met_MDS_selected%err_ps       = RawProfile%err_ps
    Met_MDS_selected%zs           = FAdoxy_to_IntAl( RawProfile%zs,scale=100._r8_)
!    Met_MDS_selected%zs           = maxval_IntAl ! missing

    do i_layer = 1, size(RawProfile%P_base)
       Met_MDS_selected%Met_Layer(i_layer)%Validity_Flag = -2 ! -1 when p,z,T all valid, 0 when everything valid

       Met_MDS_selected%Met_Layer(i_layer)%pbase = FAdoxy_to_IntAul(RawProfile%P_base(i_layer) )
       Met_MDS_selected%Met_Layer(i_layer)%ptop  = FAdoxy_to_IntAul(RawProfile%P_top (i_layer) )
       Met_MDS_selected%Met_Layer(i_layer)%pnom  = FAdoxy_to_IntAul(RawProfile%P_nom (i_layer) )

       Met_MDS_selected%Met_Layer(i_layer)%zbase = FAdoxy_to_IntAl( RawProfile%z_base(i_layer),scale=100._r8_)
       Met_MDS_selected%Met_Layer(i_layer)%ztop  = FAdoxy_to_IntAl( RawProfile%z_top (i_layer),scale=100._r8_)
       Met_MDS_selected%Met_Layer(i_layer)%znom  = FAdoxy_to_IntAl( RawProfile%z     (i_layer),scale=100._r8_)
!        Met_MDS_selected%Met_Layer(i_layer)%zbase  = maxval_IntAl ! missing
!        Met_MDS_selected%Met_Layer(i_layer)%ztop   = maxval_IntAl ! missing
!        Met_MDS_selected%Met_Layer(i_layer)%znom   = maxval_IntAl ! missing

       Met_MDS_selected%Met_Layer(i_layer)%T     = FAdoxy_to_IntAus( RawProfile%T_nom (i_layer),scale=100._r8_)
       Met_MDS_selected%Met_Layer(i_layer)%err_T = FAdoxy_to_IntAus( RawProfile%err_T (i_layer),scale=100._r8_)

       Met_MDS_selected%Met_Layer(i_layer)%u     = FAdoxy_to_IntAs( RawProfile%u (i_layer),scale=100._r8_)
       Met_MDS_selected%Met_Layer(i_layer)%v     = FAdoxy_to_IntAs( RawProfile%v (i_layer),scale=100._r8_)

       rh_positive                               = max(        RawProfile%rh(i_layer), 0._r8_ )
       Met_MDS_selected%Met_Layer(i_layer)%rh    = FAdoxy_to_IntAuc( rh_positive,scale=100._r8_)

       Met_MDS_selected%Met_Layer(i_layer)%err_rh=              RawProfile%err_rh(i_layer) * 100._FAdoxy
       Met_MDS_selected%Met_Layer(i_layer)%q     =              RawProfile%q     (i_layer)

       cc_positive                               = max(         RawProfile%cc(i_layer), 0._r8_ )
       Met_MDS_selected%Met_Layer(i_layer)%cc    = FAdoxy_to_IntAuc( cc_positive,scale=100._r8_)
       Met_MDS_selected%Met_Layer(i_layer)%clwc  =              RawProfile%clwc  (i_layer)
       Met_MDS_selected%Met_Layer(i_layer)%ciwc  =              RawProfile%ciwc  (i_layer)

    end do

  end subroutine RawProfile_to_AMD_DataStructure
    !  #]
  subroutine Copy_AMD_to_WorkingData(from_off_nadir_flag, &
                                     AMD_DSR,&
                                     L1B,WorkingData,&
                                             error_flag)
    !  #[

    logical           ,     intent(in)    :: from_off_nadir_flag ! input
    TYPE(AMD_BRC_DataType), intent(in)    :: AMD_DSR             ! input
    TYPE(L1B_BRC_DataType), intent(in)    :: L1B
    TYPE(WorkingDataType),  intent(inout) :: WorkingData ! input (for the altitudes)
                                                         ! output (working data)
    integer,                intent(out)   :: error_flag

    ! local variables
    TYPE(profile_type) :: raw_profile
    integer            :: n_model_layers      ! input

    error_flag = no_error

    ! determine n_model_layers
    n_model_layers = 0
    IF (associated(AMD_DSR%Met_MDS_off_nadir%Met_Layer)) &
       n_model_layers=maxval((/n_model_layers,size(AMD_DSR%Met_MDS_off_nadir%Met_Layer)/))
    IF (associated(AMD_DSR%Met_MDS_nadir%Met_Layer)) &
       n_model_layers=maxval((/n_model_layers,size(AMD_DSR%Met_MDS_nadir%Met_Layer)/))

    IF (n_model_layers < 1) THEN
       print *,"ERROR: maxval((/size(AMD_DSR%Met_MDS_off_nadir%Met_Layer), "
       print *,"                size(AMD_DSR%Met_MDS_nadir%Met_Layer)/))"
       print *,"n_model_layers < 1 inside Copy_AMD_to_WorkingData!!!"
       error_flag = error_programming
       return
    ENDIF

    call Init_Profile(raw_profile,n_model_layers,error_flag)
    IF (error_flag .ne. no_error) return

    call Copy_AMD_to_RawProfile(from_off_nadir_flag, AMD_DSR, &
                                raw_profile, error_flag)
    IF ( error_flag .ne. no_error ) return

    call Interp_RawProf_to_WorkingData(&
                                raw_profile, L1B, &
                                WorkingData, error_flag)
    IF ( error_flag .ne. no_error ) return

    call Delete_Profile(raw_profile)

  end subroutine Copy_AMD_to_WorkingData

    !  #]
  subroutine Copy_AMD_to_RawProfile(from_off_nadir_flag, AMD_DSR, &
                                    RawProfile, error_flag)
    !  #[
    logical           ,     intent(in)    :: from_off_nadir_flag ! input
    TYPE(AMD_BRC_DataType), intent(in)    :: AMD_DSR             ! input

    TYPE(profile_type), intent(out)   :: RawProfile          ! output
    integer,            intent(out)   :: error_flag          ! output

    ! local variables
    TYPE(AMD_Geoloc_ADS_type), pointer :: AMD_Geoloc_ADS_Selected
    TYPE(AMD_Met_MDS_type),    pointer :: AMD_Met_MDS_Selected
    integer :: i_layer, n_model_layers
    integer :: n_raw_profiles ! only needed for testoutput preservation

    error_flag = no_error

    !-------------------------------------
    ! Determine the number of model layers
    !-------------------------------------
    n_model_layers = RawProfile%nlevels

    !-------------------------------------
    ! Point to off_nadir or nadir datasets
    !-------------------------------------

    IF ( from_off_nadir_flag ) THEN
       AMD_Geoloc_ADS_selected => AMD_DSR%Geoloc_ADS_off_nadir
       AMD_Met_MDS_selected    => AMD_DSR%Met_MDS_off_nadir
    ELSE
       AMD_Geoloc_ADS_selected => AMD_DSR%Geoloc_ADS_nadir
       AMD_Met_MDS_selected    => AMD_DSR%Met_MDS_nadir
    END IF

    !---------------------------------------------------
    ! Copy from AMD_DSR to RawProfile non-array elements
    !---------------------------------------------------

    RawProfile%DateTime = AMD_Geoloc_ADS_selected%DateTime
    RawProfile%lat      = AMD_Geoloc_ADS_selected%Latitude  * 1.e-6_r8_
    RawProfile%lon      = AMD_Geoloc_ADS_selected%Longitude * 1.e-6_r8_

    RawProfile%us = AMD_Met_MDS_selected%us * 1.e-2_r8_
    RawProfile%vs = AMD_Met_MDS_selected%vs * 1.e-2_r8_
    RawProfile%ps = real(AMD_Met_MDS_selected%ps,r8_)
    RawProfile%err_ps = AMD_Met_MDS_selected%err_ps
    RawProfile%zs = AMD_Met_MDS_selected%zs * 1.e-2_r8_ ! activate when not missing

    !---------------------------------------
    ! Copy from AMD_DSR to RawProfile arrays
    !---------------------------------------

    layer_loop: DO i_layer = 1, n_model_layers
       RawProfile%z(i_layer)    = AMD_Met_MDS_selected%met_layer(i_layer)%znom * 1.e-2_r8_
       RawProfile%P_nom(i_layer)= AMD_Met_MDS_selected%met_layer(i_layer)%pnom
       RawProfile%T_nom(i_layer)= AMD_Met_MDS_selected%met_layer(i_layer)%t * 1.e-2_r8_
       RawProfile%u(i_layer)    = AMD_Met_MDS_selected%met_layer(i_layer)%u
       RawProfile%v(i_layer)    = AMD_Met_MDS_selected%met_layer(i_layer)%v
    END DO layer_loop


       ! Some prints to preserve testoutput
       ! TODO: Decide whether to remove the following 
       !       prints, resulting in testoutput changes
       !
       print *, 'Conversion from AMD_DataStructure to RawProfile ...'
       n_raw_profiles = 1
       print *,"n_raw_profiles = ",n_raw_profiles
       call TestPrintAuxiliaryData(RawProfile)

  end subroutine Copy_AMD_to_RawProfile
    !  #]
  subroutine Interp_RawProf_to_WorkingData( &
                                     raw_profile, L1B, &
                                     WorkingData, error_flag)

    !  #[
    TYPE(profile_type),     intent(in)    :: raw_profile
    TYPE(L1B_BRC_DataType), intent(in)    :: L1B
    TYPE(WorkingDataType),  intent(inout) :: WorkingData ! input (for the altitudes)
                                                         ! output (working data)
    integer,                intent(out)   :: error_flag

    ! local variables
    TYPE(WD_MieRangeGateType),      pointer :: WD_MieRangeGate 
    TYPE(WD_RayleighRangeGateType), pointer :: WD_RayleighRangeGate 
    integer :: m, r

    ! this is no longer a parameter, but the actual value, 
    ! read from the L1B file, and used for dimensioning the Working datastructure
    integer   :: N_meas
    integer   :: N_range_gates
    real(r8_) :: altitude_mid_wgs84, geoid_above_wgs84 ! Both relative to WGS84
    real(r8_) :: altitude_mid                          ! Relative to geoid

    ! to enable interpolation using the ProfileInterpolate module
    type(ProfileInterpolate_type) :: Interp_P_nom
    type(ProfileInterpolate_type) :: Interp_T_nom
    type(ProfileInterpolate_type) :: Interp_u
    type(ProfileInterpolate_type) :: Interp_v

! init
    error_flag = no_error
    N_meas = size(WorkingData%MieMeas)
    N_range_gates = size(WorkingData%MieMeas(1)%RangeGate) ! 24 ! current used nr of gates    

    ! Get the geoid offset from L1B data for conversion of
    ! L1B altitudes to NWP geopotential height in interpolations
    !
    geoid_above_wgs84 = L1B%Geoloc_ADS%Obs_Geoloc%Geoid_Separation_of_Height_Bin

    IF (.not. associated(raw_profile%z)) THEN
       print *,"ERROR in Interp_RawProf_to_WorkingData():"
       print *,"raw_profile%z seems not associated, so no NWP data is"
       print *," available! or if they were, vertical interpolation "
       print *,"is impossible....."
       error_flag = error_programming
       return
    END IF

    IF (.not. associated(raw_profile%P_nom)) THEN
       print *,"ERROR in Interp_RawProf_to_WorkingData():"
       print *,"raw_profile%P_nom seems not associated..."
       error_flag = error_programming
       return
    END IF

    IF (.not. associated(raw_profile%T_nom)) THEN
       print *,"ERROR in Interp_RawProf_to_WorkingData():"
       print *,"raw_profile%T_nom seems not associated..."
       error_flag = error_programming
       return       
    END IF

    IF (.not. associated(raw_profile%u)) THEN
       print *,"ERROR in Interp_RawProf_to_WorkingData():"
       print *,"raw_profile%u seems not associated..."
       error_flag = error_programming
       return
    END IF

    IF (.not. associated(raw_profile%v)) THEN
       print *,"ERROR in Interp_RawProf_to_WorkingData():"
       print *,"raw_profile%v seems not associated..."
       error_flag = error_programming
       return
    END IF

    call ProfileInterpolate_Init_Profile(Interp_P_nom,&
         raw_profile%z,raw_profile%P_nom,raw_profile%nlevels,&
         "pressure profile",&
         error_flag,SkipMissingData=.true.,&
         interpolation_type=interp_type_spline)
! TODO: document sensitivity to inter/extra-polation_type in TN4.1
! TODO: make such choices part of AUX_PAR_2B?
!         interpolation_type=interp_type_log,&
!         extrapolation_type_down=extrap_type_log,&
!         extrapolation_type_up=extrap_type_log)

    call ProfileInterpolate_Init_Profile(Interp_T_nom,&
         raw_profile%z,raw_profile%T_nom,raw_profile%nlevels,&
         "temperature profile",&
         error_flag,SkipMissingData=.true.,&
         interpolation_type=interp_type_spline)
! TODO: unrevert when updates to expected outputs and sensitivity checked
! TODO: make such choices part of AUX_PAR_2B?
!         interpolation_type=interp_type_spline,&
!         extrapolation_type_down=extrap_type_edgeval,&
!         extrapolation_type_up=extrap_type_edgeval)

    call ProfileInterpolate_Init_Profile(Interp_u,&
         raw_profile%z,raw_profile%u,raw_profile%nlevels,&
         "u-component profile",&
         error_flag,SkipMissingData=.true.,&
         interpolation_type=interp_type_spline,&
         extrapolation_type_down=extrap_type_edgeval,&
         extrapolation_type_up=extrap_type_edgeval)

    call ProfileInterpolate_Init_Profile(Interp_v,&
         raw_profile%z,raw_profile%v,raw_profile%nlevels,&
         "v-component profile",&
         error_flag,SkipMissingData=.true.,&
         interpolation_type=interp_type_spline,&
         extrapolation_type_down=extrap_type_edgeval,&
         extrapolation_type_up=extrap_type_edgeval)

    meas_loop: DO m=1,N_meas
       rangegate_loop: DO r=1,N_range_gates
          ! set some pointers for shorter notation to the working data structs
          WD_RayleighRangeGate => WorkingData%RayleighMeas(m)%RangeGate(r)
          WD_MieRangeGate      => WorkingData%MieMeas(     m)%RangeGate(r)

          !=============================================================
          ! copy the needed Rayleigh data
          !=============================================================

          altitude_mid_wgs84 = &
               (WorkingData%RayleighMeas(m)%RangeGate(r)%alt_start + &
                WorkingData%RayleighMeas(m)%RangeGate(r)%alt_end      )/2._r8_
          altitude_mid = altitude_mid_wgs84 - geoid_above_wgs84 ! Now above geoid


          WD_RayleighRangeGate%p = interpolate(Interp_P_nom,altitude_mid)
          WD_RayleighRangeGate%T = interpolate(Interp_T_nom,altitude_mid)
          WD_RayleighRangeGate%u = interpolate(Interp_u,    altitude_mid)
          WD_RayleighRangeGate%v = interpolate(Interp_v,    altitude_mid)

          !=============================================================
          ! copy the needed Mie data
          !=============================================================

          altitude_mid_wgs84 = &
               (WorkingData%MieMeas(m)%RangeGate(r)%alt_start + &
                WorkingData%MieMeas(m)%RangeGate(r)%alt_end     )/2._r8_
          altitude_mid = altitude_mid_wgs84 - geoid_above_wgs84 ! Now above geoid

          WD_MieRangeGate%p = interpolate(Interp_P_nom,altitude_mid)
          WD_MieRangeGate%T = interpolate(Interp_T_nom,altitude_mid)
          WD_MieRangeGate%u = interpolate(Interp_u,    altitude_mid)
          WD_MieRangeGate%v = interpolate(Interp_v,    altitude_mid)

       END DO rangegate_loop
    END DO meas_loop

    
    call ProfileInterpolateDeleteProfile(Interp_P_nom)
    call ProfileInterpolateDeleteProfile(Interp_T_nom)
    call ProfileInterpolateDeleteProfile(Interp_u)
    call ProfileInterpolateDeleteProfile(Interp_v)

  end subroutine Interp_RawProf_to_WorkingData
    !  #]
  !------------------------------------------
end module AuxiliaryModule
