module XML_module
  !  #[ Description:
  !     A module containing interface routines with the
  !     explorer_file_handling library from the 
  !     ee_cfi_software package.
  !
  !     created by:  J. de Kloe, KNMI
  !     modifications:
  !        21-Jun-2005 D. Tan     Take model_name and version_number from input
  !        03-Jan-2006 J. de Kloe added code to force setting SaveNoEmptyTags
  !        02-Mar-2006 J. de Kloe added xml_tree_add_and_set_attribute
  !                               and added use of the xml_node_type
  !        04-Apr-2006 D. Tan     Cleanup in eof/error handling
  !        12-May-2006 D. Tan     Null use of ParKind1 module
  !        21-Jul-2006 J. de Kloe adapted to ee_cfi 3.5
  !        28-Jul-2006 J. de Kloe added deletion of tempfile and use of pid
  !        09-Mar-2007 D. Tan     Update ee_mph_name, ee_sph_name, make public
  !        07-May-2007 J. de Kloe added routine xml_check_node_exists
  !        11-Jun-2007 J. de Kloe adapted to new requirement that HDR files no
  !                               longer use Earth_Explorer_File root tag
  !        20-Jul-2007 J. de Kloe added subroutine xml_tree_set_schema()
  !        05-Nov-2007 J. de Kloe added subroutine xml_tree_set_namespace()
  !        12-Nov-2007 J. de Kloe implemented routine xml_tree_get_namespace
  !                               (a workaround for now that bypasses the
  !                                ee_cfi library)
  !        16-Jan-2008 J. de Kloe phase out integer kind i_ 
  !        28-Jan-2008 J. de Kloe added initialisation of pos variable
  !                               to suppres an uninitialized warning in g95
  !  #]
  !  #[ Modules used:
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE ErrorHandler, only: no_error, error_xml, error_reading_file, &
                          error_programming, error_opening_file,&
                          error_allocate, error_c_interface
  USE c_support, only: remove_file, get_process_id
  USE LunManager, only: get_lun, free_lun
  USE numerics, only: i4_
  !  #]
  !  #[ Parameter/Variable/Type declarations
  implicit none

  ! make all module contents private, unless explicitely stated otherwise
  private

  ! some constant parameters
  integer, parameter :: max_str_len    = 512
  integer, parameter :: xml_no_c_error = 0

  ! Earth Explorer specific string constants
  character(len=*), parameter :: ee_root_name = "Earth_Explorer_File"
  character(len=*), parameter :: ee_h_name    = "Earth_Explorer_Header"
  character(len=*), parameter :: ee_fh_name   = "Fixed_Header"
  character(len=*), parameter :: ee_vh_name   = "Variable_Header"
  character(len=*), parameter :: ee_mph_name  = "Main_Product_Header"
  character(len=*), parameter :: ee_sph_name  = "Specific_Product_Header"
  character(len=*), parameter :: uri_ns_header= "http://www.esa.int/schemas/ae/"
 
  ! a little struct to allow having an identical interface 
  ! with the simple_xml module
  type xml_node_type
     private
     integer                    :: x
     character(len=max_str_len) :: xml_schema_ref
     character(len=max_str_len) :: xml_namespace_uri
  end type xml_node_type

  ! the default verbosity setting is Error (as silent as possible)
  integer, private, save :: LocalVerbosity = VerbosityError

  logical, save :: module_is_initialised = .false.
  !  #]
  !  #[ interface definitions of the external c library routines/functions
  ! making the interfaces explicit enables the fortran compiler 
  ! to check the parameter types, and thus prevents programming errors
  ! (this replaces the old external statement)
  interface
     subroutine force_SaveNoEmptyTags()
     end subroutine force_SaveNoEmptyTags
  end interface

  interface
     function xf_tree_create(c_error) result(xml_tree)
       USE numerics, only: i4_
       integer(i4_) :: c_error  ! result
       integer(i4_) :: xml_tree ! result
     end function xf_tree_create
  end interface

  interface
     subroutine xf_tree_create_root(xml_tree,name,c_error) 
       USE numerics, only: i4_
       integer(i4_),     intent(in)  :: xml_tree ! input
       character(len=*), intent(in)  :: name     ! input
       integer(i4_),     intent(out) :: c_error  ! result
     end subroutine xf_tree_create_root
  end interface

  interface
     subroutine xf_tree_write(xml_tree,filename,c_error) 
       USE numerics, only: i4_
       integer(i4_),     intent(in)  :: xml_tree ! input
       character(len=*), intent(in)  :: filename ! input
       integer(i4_),     intent(out) :: c_error  ! result
     end subroutine xf_tree_write
  end interface

  interface
     subroutine xf_set_schema(filename,xml_schema_ref,action,c_error) 
       USE numerics, only: i4_
       character(len=*), intent(in)  :: filename       ! input
       character(len=*), intent(in)  :: xml_schema_ref ! input
       integer(i4_),     intent(in)  :: action         ! input
       integer(i4_),     intent(out) :: c_error        ! result
     end subroutine xf_set_schema
  end interface

  interface
     subroutine xf_error_msg(error_message)
       character(len=*), intent(out)  :: error_message
     end subroutine xf_error_msg
  end interface

  interface
     subroutine xf_tree_cleanup_all_parser()
     end subroutine xf_tree_cleanup_all_parser
  end interface

  interface
     subroutine xf_tree_cleanup_parser(xml_tree,c_error)
       USE numerics, only: i4_
       integer(i4_),     intent(in)  :: xml_tree ! input
       integer(i4_),     intent(out) :: c_error  ! result
     end subroutine xf_tree_cleanup_parser
  end interface

  interface
     subroutine xf_tree_add_child(xml_tree,parent,name,c_error) 
       USE numerics, only: i4_
       integer(i4_),     intent(in)  :: xml_tree ! input
       character(len=*), intent(in)  :: parent   ! input
       character(len=*), intent(in)  :: name     ! input
       integer(i4_),     intent(out) :: c_error  ! result
     end subroutine xf_tree_add_child
  end interface

  interface
     subroutine xf_tree_set_string_node_value(xml_tree,&
          path,text,formatstring,c_error) 
       USE numerics, only: i4_
       integer(i4_),     intent(in)  :: xml_tree     ! input
       character(len=*), intent(in)  :: path         ! input
       character(len=*), intent(in)  :: text         ! input
       character(len=*), intent(in)  :: formatstring ! input
       integer(i4_),     intent(out) :: c_error      ! result
     end subroutine xf_tree_set_string_node_value
  end interface

  interface
     subroutine xf_tree_add_attribute(xml_tree,path,attr,c_error) 
       USE numerics, only: i4_
       integer(i4_),     intent(in)  :: xml_tree     ! input
       character(len=*), intent(in)  :: path         ! input
       character(len=*), intent(in)  :: attr         ! input
       integer(i4_),     intent(out) :: c_error      ! result
     end subroutine xf_tree_add_attribute
  end interface

  interface
     subroutine xf_tree_create_header(xml_tree,header_type,c_error) 
       USE numerics, only: i4_
       integer(i4_),     intent(in)  :: xml_tree    ! input
       integer(i4_),     intent(in)  :: header_type ! input
       integer(i4_),     intent(out) :: c_error     ! result
     end subroutine xf_tree_create_header
  end interface

  interface
     subroutine xf_tree_set_fixed_header_items(xml_tree,&
          file_name, file_description, notes, mission, file_class, file_type,&
          validity_start, validity_stop, file_version, system, creator, &
          creator_version, creation_date, c_error) 
       USE numerics, only: i4_
       integer(i4_),     intent(in)  :: xml_tree         ! input
       character(len=*), intent(in)  :: file_name        ! input
       character(len=*), intent(in)  :: file_description ! input
       character(len=*), intent(in)  :: notes            ! input
       character(len=*), intent(in)  :: mission          ! input
       character(len=*), intent(in)  :: file_class       ! input
       character(len=*), intent(in)  :: file_type        ! input
       character(len=*), intent(in)  :: validity_start   ! input
       character(len=*), intent(in)  :: validity_stop    ! input
       integer(i4_),     intent(in)  :: file_version     ! input
       character(len=*), intent(in)  :: system           ! input
       character(len=*), intent(in)  :: creator          ! input
       character(len=*), intent(in)  :: creator_version  ! input
       character(len=*), intent(in)  :: creation_date    ! input
       integer(i4_),     intent(out) :: c_error          ! result
     end subroutine xf_tree_set_fixed_header_items
  end interface

  interface
     subroutine xf_tree_get_fixed_header_items(xml_tree,&
          file_name, file_description, notes, mission, file_class, file_type,&
          validity_start, validity_stop, file_version, system, creator, &
          creator_version, creation_date, c_error) 
       USE numerics, only: i4_
       integer(i4_),     intent(in)  :: xml_tree         ! input
       character(len=*), intent(out) :: file_name        ! result
       character(len=*), intent(out) :: file_description ! result
       character(len=*), intent(out) :: notes            ! result
       character(len=*), intent(out) :: mission          ! result
       character(len=*), intent(out) :: file_class       ! result
       character(len=*), intent(out) :: file_type        ! result
       character(len=*), intent(out) :: validity_start   ! result
       character(len=*), intent(out) :: validity_stop    ! result
       integer(i4_),     intent(out) :: file_version     ! result
       character(len=*), intent(out) :: system           ! result
       character(len=*), intent(out) :: creator          ! result
       character(len=*), intent(out) :: creator_version  ! result
       character(len=*), intent(out) :: creation_date    ! result
       integer(i4_),     intent(out) :: c_error          ! result
     end subroutine xf_tree_get_fixed_header_items
  end interface

  interface
     function xf_tree_init_parser(filename, c_error) result(xml_tree)
       USE numerics, only: i4_
       integer(i4_)     :: xml_tree    ! input
       character(len=*) :: filename    ! input
       integer(i4_)     :: c_error     ! result
     end function xf_tree_init_parser
  end interface

  interface
     subroutine xf_tp_read_string_node_value(xml_tree,path,value,c_error) 
       USE numerics, only: i4_
       integer(i4_),     intent(in)  :: xml_tree    ! input
       character(len=*), intent(in)  :: path        ! input
       character(len=*), intent(out) :: value       ! result
       integer(i4_),     intent(out) :: c_error     ! result
     end subroutine xf_tp_read_string_node_value
  end interface

  interface
     subroutine xf_tree_go_to_path_node(xml_tree,path,c_error) 
       USE numerics, only: i4_
       integer(i4_),     intent(in)  :: xml_tree    ! input
       character(len=*), intent(in)  :: path        ! input
       integer(i4_),     intent(out) :: c_error     ! result
     end subroutine xf_tree_go_to_path_node
  end interface

  interface
     subroutine xf_tree_rewind(xml_tree,c_error)
       USE numerics, only: i4_
       integer(i4_),     intent(in)  :: xml_tree    ! input
       integer(i4_),     intent(out) :: c_error     ! result
    end subroutine xf_tree_rewind
  end interface

  interface
     subroutine  xf_tree_get_curr_element_name(xml_tree,name,c_error)
       USE numerics, only: i4_
       integer(i4_),     intent(in)  :: xml_tree    ! input
       character(len=*), intent(out) :: name        ! output
       integer(i4_),     intent(out) :: c_error     ! output
     end subroutine xf_tree_get_curr_element_name
  end interface

  interface
     function get_xf_header_format_eef() result(header_type)
       USE numerics, only: i4_
       integer(i4_) :: header_type ! result
     end function get_xf_header_format_eef
  end interface

  interface
     function get_xf_header_format_hdr() result(header_type)
       USE numerics, only: i4_
       integer(i4_) :: header_type ! result
     end function get_xf_header_format_hdr
  end interface

  interface
     function get_xf_max_error_msg_length() result(max_err_length)
       USE numerics, only: i4_
       integer(i4_) :: max_err_length ! result
     end function get_xf_max_error_msg_length
  end interface

  interface
     function get_xf_max_value_length() result(max_val_length)
       USE numerics, only: i4_
       integer(i4_) :: max_val_length ! result
     end function get_xf_max_value_length
  end interface

  interface
     function get_xf_max_path_length() result(max_path_length)
       USE numerics, only: i4_
       integer(i4_) :: max_path_length ! result
     end function get_xf_max_path_length
  end interface

  interface
     function get_xf_cfi_ok() result(cfi_ok_code)
       USE numerics, only: i4_
       integer(i4_) :: cfi_ok_code ! result
     end function get_xf_cfi_ok
  end interface

  interface
     function get_xf_cfi_empty_element() result(cfi_empty_element)
       USE numerics, only: i4_
       integer(i4_) :: cfi_empty_element ! result
     end function get_xf_cfi_empty_element
  end interface

  interface
     function get_xf_cfi_bad_input_argument() result(cfi_bad_input_argument)
       USE numerics, only: i4_
       integer(i4_) :: cfi_bad_input_argument ! result
     end function get_xf_cfi_bad_input_argument
  end interface

  interface
     function get_xf_change_schema_and_vers() result(xf_change_schema_and_version)
       USE numerics, only: i4_
       integer(i4_) :: xf_change_schema_and_version ! result
     end function get_xf_change_schema_and_vers
  end interface
  !  #]             
  !  #[ Module interface
  ! specify which routines are public, so are part of the interface
  ! to this module
  ! public parameters
  public :: max_str_len, ee_root_name, ee_vh_name, &
            ee_mph_name, ee_sph_name, ee_fh_name, ee_h_name, &
            uri_ns_header
  ! public types
  public ::  xml_node_type
  ! public functions/subroutines
  public :: Init_XML, create_xml_tree, delete_xml_tree, &
            xml_tree_create_header, xml_set_fixed_header, &
            xml_get_fixed_header, xml_tree_create_root, &
            xml_tree_add_child, xml_tree_add_attribute,&
            xml_tree_set_string_node_value, xml_node_get_string, &
            xml_tree_write, xml_load_file, xml_tree_print, &
            check_length_xml_item, check_string_length, &
            xml_tree_read_string_attribute, xml_add_SPH_item, &
            xml_tree_add_and_set_attribute, xml_check_node_exists, &
            xml_tree_set_namespace, xml_tree_get_namespace
  !  #]
contains
  !------------------------------------------
  ! public functions/subroutines
  subroutine Init_XML(error_flag, ChoosenVerbosity)
    !  #[

    integer,           intent(out) :: error_flag
    integer, optional, intent(in)  :: ChoosenVerbosity

    error_flag = no_error

    IF (present(ChoosenVerbosity)) LocalVerbosity = ChoosenVerbosity

    if (LocalVerbosity >= VerbosityWarn) &
         call print_verbosity_setting(LocalVerbosity,&
                                      ModuleName="Classification")

    ! get some numbers from the explorer_file_handling.h header file
    ! and check whether we have set our constants correct or not
    IF (xml_no_c_error .ne. get_xf_cfi_ok()) THEN
       print *,"ERROR: the value of the xml_no_c_error parameter is wrong!"
       print *,"the module xml_module (in ee_cfi_wrapper) will not work"
       print *,"correctly until this is fixed !!!"
       error_flag = error_programming
       return
    END IF
    IF (max_str_len .lt. get_xf_max_value_length()) THEN
       print *,"WARNING: the value of max_str_len may be too low!"
       print *,"WARNING: max_str_len = ",max_str_len
       print *,"but get_xf_max_value_length() = ",&
                    get_xf_max_value_length()
    END IF
    IF (max_str_len .lt. get_xf_max_path_length()) THEN
       print *,"WARNING: the value of max_str_len may be too low!"
       print *,"WARNING: max_str_len = ",max_str_len
       print *,"but get_xf_max_path_length() = ",&
                    get_xf_max_path_length()
    END IF
    IF (max_str_len .lt. get_xf_max_error_msg_length()) THEN
       print *,"WARNING: the value of max_str_len may be too low!"
       print *,"WARNING: max_str_len = ",max_str_len
       print *,"but get_xf_max_error_msg_length() = ",&
                    get_xf_max_error_msg_length()
    END IF

    call force_SaveNoEmptyTags()

    module_is_initialised = .true.

  end subroutine Init_XML
    !  #]
  subroutine create_xml_tree(xml_tree, error_flag)
    !  #[ create a new empty tree
    TYPE(xml_node_type), pointer :: xml_tree
    !integer, intent(out) :: xml_tree
    integer, intent(out) :: error_flag

    ! local variables
    integer      :: status
    integer(i4_) :: c_error

    error_flag = no_error

    IF (.not. module_is_initialised) THEN
       call Init_XML(error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    allocate(xml_tree, STAT=status)
    if (status .ne. 0) then
       print *,"allocation problem in function create_xml_tree"
       nullify(xml_tree)
       error_flag = error_allocate
       return
    end if

    c_error = xml_no_c_error ! init to be sure
    xml_tree%x = xf_tree_create(c_error)
    IF (c_error .lt. xml_no_c_error) THEN
       print *,"c_error = ",c_error
       print *,"occurred when calling: xf_tree_create()"
       call xml_print_error(c_error)
       error_flag = error_xml
       return
    END IF

    xml_tree%xml_schema_ref(:) = ' '
    xml_tree%xml_namespace_uri(:) = ' '

  end subroutine create_xml_tree
    !  #]
  subroutine delete_xml_tree(xml_tree, error_flag)
    !  #[ free the memory used by the xml tree
    TYPE(xml_node_type), pointer :: xml_tree
    !integer, intent(in)  :: xml_tree
    integer, intent(out) :: error_flag

    ! local variables
    integer(i4_) :: c_error

    error_flag = no_error

    IF (.not. associated(xml_tree) ) THEN
       print *,"problem: tree seems not associated ..."
       error_flag=error_programming
       return
    END IF

    ! WARNING: this xf routine seems to forget to reset the error
    ! flag in case of no errors, which may cause problems, therefore
    ! we should always initialise the error to 0 before
    ! calling any of the xf routines ....
    c_error = xml_no_c_error ! init to be sure
    call xf_tree_cleanup_parser(xml_tree%x, c_error)
    IF (c_error .lt. xml_no_c_error) THEN
       print *,"c_error = ",c_error
       print *,"occurred when calling: xf_tree_cleanup_parser()"
       call xml_print_error(c_error)
       error_flag = error_xml
       return
    END IF

    deallocate(xml_tree)
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"xml_tree has been deleted"

  end subroutine delete_xml_tree
    !  #]
  subroutine xml_tree_create_header(xml_tree, xml_root, error_flag)
    !  #[ create a blank header (fixed+variable)
    TYPE(xml_node_type), pointer  :: xml_tree
    !integer,          intent(in)  :: xml_tree
    character(len=*), intent(in)  :: xml_root
    integer,          intent(out) :: error_flag

    ! local variable
    integer(i4_) :: header_type, c_error

    error_flag = no_error

    IF (.not. associated(xml_tree) ) THEN
       print *,"problem: tree seems not associated ..."
       error_flag=error_programming
       return
    END IF

    ! NOTE: the header may exist below the<Earth_Explorer_File> root tag
    !       or just start with the file <Earth_Explorer_Header>
    !       depending on whether we are generating a .EEF or a .HDR file

    IF ( (trim(xml_root) .ne. ee_h_name)    .and. &
         (trim(xml_root) .ne. ee_root_name)       )THEN
       print *,"ERROR in xml_tree_create_header():"
       print *,"expected the xml_root to be: ",ee_root_name
       print *,"or:                          ",ee_h_name
       print *,"but found: ",trim(xml_root)
       print *,"These header routines are for reading/writing"
       print *,"the ee_cfi xml format only !"
       error_flag = error_programming
       return
    END IF

    if (trim(xml_root) .eq. ee_root_name) &
         header_type = get_xf_header_format_eef()
    if (trim(xml_root) .eq. ee_h_name) &
         header_type = get_xf_header_format_hdr()

    c_error = xml_no_c_error ! init to be sure
    call xf_tree_create_header(xml_tree%x, header_type, c_error)
    IF (c_error .lt. xml_no_c_error) THEN
       print *,"c_error = ",c_error
       print *,"occurred when calling: xf_tree_create_header()"
       call xml_print_error(c_error)
       error_flag = error_xml
       return
    END IF

  end subroutine xml_tree_create_header
    !  #]
  subroutine xml_set_fixed_header(xml_tree, path_FH, &
       file_name, file_description, &
       notes, mission, file_class, file_type, validity_start, validity_stop,&
       file_version, system, creator, creator_version, creation_date, &
       error_flag)
    !  #[
    TYPE(xml_node_type), pointer :: xml_tree     ! modified after return
    !integer,          intent(in)  :: xml_tree
    character(len=*), intent(in)  :: path_FH     ! input
    character(len=*), intent(in)  :: file_name
    character(len=*), intent(in)  :: file_description
    character(len=*), intent(in)  :: notes
    character(len=*), intent(in)  :: mission
    character(len=*), intent(in)  :: file_class
    character(len=*), intent(in)  :: file_type 
    character(len=*), intent(in)  :: validity_start
    character(len=*), intent(in)  :: validity_stop
    integer,          intent(in)  :: file_version
    character(len=*), intent(in)  :: system
    character(len=*), intent(in)  :: creator
    character(len=*), intent(in)  :: creator_version
    character(len=*), intent(in)  :: creation_date
    integer,          intent(out) :: error_flag

    ! local variables
    integer(i4_) :: c_error

    error_flag = no_error

    IF (.not. associated(xml_tree) ) THEN
       print *,"problem: tree seems not associated ..."
       error_flag=error_programming
       return
    END IF

    c_error = xml_no_c_error ! init to be sure
    call xf_tree_set_fixed_header_items(xml_tree%x, &
         string_f90_to_c(file_name),      string_f90_to_c(file_description),&
         string_f90_to_c(notes),          string_f90_to_c(mission), &
         string_f90_to_c(file_class),     string_f90_to_c(file_type), &
         string_f90_to_c(validity_start), string_f90_to_c(validity_stop) ,&
         file_version,                    string_f90_to_c(system), &
         string_f90_to_c(creator),        string_f90_to_c(creator_version), &
         string_f90_to_c(creation_date),  c_error)
    IF (c_error .lt. xml_no_c_error) THEN
       print *,"c_error = ",c_error
       print *,"occurred when calling: xf_tree_set_fixed_header_items()"
       call xml_print_error(c_error)
       error_flag = error_xml
       return
    END IF

  end subroutine xml_set_fixed_header
    !  #]

! this routine is currently not used!!!
! TODO: to make the code consistent add a corresponding routine
! into the simple_xml module as well.

  subroutine xml_get_fixed_header(xml_tree, path_FH, &
       file_name, file_description, &
       notes, mission, file_class, file_type, validity_start, validity_stop,&
       file_version, system, creator, creator_version, creation_date, &
       error_flag)
    !  #[
    TYPE(xml_node_type),        pointer :: xml_tree        ! input
    !integer,                    intent(in)  :: xml_tree
    character(len=*),           intent(in)  :: path_FH     ! input
    character(len=max_str_len), intent(out) :: file_name
    character(len=max_str_len), intent(out) :: file_description
    character(len=max_str_len), intent(out) :: notes
    character(len=max_str_len), intent(out) :: mission
    character(len=max_str_len), intent(out) :: file_class
    character(len=max_str_len), intent(out) :: file_type 
    character(len=max_str_len), intent(out) :: validity_start
    character(len=max_str_len), intent(out) :: validity_stop
    integer,                    intent(out) :: file_version
    character(len=max_str_len), intent(out) :: system
    character(len=max_str_len), intent(out) :: creator
    character(len=max_str_len), intent(out) :: creator_version
    character(len=max_str_len), intent(out) :: creation_date
    integer,                    intent(out) :: error_flag

    ! local variables
    integer(i4_) :: c_error
    character(len=max_str_len) :: &
         c_file_name, c_file_description, c_notes, c_mission, &
         c_file_class, c_file_type, c_validity_start, c_validity_stop,&
         c_system, c_creator, c_creator_version, c_creation_date

    ! init
    error_flag = no_error

    file_name(:)        = ' '
    file_description(:) = ' '
    notes(:)            = ' '
    mission(:)          = ' '
    file_class(:)       = ' '
    file_type(:)        = ' '
    validity_start(:)   = ' '
    validity_stop(:)    = ' '
    system(:)           = ' '
    creator(:)          = ' '
    creator_version(:)  = ' '
    creation_date(:)    = ' '

    IF (.not. associated(xml_tree) ) THEN
       print *,"problem: tree seems not associated ..."
       error_flag=error_programming
       return
    END IF

    c_error = xml_no_c_error ! init to be sure
    call xf_tree_get_fixed_header_items(xml_tree%x, &
         c_file_name, c_file_description, c_notes, c_mission, &
         c_file_class, c_file_type, c_validity_start, c_validity_stop,&
         file_version, c_system, c_creator, c_creator_version, &
         c_creation_date,  c_error)
    IF (c_error .lt. xml_no_c_error) THEN
       print *,"c_error = ",c_error
       print *,"occurred when calling: xf_tree_get_fixed_header_items()"
       call xml_print_error(c_error)
       error_flag = error_xml
       return
    END IF

    file_name = string_c_to_f90(c_file_name)
    file_description = string_c_to_f90(c_file_description)
    notes            = string_c_to_f90(c_notes)
    mission          = string_c_to_f90(c_mission)
    file_class       = string_c_to_f90(c_file_class)
    file_type        = string_c_to_f90(c_file_type)
    validity_start   = string_c_to_f90(c_validity_start)
    validity_stop    = string_c_to_f90(c_validity_stop)
    system           = string_c_to_f90(c_system)
    creator          = string_c_to_f90(c_creator)
    creator_version  = string_c_to_f90(c_creator_version)
    creation_date    = string_c_to_f90(c_creation_date)

  end subroutine xml_get_fixed_header
    !  #]

  subroutine xml_add_SPH_item(xml_tree, path_in, item_name, item, error_flag)
    !  #[
    TYPE(xml_node_type), pointer :: xml_tree
    !integer,             intent(in)  :: xml_tree   ! points to updated struct
    character(len=*),    intent(in)  :: path_in
    character(len=*),    intent(in)  :: item_name
    character(len=*),    intent(in)  :: item
    integer,             intent(out) :: error_flag

    ! local variables
    character(len=max_str_len) :: path

    error_flag = no_error

    IF (.not. associated(xml_tree) ) THEN
       print *,"problem: tree seems not associated ..."
       error_flag=error_programming
       return
    END IF

    
    path = trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name
    call xml_tree_add_child(xml_tree,path,item_name,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(path)//"/"//trim(item_name)
    call xml_tree_set_string_node_value(xml_tree,path,trim(adjustl(item)),&
                                        error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine xml_add_SPH_item
    !  #]                           
  subroutine xml_tree_create_root(xml_tree, name, error_flag)
    !  #[ define the root element
    TYPE(xml_node_type), pointer :: xml_tree
    !integer,          intent(in)  :: xml_tree
    character(len=*), intent(in)  :: name
    integer,          intent(out) :: error_flag

    ! local variables
    integer(i4_) :: c_error

    error_flag = no_error

    IF (.not. associated(xml_tree) ) THEN
       print *,"problem: tree seems not associated ..."
       error_flag=error_programming
       return
    END IF

    c_error = xml_no_c_error ! init to be sure
    call xf_tree_create_root(xml_tree%x, string_f90_to_c(name), c_error)
    IF (c_error .lt. xml_no_c_error) THEN
       print *,"c_error = ",c_error
       print *,"occurred when calling: xf_tree_create_root()"
       print *,"named ",trim(name)
       call xml_print_error(c_error)
       error_flag = error_xml
       return
    END IF

  end subroutine xml_tree_create_root
    !  #]
  subroutine xml_tree_add_child(xml_tree,path,name,error_flag)
    !  #[ add a child element below the given path
    TYPE(xml_node_type), pointer :: xml_tree
    !integer,          intent(in)  :: xml_tree
    character(len=*), intent(in)  :: path
    character(len=*), intent(in)  :: name
    integer,          intent(out) :: error_flag

    ! local variables
    integer(i4_) :: c_error

    error_flag = no_error

    IF (.not. associated(xml_tree) ) THEN
       print *,"problem: tree seems not associated ..."
       error_flag=error_programming
       return
    END IF

    c_error = xml_no_c_error ! init to be sure
    call xf_tree_add_child(xml_tree%x,string_f90_to_c(path),&
                           string_f90_to_c(name),c_error)
    IF (c_error .lt. xml_no_c_error) THEN
       print *,"c_error = ",c_error
       print *,"occurred when calling: xf_tree_add_child()"
       print *,"with path: ",trim(path)
       print *,"and name:  ",trim(name)
       call xml_print_error(c_error)
       error_flag = error_xml
       return
    END IF

  end subroutine xml_tree_add_child
    !  #]
  subroutine xml_tree_add_attribute(xml_tree,path,attribute,error_flag)
    !  #[ add an empty attribute to the given path
    TYPE(xml_node_type), pointer :: xml_tree
    !integer,          intent(in)  :: xml_tree
    character(len=*), intent(in)  :: path
    character(len=*), intent(in)  :: attribute
    integer,          intent(out) :: error_flag

    ! local variables
    integer(i4_) :: c_error

    error_flag = no_error

    IF (.not. associated(xml_tree) ) THEN
       print *,"problem: tree seems not associated ..."
       error_flag=error_programming
       return
    END IF

    c_error = xml_no_c_error ! init to be sure
    call xf_tree_add_attribute(xml_tree%x,string_f90_to_c(path),&
                               string_f90_to_c(attribute),c_error)
    IF (c_error .lt. xml_no_c_error) THEN
       print *,"c_error = ",c_error
       print *,"occurred when calling: xf_tree_add_attribute()"
       print *,"with path:     ",trim(path)
       print *,"and attribute: ",trim(attribute)
       call xml_print_error(c_error)
       error_flag = error_xml
       return
    END IF
    
  end subroutine xml_tree_add_attribute
    !  #]
  subroutine xml_tree_add_and_set_attribute(xml_tree,path,attr_name,&
                                            attr_value,error_flag)
    !  #[ add an attribute to the given path and fill it
    TYPE(xml_node_type), pointer :: xml_tree
    !integer,          intent(in)  :: xml_tree
    character(len=*), intent(in)  :: path
    character(len=*), intent(in)  :: attr_name
    character(len=*), intent(in)  :: attr_value
    integer,          intent(out) :: error_flag

    ! local variables
    integer(i4_) :: c_error
    character(len=10) :: strformat

    error_flag = no_error

    IF (.not. associated(xml_tree) ) THEN
       print *,"problem: tree seems not associated ..."
       error_flag=error_programming
       return
    END IF

    ! add a new empty attribute to the given path
    c_error = xml_no_c_error ! init to be sure
    call xf_tree_add_attribute(xml_tree%x,string_f90_to_c(path),&
                               string_f90_to_c(attr_name),c_error)
    IF (c_error .lt. xml_no_c_error) THEN
       print *,"c_error = ",c_error
       print *,"occurred when calling: xf_tree_add_attribute()"
       print *,"with path:     ",trim(path)
       print *,"and attribute: ",trim(attr_name)
       call xml_print_error(c_error)
       error_flag = error_xml
       return
    END IF
    
    c_error = xml_no_c_error ! init to be sure
    strformat = "%s"
    call xf_tree_set_string_node_value( &
         xml_tree%x, string_f90_to_c(trim(path)//"/@"//trim(attr_name)), &
         string_f90_to_c(attr_value), string_f90_to_c(strformat), c_error)
    IF (c_error .lt. xml_no_c_error) THEN
       print *,"c_error = ",c_error
       print *,"occurred when calling: xf_tree_set_string_node_value()"
       print *,"with path:           ",trim(path)
       print *,"and attribute name:  ",trim(attr_name)
       print *,"and attribute value: ",trim(attr_value)
       call xml_print_error(c_error)
       error_flag = error_xml
       return
    END IF
    
  end subroutine xml_tree_add_and_set_attribute
    !  #]
  subroutine xml_tree_read_string_attribute(xml_tree,path,attr_name,&
                                            attr_value,error_flag)
    !  #[ get an attribute value from a given path
    TYPE(xml_node_type), pointer :: xml_tree
    !integer,          intent(in)  :: xml_tree
    character(len=*), intent(in)  :: path
    character(len=*), intent(in)  :: attr_name
    character(len=*), intent(out) :: attr_value
    integer,          intent(out) :: error_flag

    ! local variables
    integer(i4_) :: c_error
    character(len=max_str_len) :: c_string_attr_value

    error_flag = no_error

    IF (.not. associated(xml_tree) ) THEN
       print *,"problem: tree seems not associated ..."
       error_flag=error_programming
       return
    END IF

    !print *,"testjos inside xml_tree_read_string_attribute:"
    !print *,"testjos inside: path=",trim(path)
    !print *,"testjos inside: attr_name=",trim(attr_name)
    !print *,"testjos inside: trim(path)+/@+trim(attr_name) =",&
    !        trim(path)//"/@"//trim(attr_name)

    c_error = xml_no_c_error ! init to be sure
    ! this code seems not to work with ee_cfi library 3.4
    ! this is a BUG in this library ! (still a problem for 3.7.1)
    !call xf_tree_read_string_attribute(xml_tree,string_f90_to_c(path),&
    !              string_f90_to_c(attr_name),c_string_attr_value,c_error)
    ! replace this by the following call:
    call xf_tp_read_string_node_value( &
         xml_tree%x, string_f90_to_c(trim(path)//"/@"//trim(attr_name)), &
         c_string_attr_value, c_error)
    IF (c_error .lt. xml_no_c_error) THEN
       print *,"c_error = ",c_error
       print *,"occurred when calling: xf_tp_read_string_node_value()"
       !print *,"occurred when calling: xf_tree_read_string_attribute()"
       print *,"with path:          ",trim(path)
       print *,"and attribute name: ",trim(attr_name)
       call xml_print_error(c_error)
       error_flag = error_xml
       return
    END IF

    attr_value = string_c_to_f90(c_string_attr_value)
    
  end subroutine xml_tree_read_string_attribute
    !  #]
  !------------------------------------------
  ! subroutines for namespace and schema handling
  !------------------------------------------
  subroutine xml_tree_set_namespace(xml_tree, xml_namespace_uri, error_flag,&
                                    xml_scope_name)
    !  #[ ! set the xmlns attribute in the root element
    type(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: xml_namespace_uri
    integer,             intent(out) :: error_flag
    character(len=*),    intent(in), optional :: xml_scope_name

    ! URI = uniform resource indicator

    ! the scope may be missing, in which case the global
    ! namespace xmlns="" will be set,
    ! if present it should be for example a string like "xsi"
    ! and the namespace xmlns:xsi="" will be set

    ! local variables
    character(len=max_str_len) :: xml_root_name
    character(len=max_str_len) :: attr_name
    character(len=max_str_len) :: c_string_value
    integer(i4_) :: c_error

    ! init
    error_flag = no_error

    ! rewind the xml pointer to the root tag
    call xf_tree_rewind(xml_tree%x,c_error)
    IF (c_error .lt. xml_no_c_error) THEN
       print *,"ERROR in xml_tree_set_namespace: "
       print *,"c_error = ",c_error
       print *,"occurred when calling: xf_tree_rewind()"
       call xml_print_error(c_error)
       error_flag = error_xml
       return
    END IF

    ! get the name of the root-tag 
    call xf_tree_get_curr_element_name(xml_tree%x,c_string_value,c_error)
    IF (c_error .lt. xml_no_c_error) THEN
       print *,"ERROR in xml_tree_set_namespace: "
       print *,"c_error = ",c_error
       print *,"occurred when calling: xf_tree_get_current_element_name()"
       call xml_print_error(c_error)
       error_flag = error_xml
       return
    END IF

    xml_root_name = "/"//trim(string_c_to_f90(c_string_value))

    if (present(xml_scope_name))then
       if (len_trim(xml_scope_name) .eq. 0) then
          print *,"ERROR in xml_tree_set_namespace: "
          print *,"xml_scope_name is present but seems empty!"
          print *,"this is not allowed."
          error_flag = error_xml
          return
       end if
       ! REMARK: note that this does work for ee_cfi v3.4A
       ! even though ESA has imposed the rule which disallows
       ! the colon ':' in tag names. Clearly the current library does 
       ! not test for this.
       attr_name = "xmlns:"//trim(xml_scope_name)
    else
       attr_name = "xmlns"
    end if

    if (len_trim(xml_namespace_uri) .eq. 0) then
       print *,"ERROR in xml_tree_set_namespace: "
       print *,"xml_namespace_uri seems empty!"
       print *,"this is not allowed."
       error_flag = error_xml
       return
    end if

    call xml_tree_add_and_set_attribute(xml_tree,xml_root_name,&
         attr_name,xml_namespace_uri,error_flag)
    IF (error_flag .ne. no_error) return

    xml_tree%xml_namespace_uri = xml_namespace_uri

  end subroutine xml_tree_set_namespace
    !  #]
  subroutine xml_tree_get_namespace(xml_tree,xml_namespace_uri,error_flag,&
                                    xml_scope_name)
    !  #[ ! return the xmlns attribute from the root element
    type(xml_node_type), pointer     :: xml_tree          ! input
    character(len=*),    intent(out) :: xml_namespace_uri ! result
    integer,             intent(out) :: error_flag        ! output
    character(len=*),    intent(in), optional :: xml_scope_name ! input

    ! URI = uniform resource indicator
    ! this function returns the URI for the requested scope
    ! (for example "xsi") or if not given the global namespace

    ! local variables
!    character(len=max_str_len) :: xml_root_name
!    character(len=max_str_len) :: attr_name
!    character(len=max_str_len) :: c_string_value
!    integer(i4_)               :: c_error

! the code seems not to work for now, with ee_cfi 3.4 or 3.7.1
! so use this workaround:
    error_flag = no_error

    if (len_trim(xml_tree%xml_namespace_uri) .gt. 0) then
       xml_namespace_uri = xml_tree%xml_namespace_uri
       return
    else
       ! no xmlns tag seems present, so return an error
       error_flag = error_reading_file
       return
    end if

!    ! init
!    error_flag = no_error
!    xml_namespace_uri(:) = ' '
!
!   ! rewind the xml pointer to the root tag
!    call xf_tree_rewind(xml_tree%x,c_error)
!    IF (c_error .lt. xml_no_c_error) THEN
!       print *,"ERROR in xml_tree_get_namespace: "
!       print *,"c_error = ",c_error
!       print *,"occurred when calling: xf_tree_rewind()"
!       call xml_print_error(c_error)
!       error_flag = error_xml
!       return
!    END IF
!
!    ! get the name of the root-tag
!    call xf_tree_get_curr_element_name(xml_tree%x,c_string_value,c_error)
!    IF (c_error .lt. xml_no_c_error) THEN
!       print *,"ERROR in xml_tree_get_namespace: "
!       print *,"c_error = ",c_error
!       print *,"occurred when calling: xf_tree_get_current_element_name()"
!       call xml_print_error(c_error)
!       error_flag = error_xml
!       return
!    END IF
!
!    xml_root_name = "/"//trim(string_c_to_f90(c_string_value))
!
!    if (present(xml_scope_name))then
!       if (len_trim(xml_scope_name) .eq. 0) then
!          print *,"ERROR in xml_tree_get_namespace: "
!          print *,"xml_scope_name is present but seems empty!"
!          print *,"this is not allowed."
!          error_flag = error_xml
!          return
!       end if
!       attr_name = "xmlns:"//trim(xml_scope_name)
!    else
!       attr_name = "xmlns"
!    end if
!
!print *,"testjos: xml_root_name     = ",trim(xml_root_name)
!print* ,"testjos: attr_name         = ",trim(attr_name)
!    call xml_tree_read_string_attribute(xml_tree,xml_root_name,attr_name,&
!                                        xml_namespace_uri,error_flag)
!    IF (error_flag .ne. no_error) return
!print *,"testjos: xml_namespace_uri = ",trim(xml_namespace_uri)

  end subroutine xml_tree_get_namespace
    !  #]

! NOTE that the following routines:
!      xml_tree_set_schemalocation
!      xml_tree_set_no_ns_schemaloc
! have not yet been implemented. They are available for testing in the
! simple_xml module. When needed I will implement them here as well.

  !------------------------------------------
  subroutine xml_tree_set_string_node_value(xml_tree,path,text,error_flag)
    !  #[ add a text value to a node
    TYPE(xml_node_type), pointer :: xml_tree
    !integer,          intent(in)  :: xml_tree
    character(len=*), intent(in)  :: path
    character(len=*), intent(in)  :: text
    integer,          intent(out) :: error_flag
    
    ! local variables
    character(len=10) :: strformat
    integer(i4_) :: c_error

    error_flag = no_error

    IF (.not. associated(xml_tree) ) THEN
       print *,"problem: tree seems not associated ..."
       error_flag=error_programming
       return
    END IF

    c_error = xml_no_c_error ! init to be sure
    strformat = "%s"
    call xf_tree_set_string_node_value(xml_tree%x,string_f90_to_c(path),&
         string_f90_to_c(text),string_f90_to_c(strformat),c_error)
    IF (c_error .lt. xml_no_c_error) THEN
       print *,"c_error = ",c_error
       print *,"occurred when calling: xf_tree_set_string_node_value()"
       print *,"with path: ",trim(path)
       print *,"and text:  ",trim(text)
       call xml_print_error(c_error)
       error_flag = error_xml
       return
    END IF

  end subroutine xml_tree_set_string_node_value
    !  #]
  !------------------------------------------
  subroutine xml_node_get_string(xml_tree, path, text, error_flag)
    !  #[

    TYPE(xml_node_type), pointer :: xml_tree
    !integer,          intent(in)  :: xml_tree   ! input
    character(len=*), intent(in)  :: path       ! input
    character(len=*), intent(out) :: text       ! result
    integer,          intent(out) :: error_flag ! result

    ! local variables
    integer(i4_) :: c_error
    character(len=max_str_len) :: c_string_value

    error_flag = no_error

    IF (.not. associated(xml_tree) ) THEN
       print *,"problem: tree seems not associated ..."
       error_flag=error_programming
       return
    END IF

    c_error = xml_no_c_error ! init to be sure
    text(:) = ' '

    ! name used to be:
    !  xf_tree_path_read_string_node_value_()
    ! but this exceeds the 31 char limit for fortran
    ! so I had to shorten it a bit
    call xf_tp_read_string_node_value( &
         xml_tree%x, string_f90_to_c(path), c_string_value, c_error)
    IF (c_error .lt. xml_no_c_error) THEN
       IF (c_error .eq. get_xf_cfi_empty_element() ) THEN
          ! exception: in case of an empty node, just issue a warning, no error
          IF (LocalVerbosity .ge. VerbosityWarn) THEN
             print *,"WARNING: this node seems to be empty !!!!!!!"
             print *,"Are you sure this is alright ???"
             print *,"path = ",trim(path)
          END IF
          text(:) = ' '
          return
       ELSE
          print *,"c_error = ",c_error
          print *,"occurred when calling: xf_tp_read_string_node_value()"
          print *,"with path: ",trim(path)
          call xml_print_error(c_error)
          error_flag = error_xml
          return
       END IF
    END IF

    text = string_c_to_f90(c_string_value)
    
  end subroutine xml_node_get_string

    !  #]
  subroutine xml_check_node_exists(xml_tree, path, exists, error_flag)
    !  #[
    TYPE(xml_node_type), pointer :: xml_tree    ! input
    character(len=*), intent(in)  :: path       ! input
    logical,          intent(out) :: exists     ! output
    integer,          intent(out) :: error_flag ! result

    ! local variables
    integer(i4_) :: c_error

    error_flag = no_error

    IF (.not. associated(xml_tree) ) THEN
       print *,"problem: tree seems not associated ..."
       error_flag=error_programming
       return
    END IF

    c_error = xml_no_c_error ! init to be sure

    call xf_tree_go_to_path_node( xml_tree%x, string_f90_to_c(path), c_error)
    IF (c_error .lt. xml_no_c_error) THEN
       IF (c_error .eq.  get_xf_cfi_bad_input_argument() ) THEN
          ! bad_input_argument is the expected error in case a tag is missing
          exists = .false.
          return
       ELSE
          ! for all other c_errors print the message
          exists = .false.
          print *,"c_error = ",c_error
          print *,"occurred when calling: xf_tp_read_string_node_value()"
          print *,"with path: ",trim(path)
          call xml_print_error(c_error)
          error_flag = error_xml
          return
       END IF
    END IF

    exists = .true.
    return
    
  end subroutine xml_check_node_exists
    !  #]
  !------------------------------------------
  ! write to screen or file, read from file
  !------------------------------------------
  subroutine xml_tree_print(xml_tree, error_flag)
    !  #[
    TYPE(xml_node_type), pointer :: xml_tree
    !integer, intent(in)  :: xml_tree
    integer, intent(out) :: error_flag

     ! local variables/parameters
    integer                    :: fileunit, ios, pid_nr
    character(len=max_str_len) :: text
    character(len=256)         :: tmp_filename 
    character(len=10)          :: pid_str

    ! WARNING: the xml_tree_write routine prints text using c-code.
    !          This will be bufferred separate from the fortran
    !          write/print output, and may be flushed to the
    !          screen or file in a different moment !!!!!
    !          The result may be seriously mixed up, and
    !          unpredictable which is a problem for the Test system that was
    !          implemented for the ADM-Aeolus L2B processor.
    ! Therefore I don't use STDOUT "-" anymore, but first write to a 
    ! temporary file, and then copy this file to the screen ....
    ! 
    ! This also has the advantage that the awkward ee_cfi way of specifying
    ! the schema attributes also shows up in the print results ....
    !------------------------------------------

    error_flag = no_error
    
    IF (.not. associated(xml_tree) ) THEN
       print *,"ERROR in xml_tree_print: tree seems not associated ..."
       error_flag=error_programming
       return
    END IF

    ! the filename "-" means standard output !
    !call xml_tree_write(xml_tree, "-", error_flag)

    ! get the process ID number to construct a unique filename
    ! for the temporary file (this prevent clashes if multiple instances
    ! of the software are running in parallel)
    pid_nr = get_process_id()
    write(pid_str,"(i10.10)",iostat=ios) pid_nr
    IF (ios .ne. 0) THEN
       print *,"ERROR in xml_tree_print(): "
       print *,"could not convert pid nr ",pid_nr," to a proper string ..."
       error_flag = error_c_interface
       return
    END IF
   
    tmp_filename = "/tmp/tempfile_xml_tree_print.xml."//pid_str

    call xml_tree_write(xml_tree,tmp_filename, error_flag)

    fileunit = get_lun()
    IF (fileunit .lt. 0) THEN
       ! no fileunits available, so report an error 
       error_flag = error_opening_file
       return
    END IF

    open(UNIT=fileunit, FILE=tmp_filename, FORM='formatted', &
         status="old", err=999)

    ! loop until EOF reached, then jump out of the loop to label 888
    readloop: DO
       read(fileunit,"(a)",end=888,err=995) text
       write(*,"(a)") trim(text)
    END DO readloop

888 close(fileunit)
    call free_lun(fileunit)
    
    ! delete the temporary file
    call remove_file(tmp_filename,error_flag)
    IF (error_flag .ne. no_error) return

    return

    ! error handlers
995 close(fileunit)
    call free_lun(fileunit)
    print *,"ERROR could not read from temporary file: ",trim(tmp_filename)
    print *,"which should contain the c-library  printing results"
    print *,"coming from the ee_cfi library ....."
    error_flag = error_reading_file
    return

999 call free_lun(fileunit)
    print *,"ERROR could not open temporary file: ",trim(tmp_filename)
    print *,"needed to store the c-library  printing results"
    print *,"coming from the ee_cfi library ....."
    error_flag = error_opening_file
    return

  end subroutine xml_tree_print
    !  #]
  subroutine xml_tree_write(xml_tree, filename, error_flag)
    !  #[ write the completed tree to an output file
    TYPE(xml_node_type), pointer :: xml_tree
    !integer,          intent(in)  :: xml_tree
    character(len=*), intent(in)  :: filename
    integer,          intent(out) :: error_flag

    ! local variables
    integer(i4_) :: action
    integer(i4_) :: c_error

    error_flag = no_error

    IF (.not. associated(xml_tree) ) THEN
       print *,"problem: tree seems not associated ..."
       error_flag=error_programming
       return
    END IF

    c_error = xml_no_c_error ! init to be sure
    call xf_tree_write(xml_tree%x, string_f90_to_c(filename), c_error);
    IF (c_error .lt. xml_no_c_error) THEN
       print *,"c_error = ",c_error
       print *,"occurred when calling: xf_tree_write()"
       print *,"with filename: ",trim(filename)
       call xml_print_error(c_error)
       error_flag = error_xml
       return
    END IF

    if (len_trim(xml_tree%xml_schema_ref) .gt. 0) then
       ! add a schema reference to the root tag
       action = get_xf_change_schema_and_vers()
       call xf_set_schema(string_f90_to_c(filename),&
                          string_f90_to_c(xml_tree%xml_schema_ref),&
                          action, c_error)
       IF (c_error .lt. xml_no_c_error) THEN
          print *,"c_error = ",c_error
          print *,"occurred when calling: xf_set_schema()"
          print *,"with filename:      ",trim(filename)
          print *,"and xml_schema_ref: ",trim(xml_tree%xml_schema_ref)
          call xml_print_error(c_error)
          error_flag = error_xml
          return
       END IF
    end if

  end subroutine xml_tree_write
    !  #]
  subroutine xml_load_file(filename, xml_tree, error_flag)
    !  #[
    character(len=*),    intent(in)  :: filename
    TYPE(xml_node_type), pointer     :: xml_tree
    !integer,             intent(out) :: xml_tree
    integer,             intent(out) :: error_flag

    ! local variables
    integer(i4_) :: c_error, status

    ! needed for the xmlns workaround
    integer :: fileunit, pos
    logical :: attr_found
    character(len=512) :: line
    character(len=*), parameter :: search_string = " xmlns="
    integer, parameter :: len_search_string = len(search_string)
    integer :: startquote_pos, endquote_pos

    error_flag = no_error
    pos = -1 ! needed to suppres an uninitialized warning in g95

    IF (.not. module_is_initialised) THEN
       call Init_XML(error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    allocate(xml_tree, STAT=status)
    if (status .ne. 0) then
       print *,"allocation problem in function xml_load_file"
       nullify(xml_tree)
       error_flag = error_allocate
       return
    end if

    ! workaround: 
    ! Currently it seems impossible to use the ee_cfi library to read
    ! the value of the xmlns attribute in the root tag of the xml
    ! document, do use a simple fortran workaround for this

    ! remark: opening a file from within fortran seems not always 
    !         possible after the c-code inside xf_tree_init_parser()
    !         has parsed the file. Therefore we have to inspect the
    !         file first for the xmlns attribute in the root tag.
    fileunit = get_lun()
    open(unit=fileunit,file=filename,status="old",&
         action="read",form="FORMATTED",ERR=999)

    attr_found = .false.
    search_xmlns_loop: do
       read(fileunit,"(a)",end=997,err=998) line
       ! search the string " xmlns="
       ! the string after that attribute is the one we wan
       pos=1
       charloop: do
          if (line(pos:pos+len_search_string-1) .eq. search_string) then
             attr_found = .true.
             exit search_xmlns_loop
          else
             if (pos+len_search_string-1 .lt. len_trim(line)) then
                pos=pos+1
             else
                exit charloop
             end if
          end if
       end do charloop
    end do search_xmlns_loop

    ! end of file found, but no xmlns attribute found, 
    ! this is a valid condition for now.
997 continue

    if (attr_found) then
       !print *,"attribute: "//search_string//" found"
       startquote_pos = pos+len_search_string
       !print *,"test: "//trim(line(startquote_pos:))
       endquote_pos = -1
       pos=startquote_pos+1
       findendquote_loop: do
          if (line(pos:pos) .eq. '"') then
             endquote_pos = pos
             exit findendquote_loop
          else
             if (pos .lt. len_trim(line)) then
                pos=pos+1
             else
                exit findendquote_loop
             end if
          end if
       end do findendquote_loop

       if (endquote_pos .ne. -1) then
          xml_tree%xml_namespace_uri = line(startquote_pos+1:endquote_pos-1)
       else
          print *,"ERROR: end quote of xmlns attribute not found!!!"
          error_flag = error_reading_file
          return
       end if
    else
    !   print *,"attribute: "//search_string//" NOT found !!!"
       xml_tree%xml_namespace_uri(:) = ' '
    end if

    close(unit=fileunit)
    call free_lun(fileunit)

    !print *,"TESTJOS: xml_tree%xml_namespace_uri = ["//&
    !     trim(xml_tree%xml_namespace_uri)//"]"

    c_error = xml_no_c_error ! init to be sure
    xml_tree%x = xf_tree_init_parser(string_f90_to_c(filename), c_error);
    IF (c_error .lt. xml_no_c_error) THEN
       print *,"c_error = ",c_error
       print *,"occurred when calling: xf_tree_init_parser()"
       print *,"with filename: ",trim(filename)
       call xml_print_error(c_error)
       error_flag = error_reading_file
       return
    END IF

    return

998  print *,"ERROR: unable to read from input file: ",trim(filename)
    error_flag = error_opening_file
    return

999  print *,"ERROR: unable to open input file: ",trim(filename)
    error_flag = error_opening_file
    return

  end subroutine xml_load_file
    !  #]
  !----------------------------------------
  ! routines to do some checking 
  ! (This is a copy from the version inside SimpleXML module !)
  function check_length_xml_item(text_in,path,maxlen_text,subr_name) &
       result(text_out)
    !  #[ 
    ! impose a maximum length on an xml node string content
    ! and truncate if needed
    character(len=*),           intent(in)  :: text_in
    character(len=*),           intent(in)  :: path
    integer,                    intent(in)  :: maxlen_text
    character(len=*),           intent(in)  :: subr_name
    character(len=maxlen_text)              :: text_out    ! result
    
    IF (len_trim(text_in) .gt. maxlen_text) THEN
       print *,"WARNING in ",trim(subr_name),":"
       print *,"available room for this element: ",maxlen_text
       print *,"but the xml node named: ",trim(path)
       print *,"seems to contain ",len_trim(text_in)," characters !!!"
       print *,"Truncating it...."
       text_out = text_in(1:maxlen_text)
    ELSE
       text_out = trim(text_in)
    END IF

  end function check_length_xml_item
    !  #]
  ! (This is a copy from the version inside SimpleXML module !)
  subroutine check_string_length(text,element_name,maxlen_text,subr_name) 
    !  #[ 
    character(len=*),           intent(in)  :: text
    character(len=*),           intent(in)  :: element_name
    integer,                    intent(in)  :: maxlen_text
    character(len=*),           intent(in)  :: subr_name
    
    IF (len_trim(text) .gt. maxlen_text) THEN
       print *,"WARNING in ",trim(subr_name),":"
       print *,"available room for this element: ",maxlen_text
       print *,"but the element named: ",trim(element_name)
       print *,"seems to contain ",len_trim(text)," characters !!!"
    END IF

  end subroutine check_string_length
    !  #]
  !------------------------------------------
  ! subroutines private to this module
  subroutine xml_print_error(c_error)
    !  #[
    ! the error is stored internelly in the cfi code
    ! so needs not to be passed as a parameter

    integer(i4_), intent(in) :: c_error

    ! local variables
    character(len=max_str_len) :: error_message

    ! NOTE: the list of errors is specified in:
    ! $(ee_cfi_software_3.7.1)/include/explorer_file_handling.h

    ! Currently (in version 3.7.1) this error list is as follows:
    !
    ! marker for first element
    ! XF_CFI_FIRST_ELEMENT   = -39 ! First element
    !
    ! errors
    ! XF_CFI_LOCK_ERROR      = -38 ! Could not lock other running threads
    ! XF_CFI_UNLOCK_ERROR    = -37 ! Could not unlock other blocked threads
    ! XF_CFI_VALUE_NOT_FOUND = -36 ! The requested value has not been found
    ! XF_CFI_FILENAME_TOO_LONG = -35 ! The filename is too long
    ! XF_CFI_WRONG_HEADER_ELEMENT = -34 ! Element not allowed to be
    !                                   ! included in a header  
    ! XF_CFI_WRONG_HEADER_FILE_EXTENSION = -33 ! File cant contain a header
    ! XF_CFI_WRONG_HEADER_FILE_FORMAT = -32 ! The header file contains 
    !                                       ! unwanted tags
    ! XF_CFI_EMPTY_ELEMENT   = -31 ! Element is empty
    ! XF_CFI_NO_ARRAY_PATH   = -30 ! The given path is not for an array
    ! XF_CFI_SAVING_DOC_ERROR = -29! Unable to save the XML document into disk
    ! XF_CFI_INVALID_FILE_FORMAT = -28 ! Unable to read an item. 
    !                                  ! Invalid file format
    ! XF_CFI_INVALID_FORMAT  = -27 ! Printing format provided is not valid
    ! XF_CFI_REMOVING_NODE_ERROR = -26 ! Unable to remove an element node
    ! XF_CFI_ADDING_SIBLING_ERROR = -25 ! Unable to add a sibling node
    ! XF_CFI_CREATING_NODE_ERROR = -24 ! Unable to create an element node
    ! XF_CFI_ROOT_ALREADY_EXISTS = -23 ! Root element already exists
    ! XF_CFI_CREATING_DOC_ERROR  = -22 ! Error when creating a new XML 
    !                                  ! document
    ! XF_CFI_WRONG_FILE_DESCRIPTOR = -21 ! File descriptor out of range
    ! XF_CFI_CURRENT_NODE_IS_NULL = -20 ! Current node is NULL. 
    !                                   ! Rewind is needed
    ! XF_CFI_MAX_FILES_REACHED = -19 ! The max. number of open files has been
    !                                ! reached. No more files are allowed
    ! XF_CFI_NODE_NAME_TOO_LONG = -18 ! Name of node exceeds the maximum 
    !                                 ! allowed
    ! XF_CFI_PATH_TOO_LONG   = -17 ! Path length exceeds the maximum allowed
    ! XF_CFI_NOT_A_TERMINAL_ELEMENT = -16 ! Current node has other xml
    !                                     ! elements as value
    ! XF_CFI_NO_ATTRIBUTES   = -15 ! Current node has no attributes
    ! XF_CFI_VALUE_OUT_OF_RANGE = -14 ! Element value (integer or real)
    !                                 ! is out of range
    ! XF_CFI_NOT_A_DOUBLE    = -13 ! Value to be converted is not a double
    ! XF_CFI_NOT_A_LONG      = -12 ! Value to be converted is not an integer
    ! XF_CFI_NO_ELEMENT_FOUND = -11 ! No element was found in the search
    ! XF_CFI_ROOT_NODE_IS_NULL = -10 ! Root node is NULL. Initialization is 
    !                                ! needed
    ! XF_CFI_NOT_AN_ELEMENT_NODE = -9 ! Current node is not an xml element
    !                                 ! node
    ! XF_CFI_NO_ELEMENT_REQUESTED = -8 ! There is no element name set to be 
    !                                  ! searched
    ! XF_CFI_DOC_NOT_PARSED_OR_CREATED = -7 ! No XML document has been parsed 
    !                                       ! or created
    ! XF_CFI_MEMORY_ERROR    = -6 ! Unable to ask for memory
    ! XF_CFI_GETTING_ROOT_ERROR = -5 ! Unable to get the root element
    !                                ! during initialisation
    ! XF_CFI_PARSING_FILE_ERROR = -4 ! Error during initialisation
    ! XF_CFI_BAD_INPUT_ARGUMENT = -3 ! Bad input argument
    ! XF_CFI_GENERIC_ERROR   = -2 ! Generic error
    ! XF_CFI_UNKNOWN_ERROR_CODE = -1 ! Error code for unknown error codes

    ! marks all ok
    ! XF_CFI_OK              = 0     ! OK

    ! warnings
    ! XF_CFI_ROOT_ALREADY_REACHED = 1 ! Warning used when the path goes
    !                                   up too much
    ! XF_CFI_TOO_MANY_ELEMENTS = 2 ! The number of elements in the file
    !                              ! exceeds the size of the storage variable
    ! XF_CFI_MAX_LENGTH_EXCEEDED = 3 ! Length of element name or value
    !                                ! exceeds the size of the storage variable
    ! XF_CFI_GENERIC_WARNING = 4 ! Generic warning
    ! XF_CFI_HEADER_ALREADY_EXISTS = 5 ! Header already exists
    !
    ! marker for last element
    ! XF_CFI_LAST_ELEMENT = 6 ! Last element

    ! get the error message for the reported error code
    ! call xf_error_msg(error_message) ! used by ee_cfi 3.4 of before
    call xf_basic_error_msg(c_error,error_message)  ! used from 3.5 upwards
    print *,"ERROR in module xml_module (in ee_cfi_wrapper): "
    print *,trim(string_c_to_f90(error_message))

    ! some cleanup (deallocations?)
    call xf_tree_cleanup_all_parser

    return

  end subroutine xml_print_error
    !  #]
  function string_c_to_f90(c_string) result(f90_string)
    !  #[
    character(len=*) :: c_string
    character(len=len(c_string)) :: f90_string

    ! local variables
    integer :: i, c_strlen

    ! find the terminating \0 character
    c_strlen = len(c_string)
    strloop: DO i=1,len(c_string)
       IF (c_string(i:i) .eq. char(0)) THEN
          c_strlen = i-1
          exit strloop
       END IF
    END DO strloop

    ! copy the string contents
    f90_string(1:c_strlen) = c_string(1:c_strlen)
    
    ! fill the remainder of the f90 string with spaces
    DO i=c_strlen+1,len(f90_string)
       f90_string(i:i) = ' '
    END DO

    !print *,"c_strlen = ",c_strlen," txt:",trim(f90_string)

  end function string_c_to_f90
    !  #]
  function string_f90_to_c(f90_string) result(c_string)
    !  #[
    character(len=*) :: f90_string
    character(len=len(f90_string)+1) :: c_string

    ! local variables
    integer :: n

    ! get the length of the f90 string
    n = len_trim(f90_string)

    IF (n .gt. max_str_len) THEN
       print *,"WARNING in string_f90_to_c"
       print *,"in module xml_module (in ee_cfi wrapper):"
       print *,"this string maybe larger than the c-code can handle !!!"
    END IF

    ! copy the name to a valid c string
    c_string(1:n)     = f90_string(1:n)

    ! add the terminating \0 character
    c_string(n+1:n+1) = char(0)

  end function string_f90_to_c
    !  #]
  !------------------------------------------
end module XML_module
