program L2C_processor
  !  #[ Documentation
  !     a dummy L2C processor that reads a L2B product file
  !     adds empty L2C datasets to it, and writes it again 
  !     as a L2C product file
  !
  !     written by: J. de Kloe (KNMI)
  !     created:    21-Jan-2008
  !
  !     Modifications:
  !
  !  #]
  !  #[ Modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE L2BC_DataStructure, only: L2BC_HDR_DataType, &
                                L2BC_BRC_DataType, &
                                Init_L2BC_HDR_DataStructure, &
                                Init_L2BC_BRC_DataStructure, &
                                Delete_L2BC_HDR_DataStructure, &
                                Delete_L2BC_BRC_DataStructure
  USE L2C_AssimPCD_ADS,    only: Init_L2C_AssimPCD_ADS
  USE L2C_MieVec_MDS,      only: Init_L2C_MieVec_MDS
  USE L2C_RayleighVec_MDS, only: Init_L2C_RayleighVec_MDS
  USE L2BC_Sph, only: Get_M_Meas, Get_M_Mie, Get_M_Rayleigh, &
                      AsciiWrite_L2BC_SPH
  USE FixedHeader, only: file_type_L2C_N, Fill_FH_items, Print_FH
  USE MainProductHeader, only: Fill_MPH_items, Print_MPH
  USE ReadL2BCdata, only: InitReadL2BCfileModule, ReadL2BCfile
  USE WriteL2BCdata, only: InitWriteL2BCfileModule, WriteL2BCfile
  USE StringTools, only: Split_path_and_filename
  USE ErrorHandler, only: no_error, program_abort, error_cmdline_options, &
                          error_allocate, error_writing_file, &
                          error_reading_file
  USE Logging, only: logmsg, &
                     log_error, log_warn, log_info, log_progress, log_debug
  USE Compiler_Features, only: iargc_aeolus,getarg_aeolus
  USE EE_CFI_Datatypes, only: IntAs, IntAuc, IntAl, IntAd, &
                              DateTimeType, Convert_XML_to_Num_DateTime, &
                              Convert_KVT_to_Num_DateTime
  USE LunManager, only: get_lun, free_lun, fileunit_stdout
  USE AeolusConstants, only: N_range_gates, software_version

  !  #]
  !  #[ variables
  implicit none
  integer             :: error_flag, nr_of_arguments
  integer             :: AllocStatus
  integer             :: Num_BRC, iBRC
  character(len=256)  :: L2B_Path_and_FileName ! without extension
  character(len=256)  :: L2B_FileName          ! without extension
  character(len=256)  :: Path
  character(len=4)    :: extension
  character(len=6)    :: str_cntr
  character(len=256)  :: Dummy_L2C_Proc_Params_File
  character(len=256)  :: tmp_filename_L1B
  character(len=256)  :: tmp_filename_MET
  character(len=256)  :: tmp_filename_RBC
  character(len=256)  :: tmp_filename_CLM
  character(len=256)  :: tmp_filename_CAL
  character(len=256)  :: tmp_filename_L2B_PAR
  integer(IntAuc)     :: value_IntAuc
  integer(IntAs)      :: N_Meas0
  integer(IntAs)      :: N_Obs_Mie0
  integer(IntAs)      :: N_Obs_Rayleigh0
  integer(IntAs)      :: N_Range_Gates0
  TYPE(L2BC_HDR_DataType) :: L2BC_Headers
  TYPE(L2BC_BRC_DataType), dimension(:), pointer :: L2BC
  type(DateTimeType) :: DateTimeStart
  type(DateTimeType) :: DateTimeStop
  logical            :: SwitchProductError
  !  #]
  !  #[ program code

  !  #[ initialisations
  error_flag = no_error
  !  #]
  !  #[ commandline handling

  ! read the filename without extension as 1st commandline parameter:
  nr_of_arguments = iargc_aeolus()
  IF (nr_of_arguments .ne. 1) THEN
     call Print_Usage_Cmdl_Options
     call program_abort(error_cmdline_options,"Dummy_L2C_processor")
  END IF

  call getarg_aeolus(1,L2B_Path_and_FileName)
  call Split_path_and_filename(L2B_Path_and_FileName,Path,L2B_FileName)
  print *,"L2B_Path_and_FileName = ",trim(L2B_Path_and_FileName)
  print *,"Path                  = ",trim(Path)
  print *,"L2B_FileName          = ",trim(L2B_FileName)

  ! extra sanity check
  extension = L2B_FileName(len_trim(L2B_FileName)-3:len_trim(L2B_FileName))
  IF ((extension .eq. ".DBL") .or. (extension .eq. ".HDR")) THEN
     print *,"ERROR in Dummy_L2C_processor: could not load file"
     print *,"filename seems to end with an extension: "//extension
     print *,"file: "//trim(L2B_FileName)
     print *,"but this tool requires the filename to be specified"
     print *,"without extension ....."
     call Print_Usage_Cmdl_Options
     call program_abort(error_cmdline_options,"Dummy_L2C_processor")     
  END IF

  !  #] end of commandline handling
  !  #[ reading L2B file and add dummy L2C datasets
  call InitReadL2BCfileModule(error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"InitReadL2BCfileModule")

  call Init_L2BC_HDR_DataStructure(L2BC_Headers, error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"Init_L2BC_HDR_DataStructure")

  ! read the headers only, to retrieve the nr of BRC's in this file
  call ReadL2BCfile(L2BC_Headers, L2B_FileName, error_flag, &
                    Which_DSR=1, ReadHeaders=.true.)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"ReadL2BCfile")

  call logmsg(log_progress,"Successfully read L2B file headers")

  ! Set the number of BRCs
  Num_BRC = L2BC_Headers%SPH%Geolocation_DSD%nr_of_dataset_records
  print *,"Num_BRC = ",Num_BRC

  allocate(L2BC(Num_BRC),STAT=AllocStatus)
  IF (AllocStatus .ne. 0) THEN
     call logmsg(log_error,"ERROR in Dummy_L2C_processor: "//&
          "allocation problem while allocating L2BC(:) datastructure")
     call program_abort(error_allocate,"Dummy_L2C_processor")
  END IF

  ! read all DSR's 
  brc_loop1: DO iBRC=1,Num_BRC
     !  #[ loop
     write(str_cntr,"(i6)",err=999) iBRC

     ! initialise the datastructure
     call Init_L2BC_BRC_DataStructure(L2BC(iBRC), 0, 0_IntAuc, 0_IntAuc, 0, &
                                      error_flag)
     IF (error_flag .ne. no_error) &
          call program_abort(error_flag,"Init_L2BC_BRC_DataStructure")  

     ! read all datablocks
     call ReadL2BCfile(L2BC_Headers, L2B_FileName, error_flag,     &
                       L2BC            = L2BC(iBRC),   &
                       Which_DSR       = iBRC,         &
                       ReadGeoloc      = .true.,       &
                       ReadPCD         = .true.,       &
                       ReadMie         = .true.,       &
                       ReadRayleigh    = .true.)
     IF (error_flag .ne. no_error) &
          call program_abort(error_flag,"ReadL2BCfile")

     call logmsg(log_progress,"Successfully read L2B DSR "//&
                 trim(adjustl(str_cntr)))

     allocate(L2BC(iBRC)%Assim_PCD_ADS,       &
              L2BC(iBRC)%Mie_VecWind_MDS,     &
              L2BC(iBRC)%Rayleigh_VecWind_MDS,&
              stat=AllocStatus)
     IF (AllocStatus .ne. 0) THEN
        call logmsg(log_error,"ERROR in Dummy_L2C_processor: "//&
             "allocation problem while allocating L2C datasets")
        call program_abort(error_allocate,"Dummy_L2C_processor")
     END IF

     call Get_M_Meas(    L2BC_Headers%SPH, value_IntAuc,         error_flag)
     IF (error_flag .ne. no_error) &
          call program_abort(error_flag,"Get_M_Meas")
     N_Meas0 = int(value_IntAuc,IntAs)

     call Get_M_Mie(     L2BC_Headers%SPH, value_IntAuc,      error_flag)
     IF (error_flag .ne. no_error) &
          call program_abort(error_flag,"Get_M_Mie")
     N_Obs_Mie0 = int(value_IntAuc,IntAs)

     call Get_M_Rayleigh(L2BC_Headers%SPH, value_IntAuc, error_flag)
     IF (error_flag .ne. no_error) &
          call program_abort(error_flag,"Get_M_Rayleigh")
     N_Obs_Rayleigh0 = int(value_IntAuc,IntAs)

     N_Range_Gates0 = N_range_gates ! from aeolusconstants.F90

     !print *,"test: N_Meas0         = ",N_Meas0
     !print *,"test: N_Obs_Mie0      = ",N_Obs_Mie0
     !print *,"test: N_Obs_Rayleigh0 = ",N_Obs_Rayleigh0
     !print *,"test: N_Range_Gates0  = ",N_Range_Gates0

     call Init_L2C_AssimPCD_ADS(L2BC(iBRC)%Assim_PCD_ADS, &
               N_Meas0, N_Obs_Mie0, N_Obs_Rayleigh0, N_Range_Gates0, &
               error_flag)
     IF (error_flag .ne. no_error) &
          call program_abort(error_flag,"Init_L2C_AssimPCD_ADS")

     call Init_L2C_MieVec_MDS(L2BC(iBRC)%Mie_VecWind_MDS, &
               int(N_Meas0,IntAs), int(N_Obs_Mie0,IntAs), &
               int(N_Range_Gates0,IntAs), error_flag)
     IF (error_flag .ne. no_error) &
          call program_abort(error_flag,"Init_L2C_MieVec_MDS")

     call Init_L2C_RayleighVec_MDS(L2BC(iBRC)%Rayleigh_VecWind_MDS, &
               int(N_Meas0,IntAs), int(N_Obs_Rayleigh0,IntAs), &
               int(N_Range_Gates0,IntAs), error_flag)
     IF (error_flag .ne. no_error) &
          call program_abort(error_flag,"Init_L2C_RayleighVec_MDS")

     call logmsg(log_progress,"Successfully added dummy L2C DSR "//&
                 trim(adjustl(str_cntr)))
     L2BC(iBRC)%nr_of_measurements = N_Meas0
     !  #]
  END DO brc_loop1

  call logmsg(log_progress,"Successfully read L2B file: "//trim(L2B_FileName))
  !  #]
  !  #[ modify header fields to change them from L2B to L2C
  call logmsg(log_progress,"Filling L2C specific header fields")

  DateTimeStart = Convert_KVT_to_Num_DateTime(&
                     L2BC_Headers%MPH%Sensing_Start)
  DateTimeStop  = Convert_KVT_to_Num_DateTime(&
                     L2BC_Headers%MPH%Sensing_Stop)

! don't use this, it gives rounding differences for the Fixed Header fields!
!  DateTimeStart = Convert_XML_to_Num_DateTime(&
!                  L2BC_Headers%FH%Validity_Period%Validity_Start//".000000")
!  DateTimeStop  = Convert_XML_to_Num_DateTime(&
!                  L2BC_Headers%FH%Validity_Period%Validity_Stop//".000000")
  call  Fill_FH_items(L2BC_Headers%FH, &
       file_description    = "Level 2C Product", &
       notes               = "L2C_processor, version: "//&
       software_version, &
       file_class          = L2BC_Headers%FH%file_class, &
       file_type           = file_type_L2C_N, &
       ValidityPeriodStart = DateTimeStart, &
       ValidityPeriodStop  = DateTimeStop, &
       file_version        = 1, & ! this should always be 1 for us
       system              = L2BC_Headers%FH%Source%System, &
       creator             = "AE_L2B_L2C_WIND/IPF2C_L2B_L2C", &
       creator_version     = software_version, &
       error_flag          = error_flag)
  IF (error_flag .ne. no_error)  &
       call program_abort(error_flag,"Fill_FH_items")

  select case(trim(L2BC_Headers%MPH%Product_Err))
  case("1")
     SwitchProductError = .true.
  case("0")
     SwitchProductError = .false.
  case default
      call logmsg(log_error,"ERROR in Dummy_L2C_processor: "//&
                  "Incorrect value for MPH KVT field: Product_Err = "//&
                  L2BC_Headers%MPH%Product_Err//&
                  " possible values are '0' or '1'")
      call program_abort(error_reading_file,"Fill_FH_items")
  end select

  call Fill_MPH_items(L2BC_Headers%MPH, &
       product           = L2BC_Headers%MPH%Product,&
       proc_stage        = L2BC_Headers%MPH%Proc_Stage, &
       ref_doc           = L2BC_Headers%MPH%Ref_Doc, &
       proc_center       = L2BC_Headers%MPH%Proc_Center, &
       software_version  = software_version, &
       product_err       = SwitchProductError, &
       tot_size          = 0_IntAd,& ! filled by WriteL2BCfile below
       sph_size          = 0_IntAl,& ! filled by WriteL2BCfile below
       num_DSD           = 0_IntAl,& ! filled by WriteL2BCfile below
       num_data_sets     = 0_IntAl,& ! filled by WriteL2BCfile below
       DateTimeStart     = DateTimeStart, &
       DateTimeStop      = DateTimeStop, &
       error_flag        = error_flag)
  IF (error_flag .ne. no_error)  &
       call program_abort(error_flag,"Fill_MPH_items")

  ! fill the SPH elements that are added for L2C files
  L2BC_Headers%SPH%Num_Invalid_Obs_L2C_Mie = &
       L2BC_Headers%SPH%Num_Valid_Obs_L2B_Mie
  L2BC_Headers%SPH%Num_Invalid_Obs_L2C_Ray = &
       L2BC_Headers%SPH%Num_Valid_Obs_L2B_Ray
  ! handling of the extra DSD's is done in WriteL2BCfile below

  call logmsg(log_progress,"Succesfully filled L2C specific header fields")
  !  #]
  !  #[ writing L2C product file
  call logmsg(log_progress,"Writing L2C product file")

  call InitWriteL2BCfileModule(error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"InitWriteL2BCfileModule")

  Dummy_L2C_Proc_Params_File = &
       "AE_TEST_AUX_PAR_2C_20050109T000000_20110109T000000_0009.EEF"

  ! NOTE: this intermediate step is needed. If you insert these L2BC fields
  ! directly in the call to WriteL2BCfile() apparantly there is a conflict
  ! caused by filling memory by using data from the same location.
  ! At least for the current g95 version I end up with empty strings
  ! if I do this (might be a g95 bug though)
  tmp_filename_L1B     = L2BC_Headers%SPH%L1B_DSD%Filename
  tmp_filename_MET     = L2BC_Headers%SPH%Met_DSD%Filename
  tmp_filename_RBC     = L2BC_Headers%SPH%RBC_DSD%Filename
  tmp_filename_CLM     = L2BC_Headers%SPH%CLM_DSD%Filename
  tmp_filename_CAL     = L2BC_Headers%SPH%CAL_DSD%Filename
  tmp_filename_L2B_PAR = L2BC_Headers%SPH%L2B_Param_DSD%Filename

  !print *,"test: L1B = ",trim(tmp_filename_L1B)
  !print *,"test: MET = ",trim(tmp_filename_MET)
  !print *,"test: RBC = ",trim(tmp_filename_RBC)
  !print *,"test: CLM = ",trim(tmp_filename_CLM)
  !print *,"test: CAL = ",trim(tmp_filename_CAL)
  !print *,"test: PAR = ",trim(tmp_filename_L2B_PAR)
  !print *,"test: L2B_FileName = ",trim(L2B_FileName)

  call WriteL2BCfile(Num_BRC, L2BC_Headers, L2BC, &
       N_Meas0, N_Obs_Mie0, N_Obs_Rayleigh0, &
       L2BCresultdir     = "./", &
       L1Bfilename       = tmp_filename_L1B, &
       NWPfilename       = tmp_filename_MET, &
       RBCfilename       = tmp_filename_RBC, &
       CLMfilename       = tmp_filename_CLM, &
       CALfilename       = tmp_filename_CAL, &
       Paramfilename     = tmp_filename_L2B_PAR, &
       error_flag        = error_flag, &
       L2Bfilename       = L2B_FileName, &
       Param2Cfilename   = Dummy_L2C_Proc_Params_File, &
       L2C_flag          = .true.)
  IF (error_flag .ne. no_error) call program_abort(error_flag,"WriteL2BCfile")

  call logmsg(log_progress,"Successfully writing L2C product file named: "//&
              trim(L2BC_Headers%FH%File_Name))
  !  #]
  !  #[ test prints for the header fields
  ! for testing only
  !call Print_FH(L2BC_Headers%FH,error_flag,"L2C%FH")
  !IF (error_flag .ne. no_error)  &
  !     call program_abort(error_flag,"Print_FH")

  ! for testing only
  !call Print_MPH(L2BC_Headers%MPH,error_flag,"L2C%MPH")
  !IF (error_flag .ne. no_error)  &
  !     call program_abort(error_flag,"Print_MPH")

  ! for testing only
  !call AsciiWrite_L2BC_SPH(L2BC_Headers%SPH,fileunit_stdout,"L2C%SPH",&
  !                         error_flag,L2C_flag=.true.)
  !IF (error_flag .ne. no_error)  &
  !     call program_abort(error_flag,"Print_SPH")
  !  #]
  !  #[ clean up
  brc_loop2: DO iBRC=1,Num_BRC  
     call Delete_L2BC_BRC_DataStructure(L2BC(iBRC),L2C_flag=.true.)
  END DO brc_loop2
  deallocate(L2BC)
  call Delete_L2BC_HDR_DataStructure(L2BC_Headers)
  !  #]

  !  #] end of the main program
  goto 111
  !  #[ exception/error handlers
!997 print *,"ERROR  in: Dummy_L2C_processor"
!    print *,"could not open outputfile: ",trim(filename_txt)
!    call program_abort(error_opening_file,"Dummy_L2C_processor")

!998 print *,"conversion error in: Dummy_L2C_processor"
!    print *,"could not convert the counter iBRC = ",iBRC," to a string .."
!    call program_abort(error_writing_file,"Dummy_L2C_processor")    

999 print *,"ERROR: could not convert iBRC into a string!"
  call program_abort(error_writing_file,"Dummy_L2C_processor")

  !  #] end of the error exception handler
111 continue 
  !--------------------
contains ! a helper subroutine
  !--------------------
  subroutine Print_Usage_Cmdl_Options
    !  #[ 
    print *,"============================================"
    print *,""
    print *,"Usage: "
    print *,"Dummy_L2C_processor <L2B_filename[without extension]>"
    print *,""
    print *,"============================================"

  end subroutine Print_Usage_Cmdl_Options
    !  #]
  !--------------------
end program L2C_processor
