program TestSimpleXML
  !  #[ Description:

  !     A testprogram for the SimpleXML module
  !     -it creates an xml file containing ee_cfi formatted
  !     orbit_change data
  !
  !     created:     21-05-2005, by: J. de Kloe
  !
  !     Modifications:
  !     12-May-2006  D. Tan     Null use of ParKind1 module
  !     06-Jun-2007  J. de Kloe added path parameter to xml_set_fixed_header
  !     20-Jul-2007  J. de Kloe added xml_schema_ref in call to 
  !                             xml_tree_create_header()
  !     19-Oct-2007  J. de Kloe added a nullify statement
  !     01-Nov-2007  J. de Kloe added tests for xml_tree_set_namespace,
  !                             xml_tree_set_schemalocation and
  !                             xml_tree_set_no_ns_schemaloc
  !                             remove the test for xml_tree_set_schema
  !     16-Jan-2008  J. de Kloe phase out integer kind i_
  !
  !  #]
  !  #[ modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  use XML_module
  USE ErrorHandler, only: no_error, program_abort, &
                          error_reading_file, error_allocate
  USE VerbosityControl, only: VerbosityReport, VerbosityError
  !  #]
  !  #[ variables

  implicit none

  ! general variables
  type(xml_node_type), pointer :: xml_tree
  character(len=max_str_len) :: path
  character(len=max_str_len) :: name
  character(len=max_str_len) :: text
  character(len=max_str_len) :: attr_name
  character(len=max_str_len) :: attr_value
  character(len=max_str_len) :: xml_namespace_uri
  character(len=max_str_len) :: xml_scope_name
  character(len=max_str_len) :: xml_schema_uri
  integer                    :: error_flag

  ! fixed header input variables
  character(len=max_str_len) :: file_name
  character(len=max_str_len) :: file_description
  character(len=max_str_len) :: notes
  character(len=max_str_len) :: mission
  character(len=max_str_len) :: file_class
  character(len=max_str_len) :: file_type
  character(len=max_str_len) :: validity_start
  character(len=max_str_len) :: validity_stop
  integer                    :: file_version
  character(len=max_str_len) :: system
  character(len=max_str_len) :: creator
  character(len=max_str_len) :: creator_version
  character(len=max_str_len) :: creation_date

  ! orbit change inputs
  character(len=5) :: abs_orbit
  character(len=5) :: rel_orbit
  character(len=5) :: cycl_number
  character(len=5) :: phase_number

  character(len=5)  :: rep_cycle
  character(len=5)  :: cycl_length
  character(len=5)  :: anx_longitude
  character(len=15) :: mlst
  character(len=15) :: mlst_drift

  character(len=max_str_len) :: tai_of_anx
  character(len=max_str_len) :: utc_of_anx
  character(len=max_str_len) :: ut1_of_anx

  integer :: Verbosity

  !  #]
  !  #[ program code

  ! the default verbosity setting is Err (as silent as possible)
  Verbosity = VerbosityError
!  Verbosity = VerbosityReport
  call Init_XML(error_flag, ChoosenVerbosity=Verbosity, &
                CaseSensitiveTagNames=.false.)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"Init_XML")

  ! xml-reading test
  file_name = "simple_testdata.xml"
  nullify(xml_tree)
  call xml_load_file(file_name, xml_tree, error_flag)
  IF ( (error_flag .ne. no_error) .or. &
       (.not. associated(xml_tree)     ) ) &
       call program_abort(error_reading_file,"xml_load_file")
  print *,"the xml file was successfully read"

  IF (Verbosity .ge. VerbosityReport) THEN
     print *,"==================="
     call xml_tree_print(xml_tree, error_flag)
     IF (error_flag .ne. no_error) &
          call program_abort(error_flag,"xml_tree_print")
     print *,"==================="
  END IF

  file_name = "parsing_result.xml"
  call xml_tree_write(xml_tree, file_name, error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"xml_tree_write")

  print *,"the read xml tree was successfully written to file"

  path = "/Earth_Explorer_File/Earth_Explorer_Header/Fixed_Header/Mission"
  call xml_node_get_string(xml_tree,path,text,error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"xml_node_get_string")
  print *,"result of xml_node_get_string() test:"
  print *,"path       = ",trim(path)
  print *,"node_value = ",trim(text)

  attr_name = "unit"
  path = "/Earth_Explorer_File/Data_Block/"//&
         "List_of_Orbit_Changes/Orbit_Change[1]/Cycle/MLST_Drift"
  call xml_tree_read_string_attribute(xml_tree,&
       path,attr_name,attr_value,error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"xml_tree_read_string_attribute")
  print *,"result of xml_tree_read_string_attribute() test:"
  print *,"path       = ",trim(path)
  print *,"attr_name  = ",trim(attr_name)
  print *,"attr_value = ",trim(attr_value)

  call delete_xml_tree(xml_tree,error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"delete_xml_tree")

  ! level 1
  call create_ee_xml_tree(xml_tree, error_flag)
  IF ( (.not. associated(xml_tree)) .or. &
       (error_flag .ne. no_error) ) &
       call program_abort(error_allocate,"create_ee_xml_tree")

  ! set the schema reference
!  xml_schema_ref = "Dummy_Schema.xsd"
!  call xml_tree_set_schema(xml_tree,xml_schema_ref,error_flag)

  xml_namespace_uri = "http://www.w3.org/2001/XMLSchema-instance"
  call xml_tree_set_namespace(xml_tree,xml_namespace_uri, error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"xml_tree_set_namespace")

  xml_namespace_uri = "http://www.esa.int/schemas/ae/ALD_U_T_2B_01.01"
  xml_scope_name    = "xsi"
  call xml_tree_set_namespace(xml_tree,xml_namespace_uri, error_flag,&
                              xml_scope_name)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"xml_tree_set_namespace")

  xml_schema_uri="DummySchema.xsd"
  call xml_tree_set_schemalocation(xml_tree,xml_schema_uri, error_flag, &
                                   xml_scope_name)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"xml_tree_set_schemalocation")

  xml_schema_uri="AnotherDummySchema.xsd"
  call xml_tree_set_schemalocation(xml_tree,xml_schema_uri, error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"xml_tree_set_schemalocation")

  xml_schema_uri="DummyNoNamespaceSchema.xsd"
  call xml_tree_set_no_ns_schemaloc(xml_tree,xml_schema_uri, error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"xml_tree_set_schemalocation")

  ! add empty fixed and variable headers
  call xml_tree_create_header(xml_tree, ee_root_name, error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"xml_tree_create_header")

  file_name        = "CS_TEST_MPL_ORBREF_20040701T092917_99999999T999999_0001"
  file_description = "Reference Orbit Scenario File"
  notes(:)         = ' '
  mission          = "CryoSat"
  file_class       = "TEST"
  file_type        = "MPL_ORBREF"
  validity_start   = "UTC=2001-03-12T14:00:35"
  validity_stop    = "UTC=9999-99-99T99:99:99"
  file_version     = 1
  system           = "CFI Example"
  creator          = "EXPLORER_GEN_FILES: xg_gen_osf_create"
  creator_version  = "2.0"
  creation_date    = "UTC=2002-11-30T15:14:46"

  path = "/Earth_Explorer_File/Earth_Explorer_Header/Fixed_Header"

  call xml_set_fixed_header(xml_tree, path, file_name, file_description, &
       notes, mission, file_class, file_type, validity_start, validity_stop, &
       file_version, system, creator, creator_version, creation_date, &
       error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"xml_add_fixed_header")
  
  ! the data block

  ! level 2
  path = "/Earth_Explorer_File"
  name = "Data_Block"
  call xml_tree_add_child(xml_tree,path,name,error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"xml_tree_add_child")

!  path = "/Earth_Explorer_File/Data_Block"
!  path = "../Data_Block" ! werkt !
  path = "./" ! werkt
  attr_name  = "type"
  attr_value = "xml"
  call xml_tree_add_and_set_attribute(xml_tree,path,attr_name,attr_value,&
                                      error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"xml_tree_add_and_set_attribute")

  ! level 3
  path = "/Earth_Explorer_File/Data_Block"
  name = "List_of_Orbit_Changes"
  call xml_tree_add_child(xml_tree,path,name,error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"xml_tree_add_child")

  path = "/Earth_Explorer_File/Data_Block/List_of_Orbit_Changes"
  attr_name  = "count"
  attr_value = "2"
  call xml_tree_add_and_set_attribute(xml_tree,path,attr_name,attr_value,&
                                      error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"xml_tree_add_and_set_attribute")

  ! level 4+5
  path="/Earth_Explorer_File/Data_Block/List_of_Orbit_Changes"
  abs_orbit    = "1"
  rel_orbit    = "1"
  cycl_number  = "1"
  phase_number = "1" 
  rep_cycle     = "2"
  cycl_length   = "29"
  anx_longitude = "23.6"
  mlst          = "22:17:19.999999"
  mlst_drift    = "009.000000"
  tai_of_anx = "TAI=2001-03-12T14:00:34.999999"
  utc_of_anx = "UTC=2001-03-12T14:00:02.999999"
  ut1_of_anx = "UT1=2001-03-12T14:00:02.777777"

  call add_ee_orbit_change(xml_tree, path, &
       abs_orbit, rel_orbit, cycl_number, phase_number, &
       rep_cycle, cycl_length, anx_longitude, mlst, mlst_drift, &
       tai_of_anx, utc_of_anx, ut1_of_anx, error_flag) 
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"add_ee_orbit_change")

  ! level 4+5
  abs_orbit    = "2"
  rel_orbit    = "2"
  cycl_number  = "1"
  phase_number = "1"
  rep_cycle     = "2"
  cycl_length   = "30"
  anx_longitude = "21.6"
  mlst          = "23:17:17.999999"
  mlst_drift    = "010.000000"
  tai_of_anx = "TAI=2001-03-12T15:00:34.999999"
  utc_of_anx = "UTC=2001-03-12T15:00:02.999999"
  ut1_of_anx = "UT1=2001-03-12T15:00:02.777777"

  call add_ee_orbit_change(xml_tree, path, &
       abs_orbit, rel_orbit, cycl_number, phase_number, &
       rep_cycle, cycl_length, anx_longitude, mlst, mlst_drift, &
       tai_of_anx, utc_of_anx, ut1_of_anx, error_flag) 
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"add_ee_orbit_change")

  IF (Verbosity .ge. VerbosityReport) THEN
     print *,"==================="
     call xml_tree_print(xml_tree, error_flag)
     IF (error_flag .ne. no_error) &
          call program_abort(error_flag,"xml_tree_print")
     print *,"==================="
  END IF

  call xml_tree_write(xml_tree, file_name, error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"xml_tree_write")

  print *,"the generated xml tree was successfully written to file"

  call delete_xml_tree(xml_tree,error_flag)
  IF (error_flag .ne. no_error) &
       call program_abort(error_flag,"delete_xml_tree")

  call Close_XML
  !  #]
end program TestSimpleXML
