module XML_module
  !  #[ Description:

  !     A module to
  !     -read and write xml formatted data
  !     -provide a tree-data structure to store this data
  !     intended to be a simple replacement for the ee_cfi_filehandling
  !     software which is not portable enough due to missing source code.
  !
  !     created:     21-May-2005, by: J. de Kloe, KNMI
  !
  !     Modifications:
  !     20-Feb-2006 J. de Kloe  adapted to enable having a single newline
  !                             as node text
  !     21-Feb-2006 J. de Kloe  added xml_tree_read_string_attribute()
  !     03-Mar-2006 J. de Kloe  renamed set_node_attr to
  !                             xml_tree_add_and_set_attribute
  !     05-Apr-2006 J. de Kloe  added handling for xml-comments like
  !                             <!-- comment -->
  !     05-Apr-2006 J. de Kloe  removed some bugs found while parsing 
  !                             the JobOrder.xml file
  !     13-Apr-2006 J. de Kloe  removed a small bug, in case an attribute is
  !                             not found, in get_attr_value_from_node()
  !     26-Apr-2006 J. de Kloe  Default Do_Allow_xmlns_Attr set to .TRUE.
  !                             to enable JobOrder processing, and
  !                             re-order to_uppercase(trim(.)) (pgf90 bugfix)
  !     12-May-2006 D. Tan      Null use of ParKind1 module
  !     07-Sep-2006 J. de Kloe  made xmlns:/xsi: attribute exception more 
  !                             general
  !     07-Sep-2006 J. de Kloe  rephrased an error message in case a file
  !                             cannot be opened
  !     09-Mar-2007 D. Tan      Update ee_mph_name, ee_sph_name, make public
  !     08-May-2007 J. de Kloe  added routine xml_check_node_exists
  !     06-Jun-2007 J. de Kloe  added path parameter to xml_set_fixed_header
  !                             and xml_add_SPH_item
  !     20-Jul-2007 J. de Kloe  added xml_tree_set_schema() subroutine
  !     23-Jul-2007 J. de Kloe  changed xml_tree_set_schema to use
  !                             "xsi:schemaLocation" rather than 
  !                             "xsi:noNamespaceSchemaLocation" to have
  !                             output identical to what ee_cfi generates
  !     23-Jul-2007 J. de Kloe  reversed the order in which the schema
  !                             attributes "xmlns:xsi" and "xsi:schemaLocation"
  !                             are defined.
  !     24-jul-2007 J. de Kloe  disable xml_tree_set_schema routine, to let 
  !                             the code produce the same output as when
  !                             using the ee_cfi_software with disabled
  !                             schema handling (to let it be compatible 
  !                             with ee_cfi v3.4A for now)
  !     01-Nov-2007 J. de Kloe  change empty tag handling to ESA default
  !                             add namespace handling and changed schema
  !                             setting routines, changed add_attr_to_node
  !                             to allow adding attributes in normal order
  !                             (used to be only possible in reversed order)
  !     12-Nov-2007 J. de Kloe  implemented routine xml_tree_get_namespace
  !     27-Nov-2007 J. de Kloe  adapted error handling for routine 
  !                             xml_tree_get_namespace to be more similar
  !                             to the one in ee_cfi_wrapper_module
  !     16-Jan-2008 J. de Kloe phase out integer kind i_ 
  !
  !  #]
  !  #[ Known problems/limitations:
  !
  !   Use of the <![CDATA[...]]> tag is not implemented.
  !   The ThinLayer uses this tag in its product report file to contain
  !   strings that may contain all sorts of characters, including < and >.
  !   However, since we don't have to write or read these reports in the
  !   L2BP Fortran code (we use python for that) I have not yet put any 
  !   effort in implementing this. JK
  !
  !  #]
  !  #[ Modules used:
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE LunManager, only: get_lun, free_lun, fileunit_stdout
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE ErrorHandler, only: no_error, error_reading_file, error_opening_file,&
                          error_allocate, error_programming, error_xml
  USE c_support, only: get_filesize
  USE Compiler_Features, only: retchar,tabchar,newline
  !  #]
  !  #[ types, variables and parameters

  implicit none

  ! make all module contents private, unless explicitely stated otherwise
  private

  integer, parameter :: max_str_len = 512
  ! should be large enough for all string variables.
  ! the only exception being comment sections in xml, which might have
  ! an arbitrary size

  ! see section 3.1.2 of the explorer_file_handling reference manual
  ! used for checking only, not for dimensioning variables
  ! note also, that each of these checks occurs only once in a 
  ! strategic place, so they should be easy to maintain/update
  integer, parameter :: max_node_val_len = 512
  integer, parameter :: max_attr_val_len = 512
  integer, parameter :: max_path_len     = 256
  integer, parameter :: max_err_msg_len  = 512 ! not yet used
  integer, parameter :: max_node_name_len = 64
  integer, parameter :: max_attr_name_len = 64

  type node_attr_type
     private
     character, dimension(:), pointer :: name
     character, dimension(:), pointer :: value
     type(node_attr_type), pointer :: next_attr
  end type node_attr_type

  type xml_node_type
     private
     character, dimension(:), pointer :: name
     character, dimension(:), pointer :: value
     type(node_attr_type),    pointer :: first_attr
     type(node_attr_type),    pointer :: last_attr
     type(xml_node_type),     pointer :: parent
     type(xml_node_type),     pointer :: first_child
     type(xml_node_type),     pointer :: next_sibling
     type(xml_node_type),     pointer :: root_node
     logical :: is_root_node
     ! the next item is only used for the xml_root
     type(xml_node_type),     pointer :: last_visited_node
  end type xml_node_type

  ! Earth Explorer specific string constants
  character(len=*), parameter :: ee_root_name = "Earth_Explorer_File"
  character(len=*), parameter :: ee_h_name    = "Earth_Explorer_Header"
  character(len=*), parameter :: ee_fh_name   = "Fixed_Header"
  character(len=*), parameter :: ee_vh_name   = "Variable_Header"
  character(len=*), parameter :: ee_mph_name  = "Main_Product_Header"
  character(len=*), parameter :: ee_sph_name  = "Specific_Product_Header"

  character(len=*), parameter :: uri_ns_header = "http://www.esa.int/schemas/ae/"

  integer, parameter :: elem_is_undefined  = -1
  integer, parameter :: elem_is_xml_def    =  1
  integer, parameter :: elem_is_node_start =  2 
  integer, parameter :: elem_is_node_end   =  3
  integer, parameter :: elem_is_empty_node =  4 
  integer, parameter :: elem_is_node_text  =  5
  integer, parameter :: elem_is_comment    =  6
  integer, parameter :: eof_reached        = 99

  character, parameter :: space = ' '

  ! a value to remember the module status
  logical, save :: module_is_initialised = .false.

  ! should tag names be handled case-sensitive or not?
  ! remember: the ee_cfi xml handling is also case-sensitive
  logical, save :: TagNamesCaseSensitive = .true.

  ! should empty tags <tag/> be used or not ?
  ! (using a tag pair with empty content like: <tag></tag>)
  ! ee_cfi file convention is not to use empty tags
  logical, save :: DoUseEmptyTags = .false.
!  logical, save :: DoUseEmptyTags = .true.

  logical, save :: suppress_empty_tag_strings = .false.

  ! the JobOrder file (and maybe others) seem to use some attribute names
  ! having a colon ':' in it. This switch allows this for 2 exceptions only.
  ! "xmlns:xsi" and "xsi:noNamespaceSchemaLocation"
  logical, save :: Do_Allow_xmlns_Attr = .true.

  ! the default verbosity setting is Err (as silent as possible)
  integer, private, save :: LocalVerbosity = VerbosityError

  !  #]
  !  #[ Module interface
  ! specify which routines are public, so are part of the interface
  ! to this module
  ! public parameters
  public :: max_str_len, ee_root_name, ee_fh_name, &
            ee_mph_name, ee_sph_name, ee_h_name, ee_vh_name, &
            uri_ns_header
  ! public types
  public :: xml_node_type
  ! public functions/subroutines
  public :: Init_XML, Close_XML, delete_xml_tree, &
            create_xml_tree, xml_tree_add_child, &
            xml_tree_set_string_node_value, &
            xml_tree_add_and_set_attribute, &
            xml_tree_print, xml_tree_write, &
            xml_load_file, create_ee_xml_tree, xml_set_fixed_header, &
            xml_tree_create_header, add_ee_orbit_change, &
            xml_node_get_string, check_length_xml_item, check_string_length, &
            xml_tree_read_string_attribute, xml_add_SPH_item, &
            xml_check_node_exists, &
            xml_tree_set_namespace, xml_tree_get_namespace,&
            xml_tree_set_schemalocation, &
            xml_tree_set_no_ns_schemaloc
  ! to be phased out: xml_tree_set_schema
  !  #]
contains
  !----------------------------------------
  ! init and close the module
  subroutine Init_XML(error_flag, ChoosenVerbosity, &
                      CaseSensitiveTagNames, UseEmptyTags, Allow_xmlns_Attr)
    !  #[

    integer, intent(out) :: error_flag
    integer, optional, intent(in) :: ChoosenVerbosity
    logical, optional, intent(in) :: CaseSensitiveTagNames
    logical, optional, intent(in) :: UseEmptyTags
    logical, optional, intent(in) :: Allow_xmlns_Attr

    ! the error flag is not (yet) used here, but is defined to keep this
    ! routine compatible with the one in the ee_cfi_wrapper module
    error_flag = no_error

    IF (present(ChoosenVerbosity)) LocalVerbosity = ChoosenVerbosity

    IF (present(CaseSensitiveTagNames)) &
         TagNamesCaseSensitive = CaseSensitiveTagNames

    ! should empty tags <tag/> be used or not ?
    ! (using a tag pair with empty content like: <tag></tag>)
    ! ee_cfi file convention is not to use empty tags
    IF (present(UseEmptyTags)) DoUseEmptyTags = UseEmptyTags

    ! allow xml-namespace attributes containing a colonm ':'
    IF (present(Allow_xmlns_Attr)) Do_Allow_xmlns_Attr = Allow_xmlns_Attr

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"SimpleXML module has been initialised"
       call print_verbosity_setting(LocalVerbosity,ModuleName="SimpleXML")
    END IF

    ! this was default, but since the ee_cfi library seems not to
    ! to this, set it to false for now.
    ! (if needed it can be made an optional input for this init routine)
    suppress_empty_tag_strings = .false.

    module_is_initialised = .true.

  end subroutine Init_XML
    !  #]
  subroutine Close_XML
    !  #[

    IF (.not. module_is_initialised) THEN
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"Module was not initialised, nothing to do"
       return
    END IF

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"closing SimpleXML module"
    module_is_initialised = .false.

  end subroutine Close_XML
    !  #]
  !----------------------------------------
  ! create and delete a tree
  subroutine delete_xml_tree(xml_tree, error_flag)
    !  #[
    type(xml_node_type), pointer :: xml_tree
    integer, intent(out) :: error_flag

    error_flag = no_error

    IF (.not. associated(xml_tree) ) THEN
       print *,"problem: tree seems not associated ..."
       error_flag=error_programming
       return
    END IF

    IF (.not. xml_tree%is_root_node) THEN
       print *,"problem: this node is not the root of the tree !!!"
       error_flag=error_programming
       return
    END IF

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"deleting xml_tree"
    call DeleteNode(xml_tree,error_flag)
    if (error_flag .ne. no_error) return
    deallocate(xml_tree)
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"xml_tree has been deleted"

  end subroutine Delete_xml_tree
    !  #]
  subroutine create_xml_tree(xml_tree, error_flag)
    !  #[ ! allocate and init the root-node
    type(xml_node_type), pointer :: xml_tree ! result
    integer, intent(out) :: error_flag

    ! local variables
    integer :: status

    error_flag = no_error

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"creating xml_tree"

    allocate(xml_tree, STAT=status)
    if (status .ne. 0) then
       print *,"allocation problem in function create_xml_tree"
       nullify(xml_tree)
       error_flag = error_allocate
       return
    end if

    call InitNode(xml_tree, error_flag)
    IF (error_flag .ne. no_error) return

    xml_tree%is_root_node = .true.
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"done creating xml_tree"

    ! we have a tree now with only one element,
    ! so let the last_visited pointer point to this root node
    xml_tree%last_visited_node => xml_tree

    ! dito for the pointer to the root node
    xml_tree%root_node => xml_tree

  end subroutine create_xml_tree
    !  #]
  !----------------------------------------
  ! add nodes, add a text value
  subroutine xml_tree_add_child(xml_tree,path,name,error_flag)
    !  #[
    type(xml_node_type), pointer  :: xml_tree
    character(len=*), intent(in)  :: path
    character(len=*), intent(in)  :: name
    integer,      intent(out) :: error_flag
   
    ! local variables
    type(xml_node_type), pointer :: xml_node

    error_flag = no_error

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"adding child:"
       print *,"path = ",trim(path)
       print *,"name = ",trim(name)
    END IF

    IF (.not. associated(xml_tree)) THEN
       error_flag = error_programming
       return
    END IF

    xml_node => get_xml_node(xml_tree, path)
    call add_child_to_node(xml_node,name,error_flag)

  end subroutine xml_tree_add_child
    !  #]
  subroutine xml_tree_set_string_node_value(xml_tree,path,text,error_flag)
    !  #[

    type(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: path
    character(len=*),    intent(in)  :: text
    integer,         intent(out) :: error_flag
   
    ! local variables
    type(xml_node_type), pointer :: xml_node
    integer :: n, j, status
    logical :: text_found

    error_flag = no_error

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"setting node text:"
       print *,"path = [",trim(path),"]"
       print *,"text = [",trim(text),"]"
    END IF

    IF (.not. associated(xml_tree)) THEN
       print *,"ERROR in xml_tree_set_string_node_value():"
       print *,"the xml_tree pointer seems not associated..."
       error_flag = error_programming
       return
    END IF

    xml_node => get_xml_node(xml_tree, path)
    IF (.not. associated(xml_node)) THEN
       print *,"ERROR in xml_tree_set_string_node_value():"
       print *,"the xml_node pointer seems not associated..."
       error_flag = error_programming
       return
    END IF

    ! for now, check the rule that a node with children can
    ! not have a text value
    if (associated(xml_node%first_child)) then
       ! at this point the "text" string may only contain
       ! spaces and newlines, used for nicer formatting of the xml file

       text_found = .false.
       testtextloop: DO j=1,len_trim(text)
          ! take anything except space, tab, cr and newline to be text
          IF ( (text(j:j) .ne. space)   .and. &
               (text(j:j) .ne. tabchar) .and. &
               (text(j:j) .ne. retchar) .and. &
               (text(j:j) .ne. newline)      ) THEN
             text_found = .true.
             exit testtextloop
          END IF
       END DO testtextloop

       IF (text_found) THEN
          print *,"ERROR in xml_tree_set_string_node_value: double contents !!"
          print *,"for now, the rule is that a node with children can"
          print *,"not have a text value..."
          error_flag = error_programming
          return
       END IF

       return
    end if
    
    ! deallocate the name in case it should be reset
    IF (associated(xml_node%value)) THEN
       deallocate(xml_node%value)
       nullify(xml_node%value)
    END IF

    n = len_trim(text)

    !print *,"testjos: n=",n," text = [",text,"]"

    ! zero length is allowed, but the node should just remain empty
    ! in this case, so nothing more to do
    IF (n .eq. 0) return

    ! check for the ee_cfi imposed maximum
    IF (n .gt. max_node_val_len) THEN
       print *,"Error in xml_tree_set_string_node_value:"
       print *,"node content string should not be larger than ",&
            max_node_val_len
       print *,"characters (this is an ee_cfi constraint!)"
       error_flag = error_xml
       return
    END IF

    ! allocate space for the text string
    allocate(xml_node%value(n), STAT=status)
    IF (status .ne. 0) THEN
       error_flag = error_allocate
       return
    END IF

    ! copy the name from the string to the char array
    xml_node%value(1:n) = string2chararray(text)

  end subroutine xml_tree_set_string_node_value

    !  #]
  !----------------------------------------
  ! add attributes to, or read from a node and
  subroutine xml_tree_add_and_set_attribute(xml_tree,path,attr_name,&
                                            attr_value,error_flag,at_end)
    !  #[ add a new attribute and fill it

    type(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: path
    character(len=*),    intent(in)  :: attr_name
    character(len=*),    intent(in)  :: attr_value
    integer,             intent(out) :: error_flag
    logical,   optional, intent(in)  :: at_end

    ! local variables
    type(xml_node_type), pointer :: xml_node

    error_flag = no_error

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"setting node attribute:"
       print *,"path = ",trim(path)
       print *,"attribute name  = [",trim(attr_name),"]"
       print *,"attribute value = [",trim(attr_value),"]"
    END IF

    IF (.not. associated(xml_tree)) THEN
       error_flag = error_programming
       return
    END IF

    xml_node => get_xml_node(xml_tree, path)
    IF (.not. associated(xml_node)) THEN
       error_flag = error_programming
       return
    END IF

    call check_xml_attr_name(attr_name, error_flag)
    IF (error_flag .ne. no_error) return

    call add_attr_to_node(xml_node,attr_name,attr_value,error_flag,at_end)

  end subroutine xml_tree_add_and_set_attribute

    !  #]
  subroutine xml_tree_read_string_attribute(xml_tree,path,attr_name,&
                                            attr_value,error_flag,silent)
    !  #[ read an attribute value
    ! Reading attributes is not yet fully implemented in this module.
    !  especially extending the path with @attr_name
    !  is not yet implemented

    type(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: path
    character(len=*),    intent(in)  :: attr_name
    character(len=*),    intent(out) :: attr_value
    integer,             intent(out) :: error_flag
    logical, optional,   intent(in)  :: silent

    ! local variables
    type(xml_node_type), pointer :: xml_node

    error_flag = no_error

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"reading node attribute string:"
       print *,"path = ",trim(path)
       print *,"attribute name  = [",trim(attr_name),"]"
!       print *,"attribute value = [",trim(attr_value),"]"
    END IF

    IF (.not. associated(xml_tree)) THEN
       error_flag = error_programming
       return
    END IF

    xml_node => get_xml_node(xml_tree, path)
    IF (.not. associated(xml_node)) THEN
       error_flag = error_programming
       return
    END IF

    call get_attr_value_from_node(xml_node,attr_name,attr_value,error_flag,&
                                  silent)

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"attribute value = [",trim(attr_value),"]"
    END IF

  end subroutine xml_tree_read_string_attribute
    !  #]
  !----------------------------------------
  ! read a given element
  subroutine xml_node_get_string(xml_tree,path,node_value,error_flag)
    !  #[

    type(xml_node_type), pointer     :: xml_tree   ! input
    character(len=*),    intent(in)  :: path       ! input
    character(len=*),    intent(out) :: node_value ! result
    integer,             intent(out) :: error_flag ! result
    
    ! local variables
    type(xml_node_type), pointer :: xml_node
    
    error_flag = no_error

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"getting xml element:"
       print *,"path = ",trim(path)
    END IF
    
    IF (.not. associated(xml_tree)) THEN
       error_flag = error_programming
       return
    END IF

    xml_node => get_xml_node(xml_tree, path)
    IF (.not. associated(xml_node)) THEN
       error_flag = error_programming
       return
    END IF
    
    node_value(:) = ' '
    IF (associated(xml_node%value)) THEN
       ! check whether the string fits in the available output string
       IF (size(xml_node%value) .gt. len(node_value)) THEN
          print *,"ERROR in xml_get_node_string:"
          print *,"==>The value of this node does not fit in the"
          print *,"==>provided output string !!!"
          print *,"==>size(xml_node%value) = ",size(xml_node%value)
          print *,"==>len(node_value)      = ",len(node_value)
          print *,"==>path = ",trim(path)
          print *,"trim(xml_node%value) = ",trim(chararray2string(xml_node%value))
          error_flag = error_reading_file
          return
       END IF
       node_value = chararray2string(xml_node%value)
    ELSE
       ! in this case the result is an empty string, which is a VALID result
    END IF

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"trim(node_value) = ",trim(node_value)

  end subroutine xml_node_get_string
    !  #]
  subroutine xml_check_node_exists(xml_tree,path,exists,error_flag)
    !  #[

    type(xml_node_type), pointer     :: xml_tree   ! input
    character(len=*),    intent(in)  :: path       ! input
    logical,             intent(out) :: exists     ! output
    integer,             intent(out) :: error_flag ! result
    
    ! local variables
    type(xml_node_type), pointer :: xml_node

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"getting xml element:"
       print *,"path = ",trim(path)
    END IF
    
    IF (.not. associated(xml_tree)) THEN
       error_flag = error_programming
       return
    END IF

    exists = .true.

    xml_node => get_xml_node(xml_tree, path, fail_silently=.true.)
    IF (.not. associated(xml_node)) exists = .false.

  end subroutine xml_check_node_exists
    !  #]
  !----------------------------------------
  ! write to screen or file, read from file
  subroutine xml_tree_print(xml_tree, error_flag)
    !  #[
    type(xml_node_type),   pointer     :: xml_tree
    integer,               intent(out) :: error_flag

    write(fileunit_stdout,"(a)") "<?xml version=""1.0""?>"
    call write_node(fileunit_stdout, xml_tree, error_flag)
    
  end subroutine xml_tree_print
    !  #]
  subroutine xml_tree_write(xml_tree, file_name, error_flag)
    !  #[

    type(xml_node_type),   pointer     :: xml_tree
    character(len=*),      intent(in)  :: file_name
    integer,               intent(out) :: error_flag

    ! local variable
    integer :: fileunit

    fileunit = get_lun()
    IF (fileunit .lt. 0) THEN
       ! no fileunits available, so report an error
       error_flag = error_opening_file
       return
    END IF

    ! open the output file
    open(UNIT=fileunit, FILE=file_name, FORM='formatted', &
         status="replace", err=999)
    write(fileunit,"(a)") "<?xml version=""1.0""?>"
    call write_node(fileunit, xml_tree, error_flag)
    close(UNIT=fileunit)
    call free_lun(fileunit)
    return

999 print *,"ERROR could not open outputfile: ",trim(file_name)
    error_flag = error_opening_file
    return

  end subroutine xml_tree_write

    !  #]
  subroutine xml_load_file(filename, xml_tree, error_flag)
    !  #[
    character(len=*),    intent(in)  :: filename   ! input
    type(xml_node_type), pointer     :: xml_tree   ! result
    integer,             intent(out) :: error_flag ! result

    ! local variables
    integer :: status, ios
    integer :: filesize
    integer :: fileunit
    integer :: bufferpos
    integer :: type
    character(len=1), dimension(:), pointer :: buffer
    character(len=max_str_len*10) :: text
    character(len=max_str_len*10) :: save_text
    character(len=max_str_len)    :: path       ! ee_cfi maximum is 256
    character(len=max_str_len)    :: node_text  ! ee_cfi maximum is 512
    character(len=max_str_len)    :: attr_name  ! ee_cfi maximum is 64
    character(len=max_str_len)    :: attr_value ! ee_cfi maximum is ??
    character(len=max_str_len)    :: remaining_text
    logical :: first_node

    ! init to null which will be returned in case of errors
    nullify(xml_tree)
    error_flag = no_error
    save_text(:) = ' '

    ! read the full xml file in a text buffer for easier parsing
    filesize = get_filesize(filename)
    IF (filesize .le. 0) THEN
       print *,"could not open the file: ",trim(filename)
       print *,"to retrieve its size (or size equals zero)"
       print *,"reported filesize = ",filesize
       error_flag = error_reading_file       
       return
    END IF
    allocate(buffer(filesize),STAT=status)
    IF (status .ne. 0) THEN
       print *,"allocating problem in read_xml_file():"
       print *,"could not get memory to store the comple file"
       print *,"if this file is very big, increase the limit settings ..."
       error_flag = error_allocate
       return
    END IF

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"opening xml file: ",trim(filename)
    fileunit = get_lun()
    IF (fileunit .lt. 0) THEN
       print *,"ERROR in read_xml_file():"
       print *,"no more free fileunits seem available ..."
       deallocate(buffer)
       error_flag = error_reading_file
       return ! return if no fileunits available 
    END IF

    open(unit=fileunit,file=filename,status="old",&
         action="read",access="direct",form="unformatted",&
         recl=filesize,ERR=999)
    read(fileunit,rec=1, iostat=ios) buffer
    IF (ios .ne. 0) goto 998 
    close(unit=fileunit)
    call free_lun(fileunit)

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"successfully read ",filesize," bytes into memory"
       !print *,"BUFFER = "
       !print "(a)",chararray2string(buffer)
    END IF

    ! parse the file
    bufferpos = 1 ! start at the beginning of the buffer
    first_node = .true.
    readloop: DO
       call read_next_element(buffer, filesize, bufferpos, text, type, &
                              error_flag)
       IF (error_flag .ne. no_error) THEN
          print *,"parsing error !! returning with partial result..."
          deallocate(buffer)
          return
       END IF

       IF (LocalVerbosity .ge. VerbosityReport) THEN
          print *,"inside readloop: type = ",type,&
               " trim(text) = ",trim(text),&
               " bufferpos = ",bufferpos
       END IF

       select case (type)
       case(elem_is_undefined)
          !  #[
          print *,"undefined type !!!"
          error_flag = error_programming
          deallocate(buffer)
          return
          !  #]
       case(elem_is_xml_def)
          !  #[
          IF (LocalVerbosity .ge. VerbosityReport) &
               print *,"xml_def:    ",trim(text)
          !  #]
       case(elem_is_node_start)
          !  #[
          IF (LocalVerbosity .ge. VerbosityReport) &
               print *,"node_start: ",trim(text)
          IF (first_node) THEN
             first_node = .false.
             ! this is by definition the root node
             call create_xml_tree(xml_tree, error_flag)             
             IF ( (.not. associated(xml_tree) ) .or. &
                  ( error_flag .ne. no_error  )      ) THEN 
                deallocate(buffer)
                return
             END IF

             ! also the root may have attributes
             ! (for example xmlns:xsi !) so split it as well
             call split_node_text(text, node_text, remaining_text)

             call set_node_name(xml_tree, name=trim(node_text), &
                                error_flag=error_flag)
             IF (error_flag .ne. no_error) return

          ELSE
             call split_node_text(text, node_text, remaining_text)

             ! this is a normal node, so add "./"
             path = "./"
             call xml_tree_add_child(xml_tree,path,trim(node_text),error_flag)
             IF (error_flag .ne. no_error) THEN
                deallocate(buffer)
                return
             END IF
          END IF

          IF (len_trim(remaining_text) .ne. 0) THEN
             IF (LocalVerbosity .ge. VerbosityReport) THEN
                print *,"some attributes seem to be present"
                print *,"remaining_text = ",trim(remaining_text)
             END IF
             attr_loop: DO
                text = remaining_text
                call split_attr_text(text, attr_name, attr_value, &
                                     remaining_text)
                path = "./"
                ! the at_end flag makes sure the attributes are
                ! stored in the same order as they are read from the
                ! file. This is especially important for the
                ! xmlns:xsi and xsi:schemaLocation attributes
                ! that may be present in the root node
                call xml_tree_add_and_set_attribute(xml_tree,path,&
                     attr_name,attr_value,error_flag,at_end=.true.)
                IF (error_flag .ne. no_error) THEN
                   deallocate(buffer)
                   return
                END IF

                IF (len_trim(remaining_text) .eq. 0) exit attr_loop
             END DO attr_loop
          END IF

          ! any text that is now stored, probably contains only 
          ! formatting spaces, so delete the save variable
          save_text(:) = ' '

          !  #]
       case(elem_is_node_end)
          !  #[
          IF (LocalVerbosity .ge. VerbosityReport) &
               print *,"node_end:   ",trim(text)

          ! check if the name of the closing tag corresponds to the opening tag
          IF ( trim(text(2:)) .ne. &
               trim(get_node_name(xml_tree%last_visited_node))) THEN
             print *,"ERROR while parsing XML file:"
             print *,"closing TAG               : ",trim(text)
             print *,"does not match opening TAG: ",&
                  trim(get_node_name(xml_tree%last_visited_node))
             deallocate(buffer)
             return
          END IF

          ! if there are no children, see if we have to add some text
          path = "./"
          call xml_tree_set_string_node_value(xml_tree,path,&
                             trim(save_text),error_flag)
          IF (error_flag .ne. no_error) THEN
             deallocate(buffer)
             return
          END IF

          ! the text is now stored, so delete the save variable
          save_text(:) = ' '

          ! step 1 level up to the parent
          call step_to_parent(xml_tree, error_flag)
          IF (error_flag .ne. no_error) THEN
             deallocate(buffer)
             return
          END IF
          !  #]
       case(elem_is_empty_node)
          !  #[
          IF (LocalVerbosity .ge. VerbosityReport) &
               print *,"empty_node: ",trim(text)
          ! an empty node still can have attributes, so split it
          call split_node_text(text, node_text, remaining_text)
          ! remind that the trailing / already is removed by the
          ! read_next_element routine

          ! this is a normal node, so add "./"
          path = "./"
          call xml_tree_add_child(xml_tree,path,&
                       node_text(1:len_trim(node_text)-1),error_flag)
          IF (error_flag .ne. no_error) THEN
             deallocate(buffer)
             return
          END IF

          IF (len_trim(remaining_text) .ne. 0) THEN
             IF (LocalVerbosity .ge. VerbosityReport) THEN
                print *,"some attributes seem to be present"
                print *,"remaining_text = ",trim(remaining_text)
             END IF
             attr_loop2: DO
                text = remaining_text
                call split_attr_text(text, attr_name, attr_value, &
                                     remaining_text)
                path = "./"
                call xml_tree_add_and_set_attribute(xml_tree,path,&
                     attr_name,attr_value,error_flag)
                IF (error_flag .ne. no_error) THEN
                   deallocate(buffer)
                   return
                END IF

                IF (len_trim(remaining_text) .eq. 0) exit attr_loop2
             END DO attr_loop2
          END IF

          ! step 1 level up to the parent
          call step_to_parent(xml_tree, error_flag)
          IF (error_flag .ne. no_error) THEN
             deallocate(buffer)
             return
          END IF
          !  #]
       case(elem_is_node_text)
          !  #[
          IF (LocalVerbosity .ge. VerbosityReport) &
               print *,"node_text:  [",trim(text),"]"
          ! save the text, and only add it to the node when we found
          ! the closing tag. In this way we can be sure whether there are
          ! some children or not. If there are children the text
          ! probably only contains newlines and indentation spaces 
          ! which will not be considered text.
          save_text = text

!          path = "./"
!          call xml_tree_set_string_node_value(xml_tree,path,trim(text),error_flag)
!
!          IF (error_flag .ne. no_error) THEN
!             deallocate(buffer)
!             return
!          END IF
          !  #]
       case(elem_is_comment)
          !  #[
          IF (LocalVerbosity .ge. VerbosityReport) &
               print *,"comment_text:  [",trim(text),"]"
          ! just ignore comments
          ! (or should we store them and write them when the xml_tree is
          !  written to file?? Remind that the ee_cfi xml file handling
          !  library has no provisions whatsoever to handle this, so assume
          !  we don't need it here for now)
          !  #]
       case(eof_reached)
          !  #[
          IF (LocalVerbosity .ge. VerbosityReport) &
               print *,"EOF reached, done parsing ..."
          exit readloop
          !  #]
       case default
          print *,"unknown type !!!"
          error_flag = error_programming
          deallocate(buffer)
          return
       end select

       IF (LocalVerbosity .ge. VerbosityReport) THEN
          IF (associated(xml_tree)) THEN
             print *,"last visited = ",&
                  trim(get_node_name(xml_tree%last_visited_node))
          ELSE
             print *,"xml_tree is not yet defined"
          END IF
       END IF
    END DO readloop

    deallocate(buffer)

    return
    
998 print *,"ERROR while reading inputfile: ",trim(filename)
    deallocate(buffer)
    return

999 print *,"ERROR could not open inputfile: ",trim(filename)
    deallocate(buffer)
    return

  end subroutine xml_load_file
    !  #]
  !----------------------------------------
  ! Earth Explorer specific code
  subroutine create_ee_xml_tree(xml_tree, error_flag)
    !  #[
    type(xml_node_type), pointer     :: xml_tree
    integer,             intent(out) :: error_flag ! result

    error_flag = no_error

    call create_xml_tree(xml_tree, error_flag)
    IF (error_flag .ne. no_error) return

    call set_node_name(xml_tree,name=ee_root_name,error_flag=error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine create_ee_xml_tree
    !  #]
  ! header routines
  subroutine xml_set_fixed_header(xml_tree, &
       path_in, file_name, file_description,&
       notes, mission, file_class, file_type, validity_start, validity_stop, &
       file_version, system, creator, creator_version, creation_date, &
       error_flag)
    !  #[
    type(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: path_in
    character(len=*),    intent(in)  :: file_name
    character(len=*),    intent(in)  :: file_description
    character(len=*),    intent(in)  :: notes
    character(len=*),    intent(in)  :: mission
    character(len=*),    intent(in)  :: file_class
    character(len=*),    intent(in)  :: file_type
    character(len=*),    intent(in)  :: validity_start
    character(len=*),    intent(in)  :: validity_stop
    integer,             intent(in)  :: file_version
    character(len=*),    intent(in)  :: system
    character(len=*),    intent(in)  :: creator
    character(len=*),    intent(in)  :: creator_version
    character(len=*),    intent(in)  :: creation_date
    integer,             intent(out) :: error_flag

    ! local variables
    character(len=max_str_len) :: name
    character(len=max_str_len) :: path
    character(len=max_str_len) :: file_version_string

    ! level 4
    path = path_in

    name = "File_Name"
    call xml_tree_add_child(xml_tree,path,name,error_flag)
    IF (error_flag .ne. no_error) return
    name = "File_Description"
    call xml_tree_add_child(xml_tree,path,name,error_flag)
    IF (error_flag .ne. no_error) return
    name = "Notes"
    call xml_tree_add_child(xml_tree,path,name,error_flag)
    IF (error_flag .ne. no_error) return
    name = "Mission"
    call xml_tree_add_child(xml_tree,path,name,error_flag)
    IF (error_flag .ne. no_error) return
    name = "File_Class"
    call xml_tree_add_child(xml_tree,path,name,error_flag)
    IF (error_flag .ne. no_error) return
    name = "File_Type"
    call xml_tree_add_child(xml_tree,path,name,error_flag)
    IF (error_flag .ne. no_error) return
    name = "Validity_Period"
    call xml_tree_add_child(xml_tree,path,name,error_flag)
    IF (error_flag .ne. no_error) return
    name = "File_Version"
    call xml_tree_add_child(xml_tree,path,name,error_flag)
    IF (error_flag .ne. no_error) return
    name = "Source"
    call xml_tree_add_child(xml_tree,path,name,error_flag)
    IF (error_flag .ne. no_error) return
    
    ! level 5
    path = trim(path_in)//"/Validity_Period"
!ee_r_h_fh_name

    name = "Validity_Start"
    call xml_tree_add_child(xml_tree,path,name,error_flag)
    IF (error_flag .ne. no_error) return
    name = "Validity_Stop"
    call xml_tree_add_child(xml_tree,path,name,error_flag)
    IF (error_flag .ne. no_error) return
    
    ! level 5
    path = trim(path_in)//"/Source"
    name = "System"
    call xml_tree_add_child(xml_tree,path,name,error_flag)
    IF (error_flag .ne. no_error) return
    name = "Creator"
    call xml_tree_add_child(xml_tree,path,name,error_flag)
    IF (error_flag .ne. no_error) return
    name = "Creator_Version"
    call xml_tree_add_child(xml_tree,path,name,error_flag)
    IF (error_flag .ne. no_error) return
    name = "Creation_Date"
    call xml_tree_add_child(xml_tree,path,name,error_flag)
    IF (error_flag .ne. no_error) return

    ! set the values of all these fixed header items
    path = trim(path_in)//"/File_Name"
    call xml_tree_set_string_node_value(xml_tree,path,file_name,error_flag)
    IF (error_flag .ne. no_error) return
    path = trim(path_in)//"/File_Description"
    call xml_tree_set_string_node_value(xml_tree,path,file_description,error_flag)
    IF (error_flag .ne. no_error) return
    path = trim(path_in)//"/Notes"
    call xml_tree_set_string_node_value(xml_tree,path,notes,error_flag)
    IF (error_flag .ne. no_error) return
    path = trim(path_in)//"/Mission"
    call xml_tree_set_string_node_value(xml_tree,path,mission,error_flag)
    IF (error_flag .ne. no_error) return
    path = trim(path_in)//"/File_Class"
    call xml_tree_set_string_node_value(xml_tree,path,file_class,error_flag)
    IF (error_flag .ne. no_error) return
    path = trim(path_in)//"/File_Type"
    call xml_tree_set_string_node_value(xml_tree,path,file_type,error_flag)
    IF (error_flag .ne. no_error) return
    path = trim(path_in)//"/Validity_Period/Validity_Start"
    call xml_tree_set_string_node_value(xml_tree,path,validity_start,error_flag)
    IF (error_flag .ne. no_error) return
    path = trim(path_in)//"/Validity_Period/Validity_Stop"
    call xml_tree_set_string_node_value(xml_tree,path,validity_stop,error_flag)
    IF (error_flag .ne. no_error) return

    write(file_version_string,"(i4.4)") file_version
    path = trim(path_in)//"/File_Version"
    call xml_tree_set_string_node_value(xml_tree,path,file_version_string,error_flag)
    IF (error_flag .ne. no_error) return
    path = trim(path_in)//"/Source/System"
    call xml_tree_set_string_node_value(xml_tree,path,system,error_flag)
    IF (error_flag .ne. no_error) return
    path = trim(path_in)//"/Source/Creator"
    call xml_tree_set_string_node_value(xml_tree,path,creator,error_flag)
    IF (error_flag .ne. no_error) return
    path = trim(path_in)//"/Source/Creator_Version"
    call xml_tree_set_string_node_value(xml_tree,path,creator_version,error_flag)
    IF (error_flag .ne. no_error) return
    path = trim(path_in)//"/Source/Creation_Date"
    call xml_tree_set_string_node_value(xml_tree,path,creation_date,error_flag)
    IF (error_flag .ne. no_error) return
    

  end subroutine xml_set_fixed_header
    !  #]

  !----------------------------------------
  ! subroutines for schema handling
  !----------------------------------------
  ! other new routines that might be usefull (not yet implemented):
  !xml_tree_get_schemalocation
  !xml_tree_get_no_ns_schemaloc
  
  subroutine xml_tree_set_namespace(xml_tree, xml_namespace_uri, error_flag,&
                                    xml_scope_name)
    !  #[ ! set the xmlns attribute in the root element
    type(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: xml_namespace_uri
    integer,             intent(out) :: error_flag
    character(len=*),    intent(in), optional :: xml_scope_name

    ! URI = uniform resource indicator

    ! the scope may be missing, in which case the global
    ! namespace xmlns="" will be set,
    ! if present it should be for example a string like "xsi"
    ! and the namespace xmlns:xsi="" will be set

    ! local variables
    character(len=max_str_len) :: xml_root_name
    character(len=max_str_len) :: attr_name

    ! init 
    error_flag = no_error

    ! get the name of the root-tag
    xml_root_name = get_node_name(xml_tree)

    if (present(xml_scope_name))then
       if (len_trim(xml_scope_name) .eq. 0) then
          print *,"ERROR in xml_tree_set_namespace: "
          print *,"xml_scope_name is present but seems empty!"
          print *,"this is not allowed."
          error_flag = error_xml
          return
       end if
       attr_name = "xmlns:"//trim(xml_scope_name)
    else
       attr_name = "xmlns"
    end if

    if (len_trim(xml_namespace_uri) .eq. 0) then
       print *,"ERROR in xml_tree_set_namespace: "
       print *,"xml_namespace_uri seems empty!"
       print *,"this is not allowed."
       error_flag = error_xml
       return
    end if

    ! Note that xmlns:xsi needs to be printed first, 
    ! before xsi:schemaLocation or xsi:noNamespaceSchemaLocation
    ! however, the xml_tree_add_and_set_attribute() code adds every 
    ! new attribute as first in the linked list of attributes
    ! so first add the one that needs to be printed last....
    call xml_tree_add_and_set_attribute(xml_tree,xml_root_name,&
         attr_name,xml_namespace_uri,error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine xml_tree_set_namespace
    !  #]

  subroutine xml_tree_get_namespace(xml_tree,xml_namespace_uri,error_flag,&
                                    xml_scope_name)
    !  #[ ! return the xmlns attribute from the root element
    type(xml_node_type), pointer     :: xml_tree          ! input
    character(len=*),    intent(out) :: xml_namespace_uri ! result
    integer,             intent(out) :: error_flag        ! output
    character(len=*),    intent(in), optional :: xml_scope_name    ! input 

    ! URI = uniform resource indicator
    ! this function returns the URI for the requested scope
    ! (for example "xsi") or if not given the global namespace
    
    ! local variables
    character(len=max_str_len) :: xml_root_name
    character(len=max_str_len) :: attr_name

    ! init
    error_flag = no_error
    xml_namespace_uri(:) = ' '

    ! get the name of the root-tag
    xml_root_name = "/"//get_node_name(xml_tree)

    if (present(xml_scope_name))then
       if (len_trim(xml_scope_name) .eq. 0) then
          print *,"ERROR in xml_tree_set_namespace: "
          print *,"xml_scope_name is present but seems empty!"
          print *,"this is not allowed."
           error_flag = error_xml
           return
       end if
       attr_name = "xmlns:"//trim(xml_scope_name)
    else
       attr_name = "xmlns"
    end if

    ! the namespace attribute may be absent (i.e. for L1B v.1.04 files)
    ! this should just return with an empty namespace string without
    ! printing any error or warning messages.
    ! Reporting this problem is done at a higher level.
    ! NOTE: this is a workaround!
    ! TODO: remove when these old files have been phased out
    call xml_tree_read_string_attribute(xml_tree,xml_root_name,attr_name,&
                                        xml_namespace_uri,error_flag,&
                                        silent=.true.)
    IF (error_flag .ne. no_error) return

    if (len_trim(xml_namespace_uri) .eq. 0) then
       ! no xmlns tag seems present, so return an error
       error_flag = error_reading_file
       return
    end if
    
  end subroutine xml_tree_get_namespace
    !  #]
  subroutine xml_tree_set_schemalocation(xml_tree, xml_schema_uri, error_flag,&
                                         xml_scope_name)
    !  #[ set the ref:schemaLocation or schemaLocation attribute
    type(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: xml_schema_uri
    integer,             intent(out) :: error_flag
    character(len=*),    intent(in), optional :: xml_scope_name

    ! local variables
    character(len=max_str_len) :: xml_root_name
    character(len=max_str_len) :: attr_name

    ! init 
    error_flag = no_error

    ! get the name of the root-tag
    xml_root_name = get_node_name(xml_tree)

    if (len_trim(xml_schema_uri) .eq. 0) then
       print *,"ERROR in xml_tree_set_schemalocation: "
       print *,"xml_schema_uri seems empty!"
       print *,"this is not allowed."
       error_flag = error_xml
       return
    end if

    if (present(xml_scope_name))then
       if (len_trim(xml_scope_name) .eq. 0) then
          print *,"ERROR in xml_tree_set_schemalocation: "
          print *,"xml_scope_name is present but seems empty!"
          print *,"this is not allowed."
          error_flag = error_xml
          return
       end if
       attr_name = trim(xml_scope_name)//":schemaLocation"
    else
       attr_name = "schemaLocation"
    end if

    ! Note that xmlns:xsi needs to be printed first, 
    ! before xsi:schemaLocation
    ! the "at_end" switch takes care of this

    call xml_tree_add_and_set_attribute(xml_tree,xml_root_name,&
         attr_name,xml_schema_uri,error_flag,at_end=.true.)
    IF (error_flag .ne. no_error) return

  end subroutine xml_tree_set_schemalocation
    !  #]
  subroutine xml_tree_set_no_ns_schemaloc(xml_tree, xml_schema_uri, error_flag)
    !  #[ set the noNamespaceSchemaLocation attribute
    type(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: xml_schema_uri
    integer,             intent(out) :: error_flag

    ! local variables
    character(len=max_str_len) :: xml_root_name
    character(len=max_str_len) :: attr_name

    ! init 
    error_flag = no_error

    ! get the name of the root-tag
    xml_root_name = get_node_name(xml_tree)

    if (len_trim(xml_schema_uri) .eq. 0) then
       print *,"ERROR in xml_tree_set_no_ns_schemaloc: "
       print *,"xml_schema_uri seems empty!"
       print *,"this is not allowed."
       error_flag = error_xml
       return
    end if

    attr_name = "noNamespaceSchemaLocation"

    ! Note that the global namespace xmlns needs to be printed first, 
    ! before noNamespaceSchemaLocation (I am not quite sure of this)
    ! the "at_end" switch takes care of this

    call xml_tree_add_and_set_attribute(xml_tree,xml_root_name,&
         attr_name,xml_schema_uri,error_flag,at_end=.true.)
    IF (error_flag .ne. no_error) return

  end subroutine xml_tree_set_no_ns_schemaloc
    !  #]
  subroutine xml_tree_create_header(xml_tree, xml_root, error_flag)
    !  #[
    type(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: xml_root
    integer,             intent(out) :: error_flag

    ! local variables
    character(len=max_str_len) :: path
    
    ! NOTE: the header may exist below the<Earth_Explorer_File> root tag 
    !       or just start with the file <Earth_Explorer_Header>
    !       depending on whether we are generating a .EEF or a .HDR file

    IF ( (trim(xml_root) .ne. ee_h_name)    .and. &
         (trim(xml_root) .ne. ee_root_name)       )THEN
       print *,"ERROR in xml_tree_create_header():"
       print *,"expected the xml_root to be: ",ee_root_name
       print *,"or:                          ",ee_h_name
       print *,"but found: ",trim(xml_root)
       print *,"These header routines are for reading/writing"
       print *,"the ee_cfi xml format only !"
       error_flag = error_programming
       return
    END IF

    ! level 1
    call set_node_name(xml_tree,name=xml_root,error_flag=error_flag)
    IF (error_flag .ne. no_error) return

    IF (trim(xml_root) .eq. ee_root_name) THEN
       ! level 2
       path = "/"//ee_root_name
       call xml_tree_add_child(xml_tree,path,ee_h_name,error_flag)
       IF (error_flag .ne. no_error) return
       path = "/"//ee_root_name//"/"//ee_h_name
    ELSE
       path = "/"//ee_h_name
    END IF
    
    ! level 2 or 3
    call xml_tree_add_child(xml_tree,path,ee_fh_name,error_flag)
    IF (error_flag .ne. no_error) return

    ! level 2 or 3
    call xml_tree_add_child(xml_tree,path,ee_vh_name,error_flag)
    IF (error_flag .ne. no_error) return
    
  end subroutine xml_tree_create_header
    !  #]
  subroutine xml_add_SPH_item(xml_tree, path_in, item_name, item, error_flag)
    !  #[
    TYPE(xml_node_type), pointer     :: xml_tree   ! points to updated struct
    character(len=*),    intent(in)  :: path_in
    character(len=*),    intent(in)  :: item_name
    character(len=*),    intent(in)  :: item
    integer,             intent(out) :: error_flag

    ! local variables
    character(len=max_str_len) :: path

    error_flag = no_error

    path = trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name
    call xml_tree_add_child(xml_tree,path,item_name,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(path)//"/"//trim(item_name)
    call xml_tree_set_string_node_value(xml_tree,path,trim(adjustl(item)),&
                                        error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine xml_add_SPH_item
    !  #]                           
  ! orbit_change routines to generate the testfile for the module)
  subroutine add_ee_orbit_change(xml_tree, path, &
       abs_orbit, rel_orbit, cycl_number, phase_number, &
       rep_cycle, cycl_length, anx_longitude, mlst, mlst_drift, &
       tai_of_anx, utc_of_anx, ut1_of_anx, error_flag) 
    !  #[

    type(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: path
    character(len=*),    intent(in)  :: abs_orbit
    character(len=*),    intent(in)  :: rel_orbit
    character(len=*),    intent(in)  :: cycl_number
    character(len=*),    intent(in)  :: phase_number
    character(len=*),    intent(in)  :: rep_cycle
    character(len=*),    intent(in)  :: cycl_length
    character(len=*),    intent(in)  :: anx_longitude
    character(len=*),    intent(in)  :: mlst
    character(len=*),    intent(in)  :: mlst_drift
    character(len=*),    intent(in)  :: tai_of_anx
    character(len=*),    intent(in)  :: utc_of_anx
    character(len=*),    intent(in)  :: ut1_of_anx
    integer,             intent(out) :: error_flag

    ! local variables
    character(len=max_str_len) :: tmp_path
    character(len=max_str_len) :: name
    character(len=6)           :: count_str
    integer                    :: n

    error_flag = no_error

    ! count how many orbit changes there are already for this path
    name = "Orbit_Change"
    n = get_nr_of_nodes(xml_tree, path, name)
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"nr of similar nodes = ",n
    IF (n .lt. 0) THEN
       error_flag = error_programming
       return
    END IF
    call xml_tree_add_child(xml_tree,path,name,error_flag)
    IF (error_flag .ne. no_error) return

    tmp_path = trim(path)//"/Orbit_Change"
    IF (n .gt. 0) THEN
       write(count_str,"(i6)") n+1
       tmp_path = trim(path)//'/Orbit_Change['//trim(adjustl(count_str))//']'
    END IF

    call add_ee_orbit(xml_tree, tmp_path, &
         abs_orbit, rel_orbit, cycl_number, phase_number, error_flag)
    IF (error_flag .ne. no_error) return
    call add_ee_cycle(xml_tree, tmp_path, &
         rep_cycle, cycl_length, anx_longitude, mlst, mlst_drift, error_flag)
    IF (error_flag .ne. no_error) return
    call add_ee_time_of_anx(xml_tree, tmp_path, &
         tai_of_anx, utc_of_anx, ut1_of_anx, error_flag)
    IF (error_flag .ne. no_error) return
    
  end subroutine add_ee_orbit_change
    !  #]
  !----------------------------------------
  ! routines to do some checking 
  function check_length_xml_item(text_in,path,maxlen_text,subr_name) &
       result(text_out)
    !  #[ 
    ! impose a maximum length on an xml node string content
    ! and truncate if needed
    ! this routine is used to check for example numerical types like
    ! IntAuc that have a fixed number of characters
    ! It is not intended to be used for checking if the
    ! length complies to the ee_cfi max_node_val_len constraint

    character(len=*),           intent(in)  :: text_in
    character(len=*),           intent(in)  :: path
    integer,                    intent(in)  :: maxlen_text
    character(len=*),           intent(in)  :: subr_name
    character(len=maxlen_text)              :: text_out    ! result
    
    IF (len_trim(text_in) .gt. maxlen_text) THEN
       print *,"WARNING in ",trim(subr_name),":"
       print *,"available room for this element: ",maxlen_text
       print *,"but the xml node named: ",trim(path)
       print *,"seems to contain ",len_trim(text_in)," characters !!!"
       print *,"Truncating it...."
       text_out = text_in(1:maxlen_text)
    ELSE
       text_out = trim(text_in)
    END IF

  end function check_length_xml_item
    !  #]
  subroutine check_string_length(text,element_name,maxlen_text,subr_name) 
    !  #[ 
    character(len=*),           intent(in)  :: text
    character(len=*),           intent(in)  :: element_name
    integer,                    intent(in)  :: maxlen_text
    character(len=*),           intent(in)  :: subr_name
    
    IF (len_trim(text) .gt. maxlen_text) THEN
       print *,"WARNING in ",trim(subr_name),":"
       print *,"available room for this element: ",maxlen_text
       print *,"but the element named: ",trim(element_name)
       print *,"seems to contain ",len_trim(text)," characters !!!"
    END IF

  end subroutine check_string_length
    !  #]
  !----------------------------------------
  !  internal routines
  subroutine add_ee_orbit(xml_tree, path, &
       abs_orbit, rel_orbit, cycl_number, phase_number, error_flag)
    !  #[
    type(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: path
    character(len=*),    intent(in)  :: abs_orbit
    character(len=*),    intent(in)  :: rel_orbit
    character(len=*),    intent(in)  :: cycl_number
    character(len=*),    intent(in)  :: phase_number
    integer,             intent(out) :: error_flag

    ! local variables
    character(len=max_str_len) :: name
    character(len=max_str_len) :: tmp_path

    error_flag = no_error

    name = "Orbit"
    call xml_tree_add_child(xml_tree,path,name,error_flag)
    IF (error_flag .ne. no_error) return

    tmp_path = trim(path)//"/Orbit"
    name = "Absolute_Orbit"
    call xml_tree_add_child(xml_tree,tmp_path,name,error_flag)
    IF (error_flag .ne. no_error) return
    name = "Relative_Orbit"
    call xml_tree_add_child(xml_tree,tmp_path,name,error_flag)
    IF (error_flag .ne. no_error) return
    name = "Cycle_Number"
    call xml_tree_add_child(xml_tree,tmp_path,name,error_flag)
    IF (error_flag .ne. no_error) return
    name = "Phase_Number"
    call xml_tree_add_child(xml_tree,tmp_path,name,error_flag)
    IF (error_flag .ne. no_error) return

    ! set the values of all these orbit items
    tmp_path = trim(path)//"/Orbit/Absolute_Orbit"
    call xml_tree_set_string_node_value(xml_tree,tmp_path,abs_orbit,error_flag)
    IF (error_flag .ne. no_error) return
    tmp_path = trim(path)//"/Orbit/Relative_Orbit"
    call xml_tree_set_string_node_value(xml_tree,tmp_path,rel_orbit,error_flag)
    IF (error_flag .ne. no_error) return
    tmp_path = trim(path)//"/Orbit/Cycle_Number"
    call xml_tree_set_string_node_value(xml_tree,tmp_path,cycl_number,error_flag)
    IF (error_flag .ne. no_error) return
    tmp_path = trim(path)//"/Orbit/Phase_Number"
    call xml_tree_set_string_node_value(xml_tree,tmp_path,phase_number,error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine add_ee_orbit
    !  #]
  subroutine add_ee_cycle(xml_tree, path, &
       rep_cycle, cycl_length, anx_longitude, mlst, mlst_drift, error_flag)
    !  #[
    type(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: path
    character(len=*),    intent(in)  :: rep_cycle
    character(len=*),    intent(in)  :: cycl_length
    character(len=*),    intent(in)  :: anx_longitude
    character(len=*),    intent(in)  :: mlst
    character(len=*),    intent(in)  :: mlst_drift
    integer,             intent(out) :: error_flag

    ! local variables
    character(len=max_str_len) :: name
    character(len=max_str_len) :: tmp_path
    character(len=max_str_len) :: attr_name
    character(len=max_str_len) :: attr_value

    error_flag = no_error

    name = "Cycle"
    call xml_tree_add_child(xml_tree,path,name,error_flag)
    IF (error_flag .ne. no_error) return

    tmp_path = trim(path)//"/Cycle"
    name = "Repeat_Cycle"
    call xml_tree_add_child(xml_tree,tmp_path,name,error_flag)
    IF (error_flag .ne. no_error) return
    name = "Cycle_Length"
    call xml_tree_add_child(xml_tree,tmp_path,name,error_flag)
    IF (error_flag .ne. no_error) return
    name = "ANX_Longitude"
    call xml_tree_add_child(xml_tree,tmp_path,name,error_flag)
    IF (error_flag .ne. no_error) return
    name = "MLST"
    call xml_tree_add_child(xml_tree,tmp_path,name,error_flag)
    IF (error_flag .ne. no_error) return
    name = "MLST_Drift"
    call xml_tree_add_child(xml_tree,tmp_path,name,error_flag)
    IF (error_flag .ne. no_error) return

    ! set the needed attributes
    tmp_path =  trim(path)//"/Cycle/Repeat_Cycle"
    attr_name  = "unit"
    attr_value = "day"
    call xml_tree_add_and_set_attribute(xml_tree,tmp_path,&
                                        attr_name,attr_value,error_flag)
    IF (error_flag .ne. no_error) return

    tmp_path =  trim(path)//"/Cycle/Cycle_Length"
    attr_name  = "unit"
    attr_value = "orbit"
    call xml_tree_add_and_set_attribute(xml_tree,tmp_path,&
                                        attr_name,attr_value,error_flag)
    IF (error_flag .ne. no_error) return

    tmp_path =  trim(path)//"/Cycle/ANX_Longitude"
    attr_name  = "unit"
    attr_value = "deg"
    call xml_tree_add_and_set_attribute(xml_tree,tmp_path,&
                                        attr_name,attr_value,error_flag)
    IF (error_flag .ne. no_error) return

    tmp_path =  trim(path)//"/Cycle/MLST_Drift"
    attr_name  = "unit"
    attr_value = "s/day"
    call xml_tree_add_and_set_attribute(xml_tree,tmp_path,&
                                        attr_name,attr_value,error_flag)
    IF (error_flag .ne. no_error) return

    ! set the values of all these orbit items
    tmp_path = trim(path)//"/Cycle/Repeat_Cycle"
    call xml_tree_set_string_node_value(xml_tree,tmp_path,rep_cycle,error_flag)
    IF (error_flag .ne. no_error) return
    tmp_path = trim(path)//"/Cycle/Cycle_Length"
    call xml_tree_set_string_node_value(xml_tree,tmp_path,cycl_length,&
                                        error_flag)
    IF (error_flag .ne. no_error) return
    tmp_path = trim(path)//"/Cycle/ANX_Longitude"
    call xml_tree_set_string_node_value(xml_tree,tmp_path,anx_longitude,&
                                        error_flag)
    IF (error_flag .ne. no_error) return
    tmp_path = trim(path)//"/Cycle/MLST"
    call xml_tree_set_string_node_value(xml_tree,tmp_path,mlst,error_flag)
    IF (error_flag .ne. no_error) return
    tmp_path = trim(path)//"/Cycle/MLST_Drift"
    call xml_tree_set_string_node_value(xml_tree,tmp_path,mlst_drift,&
                                        error_flag)
    IF (error_flag .ne. no_error) return
    
  end subroutine add_ee_cycle
    !  #]
  subroutine add_ee_time_of_anx(xml_tree, path, &
       tai_of_anx, utc_of_anx, ut1_of_anx, error_flag)
    !  #[
    type(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: path
    character(len=*),    intent(in)  :: tai_of_anx
    character(len=*),    intent(in)  :: utc_of_anx
    character(len=*),    intent(in)  :: ut1_of_anx
    integer,             intent(out) :: error_flag

    ! local variables
    character(len=max_str_len) :: name
    character(len=max_str_len) :: tmp_path

    error_flag = no_error

    name = "Time_of_ANX"
    call xml_tree_add_child(xml_tree,path,name,error_flag)
    IF (error_flag .ne. no_error) return

    tmp_path = trim(path)//"/Time_of_ANX"
    name = "TAI_of_ANX"
    call xml_tree_add_child(xml_tree,tmp_path,name,error_flag)
    IF (error_flag .ne. no_error) return
    name = "UTC_of_ANX"
    call xml_tree_add_child(xml_tree,tmp_path,name,error_flag)
    IF (error_flag .ne. no_error) return
    name = "UT1_of_ANX"
    call xml_tree_add_child(xml_tree,tmp_path,name,error_flag)
    IF (error_flag .ne. no_error) return

    ! set the values of all these orbit items
    tmp_path = trim(path)//"/Time_of_ANX/TAI_of_ANX"
    call xml_tree_set_string_node_value(xml_tree,tmp_path,tai_of_anx,error_flag)
    IF (error_flag .ne. no_error) return
    tmp_path = trim(path)//"/Time_of_ANX/UTC_of_ANX"
    call xml_tree_set_string_node_value(xml_tree,tmp_path,utc_of_anx,error_flag)
    IF (error_flag .ne. no_error) return
    tmp_path = trim(path)//"/Time_of_ANX/UT1_of_ANX"
    call xml_tree_set_string_node_value(xml_tree,tmp_path,ut1_of_anx,error_flag)
    IF (error_flag .ne. no_error) return
    
  end subroutine add_ee_time_of_anx
    !  #]
  subroutine read_next_element(buffer, buffersize, bufferpos, &
                               text, element_type, error_flag)
    !  #[

    ! no length checking is done in this parsing routine.
    ! This is only done when the elements are added to the xml_tree

    character(len=1), dimension(:), pointer   :: buffer
    integer,          intent(in)    :: buffersize
    integer,          intent(inout) :: bufferpos
    character(len=*), intent(out)   :: text
    integer,          intent(out)   :: element_type
    integer,          intent(out)   :: error_flag

    ! local variables
    character(len=1) :: inp_char
    integer          :: i, j, n, startpos, endpos, text_len
    logical          :: text_found

    error_flag   = no_error
    text(:)      = ' '
    element_type = elem_is_undefined
    text_found   = .false.
    startpos     = -1
    endpos       = -1
    text_len     = -1

    ! scan the file until a < or > character is found
    i = bufferpos
    scanloop: DO
       IF (i .gt. buffersize) THEN
          element_type = eof_reached
          return
       END IF

       inp_char = buffer(i)
       if ( (inp_char .eq. '>') .or. &
            (inp_char .eq. '<')      ) exit scanloop
       i = i + 1
    END DO scanloop

    select case (inp_char)
    case('>')
       print *,"parsing problem"
       error_flag = error_reading_file
       return
    case('<')
       ! see if we have scanned over some text in the above loop
       text_found = .false.
       IF (suppress_empty_tag_strings) THEN
          testtextloop: DO j=bufferpos,i-1
             ! take anything except space, tab, cr and newline to be text
             IF ( (buffer(j) .ne. space)   .and. &
                  (buffer(j) .ne. tabchar) .and. &
                  (buffer(j) .ne. retchar) .and. &
                  (buffer(j) .ne. newline)      ) THEN
                text_found = .true.
                !print *,"char ",buffer(j)," has code: ",ichar(buffer(j))
                exit testtextloop
             END IF
          END DO testtextloop
       ELSE 
          ! just threat anything as text, even spaces, newlines etc.
          ! however, if the first remaining character is < or >
          ! we have no text
          IF (i .gt.  bufferpos) text_found = .true.
       END IF

       IF (text_found) THEN
          startpos = bufferpos
          endpos   = i-1
          !print *,"a piece of node text was found"
          !print *,"startpos = ",startpos," endpos = ",endpos
          !print *,"buffer(start:endpos) = [",buffer(startpos:endpos),"]"
          !DO j=startpos,endpos
          !   print *,"char ",buffer(j)," has code: ",ichar(buffer(j))
          !END DO
       ELSE
          startpos = i+1
          ! now scan untill the following '>' character is found
          ! and return all text in between them, which should
          ! define the node name, and all attributes belonging to it
          scanloop2: DO
             IF (i .gt. buffersize) THEN
                print *,"parsing problem"
                error_flag = error_reading_file
                return
             END IF
             
             inp_char = buffer(i)
             if (inp_char .eq. '>') exit scanloop2
             i = i + 1
          END DO scanloop2
          endpos = i-1
       END IF
       
    case default
       print *,"this should never be reached"
       print *,"parsing problem"
       error_flag = error_reading_file
       return
    end select

    ! report the found text
    !print *,"startpos, endpos = ",startpos, endpos
    text_len = endpos-startpos+1

    IF (text_len .gt. len(text)) THEN
       print *,"parsing problem"
       print *,"this text/tag section is larger than the text variable"
       print *,"used by the SimpleXML module (in subroutine read_next_element)"
       print *,"Probably something is wrong with this XML file..."
       print *,"len(text) = ",len(text)
       print *,"len(text/tag section) = ",text_len
       print *,"found text/tag section: ["//&
            chararray2string(buffer(startpos:endpos))//"]"
       error_flag = error_xml
       return
    END IF

    text = chararray2string(buffer(startpos:endpos))
    !print *,"text = ",trim(text)       

    ! see what type this is
    IF (text_found) THEN
       element_type = elem_is_node_text 
       ! report the position where we stopped scanning
       bufferpos = endpos+1
    ELSE
       IF (text(1:1) .eq. '?') THEN
          element_type = elem_is_xml_def
       ELSEIF (text(1:1) .eq. '/') THEN
          element_type = elem_is_node_end
       ELSEIF ((text_len .ge. 5) .and. (text(1:3) .eq. '!--') ) THEN
          ! handle the comment tags <!-- -->
          element_type = elem_is_comment
          ! double check
          IF (text(text_len-1:text_len) .ne. '--') THEN
             print *,"parsing problem: xml comment starting with '<!--' found"
             print *,"but it seems not closed with a proper '-->' "
             print *,"trim(text) = [",trim(text),"]"
             error_flag = error_reading_file
             return
          END IF
       ELSE
          n = len_trim(text)
          IF (text(n:n) .eq. '/') THEN
             element_type = elem_is_empty_node
          ELSE
             element_type = elem_is_node_start
          END IF
       END IF

       ! report the position where we stopped scanning
       bufferpos = endpos+2
    END IF

    return

  end subroutine read_next_element
    !  #]
  subroutine split_node_text(text, node_text, remaining_text)
    !  #[
    character(len=*), intent(in)  :: text
    character(len=*), intent(out) :: node_text
    character(len=*), intent(out) :: remaining_text

    ! local variable
    integer :: i

    ! find any spaces
    space_finding_loop: DO i=1,len_trim(text)
       IF (text(i:i) .eq. ' ') exit space_finding_loop
    END DO space_finding_loop
    
    IF (i .le. len_trim(text)) THEN
       node_text = text(1:i-1)
       remaining_text = adjustl(text(i+1:len_trim(text)))
    ELSE
       node_text = text
       remaining_text(:) = ' '
    END IF

  end subroutine split_node_text
    !  #]
  subroutine split_attr_text(text, attr_name, attr_value, remaining_text)
    !  #[
    character(len=*), intent(in)  :: text
    character(len=*), intent(out) :: attr_name
    character(len=*), intent(out) :: attr_value
    character(len=*), intent(out) :: remaining_text

    ! local variable
    integer          :: i, pos_space, pos_equal
    logical          :: inside_string 
    character(len=1) :: last_found_quote

    ! find any spaces, but exclude strings inside single
    ! or double quotes
    inside_string    = .false.
    last_found_quote = " "
    pos_equal = -1
    remaining_text(:) = ' '
    space_finding_loop: DO i=1,len_trim(text)
       IF (text(i:i) .eq. "'") THEN
          IF (inside_string) THEN
             ! only terminate the string if the right quote is found
             IF (text(i:i) .eq. last_found_quote) inside_string = .false.
          ELSE
             inside_string = .true.
             last_found_quote = text(i:i)
          END IF
       END IF
       IF (text(i:i) .eq. '"') THEN
          IF (inside_string) THEN
             ! only terminate the string if the right quote is found
             IF (text(i:i) .eq. last_found_quote) inside_string = .false.
          ELSE
             inside_string = .true.
             last_found_quote = text(i:i)
          END IF
       END IF

       IF (.not. inside_string) THEN
          IF (text(i:i) .eq. '=') pos_equal = i
          IF (text(i:i) .eq. ' ') exit space_finding_loop
       END IF
    END DO space_finding_loop

    pos_space = i
    
    IF (pos_equal .eq. -1) THEN
       print *,"ERROR: an attribute should contain a = sign"
       print *,"followed by its value, but this seems not the case here..."
       print *,"attr = ",trim(text)
    END IF

    IF (pos_space .le. len_trim(text)) THEN
       attr_name  = text(1:pos_equal-1)
       IF ( (text(pos_equal+1:pos_equal+1) .eq. "'") .or. &
            (text(pos_equal+1:pos_equal+1) .eq. '"')      ) THEN
          ! remove the quotes
          attr_value = text(pos_equal+2:pos_space-2)
       ELSE
          attr_value = text(pos_equal+1:pos_space-1)
       END IF
       remaining_text = adjustl(text(pos_space+1:len_trim(text)))
    ELSE
       attr_name  = text(1:pos_equal-1)
       IF ( (text(pos_equal+1:pos_equal+1) .eq. "'") .or. &
            (text(pos_equal+1:pos_equal+1) .eq. '"')      ) THEN
          ! remove the quotes
          attr_value = text(pos_equal+2:pos_space-2)
       ELSE
          attr_value = text(pos_equal+1:pos_space-1)
       END IF
       remaining_text(:) = ' '
    END IF

  end subroutine split_attr_text
    !  #]
  function get_nr_of_nodes(xml_tree, path, name) result(count)
    !  #[
    type(xml_node_type), pointer :: xml_tree ! input
    character(len=*)             :: path     ! input
    character(len=*)             :: name     ! input
    integer                      :: count    ! result

    ! local variables
    type(xml_node_type), pointer :: parent_node
    type(xml_node_type), pointer :: this_node

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"counting nodes:"
       print *,"path = ",trim(path)
       print *,"name = ",trim(name)
    END IF

    IF (.not. associated(xml_tree)) THEN
       count = -1
       return
    END IF

    parent_node => get_xml_node(xml_tree, path)
    IF (.not. associated(parent_node)) THEN
       count = -1
       return
    END IF
    
    ! walk through the list of siblings
    count = 0
    IF (.not. associated(parent_node%first_child)) THEN
       ! if the parent has no child yet, the count is 0
       count = 0
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"found nr of nodes = ",count
       return
    ELSE ! 
       ! add the new node in the list of siblings for this level
       this_node => parent_node%first_child
       IF (trim(name) .eq. trim(get_node_name(this_node)) ) & 
            count = count + 1
       siblingloop: DO
          IF (.not. associated(this_node%next_sibling)) exit siblingloop
          ! step to the next sibling
          this_node => this_node%next_sibling
          IF (trim(name) .eq. trim(get_node_name(this_node)) ) & 
               count = count + 1
       END DO siblingloop
    END IF

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"found nr of nodes = ",count

  end function get_nr_of_nodes
    !  #]
  subroutine add_child_to_node(xml_node,name,error_flag)
    !  #[
    type(xml_node_type), pointer     :: xml_node
    character(len=*),    intent(in)  :: name
    integer,             intent(out) :: error_flag

    ! local variables
    type(xml_node_type), pointer :: new_node
    type(xml_node_type), pointer :: this_node
    integer                      :: status

    error_flag = no_error

    IF (.not. associated(xml_node)) THEN
       error_flag = error_programming
       return
    END IF

    ! get memory for the new node
    allocate(new_node,STAT=status)
    IF (status .ne. 0) THEN
       error_flag = error_allocate
       return
    END IF

    ! fill the node fields
    call InitNode(new_node,error_flag)
    IF (error_flag .ne. no_error) return

    call set_node_name(new_node,name,error_flag)
    IF (error_flag .ne. no_error) return

    ! link the node to its parent
    new_node%parent => xml_node
    IF (.not. associated(xml_node%first_child)) THEN
       ! if the parent has no child yet, this is the one
       xml_node%first_child => new_node
    ELSE ! 
       ! add the new node in the list of siblings for this level
       this_node => xml_node%first_child
       siblingloop: DO
          IF (.not. associated(this_node%next_sibling)) exit siblingloop
          ! step to the next sibling
          this_node => this_node%next_sibling
       END DO siblingloop
       this_node%next_sibling => new_node
    END IF

    ! set the pointer to the root_node
    new_node%root_node => xml_node%root_node

    ! set the last_visited_node to the correct value
    new_node%root_node%last_visited_node => new_node

  end subroutine add_child_to_node
    !  #]
  subroutine InitNode(node,error_flag)
    !  #[
    type(xml_node_type), pointer     :: node
    integer,             intent(out) :: error_flag

    error_flag = no_error

    IF (.not. associated(node)) THEN
       error_flag = error_programming
       return
    END IF

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"initialising node"
    nullify(node%name)
    nullify(node%value)
    nullify(node%first_attr)
    nullify(node%last_attr)
    nullify(node%parent)
    nullify(node%first_child)
    nullify(node%next_sibling)
    nullify(node%root_node)
    node%is_root_node = .false.
    nullify(node%last_visited_node)
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"node has been initialised"

  end subroutine InitNode
    !  #]
  recursive subroutine DeleteNode(node, error_flag)
    !  #[

    type(xml_node_type), pointer     :: node
    integer,             intent(out) :: error_flag

    error_flag = no_error

    IF (.not. associated(node)) THEN
       error_flag = error_programming
       return
    END IF

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"deleting node: ",trim(get_node_name(node))

    IF (associated(node%name)) deallocate(node%name)
    nullify(node%name)

    IF (associated(node%value)) deallocate(node%value)

    IF (associated(node%first_attr)) THEN
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"attributes will be deleted"
       call delete_attr(node%first_attr, error_flag)
    ELSE
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"no attributes to delete"
    END IF

    IF (associated(node%first_child)) THEN
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"child will be deleted first"
       call DeleteNode(node%first_child, error_flag)
       if (error_flag .ne. no_error) return
       deallocate(node%first_child)
    ELSE
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"no children to delete"
    END IF

    IF (associated(node%next_sibling)) THEN
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"sibling will be deleted first"
       call DeleteNode(node%next_sibling,error_flag)
       if (error_flag .ne. no_error) return
       deallocate(node%next_sibling)
    ELSE
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"no sibling to delete"
    END IF

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"node was deleted"

  end subroutine DeleteNode
    !  #]
  function get_node_name(node) result(name)
    !  #[

    type(xml_node_type), pointer :: node ! input
    character(len=max_str_len)   :: name ! result

    IF (.not. associated(node)) THEN
       name = "[ERROR: node pointer was not associated]"
       return
    END IF

    IF (.not. associated(node%name)) THEN
       name = "[node name is not yet undefined]"
       return
    END IF

    name(:) = ' ' ! init to an empty string first because chararray2string
                  ! only fills the string upto the length of node%name !!!!!
    name = chararray2string(node%name)
    
  end function get_node_name
    !  #]
  subroutine set_node_name(node,name,error_flag)
    !  #[
    type(xml_node_type), pointer     :: node
    character(len=*),    intent(in)  :: name       ! input
    integer,             intent(out) :: error_flag ! output

    ! local variables
    integer :: status

    error_flag = no_error

    IF (.not. associated(node)) THEN
       error_flag = error_programming
       return
    END IF

    ! check for the ee_cfi imposed maximum
    IF (len_trim(name) .gt. max_node_name_len) THEN
       print *,"Error in set_node_name:"
       print *,"node name string should not be larger than ",&
            max_node_name_len
       print *,"characters (this is an ee_cfi constraint!)"
       print *,"name = ["//trim(name)//"]"
       error_flag = error_xml
       return
    END IF

    ! check the validity of the name to see if any
    ! illegal characters (like slashes, spaces etc.) are present
    call check_xml_node_name(name, error_flag)
    IF (error_flag .ne. no_error) return

    ! deallocate the name in case it should be reset
    IF (associated(node%name)) deallocate(node%name)

    ! allocate space for this string
    allocate(node%name(len_trim(name)),STAT=status)
    if (status .ne. 0) then
       error_flag = error_allocate
       return
    end if

    ! copy the name from the string to the char array
    node%name = string2chararray(name)
    
  end subroutine set_node_name
    !  #]
  subroutine check_xml_node_name(name, error_flag)
    !  #[

    ! check the validity of the name to see if any
    ! illegal characters (like slashes, spaces etc.) are present
    character(len=*),    intent(in)  :: name       ! input
    integer,             intent(out) :: error_flag ! output

    ! local variables and parameters
    integer :: pos

    ! the current portland compiler (version 5.2) on KNMI does not support 
    ! the next construct (this is a BUG in this compiler!)
    ! and complains with the error: 
    !     PGF90-S-0069-Illegal implied DO expression (SimpleXML.F90: 1434)
    !     PGF90-S-0146-Expression must be character type (SimpleXML.F90: 1434)
    !
    !character(len=1), dimension(2*26+10+1), parameter :: allowed_chars = &
    !     (/ (char(i),i=ichar('a'),ichar('z')), &
    !        (char(i),i=ichar('A'),ichar('Z')), &
    !        (char(i),i=ichar('0'),ichar('9')), &
    !        '_' /)
    ! a workaround is to type all characters in the list ....
    character(len=1), dimension(2*26+10+1), parameter :: allowed_chars = &
         (/ 'a','b','c','d','e','f','g','h','i','j','k','l','m',&
            'n','o','p','q','r','s','t','u','v','w','x','y','z',&
            'A','B','C','D','E','F','G','H','I','J','K','L','M',&
            'N','O','P','Q','R','S','T','U','V','W','X','Y','Z',&
            '0','1','2','3','4','5','6','7','8','9', &
            '_' /)

    character(len=256), save :: allowed_char_string
    logical,            save :: first_call = .true.

    IF (first_call) THEN
       ! this needs just to be converted once
       allowed_char_string = chararray2string(allowed_chars)
       first_call = .false.
    END IF

    error_flag = no_error

    pos = verify(trim(name),trim(allowed_char_string))
    IF (pos .ne. 0) THEN
       ! some illegal character appears to be present !!!
       print *,"ERROR in check_xml_node_name: an illegal character seems to"
       print *,"be present in this name:"
       print *,"name = ",trim(name)
       print *,"pos = ",pos," name(pos:pos)=",name(pos:pos)
       print *,"allowed_chars = ",allowed_chars
       error_flag = error_programming
       return
    END IF

  end subroutine check_xml_node_name
    !  #]
  subroutine check_xml_attr_name(attr_name, error_flag)
    !  #[
    ! check the validity of the attribute name to see if any
    ! illegal characters (like slashes, spaces etc.) are present
    character(len=*),    intent(in)  :: attr_name  ! input
    integer,             intent(out) :: error_flag ! output

    ! local variables and parameters
    integer :: pos

    ! in attribute names only lowercase is allowed
    character(len=1), dimension(26), parameter :: allowed_chars = &
         (/ 'a','b','c','d','e','f','g','h','i','j','k','l','m',&
            'n','o','p','q','r','s','t','u','v','w','x','y','z' /)
    character(len=256), save :: allowed_char_string
    logical,            save :: first_call = .true.

    IF (first_call) THEN
       ! this needs just to be converted once
       allowed_char_string = chararray2string(allowed_chars)
       first_call = .false.
    END IF

    error_flag = no_error

    IF (Do_Allow_xmlns_Attr) THEN
       ! allow 2 namespace related attributes
       IF (len(attr_name) .ge. 6) THEN
          IF (attr_name(1:6) .eq. "xmlns:") return
       END IF
       IF (len(attr_name) .ge. 4) THEN
          IF (attr_name(1:4) .eq. "xsi:") return
       END IF
       ! exceptions needed to allow defining namespaces
       IF (trim(attr_name) .eq. "schemaLocation") return
       IF (trim(attr_name) .eq. "noNamespaceSchemaLocation") return
    END IF

    pos = verify(trim(attr_name),trim(allowed_char_string))
    IF (pos .ne. 0) THEN
       ! some illegal character appears to be present !!!
       print *,"ERROR in check_xml_attr_name: an illegal character seems to"
       print *,"be present in this attribute name:"
       print *,"attr_name = ",trim(attr_name)
       print *,"pos = ",pos," attr_name(pos:pos)=",attr_name(pos:pos)
       print *,"allowed_chars = ",allowed_chars
       print *,"remember that is is a ee_cfi convention to use only"
       print *,"lowercase letters in attribute names ..."
       error_flag = error_programming
       return
    END IF

  end subroutine check_xml_attr_name
    !  #]
  function chararray2string(chararray) result(text)
    !  #[

    character(len=1), dimension(:) :: chararray ! input
    character(len=size(chararray)) :: text ! output

    integer :: n, i
    n = size(chararray)

    ! copy the data from the char array to a string
    DO i=1,n
       text(i:i) = chararray(i)
    END DO

  end function chararray2string
    !  #]
  function string2chararray(text) result(chararray)
    !  #[
    character(len=*) :: text ! output
    character(len=1), dimension(len_trim(text)) :: chararray ! output

    integer :: n, i

    n = len_trim(text)
    IF (n .gt. max_str_len) n = max_str_len

    ! copy the text from the string to the char array
    ! dont use transfer here, since then the type will be converted
    ! in the same memory location, which is not how a function
    ! should behave
    DO i=1,n
        chararray(i) = text(i:i)
    END DO
    
  end function string2chararray
    !  #]
  function to_uppercase(txt_in) result(txt_out)
    !  #[
    character(len=*)           :: txt_in  ! input
    character(len=len(txt_in)) :: txt_out ! result

    ! local variables
    integer          :: i,n
    character(len=1) :: c

    txt_out(:) = ' '
    n = len_trim(txt_in)
    IF (n .gt. 0) THEN
       DO i=1,n
          c = txt_in(i:i)
          IF ( (c .ge. 'a') .and. (c .le. 'z')) THEN
             c = achar(ichar(c)+ichar('A')-ichar('a'))
          END IF
          txt_out(i:i) = c
       END DO
    END IF

  end function to_uppercase
    !  #]
  subroutine step_to_parent(xml_tree, error_flag)
    !  #[
    type(xml_node_type), pointer     :: xml_tree   ! input/output
    integer,             intent(out) :: error_flag ! output

    ! local variable
    type(xml_node_type), pointer :: this_node ! input/output

    error_flag = no_error
    IF (.not. associated(xml_tree)) THEN
       error_flag = error_programming
       return
    END IF
    
    IF (.not. xml_tree%is_root_node) THEN
       error_flag = error_programming
       return
    END IF

    this_node => xml_tree%last_visited_node
    
    IF (associated(this_node)) THEN
       xml_tree%last_visited_node => this_node%parent
    ELSE
       print *,"ERROR: last_vsited_node was not set !!!"
       error_flag = error_programming
       return       
    END IF
    
  end subroutine step_to_parent
    !  #]
  recursive function  get_xml_node(xml_node, path, fail_silently) &
       result(found_xml_node)
    !  #[
    ! find the node belonging to the first element of the path
    type(xml_node_type), pointer :: xml_node       ! input
    character(len=*) :: path                       ! input
    type(xml_node_type), pointer :: found_xml_node ! result

    logical, optional, intent(in) :: fail_silently

    ! local variables
    integer :: n1, n2, i, n, count, num_matching_tags, ios
    character(len=max_str_len) :: tag
    logical :: final_tag, do_fail_silently
    type(xml_node_type), pointer :: this_node
    type(xml_node_type), pointer :: root_node

    do_fail_silently = .false.
    if (present(fail_silently)) do_fail_silently = fail_silently

    IF (.not. associated(xml_node)) THEN
       print *,"ERROR in get_xml_node: input node seems not associated !!!"
       nullify(found_xml_node)
       return
    END IF

    ! check for the ee_cfi imposed maximum
    IF (len_trim(path) .gt. max_path_len) THEN
       print *,"Error in get_xml_node:"
       print *,"path names should not be larger than ",max_path_len
       print *,"characters (this is an ee_cfi constraint!)"
       print *,"The current path is: ["//trim(path)//"]"
       print *,"and seems to have ",len_trim(path)," characters ..."
       nullify(found_xml_node)
       return
    END IF

    ! go to the root of the tree
    root_node => xml_node%root_node

    ! if first char of the path is a slash, check if it is the root
    n1 = 1
    if (path(1:1) .eq. '/') then
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"path seems to start with a slash !"
       n1 = 2
       IF (.not. xml_node%is_root_node) THEN
          print *,"ERROR in get_xml_node: path starts with a slash"
          print *,"buth the given xml_node is not root !!!"
          nullify(found_xml_node)
          return
       END IF
    end if

    ! if the first 2 chars of the path are "./" then dont use the given
    ! xml_node, but get the last used node, and work relative from it
    IF (path(1:2) .eq. "./" ) n1 = 3

    ! if the first 3 chars of the path are "../" then dont use the given
    ! xml_node, but get the last used node, and go to its parent node,
    ! and work relative from it.
    ! (this is done below just before the sibling_loop)
    ! before entering the recursion
    IF (path(1:3) .eq. "../") n1 = 4

    ! find the next slash in the path name
    n2 = -1
    searchloop: DO i=n1,len_trim(path)
       if (path(i:i) .eq. '/') then
          n2 = i-1
          exit searchloop
       end if
    END DO searchloop

    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"n1=",n1," n2=",n2
    
    final_tag = .false.
    if (n2 .eq. -1) then
       ! no ending slash found, so the whole rest of the string 
       ! belongs to this tag
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"no ending slash found"
       n2 = len_trim(path)
       final_tag = .true.
    end if

    IF (n2 .le. n1) THEN 
       ! ok, we are done
       IF (path(1:3) .eq. "../") THEN
          IF (LocalVerbosity .ge. VerbosityReport) &
               print *,"using parent of last used node [1], n1=",n1," n2=",n2
          this_node => root_node%last_visited_node
          IF (LocalVerbosity .ge. VerbosityReport) &
               print *,"last visited node [1] = ", &
                        trim(get_node_name(this_node))
          IF (.not. associated(this_node)) THEN
             print *,"ERROR in get_xml_node() !!!"
             print *,"root_node%last_visited_node was not yet defined !!!"
             nullify(found_xml_node)
             return
          END IF
          this_node => this_node%parent
          IF (LocalVerbosity .ge. VerbosityReport) &
               print *,"parent of last visited node [1] = ", &
                        trim(get_node_name(this_node))
          found_xml_node => this_node

          ! remember the pointer to this node as the last visited one
          root_node%last_visited_node => found_xml_node
          return
       ELSE
          IF (path(1:2) .eq. "./") THEN
             IF (LocalVerbosity .ge. VerbosityReport) &
                  print *,"using last used node [2], n1=",n1," n2=",n2
             this_node => root_node%last_visited_node
             IF (LocalVerbosity .ge. VerbosityReport) &
                  print *,"last visited node [2] = ", &
                          trim(get_node_name(this_node))
             found_xml_node => this_node
             return
          ELSE
             print *,"ERROR, this code should never be reached ..."
             nullify(found_xml_node)
             return
          END IF
       END IF
    END IF

    tag = path(n1:n2)
    IF (LocalVerbosity .ge. VerbosityReport) &
         print *,"tag=",trim(tag)

    ! see if a count was present at the end of the tag
    count = 1 ! by default, take the first occurrence
    n = len_trim(tag)
    IF (tag(n:n) .eq. ']') THEN
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"a count is present !!!"
       ! find the starting bracket for this count
       bracketloop: DO i=n,1,-1
          IF (tag(i:i) .eq. '[') exit bracketloop
       END DO bracketloop
       IF (LocalVerbosity .ge. VerbosityReport) THEN
          print *,"[count] = ",tag(i:n)
          print *,"count = ",tag(i+1:n-1)
          print *,"tag = ",tag(1:n)
       END IF

       read(tag(i+1:n-1),"(i6)",iostat=ios) count 
       IF (ios .ne. 0) goto 999
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"int count = ",count

       ! we have the count now, so we can strip it from the tag string
       tag(i:n) = ' '
    ELSE
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"no count present !!!"
    END IF

    ! find the wanted tag in this node or one of its siblings

    ! if the first 3 chars of the path are "../" then go to the parent node
    ! before entering the recursion
    IF (path(1:3) .eq. "../") THEN
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"using parent of last used node [3], n1=",n1," n2=",n2
       this_node => root_node%last_visited_node
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"last visited node [3] = ", &
                    trim(get_node_name(this_node))
       IF (.not. associated(this_node)) THEN
          print *,"ERROR in get_xml_node() !!!"
          print *,"root_node%last_visited_node was not yet defined !!!"
          nullify(found_xml_node)
          return
       END IF
       this_node => this_node%parent
       IF (LocalVerbosity .ge. VerbosityReport) &
            print *,"parent of last visited node [3] = ", &
                    trim(get_node_name(this_node))

       ! pass the remainder of the path (without the leading ../)
       ! on in the recursive call to this subroutine
       found_xml_node => get_xml_node(this_node%first_child,&
                                      path(4:len_trim(path)))
       return
    ELSE
       IF (path(1:2) .eq. "./") THEN
          IF (LocalVerbosity .ge. VerbosityReport) &
               print *,"using last used node [4], n1=",n1," n2=",n2
          this_node => root_node%last_visited_node
          IF (LocalVerbosity .ge. VerbosityReport) &
               print *,"last visited node [4] = ", &
                       trim(get_node_name(this_node))
          IF (.not. associated(this_node)) THEN
             print *,"ERROR in get_xml_node() !!!"
             print *,"root_node%last_visited_node was not yet defined !!!"
             nullify(found_xml_node)
             return
          END IF

          ! pass the remainder of the path (without the leading ./)
          ! on in the recursive call to this subroutine
          found_xml_node => get_xml_node(this_node%first_child,&
                                         path(3:len_trim(path)))
          return
       ELSE
          this_node => xml_node
       END IF
    END IF

    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"entering sibling loop"
       print *,"trim(tag): ",trim(tag)

       IF (TagNamesCaseSensitive) THEN
          print *,"trim(tag): ",trim(tag)
       ELSE
          print *,"to_uppercase(trim(tag)): ",to_uppercase(trim(tag))
       END IF

       print *,"trim(get_node_name(this_node)): ",&
            trim(get_node_name(this_node))

       IF (TagNamesCaseSensitive) THEN
          print *,"node_name: ",trim(get_node_name(this_node))
       ELSE
          print *,"node_name: ",to_uppercase(trim(get_node_name(this_node)))
       END IF
    END IF

    num_matching_tags = 0
    sibling_loop: do
       IF (LocalVerbosity .ge. VerbosityReport) THEN
          IF (TagNamesCaseSensitive) THEN
             print *,"tag: ",trim(tag)
             print *,"node_name: ",trim(get_node_name(this_node))
          ELSE
             print *,"tag: ",to_uppercase(trim(tag))
             print *,"node_name: ",to_uppercase(trim(get_node_name(this_node)))
          END IF
       END IF

       IF (TagNamesCaseSensitive) THEN
          IF ( trim(tag) .eq. trim(get_node_name(this_node))) THEN
             num_matching_tags = num_matching_tags + 1
          END IF
       ELSE
          IF ( to_uppercase(trim(tag)) .eq. &
               to_uppercase(trim(get_node_name(this_node)))) THEN
             num_matching_tags = num_matching_tags + 1
          END IF
       END IF

       if (count .eq. num_matching_tags) exit sibling_loop

       ! this is not the right one, so try the next sibling
       if (.not. associated(this_node%next_sibling)) then
          if (.not. do_fail_silently) then
             print *,"ERROR in get_xml_node: tag not found !!!"
             print *,"requested tag: ",trim(tag)
             print *,"requested path: ",trim(path)
          end if
          nullify(found_xml_node)
          return
       end if
       this_node => this_node%next_sibling
    end do sibling_loop

    if (final_tag) then
       ! ok, we are done
       found_xml_node => this_node
       ! remember the pointer to this node as the last visited one
       root_node%last_visited_node => found_xml_node
    else
       if (.not. associated(this_node%first_child)) then
          if (.not. do_fail_silently) then
             print *,"ERROR in get_xml_node: tag not found !!!"
             print *,"requested tag: ...",trim(path)
          end if
          nullify(found_xml_node)
          return
       end if
       ! pass the remainder of the path (without the leading slash)
       ! on in the recursive call to this subroutine
       found_xml_node => get_xml_node(this_node%first_child,&
                                      path(n2+2:len_trim(path)))
    end if

    return

999 print *,"ERROR while reading integer from string: ",trim(tag(i+1:n-1))
    nullify(found_xml_node)
    return

  end function get_xml_node
    !  #]
  recursive subroutine write_node(fileunit, xml_node, error_flag, &
                                  recursion_level_in)
    !  #[

    integer,               intent(in) :: fileunit
    type(xml_node_type),   pointer     :: xml_node
    integer,               intent(out) :: error_flag
    integer,     optional, intent(in)  :: recursion_level_in

    ! local variables
    integer :: recursion_level
    character(len=max_str_len) :: name 
    character(len=max_str_len) :: text
    logical :: child_present
    logical :: sibling_present

    error_flag = no_error

    IF (.not. associated(xml_node)) THEN
       error_flag = error_programming
       return
    END IF

    recursion_level = 1
    if (present(recursion_level_in)) &
         recursion_level = recursion_level_in

    child_present = .false.
    if (associated(xml_node%first_child)) child_present = .true.
    
    sibling_present = .false.
    if (associated(xml_node%next_sibling)) sibling_present = .true.

    ! print the node data
    name = get_node_name(xml_node)

    if (child_present) then
       ! print the opening tag
       text = "<"//trim(name)//&
            trim(get_attributes(xml_node%first_attr))//">"
       call write_indented(fileunit, text, recursion_level)
    else
       if (associated(xml_node%value)) then
          ! print this tag, with the text string as content
          ! for now, no attributes should be present in this case 
          ! but adding them doesn't hurt
          text = "<"//trim(name)//&
               trim(get_attributes(xml_node%first_attr))//">"//&
                 trim(chararray2string(xml_node%value))//&
                 "</"//trim(name)//">"
          call write_indented(fileunit, text,recursion_level)
       else
          IF (DoUseEmptyTags) THEN
             ! print an empty tag
             text = "<"//trim(name)//&
                  trim(get_attributes(xml_node%first_attr))//"/>"
             call write_indented(fileunit, text,recursion_level)
          ELSE
             ! print a tag pair, without any content inbetween
             ! (in the ee_cfi file format this is preferred
             ! above using empty tags !!)
             text = "<"//trim(name)//&
                  trim(get_attributes(xml_node%first_attr))//">"//&
                 "</"//trim(name)//">"
             call write_indented(fileunit, text,recursion_level)
          END IF
       end if
    end if

    ! then print all children (if present)
    if (child_present) then
       call write_node(fileunit, xml_node%first_child, error_flag, &
                       recursion_level_in = recursion_level+1)
       IF (error_flag .ne. no_error) return
    end if

    ! print the closing tag
    if (child_present) then
       text = "</"//trim(name)//">"
       call write_indented(fileunit, text,recursion_level)
    end if

    ! then print the next sibling (if present)
    if (sibling_present) then
       call write_node(fileunit, xml_node%next_sibling, error_flag, &
                       recursion_level_in = recursion_level)
       IF (error_flag .ne. no_error) return
    end if

  end subroutine write_node

    !  #]
  subroutine write_indented(fileunit,text,recursion_level)
    !  #[
    integer,          intent(in) :: fileunit
    character(len=*), intent(in) :: text
    integer,          intent(in) :: recursion_level

    ! local variables
    character(len=len(text)+(recursion_level-1)*2) :: tmp_text
    integer :: n

    n=len_trim(text)

    ! fill the string with spaces
    tmp_text(:) = ' '

    ! copy the text to be printed after the desired nr of indentation spaces
    tmp_text((recursion_level-1)*2+1:(recursion_level-1)*2+n) = text(1:n)
    write(fileunit,"(a)") trim(tmp_text)

  end subroutine write_indented
    !  #]
  subroutine add_attr_to_node(xml_node,attr_name,attr_value,error_flag,at_end)
    !  #[

    type(xml_node_type), pointer     :: xml_node
    character(len=*),    intent(in)  :: attr_name
    character(len=*),    intent(in)  :: attr_value
    integer,             intent(out) :: error_flag
    logical, optional,   intent(in)  :: at_end
    
    ! local variables
    TYPE(node_attr_type), pointer :: new_attr
    TYPE(node_attr_type), pointer :: this_attr    
    integer                       :: status
    logical                       :: add_at_the_end

    error_flag = no_error

    IF (.not. associated(xml_node)) THEN
       error_flag = error_programming
       return
    END IF

    ! the xmlns:xsi attr must allways be defined before the xsi:schemaLocation
    ! and xsi:NoNamespaceSchemaLocation tags.
    ! To make sure this happens, the xmlns attributes will be added in front
    ! of the linked list of attributes, and the xsi:* attributes
    ! will be added at the end of the linked list of attributes.
    ! This is why the at_end optional input flag is needed

    add_at_the_end = .false.
    if (present(at_end)) add_at_the_end = .true.

    ! to add: check for double attributes with the same name
    ! (is that forbidden or not ?? it seems to me it should not be possible)

    ! check for the ee_cfi imposed maximum
    IF (len_trim(attr_name) .gt. max_attr_name_len) THEN
       print *,"Warning in add_attr_to_node:"
       print *,"attribute name string should not be larger than ",&
            max_attr_name_len
       print *,"characters (this is an ee_cfi constraint!)"
       error_flag = error_xml
       return
    END IF

    ! check for the ee_cfi imposed maximum
    IF (len_trim(attr_value) .gt. max_attr_val_len) THEN
       print *,"Warning in add_attr_to_node:"
       print *,"attribute value string should not be larger than ",&
            max_attr_val_len
       print *,"characters (this is an ee_cfi constraint!)"
       error_flag = error_xml
       return
    END IF

    this_attr => xml_node%first_attr
    attr_loop: DO
       IF (.not. associated(this_attr)) exit attr_loop

       IF (LocalVerbosity .ge. VerbosityReport) THEN
          print *,"already defined attr: "
          print *,"name=",trim(chararray2string(this_attr%name)),&
               " value=",trim(chararray2string(this_attr%value))
       END IF

       IF (trim(attr_name) .eq. trim(chararray2string(this_attr%name))) THEN
          print *,"ERROR in add_attr_to_node: a node cannot have two"
          print *,"attributes with identical name ..."
          print *,"new attribute name: ",trim(attr_name)
          print *,"added attribute name: ",&
               trim(chararray2string(this_attr%name))
          print *,"added attribute value: ",&
               trim(chararray2string(this_attr%value))
          error_flag = error_programming
          return
       END IF

       ! try the next attribute in the list
       this_attr => this_attr%next_attr
    END DO attr_loop

    ! get memory for the new attribute
    allocate(new_attr,STAT=status)
    IF (status .ne. 0) THEN
       error_flag = error_allocate
       return
    END IF    
    allocate(new_attr%name(len_trim(attr_name)),STAT=status)
    if (status .ne. 0) then
       error_flag = error_allocate
       return
    end if
    allocate(new_attr%value(len_trim(attr_value)),STAT=status)
    if (status .ne. 0) then
       error_flag = error_allocate
       return
    end if

    ! fill the new attr
    new_attr%name  = string2chararray(trim(attr_name))
    new_attr%value = string2chararray(trim(attr_value))

    if (.not. associated(xml_node%first_attr)) then
       ! for the first attr of this tag
       ! point the last_attr pointer to it as well
       xml_node%first_attr => new_attr
       xml_node%last_attr  => new_attr
       nullify(new_attr%next_attr)
       return
    end if

    ! link the new attribute to the data structure
    if (add_at_the_end) then
       ! add this attribute at the end of the linked list of attributes
       xml_node%last_attr%next_attr => new_attr
       xml_node%last_attr           => new_attr
       nullify(new_attr%next_attr)
    else
       ! add this attribute at the head of the linked list of attributes
       new_attr%next_attr => xml_node%first_attr
       xml_node%first_attr => new_attr
    end if

    ! done
    IF (LocalVerbosity .ge. VerbosityReport) &
       print *,"attribute successfully added"

  end subroutine add_attr_to_node
    !  #]
  subroutine get_attr_value_from_node(xml_node,attr_name,attr_value,&
                                      error_flag,silent)
    !  #[

    type(xml_node_type), pointer     :: xml_node
    character(len=*),    intent(in)  :: attr_name
    character(len=*),    intent(out) :: attr_value
    integer,             intent(out) :: error_flag
    logical, optional,   intent(in)  :: silent
 
    ! local variables
    TYPE(node_attr_type), pointer :: this_attr    
    logical :: attr_found

    error_flag = no_error

    IF (.not. associated(xml_node)) THEN
       error_flag = error_programming
       return
    END IF

    attr_found = .false.
    attr_value(:) = ' '

    this_attr => xml_node%first_attr
    attr_loop: DO
       IF (.not. associated(this_attr)) exit attr_loop

       IF (LocalVerbosity .ge. VerbosityReport) THEN
          print *,"checking attr: "
          print *,"name=",trim(chararray2string(this_attr%name)),&
               " value=",trim(chararray2string(this_attr%value))
       END IF

       IF (trim(attr_name) .eq. trim(chararray2string(this_attr%name))) THEN
          attr_found = .true.
          attr_value = chararray2string(this_attr%value)
          exit attr_loop
       END IF

       ! try the next attribute in the list
       this_attr => this_attr%next_attr
    END DO attr_loop

    IF (.not. attr_found) THEN
       if (present(silent)) then
          if (silent) then
             ! just return with the empty attr_value string
             ! NOTE: this is a workaround!
             return
          end if
       end if
       print *,"ERROR in get_attr_value_from_node():"
       print *,"==>could not find an attribute with name: ",trim(attr_name)
       print *,"in this node."
       print *,"trim(xml_node%name) = ",trim(chararray2string(xml_node%name))
       error_flag = error_reading_file
       return
    END IF

    ! done
    IF (LocalVerbosity .ge. VerbosityReport) THEN
       print *,"attribute successfully read: "
       print *,"==> this_attr%name = [", &
            trim(chararray2string(this_attr%name)),"]"
       print *,"==> this_attr%value = [", &
            trim(chararray2string(this_attr%value)),"]"
    END IF

  end subroutine get_attr_value_from_node
    !  #]
  recursive subroutine delete_attr(xml_attr, error_flag)
    !  #[
    TYPE(node_attr_type), pointer     :: xml_attr
    integer,              intent(out) :: error_flag
 
    error_flag = no_error

    ! if more attributes are present, delete them first
    IF (associated(xml_attr%next_attr)) THEN
       call delete_attr(xml_attr%next_attr, error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    IF (LocalVerbosity .ge. VerbosityReport) THEN
          print *,"deleting attribute:",&
               " name=",trim(chararray2string(xml_attr%name)),&
               " value=",trim(chararray2string(xml_attr%value))
    END IF

    deallocate(xml_attr%name)
    deallocate(xml_attr%value)
    deallocate(xml_attr)

  end subroutine delete_attr
    !  #]
  recursive function get_attributes(xml_attr) result(attr_text)
    !  #[
    TYPE(node_attr_type), pointer    :: xml_attr ! input    
    character(len=max_str_len) :: attr_text

    attr_text(:) = ' '
    IF (associated(xml_attr)) THEN
       attr_text = " "//trim(chararray2string(xml_attr%name))//&
                   '="'//trim(chararray2string(xml_attr%value))//&
                   '"'//trim(get_attributes(xml_attr%next_attr))
    END IF

  end function get_attributes
    !  #]
  !----------------------------------------

end module XML_module
