MODULE AeolusConstants
  !  #[ Documentation
  !---------------------------------------------------
  !    This module introduces some constants used by the AEOLUS L2B processor
  !
  !    Written by: Paul Poli, METEO FRANCE,  10-Jun-2005
  !
  !    Modifications:
  !    01-Sep-2005 J. de Kloe  Added software_version
  !    12-Sep-2005 D. Tan      Software_version reduced to 0.1
  !    28-Feb-2006 D. Tan      Software_version 0.2
  !    03-Apr-2006 J. de Kloe  added software_ProcName 
  !    27-Apr-2006 J. de Kloe  added software_ExecutableName 
  !    04-May-2006 D. Tan      Software_version 01.00
  !    12-May-2006 D. Tan      Null use of ParKind1 module
  !    22-May-2006 D. Tan      Software_version 01.10
  !    07-Dec-2006 D. Tan      Software_version 01.20
  !    28-Mar-2007 D. Tan      Software_version 01.30
  !    05-Jun-2007 D. Tan      Software_version 01.31
  !    15-Jun-2007 J. de Kloe  disabled some unused items, and added some
  !                            integer items defining the Aeolus (max) values
  !                            for Num_Meas/Num_Obs/Num_RangeBins
  !    05-Nov-2007 J. de Kloe  added IODD version string parameters
  !    13-Dec-2007 D. Tan      Software_version 01.32
  !    21-Jan-2008 J. de Kloe  incremented L1B IODD version
  !    21-Feb-2008 J. de Kloe  again incremented L1B IODD version
  !                            and made old version number available as well
  !    22-Feb-2008 D. Tan      RBC IODD version 01.20 (not 01.02), Ref_Doc strings.
  !    26-Feb-2008 D. Tan      Software_version 01.33
  !---------------------------------------------------
  !  #]
  !  #[ modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE Numerics, only: r8_, i2_
  !  #]
  !  #[ global constants
  IMPLICIT NONE

! TODO:
! remove the real physical constants like c and kb from this file
! and move them to a module named PhysicsConstants.F90.  JK. 27-04-2006

! REMARK: I remember having read a requirement that says that we
!         should read all these constants from an inputfile
!         but I don't remember exactly where this is stated.
! TODO: check if this is true. JK. 27-04-2006

  ! Software version for the L2BP.
  ! This version is used to fill the version field in the Fixed Header
  ! when writing ee_cfi formatted output files, like the L2B 
  ! or the Aux. Met. Data output files
  character(len=*), parameter :: software_version  = "01.33"
  ! the next name is used/reported by the Logging module
  character(len=*), parameter :: software_ProcName = "L2BP"
  ! the name that will be used to refer to the main executable program 
  ! of the L2BP by the ThinLayer 
  character(len=*), parameter :: software_TaskName = "IPF2B_L1B_L2B"

  ! define IODD version numbers used by the software
  ! these are used to construct the namespace strings, which we
  ! have to use in the root-tag of our xml output files         ! issue date
  character(len=*), parameter :: iodd_version_L1B_N   = "03.05" ! 200802??
  character(len=*), parameter :: iodd_version_L2B_N   = "01.32" ! 20070615
  character(len=*), parameter :: iodd_version_L2C_N   = iodd_version_L2B_N
  character(len=*), parameter :: iodd_version_AuxMet  = iodd_version_L2B_N
  character(len=*), parameter :: iodd_version_AuxRBC  = "01.20" ! 20070615
  character(len=*), parameter :: iodd_version_AuxClim = iodd_version_L2B_N
  character(len=*), parameter :: iodd_version_L2A_N   = "01.03" ! 20070117
  character(len=*), parameter :: iodd_version_AuxCal  = iodd_version_L2A_N
  character(len=*), parameter :: iodd_version_AuxPar  = iodd_version_L2B_N

  ! allow old L1B iodd versions for compatibility reasons
  ! (needed to read and check L1B v1.08 files)
  character(len=*), parameter :: iodd_version_L1B_N_108 = "03.04" ! 20070412

  ! Strings for the Ref_Doc tag
  character(len=*), parameter :: Ref_Doc_Root_L2BC = "L2B/L2C IODD Iss. "
  character(len=*), parameter :: Ref_Doc_Root_RBC  = "RBC IODD Iss. "

  character(len=*), parameter :: Ref_Doc_L2BC   = Ref_Doc_Root_L2BC//iodd_version_L2B_N
  character(len=*), parameter :: Ref_Doc_AuxMet = Ref_Doc_L2BC
  character(len=*), parameter :: Ref_Doc_AuxCLM = Ref_Doc_L2BC
  character(len=*), parameter :: Ref_Doc_AuxPar = Ref_Doc_L2BC
  character(len=*), parameter :: Ref_Doc_AuxRBC = Ref_Doc_Root_RBC//iodd_version_AuxRBC

  ! define maximum values for number of range gates, measurements
  ! and observations to be used to setup the L2B datastructures
  integer(i2_), parameter :: N_range_gates      = 24 ! current used nr of gates
  integer(i2_), parameter :: N_Meas_Max         = 67 ! max. nr. of meas.
  integer(i2_), parameter :: N_Obs_Max_Mie      = 2  ! max. nr. of obs. Mie
  integer(i2_), parameter :: N_Obs_Max_Rayleigh = 2  ! max. nr. of obs. Rayl
  ! NOTE: the i2_ integer type corresponds to the IntAuc type used in the 
  ! rest of the software. We should not use IntAuc here since this is defined
  ! in ee_cfi_datatypes in directory and that would give a circular dependency
  ! between the support and DataStructures directories, which cannot
  ! easily be compiled.

  ! NOTE2: besides the cloud/no_cloud classification
  ! the measurements may also be split in 2 sets if there
  ! is for example a strong wind shear (or other phenomenon that 
  ! would justify it). So multiple observations with the same
  ! observation type are possible. Therefore nr_of_obs_types
  ! is not identical to the max nr. of allowed observations!

  real(r8_), parameter :: pi=3.14159265358979323846264338328_r8_
  real(r8_), parameter :: lambda=355e-9_r8_      ! Laser wavelength [m]
  real(r8_)            :: kwvn=2._r8_*pi/lambda  ! Laser wavenumber [m^-1]
  real(r8_), parameter :: kb=1.3806503e-23_r8_   ! Boltzmann constant [J/K]
  real(r8_), parameter :: M=28.86e-3_r8_/6.022e23_r8_ 
                          ! Mass of an air molecule [kg]

  ! JK: these are table properties in the RayleighBrillouin code
  ! so they should be moved into the RayleighBrillouinProcessing.F90
  ! module I think. Paul what do you think?
  !Paul: agreed -- including these in the RBC product documentation
  !(01/26/2006). When we code the RBC, will move these in the RBC_ADS.
  ! JK: most of these have been phased out now.
  ! Note however, that df is still used by tentispectrum.F90
  ! Also note that calib_grid.F90 and calib_tenti.F90 which are currently
  ! not being compiled by our make system, use all of these.
  ! Paul, do you think these constants should remain here, or could you
  ! agree to have a local RbcConstants module in the 
  ! RayleighBrillouinProcessing directory defining these?

!  real(r8_), parameter :: FSR=10.95e9_r8_ ! Fabry-Perot Free Spectral Range [Hz]
!  real(r8_), parameter :: USR=1.5e9_r8_ ! Useful Spectral Range [Hz]
  real(r8_), parameter :: df=25.e6_r8_  ! frequency step
!  real(r8_), parameter :: dT=1._r8_     ! temperature step
!  real(r8_), parameter :: Tmin=170._r8_,Tmax=330._r8_ ! min/max temperatures
!  real(r8_), parameter :: dP=50._r8_ ! pressure step
!  real(r8_), parameter :: Pmin=10._r8_,Pmax=1100._r8_ ! min/max pressures
!  real(r8_), parameter :: dRR=0.01_r8_ ! response step

  real(r8_), parameter :: c=299792458._r8_ ! speed of light
  !  #]
  !---------------------------------------------------
END Module AeolusConstants
