MODULE ARRAYTOOLS
  !  #[ documentation
  !
  ! This module contains a handful of functions 
  ! to create and manipulate arrays
  !
  !    Jun 2005 P. Poli     Original version
  !    Aug 2005 J. de Kloe  adapted to F90 standard
  ! 12-May-2006 D. Tan      Null use of ParKind1 module
  ! 25-Jul-2006 J. de Kloe  added some remarks in LUDCMP
  ! 16-Jan-2008 J. de Kloe  phase out integer kind i_ 
  !
  !  #]
  !  #[ modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE Numerics, only: r8_
  !  #]
  !  #[ variables
  implicit none
  !  #]
  !  #[ interface
!  INTERFACE
! ...
!     The explicit interface is already defined when using a module,
!     and specifying it by hand is not needed anymore.
!     Specifying the interface is only needed when overloading several
!     routines to a single name, or when specifying an operator.
!     Specifying them with the same name again is against the F90 standard
!     and my default compiler (g95) refuses to compile the module
!     with this interface in it.
!     Therefore I removed the interface statements (JK).
! ...
!  END INTERFACE
  !  #]
CONTAINS
  !--------
  FUNCTION FARRAY1(mini,step,n) result(array_out)
    !  #[ Create a real*8 array of size n, and fill it
    ! with ascending values, starting with 0.,increasing by "step"

    ! the next 2 lines are not needed here, because they are already
    ! present in the module header, but they also don't hurt
    USE Numerics, only: r8_
    implicit none

    ! I/O variables
    real (r8_),                  intent (in)  :: mini, step
    integer ,                    intent (in)  :: n
    real (r8_),    dimension(n)               :: array_out
    
    ! local variables
    integer :: i
    integer, dimension(:), allocatable :: tabi
    
    ! Why this (somewhat) complicated trick with allocate?
    ! Why not just say, (which saves a lot of memory for large arrays):
    ! do i=1,n
    !   array_out(i)=step*(i-1)+mini
    ! end
    ! if there is an anvantage in the efficiency of this
    ! calculation by using the allocated array, a little 
    ! explanation would be in place here.

    allocate(tabi(n))
    do i=1,n
       tabi(i)=i-1
    enddo
    array_out(1:n)=step*tabi(1:n)
    array_out(1:n)=array_out(1:n)+mini
    deallocate(tabi)
    
  END FUNCTION FARRAY1
  !  #]
  FUNCTION FCALCN (miny,stepy,maxy) result (n)
    !  #[ Calculate the size of a matlab-like array [miny:stepy:maxy]
    !
    !  15/06/2005  Paul Poli    Original code

    USE Numerics, only: r8_
    implicit none
    
    ! I/O variables
    real(r8_), intent(in) :: miny,stepy,maxy
    integer               :: n
    
    ! local variables
    ! none
    
    n=floor((maxy-miny)/stepy)
    n=n+1
    
  END FUNCTION FCALCN
  !  #]
  FUNCTION FCUMPROD_R(n,tabin) result(tabout)
    !  #[ Returns the cumulative product of a real*8 array
    ! History/modifications
    ! 09/06/2005   Paul Poli, METEO FRANCE    Original code

    USE Numerics, only: r8_
    implicit none

    ! I/O variables
    integer,                 intent (in) :: n
    real(r8_), dimension(n), intent (in) :: tabin
    real(r8_), dimension(n)              :: tabout
    
    ! local variables
    integer :: i
    
    tabout(1)=tabin(1)
    if (n>1) then
       do i=2,n
          tabout(i)=tabout(i-1)*tabin(i)
       enddo
    endif
    
  END FUNCTION FCUMPROD_R
  !  #]
  FUNCTION FCUMPROD_C(n,tabin) result(tabout)
    !  #[ Returns the cumulative product of a complex array
    ! History/modifications
    ! 09/06/2005   Paul Poli, METEO FRANCE    Original code
    
    USE Numerics, only: r8_
    implicit none
    
    ! I/O variables
    integer,                    intent (in) :: n
    complex(r8_), dimension(n), intent (in) :: tabin
    complex(r8_), dimension(n)              :: tabout
    
    ! local variables
    integer :: i
    
    tabout(1)=tabin(1)
    if (n>1) then
       do i=2,n
          tabout(i)=tabout(i-1)*tabin(i)
       enddo
    endif
    
  END FUNCTION FCUMPROD_C
  !  #]
  FUNCTION FDIFF_R(n,tabin) result(tabout)
    !  #[ Return the cumulative diff of the elements in a real*8 array
    ! History/modifications
    ! 09/06/2005   Paul Poli, METEO FRANCE    Original code
    
    USE Numerics, only: r8_    
    implicit none
    
    ! I/O variables
    integer,                   intent (in) :: n
    real(r8_), dimension(n+1), intent (in) :: tabin
    real(r8_), dimension(n)                :: tabout
    
    ! local variables
    integer :: i
    
    do i=1,n
       tabout(i)=tabin(i+1)-tabin(i)
    enddo
    
  END FUNCTION FDIFF_R
  !  #]
  FUNCTION FDIFF_C(n,tabin) result(tabout)
    !  #[ Return the cumulative diff of the elements in a complex array
    ! History/modifications
    ! 09/06/2005   Paul Poli, METEO FRANCE    Original code

    USE Numerics, only: r8_
    implicit none
    
    ! I/O variables
    integer,                      intent (in) :: n
    complex(r8_), dimension(n+1), intent (in) :: tabin
    complex(r8_), dimension(n)                :: tabout
    
    ! local variables
    integer :: i

    do i=1,n
       tabout(i)=tabin(i+1)-tabin(i)
    enddo

  END FUNCTION FDIFF_C
  !  #]
  SUBROUTINE LUBKSB(A,N,NP,INDX,B)
    !  #[ Solves set of linear equations Ax=b 
    !
    ! INPUT
    !			a    : LU decomposition of A
    !			indx : permutation vector from ludcmp
    !			b    : RHS vector 
    !			n    : size of b
    !		 	np   : dimension of square matrix A		
    !
    ! OUTPUT
    !			b    : returns solution vector x 
    !
    ! See numerical recipes
    
    USE Numerics, only: r8_
    implicit none

    integer,                     intent(in)  :: n,np
    real(r8_), dimension(np,np), intent(in)  :: A
    integer,   dimension(np),    intent(in)  :: INDX
    real(r8_), dimension(np),    intent(inout) :: B

    ! local variables
    integer   :: II,I,LL,J
    real(r8_) :: SUM

    II=0
    DO I=1,N
       LL=INDX(I)
       SUM=B(LL)
       B(LL)=B(I)
       IF (II.NE.0)THEN
          DO J=II,I-1
             SUM=SUM-A(I,J)*B(J)
          END DO
       ELSE IF (SUM.NE.0.) THEN
          II=I
       ENDIF
       B(I)=SUM
    END DO

    DO I=N,1,-1
       SUM=B(I)
       IF(I.LT.N)THEN
          DO J=I+1,N
             SUM=SUM-A(I,J)*B(J)
          END DO
       ENDIF
       B(I)=SUM/A(I,I)
    END DO

    RETURN

  END SUBROUTINE LUBKSB
  !  #]
  SUBROUTINE LUDCMP(A,N,NP,INDX,D)
    !  #[ Performs LU decomposition
    ! INPUT
    !			a    : input matrix to decompose
    !			n    : size of 2D square matrix a
    !		 	np   : physical dimensions of a
    !
    ! OUTPUT
    !			indx : records row permutation
    !			d    : +/-1 depending on whether # row 
    !				interchanges even or odd
    !
    ! See numerical recipes
    
    USE Numerics, only:r8_
    implicit none

    integer,   intent(in)  ::  n, np
    real(r8_), intent(inout), dimension(1:np,1:np) ::  A
    integer,   intent(out),   dimension(1:np)      :: INDX
    real(r8_), intent(out) :: d

    integer,   parameter          :: NMAX=600
    real(r8_), parameter          :: TINY=1.0E-20_r8_
    integer                       :: I,J,K,IMAX
    real(r8_), dimension (1:NMAX) :: VV 
    real(r8_)                     :: SUM,DUM,AAMAX

    ! initialised to suppress uninitialised warning with g95
    ! see remarks in the code below.
    IMAX=-1

    D=1._r8_
    DO I=1,N
       AAMAX=0.
       DO J=1,N
          ! aamax is always 0 or larger !
          IF (ABS(A(I,J)).GT.AAMAX) AAMAX=ABS(A(I,J))
       END DO
       IF (AAMAX.EQ.0.) THEN
          PRINT *,'ludcmp: Singular matrix'
          return
       ENDIF
       ! aamax is at this point always positive
       ! so vv is always positive
       VV(I)=1._r8_/AAMAX
    END DO
    
    DO J=1,N
       
       IF (J.GT.1) THEN
          DO I=1,J-1
             SUM=A(I,J)
             IF (I.GT.1)THEN
                DO K=1,I-1
                   SUM=SUM-A(I,K)*A(K,J)
                END DO
                A(I,J)=SUM
             ENDIF
          END DO
       ENDIF
       
       AAMAX=0.
       DO I=J,N
          SUM=A(I,J)
          IF (J.GT.1)THEN
             DO K=1,J-1
                SUM=SUM-A(I,K)*A(K,J)
             END DO
             A(I,J)=SUM
          ENDIF
          ! if vv has only negative values IMAX is never set
          ! this triggers a compiler warning with g95.
          ! However, since we know vv is always positive at this point
          ! it is safe to initialise IMAX above to suppress this warning.
          DUM=VV(I)*ABS(SUM)
          IF (DUM.GE.AAMAX) THEN
             IMAX=I
             AAMAX=DUM
          ENDIF
       END DO
       
       ! WARNING: if (DUM.lt.AAMAX) in the above if-then
       ! then IMAX stays uninitialised, which may break down the code below !
       ! Paul: can you check if this can occur? JK.

       IF (J.NE.IMAX)THEN
          DO K=1,N
             DUM=A(IMAX,K)
             A(IMAX,K)=A(J,K)
             A(J,K)=DUM
          END DO
          D=-D
          VV(IMAX)=VV(J)
       ENDIF
       
       INDX(J)=IMAX
       IF(J.NE.N)THEN
          IF(A(J,J).EQ.0.)A(J,J)=TINY
          DUM=1._r8_/A(J,J)
          DO I=J+1,N
             A(I,J)=A(I,J)*DUM
          END DO
       ENDIF
       
    END DO
    
    IF(A(N,N).EQ.0.)A(N,N)=TINY
    RETURN
    
  END SUBROUTINE LUDCMP
  !  #]
  FUNCTION FINVERT (amatrix) result (amatinv)
    !  #[ Matrix inversion
    !
    !        amatinv = invert(amat)
    !     
    !   amat : 2D matrix to invert
    !   amatinv : 2D inverted matrix
    ! See numerical recipes

    USE Numerics, only:r8_
    implicit none

    ! I/O variables
    real(r8_), dimension(:,:), intent(inout) :: amatrix
    real(r8_), dimension(lbound(amatrix,1):ubound(amatrix,1), &
                         lbound(amatrix,2):ubound(amatrix,2)) :: amatinv
      
    ! local variables
    integer,   dimension (:),   allocatable :: indx
    real(r8_), dimension (:,:), allocatable :: work
    real(r8_)                               :: d
    integer                                 :: j,nsampl

    nsampl=size(amatrix,1)
    if (nsampl .ne. size(amatrix,2)) then
       print *,'can''t invert a non-square matrix'
       stop 1
    endif

    amatinv=0.
    allocate(indx(nsampl))
    allocate(work(nsampl,nsampl))
    work=amatrix
    do j = 0, nsampl-1
       amatinv(j+lbound(amatrix,1),j+lbound(amatrix,2))=1._r8_
    end do
    call ludcmp(amatrix, nsampl, nsampl, indx, d)
    do j = 0, nsampl-1
       call lubksb(amatrix, nsampl, nsampl, indx, & 
            amatinv(:,j+lbound(amatrix,2)))
    end do
    amatrix=work
    deallocate(indx)
    deallocate(work)

  END FUNCTION FINVERT
  !  #]
  FUNCTION FSIGN(x) result(y)
    !  #[ Function which returns the sign of the argument
    !
    ! 2005 Paul Poli   Original version

    ! isn't there a standard fortran function for this?

    USE Numerics, only: r8_
    implicit none
    
    ! I/O variables
    real(r8_), intent(in) :: x
    real(r8_)             :: y
    
    IF (x.gt.0._r8_) THEN
       y=1._r8_
    ELSE
       IF (x.lt.0._r8_) THEN
          y=-1._r8_
       ELSE
          y=0._r8_
       ENDIF
    ENDIF
       
  END FUNCTION FSIGN
  !  #]
  !--------
END MODULE ARRAYTOOLS

