MODULE Logging
  !  #[ Description
  !---------------------------------------------------
  ! a module to handle the production of log-messages
  ! which is one of the required interfaces of the Thin Layer.
  ! 
  ! See: ESA-ID-ACS-GS-001_PDS-GENERIC-IF-GUIDELINES_Iss_1.1_040220.pdf
  ! section 4.5, p.56-60 for specific details on this interfaces
  !---------------------------------------------------
  ! conventions:
  ! a log messages is a one line message terminated by a newline character
  ! and containing the following 8 elements (which may be of variable length),
  ! separated by a single space:
  ! 1->DateTime (format YYYY-MM-ddThh:mm:ss.nnnnnn)
  ! 2->Node Name (string without whitespaces)
  ! 3->Processor Name  (string without whitespaces)
  ! 4->Processor version (format nn.nn)
  ! 5->Process-ID (10 character integer enclosed by square brackets,
  !                format [nnnnnnnnnn])
  ! 6->Header Separator (a single colon character
  ! 7->Message Type, one of: D,I,P,W,E enclosed by square brackets
  ! 8->the actual message text
  ! Log messages are written to stdout, and for some types to stderr
  !
  ! Remark: the logmsg() routine will be frequently called when reporting
  !         a warning or an error. Therefore it is not very usefull to let
  !         it report an error_flag if something goes wrong inside it.
  !         In stead I choose to change the msg_type into an error
  !         in case something goes wrong inside this routine.
  !         (usually this indicates a programming error), and add some
  !         text describing the problem to the msg_text string
  !---------------------------------------------------
  !    Written by:  Jos de Kloe.
  !    created:     03-04-2006
  !
  ! Modifications:
  !     12-May-2006 D. Tan     Null use of ParKind1 module
  !     07-Aug-2006 J. de Kloe added init routine, to allow control by
  !                            the ThinLayer/JobOrder system
  !     27-Jul-2006 P. Poli    Added by default REALACC(6) to all logmsg 
  !                            outputs.
  !                            Optional input arguments realacc_precision 
  !                            and realacc_off to subroutine logmsg enable
  !                            to change precision or to turn off REALACC 
  !                            output.
  !     08-Aug-2006 D. Tan     Change default to no REALACC, allow realacc_on
  !                            or precision
  !     17-Aug-2006 P. Poli    Added parameter Log_Level_Default
  !     15-Feb-2007 J. de Kloe Added mechanism to abort and report fatal errors
  !                            from within functions that have no error_flag
  !     01-Mar-2007 J. de Kloe Added mechanism to export the log messages to
  !                            a private log file, independant of the ThinLayer
  !                            logging mechanism (which seems to be an 
  !                            additional new requirement for us?)
  !        Dec-2007 P. Poli    added GetFatalError() function
  !     16-Jan-2008 J. de Kloe phase out integer kind i_ 
  !
  !---------------------------------------------------
  !  #]
  !  #[ modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE ErrorHandler, only: no_error, error_programming, error_opening_file
  USE StringTools, only: space,contains_spaces,contains_crnl
  USE LunManager, only: stdout, stderr, get_lun, free_lun
  USE DateTimeMod, only: date_type, time_type, get_current_date_time,&
                         date2string_sep,time2string_sep_ms
  USE AeolusConstants, only: software_version,software_ProcName
  USE Compiler_Features, only: flush_aeolus
  USE c_support, only: get_hostname, get_process_id
  USE numerics, only: r8_
  !  #]
  !  #[ Variables and parameters
  IMPLICIT NONE

  ! types of log messages
  character(len=1), parameter :: log_debug    = "D"
  character(len=1), parameter :: log_info     = "I"
  character(len=1), parameter :: log_progress = "P"
  character(len=1), parameter :: log_warn     = "W"
  character(len=1), parameter :: log_error    = "E"

  ! possible log levels
  integer, parameter :: Log_Level_Debug    = 1
  integer, parameter :: Log_Level_Info     = 2
  integer, parameter :: Log_Level_Progress = 3
  integer, parameter :: Log_Level_Warning  = 4
  integer, parameter :: Log_Level_Error    = 5
  integer, parameter :: Log_Level_Noop     = 99
  integer, parameter :: Log_Level_Default  = Log_Level_Debug

  integer, save :: Stdout_Log_Level   = Log_Level_Info
  integer, save :: Stderr_Log_Level   = Log_Level_Noop

  ! To enable signalling of serious errors/showstoppers from within
  ! functions that cannot return an error_flag, the following flag may be
  ! set to true. The program can check at a later time if this did occur
  ! and take appropriate actions to terminate processing.
  ! This should only be used with fatal-errors.
  ! To activate it, use the optional flag SignalFatalError
  ! in the logmsg() call
  logical, save :: FatalErrorOccurred              = .false.
  integer, save :: Error_That_Caused_This_Fatality = no_error

  ! some variables to enable using a private logfile
  logical,            save :: UseOptionalLogFile  = .false.
  integer,            save :: OptionalLogFileUnit = -1
  character(len=256), save :: OptionalLogFile
  !  #]
contains
  !--------------------------------
  subroutine Init_Logging(Stdout_Log_Level_txt,Stderr_Log_Level_txt,&
                          error_flag,OptionalLogFile_in)
    !  #[
    character(len=*), intent(in)  :: Stdout_Log_Level_txt
    character(len=*), intent(in)  :: Stderr_Log_Level_txt
    integer,          intent(out) :: error_flag
    character(len=*), optional, intent(in) :: OptionalLogFile_in

    error_flag = no_error
    FatalErrorOccurred = .false.

    IF (present(OptionalLogFile_in)) THEN
       UseOptionalLogFile = .true.
       OptionalLogFile = OptionalLogFile_in
       OptionalLogFileUnit = get_lun()
       open(unit=OptionalLogFileUnit,file=OptionalLogFile,status="replace",&
            action="write",form="FORMATTED",ERR=999)
    END IF

    ! define what log messages should be printed on what channel
    select case (trim(Stdout_Log_Level_txt))
    case("DEBUG")
       Stdout_Log_Level = Log_Level_Debug
    case("INFO")
       Stdout_Log_Level = Log_Level_Info
    case("PROGRESS")
       Stdout_Log_Level = Log_Level_Progress
    case("WARNING")
       Stdout_Log_Level = Log_Level_Warning
    case("ERROR")
       Stdout_Log_Level = Log_Level_Error
    case("NOOP")
       Stdout_Log_Level = Log_Level_Noop
    case default
       ! this code should never be reached
       call logmsg(log_error,"ERROR in Init_Logging():"//&
             " unknown Stdout log level: "//trim(Stdout_Log_Level_txt)//&
             " This should have been handled in the"//&
             " readjoborderdata() routine.")
       error_flag = error_programming
       return
    end select
    call logmsg(log_info,"In Init_Logging():"//&
             " Stdout log level set to: "//trim(Stdout_Log_Level_txt))

    select case (trim(Stderr_Log_Level_txt))
    case("DEBUG")
       Stderr_Log_Level = Log_Level_Debug
    case("INFO")
       Stderr_Log_Level = Log_Level_Info
    case("PROGRESS")
       Stderr_Log_Level = Log_Level_Progress
    case("WARNING")
       Stderr_Log_Level = Log_Level_Warning
    case("ERROR")
       Stderr_Log_Level = Log_Level_Error
    case("NOOP")
       Stderr_Log_Level = Log_Level_Noop
    case default
       ! this code should never be reached
       call logmsg(log_error,"ERROR in Init_Logging():"//&
             " unknown Stderr log level: "//trim(Stderr_Log_Level_txt)//&
             " This should have been handled in the"//&
             " readjoborderdata() routine.")
       error_flag = error_programming
       return
    end select
    call logmsg(log_info,"In Init_Logging():"//&
             " Stderr log level set to: "//&
             trim(Stderr_Log_Level_txt))

    return

999 call logmsg(log_error,"ERROR in Init_Logging():"//&
             " opening optional logfile failed: "//trim(OptionalLogFile))
    error_flag = error_opening_file
    return

  end subroutine Init_Logging
    !  #]
  subroutine Close_Logging()
    !  #[
    IF (UseOptionalLogFile) THEN
       close(unit=OptionalLogFileUnit)
       call free_lun(OptionalLogFileUnit)
    END IF

  end subroutine Close_Logging
    !  #]
  subroutine logmsg(msg_type,msg_text,realacc_on,realacc_precision,&
                    SignalFatalError,FatalError)
    !  #[
    character(len=1),     intent(in) :: msg_type
    character(len=*),     intent(in) :: msg_text
    logical,    optional, intent(in) :: realacc_on
    integer,    optional, intent(in) :: realacc_precision
    logical,    optional, intent(in) :: SignalFatalError
    integer,    optional, intent(in) :: FatalError

    ! local variables
    integer                          :: pos, ios, log_level
    character(len=1)                 :: msg_type_copy
    character(len=len(msg_text)+256) :: msg_text_copy
    character(len=len(msg_text)+512) :: total_log_msg
    ! max expected length is: 26+1+20+1+20+1+5+1+12+1+3+1+len(msg_text)+256
    ! 92+len(msg_txt)+256  (+256 in case of added error texts)
    ! so 512 extra bytes should be a safe margin
    logical                          :: do_realacc
    integer                          :: do_realacc_precision
    character(len=2)                 :: str_realacc_precision

    type(date_type)   :: date_current
    type(time_type)   :: time_current
    character(len=26) :: DateTimeString
    character(len=10) :: datestring ! 10 chars, format: yyyy-mm-dd  
    character(len=15) :: timestring ! 15 chars, format: hh:mm:ss.nnnnnn
    logical, save :: first_call = .true.
    character(len=20) :: txt
    character(len=25), save :: NodeName
    character(len=20), save :: ProcName
    character(len=5),  save :: ProcVersion
    integer                 :: ProcID_nr
    character(len=12), save :: ProcID
    real(r8_)               :: version_nr

    ! is REALACC required ?
    do_realacc=.false.
    IF (present(realacc_on)) do_realacc=realacc_on
    do_realacc_precision=6
    IF (present(realacc_precision)) THEN
       do_realacc=.true.
       IF (realacc_precision .ge. 0 .and. &
         & realacc_precision .le. 20) THEN
          do_realacc_precision=realacc_precision
       ELSE
          write(stdout,*,err=998) &
          "logmsg: WARNING: realacc_precision argument out of bounds (", &
          realacc_precision,"). Setting to ",do_realacc_precision
       ENDIF
    ENDIF
    str_realacc_precision="  "
    IF (do_realacc_precision .ge. 10) THEN
       write(str_realacc_precision,"(I2)") do_realacc_precision
    ELSE
       write(str_realacc_precision(1:1),"(I1)") do_realacc_precision
    ENDIF

    ! copy the type, to be able to switch it to "error"
    ! in case something goes wrong inside this routine
    msg_type_copy = msg_type

    ! copy the msg_text to be able to add some text to it in case
    ! something goes wrong inside this routine
    msg_text_copy = msg_text

    ! get the current date and time and construct a DateTimeString from them
    ! remark: the used stringformat for datetime is yet a little different
    ! from the already defined XML adn KVT formats, so I can't reuse 
    ! those here
    call get_current_date_time(date_current,time_current)
    datestring = date2string_sep(   date_current,"-")
    timestring = time2string_sep_ms(time_current,":")

    DateTimeString = trim(datestring)//'T'//trim(timestring)
    
    ! set some variables at the first call only to save time
    IF (first_call) THEN
       first_call = .false.
       ! get the nodename
       NodeName = get_hostname()
       ! make sure it contains no spaces !
       IF (contains_spaces(NodeName)) THEN
          msg_text_copy = trim(msg_text)//&
               " ERROR in logmsg(): "//&
               "NodeName ["//trim(NodeName)//"] contains one or more spaces"//&
               "this is not allowed !!!"
          msg_type_copy = log_error
       END IF

       ! get the procname
       ProcName = software_ProcName
       ! make sure it contains no spaces !
       IF (contains_spaces(ProcName)) THEN 
          msg_text_copy = trim(msg_text)//&
               " ERROR in logmsg(): "//&
               "ProcName ["//trim(ProcName)//"] contains one or more spaces"//&
               "this is not allowed !!!"
          msg_type_copy = log_error
       END IF
       
       ! get the Proc version
       ! convert from "0.2" to "00.20"
       txt = software_version
       read(txt,*,iostat=ios) version_nr
       IF (ios .ne.0) THEN
          msg_text_copy = trim(msg_text)//" ERROR in logmsg(): "//&
               "conversion of software_version to integer string failed ..."
          msg_type_copy = log_error
       END IF
       write(ProcVersion,"(f5.2)",iostat=ios) version_nr
       IF (ios .ne.0) THEN
          msg_text_copy = trim(msg_text)//" ERROR in logmsg(): "//&
               "conversion of version_nr to string failed ..."
          msg_type_copy = log_error
       END IF

       ! make sure the first char is a zero if it is not filled
       if (ProcVersion(1:1) .eq. space) ProcVersion(1:1)="0"

       ! get the PID
       ProcID_nr = get_process_id()
       write(ProcID,"('[',i10.10,']')",iostat=ios) ProcID_nr
       IF (ios .ne.0) THEN
          msg_text_copy = trim(msg_text)//" ERROR in logmsg(): "//&
               "conversion of PID to string failed ..."
          msg_type_copy = log_error
       END IF
    END IF

    ! verify that msg_type is one of the allowed types:
    pos = verify(msg_type,"DIPWE")
    IF (pos .ne. 0) THEN
       msg_text_copy = trim(msg_text)//" ERROR in logmsg(): "//&
            "msg_type ["//msg_type//"] is not one of the allowed types "//&
            "[DIPWE], this is a programming error ..."
       msg_type_copy = log_error
    END IF

    ! verify that there are no illegal characters in the msg string 
    ! (like return and newline characters ...)
    IF (contains_crnl(msg_text_copy)) THEN
       msg_text_copy = trim(msg_text)//" ERROR in logmsg(): "//&
            "msg_text contains one or more "//&
            "carriage-return or newline characters, this is not allowed !!!"
       msg_type_copy = log_error
    END IF

    ! construct the message line
    write(total_log_msg,"(a)",err=997) &
         DateTimeString//space//&
         trim(Nodename)//space//&
         trim(ProcName)//space//&
         trim(ProcVersion)//space//&
         trim(ProcID)//& ! no space here !!!!!
         ":"//space//&
         "["//msg_type_copy//"]"//space//&
         trim(adjustl(msg_text_copy))

    ! there should be no space after the ProcID field.
    ! see email of Christophe Caspar, 3-apr-2006

    ! Poli: set log_level to default to prevent
    ! "warning: 'log_level' may be used uninitialized in this function"
    ! with g95

    log_level = Log_Level_Default

    select case (msg_type_copy)
    case(log_debug)
       log_level = Log_Level_Debug
    case(log_info)
       log_level = Log_Level_Info
    case(log_progress)
       log_level = Log_Level_Progress
    case(log_warn)
       log_level = Log_Level_Warning
    case(log_error)
       log_level = Log_Level_Error
       IF (present(SignalFatalError)) THEN
          IF (SignalFatalError) THEN
             FatalErrorOccurred = .true.
             IF (present(FatalError)) THEN
                Error_That_Caused_This_Fatality = FatalError
             ELSE
                Error_That_Caused_This_Fatality = error_programming
             END IF
          END IF
       End IF
    case default
       ! this code should never be reached
       total_log_msg = trim(total_log_msg)//" ERROR in logmsg():"//&
            " unknown msg_type: "//trim(msg_type_copy)
       log_level = Log_Level_Default
       ! no return here otherwise the msg is not printed !
    end select

    IF (log_level .ge. Stdout_Log_Level) THEN
       ! write the log message to stdout
       IF (do_realacc) &
          write(stdout,*,err=998) "REALACC("//trim(str_realacc_precision)//")"
       write(stdout,"(a)",err=998) trim(total_log_msg)
       IF (do_realacc) &
          write(stdout,*,err=998) "ENDREALACC"
       ! remind that flushing the output is a thinLayer requirement!
       ! (see section 4.5.4 of the inferface guidelines document)
       call flush_aeolus(stdout)
    END IF

    IF (log_level .ge. Stderr_Log_Level) THEN
       ! write the log message to stderr
       IF (do_realacc) &
          write(stderr,*,err=998) "REALACC("//trim(str_realacc_precision)//")"
       write(stderr,"(a)",err=999) trim(total_log_msg)
       IF (do_realacc) &
          write(stderr,*,err=998) "ENDREALACC"
       ! remind that flushing the output is a thinLayer requirement!
       ! (see section 4.5.4 of the inferface guidelines document)
       call flush_aeolus(stderr)
    END IF

    IF (UseOptionalLogFile) THEN
       ! Write the log message to our private optional logfile
       ! Note that there are no ThinLayer requirements defined for
       ! this one, so for now just use the same loglevel as is
       ! used for stdout.
       IF (do_realacc) &
          write(OptionalLogFileUnit,*,err=996) &
               "REALACC("//trim(str_realacc_precision)//")"
       write(OptionalLogFileUnit,"(a)",err=996) trim(total_log_msg)
       IF (do_realacc) &
          write(OptionalLogFileUnit,*,err=996) "ENDREALACC"
    END IF

    return

996 print *,"ERROR in logmsg(): could not write to the optional logfile "//&
         "named: "//trim(OptionalLogFile)
    return
997 print *,"ERROR in logmsg(): could not write to the string total_log_msg"
    return
998 print *,"ERROR in logmsg(): could not write to stdout !!!"
    ! so does it make sense to try and print a warning in this case?
    return
999 print *,"ERROR in logmsg(): could not write to stderr !!!"
    return

  end subroutine logmsg
    !  #]
  function CheckIfFatalErrorOccurred() result(FatalErrorOcc)
    !  #[
    logical :: FatalErrorOcc ! result
    FatalErrorOcc = FatalErrorOccurred
    return
  end function CheckIfFatalErrorOccurred
    !  #]
  function GetFatalError() result(error)
    !  #[
    integer :: error ! result
    error=Error_That_Caused_This_Fatality
    return
  end function GetFatalError
    !  #]
  !--------------------------------
end MODULE Logging
