module L2B_AuxPar_DataStructure
  !  #[ Description
  !     a module to define the L2B AuxPar GADS datastructure
  !     used by the L2B processor
  !
  !     created: 05-04-2006, by: J. de Kloe
  !
  !     Modifications
  !     10-Apr-2006 J.de Kloe  implemented WVM_Params reading
  !     24-Apr-2006 J. de Kloe fully implemented now
  !     03-May-2006 D. Tan     Remove Mie_Core_Use_Threshold 
  !                            (cf SNR_Threshold).
  !     12-May-2006 D. Tan     Null use of ParKind1 module
  !     01-Nov-2006 J. de Kloe added AMD_Matchup_Params_type
  !     03-Nov-2006 P. Poli    upgrade to L1BP v1.04
  !     05-Mar-2007 J. de Kloe added Rangebin_Mismatch_Tolerance and
  !                            implemented CLM_Matchup_Params handling
  !     07-Aug-2007 D. Tan     Fix SPR#20 (extra underscores)
  !     11-Dec-2007 D. Tan     Comment out check_length_xml_item twice
  !     17-Dec-2007 J. de Kloe implemented unit checking for Mie Core pars
  !     16-Jan-2008 J. de Kloe phase out integer kind i_ 
  !
  !  #]
  !  #[ modules used
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE FixedHeader, only: &
       FH_type, Init_FH, Fill_FH_items, AsciiWrite_FH,&
       FH_maxlen_file_description => maxlen_file_description,&
       FH_maxlen_notes            => maxlen_notes,&
       FH_maxlen_mission          => maxlen_mission,&
       FH_maxlen_file_class       => maxlen_file_class,&
       FH_maxlen_file_version     => maxlen_file_version,&
       FH_maxlen_system           => maxlen_system,&
       FH_maxlen_creator          => maxlen_creator,&
       FH_maxlen_creator_version  => maxlen_creator_version

  USE MainProductHeader, only: &
       MPH_type, Init_MPH, Fill_MPH_items, AsciiWrite_MPH, &
       MPH_maxlen_ref_doc      => maxlen_ref_doc, &
       MPH_maxlen_software_ver => maxlen_software_ver,&
       MPH_maxlen_proc_center  => maxlen_proc_center

  USE L2B_AuxPar_SPH, only: L2B_AuxPar_SPH_type, AsciiWrite_L2B_AuxPar_SPH,&
                            Delete_L2B_AuxPar_SPH
  USE EE_CFI_Datatypes, only: &
       IntAuc, &
       IntAs,  missing_indicator_IntAs,  SetIntAs_or_Missing,  maxlen_IntAs, &
       IntAus, missing_indicator_IntAus, SetIntAus_or_Missing, maxlen_IntAus,&
       IntAl,  missing_indicator_IntAl,  SetIntAl_or_missing,  maxlen_IntAl, &
       IntAul, missing_indicator_IntAul, SetIntAul_or_missing, maxlen_IntAul,&
       FAdoxy, missing_indicator_FAdoxy, SetFAdoFull_or_Missing, &
       GetIntAs, GetIntAus, GetIntAl, GetIntAul, GetFAdoxy
  USE Logging, only: logmsg,log_debug,log_info,log_progress,log_warn,log_error
  USE VerbosityControl, only: VerbosityError, VerbosityWarn, &
                              VerbosityReport, print_verbosity_setting
  USE numerics, only: r8_
  USE ErrorHandler, only: no_error, error_programming, error_allocate, &
                          error_writing_file, error_reading_file
  USE LunManager, only: fileunit_stdout
  USE XML_module, only: max_str_len, xml_node_type, xml_node_get_string,&
                        check_length_xml_item, xml_tree_read_string_attribute
  USE StringTools, only: to_lowercase
  USE ProfileInterpolate, only: profile_type, init_profile, print_profile, &
                                interp_type_linear, extrap_type_linear
  !  #]
  !  #[ Parameter and data type definitions
  implicit none
  !  #[ FH_Default_Fields
  TYPE FH_Default_Fields_Type
     character(len=FH_maxlen_file_description) :: File_Description
     character(len=FH_maxlen_notes)            :: Notes
     character(len=FH_maxlen_mission)          :: Mission
     !character(len=maxlen_mission_id)         :: Mission_Id
     ! this item seems unused, or is this the file_type ???
     character(len=FH_maxlen_file_class)       :: File_Class
     character(len=FH_maxlen_file_version)     :: File_Version
     character(len=FH_maxlen_system)           :: System
     character(len=FH_maxlen_creator)          :: Creator
     character(len=FH_maxlen_creator_version)  :: Creator_Version
  END TYPE FH_Default_Fields_Type
  !  #]
  !  #[ MPH_Default_Fields
  TYPE MPH_Default_Fields_Type
     character(len=MPH_maxlen_ref_doc)      :: Ref_Doc
     character(len=MPH_maxlen_software_ver) :: Software_Ver
     character(len=MPH_maxlen_proc_center)  :: Proc_Center
  END TYPE MPH_Default_Fields_Type
  !  #]
  !  #[ WVM Params and subtypes

  TYPE Threshold_Type
     real(FAdoxy) :: Threshold_Value ! [none] or [m-1]
     real(FAdoxy) :: Altitude ! [km]
  END TYPE Threshold_Type

  !character(len=*), parameter :: unit_Thresh_Value_Backsc_Ratio = "none"
  character(len=*), parameter :: unit_Thresh_Value_Extinction   = "m-1"
  character(len=*), parameter :: unit_Altitude                  = "km"

  TYPE Classification_Params_type
     integer(IntAuc) :: Classification_Type_Mie
     integer(IntAuc) :: Classification_Type_Rayleigh
     TYPE(Threshold_Type), dimension(:), pointer :: Threshold_BackscatterRatio
     TYPE(Threshold_Type), dimension(:), pointer :: Threshold_Extinction
     ! spare 1000 bytes
  END TYPE Classification_Params_type

  ! allowed enum values for Classification_Type_Mie/Rayleigh
  ! NOTE: these are also defined in the module DataStructures/L2B_Proc_Settings
  !       but should be moved here, if/once we phase out the commandline 
  !       switches
                                               
  integer(IntAuc), parameter :: Class_undefined         = 0
  ! used for initialisation only
  integer(IntAuc), parameter :: Class_No_Clouds         = 1
  ! dont actually do classification, just assume clear sky 
  integer(IntAuc), parameter :: Class_Copy_E2S_Input    = 2
  ! copy the input that was used for the E2S
  integer(IntAuc), parameter :: Class_Ext_Threshold     = 3  
  ! set a threshold on extinction derived from Rayleigh
  integer(IntAuc), parameter :: Class_Cal_Mie_Threshold = 4
  ! calibrate Mie using extinction and set a threshold on it
  integer(IntAuc), parameter :: Class_Rayleigh_Slope    = 5
  ! use a threshold on the slope of the Rayleigh profile
  integer(IntAuc), parameter :: Class_Mie_Slope         = 6
  ! use a threshold on the slope of the Mie profile
  integer(IntAuc), parameter :: Class_Backscat_Ratio    = 7
  ! set a threshold on the backscatter ratio
  integer(IntAuc), parameter :: Class_Adjacent_Mie_Backscatter  = 8
  integer(IntAuc), parameter :: Class_Adjacent_Rayl_Backscatter = 9
  integer(IntAuc), parameter :: Class_Tau               = 10
  ! two possible other types (see email David Tan 11-11-2005)
  ! without need for an extinction value
  ! ==>using the ratio (and not the difference) of backscatter in 
  !    adjacent height bins.     

  TYPE Optical_Properties_Params_type
     integer(IntAuc) :: ScatRatio_Method
     integer(IntAuc) :: ScatRatio_Method2
     ! spare 1000 bytes
  END TYPE Optical_Properties_Params_type

  ! allowed enum values for ScatRatio_Method and ScatRatio_Method2
  ! NOTE: these are also defined in the module DataStructures/L2B_Proc_Settings
  !       but should be removed there, if/once we phase out the commandline 
  !       switches
  ! missing_indicator, used for initialisation
  integer(IntAuc), parameter :: Scat_Ratio_Undefined     = 0
  ! get the value from the L1B input file (available for Mie rangebins only)
  integer(IntAuc), parameter :: Scat_Ratio_from_L1B      = 1
  ! calc. the value by comparing Mie and Rayleigh US
  integer(IntAuc), parameter :: Scat_Ratio_from_MieRayl  = 2
  ! calc. the value by assuming a backscatter to extinction ratio for aerosols
  integer(IntAuc), parameter :: Scat_Ratio_from_RaylOnly = 3
  ! dont use the fall-back option
  integer(IntAuc), parameter :: Scat_Ratio_Dont_Use      = 99

  TYPE Error_Quantifier_Params_type
     integer(IntAuc) :: ErrorQuantMethod_Mie
     integer(IntAuc) :: ErrorQuantMethod_Rayleigh
     ! spare 1000 bytes
  END TYPE Error_Quantifier_Params_type

  ! allowed enum values for ErrorQuantMethod_Mie
  ! NOTE: these are also defined in the module DataStructures/L2B_Proc_Settings
  !       but should be removed there, if/once we phase out the commandline 
  !       switches
  integer(IntAuc), parameter :: ErrorQuantMethod_Mie_undefined  = 0
  integer(IntAuc), parameter :: ErrorQuantMethod_Mie_1Bweighted = 1
  integer(IntAuc), parameter :: ErrorQuantMethod_Mie_core_sens  = 2
  !                             1234567890123456789012345678901

  ! allowed enum values for ErrorQuantMethod_Rayleigh
  integer(IntAuc), parameter :: ErrorQuantMethod_Ray_undefined  = 0
  integer(IntAuc), parameter :: ErrorQuantMethod_Ray_1Bweighted = 1
  integer(IntAuc), parameter :: ErrorQuantMethod_Ray_iliad_sens = 2

  integer, parameter :: WVM_maxlen_File_Type      = 49 !??? is this correct
  integer, parameter :: WVM_maxlen_Sph_Descriptor = 59 !??? is this correct

  TYPE Mie_PCD_Params_type
     real(FAdoxy)    :: Alpha_Correction
     integer(IntAs)  :: Summation_Index
  END TYPE Mie_PCD_Params_type

  TYPE Mie_Core_Algorithm_Params_type
     real(FAdoxy)    :: SNR_Threshold ! = 8. ! [dB] !
     real(FAdoxy)    :: Start_FWHM
     real(FAdoxy)    :: Residual_Error_Threshold
     integer(IntAs)  :: Max_Iterations_Lorentz_Fit
     real(FAdoxy)    :: FWHM_Upper_Threshold
     real(FAdoxy)    :: FWHM_Lower_Threshold
     real(FAdoxy)    :: Peak_Height_Upper_Threshold
     real(FAdoxy)    :: Peak_Height_Lower_Threshold
     real(FAdoxy)    :: Peak_Location_Threshold
     real(FAdoxy)    :: Nonlinear_Optimization_Thresh
     integer(IntAs)  :: Max_Iter_Nonlinear_Optimization
     !                  1234567890123456789012345678901
     integer(IntAs)  :: Num_Spectral_Sub_Samples
  END TYPE Mie_Core_Algorithm_Params_type

  ! unit definitions are used in 2 places, to check the xml file
  ! when reading it, and to print in the AsciiWrite routines.
  ! So it is probably a good idea to specify them in a single
  ! place like this, to ensure consistency through the code.
! character(len=*), parameter :: unit_SNR_Threshold = "dB" ! currently not used
  character(len=*), parameter :: unit_Start_FWHM                = "ACCD pixel"
  character(len=*), parameter :: unit_Residual_Error_Threshold  = "A.U."
  character(len=*), parameter :: unit_Upper_Threshold           = "ACCD pixel"
  character(len=*), parameter :: unit_Lower_Threshold           = "ACCD pixel"
  character(len=*), parameter :: unit_Peak_Height_Upper_Thresh  = "ACCD counts"
  character(len=*), parameter :: unit_Peak_Height_Lower_Thresh  = "ACCD counts"
  character(len=*), parameter :: unit_Peak_Location_Threshold   = "ACCD pixel"
  character(len=*), parameter :: unit_Nonlinear_Opt_Threshold   = "A.U."
  character(len=*), parameter :: unit_MaxIter_Nonlin_Opt_Thresh = "A.U."
  character(len=*), parameter :: unit_Num_Spectral_Sub_Samples  = "A.U."

  TYPE Corrupt_Data_Detect_Params_type
     integer(IntAus) :: Max_Signal_Derivative      ! [PixelLevel]
     integer(IntAus) :: Pixel_Saturation_Threshold ! [PixelLevel]
  END TYPE Corrupt_Data_Detect_Params_type

  character(len=*), parameter :: unit_Max_Signal_Derivative   = "PixelLevel"
  character(len=*), parameter :: unit_Pixel_Saturation_Thresh = "PixelLevel"

  TYPE Common_Processing_Params_type
     TYPE(Mie_PCD_Params_type)             :: Mie_PCD_Params
     TYPE(Mie_Core_Algorithm_Params_type)  :: Mie_Core_Algorithm_Params
     TYPE(Corrupt_Data_Detect_Params_type) :: Corrupt_Data_Detection_Params
     !    1234567890123456789012345678901
  END TYPE Common_Processing_Params_type

  TYPE RBC_Algorithm_Params_type
     logical :: Do_Mie_Decontamination
     ! spare 100 bytes
  END TYPE RBC_Algorithm_Params_type

  TYPE AMD_Matchup_Params_type
     integer(IntAuc) :: Matchup_Method
     integer(IntAul) :: Max_Allowed_Time_Diff ! in [s]
     integer(IntAul) :: Max_Allowed_Distance  ! in [km]
  END TYPE AMD_Matchup_Params_type

  character(len=*), parameter :: unit_Coll_Dist_Thresh = "km"
  character(len=*), parameter :: unit_Coll_Time_Diff   = "s"

  ! allowed enum values for Matchup_Method
  integer(IntAuc), parameter :: Matchup_Method_undefined  = 0
  integer(IntAuc), parameter :: Matchup_Dummy             = 1
  integer(IntAuc), parameter :: Matchup_Nearest_Neighbour = 2

  TYPE CLM_Matchup_Params_type
     integer(IntAuc) :: Matchup_Method
  END TYPE CLM_Matchup_Params_type

  ! allowed enum values for Matchup_Method
  integer(IntAuc), parameter :: CLM_Matchup_Method_undefined  = 0
  integer(IntAuc), parameter :: CLM_Matchup_Dummy             = 1
  integer(IntAuc), parameter :: CLM_Matchup_Nearest_Neighbour = 2

  TYPE WVM_Params_Type
     character(len=WVM_maxlen_File_Type)      :: File_Type
     ! shouldn't the next item be in a Sph_Default_Fields section ???
     character(len=WVM_maxlen_Sph_Descriptor) :: Sph_Descriptor
     real(FAdoxy)       :: Warmup_Pulse_Min_Height ! [m]
     real(FAdoxy)       :: Rangebin_Mismatch_Tolerance ! [m]
     logical            :: Line_of_Sight_Wind_Flag
     real(FAdoxy)       :: Mie_GroundCorr_WeightingFactor
     real(FAdoxy)       :: Rayl_GroundCorr_WeightingFactor
     !                     1234567890123456789012345678901
     integer(IntAs)     :: N_Obs_Mie_Max
     integer(IntAs)     :: N_Obs_Rayleigh_Max
     TYPE(Classification_Params_type)     :: Classification_Params
     TYPE(Optical_Properties_Params_type) :: Optical_Properties_Params
     TYPE(Error_Quantifier_Params_type)   :: Error_Quantifier_Params
     TYPE(Common_Processing_Params_type)  :: Common_Processing_Params
     TYPE(RBC_Algorithm_Params_type)      :: RBC_Algorithm_Params
     TYPE(AMD_Matchup_Params_type)        :: AMD_Matchup_Params
     TYPE(CLM_Matchup_Params_type)        :: CLM_Matchup_Params
     ! spare 500 bytes
  END TYPE WVM_Params_Type

  character(len=*), parameter :: unit_Warmup_Pulse_Min_Height = "m"
  character(len=*), parameter :: unit_Rangebin_Mismatch_Toler = "m"

  !  #]
  !  #[ Screening_Params and subtypes

  TYPE L1B_Obs_Scr_Params_type
     !                 1234567890123456789012345678901
     integer(IntAl) :: L1B_Laser_Freq_Unlocked_Thresh
     integer(IntAl) :: L1B_Ref_Pulses_Unlocked_Thresh
     real(FAdoxy)   :: L1B_Laser_Freq_Offset_Thresh    ! [GHz]
     real(FAdoxy)   :: L1B_Laser_UV_Energy_Thresh      ! [mJ]
     real(FAdoxy)   :: L1B_Laser_Freq_Offs_StdevThresh ! [GHz]
     real(FAdoxy)   :: L1B_Laser_UV_Energy_StdevThresh ! [mJ]
     real(FAdoxy)   :: L1B_Mie_Mean_Emit_Freq_Min      ! [GHz]
     real(FAdoxy)   :: L1B_Mie_Mean_Emit_Freq_Max      ! [GHz]
     real(FAdoxy)   :: L1B_Mie_Emit_Freq_Stdev_Thresh  ! [GHz] 
     real(FAdoxy)   :: L1B_Rayleigh_Mean_Emit_Freq_Min ! [GHz]
     real(FAdoxy)   :: L1B_Rayleigh_Mean_Emit_Freq_Max ! [GHz]
     real(FAdoxy)   :: L1B_RayleighEmitFreqStdevThresh ! [GHz]
     integer(IntAl) :: L1B_Sat_Not_on_Target_Thresh
     integer(IntAl) :: L1B_Mie_Corrupt_Thresh
     integer(IntAl) :: L1B_Rayleigh_Corrupt_Thresh
     integer(IntAl) :: L1B_Mie_Ref_PulsesCorruptThresh
     integer(IntAl) :: L1B_Rayl_RefPulsesCorruptThresh
     integer(IntAl) :: L1B_Mie_Invalid_Meas_Thresh
     integer(IntAl) :: L1B_Mie_Invalid_RefPulsesThresh
     integer(IntAl) :: L1B_Rayl_Invalid_Meas_Thresh
     integer(IntAl) :: L1B_Rayl_InvalidRefPulsesThresh
     !                 1234567890123456789012345678901
     ! spare 1000 bytes
  END TYPE L1B_Obs_Scr_Params_type

  ! use a common unit for all frequency related parameters
  character(len=*), parameter :: unit_Laser_Frequency = "GigaHertz"
  ! use a common unit for all laser UV energy related parameters
  character(len=*), parameter :: unit_Laser_UV_Energy = "mJ"

  TYPE L1B_Mie_Meas_Scr_Params_type
     !                 1234567890123456789012345678901
     integer(IntAl) :: L1B_MieMeasInvalidRefPlssThresh
     real(FAdoxy)   :: L1B_Avg_Laser_Freq_Offset_Min   ! [GHz]
     real(FAdoxy)   :: L1B_Avg_Laser_Freq_Offset_Max   ! [GHz]
     real(FAdoxy)   :: L1B_Avg_UV_Energy_Min           ! [mJ]
     real(FAdoxy)   :: L1B_Avg_UV_Energy_Max           ! [mJ]
     real(FAdoxy)   :: L1B_Laser_FreqOffsetStdevThresh ! [GHz]
     real(FAdoxy)   :: L1B_UV_Energy_Stdev_Thresh      ! [mJ]
     real(FAdoxy)   :: L1B_VelofAttUncertaintyErrorMin ! [m/s]
     real(FAdoxy)   :: L1B_VelofAttUncertaintyErrorMax ! [m/s]
     real(FAdoxy)   :: L1B_Mie_Mean_Emitted_Freq_Min   ! [GHz]
     real(FAdoxy)   :: L1B_Mie_Mean_Emitted_Freq_Max   ! [GHz]
     real(FAdoxy)   :: L1B_Mie_Emitted_FreqStdevThresh ! [GHz]
     real(FAdoxy)   :: L1B_Scattering_Ratio_Min
     real(FAdoxy)   :: L1B_Scattering_Ratio_Max
     real(FAdoxy)   :: L1B_Mie_SNR_Thresh
     real(FAdoxy)   :: Mie_Ground_Bin_Thickness_Thresh ! [m]
     integer(IntAs) :: Max_Signal_Derivative           ! [PixelLevel]
     integer(IntAs) :: Pixel_Saturation_Thresh         ! [PixelLevel]
     !                 1234567890123456789012345678901
     ! spare 1000 bytes
  END TYPE L1B_Mie_Meas_Scr_Params_type

  ! the next 2 are defined above, just below L1B_Obs_Scr_Params_type
  ! unit_Laser_Frequency = "GHz" 
  ! unit_Laser_UV_Energy = "mJ"

  character(len=*), parameter :: unit_Vel_of_Att_uncertainty = "m/s"
  character(len=*), parameter :: unit_bin_thickness_thresh   = "m"

  ! the next 2 are defined above, just below Corrupt_Data_Detect_Params_type
  !unit_Max_Signal_Derivative   = "PixelLevel"
  !unit_Pixel_Saturation_Thresh = "PixelLevel"

  TYPE L1B_Rayl_Meas_Scr_Params_type
     !                 1234567890123456789012345678901
     integer(IntAl) :: L1B_RayMeasInvalidRefPlssThresh
     real(FAdoxy)   :: L1B_Avg_Laser_Freq_Offset_Min   ! [GHz]
     real(FAdoxy)   :: L1B_Avg_Laser_Freq_Offset_Max   ! [GHz]
     real(FAdoxy)   :: L1B_Avg_UV_Energy_Min           ! [mJ]
     real(FAdoxy)   :: L1B_Avg_UV_Energy_Max           ! [mJ]
     real(FAdoxy)   :: L1B_Laser_FreqOffsetStdevThresh ! [GHz]
     real(FAdoxy)   :: L1B_UV_Energy_Stdev_Thresh      ! [mJ]
     real(FAdoxy)   :: L1B_VelOfAttUncertaintyErrorMin ! [m/s]
     real(FAdoxy)   :: L1B_VelOfAttUncertaintyErrorMax ! [m/s]
     real(FAdoxy)   :: L1B_Rayl_Mean_Emitted_Freq_Min  ! [GHz]
     real(FAdoxy)   :: L1B_Rayl_Mean_Emitted_Freq_Max  ! [GHz]
     real(FAdoxy)   :: L1B_Rayl_EmittedFreqStdevThresh ! [GHz]
     real(FAdoxy)   :: L1B_Rayleigh_SNR_Min
     real(FAdoxy)   :: L1B_Rayleigh_SNR_Max
     real(FAdoxy)   :: Rayl_Ground_Bin_ThicknessThresh ! [m]
     integer(IntAs) :: Pixel_Saturation_Thresh         ! [PixelLevel]
     !                 1234567890123456789012345678901
     ! spare 1000 bytes
  END TYPE L1B_Rayl_Meas_Scr_Params_type

  ! the next 2 are defined above, just below L1B_Obs_Scr_Params_type
  ! unit_Laser_Frequency = "GHz" 
  ! unit_Laser_UV_Energy = "mJ"

  ! the next 2 are defined above, just below L1B_Mie_Meas_Scr_Params_type
  ! unit_Vel_of_Att_uncertainty = "m/s"
  ! unit_bin_thickness_thresh   = "m"

  ! the next one is defined above, just below Corrupt_Data_Detect_Params_type
  ! unit_Pixel_Saturation_Thresh = "PixelLevel"

  TYPE L1B_Screening_Params_type
     TYPE(L1B_Obs_Scr_Params_type)       :: L1B_Obs_Screening_Params
     TYPE(L1B_Mie_Meas_Scr_Params_type)  :: L1B_Mie_Meas_Screening_Params
     TYPE(L1B_Rayl_Meas_Scr_Params_type) :: L1B_Rayl_Meas_Screening_Params
     !    1234567890123456789012345678901
  END TYPE L1B_Screening_Params_type

  TYPE L2B_AMD_Screening_Params_type
     real(FAdoxy) :: L2B_AMD_p_min                  ! [Pa]
     real(FAdoxy) :: L2B_AMD_p_max                  ! [Pa]
     real(FAdoxy) :: L2B_AMD_T_min                  ! [K]
     real(FAdoxy) :: L2B_AMD_T_max                  ! [K]
     ! spare 1000 bytes
  END TYPE L2B_AMD_Screening_Params_type

  character(len=*), parameter :: unit_L2B_AMD_p        = "Pa"
  character(len=*), parameter :: unit_L2B_AMD_T        = "K"

  TYPE Screening_Params_Type
     TYPE(L1B_Screening_Params_type)     :: L1B_Screening_Params
     TYPE(L2B_AMD_Screening_Params_type) :: L2B_AMD_Screening_Params
  END TYPE Screening_Params_Type
  !  #]

  TYPE L2B_AuxParData_type
     TYPE(FH_Default_Fields_Type)  :: FH_Default_Fields
     TYPE(MPH_Default_Fields_Type) :: MPH_Default_Fields
     TYPE(WVM_Params_Type)         :: WVM_Params
     TYPE(Screening_Params_Type)   :: Screening_Params
  END TYPE L2B_AuxParData_type

  TYPE L2B_AuxPar_Type
     ! default headers needed for all EE_CFI type files
     TYPE(FH_type),             pointer :: FH
     TYPE(MPH_type),            pointer :: MPH
     TYPE(L2B_AuxPar_SPH_type), pointer :: SPH
     !
     ! the attached dataset
     TYPE(L2B_AuxParData_type), pointer :: Data
  END TYPE L2B_AuxPar_Type

! REMARK: The generic FAdoxy should not be used in xml files!
!         the values for x and y should be specified !
!         This is not yet the case for most L2B_Auc_Par IODD elements
!         therefore this temporary maxlen_FAdoxy is needed

    ! temporary value
    integer, parameter :: maxlen_FAdoxy = 10

  !  #]
  !  #[ Module interface
  !     everything is public in this module
  !  #]
contains
  !-----------------------------------------------------
  !  #[ FH_Default_Fields
  subroutine Init_FH_Default_Fields(FH_Default_Fields)
    !  #[
    TYPE(FH_Default_Fields_type), intent(out) :: FH_Default_Fields
    
    FH_Default_Fields%File_Description(:) = ' '
    FH_Default_Fields%Notes(:)            = ' '
    FH_Default_Fields%Mission(:)          = ' '
    !FH_Default_Fields%Mission_Id(:)       = ' ' ! not used?
    FH_Default_Fields%File_Class(:)       = ' '
    FH_Default_Fields%File_Version(:)     = ' '
    FH_Default_Fields%System(:)           = ' '
    FH_Default_Fields%Creator(:)          = ' '
    FH_Default_Fields%Creator_Version(:)  = ' '

  end subroutine Init_FH_Default_Fields
    !  #]
  subroutine AsciiWrite_FH_Default_Fields(FH_Default_Fields, fileunit, &
                                          prefix, error_flag)
    !  #[
    TYPE(FH_Default_Fields_type), intent(in)  :: FH_Default_Fields
    integer,                      intent(in)  :: fileunit
    character(len=*),             intent(in)  :: prefix
    integer,                      intent(out) :: error_flag
    
    ! local variables
    character(len=256) :: txt

    error_flag = no_error

    txt = FH_Default_Fields%File_Description
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%File_Description = "//trim(txt)

    txt = FH_Default_Fields%Notes
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Notes            = "//trim(txt)

    txt = FH_Default_Fields%Mission
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Mission          = "//trim(txt)

    ! not used?
    !txt = FH_Default_Fields%Mission_Id
    !IF (len_trim(txt).eq.0) txt = "[empty]"
    !write(fileunit,"(a)",err=999) &
    !     trim(prefix)//"%Mission_Id       = "//trim(txt)

    txt = FH_Default_Fields%File_Class
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%File_Class       = "//trim(txt)

    txt = FH_Default_Fields%File_Version
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%File_Version     = "//trim(txt)

    txt = FH_Default_Fields%System
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%System           = "//trim(txt)

    txt = FH_Default_Fields%Creator
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Creator          = "//trim(txt)

    txt = FH_Default_Fields%Creator_Version
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Creator_Version  = "//trim(txt)

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_FH_Default_Fields "//&
                          "error while writing to outputfile")
    error_flag = error_writing_file
    return
    
  end subroutine AsciiWrite_FH_Default_Fields
    !  #]
  subroutine Get_FH_Def_Flds_from_xml_tree(xml_tree, xml_path, &
                                           FH_Default_Fields, error_flag)
    !  #[
    TYPE(xml_node_type),          pointer     :: xml_tree
    character(len=*),             intent(in)  :: xml_path
    TYPE(FH_Default_Fields_type), intent(out) :: FH_Default_Fields
    integer,                      intent(out) :: error_flag

    ! local variables
    character(len=max_str_len)  :: path
    character(len=max_str_len)  :: text

    ! init
    error_flag = no_error

    IF (.not. associated(xml_tree)) THEN
       call logmsg(log_error,"ERROR in Get_FH_Def_Flds_from_xml_tree():"//&
            "the pointer xml_tree seems not associated !!!")
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! copying the FH_Default_Fields XML elements
    !------------------------------

    path = trim(xml_path)//"/File_Description"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    FH_Default_Fields%File_Description = check_length_xml_item(text,path,&
         FH_maxlen_file_description,"Get_FH_Def_Flds_from_xml_tree")

    path = trim(xml_path)//"/Notes"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    FH_Default_Fields%Notes = check_length_xml_item(text,path,&
         FH_maxlen_notes,"Get_FH_Def_Flds_from_xml_tree")

    path = trim(xml_path)//"/Mission"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    FH_Default_Fields%Mission = check_length_xml_item(text,path,&
         FH_maxlen_mission,"Get_FH_Def_Flds_from_xml_tree")

    !path = trim(xml_path)//"/Mission_Id" ! not used?

    path = trim(xml_path)//"/File_Class"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
!    FH_Default_Fields%File_Class = check_length_xml_item(text,path,&
!         FH_maxlen_file_class,"Get_FH_Def_Flds_from_xml_tree")
!   Use of check_length_xml_item disabled 11/12/07 while File_Class is obsolete.
!   TODO: review.
    FH_Default_Fields%File_Class = text

    path = trim(xml_path)//"/File_Version"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    FH_Default_Fields%File_Version = check_length_xml_item(text,path,&
         FH_maxlen_file_version,"Get_FH_Def_Flds_from_xml_tree")

    path = trim(xml_path)//"/System"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    FH_Default_Fields%System = check_length_xml_item(text,path,&
         FH_maxlen_system,"Get_FH_Def_Flds_from_xml_tree")

    path = trim(xml_path)//"/Creator"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    FH_Default_Fields%Creator = check_length_xml_item(text,path,&
         FH_maxlen_creator,"Get_FH_Def_Flds_from_xml_tree")

    path = trim(xml_path)//"/Creator_Version"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    FH_Default_Fields%Creator_Version = check_length_xml_item(text,path,&
         FH_maxlen_creator_version,"Get_FH_Def_Flds_from_xml_tree")

    return

  end subroutine Get_FH_Def_Flds_from_xml_tree
    !  #]
  !  #]
  !-----------------------------------------------------
  !  #[ MPH_Default_Fields
  subroutine Init_MPH_Default_Fields(MPH_Default_Fields)
    !  #[
    TYPE(MPH_Default_Fields_type), intent(out) :: MPH_Default_Fields
    
    MPH_Default_Fields%Ref_Doc(:)      = ' '
    MPH_Default_Fields%Software_Ver(:) = ' '
    MPH_Default_Fields%Proc_Center(:)  = ' '

  end subroutine Init_MPH_Default_Fields
    !  #]
  subroutine AsciiWrite_MPH_Default_Fields(MPH_Default_Fields, fileunit, &
                                           prefix, error_flag)
    !  #[
    TYPE(MPH_Default_Fields_type), intent(in)  :: MPH_Default_Fields
    integer,                       intent(in)  :: fileunit
    character(len=*),              intent(in)  :: prefix
    integer,                       intent(out) :: error_flag
    
    ! local variables
    character(len=256) :: txt

    error_flag = no_error

    txt = MPH_Default_Fields%Ref_Doc
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Ref_Doc      = "//trim(txt)

    txt = MPH_Default_Fields%Software_Ver
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Software_Ver = "//trim(txt)

    txt = MPH_Default_Fields%Proc_Center
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Proc_Center  = "//trim(txt)

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_MPH_Default_Fields "//&
                          "error while writing to outputfile")
    error_flag = error_writing_file
    return
    
  end subroutine AsciiWrite_MPH_Default_Fields
    !  #]
  subroutine Get_MPH_Def_Flds_from_xml_tree(xml_tree, xml_path, &
                                            MPH_Default_Fields, error_flag)
    !  #[
    TYPE(xml_node_type),           pointer     :: xml_tree
    character(len=*),              intent(in)  :: xml_path
    TYPE(MPH_Default_Fields_type), intent(out) :: MPH_Default_Fields
    integer,                       intent(out) :: error_flag

    ! local variables
    character(len=max_str_len)  :: path
    character(len=max_str_len)  :: text

    ! init
    error_flag = no_error

    IF (.not. associated(xml_tree)) THEN
       call logmsg(log_error,"ERROR in Get_MPH_Def_Flds_from_xml_tree():"//&
            "the pointer xml_tree seems not associated !!!")
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! copying the MPH_Default_Fields XML elements
    !------------------------------

    path = trim(xml_path)//"/Ref_Doc"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH_Default_Fields%Ref_Doc = check_length_xml_item(text,path,&
         MPH_maxlen_ref_doc,"Get_MPH_Def_Flds_from_xml_tree")

    path = trim(xml_path)//"/Software_Ver"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    MPH_Default_Fields%Software_Ver = check_length_xml_item(text,path,&
         MPH_maxlen_software_ver,"Get_MPH_Def_Flds_from_xml_tree")

    path = trim(xml_path)//"/Proc_Center"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
!    MPH_Default_Fields%Proc_Center = check_length_xml_item(text,path,&
!         MPH_maxlen_proc_center,"Get_MPH_Def_Flds_from_xml_tree")
!   Use of check_length_xml_item disabled 11/12/07 while Proc_Center is obsolete.
!   TODO: review.
    MPH_Default_Fields%Proc_Center = text

    return

  end subroutine Get_MPH_Def_Flds_from_xml_tree
    !  #]
  !  #]
  !-----------------------------------------------------
  !  #[ WVM Params
  subroutine Init_Classification_Params(Classification_Params)
    !  #[
    TYPE(Classification_Params_type), intent(out) :: Classification_Params
    
    Classification_Params%Classification_Type_Mie      = Class_undefined
    Classification_Params%Classification_Type_Rayleigh = Class_undefined
    nullify(Classification_Params%Threshold_BackscatterRatio)
    nullify(Classification_Params%Threshold_Extinction)

  end subroutine Init_Classification_Params
    !  #]
  subroutine Delete_Classification_Params(Classification_Params)
    !  #[
    TYPE(Classification_Params_type), intent(inout) :: Classification_Params
    
    IF (associated(Classification_Params%Threshold_BackscatterRatio)) &
         deallocate(Classification_Params%Threshold_BackscatterRatio)
    IF (associated(Classification_Params%Threshold_Extinction)) &
         deallocate(Classification_Params%Threshold_Extinction)

  end subroutine Delete_Classification_Params
    !  #]
  subroutine Get_Class_Type_Text(Class_Type,Class_Text,error_flag)
    !  #[
    integer(IntAuc),  intent(in)  :: Class_Type
    character(len=*), intent(out) :: Class_Text
    integer,          intent(out) :: error_flag

    Class_Text(:) = ' '
    error_flag    = no_error

    select case(Class_Type)
    case(Class_undefined)
       Class_Text = "Class_undefined"
    case(Class_No_Clouds)
       Class_Text = "Class_No_Clouds"
    case(Class_Copy_E2S_Input)
       Class_Text = "Class_Copy_E2S_Input"
    case(Class_Ext_Threshold)
       Class_Text = "Class_Ext_Threshold"
    case(Class_Cal_Mie_Threshold)
       Class_Text = "Class_Cal_Mie_Threshold"
    case(Class_Rayleigh_Slope)
       Class_Text = "Class_Rayleigh_Slope"
    case(Class_Mie_Slope)
       Class_Text = "Class_Mie_Slope"
    case(Class_Backscat_Ratio)
       Class_Text = "Class_Backscat_Ratio"
    case(Class_Adjacent_Mie_Backscatter)
       Class_Text = "Class_Adjacent_Mie_Backscatter"
    case(Class_Adjacent_Rayl_Backscatter)
       Class_Text = "Class_Adjacent_Rayl_Backscatter"
    case(Class_Tau)
       Class_Text = "Class_Tau"
    case default
       call logmsg(log_error,"ERROR in Get_Class_Type_Text "//&
                   "unknown enum value for Classification_Type")
       error_flag = error_writing_file
       return
    end select

  end subroutine Get_Class_Type_Text
    !  #]
  subroutine Get_Class_Type_Code(Class_Text,Class_Type,error_flag)
    !  #[
    character(len=*), intent(in)  :: Class_Text
    integer(IntAuc),  intent(out) :: Class_Type
    integer,          intent(out) :: error_flag

    Class_Type = Class_undefined
    error_flag = no_error

    select case(trim(Class_Text))
    case("Class_undefined")
       ! this should only be used for initialisation of the datastructure
       ! as missing, and should not be present in the input file!
       call logmsg(log_error,"ERROR in Get_Class_Type_Code: "//&
         "invalid setting for Classification_Type: [Class_undefined]"//&
         " this should only be used for initialisation of the "//&
         " datastructure as missing_indicator, and should not be present"//&
         " in the L2B_Aux_Par input file!")
       error_flag = error_reading_file
       return
    case("Class_No_Clouds")
       Class_Type = Class_No_Clouds
    case("Class_Copy_E2S_Input")
       Class_Type = Class_Copy_E2S_Input
    case("Class_Ext_Threshold")
       Class_Type = Class_Ext_Threshold
    case("Class_Cal_Mie_Threshold")
       Class_Type = Class_Cal_Mie_Threshold
    case("Class_Rayleigh_Slope")
       Class_Type = Class_Rayleigh_Slope
    case("Class_Mie_Slope")
       Class_Type = Class_Mie_Slope
    case("Class_Backscat_Ratio")
       Class_Type = Class_Backscat_Ratio
    case("Class_Adjacent_Mie_Backscatter")
       Class_Type = Class_Adjacent_Mie_Backscatter
    case("Class_Adjacent_Rayl_Backscatter")
       Class_Type = Class_Adjacent_Rayl_Backscatter
    case("Class_Tau")
       Class_Type = Class_Tau
    case default
       call logmsg(log_error,"ERROR in Get_Class_Type_Code: "//&
            "invalid setting for Classification_Type: ["//&
            trim(Class_Text)//"] this Classification type is unknown")
       error_flag = error_reading_file
       return
   end select

  end subroutine Get_Class_Type_Code
    !  #]
  subroutine AsciiWrite_Classific_Params(Classification_Params,fileunit,&
                                         prefix, error_flag)
    !  #[
    TYPE(Classification_Params_type), intent(in)  :: Classification_Params
    integer,                          intent(in)  :: fileunit
    character(len=*),                 intent(in)  :: prefix
    integer,                          intent(out) :: error_flag

    ! local variables
    integer            :: s,i
    character(len=2)   :: str_cntr
    character(len=256) :: txt
    character(len=25)  :: str_FAdoxy

    error_flag = no_error

    call Get_Class_Type_Text(&
         Classification_Params%Classification_Type_Mie,txt,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Classification_Type_Mie      = "//trim(txt)

    call Get_Class_Type_Text(&
         Classification_Params%Classification_Type_Rayleigh,txt,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Classification_Type_Rayleigh = "//trim(txt)

    write(fileunit,"(a)",err=999) "REALACC(12)"

    IF (associated(Classification_Params%Threshold_BackscatterRatio)) THEN
       s = size(Classification_Params%Threshold_BackscatterRatio)
       DO i=1,s
          write(str_cntr,"(i2.2)",err=998) i

          call SetFAdoFull_or_Missing(Classification_Params%&
               Threshold_BackscatterRatio(i)%Threshold_Value,&
               str_FAdoxy,error_flag)
          if (error_flag .ne. no_error) return

          write(fileunit,"(a)",err=999) &
               trim(prefix)//"%Threshold_BackscatterRatio("//&
               trim(str_cntr)//")%Threshold_Value = "//trim(str_FAdoxy)

          call SetFAdoFull_or_Missing(Classification_Params%&
               Threshold_BackscatterRatio(i)%Altitude,&
               str_FAdoxy,error_flag)
          if (error_flag .ne. no_error) return

          write(fileunit,"(a)",err=999) &
               trim(prefix)//"%Threshold_BackscatterRatio("//&
               trim(str_cntr)//")%Altitude        = "//&
               trim(str_FAdoxy)//" ["//unit_Altitude//"]"
       END DO
    ELSE
       write(fileunit,"(a)",err=999) &
            trim(prefix)//"%Threshold_BackscatterRatio(:) [is not associated]"
    END IF

    IF (associated(Classification_Params%Threshold_Extinction)) THEN
       s = size(Classification_Params%Threshold_Extinction)
       DO i=1,s
          write(str_cntr,"(i2.2)",err=998) i

          call SetFAdoFull_or_Missing(Classification_Params%&
               Threshold_Extinction(i)%Threshold_Value,&
               str_FAdoxy,error_flag)
          if (error_flag .ne. no_error) return

          write(fileunit,"(a)",err=999) &
               trim(prefix)//"%Threshold_Extinction("//&
               trim(str_cntr)//")%Threshold_Value = "//&
               trim(str_FAdoxy)//" ["//unit_Thresh_Value_Extinction//"]"

          call SetFAdoFull_or_Missing(Classification_Params%&
               Threshold_Extinction(i)%Altitude,&
               str_FAdoxy,error_flag)
          if (error_flag .ne. no_error) return

          write(fileunit,"(a)",err=999) &
               trim(prefix)//"%Threshold_Extinction("//&
               trim(str_cntr)//")%Altitude        = "//&
               trim(str_FAdoxy)//" ["//unit_Altitude//"]"
       END DO
    ELSE
       write(fileunit,"(a)",err=999) &
            trim(prefix)//"%Threshold_Extinction(:) [is not associated]"
    END IF

    write(fileunit,"(a)",err=999) "ENDREALACC"

    return

998 call logmsg(log_error,"ERROR in AsciiWrite_Classific_Params "//&
         "could not convert the counter i  to a string"//&
         "(maybe the format is too small?)")
    error_flag = error_writing_file
    return

999 call logmsg(log_error,"ERROR in AsciiWrite_Classific_Params "//&
                          "error while writing to outputfile")
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Classific_Params
    !  #]
  subroutine Get_Class_Params_from_xml_tree(xml_tree, xml_path, &
                                            Classification_Params, error_flag)
    !  #[
    TYPE(xml_node_type),              pointer     :: xml_tree
    character(len=*),                 intent(in)  :: xml_path
    TYPE(Classification_Params_type), intent(out) :: Classification_Params
    integer,                          intent(out) :: error_flag

    ! local variables
    character(len=max_str_len) :: path
    character(len=max_str_len) :: text
    character(len=max_str_len) :: attr_value
    character(len=3)           :: str_cntr
    integer                    :: i, count, ios, AllocStatus

    ! init
    error_flag = no_error

    IF (.not. associated(xml_tree)) THEN
       call logmsg(log_error,"ERROR in Get_Class_Params_from_xml_tree():"//&
            "the pointer xml_tree seems not associated !!!")
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! copying the Classification_Params XML elements
    !------------------------------

    path = trim(xml_path)//"/Classification_Type_Mie"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    call Get_Class_Type_Code(text,&
         Classification_Params%Classification_Type_Mie,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/Classification_Type_Rayleigh"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    call Get_Class_Type_Code(text,&
         Classification_Params%Classification_Type_Rayleigh,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/List_of_BackscatterRatio_Thresholds"
    ! get the nr of points in the Threshold profile
    call xml_tree_read_string_attribute(xml_tree,path,"count",attr_value,&
                                        error_flag)
    IF (error_flag .ne. no_error) return
    
    read(attr_value,*,iostat=ios) count
    IF (ios .ne. 0) goto 999

    ! allocate the needed array for storing the Thresholds
    ! but deallocate it first in case the array was already allocated
    IF (associated(Classification_Params%Threshold_BackscatterRatio)) &
         deallocate(Classification_Params%Threshold_BackscatterRatio)
    allocate(Classification_Params%Threshold_BackscatterRatio(count),&
             stat=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       call logmsg(log_error,"ERROR in Get_Class_Params_from_xml_tree():"//&
            " memory allocation failed")
       error_flag = error_allocate
       return
    END IF

    DO i=1,count
       write(str_cntr,"(i3)",err=998) i

       path = trim(xml_path)//"/List_of_BackscatterRatio_Thresholds"//&
            "/BackscatterRatio_Threshold["//trim(adjustl(str_cntr))//"]"//&
            "/Threshold_Value"
       call GetFAdoxy_from_xml_tree(xml_tree,path,Classification_Params%&
            Threshold_BackscatterRatio(i)%Threshold_Value,error_flag)
       IF (error_flag .ne. no_error) return

       path = trim(xml_path)//"/List_of_BackscatterRatio_Thresholds"//&
            "/BackscatterRatio_Threshold["//trim(adjustl(str_cntr))//"]"//&
            "/Altitude"
       call GetFAdoxy_from_xml_tree(xml_tree,path,Classification_Params%&
            Threshold_BackscatterRatio(i)%Altitude,error_flag,&
            unit=unit_Altitude)
       IF (error_flag .ne. no_error) return

    END DO

    ! allocate the needed array for storing the Thresholds
    ! but deallocate it first in case the array was already allocated
    IF (associated(Classification_Params%Threshold_Extinction)) &
         deallocate(Classification_Params%Threshold_Extinction)
    allocate(Classification_Params%Threshold_Extinction(count),&
             stat=AllocStatus)
    IF (AllocStatus .ne. 0) THEN
       call logmsg(log_error,"ERROR in Get_Class_Params_from_xml_tree():"//&
            " memory allocation failed")
       error_flag = error_allocate
       return
    END IF

    DO i=1,count
       write(str_cntr,"(i3)",err=998) i

       path = trim(xml_path)//"/List_of_Extinction_Thresholds"//&
            "/Extinction_Threshold["//trim(adjustl(str_cntr))//"]"//&
            "/Threshold_Value"
       call GetFAdoxy_from_xml_tree(xml_tree,path,Classification_Params%&
            Threshold_Extinction(i)%Threshold_Value,error_flag,&
            unit=unit_Thresh_Value_Extinction)
       IF (error_flag .ne. no_error) return

       path = trim(xml_path)//"/List_of_Extinction_Thresholds"//&
            "/Extinction_Threshold["//trim(adjustl(str_cntr))//"]"//&
            "/Altitude"
       call GetFAdoxy_from_xml_tree(xml_tree,path,Classification_Params%&
            Threshold_Extinction(i)%Altitude,error_flag,&
            unit=unit_Altitude)
       IF (error_flag .ne. no_error) return

    END DO

    ! spare 1000 bytes

    return

998 call logmsg(log_error,"ERROR in Get_Class_Params_from_xml_tree():"//&
         " could not convert counter i to a proper counter string ")
    error_flag = error_writing_file
    return

999 call logmsg(log_error,"ERROR in Get_Class_Params_from_xml_tree():"//&
         " could not convert the count attribute string ["//trim(attr_value)//&
         " to a value for item ["//trim(path)//"]")
    error_flag = error_reading_file
    return

  end subroutine Get_Class_Params_from_xml_tree
    !  #]
  !-----------------------------------------------------
  subroutine Init_Optical_Properties_Params(Optical_Properties_Params)
    !  #[
    TYPE(Optical_Properties_Params_type), intent(out) :: &
         Optical_Properties_Params
    
    Optical_Properties_Params%ScatRatio_Method  = Scat_Ratio_Undefined
    Optical_Properties_Params%ScatRatio_Method2 = Scat_Ratio_Undefined

  end subroutine Init_Optical_Properties_Params
    !  #]
  subroutine Get_Scat_Ratio_Method_Text(Scat_Ratio_Method,&
                                   Scat_Ratio_Method_Text,error_flag)
    !  #[
    integer(IntAuc),  intent(in)  :: Scat_Ratio_Method
    character(len=*), intent(out) :: Scat_Ratio_Method_Text
    integer,          intent(out) :: error_flag

    error_flag                = no_error
    Scat_Ratio_Method_Text(:) = ' '

    select case(Scat_Ratio_Method)
    case(Scat_Ratio_Undefined)
       Scat_Ratio_Method_Text = "Scat_Ratio_Undefined"
    case(Scat_Ratio_from_L1B)
       Scat_Ratio_Method_Text = "Scat_Ratio_from_L1B"
    case(Scat_Ratio_from_MieRayl)
       Scat_Ratio_Method_Text = "Scat_Ratio_from_MieRayl"
    case(Scat_Ratio_from_RaylOnly)
       Scat_Ratio_Method_Text = "Scat_Ratio_from_RaylOnly"
    case(Scat_Ratio_Dont_Use)
       Scat_Ratio_Method_Text = "Scat_Ratio_Dont_Use"
    case default
       call logmsg(log_error,"ERROR in Get_Scat_Ratio_Method_Text "//&
                   "unknown enum value for Scat_Ratio_Method")
       error_flag = error_writing_file
       return
    end select

  end subroutine Get_Scat_Ratio_Method_Text
    !  #]
  subroutine Get_Scat_Ratio_Method_Code(Scat_Ratio_Method_Text,&
                                        Scat_Ratio_Method,error_flag)
    !  #[
    character(len=*), intent(in)  :: Scat_Ratio_Method_Text
    integer(IntAuc),  intent(out) :: Scat_Ratio_Method
    integer,          intent(out) :: error_flag

    Scat_Ratio_Method = Scat_Ratio_Undefined
    error_flag = no_error

    select case(trim(Scat_Ratio_Method_Text))
    case("Scat_Ratio_Undefined")
       ! this should only be used for initialisation of the datastructure
       ! as missing, and should not be present in the input file!
       call logmsg(log_error,"ERROR in Get_Scat_Ratio_Method_Code: "//&
         "invalid setting for Scat_Ratio_Method: [Scat_Ratio_Undefined]"//&
         " this should only be used for initialisation of the "//&
         " datastructure as missing_indicator, and should not be present"//&
         " in the L2B_Aux_Par input file!")
       error_flag = error_reading_file
       return
    case("Scat_Ratio_from_L1B")
       Scat_Ratio_Method = Scat_Ratio_from_L1B
    case("Scat_Ratio_from_MieRayl")
       Scat_Ratio_Method = Scat_Ratio_from_MieRayl
    case("Scat_Ratio_from_RaylOnly")
       Scat_Ratio_Method = Scat_Ratio_from_RaylOnly
    case("Scat_Ratio_Dont_Use")
       Scat_Ratio_Method = Scat_Ratio_Dont_Use
    case default
       call logmsg(log_error,"ERROR in Get_Scat_Ratio_Method_Code: "//&
            "invalid setting for Scat_Ratio_Method: ["//&
            trim(Scat_Ratio_Method_Text)//&
            "] this Scat_Ratio_Method is unknown")
       error_flag = error_reading_file
       return
   end select

 end subroutine Get_Scat_Ratio_Method_Code
    !  #]
  subroutine AsciiWrite_Optical_Prop_Params(Optical_Properties_Params,&
                                            fileunit,prefix, error_flag)
    !  #[
    TYPE(Optical_Properties_Params_type), &
                      intent(in)  :: Optical_Properties_Params
    integer,          intent(in)  :: fileunit
    character(len=*), intent(in)  :: prefix
    integer,          intent(out) :: error_flag

    ! local variables
    character(len=256)          :: txt

    error_flag = no_error

    call Get_Scat_Ratio_Method_Text(&
         Optical_Properties_Params%ScatRatio_Method,txt,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Scat_Ratio_Method  = "//trim(txt)

    call Get_Scat_Ratio_Method_Text(&
         Optical_Properties_Params%ScatRatio_Method2,txt,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Scat_Ratio_Method2 = "//trim(txt)

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_Classific_Params "//&
                          "error while writing to outputfile")
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Optical_Prop_Params
    !  #]
  subroutine Get_OptPropPars_from_xml_tree(xml_tree, xml_path, &
                                      Optical_Properties_Params, error_flag)
    !  #[
    TYPE(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: xml_path
    TYPE(Optical_Properties_Params_type), &
                         intent(out) :: Optical_Properties_Params
    integer,             intent(out) :: error_flag

    ! local variables
    character(len=max_str_len)  :: path
    character(len=max_str_len)  :: text

    ! init
    error_flag = no_error

    IF (.not. associated(xml_tree)) THEN
       call logmsg(log_error,"ERROR in Get_OptPropPars_from_xml_tree():"//&
            "the pointer xml_tree seems not associated !!!")
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! copying the Optical_Properties_Params XML elements
    !------------------------------

    path = trim(xml_path)//"/ScatRatio_Method"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    call Get_Scat_Ratio_Method_Code(text,&
         Optical_Properties_Params%ScatRatio_Method,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/ScatRatio_Method2"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    call Get_Scat_Ratio_Method_Code(text,&
         Optical_Properties_Params%ScatRatio_Method2,error_flag)
    IF (error_flag .ne. no_error) return

    ! spare 1000 bytes

    return

  end subroutine Get_OptPropPars_from_xml_tree
    !  #]
  !-----------------------------------------------------
  subroutine Init_Error_Quantifier_Params(Error_Quantifier_Params)
    !  #[
    TYPE(Error_Quantifier_Params_type), intent(out) :: Error_Quantifier_Params
    
    Error_Quantifier_Params%ErrorQuantMethod_Mie      = &
         ErrorQuantMethod_Mie_undefined
    Error_Quantifier_Params%ErrorQuantMethod_Rayleigh = &
         ErrorQuantMethod_Ray_undefined

  end subroutine Init_Error_Quantifier_Params
    !  #]
  subroutine Get_ErrorQuantMethod_Mie_Text(ErrorQuantMethod_Mie,&
                                   ErrorQuantMethod_Mie_Text,error_flag)
    !  #[
    integer(IntAuc),  intent(in)  :: ErrorQuantMethod_Mie
    character(len=*), intent(out) :: ErrorQuantMethod_Mie_Text
    integer,          intent(out) :: error_flag

    error_flag                = no_error
    ErrorQuantMethod_Mie_Text(:) = ' '

    select case(ErrorQuantMethod_Mie)
    case(ErrorQuantMethod_Mie_undefined)
       ErrorQuantMethod_Mie_Text = "ErrorQuantMethod_Mie_undefined"
    case(ErrorQuantMethod_Mie_1Bweighted)
       ErrorQuantMethod_Mie_Text = "ErrorQuantMethod_Mie_1Bweighted"
    case(ErrorQuantMethod_Mie_core_sens)
       ErrorQuantMethod_Mie_Text = "ErrorQuantMethod_Mie_core_sens"
    case default
       call logmsg(log_error,"ERROR in Get_ErrorQuantMethod_Mie_Text "//&
                   "unknown enum value for ErrorQuantMethod_Mie")
       error_flag = error_writing_file
       return
    end select

  end subroutine Get_ErrorQuantMethod_Mie_Text
    !  #]
  subroutine Get_ErrorQuantMethod_Rayl_Text(ErrorQuantMethod_Rayleigh,&
                                   ErrorQuantMethod_Rayleigh_Text,error_flag)
    !  #[
    integer(IntAuc),  intent(in)  :: ErrorQuantMethod_Rayleigh
    character(len=*), intent(out) :: ErrorQuantMethod_Rayleigh_Text
    integer,          intent(out) :: error_flag

    error_flag                = no_error
    ErrorQuantMethod_Rayleigh_Text(:) = ' '

    select case(ErrorQuantMethod_Rayleigh)
    case(ErrorQuantMethod_Ray_undefined)
       ErrorQuantMethod_Rayleigh_Text = "ErrorQuantMethod_Ray_undefined"
    case(ErrorQuantMethod_Ray_1Bweighted)
       ErrorQuantMethod_Rayleigh_Text = "ErrorQuantMethod_Ray_1Bweighted"
    case(ErrorQuantMethod_Ray_iliad_sens)
       ErrorQuantMethod_Rayleigh_Text = "ErrorQuantMethod_Ray_iliad_sens"
    case default
       call logmsg(log_error,"ERROR in Get_ErrorQuantMethod_Rayl_Text "//&
                   "unknown enum value for ErrorQuantMethod_Rayleigh")
       error_flag = error_writing_file
       return
    end select

  end subroutine Get_ErrorQuantMethod_Rayl_Text
    !  #]
  subroutine Get_ErrorQuantMethodMie_Code(ErrorQuantMethod_Mie_Text,&
                                          ErrorQuantMethod_Mie,error_flag)
    !  #[
    character(len=*), intent(in)  :: ErrorQuantMethod_Mie_Text
    integer(IntAuc),  intent(out) :: ErrorQuantMethod_Mie
    integer,          intent(out) :: error_flag

    ErrorQuantMethod_Mie = ErrorQuantMethod_Mie_undefined
    error_flag = no_error

    select case(trim(ErrorQuantMethod_Mie_Text))
    case("ErrorQuantMethod_Mie_undefined")
       ! this should only be used for initialisation of the datastructure
       ! as missing, and should not be present in the input file!
       call logmsg(log_error,"ERROR in Get_ErrorQuantMethodMie_Code:"//&
         " invalid setting for ErrorQuantMethod_Mie:"//&
         " [ErrorQuantMethod_Mie_undefined]"//&
         " this should only be used for initialisation of the "//&
         " datastructure as missing_indicator, and should not be present"//&
         " in the L2B_Aux_Par input file!")
       error_flag = error_reading_file
       return
    case("ErrorQuantMethod_Mie_1Bweighted")
       ErrorQuantMethod_Mie = ErrorQuantMethod_Mie_1Bweighted
    case("ErrorQuantMethod_Mie_core_sens")
       ErrorQuantMethod_Mie = ErrorQuantMethod_Mie_core_sens
    case default
       call logmsg(log_error,"ERROR in Get_ErrorQuantMethodMie_Code: "//&
            "invalid setting for ErrorQuantMethod_Mie: ["//&
            trim(ErrorQuantMethod_Mie_Text)//&
            "] this ErrorQuantMethod_Mie is unknown")
       error_flag = error_reading_file
       return
    end select

  end subroutine Get_ErrorQuantMethodMie_Code
    !  #]
  subroutine Get_ErrorQuantMethodRayl_Code(ErrorQuantMethod_Rayleigh_Text,&
                                       ErrorQuantMethod_Rayleigh,error_flag)
    !  #[
    character(len=*), intent(in)  :: ErrorQuantMethod_Rayleigh_Text
    integer(IntAuc),  intent(out) :: ErrorQuantMethod_Rayleigh
    integer,          intent(out) :: error_flag

    ErrorQuantMethod_Rayleigh = ErrorQuantMethod_Ray_undefined
    error_flag = no_error

    select case(trim(ErrorQuantMethod_Rayleigh_Text))
    case("ErrorQuantMethod_Ray_undefined")
       ! this should only be used for initialisation of the datastructure
       ! as missing, and should not be present in the input file!
       call logmsg(log_error,"ERROR in Get_ErrorQuantMethodRayl_Code:"//&
         " invalid setting for ErrorQuantMethod_Rayleigh:"//&
         " [ErrorQuantMethod_Ray_undefined]"//&
         " this should only be used for initialisation of the "//&
         " datastructure as missing_indicator, and should not be present"//&
         " in the L2B_Aux_Par input file!")
       error_flag = error_reading_file
       return
    case("ErrorQuantMethod_Ray_1Bweighted")
       ErrorQuantMethod_Rayleigh = ErrorQuantMethod_Ray_1Bweighted
    case("ErrorQuantMethod_Ray_iliad_sens")
       ErrorQuantMethod_Rayleigh = ErrorQuantMethod_Ray_iliad_sens
    case default
       call logmsg(log_error,"ERROR in Get_ErrorQuantMethodRayl_Code: "//&
            "invalid setting for ErrorQuantMethod_Rayleigh: ["//&
            trim(ErrorQuantMethod_Rayleigh_Text)//&
            "] this ErrorQuantMethod_Rayleigh is unknown")
       error_flag = error_reading_file
       return
    end select

  end subroutine Get_ErrorQuantMethodRayl_Code
    !  #]
  subroutine AsciiWrite_Error_Quant_Params(Error_Quantifier_Params,&
                                           fileunit,prefix, error_flag)
    !  #[
    TYPE(Error_Quantifier_Params_type), &
                      intent(in)  :: Error_Quantifier_Params
    integer,          intent(in)  :: fileunit
    character(len=*), intent(in)  :: prefix
    integer,          intent(out) :: error_flag

    ! local variables
    character(len=256)          :: txt

    error_flag = no_error

    call Get_ErrorQuantMethod_Mie_Text(&
         Error_Quantifier_Params%ErrorQuantMethod_Mie,txt,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%ErrorQuantMethod_Mie       = "//trim(txt)

    call Get_ErrorQuantMethod_Rayl_Text(&
         Error_Quantifier_Params%ErrorQuantMethod_Rayleigh,txt,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%ErrorQuantMethod_Rayleigh  = "//trim(txt)

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_Error_Quant_Params "//&
                          "error while writing to outputfile")
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Error_Quant_Params
    !  #]
  subroutine Get_ErrorQuantPars_from_xmltree(xml_tree, xml_path, &
                                      Error_Quantifier_Params, error_flag)
    !  #[
    TYPE(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: xml_path
    TYPE(Error_Quantifier_Params_type), &
                         intent(out) :: Error_Quantifier_Params
    integer,             intent(out) :: error_flag

    ! local variables
    character(len=max_str_len)  :: path
    character(len=max_str_len)  :: text

    ! init
    error_flag = no_error

    IF (.not. associated(xml_tree)) THEN
       call logmsg(log_error,"ERROR in Get_ErrorQuantPars_from_xmltree():"//&
            "the pointer xml_tree seems not associated !!!")
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! copying the Error_Quantifier_Params XML elements
    !------------------------------

    path = trim(xml_path)//"/ErrorQuantMethod_Mie"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    call Get_ErrorQuantMethodMie_Code(text,&
         Error_Quantifier_Params%ErrorQuantMethod_Mie,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/ErrorQuantMethod_Rayleigh"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    call Get_ErrorQuantMethodRayl_Code(text,&
         Error_Quantifier_Params%ErrorQuantMethod_Rayleigh,error_flag)
    IF (error_flag .ne. no_error) return

    ! spare 1000 bytes

    return

  end subroutine Get_ErrorQuantPars_from_xmltree
    !  #]
  !-----------------------------------------------------
  subroutine Init_Mie_PCD_Params(Mie_PCD_Params)
    !  #[
    TYPE(Mie_PCD_Params_type), intent(out) :: Mie_PCD_Params

    Mie_PCD_Params%Alpha_Correction = missing_indicator_FAdoxy
    Mie_PCD_Params%Summation_Index  = missing_indicator_IntAs

  end subroutine Init_Mie_PCD_Params
    !  #]
  subroutine AsciiWrite_Mie_PCD_Params(Mie_PCD_Params,fileunit, &
                                      prefix, error_flag)
    !  #[
    TYPE(Mie_PCD_Params_type), intent(in)  :: Mie_PCD_Params
    integer,                   intent(in)  :: fileunit
    character(len=*),          intent(in)  :: prefix
    integer,                   intent(out) :: error_flag

    ! local variables
    character(len=25)           :: str_FAdoxy
    character(len=maxlen_IntAs) :: str_IntAs

    write(fileunit,"(a)",err=999) "REALACC(12)"

    call SetFAdoFull_or_Missing(Mie_PCD_Params%Alpha_Correction,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Alpha_Correction = "//trim(str_FAdoxy)


    call SetIntAs_or_Missing(Mie_PCD_Params%Summation_Index,&
                             str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Summation_Index  = "//trim(str_IntAs)

    write(fileunit,"(a)",err=999) "ENDREALACC"

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_Mie_PCD_Params "//&
                          "error while writing to outputfile")
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Mie_PCD_Params
    !  #]
  subroutine Get_Mie_PCD_Params_from_xmltree(xml_tree, xml_path, &
                                             Mie_PCD_Params, error_flag)
    !  #[

    TYPE(xml_node_type),       pointer     :: xml_tree
    character(len=*),          intent(in)  :: xml_path
    TYPE(Mie_PCD_Params_Type), intent(out) :: Mie_PCD_Params
    integer,                   intent(out) :: error_flag

    ! local variables
    character(len=max_str_len)  :: path

    ! init
    error_flag = no_error

    IF (.not. associated(xml_tree)) THEN
       call logmsg(log_error,"ERROR in Get_Mie_PCD_Params_from_xmltree():"//&
            "the pointer xml_tree seems not associated !!!")
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! copying the Mie_PCD_Params XML elements
    !------------------------------

    path = trim(xml_path)//"/Alpha_Correction"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         Mie_PCD_Params%Alpha_Correction,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/Summation_Index"
    call GetIntAs_from_xml_tree(xml_tree,path,&
         Mie_PCD_Params%Summation_Index,error_flag)
    IF (error_flag .ne. no_error) return

    return

  end subroutine Get_Mie_PCD_Params_from_xmltree

    !  #]
  !-----------------------------------------------------
  subroutine Init_Mie_Core_Algorithm_Params(Mie_Core_Algorithm_Params)
    !  #[
    TYPE(Mie_Core_Algorithm_Params_type), intent(out) :: &
         Mie_Core_Algorithm_Params

    Mie_Core_Algorithm_Params%SNR_Threshold = missing_indicator_FAdoxy
    Mie_Core_Algorithm_Params%Start_FWHM    = missing_indicator_FAdoxy
    Mie_Core_Algorithm_Params%Residual_Error_Threshold = &
         missing_indicator_FAdoxy
    Mie_Core_Algorithm_Params%Max_Iterations_Lorentz_Fit = &
         missing_indicator_IntAs
    Mie_Core_Algorithm_Params%FWHM_Upper_Threshold = &
         missing_indicator_FAdoxy
    Mie_Core_Algorithm_Params%FWHM_Lower_Threshold = &
         missing_indicator_FAdoxy
    Mie_Core_Algorithm_Params%Peak_Height_Upper_Threshold = &
         missing_indicator_FAdoxy
    Mie_Core_Algorithm_Params%Peak_Height_Lower_Threshold = &
         missing_indicator_FAdoxy
    Mie_Core_Algorithm_Params%Peak_Location_Threshold = &
         missing_indicator_FAdoxy
    Mie_Core_Algorithm_Params%Nonlinear_Optimization_Thresh = &
         missing_indicator_FAdoxy
    Mie_Core_Algorithm_Params%Max_Iter_Nonlinear_Optimization = &
         missing_indicator_IntAs
    Mie_Core_Algorithm_Params%Num_Spectral_Sub_Samples = &
         missing_indicator_IntAs

  end subroutine Init_Mie_Core_Algorithm_Params
    !  #]
  subroutine AsciiWrite_Mie_Core_Algo_Params(Mie_Core_Algorithm_Params,&
                                             fileunit, prefix, error_flag)
    !  #[
    TYPE(Mie_Core_Algorithm_Params_type), &
                      intent(in)  :: Mie_Core_Algorithm_Params
    integer,          intent(in)  :: fileunit
    character(len=*), intent(in)  :: prefix
    integer,          intent(out) :: error_flag

    ! local variables
    character(len=25)           :: str_FAdoxy
    character(len=maxlen_IntAs) :: str_IntAs

    write(fileunit,"(a)",err=999) "REALACC(12)"

    call SetFAdoFull_or_Missing(Mie_Core_Algorithm_Params%SNR_Threshold,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%SNR_Threshold                   = "//trim(str_FAdoxy)
! this one is currently not used
!         trim(str_FAdoxy)//" ["//unit_SNR_Threshold//"]"

    call SetFAdoFull_or_Missing(Mie_Core_Algorithm_Params%Start_FWHM,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Start_FWHM                      = "//trim(str_FAdoxy)
! TODO: activate use of this unit
!         trim(str_FAdoxy)//" ["//unit_Start_FVWM//"]"

    call SetFAdoFull_or_Missing(&
         Mie_Core_Algorithm_Params%Residual_Error_Threshold,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Residual_Error_Threshold        = "//trim(str_FAdoxy)
! TODO: activate use of this unit
!         trim(str_FAdoxy)//" ["//unit_Residual_Error_Threshold//"]"

    call SetIntAs_or_Missing(&
         Mie_Core_Algorithm_Params%Max_Iterations_Lorentz_Fit,&
         str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Max_Iterations_Lorentz_Fit      = "//trim(str_IntAs)

    call SetFAdoFull_or_Missing(&
         Mie_Core_Algorithm_Params%FWHM_Upper_Threshold,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%FWHM_Upper_Threshold            = "//trim(str_FAdoxy)
! TODO: activate use of this unit
!         trim(str_FAdoxy)//" ["//unit_Upper_Threshold//"]"

    call SetFAdoFull_or_Missing(&
         Mie_Core_Algorithm_Params%FWHM_Lower_Threshold,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%FWHM_Lower_Threshold            = "//trim(str_FAdoxy)
! TODO: activate use of this unit
!         trim(str_FAdoxy)//" ["//unit_Lower_Threshold//"]"

    call SetFAdoFull_or_Missing(&
         Mie_Core_Algorithm_Params%Peak_Height_Upper_Threshold,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Peak_Height_Upper_Threshold     = "//trim(str_FAdoxy)
! TODO: activate use of this unit
!         trim(str_FAdoxy)//" ["//unit_Peak_Height_Upper_Thresh//"]"

    call SetFAdoFull_or_Missing(&
         Mie_Core_Algorithm_Params%Peak_Height_Lower_Threshold,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Peak_Height_Lower_Threshold     = "//trim(str_FAdoxy)
! TODO: activate use of this unit
!         trim(str_FAdoxy)//" ["//unit_Peak_Height_Lower_Thresh//"]"

    call SetFAdoFull_or_Missing(&
         Mie_Core_Algorithm_Params%Peak_Location_Threshold,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Peak_Location_Threshold         = "//trim(str_FAdoxy)
! TODO: activate use of this unit
!         trim(str_FAdoxy)//" ["//unit_Peak_Location_Threshold//"]"

    call SetFAdoFull_or_Missing(&
         Mie_Core_Algorithm_Params%Nonlinear_Optimization_Thresh,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Nonlinear_Optimization_Thresh   = "//trim(str_FAdoxy)
! TODO: activate use of this unit
!         trim(str_FAdoxy)//" ["//unit_Nonlinear_Opt_Threshold//"]"

    call SetIntAs_or_Missing(&
         Mie_Core_Algorithm_Params%Max_Iter_Nonlinear_Optimization,&
         str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Max_Iter_Nonlinear_Optimization = "//trim(str_IntAs)
! TODO: activate use of this unit
!         trim(str_FAdoxy)//" ["//unit_MaxIter_Nonlin_Opt_Thresh//"]"

    call SetIntAs_or_Missing(&
         Mie_Core_Algorithm_Params%Num_Spectral_Sub_Samples,&
         str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Num_Spectral_Sub_Samples        = "//trim(str_IntAs)
! TODO: activate use of this unit
!         trim(str_FAdoxy)//" ["//unit_Num_Spectral_Sub_Samples//"]"

    write(fileunit,"(a)",err=999) "ENDREALACC"

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_Mie_Core_Algo_Params "//&
                          "error while writing to outputfile")
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Mie_Core_Algo_Params
    !  #]
  subroutine GetMieCoreAlgoPars_from_xmltree(xml_tree, xml_path, &
                            Mie_Core_Algorithm_Params, error_flag)
    !  #[
    TYPE(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: xml_path
    TYPE(Mie_Core_Algorithm_Params_Type), &
                         intent(out) :: Mie_Core_Algorithm_Params
    integer,             intent(out) :: error_flag

    ! local variables
    character(len=max_str_len)  :: path

    ! init
    error_flag = no_error

    IF (.not. associated(xml_tree)) THEN
       call logmsg(log_error,"ERROR in GetMieCoreAlgoPars_from_xmltree():"//&
            "the pointer xml_tree seems not associated !!!")
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! copying the Mie_Core_Algorithm_Params XML elements
    !------------------------------

    path = trim(xml_path)//"/SNR_Threshold"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         Mie_Core_Algorithm_Params%SNR_Threshold,error_flag)
! this one is currently not used
!         unit=unit_SNR_Threshold)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/Start_FWHM"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         Mie_Core_Algorithm_Params%Start_FWHM,error_flag,&
         unit=unit_Start_FWHM)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/Residual_Error_Threshold"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         Mie_Core_Algorithm_Params%Residual_Error_Threshold,error_flag,&
         unit=unit_Residual_Error_Threshold)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/Max_Iterations_Lorentz_Fit"
    call GetIntAs_from_xml_tree(xml_tree,path,&
         Mie_Core_Algorithm_Params%Max_Iterations_Lorentz_Fit,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/FWHM_Upper_Threshold"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         Mie_Core_Algorithm_Params%FWHM_Upper_Threshold,error_flag,&
         unit=unit_Upper_Threshold)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/FWHM_Lower_Threshold"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         Mie_Core_Algorithm_Params%FWHM_Lower_Threshold,error_flag,&
         unit=unit_Lower_Threshold)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/Peak_Height_Upper_Threshold"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         Mie_Core_Algorithm_Params%Peak_Height_Upper_Threshold,error_flag,&
         unit=unit_Peak_Height_Upper_Thresh)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/Peak_Height_Lower_Threshold"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         Mie_Core_Algorithm_Params%Peak_Height_Lower_Threshold,error_flag,&
         unit=unit_Peak_Height_Lower_Thresh)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/Peak_Location_Threshold"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         Mie_Core_Algorithm_Params%Peak_Location_Threshold,error_flag,&
         unit=unit_Peak_Location_Threshold)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/Nonlinear_Optimization_Threshold"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         Mie_Core_Algorithm_Params%Nonlinear_Optimization_Thresh,error_flag,&
         unit=unit_Nonlinear_Opt_Threshold)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/Max_Iterations_Nonlinear_Optimization"
    call GetIntAs_from_xml_tree(xml_tree,path,&
         Mie_Core_Algorithm_Params%Max_Iter_Nonlinear_Optimization,error_flag,&
         unit=unit_MaxIter_Nonlin_Opt_Thresh)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/Num_Spectral_Sub_Samples"
    call GetIntAs_from_xml_tree(xml_tree,path,&
         Mie_Core_Algorithm_Params%Num_Spectral_Sub_Samples,error_flag,&
         unit_Num_Spectral_Sub_Samples)
    IF (error_flag .ne. no_error) return

    return

  end subroutine GetMieCoreAlgoPars_from_xmltree
    !  #]
  !-----------------------------------------------------
  subroutine Init_Corrupt_Data_Detect_Params(Corrupt_Data_Detection_Params)
    !  #[
    TYPE(Corrupt_Data_Detect_Params_type), intent(out) :: &
         Corrupt_Data_Detection_Params

    Corrupt_Data_Detection_Params%Max_Signal_Derivative      = &
         missing_indicator_IntAus
    Corrupt_Data_Detection_Params%Pixel_Saturation_Threshold = &
         missing_indicator_IntAus

  end subroutine Init_Corrupt_Data_Detect_Params
    !  #]
  subroutine AsciiWrite_Corr_Data_Det_Pars(Corrupt_Data_Detection_Params,&
                                           fileunit, prefix, error_flag)
    !  #[
    TYPE(Corrupt_Data_Detect_Params_type), &
                               intent(in)  :: Corrupt_Data_Detection_Params
    integer,                   intent(in)  :: fileunit
    character(len=*),          intent(in)  :: prefix
    integer,                   intent(out) :: error_flag

    ! local variables
    character(len=maxlen_IntAus) :: str_IntAus

    call SetIntAus_or_Missing(&
         Corrupt_Data_Detection_Params%Max_Signal_Derivative,&
         str_IntAus,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Max_Signal_Derivative      = "//&
         trim(str_IntAus)//" ["//unit_Max_Signal_Derivative//"]"

    call SetIntAus_or_Missing(&
         Corrupt_Data_Detection_Params%Pixel_Saturation_Threshold,&
         str_IntAus,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Pixel_Saturation_Threshold = "//&
         trim(str_IntAus)//" ["//unit_Pixel_Saturation_Thresh//"]"

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_CorrDataDetectPars "//&
                          "error while writing to outputfile")
    error_flag = error_writing_file
    return
    
  end subroutine AsciiWrite_Corr_Data_Det_Pars
  !  #]
  subroutine GetCorrDataDetPars_from_xmltree(xml_tree, xml_path, &
                               Corrupt_Data_Detection_Params, error_flag)

    !  #[

    TYPE(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: xml_path
    TYPE(Corrupt_Data_Detect_Params_Type), &
                         intent(out) :: Corrupt_Data_Detection_Params
    integer,             intent(out) :: error_flag

    ! local variables
    character(len=max_str_len)  :: path

    ! init
    error_flag = no_error

    IF (.not. associated(xml_tree)) THEN
       call logmsg(log_error,"ERROR in GetCorrDataDetPars_from_xmltree():"//&
            "the pointer xml_tree seems not associated !!!")
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! copying the Corrupt_Data_Detection_Params XML elements
    !------------------------------

    path = trim(xml_path)//"/Max_Signal_Derivative"
    call GetIntAus_from_xml_tree(xml_tree,path,&
         Corrupt_Data_Detection_Params%Max_Signal_Derivative,error_flag,&
         unit=unit_Max_Signal_Derivative)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/Pixel_Saturation_Threshold"
    call GetIntAus_from_xml_tree(xml_tree,path,&
         Corrupt_Data_Detection_Params%Pixel_Saturation_Threshold,error_flag,&
         unit=unit_Pixel_Saturation_Thresh)
    IF (error_flag .ne. no_error) return

    return

  end subroutine GetCorrDataDetPars_from_xmltree

    !  #]
  !-----------------------------------------------------
  subroutine Init_Common_Processing_Params(Common_Processing_Params)
    !  #[
    TYPE(Common_Processing_Params_type), intent(out) :: &
         Common_Processing_Params
    
    call Init_Mie_PCD_Params(Common_Processing_Params%Mie_PCD_Params)
    call Init_Mie_Core_Algorithm_Params(&
                  Common_Processing_Params%Mie_Core_Algorithm_Params)
    call Init_Corrupt_Data_Detect_Params(&
                  Common_Processing_Params%Corrupt_Data_Detection_Params)

  end subroutine Init_Common_Processing_Params
    !  #]
  subroutine AsciiWrite_Common_Proc_Params(Common_Processing_Params,fileunit, &
                                           prefix, error_flag)
    !  #[
    TYPE(Common_Processing_Params_type), &
                      intent(in)  :: Common_Processing_Params
    integer,          intent(in)  :: fileunit
    character(len=*), intent(in)  :: prefix
    integer,          intent(out) :: error_flag

    error_flag = no_error
    
    call AsciiWrite_Mie_PCD_Params(Common_Processing_Params%Mie_PCD_Params,&
         fileunit, trim(prefix)//"%Mie_PCD_Params", error_flag)
    IF (error_flag .ne. no_error) return

    call AsciiWrite_Mie_Core_Algo_Params(&
         Common_Processing_Params%Mie_Core_Algorithm_Params,&
         fileunit, trim(prefix)//"%Mie_Core_Algorithm_Params", error_flag)
    IF (error_flag .ne. no_error) return

    call AsciiWrite_Corr_Data_Det_Pars(&
         Common_Processing_Params%Corrupt_Data_Detection_Params,&
         fileunit, trim(prefix)//"%Corrupt_Data_Detection_Params", error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine AsciiWrite_Common_Proc_Params
    !  #]
  subroutine Get_CommProcParams_from_xmltree(xml_tree, xml_path, &
                                      Common_Processing_Params, error_flag)
    !  #[
    TYPE(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: xml_path
    TYPE(Common_Processing_Params_type), &
                         intent(out) :: Common_Processing_Params
    integer,             intent(out) :: error_flag

    ! init
    error_flag = no_error

    IF (.not. associated(xml_tree)) THEN
       call logmsg(log_error,"ERROR in Get_CommProcParams_from_xmltree():"//&
            "the pointer xml_tree seems not associated !!!")
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! copying the Common_Processing_Params XML elements
    !------------------------------

    call Get_Mie_PCD_Params_from_xmltree(xml_tree, &
         trim(xml_path)//"/Mie_PCD_Params", &
         Common_Processing_Params%Mie_PCD_Params, error_flag)
    IF (error_flag .ne. no_error) return

    call GetMieCoreAlgoPars_from_xmltree(xml_tree, &
         trim(xml_path)//"/Mie_Core_Algorithm_Params", &
         Common_Processing_Params%Mie_Core_Algorithm_Params, error_flag)
    IF (error_flag .ne. no_error) return

    call GetCorrDataDetPars_from_xmltree(xml_tree, &
         trim(xml_path)//"/Corrupt_Data_Detection_Params", &
         Common_Processing_Params%Corrupt_Data_Detection_Params, error_flag)
    IF (error_flag .ne. no_error) return

    return

  end subroutine Get_CommProcParams_from_xmltree
    !  #]
  !-----------------------------------------------------
  subroutine Init_RBC_Algorithm_Params(RBC_Algorithm_Params)
    !  #[
    TYPE(RBC_Algorithm_Params_type), intent(out) :: RBC_Algorithm_Params

    RBC_Algorithm_Params%Do_Mie_Decontamination = .false.

  end subroutine Init_RBC_Algorithm_Params
    !  #]
  subroutine AsciiWrite_RBC_Algorithm_Params(RBC_Algorithm_Params, &
                                             fileunit, prefix, error_flag)
    !  #[
    TYPE(RBC_Algorithm_Params_type), &
                      intent(in)  :: RBC_Algorithm_Params
    integer,          intent(in)  :: fileunit
    character(len=*), intent(in)  :: prefix
    integer,          intent(out) :: error_flag

    ! local variables
    character(len=256)          :: txt

    IF (RBC_Algorithm_Params%Do_Mie_Decontamination) THEN
       txt = "true"
    ELSE
       txt = "false"
    END IF
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Do_Mie_Decontamination = "//trim(txt)

    ! 500 spare bytes

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_RBC_Algorithm_Params "//&
                          "error while writing to outputfile")
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_RBC_Algorithm_Params
    !  #]
  subroutine Get_RBC_Alg_Pars_from_xml_tree(xml_tree, xml_path, &
                                            RBC_Algorithm_Params, error_flag)
    !  #[
    TYPE(xml_node_type),             pointer     :: xml_tree
    character(len=*),                intent(in)  :: xml_path
    TYPE(RBC_Algorithm_Params_Type), intent(out) :: RBC_Algorithm_Params
    integer,                         intent(out) :: error_flag

    ! local variables
    character(len=max_str_len)  :: path
    character(len=max_str_len)  :: text

    ! init
    error_flag = no_error

    IF (.not. associated(xml_tree)) THEN
       call logmsg(log_error,"ERROR in Get_RBC_Alg_Pars_from_xml_tree():"//&
            "the pointer xml_tree seems not associated !!!")
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! copying the RBC_Algorithm_Params XML elements
    !------------------------------

    path = trim(xml_path)//"/Do_Mie_Decontamination"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return

    select case (to_lowercase(trim(text)))
    case("true") 
       RBC_Algorithm_Params%Do_Mie_Decontamination = .true.
    case("false") 
       RBC_Algorithm_Params%Do_Mie_Decontamination = .false.
    case default
       call logmsg(log_error,"ERROR in Get_RBC_Alg_Pars_from_xml_tree():"//&
            "expected a value True or False, but detected a value ["//&
            trim(text)//"] for item ["//trim(path)//"]")
       error_flag = error_reading_file
       IF (error_flag .ne. no_error) return
    end select

    return

  end subroutine Get_RBC_Alg_Pars_from_xml_tree
    !  #]
  !-----------------------------------------------------
  subroutine Init_AMD_Matchup_Params(AMD_Matchup_Params)
    !  #[
    TYPE(AMD_Matchup_Params_type), intent(out) :: AMD_Matchup_Params

    AMD_Matchup_Params%Matchup_Method        = Matchup_Method_undefined
    AMD_Matchup_Params%Max_Allowed_Time_Diff = 0
    AMD_Matchup_Params%Max_Allowed_Distance  = 0

  end subroutine Init_AMD_Matchup_Params
    !  #]
  subroutine AsciiWrite_AMD_Matchup_Params(AMD_Matchup_Params, &
                                           fileunit, prefix, error_flag)
    !  #[
    TYPE(AMD_Matchup_Params_type), &
                      intent(in)  :: AMD_Matchup_Params
    integer,          intent(in)  :: fileunit
    character(len=*), intent(in)  :: prefix
    integer,          intent(out) :: error_flag

    ! local variables
    character(len=256)           :: txt
    character(len=maxlen_IntAul) :: str_IntAul

    select case (AMD_Matchup_Params%Matchup_Method)
    case (Matchup_Method_undefined)
       txt = "Undefined"
    case (Matchup_Dummy)
       txt = "Dummy"
    case (Matchup_Nearest_Neighbour)
       txt = "Nearest_Neighbour"
    case default 
       txt = "undefined matchup method"
    end select

    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Matchup_Method        = "//trim(txt)

    call SetIntAul_or_Missing(AMD_Matchup_Params%Max_Allowed_Time_Diff,&
                              str_IntAul,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Max_Allowed_Time_Diff = "//trim(str_IntAul)//&
         " ["//trim(unit_Coll_Time_Diff)//"]"

    call SetIntAul_or_Missing(AMD_Matchup_Params%Max_Allowed_Distance,&
                              str_IntAul,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Max_Allowed_Distance  = "//trim(str_IntAul)//&
         " ["//trim(unit_Coll_Dist_Thresh)//"]"

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_AMD_Matchup_Params "//&
                          "error while writing to outputfile")
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_AMD_Matchup_Params
    !  #]
  subroutine Get_AMDMatchupPars_fr_xml_tree(xml_tree, xml_path, &
                                            AMD_Matchup_Params, error_flag)
    !  #[
    TYPE(xml_node_type),           pointer     :: xml_tree
    character(len=*),              intent(in)  :: xml_path
    TYPE(AMD_Matchup_Params_Type), intent(out) :: AMD_Matchup_Params
    integer,                       intent(out) :: error_flag

    ! local variables
    character(len=max_str_len)  :: path
    character(len=max_str_len)  :: text

    ! init
    error_flag = no_error

    IF (.not. associated(xml_tree)) THEN
       call logmsg(log_error,"ERROR in Get_AMDMatchupPars_fr_xml_tree():"//&
            "the pointer xml_tree seems not associated !!!")
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! copying the AMD_Matchup_Params XML elements
    !------------------------------

    path = trim(xml_path)//"/Matchup_Method"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return

    select case (trim(text))
    case("Dummy") 
       AMD_Matchup_Params%Matchup_Method = Matchup_Dummy
    case("Nearest_Neighbour") 
       AMD_Matchup_Params%Matchup_Method = Matchup_Nearest_Neighbour
    case default
       call logmsg(log_error,"ERROR in Get_AMDMatchupPars_fr_xml_tree():"//&
            "unexpected value ["//&
            trim(text)//"] for item ["//trim(path)//"]")
       error_flag = error_reading_file
       IF (error_flag .ne. no_error) return
    end select

    path = trim(xml_path)//"/Max_Allowed_Time_Diff"
    call GetIntAul_from_xml_tree(xml_tree,path,&
         AMD_Matchup_Params%Max_Allowed_Time_Diff,error_flag,&
         unit=unit_Coll_Time_Diff)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/Max_Allowed_Distance"
    call GetIntAul_from_xml_tree(xml_tree,path,&
         AMD_Matchup_Params%Max_Allowed_Distance,error_flag,&
         unit=unit_Coll_Dist_Thresh)
    IF (error_flag .ne. no_error) return

    return

  end subroutine Get_AMDMatchupPars_fr_xml_tree
    !  #]
  !-----------------------------------------------------
  subroutine Init_CLM_Matchup_Params(CLM_Matchup_Params)
    !  #[
    TYPE(CLM_Matchup_Params_type), intent(out) :: CLM_Matchup_Params

    CLM_Matchup_Params%Matchup_Method        = CLM_Matchup_Method_undefined

  end subroutine Init_CLM_Matchup_Params
    !  #]
  subroutine AsciiWrite_CLM_Matchup_Params(CLM_Matchup_Params, &
                                           fileunit, prefix, error_flag)
    !  #[
    TYPE(CLM_Matchup_Params_type), &
                      intent(in)  :: CLM_Matchup_Params
    integer,          intent(in)  :: fileunit
    character(len=*), intent(in)  :: prefix
    integer,          intent(out) :: error_flag

    ! local variables
    character(len=256)           :: txt

    select case (CLM_Matchup_Params%Matchup_Method)
    case (CLM_Matchup_Method_undefined)
       txt = "Undefined"
    case (CLM_Matchup_Dummy)
       txt = "Dummy"
    case (CLM_Matchup_Nearest_Neighbour)
       txt = "Nearest_Neighbour"
    case default 
       txt = "undefined matchup method"
    end select

    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Matchup_Method        = "//trim(txt)

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_CLM_Matchup_Params "//&
                          "error while writing to outputfile")
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_CLM_Matchup_Params
    !  #]
  subroutine Get_CLMMatchupPars_fr_xml_tree(xml_tree, xml_path, &
                                            CLM_Matchup_Params, error_flag)
    !  #[
    TYPE(xml_node_type),           pointer     :: xml_tree
    character(len=*),              intent(in)  :: xml_path
    TYPE(CLM_Matchup_Params_Type), intent(out) :: CLM_Matchup_Params
    integer,                       intent(out) :: error_flag

    ! local variables
    character(len=max_str_len)  :: path
    character(len=max_str_len)  :: text

    ! init
    error_flag = no_error

    IF (.not. associated(xml_tree)) THEN
       call logmsg(log_error,"ERROR in Get_CLMMatchupPars_fr_xml_tree():"//&
            "the pointer xml_tree seems not associated !!!")
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! copying the CLM_Matchup_Params XML elements
    !------------------------------

    path = trim(xml_path)//"/Matchup_Method"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return

    select case (trim(text))
    case("Dummy") 
       CLM_Matchup_Params%Matchup_Method = CLM_Matchup_Dummy
    case("Nearest_Neighbour") 
       CLM_Matchup_Params%Matchup_Method = CLM_Matchup_Nearest_Neighbour
    case default
       call logmsg(log_error,"ERROR in Get_CLMMatchupPars_fr_xml_tree():"//&
            "unexpected value ["//&
            trim(text)//"] for item ["//trim(path)//"]")
       error_flag = error_reading_file
       IF (error_flag .ne. no_error) return
    end select

    return

  end subroutine Get_CLMMatchupPars_fr_xml_tree
    !  #]
  !-----------------------------------------------------
  subroutine Init_WVM_Params(WVM_Params)
    !  #[
    TYPE(WVM_Params_Type), intent(out) :: WVM_Params

    WVM_Params%File_Type(:)                    = ' '
    WVM_Params%Sph_Descriptor(:)               = ' '
    WVM_Params%Warmup_Pulse_Min_Height         = missing_indicator_FAdoxy
    WVM_Params%Rangebin_Mismatch_Tolerance     = missing_indicator_FAdoxy
    WVM_Params%Line_of_Sight_Wind_Flag         = .false.
    WVM_Params%Mie_GroundCorr_WeightingFactor  = missing_indicator_FAdoxy
    WVM_Params%Rayl_GroundCorr_WeightingFactor = missing_indicator_FAdoxy
    WVM_Params%N_Obs_Mie_Max                   = missing_indicator_IntAs
    WVM_Params%N_Obs_Rayleigh_Max              = missing_indicator_IntAs

    call Init_Classification_Params(WVM_Params%Classification_Params)
    call Init_Optical_Properties_Params(WVM_Params%Optical_Properties_Params)
    call Init_Error_Quantifier_Params(WVM_Params%Error_Quantifier_Params)
    call Init_Common_Processing_Params(WVM_Params%Common_Processing_Params)
    call Init_RBC_Algorithm_Params(WVM_Params%RBC_Algorithm_Params)
    call Init_AMD_Matchup_Params(WVM_Params%AMD_Matchup_Params)
    call Init_CLM_Matchup_Params(WVM_Params%CLM_Matchup_Params)

  end subroutine Init_WVM_Params
    !  #]
  subroutine Delete_WVM_Params(WVM_Params)
    !  #[
    TYPE(WVM_Params_Type), intent(inout) :: WVM_Params

    call Delete_Classification_Params(WVM_Params%Classification_Params)

  end subroutine Delete_WVM_Params
    !  #]
  subroutine AsciiWrite_WVM_Params(WVM_Params, fileunit, &
                                   prefix, error_flag)
    !  #[
    TYPE(WVM_Params_type), intent(in)  :: WVM_Params
    integer,               intent(in)  :: fileunit
    character(len=*),      intent(in)  :: prefix
    integer,               intent(out) :: error_flag
   
    ! local variables
    character(len=256)          :: txt
    character(len=25)           :: str_FAdoxy
    character(len=maxlen_IntAs) :: str_IntAs

    error_flag = no_error

    write(fileunit,"(a)",err=999) "REALACC(12)"

    txt = WVM_Params%File_Type
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%File_Type                       = "//trim(txt)

    txt = WVM_Params%Sph_Descriptor
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Sph_Descriptor                  = "//trim(txt)

    call SetFAdoFull_or_Missing(WVM_Params%Warmup_Pulse_Min_Height,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Warmup_Pulse_Min_Height         = "//&
         trim(str_FAdoxy)//" ["//unit_Warmup_Pulse_Min_Height//"]"

    call SetFAdoFull_or_Missing(WVM_Params%Rangebin_Mismatch_Tolerance,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Rangebin_Mismatch_Tolerance     = "//&
         trim(str_FAdoxy)//" ["//unit_Rangebin_Mismatch_Toler//"]"

    IF (WVM_Params%Line_of_Sight_Wind_Flag) THEN
       txt = "true (so reporting LOS wind)"
    ELSE
       txt = "false (so reporting HLOS wind)"
    END IF
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Line_of_Sight_Wind_Flag         = "//trim(txt)

    call SetFAdoFull_or_Missing(WVM_Params%Mie_GroundCorr_WeightingFactor,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Mie_GroundCorr_WeightingFactor  = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(WVM_Params%Rayl_GroundCorr_WeightingFactor,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Rayl_GroundCorr_WeightingFactor = "//trim(str_FAdoxy)

    call SetIntAs_or_Missing(WVM_Params%N_Obs_Mie_Max,str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%N_Obs_Mie_Max                   = "//trim(str_IntAs)

    call SetIntAs_or_Missing(WVM_Params%N_Obs_Rayleigh_Max,&
                             str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%N_Obs_Rayleigh_Max              = "//trim(str_IntAs)

    write(fileunit,"(a)",err=999) "ENDREALACC"

    call AsciiWrite_Classific_Params(WVM_Params%Classification_Params, &
         fileunit, trim(prefix)//"%Classification_Params", error_flag)
    IF (error_flag .ne. no_error) return

    call AsciiWrite_Optical_Prop_Params(WVM_Params%Optical_Properties_Params, &
         fileunit, trim(prefix)//"%Optical_Properties_Params", error_flag)
    IF (error_flag .ne. no_error) return

    call AsciiWrite_Error_Quant_Params(&
         WVM_Params%Error_Quantifier_Params,fileunit,&
         trim(prefix)//"%Error_Quantifier_Params", error_flag)
    IF (error_flag .ne. no_error) return

    call AsciiWrite_Common_Proc_Params(&
         WVM_Params%Common_Processing_Params,fileunit,&
         trim(prefix)//"%Common_Processing_Params", error_flag)
    IF (error_flag .ne. no_error) return

    call AsciiWrite_RBC_Algorithm_Params(&
         WVM_Params%RBC_Algorithm_Params,fileunit,&
         trim(prefix)//"%RBC_Algorithm_Params", error_flag)
    IF (error_flag .ne. no_error) return

    call AsciiWrite_AMD_Matchup_Params(&
         WVM_Params%AMD_Matchup_Params,fileunit,&
         trim(prefix)//"%AMD_Matchup_Params", error_flag)
    IF (error_flag .ne. no_error) return

    call AsciiWrite_CLM_Matchup_Params(&
         WVM_Params%CLM_Matchup_Params,fileunit,&
         trim(prefix)//"%CLM_Matchup_Params", error_flag)
    IF (error_flag .ne. no_error) return

    ! spare 500 bytes

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_WVM_Params "//&
                          "error while writing to outputfile")
    error_flag = error_writing_file
    return
   
  end subroutine AsciiWrite_WVM_Params
    !  #]
  subroutine Get_WVM_Params_from_xml_tree(xml_tree, xml_path, &
                                          WVM_Params, error_flag)
    !  #[
    TYPE(xml_node_type),   pointer     :: xml_tree
    character(len=*),      intent(in)  :: xml_path
    TYPE(WVM_Params_Type), intent(out) :: WVM_Params
    integer,               intent(out) :: error_flag

    ! local variables
    character(len=max_str_len)  :: path
    character(len=max_str_len)  :: text

    ! init
    error_flag = no_error

    IF (.not. associated(xml_tree)) THEN
       call logmsg(log_error,"ERROR in Get_WVM_Params_from_xml_tree():"//&
            "the pointer xml_tree seems not associated !!!")
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! copying the WVM_Params XML elements
    !------------------------------

    path = trim(xml_path)//"/File_Type"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    WVM_Params%File_Type = check_length_xml_item(text,path,&
         WVM_maxlen_File_Type,"Get_WVM_Params_from_xml_tree")

    path = trim(xml_path)//"/Sph_Descriptor"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    WVM_Params%Sph_Descriptor = check_length_xml_item(text,path,&
         WVM_maxlen_Sph_Descriptor,"Get_WVM_Params_from_xml_tree")

    path = trim(xml_path)//"/Warmup_Pulse_Min_Height"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         WVM_Params%Warmup_Pulse_Min_Height,error_flag,&
         unit=unit_Warmup_Pulse_Min_Height)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/Rangebin_Mismatch_Tolerance"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         WVM_Params%Rangebin_Mismatch_Tolerance,error_flag,&
         unit=unit_Rangebin_Mismatch_Toler)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/Line_of_Sight_Wind_Flag"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    select case (to_lowercase(trim(text)))
    case("true") 
       WVM_Params%Line_of_Sight_Wind_Flag = .true.
    case("false") 
       WVM_Params%Line_of_Sight_Wind_Flag = .false.
    case default
       call logmsg(log_error,"ERROR in Get_WVM_Params_from_xml_tree():"//&
            "expected a value True or False, but detected a value ["//&
            trim(text)//"] for item ["//trim(path)//"]")
    end select

    path = trim(xml_path)//"/Mie_Ground_Correction_Weighting_Factor"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         WVM_Params%Mie_GroundCorr_WeightingFactor,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/Rayleigh_Ground_Correction_Weighting_Factor"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         WVM_Params%Rayl_GroundCorr_WeightingFactor,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/N_Obs_Mie_Max"
    call GetIntAs_from_xml_tree(xml_tree,path,&
         WVM_Params%N_Obs_Mie_Max,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/N_Obs_Rayleigh_Max"
    call GetIntAs_from_xml_tree(xml_tree,path,&
         WVM_Params%N_Obs_Rayleigh_Max,error_flag)
    IF (error_flag .ne. no_error) return

    call Get_Class_Params_from_xml_tree(xml_tree, &
         trim(xml_path)//"/Classification_Params", &
         WVM_Params%Classification_Params, error_flag)
    IF (error_flag .ne. no_error) return

    call Get_OptPropPars_from_xml_tree(xml_tree, &
         trim(xml_path)//"/Optical_Properties_Params", &
         WVM_Params%Optical_Properties_Params, error_flag)
    IF (error_flag .ne. no_error) return

    call Get_ErrorQuantPars_from_xmltree(xml_tree, &
         trim(xml_path)//"/Error_Quantifier_Params", &
         WVM_Params%Error_Quantifier_Params, error_flag)
    IF (error_flag .ne. no_error) return

    call Get_CommProcParams_from_xmltree(xml_tree, &
         trim(xml_path)//"/Common_Processing_Params", &
         WVM_Params%Common_Processing_Params, error_flag)
    IF (error_flag .ne. no_error) return

    call Get_RBC_Alg_Pars_from_xml_tree(xml_tree, &
         trim(xml_path)//"/RBC_Algorithm_Params", &
         WVM_Params%RBC_Algorithm_Params, error_flag)
    IF (error_flag .ne. no_error) return

    call Get_AMDMatchupPars_fr_xml_tree(xml_tree, &
         trim(xml_path)//"/AMD_Matchup_Params", &
         WVM_Params%AMD_Matchup_Params, error_flag)
    IF (error_flag .ne. no_error) return

    call Get_CLMMatchupPars_fr_xml_tree(xml_tree, &
         trim(xml_path)//"/CLM_Matchup_Params", &
         WVM_Params%CLM_Matchup_Params, error_flag)
    IF (error_flag .ne. no_error) return

    return

  end subroutine Get_WVM_Params_from_xml_tree
    !  #]
  !  #]
  !-----------------------------------------------------
  !  #[ Screening_Params
  subroutine Init_L1B_Obs_Screening_Params(L1B_Obs_Screening_Params)
    !  #[
    TYPE(L1B_Obs_Scr_Params_type), &
         intent(out) :: L1B_Obs_Screening_Params

    L1B_Obs_Screening_Params%L1B_Laser_Freq_Unlocked_Thresh = &
         missing_indicator_IntAl
    L1B_Obs_Screening_Params%L1B_Ref_Pulses_Unlocked_Thresh = &
         missing_indicator_IntAl
    L1B_Obs_Screening_Params%L1B_Laser_Freq_Offset_Thresh = &
         missing_indicator_FAdoxy
    L1B_Obs_Screening_Params%L1B_Laser_UV_Energy_Thresh = &
         missing_indicator_FAdoxy
    L1B_Obs_Screening_Params%L1B_Laser_Freq_Offs_StdevThresh = &
         missing_indicator_FAdoxy
    L1B_Obs_Screening_Params%L1B_Laser_UV_Energy_StdevThresh = &
         missing_indicator_FAdoxy
    L1B_Obs_Screening_Params%L1B_Mie_Mean_Emit_Freq_Min = &
         missing_indicator_FAdoxy
    L1B_Obs_Screening_Params%L1B_Mie_Mean_Emit_Freq_Max = &
         missing_indicator_FAdoxy
    L1B_Obs_Screening_Params%L1B_Mie_Emit_Freq_Stdev_Thresh = &
         missing_indicator_FAdoxy
    L1B_Obs_Screening_Params%L1B_Rayleigh_Mean_Emit_Freq_Min = &
         missing_indicator_FAdoxy
    L1B_Obs_Screening_Params%L1B_Rayleigh_Mean_Emit_Freq_Max = &
         missing_indicator_FAdoxy
    L1B_Obs_Screening_Params%L1B_RayleighEmitFreqStdevThresh = &
         missing_indicator_FAdoxy
    L1B_Obs_Screening_Params%L1B_Sat_Not_on_Target_Thresh = &
         missing_indicator_IntAl
    L1B_Obs_Screening_Params%L1B_Mie_Corrupt_Thresh = &
         missing_indicator_IntAl
    L1B_Obs_Screening_Params%L1B_Rayleigh_Corrupt_Thresh = &
         missing_indicator_IntAl
    L1B_Obs_Screening_Params%L1B_Mie_Ref_PulsesCorruptThresh = &
         missing_indicator_IntAl
    L1B_Obs_Screening_Params%L1B_Rayl_RefPulsesCorruptThresh = &
         missing_indicator_IntAl
    L1B_Obs_Screening_Params%L1B_Mie_Invalid_Meas_Thresh = &
         missing_indicator_IntAl
    L1B_Obs_Screening_Params%L1B_Mie_Invalid_RefPulsesThresh = &
         missing_indicator_IntAl
    L1B_Obs_Screening_Params%L1B_Rayl_Invalid_Meas_Thresh = &
         missing_indicator_IntAl
    L1B_Obs_Screening_Params%L1B_Rayl_InvalidRefPulsesThresh = &
         missing_indicator_IntAl
    
  end subroutine Init_L1B_Obs_Screening_Params
    !  #]
  subroutine AsciiWrite_L1B_Obs_Scr_Params(&
                L1B_Obs_Screening_Params,fileunit, prefix, error_flag)
    !  #[
    TYPE(L1B_Obs_Scr_Params_type), intent(in)  :: L1B_Obs_Screening_Params
    integer,                       intent(in)  :: fileunit
    character(len=*),              intent(in)  :: prefix
    integer,                       intent(out) :: error_flag
    
    ! local variables
    character(len=25)           :: str_FAdoxy
    character(len=maxlen_IntAl) :: str_IntAl

    error_flag = no_error

    
    call SetIntAl_or_Missing(&
         L1B_Obs_Screening_Params%L1B_Laser_Freq_Unlocked_Thresh,&
         str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Laser_Freq_Unlocked_Thresh  = "//&
         trim(str_IntAl)

    call SetIntAl_or_Missing(&
         L1B_Obs_Screening_Params%L1B_Ref_Pulses_Unlocked_Thresh, &
         str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Ref_Pulses_Unlocked_Thresh  = "//&
         trim(str_IntAl)

    write(fileunit,"(a)",err=999) "REALACC(12)"

    call SetFAdoFull_or_Missing(&
         L1B_Obs_Screening_Params%L1B_Laser_Freq_Offset_Thresh,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Laser_Freq_Offset_Thresh    = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_Frequency//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Obs_Screening_Params%L1B_Laser_UV_Energy_Thresh,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Laser_UV_Energy_Thresh      = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_UV_Energy//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Obs_Screening_Params%L1B_Laser_Freq_Offs_StdevThresh,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Laser_Freq_Offs_StdevThresh = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_Frequency//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Obs_Screening_Params%L1B_Laser_UV_Energy_StdevThresh,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Laser_UV_Energy_StdevThresh = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_UV_Energy//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Obs_Screening_Params%L1B_Mie_Mean_Emit_Freq_Min,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Mie_Mean_Emit_Freq_Min      = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_Frequency//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Obs_Screening_Params%L1B_Mie_Mean_Emit_Freq_Max,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Mie_Mean_Emit_Freq_Max      = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_Frequency//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Obs_Screening_Params%L1B_Mie_Emit_Freq_Stdev_Thresh,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Mie_Emit_Freq_Stdev_Thresh  = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_Frequency//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Obs_Screening_Params%L1B_Rayleigh_Mean_Emit_Freq_Min,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Rayleigh_Mean_Emit_Freq_Min = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_Frequency//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Obs_Screening_Params%L1B_Rayleigh_Mean_Emit_Freq_Max,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Rayleigh_Mean_Emit_Freq_Max = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_Frequency//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Obs_Screening_Params%L1B_RayleighEmitFreqStdevThresh,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_RayleighEmitFreqStdevThresh = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_Frequency//"]"

    write(fileunit,"(a)",err=999) "ENDREALACC"

    call SetIntAl_or_Missing(&
         L1B_Obs_Screening_Params%L1B_Sat_Not_on_Target_Thresh,&
         str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Sat_Not_on_Target_Thresh    = "//&
         trim(str_IntAl)

    call SetIntAl_or_Missing(&
         L1B_Obs_Screening_Params%L1B_Mie_Corrupt_Thresh,&
         str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Mie_Corrupt_Thresh          = "//&
         trim(str_IntAl)

    call SetIntAl_or_Missing(&
         L1B_Obs_Screening_Params%L1B_Rayleigh_Corrupt_Thresh,&
         str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Rayleigh_Corrupt_Thresh     = "//&
         trim(str_IntAl)

    call SetIntAl_or_Missing(&
         L1B_Obs_Screening_Params%L1B_Mie_Ref_PulsesCorruptThresh,&
         str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Mie_Ref_PulsesCorruptThresh = "//&
         trim(str_IntAl)

    call SetIntAl_or_Missing(&
         L1B_Obs_Screening_Params%L1B_Rayl_RefPulsesCorruptThresh,&
         str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Rayl_RefPulsesCorruptThresh = "//&
         trim(str_IntAl)

    call SetIntAl_or_Missing(&
         L1B_Obs_Screening_Params%L1B_Mie_Invalid_Meas_Thresh,&
         str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Mie_Invalid_Meas_Thresh     = "//&
         trim(str_IntAl)

    call SetIntAl_or_Missing(&
         L1B_Obs_Screening_Params%L1B_Mie_Invalid_RefPulsesThresh,&
         str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Mie_Invalid_RefPulsesThresh = "//&
         trim(str_IntAl)

    call SetIntAl_or_Missing(&
         L1B_Obs_Screening_Params%L1B_Rayl_Invalid_Meas_Thresh,&
         str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Rayl_Invalid_Meas_Thresh    = "//&
         trim(str_IntAl)

    call SetIntAl_or_Missing(&
         L1B_Obs_Screening_Params%L1B_Rayl_InvalidRefPulsesThresh,&
         str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Rayl_InvalidRefPulsesThresh = "//&
         trim(str_IntAl)

    ! spare 1000 bytes

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_L1B_Obs_Scr_Params "//&
                          "error while writing to outputfile")
    error_flag = error_writing_file
    return
   
  end subroutine AsciiWrite_L1B_Obs_Scr_Params
    !  #]
  subroutine Get_L1BObsScrPars_from_xmltree(xml_tree, xml_path, &
                L1B_Obs_Screening_Params, error_flag)
    !  #[
    TYPE(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: xml_path
    TYPE(L1B_Obs_Scr_Params_type), &
                         intent(out) :: L1B_Obs_Screening_Params
    integer,             intent(out) :: error_flag

    ! local variables
    character(len=max_str_len)  :: path

    ! init
    error_flag = no_error

    IF (.not. associated(xml_tree)) THEN
       call logmsg(log_error,"ERROR in Get_L1BObsScrPars_from_xmltree():"//&
            "the pointer xml_tree seems not associated !!!")
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! copying the L1B_Obs_Screening_Params XML elements
    !------------------------------

    path = trim(xml_path)//"/L1B_Laser_Freq_Unlocked_Threshold"
    call GetIntAl_from_xml_tree(xml_tree,path,&
         L1B_Obs_Screening_Params%L1B_Laser_Freq_Unlocked_Thresh,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Ref_Pulses_Unlocked_Threshold"
    call GetIntAl_from_xml_tree(xml_tree,path,&
         L1B_Obs_Screening_Params%L1B_Ref_Pulses_Unlocked_Thresh,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Laser_Freq_Offset_Threshold"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Obs_Screening_Params%L1B_Laser_Freq_Offset_Thresh,&
         error_flag,unit=unit_Laser_Frequency)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Laser_UV_Energy_Threshold"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Obs_Screening_Params%L1B_Laser_UV_Energy_Thresh,&
         error_flag,unit=unit_Laser_UV_Energy)
    IF (error_flag .ne. no_error) return
    
    path = trim(xml_path)//"/L1B_Laser_Freq_Offs_Stdev_Threshold"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Obs_Screening_Params%L1B_Laser_Freq_Offs_StdevThresh,&
         error_flag,unit=unit_Laser_Frequency)
    IF (error_flag .ne. no_error) return
    
    path = trim(xml_path)//"/L1B_Laser_UV_Energy_Stdev_Threshold"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Obs_Screening_Params%L1B_Laser_UV_Energy_StdevThresh,&
         error_flag,unit=unit_Laser_UV_Energy)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Mie_Mean_Emit_Freq_Min"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Obs_Screening_Params%L1B_Mie_Mean_Emit_Freq_Min,&
         error_flag,unit=unit_Laser_Frequency)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Mie_Mean_Emit_Freq_Max"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Obs_Screening_Params%L1B_Mie_Mean_Emit_Freq_Max,&
         error_flag,unit=unit_Laser_Frequency)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Mie_Emit_Freq_Stdev_Threshold"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Obs_Screening_Params%L1B_Mie_Emit_Freq_Stdev_Thresh,&
         error_flag,unit=unit_Laser_Frequency)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Rayleigh_Mean_Emit_Freq_Min"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Obs_Screening_Params%L1B_Rayleigh_Mean_Emit_Freq_Min,&
         error_flag,unit=unit_Laser_Frequency)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Rayleigh_Mean_Emit_Freq_Max"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Obs_Screening_Params%L1B_Rayleigh_Mean_Emit_Freq_Max,&
         error_flag,unit=unit_Laser_Frequency)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Rayleigh_Emit_Freq_Stdev_Threshold"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Obs_Screening_Params%L1B_RayleighEmitFreqStdevThresh,&
         error_flag,unit=unit_Laser_Frequency)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Sat_Not_on_Target_Threshold"
    call GetIntAl_from_xml_tree(xml_tree,path,&
         L1B_Obs_Screening_Params%L1B_Sat_Not_on_Target_Thresh,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Mie_Corrupt_Threshold"
    call GetIntAl_from_xml_tree(xml_tree,path,&
         L1B_Obs_Screening_Params%L1B_Mie_Corrupt_Thresh,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Rayleigh_Corrupt_Threshold"
    call GetIntAl_from_xml_tree(xml_tree,path,&
         L1B_Obs_Screening_Params%L1B_Rayleigh_Corrupt_Thresh,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Mie_Ref_Pulses_Corrupt_Threshold"
    call GetIntAl_from_xml_tree(xml_tree,path,&
         L1B_Obs_Screening_Params%L1B_Mie_Ref_PulsesCorruptThresh,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Rayl_Ref_Pulses_Corrupt_Threshold"
    call GetIntAl_from_xml_tree(xml_tree,path,&
         L1B_Obs_Screening_Params%L1B_Rayl_RefPulsesCorruptThresh,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Mie_Invalid_Meas_Threshold"
    call GetIntAl_from_xml_tree(xml_tree,path,&
         L1B_Obs_Screening_Params%L1B_Mie_Invalid_Meas_Thresh,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Mie_Invalid_Ref_Pulses_Threshold"
    call GetIntAl_from_xml_tree(xml_tree,path,&
         L1B_Obs_Screening_Params%L1B_Mie_Invalid_RefPulsesThresh,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Rayl_Invalid_Meas_Threshold"
    call GetIntAl_from_xml_tree(xml_tree,path,&
         L1B_Obs_Screening_Params%L1B_Rayl_Invalid_Meas_Thresh,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Rayl_Invalid_Ref_Pulses_Threshold"
    call GetIntAl_from_xml_tree(xml_tree,path,&
         L1B_Obs_Screening_Params%L1B_Rayl_InvalidRefPulsesThresh,error_flag)
    IF (error_flag .ne. no_error) return

    return

  end subroutine Get_L1BObsScrPars_from_xmltree
    !  #]
  !-----------------------------------------------------
  subroutine Init_L1B_Mie_Meas_Scr_Params(L1B_Mie_Meas_Screening_Params)
    !  #[
    TYPE(L1B_Mie_Meas_Scr_Params_type), &
         intent(out) :: L1B_Mie_Meas_Screening_Params

    L1B_Mie_Meas_Screening_Params%L1B_MieMeasInvalidRefPlssThresh = &
         missing_indicator_IntAl
    L1B_Mie_Meas_Screening_Params%L1B_Avg_Laser_Freq_Offset_Min = &
         missing_indicator_FAdoxy
    L1B_Mie_Meas_Screening_Params%L1B_Avg_Laser_Freq_Offset_Max = &
         missing_indicator_FAdoxy
    L1B_Mie_Meas_Screening_Params%L1B_Avg_UV_Energy_Min = &
         missing_indicator_FAdoxy
    L1B_Mie_Meas_Screening_Params%L1B_Avg_UV_Energy_Max = &
         missing_indicator_FAdoxy
    L1B_Mie_Meas_Screening_Params%L1B_Laser_FreqOffsetStdevThresh = &
         missing_indicator_FAdoxy
    L1B_Mie_Meas_Screening_Params%L1B_UV_Energy_Stdev_Thresh = &
         missing_indicator_FAdoxy
    L1B_Mie_Meas_Screening_Params%L1B_VelofAttUncertaintyErrorMin = &
         missing_indicator_FAdoxy
    L1B_Mie_Meas_Screening_Params%L1B_VelofAttUncertaintyErrorMax = &
         missing_indicator_FAdoxy
    L1B_Mie_Meas_Screening_Params%L1B_Mie_Mean_Emitted_Freq_Min = &
         missing_indicator_FAdoxy
    L1B_Mie_Meas_Screening_Params%L1B_Mie_Mean_Emitted_Freq_Max = &
         missing_indicator_FAdoxy
    L1B_Mie_Meas_Screening_Params%L1B_Mie_Emitted_FreqStdevThresh = &
         missing_indicator_FAdoxy
    L1B_Mie_Meas_Screening_Params%L1B_Scattering_Ratio_Min = &
         missing_indicator_FAdoxy
    L1B_Mie_Meas_Screening_Params%L1B_Scattering_Ratio_Max = &
         missing_indicator_FAdoxy
    L1B_Mie_Meas_Screening_Params%L1B_Mie_SNR_Thresh = &
         missing_indicator_FAdoxy
    L1B_Mie_Meas_Screening_Params%Mie_Ground_Bin_Thickness_Thresh = &
         missing_indicator_FAdoxy
    L1B_Mie_Meas_Screening_Params%Max_Signal_Derivative = &
         missing_indicator_IntAs
    L1B_Mie_Meas_Screening_Params%Pixel_Saturation_Thresh = &
         missing_indicator_IntAs

  end subroutine Init_L1B_Mie_Meas_Scr_Params
    !  #]         
  subroutine AsciiWrite_L1B_MieMeasScrParams(&
                L1B_Mie_Meas_Screening_Params,fileunit, prefix, error_flag)
    !  #[
    TYPE(L1B_Mie_Meas_Scr_Params_type), &
                      intent(in)  :: L1B_Mie_Meas_Screening_Params
    integer,          intent(in)  :: fileunit
    character(len=*), intent(in)  :: prefix
    integer,          intent(out) :: error_flag
    
    ! local variables
    character(len=25)           :: str_FAdoxy
    character(len=maxlen_IntAl) :: str_IntAl
    character(len=maxlen_IntAs) :: str_IntAs

    error_flag = no_error
    
    call SetIntAl_or_Missing(&
         L1B_Mie_Meas_Screening_Params%L1B_MieMeasInvalidRefPlssThresh, &
         str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_MieMeasInvalidRefPlssThresh = "//&
         trim(str_IntAl)

    write(fileunit,"(a)",err=999) "REALACC(12)"

    call SetFAdoFull_or_Missing(&
         L1B_Mie_Meas_Screening_Params%L1B_Avg_Laser_Freq_Offset_Min,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Avg_Laser_Freq_Offset_Min   = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_Frequency//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Mie_Meas_Screening_Params%L1B_Avg_Laser_Freq_Offset_Max,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Avg_Laser_Freq_Offset_Max   = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_Frequency//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Mie_Meas_Screening_Params%L1B_Avg_UV_Energy_Min,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Avg_UV_Energy_Min           = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_UV_Energy//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Mie_Meas_Screening_Params%L1B_Avg_UV_Energy_Max,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Avg_UV_Energy_Max           = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_UV_Energy//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Mie_Meas_Screening_Params%L1B_Laser_FreqOffsetStdevThresh, &
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Laser_FreqOffsetStdevThresh = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_Frequency//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Mie_Meas_Screening_Params%L1B_UV_Energy_Stdev_Thresh, &
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_UV_Energy_Stdev_Thresh      = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_UV_Energy//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Mie_Meas_Screening_Params%L1B_VelofAttUncertaintyErrorMin, &
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_VelofAttUncertaintyErrorMin = "//&
         trim(str_FAdoxy)//" ["//unit_Vel_of_Att_uncertainty//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Mie_Meas_Screening_Params%L1B_VelofAttUncertaintyErrorMax, &
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_VelofAttUncertaintyErrorMax = "//&
         trim(str_FAdoxy)//" ["//unit_Vel_of_Att_uncertainty//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Mie_Meas_Screening_Params%L1B_Mie_Mean_Emitted_Freq_Min, &
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Mie_Mean_Emitted_Freq_Min   = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_Frequency//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Mie_Meas_Screening_Params%L1B_Mie_Mean_Emitted_Freq_Max, &
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Mie_Mean_Emitted_Freq_Max   = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_Frequency//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Mie_Meas_Screening_Params%L1B_Mie_Emitted_FreqStdevThresh, &
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Mie_Emitted_FreqStdevThresh = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_Frequency//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Mie_Meas_Screening_Params%L1B_Scattering_Ratio_Min, &
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Scattering_Ratio_Min        = "//&
         trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(&
         L1B_Mie_Meas_Screening_Params%L1B_Scattering_Ratio_Max, &
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Scattering_Ratio_Max        = "//&
         trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(&
         L1B_Mie_Meas_Screening_Params%L1B_Mie_SNR_Thresh, &
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Mie_SNR_Thresh              = "//&
         trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(&
         L1B_Mie_Meas_Screening_Params%Mie_Ground_Bin_Thickness_Thresh, &
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Mie_Ground_Bin_Thickness_Thresh = "//&
         trim(str_FAdoxy)//" ["//unit_bin_thickness_thresh//"]"

    write(fileunit,"(a)",err=999) "ENDREALACC"

    call SetIntAs_or_Missing(&
         L1B_Mie_Meas_Screening_Params%Max_Signal_Derivative,&
         str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Max_Signal_Derivative           = "//&
         trim(str_IntAs)//" ["//unit_Max_Signal_Derivative//"]"

    call SetIntAs_or_Missing(&
         L1B_Mie_Meas_Screening_Params%Pixel_Saturation_Thresh,&
         str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Pixel_Saturation_Thresh         = "//&
         trim(str_IntAs)//" ["//unit_Pixel_Saturation_Thresh//"]"

    ! spare 1000 bytes

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_L1B_MieMeasScrParams "//&
                          "error while writing to outputfile")
    error_flag = error_writing_file
    return
   
  end subroutine AsciiWrite_L1B_MieMeasScrParams
    !  #]
  subroutine GetL1BMieMeasScrParsFromXmlTree(xml_tree, xml_path, &
                L1B_Mie_Meas_Screening_Params, error_flag)
    !  #[
    TYPE(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: xml_path
    TYPE(L1B_Mie_Meas_Scr_Params_type), &
                         intent(out) :: L1B_Mie_Meas_Screening_Params
    integer,             intent(out) :: error_flag

    ! local variables
    character(len=max_str_len)  :: path

    ! init
    error_flag = no_error

    IF (.not. associated(xml_tree)) THEN
       call logmsg(log_error,"ERROR in GetL1BMieMeasScrParsFromXmlTree():"//&
            "the pointer xml_tree seems not associated !!!")
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! copying the L1B_Mie_Meas_Screening_Params XML elements
    !------------------------------

    path = trim(xml_path)//"/L1B_Mie_Meas_Invalid_Ref_Pulses_Threshold"
    call GetIntAl_from_xml_tree(xml_tree,path,&
         L1B_Mie_Meas_Screening_Params%L1B_MieMeasInvalidRefPlssThresh,&
         error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Avg_Laser_Freq_Offset_Min"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Mie_Meas_Screening_Params%L1B_Avg_Laser_Freq_Offset_Min,&
         error_flag,unit=unit_Laser_Frequency)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Avg_Laser_Freq_Offset_Max"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Mie_Meas_Screening_Params%L1B_Avg_Laser_Freq_Offset_Max,&
         error_flag,unit=unit_Laser_Frequency)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Avg_UV_Energy_Min"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Mie_Meas_Screening_Params%L1B_Avg_UV_Energy_Min,&
         error_flag,unit=unit_Laser_UV_Energy)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Avg_UV_Energy_Max"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Mie_Meas_Screening_Params%L1B_Avg_UV_Energy_Max,&
         error_flag,unit=unit_Laser_UV_Energy)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Laser_Freq_Offset_Stdev_Threshold"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Mie_Meas_Screening_Params%L1B_Laser_FreqOffsetStdevThresh,&
         error_flag,unit=unit_Laser_Frequency)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_UV_Energy_Stdev_Threshold"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Mie_Meas_Screening_Params%L1B_UV_Energy_Stdev_Thresh,&
         error_flag,unit=unit_Laser_UV_Energy)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Vel_of_Att_Uncertainty_Error_Min"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Mie_Meas_Screening_Params%L1B_VelofAttUncertaintyErrorMin,&
         error_flag,unit=unit_Vel_of_Att_uncertainty)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Vel_of_Att_Uncertainty_Error_Max"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Mie_Meas_Screening_Params%L1B_VelofAttUncertaintyErrorMax,&
         error_flag,unit=unit_Vel_of_Att_uncertainty)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Mie_Mean_Emitted_Freq_Min"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Mie_Meas_Screening_Params%L1B_Mie_Mean_Emitted_Freq_Min,&
         error_flag,unit=unit_Laser_Frequency)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Mie_Mean_Emitted_Freq_Max"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Mie_Meas_Screening_Params%L1B_Mie_Mean_Emitted_Freq_Max,&
         error_flag,unit=unit_Laser_Frequency)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Mie_Emitted_Freq_Stdev_Threshold"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Mie_Meas_Screening_Params%L1B_Mie_Emitted_FreqStdevThresh,&
         error_flag,unit=unit_Laser_Frequency)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Scattering_Ratio_Min"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Mie_Meas_Screening_Params%L1B_Scattering_Ratio_Min,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Scattering_Ratio_Max"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Mie_Meas_Screening_Params%L1B_Scattering_Ratio_Max,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Mie_SNR_Threshold"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Mie_Meas_Screening_Params%L1B_Mie_SNR_Thresh,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/Mie_Ground_Bin_Thickness_Threshold"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Mie_Meas_Screening_Params%Mie_Ground_Bin_Thickness_Thresh,&
         error_flag,unit=unit_bin_thickness_thresh)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/Max_Signal_Derivative"
    call GetIntAs_from_xml_tree(xml_tree,path,&
         L1B_Mie_Meas_Screening_Params%Max_Signal_Derivative,&
         error_flag,unit=unit_Max_Signal_Derivative)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/Pixel_Saturation_Threshold"
    call GetIntAs_from_xml_tree(xml_tree,path,&
         L1B_Mie_Meas_Screening_Params%Pixel_Saturation_Thresh,&
         error_flag,unit=unit_Pixel_Saturation_Thresh)
    IF (error_flag .ne. no_error) return

    return

  end subroutine GetL1BMieMeasScrParsFromXmlTree
    !  #]
  !-----------------------------------------------------
  subroutine Init_L1B_Rayl_Meas_Scr_Params(L1B_Rayl_Meas_Screening_Params)
    !  #[
    TYPE(L1B_Rayl_Meas_Scr_Params_type), &
         intent(out) :: L1B_Rayl_Meas_Screening_Params

    L1B_Rayl_Meas_Screening_Params%L1B_RayMeasInvalidRefPlssThresh = &
         missing_indicator_IntAl
    L1B_Rayl_Meas_Screening_Params%L1B_Avg_Laser_Freq_Offset_Min = &
         missing_indicator_FAdoxy
    L1B_Rayl_Meas_Screening_Params%L1B_Avg_Laser_Freq_Offset_Max = &
         missing_indicator_FAdoxy
    L1B_Rayl_Meas_Screening_Params%L1B_Avg_UV_Energy_Min = &
         missing_indicator_FAdoxy
    L1B_Rayl_Meas_Screening_Params%L1B_Avg_UV_Energy_Max = &
         missing_indicator_FAdoxy
    L1B_Rayl_Meas_Screening_Params%L1B_Laser_FreqOffsetStdevThresh = &
         missing_indicator_FAdoxy
    L1B_Rayl_Meas_Screening_Params%L1B_UV_Energy_Stdev_Thresh = &
         missing_indicator_FAdoxy
    L1B_Rayl_Meas_Screening_Params%L1B_VelOfAttUncertaintyErrorMin = &
         missing_indicator_FAdoxy
    L1B_Rayl_Meas_Screening_Params%L1B_VelOfAttUncertaintyErrorMax = &
         missing_indicator_FAdoxy
    L1B_Rayl_Meas_Screening_Params%L1B_Rayl_Mean_Emitted_Freq_Min = &
         missing_indicator_FAdoxy
    L1B_Rayl_Meas_Screening_Params%L1B_Rayl_Mean_Emitted_Freq_Max = &
         missing_indicator_FAdoxy
    L1B_Rayl_Meas_Screening_Params%L1B_Rayl_EmittedFreqStdevThresh = &
         missing_indicator_FAdoxy
    L1B_Rayl_Meas_Screening_Params%L1B_Rayleigh_SNR_Min = &
         missing_indicator_FAdoxy
    L1B_Rayl_Meas_Screening_Params%L1B_Rayleigh_SNR_Max = &
         missing_indicator_FAdoxy
    L1B_Rayl_Meas_Screening_Params%Rayl_Ground_Bin_ThicknessThresh = &
         missing_indicator_FAdoxy
    L1B_Rayl_Meas_Screening_Params%Pixel_Saturation_Thresh = &
         missing_indicator_IntAs

  end subroutine Init_L1B_Rayl_Meas_Scr_Params
    !  #]
  subroutine AsciiWrite_L1BRaylMeasScrParams(&
                L1B_Rayl_Meas_Screening_Params,fileunit, prefix, error_flag)
    !  #[
    TYPE(L1B_Rayl_Meas_Scr_Params_type), &
                      intent(in)  :: L1B_Rayl_Meas_Screening_Params
    integer,          intent(in)  :: fileunit
    character(len=*), intent(in)  :: prefix
    integer,          intent(out) :: error_flag
    
    ! local variables
    character(len=25)           :: str_FAdoxy
    character(len=maxlen_IntAl) :: str_IntAl
    character(len=maxlen_IntAs) :: str_IntAs

    error_flag = no_error
    
    call SetIntAl_or_Missing(&
         L1B_Rayl_Meas_Screening_Params%L1B_RayMeasInvalidRefPlssThresh, &
         str_IntAl,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_RayMeasInvalidRefPlssThresh = "//&
         trim(str_IntAl)

    write(fileunit,"(a)",err=999) "REALACC(12)"

    call SetFAdoFull_or_Missing(&
         L1B_Rayl_Meas_Screening_Params%L1B_Avg_Laser_Freq_Offset_Min,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Avg_Laser_Freq_Offset_Min   = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_Frequency//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Rayl_Meas_Screening_Params%L1B_Avg_Laser_Freq_Offset_Max,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Avg_Laser_Freq_Offset_Max   = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_Frequency//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Rayl_Meas_Screening_Params%L1B_Avg_UV_Energy_Min,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Avg_UV_Energy_Min           = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_UV_Energy//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Rayl_Meas_Screening_Params%L1B_Avg_UV_Energy_Max,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Avg_UV_Energy_Max           = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_UV_Energy//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Rayl_Meas_Screening_Params%L1B_Laser_FreqOffsetStdevThresh, &
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Laser_FreqOffsetStdevThresh = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_Frequency//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Rayl_Meas_Screening_Params%L1B_UV_Energy_Stdev_Thresh, &
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_UV_Energy_Stdev_Thresh      = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_UV_Energy//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Rayl_Meas_Screening_Params%L1B_VelofAttUncertaintyErrorMin, &
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_VelofAttUncertaintyErrorMin = "//&
         trim(str_FAdoxy)//" ["//unit_Vel_of_Att_uncertainty//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Rayl_Meas_Screening_Params%L1B_VelofAttUncertaintyErrorMax, &
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_VelofAttUncertaintyErrorMax = "//&
         trim(str_FAdoxy)//" ["//unit_Vel_of_Att_uncertainty//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Rayl_Meas_Screening_Params%L1B_Rayl_Mean_Emitted_Freq_Min, &
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Rayl_Mean_Emitted_Freq_Min  = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_Frequency//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Rayl_Meas_Screening_Params%L1B_Rayl_Mean_Emitted_Freq_Max, &
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Rayl_Mean_Emitted_Freq_Max  = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_Frequency//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Rayl_Meas_Screening_Params%L1B_Rayl_EmittedFreqStdevThresh, &
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Rayl_EmittedFreqStdevThresh = "//&
         trim(str_FAdoxy)//" ["//unit_Laser_Frequency//"]"

    call SetFAdoFull_or_Missing(&
         L1B_Rayl_Meas_Screening_Params%L1B_Rayleigh_SNR_Min, &
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Rayleigh_SNR_Min            = "//&
         trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(&
         L1B_Rayl_Meas_Screening_Params%L1B_Rayleigh_SNR_Max, &
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L1B_Rayleigh_SNR_Max            = "//&
         trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(&
         L1B_Rayl_Meas_Screening_Params%Rayl_Ground_Bin_ThicknessThresh, &
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Rayl_Ground_Bin_ThicknessThresh = "//&
         trim(str_FAdoxy)//" ["//unit_bin_thickness_thresh//"]"

    write(fileunit,"(a)",err=999) "ENDREALACC"

    call SetIntAs_or_Missing(&
         L1B_Rayl_Meas_Screening_Params%Pixel_Saturation_Thresh,&
         str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Pixel_Saturation_Thresh         = "//&
         trim(str_IntAs)//" ["//unit_Pixel_Saturation_Thresh//"]"

    ! spare 1000 bytes

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_L1BRaylMeasScrParams "//&
                          "error while writing to outputfile")
    error_flag = error_writing_file
    return
   
  end subroutine AsciiWrite_L1BRaylMeasScrParams
    !  #]
  subroutine GetL1BRayMeasScrParsFromXmlTree(xml_tree, xml_path, &
                L1B_Rayl_Meas_Screening_Params, error_flag)
    !  #[
    TYPE(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: xml_path
    TYPE(L1B_Rayl_Meas_Scr_Params_type), &
                         intent(out) :: L1B_Rayl_Meas_Screening_Params
    integer,             intent(out) :: error_flag

    ! local variables
    character(len=max_str_len)  :: path

    ! init
    error_flag = no_error

    IF (.not. associated(xml_tree)) THEN
       call logmsg(log_error,"ERROR in GetL1BRayMeasScrParsFromXmlTree():"//&
            "the pointer xml_tree seems not associated !!!")
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! copying the L1B_Rayl_Meas_Screening_Params XML elements
    !------------------------------

    path = trim(xml_path)//"/L1B_Rayleigh_Meas_Invalid_Ref_Pulses_Threshold"
    call GetIntAl_from_xml_tree(xml_tree,path,&
         L1B_Rayl_Meas_Screening_Params%L1B_RayMeasInvalidRefPlssThresh,&
         error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Avg_Laser_Freq_Offset_Min"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Rayl_Meas_Screening_Params%L1B_Avg_Laser_Freq_Offset_Min,&
         error_flag,unit=unit_Laser_Frequency)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Avg_Laser_Freq_Offset_Max"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Rayl_Meas_Screening_Params%L1B_Avg_Laser_Freq_Offset_Max,&
         error_flag,unit=unit_Laser_Frequency)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Avg_UV_Energy_Min"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Rayl_Meas_Screening_Params%L1B_Avg_UV_Energy_Min,&
         error_flag,unit=unit_Laser_UV_Energy)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Avg_UV_Energy_Max"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Rayl_Meas_Screening_Params%L1B_Avg_UV_Energy_Max,&
         error_flag,unit=unit_Laser_UV_Energy)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Laser_Freq_Offset_Stdev_Threshold"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Rayl_Meas_Screening_Params%L1B_Laser_FreqOffsetStdevThresh,&
         error_flag,unit=unit_Laser_Frequency)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_UV_Energy_Stdev_Threshold"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Rayl_Meas_Screening_Params%L1B_UV_Energy_Stdev_Thresh,&
         error_flag,unit=unit_Laser_UV_Energy)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Vel_of_Att_Uncertainty_Error_Min"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Rayl_Meas_Screening_Params%L1B_VelOfAttUncertaintyErrorMin,&
         error_flag,unit=unit_Vel_of_Att_uncertainty)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Vel_of_Att_Uncertainty_Error_Max"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Rayl_Meas_Screening_Params%L1B_VelOfAttUncertaintyErrorMax,&
         error_flag,unit=unit_Vel_of_Att_uncertainty)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Rayleigh_Mean_Emitted_Freq_Min"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Rayl_Meas_Screening_Params%L1B_Rayl_Mean_Emitted_Freq_Min,&
         error_flag,unit=unit_Laser_Frequency)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Rayleigh_Mean_Emitted_Freq_Max"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Rayl_Meas_Screening_Params%L1B_Rayl_Mean_Emitted_Freq_Max,&
         error_flag,unit=unit_Laser_Frequency)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Rayleigh_Emitted_Freq_Stdev_Threshold"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Rayl_Meas_Screening_Params%L1B_Rayl_EmittedFreqStdevThresh,&
         error_flag,unit=unit_Laser_Frequency)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Rayleigh_SNR_Min"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Rayl_Meas_Screening_Params%L1B_Rayleigh_SNR_Min,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L1B_Rayleigh_SNR_Max"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Rayl_Meas_Screening_Params%L1B_Rayleigh_SNR_Max,error_flag)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/Rayleigh_Ground_Bin_Thickness_Threshold"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L1B_Rayl_Meas_Screening_Params%Rayl_Ground_Bin_ThicknessThresh,&
         error_flag,unit=unit_bin_thickness_thresh)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/Pixel_Saturation_Threshold"
    call GetIntAs_from_xml_tree(xml_tree,path,&
         L1B_Rayl_Meas_Screening_Params%Pixel_Saturation_Thresh,&
         error_flag,unit=unit_Pixel_Saturation_Thresh)
    IF (error_flag .ne. no_error) return

    return

  end subroutine GetL1BRayMeasScrParsFromXmlTree
    !  #]
  !-----------------------------------------------------
  subroutine Init_L1B_Screening_Params(L1B_Screening_Params)
    !  #[
    TYPE(L1B_Screening_Params_type), intent(out) :: L1B_Screening_Params

    call Init_L1B_Obs_Screening_Params(&
         L1B_Screening_Params%L1B_Obs_Screening_Params)
    call Init_L1B_Mie_Meas_Scr_Params(&
         L1B_Screening_Params%L1B_Mie_Meas_Screening_Params)
    call Init_L1B_Rayl_Meas_Scr_Params(&
         L1B_Screening_Params%L1B_Rayl_Meas_Screening_Params)

  end subroutine Init_L1B_Screening_Params
    !  #]
  subroutine AsciiWrite_L1B_Screening_Params(L1B_Screening_Params, fileunit,&
                                             prefix, error_flag)
    !  #[
    TYPE(L1B_Screening_Params_type), intent(in)  :: L1B_Screening_Params
    integer,                         intent(in)  :: fileunit
    character(len=*),                intent(in)  :: prefix
    integer,                         intent(out) :: error_flag
    
    error_flag = no_error

    call AsciiWrite_L1B_Obs_Scr_Params(&
         L1B_Screening_Params%L1B_Obs_Screening_Params, fileunit, &
         trim(prefix)//"%L1B_Obs_Screening_Params", error_flag)
    IF (error_flag .ne. no_error) return

    call AsciiWrite_L1B_MieMeasScrParams(&
         L1B_Screening_Params%L1B_Mie_Meas_Screening_Params, fileunit, &
         trim(prefix)//"%L1B_Mie_Meas_Screening_Params", error_flag)
    IF (error_flag .ne. no_error) return

    call AsciiWrite_L1BRaylMeasScrParams(&
         L1B_Screening_Params%L1B_Rayl_Meas_Screening_Params, fileunit, &
         trim(prefix)//"%L1B_Rayl_Meas_Screening_Params", error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine AsciiWrite_L1B_Screening_Params
    !  #]
  subroutine Get_L1BScrParams_from_xml_tree(xml_tree, xml_path, &
                                 L1B_Screening_Params, error_flag)
    !  #[
    TYPE(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: xml_path
    TYPE(L1B_Screening_Params_type), &
                         intent(out) :: L1B_Screening_Params
    integer,             intent(out) :: error_flag

    ! init
    error_flag = no_error

    IF (.not. associated(xml_tree)) THEN
       call logmsg(log_error,"ERROR in Get_L1BScrParams_from_xml_tree():"//&
            "the pointer xml_tree seems not associated !!!")
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! copying the L1B_Screening_Params XML elements
    !------------------------------

    call Get_L1BObsScrPars_from_xmltree(xml_tree, &
         trim(xml_path)//"/L1B_Obs_Screening_Params", &
         L1B_Screening_Params%L1B_Obs_Screening_Params, error_flag)
    IF (error_flag .ne. no_error) return

    call GetL1BMieMeasScrParsFromXmlTree(xml_tree, &
         trim(xml_path)//"/L1B_Mie_Meas_Screening_Params", &
         L1B_Screening_Params%L1B_Mie_Meas_Screening_Params, error_flag)
    IF (error_flag .ne. no_error) return

    call GetL1BRayMeasScrParsFromXmlTree(xml_tree, &
         trim(xml_path)//"/L1B_Rayleigh_Meas_Screening_Params", &
         L1B_Screening_Params%L1B_Rayl_Meas_Screening_Params, error_flag)
    IF (error_flag .ne. no_error) return

    return

  end subroutine Get_L1BScrParams_from_xml_tree
    !  #]
  !-----------------------------------------------------
  subroutine Init_L2B_AMD_Screening_Params(L2B_AMD_Screening_Params)
    !  #[
    TYPE(L2B_AMD_Screening_Params_type), &
         intent(out) :: L2B_AMD_Screening_Params

    L2B_AMD_Screening_Params%L2B_AMD_p_min = &
         missing_indicator_FAdoxy
    L2B_AMD_Screening_Params%L2B_AMD_p_max = &
         missing_indicator_FAdoxy
    L2B_AMD_Screening_Params%L2B_AMD_T_min = &
         missing_indicator_FAdoxy
    L2B_AMD_Screening_Params%L2B_AMD_T_max = &
         missing_indicator_FAdoxy

    ! spare 1000 bytes

  end subroutine Init_L2B_AMD_Screening_Params
    !  #]
  subroutine AsciiWrite_L2B_AMD_Scr_Params(&
              L2B_AMD_Screening_Params,fileunit, prefix, error_flag)
    !  #[
    TYPE(L2B_AMD_Screening_Params_type), &
                      intent(in)  :: L2B_AMD_Screening_Params
    integer,          intent(in)  :: fileunit
    character(len=*), intent(in)  :: prefix
    integer,          intent(out) :: error_flag
    
    ! local variables
    character(len=25)           :: str_FAdoxy

    error_flag = no_error
    
    write(fileunit,"(a)",err=999) "REALACC(12)"

    call SetFAdoFull_or_Missing(&
         L2B_AMD_Screening_Params%L2B_AMD_p_min,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L2B_AMD_p_min                  = "//&
         trim(str_FAdoxy)//" ["//unit_L2B_AMD_p//"]"

    call SetFAdoFull_or_Missing(&
         L2B_AMD_Screening_Params%L2B_AMD_p_max,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L2B_AMD_p_max                  = "//&
         trim(str_FAdoxy)//" ["//unit_L2B_AMD_p//"]"

    call SetFAdoFull_or_Missing(&
         L2B_AMD_Screening_Params%L2B_AMD_T_min,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L2B_AMD_T_min                  = "//&
         trim(str_FAdoxy)//" ["//unit_L2B_AMD_T//"]"

    call SetFAdoFull_or_Missing(&
         L2B_AMD_Screening_Params%L2B_AMD_T_max,&
         str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L2B_AMD_T_max                  = "//&
         trim(str_FAdoxy)//" ["//unit_L2B_AMD_T//"]"

    write(fileunit,"(a)",err=999) "ENDREALACC"

    ! spare 1000 bytes

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_L2B_AMD_Scr_Params "//&
                          "error while writing to outputfile")
    error_flag = error_writing_file
    return
   
  end subroutine AsciiWrite_L2B_AMD_Scr_Params
    !  #]

  subroutine Get_L2BAMDScrParamsFromXmlTree(xml_tree, xml_path, &
              L2B_AMD_Screening_Params, error_flag)
    !  #[
    TYPE(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: xml_path
    TYPE(L2B_AMD_Screening_Params_type), &
                         intent(out) :: L2B_AMD_Screening_Params
    integer,             intent(out) :: error_flag

    ! local variables
    character(len=max_str_len)  :: path

    ! init
    error_flag = no_error

    IF (.not. associated(xml_tree)) THEN
       call logmsg(log_error,"ERROR in Get_L2BAMDScrParamsFromXmlTree():"//&
            "the pointer xml_tree seems not associated !!!")
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! copying the L2B_AMD_Screening_Params XML elements
    !------------------------------

    path = trim(xml_path)//"/L2B_AMD_p_min"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L2B_AMD_Screening_Params%L2B_AMD_p_min,&
         error_flag,unit=unit_L2B_AMD_p)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L2B_AMD_p_max"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L2B_AMD_Screening_Params%L2B_AMD_p_max,&
         error_flag,unit=unit_L2B_AMD_p)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L2B_AMD_T_min"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L2B_AMD_Screening_Params%L2B_AMD_T_min,&
         error_flag,unit=unit_L2B_AMD_T)
    IF (error_flag .ne. no_error) return

    path = trim(xml_path)//"/L2B_AMD_T_max"
    call GetFAdoxy_from_xml_tree(xml_tree,path,&
         L2B_AMD_Screening_Params%L2B_AMD_T_max,&
         error_flag,unit=unit_L2B_AMD_T)
    IF (error_flag .ne. no_error) return

    return

  end subroutine Get_L2BAMDScrParamsFromXmlTree
    !  #]
  !-----------------------------------------------------
  subroutine Init_Screening_Params(Screening_Params)
    !  #[
    TYPE(Screening_Params_type), intent(out) :: Screening_Params

    call Init_L1B_Screening_Params(&
                   Screening_Params%L1B_Screening_Params)
    call Init_L2B_AMD_Screening_Params(&
                   Screening_Params%L2B_AMD_Screening_Params)

  end subroutine Init_Screening_Params
    !  #]
  subroutine AsciiWrite_Screening_Params(Screening_Params, fileunit, &
                                         prefix, error_flag)
    !  #[
    TYPE(Screening_Params_type), intent(in)  :: Screening_Params
    integer,                     intent(in)  :: fileunit
    character(len=*),            intent(in)  :: prefix
    integer,                     intent(out) :: error_flag
    
    error_flag = no_error

    call AsciiWrite_L1B_Screening_Params(&
         Screening_Params%L1B_Screening_Params, fileunit, &
         trim(prefix)//"%L1B_Screening_Params", error_flag)
    IF (error_flag .ne. no_error) return

    call AsciiWrite_L2B_AMD_Scr_Params(&
         Screening_Params%L2B_AMD_Screening_Params,fileunit,&
         trim(prefix)//"%L2B_AMD_Screening_Params",error_flag)
    IF (error_flag .ne. no_error) return

    return

  end subroutine AsciiWrite_Screening_Params
    !  #]
  subroutine Get_Scr_Params_from_xml_tree(xml_tree, xml_path, &
                                          Screening_Params, error_flag)
    !  #[
    TYPE(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: xml_path
    TYPE(Screening_Params_type), &
                         intent(out) :: Screening_Params
    integer,             intent(out) :: error_flag

    ! init
    error_flag = no_error

    IF (.not. associated(xml_tree)) THEN
       call logmsg(log_error,"ERROR in Get_Scr_Params_from_xml_tree():"//&
            "the pointer xml_tree seems not associated !!!")
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! copying the Screening_Params XML elements
    !------------------------------

    call Get_L1BScrParams_from_xml_tree(xml_tree, &
         trim(xml_path)//"/L1B_Screening_Params", &
         Screening_Params%L1B_Screening_Params, error_flag)
    IF (error_flag .ne. no_error) return

    call Get_L2BAMDScrParamsFromXmlTree(xml_tree, &
         trim(xml_path)//"/L2B_AMD_Screening_Params", &
         Screening_Params%L2B_AMD_Screening_Params, error_flag)
    IF (error_flag .ne. no_error) return

    !    1234567890123456789012345678901

    return

  end subroutine Get_Scr_Params_from_xml_tree
    !  #]
  !  #]
  !-----------------------------------------------------
  subroutine Init_L2B_AuxPar_Data(Data,error_flag)
    !  #[
    TYPE(L2B_AuxParData_type), pointer     :: Data       ! input/output
    integer,                   intent(out) :: error_flag ! output

    error_flag = no_error

    IF (.not. associated(Data)) THEN
       call logmsg(log_error,"ERROR in Init_L2B_AuxPar_Data: "//&
            "the pointer to Data seems not associated")
       error_flag = error_programming
       return
    END IF
    
    call Init_FH_Default_Fields( Data%FH_Default_Fields)
    call Init_MPH_Default_Fields(Data%MPH_Default_Fields)
    call Init_WVM_Params(        Data%WVM_Params)
    call Init_Screening_Params(  Data%Screening_Params)

  end subroutine Init_L2B_AuxPar_Data
    !  #]
  subroutine Delete_L2B_AuxPar_Data(Data)
    !  #[
    TYPE(L2B_AuxParData_type), pointer     :: Data       ! input/output

    IF (.not. associated(Data)) return ! nothing to do
    
    call Delete_WVM_Params(        Data%WVM_Params)

  end subroutine Delete_L2B_AuxPar_Data
    !  #]
  subroutine AsciiWrite_L2B_AuxPar_Data(Data, fileunit, prefix, error_flag)
    !  #[
    TYPE(L2B_AuxParData_type), pointer     :: Data
    integer,                   intent(in)  :: fileunit
    character(len=*),          intent(in)  :: prefix
    integer,                   intent(out) :: error_flag
    
    error_flag = no_error

    IF (.not. associated(Data)) THEN
       write(fileunit,"(a)",err=999) trim(prefix)//" [seems not associated]"
       return
    END IF

    call AsciiWrite_FH_Default_Fields(Data%FH_Default_Fields, fileunit, &
         trim(prefix)//"%FH_Default_Fields", error_flag)
    IF (error_flag .ne. no_error) return

    call AsciiWrite_MPH_Default_Fields(Data%MPH_Default_Fields, fileunit, &
         trim(prefix)//"%MPH_Default_Fields", error_flag)
    IF (error_flag .ne. no_error) return

    call AsciiWrite_WVM_Params(Data%WVM_Params, fileunit, &
         trim(prefix)//"%WVM_Params", error_flag)
    IF (error_flag .ne. no_error) return

    call AsciiWrite_Screening_Params(Data%Screening_Params, fileunit, &
         trim(prefix)//"%Screening_Params", error_flag)
    IF (error_flag .ne. no_error) return

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_L2B_AuxPar_Data "//&
                          "error while writing to outputfile")
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2B_AuxPar_Data
    !  #]
  subroutine Get_L2BAuxParData_from_xml_tree(xml_tree,xml_path,Data,error_flag)
    !  #[
    TYPE(xml_node_type),       pointer     :: xml_tree
    character(len=*),          intent(in)  :: xml_path
    TYPE(L2B_AuxParData_type), pointer     :: Data       ! input/output
    integer,                   intent(out) :: error_flag ! output

    ! init
    error_flag = no_error

    IF (.not. associated(xml_tree)) THEN
       call logmsg(log_error,"ERROR in Get_L2BAuxParData_from_xml_tree():"//&
            "the pointer xml_tree seems not associated !!!")
       error_flag = error_programming
       return
    END IF

    !------------------------------
    ! copying the Data XML elements
    !------------------------------

    IF (.not. associated(Data)) THEN
       call logmsg(log_error,"ERROR in Init_L2B_AuxPar_Data: "//&
            "the pointer to Data seems not associated")
       error_flag = error_programming
       return
    END IF
    
    call  Get_FH_Def_Flds_from_xml_tree(xml_tree, &
         trim(xml_path)//"/FH_Default_Fields", &
         Data%FH_Default_Fields, error_flag)
    IF (error_flag .ne. no_error) return

    call  Get_MPH_Def_Flds_from_xml_tree(xml_tree, &
         trim(xml_path)//"/MPH_Default_Fields", &
         Data%MPH_Default_Fields, error_flag)
    IF (error_flag .ne. no_error) return

    call  Get_WVM_Params_from_xml_tree(xml_tree, &
         trim(xml_path)//"/WVM_Params", &
         Data%WVM_Params, error_flag)
    IF (error_flag .ne. no_error) return

    call  Get_Scr_Params_from_xml_tree(xml_tree, &
         trim(xml_path)//"/Screening_Params", &
         Data%Screening_Params, error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine Get_L2BAuxParData_from_xml_tree
    !  #]
  !-----------------------------------------------------
  subroutine Init_L2B_AuxPar(L2B_AuxPar,error_flag)
    !  #[
    TYPE(L2B_AuxPar_Type), intent(out) :: L2B_AuxPar ! output
    integer,               intent(out) :: error_flag ! output

    error_flag = no_error

    ! init the datastructure
    nullify(L2B_AuxPar%FH)
    nullify(L2B_AuxPar%MPH)
    nullify(L2B_AuxPar%SPH)

     ! the attached dataset
    nullify(L2B_AuxPar%Data)

  end subroutine Init_L2B_AuxPar
    !  #]
  subroutine Delete_L2B_AuxPar(L2B_AuxPar)
    !  #[
    TYPE(L2B_AuxPar_Type), intent(inout) :: L2B_AuxPar ! input/output

    ! clean up the header datastructures
    IF (associated(L2B_AuxPar%SPH)) call Delete_L2B_AuxPar_SPH(L2B_AuxPar%SPH)

    ! init the datastructure
    IF (associated(L2B_AuxPar%FH))  deallocate(L2B_AuxPar%FH)
    IF (associated(L2B_AuxPar%MPH)) deallocate(L2B_AuxPar%MPH)
    IF (associated(L2B_AuxPar%SPH)) deallocate(L2B_AuxPar%SPH)

    ! the attached dataset
    IF (associated(L2B_AuxPar%Data)) THEN
       call Delete_L2B_AuxPar_Data(L2B_AuxPar%Data)
       deallocate(L2B_AuxPar%Data)
    END IF

  end subroutine Delete_L2B_AuxPar
    !  #]
  subroutine Print_L2B_AuxPar(L2B_AuxPar, prefix, error_flag)
    !  #[ ascii output to screen
    TYPE(L2B_AuxPar_Type), intent(in)  :: L2B_AuxPar ! input
    character(len=*),      intent(in)  :: prefix     ! input
    integer,               intent(out) :: error_flag ! output

    error_flag = no_error

    call AsciiWrite_L2B_AuxPar(L2B_AuxPar,fileunit_stdout,prefix,error_flag)
    IF (error_flag .ne. no_error) return
    
  end subroutine Print_L2B_AuxPar
  !  #]                           
  subroutine AsciiWrite_L2B_AuxPar(L2B_AuxPar, fileunit, prefix, error_flag)
    !  #[
    TYPE(L2B_AuxPar_Type), intent(in)  :: L2B_AuxPar ! input
    integer,               intent(in)  :: fileunit   ! input
    character(len=*),      intent(in)  :: prefix     ! input
    integer,               intent(out) :: error_flag ! output
    
    error_flag = no_error

    write(fileunit,"(a)") "=================================="
    write(fileunit,"(a)") "contents of the "//trim(prefix)//" datastructure:"
    write(fileunit,"(a)") "=================================="

    call AsciiWrite_FH( L2B_AuxPar%FH, fileunit,&
                        prefix=trim(prefix)//"%FH",error_flag=error_flag)
    IF(error_flag .ne. no_error) return
    call AsciiWrite_MPH(L2B_AuxPar%MPH,fileunit,&
                        prefix=trim(prefix)//"%MPH",error_flag=error_flag)
    IF(error_flag .ne. no_error) return
    call AsciiWrite_L2B_AuxPar_SPH(L2B_AuxPar%SPH,&
                        fileunit,trim(prefix)//"%SPH",error_flag=error_flag)
    IF(error_flag .ne. no_error) return

    ! the attached dataset
    call AsciiWrite_L2B_AuxPar_Data(L2B_AuxPar%Data, fileunit, &
                                    prefix=trim(prefix)//"%Data",&
                                    error_flag=error_flag)
    IF(error_flag .ne. no_error) return

  end subroutine AsciiWrite_L2B_AuxPar
    !  #]
  !-----------------------------------------------------
  subroutine Get_BackscRatio_Thresh_Profile(Classification_Params,&
                                            BackscRatio_Thresh_Profile,&
                                            error_flag)
    !  #[
    TYPE(Classification_Params_Type), &
                        intent(in)  :: Classification_Params      ! input
    TYPE(profile_type), intent(out) :: BackscRatio_Thresh_Profile ! result
    integer,            intent(out) :: error_flag                 ! output
    
    error_flag = no_error

    IF (.not. associated(Classification_Params%Threshold_BackscatterRatio)) THEN
       call logmsg(log_error,"ERROR in Get_BackscRatio_Thresh_Profile: "//&
            "Classification_Params%Threshold_BackscatterRatio(:) "//&
            "seems not associated")
       error_flag = error_programming
       return
    END IF

    ! init the profile and convert altitude from [km] to [m]
    call init_profile(BackscRatio_Thresh_Profile,&
         Classification_Params%Threshold_BackscatterRatio(:)%Altitude*1.e3_r8_,&
         Classification_Params%Threshold_BackscatterRatio(:)%Threshold_Value,&
         size(Classification_Params%Threshold_BackscatterRatio),&
         "BackscRatio_Thresh_Profile", error_flag, &
         interpolation_type      = interp_type_linear,&
         extrapolation_type_down = extrap_type_linear,&
         extrapolation_type_up   = extrap_type_linear)
    IF (error_flag .ne. no_error) return

    !print *,"testjos: BackscRatio_Thresh_Profile"
    !call print_profile(BackscRatio_Thresh_Profile)

  end subroutine Get_BackscRatio_Thresh_Profile
    !  #]
  subroutine Get_Extinction_Thresh_Profile(Classification_Params,&
                                           Extinction_Thresh_Profile,&
                                           error_flag)
    !  #[
    TYPE(Classification_Params_Type), &
                        intent(in)  :: Classification_Params     ! input
    TYPE(profile_type), intent(out) :: Extinction_Thresh_Profile ! result
    integer,            intent(out) :: error_flag                ! output
    
    error_flag = no_error

    IF (.not. associated(Classification_Params%Threshold_Extinction)) THEN
       call logmsg(log_error,"ERROR in Get_Extinction_Thresh_Profile: "//&
            "Classification_Params%Threshold_Extinction(:) "//&
            "seems not associated")
       error_flag = error_programming
       return
    END IF

    ! init the profile and convert altitude from [km] to [m]
    call init_profile(Extinction_Thresh_Profile,&
         Classification_Params%Threshold_Extinction(:)%Altitude*1.e3_r8_,&
         Classification_Params%Threshold_Extinction(:)%Threshold_Value,&
         size(Classification_Params%Threshold_Extinction),&
         "Extinction_Thresh_Profile", error_flag, &
         interpolation_type      = interp_type_linear,&
         extrapolation_type_down = extrap_type_linear,&
         extrapolation_type_up   = extrap_type_linear)
    IF (error_flag .ne. no_error) return

    !print *,"testjos: Extinction_Thresh_Profile"
    !call print_profile(Extinction_Thresh_Profile)

  end subroutine Get_Extinction_Thresh_Profile
    !  #]
  !-----------------------------------------------------
  ! some helper routines for conversion and checking
  ! (should these be moved into the EE_CFI_Datatypes module?)
  !-----------------------------------------------------
  subroutine GetIntAs_from_xml_tree(xml_tree,xml_path,i,error_flag,unit)
    !  #[
    TYPE(xml_node_type),        pointer     :: xml_tree
    character(len=*),           intent(in)  :: xml_path
    integer(IntAs),             intent(out) :: i
    integer,                    intent(out) :: error_flag
    character(len=*), optional, intent(in)  :: unit

    ! local variables
    character(len=max_str_len)  :: text

    ! init
    error_flag = no_error
    i = missing_indicator_IntAs

    ! get the string value from the xml_tree
    call xml_node_get_string(xml_tree,xml_path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,xml_path,&
                                 maxlen_IntAs,"GetIntAs_from_xml_tree")

    ! get the integer value from the string
    call GetIntAs(text,i,error_flag)
    IF (error_flag .ne. no_error) return

    IF (present(unit)) THEN
       ! check if we have the expected physical "unit" 
       call Check_Unit_from_xml_tree(xml_tree,xml_path,unit,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

  end subroutine GetIntAs_from_xml_tree
    !  #]
  subroutine GetIntAus_from_xml_tree(xml_tree,xml_path,i,error_flag,unit)
    !  #[
    TYPE(xml_node_type),        pointer     :: xml_tree
    character(len=*),           intent(in)  :: xml_path
    integer(IntAus),            intent(out) :: i
    integer,                    intent(out) :: error_flag
    character(len=*), optional, intent(in)  :: unit

    ! local variables
    character(len=max_str_len)  :: text

    ! init
    error_flag = no_error
    i = missing_indicator_IntAus

    ! get the string value from the xml_tree
    call xml_node_get_string(xml_tree,xml_path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,xml_path,&
                                 maxlen_IntAus,"GetIntAus_from_xml_tree")

    ! get the integer value from the string
    call GetIntAus(text,i,error_flag)
    IF (error_flag .ne. no_error) return

    IF (present(unit)) THEN
       ! check if we have the expected physical "unit" 
       call Check_Unit_from_xml_tree(xml_tree,xml_path,unit,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

  end subroutine GetIntAus_from_xml_tree
    !  #]
  subroutine GetIntAl_from_xml_tree(xml_tree,xml_path,i,error_flag,unit)
    !  #[
    TYPE(xml_node_type),        pointer     :: xml_tree
    character(len=*),           intent(in)  :: xml_path
    integer(IntAl),             intent(out) :: i
    integer,                    intent(out) :: error_flag
    character(len=*), optional, intent(in)  :: unit

    ! local variables
    character(len=max_str_len)  :: text

    ! init
    error_flag = no_error
    i = missing_indicator_IntAl

    ! get the string value from the xml_tree
    call xml_node_get_string(xml_tree,xml_path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,xml_path,&
                                 maxlen_IntAl,"GetIntAl_from_xml_tree")

    ! get the integer value from the string
    call GetIntAl(text,i,error_flag)
    IF (error_flag .ne. no_error) return

    IF (present(unit)) THEN
       ! check if we have the expected physical "unit" 
       call Check_Unit_from_xml_tree(xml_tree,xml_path,unit,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

  end subroutine GetIntAl_from_xml_tree
    !  #]
  subroutine GetIntAul_from_xml_tree(xml_tree,xml_path,i,error_flag,unit)
    !  #[
    TYPE(xml_node_type),        pointer     :: xml_tree
    character(len=*),           intent(in)  :: xml_path
    integer(IntAul),            intent(out) :: i
    integer,                    intent(out) :: error_flag
    character(len=*), optional, intent(in)  :: unit

    ! local variables
    character(len=max_str_len)  :: text

    ! init
    error_flag = no_error
    i = missing_indicator_IntAul

    ! get the string value from the xml_tree
    call xml_node_get_string(xml_tree,xml_path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,xml_path,&
                                 maxlen_IntAul,"GetIntAul_from_xml_tree")

    ! get the integer value from the string
    call GetIntAul(text,i,error_flag)
    IF (error_flag .ne. no_error) return

    IF (present(unit)) THEN
       ! check if we have the expected physical "unit" 
       call Check_Unit_from_xml_tree(xml_tree,xml_path,unit,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

  end subroutine GetIntAul_from_xml_tree
    !  #]
  subroutine GetFAdoxy_from_xml_tree(xml_tree,xml_path,x,error_flag,unit)
    !  #[
    TYPE(xml_node_type),        pointer     :: xml_tree
    character(len=*),           intent(in)  :: xml_path
    real(FAdoxy),               intent(out) :: x
    integer,                    intent(out) :: error_flag
    character(len=*), optional, intent(in)  :: unit

    ! local variables
    character(len=max_str_len)  :: text

    ! init
    error_flag = no_error
    x = missing_indicator_FAdoxy

    ! get the string value from the xml_tree
    call xml_node_get_string(xml_tree,xml_path,text,error_flag)
    IF (error_flag .ne. no_error) return
    text = check_length_xml_item(text,xml_path,&
                                 maxlen_FAdoxy,"GetFAdoxy_from_xml_tree")

    ! get the real value from the string
    call GetFAdoxy(text,x,error_flag)
    IF (error_flag .ne. no_error) return

    IF (present(unit)) THEN
       ! check if we have the expected physical "unit" 
       call Check_Unit_from_xml_tree(xml_tree,xml_path,unit,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

  end subroutine GetFAdoxy_from_xml_tree
    !  #]
  subroutine Check_Unit_from_xml_tree(xml_tree,xml_path,&
                                      expected_unit, error_flag)
    !  #[
    TYPE(xml_node_type), pointer     :: xml_tree
    character(len=*),    intent(in)  :: xml_path
    character(len=*),    intent(in)  :: expected_unit
    integer,             intent(out) :: error_flag

    ! local variables
    character(len=max_str_len)  :: attr_value

    ! check if we have the expected physical "unit" 
    call xml_tree_read_string_attribute(xml_tree,xml_path,"unit",&
                                        attr_value,error_flag)
    IF (error_flag .ne. no_error) return
    
    IF (trim(attr_value) .ne. trim(expected_unit)) THEN
       call logmsg(log_error,"ERROR in Check_Unit_from_xml_tree():"//&
            "expected unit ["//trim(expected_unit)//&
            "] differs from actual unit ["//trim(attr_value)//&
            "] for item ["//trim(xml_path)//"]")
       error_flag = error_reading_file
       return
    END IF
    
  end subroutine Check_Unit_from_xml_tree
    !  #]
  !-----------------------------------------------------
end module L2B_AuxPar_DataStructure
