module L2B_PCD_ADS
  !  #[ Documentation
  !
  !     a module to handle the L2B Product Confidence Data Annotation dataset
  !     as defined in AE-IF_ECMWF-L2BP-001 (L2B IODD), version ... 
  !     sec. ..., 
  !
  !     Written by J. de Kloe, KNMI 
  !     created:     29-11-2005
  !
  !     History
  !     05-Jan-2006 P. Poli        Implemented writing of PCD_ADS Uses L2B QC
  !                                parameter table from JdK v0.2 dd 03/01/2006
  !     10-Jan-2006 P. Poli        Added reading routines (NOT tested yet)
  !     30-Jan-2006 P. Poli        Added N_Meas, N_Obs_*_Actual
  !     23-Mar-2006 J. de Kloe     introduced *_to_write parameters, 
  !                                removed fillwithzero options, and
  !                                rearranged looping in read/write routines
  !     29-Mar-2006 J. de Kloe     added all AsciiWrite* routines
  !     26-Apr-2006 D. Tan         Remove Gross_Error_Probab. (IODD Draft 1.0D)
  !     12-May-2006 D. Tan         Null use of ParKind1 module
  !     29-Jun-2006 G.J. Marseille L1B_Rayleigh_Currupt => L1B_Rayleigh_Corrupt
  !     08-Mar-2007 J. de Kloe     reintroduced asciiwrite code from march-2006
  !                                and several clean-ups
  !     15-Mar-2007 J. de Kloe     added AsciiWrite routines
  !     15-Mar-2007 J. de Kloe     major cleanup, removing all todo's
  !     15-Mar-2007 J. de Kloe     removed use of SummaryVars
  !     20-Mar-2007 J. de Kloe     filling M_Meas now during initialisation
  !                                (as was already done in all other datasets)
  !     24-Jul-2007 J. de Kloe     added L1B_Invalid_Num_Mie_Peaks parameter
  !     16-Jan-2008 J. de Kloe     phase out integer kind i_ 
  !
  !  #]
  !  #[ Modules used
!  USE DirectBinaryIO, only: write_zero_bytes_bin_file, &
!                            read_zero_bytes_bin_file
! replaced by the aliasses in EE_CFI_DataTypes:
!       write_SpareBytes_to_bin_file, read_SpareBytes_from_bin_file
  USE EE_CFI_Datatypes, only: &
       FAdoxy, c_bytelen_FAdoxy,   SetFAdoFull_or_Missing, &
       missing_indicator_FAdoxy, &
       IntAl, &
       IntAs,  c_bytelen_IntAs,  maxlen_IntAs,  SetIntAs,  &
       IntAuc, c_bytelen_IntAuc, maxlen_IntAuc, SetIntAuc, &
       maxlen_IntAucFlags, SetIntAucFlags, &
       write_IntAs_to_bin_file,  read_IntAs_from_bin_file,  &
       write_IntAuc_to_bin_file, read_IntAuc_from_bin_file, &
       write_FAdoxy_to_bin_file, read_FAdoxy_from_bin_file, &
       write_SpareBytes_to_bin_file, read_SpareBytes_from_bin_file
  ! support modules
  USE Logging, only: logmsg,log_debug,log_info,log_progress,log_warn,log_error
  USE ErrorHandler, only: no_error, error_programming, error_allocate,&
                          error_writing_file
  USE LunManager, only: fileunit_stdout, get_filename
  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  !  #]
  !  #[ parameter and type definitions
  implicit none
  !---------------------------------------------------------------------
  !  #[ datastructures for L1B_Input_Screening
  !---------------------------------------------------------------------

  TYPE L1B_Mie_Meas_Bin_Scr_Type
     !  #[ see table 21 in section 4.3.2 of the iodd
     integer(IntAuc) :: Bin_QC ! enum field, see possible values below
     integer(IntAuc) :: Bin_QC_Flags
  END TYPE L1B_Mie_Meas_Bin_Scr_Type

  ! possible values to be used inside the above enum type
  ! see table 22 in section 4.3.2 of the iodd
  integer(IntAuc), parameter :: L1B_Mie_Bin_OK               = 0
  integer(IntAuc), parameter :: L1B_Mie_Bin_Invalid          = 1
  integer(IntAuc), parameter :: L1B_Mie_Scattering_Ratio     = 2
  integer(IntAuc), parameter :: L1B_Mie_SNR                  = 3
  integer(IntAuc), parameter :: L1B_Mie_Bin_Contains_Surface = 50
  integer(IntAuc), parameter :: L1B_Mie_Contains_Surface_DEM = 51
  integer(IntAuc), parameter :: L1B_Mie_Bin_Below_Surface    = 52
  integer(IntAuc), parameter :: L1B_Mie_Below_Surface_DEM    = 53
  ! this is our missing value, its not defined in the iodd
  integer(IntAuc), parameter :: L1B_Mie_Bin_undefined        = 255
  !  #]

  TYPE L1B_Mie_Meas_Screening_Type
     !  #[ see table 19 in section 4.3.2 of the iodd
     integer(IntAuc) :: Meas_QC ! enum field, see possible values below
     integer(IntAuc) :: Meas_QC_Flags ! to be defined later
     TYPE(L1B_Mie_Meas_Bin_Scr_Type), &
          dimension(:), pointer :: Bin_Screening
     ! 1 spare byte
  END TYPE L1B_Mie_Meas_Screening_Type

  ! possible values to be used inside the above enum type
  ! see table 20 in section 4.3.2 of the iodd
  integer(IntAuc), parameter :: L1B_Mie_Meas_OK                  = 0
  integer(IntAuc), parameter :: L1B_Mie_Meas_Invalid_Ref_Pulses  = 1
  integer(IntAuc), parameter :: L1B_Mie_Avg_Laser_Freq_Offset    = 2
  integer(IntAuc), parameter :: L1B_Mie_Avg_UV_Energy            = 3
  integer(IntAuc), parameter :: L1B_Mie_Laser_Freq_Offset_Stdev  = 4
  integer(IntAuc), parameter :: L1B_Mie_UV_Energy_Stdev          = 5
  integer(IntAuc), parameter :: L1B_Mie_Vel_of_Att_Uncert_Error  = 6
  integer(IntAuc), parameter :: L1B_Mie_Mean_Emitted_Freq        = 7
  integer(IntAuc), parameter :: L1B_Mie_Emitted_Freq_Stdev       = 8
  integer(IntAuc), parameter :: L1B_Meas_Reference_Pulse_FWHM    = 9
  ! this is our missing value, its not defined in the iodd
  integer(IntAuc), parameter :: L1B_Mie_Meas_Undefined           = 255
  !  #]

  TYPE L1B_Rayl_Meas_Bin_Scr_Type
     !  #[ see table 25 in section 4.3.2 of the iodd
     integer(IntAuc) :: Bin_QC  ! enum field, see possible values below
     integer(IntAuc) :: Bin_QC_Flags
  END TYPE L1B_Rayl_Meas_Bin_Scr_Type

  ! possible values to be used inside the above type
  ! see table 26 in section 4.3.2 of the iodd
  integer(IntAuc), parameter :: L1B_Rayl_Bin_OK               = 0
  integer(IntAuc), parameter :: L1B_Rayl_Bin_Invalid          = 1
  integer(IntAuc), parameter :: L1B_Rayl_SNR_A                = 2
  integer(IntAuc), parameter :: L1B_Rayl_SNR_B                = 3
  integer(IntAuc), parameter :: L1B_Rayl_Bin_Contains_Surface = 50
  integer(IntAuc), parameter :: L1B_Rayl_Contains_Surface_DEM = 51
  integer(IntAuc), parameter :: L1B_Rayl_Bin_Below_Surface    = 52
  integer(IntAuc), parameter :: L1B_Rayl_Below_Surface_DEM    = 53
  ! this is our missing value, its not defined in the iodd
  integer(IntAuc), parameter :: L1B_Rayl_Bin_undefined        = 255
  !  #]

  TYPE L1B_Rayl_Meas_Screening_Type
     !  #[ see table 23 in section 4.3.2 of the iodd
     integer(IntAuc) :: Meas_QC ! enum field, see possible values below
     integer(IntAuc) :: Meas_QC_Flags ! to be defined later
     TYPE(L1B_Rayl_Meas_Bin_Scr_Type), dimension(:), pointer :: Bin_Screening
     ! 1 spare byte
  END TYPE L1B_Rayl_Meas_Screening_Type

  ! possible values to be used inside the above enum type
  ! see table 24 in section 4.3.2 of the iodd
  integer(IntAuc), parameter :: L1B_Rayl_Meas_OK               = 0
  integer(IntAuc), parameter :: L1B_Rayl_Meas_Invalid_Ref_Pls  = 1
  integer(IntAuc), parameter :: L1B_Rayl_Avg_Laser_Freq_Offset = 2
  integer(IntAuc), parameter :: L1B_Rayl_Avg_UV_Energy         = 3
  integer(IntAuc), parameter :: L1B_Rayl_Laser_Freq_Offs_Stdev = 4
  integer(IntAuc), parameter :: L1B_Rayl_UV_Energy_Stdev       = 5
  integer(IntAuc), parameter :: L1B_Rayl_Vel_of_Att_Uncert_Err = 6
  integer(IntAuc), parameter :: L1B_Rayl_Mean_Emitted_Freq     = 7
  integer(IntAuc), parameter :: L1B_Rayl_Emitted_Freq_Stdev    = 8
  ! this is our missing value, its not defined in the iodd
  integer(IntAuc), parameter :: L1B_Rayl_Meas_Undefined        = 255
  !  #]

  TYPE L1B_MRC_Screening_Type
     !  #[ see table 27 in section 4.3.2 of the iodd
     integer(IntAuc) :: Calibration_Valid ! overall validity
     ! others?
     ! 5 spare bytes
  END TYPE L1B_MRC_Screening_Type

  ! possible values to be used inside the above type
  ! see table [to be writtten] in section 4.3.2 of the iodd
  integer(IntAuc), parameter :: L1B_MRC_Cal_OK        = 0
  integer(IntAuc), parameter :: L1B_MRC_Cal_NotOK     = 1
  ! this is our missing value, its not defined in the iodd
  integer(IntAuc), parameter :: L1B_MRC_Cal_Undefined = 255
  !  #]

  TYPE L1B_RRC_Screening_Type
     !  #[ see table 28 in section 4.3.2 of the iodd
     integer(IntAuc) :: Calibration_Valid ! overall validity
     ! others?
     ! 5 spare bytes
  END TYPE L1B_RRC_Screening_Type

  ! possible values to be used inside the above type
  ! see table [to be writtten] in section 4.3.2 of the iodd
  integer(IntAuc), parameter :: L1B_RRC_Cal_OK        = 0
  integer(IntAuc), parameter :: L1B_RRC_Cal_NotOK     = 1
  ! this is our missing value, its not defined in the iodd
  integer(IntAuc), parameter :: L1B_RRC_Cal_Undefined = 255
  !  #]

  TYPE L1B_ZWC_Screening_Type
     !  #[ see table 29 in section 4.3.2 of the iodd
     integer(IntAuc) :: Calibration_Valid ! overall validity
     ! others?
     ! 5 spare bytes
  END TYPE L1B_ZWC_Screening_Type

  ! possible values to be used inside the above type
  ! see table [to be writtten] in section 4.3.2 of the iodd
  integer(IntAuc), parameter :: L1B_ZWC_Cal_OK        = 0
  integer(IntAuc), parameter :: L1B_ZWC_Cal_NotOK     = 1
  ! this is our missing value, its not defined in the iodd
  integer(IntAuc), parameter :: L1B_ZWC_Cal_Undefined = 255
  !  #]

  TYPE L1B_RBC_Screening_Type
     !  #[ see table 30 in section 4.3.2 of the iodd
     integer(IntAuc) :: Calibration_Valid ! overall validity
     ! others?
     ! 5 spare bytes
  END TYPE L1B_RBC_Screening_Type

  ! possible values to be used inside the above type
  ! see table [to be writtten] in section 4.3.2 of the iodd
  integer(IntAuc), parameter :: L1B_RBC_Cal_OK        = 0
  integer(IntAuc), parameter :: L1B_RBC_Cal_NotOK     = 1
  ! this is our missing value, its not defined in the iodd
  integer(IntAuc), parameter :: L1B_RBC_Cal_Undefined = 255
  !  #]

  TYPE L1B_Input_Screening_Type
     !  #[ see table 17 in section 4.3.2 of the iodd
     integer(IntAuc) :: Obs_Screening
     integer(IntAuc) :: Obs_Screening_Flags1 ! to be defined later
     integer(IntAuc) :: Obs_Screening_Flags2 ! to be defined later
     integer(IntAuc) :: Obs_Screening_Flags3 ! to be defined later
     integer(IntAuc) :: Obs_Screening_Flags4 ! to be defined later
     integer(IntAuc) :: Obs_Screening_Flags5 ! to be defined later
     TYPE(L1B_Mie_Meas_Screening_Type),  dimension(:), pointer :: Mie_Meas
     TYPE(L1B_Rayl_Meas_Screening_Type), dimension(:), pointer :: Rayl_Meas
     TYPE(L1B_MRC_Screening_Type) :: MRC_Screening
     TYPE(L1B_RRC_Screening_Type) :: RRC_Screening
     TYPE(L1B_ZWC_Screening_Type) :: ZWC_Screening
     TYPE(L1B_RBC_Screening_Type) :: RBC_Screening
     ! 20 spare bytes
  END TYPE L1B_Input_Screening_Type

  ! possible values to be used inside the above enum type
  ! file reading related
  ! see table 18 in section 4.3.2 of the iodd
  integer(IntAuc), parameter :: L1B_File_OK       = 0
  integer(IntAuc), parameter :: L1B_HDR_Corrupted = 1
  integer(IntAuc), parameter :: L1B_BRC_Corrupted = 2
  ! etc 
  ! laser related
  integer(IntAuc), parameter :: L1B_Laser_Freq_Unlocked   = 10
  integer(IntAuc), parameter :: L1B_Ref_Pulses_Unlocked   = 11
  integer(IntAuc), parameter :: L1B_Laser_Freq_Offset     = 12
  integer(IntAuc), parameter :: L1B_Laser_UV_Energy       = 13
  integer(IntAuc), parameter :: L1B_Laser_Freq_Offs_Stdev = 14
  integer(IntAuc), parameter :: L1B_Laser_UV_Energy_Stdev = 15
  integer(IntAuc), parameter :: L1B_Mie_Mean_Emit_Freq    = 16
  integer(IntAuc), parameter :: L1B_Mie_Emit_Freq_Stdev   = 17
  integer(IntAuc), parameter :: L1B_Rayl_Mean_Emit_Freq   = 18
  integer(IntAuc), parameter :: L1B_Rayl_Emit_Freq_Stdev  = 19
  ! etc
  ! satellite related
  integer(IntAuc), parameter :: L1B_Sat_Not_on_Target     = 20
  ! etc
  ! measurement related
  integer(IntAuc), parameter :: L1B_Mie_Corrupt             = 30 
  integer(IntAuc), parameter :: L1B_Rayleigh_Corrupt        = 31
  integer(IntAuc), parameter :: L1B_Mie_Ref_Pulses_Corrupt  = 32
  integer(IntAuc), parameter :: L1B_Rayl_Ref_Pulses_Corrupt = 33
  ! etc
  ! combinations
  integer(IntAuc), parameter :: L1B_Mie_Invalid_Meas        = 40
  integer(IntAuc), parameter :: L1B_Mie_Invalid_Ref_Pulses  = 41
  integer(IntAuc), parameter :: L1B_Rayl_Invalid_Meas       = 42
  integer(IntAuc), parameter :: L1B_Rayl_Invalid_Ref_Pulses = 43
  integer(IntAuc), parameter :: L1B_Invalid_Num_Mie_Peaks   = 44
  ! this is our missing value, its not defined in the iodd
  integer(IntAuc), parameter :: L1B_Obs_Scr_undefined       = 255
  ! etc
     !  #]
  !  #]
  !---------------------------------------------------------------------
  !  #[ datastructures for L2B_AMD_Screening
  !---------------------------------------------------------------------
  TYPE L2B_AMD_Collocations_Type
     !  #[ see table 33 in section 4.3.2 of the iodd
     integer(IntAuc) :: L2B_AMD_Collocation_QC ! enum field
     integer(IntAuc) :: L2B_AMD_Collocation_QC_Flags
     ! 1 spare byte
  END TYPE L2B_AMD_Collocations_Type

  ! possible values to be used inside the above enum type
  ! see table 34 in section 4.3.2 of the iodd
  integer(IntAuc), parameter :: L2B_AMD_Coll_OK        = 0
  integer(IntAuc), parameter :: L2B_AMD_Coll_Failed    = 1
  ! this is our missing value, its not defined in the iodd
  integer(IntAuc), parameter :: L2B_AMD_Coll_Undefined = 255
  !  #]

  TYPE L2B_AMD_Screening_Type
     !  #[ see table 31 in section 4.3.2 of the iodd
     integer(IntAuc) :: L2B_AMD_Screening_QC ! enum field
     integer(IntAuc) :: L2B_AMD_Screening_QC_Flags ! to be defined later
     TYPE(L2B_AMD_Collocations_Type), &
          dimension(:), pointer :: L2B_AMD_Collocations
     ! 20 spare bytes
  END TYPE L2B_AMD_Screening_Type

  ! possible values to be used inside the above enum type
  ! see table 32 in section 4.3.2 of the iodd
  integer(IntAuc), parameter :: L2B_AMD_OK               = 0
  integer(IntAuc), parameter :: L2B_AMD_Corrupt_Header   = 1
  integer(IntAuc), parameter :: L2B_AMD_Corrupt_Profile  = 2
  integer(IntAuc), parameter :: L2B_AMD_Unlikely_Profile = 3
  ! this is our missing value, its not defined in the iodd
  integer(IntAuc), parameter :: L2B_AMD_Undefined        = 255
     !  #]
  !  #]
  !---------------------------------------------------------------------
  !  #[ datastructures for Mie_Classification_QC
  !---------------------------------------------------------------------
  TYPE L2B_Mie_Bin_Classific_Type
     !  #[ see table 38 in section 4.3.2 of the iodd
     ! remark:
     ! maybe this should be split into 2 different bin types, 
     ! one for Meas and one for Obs ? Currently this is common code
     ! for both types of bins
     integer(IntAuc) :: flag1
     integer(IntAuc) :: flag2
     real(FAdoxy)    :: L2B_Reliability
     ! 1 spare byte
  END TYPE L2B_Mie_Bin_Classific_Type
     !  #]
  TYPE L2B_Mie_Meas_Classific_Type
     !  #[ see table 36 in section 4.3.2 of the iodd
     TYPE(L2B_Mie_Bin_Classific_Type), &
          dimension(:), pointer :: L2B_Mie_Meas_Bin_Classific
     ! 1 spare byte
  END TYPE L2B_Mie_Meas_Classific_Type
     !  #]
  TYPE L2B_Mie_Obs_Classific_Type
     !  #[ see table 37 in section 4.3.2 of the iodd
     integer(IntAuc) :: flag
     TYPE(L2B_Mie_Bin_Classific_Type), &
          dimension(:), pointer :: L2B_Mie_Obs_Bin_Classific
     ! 1 spare byte
  END TYPE L2B_Mie_Obs_Classific_Type
     !  #]
  TYPE Mie_Classification_QC_Type
     !  #[ see table 35 in section 4.3.2 of the iodd
     TYPE(L2B_Mie_Meas_Classific_Type), &
          dimension(:), pointer :: L2B_Mie_Meas_Classific
     TYPE(L2B_Mie_Obs_Classific_Type), &
          dimension(:), pointer :: L2B_Mie_Obs_Classific
     ! 20 spare bytes
  END TYPE Mie_Classification_QC_Type
     !  #]
  !  #]
  !---------------------------------------------------------------------
  !  #[ datastructures for Rayl_Classification_QC
  !---------------------------------------------------------------------
  TYPE L2B_Rayl_Bin_Classific_Type
     !  #[ see table 42 in section 4.3.2 of the iodd
     ! remark:
     ! maybe this should be split into 2 different bin types, 
     ! one for Meas and one for Obs ? Currently this is common code
     ! for both types of bins
     integer(IntAuc) :: flag1
     integer(IntAuc) :: flag2
     real(FAdoxy)    :: L2B_Reliability
     ! 1 spare byte
  END TYPE L2B_Rayl_Bin_Classific_Type
     !  #]
  TYPE L2B_Rayl_Meas_Classific_Type
     !  #[ see table 40 in section 4.3.2 of the iodd
     TYPE(L2B_Rayl_Bin_Classific_Type), &
          dimension(:), pointer :: L2B_Rayl_Meas_Bin_Classific
     ! 1 spare byte
  END TYPE L2B_Rayl_Meas_Classific_Type
     !  #]
  TYPE L2B_Rayl_Obs_Classific_Type
     !  #[ see table 41 in section 4.3.2 of the iodd
     integer(IntAuc) :: flag
     TYPE(L2B_Rayl_Bin_Classific_Type), &
          dimension(:), pointer :: L2B_Rayl_Obs_Bin_Classific
     ! 1 spare byte
  END TYPE L2B_Rayl_Obs_Classific_Type
     !  #]
  TYPE Rayl_Classification_QC_Type
     !  #[ see table 39 in section 4.3.2 of the iodd
     TYPE(L2B_Rayl_Meas_Classific_Type), &
          dimension(:), pointer :: L2B_Rayl_Meas_Classific
     TYPE(L2B_Rayl_Obs_Classific_Type), &
          dimension(:), pointer :: L2B_Rayl_Obs_Classific
     ! 20 spare bytes
  END TYPE Rayl_Classification_QC_Type
     !  #]
  !  #]
  !---------------------------------------------------------------------
  !  #[ datastructures for Mie_Processing_QC
  !---------------------------------------------------------------------
  TYPE Obs_Mie_Bin_QC_Type
     !  #[ see table 45 in section 4.3.2 of the iodd
     real(FAdoxy)    :: Hlos_Error_Estimate ! [m/s]
     real(FAdoxy)    :: Reliability_Number
     integer(IntAuc) :: flags
     real(FAdoxy)    :: Fitting_Amplitude
     real(FAdoxy)    :: Fitting_Residual
     real(FAdoxy)    :: Fitting_Offset
     real(FAdoxy)    :: Fitting_FWHM
     real(FAdoxy)    :: Mie_SNR
     real(FAdoxy)    :: Extinction ! [1/m]
     real(FAdoxy)    :: Scattering_Ratio
     real(FAdoxy)    :: Variability_Measure
     ! 1 spare byte
  END TYPE Obs_Mie_Bin_QC_Type
     !  #]
  TYPE Obs_Mie_QC_Type
     !  #[ see table 44 in section 4.3.2 of the iodd
     TYPE(Obs_Mie_Bin_QC_Type), dimension(:), pointer :: Obs_Mie_Bin_QC
     ! 1 spare byte
  END TYPE Obs_Mie_QC_Type
     !  #]
  TYPE Mie_Processing_QC_Type
     !  #[ see table 43 in section 4.3.2 of the iodd
     TYPE(Obs_Mie_QC_Type), dimension(:), pointer :: Obs_Mie_QC
     integer(IntAuc) :: Mie_Background_High
     ! 20 spare bytes
  END TYPE Mie_Processing_QC_Type
     !  #]
  !  #]
  !---------------------------------------------------------------------
  !  #[ datastructures for Rayleigh_Processing_QC
  !---------------------------------------------------------------------
  TYPE Obs_Rayleigh_Bin_QC_Type
     !  #[ see table 48 in section 4.3.2 of the iodd
     real(FAdoxy)    :: Hlos_Error_Estimate     ! [m/s]
     integer(IntAuc) :: flags1
     integer(IntAuc) :: flags2
     integer(IntAuc) :: flags3
     real(FAdoxy)    :: Extinction              ! [1/m]
     real(FAdoxy)    :: Scattering_Ratio
     real(FAdoxy)    :: Variability_Measure
     integer(IntAuc) :: Extinction_Method       ! enum field
     integer(IntAuc) :: Scattering_Ratio_Method ! enum field
     ! 1 spare byte
  END TYPE Obs_Rayleigh_Bin_QC_Type

  ! possible values for the above enum types
  ! see table 49 in section 4.3.2 of the iodd
  ! TODO: fill these items in the IODD table
  integer(IntAuc), parameter :: Extinction_Meth_Undefined = 0
  integer(IntAuc), parameter :: Extinction_Not_Available  = 1
  integer(IntAuc), parameter :: Extinction_From_NWP_Comp  = 2
  !...

  ! see table 50 in section 4.3.2 of the iodd
  ! TODO: fill these items in the IODD table
  integer(IntAuc), parameter :: Scat_Ratio_Meth_Undefined    = 0
  integer(IntAuc), parameter :: Scat_Ratio_From_2_Channels   = 1
  integer(IntAuc), parameter :: Scat_Ratio_Ass_Mie_ext_Ratio = 2
  ! ...
     !  #]
  TYPE Obs_Rayleigh_QC_Type
     !  #[ see table 47 in section 4.3.2 of the iodd
     TYPE(Obs_Rayleigh_Bin_QC_Type), &
          dimension(:), pointer :: Obs_Rayleigh_Bin_QC
     ! 1 spare byte
  END TYPE Obs_Rayleigh_QC_Type
     !  #]
  TYPE Rayleigh_Processing_QC_Type
     !  #[ see table 46 in section 4.3.2 of the iodd
     TYPE(Obs_Rayleigh_QC_Type), dimension(:), pointer :: Obs_Rayleigh_QC
     integer(IntAuc) :: Rayleigh_Background_High
     ! 20 spare bytes
  END TYPE Rayleigh_Processing_QC_Type
     !  #]
  !  #]
  !---------------------------------------------------------------------

  TYPE L2B_PCD_ADS_type 
     !  #[ see table 16 in section 4.3.2 of the iodd
     integer(IntAs)      :: N_Meas            ! actual nr of meas. for this BRC
     integer(IntAs)      :: N_Obs_Mie_Actual  ! actual nr of Obs. for this BRC
     integer(IntAs)      :: N_Obs_Rayleigh_Actual ! dito
! REMARK: the IntAs type of the above 3 items is not consistent with the
! IntAuc type used in the l2bc_sph for M_Mie, M_Rayleigh and M_Meas
! Should we make this consistent during the clean-up ?
! (for the code it has almost no effect, apart from a few int(*,IntAuc)
!  conversions that may be removed when we change te types)
! JK 9-mar-2007
     TYPE(L1B_Input_Screening_Type)    :: L1B_Input_Screening
     TYPE(L2B_AMD_Screening_Type)      :: L2B_AMD_Screening
     TYPE(Mie_Classification_QC_Type)  :: Mie_Classification_QC
     TYPE(Rayl_Classification_QC_Type) :: Rayl_Classification_QC
     TYPE(Mie_Processing_QC_Type)      :: Mie_Processing_QC
     TYPE(Rayleigh_Processing_QC_Type) :: Rayleigh_Processing_QC
     ! 20 spare bytes
  END TYPE L2B_PCD_ADS_type
     !  #]

  !  #]
contains
  !---------------------------------------------------------------------
  ! subroutines for the L1B_Input_Screening sub-datastructure    
  !---------------------------------------------------------------------
  !  #[

  subroutine Init_L1B_Mie_Meas_Bin_Scr(Bin_Scr)
    !  #[
    TYPE(L1B_Mie_Meas_Bin_Scr_Type), intent(out) :: Bin_Scr

    Bin_Scr%Bin_QC       = L1B_Mie_Bin_undefined
    Bin_Scr%Bin_QC_Flags = 0_IntAuc

  end subroutine Init_L1B_Mie_Meas_Bin_Scr
  !  #]
  subroutine Print_L1B_Mie_Meas_Bin_Scr(Bin_Scr)
    !  #[
    TYPE(L1B_Mie_Meas_Bin_Scr_Type), intent(in) :: Bin_Scr

    print *,"       ...%Bin_QC       = ",Bin_Scr%Bin_QC
    print *,"       ...%Bin_QC_Flags = ",Bin_Scr%Bin_QC_Flags

  end subroutine Print_L1B_Mie_Meas_Bin_Scr
    !  #]
  subroutine AsciiWrite_L1B_Mie_Meas_Bin_Scr(Bin_Scr,fileunit,&
                                             prefix,error_flag)
    !  #[
    TYPE(L1B_Mie_Meas_Bin_Scr_Type), intent(in)  :: Bin_Scr    ! input
    integer,                         intent(in)  :: fileunit   ! input
    character(len=*),                intent(in)  :: prefix     ! input
    integer,                         intent(out) :: error_flag ! output
    
    ! local variables
    character(len=maxlen_IntAuc)      :: str_IntAuc
    character(len=maxlen_IntAucFlags) :: str_IntAucFlags

    error_flag = no_error

    IF (Bin_Scr%Bin_QC .eq. L1B_Mie_Bin_undefined) THEN
       str_IntAuc = "[mi]"
    ELSE
       call SetIntAuc(Bin_Scr%Bin_QC,str_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
    END IF
    write(fileunit,"(a)",err=999) trim(prefix)//"%Bin_QC = "//trim(str_IntAuc)

    call SetIntAucFlags(Bin_Scr%Bin_QC_Flags,str_IntAucFlags,error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) trim(prefix)//"%Bin_QC_Flags = "//&
         trim(str_IntAucFlags)

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_L1B_Mie_Meas_BinScr: "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L1B_Mie_Meas_Bin_Scr
    !  #]
  function GetByteSize_L1B_Mie_MeasBin_Scr() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result

    c_bytesize = 2*c_bytelen_IntAuc ! 2 items in total

    ! typical sizes
    ! memory size = 2*2 = 4 bytes
    ! file   size = 2*1 = 2 bytes
  end function GetByteSize_L1B_Mie_MeasBin_Scr
    !  #]
  subroutine Write_L1B_Mie_Meas_Bin_Scr(Bin_Scr,fileunit,error_flag)
    !  #[
    !  Writes L1B_Mie_Meas_Bin_Screening in binary format
    !  P. Poli     09-01-2006 Original Code
    integer,                         intent(in)  :: fileunit
    TYPE(L1B_Mie_Meas_Bin_Scr_Type), intent(in)  :: Bin_Scr
    integer,                         intent(out) :: error_flag 

    !-----------------------------------
    ! writing the Bin_Screening elements
    !-----------------------------------

    error_flag = no_error

    call write_IntAuc_to_bin_file(fileunit,Bin_Scr%Bin_QC, &
         error_flag)
    IF (error_flag .ne. no_error) goto 998
    call write_IntAuc_to_bin_file(fileunit,Bin_Scr%Bin_QC_Flags, &
         error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998  call logmsg(log_error,"ABORTING subroutine Write_L1B_Mie_Meas_Bin_Scr")
    return

  end subroutine Write_L1B_Mie_Meas_Bin_Scr
    !  #]
  subroutine Read_L1B_Mie_Meas_Bin_Scr(fileunit,Bin_Screening,error_flag)
    !  #[
    !  J. de Kloe  03-01-2006 Original Code. 
    !  P. Poli     10-01-2006 Mods in line with Write_L1B_Mie_Meas_Bin
    integer,                         intent(in)    :: fileunit
    TYPE(L1B_Mie_Meas_Bin_Scr_Type), intent(inout) :: Bin_Screening
    integer,                         intent(out)   :: error_flag

    !-----------------------------------
    ! Reading the Bin_Screening elements
    !-----------------------------------

    error_flag = no_error

    call read_IntAuc_from_bin_file(fileunit,Bin_Screening%Bin_QC,&
         error_flag)
    IF (error_flag .ne. no_error) goto 998

    call read_IntAuc_from_bin_file(fileunit,Bin_Screening%Bin_QC_Flags,&
         error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Read_L1B_Mie_Meas_Bin_Scr")
    return

  end subroutine Read_L1B_Mie_Meas_Bin_Scr
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_L1B_Rayl_Meas_Bin_Scr(Bin_Scr)
    !  #[
    TYPE(L1B_Rayl_Meas_Bin_Scr_Type), intent(out) :: Bin_Scr

    Bin_Scr%Bin_QC       = L1B_Rayl_Bin_undefined
    Bin_Scr%Bin_QC_Flags = 0_IntAuc

  end subroutine Init_L1B_Rayl_Meas_Bin_Scr
  !  #]
  subroutine Print_L1B_Rayl_Meas_Bin_Scr(Bin_Scr)
    !  #[
    TYPE(L1B_Rayl_Meas_Bin_Scr_Type), intent(in) :: Bin_Scr

    print *,"       ...%Bin_QC       = ",Bin_Scr%Bin_QC
    print *,"       ...%Bin_QC_Flags = ",Bin_Scr%Bin_QC_FLags

  end subroutine Print_L1B_Rayl_Meas_Bin_Scr
    !  #]
  subroutine AsciiWrite_L1B_RaylMeasBin_Scr(Bin_Scr,fileunit,&
                                             prefix,error_flag)
    !  #[
    TYPE(L1B_Rayl_Meas_Bin_Scr_Type), intent(in)  :: Bin_Scr    ! input
    integer,                          intent(in)  :: fileunit   ! input
    character(len=*),                 intent(in)  :: prefix     ! input
    integer,                          intent(out) :: error_flag ! output

    ! local variables
    character(len=maxlen_IntAuc)      :: str_IntAuc
    character(len=maxlen_IntAucFlags) :: str_IntAucFlags

    error_flag = no_error

    IF (Bin_Scr%Bin_QC .eq. L1B_Rayl_Bin_undefined) THEN
       str_IntAuc = "[mi]"
    ELSE
       call SetIntAuc(Bin_Scr%Bin_QC,str_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
    END IF
    write(fileunit,"(a)",err=999) trim(prefix)//"%Bin_QC = "//trim(str_IntAuc)

    call SetIntAucFlags(Bin_Scr%Bin_QC_Flags,str_IntAucFlags,error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) trim(prefix)//"%Bin_QC_Flags = "//&
         trim(str_IntAucFlags)

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_L1B_RaylMeas_BinScr: "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L1B_RaylMeasBin_Scr
    !  #]
  function GetByteSize_L1B_RaylMeasBin_Scr() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result

    c_bytesize = 2*c_bytelen_IntAuc ! 2 items in total

    ! typical sizes
    ! memory size = 2*2 = 4 bytes
    ! file   size = 2*1 = 2 bytes
  end function GetByteSize_L1B_RaylMeasBin_Scr
    !  #]
  subroutine Write_L1B_Rayl_Meas_Bin_Scr(Bin_Screening,fileunit,error_flag)
    !  #[
    !  Writes L1B_Rayl_Meas_Bin_Screening in binary format
    !  P. Poli     09-01-2006 Original Code
    integer,                          intent(in)  :: fileunit
    TYPE(L1B_Rayl_Meas_Bin_Scr_Type), intent(in)  :: Bin_Screening
    integer,                          intent(out) :: error_flag 

    !-----------------------------------
    ! writing the Bin_Screening elements
    !-----------------------------------

    error_flag = no_error

    call write_IntAuc_to_bin_file(fileunit,Bin_Screening%Bin_QC, error_flag)
    IF (error_flag .ne. no_error) goto 998

    call write_IntAuc_to_bin_file(fileunit,Bin_Screening%Bin_QC_Flags, &
                                  error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Write_L1B_Rayl_Meas_Bin_Scr")
    return

  end subroutine Write_L1B_Rayl_Meas_Bin_Scr
    !  #]
  subroutine Read_L1B_Rayl_Meas_Bin_Scr(fileunit,Bin_Screening,error_flag)
    !  #[
    !  P. Poli     10-01-2006 Original Code. 
    integer,                          intent(in)    :: fileunit
    TYPE(L1B_Rayl_Meas_Bin_Scr_Type), intent(inout) :: Bin_Screening
    integer,                          intent(out)   :: error_flag

    !-----------------------------------
    ! Reading the Bin_Screening elements
    !-----------------------------------

    error_flag = no_error

    call read_IntAuc_from_bin_file(fileunit,Bin_Screening%Bin_QC,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call read_IntAuc_from_bin_file(fileunit,Bin_Screening%Bin_QC_Flags,&
                                   error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Read_L1B_Rayl_Meas_Bin_Scr")
    return

  end subroutine Read_L1B_Rayl_Meas_Bin_Scr
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_Mie_Meas_Screening(Mie_Meas, N_Range_Gates0, error_flag)
    !  #[
    TYPE(L1B_Mie_Meas_Screening_Type), intent(out) :: Mie_Meas ! output
    integer(IntAs), intent(in)  :: N_Range_Gates0 ! input
    integer,        intent(out) :: error_flag    ! output

    ! local variable
    integer :: r, AllocStatus

    error_flag = no_error

    Mie_Meas%Meas_QC = L1B_Mie_Meas_Undefined
    Mie_Meas%Meas_QC_Flags = 0_IntAuc

    IF (N_Range_Gates0 .eq. 0) THEN
       ! init them to be empty
       nullify(Mie_Meas%Bin_Screening)
    ELSE
       allocate(Mie_Meas%Bin_Screening(N_Range_Gates0),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          call logmsg(log_error,"Allocation failure in subroutine "//&
                                "Init_Mie_Meas_Screening()")
          error_flag = error_allocate
          return
       END IF
       
       DO r=1,N_Range_Gates0
          call Init_L1B_Mie_Meas_Bin_Scr(Mie_Meas%Bin_Screening(r))
       END DO
    END IF

  end subroutine Init_Mie_Meas_Screening
    !  #]
  subroutine Delete_Mie_Meas_Screening(Mie_Meas)
    !  #[
    TYPE(L1B_Mie_Meas_Screening_Type), intent(inout) :: Mie_Meas

    IF (associated(Mie_Meas%Bin_Screening)) &
         deallocate(Mie_Meas%Bin_Screening)

  end subroutine Delete_Mie_Meas_Screening
    !  #]
  subroutine Print_Mie_Meas_Screening(Mie_Meas)
    !  #[
    TYPE(L1B_Mie_Meas_Screening_Type), intent(in) :: Mie_Meas

    integer :: r,s

    print *,"   ...%Meas_QC       = ",Mie_Meas%Meas_QC
    print *,"   ...%Meas_QC_Flags = ",Mie_Meas%Meas_QC_Flags

    IF (associated(Mie_Meas%Bin_Screening)) THEN
       ! get the size of the array
       s = size(Mie_Meas%Bin_Screening)
       print "(a,i2,a)","    ==> size(Mie_Meas%Bin_Screening) = ",s,&
                        " [printing first and last one]"
       DO r=1,s,s-1
          print "('        Mie_Meas%Bin_Screening(',i2.2,')%...')",r
          call Print_L1B_Mie_Meas_Bin_Scr(Mie_Meas%Bin_Screening(r))
       END DO
    ELSE
       print *,"        Mie_Meas%Bin_Screening(:) [seems not associated]"
    END IF

  end subroutine Print_Mie_Meas_Screening
    !  #]
  subroutine AsciiWrite_Mie_Meas_Screening(Mie_Meas,fileunit,prefix,&
                                           PrintAll,error_flag)
    !  #[
    TYPE(L1B_Mie_Meas_Screening_Type), intent(in)  :: Mie_Meas   ! input
    integer,                           intent(in)  :: fileunit   ! input
    character(len=*),                  intent(in)  :: prefix     ! input
    logical,                           intent(in)  :: PrintAll   ! input
    integer,                           intent(out) :: error_flag ! output

    ! local variables
    integer                           :: r,s,step
    character(len=maxlen_IntAuc)      :: str_IntAuc
    character(len=maxlen_IntAucFlags) :: str_IntAucFlags
    character(len=2)                  :: str_cntr

    error_flag = no_error

    IF (Mie_Meas%Meas_QC .eq. L1B_Mie_Meas_Undefined) THEN
       str_IntAuc = "[mi]"
    ELSE
       call SetIntAuc(Mie_Meas%Meas_QC,str_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
    END IF
    write(fileunit,"(a)",err=999) trim(prefix)//"%Meas_QC = "//trim(str_IntAuc)

    call SetIntAucFlags(Mie_Meas%Meas_QC_Flags,str_IntAucFlags,error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) trim(prefix)//"%Meas_QC_Flags = "//&
         trim(str_IntAucFlags)

    IF (associated(Mie_Meas%Bin_Screening)) THEN
       ! get the size of the array
       s = size(Mie_Meas%Bin_Screening)
       IF (PrintAll .or. (s .le. 1) ) THEN
          step=1
       ELSE
          step = s-1
       END IF
       DO r=1,s,step
          write(str_cntr,"(i2.2)",err=998) r
          call AsciiWrite_L1B_Mie_Meas_Bin_Scr(Mie_Meas%Bin_Screening(r),&
               fileunit,&
               trim(prefix)//"%Bin_Screening("//trim(str_cntr)//")",&
               error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    ELSE
       write(fileunit,"(a)",err=999) &
            trim(prefix)//"%Bin_Screening(:) [seems not associated]"
    END IF

    return

998 call logmsg(log_error,"conversion error in "//&
         "AsciiWrite_Mie_Meas_Screening: could not convert counter r "//&
         " to a string ..")
    error_flag = error_writing_file
    return
    
999 call logmsg(log_error,"ERROR in AsciiWrite_Mie_Meas_Screening: "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Mie_Meas_Screening
    !  #]
  function GetByteSize_Mie_Meas_Screening(Mie_Meas) result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result
    TYPE(L1B_Mie_Meas_Screening_Type), intent(in) :: Mie_Meas

    c_bytesize = 2*c_bytelen_IntAuc + 1 ! spare_byte
    IF (associated(Mie_Meas%Bin_Screening)) THEN
       c_bytesize = c_bytesize + &
            int(size(Mie_Meas%Bin_Screening)*&
                GetByteSize_L1B_Mie_MeasBin_Scr(),IntAl)
    END IF

    ! typical sizes
    ! memory size = 2*2  +24*4 = 100 bytes
    ! file   size = 2*1+1+24*2 =  51 bytes
  end function GetByteSize_Mie_Meas_Screening
    !  #]
  subroutine Write_L1B_Mie_Meas_Screening(Mie_Meas,fileunit,error_flag)
    !  #[
    !  Writes L1B_Mie_Meas_Screening in binary format
    !  P. Poli     09-01-2006 Original Code
    integer,                           intent(in)  :: fileunit
    TYPE(L1B_Mie_Meas_Screening_Type), intent(in)  :: Mie_Meas
    integer,                           intent(out) :: error_flag 

    ! local variable
    integer :: s, i, nr_spare_bytes

    !------------------------------
    ! writing the Mie_Meas elements
    !------------------------------

    error_flag = no_error

    call write_IntAuc_to_bin_file(fileunit,Mie_Meas%Meas_QC,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_IntAuc_to_bin_file(fileunit,Mie_Meas%Meas_QC_Flags, &
                                  error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    IF (.not. associated(Mie_Meas%Bin_Screening)) THEN
       call logmsg(log_error,"ERROR in Write_L1B_Mie_Meas_Screening(): "//&
            "The array Mie_Meas%Bin_Screening seems not allocated !!")
       error_flag = error_programming
       return
    END IF
    s = size(Mie_Meas%Bin_Screening)
    DO i=1,s
       call Write_L1B_Mie_Meas_Bin_Scr(Mie_Meas%Bin_Screening(i), &
            fileunit,error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    ! 1 spare byte
    nr_spare_bytes = 1
    call write_SpareBytes_to_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Write_L1B_Mie_Meas_Screening")
    return

  end subroutine Write_L1B_Mie_Meas_Screening
    !  #]
  subroutine Read_L1B_Mie_Meas_Screening(fileunit,Mie_Meas,error_flag)
    !  #[
    !  P. Poli     10-01-2006 Original code.
    integer,                           intent(in)    :: fileunit
    TYPE(L1B_Mie_Meas_Screening_Type), intent(inout) :: Mie_Meas
    integer,                           intent(out)   :: error_flag

    ! local variable
    integer :: s, i, nr_spare_bytes

    !-----------------------------------
    ! Reading the Mie_Meas elements
    !-----------------------------------

    error_flag = no_error

    call read_IntAuc_from_bin_file(fileunit,Mie_Meas%Meas_QC,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call read_IntAuc_from_bin_file(fileunit,Mie_Meas%Meas_QC_Flags,error_flag)
    IF (error_flag .ne. no_error) goto 998

    IF (.not. associated(Mie_Meas%Bin_Screening)) THEN
       call logmsg(log_error,"ERROR in Read_L1B_Mie_Meas_Screening(): "//&
            "The array Mie_Meas%Bin_Screening seems not allocated !!")
       error_flag = error_programming
       return
    END IF

    s = size(Mie_Meas%Bin_Screening)
    DO i=1,s
      call Read_L1B_Mie_Meas_Bin_Scr(fileunit,Mie_Meas%Bin_Screening(i), &
                                     error_flag)
      IF (error_flag .ne. no_error) goto 998
    ENDDO

    ! 1 spare byte
    nr_spare_bytes = 1
    call read_SpareBytes_from_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Read_L1B_Mie_Meas_Screening")
    return

  end subroutine Read_L1B_Mie_Meas_Screening
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_Rayl_Meas_Screening(Rayl_Meas, N_Range_Gates0, error_flag)
    !  #[
    TYPE(L1B_Rayl_Meas_Screening_Type), intent(out) :: Rayl_Meas ! output
    integer(IntAs), intent(in)  :: N_Range_Gates0 ! input
    integer,        intent(out) :: error_flag     ! output

    ! local variable
    integer :: r, AllocStatus

    error_flag = no_error

    Rayl_Meas%Meas_QC = L1B_Rayl_Meas_Undefined
    Rayl_Meas%Meas_QC_Flags = 0_IntAuc

    IF (N_Range_Gates0 .eq. 0) THEN
       ! init them to be empty
       nullify(Rayl_Meas%Bin_Screening)
    ELSE
       allocate(Rayl_Meas%Bin_Screening(N_Range_Gates0),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          call logmsg(log_error,"Allocation failure in subroutine "//&
                                "Init_Rayl_Meas_Screening()")
          error_flag = error_allocate
          return
       END IF
       
       DO r=1,N_Range_Gates0
          call Init_L1B_Rayl_Meas_Bin_Scr(Rayl_Meas%Bin_Screening(r))
       END DO
    END IF

  end subroutine Init_Rayl_Meas_Screening
    !  #]
  subroutine Delete_Rayl_Meas_Screening(Rayl_Meas)
    !  #[
    TYPE(L1B_Rayl_Meas_Screening_Type), intent(inout) :: Rayl_Meas

    IF (associated(Rayl_Meas%Bin_Screening)) &
         deallocate(Rayl_Meas%Bin_Screening)

  end subroutine Delete_Rayl_Meas_Screening
    !  #]
  subroutine Print_Rayl_Meas_Screening(Rayl_Meas)
    !  #[
    TYPE(L1B_Rayl_Meas_Screening_Type), intent(in) :: Rayl_Meas

    integer :: r,s

    print *,"   ...%Meas_QC       = ",Rayl_Meas%Meas_QC
    print *,"   ...%Meas_QC_Flags = ",Rayl_Meas%Meas_QC_Flags

    IF (associated(Rayl_Meas%Bin_Screening)) THEN
       ! get the size of the array
       s = size(Rayl_Meas%Bin_Screening)
       print "(a,i2,a)","    ==> size(Rayl_Meas%Bin_Screening) = ",s,&
                        " [printing first and last one]"
       DO r=1,s,s-1
          print "('        Rayl_Meas%Bin_Screening(',i2.2,')%...')",r
          call Print_L1B_Rayl_Meas_Bin_Scr(Rayl_Meas%Bin_Screening(r))
       END DO
    ELSE
       print *,"        Rayl_Meas%Bin_Screening(:) [seems not associated]"
    END IF

  end subroutine Print_Rayl_Meas_Screening
    !  #]
  subroutine AsciiWrite_Rayl_Meas_Screening(Rayl_Meas,fileunit,prefix,&
                                            PrintAll,error_flag)
    !  #[
    TYPE(L1B_Rayl_Meas_Screening_Type), intent(in)  :: Rayl_Meas  ! input
    integer,                            intent(in)  :: fileunit   ! input
    character(len=*),                   intent(in)  :: prefix     ! input
    logical,                            intent(in)  :: PrintAll   ! input
    integer,                            intent(out) :: error_flag ! output

    ! local variables
    integer                           :: r,s,step
    character(len=maxlen_IntAuc)      :: str_IntAuc
    character(len=maxlen_IntAucFlags) :: str_IntAucFlags
    character(len=2)                  :: str_cntr

    error_flag = no_error

    IF (Rayl_Meas%Meas_QC .eq. L1B_Rayl_Meas_Undefined) THEN
       str_IntAuc = "[mi]"
    ELSE
       call SetIntAuc(Rayl_Meas%Meas_QC,str_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
    END IF
    write(fileunit,"(a)",err=999) trim(prefix)//"%Meas_QC = "//trim(str_IntAuc)

    call SetIntAucFlags(Rayl_Meas%Meas_QC_Flags,str_IntAucFlags,error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) trim(prefix)//"%Meas_QC_Flags = "//&
         trim(str_IntAucFlags)

    IF (associated(Rayl_Meas%Bin_Screening)) THEN
       ! get the size of the array
       s = size(Rayl_Meas%Bin_Screening)
       IF (PrintAll .or. (s .le. 1) ) THEN
          step=1
       ELSE
          step = s-1
       END IF
       DO r=1,s,step
          write(str_cntr,"(i2.2)",err=998) r
          call AsciiWrite_L1B_RaylMeasBin_Scr(Rayl_Meas%Bin_Screening(r),&
               fileunit,&
               trim(prefix)//"%Bin_Screening("//trim(str_cntr)//")",&
               error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    ELSE
       write(fileunit,"(a)",err=999) &
            trim(prefix)//"%Bin_Screening(:) [seems not associated]"
    END IF

    return

998 call logmsg(log_error,"conversion error in "//&
         "AsciiWrite_Rayl_Meas_Screening: could not convert the "//&
         "counter r to a string ..")
    error_flag = error_writing_file
    return
    
999 call logmsg(log_error,"ERROR in AsciiWrite_Rayl_Meas_Screening: "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Rayl_Meas_Screening
    !  #]
  function GetByteSize_Rayl_Meas_Screening(Rayl_Meas) result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result
    TYPE(L1B_Rayl_Meas_Screening_Type), intent(in) :: Rayl_Meas

    c_bytesize = 2*c_bytelen_IntAuc + 1 ! spare_byte
    IF (associated(Rayl_Meas%Bin_Screening)) THEN
       c_bytesize = c_bytesize + &
            int(size(Rayl_Meas%Bin_Screening)*&
                GetByteSize_L1B_RaylMeasBin_Scr(),IntAl)
    END IF

    ! typical sizes
    ! memory size = 2*2  +24*4 = 100 bytes
    ! file   size = 2*1+1+24*2 =  51 bytes
  end function GetByteSize_Rayl_Meas_Screening
    !  #]
  subroutine Write_L1B_Rayl_Meas_Screening(Rayl_Meas,fileunit,error_flag)
    !  #[
    !  Writes L1B_Rayleigh_Meas_Screening in binary format
    !  P. Poli     09-01-2006 Original Code
    integer,                            intent(in)  :: fileunit
    TYPE(L1B_Rayl_Meas_Screening_Type), intent(in)  :: Rayl_Meas
    integer,                            intent(out) :: error_flag 

    ! local variable
    integer :: s, i, nr_spare_bytes

    !-------------------------------
    ! writing the Rayl_Meas elements
    !-------------------------------

    error_flag = no_error

    call write_IntAuc_to_bin_file(fileunit,Rayl_Meas%Meas_QC,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call write_IntAuc_to_bin_file(fileunit,Rayl_Meas%Meas_QC_Flags, &
                                  error_flag)
    IF (error_flag .ne. no_error) goto 998

    IF (.not. associated(Rayl_Meas%Bin_Screening)) THEN
       call logmsg(log_error,"ERROR in Write_L1B_Rayl_Meas_Screening(): "//&
            "The array Rayl_Meas%Bin_Screening seems not allocated !!")
       error_flag = error_programming
       return
    END IF

    s = size(Rayl_Meas%Bin_Screening)
    DO i=1,s
       call Write_L1B_Rayl_Meas_Bin_Scr(Rayl_Meas%Bin_Screening(i), &
            fileunit,error_flag)
       IF (error_flag .ne. no_error) goto 998
    ENDDO

    ! 1 spare byte
    nr_spare_bytes = 1
    call write_SpareBytes_to_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Write_L1B_Rayl_Meas_Screening")
    return

  end subroutine Write_L1B_Rayl_Meas_Screening
    !  #]
  subroutine Read_L1B_Rayl_Meas_Screening(fileunit,Rayl_Meas,error_flag)
    !  #[
    !  P. Poli     10-01-2006 Original code
    integer,                            intent(in)    :: fileunit
    TYPE(L1B_Rayl_Meas_Screening_Type), intent(inout) :: Rayl_Meas
    integer,                            intent(out)   :: error_flag

    ! local variable
    integer :: s, i, nr_spare_bytes

    !-----------------------------------
    ! Reading the Rayl_Meas elements
    !-----------------------------------

    error_flag = no_error

    call read_IntAuc_from_bin_file(fileunit,Rayl_Meas%Meas_QC,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call read_IntAuc_from_bin_file(fileunit,Rayl_Meas%Meas_QC_Flags,error_flag)
    IF (error_flag .ne. no_error) goto 998

    IF (.not. associated(Rayl_Meas%Bin_Screening)) THEN
       call logmsg(log_error,"ERROR in Read_L1B_Rayl_Meas_Screening(): "//&
            "The array Rayl_Meas%Bin_Screening seems not allocated !!")
       error_flag = error_programming
       return
    END IF

    s = size(Rayl_Meas%Bin_Screening)
    DO i=1,s
       call Read_L1B_Rayl_Meas_Bin_Scr(fileunit,Rayl_Meas%Bin_Screening(i), &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    ! 1 spare byte
    nr_spare_bytes = 1
    call read_SpareBytes_from_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Read_L1B_Rayl_Meas_Screening")
    return

  end subroutine Read_L1B_Rayl_Meas_Screening
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_L1B_MRC_Screening(MRC_Screening)
    !  #[
    TYPE(L1B_MRC_Screening_Type), intent(out) :: MRC_Screening

    MRC_Screening%Calibration_Valid = L1B_MRC_Cal_Undefined

  end subroutine Init_L1B_MRC_Screening
  !  #]
  subroutine Print_L1B_MRC_Screening(MRC_Screening)
    !  #[
    TYPE(L1B_MRC_Screening_Type), intent(in) :: MRC_Screening

    print *,"   ...%MRC_Screening = ",MRC_Screening%Calibration_Valid

  end subroutine Print_L1B_MRC_Screening
    !  #]
  subroutine AsciiWrite_L1B_MRC_Screening(MRC_Screening,fileunit,&
                                          prefix,error_flag)
    !  #[
    TYPE(L1B_MRC_Screening_Type), intent(in)  :: MRC_Screening ! input
    integer,                      intent(in)  :: fileunit      ! input
    character(len=*),             intent(in)  :: prefix        ! input
    integer,                      intent(out) :: error_flag    ! output

    ! local variable
    character(len=maxlen_IntAuc)      :: str_IntAuc

    if (MRC_Screening%Calibration_Valid .eq. L1B_MRC_Cal_Undefined) then
       str_IntAuc = "[mi]"
    else
       call SetIntAuc(MRC_Screening%Calibration_Valid,str_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
    end if

    write(fileunit,"(a)",err=999) trim(prefix)//"%Calibration_Valid = "//&
                                  trim(str_IntAuc)

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_L1B_MRC_Screening: "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L1B_MRC_Screening
    !  #]
  function GetByteSize_L1B_MRC_Screening() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result

    c_bytesize = 1*c_bytelen_IntAuc + 5 ! spare_bytes
    ! 1 item only for now

    ! typical sizes
    ! memory size = 2       bytes
    ! file   size = 1+1 = 6 bytes
  end function GetByteSize_L1B_MRC_Screening
    !  #]
  subroutine Write_L1B_MRC_Screening(MRC_Screening,fileunit,error_flag)
    !  #[
    !  Writes L1B_MRC_Screening in binary format
    !  P. Poli     09-01-2006 Original Code
    integer,                        intent(in)  :: fileunit
    TYPE(L1B_MRC_Screening_Type),   intent(in)  :: MRC_Screening
    integer,                        intent(out) :: error_flag 

    ! local variable
    integer :: nr_spare_bytes

    !-----------------------------------
    ! writing the MRC_Screening elements
    !-----------------------------------

    error_flag = no_error

    call write_IntAuc_to_bin_file(fileunit,MRC_Screening%Calibration_Valid, &
                                  error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 5 spare bytes
    nr_spare_bytes = 5
    call write_SpareBytes_to_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Write_L1B_MRC_Screening")
    return

  end subroutine Write_L1B_MRC_Screening
    !  #]
  subroutine Read_L1B_MRC_Screening(fileunit,MRC_Screening,error_flag)
    !  #[
    !  P. Poli     10-01-2006 Original code.
    integer,                       intent(in)  :: fileunit
    TYPE(L1B_MRC_Screening_Type),  intent(inout) :: MRC_Screening
    integer,                       intent(out) :: error_flag

    ! local variable
    integer :: nr_spare_bytes

    !-----------------------------------
    ! Reading the MRC_Screening elements
    !-----------------------------------

    error_flag = no_error

    call read_IntAuc_from_bin_file(fileunit,MRC_Screening%Calibration_Valid, &
                                   error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 5 spare bytes
    nr_spare_bytes = 5
    call read_SpareBytes_from_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Read_L1B_MRC_Screening")
    return

  end subroutine Read_L1B_MRC_Screening
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_L1B_RRC_Screening(RRC_Screening)
    !  #[
    TYPE(L1B_RRC_Screening_Type), intent(out) :: RRC_Screening

    RRC_Screening%Calibration_Valid = L1B_RRC_Cal_Undefined

  end subroutine Init_L1B_RRC_Screening
  !  #]
  subroutine Print_L1B_RRC_Screening(RRC_Screening)
    !  #[
    TYPE(L1B_RRC_Screening_Type), intent(in) :: RRC_Screening

    print *,"   ...%RRC_Screening = ",RRC_Screening%Calibration_Valid

  end subroutine Print_L1B_RRC_Screening
    !  #]
  subroutine AsciiWrite_L1B_RRC_Screening(RRC_Screening,fileunit,&
                                          prefix,error_flag)
    !  #[
    TYPE(L1B_RRC_Screening_Type), intent(in)  :: RRC_Screening ! input
    integer,                      intent(in)  :: fileunit      ! input
    character(len=*),             intent(in)  :: prefix        ! input
    integer,                      intent(out) :: error_flag    ! output

    ! local variable
    character(len=maxlen_IntAuc)      :: str_IntAuc

    if (RRC_Screening%Calibration_Valid .eq. L1B_RRC_Cal_Undefined) then
       str_IntAuc = "[mi]"
    else
       call SetIntAuc(RRC_Screening%Calibration_Valid,str_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
    end if

    write(fileunit,"(a)",err=999) trim(prefix)//"%Calibration_Valid = "//&
                                  trim(str_IntAuc)

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_L1B_RRC_Screening: "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L1B_RRC_Screening
    !  #]
  function GetByteSize_L1B_RRC_Screening() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result

    c_bytesize = 1*c_bytelen_IntAuc + 5 ! spare_bytes
    ! 1 item only for now

    ! typical sizes
    ! memory size = 2       bytes
    ! file   size = 1+5 = 6 bytes
  end function GetByteSize_L1B_RRC_Screening
    !  #]
  subroutine Write_L1B_RRC_Screening(RRC_Screening,fileunit,error_flag)
    !  #[
    !  Writes L1B_RRC_Screening in binary format
    !  P. Poli     09-01-2006 Original Code
    integer,                        intent(in)  :: fileunit
    TYPE(L1B_RRC_Screening_Type),   intent(in)  :: RRC_Screening
    integer,                        intent(out) :: error_flag 

    ! local variable
    integer :: nr_spare_bytes

    !-----------------------------------
    ! writing the RRC_Screening elements
    !-----------------------------------

    error_flag = no_error

    call write_IntAuc_to_bin_file(fileunit,RRC_Screening%Calibration_Valid, &
                                  error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 5 spare bytes
    nr_spare_bytes = 5
    call write_SpareBytes_to_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Write_L1B_RRC_Screening")
    return

  end subroutine Write_L1B_RRC_Screening
    !  #]
  subroutine Read_L1B_RRC_Screening(fileunit,RRC_Screening,error_flag)
    !  #[
    !  P. Poli     10-01-2006 Original code.
    integer,                       intent(in)    :: fileunit
    TYPE(L1B_RRC_Screening_Type),  intent(inout) :: RRC_Screening
    integer,                       intent(out)   :: error_flag

    ! local variable
    integer :: nr_spare_bytes

    !-----------------------------------
    ! Reading the RRC_Screening elements
    !-----------------------------------

    error_flag = no_error

    call read_IntAuc_from_bin_file(fileunit,RRC_Screening%Calibration_Valid, &
                                   error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 5 spare bytes
    nr_spare_bytes = 5
    call read_SpareBytes_from_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Read_L1B_RRC_Screening")
    return

  end subroutine Read_L1B_RRC_Screening
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_L1B_ZWC_Screening(ZWC_Screening)
    !  #[
    TYPE(L1B_ZWC_Screening_Type), intent(out) :: ZWC_Screening

    ZWC_Screening%Calibration_Valid = L1B_ZWC_Cal_Undefined

  end subroutine Init_L1B_ZWC_Screening
  !  #]
  subroutine Print_L1B_ZWC_Screening(ZWC_Screening)
    !  #[
    TYPE(L1B_ZWC_Screening_Type), intent(in) :: ZWC_Screening

    print *,"   ...%ZWC_Screening = ",ZWC_Screening%Calibration_Valid

  end subroutine Print_L1B_ZWC_Screening
    !  #]
  subroutine AsciiWrite_L1B_ZWC_Screening(ZWC_Screening,fileunit,&
                                          prefix,error_flag)
    !  #[
    TYPE(L1B_ZWC_Screening_Type), intent(in)  :: ZWC_Screening ! input
    integer,                      intent(in)  :: fileunit      ! input
    character(len=*),             intent(in)  :: prefix        ! input
    integer,                      intent(out) :: error_flag    ! output

    ! local variable
    character(len=maxlen_IntAuc)      :: str_IntAuc

    if (ZWC_Screening%Calibration_Valid .eq. L1B_ZWC_Cal_Undefined) then
       str_IntAuc = "[mi]"
    else
       call SetIntAuc(ZWC_Screening%Calibration_Valid,str_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
    end if

    write(fileunit,"(a)",err=999) trim(prefix)//"%Calibration_Valid = "//&
                                  trim(str_IntAuc)

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_L1B_ZWC_Screening: "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L1B_ZWC_Screening
    !  #]
  function GetByteSize_L1B_ZWC_Screening() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result

    c_bytesize = 1*c_bytelen_IntAuc + 5 ! spare_bytes
    ! 1 item only for now

    ! typical sizes
    ! memory size = 2       bytes
    ! file   size = 1+5 = 6 bytes
  end function GetByteSize_L1B_ZWC_Screening
    !  #]
  subroutine Write_L1B_ZWC_Screening(ZWC_Screening,fileunit,error_flag)
    !  #[
    !  Writes L1B_ZWC_Screening in binary format
    !  P. Poli     09-01-2006 Original Code
    integer,                      intent(in)  :: fileunit
    TYPE(L1B_ZWC_Screening_Type), intent(in)  :: ZWC_Screening
    integer,                      intent(out) :: error_flag 

    ! local variable
    integer :: nr_spare_bytes

    !-----------------------------------
    ! writing the ZWC_Screening elements
    !-----------------------------------

    error_flag = no_error

    call write_IntAuc_to_bin_file(fileunit,ZWC_Screening%Calibration_Valid, &
                                  error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 5 spare bytes
    nr_spare_bytes = 5
    call write_SpareBytes_to_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Write_L1B_ZWC_Screening")
    return

  end subroutine Write_L1B_ZWC_Screening
    !  #]
  subroutine Read_L1B_ZWC_Screening(fileunit,ZWC_Screening,error_flag)
    !  #[
    !  P. Poli     10-01-2006 Original code.
    integer,                       intent(in)    :: fileunit
    TYPE(L1B_ZWC_Screening_Type),  intent(inout) :: ZWC_Screening
    integer,                       intent(out)   :: error_flag

    ! local variable
    integer :: nr_spare_bytes

    !-----------------------------------
    ! Reading the ZWC_Screening elements
    !-----------------------------------

    error_flag = no_error

    call read_IntAuc_from_bin_file(fileunit,ZWC_Screening%Calibration_Valid, &
                                   error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 5 spare bytes
    nr_spare_bytes = 5
    call read_SpareBytes_from_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Read_L1B_ZWC_Screening")
    return

  end subroutine Read_L1B_ZWC_Screening
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_L1B_RBC_Screening(RBC_Screening)
    !  #[
    TYPE(L1B_RBC_Screening_Type), intent(out) :: RBC_Screening

    RBC_Screening%Calibration_Valid = L1B_RBC_Cal_Undefined

  end subroutine Init_L1B_RBC_Screening
  !  #]
  subroutine Print_L1B_RBC_Screening(RBC_Screening)
    !  #[
    TYPE(L1B_RBC_Screening_Type), intent(in) :: RBC_Screening

    print *,"   ...%RBC_Screening = ",RBC_Screening%Calibration_Valid

  end subroutine Print_L1B_RBC_Screening
    !  #]
  subroutine AsciiWrite_L1B_RBC_Screening(RBC_Screening,fileunit,&
                                          prefix,error_flag)
    !  #[
    TYPE(L1B_RBC_Screening_Type), intent(in)  :: RBC_Screening ! input
    integer,                      intent(in)  :: fileunit      ! input
    character(len=*),             intent(in)  :: prefix        ! input
    integer,                      intent(out) :: error_flag    ! output

    ! local variable
    character(len=maxlen_IntAuc)      :: str_IntAuc

    if (RBC_Screening%Calibration_Valid .eq. L1B_RBC_Cal_Undefined) then
       str_IntAuc = "[mi]"
    else
       call SetIntAuc(RBC_Screening%Calibration_Valid,str_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
    end if

    write(fileunit,"(a)",err=999) trim(prefix)//"%Calibration_Valid = "//&
                                  trim(str_IntAuc)

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_L1B_RBC_Screening: "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L1B_RBC_Screening
    !  #]
  function GetByteSize_L1B_RBC_Screening() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result

    c_bytesize = 1*c_bytelen_IntAuc + 5 ! spare_bytes
    ! 1 item only for now

    ! typical sizes
    ! memory size = 2       bytes
    ! file   size = 1+5 = 6 bytes
  end function GetByteSize_L1B_RBC_Screening
    !  #]
  subroutine Write_L1B_RBC_Screening(RBC_Screening,fileunit,error_flag)
    !  #[
    !  Writes L1B_RBC_Screening in binary format
    !  P. Poli     09-01-2006 Original Code
    integer,                        intent(in)  :: fileunit
    TYPE(L1B_RBC_Screening_Type),   intent(in)  :: RBC_Screening
    integer,                        intent(out) :: error_flag 

    ! local variable
    integer :: nr_spare_bytes

    !-----------------------------------
    ! writing the RBC_Screening elements
    !-----------------------------------

    error_flag = no_error

    call write_IntAuc_to_bin_file(fileunit,RBC_Screening%Calibration_Valid, &
                                  error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 5 spare bytes
    nr_spare_bytes = 5
    call write_SpareBytes_to_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Write_L1B_RBC_Screening")
    return

  end subroutine Write_L1B_RBC_Screening
    !  #]
  subroutine Read_L1B_RBC_Screening(fileunit,RBC_Screening,error_flag)
    !  #[
    !  P. Poli     10-01-2006 Original code.
    integer,                       intent(in)    :: fileunit
    TYPE(L1B_RBC_Screening_Type),  intent(inout) :: RBC_Screening
    integer,                       intent(out)   :: error_flag

    ! local variable
    integer :: nr_spare_bytes

    !-----------------------------------
    ! Reading the RBC_Screening elements
    !-----------------------------------

    error_flag = no_error

    call read_IntAuc_from_bin_file(fileunit,RBC_Screening%Calibration_Valid, &
                                   error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 5 spare bytes
    nr_spare_bytes = 5
    call read_SpareBytes_from_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Read_L1B_RBC_Screening")
    return

  end subroutine Read_L1B_RBC_Screening
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_L1B_Input_Screening(L1B_Input_Scr, N_Meas0, N_Range_Gates0, &
                                      error_flag)
    !  #[

    TYPE(L1B_Input_Screening_Type), intent(out) :: L1B_Input_Scr ! output
    integer(IntAs), intent(in)  :: N_Meas0        ! input
    integer(IntAs), intent(in)  :: N_Range_Gates0 ! input
    integer,        intent(out) :: error_flag     ! output

    ! local variable
    integer :: m, AllocStatus

    error_flag = no_error

    L1B_Input_Scr%Obs_Screening        = L1B_Obs_Scr_undefined
    L1B_Input_Scr%Obs_Screening_Flags1 = 0_IntAuc
    L1B_Input_Scr%Obs_Screening_Flags2 = 0_IntAuc
    L1B_Input_Scr%Obs_Screening_Flags3 = 0_IntAuc
    L1B_Input_Scr%Obs_Screening_Flags4 = 0_IntAuc
    L1B_Input_Scr%Obs_Screening_Flags5 = 0_IntAuc

    IF (N_Meas0 .eq. 0) THEN
       ! init them to be empty
       nullify(L1B_Input_Scr%Mie_Meas)
       nullify(L1B_Input_Scr%Rayl_Meas)
    ELSE
       allocate(L1B_Input_Scr%Mie_Meas(N_Meas0),&
                L1B_Input_Scr%Rayl_Meas(N_Meas0),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          call logmsg(log_error,"Allocation failure in subroutine "//&
                                "Init_L1B_Input_Screening()")
          error_flag = error_allocate
          return
       END IF
       
       DO m=1,N_Meas0
          call Init_Mie_Meas_Screening(L1B_Input_Scr%Mie_Meas(m),&
                                       N_Range_Gates0,error_flag)
          call Init_Rayl_Meas_Screening(L1B_Input_Scr%Rayl_Meas(m),&
                                        N_Range_Gates0,error_flag)
       END DO
    END IF
    
    call Init_L1B_MRC_Screening(L1B_Input_Scr%MRC_Screening)
    call Init_L1B_RRC_Screening(L1B_Input_Scr%RRC_Screening)
    call Init_L1B_ZWC_Screening(L1B_Input_Scr%ZWC_Screening)
    call Init_L1B_RBC_Screening(L1B_Input_Scr%RBC_Screening)

  end subroutine Init_L1B_Input_Screening
    !  #]
  subroutine Delete_L1B_Input_Screening(L1B_Input_Scr)
    !  #[
    TYPE(L1B_Input_Screening_Type), intent(inout) :: L1B_Input_Scr

    ! local variable
    integer :: m

    IF (associated(L1B_Input_Scr%Mie_Meas)) THEN
       DO m=1,size(L1B_Input_Scr%Mie_Meas)
          call Delete_Mie_Meas_Screening(L1B_Input_Scr%Mie_Meas(m))
       END DO
       deallocate(L1B_Input_Scr%Mie_Meas)
    END IF
    IF (associated(L1B_Input_Scr%Rayl_Meas)) THEN
       DO m=1,size(L1B_Input_Scr%Rayl_Meas)
         call Delete_Rayl_Meas_Screening(L1B_Input_Scr%Rayl_Meas(m))
       END DO
       deallocate(L1B_Input_Scr%Rayl_Meas)
    END IF

  end subroutine Delete_L1B_Input_Screening
    !  #]
  subroutine Print_L1B_Input_Screening(L1B_Input_Scr)
    !  #[
    TYPE(L1B_Input_Screening_Type), intent(in) :: L1B_Input_Scr

    ! local variable
    integer :: m, s

    print *,"   L1B_Input_Scr%Obs_Screening        = ",&
         L1B_Input_Scr%Obs_Screening
    print *,"   L1B_Input_Scr%Obs_Screening_Flags1 = ",&
         L1B_Input_Scr%Obs_Screening_Flags1
    print *,"   L1B_Input_Scr%Obs_Screening_Flags2 = ",&
         L1B_Input_Scr%Obs_Screening_Flags2
    print *,"   L1B_Input_Scr%Obs_Screening_Flags3 = ",&
         L1B_Input_Scr%Obs_Screening_Flags3
    print *,"   L1B_Input_Scr%Obs_Screening_Flags4 = ",&
         L1B_Input_Scr%Obs_Screening_Flags4
    print *,"   L1B_Input_Scr%Obs_Screening_Flags5 = ",&
         L1B_Input_Scr%Obs_Screening_Flags5

    IF (associated(L1B_Input_Scr%Mie_Meas)) THEN
       ! get the size of the array
       s = size(L1B_Input_Scr%Mie_Meas)
       print "(a,i2,a)","    ==> size(L1B_Input_Scr%Mie_Meas) = ",s,&
                        " [printing first and last one]"
       DO m=1,s,s-1
          print "('    L1B_Input_Scr%Mie_Meas(',i2.2,')%...')",m
          call Print_Mie_Meas_Screening(L1B_Input_Scr%Mie_Meas(m))
       END DO
    ELSE
       print *,"    L1B_Input_Scr%Mie_Meas(:) [seems not associated]"
    END IF
    IF (associated(L1B_Input_Scr%Rayl_Meas)) THEN
       ! get the size of the array
       s = size(L1B_Input_Scr%Rayl_Meas)
       print "(a,i2,a)","    ==> size(L1B_Input_Scr%Rayl_Meas) = ",s,&
                        " [printing first and last one]"
       DO m=1,s,s-1
          print "('    L1B_Input_Scr%Rayl_Meas(',i2.2,')%...')",m
          call Print_Rayl_Meas_Screening(L1B_Input_Scr%Rayl_Meas(m))
       END DO
    ELSE
       print *,"    L1B_Input_Scr%Rayl_Meas(:) [seems not associated]"
    END IF

    print *,"   L1B_Input_Scr%MRC_Screening%..."
    call Print_L1B_MRC_Screening(L1B_Input_Scr%MRC_Screening)
    print *,"   L1B_Input_Scr%RRC_Screening%..."
    call Print_L1B_RRC_Screening(L1B_Input_Scr%RRC_Screening)
    print *,"   L1B_Input_Scr%ZWC_Screening%..."
    call Print_L1B_ZWC_Screening(L1B_Input_Scr%ZWC_Screening)
    print *,"   L1B_Input_Scr%RBC_Screening%..."
    call Print_L1B_RBC_Screening(L1B_Input_Scr%RBC_Screening)

  end subroutine Print_L1B_Input_Screening
    !  #]
  subroutine AsciiWrite_L1B_Input_Screening(L1B_Input_Scr,fileunit,&
                                            prefix,PrintAll,error_flag)
    !  #[
    TYPE(L1B_Input_Screening_Type), intent(in)  :: L1B_Input_Scr ! input
    integer,                        intent(in)  :: fileunit      ! input
    character(len=*),               intent(in)  :: prefix        ! input
    logical,                        intent(in)  :: PrintAll      ! input
    integer,                        intent(out) :: error_flag    ! output

    ! local variable
    integer :: m, s, step
    character(len=maxlen_IntAuc)      :: str_IntAuc
    character(len=maxlen_IntAucFlags) :: str_IntAucFlags
    character(len=2)                  :: str_cntr

    error_flag = no_error
   
    IF (L1B_Input_Scr%Obs_Screening .eq. L1B_Obs_Scr_undefined) THEN
       str_IntAuc = "[mi]"
    ELSE
       call SetIntAuc(L1B_Input_Scr%Obs_Screening,str_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
    END IF
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Obs_Screening = "//trim(str_IntAuc)

    call SetIntAucFlags(L1B_Input_Scr%Obs_Screening_Flags1,str_IntAucFlags,&
                        error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Obs_Screening_Flags1 = "//trim(str_IntAucFlags)

    call SetIntAucFlags(L1B_Input_Scr%Obs_Screening_Flags2,str_IntAucFlags,&
                        error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Obs_Screening_Flags2 = "//trim(str_IntAucFlags)

    call SetIntAucFlags(L1B_Input_Scr%Obs_Screening_Flags3,str_IntAucFlags,&
                        error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Obs_Screening_Flags3 = "//trim(str_IntAucFlags)

    call SetIntAucFlags(L1B_Input_Scr%Obs_Screening_Flags3,str_IntAucFlags,&
                        error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Obs_Screening_Flags3 = "//trim(str_IntAucFlags)

    call SetIntAucFlags(L1B_Input_Scr%Obs_Screening_Flags4,str_IntAucFlags,&
                        error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Obs_Screening_Flags4 = "//trim(str_IntAucFlags)

    call SetIntAucFlags(L1B_Input_Scr%Obs_Screening_Flags5,str_IntAucFlags,&
                        error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Obs_Screening_Flags5 = "//trim(str_IntAucFlags)

    IF (associated(L1B_Input_Scr%Mie_Meas)) THEN
       ! get the size of the array
       s = size(L1B_Input_Scr%Mie_Meas)
       IF (PrintAll .or. (s .le. 1) ) THEN
          step = 1
       ELSE
          step = s-1
       END IF
       DO m=1,s,step
          write(str_cntr,"(i2.2)",err=998) m
          call AsciiWrite_Mie_Meas_Screening(L1B_Input_Scr%Mie_Meas(m),&
               fileunit,trim(prefix)//"%Mie_Meas("//trim(str_cntr)//")",&
               PrintAll,error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    ELSE
       write(fileunit,"(a)",err=999) &
            trim(prefix)//"%Mie_Meas(:) [seems not associated]"
    END IF

    IF (associated(L1B_Input_Scr%Rayl_Meas)) THEN
       ! get the size of the array
       s = size(L1B_Input_Scr%Rayl_Meas)
       IF (PrintAll .or. (s .le. 1) ) THEN
          step = 1
       ELSE
          step = s-1
       END IF
       DO m=1,s,step
          write(str_cntr,"(i2.2)",err=998) m
          call AsciiWrite_Rayl_Meas_Screening(L1B_Input_Scr%Rayl_Meas(m),&
               fileunit,trim(prefix)//"%Rayl_Meas("//trim(str_cntr)//")",&
               PrintAll,error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    ELSE
       write(fileunit,"(a)",err=999) &
            trim(prefix)//"%Rayl_Meas(:) [seems not associated]"
    END IF

    call AsciiWrite_L1B_MRC_Screening(L1B_Input_Scr%MRC_Screening,&
         fileunit,trim(prefix)//"%MRC_Screening",error_flag)
    IF (error_flag .ne. no_error) return
    
    call AsciiWrite_L1B_RRC_Screening(L1B_Input_Scr%RRC_Screening,&
         fileunit,trim(prefix)//"%RRC_Screening",error_flag)
    IF (error_flag .ne. no_error) return
    
    call AsciiWrite_L1B_ZWC_Screening(L1B_Input_Scr%ZWC_Screening,&
         fileunit,trim(prefix)//"%ZWC_Screening",error_flag)
    IF (error_flag .ne. no_error) return

    call AsciiWrite_L1B_RBC_Screening(L1B_Input_Scr%RBC_Screening,&
         fileunit,trim(prefix)//"%RBC_Screening",error_flag)
    IF (error_flag .ne. no_error) return

    return

998 call logmsg(log_error,"conversion error in "//&
         "AsciiWrite_L1B_Input_Screening: could not convert the "//&
         "counter m to a string ..")
    error_flag = error_writing_file
    return
    
999 call logmsg(log_error,"ERROR in AsciiWrite_L1B_Input_Screening: "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L1B_Input_Screening
    !  #]
  function GetByteSize_L1B_Input_Screening(L1B_Input_Scr, N_Meas_to_write) &
       result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result
    TYPE(L1B_Input_Screening_Type), intent(in) :: L1B_Input_Scr
    integer(IntAuc),                intent(in) :: N_Meas_to_write

    ! local variable
    integer :: s, error_flag
    character(len=maxlen_IntAuc) :: str1_IntAuc, str2_IntAuc

    c_bytesize = int(6*c_bytelen_IntAuc,IntAl) + &
                 GetByteSize_L1B_MRC_Screening() + &
                 GetByteSize_L1B_RRC_Screening() + &
                 GetByteSize_L1B_ZWC_Screening() + &
                 GetByteSize_L1B_RBC_Screening() + 20_IntAl ! spare_bytes

    IF (associated(L1B_Input_Scr%Mie_Meas)) THEN
       s = size(L1B_Input_Scr%Mie_Meas)
       IF (N_Meas_to_write .gt. s) THEN
          call SetIntAuc(N_Meas_to_write,str1_IntAuc,error_flag)
          IF (error_flag .ne. no_error) str1_IntAuc = "[er]"
          call SetIntAuc(int(s,IntAuc), str2_IntAuc,error_flag)
          IF (error_flag .ne. no_error) str2_IntAuc = "[er]"
          call logmsg(log_warn,"WARNING in "//&
               "GetByteSize_L1B_Input_Screening(): invalid value for "//&
               "N_Meas_to_write: "//trim(str1_IntAuc)//&
               " max. possible value is: size(L1B_Input_Scr%Mie_Meas) = "//&
               trim(str2_IntAuc)//" GetByteSize result will be useless....")
       END IF

       c_bytesize = c_bytesize + &
            N_Meas_to_write*&
            GetByteSize_Mie_Meas_Screening(L1B_Input_Scr%Mie_Meas(1))
    END IF
    IF (associated(L1B_Input_Scr%Rayl_Meas)) THEN
       s = size(L1B_Input_Scr%Rayl_Meas)
       IF (N_Meas_to_write .gt. s) THEN
          call SetIntAuc(N_Meas_to_write,str1_IntAuc,error_flag)
          IF (error_flag .ne. no_error) str1_IntAuc = "[er]"
          call SetIntAuc(int(s,IntAuc), str2_IntAuc,error_flag)
          IF (error_flag .ne. no_error) str2_IntAuc = "[er]"
          call logmsg(log_warn,"WARNING in "//&
               "GetByteSize_L1B_Input_Screening(): invalid value for "//&
               "N_Meas_to_write: "//trim(str1_IntAuc)//&
               " max. possible value is: size(L1B_Input_Scr%Rayl_Meas) = "//&
               trim(str2_IntAuc)//" GetByteSize result will be useless....")
       END IF

       c_bytesize = c_bytesize + &
            N_Meas_to_write*&
            GetByteSize_Rayl_Meas_Screening(L1B_Input_Scr%Rayl_Meas(1))
    END IF

    ! typical sizes
    ! memory size = 6*2+2+2+2+2   +67*100+67*100 = 13420 bytes
    ! file   size = 6*1+6+6+6+6+20+67* 51+67* 51 = 6884 bytes
  end function GetByteSize_L1B_Input_Screening
    !  #]
  subroutine Write_L1B_Inp_Screening(fileunit, &
             L1B_Input_Scr, N_Meas_to_write, error_flag)
    !  #[
    !  Writes the L1B_Input_Screening in binary format
    !  J. de Kloe  03-01-2006 Original Code
    !  P. Poli     10-01-2006 Removed placeholders. Put actual writing in place
    !                         From L2B QC tables v0.2 03-01-2005
    
    integer,                        intent(in)  :: fileunit        ! input
    TYPE(L1B_Input_Screening_Type), intent(in)  :: L1B_Input_Scr   ! input
    integer(IntAuc),                intent(in)  :: N_Meas_to_write ! input
    integer,                        intent(out) :: error_flag      ! output

    ! local variables
    integer :: nr_spare_bytes
    integer :: i,s
    character(len=maxlen_IntAuc) :: str1_IntAuc, str2_IntAuc

    !-----------------------------------
    ! writing the L1B_Input_Scr elements
    !-----------------------------------

    error_flag = no_error

    call write_IntAuc_to_bin_file(fileunit, &
         L1B_Input_Scr%Obs_Screening,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_IntAuc_to_bin_file(fileunit, &
         L1B_Input_Scr%Obs_Screening_Flags1, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_IntAuc_to_bin_file(fileunit, &
         L1B_Input_Scr%Obs_Screening_Flags2, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_IntAuc_to_bin_file(fileunit, &
         L1B_Input_Scr%Obs_Screening_Flags3, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_IntAuc_to_bin_file(fileunit, &
         L1B_Input_Scr%Obs_Screening_Flags4, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_IntAuc_to_bin_file(fileunit, &
         L1B_Input_Scr%Obs_Screening_Flags5, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    IF (.not. associated(L1B_Input_Scr%Mie_Meas)) THEN
       call logmsg(log_error,"ERROR in Write_L1B_Inp_Screening(): "//&
            "The pointer to L1B_Input_Scr%Mie_Meas seems not associated !!")
       error_flag = error_programming
       return
    END IF

    s = size(L1B_Input_Scr%Mie_Meas)
    IF (N_Meas_to_write .gt. s) THEN
       call SetIntAuc(N_Meas_to_write,str1_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
       call SetIntAuc(int(s,IntAuc), str2_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_error,"ERROR in "//&
            "Write_L1B_Inp_Screening(): invalid value for "//&
            "N_Meas_to_write: "//trim(str1_IntAuc)//&
            " max. possible value is: size(L1B_Input_Scr%Mie_Meas) = "//&
            trim(str2_IntAuc))
       error_flag = error_programming
       return
    END IF
    
    DO i=1,N_Meas_to_write
       call Write_L1B_Mie_Meas_Screening( &
            L1B_Input_Scr%Mie_Meas(i), fileunit, error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO
    
    IF (.not. associated(L1B_Input_Scr%Rayl_Meas)) THEN
       call logmsg(log_error,"ERROR in Write_L1B_Inp_Screening(): "//&
            "The pointer to L1B_Input_Scr%Rayl_Meas seems not associated !!")
       error_flag = error_programming
       return
    END IF
    
    s = size(L1B_Input_Scr%Rayl_Meas)
    IF (N_Meas_to_write .gt. s) THEN
       call SetIntAuc(N_Meas_to_write,str1_IntAuc,error_flag)
       IF (error_flag .ne. no_error) str1_IntAuc = "[er]"
       call SetIntAuc(int(s,IntAuc), str2_IntAuc,error_flag)
       IF (error_flag .ne. no_error) str2_IntAuc = "[er]"
       call logmsg(log_error,"ERROR in "//&
            "Write_L1B_Inp_Screening(): invalid value for "//&
            "N_Meas_to_write: "//trim(str1_IntAuc)//&
            " max. possible value is: size(L1B_Input_Scr%Rayl_Meas) = "//&
            trim(str2_IntAuc))
       error_flag = error_programming
       return
    END IF

    DO i=1,N_Meas_to_write
       call Write_L1B_Rayl_Meas_Screening( &
            L1B_Input_Scr%Rayl_Meas(i), fileunit, error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    call Write_L1B_MRC_Screening( &
         L1B_Input_Scr%MRC_Screening, fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call Write_L1B_RRC_Screening( &
         L1B_Input_Scr%RRC_Screening, fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call Write_L1B_ZWC_Screening( &
         L1B_Input_Scr%ZWC_Screening, fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call Write_L1B_RBC_Screening( &
         L1B_Input_Scr%RBC_Screening, fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! 20 spare bytes
    nr_spare_bytes = 20
    call write_SpareBytes_to_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    return

998 call logmsg(log_error,"ABORTING subroutine Write_L1B_Inp_Screening")
    return

  end subroutine Write_L1B_Inp_Screening
    !  #]
  subroutine Read_L1B_Inp_Screening(fileunit, &
             L1B_Input_Scr, error_flag)
    !  #[
    !  P. Poli     10-01-2006 Original code
    integer,                        intent(in)   :: fileunit
    TYPE(L1B_Input_Screening_Type), intent(inout):: L1B_Input_Scr
    integer,                        intent(out)  :: error_flag

    ! local variable
    integer :: i,s,nr_spare_bytes

    !-----------------------------------
    ! reading the L1B_Input_Scr elements
    !-----------------------------------

    error_flag = no_error

    call read_IntAuc_from_bin_file(fileunit, &
         L1B_Input_Scr%Obs_Screening,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call read_IntAuc_from_bin_file(fileunit, &
         L1B_Input_Scr%Obs_Screening_Flags1, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call read_IntAuc_from_bin_file(fileunit, &
         L1B_Input_Scr%Obs_Screening_Flags2, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call read_IntAuc_from_bin_file(fileunit, &
         L1B_Input_Scr%Obs_Screening_Flags3, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call read_IntAuc_from_bin_file(fileunit, &
         L1B_Input_Scr%Obs_Screening_Flags4, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call read_IntAuc_from_bin_file(fileunit, &
         L1B_Input_Scr%Obs_Screening_Flags5, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    IF (.not. associated(L1B_Input_Scr%Mie_Meas)) THEN
       call logmsg(log_error,"ERROR in Read_L1B_Inp_Screening(): "//&
            "The pointer to L1B_Input_Scr%Mie_Meas seems not associated !!")
       error_flag = error_programming
       return
    END IF

    s=size(L1B_Input_Scr%Mie_Meas)
    DO i=1,s
       call Read_L1B_Mie_Meas_Screening(fileunit, &
            L1B_Input_Scr%Mie_Meas(i), error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    IF (.not. associated(L1B_Input_Scr%Rayl_Meas)) THEN
       call logmsg(log_error,"ERROR in Read_L1B_Inp_Screening(): "//&
            "The pointer to L1B_Input_Scr%Rayl_Meas seems not associated !!")
       error_flag = error_programming
       return
    END IF
    
    s=size(L1B_Input_Scr%Rayl_Meas)
    DO i=1,s
       call Read_L1B_Rayl_Meas_Screening(fileunit, &
            L1B_Input_Scr%Rayl_Meas(i), error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    call Read_L1B_MRC_Screening(fileunit, &
         L1B_Input_Scr%MRC_Screening, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call Read_L1B_RRC_Screening(fileunit, &
         L1B_Input_Scr%RRC_Screening, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call Read_L1B_ZWC_Screening(fileunit, &
         L1B_Input_Scr%ZWC_Screening, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call Read_L1B_RBC_Screening(fileunit, &
         L1B_Input_Scr%RBC_Screening, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! 20 spare bytes
    nr_spare_bytes = 20
    call read_SpareBytes_from_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Read_L1B_Inp_Screening")
    return

  end subroutine Read_L1B_Inp_Screening
    !  #]

  !  #]
  !---------------------------------------------------------------------
  ! subroutines for the L2B_AMD_Screening sub-datastructure    
  !---------------------------------------------------------------------
  !  #[
  subroutine Init_L2B_AMD_Collocation(L2B_AMD_Collocations)
    !  #[
    TYPE(L2B_AMD_Collocations_Type), intent(out) :: L2B_AMD_Collocations

    L2B_AMD_Collocations%L2B_AMD_Collocation_QC       = L2B_AMD_Coll_Undefined
    L2B_AMD_Collocations%L2B_AMD_Collocation_QC_Flags = 0

  end subroutine Init_L2B_AMD_Collocation
    !  #]
  subroutine Print_L2B_AMD_Collocations(L2B_AMD_Collocations)
    !  #[
    TYPE(L2B_AMD_Collocations_Type), intent(in) :: L2B_AMD_Collocations

    print *,"       ...%L2B_AMD_Collocation_QC       = ",&    
         L2B_AMD_Collocations%L2B_AMD_Collocation_QC
    print *,"       ...%L2B_AMD_Collocation_QC_Flags = ",&    
         L2B_AMD_Collocations%L2B_AMD_Collocation_QC_Flags

  end subroutine Print_L2B_AMD_Collocations
    !  #]
  subroutine AsciiWrite_L2B_AMD_Collocations(L2B_AMD_Collocations,fileunit,&
                                             prefix,error_flag)
    !  #[
    TYPE(L2B_AMD_Collocations_Type), intent(in)  :: L2B_AMD_Collocations
    integer,                         intent(in)  :: fileunit   ! input
    character(len=*),                intent(in)  :: prefix     ! input
    integer,                         intent(out) :: error_flag ! output

    ! local variables
    character(len=maxlen_IntAuc)      :: str_IntAuc
    character(len=maxlen_IntAucFlags) :: str_IntAucFlags

    error_flag = no_error

    if ( L2B_AMD_Collocations%L2B_AMD_Collocation_QC .eq. &
         L2B_AMD_Coll_Undefined                           ) then
       str_IntAuc = "[mi]"
    else
       call SetIntAuc(L2B_AMD_Collocations%L2B_AMD_Collocation_QC,&
            str_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
    end if
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L2B_AMD_Collocation_QC = "//trim(str_IntAuc)

    call SetIntAucFlags(L2B_AMD_Collocations%L2B_AMD_Collocation_QC_Flags,&
         str_IntAucFlags,error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L2B_AMD_Collocation_QC_Flags = "//&
         trim(str_IntAucFlags)
    
    return
    
999 call logmsg(log_error,"ERROR in AsciiWrite_L2B_AMD_Collocations: "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2B_AMD_Collocations
    !  #]
  function GetByteSize_L2B_AMD_Colloc() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result

    c_bytesize = 2*c_bytelen_IntAuc + 1 ! spare_byte
    ! 2 items in total

    ! typical sizes
    ! memory size = 2*2   = 4 bytes
    ! file   size = 2*1+1 = 3 bytes
  end function GetByteSize_L2B_AMD_Colloc
    !  #]
  subroutine Write_L2B_AMD_Collocations(fileunit,&
       L2B_AMD_Collocations, error_flag)
    !  #[
    !  Writes L2B_AMD_Collocations in binary format
    !  P. Poli     09-01-2006 Original Code
    integer,                         intent(in)  :: fileunit
    TYPE(L2B_AMD_Collocations_Type), intent(in)  :: L2B_AMD_Collocations
    integer,                         intent(out) :: error_flag 

    ! local variable
    integer :: nr_spare_bytes

    !------------------------------------------
    ! writing the L2B_AMD_Collocations elements
    !------------------------------------------

    error_flag = no_error

    call write_IntAuc_to_bin_file(fileunit, &
         L2B_AMD_Collocations%L2B_AMD_Collocation_QC, error_flag)
    IF (error_flag .ne. no_error) goto 998

    call write_IntAuc_to_bin_file(fileunit, &
         L2B_AMD_Collocations%L2B_AMD_Collocation_QC_Flags, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 1 spare byte
    nr_spare_bytes = 1
    call write_SpareBytes_to_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Write_L2B_AMD_Collocations")
    return

  end subroutine Write_L2B_AMD_Collocations
    !  #]
  subroutine Read_L2B_AMD_Collocations(fileunit,&
       L2B_AMD_Collocations,error_flag)
    !  #[
    !  P. Poli     10-01-2006 Original Code
    integer,                         intent(in)    :: fileunit
    TYPE(L2B_AMD_Collocations_Type), intent(inout) :: L2B_AMD_Collocations
    integer,                         intent(out)   :: error_flag 

    ! local variable
    integer :: nr_spare_bytes

    !------------------------------------------
    ! reading the L2B_AMD_Collocations elements
    !------------------------------------------

    error_flag = no_error

    call read_IntAuc_from_bin_file(fileunit, &
         L2B_AMD_Collocations%L2B_AMD_Collocation_QC, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call read_IntAuc_from_bin_file(fileunit, &
         L2B_AMD_Collocations%L2B_AMD_Collocation_QC_Flags, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! 1 spare byte
    nr_spare_bytes = 1
    call read_SpareBytes_from_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Read_L2B_AMD_Collocations")
    return

  end subroutine Read_L2B_AMD_Collocations
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_L2B_AMD_Screening(L2B_AMD_Screening, N_Meas0, error_flag)
    !  #[
    TYPE(L2B_AMD_Screening_Type), intent(out) :: L2B_AMD_Screening
    integer(IntAs), intent(in)  :: N_Meas0        ! input
    integer,        intent(out) :: error_flag    ! output

    ! local variable
    integer :: m, AllocStatus

    error_flag = no_error

    L2B_AMD_Screening%L2B_AMD_Screening_QC       = L2B_AMD_Undefined
    L2B_AMD_Screening%L2B_AMD_Screening_QC_Flags = 0

    IF (N_Meas0 .eq. 0) THEN
       ! init them to be empty
       nullify(L2B_AMD_Screening%L2B_AMD_Collocations)
    ELSE
       allocate(L2B_AMD_Screening%L2B_AMD_Collocations(N_Meas0),&
                STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          call logmsg(log_error,"Allocation failure in subroutine "//&
                                "Init_L2B_AMD_Screening()")
          error_flag = error_allocate
          return
       END IF
       
       DO m=1,N_Meas0
          call Init_L2B_AMD_Collocation(&
                  L2B_AMD_Screening%L2B_AMD_Collocations(m))
       END DO
    END IF

  end subroutine Init_L2B_AMD_Screening
    !  #]
  subroutine Delete_L2B_AMD_Screening(L2B_AMD_Screening)
    !  #[
    TYPE(L2B_AMD_Screening_Type), intent(inout) :: L2B_AMD_Screening

    IF (associated(L2B_AMD_Screening%L2B_AMD_Collocations)) &
         deallocate(L2B_AMD_Screening%L2B_AMD_Collocations)

  end subroutine Delete_L2B_AMD_Screening
    !  #]
  subroutine Print_L2B_AMD_Screening(L2B_AMD_Screening)
    !  #[
    TYPE(L2B_AMD_Screening_Type), intent(in) :: L2B_AMD_Screening

    ! local variable
    integer :: m, s

    print *,"   ...%L2B_AMD_Screening_QC       = ",&
         L2B_AMD_Screening%L2B_AMD_Screening_QC
    print *,"   ...%L2B_AMD_Screening_QC_Flags = ",&
         L2B_AMD_Screening%L2B_AMD_Screening_QC_Flags

    print *,"   L2B_AMD_Screening%L2B_AMD_Collocations%..."
    IF (associated(L2B_AMD_Screening%L2B_AMD_Collocations)) THEN
       ! get the size of the array
       s = size(L2B_AMD_Screening%L2B_AMD_Collocations)
       print "(a,i2,a)",&
            "    ==> size(L2B_AMD_Screening%L2B_AMD_Collocations) = ",s
       print *,"       [printing first and last one]"
       DO m=1,s,s-1
          print "('    L2B_AMD_Screening%L2B_AMD_Collocations(',"//&
               "i2.2,')%...')",m
          call Print_L2B_AMD_Collocations(&
               L2B_AMD_Screening%L2B_AMD_Collocations(m))
       END DO
    ELSE
       print *,"    L2B_AMD_Screening%L2B_AMD_Collocations(:) "//&
            "[seems not associated]"
    END IF

  end subroutine Print_L2B_AMD_Screening
    !  #]
  subroutine AsciiWrite_L2B_AMD_Screening(L2B_AMD_Screening,fileunit,&
                                          prefix,PrintAll,error_flag)
    !  #[
    TYPE(L2B_AMD_Screening_Type), intent(in)  :: L2B_AMD_Screening
    integer,                      intent(in)  :: fileunit   ! input
    character(len=*),             intent(in)  :: prefix     ! input
    logical,                      intent(in)  :: PrintAll   ! input
    integer,                      intent(out) :: error_flag ! output

    ! local variable
    integer :: m, s, step
    character(len=maxlen_IntAuc)      :: str_IntAuc
    character(len=maxlen_IntAucFlags) :: str_IntAucFlags
    character(len=2)                  :: str_cntr

    error_flag = no_error

    if ( L2B_AMD_Screening%L2B_AMD_Screening_QC .eq. L2B_AMD_Undefined) then
       str_IntAuc = "[mi]"
    else
       call SetIntAuc(L2B_AMD_Screening%L2B_AMD_Screening_QC,&
                      str_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
    end if
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L2B_AMD_Screening_QC = "//trim(str_IntAuc)

    call SetIntAucFlags(L2B_AMD_Screening%L2B_AMD_Screening_QC_Flags,&
                        str_IntAucFlags,error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L2B_AMD_Screening_QC_Flags = "//trim(str_IntAucFlags)

    IF (associated(L2B_AMD_Screening%L2B_AMD_Collocations)) THEN
       ! get the size of the array
       s = size(L2B_AMD_Screening%L2B_AMD_Collocations)
       IF (PrintAll .or. (s .le. 1) ) THEN
          step=1
       ELSE
          step = s-1
       END IF
       DO m=1,s,step
          write(str_cntr,"(i2.2)",err=998) m
          call AsciiWrite_L2B_AMD_Collocations(&
               L2B_AMD_Screening%L2B_AMD_Collocations(m),fileunit,&
               trim(prefix)//"%L2B_AMD_Collocations("//trim(str_cntr)//")",&
               error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    ELSE
       write(fileunit,"(a)",err=999) &
            trim(prefix)//"%L2B_AMD_Collocations(:) [seems not associated]"
    END IF

    return

998 call logmsg(log_error,"conversion error in "//&
         "AsciiWrite_L2B_AMD_Screening: could not convert the counter "//&
         "m to a string ..")
    error_flag = error_writing_file
    return
    
999 call logmsg(log_error,"ERROR in AsciiWrite_L2B_AMD_Screening: "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2B_AMD_Screening
    !  #]
  function GetByteSize_L2B_AMD_Screening(L2B_AMD_Screening,N_Meas_to_write) &
       result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result
    TYPE(L2B_AMD_Screening_Type), intent(in) :: L2B_AMD_Screening
    integer(IntAuc),              intent(in) :: N_Meas_to_write

    ! local variables
    integer :: s, error_flag
    character(len=maxlen_IntAuc) :: str1_IntAuc, str2_IntAuc

    c_bytesize = 2*c_bytelen_IntAuc + 20 ! spare_bytes
    IF (associated(L2B_AMD_Screening%L2B_AMD_Collocations)) THEN
       s = size(L2B_AMD_Screening%L2B_AMD_Collocations)
       IF (N_Meas_to_write .gt. s) THEN
          call SetIntAuc(N_Meas_to_write,str1_IntAuc,error_flag)
          IF (error_flag .ne. no_error) str1_IntAuc = "[er]"
          call SetIntAuc(int(s,IntAuc), str2_IntAuc,error_flag)
          IF (error_flag .ne. no_error) str2_IntAuc = "[er]"
          call logmsg(log_warn,"WARNING in "//&
               "GetByteSize_L2B_AMD_Input_Screening(): invalid value for "//&
               "N_Meas_to_write: "//trim(str1_IntAuc)//" max. possible "//& 
               "value is size(L2B_AMD_Screening%L2B_AMD_Collocations) = "//&
               trim(str2_IntAuc)//"GetByteSize result will be useless....")
       END IF
       c_bytesize = c_bytesize + &
            N_Meas_to_write*GetByteSize_L2B_AMD_Colloc()
    END IF

    ! typical sizes
    ! memory size = 2*2   +67*4 = 272 bytes
    ! file   size = 2*1+20+67*3 = 223 bytes
  end function GetByteSize_L2B_AMD_Screening
    !  #]
  subroutine Write_L2B_AMD_Screening(fileunit, &
             L2B_AMD_Screening, N_Meas_to_write, error_flag)
    !  #[

    !  Writes the L2B_AMD_Screening in binary format
    !  J. de Kloe  03-01-2006 Original Code
    !  P. Poli     10-01-2006 Removed placeholders. Put actual writing in place
    !                         From L2B QC tables v0.2 03-01-2005
    
    integer,                      intent(in)  :: fileunit          ! input
    TYPE(L2B_AMD_Screening_Type), intent(in)  :: L2B_AMD_Screening ! input
    integer(IntAuc),              intent(in)  :: N_Meas_to_write   ! input
    integer,                      intent(out) :: error_flag        ! output

    ! local variables
    integer :: i,s,nr_spare_bytes
    character(len=maxlen_IntAuc) :: str1_IntAuc, str2_IntAuc

    !---------------------------------------
    ! writing the L2B_AMD_Screening elements
    !---------------------------------------

    error_flag = no_error
    
    call write_IntAuc_to_bin_file(fileunit, &
         L2B_AMD_Screening%L2B_AMD_Screening_QC,error_flag)
    IF (error_flag .ne. no_error) goto 998
       
    call write_IntAuc_to_bin_file(fileunit, &
         L2B_AMD_Screening%L2B_AMD_Screening_QC_Flags, error_flag)
    IF (error_flag .ne. no_error) goto 998
       
    IF (.not. associated(L2B_AMD_Screening%L2B_AMD_Collocations)) THEN
       call logmsg(log_error,"ERROR in Write_L2B_AMD_Screening(): "//&
            "The pointer to L2B_AMD_Screening%L2B_AMD_Collocations "//&
            "seems not associated !!")
       error_flag = error_programming
       return
    END IF

    s = size(L2B_AMD_Screening%L2B_AMD_Collocations)
    IF (N_Meas_to_write .gt. s) THEN
       call SetIntAuc(N_Meas_to_write,str1_IntAuc,error_flag)
       IF (error_flag .ne. no_error) str1_IntAuc = "[er]"
       call SetIntAuc(int(s,IntAuc), str2_IntAuc,error_flag)
       IF (error_flag .ne. no_error) str2_IntAuc = "[er]"
       call logmsg(log_error,"ERROR in "//&
            "Write_L2B_AMD_Screening(): invalid value for "//&
            "N_Meas_to_write: "//trim(str1_IntAuc)//&
            " max. possible value is: "//&
            "size(L2B_AMD_Screening%L2B_AMD_Collocations) = "//&
            trim(str2_IntAuc))
       error_flag = error_programming
       return
    END IF

    DO i=1,N_Meas_to_write
       call Write_L2B_AMD_Collocations(fileunit, &
            L2B_AMD_Screening%L2B_AMD_Collocations(i), error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    ! 20 spare bytes
    nr_spare_bytes = 20
    call write_SpareBytes_to_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Write_L2B_AMD_Screening")
    return

  end subroutine Write_L2B_AMD_Screening

    !  #]
  subroutine Read_L2B_AMD_Screening(fileunit, L2B_AMD_Screening, error_flag)
    !  #[
    !  P. Poli     10-01-2006 Original code
    integer,                      intent(in)    :: fileunit
    TYPE(L2B_AMD_Screening_Type), intent(inout) :: L2B_AMD_Screening
    integer,                      intent(out)   :: error_flag

    ! local variable
    integer :: i,s,nr_spare_bytes

    !---------------------------------------
    ! reading the L2B_AMD_Screening elements
    !---------------------------------------

    error_flag = no_error

    call read_IntAuc_from_bin_file(fileunit, &
         L2B_AMD_Screening%L2B_AMD_Screening_QC,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call read_IntAuc_from_bin_file(fileunit, &
         L2B_AMD_Screening%L2B_AMD_Screening_QC_Flags, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    IF (.not. associated(L2B_AMD_Screening%L2B_AMD_Collocations)) THEN
       call logmsg(log_error,"ERROR in Read_L2B_AMD_Screening(): "//&
            "The pointer to L2B_AMD_Screening%L2B_AMD_Collocations "//&
            "seems not associated !!")
       error_flag = error_programming
       return
    END IF
    
    s = size(L2B_AMD_Screening%L2B_AMD_Collocations)
    DO i=1,s
       call Read_L2B_AMD_Collocations(fileunit, &
            L2B_AMD_Screening%L2B_AMD_Collocations(i), error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    ! 20 spare bytes
    nr_spare_bytes = 20
    call read_SpareBytes_from_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Read_L2B_AMD_Screening")
    return

  end subroutine Read_L2B_AMD_Screening
    !  #]
  !  #]
  !---------------------------------------------------------------------
  ! subroutines for the Mie_Classification_QC sub-datastructure    
  !---------------------------------------------------------------------
  !  #[
  subroutine Init_L2B_Mie_Bin_Classific(L2B_Mie_Bin_Classific)
    !  #[
    TYPE(L2B_Mie_Bin_Classific_Type), intent(out) :: L2B_Mie_Bin_Classific
    
    L2B_Mie_Bin_Classific%flag1 = 0_IntAuc
    L2B_Mie_Bin_Classific%flag2 = 0_IntAuc
    L2B_Mie_Bin_Classific%L2B_Reliability = missing_indicator_FAdoxy

  end subroutine Init_L2B_Mie_Bin_Classific
    !  #]
  subroutine Print_L2B_Mie_Bin_Classific(L2B_Mie_Bin_Classific)
    !  #[
    TYPE(L2B_Mie_Bin_Classific_Type), intent(in) :: L2B_Mie_Bin_Classific

    print *,"       ...%flag1 = ",L2B_Mie_Bin_Classific%flag1
    print *,"       ...%flag2 = ",L2B_Mie_Bin_Classific%flag2
    print *,"       ...%L2B_Reliability             = ",&
         L2B_Mie_Bin_Classific%L2B_Reliability

  end subroutine Print_L2B_Mie_Bin_Classific
    !  #]
  subroutine AsciiWrite_L2B_Mie_Bin_Classif(L2B_Mie_Bin_Classific,&
                                            fileunit,prefix,error_flag)
    !  #[
    TYPE(L2B_Mie_Bin_Classific_Type), intent(in)  :: L2B_Mie_Bin_Classific
    integer,                          intent(in)  :: fileunit   ! input
    character(len=*),                 intent(in)  :: prefix     ! input
    integer,                          intent(out) :: error_flag ! output

    ! local variables
    character(len=maxlen_IntAucFlags) :: str_IntAucFlags
    character(len=25)                 :: str_FAdoxy

    call SetIntAucFlags(L2B_Mie_Bin_Classific%flag1,str_IntAucFlags,error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%flag1 = "//trim(str_IntAucFlags)

    call SetIntAucFlags(L2B_Mie_Bin_Classific%flag2,str_IntAucFlags,error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%flag2 = "//trim(str_IntAucFlags)

    write(fileunit,"(a)",err=999) "REALACC(12)"

    call SetFAdoFull_or_Missing(L2B_Mie_Bin_Classific%L2B_Reliability,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L2B_Reliability = "//trim(str_FAdoxy)

    write(fileunit,"(a)",err=999) "ENDREALACC"

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_L2B_Mie_Bin_Classif "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2B_Mie_Bin_Classif
    !  #]
  function GetByteSize_L2B_Mie_Bin_Class() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result

    c_bytesize = 2*c_bytelen_IntAuc + &
                 1*c_bytelen_FAdoxy + 1 ! spare_byte
    ! 3 items in total

    ! typical sizes
    ! memory size = 2*2+8   = 12 bytes
    ! file   size = 2*1+8+1 = 11 bytes
  end function GetByteSize_L2B_Mie_Bin_Class
    !  #]
  subroutine Write_L2B_Mie_Bin_Classific(fileunit, &
             L2B_Mie_Bin_Classific, error_flag)
    !  #[
    !  Writes L2B_Mie_Meas_Bin_Classification or
    !         L2B_Mie_Obs_Bin_Classification in binary format
    !  P. Poli     09-01-2006 Original Code
    integer,                          intent(in)  :: fileunit
    TYPE(L2B_Mie_Bin_Classific_Type), intent(in)  :: L2B_Mie_Bin_Classific
    integer,                          intent(out) :: error_flag 

    ! local variable
    integer :: nr_spare_bytes

    !-------------------------------------------
    ! writing the L2B_Mie_Bin_Classific elements
    !-------------------------------------------

    error_flag = no_error

    call write_IntAuc_to_bin_file(fileunit, &
         L2B_Mie_Bin_Classific%flag1, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_IntAuc_to_bin_file(fileunit, &
         L2B_Mie_Bin_Classific%flag2, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_FAdoxy_to_bin_file(fileunit, &
         L2B_Mie_Bin_Classific%L2B_Reliability, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! 1 spare byte
    nr_spare_bytes = 1
    call write_SpareBytes_to_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Write_L2B_Mie_Bin_Classific")
    return

  end subroutine Write_L2B_Mie_Bin_Classific
    !  #]
  subroutine Read_L2B_Mie_Bin_Classific(fileunit, &
             L2B_Mie_Bin_Classific, error_flag)
    !  #[
    !  P. Poli     10-01-2006 Original Code
    integer,                          intent(in)    :: fileunit
    TYPE(L2B_Mie_Bin_Classific_Type), intent(inout) :: L2B_Mie_Bin_Classific
    integer,                          intent(out)   :: error_flag 

    ! local variable
    integer :: nr_spare_bytes

    !-------------------------------------------
    ! reading the L2B_Mie_Bin_Classific elements
    !-------------------------------------------

    error_flag = no_error

    call read_IntAuc_from_bin_file(fileunit, &
         L2B_Mie_Bin_Classific%flag1, error_flag)
    IF (error_flag .ne. no_error) goto 998

    call read_IntAuc_from_bin_file(fileunit, &
         L2B_Mie_Bin_Classific%flag2, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call read_FAdoxy_from_bin_file(fileunit, &
         L2B_Mie_Bin_Classific%L2B_Reliability, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! 1 spare byte
    nr_spare_bytes = 1
    call read_SpareBytes_from_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Read_L2B_Mie_Bin_Classific")
    return

  end subroutine Read_L2B_Mie_Bin_Classific
    !  #]
  !---------------------------------------------------------------------
  subroutine  Init_L2B_Mie_Meas_Classific(L2B_Mie_Meas_Classific,&
                                          N_Range_Gates0, error_flag)
    !  #[
    TYPE(L2B_Mie_Meas_Classific_Type), intent(out) :: L2B_Mie_Meas_Classific
    integer(IntAs), intent(in)  :: N_Range_Gates0 ! input
    integer,        intent(out) :: error_flag    ! output

    ! local variable
    integer :: r, AllocStatus

    error_flag = no_error

    IF (N_Range_Gates0 .eq. 0) THEN
       ! init them to be empty
       nullify(L2B_Mie_Meas_Classific%L2B_Mie_Meas_Bin_Classific)
    ELSE
       allocate(L2B_Mie_Meas_Classific%&
                L2B_Mie_Meas_Bin_Classific(N_Range_Gates0),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          call logmsg(log_error,"Allocation failure in subroutine"//&
                                "Init_L2B_Mie_Meas_Classific()")
          error_flag = error_allocate
          return
       END IF
       DO r=1,N_Range_Gates0
          call Init_L2B_Mie_Bin_Classific(&
                  L2B_Mie_Meas_Classific%L2B_Mie_Meas_Bin_Classific(r))
       END DO
    END IF
    
  end subroutine Init_L2B_Mie_Meas_Classific
    !  #]
  subroutine  Delete_L2B_Mie_Meas_Classific(L2B_Mie_Meas_Classific)
    !  #[
    TYPE(L2B_Mie_Meas_Classific_Type), &
         intent(inout) :: L2B_Mie_Meas_Classific

    IF (associated(L2B_Mie_Meas_Classific%L2B_Mie_Meas_Bin_Classific)) &
         deallocate(L2B_Mie_Meas_Classific%L2B_Mie_Meas_Bin_Classific)
    
  end subroutine Delete_L2B_Mie_Meas_Classific
    !  #]
  subroutine Print_L2B_Mie_Meas_Classific(L2B_Mie_Meas_Classific)
    !  #[
    TYPE(L2B_Mie_Meas_Classific_Type), intent(in) :: L2B_Mie_Meas_Classific

    integer :: r,s

    IF (associated(L2B_Mie_Meas_Classific%L2B_Mie_Meas_Bin_Classific)) THEN
       ! get the size of the array
       s = size(L2B_Mie_Meas_Classific%L2B_Mie_Meas_Bin_Classific)
       print "(a,i2,a)","    ==> size(L2B_Mie_Meas_Classific%"//&
            "L2B_Mie_Meas_Bin_Classific) = ",s
       print *,"       [printing first and last one]"
       DO r=1,s,s-1
          print "('        L2B_Mie_Meas_Classific%"//&
               "L2B_Mie_Meas_Bin_Classific(',i2.2,')%...')",r
          call Print_L2B_Mie_Bin_Classific(&
               L2B_Mie_Meas_Classific%L2B_Mie_Meas_Bin_Classific(r))
       END DO
    ELSE
       print *,"        L2B_Mie_Meas_Classific%"//&
            "L2B_Mie_Meas_Bin_Classific(:) [seems not associated]"
    END IF

  end subroutine Print_L2B_Mie_Meas_Classific
    !  #]
  subroutine AsciiWrite_L2B_Mie_Meas_Classif(L2B_Mie_Meas_Classific,&
                                        fileunit,prefix,PrintAll,error_flag)
    !  #[
    TYPE(L2B_Mie_Meas_Classific_Type), intent(in)  :: L2B_Mie_Meas_Classific
    integer,                           intent(in)  :: fileunit   ! input
    character(len=*),                  intent(in)  :: prefix     ! input
    logical,                           intent(in)  :: PrintAll   ! input
    integer,                           intent(out) :: error_flag ! output

    ! local variables
    integer          :: r,s,step
    character(len=2) :: str_cntr

    IF (associated(L2B_Mie_Meas_Classific%L2B_Mie_Meas_Bin_Classific)) THEN
       ! get the size of the array
       s = size(L2B_Mie_Meas_Classific%L2B_Mie_Meas_Bin_Classific)
       IF (PrintAll .or. (s .le. 1) ) THEN
          step=1
       ELSE
          step = s-1
       END IF
       DO r=1,s,step
          write(str_cntr,"(i2.2)",err=998) r
          call AsciiWrite_L2B_Mie_Bin_Classif(&
               L2B_Mie_Meas_Classific%L2B_Mie_Meas_Bin_Classific(r),&
               fileunit,trim(prefix)//"%L2B_Mie_Meas_Bin_Classific("//&
               trim(str_cntr)//")",error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%L2B_Mie_Meas_Bin_Classific(:) [seems not associated]"
    END IF

    return

998 call logmsg(log_error,"conversion error in "//&
         "AsciiWrite_L2B_Mie_Meas_Classif: "//&
         "could not convert the counter r to a string ..")
    error_flag = error_writing_file
    return

999 call logmsg(log_error,"ERROR in AsciiWrite_L2B_Mie_Meas_Classif: "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2B_Mie_Meas_Classif
    !  #]
  function GetByteSize_L2B_Mie_Meas_Class(L2B_Mie_Meas_Classific) &
       result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result
    TYPE(L2B_Mie_Meas_Classific_Type), intent(in) :: L2B_Mie_Meas_Classific

    c_bytesize = 1 ! spare_byte
    IF (associated(L2B_Mie_Meas_Classific%L2B_Mie_Meas_Bin_Classific)) THEN
       c_bytesize = c_bytesize + &
            int(size(L2B_Mie_Meas_Classific%L2B_Mie_Meas_Bin_Classific)*&
                GetByteSize_L2B_Mie_Bin_Class(),IntAl)
    END IF

    ! typical sizes
    ! memory size =   24*12 = 288 bytes
    ! file   size = 1+24*11 = 265 bytes
  end function GetByteSize_L2B_Mie_Meas_Class
    !  #]
  subroutine Write_L2B_Mie_Meas_Classific(fileunit, &
             L2B_Mie_Meas_Classific, error_flag)
    !  #[
    !  Writes L2B_Mie_Meas_Classification in binary format
    !  P. Poli     09-01-2006 Original Code
    integer,                           intent(in)  :: fileunit
    TYPE(L2B_Mie_Meas_Classific_Type), intent(in)  :: L2B_Mie_Meas_Classific
    integer,                           intent(out) :: error_flag 

    ! local variable
    integer :: s, i, nr_spare_bytes

    !--------------------------------------------
    ! writing the L2B_Mie_Meas_Classific elements
    !--------------------------------------------

    error_flag = no_error

    s = size(L2B_Mie_Meas_Classific%L2B_Mie_Meas_Bin_Classific)
    DO i=1,s
       call Write_L2B_Mie_Bin_Classific(fileunit, &
            L2B_Mie_Meas_Classific%L2B_Mie_Meas_Bin_Classific(i), &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    ! 1 spare byte
    nr_spare_bytes = 1
    call write_SpareBytes_to_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Write_L2B_Mie_Meas_Classific")
    return

  end subroutine Write_L2B_Mie_Meas_Classific
    !  #]
  subroutine Read_L2B_Mie_Meas_Classific(fileunit, &
             L2B_Mie_Meas_Classific, error_flag)
    !  #[
    !  P. Poli     10-01-2006 Original Code
    integer,                           intent(in)    :: fileunit
    TYPE(L2B_Mie_Meas_Classific_Type), intent(inout) :: L2B_Mie_Meas_Classific
    integer,                           intent(out)   :: error_flag 

    ! local variable
    integer :: s, i, nr_spare_bytes

    !--------------------------------------------
    ! reading the L2B_Mie_Meas_Classific elements
    !--------------------------------------------

    error_flag = no_error

    s = size(L2B_Mie_Meas_Classific%L2B_Mie_Meas_Bin_Classific)
    DO i=1,s
       call Read_L2B_Mie_Bin_Classific(fileunit, &
            L2B_Mie_Meas_Classific%L2B_Mie_Meas_Bin_Classific(i), &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO
    
    ! 1 spare byte
    nr_spare_bytes = 1
    call read_SpareBytes_from_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Read_L2B_Mie_Meas_Classific")
    return

  end subroutine Read_L2B_Mie_Meas_Classific
    !  #]
  !---------------------------------------------------------------------
  subroutine  Init_L2B_Mie_Obs_Classific(L2B_Mie_Obs_Classific,&
                                         N_Range_Gates0, error_flag)
    !  #[
    TYPE(L2B_Mie_Obs_Classific_Type), intent(out) :: L2B_Mie_Obs_Classific
    integer(IntAs), intent(in)  :: N_Range_Gates0 ! input
    integer,        intent(out) :: error_flag    ! output

    ! local variable
    integer :: r, AllocStatus

    error_flag = no_error

    L2B_Mie_Obs_Classific%flag = 0_IntAuc

    IF (N_Range_Gates0 .eq. 0) THEN
       ! init them to be empty
       nullify(L2B_Mie_Obs_Classific%L2B_Mie_Obs_Bin_Classific)
    ELSE
       allocate(L2B_Mie_Obs_Classific%&
                L2B_Mie_Obs_Bin_Classific(N_Range_Gates0),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          call logmsg(log_error,"Allocation failure in subroutine"//&
                                "Init_L2B_Mie_Obs_Classific()")
          error_flag = error_allocate
          return
       END IF
       DO r=1,N_Range_Gates0
          call Init_L2B_Mie_Bin_Classific(&
                  L2B_Mie_Obs_Classific%L2B_Mie_Obs_Bin_Classific(r))
       END DO
    END IF
    
  end subroutine Init_L2B_Mie_Obs_Classific
    !  #]
  subroutine  Delete_L2B_Mie_Obs_Classific(L2B_Mie_Obs_Classific)
    !  #[
    TYPE(L2B_Mie_Obs_Classific_Type), &
         intent(inout) :: L2B_Mie_Obs_Classific

    IF (associated(L2B_Mie_Obs_Classific%L2B_Mie_Obs_Bin_Classific)) &
        deallocate(L2B_Mie_Obs_Classific%L2B_Mie_Obs_Bin_Classific)
    
  end subroutine Delete_L2B_Mie_Obs_Classific
    !  #]
  subroutine Print_L2B_Mie_Obs_Classific(L2B_Mie_Obs_Classific)
    !  #[
    TYPE(L2B_Mie_Obs_Classific_Type), intent(in) :: L2B_Mie_Obs_Classific

    integer :: r,s

    print *,"       ...%flag       = ",L2B_Mie_Obs_Classific%flag
    IF (associated(L2B_Mie_Obs_Classific%L2B_Mie_Obs_Bin_Classific)) THEN
       ! get the size of the array
       s = size(L2B_Mie_Obs_Classific%L2B_Mie_Obs_Bin_Classific)
       print "(a,i2,a)","    ==> size(L2B_Mie_Obs_Bin_Classific%"//&
            "L2B_Mie_Obs_Bin_Classific) = ",s
       print *,"    [printing first and last one]"
       DO r=1,s,s-1
          print "('        L2B_Mie_Obs_Classific%"//&
               "L2B_Mie_Obs_Bin_Classific(',i2.2,')%...')",r
          call Print_L2B_Mie_Bin_Classific(&
               L2B_Mie_Obs_Classific%L2B_Mie_Obs_Bin_Classific(r))
       END DO
    ELSE
       print *,"        L2B_Mie_Obs_Classific%"//&
            "L2B_Mie_Obs_Bin_Classific(:) [seems not associated]"
    END IF

  end subroutine Print_L2B_Mie_Obs_Classific
    !  #]
  subroutine AsciiWrite_L2B_Mie_Obs_Classif(L2B_Mie_Obs_Classific,fileunit,&
                                            prefix,PrintAll,error_flag)
    !  #[
    TYPE(L2B_Mie_Obs_Classific_Type), intent(in)  :: L2B_Mie_Obs_Classific
    integer,                          intent(in)  :: fileunit   ! input
    character(len=*),                 intent(in)  :: prefix     ! input
    logical,                          intent(in)  :: PrintAll   ! input
    integer,                          intent(out) :: error_flag ! output

    ! local variables
    integer                      :: r,s,step
    character(len=2)             :: str_cntr
    character(len=maxlen_IntAuc) :: str_IntAuc

    call SetIntAuc(L2B_Mie_Obs_Classific%flag,str_IntAuc,error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) trim(prefix)//"%flag = "//trim(str_IntAuc)

    IF (associated(L2B_Mie_Obs_Classific%L2B_Mie_Obs_Bin_Classific)) THEN
       ! get the size of the array
       s = size(L2B_Mie_Obs_Classific%L2B_Mie_Obs_Bin_Classific)
       IF (PrintAll .or. (s .le. 1) ) THEN
          step=1
       ELSE
          step=s-1
       END IF
       DO r=1,s,step
          write(str_cntr,"(i2.2)",err=998) r
          call AsciiWrite_L2B_Mie_Bin_Classif(&
               L2B_Mie_Obs_Classific%L2B_Mie_Obs_Bin_Classific(r),&
               fileunit,trim(prefix)//"%L2B_Mie_Obs_Bin_Classific("//&
               trim(str_cntr)//")",error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%L2B_Mie_Obs_Bin_Classific(:) [seems not associated]"
    END IF

    return

998 call logmsg(log_error,"conversion error in "//&
         "AsciiWrite_L2B_Mie_Obs_Classif: "//&
         "could not convert the counter r to a string ..")
    error_flag = error_writing_file
    return

999 call logmsg(log_error,"ERROR in AsciiWrite_L2B_Mie_Obs_Classif: "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2B_Mie_Obs_Classif
    !  #]
  function GetByteSize_L2B_Mie_Obs_Class(L2B_Mie_Obs_Classific) &
       result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result
    TYPE(L2B_Mie_Obs_Classific_Type), intent(in) :: L2B_Mie_Obs_Classific

    c_bytesize = c_bytelen_IntAuc + 1 ! spare_byte
    IF (associated(L2B_Mie_Obs_Classific%L2B_Mie_Obs_Bin_Classific)) THEN
       c_bytesize = c_bytesize + &
            int(size(L2B_Mie_Obs_Classific%L2B_Mie_Obs_Bin_Classific)*&
                GetByteSize_L2B_Mie_Bin_Class(),IntAl)
    END IF

    ! typical sizes
    ! memory size = 2  +24*12 = 290 bytes
    ! file   size = 1+1+24*11 = 266 bytes
  end function GetByteSize_L2B_Mie_Obs_Class
    !  #]
  subroutine Write_L2B_Mie_Obs_Classific(fileunit, &
             L2B_Mie_Obs_Classific, error_flag)
    !  #[
    !  Writes L2B_Mie_Obs_Classification in binary format
    !  P. Poli     09-01-2006 Original Code
    integer,                          intent(in)  :: fileunit
    TYPE(L2B_Mie_Obs_Classific_Type), intent(in)  :: L2B_Mie_Obs_Classific
    integer,                          intent(out) :: error_flag 

    ! local variable
    integer :: s, i, nr_spare_bytes

    !--------------------------------------------
    ! writing the L2B_Mie_Obs_Classific elements
    !--------------------------------------------

    error_flag = no_error

    call write_IntAuc_to_bin_file(fileunit, &
         L2B_Mie_Obs_Classific%flag, error_flag)
    IF (error_flag .ne. no_error) goto 998

    s = size(L2B_Mie_Obs_Classific%L2B_Mie_Obs_Bin_Classific)
    DO i=1,s
       call Write_L2B_Mie_Bin_Classific(fileunit, &
            L2B_Mie_Obs_Classific%L2B_Mie_Obs_Bin_Classific(i), &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    ! 1 spare byte
    nr_spare_bytes = 1
    call write_SpareBytes_to_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Write_L2B_Mie_Obs_Classific")
    return

  end subroutine Write_L2B_Mie_Obs_Classific
    !  #]
  subroutine Read_L2B_Mie_Obs_Classific(fileunit, &
             L2B_Mie_Obs_Classific, error_flag)
    !  #[
    !  P. Poli     10-01-2006 Original Code
    integer,                          intent(in)    :: fileunit
    TYPE(L2B_Mie_Obs_Classific_Type), intent(inout) :: L2B_Mie_Obs_Classific
    integer,                          intent(out)   :: error_flag 

    ! local variable
    integer :: s, i, nr_spare_bytes

    !--------------------------------------------
    ! reading the L2B_Mie_Obs_Classific elements
    !--------------------------------------------

    error_flag = no_error

    call read_IntAuc_from_bin_file(fileunit, &
         L2B_Mie_Obs_Classific%flag, error_flag)
    IF (error_flag .ne. no_error) goto 998

    s = size(L2B_Mie_Obs_Classific%L2B_Mie_Obs_Bin_Classific)
    DO i=1,s
       call Read_L2B_Mie_Bin_Classific(fileunit, &
            L2B_Mie_Obs_Classific%L2B_Mie_Obs_Bin_Classific(i), &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    ! 1 spare byte
    nr_spare_bytes = 1
    call read_SpareBytes_from_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Read_L2B_Mie_Obs_Classific")
    return

  end subroutine Read_L2B_Mie_Obs_Classific
    !  #]
  !---------------------------------------------------------------------
  subroutine  Init_Mie_Classification_QC(Mie_Classification_QC,&
                     N_Range_Gates0, N_Meas0, N_Obs_Mie_Max0, error_flag)
    !  #[
    TYPE(Mie_Classification_QC_Type), intent(out) :: Mie_Classification_QC
    integer(IntAs), intent(in)  :: N_Range_Gates0 ! input
    integer(IntAs), intent(in)  :: N_Meas0        ! input
    integer(IntAs), intent(in)  :: N_Obs_Mie_Max0     ! input
    integer,        intent(out) :: error_flag    ! output

    ! local variable
    integer :: m,o, AllocStatus

    error_flag = no_error

    IF (N_Meas0 .eq. 0) THEN
       ! init them to be empty
       nullify(Mie_Classification_QC%L2B_Mie_Meas_Classific)
    ELSE
       allocate(Mie_Classification_QC%L2B_Mie_Meas_Classific(N_Meas0),&
                STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          call logmsg(log_error,"Allocation failure in subroutine "//&
                                "Init_Mie_Classification_QC()")
          error_flag = error_allocate
          return
       END IF
       DO m=1,N_Meas0
          call Init_L2B_Mie_Meas_Classific(&
                  Mie_Classification_QC%L2B_Mie_Meas_Classific(m),&
                  N_Range_Gates0,error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    END IF

    IF (N_Obs_Mie_Max0 .eq. 0) THEN
       ! init them to be empty
       nullify(Mie_Classification_QC%L2B_Mie_Obs_Classific)
    ELSE
       allocate(Mie_Classification_QC%L2B_Mie_Obs_Classific(N_Obs_Mie_Max0),&
                STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          call logmsg(log_error,"Allocation failure in subroutine "//&
                                "Init_Mie_Classification_QC()")
          error_flag = error_allocate
          return
       END IF
       DO o=1,N_Obs_Mie_Max0
          call Init_L2B_Mie_Obs_Classific(&
                  Mie_Classification_QC%L2B_Mie_Obs_Classific(o),&
                  N_Range_Gates0,error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    END IF
    
  end subroutine Init_Mie_Classification_QC
    !  #]
  subroutine  Delete_Mie_Classification_QC(Mie_Classification_QC)
    !  #[
    TYPE(Mie_Classification_QC_Type), intent(out) :: Mie_Classification_QC
    
    ! local variables
    integer :: m, o

    IF (associated(Mie_Classification_QC%L2B_Mie_Meas_Classific)) THEN
       DO m=1,size(Mie_Classification_QC%L2B_Mie_Meas_Classific)
          call Delete_L2B_Mie_Meas_Classific(&
                  Mie_Classification_QC%L2B_Mie_Meas_Classific(m))
       END DO
       deallocate(Mie_Classification_QC%L2B_Mie_Meas_Classific)
    END IF

    IF (associated(Mie_Classification_QC%L2B_Mie_Obs_Classific)) THEN
       DO o=1,size(Mie_Classification_QC%L2B_Mie_Obs_Classific)
          call Delete_L2B_Mie_Obs_Classific(&
                  Mie_Classification_QC%L2B_Mie_Obs_Classific(o))
       END DO
       deallocate(Mie_Classification_QC%L2B_Mie_Obs_Classific)
    END IF
    
  end subroutine Delete_Mie_Classification_QC
    !  #]
  subroutine Print_Mie_Classification_QC(Mie_Classification_QC)
    !  #[
    TYPE(Mie_Classification_QC_Type), intent(in) :: Mie_Classification_QC

    integer :: m,s

    print *,"Mie_Classification_QC%L2B_Mie_Meas_Classific%..."
    IF (associated(Mie_Classification_QC%L2B_Mie_Meas_Classific)) THEN
       ! get the size of the array
       s = size(Mie_Classification_QC%L2B_Mie_Meas_Classific)
       print "(a,i2,a)"," ==> size(Mie_Classification_QC%"//&
            "L2B_Mie_Meas_Classific) = ",s
       print *,"   [printing first and last one]"
       DO m=1,s,s-1
          print "('    L2B_Mie_Meas_Classific%"//&
               "L2B_Mie_Meas_Bin_Classific(',i2.2,')%...')",m
          call Print_L2B_Mie_Meas_Classific(&
               Mie_Classification_QC%L2B_Mie_Meas_Classific(m))
       END DO
    ELSE
       print *,"    L2B_Mie_Meas_Classific%"//&
            "L2B_Mie_Meas_Bin_Classific(:) [seems not associated]"
    END IF

    print *,"Mie_Classification_QC%L2B_Mie_Obs_Classific%..."
    IF (associated(Mie_Classification_QC%L2B_Mie_Obs_Classific)) THEN
       ! get the size of the array
       s = size(Mie_Classification_QC%L2B_Mie_Obs_Classific)
       print "(a,i2,a)"," ==> size(Mie_Classification_QC%"//&
            "L2B_Mie_Obs_Classific) = ",s
       print *,"   [printing first and last one]"
       DO m=1,s,s-1
          print "('    L2B_Mie_Obs_Classific%"//&
               "L2B_Mie_Obs_Bin_Classific(',i2.2,')%...')",m
          call Print_L2B_Mie_Obs_Classific(&
               Mie_Classification_QC%L2B_Mie_Obs_Classific(m))
       END DO
    ELSE
       print *,"    L2B_Mie_Obs_Classific%"//&
            "L2B_Mie_Obs_Bin_Classific(:) [seems not associated]"
    END IF

  end subroutine Print_Mie_Classification_QC
    !  #]
  subroutine AsciiWrite_Mie_Classific_QC(Mie_Classification_QC,fileunit,&
                                         prefix,PrintAll,error_flag)
    !  #[
    TYPE(Mie_Classification_QC_Type), intent(in)  :: Mie_Classification_QC
    integer,                          intent(in)  :: fileunit   ! input
    character(len=*),                 intent(in)  :: prefix     ! input
    logical,                          intent(in)  :: PrintAll   ! input
    integer,                          intent(out) :: error_flag ! output

    ! local variables
    integer          :: m,s,step
    character(len=2) :: str_cntr

    IF (associated(Mie_Classification_QC%L2B_Mie_Meas_Classific)) THEN
       ! get the size of the array
       s = size(Mie_Classification_QC%L2B_Mie_Meas_Classific)
       IF (PrintAll .or. (s .le. 1) ) THEN
          step=1
       ELSE
          step=s-1
       END IF
       DO m=1,s,step
          write(str_cntr,"(i2.2)",err=998) m
          call AsciiWrite_L2B_Mie_Meas_Classif(&
               Mie_Classification_QC%L2B_Mie_Meas_Classific(m),fileunit,&
               trim(prefix)//"%L2B_Mie_Meas_Classific("//&
               trim(str_cntr)//")",PrintAll,error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    ELSE
        write(fileunit,"(a)",err=999) trim(prefix)//&
            "%L2B_Mie_Meas_Classific(:) [seems not associated]"
    END IF

    IF (associated(Mie_Classification_QC%L2B_Mie_Obs_Classific)) THEN
       ! get the size of the array
       s = size(Mie_Classification_QC%L2B_Mie_Obs_Classific)
       IF (PrintAll .or. (s .le. 1) ) THEN
          step=1
       ELSE
          step=s-1
       END IF
       DO m=1,s,step
          write(str_cntr,"(i2.2)",err=998) m
          call AsciiWrite_L2B_Mie_Obs_Classif(&
               Mie_Classification_QC%L2B_Mie_Obs_Classific(m),fileunit,&
               trim(prefix)//"%L2B_Mie_Obs_Classific("//&
               trim(str_cntr)//")",PrintAll,error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%L2B_Mie_Obs_Classific(:) [seems not associated]"
    END IF

    return

998 call logmsg(log_error,"conversion error in "//&
         "AsciiWrite_Mie_Classific_QC: "//&
         "could not convert the counter m to a string ...")
    error_flag = error_writing_file
    return

999 call logmsg(log_error,"ERROR in AsciiWrite_Mie_Classific_QC: "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Mie_Classific_QC
    !  #]
  function GetByteSize_Mie_Classific_QC(Mie_Classification_QC, &
       N_Meas_to_write, N_Obs_Mie_to_write) result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result
    TYPE(Mie_Classification_QC_Type), intent(in) :: Mie_Classification_QC
    integer(IntAuc),                  intent(in) :: N_Meas_to_write
    integer(IntAuc),                  intent(in) :: N_Obs_Mie_to_write

    ! local variable
    integer :: s, error_flag
    character(len=maxlen_IntAuc) :: str1_IntAuc, str2_IntAuc

    c_bytesize = 20 ! spare_bytes
    IF (associated(Mie_Classification_QC%L2B_Mie_Meas_Classific)) THEN
       s = size(Mie_Classification_QC%L2B_Mie_Meas_Classific)
       IF (N_Meas_to_write .gt. s) THEN
          call SetIntAuc(N_Meas_to_write,str1_IntAuc,error_flag)
          IF (error_flag .ne. no_error) str1_IntAuc = "[er]"
          call SetIntAuc(int(s,IntAuc), str2_IntAuc,error_flag)
          IF (error_flag .ne. no_error) str2_IntAuc = "[er]"
          call logmsg(log_warn,"WARNING in "//&
               "GetByteSize_Mie_Classific_QC(): invalid value for "//&
               "N_Meas_to_write: "//trim(str1_IntAuc)//&
               " max. possible value is: "//&
               "size(Mie_Classification_QC%L2B_Mie_Meas_Classific) = "//&
               trim(str2_IntAuc)//" GetByteSize result will be useless...")
       END IF
       c_bytesize = c_bytesize + &
            N_meas_to_write * GetByteSize_L2B_Mie_Meas_Class(&
               Mie_Classification_QC%L2B_Mie_Meas_Classific(1))
    END IF
    IF (associated(Mie_Classification_QC%L2B_Mie_Obs_Classific)) THEN
       s = size(Mie_Classification_QC%L2B_Mie_Obs_Classific)
       IF (N_Obs_Mie_to_write .gt. s) THEN
          call SetIntAuc(N_Obs_Mie_to_write,str1_IntAuc,error_flag)
          IF (error_flag .ne. no_error) str1_IntAuc = "[er]"
          call SetIntAuc(int(s,IntAuc), str2_IntAuc,error_flag)
          IF (error_flag .ne. no_error) str2_IntAuc = "[er]"
          call logmsg(log_warn,"WARNING in "//&
               "GetByteSize_Mie_Classific_QC(): invalid value for "//&
               "N_Obs_Mie_to_write: "//trim(str1_IntAuc)//&
               " max. possible value is: "//&
               "size(Mie_Classification_QC%L2B_Mie_Obs_Classific) = "//&
               trim(str2_IntAuc)//" GetByteSize result will be useless...")
       ENDIF
       c_bytesize = c_bytesize + N_Obs_Mie_to_write * &
            GetByteSize_L2B_Mie_Obs_Class(&
               Mie_Classification_QC%L2B_Mie_Obs_Classific(1))
    END IF

    ! typical sizes
    ! memory size =    67*288+5*290 = 20746 bytes
    ! file   size = 20+67*265+5*266 = 19105 bytes
  end function GetByteSize_Mie_Classific_QC
    !  #]
  subroutine Write_Mie_Classification_QC(fileunit, &
             Mie_Classification_QC, N_Meas_to_write, &
             N_Obs_Mie_to_write, error_flag)
    !  #[
    !  Writes the Mie_Classification_QC in binary format
    !  J. de Kloe  03-01-2006 Original Code
    !  P. Poli     10-01-2006 Removed placeholders. Put actual writing in place
    !                         From L2B QC tables v0.2 03-01-2005
    
    integer,           intent(in) :: fileunit              ! input
    TYPE(Mie_Classification_QC_Type),&
                       intent(in) :: Mie_Classification_QC ! input
    integer(IntAuc),   intent(in) :: N_Meas_to_write       ! input
!    integer,           intent(in) :: M_Mie0 ! renamed to:
    integer(IntAuc),   intent(in) :: N_Obs_Mie_to_write    ! input
    integer,           intent(out):: error_flag            ! output

    ! local variable
    integer :: i,s,nr_spare_bytes
    character(len=maxlen_IntAuc) :: str1_IntAuc, str2_IntAuc

    !----------------------------------------
    ! writing the Mie_Classification elements
    !----------------------------------------

    error_flag = no_error
    IF (.not. associated(Mie_Classification_QC%L2B_Mie_Meas_Classific)) THEN
       call logmsg(log_error,"ERROR in Write_Mie_Classification_QC(): "//&
            "The pointer to Mie_Classification_QC%L2B_Mie_Meas_Classific "//&
            "seems not associated !!")
       error_flag = error_programming
       return
    END IF
    IF (.not. associated(Mie_Classification_QC%L2B_Mie_Obs_Classific)) THEN
       call logmsg(log_error,"ERROR in Write_Mie_Classification_QC(): "//&
            "The pointer to Mie_Classification_QC%L2B_Mie_Obs_Classific "//&
            "seems not associated !!")
       error_flag = error_programming
       return
    END IF
    
    s = size(Mie_Classification_QC%L2B_Mie_Meas_Classific)
    IF (N_Meas_to_write .gt. s) THEN
       call SetIntAuc(N_Meas_to_write,str1_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
       call SetIntAuc(int(s,IntAuc), str2_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_error,"ERROR in "//&
            "Write_Mie_Classification_QC(): invalid value for "//&
            "N_Meas_to_write: "//trim(str1_IntAuc)//&
            " max. possible value is: "//&
            "size(Mie_Classification_QC%L2B_Mie_Meas_Classific) = "//&
            trim(str2_IntAuc))
       error_flag = error_programming
       return
    END IF

    DO i=1,N_Meas_to_write
       call Write_L2B_Mie_Meas_Classific(fileunit, &
            Mie_Classification_QC%L2B_Mie_Meas_Classific(i), error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    s = size(Mie_Classification_QC%L2B_Mie_Obs_Classific)
    IF (N_Obs_Mie_to_write .gt. s) THEN
       call SetIntAuc(N_Obs_Mie_to_write,str1_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
       call SetIntAuc(int(s,IntAuc), str2_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_error,"ERROR in "//&
            "Write_Mie_Classification_QC(): invalid value for "//&
            "N_Obs_Mie_to_write: "//trim(str1_IntAuc)//&
            " max. possible value is: "//&
            "size(Mie_Classification_QC%L2B_Mie_Obs_Classific) = "//&
            trim(str2_IntAuc))
       error_flag = error_programming
       return
    END IF

    DO i=1,N_Obs_Mie_to_write
       call Write_L2B_Mie_Obs_Classific(fileunit, &
            Mie_Classification_QC%L2B_Mie_Obs_Classific(i), &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    ! 20 spare bytes
    nr_spare_bytes = 20
    call write_SpareBytes_to_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Write_Mie_Classification_QC")
    return

  end subroutine Write_Mie_Classification_QC
    !  #]
  subroutine Read_Mie_Classification_QC(fileunit, &
             Mie_Classification_QC, error_flag)
    !  #[
    !  P. Poli     10-01-2006 Original code
    integer,        intent(in)    :: fileunit              ! input
    TYPE(Mie_Classification_QC_Type),&
                    intent(inout) :: Mie_Classification_QC ! input
    integer,        intent(out)   :: error_flag            ! output

    ! local variable
    integer :: i,s,nr_spare_bytes

    !----------------------------------------
    ! writing the Mie_Classification elements
    !----------------------------------------

    error_flag = no_error
    IF (.not. associated(Mie_Classification_QC%L2B_Mie_Meas_Classific)) THEN
       call logmsg(log_error,"ERROR in Read_Mie_Classification(): "//&
            "The pointer to Mie_Classification_QC%L2B_Mie_Meas_Classific "//&
            "seems not associated !!")
       error_flag = error_programming
       return
    END IF

    s = size(Mie_Classification_QC%L2B_Mie_Meas_Classific)    
    DO i=1,s
       call Read_L2B_Mie_Meas_Classific(fileunit, &
            Mie_Classification_QC%L2B_Mie_Meas_Classific(i), error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    ! N_Obs may turn out to be zero !
    ! so this check does not make sense anymore
    !IF (.not. associated(Mie_Classification_QC%L2B_Mie_Obs_Classific)) THEN
    !   call logmsg(log_error,"ERROR in Read_Mie_Classification(): "//&
    !        "The pointer to Mie_Classification_QC%L2B_Mie_Obs_Classific "//&
    !        "seems not associated !!")
    !   error_flag = error_programming
    !   return
    !END IF
    
    IF (associated(Mie_Classification_QC%L2B_Mie_Obs_Classific)) THEN
       s = size(Mie_Classification_QC%L2B_Mie_Obs_Classific)
       DO i=1,s
          call Read_L2B_Mie_Obs_Classific(fileunit, &
               Mie_Classification_QC%L2B_Mie_Obs_Classific(i), error_flag)
          IF (error_flag .ne. no_error) goto 998
       END DO
    END IF

    ! 20 spare bytes
    nr_spare_bytes = 20
    call read_SpareBytes_from_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    return

998 call logmsg(log_error,"ABORTING subroutine Read_Mie_Classification_QC")
    return

  end subroutine Read_Mie_Classification_QC
    !  #]
  !  #]
  !---------------------------------------------------------------------
  ! subroutines for the Rayl_Classification_QC sub-datastructure    
  !---------------------------------------------------------------------
  !  #[
  subroutine Init_L2B_Rayl_Bin_Classific(L2B_Rayl_Bin_Classific)
    !  #[
    TYPE(L2B_Rayl_Bin_Classific_Type), intent(out) :: L2B_Rayl_Bin_Classific
    
    L2B_Rayl_Bin_Classific%flag1 = 0
    L2B_Rayl_Bin_Classific%flag2 = 0
    L2B_Rayl_Bin_Classific%L2B_Reliability = missing_indicator_FAdoxy

  end subroutine Init_L2B_Rayl_Bin_Classific
    !  #]
  subroutine Print_L2B_Rayl_Bin_Classific(L2B_Rayl_Bin_Classific)
    !  #[
    TYPE(L2B_Rayl_Bin_Classific_Type), intent(in) :: L2B_Rayl_Bin_Classific

    print *,"       ...%flag1 = ",L2B_Rayl_Bin_Classific%flag1
    print *,"       ...%flag2 = ",L2B_Rayl_Bin_Classific%flag2
    print *,"       ...%L2B_Reliability             = ",&
         L2B_Rayl_Bin_Classific%L2B_Reliability

  end subroutine Print_L2B_Rayl_Bin_Classific
    !  #]
  subroutine AsciiWrite_L2B_Rayl_Bin_Classif(L2B_Rayl_Bin_Classific,&
                                             fileunit,prefix,error_flag)
    !  #[
    TYPE(L2B_Rayl_Bin_Classific_Type), intent(in)  :: L2B_Rayl_Bin_Classific
    integer,                           intent(in)  :: fileunit   ! input
    character(len=*),                  intent(in)  :: prefix     ! input
    integer,                           intent(out) :: error_flag ! output

    ! local variables
    character(len=maxlen_IntAucFlags) :: str_IntAucFlags
    character(len=25)                 :: str_FAdoxy

    call SetIntAucFlags(L2B_Rayl_Bin_Classific%flag1,str_IntAucFlags,&
                        error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%flag1 = "//trim(str_IntAucFlags)

    call SetIntAucFlags(L2B_Rayl_Bin_Classific%flag2,str_IntAucFlags,&
                        error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%flag2 = "//trim(str_IntAucFlags)

    write(fileunit,"(a)",err=999) "REALACC(12)"
    call SetFAdoFull_or_Missing(L2B_Rayl_Bin_Classific%L2B_Reliability,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%L2B_Reliability = "//trim(str_FAdoxy)

    write(fileunit,"(a)",err=999) "ENDREALACC"

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_L2B_Rayl_Bin_Classif "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2B_Rayl_Bin_Classif
    !  #]
  function GetByteSize_L2B_Rayl_Bin_Class() result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result

    c_bytesize = 2*c_bytelen_IntAuc + &
                 1*c_bytelen_FAdoxy + 1 ! spare_byte
    ! 3 items in total

    ! typical sizes
    ! memory size = 2*2+8   = 12 bytes
    ! file   size = 2*1+8+1 = 11 bytes
  end function GetByteSize_L2B_Rayl_Bin_Class
    !  #]
  subroutine Write_L2B_Rayl_Bin_Classific(fileunit, &
             L2B_Rayl_Bin_Classific, error_flag)
    !  #[
    !  Writes L2B_Rayl_Obs_Bin_Classification in binary format
    !  P. Poli     09-01-2006 Original Code
    integer,                           intent(in)  :: fileunit
    TYPE(L2B_Rayl_Bin_Classific_Type), intent(in)  :: L2B_Rayl_Bin_Classific
    integer,                           intent(out) :: error_flag 

    ! local variable
    integer :: nr_spare_bytes

    !-------------------------------------------
    ! writing the L2B_Rayl_Bin_Classific elements
    !-------------------------------------------

    error_flag = no_error

    call write_IntAuc_to_bin_file(fileunit, &
         L2B_Rayl_Bin_Classific%flag1, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_IntAuc_to_bin_file(fileunit, &
         L2B_Rayl_Bin_Classific%flag2, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_FAdoxy_to_bin_file(fileunit, &
         L2B_Rayl_Bin_Classific%L2B_Reliability, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 1 spare byte
    nr_spare_bytes = 1
    call write_SpareBytes_to_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Write_L2B_Rayl_Bin_Classific")
    return

  end subroutine Write_L2B_Rayl_Bin_Classific
    !  #]
  subroutine Read_L2B_Rayl_Bin_Classific(fileunit, &
             L2B_Rayl_Bin_Classific, error_flag)
    !  #[
    !  P. Poli     10-01-2006 Original Code
    integer,                           intent(in)    :: fileunit
    TYPE(L2B_Rayl_Bin_Classific_Type), intent(inout) :: L2B_Rayl_Bin_Classific
    integer,                           intent(out)   :: error_flag 

    ! local variable
    integer :: nr_spare_bytes

    !-------------------------------------------
    ! reading the L2B_Rayl_Bin_Classific elements
    !-------------------------------------------

    error_flag = no_error

    call read_IntAuc_from_bin_file(fileunit, &
         L2B_Rayl_Bin_Classific%flag1, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call read_IntAuc_from_bin_file(fileunit, &
         L2B_Rayl_Bin_Classific%flag2, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call read_FAdoxy_from_bin_file(fileunit, &
         L2B_Rayl_Bin_Classific%L2B_Reliability, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 1 spare byte
    nr_spare_bytes = 1
    call read_SpareBytes_from_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Read_L2B_Rayl_Bin_Classific")
    return

  end subroutine Read_L2B_Rayl_Bin_Classific
    !  #]
  !---------------------------------------------------------------------
  subroutine  Init_L2B_Rayl_Meas_Classific(L2B_Rayl_Meas_Classific,&
                                           N_Range_Gates0, error_flag)
    !  #[
    TYPE(L2B_Rayl_Meas_Classific_Type), intent(out) :: L2B_Rayl_Meas_Classific
    integer(IntAs), intent(in)  :: N_Range_Gates0 ! input
    integer,        intent(out) :: error_flag     ! output

    ! local variable
    integer :: r, AllocStatus

    error_flag = no_error

    IF (N_Range_Gates0 .eq. 0) THEN
       ! init them to be empty
       nullify(L2B_Rayl_Meas_Classific%L2B_Rayl_Meas_Bin_Classific)
    ELSE
       allocate(L2B_Rayl_Meas_Classific%&
                L2B_Rayl_Meas_Bin_Classific(N_Range_Gates0),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          call logmsg(log_error,"Allocation failure in subroutine "//&
                                "Init_L2B_Rayl_Meas_Classific()")
          error_flag = error_allocate
          return
       END IF
       DO r=1,N_Range_Gates0
          call Init_L2B_Rayl_Bin_Classific(&
                  L2B_Rayl_Meas_Classific%L2B_Rayl_Meas_Bin_Classific(r))
       END DO
    END IF
    
  end subroutine Init_L2B_Rayl_Meas_Classific
    !  #]
  subroutine  Delete_L2B_Rayl_Meas_Classific(L2B_Rayl_Meas_Classific)
    !  #[
    TYPE(L2B_Rayl_Meas_Classific_Type), &
         intent(inout) :: L2B_Rayl_Meas_Classific

    IF (associated(L2B_Rayl_Meas_Classific%L2B_Rayl_Meas_Bin_Classific)) &
         deallocate(L2B_Rayl_Meas_Classific%L2B_Rayl_Meas_Bin_Classific)
    
  end subroutine Delete_L2B_Rayl_Meas_Classific
    !  #]
  subroutine Print_L2B_Rayl_Meas_Classific(L2B_Rayl_Meas_Classific)
    !  #[
    TYPE(L2B_Rayl_Meas_Classific_Type), intent(in) :: L2B_Rayl_Meas_Classific

    integer :: r,s

    IF (associated(L2B_Rayl_Meas_Classific%L2B_Rayl_Meas_Bin_Classific)) THEN
       ! get the size of the array
       s = size(L2B_Rayl_Meas_Classific%L2B_Rayl_Meas_Bin_Classific)
       print "(a,i2,a)","    ==> size(L2B_Rayl_Meas_Classific%"//&
            "L2B_Rayl_Meas_Bin_Classific) = ",s
       print *,"       [printing first and last one]"
       DO r=1,s,s-1
          print "('        L2B_Rayl_Meas_Classific%"//&
               "L2B_Rayl_Meas_Bin_Classific(',i2.2,')%...')",r
          call Print_L2B_Rayl_Bin_Classific(&
               L2B_Rayl_Meas_Classific%L2B_Rayl_Meas_Bin_Classific(r))
       END DO
    ELSE
       print *,"        L2B_Rayl_Meas_Classific%"//&
            "L2B_Rayl_Meas_Bin_Classific(:) [seems not associated]"
    END IF

  end subroutine Print_L2B_Rayl_Meas_Classific
    !  #]
  subroutine AsciiWrite_L2B_Rayl_Meas_Class(L2B_Rayl_Meas_Classific,&
                                         fileunit,prefix,PrintAll,error_flag)
    !  #[
    TYPE(L2B_Rayl_Meas_Classific_Type), intent(in)  :: L2B_Rayl_Meas_Classific
    integer,                            intent(in)  :: fileunit   ! input
    character(len=*),                   intent(in)  :: prefix     ! input
    logical,                            intent(in)  :: PrintAll   ! input
    integer,                            intent(out) :: error_flag ! output

    ! local variables
    integer          :: r,s,step
    character(len=2) :: str_cntr

    IF (associated(L2B_Rayl_Meas_Classific%L2B_Rayl_Meas_Bin_Classific)) THEN
       ! get the size of the array
       s = size(L2B_Rayl_Meas_Classific%L2B_Rayl_Meas_Bin_Classific)
       IF (PrintAll .or. (s .le. 1) ) THEN
          step=1
       ELSE
          step = s-1
       END IF
       DO r=1,s,step
          write(str_cntr,"(i2.2)",err=998) r
          call AsciiWrite_L2B_Rayl_Bin_Classif(&
               L2B_Rayl_Meas_Classific%L2B_Rayl_Meas_Bin_Classific(r),&
               fileunit,trim(prefix)//"%L2B_Rayl_Meas_Bin_Classific("//&
               trim(str_cntr)//")",error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%L2B_Rayl_Meas_Bin_Classific(:) [seems not associated]"
    END IF

    return

998 call logmsg(log_error,"conversion error in "//&
         "AsciiWrite_L2B_Rayl_Meas_Class: "//&
         "could not convert the counter r to a string ..")
    error_flag = error_writing_file
    return

999 call logmsg(log_error,"ERROR in AsciiWrite_L2B_Rayl_Meas_Class: "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2B_Rayl_Meas_Class
    !  #]
  function GetByteSize_L2B_Rayl_Meas_Class(L2B_Rayl_Meas_Classific) &
       result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result
    TYPE(L2B_Rayl_Meas_Classific_Type), intent(in) :: L2B_Rayl_Meas_Classific

    c_bytesize = 1 ! spare_byte
    IF (associated(L2B_Rayl_Meas_Classific%L2B_Rayl_Meas_Bin_Classific)) THEN
       c_bytesize = c_bytesize + &
            int(size(L2B_Rayl_Meas_Classific%L2B_Rayl_Meas_Bin_Classific)*&
                GetByteSize_L2B_Rayl_Bin_Class(),IntAl)
    END IF

    ! typical sizes
    ! memory size =   24*12 = 288 bytes
    ! file   size = 1+24*11 = 265 bytes
  end function GetByteSize_L2B_Rayl_Meas_Class
    !  #]
  subroutine Write_L2B_Rayl_Meas_Classific(fileunit, &
             L2B_Rayl_Meas_Classific, error_flag)
    !  #[
    !  Writes L2B_Rayl_Meas_Classification in binary format
    !  P. Poli     09-01-2006 Original Code
    integer,                            intent(in)  :: fileunit
    TYPE(L2B_Rayl_Meas_Classific_Type), intent(in)  :: L2B_Rayl_Meas_Classific
    integer,                            intent(out) :: error_flag 

    ! local variable
    integer :: s, i, nr_spare_bytes

    !--------------------------------------------
    ! writing the L2B_Rayl_Meas_Classific elements
    !--------------------------------------------

    error_flag = no_error

    s = size(L2B_Rayl_Meas_Classific%L2B_Rayl_Meas_Bin_Classific)
    DO i=1,s
       call Write_L2B_Rayl_Bin_Classific(fileunit, &
            L2B_Rayl_Meas_Classific%L2B_Rayl_Meas_Bin_Classific(i), &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    ! 1 spare byte
    nr_spare_bytes = 1
    call write_SpareBytes_to_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Write_L2B_Rayl_Meas_Classific")
    return

  end subroutine Write_L2B_Rayl_Meas_Classific
    !  #]
  subroutine Read_L2B_Rayl_Meas_Classific(fileunit, &
             L2B_Rayl_Meas_Classific, error_flag)
    !  #[
    !  P. Poli     10-01-2006 Original Code
    integer,                           intent(in)    :: fileunit
    TYPE(L2B_Rayl_Meas_Classific_Type),intent(inout) :: L2B_Rayl_Meas_Classific
    integer,                           intent(out)   :: error_flag 

    ! local variable
    integer :: s, i, nr_spare_bytes

    !--------------------------------------------
    ! reading the L2B_Rayl_Meas_Classific elements
    !--------------------------------------------

    error_flag = no_error
 
    s = size(L2B_Rayl_Meas_Classific%L2B_Rayl_Meas_Bin_Classific)
    DO i=1,s
       call Read_L2B_Rayl_Bin_Classific(fileunit, &
            L2B_Rayl_Meas_Classific%L2B_Rayl_Meas_Bin_Classific(i), &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO
    
    ! 1 spare byte
    nr_spare_bytes = 1
    call read_SpareBytes_from_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    return

998 call logmsg(log_error,"ABORTING subroutine Read_L2B_Rayl_Meas_Classific")
    return

  end subroutine Read_L2B_Rayl_Meas_Classific
    !  #]
  !---------------------------------------------------------------------
  subroutine  Init_L2B_Rayl_Obs_Classific(L2B_Rayl_Obs_Classific,&
                                         N_Range_Gates0, error_flag)
    !  #[
    TYPE(L2B_Rayl_Obs_Classific_Type), intent(out) :: L2B_Rayl_Obs_Classific
    integer(IntAs), intent(in)  :: N_Range_Gates0 ! input
    integer,        intent(out) :: error_flag    ! output

    ! local variable
    integer :: r, AllocStatus

    error_flag = no_error

    L2B_Rayl_Obs_Classific%flag = 0_IntAuc

    IF (N_Range_Gates0 .eq. 0) THEN
       ! init them to be empty
       nullify(L2B_Rayl_Obs_Classific%L2B_Rayl_Obs_Bin_Classific)
    ELSE
       allocate(L2B_Rayl_Obs_Classific%&
                L2B_Rayl_Obs_Bin_Classific(N_Range_Gates0),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          call logmsg(log_error,"Allocation failure in subroutine "//&
                                "Init_L2B_Rayl_Obs_Classific()")
          error_flag = error_allocate
          return
       END IF
       DO r=1,N_Range_Gates0
          call Init_L2B_Rayl_Bin_Classific(&
                  L2B_Rayl_Obs_Classific%L2B_Rayl_Obs_Bin_Classific(r))
       END DO
    END IF
    
  end subroutine Init_L2B_Rayl_Obs_Classific
    !  #]
  subroutine  Delete_L2B_Rayl_Obs_Classific(L2B_Rayl_Obs_Classific)
    !  #[
    TYPE(L2B_Rayl_Obs_Classific_Type), &
         intent(inout) :: L2B_Rayl_Obs_Classific

    IF (associated(L2B_Rayl_Obs_Classific%L2B_Rayl_Obs_Bin_Classific)) &
        deallocate(L2B_Rayl_Obs_Classific%L2B_Rayl_Obs_Bin_Classific)
    
  end subroutine Delete_L2B_Rayl_Obs_Classific
    !  #]
  subroutine Print_L2B_Rayl_Obs_Classific(L2B_Rayl_Obs_Classific)
    !  #[
    TYPE(L2B_Rayl_Obs_Classific_Type), intent(in) :: L2B_Rayl_Obs_Classific

    integer :: r,s

    print *,"       ...%flag       = ",L2B_Rayl_Obs_Classific%flag
    IF (associated(L2B_Rayl_Obs_Classific%L2B_Rayl_Obs_Bin_Classific)) THEN
       ! get the size of the array
       s = size(L2B_Rayl_Obs_Classific%L2B_Rayl_Obs_Bin_Classific)
       print "(a,i2,a)","    ==> size(L2B_Rayl_Obs_Bin_Classific%"//&
            "L2B_Rayl_Obs_Bin_Classific) = ",s
       print *,"    [printing first and last one]"
       DO r=1,s,s-1
          print "('        L2B_Rayl_Obs_Classific%"//&
               "L2B_Rayl_Obs_Bin_Classific(',i2.2,')%...')",r
          call Print_L2B_Rayl_Bin_Classific(&
               L2B_Rayl_Obs_Classific%L2B_Rayl_Obs_Bin_Classific(r))
       END DO
    ELSE
       print *,"        L2B_Rayl_Obs_Classific%"//&
            "L2B_Rayl_Obs_Bin_Classific(:) [seems not associated]"
    END IF

  end subroutine Print_L2B_Rayl_Obs_Classific
    !  #]
  subroutine AsciiWrite_L2B_Rayl_Obs_Class(L2B_Rayl_Obs_Classific,&
                                       fileunit,prefix,PrintAll,error_flag)
    !  #[
    TYPE(L2B_Rayl_Obs_Classific_Type), intent(in) :: L2B_Rayl_Obs_Classific
    integer,                          intent(in)  :: fileunit   ! input
    character(len=*),                 intent(in)  :: prefix     ! input
    logical,                          intent(in)  :: PrintAll   ! input
    integer,                          intent(out) :: error_flag ! output

    ! local variables
    integer          :: r,s,step
    character(len=2) :: str_cntr
    character(len=maxlen_IntAuc) :: str_IntAuc

    call SetIntAuc(L2B_Rayl_Obs_Classific%flag,str_IntAuc,error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) trim(prefix)//"%flag = "//trim(str_IntAuc)

    IF (associated(L2B_Rayl_Obs_Classific%L2B_Rayl_Obs_Bin_Classific)) THEN
       ! get the size of the array
       s = size(L2B_Rayl_Obs_Classific%L2B_Rayl_Obs_Bin_Classific)
       IF (PrintAll .or. (s .le. 1) ) THEN
          step=1
       ELSE
          step=s-1
       END IF
       DO r=1,s,step
          write(str_cntr,"(i2.2)",err=998) r
          call AsciiWrite_L2B_Rayl_Bin_Classif(&
               L2B_Rayl_Obs_Classific%L2B_Rayl_Obs_Bin_Classific(r),&
               fileunit,trim(prefix)//"%L2B_Rayl_Obs_Bin_Classific("//&
               trim(str_cntr)//")",error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%L2B_Rayl_Obs_Bin_Classific(:) [seems not associated]"
    END IF

    return

998 call logmsg(log_error,"conversion error in "//&
         "AsciiWrite_L2B_Rayl_Obs_Class: "//&
         "could not convert the counter r to a string ..")
    error_flag = error_writing_file
    return

999 call logmsg(log_error,"ERROR in AsciiWrite_L2B_Rayl_Obs_Class: "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2B_Rayl_Obs_Class
    !  #]
  function GetByteSize_L2B_Rayl_Obs_Class(L2B_Rayl_Obs_Classific) &
       result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result
    TYPE(L2B_Rayl_Obs_Classific_Type), intent(in) :: L2B_Rayl_Obs_Classific

    c_bytesize = c_bytelen_IntAuc + 1 ! spare_byte
    IF (associated(L2B_Rayl_Obs_Classific%L2B_Rayl_Obs_Bin_Classific)) THEN
       c_bytesize = c_bytesize + &
            int(size(L2B_Rayl_Obs_Classific%L2B_Rayl_Obs_Bin_Classific)*&
                GetByteSize_L2B_Rayl_Bin_Class(),IntAl)
    END IF

    ! typical sizes
    ! memory size = 2  +24*12 = 290 bytes
    ! file   size = 1+1+24*11 = 266 bytes
  end function GetByteSize_L2B_Rayl_Obs_Class
    !  #]
  subroutine Write_L2B_Rayl_Obs_Classific(fileunit, &
       L2B_Rayl_Obs_Classific, error_flag)
    !  #[
    !  Writes L2B_Rayl_Obs_Classification in binary format
    !  P. Poli     09-01-2006 Original Code
    integer,                           intent(in)  :: fileunit
    TYPE(L2B_Rayl_Obs_Classific_Type), intent(in)  :: L2B_Rayl_Obs_Classific
    integer,                           intent(out) :: error_flag 

    ! local variable
    integer :: s, i, nr_spare_bytes

    error_flag = no_error

    !--------------------------------------------
    ! writing the L2B_Rayl_Obs_Classific elements
    !--------------------------------------------

    error_flag = no_error

    call write_IntAuc_to_bin_file(fileunit, &
         L2B_Rayl_Obs_Classific%flag, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    s = size(L2B_Rayl_Obs_Classific%L2B_Rayl_Obs_Bin_Classific)
    DO i=1,s
       call Write_L2B_Rayl_Bin_Classific(fileunit, &
            L2B_Rayl_Obs_Classific%L2B_Rayl_Obs_Bin_Classific(i), &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    ! 1 spare byte
    nr_spare_bytes = 1
    call write_SpareBytes_to_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Write_L2B_Rayl_Obs_Classific")
    return

  end subroutine Write_L2B_Rayl_Obs_Classific
    !  #]
  subroutine Read_L2B_Rayl_Obs_Classific(fileunit, &
             L2B_Rayl_Obs_Classific, error_flag)
    !  #[
    !  P. Poli     10-01-2006 Original Code
    integer,                           intent(in)    :: fileunit
    TYPE(L2B_Rayl_Obs_Classific_Type), intent(inout) :: L2B_Rayl_Obs_Classific
    integer,                           intent(out)   :: error_flag 

    ! local variable
    integer :: s, i, nr_spare_bytes

    !--------------------------------------------
    ! reading the L2B_Rayl_Obs_Classific elements
    !--------------------------------------------

    error_flag = no_error

    call read_IntAuc_from_bin_file(fileunit, &
         L2B_Rayl_Obs_Classific%flag, error_flag)
    IF (error_flag .ne. no_error) goto 998

    s = size(L2B_Rayl_Obs_Classific%L2B_Rayl_Obs_Bin_Classific)
    DO i=1,s
       call Read_L2B_Rayl_Bin_Classific(fileunit, &
            L2B_Rayl_Obs_Classific%L2B_Rayl_Obs_Bin_Classific(i), &
            error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    ! 1 spare byte
    nr_spare_bytes = 1
    call read_SpareBytes_from_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Read_L2B_Rayl_Obs_Classific")
    return

  end subroutine Read_L2B_Rayl_Obs_Classific
    !  #]
  !---------------------------------------------------------------------
  subroutine  Init_Rayl_Classification_QC(Rayl_Classification_QC,&
                     N_Range_Gates0, N_Meas0, N_Obs_Rayleigh_Max0, error_flag)
    !  #[
    TYPE(Rayl_Classification_QC_Type), intent(out) :: Rayl_Classification_QC
    integer(IntAs), intent(in)  :: N_Range_Gates0 ! input
    integer(IntAs), intent(in)  :: N_Meas0        ! input
    integer(IntAs), intent(in)  :: N_Obs_Rayleigh_Max0     ! input
    integer,        intent(out) :: error_flag    ! output

    ! local variable
    integer :: m,o, AllocStatus

    error_flag = no_error

    IF (N_Meas0 .eq. 0) THEN
       ! init them to be empty
       nullify(Rayl_Classification_QC%L2B_Rayl_Meas_Classific)
    ELSE
       allocate(Rayl_Classification_QC%L2B_Rayl_Meas_Classific(N_Meas0),&
                STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          call logmsg(log_error,"Allocation failure in subroutine "//&
                                "Init_Rayl_Classification_QC()")
          error_flag = error_allocate
          return
       END IF
       DO m=1,N_Meas0
          call Init_L2B_Rayl_Meas_Classific(&
                  Rayl_Classification_QC%L2B_Rayl_Meas_Classific(m),&
                  N_Range_Gates0,error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    END IF

    IF (N_Obs_Rayleigh_Max0 .eq. 0) THEN
       ! init them to be empty
       nullify(Rayl_Classification_QC%L2B_Rayl_Obs_Classific)
    ELSE
       allocate(Rayl_Classification_QC%L2B_Rayl_Obs_Classific(N_Obs_Rayleigh_Max0),&
                STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          call logmsg(log_error,"Allocation failure in subroutine "//&
                                "Init_Rayl_Classification_QC()")
          error_flag = error_allocate
          return
       END IF
       DO o=1,N_Obs_Rayleigh_Max0
          call Init_L2B_Rayl_Obs_Classific(&
                  Rayl_Classification_QC%L2B_Rayl_Obs_Classific(o),&
                  N_Range_Gates0,error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    END IF
    
  end subroutine Init_Rayl_Classification_QC
    !  #]
  subroutine  Delete_Rayl_Classification_QC(Rayl_Classification_QC)
    !  #[
    TYPE(Rayl_Classification_QC_Type), intent(out) :: Rayl_Classification_QC
    
    ! local variables
    integer :: m, o

    IF (associated(Rayl_Classification_QC%L2B_Rayl_Meas_Classific)) THEN
       DO m=1,size(Rayl_Classification_QC%L2B_Rayl_Meas_Classific)
          call Delete_L2B_Rayl_Meas_Classific(&
                  Rayl_Classification_QC%L2B_Rayl_Meas_Classific(m))
       END DO
       deallocate(Rayl_Classification_QC%L2B_Rayl_Meas_Classific)
    END IF

    IF (associated(Rayl_Classification_QC%L2B_Rayl_Obs_Classific)) THEN
       DO o=1,size(Rayl_Classification_QC%L2B_Rayl_Obs_Classific)
          call Delete_L2B_Rayl_Obs_Classific(&
                  Rayl_Classification_QC%L2B_Rayl_Obs_Classific(o))
       END DO
       deallocate(Rayl_Classification_QC%L2B_Rayl_Obs_Classific)
    END IF
    
  end subroutine Delete_Rayl_Classification_QC
    !  #]
  subroutine Print_Rayl_Classification_QC(Rayl_Classification_QC)
    !  #[
    TYPE(Rayl_Classification_QC_Type), intent(in) :: Rayl_Classification_QC

    integer :: m,s

    print *,"Rayl_Classification_QC%L2B_MRayl_eas_Classific%..."
    IF (associated(Rayl_Classification_QC%L2B_Rayl_Meas_Classific)) THEN
       ! get the size of the array
       s = size(Rayl_Classification_QC%L2B_Rayl_Meas_Classific)
       print "(a,i2,a)"," ==> size(Rayl_Classification_QC%"//&
            "L2B_Rayl_Meas_Classific) = ",s
       print *,"   [printing first and last one]"
       DO m=1,s,s-1
          print "('    L2B_Rayl_Meas_Classific%"//&
               "L2B_Rayl_Meas_Bin_Classific(',i2.2,')%...')",m
          call Print_L2B_Rayl_Meas_Classific(&
               Rayl_Classification_QC%L2B_Rayl_Meas_Classific(m))
       END DO
    ELSE
       print *,"    L2B_Rayl_Meas_Classific%"//&
            "L2B_Rayl_Meas_Bin_Classific(:) [seems not associated]"
    END IF

    print *,"Rayl_Classification_QC%L2B_Rayl_Obs_Classific%..."
    IF (associated(Rayl_Classification_QC%L2B_Rayl_Obs_Classific)) THEN
       ! get the size of the array
       s = size(Rayl_Classification_QC%L2B_Rayl_Obs_Classific)
       print "(a,i2,a)"," ==> size(Rayl_Classification_QC%"//&
            "L2B_Rayl_Obs_Classific) = ",s
       print *,"   [printing first and last one]"
       DO m=1,s,s-1
          print "('    L2B_Rayl_Obs_Classific%"//&
               "L2B_Rayl_Obs_Bin_Classific(',i2.2,')%...')",m
          call Print_L2B_Rayl_Obs_Classific(&
               Rayl_Classification_QC%L2B_Rayl_Obs_Classific(m))
       END DO
    ELSE
       print *,"    L2B_Rayl_Obs_Classific%"//&
            "L2B_Rayl_Obs_Bin_Classific(:) [seems not associated]"
    END IF

  end subroutine Print_Rayl_Classification_QC
    !  #]
  subroutine AsciiWrite_Rayl_Classific_QC(Rayl_Classification_QC,fileunit,&
                                          prefix,PrintAll,error_flag)
    !  #[
    TYPE(Rayl_Classification_QC_Type), intent(in)  :: Rayl_Classification_QC
    integer,                           intent(in)  :: fileunit   ! input
    character(len=*),                  intent(in)  :: prefix     ! input
    logical,                           intent(in)  :: PrintAll   ! input
    integer,                           intent(out) :: error_flag ! output


    ! local variables
    integer          :: m,s,step
    character(len=2) :: str_cntr

    IF (associated(Rayl_Classification_QC%L2B_Rayl_Meas_Classific)) THEN
       ! get the size of the array
       s = size(Rayl_Classification_QC%L2B_Rayl_Meas_Classific)
       IF (PrintAll .or. (s .le. 1) ) THEN
          step=1
       ELSE
          step=s-1
       END IF
       DO m=1,s,step
          write(str_cntr,"(i2.2)",err=998) m
          call AsciiWrite_L2B_Rayl_Meas_Class(&
               Rayl_Classification_QC%L2B_Rayl_Meas_Classific(m),fileunit,&
               trim(prefix)//"%L2B_Rayl_Meas_Classific("//&
               trim(str_cntr)//")",PrintAll,error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%L2B_Rayl_Meas_Classific(:) [seems not associated]"
    END IF

    IF (associated(Rayl_Classification_QC%L2B_Rayl_Obs_Classific)) THEN
       ! get the size of the array
       s = size(Rayl_Classification_QC%L2B_Rayl_Obs_Classific)
       IF (PrintAll .or. (s .le. 1) ) THEN
          step=1
       ELSE
          step=s-1
       END IF
       DO m=1,s,step
          write(str_cntr,"(i2.2)",err=998) m
          call AsciiWrite_L2B_Rayl_Obs_Class(&
               Rayl_Classification_QC%L2B_Rayl_Obs_Classific(m),fileunit,&
               trim(prefix)//"%L2B_Rayl_Obs_Classific("//&
               trim(str_cntr)//")",PrintAll,error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%L2B_Rayl_Obs_Classific(:) [seems not associated]"
    END IF

    return

998 call logmsg(log_error,"conversion error in "//&
         "AsciiWrite_Rayl_Classific_QC: "//&
         "could not convert the counter m to a string ..")
    error_flag = error_writing_file
    return

999 call logmsg(log_error,"ERROR in AsciiWrite_Rayl_Classific_QC: "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Rayl_Classific_QC
    !  #]
  function GetByteSize_Rayl_Classific_QC(Rayl_Classification_QC, &
       N_Meas_to_write,N_Obs_Rayleigh_to_write) result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result
    TYPE(Rayl_Classification_QC_Type), intent(in) :: Rayl_Classification_QC
    integer(IntAuc),                   intent(in) :: N_Meas_to_write
    integer(IntAuc),                   intent(in) :: N_Obs_Rayleigh_to_write

    ! local variables
    integer :: s, error_flag
    character(len=maxlen_IntAuc) :: str1_IntAuc, str2_IntAuc

    c_bytesize = 20 ! spare_bytes
    IF (associated(Rayl_Classification_QC%L2B_Rayl_Meas_Classific)) THEN
       s = size(Rayl_Classification_QC%L2B_Rayl_Meas_Classific)
       IF (N_Meas_to_write .gt. s) THEN
          call SetIntAuc(N_Meas_to_write,str1_IntAuc,error_flag)
          IF (error_flag .ne. no_error) str1_IntAuc = "[er]"
          call SetIntAuc(int(s,IntAuc), str2_IntAuc,error_flag)
          IF (error_flag .ne. no_error) str2_IntAuc = "[er]"
          call logmsg(log_warn,"WARNING in "//&
               "GetByteSize_Rayleigh_Classific_QC(): invalid value for "//&
               "N_Meas_to_write: "//trim(str1_IntAuc)//&
               " max. possible value is: "//&
               "size(Rayl_Classification_QC%L2B_Rayl_Meas_Classific) = "//&
               trim(str2_IntAuc)//"GetByteSize result will be useless....")
       END IF
       c_bytesize = c_bytesize + &
            N_Meas_to_write * GetByteSize_L2B_Rayl_Meas_Class(&
               Rayl_Classification_QC%L2B_Rayl_Meas_Classific(1))
    END IF
    IF (associated(Rayl_Classification_QC%L2B_Rayl_Obs_Classific)) THEN
       s = size(Rayl_Classification_QC%L2B_Rayl_Obs_Classific)
       IF (N_Obs_Rayleigh_to_write .gt. s) THEN
          call SetIntAuc(N_Obs_Rayleigh_to_write,str1_IntAuc,error_flag)
          IF (error_flag .ne. no_error) str1_IntAuc = "[er]"
          call SetIntAuc(int(s,IntAuc), str2_IntAuc,error_flag)
          IF (error_flag .ne. no_error) str2_IntAuc = "[er]"
          call logmsg(log_warn,"WARNING in "//&
               "GetByteSize_Rayleigh_Classific_QC(): invalid value for "//&
               "N_Obs_Rayleigh_to_write: "//trim(str1_IntAuc)//&
               " max. possible value is: "//&
               "size(Rayl_Classification_QC%L2B_Rayl_Obs_Classific) = "//&
               trim(str2_IntAuc)//"GetByteSize result will be useless....")
       ENDIF
       c_bytesize = c_bytesize + &
            N_Obs_Rayleigh_to_write * GetByteSize_L2B_Rayl_Obs_Class(&
               Rayl_Classification_QC%L2B_Rayl_Obs_Classific(1))
    END IF

    ! typical sizes
    ! memory size =    67*288+5*290 = 20746 bytes
    ! file   size = 20+67*265+5*266 = 19105 bytes
  end function GetByteSize_Rayl_Classific_QC
    !  #]
  subroutine Write_Rayl_Classification_QC(fileunit, &
             Rayl_Classification_QC, N_Meas_to_write, &
             N_Obs_Rayleigh_to_write, error_flag)
    !  #[
    !  Writes the Rayleigh_Classification_QC in binary format
    !  J. de Kloe  03-01-2006 Original Code
    !  P. Poli     10-01-2006 Removed placeholders. Put actual writing in place
    !                         From L2B QC tables v0.2 03-01-2005
    
    integer,               intent(in) :: fileunit                 ! input
    TYPE(Rayl_Classification_QC_Type),&
                           intent(in)  :: Rayl_Classification_QC  ! input
    integer(IntAuc),       intent(in)  :: N_Meas_to_write         ! input
    integer(IntAuc),       intent(in)  :: N_Obs_Rayleigh_to_write ! input
    integer,               intent(out):: error_flag               ! output

    ! local variable
    integer :: i,s,nr_spare_bytes
    character(len=maxlen_IntAuc) :: str1_IntAuc, str2_IntAuc

    !---------------------------------------------
    ! writing the Rayleigh_Classification elements
    !---------------------------------------------

    error_flag = no_error
    IF (.not. associated(Rayl_Classification_QC%L2B_Rayl_Meas_Classific)) THEN
       call logmsg(log_error,"ERROR in Write_Rayl_Classification_QC(): "//&
            "The pointer to Rayl_Classification_QC%L2B_Rayl_Meas_Classific "//&
            "seems not associated !!")
       error_flag = error_programming
       return
    END IF
    IF (.not. associated(Rayl_Classification_QC%L2B_Rayl_Obs_Classific)) THEN
       call logmsg(log_error,"ERROR in Write_Rayl_Classification_QC(): "//&
            "The pointer to Rayl_Classification_QC%L2B_Rayl_Obs_Classific "//&
            "seems not associated !!")
       error_flag = error_programming
       return
    END IF

    s = size(Rayl_Classification_QC%L2B_Rayl_Meas_Classific)
    IF (N_Meas_to_write .gt. s) THEN
       call SetIntAuc(N_Meas_to_write,str1_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
       call SetIntAuc(int(s,IntAuc), str2_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_error,"ERROR in Write_Rayl_Classification_QC(): "//&
            "invalid value for N_Meas_to_write: "//trim(str1_IntAuc)//&
            " max. possible value is: "//&
            "size(Rayl_Classification_QC%L2B_Rayl_Meas_Classific) = "//&
            trim(str2_IntAuc))
       error_flag = error_programming
       return
    END IF  

    DO i=1,N_Meas_to_write
       call Write_L2B_Rayl_Meas_Classific(fileunit, &
            Rayl_Classification_QC%L2B_Rayl_Meas_Classific(i), error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    s = size(Rayl_Classification_QC%L2B_Rayl_Obs_Classific)
    IF (N_Obs_Rayleigh_to_write .gt. s) THEN
       call SetIntAuc(N_Obs_Rayleigh_to_write,str1_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
       call SetIntAuc(int(s,IntAuc), str2_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_error,"ERROR in "//&
            "Write_Rayl_Classification_QC(): invalid value for "//&
            "N_Obs_Rayleigh_to_write: "//trim(str1_IntAuc)//&
            " max. possible value is: "//&
            "size(Rayl_Classification_QC%L2B_Rayl_Obs_Classific) = "//&
            trim(str2_IntAuc))
       error_flag = error_programming
       return
    END IF  

    DO i=1,N_Obs_Rayleigh_to_write
       call Write_L2B_Rayl_Obs_Classific(fileunit, &
            Rayl_Classification_QC%L2B_Rayl_Obs_Classific(i), error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    ! 20 spare bytes
    nr_spare_bytes = 20
    call write_SpareBytes_to_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    return

998 call logmsg(log_error,"ABORTING subroutine Write_Rayl_Classification_QC")
    return

  end subroutine Write_Rayl_Classification_QC
    !  #]
  subroutine Read_Rayl_Classification_QC(fileunit, &
             Rayl_Classification_QC, error_flag)
    !  #[
    !  P. Poli     10-01-2006 Original code
    integer,           intent(in)    :: fileunit               ! input
    TYPE(Rayl_Classification_QC_Type),&
                       intent(inout) :: Rayl_Classification_QC ! input
    integer,           intent(out)   :: error_flag             ! output

    ! local variable
    integer :: i,s,nr_spare_bytes

    !---------------------------------------------
    ! reading the Rayleigh_Classification elements
    !---------------------------------------------

    error_flag = no_error
    IF (.not. associated(Rayl_Classification_QC%L2B_Rayl_Meas_Classific)) THEN
       call logmsg(log_error,"ERROR in Read_Rayl_Classification(): "//&
            "The pointer to Rayl_Classification_QC%L2B_Rayl_Meas_Classific "//&
            "seems not associated !!")
       error_flag = error_programming
       return
    END IF

    s = size(Rayl_Classification_QC%L2B_Rayl_Meas_Classific)
    DO i=1,s
       call Read_L2B_Rayl_Meas_Classific(fileunit, &
            Rayl_Classification_QC%L2B_Rayl_Meas_Classific(i), error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    ! N_Obs may turn out to be zero !
    ! so this check does not make sense anymore
    !IF (.not. associated(Rayl_Classification_QC%L2B_Rayl_Obs_Classific)) THEN
    !   call logmsg(log_error,"ERROR in Read_Rayl_Classification(): "//&
    !        "The pointer to Rayl_Classification_QC%L2B_Rayl_Obs_Classific "//&
    !        "seems not associated !!")
    !   error_flag = error_programming
    !   return
    !END IF
    
    IF (associated(Rayl_Classification_QC%L2B_Rayl_Obs_Classific)) THEN
       s = size(Rayl_Classification_QC%L2B_Rayl_Obs_Classific)
       DO i=1,s
          call Read_L2B_Rayl_Obs_Classific(fileunit, &
               Rayl_Classification_QC%L2B_Rayl_Obs_Classific(i), error_flag)
          IF (error_flag .ne. no_error) goto 998
       END DO
    END IF

    ! 20 spare bytes
    nr_spare_bytes = 20
    call read_SpareBytes_from_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Read_Rayl_Classification_QC")
    return

  end subroutine Read_Rayl_Classification_QC
    !  #]
  !  #]
  !---------------------------------------------------------------------
  ! subroutines for the Mie_Processing_QC sub-datastructure
  !---------------------------------------------------------------------
  !  #[
  subroutine Init_Obs_Mie_Bin_QC(Obs_Mie_Bin_QC)
    !  #[
    TYPE(Obs_Mie_Bin_QC_Type), intent(out) :: Obs_Mie_Bin_QC

    Obs_Mie_Bin_QC%Hlos_Error_Estimate = missing_indicator_FAdoxy
    Obs_Mie_Bin_QC%Reliability_Number  = missing_indicator_FAdoxy
    Obs_Mie_Bin_QC%flags = 0
    Obs_Mie_Bin_QC%Fitting_Amplitude   = missing_indicator_FAdoxy
    Obs_Mie_Bin_QC%Fitting_Residual    = missing_indicator_FAdoxy
    Obs_Mie_Bin_QC%Fitting_Offset      = missing_indicator_FAdoxy
    Obs_Mie_Bin_QC%Fitting_FWHM        = missing_indicator_FAdoxy
    Obs_Mie_Bin_QC%Mie_SNR             = missing_indicator_FAdoxy
    Obs_Mie_Bin_QC%Extinction          = missing_indicator_FAdoxy
    Obs_Mie_Bin_QC%Scattering_Ratio    = missing_indicator_FAdoxy
    Obs_Mie_Bin_QC%Variability_Measure = missing_indicator_FAdoxy

  end subroutine Init_Obs_Mie_Bin_QC
  !  #]
  subroutine Print_Obs_Mie_Bin_QC(Obs_Mie_Bin_QC)
    !  #[
    TYPE(Obs_Mie_Bin_QC_Type), intent(in) :: Obs_Mie_Bin_QC

    print *,"       ...%Hlos_Error_Estimate  = ",&
         Obs_Mie_Bin_QC%Hlos_Error_Estimate
    print *,"       ...%Reliability_Number   = ",&
         Obs_Mie_Bin_QC%Reliability_Number
    print *,"       ...%flags                = ",Obs_Mie_Bin_QC%flags
    print *,"       ...%Fitting_Amplitude    = ",&
         Obs_Mie_Bin_QC%Fitting_Amplitude
    print *,"       ...%Fitting_Residual     = ",&
         Obs_Mie_Bin_QC%Fitting_Residual
    print *,"       ...%Fitting_Offset       = ",Obs_Mie_Bin_QC%Fitting_Offset
    print *,"       ...%Fitting_FWHM         = ",Obs_Mie_Bin_QC%Fitting_FWHM
    print *,"       ...%Mie_SNR              = ",Obs_Mie_Bin_QC%Mie_SNR
    print *,"       ...%Extinction           = ",Obs_Mie_Bin_QC%Extinction
    print *,"       ...%Scattering_Ratio     = ",&
         Obs_Mie_Bin_QC%Scattering_Ratio
    print *,"       ...%Variability_Measure  = ",&
         Obs_Mie_Bin_QC%Variability_Measure

  end subroutine Print_Obs_Mie_Bin_QC
    !  #]
  subroutine AsciiWrite_Obs_Mie_Bin_QC(Obs_Mie_Bin_QC,&
                                       fileunit,prefix,error_flag)
    !  #[
    TYPE(Obs_Mie_Bin_QC_Type), intent(in)  :: Obs_Mie_Bin_QC
    integer,                   intent(in)  :: fileunit   ! input
    character(len=*),          intent(in)  :: prefix     ! input
    integer,                   intent(out) :: error_flag ! output

    ! local variables
    character(len=maxlen_IntAucFlags) :: str_IntAucFlags
    character(len=25)                 :: str_FAdoxy

    write(fileunit,"(a)",err=999) "REALACC(12)"

    call SetFAdoFull_or_Missing(Obs_Mie_Bin_QC%Hlos_Error_Estimate,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Hlos_Error_Estimate = "//trim(str_FAdoxy)//" [m/s]"

    call SetFAdoFull_or_Missing(Obs_Mie_Bin_QC%Reliability_Number,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Reliability_Number  = "//trim(str_FAdoxy)

    call SetIntAucFlags(Obs_Mie_Bin_QC%flags,str_IntAucFlags,error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%flags               = "//trim(str_IntAucFlags)

    call SetFAdoFull_or_Missing(Obs_Mie_Bin_QC%Fitting_Amplitude,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Fitting_Amplitude   = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(Obs_Mie_Bin_QC%Fitting_Residual,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Fitting_Residual    = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(Obs_Mie_Bin_QC%Fitting_Offset,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Fitting_Offset      = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(Obs_Mie_Bin_QC%Fitting_FWHM,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Fitting_FWHM        = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(Obs_Mie_Bin_QC%Mie_SNR,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Mie_SNR             = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(Obs_Mie_Bin_QC%Extinction,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Extinction          = "//trim(str_FAdoxy)//" [1/m]"

    call SetFAdoFull_or_Missing(Obs_Mie_Bin_QC%Scattering_Ratio,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Scattering_Ratio    = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(Obs_Mie_Bin_QC%Variability_Measure,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Variability_Measure = "//trim(str_FAdoxy)

    write(fileunit,"(a)",err=999) "ENDREALACC"

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_Obs_Mie_Bin_QC: "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Obs_Mie_Bin_QC
    !  #]
  function GetByteSize_Obs_Mie_Bin_QC () result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result

    c_bytesize = 10*c_bytelen_FAdoxy + &
                 1*c_bytelen_IntAuc + 1 ! spare byte
                 ! 11 items in total

    ! typical sizes
    ! memory size = 2+10*8   = 82 bytes
    ! file   size = 1+10*8+1 = 82 bytes
  end function GetByteSize_Obs_Mie_Bin_QC
    !  #]
  subroutine Write_Obs_Mie_Bin_QC(fileunit, &
             Obs_Mie_Bin_QC, error_flag)
    !  #[
    !  Writes Obs_Mie_Bin_QC in binary format
    !  P. Poli     09-01-2006 Original Code
    integer,                   intent(in)  :: fileunit
    TYPE(Obs_Mie_Bin_QC_Type), intent(in)  :: Obs_Mie_Bin_QC
    integer,                   intent(out) :: error_flag 

    ! local variable
    integer :: nr_spare_bytes

    !------------------------------------
    ! writing the Obs_Mie_Bin_QC elements
    !------------------------------------

    error_flag = no_error

    call write_FAdoxy_to_bin_file(fileunit, &
         Obs_Mie_Bin_QC%Hlos_Error_Estimate, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_FAdoxy_to_bin_file(fileunit, &
         Obs_Mie_Bin_QC%Reliability_Number, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_IntAuc_to_bin_file(fileunit, &
         Obs_Mie_Bin_QC%flags, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_FAdoxy_to_bin_file(fileunit, &
         Obs_Mie_Bin_QC%Fitting_Amplitude, error_flag)
    IF (error_flag .ne. no_error) goto 998

    call write_FAdoxy_to_bin_file(fileunit, &
         Obs_Mie_Bin_QC%Fitting_Residual, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_FAdoxy_to_bin_file(fileunit, &
         Obs_Mie_Bin_QC%Fitting_Offset, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_FAdoxy_to_bin_file(fileunit, &
         Obs_Mie_Bin_QC%Fitting_FWHM, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_FAdoxy_to_bin_file(fileunit, &
         Obs_Mie_Bin_QC%Mie_SNR, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_FAdoxy_to_bin_file(fileunit, &
         Obs_Mie_Bin_QC%Extinction, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_FAdoxy_to_bin_file(fileunit, &
         Obs_Mie_Bin_QC%Scattering_Ratio, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_FAdoxy_to_bin_file(fileunit, &
         Obs_Mie_Bin_QC%Variability_Measure, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 1 spare byte
    nr_spare_bytes = 1
    call write_SpareBytes_to_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998
   
    return

998 call logmsg(log_error,"ABORTING subroutine Write_Obs_Mie_Bin_QC")
    return

  end subroutine Write_Obs_Mie_Bin_QC
    !  #]
  subroutine Read_Obs_Mie_Bin_QC(fileunit, &
             Obs_Mie_Bin_QC, error_flag)
    !  #[
    !  P. Poli     10-01-2006 Original Code
    integer,                   intent(in)    :: fileunit
    TYPE(Obs_Mie_Bin_QC_Type), intent(inout) :: Obs_Mie_Bin_QC
    integer,                   intent(out)   :: error_flag 

    ! local variable
    integer :: nr_spare_bytes

    !------------------------------------
    ! reading the Obs_Mie_Bin_QC elements
    !------------------------------------

    error_flag = no_error

    call read_FAdoxy_from_bin_file(fileunit, &
         Obs_Mie_Bin_QC%Hlos_Error_Estimate, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call read_FAdoxy_from_bin_file(fileunit, &
         Obs_Mie_Bin_QC%Reliability_Number, error_flag)
    IF (error_flag .ne. no_error) goto 998

    call read_IntAuc_from_bin_file(fileunit, &
         Obs_Mie_Bin_QC%flags, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call read_FAdoxy_from_bin_file(fileunit, &
         Obs_Mie_Bin_QC%Fitting_Amplitude, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call read_FAdoxy_from_bin_file(fileunit, &
         Obs_Mie_Bin_QC%Fitting_Residual, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call read_FAdoxy_from_bin_file(fileunit, &
         Obs_Mie_Bin_QC%Fitting_Offset, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call read_FAdoxy_from_bin_file(fileunit, &
         Obs_Mie_Bin_QC%Fitting_FWHM, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call read_FAdoxy_from_bin_file(fileunit, &
         Obs_Mie_Bin_QC%Mie_SNR, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call read_FAdoxy_from_bin_file(fileunit, &
         Obs_Mie_Bin_QC%Extinction, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call read_FAdoxy_from_bin_file(fileunit, &
         Obs_Mie_Bin_QC%Scattering_Ratio, error_flag)
    IF (error_flag .ne. no_error) goto 998

    call read_FAdoxy_from_bin_file(fileunit, &
         Obs_Mie_Bin_QC%Variability_Measure, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 1 spare byte
    nr_spare_bytes = 1
    call read_SpareBytes_from_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Read_Obs_Mie_Bin_QC")
    return

  end subroutine Read_Obs_Mie_Bin_QC
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_Obs_Mie_QC(Obs_Mie_QC, N_Range_Gates0, error_flag)
    !  #[
    TYPE(Obs_Mie_QC_Type), intent(out) :: Obs_Mie_QC ! output
    integer(IntAs), intent(in)  :: N_Range_Gates0 ! input
    integer,        intent(out) :: error_flag    ! output

    ! local variable
    integer :: r, AllocStatus

    error_flag = no_error

    IF (N_Range_Gates0 .eq. 0) THEN
       ! init them to be empty
       nullify(Obs_Mie_QC%Obs_Mie_Bin_QC)
    ELSE
       allocate(Obs_Mie_QC%Obs_Mie_Bin_QC(N_Range_Gates0),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          call logmsg(log_error,"Allocation failure in "//&
                                "subroutine Init_Obs_Mie_QC()")
          error_flag = error_allocate
          return
       END IF
       
       DO r=1,N_Range_Gates0
          call Init_Obs_Mie_Bin_QC(Obs_Mie_QC%Obs_Mie_Bin_QC(r))
       END DO
    END IF

  end subroutine Init_Obs_Mie_QC
    !  #]
  subroutine Delete_Obs_Mie_QC(Obs_Mie_QC)
    !  #[
    TYPE(Obs_Mie_QC_Type), intent(inout) :: Obs_Mie_QC

    IF (associated(Obs_Mie_QC%Obs_Mie_Bin_QC)) &
         deallocate(Obs_Mie_QC%Obs_Mie_Bin_QC)

  end subroutine Delete_Obs_Mie_QC
    !  #]
  subroutine Print_Obs_Mie_QC(Obs_Mie_QC)
    !  #[
    TYPE(Obs_Mie_QC_Type), intent(in) :: Obs_Mie_QC

    integer :: r,s

    IF (associated(Obs_Mie_QC%Obs_Mie_Bin_QC)) THEN
       ! get the size of the array
       s = size(Obs_Mie_QC%Obs_Mie_Bin_QC)
       print "(a,i2,a)","    ==> size(Obs_Mie_QC%Obs_Mie_Bin_QC) = ",s,&
                        " [printing first and last one]"
       DO r=1,s,s-1
          print "('        Obs_Mie_QC%Obs_Mie_Bin_QC(',i2.2,')%...')",r
          call Print_Obs_Mie_Bin_QC(Obs_Mie_QC%Obs_Mie_Bin_QC(r))
       END DO
    ELSE
       print *,"        Obs_Mie_QC%Obs_Mie_Bin_QC(:) [seems not associated]"
    END IF

  end subroutine Print_Obs_Mie_QC
    !  #]
  subroutine AsciiWrite_Obs_Mie_QC(Obs_Mie_QC,fileunit,&
                                   prefix,PrintAll,error_flag)
    !  #[
    TYPE(Obs_Mie_QC_Type), intent(in)  :: Obs_Mie_QC
    integer,               intent(in)  :: fileunit   ! input
    character(len=*),      intent(in)  :: prefix     ! input
    logical,               intent(in)  :: PrintAll   ! input
    integer,               intent(out) :: error_flag ! output

    ! local variables
    integer          :: r,s,step
    character(len=2) :: str_cntr

    IF (associated(Obs_Mie_QC%Obs_Mie_Bin_QC)) THEN
       ! get the size of the array
       s = size(Obs_Mie_QC%Obs_Mie_Bin_QC)
       IF (PrintAll .or. (s .le. 1) ) THEN
          step=1
       ELSE
          step = s-1
       END IF
       DO r=1,s,step
          write(str_cntr,"(i2.2)",err=998) r
          call AsciiWrite_Obs_Mie_Bin_QC(Obs_Mie_QC%Obs_Mie_Bin_QC(r),&
               fileunit,trim(prefix)//"%Obs_Mie_Bin_QC("//trim(str_cntr)//")",&
               error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%Obs_Mie_Bin_QC(:) [seems not associated]"
    END IF

    return

998 call logmsg(log_error,"conversion error in AsciiWrite_Obs_Mie_QC: "//&
         "could not convert the counter r to a string ..")
    error_flag = error_writing_file
    return

999 call logmsg(log_error,"ERROR in AsciiWrite_Obs_Mie_QC: "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Obs_Mie_QC
    !  #]
  function GetByteSize_Obs_Mie_QC(Obs_Mie_QC) result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result
    TYPE(Obs_Mie_QC_Type), intent(in) :: Obs_Mie_QC

    c_bytesize = 1 ! spare_byte
    IF (associated(Obs_Mie_QC%Obs_Mie_Bin_QC)) THEN
       c_bytesize = c_bytesize + &
            int(size(Obs_Mie_QC%Obs_Mie_Bin_QC)*&
                GetByteSize_Obs_Mie_Bin_QC(),IntAl)
    END IF

    ! typical sizes
    ! memory size =   24*82 = 1968 bytes
    ! file   size = 1+24*82 = 1969 bytes
  end function GetByteSize_Obs_Mie_QC
    !  #]  
  subroutine Write_Obs_Mie_QC(fileunit, Obs_Mie_QC, error_flag)
    !  #[
    !  Writes Obs_Mie_QC in binary format
    !  P. Poli     09-01-2006 Original Code
    integer,               intent(in)  :: fileunit
    TYPE(Obs_Mie_QC_Type), intent(in)  :: Obs_Mie_QC
    integer,               intent(out) :: error_flag 

    ! local variable
    integer :: s, i, nr_spare_bytes

    !--------------------------------------------
    ! writing the Obs_Mie_QC elements
    !--------------------------------------------

    error_flag = no_error

    s=size(Obs_Mie_QC%Obs_Mie_Bin_QC)
    DO i=1,s
       call Write_Obs_Mie_Bin_QC(fileunit, &
            Obs_Mie_QC%Obs_Mie_Bin_QC(i), error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    ! 1 spare byte
    nr_spare_bytes = 1
    call write_SpareBytes_to_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Write_Obs_Mie_QC")
    return

  end subroutine Write_Obs_Mie_QC
    !  #]
  subroutine Read_Obs_Mie_QC(fileunit, Obs_Mie_QC, error_flag)
    !  #[
    !  P. Poli     10-01-2006 Original Code
    integer,               intent(in)    :: fileunit
    TYPE(Obs_Mie_QC_Type), intent(inout) :: Obs_Mie_QC
    integer,               intent(out)   :: error_flag 

    ! local variable
    integer :: s, i, nr_spare_bytes

    !--------------------------------------------
    ! writing the Obs_Mie_QC elements
    !--------------------------------------------

    error_flag = no_error

    s=size(Obs_Mie_QC%Obs_Mie_Bin_QC)
    DO i=1,s
       call Read_Obs_Mie_Bin_QC(fileunit, &
            Obs_Mie_QC%Obs_Mie_Bin_QC(i), error_flag)
       IF (error_flag .ne. no_error) goto 998
    ENDDO

    ! 1 spare byte
    nr_spare_bytes = 1
    call read_SpareBytes_from_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Read_Obs_Mie_QC")
    return

  end subroutine Read_Obs_Mie_QC
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_Mie_Processing_QC(Mie_Processing_QC, N_Obs_Mie_Max0, &
                                    N_Range_Gates0, error_flag)
    !  #[
    TYPE(Mie_Processing_QC_Type), intent(out) :: Mie_Processing_QC ! output
    integer(IntAs), intent(in)  :: N_Obs_Mie_Max0     ! input
    integer(IntAs), intent(in)  :: N_Range_Gates0 ! input
    integer,        intent(out) :: error_flag    ! output

    ! local variable
    integer :: o, AllocStatus

    error_flag = no_error

    IF (N_Obs_Mie_Max0 .eq. 0) THEN
       ! init them to be empty
       nullify(Mie_Processing_QC%Obs_Mie_QC)
    ELSE
       allocate(Mie_Processing_QC%Obs_Mie_QC(N_Obs_Mie_Max0),STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          call logmsg(log_error,"Allocation failure in "//&
                                "subroutine Init_Mie_Processing_QC()")
          error_flag = error_allocate
          return
       END IF
       DO o=1,N_Obs_Mie_Max0
          call Init_Obs_Mie_QC(Mie_Processing_QC%Obs_Mie_QC(o),&
                               N_Range_Gates0,error_flag)
       END DO
    END IF
    
    Mie_Processing_QC%Mie_Background_High = 0

  end subroutine Init_Mie_Processing_QC
    !  #]
  subroutine Delete_Mie_Processing_QC(Mie_Processing_QC)
    !  #[
    TYPE(Mie_Processing_QC_Type), intent(inout) :: Mie_Processing_QC

    ! local variable
    integer :: O

    IF (associated(Mie_Processing_QC%Obs_Mie_QC)) THEN
       DO O=1,size(Mie_Processing_QC%Obs_Mie_QC)
          call Delete_Obs_Mie_QC(Mie_Processing_QC%Obs_Mie_QC(O))
       END DO
       deallocate(Mie_Processing_QC%Obs_Mie_QC)
    END IF

  end subroutine Delete_Mie_Processing_QC
    !  #]
  subroutine Print_Mie_Processing_QC(Mie_Processing_QC)
    !  #[
    TYPE(Mie_Processing_QC_Type), intent(in) :: Mie_Processing_QC

    ! local variable
    integer :: o, s

    IF (associated(Mie_Processing_QC%Obs_Mie_QC)) THEN
       ! get the size of the array
       s = size(Mie_Processing_QC%Obs_Mie_QC)
       print "(a,i2,a)"," ==> size(Mie_Processing_QC%Obs_Mie_QC) = ",s,&
                        " [printing first and last one]"
       DO o=1,s,s-1
          print "('    Mie_Processing_QC%Obs_Mie_QC(',i2.2,')%...')",o
          call Print_Obs_Mie_QC(Mie_Processing_QC%Obs_Mie_QC(o))
       END DO
    ELSE
       print *," Mie_Processing_QC%Obs_Mie_QC(:) [seems not associated]"
    END IF

    print *,"Mie_Processing_QC%Mie_Background_High = ",&
         Mie_Processing_QC%Mie_Background_High

  end subroutine Print_Mie_Processing_QC
    !  #]
  subroutine AsciiWrite_Mie_Processing_QC(Mie_Processing_QC,fileunit,&
                                          prefix,PrintAll,error_flag)
    !  #[
    TYPE(Mie_Processing_QC_Type), intent(in)  :: Mie_Processing_QC
    integer,                      intent(in)  :: fileunit   ! input
    character(len=*),             intent(in)  :: prefix     ! input
    logical,                      intent(in)  :: PrintAll   ! input
    integer,                      intent(out) :: error_flag ! output

    ! local variables
    integer                      :: o, s, step
    character(len=2)             :: str_cntr
    character(len=maxlen_IntAuc) :: str_IntAuc

    IF (associated(Mie_Processing_QC%Obs_Mie_QC)) THEN
       ! get the size of the array
       s = size(Mie_Processing_QC%Obs_Mie_QC)
       IF (PrintAll .or. (s .le. 1) ) THEN
          step=1
       ELSE
          step=s-1
       END IF
       DO o=1,s,step
          write(str_cntr,"(i2.2)",err=998) o
          call AsciiWrite_Obs_Mie_QC(&
               Mie_Processing_QC%Obs_Mie_QC(o),fileunit,&
               trim(prefix)//"%Obs_Mie_QC("//trim(str_cntr)//")",&
               PrintAll,error_flag)
       END DO
    ELSE
     write(fileunit,"(a)",err=999) trim(prefix)//&
          "%Obs_Mie_QC(:) [seems not associated]"
    END IF

    write(fileunit,"(a)",err=999) "REALACC(12)"

    call SetIntAuc(Mie_Processing_QC%Mie_Background_High,str_IntAuc,error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Mie_Background_High = "//trim(str_IntAuc)

    write(fileunit,"(a)",err=999) "ENDREALACC"

    return

998 call logmsg(log_error,"conversion error in "//&
         "AsciiWrite_Mie_Processing_QC: "//&
         "could not convert the counter o to a string ..")
    error_flag = error_writing_file
    return

999 call logmsg(log_error,"ERROR in AsciiWrite_Mie_Processing_QC: "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Mie_Processing_QC
    !  #]
  function GetByteSize_Mie_Processing_QC(Mie_Processing_QC, &
                                         N_Obs_Mie_to_write) result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result
    TYPE(Mie_Processing_QC_Type), intent(in) :: Mie_Processing_QC
    integer(IntAuc),              intent(in) :: N_Obs_Mie_to_write

    ! local variable
    integer :: s, error_flag
    character(len=maxlen_IntAuc) :: str1_IntAuc, str2_IntAuc

    c_bytesize = 1*c_bytelen_IntAuc + 20 ! spare_bytes

    IF (associated(Mie_Processing_QC%Obs_Mie_QC)) THEN
       s = size(Mie_Processing_QC%Obs_Mie_QC)

       IF (N_Obs_Mie_to_write .gt. s) THEN
          call SetIntAuc(N_Obs_Mie_to_write,str1_IntAuc,error_flag)
          IF (error_flag .ne. no_error) str1_IntAuc = "[er]"
          call SetIntAuc(int(s,IntAuc), str2_IntAuc,error_flag)
          IF (error_flag .ne. no_error) str2_IntAuc = "[er]"
          call logmsg(log_warn,"WARNING in "//&
               "GetByteSize_Mie_Processing_QC(): invalid value for "//&
               "N_Obs_Mie_to_write: "//trim(str1_IntAuc)//&
               " max. possible value is: "//&
               "size(Mie_Processing_QC%Obs_Mie_QC) = "//&
               trim(str2_IntAuc)//" GetByteSize result will be useless...")
       END IF

       c_bytesize = c_bytesize + N_Obs_Mie_to_write *&
            GetByteSize_Obs_Mie_QC(Mie_Processing_QC%Obs_Mie_QC(1))
    END IF

    ! typical sizes
    ! memory size = 2   +5*1968 = 9842 bytes
    ! file   size = 1+20+5*1969 = 9866 bytes
  end function GetByteSize_Mie_Processing_QC
    !  #]
  subroutine Write_Mie_Processing_QC(fileunit, &
             Mie_Processing_QC, N_Obs_Mie_to_write, error_flag)
    !  #[
    !  Writes the Mie_Processing_QC in binary format
    !  J. de Kloe  03-01-2006 Original Code
    !  P. Poli     10-01-2006 Removed placeholders. Put actual writing in place
    !                         From L2B QC tables v0.2 03-01-2005
    
    integer,                      intent(in)  :: fileunit           ! input
    TYPE(Mie_Processing_QC_Type), intent(in)  :: Mie_Processing_QC  ! input
    integer(IntAuc),              intent(in)  :: N_Obs_Mie_to_write ! input
    integer,                      intent(out) :: error_flag         ! output

    ! local variables
    integer :: i, s, nr_spare_bytes
    character(len=maxlen_IntAuc) :: str1_IntAuc, str2_IntAuc

    !---------------------------------------
    ! writing the Mie_Processing_QC elements
    !---------------------------------------

    error_flag = no_error
    IF (.not. associated(Mie_Processing_QC%Obs_Mie_QC)) THEN
       call logmsg(log_error,"ERROR in Write_Mie_Processing_QC(): "//&
            "The pointer to this Mie_Processing_QC%Obs_Mie_QC "//&
            "seems not associated !!")
       error_flag = error_programming
       return
    END IF
    
    s = size(Mie_Processing_QC%Obs_Mie_QC)
    IF (N_Obs_Mie_to_write .gt. s) THEN
       call SetIntAuc(N_Obs_Mie_to_write,str1_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
       call SetIntAuc(int(s,IntAuc), str2_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_error,"ERROR in "//&
            "Write_Mie_Processing_QC(): invalid value for "//&
            "N_Obs_Mie_to_write: "//trim(str1_IntAuc)//&
            " max. possible value is: "//&
            "size(Mie_Processing_QC%Obs_Mie_QC) = "//&
            trim(str2_IntAuc))
       error_flag = error_programming
       return
    END IF

    DO i=1,N_Obs_Mie_to_write
       call Write_Obs_Mie_QC(fileunit, &
            Mie_Processing_QC%Obs_Mie_QC(i), error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    call write_IntAuc_to_bin_file(fileunit, &
         Mie_Processing_QC%Mie_Background_High, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! 20 spare bytes
    nr_spare_bytes = 20
    call write_SpareBytes_to_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Write_Mie_Processing_QC")
    return

  end subroutine Write_Mie_Processing_QC
    !  #]
  subroutine Read_Mie_Processing_QC(fileunit, &
             Mie_Processing_QC, error_flag)
    !  #[
    !  P. Poli     10-01-2006 Original code
    integer,                      intent(in)   :: fileunit
    TYPE(Mie_Processing_QC_Type), intent(inout):: Mie_Processing_QC
    integer,                      intent(out)  :: error_flag

    ! local variable
    integer :: i,s,nr_spare_bytes

    !---------------------------------------
    ! reading the Mie_Processing_QC elements
    !---------------------------------------

    error_flag = no_error

    ! N_Obs may turn out to be zero !
    ! so this check does not make sense anymore
    !IF (.not. associated(Mie_Processing_QC%Obs_Mie_QC)) THEN
    !   call logmsg(log_error,"ERROR in Read_Mie_Processing_QC(): "//&
    !        "The pointer to this Mie_Processing_QC%Obs_Mie_QC "//&
    !        "seems not associated !!")
    !   error_flag = error_programming
    !   return
    !END IF

    IF (associated(Mie_Processing_QC%Obs_Mie_QC)) THEN    
       s = size(Mie_Processing_QC%Obs_Mie_QC)
       DO i=1,s
          call Read_Obs_Mie_QC(fileunit, &
               Mie_Processing_QC%Obs_Mie_QC(i), error_flag)
          IF (error_flag .ne. no_error) goto 998
       END DO
    END IF

    call read_IntAuc_from_bin_file(fileunit, &
         Mie_Processing_QC%Mie_Background_High, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! 20 spare bytes
    nr_spare_bytes = 20
    call read_SpareBytes_from_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    return

998 call logmsg(log_error,"ABORTING subroutine Read_Mie_Processing_QC")
    return

  end subroutine Read_Mie_Processing_QC
    !  #]
  !  #]
  !---------------------------------------------------------------------
  ! subroutines for the Rayleigh_Processing_QC sub-datastructure
  !---------------------------------------------------------------------
  !  #[
  subroutine Init_Obs_Rayleigh_Bin_QC(Obs_Rayleigh_Bin_QC)
    !  #[
    TYPE(Obs_Rayleigh_Bin_QC_Type), intent(out) :: Obs_Rayleigh_Bin_QC

    Obs_Rayleigh_Bin_QC%Hlos_Error_Estimate     = missing_indicator_FAdoxy
    Obs_Rayleigh_Bin_QC%flags1                  = 0
    Obs_Rayleigh_Bin_QC%flags2                  = 0
    Obs_Rayleigh_Bin_QC%flags3                  = 0
    Obs_Rayleigh_Bin_QC%Extinction              = missing_indicator_FAdoxy
    Obs_Rayleigh_Bin_QC%Scattering_Ratio        = missing_indicator_FAdoxy
    Obs_Rayleigh_Bin_QC%Variability_Measure     = missing_indicator_FAdoxy
    Obs_Rayleigh_Bin_QC%Extinction_Method       = Extinction_Meth_Undefined
    Obs_Rayleigh_Bin_QC%Scattering_Ratio_Method = Scat_Ratio_Meth_Undefined

  end subroutine Init_Obs_Rayleigh_Bin_QC
  !  #]
  subroutine Print_Obs_Rayleigh_Bin_QC(Obs_Rayleigh_Bin_QC)
    !  #[
    TYPE(Obs_Rayleigh_Bin_QC_Type), intent(in) :: Obs_Rayleigh_Bin_QC

    print *,"       ...%Hlos_Error_Estimate     = ",&
         Obs_Rayleigh_Bin_QC%Hlos_Error_Estimate
    print *,"       ...%flags1                  = ",&
         Obs_Rayleigh_Bin_QC%flags1
    print *,"       ...%flags2                  = ",&
         Obs_Rayleigh_Bin_QC%flags2
    print *,"       ...%flags3                  = ",&
         Obs_Rayleigh_Bin_QC%flags3
    print *,"       ...%Extinction              = ",&
         Obs_Rayleigh_Bin_QC%Extinction
    print *,"       ...%Scattering_Ratio        = ",&
         Obs_Rayleigh_Bin_QC%Scattering_Ratio
    print *,"       ...%Variability_Measure     = ",&
         Obs_Rayleigh_Bin_QC%Variability_Measure
    print *,"       ...%Extinction_Method       = ",&
         Obs_Rayleigh_Bin_QC%Extinction_Method
    print *,"       ...%Scattering_Ratio_Method = ",&
         Obs_Rayleigh_Bin_QC%Scattering_Ratio_Method

  end subroutine Print_Obs_Rayleigh_Bin_QC
    !  #]
  subroutine AsciiWrite_Obs_Rayleigh_Bin_QC(Obs_Rayleigh_Bin_QC,fileunit,&
                                            prefix,error_flag)
    !  #[
    TYPE(Obs_Rayleigh_Bin_QC_Type), intent(in)  :: Obs_Rayleigh_Bin_QC
    integer,                        intent(in)  :: fileunit   ! input
    character(len=*),               intent(in)  :: prefix     ! input
    integer,                        intent(out) :: error_flag ! output

   ! local variables
    character(len=maxlen_IntAucFlags) :: str_IntAucFlags
    character(len=maxlen_IntAuc)      :: str_IntAuc
    character(len=25)                 :: str_FAdoxy

    write(fileunit,"(a)",err=999) "REALACC(12)"
 
    call SetFAdoFull_or_Missing(Obs_Rayleigh_Bin_QC%Hlos_Error_Estimate,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Hlos_Error_Estimate     = "//&
         trim(str_FAdoxy)//" [m/s]"

    call SetIntAucFlags(Obs_Rayleigh_Bin_QC%flags1,str_IntAucFlags,error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%flags1                  = "//trim(str_IntAucFlags)

    call SetIntAucFlags(Obs_Rayleigh_Bin_QC%flags2,str_IntAucFlags,error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%flags2                  = "//trim(str_IntAucFlags)

    call SetIntAucFlags(Obs_Rayleigh_Bin_QC%flags3,str_IntAucFlags,error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%flags3                  = "//trim(str_IntAucFlags)

    call SetFAdoFull_or_Missing(Obs_Rayleigh_Bin_QC%Extinction,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Extinction              = "//&
         trim(str_FAdoxy)//" [1/m]"

    call SetFAdoFull_or_Missing(Obs_Rayleigh_Bin_QC%Scattering_Ratio,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Scattering_Ratio        = "//trim(str_FAdoxy)

    call SetFAdoFull_or_Missing(Obs_Rayleigh_Bin_QC%Variability_Measure,&
                                str_FAdoxy,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Variability_Measure     = "//trim(str_FAdoxy)

    if ( Obs_Rayleigh_Bin_QC%Extinction_Method .eq. &
         Extinction_Meth_Undefined ) then
       str_IntAuc = "[mi]"
    else
       call SetIntAuc(Obs_Rayleigh_Bin_QC%Extinction_Method,&
                      str_IntAuc,error_flag)
       if (error_flag .ne. no_error) return
    end if
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Extinction_Method       = "//trim(str_IntAuc)

    if ( Obs_Rayleigh_Bin_QC%Scattering_Ratio_Method .eq. &
         Scat_Ratio_Meth_Undefined ) then
       str_IntAuc = "[mi]"
    else
       call SetIntAuc(Obs_Rayleigh_Bin_QC%Scattering_Ratio_Method,&
                      str_IntAuc,error_flag)
       if (error_flag .ne. no_error) return
    end if
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Scattering_Ratio_Method = "//trim(str_IntAuc)

    write(fileunit,"(a)",err=999) "ENDREALACC"

    return

999 call logmsg(log_error,"ERROR in AsciiWrite_Obs_Rayleigh_Bin_QC: "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Obs_Rayleigh_Bin_QC
    !  #]
  function GetByteSize_Obs_Rayleigh_Bin_QC () result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result

    c_bytesize = 4*c_bytelen_FAdoxy + &
                 5*c_bytelen_IntAuc + 1 ! spare byte
                 ! 9 items in total

    ! typical sizes
    ! memory size =   5*2+4*8 = 42 bytes
    ! file   size = 1+5*1+4*8 = 38 bytes
  end function GetByteSize_Obs_Rayleigh_Bin_QC
    !  #]
  subroutine Write_Obs_Rayleigh_Bin_QC(fileunit, &
             Obs_Rayleigh_Bin_QC, error_flag)
    !  #[
    !  Writes Obs_Rayleigh_Bin_QC in binary format
    !  P. Poli     09-01-2006 Original Code
    integer,                        intent(in)  :: fileunit
    TYPE(Obs_Rayleigh_Bin_QC_Type), intent(in)  :: Obs_Rayleigh_Bin_QC
    integer,                        intent(out) :: error_flag 

    ! local variable
    integer :: nr_spare_bytes

    !------------------------------------
    ! writing the Obs_Rayleigh_Bin_QC elements
    !------------------------------------

    error_flag = no_error

    call write_FAdoxy_to_bin_file(fileunit, &
         Obs_Rayleigh_Bin_QC%Hlos_Error_Estimate, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_IntAuc_to_bin_file(fileunit, &
         Obs_Rayleigh_Bin_QC%flags1, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_IntAuc_to_bin_file(fileunit, &
         Obs_Rayleigh_Bin_QC%flags2, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_IntAuc_to_bin_file(fileunit, &
         Obs_Rayleigh_Bin_QC%flags3, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_FAdoxy_to_bin_file(fileunit, &
         Obs_Rayleigh_Bin_QC%Extinction, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_FAdoxy_to_bin_file(fileunit, &
         Obs_Rayleigh_Bin_QC%Scattering_Ratio, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_FAdoxy_to_bin_file(fileunit, &
         Obs_Rayleigh_Bin_QC%Variability_Measure, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_IntAuc_to_bin_file(fileunit, &
         Obs_Rayleigh_Bin_QC%Extinction_Method, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    call write_IntAuc_to_bin_file(fileunit, &
         Obs_Rayleigh_Bin_QC%Scattering_Ratio_Method, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! 1 spare byte
    nr_spare_bytes = 1
    call write_SpareBytes_to_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Write_Obs_Rayleigh_Bin_QC")
    return

  end subroutine Write_Obs_Rayleigh_Bin_QC
    !  #]
  subroutine Read_Obs_Rayleigh_Bin_QC(fileunit, &
             Obs_Rayleigh_Bin_QC, error_flag)
    !  #[
    !  P. Poli     10-01-2006 Original Code
    integer,                        intent(in)    :: fileunit
    TYPE(Obs_Rayleigh_Bin_QC_Type), intent(inout) :: Obs_Rayleigh_Bin_QC
    integer,                        intent(out)   :: error_flag 

    ! local variable
    integer :: nr_spare_bytes

    !------------------------------------
    ! reading the Obs_Rayleigh_Bin_QC elements
    !------------------------------------

    error_flag = no_error

    call read_FAdoxy_from_bin_file(fileunit, &
         Obs_Rayleigh_Bin_QC%Hlos_Error_Estimate, error_flag)
    IF (error_flag .ne. no_error) goto 998

    call read_IntAuc_from_bin_file(fileunit, &
         Obs_Rayleigh_Bin_QC%flags1, error_flag)
    IF (error_flag .ne. no_error) goto 998

    call read_IntAuc_from_bin_file(fileunit, &
         Obs_Rayleigh_Bin_QC%flags2, error_flag)
    IF (error_flag .ne. no_error) goto 998

    call read_IntAuc_from_bin_file(fileunit, &
         Obs_Rayleigh_Bin_QC%flags3, error_flag)
    IF (error_flag .ne. no_error) goto 998

    call read_FAdoxy_from_bin_file(fileunit, &
         Obs_Rayleigh_Bin_QC%Extinction, error_flag)
    IF (error_flag .ne. no_error) goto 998

    call read_FAdoxy_from_bin_file(fileunit, &
         Obs_Rayleigh_Bin_QC%Scattering_Ratio, error_flag)
    IF (error_flag .ne. no_error) goto 998

    call read_FAdoxy_from_bin_file(fileunit, &
         Obs_Rayleigh_Bin_QC%Variability_Measure, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! integer(IntAuc) :: Extinction_Method
    call read_IntAuc_from_bin_file(fileunit, &
         Obs_Rayleigh_Bin_QC%Extinction_Method, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! integer(IntAuc) :: Scattering_Ratio_Method
    call read_IntAuc_from_bin_file(fileunit, &
         Obs_Rayleigh_Bin_QC%Scattering_Ratio_Method, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 1 spare byte
    nr_spare_bytes = 1
    call read_SpareBytes_from_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Read_Obs_Rayleigh_Bin_QC")
    return

  end subroutine Read_Obs_Rayleigh_Bin_QC
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_Obs_Rayleigh_QC(Obs_Rayleigh_QC, N_Range_Gates0, error_flag)
    !  #[
    TYPE(Obs_Rayleigh_QC_Type), intent(out) :: Obs_Rayleigh_QC ! output
    integer(IntAs), intent(in)  :: N_Range_Gates0 ! input
    integer,        intent(out) :: error_flag    ! output

    ! local variable
    integer :: r, AllocStatus

    error_flag = no_error

    IF (N_Range_Gates0 .eq. 0) THEN
       ! init them to be empty
       nullify(Obs_Rayleigh_QC%Obs_Rayleigh_Bin_QC)
    ELSE
       allocate(Obs_Rayleigh_QC%Obs_Rayleigh_Bin_QC(N_Range_Gates0),&
                STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          call logmsg(log_error,"Allocation failure in "//&
                                "subroutine Init_Obs_Rayleigh_QC()")
          error_flag = error_allocate
          return
       END IF
       
       DO r=1,N_Range_Gates0
          call Init_Obs_Rayleigh_Bin_QC(Obs_Rayleigh_QC%Obs_Rayleigh_Bin_QC(r))
       END DO
    END IF

  end subroutine Init_Obs_Rayleigh_QC
    !  #]
  subroutine Delete_Obs_Rayleigh_QC(Obs_Rayleigh_QC)
    !  #[
    TYPE(Obs_Rayleigh_QC_Type), intent(inout) :: Obs_Rayleigh_QC

    IF (associated(Obs_Rayleigh_QC%Obs_Rayleigh_Bin_QC)) &
         deallocate(Obs_Rayleigh_QC%Obs_Rayleigh_Bin_QC)

  end subroutine Delete_Obs_Rayleigh_QC
    !  #]
  subroutine Print_Obs_Rayleigh_QC(Obs_Rayleigh_QC)
    !  #[
    TYPE(Obs_Rayleigh_QC_Type), intent(in) :: Obs_Rayleigh_QC

    integer :: r,s

    IF (associated(Obs_Rayleigh_QC%Obs_Rayleigh_Bin_QC)) THEN
       ! get the size of the array
       s = size(Obs_Rayleigh_QC%Obs_Rayleigh_Bin_QC)
       print "(a,i2,a)","    ==> size(Obs_Rayleigh_QC%"//&
            "Obs_Rayleigh_Bin_QC) = ",s,&
            " [printing first and last one]"
       DO r=1,s,s-1
          print "('        Obs_Rayleigh_QC%"//&
               "Obs_Rayleigh_Bin_QC(',i2.2,')%...')",r
          call Print_Obs_Rayleigh_Bin_QC(&
               Obs_Rayleigh_QC%Obs_Rayleigh_Bin_QC(r))
       END DO
    ELSE
       print *,"        Obs_Rayleigh_QC%"//&
       "Obs_Rayleigh_Bin_QC(:) [seems not associated]"
    END IF

  end subroutine Print_Obs_Rayleigh_QC
    !  #]
  subroutine AsciiWrite_Obs_Rayleigh_QC(Obs_Rayleigh_QC,fileunit,&
                                        prefix,PrintAll,error_flag)
    !  #[
    TYPE(Obs_Rayleigh_QC_Type), intent(in) :: Obs_Rayleigh_QC
    integer,               intent(in)  :: fileunit   ! input
    character(len=*),      intent(in)  :: prefix     ! input
    logical,               intent(in)  :: PrintAll   ! input
    integer,               intent(out) :: error_flag ! output

    ! local variables
    integer          :: r,s,step
    character(len=2) :: str_cntr

    IF (associated(Obs_Rayleigh_QC%Obs_Rayleigh_Bin_QC)) THEN
       ! get the size of the array
       s = size(Obs_Rayleigh_QC%Obs_Rayleigh_Bin_QC)
       IF (PrintAll .or. (s .le. 1) ) THEN
          step=1
       ELSE
          step = s-1
       END IF
       DO r=1,s,step
          write(str_cntr,"(i2.2)",err=998) r
          call AsciiWrite_Obs_Rayleigh_Bin_QC(&
               Obs_Rayleigh_QC%Obs_Rayleigh_Bin_QC(r),fileunit,&
               trim(prefix)//"%Obs_Rayleigh_Bin_QC("//trim(str_cntr)//")",&
               error_flag)
          IF (error_flag .ne. no_error) return
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
            "%Obs_Rayleigh_Bin_QC(:) [seems not associated]"
    END IF

    return

998 call logmsg(log_error,"conversion error in AsciiWrite_Obs_Rayleigh_QC: "//&
         "could not convert the counter r to a string ..")
    error_flag = error_writing_file
    return

999 call logmsg(log_error,"ERROR in AsciiWrite_Obs_Rayleigh_QC: "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Obs_Rayleigh_QC
    !  #]
  function GetByteSize_Obs_Rayleigh_QC(Obs_Rayleigh_QC) result(c_bytesize)
    !  #[
    integer(IntAl) :: c_bytesize ! result
    TYPE(Obs_Rayleigh_QC_Type), intent(in) :: Obs_Rayleigh_QC

    c_bytesize = 1 ! spare_byte
    IF (associated(Obs_Rayleigh_QC%Obs_Rayleigh_Bin_QC)) THEN
       c_bytesize = c_bytesize + &
            int(size(Obs_Rayleigh_QC%Obs_Rayleigh_Bin_QC)*&
                GetByteSize_Obs_Rayleigh_Bin_QC(),IntAl)
    END IF

    ! typical sizes
    ! memory size =   24*42 = 1009 bytes
    ! file   size = 1+24*38 =  913 bytes
  end function GetByteSize_Obs_Rayleigh_QC
    !  #]
  subroutine Write_Obs_Rayleigh_QC(fileunit, Obs_Rayleigh_QC, error_flag)
    !  #[
    !  Writes Obs_Rayleigh_QC in binary format
    !  P. Poli     09-01-2006 Original Code
    integer,                    intent(in)  :: fileunit
    TYPE(Obs_Rayleigh_QC_Type), intent(in)  :: Obs_Rayleigh_QC
    integer,                    intent(out) :: error_flag 

    ! local variable
    integer :: i, s, nr_spare_bytes

    !--------------------------------------------
    ! writing the Obs_Rayleigh_QC elements
    !--------------------------------------------

    error_flag = no_error

    s = size(Obs_Rayleigh_QC%Obs_Rayleigh_Bin_QC)
    DO i=1,s
       call Write_Obs_Rayleigh_Bin_QC(fileunit, &
            Obs_Rayleigh_QC%Obs_Rayleigh_Bin_QC(i), error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    ! 1 spare byte
    nr_spare_bytes = 1
    call write_SpareBytes_to_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Write_Obs_Rayleigh_QC")
    return

  end subroutine Write_Obs_Rayleigh_QC
    !  #]
  subroutine Read_Obs_Rayleigh_QC(fileunit, &
             Obs_Rayleigh_QC, error_flag)
    !  #[
    !  P. Poli     10-01-2006 Original Code
    integer,                    intent(in)  :: fileunit
    TYPE(Obs_Rayleigh_QC_Type), intent(inout) :: Obs_Rayleigh_QC
    integer,                    intent(out) :: error_flag 

    ! local variable
    integer :: s, i, nr_spare_bytes

    !--------------------------------------------
    ! reading the Obs_Rayleigh_QC elements
    !--------------------------------------------

    error_flag = no_error

    s=size(Obs_Rayleigh_QC%Obs_Rayleigh_Bin_QC)
    DO i=1,s
       call Read_Obs_Rayleigh_Bin_QC(fileunit, &
            Obs_Rayleigh_QC%Obs_Rayleigh_Bin_QC(i), error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    ! 1 spare byte
    nr_spare_bytes = 1
    call read_SpareBytes_from_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Read_Obs_Rayleigh_QC")
    return

  end subroutine Read_Obs_Rayleigh_QC
    !  #]
  !---------------------------------------------------------------------
  subroutine Init_Rayleigh_Processing_QC(Rayleigh_Processing_QC, &
               N_Obs_Rayleigh_Max0, N_Range_Gates0, error_flag)
    !  #[
    TYPE(Rayleigh_Processing_QC_Type), &
         intent(out) :: Rayleigh_Processing_QC         ! output
    integer(IntAs), intent(in)  :: N_Obs_Rayleigh_Max0 ! input
    integer(IntAs), intent(in)  :: N_Range_Gates0      ! input
    integer,        intent(out) :: error_flag          ! output

    ! local variable
    integer :: o, AllocStatus

    error_flag = no_error

    IF (N_Obs_Rayleigh_Max0 .eq. 0) THEN
       ! init them to be empty
       nullify(Rayleigh_Processing_QC%Obs_Rayleigh_QC)
    ELSE
       allocate(Rayleigh_Processing_QC%Obs_Rayleigh_QC(N_Obs_Rayleigh_Max0),&
                STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          call logmsg(log_error,"Allocation failure in subroutine "//&
               "Init_Rayleigh_Processing_QC()")
          error_flag = error_allocate
          return
       END IF
       DO o=1,N_Obs_Rayleigh_Max0
          call Init_Obs_Rayleigh_QC(Rayleigh_Processing_QC%Obs_Rayleigh_QC(o),&
                               N_Range_Gates0,error_flag)
       END DO
    END IF
    
    Rayleigh_Processing_QC%Rayleigh_Background_High = 0_IntAuc

  end subroutine Init_Rayleigh_Processing_QC
    !  #]
  subroutine Delete_Rayleigh_Processing_QC(Rayleigh_Processing_QC)
    !  #[
    TYPE(Rayleigh_Processing_QC_Type), intent(inout) :: Rayleigh_Processing_QC

    ! local variable
    integer :: o

    IF (associated(Rayleigh_Processing_QC%Obs_Rayleigh_QC)) THEN
       DO o=1,size(Rayleigh_Processing_QC%Obs_Rayleigh_QC)
          call Delete_Obs_Rayleigh_QC(&
               Rayleigh_Processing_QC%Obs_Rayleigh_QC(o))
       END DO
       deallocate(Rayleigh_Processing_QC%Obs_Rayleigh_QC)
    END IF

  end subroutine Delete_Rayleigh_Processing_QC
    !  #]
  subroutine Print_Rayleigh_Processing_QC(Rayleigh_Processing_QC)
    !  #[
    TYPE(Rayleigh_Processing_QC_Type), intent(in) :: Rayleigh_Processing_QC

    ! local variable
    integer :: o, s

    IF (associated(Rayleigh_Processing_QC%Obs_Rayleigh_QC)) THEN
       ! get the size of the array
       s = size(Rayleigh_Processing_QC%Obs_Rayleigh_QC)
       print "(a,i2,a)",&
            " ==> size(Rayleigh_Processing_QC%Obs_Rayleigh_QC) = ",s,&
            " [printing first and last one]"
       DO o=1,s,s-1
          print "('    Rayleigh_Processing_QC%"//&
               "Obs_Rayleigh_QC(',i2.2,')%...')",o
          call Print_Obs_Rayleigh_QC(Rayleigh_Processing_QC%Obs_Rayleigh_QC(o))
       END DO
    ELSE
       print *," Rayleigh_Processing_QC%Obs_Rayleigh_QC(:) "//&
            "[seems not associated]"
    END IF

    print *,"Rayleigh_Processing_QC%Rayleigh_Background_High = ",&
         Rayleigh_Processing_QC%Rayleigh_Background_High

  end subroutine Print_Rayleigh_Processing_QC
    !  #]
  subroutine AsciiWrite_Rayl_Processing_QC(Rayleigh_Processing_QC, fileunit, &
                                           prefix,PrintAll,error_flag)
    !  #[
    TYPE(Rayleigh_Processing_QC_Type), intent(in)  :: Rayleigh_Processing_QC
    integer,                           intent(in)  :: fileunit   ! input
    character(len=*),                  intent(in)  :: prefix     ! input
    logical,                           intent(in)  :: PrintAll   ! input
    integer,                           intent(out) :: error_flag ! output

    ! local variables
    integer                      :: o, s, step
    character(len=2)             :: str_cntr
    character(len=maxlen_IntAuc) :: str_IntAuc

    IF (associated(Rayleigh_Processing_QC%Obs_Rayleigh_QC)) THEN
       ! get the size of the array
       s = size(Rayleigh_Processing_QC%Obs_Rayleigh_QC)
       IF (PrintAll .or. (s .le. 1) ) THEN
          step=1
       ELSE
          step=s-1
       END IF
       DO o=1,s,step
          write(str_cntr,"(i2.2)",err=998) o
          call AsciiWrite_Obs_Rayleigh_QC(&
               Rayleigh_Processing_QC%Obs_Rayleigh_QC(o),fileunit,&
               trim(prefix)//"%Obs_Rayleigh_QC("//trim(str_cntr)//")",&
               PrintAll,error_flag)
       END DO
    ELSE
       write(fileunit,"(a)",err=999) trim(prefix)//&
          "%Obs_Rayleigh_QC(:) [seems not associated]"
    END IF

    write(fileunit,"(a)",err=999) "REALACC(12)"

    call SetIntAuc(Rayleigh_Processing_QC%Rayleigh_Background_High,&
                   str_IntAuc,error_flag)
    IF (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%Rayleigh_Background_High = "//trim(str_IntAuc)

    write(fileunit,"(a)",err=999) "ENDREALACC"

    return

998 call logmsg(log_error,"conversion error in "//&
         "AsciiWrite_Rayl_Processing_QC: "//&
         "could not convert the counter o to a string ..")
    error_flag = error_writing_file
    return

999 call logmsg(log_error,"ERROR in AsciiWrite_Rayl_Processing_QC: "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_Rayl_Processing_QC
    !  #]
  function GetByteSize_Rayl_Processing_QC(Rayleigh_Processing_QC, &
                                          N_Obs_Rayleigh_to_write) &
       result(c_bytesize)
    !  #[

    integer(IntAl)                                :: c_bytesize ! result
    TYPE(Rayleigh_Processing_QC_Type), intent(in) :: Rayleigh_Processing_QC
    integer(IntAuc),                   intent(in) :: N_Obs_Rayleigh_to_write

    ! local variable
    integer :: s, error_flag
    character(len=maxlen_IntAuc) :: str1_IntAuc, str2_IntAuc

    c_bytesize = 1*c_bytelen_IntAuc + 20 ! spare_bytes

    IF (associated(Rayleigh_Processing_QC%Obs_Rayleigh_QC)) THEN
       s =  size(Rayleigh_Processing_QC%Obs_Rayleigh_QC)
       
       IF (N_Obs_Rayleigh_to_write .gt. s) then
          call SetIntAuc(N_Obs_Rayleigh_to_write,str1_IntAuc,error_flag)
          IF (error_flag .ne. no_error) str1_IntAuc = "[er]"
          call SetIntAuc(int(s,IntAuc), str2_IntAuc,error_flag)
          IF (error_flag .ne. no_error) str2_IntAuc = "[er]"
          call logmsg(log_warn,"WARNING in "//&
               "GetByteSize_Rayl_Processing_QC(): invalid value for "//&
               "N_Obs_Rayleigh_to_write: "//trim(str1_IntAuc)//&
               " max. possible value is: "//&
               "size(Rayleigh_Processing_QC%Obs_Rayleigh_QC) = "//&
               trim(str2_IntAuc)//" GetByteSize result will be useless...")
       END IF

       c_bytesize = c_bytesize + N_Obs_Rayleigh_to_write * &
            GetByteSize_Obs_Rayleigh_QC(&
                Rayleigh_Processing_QC%Obs_Rayleigh_QC(1))
    END IF

    ! typical sizes
    ! memory size = 2   +5*1009 = 5047 bytes
    ! file   size = 1+20+5* 913 = 4586 bytes
  end function GetByteSize_Rayl_Processing_QC
    !  #]
  subroutine Write_Rayleigh_Processing_QC(fileunit, &
             Rayleigh_Processing_QC, N_Obs_Rayleigh_to_write, &
             error_flag)
    !  #[
    !  Writes the Rayleigh_Processing_QC in binary format
    !  J. de Kloe  03-01-2006 Original Code
    !  P. Poli     10-01-2006 Removed placeholders. Put actual writing in place
    !                         From L2B QC tables v0.2 03-01-2005
    
    integer,         intent(in)  :: fileunit                ! input
    TYPE(Rayleigh_Processing_QC_Type), &
                     intent(in)  :: Rayleigh_Processing_QC  ! input
    integer(IntAuc), intent(in)  :: N_Obs_Rayleigh_to_write ! input
    integer,         intent(out) :: error_flag              ! output

    ! local variable
    integer :: i,s,nr_spare_bytes
    character(len=maxlen_IntAuc) :: str1_IntAuc, str2_IntAuc

    !---------------------------------------
    ! writing the Rayleigh_Processing_QC elements
    !---------------------------------------

    error_flag = no_error
    IF (.not. associated(Rayleigh_Processing_QC%Obs_Rayleigh_QC)) THEN
       call logmsg(log_error,"ERROR in Write_Rayleigh_Processing_QC(): "//&
            "The pointer to this Rayleigh_Processing_QC%Obs_Rayleigh_QC "//&
            "seems not associated !!")
       error_flag = error_programming
       return
    END IF
    
    s = size(Rayleigh_Processing_QC%Obs_Rayleigh_QC)
    IF (N_Obs_Rayleigh_to_write .gt. s) THEN
       call SetIntAuc(N_Obs_Rayleigh_to_write,str1_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
       call SetIntAuc(int(s,IntAuc), str2_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_error,"ERROR in "//&
            "Write_Rayleigh_Processing_QC(): invalid value for "//&
            "N_Obs_Rayleigh_to_write: "//trim(str1_IntAuc)//&
            " max. possible value is: "//&
            "size(Rayleigh_Processing_QC%Obs_Rayleigh_QC) = "//&
            trim(str2_IntAuc))
       error_flag = error_programming
       return
    END IF

    DO i=1,N_Obs_Rayleigh_to_write
       call Write_Obs_Rayleigh_QC(fileunit, &
            Rayleigh_Processing_QC%Obs_Rayleigh_QC(i), error_flag)
       IF (error_flag .ne. no_error) goto 998
    END DO

    call write_IntAuc_to_bin_file(fileunit, &
         Rayleigh_Processing_QC%Rayleigh_Background_High, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! 20 spare bytes
    nr_spare_bytes = 20
    call write_SpareBytes_to_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Write_Rayleigh_Processing_QC")
    return

  end subroutine Write_Rayleigh_Processing_QC
    !  #]
  subroutine Read_Rayleigh_Processing_QC(fileunit, &
             Rayleigh_Processing_QC, error_flag)
    !  #[
    !  P. Poli     10-01-2006 Original code
    integer,                          intent(in)    :: fileunit
    TYPE(Rayleigh_Processing_QC_Type),intent(inout) :: Rayleigh_Processing_QC
    integer,                          intent(out)   :: error_flag

    ! local variable
    integer :: i,s,nr_spare_bytes

    !---------------------------------------
    ! reading the Rayleigh_Processing_QC elements
    !---------------------------------------

    error_flag = no_error

    ! N_Obs may turn out to be zero !
    ! so this check does not make sense anymore
    !IF (.not. associated(Rayleigh_Processing_QC%Obs_Rayleigh_QC)) THEN
    !   call logmsg(log_error,"ERROR in Read_Rayleigh_Processing_QC(): "//&
    !        "The pointer to this Rayleigh_Processing_QC%Obs_Rayleigh_QC "//&
    !        "seems not associated !!")
    !   error_flag = error_programming
    !   return
    !END IF
    
    IF (associated(Rayleigh_Processing_QC%Obs_Rayleigh_QC)) THEN
       s = size(Rayleigh_Processing_QC%Obs_Rayleigh_QC)
       DO i=1,s
          call Read_Obs_Rayleigh_QC(fileunit, &
               Rayleigh_Processing_QC%Obs_Rayleigh_QC(i), error_flag)
          IF (error_flag .ne. no_error) goto 998
       END DO
    END IF

    call read_IntAuc_from_bin_file(fileunit, &
         Rayleigh_Processing_QC%Rayleigh_Background_High, error_flag)
    IF (error_flag .ne. no_error) goto 998
    
    ! 20 spare bytes
    nr_spare_bytes = 20
    call read_SpareBytes_from_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Read_Rayleigh_Processing_QC")
    return

  end subroutine Read_Rayleigh_Processing_QC
    !  #]
  !  #]
  !---------------------------------------------------------------------
  ! subroutines for the L2B_PCD_ADS datastructure    
  !---------------------------------------------------------------------
  !  #[
  subroutine Init_L2B_PCD_ADS(L2B_PCD_ADS, N_Meas0, &
             N_Obs_Mie_Max0, N_Obs_Rayleigh_Max0, N_Range_Gates0, error_flag)
    !  #[
    !   P. Poli    30-01-2006   Added N_Meas, N_Obs_*_Actual
    TYPE(L2B_PCD_ADS_type), pointer     :: L2B_PCD_ADS  ! input/output
    integer(IntAs),         intent(in)  :: N_Meas0             ! input
    integer(IntAs),         intent(in)  :: N_Obs_Mie_Max0      ! input
    integer(IntAs),         intent(in)  :: N_Obs_Rayleigh_Max0 ! input
    integer(IntAs),         intent(in)  :: N_Range_Gates0      ! input
    integer,                intent(out) :: error_flag   ! output

    error_flag = no_error

    IF (.not. associated(L2B_PCD_ADS)) THEN
       call logmsg(log_error,"ERROR in Init_L2B_PCD_ADS: "//&
            "pointer L2B_PCD_ADS seems not associated !!!")
       error_flag = error_programming
       return
    END IF

    ! this is a constant, defined by the BRC in the L1B input file
    L2B_PCD_ADS%N_Meas                = N_Meas0

    ! these are filled during processing, so init them to zero
    L2B_PCD_ADS%N_Obs_Mie_Actual      = 0_IntAs
    L2B_PCD_ADS%N_Obs_Rayleigh_Actual = 0_IntAs

    call Init_L1B_Input_Screening(L2B_PCD_ADS%L1B_Input_Screening,&
                                  N_Meas0, N_Range_Gates0, error_flag)
    IF (error_flag .ne. no_error) return

    call Init_L2B_AMD_Screening(L2B_PCD_ADS%L2B_AMD_Screening,&
                                N_Meas0, error_flag)
    IF (error_flag .ne. no_error) return

    call Init_Mie_Classification_QC(L2B_PCD_ADS%Mie_Classification_QC,&
                                N_Range_Gates0, N_Meas0, N_Obs_Mie_Max0, &
                                error_flag)
    IF (error_flag .ne. no_error) return

    call Init_Rayl_Classification_QC(L2B_PCD_ADS%Rayl_Classification_QC,&
                                N_Range_Gates0, N_Meas0, N_Obs_Rayleigh_Max0, &
                                error_flag)
    IF (error_flag .ne. no_error) return

    call Init_Mie_Processing_QC(L2B_PCD_ADS%Mie_Processing_QC, &
                                N_Obs_Mie_Max0, N_Range_Gates0, &
                                error_flag)
    IF (error_flag .ne. no_error) return

    call Init_Rayleigh_Processing_QC(L2B_PCD_ADS%Rayleigh_Processing_QC, &
                                     N_Obs_Rayleigh_Max0, N_Range_Gates0, &
                                     error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine Init_L2B_PCD_ADS
    !  #]
  subroutine Delete_L2B_PCD_ADS(L2B_PCD_ADS)
    !  #[
    TYPE(L2B_PCD_ADS_type), pointer :: L2B_PCD_ADS

    IF (.not. associated(L2B_PCD_ADS)) return

    call Delete_L1B_Input_Screening(L2B_PCD_ADS%L1B_Input_Screening)
    call Delete_L2B_AMD_Screening(L2B_PCD_ADS%L2B_AMD_Screening)
    call Delete_Mie_Classification_QC(L2B_PCD_ADS%Mie_Classification_QC)
    call Delete_Rayl_Classification_QC(L2B_PCD_ADS%Rayl_Classification_QC)
    call Delete_Mie_Processing_QC(L2B_PCD_ADS%Mie_Processing_QC)
    call Delete_Rayleigh_Processing_QC(L2B_PCD_ADS%Rayleigh_Processing_QC)

  end subroutine Delete_L2B_PCD_ADS
    !  #]
  subroutine Print_L2B_PCD_ADS(L2B_PCD_ADS)
    !  #[
    !   P. Poli    30-01-2006   Added N_Meas, N_Obs_*_Actual
    TYPE(L2B_PCD_ADS_type), pointer :: L2B_PCD_ADS

    print *,"-----------"
    IF (associated(L2B_PCD_ADS)) THEN
       print *,"L2B_PCD_ADS%..."
    ELSE
       print *,"L2B_PCD_ADS [seems not associated]"
       return
    END IF

    print *,"...%N_Meas                = ",L2B_PCD_ADS%N_Meas
    print *,"...%N_Obs_Mie_Actual      = ",L2B_PCD_ADS%N_Obs_Mie_Actual
    print *,"...%N_Obs_Rayleigh_Actual = ",L2B_PCD_ADS%N_Obs_Rayleigh_Actual

    print *,"-----------"
    print *,"REALACC(12)"
    print *,"L2B_PCD_ADS%L1B_Input_Screening%..."
    call Print_L1B_Input_Screening(L2B_PCD_ADS%L1B_Input_Screening)
    print *,"-----------"
    print *,"L2B_PCD_ADS%L2B_AMD_Screening%..."
    call Print_L2B_AMD_Screening(L2B_PCD_ADS%L2B_AMD_Screening)
    print *,"-----------"
    print *,"L2B_PCD_ADS%Mie_Classification_QC%..."
    call Print_Mie_Classification_QC(L2B_PCD_ADS%Mie_Classification_QC)
    print *,"-----------"
    print *,"L2B_PCD_ADS%Rayl_Classification_QC%..."
    call Print_Rayl_Classification_QC(L2B_PCD_ADS%Rayl_Classification_QC)
    print *,"-----------"
    print *,"L2B_PCD_ADS%Mie_Processing_QC%..."
    call Print_Mie_Processing_QC(L2B_PCD_ADS%Mie_Processing_QC)
    print *,"-----------"
    print *,"L2B_PCD_ADS%Rayleigh_Processing_QC%..."
    call Print_Rayleigh_Processing_QC(L2B_PCD_ADS%Rayleigh_Processing_QC)
    print *,"ENDREALACC"
    print *,"-----------"

  end subroutine Print_L2B_PCD_ADS
    !  #]
  subroutine AsciiWrite_L2B_PCD_ADS(L2B_PCD_ADS,fileunit,prefix,&
                                    PrintAll,error_flag)
    !  #[
    TYPE(L2B_PCD_ADS_type), pointer     :: L2B_PCD_ADS
    integer,                intent(in)  :: fileunit   ! input
    character(len=*),       intent(in)  :: prefix     ! input
    logical,                intent(in)  :: PrintAll   ! input
    integer,                intent(out) :: error_flag ! output

    ! local variable
    character(len=maxlen_IntAs) :: str_IntAs

    IF (.not. associated(L2B_PCD_ADS)) THEN
       call logmsg(log_error,"ERROR in AsciiWrite_L2B_PCD_ADS: "//&
            "L2B_PCD_ADS [seems not associated]")
       error_flag = error_programming
       return
    END IF

    call SetIntAs(L2B_PCD_ADS%N_Meas,str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%N_Meas                = "//trim(str_IntAs)

    call SetIntAs(L2B_PCD_ADS%N_Obs_Mie_Actual,str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%N_Obs_Mie_Actual      = "//trim(str_IntAs)

    call SetIntAs(L2B_PCD_ADS%N_Obs_Rayleigh_Actual,str_IntAs,error_flag)
    if (error_flag .ne. no_error) return
    write(fileunit,"(a)",err=999) &
         trim(prefix)//"%N_Obs_Rayleigh_Actual = "//trim(str_IntAs)

    call AsciiWrite_L1B_Input_Screening(L2B_PCD_ADS%L1B_Input_Screening,&
         fileunit,trim(prefix)//"%L1B_Input_Screening",PrintAll,error_flag)
    IF (error_flag .ne. no_error) return

    call AsciiWrite_L2B_AMD_Screening(L2B_PCD_ADS%L2B_AMD_Screening,&
         fileunit,trim(prefix)//"%L2B_AMD_Screening",PrintAll,error_flag)
    IF (error_flag .ne. no_error) return

    call AsciiWrite_Mie_Classific_QC(L2B_PCD_ADS%Mie_Classification_QC,&
         fileunit,trim(prefix)//"%Mie_Classification_QC",PrintAll,error_flag)
    IF (error_flag .ne. no_error) return

    call AsciiWrite_Rayl_Classific_QC(L2B_PCD_ADS%Rayl_Classification_QC,&
         fileunit,trim(prefix)//"%Rayl_Classification_QC",PrintAll,error_flag)
    IF (error_flag .ne. no_error) return

    call AsciiWrite_Mie_Processing_QC(L2B_PCD_ADS%Mie_Processing_QC,&
         fileunit,trim(prefix)//"%Mie_Processing_QC",PrintAll,error_flag)
    IF (error_flag .ne. no_error) return

    call AsciiWrite_Rayl_Processing_QC(L2B_PCD_ADS%Rayleigh_Processing_QC,&
         fileunit,trim(prefix)//"%Rayleigh_Processing_QC",PrintAll,error_flag)
    IF (error_flag .ne. no_error) return
    
    return

999 call logmsg(log_error,"ERROR in AsciiWrite_L2B_PCD_ADS: "//&
         "error while writing to outputfile: "//trim(get_filename(fileunit)))
    error_flag = error_writing_file
    return

  end subroutine AsciiWrite_L2B_PCD_ADS
    !  #]
  function GetByteSize_L2B_PCD_ADS(L2B_PCD_ADS,N_Meas_to_write,&
       N_Obs_Mie_to_write, N_Obs_Rayleigh_to_write) result(c_bytesize)
    !  #[
    !   P. Poli    30-01-2006   Added N_Meas_Actual, N_Obs_*_Actual
    integer(IntAl) :: c_bytesize ! result
    TYPE(L2B_PCD_ADS_type), pointer :: L2B_PCD_ADS
    integer(IntAuc), intent(in) :: N_Meas_to_write
    integer(IntAuc), intent(in) :: N_Obs_Mie_to_write
    integer(IntAuc), intent(in) :: N_Obs_Rayleigh_to_write

    c_bytesize = 0
    IF (.not. associated(L2B_PCD_ADS)) return
       
    c_bytesize = 3*c_bytelen_IntAs + 20 ! spare_bytes

    c_bytesize = c_bytesize + &
         GetByteSize_L1B_Input_Screening(L2B_PCD_ADS%L1B_Input_Screening,&
                                         N_Meas_to_write)

    c_bytesize = c_bytesize + &
         GetByteSize_L2B_AMD_Screening(L2B_PCD_ADS%L2B_AMD_Screening,&
                                       N_Meas_to_write)

    c_bytesize = c_bytesize + &
         GetByteSize_Mie_Classific_QC(L2B_PCD_ADS%Mie_Classification_QC,&
                                      N_Meas_to_write, N_Obs_Mie_to_write)

    c_bytesize = c_bytesize + &
         GetByteSize_Rayl_Classific_QC(L2B_PCD_ADS%Rayl_Classification_QC,&
                             N_Meas_to_write, N_Obs_Rayleigh_to_write)

    c_bytesize = c_bytesize + &
         GetByteSize_Mie_Processing_QC(L2B_PCD_ADS%Mie_Processing_QC,&
                                       N_Obs_Mie_to_write)

    c_bytesize = c_bytesize + &
         GetByteSize_Rayl_Processing_QC(L2B_PCD_ADS%Rayleigh_Processing_QC,&
                                        N_Obs_Rayleigh_to_write)

    ! typical sizes
    ! memory size =    3*2+13420+272+20746+20746+9842+5047 = 70079 bytes
    ! file   size = 20+3*2+ 6884+223+19105+19105+9866+4586 = 59795 bytes

  end function GetByteSize_L2B_PCD_ADS
    !  #]
  subroutine Write_L2B_PCD_ADS(fileunit,L2B_PCD_ADS,N_Meas_to_write,&
       N_Obs_Mie_to_write, N_Obs_Rayleigh_to_write,error_flag)
    !  #[
    !   Jos de Kloe    January 2006   Original code
    !
    ! History:
    !   P. Poli    10-01-2006   Replaced placeholders by actual writing
    !   P. Poli    30-01-2006   Added N_Meas_Actual, N_Obs_*_Actual

    integer,                    intent(in)  :: fileunit                ! input
    TYPE(L2B_PCD_ADS_type),     pointer     :: L2B_PCD_ADS             ! input
    integer(IntAuc),            intent(in)  :: N_Meas_to_write         ! input
    integer(IntAuc),            intent(in)  :: N_Obs_Mie_to_write      ! input
    integer(IntAuc),            intent(in)  :: N_Obs_Rayleigh_to_write ! input
    integer,                    intent(out) :: error_flag ! output

    ! local variables
    integer :: nr_spare_bytes

    error_flag = no_error
    IF (.not. associated(L2B_PCD_ADS)) THEN
       call logmsg(log_error,"ERROR in Write_L2B_PCD_ADS(): "//&
            "The pointer to this L2B_PCD_ADS seems not associated !!")
       error_flag = error_programming
       return
    END IF

    call write_IntAs_to_bin_file(fileunit, &
         L2B_PCD_ADS%N_Meas,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call write_IntAs_to_bin_file(fileunit, &
         L2B_PCD_ADS%N_Obs_Mie_Actual,error_flag)
    IF (error_flag .ne. no_error) goto 998
       
    call write_IntAs_to_bin_file(fileunit, &
         L2B_PCD_ADS%N_Obs_Rayleigh_Actual,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! L1B_Input_Screening
    call Write_L1B_Inp_Screening(fileunit, &
         L2B_PCD_ADS%L1B_Input_Screening, N_Meas_to_write, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! L2B_AMD_Screening
    call Write_L2B_AMD_Screening(fileunit, &
         L2B_PCD_ADS%L2B_AMD_Screening, N_Meas_to_write, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Mie_Classification_QC
    call Write_Mie_Classification_QC(fileunit, &
         L2B_PCD_ADS%Mie_Classification_QC, &
         N_Meas_to_write, N_Obs_Mie_to_write, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Rayleigh_Classification_QC
    call Write_Rayl_Classification_QC(fileunit, &
         L2B_PCD_ADS%Rayl_Classification_QC, &
         N_Meas_to_write, N_Obs_Rayleigh_to_write, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Mie_Processing_QC
    call Write_Mie_Processing_QC(fileunit, &
         L2B_PCD_ADS%Mie_Processing_QC, &
         N_Obs_Mie_to_write, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Rayleigh_Processing_QC
    call Write_Rayleigh_Processing_QC(fileunit, &
         L2B_PCD_ADS%Rayleigh_Processing_QC, &
         N_Obs_Rayleigh_to_write, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 20 spare bytes
    nr_spare_bytes = 20
    call write_SpareBytes_to_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Write_L2B_PCD_ADS")
    return

  end subroutine Write_L2B_PCD_ADS
    !  #]
  subroutine Read_L2B_PCD_ADS(fileunit,L2B_PCD_ADS,error_flag)
    !  #[
    !   P. Poli    10-01-2006   Original code
    !   P. Poli    30-01-2006   Added N_Meas, N_Obs_*_Actual
    integer,                intent(in)  :: fileunit
    TYPE(L2B_PCD_ADS_type), pointer     :: L2B_PCD_ADS
    integer,                intent(out) :: error_flag

    ! local variables
    integer :: nr_spare_bytes

    error_flag = no_error
    IF (.not. associated(L2B_PCD_ADS)) THEN
       call logmsg(log_error,"ERROR in Read_L2B_PCD_ADS(): "//&
            "The pointer to this L2B_PCD_ADS seems not associated !!")
       error_flag = error_programming
       return
    END IF

    ! Reading N_Meas_Actual, N_Obs_*_Actual
    call read_IntAs_from_bin_file(fileunit, &
         L2B_PCD_ADS%N_Meas,error_flag)

    IF (error_flag .ne. no_error) goto 998
    call read_IntAs_from_bin_file(fileunit, &
         L2B_PCD_ADS%N_Obs_Mie_Actual,error_flag)
    IF (error_flag .ne. no_error) goto 998

    call read_IntAs_from_bin_file(fileunit, &
         L2B_PCD_ADS%N_Obs_Rayleigh_Actual,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! L1B_Input_Screening
    call Read_L1B_Inp_Screening(fileunit, &
         L2B_PCD_ADS%L1B_Input_Screening, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! L2B_AMD_Screening
    call Read_L2B_AMD_Screening(fileunit, &
         L2B_PCD_ADS%L2B_AMD_Screening, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Mie_Classification_QC
    call Read_Mie_Classification_QC(fileunit, &
         L2B_PCD_ADS%Mie_Classification_QC, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Rayleigh_Classification_QC
    call Read_Rayl_Classification_QC(fileunit, &
         L2B_PCD_ADS%Rayl_Classification_QC, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Mie_Processing_QC
    call Read_Mie_Processing_QC(fileunit, &
         L2B_PCD_ADS%Mie_Processing_QC, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Rayleigh_Processing_QC
    call Read_Rayleigh_Processing_QC(fileunit, &
         L2B_PCD_ADS%Rayleigh_Processing_QC, error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! 20 spare bytes
    nr_spare_bytes = 20
    call read_SpareBytes_from_bin_file(fileunit,nr_spare_bytes,error_flag)
    IF (error_flag .ne. no_error) goto 998

    return

998 call logmsg(log_error,"ABORTING subroutine Read_L2B_PCD_ADS")
    return

  end subroutine Read_L2B_PCD_ADS
    !  #]
  !  #]
  !---------------------------------------------------------------------
! question:
! -if we want to follow the object oriented style that I tried to introduce
! in these modules, then indeed we should not access the elements inside the
! datastructures directly like this:
!
! L2B%PCD_ADS%Mie_Processing_QC%Obs_Mie_QC(o)%Obs_Mie_Bin_QC(r)%&
!             Hlos_Error_Estimate = some_number
!
! but in stead define little subroutines/functions to fill/retrieve these elements
! like this:
!
! call Set_L2B_PCD_Mie_ProcQC(obs=o,bin=r,Hlos_Error_Estimate=some_number)
!
! This is also done in the example below (and we do a similar thing with our
! FH, MPH, SPH datastructures).
! This however is a lot of extra work, and maybe it does not add very much to the
! readability of the code.
! David, Paul, what do you think we should do?
! JK, 12-mar-2007
    subroutine Update_L2B_PCD_N_Obs_Actual(PCD_ADS, &
         N_Obs_Mie_Actual0, N_Obs_Rayleigh_Actual0, error_flag)
    !  #[
    !  30-01-2006 P. Poli   Original code
    TYPE(L2B_PCD_ADS_Type), pointer       :: PCD_ADS
    integer(IntAs),         intent(in)    :: N_Obs_Mie_Actual0
    integer(IntAs),         intent(in)    :: N_Obs_Rayleigh_Actual0
    integer,                intent(out)   :: error_flag

    error_flag = no_error
    IF (.not.(associated(PCD_ADS))) THEN
       call logmsg(log_error,"ERROR in Update_L2B_PCD_N_Obs_Actual(): "//&
            "The pointer to this PCD_ADS seems not associated !!")
       error_flag = error_programming
       return
    END IF

    PCD_ADS%N_Obs_Mie_Actual      = N_Obs_Mie_Actual0
    PCD_ADS%N_Obs_Rayleigh_Actual = N_Obs_Rayleigh_Actual0

    end subroutine Update_L2B_PCD_N_Obs_Actual
    !  #]
end module L2B_PCD_ADS
