module L2BC_SPH
  !  #[ Documentation

  !
  ! this module defines the Specific Product Header for the ee_cfi fileformat
  ! for the L2B/L2C processor wind product data files
  ! The SPH format may differ for different filedefinitions, therefore
  ! we use a private definition for this filetype only
  !
  ! for a description, see 
  ! AE-IF-ECMWF-L2BP-001 (L2B IODD), sec.4.2, table 11
  !
  ! Written by J. de Kloe (KNMI)
  !
  ! History:
  !   28-Nov-2005 P. Poli     Added routine Write_L2B_SPH_to_KVT
  !   28-Nov-2005 P. Poli     Added number variables in SPH
  !   ...         J. de Kloe  small correction in USE statements
  !   02-Mar-2006 J. de Kloe  adapted to changes in XML_module
  !   13-Mar-2006 J. de Kloe  Add M_Meas parameter
  !   13-Mar-2006 J. de Kloe  added Read_L2B_SPH_from_KVT and 
  !                           GetByteSize_L2B_SPH_KVT routines
  !   26-Apr-2006 D. Tan      Update for IODD Draft 1.0D
  !   12-May-2006 D. Tan      Null use of ParKind1 module
  !   13-Jul-2006 D. Tan      Re-order DSDs
  !   17-Jul-2006 D. Tan      New L2B DSDs (Aeolus_Level_1B_Product 
  !                           and Clim_Product)
  !   18-Jul-2006 D. Tan      Additional L1B statistics
  !   18-Sep-2006 P. Poli     Removed remaining instances of L1B_File_Name
  !   26-Oct-2006 D. Tan      Generalize to L2BC, i.e. L2C-specific DSDs
  !   15-Nov-2006 J. de Kloe  reintroduced the Read_L2B_SPH_from_KVT
  !                           and GetByteSize_L2B_SPH_KVT code from march 13
  !   15-Nov-2006 J. de Kloe  Added 4 integer L2C items to the struct 
  !                           and implemented L2C handling in many routines
  !   01-Dec-2006 D. Tan      Change integer L2C items to bin-level
  !   20-Feb-2007 J. de Kloe  changed filling of intersection lat-lon values to
  !                           use integers as input to prevent rounding issues
  !   23-Feb-2007 J. de Kloe  some small corrections to the AsciiWrite code
  !                           and added some Get* routines
  !   02-Mar-2007 J. de Kloe  adapted to the latest IODD change (reordering of
  !                           DSD's and replaced routine add_L2B_SPH_item() by
  !                           the more generic xml_add_SPH_item defined in 
  !                           xml_module.F90
  !   09-Mar-2007 D. Tan      Use ee_sph_name.
  !   15-Mar-2007 J. de Kloe  removed use of SummaryVars
  !   07-Jun-2007 J. de Kloe  adapted to changed xml interface
  !   18-Jun-2007 J. de Kloe  cleaned Fill_L2BC_SPH_items subroutine
  !   03-Sep-2007 P. Poli     added tests to prevent allocation of 
  !                           DSDs if they have already been allocated
  !                           inside routine Get_L2BC_SPH_from_xml_tree
  !   16-Jan-2008 J. de Kloe  phase out integer kind i_ 
  !
  !  #] 
  !  #[ Modules used

  USE ParKind1, only:    ! For compatability with ECMWF-IFS build.
  USE XML_module, only: max_str_len, xml_node_type, xml_node_get_string, &
                        xml_set_fixed_header, check_length_xml_item, &
                        check_string_length, xml_tree_set_string_node_value, &
                        xml_tree_add_child, ee_vh_name, ee_sph_name, &
                        xml_tree_add_and_set_attribute, &
                        xml_add_SPH_item

  USE EE_CFI_Datatypes, only: &
       IntAuc, maxlen_IntAuc, SetIntAuc, GetIntAuc, &
       IntAus, maxlen_IntAus, SetIntAus, &
       IntAl,  maxlen_IntAl,  SetIntAl,  &
       IntAd, &
       FAdoxy, maxlen_FAdo36, SetFAdo36, &
       maxlen_FAdoxy, SetFAdoFull_or_Missing, &
       Remove_Plus_Sign, FAdoxy
  USE DataSetDescriptor, only: DSD_type, Init_DSD, AsciiWrite_DSD, &
                               Add_DSD_to_xml_SPH, Write_DSD_to_KVT, &
                               Get_DSD_from_xml_tree, Read_DSD_from_KVT, &
                               GetByteSize_DSD_KVT, Print_DSD, &
                               maxlen_DS_Name, maxlen_DS_Type
  USE KVT_module, only: write_kvt_string, write_spare_kvt_string, &
                        read_kvt_string,  read_spare_kvt_string, &
                        Compare_XML_to_KVT_item_Repl
  USE L2BC_DS_Settings, only: set_L2BC_string

  ! support modules
  USE Logging, only: logmsg,log_debug,log_info,log_progress,log_warn,log_error
  USE ErrorHandler, only: no_error, error_allocate, error_programming, &
       error_reading_file
  USE LunManager, only: fileunit_stdout
  USE Compiler_Features, only: newline
  USE AeolusConstants, only: N_range_gates, N_Meas_Max, &
                             N_Obs_Max_Mie, N_Obs_Max_Rayleigh

  !  #]
  !  #[ Type definitions and parameters
  implicit none

  ! Names of attached datasets in L2B products
  character(len=maxlen_DS_Name) ,parameter :: L2B_DS_Name_Geolocation_ADS=&
                                     "Geolocation_ADS             "
  character(len=maxlen_DS_Name) ,parameter :: L2B_DS_Name_PCD_ADS=&
                                     "Product_Confidence_Data_ADS "
  character(len=maxlen_DS_Name) ,parameter :: L2B_DS_Name_Mie_MDS=&
                                     "Mie_HLOSwind_MDS            "
  character(len=maxlen_DS_Name) ,parameter :: L2B_DS_Name_Rayleigh_MDS=&
                                     "Rayleigh_HLOSwind_MDS       "
  
  ! Names of attached datasets in L2C products
  character(len=maxlen_DS_Name) ,parameter :: L2C_DS_Name_AssimPCD_ADS=&
                                     "Assim_PCD_ADS               "
  character(len=maxlen_DS_Name) ,parameter :: L2C_DS_Name_MieVec_MDS=&
                                     "Mie_VecWind_MDS             "
  character(len=maxlen_DS_Name) ,parameter :: L2C_DS_Name_RayleighVec_MDS=&
                                     "Rayleigh_VecWind_MDS        "
  
  ! Names of reference datasets in L2B products
  character(len=maxlen_DS_Name) ,parameter :: L2B_DS_Name_L1B_Product=&
                                     "Aeolus_Level_1B_Product     "
  character(len=maxlen_DS_Name) ,parameter :: L2B_DS_Name_Met_MDS=&
                                     "Aux_Met_Product             "
  character(len=maxlen_DS_Name) ,parameter :: L2B_DS_Name_RBC_Product=&
                                     "Aeolus_RBC                  "
  character(len=maxlen_DS_Name) ,parameter :: L2B_DS_Name_CLM_Product=&
                                     "Clim_Product                "
  character(len=maxlen_DS_Name) ,parameter :: L2B_DS_Name_CAL_Product=&
                                     "Cal_Product                 "
  character(len=maxlen_DS_Name) ,parameter :: L2B_DS_Name_L2B_Param_ADS=&
                                     "Level_2B_Proc_Params        "

  ! Names of reference datasets in L2C products
  character(len=maxlen_DS_Name) ,parameter :: L2C_DS_Name_L2B_Product=&
                                     "Aeolus_Level_2B_Product     "
  character(len=maxlen_DS_Name) ,parameter :: L2C_DS_Name_L2C_Param_ADS=&
                                     "Level_2C_Proc_Params        "

  ! Types (Annotation/Measurement) of attached datasets in L2B products
  character(len=maxlen_DS_Type) ,parameter :: L2B_DS_Type_Geolocation_ADS = "A"
  character(len=maxlen_DS_Type) ,parameter :: L2B_DS_Type_PCD_ADS         = "A"
  character(len=maxlen_DS_Type) ,parameter :: L2B_DS_Type_Mie_MDS         = "M"
  character(len=maxlen_DS_Type) ,parameter :: L2B_DS_Type_Rayleigh_MDS    = "M"
  
  ! Types (Annotation/Measurement) of attached datasets in L2C products
  character(len=maxlen_DS_Type) ,parameter :: L2C_DS_Type_AssimPCD_ADS    = "A"
  character(len=maxlen_DS_Type) ,parameter :: L2C_DS_Type_MieVec_MDS      = "M"
  character(len=maxlen_DS_Type) ,parameter :: L2C_DS_Type_RayleighVec_MDS = "M"

  ! DS_Types for "reference datasets" in the L2B product file
  character(len=maxlen_DS_Type) ,parameter :: L2B_DS_Type_L1B_Product     = "R"
  character(len=maxlen_DS_Type) ,parameter :: L2B_DS_Type_Met_MDS         = "R"
  character(len=maxlen_DS_Type) ,parameter :: L2B_DS_Type_RBC_Product     = "R"
  character(len=maxlen_DS_Type) ,parameter :: L2B_DS_Type_CLM_Product     = "R"
  character(len=maxlen_DS_Type) ,parameter :: L2B_DS_Type_CAL_Product     = "R"
  character(len=maxlen_DS_Type) ,parameter :: L2B_DS_Type_L2B_Param_ADS   = "R"

  ! DS_Types for "reference datasets" in the L2C product file
  character(len=maxlen_DS_Type) ,parameter :: L2C_DS_Type_L2B_Product     = "R"
  character(len=maxlen_DS_Type) ,parameter :: L2C_DS_Type_L2C_Param_ADS   = "R"

  integer, parameter :: maxlen_SPH_Descriptor   = 28

  ! these numbers are constant for the L2B and L2C file types
  ! but since they are used both in this module, and in the 
  ! l2bc_datastructure module, it is better to use parameters for them
  character(len=2), parameter :: Num_DSD_string_L2B="10"
  character(len=2), parameter :: Num_DSD_string_L2C="15"
  integer,          parameter :: Num_DSD_L2B = 10
  integer,          parameter :: Num_DSD_L2C = 15
  integer,          parameter :: Num_DS_L2B  = 4
  integer,          parameter :: Num_DS_L2C  = 7

  TYPE L2BC_SPH_type
     !  #[
     character(len=maxlen_SPH_Descriptor)    :: SPH_Descriptor
     ! Spare_1
     character(len=maxlen_IntAus)            :: Num_BRC
     character(len=maxlen_IntAl)             :: Intersect_Start_Lat
     character(len=maxlen_IntAl)             :: Intersect_Start_Long
     character(len=maxlen_IntAl)             :: Intersect_Stop_Lat
     character(len=maxlen_IntAl)             :: Intersect_Stop_Long
     character(len=maxlen_FAdo36)            :: Sat_Track
     ! Spare_2
     character(len=maxlen_IntAuc)            :: M_Mie
     character(len=maxlen_IntAuc)            :: M_Rayleigh
     character(len=maxlen_IntAuc)            :: M_Meas ! added 22-3-2006, JK
     character(len=maxlen_IntAus)            :: Num_Valid_Mie_Profiles
     character(len=maxlen_IntAus)            :: Num_Valid_Rayleigh_Profiles
     character(len=maxlen_IntAus)            :: Num_Clear_Mie_Profiles
     character(len=maxlen_IntAus)            :: Num_Clear_Rayleigh_Profiles
     character(len=maxlen_IntAus)            :: Num_Cloud_Mie_Profiles
     character(len=maxlen_IntAus)            :: Num_Cloud_Rayleigh_Profiles
     character(len=maxlen_IntAus)            :: Num_Mie_Prof_Warm_Pulses
     character(len=maxlen_IntAus)            :: Num_Rayleigh_Prof_Warm_Pulses
     !
     character(len=maxlen_IntAus)            :: Num_Profiles_Surface_Mie
     character(len=maxlen_IntAus)            :: Num_Profiles_Surface_Ray
     !
     character(len=maxlen_IntAus)            :: Num_Valid_Obs_Prof_L1B_Mie
     character(len=maxlen_IntAus)            :: Num_Valid_Obs_Prof_L1B_Ray
     character(len=maxlen_IntAus)            :: Num_Valid_Meas_Prof_L1B_Mie
     character(len=maxlen_IntAus)            :: Num_Valid_Meas_Prof_L1B_Ray
     !
     character(len=maxlen_IntAus)            :: Num_Valid_Obs_Bins_L1B_Mie
     character(len=maxlen_IntAus)            :: Num_Valid_Obs_Bins_L1B_Ray
     character(len=maxlen_IntAus)            :: Num_Valid_Meas_Bins_L1B_Mie
     character(len=maxlen_IntAus)            :: Num_Valid_Meas_Bins_L1B_Ray
     !
     character(len=maxlen_IntAus)            :: Num_Invalid_Meas_Prof_L1B_Mie
     character(len=maxlen_IntAus)            :: Num_Invalid_Meas_Prof_L1B_Ray
     character(len=maxlen_IntAus)            :: Num_Invalid_Meas_Prof_L2B_Mie
     character(len=maxlen_IntAus)            :: Num_Invalid_Meas_Prof_L2B_Ray
     !
     character(len=maxlen_IntAus)            :: Num_Invalid_Meas_Bins_L1B_Mie
     character(len=maxlen_IntAus)            :: Num_Invalid_Meas_Bins_L1B_Ray
     character(len=maxlen_IntAus)            :: Num_Invalid_Meas_Bins_L2B_Mie
     character(len=maxlen_IntAus)            :: Num_Invalid_Meas_Bins_L2B_Ray
     !
     ! Spare_3                                  1234567890123456789012345678901
     !
     ! TODO: Renaming below to make clear that Obs refers to
     !       scalar observations at range-gate level.
     ! E.g. Num_Valid_Obs_L2B_Mie   -> Num_Valid_Obs_Bin_L2B_Mie
     !      Num_Valid_Obs_L2B_Ray   -> Num_Valid_Obs_Bin_L2B_Ray
     !      Num_Invalid_Obs_L2C_Mie -> Num_Invalid_Obs_Bin_L2C_Mie
     !      Num_Invalid_Obs_L2C_Ray -> Num_Invalid_Obs_Bin_L2C_Ray
     !
     character(len=maxlen_IntAus)            :: Num_Valid_Obs_L2B_Mie
     character(len=maxlen_IntAus)            :: Num_Valid_Obs_L2B_Ray
     !
     ! the L2C elements
     character(len=maxlen_IntAus)            :: Num_Invalid_Obs_L2C_Mie ! (L2C only)
     character(len=maxlen_IntAus)            :: Num_Invalid_Obs_L2C_Ray ! (L2C only)
     !
     ! Spare_4 (L2B and L2C)                    1234567890123456789012345678901
     !
     ! Note that the following DSD's are resorted now (2-mar-2007,JK)
     ! First the attached datasets, then the referenced datasets
     !
     ! the 4 attached L2B-related DSD's
     TYPE(DSD_type), pointer :: Geolocation_DSD
     TYPE(DSD_type), pointer :: PCD_DSD
     TYPE(DSD_type), pointer :: Mie_DSD
     TYPE(DSD_type), pointer :: Rayleigh_DSD
     ! the 3 attached L2C-related DSDs
     TYPE(DSD_type), pointer :: Assim_PCD_DSD        ! (L2C only)
     TYPE(DSD_type), pointer :: Mie_VecWind_DSD      ! (L2C only)
     TYPE(DSD_type), pointer :: Rayleigh_VecWind_DSD ! (L2C only)
     ! the 6 referenced L2B-related DSD's
     TYPE(DSD_type), pointer :: L1B_DSD
     TYPE(DSD_type), pointer :: Met_DSD
     TYPE(DSD_type), pointer :: RBC_DSD
     TYPE(DSD_type), pointer :: CLM_DSD
     TYPE(DSD_type), pointer :: CAL_DSD
     TYPE(DSD_type), pointer :: L2B_Param_DSD
     ! the 2 referenced L2C-related DSD's
     TYPE(DSD_type), pointer :: L2B_DSD              ! (L2C only)
     TYPE(DSD_type), pointer :: L2C_Param_DSD        ! (L2C only)
     !
  END TYPE L2BC_SPH_type
    !  #]

  !  #]
  !  #[ Module interface
! all is public for the moment
  !  #]
contains
  !---------------------------------
  subroutine Init_L2BC_SPH(L2BC_SPH,error_flag)
    !  #[
    TYPE(L2BC_SPH_type), pointer     :: L2BC_SPH
    integer,             intent(out) :: error_flag
    
    error_flag = no_error
    IF (.not. associated(L2BC_SPH)) THEN
       print *,"ERROR in Init_L2BC_SPH():"
       print *,"The pointer to this L2BC_SPH seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! see the remark in MainProductHeader.F90 for the reason why I use
    ! the form x(:)=' ' to initialise the strings

    L2BC_SPH%SPH_Descriptor(:)                = ' '
    ! Spare_1
    L2BC_SPH%Num_BRC(:)                       = ' '
    L2BC_SPH%Intersect_Start_Lat(:)           = ' '
    L2BC_SPH%Intersect_Start_Long(:)          = ' '
    L2BC_SPH%Intersect_Stop_Lat(:)            = ' '
    L2BC_SPH%Intersect_Stop_Long(:)           = ' '
    L2BC_SPH%Sat_Track(:)                     = ' '
    ! Spare_2
    L2BC_SPH%M_Mie(:)                         = ' '
    L2BC_SPH%M_Rayleigh(:)                    = ' '
    L2BC_SPH%M_Meas(:)                        = ' '
    L2BC_SPH%Num_Valid_Mie_Profiles(:)        = ' '
    L2BC_SPH%Num_Valid_Rayleigh_Profiles(:)   = ' '
    L2BC_SPH%Num_Clear_Mie_Profiles(:)        = ' '
    L2BC_SPH%Num_Clear_Rayleigh_Profiles(:)   = ' '
    L2BC_SPH%Num_Cloud_Mie_Profiles(:)        = ' '
    L2BC_SPH%Num_Cloud_Rayleigh_Profiles(:)   = ' '
    L2BC_SPH%Num_Mie_Prof_Warm_Pulses(:)      = ' '
    L2BC_SPH%Num_Rayleigh_Prof_Warm_Pulses(:) = ' '
    !
    L2BC_SPH%Num_Profiles_Surface_Mie(:)      = ' '
    L2BC_SPH%Num_Profiles_Surface_Ray(:)      = ' '
    !
    L2BC_SPH%Num_Valid_Obs_Prof_L1B_Mie(:)    = ' '
    L2BC_SPH%Num_Valid_Obs_Prof_L1B_Ray(:)    = ' '
    L2BC_SPH%Num_Valid_Meas_Prof_L1B_Mie(:)   = ' '
    L2BC_SPH%Num_Valid_Meas_Prof_L1B_Ray(:)   = ' '
    !
    L2BC_SPH%Num_Valid_Obs_Bins_L1B_Mie(:)    = ' '
    L2BC_SPH%Num_Valid_Obs_Bins_L1B_Ray(:)    = ' '
    L2BC_SPH%Num_Valid_Meas_Bins_L1B_Mie(:)   = ' '
    L2BC_SPH%Num_Valid_Meas_Bins_L1B_Ray(:)   = ' '
    !
    L2BC_SPH%Num_Invalid_Meas_Prof_L1B_Mie(:) = ' '
    L2BC_SPH%Num_Invalid_Meas_Prof_L1B_Ray(:) = ' '
    L2BC_SPH%Num_Invalid_Meas_Prof_L2B_Mie(:) = ' '
    L2BC_SPH%Num_Invalid_Meas_Prof_L2B_Ray(:) = ' '
    !
    L2BC_SPH%Num_Invalid_Meas_Bins_L1B_Mie(:) = ' '
    L2BC_SPH%Num_Invalid_Meas_Bins_L1B_Ray(:) = ' '
    L2BC_SPH%Num_Invalid_Meas_Bins_L2B_Mie(:) = ' '
    L2BC_SPH%Num_Invalid_Meas_Bins_L2B_Ray(:) = ' '
    ! Spare_3
    L2BC_SPH%Num_Valid_Obs_L2B_Mie(:)         = ' '
    L2BC_SPH%Num_Valid_Obs_L2B_Ray(:)         = ' '
    ! the L2C elements
    L2BC_SPH%Num_Invalid_Obs_L2C_Mie(:)       = ' '
    L2BC_SPH%Num_Invalid_Obs_L2C_Ray(:)       = ' '
    ! spare 4 

    nullify(L2BC_SPH%Geolocation_DSD) ! L2B
    nullify(L2BC_SPH%PCD_DSD)         ! L2B
    nullify(L2BC_SPH%Mie_DSD)         ! L2B
    nullify(L2BC_SPH%Rayleigh_DSD)    ! L2B

    nullify(L2BC_SPH%Assim_PCD_DSD)        ! L2C
    nullify(L2BC_SPH%Mie_VecWind_DSD)      ! L2C
    nullify(L2BC_SPH%Rayleigh_VecWind_DSD) ! L2C

    nullify(L2BC_SPH%L1B_DSD)       ! L2B
    nullify(L2BC_SPH%Met_DSD)       ! L2B
    nullify(L2BC_SPH%RBC_DSD)       ! L2B
    nullify(L2BC_SPH%CLM_DSD)       ! L2B
    nullify(L2BC_SPH%CAL_DSD)       ! L2B
    nullify(L2BC_SPH%L2B_Param_DSD) ! L2B

    nullify(L2BC_SPH%L2B_DSD)       ! L2C
    nullify(L2BC_SPH%L2C_Param_DSD) ! L2C

  end subroutine Init_L2BC_SPH
    !  #]
  subroutine Print_L2BC_SPH(L2BC_SPH,L2C_flag,error_flag)
    !  #[
    TYPE(L2BC_SPH_type), pointer     :: L2BC_SPH   ! input
    integer,             intent(out) :: error_flag ! output
    logical,   optional, intent(in)  :: L2C_flag   ! switch

    ! local variable
    character(len=256) :: txt
    logical            :: L2C_flag_local
    character(len=3)   :: L2BC_string
    character(len=7)   :: L2BCsph_string

    ! init
    error_flag = no_error

    ! interpret optional arugment(s)
    L2C_flag_local=.false.
    IF (present(L2C_flag)) L2C_flag_local=L2C_flag

    L2BC_string    = set_L2BC_string(L2C_flag_local)
    L2BCsph_string = L2BC_string//"_SPH"

    IF (.not. associated(L2BC_SPH)) THEN
       print *,"[pointer to "//L2BCsph_string//&
               " is not associated so nothing to print]"
       return
    END IF

    txt = L2BC_SPH%SPH_Descriptor
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%SPH_Descriptor                = ",trim(txt)

    ! Spare_1

    txt = L2BC_SPH%Num_BRC
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_BRC                       = ",trim(txt)

    txt = L2BC_SPH%Intersect_Start_Lat
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Intersect_Start_Lat           = ",trim(txt)

    txt = L2BC_SPH%Intersect_Start_Long
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Intersect_Start_Long          = ",trim(txt)

    txt = L2BC_SPH%Intersect_Stop_Lat
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Intersect_Stop_Lat            = ",trim(txt)

    txt = L2BC_SPH%Intersect_Stop_Long
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Intersect_Stop_Long           = ",trim(txt)

    ! Spare_2

    txt = L2BC_SPH%Sat_Track
!    txt = Remove_Plus_Sign(L2BC_SPH%Sat_Track)
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Sat_Track                     = ",trim(txt)

    txt = L2BC_SPH%M_Mie
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%M_Mie                         = ",trim(txt)

    txt = L2BC_SPH%M_Rayleigh
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%M_Rayleigh                    = ",trim(txt)

    txt = L2BC_SPH%M_Meas
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%M_Meas                        = ",trim(txt)

    txt = L2BC_SPH%Num_Valid_Mie_Profiles
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Valid_Mie_Profiles        = ",trim(txt)

    txt = L2BC_SPH%Num_Valid_Rayleigh_Profiles
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Valid_Rayleigh_Profiles   = ",trim(txt)

    txt = L2BC_SPH%Num_Clear_Mie_Profiles
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Clear_Mie_Profiles        = ",trim(txt)

    txt = L2BC_SPH%Num_Clear_Rayleigh_Profiles
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Clear_Rayleigh_Profiles   = ",trim(txt)

    txt = L2BC_SPH%Num_Cloud_Mie_Profiles
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Cloud_Mie_Profiles        = ",trim(txt)

    txt = L2BC_SPH%Num_Cloud_Rayleigh_Profiles
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Cloud_Rayleigh_Profiles   = ",trim(txt)

    txt = L2BC_SPH%Num_Mie_Prof_Warm_Pulses
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Mie_Prof_Warm_Pulses      = ",trim(txt)

    txt = L2BC_SPH%Num_Rayleigh_Prof_Warm_Pulses
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Rayleigh_Prof_Warm_Pulses = ",trim(txt)

    txt = L2BC_SPH%Num_Profiles_Surface_Mie
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Profiles_Surface_Mie      = ",trim(txt)

    txt = L2BC_SPH%Num_Profiles_Surface_Ray
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Profiles_Surface_Ray      = ",trim(txt)

    !
    txt = L2BC_SPH%Num_Valid_Obs_Prof_L1B_Mie
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Valid_Obs_Prof_L1B_Mie  = ",trim(txt)

    txt = L2BC_SPH%Num_Valid_Obs_Prof_L1B_Ray
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Valid_Obs_Prof_L1B_Ray  = ",trim(txt)

    txt = L2BC_SPH%Num_Valid_Meas_Prof_L1B_Mie
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Valid_Meas_Prof_L1B_Mie = ",trim(txt)

    txt = L2BC_SPH%Num_Valid_Meas_Prof_L1B_Ray
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Valid_Meas_Prof_L1B_Ray = ",trim(txt)

    !
    txt = L2BC_SPH%Num_Valid_Obs_Bins_L1B_Mie
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Valid_Obs_Bins_L1B_Mie  = ",trim(txt)

    txt = L2BC_SPH%Num_Valid_Obs_Bins_L1B_Ray
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Valid_Obs_Bins_L1B_Ray  = ",trim(txt)

    txt = L2BC_SPH%Num_Valid_Meas_Bins_L1B_Mie
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Valid_Meas_Bins_L1B_Mie = ",trim(txt)

    txt = L2BC_SPH%Num_Valid_Meas_Bins_L1B_Ray
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Valid_Meas_Bins_L1B_Ray = ",trim(txt)

    !
    txt = L2BC_SPH%Num_Invalid_Meas_Prof_L1B_Mie
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Invalid_Meas_Prof_L1B_Mie = ",trim(txt)

    txt = L2BC_SPH%Num_Invalid_Meas_Prof_L1B_Ray
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Invalid_Meas_Prof_L1B_Ray = ",trim(txt)

    txt = L2BC_SPH%Num_Invalid_Meas_Prof_L2B_Mie
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Invalid_Meas_Prof_L2B_Mie = ",trim(txt)

    txt = L2BC_SPH%Num_Invalid_Meas_Prof_L2B_Ray
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Invalid_Meas_Prof_L2B_Ray = ",trim(txt)

    !
    txt = L2BC_SPH%Num_Invalid_Meas_Bins_L1B_Mie
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Invalid_Meas_Bins_L1B_Mie = ",trim(txt)

    txt = L2BC_SPH%Num_Invalid_Meas_Bins_L1B_Ray
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Invalid_Meas_Bins_L1B_Ray = ",trim(txt)

    txt = L2BC_SPH%Num_Invalid_Meas_Bins_L2B_Mie
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Invalid_Meas_Bins_L2B_Mie = ",trim(txt)

    txt = L2BC_SPH%Num_Invalid_Meas_Bins_L2B_Ray
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Invalid_Meas_Bins_L2B_Ray = ",trim(txt)

    ! Spare_3
    
    txt = L2BC_SPH%Num_Valid_Obs_L2B_Mie
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Valid_Obs_L2B_Mie         = ",trim(txt)

    txt = L2BC_SPH%Num_Valid_Obs_L2B_Ray
    IF (len_trim(txt).eq.0) txt = "[empty]"
    print *,"...%Num_Valid_Obs_L2B_Ray         = ",trim(txt)
    
    IF (L2C_flag_local) THEN
       txt = L2BC_SPH%Num_Invalid_Obs_L2C_Mie
       IF (len_trim(txt).eq.0) txt = "[empty]"
       print *,"...%Num_Invalid_Obs_L2C_Mie       = ",trim(txt)

       txt = L2BC_SPH%Num_Invalid_Obs_L2C_Ray
       IF (len_trim(txt).eq.0) txt = "[empty]"
       print *,"...%Num_Invalid_Obs_L2C_Ray       = ",trim(txt)
    END IF

    ! Spare_4

    ! the list of DSD's
    print *,L2BCsph_string//"%Geolocation_DSD%..."
    call Print_DSD(L2BC_SPH%Geolocation_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    print *,L2BCsph_string//"%PCD_DS%..."
    call Print_DSD(L2BC_SPH%PCD_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    print *,L2BCsph_string//"%Mie_DSD%..."
    call Print_DSD(L2BC_SPH%Mie_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    print *,L2BCsph_string//"%Rayleigh_DSD%..."
    call Print_DSD(L2BC_SPH%Rayleigh_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    
    IF (L2C_flag_local) THEN
       print *,L2BCsph_string//"%Assim_PCD_DS%..."
       call Print_DSD(L2BC_SPH%Assim_PCD_DSD,"...",error_flag)
       if (error_flag .ne. no_error) return
       print *,L2BCsph_string//"%Mie_VecWind_DSD%..."
       call Print_DSD(L2BC_SPH%Mie_VecWind_DSD,"...",error_flag)
       if (error_flag .ne. no_error) return
       print *,L2BCsph_string//"%Rayleigh_VecWind_DSD%..."
       call Print_DSD(L2BC_SPH%Rayleigh_VecWind_DSD,"...",error_flag)
       if (error_flag .ne. no_error) return
    END IF
    
    print *,L2BCsph_string//"%L1B_DSD%..."
    call Print_DSD(L2BC_SPH%L1B_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    print *,L2BCsph_string//"%Met_DSD%..."
    call Print_DSD(L2BC_SPH%Met_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    print *,L2BCsph_string//"%RBC_DSD%..."
    call Print_DSD(L2BC_SPH%RBC_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    print *,L2BCsph_string//"%CLM_DSD%..."
    call Print_DSD(L2BC_SPH%CLM_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    print *,L2BCsph_string//"%CAL_DSD%..."
    call Print_DSD(L2BC_SPH%CAL_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return
    print *,L2BCsph_string//"%L2B_Param_DSD%..."
    call Print_DSD(L2BC_SPH%L2B_Param_DSD,"...",error_flag)
    if (error_flag .ne. no_error) return

! TODO: (possibly) consistency_sph(L2BC,L2C_flag_local)

    IF (L2C_flag_local) THEN
       print *,L2BCsph_string//"%L2B_DSD%..."
       call Print_DSD(L2BC_SPH%L2B_DSD,"...",error_flag)
       if (error_flag .ne. no_error) return
       print *,L2BCsph_string//"%L2C_Param_DSD%..."
       call Print_DSD(L2BC_SPH%L2C_Param_DSD,"...",error_flag)
       if (error_flag .ne. no_error) return
    ENDIF
    
  end subroutine Print_L2BC_SPH
    !  #]
  subroutine AsciiWrite_L2BC_SPH(L2BC_SPH,fileunit,prefix,&
                                 error_flag,L2C_flag)
    !  #[
    TYPE(L2BC_SPH_type), pointer     :: L2BC_SPH
    integer,             intent(in)  :: fileunit
    character(len=*),    intent(in)  :: prefix
    integer,             intent(out) :: error_flag
    logical,   optional, intent(in)  :: L2C_flag

    ! local variable
    character(len=256) :: txt
    logical            :: L2C_flag_local

    error_flag = no_error

    ! interpret optional arugment(s)
    L2C_flag_local=.false.
    IF (present(L2C_flag)) L2C_flag_local=L2C_flag

    IF (.not. associated(L2BC_SPH)) THEN
       write(fileunit,"(a)") trim(prefix)//"[seems not associated]"
       return
    END IF

    txt = L2BC_SPH%SPH_Descriptor
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Sph_Descriptor                = "//trim(txt)
!         trim(prefix)//"%SPH_Descriptor                = "//trim(txt)

    ! Spare_1

    txt = L2BC_SPH%Num_BRC
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_BRC                       = "//trim(txt)

    txt = L2BC_SPH%Intersect_Start_Lat
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Intersect_Start_Lat           = "//trim(txt)

    txt = L2BC_SPH%Intersect_Start_Long
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Intersect_Start_Long          = "//trim(txt)

    txt = L2BC_SPH%Intersect_Stop_Lat
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Intersect_Stop_Lat            = "//trim(txt)

    txt = L2BC_SPH%Intersect_Stop_Long
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Intersect_Stop_Long           = "//trim(txt)

    ! Spare_2

    txt = L2BC_SPH%Sat_Track
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Sat_Track                     = "//trim(txt)

    txt = L2BC_SPH%M_Mie
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%M_Mie                         = "//trim(txt)

    txt = L2BC_SPH%M_Rayleigh
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%M_Rayleigh                    = "//trim(txt)

    txt = L2BC_SPH%M_Meas
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%M_Meas                        = "//trim(txt)

    txt = L2BC_SPH%Num_Valid_Mie_Profiles
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Valid_Mie_Profiles        = "//trim(txt)

    txt = L2BC_SPH%Num_Valid_Rayleigh_Profiles
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Valid_Rayleigh_Profiles   = "//trim(txt)

    txt = L2BC_SPH%Num_Clear_Mie_Profiles
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Clear_Mie_Profiles        = "//trim(txt)

    txt = L2BC_SPH%Num_Clear_Rayleigh_Profiles
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Clear_Rayleigh_Profiles   = "//trim(txt)

    txt = L2BC_SPH%Num_Cloud_Mie_Profiles
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Cloud_Mie_Profiles        = "//trim(txt)

    txt = L2BC_SPH%Num_Cloud_Rayleigh_Profiles
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Cloud_Rayleigh_Profiles   = "//trim(txt)

    txt = L2BC_SPH%Num_Mie_Prof_Warm_Pulses
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Mie_Prof_Warm_Pulses      = "//trim(txt)

    txt = L2BC_SPH%Num_Rayleigh_Prof_Warm_Pulses
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Rayleigh_Prof_Warm_Pulses = "//trim(txt)

    txt = L2BC_SPH%Num_Profiles_Surface_Mie
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Profiles_Surface_Mie      = "//trim(txt)

    txt = L2BC_SPH%Num_Profiles_Surface_Ray
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Profiles_Surface_Ray      = "//trim(txt)
    !
    txt = L2BC_SPH%Num_Valid_Obs_Prof_L1B_Mie
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Valid_Obs_Prof_L1B_Mie    = "//trim(txt)

    txt = L2BC_SPH%Num_Valid_Obs_Prof_L1B_Ray
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Valid_Obs_Prof_L1B_Ray    = "//trim(txt)

    txt = L2BC_SPH%Num_Valid_Meas_Prof_L1B_Mie
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Valid_Meas_Prof_L1B_Mie   = "//trim(txt)

    txt = L2BC_SPH%Num_Valid_Meas_Prof_L1B_Ray
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Valid_Meas_Prof_L1B_Ray   = "//trim(txt)

    txt = L2BC_SPH%Num_Valid_Obs_Bins_L1B_Mie
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Valid_Obs_Bins_L1B_Mie    = "//trim(txt)

    txt = L2BC_SPH%Num_Valid_Obs_Bins_L1B_Ray
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Valid_Obs_Bins_L1B_Ray    = "//trim(txt)

    txt = L2BC_SPH%Num_Valid_Meas_Bins_L1B_Mie
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Valid_Meas_Bins_L1B_Mie   = "//trim(txt)

    txt = L2BC_SPH%Num_Valid_Meas_Bins_L1B_Ray
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Valid_Meas_Bins_L1B_Ray   = "//trim(txt)

    txt = L2BC_SPH%Num_Invalid_Meas_Prof_L1B_Mie
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Invalid_Meas_Prof_L1B_Mie = "//trim(txt)

    txt = L2BC_SPH%Num_Invalid_Meas_Prof_L1B_Ray
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Invalid_Meas_Prof_L1B_Ray = "//trim(txt)

    txt = L2BC_SPH%Num_Invalid_Meas_Prof_L2B_Mie
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Invalid_Meas_Prof_L2B_Mie = "//trim(txt)

    txt = L2BC_SPH%Num_Invalid_Meas_Prof_L2B_Ray
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Invalid_Meas_Prof_L2B_Ray = "//trim(txt)

    txt = L2BC_SPH%Num_Invalid_Meas_Bins_L1B_Mie
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Invalid_Meas_Bins_L1B_Mie = "//trim(txt)

    txt = L2BC_SPH%Num_Invalid_Meas_Bins_L1B_Ray
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Invalid_Meas_Bins_L1B_Ray = "//trim(txt)

    txt = L2BC_SPH%Num_Invalid_Meas_Bins_L2B_Mie
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Invalid_Meas_Bins_L2B_Mie = "//trim(txt)

    txt = L2BC_SPH%Num_Invalid_Meas_Bins_L2B_Ray
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Invalid_Meas_Bins_L2B_Ray = "//trim(txt)

    ! Spare_3

    txt = L2BC_SPH%Num_Valid_Obs_L2B_Mie
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Valid_Obs_L2B_Mie         = "//trim(txt)
    
    txt = L2BC_SPH%Num_Valid_Obs_L2B_Ray
    IF (len_trim(txt).eq.0) txt = "[empty]"
    write(fileunit,"(a)") &
         trim(prefix)//"%Num_Valid_Obs_L2B_Ray         = "//trim(txt)

    IF (L2C_flag_local) THEN
       txt = L2BC_SPH%Num_Invalid_Obs_L2C_Mie
       IF (len_trim(txt).eq.0) txt = "[empty]"
       write(fileunit,"(a)") &
            trim(prefix)//"%Num_Invalid_Obs_L2C_Mie       = "//trim(txt)

       txt = L2BC_SPH%Num_Invalid_Obs_L2C_Ray
       IF (len_trim(txt).eq.0) txt = "[empty]"
       write(fileunit,"(a)") &
            trim(prefix)//"%Num_Invalid_Obs_L2C_Ray       = "//trim(txt)
    END IF

    ! Spare_4

    call AsciiWrite_DSD(L2BC_SPH%Geolocation_DSD,fileunit,&
         trim(prefix)//"%Geolocation_DSD",error_flag)
    IF (error_flag .ne. no_error) return
    call AsciiWrite_DSD(L2BC_SPH%PCD_DSD, fileunit, &
         trim(prefix)//"%PCD_DSD",error_flag)
    IF (error_flag .ne. no_error) return
    call AsciiWrite_DSD(L2BC_SPH%Mie_DSD, fileunit, &
         trim(prefix)//"%Mie_DSD",error_flag)
    IF (error_flag .ne. no_error) return
    call AsciiWrite_DSD(L2BC_SPH%Rayleigh_DSD, fileunit, &
         trim(prefix)//"%Rayleigh_DSD",error_flag)
    IF (error_flag .ne. no_error) return

    IF (L2C_flag_local) THEN
       call AsciiWrite_DSD(L2BC_SPH%Assim_PCD_DSD, fileunit, &
            trim(prefix)//"%Assim_PCD_DS",error_flag)
       IF (error_flag .ne. no_error) return
       call AsciiWrite_DSD(L2BC_SPH%Mie_VecWind_DSD, fileunit, &
            trim(prefix)//"%Mie_VecWind_DSD",error_flag)
       IF (error_flag .ne. no_error) return
       call AsciiWrite_DSD(L2BC_SPH%Rayleigh_VecWind_DSD, fileunit, &
         trim(prefix)//"%Rayleigh_VecWind_DSD",error_flag)
       IF (error_flag .ne. no_error) return
    END IF
    
    call AsciiWrite_DSD(L2BC_SPH%L1B_DSD, fileunit, &
         trim(prefix)//"%L1B_DSD",error_flag)
    IF (error_flag .ne. no_error) return
    call AsciiWrite_DSD(L2BC_SPH%Met_DSD, fileunit, &
         trim(prefix)//"%Met_DSD",error_flag)
    IF (error_flag .ne. no_error) return
    call AsciiWrite_DSD(L2BC_SPH%RBC_DSD, fileunit, &
         trim(prefix)//"%RBC_DSD",error_flag)
    IF (error_flag .ne. no_error) return
    call AsciiWrite_DSD(L2BC_SPH%CLM_DSD, fileunit, &
         trim(prefix)//"%CLM_DSD",error_flag)
    IF (error_flag .ne. no_error) return
    call AsciiWrite_DSD(L2BC_SPH%CAL_DSD, fileunit, &
         trim(prefix)//"%CAL_DSD",error_flag)
    IF (error_flag .ne. no_error) return
    call AsciiWrite_DSD(L2BC_SPH%L2B_Param_DSD, fileunit, &
         trim(prefix)//"%L2B_Param_DSD",error_flag)
    IF (error_flag .ne. no_error) return

! TODO: (possibly) consistency_sph(L2BC,L2C_flag_local)

   IF (L2C_flag_local) THEN
      call AsciiWrite_DSD(L2BC_SPH%L2B_DSD, fileunit, &
         trim(prefix)//"%L2B_DSD",error_flag)
      IF (error_flag .ne. no_error) return
      call AsciiWrite_DSD(L2BC_SPH%L2C_Param_DSD, fileunit, &
         trim(prefix)//"%L2C_Param_DSD",error_flag)
      IF (error_flag .ne. no_error) return
   ENDIF

 end subroutine AsciiWrite_L2BC_SPH
    !  #]
  subroutine Delete_L2BC_SPH(L2BC_SPH)
    !  #[

    TYPE(L2BC_SPH_type), pointer :: L2BC_SPH

    ! the list of 9 L2B-related DSD's

    IF (associated(L2BC_SPH%Geolocation_DSD)) &
         deallocate(L2BC_SPH%Geolocation_DSD)
    IF (associated(L2BC_SPH%PCD_DSD)) &
         deallocate(L2BC_SPH%PCD_DSD)
    IF (associated(L2BC_SPH%Mie_DSD)) &
         deallocate(L2BC_SPH%Mie_DSD)
    IF (associated(L2BC_SPH%Rayleigh_DSD)) &
         deallocate(L2BC_SPH%Rayleigh_DSD)

    IF (associated(L2BC_SPH%Assim_PCD_DSD)) &
         deallocate(L2BC_SPH%Assim_PCD_DSD)
    IF (associated(L2BC_SPH%Mie_VecWind_DSD)) &
         deallocate(L2BC_SPH%Mie_VecWind_DSD)
    IF (associated(L2BC_SPH%Rayleigh_VecWind_DSD)) &
         deallocate(L2BC_SPH%Rayleigh_VecWind_DSD)

    IF (associated(L2BC_SPH%L1B_DSD)) &
         deallocate(L2BC_SPH%L1B_DSD)
    IF (associated(L2BC_SPH%Met_DSD)) &
         deallocate(L2BC_SPH%Met_DSD)
    IF (associated(L2BC_SPH%RBC_DSD)) &
         deallocate(L2BC_SPH%RBC_DSD)
    IF (associated(L2BC_SPH%CLM_DSD)) &
         deallocate(L2BC_SPH%CLM_DSD)
    IF (associated(L2BC_SPH%CAL_DSD)) &
         deallocate(L2BC_SPH%CAL_DSD)
    IF (associated(L2BC_SPH%L2B_Param_DSD)) &
         deallocate(L2BC_SPH%L2B_Param_DSD)

    IF (associated(L2BC_SPH%L2B_DSD)) &
         deallocate(L2BC_SPH%L2B_DSD)
    IF (associated(L2BC_SPH%L2C_Param_DSD)) &
         deallocate(L2BC_SPH%L2C_Param_DSD)

  end subroutine Delete_L2BC_SPH
    !  #]
  subroutine Fill_L2BC_SPH_items(L2B_SPH, &
       SPH_Descriptor, Num_BRC, &
       Intersect_Start_Lat, Intersect_Start_Long, &
       Intersect_Stop_Lat, Intersect_Stop_Long, &
       Sat_Track, M_Mie0, M_Rayleigh0, M_Meas0, &
       Num_Valid_Mie_Profiles,        Num_Valid_Rayleigh_Profiles,&
       Num_Clear_Mie_Profiles,        Num_Clear_Rayleigh_Profiles,&
       Num_Cloud_Mie_Profiles,        Num_Cloud_Rayleigh_Profiles,&
       Num_Mie_Prof_Warm_Pulses,      Num_Rayleigh_Prof_Warm_Pulses,&
       Num_Profiles_Surface_Mie,      Num_Profiles_Surface_Ray,&
       Num_Valid_Obs_Prof_L1B_Mie,    Num_Valid_Obs_Prof_L1B_Ray, &
       Num_Valid_Meas_Prof_L1B_Mie,   Num_Valid_Meas_Prof_L1B_Ray,&
       Num_Valid_Obs_Bins_L1B_Mie,    Num_Valid_Obs_Bins_L1B_Ray, &
       Num_Valid_Meas_Bins_L1B_Mie,   Num_Valid_Meas_Bins_L1B_Ray,&
       Num_Invalid_Meas_Prof_L1B_Mie, Num_Invalid_Meas_Prof_L1B_Ray,&
       Num_Invalid_Meas_Prof_L2B_Mie, Num_Invalid_Meas_Prof_L2B_Ray,&
       Num_Invalid_Meas_Bins_L1B_Mie, Num_Invalid_Meas_Bins_L1B_Ray,&
       Num_Invalid_Meas_Bins_L2B_Mie, Num_Invalid_Meas_Bins_L2B_Ray,&
       Num_Valid_Obs_L2B_Mie,         Num_Valid_Obs_L2B_Ray,&
       error_flag, &
       ! the following optional flag and 2 optional items are the
       ! additional L2C SPH parameters
       L2C_flag,&
       Num_Invalid_Obs_L2C_Mie,       Num_Invalid_Obs_L2C_Ray)
    !  #[

    ! Remark: the DSD structures are not yet filled inside this routine
    ! To get the needed sizes the GetByteSize functions are used,
    ! this is done in the WriteL2Bdata module.
    ! If we would do it here, the SPH module would become dependant
    ! on the L2B_Datastructure module, which in turn is dependent on this
    ! SPH module, and this cyclic dependency cannot be properly compiled.

    TYPE(L2BC_SPH_type), pointer  :: L2B_SPH       ! filled on output
    character(len=*), intent(in)  :: SPH_Descriptor
    integer(IntAus),  intent(in)  :: Num_BRC
    integer(IntAl),   intent(in)  :: Intersect_Start_Lat
    integer(IntAl),   intent(in)  :: Intersect_Start_Long
    integer(IntAl),   intent(in)  :: Intersect_Stop_Lat
    integer(IntAl),   intent(in)  :: Intersect_Stop_Long
    real(FAdoxy),     intent(in)  :: Sat_Track
    integer(IntAuc),  intent(in)  :: M_Mie0
    integer(IntAuc),  intent(in)  :: M_Rayleigh0
    integer(IntAuc),  intent(in)  :: M_Meas0
    integer(IntAus),  intent(in)  :: Num_Valid_Mie_Profiles
    integer(IntAus),  intent(in)  :: Num_Valid_Rayleigh_Profiles
    integer(IntAus),  intent(in)  :: Num_Clear_Mie_Profiles
    integer(IntAus),  intent(in)  :: Num_Clear_Rayleigh_Profiles
    integer(IntAus),  intent(in)  :: Num_Cloud_Mie_Profiles
    integer(IntAus),  intent(in)  :: Num_Cloud_Rayleigh_Profiles
    integer(IntAus),  intent(in)  :: Num_Mie_Prof_Warm_Pulses
    integer(IntAus),  intent(in)  :: Num_Rayleigh_Prof_Warm_Pulses
    integer(IntAus),  intent(in)  :: Num_Profiles_Surface_Mie
    integer(IntAus),  intent(in)  :: Num_Profiles_Surface_Ray
    integer(IntAus),  intent(in)  :: Num_Valid_Obs_Prof_L1B_Mie
    integer(IntAus),  intent(in)  :: Num_Valid_Obs_Prof_L1B_Ray
    integer(IntAus),  intent(in)  :: Num_Valid_Meas_Prof_L1B_Mie
    integer(IntAus),  intent(in)  :: Num_Valid_Meas_Prof_L1B_Ray
    integer(IntAus),  intent(in)  :: Num_Valid_Obs_Bins_L1B_Mie
    integer(IntAus),  intent(in)  :: Num_Valid_Obs_Bins_L1B_Ray
    integer(IntAus),  intent(in)  :: Num_Valid_Meas_Bins_L1B_Mie
    integer(IntAus),  intent(in)  :: Num_Valid_Meas_Bins_L1B_Ray
    integer(IntAus),  intent(in)  :: Num_Invalid_Meas_Prof_L1B_Mie
    integer(IntAus),  intent(in)  :: Num_Invalid_Meas_Prof_L1B_Ray
    integer(IntAus),  intent(in)  :: Num_Invalid_Meas_Prof_L2B_Mie
    integer(IntAus),  intent(in)  :: Num_Invalid_Meas_Prof_L2B_Ray
    integer(IntAus),  intent(in)  :: Num_Invalid_Meas_Bins_L1B_Mie
    integer(IntAus),  intent(in)  :: Num_Invalid_Meas_Bins_L1B_Ray
    integer(IntAus),  intent(in)  :: Num_Invalid_Meas_Bins_L2B_Mie
    integer(IntAus),  intent(in)  :: Num_Invalid_Meas_Bins_L2B_Ray
    integer(IntAus),  intent(in)  :: Num_Valid_Obs_L2B_Mie
    integer(IntAus),  intent(in)  :: Num_Valid_Obs_L2B_Ray
    integer,          intent(out) :: error_flag    ! result
    ! L2C items
    integer(IntAus), optional, intent(in) :: Num_Invalid_Obs_L2C_Mie
    integer(IntAus), optional, intent(in) :: Num_Invalid_Obs_L2C_Ray
    logical,         optional, intent(in) :: L2C_flag

    ! local variables
    logical :: L2C_flag_local
    integer(IntAus) :: MaxNumL2BMeasPossible
    integer(IntAus) :: MaxNumL2BMieProfPossible
    integer(IntAus) :: MaxNumL2BRayProfPossible
    integer(IntAus) :: MaxNumL2BMieBinsPossible
    integer(IntAus) :: MaxNumL2BRayBinsPossible
    integer(IntAus) :: MaxNumL1BObsProfPossible
    integer(IntAus) :: MaxNumL1BMeasPossible
    integer(IntAus) :: MaxNumL1BObsBinsPossible
    integer(IntAus) :: MaxNumL1BMeasBinsPossible
    character(len=maxlen_IntAuc) :: txt_IntAuc
    character(len=maxlen_IntAus) :: txt_IntAus
    character(len=maxlen_FAdoxy) :: txt_FAdoxy

    ! init
    error_flag = no_error

    MaxNumL2BMeasPossible    = Num_BRC*M_Meas0
    MaxNumL2BMieProfPossible = Num_BRC*M_Mie0
    MaxNumL2BRayProfPossible = Num_BRC*M_Rayleigh0
    MaxNumL2BMieBinsPossible = Num_BRC*M_Mie0     *N_range_gates
    MaxNumL2BRayBinsPossible = Num_BRC*M_Rayleigh0*N_range_gates

    MaxNumL1BObsProfPossible  = Num_BRC
    MaxNumL1BMeasPossible     = Num_BRC*M_Meas0
    MaxNumL1BObsBinsPossible  = Num_BRC*N_range_gates
    MaxNumL1BMeasBinsPossible = Num_BRC*M_Meas0*N_range_gates

    ! interpret optional arugment(s)
    L2C_flag_local=.false.
    IF (present(L2C_flag)) L2C_flag_local=L2C_flag

    IF (.not. associated(L2B_SPH)) THEN
       print *,"ERROR in Fill_L2BC_SPH_items():"
       print *,"The pointer to this L2B_SPH seems not associated !!"
       error_flag = error_programming
       return
    END IF

    call  check_string_length(SPH_Descriptor,"Sph_Descriptor",&
         maxlen_SPH_Descriptor,"Fill_L2BC_SPH_items") 
    L2B_SPH%SPH_Descriptor = SPH_Descriptor

    IF ( (Num_BRC .lt. 0) .or. (Num_BRC .gt. 9999)) THEN
       call SetIntAus(Num_BRC,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value for Num_BRC, are you sure this is correct? "//&
            "Num_BRC = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_BRC,L2B_SPH%Num_BRC,error_flag)
    IF (error_flag .ne. no_error) return

    call SetIntAl(Intersect_Start_Lat,L2B_SPH%Intersect_Start_Lat,error_flag)
    IF (error_flag .ne. no_error) return

    call SetIntAl(Intersect_Start_Long,L2B_SPH%Intersect_Start_Long,error_flag)
    IF (error_flag .ne. no_error) return

    call SetIntAl(Intersect_Stop_Lat,L2B_SPH%Intersect_Stop_Lat,error_flag)
    IF (error_flag .ne. no_error) return

    call SetIntAl(Intersect_Stop_Long,L2B_SPH%Intersect_Stop_Long,error_flag)
    IF (error_flag .ne. no_error) return

    IF (Sat_Track .lt. -360. .or. Sat_Track .gt. 360.) THEN
       call SetFAdoFull_or_Missing(Sat_Track,txt_FAdoxy,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "Sat_Track ="//trim(txt_FAdoxy),realacc_precision=12)
    END IF
    call SetFAdo36(Sat_Track,L2B_SPH%Sat_Track,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (M_Mie0 .lt. 0) .or. &
         (M_Mie0 .gt. N_Obs_Max_Mie) ) THEN
       call SetIntAuc(M_Mie0,txt_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value M_Mie0 = "//trim(txt_IntAuc))
    END IF
    call SetIntAuc(M_Mie0,L2B_SPH%M_Mie,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (M_Rayleigh0 .lt. 0) .or. &
         (M_Rayleigh0 .gt. N_Obs_Max_Rayleigh) ) THEN
       call SetIntAuc(M_Rayleigh0,txt_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value M_Rayleigh0 = "//trim(txt_IntAuc))
    END IF
    call SetIntAuc(M_Rayleigh0,L2B_SPH%M_Rayleigh,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (M_Meas0 .lt. 0) .or. &
         (M_Meas0 .gt. N_Meas_Max) ) THEN
       call SetIntAuc(M_Meas0,txt_IntAuc,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value M_Meas0 = "//trim(txt_IntAuc))
    END IF
    call SetIntAuc(M_Meas0,L2B_SPH%M_Meas,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (Num_Valid_Mie_Profiles .lt. 0) .or. &
         (Num_Valid_Mie_Profiles .gt. MaxNumL2BMieProfPossible) ) THEN
       call SetIntAus(Num_Valid_Mie_Profiles,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Valid_Mie_Profiles = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Valid_Mie_Profiles,&
                   L2B_SPH%Num_Valid_Mie_Profiles,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (Num_Valid_Rayleigh_Profiles .lt. 0) .or. &
         (Num_Valid_Rayleigh_Profiles .gt. MaxNumL2BRayProfPossible) ) THEN
       call SetIntAus(Num_Valid_Rayleigh_Profiles,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Valid_Rayleigh_Profiles = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Valid_Rayleigh_Profiles,&
                   L2B_SPH%Num_Valid_Rayleigh_Profiles,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (Num_Clear_Mie_Profiles .lt. 0) .or. &
         (Num_Clear_Mie_Profiles .gt. MaxNumL2BMieProfPossible) ) THEN
       call SetIntAus(Num_Clear_Mie_Profiles,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Clear_Mie_Profiles = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Clear_Mie_Profiles,&
                   L2B_SPH%Num_Clear_Mie_Profiles,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (Num_Clear_Rayleigh_Profiles .lt. 0) .or. &
         (Num_Clear_Rayleigh_Profiles .gt. MaxNumL2BRayProfPossible) ) THEN
       call SetIntAus(Num_Clear_Rayleigh_Profiles,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Clear_Rayleigh_Profiles = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Clear_Rayleigh_Profiles,&
                   L2B_SPH%Num_Clear_Rayleigh_Profiles,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (Num_Cloud_Mie_Profiles .lt. 0) .or. &
         (Num_Cloud_Mie_Profiles .gt. MaxNumL2BMieProfPossible) ) THEN
       call SetIntAus(Num_Cloud_Mie_Profiles,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Cloud_Mie_Profiles = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Cloud_Mie_Profiles,&
                   L2B_SPH%Num_Cloud_Mie_Profiles,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (Num_Cloud_Rayleigh_Profiles .lt. 0) .or. &
         (Num_Cloud_Rayleigh_Profiles .gt. MaxNumL2BRayProfPossible) ) THEN
       call SetIntAus(Num_Cloud_Rayleigh_Profiles,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Cloud_Rayleigh_Profiles = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Cloud_Rayleigh_Profiles,&
                   L2B_SPH%Num_Cloud_Rayleigh_Profiles,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (Num_Mie_Prof_Warm_Pulses .lt. 0) .or. &
         (Num_Mie_Prof_Warm_Pulses .gt. MaxNumL2BMieProfPossible) ) THEN
       call SetIntAus(Num_Mie_Prof_Warm_Pulses,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Mie_Prof_Warm_Pulses = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Mie_Prof_Warm_Pulses,&
                   L2B_SPH%Num_Mie_Prof_Warm_Pulses,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (Num_Rayleigh_Prof_Warm_Pulses .lt. 0) .or. &
         (Num_Rayleigh_Prof_Warm_Pulses .gt. MaxNumL2BRayProfPossible) ) THEN
       call SetIntAus(Num_Rayleigh_Prof_Warm_Pulses,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Rayleigh_Prof_Warm_Pulses = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Rayleigh_Prof_Warm_Pulses,&
                   L2B_SPH%Num_Rayleigh_Prof_Warm_Pulses,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (Num_Profiles_Surface_Mie .lt. 0) .or. &
         (Num_Profiles_Surface_Mie .gt. MaxNumL2BMieProfPossible) ) THEN
       call SetIntAus(Num_Profiles_Surface_Mie,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Profiles_Surface_Mie = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Profiles_Surface_Mie,&
                   L2B_SPH%Num_Profiles_Surface_Mie,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (Num_Profiles_Surface_Ray .lt. 0) .or. &
         (Num_Profiles_Surface_Ray .gt. MaxNumL2BRayProfPossible) ) THEN
       call SetIntAus(Num_Profiles_Surface_Ray,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Profiles_Surface_Ray = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Profiles_Surface_Ray,&
                   L2B_SPH%Num_Profiles_Surface_Ray,error_flag)
    IF (error_flag .ne. no_error) return

    !

    IF ( (Num_Valid_Obs_Prof_L1B_Mie .lt. 0) .or. &
         (Num_Valid_Obs_Prof_L1B_Mie .gt. MaxNumL1BObsProfPossible) ) THEN
       call SetIntAus(Num_Valid_Obs_Prof_L1B_Mie,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Valid_Obs_Prof_L1B_Mie = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Valid_Obs_Prof_L1B_Mie,&
                   L2B_SPH%Num_Valid_Obs_Prof_L1B_Mie,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (Num_Valid_Obs_Prof_L1B_Ray .lt. 0) .or. &
         (Num_Valid_Obs_Prof_L1B_Ray .gt. MaxNumL1BObsProfPossible) ) THEN
       call SetIntAus(Num_Valid_Obs_Prof_L1B_Ray,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Valid_Obs_Prof_L1B_Ray = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Valid_Obs_Prof_L1B_Ray,&
                   L2B_SPH%Num_Valid_Obs_Prof_L1B_Ray,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (Num_Valid_Meas_Prof_L1B_Mie .lt. 0) .or. &
         (Num_Valid_Meas_Prof_L1B_Mie .gt. MaxNumL1BMeasPossible) ) THEN
       call SetIntAus(Num_Valid_Meas_Prof_L1B_Mie,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Valid_Meas_Prof_L1B_Mie = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Valid_Meas_Prof_L1B_Mie,&
                   L2B_SPH%Num_Valid_Meas_Prof_L1B_Mie,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (Num_Valid_Meas_Prof_L1B_Ray .lt. 0) .or. &
         (Num_Valid_Meas_Prof_L1B_Ray .gt. MaxNumL1BMeasPossible) ) THEN
       call SetIntAus(Num_Valid_Meas_Prof_L1B_Ray,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Valid_Meas_Prof_L1B_Ray = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Valid_Meas_Prof_L1B_Ray,&
                   L2B_SPH%Num_Valid_Meas_Prof_L1B_Ray,error_flag)
    IF (error_flag .ne. no_error) return

    !

    IF ( (Num_Valid_Obs_Bins_L1B_Mie .lt. 0) .or. &
         (Num_Valid_Obs_Bins_L1B_Mie .gt. MaxNumL1BObsBinsPossible) ) THEN
       call SetIntAus(Num_Valid_Obs_Bins_L1B_Mie,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Valid_Obs_Bins_L1B_Mie = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Valid_Obs_Bins_L1B_Mie,&
                   L2B_SPH%Num_Valid_Obs_Bins_L1B_Mie,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (Num_Valid_Obs_Bins_L1B_Ray .lt. 0) .or. &
         (Num_Valid_Obs_Bins_L1B_Ray .gt. MaxNumL1BObsBinsPossible) ) THEN
       call SetIntAus(Num_Valid_Obs_Bins_L1B_Ray,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Valid_Obs_Bins_L1B_Ray = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Valid_Obs_Bins_L1B_Ray,&
                   L2B_SPH%Num_Valid_Obs_Bins_L1B_Ray,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (Num_Valid_Meas_Bins_L1B_Mie .lt. 0) .or. &
         (Num_Valid_Meas_Bins_L1B_Mie .gt. MaxNumL1BMeasBinsPossible) ) THEN
       call SetIntAus(Num_Valid_Meas_Bins_L1B_Mie,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Valid_Meas_Bins_L1B_Mie = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Valid_Meas_Bins_L1B_Mie,&
                   L2B_SPH%Num_Valid_Meas_Bins_L1B_Mie,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (Num_Valid_Meas_Bins_L1B_Ray .lt. 0) .or. &
         (Num_Valid_Meas_Bins_L1B_Ray .gt. MaxNumL1BMeasBinsPossible) ) THEN
       call SetIntAus(Num_Valid_Meas_Bins_L1B_Ray,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Valid_Meas_Bins_L1B_Ray = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Valid_Meas_Bins_L1B_Ray,&
                   L2B_SPH%Num_Valid_Meas_Bins_L1B_Ray,error_flag)
    IF (error_flag .ne. no_error) return

    !

    IF ( (Num_Invalid_Meas_Prof_L1B_Mie .lt. 0) .or. &
         (Num_Invalid_Meas_Prof_L1B_Mie .gt. MaxNumL1BMeasPossible) ) THEN
       call SetIntAus(Num_Invalid_Meas_Prof_L1B_Mie,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Invalid_Meas_Prof_L1B_Mie = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Invalid_Meas_Prof_L1B_Mie,&
                   L2B_SPH%Num_Invalid_Meas_Prof_L1B_Mie,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (Num_Invalid_Meas_Prof_L1B_Ray .lt. 0) .or. &
         (Num_Invalid_Meas_Prof_L1B_Ray .gt. MaxNumL1BMeasPossible) ) THEN
       call SetIntAus(Num_Invalid_Meas_Prof_L1B_Ray,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Invalid_Meas_Prof_L1B_Ray = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Invalid_Meas_Prof_L1B_Ray,&
                   L2B_SPH%Num_Invalid_Meas_Prof_L1B_Ray,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (Num_Invalid_Meas_Prof_L2B_Mie .lt. 0) .or. &
         (Num_Invalid_Meas_Prof_L2B_Mie .gt. MaxNumL2BMeasPossible) ) THEN
       call SetIntAus(Num_Invalid_Meas_Prof_L2B_Mie,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Invalid_Meas_Prof_L2B_Mie = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Invalid_Meas_Prof_L2B_Mie,&
                   L2B_SPH%Num_Invalid_Meas_Prof_L2B_Mie,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (Num_Invalid_Meas_Prof_L2B_Ray .lt. 0) .or. &
         (Num_Invalid_Meas_Prof_L2B_Ray .gt. MaxNumL2BMeasPossible) ) THEN
       call SetIntAus(Num_Invalid_Meas_Prof_L2B_Ray,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Invalid_Meas_Prof_L2B_Ray = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Invalid_Meas_Prof_L2B_Ray,&
                   L2B_SPH%Num_Invalid_Meas_Prof_L2B_Ray,error_flag)
    IF (error_flag .ne. no_error) return

    !

    IF ( (Num_Invalid_Meas_Bins_L1B_Mie .lt. 0) .or. &
         (Num_Invalid_Meas_Bins_L1B_Mie .gt. MaxNumL1BMeasBinsPossible) ) THEN
       call SetIntAus(Num_Invalid_Meas_Bins_L1B_Mie,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Invalid_Meas_Bins_L1B_Mie = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Invalid_Meas_Bins_L1B_Mie,&
                   L2B_SPH%Num_Invalid_Meas_Bins_L1B_Mie,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (Num_Invalid_Meas_Bins_L1B_Ray .lt. 0) .or. &
         (Num_Invalid_Meas_Bins_L1B_Ray .gt. MaxNumL1BMeasBinsPossible) ) THEN
       call SetIntAus(Num_Invalid_Meas_Bins_L1B_Ray,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Invalid_Meas_Bins_L1B_Ray = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Invalid_Meas_Bins_L1B_Ray,&
                   L2B_SPH%Num_Invalid_Meas_Bins_L1B_Ray,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (Num_Invalid_Meas_Bins_L2B_Mie .lt. 0) .or. &
         (Num_Invalid_Meas_Bins_L2B_Mie .gt. MaxNumL1BMeasBinsPossible) ) THEN
       call SetIntAus(Num_Invalid_Meas_Bins_L2B_Mie,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Invalid_Meas_Bins_L2B_Mie = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Invalid_Meas_Bins_L2B_Mie,&
                   L2B_SPH%Num_Invalid_Meas_Bins_L2B_Mie,error_flag)
    IF (error_flag .ne. no_error) return

    IF ( (Num_Invalid_Meas_Bins_L2B_Ray .lt. 0) .or. &
         (Num_Invalid_Meas_Bins_L2B_Ray .gt. MaxNumL1BMeasBinsPossible) ) THEN
       call SetIntAus(Num_Invalid_Meas_Bins_L2B_Ray,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Invalid_Meas_Bins_L2B_Ray = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Invalid_Meas_Bins_L2B_Ray,&
                   L2B_SPH%Num_Invalid_Meas_Bins_L2B_Ray,error_flag)
    IF (error_flag .ne. no_error) return
    
    IF ( (Num_Valid_Obs_L2B_Mie .lt. 0) .or. &
         (Num_Valid_Obs_L2B_Mie .gt. MaxNumL2BMieBinsPossible) ) THEN
       call SetIntAus(Num_Valid_Obs_L2B_Mie,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Valid_Obs_L2B_Mie = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Valid_Obs_L2B_Mie,&
         L2B_SPH%Num_Valid_Obs_L2B_Mie,error_flag)
    IF (error_flag .ne. no_error) return
    
    IF ( (Num_Valid_Obs_L2B_Ray .lt. 0) .or. &
         (Num_Valid_Obs_L2B_Ray .gt. MaxNumL2BRayBinsPossible) ) THEN
       call SetIntAus(Num_Valid_Obs_L2B_Ray,txt_IntAus,error_flag)
       IF (error_flag .ne. no_error) return
       call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
            "unlikely value Num_Valid_Obs_L2B_Ray = "//trim(txt_IntAus))
    END IF
    call SetIntAus(Num_Valid_Obs_L2B_Ray,&
         L2B_SPH%Num_Valid_Obs_L2B_Ray,error_flag)
    IF (error_flag .ne. no_error) return
    
    IF (L2C_flag_local) THEN

       IF ( .not. present(Num_Invalid_Obs_L2C_Mie) .or. &
            .not. present(Num_Invalid_Obs_L2C_Ray)       ) THEN
          call logmsg(log_error,"ERROR in Fill_L2BC_SPH_items: "//&
               "when filling the L2C_SPH the parameters "//&
               "Num_Invalid_Obs_L2C_Mie and Num_Invalid_Obs_L2C_Ray "//&
               "should both be present, but one of them is missing in "//&
               "the call to Fill_L2BC_SPH_items()....")
          error_flag = error_programming
       END IF

       IF ( (Num_Invalid_Obs_L2C_Mie .lt. 0) .or. &
            (Num_Invalid_Obs_L2C_Mie .gt. MaxNumL2BMieBinsPossible) ) THEN
          call SetIntAus(Num_Invalid_Obs_L2C_Mie,txt_IntAus,error_flag)
          IF (error_flag .ne. no_error) return
          call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
               "unlikely value Num_Invalid_Obs_L2C_Mie = "//trim(txt_IntAus))
       END IF
       call SetIntAus(Num_Invalid_Obs_L2C_Mie,&
            L2B_SPH%Num_Invalid_Obs_L2C_Mie,error_flag)
       IF (error_flag .ne. no_error) return

       IF ( (Num_Invalid_Obs_L2C_Ray .lt. 0) .or. &
            (Num_Invalid_Obs_L2C_Ray .gt. MaxNumL2BRayBinsPossible) ) THEN
          call SetIntAus(Num_Invalid_Obs_L2C_Ray,txt_IntAus,error_flag)
          IF (error_flag .ne. no_error) return
          call logmsg(log_warn,"WARNING in Fill_L2BC_SPH_items(): "//&
               "unlikely value Num_Invalid_Obs_L2C_Ray = "//trim(txt_IntAus))
       END IF
       call SetIntAus(Num_Invalid_Obs_L2C_Ray,&
            L2B_SPH%Num_Invalid_Obs_L2C_Ray,error_flag)
       IF (error_flag .ne. no_error) return
    
    END IF
    
  end subroutine Fill_L2BC_SPH_items
    !  #]
  subroutine Copy_L2BC_SPH_to_xml_tree(L2BC_SPH, path_in, &
                                       xml_tree, error_flag, L2C_flag)
    !  #[
    TYPE(L2BC_SPH_type), pointer     :: L2BC_SPH    ! input
    character(len=*),    intent(in)  :: path_in
    TYPE(xml_node_type), pointer     :: xml_tree    ! points to updated struct
    integer,             intent(out) :: error_flag  ! result
    logical,   optional, intent(in)  :: L2C_flag

    ! local variables
    character(len=max_str_len) :: path
    character(len=max_str_len) :: path_dsd
    integer                    :: which_DSD
    character(len=max_str_len) :: dsd_count_string
    logical     :: L2C_flag_local

    ! interpret optional arugment(s)
    L2C_flag_local=.false.
    IF (present(L2C_flag)) L2C_flag_local=L2C_flag

    error_flag = no_error

    ! add the <Specific_Product_Header></Specific_Product_Header> tags
    path = trim(path_in)//"/"//ee_vh_name
    call xml_tree_add_child(xml_tree,path,ee_sph_name,error_flag)
    IF (error_flag .ne. no_error) return

    ! fill all items inside the L2BC_SPH
    call xml_add_SPH_item(xml_tree, path_in, "Sph_Descriptor",&
                          L2BC_SPH%SPH_Descriptor, error_flag)
    IF (error_flag .ne. no_error) return

    ! Spare_1
    call xml_add_SPH_item(xml_tree, path_in, "Spare_1",newline,error_flag)
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_BRC",&
                          Remove_Plus_Sign(L2BC_SPH%Num_BRC),error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Intersect_Start_Lat",&
                          Remove_Plus_Sign(L2BC_SPH%Intersect_Start_Lat),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Intersect_Start_Long",&
                          Remove_Plus_Sign(L2BC_SPH%Intersect_Start_Long),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Intersect_Stop_Lat",&
                          Remove_Plus_Sign(L2BC_SPH%Intersect_Stop_Lat),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Intersect_Stop_Long",&
                          Remove_Plus_Sign(L2BC_SPH%Intersect_Stop_Long),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Sat_Track",&
                          Remove_Plus_Sign(L2BC_SPH%Sat_Track),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    ! Spare_2
    call xml_add_SPH_item(xml_tree, path_in, "Spare_2",&
                          newline,error_flag)
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "M_Mie",&
                          Remove_Plus_Sign(L2BC_SPH%M_Mie),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "M_Rayleigh",&
                          Remove_Plus_Sign(L2BC_SPH%M_Rayleigh),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "M_Meas",&
                          Remove_Plus_Sign(L2BC_SPH%M_Meas),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Valid_Mie_Profiles",&
                          Remove_Plus_Sign(L2BC_SPH%Num_Valid_Mie_Profiles),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Valid_Rayleigh_Profiles",&
                      Remove_Plus_Sign(L2BC_SPH%Num_Valid_Rayleigh_Profiles),&
                      error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Clear_Mie_Profiles",&
                          Remove_Plus_Sign(L2BC_SPH%Num_Clear_Mie_Profiles),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Clear_Rayleigh_Profiles",&
                       Remove_Plus_Sign(L2BC_SPH%Num_Clear_Rayleigh_Profiles),&
                       error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Cloud_Mie_Profiles",&
                          Remove_Plus_Sign(L2BC_SPH%Num_Cloud_Mie_Profiles),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Cloud_Rayleigh_Profiles",&
                       Remove_Plus_Sign(L2BC_SPH%Num_Cloud_Rayleigh_Profiles),&
                       error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Mie_Prof_Warm_Pulses",&
                          Remove_Plus_Sign(L2BC_SPH%Num_Mie_Prof_Warm_Pulses),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Rayleigh_Prof_Warm_Pulses",&
                     Remove_Plus_Sign(L2BC_SPH%Num_Rayleigh_Prof_Warm_Pulses),&
                     error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Profiles_Surface_Mie",&
                          Remove_Plus_Sign(L2BC_SPH%Num_Profiles_Surface_Mie),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Profiles_Surface_Ray",&
                          Remove_Plus_Sign(L2BC_SPH%Num_Profiles_Surface_Ray),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    !
    call xml_add_SPH_item(xml_tree, path_in, "Num_Valid_Obs_Prof_L1B_Mie",&
                          Remove_Plus_Sign(L2BC_SPH%Num_Valid_Obs_Prof_L1B_Mie),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Valid_Obs_Prof_L1B_Ray",&
                          Remove_Plus_Sign(L2BC_SPH%Num_Valid_Obs_Prof_L1B_Ray),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Valid_Meas_Prof_L1B_Mie",&
                          Remove_Plus_Sign(L2BC_SPH%Num_Valid_Meas_Prof_L1B_Mie),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Valid_Meas_Prof_L1B_Ray",&
                          Remove_Plus_Sign(L2BC_SPH%Num_Valid_Meas_Prof_L1B_Ray),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    !
    call xml_add_SPH_item(xml_tree, path_in, "Num_Valid_Obs_Bins_L1B_Mie",&
                          Remove_Plus_Sign(L2BC_SPH%Num_Valid_Obs_Bins_L1B_Mie),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Valid_Obs_Bins_L1B_Ray",&
                          Remove_Plus_Sign(L2BC_SPH%Num_Valid_Obs_Bins_L1B_Ray),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Valid_Meas_Bins_L1B_Mie",&
                          Remove_Plus_Sign(L2BC_SPH%Num_Valid_Meas_Bins_L1B_Mie),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Valid_Meas_Bins_L1B_Ray",&
                          Remove_Plus_Sign(L2BC_SPH%Num_Valid_Meas_Bins_L1B_Ray),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    !
    call xml_add_SPH_item(xml_tree, path_in, "Num_Invalid_Meas_Prof_L1B_Mie",&
                          Remove_Plus_Sign(L2BC_SPH%Num_Invalid_Meas_Prof_L1B_Mie),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Invalid_Meas_Prof_L1B_Ray",&
                          Remove_Plus_Sign(L2BC_SPH%Num_Invalid_Meas_Prof_L1B_Ray),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Invalid_Meas_Prof_L2B_Mie",&
                          Remove_Plus_Sign(L2BC_SPH%Num_Invalid_Meas_Prof_L2B_Mie),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Invalid_Meas_Prof_L2B_Ray",&
                          Remove_Plus_Sign(L2BC_SPH%Num_Invalid_Meas_Prof_L2B_Ray),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    !
    call xml_add_SPH_item(xml_tree, path_in, "Num_Invalid_Meas_Bins_L1B_Mie",&
                          Remove_Plus_Sign(L2BC_SPH%Num_Invalid_Meas_Bins_L1B_Mie),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Invalid_Meas_Bins_L1B_Ray",&
                          Remove_Plus_Sign(L2BC_SPH%Num_Invalid_Meas_Bins_L1B_Ray),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Invalid_Meas_Bins_L2B_Mie",&
                          Remove_Plus_Sign(L2BC_SPH%Num_Invalid_Meas_Bins_L2B_Mie),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Invalid_Meas_Bins_L2B_Ray",&
                          Remove_Plus_Sign(L2BC_SPH%Num_Invalid_Meas_Bins_L2B_Ray),&
                          error_flag )
    IF (error_flag .ne. no_error) return

    ! Spare_3
    call xml_add_SPH_item(xml_tree, path_in, "Spare_3",&
                          newline,error_flag)
    IF (error_flag .ne. no_error) return

    call xml_add_SPH_item(xml_tree, path_in, "Num_Valid_Obs_L2B_Mie",&
                          Remove_Plus_Sign(L2BC_SPH%Num_Valid_Obs_L2B_Mie),error_flag )
    IF (error_flag .ne. no_error) return
       
    call xml_add_SPH_item(xml_tree, path_in, "Num_Valid_Obs_L2B_Ray",&
                          Remove_Plus_Sign(L2BC_SPH%Num_Valid_Obs_L2B_Ray),error_flag )
    IF (error_flag .ne. no_error) return
       
    IF (L2C_flag_local) THEN
       call xml_add_SPH_item(xml_tree, path_in, "Num_Invalid_Obs_L2C_Mie",&
                             Remove_Plus_Sign(L2BC_SPH%Num_Invalid_Obs_L2C_Mie),&
                             error_flag )
       IF (error_flag .ne. no_error) return
       
       call xml_add_SPH_item(xml_tree, path_in, "Num_Invalid_Obs_L2C_Ray",&
                             Remove_Plus_Sign(L2BC_SPH%Num_Invalid_Obs_L2C_Ray),&
                             error_flag )
       IF (error_flag .ne. no_error) return
       
    END IF

    ! Spare_4
    call xml_add_SPH_item(xml_tree, path_in, "Spare_4",newline,error_flag)
    IF (error_flag .ne. no_error) return

    ! add the <List_of_DSDs></List_of_DSDs> tags
    path = trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name
    call xml_tree_add_child(xml_tree,path,"List_of_Dsds",error_flag)
    IF (error_flag .ne. no_error) return

    IF (L2C_flag_local) THEN
       dsd_count_string=Num_DSD_string_L2C
    ELSE
       dsd_count_string=Num_DSD_string_L2B
    END IF
    path_dsd = trim(path)//"/List_of_Dsds"
    call xml_tree_add_and_set_attribute(xml_tree,path_dsd,"count",&
                                        trim(dsd_count_string),&
                                        error_flag)
    which_DSD = 0

    which_DSD = which_DSD + 1
    call Add_DSD_to_xml_SPH(xml_tree, path_in,L2BC_SPH%Geolocation_DSD, which_DSD, &
                            error_flag)
    IF (error_flag .ne. no_error) return

    which_DSD = which_DSD + 1
    call Add_DSD_to_xml_SPH(xml_tree, path_in, L2BC_SPH%PCD_DSD, which_DSD, error_flag)
    IF (error_flag .ne. no_error) return

    which_DSD = which_DSD + 1
    call Add_DSD_to_xml_SPH(xml_tree, path_in, L2BC_SPH%Mie_DSD, which_DSD, error_flag)
    IF (error_flag .ne. no_error) return

    which_DSD = which_DSD + 1
    call Add_DSD_to_xml_SPH(xml_tree, path_in, L2BC_SPH%Rayleigh_DSD, which_DSD, &
                            error_flag)
    IF (error_flag .ne. no_error) return

    IF (L2C_flag_local) THEN
       which_DSD = which_DSD + 1
       call Add_DSD_to_xml_SPH(xml_tree, path_in, L2BC_SPH%Assim_PCD_DSD, which_DSD, &
                               error_flag)
       IF (error_flag .ne. no_error) return

       which_DSD = which_DSD + 1
       call Add_DSD_to_xml_SPH(xml_tree, path_in, L2BC_SPH%Mie_VecWind_DSD, which_DSD, &
                               error_flag)
       IF (error_flag .ne. no_error) return

       which_DSD = which_DSD + 1
       call Add_DSD_to_xml_SPH(xml_tree, path_in, L2BC_SPH%Rayleigh_VecWind_DSD, &
                               which_DSD, error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    which_DSD = which_DSD + 1
    call Add_DSD_to_xml_SPH(xml_tree, path_in, L2BC_SPH%L1B_DSD, which_DSD, error_flag)
    IF (error_flag .ne. no_error) return

    which_DSD = which_DSD + 1
    call Add_DSD_to_xml_SPH(xml_tree, path_in, L2BC_SPH%Met_DSD, which_DSD, error_flag)
    IF (error_flag .ne. no_error) return

    which_DSD = which_DSD + 1
    call Add_DSD_to_xml_SPH(xml_tree, path_in, L2BC_SPH%RBC_DSD, which_DSD, error_flag)
    IF (error_flag .ne. no_error) return

    which_DSD = which_DSD + 1
    call Add_DSD_to_xml_SPH(xml_tree, path_in, L2BC_SPH%CLM_DSD, which_DSD, error_flag)
    IF (error_flag .ne. no_error) return

    which_DSD = which_DSD + 1
    call Add_DSD_to_xml_SPH(xml_tree, path_in, L2BC_SPH%CAL_DSD, which_DSD, error_flag)
    IF (error_flag .ne. no_error) return

    which_DSD = which_DSD + 1
    call Add_DSD_to_xml_SPH(xml_tree, path_in, L2BC_SPH%L2B_Param_DSD, which_DSD, &
                            error_flag)
    IF (error_flag .ne. no_error) return

    ! TODO: consistency_DSDs(L2BC_SPH,L2C_flag_local)

    IF (L2C_flag_local) THEN
       which_DSD = which_DSD + 1
       call Add_DSD_to_xml_SPH(xml_tree, path_in, L2BC_SPH%L2B_DSD, which_DSD, &
                               error_flag)
       IF (error_flag .ne. no_error) return

       which_DSD = which_DSD + 1
       call Add_DSD_to_xml_SPH(xml_tree, path_in, L2BC_SPH%L2C_Param_DSD, which_DSD, &
                               error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    return

  end subroutine Copy_L2BC_SPH_to_xml_tree
    !  #]
  subroutine Get_L2BC_SPH_from_xml_tree(xml_tree, path_in, &
                                        L2BC_SPH, error_flag, L2C_flag)
    !  #[
    TYPE(xml_node_type), pointer     :: xml_tree   ! input
    character(len=*),    intent(in)  :: path_in    ! input
    TYPE(L2BC_SPH_type), pointer     :: L2BC_SPH   ! result
    integer,             intent(out) :: error_flag ! output
    logical,  optional,  intent(in)  :: L2C_flag   ! switch (opt. input)

    ! local variables
    character(len=max_str_len) :: path
    character(len=max_str_len) :: text
    integer                    :: AllocStatus, which_DSD
    logical                    :: L2C_flag_local

    ! init
    error_flag = no_error

    ! interpret optional arugment(s)
    L2C_flag_local=.false.
    IF (present(L2C_flag)) L2C_flag_local=L2C_flag

    IF (.not. associated(L2BC_SPH)) THEN
       print *,"ERROR in Get_L2BC_SPH_from_xml_tree():"
       print *,"the pointer L2BC_SPH seems not associated !!!"
       error_flag = error_programming
       return
    END IF

    IF (.not. associated(xml_tree)) THEN
       print *,"ERROR in Get_L2BC_SPH_from_xml_tree():"
       print *,"the pointer xml_tree seems not associated !!!"
       error_flag = error_programming
       return
    END IF

    !----------------------------------
    ! copying the L2BC_SPH XML elements
    !----------------------------------

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//"/Sph_Descriptor"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%SPH_Descriptor = &
         check_length_xml_item(text,path,maxlen_SPH_Descriptor,&
                               "Get_L2BC_SPH_from_xml_tree")

    ! Spare_1: no need to read this one

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//"/Num_BRC"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_BRC = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Intersect_Start_Lat"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Intersect_Start_Lat = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L2BC_SPH_from_xml_tree")
    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Intersect_Start_Long"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Intersect_Start_Long = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Intersect_Stop_Lat"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Intersect_Stop_Lat = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Intersect_Stop_Long"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Intersect_Stop_Long = &
         check_length_xml_item(text,path,maxlen_IntAl,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Sat_Track"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Sat_Track = &
         check_length_xml_item(text,path,maxlen_FAdo36,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/M_Mie"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%M_Mie = &
         check_length_xml_item(text,path,maxlen_IntAuc,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/M_Meas"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%M_Meas = &
         check_length_xml_item(text,path,maxlen_IntAuc,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/M_Rayleigh"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%M_Rayleigh = &
         check_length_xml_item(text,path,maxlen_IntAuc,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Valid_Mie_Profiles"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Valid_Mie_Profiles = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Valid_Rayleigh_Profiles"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Valid_Rayleigh_Profiles = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Clear_Mie_Profiles"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Clear_Mie_Profiles = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Clear_Rayleigh_Profiles"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Clear_Rayleigh_Profiles = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Cloud_Mie_Profiles"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Cloud_Mie_Profiles = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Cloud_Rayleigh_Profiles"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Cloud_Rayleigh_Profiles = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Mie_Prof_Warm_Pulses"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Mie_Prof_Warm_Pulses = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Rayleigh_Prof_Warm_Pulses"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Rayleigh_Prof_Warm_Pulses = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Profiles_Surface_Mie"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Profiles_Surface_Mie = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Profiles_Surface_Ray"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Profiles_Surface_Ray = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Valid_Obs_Prof_L1B_Mie"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Valid_Obs_Prof_L1B_Mie = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Valid_Obs_Prof_L1B_Ray"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Valid_Obs_Prof_L1B_Ray = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Valid_Meas_Prof_L1B_Mie"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Valid_Meas_Prof_L1B_Mie = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Valid_Meas_Prof_L1B_Ray"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Valid_Meas_Prof_L1B_Ray = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Valid_Obs_Bins_L1B_Mie"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Valid_Obs_Bins_L1B_Mie = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Valid_Obs_Bins_L1B_Ray"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Valid_Obs_Bins_L1B_Ray = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Valid_Meas_Bins_L1B_Mie"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Valid_Meas_Bins_L1B_Mie = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Valid_Meas_Bins_L1B_Ray"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Valid_Meas_Bins_L1B_Ray = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Invalid_Meas_Prof_L1B_Mie"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Invalid_Meas_Prof_L1B_Mie = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Invalid_Meas_Prof_L1B_Ray"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Invalid_Meas_Prof_L1B_Ray = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Invalid_Meas_Prof_L2B_Mie"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Invalid_Meas_Prof_L2B_Mie = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Invalid_Meas_Prof_L2B_Ray"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Invalid_Meas_Prof_L2B_Ray = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Invalid_Meas_Bins_L1B_Mie"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Invalid_Meas_Bins_L1B_Mie = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Invalid_Meas_Bins_L1B_Ray"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Invalid_Meas_Bins_L1B_Ray = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Invalid_Meas_Bins_L2B_Mie"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Invalid_Meas_Bins_L2B_Mie= &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Invalid_Meas_Bins_L2B_Ray"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Invalid_Meas_Bins_L2B_Ray = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    ! Spare_3: no need to read this one

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Valid_Obs_L2B_Mie"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Valid_Obs_L2B_Mie = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
         "/Num_Valid_Obs_L2B_Ray"
    call xml_node_get_string(xml_tree,path,text,error_flag)
    IF (error_flag .ne. no_error) return
    L2BC_SPH%Num_Valid_Obs_L2B_Ray = &
         check_length_xml_item(text,path,maxlen_IntAus,&
                               "Get_L2BC_SPH_from_xml_tree")

    IF (L2C_flag_local) THEN
       path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
            "/Num_Invalid_Obs_L2C_Mie"
       call xml_node_get_string(xml_tree,path,text,error_flag)
       IF (error_flag .ne. no_error) return
       L2BC_SPH%Num_Invalid_Obs_L2C_Mie = &
            check_length_xml_item(text,path,maxlen_IntAus,&
                                  "Get_L2BC_SPH_from_xml_tree")

       path  =trim(path_in)//"/"//ee_vh_name//"/"//ee_sph_name//&
            "/Num_Invalid_Obs_L2C_Ray"
       call xml_node_get_string(xml_tree,path,text,error_flag)
       IF (error_flag .ne. no_error) return
       L2BC_SPH%Num_Invalid_Obs_L2C_Ray = &
            check_length_xml_item(text,path,maxlen_IntAus,&
                                  "Get_L2BC_SPH_from_xml_tree")

       ! Spare_4: no need to read this one
    END IF

    which_DSD = 0
    !  #[ DSD 1 Geolocation_DSD

    which_DSD = which_DSD + 1
    IF (.not.associated(L2BC_SPH%Geolocation_DSD)) THEN
       allocate(L2BC_SPH%Geolocation_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L2BC_SPH%Geolocation_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L2BC_SPH%Geolocation_DSD,which_DSD,&
                               L2B_DS_Name_Geolocation_ADS,&
                               L2B_DS_Type_Geolocation_ADS,error_flag)
    IF (error_flag .ne. no_error) return

    !  #]
    !  #[ DSD 2 PCD_DSD
    which_DSD = which_DSD + 1
    IF (.not.associated(L2BC_SPH%PCD_DSD)) THEN
       allocate(L2BC_SPH%PCD_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L2BC_SPH%PCD_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L2BC_SPH%PCD_DSD,which_DSD,&
                               L2B_DS_Name_PCD_ADS, &
                               L2B_DS_Type_PCD_ADS, error_flag)
    IF (error_flag .ne. no_error) return
    !  #]
    !  #[ DSD 3 Mie_DSD
    which_DSD = which_DSD + 1
    IF (.not.associated(L2BC_SPH%Mie_DSD)) THEN
       allocate(L2BC_SPH%Mie_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L2BC_SPH%Mie_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L2BC_SPH%Mie_DSD,which_DSD,&
                               L2B_DS_Name_Mie_MDS,&
                               L2B_DS_Type_Mie_MDS, error_flag)
    IF (error_flag .ne. no_error) return
    !  #]
    !  #[ DSD 4 Rayleigh_DSD
    which_DSD = which_DSD + 1
    IF (.not.associated(L2BC_SPH%Rayleigh_DSD)) THEN
       allocate(L2BC_SPH%Rayleigh_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L2BC_SPH%Rayleigh_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L2BC_SPH%Rayleigh_DSD,which_DSD,&
                               L2B_DS_Name_Rayleigh_MDS, &
                               L2B_DS_Type_Rayleigh_MDS, error_flag)
    IF (error_flag .ne. no_error) return
    !  #]

    IF (L2C_flag_local) THEN
       !  #[ DSD 5 Assim_PCD_DSD
       which_DSD = which_DSD + 1
       IF (.not.associated(L2BC_SPH%Assim_PCD_DSD)) THEN
          allocate(L2BC_SPH%Assim_PCD_DSD,STAT=AllocStatus)
          IF (AllocStatus .ne. 0) THEN
             error_flag = error_allocate
             return
          END IF
       
          call Init_DSD(L2BC_SPH%Assim_PCD_DSD,error_flag)
          IF (error_flag .ne. no_error) return
       END IF
       
       ! get the DSD from the xml_tree structure
       call Get_DSD_from_xml_tree(xml_tree,path_in,&
                                  L2BC_SPH%Assim_PCD_DSD,which_DSD,&
                                  L2C_DS_Name_AssimPCD_ADS,&
                                  L2C_DS_Type_AssimPCD_ADS,error_flag)
       IF (error_flag .ne. no_error) return
       !  #]
       !  #[ DSD 6 Mie_VecWind_DSD
       which_DSD = which_DSD + 1
       IF (.not.associated(L2BC_SPH%Mie_VecWind_DSD)) THEN
          allocate(L2BC_SPH%Mie_VecWind_DSD,STAT=AllocStatus)
          IF (AllocStatus .ne. 0) THEN
             error_flag = error_allocate
             return
          END IF
       
          call Init_DSD(L2BC_SPH%Mie_VecWind_DSD,error_flag)
          IF (error_flag .ne. no_error) return
       END IF
       
       ! get the DSD from the xml_tree structure
       call Get_DSD_from_xml_tree(xml_tree,path_in,&
                                 L2BC_SPH%Mie_VecWind_DSD,which_DSD,&
                                 L2C_DS_Name_MieVec_MDS,&
                                 L2C_DS_Type_MieVec_MDS, error_flag)
       IF (error_flag .ne. no_error) return
       !  #]
       !  #[ DSD 7 Rayleigh_VecWind_DSD
       which_DSD = which_DSD + 1
       IF (.not.associated(L2BC_SPH%Rayleigh_VecWind_DSD)) THEN
          allocate(L2BC_SPH%Rayleigh_VecWind_DSD,STAT=AllocStatus)
          IF (AllocStatus .ne. 0) THEN
             error_flag = error_allocate
             return
          END IF
       
          call Init_DSD(L2BC_SPH%Rayleigh_VecWind_DSD,error_flag)
          IF (error_flag .ne. no_error) return
       END IF
       
       ! get the DSD from the xml_tree structure
       call Get_DSD_from_xml_tree(xml_tree,path_in, &
                                  L2BC_SPH%Rayleigh_VecWind_DSD,&
                                  which_DSD,&
                                  L2C_DS_Name_RayleighVec_MDS, &
                                  L2C_DS_Type_RayleighVec_MDS, error_flag)
       IF (error_flag .ne. no_error) return
       !  #]
    END IF

    !  #[ DSD  5 or  8 L1B_DSD
    which_DSD = which_DSD + 1
    IF (.not.associated(L2BC_SPH%L1B_DSD)) THEN
       allocate(L2BC_SPH%L1B_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L2BC_SPH%L1B_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L2BC_SPH%L1B_DSD,which_DSD,&
                               L2B_DS_Name_L1B_Product, &
                               L2B_DS_Type_L1B_Product, error_flag)
    IF (error_flag .ne. no_error) return
    !  #]
    !  #[ DSD  6 or  9 Met_DSD
    which_DSD = which_DSD + 1
    IF (.not.associated(L2BC_SPH%Met_DSD)) THEN
       allocate(L2BC_SPH%Met_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L2BC_SPH%Met_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L2BC_SPH%Met_DSD,which_DSD,&
                               L2B_DS_Name_Met_MDS,&
                               L2B_DS_Type_Met_MDS, error_flag)
    IF (error_flag .ne. no_error) return
    !  #]
    !  #[ DSD  7 or 10 RBC_DSD
    which_DSD = which_DSD + 1
    IF (.not.associated(L2BC_SPH%RBC_DSD)) THEN
       allocate(L2BC_SPH%RBC_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L2BC_SPH%RBC_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L2BC_SPH%RBC_DSD,which_DSD,&
                               L2B_DS_Name_RBC_Product, &
                               L2B_DS_Type_RBC_Product, error_flag)
    IF (error_flag .ne. no_error) return
    !  #]
    !  #[ DSD  8 or 11 CLM_DSD
    which_DSD = which_DSD + 1
    IF (.not.associated(L2BC_SPH%CLM_DSD)) THEN
       allocate(L2BC_SPH%CLM_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L2BC_SPH%CLM_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L2BC_SPH%CLM_DSD,which_DSD,&
                               L2B_DS_Name_CLM_Product, &
                               L2B_DS_Type_CLM_Product, error_flag)
    IF (error_flag .ne. no_error) return
    !  #]
    !  #[ DSD  9 or 12 CLM_DSD
    which_DSD = which_DSD + 1
    IF (.not.associated(L2BC_SPH%CAL_DSD)) THEN
       allocate(L2BC_SPH%CAL_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L2BC_SPH%CAL_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L2BC_SPH%CAL_DSD,which_DSD,&
                               L2B_DS_Name_CAL_Product, &
                               L2B_DS_Type_CAL_Product, error_flag)
    IF (error_flag .ne. no_error) return
    !  #]
    !  #[ DSD 10 or 13 L2B_Param_DSD
    which_DSD = which_DSD + 1
    IF (.not.associated(L2BC_SPH%L2B_Param_DSD)) THEN
       allocate(L2BC_SPH%L2B_Param_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          error_flag = error_allocate
          return
       END IF

       call Init_DSD(L2BC_SPH%L2B_Param_DSD,error_flag)
       IF (error_flag .ne. no_error) return
    END IF

    ! get the DSD from the xml_tree structure
    call Get_DSD_from_xml_tree(xml_tree,path_in,&
                               L2BC_SPH%L2B_Param_DSD,which_DSD,&
                               L2B_DS_Name_L2B_Param_ADS, &
                               L2B_DS_Type_L2B_Param_ADS, error_flag)
    IF (error_flag .ne. no_error) return
    !  #]

    IF (L2C_flag_local) THEN
       !  #[ DSD 14 L2B_DSD
       which_DSD = which_DSD + 1
       IF (.not.associated(L2BC_SPH%L2B_DSD)) THEN
          allocate(L2BC_SPH%L2B_DSD,STAT=AllocStatus)
          IF (AllocStatus .ne. 0) THEN
             error_flag = error_allocate
             return
          END IF
       
          call Init_DSD(L2BC_SPH%L2B_DSD,error_flag)
          IF (error_flag .ne. no_error) return
       END IF
       
       ! get the DSD from the xml_tree structure
       call Get_DSD_from_xml_tree(xml_tree,path_in,L2BC_SPH%L2B_DSD,which_DSD,&
                                  L2C_DS_Name_L2B_Product,&
                                  L2C_DS_Type_L2B_Product, error_flag)
       IF (error_flag .ne. no_error) return
       !  #]
       !  #[ DSD 15 L2C_Param_DSD
       which_DSD = which_DSD + 1
       IF (.not.associated(L2BC_SPH%L2C_Param_DSD)) THEN
          allocate(L2BC_SPH%L2C_Param_DSD,STAT=AllocStatus)
          IF (AllocStatus .ne. 0) THEN
             error_flag = error_allocate
             return
          END IF
       
          call Init_DSD(L2BC_SPH%L2C_Param_DSD,error_flag)
          IF (error_flag .ne. no_error) return
       END IF
       
       ! get the DSD from the xml_tree structure
       call Get_DSD_from_xml_tree(xml_tree,path_in,&
                                  L2BC_SPH%L2C_Param_DSD,which_DSD,&
                                  L2C_DS_Name_L2C_Param_ADS, &
                                  L2C_DS_Type_L2C_Param_ADS, error_flag)
       IF (error_flag .ne. no_error) return
       !  #]
    END IF

  end subroutine Get_L2BC_SPH_from_xml_tree
    !  #]
  !---------------------------------
  subroutine Write_L2BC_SPH_to_KVT(L2BC_SPH, fileunit, error_flag, L2C_flag)
    !  #[
    ! Writes the L2B/L2C Specific Product Header (KVT values).
    ! 28-11-2005   P. Poli  Original code. From TN2.1 v1.2 dated 06-09-2005

    TYPE(L2BC_SPH_type), pointer   :: L2BC_SPH
    integer,             intent(in)  :: fileunit
    integer,             intent(out) :: error_flag
    logical, optional,   intent(in)  :: L2C_flag

    ! local variables
    character(len=64) :: varname
    character(len=64) :: value
    character(len=64) :: phys_unit
    integer           :: max_value_length,begpos,endpos
    integer(IntAl)    :: size_SPH_calc
    logical           :: L2C_flag_local

    ! interpret optional arugment(s)
    L2C_flag_local=.false.
    IF (present(L2C_flag)) L2C_flag_local=L2C_flag

    !---------------------------------
    ! Writing the L2B SPH KVT elements
    !---------------------------------

    error_flag = no_error
    IF (.not. associated(L2BC_SPH)) THEN
       print *,"ERROR in Write_L2BC_SPH_to_KVT():"
       print *,"The pointer to this SPH seems not associated !!"
       error_flag = error_programming
       return
    END IF

    ! save start position
    inquire(unit=fileunit,nextrec=begpos)

       !  #[ SPH_Descriptor 
       varname="Sph_Descriptor"
       value=L2BC_SPH%SPH_Descriptor
       max_value_length=maxlen_SPH_Descriptor
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag,&
                             use_quotes=.true.)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Spare_1
       ! Spare_1: this is a spare value without content
       max_value_length = 40
       call write_spare_kvt_string(fileunit,max_value_length,error_flag)
       IF (error_flag .ne. no_error) return
       !  #]
       !  #[ Num_BRC
       varname="Num_BRC"
       value=L2BC_SPH%Num_BRC
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Intersect_Start_Lat
       varname="Intersect_Start_Lat"
       value=L2BC_SPH%Intersect_Start_Lat
       max_value_length=maxlen_IntAl  ! =11
       phys_unit="10-6DegN"
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag, &
                             phys_unit=phys_unit)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Intersect_Start_Long
       varname="Intersect_Start_Long"
       value=L2BC_SPH%Intersect_Start_Long
       max_value_length=maxlen_IntAl  ! =11
       phys_unit="10-6DegE"
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag, &
       phys_unit=phys_unit)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Intersect_Stop_Lat
       varname="Intersect_Stop_Lat"
       value=L2BC_SPH%Intersect_Stop_Lat
       max_value_length=maxlen_IntAl  ! =11
       phys_unit="10-6DegN"
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag, &
                             phys_unit=phys_unit )
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Intersect_Stop_Long
       varname="Intersect_Stop_Long"
       value=L2BC_SPH%Intersect_Stop_Long
       max_value_length=maxlen_IntAl  ! =11
       phys_unit="10-6DegE"
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag, &
                             phys_unit=phys_unit)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Sat_Track

       varname="Sat_Track"
       value=L2BC_SPH%Sat_Track
       max_value_length=maxlen_FAdo36  ! =11
       phys_unit="deg"
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag, &
                             phys_unit=phys_unit)
       IF (error_flag .ne. no_error) goto 998

       !  #]
       !  #[ Spare_2: 
       ! Spare_2: this is a spare value without content
       max_value_length = 40
       call write_spare_kvt_string(fileunit,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ M_Mie
       varname="M_Mie"
       value=L2BC_SPH%M_Mie
       max_value_length=maxlen_IntAuc
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ M_Rayleigh
       varname="M_Rayleigh"
       value=L2BC_SPH%M_Rayleigh
       max_value_length=maxlen_IntAuc
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ M_Meas
       varname="M_Meas"
       value=L2BC_SPH%M_Meas
       max_value_length=maxlen_IntAuc
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Valid_Mie_Profiles
       varname="Num_Valid_Mie_Profiles"
       value=L2BC_SPH%Num_Valid_Mie_Profiles
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Valid_Rayleigh_Profiles
       varname="Num_Valid_Rayleigh_Profiles"
       value=L2BC_SPH%Num_Valid_Rayleigh_Profiles
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Clear_Mie_Profiles
       varname="Num_Clear_Mie_Profiles"
       value=L2BC_SPH%Num_Clear_Mie_Profiles
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Clear_Rayleigh_Profiles
       varname="Num_Clear_Rayleigh_Profiles"
       value=L2BC_SPH%Num_Clear_Rayleigh_Profiles
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Cloud_Mie_Profiles
       varname="Num_Cloud_Mie_Profiles"
       value=L2BC_SPH%Num_Cloud_Mie_Profiles
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Cloud_Rayleigh_Profiles
       varname="Num_Cloud_Rayleigh_Profiles"
       value=L2BC_SPH%Num_Cloud_Rayleigh_Profiles
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Mie_Prof_Warm_Pulses
       varname="Num_Mie_Prof_Warm_Pulses"
       value=L2BC_SPH%Num_Mie_Prof_Warm_Pulses
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Rayleigh_Prof_Warm_Pulses
       varname="Num_Rayleigh_Prof_Warm_Pulses"
       value=L2BC_SPH%Num_Rayleigh_Prof_Warm_Pulses
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Profiles_Surface_Mie
       varname="Num_Profiles_Surface_Mie"
       value=L2BC_SPH%Num_Profiles_Surface_Mie
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Profiles_Surface_Ray
       varname="Num_Profiles_Surface_Ray"
       value=L2BC_SPH%Num_Profiles_Surface_Ray
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Valid_Obs_Prof_L1B_Mie
       varname="Num_Valid_Obs_Prof_L1B_Mie"
       value=L2BC_SPH%Num_Valid_Obs_Prof_L1B_Mie
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Valid_Obs_Prof_L1B_Ray
       varname="Num_Valid_Obs_Prof_L1B_Ray"
       value=L2BC_SPH%Num_Valid_Obs_Prof_L1B_Ray
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Valid_Meas_Prof_L1B_Mie
       varname="Num_Valid_Meas_Prof_L1B_Mie"
       value=L2BC_SPH%Num_Valid_Meas_Prof_L1B_Mie
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Valid_Meas_Prof_L1B_Ray
       varname="Num_Valid_Meas_Prof_L1B_Ray"
       value=L2BC_SPH%Num_Valid_Meas_Prof_L1B_Ray
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Valid_Obs_Bins_L1B_Mie
       varname="Num_Valid_Obs_Bins_L1B_Mie"
       value=L2BC_SPH%Num_Valid_Obs_Bins_L1B_Mie
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Valid_Obs_Bins_L1B_Ray
       varname="Num_Valid_Obs_Bins_L1B_Ray"
       value=L2BC_SPH%Num_Valid_Obs_Bins_L1B_Ray
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Valid_Meas_Bins_L1B_Mie
       varname="Num_Valid_Meas_Bins_L1B_Mie"
       value=L2BC_SPH%Num_Valid_Meas_Bins_L1B_Mie
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Valid_Meas_Bins_L1B_Ray
       varname="Num_Valid_Meas_Bins_L1B_Ray"
       value=L2BC_SPH%Num_Valid_Meas_Bins_L1B_Ray
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Invalid_Meas_Prof_L1B_Mie
       varname="Num_Invalid_Meas_Prof_L1B_Mie"
       value=L2BC_SPH%Num_Invalid_Meas_Prof_L1B_Mie
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Invalid_Meas_Prof_L1B_Ray
       varname="Num_Invalid_Meas_Prof_L1B_Ray"
       value=L2BC_SPH%Num_Invalid_Meas_Prof_L1B_Ray
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Invalid_Meas_Prof_L2B_Mie
       varname="Num_Invalid_Meas_Prof_L2B_Mie"
       value=L2BC_SPH%Num_Invalid_Meas_Prof_L2B_Mie
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Invalid_Meas_Prof_L2B_Ray
       varname="Num_Invalid_Meas_Prof_L2B_Ray"
       value=L2BC_SPH%Num_Invalid_Meas_Prof_L2B_Ray
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Invalid_Meas_Bins_L1B_Mie
       varname="Num_Invalid_Meas_Bins_L1B_Mie"
       value=L2BC_SPH%Num_Invalid_Meas_Bins_L1B_Mie
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Invalid_Meas_Bins_L1B_Ray
       varname="Num_Invalid_Meas_Bins_L1B_Ray"
       value=L2BC_SPH%Num_Invalid_Meas_Bins_L1B_Ray
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Invalid_Meas_Bins_L2B_Mie
       varname="Num_Invalid_Meas_Bins_L2B_Mie"
       value=L2BC_SPH%Num_Invalid_Meas_Bins_L2B_Mie
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Invalid_Meas_Bins_L2B_Ray
       varname="Num_Invalid_Meas_Bins_L2B_Ray"
       value=L2BC_SPH%Num_Invalid_Meas_Bins_L2B_Ray
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Spare_3
       ! Spare_3: this is a spare value without content
       max_value_length = 40
       call write_spare_kvt_string(fileunit,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Valid_Obs_L2B_Mie
       varname="Num_Valid_Obs_L2B_Mie"
       value=L2BC_SPH%Num_Valid_Obs_L2B_Mie
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       !  #[ Num_Valid_Obs_L2B_Ray
       varname="Num_Valid_Obs_L2B_Ray"
       value=L2BC_SPH%Num_Valid_Obs_L2B_Ray
       max_value_length=maxlen_IntAus
       call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]
       IF (L2C_flag_local) THEN
          !  #[ Num_Invalid_Obs_L2C_Mie
          varname="Num_Invalid_Obs_L2C_Mie"
          value=L2BC_SPH%Num_Invalid_Obs_L2C_Mie
          max_value_length=maxlen_IntAus
          call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
          IF (error_flag .ne. no_error) goto 998
          !  #]
          !  #[ Num_Invalid_Obs_L2C_Ray
          varname="Num_Invalid_Obs_L2C_Ray"
          value=L2BC_SPH%Num_Invalid_Obs_L2C_Ray
          max_value_length=maxlen_IntAus
          call write_kvt_string(fileunit,varname,value,max_value_length,error_flag)
          IF (error_flag .ne. no_error) goto 998
          !  #]
       END IF
       !  #[ Spare_4
       ! Spare_4: this is a spare value without content
       max_value_length = 40
       call write_spare_kvt_string(fileunit,max_value_length,error_flag)
       IF (error_flag .ne. no_error) goto 998
       !  #]

    ! Write Geolocation_DSD
    call Write_DSD_to_KVT(L2BC_SPH%Geolocation_DSD,fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Write PCD_DSD
    call Write_DSD_to_KVT(L2BC_SPH%PCD_DSD,fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Write Mie_DSD
    call Write_DSD_to_KVT(L2BC_SPH%Mie_DSD,fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Write Rayleigh_DSD
    call Write_DSD_to_KVT(L2BC_SPH%Rayleigh_DSD,fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 998

    IF (L2C_flag_local) THEN
       ! Write Assim_PCD_DSD
       call Write_DSD_to_KVT(L2BC_SPH%Assim_PCD_DSD,fileunit,error_flag)
       IF (error_flag .ne. no_error) goto 998
       
       ! Write Mie_VecWind_DSD
       call Write_DSD_to_KVT(L2BC_SPH%Mie_VecWind_DSD,fileunit,error_flag)
       IF (error_flag .ne. no_error) goto 998
       
       ! Write Rayleigh_VecWind_DSD
       call Write_DSD_to_KVT(L2BC_SPH%Rayleigh_VecWind_DSD,&
                                 fileunit,error_flag)
       IF (error_flag .ne. no_error) goto 998
    END IF

    ! Write L1B_DSD
    call Write_DSD_to_KVT(L2BC_SPH%L1B_DSD,fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Write Met_DSD
    call Write_DSD_to_KVT(L2BC_SPH%Met_DSD,fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Write RBC_DSD
    call Write_DSD_to_KVT(L2BC_SPH%RBC_DSD,fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Write CLM_DSD
    call Write_DSD_to_KVT(L2BC_SPH%CLM_DSD,fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Write CAL_DSD
    call Write_DSD_to_KVT(L2BC_SPH%CAL_DSD,fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! Write L2B_Param_DSD
    call Write_DSD_to_KVT(L2BC_SPH%L2B_Param_DSD,fileunit,error_flag)
    IF (error_flag .ne. no_error) goto 998

    ! TODO: (possibly) consistency_DSDs(L2BC_SPH,L2C_flag_local)

    IF (L2C_flag_local) THEN
       ! Write L2B_DSD
       call Write_DSD_to_KVT(L2BC_SPH%L2B_DSD,fileunit,error_flag)
       IF (error_flag .ne. no_error) goto 998
       
       ! Write L2C_Param_DSD
       call Write_DSD_to_KVT(L2BC_SPH%L2C_Param_DSD,fileunit,error_flag)
       IF (error_flag .ne. no_error) goto 998
    END IF

    ! save end position
    inquire(unit=fileunit,nextrec=endpos)

    ! calculate L2BC SPH Size
    size_SPH_calc = int(endpos-begpos,IntAl)

    ! compare size with GetByteSize result
    IF (size_SPH_calc .ne. GetByteSize_L2B_SPH_KVT(L2C_flag)) THEN
       call logmsg(log_error,"ERROR in Write_L2BC_SPH_to_KVT: "//&
            "SPH size calculated by GetByteSize_L2B_SPH_KVT does not "//&
            "match actual nr of bytes written to file !!!")
       error_flag = error_programming
       return
    ENDIF

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Write_L2BC_SPH_to_KVT"
    return

  end subroutine Write_L2BC_SPH_to_KVT
    !  #]
  subroutine Read_L2BC_SPH_from_KVT(L2BC_SPH, fileunit, error_flag, &
                                    Compare, L2C_flag)
    !  #[
    ! Reads the L2B or L2C Specific Product Header (KVT values).

    ! if Compare=.true. then
    ! assume the values are already read from the XML file
    ! so read and compare them, and then replace the ones in memory by
    ! those read, because those read are in KVT format, which is the
    ! convention of choice.

    TYPE(L2BC_SPH_type), pointer     :: L2BC_SPH
    integer,             intent(in)  :: fileunit
    integer,             intent(out) :: error_flag
    logical,   optional, intent(in)  :: Compare
    logical,   optional, intent(in)  :: L2C_flag

    ! local variables
    character(len=max_str_len) :: exp_name
    character(len=max_str_len) :: value
    character(len=max_str_len) :: exp_phys_unit
    integer                    :: exp_val_length
    integer                    :: AllocStatus
    integer                    :: filepos
    integer(IntAd)             :: DS_distance
    logical                    :: do_Compare
    logical                    :: order_as_expected
    logical                    :: distance_check_ok
    logical                    :: L2C_flag_local
    character(len=3)           :: L2BC_string
    character(len=7)           :: L2BCsph_string

    !---------------------------------
    ! Reading the L2B SPH KVT elements
    !---------------------------------

    ! init
    error_flag = no_error

    ! interpret optional arugment(s)
    L2C_flag_local=.false.
    IF (present(L2C_flag)) L2C_flag_local=L2C_flag
    L2BC_string    = set_L2BC_string(L2C_flag_local)
    L2BCsph_string = L2BC_string//"_SPH"

    IF (.not. associated(L2BC_SPH)) THEN
       print *,"ERROR in Read_L2BC_SPH_from_KVT():"
       print *,"The pointer to this L2BC_SPH seems not associated !!"
       error_flag = error_programming
       return
    END IF

    do_Compare = .false.
    IF (present(Compare)) do_Compare = Compare

    IF (do_Compare) THEN
       print *,"Read_L2BC_SPH_from_KVT: reading and comparing "//&
            trim(L2BC_string)//" SPH-KVT"
       print *,"items with the XML items in memory"
    ELSE
       print *,"Read_L2BC_MPH_from_KVT: reading "//&
            trim(L2BC_string)//" SPH-KVT items "
    END IF

    !  #[ read the SPH_Descriptor 
    exp_val_length = len(L2BC_SPH%SPH_Descriptor)
    exp_name       = "Sph_Descriptor"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,expect_quotes=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%SPH_Descriptor, value, &
            trim(L2BCsph_string)//"%Sph_Descriptor",error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%SPH_Descriptor = value
    END IF
    !  #]

    !  #[ Read Spare_1: this is a spare value without content
    exp_val_length = 40
    call read_spare_kvt_string(fileunit,exp_val_length,error_flag)
    IF (error_flag .ne. no_error) goto 998
    !  #]

    !  #[ read Num_BRC
    exp_val_length = len(L2BC_SPH%Num_BRC)
    exp_name       = "Num_BRC"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_BRC, value, &
            trim(L2BCsph_string)//"%Num_BRC",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_BRC = value
    END IF
    !  #]
    !  #[ read Intersect_Start_Lat
    exp_val_length = len(L2BC_SPH%Intersect_Start_Lat)
    exp_name       = "Intersect_Start_Lat"
    exp_phys_unit  = "10-6DegN"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.,&
         expected_phys_unit=exp_phys_unit)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Intersect_Start_Lat, value, &
            trim(L2BCsph_string)//"%Intersect_Start_Lat",error_flag,&
            CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Intersect_Start_Lat = value
    END IF
    !  #]
    !  #[ read Intersect_Start_Long
    exp_val_length = len(L2BC_SPH%Intersect_Start_Long)
    exp_name       = "Intersect_Start_Long"
    exp_phys_unit  = "10-6DegE"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.,&
         expected_phys_unit=exp_phys_unit)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Intersect_Start_Long,value,&
            trim(L2BCsph_string)//"%Intersect_Start_Long",error_flag,&
            CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Intersect_Start_Long = value
    END IF
    !  #]
    !  #[ read Intersect_Stop_Lat
    exp_val_length = len(L2BC_SPH%Intersect_Stop_Lat)
    exp_name       = "Intersect_Stop_Lat"
    exp_phys_unit  = "10-6DegN"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.,&
         expected_phys_unit=exp_phys_unit)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Intersect_Stop_Lat, value, &
            trim(L2BCsph_string)//"%Intersect_Stop_Lat",error_flag,&
            CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Intersect_Stop_Lat = value
    END IF
    !  #]
    !  #[ read Intersect_Stop_Long
    exp_val_length = len(L2BC_SPH%Intersect_Stop_Long)
    exp_name       = "Intersect_Stop_Long"
    exp_phys_unit  = "10-6DegE"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.,&
         expected_phys_unit=exp_phys_unit)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Intersect_Stop_Long,value,&
            trim(L2BCsph_string)//"%Intersect_Stop_Long",error_flag,&
            CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Intersect_Stop_Long = value
    END IF
    !  #]
    !  #[ read Sat_Track
    exp_val_length = len(L2BC_SPH%Sat_Track)
    exp_name       = "Sat_Track"
    exp_phys_unit  = "deg"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.,&
         expected_phys_unit=exp_phys_unit)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Sat_Track, value, &
            trim(L2BCsph_string)//"%Sat_Track",error_flag,CompareAsReal=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Sat_Track = value
    END IF
    !  #]

    !  #[ Read Spare_2: this is a spare value without content
    exp_val_length = 40
    call read_spare_kvt_string(fileunit,exp_val_length,error_flag)
    IF (error_flag .ne. no_error) goto 998
    !  #]

    !  #[ read M_Mie
    exp_val_length = len(L2BC_SPH%M_Mie)
    exp_name       = "M_Mie"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%M_Mie, value, &
            trim(L2BCsph_string)//"%M_Mie",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%M_Mie = value
    END IF
    !  #]
    !  #[ read M_Rayleigh
    exp_val_length = len(L2BC_SPH%M_Rayleigh)
    exp_name       = "M_Rayleigh"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%M_Rayleigh, value, &
            trim(L2BCsph_string)//"%M_Rayleigh",error_flag,&
            CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%M_Rayleigh = value
    END IF
    !  #]
    !  #[ read M_Meas
    exp_val_length = len(L2BC_SPH%M_Meas)
    exp_name       = "M_Meas"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%M_Meas, value, &
            trim(L2BCsph_string)//"%M_Meas",error_flag,CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%M_Meas = value
    END IF
    !  #]
    !  #[ read Num_Valid_Mie_Profiles
    exp_val_length = len(L2BC_SPH%Num_Valid_Mie_Profiles)
    exp_name       = "Num_Valid_Mie_Profiles"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Valid_Mie_Profiles, &
            value, trim(L2BCsph_string)//"%Num_Valid_Mie_Profiles",error_flag,&
            CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Valid_Mie_Profiles = value
    END IF
    !  #]
    !  #[ read Num_Valid_Rayleigh_Profiles
    exp_val_length = len(L2BC_SPH%Num_Valid_Rayleigh_Profiles)
    exp_name       = "Num_Valid_Rayleigh_Profiles"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Valid_Rayleigh_Profiles, &
            value, trim(L2BCsph_string)//"%Num_Valid_Rayleigh_Profiles",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Valid_Rayleigh_Profiles = value
    END IF
    !  #]
    !  #[ read Num_Clear_Mie_Profiles
    exp_val_length = len(L2BC_SPH%Num_Clear_Mie_Profiles)
    exp_name       = "Num_Clear_Mie_Profiles"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Clear_Mie_Profiles, &
            value, trim(L2BCsph_string)//"%Num_Clear_Mie_Profiles",error_flag,&
            CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Clear_Mie_Profiles = value
    END IF
    !  #]
    !  #[ read Num_Clear_Rayleigh_Profiles
    exp_val_length = len(L2BC_SPH%Num_Clear_Rayleigh_Profiles)
    exp_name       = "Num_Clear_Rayleigh_Profiles"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Clear_Rayleigh_Profiles, &
            value, trim(L2BCsph_string)//"%Num_Clear_Rayleigh_Profiles",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Clear_Rayleigh_Profiles = value
    END IF
    !  #]
    !  #[ read Num_Cloud_Mie_Profiles
    exp_val_length = len(L2BC_SPH%Num_Cloud_Mie_Profiles)
    exp_name       = "Num_Cloud_Mie_Profiles"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Cloud_Mie_Profiles, &
            value, trim(L2BCsph_string)//"%Num_Cloud_Mie_Profiles",error_flag,&
            CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Cloud_Mie_Profiles = value
    END IF
    !  #]
    !  #[ read Num_Cloud_Rayleigh_Profiles
    exp_val_length = len(L2BC_SPH%Num_Cloud_Rayleigh_Profiles)
    exp_name       = "Num_Cloud_Rayleigh_Profiles"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Cloud_Rayleigh_Profiles, &
            value, trim(L2BCsph_string)//"%Num_Cloud_Rayleigh_Profiles",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Cloud_Rayleigh_Profiles = value
    END IF
    !  #]
    !  #[ read Num_Mie_Prof_Warm_Pulses
    exp_val_length = len(L2BC_SPH%Num_Mie_Prof_Warm_Pulses)
    exp_name       = "Num_Mie_Prof_Warm_Pulses"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Mie_Prof_Warm_Pulses,&
            value, trim(L2BCsph_string)//"%Num_Mie_Prof_Warm_Pulses",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Mie_Prof_Warm_Pulses = value
    END IF
    !  #]
    !  #[ read Num_Rayleigh_Prof_Warm_Pulses
    exp_val_length = len(L2BC_SPH%Num_Rayleigh_Prof_Warm_Pulses)
    exp_name       = "Num_Rayleigh_Prof_Warm_Pulses"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(&
            L2BC_SPH%Num_Rayleigh_Prof_Warm_Pulses,&
            value, trim(L2BCsph_string)//"%Num_Rayleigh_Prof_Warm_Pulses",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Rayleigh_Prof_Warm_Pulses = value
    END IF
    !  #]
    !  #[ read Num_Profiles_Surface_Mie
    exp_val_length = len(L2BC_SPH%Num_Profiles_Surface_Mie)
    exp_name       = "Num_Profiles_Surface_Mie"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Profiles_Surface_Mie,&
            value, trim(L2BCsph_string)//"%Num_Profiles_Surface_Mie",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Profiles_Surface_Mie = value
    END IF
    !  #]
    !  #[ read Num_Profiles_Surface_Ray
    exp_val_length = len(L2BC_SPH%Num_Profiles_Surface_Ray)
    exp_name       = "Num_Profiles_Surface_Ray"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Profiles_Surface_Ray,&
            value, trim(L2BCsph_string)//"%Num_Profiles_Surface_Ray",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Profiles_Surface_Ray = value
    END IF
    !  #]
    !  #[ read Num_Valid_Obs_Prof_L1B_Mie
    exp_val_length = len(L2BC_SPH%Num_Valid_Obs_Prof_L1B_Mie)
    exp_name       = "Num_Valid_Obs_Prof_L1B_Mie"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Valid_Obs_Prof_L1B_Mie,&
            value, trim(L2BCsph_string)//"%Num_Valid_Obs_Prof_L1B_Mie",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Valid_Obs_Prof_L1B_Mie = value
    END IF
    !  #]
    !  #[ read Num_Valid_Obs_Prof_L1B_Ray
    exp_val_length = len(L2BC_SPH%Num_Valid_Obs_Prof_L1B_Ray)
    exp_name       = "Num_Valid_Obs_Prof_L1B_Ray"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Valid_Obs_Prof_L1B_Ray,&
            value, trim(L2BCsph_string)//"%Num_Valid_Obs_Prof_L1B_Ray",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Valid_Obs_Prof_L1B_Ray = value
    END IF
    !  #]
    !  #[ read Num_Valid_Meas_Prof_L1B_Mie
    exp_val_length = len(L2BC_SPH%Num_Valid_Meas_Prof_L1B_Mie)
    exp_name       = "Num_Valid_Meas_Prof_L1B_Mie"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Valid_Meas_Prof_L1B_Mie,&
            value, trim(L2BCsph_string)//"%Num_Valid_Meas_Prof_L1B_Mie",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Valid_Meas_Prof_L1B_Mie = value
    END IF
    !  #]
    !  #[ read Num_Valid_Meas_Prof_L1B_Ray
    exp_val_length = len(L2BC_SPH%Num_Valid_Meas_Prof_L1B_Ray)
    exp_name       = "Num_Valid_Meas_Prof_L1B_Ray"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Valid_Meas_Prof_L1B_Ray,&
            value, trim(L2BCsph_string)//"%Num_Valid_Meas_Prof_L1B_Ray",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Valid_Meas_Prof_L1B_Ray = value
    END IF
    !  #]
    !  #[ read Num_Valid_Obs_Bins_L1B_Mie
    exp_val_length = len(L2BC_SPH%Num_Valid_Obs_Bins_L1B_Mie)
    exp_name       = "Num_Valid_Obs_Bins_L1B_Mie"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Valid_Obs_Bins_L1B_Mie,&
            value, trim(L2BCsph_string)//"%Num_Valid_Obs_Bins_L1B_Mie",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Valid_Obs_Bins_L1B_Mie = value
    END IF
    !  #]
    !  #[ read Num_Valid_Obs_Bins_L1B_Ray
    exp_val_length = len(L2BC_SPH%Num_Valid_Obs_Bins_L1B_Ray)
    exp_name       = "Num_Valid_Obs_Bins_L1B_Ray"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Valid_Obs_Bins_L1B_Ray,&
            value, trim(L2BCsph_string)//"%Num_Valid_Obs_Bins_L1B_Ray",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Valid_Obs_Bins_L1B_Ray = value
    END IF
    !  #]
    !  #[ read Num_Valid_Meas_Bins_L1B_Mie
    exp_val_length = len(L2BC_SPH%Num_Valid_Meas_Bins_L1B_Mie)
    exp_name       = "Num_Valid_Meas_Bins_L1B_Mie"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Valid_Meas_Bins_L1B_Mie,&
            value, trim(L2BCsph_string)//"%Num_Valid_Meas_Bins_L1B_Mie",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Valid_Meas_Bins_L1B_Mie = value
    END IF
    !  #]
    !  #[ read Num_Valid_Meas_Bins_L1B_Ray
    exp_val_length = len(L2BC_SPH%Num_Valid_Meas_Bins_L1B_Ray)
    exp_name       = "Num_Valid_Meas_Bins_L1B_Ray"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Valid_Meas_Bins_L1B_Ray,&
            value, trim(L2BCsph_string)//"%Num_Valid_Meas_Bins_L1B_Ray",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Valid_Meas_Bins_L1B_Ray = value
    END IF
    !  #]
    !  #[ read Num_Invalid_Meas_Prof_L1B_Mie
    exp_val_length = len(L2BC_SPH%Num_Invalid_Meas_Prof_L1B_Mie)
    exp_name       = "Num_Invalid_Meas_Prof_L1B_Mie"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Invalid_Meas_Prof_L1B_Mie,&
            value, trim(L2BCsph_string)//"%Num_Invalid_Meas_Prof_L1B_Mie",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Invalid_Meas_Prof_L1B_Mie = value
    END IF
    !  #]
    !  #[ read Num_Invalid_Meas_Prof_L1B_Ray
    exp_val_length = len(L2BC_SPH%Num_Invalid_Meas_Prof_L1B_Ray)
    exp_name       = "Num_Invalid_Meas_Prof_L1B_Ray"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Invalid_Meas_Prof_L1B_Ray,&
            value, trim(L2BCsph_string)//"%Num_Invalid_Meas_Prof_L1B_Ray",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Invalid_Meas_Prof_L1B_Ray = value
    END IF
    !  #]
    !  #[ read Num_Invalid_Meas_Prof_L2B_Mie
    exp_val_length = len(L2BC_SPH%Num_Invalid_Meas_Prof_L2B_Mie)
    exp_name       = "Num_Invalid_Meas_Prof_L2B_Mie"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Invalid_Meas_Prof_L2B_Mie,&
            value, trim(L2BCsph_string)//"%Num_Invalid_Meas_Prof_L2B_Mie",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Invalid_Meas_Prof_L2B_Mie = value
    END IF
    !  #]
    !  #[ read Num_Invalid_Meas_Prof_L2B_Ray
    exp_val_length = len(L2BC_SPH%Num_Invalid_Meas_Prof_L2B_Ray)
    exp_name       = "Num_Invalid_Meas_Prof_L2B_Ray"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Invalid_Meas_Prof_L2B_Ray,&
            value, trim(L2BCsph_string)//"%Num_Invalid_Meas_Prof_L2B_Ray",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Invalid_Meas_Prof_L2B_Ray = value
    END IF
    !  #]
    !  #[ read Num_Invalid_Meas_Bins_L1B_Mie
    exp_val_length = len(L2BC_SPH%Num_Invalid_Meas_Bins_L1B_Mie)
    exp_name       = "Num_Invalid_Meas_Bins_L1B_Mie"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Invalid_Meas_Bins_L1B_Mie,&
            value, trim(L2BCsph_string)//"%Num_Invalid_Meas_Bins_L1B_Mie",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Invalid_Meas_Bins_L1B_Mie = value
    END IF
    !  #]
    !  #[ read Num_Invalid_Meas_Bins_L1B_Ray
    exp_val_length = len(L2BC_SPH%Num_Invalid_Meas_Bins_L1B_Ray)
    exp_name       = "Num_Invalid_Meas_Bins_L1B_Ray"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Invalid_Meas_Bins_L1B_Ray,&
            value, trim(L2BCsph_string)//"%Num_Invalid_Meas_Bins_L1B_Ray",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Invalid_Meas_Bins_L1B_Ray = value
    END IF
    !  #]
    !  #[ read Num_Invalid_Meas_Bins_L2B_Mie
    exp_val_length = len(L2BC_SPH%Num_Invalid_Meas_Bins_L2B_Mie)
    exp_name       = "Num_Invalid_Meas_Bins_L2B_Mie"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Invalid_Meas_Bins_L2B_Mie,&
            value, trim(L2BCsph_string)//"%Num_Invalid_Meas_Bins_L2B_Mie",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Invalid_Meas_Bins_L2B_Mie = value
    END IF
    !  #]
    !  #[ read Num_Invalid_Meas_Bins_L2B_Ray
    exp_val_length = len(L2BC_SPH%Num_Invalid_Meas_Bins_L2B_Ray)
    exp_name       = "Num_Invalid_Meas_Bins_L2B_Ray"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998

    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Invalid_Meas_Bins_L2B_Ray,&
            value, trim(L2BCsph_string)//"%Num_Invalid_Meas_Bins_L2B_Ray",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Invalid_Meas_Bins_L2B_Ray = value
    END IF
    !  #]

    !  #[ Read Spare_3: this is a spare value without content
    exp_val_length = 40
    call read_spare_kvt_string(fileunit,exp_val_length,error_flag)
    IF (error_flag .ne. no_error) goto 998
    !  #]

    !  #[ read Num_Valid_Obs_L2B_Mie
    exp_val_length = len(L2BC_SPH%Num_Valid_Obs_L2B_Mie)
    exp_name       = "Num_Valid_Obs_L2B_Mie"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998
    
    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Valid_Obs_L2B_Mie,&
            value, trim(L2BCsph_string)//"%Num_Valid_Obs_L2B_Mie",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Valid_Obs_L2B_Mie = value
    END IF
    !  #]
    !  #[ read Num_Valid_Obs_L2B_Ray
    exp_val_length = len(L2BC_SPH%Num_Valid_Obs_L2B_Ray)
    exp_name       = "Num_Valid_Obs_L2B_Ray"
    call read_kvt_string(fileunit,exp_name,value,error_flag,&
         expected_value_length=exp_val_length,is_numerical=.true.)
    IF (error_flag .ne. no_error) goto 998
    
    IF (do_Compare) THEN
       call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Valid_Obs_L2B_Ray,&
            value, trim(L2BCsph_string)//"%Num_Valid_Obs_L2B_Ray",&
            error_flag, CompareAsInteger=.true.)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       L2BC_SPH%Num_Valid_Obs_L2B_Ray = value
    END IF
    !  #]

    IF (L2C_flag_local) THEN
       !  #[ read Num_Invalid_Obs_L2C_Mie
       exp_val_length = len(L2BC_SPH%Num_Invalid_Obs_L2C_Mie)
       exp_name       = "Num_Invalid_Obs_L2C_Mie"
       call read_kvt_string(fileunit,exp_name,value,error_flag,&
            expected_value_length=exp_val_length,is_numerical=.true.)
       IF (error_flag .ne. no_error) goto 998
       
       IF (do_Compare) THEN
          call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Invalid_Obs_L2C_Mie,&
               value, trim(L2BCsph_string)//"%Num_Invalid_Obs_L2C_Mie",&
               error_flag, CompareAsInteger=.true.)
          IF (error_flag .ne. no_error) goto 998
       ELSE
          L2BC_SPH%Num_Invalid_Obs_L2C_Mie = value
       END IF
       !  #]
       !  #[ read Num_Invalid_Obs_L2C_Ray
       exp_val_length = len(L2BC_SPH%Num_Invalid_Obs_L2C_Ray)
       exp_name       = "Num_Invalid_Obs_L2C_Ray"
       call read_kvt_string(fileunit,exp_name,value,error_flag,&
            expected_value_length=exp_val_length,is_numerical=.true.)
       IF (error_flag .ne. no_error) goto 998
       
       IF (do_Compare) THEN
          call Compare_XML_to_KVT_item_Repl(L2BC_SPH%Num_Invalid_Obs_L2C_Ray,&
               value, trim(L2BCsph_string)//"%Num_Invalid_Obs_L2C_Ray",&
               error_flag, CompareAsInteger=.true.)
          IF (error_flag .ne. no_error) goto 998
       ELSE
          L2BC_SPH%Num_Invalid_Obs_L2C_Ray = value
       END IF
       !  #]
    END IF

    !  #[ Read Spare_4: this is a spare value without content
    exp_val_length = 40
    call read_spare_kvt_string(fileunit,exp_val_length,error_flag)
    IF (error_flag .ne. no_error) goto 998
    !  #]
    
    !  #[ read the Geolocation_DSD
    IF (.not. associated(L2BC_SPH%Geolocation_DSD)) THEN
       ! allow this, to enable the possibility of reading the dbl file only,
       ! without reading the HDR files.
       allocate(L2BC_SPH%Geolocation_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L2BC_SPH_from_KVT()"
          error_flag = error_allocate
          goto 998
       END IF
       call Init_DSD(L2BC_SPH%Geolocation_DSD,error_flag)
       IF (error_flag .ne. no_error) goto 998

       ! Read the Geolocation_DSD without comparing
       call Read_DSD_from_KVT(L2BC_SPH%Geolocation_DSD,fileunit,&
            L2B_DS_Name_Geolocation_ADS,L2B_DS_Type_Geolocation_ADS,error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the Geolocation_DSD and compare if needed
       call Read_DSD_from_KVT(L2BC_SPH%Geolocation_DSD,fileunit,&
            L2B_DS_Name_Geolocation_ADS,L2B_DS_Type_Geolocation_ADS,error_flag,Compare)
       IF (error_flag .ne. no_error) goto 998
    END IF
    !  #]
    !  #[ read the PCD_DSD
    IF (.not. associated(L2BC_SPH%PCD_DSD)) THEN
       ! allow this, to enable the possibility of reading the dbl file only,
       ! without reading the HDR files.
       allocate(L2BC_SPH%PCD_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L2BC_SPH_from_KVT()"
          error_flag = error_allocate
          goto 998
       END IF
       call Init_DSD(L2BC_SPH%PCD_DSD,error_flag)
       IF (error_flag .ne. no_error) goto 998

       ! Read the PCD_DSD without comparing
       call Read_DSD_from_KVT(L2BC_SPH%PCD_DSD,fileunit,&
            L2B_DS_Name_PCD_ADS,L2B_DS_Type_PCD_ADS,error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the PCD_DSD and compare if needed
       call Read_DSD_from_KVT(L2BC_SPH%PCD_DSD,fileunit,&
            L2B_DS_Name_PCD_ADS,L2B_DS_Type_PCD_ADS,error_flag,Compare)
       IF (error_flag .ne. no_error) goto 998
    END IF
    !  #]
    !  #[ read the Mie_DSD
    IF (.not. associated(L2BC_SPH%Mie_DSD)) THEN
       ! allow this, to enable the possibility of reading the dbl file only,
       ! without reading the HDR files.
       allocate(L2BC_SPH%Mie_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L2BC_SPH_from_KVT()"
          error_flag = error_allocate
          goto 998
       END IF
       call Init_DSD(L2BC_SPH%Mie_DSD,error_flag)
       IF (error_flag .ne. no_error) goto 998

       ! Read the Mie_DSD without comparing
       call Read_DSD_from_KVT(L2BC_SPH%Mie_DSD,fileunit,&
            L2B_DS_Name_Mie_MDS,L2B_DS_Type_Mie_MDS,error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the Mie_DSD and compare if needed
       call Read_DSD_from_KVT(L2BC_SPH%Mie_DSD,fileunit,&
            L2B_DS_Name_Mie_MDS,L2B_DS_Type_Mie_MDS,error_flag,Compare)
       IF (error_flag .ne. no_error) goto 998
    END IF
    !  #]
    !  #[ read the Rayleigh_DSD
    IF (.not. associated(L2BC_SPH%Rayleigh_DSD)) THEN
       ! allow this, to enable the possibility of reading the dbl file only,
       ! without reading the HDR files.
       allocate(L2BC_SPH%Rayleigh_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L2BC_SPH_from_KVT()"
          error_flag = error_allocate
          goto 998
       END IF
       call Init_DSD(L2BC_SPH%Rayleigh_DSD,error_flag)
       IF (error_flag .ne. no_error) goto 998

       ! Read the Mie_DSD without comparing
       call Read_DSD_from_KVT(L2BC_SPH%Rayleigh_DSD,fileunit,&
            L2B_DS_Name_Rayleigh_MDS,L2B_DS_Type_Rayleigh_MDS,error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the Mie_DSD and compare if needed
       call Read_DSD_from_KVT(L2BC_SPH%Rayleigh_DSD,fileunit,&
            L2B_DS_Name_Rayleigh_MDS,L2B_DS_Type_Rayleigh_MDS,error_flag,Compare)
       IF (error_flag .ne. no_error) goto 998
    END IF
    !  #]

    IF (L2C_flag_local) THEN
       !  #[ read the Assim_PCD_DSD
       IF (.not. associated(L2BC_SPH%Assim_PCD_DSD)) THEN
          ! allow this, to enable the possibility of reading the dbl file only,
          ! without reading the HDR files.
          allocate(L2BC_SPH%Assim_PCD_DSD,STAT=AllocStatus)
          IF (AllocStatus .ne. 0) THEN
             print *,"Allocation failure in subroutine Read_L2BC_SPH_from_KVT()"
             error_flag = error_allocate
             goto 998
          END IF
          call Init_DSD(L2BC_SPH%Assim_PCD_DSD,error_flag)
          IF (error_flag .ne. no_error) goto 998
          
          ! Read the Assim_PCD_DSD without comparing
          call Read_DSD_from_KVT(L2BC_SPH%Assim_PCD_DSD,fileunit,&
               L2C_DS_Name_AssimPCD_ADS,L2C_DS_Type_AssimPCD_ADS,error_flag)
          IF (error_flag .ne. no_error) goto 998
       ELSE
          ! Read the Assim_PCD_DSD and compare if needed
          call Read_DSD_from_KVT(L2BC_SPH%Assim_PCD_DSD,fileunit,&
               L2C_DS_Name_AssimPCD_ADS,L2C_DS_Type_AssimPCD_ADS,error_flag,Compare)
          IF (error_flag .ne. no_error) goto 998
       END IF
       !  #]
       !  #[ read the Mie_VecWind_DSD
       IF (.not. associated(L2BC_SPH%Mie_VecWind_DSD)) THEN
          ! allow this, to enable the possibility of reading the dbl file only,
          ! without reading the HDR files.
          allocate(L2BC_SPH%Mie_VecWind_DSD,STAT=AllocStatus)
          IF (AllocStatus .ne. 0) THEN
             print *,"Allocation failure in subroutine Read_L2BC_SPH_from_KVT()"
             error_flag = error_allocate
             goto 998
          END IF
          call Init_DSD(L2BC_SPH%Mie_VecWind_DSD,error_flag)
          IF (error_flag .ne. no_error) goto 998
          
          ! Read the Mie_VecWind_DSD without comparing
          call Read_DSD_from_KVT(L2BC_SPH%Mie_VecWind_DSD,fileunit,&
               L2C_DS_Name_MieVec_MDS,L2C_DS_Type_MieVec_MDS,error_flag)
          IF (error_flag .ne. no_error) goto 998
       ELSE
          ! Read the Mie_VecWind_DSD and compare if needed
          call Read_DSD_from_KVT(L2BC_SPH%Mie_VecWind_DSD,fileunit,&
               L2C_DS_Name_MieVec_MDS,L2C_DS_Type_MieVec_MDS,error_flag,Compare)
          IF (error_flag .ne. no_error) goto 998
       END IF
       !  #]
       !  #[ read the Rayleigh_VecWind_DSD
       IF (.not. associated(L2BC_SPH%Rayleigh_VecWind_DSD)) THEN
          ! allow this, to enable the possibility of reading the dbl file only,
          ! without reading the HDR files.
          allocate(L2BC_SPH%Rayleigh_VecWind_DSD,STAT=AllocStatus)
          IF (AllocStatus .ne. 0) THEN
             print *,"Allocation failure in subroutine Read_L2BC_SPH_from_KVT()"
             error_flag = error_allocate
             goto 998
          END IF
          call Init_DSD(L2BC_SPH%Rayleigh_VecWind_DSD,error_flag)
          IF (error_flag .ne. no_error) goto 998
          
          ! Read the Rayleigh_VecWind_DSD without comparing
          call Read_DSD_from_KVT(L2BC_SPH%Rayleigh_VecWind_DSD,fileunit,&
               L2C_DS_Name_RayleighVec_MDS,L2C_DS_Type_RayleighVec_MDS,error_flag)
          IF (error_flag .ne. no_error) goto 998
       ELSE
          ! Read the Rayleigh_VecWind_DSD and compare if needed
          call Read_DSD_from_KVT(L2BC_SPH%Rayleigh_VecWind_DSD,fileunit,&
               L2C_DS_Name_RayleighVec_MDS,L2C_DS_Type_RayleighVec_MDS,error_flag,Compare)
          IF (error_flag .ne. no_error) goto 998
       END IF
       !  #]
    END IF

    !  #[ read the L1B_DSD

    IF (.not. associated(L2BC_SPH%L1B_DSD)) THEN
       ! allow this, to enable the possibility of reading the dbl file only,
       ! without reading the HDR files.
       allocate(L2BC_SPH%L1B_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L2BC_SPH_from_KVT()"
          error_flag = error_allocate
          goto 998
       END IF
       call Init_DSD(L2BC_SPH%L1B_DSD,error_flag)
       IF (error_flag .ne. no_error) goto 998

       ! Read the L1B_DSD without comparing
       call Read_DSD_from_KVT(L2BC_SPH%L1B_DSD,fileunit,&
            L2B_DS_Name_L1B_Product,L2B_DS_Type_L1B_Product,error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the L1B_DSD and compare if needed
       call Read_DSD_from_KVT(L2BC_SPH%L1B_DSD,fileunit,&
            L2B_DS_Name_L1B_Product,L2B_DS_Type_L1B_Product,error_flag,Compare)
       IF (error_flag .ne. no_error) goto 998
    END IF

    !  #]
    !  #[ read the Met_DSD
    IF (.not. associated(L2BC_SPH%Met_DSD)) THEN
       ! allow this, to enable the possibility of reading the dbl file only,
       ! without reading the HDR files.
       allocate(L2BC_SPH%Met_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L2BC_SPH_from_KVT()"
          error_flag = error_allocate
          goto 998
       END IF
       call Init_DSD(L2BC_SPH%Met_DSD,error_flag)
       IF (error_flag .ne. no_error) goto 998

       ! Read the Met_DSD without comparing
       call Read_DSD_from_KVT(L2BC_SPH%Met_DSD,fileunit,&
            L2B_DS_Name_Met_MDS,L2B_DS_Type_Met_MDS,error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the Met_DSD and compare if needed
       call Read_DSD_from_KVT(L2BC_SPH%Met_DSD,fileunit,&
            L2B_DS_Name_Met_MDS,L2B_DS_Type_Met_MDS,error_flag,Compare)
       IF (error_flag .ne. no_error) goto 998
    END IF
    !  #]
    !  #[ read the RBC_DSD

    IF (.not. associated(L2BC_SPH%RBC_DSD)) THEN
       ! allow this, to enable the possibility of reading the dbl file only,
       ! without reading the HDR files.
       allocate(L2BC_SPH%RBC_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L2BC_SPH_from_KVT()"
          error_flag = error_allocate
          goto 998
       END IF
       call Init_DSD(L2BC_SPH%RBC_DSD,error_flag)
       IF (error_flag .ne. no_error) goto 998

       ! Read the RBC_DSD without comparing
       call Read_DSD_from_KVT(L2BC_SPH%RBC_DSD,fileunit,&
            L2B_DS_Name_RBC_Product,L2B_DS_Type_RBC_Product,error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the RBC_DSD and compare if needed
       call Read_DSD_from_KVT(L2BC_SPH%RBC_DSD,fileunit,&
            L2B_DS_Name_RBC_Product,L2B_DS_Type_RBC_Product,error_flag,Compare)
       IF (error_flag .ne. no_error) goto 998
    END IF

    !  #]
    !  #[ read the CLM_DSD

    IF (.not. associated(L2BC_SPH%CLM_DSD)) THEN
       ! allow this, to enable the possibility of reading the dbl file only,
       ! without reading the HDR files.
       allocate(L2BC_SPH%CLM_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L2BC_SPH_from_KVT()"
          error_flag = error_allocate
          goto 998
       END IF
       call Init_DSD(L2BC_SPH%CLM_DSD,error_flag)
       IF (error_flag .ne. no_error) goto 998

       ! Read the CLM_DSD without comparing
       call Read_DSD_from_KVT(L2BC_SPH%CLM_DSD,fileunit,&
            L2B_DS_Name_CLM_Product,L2B_DS_Type_CLM_Product,error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the CLM_DSD and compare if needed
       call Read_DSD_from_KVT(L2BC_SPH%CLM_DSD,fileunit,&
            L2B_DS_Name_CLM_Product,L2B_DS_Type_CLM_Product,error_flag,Compare)
       IF (error_flag .ne. no_error) goto 998
    END IF

    !  #]
    !  #[ read the CAL_DSD
    IF (.not. associated(L2BC_SPH%CAL_DSD)) THEN
       ! allow this, to enable the possibility of reading the dbl file only,
       ! without reading the HDR files.
       allocate(L2BC_SPH%CAL_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L2BC_SPH_from_KVT()"
          error_flag = error_allocate
          goto 998
       END IF
       call Init_DSD(L2BC_SPH%CAL_DSD,error_flag)
       IF (error_flag .ne. no_error) goto 998

       ! Read the CAL_DSD without comparing
       call Read_DSD_from_KVT(L2BC_SPH%CAL_DSD,fileunit,&
            L2B_DS_Name_CAL_Product,L2B_DS_Type_CAL_Product,error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the CAL_DSD and compare if needed
       call Read_DSD_from_KVT(L2BC_SPH%CAL_DSD,fileunit,&
            L2B_DS_Name_CAL_Product,L2B_DS_Type_CAL_Product,error_flag,Compare)
       IF (error_flag .ne. no_error) goto 998
    END IF
    !  #]
    !  #[ read the L2B_Param_DSD
    IF (.not. associated(L2BC_SPH%L2B_Param_DSD)) THEN
       ! allow this, to enable the possibility of reading the dbl file only,
       ! without reading the HDR files.
       allocate(L2BC_SPH%L2B_Param_DSD,STAT=AllocStatus)
       IF (AllocStatus .ne. 0) THEN
          print *,"Allocation failure in subroutine Read_L2BC_SPH_from_KVT()"
          error_flag = error_allocate
          goto 998
       END IF
       call Init_DSD(L2BC_SPH%L2B_Param_DSD,error_flag)
       IF (error_flag .ne. no_error) goto 998

       ! Read the L2B_Param_DSD without comparing
       call Read_DSD_from_KVT(L2BC_SPH%L2B_Param_DSD,fileunit,&
            L2B_DS_Name_L2B_Param_ADS,L2B_DS_Type_L2B_Param_ADS,error_flag)
       IF (error_flag .ne. no_error) goto 998
    ELSE
       ! Read the L2B_Param_DSD and compare if needed
       call Read_DSD_from_KVT(L2BC_SPH%L2B_Param_DSD,fileunit,&
            L2B_DS_Name_L2B_Param_ADS,L2B_DS_Type_L2B_Param_ADS,error_flag,Compare)
       IF (error_flag .ne. no_error) goto 998
    END IF
    !  #]

    IF (L2C_flag_local) THEN
       !  #[ read the L2B_DSD
       IF (.not. associated(L2BC_SPH%L2B_DSD)) THEN
          ! allow this, to enable the possibility of reading the dbl file only,
          ! without reading the HDR files.
          allocate(L2BC_SPH%L2B_DSD,STAT=AllocStatus)
          IF (AllocStatus .ne. 0) THEN
             print *,"Allocation failure in subroutine Read_L2BC_SPH_from_KVT()"
             error_flag = error_allocate
             goto 998
          END IF
          call Init_DSD(L2BC_SPH%L2B_DSD,error_flag)
          IF (error_flag .ne. no_error) goto 998
          
          ! Read the L2B_DSD without comparing
          call Read_DSD_from_KVT(L2BC_SPH%L2B_DSD,fileunit,&
               L2C_DS_Name_L2B_Product,L2C_DS_Type_L2B_Product,error_flag)
          IF (error_flag .ne. no_error) goto 998
       ELSE
          ! Read the L2B_DSD and compare if needed
          call Read_DSD_from_KVT(L2BC_SPH%L2B_DSD,fileunit,&
               L2C_DS_Name_L2B_Product,L2C_DS_Type_L2B_Product,error_flag,Compare)
          IF (error_flag .ne. no_error) goto 998
       END IF
       !  #]
       !  #[ read the L2C_Param_DSD
       IF (.not. associated(L2BC_SPH%L2C_Param_DSD)) THEN
          ! allow this, to enable the possibility of reading the dbl file only,
          ! without reading the HDR files.
          allocate(L2BC_SPH%L2C_Param_DSD,STAT=AllocStatus)
          IF (AllocStatus .ne. 0) THEN
             print *,"Allocation failure in subroutine Read_L2BC_SPH_from_KVT()"
             error_flag = error_allocate
             goto 998
          END IF
          call Init_DSD(L2BC_SPH%L2C_Param_DSD,error_flag)
          IF (error_flag .ne. no_error) goto 998
          
          ! Read the L2C_Param_DSD without comparing
          call Read_DSD_from_KVT(L2BC_SPH%L2C_Param_DSD,fileunit,&
               L2C_DS_Name_L2C_Param_ADS,L2C_DS_Type_L2C_Param_ADS,error_flag)
          IF (error_flag .ne. no_error) goto 998
       ELSE
          ! Read the L2C_Param_DSD and compare if needed
          call Read_DSD_from_KVT(L2BC_SPH%L2C_Param_DSD,fileunit,&
               L2C_DS_Name_L2C_Param_ADS,L2C_DS_Type_L2C_Param_ADS,error_flag,Compare)
          IF (error_flag .ne. no_error) goto 998
       END IF
       !  #]
    END IF

    !  #[ check if the DS's are in the expected order
    order_as_expected = .true.
    IF ( L2BC_SPH%Geolocation_DSD%start_of_ds .gt. &
         L2BC_SPH%PCD_DSD%start_of_ds) order_as_expected = .false.
    IF ( L2BC_SPH%PCD_DSD%start_of_ds .gt. &
         L2BC_SPH%Mie_DSD%start_of_ds) order_as_expected = .false.
    IF ( L2BC_SPH%Mie_DSD%start_of_ds .gt. &
         L2BC_SPH%Rayleigh_DSD%start_of_ds) order_as_expected = .false.

    IF (L2C_flag_local) THEN
       IF ( L2BC_SPH%Rayleigh_DSD%start_of_ds .gt. &
            L2BC_SPH%Assim_PCD_DSD%start_of_ds) order_as_expected = .false.
       IF ( L2BC_SPH%Assim_PCD_DSD%start_of_ds .gt. &
            L2BC_SPH%Mie_VecWind_DSD%start_of_ds) order_as_expected = .false.
       IF ( L2BC_SPH%Mie_VecWind_DSD%start_of_ds .gt. &
            L2BC_SPH%Rayleigh_VecWind_DSD%start_of_ds) &
            order_as_expected = .false.
    END IF

    ! the L1B_DSD, Met_DSD, RBC_DSD, CLM_DSD and L2B_Param_DSD point to
    ! external files so their filepointer start values cannot be checked

    ! the L2B_DSD and L2C_Param_DSD point to
    ! external files so their filepointer start values cannot be checked

    print *,"-------------------------------------------------"
    IF (order_as_expected) THEN
       print *,"DSD order is OK"
    ELSE
       call logmsg(log_error,"ERROR in Read_L2BC_SPH_from_KVT: "//&
            "order of DSD's is not as expected for this file")
       error_flag = error_reading_file
       return
    END IF
    !  #]

    !  #[ check if the DS's are following each other without gaps
    
    distance_check_ok = .true.
       
    ! request a filepointer pointing to the first byte AFTER
    ! the last byte of the KVT header
    inquire(unit=fileunit,nextrec=filepos)
    DS_distance = filepos - L2BC_SPH%Geolocation_DSD%start_of_ds
    IF ( DS_distance .ne. 0) THEN
       print *,"WARNING: there seem to be unexpected extra bytes"
       print *,"between the KVT header and the Geolocation_ADS"
       print *,"DS_distance = ",DS_distance
       print *,"filepos = ",filepos
       print *,trim(L2BCsph_string)//"%Geolocation_DSD%start_of_ds = ",&
            L2BC_SPH%Geolocation_DSD%start_of_ds
       distance_check_ok = .false.
    END IF
    
    DS_distance = L2BC_SPH%PCD_DSD%start_of_ds - &
         L2BC_SPH%Geolocation_DSD%end_of_ds - 1
    IF ( DS_distance .ne. 0) THEN
       print *,"WARNING: there seem to be unexpected extra bytes"
       print *,"between the Geolocation_ADS and the PCD_ADS"
       print *,"DS_distance = ",DS_distance
       distance_check_ok = .false.
    END IF
    
    DS_distance = L2BC_SPH%Mie_DSD%start_of_ds - &
         L2BC_SPH%PCD_DSD%end_of_ds - 1
    IF ( DS_distance .ne. 0) THEN
       print *,"WARNING: there seem to be unexpected extra bytes"
       print *,"between the PCD_ADS and the Mie_MDS"
       print *,"DS_distance = ",DS_distance
       distance_check_ok = .false.
    END IF

    DS_distance = L2BC_SPH%Rayleigh_DSD%start_of_ds - &
         L2BC_SPH%Mie_DSD%end_of_ds - 1
    IF ( DS_distance .ne. 0) THEN
       print *,"WARNING: there seem to be unexpected extra bytes"
       print *,"between the Mie_ADS and the Rayleigh_MDS"
       print *,"DS_distance = ",DS_distance
       distance_check_ok = .false.
    END IF
    
    IF (L2C_flag_local) THEN
       DS_distance = L2BC_SPH%Assim_PCD_DSD%start_of_ds - &
            L2BC_SPH%Rayleigh_DSD%end_of_ds - 1
       IF ( DS_distance .ne. 0) THEN
          print *,"WARNING: there seem to be unexpected extra bytes"
          print *,"between the Rayleigh_ADS and the Assim_PCD_ADS"
          print *,"DS_distance = ",DS_distance
          distance_check_ok = .false.
       END IF

       DS_distance = L2BC_SPH%Mie_VecWind_DSD%start_of_ds - &
            L2BC_SPH%Assim_PCD_DSD%end_of_ds - 1
       IF ( DS_distance .ne. 0) THEN
          print *,"WARNING: there seem to be unexpected extra bytes"
          print *,"between the Assim_PCD_ADS and the Mie_VecWind_MDS"
          print *,"DS_distance = ",DS_distance
          distance_check_ok = .false.
       END IF

       DS_distance = L2BC_SPH%Rayleigh_VecWind_DSD%start_of_ds - &
            L2BC_SPH%Mie_VecWind_DSD%end_of_ds - 1
       IF ( DS_distance .ne. 0) THEN
          print *,"WARNING: there seem to be unexpected extra bytes"
          print *,"between the Mie_VecWind_MDS and the Rayleigh_VecWind_MDS"
          print *,"DS_distance = ",DS_distance
          distance_check_ok = .false.
       END IF

    END IF
       
    ! the L1B_DSD, Met_DSD, RBC_DSD, CLM_DSD and L2B_Param_DSD point to 
    ! external files so their distance values have no meaning

    ! the L2B_DSD and L2C_Param_DSD point to
    ! external files so their distance values have no meaning

    IF (distance_check_ok) THEN
       print *,"DS distance check OK"
    ELSE
       call logmsg(log_error,"ERROR in Read_L2BC_SPH_from_KVT: DS"//&
            " distance check failed. The end_of_ds field in one DSD seems"//&
            " not consistent with start_of_ds field in the next DSD")
       error_flag = error_reading_file
       return
    END IF
    !  #]

    ! check if the datasets have the expected size:
    !
    ! I wish to use the GetByteSize functions for this check,
    ! which makes the module dependent on the actual dataset,
    ! which can only be allocated after the SPH has been read !!!
    ! therefore this test is moved to the ReadL2Bdata module.
    ! JK. 14-aug-2006

    !------------------------------
    ! done
    !------------------------------
    return

998 print *,"ABORTING subroutine Read_L2BC_SPH_to_KVT"
    return

  end subroutine Read_L2BC_SPH_from_KVT
    !  #]
  !---------------------------------
  function GetByteSize_L2B_SPH_KVT(L2C_flag) result(KVT_bytesize)
    !  #[
    ! optional input parameter
    logical,   optional, intent(in)  :: L2C_flag

    ! result
    integer(IntAl) :: KVT_bytesize

    ! local variable
    logical        :: L2C_flag_local

    ! interpret optional arugment(s)
    L2C_flag_local=.false.
    IF (present(L2C_flag)) L2C_flag_local=L2C_flag

    ! KVT sizes have size(key)+size(=)+size(value)+size(terminator)
    KVT_bytesize = &
         !     len(key)                len(=)  len(value)      len(unit)   len(terminator)
         !                                                  or len(quotes)
         int(len("Sph_Descriptor")       + 1 + maxlen_SPH_Descriptor + 2         + 1 + &
             40+1+& ! spare_1
             len("Num_BRC")              + 1 + maxlen_IntAus                     + 1 + &
             len("Intersect_Start_Lat")  + 1 + maxlen_IntAl  + len("<10-6DegN>") + 1 + &
             len("Intersect_Start_Long") + 1 + maxlen_IntAl  + len("<10-6DegE>") + 1 + &
             len("Intersect_Stop_Lat")   + 1 + maxlen_IntAl  + len("<10-6DegN>") + 1 + &
             len("Intersect_Stop_Long")  + 1 + maxlen_IntAl  + len("<10-6DegE>") + 1 + &
             len("Sat_Track")            + 1 + maxlen_FAdo36 + len("<deg>")      + 1,IntAl)
 
    KVT_bytesize = KVT_bytesize + &         
         int(40+1+& ! spare_2
             len("M_Mie")                + 1 + maxlen_IntAuc                     + 1 + &
             len("M_Rayleigh")           + 1 + maxlen_IntAuc                     + 1 + &
             len("M_Meas")               + 1 + maxlen_IntAuc                     + 1 + &
             len("Num_Valid_Mie_Profiles")        + 1 + maxlen_IntAus            + 1 + &
             len("Num_Valid_Rayleigh_Profiles")   + 1 + maxlen_IntAus            + 1 + &
             len("Num_Clear_Mie_Profiles")        + 1 + maxlen_IntAus            + 1 + &
             len("Num_Clear_Rayleigh_Profiles")   + 1 + maxlen_IntAus            + 1 + &
             len("Num_Cloud_Mie_Profiles")        + 1 + maxlen_IntAus            + 1 + &
             len("Num_Cloud_Rayleigh_Profiles")   + 1 + maxlen_IntAus            + 1 + &
             len("Num_Mie_Prof_Warm_Pulses")      + 1 + maxlen_IntAus            + 1 + &
             len("Num_Rayleigh_Prof_Warm_Pulses") + 1 + maxlen_IntAus            + 1 + &
             len("Num_Profiles_Surface_Mie")      + 1 + maxlen_IntAus            + 1 + &
             len("Num_Profiles_Surface_Ray")      + 1 + maxlen_IntAus            + 1,IntAl)
    KVT_bytesize = KVT_bytesize + &         
         int(len("Num_Valid_Obs_Prof_L1B_Mie")    + 1 + maxlen_IntAus            + 1 + &
             len("Num_Valid_Obs_Prof_L1B_Ray")    + 1 + maxlen_IntAus            + 1 + &
             len("Num_Valid_Meas_Prof_L1B_Mie")   + 1 + maxlen_IntAus            + 1 + &
             len("Num_Valid_Meas_Prof_L1B_Ray")   + 1 + maxlen_IntAus            + 1 + &
             len("Num_Valid_Obs_Bins_L1B_Mie")    + 1 + maxlen_IntAus            + 1 + &
             len("Num_Valid_Obs_Bins_L1B_Ray")    + 1 + maxlen_IntAus            + 1 + &
             len("Num_Valid_Meas_Bins_L1B_Mie")   + 1 + maxlen_IntAus            + 1 + &
             len("Num_Valid_Meas_Bins_L1B_Ray")   + 1 + maxlen_IntAus            + 1 + &
             len("Num_Invalid_Meas_Prof_L1B_Mie") + 1 + maxlen_IntAus            + 1 + &
             len("Num_Invalid_Meas_Prof_L1B_Ray") + 1 + maxlen_IntAus            + 1 + &
             len("Num_Invalid_Meas_Prof_L2B_Mie") + 1 + maxlen_IntAus            + 1 + &
             len("Num_Invalid_Meas_Prof_L2B_Ray") + 1 + maxlen_IntAus            + 1 + &
             len("Num_Invalid_Meas_Bins_L1B_Mie") + 1 + maxlen_IntAus            + 1 + &
             len("Num_Invalid_Meas_Bins_L1B_Ray") + 1 + maxlen_IntAus            + 1 + &
             len("Num_Invalid_Meas_Bins_L2B_Mie") + 1 + maxlen_IntAus            + 1 + &
             len("Num_Invalid_Meas_Bins_L2B_Ray") + 1 + maxlen_IntAus            + 1,IntAl)
    KVT_bytesize = KVT_bytesize + &         
         int(40+1+& ! spare_3
             len("Num_Valid_Obs_L2B_Mie")         + 1 + maxlen_IntAus            + 1 + &
             len("Num_Valid_Obs_L2B_Ray")         + 1 + maxlen_IntAus            + 1 + &
             ! the optional L2C items are added below
             40+1+& ! spare_4
             Num_DSD_L2B*GetByteSize_DSD_KVT(),IntAl) ! the 10 L2B DSD's
    ! the optional L2C DSD's are added below
    
    
    IF (L2C_flag_local) THEN
       KVT_bytesize = KVT_bytesize + &
            ! len(key)                len(=)  len(value)      len(unit)   len(terminator)
            !                                              or len(quotes)
            int(len("Num_Invalid_Obs_L2C_Mie") + 1 + maxlen_IntAus + 1 + &
                len("Num_Invalid_Obs_L2B_Ray") + 1 + maxlen_IntAus + 1 + &
                (Num_DSD_L2C-Num_DSD_L2B)*GetByteSize_DSD_KVT(),IntAl) ! the 5 L2C DSD's
    END IF

  end function GetByteSize_L2B_SPH_KVT
    !  #]
  subroutine Get_M_Meas(L2BC_SPH,M_Meas,error_flag)
    !  #[
    TYPE(L2BC_SPH_type), pointer     :: L2BC_SPH   ! input
    integer(IntAuc),     intent(out) :: M_Meas     ! output
    integer,             intent(out) :: error_flag ! output

    error_flag = no_error
    call GetIntAuc(L2BC_SPH%M_Meas,M_Meas,error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine Get_M_Meas
    !  #]
  subroutine Get_M_Mie(L2BC_SPH,M_Mie,error_flag)
    !  #[
    TYPE(L2BC_SPH_type), pointer     :: L2BC_SPH   ! input
    integer(IntAuc),     intent(out) :: M_Mie      ! output
    integer,             intent(out) :: error_flag ! output

    error_flag = no_error
    call GetIntAuc(L2BC_SPH%M_Mie,M_Mie,error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine Get_M_Mie
    !  #]
  subroutine Get_M_Rayleigh(L2BC_SPH,M_Rayleigh,error_flag)
    !  #[
    TYPE(L2BC_SPH_type), pointer     :: L2BC_SPH   ! input
    integer(IntAuc),     intent(out) :: M_Rayleigh ! output
    integer,             intent(out) :: error_flag ! output

    error_flag = no_error
    call GetIntAuc(L2BC_SPH%M_Rayleigh,M_Rayleigh,error_flag)
    IF (error_flag .ne. no_error) return

  end subroutine Get_M_Rayleigh
    !  #]
  !---------------------------------
end module L2BC_SPH
