% ReadBinaryData : Reader for binary L1b,L2a,L2b,L2c,MET,RBC,CLM,CAL data
% 
% xxxData = ReadBinaryData(filename,flags) 
% is a structure containing the Data Sets (selected by the variable flags ),
% the MPH, and the SPH,
% of the L1b or L2a or L2b or L2c or MET or RBC or CLM data file found 
% in 'filename' [where filename = AE_*.DBL]
% 
% Examples : 
% L2cData = ReadBinaryData(L2cfilename,'ALL') will read all the Data Sets of the file.
% L2bData = ReadBinaryData(L2bfilename,'Geo_ADS') will read only the 
%           Geolocation Data Set of the L2b file
% L1bData = ReadBinaryData(L1bfilename,'US_DSR','WV_DSR') will read the Useful
%           Signal Data Set and the Wind Velocity Data Set
% L2aData = ReadBinaryData(L2afilename,'PCD_ADS') will read the Product 
%           Confidence Dataset
% MetData = ReadBinaryData(metfilename,'ALL') will read all the met datasets
%
% Depending on the filename, the present software
% reads either L1b, L2a, L2b, L2c, MET, RBC , CLM or CAL data.
%  
% Flags can take the values : (case-insensitive)

% - for L1b: 'Geo_ADS','PCD_ADS','GWD_ADS','Meas_ADS','CCD_GADS','US_MDS','WV_MDS'
% - for L2a: 'Geo_ADS','PCD_ADS','Opt_ADS'
% - for L2b: 'Geo_ADS','PCD_ADS','Mie_MDS','Rayleigh_MDS'
% - for L2c: 'Geo_ADS','PCD_ADS','Mie_MDS','Rayleigh_MDS', 'Assim_PCD_ADS', 'Mie_Vec_MDS', 'Rayleigh_Vec_MDS'
% - for MET: 'Geo_ADS1','Geo_ADS2','Met_MDS1','Met_MDS2'
% - for RBC: 'RB_ADS'
% - for CLM: 'Clim_ADS'
% - for CAL: 'Cal_ADS'
% - for any file type: 'ALL' 
% 

 
% HISTORY
% 
% 09 Feb 2007   P Poli              New code. Reads all datatypes.
%                                   L1b reading from Marie-Laure Denneulin,
%                                   L2b, L2c, MET, RBC from Paul Poli with mods by David Tan
%                                   (see history for each function).
%                                   Main program shares as many resources as possible.
% 05 Apr 2007   P Poli              Upgrade to L2bP 1.3 and add L2c reading routines
% 25 Apr 2007   P Poli              Add reader for AUX_CLM files
% 26 Apr 2007   ML Denneulin        Add reader for L2a data
% 15 May 2007   P Poli              Include ReadMPH, upgrade to L1B v1.06, 
%                                   keeping compatibility with L1B v1.05
% 07 Jun 2007   ML Denneulin        Add reader for CAL data
% 29 Aug 2007   P Poli              Upgrade to L1B v1.07
% 03 Dec 2007   P Poli              Add compatibility with L1B v1.04
% 22 Jan 2008   P Poli              Upgrade to L1B v1.08
% 14 Feb 2008   P Poli              Fixed bug in reading Met data 
%                                   (Met_MDS2 overwritten by Met_MDS1)
% 28 Feb 2008   P Poli              Upgrade to L1B V1.09
function xxxData = ReadBinaryData(filename,varargin);

% Initialize structure
xxxData = [];

% Indicate here the IODD document reference
iodd_reference1b_v104='3.1_30Oct2006'; % E2S and L1B
iodd_reference1b_v105='3.2_22Jan2007'; % E2S and L1B
iodd_reference1b_v106='3.3_12Apr2007'; % E2S and L1B
iodd_reference1b_v108='3.4_14Jan2008'; % E2S and L1B
iodd_reference1b_v109='3.5_??Feb2008'; % E2S and L1B
iodd_reference2bc='1.3_23Feb2007'; % L2B/C
iodd_reference2a='1.3_17Jan2007'; % L2A
%                1234567890123 use 13 characters

% First step: determine file type: L1b, L2a, L2b, L2c, MET, RBC, CLM or CAL data ?
pos_ALD_U_=findstr('ALD_U_',filename);
if (isempty(pos_ALD_U_)),
	pos_AUX_=findstr('AUX_',filename);
	if (isempty(pos_AUX_)),
		fprintf('Error: filename is not an ADM Aeolus filename\n');
		return;
	end
	filetype=filename([pos_AUX_+4:pos_AUX_+6]);
	switch filetype
	case 'MET'
		fprintf('Read MET data \n');
	case 'RBC'
		fprintf('Read RBC data \n');
	case 'CLM'
		fprintf('Read CLM data \n');
	case 'CAL'
		fprintf('Read CAL data \n');
	otherwise
		fprintf('Error: AUX filetype is not CLM, RBC, MET, but %s\n',filetype);
		return
	end
else
	filetype=filename([pos_ALD_U_+8:pos_ALD_U_+9]);
	switch filetype
	case '1B'
		fprintf('Read L1b data \n');
	case '2A'
		fprintf('Read L2a data \n');
	case '2B'
		fprintf('Read L2b data \n');
	case '2C'
		fprintf('Read L2c data \n');
	otherwise
		fprintf('Error: Lxx filetype is neither L1b, nor L2a/b/c, but %s\n',filetype);
		return
	end
end

fprintf('Trying to open file %s \n', filename);

%  Open the selected data file in mode 'read' ('r') and 'big endian ordering' ('b')
fid = fopen(filename,'r','b');
if fid==-1,
        fprintf('Error : cannot open selected file\n');
	return;
end

% read MPH
nbByteMPH = 1247; % v1.07 or earlier
MPH = fread(fid, nbByteMPH, '*char');
[xxxData.MPH,flagv108]=ReadMPH(MPH);
if (flagv108 == 1),
	% read 4 extra characters for 1.08: nbByteMPH = 1251
	decv108 = fread(fid, 4,'*char');
end
nbByteSPH = xxxData.MPH.Sph_Size;
nbDS =  xxxData.MPH.Num_Data_Sets;
nbDSD =  xxxData.MPH.Num_Dsd;
version = xxxData.MPH.Software_Ver;
% expected information
switch filetype
case '1B'
	switch version
	case {'ADM_L1bP/1.04 ','ADM_L1bP/1.05 '}
		expected_SPH_size=6005;
		expected_DS_Number=7;
		expected_DSD_Number=15;
		expected_DS_Names{1} ='Geolocation_ADS             ';
		expected_DS_Names{2} ='Product_Confidence_Data_ADS ';
		expected_DS_Names{3} ='Ground_Wind_Detection_ADS   ';
		expected_DS_Names{4} ='Measurement_ADS             ';
		expected_DS_Names{5} ='Calibration_Char_GADS       ';
		expected_DS_Names{6} ='Useful_Signal_MDS           ';
		expected_DS_Names{7} ='Wind_Velocity_MDS           ';
		expected_DS_Names{8} ='Aeolus_Level_1A_Product     ';
		expected_DS_Names{9} ='Aeolus_MRC                  ';
		expected_DS_Names{10}='Aeolus_RRC                  ';
		expected_DS_Names{11}='Aeolus_ZWC                  ';
		expected_DS_Names{12}='Level_1B_Proc_Params        ';
		expected_DS_Names{13}='Satellite_Characterisation  ';
		expected_DS_Names{14}='Orbit_Scenario_File         ';
		expected_DS_Names{15}='Geoid_Model                 ';
		map_DS_to_DSD=[1,2,3,4,5,6,7];
		if (version == 'ADM_L1bP/1.04 ')
			iodd_reference=iodd_reference1b_v104;
		end
		if (version == 'ADM_L1bP/1.05 ')
			iodd_reference=iodd_reference1b_v105;
		end
	case {'ADM_L1bP/1.06 ','ADM_L1bP/1.07 ','ADM_L1bP/1.08 ','ADM_L1bP/1.09 '}
		expected_SPH_size=6293;
		expected_DS_Number=7;
		expected_DSD_Number=16;
		expected_DS_Names{1} ='Geolocation_ADS             ';
		expected_DS_Names{2} ='Product_Confidence_Data_ADS ';
		expected_DS_Names{3} ='Ground_Wind_Detection_ADS   ';
		expected_DS_Names{4} ='Measurement_ADS             ';
		expected_DS_Names{5} ='Calibration_Char_GADS       ';
		expected_DS_Names{6} ='Useful_Signal_MDS           ';
		expected_DS_Names{7} ='Wind_Velocity_MDS           ';
		expected_DS_Names{8} ='Aeolus_Level_1A_Product     ';
		expected_DS_Names{9} ='Aeolus_MRC                  ';
		expected_DS_Names{10}='Aeolus_RRC                  ';
		expected_DS_Names{11}='Aeolus_ZWC                  ';
		expected_DS_Names{12}='Level_1B_Proc_Params        ';
		expected_DS_Names{13}='Satellite_Characterisation  ';
		expected_DS_Names{14}='Orbit_Scenario_File         ';
		expected_DS_Names{15}='Digital_Elevation_Model     ';
		expected_DS_Names{16}='Geoid_Model                 ';
		map_DS_to_DSD=[1,2,3,4,5,6,7];
		if (version == 'ADM_L1bP/1.06 ')
			iodd_reference=iodd_reference1b_v106;
		end
		if (version == 'ADM_L1bP/1.07 ')
			iodd_reference=iodd_reference1b_v106;
		end
		if (version == 'ADM_L1bP/1.08 ')
			iodd_reference=iodd_reference1b_v108;
			if (flagv108 ~= 1),
				fprintf('Error: L1B file version should be v1.08 according to MPH\n');
				fprintf('but length of MPH field ACQUISITION_STATION is 20 characters\n');
				return
			end
		end
		if (version == 'ADM_L1bP/1.09 ')
			iodd_reference=iodd_reference1b_v109
		end
	otherwise
		fprintf('Error: L1B file version is not supported: %s\n',version);
		return
	end
case '2A'
	expected_SPH_size = 2731;
	expected_DS_Number = 3;
	expected_DSD_Number = 8;
	expected_DS_Names{1} = 'Geolocation_ADS             ';
	expected_DS_Names{2} = 'PCD_ADS                     ';
	expected_DS_Names{3} = 'Opt_MDS                     ';
	expected_DS_Names{4} = 'L1B_Input_ADS               ';
	expected_DS_Names{5} = 'Met_ADS                     ';
	expected_DS_Names{6} = 'CAL_ADS                     ';
	expected_DS_Names{7} = 'Aer_ext_to_bck              ';
	expected_DS_Names{8} = 'L2A_Param_ADS               ';
	map_DS_to_DSD=[1,2,3];
	iodd_reference=iodd_reference2a;
case '2B'
	expected_SPH_size=4297;
	expected_DS_Number=4;
	expected_DSD_Number=10;
	expected_DS_Names{1} ='Geolocation_ADS             ';
	expected_DS_Names{2} ='Product_Confidence_Data_ADS ';
	expected_DS_Names{3} ='Mie_HLOSwind_MDS            ';
	expected_DS_Names{4} ='Rayleigh_HLOSwind_MDS       ';
	expected_DS_Names{5} ='Aeolus_Level_1B_Product     ';
	expected_DS_Names{6} ='Aux_Met_Product             ';
	expected_DS_Names{7} ='Aeolus_RBC                  ';
	expected_DS_Names{8} ='Clim_Product                ';
	expected_DS_Names{9} ='Cal_Product                 ';
	expected_DS_Names{10}='Level_2B_Proc_Params        ';
	map_DS_to_DSD=[1,2,3,4];
	iodd_reference=iodd_reference2bc;
case '2C'
	expected_SPH_size=5799;
	expected_DS_Number=7;
	expected_DSD_Number=15;
	expected_DS_Names{1} ='Geolocation_ADS             ';
	expected_DS_Names{2} ='Product_Confidence_Data_ADS ';
	expected_DS_Names{3} ='Mie_HLOSwind_MDS            ';
	expected_DS_Names{4} ='Rayleigh_HLOSwind_MDS       ';
	expected_DS_Names{5} ='Assim_PCD_ADS               ';
	expected_DS_Names{6} ='Mie_VecWind_MDS             ';
	expected_DS_Names{7} ='Rayleigh_VecWind_MDS        ';
	expected_DS_Names{8} ='Aeolus_Level_1B_Product     ';
	expected_DS_Names{9} ='Aux_Met_Product             ';
	expected_DS_Names{10}='Aeolus_RBC                  ';
	expected_DS_Names{11}='Clim_Product                ';
	expected_DS_Names{12}='Cal_Product                 ';
	expected_DS_Names{13}='Level_2B_Proc_Params        ';
	expected_DS_Names{14}='Aeolus_Level_2B_Product     ';
	expected_DS_Names{15}='Level_2C_Proc_Params        ';
	map_DS_to_DSD=[1,2,3,4,5,6,7];
	iodd_reference=iodd_reference2bc;
case 'MET'
	expected_SPH_size=4623;
	expected_DS_Number=4;
	expected_DSD_Number=14;
	expected_DS_Names{1} ='Geolocation_ADS1 off-nadir  ';
	expected_DS_Names{2} ='Geolocation_ADS2 nadir      ';
	expected_DS_Names{3} ='Meteorological DS1 off-nadir';
	expected_DS_Names{4} ='Meteorological DS2 nadir    ';
	expected_DS_Names{5} ='Input DS 01                 ';
	expected_DS_Names{6} ='Input DS 02                 ';
	expected_DS_Names{7} ='Input DS 03                 ';
	expected_DS_Names{8} ='Input DS 04                 ';
	expected_DS_Names{9} ='Input DS 05                 ';
	expected_DS_Names{10} ='Input DS 06                 ';
	expected_DS_Names{11} ='Input DS 07                 ';
	expected_DS_Names{12} ='Input DS 08                 ';
	expected_DS_Names{13} ='Input DS 09                 ';
	expected_DS_Names{14} ='Input DS 10                 ';
	map_DS_to_DSD=[1,2,3,4];
	iodd_reference=iodd_reference2bc;
case 'RBC'
	expected_SPH_size=1398;
	expected_DS_Number=1;
	expected_DSD_Number=2;
	expected_DS_Names{1} ='Rayleigh_Brillouin_ADS      ';
	expected_DS_Names{2} ='Fabry_Perot_MDS             ';
	map_DS_to_DSD=[1];
	iodd_reference=iodd_reference2bc;
case 'CLM'
	expected_SPH_size=486;
	expected_DS_Number=1;
	expected_DSD_Number=1;
	expected_DS_Names{1} ='AuxClim_ADS                 ';
	map_DS_to_DSD=[1];
	iodd_reference=iodd_reference2bc;
case 'CAL'
	expected_SPH_size=677;
	expected_DS_Number=1;
	expected_DSD_Number=1;
	expected_DS_Names{1}='CAL_ADS                     ';
	map_DS_to_DSD=[1];
	iodd_reference=iodd_reference2a;
end

% compare MPH with expected information
if nbByteSPH~=expected_SPH_size,
    fprintf('Error when reading MPH: The SPH size has changed. Check file format.\n');
    return;
end
if (nbDS ~= expected_DS_Number),
    fprintf('Error when reading SPH: The number of DS has changed. Check file format.\n');
    return;
end
if (nbDSD ~= expected_DSD_Number),
    fprintf('Error when reading SPH: The number of DSDs has changed. Check file format.\n');
    return;
end

% read SPH
SPH_chain = fread(fid, nbByteSPH, '*char');
switch filetype
case '1B'
	SPH1=SPH_chain(1:1685);
	DSD_st=SPH_chain(1686:end);
	for k = 1:expected_DS_Number,
		DSD(k)=ReadDSD(DSD_st([1:288]+(k-1)*288)');
	end
	xxxData.NFOrder=str2num(SPH1(1574:1583)');
	xxxData.Nmax = str2num(SPH1(351:361)');
	xxxData.nbDSR = str2num(DSD_st(198:208)');
	switch version
	case {'ADM_L1bP/1.04 ','ADM_L1bP/1.05 ','ADM_L1bP/1.06 ','ADM_L1bP/1.07 '}
		expected_iodd_sizes= ...
		[2244+xxxData.Nmax*924,   ...
		 3799+xxxData.Nmax*2022,  ...
		 130+xxxData.Nmax*20,     ...
		 220+xxxData.Nmax*1061,   ...
		 6390+xxxData.NFOrder*32, ...
		 662+xxxData.Nmax*650,    ...
		 447+xxxData.Nmax*434];
	case {'ADM_L1bP/1.08 ','ADM_L1bP/1.09 '}
		expected_iodd_sizes= ...
		[2244+xxxData.Nmax*924,   ...
		 3799+xxxData.Nmax*2022,  ...
		 138+xxxData.Nmax*20,     ...
		 220+xxxData.Nmax*1061,   ...
		 6390+xxxData.NFOrder*32, ...
		 662+xxxData.Nmax*650,    ...
		 447+xxxData.Nmax*434];
	otherwise
		fprintf('Error: L1B file version is not supported: %s\n',version);
		return
	end
case '2A'
	SPH1 = SPH_chain(1:427);
	DSD_st = SPH_chain(428:end);
	for k = 1:expected_DS_Number,
		DSD(k)=ReadDSD(DSD_st([1:288]+(k-1)*288)');
	end
	xxxData.maxNumProf=str2num(SPH1(356:360)');
	xxxData.nbDSR = str2num(SPH1(96:101)');
	current_pos=ftell(fid);
	fseek(fid,xxxData.nbDSR*(18+1452*xxxData.maxNumProf)+12,'cof');
	for i=1:xxxData.nbDSR,       
		numMeas(i)=double(fread(fid,1,'*int16'));
		fseek(fid,110+11*numMeas(i)+971*xxxData.maxNumProf,'cof');
	end
	fseek(fid,current_pos,'bof');
	xxxData.numMeas=numMeas;
	for i=1:xxxData.nbDSR,
		expected_iodd_sizes(i,1)=18+xxxData.maxNumProf*1452;
		expected_iodd_sizes(i,2)=112+11*xxxData.numMeas(i)+971*xxxData.maxNumProf;
		expected_iodd_sizes(i,3)=16+72*xxxData.numMeas(i)+2164*xxxData.maxNumProf;
	end	
case '2B'
	SPH=ReadL2bcSPH(SPH_chain,false);
	DSD=SPH.DSD;
	xxxData.SPH=SPH;
	xxxData.Nmax  = SPH.M_Meas;
	xxxData.nbDSR = SPH.Num_BRC;
	xxxData.M_Mie = SPH.M_Mie;
	xxxData.M_Rayleigh = SPH.M_Rayleigh;
	expected_iodd_sizes= ...
	[20+xxxData.M_Mie*1452+1452*xxxData.M_Rayleigh,   ...
	 180+2235*xxxData.M_Mie+1179*xxxData.M_Rayleigh+635*xxxData.Nmax, ...
	 18+253*xxxData.M_Mie+72*xxxData.Nmax, ...
	 18+637*xxxData.M_Rayleigh+72*xxxData.Nmax];
case '2C'
	SPH = ReadL2bcSPH(SPH_chain,true);
	DSD=SPH.DSD;
	xxxData.SPH=SPH;
	xxxData.Nmax  = SPH.M_Meas;
	xxxData.nbDSR = SPH.Num_BRC
	xxxData.M_Mie = SPH.M_Mie;
	xxxData.M_Rayleigh = SPH.M_Rayleigh;
	expected_iodd_sizes= ...
	[20+xxxData.M_Mie*1452+1452*xxxData.M_Rayleigh,   ...
	 180+2235*xxxData.M_Mie+1179*xxxData.M_Rayleigh+635*xxxData.Nmax, ...
	 18+253*xxxData.M_Mie+72*xxxData.Nmax, ...
	 18+637*xxxData.M_Rayleigh+72*xxxData.Nmax, ...
	 52+2293*xxxData.M_Mie+2293*xxxData.M_Rayleigh, ...
	 16+733*xxxData.M_Mie, ...
	 16+733*xxxData.M_Rayleigh];
case 'MET'
	SPH = ReadMETSPH(SPH_chain);
	DSD=SPH.DSD;
	xxxData.SPH=SPH;
	xxxData.Num_of_Model_Layers=SPH.Num_of_Model_Layers;
	xxxData.Num_Records_in_DS1=SPH.Num_Records_in_DS1;
	xxxData.Num_Records_in_DS2=SPH.Num_Records_in_DS2;
	expected_iodd_sizes= ...
	[24, ...
	 24, ...
	 26+xxxData.Num_of_Model_Layers*71, ...
	 26+xxxData.Num_of_Model_Layers*71];
case 'RBC'
	SPH = ReadRBCSPH(SPH_chain);
	DSD=SPH.DSD;
	xxxData.SPH=SPH;
	xxxData.Num_P=SPH.Num_P;
	xxxData.Num_T=SPH.Num_T;
	xxxData.Num_F=SPH.Num_F;
	xxxData.Num_FP=SPH.Num_FP;
	xxxData.Num_fd=SPH.Num_fd;
	xxxData.Num_RR=SPH.Num_RR;
	expected_iodd_sizes= ...
	[xxxData.Num_P*4+xxxData.Num_T*2+xxxData.Num_F*8+ ...
	 xxxData.Num_P*xxxData.Num_T*xxxData.Num_F*8+ ...
	 xxxData.Num_FP*(8+8+8)+xxxData.Num_fd*8+ ...
	 xxxData.Num_RR*8+xxxData.Num_P*xxxData.Num_T*(xxxData.Num_RR*(8+8))+ ...
	 xxxData.Num_P*xxxData.Num_T*xxxData.Num_fd*(4+4)+xxxData.Num_RR*8];
case 'CLM'
	SPH = ReadCLMSPH(SPH_chain);
	DSD=SPH.DSD;
	xxxData.SPH=SPH;
	expected_iodd_sizes=[xxxData.SPH.DSD(1).DS_Size];
case 'CAL'
	SPH=SPH_chain(1:389);
	DSD_st=SPH_chain(390:677);
	for k = 1:expected_DS_Number,
		DSD(k)=ReadDSD(DSD_st([1:288]+(k-1)*288)');
	end
	xxxData.SPH=SPH;
	xxxData.Num_P=str2num(SPH(131:134)');
	xxxData.Num_T=str2num(SPH(142:145)');
	xxxData.Num_fd=str2num(SPH(154:157)');
	xxxData.Num_FP=str2num(SPH(166:170)');
	expected_iodd_sizes=...
	[16+xxxData.Num_P*4+xxxData.Num_T*2+xxxData.Num_fd*24+...
	xxxData.Num_FP*32+xxxData.Num_P*xxxData.Num_T*xxxData.Num_fd*16];
otherwise
	fprintf('Programming error: filetype not known: %s\n',filetype);
	return;
end
if ~strcmp(filetype,'2A'),
	for k = 1:expected_DS_Number,
		DSD_Number=map_DS_to_DSD(k);
		if (strcmp(DSD(DSD_Number).DS_Name,expected_DS_Names{k}) == 0),
			fprintf('Error: DSD %02d is |%s|\n',k,DSD(DSD_Number).DS_Name);
			fprintf('Expected         |%s|\n',expected_DS_Names{k}')
			return;
		end
		expected_iodd_sizes(k)=expected_iodd_sizes(k)*DSD(DSD_Number).Num_DSR;
	end
end

fprintf('List of DSDs found in SPH is consistent with IODD %s\n',iodd_reference);
fprintf('------------------------------------------------------------------------\n');
switch filetype
case '1B'
	fprintf('Number of|BRCs | Meas per BRC | harmonics | DSDs | datasets\n')
	fprintf('Found    |%4d | %12d | %9d | %4d | %8d\n',xxxData.nbDSR,xxxData.Nmax,xxxData.NFOrder,nbDSD,nbDS);
case {'2A'}
	fprintf('Number of|BRCs | Max Number Profiles | DSDs | datasets\n')
	fprintf('Found    |%4d | %19d | %4d | %8d\n',xxxData.nbDSR,xxxData.maxNumProf,nbDSD,nbDS);
case {'2B','2C'}
	fprintf('Number of|BRCs | Meas per BRC | M_Mie | M_Rayleigh | DSDs | datasets\n')
	fprintf('Found    |%4d | %12d | %5d | %10d | %4d | %8d\n',xxxData.nbDSR,xxxData.Nmax,xxxData.M_Mie,xxxData.M_Rayleigh,nbDSD,nbDS);
case 'MET'
	fprintf('Number of|BRCs off-nadir | BRCs nadir | model layers | DSDs | datasets\n')
	fprintf('Found    |%14d | %10d | %12d | %4d | %8d\n', ...
	   xxxData.Num_Records_in_DS1,xxxData.Num_Records_in_DS2,xxxData.Num_of_Model_Layers,nbDSD,nbDS);
case 'RBC'
	fprintf('Number of|    P |    T |    F |   FP |   fd |   RR | DSDs | datasets\n');
	fprintf('Found    | %4d | %4d | %4d | %4d | %4d | %4d | %4d | %8d\n', ...
	   xxxData.Num_P,xxxData.Num_T,xxxData.Num_F,xxxData.Num_FP,xxxData.Num_fd,xxxData.Num_RR,nbDSD,nbDS);
case 'CLM'
	fprintf('Number of| DSDs | datasets\n');
	fprintf('Found    | %4d | %8d\n', ...
	   nbDSD,nbDS);
case 'CAL'
       	fprintf('Number of|    P |    T |    fd |   FP | DSDs | datasets\n');
	fprintf('Found	  | %4d | %4d | %4d | %4d | %4d | %8d\n',...
	   xxxData.Num_P,xxxData.Num_T,xxxData.Num_fd,xxxData.Num_FP,nbDSD,nbDS);
otherwise
	fprintf('Programming error: filetype %s not recognized\n',filetype);
	return;
end
fprintf('------------------------------------------------------------------------\n');

% interpret user inputs
DSRs_to_read=varargin;
if (length(DSRs_to_read)==1),
    DSR_to_read = DSRs_to_read{1};
    switch upper(DSR_to_read)
    case 'ALL'
		switch filetype
		case '1B'
		        DSRs_to_read=[];
		        DSRs_to_read{1}='GEO_ADS';
		        DSRs_to_read{2}='PCD_ADS';
		        DSRs_to_read{3}='GWD_ADS';
		        DSRs_to_read{4}='MEAS_ADS';
		        DSRs_to_read{5}='CCD_GADS';
		        DSRs_to_read{6}='US_MDS';
		        DSRs_to_read{7}='WV_MDS';
		case '2A'
			DSRs_to_read{1}='GEO_ADS';
			DSRs_to_read{2}='PCD_ADS';
			DSRs_to_read{3}='OPT_MDS';
		case '2B'
		        DSRs_to_read=[];
		        DSRs_to_read{1}='GEO_ADS';
		        DSRs_to_read{2}='PCD_ADS';
		        DSRs_to_read{3}='MIE_MDS';
		        DSRs_to_read{4}='RAYLEIGH_MDS';
		case '2C'
		        DSRs_to_read=[];
		        DSRs_to_read{1}='GEO_ADS';
		        DSRs_to_read{2}='PCD_ADS';
		        DSRs_to_read{3}='MIE_MDS';
		        DSRs_to_read{4}='RAYLEIGH_MDS';
		        DSRs_to_read{5}='ASSIM_PCD_ADS';
		        DSRs_to_read{6}='MIE_VEC_MDS';
		        DSRs_to_read{7}='RAYLEIGH_VEC_MDS';
		case 'MET'
			DSRs_to_read=[];
			DSRs_to_read{1}='GEO_ADS1';
			DSRs_to_read{2}='GEO_ADS2';
			DSRs_to_read{3}='MET_MDS1';
			DSRs_to_read{4}='MET_MDS2';
		case 'RBC'
			DSRs_to_read=[];
			DSRs_to_read{1}='RB_ADS';
		case 'CLM'
			DSRs_to_read=[];
			DSRs_to_read{1}='CLIM_ADS';
		case 'CAL'
			DSRs_to_read=[];
			DSRs_to_read{1}='CAL_ADS';
		otherwise
			fprintf('Programming error: filetype is not recognized: %s\n',filetype);
			return;
		end
    end
end

if (length(DSRs_to_read)==0),
	fprintf('Error: check input. No dataset requested.\n');
	return;
end

% loop through user-requested data sets
fprintf('-----------------------------------------------------------------------------\n');
fprintf('DS_Name                     |  Size(IODD) |   Size(DSD) |  Bytes read |Result\n');
fprintf('                            |%s| from header |  from file  |\n',iodd_reference);
fprintf('-----------------------------------------------------------------------------\n');
%       'Product_Confidence_Data_ADS    0123456789  0123456789  0123456789
%       '%s   %010d  %010d  %010d  %s\n'
readok=true;
for i=1:length(DSRs_to_read),
    % identify DS number
    DSR_to_read = DSRs_to_read{i};
    switch upper(DSR_to_read)
        case {'GEO_ADS','GEO_ADS1','RB_ADS','CLIM_ADS','CAL_ADS'}
		DS_Number=1;
        case {'PCD_ADS','GEO_ADS2'}
		DS_Number=2;
        case {'GWD_ADS','MIE_MDS','MET_MDS1','OPT_MDS'}
		DS_Number=3;
        case {'MEAS_ADS','RAYLEIGH_MDS','MET_MDS2'}
		DS_Number=4;
        case {'CCD_GADS','ASSIM_PCD_ADS'}
		DS_Number=5;
        case {'US_MDS','MIE_VEC_MDS'}
		DS_Number=6;
        case {'WV_MDS','RAYLEIGH_VEC_MDS'}
		DS_Number=7;
        otherwise
            fprintf('Error: check dataset name argument: %s \n', DSR_to_read);
	    return;
    end
    DSD_Number=map_DS_to_DSD(DS_Number);
    % move pointer to beginning of DS
    status = fseek(fid,DSD(DSD_Number).DS_Offset, 'bof');
    if status~=0,
	    fprintf('Error while reading DS %s. DS_Offset seems wrong\n',DSR_to_read);
	    return;
    end
    % start reading (save file position)
    Start_of_DS=ftell(fid);
    % select reading routine
    switch DS_Number
    case 1
	    switch filetype
	    case '1B'
	            xxxData.Geolocation_ADS = L1b_Read_Geo_ADS(fid,DSD(DSD_Number).Num_DSR,xxxData.Nmax);
	    case '2A'
		    xxxData.Geolocation_ADS = L2a_Read_Geo_ADS(fid,xxxData.nbDSR,xxxData.maxNumProf);
	    case '2B'	    
	            xxxData.Geolocation_ADS = L2b_Read_Geo_ADS(fid,DSD(DSD_Number).Num_DSR,xxxData.M_Mie,xxxData.M_Rayleigh);		    
	    case 'MET'
		    xxxData.Geolocation_ADS1 = MET_Read_Geo_ADS(fid,DSD(DSD_Number).Num_DSR);
	    case 'RBC'
		    xxxData.Rayleigh_Brillouin_ADS = RBC_Read_RB_ADS(fid,DSD(DSD_Number).Num_DSR, ...
		    	xxxData.Num_P,xxxData.Num_T,xxxData.Num_F,xxxData.Num_FP,xxxData.Num_fd,xxxData.Num_RR);
            case 'CLM'
		    xxxData.Clim_ADS = CLM_Read_ADS(fid,DSD(DSD_Number).Num_DSR);
	    case 'CAL'
		    xxxData.Cal_ADS = Cal_Read_ADS(fid,DSD(DSD_Number).Num_DSR,...
		    	xxxData.Num_P,xxxData.Num_T,xxxData.Num_fd,xxxData.Num_FP);
	    otherwise
		    fprintf('Error: filetype %s not recognized for DS %s\n',filetype,DSR_to_read);
		    return;
	    end
    case 2
	    switch filetype
	    case '1B'
		    xxxData.Product_Confidence_Data_ADS = L1b_Read_PCD_ADS(fid,DSD(DSD_Number).Num_DSR, xxxData.Nmax, version);
	    case '2A'
		    xxxData.Product_Confidence_Data_ADS = L2a_Read_PCD_ADS(fid,xxxData.nbDSR,xxxData.maxNumProf);
	    case {'2B','2C'}
	            xxxData.Product_Confidence_Data_ADS = L2b_Read_PCD_ADS(fid,DSD(DSD_Number).Num_DSR, xxxData.M_Mie, xxxData.M_Rayleigh, xxxData.Nmax);
	    case 'MET'
		    xxxData.Geolocation_ADS2 = MET_Read_Geo_ADS(fid,DSD(DSD_Number).Num_DSR);
	    end
    case 3
	    switch filetype
	    case '1B'
	            xxxData.Ground_Wind_Detection_ADS = L1b_Read_GWD_ADS(fid,DSD(DSD_Number).Num_DSR, xxxData.Nmax,version);
	    case '2A'
		    xxxData.Optical_Properties_ADS = L2a_Read_OPT_ADS(fid,xxxData.nbDSR,xxxData.maxNumProf);
	    case {'2B','2C'}
	            xxxData.Mie_HLOSwind_MDS = L2b_Read_Mie_MDS(fid,DSD(DSD_Number).Num_DSR, xxxData.M_Mie, xxxData.Nmax);
	    case 'MET'
		    xxxData.Met_MDS1 = MET_Read_Met_MDS(fid,DSD(DSD_Number).Num_DSR,xxxData.Num_of_Model_Layers);
	    end
    case 4
	    switch filetype
	    case '1B'
		    xxxData.Measurement_ADS = L1b_Read_Meas_ADS(fid,DSD(DSD_Number).Num_DSR, xxxData.Nmax);
	    case {'2B','2C'}
		    xxxData.Rayleigh_HLOSwind_MDS = L2b_Read_Rayleigh_MDS(fid,DSD(DSD_Number).Num_DSR, xxxData.M_Rayleigh, xxxData.Nmax);
	    case 'MET'
		    xxxData.Met_MDS2 = MET_Read_Met_MDS(fid,DSD(DSD_Number).Num_DSR,xxxData.Num_of_Model_Layers);
	    end
    case 5
	    switch filetype
	    case '1B'
	            xxxData.Calibration_Char_Data_GADS = L1b_Read_CCD_GADS(fid);
	    case '2C'
	            xxxData.Assim_PCD_ADS = L2c_Read_Assim_PCD_ADS(fid,DSD(DSD_Number).Num_DSR, xxxData.M_Mie, xxxData.M_Rayleigh);
	    end
    case 6
	    switch filetype
	    case '1B'
	            xxxData.Useful_Signal_MDS = L1b_Read_US_MDS(fid,DSD(DSD_Number).Num_DSR, xxxData.Nmax);
	    case '2C'
		    xxxData.Mie_VecWind_MDS = L2c_Read_Mie_Vec_MDS(fid,DSD(DSD_Number).Num_DSR, xxxData.M_Mie);
	    end
    case 7
	    switch filetype
	    case '1B'
		    xxxData.Wind_Velocity_MDS = L1b_Read_WV_ADS(fid,DSD(DSD_Number).Num_DSR, xxxData.Nmax);
	    case '2C'
		    xxxData.Rayleigh_VecWind_MDS = L2c_Read_Rayleigh_Vec_MDS(fid,DSD(DSD_Number).Num_DSR, xxxData.M_Rayleigh);
	    end
    otherwise
	    fprintf('Programming Error: requesting DSD_Number out-of-range');
	    return;
    end
    if ~strcmp(filetype,'2A')
	% check DS size
	End_of_DS=ftell(fid);
	Size_of_DS=(End_of_DS-Start_of_DS);
	if (Size_of_DS ~= DSD(DSD_Number).DS_Size | Size_of_DS ~= expected_iodd_sizes(DS_Number)),
		status='FAIL';
		readok=false;
	else
		status=' OK ';
	end
	% print size report (useful to catch IODD or software errors)
    	fprintf('%s|  %010d |  %010d |  %010d | %s\n', ...
     		 DSD(DSD_Number).DS_Name,expected_iodd_sizes(DS_Number),DSD(DSD_Number).DS_Size,Size_of_DS,status);
    else
	% check DS size
	End_of_DS=ftell(fid);
    	Size_of_DS=(End_of_DS-Start_of_DS);
	if (Size_of_DS ~= DSD(DSD_Number).DS_Size | Size_of_DS ~= sum(expected_iodd_sizes(:,DS_Number))),
		status='FAIL';
		readok=false;
	else
		status=' OK ';
	end
	% print size report (useful to catch IODD or software errors)
    	fprintf('%s|  %010d |  %010d |  %010d | %s\n', ...
     		 DSD(DSD_Number).DS_Name,sum(expected_iodd_sizes(:,DS_Number),1),DSD(DSD_Number).DS_Size,Size_of_DS,status);
    end

end
fprintf('-----------------------------------------------------------------------------\n');

fclose (fid);
fprintf('Reading terminated normally: ')
if (~readok),
	fprintf('ERROR: Check the matlab reading code versus the IODD %s\n',iodd_reference);
else
	fprintf('no error found\n')
end
end % function

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       MATLAB CODE TO READ MAIN PRODUCT HEADER
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% HISTORY
%
% 04/2006 P. Poli    Original code.
% 05/2007 P. Poli    Moved to ReadBinaryData
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DOCUMENTATION
%
% The function ReadMPH decodes the string of characters
% making up the Main Product Header (MPH).
%
% The output is a structure containing the elements of 
% the MPH.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [MPH,flagv108] = ReadMPH(MPH_chain)

MPH1=MPH_chain(1:1247);

% decode variables contained in MPH

MPH.Product=(MPH1(9+[1:62])');
MPH.Proc_Stage=(MPH1(84+[1:1])');
MPH.Ref_Doc=(MPH1(95+[1:23])');
%MPH.Spare_1
% determine v1.08 or earlier
if (MPH1((182+21):(182+21)) == '"'),
	MPH.Acquisition_Station=(MPH1(182+[1:20])');
	decv108=0;
	flagv108=0;
else
	MPH.Acquisition_Station=(MPH1(182+[1:24])');
	decv108=4;
	flagv108=1;
end

MPH.Proc_Center=(MPH1(decv108+217+[1:6])');
MPH.Proc_Time=(MPH1(decv108+236+[1:27])');
MPH.Software_Ver=(MPH1(decv108+279+[1:14])');
%MPH.Spare_2
MPH.Sensing_Start=(MPH1(decv108+351+[1:27])');
MPH.Sensing_Stop=(MPH1(decv108+394+[1:27])');
%MPH.Spare_3
MPH.Phase=(MPH1(decv108+470+[1:1])');
MPH.Cycle=str2num(MPH1(decv108+478+[1:4])');
MPH.Rel_Orbit=str2num(MPH1(decv108+493+[1:6])');
MPH.Abs_Orbit=str2num(MPH1(decv108+510+[1:6])');
MPH.State_Vector_Time=(MPH1(decv108+536+[1:27])');
MPH.Delta_UT1=str2num(MPH1(decv108+575+[1:8])');
MPH.X_Position=str2num(MPH1(decv108+598+[1:12])');
MPH.Y_Position=str2num(MPH1(decv108+625+[1:12])');
MPH.Z_Position=str2num(MPH1(decv108+652+[1:12])');
MPH.X_Velocity=str2num(MPH1(decv108+679+[1:12])');
MPH.Y_Velocity=str2num(MPH1(decv108+708+[1:12])');
MPH.Z_Velocity=str2num(MPH1(decv108+737+[1:12])');
MPH.Vector_Source=(MPH1(decv108+770+[1:2])');
%MPH.Spare_4
MPH.Utc_Sbt_Time=(MPH1(decv108+829+[1:27])');
MPH.Sat_Binary_Time=str2num(MPH1(decv108+874+[1:11])');
MPH.Clock_Step=str2num(MPH1(decv108+897+[1:11])');
%MPH.Spare_5
MPH.Leap_Utc=(MPH1(decv108+956+[1:27])');
MPH.Leap_Sign=str2num(MPH1(decv108+995+[1:4])');
MPH.Leap_Err=(MPH1(decv108+1009+[1:1])');
%MPH.Spare_6
MPH.Product_Err=(MPH1(decv108+1064+[1:1])');
MPH.Tot_Size=str2num(MPH1(decv108+1075+[1:21])');
MPH.Sph_Size=str2num(MPH1(decv108+1113+[1:11])');
MPH.Num_Dsd=str2num(MPH1(decv108+1140+[1:11])');
MPH.Dsd_Size=str2num(MPH1(decv108+1161+[1:11])');
MPH.Num_Data_Sets=str2num(MPH1(decv108+1194+[1:11])');
%MPH.Spare_7

end % end function ReadMPH

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       MATLAB CODE TO READ DSDs
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% HISTORY
%
% 03/2006 P. Poli    Original code.
% 07/2006 D. Tan     Adjust offsets by 1.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DOCUMENTATION
%
% The function ReadDSD decodes the string of characters
% making up the Data Set Descriptors (DSD).
%
% The output is a structure containing the elements of the DSD:
% DSD=struct('DS_Name', {}, 'Filename', {}, 'DS_Offset', {}, ...
%            'DS_Size', {}, 'Num_DSR', {}, 'DSR_Size', {}, 'Byte_Order', {});
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [DSD] = ReadDSD(DSD_Chain);

DSD.DS_Name   =DSD_Chain(9+[1:28]);
DSD.Filename  =DSD_Chain(59+[1:62]);
DSD.DS_Offset =str2num(DSD_Chain(133+[1:21]));
DSD.DS_Size   =str2num(DSD_Chain(170+[1:11]));
DSD.Num_DSR   =str2num(DSD_Chain(197+[1:11]));
DSD.DSR_Size  =str2num(DSD_Chain(218+[1:11]));
DSD.Byte_Order=str2num(DSD_Chain(249+[1:4]));

end % function

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       MATLAB CODE TO READ L2b/c SPH
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% HISTORY
%
% 11/2005 P. Poli    Original code.
% 12/2005 P. Poli    Code interfaced with L2BP version
%                    NewClassification and L2B writing,
%                    consistent with TN2.1 v1.2 09-06-2005
% 03/2006 P. Poli    Upgrade to read outputs from L2BP V0.2
% 04/2006 P. Poli    Upgrade to IODD AE_IF_ECMWF_L2BP_001_20060427
% 05/2006 D. Tan     Upgrade to IODD AE_IF_ECMWF_L2BP_001_20060502 (M_Meas)
% 05/2006 D. Tan     Reduce SPH_Chain length by one
% 07/2006 D. Tan     Re-order DSDs, increase to 9, and rename; additional L1B stats
% 02/2007 P. Poli    Read also L2c (add logical argument flag_2c)
%                    Use IODD variable names
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DOCUMENTATION
%
% The function ReadL2bcSPH decodes the string of characters
% making up the L2B Specific Product Header (SPH).
%
% The output is a structure containing the elements of 
% the SPH.
%
% Second input argument flag_2c: logical:
%   Set to false in order to read L2b data
%   Set to true  in order to read L2c data
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function SPH = ReadL2bcSPH(SPH_chain,flag_2c)

SPH1=SPH_chain(1:1417);
if (flag_2c),
SPH1=SPH_chain(1:1479);
end

% decode variables contained in SPH

SPH.Sph_Descriptor=(SPH1(16+[1:28])');
SPH.Num_BRC=str2num(SPH1(95+[1:6])');
SPH.Intersect_Start_Lat=str2num(SPH1(122+[1:11])');
SPH.Intersect_Start_Long=str2num(SPH1(165+[1:11])');
SPH.Intersect_Stop_Lat=str2num(SPH1(206+[1:11])');
SPH.Intersect_Stop_Long=str2num(SPH1(248+[1:11])');
SPH.Sat_Track=str2num(SPH1(280+[1:11])');
SPH.M_Mie=str2num(SPH1(344+[1:4])');
SPH.M_Rayleigh=str2num(SPH1(360+[1:4])');
SPH.M_Meas=str2num(SPH1(372+[1:4])');
SPH.Num_Valid_Mie_Profiles=str2num(SPH1(400+[1:6])');
SPH.Num_Valid_Rayleigh_Profiles=str2num(SPH1(435+[1:6])');
SPH.Num_Clear_Mie_Profiles=str2num(SPH1(465+[1:6])');
SPH.Num_Clear_Rayleigh_Profiles=str2num(SPH1(500+[1:6])');
SPH.Num_Cloud_Mie_Profiles=str2num(SPH1(530+[1:6])');
SPH.Num_Cloud_Rayleigh_Profiles=str2num(SPH1(565+[1:6])');
SPH.Num_Mie_Prof_Warm_Pulses=str2num(SPH1(597+[1:6])');
SPH.Num_Rayleigh_Prof_Warm_Pulses=str2num(SPH1(634+[1:6])');
SPH.Num_Profiles_Surface_Mie=str2num(SPH1(666+[1:6])');
SPH.Num_Profiles_Surface_Ray=str2num(SPH1(698+[1:6])');
SPH.Num_Valid_Obs_Prof_L1B_Mie =str2num(SPH1(732+[1:6])');
SPH.Num_Valid_Obs_Prof_L1B_Ray =str2num(SPH1(766+[1:6])');
SPH.Num_Valid_Meas_Prof_L1B_Mie=str2num(SPH1(801+[1:6])');
SPH.Num_Valid_Meas_Prof_L1B_Ray=str2num(SPH1(836+[1:6])');
SPH.Num_Valid_Obs_Bins_L1B_Mie =str2num(SPH1(870+[1:6])');
SPH.Num_Valid_Obs_Bins_L1B_Ray =str2num(SPH1(904+[1:6])');
SPH.Num_Valid_Meas_Bins_L1B_Mie=str2num(SPH1(939+[1:6])');
SPH.Num_Valid_Meas_Bins_L1B_Ray=str2num(SPH1(974+[1:6])');
SPH.Num_Invalid_Meas_Prof_L1B_Mie=str2num(SPH1(1011+[1:6])');
SPH.Num_Invalid_Meas_Prof_L1B_Ray=str2num(SPH1(1048+[1:6])');
SPH.Num_Invalid_Meas_Prof_L2B_Mie=str2num(SPH1(1085+[1:6])');
SPH.Num_Invalid_Meas_Prof_L2B_Ray=str2num(SPH1(1122+[1:6])');
SPH.Num_Invalid_Meas_Bins_L1B_Mie=str2num(SPH1(1159+[1:6])');
SPH.Num_Invalid_Meas_Bins_L1B_Ray=str2num(SPH1(1196+[1:6])');
SPH.Num_Invalid_Meas_Bins_L2B_Mie=str2num(SPH1(1233+[1:6])');
SPH.Num_Invalid_Meas_Bins_L2B_Ray=str2num(SPH1(1270+[1:6])');
SPH.Num_Valid_Obs_L2B_Mie=str2num(SPH1(1340+[1:6])');
SPH.Num_Valid_Obs_L2B_Ray=str2num(SPH1(1369+[1:6])');

if (flag_2c),
SPH.Num_Invalid_Meas_Bins_L2C_Mie=str2num(SPH1(1400+[1:6])');
SPH.Num_Invalid_Meas_Bins_L2C_Ray=str2num(SPH1(1431+[1:6])');
end

% Read DSDs
DSDs=SPH_chain(1418:end);
if (flag_2c),
DSDs=SPH_chain(1480:end);
end
for k=1:9,
	SPH.DSD(k)=ReadDSD(DSDs([288*(k-1)+1:288*(k-1)+288])');
end
SPH.Geoloc_ADS_DSD=SPH.DSD(1);
SPH.PCD_ADS_DSD=SPH.DSD(2);
SPH.Mie_MDS_DSD=SPH.DSD(3);
SPH.Ray_MDS_DSD=SPH.DSD(4);
SPH.L1B_Product_DSD=SPH.DSD(5);
SPH.Met_Product_DSD=SPH.DSD(6);
SPH.RBC_Product_DSD=SPH.DSD(7);
SPH.CLM_Product_DSD=SPH.DSD(8);
SPH.L2B_Param_ADS_DSD=SPH.DSD(9);
if (flag_2c)
for k=10:14,
	SPH.DSD(k)=ReadDSD(DSDs([288*(k-1)+1:288*(k-1)+288])');
end
SPH.L2B_Product_DSD=SPH.DSD(10);
SPH.L2C_Param_ADS_DSD=SPH.DSD(11);
SPH.Assim_PCD_ADS_DSD=SPH.DSD(12);
SPH.Mie_VecWind_MDS_DSD=SPH.DSD(13);
SPH.Rayleigh_VecWind_MDS_DSD=SPH.DSD(14);
end
% ]

end %function

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       MATLAB CODE TO READ L2b Geolocation ADS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% 02/2007 P. Poli   New code. Uses IODD names.

function Geolocation_ADS = L2b_Read_Geo_ADS(fid,nbDSR_Geo,M_Mie,M_Rayleigh)

Geolocation_ADS = [];

for idsr = 1:nbDSR_Geo,
    Geolocation_ADS.DSR(idsr).Start_of_Obs_Time(1) = fread(fid, 1,'*int32');
    Geolocation_ADS.DSR(idsr).Start_of_Obs_Time(2:3) = fread(fid, 2,'*uint32');
    Geolocation_ADS.DSR(idsr).N_Obs_Mie_Actual = fread(fid, 1,'*int16','ieee-be');
    Geolocation_ADS.DSR(idsr).N_Obs_Rayleigh_Actual = fread(fid, 1,'*int16','ieee-be');
    % Mie_Geolocation
    for iprof = 1:M_Mie,
	    for ibin=1:24,
		    Geolocation_ADS.DSR(idsr).Mie_Geolocation(iprof).List_of_Profile_Height_Bin_Geolocation(ibin).Latitude_Start = fread(fid, 1,'*int32','ieee-be');
		    Geolocation_ADS.DSR(idsr).Mie_Geolocation(iprof).List_of_Profile_Height_Bin_Geolocation(ibin).Latitude_Stop = fread(fid, 1,'*int32','ieee-be');
		    Geolocation_ADS.DSR(idsr).Mie_Geolocation(iprof).List_of_Profile_Height_Bin_Geolocation(ibin).Latitude_COG = fread(fid, 1,'*int32','ieee-be');
		    Geolocation_ADS.DSR(idsr).Mie_Geolocation(iprof).List_of_Profile_Height_Bin_Geolocation(ibin).Longitude_Start = fread(fid, 1,'*int32','ieee-be');
		    Geolocation_ADS.DSR(idsr).Mie_Geolocation(iprof).List_of_Profile_Height_Bin_Geolocation(ibin).Longitude_Stop = fread(fid, 1,'*int32','ieee-be');
		    Geolocation_ADS.DSR(idsr).Mie_Geolocation(iprof).List_of_Profile_Height_Bin_Geolocation(ibin).Longitude_COG = fread(fid, 1,'*int32','ieee-be');
		    Geolocation_ADS.DSR(idsr).Mie_Geolocation(iprof).List_of_Profile_Height_Bin_Geolocation(ibin).Altitude_Bottom = fread(fid, 1,'*int32','ieee-be');
		    Geolocation_ADS.DSR(idsr).Mie_Geolocation(iprof).List_of_Profile_Height_Bin_Geolocation(ibin).Altitude_Top = fread(fid, 1,'*int32','ieee-be');
		    Geolocation_ADS.DSR(idsr).Mie_Geolocation(iprof).List_of_Profile_Height_Bin_Geolocation(ibin).Altitude_COG = fread(fid, 1,'*int32','ieee-be');
		    Geolocation_ADS.DSR(idsr).Mie_Geolocation(iprof).List_of_Profile_Height_Bin_Geolocation(ibin).LOS_Azimuth = fread(fid, 1,'*float64','ieee-be');
		    Geolocation_ADS.DSR(idsr).Mie_Geolocation(iprof).List_of_Profile_Height_Bin_Geolocation(ibin).LOS_Elevation = fread(fid, 1,'*float64','ieee-be');
		    Geolocation_ADS.DSR(idsr).Mie_Geolocation(iprof).List_of_Profile_Height_Bin_Geolocation(ibin).LOS_Satellite_Velocity = fread(fid, 1,'*float64','ieee-be');
	    end
	    Geolocation_ADS.DSR(idsr).Mie_Geolocation(iprof).Latitude_of_DEM_Intersection = fread(fid, 1,'*int32','ieee-be');
	    Geolocation_ADS.DSR(idsr).Mie_Geolocation(iprof).Longitude_of_DEM_Intersection = fread(fid, 1,'*int32','ieee-be');
	    Geolocation_ADS.DSR(idsr).Mie_Geolocation(iprof).Altitude_of_DEM_Intersection = fread(fid, 1,'*int32','ieee-be');
    end
    % Rayleigh geolocation
    for iprof = 1:M_Rayleigh,
	    for ibin=1:24,
		    Geolocation_ADS.DSR(idsr).Rayleigh_Geolocation(iprof).List_of_Profile_Height_Bin_Geolocation(ibin).Latitude_Start = fread(fid, 1,'*int32','ieee-be');
		    Geolocation_ADS.DSR(idsr).Rayleigh_Geolocation(iprof).List_of_Profile_Height_Bin_Geolocation(ibin).Latitude_Stop = fread(fid, 1,'*int32','ieee-be');
		    Geolocation_ADS.DSR(idsr).Rayleigh_Geolocation(iprof).List_of_Profile_Height_Bin_Geolocation(ibin).Latitude_COG = fread(fid, 1,'*int32','ieee-be');
		    Geolocation_ADS.DSR(idsr).Rayleigh_Geolocation(iprof).List_of_Profile_Height_Bin_Geolocation(ibin).Longitude_Start = fread(fid, 1,'*int32','ieee-be');
		    Geolocation_ADS.DSR(idsr).Rayleigh_Geolocation(iprof).List_of_Profile_Height_Bin_Geolocation(ibin).Longitude_Stop = fread(fid, 1,'*int32','ieee-be');
		    Geolocation_ADS.DSR(idsr).Rayleigh_Geolocation(iprof).List_of_Profile_Height_Bin_Geolocation(ibin).Longitude_COG = fread(fid, 1,'*int32','ieee-be');
		    Geolocation_ADS.DSR(idsr).Rayleigh_Geolocation(iprof).List_of_Profile_Height_Bin_Geolocation(ibin).Altitude_Bottom = fread(fid, 1,'*int32','ieee-be');
		    Geolocation_ADS.DSR(idsr).Rayleigh_Geolocation(iprof).List_of_Profile_Height_Bin_Geolocation(ibin).Altitude_Top = fread(fid, 1,'*int32','ieee-be');
		    Geolocation_ADS.DSR(idsr).Rayleigh_Geolocation(iprof).List_of_Profile_Height_Bin_Geolocation(ibin).Altitude_COG = fread(fid, 1,'*int32','ieee-be');
		    Geolocation_ADS.DSR(idsr).Rayleigh_Geolocation(iprof).List_of_Profile_Height_Bin_Geolocation(ibin).LOS_Azimuth = fread(fid, 1,'*float64','ieee-be');
		    Geolocation_ADS.DSR(idsr).Rayleigh_Geolocation(iprof).List_of_Profile_Height_Bin_Geolocation(ibin).LOS_Elevation = fread(fid, 1,'*float64','ieee-be');
		    Geolocation_ADS.DSR(idsr).Rayleigh_Geolocation(iprof).List_of_Profile_Height_Bin_Geolocation(ibin).LOS_Satellite_Velocity = fread(fid, 1,'*float64','ieee-be');
	    end
	    Geolocation_ADS.DSR(idsr).Rayleigh_Geolocation(iprof).Latitude_of_DEM_Intersection = fread(fid, 1,'*int32','ieee-be');
	    Geolocation_ADS.DSR(idsr).Rayleigh_Geolocation(iprof).Longitude_of_DEM_Intersection = fread(fid, 1,'*int32','ieee-be');
	    Geolocation_ADS.DSR(idsr).Rayleigh_Geolocation(iprof).Altitude_of_DEM_Intersection = fread(fid, 1,'*int32','ieee-be');
    end
    Geolocation_ADS.DSR(idsr).WGS84_to_Geoid_Altitude = fread(fid, 1,'*int32','ieee-be');
end % for idsr

end % function

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%% Product Confidence %%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% 02/2007 P. Poli   New code. Uses IODD names.

function Product_Confidence_Data_ADS = L2b_Read_PCD_ADS(fid,nbDSR_PCD,M_Mie,M_Rayleigh,Nmax)

Product_Confidence_Data_ADS = [];

for idsr = 1:nbDSR_PCD,
    
    Product_Confidence_Data_ADS.DSR(idsr).N_Meas = fread(fid, 1,'*int16','ieee-be');
    Product_Confidence_Data_ADS.DSR(idsr).N_Obs_Mie_Actual = fread(fid, 1,'*int16','ieee-be');
    Product_Confidence_Data_ADS.DSR(idsr).N_Obs_Rayleigh_Actual = fread(fid, 1,'*int16','ieee-be');
    
    % L1B_Input_Screening
    Product_Confidence_Data_ADS.DSR(idsr).L1B_Input_Screening.L1B_Obs_Screening = fread(fid, 1,'uint8','ieee-be');
    Product_Confidence_Data_ADS.DSR(idsr).L1B_Input_Screening.L1B_Obs_Screening_Flags1 = fread(fid, 1,'uint8','ieee-be');
    Product_Confidence_Data_ADS.DSR(idsr).L1B_Input_Screening.L1B_Obs_Screening_Flags2 = fread(fid, 1,'uint8','ieee-be');
    Product_Confidence_Data_ADS.DSR(idsr).L1B_Input_Screening.L1B_Obs_Screening_Flags3 = fread(fid, 1,'uint8','ieee-be');
    Product_Confidence_Data_ADS.DSR(idsr).L1B_Input_Screening.L1B_Obs_Screening_Flags4 = fread(fid, 1,'uint8','ieee-be');
    Product_Confidence_Data_ADS.DSR(idsr).L1B_Input_Screening.L1B_Obs_Screening_Flags5 = fread(fid, 1,'uint8','ieee-be');
    for imeas = 1:Nmax,
	    Product_Confidence_Data_ADS.DSR(idsr).L1B_Input_Screening.L1B_Mie_Meas_Screening(imeas).L1B_Meas_QC = fread(fid, 1,'uint8','ieee-be');
	    Product_Confidence_Data_ADS.DSR(idsr).L1B_Input_Screening.L1B_Mie_Meas_Screening(imeas).L1B_Meas_QC_Flags = fread(fid, 1,'uint8','ieee-be');
	    for ibin=1:24
		    Product_Confidence_Data_ADS.DSR(idsr).L1B_Input_Screening.L1B_Mie_Meas_Screening(imeas).L1B_Mie_Meas_Bin_Screening(ibin).L1B_Mie_Meas_Bin_QC = fread(fid, 1,'uint8','ieee-be');
		    Product_Confidence_Data_ADS.DSR(idsr).L1B_Input_Screening.L1B_Mie_Meas_Screening(imeas).L1B_Mie_Meas_Bin_Screening(ibin).L1B_Mie_Meas_Bin_QC_Flags = fread(fid, 1,'uint8','ieee-be');
	    end
	    % Spare_1
	    fseek(fid,1,'cof');
    end
    for imeas = 1:Nmax,
	    Product_Confidence_Data_ADS.DSR(idsr).L1B_Input_Screening.L1B_Rayleigh_Meas_Screening(imeas).L1B_Meas_QC = fread(fid, 1,'uint8','ieee-be');
	    Product_Confidence_Data_ADS.DSR(idsr).L1B_Input_Screening.L1B_Rayleigh_Meas_Screening(imeas).L1B_Meas_QC_Flags = fread(fid, 1,'uint8','ieee-be');
	    for ibin=1:24
		    Product_Confidence_Data_ADS.DSR(idsr).L1B_Input_Screening.L1B_Rayleigh_Meas_Screening(imeas).L1B_Rayleigh_Meas_Bin_Screening(ibin).L1B_Rayleigh_Meas_Bin_QC = fread(fid, 1,'uint8','ieee-be');
		    Product_Confidence_Data_ADS.DSR(idsr).L1B_Input_Screening.L1B_Rayleigh_Meas_Screening(imeas).L1B_Rayleigh_Meas_Bin_Screening(ibin).L1B_Rayleigh_Meas_Bin_QC_Flags = fread(fid, 1,'uint8','ieee-be');
	    end
	    % Spare_1
	    fseek(fid,1,'cof');
    end
    Product_Confidence_Data_ADS.DSR(idsr).L1B_MRC_Screening.L1B_MRC_Valid = fread(fid, 1,'uint8','ieee-be');
    % Spare_1
    fseek(fid,5,'cof');
    Product_Confidence_Data_ADS.DSR(idsr).L1B_RRC_Screening.L1B_RRC_Valid = fread(fid, 1,'uint8','ieee-be');
    % Spare_1
    fseek(fid,5,'cof');
    Product_Confidence_Data_ADS.DSR(idsr).L1B_ZWC_Screening.L1B_ZWC_Valid = fread(fid, 1,'uint8','ieee-be');
    % Spare_1
    fseek(fid,5,'cof');
    Product_Confidence_Data_ADS.DSR(idsr).L1B_RBC_Screening.L1B_RBC_Valid = fread(fid, 1,'uint8','ieee-be');
    % Spare_1
    fseek(fid,5,'cof');
    % Spare_1
    fseek(fid,20,'cof');

    % L2B_AMD_Screening
    Product_Confidence_Data_ADS.DSR(idsr).L2B_AMD_Screening.L2B_AMD_Screening_QC = fread(fid, 1,'uint8','ieee-be');
    Product_Confidence_Data_ADS.DSR(idsr).L2B_AMD_Screening.L2B_AMD_Screening_QC_Flags = fread(fid, 1,'uint8','ieee-be');
    for imeas=1:Nmax,
	    Product_Confidence_Data_ADS.DSR(idsr).L2B_AMD_Screening.L2B_AMD_Collocations(imeas).L2B_AMD_Collocation_QC = fread(fid, 1,'uint8','ieee-be');
	    Product_Confidence_Data_ADS.DSR(idsr).L2B_AMD_Screening.L2B_AMD_Collocations(imeas).L2B_AMD_Collocation_QC_Flags = fread(fid, 1,'uint8','ieee-be');
	    % Spare_1
	    fseek(fid,1,'cof');
    end
    % Spare_1
    fseek(fid,20,'cof');

    % L2B_Mie_Classification_QC
    for imeas=1:Nmax,
	    for ibin=1:24,
		    Product_Confidence_Data_ADS.DSR(idsr).L2B_Mie_Classification_QC.L2B_Mie_Meas_Classification(imeas).L2B_Mie_Meas_Bin_Classification(ibin).L2B_Mie_Meas_Class_Flags1 = fread(fid, 1,'uint8','ieee-be');
		    Product_Confidence_Data_ADS.DSR(idsr).L2B_Mie_Classification_QC.L2B_Mie_Meas_Classification(imeas).L2B_Mie_Meas_Bin_Classification(ibin).L2B_Mie_Meas_Class_Flags2 = fread(fid, 1,'uint8','ieee-be');
		    Product_Confidence_Data_ADS.DSR(idsr).L2B_Mie_Classification_QC.L2B_Mie_Meas_Classification(imeas).L2B_Mie_Meas_Bin_Classification(ibin).L2B_Reliability = fread(fid, 1,'float64','ieee-be');
		    % Spare_1
		    fseek(fid,1,'cof');
	    end
	    % Spare_1
	    fseek(fid,1,'cof');
    end
    for iprof=1:M_Mie,
	    Product_Confidence_Data_ADS.DSR(idsr).L2B_Mie_Classification_QC.L2B_Mie_Obs_Classification(iprof).L2B_Mie_Obs_Class_Flags = fread(fid, 1,'uint8','ieee-be');
	    for ibin=1:24,
		    Product_Confidence_Data_ADS.DSR(idsr).L2B_Mie_Classification_QC.L2B_Mie_Obs_Classification(iprof).L2B_Mie_Obs_Bin_Classification(ibin).L2B_Mie_Meas_Class_Flags1 = fread(fid, 1,'uint8','ieee-be');
		    Product_Confidence_Data_ADS.DSR(idsr).L2B_Mie_Classification_QC.L2B_Mie_Obs_Classification(iprof).L2B_Mie_Obs_Bin_Classification(ibin).L2B_Mie_Meas_Class_Flags2 = fread(fid, 1,'uint8','ieee-be');
		    Product_Confidence_Data_ADS.DSR(idsr).L2B_Mie_Classification_QC.L2B_Mie_Obs_Classification(iprof).L2B_Mie_Obs_Bin_Classification(ibin).L2B_Reliability = fread(fid, 1,'float64','ieee-be');
		    % Spare_1
		    fseek(fid,1,'cof');
	    end
	    % Spare_1
	    fseek(fid,1,'cof');
    end
    % Spare_1
    fseek(fid,20,'cof');

    % L2B_Rayleigh_Classification_QC
    for imeas=1:Nmax,
	    for ibin=1:24,
		    Product_Confidence_Data_ADS.DSR(idsr).L2B_Rayleigh_Classification_QC.L2B_Rayleigh_Meas_Classification(imeas).L2B_Rayleigh_Meas_Bin_Classification(ibin).L2B_Rayleigh_Meas_Class_Flags1 = fread(fid, 1,'uint8','ieee-be');
		    Product_Confidence_Data_ADS.DSR(idsr).L2B_Rayleigh_Classification_QC.L2B_Rayleigh_Meas_Classification(imeas).L2B_Rayleigh_Meas_Bin_Classification(ibin).L2B_Rayleigh_Meas_Class_Flags2 = fread(fid, 1,'uint8','ieee-be');
		    Product_Confidence_Data_ADS.DSR(idsr).L2B_Rayleigh_Classification_QC.L2B_Rayleigh_Meas_Classification(imeas).L2B_Rayleigh_Meas_Bin_Classification(ibin).L2B_Reliability = fread(fid, 1,'float64','ieee-be');
		    % Spare_1
		    fseek(fid,1,'cof');
	    end
	    % Spare_1
	    fseek(fid,1,'cof');
    end
    for iprof=1:M_Rayleigh,
	    Product_Confidence_Data_ADS.DSR(idsr).L2B_Rayleigh_Classification_QC.L2B_Rayleigh_Obs_Classification(iprof).L2B_Rayleigh_Obs_Class_Flags = fread(fid, 1,'uint8','ieee-be');
	    for ibin=1:24,
		    Product_Confidence_Data_ADS.DSR(idsr).L2B_Rayleigh_Classification_QC.L2B_Rayleigh_Obs_Classification(iprof).L2B_Rayleigh_Obs_Bin_Classification(ibin).L2B_Rayleigh_Meas_Class_Flags1 = fread(fid, 1,'uint8','ieee-be');
		    Product_Confidence_Data_ADS.DSR(idsr).L2B_Rayleigh_Classification_QC.L2B_Rayleigh_Obs_Classification(iprof).L2B_Rayleigh_Obs_Bin_Classification(ibin).L2B_Rayleigh_Meas_Class_Flags2 = fread(fid, 1,'uint8','ieee-be');
		    Product_Confidence_Data_ADS.DSR(idsr).L2B_Rayleigh_Classification_QC.L2B_Rayleigh_Obs_Classification(iprof).L2B_Rayleigh_Obs_Bin_Classification(ibin).L2B_Reliability = fread(fid, 1,'float64','ieee-be');
		    % Spare_1
		    fseek(fid,1,'cof');
	    end
	    % Spare_1
	    fseek(fid,1,'cof');
    end
    % Spare_1
    fseek(fid,20,'cof');

    % L2B_Mie_Processing_QC
    for iprof=1:M_Mie,
	    for ibin=1:24,
		     Product_Confidence_Data_ADS.DSR(idsr).L2B_Mie_Processing_QC.L2B_Obs_Mie_QC(iprof).L2B_Obs_Mie_Bin_QC(ibin).Hlos_Error_Estimate = fread(fid, 1,'float64','ieee-be');
		     Product_Confidence_Data_ADS.DSR(idsr).L2B_Mie_Processing_QC.L2B_Obs_Mie_QC(iprof).L2B_Obs_Mie_Bin_QC(ibin).Reliability_Number = fread(fid, 1,'float64','ieee-be');
		     Product_Confidence_Data_ADS.DSR(idsr).L2B_Mie_Processing_QC.L2B_Obs_Mie_QC(iprof).L2B_Obs_Mie_Bin_QC(ibin).L2B_Obs_Mie_Bin_Flags = fread(fid, 1,'uint8','ieee-be');
		     Product_Confidence_Data_ADS.DSR(idsr).L2B_Mie_Processing_QC.L2B_Obs_Mie_QC(iprof).L2B_Obs_Mie_Bin_QC(ibin).Fitting_Amplitude = fread(fid, 1,'float64','ieee-be');
		     Product_Confidence_Data_ADS.DSR(idsr).L2B_Mie_Processing_QC.L2B_Obs_Mie_QC(iprof).L2B_Obs_Mie_Bin_QC(ibin).Fitting_Residual = fread(fid, 1,'float64','ieee-be');
		     Product_Confidence_Data_ADS.DSR(idsr).L2B_Mie_Processing_QC.L2B_Obs_Mie_QC(iprof).L2B_Obs_Mie_Bin_QC(ibin).Fitting_Offset = fread(fid, 1,'float64','ieee-be');
		     Product_Confidence_Data_ADS.DSR(idsr).L2B_Mie_Processing_QC.L2B_Obs_Mie_QC(iprof).L2B_Obs_Mie_Bin_QC(ibin).Fitting_FWHM = fread(fid, 1,'float64','ieee-be');
		     Product_Confidence_Data_ADS.DSR(idsr).L2B_Mie_Processing_QC.L2B_Obs_Mie_QC(iprof).L2B_Obs_Mie_Bin_QC(ibin).Mie_SNR = fread(fid, 1,'float64','ieee-be');
		     Product_Confidence_Data_ADS.DSR(idsr).L2B_Mie_Processing_QC.L2B_Obs_Mie_QC(iprof).L2B_Obs_Mie_Bin_QC(ibin).Extinction = fread(fid, 1,'float64','ieee-be');
		     Product_Confidence_Data_ADS.DSR(idsr).L2B_Mie_Processing_QC.L2B_Obs_Mie_QC(iprof).L2B_Obs_Mie_Bin_QC(ibin).Scattering_Ratio = fread(fid, 1,'float64','ieee-be');
		     Product_Confidence_Data_ADS.DSR(idsr).L2B_Mie_Processing_QC.L2B_Obs_Mie_QC(iprof).L2B_Obs_Mie_Bin_QC(ibin).Variability_Measure = fread(fid, 1,'float64','ieee-be');
		     % Spare_1
		     fseek(fid,1,'cof');
	     end
	     % Spare_1
	     fseek(fid,1,'cof');
     end
     Product_Confidence_Data_ADS.DSR(idsr).L2B_Mie_Processing_QC.Mie_Background_High = fread(fid, 1,'uint8','ieee-be');
     % Spare_1
     fseek(fid,20,'cof');

    % L2B_Rayleigh_Processing_QC
    for iprof=1:M_Rayleigh,
	    for ibin=1:24,
		     Product_Confidence_Data_ADS.DSR(idsr).L2B_Rayleigh_Processing_QC.L2B_Obs_Rayleigh_QC(iprof).L2B_Obs_Rayleigh_Bin_QC(ibin).Hlos_Error_Estimate = fread(fid, 1,'float64','ieee-be');
		     Product_Confidence_Data_ADS.DSR(idsr).L2B_Rayleigh_Processing_QC.L2B_Obs_Rayleigh_QC(iprof).L2B_Obs_Rayleigh_Bin_QC(ibin).L2B_Obs_Rayleigh_Bin_Flags = fread(fid, 3,'uint8','ieee-be');
		     Product_Confidence_Data_ADS.DSR(idsr).L2B_Rayleigh_Processing_QC.L2B_Obs_Rayleigh_QC(iprof).L2B_Obs_Rayleigh_Bin_QC(ibin).Extinction = fread(fid, 1,'float64','ieee-be');
		     Product_Confidence_Data_ADS.DSR(idsr).L2B_Rayleigh_Processing_QC.L2B_Obs_Rayleigh_QC(iprof).L2B_Obs_Rayleigh_Bin_QC(ibin).Scattering_Ratio = fread(fid, 1,'float64','ieee-be');
		     Product_Confidence_Data_ADS.DSR(idsr).L2B_Rayleigh_Processing_QC.L2B_Obs_Rayleigh_QC(iprof).L2B_Obs_Rayleigh_Bin_QC(ibin).Variability_Measure = fread(fid, 1,'float64','ieee-be');
		     Product_Confidence_Data_ADS.DSR(idsr).L2B_Rayleigh_Processing_QC.L2B_Obs_Rayleigh_QC(iprof).L2B_Obs_Rayleigh_Bin_QC(ibin).Extinction_Method = fread(fid, 1,'uint8','ieee-be');
		     Product_Confidence_Data_ADS.DSR(idsr).L2B_Rayleigh_Processing_QC.L2B_Obs_Rayleigh_QC(iprof).L2B_Obs_Rayleigh_Bin_QC(ibin).Scattering_Ratio_Method = fread(fid, 1,'uint8','ieee-be');
		     % Spare_1
		     fseek(fid,1,'cof');
	     end
	     % Spare_1
	     fseek(fid,1,'cof');
     end
     Product_Confidence_Data_ADS.DSR(idsr).L2B_Rayleigh_Processing_QC.Rayleigh_Background_High = fread(fid, 1,'uint8','ieee-be');
     % Spare_1
     fseek(fid,20,'cof');

     % Spare_1
     fseek(fid,20,'cof');

end % for idsr

end % function
    
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% %%%%%%%%%%%% Mie_HLOSwind_MDS         %%%%%%%%%%%%%
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% 02/2007 P. Poli   New code. Uses IODD names.
% 04/2007 P. Poli   Upgrade to L2bP 1.3 (add Spare_1 after Obs_Type)

function Mie_HLOSwind_MDS = L2b_Read_Mie_MDS(fid,nbDSR_GWD,M_Mie,Nmax)

Mie_HLOSwind_MDS = [];

for idsr = 1:nbDSR_GWD,
    Mie_HLOSwind_MDS.DSR(idsr).Start_of_Obs_Time (1) = fread(fid, 1,'*int32');
    Mie_HLOSwind_MDS.DSR(idsr).Start_of_Obs_Time (2:3) = fread(fid, 2,'*uint32');
    Mie_HLOSwind_MDS.DSR(idsr).N_Meas = fread(fid, 1 ,'int16', 'ieee-be');
    Mie_HLOSwind_MDS.DSR(idsr).N_Obs_Mie_Actual = fread(fid, 1 ,'int16', 'ieee-be');
    Mie_HLOSwind_MDS.DSR(idsr).P = fread(fid, 1 ,'int16', 'ieee-be');
    for ibin=1:24,
	    for imeas=1:Nmax,
		    Mie_HLOSwind_MDS.DSR(idsr).Map_of_L1_Measurements_Used(ibin,imeas).L1_Measurement_Used = fread(fid, 1,'uint8','ieee-be');
	    end
    end
    for ibin=1:24,
	    for imeas=1:Nmax,
		    Mie_HLOSwind_MDS.DSR(idsr).List_of_L1_Measurement_Weights(ibin,imeas).L1_Measurement_Weights = fread(fid, 1,'uint16','ieee-be');
	    end
    end
    for iprof=1:M_Mie,
	    Mie_HLOSwind_MDS.DSR(idsr).List_of_Mie_Profiles(iprof).Obs_Type = fread(fid, 1,'int8','ieee-be');
	    Mie_HLOSwind_MDS.DSR(idsr).List_of_Mie_Profiles(iprof).Spare_1  = fread(fid,36,'uint8','ieee-be');
	    for ibin=1:24,
		    Mie_HLOSwind_MDS.DSR(idsr).List_of_Mie_Profiles(iprof).List_of_Mie_Height_Bin_Winds(ibin).Validity_Flag = fread(fid, 1,'int8','ieee-be');
		    Mie_HLOSwind_MDS.DSR(idsr).List_of_Mie_Profiles(iprof).List_of_Mie_Height_Bin_Winds(ibin).Mie_Wind_Velocity = fread(fid, 1,'int16','ieee-be');
		    Mie_HLOSwind_MDS.DSR(idsr).List_of_Mie_Profiles(iprof).List_of_Mie_Height_Bin_Winds(ibin).Mie_Error_Quantifier = fread(fid, 1,'int16','ieee-be');
		    Mie_HLOSwind_MDS.DSR(idsr).List_of_Mie_Profiles(iprof).List_of_Mie_Height_Bin_Winds(ibin).Integration_Length = fread(fid, 1,'uint32','ieee-be');
	    end
    end
end % for idsr

end % function

% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% %%%%%%%%%%%% Rayleigh_HLOSwind_MDS    %%%%%%%%%%%%%
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% 02/2007 P. Poli   New code. Uses IODD names.
% 04/2007 P. Poli   Upgrade to L2bP 1.3 (add Spare_1 after Obs_Type)

function Rayleigh_HLOSwind_MDS = L2b_Read_Rayleigh_MDS(fid,nbDSR_GWD,M_Rayleigh,Nmax)

Rayleigh_HLOSwind_MDS = [];

for idsr = 1:nbDSR_GWD,
    Rayleigh_HLOSwind_MDS.DSR(idsr).Start_of_Obs_Time (1) = fread(fid, 1,'*int32');
    Rayleigh_HLOSwind_MDS.DSR(idsr).Start_of_Obs_Time (2:3) = fread(fid, 2,'*uint32');
    Rayleigh_HLOSwind_MDS.DSR(idsr).N_Meas = fread(fid, 1 ,'int16', 'ieee-be');
    Rayleigh_HLOSwind_MDS.DSR(idsr).N_Obs_Rayleigh_Actual = fread(fid, 1 ,'int16', 'ieee-be');
    Rayleigh_HLOSwind_MDS.DSR(idsr).P = fread(fid, 1 ,'int16', 'ieee-be');
    for ibin=1:24,
	    for imeas=1:Nmax,
		    Rayleigh_HLOSwind_MDS.DSR(idsr).Map_of_L1_Measurements_Used(ibin,imeas).L1_Measurement_Used = fread(fid, 1,'uint8','ieee-be');
	    end
    end
    for ibin=1:24,
	    for imeas=1:Nmax,
		    Rayleigh_HLOSwind_MDS.DSR(idsr).List_of_L1_Measurement_Weights(ibin,imeas).L1_Measurement_Weights = fread(fid, 1,'uint16','ieee-be');
	    end
    end
    for iprof=1:M_Rayleigh,
	    Rayleigh_HLOSwind_MDS.DSR(idsr).List_of_Rayleigh_Profiles(iprof).Obs_Type = fread(fid, 1,'int8','ieee-be');
	    Rayleigh_HLOSwind_MDS.DSR(idsr).List_of_Rayleigh_Profiles(iprof).Spare_1  = fread(fid,36,'uint8','ieee-be');
	    for ibin=1:24,
		    Rayleigh_HLOSwind_MDS.DSR(idsr).List_of_Rayleigh_Profiles(iprof). ...
		    List_of_Rayleigh_Height_Bin_Winds(ibin).Validity_Flag = fread(fid, 1,'int8','ieee-be');
		    Rayleigh_HLOSwind_MDS.DSR(idsr).List_of_Rayleigh_Profiles(iprof). ...
		    List_of_Rayleigh_Height_Bin_Winds(ibin).Rayleigh_Wind_Velocity = fread(fid, 1,'int16','ieee-be');
		    Rayleigh_HLOSwind_MDS.DSR(idsr).List_of_Rayleigh_Profiles(iprof). ...
		    List_of_Rayleigh_Height_Bin_Winds(ibin).Rayleigh_Wind_to_Pressure = fread(fid, 1,'int16','ieee-be');
		    Rayleigh_HLOSwind_MDS.DSR(idsr).List_of_Rayleigh_Profiles(iprof). ...
		    List_of_Rayleigh_Height_Bin_Winds(ibin).Rayleigh_Wind_to_Temperature = fread(fid, 1,'int16','ieee-be');
		    Rayleigh_HLOSwind_MDS.DSR(idsr).List_of_Rayleigh_Profiles(iprof). ...
		    List_of_Rayleigh_Height_Bin_Winds(ibin).Rayleigh_Wind_to_Backscatter_Ratio = fread(fid, 1,'int16','ieee-be');
		    Rayleigh_HLOSwind_MDS.DSR(idsr).List_of_Rayleigh_Profiles(iprof). ...
		    List_of_Rayleigh_Height_Bin_Winds(ibin).Reference_Pressure = fread(fid, 1,'uint32','ieee-be');
		    Rayleigh_HLOSwind_MDS.DSR(idsr).List_of_Rayleigh_Profiles(iprof). ...
		    List_of_Rayleigh_Height_Bin_Winds(ibin).Reference_Temperature = fread(fid, 1,'uint16','ieee-be');
		    Rayleigh_HLOSwind_MDS.DSR(idsr).List_of_Rayleigh_Profiles(iprof). ...
		    List_of_Rayleigh_Height_Bin_Winds(ibin).Reference_Backscatter_Ratio = fread(fid, 1,'uint32','ieee-be');
		    Rayleigh_HLOSwind_MDS.DSR(idsr).List_of_Rayleigh_Profiles(iprof). ...
		    List_of_Rayleigh_Height_Bin_Winds(ibin).Rayleigh_Error_Quantifier = fread(fid, 1,'uint16','ieee-be');
		    Rayleigh_HLOSwind_MDS.DSR(idsr).List_of_Rayleigh_Profiles(iprof). ...
		    List_of_Rayleigh_Height_Bin_Winds(ibin).Integration_Length = fread(fid, 1,'uint32','ieee-be');
	    end
    end
end % for idsr

end % function

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%% Assimilation Product Confidence %%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% 04/2007 P. Poli   Original code.

function Assim_PCD_ADS = L2c_Read_Assim_PCD_ADS(fid,nbDSR_Assim_PCD,M_Mie,M_Rayleigh)

Assim_PCD_ADS = [];

for idsr = 1:nbDSR_Assim_PCD,
    
    Assim_PCD_ADS.DSR(idsr).N_Meas = fread(fid, 1,'*int16','ieee-be');
    Assim_PCD_ADS.DSR(idsr).N_Obs_Mie_Actual = fread(fid, 1,'*int16','ieee-be');
    Assim_PCD_ADS.DSR(idsr).N_Obs_Rayleigh_Actual = fread(fid, 1,'*int16','ieee-be');
    
    % L2B_Input_Screening
    Assim_PCD_ADS.DSR(idsr).L2B_Input_Screening.L2B_Obs_Screening = fread(fid, 1,'uint8','ieee-be');
    Assim_PCD_ADS.DSR(idsr).L2B_Input_Screening.L2B_Obs_Screening_Flags1 = fread(fid, 1,'uint8','ieee-be');
    Assim_PCD_ADS.DSR(idsr).L2B_Input_Screening.L2B_Obs_Screening_Flags2 = fread(fid, 1,'uint8','ieee-be');
    Assim_PCD_ADS.DSR(idsr).L2B_Input_Screening.L2B_Obs_Screening_Flags3 = fread(fid, 1,'uint8','ieee-be');
    Assim_PCD_ADS.DSR(idsr).L2B_Input_Screening.L2B_Obs_Screening_Flags4 = fread(fid, 1,'uint8','ieee-be');
    Assim_PCD_ADS.DSR(idsr).L2B_Input_Screening.L2B_Obs_Screening_Flags5 = fread(fid, 1,'uint8','ieee-be');
    % Spare_1
    fseek(fid,20,'cof');

    % List_of_L2C_Mie_Quality_Params
    for iprof=1:M_Mie,
	    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Mie_Quality_Params(iprof).Obs_Type = fread(fid, 1,'uint8','ieee-be');
	    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Mie_Quality_Params(iprof).Spare_1  = fread(fid,36,'uint8','ieee-be');
	    for ibin=1:24,
		    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Mie_Quality_Params(iprof).List_of_L2C_Mie_Height_Bin_Quality_Params(ibin). ...
		    L2B_Mie_Obs_Screening.L2B_Mie_Obs_QC = fread(fid, 1,'uint8','ieee-be');
		    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Mie_Quality_Params(iprof).List_of_L2C_Mie_Height_Bin_Quality_Params(ibin). ...
		    L2B_Mie_Obs_Screening.L2B_Mie_Obs_QC_Flags = fread(fid, 1,'uint8','ieee-be');
		    % Spare_1
		    fseek(fid,16,'cof');
		    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Mie_Quality_Params(iprof).List_of_L2C_Mie_Height_Bin_Quality_Params(ibin). ...
		    Assimilation_Model_PCD.HLOS_observation_errors.persistence_error = fread(fid, 1,'uint16','ieee-be');
		    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Mie_Quality_Params(iprof).List_of_L2C_Mie_Height_Bin_Quality_Params(ibin). ...
		    Assimilation_Model_PCD.HLOS_observation_errors.representativity_error = fread(fid, 1,'uint16','ieee-be');
		    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Mie_Quality_Params(iprof).List_of_L2C_Mie_Height_Bin_Quality_Params(ibin). ...
		    Assimilation_Model_PCD.HLOS_observation_errors.final_error = fread(fid, 1,'uint16','ieee-be');
		    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Mie_Quality_Params(iprof).List_of_L2C_Mie_Height_Bin_Quality_Params(ibin). ...
		    Assimilation_Model_PCD.HLOS_observation_errors.estimated_obs_bias = fread(fid, 1,'*int16','ieee-be');
		    % Spare_1
		    fseek(fid,20,'cof');
		    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Mie_Quality_Params(iprof).List_of_L2C_Mie_Height_Bin_Quality_Params(ibin). ...
		    Assimilation_Model_PCD.Background_hlos = fread(fid, 1,'*int16','ieee-be');
		    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Mie_Quality_Params(iprof).List_of_L2C_Mie_Height_Bin_Quality_Params(ibin). ...
		    Assimilation_Model_PCD.Background_hlos_error = fread(fid, 1,'uint16','ieee-be');
		    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Mie_Quality_Params(iprof).List_of_L2C_Mie_Height_Bin_Quality_Params(ibin). ...
		    Assimilation_Model_PCD.L2B_hlos_reliability = fread(fid, 1,'float64','ieee-be');
		    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Mie_Quality_Params(iprof).List_of_L2C_Mie_Height_Bin_Quality_Params(ibin). ...
		    Assimilation_Model_PCD.Analysis_hlos = fread(fid, 1,'*int16','ieee-be');
		    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Mie_Quality_Params(iprof).List_of_L2C_Mie_Height_Bin_Quality_Params(ibin). ...
		    Assimilation_Model_PCD.Zonal_wind_background_error = fread(fid, 1,'uint16','ieee-be');
		    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Mie_Quality_Params(iprof).List_of_L2C_Mie_Height_Bin_Quality_Params(ibin). ...
		    Assimilation_Model_PCD.Meridional_wind_background_error = fread(fid, 1,'uint16','ieee-be');
		    % Spare_1
		    fseek(fid,20,'cof');
		    % Spare_1
		    fseek(fid,10,'cof');
	     end
    end

    % List_of_L2C_Rayleigh_Quality_Params
    for iprof=1:M_Rayleigh,
	    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Rayleigh_Quality_Params(iprof).Obs_Type = fread(fid, 1,'uint8','ieee-be');
	    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Rayleigh_Quality_Params(iprof).Spare_1  = fread(fid,36,'uint8','ieee-be');
	    for ibin=1:24,
		    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Rayleigh_Quality_Params(iprof).List_of_L2C_Rayleigh_Height_Bin_Quality_Params(ibin). ...
		    L2B_Rayleigh_Obs_Screening.L2B_Rayleigh_Obs_QC = fread(fid, 1,'uint8','ieee-be');
		    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Rayleigh_Quality_Params(iprof).List_of_L2C_Rayleigh_Height_Bin_Quality_Params(ibin). ...
		    L2B_Rayleigh_Obs_Screening.L2B_Rayleigh_Obs_QC_Flags = fread(fid, 1,'uint8','ieee-be');
		    % Spare_1
		    fseek(fid,16,'cof');
		    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Rayleigh_Quality_Params(iprof).List_of_L2C_Rayleigh_Height_Bin_Quality_Params(ibin). ...
		    Assimilation_Model_PCD.HLOS_observation_errors.persistence_error = fread(fid, 1,'uint16','ieee-be');
		    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Rayleigh_Quality_Params(iprof).List_of_L2C_Rayleigh_Height_Bin_Quality_Params(ibin). ...
		    Assimilation_Model_PCD.HLOS_observation_errors.representativity_error = fread(fid, 1,'uint16','ieee-be');
		    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Rayleigh_Quality_Params(iprof).List_of_L2C_Rayleigh_Height_Bin_Quality_Params(ibin). ...
		    Assimilation_Model_PCD.HLOS_observation_errors.final_error = fread(fid, 1,'uint16','ieee-be');
		    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Rayleigh_Quality_Params(iprof).List_of_L2C_Rayleigh_Height_Bin_Quality_Params(ibin). ...
		    Assimilation_Model_PCD.HLOS_observation_errors.estimated_obs_bias = fread(fid, 1,'*int16','ieee-be');
		    % Spare_1
		    fseek(fid,20,'cof');
		    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Rayleigh_Quality_Params(iprof).List_of_L2C_Rayleigh_Height_Bin_Quality_Params(ibin). ...
		    Assimilation_Model_PCD.Background_hlos = fread(fid, 1,'*int16','ieee-be');
		    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Rayleigh_Quality_Params(iprof).List_of_L2C_Rayleigh_Height_Bin_Quality_Params(ibin). ...
		    Assimilation_Model_PCD.Background_hlos_error = fread(fid, 1,'uint16','ieee-be');
		    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Rayleigh_Quality_Params(iprof).List_of_L2C_Rayleigh_Height_Bin_Quality_Params(ibin). ...
		    Assimilation_Model_PCD.L2B_hlos_reliability = fread(fid, 1,'float64','ieee-be');
		    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Rayleigh_Quality_Params(iprof).List_of_L2C_Rayleigh_Height_Bin_Quality_Params(ibin). ...
		    Assimilation_Model_PCD.Analysis_hlos = fread(fid, 1,'*int16','ieee-be');
		    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Rayleigh_Quality_Params(iprof).List_of_L2C_Rayleigh_Height_Bin_Quality_Params(ibin). ...
		    Assimilation_Model_PCD.Zonal_wind_background_error = fread(fid, 1,'uint16','ieee-be');
		    Assim_PCD_ADS.DSR(idsr).List_of_L2C_Rayleigh_Quality_Params(iprof).List_of_L2C_Rayleigh_Height_Bin_Quality_Params(ibin). ...
		    Assimilation_Model_PCD.Meridional_wind_background_error = fread(fid, 1,'uint16','ieee-be');
		    % Spare_1
		    fseek(fid,20,'cof');
		    % Spare_1
		    fseek(fid,10,'cof');
	     end
    end

    % Spare_1
    fseek(fid,20,'cof');
end % for idsr

end % function

% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% %%%%%%%%%%% L2C Mie_VecWind_MDS         %%%%%%%%%%%
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% 04/2007 P. Poli   Original code

function Mie_Vecwind_MDS = L2c_Read_Mie_Vec_MDS(fid,nbDSR,M_Mie)

Mie_Vecwind_MDS = [];

for idsr = 1:nbDSR,
    Mie_Vecwind_MDS.DSR(idsr).Start_of_Obs_Time (1) = fread(fid, 1,'*int32');
    Mie_Vecwind_MDS.DSR(idsr).Start_of_Obs_Time (2:3) = fread(fid, 2,'*uint32');
    Mie_Vecwind_MDS.DSR(idsr).N_Meas = fread(fid, 1 ,'int16', 'ieee-be');
    Mie_Vecwind_MDS.DSR(idsr).N_Obs_Mie_Actual = fread(fid, 1 ,'int16', 'ieee-be');
    for iprof=1:M_Mie,
	    Mie_Vecwind_MDS.DSR(idsr).List_of_Mie_Profiles(iprof).Obs_Type = fread(fid, 1 ,'uint8', 'ieee-be');
	    Mie_Vecwind_MDS.DSR(idsr).List_of_Mie_Profiles(iprof).Spare_1  = fread(fid,36 ,'uint8', 'ieee-be');
	    for ibin=1:24,
		    Mie_Vecwind_MDS.DSR(idsr).List_of_Mie_Profiles(iprof).List_of_Mie_Height_Bin_VecWinds(ibin). ...
		    Validity_Flag = fread(fid, 1,'int8','ieee-be');
		    Mie_Vecwind_MDS.DSR(idsr).List_of_Mie_Profiles(iprof).List_of_Mie_Height_Bin_VecWinds(ibin). ...
		    Background_Zonal_Zind_Velocity = fread(fid, 1,'int16','ieee-be');
		    Mie_Vecwind_MDS.DSR(idsr).List_of_Mie_Profiles(iprof).List_of_Mie_Height_Bin_VecWinds(ibin). ...
		    Background_Meridional_Wind_Velocity = fread(fid, 1,'int16','ieee-be');
		    Mie_Vecwind_MDS.DSR(idsr).List_of_Mie_Profiles(iprof).List_of_Mie_Height_Bin_VecWinds(ibin). ...
		    Analysis_Zonal_Zind_Velocity = fread(fid, 1,'int16','ieee-be');
		    Mie_Vecwind_MDS.DSR(idsr).List_of_Mie_Profiles(iprof).List_of_Mie_Height_Bin_VecWinds(ibin). ...
		    Analysis_Meridional_Wind_Velocity = fread(fid, 1,'int16','ieee-be');
		    % Spare_1
		    fseek(fid,20,'cof');
	    end
    end

end % for idsr

end % function

% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% %%%%%%%%%%% L2C Rayleigh_VecWind_MDS    %%%%%%%%%%%
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% 04/2007 P. Poli   Original code

function Rayleigh_Vecwind_MDS = L2c_Read_Rayleigh_Vec_MDS(fid,nbDSR,M_Rayleigh)

Rayleigh_Vecwind_MDS = [];

for idsr = 1:nbDSR,
    Rayleigh_Vecwind_MDS.DSR(idsr).Start_of_Obs_Time (1) = fread(fid, 1,'*int32');
    Rayleigh_Vecwind_MDS.DSR(idsr).Start_of_Obs_Time (2:3) = fread(fid, 2,'*uint32');
    Rayleigh_Vecwind_MDS.DSR(idsr).N_Meas = fread(fid, 1 ,'int16', 'ieee-be');
    Rayleigh_Vecwind_MDS.DSR(idsr).N_Obs_Rayleigh_Actual = fread(fid, 1 ,'int16', 'ieee-be');
    for iprof=1:M_Rayleigh,
	    Rayleigh_Vecwind_MDS.DSR(idsr).List_of_Rayleigh_Profiles(iprof).Obs_Type = fread(fid, 1 ,'uint8', 'ieee-be');
	    Rayleigh_Vecwind_MDS.DSR(idsr).List_of_Rayleigh_Profiles(iprof).Spare_1  = fread(fid,36 ,'uint8', 'ieee-be');
	    for ibin=1:24,
		    Rayleigh_Vecwind_MDS.DSR(idsr).List_of_Rayleigh_Profiles(iprof).List_of_Rayleigh_Height_Bin_VecWinds(ibin). ...
		    Validity_Flag = fread(fid, 1,'int8','ieee-be');
		    Rayleigh_Vecwind_MDS.DSR(idsr).List_of_Rayleigh_Profiles(iprof).List_of_Rayleigh_Height_Bin_VecWinds(ibin). ...
		    Background_Zonal_Zind_Velocity = fread(fid, 1,'int16','ieee-be');
		    Rayleigh_Vecwind_MDS.DSR(idsr).List_of_Rayleigh_Profiles(iprof).List_of_Rayleigh_Height_Bin_VecWinds(ibin). ...
		    Background_Meridional_Wind_Velocity = fread(fid, 1,'int16','ieee-be');
		    Rayleigh_Vecwind_MDS.DSR(idsr).List_of_Rayleigh_Profiles(iprof).List_of_Rayleigh_Height_Bin_VecWinds(ibin). ...
		    Analysis_Zonal_Zind_Velocity = fread(fid, 1,'int16','ieee-be');
		    Rayleigh_Vecwind_MDS.DSR(idsr).List_of_Rayleigh_Profiles(iprof).List_of_Rayleigh_Height_Bin_VecWinds(ibin). ...
		    Analysis_Meridional_Wind_Velocity = fread(fid, 1,'int16','ieee-be');
		    % Spare_1
		    fseek(fid,20,'cof');
	    end
    end
end % for idsr

end % function

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%  MATLAB CODE TO READ L1b DATA   %%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% HISTORY
% 
% 20 Dec 2005   M-L Denneulin       Original code
% 23 Jan 2006   M-L Denneulin       Modification of format for several ADSR to be compliant with the DLR
%                                   version 1.2 of the L1B processor      
% Nov 2006      M-L Denneulin       Modify the structure of the program to
%                                   be able to read only a part of the file
%                                   Modification of format to be compliant
%                                   with the IODD of the L1 processor v1.04
%                                   (ADM-IC-52-1666-3.1 du 30 Oct 2006)
% 15 Nov 2006   M-L Denneulin       Correction of a bug in Ground Wind Detection ADS    
% 30 Nov 2006   D Tan               Post-FAT PCD_ADS and Meas_ADS.
% 18 Dec 2006   D Tan               Reduce NumFrequencySteps from 60 to 40
% 08 Feb 2007   P Poli              Extracted the functions below from program and 
%                                   moved common instructions to main program

function Geolocation_ADS = L1b_Read_Geo_ADS(fid, nbDSR_Geo, Nmax)

Geolocation_ADS = [];

for idsr = 1:nbDSR_Geo,
    Geolocation_ADS.DSR(idsr).Start_of_Observation_Time(1) = fread(fid, 1,'*int32');
    Geolocation_ADS.DSR(idsr).Start_of_Observation_Time(2:3) = fread(fid, 2,'*uint32');
    % OBSERVATION AOCS
    Geolocation_ADS.DSR(idsr).Observation_AOCS.Observation_Centroid_Time = fread(fid, 3,'*int32');
    Geolocation_ADS.DSR(idsr).Observation_AOCS.X_Position = fread(fid, 1,'*float64');
    Geolocation_ADS.DSR(idsr).Observation_AOCS.Y_Position = fread(fid, 1,'*float64');
    Geolocation_ADS.DSR(idsr).Observation_AOCS.Z_Position = fread(fid, 1,'*float64');
    Geolocation_ADS.DSR(idsr).Observation_AOCS.X_Velocity = fread(fid, 1, '*float64');
    Geolocation_ADS.DSR(idsr).Observation_AOCS.Y_velocity = fread(fid, 1, '*float64');
    Geolocation_ADS.DSR(idsr).Observation_AOCS.Z_velocity = fread(fid, 1, '*float64');
    Geolocation_ADS.DSR(idsr).Observation_AOCS.Roll_Angle = fread(fid, 1, '*float64');
    Geolocation_ADS.DSR(idsr).Observation_AOCS.Pitch_Angle = fread(fid, 1, '*float64');
    Geolocation_ADS.DSR(idsr).Observation_AOCS.Yaw_Angle = fread(fid, 1, '*float64');
    %spare
    fseek(fid,8,'cof');
    % LIST OF MEASUREMENT AOCS
    for imeas = 1:Nmax,
        Geolocation_ADS.DSR(idsr).List_of_Measurement_AOCSs.Measurement_AOCS(imeas).Measurement_Centroid_Time(1) = fread(fid, 1,'*int32');
        Geolocation_ADS.DSR(idsr).List_of_Measurement_AOCSs.Measurement_AOCS(imeas).Measurement_Centroid_Time(2:3) = fread(fid, 2,'*uint32');
        Geolocation_ADS.DSR(idsr).List_of_Measurement_AOCSs.Measurement_AOCS(imeas).X_Position = fread(fid, 1,'*float64');
        Geolocation_ADS.DSR(idsr).List_of_Measurement_AOCSs.Measurement_AOCS(imeas).Y_Position = fread(fid, 1,'*float64');
        Geolocation_ADS.DSR(idsr).List_of_Measurement_AOCSs.Measurement_AOCS(imeas).Z_Position = fread(fid, 1,'*float64');
        Geolocation_ADS.DSR(idsr).List_of_Measurement_AOCSs.Measurement_AOCS(imeas).X_Velocity = fread(fid, 1, '*float64');
        Geolocation_ADS.DSR(idsr).List_of_Measurement_AOCSs.Measurement_AOCS(imeas).Y_Velocity = fread(fid, 1, '*float64');
        Geolocation_ADS.DSR(idsr).List_of_Measurement_AOCSs.Measurement_AOCS(imeas).Z_Velocity = fread(fid, 1, '*float64');
        Geolocation_ADS.DSR(idsr).List_of_Measurement_AOCSs.Measurement_AOCS(imeas).Roll_Angle = fread(fid, 1, '*float64');
        Geolocation_ADS.DSR(idsr).List_of_Measurement_AOCSs.Measurement_AOCS(imeas).Pitch_Angle = fread(fid, 1, '*float64');
        Geolocation_ADS.DSR(idsr).List_of_Measurement_AOCSs.Measurement_AOCS(imeas).Yaw_Angle = fread(fid, 1, '*float64');
        fseek(fid, 8, 'cof');
    end

    % OBSERVATION GEOLOCATION
    %	Observation Mie Geolocation
    for ibin = 1:25,
        Geolocation_ADS.DSR(idsr).Observation_Geolocation.Observation_Mie_Geolocation.List_of_Observation_Geolocation_of_Height_Bins.Observation_Geolocation_of_Height_Bin(ibin).Longitude_of_Height_Bin = fread(fid, 1,'*int32');
        Geolocation_ADS.DSR(idsr).Observation_Geolocation.Observation_Mie_Geolocation.List_of_Observation_Geolocation_of_Height_Bins.Observation_Geolocation_of_Height_Bin(ibin).Latitude_of_Height_Bin = fread(fid, 1,'*int32');
        Geolocation_ADS.DSR(idsr).Observation_Geolocation.Observation_Mie_Geolocation.List_of_Observation_Geolocation_of_Height_Bins.Observation_Geolocation_of_Height_Bin(ibin).Altitude_of_Height_Bin = fread(fid, 1,'*float64');
        Geolocation_ADS.DSR(idsr).Observation_Geolocation.Observation_Mie_Geolocation.List_of_Observation_Geolocation_of_Height_Bins.Observation_Geolocation_of_Height_Bin(ibin).Topocentric_Azimuth_of_Height_Bin = fread(fid, 1,'*float64');
        Geolocation_ADS.DSR(idsr).Observation_Geolocation.Observation_Mie_Geolocation.List_of_Observation_Geolocation_of_Height_Bins.Observation_Geolocation_of_Height_Bin(ibin).Topocentric_Elevation_of_Height_Bin = fread(fid, 1,'*float64');
        Geolocation_ADS.DSR(idsr).Observation_Geolocation.Observation_Mie_Geolocation.List_of_Observation_Geolocation_of_Height_Bins.Observation_Geolocation_of_Height_Bin(ibin).Target_to_Sun_Visibility_Flag = fread(fid, 1,'*int16');
        Geolocation_ADS.DSR(idsr).Observation_Geolocation.Observation_Mie_Geolocation.List_of_Observation_Geolocation_of_Height_Bins.Observation_Geolocation_of_Height_Bin(ibin).Satellite_Range_of_Height_Bin = fread(fid, 1, '*float64');
    end
    % Observation Rayleigh Geolocation
    for ibin = 1:25,
        Geolocation_ADS.DSR(idsr).Observation_Geolocation.Observation_Rayleigh_Geolocation.List_of_Observation_Geolocation_of_Height_Bins.Observation_Geolocation_of_Height_Bin(ibin).Longitude_of_Height_Bin = fread(fid, 1,'*int32');
        Geolocation_ADS.DSR(idsr).Observation_Geolocation.Observation_Rayleigh_Geolocation.List_of_Observation_Geolocation_of_Height_Bins.Observation_Geolocation_of_Height_Bin(ibin).Latitude_of_Height_Bin = fread(fid, 1,'*int32');
        Geolocation_ADS.DSR(idsr).Observation_Geolocation.Observation_Rayleigh_Geolocation.List_of_Observation_Geolocation_of_Height_Bins.Observation_Geolocation_of_Height_Bin(ibin).Altitude_of_Height_Bin = fread(fid, 1,'*float64');
        Geolocation_ADS.DSR(idsr).Observation_Geolocation.Observation_Rayleigh_Geolocation.List_of_Observation_Geolocation_of_Height_Bins.Observation_Geolocation_of_Height_Bin(ibin).Topocentric_Azimuth_of_Height_Bin = fread(fid, 1,'*float64');
        Geolocation_ADS.DSR(idsr).Observation_Geolocation.Observation_Rayleigh_Geolocation.List_of_Observation_Geolocation_of_Height_Bins.Observation_Geolocation_of_Height_Bin(ibin).Topocentric_Elevation_of_Height_Bin = fread(fid, 1,'*float64');
        Geolocation_ADS.DSR(idsr).Observation_Geolocation.Observation_Rayleigh_Geolocation.List_of_Observation_Geolocation_of_Height_Bins.Observation_Geolocation_of_Height_Bin(ibin).Target_to_Sun_Visibility_Flag = fread(fid, 1,'*int16');
        Geolocation_ADS.DSR(idsr).Observation_Geolocation.Observation_Rayleigh_Geolocation.List_of_Observation_Geolocation_of_Height_Bins.Observation_Geolocation_of_Height_Bin(ibin).Satellite_Range_of_Height_Bin = fread(fid, 1, '*float64');
    end
    % Geolocation of DEM intersection
    Geolocation_ADS.DSR(idsr).Observation_Geolocation.Geolocation_of_DEM_Intersection.Latitude_of_DEM_Intersection = fread(fid, 1,'*int32');
    Geolocation_ADS.DSR(idsr).Observation_Geolocation.Geolocation_of_DEM_Intersection.Longitude_of_DEM_Intersection = fread(fid, 1,'*int32');
    Geolocation_ADS.DSR(idsr).Observation_Geolocation.Geolocation_of_DEM_Intersection.Altitude_of_DEM_Intersection = fread(fid, 1,'*float64');

    % Los Velocity
    Geolocation_ADS.DSR(idsr).Observation_Geolocation.Line_of_Sight_Velocity = fread(fid, 1,'*float64');
    %spare3 = fread(fid, 64, '*ubit1');

    % Geoid separation
    Geolocation_ADS.DSR(idsr).Observation_Geolocation.Geoid_Separation = fread(fid, 1, '*float64');
    fseek(fid,8,'cof');


    % LIST OF MEASUREMENT GEOLOCATION
    for imeas = 1:Nmax,
        % Measurement Mie Geolocation
        for ibin = 1:25,
            Geolocation_ADS.DSR(idsr).List_of_Measurement_Geolocations.Measurement_Geolocation(imeas).Mie_Geolocation.List_of_Geolocation_Height_Bins.Geolocation_Height_Bin(ibin).Longitude_of_Height_Bin = fread(fid, 1,'*int32');
            Geolocation_ADS.DSR(idsr).List_of_Measurement_Geolocations.Measurement_Geolocation(imeas).Mie_Geolocation.List_of_Geolocation_Height_Bins.Geolocation_Height_Bin(ibin).Latitude_of_Height_Bin = fread(fid, 1,'*int32');
            Geolocation_ADS.DSR(idsr).List_of_Measurement_Geolocations.Measurement_Geolocation(imeas).Mie_Geolocation.List_of_Geolocation_Height_Bins.Geolocation_Height_Bin(ibin).Altitude_of_Height_Bin = fread(fid, 1,'*float64');
        end
        % Measurement Rayleigh Geolocation
        for ibin = 1:25,
            Geolocation_ADS.DSR(idsr).List_of_Measurement_Geolocations.Measurement_Geolocation(imeas).Rayleigh_Geolocation.List_of_Geolocation_Height_Bins.Geolocation_Height_Bin(ibin).Longitude_of_Height_Bin = fread(fid, 1,'*int32');
            Geolocation_ADS.DSR(idsr).List_of_Measurement_Geolocations.Measurement_Geolocation(imeas).Rayleigh_Geolocation.List_of_Geolocation_Height_Bins.Geolocation_Height_Bin(ibin).Latitude_of_Height_Bin = fread(fid, 1,'*int32');
            Geolocation_ADS.DSR(idsr).List_of_Measurement_Geolocations.Measurement_Geolocation(imeas).Rayleigh_Geolocation.List_of_Geolocation_Height_Bins.Geolocation_Height_Bin(ibin).Altitude_of_Height_Bin = fread(fid, 1,'*float64');
        end
        % Geolocation of DEM intersection
        Geolocation_ADS.DSR(idsr).List_of_Measurement_Geolocations.Measurement_Geolocation(imeas).Geolocation_of_DEM_Intersection.Latitude_DEM_Intersection = fread(fid, 1,'*int32');
        Geolocation_ADS.DSR(idsr).List_of_Measurement_Geolocations.Measurement_Geolocation(imeas).Geolocation_of_DEM_Intersection.Longitude_DEM_Intersection = fread(fid, 1,'*int32');
        Geolocation_ADS.DSR(idsr).List_of_Measurement_Geolocations.Measurement_Geolocation(imeas).Geolocation_of_DEM_Intersection.Altitude_DEM_Intersection = fread(fid, 1,'*float64');
        %Geolocation_ADS.DSR(idsr).MeasGeolocation(imeas).DemGeolocation.TopoAzDemIntersec = fread(fid, 1,'*float64');
        %Geolocation_ADS.DSR(idsr).MeasGeolocation(imeas).DemGeolocation.TopoElevDemIntersec = fread(fid, 1,'*float64');

        % AOCS Los Velocity
        Geolocation_ADS.DSR(idsr).List_of_Measurement_Geolocations.Measurement_Geolocation(imeas).Line_of_Sight_Velocity = fread(fid, 1,'*float64');
        fseek(fid,8,'cof');

    end

end % for idsr

end % function

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%% L1b Product Confidence %%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function Product_Confidence_Data_ADS = L1b_Read_PCD_ADS(fid,nbDSR_PCD,Nmax,version)

Product_Confidence_Data_ADS = [];

for idsr = 1:nbDSR_PCD,
    
    Product_Confidence_Data_ADS.DSR(idsr).Start_of_Observation_Time(1) = fread(fid, 1,'*int32');
    Product_Confidence_Data_ADS.DSR(idsr).Start_of_Observation_Time(2:3) = fread(fid, 2,'*uint32');
    Product_Confidence_Data_ADS.DSR(idsr).N = fread(fid, 1, '*uint16');
    Product_Confidence_Data_ADS.DSR(idsr).P = fread(fid, 1, '*uint16');
    %spare
    fseek(fid, 8,'cof');

    % OBSERVATION PCD
    Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.Num_Measurement_Laser_Freq_Unlocked = fread(fid, 1,'*int32');
    Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.Num_Reference_Pulse_Laser_Freq_Unlocked = fread(fid, 1,'*int32');
    Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.Num_Sat_Not_on_Target_Measurements = fread(fid, 1,'*int32');
    Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.Num_Corrupt_Mie_Measurements = fread(fid, 1,'*int32');
    Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.Num_Corrupt_Rayleigh_Measurements = fread(fid, 1,'*int32');
    Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.Num_Corrupt_Mie_Reference_Pulses = fread(fid, 1,'*int32');
    Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.Num_Corrupt_Rayleigh_Reference_Pulses = fread(fid, 1,'*int32');
    Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.Avg_Laser_Frequency_Offset = fread(fid, 1,'*float64');
    Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.Avg_UV_Energy = fread(fid, 1,'*float64');
    Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.Laser_Frequency_Offset_Std_Dev = fread(fid, 1,'*float64');
    Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.UV_Energy_Std_Dev = fread(fid, 1,'*float64');
    switch version
    case {'ADM_L1bP/1.04 ','ADM_L1bP/1.05 '}
	    Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.Num_Invalid_Mie_Peak = fread(fid, 1,'*int32');
    case {'ADM_L1bP/1.06 ','ADM_L1bP/1.07 ','ADM_L1bP/1.08 ','ADM_L1bP/1.09 '}
	    Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.Num_Mie_Peak_Invalid = fread(fid, 1,'*int32');
    end

    %    List of Observation Alt Bin PCD
    for ibin = 1:25,
        Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.List_of_Observation_Alt_Bin_PCDs.Observation_Alt_Bin_PCD(ibin).ErrorQuantifier.ErrorQuantifierMie = fread(fid, 1,'*float64');
        Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.List_of_Observation_Alt_Bin_PCDs.Observation_Alt_Bin_PCD(ibin).ErrorQuantifier.ErrorQuantifierRay = fread(fid, 1,'*float64');
        Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.List_of_Observation_Alt_Bin_PCDs.Observation_Alt_Bin_PCD(ibin).ErrorQuantifier.ErrorQuantifierRayA = fread(fid, 1,'*float64');
        Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.List_of_Observation_Alt_Bin_PCDs.Observation_Alt_Bin_PCD(ibin).ErrorQuantifier.ErrorQuantifierRayB = fread(fid, 1,'*float64');

        Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.List_of_Observation_Alt_Bin_PCDs.Observation_Alt_Bin_PCD(ibin).Mie_Wind_Velocity_Std_Dev = fread(fid, 1,'*float64');
        Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.List_of_Observation_Alt_Bin_PCDs.Observation_Alt_Bin_PCD(ibin).Rayleigh_Wind_Velocity_Std_Dev = fread(fid, 1,'*float64');
        Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.List_of_Observation_Alt_Bin_PCDs.Observation_Alt_Bin_PCD(ibin).Mie_Useful_Signal_Std_Dev = fread(fid, 1,'*float64');
        Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.List_of_Observation_Alt_Bin_PCDs.Observation_Alt_Bin_PCD(ibin).Rayleigh_Useful_Signal_Channel_A_Std_Dev = fread(fid, 1,'*float64');
        Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.List_of_Observation_Alt_Bin_PCDs.Observation_Alt_Bin_PCD(ibin).Rayleigh_Useful_Signal_Channel_B_Std_Dev = fread(fid, 1,'*float64');
        % Mie core Characteristic
        Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.List_of_Observation_Alt_Bin_PCDs.Observation_Alt_Bin_PCD(ibin).Mie_Core_Characteristic.PeakPosition = fread(fid, 1, '*float64');
        Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.List_of_Observation_Alt_Bin_PCDs.Observation_Alt_Bin_PCD(ibin).Mie_Core_Characteristic.FWHM = fread(fid, 1, '*float64');
        Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.List_of_Observation_Alt_Bin_PCDs.Observation_Alt_Bin_PCD(ibin).Mie_Core_Characteristic.Amplitude = fread(fid, 1, '*float64');
        Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.List_of_Observation_Alt_Bin_PCDs.Observation_Alt_Bin_PCD(ibin).Mie_Core_Characteristic.Offset = fread(fid, 1, '*float64');
        Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.List_of_Observation_Alt_Bin_PCDs.Observation_Alt_Bin_PCD(ibin).Mie_Core_Characteristic.ErrorFlag = fread(fid, 1, '*uint8');
        Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.List_of_Observation_Alt_Bin_PCDs.Observation_Alt_Bin_PCD(ibin).Mie_Core_Characteristic.ResidualError = fread(fid, 1, '*float64');
        Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.List_of_Observation_Alt_Bin_PCDs.Observation_Alt_Bin_PCD(ibin).Mie_Core_Characteristic.NumIterations = fread(fid, 1, '*uint8');


        Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.List_of_Observation_Alt_Bin_PCDs.Observation_Alt_Bin_PCD(ibin).Scattering_Ratio_Mie = fread(fid, 1, '*float64');
        Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.List_of_Observation_Alt_Bin_PCDs.Observation_Alt_Bin_PCD(ibin).Mie_Signal_to_Noise_Ratio = fread(fid, 1, '*float64');
        Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.List_of_Observation_Alt_Bin_PCDs.Observation_Alt_Bin_PCD(ibin).Rayleigh_Signal_to_Noise_Ratio_Channel_A = fread(fid, 1, '*float64');
        Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.List_of_Observation_Alt_Bin_PCDs.Observation_Alt_Bin_PCD(ibin).Rayleigh_Signal_to_Noise_Ratio_Channel_B = fread(fid, 1, '*float64');
    end % for ibin

    Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.Num_Mie_Invalid_Measurements = fread(fid, 1,'*int32');
    Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.Num_Mie_Invalid_Reference_Pulse = fread(fid, 1,'*int32');
    Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.Num_Rayleigh_Invalid_Measurements = fread(fid, 1,'*int32');
    Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.Num_Rayleigh_Invalid_Reference_Pulse = fread(fid, 1,'*int32');
    Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.Mie_Mean_Emitted_Frequency = fread(fid, 1, '*float64');
    Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.Mie_Emitted_Frequency_Std_Dev = fread(fid, 1, '*float64');
    Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.Ray_Mean_Emitted_Frequency = fread(fid, 1, '*float64');
    Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.Ray_Emitted_Frequency_Std_Dev = fread(fid, 1, '*float64');
    Product_Confidence_Data_ADS.DSR(idsr).Observation_PCD.MultiMode_Ratio_Quality_Flag = fread(fid,1 ,'*uint8');          %L1b 1.04
    %spare
    fseek(fid,4,'cof');

    % LIST OF MEASUREMENT PCD
    for imeas = 1:Nmax,
        Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).Num_Mie_Invalid_Reference_Pulse = fread(fid, 1,'*int32');
        Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).Num_Rayleigh_Invalid_Reference_Pulse = fread(fid, 1,'*int32');
        Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).Avg_Laser_Frequency_Offset  = fread(fid, 1, '*float64');
        Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).Avg_UV_Energy = fread(fid, 1, '*float64');
        Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).Laser_Frequency_Offset_Std_Dev = fread(fid, 1, '*float64');
        Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).UV_Energy_Std_Dev = fread(fid, 1, '*float64');
        
        % List of Measurement ALtitude bins PCD
        for ibin = 1:25,
            Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).List_of_Meas_Alt_Bin_PCDs.Meas_Alt_Bin_PCD(ibin).Mie_Measurement_Invalid = fread(fid, 1,'*uint8');
            Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).List_of_Meas_Alt_Bin_PCDs.Meas_Alt_Bin_PCD(ibin).Rayleigh_Measurement_Invalid = fread(fid, 1,'*uint8');
	    switch version
	    case {'ADM_L1bP/1.04 ','ADM_L1bP/1.05 '}
		    Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).List_of_Meas_Alt_Bin_PCDs.Meas_Alt_Bin_PCD(ibin).Invalid_Peak_Mie = fread(fid, 1,'*uint8');
	    case {'ADM_L1bP/1.06 ','ADM_L1bP/1.07 ','ADM_L1bP/1.08 ','ADM_L1bP/1.09 '}
		    Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).List_of_Meas_Alt_Bin_PCDs.Meas_Alt_Bin_PCD(ibin).Mie_Peak_Invalid = fread(fid, 1,'*uint8');
	    end

            Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).List_of_Meas_Alt_Bin_PCDs.Meas_Alt_Bin_PCD(ibin).Mie_Core_Characteristic.Peak_Position = fread(fid, 1, '*float64');
            Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).List_of_Meas_Alt_Bin_PCDs.Meas_Alt_Bin_PCD(ibin).Mie_Core_Characteristic.FWHM = fread(fid, 1, '*float64');
            Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).List_of_Meas_Alt_Bin_PCDs.Meas_Alt_Bin_PCD(ibin).Mie_Core_Characteristic.Amplitude = fread(fid, 1, '*float64');
            Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).List_of_Meas_Alt_Bin_PCDs.Meas_Alt_Bin_PCD(ibin).Mie_Core_Characteristic.Offset = fread(fid, 1, '*float64');
            Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).List_of_Meas_Alt_Bin_PCDs.Meas_Alt_Bin_PCD(ibin).Mie_Core_Characteristic.Error_Flag = fread(fid, 1, '*uint8');
            Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).List_of_Meas_Alt_Bin_PCDs.Meas_Alt_Bin_PCD(ibin).Mie_Core_Characteristic.Residual_Error = fread(fid, 1, '*float64');
            Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).List_of_Meas_Alt_Bin_PCDs.Meas_Alt_Bin_PCD(ibin).Mie_Core_Characteristic.Num_Iterations = fread(fid, 1, '*uint8');

            Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).List_of_Meas_Alt_Bin_PCDs.Meas_Alt_Bin_PCD(ibin).Scattering_Ratio_Mie = fread(fid, 1, '*float64');
            Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).List_of_Meas_Alt_Bin_PCDs.Meas_Alt_Bin_PCD(ibin).Mie_Signal_to_Noise_Ratio = fread(fid, 1, '*float64');
            Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).List_of_Meas_Alt_Bin_PCDs.Meas_Alt_Bin_PCD(ibin).Rayleigh_Signal_to_Noise_Ratio_Channel_A = fread(fid, 1, '*float64');
            Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).List_of_Meas_Alt_Bin_PCDs.Meas_Alt_Bin_PCD(ibin).Rayleigh_Signal_to_Noise_Ratio_Channel_B = fread(fid, 1, '*float64');
        end

        Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).Velocity_of_Attitude_Uncertainty_Error = fread(fid, 1,'*float64');
        Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).Mie_Mean_Emitted_Frequency = fread(fid, 1,'*float64');
        Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).Mie_Emitted_Frequency_Std_Dev = fread(fid, 1,'*float64');
        Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).Reference_Pulse_FWHM = fread(fid, 1,'*float64');
        Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).Rayleigh_Mean_Emitted_Frequency = fread(fid, 1,'*float64');
        Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).Rayleigh_Emitted_Frequency_Std_Dev = fread(fid, 1,'*float64');
        Product_Confidence_Data_ADS.DSR(idsr).List_of_Measurement_PCDs.Measurement_PCD(imeas).UV_Energy_Quality_Flag = fread(fid, 1, '*uint8');   % L1b v1.04
        %spare
        fseek(fid, 8,'cof');
    end % for Nmax
    %spare
    fseek(fid, 8,'cof');
end % for idsr

end % function
    
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% %%%%%%%% L1b Ground Wind Detection ADS%%%%%%%%%%%%%
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 
function Ground_Wind_Detection_ADS = L1b_Read_GWD_ADS(fid,nbDSR_GWD,Nmax,version)

Ground_Wind_Detection_ADS = [];

for idsr = 1:nbDSR_GWD,
    ftell(fid);
    Ground_Wind_Detection_ADS.DSR(idsr).Start_of_Observation_Time(1) = fread(fid, 1,'*int32');
    Ground_Wind_Detection_ADS.DSR(idsr).Start_of_Observation_Time(2:3) = fread(fid, 2,'*uint32');
    Ground_Wind_Detection_ADS.DSR(idsr).Mie_Ground_Correction_Velocity = fread(fid, 1 ,'*float64');
    Ground_Wind_Detection_ADS.DSR(idsr).Rayleigh_Ground_Correction_Velocity = fread(fid, 1 ,'*float64');
    Ground_Wind_Detection_ADS.DSR(idsr).Updated_Mie_Ground_Correction_Velocity = fread(fid, 1,'*uint8');
    Ground_Wind_Detection_ADS.DSR(idsr).Updated_Rayleigh_Ground_Correction_Velocity = fread(fid, 1,'*uint8');
    Ground_Wind_Detection_ADS.DSR(idsr).Mie_Average_Ground_Wind_Bin_Altitude = fread(fid, 1 ,'*float64');
    Ground_Wind_Detection_ADS.DSR(idsr).Ray_Average_Ground_Wind_Bin_Altitude = fread(fid, 1 ,'*float64');
    Ground_Wind_Detection_ADS.DSR(idsr).Mie_Average_Ground_Wind_Bin_Thickness_Above_DEM = fread(fid, 1 ,'*float64');
    Ground_Wind_Detection_ADS.DSR(idsr).Rayleigh_Average_Ground_Wind_Bin_Thickness_Above_DEM = fread(fid, 1 ,'*float64');

    % Validation criteria
    Ground_Wind_Detection_ADS.DSR(idsr).Validation_Criteria.Min_Num_Mie_Ground_Echo_Measurement_Bins = fread(fid, 1,'*uint8');
    Ground_Wind_Detection_ADS.DSR(idsr).Validation_Criteria.Min_Num_Rayleigh_Ground_Echo_Measurement_Bins = fread(fid, 1,'*uint8');
    Ground_Wind_Detection_ADS.DSR(idsr).Validation_Criteria.Num_Mie_Ground_Bins = fread(fid, 1,'*uint8');
    Ground_Wind_Detection_ADS.DSR(idsr).Validation_Criteria.Num_Rayleigh_Ground_Bins = fread(fid, 1,'*uint8');
    fseek(fid, 8, 'cof');

    % LIST OF MEASUREMENTS GROUND WIND DETECTION
    switch version
    case {'ADM_L1bP/1.04 ','ADM_L1bP/1.05 '}
	    for imeas = 1:Nmax,
	        % Mie Measurement Ground Wind bin
	        Ground_Wind_Detection_ADS.DSR(idsr).List_of_Measurement_Ground_Wind_Detections.Measurement_Ground_Wind_Detection(imeas).Mie_Measurement_Ground_Wind_Bin.Ground_Wind_Detected =...
	            fread(fid, 1, '*uint8');
	        Ground_Wind_Detection_ADS.DSR(idsr).List_of_Measurement_Ground_Wind_Detections.Measurement_Ground_Wind_Detection(imeas).Mie_Measurement_Ground_Wind_Bin.Ground_Bin_Number =...
	            fread(fid, 1, '*uint8');
	        Ground_Wind_Detection_ADS.DSR(idsr).List_of_Measurement_Ground_Wind_Detections.Measurement_Ground_Wind_Detection(imeas).Mie_Measurement_Ground_Wind_Bin.Ground_Bin_Thickness_Above_DEM =...
	            fread(fid, 1 ,'*float64');
	        % Rayleigh Measurement Ground Wind bin
	        Ground_Wind_Detection_ADS.DSR(idsr).List_of_Measurement_Ground_Wind_Detections.Measurement_Ground_Wind_Detection(imeas).Rayleigh_Measurement_Ground_Wind_Bin.Ground_Wind_Detected =...
	            fread(fid, 1, '*uint8');
	        Ground_Wind_Detection_ADS.DSR(idsr).List_of_Measurement_Ground_Wind_Detections.Measurement_Ground_Wind_Detection(imeas).Rayleigh_Measurement_Ground_Wind_Bin.Ground_Bin_Number =...
	            fread(fid, 1, '*uint8');
	        Ground_Wind_Detection_ADS.DSR(idsr).List_of_Measurement_Ground_Wind_Detections.Measurement_Ground_Wind_Detection(imeas).Rayleigh_Measurement_Ground_Wind_Bin.Ground_Bin_Thickness_Above_DEM =...
	            fread(fid, 1 ,'*float64');
	    end %for imeas
    case {'ADM_L1bP/1.06 ','ADM_L1bP/1.07 ','ADM_L1bP/1.08 ','ADM_L1bP/1.09 '}
	    for imeas = 1:Nmax,
	        % Mie Measurement Ground Wind bin
	        Ground_Wind_Detection_ADS.DSR(idsr).List_of_Measurement_Ground_Wind_Detections.Mie_Measurement_Ground_Wind_Bin(imeas).Ground_Wind_Detected =...
	            fread(fid, 1, '*uint8');
	        Ground_Wind_Detection_ADS.DSR(idsr).List_of_Measurement_Ground_Wind_Detections.Mie_Measurement_Ground_Wind_Bin(imeas).Ground_Bin_Number =...
	            fread(fid, 1, '*uint8');
	        Ground_Wind_Detection_ADS.DSR(idsr).List_of_Measurement_Ground_Wind_Detections.Mie_Measurement_Ground_Wind_Bin(imeas).Ground_Bin_Thickness_Above_DEM =...
	            fread(fid, 1 ,'*float64');
	        % Rayleigh Measurement Ground Wind bin
	        Ground_Wind_Detection_ADS.DSR(idsr).List_of_Measurement_Ground_Wind_Detections.Rayleigh_Measurement_Ground_Wind_Bin(imeas).Ground_Wind_Detected =...
	            fread(fid, 1, '*uint8');
	        Ground_Wind_Detection_ADS.DSR(idsr).List_of_Measurement_Ground_Wind_Detections.Rayleigh_Measurement_Ground_Wind_Bin(imeas).Ground_Bin_Number =...
	            fread(fid, 1, '*uint8');
	        Ground_Wind_Detection_ADS.DSR(idsr).List_of_Measurement_Ground_Wind_Detections.Rayleigh_Measurement_Ground_Wind_Bin(imeas).Ground_Bin_Thickness_Above_DEM =...
	            fread(fid, 1 ,'*float64');
	    end %for imeas
    end

    Ground_Wind_Detection_ADS.DSR(idsr).Mie_Ground_Correction_Weighting_Factor = fread(fid, 1 ,'*float64');
    Ground_Wind_Detection_ADS.DSR(idsr).Rayleigh_Ground_Correction_Weighting_Factor = fread(fid, 1 ,'*float64');
    switch version
    case {'ADM_L1bP/1.04 ','ADM_L1bP/1.05 ','ADM_L1bP/1.06 ','ADM_L1bP/1.07 '}
	    Ground_Wind_Detection_ADS.DSR(idsr).Mie_Channel_Total_Zero_Wind_Correction = fread(fid, 1,'*int32');                   %L1b 1.04
	    Ground_Wind_Detection_ADS.DSR(idsr).Rayleigh_Channel_Total_Zero_Wind_Correction = fread(fid, 1,'*int32');              %L1b 1.04
    case {'ADM_L1bP/1.08 ','ADM_L1bP/1.09 '}
	    Ground_Wind_Detection_ADS.DSR(idsr).Mie_Channel_Total_Zero_Wind_Correction = fread(fid, 1,'*float64');                   %L1b 1.08
	    Ground_Wind_Detection_ADS.DSR(idsr).Rayleigh_Channel_Total_Zero_Wind_Correction = fread(fid, 1,'*float64');              %L1b 1.08
    end
    Ground_Wind_Detection_ADS.DSR(idsr).HBE_Mie_Ground_Correction_Velocity = fread(fid, 1,'*float64');                     %L1b 1.04
    Ground_Wind_Detection_ADS.DSR(idsr).HBE_Rayleigh_Ground_Correction_Velocity = fread(fid, 1,'*float64');                %L1b 1.04
    %spare
    fseek(fid, 16, 'cof');
end % for idsr

end % function

% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% %%%%%%%%% L1b  Measurement ADS %%%%%%%%%%%%%%%%%%%%
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function Measurement_ADS = L1b_Read_Meas_ADS(fid,nbDSR_Meas,Nmax)

Measurement_ADS = [];

for idsr = 1:nbDSR_Meas,
    Measurement_ADS.DSR(idsr).Start_of_Observation_Time(1) = fread(fid, 1,'*int32');
    Measurement_ADS.DSR(idsr).Start_of_Observation_Time(2:3) = fread(fid, 2,'*uint32');
    Measurement_ADS.DSR(idsr).Num_of_Reference_Pulses = fread(fid, 1, '*uint32');
    %spare
    fseek(fid,4,'cof');
    % LIST OF MIE REFERENCE PULSES
    for imeas = 1:Nmax,
        Measurement_ADS.DSR(idsr).List_of_Mie_Reference_Pulse.Mie_Reference_Pulse(imeas).col = fread(fid, 20, '*uint16');
    end
    % LIST OF RAYLEIGH REFERENCE PULSES CHANNEL A
    for imeas=1:Nmax,
        Measurement_ADS.DSR(idsr).List_of_Rayleigh_Reference_Pulse_A.Rayleigh_Reference_Pulse_A(imeas).col = fread(fid, 1,'*float64');
    end
    % LIST OF RAYLEIGH REFERENCE PULSES CHANNEL B
    for imeas=1:Nmax,
        Measurement_ADS.DSR(idsr).List_of_Rayleigh_Reference_Pulse_B.Rayleigh_Reference_Pulse_B(imeas).data = fread(fid, 1,'*float64');
    end
    % MIE MEASUREMENT DATA
    for imeas=1:Nmax,
        Measurement_ADS.DSR(idsr).List_of_Mie_Measurement_Data.Mie_Measurement_Data(imeas).data = fread(fid, [20,25], '*uint16');
    end
    % MIE TIME DELAYS
    for ibin=1:24,
        Measurement_ADS.DSR(idsr).Mie_Time_Delays.List_of_Bin_Layer_Integration_Time.Bin_Layer_Integration_Time(ibin).time = fread(fid, 1, '*int32');
    end
    Measurement_ADS.DSR(idsr).Mie_Time_Delays.Background_Integration_Time = fread(fid, 1, '*int32');

    % RAYLEIGH TIME DELAYS
    for ibin=1:24,
        Measurement_ADS.DSR(idsr).Rayleigh_Time_Delays.List_of_Bin_Layers_Integration_Time.Bin_Layers_Integration_Time(ibin).time = fread(fid, 1, '*int32');
    end
    Measurement_ADS.DSR(idsr).Rayleigh_Time_Delays.Background_Integration_Time = fread(fid, 1, '*int32');

    % LIST OF MEASUREMENT VALIDITY INDICATOR
    for imeas = 1:Nmax,
        Measurement_ADS.DSR(idsr).List_of_Measurement_Validity_Indicators.Measurement_Validity_Indicators(imeas).Measurement_Data_Present = fread(fid, 1, '*uint8');
        Measurement_ADS.DSR(idsr).List_of_Measurement_Validity_Indicators.Measurement_Validity_Indicators(imeas).Mie_Measurement_Sp_Valid = fread(fid, 1, '*uint8');
        Measurement_ADS.DSR(idsr).List_of_Measurement_Validity_Indicators.Measurement_Validity_Indicators(imeas).Rayleigh_Measurement_Sp_Valid = fread(fid, 1, '*uint8');
        Measurement_ADS.DSR(idsr).List_of_Measurement_Validity_Indicators.Measurement_Validity_Indicators(imeas).Measurement_Laser_Freq_Locked = fread(fid, 1, '*uint8');
        Measurement_ADS.DSR(idsr).List_of_Measurement_Validity_Indicators.Measurement_Validity_Indicators(imeas).Spacecraft_Attitude_On_Target = fread(fid, 1, '*uint8');
    end

end %idsr

end % function
% 
% 
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% %%% L1b Calibration and Characterization GADS %%%%%%%
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function Calibration_Char_Data_GADS = L1b_Read_CCD_GADS(fid)

NumFrequencySteps=40;

Calibration_Char_Data_GADS = [];

Calibration_Char_Data_GADS.MRC_First_Start_of_Observation_Time(1) = fread(fid, 1,'*int32');
Calibration_Char_Data_GADS.MRC_First_Start_of_Observation_Time(2:3) = fread(fid, 2,'*uint32');
Calibration_Char_Data_GADS.MRC_Last_Start_of_Observation_Time(1) = fread(fid, 1,'*int32');
Calibration_Char_Data_GADS.MRC_Last_Start_of_Observation_Time(2:3) = fread(fid, 2,'*uint32');
Calibration_Char_Data_GADS.RRC_First_Start_of_Observation_Time(1) = fread(fid, 1,'*int32');
Calibration_Char_Data_GADS.RRC_First_Start_of_Observation_Time(2:3) = fread(fid, 2,'*uint32');
Calibration_Char_Data_GADS.RRC_Last_Start_of_Observation_Time(1) = fread(fid, 1,'*int32');
Calibration_Char_Data_GADS.RRC_Last_Start_of_Observation_Time(2:3) = fread(fid, 2,'*uint32');
%%%%%%%% L1B CHARACTERIZATION DATA %%%%%%%%
%%%% Satellite Characterisation Data %%%%
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Satellite_Characterisation_Data.Laser_Wavelength = fread(fid, 1, '*float64');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Satellite_Characterisation_Data.ErrorQuantifiers.Mie_Error_Quantifier_K1 = fread(fid, 1, '*float64');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Satellite_Characterisation_Data.ErrorQuantifiers.Mie_Error_Quantifier_K2 = fread(fid, 1, '*float64');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Satellite_Characterisation_Data.ErrorQuantifiers.Mie_Error_Quantifier_K3 = fread(fid, 1, '*float64');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Satellite_Characterisation_Data.ErrorQuantifiers.Rayleigh_Error_Quantifier_Ka2 = fread(fid, 1, '*float64');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Satellite_Characterisation_Data.ErrorQuantifiers.Rayleigh_Error_Quantifier_Ka3 = fread(fid, 1, '*float64');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Satellite_Characterisation_Data.ErrorQuantifiers.Rayleigh_Error_Quantifier_Kb2 = fread(fid, 1, '*float64');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Satellite_Characterisation_Data.ErrorQuantifiers.Rayleigh_Error_Quantifier_Kb3 = fread(fid, 1, '*float64');
for i=1:16,
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Satellite_Characterisation_Data.List_of_Tripod_Obscuration_Correction.Tripod_Obscuration_Correction(i).corr = fread(fid, 1, '*float64');
end
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Satellite_Characterisation_Data.NF_Order = fread(fid,1, '*uint32');  %L1b 1.04
NFplus1 = Calibration_Char_Data_GADS.L1B_Characterisation_Data.Satellite_Characterisation_Data.NF_Order + 1;
for i=1:NFplus1,
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Satellite_Characterisation_Data.List_of_Mie_Harmonic_Bias_Coefficient_A.Mie_Harmonic_Bias_Coefficient_A(i).Harmonic_Bias_Coefficient_A = fread(fid, 1, '*float64');
end
for i=1:NFplus1,
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Satellite_Characterisation_Data.List_of_Mie_Harmonic_Bias_Coefficient_B.Mie_Harmonic_Bias_Coefficient_B(i).Harmonic_Bias_Coefficient_B = fread(fid, 1, '*float64');
end
for i=1:NFplus1,
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Satellite_Characterisation_Data.List_of_Ray_Harmonic_Bias_Coefficient_A.Ray_Harmonic_Bias_Coefficient_A(i).Harmonic_Bias_Coefficient_A = fread(fid, 1, '*float64');
end
for i=1:NFplus1,
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Satellite_Characterisation_Data.List_of_Ray_Harmonic_Bias_Coefficient_B.Ray_Harmonic_Bias_Coefficient_B(i).Harmonic_Bias_Coefficient_B = fread(fid, 1, '*float64');
end

%%%% Mie Response Calibration Data %%%%
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.Calibration_Valid = fread(fid, 1, '*uint8');
%% List of mie frequency step results
for istep=1:NumFrequencySteps,
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.List_of_Mie_Frequency_Step_Results.Mie_Frequency_Step_Result(istep).Frequency_Offset = fread(fid, 1, '*float64');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.List_of_Mie_Frequency_Step_Results.Mie_Frequency_Step_Result(istep).Frequency_Valid = fread(fid, 1, '*uint8');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.List_of_Mie_Frequency_Step_Results.Mie_Frequency_Step_Result(istep).Measurement_Response_Valid = fread(fid, 1, '*uint8');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.List_of_Mie_Frequency_Step_Results.Mie_Frequency_Step_Result(istep).Reference_Pulse_Response_Valid = fread(fid, 1, '*uint8');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.List_of_Mie_Frequency_Step_Results.Mie_Frequency_Step_Result(istep).Measurement_Response = fread(fid, 1, '*float64');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.List_of_Mie_Frequency_Step_Results.Mie_Frequency_Step_Result(istep).Measurement_Error_Mie_Response = fread(fid, 1, '*float64');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.List_of_Mie_Frequency_Step_Results.Mie_Frequency_Step_Result(istep).Reference_Pulse_Response = fread(fid, 1, '*float64');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.List_of_Mie_Frequency_Step_Results.Mie_Frequency_Step_Result(istep).Reference_Pulse_Error_Mie_Response = fread(fid, 1, '*float64');
    % Frequency Step Data Statistics
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.List_of_Mie_Frequency_Step_Results.Mie_Frequency_Step_Result(istep).Frequency_Step_Data_Statistic.Num_Valid_Measurements = fread(fid, 1, '*int32');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.List_of_Mie_Frequency_Step_Results.Mie_Frequency_Step_Result(istep).Frequency_Step_Data_Statistic.Num_Measurements_Usable = fread(fid, 1, '*int32');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.List_of_Mie_Frequency_Step_Results.Mie_Frequency_Step_Result(istep).Frequency_Step_Data_Statistic.Num_Reference_Pulses_Usable = fread(fid, 1, '*int32');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.List_of_Mie_Frequency_Step_Results.Mie_Frequency_Step_Result(istep).Frequency_Step_Data_Statistic.Num_Measurement_Laser_Freq_Unlocked = fread(fid, 1, '*int32');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.List_of_Mie_Frequency_Step_Results.Mie_Frequency_Step_Result(istep).Frequency_Step_Data_Statistic.Num_Reference_Pulse_Laser_Freq_Unlocked = fread(fid, 1, '*int32');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.List_of_Mie_Frequency_Step_Results.Mie_Frequency_Step_Result(istep).Frequency_Step_Data_Statistic.Num_Sat_Not_on_Target_Measurements = fread(fid, 1, '*int32');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.List_of_Mie_Frequency_Step_Results.Mie_Frequency_Step_Result(istep).Frequency_Step_Data_Statistic.Num_Corrupt_Measurements = fread(fid, 1, '*int32');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.List_of_Mie_Frequency_Step_Results.Mie_Frequency_Step_Result(istep).Frequency_Step_Data_Statistic.Num_Corrupt_Reference_Pulses = fread(fid, 1, '*int32');
end
%% Mie Measurement Response Calibration
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.Mie_Measurement_Response_Calibration.Measurement_Mean_Sensitivity = fread(fid, 1, '*float64');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.Mie_Measurement_Response_Calibration.Measurement_Zero_Frequency = fread(fid, 1, '*float64');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.Mie_Measurement_Response_Calibration.Measurement_Error_Mie_Response_Std_Dev = fread(fid, 1, '*float64');

%% Mie Reference Pulse Response Calibration
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.Mie_Reference_Pulse_Response_Calibration.Reference_Pulse_Mean_Sensitivity = fread(fid, 1, '*float64');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.Mie_Reference_Pulse_Response_Calibration.Reference_Pulse_Zero_Frequency = fread(fid, 1, '*float64');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.Mie_Reference_Pulse_Response_Calibration.Reference_Pulse_Error_Mie_Response_Std_Dev = fread(fid, 1, '*float64');

%% Measurement Calibration Validity
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.Measurement_Calibration_Validity.Mean_Sensitivity_Valid = fread(fid, 1, '*uint8');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.Measurement_Calibration_Validity.Error_Response_Std_Dev_Valid = fread(fid, 1, '*uint8');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.Measurement_Calibration_Validity.Zero_Freq_Response_Valid = fread(fid, 1, '*uint8');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.Measurement_Calibration_Validity.Data_Monotonic = fread(fid, 1, '*uint8');

%% Reference Pulse Calibration Validity
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.Reference_Pulse_Calibration_Validity.Mean_Sensitivity_Valid = fread(fid, 1, '*uint8');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.Reference_Pulse_Calibration_Validity.Error_Response_Std_Dev_Valid = fread(fid, 1, '*uint8');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.Reference_Pulse_Calibration_Validity.Zero_Freq_Response_Valid = fread(fid, 1, '*uint8');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Mie_Response_Calibration_Data.Reference_Pulse_Calibration_Validity.Data_Monotonic = fread(fid, 1, '*uint8');

%%%% Rayleigh Response Calibration Data %%%%
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.Calibration_Valid = fread(fid, 1, '*uint8');
%% List of rayleigh frequency step results
for istep=1:NumFrequencySteps,
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.List_of_Rayleigh_Frequency_Step_Results.Rayleigh_Frequency_Step_Result(istep).Frequency_Offset = fread(fid, 1, '*float64');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.List_of_Rayleigh_Frequency_Step_Results.Rayleigh_Frequency_Step_Result(istep).Frequency_Valid = fread(fid, 1, '*uint8');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.List_of_Rayleigh_Frequency_Step_Results.Rayleigh_Frequency_Step_Result(istep).Measurment_Response_Valid = fread(fid, 1, '*uint8');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.List_of_Rayleigh_Frequency_Step_Results.Rayleigh_Frequency_Step_Result(istep).Reference_Pulse_Response_Valid = fread(fid, 1, '*uint8');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.List_of_Rayleigh_Frequency_Step_Results.Rayleigh_Frequency_Step_Result(istep).Measurement_Response = fread(fid, 1, '*float64');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.List_of_Rayleigh_Frequency_Step_Results.Rayleigh_Frequency_Step_Result(istep).Measurement_Error_Rayleigh_Response = fread(fid, 1, '*float64');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.List_of_Rayleigh_Frequency_Step_Results.Rayleigh_Frequency_Step_Result(istep).Reference_Pulse_Response = fread(fid, 1, '*float64');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.List_of_Rayleigh_Frequency_Step_Results.Rayleigh_Frequency_Step_Result(istep).Reference_Pulse_Error_Rayleigh_Response = fread(fid, 1, '*float64');
    % Frequency Step Data Statistics
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.List_of_Rayleigh_Frequency_Step_Results.Rayleigh_Frequency_Step_Result(istep).Frequency_Step_Data_Statistic.Num_Valid_Measurements = fread(fid, 1, '*int32');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.List_of_Rayleigh_Frequency_Step_Results.Rayleigh_Frequency_Step_Result(istep).Frequency_Step_Data_Statistic.Num_Measurements_Usable = fread(fid, 1, '*int32');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.List_of_Rayleigh_Frequency_Step_Results.Rayleigh_Frequency_Step_Result(istep).Frequency_Step_Data_Statistic.Num_Reference_Pulses_Usable = fread(fid, 1, '*int32');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.List_of_Rayleigh_Frequency_Step_Results.Rayleigh_Frequency_Step_Result(istep).Frequency_Step_Data_Statistic.Num_Measurement_Laser_Freq_Unlocked = fread(fid, 1, '*int32');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.List_of_Rayleigh_Frequency_Step_Results.Rayleigh_Frequency_Step_Result(istep).Frequency_Step_Data_Statistic.Num_Reference_Pulse_Laser_Freq_Unlocked = fread(fid, 1, '*int32');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.List_of_Rayleigh_Frequency_Step_Results.Rayleigh_Frequency_Step_Result(istep).Frequency_Step_Data_Statistic.Num_Sat_Not_on_Target_Measurements = fread(fid, 1, '*int32');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.List_of_Rayleigh_Frequency_Step_Results.Rayleigh_Frequency_Step_Result(istep).Frequency_Step_Data_Statistic.Num_Corrupt_Measurements = fread(fid, 1, '*int32');
    Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.List_of_Rayleigh_Frequency_Step_Results.Rayleigh_Frequency_Step_Result(istep).Frequency_Step_Data_Statistic.Num_Corrupt_Reference_Pulses = fread(fid, 1, '*int32');
end
%% Rayleigh Measurement Response Calibration
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.Rayleigh_Measurement_Response_Calibration.Measurement_Mean_Sensitivity = fread(fid, 1, '*float64');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.Rayleigh_Measurement_Response_Calibration.Measurement_Zero_Frequency = fread(fid, 1, '*float64');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.Rayleigh_Measurement_Response_Calibration.Measurement_Error_Rayleigh_Response_Std_Dev = fread(fid, 1, '*float64');

%% Rayleigh Reference Pulse Response Calibration
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.Rayleigh_Reference_Pulse_Response_Calibration.Reference_Pulse_Mean_Sensitivity = fread(fid, 1, '*float64');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.Rayleigh_Reference_Pulse_Response_Calibration.Reference_Pulse_Zero_Frequency = fread(fid, 1, '*float64');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.Rayleigh_Reference_Pulse_Response_Calibration.Reference_Pulse_Error_Rayleigh_Response_Std_Dev = fread(fid, 1, '*float64');

%% Measurement Calibration Validity
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.Measurement_Calibration_Validity.Mean_Sensitivity_Valid = fread(fid, 1, '*uint8');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.Measurement_Calibration_Validity.Error_Response_Std_Dev_Valid = fread(fid, 1, '*uint8');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.Measurement_Calibration_Validity.Zero_Freq_Response_Valid = fread(fid, 1, '*uint8');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.Measurement_Calibration_Validity.Data_Monotonic = fread(fid, 1, '*uint8');

%% Reference Pulse Calibration Validity
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.Reference_Pulse_Calibration_Validity.Mean_Sensitivity_Valid = fread(fid, 1, '*uint8');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.Reference_Pulse_Calibration_Validity.Error_Response_Std_Dev_Valid = fread(fid, 1, '*uint8');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.Reference_Pulse_Calibration_Validity.Zero_Freq_Response_Valid = fread(fid, 1, '*uint8');
Calibration_Char_Data_GADS.L1B_Characterisation_Data.Rayleigh_Response_Calibration_Data.Reference_Pulse_Calibration_Validity.Data_Monotonic = fread(fid, 1, '*uint8');

end % function


% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% %%%%%%%% L1b Useful Signal MDS %%%%%%%%%%%%%%%%%%%%
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function Useful_Signal_MDS = L1b_Read_US_MDS(fid,nbDSR_US,Nmax)

Useful_Signal_MDS = [];

for idsr = 1:nbDSR_US,
    Useful_Signal_MDS.DSR(idsr).Start_of_Observation_Time(1) = fread(fid, 1,'*int32');
    Useful_Signal_MDS.DSR(idsr).Start_of_Observation_Time(2:3) = fread(fid, 2,'*uint32');
    % Observation Useful Signals
    for ibin = 1:25,
        Useful_Signal_MDS.DSR(idsr).Observation_Useful_Signal.List_of_Mie_Altitude_Bin_Useful_Signal_Infos.Mie_Altitude_Bin_Useful_Signal_Info(ibin).Data_Quality_Flag = fread(fid, 8, '*ubit1');
        Useful_Signal_MDS.DSR(idsr).Observation_Useful_Signal.List_of_Mie_Altitude_Bin_Useful_Signal_Infos.Mie_Altitude_Bin_Useful_Signal_Info(ibin).Useful_Signal = fread(fid, 1, '*float64');
    end %for index
    for ibin = 1:25,
        Useful_Signal_MDS.DSR(idsr).Observation_Useful_Signal.List_of_Rayleigh_Altitude_Bin_Useful_Signal_Infos.Rayleigh_Altitude_Bin_Useful_Signal_Info(ibin).Data_Quality_Flag = fread(fid, 8, '*ubit1');
        Useful_Signal_MDS.DSR(idsr).Observation_Useful_Signal.List_of_Rayleigh_Altitude_Bin_Useful_Signal_Infos.Rayleigh_Altitude_Bin_Useful_Signal_Info(ibin).Useful_Signal_Channel_A = fread(fid, 1, '*float64');
        Useful_Signal_MDS.DSR(idsr).Observation_Useful_Signal.List_of_Rayleigh_Altitude_Bin_Useful_Signal_Infos.Rayleigh_Altitude_Bin_Useful_Signal_Info(ibin).Useful_Signal_Channel_B = fread(fid, 1, '*float64');
    end %for index

    % List of Measurement Useful Signals
    for imeas = 1: Nmax,
        for ibin = 1:25,
            Useful_Signal_MDS.DSR(idsr).List_of_Measurement_Useful_Signal.Measurement_Useful_Signal(imeas).List_of_Mie_Altitude_Bin_Useful_Signal_Infos.Mie_Altitude_Bin_Useful_Signal_Info(ibin).Data_Quality_Flag = fread(fid, 8, '*ubit1');
            Useful_Signal_MDS.DSR(idsr).List_of_Measurement_Useful_Signal.Measurement_Useful_Signal(imeas).List_of_Mie_Altitude_Bin_Useful_Signal_Infos.Mie_Altitude_Bin_Useful_Signal_Info(ibin).Useful_Signal = fread(fid, 1, '*float64');
        end %for index
        for ibin = 1:25,
            Useful_Signal_MDS.DSR(idsr).List_of_Measurement_Useful_Signal.Measurement_Useful_Signal(imeas).List_of_Rayleigh_Altitude_Bin_Useful_Signal_Infos.Rayleigh_Altitude_Bin_Useful_Signal_Info(ibin).Data_Quality_Flag = fread(fid, 8, '*ubit1');
            Useful_Signal_MDS.DSR(idsr).List_of_Measurement_Useful_Signal.Measurement_Useful_Signal(imeas).List_of_Rayleigh_Altitude_Bin_Useful_Signal_Infos.Rayleigh_Altitude_Bin_Useful_Signal_Info(ibin).Useful_Signal_Channel_A = fread(fid, 1, '*float64');
            Useful_Signal_MDS.DSR(idsr).List_of_Measurement_Useful_Signal.Measurement_Useful_Signal(imeas).List_of_Rayleigh_Altitude_Bin_Useful_Signal_Infos.Rayleigh_Altitude_Bin_Useful_Signal_Info(ibin).Useful_Signal_Channel_B = fread(fid, 1, '*float64');
        end %for ibin
    end % for imeas

end %for idsr

end % function
% 
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% %%%%%%%% L1b Wind Velocity MDS %%%%%%%%%%%%%%%%%%%%
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function  Wind_Velocity_MDS = L1b_Read_WV_ADS(fid,nbDSR_WV,Nmax)

Wind_Velocity_MDS = [];

for idsr = 1:nbDSR_WV,
    Wind_Velocity_MDS.DSR(idsr).Start_of_Observation_Time(1) = fread(fid, 1,'*int32');
    Wind_Velocity_MDS.DSR(idsr).Start_of_Observation_Time(2:3) = fread(fid, 2,'*uint32');
    Wind_Velocity_MDS.DSR(idsr).Line_of_Sight_Wind_Flag = fread(fid, 1,'*int8');

    %%%%%%%% Observation Wind Profile %%%%%%%%
    Wind_Velocity_MDS.DSR(idsr).Observation_Wind_Profile.Mie_Reference_Pulse_Quality_Flag = fread(fid, 8, '*ubit1');
    Wind_Velocity_MDS.DSR(idsr).Observation_Wind_Profile.Rayleigh_Reference_Pulse_Quality_Flag = fread(fid, 8, '*ubit1');

    % List of Mie Altitude Bin Wind
    for ibin = 1:24,
        Wind_Velocity_MDS.DSR(idsr).Observation_Wind_Profile.List_of_Mie_Altitude_Bin_Wind_Infos.Mie_Altitude_Bin_Wind_Info(ibin).Bin_Quality_Flag = fread(fid, 8, '*ubit1');
        Wind_Velocity_MDS.DSR(idsr).Observation_Wind_Profile.List_of_Mie_Altitude_Bin_Wind_Infos.Mie_Altitude_Bin_Wind_Info(ibin).Wind_Velocity = fread(fid, 1, '*float64');
    end

    % List of Rayleigh Bin Wind
    for ibin = 1:24,
        Wind_Velocity_MDS.DSR(idsr).Observation_Wind_Profile.List_of_Rayleigh_Altitude_Bin_Wind_Infos.Rayleigh_Altitude_Bin_Wind_Info(ibin).Bin_Quality_Flag = fread(fid, 8, '*ubit1');
        Wind_Velocity_MDS.DSR(idsr).Observation_Wind_Profile.List_of_Rayleigh_Altitude_Bin_Wind_Infos.Rayleigh_Altitude_Bin_Wind_Info(ibin).Wind_Velocity = fread(fid, 1, '*float64');
    end

    %%%%%% List of Measurement Wind Profile %%%%%%%%%%
    for imeas = 1:Nmax,
        Wind_Velocity_MDS.DSR(idsr).List_of_Measurement_Wind_Profile.Measurement_Wind_Profile(imeas).Mie_Reference_Pulse_Quality_Flag = fread(fid, 8, '*ubit1');
        Wind_Velocity_MDS.DSR(idsr).List_of_Measurement_Wind_Profile.Measurement_Wind_Profile(imeas).Rayleigh_Reference_Pulse_Quality_Flag = fread(fid, 8, '*ubit1');

        % List of Mie Altitude Bin Wind Infos
        for ibin = 1:24,
            Wind_Velocity_MDS.DSR(idsr).List_of_Measurement_Wind_Profile.Measurement_Wind_Profile(imeas).List_of_Mie_Altitude_Bin_Wind_Infos.Mie_Altitude_Bin_Wind_Info(ibin).Bin_Quality_Flag = fread(fid, 8, '*ubit1');
            Wind_Velocity_MDS.DSR(idsr).List_of_Measurement_Wind_Profile.Measurement_Wind_Profile(imeas).List_of_Mie_Altitude_Bin_Wind_Infos.Mie_Altitude_Bin_Wind_Info(ibin).Wind_Velocity = fread(fid, 1, '*float64');
        end
        % List of Ray Altitude Bin Wind Infos
        for ibin = 1:24,
            Wind_Velocity_MDS.DSR(idsr).List_of_Measurement_Wind_Profile.Measurement_Wind_Profile(imeas).Bin_Quality_Flag(ibin).data = fread(fid, 8, '*ubit1');
            Wind_Velocity_MDS.DSR(idsr).List_of_Measurement_Wind_Profile.Measurement_Wind_Profile(imeas).List_of_Rayleigh_Altitude_Bin_Wind_Infos.Rayleigh_Altitude_Bin_Wind_Info(ibin).Wind_Velocity = fread(fid, 1, '*float64');
        end
    end

end %for idsr

end %function
% 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       MATLAB CODE TO READ MET SPH 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% HISTORY
%
% 03/2006 P. Poli    Original code.
% 04/2006 P. Poli    Upgrade to IODD AE_IF_ECMWF_L2BP_001_20060427
% 07/2006 D. Tan     Adjust offsets by 1.
% 02/2007 P. Poli    Use IODD variable names
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DOCUMENTATION
%
% The function ReadMETSPH decodes the string of characters
% making up the L2B Auxiliary Meteorological Data (AMD)
% Specific Product Header (SPH).
%
% The output is a structure containing the elements of 
% the L2B Aux Met Data SPH.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function SPH = ReadMETSPH(SPH_chain);

SPH1=SPH_chain(1:591);

% decode variables contained in SPH

SPH.Sph_Descriptor=(SPH1(16+[1:28])');
SPH.Ref_NWP_Suite=(SPH1(102+[1:20])');
SPH.Fcst_Initial_Time=(SPH1(143+[1:27])');
SPH.Model_Timestep=str2num((SPH1(187+[1:11])'));
SPH.Model_Grid_Type=(SPH1(218+[1:2])');
SPH.Model_Resol_Par1=str2num((SPH1(238+[1:6])'));
SPH.Model_Resol_Par2=str2num((SPH1(262+[1:6])'));
SPH.Num_of_Model_Layers=str2num((SPH1(289+[1:6])'));
SPH.Num_Records_in_DS1=str2num((SPH1(315+[1:11])'));
SPH.Num_Records_in_DS2=str2num((SPH1(346+[1:11])'));
SPH.Num_Avail_L1B_Obs=str2num((SPH1(376+[1:11])'));
SPH.Num_Missing_L1B_Obs=str2num((SPH1(408+[1:11])'));
SPH.Num_Computed_Locations=str2num((SPH1(443+[1:11])'));
SPH.Num_Input_Files=str2num((SPH1(512+[1:6])'));
SPH.Num_Files_Predicted_Orbit=str2num((SPH1(543+[1:6])'));

% Read DSDs
DSDs=SPH_chain(592:end);
for k=1:4,
	SPH.DSD(k)=ReadDSD(DSDs([288*(k-1)+1:288*(k-1)+288])');
end
SPH.GeoADS1_DSD=SPH.DSD(1);
SPH.GeoADS2_DSD=SPH.DSD(2);
SPH.MDS1_DSD=SPH.DSD(3);
SPH.MDS2_DSD=SPH.DSD(4);
%for idsd=1:SPH.Num_Input_Files
%	k=k+1;
%	SPH.Inp_DSD(idsd)=ReadDSD(DSDs([288*(k-1)+1:288*(k-1)+288])');
%end

end % function

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       MATLAB CODE TO READ MET Geolocation ADS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% 02/2007 P. Poli   Rewrote entire code using IODD names

function Geolocation_ADS = MET_Read_Geo_ADS(fid,nbDSR_Geo)

Geolocation_ADS = [];

for idsr = 1:nbDSR_Geo,
    Geolocation_ADS.DSR(idsr).AMD_DateTime(1) = fread(fid, 1,'*int32');
    Geolocation_ADS.DSR(idsr).AMD_DateTime(2:3) = fread(fid, 2,'*uint32');
    Geolocation_ADS.DSR(idsr).AMD_Latitude = fread(fid, 1,'*int32','ieee-be');
    Geolocation_ADS.DSR(idsr).AMD_Longitude = fread(fid, 1,'*int32','ieee-be');
    Geolocation_ADS.DSR(idsr).AMD_zg = fread(fid, 1,'*int32','ieee-be');
end %for idsr

end %function

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       MATLAB CODE TO READ MET Met MDS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% 02/2007 P. Poli   Rewrote entire code using IODD names

function Met_MDS = MET_Read_Met_MDS(fid,nbDSR,Num_of_Model_Layers)

Met_MDS = [];

for idsr = 1:nbDSR,
	% Spare_1
	fseek(fid,2,'cof');
	Met_MDS.DSR(idsr).AMD_us = fread(fid, 1,'*int16','ieee-be');
	Met_MDS.DSR(idsr).AMD_vs = fread(fid, 1,'*int16','ieee-be');
	Met_MDS.DSR(idsr).AMD_ps = fread(fid, 1,'uint32','ieee-be');
	Met_MDS.DSR(idsr).AMD_err_ps = fread(fid, 1,'float64','ieee-be');
	Met_MDS.DSR(idsr).AMD_zs = fread(fid, 1,'int32','ieee-be');
	% Spare_2
	fseek(fid,2,'cof');
	for ilay=1:Num_of_Model_Layers,
		Met_MDS.DSR(idsr).List_of_Profile_Data(ilay).AMD_Validity_Flag = fread(fid, 1,'*int8','ieee-be');
		Met_MDS.DSR(idsr).List_of_Profile_Data(ilay).AMD_pbase = fread(fid, 1,'uint32','ieee-be');
		Met_MDS.DSR(idsr).List_of_Profile_Data(ilay).AMD_ptop = fread(fid, 1,'uint32','ieee-be');
		Met_MDS.DSR(idsr).List_of_Profile_Data(ilay).AMD_pnom = fread(fid, 1,'uint32','ieee-be');
		Met_MDS.DSR(idsr).List_of_Profile_Data(ilay).AMD_zbase = fread(fid, 1,'*int32','ieee-be');
		Met_MDS.DSR(idsr).List_of_Profile_Data(ilay).AMD_ztop = fread(fid, 1,'*int32','ieee-be');
		Met_MDS.DSR(idsr).List_of_Profile_Data(ilay).AMD_znom = fread(fid, 1,'*int32','ieee-be');
		Met_MDS.DSR(idsr).List_of_Profile_Data(ilay).AMD_T = fread(fid, 1,'uint16','ieee-be');
		Met_MDS.DSR(idsr).List_of_Profile_Data(ilay).AMD_err_T = fread(fid, 1,'uint16','ieee-be');
		Met_MDS.DSR(idsr).List_of_Profile_Data(ilay).AMD_u = fread(fid, 1,'*int16','ieee-be');
		Met_MDS.DSR(idsr).List_of_Profile_Data(ilay).AMD_v = fread(fid, 1,'*int16','ieee-be');
		% Spare_1
		fseek(fid,4,'cof');
		Met_MDS.DSR(idsr).List_of_Profile_Data(ilay).AMD_rh = fread(fid, 1,'uint8','ieee-be');
		Met_MDS.DSR(idsr).List_of_Profile_Data(ilay).AMD_err_rh = fread(fid, 1,'float64','ieee-be');
		Met_MDS.DSR(idsr).List_of_Profile_Data(ilay).AMD_q = fread(fid, 1,'float64','ieee-be');
		Met_MDS.DSR(idsr).List_of_Profile_Data(ilay).AMD_cc = fread(fid, 1,'uint8','ieee-be');
		Met_MDS.DSR(idsr).List_of_Profile_Data(ilay).AMD_clwc = fread(fid, 1,'float64','ieee-be');
		Met_MDS.DSR(idsr).List_of_Profile_Data(ilay).AMD_ciwc = fread(fid, 1,'float64','ieee-be');
	end
	% Spare_3
	fseek(fid,2,'cof');
end % idsr

end % function

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       MATLAB CODE TO READ RBC SPH              
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% HISTORY
%
% 04/2006 P. Poli    Original code (IODD AE_IF_ECMWF_L2BP_001_20060427)
% 07/2006 D. Tan     Adjust offsets by 1
% 02/2007 P. Poli    Use IODD variable names
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DOCUMENTATION
%
% The function ReadRBCSPH decodes the string of characters
% making up the L2B Rayleigh Brillouin Correction (RBC)
% Specific Product Header (SPH).
%
% The output is a structure containing the elements of 
% the L2B RBC SPH.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function SPH = ReadRBCSPH(SPH_chain);

SPH1=SPH_chain(1:822);

% decode variables contained in SPH

SPH.Sph_Descriptor=(SPH1(16+[1:28])');
%Spare_1
SPH.Ref_RBC_Suite=(SPH1(102+[1:20])');
SPH.Num_P=str2num(SPH1(130+[1:6])');
SPH.Num_T=str2num(SPH1(143+[1:6])');
SPH.Num_F=str2num(SPH1(156+[1:6])');
SPH.Num_FP=str2num(SPH1(170+[1:6])');
SPH.Num_fd=str2num(SPH1(184+[1:6])');
SPH.Num_RR=str2num(SPH1(198+[1:6])');
%Spare_2
SPH.P_min=str2num(SPH1(252+[1:11])');
SPH.P_max=str2num(SPH1(274+[1:11])');
SPH.T_min=str2num(SPH1(296+[1:6])');
SPH.T_max=str2num(SPH1(316+[1:6])');
SPH.Laser_Freq_Offset_Start=str2num(SPH1(354+[1:13])');
SPH.Laser_Freq_Offset_Stop=str2num(SPH1(396+[1:13])');
SPH.Total_Num_of_Observations=str2num(SPH1(441+[1:11])');
SPH.Total_Num_of_Measurements=str2num(SPH1(479+[1:11])');
SPH.Total_Num_of_Reference_Pulses=str2num(SPH1(521+[1:11])');
SPH.Total_Num_of_Corrupt_Mie_Meas=str2num(SPH1(563+[1:11])');
SPH.Total_Num_of_Corrupt_Ray_Meas=str2num(SPH1(605+[1:11])');
SPH.Total_Num_of_Corrupt_Mie_RefP=str2num(SPH1(647+[1:11])');
SPH.Total_Num_of_Corrupt_Ray_RefP=str2num(SPH1(689+[1:11])');
SPH.Average_Error_FP_Response_A=str2num(SPH1(729+[1:11])');
SPH.Average_Error_FP_Response_B=str2num(SPH1(769+[1:11])');
%Spare_3

% Read DSDs
DSDs=SPH_chain(823:end);
for k=1:2,
	SPH.DSD(k)=ReadDSD(DSDs([288*(k-1)+1:288*(k-1)+288])');
end
SPH.RBC_DS_DSD=SPH.DSD(1);
SPH.RBC_FP_DSD=SPH.DSD(2);

end %function

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DOCUMENTATION
%
% The function ReadCLMSPH decodes the string of characters
% making up the L2B Auxiliary Climatology file (CLM)
% Specific Product Header (SPH).
%
% The output is a structure containing the elements of 
% the L2B CLM SPH.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function SPH = ReadCLMSPH(SPH_chain);

SPH1=SPH_chain(1:486);

% decode variables contained in SPH

SPH.Sph_Descriptor=(SPH1(16+[1:28])');
%Spare_1
SPH.Ref_CLM_Suite=(SPH1(105+[1:50])');
%%Spare_2

% Read DSDs
DSDs=SPH_chain(199:end);
for k=1:1,
	SPH.DSD(k)=ReadDSD(DSDs([288*(k-1)+1:288*(k-1)+288])');
end
SPH.RBC_DS_DSD=SPH.DSD(1);

end %function

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       MATLAB CODE TO READ RBC Rayleigh-Brillouin dataset
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% HISTORY
%
% 04/2006 P. Poli    Original code (IODD AE_IF_ECMWF_L2BP_001_20060427)
% 05/2006 D. Tan     Account for one-byte shift in offset defn/use
% 07/2006 D. Tan     Upgrade to new format (NA, NB, Fint)
% 08/2006 D. Tan     IODD order for NA and NB
% 02/2007 P. Poli    Rewrote entire code using IODD names. Removed wrapper part.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% The function RBC_Read_RB_ADS reads the RB DS of a RBC file
% opened by the calling program and whose file id is used
% as input.
%
% The output is a structure containing the elements of 
% the Rayleigh Brillouin data set.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


function RB_ADS = RBC_Read_RB_ADS(fid,nbDSR,Num_P,Num_T,Num_F,Num_FP,Num_fd,Num_RR);

for idsr=1:nbDSR,
	for m=1:Num_P,
		RB_ADS.DSR(idsr).P_Grid(m) = fread(fid, 1,'uint32','ieee-be');
	end
	for i=1:Num_T,
		RB_ADS.DSR(idsr).T_Grid(i) = fread(fid, 1,'uint16','ieee-be');
	end
	for i=1:Num_F
		RB_ADS.DSR(idsr).F_Gridtmp(i) = fread(fid, 1,'int64','ieee-be');
	end
	for i=1:Num_P
		for j=1:Num_T
			for k=1:Num_F
				RB_ADS.DSR(idsr).Spec_Grid_PTF(i).Spec_Grid_PT(j).Spec_Grid_F(k) = ...
				fread(fid, 1,'float64','ieee-be');
			end
		end
	end
	for i=1:Num_FP
		RB_ADS.DSR(idsr).F_FP(i).val = fread(fid, 1,'int64','ieee-be');
	end
	for i=1:Num_FP
		RB_ADS.DSR(idsr).TA_FP(i).val = fread(fid, 1,'float64','ieee-be');
	end
	for i=1:Num_FP
		RB_ADS.DSR(idsr).TB_FP(i).val = fread(fid, 1,'float64','ieee-be');
	end
	for i=1:Num_fd
		RB_ADS.DSR(idsr).fd(i).val = fread(fid, 1,'int64','ieee-be');
	end
	for i=1:Num_RR
		RB_ADS.DSR(idsr).RR(i).val = fread(fid, 1,'float64','ieee-be');
	end
	for i=1:Num_P
		for j=1:Num_T
			for k=1:Num_RR
				RB_ADS.DSR(idsr).F_Calib_PTR(i).F_Calib_TR(j).F_Calib_R(k).val = ...
				fread(fid, 1,'float64','ieee-be');
			end
			for k=1:Num_RR
				RB_ADS.DSR(idsr).F_Calib_PTR(i).F_Calib_TR(j).F_Calib_R_error(k).val = ...
				fread(fid, 1,'float64','ieee-be');
			end
		end
	end
	for i=1:Num_P
		for j=1:Num_T
			for k=1:Num_fd
				RB_ADS.DSR(idsr).Nab_PTFd(i).Nab_TFd(j).Na_Fd(k).val = ...
	                        fread(fid, 1,'uint32','ieee-be');
			end
			for k=1:Num_fd
				RB_ADS.DSR(idsr).Nab_PTFd(i).Nab_TFd(j).Nb_Fd(k).val = ...
	                        fread(fid, 1,'uint32','ieee-be');
			end
		end
	end
	for i=1:Num_RR
		RB_ADS.DSR(idsr).Fint_R(i).val = fread(fid, 1,'int64','ieee-be');
	end
end % idsr

end % function

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       MATLAB CODE TO READ CLM Rayleigh-Brillouin dataset
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% HISTORY
%
% 04/2007 P. Poli    Original code (IODD AE_IF_ECMWF_L2BP_001_20070223)
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% The function CLM_Read_ADS reads the DS of a CLM file
% opened by the calling program and whose file id is used
% as input.
%
% The output is a structure containing the elements of 
% the climatology data set.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


function CLIM_ADS = CLM_Read_ADS(fid,nbDSR);

for idsr=1:nbDSR,
	CLIM_ADS.DSR(idsr).Num_DateTime_Ranges = fread(fid, 1,'*int16','ieee-be');
	for t=1:CLIM_ADS.DSR(idsr).Num_DateTime_Ranges,
	  CLIM_ADS.DSR(idsr).ClimDateLatLonAlt(t).StartDateTime(1)   = fread(fid, 1,'*int32');
          CLIM_ADS.DSR(idsr).ClimDateLatLonAlt(t).StartDateTime(2:3) = fread(fid, 2,'*uint32');
	  CLIM_ADS.DSR(idsr).ClimDateLatLonAlt(t).EndDateTime(1)     = fread(fid, 1,'*int32');
          CLIM_ADS.DSR(idsr).ClimDateLatLonAlt(t).EndDateTime(2:3)   = fread(fid, 2,'*uint32');
	  CLIM_ADS.DSR(idsr).ClimDateLatLonAlt(t).Num_Latitude_Ranges = fread(fid, 1,'*int16','ieee-be');
	  for i=1:CLIM_ADS.DSR(idsr).ClimDateLatLonAlt(t).Num_Latitude_Ranges
	    CLIM_ADS.DSR(idsr).ClimDateLatLonAlt(t).ClimLatLonAlt(i).StartLatitude = fread(fid, 1,'*int32','ieee-be');
	    CLIM_ADS.DSR(idsr).ClimDateLatLonAlt(t).ClimLatLonAlt(i).EndLatitude   = fread(fid, 1,'*int32','ieee-be');
	    CLIM_ADS.DSR(idsr).ClimDateLatLonAlt(t).ClimLatLonAlt(i).Num_Longitude_Ranges = fread(fid, 1,'*int16','ieee-be');
	    for j=1:CLIM_ADS.DSR(idsr).ClimDateLatLonAlt(t).ClimLatLonAlt(i).Num_Longitude_Ranges
	      CLIM_ADS.DSR(idsr).ClimDateLatLonAlt(t).ClimLatLonAlt(i).ClimLonAlt(j).StartLongitude = fread(fid, 1,'*int32','ieee-be');
	      CLIM_ADS.DSR(idsr).ClimDateLatLonAlt(t).ClimLatLonAlt(i).ClimLonAlt(j).EndLongitude   = fread(fid, 1,'*int32','ieee-be');
	      CLIM_ADS.DSR(idsr).ClimDateLatLonAlt(t).ClimLatLonAlt(i).ClimLonAlt(j).Num_Altitude_Ranges = fread(fid, 1,'*int16','ieee-be');
	      for k=1:CLIM_ADS.DSR(idsr).ClimDateLatLonAlt(t).ClimLatLonAlt(i).ClimLonAlt(j).Num_Altitude_Ranges
	        CLIM_ADS.DSR(idsr).ClimDateLatLonAlt(t).ClimLatLonAlt(i).ClimLonAlt(j).ClimAlt(k).StartAltitude = fread(fid, 1,'*int32','ieee-be');
		CLIM_ADS.DSR(idsr).ClimDateLatLonAlt(t).ClimLatLonAlt(i).ClimLonAlt(j).ClimAlt(k).EndAltitude   = fread(fid, 1,'*int32','ieee-be');
		CLIM_ADS.DSR(idsr).ClimDateLatLonAlt(t).ClimLatLonAlt(i).ClimLonAlt(j).ClimAlt(k).S             = fread(fid, 1,'*int32','ieee-be');
		CLIM_ADS.DSR(idsr).ClimDateLatLonAlt(t).ClimLatLonAlt(i).ClimLonAlt(j).ClimAlt(k).S_stdev       = fread(fid, 1,'*int32','ieee-be');
	      end
	    end
	  end
	end
end % idsr

end % function

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%	MATLAB CODE TO READ L2A GEOLOCATION DS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% HISTORY
%
% Oct 2006      ML Denneulin    Original code
% Dec 2006      ML Denneulin    Corrections of inconsistencies with the
%                               IODD
% Apr 2007      ML Denneulin    Insertion into ReadBinaryData
function GEO_ADS = L2a_Read_Geo_ADS(fid,nbObs,Max_Num_Prof)
for iObs=1:nbObs,
         
    % Read date and and time of observation
    GEO_ADS.Obs(iObs).Start_of_Obs_Time(1) = fread(fid,1,'*int32','ieee-be');
    GEO_ADS.Obs(iObs).Start_of_Obs_Time(2:3) = fread(fid,2,'*uint32','ieee-be');
    % Read the number of meaningful profiles of the current BRC
    GEO_ADS.Obs(iObs).N_Prof_Actual = fread(fid,1,'*int16','ieee-be');
    
    % Read profiles geolocations
    for jprof = 1:Max_Num_Prof,
        % Read the list of geolocation parameters for 24 heigth bins :
        % Latitude, longitude and altitude for the 
        % start, stop and center of gravity time of the observation,
        % los velocity, los elevation and los satellite velocity
        for jbin=1:24, 
            GEO_ADS.Obs(iObs).List_of_Profile_Geolocation.Profile_Geolocation(jprof)...
                .List_of_Profile_Height_Bin_Geolocation.Profile_Heigth_Bin_Geolocation(jbin)...
                .Latitude_Start = fread(fid,1,'*int32','ieee-be');
            
            GEO_ADS.Obs(iObs).List_of_Profile_Geolocation.Profile_Geolocation(jprof)...
                .List_of_Profile_Height_Bin_Geolocation.Profile_Heigth_Bin_Geolocation(jbin)...
                .Latitude_Stop = fread(fid,1,'*int32','ieee-be');
            
            GEO_ADS.Obs(iObs).List_of_Profile_Geolocation.Profile_Geolocation(jprof)...
                .List_of_Profile_Height_Bin_Geolocation.Profile_Heigth_Bin_Geolocation(jbin)...
                .Latitude_COG = fread(fid,1,'*int32','ieee-be');
            
            GEO_ADS.Obs(iObs).List_of_Profile_Geolocation.Profile_Geolocation(jprof)...
                .List_of_Profile_Height_Bin_Geolocation.Profile_Heigth_Bin_Geolocation(jbin)...
                .Longitude_Start = fread(fid,1,'*int32','ieee-be');
            
            GEO_ADS.Obs(iObs).List_of_Profile_Geolocation.Profile_Geolocation(jprof)...
                .List_of_Profile_Height_Bin_Geolocation.Profile_Heigth_Bin_Geolocation(jbin)...
                .Longitude_Stop = fread(fid,1,'*int32','ieee-be');
            
            GEO_ADS.Obs(iObs).List_of_Profile_Geolocation.Profile_Geolocation(jprof)...
                .List_of_Profile_Height_Bin_Geolocation.Profile_Heigth_Bin_Geolocation(jbin)...
                .Longitude_COG = fread(fid,1,'*int32','ieee-be');
            
            GEO_ADS.Obs(iObs).List_of_Profile_Geolocation.Profile_Geolocation(jprof)...
                .List_of_Profile_Height_Bin_Geolocation.Profile_Heigth_Bin_Geolocation(jbin)...
                .Altitude_Bottom = fread(fid,1,'*int32','ieee-be');
            
            GEO_ADS.Obs(iObs).List_of_Profile_Geolocation.Profile_Geolocation(jprof)...
                .List_of_Profile_Height_Bin_Geolocation.Profile_Heigth_Bin_Geolocation(jbin)...
                .Altitude_Top = fread(fid, 1,'*int32','ieee-be');
            
            GEO_ADS.Obs(iObs).List_of_Profile_Geolocation.Profile_Geolocation(jprof)...
                .List_of_Profile_Height_Bin_Geolocation.Profile_Heigth_Bin_Geolocation(jbin)...
                .Altitude_COG = fread(fid, 1,'*int32','ieee-be');
            
            GEO_ADS.Obs(iObs).List_of_Profile_Geolocation.Profile_Geolocation(jprof)...
                .List_of_Profile_Height_Bin_Geolocation.Profile_Heigth_Bin_Geolocation(jbin)...
                .LOS_Azimuth = fread(fid,1,'*float64','ieee-be');
            
            GEO_ADS.Obs(iObs).List_of_Profile_Geolocation.Profile_Geolocation(jprof)...
                .List_of_Profile_Height_Bin_Geolocation.Profile_Heigth_Bin_Geolocation(jbin)...
                .LOS_Elevation = fread(fid, 1,'*float64','ieee-be');
            
            GEO_ADS.Obs(iObs).List_of_Profile_Geolocation.Profile_Geolocation(jprof)...
                .List_of_Profile_Height_Bin_Geolocation.Profile_Heigth_Bin_Geolocation(jbin)...
                .LOS_Satellite_Velocity = fread(fid, 1,'*float64','ieee-be');
        end
        % Read geolocation (Latitude, Longitude, Altitude) of the intersection 
        % between the DEM and the LOS 
        GEO_ADS.Obs(iObs).List_of_Profile_Geolocation.Profile_Geolocation(jprof)...
            .Latitude_of_DEM_Intersection = fread(fid, 1,'*int32','ieee-be');
        
        GEO_ADS.Obs(iObs).List_of_Profile_Geolocation.Profile_Geolocation(jprof)...
            .Longitude_of_DEM_Intersection = fread(fid, 1,'*int32','ieee-be');
        
        GEO_ADS.Obs(iObs).List_of_Profile_Geolocation.Profile_Geolocation(jprof)...
            .Altitude_of_DEM_Intersection = fread(fid, 1,'*int32','ieee-be');
    end
    % Read the difference between the DEM and the WGS84 ellipsoid
    GEO_ADS.Obs(iObs).WGS84_to_Geoid_Altitude = fread(fid, 1, '*int32','ieee-be');

end

end %function

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%	MATLAB CODE TO READ L2A PRODUCT CONFIDENCE ADS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% HISTORY
%
% Oct 2006      ML Denneulin    Original code
% Dec 2006      ML Denneulin    Corrections of inconsistencies with the
%                               IODD
% Apr 2007      ML Denneulin    Insertion into ReadBinaryData
function PCD_ADS = L2a_Read_PCD_ADS(fid,nbObs,Max_Num_Prof)

for iObs=1:nbObs,    
    PCD_ADS.Obs(iObs).Start_of_Obs_Time(1) = fread(fid,1,'*int32','ieee-be');
    PCD_ADS.Obs(iObs).Start_of_Obs_Time(2:3) = fread(fid,2,'*uint32','ieee-be');
   % Read the number of measurements
    PCD_ADS.Obs(iObs).N_Meas = fread(fid, 1,'*int16','ieee-be');
    N_Meas = PCD_ADS.Obs(iObs).N_Meas;
    % Read the number of meaningful sets of optical profiles in the current
    % BRC
    PCD_ADS.Obs(iObs).N_Prof_Actual = fread(fid, 1,'*int16','ieee-be');
    
    % Read the value of L1B_Obs_Screening describing a problem which
    % prevent using this whole BRC
    PCD_ADS.Obs(iObs).L1B_Input_Screening.L1B_Obs_Screening = fread(fid, 1,'*uint8','ieee-be');

    % Read the values of L1B_Obs_Screening_Flags
    for kflag=1:40,
        PCD_ADS.Obs(iObs).L1B_Input_Screening.(sprintf('L1B_OBS_Screening_Flags_%d',kflag)) = fread(fid,1, '*ubit1','ieee-be');
    end

    for knmeas=1:N_Meas,
        PCD_ADS.Obs(iObs).L1B_Input_Screening.List_of_L1B_Mie_Meas_Screening...
            .L1B_Mie_Meas_Screening(knmeas).L1B_Mie_Meas_QC = ...
            fread(fid,1,'*uint16','ieee-be');
        
        PCD_ADS.Obs(iObs).L1B_Input_Screening.List_of_L1B_Mie_Meas_Screening...
            .L1B_Mie_Meas_Screening(knmeas).L1B_Mie_Meas_QC_Flags = ...
            fread(fid,1,'*ubit8','ieee-be');

        fseek(fid, 1,'cof');% Skip spare table 13
    end
   
    for knmeas=1:N_Meas,
        PCD_ADS.Obs(iObs).L1B_Input_Screening.List_of_L1B_Rayleigh_Meas_Screening...
            .L1B_Rayleigh_Meas_Screening(knmeas).L1B_Rayleigh_Meas_QC = ...
            fread(fid,1,'*uint16','ieee-be');
        
        PCD_ADS.Obs(iObs).L1B_Input_Screening.List_of_L1B_Rayleigh_Meas_Screening...
            .L1B_Mie_Rayleigh_Screening(knmeas).L1B_Rayleigh_Meas_QC_Flags = ...
            fread(fid,1,'*ubit8','ieee-be');
        fseek(fid, 1, 'cof');% Skip spare table 15
    end
    
    fseek(fid, 1, 'cof');% Skip spare table 11
    
   % L1B CAL SCREENING %%%%%%%%%%%%%%%%%%%

   PCD_ADS.Obs(iObs).L1B_Input_Screening.CAL_Screening.CAL_Valid = fread(fid, 1,'*ubit8','ieee-be');
   fseek(fid,5,'cof');% Skip spare table 17

   % L2B AMD SCREENING %%%%%%%%%%%%%%%%%%%
   PCD_ADS.Obs(iObs).L2B_AMD_Screening.L2B_AMD_Screening_QC = fread(fid, 1,'*uint8','ieee-be');
   PCD_ADS.Obs(iObs).L2B_AMD_Screening.L2B_AMD_Screening_QC_Flags = fread(fid, 1,'*ubit8','ieee-be');
   for k=1:N_Meas,
       PCD_ADS.Obs(iObs).L2B_AMD_Screening.L2B_AMD_Collocations(k).L2B_AMD_Collocation_QC...
           = fread(fid, 1,'*uint8','ieee-be');
       
       PCD_ADS.Obs(iObs).L2B_AMD_Screening.L2B_AMD_Collocations(k).L2B_AMD_Collocation_QC...
           = fread(fid, 1,'*ubit8','ieee-be');
       fseek(fid,1,'cof');% Skip spare table 20
   end
   fseek(fid, 20,'cof');%Skip spare table 18
   
   % L2A CLASSIFICATION QC
   for k=1:Max_Num_Prof,
        PCD_ADS.Obs(iObs).L2A_Classification_QC.L2A_Prof_Classification(k)...
            .L2A_Prof_Class_Flags = fread(fid,1, '*ubit8','ieee-be');
        PCD_ADS.Obs(iObs).L2A_Classification_QC.L2A_Prof_Classification(k)...
            .L2A_Prof_Class_Reliability = fread(fid,1, '*float64','ieee-be');
        fseek(fid,1,'cof');%Skip spare table 23
    end
    fseek(fid, 20,'cof');%Skip spare table 22
  
    % L2A PROCESSING QC
    for k=1:Max_Num_Prof,
        for lbin = 1:24,
            PCD_ADS.Obs(iObs).L2A_Processing_QC.L2A_Prof_Proc_QC(k).L2A_Prof_Proc_Bin_QC(lbin).data = fread(fid,40,'int8','ieee-be');
        end    
        fseek(fid,1,'cof');%Skip spare table 25
    end
    PCD_ADS.Obs(iObs).L2A_Processing_QC.Backgroung_High = fread(fid,1,'*uint8','ieee-be');
    fseek(fid, 20, 'cof');%Skip spare table 24
   fseek(fid, 20, 'cof');%Skip spare table 10

end

end %function

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%	MATLAB CODE TO READ L2A PRODUCT CONFIDENCE ADS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% HISTORY
%
% Oct 2006      ML Denneulin    Original code
% Dec 2006      ML Denneulin    Corrections of inconsistencies with the
%                               IODD
% Apr 2007      ML Denneulin    Insertion into ReadBinaryData
function OPT_ADS=L2a_Read_OPT_ADS(fid,nbObs,Max_Num_Prof)

for iObs=1:nbObs,   
    OPT_ADS.Obs(iObs).Start_of_Obs_Time(1) = fread(fid,1,'*int32','ieee-be');
    OPT_ADS.Obs(iObs).Start_of_Obs_Time(2:3) = fread(fid,2,'*uint32','ieee-be');
    OPT_ADS.Obs(iObs).N_Meas = fread(fid,1,'*int16','ieee-be');
    N_Meas = double(OPT_ADS.Obs(iObs).N_Meas);
    OPT_ADS.Obs(iObs).P = fread(fid, 1,'*int16','ieee-be');
    OPT_ADS.Obs(iObs).Maps_of_L1B_Measurements_Used.L1B_Measurement_Used = fread(fid, N_Meas*24, '*int8','ieee-be');
    OPT_ADS.Obs(iObs).List_of_L1_Measurement_Weights.L1B_Measurement_Weight = fread(fid, N_Meas*24, '*int16','ieee-be');
    
    for kprof = 1:Max_Num_Prof,
        OPT_ADS.Obs(iObs).List_of_Optical_Profiles.Optical_Profile(kprof).Algorithm = fread(fid, 3,'*char','ieee-be');
        OPT_ADS.Obs(iObs).List_of_Optical_Profiles.Optical_Profile(kprof).Prof_Type = fread(fid, 1, '*uint8','ieee-be');
        
        for lbin = 1:24,
            OPT_ADS.Obs(iObs).List_of_Optical_Profiles.Optical_Profile(kprof)...
                .List_of_Height_Bin_Opt.HeightBinOpt(lbin).ValidityFlag ...
                = fread(fid, 1, '*int8','ieee-be');
            
            OPT_ADS.Obs(iObs).List_of_Optical_Profiles.Optical_Profile(kprof)...
                .List_of_Height_Bin_Opt.HeightBinOpt(lbin).Reference_Pressure ...
                = fread(fid, 1, '*uint32','ieee-be');
            OPT_ADS.Obs(iObs).List_of_Optical_Profiles.Optical_Profile(kprof)...
                .List_of_Height_Bin_Opt.HeightBinOpt(lbin).Reference_Temperature ...
                = fread(fid, 1, '*uint16','ieee-be');
            OPT_ADS.Obs(iObs).List_of_Optical_Profiles.Optical_Profile(kprof)...
                .List_of_Height_Bin_Opt.HeightBinOpt(lbin).Reference_Hlos_Wind ...
                = fread(fid, 1, '*int16','ieee-be');
            OPT_ADS.Obs(iObs).List_of_Optical_Profiles.Optical_Profile(kprof)...
                .List_of_Height_Bin_Opt.HeightBinOpt(lbin).Opt_mol_back ...
                = fread(fid, 1, '*float64','ieee-be');
            OPT_ADS.Obs(iObs).List_of_Optical_Profiles.Optical_Profile(kprof)...
                .List_of_Height_Bin_Opt.HeightBinOpt(lbin).Opt_aer_back ...
                = fread(fid, 1, 'float64','ieee-be');
            OPT_ADS.Obs(iObs).List_of_Optical_Profiles.Optical_Profile(kprof)...
                .List_of_Height_Bin_Opt.HeightBinOpt(lbin).Opt_mol_ext ...
                = fread(fid, 1, '*float64','ieee-be');
            OPT_ADS.Obs(iObs).List_of_Optical_Profiles.Optical_Profile(kprof)...
                .List_of_Height_Bin_Opt.HeightBinOpt(lbin).Opt_aer_ext ...
                = fread(fid, 1, '*float64','ieee-be');
            OPT_ADS.Obs(iObs).List_of_Optical_Profiles.Optical_Profile(kprof)...
                .List_of_Height_Bin_Opt.HeightBinOpt(lbin).Scat_Ratio ...
                = fread(fid, 1, '*uint32','ieee-be');
            OPT_ADS.Obs(iObs).List_of_Optical_Profiles.Optical_Profile(kprof)...
                .List_of_Height_Bin_Opt.HeightBinOpt(lbin).Comp_aer_bck_to_ext ...
                = fread(fid, 1, '*int8','ieee-be');
            OPT_ADS.Obs(iObs).List_of_Optical_Profiles.Optical_Profile(kprof)...
                .List_of_Height_Bin_Opt.HeightBinOpt(lbin).Aer_bck_to_ext ...
                = fread(fid, 1, '*uint16','ieee-be');
            OPT_ADS.Obs(iObs).List_of_Optical_Profiles.Optical_Profile(kprof)...
                .List_of_Height_Bin_Opt.HeightBinOpt(lbin).Opt_mol_bck_err ...
                = fread(fid, 1, '*float64','ieee-be');
            OPT_ADS.Obs(iObs).List_of_Optical_Profiles.Optical_Profile(kprof)...
                .List_of_Height_Bin_Opt.HeightBinOpt(lbin).Opt_aer_bck_err ...
                = fread(fid, 1, '*float64','ieee-be');
            OPT_ADS.Obs(iObs).List_of_Optical_Profiles.Optical_Profile(kprof)...
                .List_of_Height_Bin_Opt.HeightBinOpt(lbin).Opt_mol_ext_err ...
                = fread(fid, 1, '*float64','ieee-be');
            OPT_ADS.Obs(iObs).List_of_Optical_Profiles.Optical_Profile(kprof)...
                .List_of_Height_Bin_Opt.HeightBinOpt(lbin).Opt_aer_ext_err ...
                = fread(fid, 1, '*float64','ieee-be');
            OPT_ADS.Obs(iObs).List_of_Optical_Profiles.Optical_Profile(kprof)...
                .List_of_Height_Bin_Opt.HeightBinOpt(lbin).Scat_Ratio_err ...
                = fread(fid, 1, '*uint32','ieee-be');
            OPT_ADS.Obs(iObs).List_of_Optical_Profiles.Optical_Profile(kprof)...
                .List_of_Height_Bin_Opt.HeightBinOpt(lbin).Aer_bck_to_ext_err...
                = fread(fid, 1, '*uint16','ieee-be');
            OPT_ADS.Obs(iObs).List_of_Optical_Profiles.Optical_Profile(kprof)...
                .List_of_Height_Bin_Opt.HeightBinOpt(lbin).Integration_Length ...
                = fread(fid, 1, '*uint32','ieee-be');
    
        end
    end

end
end %function


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%	MATLAB CODE TO READ AUX CAL ADS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% HISTORY
%
% Sep 2006   ML Denneulin    Original code
% Feb 2007   ML Denneulin    the size of the vector Num_FP is now reported on 5 digits



function CAL_ADS = Cal_Read_ADS(fid,Num_DSR,Num_P,Num_T,Num_fd,Num_FP);


CAL_ADS.P_Grid = fread(fid,Num_P,'uint32');
CAL_ADS.T_Grid = fread(fid,Num_T,'uint16');
CAL_ADS.fd_Grid = fread(fid,Num_fd, 'int64');
CAL_ADS.F_FP = fread(fid,Num_FP,'int64');
% Read double Fabry-Perot transmissions
CAL_ADS.Ta_FP = fread(fid,Num_FP,'float64');
CAL_ADS.Tb_FP = fread(fid,Num_FP,'float64');

% Read Fizeau transmission function
CAL_ADS.Tmie_FP = fread(fid,Num_FP,'float64');

%Read calibration coefficients for Rayleigh channel
CAL_ADS.Cal_Coeff_Ray.Kray = fread(fid,1,'float64');
for iP=1:Num_P,
    for iT= 1 :Num_T,
        for ifd=1:Num_fd,
            CAL_ADS.Cal_Coeff_Ray.List_of_Cal_Coeff_Ray_P.List_of_Cal_Coeff_Ray_PT(iP)...
                .List_of_Cal_Coeff_RayPTfd(iT).C1(ifd) = fread(fid,1,'float64');

            CAL_ADS.Cal_Coeff_Ray.List_of_Cal_Coeff_Ray_P.List_of_Cal_Coeff_Ray_PT(iP)...
                .List_of_Cal_Coeff_RayPTfd(iT).C4(ifd) = fread(fid,1,'float64');
        end
    end
end
%Read calibration coefficients for Mie channel
CAL_ADS.Cal_Coeff_Mie.Kmie = fread(fid,1,'float64');

for jfd=1:Num_fd,
    CAL_ADS.Cal_Coeff_Mie.List_of_Cal_Coeff_Mie_fd.C2(jfd) = fread(fid,1,'float64');
    CAL_ADS.Cal_Coeff_Mie.List_of_Cal_Coeff_Mie_fd.C3(jfd)=fread(fid,1,'float64');
end

end %function

